<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8" />
    <title>Welcome to Arize Phoenix</title>
  </head>
  <body>
    <h1>Welcome to Arize Phoenix!</h1>
    <p>Hi {{ name }}, please click the link below to get started:</p>
    <a id="welcome-url" href="{{ welcome_url }}">Get Started</a>
  </body>
</html>
