var oOe=Object.defineProperty;var BB=e=>{throw TypeError(e)};var sOe=(e,t,n)=>t in e?oOe(e,t,{enumerable:!0,configurable:!0,writable:!0,value:n}):e[t]=n;var P_=(e,t,n)=>sOe(e,typeof t!="symbol"?t+"":t,n),Y5=(e,t,n)=>t.has(e)||BB("Cannot "+n);var Ot=(e,t,n)=>(Y5(e,t,"read from private field"),n?n.call(e):t.get(e)),Yi=(e,t,n)=>t.has(e)?BB("Cannot add the same private member more than once"):t instanceof WeakSet?t.add(e):t.set(e,n),wn=(e,t,n,r)=>(Y5(e,t,"write to private field"),r?r.call(e,n):t.set(e,n),n),Zn=(e,t,n)=>(Y5(e,t,"access private method"),n);import{W as cOe,R as fOe,O as JO,P as e3,S as dOe,a as G5,b as hOe,c as pOe,B as mOe,C as gOe,L as vOe,d as hJ,N as yOe,A as bOe,e as QB,f as OC,g as OOe,U as xOe,V as ci,h as ao,i as SOe,T as _Oe,D as EOe,j as DOe,k as $Oe,l as COe,m as zB,Q as qB,M as gO,n as vO,o as wOe,p as ROe,I as AOe,F as e9,q as t9,r as VO,s as TOe,t as FF,u as pJ,v as kOe,w as VB,x as UB,y as POe,z as j3,E as MOe,G as FOe,H as mJ,J as NOe,K as IOe,X as LOe,Y as BOe}from"./vendor-three-CmB8bl_y.js";function QOe(e,t){for(var n=0;n<t.length;n++){const r=t[n];if(typeof r!="string"&&!Array.isArray(r)){for(const i in r)if(i!=="default"&&!(i in e)){const a=Object.getOwnPropertyDescriptor(r,i);a&&Object.defineProperty(e,i,a.get?a:{enumerable:!0,get:()=>r[i]})}}}return Object.freeze(Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}))}var Xd=typeof globalThis<"u"?globalThis:typeof window<"u"?window:typeof global<"u"?global:typeof self<"u"?self:{};function da(e){return e&&e.__esModule&&Object.prototype.hasOwnProperty.call(e,"default")?e.default:e}function zOe(e){if(Object.prototype.hasOwnProperty.call(e,"__esModule"))return e;var t=e.default;if(typeof t=="function"){var n=function r(){return this instanceof r?Reflect.construct(t,arguments,this.constructor):t.apply(this,arguments)};n.prototype=t.prototype}else n={};return Object.defineProperty(n,"__esModule",{value:!0}),Object.keys(e).forEach(function(r){var i=Object.getOwnPropertyDescriptor(e,r);Object.defineProperty(n,r,i.get?i:{enumerable:!0,get:function(){return e[r]}})}),n}var K5={exports:{}},H4={};/**
 * @license React
 * react-jsx-runtime.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var jB;function qOe(){if(jB)return H4;jB=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.fragment");function n(r,i,a){var u=null;if(a!==void 0&&(u=""+a),i.key!==void 0&&(u=""+i.key),"key"in i){a={};for(var l in i)l!=="key"&&(a[l]=i[l])}else a=i;return i=a.ref,{$$typeof:e,type:r,key:u,ref:i!==void 0?i:null,props:a}}return H4.Fragment=t,H4.jsx=n,H4.jsxs=n,H4}var ZB;function VOe(){return ZB||(ZB=1,K5.exports=qOe()),K5.exports}var Qr=VOe(),W5={exports:{}},or={};/**
 * @license React
 * react.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var HB;function UOe(){if(HB)return or;HB=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.portal"),n=Symbol.for("react.fragment"),r=Symbol.for("react.strict_mode"),i=Symbol.for("react.profiler"),a=Symbol.for("react.consumer"),u=Symbol.for("react.context"),l=Symbol.for("react.forward_ref"),o=Symbol.for("react.suspense"),s=Symbol.for("react.memo"),c=Symbol.for("react.lazy"),d=Symbol.for("react.activity"),h=Symbol.iterator;function m(G){return G===null||typeof G!="object"?null:(G=h&&G[h]||G["@@iterator"],typeof G=="function"?G:null)}var y={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},v=Object.assign,S={};function _(G,X,B){this.props=G,this.context=X,this.refs=S,this.updater=B||y}_.prototype.isReactComponent={},_.prototype.setState=function(G,X){if(typeof G!="object"&&typeof G!="function"&&G!=null)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,G,X,"setState")},_.prototype.forceUpdate=function(G){this.updater.enqueueForceUpdate(this,G,"forceUpdate")};function E(){}E.prototype=_.prototype;function $(G,X,B){this.props=G,this.context=X,this.refs=S,this.updater=B||y}var k=$.prototype=new E;k.constructor=$,v(k,_.prototype),k.isPureReactComponent=!0;var A=Array.isArray;function w(){}var R={H:null,A:null,T:null,S:null},F=Object.prototype.hasOwnProperty;function I(G,X,B){var ne=B.ref;return{$$typeof:e,type:G,key:X,ref:ne!==void 0?ne:null,props:B}}function U(G,X){return I(G.type,X,G.props)}function q(G){return typeof G=="object"&&G!==null&&G.$$typeof===e}function V(G){var X={"=":"=0",":":"=2"};return"$"+G.replace(/[=:]/g,function(B){return X[B]})}var N=/\/+/g;function j(G,X){return typeof G=="object"&&G!==null&&G.key!=null?V(""+G.key):X.toString(36)}function W(G){switch(G.status){case"fulfilled":return G.value;case"rejected":throw G.reason;default:switch(typeof G.status=="string"?G.then(w,w):(G.status="pending",G.then(function(X){G.status==="pending"&&(G.status="fulfilled",G.value=X)},function(X){G.status==="pending"&&(G.status="rejected",G.reason=X)})),G.status){case"fulfilled":return G.value;case"rejected":throw G.reason}}throw G}function K(G,X,B,ne,oe){var ie=typeof G;(ie==="undefined"||ie==="boolean")&&(G=null);var de=!1;if(G===null)de=!0;else switch(ie){case"bigint":case"string":case"number":de=!0;break;case"object":switch(G.$$typeof){case e:case t:de=!0;break;case c:return de=G._init,K(de(G._payload),X,B,ne,oe)}}if(de)return oe=oe(G),de=ne===""?"."+j(G,0):ne,A(oe)?(B="",de!=null&&(B=de.replace(N,"$&/")+"/"),K(oe,X,B,"",function(me){return me})):oe!=null&&(q(oe)&&(oe=U(oe,B+(oe.key==null||G&&G.key===oe.key?"":(""+oe.key).replace(N,"$&/")+"/")+de)),X.push(oe)),1;de=0;var ge=ne===""?".":ne+":";if(A(G))for(var ye=0;ye<G.length;ye++)ne=G[ye],ie=ge+j(ne,ye),de+=K(ne,X,B,ie,oe);else if(ye=m(G),typeof ye=="function")for(G=ye.call(G),ye=0;!(ne=G.next()).done;)ne=ne.value,ie=ge+j(ne,ye++),de+=K(ne,X,B,ie,oe);else if(ie==="object"){if(typeof G.then=="function")return K(W(G),X,B,ne,oe);throw X=String(G),Error("Objects are not valid as a React child (found: "+(X==="[object Object]"?"object with keys {"+Object.keys(G).join(", ")+"}":X)+"). If you meant to render a collection of children, use an array instead.")}return de}function J(G,X,B){if(G==null)return G;var ne=[],oe=0;return K(G,ne,"","",function(ie){return X.call(B,ie,oe++)}),ne}function te(G){if(G._status===-1){var X=G._result;X=X(),X.then(function(B){(G._status===0||G._status===-1)&&(G._status=1,G._result=B)},function(B){(G._status===0||G._status===-1)&&(G._status=2,G._result=B)}),G._status===-1&&(G._status=0,G._result=X)}if(G._status===1)return G._result.default;throw G._result}var H=typeof reportError=="function"?reportError:function(G){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var X=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof G=="object"&&G!==null&&typeof G.message=="string"?String(G.message):String(G),error:G});if(!window.dispatchEvent(X))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",G);return}console.error(G)},M={map:J,forEach:function(G,X,B){J(G,function(){X.apply(this,arguments)},B)},count:function(G){var X=0;return J(G,function(){X++}),X},toArray:function(G){return J(G,function(X){return X})||[]},only:function(G){if(!q(G))throw Error("React.Children.only expected to receive a single React element child.");return G}};return or.Activity=d,or.Children=M,or.Component=_,or.Fragment=n,or.Profiler=i,or.PureComponent=$,or.StrictMode=r,or.Suspense=o,or.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=R,or.__COMPILER_RUNTIME={__proto__:null,c:function(G){return R.H.useMemoCache(G)}},or.cache=function(G){return function(){return G.apply(null,arguments)}},or.cacheSignal=function(){return null},or.cloneElement=function(G,X,B){if(G==null)throw Error("The argument must be a React element, but you passed "+G+".");var ne=v({},G.props),oe=G.key;if(X!=null)for(ie in X.key!==void 0&&(oe=""+X.key),X)!F.call(X,ie)||ie==="key"||ie==="__self"||ie==="__source"||ie==="ref"&&X.ref===void 0||(ne[ie]=X[ie]);var ie=arguments.length-2;if(ie===1)ne.children=B;else if(1<ie){for(var de=Array(ie),ge=0;ge<ie;ge++)de[ge]=arguments[ge+2];ne.children=de}return I(G.type,oe,ne)},or.createContext=function(G){return G={$$typeof:u,_currentValue:G,_currentValue2:G,_threadCount:0,Provider:null,Consumer:null},G.Provider=G,G.Consumer={$$typeof:a,_context:G},G},or.createElement=function(G,X,B){var ne,oe={},ie=null;if(X!=null)for(ne in X.key!==void 0&&(ie=""+X.key),X)F.call(X,ne)&&ne!=="key"&&ne!=="__self"&&ne!=="__source"&&(oe[ne]=X[ne]);var de=arguments.length-2;if(de===1)oe.children=B;else if(1<de){for(var ge=Array(de),ye=0;ye<de;ye++)ge[ye]=arguments[ye+2];oe.children=ge}if(G&&G.defaultProps)for(ne in de=G.defaultProps,de)oe[ne]===void 0&&(oe[ne]=de[ne]);return I(G,ie,oe)},or.createRef=function(){return{current:null}},or.forwardRef=function(G){return{$$typeof:l,render:G}},or.isValidElement=q,or.lazy=function(G){return{$$typeof:c,_payload:{_status:-1,_result:G},_init:te}},or.memo=function(G,X){return{$$typeof:s,type:G,compare:X===void 0?null:X}},or.startTransition=function(G){var X=R.T,B={};R.T=B;try{var ne=G(),oe=R.S;oe!==null&&oe(B,ne),typeof ne=="object"&&ne!==null&&typeof ne.then=="function"&&ne.then(w,H)}catch(ie){H(ie)}finally{X!==null&&B.types!==null&&(X.types=B.types),R.T=X}},or.unstable_useCacheRefresh=function(){return R.H.useCacheRefresh()},or.use=function(G){return R.H.use(G)},or.useActionState=function(G,X,B){return R.H.useActionState(G,X,B)},or.useCallback=function(G,X){return R.H.useCallback(G,X)},or.useContext=function(G){return R.H.useContext(G)},or.useDebugValue=function(){},or.useDeferredValue=function(G,X){return R.H.useDeferredValue(G,X)},or.useEffect=function(G,X){return R.H.useEffect(G,X)},or.useEffectEvent=function(G){return R.H.useEffectEvent(G)},or.useId=function(){return R.H.useId()},or.useImperativeHandle=function(G,X,B){return R.H.useImperativeHandle(G,X,B)},or.useInsertionEffect=function(G,X){return R.H.useInsertionEffect(G,X)},or.useLayoutEffect=function(G,X){return R.H.useLayoutEffect(G,X)},or.useMemo=function(G,X){return R.H.useMemo(G,X)},or.useOptimistic=function(G,X){return R.H.useOptimistic(G,X)},or.useReducer=function(G,X,B){return R.H.useReducer(G,X,B)},or.useRef=function(G){return R.H.useRef(G)},or.useState=function(G){return R.H.useState(G)},or.useSyncExternalStore=function(G,X,B){return R.H.useSyncExternalStore(G,X,B)},or.useTransition=function(){return R.H.useTransition()},or.version="19.2.0",or}var XB;function ir(){return XB||(XB=1,W5.exports=UOe()),W5.exports}var P=ir();const be=da(P),wD=QOe({__proto__:null,default:be},[P]);function jOe(e){if(e.sheet)return e.sheet;for(var t=0;t<document.styleSheets.length;t++)if(document.styleSheets[t].ownerNode===e)return document.styleSheets[t]}function ZOe(e){var t=document.createElement("style");return t.setAttribute("data-emotion",e.key),e.nonce!==void 0&&t.setAttribute("nonce",e.nonce),t.appendChild(document.createTextNode("")),t.setAttribute("data-s",""),t}var HOe=(function(){function e(n){var r=this;this._insertTag=function(i){var a;r.tags.length===0?r.insertionPoint?a=r.insertionPoint.nextSibling:r.prepend?a=r.container.firstChild:a=r.before:a=r.tags[r.tags.length-1].nextSibling,r.container.insertBefore(i,a),r.tags.push(i)},this.isSpeedy=n.speedy===void 0?!0:n.speedy,this.tags=[],this.ctr=0,this.nonce=n.nonce,this.key=n.key,this.container=n.container,this.prepend=n.prepend,this.insertionPoint=n.insertionPoint,this.before=null}var t=e.prototype;return t.hydrate=function(r){r.forEach(this._insertTag)},t.insert=function(r){this.ctr%(this.isSpeedy?65e3:1)===0&&this._insertTag(ZOe(this));var i=this.tags[this.tags.length-1];if(this.isSpeedy){var a=jOe(i);try{a.insertRule(r,a.cssRules.length)}catch{}}else i.appendChild(document.createTextNode(r));this.ctr++},t.flush=function(){this.tags.forEach(function(r){var i;return(i=r.parentNode)==null?void 0:i.removeChild(r)}),this.tags=[],this.ctr=0},e})(),gs="-ms-",RD="-moz-",Pi="-webkit-",gJ="comm",NF="rule",IF="decl",XOe="@import",vJ="@keyframes",YOe="@layer",GOe=Math.abs,xC=String.fromCharCode,KOe=Object.assign;function WOe(e,t){return Io(e,0)^45?(((t<<2^Io(e,0))<<2^Io(e,1))<<2^Io(e,2))<<2^Io(e,3):0}function yJ(e){return e.trim()}function JOe(e,t){return(e=t.exec(e))?e[0]:e}function Mi(e,t,n){return e.replace(t,n)}function n9(e,t){return e.indexOf(t)}function Io(e,t){return e.charCodeAt(t)|0}function sx(e,t,n){return e.slice(t,n)}function h0(e){return e.length}function LF(e){return e.length}function M_(e,t){return t.push(e),e}function e3e(e,t){return e.map(t).join("")}var SC=1,E3=1,bJ=0,of=0,bl=0,Z3="";function _C(e,t,n,r,i,a,u){return{value:e,root:t,parent:n,type:r,props:i,children:a,line:SC,column:E3,length:u,return:""}}function X4(e,t){return KOe(_C("",null,null,"",null,null,0),e,{length:-e.length},t)}function t3e(){return bl}function n3e(){return bl=of>0?Io(Z3,--of):0,E3--,bl===10&&(E3=1,SC--),bl}function jf(){return bl=of<bJ?Io(Z3,of++):0,E3++,bl===10&&(E3=1,SC++),bl}function E0(){return Io(Z3,of)}function ZE(){return of}function uS(e,t){return sx(Z3,e,t)}function cx(e){switch(e){case 0:case 9:case 10:case 13:case 32:return 5;case 33:case 43:case 44:case 47:case 62:case 64:case 126:case 59:case 123:case 125:return 4;case 58:return 3;case 34:case 39:case 40:case 91:return 2;case 41:case 93:return 1}return 0}function OJ(e){return SC=E3=1,bJ=h0(Z3=e),of=0,[]}function xJ(e){return Z3="",e}function HE(e){return yJ(uS(of-1,r9(e===91?e+2:e===40?e+1:e)))}function r3e(e){for(;(bl=E0())&&bl<33;)jf();return cx(e)>2||cx(bl)>3?"":" "}function i3e(e,t){for(;--t&&jf()&&!(bl<48||bl>102||bl>57&&bl<65||bl>70&&bl<97););return uS(e,ZE()+(t<6&&E0()==32&&jf()==32))}function r9(e){for(;jf();)switch(bl){case e:return of;case 34:case 39:e!==34&&e!==39&&r9(bl);break;case 40:e===41&&r9(e);break;case 92:jf();break}return of}function a3e(e,t){for(;jf()&&e+bl!==57;)if(e+bl===84&&E0()===47)break;return"/*"+uS(t,of-1)+"*"+xC(e===47?e:jf())}function u3e(e){for(;!cx(E0());)jf();return uS(e,of)}function l3e(e){return xJ(XE("",null,null,null,[""],e=OJ(e),0,[0],e))}function XE(e,t,n,r,i,a,u,l,o){for(var s=0,c=0,d=u,h=0,m=0,y=0,v=1,S=1,_=1,E=0,$="",k=i,A=a,w=r,R=$;S;)switch(y=E,E=jf()){case 40:if(y!=108&&Io(R,d-1)==58){n9(R+=Mi(HE(E),"&","&\f"),"&\f")!=-1&&(_=-1);break}case 34:case 39:case 91:R+=HE(E);break;case 9:case 10:case 13:case 32:R+=r3e(y);break;case 92:R+=i3e(ZE()-1,7);continue;case 47:switch(E0()){case 42:case 47:M_(o3e(a3e(jf(),ZE()),t,n),o);break;default:R+="/"}break;case 123*v:l[s++]=h0(R)*_;case 125*v:case 59:case 0:switch(E){case 0:case 125:S=0;case 59+c:_==-1&&(R=Mi(R,/\f/g,"")),m>0&&h0(R)-d&&M_(m>32?GB(R+";",r,n,d-1):GB(Mi(R," ","")+";",r,n,d-2),o);break;case 59:R+=";";default:if(M_(w=YB(R,t,n,s,c,i,l,$,k=[],A=[],d),a),E===123)if(c===0)XE(R,t,w,w,k,a,d,l,A);else switch(h===99&&Io(R,3)===110?100:h){case 100:case 108:case 109:case 115:XE(e,w,w,r&&M_(YB(e,w,w,0,0,i,l,$,i,k=[],d),A),i,A,d,l,r?k:A);break;default:XE(R,w,w,w,[""],A,0,l,A)}}s=c=m=0,v=_=1,$=R="",d=u;break;case 58:d=1+h0(R),m=y;default:if(v<1){if(E==123)--v;else if(E==125&&v++==0&&n3e()==125)continue}switch(R+=xC(E),E*v){case 38:_=c>0?1:(R+="\f",-1);break;case 44:l[s++]=(h0(R)-1)*_,_=1;break;case 64:E0()===45&&(R+=HE(jf())),h=E0(),c=d=h0($=R+=u3e(ZE())),E++;break;case 45:y===45&&h0(R)==2&&(v=0)}}return a}function YB(e,t,n,r,i,a,u,l,o,s,c){for(var d=i-1,h=i===0?a:[""],m=LF(h),y=0,v=0,S=0;y<r;++y)for(var _=0,E=sx(e,d+1,d=GOe(v=u[y])),$=e;_<m;++_)($=yJ(v>0?h[_]+" "+E:Mi(E,/&\f/g,h[_])))&&(o[S++]=$);return _C(e,t,n,i===0?NF:l,o,s,c)}function o3e(e,t,n){return _C(e,t,n,gJ,xC(t3e()),sx(e,2,-2),0)}function GB(e,t,n,r){return _C(e,t,n,IF,sx(e,0,r),sx(e,r+1,-1),r)}function t3(e,t){for(var n="",r=LF(e),i=0;i<r;i++)n+=t(e[i],i,e,t)||"";return n}function s3e(e,t,n,r){switch(e.type){case YOe:if(e.children.length)break;case XOe:case IF:return e.return=e.return||e.value;case gJ:return"";case vJ:return e.return=e.value+"{"+t3(e.children,r)+"}";case NF:e.value=e.props.join(",")}return h0(n=t3(e.children,r))?e.return=e.value+"{"+n+"}":""}function c3e(e){var t=LF(e);return function(n,r,i,a){for(var u="",l=0;l<t;l++)u+=e[l](n,r,i,a)||"";return u}}function f3e(e){return function(t){t.root||(t=t.return)&&e(t)}}function d3e(e){var t=Object.create(null);return function(n){return t[n]===void 0&&(t[n]=e(n)),t[n]}}var h3e=function(t,n,r){for(var i=0,a=0;i=a,a=E0(),i===38&&a===12&&(n[r]=1),!cx(a);)jf();return uS(t,of)},p3e=function(t,n){var r=-1,i=44;do switch(cx(i)){case 0:i===38&&E0()===12&&(n[r]=1),t[r]+=h3e(of-1,n,r);break;case 2:t[r]+=HE(i);break;case 4:if(i===44){t[++r]=E0()===58?"&\f":"",n[r]=t[r].length;break}default:t[r]+=xC(i)}while(i=jf());return t},m3e=function(t,n){return xJ(p3e(OJ(t),n))},KB=new WeakMap,g3e=function(t){if(!(t.type!=="rule"||!t.parent||t.length<1)){for(var n=t.value,r=t.parent,i=t.column===r.column&&t.line===r.line;r.type!=="rule";)if(r=r.parent,!r)return;if(!(t.props.length===1&&n.charCodeAt(0)!==58&&!KB.get(r))&&!i){KB.set(t,!0);for(var a=[],u=m3e(n,a),l=r.props,o=0,s=0;o<u.length;o++)for(var c=0;c<l.length;c++,s++)t.props[s]=a[o]?u[o].replace(/&\f/g,l[c]):l[c]+" "+u[o]}}},v3e=function(t){if(t.type==="decl"){var n=t.value;n.charCodeAt(0)===108&&n.charCodeAt(2)===98&&(t.return="",t.value="")}};function SJ(e,t){switch(WOe(e,t)){case 5103:return Pi+"print-"+e+e;case 5737:case 4201:case 3177:case 3433:case 1641:case 4457:case 2921:case 5572:case 6356:case 5844:case 3191:case 6645:case 3005:case 6391:case 5879:case 5623:case 6135:case 4599:case 4855:case 4215:case 6389:case 5109:case 5365:case 5621:case 3829:return Pi+e+e;case 5349:case 4246:case 4810:case 6968:case 2756:return Pi+e+RD+e+gs+e+e;case 6828:case 4268:return Pi+e+gs+e+e;case 6165:return Pi+e+gs+"flex-"+e+e;case 5187:return Pi+e+Mi(e,/(\w+).+(:[^]+)/,Pi+"box-$1$2"+gs+"flex-$1$2")+e;case 5443:return Pi+e+gs+"flex-item-"+Mi(e,/flex-|-self/,"")+e;case 4675:return Pi+e+gs+"flex-line-pack"+Mi(e,/align-content|flex-|-self/,"")+e;case 5548:return Pi+e+gs+Mi(e,"shrink","negative")+e;case 5292:return Pi+e+gs+Mi(e,"basis","preferred-size")+e;case 6060:return Pi+"box-"+Mi(e,"-grow","")+Pi+e+gs+Mi(e,"grow","positive")+e;case 4554:return Pi+Mi(e,/([^-])(transform)/g,"$1"+Pi+"$2")+e;case 6187:return Mi(Mi(Mi(e,/(zoom-|grab)/,Pi+"$1"),/(image-set)/,Pi+"$1"),e,"")+e;case 5495:case 3959:return Mi(e,/(image-set\([^]*)/,Pi+"$1$`$1");case 4968:return Mi(Mi(e,/(.+:)(flex-)?(.*)/,Pi+"box-pack:$3"+gs+"flex-pack:$3"),/s.+-b[^;]+/,"justify")+Pi+e+e;case 4095:case 3583:case 4068:case 2532:return Mi(e,/(.+)-inline(.+)/,Pi+"$1$2")+e;case 8116:case 7059:case 5753:case 5535:case 5445:case 5701:case 4933:case 4677:case 5533:case 5789:case 5021:case 4765:if(h0(e)-1-t>6)switch(Io(e,t+1)){case 109:if(Io(e,t+4)!==45)break;case 102:return Mi(e,/(.+:)(.+)-([^]+)/,"$1"+Pi+"$2-$3$1"+RD+(Io(e,t+3)==108?"$3":"$2-$3"))+e;case 115:return~n9(e,"stretch")?SJ(Mi(e,"stretch","fill-available"),t)+e:e}break;case 4949:if(Io(e,t+1)!==115)break;case 6444:switch(Io(e,h0(e)-3-(~n9(e,"!important")&&10))){case 107:return Mi(e,":",":"+Pi)+e;case 101:return Mi(e,/(.+:)([^;!]+)(;|!.+)?/,"$1"+Pi+(Io(e,14)===45?"inline-":"")+"box$3$1"+Pi+"$2$3$1"+gs+"$2box$3")+e}break;case 5936:switch(Io(e,t+11)){case 114:return Pi+e+gs+Mi(e,/[svh]\w+-[tblr]{2}/,"tb")+e;case 108:return Pi+e+gs+Mi(e,/[svh]\w+-[tblr]{2}/,"tb-rl")+e;case 45:return Pi+e+gs+Mi(e,/[svh]\w+-[tblr]{2}/,"lr")+e}return Pi+e+gs+e+e}return e}var y3e=function(t,n,r,i){if(t.length>-1&&!t.return)switch(t.type){case IF:t.return=SJ(t.value,t.length);break;case vJ:return t3([X4(t,{value:Mi(t.value,"@","@"+Pi)})],i);case NF:if(t.length)return e3e(t.props,function(a){switch(JOe(a,/(::plac\w+|:read-\w+)/)){case":read-only":case":read-write":return t3([X4(t,{props:[Mi(a,/:(read-\w+)/,":"+RD+"$1")]})],i);case"::placeholder":return t3([X4(t,{props:[Mi(a,/:(plac\w+)/,":"+Pi+"input-$1")]}),X4(t,{props:[Mi(a,/:(plac\w+)/,":"+RD+"$1")]}),X4(t,{props:[Mi(a,/:(plac\w+)/,gs+"input-$1")]})],i)}return""})}},b3e=[y3e],O3e=function(t){var n=t.key;if(n==="css"){var r=document.querySelectorAll("style[data-emotion]:not([data-s])");Array.prototype.forEach.call(r,function(v){var S=v.getAttribute("data-emotion");S.indexOf(" ")!==-1&&(document.head.appendChild(v),v.setAttribute("data-s",""))})}var i=t.stylisPlugins||b3e,a={},u,l=[];u=t.container||document.head,Array.prototype.forEach.call(document.querySelectorAll('style[data-emotion^="'+n+' "]'),function(v){for(var S=v.getAttribute("data-emotion").split(" "),_=1;_<S.length;_++)a[S[_]]=!0;l.push(v)});var o,s=[g3e,v3e];{var c,d=[s3e,f3e(function(v){c.insert(v)})],h=c3e(s.concat(i,d)),m=function(S){return t3(l3e(S),h)};o=function(S,_,E,$){c=E,m(S?S+"{"+_.styles+"}":_.styles),$&&(y.inserted[_.name]=!0)}}var y={key:n,sheet:new HOe({key:n,container:u,nonce:t.nonce,speedy:t.speedy,prepend:t.prepend,insertionPoint:t.insertionPoint}),nonce:t.nonce,inserted:a,registered:{},insert:o};return y.sheet.hydrate(l),y};function sf(){return sf=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)({}).hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},sf.apply(null,arguments)}var J5={exports:{}},qi={};/** @license React v16.13.1
 * react-is.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var WB;function x3e(){if(WB)return qi;WB=1;var e=typeof Symbol=="function"&&Symbol.for,t=e?Symbol.for("react.element"):60103,n=e?Symbol.for("react.portal"):60106,r=e?Symbol.for("react.fragment"):60107,i=e?Symbol.for("react.strict_mode"):60108,a=e?Symbol.for("react.profiler"):60114,u=e?Symbol.for("react.provider"):60109,l=e?Symbol.for("react.context"):60110,o=e?Symbol.for("react.async_mode"):60111,s=e?Symbol.for("react.concurrent_mode"):60111,c=e?Symbol.for("react.forward_ref"):60112,d=e?Symbol.for("react.suspense"):60113,h=e?Symbol.for("react.suspense_list"):60120,m=e?Symbol.for("react.memo"):60115,y=e?Symbol.for("react.lazy"):60116,v=e?Symbol.for("react.block"):60121,S=e?Symbol.for("react.fundamental"):60117,_=e?Symbol.for("react.responder"):60118,E=e?Symbol.for("react.scope"):60119;function $(A){if(typeof A=="object"&&A!==null){var w=A.$$typeof;switch(w){case t:switch(A=A.type,A){case o:case s:case r:case a:case i:case d:return A;default:switch(A=A&&A.$$typeof,A){case l:case c:case y:case m:case u:return A;default:return w}}case n:return w}}}function k(A){return $(A)===s}return qi.AsyncMode=o,qi.ConcurrentMode=s,qi.ContextConsumer=l,qi.ContextProvider=u,qi.Element=t,qi.ForwardRef=c,qi.Fragment=r,qi.Lazy=y,qi.Memo=m,qi.Portal=n,qi.Profiler=a,qi.StrictMode=i,qi.Suspense=d,qi.isAsyncMode=function(A){return k(A)||$(A)===o},qi.isConcurrentMode=k,qi.isContextConsumer=function(A){return $(A)===l},qi.isContextProvider=function(A){return $(A)===u},qi.isElement=function(A){return typeof A=="object"&&A!==null&&A.$$typeof===t},qi.isForwardRef=function(A){return $(A)===c},qi.isFragment=function(A){return $(A)===r},qi.isLazy=function(A){return $(A)===y},qi.isMemo=function(A){return $(A)===m},qi.isPortal=function(A){return $(A)===n},qi.isProfiler=function(A){return $(A)===a},qi.isStrictMode=function(A){return $(A)===i},qi.isSuspense=function(A){return $(A)===d},qi.isValidElementType=function(A){return typeof A=="string"||typeof A=="function"||A===r||A===s||A===a||A===i||A===d||A===h||typeof A=="object"&&A!==null&&(A.$$typeof===y||A.$$typeof===m||A.$$typeof===u||A.$$typeof===l||A.$$typeof===c||A.$$typeof===S||A.$$typeof===_||A.$$typeof===E||A.$$typeof===v)},qi.typeOf=$,qi}var JB;function S3e(){return JB||(JB=1,J5.exports=x3e()),J5.exports}var e6,eQ;function _3e(){if(eQ)return e6;eQ=1;var e=S3e(),t={childContextTypes:!0,contextType:!0,contextTypes:!0,defaultProps:!0,displayName:!0,getDefaultProps:!0,getDerivedStateFromError:!0,getDerivedStateFromProps:!0,mixins:!0,propTypes:!0,type:!0},n={name:!0,length:!0,prototype:!0,caller:!0,callee:!0,arguments:!0,arity:!0},r={$$typeof:!0,render:!0,defaultProps:!0,displayName:!0,propTypes:!0},i={$$typeof:!0,compare:!0,defaultProps:!0,displayName:!0,propTypes:!0,type:!0},a={};a[e.ForwardRef]=r,a[e.Memo]=i;function u(y){return e.isMemo(y)?i:a[y.$$typeof]||t}var l=Object.defineProperty,o=Object.getOwnPropertyNames,s=Object.getOwnPropertySymbols,c=Object.getOwnPropertyDescriptor,d=Object.getPrototypeOf,h=Object.prototype;function m(y,v,S){if(typeof v!="string"){if(h){var _=d(v);_&&_!==h&&m(y,_,S)}var E=o(v);s&&(E=E.concat(s(v)));for(var $=u(y),k=u(v),A=0;A<E.length;++A){var w=E[A];if(!n[w]&&!(S&&S[w])&&!(k&&k[w])&&!($&&$[w])){var R=c(v,w);try{l(y,w,R)}catch{}}}}return y}return e6=m,e6}_3e();var E3e=!0;function D3e(e,t,n){var r="";return n.split(" ").forEach(function(i){e[i]!==void 0?t.push(e[i]+";"):i&&(r+=i+" ")}),r}var _J=function(t,n,r){var i=t.key+"-"+n.name;(r===!1||E3e===!1)&&t.registered[i]===void 0&&(t.registered[i]=n.styles)},EJ=function(t,n,r){_J(t,n,r);var i=t.key+"-"+n.name;if(t.inserted[n.name]===void 0){var a=n;do t.insert(n===a?"."+i:"",a,t.sheet,!0),a=a.next;while(a!==void 0)}};function $3e(e){for(var t=0,n,r=0,i=e.length;i>=4;++r,i-=4)n=e.charCodeAt(r)&255|(e.charCodeAt(++r)&255)<<8|(e.charCodeAt(++r)&255)<<16|(e.charCodeAt(++r)&255)<<24,n=(n&65535)*1540483477+((n>>>16)*59797<<16),n^=n>>>24,t=(n&65535)*1540483477+((n>>>16)*59797<<16)^(t&65535)*1540483477+((t>>>16)*59797<<16);switch(i){case 3:t^=(e.charCodeAt(r+2)&255)<<16;case 2:t^=(e.charCodeAt(r+1)&255)<<8;case 1:t^=e.charCodeAt(r)&255,t=(t&65535)*1540483477+((t>>>16)*59797<<16)}return t^=t>>>13,t=(t&65535)*1540483477+((t>>>16)*59797<<16),((t^t>>>15)>>>0).toString(36)}var C3e={animationIterationCount:1,aspectRatio:1,borderImageOutset:1,borderImageSlice:1,borderImageWidth:1,boxFlex:1,boxFlexGroup:1,boxOrdinalGroup:1,columnCount:1,columns:1,flex:1,flexGrow:1,flexPositive:1,flexShrink:1,flexNegative:1,flexOrder:1,gridRow:1,gridRowEnd:1,gridRowSpan:1,gridRowStart:1,gridColumn:1,gridColumnEnd:1,gridColumnSpan:1,gridColumnStart:1,msGridRow:1,msGridRowSpan:1,msGridColumn:1,msGridColumnSpan:1,fontWeight:1,lineHeight:1,opacity:1,order:1,orphans:1,scale:1,tabSize:1,widows:1,zIndex:1,zoom:1,WebkitLineClamp:1,fillOpacity:1,floodOpacity:1,stopOpacity:1,strokeDasharray:1,strokeDashoffset:1,strokeMiterlimit:1,strokeOpacity:1,strokeWidth:1},w3e=/[A-Z]|^ms/g,R3e=/_EMO_([^_]+?)_([^]*?)_EMO_/g,DJ=function(t){return t.charCodeAt(1)===45},tQ=function(t){return t!=null&&typeof t!="boolean"},t6=d3e(function(e){return DJ(e)?e:e.replace(w3e,"-$&").toLowerCase()}),nQ=function(t,n){switch(t){case"animation":case"animationName":if(typeof n=="string")return n.replace(R3e,function(r,i,a){return p0={name:i,styles:a,next:p0},i})}return C3e[t]!==1&&!DJ(t)&&typeof n=="number"&&n!==0?n+"px":n};function fx(e,t,n){if(n==null)return"";var r=n;if(r.__emotion_styles!==void 0)return r;switch(typeof n){case"boolean":return"";case"object":{var i=n;if(i.anim===1)return p0={name:i.name,styles:i.styles,next:p0},i.name;var a=n;if(a.styles!==void 0){var u=a.next;if(u!==void 0)for(;u!==void 0;)p0={name:u.name,styles:u.styles,next:p0},u=u.next;var l=a.styles+";";return l}return A3e(e,t,n)}case"function":{if(e!==void 0){var o=p0,s=n(e);return p0=o,fx(e,t,s)}break}}var c=n;return c}function A3e(e,t,n){var r="";if(Array.isArray(n))for(var i=0;i<n.length;i++)r+=fx(e,t,n[i])+";";else for(var a in n){var u=n[a];if(typeof u!="object"){var l=u;tQ(l)&&(r+=t6(a)+":"+nQ(a,l)+";")}else if(Array.isArray(u)&&typeof u[0]=="string"&&t==null)for(var o=0;o<u.length;o++)tQ(u[o])&&(r+=t6(a)+":"+nQ(a,u[o])+";");else{var s=fx(e,t,u);switch(a){case"animation":case"animationName":{r+=t6(a)+":"+s+";";break}default:r+=a+"{"+s+"}"}}}return r}var rQ=/label:\s*([^\s;{]+)\s*(;|$)/g,p0;function BF(e,t,n){if(e.length===1&&typeof e[0]=="object"&&e[0]!==null&&e[0].styles!==void 0)return e[0];var r=!0,i="";p0=void 0;var a=e[0];if(a==null||a.raw===void 0)r=!1,i+=fx(n,t,a);else{var u=a;i+=u[0]}for(var l=1;l<e.length;l++)if(i+=fx(n,t,e[l]),r){var o=a;i+=o[l]}rQ.lastIndex=0;for(var s="",c;(c=rQ.exec(i))!==null;)s+="-"+c[1];var d=$3e(i)+s;return{name:d,styles:i,next:p0}}var T3e=function(t){return t()},$J=wD.useInsertionEffect?wD.useInsertionEffect:!1,k3e=$J||T3e,iQ=$J||P.useLayoutEffect,CJ=P.createContext(typeof HTMLElement<"u"?O3e({key:"css"}):null);CJ.Provider;var wJ=function(t){return P.forwardRef(function(n,r){var i=P.useContext(CJ);return t(n,i,r)})},RJ=P.createContext({}),lS={}.hasOwnProperty,i9="__EMOTION_TYPE_PLEASE_DO_NOT_USE__",QF=function(t,n){var r={};for(var i in n)lS.call(n,i)&&(r[i]=n[i]);return r[i9]=t,r},P3e=function(t){var n=t.cache,r=t.serialized,i=t.isStringTag;return _J(n,r,i),k3e(function(){return EJ(n,r,i)}),null},M3e=wJ(function(e,t,n){var r=e.css;typeof r=="string"&&t.registered[r]!==void 0&&(r=t.registered[r]);var i=e[i9],a=[r],u="";typeof e.className=="string"?u=D3e(t.registered,a,e.className):e.className!=null&&(u=e.className+" ");var l=BF(a,void 0,P.useContext(RJ));u+=t.key+"-"+l.name;var o={};for(var s in e)lS.call(e,s)&&s!=="css"&&s!==i9&&(o[s]=e[s]);return o.className=u,n&&(o.ref=n),P.createElement(P.Fragment,null,P.createElement(P3e,{cache:t,serialized:l,isStringTag:typeof i=="string"}),P.createElement(i,o))}),zF=M3e,trt=Qr.Fragment,nrt=function(t,n,r){return lS.call(n,"css")?Qr.jsx(zF,QF(t,n),r):Qr.jsx(t,n,r)},rrt=function(t,n,r){return lS.call(n,"css")?Qr.jsxs(zF,QF(t,n),r):Qr.jsxs(t,n,r)},n6={exports:{}},Y4={},r6={exports:{}},i6={};/**
 * @license React
 * scheduler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var aQ;function F3e(){return aQ||(aQ=1,(function(e){function t(K,J){var te=K.length;K.push(J);e:for(;0<te;){var H=te-1>>>1,M=K[H];if(0<i(M,J))K[H]=J,K[te]=M,te=H;else break e}}function n(K){return K.length===0?null:K[0]}function r(K){if(K.length===0)return null;var J=K[0],te=K.pop();if(te!==J){K[0]=te;e:for(var H=0,M=K.length,G=M>>>1;H<G;){var X=2*(H+1)-1,B=K[X],ne=X+1,oe=K[ne];if(0>i(B,te))ne<M&&0>i(oe,B)?(K[H]=oe,K[ne]=te,H=ne):(K[H]=B,K[X]=te,H=X);else if(ne<M&&0>i(oe,te))K[H]=oe,K[ne]=te,H=ne;else break e}}return J}function i(K,J){var te=K.sortIndex-J.sortIndex;return te!==0?te:K.id-J.id}if(e.unstable_now=void 0,typeof performance=="object"&&typeof performance.now=="function"){var a=performance;e.unstable_now=function(){return a.now()}}else{var u=Date,l=u.now();e.unstable_now=function(){return u.now()-l}}var o=[],s=[],c=1,d=null,h=3,m=!1,y=!1,v=!1,S=!1,_=typeof setTimeout=="function"?setTimeout:null,E=typeof clearTimeout=="function"?clearTimeout:null,$=typeof setImmediate<"u"?setImmediate:null;function k(K){for(var J=n(s);J!==null;){if(J.callback===null)r(s);else if(J.startTime<=K)r(s),J.sortIndex=J.expirationTime,t(o,J);else break;J=n(s)}}function A(K){if(v=!1,k(K),!y)if(n(o)!==null)y=!0,w||(w=!0,V());else{var J=n(s);J!==null&&W(A,J.startTime-K)}}var w=!1,R=-1,F=5,I=-1;function U(){return S?!0:!(e.unstable_now()-I<F)}function q(){if(S=!1,w){var K=e.unstable_now();I=K;var J=!0;try{e:{y=!1,v&&(v=!1,E(R),R=-1),m=!0;var te=h;try{t:{for(k(K),d=n(o);d!==null&&!(d.expirationTime>K&&U());){var H=d.callback;if(typeof H=="function"){d.callback=null,h=d.priorityLevel;var M=H(d.expirationTime<=K);if(K=e.unstable_now(),typeof M=="function"){d.callback=M,k(K),J=!0;break t}d===n(o)&&r(o),k(K)}else r(o);d=n(o)}if(d!==null)J=!0;else{var G=n(s);G!==null&&W(A,G.startTime-K),J=!1}}break e}finally{d=null,h=te,m=!1}J=void 0}}finally{J?V():w=!1}}}var V;if(typeof $=="function")V=function(){$(q)};else if(typeof MessageChannel<"u"){var N=new MessageChannel,j=N.port2;N.port1.onmessage=q,V=function(){j.postMessage(null)}}else V=function(){_(q,0)};function W(K,J){R=_(function(){K(e.unstable_now())},J)}e.unstable_IdlePriority=5,e.unstable_ImmediatePriority=1,e.unstable_LowPriority=4,e.unstable_NormalPriority=3,e.unstable_Profiling=null,e.unstable_UserBlockingPriority=2,e.unstable_cancelCallback=function(K){K.callback=null},e.unstable_forceFrameRate=function(K){0>K||125<K?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):F=0<K?Math.floor(1e3/K):5},e.unstable_getCurrentPriorityLevel=function(){return h},e.unstable_next=function(K){switch(h){case 1:case 2:case 3:var J=3;break;default:J=h}var te=h;h=J;try{return K()}finally{h=te}},e.unstable_requestPaint=function(){S=!0},e.unstable_runWithPriority=function(K,J){switch(K){case 1:case 2:case 3:case 4:case 5:break;default:K=3}var te=h;h=K;try{return J()}finally{h=te}},e.unstable_scheduleCallback=function(K,J,te){var H=e.unstable_now();switch(typeof te=="object"&&te!==null?(te=te.delay,te=typeof te=="number"&&0<te?H+te:H):te=H,K){case 1:var M=-1;break;case 2:M=250;break;case 5:M=1073741823;break;case 4:M=1e4;break;default:M=5e3}return M=te+M,K={id:c++,callback:J,priorityLevel:K,startTime:te,expirationTime:M,sortIndex:-1},te>H?(K.sortIndex=te,t(s,K),n(o)===null&&K===n(s)&&(v?(E(R),R=-1):v=!0,W(A,te-H))):(K.sortIndex=M,t(o,K),y||m||(y=!0,w||(w=!0,V()))),K},e.unstable_shouldYield=U,e.unstable_wrapCallback=function(K){var J=h;return function(){var te=h;h=J;try{return K.apply(this,arguments)}finally{h=te}}}})(i6)),i6}var uQ;function N3e(){return uQ||(uQ=1,r6.exports=F3e()),r6.exports}var a6={exports:{}},ps={};/**
 * @license React
 * react-dom.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var lQ;function I3e(){if(lQ)return ps;lQ=1;var e=ir();function t(o){var s="https://react.dev/errors/"+o;if(1<arguments.length){s+="?args[]="+encodeURIComponent(arguments[1]);for(var c=2;c<arguments.length;c++)s+="&args[]="+encodeURIComponent(arguments[c])}return"Minified React error #"+o+"; visit "+s+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function n(){}var r={d:{f:n,r:function(){throw Error(t(522))},D:n,C:n,L:n,m:n,X:n,S:n,M:n},p:0,findDOMNode:null},i=Symbol.for("react.portal");function a(o,s,c){var d=3<arguments.length&&arguments[3]!==void 0?arguments[3]:null;return{$$typeof:i,key:d==null?null:""+d,children:o,containerInfo:s,implementation:c}}var u=e.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function l(o,s){if(o==="font")return"";if(typeof s=="string")return s==="use-credentials"?s:""}return ps.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=r,ps.createPortal=function(o,s){var c=2<arguments.length&&arguments[2]!==void 0?arguments[2]:null;if(!s||s.nodeType!==1&&s.nodeType!==9&&s.nodeType!==11)throw Error(t(299));return a(o,s,null,c)},ps.flushSync=function(o){var s=u.T,c=r.p;try{if(u.T=null,r.p=2,o)return o()}finally{u.T=s,r.p=c,r.d.f()}},ps.preconnect=function(o,s){typeof o=="string"&&(s?(s=s.crossOrigin,s=typeof s=="string"?s==="use-credentials"?s:"":void 0):s=null,r.d.C(o,s))},ps.prefetchDNS=function(o){typeof o=="string"&&r.d.D(o)},ps.preinit=function(o,s){if(typeof o=="string"&&s&&typeof s.as=="string"){var c=s.as,d=l(c,s.crossOrigin),h=typeof s.integrity=="string"?s.integrity:void 0,m=typeof s.fetchPriority=="string"?s.fetchPriority:void 0;c==="style"?r.d.S(o,typeof s.precedence=="string"?s.precedence:void 0,{crossOrigin:d,integrity:h,fetchPriority:m}):c==="script"&&r.d.X(o,{crossOrigin:d,integrity:h,fetchPriority:m,nonce:typeof s.nonce=="string"?s.nonce:void 0})}},ps.preinitModule=function(o,s){if(typeof o=="string")if(typeof s=="object"&&s!==null){if(s.as==null||s.as==="script"){var c=l(s.as,s.crossOrigin);r.d.M(o,{crossOrigin:c,integrity:typeof s.integrity=="string"?s.integrity:void 0,nonce:typeof s.nonce=="string"?s.nonce:void 0})}}else s==null&&r.d.M(o)},ps.preload=function(o,s){if(typeof o=="string"&&typeof s=="object"&&s!==null&&typeof s.as=="string"){var c=s.as,d=l(c,s.crossOrigin);r.d.L(o,c,{crossOrigin:d,integrity:typeof s.integrity=="string"?s.integrity:void 0,nonce:typeof s.nonce=="string"?s.nonce:void 0,type:typeof s.type=="string"?s.type:void 0,fetchPriority:typeof s.fetchPriority=="string"?s.fetchPriority:void 0,referrerPolicy:typeof s.referrerPolicy=="string"?s.referrerPolicy:void 0,imageSrcSet:typeof s.imageSrcSet=="string"?s.imageSrcSet:void 0,imageSizes:typeof s.imageSizes=="string"?s.imageSizes:void 0,media:typeof s.media=="string"?s.media:void 0})}},ps.preloadModule=function(o,s){if(typeof o=="string")if(s){var c=l(s.as,s.crossOrigin);r.d.m(o,{as:typeof s.as=="string"&&s.as!=="script"?s.as:void 0,crossOrigin:c,integrity:typeof s.integrity=="string"?s.integrity:void 0})}else r.d.m(o)},ps.requestFormReset=function(o){r.d.r(o)},ps.unstable_batchedUpdates=function(o,s){return o(s)},ps.useFormState=function(o,s,c){return u.H.useFormState(o,s,c)},ps.useFormStatus=function(){return u.H.useHostTransitionStatus()},ps.version="19.2.0",ps}var oQ;function AJ(){if(oQ)return a6.exports;oQ=1;function e(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(t){console.error(t)}}return e(),a6.exports=I3e(),a6.exports}/**
 * @license React
 * react-dom-client.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var sQ;function L3e(){if(sQ)return Y4;sQ=1;var e=N3e(),t=ir(),n=AJ();function r(f){var p="https://react.dev/errors/"+f;if(1<arguments.length){p+="?args[]="+encodeURIComponent(arguments[1]);for(var O=2;O<arguments.length;O++)p+="&args[]="+encodeURIComponent(arguments[O])}return"Minified React error #"+f+"; visit "+p+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function i(f){return!(!f||f.nodeType!==1&&f.nodeType!==9&&f.nodeType!==11)}function a(f){var p=f,O=f;if(f.alternate)for(;p.return;)p=p.return;else{f=p;do p=f,(p.flags&4098)!==0&&(O=p.return),f=p.return;while(f)}return p.tag===3?O:null}function u(f){if(f.tag===13){var p=f.memoizedState;if(p===null&&(f=f.alternate,f!==null&&(p=f.memoizedState)),p!==null)return p.dehydrated}return null}function l(f){if(f.tag===31){var p=f.memoizedState;if(p===null&&(f=f.alternate,f!==null&&(p=f.memoizedState)),p!==null)return p.dehydrated}return null}function o(f){if(a(f)!==f)throw Error(r(188))}function s(f){var p=f.alternate;if(!p){if(p=a(f),p===null)throw Error(r(188));return p!==f?null:f}for(var O=f,D=p;;){var Q=O.return;if(Q===null)break;var Y=Q.alternate;if(Y===null){if(D=Q.return,D!==null){O=D;continue}break}if(Q.child===Y.child){for(Y=Q.child;Y;){if(Y===O)return o(Q),f;if(Y===D)return o(Q),p;Y=Y.sibling}throw Error(r(188))}if(O.return!==D.return)O=Q,D=Y;else{for(var ce=!1,_e=Q.child;_e;){if(_e===O){ce=!0,O=Q,D=Y;break}if(_e===D){ce=!0,D=Q,O=Y;break}_e=_e.sibling}if(!ce){for(_e=Y.child;_e;){if(_e===O){ce=!0,O=Y,D=Q;break}if(_e===D){ce=!0,D=Y,O=Q;break}_e=_e.sibling}if(!ce)throw Error(r(189))}}if(O.alternate!==D)throw Error(r(190))}if(O.tag!==3)throw Error(r(188));return O.stateNode.current===O?f:p}function c(f){var p=f.tag;if(p===5||p===26||p===27||p===6)return f;for(f=f.child;f!==null;){if(p=c(f),p!==null)return p;f=f.sibling}return null}var d=Object.assign,h=Symbol.for("react.element"),m=Symbol.for("react.transitional.element"),y=Symbol.for("react.portal"),v=Symbol.for("react.fragment"),S=Symbol.for("react.strict_mode"),_=Symbol.for("react.profiler"),E=Symbol.for("react.consumer"),$=Symbol.for("react.context"),k=Symbol.for("react.forward_ref"),A=Symbol.for("react.suspense"),w=Symbol.for("react.suspense_list"),R=Symbol.for("react.memo"),F=Symbol.for("react.lazy"),I=Symbol.for("react.activity"),U=Symbol.for("react.memo_cache_sentinel"),q=Symbol.iterator;function V(f){return f===null||typeof f!="object"?null:(f=q&&f[q]||f["@@iterator"],typeof f=="function"?f:null)}var N=Symbol.for("react.client.reference");function j(f){if(f==null)return null;if(typeof f=="function")return f.$$typeof===N?null:f.displayName||f.name||null;if(typeof f=="string")return f;switch(f){case v:return"Fragment";case _:return"Profiler";case S:return"StrictMode";case A:return"Suspense";case w:return"SuspenseList";case I:return"Activity"}if(typeof f=="object")switch(f.$$typeof){case y:return"Portal";case $:return f.displayName||"Context";case E:return(f._context.displayName||"Context")+".Consumer";case k:var p=f.render;return f=f.displayName,f||(f=p.displayName||p.name||"",f=f!==""?"ForwardRef("+f+")":"ForwardRef"),f;case R:return p=f.displayName||null,p!==null?p:j(f.type)||"Memo";case F:p=f._payload,f=f._init;try{return j(f(p))}catch{}}return null}var W=Array.isArray,K=t.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,J=n.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,te={pending:!1,data:null,method:null,action:null},H=[],M=-1;function G(f){return{current:f}}function X(f){0>M||(f.current=H[M],H[M]=null,M--)}function B(f,p){M++,H[M]=f.current,f.current=p}var ne=G(null),oe=G(null),ie=G(null),de=G(null);function ge(f,p){switch(B(ie,p),B(oe,f),B(ne,null),p.nodeType){case 9:case 11:f=(f=p.documentElement)&&(f=f.namespaceURI)?ob(f):0;break;default:if(f=p.tagName,p=p.namespaceURI)p=ob(p),f=ds(p,f);else switch(f){case"svg":f=1;break;case"math":f=2;break;default:f=0}}X(ne),B(ne,f)}function ye(){X(ne),X(oe),X(ie)}function me(f){f.memoizedState!==null&&B(de,f);var p=ne.current,O=ds(p,f.type);p!==O&&(B(oe,f),B(ne,O))}function Ee(f){oe.current===f&&(X(ne),X(oe)),de.current===f&&(X(de),cv._currentValue=te)}var se,pe;function ve(f){if(se===void 0)try{throw Error()}catch(O){var p=O.stack.trim().match(/\n( *(at )?)/);se=p&&p[1]||"",pe=-1<O.stack.indexOf(`
    at`)?" (<anonymous>)":-1<O.stack.indexOf("@")?"@unknown:0:0":""}return`
`+se+f+pe}var we=!1;function Me(f,p){if(!f||we)return"";we=!0;var O=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var D={DetermineComponentFrameRoot:function(){try{if(p){var xt=function(){throw Error()};if(Object.defineProperty(xt.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(xt,[])}catch(ft){var ot=ft}Reflect.construct(f,[],xt)}else{try{xt.call()}catch(ft){ot=ft}f.call(xt.prototype)}}else{try{throw Error()}catch(ft){ot=ft}(xt=f())&&typeof xt.catch=="function"&&xt.catch(function(){})}}catch(ft){if(ft&&ot&&typeof ft.stack=="string")return[ft.stack,ot.stack]}return[null,null]}};D.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var Q=Object.getOwnPropertyDescriptor(D.DetermineComponentFrameRoot,"name");Q&&Q.configurable&&Object.defineProperty(D.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var Y=D.DetermineComponentFrameRoot(),ce=Y[0],_e=Y[1];if(ce&&_e){var Le=ce.split(`
`),ut=_e.split(`
`);for(Q=D=0;D<Le.length&&!Le[D].includes("DetermineComponentFrameRoot");)D++;for(;Q<ut.length&&!ut[Q].includes("DetermineComponentFrameRoot");)Q++;if(D===Le.length||Q===ut.length)for(D=Le.length-1,Q=ut.length-1;1<=D&&0<=Q&&Le[D]!==ut[Q];)Q--;for(;1<=D&&0<=Q;D--,Q--)if(Le[D]!==ut[Q]){if(D!==1||Q!==1)do if(D--,Q--,0>Q||Le[D]!==ut[Q]){var pt=`
`+Le[D].replace(" at new "," at ");return f.displayName&&pt.includes("<anonymous>")&&(pt=pt.replace("<anonymous>",f.displayName)),pt}while(1<=D&&0<=Q);break}}}finally{we=!1,Error.prepareStackTrace=O}return(O=f?f.displayName||f.name:"")?ve(O):""}function Re(f,p){switch(f.tag){case 26:case 27:case 5:return ve(f.type);case 16:return ve("Lazy");case 13:return f.child!==p&&p!==null?ve("Suspense Fallback"):ve("Suspense");case 19:return ve("SuspenseList");case 0:case 15:return Me(f.type,!1);case 11:return Me(f.type.render,!1);case 1:return Me(f.type,!0);case 31:return ve("Activity");default:return""}}function et(f){try{var p="",O=null;do p+=Re(f,O),O=f,f=f.return;while(f);return p}catch(D){return`
Error generating stack: `+D.message+`
`+D.stack}}var tt=Object.prototype.hasOwnProperty,Oe=e.unstable_scheduleCallback,$e=e.unstable_cancelCallback,Ae=e.unstable_shouldYield,He=e.unstable_requestPaint,je=e.unstable_now,Ye=e.unstable_getCurrentPriorityLevel,Ne=e.unstable_ImmediatePriority,ze=e.unstable_UserBlockingPriority,at=e.unstable_NormalPriority,mt=e.unstable_LowPriority,Mt=e.unstable_IdlePriority,bt=e.log,zt=e.unstable_setDisableYieldValue,qt=null,Xt=null;function nn(f){if(typeof bt=="function"&&zt(f),Xt&&typeof Xt.setStrictMode=="function")try{Xt.setStrictMode(qt,f)}catch{}}var Fe=Math.clz32?Math.clz32:Qt,ht=Math.log,kt=Math.LN2;function Qt(f){return f>>>=0,f===0?32:31-(ht(f)/kt|0)|0}var en=256,In=262144,rn=4194304;function Xn(f){var p=f&42;if(p!==0)return p;switch(f&-f){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:return 128;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:return f&261888;case 262144:case 524288:case 1048576:case 2097152:return f&3932160;case 4194304:case 8388608:case 16777216:case 33554432:return f&62914560;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return f}}function br(f,p,O){var D=f.pendingLanes;if(D===0)return 0;var Q=0,Y=f.suspendedLanes,ce=f.pingedLanes;f=f.warmLanes;var _e=D&134217727;return _e!==0?(D=_e&~Y,D!==0?Q=Xn(D):(ce&=_e,ce!==0?Q=Xn(ce):O||(O=_e&~f,O!==0&&(Q=Xn(O))))):(_e=D&~Y,_e!==0?Q=Xn(_e):ce!==0?Q=Xn(ce):O||(O=D&~f,O!==0&&(Q=Xn(O)))),Q===0?0:p!==0&&p!==Q&&(p&Y)===0&&(Y=Q&-Q,O=p&-p,Y>=O||Y===32&&(O&4194048)!==0)?p:Q}function Or(f,p){return(f.pendingLanes&~(f.suspendedLanes&~f.pingedLanes)&p)===0}function Cr(f,p){switch(f){case 1:case 2:case 4:case 8:case 64:return p+250;case 16:case 32:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return p+5e3;case 4194304:case 8388608:case 16777216:case 33554432:return-1;case 67108864:case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function wr(){var f=rn;return rn<<=1,(rn&62914560)===0&&(rn=4194304),f}function Gr(f){for(var p=[],O=0;31>O;O++)p.push(f);return p}function ji(f,p){f.pendingLanes|=p,p!==268435456&&(f.suspendedLanes=0,f.pingedLanes=0,f.warmLanes=0)}function ai(f,p,O,D,Q,Y){var ce=f.pendingLanes;f.pendingLanes=O,f.suspendedLanes=0,f.pingedLanes=0,f.warmLanes=0,f.expiredLanes&=O,f.entangledLanes&=O,f.errorRecoveryDisabledLanes&=O,f.shellSuspendCounter=0;var _e=f.entanglements,Le=f.expirationTimes,ut=f.hiddenUpdates;for(O=ce&~O;0<O;){var pt=31-Fe(O),xt=1<<pt;_e[pt]=0,Le[pt]=-1;var ot=ut[pt];if(ot!==null)for(ut[pt]=null,pt=0;pt<ot.length;pt++){var ft=ot[pt];ft!==null&&(ft.lane&=-536870913)}O&=~xt}D!==0&&ar(f,D,0),Y!==0&&Q===0&&f.tag!==0&&(f.suspendedLanes|=Y&~(ce&~p))}function ar(f,p,O){f.pendingLanes|=p,f.suspendedLanes&=~p;var D=31-Fe(p);f.entangledLanes|=p,f.entanglements[D]=f.entanglements[D]|1073741824|O&261930}function Te(f,p){var O=f.entangledLanes|=p;for(f=f.entanglements;O;){var D=31-Fe(O),Q=1<<D;Q&p|f[D]&p&&(f[D]|=p),O&=~Q}}function Ve(f,p){var O=p&-p;return O=(O&42)!==0?1:nt(O),(O&(f.suspendedLanes|p))!==0?0:O}function nt(f){switch(f){case 2:f=1;break;case 8:f=4;break;case 32:f=16;break;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:f=128;break;case 268435456:f=134217728;break;default:f=0}return f}function gt(f){return f&=-f,2<f?8<f?(f&134217727)!==0?32:268435456:8:2}function Ct(){var f=J.p;return f!==0?f:(f=window.event,f===void 0?32:C_(f.type))}function Vt(f,p){var O=J.p;try{return J.p=f,p()}finally{J.p=O}}var Yt=Math.random().toString(36).slice(2),Dt="__reactFiber$"+Yt,Lt="__reactProps$"+Yt,an="__reactContainer$"+Yt,_n="__reactEvents$"+Yt,sn="__reactListeners$"+Yt,Ln="__reactHandles$"+Yt,er="__reactResources$"+Yt,Di="__reactMarker$"+Yt;function $i(f){delete f[Dt],delete f[Lt],delete f[_n],delete f[sn],delete f[Ln]}function hr(f){var p=f[Dt];if(p)return p;for(var O=f.parentNode;O;){if(p=O[an]||O[Dt]){if(O=p.alternate,p.child!==null||O!==null&&O.child!==null)for(f=c_(f);f!==null;){if(O=f[Dt])return O;f=c_(f)}return p}f=O,O=f.parentNode}return null}function wt(f){if(f=f[Dt]||f[an]){var p=f.tag;if(p===5||p===6||p===13||p===31||p===26||p===27||p===3)return f}return null}function Kr(f){var p=f.tag;if(p===5||p===26||p===27||p===6)return f.stateNode;throw Error(r(33))}function Zi(f){var p=f[er];return p||(p=f[er]={hoistableStyles:new Map,hoistableScripts:new Map}),p}function yi(f){f[Di]=!0}var Ca=new Set,Xa={};function Au(f,p){Tu(f,p),Tu(f+"Capture",p)}function Tu(f,p){for(Xa[f]=p,f=0;f<p.length;f++)Ca.add(p[f])}var Nl=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),gc={},xl={};function ur(f){return tt.call(xl,f)?!0:tt.call(gc,f)?!1:Nl.test(f)?xl[f]=!0:(gc[f]=!0,!1)}function Rr(f,p,O){if(ur(p))if(O===null)f.removeAttribute(p);else{switch(typeof O){case"undefined":case"function":case"symbol":f.removeAttribute(p);return;case"boolean":var D=p.toLowerCase().slice(0,5);if(D!=="data-"&&D!=="aria-"){f.removeAttribute(p);return}}f.setAttribute(p,""+O)}}function Sl(f,p,O){if(O===null)f.removeAttribute(p);else{switch(typeof O){case"undefined":case"function":case"symbol":case"boolean":f.removeAttribute(p);return}f.setAttribute(p,""+O)}}function tr(f,p,O,D){if(D===null)f.removeAttribute(O);else{switch(typeof D){case"undefined":case"function":case"symbol":case"boolean":f.removeAttribute(O);return}f.setAttributeNS(p,O,""+D)}}function Bn(f){switch(typeof f){case"bigint":case"boolean":case"number":case"string":case"undefined":return f;case"object":return f;default:return""}}function ku(f){var p=f.type;return(f=f.nodeName)&&f.toLowerCase()==="input"&&(p==="checkbox"||p==="radio")}function gf(f,p,O){var D=Object.getOwnPropertyDescriptor(f.constructor.prototype,p);if(!f.hasOwnProperty(p)&&typeof D<"u"&&typeof D.get=="function"&&typeof D.set=="function"){var Q=D.get,Y=D.set;return Object.defineProperty(f,p,{configurable:!0,get:function(){return Q.call(this)},set:function(ce){O=""+ce,Y.call(this,ce)}}),Object.defineProperty(f,p,{enumerable:D.enumerable}),{getValue:function(){return O},setValue:function(ce){O=""+ce},stopTracking:function(){f._valueTracker=null,delete f[p]}}}}function Ds(f){if(!f._valueTracker){var p=ku(f)?"checked":"value";f._valueTracker=gf(f,p,""+f[p])}}function qo(f){if(!f)return!1;var p=f._valueTracker;if(!p)return!0;var O=p.getValue(),D="";return f&&(D=ku(f)?f.checked?"true":"false":f.value),f=D,f!==O?(p.setValue(f),!0):!1}function Il(f){if(f=f||(typeof document<"u"?document:void 0),typeof f>"u")return null;try{return f.activeElement||f.body}catch{return f.body}}var sh=/[\n"\\]/g;function Ya(f){return f.replace(sh,function(p){return"\\"+p.charCodeAt(0).toString(16)+" "})}function rd(f,p,O,D,Q,Y,ce,_e){f.name="",ce!=null&&typeof ce!="function"&&typeof ce!="symbol"&&typeof ce!="boolean"?f.type=ce:f.removeAttribute("type"),p!=null?ce==="number"?(p===0&&f.value===""||f.value!=p)&&(f.value=""+Bn(p)):f.value!==""+Bn(p)&&(f.value=""+Bn(p)):ce!=="submit"&&ce!=="reset"||f.removeAttribute("value"),p!=null?id(f,ce,Bn(p)):O!=null?id(f,ce,Bn(O)):D!=null&&f.removeAttribute("value"),Q==null&&Y!=null&&(f.defaultChecked=!!Y),Q!=null&&(f.checked=Q&&typeof Q!="function"&&typeof Q!="symbol"),_e!=null&&typeof _e!="function"&&typeof _e!="symbol"&&typeof _e!="boolean"?f.name=""+Bn(_e):f.removeAttribute("name")}function $s(f,p,O,D,Q,Y,ce,_e){if(Y!=null&&typeof Y!="function"&&typeof Y!="symbol"&&typeof Y!="boolean"&&(f.type=Y),p!=null||O!=null){if(!(Y!=="submit"&&Y!=="reset"||p!=null)){Ds(f);return}O=O!=null?""+Bn(O):"",p=p!=null?""+Bn(p):O,_e||p===f.value||(f.value=p),f.defaultValue=p}D=D??Q,D=typeof D!="function"&&typeof D!="symbol"&&!!D,f.checked=_e?f.checked:!!D,f.defaultChecked=!!D,ce!=null&&typeof ce!="function"&&typeof ce!="symbol"&&typeof ce!="boolean"&&(f.name=ce),Ds(f)}function id(f,p,O){p==="number"&&Il(f.ownerDocument)===f||f.defaultValue===""+O||(f.defaultValue=""+O)}function vc(f,p,O,D){if(f=f.options,p){p={};for(var Q=0;Q<O.length;Q++)p["$"+O[Q]]=!0;for(O=0;O<f.length;O++)Q=p.hasOwnProperty("$"+f[O].value),f[O].selected!==Q&&(f[O].selected=Q),Q&&D&&(f[O].defaultSelected=!0)}else{for(O=""+Bn(O),p=null,Q=0;Q<f.length;Q++){if(f[Q].value===O){f[Q].selected=!0,D&&(f[Q].defaultSelected=!0);return}p!==null||f[Q].disabled||(p=f[Q])}p!==null&&(p.selected=!0)}}function yc(f,p,O){if(p!=null&&(p=""+Bn(p),p!==f.value&&(f.value=p),O==null)){f.defaultValue!==p&&(f.defaultValue=p);return}f.defaultValue=O!=null?""+Bn(O):""}function Vo(f,p,O,D){if(p==null){if(D!=null){if(O!=null)throw Error(r(92));if(W(D)){if(1<D.length)throw Error(r(93));D=D[0]}O=D}O==null&&(O=""),p=O}O=Bn(p),f.defaultValue=O,D=f.textContent,D===O&&D!==""&&D!==null&&(f.value=D),Ds(f)}function Pu(f,p){if(p){var O=f.firstChild;if(O&&O===f.lastChild&&O.nodeType===3){O.nodeValue=p;return}}f.textContent=p}var Cs=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function ws(f,p,O){var D=p.indexOf("--")===0;O==null||typeof O=="boolean"||O===""?D?f.setProperty(p,""):p==="float"?f.cssFloat="":f[p]="":D?f.setProperty(p,O):typeof O!="number"||O===0||Cs.has(p)?p==="float"?f.cssFloat=O:f[p]=(""+O).trim():f[p]=O+"px"}function j0(f,p,O){if(p!=null&&typeof p!="object")throw Error(r(62));if(f=f.style,O!=null){for(var D in O)!O.hasOwnProperty(D)||p!=null&&p.hasOwnProperty(D)||(D.indexOf("--")===0?f.setProperty(D,""):D==="float"?f.cssFloat="":f[D]="");for(var Q in p)D=p[Q],p.hasOwnProperty(Q)&&O[Q]!==D&&ws(f,Q,D)}else for(var Y in p)p.hasOwnProperty(Y)&&ws(f,Y,p[Y])}function vf(f){if(f.indexOf("-")===-1)return!1;switch(f){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var Z0=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),Op=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function yf(f){return Op.test(""+f)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":f}function ho(){}var ch=null;function fh(f){return f=f.target||f.srcElement||window,f.correspondingUseElement&&(f=f.correspondingUseElement),f.nodeType===3?f.parentNode:f}var Rs=null,bc=null;function H0(f){var p=wt(f);if(p&&(f=p.stateNode)){var O=f[Lt]||null;e:switch(f=p.stateNode,p.type){case"input":if(rd(f,O.value,O.defaultValue,O.defaultValue,O.checked,O.defaultChecked,O.type,O.name),p=O.name,O.type==="radio"&&p!=null){for(O=f;O.parentNode;)O=O.parentNode;for(O=O.querySelectorAll('input[name="'+Ya(""+p)+'"][type="radio"]'),p=0;p<O.length;p++){var D=O[p];if(D!==f&&D.form===f.form){var Q=D[Lt]||null;if(!Q)throw Error(r(90));rd(D,Q.value,Q.defaultValue,Q.defaultValue,Q.checked,Q.defaultChecked,Q.type,Q.name)}}for(p=0;p<O.length;p++)D=O[p],D.form===f.form&&qo(D)}break e;case"textarea":yc(f,O.value,O.defaultValue);break e;case"select":p=O.value,p!=null&&vc(f,!!O.multiple,p,!1)}}}var xp=!1;function Sp(f,p,O){if(xp)return f(p,O);xp=!0;try{var D=f(p);return D}finally{if(xp=!1,(Rs!==null||bc!==null)&&(De(),Rs&&(p=Rs,f=bc,bc=Rs=null,H0(p),f)))for(p=0;p<f.length;p++)H0(f[p])}}function As(f,p){var O=f.stateNode;if(O===null)return null;var D=O[Lt]||null;if(D===null)return null;O=D[p];e:switch(p){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(D=!D.disabled)||(f=f.type,D=!(f==="button"||f==="input"||f==="select"||f==="textarea")),f=!D;break e;default:f=!1}if(f)return null;if(O&&typeof O!="function")throw Error(r(231,p,typeof O));return O}var Uo=!(typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"),_p=!1;if(Uo)try{var dh={};Object.defineProperty(dh,"passive",{get:function(){_p=!0}}),window.addEventListener("test",dh,dh),window.removeEventListener("test",dh,dh)}catch{_p=!1}var Oc=null,hh=null,Mr=null;function Qn(){if(Mr)return Mr;var f,p=hh,O=p.length,D,Q="value"in Oc?Oc.value:Oc.textContent,Y=Q.length;for(f=0;f<O&&p[f]===Q[f];f++);var ce=O-f;for(D=1;D<=ce&&p[O-D]===Q[Y-D];D++);return Mr=Q.slice(f,1<D?1-D:void 0)}function xc(f){var p=f.keyCode;return"charCode"in f?(f=f.charCode,f===0&&p===13&&(f=13)):f=p,f===10&&(f=13),32<=f||f===13?f:0}function jo(){return!0}function Sc(){return!1}function ra(f){function p(O,D,Q,Y,ce){this._reactName=O,this._targetInst=Q,this.type=D,this.nativeEvent=Y,this.target=ce,this.currentTarget=null;for(var _e in f)f.hasOwnProperty(_e)&&(O=f[_e],this[_e]=O?O(Y):Y[_e]);return this.isDefaultPrevented=(Y.defaultPrevented!=null?Y.defaultPrevented:Y.returnValue===!1)?jo:Sc,this.isPropagationStopped=Sc,this}return d(p.prototype,{preventDefault:function(){this.defaultPrevented=!0;var O=this.nativeEvent;O&&(O.preventDefault?O.preventDefault():typeof O.returnValue!="unknown"&&(O.returnValue=!1),this.isDefaultPrevented=jo)},stopPropagation:function(){var O=this.nativeEvent;O&&(O.stopPropagation?O.stopPropagation():typeof O.cancelBubble!="unknown"&&(O.cancelBubble=!0),this.isPropagationStopped=jo)},persist:function(){},isPersistent:jo}),p}var Ts={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(f){return f.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},ad=ra(Ts),Ll=d({},Ts,{view:0,detail:0}),ud=ra(Ll),ia,ks,Bl,ld=d({},Ll,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:Ps,button:0,buttons:0,relatedTarget:function(f){return f.relatedTarget===void 0?f.fromElement===f.srcElement?f.toElement:f.fromElement:f.relatedTarget},movementX:function(f){return"movementX"in f?f.movementX:(f!==Bl&&(Bl&&f.type==="mousemove"?(ia=f.screenX-Bl.screenX,ks=f.screenY-Bl.screenY):ks=ia=0,Bl=f),ia)},movementY:function(f){return"movementY"in f?f.movementY:ks}}),Ep=ra(ld),ui=d({},ld,{dataTransfer:0}),X0=ra(ui),Y0=d({},Ll,{relatedTarget:0}),od=ra(Y0),G0=d({},Ts,{animationName:0,elapsedTime:0,pseudoElement:0}),_c=ra(G0),sd=d({},Ts,{clipboardData:function(f){return"clipboardData"in f?f.clipboardData:window.clipboardData}}),ha=ra(sd),cd=d({},Ts,{data:0}),Ba=ra(cd),Eg={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},Dp={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},Zo={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function ph(f){var p=this.nativeEvent;return p.getModifierState?p.getModifierState(f):(f=Zo[f])?!!p[f]:!1}function Ps(){return ph}var bi=d({},Ll,{key:function(f){if(f.key){var p=Eg[f.key]||f.key;if(p!=="Unidentified")return p}return f.type==="keypress"?(f=xc(f),f===13?"Enter":String.fromCharCode(f)):f.type==="keydown"||f.type==="keyup"?Dp[f.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:Ps,charCode:function(f){return f.type==="keypress"?xc(f):0},keyCode:function(f){return f.type==="keydown"||f.type==="keyup"?f.keyCode:0},which:function(f){return f.type==="keypress"?xc(f):f.type==="keydown"||f.type==="keyup"?f.keyCode:0}}),Ho=ra(bi),$p=d({},ld,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0}),Cp=ra($p),wp=d({},Ll,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:Ps}),Ms=ra(wp),Rp=d({},Ts,{propertyName:0,elapsedTime:0,pseudoElement:0}),Dy=ra(Rp),K0=d({},ld,{deltaX:function(f){return"deltaX"in f?f.deltaX:"wheelDeltaX"in f?-f.wheelDeltaX:0},deltaY:function(f){return"deltaY"in f?f.deltaY:"wheelDeltaY"in f?-f.wheelDeltaY:"wheelDelta"in f?-f.wheelDelta:0},deltaZ:0,deltaMode:0}),po=ra(K0),Ku=d({},Ts,{newState:0,oldState:0}),W0=ra(Ku),mh=[9,13,27,32],Ec=Uo&&"CompositionEvent"in window,Xo=null;Uo&&"documentMode"in document&&(Xo=document.documentMode);var Fs=Uo&&"TextEvent"in window&&!Xo,Qa=Uo&&(!Ec||Xo&&8<Xo&&11>=Xo),gh=" ",vh=!1;function fd(f,p){switch(f){case"keyup":return mh.indexOf(p.keyCode)!==-1;case"keydown":return p.keyCode!==229;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function dd(f){return f=f.detail,typeof f=="object"&&"data"in f?f.data:null}var Ns=!1;function pu(f,p){switch(f){case"compositionend":return dd(p);case"keypress":return p.which!==32?null:(vh=!0,gh);case"textInput":return f=p.data,f===gh&&vh?null:f;default:return null}}function Ap(f,p){if(Ns)return f==="compositionend"||!Ec&&fd(f,p)?(f=Qn(),Mr=hh=Oc=null,Ns=!1,f):null;switch(f){case"paste":return null;case"keypress":if(!(p.ctrlKey||p.altKey||p.metaKey)||p.ctrlKey&&p.altKey){if(p.char&&1<p.char.length)return p.char;if(p.which)return String.fromCharCode(p.which)}return null;case"compositionend":return Qa&&p.locale!=="ko"?null:p.data;default:return null}}var Ga={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function Tp(f){var p=f&&f.nodeName&&f.nodeName.toLowerCase();return p==="input"?!!Ga[f.type]:p==="textarea"}function hd(f,p,O,D){Rs?bc?bc.push(D):bc=[D]:Rs=D,p=ib(p,"onChange"),0<p.length&&(O=new ad("onChange","change",null,O,D),f.push({event:O,listeners:p}))}var bf=null,pd=null;function Wu(f){t_(f,0)}function mo(f){var p=Kr(f);if(qo(p))return f}function J0(f,p){if(f==="change")return p}var mu=!1;if(Uo){var Of;if(Uo){var Ar="oninput"in document;if(!Ar){var yh=document.createElement("div");yh.setAttribute("oninput","return;"),Ar=typeof yh.oninput=="function"}Of=Ar}else Of=!1;mu=Of&&(!document.documentMode||9<document.documentMode)}function kp(){bf&&(bf.detachEvent("onpropertychange",go),pd=bf=null)}function go(f){if(f.propertyName==="value"&&mo(pd)){var p=[];hd(p,pd,f,fh(f)),Sp(Wu,p)}}function Kn(f,p,O){f==="focusin"?(kp(),bf=p,pd=O,bf.attachEvent("onpropertychange",go)):f==="focusout"&&kp()}function _l(f){if(f==="selectionchange"||f==="keyup"||f==="keydown")return mo(pd)}function Dg(f,p){if(f==="click")return mo(p)}function Ii(f,p){if(f==="input"||f==="change")return mo(p)}function gu(f,p){return f===p&&(f!==0||1/f===1/p)||f!==f&&p!==p}var pr=typeof Object.is=="function"?Object.is:gu;function xf(f,p){if(pr(f,p))return!0;if(typeof f!="object"||f===null||typeof p!="object"||p===null)return!1;var O=Object.keys(f),D=Object.keys(p);if(O.length!==D.length)return!1;for(D=0;D<O.length;D++){var Q=O[D];if(!tt.call(p,Q)||!pr(f[Q],p[Q]))return!1}return!0}function em(f){for(;f&&f.firstChild;)f=f.firstChild;return f}function tm(f,p){var O=em(f);f=0;for(var D;O;){if(O.nodeType===3){if(D=f+O.textContent.length,f<=p&&D>=p)return{node:O,offset:p-f};f=D}e:{for(;O;){if(O.nextSibling){O=O.nextSibling;break e}O=O.parentNode}O=void 0}O=em(O)}}function xr(f,p){return f&&p?f===p?!0:f&&f.nodeType===3?!1:p&&p.nodeType===3?xr(f,p.parentNode):"contains"in f?f.contains(p):f.compareDocumentPosition?!!(f.compareDocumentPosition(p)&16):!1:!1}function nm(f){f=f!=null&&f.ownerDocument!=null&&f.ownerDocument.defaultView!=null?f.ownerDocument.defaultView:window;for(var p=Il(f.document);p instanceof f.HTMLIFrameElement;){try{var O=typeof p.contentWindow.location.href=="string"}catch{O=!1}if(O)f=p.contentWindow;else break;p=Il(f.document)}return p}function Sf(f){var p=f&&f.nodeName&&f.nodeName.toLowerCase();return p&&(p==="input"&&(f.type==="text"||f.type==="search"||f.type==="tel"||f.type==="url"||f.type==="password")||p==="textarea"||f.contentEditable==="true")}var Dc=Uo&&"documentMode"in document&&11>=document.documentMode,Xe=null,dt=null,rt=null,Bt=!1;function En(f,p,O){var D=O.window===O?O.document:O.nodeType===9?O:O.ownerDocument;Bt||Xe==null||Xe!==Il(D)||(D=Xe,"selectionStart"in D&&Sf(D)?D={start:D.selectionStart,end:D.selectionEnd}:(D=(D.ownerDocument&&D.ownerDocument.defaultView||window).getSelection(),D={anchorNode:D.anchorNode,anchorOffset:D.anchorOffset,focusNode:D.focusNode,focusOffset:D.focusOffset}),rt&&xf(rt,D)||(rt=D,D=ib(dt,"onSelect"),0<D.length&&(p=new ad("onSelect","select",null,p,O),f.push({event:p,listeners:D}),p.target=Xe)))}function lr(f,p){var O={};return O[f.toLowerCase()]=p.toLowerCase(),O["Webkit"+f]="webkit"+p,O["Moz"+f]="moz"+p,O}var Li={animationend:lr("Animation","AnimationEnd"),animationiteration:lr("Animation","AnimationIteration"),animationstart:lr("Animation","AnimationStart"),transitionrun:lr("Transition","TransitionRun"),transitionstart:lr("Transition","TransitionStart"),transitioncancel:lr("Transition","TransitionCancel"),transitionend:lr("Transition","TransitionEnd")},zr={},Ka={};Uo&&(Ka=document.createElement("div").style,"AnimationEvent"in window||(delete Li.animationend.animation,delete Li.animationiteration.animation,delete Li.animationstart.animation),"TransitionEvent"in window||delete Li.transitionend.transition);function $c(f){if(zr[f])return zr[f];if(!Li[f])return f;var p=Li[f],O;for(O in p)if(p.hasOwnProperty(O)&&O in Ka)return zr[f]=p[O];return f}var Ci=$c("animationend"),md=$c("animationiteration"),Ql=$c("animationstart"),vo=$c("transitionrun"),yo=$c("transitionstart"),_f=$c("transitioncancel"),Sr=$c("transitionend"),Yo=new Map,bh="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll toggle touchMove waiting wheel".split(" ");bh.push("scrollEnd");function pa(f,p){Yo.set(f,p),Au(p,[f])}var Go=typeof reportError=="function"?reportError:function(f){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var p=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof f=="object"&&f!==null&&typeof f.message=="string"?String(f.message):String(f),error:f});if(!window.dispatchEvent(p))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",f);return}console.error(f)},wa=[],bo=0,Is=0;function Wa(){for(var f=bo,p=Is=bo=0;p<f;){var O=wa[p];wa[p++]=null;var D=wa[p];wa[p++]=null;var Q=wa[p];wa[p++]=null;var Y=wa[p];if(wa[p++]=null,D!==null&&Q!==null){var ce=D.pending;ce===null?Q.next=Q:(Q.next=ce.next,ce.next=Q),D.pending=Q}Y!==0&&wc(O,Q,Y)}}function Ju(f,p,O,D){wa[bo++]=f,wa[bo++]=p,wa[bo++]=O,wa[bo++]=D,Is|=D,f.lanes|=D,f=f.alternate,f!==null&&(f.lanes|=D)}function Cc(f,p,O,D){return Ju(f,p,O,D),Rc(f)}function el(f,p){return Ju(f,null,null,p),Rc(f)}function wc(f,p,O){f.lanes|=O;var D=f.alternate;D!==null&&(D.lanes|=O);for(var Q=!1,Y=f.return;Y!==null;)Y.childLanes|=O,D=Y.alternate,D!==null&&(D.childLanes|=O),Y.tag===22&&(f=Y.stateNode,f===null||f._visibility&1||(Q=!0)),f=Y,Y=Y.return;return f.tag===3?(Y=f.stateNode,Q&&p!==null&&(Q=31-Fe(O),f=Y.hiddenUpdates,D=f[Q],D===null?f[Q]=[p]:D.push(p),p.lane=O|536870912),Y):null}function Rc(f){if(50<Bh)throw Bh=0,Qh=null,Error(r(185));for(var p=f.return;p!==null;)f=p,p=f.return;return f.tag===3?f.stateNode:null}var vu={};function gd(f,p,O,D){this.tag=f,this.key=O,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=p,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=D,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function aa(f,p,O,D){return new gd(f,p,O,D)}function Ji(f){return f=f.prototype,!(!f||!f.isReactComponent)}function tl(f,p){var O=f.alternate;return O===null?(O=aa(f.tag,p,f.key,f.mode),O.elementType=f.elementType,O.type=f.type,O.stateNode=f.stateNode,O.alternate=f,f.alternate=O):(O.pendingProps=p,O.type=f.type,O.flags=0,O.subtreeFlags=0,O.deletions=null),O.flags=f.flags&65011712,O.childLanes=f.childLanes,O.lanes=f.lanes,O.child=f.child,O.memoizedProps=f.memoizedProps,O.memoizedState=f.memoizedState,O.updateQueue=f.updateQueue,p=f.dependencies,O.dependencies=p===null?null:{lanes:p.lanes,firstContext:p.firstContext},O.sibling=f.sibling,O.index=f.index,O.ref=f.ref,O.refCleanup=f.refCleanup,O}function yu(f,p){f.flags&=65011714;var O=f.alternate;return O===null?(f.childLanes=0,f.lanes=p,f.child=null,f.subtreeFlags=0,f.memoizedProps=null,f.memoizedState=null,f.updateQueue=null,f.dependencies=null,f.stateNode=null):(f.childLanes=O.childLanes,f.lanes=O.lanes,f.child=O.child,f.subtreeFlags=0,f.deletions=null,f.memoizedProps=O.memoizedProps,f.memoizedState=O.memoizedState,f.updateQueue=O.updateQueue,f.type=O.type,p=O.dependencies,f.dependencies=p===null?null:{lanes:p.lanes,firstContext:p.firstContext}),f}function Oh(f,p,O,D,Q,Y){var ce=0;if(D=f,typeof f=="function")Ji(f)&&(ce=1);else if(typeof f=="string")ce=Q5(f,O,ne.current)?26:f==="html"||f==="head"||f==="body"?27:5;else e:switch(f){case I:return f=aa(31,O,p,Q),f.elementType=I,f.lanes=Y,f;case v:return zl(O.children,Q,Y,p);case S:ce=8,Q|=24;break;case _:return f=aa(12,O,p,Q|2),f.elementType=_,f.lanes=Y,f;case A:return f=aa(13,O,p,Q),f.elementType=A,f.lanes=Y,f;case w:return f=aa(19,O,p,Q),f.elementType=w,f.lanes=Y,f;default:if(typeof f=="object"&&f!==null)switch(f.$$typeof){case $:ce=10;break e;case E:ce=9;break e;case k:ce=11;break e;case R:ce=14;break e;case F:ce=16,D=null;break e}ce=29,O=Error(r(130,f===null?"null":typeof f,"")),D=null}return p=aa(ce,O,p,Q),p.elementType=f,p.type=D,p.lanes=Y,p}function zl(f,p,O,D){return f=aa(7,f,D,p),f.lanes=O,f}function Ef(f,p,O){return f=aa(6,f,null,p),f.lanes=O,f}function xh(f){var p=aa(18,null,null,0);return p.stateNode=f,p}function bu(f,p,O){return p=aa(4,f.children!==null?f.children:[],f.key,p),p.lanes=O,p.stateNode={containerInfo:f.containerInfo,pendingChildren:null,implementation:f.implementation},p}var Pp=new WeakMap;function Mu(f,p){if(typeof f=="object"&&f!==null){var O=Pp.get(f);return O!==void 0?O:(p={value:f,source:p,stack:et(p)},Pp.set(f,p),p)}return{value:f,source:p,stack:et(p)}}var ql=[],Ac=0,Bi=null,wi=0,Fu=[],dn=0,Nu=null,Ko=1,ue="";function L(f,p){ql[Ac++]=wi,ql[Ac++]=Bi,Bi=f,wi=p}function re(f,p,O){Fu[dn++]=Ko,Fu[dn++]=ue,Fu[dn++]=Nu,Nu=f;var D=Ko;f=ue;var Q=32-Fe(D)-1;D&=~(1<<Q),O+=1;var Y=32-Fe(p)+Q;if(30<Y){var ce=Q-Q%5;Y=(D&(1<<ce)-1).toString(32),D>>=ce,Q-=ce,Ko=1<<32-Fe(p)+Q|O<<Q|D,ue=Y+f}else Ko=1<<Y|O<<Q|D,ue=f}function fe(f){f.return!==null&&(L(f,1),re(f,1,0))}function xe(f){for(;f===Bi;)Bi=ql[--Ac],ql[Ac]=null,wi=ql[--Ac],ql[Ac]=null;for(;f===Nu;)Nu=Fu[--dn],Fu[dn]=null,ue=Fu[--dn],Fu[dn]=null,Ko=Fu[--dn],Fu[dn]=null}function ke(f,p){Fu[dn++]=Ko,Fu[dn++]=ue,Fu[dn++]=Nu,Ko=p.id,ue=p.overflow,Nu=f}var Qe=null,Be=null,Ze=!1,Rt=null,Kt=!1,zn=Error(r(519));function ln(f){var p=Error(r(418,1<arguments.length&&arguments[1]!==void 0&&arguments[1]?"text":"HTML",""));throw Vl(Mu(p,f)),zn}function mn(f){var p=f.stateNode,O=f.type,D=f.memoizedProps;switch(p[Dt]=f,p[Lt]=D,O){case"dialog":kr("cancel",p),kr("close",p);break;case"iframe":case"object":case"embed":kr("load",p);break;case"video":case"audio":for(O=0;O<rb.length;O++)kr(rb[O],p);break;case"source":kr("error",p);break;case"img":case"image":case"link":kr("error",p),kr("load",p);break;case"details":kr("toggle",p);break;case"input":kr("invalid",p),$s(p,D.value,D.defaultValue,D.checked,D.defaultChecked,D.type,D.name,!0);break;case"select":kr("invalid",p);break;case"textarea":kr("invalid",p),Vo(p,D.value,D.defaultValue,D.children)}O=D.children,typeof O!="string"&&typeof O!="number"&&typeof O!="bigint"||p.textContent===""+O||D.suppressHydrationWarning===!0||i_(p.textContent,O)?(D.popover!=null&&(kr("beforetoggle",p),kr("toggle",p)),D.onScroll!=null&&kr("scroll",p),D.onScrollEnd!=null&&kr("scrollend",p),D.onClick!=null&&(p.onclick=ho),p=!0):p=!1,p||ln(f,!0)}function Ri(f){for(Qe=f.return;Qe;)switch(Qe.tag){case 5:case 31:case 13:Kt=!1;return;case 27:case 3:Kt=!0;return;default:Qe=Qe.return}}function za(f){if(f!==Qe)return!1;if(!Ze)return Ri(f),Ze=!0,!1;var p=f.tag,O;if((O=p!==3&&p!==27)&&((O=p===5)&&(O=f.type,O=!(O!=="form"&&O!=="button")||Gp(f.type,f.memoizedProps)),O=!O),O&&Be&&ln(f),Ri(f),p===13){if(f=f.memoizedState,f=f!==null?f.dehydrated:null,!f)throw Error(r(317));Be=s_(f)}else if(p===31){if(f=f.memoizedState,f=f!==null?f.dehydrated:null,!f)throw Error(r(317));Be=s_(f)}else p===27?(p=Be,Vn(f.type)?(f=I4,I4=null,Be=f):Be=p):Be=Qe?Zc(f.stateNode.nextSibling):null;return!0}function Se(){Be=Qe=null,Ze=!1}function El(){var f=Rt;return f!==null&&(dl===null?dl=f:dl.push.apply(dl,f),Rt=null),f}function Vl(f){Rt===null?Rt=[f]:Rt.push(f)}var Iu=G(null),qn=null,Ja=null;function Ls(f,p,O){B(Iu,p._currentValue),p._currentValue=O}function Ul(f){f._currentValue=Iu.current,X(Iu)}function Lu(f,p,O){for(;f!==null;){var D=f.alternate;if((f.childLanes&p)!==p?(f.childLanes|=p,D!==null&&(D.childLanes|=p)):D!==null&&(D.childLanes&p)!==p&&(D.childLanes|=p),f===O)break;f=f.return}}function Tc(f,p,O,D){var Q=f.child;for(Q!==null&&(Q.return=f);Q!==null;){var Y=Q.dependencies;if(Y!==null){var ce=Q.child;Y=Y.firstContext;e:for(;Y!==null;){var _e=Y;Y=Q;for(var Le=0;Le<p.length;Le++)if(_e.context===p[Le]){Y.lanes|=O,_e=Y.alternate,_e!==null&&(_e.lanes|=O),Lu(Y.return,O,f),D||(ce=null);break e}Y=_e.next}}else if(Q.tag===18){if(ce=Q.return,ce===null)throw Error(r(341));ce.lanes|=O,Y=ce.alternate,Y!==null&&(Y.lanes|=O),Lu(ce,O,f),ce=null}else ce=Q.child;if(ce!==null)ce.return=Q;else for(ce=Q;ce!==null;){if(ce===f){ce=null;break}if(Q=ce.sibling,Q!==null){Q.return=ce.return,ce=Q;break}ce=ce.return}Q=ce}}function jl(f,p,O,D){f=null;for(var Q=p,Y=!1;Q!==null;){if(!Y){if((Q.flags&524288)!==0)Y=!0;else if((Q.flags&262144)!==0)break}if(Q.tag===10){var ce=Q.alternate;if(ce===null)throw Error(r(387));if(ce=ce.memoizedProps,ce!==null){var _e=Q.type;pr(Q.pendingProps.value,ce.value)||(f!==null?f.push(_e):f=[_e])}}else if(Q===de.current){if(ce=Q.alternate,ce===null)throw Error(r(387));ce.memoizedState.memoizedState!==Q.memoizedState.memoizedState&&(f!==null?f.push(cv):f=[cv])}Q=Q.return}f!==null&&Tc(p,f,O,D),p.flags|=262144}function vd(f){for(f=f.firstContext;f!==null;){if(!pr(f.context._currentValue,f.memoizedValue))return!0;f=f.next}return!1}function Sh(f){qn=f,Ja=null,f=f.dependencies,f!==null&&(f.firstContext=null)}function Ou(f){return $y(qn,f)}function Oo(f,p){return qn===null&&Sh(f),$y(f,p)}function $y(f,p){var O=p._currentValue;if(p={context:p,memoizedValue:O,next:null},Ja===null){if(f===null)throw Error(r(308));Ja=p,f.dependencies={lanes:0,firstContext:p},f.flags|=524288}else Ja=Ja.next=p;return O}var I1=typeof AbortController<"u"?AbortController:function(){var f=[],p=this.signal={aborted:!1,addEventListener:function(O,D){f.push(D)}};this.abort=function(){p.aborted=!0,f.forEach(function(O){return O()})}},L1=e.unstable_scheduleCallback,Cy=e.unstable_NormalPriority,gn={$$typeof:$,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function Bs(){return{controller:new I1,data:new Map,refCount:0}}function Bu(f){f.refCount--,f.refCount===0&&L1(Cy,function(){f.controller.abort()})}var _h=null,wy=0,Df=0,Qs=null;function Wo(f,p){if(_h===null){var O=_h=[];wy=0,Df=Ws(),Qs={status:"pending",value:void 0,then:function(D){O.push(D)}}}return wy++,p.then(Ry,Ry),p}function Ry(){if(--wy===0&&_h!==null){Qs!==null&&(Qs.status="fulfilled");var f=_h;_h=null,Df=0,Qs=null;for(var p=0;p<f.length;p++)(0,f[p])()}}function Ay(f,p){var O=[],D={status:"pending",value:null,reason:null,then:function(Q){O.push(Q)}};return f.then(function(){D.status="fulfilled",D.value=p;for(var Q=0;Q<O.length;Q++)(0,O[Q])(p)},function(Q){for(D.status="rejected",D.reason=Q,Q=0;Q<O.length;Q++)(0,O[Q])(void 0)}),D}var Jo=K.S;K.S=function(f,p){Zp=je(),typeof p=="object"&&p!==null&&typeof p.then=="function"&&Wo(f,p),Jo!==null&&Jo(f,p)};var yd=G(null);function kc(){var f=yd.current;return f!==null?f:li.pooledCache}function Mp(f,p){p===null?B(yd,yd.current):B(yd,p.pool)}function $g(){var f=kc();return f===null?null:{parent:gn._currentValue,pool:f}}var bd=Error(r(460)),Fp=Error(r(474)),es=Error(r(542)),rm={then:function(){}};function B1(f){return f=f.status,f==="fulfilled"||f==="rejected"}function im(f,p,O){switch(O=f[O],O===void 0?f.push(p):O!==p&&(p.then(ho,ho),p=O),p.status){case"fulfilled":return p.value;case"rejected":throw f=p.reason,wg(f),f;default:if(typeof p.status=="string")p.then(ho,ho);else{if(f=li,f!==null&&100<f.shellSuspendCounter)throw Error(r(482));f=p,f.status="pending",f.then(function(D){if(p.status==="pending"){var Q=p;Q.status="fulfilled",Q.value=D}},function(D){if(p.status==="pending"){var Q=p;Q.status="rejected",Q.reason=D}})}switch(p.status){case"fulfilled":return p.value;case"rejected":throw f=p.reason,wg(f),f}throw xn=p,bd}}function zs(f){try{var p=f._init;return p(f._payload)}catch(O){throw O!==null&&typeof O=="object"&&typeof O.then=="function"?(xn=O,bd):O}}var xn=null;function Cg(){if(xn===null)throw Error(r(459));var f=xn;return xn=null,f}function wg(f){if(f===bd||f===es)throw Error(r(483))}var Od=null,nl=0;function Np(f){var p=nl;return nl+=1,Od===null&&(Od=[]),im(Od,f,p)}function qs(f,p){p=p.props.ref,f.ref=p!==void 0?p:null}function rl(f,p){throw p.$$typeof===h?Error(r(525)):(f=Object.prototype.toString.call(p),Error(r(31,f==="[object Object]"?"object with keys {"+Object.keys(p).join(", ")+"}":f)))}function Ip(f){function p(Ke,Ue){if(f){var it=Ke.deletions;it===null?(Ke.deletions=[Ue],Ke.flags|=16):it.push(Ue)}}function O(Ke,Ue){if(!f)return null;for(;Ue!==null;)p(Ke,Ue),Ue=Ue.sibling;return null}function D(Ke){for(var Ue=new Map;Ke!==null;)Ke.key!==null?Ue.set(Ke.key,Ke):Ue.set(Ke.index,Ke),Ke=Ke.sibling;return Ue}function Q(Ke,Ue){return Ke=tl(Ke,Ue),Ke.index=0,Ke.sibling=null,Ke}function Y(Ke,Ue,it){return Ke.index=it,f?(it=Ke.alternate,it!==null?(it=it.index,it<Ue?(Ke.flags|=67108866,Ue):it):(Ke.flags|=67108866,Ue)):(Ke.flags|=1048576,Ue)}function ce(Ke){return f&&Ke.alternate===null&&(Ke.flags|=67108866),Ke}function _e(Ke,Ue,it,yt){return Ue===null||Ue.tag!==6?(Ue=Ef(it,Ke.mode,yt),Ue.return=Ke,Ue):(Ue=Q(Ue,it),Ue.return=Ke,Ue)}function Le(Ke,Ue,it,yt){var Sn=it.type;return Sn===v?pt(Ke,Ue,it.props.children,yt,it.key):Ue!==null&&(Ue.elementType===Sn||typeof Sn=="object"&&Sn!==null&&Sn.$$typeof===F&&zs(Sn)===Ue.type)?(Ue=Q(Ue,it.props),qs(Ue,it),Ue.return=Ke,Ue):(Ue=Oh(it.type,it.key,it.props,null,Ke.mode,yt),qs(Ue,it),Ue.return=Ke,Ue)}function ut(Ke,Ue,it,yt){return Ue===null||Ue.tag!==4||Ue.stateNode.containerInfo!==it.containerInfo||Ue.stateNode.implementation!==it.implementation?(Ue=bu(it,Ke.mode,yt),Ue.return=Ke,Ue):(Ue=Q(Ue,it.children||[]),Ue.return=Ke,Ue)}function pt(Ke,Ue,it,yt,Sn){return Ue===null||Ue.tag!==7?(Ue=zl(it,Ke.mode,yt,Sn),Ue.return=Ke,Ue):(Ue=Q(Ue,it),Ue.return=Ke,Ue)}function xt(Ke,Ue,it){if(typeof Ue=="string"&&Ue!==""||typeof Ue=="number"||typeof Ue=="bigint")return Ue=Ef(""+Ue,Ke.mode,it),Ue.return=Ke,Ue;if(typeof Ue=="object"&&Ue!==null){switch(Ue.$$typeof){case m:return it=Oh(Ue.type,Ue.key,Ue.props,null,Ke.mode,it),qs(it,Ue),it.return=Ke,it;case y:return Ue=bu(Ue,Ke.mode,it),Ue.return=Ke,Ue;case F:return Ue=zs(Ue),xt(Ke,Ue,it)}if(W(Ue)||V(Ue))return Ue=zl(Ue,Ke.mode,it,null),Ue.return=Ke,Ue;if(typeof Ue.then=="function")return xt(Ke,Np(Ue),it);if(Ue.$$typeof===$)return xt(Ke,Oo(Ke,Ue),it);rl(Ke,Ue)}return null}function ot(Ke,Ue,it,yt){var Sn=Ue!==null?Ue.key:null;if(typeof it=="string"&&it!==""||typeof it=="number"||typeof it=="bigint")return Sn!==null?null:_e(Ke,Ue,""+it,yt);if(typeof it=="object"&&it!==null){switch(it.$$typeof){case m:return it.key===Sn?Le(Ke,Ue,it,yt):null;case y:return it.key===Sn?ut(Ke,Ue,it,yt):null;case F:return it=zs(it),ot(Ke,Ue,it,yt)}if(W(it)||V(it))return Sn!==null?null:pt(Ke,Ue,it,yt,null);if(typeof it.then=="function")return ot(Ke,Ue,Np(it),yt);if(it.$$typeof===$)return ot(Ke,Ue,Oo(Ke,it),yt);rl(Ke,it)}return null}function ft(Ke,Ue,it,yt,Sn){if(typeof yt=="string"&&yt!==""||typeof yt=="number"||typeof yt=="bigint")return Ke=Ke.get(it)||null,_e(Ue,Ke,""+yt,Sn);if(typeof yt=="object"&&yt!==null){switch(yt.$$typeof){case m:return Ke=Ke.get(yt.key===null?it:yt.key)||null,Le(Ue,Ke,yt,Sn);case y:return Ke=Ke.get(yt.key===null?it:yt.key)||null,ut(Ue,Ke,yt,Sn);case F:return yt=zs(yt),ft(Ke,Ue,it,yt,Sn)}if(W(yt)||V(yt))return Ke=Ke.get(it)||null,pt(Ue,Ke,yt,Sn,null);if(typeof yt.then=="function")return ft(Ke,Ue,it,Np(yt),Sn);if(yt.$$typeof===$)return ft(Ke,Ue,it,Oo(Ue,yt),Sn);rl(Ue,yt)}return null}function on(Ke,Ue,it,yt){for(var Sn=null,oi=null,hn=Ue,fr=Ue=0,Br=null;hn!==null&&fr<it.length;fr++){hn.index>fr?(Br=hn,hn=null):Br=hn.sibling;var si=ot(Ke,hn,it[fr],yt);if(si===null){hn===null&&(hn=Br);break}f&&hn&&si.alternate===null&&p(Ke,hn),Ue=Y(si,Ue,fr),oi===null?Sn=si:oi.sibling=si,oi=si,hn=Br}if(fr===it.length)return O(Ke,hn),Ze&&L(Ke,fr),Sn;if(hn===null){for(;fr<it.length;fr++)hn=xt(Ke,it[fr],yt),hn!==null&&(Ue=Y(hn,Ue,fr),oi===null?Sn=hn:oi.sibling=hn,oi=hn);return Ze&&L(Ke,fr),Sn}for(hn=D(hn);fr<it.length;fr++)Br=ft(hn,Ke,fr,it[fr],yt),Br!==null&&(f&&Br.alternate!==null&&hn.delete(Br.key===null?fr:Br.key),Ue=Y(Br,Ue,fr),oi===null?Sn=Br:oi.sibling=Br,oi=Br);return f&&hn.forEach(function(n0){return p(Ke,n0)}),Ze&&L(Ke,fr),Sn}function Rn(Ke,Ue,it,yt){if(it==null)throw Error(r(151));for(var Sn=null,oi=null,hn=Ue,fr=Ue=0,Br=null,si=it.next();hn!==null&&!si.done;fr++,si=it.next()){hn.index>fr?(Br=hn,hn=null):Br=hn.sibling;var n0=ot(Ke,hn,si.value,yt);if(n0===null){hn===null&&(hn=Br);break}f&&hn&&n0.alternate===null&&p(Ke,hn),Ue=Y(n0,Ue,fr),oi===null?Sn=n0:oi.sibling=n0,oi=n0,hn=Br}if(si.done)return O(Ke,hn),Ze&&L(Ke,fr),Sn;if(hn===null){for(;!si.done;fr++,si=it.next())si=xt(Ke,si.value,yt),si!==null&&(Ue=Y(si,Ue,fr),oi===null?Sn=si:oi.sibling=si,oi=si);return Ze&&L(Ke,fr),Sn}for(hn=D(hn);!si.done;fr++,si=it.next())si=ft(hn,Ke,fr,si.value,yt),si!==null&&(f&&si.alternate!==null&&hn.delete(si.key===null?fr:si.key),Ue=Y(si,Ue,fr),oi===null?Sn=si:oi.sibling=si,oi=si);return f&&hn.forEach(function(X5){return p(Ke,X5)}),Ze&&L(Ke,fr),Sn}function zi(Ke,Ue,it,yt){if(typeof it=="object"&&it!==null&&it.type===v&&it.key===null&&(it=it.props.children),typeof it=="object"&&it!==null){switch(it.$$typeof){case m:e:{for(var Sn=it.key;Ue!==null;){if(Ue.key===Sn){if(Sn=it.type,Sn===v){if(Ue.tag===7){O(Ke,Ue.sibling),yt=Q(Ue,it.props.children),yt.return=Ke,Ke=yt;break e}}else if(Ue.elementType===Sn||typeof Sn=="object"&&Sn!==null&&Sn.$$typeof===F&&zs(Sn)===Ue.type){O(Ke,Ue.sibling),yt=Q(Ue,it.props),qs(yt,it),yt.return=Ke,Ke=yt;break e}O(Ke,Ue);break}else p(Ke,Ue);Ue=Ue.sibling}it.type===v?(yt=zl(it.props.children,Ke.mode,yt,it.key),yt.return=Ke,Ke=yt):(yt=Oh(it.type,it.key,it.props,null,Ke.mode,yt),qs(yt,it),yt.return=Ke,Ke=yt)}return ce(Ke);case y:e:{for(Sn=it.key;Ue!==null;){if(Ue.key===Sn)if(Ue.tag===4&&Ue.stateNode.containerInfo===it.containerInfo&&Ue.stateNode.implementation===it.implementation){O(Ke,Ue.sibling),yt=Q(Ue,it.children||[]),yt.return=Ke,Ke=yt;break e}else{O(Ke,Ue);break}else p(Ke,Ue);Ue=Ue.sibling}yt=bu(it,Ke.mode,yt),yt.return=Ke,Ke=yt}return ce(Ke);case F:return it=zs(it),zi(Ke,Ue,it,yt)}if(W(it))return on(Ke,Ue,it,yt);if(V(it)){if(Sn=V(it),typeof Sn!="function")throw Error(r(150));return it=Sn.call(it),Rn(Ke,Ue,it,yt)}if(typeof it.then=="function")return zi(Ke,Ue,Np(it),yt);if(it.$$typeof===$)return zi(Ke,Ue,Oo(Ke,it),yt);rl(Ke,it)}return typeof it=="string"&&it!==""||typeof it=="number"||typeof it=="bigint"?(it=""+it,Ue!==null&&Ue.tag===6?(O(Ke,Ue.sibling),yt=Q(Ue,it),yt.return=Ke,Ke=yt):(O(Ke,Ue),yt=Ef(it,Ke.mode,yt),yt.return=Ke,Ke=yt),ce(Ke)):O(Ke,Ue)}return function(Ke,Ue,it,yt){try{nl=0;var Sn=zi(Ke,Ue,it,yt);return Od=null,Sn}catch(hn){if(hn===bd||hn===es)throw hn;var oi=aa(29,hn,null,Ke.mode);return oi.lanes=yt,oi.return=Ke,oi}finally{}}}var xd=Ip(!0),Ty=Ip(!1),ts=!1;function Pc(f){f.updateQueue={baseState:f.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function Rg(f,p){f=f.updateQueue,p.updateQueue===f&&(p.updateQueue={baseState:f.baseState,firstBaseUpdate:f.firstBaseUpdate,lastBaseUpdate:f.lastBaseUpdate,shared:f.shared,callbacks:null})}function Mc(f){return{lane:f,tag:0,payload:null,callback:null,next:null}}function ns(f,p,O){var D=f.updateQueue;if(D===null)return null;if(D=D.shared,(Nr&2)!==0){var Q=D.pending;return Q===null?p.next=p:(p.next=Q.next,Q.next=p),D.pending=p,p=Rc(f),wc(f,null,O),p}return Ju(f,D,p,O),Rc(f)}function Sd(f,p,O){if(p=p.updateQueue,p!==null&&(p=p.shared,(O&4194048)!==0)){var D=p.lanes;D&=f.pendingLanes,O|=D,p.lanes=O,Te(f,O)}}function am(f,p){var O=f.updateQueue,D=f.alternate;if(D!==null&&(D=D.updateQueue,O===D)){var Q=null,Y=null;if(O=O.firstBaseUpdate,O!==null){do{var ce={lane:O.lane,tag:O.tag,payload:O.payload,callback:null,next:null};Y===null?Q=Y=ce:Y=Y.next=ce,O=O.next}while(O!==null);Y===null?Q=Y=p:Y=Y.next=p}else Q=Y=p;O={baseState:D.baseState,firstBaseUpdate:Q,lastBaseUpdate:Y,shared:D.shared,callbacks:D.callbacks},f.updateQueue=O;return}f=O.lastBaseUpdate,f===null?O.firstBaseUpdate=p:f.next=p,O.lastBaseUpdate=p}var eu=!1;function Eh(){if(eu){var f=Qs;if(f!==null)throw f}}function _d(f,p,O,D){eu=!1;var Q=f.updateQueue;ts=!1;var Y=Q.firstBaseUpdate,ce=Q.lastBaseUpdate,_e=Q.shared.pending;if(_e!==null){Q.shared.pending=null;var Le=_e,ut=Le.next;Le.next=null,ce===null?Y=ut:ce.next=ut,ce=Le;var pt=f.alternate;pt!==null&&(pt=pt.updateQueue,_e=pt.lastBaseUpdate,_e!==ce&&(_e===null?pt.firstBaseUpdate=ut:_e.next=ut,pt.lastBaseUpdate=Le))}if(Y!==null){var xt=Q.baseState;ce=0,pt=ut=Le=null,_e=Y;do{var ot=_e.lane&-536870913,ft=ot!==_e.lane;if(ft?(_r&ot)===ot:(D&ot)===ot){ot!==0&&ot===Df&&(eu=!0),pt!==null&&(pt=pt.next={lane:0,tag:_e.tag,payload:_e.payload,callback:null,next:null});e:{var on=f,Rn=_e;ot=p;var zi=O;switch(Rn.tag){case 1:if(on=Rn.payload,typeof on=="function"){xt=on.call(zi,xt,ot);break e}xt=on;break e;case 3:on.flags=on.flags&-65537|128;case 0:if(on=Rn.payload,ot=typeof on=="function"?on.call(zi,xt,ot):on,ot==null)break e;xt=d({},xt,ot);break e;case 2:ts=!0}}ot=_e.callback,ot!==null&&(f.flags|=64,ft&&(f.flags|=8192),ft=Q.callbacks,ft===null?Q.callbacks=[ot]:ft.push(ot))}else ft={lane:ot,tag:_e.tag,payload:_e.payload,callback:_e.callback,next:null},pt===null?(ut=pt=ft,Le=xt):pt=pt.next=ft,ce|=ot;if(_e=_e.next,_e===null){if(_e=Q.shared.pending,_e===null)break;ft=_e,_e=ft.next,ft.next=null,Q.lastBaseUpdate=ft,Q.shared.pending=null}}while(!0);pt===null&&(Le=xt),Q.baseState=Le,Q.firstBaseUpdate=ut,Q.lastBaseUpdate=pt,Y===null&&(Q.shared.lanes=0),Ro|=ce,f.lanes=ce,f.memoizedState=xt}}function tu(f,p){if(typeof f!="function")throw Error(r(191,f));f.call(p)}function nu(f,p){var O=f.callbacks;if(O!==null)for(f.callbacks=null,f=0;f<O.length;f++)tu(O[f],p)}var il=G(null),Ra=G(0);function Ag(f,p){f=la,B(Ra,f),B(il,p),la=f|p.baseLanes}function Qu(){B(Ra,la),B(il,il.current)}function xu(){la=Ra.current,X(il),X(Ra)}var ru=G(null),rs=null;function Ed(f){var p=f.alternate;B(ma,ma.current&1),B(ru,f),rs===null&&(p===null||il.current!==null||p.memoizedState!==null)&&(rs=f)}function um(f){B(ma,ma.current),B(ru,f),rs===null&&(rs=f)}function Q1(f){f.tag===22?(B(ma,ma.current),B(ru,f),rs===null&&(rs=f)):xo()}function xo(){B(ma,ma.current),B(ru,ru.current)}function Zl(f){X(ru),rs===f&&(rs=null),X(ma)}var ma=G(0);function Tg(f){for(var p=f;p!==null;){if(p.tag===13){var O=p.memoizedState;if(O!==null&&(O=O.dehydrated,O===null||rO(O)||iv(O)))return p}else if(p.tag===19&&(p.memoizedProps.revealOrder==="forwards"||p.memoizedProps.revealOrder==="backwards"||p.memoizedProps.revealOrder==="unstable_legacy-backwards"||p.memoizedProps.revealOrder==="together")){if((p.flags&128)!==0)return p}else if(p.child!==null){p.child.return=p,p=p.child;continue}if(p===f)break;for(;p.sibling===null;){if(p.return===null||p.return===f)return null;p=p.return}p.sibling.return=p.return,p=p.sibling}return null}var Su=0,An=null,Oi=null,ga=null,lm=!1,Fc=!1,Dh=!1,kg=0,Lp=0,is=null,O4=0;function ea(){throw Error(r(321))}function Pg(f,p){if(p===null)return!1;for(var O=0;O<p.length&&O<f.length;O++)if(!pr(f[O],p[O]))return!1;return!0}function Mg(f,p,O,D,Q,Y){return Su=Y,An=p,p.memoizedState=null,p.updateQueue=null,p.lanes=0,K.H=f===null||f.memoizedState===null?Z1:Rh,Dh=!1,Y=O(D,Q),Dh=!1,Fc&&(Y=om(p,O,D,Q)),ky(f),Y}function ky(f){K.H=Rd;var p=Oi!==null&&Oi.next!==null;if(Su=0,ga=Oi=An=null,lm=!1,Lp=0,is=null,p)throw Error(r(300));f===null||xi||(f=f.dependencies,f!==null&&vd(f)&&(xi=!0))}function om(f,p,O,D){An=f;var Q=0;do{if(Fc&&(is=null),Lp=0,Fc=!1,25<=Q)throw Error(r(301));if(Q+=1,ga=Oi=null,f.updateQueue!=null){var Y=f.updateQueue;Y.lastEffect=null,Y.events=null,Y.stores=null,Y.memoCache!=null&&(Y.memoCache.index=0)}K.H=Ad,Y=p(O,D)}while(Fc);return Y}function x4(){var f=K.H,p=f.useState()[0];return p=typeof p.then=="function"?Bp(p):p,f=f.useState()[0],(Oi!==null?Oi.memoizedState:null)!==f&&(An.flags|=1024),p}function Fg(){var f=kg!==0;return kg=0,f}function Ng(f,p,O){p.updateQueue=f.updateQueue,p.flags&=-2053,f.lanes&=~O}function Py(f){if(lm){for(f=f.memoizedState;f!==null;){var p=f.queue;p!==null&&(p.pending=null),f=f.next}lm=!1}Su=0,ga=Oi=An=null,Fc=!1,Lp=kg=0,is=null}function al(){var f={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return ga===null?An.memoizedState=ga=f:ga=ga.next=f,ga}function Aa(){if(Oi===null){var f=An.alternate;f=f!==null?f.memoizedState:null}else f=Oi.next;var p=ga===null?An.memoizedState:ga.next;if(p!==null)ga=p,Oi=f;else{if(f===null)throw An.alternate===null?Error(r(467)):Error(r(310));Oi=f,f={memoizedState:Oi.memoizedState,baseState:Oi.baseState,baseQueue:Oi.baseQueue,queue:Oi.queue,next:null},ga===null?An.memoizedState=ga=f:ga=ga.next=f}return ga}function $h(){return{lastEffect:null,events:null,stores:null,memoCache:null}}function Bp(f){var p=Lp;return Lp+=1,is===null&&(is=[]),f=im(is,f,p),p=An,(ga===null?p.memoizedState:ga.next)===null&&(p=p.alternate,K.H=p===null||p.memoizedState===null?Z1:Rh),f}function Qp(f){if(f!==null&&typeof f=="object"){if(typeof f.then=="function")return Bp(f);if(f.$$typeof===$)return Ou(f)}throw Error(r(438,String(f)))}function My(f){var p=null,O=An.updateQueue;if(O!==null&&(p=O.memoCache),p==null){var D=An.alternate;D!==null&&(D=D.updateQueue,D!==null&&(D=D.memoCache,D!=null&&(p={data:D.data.map(function(Q){return Q.slice()}),index:0})))}if(p==null&&(p={data:[],index:0}),O===null&&(O=$h(),An.updateQueue=O),O.memoCache=p,O=p.data[p.index],O===void 0)for(O=p.data[p.index]=Array(f),D=0;D<f;D++)O[D]=U;return p.index++,O}function Vs(f,p){return typeof p=="function"?p(f):p}function Jn(f){var p=Aa();return Fy(p,Oi,f)}function Fy(f,p,O){var D=f.queue;if(D===null)throw Error(r(311));D.lastRenderedReducer=O;var Q=f.baseQueue,Y=D.pending;if(Y!==null){if(Q!==null){var ce=Q.next;Q.next=Y.next,Y.next=ce}p.baseQueue=Q=Y,D.pending=null}if(Y=f.baseState,Q===null)f.memoizedState=Y;else{p=Q.next;var _e=ce=null,Le=null,ut=p,pt=!1;do{var xt=ut.lane&-536870913;if(xt!==ut.lane?(_r&xt)===xt:(Su&xt)===xt){var ot=ut.revertLane;if(ot===0)Le!==null&&(Le=Le.next={lane:0,revertLane:0,gesture:null,action:ut.action,hasEagerState:ut.hasEagerState,eagerState:ut.eagerState,next:null}),xt===Df&&(pt=!0);else if((Su&ot)===ot){ut=ut.next,ot===Df&&(pt=!0);continue}else xt={lane:0,revertLane:ut.revertLane,gesture:null,action:ut.action,hasEagerState:ut.hasEagerState,eagerState:ut.eagerState,next:null},Le===null?(_e=Le=xt,ce=Y):Le=Le.next=xt,An.lanes|=ot,Ro|=ot;xt=ut.action,Dh&&O(Y,xt),Y=ut.hasEagerState?ut.eagerState:O(Y,xt)}else ot={lane:xt,revertLane:ut.revertLane,gesture:ut.gesture,action:ut.action,hasEagerState:ut.hasEagerState,eagerState:ut.eagerState,next:null},Le===null?(_e=Le=ot,ce=Y):Le=Le.next=ot,An.lanes|=xt,Ro|=xt;ut=ut.next}while(ut!==null&&ut!==p);if(Le===null?ce=Y:Le.next=_e,!pr(Y,f.memoizedState)&&(xi=!0,pt&&(O=Qs,O!==null)))throw O;f.memoizedState=Y,f.baseState=ce,f.baseQueue=Le,D.lastRenderedState=Y}return Q===null&&(D.lanes=0),[f.memoizedState,D.dispatch]}function Ig(f){var p=Aa(),O=p.queue;if(O===null)throw Error(r(311));O.lastRenderedReducer=f;var D=O.dispatch,Q=O.pending,Y=p.memoizedState;if(Q!==null){O.pending=null;var ce=Q=Q.next;do Y=f(Y,ce.action),ce=ce.next;while(ce!==Q);pr(Y,p.memoizedState)||(xi=!0),p.memoizedState=Y,p.baseQueue===null&&(p.baseState=Y),O.lastRenderedState=Y}return[Y,D]}function Ch(f,p,O){var D=An,Q=Aa(),Y=Ze;if(Y){if(O===void 0)throw Error(r(407));O=O()}else O=p();var ce=!pr((Oi||Q).memoizedState,O);if(ce&&(Q.memoizedState=O,xi=!0),Q=Q.queue,fm(Ly.bind(null,D,Q,f),[f]),Q.getSnapshot!==p||ce||ga!==null&&ga.memoizedState.tag&1){if(D.flags|=2048,So(9,{destroy:void 0},Iy.bind(null,D,Q,O,p),null),li===null)throw Error(r(349));Y||(Su&127)!==0||Ny(D,p,O)}return O}function Ny(f,p,O){f.flags|=16384,f={getSnapshot:p,value:O},p=An.updateQueue,p===null?(p=$h(),An.updateQueue=p,p.stores=[f]):(O=p.stores,O===null?p.stores=[f]:O.push(f))}function Iy(f,p,O,D){p.value=O,p.getSnapshot=D,Hl(p)&&Lg(f)}function Ly(f,p,O){return O(function(){Hl(p)&&Lg(f)})}function Hl(f){var p=f.getSnapshot;f=f.value;try{var O=p();return!pr(f,O)}catch{return!0}}function Lg(f){var p=el(f,2);p!==null&&b(p,f,2)}function Dd(f){var p=al();if(typeof f=="function"){var O=f;if(f=O(),Dh){nn(!0);try{O()}finally{nn(!1)}}}return p.memoizedState=p.baseState=f,p.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Vs,lastRenderedState:f},p}function Bg(f,p,O,D){return f.baseState=O,Fy(f,Oi,typeof D=="function"?D:Vs)}function z1(f,p,O,D,Q){if(Ic(f))throw Error(r(485));if(f=p.action,f!==null){var Y={payload:Q,action:f,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(ce){Y.listeners.push(ce)}};K.T!==null?O(!0):Y.isTransition=!1,D(Y),O=p.pending,O===null?(Y.next=p.pending=Y,By(p,Y)):(Y.next=O.next,p.pending=O.next=Y)}}function By(f,p){var O=p.action,D=p.payload,Q=f.state;if(p.isTransition){var Y=K.T,ce={};K.T=ce;try{var _e=O(Q,D),Le=K.S;Le!==null&&Le(ce,_e),Xl(f,p,_e)}catch(ut){zp(f,p,ut)}finally{Y!==null&&ce.types!==null&&(Y.types=ce.types),K.T=Y}}else try{Y=O(Q,D),Xl(f,p,Y)}catch(ut){zp(f,p,ut)}}function Xl(f,p,O){O!==null&&typeof O=="object"&&typeof O.then=="function"?O.then(function(D){Nc(f,p,D)},function(D){return zp(f,p,D)}):Nc(f,p,O)}function Nc(f,p,O){p.status="fulfilled",p.value=O,Qy(p),f.state=O,p=f.pending,p!==null&&(O=p.next,O===p?f.pending=null:(O=O.next,p.next=O,By(f,O)))}function zp(f,p,O){var D=f.pending;if(f.pending=null,D!==null){D=D.next;do p.status="rejected",p.reason=O,Qy(p),p=p.next;while(p!==D)}f.action=null}function Qy(f){f=f.listeners;for(var p=0;p<f.length;p++)(0,f[p])()}function sm(f,p){return p}function zy(f,p){if(Ze){var O=li.formState;if(O!==null){e:{var D=An;if(Ze){if(Be){t:{for(var Q=Be,Y=Kt;Q.nodeType!==8;){if(!Y){Q=null;break t}if(Q=Zc(Q.nextSibling),Q===null){Q=null;break t}}Y=Q.data,Q=Y==="F!"||Y==="F"?Q:null}if(Q){Be=Zc(Q.nextSibling),D=Q.data==="F!";break e}}ln(D)}D=!1}D&&(p=O[0])}}return O=al(),O.memoizedState=O.baseState=p,D={pending:null,lanes:0,dispatch:null,lastRenderedReducer:sm,lastRenderedState:p},O.queue=D,O=Yy.bind(null,An,D),D.dispatch=O,D=Dd(!1),Y=wd.bind(null,An,!1,D.queue),D=al(),Q={state:p,dispatch:null,action:f,pending:null},D.queue=Q,O=z1.bind(null,An,Q,Y,O),Q.dispatch=O,D.memoizedState=f,[p,O,!1]}function Qg(f){var p=Aa();return qy(p,Oi,f)}function qy(f,p,O){if(p=Fy(f,p,sm)[0],f=Jn(Vs)[0],typeof p=="object"&&p!==null&&typeof p.then=="function")try{var D=Bp(p)}catch(ce){throw ce===bd?es:ce}else D=p;p=Aa();var Q=p.queue,Y=Q.dispatch;return O!==p.memoizedState&&(An.flags|=2048,So(9,{destroy:void 0},Vy.bind(null,Q,O),null)),[D,Y,f]}function Vy(f,p){f.action=p}function zg(f){var p=Aa(),O=Oi;if(O!==null)return qy(p,O,f);Aa(),p=p.memoizedState,O=Aa();var D=O.queue.dispatch;return O.memoizedState=f,[p,D,!1]}function So(f,p,O,D){return f={tag:f,create:O,deps:D,inst:p,next:null},p=An.updateQueue,p===null&&(p=$h(),An.updateQueue=p),O=p.lastEffect,O===null?p.lastEffect=f.next=f:(D=O.next,O.next=f,f.next=D,p.lastEffect=f),f}function q1(){return Aa().memoizedState}function Us(f,p,O,D){var Q=al();An.flags|=f,Q.memoizedState=So(1|p,{destroy:void 0},O,D===void 0?null:D)}function cm(f,p,O,D){var Q=Aa();D=D===void 0?null:D;var Y=Q.memoizedState.inst;Oi!==null&&D!==null&&Pg(D,Oi.memoizedState.deps)?Q.memoizedState=So(p,Y,O,D):(An.flags|=f,Q.memoizedState=So(1|p,Y,O,D))}function Uy(f,p){Us(8390656,8,f,p)}function fm(f,p){cm(2048,8,f,p)}function V1(f){An.flags|=4;var p=An.updateQueue;if(p===null)p=$h(),An.updateQueue=p,p.events=[f];else{var O=p.events;O===null?p.events=[f]:O.push(f)}}function U1(f){var p=Aa().memoizedState;return V1({ref:p,nextImpl:f}),function(){if((Nr&2)!==0)throw Error(r(440));return p.impl.apply(void 0,arguments)}}function j1(f,p){return cm(4,2,f,p)}function Yl(f,p){return cm(4,4,f,p)}function qg(f,p){if(typeof p=="function"){f=f();var O=p(f);return function(){typeof O=="function"?O():p(null)}}if(p!=null)return f=f(),p.current=f,function(){p.current=null}}function Vg(f,p,O){O=O!=null?O.concat([f]):null,cm(4,4,qg.bind(null,p,f),O)}function dm(){}function Ug(f,p){var O=Aa();p=p===void 0?null:p;var D=O.memoizedState;return p!==null&&Pg(p,D[1])?D[0]:(O.memoizedState=[f,p],f)}function ul(f,p){var O=Aa();p=p===void 0?null:p;var D=O.memoizedState;if(p!==null&&Pg(p,D[1]))return D[0];if(D=f(),Dh){nn(!0);try{f()}finally{nn(!1)}}return O.memoizedState=[D,p],D}function as(f,p,O){return O===void 0||(Su&1073741824)!==0&&(_r&261930)===0?f.memoizedState=p:(f.memoizedState=O,f=g(),An.lanes|=f,Ro|=f,O)}function jy(f,p,O,D){return pr(O,p)?O:il.current!==null?(f=as(f,O,D),pr(f,p)||(xi=!0),f):(Su&42)===0||(Su&1073741824)!==0&&(_r&261930)===0?(xi=!0,f.memoizedState=O):(f=g(),An.lanes|=f,Ro|=f,p)}function Zy(f,p,O,D,Q){var Y=J.p;J.p=Y!==0&&8>Y?Y:8;var ce=K.T,_e={};K.T=_e,wd(f,!1,p,O);try{var Le=Q(),ut=K.S;if(ut!==null&&ut(_e,Le),Le!==null&&typeof Le=="object"&&typeof Le.then=="function"){var pt=Ay(Le,D);wh(f,p,pt,wl(f))}else wh(f,p,D,wl(f))}catch(xt){wh(f,p,{then:function(){},status:"rejected",reason:xt},wl())}finally{J.p=Y,ce!==null&&_e.types!==null&&(ce.types=_e.types),K.T=ce}}function jg(){}function $d(f,p,O,D){if(f.tag!==5)throw Error(r(476));var Q=Hy(f).queue;Zy(f,Q,p,te,O===null?jg:function(){return Zg(f),O(D)})}function Hy(f){var p=f.memoizedState;if(p!==null)return p;p={memoizedState:te,baseState:te,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Vs,lastRenderedState:te},next:null};var O={};return p.next={memoizedState:O,baseState:O,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Vs,lastRenderedState:O},next:null},f.memoizedState=p,f=f.alternate,f!==null&&(f.memoizedState=p),p}function Zg(f){var p=Hy(f);p.next===null&&(p=f.alternate.memoizedState),wh(f,p.next.queue,{},wl())}function Xy(){return Ou(cv)}function _u(){return Aa().memoizedState}function Cd(){return Aa().memoizedState}function hm(f){for(var p=f.return;p!==null;){switch(p.tag){case 24:case 3:var O=wl();f=Mc(O);var D=ns(p,f,O);D!==null&&(b(D,p,O),Sd(D,p,O)),p={cache:Bs()},f.payload=p;return}p=p.return}}function S4(f,p,O){var D=wl();O={lane:D,revertLane:0,gesture:null,action:O,hasEagerState:!1,eagerState:null,next:null},Ic(f)?Lc(p,O):(O=Cc(f,p,O,D),O!==null&&(b(O,f,D),ll(O,p,D)))}function Yy(f,p,O){var D=wl();wh(f,p,O,D)}function wh(f,p,O,D){var Q={lane:D,revertLane:0,gesture:null,action:O,hasEagerState:!1,eagerState:null,next:null};if(Ic(f))Lc(p,Q);else{var Y=f.alternate;if(f.lanes===0&&(Y===null||Y.lanes===0)&&(Y=p.lastRenderedReducer,Y!==null))try{var ce=p.lastRenderedState,_e=Y(ce,O);if(Q.hasEagerState=!0,Q.eagerState=_e,pr(_e,ce))return Ju(f,p,Q,0),li===null&&Wa(),!1}catch{}finally{}if(O=Cc(f,p,Q,D),O!==null)return b(O,f,D),ll(O,p,D),!0}return!1}function wd(f,p,O,D){if(D={lane:2,revertLane:Ws(),gesture:null,action:D,hasEagerState:!1,eagerState:null,next:null},Ic(f)){if(p)throw Error(r(479))}else p=Cc(f,O,D,2),p!==null&&b(p,f,2)}function Ic(f){var p=f.alternate;return f===An||p!==null&&p===An}function Lc(f,p){Fc=lm=!0;var O=f.pending;O===null?p.next=p:(p.next=O.next,O.next=p),f.pending=p}function ll(f,p,O){if((O&4194048)!==0){var D=p.lanes;D&=f.pendingLanes,O|=D,p.lanes=O,Te(f,O)}}var Rd={readContext:Ou,use:Qp,useCallback:ea,useContext:ea,useEffect:ea,useImperativeHandle:ea,useLayoutEffect:ea,useInsertionEffect:ea,useMemo:ea,useReducer:ea,useRef:ea,useState:ea,useDebugValue:ea,useDeferredValue:ea,useTransition:ea,useSyncExternalStore:ea,useId:ea,useHostTransitionStatus:ea,useFormState:ea,useActionState:ea,useOptimistic:ea,useMemoCache:ea,useCacheRefresh:ea};Rd.useEffectEvent=ea;var Z1={readContext:Ou,use:Qp,useCallback:function(f,p){return al().memoizedState=[f,p===void 0?null:p],f},useContext:Ou,useEffect:Uy,useImperativeHandle:function(f,p,O){O=O!=null?O.concat([f]):null,Us(4194308,4,qg.bind(null,p,f),O)},useLayoutEffect:function(f,p){return Us(4194308,4,f,p)},useInsertionEffect:function(f,p){Us(4,2,f,p)},useMemo:function(f,p){var O=al();p=p===void 0?null:p;var D=f();if(Dh){nn(!0);try{f()}finally{nn(!1)}}return O.memoizedState=[D,p],D},useReducer:function(f,p,O){var D=al();if(O!==void 0){var Q=O(p);if(Dh){nn(!0);try{O(p)}finally{nn(!1)}}}else Q=p;return D.memoizedState=D.baseState=Q,f={pending:null,lanes:0,dispatch:null,lastRenderedReducer:f,lastRenderedState:Q},D.queue=f,f=f.dispatch=S4.bind(null,An,f),[D.memoizedState,f]},useRef:function(f){var p=al();return f={current:f},p.memoizedState=f},useState:function(f){f=Dd(f);var p=f.queue,O=Yy.bind(null,An,p);return p.dispatch=O,[f.memoizedState,O]},useDebugValue:dm,useDeferredValue:function(f,p){var O=al();return as(O,f,p)},useTransition:function(){var f=Dd(!1);return f=Zy.bind(null,An,f.queue,!0,!1),al().memoizedState=f,[!1,f]},useSyncExternalStore:function(f,p,O){var D=An,Q=al();if(Ze){if(O===void 0)throw Error(r(407));O=O()}else{if(O=p(),li===null)throw Error(r(349));(_r&127)!==0||Ny(D,p,O)}Q.memoizedState=O;var Y={value:O,getSnapshot:p};return Q.queue=Y,Uy(Ly.bind(null,D,Y,f),[f]),D.flags|=2048,So(9,{destroy:void 0},Iy.bind(null,D,Y,O,p),null),O},useId:function(){var f=al(),p=li.identifierPrefix;if(Ze){var O=ue,D=Ko;O=(D&~(1<<32-Fe(D)-1)).toString(32)+O,p="_"+p+"R_"+O,O=kg++,0<O&&(p+="H"+O.toString(32)),p+="_"}else O=O4++,p="_"+p+"r_"+O.toString(32)+"_";return f.memoizedState=p},useHostTransitionStatus:Xy,useFormState:zy,useActionState:zy,useOptimistic:function(f){var p=al();p.memoizedState=p.baseState=f;var O={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return p.queue=O,p=wd.bind(null,An,!0,O),O.dispatch=p,[f,p]},useMemoCache:My,useCacheRefresh:function(){return al().memoizedState=hm.bind(null,An)},useEffectEvent:function(f){var p=al(),O={impl:f};return p.memoizedState=O,function(){if((Nr&2)!==0)throw Error(r(440));return O.impl.apply(void 0,arguments)}}},Rh={readContext:Ou,use:Qp,useCallback:Ug,useContext:Ou,useEffect:fm,useImperativeHandle:Vg,useInsertionEffect:j1,useLayoutEffect:Yl,useMemo:ul,useReducer:Jn,useRef:q1,useState:function(){return Jn(Vs)},useDebugValue:dm,useDeferredValue:function(f,p){var O=Aa();return jy(O,Oi.memoizedState,f,p)},useTransition:function(){var f=Jn(Vs)[0],p=Aa().memoizedState;return[typeof f=="boolean"?f:Bp(f),p]},useSyncExternalStore:Ch,useId:_u,useHostTransitionStatus:Xy,useFormState:Qg,useActionState:Qg,useOptimistic:function(f,p){var O=Aa();return Bg(O,Oi,f,p)},useMemoCache:My,useCacheRefresh:Cd};Rh.useEffectEvent=U1;var Ad={readContext:Ou,use:Qp,useCallback:Ug,useContext:Ou,useEffect:fm,useImperativeHandle:Vg,useInsertionEffect:j1,useLayoutEffect:Yl,useMemo:ul,useReducer:Ig,useRef:q1,useState:function(){return Ig(Vs)},useDebugValue:dm,useDeferredValue:function(f,p){var O=Aa();return Oi===null?as(O,f,p):jy(O,Oi.memoizedState,f,p)},useTransition:function(){var f=Ig(Vs)[0],p=Aa().memoizedState;return[typeof f=="boolean"?f:Bp(f),p]},useSyncExternalStore:Ch,useId:_u,useHostTransitionStatus:Xy,useFormState:zg,useActionState:zg,useOptimistic:function(f,p){var O=Aa();return Oi!==null?Bg(O,Oi,f,p):(O.baseState=f,[f,O.queue.dispatch])},useMemoCache:My,useCacheRefresh:Cd};Ad.useEffectEvent=U1;function Ah(f,p,O,D){p=f.memoizedState,O=O(D,p),O=O==null?p:d({},p,O),f.memoizedState=O,f.lanes===0&&(f.updateQueue.baseState=O)}var Th={enqueueSetState:function(f,p,O){f=f._reactInternals;var D=wl(),Q=Mc(D);Q.payload=p,O!=null&&(Q.callback=O),p=ns(f,Q,D),p!==null&&(b(p,f,D),Sd(p,f,D))},enqueueReplaceState:function(f,p,O){f=f._reactInternals;var D=wl(),Q=Mc(D);Q.tag=1,Q.payload=p,O!=null&&(Q.callback=O),p=ns(f,Q,D),p!==null&&(b(p,f,D),Sd(p,f,D))},enqueueForceUpdate:function(f,p){f=f._reactInternals;var O=wl(),D=Mc(O);D.tag=2,p!=null&&(D.callback=p),p=ns(f,D,O),p!==null&&(b(p,f,O),Sd(p,f,O))}};function H1(f,p,O,D,Q,Y,ce){return f=f.stateNode,typeof f.shouldComponentUpdate=="function"?f.shouldComponentUpdate(D,Y,ce):p.prototype&&p.prototype.isPureReactComponent?!xf(O,D)||!xf(Q,Y):!0}function Gy(f,p,O,D){f=p.state,typeof p.componentWillReceiveProps=="function"&&p.componentWillReceiveProps(O,D),typeof p.UNSAFE_componentWillReceiveProps=="function"&&p.UNSAFE_componentWillReceiveProps(O,D),p.state!==f&&Th.enqueueReplaceState(p,p.state,null)}function Hi(f,p){var O=p;if("ref"in p){O={};for(var D in p)D!=="ref"&&(O[D]=p[D])}if(f=f.defaultProps){O===p&&(O=d({},O));for(var Q in f)O[Q]===void 0&&(O[Q]=f[Q])}return O}function Hg(f){Go(f)}function Ky(f){console.error(f)}function Xg(f){Go(f)}function pm(f,p){try{var O=f.onUncaughtError;O(p.value,{componentStack:p.stack})}catch(D){setTimeout(function(){throw D})}}function X1(f,p,O){try{var D=f.onCaughtError;D(O.value,{componentStack:O.stack,errorBoundary:p.tag===1?p.stateNode:null})}catch(Q){setTimeout(function(){throw Q})}}function Wy(f,p,O){return O=Mc(O),O.tag=3,O.payload={element:null},O.callback=function(){pm(f,p)},O}function Gl(f){return f=Mc(f),f.tag=3,f}function ol(f,p,O,D){var Q=O.type.getDerivedStateFromError;if(typeof Q=="function"){var Y=D.value;f.payload=function(){return Q(Y)},f.callback=function(){X1(p,O,D)}}var ce=O.stateNode;ce!==null&&typeof ce.componentDidCatch=="function"&&(f.callback=function(){X1(p,O,D),typeof Q!="function"&&(Uc===null?Uc=new Set([this]):Uc.add(this));var _e=D.stack;this.componentDidCatch(D.value,{componentStack:_e!==null?_e:""})})}function Dl(f,p,O,D,Q){if(O.flags|=32768,D!==null&&typeof D=="object"&&typeof D.then=="function"){if(p=O.alternate,p!==null&&jl(p,O,Q,!0),O=ru.current,O!==null){switch(O.tag){case 31:case 13:return rs===null?Hn():O.alternate===null&&ti===0&&(ti=3),O.flags&=-257,O.flags|=65536,O.lanes=Q,D===rm?O.flags|=16384:(p=O.updateQueue,p===null?O.updateQueue=new Set([D]):p.add(D),D4(f,D,Q)),!1;case 22:return O.flags|=65536,D===rm?O.flags|=16384:(p=O.updateQueue,p===null?(p={transitions:null,markerInstances:null,retryQueue:new Set([D])},O.updateQueue=p):(O=p.retryQueue,O===null?p.retryQueue=new Set([D]):O.add(D)),D4(f,D,Q)),!1}throw Error(r(435,O.tag))}return D4(f,D,Q),Hn(),!1}if(Ze)return p=ru.current,p!==null?((p.flags&65536)===0&&(p.flags|=256),p.flags|=65536,p.lanes=Q,D!==zn&&(f=Error(r(422),{cause:D}),Vl(Mu(f,O)))):(D!==zn&&(p=Error(r(423),{cause:D}),Vl(Mu(p,O))),f=f.current.alternate,f.flags|=65536,Q&=-Q,f.lanes|=Q,D=Mu(D,O),Q=Wy(f.stateNode,D,Q),am(f,Q),ti!==4&&(ti=2)),!1;var Y=Error(r(520),{cause:D});if(Y=Mu(Y,O),Gs===null?Gs=[Y]:Gs.push(Y),ti!==4&&(ti=2),p===null)return!0;D=Mu(D,O),O=p;do{switch(O.tag){case 3:return O.flags|=65536,f=Q&-Q,O.lanes|=f,f=Wy(O.stateNode,D,f),am(O,f),!1;case 1:if(p=O.type,Y=O.stateNode,(O.flags&128)===0&&(typeof p.getDerivedStateFromError=="function"||Y!==null&&typeof Y.componentDidCatch=="function"&&(Uc===null||!Uc.has(Y))))return O.flags|=65536,Q&=-Q,O.lanes|=Q,Q=Gl(Q),ol(Q,f,O,D),am(O,Q),!1}O=O.return}while(O!==null);return!1}var qp=Error(r(461)),xi=!1;function Ai(f,p,O,D){p.child=f===null?Ty(p,null,O,D):xd(p,f.child,O,D)}function On(f,p,O,D,Q){O=O.render;var Y=p.ref;if("ref"in D){var ce={};for(var _e in D)_e!=="ref"&&(ce[_e]=D[_e])}else ce=D;return Sh(p),D=Mg(f,p,O,ce,Y,Q),_e=Fg(),f!==null&&!xi?(Ng(f,p,Q),os(f,p,Q)):(Ze&&_e&&fe(p),p.flags|=1,Ai(f,p,D,Q),p.child)}function Td(f,p,O,D,Q){if(f===null){var Y=O.type;return typeof Y=="function"&&!Ji(Y)&&Y.defaultProps===void 0&&O.compare===null?(p.tag=15,p.type=Y,$l(f,p,Y,D,Q)):(f=Oh(O.type,null,D,p,p.mode,Q),f.ref=p.ref,f.return=p,p.child=f)}if(Y=f.child,!Cf(f,Q)){var ce=Y.memoizedProps;if(O=O.compare,O=O!==null?O:xf,O(ce,D)&&f.ref===p.ref)return os(f,p,Q)}return p.flags|=1,f=tl(Y,D),f.ref=p.ref,f.return=p,p.child=f}function $l(f,p,O,D,Q){if(f!==null){var Y=f.memoizedProps;if(xf(Y,D)&&f.ref===p.ref)if(xi=!1,p.pendingProps=D=Y,Cf(f,Q))(f.flags&131072)!==0&&(xi=!0);else return p.lanes=f.lanes,os(f,p,Q)}return kh(f,p,O,D,Q)}function iu(f,p,O,D){var Q=D.children,Y=f!==null?f.memoizedState:null;if(f===null&&p.stateNode===null&&(p.stateNode={_visibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null}),D.mode==="hidden"){if((p.flags&128)!==0){if(Y=Y!==null?Y.baseLanes|O:O,f!==null){for(D=p.child=f.child,Q=0;D!==null;)Q=Q|D.lanes|D.childLanes,D=D.sibling;D=Q&~Y}else D=0,p.child=null;return us(f,p,Y,O,D)}if((O&536870912)!==0)p.memoizedState={baseLanes:0,cachePool:null},f!==null&&Mp(p,Y!==null?Y.cachePool:null),Y!==null?Ag(p,Y):Qu(),Q1(p);else return D=p.lanes=536870912,us(f,p,Y!==null?Y.baseLanes|O:O,O,D)}else Y!==null?(Mp(p,Y.cachePool),Ag(p,Y),xo(),p.memoizedState=null):(f!==null&&Mp(p,null),Qu(),xo());return Ai(f,p,Q,O),p.child}function js(f,p){return f!==null&&f.tag===22||p.stateNode!==null||(p.stateNode={_visibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null}),p.sibling}function us(f,p,O,D,Q){var Y=kc();return Y=Y===null?null:{parent:gn._currentValue,pool:Y},p.memoizedState={baseLanes:O,cachePool:Y},f!==null&&Mp(p,null),Qu(),Q1(p),f!==null&&jl(f,p,D,!0),p.childLanes=Q,null}function _o(f,p){return p=Vp({mode:p.mode,children:p.children},f.mode),p.ref=f.ref,f.child=p,p.return=f,p}function qr(f,p,O){return xd(p,f.child,null,O),f=_o(p,p.pendingProps),f.flags|=2,Zl(p),p.memoizedState=null,f}function mm(f,p,O){var D=p.pendingProps,Q=(p.flags&128)!==0;if(p.flags&=-129,f===null){if(Ze){if(D.mode==="hidden")return f=_o(p,D),p.lanes=536870912,js(null,f);if(um(p),(f=Be)?(f=o_(f,Kt),f=f!==null&&f.data==="&"?f:null,f!==null&&(p.memoizedState={dehydrated:f,treeContext:Nu!==null?{id:Ko,overflow:ue}:null,retryLane:536870912,hydrationErrors:null},O=xh(f),O.return=p,p.child=O,Qe=p,Be=null)):f=null,f===null)throw ln(p);return p.lanes=536870912,null}return _o(p,D)}var Y=f.memoizedState;if(Y!==null){var ce=Y.dehydrated;if(um(p),Q)if(p.flags&256)p.flags&=-257,p=qr(f,p,O);else if(p.memoizedState!==null)p.child=f.child,p.flags|=128,p=null;else throw Error(r(558));else if(xi||jl(f,p,O,!1),Q=(O&f.childLanes)!==0,xi||Q){if(D=li,D!==null&&(ce=Ve(D,O),ce!==0&&ce!==Y.retryLane))throw Y.retryLane=ce,el(f,ce),b(D,f,ce),qp;Hn(),p=qr(f,p,O)}else f=Y.treeContext,Be=Zc(ce.nextSibling),Qe=p,Ze=!0,Rt=null,Kt=!1,f!==null&&ke(p,f),p=_o(p,D),p.flags|=4096;return p}return f=tl(f.child,{mode:D.mode,children:D.children}),f.ref=p.ref,p.child=f,f.return=p,f}function Zs(f,p){var O=p.ref;if(O===null)f!==null&&f.ref!==null&&(p.flags|=4194816);else{if(typeof O!="function"&&typeof O!="object")throw Error(r(284));(f===null||f.ref!==O)&&(p.flags|=4194816)}}function kh(f,p,O,D,Q){return Sh(p),O=Mg(f,p,O,D,void 0,Q),D=Fg(),f!==null&&!xi?(Ng(f,p,Q),os(f,p,Q)):(Ze&&D&&fe(p),p.flags|=1,Ai(f,p,O,Q),p.child)}function sl(f,p,O,D,Q,Y){return Sh(p),p.updateQueue=null,O=om(p,D,O,Q),ky(f),D=Fg(),f!==null&&!xi?(Ng(f,p,Y),os(f,p,Y)):(Ze&&D&&fe(p),p.flags|=1,Ai(f,p,O,Y),p.child)}function Eu(f,p,O,D,Q){if(Sh(p),p.stateNode===null){var Y=vu,ce=O.contextType;typeof ce=="object"&&ce!==null&&(Y=Ou(ce)),Y=new O(D,Y),p.memoizedState=Y.state!==null&&Y.state!==void 0?Y.state:null,Y.updater=Th,p.stateNode=Y,Y._reactInternals=p,Y=p.stateNode,Y.props=D,Y.state=p.memoizedState,Y.refs={},Pc(p),ce=O.contextType,Y.context=typeof ce=="object"&&ce!==null?Ou(ce):vu,Y.state=p.memoizedState,ce=O.getDerivedStateFromProps,typeof ce=="function"&&(Ah(p,O,ce,D),Y.state=p.memoizedState),typeof O.getDerivedStateFromProps=="function"||typeof Y.getSnapshotBeforeUpdate=="function"||typeof Y.UNSAFE_componentWillMount!="function"&&typeof Y.componentWillMount!="function"||(ce=Y.state,typeof Y.componentWillMount=="function"&&Y.componentWillMount(),typeof Y.UNSAFE_componentWillMount=="function"&&Y.UNSAFE_componentWillMount(),ce!==Y.state&&Th.enqueueReplaceState(Y,Y.state,null),_d(p,D,Y,Q),Eh(),Y.state=p.memoizedState),typeof Y.componentDidMount=="function"&&(p.flags|=4194308),D=!0}else if(f===null){Y=p.stateNode;var _e=p.memoizedProps,Le=Hi(O,_e);Y.props=Le;var ut=Y.context,pt=O.contextType;ce=vu,typeof pt=="object"&&pt!==null&&(ce=Ou(pt));var xt=O.getDerivedStateFromProps;pt=typeof xt=="function"||typeof Y.getSnapshotBeforeUpdate=="function",_e=p.pendingProps!==_e,pt||typeof Y.UNSAFE_componentWillReceiveProps!="function"&&typeof Y.componentWillReceiveProps!="function"||(_e||ut!==ce)&&Gy(p,Y,D,ce),ts=!1;var ot=p.memoizedState;Y.state=ot,_d(p,D,Y,Q),Eh(),ut=p.memoizedState,_e||ot!==ut||ts?(typeof xt=="function"&&(Ah(p,O,xt,D),ut=p.memoizedState),(Le=ts||H1(p,O,Le,D,ot,ut,ce))?(pt||typeof Y.UNSAFE_componentWillMount!="function"&&typeof Y.componentWillMount!="function"||(typeof Y.componentWillMount=="function"&&Y.componentWillMount(),typeof Y.UNSAFE_componentWillMount=="function"&&Y.UNSAFE_componentWillMount()),typeof Y.componentDidMount=="function"&&(p.flags|=4194308)):(typeof Y.componentDidMount=="function"&&(p.flags|=4194308),p.memoizedProps=D,p.memoizedState=ut),Y.props=D,Y.state=ut,Y.context=ce,D=Le):(typeof Y.componentDidMount=="function"&&(p.flags|=4194308),D=!1)}else{Y=p.stateNode,Rg(f,p),ce=p.memoizedProps,pt=Hi(O,ce),Y.props=pt,xt=p.pendingProps,ot=Y.context,ut=O.contextType,Le=vu,typeof ut=="object"&&ut!==null&&(Le=Ou(ut)),_e=O.getDerivedStateFromProps,(ut=typeof _e=="function"||typeof Y.getSnapshotBeforeUpdate=="function")||typeof Y.UNSAFE_componentWillReceiveProps!="function"&&typeof Y.componentWillReceiveProps!="function"||(ce!==xt||ot!==Le)&&Gy(p,Y,D,Le),ts=!1,ot=p.memoizedState,Y.state=ot,_d(p,D,Y,Q),Eh();var ft=p.memoizedState;ce!==xt||ot!==ft||ts||f!==null&&f.dependencies!==null&&vd(f.dependencies)?(typeof _e=="function"&&(Ah(p,O,_e,D),ft=p.memoizedState),(pt=ts||H1(p,O,pt,D,ot,ft,Le)||f!==null&&f.dependencies!==null&&vd(f.dependencies))?(ut||typeof Y.UNSAFE_componentWillUpdate!="function"&&typeof Y.componentWillUpdate!="function"||(typeof Y.componentWillUpdate=="function"&&Y.componentWillUpdate(D,ft,Le),typeof Y.UNSAFE_componentWillUpdate=="function"&&Y.UNSAFE_componentWillUpdate(D,ft,Le)),typeof Y.componentDidUpdate=="function"&&(p.flags|=4),typeof Y.getSnapshotBeforeUpdate=="function"&&(p.flags|=1024)):(typeof Y.componentDidUpdate!="function"||ce===f.memoizedProps&&ot===f.memoizedState||(p.flags|=4),typeof Y.getSnapshotBeforeUpdate!="function"||ce===f.memoizedProps&&ot===f.memoizedState||(p.flags|=1024),p.memoizedProps=D,p.memoizedState=ft),Y.props=D,Y.state=ft,Y.context=Le,D=pt):(typeof Y.componentDidUpdate!="function"||ce===f.memoizedProps&&ot===f.memoizedState||(p.flags|=4),typeof Y.getSnapshotBeforeUpdate!="function"||ce===f.memoizedProps&&ot===f.memoizedState||(p.flags|=1024),D=!1)}return Y=D,Zs(f,p),D=(p.flags&128)!==0,Y||D?(Y=p.stateNode,O=D&&typeof O.getDerivedStateFromError!="function"?null:Y.render(),p.flags|=1,f!==null&&D?(p.child=xd(p,f.child,null,Q),p.child=xd(p,null,O,Q)):Ai(f,p,O,Q),p.memoizedState=Y.state,f=p.child):f=os(f,p,Q),f}function Vr(f,p,O,D){return Se(),p.flags|=256,Ai(f,p,O,D),p.child}var Eo={dehydrated:null,treeContext:null,retryLane:0,hydrationErrors:null};function ls(f){return{baseLanes:f,cachePool:$g()}}function Do(f,p,O){return f=f!==null?f.childLanes&~O:0,p&&(f|=Ir),f}function Ti(f,p,O){var D=p.pendingProps,Q=!1,Y=(p.flags&128)!==0,ce;if((ce=Y)||(ce=f!==null&&f.memoizedState===null?!1:(ma.current&2)!==0),ce&&(Q=!0,p.flags&=-129),ce=(p.flags&32)!==0,p.flags&=-33,f===null){if(Ze){if(Q?Ed(p):xo(),(f=Be)?(f=o_(f,Kt),f=f!==null&&f.data!=="&"?f:null,f!==null&&(p.memoizedState={dehydrated:f,treeContext:Nu!==null?{id:Ko,overflow:ue}:null,retryLane:536870912,hydrationErrors:null},O=xh(f),O.return=p,p.child=O,Qe=p,Be=null)):f=null,f===null)throw ln(p);return iv(f)?p.lanes=32:p.lanes=536870912,null}var _e=D.children;return D=D.fallback,Q?(xo(),Q=p.mode,_e=Vp({mode:"hidden",children:_e},Q),D=zl(D,Q,O,null),_e.return=p,D.return=p,_e.sibling=D,p.child=_e,D=p.child,D.memoizedState=ls(O),D.childLanes=Do(f,ce,O),p.memoizedState=Eo,js(null,D)):(Ed(p),Bc(p,_e))}var Le=f.memoizedState;if(Le!==null&&(_e=Le.dehydrated,_e!==null)){if(Y)p.flags&256?(Ed(p),p.flags&=-257,p=kd(f,p,O)):p.memoizedState!==null?(xo(),p.child=f.child,p.flags|=128,p=null):(xo(),_e=D.fallback,Q=p.mode,D=Vp({mode:"visible",children:D.children},Q),_e=zl(_e,Q,O,null),_e.flags|=2,D.return=p,_e.return=p,D.sibling=_e,p.child=D,xd(p,f.child,null,O),D=p.child,D.memoizedState=ls(O),D.childLanes=Do(f,ce,O),p.memoizedState=Eo,p=js(null,D));else if(Ed(p),iv(_e)){if(ce=_e.nextSibling&&_e.nextSibling.dataset,ce)var ut=ce.dgst;ce=ut,D=Error(r(419)),D.stack="",D.digest=ce,Vl({value:D,source:null,stack:null}),p=kd(f,p,O)}else if(xi||jl(f,p,O,!1),ce=(O&f.childLanes)!==0,xi||ce){if(ce=li,ce!==null&&(D=Ve(ce,O),D!==0&&D!==Le.retryLane))throw Le.retryLane=D,el(f,D),b(ce,f,D),qp;rO(_e)||Hn(),p=kd(f,p,O)}else rO(_e)?(p.flags|=192,p.child=f.child,p=null):(f=Le.treeContext,Be=Zc(_e.nextSibling),Qe=p,Ze=!0,Rt=null,Kt=!1,f!==null&&ke(p,f),p=Bc(p,D.children),p.flags|=4096);return p}return Q?(xo(),_e=D.fallback,Q=p.mode,Le=f.child,ut=Le.sibling,D=tl(Le,{mode:"hidden",children:D.children}),D.subtreeFlags=Le.subtreeFlags&65011712,ut!==null?_e=tl(ut,_e):(_e=zl(_e,Q,O,null),_e.flags|=2),_e.return=p,D.return=p,D.sibling=_e,p.child=D,js(null,D),D=p.child,_e=f.child.memoizedState,_e===null?_e=ls(O):(Q=_e.cachePool,Q!==null?(Le=gn._currentValue,Q=Q.parent!==Le?{parent:Le,pool:Le}:Q):Q=$g(),_e={baseLanes:_e.baseLanes|O,cachePool:Q}),D.memoizedState=_e,D.childLanes=Do(f,ce,O),p.memoizedState=Eo,js(f.child,D)):(Ed(p),O=f.child,f=O.sibling,O=tl(O,{mode:"visible",children:D.children}),O.return=p,O.sibling=null,f!==null&&(ce=p.deletions,ce===null?(p.deletions=[f],p.flags|=16):ce.push(f)),p.child=O,p.memoizedState=null,O)}function Bc(f,p){return p=Vp({mode:"visible",children:p},f.mode),p.return=f,f.child=p}function Vp(f,p){return f=aa(22,f,null,p),f.lanes=0,f}function kd(f,p,O){return xd(p,f.child,null,O),f=Bc(p,p.pendingProps.children),f.flags|=2,p.memoizedState=null,f}function Ph(f,p,O){f.lanes|=p;var D=f.alternate;D!==null&&(D.lanes|=p),Lu(f.return,p,O)}function gm(f,p,O,D,Q,Y){var ce=f.memoizedState;ce===null?f.memoizedState={isBackwards:p,rendering:null,renderingStartTime:0,last:D,tail:O,tailMode:Q,treeForkCount:Y}:(ce.isBackwards=p,ce.rendering=null,ce.renderingStartTime=0,ce.last=D,ce.tail=O,ce.tailMode=Q,ce.treeForkCount=Y)}function $f(f,p,O){var D=p.pendingProps,Q=D.revealOrder,Y=D.tail;D=D.children;var ce=ma.current,_e=(ce&2)!==0;if(_e?(ce=ce&1|2,p.flags|=128):ce&=1,B(ma,ce),Ai(f,p,D,O),D=Ze?wi:0,!_e&&f!==null&&(f.flags&128)!==0)e:for(f=p.child;f!==null;){if(f.tag===13)f.memoizedState!==null&&Ph(f,O,p);else if(f.tag===19)Ph(f,O,p);else if(f.child!==null){f.child.return=f,f=f.child;continue}if(f===p)break e;for(;f.sibling===null;){if(f.return===null||f.return===p)break e;f=f.return}f.sibling.return=f.return,f=f.sibling}switch(Q){case"forwards":for(O=p.child,Q=null;O!==null;)f=O.alternate,f!==null&&Tg(f)===null&&(Q=O),O=O.sibling;O=Q,O===null?(Q=p.child,p.child=null):(Q=O.sibling,O.sibling=null),gm(p,!1,Q,O,Y,D);break;case"backwards":case"unstable_legacy-backwards":for(O=null,Q=p.child,p.child=null;Q!==null;){if(f=Q.alternate,f!==null&&Tg(f)===null){p.child=Q;break}f=Q.sibling,Q.sibling=O,O=Q,Q=f}gm(p,!0,O,null,Y,D);break;case"together":gm(p,!1,null,null,void 0,D);break;default:p.memoizedState=null}return p.child}function os(f,p,O){if(f!==null&&(p.dependencies=f.dependencies),Ro|=p.lanes,(O&p.childLanes)===0)if(f!==null){if(jl(f,p,O,!1),(O&p.childLanes)===0)return null}else return null;if(f!==null&&p.child!==f.child)throw Error(r(153));if(p.child!==null){for(f=p.child,O=tl(f,f.pendingProps),p.child=O,O.return=p;f.sibling!==null;)f=f.sibling,O=O.sibling=tl(f,f.pendingProps),O.return=p;O.sibling=null}return p.child}function Cf(f,p){return(f.lanes&p)!==0?!0:(f=f.dependencies,!!(f!==null&&vd(f)))}function vm(f,p,O){switch(p.tag){case 3:ge(p,p.stateNode.containerInfo),Ls(p,gn,f.memoizedState.cache),Se();break;case 27:case 5:me(p);break;case 4:ge(p,p.stateNode.containerInfo);break;case 10:Ls(p,p.type,p.memoizedProps.value);break;case 31:if(p.memoizedState!==null)return p.flags|=128,um(p),null;break;case 13:var D=p.memoizedState;if(D!==null)return D.dehydrated!==null?(Ed(p),p.flags|=128,null):(O&p.child.childLanes)!==0?Ti(f,p,O):(Ed(p),f=os(f,p,O),f!==null?f.sibling:null);Ed(p);break;case 19:var Q=(f.flags&128)!==0;if(D=(O&p.childLanes)!==0,D||(jl(f,p,O,!1),D=(O&p.childLanes)!==0),Q){if(D)return $f(f,p,O);p.flags|=128}if(Q=p.memoizedState,Q!==null&&(Q.rendering=null,Q.tail=null,Q.lastEffect=null),B(ma,ma.current),D)break;return null;case 22:return p.lanes=0,iu(f,p,O,p.pendingProps);case 24:Ls(p,gn,f.memoizedState.cache)}return os(f,p,O)}function Yg(f,p,O){if(f!==null)if(f.memoizedProps!==p.pendingProps)xi=!0;else{if(!Cf(f,O)&&(p.flags&128)===0)return xi=!1,vm(f,p,O);xi=(f.flags&131072)!==0}else xi=!1,Ze&&(p.flags&1048576)!==0&&re(p,wi,p.index);switch(p.lanes=0,p.tag){case 16:e:{var D=p.pendingProps;if(f=zs(p.elementType),p.type=f,typeof f=="function")Ji(f)?(D=Hi(f,D),p.tag=1,p=Eu(null,p,f,D,O)):(p.tag=0,p=kh(null,p,f,D,O));else{if(f!=null){var Q=f.$$typeof;if(Q===k){p.tag=11,p=On(null,p,f,D,O);break e}else if(Q===R){p.tag=14,p=Td(null,p,f,D,O);break e}}throw p=j(f)||f,Error(r(306,p,""))}}return p;case 0:return kh(f,p,p.type,p.pendingProps,O);case 1:return D=p.type,Q=Hi(D,p.pendingProps),Eu(f,p,D,Q,O);case 3:e:{if(ge(p,p.stateNode.containerInfo),f===null)throw Error(r(387));D=p.pendingProps;var Y=p.memoizedState;Q=Y.element,Rg(f,p),_d(p,D,null,O);var ce=p.memoizedState;if(D=ce.cache,Ls(p,gn,D),D!==Y.cache&&Tc(p,[gn],O,!0),Eh(),D=ce.element,Y.isDehydrated)if(Y={element:D,isDehydrated:!1,cache:ce.cache},p.updateQueue.baseState=Y,p.memoizedState=Y,p.flags&256){p=Vr(f,p,D,O);break e}else if(D!==Q){Q=Mu(Error(r(424)),p),Vl(Q),p=Vr(f,p,D,O);break e}else{switch(f=p.stateNode.containerInfo,f.nodeType){case 9:f=f.body;break;default:f=f.nodeName==="HTML"?f.ownerDocument.body:f}for(Be=Zc(f.firstChild),Qe=p,Ze=!0,Rt=null,Kt=!0,O=Ty(p,null,D,O),p.child=O;O;)O.flags=O.flags&-3|4096,O=O.sibling}else{if(Se(),D===Q){p=os(f,p,O);break e}Ai(f,p,D,O)}p=p.child}return p;case 26:return Zs(f,p),f===null?(O=p_(p.type,null,p.pendingProps,null))?p.memoizedState=O:Ze||(O=p.type,f=p.pendingProps,D=lb(ie.current).createElement(O),D[Dt]=p,D[Lt]=f,Rl(D,O,f),yi(D),p.stateNode=D):p.memoizedState=p_(p.type,f.memoizedProps,p.pendingProps,f.memoizedState),null;case 27:return me(p),f===null&&Ze&&(D=p.stateNode=f_(p.type,p.pendingProps,ie.current),Qe=p,Kt=!0,Q=Be,Vn(p.type)?(I4=Q,Be=Zc(D.firstChild)):Be=Q),Ai(f,p,p.pendingProps.children,O),Zs(f,p),f===null&&(p.flags|=4194304),p.child;case 5:return f===null&&Ze&&((Q=D=Be)&&(D=T5(D,p.type,p.pendingProps,Kt),D!==null?(p.stateNode=D,Qe=p,Be=Zc(D.firstChild),Kt=!1,Q=!0):Q=!1),Q||ln(p)),me(p),Q=p.type,Y=p.pendingProps,ce=f!==null?f.memoizedProps:null,D=Y.children,Gp(Q,Y)?D=null:ce!==null&&Gp(Q,ce)&&(p.flags|=32),p.memoizedState!==null&&(Q=Mg(f,p,x4,null,null,O),cv._currentValue=Q),Zs(f,p),Ai(f,p,D,O),p.child;case 6:return f===null&&Ze&&((f=O=Be)&&(O=_i(O,p.pendingProps,Kt),O!==null?(p.stateNode=O,Qe=p,Be=null,f=!0):f=!1),f||ln(p)),null;case 13:return Ti(f,p,O);case 4:return ge(p,p.stateNode.containerInfo),D=p.pendingProps,f===null?p.child=xd(p,null,D,O):Ai(f,p,D,O),p.child;case 11:return On(f,p,p.type,p.pendingProps,O);case 7:return Ai(f,p,p.pendingProps,O),p.child;case 8:return Ai(f,p,p.pendingProps.children,O),p.child;case 12:return Ai(f,p,p.pendingProps.children,O),p.child;case 10:return D=p.pendingProps,Ls(p,p.type,D.value),Ai(f,p,D.children,O),p.child;case 9:return Q=p.type._context,D=p.pendingProps.children,Sh(p),Q=Ou(Q),D=D(Q),p.flags|=1,Ai(f,p,D,O),p.child;case 14:return Td(f,p,p.type,p.pendingProps,O);case 15:return $l(f,p,p.type,p.pendingProps,O);case 19:return $f(f,p,O);case 31:return mm(f,p,O);case 22:return iu(f,p,O,p.pendingProps);case 24:return Sh(p),D=Ou(gn),f===null?(Q=kc(),Q===null&&(Q=li,Y=Bs(),Q.pooledCache=Y,Y.refCount++,Y!==null&&(Q.pooledCacheLanes|=O),Q=Y),p.memoizedState={parent:D,cache:Q},Pc(p),Ls(p,gn,Q)):((f.lanes&O)!==0&&(Rg(f,p),_d(p,null,null,O),Eh()),Q=f.memoizedState,Y=p.memoizedState,Q.parent!==D?(Q={parent:D,cache:D},p.memoizedState=Q,p.lanes===0&&(p.memoizedState=p.updateQueue.baseState=Q),Ls(p,gn,D)):(D=Y.cache,Ls(p,gn,D),D!==Q.cache&&Tc(p,[gn],O,!0))),Ai(f,p,p.pendingProps.children,O),p.child;case 29:throw p.pendingProps}throw Error(r(156,p.tag))}function Kl(f){f.flags|=4}function Pd(f,p,O,D,Q){if((p=(f.mode&32)!==0)&&(p=!1),p){if(f.flags|=16777216,(Q&335544128)===Q)if(f.stateNode.complete)f.flags|=8192;else if(Pt())f.flags|=8192;else throw xn=rm,Fp}else f.flags&=-16777217}function Qc(f,p){if(p.type!=="stylesheet"||(p.state.loading&4)!==0)f.flags&=-16777217;else if(f.flags|=16777216,!b_(p))if(Pt())f.flags|=8192;else throw xn=rm,Fp}function Mh(f,p){p!==null&&(f.flags|=4),f.flags&16384&&(p=f.tag!==22?wr():536870912,f.lanes|=p,Ao|=p)}function ym(f,p){if(!Ze)switch(f.tailMode){case"hidden":p=f.tail;for(var O=null;p!==null;)p.alternate!==null&&(O=p),p=p.sibling;O===null?f.tail=null:O.sibling=null;break;case"collapsed":O=f.tail;for(var D=null;O!==null;)O.alternate!==null&&(D=O),O=O.sibling;D===null?p||f.tail===null?f.tail=null:f.tail.sibling=null:D.sibling=null}}function Wr(f){var p=f.alternate!==null&&f.alternate.child===f.child,O=0,D=0;if(p)for(var Q=f.child;Q!==null;)O|=Q.lanes|Q.childLanes,D|=Q.subtreeFlags&65011712,D|=Q.flags&65011712,Q.return=f,Q=Q.sibling;else for(Q=f.child;Q!==null;)O|=Q.lanes|Q.childLanes,D|=Q.subtreeFlags,D|=Q.flags,Q.return=f,Q=Q.sibling;return f.subtreeFlags|=D,f.childLanes=O,p}function Gg(f,p,O){var D=p.pendingProps;switch(xe(p),p.tag){case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return Wr(p),null;case 1:return Wr(p),null;case 3:return O=p.stateNode,D=null,f!==null&&(D=f.memoizedState.cache),p.memoizedState.cache!==D&&(p.flags|=2048),Ul(gn),ye(),O.pendingContext&&(O.context=O.pendingContext,O.pendingContext=null),(f===null||f.child===null)&&(za(p)?Kl(p):f===null||f.memoizedState.isDehydrated&&(p.flags&256)===0||(p.flags|=1024,El())),Wr(p),null;case 26:var Q=p.type,Y=p.memoizedState;return f===null?(Kl(p),Y!==null?(Wr(p),Qc(p,Y)):(Wr(p),Pd(p,Q,null,D,O))):Y?Y!==f.memoizedState?(Kl(p),Wr(p),Qc(p,Y)):(Wr(p),p.flags&=-16777217):(f=f.memoizedProps,f!==D&&Kl(p),Wr(p),Pd(p,Q,f,D,O)),null;case 27:if(Ee(p),O=ie.current,Q=p.type,f!==null&&p.stateNode!=null)f.memoizedProps!==D&&Kl(p);else{if(!D){if(p.stateNode===null)throw Error(r(166));return Wr(p),null}f=ne.current,za(p)?mn(p):(f=f_(Q,D,O),p.stateNode=f,Kl(p))}return Wr(p),null;case 5:if(Ee(p),Q=p.type,f!==null&&p.stateNode!=null)f.memoizedProps!==D&&Kl(p);else{if(!D){if(p.stateNode===null)throw Error(r(166));return Wr(p),null}if(Y=ne.current,za(p))mn(p);else{var ce=lb(ie.current);switch(Y){case 1:Y=ce.createElementNS("http://www.w3.org/2000/svg",Q);break;case 2:Y=ce.createElementNS("http://www.w3.org/1998/Math/MathML",Q);break;default:switch(Q){case"svg":Y=ce.createElementNS("http://www.w3.org/2000/svg",Q);break;case"math":Y=ce.createElementNS("http://www.w3.org/1998/Math/MathML",Q);break;case"script":Y=ce.createElement("div"),Y.innerHTML="<script><\/script>",Y=Y.removeChild(Y.firstChild);break;case"select":Y=typeof D.is=="string"?ce.createElement("select",{is:D.is}):ce.createElement("select"),D.multiple?Y.multiple=!0:D.size&&(Y.size=D.size);break;default:Y=typeof D.is=="string"?ce.createElement(Q,{is:D.is}):ce.createElement(Q)}}Y[Dt]=p,Y[Lt]=D;e:for(ce=p.child;ce!==null;){if(ce.tag===5||ce.tag===6)Y.appendChild(ce.stateNode);else if(ce.tag!==4&&ce.tag!==27&&ce.child!==null){ce.child.return=ce,ce=ce.child;continue}if(ce===p)break e;for(;ce.sibling===null;){if(ce.return===null||ce.return===p)break e;ce=ce.return}ce.sibling.return=ce.return,ce=ce.sibling}p.stateNode=Y;e:switch(Rl(Y,Q,D),Q){case"button":case"input":case"select":case"textarea":D=!!D.autoFocus;break e;case"img":D=!0;break e;default:D=!1}D&&Kl(p)}}return Wr(p),Pd(p,p.type,f===null?null:f.memoizedProps,p.pendingProps,O),null;case 6:if(f&&p.stateNode!=null)f.memoizedProps!==D&&Kl(p);else{if(typeof D!="string"&&p.stateNode===null)throw Error(r(166));if(f=ie.current,za(p)){if(f=p.stateNode,O=p.memoizedProps,D=null,Q=Qe,Q!==null)switch(Q.tag){case 27:case 5:D=Q.memoizedProps}f[Dt]=p,f=!!(f.nodeValue===O||D!==null&&D.suppressHydrationWarning===!0||i_(f.nodeValue,O)),f||ln(p,!0)}else f=lb(f).createTextNode(D),f[Dt]=p,p.stateNode=f}return Wr(p),null;case 31:if(O=p.memoizedState,f===null||f.memoizedState!==null){if(D=za(p),O!==null){if(f===null){if(!D)throw Error(r(318));if(f=p.memoizedState,f=f!==null?f.dehydrated:null,!f)throw Error(r(557));f[Dt]=p}else Se(),(p.flags&128)===0&&(p.memoizedState=null),p.flags|=4;Wr(p),f=!1}else O=El(),f!==null&&f.memoizedState!==null&&(f.memoizedState.hydrationErrors=O),f=!0;if(!f)return p.flags&256?(Zl(p),p):(Zl(p),null);if((p.flags&128)!==0)throw Error(r(558))}return Wr(p),null;case 13:if(D=p.memoizedState,f===null||f.memoizedState!==null&&f.memoizedState.dehydrated!==null){if(Q=za(p),D!==null&&D.dehydrated!==null){if(f===null){if(!Q)throw Error(r(318));if(Q=p.memoizedState,Q=Q!==null?Q.dehydrated:null,!Q)throw Error(r(317));Q[Dt]=p}else Se(),(p.flags&128)===0&&(p.memoizedState=null),p.flags|=4;Wr(p),Q=!1}else Q=El(),f!==null&&f.memoizedState!==null&&(f.memoizedState.hydrationErrors=Q),Q=!0;if(!Q)return p.flags&256?(Zl(p),p):(Zl(p),null)}return Zl(p),(p.flags&128)!==0?(p.lanes=O,p):(O=D!==null,f=f!==null&&f.memoizedState!==null,O&&(D=p.child,Q=null,D.alternate!==null&&D.alternate.memoizedState!==null&&D.alternate.memoizedState.cachePool!==null&&(Q=D.alternate.memoizedState.cachePool.pool),Y=null,D.memoizedState!==null&&D.memoizedState.cachePool!==null&&(Y=D.memoizedState.cachePool.pool),Y!==Q&&(D.flags|=2048)),O!==f&&O&&(p.child.flags|=8192),Mh(p,p.updateQueue),Wr(p),null);case 4:return ye(),f===null&&A4(p.stateNode.containerInfo),Wr(p),null;case 10:return Ul(p.type),Wr(p),null;case 19:if(X(ma),D=p.memoizedState,D===null)return Wr(p),null;if(Q=(p.flags&128)!==0,Y=D.rendering,Y===null)if(Q)ym(D,!1);else{if(ti!==0||f!==null&&(f.flags&128)!==0)for(f=p.child;f!==null;){if(Y=Tg(f),Y!==null){for(p.flags|=128,ym(D,!1),f=Y.updateQueue,p.updateQueue=f,Mh(p,f),p.subtreeFlags=0,f=O,O=p.child;O!==null;)yu(O,f),O=O.sibling;return B(ma,ma.current&1|2),Ze&&L(p,D.treeForkCount),p.child}f=f.sibling}D.tail!==null&&je()>hl&&(p.flags|=128,Q=!0,ym(D,!1),p.lanes=4194304)}else{if(!Q)if(f=Tg(Y),f!==null){if(p.flags|=128,Q=!0,f=f.updateQueue,p.updateQueue=f,Mh(p,f),ym(D,!0),D.tail===null&&D.tailMode==="hidden"&&!Y.alternate&&!Ze)return Wr(p),null}else 2*je()-D.renderingStartTime>hl&&O!==536870912&&(p.flags|=128,Q=!0,ym(D,!1),p.lanes=4194304);D.isBackwards?(Y.sibling=p.child,p.child=Y):(f=D.last,f!==null?f.sibling=Y:p.child=Y,D.last=Y)}return D.tail!==null?(f=D.tail,D.rendering=f,D.tail=f.sibling,D.renderingStartTime=je(),f.sibling=null,O=ma.current,B(ma,Q?O&1|2:O&1),Ze&&L(p,D.treeForkCount),f):(Wr(p),null);case 22:case 23:return Zl(p),xu(),D=p.memoizedState!==null,f!==null?f.memoizedState!==null!==D&&(p.flags|=8192):D&&(p.flags|=8192),D?(O&536870912)!==0&&(p.flags&128)===0&&(Wr(p),p.subtreeFlags&6&&(p.flags|=8192)):Wr(p),O=p.updateQueue,O!==null&&Mh(p,O.retryQueue),O=null,f!==null&&f.memoizedState!==null&&f.memoizedState.cachePool!==null&&(O=f.memoizedState.cachePool.pool),D=null,p.memoizedState!==null&&p.memoizedState.cachePool!==null&&(D=p.memoizedState.cachePool.pool),D!==O&&(p.flags|=2048),f!==null&&X(yd),null;case 24:return O=null,f!==null&&(O=f.memoizedState.cache),p.memoizedState.cache!==O&&(p.flags|=2048),Ul(gn),Wr(p),null;case 25:return null;case 30:return null}throw Error(r(156,p.tag))}function Md(f,p){switch(xe(p),p.tag){case 1:return f=p.flags,f&65536?(p.flags=f&-65537|128,p):null;case 3:return Ul(gn),ye(),f=p.flags,(f&65536)!==0&&(f&128)===0?(p.flags=f&-65537|128,p):null;case 26:case 27:case 5:return Ee(p),null;case 31:if(p.memoizedState!==null){if(Zl(p),p.alternate===null)throw Error(r(340));Se()}return f=p.flags,f&65536?(p.flags=f&-65537|128,p):null;case 13:if(Zl(p),f=p.memoizedState,f!==null&&f.dehydrated!==null){if(p.alternate===null)throw Error(r(340));Se()}return f=p.flags,f&65536?(p.flags=f&-65537|128,p):null;case 19:return X(ma),null;case 4:return ye(),null;case 10:return Ul(p.type),null;case 22:case 23:return Zl(p),xu(),f!==null&&X(yd),f=p.flags,f&65536?(p.flags=f&-65537|128,p):null;case 24:return Ul(gn),null;case 25:return null;default:return null}}function wf(f,p){switch(xe(p),p.tag){case 3:Ul(gn),ye();break;case 26:case 27:case 5:Ee(p);break;case 4:ye();break;case 31:p.memoizedState!==null&&Zl(p);break;case 13:Zl(p);break;case 19:X(ma);break;case 10:Ul(p.type);break;case 22:case 23:Zl(p),xu(),f!==null&&X(yd);break;case 24:Ul(gn)}}function Hs(f,p){try{var O=p.updateQueue,D=O!==null?O.lastEffect:null;if(D!==null){var Q=D.next;O=Q;do{if((O.tag&f)===f){D=void 0;var Y=O.create,ce=O.inst;D=Y(),ce.destroy=D}O=O.next}while(O!==Q)}}catch(_e){ki(p,p.return,_e)}}function $o(f,p,O){try{var D=p.updateQueue,Q=D!==null?D.lastEffect:null;if(Q!==null){var Y=Q.next;D=Y;do{if((D.tag&f)===f){var ce=D.inst,_e=ce.destroy;if(_e!==void 0){ce.destroy=void 0,Q=p;var Le=O,ut=_e;try{ut()}catch(pt){ki(Q,Le,pt)}}}D=D.next}while(D!==Y)}}catch(pt){ki(p,p.return,pt)}}function Xs(f){var p=f.updateQueue;if(p!==null){var O=f.stateNode;try{nu(p,O)}catch(D){ki(f,f.return,D)}}}function Kg(f,p,O){O.props=Hi(f.type,f.memoizedProps),O.state=f.memoizedState;try{O.componentWillUnmount()}catch(D){ki(f,p,D)}}function au(f,p){try{var O=f.ref;if(O!==null){switch(f.tag){case 26:case 27:case 5:var D=f.stateNode;break;case 30:D=f.stateNode;break;default:D=f.stateNode}typeof O=="function"?f.refCleanup=O(D):O.current=D}}catch(Q){ki(f,p,Q)}}function uu(f,p){var O=f.ref,D=f.refCleanup;if(O!==null)if(typeof D=="function")try{D()}catch(Q){ki(f,p,Q)}finally{f.refCleanup=null,f=f.alternate,f!=null&&(f.refCleanup=null)}else if(typeof O=="function")try{O(null)}catch(Q){ki(f,p,Q)}else O.current=null}function Co(f){var p=f.type,O=f.memoizedProps,D=f.stateNode;try{e:switch(p){case"button":case"input":case"select":case"textarea":O.autoFocus&&D.focus();break e;case"img":O.src?D.src=O.src:O.srcSet&&(D.srcset=O.srcSet)}}catch(Q){ki(f,f.return,Q)}}function wo(f,p,O){try{var D=f.stateNode;w5(D,f.type,O,p),D[Lt]=p}catch(Q){ki(f,f.return,Q)}}function qa(f){return f.tag===5||f.tag===3||f.tag===26||f.tag===27&&Vn(f.type)||f.tag===4}function zc(f){e:for(;;){for(;f.sibling===null;){if(f.return===null||qa(f.return))return null;f=f.return}for(f.sibling.return=f.return,f=f.sibling;f.tag!==5&&f.tag!==6&&f.tag!==18;){if(f.tag===27&&Vn(f.type)||f.flags&2||f.child===null||f.tag===4)continue e;f.child.return=f,f=f.child}if(!(f.flags&2))return f.stateNode}}function nr(f,p,O){var D=f.tag;if(D===5||D===6)f=f.stateNode,p?(O.nodeType===9?O.body:O.nodeName==="HTML"?O.ownerDocument.body:O).insertBefore(f,p):(p=O.nodeType===9?O.body:O.nodeName==="HTML"?O.ownerDocument.body:O,p.appendChild(f),O=O._reactRootContainer,O!=null||p.onclick!==null||(p.onclick=ho));else if(D!==4&&(D===27&&Vn(f.type)&&(O=f.stateNode,p=null),f=f.child,f!==null))for(nr(f,p,O),f=f.sibling;f!==null;)nr(f,p,O),f=f.sibling}function Jr(f,p,O){var D=f.tag;if(D===5||D===6)f=f.stateNode,p?O.insertBefore(f,p):O.appendChild(f);else if(D!==4&&(D===27&&Vn(f.type)&&(O=f.stateNode),f=f.child,f!==null))for(Jr(f,p,O),f=f.sibling;f!==null;)Jr(f,p,O),f=f.sibling}function Ta(f){var p=f.stateNode,O=f.memoizedProps;try{for(var D=f.type,Q=p.attributes;Q.length;)p.removeAttributeNode(Q[0]);Rl(p,D,O),p[Dt]=f,p[Lt]=O}catch(Y){ki(f,f.return,Y)}}var ss=!1,Xi=!1,Rf=!1,bm=typeof WeakSet=="function"?WeakSet:Set,va=null;function Up(f,p){if(f=f.containerInfo,ub=cO,f=nm(f),Sf(f)){if("selectionStart"in f)var O={start:f.selectionStart,end:f.selectionEnd};else e:{O=(O=f.ownerDocument)&&O.defaultView||window;var D=O.getSelection&&O.getSelection();if(D&&D.rangeCount!==0){O=D.anchorNode;var Q=D.anchorOffset,Y=D.focusNode;D=D.focusOffset;try{O.nodeType,Y.nodeType}catch{O=null;break e}var ce=0,_e=-1,Le=-1,ut=0,pt=0,xt=f,ot=null;t:for(;;){for(var ft;xt!==O||Q!==0&&xt.nodeType!==3||(_e=ce+Q),xt!==Y||D!==0&&xt.nodeType!==3||(Le=ce+D),xt.nodeType===3&&(ce+=xt.nodeValue.length),(ft=xt.firstChild)!==null;)ot=xt,xt=ft;for(;;){if(xt===f)break t;if(ot===O&&++ut===Q&&(_e=ce),ot===Y&&++pt===D&&(Le=ce),(ft=xt.nextSibling)!==null)break;xt=ot,ot=xt.parentNode}xt=ft}O=_e===-1||Le===-1?null:{start:_e,end:Le}}else O=null}O=O||{start:0,end:0}}else O=null;for(M4={focusedElem:f,selectionRange:O},cO=!1,va=p;va!==null;)if(p=va,f=p.child,(p.subtreeFlags&1028)!==0&&f!==null)f.return=p,va=f;else for(;va!==null;){switch(p=va,Y=p.alternate,f=p.flags,p.tag){case 0:if((f&4)!==0&&(f=p.updateQueue,f=f!==null?f.events:null,f!==null))for(O=0;O<f.length;O++)Q=f[O],Q.ref.impl=Q.nextImpl;break;case 11:case 15:break;case 1:if((f&1024)!==0&&Y!==null){f=void 0,O=p,Q=Y.memoizedProps,Y=Y.memoizedState,D=O.stateNode;try{var on=Hi(O.type,Q);f=D.getSnapshotBeforeUpdate(on,Y),D.__reactInternalSnapshotBeforeUpdate=f}catch(Rn){ki(O,O.return,Rn)}}break;case 3:if((f&1024)!==0){if(f=p.stateNode.containerInfo,O=f.nodeType,O===9)nO(f);else if(O===1)switch(f.nodeName){case"HEAD":case"HTML":case"BODY":nO(f);break;default:f.textContent=""}}break;case 5:case 26:case 27:case 6:case 4:case 17:break;default:if((f&1024)!==0)throw Error(r(163))}if(f=p.sibling,f!==null){f.return=p.return,va=f;break}va=p.return}}function Jy(f,p,O){var D=O.flags;switch(O.tag){case 0:case 11:case 15:Af(f,O),D&4&&Hs(5,O);break;case 1:if(Af(f,O),D&4)if(f=O.stateNode,p===null)try{f.componentDidMount()}catch(ce){ki(O,O.return,ce)}else{var Q=Hi(O.type,p.memoizedProps);p=p.memoizedState;try{f.componentDidUpdate(Q,p,f.__reactInternalSnapshotBeforeUpdate)}catch(ce){ki(O,O.return,ce)}}D&64&&Xs(O),D&512&&au(O,O.return);break;case 3:if(Af(f,O),D&64&&(f=O.updateQueue,f!==null)){if(p=null,O.child!==null)switch(O.child.tag){case 27:case 5:p=O.child.stateNode;break;case 1:p=O.child.stateNode}try{nu(f,p)}catch(ce){ki(O,O.return,ce)}}break;case 27:p===null&&D&4&&Ta(O);case 26:case 5:Af(f,O),p===null&&D&4&&Co(O),D&512&&au(O,O.return);break;case 12:Af(f,O);break;case 31:Af(f,O),D&4&&xm(f,O);break;case 13:Af(f,O),D&4&&eb(f,O),D&64&&(f=O.memoizedState,f!==null&&(f=f.dehydrated,f!==null&&(O=K1.bind(null,O),k5(f,O))));break;case 22:if(D=O.memoizedState!==null||ss,!D){p=p!==null&&p.memoizedState!==null||Xi,Q=ss;var Y=Xi;ss=D,(Xi=p)&&!Y?ei(f,O,(O.subtreeFlags&8772)!==0):Af(f,O),ss=Q,Xi=Y}break;case 30:break;default:Af(f,O)}}function Om(f){var p=f.alternate;p!==null&&(f.alternate=null,Om(p)),f.child=null,f.deletions=null,f.sibling=null,f.tag===5&&(p=f.stateNode,p!==null&&$i(p)),f.stateNode=null,f.return=null,f.dependencies=null,f.memoizedProps=null,f.memoizedState=null,f.pendingProps=null,f.stateNode=null,f.updateQueue=null}var Fr=null,Va=!1;function cl(f,p,O){for(O=O.child;O!==null;)Fd(f,p,O),O=O.sibling}function Fd(f,p,O){if(Xt&&typeof Xt.onCommitFiberUnmount=="function")try{Xt.onCommitFiberUnmount(qt,O)}catch{}switch(O.tag){case 26:Xi||uu(O,p),cl(f,p,O),O.memoizedState?O.memoizedState.count--:O.stateNode&&(O=O.stateNode,O.parentNode.removeChild(O));break;case 27:Xi||uu(O,p);var D=Fr,Q=Va;Vn(O.type)&&(Fr=O.stateNode,Va=!1),cl(f,p,O),sb(O.stateNode),Fr=D,Va=Q;break;case 5:Xi||uu(O,p);case 6:if(D=Fr,Q=Va,Fr=null,cl(f,p,O),Fr=D,Va=Q,Fr!==null)if(Va)try{(Fr.nodeType===9?Fr.body:Fr.nodeName==="HTML"?Fr.ownerDocument.body:Fr).removeChild(O.stateNode)}catch(Y){ki(O,p,Y)}else try{Fr.removeChild(O.stateNode)}catch(Y){ki(O,p,Y)}break;case 18:Fr!==null&&(Va?(f=Fr,N4(f.nodeType===9?f.body:f.nodeName==="HTML"?f.ownerDocument.body:f,O.stateNode),dv(f)):N4(Fr,O.stateNode));break;case 4:D=Fr,Q=Va,Fr=O.stateNode.containerInfo,Va=!0,cl(f,p,O),Fr=D,Va=Q;break;case 0:case 11:case 14:case 15:$o(2,O,p),Xi||$o(4,O,p),cl(f,p,O);break;case 1:Xi||(uu(O,p),D=O.stateNode,typeof D.componentWillUnmount=="function"&&Kg(O,p,D)),cl(f,p,O);break;case 21:cl(f,p,O);break;case 22:Xi=(D=Xi)||O.memoizedState!==null,cl(f,p,O),Xi=D;break;default:cl(f,p,O)}}function xm(f,p){if(p.memoizedState===null&&(f=p.alternate,f!==null&&(f=f.memoizedState,f!==null))){f=f.dehydrated;try{dv(f)}catch(O){ki(p,p.return,O)}}}function eb(f,p){if(p.memoizedState===null&&(f=p.alternate,f!==null&&(f=f.memoizedState,f!==null&&(f=f.dehydrated,f!==null))))try{dv(f)}catch(O){ki(p,p.return,O)}}function Y1(f){switch(f.tag){case 31:case 13:case 19:var p=f.stateNode;return p===null&&(p=f.stateNode=new bm),p;case 22:return f=f.stateNode,p=f._retryCache,p===null&&(p=f._retryCache=new bm),p;default:throw Error(r(435,f.tag))}}function ka(f,p){var O=Y1(f);p.forEach(function(D){if(!O.has(D)){O.add(D);var Q=E5.bind(null,f,D);D.then(Q,Q)}})}function fl(f,p){var O=p.deletions;if(O!==null)for(var D=0;D<O.length;D++){var Q=O[D],Y=f,ce=p,_e=ce;e:for(;_e!==null;){switch(_e.tag){case 27:if(Vn(_e.type)){Fr=_e.stateNode,Va=!1;break e}break;case 5:Fr=_e.stateNode,Va=!1;break e;case 3:case 4:Fr=_e.stateNode.containerInfo,Va=!0;break e}_e=_e.return}if(Fr===null)throw Error(r(160));Fd(Y,ce,Q),Fr=null,Va=!1,Y=Q.alternate,Y!==null&&(Y.return=null),Q.return=null}if(p.subtreeFlags&13886)for(p=p.child;p!==null;)Nd(p,f),p=p.sibling}var Wl=null;function Nd(f,p){var O=f.alternate,D=f.flags;switch(f.tag){case 0:case 11:case 14:case 15:fl(p,f),Ur(f),D&4&&($o(3,f,f.return),Hs(3,f),$o(5,f,f.return));break;case 1:fl(p,f),Ur(f),D&512&&(Xi||O===null||uu(O,O.return)),D&64&&ss&&(f=f.updateQueue,f!==null&&(D=f.callbacks,D!==null&&(O=f.shared.hiddenCallbacks,f.shared.hiddenCallbacks=O===null?D:O.concat(D))));break;case 26:var Q=Wl;if(fl(p,f),Ur(f),D&512&&(Xi||O===null||uu(O,O.return)),D&4){var Y=O!==null?O.memoizedState:null;if(D=f.memoizedState,O===null)if(D===null)if(f.stateNode===null){e:{D=f.type,O=f.memoizedProps,Q=Q.ownerDocument||Q;t:switch(D){case"title":Y=Q.getElementsByTagName("title")[0],(!Y||Y[Di]||Y[Dt]||Y.namespaceURI==="http://www.w3.org/2000/svg"||Y.hasAttribute("itemprop"))&&(Y=Q.createElement(D),Q.head.insertBefore(Y,Q.querySelector("head > title"))),Rl(Y,D,O),Y[Dt]=f,yi(Y),D=Y;break e;case"link":var ce=v_("link","href",Q).get(D+(O.href||""));if(ce){for(var _e=0;_e<ce.length;_e++)if(Y=ce[_e],Y.getAttribute("href")===(O.href==null||O.href===""?null:O.href)&&Y.getAttribute("rel")===(O.rel==null?null:O.rel)&&Y.getAttribute("title")===(O.title==null?null:O.title)&&Y.getAttribute("crossorigin")===(O.crossOrigin==null?null:O.crossOrigin)){ce.splice(_e,1);break t}}Y=Q.createElement(D),Rl(Y,D,O),Q.head.appendChild(Y);break;case"meta":if(ce=v_("meta","content",Q).get(D+(O.content||""))){for(_e=0;_e<ce.length;_e++)if(Y=ce[_e],Y.getAttribute("content")===(O.content==null?null:""+O.content)&&Y.getAttribute("name")===(O.name==null?null:O.name)&&Y.getAttribute("property")===(O.property==null?null:O.property)&&Y.getAttribute("http-equiv")===(O.httpEquiv==null?null:O.httpEquiv)&&Y.getAttribute("charset")===(O.charSet==null?null:O.charSet)){ce.splice(_e,1);break t}}Y=Q.createElement(D),Rl(Y,D,O),Q.head.appendChild(Y);break;default:throw Error(r(468,D))}Y[Dt]=f,yi(Y),D=Y}f.stateNode=D}else y_(Q,f.type,f.stateNode);else f.stateNode=g_(Q,D,f.memoizedProps);else Y!==D?(Y===null?O.stateNode!==null&&(O=O.stateNode,O.parentNode.removeChild(O)):Y.count--,D===null?y_(Q,f.type,f.stateNode):g_(Q,D,f.memoizedProps)):D===null&&f.stateNode!==null&&wo(f,f.memoizedProps,O.memoizedProps)}break;case 27:fl(p,f),Ur(f),D&512&&(Xi||O===null||uu(O,O.return)),O!==null&&D&4&&wo(f,f.memoizedProps,O.memoizedProps);break;case 5:if(fl(p,f),Ur(f),D&512&&(Xi||O===null||uu(O,O.return)),f.flags&32){Q=f.stateNode;try{Pu(Q,"")}catch(on){ki(f,f.return,on)}}D&4&&f.stateNode!=null&&(Q=f.memoizedProps,wo(f,Q,O!==null?O.memoizedProps:Q)),D&1024&&(Rf=!0);break;case 6:if(fl(p,f),Ur(f),D&4){if(f.stateNode===null)throw Error(r(162));D=f.memoizedProps,O=f.stateNode;try{O.nodeValue=D}catch(on){ki(f,f.return,on)}}break;case 3:if(cb=null,Q=Wl,Wl=iO(p.containerInfo),fl(p,f),Wl=Q,Ur(f),D&4&&O!==null&&O.memoizedState.isDehydrated)try{dv(p.containerInfo)}catch(on){ki(f,f.return,on)}Rf&&(Rf=!1,G1(f));break;case 4:D=Wl,Wl=iO(f.stateNode.containerInfo),fl(p,f),Ur(f),Wl=D;break;case 12:fl(p,f),Ur(f);break;case 31:fl(p,f),Ur(f),D&4&&(D=f.updateQueue,D!==null&&(f.updateQueue=null,ka(f,D)));break;case 13:fl(p,f),Ur(f),f.child.flags&8192&&f.memoizedState!==null!=(O!==null&&O.memoizedState!==null)&&(Vc=je()),D&4&&(D=f.updateQueue,D!==null&&(f.updateQueue=null,ka(f,D)));break;case 22:Q=f.memoizedState!==null;var Le=O!==null&&O.memoizedState!==null,ut=ss,pt=Xi;if(ss=ut||Q,Xi=pt||Le,fl(p,f),Xi=pt,ss=ut,Ur(f),D&8192)e:for(p=f.stateNode,p._visibility=Q?p._visibility&-2:p._visibility|1,Q&&(O===null||Le||ss||Xi||Fh(f)),O=null,p=f;;){if(p.tag===5||p.tag===26){if(O===null){Le=O=p;try{if(Y=Le.stateNode,Q)ce=Y.style,typeof ce.setProperty=="function"?ce.setProperty("display","none","important"):ce.display="none";else{_e=Le.stateNode;var xt=Le.memoizedProps.style,ot=xt!=null&&xt.hasOwnProperty("display")?xt.display:null;_e.style.display=ot==null||typeof ot=="boolean"?"":(""+ot).trim()}}catch(on){ki(Le,Le.return,on)}}}else if(p.tag===6){if(O===null){Le=p;try{Le.stateNode.nodeValue=Q?"":Le.memoizedProps}catch(on){ki(Le,Le.return,on)}}}else if(p.tag===18){if(O===null){Le=p;try{var ft=Le.stateNode;Q?ec(ft,!0):ec(Le.stateNode,!1)}catch(on){ki(Le,Le.return,on)}}}else if((p.tag!==22&&p.tag!==23||p.memoizedState===null||p===f)&&p.child!==null){p.child.return=p,p=p.child;continue}if(p===f)break e;for(;p.sibling===null;){if(p.return===null||p.return===f)break e;O===p&&(O=null),p=p.return}O===p&&(O=null),p.sibling.return=p.return,p=p.sibling}D&4&&(D=f.updateQueue,D!==null&&(O=D.retryQueue,O!==null&&(D.retryQueue=null,ka(f,O))));break;case 19:fl(p,f),Ur(f),D&4&&(D=f.updateQueue,D!==null&&(f.updateQueue=null,ka(f,D)));break;case 30:break;case 21:break;default:fl(p,f),Ur(f)}}function Ur(f){var p=f.flags;if(p&2){try{for(var O,D=f.return;D!==null;){if(qa(D)){O=D;break}D=D.return}if(O==null)throw Error(r(160));switch(O.tag){case 27:var Q=O.stateNode,Y=zc(f);Jr(f,Y,Q);break;case 5:var ce=O.stateNode;O.flags&32&&(Pu(ce,""),O.flags&=-33);var _e=zc(f);Jr(f,_e,ce);break;case 3:case 4:var Le=O.stateNode.containerInfo,ut=zc(f);nr(f,ut,Le);break;default:throw Error(r(161))}}catch(pt){ki(f,f.return,pt)}f.flags&=-3}p&4096&&(f.flags&=-4097)}function G1(f){if(f.subtreeFlags&1024)for(f=f.child;f!==null;){var p=f;G1(p),p.tag===5&&p.flags&1024&&p.stateNode.reset(),f=f.sibling}}function Af(f,p){if(p.subtreeFlags&8772)for(p=p.child;p!==null;)Jy(f,p.alternate,p),p=p.sibling}function Fh(f){for(f=f.child;f!==null;){var p=f;switch(p.tag){case 0:case 11:case 14:case 15:$o(4,p,p.return),Fh(p);break;case 1:uu(p,p.return);var O=p.stateNode;typeof O.componentWillUnmount=="function"&&Kg(p,p.return,O),Fh(p);break;case 27:sb(p.stateNode);case 26:case 5:uu(p,p.return),Fh(p);break;case 22:p.memoizedState===null&&Fh(p);break;case 30:Fh(p);break;default:Fh(p)}f=f.sibling}}function ei(f,p,O){for(O=O&&(p.subtreeFlags&8772)!==0,p=p.child;p!==null;){var D=p.alternate,Q=f,Y=p,ce=Y.flags;switch(Y.tag){case 0:case 11:case 15:ei(Q,Y,O),Hs(4,Y);break;case 1:if(ei(Q,Y,O),D=Y,Q=D.stateNode,typeof Q.componentDidMount=="function")try{Q.componentDidMount()}catch(ut){ki(D,D.return,ut)}if(D=Y,Q=D.updateQueue,Q!==null){var _e=D.stateNode;try{var Le=Q.shared.hiddenCallbacks;if(Le!==null)for(Q.shared.hiddenCallbacks=null,Q=0;Q<Le.length;Q++)tu(Le[Q],_e)}catch(ut){ki(D,D.return,ut)}}O&&ce&64&&Xs(Y),au(Y,Y.return);break;case 27:Ta(Y);case 26:case 5:ei(Q,Y,O),O&&D===null&&ce&4&&Co(Y),au(Y,Y.return);break;case 12:ei(Q,Y,O);break;case 31:ei(Q,Y,O),O&&ce&4&&xm(Q,Y);break;case 13:ei(Q,Y,O),O&&ce&4&&eb(Q,Y);break;case 22:Y.memoizedState===null&&ei(Q,Y,O),au(Y,Y.return);break;case 30:break;default:ei(Q,Y,O)}p=p.sibling}}function Wg(f,p){var O=null;f!==null&&f.memoizedState!==null&&f.memoizedState.cachePool!==null&&(O=f.memoizedState.cachePool.pool),f=null,p.memoizedState!==null&&p.memoizedState.cachePool!==null&&(f=p.memoizedState.cachePool.pool),f!==O&&(f!=null&&f.refCount++,O!=null&&Bu(O))}function Tf(f,p){f=null,p.alternate!==null&&(f=p.alternate.memoizedState.cache),p=p.memoizedState.cache,p!==f&&(p.refCount++,f!=null&&Bu(f))}function lu(f,p,O,D){if(p.subtreeFlags&10256)for(p=p.child;p!==null;)ua(f,p,O,D),p=p.sibling}function ua(f,p,O,D){var Q=p.flags;switch(p.tag){case 0:case 11:case 15:lu(f,p,O,D),Q&2048&&Hs(9,p);break;case 1:lu(f,p,O,D);break;case 3:lu(f,p,O,D),Q&2048&&(f=null,p.alternate!==null&&(f=p.alternate.memoizedState.cache),p=p.memoizedState.cache,p!==f&&(p.refCount++,f!=null&&Bu(f)));break;case 12:if(Q&2048){lu(f,p,O,D),f=p.stateNode;try{var Y=p.memoizedProps,ce=Y.id,_e=Y.onPostCommit;typeof _e=="function"&&_e(ce,p.alternate===null?"mount":"update",f.passiveEffectDuration,-0)}catch(Le){ki(p,p.return,Le)}}else lu(f,p,O,D);break;case 31:lu(f,p,O,D);break;case 13:lu(f,p,O,D);break;case 23:break;case 22:Y=p.stateNode,ce=p.alternate,p.memoizedState!==null?Y._visibility&2?lu(f,p,O,D):jp(f,p):Y._visibility&2?lu(f,p,O,D):(Y._visibility|=2,qc(f,p,O,D,(p.subtreeFlags&10256)!==0||!1)),Q&2048&&Wg(ce,p);break;case 24:lu(f,p,O,D),Q&2048&&Tf(p.alternate,p);break;default:lu(f,p,O,D)}}function qc(f,p,O,D,Q){for(Q=Q&&((p.subtreeFlags&10256)!==0||!1),p=p.child;p!==null;){var Y=f,ce=p,_e=O,Le=D,ut=ce.flags;switch(ce.tag){case 0:case 11:case 15:qc(Y,ce,_e,Le,Q),Hs(8,ce);break;case 23:break;case 22:var pt=ce.stateNode;ce.memoizedState!==null?pt._visibility&2?qc(Y,ce,_e,Le,Q):jp(Y,ce):(pt._visibility|=2,qc(Y,ce,_e,Le,Q)),Q&&ut&2048&&Wg(ce.alternate,ce);break;case 24:qc(Y,ce,_e,Le,Q),Q&&ut&2048&&Tf(ce.alternate,ce);break;default:qc(Y,ce,_e,Le,Q)}p=p.sibling}}function jp(f,p){if(p.subtreeFlags&10256)for(p=p.child;p!==null;){var O=f,D=p,Q=D.flags;switch(D.tag){case 22:jp(O,D),Q&2048&&Wg(D.alternate,D);break;case 24:jp(O,D),Q&2048&&Tf(D.alternate,D);break;default:jp(O,D)}p=p.sibling}}var ya=8192;function Nh(f,p,O){if(f.subtreeFlags&ya)for(f=f.child;f!==null;)Pa(f,p,O),f=f.sibling}function Pa(f,p,O){switch(f.tag){case 26:Nh(f,p,O),f.flags&ya&&f.memoizedState!==null&&sv(O,Wl,f.memoizedState,f.memoizedProps);break;case 5:Nh(f,p,O);break;case 3:case 4:var D=Wl;Wl=iO(f.stateNode.containerInfo),Nh(f,p,O),Wl=D;break;case 22:f.memoizedState===null&&(D=f.alternate,D!==null&&D.memoizedState!==null?(D=ya,ya=16777216,Nh(f,p,O),ya=D):Nh(f,p,O));break;default:Nh(f,p,O)}}function Jl(f){var p=f.alternate;if(p!==null&&(f=p.child,f!==null)){p.child=null;do p=f.sibling,f.sibling=null,f=p;while(f!==null)}}function Cl(f){var p=f.deletions;if((f.flags&16)!==0){if(p!==null)for(var O=0;O<p.length;O++){var D=p[O];va=D,Sm(D,f)}Jl(f)}if(f.subtreeFlags&10256)for(f=f.child;f!==null;)Ih(f),f=f.sibling}function Ih(f){switch(f.tag){case 0:case 11:case 15:Cl(f),f.flags&2048&&$o(9,f,f.return);break;case 3:Cl(f);break;case 12:Cl(f);break;case 22:var p=f.stateNode;f.memoizedState!==null&&p._visibility&2&&(f.return===null||f.return.tag!==13)?(p._visibility&=-3,Jg(f)):Cl(f);break;default:Cl(f)}}function Jg(f){var p=f.deletions;if((f.flags&16)!==0){if(p!==null)for(var O=0;O<p.length;O++){var D=p[O];va=D,Sm(D,f)}Jl(f)}for(f=f.child;f!==null;){switch(p=f,p.tag){case 0:case 11:case 15:$o(8,p,p.return),Jg(p);break;case 22:O=p.stateNode,O._visibility&2&&(O._visibility&=-3,Jg(p));break;default:Jg(p)}f=f.sibling}}function Sm(f,p){for(;va!==null;){var O=va;switch(O.tag){case 0:case 11:case 15:$o(8,O,p);break;case 23:case 22:if(O.memoizedState!==null&&O.memoizedState.cachePool!==null){var D=O.memoizedState.cachePool.pool;D!=null&&D.refCount++}break;case 24:Bu(O.memoizedState.cache)}if(D=O.child,D!==null)D.return=O,va=D;else e:for(O=f;va!==null;){D=va;var Q=D.sibling,Y=D.return;if(Om(D),D===O){va=null;break e}if(Q!==null){Q.return=Y,va=Q;break e}va=Y}}}var ev={getCacheForType:function(f){var p=Ou(gn),O=p.data.get(f);return O===void 0&&(O=f(),p.data.set(f,O)),O},cacheSignal:function(){return Ou(gn).controller.signal}},tv=typeof WeakMap=="function"?WeakMap:Map,Nr=0,li=null,rr=null,_r=0,vn=0,mr=null,Yn=!1,cr=!1,Si=!1,la=0,ti=0,Ro=0,kf=0,Ys=0,Ir=0,Ao=0,Gs=null,dl=null,eo=!1,Vc=0,Zp=0,hl=1/0,Hp=null,Uc=null,oa=0,Ks=null,To=null,cs=0,Pf=0,Lh=null,nv=null,Bh=0,Qh=null;function wl(){return(Nr&2)!==0&&_r!==0?_r&-_r:K.T!==null?Ws():Ct()}function g(){if(Ir===0)if((_r&536870912)===0||Ze){var f=In;In<<=1,(In&3932160)===0&&(In=262144),Ir=f}else Ir=536870912;return f=ru.current,f!==null&&(f.flags|=32),Ir}function b(f,p,O){(f===li&&(vn===2||vn===9)||f.cancelPendingCommit!==null)&&(We(f,0),le(f,_r,Ir,!1)),ji(f,O),((Nr&2)===0||f!==li)&&(f===li&&((Nr&2)===0&&(kf|=O),ti===4&&le(f,_r,Ir,!1)),Id(f))}function T(f,p,O){if((Nr&6)!==0)throw Error(r(327));var D=!O&&(p&127)===0&&(p&f.expiredLanes)===0||Or(f,p),Q=D?zh(f,p):to(f,p,!0),Y=D;do{if(Q===0){cr&&!D&&le(f,p,0,!1);break}else{if(O=f.current.alternate,Y&&!ee(O)){Q=to(f,p,!1),Y=!1;continue}if(Q===2){if(Y=p,f.errorRecoveryDisabledLanes&Y)var ce=0;else ce=f.pendingLanes&-536870913,ce=ce!==0?ce:ce&536870912?536870912:0;if(ce!==0){p=ce;e:{var _e=f;Q=Gs;var Le=_e.current.memoizedState.isDehydrated;if(Le&&(We(_e,ce).flags|=256),ce=to(_e,ce,!1),ce!==2){if(Si&&!Le){_e.errorRecoveryDisabledLanes|=Y,kf|=Y,Q=4;break e}Y=dl,dl=Q,Y!==null&&(dl===null?dl=Y:dl.push.apply(dl,Y))}Q=ce}if(Y=!1,Q!==2)continue}}if(Q===1){We(f,0),le(f,p,0,!0);break}e:{switch(D=f,Y=Q,Y){case 0:case 1:throw Error(r(345));case 4:if((p&4194048)!==p)break;case 6:le(D,p,Ir,!Yn);break e;case 2:dl=null;break;case 3:case 5:break;default:throw Error(r(329))}if((p&62914560)===p&&(Q=Vc+0-je(),10<Q)){if(le(D,p,Ir,!Yn),br(D,0,!0)!==0)break e;cs=p,D.timeoutHandle=a_(z.bind(null,D,O,dl,Hp,eo,p,Ir,kf,Ao,Yn,Y,"Throttled",-0,0),Q);break e}z(D,O,dl,Hp,eo,p,Ir,kf,Ao,Yn,Y,null,-0,0)}}break}while(!0);Id(f)}function z(f,p,O,D,Q,Y,ce,_e,Le,ut,pt,xt,ot,ft){if(f.timeoutHandle=-1,xt=p.subtreeFlags,xt&8192||(xt&16785408)===16785408){xt={stylesheets:null,count:0,imgCount:0,imgBytes:0,suspenseyImages:[],waitingForImages:!0,waitingForViewTransition:!1,unsuspend:ho},Pa(p,Y,xt);var on=(Y&62914560)===Y?Vc-je():(Y&4194048)===Y?Zp-je():0;if(on=z5(xt,on),on!==null){cs=Y,f.cancelPendingCommit=on(Tn.bind(null,f,p,Y,O,D,Q,ce,_e,Le,pt,xt,null,ot,ft)),le(f,Y,ce,!ut);return}}Tn(f,p,Y,O,D,Q,ce,_e,Le)}function ee(f){for(var p=f;;){var O=p.tag;if((O===0||O===11||O===15)&&p.flags&16384&&(O=p.updateQueue,O!==null&&(O=O.stores,O!==null)))for(var D=0;D<O.length;D++){var Q=O[D],Y=Q.getSnapshot;Q=Q.value;try{if(!pr(Y(),Q))return!1}catch{return!1}}if(O=p.child,p.subtreeFlags&16384&&O!==null)O.return=p,p=O;else{if(p===f)break;for(;p.sibling===null;){if(p.return===null||p.return===f)return!0;p=p.return}p.sibling.return=p.return,p=p.sibling}}return!0}function le(f,p,O,D){p&=~Ys,p&=~kf,f.suspendedLanes|=p,f.pingedLanes&=~p,D&&(f.warmLanes|=p),D=f.expirationTimes;for(var Q=p;0<Q;){var Y=31-Fe(Q),ce=1<<Y;D[Y]=-1,Q&=~ce}O!==0&&ar(f,O,p)}function De(){return(Nr&6)===0?(nb(0),!1):!0}function Ie(){if(rr!==null){if(vn===0)var f=rr.return;else f=rr,Ja=qn=null,Py(f),Od=null,nl=0,f=rr;for(;f!==null;)wf(f.alternate,f),f=f.return;rr=null}}function We(f,p){var O=f.timeoutHandle;O!==-1&&(f.timeoutHandle=-1,A5(O)),O=f.cancelPendingCommit,O!==null&&(f.cancelPendingCommit=null,O()),cs=0,Ie(),li=f,rr=O=tl(f.current,null),_r=p,vn=0,mr=null,Yn=!1,cr=Or(f,p),Si=!1,Ao=Ir=Ys=kf=Ro=ti=0,dl=Gs=null,eo=!1,(p&8)!==0&&(p|=p&32);var D=f.entangledLanes;if(D!==0)for(f=f.entanglements,D&=p;0<D;){var Q=31-Fe(D),Y=1<<Q;p|=f[Q],D&=~Y}return la=p,Wa(),O}function vt(f,p){An=null,K.H=Rd,p===bd||p===es?(p=Cg(),vn=3):p===Fp?(p=Cg(),vn=4):vn=p===qp?8:p!==null&&typeof p=="object"&&typeof p.then=="function"?6:1,mr=p,rr===null&&(ti=1,pm(f,Mu(p,f.current)))}function Pt(){var f=ru.current;return f===null?!0:(_r&4194048)===_r?rs===null:(_r&62914560)===_r||(_r&536870912)!==0?f===rs:!1}function jt(){var f=K.H;return K.H=Rd,f===null?Rd:f}function Wt(){var f=K.A;return K.A=ev,f}function Hn(){ti=4,Yn||(_r&4194048)!==_r&&ru.current!==null||(cr=!0),(Ro&134217727)===0&&(kf&134217727)===0||li===null||le(li,_r,Ir,!1)}function to(f,p,O){var D=Nr;Nr|=2;var Q=jt(),Y=Wt();(li!==f||_r!==p)&&(Hp=null,We(f,p)),p=!1;var ce=ti;e:do try{if(vn!==0&&rr!==null){var _e=rr,Le=mr;switch(vn){case 8:Ie(),ce=6;break e;case 3:case 2:case 9:case 6:ru.current===null&&(p=!0);var ut=vn;if(vn=0,mr=null,Tt(f,_e,Le,ut),O&&cr){ce=0;break e}break;default:ut=vn,vn=0,mr=null,Tt(f,_e,Le,ut)}}_m(),ce=ti;break}catch(pt){vt(f,pt)}while(!0);return p&&f.shellSuspendCounter++,Ja=qn=null,Nr=D,K.H=Q,K.A=Y,rr===null&&(li=null,_r=0,Wa()),ce}function _m(){for(;rr!==null;)Ge(rr)}function zh(f,p){var O=Nr;Nr|=2;var D=jt(),Q=Wt();li!==f||_r!==p?(Hp=null,hl=je()+500,We(f,p)):cr=Or(f,p);e:do try{if(vn!==0&&rr!==null){p=rr;var Y=mr;t:switch(vn){case 1:vn=0,mr=null,Tt(f,p,Y,1);break;case 2:case 9:if(B1(Y)){vn=0,mr=null,ct(p);break}p=function(){vn!==2&&vn!==9||li!==f||(vn=7),Id(f)},Y.then(p,p);break e;case 3:vn=7;break e;case 4:vn=5;break e;case 7:B1(Y)?(vn=0,mr=null,ct(p)):(vn=0,mr=null,Tt(f,p,Y,7));break;case 5:var ce=null;switch(rr.tag){case 26:ce=rr.memoizedState;case 5:case 27:var _e=rr;if(ce?b_(ce):_e.stateNode.complete){vn=0,mr=null;var Le=_e.sibling;if(Le!==null)rr=Le;else{var ut=_e.return;ut!==null?(rr=ut,pn(ut)):rr=null}break t}}vn=0,mr=null,Tt(f,p,Y,5);break;case 6:vn=0,mr=null,Tt(f,p,Y,6);break;case 8:Ie(),ti=6;break e;default:throw Error(r(462))}}lt();break}catch(pt){vt(f,pt)}while(!0);return Ja=qn=null,K.H=D,K.A=Q,Nr=O,rr!==null?0:(li=null,_r=0,Wa(),ti)}function lt(){for(;rr!==null&&!Ae();)Ge(rr)}function Ge(f){var p=Yg(f.alternate,f,la);f.memoizedProps=f.pendingProps,p===null?pn(f):rr=p}function ct(f){var p=f,O=p.alternate;switch(p.tag){case 15:case 0:p=sl(O,p,p.pendingProps,p.type,void 0,_r);break;case 11:p=sl(O,p,p.pendingProps,p.type.render,p.ref,_r);break;case 5:Py(p);default:wf(O,p),p=rr=yu(p,la),p=Yg(O,p,la)}f.memoizedProps=f.pendingProps,p===null?pn(f):rr=p}function Tt(f,p,O,D){Ja=qn=null,Py(p),Od=null,nl=0;var Q=p.return;try{if(Dl(f,Q,p,O,_r)){ti=1,pm(f,Mu(O,f.current)),rr=null;return}}catch(Y){if(Q!==null)throw rr=Q,Y;ti=1,pm(f,Mu(O,f.current)),rr=null;return}p.flags&32768?(Ze||D===1?f=!0:cr||(_r&536870912)!==0?f=!1:(Yn=f=!0,(D===2||D===9||D===3||D===6)&&(D=ru.current,D!==null&&D.tag===13&&(D.flags|=16384))),pi(p,f)):pn(p)}function pn(f){var p=f;do{if((p.flags&32768)!==0){pi(p,Yn);return}f=p.return;var O=Gg(p.alternate,p,la);if(O!==null){rr=O;return}if(p=p.sibling,p!==null){rr=p;return}rr=p=f}while(p!==null);ti===0&&(ti=5)}function pi(f,p){do{var O=Md(f.alternate,f);if(O!==null){O.flags&=32767,rr=O;return}if(O=f.return,O!==null&&(O.flags|=32768,O.subtreeFlags=0,O.deletions=null),!p&&(f=f.sibling,f!==null)){rr=f;return}rr=f=O}while(f!==null);ti=6,rr=null}function Tn(f,p,O,D,Q,Y,ce,_e,Le){f.cancelPendingCommit=null;do Em();while(oa!==0);if((Nr&6)!==0)throw Error(r(327));if(p!==null){if(p===f.current)throw Error(r(177));if(Y=p.lanes|p.childLanes,Y|=Is,ai(f,O,Y,ce,_e,Le),f===li&&(rr=li=null,_r=0),To=p,Ks=f,cs=O,Pf=Y,Lh=Q,nv=D,(p.subtreeFlags&10256)!==0||(p.flags&10256)!==0?(f.callbackNode=null,f.callbackPriority=0,D5(at,function(){return _4(),null})):(f.callbackNode=null,f.callbackPriority=0),D=(p.flags&13878)!==0,(p.subtreeFlags&13878)!==0||D){D=K.T,K.T=null,Q=J.p,J.p=2,ce=Nr,Nr|=4;try{Up(f,p,O)}finally{Nr=ce,J.p=Q,K.T=D}}oa=1,Lr(),Ua(),Tr()}}function Lr(){if(oa===1){oa=0;var f=Ks,p=To,O=(p.flags&13878)!==0;if((p.subtreeFlags&13878)!==0||O){O=K.T,K.T=null;var D=J.p;J.p=2;var Q=Nr;Nr|=4;try{Nd(p,f);var Y=M4,ce=nm(f.containerInfo),_e=Y.focusedElem,Le=Y.selectionRange;if(ce!==_e&&_e&&_e.ownerDocument&&xr(_e.ownerDocument.documentElement,_e)){if(Le!==null&&Sf(_e)){var ut=Le.start,pt=Le.end;if(pt===void 0&&(pt=ut),"selectionStart"in _e)_e.selectionStart=ut,_e.selectionEnd=Math.min(pt,_e.value.length);else{var xt=_e.ownerDocument||document,ot=xt&&xt.defaultView||window;if(ot.getSelection){var ft=ot.getSelection(),on=_e.textContent.length,Rn=Math.min(Le.start,on),zi=Le.end===void 0?Rn:Math.min(Le.end,on);!ft.extend&&Rn>zi&&(ce=zi,zi=Rn,Rn=ce);var Ke=tm(_e,Rn),Ue=tm(_e,zi);if(Ke&&Ue&&(ft.rangeCount!==1||ft.anchorNode!==Ke.node||ft.anchorOffset!==Ke.offset||ft.focusNode!==Ue.node||ft.focusOffset!==Ue.offset)){var it=xt.createRange();it.setStart(Ke.node,Ke.offset),ft.removeAllRanges(),Rn>zi?(ft.addRange(it),ft.extend(Ue.node,Ue.offset)):(it.setEnd(Ue.node,Ue.offset),ft.addRange(it))}}}}for(xt=[],ft=_e;ft=ft.parentNode;)ft.nodeType===1&&xt.push({element:ft,left:ft.scrollLeft,top:ft.scrollTop});for(typeof _e.focus=="function"&&_e.focus(),_e=0;_e<xt.length;_e++){var yt=xt[_e];yt.element.scrollLeft=yt.left,yt.element.scrollTop=yt.top}}cO=!!ub,M4=ub=null}finally{Nr=Q,J.p=D,K.T=O}}f.current=p,oa=2}}function Ua(){if(oa===2){oa=0;var f=Ks,p=To,O=(p.flags&8772)!==0;if((p.subtreeFlags&8772)!==0||O){O=K.T,K.T=null;var D=J.p;J.p=2;var Q=Nr;Nr|=4;try{Jy(f,p.alternate,p)}finally{Nr=Q,J.p=D,K.T=O}}oa=3}}function Tr(){if(oa===4||oa===3){oa=0,He();var f=Ks,p=To,O=cs,D=nv;(p.subtreeFlags&10256)!==0||(p.flags&10256)!==0?oa=5:(oa=0,To=Ks=null,jc(f,f.pendingLanes));var Q=f.pendingLanes;if(Q===0&&(Uc=null),gt(O),p=p.stateNode,Xt&&typeof Xt.onCommitFiberRoot=="function")try{Xt.onCommitFiberRoot(qt,p,void 0,(p.current.flags&128)===128)}catch{}if(D!==null){p=K.T,Q=J.p,J.p=2,K.T=null;try{for(var Y=f.onRecoverableError,ce=0;ce<D.length;ce++){var _e=D[ce];Y(_e.value,{componentStack:_e.stack})}}finally{K.T=p,J.p=Q}}(cs&3)!==0&&Em(),Id(f),Q=f.pendingLanes,(O&261930)!==0&&(Q&42)!==0?f===Qh?Bh++:(Bh=0,Qh=f):Bh=0,nb(0)}}function jc(f,p){(f.pooledCacheLanes&=p)===0&&(p=f.pooledCache,p!=null&&(f.pooledCache=null,Bu(p)))}function Em(){return Lr(),Ua(),Tr(),_4()}function _4(){if(oa!==5)return!1;var f=Ks,p=Pf;Pf=0;var O=gt(cs),D=K.T,Q=J.p;try{J.p=32>O?32:O,K.T=null,O=Lh,Lh=null;var Y=Ks,ce=cs;if(oa=0,To=Ks=null,cs=0,(Nr&6)!==0)throw Error(r(331));var _e=Nr;if(Nr|=4,Ih(Y.current),ua(Y,Y.current,ce,O),Nr=_e,nb(0,!1),Xt&&typeof Xt.onPostCommitFiberRoot=="function")try{Xt.onPostCommitFiberRoot(qt,Y)}catch{}return!0}finally{J.p=Q,K.T=D,jc(f,p)}}function E4(f,p,O){p=Mu(O,p),p=Wy(f.stateNode,p,2),f=ns(f,p,2),f!==null&&(ji(f,2),Id(f))}function ki(f,p,O){if(f.tag===3)E4(f,f,O);else for(;p!==null;){if(p.tag===3){E4(p,f,O);break}else if(p.tag===1){var D=p.stateNode;if(typeof p.type.getDerivedStateFromError=="function"||typeof D.componentDidCatch=="function"&&(Uc===null||!Uc.has(D))){f=Mu(O,f),O=Gl(2),D=ns(p,O,2),D!==null&&(ol(O,D,p,f),ji(D,2),Id(D));break}}p=p.return}}function D4(f,p,O){var D=f.pingCache;if(D===null){D=f.pingCache=new tv;var Q=new Set;D.set(p,Q)}else Q=D.get(p),Q===void 0&&(Q=new Set,D.set(p,Q));Q.has(O)||(Si=!0,Q.add(O),f=_5.bind(null,f,p,O),p.then(f,f))}function _5(f,p,O){var D=f.pingCache;D!==null&&D.delete(p),f.pingedLanes|=f.suspendedLanes&O,f.warmLanes&=~O,li===f&&(_r&O)===O&&(ti===4||ti===3&&(_r&62914560)===_r&&0>je()-Vc?(Nr&2)===0&&We(f,0):Ys|=O,Ao===_r&&(Ao=0)),Id(f)}function tb(f,p){p===0&&(p=wr()),f=el(f,p),f!==null&&(ji(f,p),Id(f))}function K1(f){var p=f.memoizedState,O=0;p!==null&&(O=p.retryLane),tb(f,O)}function E5(f,p){var O=0;switch(f.tag){case 31:case 13:var D=f.stateNode,Q=f.memoizedState;Q!==null&&(O=Q.retryLane);break;case 19:D=f.stateNode;break;case 22:D=f.stateNode._retryCache;break;default:throw Error(r(314))}D!==null&&D.delete(p),tb(f,O)}function D5(f,p){return Oe(f,p)}var rv=null,Dm=null,$4=!1,W1=!1,C4=!1,Xp=0;function Id(f){f!==Dm&&f.next===null&&(Dm===null?rv=Dm=f:Dm=Dm.next=f),W1=!0,$4||($4=!0,C5())}function nb(f,p){if(!C4&&W1){C4=!0;do for(var O=!1,D=rv;D!==null;){if(f!==0){var Q=D.pendingLanes;if(Q===0)var Y=0;else{var ce=D.suspendedLanes,_e=D.pingedLanes;Y=(1<<31-Fe(42|f)+1)-1,Y&=Q&~(ce&~_e),Y=Y&201326741?Y&201326741|1:Y?Y|2:0}Y!==0&&(O=!0,WS(D,Y))}else Y=_r,Y=br(D,D===li?Y:0,D.cancelPendingCommit!==null||D.timeoutHandle!==-1),(Y&3)===0||Or(D,Y)||(O=!0,WS(D,Y));D=D.next}while(O);C4=!1}}function $5(){YS()}function YS(){W1=$4=!1;var f=0;Xp!==0&&R5()&&(f=Xp);for(var p=je(),O=null,D=rv;D!==null;){var Q=D.next,Y=GS(D,p);Y===0?(D.next=null,O===null?rv=Q:O.next=Q,Q===null&&(Dm=O)):(O=D,(f!==0||(Y&3)!==0)&&(W1=!0)),D=Q}oa!==0&&oa!==5||nb(f),Xp!==0&&(Xp=0)}function GS(f,p){for(var O=f.suspendedLanes,D=f.pingedLanes,Q=f.expirationTimes,Y=f.pendingLanes&-62914561;0<Y;){var ce=31-Fe(Y),_e=1<<ce,Le=Q[ce];Le===-1?((_e&O)===0||(_e&D)!==0)&&(Q[ce]=Cr(_e,p)):Le<=p&&(f.expiredLanes|=_e),Y&=~_e}if(p=li,O=_r,O=br(f,f===p?O:0,f.cancelPendingCommit!==null||f.timeoutHandle!==-1),D=f.callbackNode,O===0||f===p&&(vn===2||vn===9)||f.cancelPendingCommit!==null)return D!==null&&D!==null&&$e(D),f.callbackNode=null,f.callbackPriority=0;if((O&3)===0||Or(f,O)){if(p=O&-O,p===f.callbackPriority)return p;switch(D!==null&&$e(D),gt(O)){case 2:case 8:O=ze;break;case 32:O=at;break;case 268435456:O=Mt;break;default:O=at}return D=KS.bind(null,f),O=Oe(O,D),f.callbackPriority=p,f.callbackNode=O,p}return D!==null&&D!==null&&$e(D),f.callbackPriority=2,f.callbackNode=null,2}function KS(f,p){if(oa!==0&&oa!==5)return f.callbackNode=null,f.callbackPriority=0,null;var O=f.callbackNode;if(Em()&&f.callbackNode!==O)return null;var D=_r;return D=br(f,f===li?D:0,f.cancelPendingCommit!==null||f.timeoutHandle!==-1),D===0?null:(T(f,D,p),GS(f,je()),f.callbackNode!=null&&f.callbackNode===O?KS.bind(null,f):null)}function WS(f,p){if(Em())return null;T(f,p,!0)}function C5(){l_(function(){(Nr&6)!==0?Oe(Ne,$5):YS()})}function Ws(){if(Xp===0){var f=Df;f===0&&(f=en,en<<=1,(en&261888)===0&&(en=256)),Xp=f}return Xp}function JS(f){return f==null||typeof f=="symbol"||typeof f=="boolean"?null:typeof f=="function"?f:yf(""+f)}function e_(f,p){var O=p.ownerDocument.createElement("input");return O.name=p.name,O.value=p.value,f.id&&O.setAttribute("form",f.id),p.parentNode.insertBefore(O,p),f=new FormData(f),O.parentNode.removeChild(O),f}function $m(f,p,O,D,Q){if(p==="submit"&&O&&O.stateNode===Q){var Y=JS((Q[Lt]||null).action),ce=D.submitter;ce&&(p=(p=ce[Lt]||null)?JS(p.formAction):ce.getAttribute("formAction"),p!==null&&(Y=p,ce=null));var _e=new ad("action","action",null,D,Q);f.push({event:_e,listeners:[{instance:null,listener:function(){if(D.defaultPrevented){if(Xp!==0){var Le=ce?e_(Q,ce):new FormData(Q);$d(O,{pending:!0,data:Le,method:Q.method,action:Y},null,Le)}}else typeof Y=="function"&&(_e.preventDefault(),Le=ce?e_(Q,ce):new FormData(Q),$d(O,{pending:!0,data:Le,method:Q.method,action:Y},Y,Le))},currentTarget:Q}]})}}for(var Wn=0;Wn<bh.length;Wn++){var w4=bh[Wn],fs=w4.toLowerCase(),ou=w4[0].toUpperCase()+w4.slice(1);pa(fs,"on"+ou)}pa(Ci,"onAnimationEnd"),pa(md,"onAnimationIteration"),pa(Ql,"onAnimationStart"),pa("dblclick","onDoubleClick"),pa("focusin","onFocus"),pa("focusout","onBlur"),pa(vo,"onTransitionRun"),pa(yo,"onTransitionStart"),pa(_f,"onTransitionCancel"),pa(Sr,"onTransitionEnd"),Tu("onMouseEnter",["mouseout","mouseover"]),Tu("onMouseLeave",["mouseout","mouseover"]),Tu("onPointerEnter",["pointerout","pointerover"]),Tu("onPointerLeave",["pointerout","pointerover"]),Au("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),Au("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),Au("onBeforeInput",["compositionend","keypress","textInput","paste"]),Au("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),Au("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),Au("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var rb="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),Yp=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(rb));function t_(f,p){p=(p&4)!==0;for(var O=0;O<f.length;O++){var D=f[O],Q=D.event;D=D.listeners;e:{var Y=void 0;if(p)for(var ce=D.length-1;0<=ce;ce--){var _e=D[ce],Le=_e.instance,ut=_e.currentTarget;if(_e=_e.listener,Le!==Y&&Q.isPropagationStopped())break e;Y=_e,Q.currentTarget=ut;try{Y(Q)}catch(pt){Go(pt)}Q.currentTarget=null,Y=Le}else for(ce=0;ce<D.length;ce++){if(_e=D[ce],Le=_e.instance,ut=_e.currentTarget,_e=_e.listener,Le!==Y&&Q.isPropagationStopped())break e;Y=_e,Q.currentTarget=ut;try{Y(Q)}catch(pt){Go(pt)}Q.currentTarget=null,Y=Le}}}}function kr(f,p){var O=p[_n];O===void 0&&(O=p[_n]=new Set);var D=f+"__bubble";O.has(D)||(eO(p,f,2,!1),O.add(D))}function R4(f,p,O){var D=0;p&&(D|=4),eO(O,f,D,p)}var J1="_reactListening"+Math.random().toString(36).slice(2);function A4(f){if(!f[J1]){f[J1]=!0,Ca.forEach(function(O){O!=="selectionchange"&&(Yp.has(O)||R4(O,!1,f),R4(O,!0,f))});var p=f.nodeType===9?f:f.ownerDocument;p===null||p[J1]||(p[J1]=!0,R4("selectionchange",!1,p))}}function eO(f,p,O,D){switch(C_(p)){case 2:var Q=V5;break;case 8:Q=U5;break;default:Q=q4}O=Q.bind(null,p,O,f),Q=void 0,!_p||p!=="touchstart"&&p!=="touchmove"&&p!=="wheel"||(Q=!0),D?Q!==void 0?f.addEventListener(p,O,{capture:!0,passive:Q}):f.addEventListener(p,O,!0):Q!==void 0?f.addEventListener(p,O,{passive:Q}):f.addEventListener(p,O,!1)}function T4(f,p,O,D,Q){var Y=D;if((p&1)===0&&(p&2)===0&&D!==null)e:for(;;){if(D===null)return;var ce=D.tag;if(ce===3||ce===4){var _e=D.stateNode.containerInfo;if(_e===Q)break;if(ce===4)for(ce=D.return;ce!==null;){var Le=ce.tag;if((Le===3||Le===4)&&ce.stateNode.containerInfo===Q)return;ce=ce.return}for(;_e!==null;){if(ce=hr(_e),ce===null)return;if(Le=ce.tag,Le===5||Le===6||Le===26||Le===27){D=Y=ce;continue e}_e=_e.parentNode}}D=D.return}Sp(function(){var ut=Y,pt=fh(O),xt=[];e:{var ot=Yo.get(f);if(ot!==void 0){var ft=ad,on=f;switch(f){case"keypress":if(xc(O)===0)break e;case"keydown":case"keyup":ft=Ho;break;case"focusin":on="focus",ft=od;break;case"focusout":on="blur",ft=od;break;case"beforeblur":case"afterblur":ft=od;break;case"click":if(O.button===2)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":ft=Ep;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":ft=X0;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":ft=Ms;break;case Ci:case md:case Ql:ft=_c;break;case Sr:ft=Dy;break;case"scroll":case"scrollend":ft=ud;break;case"wheel":ft=po;break;case"copy":case"cut":case"paste":ft=ha;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":ft=Cp;break;case"toggle":case"beforetoggle":ft=W0}var Rn=(p&4)!==0,zi=!Rn&&(f==="scroll"||f==="scrollend"),Ke=Rn?ot!==null?ot+"Capture":null:ot;Rn=[];for(var Ue=ut,it;Ue!==null;){var yt=Ue;if(it=yt.stateNode,yt=yt.tag,yt!==5&&yt!==26&&yt!==27||it===null||Ke===null||(yt=As(Ue,Ke),yt!=null&&Rn.push(Js(Ue,yt,it))),zi)break;Ue=Ue.return}0<Rn.length&&(ot=new ft(ot,on,null,O,pt),xt.push({event:ot,listeners:Rn}))}}if((p&7)===0){e:{if(ot=f==="mouseover"||f==="pointerover",ft=f==="mouseout"||f==="pointerout",ot&&O!==ch&&(on=O.relatedTarget||O.fromElement)&&(hr(on)||on[an]))break e;if((ft||ot)&&(ot=pt.window===pt?pt:(ot=pt.ownerDocument)?ot.defaultView||ot.parentWindow:window,ft?(on=O.relatedTarget||O.toElement,ft=ut,on=on?hr(on):null,on!==null&&(zi=a(on),Rn=on.tag,on!==zi||Rn!==5&&Rn!==27&&Rn!==6)&&(on=null)):(ft=null,on=ut),ft!==on)){if(Rn=Ep,yt="onMouseLeave",Ke="onMouseEnter",Ue="mouse",(f==="pointerout"||f==="pointerover")&&(Rn=Cp,yt="onPointerLeave",Ke="onPointerEnter",Ue="pointer"),zi=ft==null?ot:Kr(ft),it=on==null?ot:Kr(on),ot=new Rn(yt,Ue+"leave",ft,O,pt),ot.target=zi,ot.relatedTarget=it,yt=null,hr(pt)===ut&&(Rn=new Rn(Ke,Ue+"enter",on,O,pt),Rn.target=it,Rn.relatedTarget=zi,yt=Rn),zi=yt,ft&&on)t:{for(Rn=tO,Ke=ft,Ue=on,it=0,yt=Ke;yt;yt=Rn(yt))it++;yt=0;for(var Sn=Ue;Sn;Sn=Rn(Sn))yt++;for(;0<it-yt;)Ke=Rn(Ke),it--;for(;0<yt-it;)Ue=Rn(Ue),yt--;for(;it--;){if(Ke===Ue||Ue!==null&&Ke===Ue.alternate){Rn=Ke;break t}Ke=Rn(Ke),Ue=Rn(Ue)}Rn=null}else Rn=null;ft!==null&&ba(xt,ot,ft,Rn,!1),on!==null&&zi!==null&&ba(xt,zi,on,Rn,!0)}}e:{if(ot=ut?Kr(ut):window,ft=ot.nodeName&&ot.nodeName.toLowerCase(),ft==="select"||ft==="input"&&ot.type==="file")var oi=J0;else if(Tp(ot))if(mu)oi=Ii;else{oi=_l;var hn=Kn}else ft=ot.nodeName,!ft||ft.toLowerCase()!=="input"||ot.type!=="checkbox"&&ot.type!=="radio"?ut&&vf(ut.elementType)&&(oi=J0):oi=Dg;if(oi&&(oi=oi(f,ut))){hd(xt,oi,O,pt);break e}hn&&hn(f,ot,ut),f==="focusout"&&ut&&ot.type==="number"&&ut.memoizedProps.value!=null&&id(ot,"number",ot.value)}switch(hn=ut?Kr(ut):window,f){case"focusin":(Tp(hn)||hn.contentEditable==="true")&&(Xe=hn,dt=ut,rt=null);break;case"focusout":rt=dt=Xe=null;break;case"mousedown":Bt=!0;break;case"contextmenu":case"mouseup":case"dragend":Bt=!1,En(xt,O,pt);break;case"selectionchange":if(Dc)break;case"keydown":case"keyup":En(xt,O,pt)}var fr;if(Ec)e:{switch(f){case"compositionstart":var Br="onCompositionStart";break e;case"compositionend":Br="onCompositionEnd";break e;case"compositionupdate":Br="onCompositionUpdate";break e}Br=void 0}else Ns?fd(f,O)&&(Br="onCompositionEnd"):f==="keydown"&&O.keyCode===229&&(Br="onCompositionStart");Br&&(Qa&&O.locale!=="ko"&&(Ns||Br!=="onCompositionStart"?Br==="onCompositionEnd"&&Ns&&(fr=Qn()):(Oc=pt,hh="value"in Oc?Oc.value:Oc.textContent,Ns=!0)),hn=ib(ut,Br),0<hn.length&&(Br=new Ba(Br,f,null,O,pt),xt.push({event:Br,listeners:hn}),fr?Br.data=fr:(fr=dd(O),fr!==null&&(Br.data=fr)))),(fr=Fs?pu(f,O):Ap(f,O))&&(Br=ib(ut,"onBeforeInput"),0<Br.length&&(hn=new Ba("onBeforeInput","beforeinput",null,O,pt),xt.push({event:hn,listeners:Br}),hn.data=fr)),$m(xt,f,ut,O,pt)}t_(xt,p)})}function Js(f,p,O){return{instance:f,listener:p,currentTarget:O}}function ib(f,p){for(var O=p+"Capture",D=[];f!==null;){var Q=f,Y=Q.stateNode;if(Q=Q.tag,Q!==5&&Q!==26&&Q!==27||Y===null||(Q=As(f,O),Q!=null&&D.unshift(Js(f,Q,Y)),Q=As(f,p),Q!=null&&D.push(Js(f,Q,Y))),f.tag===3)return D;f=f.return}return[]}function tO(f){if(f===null)return null;do f=f.return;while(f&&f.tag!==5&&f.tag!==27);return f||null}function ba(f,p,O,D,Q){for(var Y=p._reactName,ce=[];O!==null&&O!==D;){var _e=O,Le=_e.alternate,ut=_e.stateNode;if(_e=_e.tag,Le!==null&&Le===D)break;_e!==5&&_e!==26&&_e!==27||ut===null||(Le=ut,Q?(ut=As(O,Y),ut!=null&&ce.unshift(Js(O,ut,Le))):Q||(ut=As(O,Y),ut!=null&&ce.push(Js(O,ut,Le)))),O=O.return}ce.length!==0&&f.push({event:p,listeners:ce})}var ja=/\r\n?/g,n_=/\u0000|\uFFFD/g;function r_(f){return(typeof f=="string"?f:""+f).replace(ja,`
`).replace(n_,"")}function i_(f,p){return p=r_(p),r_(f)===p}function Qi(f,p,O,D,Q,Y){switch(O){case"children":typeof D=="string"?p==="body"||p==="textarea"&&D===""||Pu(f,D):(typeof D=="number"||typeof D=="bigint")&&p!=="body"&&Pu(f,""+D);break;case"className":Sl(f,"class",D);break;case"tabIndex":Sl(f,"tabindex",D);break;case"dir":case"role":case"viewBox":case"width":case"height":Sl(f,O,D);break;case"style":j0(f,D,Y);break;case"data":if(p!=="object"){Sl(f,"data",D);break}case"src":case"href":if(D===""&&(p!=="a"||O!=="href")){f.removeAttribute(O);break}if(D==null||typeof D=="function"||typeof D=="symbol"||typeof D=="boolean"){f.removeAttribute(O);break}D=yf(""+D),f.setAttribute(O,D);break;case"action":case"formAction":if(typeof D=="function"){f.setAttribute(O,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}else typeof Y=="function"&&(O==="formAction"?(p!=="input"&&Qi(f,p,"name",Q.name,Q,null),Qi(f,p,"formEncType",Q.formEncType,Q,null),Qi(f,p,"formMethod",Q.formMethod,Q,null),Qi(f,p,"formTarget",Q.formTarget,Q,null)):(Qi(f,p,"encType",Q.encType,Q,null),Qi(f,p,"method",Q.method,Q,null),Qi(f,p,"target",Q.target,Q,null)));if(D==null||typeof D=="symbol"||typeof D=="boolean"){f.removeAttribute(O);break}D=yf(""+D),f.setAttribute(O,D);break;case"onClick":D!=null&&(f.onclick=ho);break;case"onScroll":D!=null&&kr("scroll",f);break;case"onScrollEnd":D!=null&&kr("scrollend",f);break;case"dangerouslySetInnerHTML":if(D!=null){if(typeof D!="object"||!("__html"in D))throw Error(r(61));if(O=D.__html,O!=null){if(Q.children!=null)throw Error(r(60));f.innerHTML=O}}break;case"multiple":f.multiple=D&&typeof D!="function"&&typeof D!="symbol";break;case"muted":f.muted=D&&typeof D!="function"&&typeof D!="symbol";break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":break;case"autoFocus":break;case"xlinkHref":if(D==null||typeof D=="function"||typeof D=="boolean"||typeof D=="symbol"){f.removeAttribute("xlink:href");break}O=yf(""+D),f.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",O);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":D!=null&&typeof D!="function"&&typeof D!="symbol"?f.setAttribute(O,""+D):f.removeAttribute(O);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":D&&typeof D!="function"&&typeof D!="symbol"?f.setAttribute(O,""):f.removeAttribute(O);break;case"capture":case"download":D===!0?f.setAttribute(O,""):D!==!1&&D!=null&&typeof D!="function"&&typeof D!="symbol"?f.setAttribute(O,D):f.removeAttribute(O);break;case"cols":case"rows":case"size":case"span":D!=null&&typeof D!="function"&&typeof D!="symbol"&&!isNaN(D)&&1<=D?f.setAttribute(O,D):f.removeAttribute(O);break;case"rowSpan":case"start":D==null||typeof D=="function"||typeof D=="symbol"||isNaN(D)?f.removeAttribute(O):f.setAttribute(O,D);break;case"popover":kr("beforetoggle",f),kr("toggle",f),Rr(f,"popover",D);break;case"xlinkActuate":tr(f,"http://www.w3.org/1999/xlink","xlink:actuate",D);break;case"xlinkArcrole":tr(f,"http://www.w3.org/1999/xlink","xlink:arcrole",D);break;case"xlinkRole":tr(f,"http://www.w3.org/1999/xlink","xlink:role",D);break;case"xlinkShow":tr(f,"http://www.w3.org/1999/xlink","xlink:show",D);break;case"xlinkTitle":tr(f,"http://www.w3.org/1999/xlink","xlink:title",D);break;case"xlinkType":tr(f,"http://www.w3.org/1999/xlink","xlink:type",D);break;case"xmlBase":tr(f,"http://www.w3.org/XML/1998/namespace","xml:base",D);break;case"xmlLang":tr(f,"http://www.w3.org/XML/1998/namespace","xml:lang",D);break;case"xmlSpace":tr(f,"http://www.w3.org/XML/1998/namespace","xml:space",D);break;case"is":Rr(f,"is",D);break;case"innerText":case"textContent":break;default:(!(2<O.length)||O[0]!=="o"&&O[0]!=="O"||O[1]!=="n"&&O[1]!=="N")&&(O=Z0.get(O)||O,Rr(f,O,D))}}function k4(f,p,O,D,Q,Y){switch(O){case"style":j0(f,D,Y);break;case"dangerouslySetInnerHTML":if(D!=null){if(typeof D!="object"||!("__html"in D))throw Error(r(61));if(O=D.__html,O!=null){if(Q.children!=null)throw Error(r(60));f.innerHTML=O}}break;case"children":typeof D=="string"?Pu(f,D):(typeof D=="number"||typeof D=="bigint")&&Pu(f,""+D);break;case"onScroll":D!=null&&kr("scroll",f);break;case"onScrollEnd":D!=null&&kr("scrollend",f);break;case"onClick":D!=null&&(f.onclick=ho);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":break;case"innerText":case"textContent":break;default:if(!Xa.hasOwnProperty(O))e:{if(O[0]==="o"&&O[1]==="n"&&(Q=O.endsWith("Capture"),p=O.slice(2,Q?O.length-7:void 0),Y=f[Lt]||null,Y=Y!=null?Y[O]:null,typeof Y=="function"&&f.removeEventListener(p,Y,Q),typeof D=="function")){typeof Y!="function"&&Y!==null&&(O in f?f[O]=null:f.hasAttribute(O)&&f.removeAttribute(O)),f.addEventListener(p,D,Q);break e}O in f?f[O]=D:D===!0?f.setAttribute(O,""):Rr(f,O,D)}}}function Rl(f,p,O){switch(p){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":kr("error",f),kr("load",f);var D=!1,Q=!1,Y;for(Y in O)if(O.hasOwnProperty(Y)){var ce=O[Y];if(ce!=null)switch(Y){case"src":D=!0;break;case"srcSet":Q=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(r(137,p));default:Qi(f,p,Y,ce,O,null)}}Q&&Qi(f,p,"srcSet",O.srcSet,O,null),D&&Qi(f,p,"src",O.src,O,null);return;case"input":kr("invalid",f);var _e=Y=ce=Q=null,Le=null,ut=null;for(D in O)if(O.hasOwnProperty(D)){var pt=O[D];if(pt!=null)switch(D){case"name":Q=pt;break;case"type":ce=pt;break;case"checked":Le=pt;break;case"defaultChecked":ut=pt;break;case"value":Y=pt;break;case"defaultValue":_e=pt;break;case"children":case"dangerouslySetInnerHTML":if(pt!=null)throw Error(r(137,p));break;default:Qi(f,p,D,pt,O,null)}}$s(f,Y,_e,Le,ut,ce,Q,!1);return;case"select":kr("invalid",f),D=ce=Y=null;for(Q in O)if(O.hasOwnProperty(Q)&&(_e=O[Q],_e!=null))switch(Q){case"value":Y=_e;break;case"defaultValue":ce=_e;break;case"multiple":D=_e;default:Qi(f,p,Q,_e,O,null)}p=Y,O=ce,f.multiple=!!D,p!=null?vc(f,!!D,p,!1):O!=null&&vc(f,!!D,O,!0);return;case"textarea":kr("invalid",f),Y=Q=D=null;for(ce in O)if(O.hasOwnProperty(ce)&&(_e=O[ce],_e!=null))switch(ce){case"value":D=_e;break;case"defaultValue":Q=_e;break;case"children":Y=_e;break;case"dangerouslySetInnerHTML":if(_e!=null)throw Error(r(91));break;default:Qi(f,p,ce,_e,O,null)}Vo(f,D,Q,Y);return;case"option":for(Le in O)if(O.hasOwnProperty(Le)&&(D=O[Le],D!=null))switch(Le){case"selected":f.selected=D&&typeof D!="function"&&typeof D!="symbol";break;default:Qi(f,p,Le,D,O,null)}return;case"dialog":kr("beforetoggle",f),kr("toggle",f),kr("cancel",f),kr("close",f);break;case"iframe":case"object":kr("load",f);break;case"video":case"audio":for(D=0;D<rb.length;D++)kr(rb[D],f);break;case"image":kr("error",f),kr("load",f);break;case"details":kr("toggle",f);break;case"embed":case"source":case"link":kr("error",f),kr("load",f);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(ut in O)if(O.hasOwnProperty(ut)&&(D=O[ut],D!=null))switch(ut){case"children":case"dangerouslySetInnerHTML":throw Error(r(137,p));default:Qi(f,p,ut,D,O,null)}return;default:if(vf(p)){for(pt in O)O.hasOwnProperty(pt)&&(D=O[pt],D!==void 0&&k4(f,p,pt,D,O,void 0));return}}for(_e in O)O.hasOwnProperty(_e)&&(D=O[_e],D!=null&&Qi(f,p,_e,D,O,null))}function w5(f,p,O,D){switch(p){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var Q=null,Y=null,ce=null,_e=null,Le=null,ut=null,pt=null;for(ft in O){var xt=O[ft];if(O.hasOwnProperty(ft)&&xt!=null)switch(ft){case"checked":break;case"value":break;case"defaultValue":Le=xt;default:D.hasOwnProperty(ft)||Qi(f,p,ft,null,D,xt)}}for(var ot in D){var ft=D[ot];if(xt=O[ot],D.hasOwnProperty(ot)&&(ft!=null||xt!=null))switch(ot){case"type":Y=ft;break;case"name":Q=ft;break;case"checked":ut=ft;break;case"defaultChecked":pt=ft;break;case"value":ce=ft;break;case"defaultValue":_e=ft;break;case"children":case"dangerouslySetInnerHTML":if(ft!=null)throw Error(r(137,p));break;default:ft!==xt&&Qi(f,p,ot,ft,D,xt)}}rd(f,ce,_e,Le,ut,pt,Y,Q);return;case"select":ft=ce=_e=ot=null;for(Y in O)if(Le=O[Y],O.hasOwnProperty(Y)&&Le!=null)switch(Y){case"value":break;case"multiple":ft=Le;default:D.hasOwnProperty(Y)||Qi(f,p,Y,null,D,Le)}for(Q in D)if(Y=D[Q],Le=O[Q],D.hasOwnProperty(Q)&&(Y!=null||Le!=null))switch(Q){case"value":ot=Y;break;case"defaultValue":_e=Y;break;case"multiple":ce=Y;default:Y!==Le&&Qi(f,p,Q,Y,D,Le)}p=_e,O=ce,D=ft,ot!=null?vc(f,!!O,ot,!1):!!D!=!!O&&(p!=null?vc(f,!!O,p,!0):vc(f,!!O,O?[]:"",!1));return;case"textarea":ft=ot=null;for(_e in O)if(Q=O[_e],O.hasOwnProperty(_e)&&Q!=null&&!D.hasOwnProperty(_e))switch(_e){case"value":break;case"children":break;default:Qi(f,p,_e,null,D,Q)}for(ce in D)if(Q=D[ce],Y=O[ce],D.hasOwnProperty(ce)&&(Q!=null||Y!=null))switch(ce){case"value":ot=Q;break;case"defaultValue":ft=Q;break;case"children":break;case"dangerouslySetInnerHTML":if(Q!=null)throw Error(r(91));break;default:Q!==Y&&Qi(f,p,ce,Q,D,Y)}yc(f,ot,ft);return;case"option":for(var on in O)if(ot=O[on],O.hasOwnProperty(on)&&ot!=null&&!D.hasOwnProperty(on))switch(on){case"selected":f.selected=!1;break;default:Qi(f,p,on,null,D,ot)}for(Le in D)if(ot=D[Le],ft=O[Le],D.hasOwnProperty(Le)&&ot!==ft&&(ot!=null||ft!=null))switch(Le){case"selected":f.selected=ot&&typeof ot!="function"&&typeof ot!="symbol";break;default:Qi(f,p,Le,ot,D,ft)}return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var Rn in O)ot=O[Rn],O.hasOwnProperty(Rn)&&ot!=null&&!D.hasOwnProperty(Rn)&&Qi(f,p,Rn,null,D,ot);for(ut in D)if(ot=D[ut],ft=O[ut],D.hasOwnProperty(ut)&&ot!==ft&&(ot!=null||ft!=null))switch(ut){case"children":case"dangerouslySetInnerHTML":if(ot!=null)throw Error(r(137,p));break;default:Qi(f,p,ut,ot,D,ft)}return;default:if(vf(p)){for(var zi in O)ot=O[zi],O.hasOwnProperty(zi)&&ot!==void 0&&!D.hasOwnProperty(zi)&&k4(f,p,zi,void 0,D,ot);for(pt in D)ot=D[pt],ft=O[pt],!D.hasOwnProperty(pt)||ot===ft||ot===void 0&&ft===void 0||k4(f,p,pt,ot,D,ft);return}}for(var Ke in O)ot=O[Ke],O.hasOwnProperty(Ke)&&ot!=null&&!D.hasOwnProperty(Ke)&&Qi(f,p,Ke,null,D,ot);for(xt in D)ot=D[xt],ft=O[xt],!D.hasOwnProperty(xt)||ot===ft||ot==null&&ft==null||Qi(f,p,xt,ot,D,ft)}function P4(f){switch(f){case"css":case"script":case"font":case"img":case"image":case"input":case"link":return!0;default:return!1}}function ab(){if(typeof performance.getEntriesByType=="function"){for(var f=0,p=0,O=performance.getEntriesByType("resource"),D=0;D<O.length;D++){var Q=O[D],Y=Q.transferSize,ce=Q.initiatorType,_e=Q.duration;if(Y&&_e&&P4(ce)){for(ce=0,_e=Q.responseEnd,D+=1;D<O.length;D++){var Le=O[D],ut=Le.startTime;if(ut>_e)break;var pt=Le.transferSize,xt=Le.initiatorType;pt&&P4(xt)&&(Le=Le.responseEnd,ce+=pt*(Le<_e?1:(_e-ut)/(Le-ut)))}if(--D,p+=8*(Y+ce)/(Q.duration/1e3),f++,10<f)break}}if(0<f)return p/f/1e6}return navigator.connection&&(f=navigator.connection.downlink,typeof f=="number")?f:5}var ub=null,M4=null;function lb(f){return f.nodeType===9?f:f.ownerDocument}function ob(f){switch(f){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function ds(f,p){if(f===0)switch(p){case"svg":return 1;case"math":return 2;default:return 0}return f===1&&p==="foreignObject"?0:f}function Gp(f,p){return f==="textarea"||f==="noscript"||typeof p.children=="string"||typeof p.children=="number"||typeof p.children=="bigint"||typeof p.dangerouslySetInnerHTML=="object"&&p.dangerouslySetInnerHTML!==null&&p.dangerouslySetInnerHTML.__html!=null}var F4=null;function R5(){var f=window.event;return f&&f.type==="popstate"?f===F4?!1:(F4=f,!0):(F4=null,!1)}var a_=typeof setTimeout=="function"?setTimeout:void 0,A5=typeof clearTimeout=="function"?clearTimeout:void 0,u_=typeof Promise=="function"?Promise:void 0,l_=typeof queueMicrotask=="function"?queueMicrotask:typeof u_<"u"?function(f){return u_.resolve(null).then(f).catch(qh)}:a_;function qh(f){setTimeout(function(){throw f})}function Vn(f){return f==="head"}function N4(f,p){var O=p,D=0;do{var Q=O.nextSibling;if(f.removeChild(O),Q&&Q.nodeType===8)if(O=Q.data,O==="/$"||O==="/&"){if(D===0){f.removeChild(Q),dv(p);return}D--}else if(O==="$"||O==="$?"||O==="$~"||O==="$!"||O==="&")D++;else if(O==="html")sb(f.ownerDocument.documentElement);else if(O==="head"){O=f.ownerDocument.head,sb(O);for(var Y=O.firstChild;Y;){var ce=Y.nextSibling,_e=Y.nodeName;Y[Di]||_e==="SCRIPT"||_e==="STYLE"||_e==="LINK"&&Y.rel.toLowerCase()==="stylesheet"||O.removeChild(Y),Y=ce}}else O==="body"&&sb(f.ownerDocument.body);O=Q}while(O);dv(p)}function ec(f,p){var O=f;f=0;do{var D=O.nextSibling;if(O.nodeType===1?p?(O._stashedDisplay=O.style.display,O.style.display="none"):(O.style.display=O._stashedDisplay||"",O.getAttribute("style")===""&&O.removeAttribute("style")):O.nodeType===3&&(p?(O._stashedText=O.nodeValue,O.nodeValue=""):O.nodeValue=O._stashedText||""),D&&D.nodeType===8)if(O=D.data,O==="/$"){if(f===0)break;f--}else O!=="$"&&O!=="$?"&&O!=="$~"&&O!=="$!"||f++;O=D}while(O)}function nO(f){var p=f.firstChild;for(p&&p.nodeType===10&&(p=p.nextSibling);p;){var O=p;switch(p=p.nextSibling,O.nodeName){case"HTML":case"HEAD":case"BODY":nO(O),$i(O);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if(O.rel.toLowerCase()==="stylesheet")continue}f.removeChild(O)}}function T5(f,p,O,D){for(;f.nodeType===1;){var Q=O;if(f.nodeName.toLowerCase()!==p.toLowerCase()){if(!D&&(f.nodeName!=="INPUT"||f.type!=="hidden"))break}else if(D){if(!f[Di])switch(p){case"meta":if(!f.hasAttribute("itemprop"))break;return f;case"link":if(Y=f.getAttribute("rel"),Y==="stylesheet"&&f.hasAttribute("data-precedence"))break;if(Y!==Q.rel||f.getAttribute("href")!==(Q.href==null||Q.href===""?null:Q.href)||f.getAttribute("crossorigin")!==(Q.crossOrigin==null?null:Q.crossOrigin)||f.getAttribute("title")!==(Q.title==null?null:Q.title))break;return f;case"style":if(f.hasAttribute("data-precedence"))break;return f;case"script":if(Y=f.getAttribute("src"),(Y!==(Q.src==null?null:Q.src)||f.getAttribute("type")!==(Q.type==null?null:Q.type)||f.getAttribute("crossorigin")!==(Q.crossOrigin==null?null:Q.crossOrigin))&&Y&&f.hasAttribute("async")&&!f.hasAttribute("itemprop"))break;return f;default:return f}}else if(p==="input"&&f.type==="hidden"){var Y=Q.name==null?null:""+Q.name;if(Q.type==="hidden"&&f.getAttribute("name")===Y)return f}else return f;if(f=Zc(f.nextSibling),f===null)break}return null}function _i(f,p,O){if(p==="")return null;for(;f.nodeType!==3;)if((f.nodeType!==1||f.nodeName!=="INPUT"||f.type!=="hidden")&&!O||(f=Zc(f.nextSibling),f===null))return null;return f}function o_(f,p){for(;f.nodeType!==8;)if((f.nodeType!==1||f.nodeName!=="INPUT"||f.type!=="hidden")&&!p||(f=Zc(f.nextSibling),f===null))return null;return f}function rO(f){return f.data==="$?"||f.data==="$~"}function iv(f){return f.data==="$!"||f.data==="$?"&&f.ownerDocument.readyState!=="loading"}function k5(f,p){var O=f.ownerDocument;if(f.data==="$~")f._reactRetry=p;else if(f.data!=="$?"||O.readyState!=="loading")p();else{var D=function(){p(),O.removeEventListener("DOMContentLoaded",D)};O.addEventListener("DOMContentLoaded",D),f._reactRetry=D}}function Zc(f){for(;f!=null;f=f.nextSibling){var p=f.nodeType;if(p===1||p===3)break;if(p===8){if(p=f.data,p==="$"||p==="$!"||p==="$?"||p==="$~"||p==="&"||p==="F!"||p==="F")break;if(p==="/$"||p==="/&")return null}}return f}var I4=null;function s_(f){f=f.nextSibling;for(var p=0;f;){if(f.nodeType===8){var O=f.data;if(O==="/$"||O==="/&"){if(p===0)return Zc(f.nextSibling);p--}else O!=="$"&&O!=="$!"&&O!=="$?"&&O!=="$~"&&O!=="&"||p++}f=f.nextSibling}return null}function c_(f){f=f.previousSibling;for(var p=0;f;){if(f.nodeType===8){var O=f.data;if(O==="$"||O==="$!"||O==="$?"||O==="$~"||O==="&"){if(p===0)return f;p--}else O!=="/$"&&O!=="/&"||p++}f=f.previousSibling}return null}function f_(f,p,O){switch(p=lb(O),f){case"html":if(f=p.documentElement,!f)throw Error(r(452));return f;case"head":if(f=p.head,!f)throw Error(r(453));return f;case"body":if(f=p.body,!f)throw Error(r(454));return f;default:throw Error(r(451))}}function sb(f){for(var p=f.attributes;p.length;)f.removeAttributeNode(p[0]);$i(f)}var Hc=new Map,d_=new Set;function iO(f){return typeof f.getRootNode=="function"?f.getRootNode():f.nodeType===9?f:f.ownerDocument}var Vh=J.d;J.d={f:P5,r:M5,D:L4,C:F5,L:N5,m:I5,X:hs,S:pl,M:L5};function P5(){var f=Vh.f(),p=De();return f||p}function M5(f){var p=wt(f);p!==null&&p.tag===5&&p.type==="form"?Zg(p):Vh.r(f)}var Kp=typeof document>"u"?null:document;function h_(f,p,O){var D=Kp;if(D&&typeof p=="string"&&p){var Q=Ya(p);Q='link[rel="'+f+'"][href="'+Q+'"]',typeof O=="string"&&(Q+='[crossorigin="'+O+'"]'),d_.has(Q)||(d_.add(Q),f={rel:f,crossOrigin:O,href:p},D.querySelector(Q)===null&&(p=D.createElement("link"),Rl(p,"link",f),yi(p),D.head.appendChild(p)))}}function L4(f){Vh.D(f),h_("dns-prefetch",f,null)}function F5(f,p){Vh.C(f,p),h_("preconnect",f,p)}function N5(f,p,O){Vh.L(f,p,O);var D=Kp;if(D&&f&&p){var Q='link[rel="preload"][as="'+Ya(p)+'"]';p==="image"&&O&&O.imageSrcSet?(Q+='[imagesrcset="'+Ya(O.imageSrcSet)+'"]',typeof O.imageSizes=="string"&&(Q+='[imagesizes="'+Ya(O.imageSizes)+'"]')):Q+='[href="'+Ya(f)+'"]';var Y=Q;switch(p){case"style":Y=av(f);break;case"script":Y=lv(f)}Hc.has(Y)||(f=d({rel:"preload",href:p==="image"&&O&&O.imageSrcSet?void 0:f,as:p},O),Hc.set(Y,f),D.querySelector(Q)!==null||p==="style"&&D.querySelector(uv(Y))||p==="script"&&D.querySelector(ov(Y))||(p=D.createElement("link"),Rl(p,"link",f),yi(p),D.head.appendChild(p)))}}function I5(f,p){Vh.m(f,p);var O=Kp;if(O&&f){var D=p&&typeof p.as=="string"?p.as:"script",Q='link[rel="modulepreload"][as="'+Ya(D)+'"][href="'+Ya(f)+'"]',Y=Q;switch(D){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":Y=lv(f)}if(!Hc.has(Y)&&(f=d({rel:"modulepreload",href:f},p),Hc.set(Y,f),O.querySelector(Q)===null)){switch(D){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(O.querySelector(ov(Y)))return}D=O.createElement("link"),Rl(D,"link",f),yi(D),O.head.appendChild(D)}}}function pl(f,p,O){Vh.S(f,p,O);var D=Kp;if(D&&f){var Q=Zi(D).hoistableStyles,Y=av(f);p=p||"default";var ce=Q.get(Y);if(!ce){var _e={loading:0,preload:null};if(ce=D.querySelector(uv(Y)))_e.loading=5;else{f=d({rel:"stylesheet",href:f,"data-precedence":p},O),(O=Hc.get(Y))&&B4(f,O);var Le=ce=D.createElement("link");yi(Le),Rl(Le,"link",f),Le._p=new Promise(function(ut,pt){Le.onload=ut,Le.onerror=pt}),Le.addEventListener("load",function(){_e.loading|=1}),Le.addEventListener("error",function(){_e.loading|=2}),_e.loading|=4,aO(ce,p,D)}ce={type:"stylesheet",instance:ce,count:1,state:_e},Q.set(Y,ce)}}}function hs(f,p){Vh.X(f,p);var O=Kp;if(O&&f){var D=Zi(O).hoistableScripts,Q=lv(f),Y=D.get(Q);Y||(Y=O.querySelector(ov(Q)),Y||(f=d({src:f,async:!0},p),(p=Hc.get(Q))&&uO(f,p),Y=O.createElement("script"),yi(Y),Rl(Y,"link",f),O.head.appendChild(Y)),Y={type:"script",instance:Y,count:1,state:null},D.set(Q,Y))}}function L5(f,p){Vh.M(f,p);var O=Kp;if(O&&f){var D=Zi(O).hoistableScripts,Q=lv(f),Y=D.get(Q);Y||(Y=O.querySelector(ov(Q)),Y||(f=d({src:f,async:!0,type:"module"},p),(p=Hc.get(Q))&&uO(f,p),Y=O.createElement("script"),yi(Y),Rl(Y,"link",f),O.head.appendChild(Y)),Y={type:"script",instance:Y,count:1,state:null},D.set(Q,Y))}}function p_(f,p,O,D){var Q=(Q=ie.current)?iO(Q):null;if(!Q)throw Error(r(446));switch(f){case"meta":case"title":return null;case"style":return typeof O.precedence=="string"&&typeof O.href=="string"?(p=av(O.href),O=Zi(Q).hoistableStyles,D=O.get(p),D||(D={type:"style",instance:null,count:0,state:null},O.set(p,D)),D):{type:"void",instance:null,count:0,state:null};case"link":if(O.rel==="stylesheet"&&typeof O.href=="string"&&typeof O.precedence=="string"){f=av(O.href);var Y=Zi(Q).hoistableStyles,ce=Y.get(f);if(ce||(Q=Q.ownerDocument||Q,ce={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},Y.set(f,ce),(Y=Q.querySelector(uv(f)))&&!Y._p&&(ce.instance=Y,ce.state.loading=5),Hc.has(f)||(O={rel:"preload",as:"style",href:O.href,crossOrigin:O.crossOrigin,integrity:O.integrity,media:O.media,hrefLang:O.hrefLang,referrerPolicy:O.referrerPolicy},Hc.set(f,O),Y||B5(Q,f,O,ce.state))),p&&D===null)throw Error(r(528,""));return ce}if(p&&D!==null)throw Error(r(529,""));return null;case"script":return p=O.async,O=O.src,typeof O=="string"&&p&&typeof p!="function"&&typeof p!="symbol"?(p=lv(O),O=Zi(Q).hoistableScripts,D=O.get(p),D||(D={type:"script",instance:null,count:0,state:null},O.set(p,D)),D):{type:"void",instance:null,count:0,state:null};default:throw Error(r(444,f))}}function av(f){return'href="'+Ya(f)+'"'}function uv(f){return'link[rel="stylesheet"]['+f+"]"}function m_(f){return d({},f,{"data-precedence":f.precedence,precedence:null})}function B5(f,p,O,D){f.querySelector('link[rel="preload"][as="style"]['+p+"]")?D.loading=1:(p=f.createElement("link"),D.preload=p,p.addEventListener("load",function(){return D.loading|=1}),p.addEventListener("error",function(){return D.loading|=2}),Rl(p,"link",O),yi(p),f.head.appendChild(p))}function lv(f){return'[src="'+Ya(f)+'"]'}function ov(f){return"script[async]"+f}function g_(f,p,O){if(p.count++,p.instance===null)switch(p.type){case"style":var D=f.querySelector('style[data-href~="'+Ya(O.href)+'"]');if(D)return p.instance=D,yi(D),D;var Q=d({},O,{"data-href":O.href,"data-precedence":O.precedence,href:null,precedence:null});return D=(f.ownerDocument||f).createElement("style"),yi(D),Rl(D,"style",Q),aO(D,O.precedence,f),p.instance=D;case"stylesheet":Q=av(O.href);var Y=f.querySelector(uv(Q));if(Y)return p.state.loading|=4,p.instance=Y,yi(Y),Y;D=m_(O),(Q=Hc.get(Q))&&B4(D,Q),Y=(f.ownerDocument||f).createElement("link"),yi(Y);var ce=Y;return ce._p=new Promise(function(_e,Le){ce.onload=_e,ce.onerror=Le}),Rl(Y,"link",D),p.state.loading|=4,aO(Y,O.precedence,f),p.instance=Y;case"script":return Y=lv(O.src),(Q=f.querySelector(ov(Y)))?(p.instance=Q,yi(Q),Q):(D=O,(Q=Hc.get(Y))&&(D=d({},O),uO(D,Q)),f=f.ownerDocument||f,Q=f.createElement("script"),yi(Q),Rl(Q,"link",D),f.head.appendChild(Q),p.instance=Q);case"void":return null;default:throw Error(r(443,p.type))}else p.type==="stylesheet"&&(p.state.loading&4)===0&&(D=p.instance,p.state.loading|=4,aO(D,O.precedence,f));return p.instance}function aO(f,p,O){for(var D=O.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),Q=D.length?D[D.length-1]:null,Y=Q,ce=0;ce<D.length;ce++){var _e=D[ce];if(_e.dataset.precedence===p)Y=_e;else if(Y!==Q)break}Y?Y.parentNode.insertBefore(f,Y.nextSibling):(p=O.nodeType===9?O.head:O,p.insertBefore(f,p.firstChild))}function B4(f,p){f.crossOrigin==null&&(f.crossOrigin=p.crossOrigin),f.referrerPolicy==null&&(f.referrerPolicy=p.referrerPolicy),f.title==null&&(f.title=p.title)}function uO(f,p){f.crossOrigin==null&&(f.crossOrigin=p.crossOrigin),f.referrerPolicy==null&&(f.referrerPolicy=p.referrerPolicy),f.integrity==null&&(f.integrity=p.integrity)}var cb=null;function v_(f,p,O){if(cb===null){var D=new Map,Q=cb=new Map;Q.set(O,D)}else Q=cb,D=Q.get(O),D||(D=new Map,Q.set(O,D));if(D.has(f))return D;for(D.set(f,null),O=O.getElementsByTagName(f),Q=0;Q<O.length;Q++){var Y=O[Q];if(!(Y[Di]||Y[Dt]||f==="link"&&Y.getAttribute("rel")==="stylesheet")&&Y.namespaceURI!=="http://www.w3.org/2000/svg"){var ce=Y.getAttribute(p)||"";ce=f+ce;var _e=D.get(ce);_e?_e.push(Y):D.set(ce,[Y])}}return D}function y_(f,p,O){f=f.ownerDocument||f,f.head.insertBefore(O,p==="title"?f.querySelector("head > title"):null)}function Q5(f,p,O){if(O===1||p.itemProp!=null)return!1;switch(f){case"meta":case"title":return!0;case"style":if(typeof p.precedence!="string"||typeof p.href!="string"||p.href==="")break;return!0;case"link":if(typeof p.rel!="string"||typeof p.href!="string"||p.href===""||p.onLoad||p.onError)break;switch(p.rel){case"stylesheet":return f=p.disabled,typeof p.precedence=="string"&&f==null;default:return!0}case"script":if(p.async&&typeof p.async!="function"&&typeof p.async!="symbol"&&!p.onLoad&&!p.onError&&p.src&&typeof p.src=="string")return!0}return!1}function b_(f){return!(f.type==="stylesheet"&&(f.state.loading&3)===0)}function sv(f,p,O,D){if(O.type==="stylesheet"&&(typeof D.media!="string"||matchMedia(D.media).matches!==!1)&&(O.state.loading&4)===0){if(O.instance===null){var Q=av(D.href),Y=p.querySelector(uv(Q));if(Y){p=Y._p,p!==null&&typeof p=="object"&&typeof p.then=="function"&&(f.count++,f=lO.bind(f),p.then(f,f)),O.state.loading|=4,O.instance=Y,yi(Y);return}Y=p.ownerDocument||p,D=m_(D),(Q=Hc.get(Q))&&B4(D,Q),Y=Y.createElement("link"),yi(Y);var ce=Y;ce._p=new Promise(function(_e,Le){ce.onload=_e,ce.onerror=Le}),Rl(Y,"link",D),O.instance=Y}f.stylesheets===null&&(f.stylesheets=new Map),f.stylesheets.set(O,p),(p=O.state.preload)&&(O.state.loading&3)===0&&(f.count++,O=lO.bind(f),p.addEventListener("load",O),p.addEventListener("error",O))}}var Q4=0;function z5(f,p){return f.stylesheets&&f.count===0&&sO(f,f.stylesheets),0<f.count||0<f.imgCount?function(O){var D=setTimeout(function(){if(f.stylesheets&&sO(f,f.stylesheets),f.unsuspend){var Y=f.unsuspend;f.unsuspend=null,Y()}},6e4+p);0<f.imgBytes&&Q4===0&&(Q4=62500*ab());var Q=setTimeout(function(){if(f.waitingForImages=!1,f.count===0&&(f.stylesheets&&sO(f,f.stylesheets),f.unsuspend)){var Y=f.unsuspend;f.unsuspend=null,Y()}},(f.imgBytes>Q4?50:800)+p);return f.unsuspend=O,function(){f.unsuspend=null,clearTimeout(D),clearTimeout(Q)}}:null}function lO(){if(this.count--,this.count===0&&(this.imgCount===0||!this.waitingForImages)){if(this.stylesheets)sO(this,this.stylesheets);else if(this.unsuspend){var f=this.unsuspend;this.unsuspend=null,f()}}}var oO=null;function sO(f,p){f.stylesheets=null,f.unsuspend!==null&&(f.count++,oO=new Map,p.forEach(O_,f),oO=null,lO.call(f))}function O_(f,p){if(!(p.state.loading&4)){var O=oO.get(f);if(O)var D=O.get(null);else{O=new Map,oO.set(f,O);for(var Q=f.querySelectorAll("link[data-precedence],style[data-precedence]"),Y=0;Y<Q.length;Y++){var ce=Q[Y];(ce.nodeName==="LINK"||ce.getAttribute("media")!=="not all")&&(O.set(ce.dataset.precedence,ce),D=ce)}D&&O.set(null,D)}Q=p.instance,ce=Q.getAttribute("data-precedence"),Y=O.get(ce)||D,Y===D&&O.set(null,Q),O.set(ce,Q),this.count++,D=lO.bind(this),Q.addEventListener("load",D),Q.addEventListener("error",D),Y?Y.parentNode.insertBefore(Q,Y.nextSibling):(f=f.nodeType===9?f.head:f,f.insertBefore(Q,f.firstChild)),p.state.loading|=4}}var cv={$$typeof:$,Provider:null,Consumer:null,_currentValue:te,_currentValue2:te,_threadCount:0};function q5(f,p,O,D,Q,Y,ce,_e,Le){this.tag=1,this.containerInfo=f,this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=Gr(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=Gr(0),this.hiddenUpdates=Gr(null),this.identifierPrefix=D,this.onUncaughtError=Q,this.onCaughtError=Y,this.onRecoverableError=ce,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=Le,this.incompleteTransitions=new Map}function x_(f,p,O,D,Q,Y,ce,_e,Le,ut,pt,xt){return f=new q5(f,p,O,ce,Le,ut,pt,xt,_e),p=1,Y===!0&&(p|=24),Y=aa(3,null,null,p),f.current=Y,Y.stateNode=f,p=Bs(),p.refCount++,f.pooledCache=p,p.refCount++,Y.memoizedState={element:D,isDehydrated:O,cache:p},Pc(Y),f}function S_(f){return f?(f=vu,f):vu}function __(f,p,O,D,Q,Y){Q=S_(Q),D.context===null?D.context=Q:D.pendingContext=Q,D=Mc(p),D.payload={element:O},Y=Y===void 0?null:Y,Y!==null&&(D.callback=Y),O=ns(f,D,p),O!==null&&(b(O,f,p),Sd(O,f,p))}function E_(f,p){if(f=f.memoizedState,f!==null&&f.dehydrated!==null){var O=f.retryLane;f.retryLane=O!==0&&O<p?O:p}}function z4(f,p){E_(f,p),(f=f.alternate)&&E_(f,p)}function D_(f){if(f.tag===13||f.tag===31){var p=el(f,67108864);p!==null&&b(p,f,67108864),z4(f,67108864)}}function $_(f){if(f.tag===13||f.tag===31){var p=wl();p=nt(p);var O=el(f,p);O!==null&&b(O,f,p),z4(f,p)}}var cO=!0;function V5(f,p,O,D){var Q=K.T;K.T=null;var Y=J.p;try{J.p=2,q4(f,p,O,D)}finally{J.p=Y,K.T=Q}}function U5(f,p,O,D){var Q=K.T;K.T=null;var Y=J.p;try{J.p=8,q4(f,p,O,D)}finally{J.p=Y,K.T=Q}}function q4(f,p,O,D){if(cO){var Q=V4(D);if(Q===null)T4(f,p,D,fO,O),w_(f,D);else if(R_(Q,f,p,O,D))D.stopPropagation();else if(w_(f,D),p&4&&-1<j5.indexOf(f)){for(;Q!==null;){var Y=wt(Q);if(Y!==null)switch(Y.tag){case 3:if(Y=Y.stateNode,Y.current.memoizedState.isDehydrated){var ce=Xn(Y.pendingLanes);if(ce!==0){var _e=Y;for(_e.pendingLanes|=2,_e.entangledLanes|=2;ce;){var Le=1<<31-Fe(ce);_e.entanglements[1]|=Le,ce&=~Le}Id(Y),(Nr&6)===0&&(hl=je()+500,nb(0))}}break;case 31:case 13:_e=el(Y,2),_e!==null&&b(_e,Y,2),De(),z4(Y,2)}if(Y=V4(D),Y===null&&T4(f,p,D,fO,O),Y===Q)break;Q=Y}Q!==null&&D.stopPropagation()}else T4(f,p,D,null,O)}}function V4(f){return f=fh(f),U4(f)}var fO=null;function U4(f){if(fO=null,f=hr(f),f!==null){var p=a(f);if(p===null)f=null;else{var O=p.tag;if(O===13){if(f=u(p),f!==null)return f;f=null}else if(O===31){if(f=l(p),f!==null)return f;f=null}else if(O===3){if(p.stateNode.current.memoizedState.isDehydrated)return p.tag===3?p.stateNode.containerInfo:null;f=null}else p!==f&&(f=null)}}return fO=f,null}function C_(f){switch(f){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(Ye()){case Ne:return 2;case ze:return 8;case at:case mt:return 32;case Mt:return 268435456;default:return 32}default:return 32}}var j4=!1,Wp=null,Jp=null,e0=null,fb=new Map,db=new Map,t0=[],j5="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function w_(f,p){switch(f){case"focusin":case"focusout":Wp=null;break;case"dragenter":case"dragleave":Jp=null;break;case"mouseover":case"mouseout":e0=null;break;case"pointerover":case"pointerout":fb.delete(p.pointerId);break;case"gotpointercapture":case"lostpointercapture":db.delete(p.pointerId)}}function Cm(f,p,O,D,Q,Y){return f===null||f.nativeEvent!==Y?(f={blockedOn:p,domEventName:O,eventSystemFlags:D,nativeEvent:Y,targetContainers:[Q]},p!==null&&(p=wt(p),p!==null&&D_(p)),f):(f.eventSystemFlags|=D,p=f.targetContainers,Q!==null&&p.indexOf(Q)===-1&&p.push(Q),f)}function R_(f,p,O,D,Q){switch(p){case"focusin":return Wp=Cm(Wp,f,p,O,D,Q),!0;case"dragenter":return Jp=Cm(Jp,f,p,O,D,Q),!0;case"mouseover":return e0=Cm(e0,f,p,O,D,Q),!0;case"pointerover":var Y=Q.pointerId;return fb.set(Y,Cm(fb.get(Y)||null,f,p,O,D,Q)),!0;case"gotpointercapture":return Y=Q.pointerId,db.set(Y,Cm(db.get(Y)||null,f,p,O,D,Q)),!0}return!1}function Z4(f){var p=hr(f.target);if(p!==null){var O=a(p);if(O!==null){if(p=O.tag,p===13){if(p=u(O),p!==null){f.blockedOn=p,Vt(f.priority,function(){$_(O)});return}}else if(p===31){if(p=l(O),p!==null){f.blockedOn=p,Vt(f.priority,function(){$_(O)});return}}else if(p===3&&O.stateNode.current.memoizedState.isDehydrated){f.blockedOn=O.tag===3?O.stateNode.containerInfo:null;return}}}f.blockedOn=null}function dO(f){if(f.blockedOn!==null)return!1;for(var p=f.targetContainers;0<p.length;){var O=V4(f.nativeEvent);if(O===null){O=f.nativeEvent;var D=new O.constructor(O.type,O);ch=D,O.target.dispatchEvent(D),ch=null}else return p=wt(O),p!==null&&D_(p),f.blockedOn=O,!1;p.shift()}return!0}function A_(f,p,O){dO(f)&&O.delete(p)}function Z5(){j4=!1,Wp!==null&&dO(Wp)&&(Wp=null),Jp!==null&&dO(Jp)&&(Jp=null),e0!==null&&dO(e0)&&(e0=null),fb.forEach(A_),db.forEach(A_)}function fv(f,p){f.blockedOn===p&&(f.blockedOn=null,j4||(j4=!0,e.unstable_scheduleCallback(e.unstable_NormalPriority,Z5)))}var hO=null;function T_(f){hO!==f&&(hO=f,e.unstable_scheduleCallback(e.unstable_NormalPriority,function(){hO===f&&(hO=null);for(var p=0;p<f.length;p+=3){var O=f[p],D=f[p+1],Q=f[p+2];if(typeof D!="function"){if(U4(D||O)===null)continue;break}var Y=wt(O);Y!==null&&(f.splice(p,3),p-=3,$d(Y,{pending:!0,data:Q,method:O.method,action:D},D,Q))}}))}function dv(f){function p(Le){return fv(Le,f)}Wp!==null&&fv(Wp,f),Jp!==null&&fv(Jp,f),e0!==null&&fv(e0,f),fb.forEach(p),db.forEach(p);for(var O=0;O<t0.length;O++){var D=t0[O];D.blockedOn===f&&(D.blockedOn=null)}for(;0<t0.length&&(O=t0[0],O.blockedOn===null);)Z4(O),O.blockedOn===null&&t0.shift();if(O=(f.ownerDocument||f).$$reactFormReplay,O!=null)for(D=0;D<O.length;D+=3){var Q=O[D],Y=O[D+1],ce=Q[Lt]||null;if(typeof Y=="function")ce||T_(O);else if(ce){var _e=null;if(Y&&Y.hasAttribute("formAction")){if(Q=Y,ce=Y[Lt]||null)_e=ce.formAction;else if(U4(Q)!==null)continue}else _e=ce.action;typeof _e=="function"?O[D+1]=_e:(O.splice(D,3),D-=3),T_(O)}}}function ko(){function f(Y){Y.canIntercept&&Y.info==="react-transition"&&Y.intercept({handler:function(){return new Promise(function(ce){return Q=ce})},focusReset:"manual",scroll:"manual"})}function p(){Q!==null&&(Q(),Q=null),D||setTimeout(O,20)}function O(){if(!D&&!navigation.transition){var Y=navigation.currentEntry;Y&&Y.url!=null&&navigation.navigate(Y.url,{state:Y.getState(),info:"react-transition",history:"replace"})}}if(typeof navigation=="object"){var D=!1,Q=null;return navigation.addEventListener("navigate",f),navigation.addEventListener("navigatesuccess",p),navigation.addEventListener("navigateerror",p),setTimeout(O,100),function(){D=!0,navigation.removeEventListener("navigate",f),navigation.removeEventListener("navigatesuccess",p),navigation.removeEventListener("navigateerror",p),Q!==null&&(Q(),Q=null)}}}function hb(f){this._internalRoot=f}pO.prototype.render=hb.prototype.render=function(f){var p=this._internalRoot;if(p===null)throw Error(r(409));var O=p.current,D=wl();__(O,D,f,p,null,null)},pO.prototype.unmount=hb.prototype.unmount=function(){var f=this._internalRoot;if(f!==null){this._internalRoot=null;var p=f.containerInfo;__(f.current,2,null,f,null,null),De(),p[an]=null}};function pO(f){this._internalRoot=f}pO.prototype.unstable_scheduleHydration=function(f){if(f){var p=Ct();f={blockedOn:null,target:f,priority:p};for(var O=0;O<t0.length&&p!==0&&p<t0[O].priority;O++);t0.splice(O,0,f),O===0&&Z4(f)}};var k_=t.version;if(k_!=="19.2.0")throw Error(r(527,k_,"19.2.0"));J.findDOMNode=function(f){var p=f._reactInternals;if(p===void 0)throw typeof f.render=="function"?Error(r(188)):(f=Object.keys(f).join(","),Error(r(268,f)));return f=s(p),f=f!==null?c(f):null,f=f===null?null:f.stateNode,f};var H5={bundleType:0,version:"19.2.0",rendererPackageName:"react-dom",currentDispatcherRef:K,reconcilerVersion:"19.2.0"};if(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__<"u"){var mO=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!mO.isDisabled&&mO.supportsFiber)try{qt=mO.inject(H5),Xt=mO}catch{}}return Y4.createRoot=function(f,p){if(!i(f))throw Error(r(299));var O=!1,D="",Q=Hg,Y=Ky,ce=Xg;return p!=null&&(p.unstable_strictMode===!0&&(O=!0),p.identifierPrefix!==void 0&&(D=p.identifierPrefix),p.onUncaughtError!==void 0&&(Q=p.onUncaughtError),p.onCaughtError!==void 0&&(Y=p.onCaughtError),p.onRecoverableError!==void 0&&(ce=p.onRecoverableError)),p=x_(f,1,!1,null,null,O,D,null,Q,Y,ce,ko),f[an]=p.current,A4(f),new hb(p)},Y4.hydrateRoot=function(f,p,O){if(!i(f))throw Error(r(299));var D=!1,Q="",Y=Hg,ce=Ky,_e=Xg,Le=null;return O!=null&&(O.unstable_strictMode===!0&&(D=!0),O.identifierPrefix!==void 0&&(Q=O.identifierPrefix),O.onUncaughtError!==void 0&&(Y=O.onUncaughtError),O.onCaughtError!==void 0&&(ce=O.onCaughtError),O.onRecoverableError!==void 0&&(_e=O.onRecoverableError),O.formState!==void 0&&(Le=O.formState)),p=x_(f,1,!0,p,O??null,D,Q,Le,Y,ce,_e,ko),p.context=S_(null),O=p.current,D=wl(),D=nt(D),Q=Mc(D),Q.callback=null,ns(O,Q,D),O=D,p.current.lanes=O,ji(p,O),Id(p),f[an]=p.current,A4(f),new pO(p)},Y4.version="19.2.0",Y4}var cQ;function B3e(){if(cQ)return n6.exports;cQ=1;function e(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(t){console.error(t)}}return e(),n6.exports=L3e(),n6.exports}var TJ=B3e();const irt=da(TJ);var u6,fQ;function oy(){if(fQ)return u6;fQ=1;var e={ACTOR_CHANGE:"ActorChange",CATCH_FIELD:"CatchField",CONDITION:"Condition",CLIENT_COMPONENT:"ClientComponent",CLIENT_EDGE_TO_SERVER_OBJECT:"ClientEdgeToServerObject",CLIENT_EDGE_TO_CLIENT_OBJECT:"ClientEdgeToClientObject",CLIENT_EXTENSION:"ClientExtension",DEFER:"Defer",CONNECTION:"Connection",FRAGMENT:"Fragment",FRAGMENT_SPREAD:"FragmentSpread",INLINE_DATA_FRAGMENT_SPREAD:"InlineDataFragmentSpread",INLINE_DATA_FRAGMENT:"InlineDataFragment",INLINE_FRAGMENT:"InlineFragment",LINKED_FIELD:"LinkedField",LINKED_HANDLE:"LinkedHandle",LITERAL:"Literal",LIST_VALUE:"ListValue",LOCAL_ARGUMENT:"LocalArgument",MODULE_IMPORT:"ModuleImport",ALIASED_FRAGMENT_SPREAD:"AliasedFragmentSpread",ALIASED_INLINE_FRAGMENT_SPREAD:"AliasedInlineFragmentSpread",RELAY_RESOLVER:"RelayResolver",RELAY_LIVE_RESOLVER:"RelayLiveResolver",REQUIRED_FIELD:"RequiredField",OBJECT_VALUE:"ObjectValue",OPERATION:"Operation",REQUEST:"Request",ROOT_ARGUMENT:"RootArgument",SCALAR_FIELD:"ScalarField",SCALAR_HANDLE:"ScalarHandle",SPLIT_OPERATION:"SplitOperation",STREAM:"Stream",TYPE_DISCRIMINATOR:"TypeDiscriminator",UPDATABLE_QUERY:"UpdatableQuery",VARIABLE:"Variable"};return u6=e,u6}var l6,dQ;function un(){if(dQ)return l6;dQ=1;var e=function(t,n,r,i,a,u,l,o){if(!t){var s;if(n===void 0)s=new Error("Minified exception occurred; use the non-minified dev environment for the full error message and additional helpful warnings.");else{var c=[r,i,a,u,l,o],d=0;s=new Error(n.replace(/%s/g,function(){return c[d++]})),s.name="Invariant Violation"}throw s.framesToPop=1,s}};return l6=e,l6}var o6,hQ;function Q3e(){if(hQ)return o6;hQ=1;function e(n){return function(){return n}}var t=function(){};return t.thatReturns=e,t.thatReturnsFalse=e(!1),t.thatReturnsTrue=e(!0),t.thatReturnsNull=e(null),t.thatReturnsThis=function(){return this},t.thatReturnsArgument=function(n){return n},o6=t,o6}var s6,pQ;function Vi(){if(pQ)return s6;pQ=1;var e=Q3e(),t=e;return s6=t,s6}var c6,mQ;function dp(){if(mQ)return c6;mQ=1;var e=oy(),t=un();Vi();function n(y){t(!1)}function r(y){var v=y;return typeof v=="function"?v=v():v.default&&(v=v.default),v}function i(y){var v=r(y);return typeof v=="object"&&v!==null&&v.kind===e.FRAGMENT}function a(y){var v=r(y);return typeof v=="object"&&v!==null&&v.kind===e.REQUEST}function u(y){var v=r(y);return typeof v=="object"&&v!==null&&v.kind===e.UPDATABLE_QUERY}function l(y){var v=r(y);return typeof v=="object"&&v!==null&&v.kind===e.INLINE_DATA_FRAGMENT}function o(y){var v=r(y);return i(v)||t(!1),v}function s(y){var v,S=o(y),_=(v=S.metadata)===null||v===void 0?void 0:v.refetch,E=_==null?void 0:_.connection;return _===null||typeof _!="object"||E===null||typeof E!="object"?null:S}function c(y){var v,S=o(y),_=(v=S.metadata)===null||v===void 0?void 0:v.refetch;return _===null||typeof _!="object"?null:S}function d(y){var v=r(y);return a(v)||t(!1),v}function h(y){var v=r(y);return u(v)||t(!1),v}function m(y){var v=r(y);return l(v)||t(!1),v}return c6={getFragment:o,getNode:r,getPaginationFragment:s,getRefetchableFragment:c,getRequest:d,getUpdatableQuery:h,getInlineDataFragment:m,graphql:n,isFragment:i,isRequest:a,isUpdatableQuery:u,isInlineDataFragment:l},c6}var f6={exports:{}},gQ;function $n(){return gQ||(gQ=1,(function(e){function t(n){return n&&n.__esModule?n:{default:n}}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})(f6)),f6.exports}var d6={exports:{}},h6={exports:{}},p6={exports:{}},m6={exports:{}},vQ;function kJ(){return vQ||(vQ=1,(function(e){function t(n){"@babel/helpers - typeof";return e.exports=t=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(r){return typeof r}:function(r){return r&&typeof Symbol=="function"&&r.constructor===Symbol&&r!==Symbol.prototype?"symbol":typeof r},e.exports.__esModule=!0,e.exports.default=e.exports,t(n)}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})(m6)),m6.exports}var g6={exports:{}},yQ;function z3e(){return yQ||(yQ=1,(function(e){var t=kJ().default;function n(r,i){if(t(r)!="object"||!r)return r;var a=r[Symbol.toPrimitive];if(a!==void 0){var u=a.call(r,i||"default");if(t(u)!="object")return u;throw new TypeError("@@toPrimitive must return a primitive value.")}return(i==="string"?String:Number)(r)}e.exports=n,e.exports.__esModule=!0,e.exports.default=e.exports})(g6)),g6.exports}var bQ;function PJ(){return bQ||(bQ=1,(function(e){var t=kJ().default,n=z3e();function r(i){var a=n(i,"string");return t(a)=="symbol"?a:a+""}e.exports=r,e.exports.__esModule=!0,e.exports.default=e.exports})(p6)),p6.exports}var OQ;function mc(){return OQ||(OQ=1,(function(e){var t=PJ();function n(r,i,a){return(i=t(i))in r?Object.defineProperty(r,i,{value:a,enumerable:!0,configurable:!0,writable:!0}):r[i]=a,r}e.exports=n,e.exports.__esModule=!0,e.exports.default=e.exports})(h6)),h6.exports}var xQ;function Ui(){return xQ||(xQ=1,(function(e){var t=mc();function n(i,a){var u=Object.keys(i);if(Object.getOwnPropertySymbols){var l=Object.getOwnPropertySymbols(i);a&&(l=l.filter(function(o){return Object.getOwnPropertyDescriptor(i,o).enumerable})),u.push.apply(u,l)}return u}function r(i){for(var a=1;a<arguments.length;a++){var u=arguments[a]!=null?arguments[a]:{};a%2?n(Object(u),!0).forEach(function(l){t(i,l,u[l])}):Object.getOwnPropertyDescriptors?Object.defineProperties(i,Object.getOwnPropertyDescriptors(u)):n(Object(u)).forEach(function(l){Object.defineProperty(i,l,Object.getOwnPropertyDescriptor(u,l))})}return i}e.exports=r,e.exports.__esModule=!0,e.exports.default=e.exports})(d6)),d6.exports}var v6={exports:{}},y6={exports:{}},b6={exports:{}},SQ;function MJ(){return SQ||(SQ=1,(function(e){function t(n,r){(r==null||r>n.length)&&(r=n.length);for(var i=0,a=Array(r);i<r;i++)a[i]=n[i];return a}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})(b6)),b6.exports}var _Q;function FJ(){return _Q||(_Q=1,(function(e){var t=MJ();function n(r,i){if(r){if(typeof r=="string")return t(r,i);var a={}.toString.call(r).slice(8,-1);return a==="Object"&&r.constructor&&(a=r.constructor.name),a==="Map"||a==="Set"?Array.from(r):a==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(a)?t(r,i):void 0}}e.exports=n,e.exports.__esModule=!0,e.exports.default=e.exports})(y6)),y6.exports}var EQ;function Cu(){return EQ||(EQ=1,(function(e){var t=FJ();function n(r,i){var a=typeof Symbol<"u"&&r[Symbol.iterator]||r["@@iterator"];if(!a){if(Array.isArray(r)||(a=t(r))||i&&r&&typeof r.length=="number"){a&&(r=a);var u=0,l=function(){};return{s:l,n:function(){return u>=r.length?{done:!0}:{done:!1,value:r[u++]}},e:function(h){throw h},f:l}}throw new TypeError(`Invalid attempt to iterate non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}var o,s=!0,c=!1;return{s:function(){a=a.call(r)},n:function(){var h=a.next();return s=h.done,h},e:function(h){c=!0,o=h},f:function(){try{s||a.return==null||a.return()}finally{if(c)throw o}}}}e.exports=n,e.exports.__esModule=!0,e.exports.default=e.exports})(v6)),v6.exports}var O6,DQ;function oS(){if(DQ)return O6;DQ=1;var e=$n().default,t=e(Ui()),n=e(Cu()),r=un();function i(l,o,s){var c=(0,n.default)(o),d;try{for(c.s();!(d=c.n()).done;){var h=d.value;l.relayFieldLogger((0,t.default)((0,t.default)({},h),{},{uiContext:s}))}}catch(S){c.e(S)}finally{c.f()}var m=(0,n.default)(o),y;try{for(m.s();!(y=m.n()).done;){var v=y.value;if(a(v))switch(v.kind){case"relay_resolver.error":throw new Error("Relay: Resolver error at path '".concat(v.fieldPath,"' in '").concat(v.owner,"'. Message: ").concat(v.error.message));case"relay_field_payload.error":throw new Error("Relay: Unexpected response payload - check server logs for details.");case"missing_expected_data.throw":throw new Error("Relay: Missing expected data at path '".concat(v.fieldPath,"' in '").concat(v.owner,"'."));case"missing_required_field.throw":throw new Error("Relay: Missing @required value at path '".concat(v.fieldPath,"' in '").concat(v.owner,"'."));case"missing_required_field.log":case"missing_expected_data.log":break;default:v.kind,r(!1)}}}catch(S){m.e(S)}finally{m.f()}}function a(l){switch(l.kind){case"relay_resolver.error":case"relay_field_payload.error":return l.shouldThrow&&!l.handled;case"missing_expected_data.throw":case"missing_required_field.throw":return!l.handled;case"missing_required_field.log":case"missing_expected_data.log":return!1;default:throw l.kind,new Error("Relay: Unexpected event kind")}}function u(l,o,s){o!=null&&i(l,o,s)}return O6={handlePotentialSnapshotErrors:u,eventShouldThrow:a},O6}var x6={exports:{}},S6={exports:{}},$Q;function q3e(){return $Q||($Q=1,(function(e){var t=MJ();function n(r){if(Array.isArray(r))return t(r)}e.exports=n,e.exports.__esModule=!0,e.exports.default=e.exports})(S6)),S6.exports}var _6={exports:{}},CQ;function V3e(){return CQ||(CQ=1,(function(e){function t(n){if(typeof Symbol<"u"&&n[Symbol.iterator]!=null||n["@@iterator"]!=null)return Array.from(n)}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})(_6)),_6.exports}var E6={exports:{}},wQ;function U3e(){return wQ||(wQ=1,(function(e){function t(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})(E6)),E6.exports}var RQ;function sy(){return RQ||(RQ=1,(function(e){var t=q3e(),n=V3e(),r=FJ(),i=U3e();function a(u){return t(u)||n(u)||r(u)||i()}e.exports=a,e.exports.__esModule=!0,e.exports.default=e.exports})(x6)),x6.exports}var D6,AQ;function NJ(){return AQ||(AQ=1,D6={DEFAULT_HANDLE_KEY:""}),D6}var $6,TQ;function qF(){if(TQ)return $6;TQ=1;var e=NJ(),t=e.DEFAULT_HANDLE_KEY,n=un();function r(i,a,u){return a&&a!==t?"__".concat(a,"_").concat(i):(u==null&&n(!1),"__".concat(u,"_").concat(i))}return $6=r,$6}var C6,kQ;function nh(){if(kQ)return C6;kQ=1;var e={ENABLE_VARIABLE_CONNECTION_KEY:!1,ENABLE_RELAY_RESOLVERS:!1,ENABLE_GETFRAGMENTIDENTIFIER_OPTIMIZATION:!1,ENABLE_FRIENDLY_QUERY_NAME_GQL_URL:!1,ENABLE_DO_NOT_WRAP_LIVE_QUERY:!1,ENABLE_NOTIFY_SUBSCRIPTION:!1,BATCH_ASYNC_MODULE_UPDATES_FN:null,ENABLE_CONTAINERS_SUBSCRIBE_ON_COMMIT:!1,MAX_DATA_ID_LENGTH:null,STRING_INTERN_LEVEL:0,LOG_MISSING_RECORDS_IN_PROD:!1,ENABLE_STORE_ID_COLLISION_LOGGING:!1,ENABLE_NONCOMPLIANT_ERROR_HANDLING_ON_LISTS:!1,ENABLE_LOOSE_SUBSCRIPTION_ATTRIBUTION:!1,ENABLE_OPERATION_TRACKER_OPTIMISTIC_UPDATES:!1,ENABLE_RELAY_OPERATION_TRACKER_SUSPENSE:!1,PROCESS_OPTIMISTIC_UPDATE_BEFORE_SUBSCRIPTION:!1,MARK_RESOLVER_VALUES_AS_CLEAN_AFTER_FRAGMENT_REREAD:!1,ENABLE_CYLE_DETECTION_IN_VARIABLES:!1,ENABLE_ACTIVITY_COMPATIBILITY:!1,ENABLE_READ_TIME_RESOLVER_STORAGE_KEY_PREFIX:!0,ENABLE_USE_PAGINATION_IS_LOADING_FIX:!1,DISALLOW_NESTED_UPDATES:!1,ENABLE_TYPENAME_PREFIXED_DATA_ID:!1,ENABLE_UI_CONTEXT_ON_RELAY_LOGGER:!1,CHECK_ALL_FRAGMENTS_FOR_MISSING_CLIENT_EDGES:!1};return C6=e,C6}var w6,PQ;function _1(){if(PQ)return w6;PQ=1;function e(n){if(!n||typeof n!="object")return n;if(Array.isArray(n))return n.map(e);for(var r=Object.keys(n).sort(),i={},a=0;a<r.length;a++)i[r[a]]=e(n[r[a]]);return i}function t(n){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:new Set;if(!n||typeof n!="object")return!1;if(r.has(n))return!0;var i=new Set(r);i.add(n);var a=Array.isArray(n)?n:Object.values(n);return a.some(function(u){return t(u,i)})}return w6={stableCopy:e,hasCycle:t},w6}var R6,MQ;function Da(){if(MQ)return R6;MQ=1;var e=$n().default,t=e(sy()),n=qF(),r=oy(),i=nh(),a=_1(),u=a.stableCopy,l=un(),o=r.VARIABLE,s=r.LITERAL,c=r.OBJECT_VALUE,d=r.LIST_VALUE,h="__errors",m="__module_component_",y="__module_operation_",v="$r:";function S(V,N){if(V.kind===o)return F(V.variableName,N);if(V.kind===s)return V.value;if(V.kind===c){var j={};return V.fields.forEach(function(K){j[K.name]=S(K,N)}),j}else if(V.kind===d){var W=[];return V.items.forEach(function(K){K!=null&&W.push(S(K,N))}),W}}function _(V,N,j){var W={};return j&&(W[q.FRAGMENT_POINTER_IS_WITHIN_UNMATCHED_TYPE_REFINEMENT]=!0),V&&V.forEach(function(K){W[K.name]=S(K,N)}),W}function E(V,N){var j=V.dynamicKey,W=V.handle,K=V.key,J=V.name,te=V.args,H=V.filters,M=n(W,K,J),G=null;return te&&H&&te.length!==0&&H.length!==0&&(G=te.filter(function(X){return H.indexOf(X.name)>-1})),j&&(G=G!=null?[j].concat((0,t.default)(G)):[j]),G===null?M:R(M,_(G,N))}function $(V,N){if(V.storageKey)return V.storageKey;var j=A(V),W=V.name;return j&&j.length!==0?R(W,_(j,N)):W}function k(V,N){var j=$(V,N);return i.ENABLE_READ_TIME_RESOLVER_STORAGE_KEY_PREFIX?"$r:"+j:j}function A(V){if(V.kind==="RelayResolver"||V.kind==="RelayLiveResolver"){var N;if(V.args==null){var j;return(j=V.fragment)===null||j===void 0?void 0:j.args}return((N=V.fragment)===null||N===void 0?void 0:N.args)==null?V.args:V.args.concat(V.fragment.args)}var W=typeof V.args>"u"?void 0:V.args;return W}function w(V,N){return R(V,u(N))}function R(V,N){if(!N)return V;var j=[];for(var W in N)if(N.hasOwnProperty(W)){var K=N[W];if(K!=null){var J;j.push(W+":"+((J=JSON.stringify(K))!==null&&J!==void 0?J:"undefined"))}}return j.length===0?V:V+"(".concat(j.join(","),")")}function F(V,N){return N.hasOwnProperty(V)||l(!1),u(N[V])}function I(V){return"".concat(m).concat(V)}function U(V){return"".concat(y).concat(V)}var q={ACTOR_IDENTIFIER_KEY:"__actorIdentifier",CLIENT_EDGE_TRAVERSAL_PATH:"__clientEdgeTraversalPath",FRAGMENTS_KEY:"__fragments",FRAGMENT_OWNER_KEY:"__fragmentOwner",FRAGMENT_POINTER_IS_WITHIN_UNMATCHED_TYPE_REFINEMENT:"$isWithinUnmatchedTypeRefinement",FRAGMENT_PROP_NAME_KEY:"__fragmentPropName",MODULE_COMPONENT_KEY:"__module_component",ERRORS_KEY:h,ID_KEY:"__id",REF_KEY:"__ref",REFS_KEY:"__refs",ROOT_ID:"client:root",ROOT_TYPE:"__Root",TYPENAME_KEY:"__typename",INVALIDATED_AT_KEY:"__invalidated_at",RELAY_RESOLVER_VALUE_KEY:"__resolverValue",RELAY_RESOLVER_INVALIDATION_KEY:"__resolverValueMayBeInvalid",RELAY_RESOLVER_SNAPSHOT_KEY:"__resolverSnapshot",RELAY_RESOLVER_ERROR_KEY:"__resolverError",RELAY_RESOLVER_OUTPUT_TYPE_RECORD_IDS:"__resolverOutputTypeRecordIDs",RELAY_READ_TIME_RESOLVER_KEY_PREFIX:v,formatStorageKey:R,getArgumentValue:S,getArgumentValues:_,getHandleStorageKey:E,getStorageKey:$,getReadTimeResolverStorageKey:k,getStableStorageKey:w,getModuleComponentKey:I,getModuleOperationKey:U};return R6=q,R6}var A6,FQ;function cy(){if(FQ)return A6;FQ=1;var e=$n().default,t=e(Ui()),n=Da(),r=n.getArgumentValues,i=un();function a(o,s,c){if(o.argumentDefinitions==null)return c;var d;return o.argumentDefinitions.forEach(function(h){if(!c.hasOwnProperty(h.name))switch(d=d||(0,t.default)({},c),h.kind){case"LocalArgument":d[h.name]=h.defaultValue;break;case"RootArgument":if(!s.hasOwnProperty(h.name)){d[h.name]=void 0;break}d[h.name]=s[h.name];break;default:i(!1)}}),d||c}function u(o,s,c){var d={};return o.argumentDefinitions.forEach(function(h){var m=h.defaultValue;c[h.name]!=null&&(m=c[h.name]),d[h.name]=m}),s!=null&&Object.keys(s).forEach(function(h){d[h]=s[h].get()}),d}function l(o,s,c){if(s==null)return o;var d=(0,t.default)({},o),h=c?r(c,o):{};return s.forEach(function(m){var y,v=(y=h[m.name])!==null&&y!==void 0?y:m.defaultValue;d[m.name]=v}),d}return A6={getLocalVariables:l,getFragmentVariables:a,getOperationVariables:u},A6}var T6,NQ;function ed(){if(NQ)return T6;NQ=1;var e=[],t=[];/**
 * Checks if two values are equal. Values may be primitives, arrays, or objects.
 * Returns true if both arguments have the same keys and values.
 *
 * @see http://underscorejs.org
 * @copyright 2009-2013 Jeremy Ashkenas, DocumentCloud Inc.
 * @license MIT
 */function n(i,a){var u=e.length?e.pop():[],l=t.length?t.pop():[],o=r(i,a,u,l);return u.length=0,l.length=0,e.push(u),t.push(l),o}function r(i,a,u,l){if(i===a)return i!==0||1/i==1/a;if(i==null||a==null||typeof i!="object"||typeof a!="object")return!1;var o=Object.prototype.toString,s=o.call(i);if(s!=o.call(a))return!1;switch(s){case"[object String]":return i==String(a);case"[object Number]":return isNaN(i)||isNaN(a)?!1:i==Number(a);case"[object Date]":case"[object Boolean]":return+i==+a;case"[object RegExp]":return i.source==a.source&&i.global==a.global&&i.multiline==a.multiline&&i.ignoreCase==a.ignoreCase}for(var c=u.length;c--;)if(u[c]==i)return l[c]==a;u.push(i),l.push(a);var d=0;if(s==="[object Array]"){if(d=i.length,d!==a.length)return!1;for(;d--;)if(!r(i[d],a[d],u,l))return!1}else{if(i.constructor!==a.constructor)return!1;if(i.hasOwnProperty("valueOf")&&a.hasOwnProperty("valueOf"))return i.valueOf()==a.valueOf();var h=Object.keys(i);if(h.length!=Object.keys(a).length)return!1;for(var m=0;m<h.length;m++)if(!r(i[h[m]],a[h[m]],u,l))return!1}return u.pop(),l.pop(),!0}return T6=n,T6}var k6,IQ;function hp(){if(IQ)return k6;IQ=1;var e=cy(),t=e.getFragmentVariables,n=Da(),r=n.CLIENT_EDGE_TRAVERSAL_PATH,i=n.FRAGMENT_OWNER_KEY,a=n.FRAGMENT_POINTER_IS_WITHIN_UNMATCHED_TYPE_REFINEMENT,u=n.FRAGMENTS_KEY,l=n.ID_KEY,o=ed(),s=un();Vi();function c(V,N){typeof N=="object"&&N!==null&&!Array.isArray(N)||s(!1);var j=N[l],W=N[u],K=N[i],J=N[r];if(typeof j=="string"&&typeof W=="object"&&W!==null&&typeof W[V.name]=="object"&&W[V.name]!==null&&typeof K=="object"&&K!==null&&(J==null||Array.isArray(J))){var te=K,H=J,M=W[V.name],G=t(V,te.variables,M),X=M[a]===!0;return U(V,j,G,te,X,H)}return null}function d(V,N){var j=null;return N.forEach(function(W,K){var J=W!=null?c(V,W):null;J!=null&&(j=j||[],j.push(J))}),j==null?null:{kind:"PluralReaderSelector",selectors:j}}function h(V,N){return N==null?N:V.metadata&&V.metadata.plural===!0?(Array.isArray(N)||s(!1),d(V,N)):(Array.isArray(N)&&s(!1),c(V,N))}function m(V,N){var j={};for(var W in V)if(V.hasOwnProperty(W)){var K=V[W],J=N[W];j[W]=h(K,J)}return j}function y(V,N){var j={};for(var W in V)if(V.hasOwnProperty(W)){var K=V[W],J=N[W];j[W]=v(K,J)}return j}function v(V,N){return N==null?N:V.metadata&&V.metadata.plural===!0?(Array.isArray(N)||s(!1),S(V,N)):(Array.isArray(N)&&s(!1),_(V,N))}function S(V,N){var j=null;return N.forEach(function(W){var K=W!=null?_(V,W):null;K!=null&&(j=j||[],j.push(K))}),j}function _(V,N){typeof N=="object"&&N!==null&&!Array.isArray(N)||s(!1);var j=N[l];return typeof j=="string"?j:null}function E(V,N){var j={};for(var W in V)if(V.hasOwnProperty(W)){var K=V[W],J=N[W],te=$(K,J);Object.assign(j,te)}return j}function $(V,N){var j;return N==null?{}:((j=V.metadata)===null||j===void 0?void 0:j.plural)===!0?(Array.isArray(N)||s(!1),A(V,N)):(Array.isArray(N)&&s(!1),k(V,N)||{})}function k(V,N){var j=c(V,N);return j?j.variables:null}function A(V,N){var j={};return N.forEach(function(W,K){if(W!=null){var J=k(V,W);J!=null&&Object.assign(j,J)}}),j}function w(V,N){return V.dataID===N.dataID&&V.node===N.node&&o(V.variables,N.variables)&&R(V.owner,N.owner)&&V.isWithinUnmatchedTypeRefinement===N.isWithinUnmatchedTypeRefinement&&F(V.clientEdgeTraversalPath,N.clientEdgeTraversalPath)}function R(V,N){return V===N?!0:V.identifier===N.identifier&&o(V.cacheConfig,N.cacheConfig)}function F(V,N){if(V===N)return!0;if(V==null||N==null||V.length!==N.length)return!1;for(var j=V.length;j--;){var W=V[j],K=N[j];if(W!==K&&(W==null||K==null||W.clientEdgeDestinationID!==K.clientEdgeDestinationID||W.readerClientEdge!==K.readerClientEdge))return!1}return!0}function I(V,N){return V===N?!0:V==null?N==null:N==null?V==null:V.kind==="SingularReaderSelector"&&N.kind==="SingularReaderSelector"?w(V,N):V.kind==="PluralReaderSelector"&&N.kind==="PluralReaderSelector"?V.selectors.length===N.selectors.length&&V.selectors.every(function(j,W){return w(j,N.selectors[W])}):!1}function U(V,N,j,W){var K=arguments.length>4&&arguments[4]!==void 0?arguments[4]:!1,J=arguments.length>5?arguments[5]:void 0;return{kind:"SingularReaderSelector",dataID:N,isWithinUnmatchedTypeRefinement:K,clientEdgeTraversalPath:J??null,node:V,variables:j,owner:W}}function q(V,N,j){return{dataID:N,node:V,variables:j}}return k6={areEqualSelectors:I,createReaderSelector:U,createNormalizationSelector:q,getDataIDsFromFragment:v,getDataIDsFromObject:y,getSingularSelector:c,getPluralSelector:d,getSelector:h,getSelectorsFromObject:m,getVariablesFromSingularFragment:k,getVariablesFromPluralFragment:A,getVariablesFromFragment:$,getVariablesFromObject:E},k6}var P6,LQ;function VF(){if(LQ)return P6;LQ=1;var e=dp(),t=e.getFragment,n=oS(),r=n.eventShouldThrow,i=hp(),a=i.getSelector,u=un(),l=[];function o(d,h){l.push(d);try{return h()}finally{l.pop()}}function s(d,h){if(!l.length)throw new Error("readFragment should be called only from within a Relay Resolver function.");var m=l[l.length-1],y=t(d),v=a(y,h);v==null&&u(!1),v.kind!=="SingularReaderSelector"&&u(!1);var S=m.getDataForResolverFragment(v,h),_=S.data,E=S.isMissingData,$=S.fieldErrors;if(E||$!=null&&$.some(r))throw c;return _}var c={};return P6={readFragment:s,withResolverContext:o,RESOLVER_FRAGMENT_ERRORED_SENTINEL:c},P6}var M6,BQ;function j3e(){if(BQ)return M6;BQ=1;var e=VF(),t=e.readFragment,n=un();function r(i,a,u,l){var o=a;return function(s,c,d){var h=t(i,s);if(u!=null){if(h==null)if(l===!0)n(!1);else return o(null,c,d);if(u in h)return l===!0&&h[u]==null&&n(!1),o(h[u],c,d);n(!1)}else return o(null,c,d)}}return M6=r,M6}var F6={exports:{}},QQ;function IJ(){return QQ||(QQ=1,(function(e){function t(r,i,a,u,l,o,s){try{var c=r[o](s),d=c.value}catch(h){return void a(h)}c.done?i(d):Promise.resolve(d).then(u,l)}function n(r){return function(){var i=this,a=arguments;return new Promise(function(u,l){var o=r.apply(i,a);function s(d){t(o,u,l,s,c,"next",d)}function c(d){t(o,u,l,s,c,"throw",d)}s(void 0)})}}e.exports=n,e.exports.__esModule=!0,e.exports.default=e.exports})(F6)),F6.exports}var N6,zQ;function LJ(){if(zQ)return N6;zQ=1;function e(t){return t!=null&&typeof t=="object"&&typeof t.then=="function"}return N6=e,N6}var I6,qQ;function fy(){if(qQ)return I6;qQ=1;var e=LJ(),t=o,n=(function(){s.create=function(h){return new s(h)};function s(d){this._source=d}s.onUnhandledError=function(h){t=h},s.from=function(h){return r(h)?i(h):e(h)?a(h):u(h)};var c=s.prototype;return c.catch=function(h){var m=this;return s.create(function(y){var v;return m.subscribe({start:function(_){v=_},next:y.next,complete:y.complete,error:function(_){try{h(_).subscribe({start:function($){v=$},next:y.next,complete:y.complete,error:y.error})}catch(E){y.error(E,!0)}}}),function(){return v.unsubscribe()}})},c.concat=function(h){var m=this;return s.create(function(y){var v;return m.subscribe({start:function(_){v=_},next:y.next,error:y.error,complete:function(){v=h.subscribe(y)}}),function(){v&&v.unsubscribe()}})},c.do=function(h){var m=this;return s.create(function(y){var v=function(_){return function(){try{h[_]&&h[_].apply(h,arguments)}catch(E){t(E,!0)}y[_]&&y[_].apply(y,arguments)}};return m.subscribe({start:v("start"),next:v("next"),error:v("error"),complete:v("complete"),unsubscribe:v("unsubscribe")})})},c.finally=function(h){var m=this;return s.create(function(y){var v=m.subscribe(y);return function(){v.unsubscribe(),h()}})},c.ifEmpty=function(h){var m=this;return s.create(function(y){var v=!1,S;return S=m.subscribe({next:function(E){v=!0,y.next(E)},error:y.error,complete:function(){v?y.complete():S=h.subscribe(y)}}),function(){S&&S.unsubscribe()}})},c.subscribe=function(h){return l(this._source,h)},c.map=function(h){var m=this;return s.create(function(y){var v=m.subscribe({complete:y.complete,error:y.error,next:function(_){try{var E=h(_);y.next(E)}catch($){y.error($,!0)}}});return function(){v.unsubscribe()}})},c.mergeMap=function(h){var m=this;return s.create(function(y){var v=[];function S(E){this._sub=E,v.push(E)}function _(){v.splice(v.indexOf(this._sub),1),v.length===0&&y.complete()}return m.subscribe({start:S,next:function($){try{y.closed||s.from(h($)).subscribe({start:S,next:y.next,error:y.error,complete:_})}catch(k){y.error(k,!0)}},error:y.error,complete:_}),function(){v.forEach(function(E){return E.unsubscribe()}),v.length=0}})},c.poll=function(h){var m=this;return s.create(function(y){var v,S,_=function E(){v=m.subscribe({next:y.next,error:y.error,complete:function(){S=setTimeout(E,h)}})};return _(),function(){clearTimeout(S),v.unsubscribe()}})},c.toPromise=function(){var h=this;return new Promise(function(m,y){var v=!1;h.subscribe({next:function(_){v||(v=!0,m(_))},error:y,complete:m})})},s})();function r(s){return typeof s=="object"&&s!==null&&typeof s.subscribe=="function"}function i(s){return s instanceof n?s:n.create(function(c){return s.subscribe(c)})}function a(s){return n.create(function(c){s.then(function(d){c.next(d),c.complete()},c.error)})}function u(s){return n.create(function(c){c.next(s),c.complete()})}function l(s,c){var d=!1,h,m=function(E){return Object.defineProperty(E,"closed",{get:function(){return d}})};function y(){if(h){if(h.unsubscribe)h.unsubscribe();else try{h()}catch(_){t(_,!0)}h=void 0}}var v=m({unsubscribe:function(){if(!d){d=!0;try{c.unsubscribe&&c.unsubscribe(v)}catch(E){t(E,!0)}finally{y()}}}});try{c.start&&c.start(v)}catch(_){t(_,!0)}if(d)return v;var S=m({next:function(E){if(!d&&c.next)try{c.next(E)}catch($){t($,!0)}},error:function(E,$){if(d||!c.error)d=!0,t(E,$||!1),y();else{d=!0;try{c.error(E)}catch(k){t(k,!0)}finally{y()}}},complete:function(){if(!d){d=!0;try{c.complete&&c.complete()}catch(E){t(E,!0)}finally{y()}}}});try{h=s(S)}catch(_){S.error(_,!0)}return d&&y(),v}function o(s,c){}return I6=n,I6}var L6,VQ;function BJ(){if(VQ)return L6;VQ=1;var e=$n().default,t=e(Cu()),n=e(mc()),r=fy(),i=un(),a=(function(){function u(){var o=this;(0,n.default)(this,"_complete",!1),(0,n.default)(this,"_events",[]),(0,n.default)(this,"_sinks",new Set),(0,n.default)(this,"_subscription",[]),this._observable=r.create(function(s){o._sinks.add(s);for(var c=o._events,d=0;d<c.length&&!s.closed;d++){var h=c[d];switch(h.kind){case"complete":s.complete();break;case"error":s.error(h.error);break;case"next":s.next(h.data);break;default:h.kind,i(!1)}}return function(){o._sinks.delete(s)}})}var l=u.prototype;return l.complete=function(){this._complete!==!0&&(this._complete=!0,this._events.push({kind:"complete"}),this._sinks.forEach(function(s){return s.complete()}))},l.error=function(s){this._complete!==!0&&(this._complete=!0,this._events.push({kind:"error",error:s}),this._sinks.forEach(function(c){return c.error(s)}))},l.next=function(s){this._complete!==!0&&(this._events.push({kind:"next",data:s}),this._sinks.forEach(function(c){return c.next(s)}))},l.subscribe=function(s){var c=this._observable.subscribe(s);return this._subscription.push(c),c},l.unsubscribe=function(){var s=(0,t.default)(this._subscription),c;try{for(s.s();!(c=s.n()).done;){var d=c.value;d.unsubscribe()}}catch(h){s.e(h)}finally{s.f()}this._subscription=[]},l.getObserverCount=function(){return this._sinks.size},u})();return L6=a,L6}var B6,UQ;function EC(){if(UQ)return B6;UQ=1;var e=fy(),t=BJ(),n=un(),r=typeof WeakMap=="function",i=r?new WeakMap:new Map;function a(m,y){return u(m,y.request.identifier,function(){return m.execute({operation:y})})}function u(m,y,v){return e.create(function(S){var _=d(m),E=_.get(y);return E||v().finally(function(){return _.delete(y)}).subscribe({start:function(k){E={identifier:y,subject:new t,subjectForInFlightStatus:new t,subscription:k,promise:null},_.set(y,E)},next:function(k){var A=h(_,y);A.subject.next(k),A.subjectForInFlightStatus.next(k)},error:function(k){var A=h(_,y);A.subject.error(k),A.subjectForInFlightStatus.error(k)},complete:function(){var k=h(_,y);k.subject.complete(),k.subjectForInFlightStatus.complete()},unsubscribe:function(k){var A=h(_,y);A.subject.unsubscribe(),A.subjectForInFlightStatus.unsubscribe()}}),E==null&&n(!1),l(_,E).subscribe(S)})}function l(m,y){return e.create(function(v){var S=y.subject.subscribe(v);return function(){S.unsubscribe();var _=m.get(y.identifier);if(_){var E=_.subscription;E!=null&&_.subject.getObserverCount()===0&&(E.unsubscribe(),m.delete(y.identifier))}}})}function o(m,y,v){return e.create(function(S){var _=v.subjectForInFlightStatus.subscribe({error:S.error,next:function($){if(!m.isRequestActive(v.identifier)){S.complete();return}S.next()},complete:S.complete,unsubscribe:S.complete});return function(){_.unsubscribe()}})}function s(m,y){var v=d(m),S=v.get(y.identifier);if(!S||!m.isRequestActive(S.identifier))return null;var _=new Promise(function(E,$){var k=!1;o(m,v,S).subscribe({complete:E,error:$,next:function(w){k&&E(w)}}),k=!0});return _}function c(m,y){var v=d(m),S=v.get(y.identifier);return!S||!m.isRequestActive(S.identifier)?null:o(m,v,S)}function d(m){var y=i.get(m);if(y!=null)return y;var v=new Map;return i.set(m,v),v}function h(m,y){var v=m.get(y);return v==null&&n(!1),v}return B6={fetchQuery:a,fetchQueryDeduped:u,getPromiseForActiveRequest:s,getObservableForActiveRequest:c},B6}var Q6,jQ;function UF(){if(jQ)return Q6;jQ=1;var e=$n().default,t=IJ().default,n=e(Cu()),r=fy(),i=EC(),a=i.getObservableForActiveRequest,u=dp(),l=u.getFragment,o=oS(),s=o.handlePotentialSnapshotErrors,c=hp(),d=c.getSelector,h=un();function m(k,A,w){return y.apply(this,arguments)}function y(){return y=t(function*(k,A,w){var R;try{var F,I=yield new Promise(function(q,V){R=v(k,A,w).subscribe({next:function(j){j.state==="ok"?q(j.value):j.state==="error"&&V(j.error)}})});return(F=R)===null||F===void 0||F.unsubscribe(),I}catch(q){var U;throw(U=R)===null||U===void 0||U.unsubscribe(),q}}),y.apply(this,arguments)}function v(k,A,w){var R,F=l(A),I=d(F,w);switch(((R=F.metadata)===null||R===void 0?void 0:R.hasClientEdges)!=null&&h(!1),I==null&&h(!1),I.kind){case"SingularReaderSelector":return S(k,A,I);case"PluralReaderSelector":return _(k,A,I)}h(!1)}function S(k,A,w){var R=k.lookup(w);return r.create(function(F){F.next(E(k,A,w.owner,R));var I=k.subscribe(R,function(U){F.next(E(k,A,w.owner,U))});return function(){return I.dispose()}})}function _(k,A,w){var R=w.selectors.map(function(F){return k.lookup(F)});return r.create(function(F){var I=R.map(function(q,V){return E(k,A,w.selectors[V].owner,q)});F.next($(I));var U=R.map(function(q,V){return k.subscribe(q,function(N){I[V]=E(k,A,w.selectors[V].owner,N),F.next($(I))})});return function(){return U.forEach(function(q){return q.dispose()})}})}function E(k,A,w,R){var F=R.missingLiveResolverFields!=null&&R.missingLiveResolverFields.length>0,I=R.missingClientEdges!=null&&R.missingClientEdges.length>0;if(F||I)return{state:"loading"};if(R.isMissingData&&(a(k,w)!=null||k.getOperationTracker().getPendingOperationsAffectingOwner(w)!=null))return{state:"loading"};try{s(k,R.fieldErrors)}catch(U){return{error:U,state:"error"}}return R.data==null&&h(!1),{state:"ok",value:R.data}}function $(k){var A=[],w=(0,n.default)(k),R;try{for(w.s();!(R=w.n()).done;){var F=R.value;if(F.state==="ok")A.push(F.value);else return F}}catch(I){w.e(I)}finally{w.f()}return{state:"ok",value:A}}return Q6={observeFragment:v,waitForFragmentData:m},Q6}var z6,ZQ;function sS(){if(ZQ)return z6;ZQ=1;function e(n){return t(n)&&(Object.freeze(n),Object.getOwnPropertyNames(n).forEach(function(r){var i=n[r];i&&typeof i=="object"&&!Object.isFrozen(i)&&e(i)})),n}function t(n){return n!=null&&(Array.isArray(n)||typeof n=="object"&&n.constructor===Object)}return z6=e,z6}var q6,HQ;function QJ(){if(HQ)return q6;HQ=1;var e=_1(),t=e.stableCopy,n=un();function r(i,a){var u=i.cacheID!=null?i.cacheID:i.id;return u==null&&n(!1),u+JSON.stringify(t(a))}return q6=r,q6}var V6,XQ;function dy(){if(XQ)return V6;XQ=1,sS();var e=QJ(),t=nh(),n=_1(),r=n.hasCycle,i=cy(),a=i.getOperationVariables,u=hp(),l=u.createNormalizationSelector,o=u.createReaderSelector,s=Da(),c=s.ROOT_ID,d=un();function h(y,v,S){var _=arguments.length>3&&arguments[3]!==void 0?arguments[3]:c,E=y.operation,$=a(E,y.params.providedVariables,v);t.ENABLE_CYLE_DETECTION_IN_VARIABLES&&r($)&&d(!1);var k=m(y,$,S),A={fragment:o(y.fragment,_,$,k),request:k,root:l(E,_,$)};return A}function m(y,v,S){var _={identifier:e(y.params,v),node:y,variables:v,cacheConfig:S};return _}return V6={createOperationDescriptor:h,createRequestDescriptor:m},V6}var U6,YQ;function Z3e(){if(YQ)return U6;YQ=1;var e=$n().default,t=e(mc()),n=UF(),r=n.observeFragment,i=dy(),a=i.createOperationDescriptor;function u(l,o,s){var c=a(o,s),d={__id:c.fragment.dataID,__fragments:(0,t.default)({},c.fragment.node.name,c.request.variables),__fragmentOwner:c.request},h=c.request.node.fragment;return r(l,h,d)}return U6={observeQuery:u},U6}var j6,GQ;function H3e(){if(GQ)return j6;GQ=1;var e=IJ().default,t=UF(),n=t.observeFragment;function r(a,u,l){return i.apply(this,arguments)}function i(){return i=e(function*(a,u,l){var o;try{var s,c=yield new Promise(function(h,m){o=n(a,u,l).subscribe({next:function(v){v.state==="ok"?h(v.value):v.state==="error"&&m(v.error)}})});return(s=o)===null||s===void 0||s.unsubscribe(),c}catch(h){var d;throw(d=o)===null||d===void 0||d.unsubscribe(),h}}),i.apply(this,arguments)}return j6={waitForFragmentData:r},j6}var Z6,KQ;function X3e(){if(KQ)return Z6;KQ=1;var e=j3e(),t=UF(),n=t.observeFragment,r=Z3e(),i=r.observeQuery,a=H3e(),u=a.waitForFragmentData;function l(s){return s.ok===!0}function o(s){return s.ok===!1}return Z6={resolverDataInjector:e,isValueResult:l,isErrorResult:o,observeQuery:i,observeFragment:n,waitForFragmentData:u},Z6}var H6,WQ;function zJ(){if(WQ)return H6;WQ=1;var e=new Map,t=1,n=a(),r="	",i="\v";function a(){for(var o=new Set,s=0;s<10;++s)o.add(s.toString());return o}function u(o){return o[0]===r&&n.has(o[1])||o[0]===i?i+o:o}function l(o,s){if(s==null||o.length<s)return u(o);var c=e.get(o);return c!=null||(c=r+t++,e.set(o,c)),c}return H6={intern:l},H6}var X6,JQ;function N0(){if(JQ)return X6;JQ=1;var e=nh(),t=zJ(),n=t.intern,r="client:";function i(s,c,d){var h=e.STRING_INTERN_LEVEL<=0?s:n(s,e.MAX_DATA_ID_LENGTH),m=h+":"+c;return d!=null&&(m+=":"+d),m.indexOf(r)!==0&&(m=r+m),m}function a(s){return s.indexOf(r)===0}var u=0;function l(){return"".concat(r,"local:").concat(u++)}function o(s,c,d){var h="".concat(r).concat(s,":").concat(c);return d!=null&&(h+=":"+d),h}return X6={generateClientID:i,generateClientObjectClientID:o,generateUniqueClientID:l,isClientID:a},X6}var Y6,ez;function jF(){if(ez)return Y6;ez=1;var e={after:!0,before:!0,find:!0,first:!0,last:!0,surrounds:!0},t={CURSOR:"cursor",EDGES:"edges",END_CURSOR:"endCursor",HAS_NEXT_PAGE:"hasNextPage",HAS_PREV_PAGE:"hasPreviousPage",NODE:"node",PAGE_INFO_TYPE:"PageInfo",PAGE_INFO:"pageInfo",START_CURSOR:"startCursor"},n={inject:function(i){t=i},get:function(){return t},isConnectionCall:function(i){return e.hasOwnProperty(i.name)}};return Y6=n,Y6}var G6,tz;function DC(){if(tz)return G6;tz=1;var e=N0(),t=e.generateClientID,n=Da(),r=n.getStableStorageKey,i=qF(),a=jF(),u=un();Vi();var l="connection",o="__connection_next_edge_index";function s(E,$){var k=E.get($.dataID);if(k){var A=a.get(),w=A.EDGES,R=A.END_CURSOR,F=A.HAS_NEXT_PAGE,I=A.HAS_PREV_PAGE,U=A.PAGE_INFO,q=A.PAGE_INFO_TYPE,V=A.START_CURSOR,N=k.getLinkedRecord($.fieldKey),j=N&&N.getLinkedRecord(U);if(!N){k.setValue(null,$.handleKey,void 0,k.getErrors($.fieldKey));return}var W=t(k.getDataID(),$.handleKey),K=k.getLinkedRecord($.handleKey),J=K??E.get(W),te=J&&J.getLinkedRecord(U);if(J){K==null&&k.setLinkedRecord(J,$.handleKey);var G=J,X=N.getLinkedRecords(w);X&&(X=X.map(function(Me){return S(E,G,Me)}));var B=G.getLinkedRecords(w),ne=G.getLinkedRecord(U);G.copyFieldsFrom(N),B&&G.setLinkedRecords(B,w),ne&&G.setLinkedRecord(ne,U);var oe=[],ie=$.args;if(B&&X)if(ie.after!=null){var de,ge=(de=te)===null||de===void 0?void 0:de.getValue(R),ye=j==null?void 0:j.getValue(R),me=te&&ie.after===ge,Ee=te&&ge===ye;if(me||Ee){var se=new Set;_(B,oe,se),_(X,oe,se)}else return}else if(ie.before!=null)if(te&&ie.before===te.getValue(V)){var pe=new Set;_(X,oe,pe),_(B,oe,pe)}else return;else oe=X;else X?oe=X:oe=B;if(oe!=null&&oe!==B&&G.setLinkedRecords(oe,w),te&&j){if(ie.after==null&&ie.before==null)te.copyFieldsFrom(j);else if(ie.before!=null||ie.after==null&&ie.last){te.setValue(!!j.getValue(I),I);var ve=j.getValue(V);typeof ve=="string"&&te.setValue(ve,V)}else if(ie.after!=null||ie.before==null&&ie.first){te.setValue(!!j.getValue(F),F);var we=j.getValue(R);typeof we=="string"&&te.setValue(we,R)}}}else{var H=E.create(W,N.getType());H.setValue(0,o),H.copyFieldsFrom(N);var M=N.getLinkedRecords(w);M&&(M=M.map(function(Me){return S(E,H,Me)}),H.setLinkedRecords(M,w)),k.setLinkedRecord(H,$.handleKey),te=E.create(t(H.getDataID(),U),q),te.setValue(!1,F),te.setValue(!1,I),te.setValue(null,R),te.setValue(null,V),j&&te.copyFieldsFrom(j),H.setLinkedRecord(te,U)}}}function c(E,$,k){var A=i(l,$,null);return E.getLinkedRecord(A,k)}function d(E,$,k){var A=i(l,$,null),w=r(A,k);return t(E,w)}function h(E,$,k){var A=a.get(),w=A.CURSOR,R=A.EDGES,F=E.getLinkedRecords(R);if(!F){E.setLinkedRecords([$],R);return}var I;if(k==null)I=F.concat($);else{I=[];for(var U=!1,q=0;q<F.length;q++){var V=F[q];if(I.push(V),V!=null){var N=V.getValue(w);k===N&&(I.push($),U=!0)}}U||I.push($)}E.setLinkedRecords(I,R)}function m(E,$,k,A){var w=a.get(),R=w.NODE,F=t($.getDataID(),k.getDataID()),I=E.get(F);return I||(I=E.create(F,A)),I.setLinkedRecord(k,R),I.getValue("cursor")==null&&I.setValue(null,"cursor"),I}function y(E,$,k){var A=a.get(),w=A.CURSOR,R=A.EDGES,F=E.getLinkedRecords(R);if(!F){E.setLinkedRecords([$],R);return}var I;if(k==null)I=[$].concat(F);else{I=[];for(var U=!1,q=0;q<F.length;q++){var V=F[q];if(V!=null){var N=V.getValue(w);k===N&&(I.push($),U=!0)}I.push(V)}U||I.unshift($)}E.setLinkedRecords(I,R)}function v(E,$){var k=a.get(),A=k.EDGES,w=k.NODE,R=E.getLinkedRecords(A);if(R){for(var F,I=0;I<R.length;I++){var U=R[I],q=U&&U.getLinkedRecord(w);q!=null&&q.getDataID()===$?F===void 0&&(F=R.slice(0,I)):F!==void 0&&F.push(U)}F!==void 0&&E.setLinkedRecords(F,A)}}function S(E,$,k){if(k==null)return k;var A=a.get(),w=A.EDGES,R=$.getValue(o);typeof R!="number"&&u(!1);var F=t($.getDataID(),w,R),I=E.create(F,k.getType());return I.copyFieldsFrom(k),I.getValue("cursor")==null&&I.setValue(null,"cursor"),$.setValue(R+1,o),I}function _(E,$,k){for(var A=a.get(),w=A.NODE,R=0;R<E.length;R++){var F=E[R];if(F){var I=F.getLinkedRecord(w),U=I&&I.getDataID();if(U){if(k.has(U))continue;k.add(U)}$.push(F)}}}return G6={buildConnectionEdge:S,createEdge:m,deleteNode:v,getConnection:c,getConnectionID:d,insertEdgeAfter:h,insertEdgeBefore:y,update:s},G6}var K6,nz;function qJ(){if(nz)return K6;nz=1;var e=$n().default,t=e(Cu()),n=DC(),r=jF(),i=un(),a=Vi(),u={update:function(v,S){var _=v.get(S.dataID);if(_!=null){var E=_.getValue(S.fieldKey);typeof E=="string"?v.delete(E):Array.isArray(E)&&E.forEach(function($){typeof $=="string"&&v.delete($)})}}},l={update:function(v,S){var _=v.get(S.dataID);if(_!=null){var E=S.handleArgs.connections;E==null&&i(!1);var $=_.getValue(S.fieldKey),k=Array.isArray($)?$:[$];k.forEach(function(A){if(typeof A=="string"){var w=(0,t.default)(E),R;try{for(w.s();!(R=w.n()).done;){var F=R.value,I=v.get(F);I!=null&&n.deleteNode(I,A)}}catch(U){w.e(U)}finally{w.f()}}})}}},o={update:h(n.insertEdgeAfter)},s={update:h(n.insertEdgeBefore)},c={update:m(n.insertEdgeAfter)},d={update:m(n.insertEdgeBefore)};function h(y){return function(v,S){var _,E=v.get(S.dataID);if(E!=null){var $=S.handleArgs.connections;$==null&&i(!1);var k,A;try{k=E.getLinkedRecord(S.fieldKey)}catch{}if(!k)try{A=E.getLinkedRecords(S.fieldKey)}catch{}if(!(k==null&&A==null)){var w=r.get(),R=w.NODE,F=w.EDGES,I=(_=A)!==null&&_!==void 0?_:[k],U=(0,t.default)(I),q;try{var V=function(){var K=q.value;if(K==null)return"continue";var J=K.getLinkedRecord("node");if(!J)return"continue";var te=J.getDataID(),H=(0,t.default)($),M;try{for(H.s();!(M=H.n()).done;){var G=M.value,X=v.get(G);if(X!=null){var B=(N=X.getLinkedRecords(F))===null||N===void 0?void 0:N.some(function(oe){var ie;return(oe==null||(ie=oe.getLinkedRecord(R))===null||ie===void 0?void 0:ie.getDataID())===te});if(!B){var ne=n.buildConnectionEdge(v,X,K);ne==null&&i(!1),y(X,ne)}}}}catch(oe){H.e(oe)}finally{H.f()}};for(U.s();!(q=U.n()).done;)var N,j=V()}catch(W){U.e(W)}finally{U.f()}}}}}function m(y){return function(v,S){var _,E=v.get(S.dataID);if(E!=null){var $=S.handleArgs,k=$.connections,A=$.edgeTypeName;k==null&&i(!1),A==null&&i(!1);var w,R;try{w=E.getLinkedRecord(S.fieldKey)}catch{}if(!w)try{R=E.getLinkedRecords(S.fieldKey)}catch{}if(!(w==null&&R==null)){var F=r.get(),I=F.NODE,U=F.EDGES,q=(_=R)!==null&&_!==void 0?_:[w],V=(0,t.default)(q),N;try{var j=function(){var te=N.value;if(te==null)return"continue";var H=te.getDataID(),M=(0,t.default)(k),G;try{for(M.s();!(G=M.n()).done;){var X=G.value,B=v.get(X);if(B!=null){var ne=(W=B.getLinkedRecords(U))===null||W===void 0?void 0:W.some(function(ie){var de;return(ie==null||(de=ie.getLinkedRecord(I))===null||de===void 0?void 0:de.getDataID())===H});if(!ne){var oe=n.createEdge(v,B,te,A);oe==null&&i(!1),y(B,oe)}}}}catch(ie){M.e(ie)}finally{M.f()}};for(V.s();!(N=V.n()).done;)var W,K=j()}catch(J){V.e(J)}finally{V.f()}}}}}return K6={AppendEdgeHandler:o,DeleteRecordHandler:u,PrependEdgeHandler:s,AppendNodeHandler:c,PrependNodeHandler:d,DeleteEdgeHandler:l},K6}var W6,rz;function VJ(){if(rz)return W6;rz=1;var e=DC(),t=qJ(),n=un();function r(i){switch(i){case"connection":return e;case"deleteRecord":return t.DeleteRecordHandler;case"deleteEdge":return t.DeleteEdgeHandler;case"appendEdge":return t.AppendEdgeHandler;case"prependEdge":return t.PrependEdgeHandler;case"appendNode":return t.AppendNodeHandler;case"prependNode":return t.PrependNodeHandler}n(!1)}return W6=r,W6}var J6,iz;function ZF(){if(iz)return J6;iz=1;function e(t){return!!(t&&t["@@RelayModernEnvironment"])}return J6=e,J6}var eR,az;function $C(){if(az)return eR;az=1;var e=$n().default,t=e(Cu()),n=DC(),r=Vi(),i=Object.freeze({RANGE_ADD:"RANGE_ADD",RANGE_DELETE:"RANGE_DELETE",NODE_DELETE:"NODE_DELETE"}),a=Object.freeze({APPEND:"append",PREPEND:"prepend"});function u(h,m,y,v){var S=y?[y]:[],_=v?[v]:[];return h.forEach(function(E){switch(E.type){case"NODE_DELETE":var $=l(E,m);$&&(S.push($),_.push($));break;case"RANGE_ADD":var k=o(E,m);k&&(S.push(k),_.push(k));break;case"RANGE_DELETE":var A=s(E,m);A&&(S.push(A),_.push(A));break}}),{optimisticUpdater:function($,k){S.forEach(function(A){A($,k)})},updater:function($,k){_.forEach(function(A){A($,k)})}}}function l(h,m){var y=h.deletedIDFieldName,v=d(m);return v?function(S,_){var E=S.getRootField(v);if(E){var $=E.getValue(y),k=Array.isArray($)?$:[$];k.forEach(function(A){A&&typeof A=="string"&&S.delete(A)})}}:null}function o(h,m){var y=h.parentID,v=h.connectionInfo,S=h.edgeName;if(!y)return null;var _=d(m);return!v||!_?null:function(E,$){var k=E.get(y);if(k){var A=E.getRootField(_);if(A){var w=A.getLinkedRecord(S),R=(0,t.default)(v),F;try{for(R.s();!(F=R.n()).done;){var I=F.value;if(w){var U=n.getConnection(k,I.key,I.filters);if(U){var q=n.buildConnectionEdge(E,U,w);if(q)switch(I.rangeBehavior){case"append":n.insertEdgeAfter(U,q);break;case"prepend":n.insertEdgeBefore(U,q);break;default:break}}}}}catch(V){R.e(V)}finally{R.f()}}}}}function s(h,m){var y=h.parentID,v=h.connectionKeys,S=h.pathToConnection,_=h.deletedIDFieldName;if(!y)return null;var E=d(m);return E?function($,k){if(k){var A=[],w=k[E];if(w&&Array.isArray(_)){var R=(0,t.default)(_),F;try{for(R.s();!(F=R.n()).done;){var I=F.value;w&&typeof w=="object"&&(w=w[I])}}catch(U){R.e(U)}finally{R.f()}Array.isArray(w)?w.forEach(function(U){U&&U.id&&typeof U=="object"&&typeof U.id=="string"&&A.push(U.id)}):w&&w.id&&typeof w.id=="string"&&A.push(w.id)}else w&&typeof _=="string"&&typeof w=="object"&&(w=w[_],typeof w=="string"?A.push(w):Array.isArray(w)&&w.forEach(function(U){typeof U=="string"&&A.push(U)}));c(y,v,S,$,A)}}:null}function c(h,m,y,v,S){var _=v.get(h);if(_&&!(y.length<2)){for(var E=_,$=1;$<y.length-1;$++)E&&(E=E.getLinkedRecord(y[$]));if(!(!m||!E)){var k=(0,t.default)(m),A;try{var w=function(){var F=A.value,I=n.getConnection(E,F.key,F.filters);I&&S.forEach(function(U){n.deleteNode(I,U)})};for(k.s();!(A=k.n()).done;)w()}catch(R){k.e(R)}finally{k.f()}}}}function d(h){return h.fragment.selections&&h.fragment.selections.length>0&&h.fragment.selections[0].kind==="LinkedField"?h.fragment.selections[0].name:null}return eR={MutationTypes:i,RangeOperations:a,convert:u},eR}var tR,uz;function Y3e(){if(uz)return tR;uz=1;var e=dp(),t=e.getRequest,n=ZF(),r=dy(),i=r.createOperationDescriptor,a=$C(),u=un();function l(o,s){n(o)||u(!1);var c=t(s.mutation);if(c.params.operationKind!=="mutation")throw new Error("commitMutation: Expected mutation operation");var d=s.optimisticUpdater,h=s.configs,m=s.optimisticResponse,y=s.variables,v=i(c,y);if(h){var S=a.convert(h,c,d);d=S.optimisticUpdater}return o.applyMutation({operation:v,response:m,updater:d})}return tR=l,tR}var nR,lz;function G3e(){if(lz)return nR;lz=1;function e(t,n){t.commitUpdate(n)}return nR=e,nR}var rR,oz;function K3e(){if(oz)return rR;oz=1;var e=$n().default;e(Ui()),Vi();var t=function(){};return rR=t,rR}var iR,sz;function W3e(){if(sz)return iR;sz=1;var e=$n().default,t=e(sy()),n=dp(),r=n.getRequest,i=N0(),a=i.generateUniqueClientID,u=ZF(),l=dy(),o=l.createOperationDescriptor,s=$C();K3e();var c=un();Vi();function d(h,m){u(h)||c(!1);var y=r(m.mutation);if(y.params.operationKind!=="mutation")throw new Error("commitMutation: Expected mutation operation");if(y.kind!=="Request")throw new Error("commitMutation: Expected mutation to be of type request");var v=m.optimisticResponse,S=m.optimisticUpdater,_=m.updater,E=m.configs,$=m.cacheConfig,k=m.onError,A=m.onUnsubscribe,w=m.variables,R=m.uploadables,F=o(y,w,$,a());if(typeof v=="function"&&(v=v()),E){var I=s.convert(E,y,S,_);S=I.optimisticUpdater,_=I.updater}var U=[],q=h.executeMutation({operation:F,optimisticResponse:v,optimisticUpdater:S,updater:_,uploadables:R}).subscribe({next:function(N){var j;Array.isArray(N)?N.forEach(function(W){W.errors&&U.push.apply(U,(0,t.default)(W.errors))}):N.errors&&U.push.apply(U,(0,t.default)(N.errors)),(j=m.onNext)===null||j===void 0||j.call(m)},complete:function(){var N=m.onCompleted;if(N){var j=h.lookup(F.fragment);N(j.data,U.length!==0?U:null)}},error:k,unsubscribe:A});return{dispose:q.unsubscribe}}return iR=d,iR}var aR,cz;function UJ(){if(cz)return aR;cz=1,ed(),Vi();var e=typeof WeakMap=="function",t=e?new WeakMap:new Map;function n(r,i){if(i!=null){var a={};return Object.assign(a,r),Object.keys(i).forEach(function(u){var l=i[u].get,o=l();if(!t.has(l))t.set(l,o),a[u]=o;else{var s=t.get(l);a[u]=s}}),a}else return r}return n.tests_only_resetDebugCache=void 0,aR=n,aR}var uR,fz;function J3e(){if(fz)return uR;fz=1;var e=fy();function t(n){return function(i,a,u,l,o){var s=n(i,a,u,l,o);return s instanceof Error?e.create(function(c){return c.error(s)}):e.from(s)}}return uR={convertFetch:t},uR}var lR,dz;function e4e(){if(dz)return lR;dz=1;var e=UJ(),t=J3e(),n=t.convertFetch,r=un();function i(a,u){var l=n(a);function o(s,c,d,h,m){var y=e(c,s.providedVariables);if(s.operationKind==="subscription")return u||r(!1),h&&r(!1),u(s,y,d);var v=d.poll;return v!=null?(h&&r(!1),l(s,y,{force:!0}).poll(v)):l(s,y,d,h,m)}return{execute:o}}return lR={create:i},lR}var oR,hz;function t4e(){if(hz)return oR;hz=1;var e=$n().default,t=e(Ui()),n=_1(),r=n.stableCopy,i=un(),a=(function(){function o(c){var d=c.size,h=c.ttl;d>0||i(!1),h>0||i(!1),this._responses=new Map,this._size=d,this._ttl=h}var s=o.prototype;return s.clear=function(){this._responses.clear()},s.get=function(d,h){var m=this,y=u(d,h);this._responses.forEach(function(S,_){l(S.fetchTime,m._ttl)||m._responses.delete(_)});var v=this._responses.get(y);return v==null?null:Array.isArray(v.payload)?v.payload.map(function(S){return(0,t.default)((0,t.default)({},S),{},{extensions:(0,t.default)((0,t.default)({},S.extensions),{},{cacheTimestamp:v.fetchTime})})}):(0,t.default)((0,t.default)({},v.payload),{},{extensions:(0,t.default)((0,t.default)({},v.payload.extensions),{},{cacheTimestamp:v.fetchTime})})},s.set=function(d,h,m){var y=Date.now(),v=u(d,h);if(this._responses.delete(v),this._responses.set(v,{fetchTime:y,payload:m}),this._responses.size>this._size){var S=this._responses.keys().next();S.done||this._responses.delete(S.value)}},o})();function u(o,s){return JSON.stringify(r({queryID:o,variables:s}))}function l(o,s){return o+s>=Date.now()}return oR=a,oR}var sR,pz;function n4e(){if(pz)return sR;pz=1;var e=$n().default,t=e(Ui()),n=fy(),r=dy(),i=r.createOperationDescriptor,a=oS(),u=a.handlePotentialSnapshotErrors,l=EC(),o=dp(),s=o.getRequest,c=un();function d(m,y,v,S){var _,E=s(y);E.params.operationKind!=="query"&&c(!1);var $=(0,t.default)({force:!0},S==null?void 0:S.networkCacheConfig),k=i(E,v,$),A=(_=S==null?void 0:S.fetchPolicy)!==null&&_!==void 0?_:"network-only";function w(R){return u(m,R.fieldErrors),R.data}switch(A){case"network-only":return h(m,k).map(w);case"store-or-network":return m.check(k).status==="available"?n.from(m.lookup(k.fragment)).map(w):h(m,k).map(w);default:throw new Error("fetchQuery: Invalid fetchPolicy "+A)}}function h(m,y){return l.fetchQuery(m,y).map(function(){return m.lookup(y.fragment)})}return sR=d,sR}var cR,mz;function r4e(){if(mz)return cR;mz=1;var e=dy(),t=e.createOperationDescriptor,n=dp(),r=n.getRequest;function i(a,u,l,o){var s=r(u);if(s.params.operationKind!=="query")throw new Error("fetchQuery: Expected query operation");var c=t(s,l,o);return a.execute({operation:c}).map(function(){return a.lookup(c.fragment).data}).toPromise()}return cR=i,cR}var fR,gz;function i4e(){if(gz)return fR;gz=1;var e=(function(){function n(){this._preloadableQueries=new Map,this._callbacks=new Map}var r=n.prototype;return r.set=function(a,u){this._preloadableQueries.set(a,u);var l=this._callbacks.get(a);l!=null&&l.forEach(function(o){try{o(u)}catch(s){setTimeout(function(){throw s},0)}})},r.get=function(a){return this._preloadableQueries.get(a)},r.onLoad=function(a,u){var l,o=(l=this._callbacks.get(a))!==null&&l!==void 0?l:new Set;o.add(u);var s=function(){o.delete(u)};return this._callbacks.set(a,o),{dispose:s}},r.clear=function(){this._preloadableQueries.clear()},n})(),t=new e;return fR=t,fR}var dR,vz;function jJ(){if(vz)return dR;vz=1;var e=EC(),t=e.getPromiseForActiveRequest;function n(r,i,a){var u,l,o=[],s=t(r,a);if(s!=null)o=[a];else{var c,d,h=r.getOperationTracker(),m=h.getPendingOperationsAffectingOwner(a);o=(c=m==null?void 0:m.pendingOperations)!==null&&c!==void 0?c:[],s=(d=m==null?void 0:m.promise)!==null&&d!==void 0?d:null}if(!s)return null;var y=(u=(l=o)===null||l===void 0?void 0:l.map(function(_){return _.node.params.name}).join(","))!==null&&u!==void 0?u:null;(y==null||y.length===0)&&(y="Unknown pending operation");var v=i.name,S=y===v?"Relay(".concat(y,")"):"Relay(".concat(y,":").concat(v,")");return s.displayName=S,r.__log({name:"pendingoperation.found",fragment:i,fragmentOwner:a,pendingOperations:o}),{promise:s,pendingOperations:o}}return dR=n,dR}var hR,yz;function ZJ(){if(yz)return hR;yz=1;function e(t,n){return t===n&&(t===null||typeof t!="object")}return hR=e,hR}var pR,bz;function CC(){if(bz)return pR;bz=1;function e(n,r){return t(n,r,!0)}function t(n,r,i){if(n===r||typeof n!="object"||!n||n.constructor!==Object&&!Array.isArray(n)||typeof r!="object"||!r||r.constructor!==Object&&!Array.isArray(r))return r;var a=!1,u=Array.isArray(n)?n:null,l=Array.isArray(r)?r:null;if(u&&l){var o=i&&!Object.isFrozen(l);a=l.reduce(function(y,v,S){var _=u[S],E=t(_,v,o);return E!==l[S]&&o&&(l[S]=E),y&&E===u[S]},!0)&&u.length===l.length}else if(!u&&!l){var s=n,c=r,d=Object.keys(s),h=Object.keys(c),m=i&&!Object.isFrozen(c);a=h.reduce(function(y,v){var S=s[v],_=t(S,c[v],m);return _!==c[v]&&m&&(c[v]=_),y&&_===s[v]},!0)&&d.length===h.length}return a?n:r}return pR=e,pR}var mR,Oz;function a4e(){if(Oz)return mR;Oz=1;var e=$n().default,t=e(Ui()),n=e(mc()),r=jJ(),i=oS(),a=i.handlePotentialSnapshotErrors,u=ZJ(),l=CC(),o=nh(),s=dy(),c=s.createRequestDescriptor,d=hp(),h=d.areEqualSelectors,m=d.createReaderSelector,y=d.getSelectorsFromObject,v=ed(),S=un();Vi();var _=(function(){function A(R,F,I,U,q){var V=this;(0,n.default)(this,"_onChange",function(){V._stale=!0,typeof V._callback=="function"&&V._callback()}),this._callback=U,this._context=R,this._data={},this._fragments=F,this._props={},this._resolvers={},this._stale=!1,this._rootIsQueryRenderer=q,this.setProps(I)}var w=A.prototype;return w.dispose=function(){for(var F in this._resolvers)this._resolvers.hasOwnProperty(F)&&k(this._resolvers[F])},w.resolve=function(){if(this._stale){var F=this._data,I;for(var U in this._resolvers)if(this._resolvers.hasOwnProperty(U)){var q=this._resolvers[U],V=F[U];if(q){var N=q.resolve();(I||N!==V)&&(I=I||(0,t.default)({},F),I[U]=N)}else{var j=this._props[U],W=j!==void 0?j:null;(I||!u(W,V))&&(I=I||(0,t.default)({},F),I[U]=W)}}this._data=I||F,this._stale=!1}return this._data},w.setCallback=function(F,I){this._callback=I,o.ENABLE_CONTAINERS_SUBSCRIBE_ON_COMMIT===!0&&this.setProps(F)},w.setProps=function(F){this._props={};var I=y(this._fragments,F);for(var U in I)if(I.hasOwnProperty(U)){var q=I[U],V=this._resolvers[U];q==null?(V!=null&&V.dispose(),V=null):q.kind==="PluralReaderSelector"?V==null?V=new $(this._context.environment,this._rootIsQueryRenderer,q,this._callback!=null,this._onChange):(V instanceof $||S(!1),V.setSelector(q)):V==null?V=new E(this._context.environment,this._rootIsQueryRenderer,q,this._callback!=null,this._onChange):(V instanceof E||S(!1),V.setSelector(q)),this._props[U]=F[U],this._resolvers[U]=V}this._stale=!0},w.setVariables=function(F,I){for(var U in this._resolvers)if(this._resolvers.hasOwnProperty(U)){var q=this._resolvers[U];q&&q.setVariables(F,I)}this._stale=!0},A})(),E=(function(){function A(R,F,I,U,q){var V=this;(0,n.default)(this,"_onChange",function(j){V._data=j.data,V._isMissingData=j.isMissingData,V._fieldErrors=j.fieldErrors,V._callback()});var N=R.lookup(I);this._callback=q,this._data=N.data,this._isMissingData=N.isMissingData,this._fieldErrors=N.fieldErrors,this._environment=R,this._rootIsQueryRenderer=F,this._selector=I,o.ENABLE_CONTAINERS_SUBSCRIBE_ON_COMMIT===!0?U&&(this._subscription=R.subscribe(N,this._onChange)):this._subscription=R.subscribe(N,this._onChange)}var w=A.prototype;return w.dispose=function(){this._subscription&&(this._subscription.dispose(),this._subscription=null)},w.resolve=function(){if(this._isMissingData===!0){var F=r(this._environment,this._selector.node,this._selector.owner),I=F==null?void 0:F.promise;if(I!=null&&!this._rootIsQueryRenderer){var U,q=(U=F==null?void 0:F.pendingOperations)!==null&&U!==void 0?U:[];throw this._environment.__log({name:"suspense.fragment",data:this._data,fragment:this._selector.node,isRelayHooks:!1,isMissingData:this._isMissingData,isPromiseCached:!1,pendingOperations:q}),I}}return a(this._environment,this._fieldErrors),this._data},w.setSelector=function(F){if(!(this._subscription!=null&&h(F,this._selector))){this.dispose();var I=this._environment.lookup(F);this._data=l(this._data,I.data),this._isMissingData=I.isMissingData,this._fieldErrors=I.fieldErrors,this._selector=F,this._subscription=this._environment.subscribe(I,this._onChange)}},w.setVariables=function(F,I){if(!v(F,this._selector.variables)){var U=c(I,F),q=m(this._selector.node,this._selector.dataID,F,U);this.setSelector(q)}},A})(),$=(function(){function A(R,F,I,U,q){var V=this;(0,n.default)(this,"_onChange",function(N){V._stale=!0,V._callback()}),this._callback=q,this._data=[],this._environment=R,this._resolvers=[],this._stale=!0,this._rootIsQueryRenderer=F,this._subscribeOnConstruction=U,this.setSelector(I)}var w=A.prototype;return w.dispose=function(){this._resolvers.forEach(k)},w.resolve=function(){if(this._stale){for(var F=this._data,I,U=0;U<this._resolvers.length;U++){var q=F[U],V=this._resolvers[U].resolve();(I||V!==q)&&(I=I||F.slice(0,U),I.push(V))}!I&&this._resolvers.length!==F.length&&(I=F.slice(0,this._resolvers.length)),this._data=I||F,this._stale=!1}return this._data},w.setSelector=function(F){for(var I=F.selectors;this._resolvers.length>I.length;){var U=this._resolvers.pop();U.dispose()}for(var q=0;q<I.length;q++)q<this._resolvers.length?this._resolvers[q].setSelector(I[q]):this._resolvers[q]=new E(this._environment,this._rootIsQueryRenderer,I[q],this._subscribeOnConstruction,this._onChange);this._stale=!0},w.setVariables=function(F,I){this._resolvers.forEach(function(U){return U.setVariables(F,I)}),this._stale=!0},A})();function k(A){A&&A.dispose()}return mR=_,mR}var gR,xz;function u4e(){if(xz)return gR;xz=1;var e=a4e();Vi();function t(n,r,i,a,u,l){return new e(n,i,a,l,u)}return gR=t,gR}var vR,Sz;function l4e(){if(Sz)return vR;Sz=1;var e=un(),t,n;function r(i){return t||(t=i.createContext(null),n=i),i!==n&&e(!1),t}return vR=r,vR}var yR,_z;function o4e(){if(_z)return yR;_z=1;var e=un(),t,n;function r(i){return t||(t=i.createContext(null),n=i),i!==n&&e(!1),t}return yR=r,yR}var bR,Ez;function wC(){if(Ez)return bR;Ez=1;var e=Object.freeze({__LIVE_RESOLVER_SUSPENSE_SENTINEL:!0});function t(){return e}function n(r){return r===e}return bR={isSuspenseSentinel:n,suspenseSentinel:t},bR}var OR={exports:{}},Dz;function H3(){return Dz||(Dz=1,(function(e){function t(n,r){if(n==null)return{};var i={};for(var a in n)if({}.hasOwnProperty.call(n,a)){if(r.indexOf(a)!==-1)continue;i[a]=n[a]}return i}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})(OR)),OR.exports}var xR,$z;function hf(){if($z)return xR;$z=1;var e=$n().default,t=e(mc()),n=e(H3()),r=e(PJ()),i=e(Ui()),a=sS(),u=N0(),l=u.generateClientObjectClientID;u.isClientID;var o=wC(),s=o.isSuspenseSentinel,c=Da(),d=c.ACTOR_IDENTIFIER_KEY,h=c.ERRORS_KEY,m=c.ID_KEY,y=c.INVALIDATED_AT_KEY,v=c.REF_KEY,S=c.REFS_KEY,_=c.RELAY_RESOLVER_VALUE_KEY;c.ROOT_ID;var E=c.TYPENAME_KEY,$=ed(),k=un();Vi();function A(se){return(0,i.default)({},se)}function w(se,pe){for(var ve in se)se.hasOwnProperty(ve)&&ve!==m&&ve!==E&&(pe[ve]=se[ve])}function R(se,pe){var ve={};return ve[m]=se,ve[E]=pe,ve}function F(se){return se}function I(se){return se[m]}function U(se){return h in se?Object.keys(se).filter(function(pe){return pe!==h}):Object.keys(se)}function q(se){return se[E]}function V(se,pe){var ve;return(ve=se[h])===null||ve===void 0?void 0:ve[pe]}function N(se,pe){var ve=se[pe];return ve&&typeof ve=="object"&&(!ve.hasOwnProperty(v)&&!ve.hasOwnProperty(S)||k(!1)),ve}function j(se,pe){return pe in se}function W(se,pe){var ve=se[pe];if(ve==null)return ve;var we=ve;return typeof we=="object"&&we&&typeof we[v]=="string"||k(!1),we[v]}function K(se,pe){var ve=se[pe];if(ve==null)return!1;var we=ve;return typeof we=="object"&&we&&typeof we[v]=="string"}function J(se,pe){var ve=se[pe];return ve==null?ve:(typeof ve=="object"&&Array.isArray(ve[S])||k(!1),ve[S])}function te(se,pe){var ve=se[pe];return ve==null?!1:typeof ve=="object"&&Array.isArray(ve[S])&&ve[S].every(function(we){return typeof we=="string"})}function H(se){if(se==null)return null;var pe=se[y];return typeof pe!="number"?null:pe}function M(se,pe){var ve,we=se[h],Me=pe[h],Re=null;if(we==null&&Me==null){var et;for(var tt in pe)(Re||!$(se[tt],pe[tt]))&&(Re=Re!==null?Re:(0,i.default)({},se),Re[tt]=pe[tt]);return(et=Re)!==null&&et!==void 0?et:se}for(var Oe in pe)if(Oe!==h){var $e=pe[Oe],Ae=Me==null?void 0:Me[Oe];if(Re==null){var He=se[Oe],je=we==null?void 0:we[Oe];if($(He,$e)&&$(je,Ae))continue;Re=(0,i.default)({},se),we!=null&&(Re[h]=(0,i.default)({},we))}ne(Re,Oe,$e),B(Re,Oe,Ae)}return(ve=Re)!==null&&ve!==void 0?ve:se}function G(se,pe){if(h in se||h in pe){var ve=se[h],we=(0,n.default)(se,[h].map(r.default)),Me=pe[h],Re=(0,n.default)(pe,[h].map(r.default)),et=(0,i.default)((0,i.default)({},we),Re);if(ve==null&&Me==null)return et;var tt={};for(var Oe in ve)Re.hasOwnProperty(Oe)||(tt[Oe]=ve[Oe]);for(var $e in Me)tt[$e]=Me[$e];for(var Ae in tt){et[h]=tt;break}return et}else return(0,i.default)((0,i.default)({},se),pe)}function X(se){a(se)}function B(se,pe,ve){var we=se[h];if(ve!=null&&ve.length>0)we==null?se[h]=(0,t.default)({},pe,ve):we[pe]=ve;else if(we!=null&&delete we[pe]){for(var Me in we)if(we.hasOwnProperty(Me))return;delete se[h]}}function ne(se,pe,ve){se[pe]=ve}function oe(se,pe,ve){var we={};we[v]=ve,se[pe]=we}function ie(se,pe,ve){var we={};we[S]=ve,se[pe]=we}function de(se,pe,ve,we){var Me={};Me[v]=we,Me[d]=ve,se[pe]=Me}function ge(se,pe){var ve=se[pe];return ve==null?ve:(typeof ve=="object"&&typeof ve[v]=="string"&&ve[d]!=null||k(!1),[ve[d],ve[v]])}function ye(se,pe){var ve=N(se,_);return ve==null||s(ve)?null:(typeof ve=="object"&&(ve=ve.id),typeof ve!="string"&&k(!1),l(pe,ve))}function me(se,pe){var ve=N(se,_);return ve==null||s(ve)?null:(Array.isArray(ve)||k(!1),ve.map(function(we){return we==null?null:(typeof we=="object"&&(we=we.id),typeof we!="string"&&k(!1),l(pe,we))}))}function Ee(se){return se}return xR={clone:A,copyFields:w,create:R,freeze:X,fromObject:F,getDataID:I,getErrors:V,getFields:U,getInvalidationEpoch:H,getLinkedRecordID:W,getLinkedRecordIDs:J,getType:q,getValue:N,hasValue:j,hasLinkedRecordID:K,hasLinkedRecordIDs:te,merge:G,setErrors:B,setValue:ne,setLinkedRecordID:oe,setLinkedRecordIDs:ie,update:M,getActorLinkedRecordID:ge,setActorLinkedRecordID:de,getResolverLinkedRecordID:ye,getResolverLinkedRecordIDs:me,toJSON:Ee},xR}var SR,Cz;function RC(){if(Cz)return SR;Cz=1;var e={EXISTENT:"EXISTENT",NONEXISTENT:"NONEXISTENT",UNKNOWN:"UNKNOWN"};return SR=e,SR}var _R,wz;function E1(){if(wz)return _R;wz=1;var e=$n().default,t=e(Cu()),n=hf(),r=RC(),i=r.EXISTENT,a=r.NONEXISTENT,u=r.UNKNOWN,l=(function(){function o(c){var d=this;this._records=new Map,c!=null&&Object.keys(c).forEach(function(h){var m=c[h],y=n.fromObject(m);d._records.set(h,y)})}o.create=function(d){return new o(d)};var s=o.prototype;return s.clear=function(){this._records=new Map},s.delete=function(d){this._records.set(d,null)},s.get=function(d){return this._records.get(d)},s.getRecordIDs=function(){return Array.from(this._records.keys())},s.getStatus=function(d){return this._records.has(d)?this._records.get(d)==null?a:i:u},s.has=function(d){return this._records.has(d)},s.remove=function(d){this._records.delete(d)},s.set=function(d,h){this._records.set(d,h)},s.size=function(){return this._records.size},s.toJSON=function(){var d={},h=(0,t.default)(this._records),m;try{for(h.s();!(m=h.n()).done;){var y=m.value,v=y[0],S=y[1];d[v]=n.toJSON(S)}}catch(_){h.e(_)}finally{h.f()}return d},o})();return _R=l,_R}var ER,Rz;function HF(){if(Rz)return ER;Rz=1;var e=un(),t="INTERNAL_ACTOR_IDENTIFIER_DO_NOT_USE";function n(r){r!==t&&e(!1)}return ER={assertInternalActorIdentifier:n,getActorIdentifier:function(i){return i},getDefaultActorIdentifier:function(){throw new Error("Not Implemented")},INTERNAL_ACTOR_IDENTIFIER_DO_NOT_USE:t},ER}var DR,Az;function s4e(){if(Az)return DR;Az=1;var e="actor_key",t=HF(),n=t.getActorIdentifier;function r(i){if(i!=null&&typeof i=="object"&&typeof i[e]=="string")return n(i[e])}return DR={ACTOR_IDENTIFIER_FIELD_NAME:e,getActorIdentifierFromPayload:r},DR}var $R,Tz;function c4e(){if(Tz)return $R;Tz=1;var e=$n().default,t=e(Ui()),n=e(H3()),r=e(Cu()),i=e(sy()),a=["path","locations"],u=Symbol("$SELF");function l(d){if(d==null)return null;var h=new Map,m=(0,r.default)(d),y;try{for(m.s();!(y=m.n()).done;){var v=y.value,S=v.path,_=v.locations,E=(0,n.default)(v,a);if(S!=null){var $=S.length;if($!==0){for(var k=$-1,A=h,w=0;w<k;w++){var R=S[w],F=A.get(R);if(F instanceof Map){A=F;continue}var I=new Map;Array.isArray(F)&&I.set(u,F),A.set(R,I),A=I}var U=S[k],q=A.get(U);q instanceof Map&&(A=q,q=A.get(U),U=u),Array.isArray(q)?q.push(E):A.set(U,[E])}}}}catch(V){m.e(V)}finally{m.f()}return h}function o(d,h){var m=d.get(h);if(m==null)return null;if(Array.isArray(m))return m;var y=[];return s(m,y),y}function s(d,h){var m=(0,r.default)(d),y;try{for(m.s();!(y=m.n()).done;){var v=y.value,S=v[0],_=v[1],E=h.length;if(Array.isArray(_)?h.push.apply(h,(0,i.default)(_)):s(_,h),S!==u)for(var $=h.length,k=E;k<$;k++){var A=h[k];A.path==null?h[k]=(0,t.default)((0,t.default)({},A),{},{path:[S]}):A.path.unshift(S)}}}catch(w){m.e(w)}finally{m.f()}}function c(d,h){var m=d.get(h);return m instanceof Map?m:null}return $R={SELF:u,buildErrorTrie:l,getNestedErrorTrieByKey:c,getErrorsByKey:o},$R}var CR,kz;function AC(){if(kz)return CR;kz=1;var e="client:__type:",t="__TypeSchema";function n(i){return e+i}function r(i){return i.indexOf(e)===0}return CR={generateTypeID:n,isTypeID:r,TYPE_SCHEMA_TYPE:t},CR}var wR,Pz;function HJ(){if(Pz)return wR;Pz=1;var e=$n().default,t=e(Cu()),n=e(sy()),r=s4e();r.ACTOR_IDENTIFIER_FIELD_NAME;var i=r.getActorIdentifierFromPayload,a=nh(),u=N0(),l=u.generateClientID,o=u.isClientID,s=cy(),c=s.getLocalVariables,d=c4e(),h=d.buildErrorTrie,m=d.getErrorsByKey,y=d.getNestedErrorTrieByKey,v=hf(),S=hp(),_=S.createNormalizationSelector,E=Da(),$=E.ROOT_ID,k=E.TYPENAME_KEY,A=E.getArgumentValues,w=E.getHandleStorageKey,R=E.getModuleComponentKey,F=E.getModuleOperationKey,I=E.getStorageKey,U=AC(),q=U.TYPE_SCHEMA_TYPE,V=U.generateTypeID;ed();var N=un();Vi();function j(K,J,te,H,M,G){var X=J.dataID,B=J.node,ne=J.variables,oe=new W(K,ne,H,G??!1);return oe.normalizeResponse(B,X,te,M)}var W=(function(){function K(te,H,M,G){this._actorIdentifier=M.actorIdentifier,this._getDataId=M.getDataID,this._handleFieldPayloads=[],this._treatMissingFieldsAsNull=M.treatMissingFieldsAsNull,this._incrementalPlaceholders=[],this._isClientExtension=!1,this._isUnmatchedAbstractType=!1,this._useExecTimeResolvers=G,this._followupPayloads=[],this._path=M.path?(0,n.default)(M.path):[],this._recordSource=te,this._variables=H,this._shouldProcessClientComponents=M.shouldProcessClientComponents,this._log=M.log}var J=K.prototype;return J.normalizeResponse=function(H,M,G,X){var B=this._recordSource.get(M);return B||N(!1),this._assignClientAbstractTypes(H),this._errorTrie=h(X),this._traverseSelections(H,B,G),{errors:X,fieldPayloads:this._handleFieldPayloads,incrementalPlaceholders:this._incrementalPlaceholders,followupPayloads:this._followupPayloads,source:this._recordSource,isFinal:!1}},J._assignClientAbstractTypes=function(H){var M=H.clientAbstractTypes;if(M!=null)for(var G=0,X=Object.keys(M);G<X.length;G++){var B=X[G],ne=(0,t.default)(M[B]),oe;try{for(ne.s();!(oe=ne.n()).done;){var ie=oe.value,de=V(ie),ge=this._recordSource.get(de);ge==null&&(ge=v.create(de,q),this._recordSource.set(de,ge)),v.setValue(ge,B,!0)}}catch(ye){ne.e(ye)}finally{ne.f()}}},J._getVariableValue=function(H){return this._variables.hasOwnProperty(H)||N(!1),this._variables[H]},J._getRecordType=function(H){var M=H[k];return M==null&&N(!1),M},J._traverseSelections=function(H,M,G){for(var X=0;X<H.selections.length;X++){var B=H.selections[X];switch(B.kind){case"ScalarField":case"LinkedField":this._normalizeField(B,M,G);break;case"Condition":var ne=!!this._getVariableValue(B.condition);ne===B.passingValue&&this._traverseSelections(B,M,G);break;case"FragmentSpread":{var oe=this._variables;this._variables=c(this._variables,B.fragment.argumentDefinitions,B.args),this._traverseSelections(B.fragment,M,G),this._variables=oe;break}case"InlineFragment":{var ie=B.abstractKey;if(ie==null){var de=v.getType(M);de===B.type&&this._traverseSelections(B,M,G)}else{var ge=Object.prototype.hasOwnProperty.call(G,ie),ye=v.getType(M),me=V(ye),Ee=this._recordSource.get(me);Ee==null&&(Ee=v.create(me,q),this._recordSource.set(me,Ee)),v.setValue(Ee,ie,ge),ge&&this._traverseSelections(B,M,G)}break}case"TypeDiscriminator":{var se=B.abstractKey,pe=Object.prototype.hasOwnProperty.call(G,se),ve=v.getType(M),we=V(ve),Me=this._recordSource.get(we);Me==null&&(Me=v.create(we,q),this._recordSource.set(we,Me)),v.setValue(Me,se,pe);break}case"LinkedHandle":case"ScalarHandle":var Re=B.args?A(B.args,this._variables):{},et=I(B,this._variables),tt=w(B,this._variables);this._handleFieldPayloads.push({args:Re,dataID:v.getDataID(M),fieldKey:et,handle:B.handle,handleKey:tt,handleArgs:B.handleArgs?A(B.handleArgs,this._variables):{}});break;case"ModuleImport":this._normalizeModuleImport(B,M,G);break;case"Defer":this._normalizeDefer(B,M,G);break;case"Stream":this._normalizeStream(B,M,G);break;case"ClientExtension":var Oe=this._isClientExtension;this._isClientExtension=!0,this._traverseSelections(B,M,G),this._isClientExtension=Oe;break;case"ClientComponent":if(this._shouldProcessClientComponents===!1)break;this._traverseSelections(B.fragment,M,G);break;case"ActorChange":this._normalizeActorChange(B,M,G);break;case"RelayResolver":case"RelayLiveResolver":this._useExecTimeResolvers||this._normalizeResolver(B,M,G);break;case"ClientEdgeToClientObject":this._useExecTimeResolvers||this._normalizeResolver(B.backingField,M,G);break;default:N(!1)}}},J._normalizeResolver=function(H,M,G){H.fragment!=null&&this._traverseSelections(H.fragment,M,G)},J._normalizeDefer=function(H,M,G){var X=H.if===null||this._getVariableValue(H.if);X===!1?this._traverseSelections(H,M,G):this._incrementalPlaceholders.push({kind:"defer",data:G,label:H.label,path:(0,n.default)(this._path),selector:_(H,v.getDataID(M),this._variables),typeName:v.getType(M),actorIdentifier:this._actorIdentifier})},J._normalizeStream=function(H,M,G){this._traverseSelections(H,M,G);var X=H.if===null||this._getVariableValue(H.if);X===!0&&this._incrementalPlaceholders.push({kind:"stream",label:H.label,path:(0,n.default)(this._path),parentID:v.getDataID(M),node:H,variables:this._variables,actorIdentifier:this._actorIdentifier})},J._normalizeModuleImport=function(H,M,G){typeof G=="object"&&G||N(!1);var X=v.getType(M),B=R(H.documentName),ne=H.componentModuleProvider||G[B];v.setValue(M,B,ne??null);var oe=F(H.documentName),ie=H.operationModuleProvider||G[oe];v.setValue(M,oe,ie??null),ie!=null&&this._followupPayloads.push({kind:"ModuleImportPayload",args:H.args,data:G,dataID:v.getDataID(M),operationReference:ie,path:(0,n.default)(this._path),typeName:X,variables:this._variables,actorIdentifier:this._actorIdentifier})},J._normalizeField=function(H,M,G){typeof G=="object"&&G||N(!1);var X=H.alias||H.name,B=I(H,this._variables),ne=G[X],oe=a.ENABLE_NONCOMPLIANT_ERROR_HANDLING_ON_LISTS&&Array.isArray(ne)&&ne.length===0;if(ne==null||oe){if(ne===void 0){var ie=this._isClientExtension||this._isUnmatchedAbstractType;if(ie)return;if(!this._treatMissingFieldsAsNull)return}H.kind==="ScalarField"&&this._validateConflictingFieldsWithIdenticalId(M,B,null),oe?H.kind==="LinkedField"?v.setLinkedRecordIDs(M,B,[]):v.setValue(M,B,[]):v.setValue(M,B,null);var de=this._errorTrie;if(de!=null){var ge=m(de,X);ge!=null&&v.setErrors(M,B,ge)}return}if(H.kind==="ScalarField")this._validateConflictingFieldsWithIdenticalId(M,B,ne),v.setValue(M,B,ne);else if(H.kind==="LinkedField"){this._path.push(X);var ye=this._errorTrie;this._errorTrie=ye==null?null:y(ye,X),H.plural?this._normalizePluralLink(H,M,B,ne):this._normalizeLink(H,M,B,ne),this._errorTrie=ye,this._path.pop()}else N(!1)},J._normalizeActorChange=function(H,M,G){var X,B=H.linkedField;typeof G=="object"&&G||N(!1);var ne=B.alias||B.name,oe=I(B,this._variables),ie=G[ne];if(ie==null){if(ie===void 0){var de=this._isClientExtension||this._isUnmatchedAbstractType;if(de)return;if(!this._treatMissingFieldsAsNull)return}v.setValue(M,oe,null);return}var ge=i(ie);if(ge==null){v.setValue(M,oe,null);return}var ye=(X=B.concreteType)!==null&&X!==void 0?X:this._getRecordType(ie),me=this._getDataId(ie,ye)||v.getLinkedRecordID(M,oe)||l(v.getDataID(M),oe);typeof me!="string"&&N(!1),v.setActorLinkedRecordID(M,oe,ge,me),this._followupPayloads.push({kind:"ActorPayload",data:ie,dataID:me,path:[].concat((0,n.default)(this._path),[ne]),typeName:ye,variables:this._variables,node:B,actorIdentifier:ge})},J._normalizeLink=function(H,M,G,X){var B;typeof X=="object"&&X||N(!1);var ne=this._getDataId(X,(B=H.concreteType)!==null&&B!==void 0?B:this._getRecordType(X))||v.getLinkedRecordID(M,G)||l(v.getDataID(M),G);typeof ne!="string"&&N(!1),this._validateConflictingLinkedFieldsWithIdenticalId(v.getLinkedRecordID(M,G),ne,G),v.setLinkedRecordID(M,G,ne);var oe=this._recordSource.get(ne);if(oe)this._validateRecordType(oe,H,X);else{var ie=H.concreteType||this._getRecordType(X);oe=v.create(ne,ie),this._recordSource.set(ne,oe)}this._traverseSelections(H,oe,X)},J._normalizePluralLink=function(H,M,G,X){var B=this;Array.isArray(X)||N(!1);var ne=v.getLinkedRecordIDs(M,G),oe=[];X.forEach(function(ie,de){var ge;if(ie==null){oe.push(ie);return}B._path.push(String(de));var ye=B._errorTrie;B._errorTrie=ye==null?null:y(ye,de),typeof ie!="object"&&N(!1);var me=B._getDataId(ie,(ge=H.concreteType)!==null&&ge!==void 0?ge:B._getRecordType(ie))||ne&&ne[de]||l(v.getDataID(M),G,de);typeof me!="string"&&N(!1),oe.push(me);var Ee=B._recordSource.get(me);if(Ee)B._validateRecordType(Ee,H,ie);else{var se=H.concreteType||B._getRecordType(ie);Ee=v.create(me,se),B._recordSource.set(me,Ee)}ne&&B._validateConflictingLinkedFieldsWithIdenticalId(ne[de],me,G),B._traverseSelections(H,Ee,ie),B._errorTrie=ye,B._path.pop()}),v.setLinkedRecordIDs(M,G,oe)},J._validateRecordType=function(H,M,G){if(a.ENABLE_STORE_ID_COLLISION_LOGGING){var X,B=(X=M.concreteType)!==null&&X!==void 0?X:this._getRecordType(G),ne=v.getDataID(H),oe=o(ne)&&ne!==$||v.getType(H)===B;if(!oe){var ie={name:"idCollision.typename",previous_typename:v.getType(H),new_typename:B};this._log!=null&&this._log(ie)}}},J._validateConflictingFieldsWithIdenticalId=function(H,M,G){},J._validateConflictingLinkedFieldsWithIdenticalId=function(H,M,G){},K})();return wR={normalize:j},wR}var RR,Mz;function XJ(){if(Mz)return RR;Mz=1;var e=$n().default,t=e(Ui()),n=e(hf()),r=e(E1()),i=e(HJ());function a(u,l,o,s,c){var d,h=u.data,m=u.errors,y=r.default.create(),v=n.default.create(l.dataID,o);y.set(l.dataID,v);var S=i.default.normalize(y,l,h,s,m,c);return(0,t.default)((0,t.default)({},S),{},{isFinal:((d=u.extensions)===null||d===void 0?void 0:d.is_final)===!0})}return RR=a,RR}var AR,Fz;function f4e(){if(Fz)return AR;Fz=1;var e=dp(),t=e.getInlineDataFragment,n=Da(),r=n.FRAGMENTS_KEY,i=un();function a(u,l){var o,s=t(u);if(l==null)return l;typeof l!="object"&&i(!1);var c=(o=l[r])===null||o===void 0?void 0:o[s.name];return c==null&&i(!1),c}return AR=a,AR}var TR,Nz;function YJ(){if(Nz)return TR;Nz=1;var e=1e5;function t(){return e++}return TR=t,TR}var kR,Iz;function d4e(){if(Iz)return kR;Iz=1;var e=YJ();function t(n,r){return{execute:function(a,u,l,o,s,c,d,h){var m=e(),y={start:function(_){n.__log({name:"network.start",networkRequestId:m,params:a,variables:u,cacheConfig:l})},next:function(_){n.__log({name:"network.next",networkRequestId:m,response:_})},error:function(_){n.__log({name:"network.error",networkRequestId:m,error:_})},complete:function(){n.__log({name:"network.complete",networkRequestId:m})},unsubscribe:function(){n.__log({name:"network.unsubscribe",networkRequestId:m})}},v=function(_){n.__log({name:"network.info",networkRequestId:m,info:_})};return r.execute(a,u,l,o,v,c,d,h).do(y)}}}return kR=t,kR}var PR,Lz;function GJ(){if(Lz)return PR;Lz=1;var e=$n().default,t=e(Cu()),n=un(),r=(function(){function i(){this._ownersToPendingOperations=new Map,this._pendingOperationsToOwners=new Map,this._ownersToPendingPromise=new Map}var a=i.prototype;return a.update=function(l,o){if(o.size!==0){var s=l.identifier,c=new Set,d=(0,t.default)(o),h;try{for(d.s();!(h=d.n()).done;){var m=h.value,y=m.identifier,v=this._ownersToPendingOperations.get(y);v!=null?v.has(s)||(v.set(s,l),c.add(y)):(this._ownersToPendingOperations.set(y,new Map([[s,l]])),c.add(y))}}catch(k){d.e(k)}finally{d.f()}if(c.size!==0){var S=this._pendingOperationsToOwners.get(s)||new Set,_=(0,t.default)(c),E;try{for(_.s();!(E=_.n()).done;){var $=E.value;this._resolveOwnerResolvers($),S.add($)}}catch(k){_.e(k)}finally{_.f()}this._pendingOperationsToOwners.set(s,S)}}},a.complete=function(l){var o=l.identifier,s=this._pendingOperationsToOwners.get(o);if(s!=null){var c=new Set,d=new Set,h=(0,t.default)(s),m;try{for(h.s();!(m=h.n()).done;){var y=m.value,v=this._ownersToPendingOperations.get(y);v&&(v.delete(o),v.size>0?d.add(y):c.add(y))}}catch(w){h.e(w)}finally{h.f()}var S=(0,t.default)(c),_;try{for(S.s();!(_=S.n()).done;){var E=_.value;this._resolveOwnerResolvers(E),this._ownersToPendingOperations.delete(E)}}catch(w){S.e(w)}finally{S.f()}var $=(0,t.default)(d),k;try{for($.s();!(k=$.n()).done;){var A=k.value;this._resolveOwnerResolvers(A)}}catch(w){$.e(w)}finally{$.f()}this._pendingOperationsToOwners.delete(o)}},a._resolveOwnerResolvers=function(l){var o=this._ownersToPendingPromise.get(l);o!=null&&o.resolve(),this._ownersToPendingPromise.delete(l)},a.getPendingOperationsAffectingOwner=function(l){var o=l.identifier,s=this._ownersToPendingOperations.get(o);if(s==null||s.size===0)return null;var c=this._ownersToPendingPromise.get(o);if(c!=null)return{promise:c.promise,pendingOperations:c.pendingOperations};var d,h=new Promise(function(y){d=y});d==null&&n(!1);var m=Array.from(s.values());return this._ownersToPendingPromise.set(o,{promise:h,resolve:d,pendingOperations:m}),{promise:h,pendingOperations:m}},i})();return PR=r,PR}var MR,Bz;function h4e(){if(Bz)return MR;Bz=1;function e(t){var n=typeof Xd<"u"?Xd:typeof window<"u"?window:void 0,r=n&&n.__RELAY_DEVTOOLS_HOOK__;r&&r.registerEnvironment(t)}return MR=e,MR}var FR,Qz;function XF(){if(Qz)return FR;Qz=1;var e=N0(),t=e.generateClientID,n=Da(),r=n.ROOT_ID,i=t(r,"viewer"),a="Viewer";return FR={VIEWER_ID:i,VIEWER_TYPE:a},FR}var NR,zz;function KJ(){if(zz)return NR;zz=1;var e=XF(),t=e.VIEWER_ID,n=e.VIEWER_TYPE;function r(i,a){return a===n&&i.id==null?t:i.id}return NR=r,NR}var IR,qz;function p4e(){if(qz)return IR;qz=1;var e=function(n){};return IR=e,IR}var LR,Vz;function YF(){if(Vz)return LR;Vz=1;var e=oy(),t=e.REQUEST,n=e.SPLIT_OPERATION;function r(i){switch(i.kind){case t:return i.operation;case n:default:return i}}return LR=r,LR}var BR,Uz;function WJ(){if(Uz)return BR;Uz=1;function e(t,n,r){for(var i=arguments.length,a=new Array(i>3?i-3:0),u=3;u<i;u++)a[u-3]=arguments[u];var l=0,o=r.replace(/%s/g,function(){return String(a[l++])}),s=new Error(o),c=Object.assign(s,{name:n,messageFormat:r,messageParams:a,type:t,taalOpcodes:[2,2]});if(c.stack===void 0)try{throw c}catch{}return c}return BR={create:function(n,r){for(var i=arguments.length,a=new Array(i>2?i-2:0),u=2;u<i;u++)a[u-2]=arguments[u];return e.apply(void 0,["error",n,r].concat(a))},createWarning:function(n,r){for(var i=arguments.length,a=new Array(i>2?i-2:0),u=2;u<i;u++)a[u-2]=arguments[u];return e.apply(void 0,["warn",n,r].concat(a))}},BR}var QR,jz;function m4e(){if(jz)return QR;jz=1;var e,t,n=typeof window<"u"&&typeof((e=window)===null||e===void 0||(t=e.performance)===null||t===void 0?void 0:t.now)=="function";function r(){return n?window.performance.now():Date.now()}function i(a){var u=r(),l=a();return[u,r()-u,l]}return QR=i,QR}var zR,Zz;function g4e(){if(Zz)return zR;Zz=1;var e=$n().default,t=e(Cu()),n=e(sy()),r=fy(),i=YJ(),a=YF(),u=WJ(),l=nh(),o=_1(),s=o.stableCopy,c=m4e(),d=N0(),h=d.generateClientID,m=d.generateUniqueClientID,y=cy(),v=y.getLocalVariables,S=hf(),_=hp(),E=_.createNormalizationSelector,$=_.createReaderSelector,k=E1(),A=Da(),w=A.ROOT_TYPE,R=A.TYPENAME_KEY,F=A.getStorageKey,I=un(),U=Vi();function q(K){return new V(K)}var V=(function(){function K(te){var H=this,M,G,X,B=te.actorIdentifier,ne=te.getDataID,oe=te.getPublishQueue,ie=te.getStore,de=te.isClientPayload,ge=te.operation,ye=te.operationExecutions,me=te.operationLoader,Ee=te.operationTracker,se=te.optimisticConfig,pe=te.scheduler,ve=te.shouldProcessClientComponents,we=te.sink,Me=te.source,Re=te.treatMissingFieldsAsNull,et=te.updater,tt=te.log,Oe=te.normalizeResponse;this._actorIdentifier=B,this._getDataID=ne,this._treatMissingFieldsAsNull=Re,this._incrementalPayloadsPending=!1,this._incrementalResults=new Map,this._log=tt,this._executeId=i(),this._nextSubscriptionId=0,this._operation=ge,this._operationExecutions=ye,this._operationLoader=me,this._operationTracker=Ee,this._operationUpdateEpochs=new Map,this._optimisticUpdates=null,this._useExecTimeResolvers=(M=(G=this._operation.request.node.operation.use_exec_time_resolvers)!==null&&G!==void 0?G:((X=this._operation.request.node.operation.exec_time_resolvers_enabled_provider)===null||X===void 0?void 0:X.get())===!0)!==null&&M!==void 0?M:!1,this._execTimeResolverResponseComplete=!1,this._pendingModulePayloadsCount=0,this._getPublishQueue=oe,this._scheduler=pe,this._sink=we,this._source=new Map,this._state="started",this._getStore=ie,this._subscriptions=new Map,this._updater=et,this._isClientPayload=de===!0,this._isSubscriptionOperation=this._operation.request.node.params.operationKind==="subscription",this._shouldProcessClientComponents=ve,this._retainDisposables=new Map,this._seenActors=new Set,this._completeFns=[],this._normalizeResponse=Oe;var $e=this._nextSubscriptionId++;l.PROCESS_OPTIMISTIC_UPDATE_BEFORE_SUBSCRIPTION&&se!=null&&this._processOptimisticResponse(se.response!=null?{data:se.response}:null,se.updater,!1),Me.subscribe({complete:function(){return H._complete($e)},error:function(He){return H._error(He)},next:function(He){try{H._next($e,He)}catch(je){we.error(je)}},start:function(He){var je;H._start($e,He),H._log({name:"execute.start",executeId:H._executeId,params:H._operation.request.node.params,variables:H._operation.request.variables,cacheConfig:(je=H._operation.request.cacheConfig)!==null&&je!==void 0?je:{}})},unsubscribe:function(){H._log({name:"execute.unsubscribe",executeId:H._executeId})}}),!l.PROCESS_OPTIMISTIC_UPDATE_BEFORE_SUBSCRIPTION&&se!=null&&this._processOptimisticResponse(se.response!=null?{data:se.response}:null,se.updater,!1)}var J=K.prototype;return J.cancel=function(){var H=this;if(this._state!=="completed"){this._state="completed",this._operationExecutions.delete(this._operation.request.identifier),this._subscriptions.size!==0&&(this._subscriptions.forEach(function(G){return G.unsubscribe()}),this._subscriptions.clear());var M=this._optimisticUpdates;M!==null&&(this._optimisticUpdates=null,M.forEach(function(G){return H._getPublishQueueAndSaveActor().revertUpdate(G)}),this._runPublishQueue()),this._incrementalResults.clear(),this._asyncStoreUpdateDisposable!=null&&(this._asyncStoreUpdateDisposable.dispose(),this._asyncStoreUpdateDisposable=null),this._completeFns=[],this._completeOperationTracker(),this._disposeRetainedData()}},J._updateActiveState=function(){var H;switch(this._state){case"started":{H="active";break}case"loading_incremental":{H="active";break}case"completed":{H="inactive";break}case"loading_final":{H=this._pendingModulePayloadsCount>0||this._useExecTimeResolvers&&!this._execTimeResolverResponseComplete?"active":"inactive";break}default:this._state,I(!1)}this._operationExecutions.set(this._operation.request.identifier,H)},J._schedule=function(H,M){var G=this,X=this._scheduler;if(X!=null){var B=this._nextSubscriptionId++;r.create(function(ne){var oe=X.schedule(function(){try{H(),ne.complete()}catch(ie){ne.error(ie)}},M);return function(){return X.cancel(oe)}}).subscribe({complete:function(){return G._complete(B)},error:function(oe){return G._error(oe)},start:function(oe){return G._start(B,oe)}})}else H()},J._complete=function(H){this._subscriptions.delete(H),this._subscriptions.size===0&&(this.cancel(),this._sink.complete(),this._log({name:"execute.complete",executeId:this._executeId}))},J._error=function(H){this.cancel(),this._sink.error(H),this._log({name:"execute.error",executeId:this._executeId,error:H})},J._start=function(H,M){this._subscriptions.set(H,M),this._updateActiveState()},J._next=function(H,M){var G=this,X=this._state==="loading_incremental"?"low":"default";this._schedule(function(){G._log({name:"execute.next.start",executeId:G._executeId,response:M,operation:G._operation}),G._handleNext(M),G._maybeCompleteSubscriptionOperationTracking(),G._log({name:"execute.next.end",executeId:G._executeId,response:M,operation:G._operation})},X)},J._handleErrorResponse=function(H){var M=this,G=[];return H.forEach(function(X){if(!(X.data===null&&X.extensions!=null&&!X.hasOwnProperty("errors")))if(X.data==null){var B=X.hasOwnProperty("errors")&&X.errors!=null?X.errors:null,ne=B?B.map(function(de){var ge=de.message;return ge}).join(`
`):"(No errors)",oe=u.create("RelayNetwork","No data returned for operation `"+M._operation.request.node.params.name+"`, got error(s):\n"+ne+"\n\nSee the error `source` property for more information.");throw oe.source={errors:B,operation:M._operation.request.node,variables:M._operation.request.variables},oe.stack,oe}else{var ie=X;G.push(ie)}}),G},J._handleOptimisticResponses=function(H){var M;if(H.length>1)return H.some(function(B){var ne;return((ne=B.extensions)===null||ne===void 0?void 0:ne.isOptimistic)===!0})&&I(!1),!1;var G=H[0],X=((M=G.extensions)===null||M===void 0?void 0:M.isOptimistic)===!0;return X&&this._state!=="started"&&I(!1),X?(this._processOptimisticResponse(G,null,this._treatMissingFieldsAsNull),this._sink.next(G),!0):!1},J._handleNext=function(H){if(this._state!=="completed"){this._seenActors.clear();var M=Array.isArray(H)?H:[H],G=this._handleErrorResponse(M);if(G.length===0){var X=M.some(function(Oe){var $e;return(($e=Oe.extensions)===null||$e===void 0?void 0:$e.is_final)===!0});X&&(this._state="loading_final",this._updateActiveState(),this._incrementalPayloadsPending=!1),this._sink.next(H);return}var B=this._handleOptimisticResponses(G);if(!B){var ne=N(G),oe=ne[0],ie=ne[1],de=ne[2],ge=oe.length>0,ye=de.length>0;if(ge){if(this._isSubscriptionOperation){var me=m();this._operation={request:this._operation.request,fragment:$(this._operation.fragment.node,me,this._operation.fragment.variables,this._operation.fragment.owner),root:E(this._operation.root.node,me,this._operation.root.variables)}}var Ee=this._processResponses(oe);this._processPayloadFollowups(Ee)}if(ye)for(var se=[],pe=0;pe<de.length;pe++){var ve,we=de[pe],Me=new k(we.data),X=((ve=we.extensions)===null||ve===void 0?void 0:ve.is_final)===!0,Re={errors:[],fieldPayloads:[],followupPayloads:[],incrementalPlaceholders:[],isFinal:X,source:Me};this._getPublishQueueAndSaveActor().commitPayload(this._operation,Re,this._updater),se.push(Re),this._execTimeResolverResponseComplete=X,X&&this._updateActiveState()}if(ie.length>0){var et=this._processIncrementalResponses(ie);this._processPayloadFollowups(et)}this._isSubscriptionOperation&&(G[0].extensions==null?G[0].extensions={__relay_subscription_root_id:this._operation.fragment.dataID}:G[0].extensions.__relay_subscription_root_id=this._operation.fragment.dataID);var tt=this._runPublishQueue(ge||ye?this._operation:void 0);ge&&this._incrementalPayloadsPending&&this._retainData(),this._updateOperationTracker(tt),this._sink.next(H)}}},J._processOptimisticResponse=function(H,M,G){var X=this;if(this._optimisticUpdates!==null&&I(!1),!(H==null&&M==null)){var B=[];if(H){var ne=this._normalizeResponse(H,this._operation.root,w,{actorIdentifier:this._actorIdentifier,getDataID:this._getDataID,log:this._log,path:[],shouldProcessClientComponents:this._shouldProcessClientComponents,treatMissingFieldsAsNull:G},this._useExecTimeResolvers);W(ne),B.push({operation:this._operation,payload:ne,updater:M}),this._processOptimisticFollowups(ne,B)}else M&&B.push({operation:this._operation,payload:{errors:null,fieldPayloads:null,incrementalPlaceholders:null,followupPayloads:null,source:k.create(),isFinal:!1},updater:M});this._optimisticUpdates=B,B.forEach(function(ie){return X._getPublishQueueAndSaveActor().applyUpdate(ie)});var oe=this._runPublishQueue();l.ENABLE_OPERATION_TRACKER_OPTIMISTIC_UPDATES&&this._updateOperationTracker(oe)}},J._processOptimisticFollowups=function(H,M){if(H.followupPayloads&&H.followupPayloads.length){var G=H.followupPayloads,X=(0,t.default)(G),B;try{for(X.s();!(B=X.n()).done;){var ne=B.value;switch(ne.kind){case"ModuleImportPayload":var oe=this._expectOperationLoader(),ie=oe.get(ne.operationReference);if(ie==null)this._processAsyncOptimisticModuleImport(ne);else{var de=this._processOptimisticModuleImport(ie,ne);M.push.apply(M,(0,n.default)(de))}break;case"ActorPayload":break;default:I(!1)}}}catch(ge){X.e(ge)}finally{X.f()}}},J._normalizeFollowupPayload=function(H,M){var G;M.kind==="SplitOperation"&&H.kind==="ModuleImportPayload"?G=v(H.variables,M.argumentDefinitions,H.args):G=H.variables;var X=E(M,H.dataID,G),B={data:H.data,extensions:this._state==="loading_final"?{is_final:!0}:void 0};return this._normalizeResponse(B,X,H.typeName,{actorIdentifier:this._actorIdentifier,getDataID:this._getDataID,log:this._log,path:H.path,treatMissingFieldsAsNull:this._treatMissingFieldsAsNull,shouldProcessClientComponents:this._shouldProcessClientComponents},this._useExecTimeResolvers)},J._processOptimisticModuleImport=function(H,M){var G=a(H),X=[],B=this._normalizeFollowupPayload(M,G);return W(B),X.push({operation:this._operation,payload:B,updater:null}),this._processOptimisticFollowups(B,X),X},J._processAsyncOptimisticModuleImport=function(H){var M=this;this._expectOperationLoader().load(H.operationReference).then(function(G){if(!(G==null||M._state!=="started")){var X=M._processOptimisticModuleImport(G,H);if(X.forEach(function(ne){return M._getPublishQueueAndSaveActor().applyUpdate(ne)}),M._optimisticUpdates!=null){var B;(B=M._optimisticUpdates).push.apply(B,(0,n.default)(X)),M._runPublishQueue()}}})},J._processResponses=function(H){var M=this;return this._log({name:"execute.normalize.start",operation:this._operation}),this._optimisticUpdates!==null&&(this._optimisticUpdates.forEach(function(G){M._getPublishQueueAndSaveActor().revertUpdate(G)}),this._optimisticUpdates=null),this._incrementalPayloadsPending=!1,this._incrementalResults.clear(),this._source.clear(),H.map(function(G){var X=M._normalizeResponse(G,M._operation.root,w,{actorIdentifier:M._actorIdentifier,getDataID:M._getDataID,log:M._log,path:[],treatMissingFieldsAsNull:M._treatMissingFieldsAsNull,shouldProcessClientComponents:M._shouldProcessClientComponents},M._useExecTimeResolvers);return M._getPublishQueueAndSaveActor().commitPayload(M._operation,X,M._updater),M._log({name:"execute.normalize.end",operation:M._operation}),X})},J._processPayloadFollowups=function(H){var M=this;this._state!=="completed"&&H.forEach(function(G){var X=G.incrementalPlaceholders,B=G.followupPayloads,ne=G.isFinal;if(M._state=ne?"loading_final":"loading_incremental",M._updateActiveState(),ne&&(M._incrementalPayloadsPending=!1),B&&B.length!==0&&B.forEach(function(ie){var de,ge=M._actorIdentifier;M._actorIdentifier=(de=ie.actorIdentifier)!==null&&de!==void 0?de:M._actorIdentifier,M._processFollowupPayload(ie),M._actorIdentifier=ge}),X&&X.length!==0&&(M._incrementalPayloadsPending=M._state!=="loading_final",X.forEach(function(ie){var de,ge=M._actorIdentifier;M._actorIdentifier=(de=ie.actorIdentifier)!==null&&de!==void 0?de:M._actorIdentifier,M._processIncrementalPlaceholder(G,ie),M._actorIdentifier=ge}),M._isClientPayload||M._state==="loading_final")){var oe=[];X.forEach(function(ie){ie.kind==="defer"&&oe.push(M._processDeferResponse(ie.label,ie.path,ie,{data:ie.data,extensions:{is_final:!0}}))}),oe.length>0&&M._processPayloadFollowups(oe)}})},J._maybeCompleteSubscriptionOperationTracking=function(){!this._isSubscriptionOperation&&!(this._useExecTimeResolvers&&this._execTimeResolverResponseComplete&&this._state==="loading_final")||this._pendingModulePayloadsCount===0&&this._incrementalPayloadsPending===!1&&this._completeOperationTracker()},J._processFollowupPayload=function(H){var M=this;switch(H.kind){case"ModuleImportPayload":var G=this._expectOperationLoader(),X=G.get(H.operationReference);if(X!=null)this._processFollowupPayloadWithNormalizationNode(H,a(X));else{var B=this._nextSubscriptionId++;this._pendingModulePayloadsCount++;var ne=function(){M._pendingModulePayloadsCount--,M._maybeCompleteSubscriptionOperationTracking()},oe=r.from(new Promise(function(ie,de){G.load(H.operationReference).then(ie,de)}));r.create(function(ie){var de,ge=oe.subscribe({next:function(me){if(me!=null){var Ee=function(){try{var ve=a(me),we=l.BATCH_ASYNC_MODULE_UPDATES_FN,Me=we!=null&&M._pendingModulePayloadsCount>1,Re=c(function(){if(M._handleFollowupPayload(H,ve),Me)M._scheduleAsyncStoreUpdate(we,ie.complete);else{var Oe=M._runPublishQueue();M._updateOperationTracker(Oe)}}),et=Re[0],tt=Re[1];M._log({name:"execute.async.module",executeId:M._executeId,operationName:ve.name,duration:tt}),Me||ie.complete()}catch(Oe){ie.error(Oe)}},se=M._scheduler;se==null?Ee():de=se.schedule(Ee)}else ie.complete()},error:ie.error});return function(){ge.unsubscribe(),M._scheduler!=null&&de!=null&&M._scheduler.cancel(de)}}).subscribe({complete:function(){M._complete(B),ne()},error:function(de){M._error(de),ne()},start:function(de){return M._start(B,de)}})}break;case"ActorPayload":this._processFollowupPayloadWithNormalizationNode(H,H.node);break;default:I(!1)}},J._processFollowupPayloadWithNormalizationNode=function(H,M){this._handleFollowupPayload(H,M),this._maybeCompleteSubscriptionOperationTracking()},J._handleFollowupPayload=function(H,M){var G=this._normalizeFollowupPayload(H,M);this._getPublishQueueAndSaveActor().commitPayload(this._operation,G),this._processPayloadFollowups([G])},J._processIncrementalPlaceholder=function(H,M){var G,X=M.label,B=M.path,ne=B.map(String).join("."),oe=this._incrementalResults.get(X);oe==null&&(oe=new Map,this._incrementalResults.set(X,oe));var ie=oe.get(ne),de=ie!=null&&ie.kind==="response"?ie.responses:null;oe.set(ne,{kind:"placeholder",placeholder:M});var ge;M.kind==="stream"?ge=M.parentID:M.kind==="defer"?ge=M.selector.dataID:I(!1);var ye=H.source.get(ge),me=((G=H.fieldPayloads)!==null&&G!==void 0?G:[]).filter(function(Re){var et=h(Re.dataID,Re.fieldKey);return Re.dataID===ge||et===ge});ye==null&&I(!1);var Ee,se,pe=this._source.get(ge);if(pe!=null){Ee=S.update(pe.record,ye);var ve=new Map,we=function(et){var tt=j(et);ve.set(tt,et)};pe.fieldPayloads.forEach(we),me.forEach(we),se=Array.from(ve.values())}else Ee=ye,se=me;if(this._source.set(ge,{record:Ee,fieldPayloads:se}),de!=null){var Me=this._processIncrementalResponses(de);this._processPayloadFollowups(Me)}},J._processIncrementalResponses=function(H){var M=this,G=[];return H.forEach(function(X){var B=X.label,ne=X.path,oe=X.response,ie=M._incrementalResults.get(B);if(ie==null&&(ie=new Map,M._incrementalResults.set(B,ie)),B.indexOf("$defer$")!==-1){var de=ne.map(String).join("."),ge=ie.get(de);if(ge==null){ge={kind:"response",responses:[X]},ie.set(de,ge);return}else if(ge.kind==="response"){ge.responses.push(X);return}var ye=ge.placeholder;ye.kind!=="defer"&&I(!1),G.push(M._processDeferResponse(B,ne,ye,oe))}else{var me=ne.slice(0,-2).map(String).join("."),Ee=ie.get(me);if(Ee==null){Ee={kind:"response",responses:[X]},ie.set(me,Ee);return}else if(Ee.kind==="response"){Ee.responses.push(X);return}var se=Ee.placeholder;se.kind!=="stream"&&I(!1),G.push(M._processStreamResponse(B,ne,se,oe))}}),G},J._processDeferResponse=function(H,M,G,X){var B,ne=G.selector.dataID,oe=this._actorIdentifier;this._actorIdentifier=(B=G.actorIdentifier)!==null&&B!==void 0?B:this._actorIdentifier;var ie=this._normalizeResponse(X,G.selector,G.typeName,{actorIdentifier:this._actorIdentifier,getDataID:this._getDataID,log:this._log,path:G.path,treatMissingFieldsAsNull:this._treatMissingFieldsAsNull,shouldProcessClientComponents:this._shouldProcessClientComponents},this._useExecTimeResolvers);this._getPublishQueueAndSaveActor().commitPayload(this._operation,ie);var de=this._source.get(ne);de==null&&I(!1);var ge=de.fieldPayloads;if(ge.length!==0){var ye,me={errors:null,fieldPayloads:ge,incrementalPlaceholders:null,followupPayloads:null,source:k.create(),isFinal:((ye=X.extensions)===null||ye===void 0?void 0:ye.is_final)===!0};this._getPublishQueueAndSaveActor().commitPayload(this._operation,me)}return this._actorIdentifier=oe,ie},J._processStreamResponse=function(H,M,G,X){var B=G.parentID,ne=G.node,oe=G.variables,ie=G.actorIdentifier,de=this._actorIdentifier;this._actorIdentifier=ie??this._actorIdentifier;var ge=ne.selections[0];ge!=null&&ge.kind==="LinkedField"&&ge.plural===!0||I(!1);var ye=this._normalizeStreamItem(X,B,ge,oe,M,G.path),me=ye.fieldPayloads,Ee=ye.itemID,se=ye.itemIndex,pe=ye.prevIDs,ve=ye.relayPayload,we=ye.storageKey;if(this._getPublishQueueAndSaveActor().commitPayload(this._operation,ve,function(Re){var et=Re.get(B);if(et!=null){var tt=et.getLinkedRecords(we);if(tt!=null&&!(tt.length!==pe.length||tt.some(function($e,Ae){return pe[Ae]!==($e&&$e.getDataID())}))){var Oe=(0,n.default)(tt);Oe[se]=Re.get(Ee),et.setLinkedRecords(Oe,we)}}}),me.length!==0){var Me={errors:null,fieldPayloads:me,incrementalPlaceholders:null,followupPayloads:null,source:k.create(),isFinal:!1};this._getPublishQueueAndSaveActor().commitPayload(this._operation,Me)}return this._actorIdentifier=de,ve},J._normalizeStreamItem=function(H,M,G,X,B,ne){var oe,ie,de,ge,ye=H.data;typeof ye!="object"&&I(!1);var me=(oe=G.alias)!==null&&oe!==void 0?oe:G.name,Ee=F(G,X),se=this._source.get(M);se==null&&I(!1);var pe=se.record,ve=se.fieldPayloads,we=S.getLinkedRecordIDs(pe,Ee);we==null&&I(!1);var Me=B[B.length-1],Re=parseInt(Me,10);Re===Me&&Re>=0||I(!1);var et=(ie=G.concreteType)!==null&&ie!==void 0?ie:ye[R];typeof et!="string"&&I(!1);var tt=(de=(ge=this._getDataID(ye,et))!==null&&ge!==void 0?ge:we==null?void 0:we[Re])!==null&&de!==void 0?de:h(M,Ee,Re);typeof tt!="string"&&I(!1);var Oe=E(G,tt,X),$e=S.clone(pe),Ae=(0,n.default)(we);Ae[Re]=tt,S.setLinkedRecordIDs($e,Ee,Ae),this._source.set(M,{record:$e,fieldPayloads:ve});var He=this._normalizeResponse(H,Oe,et,{actorIdentifier:this._actorIdentifier,getDataID:this._getDataID,log:this._log,path:[].concat((0,n.default)(ne),[me,String(Re)]),treatMissingFieldsAsNull:this._treatMissingFieldsAsNull,shouldProcessClientComponents:this._shouldProcessClientComponents},this._useExecTimeResolvers);return{fieldPayloads:ve,itemID:tt,itemIndex:Re,prevIDs:we,relayPayload:He,storageKey:Ee}},J._scheduleAsyncStoreUpdate=function(H,M){var G=this;this._completeFns.push(M),this._asyncStoreUpdateDisposable==null&&(this._asyncStoreUpdateDisposable=H(function(){G._asyncStoreUpdateDisposable=null;var X=G._runPublishQueue();G._updateOperationTracker(X);var B=(0,t.default)(G._completeFns),ne;try{for(B.s();!(ne=B.n()).done;){var oe=ne.value;oe()}}catch(ie){B.e(ie)}finally{B.f()}G._completeFns=[]}))},J._updateOperationTracker=function(H){H!=null&&H.length>0&&this._operationTracker.update(this._operation.request,new Set(H))},J._completeOperationTracker=function(){this._operationTracker.complete(this._operation.request)},J._getPublishQueueAndSaveActor=function(){return this._seenActors.add(this._actorIdentifier),this._getPublishQueue(this._actorIdentifier)},J._getActorsToVisit=function(){return this._seenActors.size===0?new Set([this._actorIdentifier]):this._seenActors},J._runPublishQueue=function(H){var M=new Set,G=(0,t.default)(this._getActorsToVisit()),X;try{for(G.s();!(X=G.n()).done;){var B=X.value,ne=this._getPublishQueue(B).run(H);ne.forEach(function(oe){return M.add(oe)})}}catch(oe){G.e(oe)}finally{G.f()}return Array.from(M)},J._retainData=function(){var H=(0,t.default)(this._getActorsToVisit()),M;try{for(H.s();!(M=H.n()).done;){var G=M.value;this._retainDisposables.has(G)||this._retainDisposables.set(G,this._getStore(G).retain(this._operation))}}catch(X){H.e(X)}finally{H.f()}},J._disposeRetainedData=function(){var H=(0,t.default)(this._retainDisposables.values()),M;try{for(H.s();!(M=H.n()).done;){var G=M.value;G.dispose()}}catch(X){H.e(X)}finally{H.f()}this._retainDisposables.clear()},J._expectOperationLoader=function(){var H=this._operationLoader;return H||I(!1),H},K})();function N(K){var J=[],te=[],H=[];return K.forEach(function(M){var G;if(M.path!=null||M.label!=null){var X=M.label,B=M.path;(X==null||B==null)&&I(!1),te.push({label:X,path:B,response:M})}else((G=M.extensions)===null||G===void 0?void 0:G.is_normalized)===!0?H.push(M):J.push(M)}),[J,te,H]}function j(K){var J;return(J=JSON.stringify(s(K)))!==null&&J!==void 0?J:""}function W(K){var J=K.incrementalPlaceholders;J!=null&&J.length!==0&&I(!1)}return zR={execute:q},zR}var qR,Hz;function v4e(){if(Hz)return qR;Hz=1;var e=Promise.resolve();function t(r){e.then(r).catch(n)}function n(r){setTimeout(function(){throw r},0)}return qR=t,qR}var VR,Xz;function JJ(){if(Xz)return VR;Xz=1;var e=hf(),t=RC(),n=t.EXISTENT,r=un(),i=(function(){function a(l,o){this.__sources=[o,l],this._base=l,this._sink=o}var u=a.prototype;return u.unstable_getRawRecordWithChanges=function(o){var s=this._base.get(o),c=this._sink.get(o);if(c===void 0){if(s==null)return s;var d=e.clone(s);return d}else{if(c===null)return null;if(s!=null){var h=e.update(s,c);return h}else{var m=e.clone(c);return m}}},u._getSinkRecord=function(o){var s=this._sink.get(o);if(!s){var c=this._base.get(o);c||r(!1),s=e.create(o,e.getType(c)),this._sink.set(o,s)}return s},u.copyFields=function(o,s){var c=this._sink.get(o),d=this._base.get(o);c||d||r(!1);var h=this._getSinkRecord(s);d&&e.copyFields(d,h),c&&e.copyFields(c,h)},u.copyFieldsFromRecord=function(o,s){var c=this._getSinkRecord(s);e.copyFields(o,c)},u.create=function(o,s){this._base.getStatus(o)!==n&&this._sink.getStatus(o)!==n||r(!1);var c=e.create(o,s);this._sink.set(o,c)},u.delete=function(o){this._sink.delete(o)},u.getStatus=function(o){return this._sink.has(o)?this._sink.getStatus(o):this._base.getStatus(o)},u.getType=function(o){for(var s=0;s<this.__sources.length;s++){var c=this.__sources[s].get(o);if(c)return e.getType(c);if(c===null)return null}},u.getValue=function(o,s){for(var c=0;c<this.__sources.length;c++){var d=this.__sources[c].get(o);if(d){var h=e.getValue(d,s);if(h!==void 0)return h}else if(d===null)return null}},u.setValue=function(o,s,c){var d=this._getSinkRecord(o);e.setValue(d,s,c)},u.getErrors=function(o,s){for(var c=0;c<this.__sources.length;c++){var d=this.__sources[c].get(o);if(d){var h=e.getErrors(d,s);if(h!==void 0)return h}else if(d===null)return null}},u.setErrors=function(o,s,c){var d=this._getSinkRecord(o);e.setErrors(d,s,c)},u.getLinkedRecordID=function(o,s){for(var c=0;c<this.__sources.length;c++){var d=this.__sources[c].get(o);if(d){var h=e.getLinkedRecordID(d,s);if(h!==void 0)return h}else if(d===null)return null}},u.setLinkedRecordID=function(o,s,c){var d=this._getSinkRecord(o);e.setLinkedRecordID(d,s,c)},u.getLinkedRecordIDs=function(o,s){for(var c=0;c<this.__sources.length;c++){var d=this.__sources[c].get(o);if(d){var h=e.getLinkedRecordIDs(d,s);if(h!==void 0)return h}else if(d===null)return null}},u.setLinkedRecordIDs=function(o,s,c){var d=this._getSinkRecord(o);e.setLinkedRecordIDs(d,s,c)},a})();return VR=i,VR}var UR,Yz;function eee(){if(Yz)return UR;Yz=1;var e=$n().default,t=e(Cu()),n=Da(),r=n.getArgumentValues,i=["id","__id","__typename","js"];function a(y,v,S,_,E){var $={};return u($,y,v,S,_,E),$}function u(y,v,S,_,E,$){var k,A=(0,t.default)(_),w;try{var R=function(){var q=w.value;switch(q.kind){case"LinkedField":q.plural?Object.defineProperty(y,(F=q.alias)!==null&&F!==void 0?F:q.name,{get:s(q,S,v,E,$),set:l(q,S,v,E)}):Object.defineProperty(y,(I=q.alias)!==null&&I!==void 0?I:q.name,{get:c(q,S,v,E,$),set:o(q,S,v,E)});break;case"ScalarField":var V=(k=q.alias)!==null&&k!==void 0?k:q.name;Object.defineProperty(y,V,{get:function(){var j,W=r((j=q.args)!==null&&j!==void 0?j:[],S),K=v.getValue(q.name,W);return K==null&&(K=m(q,W,v,E,$)),K},set:i.includes(q.name)?void 0:function(N){var j,W=r((j=q.args)!==null&&j!==void 0?j:[],S);v.setValue__UNSAFE(N,q.name,W)}});break;case"InlineFragment":v.getType()===q.type&&u(y,v,S,q.selections,E,$);break;case"ClientExtension":u(y,v,S,q.selections,E,$);break;case"FragmentSpread":break;case"Condition":case"ActorChange":case"InlineDataFragmentSpread":case"AliasedInlineFragmentSpread":case"ClientEdgeToClientObject":case"ClientEdgeToServerObject":case"Defer":case"ModuleImport":case"RequiredField":case"CatchField":case"Stream":case"RelayResolver":case"RelayLiveResolver":throw new Error("Encountered an unexpected ReaderSelection variant in RelayRecordSourceProxy. This indicates a bug in Relay.");default:throw q.kind,new Error("Encountered an unexpected ReaderSelection variant in RelayRecordSourceProxy. This indicates a bug in Relay.")}};for(A.s();!(w=A.n()).done;){var F,I;R()}}catch(U){A.e(U)}finally{A.f()}}function l(y,v,S,_){return function($){var k,A=r((k=y.args)!==null&&k!==void 0?k:[],v);if($==null)throw new Error("Do not assign null to plural linked fields; assign an empty array instead.");var w=$.map(function(R){if(R==null)throw new Error("When assigning an array of items, none of the items should be null or undefined.");var F=R.__id;if(F==null)throw new Error("The __id field must be present on each item passed to the setter. This indicates a bug in Relay.");var I=_.get(F);if(I==null)throw new Error("Did not find item with data id ".concat(F," in the store."));return I});S.setLinkedRecords(w,y.name,A)}}function o(y,v,S,_){return function($){var k,A=r((k=y.args)!==null&&k!==void 0?k:[],v);if($==null)S.setValue($,y.name,A);else{var w=$.__id;if(w==null)throw new Error("The __id field must be present on the argument. This indicates a bug in Relay.");var R=_.get(w);if(R==null)throw new Error("Did not find item with data id ".concat(w," in the store."));S.setLinkedRecord(R,y.name,A)}}}function s(y,v,S,_,E){return function(){var $,k=r(($=y.args)!==null&&$!==void 0?$:[],v),A=S.getLinkedRecords(y.name,k);return A===void 0&&(A=h(y,k,S,_,E)),A!=null?A.map(function(w){if(w!=null){var R={};return u(R,w,v,y.selections,_,E),R}else return w}):A}}function c(y,v,S,_,E){return function(){var $,k=r(($=y.args)!==null&&$!==void 0?$:[],v),A=S.getLinkedRecord(y.name,k);if(A===void 0&&(A=d(y,k,S,_,E)),A!=null){var w={};return u(w,A,v,y.selections,_,E),w}else return A}}function d(y,v,S,_,E){var $=(0,t.default)(E),k;try{for($.s();!(k=$.n()).done;){var A=k.value;if(A.kind==="linked"){var w=A.handle(y,S,v,_);if(w!=null)return _.get(w)}}}catch(R){$.e(R)}finally{$.f()}}function h(y,v,S,_,E){var $=(0,t.default)(E),k;try{for($.s();!(k=$.n()).done;){var A=k.value;if(A.kind==="pluralLinked"){var w=A.handle(y,S,v,_);if(w!=null)return w.map(function(R){if(R!=null)return _.get(R)})}}}catch(R){$.e(R)}finally{$.f()}}function m(y,v,S,_,E){var $=(0,t.default)(E),k;try{for($.s();!(k=$.n()).done;){var A=k.value;if(A.kind==="scalar"){var w=A.handle(y,S,v,_);if(w!==void 0)return w}}}catch(R){$.e(R)}finally{$.f()}}return UR={createUpdatableProxy:a},UR}var jR,Gz;function tee(){if(Gz)return jR;Gz=1;var e=dp(),t=e.getFragment,n=hp(),r=n.getVariablesFromFragment,i=Da(),a=i.ID_KEY,u=eee(),l=u.createUpdatableProxy,o=un();function s(c,d,h,m){var y=t(c),v=r(y,d),S=d[a],_=h.get(S);return _==null&&o(!1),{updatableData:l(_,v,y.selections,h,m)}}return jR={readUpdatableFragment:s},jR}var ZR,Kz;function nee(){if(Kz)return ZR;Kz=1;var e=dp(),t=e.getUpdatableQuery,n=eee(),r=n.createUpdatableProxy;function i(a,u,l,o){var s=t(a);return{updatableData:r(l.getRoot(),u,s.fragment.selections,l,o)}}return ZR={readUpdatableQuery:i},ZR}var HR,Wz;function y4e(){if(Wz)return HR;Wz=1;var e=N0(),t=e.generateClientID,n=Da(),r=n.getStableStorageKey,i=un(),a=(function(){function l(s,c,d){this._dataID=d,this._mutator=c,this._source=s}var o=l.prototype;return o.copyFieldsFrom=function(c){this._mutator.copyFields(c.getDataID(),this._dataID)},o.getDataID=function(){return this._dataID},o.getType=function(){var c=this._mutator.getType(this._dataID);return c==null&&i(!1),c},o.getValue=function(c,d){var h=r(c,d);return this._mutator.getValue(this._dataID,h)},o.setValue=function(c,d,h,m){return u(c)||i(!1),this.setValue__UNSAFE(c,d,h,m)},o.getErrors=function(c,d){var h=r(c,d);return this._mutator.getErrors(this._dataID,h)},o.setValue__UNSAFE=function(c,d,h,m){var y=r(d,h);return this._mutator.setValue(this._dataID,y,c),m!=null&&(m.length===0?this._mutator.setErrors(this._dataID,y):this._mutator.setErrors(this._dataID,y,m)),this},o.getLinkedRecord=function(c,d){var h=r(c,d),m=this._mutator.getLinkedRecordID(this._dataID,h);return m!=null?this._source.get(m):m},o.setLinkedRecord=function(c,d,h){c instanceof l||i(!1);var m=r(d,h),y=c.getDataID();return this._mutator.setLinkedRecordID(this._dataID,m,y),this},o.getOrCreateLinkedRecord=function(c,d,h){var m=this.getLinkedRecord(c,h);if(!m){var y,v=r(c,h),S=t(this.getDataID(),v);m=(y=this._source.get(S))!==null&&y!==void 0?y:this._source.create(S,d),this.setLinkedRecord(m,c,h)}return m},o.getLinkedRecords=function(c,d){var h=this,m=r(c,d),y=this._mutator.getLinkedRecordIDs(this._dataID,m);return y==null?y:y.map(function(v){return v!=null?h._source.get(v):v})},o.setLinkedRecords=function(c,d,h){Array.isArray(c)||i(!1);var m=r(d,h),y=c.map(function(v){return v&&v.getDataID()});return this._mutator.setLinkedRecordIDs(this._dataID,m,y),this},o.invalidateRecord=function(){this._source.markIDForInvalidation(this._dataID)},l})();function u(l){return l==null||typeof l!="object"||Array.isArray(l)&&l.every(u)}return HR=a,HR}var XR,Jz;function ree(){if(Jz)return XR;Jz=1;var e=hf(),t=RC(),n=t.EXISTENT,r=t.NONEXISTENT,i=Da(),a=i.ROOT_ID,u=i.ROOT_TYPE,l=tee(),o=l.readUpdatableFragment,s=nee(),c=s.readUpdatableQuery,d=y4e(),h=un(),m=(function(){function y(S,_,E,$,k){this.__mutator=S,this._handlerProvider=E||null,this._proxies={},this._getDataID=_,this._invalidatedStore=!1,this._idsMarkedForInvalidation=new Set,this._missingFieldHandlers=$,this._log=k??function(A){}}var v=y.prototype;return v.publishSource=function(_,E){var $=this,k=_.getRecordIDs();k.forEach(function(A){var w=_.getStatus(A);if(w===n){var R=_.get(A);R&&($.__mutator.getStatus(A)!==n&&$.create(A,e.getType(R)),$.__mutator.copyFieldsFromRecord(R,A))}else w===r&&$.delete(A)}),E&&E.length&&E.forEach(function(A){var w=$._handlerProvider&&$._handlerProvider(A.handle);w||h(!1),w.update($,A)})},v.create=function(_,E){this.__mutator.create(_,E),delete this._proxies[_];var $=this.get(_);return $||h(!1),$},v.delete=function(_){_===a&&h(!1),delete this._proxies[_],this.__mutator.delete(_)},v.get=function(_){if(!this._proxies.hasOwnProperty(_)){var E=this.__mutator.getStatus(_);E===n?this._proxies[_]=new d(this,this.__mutator,_):this._proxies[_]=E===r?null:void 0}return this._proxies[_]},v.getRoot=function(){var _=this.get(a);return _||(_=this.create(a,u)),_&&_.getType()===u||h(!1),_},v.invalidateStore=function(){this._invalidatedStore=!0},v.isStoreMarkedForInvalidation=function(){return this._invalidatedStore},v.markIDForInvalidation=function(_){this._idsMarkedForInvalidation.add(_)},v.getIDsMarkedForInvalidation=function(){return this._idsMarkedForInvalidation},v.readUpdatableQuery=function(_,E){return c(_,E,this,this._missingFieldHandlers)},v.readUpdatableFragment=function(_,E){return o(_,E,this,this._missingFieldHandlers)},y})();return XR=m,XR}var YR,eq;function iee(){if(eq)return YR;eq=1;var e=oy(),t=e.LINKED_FIELD,n=Da(),r=n.getHandleStorageKey,i=ed(),a=un();function u(l,o,s){var c=o.find(function(h){return h.kind===t&&h.name===l.name&&h.alias===l.alias&&i(h.args,l.args)});c&&c.kind===t||a(!1);var d=r(l,s);return{kind:"LinkedField",alias:c.alias,name:d,storageKey:d,args:null,concreteType:c.concreteType,plural:c.plural,selections:c.selections}}return YR=u,YR}var GR,tq;function b4e(){if(tq)return GR;tq=1;var e=oy(),t=e.SCALAR_FIELD,n=Da(),r=n.getHandleStorageKey,i=ed(),a=un();function u(l,o,s){var c=o.find(function(h){return h.kind===t&&h.name===l.name&&h.alias===l.alias&&i(h.args,l.args)});c&&c.kind===t||a(!1);var d=r(l,s);return{kind:"ScalarField",alias:c.alias,name:d,storageKey:d,args:null}}return GR=u,GR}var KR,nq;function O4e(){if(nq)return KR;nq=1;var e=$n().default,t=e(Cu()),n=JJ(),r=ree(),i=YF(),a=N0(),u=a.isClientID,l=iee(),o=b4e(),s=cy(),c=s.getLocalVariables,d=hf(),h=RC(),m=h.EXISTENT,y=h.UNKNOWN,v=Da(),S=AC(),_=S.TYPE_SCHEMA_TYPE,E=S.generateTypeID,$=un(),k=v.getModuleOperationKey,A=v.getStorageKey,w=v.getArgumentValues;function R(I,U,q,V,N,j,W,K,J,te){J!=null&&J({name:"store.datachecker.start",selector:V});var H=V.dataID,M=V.node,G=V.variables,X=new F(I,U,q,G,N,j,W,K,J,te),B=X.check(M,H);return J!=null&&J({name:"store.datachecker.end",selector:V}),B}var F=(function(){function I(q,V,N,j,W,K,J,te,H,M){this._getSourceForActor=q,this._getTargetForActor=V,this._getDataID=J,this._source=q(N),this._mutatorRecordSourceProxyCache=new Map;var G=this._getMutatorAndRecordProxyForActor(N),X=G[0],B=G[1];this._useExecTimeResolvers=M??!1,this._mostRecentlyInvalidatedAt=null,this._handlers=W,this._mutator=X,this._operationLoader=K??null,this._recordSourceProxy=B,this._recordWasMissing=!1,this._variables=j,this._shouldProcessClientComponents=te,this._log=H}var U=I.prototype;return U._getMutatorAndRecordProxyForActor=function(V){var N=this._mutatorRecordSourceProxyCache.get(V);if(N==null){var j=this._getTargetForActor(V),W=new n(this._getSourceForActor(V),j),K=new r(W,this._getDataID,void 0,this._handlers,this._log);N=[W,K],this._mutatorRecordSourceProxyCache.set(V,N)}return N},U.check=function(V,N){return this._assignClientAbstractTypes(V),this._traverse(V,N),this._recordWasMissing===!0?{status:"missing",mostRecentlyInvalidatedAt:this._mostRecentlyInvalidatedAt}:{status:"available",mostRecentlyInvalidatedAt:this._mostRecentlyInvalidatedAt}},U._getVariableValue=function(V){return this._variables.hasOwnProperty(V)||$(!1),this._variables[V]},U._handleMissing=function(){this._recordWasMissing=!0},U._handleMissingScalarField=function(V,N){if(!(V.name==="id"&&V.alias==null&&u(N))){var j=V.args!=null?w(V.args,this._variables):{},W=(0,t.default)(this._handlers),K;try{for(W.s();!(K=W.n()).done;){var J=K.value;if(J.kind==="scalar"){var te=J.handle(V,this._recordSourceProxy.get(N),j,this._recordSourceProxy);if(te!==void 0)return te}}}catch(H){W.e(H)}finally{W.f()}this._handleMissing()}},U._handleMissingLinkField=function(V,N){var j=V.args!=null?w(V.args,this._variables):{},W=(0,t.default)(this._handlers),K;try{for(W.s();!(K=W.n()).done;){var J=K.value;if(J.kind==="linked"){var te=J.handle(V,this._recordSourceProxy.get(N),j,this._recordSourceProxy);if(te!==void 0&&(te===null||this._mutator.getStatus(te)===m))return te}}}catch(H){W.e(H)}finally{W.f()}this._handleMissing()},U._handleMissingPluralLinkField=function(V,N){var j=this,W=V.args!=null?w(V.args,this._variables):{},K=(0,t.default)(this._handlers),J;try{for(K.s();!(J=K.n()).done;){var te=J.value;if(te.kind==="pluralLinked"){var H=te.handle(V,this._recordSourceProxy.get(N),W,this._recordSourceProxy);if(H!=null){var M=H.every(function(G){return G!=null&&j._mutator.getStatus(G)===m});if(M)return H}else if(H===null)return null}}}catch(G){K.e(G)}finally{K.f()}this._handleMissing()},U._traverse=function(V,N){var j=this._mutator.getStatus(N);if(j===y&&this._handleMissing(),j===m){var W=this._source.get(N),K=d.getInvalidationEpoch(W);K!=null&&(this._mostRecentlyInvalidatedAt=this._mostRecentlyInvalidatedAt!=null?Math.max(this._mostRecentlyInvalidatedAt,K):K),this._traverseSelections(V.selections,N)}},U._traverseSelections=function(V,N){var j=this;V.forEach(function(W){switch(W.kind){case"ScalarField":j._checkScalar(W,N);break;case"LinkedField":W.plural?j._checkPluralLink(W,N):j._checkLink(W,N);break;case"ActorChange":j._checkActorChange(W.linkedField,N);break;case"Condition":var K=!!j._getVariableValue(W.condition);K===W.passingValue&&j._traverseSelections(W.selections,N);break;case"InlineFragment":{var J=W.abstractKey;if(J==null){var te=j._mutator.getType(N);te===W.type&&j._traverseSelections(W.selections,N)}else{var H=j._mutator.getType(N);H==null&&$(!1);var M=E(H),G=j._mutator.getValue(M,J);G===!0?j._traverseSelections(W.selections,N):G==null&&j._handleMissing()}break}case"LinkedHandle":{var X=l(W,V,j._variables);X.plural?j._checkPluralLink(X,N):j._checkLink(X,N);break}case"ScalarHandle":{var B=o(W,V,j._variables);j._checkScalar(B,N);break}case"ModuleImport":j._checkModuleImport(W,N);break;case"Defer":case"Stream":j._traverseSelections(W.selections,N);break;case"FragmentSpread":var ne=j._variables;j._variables=c(j._variables,W.fragment.argumentDefinitions,W.args),j._traverseSelections(W.fragment.selections,N),j._variables=ne;break;case"ClientExtension":var oe=j._recordWasMissing;j._traverseSelections(W.selections,N),j._recordWasMissing=oe;break;case"TypeDiscriminator":var ie=W.abstractKey,de=j._mutator.getType(N);de==null&&$(!1);var ge=E(de),ye=j._mutator.getValue(ge,ie);ye==null&&j._handleMissing();break;case"ClientComponent":if(j._shouldProcessClientComponents===!1)break;j._traverseSelections(W.fragment.selections,N);break;case"RelayResolver":case"RelayLiveResolver":j._useExecTimeResolvers||j._checkResolver(W,N);break;case"ClientEdgeToClientObject":j._useExecTimeResolvers||j._checkResolver(W.backingField,N);break;default:$(!1)}})},U._checkResolver=function(V,N){V.fragment&&this._traverseSelections([V.fragment],N)},U._checkModuleImport=function(V,N){var j=this._operationLoader;j===null&&$(!1);var W=k(V.documentName),K=this._mutator.getValue(N,W);if(K==null){K===void 0&&this._handleMissing();return}var J=j.get(K);if(J!=null){var te=i(J),H=this._variables;this._variables=c(this._variables,te.argumentDefinitions,V.args),this._traverse(te,N),this._variables=H}else this._handleMissing()},U._checkScalar=function(V,N){var j=A(V,this._variables),W=this._mutator.getValue(N,j);W===void 0&&(W=this._handleMissingScalarField(V,N),W!==void 0&&this._mutator.setValue(N,j,W))},U._checkLink=function(V,N){var j=A(V,this._variables),W=this._mutator.getLinkedRecordID(N,j);W===void 0&&(W=this._handleMissingLinkField(V,N),W!=null?this._mutator.setLinkedRecordID(N,j,W):W===null&&this._mutator.setValue(N,j,null)),W!=null&&this._traverse(V,W)},U._checkPluralLink=function(V,N){var j=this,W=A(V,this._variables),K=this._mutator.getLinkedRecordIDs(N,W);K===void 0&&(K=this._handleMissingPluralLinkField(V,N),K!=null?this._mutator.setLinkedRecordIDs(N,W,K):K===null&&this._mutator.setValue(N,W,null)),K&&K.forEach(function(J){J!=null&&j._traverse(V,J)})},U._checkActorChange=function(V,N){var j=A(V,this._variables),W=this._source.get(N),K=W!=null?d.getActorLinkedRecordID(W,j):W;if(K==null)K===void 0&&this._handleMissing();else{var J=K[0],te=K[1],H=this._source,M=this._mutator,G=this._recordSourceProxy,X=this._getMutatorAndRecordProxyForActor(J),B=X[0],ne=X[1];this._source=this._getSourceForActor(J),this._mutator=B,this._recordSourceProxy=ne,this._assignClientAbstractTypes(V),this._traverse(V,te),this._source=H,this._mutator=M,this._recordSourceProxy=G}},U._assignClientAbstractTypes=function(V){var N=V.clientAbstractTypes;if(N!=null)for(var j=0,W=Object.keys(N);j<W.length;j++){var K=W[j],J=(0,t.default)(N[K]),te;try{for(J.s();!(te=J.n()).done;){var H=te.value,M=E(H);this._source.get(M)==null&&this._mutator.create(M,_),this._mutator.getValue(M,K)==null&&this._mutator.setValue(M,K,!0)}}catch(G){J.e(G)}finally{J.f()}}},I})();return KR={check:R},KR}var WR,rq;function x4e(){return rq||(rq=1,WR=function(t){typeof t=="object"&&t!=null&&(Array.isArray(t)||t.constructor===Object)&&Object.freeze(t)}),WR}var JR,iq;function aee(){if(iq)return JR;iq=1;var e=hf(),t=Da(),n=t.RELAY_RESOLVER_OUTPUT_TYPE_RECORD_IDS,r=un();function i(a){var u=e.getValue(a,n);return u==null?null:(u instanceof Set||r(!1),u)}return JR=i,JR}var eA,aq;function S4e(){if(aq)return eA;aq=1;function e(t){return t!=null&&typeof t=="object"&&typeof t.read=="function"&&typeof t.subscribe=="function"}return eA=e,eA}var tA,uq;function _4e(){if(uq)return tA;uq=1;var e=$n().default,t=e(Cu()),n=CC(),r=oy(),i=r.RELAY_LIVE_RESOLVER,a=nh(),u=x4e(),l=N0(),o=l.generateClientID,s=l.generateClientObjectClientID,c=hf(),d=hp(),h=d.createNormalizationSelector,m=E1(),y=HJ(),v=y.normalize,S=Da(),_=S.RELAY_RESOLVER_ERROR_KEY,E=S.RELAY_RESOLVER_INVALIDATION_KEY,$=S.RELAY_RESOLVER_OUTPUT_TYPE_RECORD_IDS,k=S.RELAY_RESOLVER_SNAPSHOT_KEY,A=S.RELAY_RESOLVER_VALUE_KEY,w=S.getReadTimeResolverStorageKey,R=S.getStorageKey,F=aee(),I=S4e(),U=wC(),q=U.isSuspenseSentinel,V=un();Vi();var N="__resolverLiveStateSubscription",j="__resolverLiveStateValue",W="__resolverLiveStateDirty",K="__RELAY_RESOLVER__",J="__relay_model_instance";function te(me,Ee,se){var pe=me.get(Ee);pe||(pe=new Set,me.set(Ee,pe)),pe.add(se)}var H=(function(){function me(se,pe){this._resolverIDToRecordIDs=new Map,this._recordIDToResolverIDs=new Map,this._getRecordSource=se,this._store=pe,this._handlingBatch=!1,this._liveResolverBatchRecordSource=null}var Ee=me.prototype;return Ee.readFromCacheOrEvaluate=function(pe,ve,we,Me,Re){var et=this._getRecordSource(),tt=de(et,pe),Oe=w(ve,we),$e=c.getLinkedRecordID(tt,Oe),Ae=$e==null?null:et.get($e),He;if(Ae==null||this._isInvalid(Ae,Re)){var je;Ae!=null&&ie(Ae),$e=(je=$e)!==null&&je!==void 0?je:o(pe,Oe),Ae=c.create($e,K);var Ye=Me();if(c.setValue(Ae,k,Ye.snapshot),c.setValue(Ae,_,Ye.error),ve.kind===i){if(Ye.resolverResult!=null){Ye.error!=null&&V(!1);var Ne=Ye.resolverResult;He=this._setLiveStateValue(Ae,$e,Ne,ve,we)}}else He=this._setResolverValue(Ae,Ye.resolverResult,ve,we);et.set($e,Ae);var ze=de(et,pe),at=c.clone(ze);if(c.setLinkedRecordID(at,Oe,$e),et.set(pe,at),ve.fragment!=null){var mt,Mt=R(ve.fragment,we),bt=o(pe,Mt);te(this._resolverIDToRecordIDs,bt,$e),te(this._recordIDToResolverIDs,pe,bt);var zt=(mt=Ye.snapshot)===null||mt===void 0?void 0:mt.seenRecords;if(zt!=null){var qt=(0,t.default)(zt),Xt;try{for(qt.s();!(Xt=qt.n()).done;){var nn=Xt.value;te(this._recordIDToResolverIDs,nn,bt)}}catch(Xn){qt.e(Xn)}finally{qt.f()}}}}else if(ve.kind===i&&c.getValue(Ae,W)){var Fe;$e=(Fe=$e)!==null&&Fe!==void 0?Fe:o(pe,Oe),Ae=c.clone(Ae);var ht=c.getValue(Ae,j);I(ht)||V(!1),He=this._setLiveResolverValue(Ae,ht,ve,we),c.setValue(Ae,W,!1),et.set($e,Ae)}var kt=this._getResolverValue(Ae),Qt=c.getValue(Ae,k),en=c.getValue(Ae,_),In=null;if(q(kt)){var rn;In=(rn=$e)!==null&&rn!==void 0?rn:o(pe,Oe)}return[kt,$e,en,Qt,In,He]},Ee.getLiveResolverPromise=function(pe){var ve=this._getRecordSource(),we=ve.get(pe);we==null&&V(!1);var Me=c.getValue(we,j);return new Promise(function(Re){var et=Me.subscribe(function(){et(),Re()})})},Ee._setLiveStateValue=function(pe,ve,we,Me,Re){var et=this._makeLiveStateHandler(ve),tt=we.subscribe(et);c.setValue(pe,j,we);var Oe=this._setLiveResolverValue(pe,we,Me,Re);return c.setValue(pe,W,!1),c.setValue(pe,N,tt),Oe},Ee._makeLiveStateHandler=function(pe){var ve=this;return function(){var we=ve._getRecordSource(),Me=we.get(pe);if(Me&&c.hasValue(Me,j)){var Re=c.clone(Me);c.setValue(Re,W,!0),ve._setLiveResolverUpdate(pe,Re)}}},Ee._setLiveResolverUpdate=function(pe,ve){if(this._handlingBatch)this._liveResolverBatchRecordSource==null&&(this._liveResolverBatchRecordSource=m.create()),this._liveResolverBatchRecordSource.set(pe,ve);else{var we=m.create();we.set(pe,ve),this._store.publish(we),this._store.notify()}},Ee.batchLiveStateUpdates=function(pe){this._handlingBatch&&V(!1),this._handlingBatch=!0;try{pe()}finally{this._liveResolverBatchRecordSource!=null&&(this._store.publish(this._liveResolverBatchRecordSource),this._store.notify()),this._liveResolverBatchRecordSource=null,this._handlingBatch=!1}},Ee._setLiveResolverValue=function(pe,ve,we,Me){var Re=null,et=null;try{Re=ve.read()}catch(tt){et=tt}return c.setValue(pe,_,et),this._setResolverValue(pe,Re,we,Me)},Ee._setResolverValue=function(pe,ve,we,Me){var Re=we.normalizationInfo,et=null;if(ve!=null&&Re!=null&&!q(ve)){var tt,Oe=F(pe),$e=new Set,Ae=this._getRecordSource();if(Re.plural){Array.isArray(ve)||V(!1),tt=[];for(var He=m.create(),je=0;je<ve.length;je++){var Ye=ve[je];if(Ye!=null){typeof Ye!="object"&&V(!1);var Ne=ye(Re,Ye),ze=s(Ne,c.getDataID(pe),je),at=this._normalizeOutputTypeValue(ze,Ye,Me,Re,[we.path,String(je)],Ne),mt=(0,t.default)(at.getRecordIDs()),Mt;try{for(mt.s();!(Mt=mt.n()).done;){var bt=Mt.value;He.set(bt,de(at,bt)),$e.add(bt)}}catch(kt){mt.e(kt)}finally{mt.f()}tt.push(ze)}}et=M(Ae,He,Oe)}else{typeof ve!="object"&&V(!1);var zt=ye(Re,ve),qt=s(zt,c.getDataID(pe)),Xt=this._normalizeOutputTypeValue(qt,ve,Me,Re,[we.path],zt),nn=(0,t.default)(Xt.getRecordIDs()),Fe;try{for(nn.s();!(Fe=nn.n()).done;){var ht=Fe.value;$e.add(ht)}}catch(kt){nn.e(kt)}finally{nn.f()}tt=qt,et=M(Ae,Xt,Oe)}c.setValue(pe,$,$e),u(tt),c.setValue(pe,A,tt)}else u(ve),c.setValue(pe,A,ve);return et},Ee.notifyUpdatedSubscribers=function(pe){this._store.__notifyUpdatedSubscribers(pe)},Ee._getResolverValue=function(pe){return c.getValue(pe,A)},Ee.invalidateDataIDs=function(pe){for(var ve=this._getRecordSource(),we=new Set,Me=Array.from(pe);Me.length;){var Re=Me.pop();we.add(Re),pe.add(Re);var et=this._recordIDToResolverIDs.get(Re);if(et!=null){var tt=(0,t.default)(et),Oe;try{for(tt.s();!(Oe=tt.n()).done;){var $e=Oe.value;if(!we.has($e)){we.add($e);var Ae=this._resolverIDToRecordIDs.get($e);if(Ae==null)continue;var He=(0,t.default)(Ae),je;try{for(He.s();!(je=He.n()).done;){var Ye=je.value;X(Ye,ve),we.has(Ye)||(we.add(Ye),Me.push(Ye))}}catch(Ne){He.e(Ne)}finally{He.f()}}}}catch(Ne){tt.e(Ne)}finally{tt.f()}}}},Ee._isInvalid=function(pe,ve){if(!c.getValue(pe,E))return!1;var we=c.getValue(pe,k),Me=we==null?void 0:we.data,Re=we==null?void 0:we.selector;if(Me==null||Re==null)return!0;var et=ve(Re),tt=et.data,Oe=n(Me,tt);if(Oe!==Me)return!0;if(a.MARK_RESOLVER_VALUES_AS_CLEAN_AFTER_FRAGMENT_REREAD){var $e=c.clone(pe);c.setValue($e,E,!1);var Ae=this._getRecordSource();Ae.set(c.getDataID(pe),$e)}return!1},Ee._normalizeOutputTypeValue=function(pe,ve,we,Me,Re,et){var tt=m.create();switch(Me.kind){case"OutputType":{var Oe=c.create(pe,et);tt.set(pe,Oe);var $e=h(Me.normalizationNode,pe,we),Ae=!1,He=this._store.__getNormalizationOptions(Re);return v(tt,$e,ve,He,void 0,Ae).source}case"WeakModel":{var je=c.create(pe,et);return c.setValue(je,J,ve),tt.set(pe,je),tt}default:Me.kind,V(!1)}},Ee.ensureClientRecord=function(pe,ve){var we=s(ve,pe),Me=this._getRecordSource();if(!Me.has(we)){var Re=c.create(we,ve);c.setValue(Re,"id",pe),Me.set(we,Re)}return we},Ee.unsubscribeFromLiveResolverRecords=function(pe){return ne(this._getRecordSource(),pe)},Ee.invalidateResolverRecords=function(pe){if(pe.size!==0){var ve=(0,t.default)(pe),we;try{for(ve.s();!(we=ve.n()).done;){var Me=we.value,Re=this._getRecordSource().get(Me);Re!=null&&oe(Re)&&this._getRecordSource().delete(Me)}}catch(et){ve.e(et)}finally{ve.f()}}},me})();function M(me,Ee,se){var pe=new Set;if(se!=null){var ve=(0,t.default)(se),we;try{for(ve.s();!(we=ve.n()).done;){var Me=we.value;Ee.has(Me)||(pe.add(Me),me.remove(Me))}}catch(He){ve.e(He)}finally{ve.f()}}var Re=(0,t.default)(Ee.getRecordIDs()),et;try{for(Re.s();!(et=Re.n()).done;){var tt=et.value,Oe=de(Ee,tt);if(me.has(tt)){var $e=de(me,tt),Ae=c.update($e,Oe);Ae!==$e&&(pe.add(tt),me.set(tt,Ae),B($e,me))}else me.set(tt,Oe)}}catch(He){Re.e(He)}finally{Re.f()}return pe}function G(me){var Ee=new Set;return c.getFields(me).forEach(function(se){if(c.hasLinkedRecordID(me,se)){var pe=c.getLinkedRecordID(me,se);pe!=null&&Ee.add(pe)}else if(c.hasLinkedRecordIDs(me,se)){var ve;(ve=c.getLinkedRecordIDs(me,se))===null||ve===void 0||ve.forEach(function(we){we!=null&&Ee.add(we)})}}),Ee}function X(me,Ee){var se=Ee.get(me);if(se){var pe=c.clone(se);c.setValue(pe,E,!0),Ee.set(me,pe)}}function B(me,Ee){var se=G(me),pe=(0,t.default)(se),ve;try{for(pe.s();!(ve=pe.n()).done;){var we=ve.value,Me=Ee.get(we);Me!=null&&oe(Me)&&X(we,Ee)}}catch(Re){pe.e(Re)}finally{pe.f()}}function ne(me,Ee){if(Ee.size!==0){var se=(0,t.default)(Ee),pe;try{for(se.s();!(pe=se.n()).done;){var ve=pe.value,we=me.get(ve);we!=null&&oe(we)&&ie(we)}}catch(Me){se.e(Me)}finally{se.f()}}}function oe(me){return c.getType(me)===K}function ie(me){var Ee=c.getValue(me,N);Ee!=null&&Ee()}function de(me,Ee){var se=me.get(Ee);return se==null&&V(!1),se}function ge(me){return me}function ye(me,Ee){var se,pe=(se=me.concreteType)!==null&&se!==void 0?se:Ee.__typename;return pe==null&&V(!1),pe}return tA={LiveResolverCache:H,getUpdatedDataIDs:ge,RELAY_RESOLVER_LIVE_STATE_SUBSCRIPTION_KEY:N},tA}var nA,lq;function E4e(){if(lq)return nA;lq=1;var e=$n().default,t=e(Ui()),n=hf(),r=E1(),i=un(),a=n.fromObject(Object.freeze({__UNPUBLISH_RECORD_SENTINEL:!0})),u=(function(){function s(d){this._base=d,this._sink=r.create()}var c=s.prototype;return c.has=function(h){if(this._sink.has(h)){var m=this._sink.get(h);return m!==a}else return this._base.has(h)},c.get=function(h){if(this._sink.has(h)){var m=this._sink.get(h);return m===a?void 0:m}else return this._base.get(h)},c.getStatus=function(h){var m=this.get(h);return m===void 0?"UNKNOWN":m===null?"NONEXISTENT":"EXISTENT"},c.clear=function(){this._base=r.create(),this._sink.clear()},c.delete=function(h){this._sink.delete(h)},c.remove=function(h){this._sink.set(h,a)},c.set=function(h,m){this._sink.set(h,m)},c.getRecordIDs=function(){return Object.keys(this.toJSON())},c.size=function(){return Object.keys(this.toJSON()).length},c.toJSON=function(){var h=this,m=(0,t.default)({},this._base.toJSON());return this._sink.getRecordIDs().forEach(function(y){var v=h.get(y);v===void 0?delete m[y]:m[y]=n.toJSON(v)}),m},c.getOptimisticRecordIDs=function(){return new Set(this._sink.getRecordIDs())},s})();function l(s){return new u(s)}function o(s){return s instanceof u||i(!1),s.getOptimisticRecordIDs()}return nA={create:l,getOptimisticRecordIDs:o},nA}var rA,oq;function D4e(){if(oq)return rA;oq=1;var e=oy(),t=e.RELAY_LIVE_RESOLVER,n=un(),r=(function(){function i(){}var a=i.prototype;return a.readFromCacheOrEvaluate=function(l,o,s,c,d){o.kind===t&&n(!1);var h=c(),m=h.resolverResult,y=h.snapshot,v=h.error;return[m,void 0,v,y,void 0,void 0]},a.invalidateDataIDs=function(l){},a.ensureClientRecord=function(l,o){n(!1)},a.notifyUpdatedSubscribers=function(l){},i})();return rA={NoopResolverCache:r},rA}var iA,sq;function GF(){if(sq)return iA;sq=1;var e=$n().default,t=e(Ui()),n=e(mc()),r=e(H3()),i=e(sy()),a=["message"],u=nh(),l=wC(),o=l.isSuspenseSentinel,s=cy(),c=hf(),d=Da(),h=d.CLIENT_EDGE_TRAVERSAL_PATH,m=d.FRAGMENT_OWNER_KEY,y=d.FRAGMENT_PROP_NAME_KEY,v=d.FRAGMENTS_KEY,S=d.ID_KEY,_=d.MODULE_COMPONENT_KEY,E=d.ROOT_ID,$=d.getArgumentValues,k=d.getModuleComponentKey,A=d.getStorageKey,w=D4e(),R=w.NoopResolverCache,F=VF(),I=F.RESOLVER_FRAGMENT_ERRORED_SENTINEL,U=F.withResolverContext,q=AC(),V=q.generateTypeID,N=un();function j(H,M,G,X){var B=new W(H,M,G??new R,X);return B.read()}var W=(function(){function H(G,X,B,ne){var oe,ie,de,ge;this._clientEdgeTraversalPath=(oe=X.clientEdgeTraversalPath)!==null&&oe!==void 0&&oe.length?(0,i.default)(X.clientEdgeTraversalPath):[],this._missingClientEdges=[],this._missingLiveResolverFields=[],this._isMissingData=!1,this._isWithinUnmatchedTypeRefinement=!1,this._fieldErrors=null,this._owner=X.owner,this._useExecTimeResolvers=(ie=(de=this._owner.node.operation.use_exec_time_resolvers)!==null&&de!==void 0?de:((ge=this._owner.node.operation.exec_time_resolvers_enabled_provider)===null||ge===void 0?void 0:ge.get())===!0)!==null&&ie!==void 0?ie:!1,this._recordSource=G,this._seenRecords=new Set,this._selector=X,this._variables=X.variables,this._resolverCache=B,this._fragmentName=X.node.name,this._updatedDataIDs=new Set,this._resolverContext=ne}var M=H.prototype;return M.read=function(){var X,B=this._selector,ne=B.node,oe=B.dataID,ie=B.isWithinUnmatchedTypeRefinement,de=ne.abstractKey,ge=this._recordSource.get(oe),ye=!ie;if(ye&&de==null&&ge!=null&&(this._recordMatchesTypeCondition(ge,ne.type)||(ye=!1)),ye&&de!=null&&ge!=null){var me=this._implementsInterface(ge,de);me===!1&&(ye=!1)}this._isWithinUnmatchedTypeRefinement=!ye;var Ee=this._traverse(ne,oe,null),se=(X=this._selector.node.metadata)===null||X===void 0?void 0:X.catchTo;return se!=null&&(Ee=this._catchErrors(Ee,se,null)),this._updatedDataIDs.size>0&&(this._resolverCache.notifyUpdatedSubscribers(this._updatedDataIDs),this._updatedDataIDs.clear()),{data:Ee,isMissingData:this._isMissingData&&ye,missingClientEdges:this._missingClientEdges.length?this._missingClientEdges:null,missingLiveResolverFields:this._missingLiveResolverFields,seenRecords:this._seenRecords,selector:this._selector,fieldErrors:this._fieldErrors}},M._maybeAddFieldErrors=function(X,B){var ne=c.getErrors(X,B);if(ne!=null){var oe=this._fragmentName;this._fieldErrors==null&&(this._fieldErrors=[]);for(var ie=0;ie<ne.length;ie++){var de,ge,ye,me=ne[ie];this._fieldErrors.push({kind:"relay_field_payload.error",owner:oe,fieldPath:((de=me.path)!==null&&de!==void 0?de:[]).join("."),error:me,shouldThrow:(ge=(ye=this._selector.node.metadata)===null||ye===void 0?void 0:ye.throwOnFieldError)!==null&&ge!==void 0?ge:!1,handled:!1,uiContext:void 0})}}},M._markDataAsMissing=function(X){var B,ne;if(!this._isWithinUnmatchedTypeRefinement){this._fieldErrors==null&&(this._fieldErrors=[]);var oe=this._fragmentName;if(this._fieldErrors.push((B=(ne=this._selector.node.metadata)===null||ne===void 0?void 0:ne.throwOnFieldError)!==null&&B!==void 0&&B?{kind:"missing_expected_data.throw",owner:oe,fieldPath:X,handled:!1,uiContext:void 0}:{kind:"missing_expected_data.log",owner:oe,fieldPath:X,uiContext:void 0}),this._isMissingData=!0,this._clientEdgeTraversalPath.length){var ie=this._clientEdgeTraversalPath[this._clientEdgeTraversalPath.length-1];ie!==null&&this._missingClientEdges.push({request:ie.readerClientEdge.operation,clientEdgeDestinationID:ie.clientEdgeDestinationID})}}},M._traverse=function(X,B,ne){var oe=this._recordSource.get(B);if(this._seenRecords.add(B),oe==null)return oe===void 0&&this._markDataAsMissing("<record>"),oe;var ie=ne||{},de=this._traverseSelections(X.selections,oe,ie);return de?ie:null},M._getVariableValue=function(X){return this._variables.hasOwnProperty(X)||N(!1),this._variables[X]},M._maybeReportUnexpectedNull=function(X){if(X.action!=="NONE"){var B=this._fragmentName;this._fieldErrors==null&&(this._fieldErrors=[]);var ne;if(X.field.linkedField!=null){var oe;ne=(oe=X.field.linkedField.alias)!==null&&oe!==void 0?oe:X.field.linkedField.name}else{var ie;ne=(ie=X.field.alias)!==null&&ie!==void 0?ie:X.field.name}switch(X.action){case"THROW":this._fieldErrors.push({kind:"missing_required_field.throw",fieldPath:ne,owner:B,handled:!1,uiContext:void 0});return;case"LOG":this._fieldErrors.push({kind:"missing_required_field.log",fieldPath:ne,owner:B,uiContext:void 0});return;default:X.action}}},M._handleRequiredFieldValue=function(X,B){return B==null?(this._maybeReportUnexpectedNull(X),!1):!0},M._catchErrors=function(X,B,ne){var oe=X;switch(B){case"RESULT":oe=this._asResult(X);break;case"NULL":this._fieldErrors!=null&&this._fieldErrors.length>0&&(oe=null);break}var ie=this._fieldErrors;if(this._fieldErrors=ne,ie!=null){this._fieldErrors==null&&(this._fieldErrors=[]);for(var de=0;de<ie.length;de++)this._fieldErrors.push(K(ie[de]))}return oe},M._asResult=function(X){if(this._fieldErrors==null||this._fieldErrors.length===0)return{ok:!0,value:X};var B=this._fieldErrors.map(function(ne){switch(ne.kind){case"relay_field_payload.error":var oe=ne.error;oe.message;var ie=(0,r.default)(oe,a);return ie;case"missing_expected_data.throw":case"missing_expected_data.log":return{path:ne.fieldPath.split(".")};case"relay_resolver.error":return{message:"Relay: Error in resolver for field at ".concat(ne.fieldPath," in ").concat(ne.owner)};case"missing_required_field.throw":return{message:"Relay: Missing @required value at path '".concat(ne.fieldPath,"' in '").concat(ne.owner,"'.")};case"missing_required_field.log":return null;default:ne.kind,N(!1)}}).filter(Boolean);return{ok:!1,errors:B}},M._traverseSelections=function(X,B,ne){for(var oe=0;oe<X.length;oe++){var ie=X[oe];switch(ie.kind){case"RequiredField":var de=this._readClientSideDirectiveField(ie,B,ne);if(!this._handleRequiredFieldValue(ie,de))return!1;break;case"CatchField":{var ge,ye,me,Ee=this._fieldErrors;this._fieldErrors=null;var se=this._readClientSideDirectiveField(ie,B,ne),pe=(ge=(ye=ie.field)===null||ye===void 0?void 0:ye.backingField)!==null&&ge!==void 0?ge:ie.field,ve=(me=pe==null?void 0:pe.alias)!==null&&me!==void 0?me:pe==null?void 0:pe.name;ve==null&&N(!1),ne[ve]=this._catchErrors(se,ie.to,Ee);break}case"ScalarField":this._readScalar(ie,B,ne);break;case"LinkedField":ie.plural?this._readPluralLink(ie,B,ne):this._readLink(ie,B,ne);break;case"Condition":var we=!!this._getVariableValue(ie.condition);if(we===ie.passingValue){var Me=this._traverseSelections(ie.selections,B,ne);if(!Me)return!1}break;case"InlineFragment":{var Re=this._readInlineFragment(ie,B,ne,!1);if(Re===!1)return!1;break}case"RelayLiveResolver":case"RelayResolver":{this._useExecTimeResolvers?this._readScalar(ie,B,ne):this._readResolverField(ie,B,ne);break}case"FragmentSpread":this._createFragmentPointer(ie,B,ne);break;case"AliasedInlineFragmentSpread":{this._readAliasedInlineFragment(ie,B,ne);break}case"ModuleImport":this._readModuleImport(ie,B,ne);break;case"InlineDataFragmentSpread":this._createInlineDataOrResolverFragmentPointer(ie,B,ne);break;case"Defer":case"ClientExtension":{var et=this._isMissingData,tt=this._missingClientEdges.length;this._clientEdgeTraversalPath.push(null);var Oe=this._traverseSelections(ie.selections,B,ne);if(this._isMissingData=et||this._missingClientEdges.length>tt||this._missingLiveResolverFields.length>0,this._clientEdgeTraversalPath.pop(),!Oe)return!1;break}case"Stream":{var $e=this._traverseSelections(ie.selections,B,ne);if(!$e)return!1;break}case"ActorChange":this._readActorChange(ie,B,ne);break;case"ClientEdgeToClientObject":case"ClientEdgeToServerObject":if(this._useExecTimeResolvers&&(ie.backingField.kind==="RelayResolver"||ie.backingField.kind==="RelayLiveResolver")){var Ae=ie.linkedField;Ae.plural?this._readPluralLink(Ae,B,ne):this._readLink(Ae,B,ne)}else this._readClientEdge(ie,B,ne);break;default:N(!1)}}return!0},M._readClientSideDirectiveField=function(X,B,ne){switch(X.field.kind){case"ScalarField":return this._readScalar(X.field,B,ne);case"LinkedField":return X.field.plural?this._readPluralLink(X.field,B,ne):this._readLink(X.field,B,ne);case"RelayResolver":case"RelayLiveResolver":return this._useExecTimeResolvers?this._readScalar(X.field,B,ne):this._readResolverField(X.field,B,ne);case"ClientEdgeToClientObject":case"ClientEdgeToServerObject":if(this._useExecTimeResolvers&&(X.field.backingField.kind==="RelayResolver"||X.field.backingField.kind==="RelayLiveResolver")){var oe=X.field;return oe.linkedField.plural?this._readPluralLink(oe.linkedField,B,ne):this._readLink(oe.linkedField,B,ne)}else return this._readClientEdge(X.field,B,ne);case"AliasedInlineFragmentSpread":return this._readAliasedInlineFragment(X.field,B,ne);default:X.field.kind,N(!1)}},M._readResolverField=function(X,B,ne){var oe,ie=c.getDataID(B),de=this._fieldErrors;this._fieldErrors=null;var ge=this._readResolverFieldImpl(X,ie),ye=(oe=X.alias)!==null&&oe!==void 0?oe:X.name;return this._prependPreviousErrors(de,ye),ne[ye]=ge,ge},M._readResolverFieldImpl=function(X,B){var ne=this,oe=X.fragment,ie,de=function(Re){return ie!=null?{data:ie.data,isMissingData:ie.isMissingData,fieldErrors:ie.fieldErrors}:(ie=j(ne._recordSource,Re,ne._resolverCache),{data:ie.data,isMissingData:ie.isMissingData,fieldErrors:ie.fieldErrors})},ge=function(){if(oe!=null){var Re={__id:B,__fragmentOwner:ne._owner,__fragments:(0,n.default)({},oe.name,oe.args?$(oe.args,ne._variables):{})};ne._clientEdgeTraversalPath.length>0&&ne._clientEdgeTraversalPath[ne._clientEdgeTraversalPath.length-1]!==null&&(Re[h]=(0,i.default)(ne._clientEdgeTraversalPath));var et={getDataForResolverFragment:de};return U(et,function(){var Ae=J(X,ne._variables,Re,ne._resolverContext),He=Ae[0],je=Ae[1];return{resolverResult:He,snapshot:ie,error:je}})}else{var tt=J(X,ne._variables,null,ne._resolverContext),Oe=tt[0],$e=tt[1];return{resolverResult:Oe,snapshot:void 0,error:$e}}},ye=this._resolverCache.readFromCacheOrEvaluate(B,X,this._variables,ge,de),me=ye[0],Ee=ye[1],se=ye[2],pe=ye[3],ve=ye[4],we=ye[5];return this._propagateResolverMetadata(X.path,pe,se,Ee,ve,we),me},M._propagateResolverMetadata=function(X,B,ne,oe,ie,de){var ge=this;if(B!=null){if(B.missingClientEdges!=null)for(var ye=0;ye<B.missingClientEdges.length;ye++){var me=B.missingClientEdges[ye];this._missingClientEdges.push(me)}if(B.missingLiveResolverFields!=null){this._isMissingData=this._isMissingData||B.missingLiveResolverFields.length>0;for(var Ee=0;Ee<B.missingLiveResolverFields.length;Ee++){var se=B.missingLiveResolverFields[Ee];this._missingLiveResolverFields.push(se)}}if(B.fieldErrors!=null){this._fieldErrors==null&&(this._fieldErrors=[]);for(var pe=0;pe<B.fieldErrors.length;pe++){var ve,we=B.fieldErrors[pe];((ve=this._selector.node.metadata)===null||ve===void 0?void 0:ve.throwOnFieldError)===!0?this._fieldErrors.push(we):this._fieldErrors.push(K(we))}}this._isMissingData=this._isMissingData||B.isMissingData}if(ne){var Me,Re,et={kind:"relay_resolver.error",fieldPath:X,owner:this._fragmentName,error:ne,shouldThrow:(Me=(Re=this._selector.node.metadata)===null||Re===void 0?void 0:Re.throwOnFieldError)!==null&&Me!==void 0?Me:!1,handled:!1,uiContext:void 0};this._fieldErrors==null?this._fieldErrors=[et]:this._fieldErrors.push(et)}oe!=null&&this._seenRecords.add(oe),ie!=null&&(this._isMissingData=!0,this._missingLiveResolverFields.push(ie)),de!=null&&de.forEach(function(tt){ge._updatedDataIDs.add(tt)})},M._readClientEdge=function(X,B,ne){var oe=this,ie,de=X.backingField;de.kind==="ClientExtension"&&N(!1);var ge=(ie=de.alias)!==null&&ie!==void 0?ie:de.name,ye={};this._traverseSelections([de],B,ye);var me=ye[ge];if(me==null||o(me))return ne[ge]=me,me;if(X.linkedField.plural){Array.isArray(me)||N(!1);var Ee;X.kind!=="ClientEdgeToClientObject"&&N(!1),X.backingField.normalizationInfo==null?Ee=me.map(function(je){var Ye,Ne=(Ye=X.concreteType)!==null&&Ye!==void 0?Ye:je.__typename;typeof Ne!="string"&&N(!1);var ze=te(je,de.path,oe._owner.identifier),at=oe._resolverCache.ensureClientRecord(ze,Ne),mt=X.modelResolvers;if(mt!=null){var Mt=mt[Ne];Mt===void 0&&N(!1);var bt=oe._readResolverFieldImpl(Mt,at);return bt!=null?at:null}return at}):Ee=me.map(function(je){return te(je,de.path,oe._owner.identifier)}),this._clientEdgeTraversalPath.push(null);var se=this._readLinkedIds(X.linkedField,Ee,B,ne);return this._clientEdgeTraversalPath.pop(),ne[ge]=se,se}else{var pe,ve=te(me,de.path,this._owner.identifier),we,Me=(pe=X.concreteType)!==null&&pe!==void 0?pe:me.__typename,Re;X.kind==="ClientEdgeToClientObject"?X.backingField.normalizationInfo==null?(typeof Me!="string"&&N(!1),we=this._resolverCache.ensureClientRecord(ve,Me),Re=null):(we=ve,Re=null):(we=ve,Re={readerClientEdge:X,clientEdgeDestinationID:ve});var et=X.modelResolvers;if(et!=null){typeof Me!="string"&&N(!1);var tt=et[Me];tt===void 0&&N(!1);var Oe=this._readResolverFieldImpl(tt,we);if(Oe==null)return ne[ge]=null,null}this._clientEdgeTraversalPath.push(Re);var $e=ne[ge];$e==null||typeof $e=="object"||N(!1);var Ae=this._fieldErrors;this._fieldErrors=null;var He=this._traverse(X.linkedField,we,$e);return this._prependPreviousErrors(Ae,ge),this._clientEdgeTraversalPath.pop(),ne[ge]=He,He}},M._readScalar=function(X,B,ne){var oe,ie=(oe=X.alias)!==null&&oe!==void 0?oe:X.name,de=A(X,this._variables),ge=c.getValue(B,de);return ge===null||u.ENABLE_NONCOMPLIANT_ERROR_HANDLING_ON_LISTS&&Array.isArray(ge)&&ge.length===0?this._maybeAddFieldErrors(B,de):ge===void 0&&this._markDataAsMissing(ie),ne[ie]=ge,ge},M._readLink=function(X,B,ne){var oe,ie=(oe=X.alias)!==null&&oe!==void 0?oe:X.name,de=A(X,this._variables),ge=c.getLinkedRecordID(B,de);if(ge==null)return ne[ie]=ge,ge===null?this._maybeAddFieldErrors(B,de):ge===void 0&&this._markDataAsMissing(ie),ge;var ye=ne[ie];ye==null||typeof ye=="object"||N(!1);var me=this._fieldErrors;this._fieldErrors=null;var Ee=this._traverse(X,ge,ye);return this._prependPreviousErrors(me,ie),ne[ie]=Ee,Ee},M._prependPreviousErrors=function(X,B){if(this._fieldErrors!=null){for(var ne=0;ne<this._fieldErrors.length;ne++){var oe=this._fieldErrors[ne];oe.owner===this._fragmentName&&(oe.kind==="missing_expected_data.throw"||oe.kind==="missing_expected_data.log"||oe.kind==="missing_required_field.throw"||oe.kind==="missing_required_field.log")&&(oe.fieldPath="".concat(B,".").concat(oe.fieldPath))}if(X!=null){for(var ie=this._fieldErrors.length-1;ie>=0;ie--)X.push(this._fieldErrors[ie]);this._fieldErrors=X}}else this._fieldErrors=X},M._readActorChange=function(X,B,ne){var oe,ie=(oe=X.alias)!==null&&oe!==void 0?oe:X.name,de=A(X,this._variables),ge=c.getActorLinkedRecordID(B,de);if(ge==null)return ne[ie]=ge,ge===void 0?this._markDataAsMissing(ie):ge===null&&this._maybeAddFieldErrors(B,de),ne[ie];var ye=ge[0],me=ge[1],Ee={};return this._createFragmentPointer(X.fragmentSpread,c.fromObject({__id:me}),Ee),ne[ie]={__fragmentRef:Ee,__viewer:ye},ne[ie]},M._readPluralLink=function(X,B,ne){var oe=A(X,this._variables),ie=c.getLinkedRecordIDs(B,oe);return(ie===null||u.ENABLE_NONCOMPLIANT_ERROR_HANDLING_ON_LISTS&&Array.isArray(ie)&&ie.length===0)&&this._maybeAddFieldErrors(B,oe),this._readLinkedIds(X,ie,B,ne)},M._readLinkedIds=function(X,B,ne,oe){var ie=this,de,ge=(de=X.alias)!==null&&de!==void 0?de:X.name;if(B==null)return oe[ge]=B,B===void 0&&this._markDataAsMissing(ge),B;var ye=oe[ge];ye==null||Array.isArray(ye)||N(!1);var me=this._fieldErrors;this._fieldErrors=null;var Ee=ye||[];return B.forEach(function(se,pe){if(se==null){se===void 0&&ie._markDataAsMissing(String(pe)),Ee[pe]=se;return}var ve=Ee[pe];ve==null||typeof ve=="object"||N(!1);var we=ie._fieldErrors;ie._fieldErrors=null,Ee[pe]=ie._traverse(X,se,ve),ie._prependPreviousErrors(we,pe)}),this._prependPreviousErrors(me,ge),oe[ge]=Ee,Ee},M._readModuleImport=function(X,B,ne){var oe=k(X.documentName),ie=c.getValue(B,oe),de=ie!==void 0?ie:X.componentModuleProvider;if(de==null){de===void 0&&this._markDataAsMissing("<module-import>");return}this._createFragmentPointer({kind:"FragmentSpread",name:X.fragmentName,args:X.args},B,ne),ne[y]=X.fragmentPropName,ne[_]=de},M._readAliasedInlineFragment=function(X,B,ne){var oe=this._fieldErrors;this._fieldErrors=null;var ie=this._readInlineFragment(X.fragment,B,{},!0);this._prependPreviousErrors(oe,X.name),ie===!1&&(ie=null),ne[X.name]=ie},M._readInlineFragment=function(X,B,ne,oe){if(X.type==null){var ie=this._traverseSelections(X.selections,B,ne);return ie===!1?!1:ne}var de=X.abstractKey;if(de==null)if(this._recordMatchesTypeCondition(B,X.type)){var ie=this._traverseSelections(X.selections,B,ne);if(!ie)return!1}else return null;else{var ge=this._implementsInterface(B,de);if(ge===!1&&oe)return null;var ye=this._isMissingData,me=this._isWithinUnmatchedTypeRefinement;this._isWithinUnmatchedTypeRefinement=me||ge===!1;var Ee=this._traverseSelections(X.selections,B,ne);if(this._isWithinUnmatchedTypeRefinement=me,ge===!1)return this._isMissingData=ye,null;if(ge==null)return;if(Ee===!1)return!1}return ne},M._recordMatchesTypeCondition=function(X,B){var ne=c.getType(X);return ne!=null&&ne===B||c.getDataID(X)===E},M._createFragmentPointer=function(X,B,ne){var oe=ne[v];oe==null&&(oe=ne[v]={}),typeof oe=="object"&&oe!=null||N(!1),ne[S]==null&&(ne[S]=c.getDataID(B)),oe[X.name]=$(X.args,this._variables,this._isWithinUnmatchedTypeRefinement),ne[m]=this._owner,this._clientEdgeTraversalPath.length>0&&this._clientEdgeTraversalPath[this._clientEdgeTraversalPath.length-1]!==null&&(ne[h]=(0,i.default)(this._clientEdgeTraversalPath))},M._createInlineDataOrResolverFragmentPointer=function(X,B,ne){var oe=ne[v];oe==null&&(oe=ne[v]={}),typeof oe=="object"&&oe!=null||N(!1),ne[S]==null&&(ne[S]=c.getDataID(B));var ie={},de=this._fragmentName;this._fragmentName=X.name;var ge=this._variables,ye=X.args?$(X.args,this._variables):{};this._variables=s.getFragmentVariables(X,this._owner.variables,ye),this._traverseSelections(X.selections,B,ie),this._variables=ge,this._fragmentName=de,oe[X.name]=ie},M._implementsInterface=function(X,B){var ne=c.getType(X),oe=this._recordSource.get(V(ne)),ie=oe!=null?c.getValue(oe,B):null;return ie==null&&this._markDataAsMissing("<abstract-type-hint>"),ie},H})();function K(H){switch(H.kind){case"missing_expected_data.throw":case"missing_required_field.throw":case"relay_field_payload.error":case"relay_resolver.error":return(0,t.default)((0,t.default)({},H),{},{handled:!0});case"missing_expected_data.log":case"missing_required_field.log":return H;default:H.kind,N(!1)}}function J(H,M,G,X){var B=typeof H.resolverModule=="function"?H.resolverModule:H.resolverModule.default,ne=null,oe=null;try{var ie=[];H.fragment!=null&&ie.push(G);var de=H.args?$(H.args,M):void 0;ie.push(de),ie.push(X),ne=B.apply(null,ie)}catch(ge){ne=null,ge!==I&&(oe=ge)}return[ne,oe]}function te(H,M,G){if(typeof H=="string")return H;if(typeof H=="object"&&H!=null&&typeof H.id=="string")return H.id;N(!1)}return iA={read:j},iA}var aA,cq;function $4e(){if(cq)return aA;cq=1;var e=$n().default,t=e(Cu()),n=YF(),r=iee(),i=aee(),a=cy(),u=a.getLocalVariables,l=hf(),o=Da(),s=AC(),c=s.generateTypeID,d=un(),h=o.getReadTimeResolverStorageKey,m=o.getStorageKey,y=o.getModuleOperationKey;function v(_,E,$,k,A,w){var R=E.dataID,F=E.node,I=E.variables,U=new S(_,I,$,k,A,w);U.mark(F,R)}var S=(function(){function _($,k,A,w,R,F){this._operationLoader=w??null,this._operationName=null,this._useExecTimeResolvers=F??!1,this._recordSource=$,this._references=A,this._variables=k,this._shouldProcessClientComponents=R}var E=_.prototype;return E.mark=function(k,A){(k.kind==="Operation"||k.kind==="SplitOperation")&&(this._operationName=k.name),this._traverse(k,A)},E._traverse=function(k,A){this._references.add(A);var w=this._recordSource.get(A);w!=null&&this._traverseSelections(k.selections,w)},E._getVariableValue=function(k){return this._variables.hasOwnProperty(k)||d(!1),this._variables[k]},E._traverseSelections=function(k,A){var w=this;k.forEach(function(R){switch(R.kind){case"ActorChange":w._traverseLink(R.linkedField,A);break;case"LinkedField":R.plural?w._traversePluralLink(R,A):w._traverseLink(R,A);break;case"Condition":var F=!!w._getVariableValue(R.condition);F===R.passingValue&&w._traverseSelections(R.selections,A);break;case"InlineFragment":if(R.abstractKey==null){var I=l.getType(A);(I!=null&&I===R.type||I===o.ROOT_TYPE)&&w._traverseSelections(R.selections,A)}else{var U=l.getType(A),q=c(U);w._references.add(q),w._traverseSelections(R.selections,A)}break;case"FragmentSpread":var V=w._variables;w._variables=u(w._variables,R.fragment.argumentDefinitions,R.args),w._traverseSelections(R.fragment.selections,A),w._variables=V;break;case"LinkedHandle":var N=r(R,k,w._variables);N.plural?w._traversePluralLink(N,A):w._traverseLink(N,A);break;case"Defer":case"Stream":w._traverseSelections(R.selections,A);break;case"ScalarField":case"ScalarHandle":break;case"TypeDiscriminator":{var j=l.getType(A),W=c(j);w._references.add(W);break}case"ModuleImport":w._traverseModuleImport(R,A);break;case"ClientExtension":w._traverseSelections(R.selections,A);break;case"ClientComponent":if(w._shouldProcessClientComponents===!1)break;w._traverseSelections(R.fragment.selections,A);break;case"RelayResolver":case"RelayLiveResolver":w._traverseResolverField(R,A);break;case"ClientEdgeToClientObject":w._traverseClientEdgeToClientObject(R,A);break;default:d(!1)}})},E._traverseClientEdgeToClientObject=function(k,A){if(this._useExecTimeResolvers){this._traverseLink(k.linkedField,A);return}var w=this._traverseResolverField(k.backingField,A);if(w!=null){var R=this._recordSource.get(w);if(R!=null)if(k.backingField.isOutputType){var F=i(R);if(F!=null){var I=(0,t.default)(F),U;try{for(I.s();!(U=I.n()).done;){var q=U.value;this._references.add(q)}}catch(H){I.e(H)}finally{I.f()}}}else{var V=k.linkedField,N=V.concreteType;if(N==null)return;if(V.plural){var j=l.getResolverLinkedRecordIDs(R,N);if(j!=null){var W=(0,t.default)(j),K;try{for(W.s();!(K=W.n()).done;){var J=K.value;J!=null&&this._traverse(V,J)}}catch(H){W.e(H)}finally{W.f()}}}else{var te=l.getResolverLinkedRecordID(R,N);te!=null&&this._traverse(V,te)}}}},E._traverseResolverField=function(k,A){if(!this._useExecTimeResolvers){var w=h(k,this._variables),R=l.getLinkedRecordID(A,w);R!=null&&this._references.add(R);var F=k.fragment;return F!=null&&this._traverseSelections([F],A),R}},E._traverseModuleImport=function(k,A){var w=this._operationLoader;w===null&&d(!1);var R=y(k.documentName),F=l.getValue(A,R);if(F!=null){var I=w.get(F);if(I!=null){var U=n(I),q=this._variables;this._variables=u(this._variables,U.argumentDefinitions,k.args),this._traverseSelections(U.selections,A),this._variables=q}}},E._traverseLink=function(k,A){var w=m(k,this._variables),R=l.getLinkedRecordID(A,w);R!=null&&this._traverse(k,R)},E._traversePluralLink=function(k,A){var w=this,R=m(k,this._variables),F=l.getLinkedRecordIDs(A,R);F!=null&&F.forEach(function(I){I!=null&&w._traverse(k,I)})},_})();return aA={mark:v},aA}var uA,fq;function C4e(){if(fq)return uA;fq=1;var e=Symbol.iterator;function t(n,r){for(var i=n[e](),a=i.next();!a.done;){var u=a.value;if(r.has(u))return!0;a=i.next()}return!1}return uA=t,uA}var lA,dq;function w4e(){if(dq)return lA;dq=1;var e=Da(),t=e.ROOT_ID,n=XF(),r=n.VIEWER_ID,i=Symbol.iterator;function a(u,l){for(var o=u[i](),s=o.next();!s.done;){var c=s.value;if(l.has(c)&&c!==t&&c!==r)return!0;s=o.next()}return!1}return lA=a,lA}var oA,hq;function R4e(){if(hq)return oA;hq=1,sS();var e=CC(),t=nh(),n=C4e(),r=w4e(),i=GF(),a=(function(){function u(o,s,c){this._subscriptions=new Set,this.__log=o,this._resolverCache=s,this._resolverContext=c}var l=u.prototype;return l.subscribe=function(s,c){var d=this,h={backup:null,callback:c,snapshot:s,stale:!1},m=function(){d._subscriptions.delete(h)};return this._subscriptions.add(h),{dispose:m}},l.snapshotSubscriptions=function(s){var c=this;this._subscriptions.forEach(function(d){if(!d.stale){d.backup=d.snapshot;return}var h=d.snapshot,m=i.read(s,h.selector,c._resolverCache,c._resolverContext),y=e(h.data,m.data);m.data=y,d.backup=m})},l.restoreSubscriptions=function(){this._subscriptions.forEach(function(s){var c=s.backup;s.backup=null,c?(c.data!==s.snapshot.data&&(s.stale=!0),s.snapshot={data:s.snapshot.data,isMissingData:c.isMissingData,missingClientEdges:c.missingClientEdges,missingLiveResolverFields:c.missingLiveResolverFields,seenRecords:c.seenRecords,selector:c.selector,fieldErrors:c.fieldErrors}):s.stale=!0})},l.updateSubscriptions=function(s,c,d,h){var m=this,y=c.size!==0;this._subscriptions.forEach(function(v){var S=m._updateSubscription(s,v,c,y,h);S!=null&&d.push(S)})},l._updateSubscription=function(s,c,d,h,m){var y=c.backup,v=c.callback,S=c.snapshot,_=c.stale,E=h&&n(S.seenRecords,d);if(!(!_&&!E)){var $=E||!y?i.read(s,S.selector,this._resolverCache,this._resolverContext):y,k=e(S.data,$.data);if($={data:k,isMissingData:$.isMissingData,missingClientEdges:$.missingClientEdges,missingLiveResolverFields:$.missingLiveResolverFields,seenRecords:$.seenRecords,selector:$.selector,fieldErrors:$.fieldErrors},c.snapshot=$,c.stale=!1,$.data!==S.data)return this.__log&&t.ENABLE_NOTIFY_SUBSCRIPTION&&this.__log({name:"store.notify.subscription",sourceOperation:m,snapshot:S,nextSnapshot:$}),v($),S.selector.owner;if(t.ENABLE_LOOSE_SUBSCRIPTION_ATTRIBUTION&&(_||r(S.seenRecords,d)))return S.selector.owner}},l.size=function(){return this._subscriptions.size},u})();return oA=a,oA}var sA,pq;function uee(){if(pq)return sA;pq=1;var e=$n().default,t=e(Cu()),n=e(mc()),r=HF(),i=r.INTERNAL_ACTOR_IDENTIFIER_DO_NOT_USE,a=r.assertInternalActorIdentifier;sS();var u=v4e(),l=O4e(),o=KJ(),s=_4e(),c=s.LiveResolverCache,d=s.RELAY_RESOLVER_LIVE_STATE_SUBSCRIPTION_KEY,h=s.getUpdatedDataIDs,m=hf(),y=E4e(),v=GF(),S=$4e(),_=R4e(),E=Da(),$=Da(),k=$.ROOT_ID,A=$.ROOT_TYPE,w=un(),R=10,F=(function(){function V(j,W){var K=this,J,te,H,M,G,X,B,ne;(0,n.default)(this,"_gcStep",function(){K._gcRun&&(K._gcRun.next().done?K._gcRun=null:K._gcScheduler(K._gcStep))}),this._currentWriteEpoch=0,this._gcHoldCounter=0,this._gcReleaseBufferSize=(J=W==null?void 0:W.gcReleaseBufferSize)!==null&&J!==void 0?J:R,this._shouldRetainWithinTTL_EXPERIMENTAL=(te=W==null?void 0:W.shouldRetainWithinTTL_EXPERIMENTAL)!==null&&te!==void 0?te:!1,this._gcRun=null,this._gcScheduler=(H=W==null?void 0:W.gcScheduler)!==null&&H!==void 0?H:u,this._getDataID=(M=W==null?void 0:W.getDataID)!==null&&M!==void 0?M:o,this._globalInvalidationEpoch=null,this._invalidationSubscriptions=new Set,this._invalidatedRecordIDs=new Set,this.__log=(G=W==null?void 0:W.log)!==null&&G!==void 0?G:null,this._queryCacheExpirationTime=W==null?void 0:W.queryCacheExpirationTime,this._operationLoader=(X=W==null?void 0:W.operationLoader)!==null&&X!==void 0?X:null,this._optimisticSource=null,this._recordSource=j,this._releaseBuffer=[],this._roots=new Map,this._shouldScheduleGC=!1,this._resolverCache=new c(function(){return K._getMutableRecordSource()},this),this._resolverContext=W==null?void 0:W.resolverContext,this._storeSubscriptions=new _(W==null?void 0:W.log,this._resolverCache,this._resolverContext),this._updatedRecordIDs=new Set,this._shouldProcessClientComponents=(B=W==null?void 0:W.shouldProcessClientComponents)!==null&&B!==void 0?B:!1,this._treatMissingFieldsAsNull=(ne=W==null?void 0:W.treatMissingFieldsAsNull)!==null&&ne!==void 0?ne:!1,this._actorIdentifier=W==null?void 0:W.actorIdentifier,I(this._recordSource)}var N=V.prototype;return N.getSource=function(){var W;return(W=this._optimisticSource)!==null&&W!==void 0?W:this._recordSource},N.getOperationLoader=function(){return this._operationLoader},N._getMutableRecordSource=function(){var W;return(W=this._optimisticSource)!==null&&W!==void 0?W:this._recordSource},N.getLiveResolverPromise=function(W){return this._resolverCache.getLiveResolverPromise(W)},N.batchLiveStateUpdates=function(W){this.__log!=null&&this.__log({name:"liveresolver.batch.start"});try{this._resolverCache.batchLiveStateUpdates(W)}finally{this.__log!=null&&this.__log({name:"liveresolver.batch.end"})}},N.check=function(W,K){var J,te,H,M,G,X,B,ne=W.root,oe=this._getMutableRecordSource(),ie=this._globalInvalidationEpoch,de=(J=(te=W.request.node.operation.use_exec_time_resolvers)!==null&&te!==void 0?te:((H=W.request.node.operation.exec_time_resolvers_enabled_provider)===null||H===void 0?void 0:H.get())===!0)!==null&&J!==void 0?J:!1,ge=this._roots.get(W.request.identifier),ye=ge!=null?ge.epoch:null;if(ie!=null&&(ye==null||ye<=ie))return{status:"stale"};var me=(M=K==null?void 0:K.handlers)!==null&&M!==void 0?M:[],Ee=(G=K==null?void 0:K.getSourceForActor)!==null&&G!==void 0?G:function(ve){return a(ve),oe},se=(X=K==null?void 0:K.getTargetForActor)!==null&&X!==void 0?X:function(ve){return a(ve),oe},pe=l.check(Ee,se,(B=K==null?void 0:K.defaultActorIdentifier)!==null&&B!==void 0?B:i,ne,me,this._operationLoader,this._getDataID,this._shouldProcessClientComponents,this.__log,de);return q(pe,ye,ge==null?void 0:ge.fetchTime,this._queryCacheExpirationTime)},N.retain=function(W){var K=this,J=W.request.identifier,te=!1,H=function(){if(!te){te=!0;var X=K._roots.get(J);if(X!=null&&(X.refCount--,X.refCount===0)){var B=K._queryCacheExpirationTime,ne=X.fetchTime!=null&&B!=null&&X.fetchTime<=Date.now()-B;if(ne)K._shouldRetainWithinTTL_EXPERIMENTAL||K._roots.delete(J),K.scheduleGC();else if(K._releaseBuffer.push(J),K._releaseBuffer.length>K._gcReleaseBufferSize){var oe=K._releaseBuffer.shift();K._shouldRetainWithinTTL_EXPERIMENTAL||K._roots.delete(oe),K.scheduleGC()}}}},M=this._roots.get(J);return M!=null?(M.refCount===0&&(this._releaseBuffer=this._releaseBuffer.filter(function(G){return G!==J})),M.refCount+=1):this._roots.set(J,{operation:W,refCount:1,epoch:null,fetchTime:null}),{dispose:H}},N.lookup=function(W){var K=this.__log;K!=null&&K({name:"store.lookup.start",selector:W});var J=this.getSource(),te=v.read(J,W,this._resolverCache,this._resolverContext);return K!=null&&K({name:"store.lookup.end",selector:W}),te},N.notify=function(W,K){var J=this,te=this.__log;te!=null&&te({name:"store.notify.start",sourceOperation:W}),this._currentWriteEpoch++,K===!0&&(this._globalInvalidationEpoch=this._currentWriteEpoch),this._resolverCache.invalidateDataIDs(this._updatedRecordIDs);var H=this.getSource(),M=[];if(this._storeSubscriptions.updateSubscriptions(H,this._updatedRecordIDs,M,W),this._invalidationSubscriptions.forEach(function(ne){J._updateInvalidationSubscription(ne,K===!0)}),W!=null){var G=W.request.identifier,X=this._roots.get(G);if(X!=null)X.epoch=this._currentWriteEpoch,X.fetchTime=Date.now();else if(W.request.node.params.operationKind==="query"&&this._gcReleaseBufferSize>0&&this._releaseBuffer.length<this._gcReleaseBufferSize){var B={operation:W,refCount:0,epoch:this._currentWriteEpoch,fetchTime:Date.now()};this._releaseBuffer.push(G),this._roots.set(G,B)}}return te!=null&&te({name:"store.notify.complete",sourceOperation:W,updatedRecordIDs:this._updatedRecordIDs,invalidatedRecordIDs:this._invalidatedRecordIDs,subscriptionsSize:this._storeSubscriptions.size(),updatedOwners:M}),this._updatedRecordIDs.clear(),this._invalidatedRecordIDs.clear(),M},N.publish=function(W,K){var J=this._getMutableRecordSource();U(J,W,this._currentWriteEpoch+1,K,this._updatedRecordIDs,this._invalidatedRecordIDs);var te=this.__log;te!=null&&te({name:"store.publish",source:W,optimistic:J===this._optimisticSource})},N.subscribe=function(W,K){return this._storeSubscriptions.subscribe(W,K)},N.holdGC=function(){var W=this;this._gcRun&&(this._gcRun=null,this._shouldScheduleGC=!0),this._gcHoldCounter++;var K=function(){W._gcHoldCounter>0&&(W._gcHoldCounter--,W._gcHoldCounter===0&&W._shouldScheduleGC&&(W.scheduleGC(),W._shouldScheduleGC=!1))};return{dispose:K}},N.toJSON=function(){return"RelayModernStore()"},N.getEpoch=function(){return this._currentWriteEpoch},N.__getUpdatedRecordIDs=function(){return this._updatedRecordIDs},N.lookupInvalidationState=function(W){var K=this,J=new Map;return W.forEach(function(te){var H,M=K.getSource().get(te);J.set(te,(H=m.getInvalidationEpoch(M))!==null&&H!==void 0?H:null)}),J.set("global",this._globalInvalidationEpoch),{dataIDs:W,invalidations:J}},N.checkInvalidationState=function(W){var K=this.lookupInvalidationState(W.dataIDs),J=K.invalidations,te=W.invalidations;if(J.get("global")!==te.get("global"))return!0;var H=(0,t.default)(W.dataIDs),M;try{for(H.s();!(M=H.n()).done;){var G=M.value;if(J.get(G)!==te.get(G))return!0}}catch(X){H.e(X)}finally{H.f()}return!1},N.subscribeToInvalidationState=function(W,K){var J=this,te={callback:K,invalidationState:W},H=function(){J._invalidationSubscriptions.delete(te)};return this._invalidationSubscriptions.add(te),{dispose:H}},N._updateInvalidationSubscription=function(W,K){var J=this,te=W.callback,H=W.invalidationState,M=H.dataIDs,G=K||M.some(function(X){return J._invalidatedRecordIDs.has(X)});G&&te()},N.snapshot=function(){this._optimisticSource!=null&&w(!1);var W=this.__log;W!=null&&W({name:"store.snapshot"}),this._storeSubscriptions.snapshotSubscriptions(this.getSource()),this._gcRun&&(this._gcRun=null,this._shouldScheduleGC=!0),this._optimisticSource=y.create(this.getSource())},N.restore=function(){var W=this._optimisticSource;W||w(!1);var K=this.__log;K!=null&&K({name:"store.restore"});var J=y.getOptimisticRecordIDs(W);this._resolverCache.unsubscribeFromLiveResolverRecords(J),this._optimisticSource=null,this._shouldScheduleGC&&this.scheduleGC(),this._storeSubscriptions.restoreSubscriptions(),this._resolverCache.invalidateResolverRecords(J)},N.scheduleGC=function(){if(this._gcHoldCounter>0){this._shouldScheduleGC=!0;return}this._gcRun||(this._gcRun=this._collect(),this._gcScheduler(this._gcStep))},N.__gc=function(){if(this._optimisticSource==null)for(var W=this._collect();!W.next().done;);},N._collect=function*(){if(!(this._shouldRetainWithinTTL_EXPERIMENTAL&&this._queryCacheExpirationTime==null)){var W=this.__log;e:for(;;){W!=null&&W({name:"store.gc.start"});var K=this._currentWriteEpoch,J=new Set,te=(0,t.default)(this._roots.entries()),H;try{for(te.s();!(H=te.n()).done;){var M,G,X,B=H.value,ne=B[0],oe=B[1],ie=oe.operation,de=oe.refCount,ge=oe.fetchTime;if(this._shouldRetainWithinTTL_EXPERIMENTAL){var ye=this._queryCacheExpirationTime;ye==null&&w(!1);var me=ge==null||ge<=Date.now()-ye,Ee=me&&de===0&&!this._releaseBuffer.includes(ne);if(Ee)continue}var se=ie.root,pe=(M=(G=ie.request.node.operation.use_exec_time_resolvers)!==null&&G!==void 0?G:((X=ie.request.node.operation.exec_time_resolvers_enabled_provider)===null||X===void 0?void 0:X.get())===!0)!==null&&M!==void 0?M:!1;if(S.mark(this._recordSource,se,J,this._operationLoader,this._shouldProcessClientComponents,pe),yield,K!==this._currentWriteEpoch){W!=null&&W({name:"store.gc.interrupted"});continue e}}}catch(tt){te.e(tt)}finally{te.f()}for(var ve=this._recordSource.getRecordIDs(),we=0;we<ve.length;we++){var Me=ve[we];if(!J.has(Me)){var Re=this._recordSource.get(Me);if(Re!=null){var et=m.getValue(Re,d);et!=null&&et()}this._recordSource.remove(Me),this._shouldRetainWithinTTL_EXPERIMENTAL&&this._roots.delete(Me)}}W!=null&&W({name:"store.gc.end",references:J});return}}},N.__getNormalizationOptions=function(W){return{path:W,getDataID:this._getDataID,log:this.__log,treatMissingFieldsAsNull:this._treatMissingFieldsAsNull,shouldProcessClientComponents:this._shouldProcessClientComponents,actorIdentifier:this._actorIdentifier}},N.__notifyUpdatedSubscribers=function(W){var K=h(W),J=this._updatedRecordIDs;this._updatedRecordIDs=K,this.notify(),this._updatedRecordIDs=J},V})();function I(V){if(!V.has(k)){var N=m.create(k,A);V.set(k,N)}}function U(V,N,j,W,K,J){W&&W.forEach(function(ne){var oe=V.get(ne),ie=N.get(ne);if(ie!==null){var de;oe!=null?de=m.clone(oe):de=ie!=null?m.clone(ie):null,de&&(m.setValue(de,E.INVALIDATED_AT_KEY,j),J.add(ne),V.set(ne,de))}});for(var te=N.getRecordIDs(),H=0;H<te.length;H++){var M=te[H],G=N.get(M),X=V.get(M);if(G&&X){var B=m.update(X,G);B!==X&&(K.add(M),V.set(M,B))}else G===null?(V.delete(M),X!==null&&K.add(M)):G&&(V.set(M,G),K.add(M))}}function q(V,N,j,W){var K=V.mostRecentlyInvalidatedAt,J=V.status;if(typeof K=="number"&&(N==null||K>N))return{status:"stale"};if(J==="missing")return{status:"missing"};if(j!=null&&W!=null){var te=j<=Date.now()-W;if(te)return{status:"stale"}}return{status:"available",fetchTime:j??null}}return sA=F,sA}var cA,mq;function A4e(){if(mq)return cA;mq=1;var e=Da(),t=e.ROOT_TYPE,n=e.getStorageKey,r=tee(),i=r.readUpdatableFragment,a=nee(),u=a.readUpdatableQuery,l=un(),o=(function(){function s(d,h,m,y){this.__mutator=d,this.__recordSource=h,this._readSelector=m,this._missingFieldHandlers=y}var c=s.prototype;return c.create=function(h,m){return this.__recordSource.create(h,m)},c.delete=function(h){this.__recordSource.delete(h)},c.get=function(h){return this.__recordSource.get(h)},c.getRoot=function(){return this.__recordSource.getRoot()},c.getOperationRoot=function(){var h=this.__recordSource.get(this._readSelector.dataID);return h||(h=this.__recordSource.create(this._readSelector.dataID,t)),h},c._getRootField=function(h,m,y){var v=h.node.selections.find(function(S){return S.kind==="LinkedField"&&S.name===m||S.kind==="RequiredField"&&S.field.name===m});return v&&v.kind==="RequiredField"&&(v=v.field),v&&v.kind==="LinkedField"||l(!1),v.plural!==y&&l(!1),v},c.getRootField=function(h){var m=this._getRootField(this._readSelector,h,!1),y=n(m,this._readSelector.variables);return this.getOperationRoot().getLinkedRecord(y)},c.getPluralRootField=function(h){var m=this._getRootField(this._readSelector,h,!0),y=n(m,this._readSelector.variables);return this.getOperationRoot().getLinkedRecords(y)},c.invalidateStore=function(){this.__recordSource.invalidateStore()},c.readUpdatableQuery=function(h,m){return u(h,m,this,this._missingFieldHandlers)},c.readUpdatableFragment=function(h,m){return i(h,m,this,this._missingFieldHandlers)},s})();return cA=o,cA}var fA,gq;function T4e(){if(gq)return fA;gq=1;var e,t,n=JJ(),r=ree(),i=A4e(),a=nh(),u=GF(),l=E1(),o=un();Vi();var s=typeof Xd<"u"?Xd:typeof window<"u"?window:void 0,c=(e=s==null||(t=s.ErrorUtils)===null||t===void 0?void 0:t.applyWithGuard)!==null&&e!==void 0?e:function(m,y,v,S,_){return m.apply(y,v)},d=(function(){function m(v,S,_,E,$){this._hasStoreSnapshot=!1,this._handlerProvider=S||null,this._pendingBackupRebase=!1,this._pendingData=new Set,this._pendingOptimisticUpdates=new Set,this._store=v,this._appliedOptimisticUpdates=new Set,this._gcHold=null,this._getDataID=_,this._missingFieldHandlers=E,this._log=$}var y=m.prototype;return y.applyUpdate=function(S){!this._appliedOptimisticUpdates.has(S)&&!this._pendingOptimisticUpdates.has(S)||o(!1),this._pendingOptimisticUpdates.add(S)},y.revertUpdate=function(S){this._pendingOptimisticUpdates.has(S)?this._pendingOptimisticUpdates.delete(S):this._appliedOptimisticUpdates.has(S)&&(this._pendingBackupRebase=!0,this._appliedOptimisticUpdates.delete(S))},y.revertAll=function(){this._pendingBackupRebase=!0,this._pendingOptimisticUpdates.clear(),this._appliedOptimisticUpdates.clear()},y.commitPayload=function(S,_,E){this._pendingBackupRebase=!0,this._pendingData.add({kind:"payload",operation:S,payload:_,updater:E})},y.commitUpdate=function(S){this._pendingBackupRebase=!0,this._pendingData.add({kind:"updater",updater:S})},y.commitSource=function(S){this._pendingBackupRebase=!0,this._pendingData.add({kind:"source",source:S})},y.run=function(S){var _=this._appliedOptimisticUpdates===0&&!!this._gcHold,E=!this._pendingBackupRebase&&this._pendingOptimisticUpdates.size===0&&!_;if(a.DISALLOW_NESTED_UPDATES&&this._isRunning===!0&&o(!1),this._isRunning=!0,E)return this._isRunning=!1,[];this._pendingBackupRebase&&this._hasStoreSnapshot&&(this._store.restore(),this._hasStoreSnapshot=!1);var $=this._commitData();return(this._pendingOptimisticUpdates.size||this._pendingBackupRebase&&this._appliedOptimisticUpdates.size)&&(this._hasStoreSnapshot||(this._store.snapshot(),this._hasStoreSnapshot=!0),this._applyUpdates()),this._pendingBackupRebase=!1,this._appliedOptimisticUpdates.size>0?this._gcHold||(this._gcHold=this._store.holdGC()):this._gcHold&&(this._gcHold.dispose(),this._gcHold=null),this._isRunning=!1,this._store.notify(S,$)},y._publishSourceFromPayload=function(S){var _=this,E=S.payload,$=S.operation,k=S.updater,A=E.source,w=E.fieldPayloads,R=new n(this._store.getSource(),A),F=new r(R,this._getDataID,this._handlerProvider,this._missingFieldHandlers,this._log);if(w&&w.length&&w.forEach(function(N){var j=_._handlerProvider&&_._handlerProvider(N.handle);j||o(!1),j.update(F,N)}),k){var I=$.fragment;I==null&&o(!1);var U=new i(R,F,I,this._missingFieldHandlers),q=h(A,I);k(U,q)}var V=F.getIDsMarkedForInvalidation();return this._store.publish(A,V),F.isStoreMarkedForInvalidation()},y._commitData=function(){var S=this;if(!this._pendingData.size)return!1;var _=!1;return this._pendingData.forEach(function(E){if(E.kind==="payload"){var $=S._publishSourceFromPayload(E);_=_||$}else if(E.kind==="source"){var k=E.source;S._store.publish(k)}else{var A=E.updater,w=l.create(),R=new n(S._store.getSource(),w),F=new r(R,S._getDataID,S._handlerProvider,S._missingFieldHandlers,S._log);c(A,null,[F],null,"RelayPublishQueue:commitData"),_=_||F.isStoreMarkedForInvalidation();var I=F.getIDsMarkedForInvalidation();S._store.publish(w,I)}}),this._pendingData.clear(),_},y._applyUpdates=function(){var S=this,_=l.create(),E=new n(this._store.getSource(),_),$=new r(E,this._getDataID,this._handlerProvider,this._missingFieldHandlers,this._log),k=function(w){if(w.storeUpdater){var R=w.storeUpdater;c(R,null,[$],null,"RelayPublishQueue:applyUpdates")}else{var F=w.operation,I=w.payload,U=w.updater,q=I.source,V=I.fieldPayloads;if(q&&$.publishSource(q,V),U){var N;q&&(N=h(q,F.fragment));var j=new i(E,$,F.fragment,S._missingFieldHandlers);c(U,null,[j,N],null,"RelayPublishQueue:applyUpdates")}}};this._pendingBackupRebase&&this._appliedOptimisticUpdates.size&&this._appliedOptimisticUpdates.forEach(k),this._pendingOptimisticUpdates.size&&(this._pendingOptimisticUpdates.forEach(function(A){k(A),S._appliedOptimisticUpdates.add(A)}),this._pendingOptimisticUpdates.clear()),this._store.publish(_)},m})();function h(m,y){var v=u.read(m,y).data;return v}return fA=d,fA}var dA,vq;function k4e(){if(vq)return dA;vq=1;var e=$n().default,t=e(Ui()),n=VJ(),r=HF(),i=r.INTERNAL_ACTOR_IDENTIFIER_DO_NOT_USE,a=r.assertInternalActorIdentifier,u=fy(),l=d4e(),o=GJ(),s=h4e(),c=KJ(),d=p4e(),h=XJ(),m=g4e(),y=uee(),v=T4e(),S=E1();un();var _=(function(){function k(w){var R=this,F,I,U,q,V,N,j,W,K,J,te;this.configName=w.configName,this._treatMissingFieldsAsNull=w.treatMissingFieldsAsNull===!0;var H=w.operationLoader,M=(F=w.store)!==null&&F!==void 0?F:new y(new S,{log:w.log,operationLoader:w.operationLoader,getDataID:w.getDataID,shouldProcessClientComponents:w.shouldProcessClientComponents});this.__log=(I=w.log)!==null&&I!==void 0?I:$,this.relayFieldLogger=(U=w.relayFieldLogger)!==null&&U!==void 0?U:d,this._defaultRenderPolicy=(q=w.UNSTABLE_defaultRenderPolicy)!==null&&q!==void 0?q:"partial",this._operationLoader=H,this._operationExecutions=new Map,this._network=l(this,w.network),this._getDataID=(V=w.getDataID)!==null&&V!==void 0?V:c,this._missingFieldHandlers=(N=w.missingFieldHandlers)!==null&&N!==void 0?N:[],this._publishQueue=new v(M,(j=w.handlerProvider)!==null&&j!==void 0?j:n,this._getDataID,this._missingFieldHandlers,this.__log),this._scheduler=(W=w.scheduler)!==null&&W!==void 0?W:null,this._store=M,this.options=w.options,this._isServer=(K=w.isServer)!==null&&K!==void 0?K:!1,this._normalizeResponse=(J=w.normalizeResponse)!==null&&J!==void 0?J:h,this.__setNet=function(G){return R._network=l(R,G)},this._operationTracker=(te=w.operationTracker)!==null&&te!==void 0?te:new o,this._shouldProcessClientComponents=w.shouldProcessClientComponents,s(this)}var A=k.prototype;return A.getStore=function(){return this._store},A.getNetwork=function(){return this._network},A.getOperationTracker=function(){return this._operationTracker},A.getScheduler=function(){return this._scheduler},A.isRequestActive=function(R){var F=this._operationExecutions.get(R);return F==="active"},A.UNSTABLE_getDefaultRenderPolicy=function(){return this._defaultRenderPolicy},A.applyUpdate=function(R){var F=this,I=function(){F._scheduleUpdates(function(){F._publishQueue.revertUpdate(R),F._publishQueue.run()})};return this._scheduleUpdates(function(){F._publishQueue.applyUpdate(R),F._publishQueue.run()}),{dispose:I}},A.revertUpdate=function(R){var F=this;this._scheduleUpdates(function(){F._publishQueue.revertUpdate(R),F._publishQueue.run()})},A.replaceUpdate=function(R,F){var I=this;this._scheduleUpdates(function(){I._publishQueue.revertUpdate(R),I._publishQueue.applyUpdate(F),I._publishQueue.run()})},A.applyMutation=function(R){var F=this._execute({createSource:function(){return u.create(function(U){})},isClientPayload:!1,operation:R.operation,optimisticConfig:R,updater:null}).subscribe({});return{dispose:function(){return F.unsubscribe()}}},A.check=function(R){return this._missingFieldHandlers.length===0&&!E(R)?this._store.check(R):this._checkSelectorAndHandleMissingFields(R,this._missingFieldHandlers)},A.commitPayload=function(R,F){this._execute({createSource:function(){return u.from({data:F})},isClientPayload:!0,operation:R,optimisticConfig:null,updater:null}).subscribe({})},A.commitUpdate=function(R){var F=this;this._scheduleUpdates(function(){F._publishQueue.commitUpdate(R),F._publishQueue.run()})},A.lookup=function(R){return this._store.lookup(R)},A.subscribe=function(R,F){return this._store.subscribe(R,F)},A.retain=function(R){return this._store.retain(R)},A.isServer=function(){return this._isServer},A._checkSelectorAndHandleMissingFields=function(R,F){var I=this,U=S.create(),q=this._store.getSource(),V=this._store.check(R,{handlers:F,defaultActorIdentifier:i,getSourceForActor:function(j){return a(j),q},getTargetForActor:function(j){return a(j),U}});return U.size()>0&&this._scheduleUpdates(function(){I._publishQueue.commitSource(U),I._publishQueue.run()}),V},A._scheduleUpdates=function(R){var F=this._scheduler;F!=null?F.schedule(R):R()},A.execute=function(R){var F=this,I=R.operation;return this._execute({createSource:function(){return F.getNetwork().execute(I.request.node.params,I.request.variables,I.request.cacheConfig||{},null,void 0,void 0,void 0,function(){return F.check(I)})},isClientPayload:!1,operation:I,optimisticConfig:null,updater:null})},A.executeSubscription=function(R){var F=this,I=R.operation,U=R.updater;return this._execute({createSource:function(){return F.getNetwork().execute(I.request.node.params,I.request.variables,I.request.cacheConfig||{},null)},isClientPayload:!1,operation:I,optimisticConfig:null,updater:U})},A.executeMutation=function(R){var F=this,I=R.operation,U=R.optimisticResponse,q=R.optimisticUpdater,V=R.updater,N=R.uploadables,j;return(U||q)&&(j={operation:I,response:U,updater:q}),this._execute({createSource:function(){return F.getNetwork().execute(I.request.node.params,I.request.variables,(0,t.default)((0,t.default)({},I.request.cacheConfig),{},{force:!0}),N)},isClientPayload:!1,operation:I,optimisticConfig:j,updater:V})},A.executeWithSource=function(R){var F=R.operation,I=R.source;return this._execute({createSource:function(){return I},isClientPayload:!1,operation:F,optimisticConfig:null,updater:null})},A.toJSON=function(){var R;return"RelayModernEnvironment(".concat((R=this.configName)!==null&&R!==void 0?R:"",")")},A._execute=function(R){var F=this,I=R.createSource,U=R.isClientPayload,q=R.operation,V=R.optimisticConfig,N=R.updater,j=this._publishQueue,W=this._store;return u.create(function(K){var J=m.execute({actorIdentifier:i,getDataID:F._getDataID,isClientPayload:U,log:F.__log,operation:q,operationExecutions:F._operationExecutions,operationLoader:F._operationLoader,operationTracker:F._operationTracker,optimisticConfig:V,getPublishQueue:function(H){return a(H),j},scheduler:F._scheduler,shouldProcessClientComponents:F._shouldProcessClientComponents,sink:K,source:I(),getStore:function(H){return a(H),W},treatMissingFieldsAsNull:F._treatMissingFieldsAsNull,updater:N,normalizeResponse:F._normalizeResponse});return function(){return J.cancel()}})},k})();function E(k){return k.root.node.kind==="Operation"&&k.root.node.clientAbstractTypes!=null}_.prototype["@@RelayModernEnvironment"]=!0;function $(){}return dA=_,dA}var hA,yq;function P4e(){if(yq)return hA;yq=1;var e=$C(),t=dp(),n=t.getRequest,r=dy(),i=r.createOperationDescriptor,a=hp(),u=a.createReaderSelector;Vi();function l(o,s){var c=n(s.subscription);if(c.params.operationKind!=="subscription")throw new Error("requestSubscription: Must use Subscription operation");var d=s.configs,h=s.onCompleted,m=s.onError,y=s.onNext,v=s.variables,S=s.cacheConfig,_=i(c,v,S),E=d?e.convert(d,c,null,s.updater):s,$=E.updater,k=o.executeSubscription({operation:_,updater:$}).subscribe({next:function(w){if(y!=null){var R=_.fragment,F;if(Array.isArray(w)){var I,U;F=(I=w[0])===null||I===void 0||(U=I.extensions)===null||U===void 0?void 0:U.__relay_subscription_root_id}else{var q;F=(q=w.extensions)===null||q===void 0?void 0:q.__relay_subscription_root_id}typeof F=="string"&&(R=u(R.node,F,R.variables,R.owner));var V=o.lookup(R).data;y(V)}},error:m,complete:h});return{dispose:k.unsubscribe}}return hA=l,hA}var pA,bq;function M4e(){if(bq)return pA;bq=1;var e=$n().default,t=e(Ui()),n=Da(),r=n.getModuleComponentKey,i=n.getModuleOperationKey;function a(u,l,o,s){var c=(0,t.default)({},s);return c[r(u)]=o,c[i(u)]=l,c}return pA=a,pA}var mA,Oq;function F4e(){if(Oq)return mA;Oq=1;var e=Object.prototype.hasOwnProperty;function t(n){for(var r in n)if(e.call(n,r))return!1;return!0}return mA=t,mA}var gA,xq;function N4e(){if(xq)return gA;xq=1;var e=hp(),t=e.getDataIDsFromFragment,n=e.getSelector,r=e.getVariablesFromFragment,i=F4e(),a=nh(),u=_1(),l=u.stableCopy,o=zJ(),s=o.intern;function c(d,h){var m=n(d,h),y=m==null?"null":m.kind==="SingularReaderSelector"?m.owner.identifier:"["+m.selectors.map(function(k){return k.owner.identifier}).join(",")+"]",v=r(d,h),S=t(d,h);if(a.ENABLE_GETFRAGMENTIDENTIFIER_OPTIMIZATION){var _=typeof S>"u"?"missing":S==null?"null":Array.isArray(S)?"["+S.join(",")+"]":S;return _=a.STRING_INTERN_LEVEL<=1?_:s(_,a.MAX_DATA_ID_LENGTH),y+"/"+d.name+"/"+(v==null||i(v)?"{}":JSON.stringify(l(v)))+"/"+_}else{var E,$=(E=JSON.stringify(S))!==null&&E!==void 0?E:"missing";return $=a.STRING_INTERN_LEVEL<=1?$:s($,a.MAX_DATA_ID_LENGTH),y+"/"+d.name+"/"+JSON.stringify(l(v))+"/"+$}}return gA=c,gA}var vA,Sq;function lee(){if(Sq)return vA;Sq=1;var e=un();function t(n,r){var i,a;((i=n.metadata)===null||i===void 0?void 0:i.plural)===!0&&e(!1);var u=(a=n.metadata)===null||a===void 0?void 0:a.refetch;u==null&&e(!1);var l=u.operation.default?u.operation.default:u.operation,o=u.fragmentPathInResult;typeof l=="string"&&e(!1);var s=u.identifierInfo;return s!=null&&(s.identifierField==null||typeof s.identifierField=="string"||e(!1),s.identifierQueryVariableName==null||typeof s.identifierQueryVariableName=="string"||e(!1)),{fragmentRefPathInResponse:o,identifierInfo:s,refetchableRequest:l,refetchMetadata:u}}return vA=t,vA}var yA,_q;function I4e(){if(_q)return yA;_q=1;var e=lee(),t=un();function n(r,i){var a,u,l=e(r,i),o=l.refetchableRequest,s=l.refetchMetadata,c=s.connection;c==null&&t(!1);var d=c.path,h=((a=(u=r.metadata)===null||u===void 0?void 0:u.connection)!==null&&a!==void 0?a:[])[0];h==null&&t(!1);var m=s.identifierInfo;return(m==null?void 0:m.identifierField)==null||typeof m.identifierField=="string"||t(!1),{connectionPathInFragmentData:d,identifierField:m==null?void 0:m.identifierField,paginationRequest:o,paginationMetadata:c,stream:h.stream===!0}}return yA=n,yA}var bA,Eq;function L4e(){if(Eq)return bA;Eq=1;var e=$n().default,t=e(mc()),n=e(Ui()),r=un();Vi();function i(a,u,l,o,s,c){var d,h=c.backward,m=c.forward;if(a==="backward"){var y;h!=null&&h.count!=null&&h.cursor!=null||r(!1);var v=(0,n.default)((0,n.default)((0,n.default)({},o),s),{},(y={},(0,t.default)(y,h.cursor,l),(0,t.default)(y,h.count,u),y));return m&&m.cursor&&(v[m.cursor]=null),m&&m.count&&(v[m.count]=null),v}m!=null&&m.count!=null&&m.cursor!=null||r(!1);var S=(0,n.default)((0,n.default)((0,n.default)({},o),s),{},(d={},(0,t.default)(d,m.cursor,l),(0,t.default)(d,m.count,u),d));return h&&h.cursor&&(S[h.cursor]=null),h&&h.count&&(S[h.count]=null),S}return bA=i,bA}var OA,Dq;function B4e(){if(Dq)return OA;Dq=1;var e=$n().default,t=e(Cu()),n=un();function r(i,a){var u=i,l=(0,t.default)(a),o;try{for(l.s();!(o=l.n()).done;){var s=o.value;if(u==null)return null;typeof s=="number"?(Array.isArray(u)||n(!1),u=u[s]):(typeof u=="object"&&!Array.isArray(u)||n(!1),u=u[s])}}catch(c){l.e(c)}finally{l.f()}return u}return OA=r,OA}var xA,$q;function Q4e(){if($q)return xA;$q=1;var e={},t={stop:function(){}},n={profile:function(a,u){var l=e[a];if(l&&l.length>0){for(var o=[],s=l.length-1;s>=0;s--){var c=l[s](a,u);o.unshift(c)}return{stop:function(h){o.forEach(function(m){return m(h)})}}}return t},attachProfileHandler:function(a,u){e.hasOwnProperty(a)||(e[a]=[]),e[a].push(u)},detachProfileHandler:function(a,u){e.hasOwnProperty(a)&&r(e[a],u)}};function r(i,a){var u=i.indexOf(a);u!==-1&&i.splice(u,1)}return xA=n,xA}var SA,Cq;function z4e(){if(Cq)return SA;Cq=1;var e=X3e(),t=e.isErrorResult,n=e.isValueResult,r=DC(),i=jF(),a=qJ(),u=VJ(),l=Y3e(),o=G3e(),s=W3e(),c=$C(),d=e4e(),h=fy(),m=t4e(),y=n4e(),v=r4e(),S=EC(),_=dp(),E=i4e(),$=N0(),k=$.generateClientID,A=$.generateUniqueClientID,w=$.isClientID,R=u4e(),F=l4e(),I=o4e(),U=ZF(),q=wC(),V=q.isSuspenseSentinel,N=q.suspenseSentinel,j=XJ(),W=f4e(),K=cy(),J=k4e(),te=dy(),H=hf(),M=hp(),G=uee(),X=GJ(),B=E1(),ne=Da(),oe=VF(),ie=XF(),de=P4e(),ge=M4e(),ye=sS(),me=N4e(),Ee=I4e(),se=L4e(),pe=jJ(),ve=lee(),we=qF(),Me=QJ(),Re=B4e(),et=oS(),tt=et.handlePotentialSnapshotErrors,Oe=LJ(),$e=ZJ(),Ae=CC(),He=oy(),je=NJ(),Ye=WJ(),Ne=nh(),ze=Q4e(),at=BJ(),mt=_1(),Mt=mt.hasCycle,bt=mt.stableCopy,zt=UJ();return SA={Environment:J,Network:d,Observable:h,QueryResponseCache:m,RecordSource:B,Record:H,ReplaySubject:at,Store:G,areEqualSelectors:M.areEqualSelectors,createFragmentSpecResolver:R,createNormalizationSelector:M.createNormalizationSelector,createOperationDescriptor:te.createOperationDescriptor,createReaderSelector:M.createReaderSelector,createRequestDescriptor:te.createRequestDescriptor,getArgumentValues:ne.getArgumentValues,getDataIDsFromFragment:M.getDataIDsFromFragment,getDataIDsFromObject:M.getDataIDsFromObject,getNode:_.getNode,getFragment:_.getFragment,getInlineDataFragment:_.getInlineDataFragment,getModuleComponentKey:ne.getModuleComponentKey,getModuleOperationKey:ne.getModuleOperationKey,getPaginationFragment:_.getPaginationFragment,getPluralSelector:M.getPluralSelector,getRefetchableFragment:_.getRefetchableFragment,getRequest:_.getRequest,getRequestIdentifier:Me,getSelector:M.getSelector,getSelectorsFromObject:M.getSelectorsFromObject,getSingularSelector:M.getSingularSelector,getStorageKey:ne.getStorageKey,getVariablesFromFragment:M.getVariablesFromFragment,getVariablesFromObject:M.getVariablesFromObject,getVariablesFromPluralFragment:M.getVariablesFromPluralFragment,getVariablesFromSingularFragment:M.getVariablesFromSingularFragment,handlePotentialSnapshotErrors:tt,graphql:_.graphql,isErrorResult:t,isValueResult:n,isFragment:_.isFragment,isInlineDataFragment:_.isInlineDataFragment,isSuspenseSentinel:V,suspenseSentinel:N,isRequest:_.isRequest,readInlineData:W,readFragment:oe.readFragment,MutationTypes:c.MutationTypes,RangeOperations:c.RangeOperations,DefaultHandlerProvider:u,ConnectionHandler:r,MutationHandlers:a,VIEWER_ID:ie.VIEWER_ID,VIEWER_TYPE:ie.VIEWER_TYPE,applyOptimisticMutation:l,commitLocalUpdate:o,commitMutation:s,fetchQuery:y,fetchQuery_DEPRECATED:v,isRelayModernEnvironment:U,requestSubscription:de,ConnectionInterface:i,PreloadableQueryRegistry:E,RelayProfiler:ze,createPayloadFor3DField:ge,RelayConcreteNode:He,RelayError:Ye,RelayFeatureFlags:Ne,DEFAULT_HANDLE_KEY:je.DEFAULT_HANDLE_KEY,FRAGMENTS_KEY:ne.FRAGMENTS_KEY,FRAGMENT_OWNER_KEY:ne.FRAGMENT_OWNER_KEY,ID_KEY:ne.ID_KEY,REF_KEY:ne.REF_KEY,REFS_KEY:ne.REFS_KEY,ROOT_ID:ne.ROOT_ID,ROOT_TYPE:ne.ROOT_TYPE,TYPENAME_KEY:ne.TYPENAME_KEY,deepFreeze:ye,generateClientID:k,generateUniqueClientID:A,getRelayHandleKey:we,isClientID:w,isPromise:Oe,isScalarAndEqual:$e,recycleNodesInto:Ae,stableCopy:bt,hasCycle:Mt,getFragmentIdentifier:me,getRefetchMetadata:ve,getPaginationMetadata:Ee,getPaginationVariables:se,getPendingOperationsForFragment:pe,getValueAtPath:Re,__internal:{ResolverFragments:oe,OperationTracker:X,createRelayContext:F,createRelayLoggingContext:I,getOperationVariables:K.getOperationVariables,getLocalVariables:K.getLocalVariables,fetchQuery:S.fetchQuery,fetchQueryDeduped:S.fetchQueryDeduped,getPromiseForActiveRequest:S.getPromiseForActiveRequest,getObservableForActiveRequest:S.getObservableForActiveRequest,normalizeResponse:j,withProvidedVariables:zt}},SA}var _A,wq;function ii(){return wq||(wq=1,_A=z4e()),_A}var EA,Rq;function hy(){if(Rq)return EA;Rq=1;var e=ir(),t=ii(),n=t.__internal.createRelayContext;return EA=n(e),EA}var DA={exports:{}},Aq;function TC(){return Aq||(Aq=1,(function(e){function t(){return e.exports=t=Object.assign?Object.assign.bind():function(n){for(var r=1;r<arguments.length;r++){var i=arguments[r];for(var a in i)({}).hasOwnProperty.call(i,a)&&(n[a]=i[a])}return n},e.exports.__esModule=!0,e.exports.default=e.exports,t.apply(null,arguments)}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})(DA)),DA.exports}var $A={exports:{}},Tq;function kC(){return Tq||(Tq=1,(function(e){function t(n){if(n===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return n}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})($A)),$A.exports}var CA={exports:{}},wA={exports:{}},kq;function q4e(){return kq||(kq=1,(function(e){function t(n,r){return e.exports=t=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(i,a){return i.__proto__=a,i},e.exports.__esModule=!0,e.exports.default=e.exports,t(n,r)}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})(wA)),wA.exports}var Pq;function PC(){return Pq||(Pq=1,(function(e){var t=q4e();function n(r,i){r.prototype=Object.create(i.prototype),r.prototype.constructor=r,t(r,i)}e.exports=n,e.exports.__esModule=!0,e.exports.default=e.exports})(CA)),CA.exports}var RA,Mq;function V4e(){if(Mq)return RA;Mq=1;var e=un();function t(n,r){r&&typeof r=="object"||e(!1);for(var i in r)if(r.hasOwnProperty(i)){var a=r[i];a&&(typeof a=="object"||typeof a=="function")||e(!1)}}return RA=t,RA}var AA,Fq;function MC(){if(Fq)return AA;Fq=1;function e(n){return n.displayName||n.name||"Component"}function t(n){return"Relay("+e(n)+")"}return AA={getComponentName:e,getContainerName:t},AA}var TA,Nq;function KF(){if(Nq)return TA;Nq=1;var e=ir();return TA=e.createContext({rootIsQueryRenderer:!1}),TA}var kA,Iq;function WF(){if(Iq)return kA;Iq=1;var e=$n().default,t=e(TC()),n=V4e(),r=MC(),i=r.getComponentName,a=r.getContainerName,u=hy(),l=KF(),o=un(),s=ir(),c=ii(),d=c.getFragment,h=s.useContext;function m(y,v,S){var _=a(y);n(i(y),v);var E={};for(var $ in v)E[$]=d(v[$]);var k=S(y,E);k.displayName=_;function A(R,F){var I,U=h(u);U==null&&o(!1);var q=h(l);return s.createElement(k,(0,t.default)({},R,{__relayContext:U,__rootIsQueryRenderer:(I=q==null?void 0:q.rootIsQueryRenderer)!==null&&I!==void 0?I:!1,componentRef:R.componentRef||F}))}A.displayName=_;var w=s.forwardRef(A);return w}return kA=m,kA}var PA,Lq;function U4e(){if(Lq)return PA;Lq=1;function e(t){return typeof t=="object"&&t!==null&&typeof t.check=="function"&&typeof t.lookup=="function"&&typeof t.retain=="function"&&typeof t.execute=="function"&&typeof t.subscribe=="function"}return PA=e,PA}var MA,Bq;function JF(){if(Bq)return MA;Bq=1;var e=U4e(),t=un();function n(i){return r(i)||t(!1),i}function r(i){return typeof i=="object"&&i!==null&&!Array.isArray(i)&&e(i.environment)}return MA={assertRelayContext:n,isRelayContext:r},MA}var FA,Qq;function j4e(){if(Qq)return FA;Qq=1;var e=$n().default,t=e(TC()),n=e(H3()),r=e(kC()),i=e(PC()),a=e(mc()),u=["componentRef","__relayContext","__rootIsQueryRenderer"],l=WF(),o=MC(),s=o.getContainerName,c=JF(),d=c.assertRelayContext,h=ed(),m=ir(),y=ii(),v=y.createFragmentSpecResolver,S=y.getDataIDsFromObject,_=y.isScalarAndEqual;function E(A,w){var R,F=s(A);return R=(function(I){(0,i.default)(U,I);function U(V){var N,j;j=I.call(this,V)||this,(0,a.default)((0,r.default)(j),"_handleFragmentDataUpdate",function(){var te=j.state.resolver;j.setState(function(H){return te===H.resolver?{data:H.resolver.resolve(),relayProp:$(H.relayProp.environment)}:null})});var W=d(V.__relayContext),K=(N=V.__rootIsQueryRenderer)!==null&&N!==void 0?N:!1,J=v(W,F,w,V,K);return j.state={data:J.resolve(),prevProps:V,prevPropsContext:W,relayProp:$(W.environment),resolver:J},j}U.getDerivedStateFromProps=function(N,j){var W,K=j.prevProps,J=d(N.__relayContext),te=(W=N.__rootIsQueryRenderer)!==null&&W!==void 0?W:!1,H=S(w,K),M=S(w,N),G=j.resolver;if(j.prevPropsContext.environment!==J.environment||!h(H,M))return G=v(J,F,w,N,te),{data:G.resolve(),prevPropsContext:J,prevProps:N,relayProp:$(J.environment),resolver:G};G.setProps(N);var X=G.resolve();return X!==j.data?{data:X,prevProps:N,prevPropsContext:J,relayProp:$(J.environment)}:null};var q=U.prototype;return q.componentDidMount=function(){this._subscribeToNewResolverAndRerenderIfStoreHasChanged()},q.componentDidUpdate=function(N,j){this.state.resolver!==j.resolver?(j.resolver.dispose(),this._subscribeToNewResolverAndRerenderIfStoreHasChanged()):this._rerenderIfStoreHasChanged()},q.componentWillUnmount=function(){this.state.resolver.dispose()},q.shouldComponentUpdate=function(N,j){if(j.data!==this.state.data)return!0;for(var W=Object.keys(N),K=0;K<W.length;K++){var J=W[K];if(J==="__relayContext"){if(j.prevPropsContext.environment!==this.state.prevPropsContext.environment)return!0}else if(!w.hasOwnProperty(J)&&!_(N[J],this.props[J]))return!0}return!1},q._rerenderIfStoreHasChanged=function(){var N=this.state,j=N.data,W=N.resolver,K=W.resolve();j!==K&&this.setState({data:K})},q._subscribeToNewResolverAndRerenderIfStoreHasChanged=function(){var N=this.state,j=N.data,W=N.resolver,K=W.resolve();W.setCallback(this.props,this._handleFragmentDataUpdate),j!==K&&this.setState({data:K})},q.render=function(){var N=this.props,j=N.componentRef;N.__relayContext,N.__rootIsQueryRenderer;var W=(0,n.default)(N,u);return m.createElement(A,(0,t.default)({},W,this.state.data,{ref:j,relay:this.state.relayProp}))},U})(m.Component),(0,a.default)(R,"displayName",F),R}function $(A){return{environment:A}}function k(A,w){return l(A,w,E)}return FA={createContainer:k},FA}var NA,zq;function Z4e(){if(zq)return NA;zq=1;var e=hy(),t=KF(),n=ed(),r=ir(),i=ii(),a=i.createOperationDescriptor;i.deepFreeze;var u=i.getRequest,l=r.useLayoutEffect,o=r.useState,s=r.useRef,c=r.useMemo,d={rootIsQueryRenderer:!0};function h(y){var v=r.useRef(y);return n(v.current,y)||(v.current=y),v.current}function m(y){var v=y.environment,S=y.query,_=y.variables,E=y.render,$=h(_),k=c(function(){var q=u(S);return a(q,$)},[S,$]),A=c(function(){return{environment:v}},[v]),w=s(null),R=o(null),F=R[1],I=s(null),U=c(function(){v.check(k);var q=v.lookup(k.fragment);w.current=q.data;var V=v.retain(k),N=v.subscribe(q,function(K){w.current=K.data,F(w.current)}),j=!1;function W(){j||(j=!0,I.current=null,V.dispose(),N.dispose())}return I.current&&I.current(),I.current=W,q},[v,k]);return l(function(){var q=I.current;return function(){q&&q()}},[U]),r.createElement(e.Provider,{value:A},r.createElement(t.Provider,{value:d},E({props:w.current})))}return NA=m,NA}var IA,qq;function oee(){if(qq)return IA;qq=1;var e=$n().default,t=e(Ui()),n=ii(),r=n.getSelector;function i(a,u){var l={};return Object.keys(a).forEach(function(o){var s,c,d,h=a[o],m=u[o],y=r(h,m),v=y!=null&&y.kind==="PluralReaderSelector"?(s=(c=y.selectors[0])===null||c===void 0?void 0:c.owner.variables)!==null&&s!==void 0?s:{}:(d=y==null?void 0:y.owner.variables)!==null&&d!==void 0?d:{};l=(0,t.default)((0,t.default)({},l),v)}),l}return IA=i,IA}var LA,Vq;function eN(){if(Vq)return LA;Vq=1;var e=$n().default,t=e(mc()),n=un(),r=ii(),i=r.__internal.fetchQuery,a=r.createOperationDescriptor,u=r.isRelayModernEnvironment,l=(function(){function o(c){(0,t.default)(this,"_selectionReferences",[]),(0,t.default)(this,"_didFetchFinish",!1),c!=null&&(this._cacheSelectionReference=c.cacheSelectionReference,this._selectionReferences=c.selectionReferences)}var s=o.prototype;return s.getSelectionReferences=function(){return{cacheSelectionReference:this._cacheSelectionReference,selectionReferences:this._selectionReferences}},s.lookupInStore=function(d,h,m){return(m==="store-and-network"||m==="store-or-network")&&d.check(h).status==="available"?(this._retainCachedOperation(d,h),d.lookup(h.fragment)):null},s.getFetchResult=function(){if(this._didFetchFinish){if(this._error!=null)return{error:this._error};if(this._snapshot!=null)return{snapshot:this._snapshot}}else return null},s.execute=function(d){var h=this,m=d.environment,y=d.operation,v=d.preservePreviousReferences,S=v===void 0?!1:v,_=m.retain(y),E=function(){h._selectionReferences=h._selectionReferences.concat(_)},$=function(){S||h.disposeSelectionReferences(),h._selectionReferences=h._selectionReferences.concat(_)},k=function(){h._selectionReferences=h._selectionReferences.concat(_)};return u(m)?i(m,y).do({error:E,complete:$,unsubscribe:k}):m.execute({operation:y}).do({error:E,complete:$,unsubscribe:k})},s.setOnDataChange=function(d){this._fetchOptions||n(!1),typeof d=="function"&&(this._fetchOptions.onDataChangeCallbacks=this._fetchOptions.onDataChangeCallbacks||[],this._fetchOptions.onDataChangeCallbacks.push(d),this._didFetchFinish&&(this._error!=null?d({error:this._error}):this._snapshot!=null&&d({snapshot:this._snapshot})))},s.fetch=function(d,h){var m=this,y=d.environment,v=d.operation,S=d.onDataChange,_=!1,E;this.disposeRequest();var $=this._fetchOptions&&this._fetchOptions.onDataChangeCallbacks;this._fetchOptions={environment:y,onDataChangeCallbacks:$||[],operation:v},S&&this._fetchOptions.onDataChangeCallbacks.indexOf(S)===-1&&this._fetchOptions.onDataChangeCallbacks.push(S);var k=h?a(v.request.node,v.request.variables,h):v,A=this.execute({environment:y,operation:k}).finally(function(){m._pendingRequest=null}).subscribe({next:function(){m._didFetchFinish=!0,m._error=null,m._onQueryDataAvailable({notifyFirstResult:_})},error:function(R){m._didFetchFinish=!0,m._error=R,m._snapshot=null;var F=m._fetchOptions&&m._fetchOptions.onDataChangeCallbacks;_?F&&F.forEach(function(I){I({error:R})}):E=R}});if(this._pendingRequest={dispose:function(){A.unsubscribe()}},_=!0,E)throw E;return this._snapshot},s.retry=function(d){return this._fetchOptions||n(!1),this.fetch({environment:this._fetchOptions.environment,operation:this._fetchOptions.operation,onDataChange:null},d)},s.dispose=function(){this.disposeRequest(),this.disposeSelectionReferences()},s.disposeRequest=function(){this._error=null,this._snapshot=null,this._pendingRequest&&this._pendingRequest.dispose(),this._rootSubscription&&(this._rootSubscription.dispose(),this._rootSubscription=null)},s._retainCachedOperation=function(d,h){this._disposeCacheSelectionReference(),this._cacheSelectionReference=d.retain(h)},s._disposeCacheSelectionReference=function(){this._cacheSelectionReference&&this._cacheSelectionReference.dispose(),this._cacheSelectionReference=null},s.disposeSelectionReferences=function(){this._disposeCacheSelectionReference(),this._selectionReferences.forEach(function(d){return d.dispose()}),this._selectionReferences=[]},s._onQueryDataAvailable=function(d){var h=this,m=d.notifyFirstResult;this._fetchOptions||n(!1);var y=this._fetchOptions,v=y.environment,S=y.onDataChangeCallbacks,_=y.operation;if(!this._snapshot&&(this._snapshot=v.lookup(_.fragment),this._rootSubscription=v.subscribe(this._snapshot,function($){if(h._fetchOptions!=null){var k=h._fetchOptions.onDataChangeCallbacks;Array.isArray(k)&&k.forEach(function(A){return A({snapshot:$})})}}),this._snapshot&&m&&Array.isArray(S))){var E=this._snapshot;S.forEach(function($){return $({snapshot:E})})}},o})();return LA=l,LA}var BA,Uq;function H4e(){if(Uq)return BA;Uq=1;var e=$n().default,t=e(TC()),n=e(H3()),r=e(kC()),i=e(PC()),a=e(mc()),u=e(Ui()),l=["componentRef"],o=["componentRef","__relayContext","__rootIsQueryRenderer"],s=["componentRef","__relayContext","__rootIsQueryRenderer"],c=WF(),d=oee(),h=MC(),m=h.getComponentName,y=h.getContainerName,v=hy(),S=eN(),_=JF(),E=_.assertRelayContext,$=ed(),k=un(),A=ir(),w=ii(),R=w.ConnectionInterface,F=w.Observable,I=w.RelayFeatureFlags,U=w.createFragmentSpecResolver,q=w.createOperationDescriptor,V=w.getDataIDsFromObject,N=w.getRequest,j=w.getVariablesFromObject,W=w.isScalarAndEqual;Vi();var K="forward";function J(B){var ne=B.path;return ne||k(!1),function(oe){for(var ie=oe[B.fragmentName],de=0;de<ne.length;de++){if(!ie||typeof ie!="object")return null;ie=ie[ne[de]]}return ie}}function te(B){var ne=B.count;return ne||k(!1),function(oe,ie){return(0,u.default)((0,u.default)({},oe),{},(0,a.default)({},ne,ie))}}function H(B){var ne=null,oe=!1;for(var ie in B){var de=B[ie],ge=de.metadata&&de.metadata.connection;de.metadata!==void 0&&(oe=!0),ge&&(ge.length!==1&&k(!1),ne&&k(!1),ne=(0,u.default)((0,u.default)({},ge[0]),{},{fragmentName:ie}))}return!oe||ne!==null||k(!1),ne||{}}function M(B){return typeof B=="function"?{error:B,complete:B,unsubscribe:function(oe){typeof B=="function"&&B()}}:B||{}}function G(B,ne,oe){var ie;m(B);var de=y(B),ge=H(ne),ye=oe.getConnectionFromProps||J(ge),me=oe.direction||ge.direction;me||k(!1);var Ee=oe.getFragmentVariables||te(ge);return ie=(function(se){(0,i.default)(pe,se);function pe(we){var Me,Re;Re=se.call(this,we)||this,(0,a.default)((0,r.default)(Re),"_handleFragmentDataUpdate",function(){Re.setState({data:Re._resolver.resolve()})}),(0,a.default)((0,r.default)(Re),"_hasMore",function(){var Oe=Re._getConnectionData();return!!(Oe&&Oe.hasMore&&Oe.cursor)}),(0,a.default)((0,r.default)(Re),"_isLoading",function(){return!!Re._refetchSubscription}),(0,a.default)((0,r.default)(Re),"_refetchConnection",function(Oe,$e,Ae){if(!Re._canFetchPage("refetchConnection"))return{dispose:function(){}};Re._refetchVariables=Ae;var He={count:Oe,cursor:null,totalCount:Oe},je=Re._fetchPage(He,M($e),{force:!0});return{dispose:je.unsubscribe}}),(0,a.default)((0,r.default)(Re),"_loadMore",function(Oe,$e,Ae){if(!Re._canFetchPage("loadMore"))return{dispose:function(){}};var He=M($e),je=Re._getConnectionData();if(!je)return F.create(function(Mt){return Mt.complete()}).subscribe(He),null;var Ye=je.edgeCount+Oe;if(Ae&&Ae.force)return Re._refetchConnection(Ye,$e);var Ne=R.get();Ne.END_CURSOR,Ne.START_CURSOR;var ze=je.cursor,at={count:Oe,cursor:ze,totalCount:Ye},mt=Re._fetchPage(at,He,Ae);return{dispose:mt.unsubscribe}});var et=E(we.__relayContext),tt=(Me=we.__rootIsQueryRenderer)!==null&&Me!==void 0?Me:!1;return Re._isARequestInFlight=!1,Re._refetchSubscription=null,Re._refetchVariables=null,I.ENABLE_CONTAINERS_SUBSCRIBE_ON_COMMIT===!0?Re._resolver=U(et,de,ne,we,tt):Re._resolver=U(et,de,ne,we,tt,Re._handleFragmentDataUpdate),Re.state={data:Re._resolver.resolve(),prevContext:et,contextForChildren:et,relayProp:Re._buildRelayProp(et),resolverGeneration:0},Re._isUnmounted=!1,Re._hasFetched=!1,Re}var ve=pe.prototype;return ve.componentDidMount=function(){this._isUnmounted=!1,I.ENABLE_CONTAINERS_SUBSCRIBE_ON_COMMIT===!0&&this._subscribeToNewResolverAndRerenderIfStoreHasChanged()},ve.componentDidUpdate=function(Me,Re){I.ENABLE_CONTAINERS_SUBSCRIBE_ON_COMMIT===!0&&(Re.resolverGeneration!==this.state.resolverGeneration?this._subscribeToNewResolverAndRerenderIfStoreHasChanged():this._rerenderIfStoreHasChanged())},ve.UNSAFE_componentWillReceiveProps=function(Me){var Re=this,et,tt=E(Me.__relayContext),Oe=(et=Me.__rootIsQueryRenderer)!==null&&et!==void 0?et:!1,$e=V(ne,this.props),Ae=V(ne,Me),He=d(ne,this.props),je=d(ne,Me);tt.environment!==this.state.prevContext.environment||!$(He,je)||!$($e,Ae)?(this._cleanup(),I.ENABLE_CONTAINERS_SUBSCRIBE_ON_COMMIT===!0?this._resolver=U(tt,de,ne,Me,Oe):this._resolver=U(tt,de,ne,Me,Oe,this._handleFragmentDataUpdate),this.setState(function(Ne){return{prevContext:tt,contextForChildren:tt,relayProp:Re._buildRelayProp(tt),resolverGeneration:Ne.resolverGeneration+1}})):this._hasFetched||this._resolver.setProps(Me);var Ye=this._resolver.resolve();Ye!==this.state.data&&this.setState({data:Ye})},ve.componentWillUnmount=function(){this._isUnmounted=!0,this._cleanup()},ve.shouldComponentUpdate=function(Me,Re){if(Re.data!==this.state.data||Re.relayProp!==this.state.relayProp||Re.resolverGeneration!==this.state.resolverGeneration)return!0;for(var et=Object.keys(Me),tt=0;tt<et.length;tt++){var Oe=et[tt];if(Oe==="__relayContext"){if(Re.prevContext.environment!==this.state.prevContext.environment)return!0}else if(!ne.hasOwnProperty(Oe)&&!W(Me[Oe],this.props[Oe]))return!0}return!1},ve._buildRelayProp=function(Me){return{hasMore:this._hasMore,isLoading:this._isLoading,loadMore:this._loadMore,refetchConnection:this._refetchConnection,environment:Me.environment}},ve._rerenderIfStoreHasChanged=function(){var Me=this.state.data,Re=this._resolver.resolve();Me!==Re&&this.setState({data:Re})},ve._subscribeToNewResolverAndRerenderIfStoreHasChanged=function(){var Me=this.state.data,Re=this._resolver.resolve();this._resolver.setCallback(this.props,this._handleFragmentDataUpdate),Me!==Re&&this.setState({data:Re})},ve._getConnectionData=function(){var Me=this.props;Me.componentRef;var Re=(0,n.default)(Me,l),et=(0,u.default)((0,u.default)({},Re),this.state.data),tt=ye(et);if(tt==null)return null;var Oe=R.get(),$e=Oe.EDGES,Ae=Oe.PAGE_INFO,He=Oe.HAS_NEXT_PAGE,je=Oe.HAS_PREV_PAGE,Ye=Oe.END_CURSOR,Ne=Oe.START_CURSOR;typeof tt!="object"&&k(!1);var ze=tt[$e],at=tt[Ae];if(ze==null||at==null)return null;Array.isArray(ze)||k(!1),typeof at!="object"&&k(!1);var mt=me===K?at[He]:at[je],Mt=me===K?at[Ye]:at[Ne];return typeof mt!="boolean"||ze.length!==0&&typeof Mt>"u"?null:{cursor:Mt,edgeCount:ze.length,hasMore:mt}},ve._getQueryFetcher=function(){return this._queryFetcher||(this._queryFetcher=new S),this._queryFetcher},ve._canFetchPage=function(Me){return!this._isUnmounted},ve._fetchPage=function(Me,Re,et){var tt=this,Oe=E(this.props.__relayContext),$e=Oe.environment,Ae=this.props;Ae.componentRef,Ae.__relayContext,Ae.__rootIsQueryRenderer;var He=(0,n.default)(Ae,o),je=(0,u.default)((0,u.default)({},He),this.state.data),Ye,Ne=d(ne,He);Ye=j(ne,He),Ye=(0,u.default)((0,u.default)((0,u.default)({},Ne),Ye),this._refetchVariables);var ze=oe.getVariables(je,{count:Me.count,cursor:Me.cursor},Ye);typeof ze=="object"&&ze!==null||k(!1),ze=(0,u.default)((0,u.default)({},ze),this._refetchVariables),Ye=(0,u.default)((0,u.default)({},ze),Ye);var at=et?{force:!!et.force}:void 0;at!=null&&(et==null?void 0:et.metadata)!=null&&(at.metadata=et==null?void 0:et.metadata);var mt=N(oe.query),Mt=q(mt,ze,at),bt=null;this._refetchSubscription&&this._refetchSubscription.unsubscribe(),this._hasFetched=!0;var zt=function(nn,Fe){var ht=tt._resolver.resolve();tt._resolver.setVariables(Ee(Ye,Me.totalCount),Mt.request.node);var kt=tt._resolver.resolve();$(ht,kt)?Fe():tt.setState({data:kt,contextForChildren:{environment:tt.props.__relayContext.environment}},Fe)},qt=function(){tt._refetchSubscription===bt&&(tt._refetchSubscription=null,tt._isARequestInFlight=!1)};return this._isARequestInFlight=!0,bt=this._getQueryFetcher().execute({environment:$e,operation:Mt,preservePreviousReferences:!0}).mergeMap(function(Xt){return F.create(function(nn){zt(Xt,function(){nn.next(),nn.complete()})})}).do({error:qt,complete:qt,unsubscribe:qt}).subscribe(Re||{}),this._refetchSubscription=this._isARequestInFlight?bt:null,bt},ve._cleanup=function(){this._resolver.dispose(),this._refetchVariables=null,this._hasFetched=!1,this._refetchSubscription&&(this._refetchSubscription.unsubscribe(),this._refetchSubscription=null,this._isARequestInFlight=!1),this._queryFetcher&&this._queryFetcher.dispose()},ve.render=function(){var Me=this.props,Re=Me.componentRef;Me.__relayContext,Me.__rootIsQueryRenderer;var et=(0,n.default)(Me,s);return A.createElement(v.Provider,{value:this.state.contextForChildren},A.createElement(B,(0,t.default)({},et,this.state.data,{ref:Re,relay:this.state.relayProp})))},pe})(A.Component),(0,a.default)(ie,"displayName",de),ie}function X(B,ne,oe){return c(B,ne,function(ie,de){return G(ie,de,oe)})}return BA={createContainer:X},BA}var QA,jq;function X4e(){if(jq)return QA;jq=1;var e=$n().default,t=e(Ui()),n=e(kC()),r=e(PC()),i=e(mc()),a=hy(),u=eN(),l=KF(),o=ed(),s=ir(),c=ii(),d=c.createOperationDescriptor;c.deepFreeze;var h=c.getRequest,m={},y={rootIsQueryRenderer:!0},v=(function(w){(0,r.default)(R,w);function R(I){var U;U=w.call(this,I)||this,(0,i.default)((0,n.default)(U),"_handleDataChange",function(K){var J=K.error==null?null:K.error,te=K.snapshot==null?null:K.snapshot;U.setState(function(H){var M=H.requestCacheKey;return M&&delete m[M],te===H.snapshot&&J===H.error?null:{renderProps:E(J,te,H.queryFetcher,H.retryCallbacks),snapshot:te,requestCacheKey:null}})});var q={handleDataChange:null,handleRetryAfterError:null},V,N;if(I.query){var j=I.query,W=h(j);N=$(W.params,I.variables),V=m[N]?m[N].queryFetcher:new u}else V=new u;return U._maybeHiddenOrFastRefresh=!1,U.state=(0,t.default)({prevPropsEnvironment:I.environment,prevPropsVariables:I.variables,prevQuery:I.query,queryFetcher:V,retryCallbacks:q},A(I,V,q,N)),U}R.getDerivedStateFromProps=function(U,q){return q.prevQuery!==U.query||q.prevPropsEnvironment!==U.environment||!o(q.prevPropsVariables,U.variables)?k(U,q):null};var F=R.prototype;return F.componentDidMount=function(){var U=this;if(this._maybeHiddenOrFastRefresh===!0){this._maybeHiddenOrFastRefresh=!1,this.setState(function(W){var K=k(U.props,W),J=K.requestCacheKey,te=K.queryFetcher;return J!=null&&m[J]!=null&&te.setOnDataChange(U._handleDataChange),K});return}var q=this.state,V=q.retryCallbacks,N=q.queryFetcher,j=q.requestCacheKey;j&&delete m[j],V.handleDataChange=this._handleDataChange,V.handleRetryAfterError=function(W){return U.setState(function(K){var J=K.requestCacheKey;return J&&delete m[J],{renderProps:S(),requestCacheKey:null}})},this.props.query&&N.setOnDataChange(this._handleDataChange)},F.componentDidUpdate=function(U,q){var V=this.state,N=V.queryFetcher,j=V.requestCacheKey;j&&(delete m[j],delete this.state.requestCacheKey),this.props.query&&N!==q.queryFetcher&&N.setOnDataChange(this._handleDataChange)},F.componentWillUnmount=function(){this.state.queryFetcher.dispose(),this._maybeHiddenOrFastRefresh=!0},F.shouldComponentUpdate=function(U,q){return U.render!==this.props.render||q.renderProps!==this.state.renderProps},F.render=function(){var U=this.state,q=U.renderProps,V=U.relayContext;return s.createElement(a.Provider,{value:V},s.createElement(l.Provider,{value:y},this.props.render(q)))},R})(s.Component);function S(){return{error:null,props:null,retry:null}}function _(){return{error:null,props:{},retry:null}}function E(w,R,F,I){return{error:w||null,props:R?R.data:null,retry:function(q){var V=F.retry(q);V&&typeof I.handleDataChange=="function"?I.handleDataChange({snapshot:V}):w&&typeof I.handleRetryAfterError=="function"&&I.handleRetryAfterError(w)}}}function $(w,R){return JSON.stringify({id:w.cacheID?w.cacheID:w.id,variables:R})}function k(w,R){var F=w.query,I=R.queryFetcher.getSelectionReferences();R.queryFetcher.disposeRequest();var U;if(F){var q=h(F),V=$(q.params,w.variables);U=m[V]?m[V].queryFetcher:new u(I)}else U=new u(I);return(0,t.default)({prevQuery:w.query,prevPropsEnvironment:w.environment,prevPropsVariables:w.variables,queryFetcher:U},A(w,U,R.retryCallbacks))}function A(w,R,F,I){var U=w.environment,q=w.query,V=w.variables,N=w.cacheConfig,j=U;if(q){var W=h(q),K=d(W,V,N),J={environment:j};if(typeof I=="string"&&m[I]){var te=m[I].snapshot;return te?{error:null,relayContext:J,renderProps:E(null,te,R,F),snapshot:te,requestCacheKey:I}:{error:null,relayContext:J,renderProps:S(),snapshot:null,requestCacheKey:I}}try{var H=R.lookupInStore(j,K,w.fetchPolicy),M=R.fetch({environment:j,onDataChange:null,operation:K}),G=M||H;return I=I||$(W.params,w.variables),m[I]={queryFetcher:R,snapshot:G},G?{error:null,relayContext:J,renderProps:E(null,G,R,F),snapshot:G,requestCacheKey:I}:{error:null,relayContext:J,renderProps:S(),snapshot:null,requestCacheKey:I}}catch(B){return{error:B,relayContext:J,renderProps:E(B,null,R,F),snapshot:null,requestCacheKey:I}}}else{R.dispose();var X={environment:j};return{error:null,relayContext:X,renderProps:_(),requestCacheKey:null}}}return QA=v,QA}var zA,Zq;function Y4e(){if(Zq)return zA;Zq=1;var e=$n().default,t=e(TC()),n=e(H3()),r=e(Ui()),i=e(kC()),a=e(PC()),u=e(mc()),l=["componentRef","__relayContext","__rootIsQueryRenderer"],o=WF(),s=oee(),c=MC(),d=c.getContainerName,h=hy(),m=eN(),y=JF(),v=y.assertRelayContext,S=ed(),_=ir(),E=ii(),$=E.Observable,k=E.createFragmentSpecResolver,A=E.createOperationDescriptor,w=E.getDataIDsFromObject,R=E.getRequest,F=E.getVariablesFromObject,I=E.isScalarAndEqual;Vi();function U(N,j,W){var K,J=d(N);return K=(function(te){(0,a.default)(H,te);function H(G){var X,B;B=te.call(this,G)||this,(0,u.default)((0,i.default)(B),"_handleFragmentDataUpdate",function(){var de=B.state.resolver;B.setState(function(ge){return de===ge.resolver?{data:ge.resolver.resolve()}:null})}),(0,u.default)((0,i.default)(B),"_refetch",function(de,ge,ye,me){if(B._isUnmounted)return{dispose:function(){}};var Ee=v(B.props.__relayContext),se=Ee.environment,pe=s(j,B.props),ve=typeof de=="function"?de(B._getFragmentVariables()):de;ve=(0,r.default)((0,r.default)({},pe),ve);var we=ge?(0,r.default)((0,r.default)({},ve),ge):ve,Me=me?{force:!!me.force}:void 0;Me!=null&&(me==null?void 0:me.metadata)!=null&&(Me.metadata=me==null?void 0:me.metadata);var Re=typeof ye=="function"?{next:ye,error:ye}:ye||{},et=R(W),tt=A(et,ve,Me);B.state.localVariables=ve,B._refetchSubscription&&B._refetchSubscription.unsubscribe();var Oe,$e=B._getQueryFetcher().lookupInStore(se,tt,me==null?void 0:me.fetchPolicy);return $e!=null?(B.state.resolver.setVariables(we,tt.request.node),B.setState(function(Ae){return{data:Ae.resolver.resolve(),contextForChildren:{environment:B.props.__relayContext.environment}}},function(){Re.next&&Re.next(),Re.complete&&Re.complete()}),{dispose:function(){}}):(B._getQueryFetcher().execute({environment:se,operation:tt,preservePreviousReferences:!0}).mergeMap(function(Ae){return B.state.resolver.setVariables(we,tt.request.node),$.create(function(He){return B.setState(function(je){return{data:je.resolver.resolve(),contextForChildren:{environment:B.props.__relayContext.environment}}},function(){He.next(),He.complete()})})}).finally(function(){B._refetchSubscription===Oe&&(B._refetchSubscription=null)}).subscribe((0,r.default)((0,r.default)({},Re),{},{start:function(He){B._refetchSubscription=Oe=He,Re.start&&Re.start(He)}})),{dispose:function(){Oe&&Oe.unsubscribe()}})});var ne=v(G.__relayContext),oe=(X=G.__rootIsQueryRenderer)!==null&&X!==void 0?X:!1;B._refetchSubscription=null;var ie=k(ne,J,j,G,oe);return B.state={data:ie.resolve(),localVariables:null,prevProps:G,prevPropsContext:ne,contextForChildren:ne,relayProp:q(ne.environment,B._refetch),resolver:ie},B._isUnmounted=!1,B}var M=H.prototype;return M.componentDidMount=function(){this._isUnmounted=!1,this._subscribeToNewResolverAndRerenderIfStoreHasChanged()},M.componentDidUpdate=function(X,B){this.state.resolver!==B.resolver?(B.resolver.dispose(),this._queryFetcher&&this._queryFetcher.dispose(),this._refetchSubscription&&this._refetchSubscription.unsubscribe(),this._subscribeToNewResolverAndRerenderIfStoreHasChanged()):this._rerenderIfStoreHasChanged()},H.getDerivedStateFromProps=function(X,B){var ne,oe=B.prevProps,ie=v(X.__relayContext),de=(ne=X.__rootIsQueryRenderer)!==null&&ne!==void 0?ne:!1,ge=w(j,oe),ye=w(j,X),me=s(j,oe),Ee=s(j,X),se=B.resolver;if(B.prevPropsContext.environment!==ie.environment||!S(me,Ee)||!S(ge,ye))return se=k(ie,J,j,X,de),{data:se.resolve(),localVariables:null,prevProps:X,prevPropsContext:ie,contextForChildren:ie,relayProp:q(ie.environment,B.relayProp.refetch),resolver:se};B.localVariables||se.setProps(X);var pe=se.resolve();return pe!==B.data?{data:pe,prevProps:X}:null},M.componentWillUnmount=function(){this._isUnmounted=!0,this.state.resolver.dispose(),this._queryFetcher&&this._queryFetcher.dispose(),this._refetchSubscription&&this._refetchSubscription.unsubscribe()},M.shouldComponentUpdate=function(X,B){if(B.data!==this.state.data||B.relayProp!==this.state.relayProp)return!0;for(var ne=Object.keys(X),oe=0;oe<ne.length;oe++){var ie=ne[oe];if(ie==="__relayContext"){if(this.state.prevPropsContext.environment!==B.prevPropsContext.environment)return!0}else if(!j.hasOwnProperty(ie)&&!I(X[ie],this.props[ie]))return!0}return!1},M._rerenderIfStoreHasChanged=function(){var X=this.state,B=X.data,ne=X.resolver,oe=ne.resolve();B!==oe&&this.setState({data:oe})},M._subscribeToNewResolverAndRerenderIfStoreHasChanged=function(){var X=this.state,B=X.data,ne=X.resolver,oe=ne.resolve();ne.setCallback(this.props,this._handleFragmentDataUpdate),B!==oe&&this.setState({data:oe})},M._getFragmentVariables=function(){return F(j,this.props)},M._getQueryFetcher=function(){return this._queryFetcher||(this._queryFetcher=new m),this._queryFetcher},M.render=function(){var X=this.props,B=X.componentRef;X.__relayContext,X.__rootIsQueryRenderer;var ne=(0,n.default)(X,l),oe=this.state,ie=oe.relayProp,de=oe.contextForChildren;return _.createElement(h.Provider,{value:de},_.createElement(N,(0,t.default)({},ne,this.state.data,{ref:B,relay:ie})))},H})(_.Component),(0,u.default)(K,"displayName",J),K}function q(N,j){return{environment:N,refetch:j}}function V(N,j,W){return o(N,j,function(K,J){return U(K,J,W)})}return zA={createContainer:V},zA}var qA,Hq;function FC(){if(Hq)return qA;Hq=1;var e=ir(),t=e.createContext({wrapPrepareQueryResource:function(r){return r()}});return qA=t,qA}var VA,Xq;function co(){if(Xq)return VA;Xq=1;var e=hy(),t=un(),n=ir(),r=n.useContext;function i(){var a=r(e);return a==null&&t(!1),a.environment}return VA=i,VA}var UA,Yq;function G4e(){if(Yq)return UA;Yq=1;var e=FC(),t=co(),n=ir(),r=ir(),i=r.useContext,a=r.useEffect;Vi();function u(l){var o=l.entryPointReference,s=l.props,c=o.getComponent,d=o.queries,h=o.entryPoints,m=o.extraProps,y=o.rootModuleID,v=c(),S=i(e),_=t();return a(function(){_.__log({name:"entrypoint.root.consume",profilerContext:S,rootModuleID:y})},[_,S,y]),n.createElement(v,{entryPoints:h,extraProps:m,props:s,queries:d})}return UA=u,UA}var jA,Gq;function NC(){if(Gq)return jA;Gq=1;var e=$n().default,t=e(Ui()),n=un(),r=ii(),i=r.__internal.fetchQueryDeduped,a=r.Observable,u=r.PreloadableQueryRegistry,l=r.ReplaySubject,o=r.createOperationDescriptor,s=r.getRequest,c=r.getRequestIdentifier,d=100001;function h(m,y,v,S,_){var E;d++;var $=(E=S==null?void 0:S.fetchPolicy)!==null&&E!==void 0?E:"store-or-network",k=(0,t.default)((0,t.default)({},S==null?void 0:S.networkCacheConfig),{},{force:!0}),A,w=!1,R=function(pe,ve){return w=!0,m.executeWithSource({operation:pe,source:ve})},F=new l,I=a.create(function(se){return F.subscribe(se)}),U,q=null,V=!1,N=function(pe,ve){V=!0;var we=new l,Me="raw-network-request-"+c(pe,v),Re=i(m,Me,function(){var Oe=m.getNetwork();return Oe.execute(pe,v,k,void 0,void 0,void 0,void 0,ve)}),et=Re.subscribe({error:function($e){q=$e,we.error($e)},next:function($e){we.next($e)},complete:function(){we.complete()}}),tt=et.unsubscribe;return U=tt,a.create(function(Oe){var $e=we.subscribe(Oe);return function(){$e.unsubscribe(),U()}})},j,W=function(pe,ve){V=!0;var we=i(m,pe.request.identifier,ve).subscribe({error:function(Re){F.error(Re)},next:function(Re){F.next(Re)},complete:function(){F.complete()}});j=we.unsubscribe},K=function(pe){var ve=o(pe,v,k);if(A=m.retain(ve),$!=="store-only"){var we=$!=="store-or-network"||m.check(ve).status!=="available";we&&W(ve,function(){var Me=N(pe.params,function(){return m.check(ve)}),Re=R(ve,Me);return Re})}},J,te,H;if(y.kind==="PreloadableConcreteRequest"){var M=y;J=M.params;var G=J;H=G.id,H===null&&n(!1);var X=u.get(H);if(X!=null)K(X);else{var B=$==="store-only"?null:N(J),ne=u.onLoad(H,function(se){te();var pe=o(se,v,k);A=m.retain(pe),B!=null&&W(pe,function(){return R(pe,B)})});te=ne.dispose}}else{var oe=y,ie=s(oe);J=ie.params,H=J.cacheID!=null?J.cacheID:J.id,K(ie)}var de=!1,ge=!1,ye=!1,me=function(){ge||(A&&A.dispose(),ge=!0)},Ee=function(){ye||(w?j&&j():U&&U(),te&&te(),ye=!0)};return{kind:"PreloadedQuery",environment:m,environmentProviderOptions:_,dispose:function(){de||(me(),Ee(),de=!0)},releaseQuery:me,cancelNetworkRequest:Ee,fetchKey:d,id:H,get isDisposed(){return de||ge},get networkError(){return q},name:J.name,networkCacheConfig:k,fetchPolicy:$,source:V?I:void 0,variables:v}}return jA={loadQuery:h},jA}var ZA,Kq;function see(){if(Kq)return ZA;Kq=1;var e=NC(),t=e.loadQuery;function n(r,i,a){var u=null;i.root.getModuleIfRequired()==null&&(u=i.root.load());var l=i.getPreloadProps(a),o=l.queries,s=l.entryPoints,c=l.extraProps,d={},h={};if(o!=null){var m=Object.keys(o);m.forEach(function(S){var _=o[S];if(_!=null){var E=_.environmentProviderOptions,$=_.options,k=_.parameters,A=_.variables;if(($==null?void 0:$.includeIf)!==!1){var w=r.getEnvironment(E);d[S]=t(w,k,A,{fetchPolicy:$==null?void 0:$.fetchPolicy,networkCacheConfig:$==null?void 0:$.networkCacheConfig,__nameForWarning:"loadEntryPoint"},E)}}})}if(s!=null){var y=Object.keys(s);y.forEach(function(S){var _=s[S];if(_!=null){var E=_.entryPoint,$=_.entryPointParams;h[S]=n(r,E,$)}})}var v=!1;return{dispose:function(){v||(d!=null&&Object.values(d).forEach(function(_){var E=_.dispose;E()}),h!=null&&Object.values(h).forEach(function(_){var E=_.dispose;E()}),v=!0)},entryPoints:h,extraProps:c??null,getComponent:function(){var _=i.root.getModuleIfRequired();if(_==null){var E;throw u=(E=u)!==null&&E!==void 0?E:i.root.load(),u}var $=_.default!=null?_.default:_;return $},get isDisposed(){return v},queries:d,rootModuleID:i.root.getModuleId()}}return ZA=n,ZA}var HA,Wq;function K4e(){if(Wq)return HA;Wq=1;var e=hy(),t=ir(),n=t.useMemo;function r(i){var a=i.children,u=i.environment,l=i.getEnvironmentForActor,o=n(function(){return{environment:u,getEnvironmentForActor:l}},[u,l]);return t.createElement(e.Provider,{value:o},a)}return HA=r,HA}var XA,Jq;function W4e(){if(Jq)return XA;Jq=1;var e=un(),t=(function(){function r(a){this._capacity=a,this._capacity>0||e(!1),this._map=new Map}var i=r.prototype;return i.set=function(u,l){if(this._map.delete(u),this._map.set(u,l),this._map.size>this._capacity){var o=this._map.keys().next();o.done||this._map.delete(o.value)}},i.get=function(u){var l=this._map.get(u);return l!=null&&(this._map.delete(u),this._map.set(u,l)),l},i.has=function(u){return this._map.has(u)},i.delete=function(u){this._map.delete(u)},i.size=function(){return this._map.size},i.capacity=function(){return this._capacity-this._map.size},i.clear=function(){this._map.clear()},r})();function n(r){return new t(r)}return XA={create:n},XA}var YA,eV;function J4e(){if(eV)return YA;eV=1;var e=$n().default,t=e(mc());Vi();var n=300*1e3,r=(function(){function i(u){var l=this;(0,t.default)(this,"_retainCount",0),(0,t.default)(this,"_retainDisposable",null),(0,t.default)(this,"_releaseTemporaryRetain",null),this._retain=function(o){return l._retainCount++,l._retainCount===1&&(l._retainDisposable=u(o)),{dispose:function(){l._retainCount=Math.max(0,l._retainCount-1),l._retainCount===0&&l._retainDisposable!=null&&(l._retainDisposable.dispose(),l._retainDisposable=null)}}}}var a=i.prototype;return a.temporaryRetain=function(l){var o=this,s;if(l.isServer())return{dispose:function(){}};var c=this._retain(l),d=null,h=function(){clearTimeout(d),d=null,o._releaseTemporaryRetain=null,c.dispose()};return d=setTimeout(h,n),(s=this._releaseTemporaryRetain)===null||s===void 0||s.call(this),this._releaseTemporaryRetain=h,{dispose:function(){var y;(y=o._releaseTemporaryRetain)===null||y===void 0||y.call(o)}}},a.permanentRetain=function(l){var o=this._retain(l);return this.releaseTemporaryRetain(),o},a.releaseTemporaryRetain=function(){var l;(l=this._releaseTemporaryRetain)===null||l===void 0||l.call(this),this._releaseTemporaryRetain=null},a.getRetainCount=function(){return this._retainCount},i})();return YA=r,YA}var GA,tV;function cS(){if(tV)return GA;tV=1;var e=$n().default,t=e(Ui()),n=e(mc()),r=W4e(),i=J4e(),a=un(),u=ii(),l=u.isPromise;Vi();var o=1e3,s="store-or-network",c="store-and-network",d=typeof WeakMap=="function";function h(A){return A.request.node.params.metadata.live!==void 0}function m(A,w,R,F,I){var U=R??(h(w)?c:s),q=F??A.UNSTABLE_getDefaultRenderPolicy(),V="".concat(U,"-").concat(q,"-").concat(w.request.identifier);return I!=null?"".concat(V,"-").concat(I):V}function y(A,w){var R={__id:A.fragment.dataID,__fragments:(0,n.default)({},A.fragment.node.name,A.request.variables),__fragmentOwner:A.request};return{cacheIdentifier:w,fragmentNode:A.request.node.fragment,fragmentRef:R,operation:A}}var v=2e5;function S(A,w,R,F,I,U){var q=h(w),V=F,N=I,j=new i(function(K){var J=K.retain(w);return{dispose:function(){q&&N!=null&&N.unsubscribe(),J.dispose(),U(W)}}}),W={cacheIdentifier:A,id:v++,processedPayloadsCount:0,operationAvailability:R,getValue:function(){return V},setValue:function(J){V=J},setNetworkSubscription:function(J){q&&N!=null&&N.unsubscribe(),N=J},temporaryRetain:function(J){return j.temporaryRetain(J)},permanentRetain:function(J){return j.permanentRetain(J)},releaseTemporaryRetain:function(){j.releaseTemporaryRetain()}};return W}var _=(function(){function A(R){var F=this;(0,n.default)(this,"_clearCacheEntry",function(I){F._cache.delete(I.cacheIdentifier)}),this._environment=R,this._cache=r.create(o)}var w=A.prototype;return w.prepare=function(F,I,U,q,V,N,j){var W=m(this._environment,F,U,q,N);return this.prepareWithIdentifier(W,F,I,U,q,V,j)},w.prepareWithIdentifier=function(F,I,U,q,V,N,j){var W=this._environment,K=q??(h(I)?c:s),J=V??W.UNSTABLE_getDefaultRenderPolicy(),te=this._cache.get(F),H=null,M=te!=null;te==null&&(te=this._fetchAndSaveQuery(F,I,U,K,J,j,(0,t.default)((0,t.default)({},N),{},{unsubscribe:function(B){H!=null&&H.dispose();var ne=N==null?void 0:N.unsubscribe;ne&&ne(B)}}))),H=te.temporaryRetain(W);var G=te.getValue();if(l(G))throw W.__log({name:"suspense.query",fetchPolicy:K,isPromiseCached:M,operation:I,queryAvailability:te.operationAvailability,renderPolicy:J}),G;if(G instanceof Error)throw G;return G},w.retain=function(F,I){var U=this._environment,q=F.cacheIdentifier,V=F.operation,N=this._getOrCreateCacheEntry(q,V,null,F,null),j=N.permanentRetain(U);return U.__log({name:"queryresource.retain",profilerContext:I,resourceID:N.id}),{dispose:function(){j.dispose()}}},w.releaseTemporaryRetain=function(F){var I=this._cache.get(F.cacheIdentifier);I!=null&&I.releaseTemporaryRetain()},w.TESTS_ONLY__getCacheEntry=function(F,I,U,q){var V=this._environment,N=m(V,F,I,U,q);return this._cache.get(N)},w._getOrCreateCacheEntry=function(F,I,U,q,V){var N=this._cache.get(F);return N==null&&(N=S(F,I,U,q,V,this._clearCacheEntry),this._cache.set(F,N)),N},w._fetchAndSaveQuery=function(F,I,U,q,V,N,j){var W=this,K=this._environment,J=K.check(I),te=J.status,H=te==="available",M=H||V==="partial"&&te!=="stale",G,X,B=function(){};switch(q){case"store-only":{G=!1,X=!0;break}case"store-or-network":{G=!H,X=M;break}case"store-and-network":{G=!0,X=M;break}case"network-only":default:{G=!0,X=!1;break}}if(X){var ne=y(I,F),oe=S(F,I,J,ne,null,this._clearCacheEntry);this._cache.set(F,oe)}if(G){var ie=y(I,F),de;U.subscribe({start:function(pe){de=pe;var ve=W._cache.get(F);ve&&ve.setNetworkSubscription(de);var we=j==null?void 0:j.start;if(we){var Me=(0,t.default)((0,t.default)({},pe),{},{unsubscribe:function(){h(I)&&pe.unsubscribe()}});we(Me)}},next:function(){var pe=W._getOrCreateCacheEntry(F,I,J,ie,de);pe.processedPayloadsCount+=1,pe.setValue(ie),B();var ve=j==null?void 0:j.next;if(ve!=null){var we=K.lookup(I.fragment);ve(we)}},error:function(pe){var ve=W._getOrCreateCacheEntry(F,I,J,pe,de);ve.processedPayloadsCount===0&&ve.setValue(pe),B(),de=null,ve.setNetworkSubscription(null);var we=j==null?void 0:j.error;we&&we(pe)},complete:function(){B(),de=null;var pe=W._cache.get(F);pe&&pe.setNetworkSubscription(null);var ve=j==null?void 0:j.complete;ve&&ve()},unsubscribe:j==null?void 0:j.unsubscribe});var ge=this._cache.get(F);if(!ge){var ye=new Promise(function(se){B=se});ye.displayName="Relay("+I.fragment.node.name+")",ge=S(F,I,J,ye,de,this._clearCacheEntry),this._cache.set(F,ge)}}else{var me=j==null?void 0:j.complete;me&&me()}var Ee=this._cache.get(F);return Ee==null&&a(!1),K.__log({name:"queryresource.fetch",resourceID:Ee.id,operation:I,profilerContext:N,fetchPolicy:q,renderPolicy:V,queryAvailability:J,shouldFetch:G}),Ee},A})();function E(A){return new _(A)}var $=d?new WeakMap:new Map;function k(A){var w=$.get(A);if(w)return w;var R=E(A);return $.set(A,R),R}return GA={createQueryResource:E,getQueryResourceForEnvironment:k,getQueryCacheIdentifier:m},GA}var KA,nV;function cee(){if(nV)return KA;nV=1;var e=ir(),t=e.useMemo;function n(r){return t(function(){return{current:r}},[])}return KA=n,KA}var WA,rV;function fee(){if(rV)return WA;rV=1;var e=cee(),t=ir(),n=t.useCallback,r=t.useEffect;function i(){var a=e(null),u=e(!1),l=n(function(){a.current!=null&&(a.current.unsubscribe(),a.current=null),u.current=!1},[]),o=n(function(c){a.current=c,u.current=!0},[]),s=n(function(){a.current=null,u.current=!1},[]);return r(function(){return l},[l]),{isFetchingRef:u,startFetch:o,disposeFetch:l,completeFetch:s}}return WA=i,WA}var JA,iV;function e2e(){if(iV)return JA;iV=1;var e=ir(),t=ii(),n=t.__internal.createRelayLoggingContext;return JA=n(e),JA}var eT,aV;function dee(){if(aV)return eT;aV=1;var e=e2e(),t=ir(),n=t.useContext;function r(){return n(e)}return eT=r,eT}var tT,uV;function t2e(){if(uV)return tT;uV=1;var e=$n().default,t=e(Ui()),n=e(Cu()),r=e(sy()),i=cS(),a=i.getQueryResourceForEnvironment,u=co(),l=dee(),o=un(),s=ir();s.useDebugValue;var c=s.useEffect,d=s.useMemo,h=s.useRef,m=s.useState,y=ii(),v=y.__internal,S=v.fetchQuery,_=v.getPromiseForActiveRequest,E=y.RelayFeatureFlags,$=y.areEqualSelectors,k=y.createOperationDescriptor,A=y.getPendingOperationsForFragment,w=y.getSelector,R=y.getVariablesFromFragment,F=y.handlePotentialSnapshotErrors,I=y.recycleNodesInto;Vi();function U(H){return H.kind==="bailout"?!1:H.kind==="singular"?H.snapshot.isMissingData:H.snapshots.some(function(M){return M.isMissingData})}function q(H){if(H.kind==="bailout")return null;if(H.kind==="singular"){var M;return(M=H.snapshot.missingClientEdges)!==null&&M!==void 0?M:null}else{var G=null,X=(0,n.default)(H.snapshots),B;try{for(X.s();!(B=X.n()).done;){var ne=B.value;if(ne.missingClientEdges){var oe;G=(oe=G)!==null&&oe!==void 0?oe:[];var ie=(0,n.default)(ne.missingClientEdges),de;try{for(ie.s();!(de=ie.n()).done;){var ge=de.value;G.push(ge)}}catch(ye){ie.e(ye)}finally{ie.f()}}}}catch(ye){X.e(ye)}finally{X.f()}return G}}function V(H){if(H.kind==="bailout")return null;if(H.kind==="singular"){var M;return(M=H.snapshot.missingLiveResolverFields)!==null&&M!==void 0?M:null}else{var G=null,X=(0,n.default)(H.snapshots),B;try{for(X.s();!(B=X.n()).done;){var ne=B.value;if(ne.missingLiveResolverFields){var oe;G=(oe=G)!==null&&oe!==void 0?oe:[];var ie=(0,n.default)(ne.missingLiveResolverFields),de;try{for(ie.s();!(de=ie.n()).done;){var ge=de.value;G.push(ge)}}catch(ye){ie.e(ye)}finally{ie.f()}}}}catch(ye){X.e(ye)}finally{X.f()}return G}}function N(H,M,G){if(M.kind==="singular")F(H,M.snapshot.fieldErrors,G);else if(M.kind==="plural"){var X=(0,n.default)(M.snapshots),B;try{for(X.s();!(B=X.n()).done;){var ne=B.value;F(H,ne.fieldErrors,G)}}catch(oe){X.e(oe)}finally{X.f()}}}function j(H,M){if(M.kind==="bailout")return null;var G=H.getStore().getEpoch();if(G===M.epoch)return null;if(M.kind==="singular"){var X=H.lookup(M.snapshot.selector),B=I(M.snapshot.data,X.data),ne={data:B,isMissingData:X.isMissingData,missingClientEdges:X.missingClientEdges,missingLiveResolverFields:X.missingLiveResolverFields,seenRecords:X.seenRecords,selector:X.selector,fieldErrors:X.fieldErrors};return[B!==M.snapshot.data,{kind:"singular",snapshot:ne,epoch:G}]}else{for(var oe=!1,ie=[],de=0;de<M.snapshots.length;de++){var ge=M.snapshots[de],ye=H.lookup(ge.selector),me=I(ge.data,ye.data),Ee={data:me,isMissingData:ye.isMissingData,missingClientEdges:ye.missingClientEdges,missingLiveResolverFields:ye.missingLiveResolverFields,seenRecords:ye.seenRecords,selector:ye.selector,fieldErrors:ye.fieldErrors};me!==ge.data&&(oe=!0),ie.push(Ee)}return ie.length!==M.snapshots.length&&o(!1),[oe,{kind:"plural",snapshots:ie,epoch:G}]}}function W(H,M,G,X,B){var ne=R(M,G),oe=(0,t.default)((0,t.default)({},ne),{},{id:X.clientEdgeDestinationID}),ie=k(X.request,oe,B==null?void 0:B.networkCacheConfig),de=a(H),ge=de.prepare(ie,S(H,ie),B==null?void 0:B.fetchPolicy);return[ge,_(H,ie.request)]}function K(H,M,G,X){if(M.kind==="bailout")return function(){};if(M.kind==="singular"){var B=H.subscribe(M.snapshot,function(oe){G(function(ie){if(ie.kind!=="singular"||ie.snapshot.selector!==oe.selector){var de=j(H,ie);if(de!=null){var ge=de[0],ye=de[1];return H.__log({name:"useFragment.subscription.missedUpdates",hasDataChanges:ge}),X.current=ge,ge?ye:ie}else return ie}return X.current=!0,{kind:"singular",snapshot:oe,epoch:H.getStore().getEpoch()}})});return function(){B.dispose()}}else{var ne=M.snapshots.map(function(oe,ie){return H.subscribe(oe,function(de){G(function(ge){var ye;if(ge.kind!=="plural"||((ye=ge.snapshots[ie])===null||ye===void 0?void 0:ye.selector)!==de.selector){var me=j(H,ge);if(me!=null){var Ee=me[0],se=me[1];return H.__log({name:"useFragment.subscription.missedUpdates",hasDataChanges:Ee}),X.current=X.current||Ee,Ee?se:ge}else return ge}var pe=(0,r.default)(ge.snapshots);return pe[ie]=de,X.current=!0,{kind:"plural",snapshots:pe,epoch:H.getStore().getEpoch()}})})});return function(){var oe=(0,n.default)(ne),ie;try{for(oe.s();!(ie=oe.n()).done;){var de=ie.value;de.dispose()}}catch(ge){oe.e(ge)}finally{oe.f()}}}}function J(H,M){return M==null?{kind:"bailout"}:M.kind==="PluralReaderSelector"?{kind:"plural",snapshots:M.selectors.map(function(G){return H.lookup(G)}),epoch:H.getStore().getEpoch()}:{kind:"singular",snapshot:H.lookup(M),epoch:H.getStore().getEpoch()}}function te(H,M,G,X){var B,ne,oe=d(function(){return w(H,M)},[H,M]),ie=(H==null||(B=H.metadata)===null||B===void 0?void 0:B.plural)===!0;ie?M==null||Array.isArray(M)||o(!1):Array.isArray(M)&&o(!1),M==null||ie&&Array.isArray(M)&&M.length===0||oe!=null||o(!1);var de=u(),ge;E.ENABLE_UI_CONTEXT_ON_RELAY_LOGGER&&(ge=l());var ye=m(function(){return J(de,oe)}),me=ye[0],Ee=ye[1],se=me,pe=m(se),ve=pe[0],we=pe[1],Me=ve,Re=m(oe),et=Re[0],tt=Re[1],Oe=m(de),$e=Oe[0],Ae=Oe[1];if(!$(oe,et)||de!==$e){tt(oe),Ae(de);var He=J(de,oe);Ee(He),we(He),se=He,Me=He}var je=h(!1);if(c(function(){je.current=oe},[oe]),((ne=H.metadata)===null||ne===void 0?void 0:ne.hasClientEdges)===!0||E.CHECK_ALL_FRAGMENTS_FOR_MISSING_CLIENT_EDGES){var Ye=d(function(){var ht=q(se),kt,Qt=[];if(ht!=null&&ht.length){kt=[];var en=(0,n.default)(ht),In;try{for(en.s();!(In=en.n()).done;){var rn=In.value,Xn=W(de,H,M,rn,X),br=Xn[0],Or=Xn[1];kt.push(br),Or!=null&&Qt.push(Or)}}catch(Cr){en.e(Cr)}finally{en.f()}}return[kt,Qt]},[se,de,H,M,X]),Ne=Ye[0],ze=Ye[1];if(ze.length)throw Promise.all(ze);c(function(){var ht=a(de);if(Ne!=null&&Ne.length){var kt=[],Qt=(0,n.default)(Ne),en;try{for(Qt.s();!(en=Qt.n()).done;){var In=en.value;kt.push(ht.retain(In))}}catch(rn){Qt.e(rn)}finally{Qt.f()}return function(){for(var rn=0,Xn=kt;rn<Xn.length;rn++){var br=Xn[rn];br.dispose()}}}},[de,Ne])}if(U(se)){var at=V(se);if(at!=null&&at.length>0)throw Promise.all(at.map(function(ht){return de.getStore().getLiveResolverPromise(ht)}));if(E.ENABLE_RELAY_OPERATION_TRACKER_SUSPENSE||de!==$e||!je.current||!$(je.current,oe)){oe==null&&o(!1);var mt=oe.kind==="PluralReaderSelector"?oe.selectors[0].owner:oe.owner,Mt=A(de,H,mt);if(Mt)throw Mt.promise}}N(de,se,ge);var bt=h(!1);if(c(function(){var ht=Me,kt=j(de,Me);if(kt!==null){var Qt=kt[0],en=kt[1];Qt&&Ee(en),ht=en}return K(de,ht,Ee,bt)},[de,Me]),bt.current){var zt=j(de,se);if(zt!=null){var qt=zt[0],Xt=zt[1];qt&&(Ee(Xt),se=Xt)}bt.current=!1}var nn;if(ie){var Fe=M==null;nn=d(function(){return se.kind==="bailout"?Fe?null:[]:(se.kind!=="plural"&&o(!1),se.snapshots.map(function(ht){return ht.data}))},[se,Fe])}else se.kind==="bailout"?nn=null:(se.kind!=="singular"&&o(!1),nn=se.snapshot.data);return E.LOG_MISSING_RECORDS_IN_PROD&&M!=null&&(nn===void 0||Array.isArray(nn)&&nn.length>0&&nn.every(function(ht){return ht===void 0})),nn}return tT=te,tT}var nT,lV;function n2e(){if(lV)return nT;lV=1;var e=$n().default,t=e(Ui()),n=e(Cu()),r=e(sy()),i=cS(),a=i.getQueryResourceForEnvironment,u=co(),l=dee(),o=un(),s=ir();s.useDebugValue;var c=s.useEffect,d=s.useMemo,h=s.useRef,m=s.useState,y=ii(),v=y.__internal,S=v.fetchQuery,_=v.getPromiseForActiveRequest,E=y.RelayFeatureFlags,$=y.areEqualSelectors,k=y.createOperationDescriptor,A=y.getPendingOperationsForFragment,w=y.getSelector,R=y.getVariablesFromFragment,F=y.handlePotentialSnapshotErrors,I=y.recycleNodesInto;Vi();function U(H){return H.kind==="bailout"?!1:H.kind==="singular"?H.snapshot.isMissingData:H.snapshots.some(function(M){return M.isMissingData})}function q(H){if(H.kind==="bailout")return null;if(H.kind==="singular"){var M;return(M=H.snapshot.missingClientEdges)!==null&&M!==void 0?M:null}else{var G=null,X=(0,n.default)(H.snapshots),B;try{for(X.s();!(B=X.n()).done;){var ne=B.value;if(ne.missingClientEdges){var oe;G=(oe=G)!==null&&oe!==void 0?oe:[];var ie=(0,n.default)(ne.missingClientEdges),de;try{for(ie.s();!(de=ie.n()).done;){var ge=de.value;G.push(ge)}}catch(ye){ie.e(ye)}finally{ie.f()}}}}catch(ye){X.e(ye)}finally{X.f()}return G}}function V(H){if(H.kind==="bailout")return null;if(H.kind==="singular"){var M;return(M=H.snapshot.missingLiveResolverFields)!==null&&M!==void 0?M:null}else{var G=null,X=(0,n.default)(H.snapshots),B;try{for(X.s();!(B=X.n()).done;){var ne=B.value;if(ne.missingLiveResolverFields){var oe;G=(oe=G)!==null&&oe!==void 0?oe:[];var ie=(0,n.default)(ne.missingLiveResolverFields),de;try{for(ie.s();!(de=ie.n()).done;){var ge=de.value;G.push(ge)}}catch(ye){ie.e(ye)}finally{ie.f()}}}}catch(ye){X.e(ye)}finally{X.f()}return G}}function N(H,M,G){if(M.kind==="singular")F(H,M.snapshot.fieldErrors,G);else if(M.kind==="plural"){var X=(0,n.default)(M.snapshots),B;try{for(X.s();!(B=X.n()).done;){var ne=B.value;F(H,ne.fieldErrors,G)}}catch(oe){X.e(oe)}finally{X.f()}}}function j(H,M){if(M.kind==="bailout")return null;var G=H.getStore().getEpoch();if(G===M.epoch)return null;if(M.kind==="singular"){var X=H.lookup(M.snapshot.selector),B=I(M.snapshot.data,X.data),ne={data:B,isMissingData:X.isMissingData,missingClientEdges:X.missingClientEdges,missingLiveResolverFields:X.missingLiveResolverFields,seenRecords:X.seenRecords,selector:X.selector,fieldErrors:X.fieldErrors};return[B!==M.snapshot.data,{kind:"singular",snapshot:ne,epoch:G,selector:M.selector,environment:M.environment}]}else{for(var oe=!1,ie=[],de=0;de<M.snapshots.length;de++){var ge=M.snapshots[de],ye=H.lookup(ge.selector),me=I(ge.data,ye.data),Ee={data:me,isMissingData:ye.isMissingData,missingClientEdges:ye.missingClientEdges,missingLiveResolverFields:ye.missingLiveResolverFields,seenRecords:ye.seenRecords,selector:ye.selector,fieldErrors:ye.fieldErrors};me!==ge.data&&(oe=!0),ie.push(Ee)}return ie.length!==M.snapshots.length&&o(!1),[oe,{kind:"plural",snapshots:ie,epoch:G,selector:M.selector,environment:M.environment}]}}function W(H,M,G,X,B){var ne=R(M,G),oe=(0,t.default)((0,t.default)({},ne),{},{id:X.clientEdgeDestinationID}),ie=k(X.request,oe,B==null?void 0:B.networkCacheConfig),de=a(H),ge=de.prepare(ie,S(H,ie),B==null?void 0:B.fetchPolicy),ye=_(H,ie.request);return ye!=null&&ye.displayName==null&&(ye.displayName=X.request.params.name),[ge,ye]}function K(H,M,G){if(M.kind==="bailout")return function(){};if(M.kind==="singular"){var X=H.subscribe(M.snapshot,function(ne){G(function(oe){var ie=null;if(oe.kind!=="singular"||oe.snapshot.selector!==ne.selector||oe.environment!==H){var de=j(oe.environment,oe);if(de!=null){var ge=de[0],ye=de[1];H.__log({name:"useFragment.subscription.missedUpdates",hasDataChanges:ge}),ie=ge?ye:oe}else ie=oe}else ie={kind:"singular",snapshot:ne,epoch:H.getStore().getEpoch(),selector:M.selector,environment:M.environment};return ie})});return function(){X.dispose()}}else{var B=M.snapshots.map(function(ne,oe){return H.subscribe(ne,function(ie){G(function(de){var ge,ye=null;if(de.kind!=="plural"||((ge=de.snapshots[oe])===null||ge===void 0?void 0:ge.selector)!==ie.selector||de.environment!==H){var me=j(de.environment,de);if(me!=null){var Ee=me[0],se=me[1];H.__log({name:"useFragment.subscription.missedUpdates",hasDataChanges:Ee}),ye=Ee?se:de}else ye=de}else{var pe=(0,r.default)(de.snapshots);pe[oe]=ie,ye={kind:"plural",snapshots:pe,epoch:H.getStore().getEpoch(),selector:M.selector,environment:M.environment}}return ye})})});return function(){var ne=(0,n.default)(B),oe;try{for(ne.s();!(oe=ne.n()).done;){var ie=oe.value;ie.dispose()}}catch(de){ne.e(de)}finally{ne.f()}}}}function J(H,M){return M==null?{kind:"bailout",environment:H}:M.kind==="PluralReaderSelector"?{kind:"plural",snapshots:M.selectors.map(function(G){return H.lookup(G)}),epoch:H.getStore().getEpoch(),selector:M,environment:H}:{kind:"singular",snapshot:H.lookup(M),epoch:H.getStore().getEpoch(),selector:M,environment:H}}function te(H,M,G,X){var B,ne,oe=d(function(){return w(H,M)},[H,M]),ie=(H==null||(B=H.metadata)===null||B===void 0?void 0:B.plural)===!0;ie?M==null||Array.isArray(M)||o(!1):Array.isArray(M)&&o(!1),M==null||ie&&Array.isArray(M)&&M.length===0||oe!=null||o(!1);var de=u(),ge;E.ENABLE_UI_CONTEXT_ON_RELAY_LOGGER&&(ge=l());var ye=m(function(){return J(de,oe)}),me=ye[0],Ee=ye[1],se=me,pe=se.environment;if(!$(oe,se.selector)||de!==se.environment){var ve=J(de,oe);Ee(ve),se=ve}var we=h(!1);if(c(function(){we.current=oe},[oe]),((ne=H.metadata)===null||ne===void 0?void 0:ne.hasClientEdges)===!0||E.CHECK_ALL_FRAGMENTS_FOR_MISSING_CLIENT_EDGES){var Me=d(function(){var ze=q(se),at,mt=[];if(ze!=null&&ze.length){at=[];var Mt=(0,n.default)(ze),bt;try{for(Mt.s();!(bt=Mt.n()).done;){var zt=bt.value,qt=W(de,H,M,zt,X),Xt=qt[0],nn=qt[1];at.push(Xt),nn!=null&&mt.push(nn)}}catch(Fe){Mt.e(Fe)}finally{Mt.f()}}return[at,mt]},[se,de,H,M,X]),Re=Me[0],et=Me[1];if(et.length){var tt=Promise.all(et);throw tt.displayName="RelayClientEdge(".concat(et.map(function(ze){return ze.displayName}).join(","),")"),tt}c(function(){var ze=a(de);if(Re!=null&&Re.length){var at=[],mt=(0,n.default)(Re),Mt;try{for(mt.s();!(Mt=mt.n()).done;){var bt=Mt.value;at.push(ze.retain(bt))}}catch(zt){mt.e(zt)}finally{mt.f()}return function(){for(var zt=0,qt=at;zt<qt.length;zt++){var Xt=qt[zt];Xt.dispose()}}}},[de,Re])}if(U(se)){var Oe=V(se);if(Oe!=null&&Oe.length>0){var $e=Promise.all(Oe.map(function(ze){return de.getStore().getLiveResolverPromise(ze)}));throw $e.displayName="RelayLiveResolver("+H.name+")",$e}if(E.ENABLE_RELAY_OPERATION_TRACKER_SUSPENSE||de!==pe||!we.current||!$(we.current,oe)){oe==null&&o(!1);var Ae=oe.kind==="PluralReaderSelector"?oe.selectors[0].owner:oe.owner,He=A(de,H,Ae);if(He)throw He.promise}}N(de,se,ge);var je=h(null);c(function(){var ze=je.current;if(ze!=null){if(se.environment===ze.environment&&se.selector===ze.selector)return;ze.dispose()}if(se.kind!=="bailout"){var at=se,mt=j(se.environment,se);if(mt!==null){var Mt=mt[0],bt=mt[1];if(Mt){Ee(bt);return}at=bt}var zt=K(se.environment,at,Ee);je.current={dispose:zt,selector:se.selector,environment:se.environment}}},[se]),c(function(){if(je.current==null&&se.kind!=="bailout"){var ze=K(se.environment,se,Ee);je.current={dispose:ze,selector:se.selector,environment:se.environment}}return function(){var at;(at=je.current)===null||at===void 0||at.dispose(),je.current=null}},[]);var Ye;if(ie){var Ne=M==null;Ye=d(function(){return se.kind==="bailout"?Ne?null:[]:(se.kind!=="plural"&&o(!1),se.snapshots.map(function(ze){return ze.data}))},[se,Ne])}else se.kind==="bailout"?Ye=null:(se.kind!=="singular"&&o(!1),Ye=se.snapshot.data);return E.LOG_MISSING_RECORDS_IN_PROD&&M!=null&&(Ye===void 0||Array.isArray(Ye)&&Ye.length>0&&Ye.every(function(ze){return ze===void 0})),Ye}return nT=te,nT}var rT,oV;function tN(){if(oV)return rT;oV=1;var e=$n().default,t=e(t2e()),n=e(n2e()),r=ii();function i(a,u,l,o){return r.RelayFeatureFlags.ENABLE_ACTIVITY_COMPATIBILITY?(0,n.default)(a,u,l,o):(0,t.default)(a,u,l,o)}return rT=i,rT}var iT,sV;function hee(){if(sV)return iT;sV=1;var e=FC(),t=cS(),n=t.getQueryCacheIdentifier,r=t.getQueryResourceForEnvironment,i=fee(),a=tN(),u=co(),l=ir(),o=l.useContext,s=l.useEffect,c=l.useState,d=l.useRef;function h(m){var y=m.query,v=m.componentDisplayName,S=m.fetchObservable,_=m.fetchPolicy,E=m.fetchKey,$=m.renderPolicy,k=u(),A=o(e),w=r(k),R=c(0),F=R[0],I=R[1],U=i(),q=U.startFetch,V=U.completeFetch,N="".concat(F,"-").concat(E??""),j=n(k,y,_,$,N),W=A.wrapPrepareQueryResource(function(){return w.prepareWithIdentifier(j,y,S,_,$,{start:q,complete:V,error:V},A)}),K=d(!1);s(function(){return function(){K.current=!0}},[]),s(function(){if(K.current===!0){K.current=!1,I(function(G){return G+1});return}var M=w.retain(W,A);return function(){M.dispose()}},[k,j]),s(function(){w.releaseTemporaryRetain(W)});var J=W.fragmentNode,te=W.fragmentRef,H=a(J,te,v);return H}return iT=h,iT}var aT,cV;function r2e(){if(cV)return aT;cV=1;var e=ed(),t=ir(),n=t.useState;function r(i){var a=n(i),u=a[0],l=a[1];return e(i,u)?u:(l(i),i)}return aT=r,aT}var uT,fV;function pee(){if(fV)return uT;fV=1;var e=r2e(),t=ir(),n=ii(),r=n.createOperationDescriptor,i=n.getRequest,a=t.useMemo;function u(l,o,s){var c=e(o),d=e(s||{});return a(function(){return r(i(l),c,d)},[l,c,d])}return uT=u,uT}var lT,dV;function mee(){if(dV)return lT;dV=1;var e=hee(),t=pee(),n=co(),r=ii(),i=r.__internal.fetchQuery;function a(u,l,o){var s=n(),c=t(u,l,o&&o.networkCacheConfig?o.networkCacheConfig:{force:!0}),d=e({componentDisplayName:"useLazyLoadQuery()",fetchKey:o==null?void 0:o.fetchKey,fetchObservable:i(s,c),fetchPolicy:o==null?void 0:o.fetchPolicy,query:c,renderPolicy:o==null?void 0:o.UNSTABLE_renderPolicy});return d}return lT=a,lT}var oT,hV;function i2e(){if(hV)return oT;hV=1;var e=$n().default,t=e(Ui()),n=mee();function r(i,a,u){var l=i;return n(l,a,(0,t.default)((0,t.default)({},u),{},{fetchPolicy:"store-only"}))}return oT=r,oT}var sT,pV;function D1(){if(pV)return sT;pV=1;var e=ir(),t=e.useEffect,n=e.useRef;function r(){var i=n(!0);return t(function(){return i.current=!0,function(){i.current=!1}},[]),i}return sT=r,sT}var cT,mV;function a2e(){if(mV)return cT;mV=1;var e=$n().default,t=e(Cu()),n=see(),r=D1(),i=ir(),a=i.useCallback,u=i.useEffect,l=i.useRef,o=i.useState,s={kind:"NullEntryPointReference"};function c(d,h,m){var y,v,S,_,E=(y=m==null||(v=m.TEST_ONLY__initialEntryPointData)===null||v===void 0?void 0:v.entryPointReference)!==null&&y!==void 0?y:s,$=(S=m==null||(_=m.TEST_ONLY__initialEntryPointData)===null||_===void 0?void 0:_.entryPointParams)!==null&&S!==void 0?S:null,k=r(),A=l(new Set([E])),w=o(E),R=w[0],F=w[1],I=o($),U=I[0],q=I[1],V=a(function(){if(k.current){var W={kind:"NullEntryPointReference"};A.current.add(W),F(W)}},[F,k]),N=a(function(W){if(k.current){var K=n(d,h,W);A.current.add(K),F(K),q(W)}},[d,h,F,k]),j=l(!1);return u(function(){return function(){j.current=!0}},[]),u(function(){if(j.current===!0){j.current=!1,R.kind!=="NullEntryPointReference"&&U!=null&&N(U);return}var W=A.current;if(k.current){var K=(0,t.default)(W),J;try{for(K.s();!(J=K.n()).done;){var te=J.value;if(te===R)break;W.delete(te),te.kind!=="NullEntryPointReference"&&te.dispose()}}catch(H){K.e(H)}finally{K.f()}}},[R,U,N,k]),u(function(){return function(){var K=(0,t.default)(A.current),J;try{for(K.s();!(J=K.n()).done;){var te=J.value;te.kind!=="NullEntryPointReference"&&te.dispose()}}catch(H){K.e(H)}finally{K.f()}}},[]),[R.kind==="NullEntryPointReference"?null:R,N,V]}return cT=c,cT}var fT,gV;function IC(){if(gV)return fT;gV=1,cee(),Vi();function e(t,n){}return fT=e,fT}var dT,vV;function gee(){if(vV)return dT;vV=1;var e=tN(),t=IC(),n=ir();n.useDebugValue;var r=ii(),i=r.getFragment;function a(u,l){var o=i(u);t(o,"first argument of useFragment()");var s=e(o,l,"useFragment()");return s}return dT=a,dT}var hT,yV;function u2e(){if(yV)return hT;yV=1;var e=$n().default,t=e(Ui()),n=D1(),r=co(),i=ir(),a=ii(),u=a.commitMutation,l=i.useState,o=i.useEffect,s=i.useRef,c=i.useCallback;function d(h){var m=arguments.length>1&&arguments[1]!==void 0?arguments[1]:u,y=r(),v=n(),S=s(y),_=s(h),E=s(new Set),$=l(!1),k=$[0],A=$[1],w=c(function(F){S.current===y&&_.current===h&&(E.current.delete(F),v.current&&A(E.current.size>0))},[y,v,h]);o(function(){(S.current!==y||_.current!==h)&&(E.current=new Set,v.current&&A(!1),S.current=y,_.current=h)},[y,v,h]);var R=c(function(F){v.current&&A(!0);var I=m(y,(0,t.default)((0,t.default)({},F),{},{mutation:h,onCompleted:function(q,V){var N;w(I),(N=F.onCompleted)===null||N===void 0||N.call(F,q,V)},onError:function(q){var V;w(I),(V=F.onError)===null||V===void 0||V.call(F,q)},onUnsubscribe:function(){var q;w(I),(q=F.onUnsubscribe)===null||q===void 0||q.call(F)},onNext:function(){var q;(q=F.onNext)===null||q===void 0||q.call(F)}}));return E.current.add(I),I},[w,m,y,v,h]);return[R,k]}return hT=d,hT}var pT,bV;function vee(){if(bV)return pT;bV=1;var e=un(),t=ii(),n=t.ConnectionInterface,r=t.getValueAtPath;function i(a,u,l,o){var s,c,d=n.get(),h=d.EDGES,m=d.PAGE_INFO,y=d.HAS_NEXT_PAGE,v=d.HAS_PREV_PAGE,S=d.END_CURSOR,_=d.START_CURSOR,E=r(l,o);if(E==null)return{cursor:null,hasMore:!1};typeof E!="object"&&e(!1);var $=E[h],k=E[m];if($==null||k==null)return{cursor:null,hasMore:!1};Array.isArray($)||e(!1),typeof k!="object"&&e(!1);var A=a==="forward"?(s=k[S])!==null&&s!==void 0?s:null:(c=k[_])!==null&&c!==void 0?c:null;A===null||typeof A=="string"||e(!1);var w;return a==="forward"?w=A!=null&&k[y]===!0:w=A!=null&&k[v]===!0,{cursor:A,hasMore:w}}return pT=i,pT}var mT,OV;function yee(){if(OV)return mT;OV=1;var e=co(),t=un(),n=ir(),r=ii(),i=r.__internal.getObservableForActiveRequest,a=r.getSelector,u=n.useEffect,l=n.useState,o=n.useMemo;function s(c,d){var h=e(),m=o(function(){var _=a(c,d);return _==null?null:(_.kind!=="SingularReaderSelector"&&t(!1),i(h,_.owner))},[h,c,d]),y=l(m!=null),v=y[0],S=y[1];return u(function(){var _;if(S(m!=null),m!=null){var E=function(){S(!1)};_=m.subscribe({complete:E,error:E})}return function(){_&&_.unsubscribe()}},[m]),v}return mT=s,mT}var gT,xV;function l2e(){if(xV)return gT;xV=1;var e=$n().default,t=e(Ui()),n=vee(),r=D1(),i=yee(),a=co(),u=un(),l=ir(),o=l.useCallback,s=l.useRef,c=l.useState,d=ii(),h=d.__internal.fetchQuery,m=d.createOperationDescriptor,y=d.getPaginationVariables,v=d.getRefetchMetadata,S=d.getSelector;Vi();function _(E){var $=E.direction,k=E.fragmentNode,A=E.fragmentRef,w=E.fragmentIdentifier,R=E.fragmentData,F=E.connectionPathInFragmentData,I=E.paginationRequest,U=E.paginationMetadata,q=E.componentDisplayName,V=E.observer,N=E.onReset,j=a(),W=v(k,q),K=W.identifierInfo,J=(K==null?void 0:K.identifierField)!=null&&R!=null&&typeof R=="object"?R[K.identifierField]:null,te=s({kind:"none"}),H=c(j),M=H[0],G=H[1],X=c(w),B=X[0],ne=X[1],oe=i(k,A),ie=o(function(){te.current.kind==="fetching"&&te.current.subscription.unsubscribe(),te.current={kind:"none"}},[]),de=j!==M||w!==B;de&&(ie(),N(),G(j),ne(w));var ge=n($,k,R,F),ye=ge.cursor,me=ge.hasMore,Ee=R==null||oe,se=r(),pe=o(function(ve,we){var Me=we==null?void 0:we.onComplete;if(se.current!==!0)return{dispose:function(){}};var Re=S(k,A);if(te.current.kind==="fetching"||Ee)return Me&&Me(null),{dispose:function(){}};Re!=null&&Re.kind!=="PluralReaderSelector"||u(!1);var et=Re.owner.variables,tt=Re.variables,Oe=we==null?void 0:we.UNSTABLE_extraVariables,$e=(0,t.default)((0,t.default)({},et),tt),Ae=y($,ve,ye,$e,(0,t.default)({},Oe),U);K!=null&&(Ae[K.identifierQueryVariableName]=J);var He=m(I,Ae,{force:!0});return h(j,He).subscribe((0,t.default)((0,t.default)({},V),{},{start:function(Ye){te.current={kind:"fetching",subscription:Ye},V.start&&V.start(Ye)},complete:function(){te.current={kind:"none"},V.complete&&V.complete(),Me&&Me(null)},error:function(Ye){te.current={kind:"none"},V.complete&&V.complete(),Me&&Me(Ye)}})),{dispose:function(){}}},[j,J,$,ye,Ee,k.name,A,q]);return[pe,me,ie]}return gT=_,gT}var vT,SV;function bee(){if(SV)return vT;SV=1;var e=$n().default,t=e(Ui()),n=vee(),r=fee(),i=D1(),a=yee(),u=l2e(),l=co(),o=un(),s=ir(),c=s.useCallback,d=s.useEffect,h=s.useState,m=ii(),y=m.__internal.fetchQuery,v=m.RelayFeatureFlags,S=m.createOperationDescriptor,_=m.getPaginationVariables,E=m.getRefetchMetadata,$=m.getSelector;Vi();function k(w){return v.ENABLE_ACTIVITY_COMPATIBILITY?u(w):A(w)}function A(w){var R=w.direction,F=w.fragmentNode,I=w.fragmentRef,U=w.fragmentIdentifier,q=w.fragmentData,V=w.connectionPathInFragmentData,N=w.paginationRequest,j=w.paginationMetadata,W=w.componentDisplayName,K=w.observer,J=w.onReset,te=l(),H=r(),M=H.isFetchingRef,G=H.startFetch,X=H.disposeFetch,B=H.completeFetch,ne=E(F,W),oe=ne.identifierInfo,ie=(oe==null?void 0:oe.identifierField)!=null&&q!=null&&typeof q=="object"?q[oe.identifierField]:null,de=i(),ge=h(te),ye=ge[0],me=ge[1],Ee=h(U),se=Ee[0],pe=Ee[1],ve=a(F,I),we=te!==ye||U!==se;we&&(X(),J(),me(te),pe(U));var Me=n(R,F,q,V),Re=Me.cursor,et=Me.hasMore;d(function(){return function(){X()}},[X]);var tt=q==null||ve,Oe=c(function($e,Ae){var He=Ae==null?void 0:Ae.onComplete;if(de.current!==!0)return{dispose:function(){}};var je=$(F,I);if(M.current===!0||tt)return He&&He(null),{dispose:function(){}};je!=null&&je.kind!=="PluralReaderSelector"||o(!1);var Ye=je.owner.variables,Ne=je.variables,ze=Ae==null?void 0:Ae.UNSTABLE_extraVariables,at=(0,t.default)((0,t.default)({},Ye),Ne),mt=_(R,$e,Re,at,(0,t.default)({},ze),j);oe!=null&&(mt[oe.identifierQueryVariableName]=ie);var Mt=S(N,mt,{force:!0});return y(te,Mt).subscribe((0,t.default)((0,t.default)({},K),{},{start:function(zt){G(zt),K.start&&K.start(zt)},complete:function(){B(),K.complete&&K.complete(),He&&He(null)},error:function(zt){B(),K.error&&K.error(zt),He&&He(zt)}})),{dispose:X}},[te,ie,R,Re,G,X,B,M,tt,F.name,I,W]);return[Oe,et,X]}return vT=k,vT}var yT,_V;function o2e(){if(_V)return yT;_V=1;var e=$n().default,t=e(Ui()),n=e(Cu()),r=cS(),i=r.getQueryResourceForEnvironment,a=un(),u=ii(),l=u.__internal.fetchQuery,o=u.RelayFeatureFlags,s=u.createOperationDescriptor,c=u.getPendingOperationsForFragment,d=u.getSelector,h=u.getVariablesFromFragment,m=u.handlePotentialSnapshotErrors;Vi();function y(k){return k.kind==="bailout"?!1:k.kind==="singular"?k.snapshot.isMissingData:k.snapshots.some(function(A){return A.isMissingData})}function v(k){if(k.kind==="bailout")return null;if(k.kind==="singular"){var A;return(A=k.snapshot.missingClientEdges)!==null&&A!==void 0?A:null}else{var w=null,R=(0,n.default)(k.snapshots),F;try{for(R.s();!(F=R.n()).done;){var I=F.value;if(I.missingClientEdges){var U;w=(U=w)!==null&&U!==void 0?U:[];var q=(0,n.default)(I.missingClientEdges),V;try{for(q.s();!(V=q.n()).done;){var N=V.value;w.push(N)}}catch(j){q.e(j)}finally{q.f()}}}}catch(j){R.e(j)}finally{R.f()}return w}}function S(k,A){if(A.kind==="singular")m(k,A.snapshot.fieldErrors);else if(A.kind==="plural"){var w=(0,n.default)(A.snapshots),R;try{for(w.s();!(R=w.n()).done;){var F=R.value;m(k,F.fieldErrors)}}catch(I){w.e(I)}finally{w.f()}}}function _(k,A,w,R,F){var I=h(A,w),U=(0,t.default)((0,t.default)({},I),{},{id:R.clientEdgeDestinationID}),q=s(R.request,U,F==null?void 0:F.networkCacheConfig),V=i(k);return V.prepare(q,l(k,q),F==null?void 0:F.fetchPolicy)}function E(k,A){return A==null?{kind:"bailout"}:A.kind==="PluralReaderSelector"?A.selectors.length===0?{kind:"bailout"}:{kind:"plural",snapshots:A.selectors.map(function(w){return k.lookup(w)}),epoch:k.getStore().getEpoch()}:{kind:"singular",snapshot:k.lookup(A),epoch:k.getStore().getEpoch()}}function $(k,A,w,R,F,I){var U,q,V=d(A,w),N=(A==null||(U=A.metadata)===null||U===void 0?void 0:U.plural)===!0;N?w==null||Array.isArray(w)||a(!1):Array.isArray(w)&&a(!1),w==null||N&&Array.isArray(w)&&w.length===0||V!=null||a(!1);var j=E(k,V),W=null;if(((q=A.metadata)===null||q===void 0?void 0:q.hasClientEdges)===!0||o.CHECK_ALL_FRAGMENTS_FOR_MISSING_CLIENT_EDGES){var K=v(j);if(K!=null&&K.length){W=[];var J=(0,n.default)(K),te;try{for(J.s();!(te=J.n()).done;){var H=te.value;W.push(_(k,A,w,H,F))}}catch(B){J.e(B)}finally{J.f()}}}if(y(j)){V==null&&a(!1);var M=V.kind==="PluralReaderSelector"?V.selectors[0].owner:V.owner,G=c(k,A,M);if(G)throw G.promise;S(k,j)}var X;return j.kind==="bailout"?X=N?[]:null:j.kind==="singular"?X=j.snapshot.data:X=j.snapshots.map(function(B){return B.data}),o.LOG_MISSING_RECORDS_IN_PROD&&w!=null&&(X===void 0||Array.isArray(X)&&X.length>0&&X.every(function(B){return B===void 0})),{data:X,clientEdgeQueries:W}}return yT=$,yT}var bT,EV;function s2e(){if(EV)return bT;EV=1;var e=$n().default,t=e(Cu()),n=NC(),r=n.loadQuery,i=D1(),a=co(),u=ir(),l=u.useCallback,o=u.useEffect,s=u.useInsertionEffect,c=u.useRef,d=u.useState,h=ii(),m=h.getRequest,y={kind:"NullQueryReference"};function v(E){if(E.kind==="PreloadableConcreteRequest")return E.params.metadata.live!==void 0;var $=m(E);return $.params.metadata.live!==void 0}var S=1e3*60*5;function _(E,$){var k=$??y,A=a(),w=i(),R=c(null);R.current==null&&(R.current=new Set([k]));var F=d(function(){return k}),I=F[0],U=F[1],q=d(function(){return k}),V=q[0],N=q[1];if(k!==V){var j;(j=R.current)===null||j===void 0||j.add(k),N(k),U(k)}var W=l(function(){if(w.current){var H;(H=R.current)===null||H===void 0||H.add(y),U(y)}},[w]),K=l(function(H,M){var G,X;if(w.current){var B=M!=null&&M.hasOwnProperty("__environment")?{fetchPolicy:M.fetchPolicy,networkCacheConfig:M.networkCacheConfig,__nameForWarning:M.__nameForWarning}:M,ne=r((G=M==null?void 0:M.__environment)!==null&&G!==void 0?G:A,E,H,B);(X=R.current)===null||X===void 0||X.add(ne),U(ne)}},[A,E,U,w]),J=l(function(M,G){var X,B=(X=R.current)!==null&&X!==void 0?X:new Set,ne=(0,t.default)(B),oe;try{for(ne.s();!(oe=ne.n()).done;){var ie=oe.value;ie!==G&&ie.kind!=="NullQueryReference"&&(v(M)?ie.dispose&&ie.dispose():ie.releaseQuery&&ie.releaseQuery())}}catch(de){ne.e(de)}finally{ne.f()}},[]),te=c(null);return o(function(){return J(E,I),te.current!=null&&(clearTimeout(te.current),te.current=null),function(){te.current=setTimeout(function(){J(E,null)},S)}},[E,I]),s(function(){return function(){te.current&&clearTimeout(te.current),te.current=null,J(E,null)}},[E]),[I.kind==="NullQueryReference"?null:I,K,W]}return bT=_,bT}var OT,DV;function Oee(){if(DV)return OT;DV=1;var e=$n().default,t=e(Cu()),n=NC(),r=n.loadQuery,i=D1(),a=s2e(),u=co(),l=ir(),o=l.useCallback,s=l.useEffect,c=l.useRef,d=l.useState,h=ii(),m=h.RelayFeatureFlags,y=h.getRequest,v={kind:"NullQueryReference"};function S($){if($.kind==="PreloadableConcreteRequest")return $.params.metadata.live!==void 0;var k=y($);return k.params.metadata.live!==void 0}function _($,k){return m.ENABLE_ACTIVITY_COMPATIBILITY?a($,k):E($,k)}function E($,k){var A=k??v,w=u(),R=i(),F=c(new Set([A])),I=d(function(){return A}),U=I[0],q=I[1],V=d(function(){return A}),N=V[0],j=V[1];A!==N&&(F.current.add(A),j(A),q(A));var W=o(function(){R.current&&(F.current.add(v),q(v))},[R]),K=o(function(te,H){var M=H!=null&&H.hasOwnProperty("__environment")?{fetchPolicy:H.fetchPolicy,networkCacheConfig:H.networkCacheConfig,__nameForWarning:H.__nameForWarning}:H;if(R.current){var G,X=r((G=H==null?void 0:H.__environment)!==null&&G!==void 0?G:w,$,te,M);F.current.add(X),q(X)}},[w,$,q,R]),J=c(!1);return s(function(){return function(){J.current=!0}},[]),s(function(){if(J.current===!0){J.current=!1,U.kind!=="NullQueryReference"&&K(U.variables,{fetchPolicy:U.fetchPolicy,networkCacheConfig:U.networkCacheConfig});return}var te=F.current;if(R.current){var H=(0,t.default)(te),M;try{for(H.s();!(M=H.n()).done;){var G=M.value;if(G===U)break;te.delete(G),G.kind!=="NullQueryReference"&&(S($)?G.dispose&&G.dispose():G.releaseQuery&&G.releaseQuery())}}catch(X){H.e(X)}finally{H.f()}}},[U,R,K,$]),s(function(){return function(){var H=(0,t.default)(F.current),M;try{for(H.s();!(M=H.n()).done;){var G=M.value;G.kind!=="NullQueryReference"&&(S($)?G.dispose&&G.dispose():G.releaseQuery&&G.releaseQuery())}}catch(X){H.e(X)}finally{H.f()}}},[$]),[U.kind==="NullQueryReference"?null:U,K,W]}return OT=_,OT}var xT,$V;function nN(){if($V)return xT;$V=1;var e=$n().default,t=e(Ui()),n=FC(),r=cS(),i=r.getQueryResourceForEnvironment,a=o2e(),u=tN(),l=D1(),o=Oee(),s=co(),c=un(),d=ir(),h=d.useCallback,m=d.useContext,y=d.useReducer,v=ii(),S=v.__internal.fetchQuery,_=v.createOperationDescriptor,E=v.getFragmentIdentifier,$=v.getRefetchMetadata,k=v.getSelector,A=v.getValueAtPath;Vi();function w(I,U){switch(U.type){case"refetch":{var q;return(0,t.default)((0,t.default)({},I),{},{fetchPolicy:U.fetchPolicy,mirroredEnvironment:(q=U.refetchEnvironment)!==null&&q!==void 0?q:I.mirroredEnvironment,onComplete:U.onComplete,refetchEnvironment:U.refetchEnvironment,refetchQuery:U.refetchQuery,renderPolicy:U.renderPolicy})}case"reset":return{fetchPolicy:void 0,mirroredEnvironment:U.environment,mirroredFragmentIdentifier:U.fragmentIdentifier,onComplete:void 0,refetchQuery:null,renderPolicy:void 0};default:throw U.type,new Error("useRefetchableFragmentNode: Unexpected action type")}}function R(I,U,q){var V=s(),N=$(I,q),j=N.refetchableRequest,W=N.fragmentRefPathInResponse,K=N.identifierInfo,J=E(I,U),te=y(w,{fetchPolicy:void 0,mirroredEnvironment:V,mirroredFragmentIdentifier:J,onComplete:void 0,refetchEnvironment:null,refetchQuery:null,renderPolicy:void 0}),H=te[0],M=te[1],G=H.fetchPolicy,X=H.mirroredEnvironment,B=H.mirroredFragmentIdentifier,ne=H.onComplete,oe=H.refetchEnvironment,ie=H.refetchQuery,de=H.renderPolicy,ge=oe??V,ye=i(ge),me=m(n),Ee=ge!==X||J!==B,se=o(j),pe=se[0],ve=se[1],we=se[2],Me=U;if(Ee)M({type:"reset",environment:ge,fragmentIdentifier:J}),we();else if(ie!=null&&pe!=null){var Re=function(Ye){ne&&ne(Ye??null)},et=pe.source!=null?pe.source:S(ge,ie),tt=me.wrapPrepareQueryResource(function(){return ye.prepare(ie,et,G,de,{error:Re,complete:function(){Re()}},pe.fetchKey,me)}),Oe=a(ge,tt.fragmentNode,tt.fragmentRef,q).data;Oe==null&&c(!1);var $e=A(Oe,W);Me=$e}var Ae=u(I,Me,q),He=F(q,M,we,Ae,J,I,W,K,ve,U,j);return{fragmentData:Ae,fragmentRef:Me,refetch:He}}function F(I,U,q,V,N,j,W,K,J,te,H){var M=l(),G=(K==null?void 0:K.identifierField)!=null&&V!=null&&typeof V=="object"?V[K.identifierField]:null;return h(function(X,B){if(M.current!==!0)return{dispose:function(){}};var ne=B==null?void 0:B.__environment,oe=B==null?void 0:B.fetchPolicy,ie=B==null?void 0:B.UNSTABLE_renderPolicy,de=B==null?void 0:B.onComplete,ge=k(j,te),ye,me;if(ge==null)ye={},me={};else if(ge.kind==="PluralReaderSelector"){var Ee,se,pe,ve;ye=(Ee=(se=ge.selectors[0])===null||se===void 0?void 0:se.owner.variables)!==null&&Ee!==void 0?Ee:{},me=(pe=(ve=ge.selectors[0])===null||ve===void 0?void 0:ve.variables)!==null&&pe!==void 0?pe:{}}else ye=ge.owner.variables,me=ge.variables;var we=(0,t.default)((0,t.default)((0,t.default)({},ye),me),X);K!=null&&!X.hasOwnProperty(K.identifierQueryVariableName)&&(we[K.identifierQueryVariableName]=G);var Me=_(H,we,{force:!0});return J(Me.request.variables,{fetchPolicy:oe,__environment:ne,__nameForWarning:"refetch"}),U({type:"refetch",fetchPolicy:oe,onComplete:de,refetchEnvironment:ne,refetchQuery:Me,renderPolicy:ie}),{dispose:q}},[N,U,q,G,J])}return xT=R,xT}var ST,CV;function c2e(){if(CV)return ST;CV=1;var e=$n().default,t=e(Ui()),n=bee(),r=nN(),i=co(),a=IC(),u=ir(),l=u.useCallback;u.useDebugValue;var o=u.useState,s=ii(),c=s.RelayFeatureFlags,d=s.getFragment,h=s.getFragmentIdentifier,m=s.getPaginationMetadata;function y(S,_){var E=d(S);a(E,"first argument of usePaginationFragment()");var $="usePaginationFragment()",k=m(E,$),A=k.connectionPathInFragmentData,w=k.paginationRequest,R=k.paginationMetadata,F=r(E,_,$),I=F.fragmentData,U=F.fragmentRef,q=F.refetch,V=h(E,U),N=v({componentDisplayName:$,connectionPathInFragmentData:A,direction:"backward",fragmentData:I,fragmentIdentifier:V,fragmentNode:E,fragmentRef:U,paginationMetadata:R,paginationRequest:w}),j=N[0],W=N[1],K=N[2],J=N[3],te=v({componentDisplayName:$,connectionPathInFragmentData:A,direction:"forward",fragmentData:I,fragmentIdentifier:V,fragmentNode:E,fragmentRef:U,paginationMetadata:R,paginationRequest:w}),H=te[0],M=te[1],G=te[2],X=te[3],B=l(function(ne,oe){return X(),J(),q(ne,(0,t.default)((0,t.default)({},oe),{},{__environment:void 0}))},[X,J,q]);return{data:I,loadNext:H,loadPrevious:j,hasNext:M,hasPrevious:W,isLoadingNext:G,isLoadingPrevious:K,refetch:B}}function v(S){var _=i(),E=o(!1),$=E[0],k=E[1],A=function(N){var j,W=(j=_.getScheduler())===null||j===void 0?void 0:j.schedule;W?W(function(){k(N)}):k(N)},w={start:function(){return A(!0)},complete:function(){return A(!1)},error:function(){return A(!1)},unsubscribe:c.ENABLE_USE_PAGINATION_IS_LOADING_FIX?function(){return A(!1)}:void 0},R=function(){return A(!1)},F=n((0,t.default)((0,t.default)({},S),{},{observer:w,onReset:R})),I=F[0],U=F[1],q=F[2];return[I,U,$,q]}return ST=y,ST}var _T,wV;function f2e(){if(wV)return _T;wV=1;var e=$n().default,t=e(Ui()),n=gee(),r=bee(),i=nN(),a=co(),u=IC(),l=un(),o=ir(),s=o.useCallback;o.useDebugValue;var c=o.useEffect,d=o.useLayoutEffect,h=o.useMemo,m=o.useRef,y=o.useState,v=ii(),S=v.getFragment,_=v.getFragmentIdentifier,E=v.getPaginationMetadata,$=ii(),k=$.ConnectionInterface,A=$.RelayFeatureFlags,w=$.getSelector,R=$.getValueAtPath;function F(I,U,q,V,N){var j,W,K=arguments.length>5&&arguments[5]!==void 0?arguments[5]:1,J=arguments.length>6&&arguments[6]!==void 0?arguments[6]:!1,te=S(I);u(te,"first argument of usePrefetchableForwardPaginationFragment()");var H="usePrefetchableForwardPaginationFragment()",M=E(te,H),G=M.connectionPathInFragmentData,X=M.paginationRequest,B=M.paginationMetadata,ne=i(te,U,H),oe=ne.fragmentData,ie=ne.fragmentRef,de=ne.refetch,ge=_(te,ie),ye=h(function(){var Qt=R(oe,G);if(Qt==null)return null;var en=k.get(),In=en.EDGES;return Qt[In]},[G,oe]),me=ye==null?-1:ye.length,Ee=y(V??me),se=Ee[0],pe=Ee[1],ve=se;se===-1&&me!==-1&&(ve=V??me,pe(ve));var we=a(),Me=y(!1),Re=Me[0],et=Me[1],tt=y(!1),Oe=tt[0],$e=tt[1],Ae=m(0),He=s(function(Qt){var en,In=(en=we.getScheduler())===null||en===void 0?void 0:en.schedule;In?In(function(){et(Qt)}):et(Qt)},[we]),je=m(!1),Ye=h(function(){function Qt(){je.current=!1,He(!1)}return{start:function(){je.current=!0,et(!0)},complete:Qt,error:Qt,unsubscribe:A.ENABLE_USE_PAGINATION_IS_LOADING_FIX?Qt:void 0}},[He]),Ne=s(function(){if(!Oe){var Qt,en=(Qt=we.getScheduler())===null||Qt===void 0?void 0:Qt.schedule;en?en(function(){pe(-1)}):pe(-1)}je.current=!1,He(!1)},[we,Oe,He]),ze=r({componentDisplayName:H,connectionPathInFragmentData:G,direction:"forward",fragmentData:oe,fragmentIdentifier:ge,fragmentNode:te,fragmentRef:ie,paginationMetadata:B,paginationRequest:X,observer:Ye,onReset:Ne}),at=ze[0],mt=ze[1],Mt=ze[2];d(function(){Ae.current=me-ve},[ve,me]);var bt=N==null?void 0:N.UNSTABLE_extraVariables,zt=N==null?void 0:N.onComplete,qt=s(function(Qt,en){(!je.current||Ae.current>=0)&&(Ae.current-=Qt,pe(function(In){return In+Qt}),!je.current&&Ae.current<0&&at(Math.max(K,Math.min(Qt,q-Ae.current)),en??{onComplete:zt,UNSTABLE_extraVariables:typeof bt=="function"?bt({hasNext:mt,data:oe,getServerEdges:function(){var rn=w(Xt,ye);return rn==null?[]:(rn.kind!=="PluralReaderSelector"&&l(!1),rn.selectors.map(function(Xn){return we.lookup(Xn).data}))}}):bt}))},[q,at,K,ye,oe,bt,zt]),Xt=(j=I.metadata)===null||j===void 0||(W=j.refetch)===null||W===void 0?void 0:W.edgesFragment;Xt==null&&l(!1),c(function(){if(!je.current&&!Re&&!Oe&&!J&&mt&&(me-ve<q||ve>me)){var Qt=zt;at(Math.max(q-Math.max(me-ve,0),ve-me,K),{onComplete:Qt,UNSTABLE_extraVariables:typeof bt=="function"?bt({hasNext:mt,data:oe,getServerEdges:function(){var In=w(Xt,ye);return In==null?[]:(In.kind!=="PluralReaderSelector"&&l(!1),In.selectors.map(function(rn){return we.lookup(rn).data}))}}):bt})}},[mt,q,Oe,at,ve,bt,zt,me,ye,Re,K,we,Xt]);var nn=Math.min(ve,me),Fe=h(function(){var Qt;return(Qt=ye==null?void 0:ye.slice(0,nn))!==null&&Qt!==void 0?Qt:[]},[ye,nn]),ht=n(Xt,Fe),kt=s(function(Qt,en){return Mt(),$e(!0),de(Qt,(0,t.default)((0,t.default)({},en),{},{onComplete:function(rn){var Xn,br,Or=(Xn=we.getScheduler())===null||Xn===void 0?void 0:Xn.schedule;Or?Or(function(){$e(!1),pe(-1)}):($e(!1),pe(-1)),en==null||(br=en.onComplete)===null||br===void 0||br.call(en,rn)},__environment:void 0}))},[Mt,we,de]);return{edges:ht,data:oe,loadNext:qt,hasNext:mt||me>ve,isLoadingNext:Re&&ve>me,refetch:kt}}return _T=F,_T}var ET,RV;function d2e(){if(RV)return ET;RV=1;var e=hee(),t=pee(),n=co(),r=un(),i=ir();i.useDebugValue;var a=ii(),u=a.__internal,l=u.fetchQueryDeduped,o=u.fetchQuery;Vi();function s(c,d,h){var m=n(),y=d.fetchKey,v=d.fetchPolicy,S=d.source,_=d.variables,E=d.networkCacheConfig,$=t(c,_,E),k;if(d.kind==="PreloadedQuery_DEPRECATED")$.request.node.params.name!==d.name&&r(!1),k={componentDisplayName:"usePreloadedQuery()",fetchKey:y,fetchObservable:l(m,$.request.identifier,function(){return m===d.environment&&S!=null?m.executeWithSource({operation:$,source:S}):m.execute({operation:$})}),fetchPolicy:v,query:$,renderPolicy:h==null?void 0:h.UNSTABLE_renderPolicy};else{var A=o(m,$),w;S!=null&&m===d.environment?w=S.ifEmpty(A):(d.environment,w=A),k={componentDisplayName:"usePreloadedQuery()",fetchObservable:w,fetchKey:y,fetchPolicy:v,query:$,renderPolicy:h==null?void 0:h.UNSTABLE_renderPolicy}}var R=e(k);return R}return ET=s,ET}var DT,AV;function h2e(){if(AV)return DT;AV=1;var e=nN(),t=IC(),n=ir();n.useDebugValue;var r=ii(),i=r.getFragment;function a(u,l){var o=i(u);t(o,"first argument of useRefetchableFragment()");var s=e(o,l,"useRefetchableFragment()"),c=s.fragmentData,d=s.refetch;return[c,d]}return DT=a,DT}var $T,TV;function p2e(){if(TV)return $T;TV=1;var e=co(),t=ir(),n=t.useEffect,r=t.useRef;function i(a,u){var l=e(),o=r(null),s=Array.from(a).sort().join("");return n(function(){var c=l.getStore(),d=c.lookupInvalidationState(a),h=c.subscribeToInvalidationState(d,u);return o.current=h,function(){return h.dispose()}},[s,u,l]),{dispose:function(){o.current!=null&&o.current.dispose()}}}return $T=i,$T}var CT,kV;function m2e(){if(kV)return CT;kV=1;var e=co(),t=ir(),n=t.useEffect,r=ii(),i=r.requestSubscription;function a(u,l){var o=l??i,s=e();n(function(){var c=o(s,u),d=c.dispose;return d},[s,u,o])}return CT=a,CT}var wT,PV;function g2e(){if(PV)return wT;PV=1;var e=hy(),t=j4e(),n=Z4e(),r=H4e(),i=X4e(),a=Y4e(),u=G4e(),l=see(),o=NC(),s=o.loadQuery,c=FC(),d=K4e(),h=i2e(),m=a2e(),y=gee(),v=mee(),S=u2e(),_=c2e(),E=f2e(),$=d2e(),k=Oee(),A=h2e(),w=co(),R=p2e(),F=m2e(),I=ii();return wT={ConnectionHandler:I.ConnectionHandler,QueryRenderer:i,LocalQueryRenderer:n,MutationTypes:I.MutationTypes,RangeOperations:I.RangeOperations,ReactRelayContext:e,applyOptimisticMutation:I.applyOptimisticMutation,commitLocalUpdate:I.commitLocalUpdate,commitMutation:I.commitMutation,createFragmentContainer:t.createContainer,createPaginationContainer:r.createContainer,createRefetchContainer:a.createContainer,fetchQuery_DEPRECATED:I.fetchQuery_DEPRECATED,graphql:I.graphql,readInlineData:I.readInlineData,requestSubscription:I.requestSubscription,EntryPointContainer:u,RelayEnvironmentProvider:d,ProfilerContext:c,fetchQuery:I.fetchQuery,loadQuery:s,loadEntryPoint:l,useClientQuery:h,useFragment:y,useLazyLoadQuery:v,useEntryPointLoader:m,useQueryLoader:k,useMutation:S,usePaginationFragment:_,usePreloadedQuery:$,useRefetchableFragment:A,usePrefetchableForwardPaginationFragment:E,useRelayEnvironment:w,useSubscribeToInvalidationState:R,useSubscription:F},wT}var RT,MV;function v2e(){return MV||(MV=1,RT=g2e()),RT}var art=v2e();function urt(e,t,n=Object.is){const r=e.getState(),[[i,a],u]=P.useReducer((l,o)=>{if(o)return o;const s=e.getState();if(Object.is(l[2],s)&&l[1]===e)return l;const c=t(s);return n(l[0],c)&&l[1]===e?l:[c,e,s]},void 0,()=>[t(r),e,r]);if(P.useEffect(()=>{const l=e.subscribe(()=>u());return u(),l},[e]),a!==e){const l=t(r);return u([l,e,r]),l}return i}const y2e={},FV=e=>{let t;const n=new Set,r=(c,d)=>{const h=typeof c=="function"?c(t):c;if(!Object.is(h,t)){const m=t;t=d??(typeof h!="object"||h===null)?h:Object.assign({},t,h),n.forEach(y=>y(t,m))}},i=()=>t,o={setState:r,getState:i,getInitialState:()=>s,subscribe:c=>(n.add(c),()=>n.delete(c)),destroy:()=>{(y2e?"production":void 0)!=="production"&&console.warn("[DEPRECATED] The `destroy` method will be unsupported in a future version. Instead use unsubscribe function returned by subscribe. Everything will be garbage-collected if store is garbage-collected."),n.clear()}},s=t=e(r,i,o);return o},b2e=e=>e?FV(e):FV;var AT={exports:{}},TT={},kT={exports:{}},PT={};/**
 * @license React
 * use-sync-external-store-shim.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var NV;function O2e(){if(NV)return PT;NV=1;var e=ir();function t(d,h){return d===h&&(d!==0||1/d===1/h)||d!==d&&h!==h}var n=typeof Object.is=="function"?Object.is:t,r=e.useState,i=e.useEffect,a=e.useLayoutEffect,u=e.useDebugValue;function l(d,h){var m=h(),y=r({inst:{value:m,getSnapshot:h}}),v=y[0].inst,S=y[1];return a(function(){v.value=m,v.getSnapshot=h,o(v)&&S({inst:v})},[d,m,h]),i(function(){return o(v)&&S({inst:v}),d(function(){o(v)&&S({inst:v})})},[d]),u(m),m}function o(d){var h=d.getSnapshot;d=d.value;try{var m=h();return!n(d,m)}catch{return!0}}function s(d,h){return h()}var c=typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"?s:l;return PT.useSyncExternalStore=e.useSyncExternalStore!==void 0?e.useSyncExternalStore:c,PT}var IV;function xee(){return IV||(IV=1,kT.exports=O2e()),kT.exports}/**
 * @license React
 * use-sync-external-store-shim/with-selector.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var LV;function x2e(){if(LV)return TT;LV=1;var e=ir(),t=xee();function n(s,c){return s===c&&(s!==0||1/s===1/c)||s!==s&&c!==c}var r=typeof Object.is=="function"?Object.is:n,i=t.useSyncExternalStore,a=e.useRef,u=e.useEffect,l=e.useMemo,o=e.useDebugValue;return TT.useSyncExternalStoreWithSelector=function(s,c,d,h,m){var y=a(null);if(y.current===null){var v={hasValue:!1,value:null};y.current=v}else v=y.current;y=l(function(){function _(w){if(!E){if(E=!0,$=w,w=h(w),m!==void 0&&v.hasValue){var R=v.value;if(m(R,w))return k=R}return k=w}if(R=k,r($,w))return R;var F=h(w);return m!==void 0&&m(R,F)?($=w,R):($=w,k=F)}var E=!1,$,k,A=d===void 0?null:d;return[function(){return _(c())},A===null?void 0:function(){return _(A())}]},[c,d,h,m]);var S=i(s,y[0],y[1]);return u(function(){v.hasValue=!0,v.value=S},[S]),o(S),S},TT}var BV;function S2e(){return BV||(BV=1,AT.exports=x2e()),AT.exports}var _2e=S2e();const See=da(_2e),_ee={},{useDebugValue:E2e}=be,{useSyncExternalStoreWithSelector:D2e}=See;let QV=!1;const $2e=e=>e;function C2e(e,t=$2e,n){(_ee?"production":void 0)!=="production"&&n&&!QV&&(console.warn("[DEPRECATED] Use `createWithEqualityFn` instead of `create` or use `useStoreWithEqualityFn` instead of `useStore`. They can be imported from 'zustand/traditional'. https://github.com/pmndrs/zustand/discussions/1937"),QV=!0);const r=D2e(e.subscribe,e.getState,e.getServerState||e.getInitialState,t,n);return E2e(r),r}const zV=e=>{(_ee?"production":void 0)!=="production"&&typeof e!="function"&&console.warn("[DEPRECATED] Passing a vanilla store will be unsupported in a future version. Instead use `import { useStore } from 'zustand'`.");const t=typeof e=="function"?b2e(e):e,n=(r,i)=>C2e(t,r,i);return Object.assign(n,t),n},lrt=e=>e?zV(e):zV,YE={BASE_URL:"/",DEV:!1,MODE:"production",PROD:!0,SSR:!1},a9=new Map,F_=e=>{const t=a9.get(e);return t?Object.fromEntries(Object.entries(t.stores).map(([n,r])=>[n,r.getState()])):{}},w2e=(e,t,n)=>{if(e===void 0)return{type:"untracked",connection:t.connect(n)};const r=a9.get(n.name);if(r)return{type:"tracked",store:e,...r};const i={connection:t.connect(n),stores:{}};return a9.set(n.name,i),{type:"tracked",store:e,...i}},R2e=(e,t={})=>(n,r,i)=>{const{enabled:a,anonymousActionType:u,store:l,...o}=t;let s;try{s=(a??(YE?"production":void 0)!=="production")&&window.__REDUX_DEVTOOLS_EXTENSION__}catch{}if(!s)return(YE?"production":void 0)!=="production"&&a&&console.warn("[zustand devtools middleware] Please install/enable Redux devtools extension"),e(n,r,i);const{connection:c,...d}=w2e(l,s,o);let h=!0;i.setState=(v,S,_)=>{const E=n(v,S);if(!h)return E;const $=_===void 0?{type:u||"anonymous"}:typeof _=="string"?{type:_}:_;return l===void 0?(c==null||c.send($,r()),E):(c==null||c.send({...$,type:`${l}/${$.type}`},{...F_(o.name),[l]:i.getState()}),E)};const m=(...v)=>{const S=h;h=!1,n(...v),h=S},y=e(i.setState,r,i);if(d.type==="untracked"?c==null||c.init(y):(d.stores[d.store]=i,c==null||c.init(Object.fromEntries(Object.entries(d.stores).map(([v,S])=>[v,v===d.store?y:S.getState()])))),i.dispatchFromDevtools&&typeof i.dispatch=="function"){let v=!1;const S=i.dispatch;i.dispatch=(..._)=>{(YE?"production":void 0)!=="production"&&_[0].type==="__setState"&&!v&&(console.warn('[zustand devtools middleware] "__setState" action type is reserved to set state from the devtools. Avoid using it.'),v=!0),S(..._)}}return c.subscribe(v=>{var S;switch(v.type){case"ACTION":if(typeof v.payload!="string"){console.error("[zustand devtools middleware] Unsupported action format");return}return MT(v.payload,_=>{if(_.type==="__setState"){if(l===void 0){m(_.state);return}Object.keys(_.state).length!==1&&console.error(`
                    [zustand devtools middleware] Unsupported __setState action format. 
                    When using 'store' option in devtools(), the 'state' should have only one key, which is a value of 'store' that was passed in devtools(),
                    and value of this only key should be a state object. Example: { "type": "__setState", "state": { "abc123Store": { "foo": "bar" } } }
                    `);const E=_.state[l];if(E==null)return;JSON.stringify(i.getState())!==JSON.stringify(E)&&m(E);return}i.dispatchFromDevtools&&typeof i.dispatch=="function"&&i.dispatch(_)});case"DISPATCH":switch(v.payload.type){case"RESET":return m(y),l===void 0?c==null?void 0:c.init(i.getState()):c==null?void 0:c.init(F_(o.name));case"COMMIT":if(l===void 0){c==null||c.init(i.getState());return}return c==null?void 0:c.init(F_(o.name));case"ROLLBACK":return MT(v.state,_=>{if(l===void 0){m(_),c==null||c.init(i.getState());return}m(_[l]),c==null||c.init(F_(o.name))});case"JUMP_TO_STATE":case"JUMP_TO_ACTION":return MT(v.state,_=>{if(l===void 0){m(_);return}JSON.stringify(i.getState())!==JSON.stringify(_[l])&&m(_[l])});case"IMPORT_STATE":{const{nextLiftedState:_}=v.payload,E=(S=_.computedStates.slice(-1)[0])==null?void 0:S.state;if(!E)return;m(l===void 0?E:E[l]),c==null||c.send(null,_);return}case"PAUSE_RECORDING":return h=!h}return}}),y},ort=R2e,MT=(e,t)=>{let n;try{n=JSON.parse(e)}catch(r){console.error("[zustand devtools middleware] Could not parse the received json",r)}n!==void 0&&t(n)};function A2e(e,t){let n;try{n=e()}catch{return}return{getItem:i=>{var a;const u=o=>o===null?null:JSON.parse(o,void 0),l=(a=n.getItem(i))!=null?a:null;return l instanceof Promise?l.then(u):u(l)},setItem:(i,a)=>n.setItem(i,JSON.stringify(a,void 0)),removeItem:i=>n.removeItem(i)}}const dx=e=>t=>{try{const n=e(t);return n instanceof Promise?n:{then(r){return dx(r)(n)},catch(r){return this}}}catch(n){return{then(r){return this},catch(r){return dx(r)(n)}}}},T2e=(e,t)=>(n,r,i)=>{let a={getStorage:()=>localStorage,serialize:JSON.stringify,deserialize:JSON.parse,partialize:S=>S,version:0,merge:(S,_)=>({..._,...S}),...t},u=!1;const l=new Set,o=new Set;let s;try{s=a.getStorage()}catch{}if(!s)return e((...S)=>{console.warn(`[zustand persist middleware] Unable to update item '${a.name}', the given storage is currently unavailable.`),n(...S)},r,i);const c=dx(a.serialize),d=()=>{const S=a.partialize({...r()});let _;const E=c({state:S,version:a.version}).then($=>s.setItem(a.name,$)).catch($=>{_=$});if(_)throw _;return E},h=i.setState;i.setState=(S,_)=>{h(S,_),d()};const m=e((...S)=>{n(...S),d()},r,i);let y;const v=()=>{var S;if(!s)return;u=!1,l.forEach(E=>E(r()));const _=((S=a.onRehydrateStorage)==null?void 0:S.call(a,r()))||void 0;return dx(s.getItem.bind(s))(a.name).then(E=>{if(E)return a.deserialize(E)}).then(E=>{if(E)if(typeof E.version=="number"&&E.version!==a.version){if(a.migrate)return a.migrate(E.state,E.version);console.error("State loaded from storage couldn't be migrated since no migrate function was provided")}else return E.state}).then(E=>{var $;return y=a.merge(E,($=r())!=null?$:m),n(y,!0),d()}).then(()=>{_==null||_(y,void 0),u=!0,o.forEach(E=>E(y))}).catch(E=>{_==null||_(void 0,E)})};return i.persist={setOptions:S=>{a={...a,...S},S.getStorage&&(s=S.getStorage())},clearStorage:()=>{s==null||s.removeItem(a.name)},getOptions:()=>a,rehydrate:()=>v(),hasHydrated:()=>u,onHydrate:S=>(l.add(S),()=>{l.delete(S)}),onFinishHydration:S=>(o.add(S),()=>{o.delete(S)})},v(),y||m},k2e=(e,t)=>(n,r,i)=>{let a={storage:A2e(()=>localStorage),partialize:v=>v,version:0,merge:(v,S)=>({...S,...v}),...t},u=!1;const l=new Set,o=new Set;let s=a.storage;if(!s)return e((...v)=>{console.warn(`[zustand persist middleware] Unable to update item '${a.name}', the given storage is currently unavailable.`),n(...v)},r,i);const c=()=>{const v=a.partialize({...r()});return s.setItem(a.name,{state:v,version:a.version})},d=i.setState;i.setState=(v,S)=>{d(v,S),c()};const h=e((...v)=>{n(...v),c()},r,i);i.getInitialState=()=>h;let m;const y=()=>{var v,S;if(!s)return;u=!1,l.forEach(E=>{var $;return E(($=r())!=null?$:h)});const _=((S=a.onRehydrateStorage)==null?void 0:S.call(a,(v=r())!=null?v:h))||void 0;return dx(s.getItem.bind(s))(a.name).then(E=>{if(E)if(typeof E.version=="number"&&E.version!==a.version){if(a.migrate)return[!0,a.migrate(E.state,E.version)];console.error("State loaded from storage couldn't be migrated since no migrate function was provided")}else return[!1,E.state];return[!1,void 0]}).then(E=>{var $;const[k,A]=E;if(m=a.merge(A,($=r())!=null?$:h),n(m,!0),k)return c()}).then(()=>{_==null||_(m,void 0),m=r(),u=!0,o.forEach(E=>E(m))}).catch(E=>{_==null||_(void 0,E)})};return i.persist={setOptions:v=>{a={...a,...v},v.storage&&(s=v.storage)},clearStorage:()=>{s==null||s.removeItem(a.name)},getOptions:()=>a,rehydrate:()=>y(),hasHydrated:()=>u,onHydrate:v=>(l.add(v),()=>{l.delete(v)}),onFinishHydration:v=>(o.add(v),()=>{o.delete(v)})},a.skipHydration||y(),m||h},P2e=(e,t)=>"getStorage"in t||"serialize"in t||"deserialize"in t?((YE?"production":void 0)!=="production"&&console.warn("[DEPRECATED] `getStorage`, `serialize` and `deserialize` options are deprecated. Use `storage` option instead."),T2e(e,t)):k2e(e,t),srt=P2e;var mi;(function(e){e.assertEqual=i=>{};function t(i){}e.assertIs=t;function n(i){throw new Error}e.assertNever=n,e.arrayToEnum=i=>{const a={};for(const u of i)a[u]=u;return a},e.getValidEnumValues=i=>{const a=e.objectKeys(i).filter(l=>typeof i[i[l]]!="number"),u={};for(const l of a)u[l]=i[l];return e.objectValues(u)},e.objectValues=i=>e.objectKeys(i).map(function(a){return i[a]}),e.objectKeys=typeof Object.keys=="function"?i=>Object.keys(i):i=>{const a=[];for(const u in i)Object.prototype.hasOwnProperty.call(i,u)&&a.push(u);return a},e.find=(i,a)=>{for(const u of i)if(a(u))return u},e.isInteger=typeof Number.isInteger=="function"?i=>Number.isInteger(i):i=>typeof i=="number"&&Number.isFinite(i)&&Math.floor(i)===i;function r(i,a=" | "){return i.map(u=>typeof u=="string"?`'${u}'`:u).join(a)}e.joinValues=r,e.jsonStringifyReplacer=(i,a)=>typeof a=="bigint"?a.toString():a})(mi||(mi={}));var qV;(function(e){e.mergeShapes=(t,n)=>({...t,...n})})(qV||(qV={}));const cn=mi.arrayToEnum(["string","nan","number","integer","float","boolean","date","bigint","symbol","function","undefined","null","array","object","unknown","promise","void","never","map","set"]),yv=e=>{switch(typeof e){case"undefined":return cn.undefined;case"string":return cn.string;case"number":return Number.isNaN(e)?cn.nan:cn.number;case"boolean":return cn.boolean;case"function":return cn.function;case"bigint":return cn.bigint;case"symbol":return cn.symbol;case"object":return Array.isArray(e)?cn.array:e===null?cn.null:e.then&&typeof e.then=="function"&&e.catch&&typeof e.catch=="function"?cn.promise:typeof Map<"u"&&e instanceof Map?cn.map:typeof Set<"u"&&e instanceof Set?cn.set:typeof Date<"u"&&e instanceof Date?cn.date:cn.object;default:return cn.unknown}},Nt=mi.arrayToEnum(["invalid_type","invalid_literal","custom","invalid_union","invalid_union_discriminator","invalid_enum_value","unrecognized_keys","invalid_arguments","invalid_return_type","invalid_date","invalid_string","too_small","too_big","invalid_intersection_types","not_multiple_of","not_finite"]);class og extends Error{get errors(){return this.issues}constructor(t){super(),this.issues=[],this.addIssue=r=>{this.issues=[...this.issues,r]},this.addIssues=(r=[])=>{this.issues=[...this.issues,...r]};const n=new.target.prototype;Object.setPrototypeOf?Object.setPrototypeOf(this,n):this.__proto__=n,this.name="ZodError",this.issues=t}format(t){const n=t||function(a){return a.message},r={_errors:[]},i=a=>{for(const u of a.issues)if(u.code==="invalid_union")u.unionErrors.map(i);else if(u.code==="invalid_return_type")i(u.returnTypeError);else if(u.code==="invalid_arguments")i(u.argumentsError);else if(u.path.length===0)r._errors.push(n(u));else{let l=r,o=0;for(;o<u.path.length;){const s=u.path[o];o===u.path.length-1?(l[s]=l[s]||{_errors:[]},l[s]._errors.push(n(u))):l[s]=l[s]||{_errors:[]},l=l[s],o++}}};return i(this),r}static assert(t){if(!(t instanceof og))throw new Error(`Not a ZodError: ${t}`)}toString(){return this.message}get message(){return JSON.stringify(this.issues,mi.jsonStringifyReplacer,2)}get isEmpty(){return this.issues.length===0}flatten(t=n=>n.message){const n={},r=[];for(const i of this.issues)if(i.path.length>0){const a=i.path[0];n[a]=n[a]||[],n[a].push(t(i))}else r.push(t(i));return{formErrors:r,fieldErrors:n}}get formErrors(){return this.flatten()}}og.create=e=>new og(e);const u9=(e,t)=>{let n;switch(e.code){case Nt.invalid_type:e.received===cn.undefined?n="Required":n=`Expected ${e.expected}, received ${e.received}`;break;case Nt.invalid_literal:n=`Invalid literal value, expected ${JSON.stringify(e.expected,mi.jsonStringifyReplacer)}`;break;case Nt.unrecognized_keys:n=`Unrecognized key(s) in object: ${mi.joinValues(e.keys,", ")}`;break;case Nt.invalid_union:n="Invalid input";break;case Nt.invalid_union_discriminator:n=`Invalid discriminator value. Expected ${mi.joinValues(e.options)}`;break;case Nt.invalid_enum_value:n=`Invalid enum value. Expected ${mi.joinValues(e.options)}, received '${e.received}'`;break;case Nt.invalid_arguments:n="Invalid function arguments";break;case Nt.invalid_return_type:n="Invalid function return type";break;case Nt.invalid_date:n="Invalid date";break;case Nt.invalid_string:typeof e.validation=="object"?"includes"in e.validation?(n=`Invalid input: must include "${e.validation.includes}"`,typeof e.validation.position=="number"&&(n=`${n} at one or more positions greater than or equal to ${e.validation.position}`)):"startsWith"in e.validation?n=`Invalid input: must start with "${e.validation.startsWith}"`:"endsWith"in e.validation?n=`Invalid input: must end with "${e.validation.endsWith}"`:mi.assertNever(e.validation):e.validation!=="regex"?n=`Invalid ${e.validation}`:n="Invalid";break;case Nt.too_small:e.type==="array"?n=`Array must contain ${e.exact?"exactly":e.inclusive?"at least":"more than"} ${e.minimum} element(s)`:e.type==="string"?n=`String must contain ${e.exact?"exactly":e.inclusive?"at least":"over"} ${e.minimum} character(s)`:e.type==="number"?n=`Number must be ${e.exact?"exactly equal to ":e.inclusive?"greater than or equal to ":"greater than "}${e.minimum}`:e.type==="bigint"?n=`Number must be ${e.exact?"exactly equal to ":e.inclusive?"greater than or equal to ":"greater than "}${e.minimum}`:e.type==="date"?n=`Date must be ${e.exact?"exactly equal to ":e.inclusive?"greater than or equal to ":"greater than "}${new Date(Number(e.minimum))}`:n="Invalid input";break;case Nt.too_big:e.type==="array"?n=`Array must contain ${e.exact?"exactly":e.inclusive?"at most":"less than"} ${e.maximum} element(s)`:e.type==="string"?n=`String must contain ${e.exact?"exactly":e.inclusive?"at most":"under"} ${e.maximum} character(s)`:e.type==="number"?n=`Number must be ${e.exact?"exactly":e.inclusive?"less than or equal to":"less than"} ${e.maximum}`:e.type==="bigint"?n=`BigInt must be ${e.exact?"exactly":e.inclusive?"less than or equal to":"less than"} ${e.maximum}`:e.type==="date"?n=`Date must be ${e.exact?"exactly":e.inclusive?"smaller than or equal to":"smaller than"} ${new Date(Number(e.maximum))}`:n="Invalid input";break;case Nt.custom:n="Invalid input";break;case Nt.invalid_intersection_types:n="Intersection results could not be merged";break;case Nt.not_multiple_of:n=`Number must be a multiple of ${e.multipleOf}`;break;case Nt.not_finite:n="Number must be finite";break;default:n=t.defaultError,mi.assertNever(e)}return{message:n}};let M2e=u9;function F2e(){return M2e}const N2e=e=>{const{data:t,path:n,errorMaps:r,issueData:i}=e,a=[...n,...i.path||[]],u={...i,path:a};if(i.message!==void 0)return{...i,path:a,message:i.message};let l="";const o=r.filter(s=>!!s).slice().reverse();for(const s of o)l=s(u,{data:t,defaultError:l}).message;return{...i,path:a,message:l}};function tn(e,t){const n=F2e(),r=N2e({issueData:t,data:e.data,path:e.path,errorMaps:[e.common.contextualErrorMap,e.schemaErrorMap,n,n===u9?void 0:u9].filter(i=>!!i)});e.common.issues.push(r)}class hc{constructor(){this.value="valid"}dirty(){this.value==="valid"&&(this.value="dirty")}abort(){this.value!=="aborted"&&(this.value="aborted")}static mergeArray(t,n){const r=[];for(const i of n){if(i.status==="aborted")return Gn;i.status==="dirty"&&t.dirty(),r.push(i.value)}return{status:t.value,value:r}}static async mergeObjectAsync(t,n){const r=[];for(const i of n){const a=await i.key,u=await i.value;r.push({key:a,value:u})}return hc.mergeObjectSync(t,r)}static mergeObjectSync(t,n){const r={};for(const i of n){const{key:a,value:u}=i;if(a.status==="aborted"||u.status==="aborted")return Gn;a.status==="dirty"&&t.dirty(),u.status==="dirty"&&t.dirty(),a.value!=="__proto__"&&(typeof u.value<"u"||i.alwaysSet)&&(r[a.value]=u.value)}return{status:t.value,value:r}}}const Gn=Object.freeze({status:"aborted"}),A2=e=>({status:"dirty",value:e}),rh=e=>({status:"valid",value:e}),VV=e=>e.status==="aborted",UV=e=>e.status==="dirty",D3=e=>e.status==="valid",AD=e=>typeof Promise<"u"&&e instanceof Promise;var bn;(function(e){e.errToObj=t=>typeof t=="string"?{message:t}:t||{},e.toString=t=>typeof t=="string"?t:t==null?void 0:t.message})(bn||(bn={}));class A0{constructor(t,n,r,i){this._cachedPath=[],this.parent=t,this.data=n,this._path=r,this._key=i}get path(){return this._cachedPath.length||(Array.isArray(this._key)?this._cachedPath.push(...this._path,...this._key):this._cachedPath.push(...this._path,this._key)),this._cachedPath}}const jV=(e,t)=>{if(D3(t))return{success:!0,data:t.value};if(!e.common.issues.length)throw new Error("Validation failed but no issues detected.");return{success:!1,get error(){if(this._error)return this._error;const n=new og(e.common.issues);return this._error=n,this._error}}};function yr(e){if(!e)return{};const{errorMap:t,invalid_type_error:n,required_error:r,description:i}=e;if(t&&(n||r))throw new Error(`Can't use "invalid_type_error" or "required_error" in conjunction with custom error map.`);return t?{errorMap:t,description:i}:{errorMap:(u,l)=>{const{message:o}=e;return u.code==="invalid_enum_value"?{message:o??l.defaultError}:typeof l.data>"u"?{message:o??r??l.defaultError}:u.code!=="invalid_type"?{message:l.defaultError}:{message:o??n??l.defaultError}},description:i}}class Yr{get description(){return this._def.description}_getType(t){return yv(t.data)}_getOrReturnCtx(t,n){return n||{common:t.parent.common,data:t.data,parsedType:yv(t.data),schemaErrorMap:this._def.errorMap,path:t.path,parent:t.parent}}_processInputParams(t){return{status:new hc,ctx:{common:t.parent.common,data:t.data,parsedType:yv(t.data),schemaErrorMap:this._def.errorMap,path:t.path,parent:t.parent}}}_parseSync(t){const n=this._parse(t);if(AD(n))throw new Error("Synchronous parse encountered promise.");return n}_parseAsync(t){const n=this._parse(t);return Promise.resolve(n)}parse(t,n){const r=this.safeParse(t,n);if(r.success)return r.data;throw r.error}safeParse(t,n){const r={common:{issues:[],async:(n==null?void 0:n.async)??!1,contextualErrorMap:n==null?void 0:n.errorMap},path:(n==null?void 0:n.path)||[],schemaErrorMap:this._def.errorMap,parent:null,data:t,parsedType:yv(t)},i=this._parseSync({data:t,path:r.path,parent:r});return jV(r,i)}"~validate"(t){var r,i;const n={common:{issues:[],async:!!this["~standard"].async},path:[],schemaErrorMap:this._def.errorMap,parent:null,data:t,parsedType:yv(t)};if(!this["~standard"].async)try{const a=this._parseSync({data:t,path:[],parent:n});return D3(a)?{value:a.value}:{issues:n.common.issues}}catch(a){(i=(r=a==null?void 0:a.message)==null?void 0:r.toLowerCase())!=null&&i.includes("encountered")&&(this["~standard"].async=!0),n.common={issues:[],async:!0}}return this._parseAsync({data:t,path:[],parent:n}).then(a=>D3(a)?{value:a.value}:{issues:n.common.issues})}async parseAsync(t,n){const r=await this.safeParseAsync(t,n);if(r.success)return r.data;throw r.error}async safeParseAsync(t,n){const r={common:{issues:[],contextualErrorMap:n==null?void 0:n.errorMap,async:!0},path:(n==null?void 0:n.path)||[],schemaErrorMap:this._def.errorMap,parent:null,data:t,parsedType:yv(t)},i=this._parse({data:t,path:r.path,parent:r}),a=await(AD(i)?i:Promise.resolve(i));return jV(r,a)}refine(t,n){const r=i=>typeof n=="string"||typeof n>"u"?{message:n}:typeof n=="function"?n(i):n;return this._refinement((i,a)=>{const u=t(i),l=()=>a.addIssue({code:Nt.custom,...r(i)});return typeof Promise<"u"&&u instanceof Promise?u.then(o=>o?!0:(l(),!1)):u?!0:(l(),!1)})}refinement(t,n){return this._refinement((r,i)=>t(r)?!0:(i.addIssue(typeof n=="function"?n(r,i):n),!1))}_refinement(t){return new n1({schema:this,typeName:It.ZodEffects,effect:{type:"refinement",refinement:t}})}superRefine(t){return this._refinement(t)}constructor(t){this.spa=this.safeParseAsync,this._def=t,this.parse=this.parse.bind(this),this.safeParse=this.safeParse.bind(this),this.parseAsync=this.parseAsync.bind(this),this.safeParseAsync=this.safeParseAsync.bind(this),this.spa=this.spa.bind(this),this.refine=this.refine.bind(this),this.refinement=this.refinement.bind(this),this.superRefine=this.superRefine.bind(this),this.optional=this.optional.bind(this),this.nullable=this.nullable.bind(this),this.nullish=this.nullish.bind(this),this.array=this.array.bind(this),this.promise=this.promise.bind(this),this.or=this.or.bind(this),this.and=this.and.bind(this),this.transform=this.transform.bind(this),this.brand=this.brand.bind(this),this.default=this.default.bind(this),this.catch=this.catch.bind(this),this.describe=this.describe.bind(this),this.pipe=this.pipe.bind(this),this.readonly=this.readonly.bind(this),this.isNullable=this.isNullable.bind(this),this.isOptional=this.isOptional.bind(this),this["~standard"]={version:1,vendor:"zod",validate:n=>this["~validate"](n)}}optional(){return rg.create(this,this._def)}nullable(){return r1.create(this,this._def)}nullish(){return this.nullable().optional()}array(){return D0.create(this)}promise(){return LD.create(this,this._def)}or(t){return PD.create([this,t],this._def)}and(t){return MD.create(this,t,this._def)}transform(t){return new n1({...yr(this._def),schema:this,typeName:It.ZodEffects,effect:{type:"transform",transform:t}})}default(t){const n=typeof t=="function"?t:()=>t;return new BD({...yr(this._def),innerType:this,defaultValue:n,typeName:It.ZodDefault})}brand(){return new Cee({typeName:It.ZodBranded,type:this,...yr(this._def)})}catch(t){const n=typeof t=="function"?t:()=>t;return new QD({...yr(this._def),innerType:this,catchValue:n,typeName:It.ZodCatch})}describe(t){const n=this.constructor;return new n({...this._def,description:t})}pipe(t){return iN.create(this,t)}readonly(){return zD.create(this)}isOptional(){return this.safeParse(void 0).success}isNullable(){return this.safeParse(null).success}}const I2e=/^c[^\s-]{8,}$/i,L2e=/^[0-9a-z]+$/,B2e=/^[0-9A-HJKMNP-TV-Z]{26}$/i,Q2e=/^[0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12}$/i,z2e=/^[a-z0-9_-]{21}$/i,q2e=/^[A-Za-z0-9-_]+\.[A-Za-z0-9-_]+\.[A-Za-z0-9-_]*$/,V2e=/^[-+]?P(?!$)(?:(?:[-+]?\d+Y)|(?:[-+]?\d+[.,]\d+Y$))?(?:(?:[-+]?\d+M)|(?:[-+]?\d+[.,]\d+M$))?(?:(?:[-+]?\d+W)|(?:[-+]?\d+[.,]\d+W$))?(?:(?:[-+]?\d+D)|(?:[-+]?\d+[.,]\d+D$))?(?:T(?=[\d+-])(?:(?:[-+]?\d+H)|(?:[-+]?\d+[.,]\d+H$))?(?:(?:[-+]?\d+M)|(?:[-+]?\d+[.,]\d+M$))?(?:[-+]?\d+(?:[.,]\d+)?S)?)??$/,U2e=/^(?!\.)(?!.*\.\.)([A-Z0-9_'+\-\.]*)[A-Z0-9_+-]@([A-Z0-9][A-Z0-9\-]*\.)+[A-Z]{2,}$/i,j2e="^(\\p{Extended_Pictographic}|\\p{Emoji_Component})+$";let FT;const Z2e=/^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])$/,H2e=/^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\/(3[0-2]|[12]?[0-9])$/,X2e=/^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))$/,Y2e=/^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))\/(12[0-8]|1[01][0-9]|[1-9]?[0-9])$/,G2e=/^([0-9a-zA-Z+/]{4})*(([0-9a-zA-Z+/]{2}==)|([0-9a-zA-Z+/]{3}=))?$/,K2e=/^([0-9a-zA-Z-_]{4})*(([0-9a-zA-Z-_]{2}(==)?)|([0-9a-zA-Z-_]{3}(=)?))?$/,Eee="((\\d\\d[2468][048]|\\d\\d[13579][26]|\\d\\d0[48]|[02468][048]00|[13579][26]00)-02-29|\\d{4}-((0[13578]|1[02])-(0[1-9]|[12]\\d|3[01])|(0[469]|11)-(0[1-9]|[12]\\d|30)|(02)-(0[1-9]|1\\d|2[0-8])))",W2e=new RegExp(`^${Eee}$`);function Dee(e){let t="[0-5]\\d";e.precision?t=`${t}\\.\\d{${e.precision}}`:e.precision==null&&(t=`${t}(\\.\\d+)?`);const n=e.precision?"+":"?";return`([01]\\d|2[0-3]):[0-5]\\d(:${t})${n}`}function J2e(e){return new RegExp(`^${Dee(e)}$`)}function exe(e){let t=`${Eee}T${Dee(e)}`;const n=[];return n.push(e.local?"Z?":"Z"),e.offset&&n.push("([+-]\\d{2}:?\\d{2})"),t=`${t}(${n.join("|")})`,new RegExp(`^${t}$`)}function txe(e,t){return!!((t==="v4"||!t)&&Z2e.test(e)||(t==="v6"||!t)&&X2e.test(e))}function nxe(e,t){if(!q2e.test(e))return!1;try{const[n]=e.split(".");if(!n)return!1;const r=n.replace(/-/g,"+").replace(/_/g,"/").padEnd(n.length+(4-n.length%4)%4,"="),i=JSON.parse(atob(r));return!(typeof i!="object"||i===null||"typ"in i&&(i==null?void 0:i.typ)!=="JWT"||!i.alg||t&&i.alg!==t)}catch{return!1}}function rxe(e,t){return!!((t==="v4"||!t)&&H2e.test(e)||(t==="v6"||!t)&&Y2e.test(e))}class Km extends Yr{_parse(t){if(this._def.coerce&&(t.data=String(t.data)),this._getType(t)!==cn.string){const a=this._getOrReturnCtx(t);return tn(a,{code:Nt.invalid_type,expected:cn.string,received:a.parsedType}),Gn}const r=new hc;let i;for(const a of this._def.checks)if(a.kind==="min")t.data.length<a.value&&(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.too_small,minimum:a.value,type:"string",inclusive:!0,exact:!1,message:a.message}),r.dirty());else if(a.kind==="max")t.data.length>a.value&&(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.too_big,maximum:a.value,type:"string",inclusive:!0,exact:!1,message:a.message}),r.dirty());else if(a.kind==="length"){const u=t.data.length>a.value,l=t.data.length<a.value;(u||l)&&(i=this._getOrReturnCtx(t,i),u?tn(i,{code:Nt.too_big,maximum:a.value,type:"string",inclusive:!0,exact:!0,message:a.message}):l&&tn(i,{code:Nt.too_small,minimum:a.value,type:"string",inclusive:!0,exact:!0,message:a.message}),r.dirty())}else if(a.kind==="email")U2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"email",code:Nt.invalid_string,message:a.message}),r.dirty());else if(a.kind==="emoji")FT||(FT=new RegExp(j2e,"u")),FT.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"emoji",code:Nt.invalid_string,message:a.message}),r.dirty());else if(a.kind==="uuid")Q2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"uuid",code:Nt.invalid_string,message:a.message}),r.dirty());else if(a.kind==="nanoid")z2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"nanoid",code:Nt.invalid_string,message:a.message}),r.dirty());else if(a.kind==="cuid")I2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"cuid",code:Nt.invalid_string,message:a.message}),r.dirty());else if(a.kind==="cuid2")L2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"cuid2",code:Nt.invalid_string,message:a.message}),r.dirty());else if(a.kind==="ulid")B2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"ulid",code:Nt.invalid_string,message:a.message}),r.dirty());else if(a.kind==="url")try{new URL(t.data)}catch{i=this._getOrReturnCtx(t,i),tn(i,{validation:"url",code:Nt.invalid_string,message:a.message}),r.dirty()}else a.kind==="regex"?(a.regex.lastIndex=0,a.regex.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"regex",code:Nt.invalid_string,message:a.message}),r.dirty())):a.kind==="trim"?t.data=t.data.trim():a.kind==="includes"?t.data.includes(a.value,a.position)||(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.invalid_string,validation:{includes:a.value,position:a.position},message:a.message}),r.dirty()):a.kind==="toLowerCase"?t.data=t.data.toLowerCase():a.kind==="toUpperCase"?t.data=t.data.toUpperCase():a.kind==="startsWith"?t.data.startsWith(a.value)||(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.invalid_string,validation:{startsWith:a.value},message:a.message}),r.dirty()):a.kind==="endsWith"?t.data.endsWith(a.value)||(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.invalid_string,validation:{endsWith:a.value},message:a.message}),r.dirty()):a.kind==="datetime"?exe(a).test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.invalid_string,validation:"datetime",message:a.message}),r.dirty()):a.kind==="date"?W2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.invalid_string,validation:"date",message:a.message}),r.dirty()):a.kind==="time"?J2e(a).test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.invalid_string,validation:"time",message:a.message}),r.dirty()):a.kind==="duration"?V2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"duration",code:Nt.invalid_string,message:a.message}),r.dirty()):a.kind==="ip"?txe(t.data,a.version)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"ip",code:Nt.invalid_string,message:a.message}),r.dirty()):a.kind==="jwt"?nxe(t.data,a.alg)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"jwt",code:Nt.invalid_string,message:a.message}),r.dirty()):a.kind==="cidr"?rxe(t.data,a.version)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"cidr",code:Nt.invalid_string,message:a.message}),r.dirty()):a.kind==="base64"?G2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"base64",code:Nt.invalid_string,message:a.message}),r.dirty()):a.kind==="base64url"?K2e.test(t.data)||(i=this._getOrReturnCtx(t,i),tn(i,{validation:"base64url",code:Nt.invalid_string,message:a.message}),r.dirty()):mi.assertNever(a);return{status:r.value,value:t.data}}_regex(t,n,r){return this.refinement(i=>t.test(i),{validation:n,code:Nt.invalid_string,...bn.errToObj(r)})}_addCheck(t){return new Km({...this._def,checks:[...this._def.checks,t]})}email(t){return this._addCheck({kind:"email",...bn.errToObj(t)})}url(t){return this._addCheck({kind:"url",...bn.errToObj(t)})}emoji(t){return this._addCheck({kind:"emoji",...bn.errToObj(t)})}uuid(t){return this._addCheck({kind:"uuid",...bn.errToObj(t)})}nanoid(t){return this._addCheck({kind:"nanoid",...bn.errToObj(t)})}cuid(t){return this._addCheck({kind:"cuid",...bn.errToObj(t)})}cuid2(t){return this._addCheck({kind:"cuid2",...bn.errToObj(t)})}ulid(t){return this._addCheck({kind:"ulid",...bn.errToObj(t)})}base64(t){return this._addCheck({kind:"base64",...bn.errToObj(t)})}base64url(t){return this._addCheck({kind:"base64url",...bn.errToObj(t)})}jwt(t){return this._addCheck({kind:"jwt",...bn.errToObj(t)})}ip(t){return this._addCheck({kind:"ip",...bn.errToObj(t)})}cidr(t){return this._addCheck({kind:"cidr",...bn.errToObj(t)})}datetime(t){return typeof t=="string"?this._addCheck({kind:"datetime",precision:null,offset:!1,local:!1,message:t}):this._addCheck({kind:"datetime",precision:typeof(t==null?void 0:t.precision)>"u"?null:t==null?void 0:t.precision,offset:(t==null?void 0:t.offset)??!1,local:(t==null?void 0:t.local)??!1,...bn.errToObj(t==null?void 0:t.message)})}date(t){return this._addCheck({kind:"date",message:t})}time(t){return typeof t=="string"?this._addCheck({kind:"time",precision:null,message:t}):this._addCheck({kind:"time",precision:typeof(t==null?void 0:t.precision)>"u"?null:t==null?void 0:t.precision,...bn.errToObj(t==null?void 0:t.message)})}duration(t){return this._addCheck({kind:"duration",...bn.errToObj(t)})}regex(t,n){return this._addCheck({kind:"regex",regex:t,...bn.errToObj(n)})}includes(t,n){return this._addCheck({kind:"includes",value:t,position:n==null?void 0:n.position,...bn.errToObj(n==null?void 0:n.message)})}startsWith(t,n){return this._addCheck({kind:"startsWith",value:t,...bn.errToObj(n)})}endsWith(t,n){return this._addCheck({kind:"endsWith",value:t,...bn.errToObj(n)})}min(t,n){return this._addCheck({kind:"min",value:t,...bn.errToObj(n)})}max(t,n){return this._addCheck({kind:"max",value:t,...bn.errToObj(n)})}length(t,n){return this._addCheck({kind:"length",value:t,...bn.errToObj(n)})}nonempty(t){return this.min(1,bn.errToObj(t))}trim(){return new Km({...this._def,checks:[...this._def.checks,{kind:"trim"}]})}toLowerCase(){return new Km({...this._def,checks:[...this._def.checks,{kind:"toLowerCase"}]})}toUpperCase(){return new Km({...this._def,checks:[...this._def.checks,{kind:"toUpperCase"}]})}get isDatetime(){return!!this._def.checks.find(t=>t.kind==="datetime")}get isDate(){return!!this._def.checks.find(t=>t.kind==="date")}get isTime(){return!!this._def.checks.find(t=>t.kind==="time")}get isDuration(){return!!this._def.checks.find(t=>t.kind==="duration")}get isEmail(){return!!this._def.checks.find(t=>t.kind==="email")}get isURL(){return!!this._def.checks.find(t=>t.kind==="url")}get isEmoji(){return!!this._def.checks.find(t=>t.kind==="emoji")}get isUUID(){return!!this._def.checks.find(t=>t.kind==="uuid")}get isNANOID(){return!!this._def.checks.find(t=>t.kind==="nanoid")}get isCUID(){return!!this._def.checks.find(t=>t.kind==="cuid")}get isCUID2(){return!!this._def.checks.find(t=>t.kind==="cuid2")}get isULID(){return!!this._def.checks.find(t=>t.kind==="ulid")}get isIP(){return!!this._def.checks.find(t=>t.kind==="ip")}get isCIDR(){return!!this._def.checks.find(t=>t.kind==="cidr")}get isBase64(){return!!this._def.checks.find(t=>t.kind==="base64")}get isBase64url(){return!!this._def.checks.find(t=>t.kind==="base64url")}get minLength(){let t=null;for(const n of this._def.checks)n.kind==="min"&&(t===null||n.value>t)&&(t=n.value);return t}get maxLength(){let t=null;for(const n of this._def.checks)n.kind==="max"&&(t===null||n.value<t)&&(t=n.value);return t}}Km.create=e=>new Km({checks:[],typeName:It.ZodString,coerce:(e==null?void 0:e.coerce)??!1,...yr(e)});function ixe(e,t){const n=(e.toString().split(".")[1]||"").length,r=(t.toString().split(".")[1]||"").length,i=n>r?n:r,a=Number.parseInt(e.toFixed(i).replace(".","")),u=Number.parseInt(t.toFixed(i).replace(".",""));return a%u/10**i}class $3 extends Yr{constructor(){super(...arguments),this.min=this.gte,this.max=this.lte,this.step=this.multipleOf}_parse(t){if(this._def.coerce&&(t.data=Number(t.data)),this._getType(t)!==cn.number){const a=this._getOrReturnCtx(t);return tn(a,{code:Nt.invalid_type,expected:cn.number,received:a.parsedType}),Gn}let r;const i=new hc;for(const a of this._def.checks)a.kind==="int"?mi.isInteger(t.data)||(r=this._getOrReturnCtx(t,r),tn(r,{code:Nt.invalid_type,expected:"integer",received:"float",message:a.message}),i.dirty()):a.kind==="min"?(a.inclusive?t.data<a.value:t.data<=a.value)&&(r=this._getOrReturnCtx(t,r),tn(r,{code:Nt.too_small,minimum:a.value,type:"number",inclusive:a.inclusive,exact:!1,message:a.message}),i.dirty()):a.kind==="max"?(a.inclusive?t.data>a.value:t.data>=a.value)&&(r=this._getOrReturnCtx(t,r),tn(r,{code:Nt.too_big,maximum:a.value,type:"number",inclusive:a.inclusive,exact:!1,message:a.message}),i.dirty()):a.kind==="multipleOf"?ixe(t.data,a.value)!==0&&(r=this._getOrReturnCtx(t,r),tn(r,{code:Nt.not_multiple_of,multipleOf:a.value,message:a.message}),i.dirty()):a.kind==="finite"?Number.isFinite(t.data)||(r=this._getOrReturnCtx(t,r),tn(r,{code:Nt.not_finite,message:a.message}),i.dirty()):mi.assertNever(a);return{status:i.value,value:t.data}}gte(t,n){return this.setLimit("min",t,!0,bn.toString(n))}gt(t,n){return this.setLimit("min",t,!1,bn.toString(n))}lte(t,n){return this.setLimit("max",t,!0,bn.toString(n))}lt(t,n){return this.setLimit("max",t,!1,bn.toString(n))}setLimit(t,n,r,i){return new $3({...this._def,checks:[...this._def.checks,{kind:t,value:n,inclusive:r,message:bn.toString(i)}]})}_addCheck(t){return new $3({...this._def,checks:[...this._def.checks,t]})}int(t){return this._addCheck({kind:"int",message:bn.toString(t)})}positive(t){return this._addCheck({kind:"min",value:0,inclusive:!1,message:bn.toString(t)})}negative(t){return this._addCheck({kind:"max",value:0,inclusive:!1,message:bn.toString(t)})}nonpositive(t){return this._addCheck({kind:"max",value:0,inclusive:!0,message:bn.toString(t)})}nonnegative(t){return this._addCheck({kind:"min",value:0,inclusive:!0,message:bn.toString(t)})}multipleOf(t,n){return this._addCheck({kind:"multipleOf",value:t,message:bn.toString(n)})}finite(t){return this._addCheck({kind:"finite",message:bn.toString(t)})}safe(t){return this._addCheck({kind:"min",inclusive:!0,value:Number.MIN_SAFE_INTEGER,message:bn.toString(t)})._addCheck({kind:"max",inclusive:!0,value:Number.MAX_SAFE_INTEGER,message:bn.toString(t)})}get minValue(){let t=null;for(const n of this._def.checks)n.kind==="min"&&(t===null||n.value>t)&&(t=n.value);return t}get maxValue(){let t=null;for(const n of this._def.checks)n.kind==="max"&&(t===null||n.value<t)&&(t=n.value);return t}get isInt(){return!!this._def.checks.find(t=>t.kind==="int"||t.kind==="multipleOf"&&mi.isInteger(t.value))}get isFinite(){let t=null,n=null;for(const r of this._def.checks){if(r.kind==="finite"||r.kind==="int"||r.kind==="multipleOf")return!0;r.kind==="min"?(n===null||r.value>n)&&(n=r.value):r.kind==="max"&&(t===null||r.value<t)&&(t=r.value)}return Number.isFinite(n)&&Number.isFinite(t)}}$3.create=e=>new $3({checks:[],typeName:It.ZodNumber,coerce:(e==null?void 0:e.coerce)||!1,...yr(e)});class hx extends Yr{constructor(){super(...arguments),this.min=this.gte,this.max=this.lte}_parse(t){if(this._def.coerce)try{t.data=BigInt(t.data)}catch{return this._getInvalidInput(t)}if(this._getType(t)!==cn.bigint)return this._getInvalidInput(t);let r;const i=new hc;for(const a of this._def.checks)a.kind==="min"?(a.inclusive?t.data<a.value:t.data<=a.value)&&(r=this._getOrReturnCtx(t,r),tn(r,{code:Nt.too_small,type:"bigint",minimum:a.value,inclusive:a.inclusive,message:a.message}),i.dirty()):a.kind==="max"?(a.inclusive?t.data>a.value:t.data>=a.value)&&(r=this._getOrReturnCtx(t,r),tn(r,{code:Nt.too_big,type:"bigint",maximum:a.value,inclusive:a.inclusive,message:a.message}),i.dirty()):a.kind==="multipleOf"?t.data%a.value!==BigInt(0)&&(r=this._getOrReturnCtx(t,r),tn(r,{code:Nt.not_multiple_of,multipleOf:a.value,message:a.message}),i.dirty()):mi.assertNever(a);return{status:i.value,value:t.data}}_getInvalidInput(t){const n=this._getOrReturnCtx(t);return tn(n,{code:Nt.invalid_type,expected:cn.bigint,received:n.parsedType}),Gn}gte(t,n){return this.setLimit("min",t,!0,bn.toString(n))}gt(t,n){return this.setLimit("min",t,!1,bn.toString(n))}lte(t,n){return this.setLimit("max",t,!0,bn.toString(n))}lt(t,n){return this.setLimit("max",t,!1,bn.toString(n))}setLimit(t,n,r,i){return new hx({...this._def,checks:[...this._def.checks,{kind:t,value:n,inclusive:r,message:bn.toString(i)}]})}_addCheck(t){return new hx({...this._def,checks:[...this._def.checks,t]})}positive(t){return this._addCheck({kind:"min",value:BigInt(0),inclusive:!1,message:bn.toString(t)})}negative(t){return this._addCheck({kind:"max",value:BigInt(0),inclusive:!1,message:bn.toString(t)})}nonpositive(t){return this._addCheck({kind:"max",value:BigInt(0),inclusive:!0,message:bn.toString(t)})}nonnegative(t){return this._addCheck({kind:"min",value:BigInt(0),inclusive:!0,message:bn.toString(t)})}multipleOf(t,n){return this._addCheck({kind:"multipleOf",value:t,message:bn.toString(n)})}get minValue(){let t=null;for(const n of this._def.checks)n.kind==="min"&&(t===null||n.value>t)&&(t=n.value);return t}get maxValue(){let t=null;for(const n of this._def.checks)n.kind==="max"&&(t===null||n.value<t)&&(t=n.value);return t}}hx.create=e=>new hx({checks:[],typeName:It.ZodBigInt,coerce:(e==null?void 0:e.coerce)??!1,...yr(e)});class l9 extends Yr{_parse(t){if(this._def.coerce&&(t.data=!!t.data),this._getType(t)!==cn.boolean){const r=this._getOrReturnCtx(t);return tn(r,{code:Nt.invalid_type,expected:cn.boolean,received:r.parsedType}),Gn}return rh(t.data)}}l9.create=e=>new l9({typeName:It.ZodBoolean,coerce:(e==null?void 0:e.coerce)||!1,...yr(e)});class TD extends Yr{_parse(t){if(this._def.coerce&&(t.data=new Date(t.data)),this._getType(t)!==cn.date){const a=this._getOrReturnCtx(t);return tn(a,{code:Nt.invalid_type,expected:cn.date,received:a.parsedType}),Gn}if(Number.isNaN(t.data.getTime())){const a=this._getOrReturnCtx(t);return tn(a,{code:Nt.invalid_date}),Gn}const r=new hc;let i;for(const a of this._def.checks)a.kind==="min"?t.data.getTime()<a.value&&(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.too_small,message:a.message,inclusive:!0,exact:!1,minimum:a.value,type:"date"}),r.dirty()):a.kind==="max"?t.data.getTime()>a.value&&(i=this._getOrReturnCtx(t,i),tn(i,{code:Nt.too_big,message:a.message,inclusive:!0,exact:!1,maximum:a.value,type:"date"}),r.dirty()):mi.assertNever(a);return{status:r.value,value:new Date(t.data.getTime())}}_addCheck(t){return new TD({...this._def,checks:[...this._def.checks,t]})}min(t,n){return this._addCheck({kind:"min",value:t.getTime(),message:bn.toString(n)})}max(t,n){return this._addCheck({kind:"max",value:t.getTime(),message:bn.toString(n)})}get minDate(){let t=null;for(const n of this._def.checks)n.kind==="min"&&(t===null||n.value>t)&&(t=n.value);return t!=null?new Date(t):null}get maxDate(){let t=null;for(const n of this._def.checks)n.kind==="max"&&(t===null||n.value<t)&&(t=n.value);return t!=null?new Date(t):null}}TD.create=e=>new TD({checks:[],coerce:(e==null?void 0:e.coerce)||!1,typeName:It.ZodDate,...yr(e)});class ZV extends Yr{_parse(t){if(this._getType(t)!==cn.symbol){const r=this._getOrReturnCtx(t);return tn(r,{code:Nt.invalid_type,expected:cn.symbol,received:r.parsedType}),Gn}return rh(t.data)}}ZV.create=e=>new ZV({typeName:It.ZodSymbol,...yr(e)});class o9 extends Yr{_parse(t){if(this._getType(t)!==cn.undefined){const r=this._getOrReturnCtx(t);return tn(r,{code:Nt.invalid_type,expected:cn.undefined,received:r.parsedType}),Gn}return rh(t.data)}}o9.create=e=>new o9({typeName:It.ZodUndefined,...yr(e)});class kD extends Yr{_parse(t){if(this._getType(t)!==cn.null){const r=this._getOrReturnCtx(t);return tn(r,{code:Nt.invalid_type,expected:cn.null,received:r.parsedType}),Gn}return rh(t.data)}}kD.create=e=>new kD({typeName:It.ZodNull,...yr(e)});class HV extends Yr{constructor(){super(...arguments),this._any=!0}_parse(t){return rh(t.data)}}HV.create=e=>new HV({typeName:It.ZodAny,...yr(e)});class s9 extends Yr{constructor(){super(...arguments),this._unknown=!0}_parse(t){return rh(t.data)}}s9.create=e=>new s9({typeName:It.ZodUnknown,...yr(e)});class Yv extends Yr{_parse(t){const n=this._getOrReturnCtx(t);return tn(n,{code:Nt.invalid_type,expected:cn.never,received:n.parsedType}),Gn}}Yv.create=e=>new Yv({typeName:It.ZodNever,...yr(e)});class XV extends Yr{_parse(t){if(this._getType(t)!==cn.undefined){const r=this._getOrReturnCtx(t);return tn(r,{code:Nt.invalid_type,expected:cn.void,received:r.parsedType}),Gn}return rh(t.data)}}XV.create=e=>new XV({typeName:It.ZodVoid,...yr(e)});class D0 extends Yr{_parse(t){const{ctx:n,status:r}=this._processInputParams(t),i=this._def;if(n.parsedType!==cn.array)return tn(n,{code:Nt.invalid_type,expected:cn.array,received:n.parsedType}),Gn;if(i.exactLength!==null){const u=n.data.length>i.exactLength.value,l=n.data.length<i.exactLength.value;(u||l)&&(tn(n,{code:u?Nt.too_big:Nt.too_small,minimum:l?i.exactLength.value:void 0,maximum:u?i.exactLength.value:void 0,type:"array",inclusive:!0,exact:!0,message:i.exactLength.message}),r.dirty())}if(i.minLength!==null&&n.data.length<i.minLength.value&&(tn(n,{code:Nt.too_small,minimum:i.minLength.value,type:"array",inclusive:!0,exact:!1,message:i.minLength.message}),r.dirty()),i.maxLength!==null&&n.data.length>i.maxLength.value&&(tn(n,{code:Nt.too_big,maximum:i.maxLength.value,type:"array",inclusive:!0,exact:!1,message:i.maxLength.message}),r.dirty()),n.common.async)return Promise.all([...n.data].map((u,l)=>i.type._parseAsync(new A0(n,u,n.path,l)))).then(u=>hc.mergeArray(r,u));const a=[...n.data].map((u,l)=>i.type._parseSync(new A0(n,u,n.path,l)));return hc.mergeArray(r,a)}get element(){return this._def.type}min(t,n){return new D0({...this._def,minLength:{value:t,message:bn.toString(n)}})}max(t,n){return new D0({...this._def,maxLength:{value:t,message:bn.toString(n)}})}length(t,n){return new D0({...this._def,exactLength:{value:t,message:bn.toString(n)}})}nonempty(t){return this.min(1,t)}}D0.create=(e,t)=>new D0({type:e,minLength:null,maxLength:null,exactLength:null,typeName:It.ZodArray,...yr(t)});function NO(e){if(e instanceof gl){const t={};for(const n in e.shape){const r=e.shape[n];t[n]=rg.create(NO(r))}return new gl({...e._def,shape:()=>t})}else return e instanceof D0?new D0({...e._def,type:NO(e.element)}):e instanceof rg?rg.create(NO(e.unwrap())):e instanceof r1?r1.create(NO(e.unwrap())):e instanceof e1?e1.create(e.items.map(t=>NO(t))):e}class gl extends Yr{constructor(){super(...arguments),this._cached=null,this.nonstrict=this.passthrough,this.augment=this.extend}_getCached(){if(this._cached!==null)return this._cached;const t=this._def.shape(),n=mi.objectKeys(t);return this._cached={shape:t,keys:n},this._cached}_parse(t){if(this._getType(t)!==cn.object){const s=this._getOrReturnCtx(t);return tn(s,{code:Nt.invalid_type,expected:cn.object,received:s.parsedType}),Gn}const{status:r,ctx:i}=this._processInputParams(t),{shape:a,keys:u}=this._getCached(),l=[];if(!(this._def.catchall instanceof Yv&&this._def.unknownKeys==="strip"))for(const s in i.data)u.includes(s)||l.push(s);const o=[];for(const s of u){const c=a[s],d=i.data[s];o.push({key:{status:"valid",value:s},value:c._parse(new A0(i,d,i.path,s)),alwaysSet:s in i.data})}if(this._def.catchall instanceof Yv){const s=this._def.unknownKeys;if(s==="passthrough")for(const c of l)o.push({key:{status:"valid",value:c},value:{status:"valid",value:i.data[c]}});else if(s==="strict")l.length>0&&(tn(i,{code:Nt.unrecognized_keys,keys:l}),r.dirty());else if(s!=="strip")throw new Error("Internal ZodObject error: invalid unknownKeys value.")}else{const s=this._def.catchall;for(const c of l){const d=i.data[c];o.push({key:{status:"valid",value:c},value:s._parse(new A0(i,d,i.path,c)),alwaysSet:c in i.data})}}return i.common.async?Promise.resolve().then(async()=>{const s=[];for(const c of o){const d=await c.key,h=await c.value;s.push({key:d,value:h,alwaysSet:c.alwaysSet})}return s}).then(s=>hc.mergeObjectSync(r,s)):hc.mergeObjectSync(r,o)}get shape(){return this._def.shape()}strict(t){return bn.errToObj,new gl({...this._def,unknownKeys:"strict",...t!==void 0?{errorMap:(n,r)=>{var a,u;const i=((u=(a=this._def).errorMap)==null?void 0:u.call(a,n,r).message)??r.defaultError;return n.code==="unrecognized_keys"?{message:bn.errToObj(t).message??i}:{message:i}}}:{}})}strip(){return new gl({...this._def,unknownKeys:"strip"})}passthrough(){return new gl({...this._def,unknownKeys:"passthrough"})}extend(t){return new gl({...this._def,shape:()=>({...this._def.shape(),...t})})}merge(t){return new gl({unknownKeys:t._def.unknownKeys,catchall:t._def.catchall,shape:()=>({...this._def.shape(),...t._def.shape()}),typeName:It.ZodObject})}setKey(t,n){return this.augment({[t]:n})}catchall(t){return new gl({...this._def,catchall:t})}pick(t){const n={};for(const r of mi.objectKeys(t))t[r]&&this.shape[r]&&(n[r]=this.shape[r]);return new gl({...this._def,shape:()=>n})}omit(t){const n={};for(const r of mi.objectKeys(this.shape))t[r]||(n[r]=this.shape[r]);return new gl({...this._def,shape:()=>n})}deepPartial(){return NO(this)}partial(t){const n={};for(const r of mi.objectKeys(this.shape)){const i=this.shape[r];t&&!t[r]?n[r]=i:n[r]=i.optional()}return new gl({...this._def,shape:()=>n})}required(t){const n={};for(const r of mi.objectKeys(this.shape))if(t&&!t[r])n[r]=this.shape[r];else{let a=this.shape[r];for(;a instanceof rg;)a=a._def.innerType;n[r]=a}return new gl({...this._def,shape:()=>n})}keyof(){return $ee(mi.objectKeys(this.shape))}}gl.create=(e,t)=>new gl({shape:()=>e,unknownKeys:"strip",catchall:Yv.create(),typeName:It.ZodObject,...yr(t)});gl.strictCreate=(e,t)=>new gl({shape:()=>e,unknownKeys:"strict",catchall:Yv.create(),typeName:It.ZodObject,...yr(t)});gl.lazycreate=(e,t)=>new gl({shape:e,unknownKeys:"strip",catchall:Yv.create(),typeName:It.ZodObject,...yr(t)});class PD extends Yr{_parse(t){const{ctx:n}=this._processInputParams(t),r=this._def.options;function i(a){for(const l of a)if(l.result.status==="valid")return l.result;for(const l of a)if(l.result.status==="dirty")return n.common.issues.push(...l.ctx.common.issues),l.result;const u=a.map(l=>new og(l.ctx.common.issues));return tn(n,{code:Nt.invalid_union,unionErrors:u}),Gn}if(n.common.async)return Promise.all(r.map(async a=>{const u={...n,common:{...n.common,issues:[]},parent:null};return{result:await a._parseAsync({data:n.data,path:n.path,parent:u}),ctx:u}})).then(i);{let a;const u=[];for(const o of r){const s={...n,common:{...n.common,issues:[]},parent:null},c=o._parseSync({data:n.data,path:n.path,parent:s});if(c.status==="valid")return c;c.status==="dirty"&&!a&&(a={result:c,ctx:s}),s.common.issues.length&&u.push(s.common.issues)}if(a)return n.common.issues.push(...a.ctx.common.issues),a.result;const l=u.map(o=>new og(o));return tn(n,{code:Nt.invalid_union,unionErrors:l}),Gn}}get options(){return this._def.options}}PD.create=(e,t)=>new PD({options:e,typeName:It.ZodUnion,...yr(t)});const km=e=>e instanceof ND?km(e.schema):e instanceof n1?km(e.innerType()):e instanceof ID?[e.value]:e instanceof t1?e.options:e instanceof f9?mi.objectValues(e.enum):e instanceof BD?km(e._def.innerType):e instanceof o9?[void 0]:e instanceof kD?[null]:e instanceof rg?[void 0,...km(e.unwrap())]:e instanceof r1?[null,...km(e.unwrap())]:e instanceof Cee||e instanceof zD?km(e.unwrap()):e instanceof QD?km(e._def.innerType):[];class rN extends Yr{_parse(t){const{ctx:n}=this._processInputParams(t);if(n.parsedType!==cn.object)return tn(n,{code:Nt.invalid_type,expected:cn.object,received:n.parsedType}),Gn;const r=this.discriminator,i=n.data[r],a=this.optionsMap.get(i);return a?n.common.async?a._parseAsync({data:n.data,path:n.path,parent:n}):a._parseSync({data:n.data,path:n.path,parent:n}):(tn(n,{code:Nt.invalid_union_discriminator,options:Array.from(this.optionsMap.keys()),path:[r]}),Gn)}get discriminator(){return this._def.discriminator}get options(){return this._def.options}get optionsMap(){return this._def.optionsMap}static create(t,n,r){const i=new Map;for(const a of n){const u=km(a.shape[t]);if(!u.length)throw new Error(`A discriminator value for key \`${t}\` could not be extracted from all schema options`);for(const l of u){if(i.has(l))throw new Error(`Discriminator property ${String(t)} has duplicate value ${String(l)}`);i.set(l,a)}}return new rN({typeName:It.ZodDiscriminatedUnion,discriminator:t,options:n,optionsMap:i,...yr(r)})}}function c9(e,t){const n=yv(e),r=yv(t);if(e===t)return{valid:!0,data:e};if(n===cn.object&&r===cn.object){const i=mi.objectKeys(t),a=mi.objectKeys(e).filter(l=>i.indexOf(l)!==-1),u={...e,...t};for(const l of a){const o=c9(e[l],t[l]);if(!o.valid)return{valid:!1};u[l]=o.data}return{valid:!0,data:u}}else if(n===cn.array&&r===cn.array){if(e.length!==t.length)return{valid:!1};const i=[];for(let a=0;a<e.length;a++){const u=e[a],l=t[a],o=c9(u,l);if(!o.valid)return{valid:!1};i.push(o.data)}return{valid:!0,data:i}}else return n===cn.date&&r===cn.date&&+e==+t?{valid:!0,data:e}:{valid:!1}}class MD extends Yr{_parse(t){const{status:n,ctx:r}=this._processInputParams(t),i=(a,u)=>{if(VV(a)||VV(u))return Gn;const l=c9(a.value,u.value);return l.valid?((UV(a)||UV(u))&&n.dirty(),{status:n.value,value:l.data}):(tn(r,{code:Nt.invalid_intersection_types}),Gn)};return r.common.async?Promise.all([this._def.left._parseAsync({data:r.data,path:r.path,parent:r}),this._def.right._parseAsync({data:r.data,path:r.path,parent:r})]).then(([a,u])=>i(a,u)):i(this._def.left._parseSync({data:r.data,path:r.path,parent:r}),this._def.right._parseSync({data:r.data,path:r.path,parent:r}))}}MD.create=(e,t,n)=>new MD({left:e,right:t,typeName:It.ZodIntersection,...yr(n)});class e1 extends Yr{_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.parsedType!==cn.array)return tn(r,{code:Nt.invalid_type,expected:cn.array,received:r.parsedType}),Gn;if(r.data.length<this._def.items.length)return tn(r,{code:Nt.too_small,minimum:this._def.items.length,inclusive:!0,exact:!1,type:"array"}),Gn;!this._def.rest&&r.data.length>this._def.items.length&&(tn(r,{code:Nt.too_big,maximum:this._def.items.length,inclusive:!0,exact:!1,type:"array"}),n.dirty());const a=[...r.data].map((u,l)=>{const o=this._def.items[l]||this._def.rest;return o?o._parse(new A0(r,u,r.path,l)):null}).filter(u=>!!u);return r.common.async?Promise.all(a).then(u=>hc.mergeArray(n,u)):hc.mergeArray(n,a)}get items(){return this._def.items}rest(t){return new e1({...this._def,rest:t})}}e1.create=(e,t)=>{if(!Array.isArray(e))throw new Error("You must pass an array of schemas to z.tuple([ ... ])");return new e1({items:e,typeName:It.ZodTuple,rest:null,...yr(t)})};class FD extends Yr{get keySchema(){return this._def.keyType}get valueSchema(){return this._def.valueType}_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.parsedType!==cn.object)return tn(r,{code:Nt.invalid_type,expected:cn.object,received:r.parsedType}),Gn;const i=[],a=this._def.keyType,u=this._def.valueType;for(const l in r.data)i.push({key:a._parse(new A0(r,l,r.path,l)),value:u._parse(new A0(r,r.data[l],r.path,l)),alwaysSet:l in r.data});return r.common.async?hc.mergeObjectAsync(n,i):hc.mergeObjectSync(n,i)}get element(){return this._def.valueType}static create(t,n,r){return n instanceof Yr?new FD({keyType:t,valueType:n,typeName:It.ZodRecord,...yr(r)}):new FD({keyType:Km.create(),valueType:t,typeName:It.ZodRecord,...yr(n)})}}class YV extends Yr{get keySchema(){return this._def.keyType}get valueSchema(){return this._def.valueType}_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.parsedType!==cn.map)return tn(r,{code:Nt.invalid_type,expected:cn.map,received:r.parsedType}),Gn;const i=this._def.keyType,a=this._def.valueType,u=[...r.data.entries()].map(([l,o],s)=>({key:i._parse(new A0(r,l,r.path,[s,"key"])),value:a._parse(new A0(r,o,r.path,[s,"value"]))}));if(r.common.async){const l=new Map;return Promise.resolve().then(async()=>{for(const o of u){const s=await o.key,c=await o.value;if(s.status==="aborted"||c.status==="aborted")return Gn;(s.status==="dirty"||c.status==="dirty")&&n.dirty(),l.set(s.value,c.value)}return{status:n.value,value:l}})}else{const l=new Map;for(const o of u){const s=o.key,c=o.value;if(s.status==="aborted"||c.status==="aborted")return Gn;(s.status==="dirty"||c.status==="dirty")&&n.dirty(),l.set(s.value,c.value)}return{status:n.value,value:l}}}}YV.create=(e,t,n)=>new YV({valueType:t,keyType:e,typeName:It.ZodMap,...yr(n)});class px extends Yr{_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.parsedType!==cn.set)return tn(r,{code:Nt.invalid_type,expected:cn.set,received:r.parsedType}),Gn;const i=this._def;i.minSize!==null&&r.data.size<i.minSize.value&&(tn(r,{code:Nt.too_small,minimum:i.minSize.value,type:"set",inclusive:!0,exact:!1,message:i.minSize.message}),n.dirty()),i.maxSize!==null&&r.data.size>i.maxSize.value&&(tn(r,{code:Nt.too_big,maximum:i.maxSize.value,type:"set",inclusive:!0,exact:!1,message:i.maxSize.message}),n.dirty());const a=this._def.valueType;function u(o){const s=new Set;for(const c of o){if(c.status==="aborted")return Gn;c.status==="dirty"&&n.dirty(),s.add(c.value)}return{status:n.value,value:s}}const l=[...r.data.values()].map((o,s)=>a._parse(new A0(r,o,r.path,s)));return r.common.async?Promise.all(l).then(o=>u(o)):u(l)}min(t,n){return new px({...this._def,minSize:{value:t,message:bn.toString(n)}})}max(t,n){return new px({...this._def,maxSize:{value:t,message:bn.toString(n)}})}size(t,n){return this.min(t,n).max(t,n)}nonempty(t){return this.min(1,t)}}px.create=(e,t)=>new px({valueType:e,minSize:null,maxSize:null,typeName:It.ZodSet,...yr(t)});class ND extends Yr{get schema(){return this._def.getter()}_parse(t){const{ctx:n}=this._processInputParams(t);return this._def.getter()._parse({data:n.data,path:n.path,parent:n})}}ND.create=(e,t)=>new ND({getter:e,typeName:It.ZodLazy,...yr(t)});class ID extends Yr{_parse(t){if(t.data!==this._def.value){const n=this._getOrReturnCtx(t);return tn(n,{received:n.data,code:Nt.invalid_literal,expected:this._def.value}),Gn}return{status:"valid",value:t.data}}get value(){return this._def.value}}ID.create=(e,t)=>new ID({value:e,typeName:It.ZodLiteral,...yr(t)});function $ee(e,t){return new t1({values:e,typeName:It.ZodEnum,...yr(t)})}class t1 extends Yr{_parse(t){if(typeof t.data!="string"){const n=this._getOrReturnCtx(t),r=this._def.values;return tn(n,{expected:mi.joinValues(r),received:n.parsedType,code:Nt.invalid_type}),Gn}if(this._cache||(this._cache=new Set(this._def.values)),!this._cache.has(t.data)){const n=this._getOrReturnCtx(t),r=this._def.values;return tn(n,{received:n.data,code:Nt.invalid_enum_value,options:r}),Gn}return rh(t.data)}get options(){return this._def.values}get enum(){const t={};for(const n of this._def.values)t[n]=n;return t}get Values(){const t={};for(const n of this._def.values)t[n]=n;return t}get Enum(){const t={};for(const n of this._def.values)t[n]=n;return t}extract(t,n=this._def){return t1.create(t,{...this._def,...n})}exclude(t,n=this._def){return t1.create(this.options.filter(r=>!t.includes(r)),{...this._def,...n})}}t1.create=$ee;class f9 extends Yr{_parse(t){const n=mi.getValidEnumValues(this._def.values),r=this._getOrReturnCtx(t);if(r.parsedType!==cn.string&&r.parsedType!==cn.number){const i=mi.objectValues(n);return tn(r,{expected:mi.joinValues(i),received:r.parsedType,code:Nt.invalid_type}),Gn}if(this._cache||(this._cache=new Set(mi.getValidEnumValues(this._def.values))),!this._cache.has(t.data)){const i=mi.objectValues(n);return tn(r,{received:r.data,code:Nt.invalid_enum_value,options:i}),Gn}return rh(t.data)}get enum(){return this._def.values}}f9.create=(e,t)=>new f9({values:e,typeName:It.ZodNativeEnum,...yr(t)});class LD extends Yr{unwrap(){return this._def.type}_parse(t){const{ctx:n}=this._processInputParams(t);if(n.parsedType!==cn.promise&&n.common.async===!1)return tn(n,{code:Nt.invalid_type,expected:cn.promise,received:n.parsedType}),Gn;const r=n.parsedType===cn.promise?n.data:Promise.resolve(n.data);return rh(r.then(i=>this._def.type.parseAsync(i,{path:n.path,errorMap:n.common.contextualErrorMap})))}}LD.create=(e,t)=>new LD({type:e,typeName:It.ZodPromise,...yr(t)});class n1 extends Yr{innerType(){return this._def.schema}sourceType(){return this._def.schema._def.typeName===It.ZodEffects?this._def.schema.sourceType():this._def.schema}_parse(t){const{status:n,ctx:r}=this._processInputParams(t),i=this._def.effect||null,a={addIssue:u=>{tn(r,u),u.fatal?n.abort():n.dirty()},get path(){return r.path}};if(a.addIssue=a.addIssue.bind(a),i.type==="preprocess"){const u=i.transform(r.data,a);if(r.common.async)return Promise.resolve(u).then(async l=>{if(n.value==="aborted")return Gn;const o=await this._def.schema._parseAsync({data:l,path:r.path,parent:r});return o.status==="aborted"?Gn:o.status==="dirty"||n.value==="dirty"?A2(o.value):o});{if(n.value==="aborted")return Gn;const l=this._def.schema._parseSync({data:u,path:r.path,parent:r});return l.status==="aborted"?Gn:l.status==="dirty"||n.value==="dirty"?A2(l.value):l}}if(i.type==="refinement"){const u=l=>{const o=i.refinement(l,a);if(r.common.async)return Promise.resolve(o);if(o instanceof Promise)throw new Error("Async refinement encountered during synchronous parse operation. Use .parseAsync instead.");return l};if(r.common.async===!1){const l=this._def.schema._parseSync({data:r.data,path:r.path,parent:r});return l.status==="aborted"?Gn:(l.status==="dirty"&&n.dirty(),u(l.value),{status:n.value,value:l.value})}else return this._def.schema._parseAsync({data:r.data,path:r.path,parent:r}).then(l=>l.status==="aborted"?Gn:(l.status==="dirty"&&n.dirty(),u(l.value).then(()=>({status:n.value,value:l.value}))))}if(i.type==="transform")if(r.common.async===!1){const u=this._def.schema._parseSync({data:r.data,path:r.path,parent:r});if(!D3(u))return Gn;const l=i.transform(u.value,a);if(l instanceof Promise)throw new Error("Asynchronous transform encountered during synchronous parse operation. Use .parseAsync instead.");return{status:n.value,value:l}}else return this._def.schema._parseAsync({data:r.data,path:r.path,parent:r}).then(u=>D3(u)?Promise.resolve(i.transform(u.value,a)).then(l=>({status:n.value,value:l})):Gn);mi.assertNever(i)}}n1.create=(e,t,n)=>new n1({schema:e,typeName:It.ZodEffects,effect:t,...yr(n)});n1.createWithPreprocess=(e,t,n)=>new n1({schema:t,effect:{type:"preprocess",transform:e},typeName:It.ZodEffects,...yr(n)});class rg extends Yr{_parse(t){return this._getType(t)===cn.undefined?rh(void 0):this._def.innerType._parse(t)}unwrap(){return this._def.innerType}}rg.create=(e,t)=>new rg({innerType:e,typeName:It.ZodOptional,...yr(t)});class r1 extends Yr{_parse(t){return this._getType(t)===cn.null?rh(null):this._def.innerType._parse(t)}unwrap(){return this._def.innerType}}r1.create=(e,t)=>new r1({innerType:e,typeName:It.ZodNullable,...yr(t)});class BD extends Yr{_parse(t){const{ctx:n}=this._processInputParams(t);let r=n.data;return n.parsedType===cn.undefined&&(r=this._def.defaultValue()),this._def.innerType._parse({data:r,path:n.path,parent:n})}removeDefault(){return this._def.innerType}}BD.create=(e,t)=>new BD({innerType:e,typeName:It.ZodDefault,defaultValue:typeof t.default=="function"?t.default:()=>t.default,...yr(t)});class QD extends Yr{_parse(t){const{ctx:n}=this._processInputParams(t),r={...n,common:{...n.common,issues:[]}},i=this._def.innerType._parse({data:r.data,path:r.path,parent:{...r}});return AD(i)?i.then(a=>({status:"valid",value:a.status==="valid"?a.value:this._def.catchValue({get error(){return new og(r.common.issues)},input:r.data})})):{status:"valid",value:i.status==="valid"?i.value:this._def.catchValue({get error(){return new og(r.common.issues)},input:r.data})}}removeCatch(){return this._def.innerType}}QD.create=(e,t)=>new QD({innerType:e,typeName:It.ZodCatch,catchValue:typeof t.catch=="function"?t.catch:()=>t.catch,...yr(t)});class GV extends Yr{_parse(t){if(this._getType(t)!==cn.nan){const r=this._getOrReturnCtx(t);return tn(r,{code:Nt.invalid_type,expected:cn.nan,received:r.parsedType}),Gn}return{status:"valid",value:t.data}}}GV.create=e=>new GV({typeName:It.ZodNaN,...yr(e)});class Cee extends Yr{_parse(t){const{ctx:n}=this._processInputParams(t),r=n.data;return this._def.type._parse({data:r,path:n.path,parent:n})}unwrap(){return this._def.type}}class iN extends Yr{_parse(t){const{status:n,ctx:r}=this._processInputParams(t);if(r.common.async)return(async()=>{const a=await this._def.in._parseAsync({data:r.data,path:r.path,parent:r});return a.status==="aborted"?Gn:a.status==="dirty"?(n.dirty(),A2(a.value)):this._def.out._parseAsync({data:a.value,path:r.path,parent:r})})();{const i=this._def.in._parseSync({data:r.data,path:r.path,parent:r});return i.status==="aborted"?Gn:i.status==="dirty"?(n.dirty(),{status:"dirty",value:i.value}):this._def.out._parseSync({data:i.value,path:r.path,parent:r})}}static create(t,n){return new iN({in:t,out:n,typeName:It.ZodPipeline})}}class zD extends Yr{_parse(t){const n=this._def.innerType._parse(t),r=i=>(D3(i)&&(i.value=Object.freeze(i.value)),i);return AD(n)?n.then(i=>r(i)):r(n)}unwrap(){return this._def.innerType}}zD.create=(e,t)=>new zD({innerType:e,typeName:It.ZodReadonly,...yr(t)});var It;(function(e){e.ZodString="ZodString",e.ZodNumber="ZodNumber",e.ZodNaN="ZodNaN",e.ZodBigInt="ZodBigInt",e.ZodBoolean="ZodBoolean",e.ZodDate="ZodDate",e.ZodSymbol="ZodSymbol",e.ZodUndefined="ZodUndefined",e.ZodNull="ZodNull",e.ZodAny="ZodAny",e.ZodUnknown="ZodUnknown",e.ZodNever="ZodNever",e.ZodVoid="ZodVoid",e.ZodArray="ZodArray",e.ZodObject="ZodObject",e.ZodUnion="ZodUnion",e.ZodDiscriminatedUnion="ZodDiscriminatedUnion",e.ZodIntersection="ZodIntersection",e.ZodTuple="ZodTuple",e.ZodRecord="ZodRecord",e.ZodMap="ZodMap",e.ZodSet="ZodSet",e.ZodFunction="ZodFunction",e.ZodLazy="ZodLazy",e.ZodLiteral="ZodLiteral",e.ZodEnum="ZodEnum",e.ZodEffects="ZodEffects",e.ZodNativeEnum="ZodNativeEnum",e.ZodOptional="ZodOptional",e.ZodNullable="ZodNullable",e.ZodDefault="ZodDefault",e.ZodCatch="ZodCatch",e.ZodPromise="ZodPromise",e.ZodBranded="ZodBranded",e.ZodPipeline="ZodPipeline",e.ZodReadonly="ZodReadonly"})(It||(It={}));const crt=Km.create,frt=$3.create,drt=l9.create,hrt=kD.create,prt=s9.create;Yv.create;const mrt=D0.create,grt=gl.create,vrt=PD.create,yrt=rN.create;MD.create;e1.create;const brt=FD.create,Ort=ND.create,xrt=ID.create,Srt=t1.create;LD.create;rg.create;r1.create;const _rt=Gn,wee=1024;let axe=0,NT=class{constructor(t,n){this.from=t,this.to=n}},Ea=class{constructor(t={}){this.id=axe++,this.perNode=!!t.perNode,this.deserialize=t.deserialize||(()=>{throw new Error("This node type doesn't define a deserialize function")})}add(t){if(this.perNode)throw new RangeError("Can't add per-node props to node types");return typeof t!="function"&&(t=sg.match(t)),n=>{let r=t(n);return r===void 0?null:[this,r]}}};Ea.closedBy=new Ea({deserialize:e=>e.split(" ")});Ea.openedBy=new Ea({deserialize:e=>e.split(" ")});Ea.group=new Ea({deserialize:e=>e.split(" ")});Ea.isolate=new Ea({deserialize:e=>{if(e&&e!="rtl"&&e!="ltr"&&e!="auto")throw new RangeError("Invalid value for isolate: "+e);return e||"auto"}});Ea.contextHash=new Ea({perNode:!0});Ea.lookAhead=new Ea({perNode:!0});Ea.mounted=new Ea({perNode:!0});let qD=class{constructor(t,n,r){this.tree=t,this.overlay=n,this.parser=r}static get(t){return t&&t.props&&t.props[Ea.mounted.id]}};const uxe=Object.create(null);let sg=class Ree{constructor(t,n,r,i=0){this.name=t,this.props=n,this.id=r,this.flags=i}static define(t){let n=t.props&&t.props.length?Object.create(null):uxe,r=(t.top?1:0)|(t.skipped?2:0)|(t.error?4:0)|(t.name==null?8:0),i=new Ree(t.name||"",n,t.id,r);if(t.props){for(let a of t.props)if(Array.isArray(a)||(a=a(i)),a){if(a[0].perNode)throw new RangeError("Can't store a per-node prop on a node type");n[a[0].id]=a[1]}}return i}prop(t){return this.props[t.id]}get isTop(){return(this.flags&1)>0}get isSkipped(){return(this.flags&2)>0}get isError(){return(this.flags&4)>0}get isAnonymous(){return(this.flags&8)>0}is(t){if(typeof t=="string"){if(this.name==t)return!0;let n=this.prop(Ea.group);return n?n.indexOf(t)>-1:!1}return this.id==t}static match(t){let n=Object.create(null);for(let r in t)for(let i of r.split(" "))n[i]=t[r];return r=>{for(let i=r.prop(Ea.group),a=-1;a<(i?i.length:0);a++){let u=n[a<0?r.name:i[a]];if(u)return u}}}};sg.none=new sg("",Object.create(null),0,8);class aN{constructor(t){this.types=t;for(let n=0;n<t.length;n++)if(t[n].id!=n)throw new RangeError("Node type ids should correspond to array positions when creating a node set")}extend(...t){let n=[];for(let r of this.types){let i=null;for(let a of t){let u=a(r);u&&(i||(i=Object.assign({},r.props)),i[u[0].id]=u[1])}n.push(i?new sg(r.name,i,r.id,r.flags):r)}return new aN(n)}}const N_=new WeakMap,KV=new WeakMap;var Lo;(function(e){e[e.ExcludeBuffers=1]="ExcludeBuffers",e[e.IncludeAnonymous=2]="IncludeAnonymous",e[e.IgnoreMounts=4]="IgnoreMounts",e[e.IgnoreOverlays=8]="IgnoreOverlays"})(Lo||(Lo={}));let ep=class d9{constructor(t,n,r,i,a){if(this.type=t,this.children=n,this.positions=r,this.length=i,this.props=null,a&&a.length){this.props=Object.create(null);for(let[u,l]of a)this.props[typeof u=="number"?u:u.id]=l}}toString(){let t=qD.get(this);if(t&&!t.overlay)return t.tree.toString();let n="";for(let r of this.children){let i=r.toString();i&&(n&&(n+=","),n+=i)}return this.type.name?(/\W/.test(this.type.name)&&!this.type.isError?JSON.stringify(this.type.name):this.type.name)+(n.length?"("+n+")":""):n}cursor(t=0){return new p9(this.topNode,t)}cursorAt(t,n=0,r=0){let i=N_.get(this)||this.topNode,a=new p9(i);return a.moveTo(t,n),N_.set(this,a._tree),a}get topNode(){return new T0(this,0,0,null)}resolve(t,n=0){let r=mx(N_.get(this)||this.topNode,t,n,!1);return N_.set(this,r),r}resolveInner(t,n=0){let r=mx(KV.get(this)||this.topNode,t,n,!0);return KV.set(this,r),r}resolveStack(t,n=0){return cxe(this,t,n)}iterate(t){let{enter:n,leave:r,from:i=0,to:a=this.length}=t,u=t.mode||0,l=(u&Lo.IncludeAnonymous)>0;for(let o=this.cursor(u|Lo.IncludeAnonymous);;){let s=!1;if(o.from<=a&&o.to>=i&&(!l&&o.type.isAnonymous||n(o)!==!1)){if(o.firstChild())continue;s=!0}for(;s&&r&&(l||!o.type.isAnonymous)&&r(o),!o.nextSibling();){if(!o.parent())return;s=!0}}}prop(t){return t.perNode?this.props?this.props[t.id]:void 0:this.type.prop(t)}get propValues(){let t=[];if(this.props)for(let n in this.props)t.push([+n,this.props[n]]);return t}balance(t={}){return this.children.length<=8?this:lN(sg.none,this.children,this.positions,0,this.children.length,0,this.length,(n,r,i)=>new d9(this.type,n,r,i,this.propValues),t.makeTree||((n,r,i)=>new d9(sg.none,n,r,i)))}static build(t){return fxe(t)}};ep.empty=new ep(sg.none,[],[],0);let lxe=class Aee{constructor(t,n){this.buffer=t,this.index=n}get id(){return this.buffer[this.index-4]}get start(){return this.buffer[this.index-3]}get end(){return this.buffer[this.index-2]}get size(){return this.buffer[this.index-1]}get pos(){return this.index}next(){this.index-=4}fork(){return new Aee(this.buffer,this.index)}},C3=class Tee{constructor(t,n,r){this.buffer=t,this.length=n,this.set=r}get type(){return sg.none}toString(){let t=[];for(let n=0;n<this.buffer.length;)t.push(this.childString(n)),n=this.buffer[n+3];return t.join(",")}childString(t){let n=this.buffer[t],r=this.buffer[t+3],i=this.set.types[n],a=i.name;if(/\W/.test(a)&&!i.isError&&(a=JSON.stringify(a)),t+=4,r==t)return a;let u=[];for(;t<r;)u.push(this.childString(t)),t=this.buffer[t+3];return a+"("+u.join(",")+")"}findChild(t,n,r,i,a){let{buffer:u}=this,l=-1;for(let o=t;o!=n&&!(kee(a,i,u[o+1],u[o+2])&&(l=o,r>0));o=u[o+3]);return l}slice(t,n,r){let i=this.buffer,a=new Uint16Array(n-t),u=0;for(let l=t,o=0;l<n;){a[o++]=i[l++],a[o++]=i[l++]-r;let s=a[o++]=i[l++]-r;a[o++]=i[l++]-t,u=Math.max(u,s)}return new Tee(a,u,this.set)}};function kee(e,t,n,r){switch(e){case-2:return n<t;case-1:return r>=t&&n<t;case 0:return n<t&&r>t;case 1:return n<=t&&r>t;case 2:return r>t;case 4:return!0}}function mx(e,t,n,r){for(var i;e.from==e.to||(n<1?e.from>=t:e.from>t)||(n>-1?e.to<=t:e.to<t);){let u=!r&&e instanceof T0&&e.index<0?null:e.parent;if(!u)return e;e=u}let a=r?0:Lo.IgnoreOverlays;if(r)for(let u=e,l=u.parent;l;u=l,l=u.parent)u instanceof T0&&u.index<0&&((i=l.enter(t,n,a))===null||i===void 0?void 0:i.from)!=u.from&&(e=l);for(;;){let u=e.enter(t,n,a);if(!u)return e;e=u}}let Pee=class{cursor(t=0){return new p9(this,t)}getChild(t,n=null,r=null){let i=WV(this,t,n,r);return i.length?i[0]:null}getChildren(t,n=null,r=null){return WV(this,t,n,r)}resolve(t,n=0){return mx(this,t,n,!1)}resolveInner(t,n=0){return mx(this,t,n,!0)}matchContext(t){return h9(this.parent,t)}enterUnfinishedNodesBefore(t){let n=this.childBefore(t),r=this;for(;n;){let i=n.lastChild;if(!i||i.to!=n.to)break;i.type.isError&&i.from==i.to?(r=n,n=i.prevSibling):n=i}return r}get node(){return this}get next(){return this.parent}},T0=class GE extends Pee{constructor(t,n,r,i){super(),this._tree=t,this.from=n,this.index=r,this._parent=i}get type(){return this._tree.type}get name(){return this._tree.type.name}get to(){return this.from+this._tree.length}nextChild(t,n,r,i,a=0){for(let u=this;;){for(let{children:l,positions:o}=u._tree,s=n>0?l.length:-1;t!=s;t+=n){let c=l[t],d=o[t]+u.from;if(kee(i,r,d,d+c.length)){if(c instanceof C3){if(a&Lo.ExcludeBuffers)continue;let h=c.findChild(0,c.buffer.length,n,r-d,i);if(h>-1)return new gx(new oxe(u,c,t,d),null,h)}else if(a&Lo.IncludeAnonymous||!c.type.isAnonymous||uN(c)){let h;if(!(a&Lo.IgnoreMounts)&&(h=qD.get(c))&&!h.overlay)return new GE(h.tree,d,t,u);let m=new GE(c,d,t,u);return a&Lo.IncludeAnonymous||!m.type.isAnonymous?m:m.nextChild(n<0?c.children.length-1:0,n,r,i)}}}if(a&Lo.IncludeAnonymous||!u.type.isAnonymous||(u.index>=0?t=u.index+n:t=n<0?-1:u._parent._tree.children.length,u=u._parent,!u))return null}}get firstChild(){return this.nextChild(0,1,0,4)}get lastChild(){return this.nextChild(this._tree.children.length-1,-1,0,4)}childAfter(t){return this.nextChild(0,1,t,2)}childBefore(t){return this.nextChild(this._tree.children.length-1,-1,t,-2)}enter(t,n,r=0){let i;if(!(r&Lo.IgnoreOverlays)&&(i=qD.get(this._tree))&&i.overlay){let a=t-this.from;for(let{from:u,to:l}of i.overlay)if((n>0?u<=a:u<a)&&(n<0?l>=a:l>a))return new GE(i.tree,i.overlay[0].from+this.from,-1,this)}return this.nextChild(0,1,t,n,r)}nextSignificantParent(){let t=this;for(;t.type.isAnonymous&&t._parent;)t=t._parent;return t}get parent(){return this._parent?this._parent.nextSignificantParent():null}get nextSibling(){return this._parent&&this.index>=0?this._parent.nextChild(this.index+1,1,0,4):null}get prevSibling(){return this._parent&&this.index>=0?this._parent.nextChild(this.index-1,-1,0,4):null}get tree(){return this._tree}toTree(){return this._tree}toString(){return this._tree.toString()}};function WV(e,t,n,r){let i=e.cursor(),a=[];if(!i.firstChild())return a;if(n!=null){for(let u=!1;!u;)if(u=i.type.is(n),!i.nextSibling())return a}for(;;){if(r!=null&&i.type.is(r))return a;if(i.type.is(t)&&a.push(i.node),!i.nextSibling())return r==null?a:[]}}function h9(e,t,n=t.length-1){for(let r=e;n>=0;r=r.parent){if(!r)return!1;if(!r.type.isAnonymous){if(t[n]&&t[n]!=r.name)return!1;n--}}return!0}let oxe=class{constructor(t,n,r,i){this.parent=t,this.buffer=n,this.index=r,this.start=i}},gx=class T2 extends Pee{get name(){return this.type.name}get from(){return this.context.start+this.context.buffer.buffer[this.index+1]}get to(){return this.context.start+this.context.buffer.buffer[this.index+2]}constructor(t,n,r){super(),this.context=t,this._parent=n,this.index=r,this.type=t.buffer.set.types[t.buffer.buffer[r]]}child(t,n,r){let{buffer:i}=this.context,a=i.findChild(this.index+4,i.buffer[this.index+3],t,n-this.context.start,r);return a<0?null:new T2(this.context,this,a)}get firstChild(){return this.child(1,0,4)}get lastChild(){return this.child(-1,0,4)}childAfter(t){return this.child(1,t,2)}childBefore(t){return this.child(-1,t,-2)}enter(t,n,r=0){if(r&Lo.ExcludeBuffers)return null;let{buffer:i}=this.context,a=i.findChild(this.index+4,i.buffer[this.index+3],n>0?1:-1,t-this.context.start,n);return a<0?null:new T2(this.context,this,a)}get parent(){return this._parent||this.context.parent.nextSignificantParent()}externalSibling(t){return this._parent?null:this.context.parent.nextChild(this.context.index+t,t,0,4)}get nextSibling(){let{buffer:t}=this.context,n=t.buffer[this.index+3];return n<(this._parent?t.buffer[this._parent.index+3]:t.buffer.length)?new T2(this.context,this._parent,n):this.externalSibling(1)}get prevSibling(){let{buffer:t}=this.context,n=this._parent?this._parent.index+4:0;return this.index==n?this.externalSibling(-1):new T2(this.context,this._parent,t.findChild(n,this.index,-1,0,4))}get tree(){return null}toTree(){let t=[],n=[],{buffer:r}=this.context,i=this.index+4,a=r.buffer[this.index+3];if(a>i){let u=r.buffer[this.index+1];t.push(r.slice(i,a,u)),n.push(0)}return new ep(this.type,t,n,this.to-this.from)}toString(){return this.context.buffer.childString(this.index)}};function Mee(e){if(!e.length)return null;let t=0,n=e[0];for(let a=1;a<e.length;a++){let u=e[a];(u.from>n.from||u.to<n.to)&&(n=u,t=a)}let r=n instanceof T0&&n.index<0?null:n.parent,i=e.slice();return r?i[t]=r:i.splice(t,1),new sxe(i,n)}let sxe=class{constructor(t,n){this.heads=t,this.node=n}get next(){return Mee(this.heads)}};function cxe(e,t,n){let r=e.resolveInner(t,n),i=null;for(let a=r instanceof T0?r:r.context.parent;a;a=a.parent)if(a.index<0){let u=a.parent;(i||(i=[r])).push(u.resolve(t,n)),a=u}else{let u=qD.get(a.tree);if(u&&u.overlay&&u.overlay[0].from<=t&&u.overlay[u.overlay.length-1].to>=t){let l=new T0(u.tree,u.overlay[0].from+a.from,-1,a);(i||(i=[r])).push(mx(l,t,n,!1))}}return i?Mee(i):r}let p9=class{get name(){return this.type.name}constructor(t,n=0){if(this.mode=n,this.buffer=null,this.stack=[],this.index=0,this.bufferNode=null,t instanceof T0)this.yieldNode(t);else{this._tree=t.context.parent,this.buffer=t.context;for(let r=t._parent;r;r=r._parent)this.stack.unshift(r.index);this.bufferNode=t,this.yieldBuf(t.index)}}yieldNode(t){return t?(this._tree=t,this.type=t.type,this.from=t.from,this.to=t.to,!0):!1}yieldBuf(t,n){this.index=t;let{start:r,buffer:i}=this.buffer;return this.type=n||i.set.types[i.buffer[t]],this.from=r+i.buffer[t+1],this.to=r+i.buffer[t+2],!0}yield(t){return t?t instanceof T0?(this.buffer=null,this.yieldNode(t)):(this.buffer=t.context,this.yieldBuf(t.index,t.type)):!1}toString(){return this.buffer?this.buffer.buffer.childString(this.index):this._tree.toString()}enterChild(t,n,r){if(!this.buffer)return this.yield(this._tree.nextChild(t<0?this._tree._tree.children.length-1:0,t,n,r,this.mode));let{buffer:i}=this.buffer,a=i.findChild(this.index+4,i.buffer[this.index+3],t,n-this.buffer.start,r);return a<0?!1:(this.stack.push(this.index),this.yieldBuf(a))}firstChild(){return this.enterChild(1,0,4)}lastChild(){return this.enterChild(-1,0,4)}childAfter(t){return this.enterChild(1,t,2)}childBefore(t){return this.enterChild(-1,t,-2)}enter(t,n,r=this.mode){return this.buffer?r&Lo.ExcludeBuffers?!1:this.enterChild(1,t,n):this.yield(this._tree.enter(t,n,r))}parent(){if(!this.buffer)return this.yieldNode(this.mode&Lo.IncludeAnonymous?this._tree._parent:this._tree.parent);if(this.stack.length)return this.yieldBuf(this.stack.pop());let t=this.mode&Lo.IncludeAnonymous?this.buffer.parent:this.buffer.parent.nextSignificantParent();return this.buffer=null,this.yieldNode(t)}sibling(t){if(!this.buffer)return this._tree._parent?this.yield(this._tree.index<0?null:this._tree._parent.nextChild(this._tree.index+t,t,0,4,this.mode)):!1;let{buffer:n}=this.buffer,r=this.stack.length-1;if(t<0){let i=r<0?0:this.stack[r]+4;if(this.index!=i)return this.yieldBuf(n.findChild(i,this.index,-1,0,4))}else{let i=n.buffer[this.index+3];if(i<(r<0?n.buffer.length:n.buffer[this.stack[r]+3]))return this.yieldBuf(i)}return r<0?this.yield(this.buffer.parent.nextChild(this.buffer.index+t,t,0,4,this.mode)):!1}nextSibling(){return this.sibling(1)}prevSibling(){return this.sibling(-1)}atLastNode(t){let n,r,{buffer:i}=this;if(i){if(t>0){if(this.index<i.buffer.buffer.length)return!1}else for(let a=0;a<this.index;a++)if(i.buffer.buffer[a+3]<this.index)return!1;({index:n,parent:r}=i)}else({index:n,_parent:r}=this._tree);for(;r;{index:n,_parent:r}=r)if(n>-1)for(let a=n+t,u=t<0?-1:r._tree.children.length;a!=u;a+=t){let l=r._tree.children[a];if(this.mode&Lo.IncludeAnonymous||l instanceof C3||!l.type.isAnonymous||uN(l))return!1}return!0}move(t,n){if(n&&this.enterChild(t,0,4))return!0;for(;;){if(this.sibling(t))return!0;if(this.atLastNode(t)||!this.parent())return!1}}next(t=!0){return this.move(1,t)}prev(t=!0){return this.move(-1,t)}moveTo(t,n=0){for(;(this.from==this.to||(n<1?this.from>=t:this.from>t)||(n>-1?this.to<=t:this.to<t))&&this.parent(););for(;this.enterChild(1,t,n););return this}get node(){if(!this.buffer)return this._tree;let t=this.bufferNode,n=null,r=0;if(t&&t.context==this.buffer)e:for(let i=this.index,a=this.stack.length;a>=0;){for(let u=t;u;u=u._parent)if(u.index==i){if(i==this.index)return u;n=u,r=a+1;break e}i=this.stack[--a]}for(let i=r;i<this.stack.length;i++)n=new gx(this.buffer,n,this.stack[i]);return this.bufferNode=new gx(this.buffer,n,this.index)}get tree(){return this.buffer?null:this._tree._tree}iterate(t,n){for(let r=0;;){let i=!1;if(this.type.isAnonymous||t(this)!==!1){if(this.firstChild()){r++;continue}this.type.isAnonymous||(i=!0)}for(;;){if(i&&n&&n(this),i=this.type.isAnonymous,!r)return;if(this.nextSibling())break;this.parent(),r--,i=!0}}}matchContext(t){if(!this.buffer)return h9(this.node.parent,t);let{buffer:n}=this.buffer,{types:r}=n.set;for(let i=t.length-1,a=this.stack.length-1;i>=0;a--){if(a<0)return h9(this._tree,t,i);let u=r[n.buffer[this.stack[a]]];if(!u.isAnonymous){if(t[i]&&t[i]!=u.name)return!1;i--}}return!0}};function uN(e){return e.children.some(t=>t instanceof C3||!t.type.isAnonymous||uN(t))}function fxe(e){var t;let{buffer:n,nodeSet:r,maxBufferLength:i=wee,reused:a=[],minRepeatType:u=r.types.length}=e,l=Array.isArray(n)?new lxe(n,n.length):n,o=r.types,s=0,c=0;function d(A,w,R,F,I,U){let{id:q,start:V,end:N,size:j}=l,W=c,K=s;for(;j<0;)if(l.next(),j==-1){let G=a[q];R.push(G),F.push(V-A);return}else if(j==-3){s=q;return}else if(j==-4){c=q;return}else throw new RangeError(`Unrecognized record size: ${j}`);let J=o[q],te,H,M=V-A;if(N-V<=i&&(H=S(l.pos-w,I))){let G=new Uint16Array(H.size-H.skip),X=l.pos-H.size,B=G.length;for(;l.pos>X;)B=_(H.start,G,B);te=new C3(G,N-H.start,r),M=H.start-A}else{let G=l.pos-j;l.next();let X=[],B=[],ne=q>=u?q:-1,oe=0,ie=N;for(;l.pos>G;)ne>=0&&l.id==ne&&l.size>=0?(l.end<=ie-i&&(y(X,B,V,oe,l.end,ie,ne,W,K),oe=X.length,ie=l.end),l.next()):U>2500?h(V,G,X,B):d(V,G,X,B,ne,U+1);if(ne>=0&&oe>0&&oe<X.length&&y(X,B,V,oe,V,ie,ne,W,K),X.reverse(),B.reverse(),ne>-1&&oe>0){let de=m(J,K);te=lN(J,X,B,0,X.length,0,N-V,de,de)}else te=v(J,X,B,N-V,W-N,K)}R.push(te),F.push(M)}function h(A,w,R,F){let I=[],U=0,q=-1;for(;l.pos>w;){let{id:V,start:N,end:j,size:W}=l;if(W>4)l.next();else{if(q>-1&&N<q)break;q<0&&(q=j-i),I.push(V,N,j),U++,l.next()}}if(U){let V=new Uint16Array(U*4),N=I[I.length-2];for(let j=I.length-3,W=0;j>=0;j-=3)V[W++]=I[j],V[W++]=I[j+1]-N,V[W++]=I[j+2]-N,V[W++]=W;R.push(new C3(V,I[2]-N,r)),F.push(N-A)}}function m(A,w){return(R,F,I)=>{let U=0,q=R.length-1,V,N;if(q>=0&&(V=R[q])instanceof ep){if(!q&&V.type==A&&V.length==I)return V;(N=V.prop(Ea.lookAhead))&&(U=F[q]+V.length+N)}return v(A,R,F,I,U,w)}}function y(A,w,R,F,I,U,q,V,N){let j=[],W=[];for(;A.length>F;)j.push(A.pop()),W.push(w.pop()+R-I);A.push(v(r.types[q],j,W,U-I,V-U,N)),w.push(I-R)}function v(A,w,R,F,I,U,q){if(U){let V=[Ea.contextHash,U];q=q?[V].concat(q):[V]}if(I>25){let V=[Ea.lookAhead,I];q=q?[V].concat(q):[V]}return new ep(A,w,R,F,q)}function S(A,w){let R=l.fork(),F=0,I=0,U=0,q=R.end-i,V={size:0,start:0,skip:0};e:for(let N=R.pos-A;R.pos>N;){let j=R.size;if(R.id==w&&j>=0){V.size=F,V.start=I,V.skip=U,U+=4,F+=4,R.next();continue}let W=R.pos-j;if(j<0||W<N||R.start<q)break;let K=R.id>=u?4:0,J=R.start;for(R.next();R.pos>W;){if(R.size<0)if(R.size==-3)K+=4;else break e;else R.id>=u&&(K+=4);R.next()}I=J,F+=j,U+=K}return(w<0||F==A)&&(V.size=F,V.start=I,V.skip=U),V.size>4?V:void 0}function _(A,w,R){let{id:F,start:I,end:U,size:q}=l;if(l.next(),q>=0&&F<u){let V=R;if(q>4){let N=l.pos-(q-4);for(;l.pos>N;)R=_(A,w,R)}w[--R]=V,w[--R]=U-A,w[--R]=I-A,w[--R]=F}else q==-3?s=F:q==-4&&(c=F);return R}let E=[],$=[];for(;l.pos>0;)d(e.start||0,e.bufferStart||0,E,$,-1,0);let k=(t=e.length)!==null&&t!==void 0?t:E.length?$[0]+E[0].length:0;return new ep(o[e.topID],E.reverse(),$.reverse(),k)}const JV=new WeakMap;function KE(e,t){if(!e.isAnonymous||t instanceof C3||t.type!=e)return 1;let n=JV.get(t);if(n==null){n=1;for(let r of t.children){if(r.type!=e||!(r instanceof ep)){n=1;break}n+=KE(e,r)}JV.set(t,n)}return n}function lN(e,t,n,r,i,a,u,l,o){let s=0;for(let y=r;y<i;y++)s+=KE(e,t[y]);let c=Math.ceil(s*1.5/8),d=[],h=[];function m(y,v,S,_,E){for(let $=S;$<_;){let k=$,A=v[$],w=KE(e,y[$]);for($++;$<_;$++){let R=KE(e,y[$]);if(w+R>=c)break;w+=R}if($==k+1){if(w>c){let R=y[k];m(R.children,R.positions,0,R.children.length,v[k]+E);continue}d.push(y[k])}else{let R=v[$-1]+y[$-1].length-A;d.push(lN(e,y,v,k,$,A,R,null,o))}h.push(A+E-a)}}return m(t,n,r,i,0),(l||o)(d,h,u)}class Trt{constructor(){this.map=new WeakMap}setBuffer(t,n,r){let i=this.map.get(t);i||this.map.set(t,i=new Map),i.set(n,r)}getBuffer(t,n){let r=this.map.get(t);return r&&r.get(n)}set(t,n){t instanceof gx?this.setBuffer(t.context.buffer,t.index,n):t instanceof T0&&this.map.set(t.tree,n)}get(t){return t instanceof gx?this.getBuffer(t.context.buffer,t.index):t instanceof T0?this.map.get(t.tree):void 0}cursorSet(t,n){t.buffer?this.setBuffer(t.buffer.buffer,t.index,n):this.map.set(t.tree,n)}cursorGet(t){return t.buffer?this.getBuffer(t.buffer.buffer,t.index):this.map.get(t.tree)}}let krt=class m9{constructor(t,n,r,i,a=!1,u=!1){this.from=t,this.to=n,this.tree=r,this.offset=i,this.open=(a?1:0)|(u?2:0)}get openStart(){return(this.open&1)>0}get openEnd(){return(this.open&2)>0}static addTree(t,n=[],r=!1){let i=[new m9(0,t.length,t,0,!1,r)];for(let a of n)a.to>t.length&&i.push(a);return i}static applyChanges(t,n,r=128){if(!n.length)return t;let i=[],a=1,u=t.length?t[0]:null;for(let l=0,o=0,s=0;;l++){let c=l<n.length?n[l]:null,d=c?c.fromA:1e9;if(d-o>=r)for(;u&&u.from<d;){let h=u;if(o>=h.from||d<=h.to||s){let m=Math.max(h.from,o)-s,y=Math.min(h.to,d)-s;h=m>=y?null:new m9(m,y,h.tree,h.offset+s,l>0,!!c)}if(h&&i.push(h),u.to>d)break;u=a<t.length?t[a++]:null}if(!c)break;o=c.toA,s=c.toA-c.toB}return i}},dxe=class{startParse(t,n,r){return typeof t=="string"&&(t=new hxe(t)),r=r?r.length?r.map(i=>new NT(i.from,i.to)):[new NT(0,0)]:[new NT(0,t.length)],this.createParse(t,n||[],r)}parse(t,n,r){let i=this.startParse(t,n,r);for(;;){let a=i.advance();if(a)return a}}},hxe=class{constructor(t){this.string=t}get length(){return this.string.length}chunk(t){return this.string.slice(t)}get lineChunks(){return!1}read(t,n){return this.string.slice(t,n)}};new Ea({perNode:!0});var I_={},eU;function Frt(){if(eU)return I_;eU=1;let e=[],t=[];(()=>{let h="lc,34,7n,7,7b,19,,,,2,,2,,,20,b,1c,l,g,,2t,7,2,6,2,2,,4,z,,u,r,2j,b,1m,9,9,,o,4,,9,,3,,5,17,3,3b,f,,w,1j,,,,4,8,4,,3,7,a,2,t,,1m,,,,2,4,8,,9,,a,2,q,,2,2,1l,,4,2,4,2,2,3,3,,u,2,3,,b,2,1l,,4,5,,2,4,,k,2,m,6,,,1m,,,2,,4,8,,7,3,a,2,u,,1n,,,,c,,9,,14,,3,,1l,3,5,3,,4,7,2,b,2,t,,1m,,2,,2,,3,,5,2,7,2,b,2,s,2,1l,2,,,2,4,8,,9,,a,2,t,,20,,4,,2,3,,,8,,29,,2,7,c,8,2q,,2,9,b,6,22,2,r,,,,,,1j,e,,5,,2,5,b,,10,9,,2u,4,,6,,2,2,2,p,2,4,3,g,4,d,,2,2,6,,f,,jj,3,qa,3,t,3,t,2,u,2,1s,2,,7,8,,2,b,9,,19,3,3b,2,y,,3a,3,4,2,9,,6,3,63,2,2,,1m,,,7,,,,,2,8,6,a,2,,1c,h,1r,4,1c,7,,,5,,14,9,c,2,w,4,2,2,,3,1k,,,2,3,,,3,1m,8,2,2,48,3,,d,,7,4,,6,,3,2,5i,1m,,5,ek,,5f,x,2da,3,3x,,2o,w,fe,6,2x,2,n9w,4,,a,w,2,28,2,7k,,3,,4,,p,2,5,,47,2,q,i,d,,12,8,p,b,1a,3,1c,,2,4,2,2,13,,1v,6,2,2,2,2,c,,8,,1b,,1f,,,3,2,2,5,2,,,16,2,8,,6m,,2,,4,,fn4,,kh,g,g,g,a6,2,gt,,6a,,45,5,1ae,3,,2,5,4,14,3,4,,4l,2,fx,4,ar,2,49,b,4w,,1i,f,1k,3,1d,4,2,2,1x,3,10,5,,8,1q,,c,2,1g,9,a,4,2,,2n,3,2,,,2,6,,4g,,3,8,l,2,1l,2,,,,,m,,e,7,3,5,5f,8,2,3,,,n,,29,,2,6,,,2,,,2,,2,6j,,2,4,6,2,,2,r,2,2d,8,2,,,2,2y,,,,2,6,,,2t,3,2,4,,5,77,9,,2,6t,,a,2,,,4,,40,4,2,2,4,,w,a,14,6,2,4,8,,9,6,2,3,1a,d,,2,ba,7,,6,,,2a,m,2,7,,2,,2,3e,6,3,,,2,,7,,,20,2,3,,,,9n,2,f0b,5,1n,7,t4,,1r,4,29,,f5k,2,43q,,,3,4,5,8,8,2,7,u,4,44,3,1iz,1j,4,1e,8,,e,,m,5,,f,11s,7,,h,2,7,,2,,5,79,7,c5,4,15s,7,31,7,240,5,gx7k,2o,3k,6o".split(",").map(m=>m?parseInt(m,36):1);for(let m=0,y=0;m<h.length;m++)(m%2?t:e).push(y=y+h[m])})();function n(h){if(h<768)return!1;for(let m=0,y=e.length;;){let v=m+y>>1;if(h<e[v])y=v;else if(h>=t[v])m=v+1;else return!0;if(m==y)return!1}}function r(h){return h>=127462&&h<=127487}const i=8205;function a(h,m,y=!0,v=!0){return(y?u:l)(h,m,v)}function u(h,m,y){if(m==h.length)return m;m&&s(h.charCodeAt(m))&&c(h.charCodeAt(m-1))&&m--;let v=o(h,m);for(m+=d(v);m<h.length;){let S=o(h,m);if(v==i||S==i||y&&n(S))m+=d(S),v=S;else if(r(S)){let _=0,E=m-2;for(;E>=0&&r(o(h,E));)_++,E-=2;if(_%2==0)break;m+=2}else break}return m}function l(h,m,y){for(;m>0;){let v=u(h,m-2,y);if(v<m)return v;m--}return 0}function o(h,m){let y=h.charCodeAt(m);if(!c(y)||m+1==h.length)return y;let v=h.charCodeAt(m+1);return s(v)?(y-55296<<10)+(v-56320)+65536:y}function s(h){return h>=56320&&h<57344}function c(h){return h>=55296&&h<56320}function d(h){return h<65536?1:2}return I_.findClusterBreak=a,I_.isExtendingChar=n,I_}const g9="ͼ",tU=typeof Symbol>"u"?"__"+g9:Symbol.for(g9),v9=typeof Symbol>"u"?"__styleSet"+Math.floor(Math.random()*1e8):Symbol("styleSet"),nU=typeof globalThis<"u"?globalThis:typeof window<"u"?window:{};let Nrt=class{constructor(t,n){this.rules=[];let{finish:r}=n||{};function i(u){return/^@/.test(u)?[u]:u.split(/,\s*/)}function a(u,l,o,s){let c=[],d=/^@(\w+)\b/.exec(u[0]),h=d&&d[1]=="keyframes";if(d&&l==null)return o.push(u[0]+";");for(let m in l){let y=l[m];if(/&/.test(m))a(m.split(/,\s*/).map(v=>u.map(S=>v.replace(/&/,S))).reduce((v,S)=>v.concat(S)),y,o);else if(y&&typeof y=="object"){if(!d)throw new RangeError("The value of a property ("+m+") should be a primitive value.");a(i(m),y,c,h)}else y!=null&&c.push(m.replace(/_.*/,"").replace(/[A-Z]/g,v=>"-"+v.toLowerCase())+": "+y+";")}(c.length||h)&&o.push((r&&!d&&!s?u.map(r):u).join(", ")+" {"+c.join(" ")+"}")}for(let u in t)a(i(u),t[u],this.rules)}getRules(){return this.rules.join(`
`)}static newName(){let t=nU[tU]||1;return nU[tU]=t+1,g9+t.toString(36)}static mount(t,n,r){let i=t[v9],a=r&&r.nonce;i?a&&i.setNonce(a):i=new pxe(t,a),i.mount(Array.isArray(n)?n:[n],t)}},rU=new Map,pxe=class{constructor(t,n){let r=t.ownerDocument||t,i=r.defaultView;if(!t.head&&t.adoptedStyleSheets&&i.CSSStyleSheet){let a=rU.get(r);if(a)return t[v9]=a;this.sheet=new i.CSSStyleSheet,rU.set(r,this)}else this.styleTag=r.createElement("style"),n&&this.styleTag.setAttribute("nonce",n);this.modules=[],t[v9]=this}mount(t,n){let r=this.sheet,i=0,a=0;for(let u=0;u<t.length;u++){let l=t[u],o=this.modules.indexOf(l);if(o<a&&o>-1&&(this.modules.splice(o,1),a--,o=-1),o==-1){if(this.modules.splice(a++,0,l),r)for(let s=0;s<l.rules.length;s++)r.insertRule(l.rules[s],i++)}else{for(;a<o;)i+=this.modules[a++].rules.length;i+=l.rules.length,a++}}if(r)n.adoptedStyleSheets.indexOf(this.sheet)<0&&(n.adoptedStyleSheets=[this.sheet,...n.adoptedStyleSheets]);else{let u="";for(let o=0;o<this.modules.length;o++)u+=this.modules[o].getRules()+`
`;this.styleTag.textContent=u;let l=n.head||n;this.styleTag.parentNode!=l&&l.insertBefore(this.styleTag,l.firstChild)}}setNonce(t){this.styleTag&&this.styleTag.getAttribute("nonce")!=t&&this.styleTag.setAttribute("nonce",t)}};var i1={8:"Backspace",9:"Tab",10:"Enter",12:"NumLock",13:"Enter",16:"Shift",17:"Control",18:"Alt",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",44:"PrintScreen",45:"Insert",46:"Delete",59:";",61:"=",91:"Meta",92:"Meta",106:"*",107:"+",108:",",109:"-",110:".",111:"/",144:"NumLock",145:"ScrollLock",160:"Shift",161:"Shift",162:"Control",163:"Control",164:"Alt",165:"Alt",173:"-",186:";",187:"=",188:",",189:"-",190:".",191:"/",192:"`",219:"[",220:"\\",221:"]",222:"'"},VD={48:")",49:"!",50:"@",51:"#",52:"$",53:"%",54:"^",55:"&",56:"*",57:"(",59:":",61:"+",173:"_",186:":",187:"+",188:"<",189:"_",190:">",191:"?",192:"~",219:"{",220:"|",221:"}",222:'"'},mxe=typeof navigator<"u"&&/Mac/.test(navigator.platform),gxe=typeof navigator<"u"&&/MSIE \d|Trident\/(?:[7-9]|\d{2,})\..*rv:(\d+)/.exec(navigator.userAgent);for(var Bo=0;Bo<10;Bo++)i1[48+Bo]=i1[96+Bo]=String(Bo);for(var Bo=1;Bo<=24;Bo++)i1[Bo+111]="F"+Bo;for(var Bo=65;Bo<=90;Bo++)i1[Bo]=String.fromCharCode(Bo+32),VD[Bo]=String.fromCharCode(Bo);for(var IT in i1)VD.hasOwnProperty(IT)||(VD[IT]=i1[IT]);function Brt(e){var t=mxe&&e.metaKey&&e.shiftKey&&!e.ctrlKey&&!e.altKey||gxe&&e.shiftKey&&e.key&&e.key.length==1||e.key=="Unidentified",n=!t&&e.key||(e.shiftKey?VD:i1)[e.keyCode]||e.key||"Unidentified";return n=="Esc"&&(n="Escape"),n=="Del"&&(n="Delete"),n=="Left"&&(n="ArrowLeft"),n=="Up"&&(n="ArrowUp"),n=="Right"&&(n="ArrowRight"),n=="Down"&&(n="ArrowDown"),n}function Qrt(){var e=arguments[0];typeof e=="string"&&(e=document.createElement(e));var t=1,n=arguments[1];if(n&&typeof n=="object"&&n.nodeType==null&&!Array.isArray(n)){for(var r in n)if(Object.prototype.hasOwnProperty.call(n,r)){var i=n[r];typeof i=="string"?e.setAttribute(r,i):i!=null&&(e[r]=i)}t++}for(;t<arguments.length;t++)Fee(e,arguments[t]);return e}function Fee(e,t){if(typeof t=="string")e.appendChild(document.createTextNode(t));else if(t!=null)if(t.nodeType!=null)e.appendChild(t);else if(Array.isArray(t))for(var n=0;n<t.length;n++)Fee(e,t[n]);else throw new RangeError("Unsupported child node: "+t)}let vxe=0,zm=class y9{constructor(t,n,r,i){this.name=t,this.set=n,this.base=r,this.modified=i,this.id=vxe++}toString(){let{name:t}=this;for(let n of this.modified)n.name&&(t=`${n.name}(${t})`);return t}static define(t,n){let r=typeof t=="string"?t:"?";if(t instanceof y9&&(n=t),n!=null&&n.base)throw new Error("Can not derive from a modified tag");let i=new y9(r,[],null,[]);if(i.set.push(i),n)for(let a of n.set)i.set.push(a);return i}static defineModifier(t){let n=new iU(t);return r=>r.modified.indexOf(n)>-1?r:iU.get(r.base||r,r.modified.concat(n).sort((i,a)=>i.id-a.id))}},yxe=0,iU=class Nee{constructor(t){this.name=t,this.instances=[],this.id=yxe++}static get(t,n){if(!n.length)return t;let r=n[0].instances.find(l=>l.base==t&&bxe(n,l.modified));if(r)return r;let i=[],a=new zm(t.name,i,t,n);for(let l of n)l.instances.push(a);let u=Oxe(n);for(let l of t.set)if(!l.modified.length)for(let o of u)i.push(Nee.get(l,o));return a}};function bxe(e,t){return e.length==t.length&&e.every((n,r)=>n==t[r])}function Oxe(e){let t=[[]];for(let n=0;n<e.length;n++)for(let r=0,i=t.length;r<i;r++)t.push(t[r].concat(e[n]));return t.sort((n,r)=>r.length-n.length)}function oN(e){let t=Object.create(null);for(let n in e){let r=e[n];Array.isArray(r)||(r=[r]);for(let i of n.split(" "))if(i){let a=[],u=2,l=i;for(let d=0;;){if(l=="..."&&d>0&&d+3==i.length){u=1;break}let h=/^"(?:[^"\\]|\\.)*?"|[^\/!]+/.exec(l);if(!h)throw new RangeError("Invalid path: "+i);if(a.push(h[0]=="*"?"":h[0][0]=='"'?JSON.parse(h[0]):h[0]),d+=h[0].length,d==i.length)break;let m=i[d++];if(d==i.length&&m=="!"){u=0;break}if(m!="/")throw new RangeError("Invalid path: "+i);l=i.slice(d)}let o=a.length-1,s=a[o];if(!s)throw new RangeError("Invalid path: "+i);let c=new UD(r,u,o>0?a.slice(0,o):null);t[s]=c.sort(t[s])}}return Iee.add(t)}const Iee=new Ea;let UD=class{constructor(t,n,r,i){this.tags=t,this.mode=n,this.context=r,this.next=i}get opaque(){return this.mode==0}get inherit(){return this.mode==1}sort(t){return!t||t.depth<this.depth?(this.next=t,this):(t.next=this.sort(t.next),t)}get depth(){return this.context?this.context.length:0}};UD.empty=new UD([],2,null);function xxe(e,t){let n=Object.create(null);for(let a of e)if(!Array.isArray(a.tag))n[a.tag.id]=a.class;else for(let u of a.tag)n[u.id]=a.class;let{scope:r,all:i=null}=t||{};return{style:a=>{let u=i;for(let l of a)for(let o of l.set){let s=n[o.id];if(s){u=u?u+" "+s:s;break}}return u},scope:r}}function Sxe(e,t){let n=null;for(let r of e){let i=r.style(t);i&&(n=n?n+" "+i:i)}return n}function qrt(e,t,n,r=0,i=e.length){let a=new _xe(r,Array.isArray(t)?t:[t],n);a.highlightRange(e.cursor(),r,i,"",a.highlighters),a.flush(i)}let _xe=class{constructor(t,n,r){this.at=t,this.highlighters=n,this.span=r,this.class=""}startSpan(t,n){n!=this.class&&(this.flush(t),t>this.at&&(this.at=t),this.class=n)}flush(t){t>this.at&&this.class&&this.span(this.at,t,this.class)}highlightRange(t,n,r,i,a){let{type:u,from:l,to:o}=t;if(l>=r||o<=n)return;u.isTop&&(a=this.highlighters.filter(m=>!m.scope||m.scope(u)));let s=i,c=Exe(t)||UD.empty,d=Sxe(a,c.tags);if(d&&(s&&(s+=" "),s+=d,c.mode==1&&(i+=(i?" ":"")+d)),this.startSpan(Math.max(n,l),s),c.opaque)return;let h=t.tree&&t.tree.prop(Ea.mounted);if(h&&h.overlay){let m=t.node.enter(h.overlay[0].from+l,1),y=this.highlighters.filter(S=>!S.scope||S.scope(h.tree.type)),v=t.firstChild();for(let S=0,_=l;;S++){let E=S<h.overlay.length?h.overlay[S]:null,$=E?E.from+l:o,k=Math.max(n,_),A=Math.min(r,$);if(k<A&&v)for(;t.from<A&&(this.highlightRange(t,k,A,i,a),this.startSpan(Math.min(A,t.to),s),!(t.to>=$||!t.nextSibling())););if(!E||$>r)break;_=E.to+l,_>n&&(this.highlightRange(m.cursor(),Math.max(n,E.from+l),Math.min(r,_),"",y),this.startSpan(Math.min(r,_),s))}v&&t.parent()}else if(t.firstChild()){h&&(i="");do if(!(t.to<=n)){if(t.from>=r)break;this.highlightRange(t,n,r,i,a),this.startSpan(Math.min(r,t.to),s)}while(t.nextSibling());t.parent()}}};function Exe(e){let t=e.type.prop(Iee);for(;t&&t.context&&!e.matchContext(t.context);)t=t.next;return t||null}const Zt=zm.define,L_=Zt(),bv=Zt(),aU=Zt(bv),uU=Zt(bv),Ov=Zt(),B_=Zt(Ov),LT=Zt(Ov),c0=Zt(),pb=Zt(c0),r0=Zt(),i0=Zt(),b9=Zt(),G4=Zt(b9),Q_=Zt(),st={comment:L_,lineComment:Zt(L_),blockComment:Zt(L_),docComment:Zt(L_),name:bv,variableName:Zt(bv),typeName:aU,tagName:Zt(aU),propertyName:uU,attributeName:Zt(uU),className:Zt(bv),labelName:Zt(bv),namespace:Zt(bv),macroName:Zt(bv),literal:Ov,string:B_,docString:Zt(B_),character:Zt(B_),attributeValue:Zt(B_),number:LT,integer:Zt(LT),float:Zt(LT),bool:Zt(Ov),regexp:Zt(Ov),escape:Zt(Ov),color:Zt(Ov),url:Zt(Ov),keyword:r0,self:Zt(r0),null:Zt(r0),atom:Zt(r0),unit:Zt(r0),modifier:Zt(r0),operatorKeyword:Zt(r0),controlKeyword:Zt(r0),definitionKeyword:Zt(r0),moduleKeyword:Zt(r0),operator:i0,derefOperator:Zt(i0),arithmeticOperator:Zt(i0),logicOperator:Zt(i0),bitwiseOperator:Zt(i0),compareOperator:Zt(i0),updateOperator:Zt(i0),definitionOperator:Zt(i0),typeOperator:Zt(i0),controlOperator:Zt(i0),punctuation:b9,separator:Zt(b9),bracket:G4,angleBracket:Zt(G4),squareBracket:Zt(G4),paren:Zt(G4),brace:Zt(G4),content:c0,heading:pb,heading1:Zt(pb),heading2:Zt(pb),heading3:Zt(pb),heading4:Zt(pb),heading5:Zt(pb),heading6:Zt(pb),contentSeparator:Zt(c0),list:Zt(c0),quote:Zt(c0),emphasis:Zt(c0),strong:Zt(c0),link:Zt(c0),monospace:Zt(c0),strikethrough:Zt(c0),inserted:Zt(),deleted:Zt(),changed:Zt(),invalid:Zt(),meta:Q_,documentMeta:Zt(Q_),annotation:Zt(Q_),processingInstruction:Zt(Q_),definition:zm.defineModifier("definition"),constant:zm.defineModifier("constant"),function:zm.defineModifier("function"),standard:zm.defineModifier("standard"),local:zm.defineModifier("local"),special:zm.defineModifier("special")};for(let e in st){let t=st[e];t instanceof zm&&(t.name=e)}xxe([{tag:st.link,class:"tok-link"},{tag:st.heading,class:"tok-heading"},{tag:st.emphasis,class:"tok-emphasis"},{tag:st.strong,class:"tok-strong"},{tag:st.keyword,class:"tok-keyword"},{tag:st.atom,class:"tok-atom"},{tag:st.bool,class:"tok-bool"},{tag:st.url,class:"tok-url"},{tag:st.labelName,class:"tok-labelName"},{tag:st.inserted,class:"tok-inserted"},{tag:st.deleted,class:"tok-deleted"},{tag:st.literal,class:"tok-literal"},{tag:st.string,class:"tok-string"},{tag:st.number,class:"tok-number"},{tag:[st.regexp,st.escape,st.special(st.string)],class:"tok-string2"},{tag:st.variableName,class:"tok-variableName"},{tag:st.local(st.variableName),class:"tok-variableName tok-local"},{tag:st.definition(st.variableName),class:"tok-variableName tok-definition"},{tag:st.special(st.variableName),class:"tok-variableName2"},{tag:st.definition(st.propertyName),class:"tok-propertyName tok-definition"},{tag:st.typeName,class:"tok-typeName"},{tag:st.namespace,class:"tok-namespace"},{tag:st.className,class:"tok-className"},{tag:st.macroName,class:"tok-macroName"},{tag:st.propertyName,class:"tok-propertyName"},{tag:st.operator,class:"tok-operator"},{tag:st.comment,class:"tok-comment"},{tag:st.meta,class:"tok-meta"},{tag:st.invalid,class:"tok-invalid"},{tag:st.punctuation,class:"tok-punctuation"}]);var lU={};class jD{constructor(t,n,r,i,a,u,l,o,s,c=0,d){this.p=t,this.stack=n,this.state=r,this.reducePos=i,this.pos=a,this.score=u,this.buffer=l,this.bufferBase=o,this.curContext=s,this.lookAhead=c,this.parent=d}toString(){return`[${this.stack.filter((t,n)=>n%3==0).concat(this.state)}]@${this.pos}${this.score?"!"+this.score:""}`}static start(t,n,r=0){let i=t.parser.context;return new jD(t,[],n,r,r,0,[],0,i?new oU(i,i.start):null,0,null)}get context(){return this.curContext?this.curContext.context:null}pushState(t,n){this.stack.push(this.state,n,this.bufferBase+this.buffer.length),this.state=t}reduce(t){var n;let r=t>>19,i=t&65535,{parser:a}=this.p,u=this.reducePos<this.pos-25;u&&this.setLookAhead(this.pos);let l=a.dynamicPrecedence(i);if(l&&(this.score+=l),r==0){this.pushState(a.getGoto(this.state,i,!0),this.reducePos),i<a.minRepeatTerm&&this.storeNode(i,this.reducePos,this.reducePos,u?8:4,!0),this.reduceContext(i,this.reducePos);return}let o=this.stack.length-(r-1)*3-(t&262144?6:0),s=o?this.stack[o-2]:this.p.ranges[0].from,c=this.reducePos-s;c>=2e3&&!(!((n=this.p.parser.nodeSet.types[i])===null||n===void 0)&&n.isAnonymous)&&(s==this.p.lastBigReductionStart?(this.p.bigReductionCount++,this.p.lastBigReductionSize=c):this.p.lastBigReductionSize<c&&(this.p.bigReductionCount=1,this.p.lastBigReductionStart=s,this.p.lastBigReductionSize=c));let d=o?this.stack[o-1]:0,h=this.bufferBase+this.buffer.length-d;if(i<a.minRepeatTerm||t&131072){let m=a.stateFlag(this.state,1)?this.pos:this.reducePos;this.storeNode(i,s,m,h+4,!0)}if(t&262144)this.state=this.stack[o];else{let m=this.stack[o-3];this.state=a.getGoto(m,i,!0)}for(;this.stack.length>o;)this.stack.pop();this.reduceContext(i,s)}storeNode(t,n,r,i=4,a=!1){if(t==0&&(!this.stack.length||this.stack[this.stack.length-1]<this.buffer.length+this.bufferBase)){let u=this,l=this.buffer.length;if(l==0&&u.parent&&(l=u.bufferBase-u.parent.bufferBase,u=u.parent),l>0&&u.buffer[l-4]==0&&u.buffer[l-1]>-1){if(n==r)return;if(u.buffer[l-2]>=n){u.buffer[l-2]=r;return}}}if(!a||this.pos==r)this.buffer.push(t,n,r,i);else{let u=this.buffer.length;if(u>0&&this.buffer[u-4]!=0){let l=!1;for(let o=u;o>0&&this.buffer[o-2]>r;o-=4)if(this.buffer[o-1]>=0){l=!0;break}if(l)for(;u>0&&this.buffer[u-2]>r;)this.buffer[u]=this.buffer[u-4],this.buffer[u+1]=this.buffer[u-3],this.buffer[u+2]=this.buffer[u-2],this.buffer[u+3]=this.buffer[u-1],u-=4,i>4&&(i-=4)}this.buffer[u]=t,this.buffer[u+1]=n,this.buffer[u+2]=r,this.buffer[u+3]=i}}shift(t,n,r,i){if(t&131072)this.pushState(t&65535,this.pos);else if((t&262144)==0){let a=t,{parser:u}=this.p;(i>this.pos||n<=u.maxNode)&&(this.pos=i,u.stateFlag(a,1)||(this.reducePos=i)),this.pushState(a,r),this.shiftContext(n,r),n<=u.maxNode&&this.buffer.push(n,r,i,4)}else this.pos=i,this.shiftContext(n,r),n<=this.p.parser.maxNode&&this.buffer.push(n,r,i,4)}apply(t,n,r,i){t&65536?this.reduce(t):this.shift(t,n,r,i)}useNode(t,n){let r=this.p.reused.length-1;(r<0||this.p.reused[r]!=t)&&(this.p.reused.push(t),r++);let i=this.pos;this.reducePos=this.pos=i+t.length,this.pushState(n,i),this.buffer.push(r,i,this.reducePos,-1),this.curContext&&this.updateContext(this.curContext.tracker.reuse(this.curContext.context,t,this,this.p.stream.reset(this.pos-t.length)))}split(){let t=this,n=t.buffer.length;for(;n>0&&t.buffer[n-2]>t.reducePos;)n-=4;let r=t.buffer.slice(n),i=t.bufferBase+n;for(;t&&i==t.bufferBase;)t=t.parent;return new jD(this.p,this.stack.slice(),this.state,this.reducePos,this.pos,this.score,r,i,this.curContext,this.lookAhead,t)}recoverByDelete(t,n){let r=t<=this.p.parser.maxNode;r&&this.storeNode(t,this.pos,n,4),this.storeNode(0,this.pos,n,r?8:4),this.pos=this.reducePos=n,this.score-=190}canShift(t){for(let n=new Dxe(this);;){let r=this.p.parser.stateSlot(n.state,4)||this.p.parser.hasAction(n.state,t);if(r==0)return!1;if((r&65536)==0)return!0;n.reduce(r)}}recoverByInsert(t){if(this.stack.length>=300)return[];let n=this.p.parser.nextStates(this.state);if(n.length>8||this.stack.length>=120){let i=[];for(let a=0,u;a<n.length;a+=2)(u=n[a+1])!=this.state&&this.p.parser.hasAction(u,t)&&i.push(n[a],u);if(this.stack.length<120)for(let a=0;i.length<8&&a<n.length;a+=2){let u=n[a+1];i.some((l,o)=>o&1&&l==u)||i.push(n[a],u)}n=i}let r=[];for(let i=0;i<n.length&&r.length<4;i+=2){let a=n[i+1];if(a==this.state)continue;let u=this.split();u.pushState(a,this.pos),u.storeNode(0,u.pos,u.pos,4,!0),u.shiftContext(n[i],this.pos),u.reducePos=this.pos,u.score-=200,r.push(u)}return r}forceReduce(){let{parser:t}=this.p,n=t.stateSlot(this.state,5);if((n&65536)==0)return!1;if(!t.validAction(this.state,n)){let r=n>>19,i=n&65535,a=this.stack.length-r*3;if(a<0||t.getGoto(this.stack[a],i,!1)<0){let u=this.findForcedReduction();if(u==null)return!1;n=u}this.storeNode(0,this.pos,this.pos,4,!0),this.score-=100}return this.reducePos=this.pos,this.reduce(n),!0}findForcedReduction(){let{parser:t}=this.p,n=[],r=(i,a)=>{if(!n.includes(i))return n.push(i),t.allActions(i,u=>{if(!(u&393216))if(u&65536){let l=(u>>19)-a;if(l>1){let o=u&65535,s=this.stack.length-l*3;if(s>=0&&t.getGoto(this.stack[s],o,!1)>=0)return l<<19|65536|o}}else{let l=r(u,a+1);if(l!=null)return l}})};return r(this.state,0)}forceAll(){for(;!this.p.parser.stateFlag(this.state,2);)if(!this.forceReduce()){this.storeNode(0,this.pos,this.pos,4,!0);break}return this}get deadEnd(){if(this.stack.length!=3)return!1;let{parser:t}=this.p;return t.data[t.stateSlot(this.state,1)]==65535&&!t.stateSlot(this.state,4)}restart(){this.storeNode(0,this.pos,this.pos,4,!0),this.state=this.stack[0],this.stack.length=0}sameState(t){if(this.state!=t.state||this.stack.length!=t.stack.length)return!1;for(let n=0;n<this.stack.length;n+=3)if(this.stack[n]!=t.stack[n])return!1;return!0}get parser(){return this.p.parser}dialectEnabled(t){return this.p.parser.dialect.flags[t]}shiftContext(t,n){this.curContext&&this.updateContext(this.curContext.tracker.shift(this.curContext.context,t,this,this.p.stream.reset(n)))}reduceContext(t,n){this.curContext&&this.updateContext(this.curContext.tracker.reduce(this.curContext.context,t,this,this.p.stream.reset(n)))}emitContext(){let t=this.buffer.length-1;(t<0||this.buffer[t]!=-3)&&this.buffer.push(this.curContext.hash,this.pos,this.pos,-3)}emitLookAhead(){let t=this.buffer.length-1;(t<0||this.buffer[t]!=-4)&&this.buffer.push(this.lookAhead,this.pos,this.pos,-4)}updateContext(t){if(t!=this.curContext.context){let n=new oU(this.curContext.tracker,t);n.hash!=this.curContext.hash&&this.emitContext(),this.curContext=n}}setLookAhead(t){t>this.lookAhead&&(this.emitLookAhead(),this.lookAhead=t)}close(){this.curContext&&this.curContext.tracker.strict&&this.emitContext(),this.lookAhead>0&&this.emitLookAhead()}}class oU{constructor(t,n){this.tracker=t,this.context=n,this.hash=t.strict?t.hash(n):0}}class Dxe{constructor(t){this.start=t,this.state=t.state,this.stack=t.stack,this.base=this.stack.length}reduce(t){let n=t&65535,r=t>>19;r==0?(this.stack==this.start.stack&&(this.stack=this.stack.slice()),this.stack.push(this.state,0,0),this.base+=3):this.base-=(r-1)*3;let i=this.start.p.parser.getGoto(this.stack[this.base-3],n,!0);this.state=i}}class ZD{constructor(t,n,r){this.stack=t,this.pos=n,this.index=r,this.buffer=t.buffer,this.index==0&&this.maybeNext()}static create(t,n=t.bufferBase+t.buffer.length){return new ZD(t,n,n-t.bufferBase)}maybeNext(){let t=this.stack.parent;t!=null&&(this.index=this.stack.bufferBase-t.bufferBase,this.stack=t,this.buffer=t.buffer)}get id(){return this.buffer[this.index-4]}get start(){return this.buffer[this.index-3]}get end(){return this.buffer[this.index-2]}get size(){return this.buffer[this.index-1]}next(){this.index-=4,this.pos-=4,this.index==0&&this.maybeNext()}fork(){return new ZD(this.stack,this.pos,this.index)}}function k2(e,t=Uint16Array){if(typeof e!="string")return e;let n=null;for(let r=0,i=0;r<e.length;){let a=0;for(;;){let u=e.charCodeAt(r++),l=!1;if(u==126){a=65535;break}u>=92&&u--,u>=34&&u--;let o=u-32;if(o>=46&&(o-=46,l=!0),a+=o,l)break;a*=46}n?n[i++]=a:n=new t(a)}return n}class WE{constructor(){this.start=-1,this.value=-1,this.end=-1,this.extended=-1,this.lookAhead=0,this.mask=0,this.context=0}}const sU=new WE;class $xe{constructor(t,n){this.input=t,this.ranges=n,this.chunk="",this.chunkOff=0,this.chunk2="",this.chunk2Pos=0,this.next=-1,this.token=sU,this.rangeIndex=0,this.pos=this.chunkPos=n[0].from,this.range=n[0],this.end=n[n.length-1].to,this.readNext()}resolveOffset(t,n){let r=this.range,i=this.rangeIndex,a=this.pos+t;for(;a<r.from;){if(!i)return null;let u=this.ranges[--i];a-=r.from-u.to,r=u}for(;n<0?a>r.to:a>=r.to;){if(i==this.ranges.length-1)return null;let u=this.ranges[++i];a+=u.from-r.to,r=u}return a}clipPos(t){if(t>=this.range.from&&t<this.range.to)return t;for(let n of this.ranges)if(n.to>t)return Math.max(t,n.from);return this.end}peek(t){let n=this.chunkOff+t,r,i;if(n>=0&&n<this.chunk.length)r=this.pos+t,i=this.chunk.charCodeAt(n);else{let a=this.resolveOffset(t,1);if(a==null)return-1;if(r=a,r>=this.chunk2Pos&&r<this.chunk2Pos+this.chunk2.length)i=this.chunk2.charCodeAt(r-this.chunk2Pos);else{let u=this.rangeIndex,l=this.range;for(;l.to<=r;)l=this.ranges[++u];this.chunk2=this.input.chunk(this.chunk2Pos=r),r+this.chunk2.length>l.to&&(this.chunk2=this.chunk2.slice(0,l.to-r)),i=this.chunk2.charCodeAt(0)}}return r>=this.token.lookAhead&&(this.token.lookAhead=r+1),i}acceptToken(t,n=0){let r=n?this.resolveOffset(n,-1):this.pos;if(r==null||r<this.token.start)throw new RangeError("Token end out of bounds");this.token.value=t,this.token.end=r}acceptTokenTo(t,n){this.token.value=t,this.token.end=n}getChunk(){if(this.pos>=this.chunk2Pos&&this.pos<this.chunk2Pos+this.chunk2.length){let{chunk:t,chunkPos:n}=this;this.chunk=this.chunk2,this.chunkPos=this.chunk2Pos,this.chunk2=t,this.chunk2Pos=n,this.chunkOff=this.pos-this.chunkPos}else{this.chunk2=this.chunk,this.chunk2Pos=this.chunkPos;let t=this.input.chunk(this.pos),n=this.pos+t.length;this.chunk=n>this.range.to?t.slice(0,this.range.to-this.pos):t,this.chunkPos=this.pos,this.chunkOff=0}}readNext(){return this.chunkOff>=this.chunk.length&&(this.getChunk(),this.chunkOff==this.chunk.length)?this.next=-1:this.next=this.chunk.charCodeAt(this.chunkOff)}advance(t=1){for(this.chunkOff+=t;this.pos+t>=this.range.to;){if(this.rangeIndex==this.ranges.length-1)return this.setDone();t-=this.range.to-this.pos,this.range=this.ranges[++this.rangeIndex],this.pos=this.range.from}return this.pos+=t,this.pos>=this.token.lookAhead&&(this.token.lookAhead=this.pos+1),this.readNext()}setDone(){return this.pos=this.chunkPos=this.end,this.range=this.ranges[this.rangeIndex=this.ranges.length-1],this.chunk="",this.next=-1}reset(t,n){if(n?(this.token=n,n.start=t,n.lookAhead=t+1,n.value=n.extended=-1):this.token=sU,this.pos!=t){if(this.pos=t,t==this.end)return this.setDone(),this;for(;t<this.range.from;)this.range=this.ranges[--this.rangeIndex];for(;t>=this.range.to;)this.range=this.ranges[++this.rangeIndex];t>=this.chunkPos&&t<this.chunkPos+this.chunk.length?this.chunkOff=t-this.chunkPos:(this.chunk="",this.chunkOff=0),this.readNext()}return this}read(t,n){if(t>=this.chunkPos&&n<=this.chunkPos+this.chunk.length)return this.chunk.slice(t-this.chunkPos,n-this.chunkPos);if(t>=this.chunk2Pos&&n<=this.chunk2Pos+this.chunk2.length)return this.chunk2.slice(t-this.chunk2Pos,n-this.chunk2Pos);if(t>=this.range.from&&n<=this.range.to)return this.input.read(t,n);let r="";for(let i of this.ranges){if(i.from>=n)break;i.to>t&&(r+=this.input.read(Math.max(i.from,t),Math.min(i.to,n)))}return r}}class n3{constructor(t,n){this.data=t,this.id=n}token(t,n){let{parser:r}=n.p;Lee(this.data,t,n,this.id,r.data,r.tokenPrecTable)}}n3.prototype.contextual=n3.prototype.fallback=n3.prototype.extend=!1;class O9{constructor(t,n,r){this.precTable=n,this.elseToken=r,this.data=typeof t=="string"?k2(t):t}token(t,n){let r=t.pos,i=0;for(;;){let a=t.next<0,u=t.resolveOffset(1,1);if(Lee(this.data,t,n,0,this.data,this.precTable),t.token.value>-1)break;if(this.elseToken==null)return;if(a||i++,u==null)break;t.reset(u,t.token)}i&&(t.reset(r,t.token),t.acceptToken(this.elseToken,i))}}O9.prototype.contextual=n3.prototype.fallback=n3.prototype.extend=!1;class vg{constructor(t,n={}){this.token=t,this.contextual=!!n.contextual,this.fallback=!!n.fallback,this.extend=!!n.extend}}function Lee(e,t,n,r,i,a){let u=0,l=1<<r,{dialect:o}=n.p.parser;e:for(;(l&e[u])!=0;){let s=e[u+1];for(let m=u+3;m<s;m+=2)if((e[m+1]&l)>0){let y=e[m];if(o.allows(y)&&(t.token.value==-1||t.token.value==y||Cxe(y,t.token.value,i,a))){t.acceptToken(y);break}}let c=t.next,d=0,h=e[u+2];if(t.next<0&&h>d&&e[s+h*3-3]==65535){u=e[s+h*3-1];continue e}for(;d<h;){let m=d+h>>1,y=s+m+(m<<1),v=e[y],S=e[y+1]||65536;if(c<v)h=m;else if(c>=S)d=m+1;else{u=e[y+2],t.advance();continue e}}break}}function cU(e,t,n){for(let r=t,i;(i=e[r])!=65535;r++)if(i==n)return r-t;return-1}function Cxe(e,t,n,r){let i=cU(n,r,t);return i<0||cU(n,r,e)<i}const Nf=typeof process<"u"&&lU&&/\bparse\b/.test(lU.LOG);let BT=null;function fU(e,t,n){let r=e.cursor(Lo.IncludeAnonymous);for(r.moveTo(t);;)if(!(n<0?r.childBefore(t):r.childAfter(t)))for(;;){if((n<0?r.to<t:r.from>t)&&!r.type.isError)return n<0?Math.max(0,Math.min(r.to-1,t-25)):Math.min(e.length,Math.max(r.from+1,t+25));if(n<0?r.prevSibling():r.nextSibling())break;if(!r.parent())return n<0?0:e.length}}class wxe{constructor(t,n){this.fragments=t,this.nodeSet=n,this.i=0,this.fragment=null,this.safeFrom=-1,this.safeTo=-1,this.trees=[],this.start=[],this.index=[],this.nextFragment()}nextFragment(){let t=this.fragment=this.i==this.fragments.length?null:this.fragments[this.i++];if(t){for(this.safeFrom=t.openStart?fU(t.tree,t.from+t.offset,1)-t.offset:t.from,this.safeTo=t.openEnd?fU(t.tree,t.to+t.offset,-1)-t.offset:t.to;this.trees.length;)this.trees.pop(),this.start.pop(),this.index.pop();this.trees.push(t.tree),this.start.push(-t.offset),this.index.push(0),this.nextStart=this.safeFrom}else this.nextStart=1e9}nodeAt(t){if(t<this.nextStart)return null;for(;this.fragment&&this.safeTo<=t;)this.nextFragment();if(!this.fragment)return null;for(;;){let n=this.trees.length-1;if(n<0)return this.nextFragment(),null;let r=this.trees[n],i=this.index[n];if(i==r.children.length){this.trees.pop(),this.start.pop(),this.index.pop();continue}let a=r.children[i],u=this.start[n]+r.positions[i];if(u>t)return this.nextStart=u,null;if(a instanceof ep){if(u==t){if(u<this.safeFrom)return null;let l=u+a.length;if(l<=this.safeTo){let o=a.prop(Ea.lookAhead);if(!o||l+o<this.fragment.to)return a}}this.index[n]++,u+a.length>=Math.max(this.safeFrom,t)&&(this.trees.push(a),this.start.push(u),this.index.push(0))}else this.index[n]++,this.nextStart=u+a.length}}}class Rxe{constructor(t,n){this.stream=n,this.tokens=[],this.mainToken=null,this.actions=[],this.tokens=t.tokenizers.map(r=>new WE)}getActions(t){let n=0,r=null,{parser:i}=t.p,{tokenizers:a}=i,u=i.stateSlot(t.state,3),l=t.curContext?t.curContext.hash:0,o=0;for(let s=0;s<a.length;s++){if((1<<s&u)==0)continue;let c=a[s],d=this.tokens[s];if(!(r&&!c.fallback)&&((c.contextual||d.start!=t.pos||d.mask!=u||d.context!=l)&&(this.updateCachedToken(d,c,t),d.mask=u,d.context=l),d.lookAhead>d.end+25&&(o=Math.max(d.lookAhead,o)),d.value!=0)){let h=n;if(d.extended>-1&&(n=this.addActions(t,d.extended,d.end,n)),n=this.addActions(t,d.value,d.end,n),!c.extend&&(r=d,n>h))break}}for(;this.actions.length>n;)this.actions.pop();return o&&t.setLookAhead(o),!r&&t.pos==this.stream.end&&(r=new WE,r.value=t.p.parser.eofTerm,r.start=r.end=t.pos,n=this.addActions(t,r.value,r.end,n)),this.mainToken=r,this.actions}getMainToken(t){if(this.mainToken)return this.mainToken;let n=new WE,{pos:r,p:i}=t;return n.start=r,n.end=Math.min(r+1,i.stream.end),n.value=r==i.stream.end?i.parser.eofTerm:0,n}updateCachedToken(t,n,r){let i=this.stream.clipPos(r.pos);if(n.token(this.stream.reset(i,t),r),t.value>-1){let{parser:a}=r.p;for(let u=0;u<a.specialized.length;u++)if(a.specialized[u]==t.value){let l=a.specializers[u](this.stream.read(t.start,t.end),r);if(l>=0&&r.p.parser.dialect.allows(l>>1)){(l&1)==0?t.value=l>>1:t.extended=l>>1;break}}}else t.value=0,t.end=this.stream.clipPos(i+1)}putAction(t,n,r,i){for(let a=0;a<i;a+=3)if(this.actions[a]==t)return i;return this.actions[i++]=t,this.actions[i++]=n,this.actions[i++]=r,i}addActions(t,n,r,i){let{state:a}=t,{parser:u}=t.p,{data:l}=u;for(let o=0;o<2;o++)for(let s=u.stateSlot(a,o?2:1);;s+=3){if(l[s]==65535)if(l[s+1]==1)s=qm(l,s+2);else{i==0&&l[s+1]==2&&(i=this.putAction(qm(l,s+2),n,r,i));break}l[s]==n&&(i=this.putAction(qm(l,s+1),n,r,i))}return i}}class Axe{constructor(t,n,r,i){this.parser=t,this.input=n,this.ranges=i,this.recovering=0,this.nextStackID=9812,this.minStackPos=0,this.reused=[],this.stoppedAt=null,this.lastBigReductionStart=-1,this.lastBigReductionSize=0,this.bigReductionCount=0,this.stream=new $xe(n,i),this.tokens=new Rxe(t,this.stream),this.topTerm=t.top[1];let{from:a}=i[0];this.stacks=[jD.start(this,t.top[0],a)],this.fragments=r.length&&this.stream.end-a>t.bufferLength*4?new wxe(r,t.nodeSet):null}get parsedPos(){return this.minStackPos}advance(){let t=this.stacks,n=this.minStackPos,r=this.stacks=[],i,a;if(this.bigReductionCount>300&&t.length==1){let[u]=t;for(;u.forceReduce()&&u.stack.length&&u.stack[u.stack.length-2]>=this.lastBigReductionStart;);this.bigReductionCount=this.lastBigReductionSize=0}for(let u=0;u<t.length;u++){let l=t[u];for(;;){if(this.tokens.mainToken=null,l.pos>n)r.push(l);else{if(this.advanceStack(l,r,t))continue;{i||(i=[],a=[]),i.push(l);let o=this.tokens.getMainToken(l);a.push(o.value,o.end)}}break}}if(!r.length){let u=i&&kxe(i);if(u)return Nf&&console.log("Finish with "+this.stackID(u)),this.stackToTree(u);if(this.parser.strict)throw Nf&&i&&console.log("Stuck with token "+(this.tokens.mainToken?this.parser.getName(this.tokens.mainToken.value):"none")),new SyntaxError("No parse at "+n);this.recovering||(this.recovering=5)}if(this.recovering&&i){let u=this.stoppedAt!=null&&i[0].pos>this.stoppedAt?i[0]:this.runRecovery(i,a,r);if(u)return Nf&&console.log("Force-finish "+this.stackID(u)),this.stackToTree(u.forceAll())}if(this.recovering){let u=this.recovering==1?1:this.recovering*3;if(r.length>u)for(r.sort((l,o)=>o.score-l.score);r.length>u;)r.pop();r.some(l=>l.reducePos>n)&&this.recovering--}else if(r.length>1){e:for(let u=0;u<r.length-1;u++){let l=r[u];for(let o=u+1;o<r.length;o++){let s=r[o];if(l.sameState(s)||l.buffer.length>500&&s.buffer.length>500)if((l.score-s.score||l.buffer.length-s.buffer.length)>0)r.splice(o--,1);else{r.splice(u--,1);continue e}}}r.length>12&&r.splice(12,r.length-12)}this.minStackPos=r[0].pos;for(let u=1;u<r.length;u++)r[u].pos<this.minStackPos&&(this.minStackPos=r[u].pos);return null}stopAt(t){if(this.stoppedAt!=null&&this.stoppedAt<t)throw new RangeError("Can't move stoppedAt forward");this.stoppedAt=t}advanceStack(t,n,r){let i=t.pos,{parser:a}=this,u=Nf?this.stackID(t)+" -> ":"";if(this.stoppedAt!=null&&i>this.stoppedAt)return t.forceReduce()?t:null;if(this.fragments){let s=t.curContext&&t.curContext.tracker.strict,c=s?t.curContext.hash:0;for(let d=this.fragments.nodeAt(i);d;){let h=this.parser.nodeSet.types[d.type.id]==d.type?a.getGoto(t.state,d.type.id):-1;if(h>-1&&d.length&&(!s||(d.prop(Ea.contextHash)||0)==c))return t.useNode(d,h),Nf&&console.log(u+this.stackID(t)+` (via reuse of ${a.getName(d.type.id)})`),!0;if(!(d instanceof ep)||d.children.length==0||d.positions[0]>0)break;let m=d.children[0];if(m instanceof ep&&d.positions[0]==0)d=m;else break}}let l=a.stateSlot(t.state,4);if(l>0)return t.reduce(l),Nf&&console.log(u+this.stackID(t)+` (via always-reduce ${a.getName(l&65535)})`),!0;if(t.stack.length>=8400)for(;t.stack.length>6e3&&t.forceReduce(););let o=this.tokens.getActions(t);for(let s=0;s<o.length;){let c=o[s++],d=o[s++],h=o[s++],m=s==o.length||!r,y=m?t:t.split(),v=this.tokens.mainToken;if(y.apply(c,d,v?v.start:y.pos,h),Nf&&console.log(u+this.stackID(y)+` (via ${(c&65536)==0?"shift":`reduce of ${a.getName(c&65535)}`} for ${a.getName(d)} @ ${i}${y==t?"":", split"})`),m)return!0;y.pos>i?n.push(y):r.push(y)}return!1}advanceFully(t,n){let r=t.pos;for(;;){if(!this.advanceStack(t,null,null))return!1;if(t.pos>r)return dU(t,n),!0}}runRecovery(t,n,r){let i=null,a=!1;for(let u=0;u<t.length;u++){let l=t[u],o=n[u<<1],s=n[(u<<1)+1],c=Nf?this.stackID(l)+" -> ":"";if(l.deadEnd&&(a||(a=!0,l.restart(),Nf&&console.log(c+this.stackID(l)+" (restarted)"),this.advanceFully(l,r))))continue;let d=l.split(),h=c;for(let m=0;d.forceReduce()&&m<10&&(Nf&&console.log(h+this.stackID(d)+" (via force-reduce)"),!this.advanceFully(d,r));m++)Nf&&(h=this.stackID(d)+" -> ");for(let m of l.recoverByInsert(o))Nf&&console.log(c+this.stackID(m)+" (via recover-insert)"),this.advanceFully(m,r);this.stream.end>l.pos?(s==l.pos&&(s++,o=0),l.recoverByDelete(o,s),Nf&&console.log(c+this.stackID(l)+` (via recover-delete ${this.parser.getName(o)})`),dU(l,r)):(!i||i.score<l.score)&&(i=l)}return i}stackToTree(t){return t.close(),ep.build({buffer:ZD.create(t),nodeSet:this.parser.nodeSet,topID:this.topTerm,maxBufferLength:this.parser.bufferLength,reused:this.reused,start:this.ranges[0].from,length:t.pos-this.ranges[0].from,minRepeatType:this.parser.minRepeatTerm})}stackID(t){let n=(BT||(BT=new WeakMap)).get(t);return n||BT.set(t,n=String.fromCodePoint(this.nextStackID++)),n+t}}function dU(e,t){for(let n=0;n<t.length;n++){let r=t[n];if(r.pos==e.pos&&r.sameState(e)){t[n].score<e.score&&(t[n]=e);return}}t.push(e)}class Txe{constructor(t,n,r){this.source=t,this.flags=n,this.disabled=r}allows(t){return!this.disabled||this.disabled[t]==0}}const QT=e=>e;class Bee{constructor(t){this.start=t.start,this.shift=t.shift||QT,this.reduce=t.reduce||QT,this.reuse=t.reuse||QT,this.hash=t.hash||(()=>0),this.strict=t.strict!==!1}}class w3 extends dxe{constructor(t){if(super(),this.wrappers=[],t.version!=14)throw new RangeError(`Parser version (${t.version}) doesn't match runtime version (14)`);let n=t.nodeNames.split(" ");this.minRepeatTerm=n.length;for(let l=0;l<t.repeatNodeCount;l++)n.push("");let r=Object.keys(t.topRules).map(l=>t.topRules[l][1]),i=[];for(let l=0;l<n.length;l++)i.push([]);function a(l,o,s){i[l].push([o,o.deserialize(String(s))])}if(t.nodeProps)for(let l of t.nodeProps){let o=l[0];typeof o=="string"&&(o=Ea[o]);for(let s=1;s<l.length;){let c=l[s++];if(c>=0)a(c,o,l[s++]);else{let d=l[s+-c];for(let h=-c;h>0;h--)a(l[s++],o,d);s++}}}this.nodeSet=new aN(n.map((l,o)=>sg.define({name:o>=this.minRepeatTerm?void 0:l,id:o,props:i[o],top:r.indexOf(o)>-1,error:o==0,skipped:t.skippedNodes&&t.skippedNodes.indexOf(o)>-1}))),t.propSources&&(this.nodeSet=this.nodeSet.extend(...t.propSources)),this.strict=!1,this.bufferLength=wee;let u=k2(t.tokenData);this.context=t.context,this.specializerSpecs=t.specialized||[],this.specialized=new Uint16Array(this.specializerSpecs.length);for(let l=0;l<this.specializerSpecs.length;l++)this.specialized[l]=this.specializerSpecs[l].term;this.specializers=this.specializerSpecs.map(hU),this.states=k2(t.states,Uint32Array),this.data=k2(t.stateData),this.goto=k2(t.goto),this.maxTerm=t.maxTerm,this.tokenizers=t.tokenizers.map(l=>typeof l=="number"?new n3(u,l):l),this.topRules=t.topRules,this.dialects=t.dialects||{},this.dynamicPrecedences=t.dynamicPrecedences||null,this.tokenPrecTable=t.tokenPrec,this.termNames=t.termNames||null,this.maxNode=this.nodeSet.types.length-1,this.dialect=this.parseDialect(),this.top=this.topRules[Object.keys(this.topRules)[0]]}createParse(t,n,r){let i=new Axe(this,t,n,r);for(let a of this.wrappers)i=a(i,t,n,r);return i}getGoto(t,n,r=!1){let i=this.goto;if(n>=i[0])return-1;for(let a=i[n+1];;){let u=i[a++],l=u&1,o=i[a++];if(l&&r)return o;for(let s=a+(u>>1);a<s;a++)if(i[a]==t)return o;if(l)return-1}}hasAction(t,n){let r=this.data;for(let i=0;i<2;i++)for(let a=this.stateSlot(t,i?2:1),u;;a+=3){if((u=r[a])==65535)if(r[a+1]==1)u=r[a=qm(r,a+2)];else{if(r[a+1]==2)return qm(r,a+2);break}if(u==n||u==0)return qm(r,a+1)}return 0}stateSlot(t,n){return this.states[t*6+n]}stateFlag(t,n){return(this.stateSlot(t,0)&n)>0}validAction(t,n){return!!this.allActions(t,r=>r==n?!0:null)}allActions(t,n){let r=this.stateSlot(t,4),i=r?n(r):void 0;for(let a=this.stateSlot(t,1);i==null;a+=3){if(this.data[a]==65535)if(this.data[a+1]==1)a=qm(this.data,a+2);else break;i=n(qm(this.data,a+1))}return i}nextStates(t){let n=[];for(let r=this.stateSlot(t,1);;r+=3){if(this.data[r]==65535)if(this.data[r+1]==1)r=qm(this.data,r+2);else break;if((this.data[r+2]&1)==0){let i=this.data[r+1];n.some((a,u)=>u&1&&a==i)||n.push(this.data[r],i)}}return n}configure(t){let n=Object.assign(Object.create(w3.prototype),this);if(t.props&&(n.nodeSet=this.nodeSet.extend(...t.props)),t.top){let r=this.topRules[t.top];if(!r)throw new RangeError(`Invalid top rule name ${t.top}`);n.top=r}return t.tokenizers&&(n.tokenizers=this.tokenizers.map(r=>{let i=t.tokenizers.find(a=>a.from==r);return i?i.to:r})),t.specializers&&(n.specializers=this.specializers.slice(),n.specializerSpecs=this.specializerSpecs.map((r,i)=>{let a=t.specializers.find(l=>l.from==r.external);if(!a)return r;let u=Object.assign(Object.assign({},r),{external:a.to});return n.specializers[i]=hU(u),u})),t.contextTracker&&(n.context=t.contextTracker),t.dialect&&(n.dialect=this.parseDialect(t.dialect)),t.strict!=null&&(n.strict=t.strict),t.wrap&&(n.wrappers=n.wrappers.concat(t.wrap)),t.bufferLength!=null&&(n.bufferLength=t.bufferLength),n}hasWrappers(){return this.wrappers.length>0}getName(t){return this.termNames?this.termNames[t]:String(t<=this.maxNode&&this.nodeSet.types[t].name||t)}get eofTerm(){return this.maxNode+1}get topNode(){return this.nodeSet.types[this.top[1]]}dynamicPrecedence(t){let n=this.dynamicPrecedences;return n==null?0:n[t]||0}parseDialect(t){let n=Object.keys(this.dialects),r=n.map(()=>!1);if(t)for(let a of t.split(" ")){let u=n.indexOf(a);u>=0&&(r[u]=!0)}let i=null;for(let a=0;a<n.length;a++)if(!r[a])for(let u=this.dialects[n[a]],l;(l=this.data[u++])!=65535;)(i||(i=new Uint8Array(this.maxTerm+1)))[l]=1;return new Txe(t,r,i)}static deserialize(t){return new w3(t)}}function qm(e,t){return e[t]|e[t+1]<<16}function kxe(e){let t=null;for(let n of e){let r=n.p.stoppedAt;(n.pos==n.p.stream.end||r!=null&&n.pos>r)&&n.p.parser.stateFlag(n.state,2)&&(!t||t.score<n.score)&&(t=n)}return t}function hU(e){if(e.external){let t=e.extend?1:0;return(n,r)=>e.external(n,r)<<1|t}return e.get}const Pxe=Symbol("Let zodToJsonSchema decide on which parser to use"),pU={name:void 0,$refStrategy:"root",basePath:["#"],effectStrategy:"input",pipeStrategy:"all",dateStrategy:"format:date-time",mapStrategy:"entries",removeAdditionalStrategy:"passthrough",allowedAdditionalProperties:!0,rejectedAdditionalProperties:!1,definitionPath:"definitions",target:"jsonSchema7",strictUnions:!1,definitions:{},errorMessages:!1,markdownDescription:!1,patternStrategy:"escape",applyRegexFlags:!1,emailStrategy:"format:email",base64Strategy:"contentEncoding:base64",nameStrategy:"ref",openAiAnyTypeName:"OpenAiAnyType"},Mxe=e=>typeof e=="string"?{...pU,name:e}:{...pU,...e},Fxe=e=>{const t=Mxe(e),n=t.name!==void 0?[...t.basePath,t.definitionPath,t.name]:t.basePath;return{...t,flags:{hasReferencedOpenAiAnyType:!1},currentPath:n,propertyPath:void 0,seen:new Map(Object.entries(t.definitions).map(([r,i])=>[i._def,{def:i._def,path:[...t.basePath,t.definitionPath,r],jsonSchema:void 0}]))}};function Qee(e,t,n,r){r!=null&&r.errorMessages&&n&&(e.errorMessage={...e.errorMessage,[t]:n})}function Ki(e,t,n,r,i){e[t]=n,Qee(e,t,r,i)}const zee=(e,t)=>{let n=0;for(;n<e.length&&n<t.length&&e[n]===t[n];n++);return[(e.length-n).toString(),...t.slice(n)].join("/")};function cf(e){if(e.target!=="openAi")return{};const t=[...e.basePath,e.definitionPath,e.openAiAnyTypeName];return e.flags.hasReferencedOpenAiAnyType=!0,{$ref:e.$refStrategy==="relative"?zee(t,e.currentPath):t.join("/")}}function Nxe(e,t){var r,i,a;const n={type:"array"};return(r=e.type)!=null&&r._def&&((a=(i=e.type)==null?void 0:i._def)==null?void 0:a.typeName)!==It.ZodAny&&(n.items=Fi(e.type._def,{...t,currentPath:[...t.currentPath,"items"]})),e.minLength&&Ki(n,"minItems",e.minLength.value,e.minLength.message,t),e.maxLength&&Ki(n,"maxItems",e.maxLength.value,e.maxLength.message,t),e.exactLength&&(Ki(n,"minItems",e.exactLength.value,e.exactLength.message,t),Ki(n,"maxItems",e.exactLength.value,e.exactLength.message,t)),n}function Ixe(e,t){const n={type:"integer",format:"int64"};if(!e.checks)return n;for(const r of e.checks)switch(r.kind){case"min":t.target==="jsonSchema7"?r.inclusive?Ki(n,"minimum",r.value,r.message,t):Ki(n,"exclusiveMinimum",r.value,r.message,t):(r.inclusive||(n.exclusiveMinimum=!0),Ki(n,"minimum",r.value,r.message,t));break;case"max":t.target==="jsonSchema7"?r.inclusive?Ki(n,"maximum",r.value,r.message,t):Ki(n,"exclusiveMaximum",r.value,r.message,t):(r.inclusive||(n.exclusiveMaximum=!0),Ki(n,"maximum",r.value,r.message,t));break;case"multipleOf":Ki(n,"multipleOf",r.value,r.message,t);break}return n}function Lxe(){return{type:"boolean"}}function qee(e,t){return Fi(e.type._def,t)}const Bxe=(e,t)=>Fi(e.innerType._def,t);function Vee(e,t,n){const r=n??t.dateStrategy;if(Array.isArray(r))return{anyOf:r.map((i,a)=>Vee(e,t,i))};switch(r){case"string":case"format:date-time":return{type:"string",format:"date-time"};case"format:date":return{type:"string",format:"date"};case"integer":return Qxe(e,t)}}const Qxe=(e,t)=>{const n={type:"integer",format:"unix-time"};if(t.target==="openApi3")return n;for(const r of e.checks)switch(r.kind){case"min":Ki(n,"minimum",r.value,r.message,t);break;case"max":Ki(n,"maximum",r.value,r.message,t);break}return n};function zxe(e,t){return{...Fi(e.innerType._def,t),default:e.defaultValue()}}function qxe(e,t){return t.effectStrategy==="input"?Fi(e.schema._def,t):cf(t)}function Vxe(e){return{type:"string",enum:Array.from(e.values)}}const Uxe=e=>"type"in e&&e.type==="string"?!1:"allOf"in e;function jxe(e,t){const n=[Fi(e.left._def,{...t,currentPath:[...t.currentPath,"allOf","0"]}),Fi(e.right._def,{...t,currentPath:[...t.currentPath,"allOf","1"]})].filter(a=>!!a);let r=t.target==="jsonSchema2019-09"?{unevaluatedProperties:!1}:void 0;const i=[];return n.forEach(a=>{if(Uxe(a))i.push(...a.allOf),a.unevaluatedProperties===void 0&&(r=void 0);else{let u=a;if("additionalProperties"in a&&a.additionalProperties===!1){const{additionalProperties:l,...o}=a;u=o}else r=void 0;i.push(u)}}),i.length?{allOf:i,...r}:void 0}function Zxe(e,t){const n=typeof e.value;return n!=="bigint"&&n!=="number"&&n!=="boolean"&&n!=="string"?{type:Array.isArray(e.value)?"array":"object"}:t.target==="openApi3"?{type:n==="bigint"?"integer":n,enum:[e.value]}:{type:n==="bigint"?"integer":n,const:e.value}}let zT;const jh={cuid:/^[cC][^\s-]{8,}$/,cuid2:/^[0-9a-z]+$/,ulid:/^[0-9A-HJKMNP-TV-Z]{26}$/,email:/^(?!\.)(?!.*\.\.)([a-zA-Z0-9_'+\-\.]*)[a-zA-Z0-9_+-]@([a-zA-Z0-9][a-zA-Z0-9\-]*\.)+[a-zA-Z]{2,}$/,emoji:()=>(zT===void 0&&(zT=RegExp("^(\\p{Extended_Pictographic}|\\p{Emoji_Component})+$","u")),zT),uuid:/^[0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12}$/,ipv4:/^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])$/,ipv4Cidr:/^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\/(3[0-2]|[12]?[0-9])$/,ipv6:/^(([a-f0-9]{1,4}:){7}|::([a-f0-9]{1,4}:){0,6}|([a-f0-9]{1,4}:){1}:([a-f0-9]{1,4}:){0,5}|([a-f0-9]{1,4}:){2}:([a-f0-9]{1,4}:){0,4}|([a-f0-9]{1,4}:){3}:([a-f0-9]{1,4}:){0,3}|([a-f0-9]{1,4}:){4}:([a-f0-9]{1,4}:){0,2}|([a-f0-9]{1,4}:){5}:([a-f0-9]{1,4}:){0,1})([a-f0-9]{1,4}|(((25[0-5])|(2[0-4][0-9])|(1[0-9]{2})|([0-9]{1,2}))\.){3}((25[0-5])|(2[0-4][0-9])|(1[0-9]{2})|([0-9]{1,2})))$/,ipv6Cidr:/^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))\/(12[0-8]|1[01][0-9]|[1-9]?[0-9])$/,base64:/^([0-9a-zA-Z+/]{4})*(([0-9a-zA-Z+/]{2}==)|([0-9a-zA-Z+/]{3}=))?$/,base64url:/^([0-9a-zA-Z-_]{4})*(([0-9a-zA-Z-_]{2}(==)?)|([0-9a-zA-Z-_]{3}(=)?))?$/,nanoid:/^[a-zA-Z0-9_-]{21}$/,jwt:/^[A-Za-z0-9-_]+\.[A-Za-z0-9-_]+\.[A-Za-z0-9-_]*$/};function Uee(e,t){const n={type:"string"};if(e.checks)for(const r of e.checks)switch(r.kind){case"min":Ki(n,"minLength",typeof n.minLength=="number"?Math.max(n.minLength,r.value):r.value,r.message,t);break;case"max":Ki(n,"maxLength",typeof n.maxLength=="number"?Math.min(n.maxLength,r.value):r.value,r.message,t);break;case"email":switch(t.emailStrategy){case"format:email":Zh(n,"email",r.message,t);break;case"format:idn-email":Zh(n,"idn-email",r.message,t);break;case"pattern:zod":nc(n,jh.email,r.message,t);break}break;case"url":Zh(n,"uri",r.message,t);break;case"uuid":Zh(n,"uuid",r.message,t);break;case"regex":nc(n,r.regex,r.message,t);break;case"cuid":nc(n,jh.cuid,r.message,t);break;case"cuid2":nc(n,jh.cuid2,r.message,t);break;case"startsWith":nc(n,RegExp(`^${qT(r.value,t)}`),r.message,t);break;case"endsWith":nc(n,RegExp(`${qT(r.value,t)}$`),r.message,t);break;case"datetime":Zh(n,"date-time",r.message,t);break;case"date":Zh(n,"date",r.message,t);break;case"time":Zh(n,"time",r.message,t);break;case"duration":Zh(n,"duration",r.message,t);break;case"length":Ki(n,"minLength",typeof n.minLength=="number"?Math.max(n.minLength,r.value):r.value,r.message,t),Ki(n,"maxLength",typeof n.maxLength=="number"?Math.min(n.maxLength,r.value):r.value,r.message,t);break;case"includes":{nc(n,RegExp(qT(r.value,t)),r.message,t);break}case"ip":{r.version!=="v6"&&Zh(n,"ipv4",r.message,t),r.version!=="v4"&&Zh(n,"ipv6",r.message,t);break}case"base64url":nc(n,jh.base64url,r.message,t);break;case"jwt":nc(n,jh.jwt,r.message,t);break;case"cidr":{r.version!=="v6"&&nc(n,jh.ipv4Cidr,r.message,t),r.version!=="v4"&&nc(n,jh.ipv6Cidr,r.message,t);break}case"emoji":nc(n,jh.emoji(),r.message,t);break;case"ulid":{nc(n,jh.ulid,r.message,t);break}case"base64":{switch(t.base64Strategy){case"format:binary":{Zh(n,"binary",r.message,t);break}case"contentEncoding:base64":{Ki(n,"contentEncoding","base64",r.message,t);break}case"pattern:zod":{nc(n,jh.base64,r.message,t);break}}break}case"nanoid":nc(n,jh.nanoid,r.message,t)}return n}function qT(e,t){return t.patternStrategy==="escape"?Xxe(e):e}const Hxe=new Set("ABCDEFGHIJKLMNOPQRSTUVXYZabcdefghijklmnopqrstuvxyz0123456789");function Xxe(e){let t="";for(let n=0;n<e.length;n++)Hxe.has(e[n])||(t+="\\"),t+=e[n];return t}function Zh(e,t,n,r){var i;e.format||(i=e.anyOf)!=null&&i.some(a=>a.format)?(e.anyOf||(e.anyOf=[]),e.format&&(e.anyOf.push({format:e.format,...e.errorMessage&&r.errorMessages&&{errorMessage:{format:e.errorMessage.format}}}),delete e.format,e.errorMessage&&(delete e.errorMessage.format,Object.keys(e.errorMessage).length===0&&delete e.errorMessage)),e.anyOf.push({format:t,...n&&r.errorMessages&&{errorMessage:{format:n}}})):Ki(e,"format",t,n,r)}function nc(e,t,n,r){var i;e.pattern||(i=e.allOf)!=null&&i.some(a=>a.pattern)?(e.allOf||(e.allOf=[]),e.pattern&&(e.allOf.push({pattern:e.pattern,...e.errorMessage&&r.errorMessages&&{errorMessage:{pattern:e.errorMessage.pattern}}}),delete e.pattern,e.errorMessage&&(delete e.errorMessage.pattern,Object.keys(e.errorMessage).length===0&&delete e.errorMessage)),e.allOf.push({pattern:mU(t,r),...n&&r.errorMessages&&{errorMessage:{pattern:n}}})):Ki(e,"pattern",mU(t,r),n,r)}function mU(e,t){var o;if(!t.applyRegexFlags||!e.flags)return e.source;const n={i:e.flags.includes("i"),m:e.flags.includes("m"),s:e.flags.includes("s")},r=n.i?e.source.toLowerCase():e.source;let i="",a=!1,u=!1,l=!1;for(let s=0;s<r.length;s++){if(a){i+=r[s],a=!1;continue}if(n.i){if(u){if(r[s].match(/[a-z]/)){l?(i+=r[s],i+=`${r[s-2]}-${r[s]}`.toUpperCase(),l=!1):r[s+1]==="-"&&((o=r[s+2])!=null&&o.match(/[a-z]/))?(i+=r[s],l=!0):i+=`${r[s]}${r[s].toUpperCase()}`;continue}}else if(r[s].match(/[a-z]/)){i+=`[${r[s]}${r[s].toUpperCase()}]`;continue}}if(n.m){if(r[s]==="^"){i+=`(^|(?<=[\r
]))`;continue}else if(r[s]==="$"){i+=`($|(?=[\r
]))`;continue}}if(n.s&&r[s]==="."){i+=u?`${r[s]}\r
`:`[${r[s]}\r
]`;continue}i+=r[s],r[s]==="\\"?a=!0:u&&r[s]==="]"?u=!1:!u&&r[s]==="["&&(u=!0)}try{new RegExp(i)}catch{return console.warn(`Could not convert regex pattern at ${t.currentPath.join("/")} to a flag-independent form! Falling back to the flag-ignorant source`),e.source}return i}function jee(e,t){var r,i,a,u,l,o;if(t.target==="openAi"&&console.warn("Warning: OpenAI may not support records in schemas! Try an array of key-value pairs instead."),t.target==="openApi3"&&((r=e.keyType)==null?void 0:r._def.typeName)===It.ZodEnum)return{type:"object",required:e.keyType._def.values,properties:e.keyType._def.values.reduce((s,c)=>({...s,[c]:Fi(e.valueType._def,{...t,currentPath:[...t.currentPath,"properties",c]})??cf(t)}),{}),additionalProperties:t.rejectedAdditionalProperties};const n={type:"object",additionalProperties:Fi(e.valueType._def,{...t,currentPath:[...t.currentPath,"additionalProperties"]})??t.allowedAdditionalProperties};if(t.target==="openApi3")return n;if(((i=e.keyType)==null?void 0:i._def.typeName)===It.ZodString&&((a=e.keyType._def.checks)!=null&&a.length)){const{type:s,...c}=Uee(e.keyType._def,t);return{...n,propertyNames:c}}else{if(((u=e.keyType)==null?void 0:u._def.typeName)===It.ZodEnum)return{...n,propertyNames:{enum:e.keyType._def.values}};if(((l=e.keyType)==null?void 0:l._def.typeName)===It.ZodBranded&&e.keyType._def.type._def.typeName===It.ZodString&&((o=e.keyType._def.type._def.checks)!=null&&o.length)){const{type:s,...c}=qee(e.keyType._def,t);return{...n,propertyNames:c}}}return n}function Yxe(e,t){if(t.mapStrategy==="record")return jee(e,t);const n=Fi(e.keyType._def,{...t,currentPath:[...t.currentPath,"items","items","0"]})||cf(t),r=Fi(e.valueType._def,{...t,currentPath:[...t.currentPath,"items","items","1"]})||cf(t);return{type:"array",maxItems:125,items:{type:"array",items:[n,r],minItems:2,maxItems:2}}}function Gxe(e){const t=e.values,r=Object.keys(e.values).filter(a=>typeof t[t[a]]!="number").map(a=>t[a]),i=Array.from(new Set(r.map(a=>typeof a)));return{type:i.length===1?i[0]==="string"?"string":"number":["string","number"],enum:r}}function Kxe(e){return e.target==="openAi"?void 0:{not:cf({...e,currentPath:[...e.currentPath,"not"]})}}function Wxe(e){return e.target==="openApi3"?{enum:["null"],nullable:!0}:{type:"null"}}const HD={ZodString:"string",ZodNumber:"number",ZodBigInt:"integer",ZodBoolean:"boolean",ZodNull:"null"};function Jxe(e,t){if(t.target==="openApi3")return gU(e,t);const n=e.options instanceof Map?Array.from(e.options.values()):e.options;if(n.every(r=>r._def.typeName in HD&&(!r._def.checks||!r._def.checks.length))){const r=n.reduce((i,a)=>{const u=HD[a._def.typeName];return u&&!i.includes(u)?[...i,u]:i},[]);return{type:r.length>1?r:r[0]}}else if(n.every(r=>r._def.typeName==="ZodLiteral"&&!r.description)){const r=n.reduce((i,a)=>{const u=typeof a._def.value;switch(u){case"string":case"number":case"boolean":return[...i,u];case"bigint":return[...i,"integer"];case"object":if(a._def.value===null)return[...i,"null"];case"symbol":case"undefined":case"function":default:return i}},[]);if(r.length===n.length){const i=r.filter((a,u,l)=>l.indexOf(a)===u);return{type:i.length>1?i:i[0],enum:n.reduce((a,u)=>a.includes(u._def.value)?a:[...a,u._def.value],[])}}}else if(n.every(r=>r._def.typeName==="ZodEnum"))return{type:"string",enum:n.reduce((r,i)=>[...r,...i._def.values.filter(a=>!r.includes(a))],[])};return gU(e,t)}const gU=(e,t)=>{const n=(e.options instanceof Map?Array.from(e.options.values()):e.options).map((r,i)=>Fi(r._def,{...t,currentPath:[...t.currentPath,"anyOf",`${i}`]})).filter(r=>!!r&&(!t.strictUnions||typeof r=="object"&&Object.keys(r).length>0));return n.length?{anyOf:n}:void 0};function eSe(e,t){if(["ZodString","ZodNumber","ZodBigInt","ZodBoolean","ZodNull"].includes(e.innerType._def.typeName)&&(!e.innerType._def.checks||!e.innerType._def.checks.length))return t.target==="openApi3"?{type:HD[e.innerType._def.typeName],nullable:!0}:{type:[HD[e.innerType._def.typeName],"null"]};if(t.target==="openApi3"){const r=Fi(e.innerType._def,{...t,currentPath:[...t.currentPath]});return r&&"$ref"in r?{allOf:[r],nullable:!0}:r&&{...r,nullable:!0}}const n=Fi(e.innerType._def,{...t,currentPath:[...t.currentPath,"anyOf","0"]});return n&&{anyOf:[n,{type:"null"}]}}function tSe(e,t){const n={type:"number"};if(!e.checks)return n;for(const r of e.checks)switch(r.kind){case"int":n.type="integer",Qee(n,"type",r.message,t);break;case"min":t.target==="jsonSchema7"?r.inclusive?Ki(n,"minimum",r.value,r.message,t):Ki(n,"exclusiveMinimum",r.value,r.message,t):(r.inclusive||(n.exclusiveMinimum=!0),Ki(n,"minimum",r.value,r.message,t));break;case"max":t.target==="jsonSchema7"?r.inclusive?Ki(n,"maximum",r.value,r.message,t):Ki(n,"exclusiveMaximum",r.value,r.message,t):(r.inclusive||(n.exclusiveMaximum=!0),Ki(n,"maximum",r.value,r.message,t));break;case"multipleOf":Ki(n,"multipleOf",r.value,r.message,t);break}return n}function nSe(e,t){const n=t.target==="openAi",r={type:"object",properties:{}},i=[],a=e.shape();for(const l in a){let o=a[l];if(o===void 0||o._def===void 0)continue;let s=iSe(o);s&&n&&(o._def.typeName==="ZodOptional"&&(o=o._def.innerType),o.isNullable()||(o=o.nullable()),s=!1);const c=Fi(o._def,{...t,currentPath:[...t.currentPath,"properties",l],propertyPath:[...t.currentPath,"properties",l]});c!==void 0&&(r.properties[l]=c,s||i.push(l))}i.length&&(r.required=i);const u=rSe(e,t);return u!==void 0&&(r.additionalProperties=u),r}function rSe(e,t){if(e.catchall._def.typeName!=="ZodNever")return Fi(e.catchall._def,{...t,currentPath:[...t.currentPath,"additionalProperties"]});switch(e.unknownKeys){case"passthrough":return t.allowedAdditionalProperties;case"strict":return t.rejectedAdditionalProperties;case"strip":return t.removeAdditionalStrategy==="strict"?t.allowedAdditionalProperties:t.rejectedAdditionalProperties}}function iSe(e){try{return e.isOptional()}catch{return!0}}const aSe=(e,t)=>{var r;if(t.currentPath.toString()===((r=t.propertyPath)==null?void 0:r.toString()))return Fi(e.innerType._def,t);const n=Fi(e.innerType._def,{...t,currentPath:[...t.currentPath,"anyOf","1"]});return n?{anyOf:[{not:cf(t)},n]}:cf(t)},uSe=(e,t)=>{if(t.pipeStrategy==="input")return Fi(e.in._def,t);if(t.pipeStrategy==="output")return Fi(e.out._def,t);const n=Fi(e.in._def,{...t,currentPath:[...t.currentPath,"allOf","0"]}),r=Fi(e.out._def,{...t,currentPath:[...t.currentPath,"allOf",n?"1":"0"]});return{allOf:[n,r].filter(i=>i!==void 0)}};function lSe(e,t){return Fi(e.type._def,t)}function oSe(e,t){const r={type:"array",uniqueItems:!0,items:Fi(e.valueType._def,{...t,currentPath:[...t.currentPath,"items"]})};return e.minSize&&Ki(r,"minItems",e.minSize.value,e.minSize.message,t),e.maxSize&&Ki(r,"maxItems",e.maxSize.value,e.maxSize.message,t),r}function sSe(e,t){return e.rest?{type:"array",minItems:e.items.length,items:e.items.map((n,r)=>Fi(n._def,{...t,currentPath:[...t.currentPath,"items",`${r}`]})).reduce((n,r)=>r===void 0?n:[...n,r],[]),additionalItems:Fi(e.rest._def,{...t,currentPath:[...t.currentPath,"additionalItems"]})}:{type:"array",minItems:e.items.length,maxItems:e.items.length,items:e.items.map((n,r)=>Fi(n._def,{...t,currentPath:[...t.currentPath,"items",`${r}`]})).reduce((n,r)=>r===void 0?n:[...n,r],[])}}function cSe(e){return{not:cf(e)}}function fSe(e){return cf(e)}const dSe=(e,t)=>Fi(e.innerType._def,t),hSe=(e,t,n)=>{switch(t){case It.ZodString:return Uee(e,n);case It.ZodNumber:return tSe(e,n);case It.ZodObject:return nSe(e,n);case It.ZodBigInt:return Ixe(e,n);case It.ZodBoolean:return Lxe();case It.ZodDate:return Vee(e,n);case It.ZodUndefined:return cSe(n);case It.ZodNull:return Wxe(n);case It.ZodArray:return Nxe(e,n);case It.ZodUnion:case It.ZodDiscriminatedUnion:return Jxe(e,n);case It.ZodIntersection:return jxe(e,n);case It.ZodTuple:return sSe(e,n);case It.ZodRecord:return jee(e,n);case It.ZodLiteral:return Zxe(e,n);case It.ZodEnum:return Vxe(e);case It.ZodNativeEnum:return Gxe(e);case It.ZodNullable:return eSe(e,n);case It.ZodOptional:return aSe(e,n);case It.ZodMap:return Yxe(e,n);case It.ZodSet:return oSe(e,n);case It.ZodLazy:return()=>e.getter()._def;case It.ZodPromise:return lSe(e,n);case It.ZodNaN:case It.ZodNever:return Kxe(n);case It.ZodEffects:return qxe(e,n);case It.ZodAny:return cf(n);case It.ZodUnknown:return fSe(n);case It.ZodDefault:return zxe(e,n);case It.ZodBranded:return qee(e,n);case It.ZodReadonly:return dSe(e,n);case It.ZodCatch:return Bxe(e,n);case It.ZodPipeline:return uSe(e,n);case It.ZodFunction:case It.ZodVoid:case It.ZodSymbol:return;default:return(r=>{})()}};function Fi(e,t,n=!1){var l;const r=t.seen.get(e);if(t.override){const o=(l=t.override)==null?void 0:l.call(t,e,t,r,n);if(o!==Pxe)return o}if(r&&!n){const o=pSe(r,t);if(o!==void 0)return o}const i={def:e,path:t.currentPath,jsonSchema:void 0};t.seen.set(e,i);const a=hSe(e,e.typeName,t),u=typeof a=="function"?Fi(a(),t):a;if(u&&mSe(e,t,u),t.postProcess){const o=t.postProcess(u,e,t);return i.jsonSchema=u,o}return i.jsonSchema=u,u}const pSe=(e,t)=>{switch(t.$refStrategy){case"root":return{$ref:e.path.join("/")};case"relative":return{$ref:zee(t.currentPath,e.path)};case"none":case"seen":return e.path.length<t.currentPath.length&&e.path.every((n,r)=>t.currentPath[r]===n)?(console.warn(`Recursive reference detected at ${t.currentPath.join("/")}! Defaulting to any`),cf(t)):t.$refStrategy==="seen"?cf(t):void 0}},mSe=(e,t,n)=>(e.description&&(n.description=e.description,t.markdownDescription&&(n.markdownDescription=e.description)),n),Urt=(e,t)=>{const n=Fxe(t);let r=typeof t=="object"&&t.definitions?Object.entries(t.definitions).reduce((o,[s,c])=>({...o,[s]:Fi(c._def,{...n,currentPath:[...n.basePath,n.definitionPath,s]},!0)??cf(n)}),{}):void 0;const i=typeof t=="string"?t:(t==null?void 0:t.nameStrategy)==="title"||t==null?void 0:t.name,a=Fi(e._def,i===void 0?n:{...n,currentPath:[...n.basePath,n.definitionPath,i]},!1)??cf(n),u=typeof t=="object"&&t.name!==void 0&&t.nameStrategy==="title"?t.name:void 0;u!==void 0&&(a.title=u),n.flags.hasReferencedOpenAiAnyType&&(r||(r={}),r[n.openAiAnyTypeName]||(r[n.openAiAnyTypeName]={type:["string","number","integer","boolean","array","null"],items:{$ref:n.$refStrategy==="relative"?"1":[...n.basePath,n.definitionPath,n.openAiAnyTypeName].join("/")}}));const l=i===void 0?r?{...a,[n.definitionPath]:r}:a:{$ref:[...n.$refStrategy==="relative"?[]:n.basePath,n.definitionPath,i].join("/"),[n.definitionPath]:{...r,[i]:a}};return n.target==="jsonSchema7"?l.$schema="http://json-schema.org/draft-07/schema#":(n.target==="jsonSchema2019-09"||n.target==="openAi")&&(l.$schema="https://json-schema.org/draft/2019-09/schema#"),n.target==="openAi"&&("anyOf"in l||"oneOf"in l||"allOf"in l||"type"in l&&Array.isArray(l.type))&&console.warn("Warning: OpenAI may not support schemas with unions as roots! Try wrapping it in an object property."),l},jrt={output:"output",llm:"llm",retrieval:"retrieval",reranker:"reranker",message:"message",document:"document",embedding:"embedding",tool:"tool",tool_call:"tool_call",user:"user"},Zrt={provider:"provider",model_name:"model_name",input_messages:"input_messages",output_messages:"output_messages",invocation_parameters:"invocation_parameters",prompts:"prompts",prompt_template:"prompt_template",tools:"tools"},Hrt={documents:"documents"},Xrt={input_documents:"input_documents",output_documents:"output_documents",query:"query"},Yrt={embeddings:"embeddings",text:"text",model_name:"model_name"},Grt={name:"name",description:"description",parameters:"parameters",json_schema:"json_schema"},Krt={role:"role",content:"content",contents:"contents",name:"name",function_call_name:"function_call_name",function_call_arguments_json:"function_call_arguments_json",tool_calls:"tool_calls",tool_call_id:"tool_call_id"},Wrt={id:"id",content:"content",score:"score",metadata:"metadata"},Jrt={id:"id"};var vU;(function(e){e.LLM="LLM",e.CHAIN="CHAIN",e.TOOL="TOOL",e.RETRIEVER="RETRIEVER",e.RERANKER="RERANKER",e.EMBEDDING="EMBEDDING",e.AGENT="AGENT",e.GUARDRAIL="GUARDRAIL",e.EVALUATOR="EVALUATOR"})(vU||(vU={}));var yU;(function(e){e.TEXT="text/plain",e.JSON="application/json",e.AUDIO_WAV="audio/wav"})(yU||(yU={}));var bU;(function(e){e.OPENAI="openai",e.ANTHROPIC="anthropic",e.MISTRALAI="mistralai",e.COHERE="cohere",e.VERTEXAI="vertexai"})(bU||(bU={}));var OU;(function(e){e.OPENAI="openai",e.ANTHROPIC="anthropic",e.MISTRALAI="mistralai",e.COHERE="cohere",e.GOOGLE="google",e.AWS="aws",e.AZURE="azure"})(OU||(OU={}));function x9(){return x9=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)({}).hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},x9.apply(null,arguments)}var Zee=["shift","alt","meta","mod","ctrl"],gSe={esc:"escape",return:"enter",".":"period",",":"comma","-":"slash"," ":"space","`":"backquote","#":"backslash","+":"bracketright",ShiftLeft:"shift",ShiftRight:"shift",AltLeft:"alt",AltRight:"alt",MetaLeft:"meta",MetaRight:"meta",OSLeft:"meta",OSRight:"meta",ControlLeft:"ctrl",ControlRight:"ctrl"};function Mv(e){return(e&&gSe[e]||e||"").trim().toLowerCase().replace(/key|digit|numpad|arrow/,"")}function vSe(e){return Zee.includes(e)}function VT(e,t){return t===void 0&&(t=","),e.split(t)}function UT(e,t,n){t===void 0&&(t="+");var r=e.toLocaleLowerCase().split(t).map(function(u){return Mv(u)}),i={alt:r.includes("alt"),ctrl:r.includes("ctrl")||r.includes("control"),shift:r.includes("shift"),meta:r.includes("meta"),mod:r.includes("mod")},a=r.filter(function(u){return!Zee.includes(u)});return x9({},i,{keys:a,description:n,hotkey:e})}(function(){typeof document<"u"&&(document.addEventListener("keydown",function(e){e.key!==void 0&&Hee([Mv(e.key),Mv(e.code)])}),document.addEventListener("keyup",function(e){e.key!==void 0&&Xee([Mv(e.key),Mv(e.code)])})),typeof window<"u"&&window.addEventListener("blur",function(){Fv.clear()})})();var Fv=new Set;function sN(e){return Array.isArray(e)}function ySe(e,t){t===void 0&&(t=",");var n=sN(e)?e:e.split(t);return n.every(function(r){return Fv.has(r.trim().toLowerCase())})}function Hee(e){var t=Array.isArray(e)?e:[e];Fv.has("meta")&&Fv.forEach(function(n){return!vSe(n)&&Fv.delete(n.toLowerCase())}),t.forEach(function(n){return Fv.add(n.toLowerCase())})}function Xee(e){var t=Array.isArray(e)?e:[e];e==="meta"?Fv.clear():t.forEach(function(n){return Fv.delete(n.toLowerCase())})}function bSe(e,t,n){(typeof n=="function"&&n(e,t)||n===!0)&&e.preventDefault()}function OSe(e,t,n){return typeof n=="function"?n(e,t):n===!0||n===void 0}function xSe(e){return Yee(e,["input","textarea","select"])}function Yee(e,t){t===void 0&&(t=!1);var n=e.target,r=e.composed,i=null;return SSe(n)&&r?i=e.composedPath()[0]&&e.composedPath()[0].tagName:i=n&&n.tagName,sN(t)?!!(i&&t&&t.some(function(a){var u;return a.toLowerCase()===((u=i)==null?void 0:u.toLowerCase())})):!!(i&&t&&t)}function SSe(e){return!!e.tagName&&!e.tagName.startsWith("-")&&e.tagName.includes("-")}function _Se(e,t){return e.length===0&&t?(console.warn('A hotkey has the "scopes" option set, however no active scopes were found. If you want to use the global scopes feature, you need to wrap your app in a <HotkeysProvider>'),!0):t?e.some(function(n){return t.includes(n)})||e.includes("*"):!0}var ESe=function(t,n,r){r===void 0&&(r=!1);var i=n.alt,a=n.meta,u=n.mod,l=n.shift,o=n.ctrl,s=n.keys,c=t.key,d=t.code,h=t.ctrlKey,m=t.metaKey,y=t.shiftKey,v=t.altKey,S=Mv(d),_=c.toLowerCase();if(!(s!=null&&s.includes(S))&&!(s!=null&&s.includes(_))&&!["ctrl","control","unknown","meta","alt","shift","os"].includes(S))return!1;if(!r){if(i===!v&&_!=="alt"||l===!y&&_!=="shift")return!1;if(u){if(!m&&!h)return!1}else if(a===!m&&_!=="meta"&&_!=="os"||o===!h&&_!=="ctrl"&&_!=="control")return!1}return s&&s.length===1&&(s.includes(_)||s.includes(S))?!0:s?ySe(s):!s},DSe=P.createContext(void 0),$Se=function(){return P.useContext(DSe)};function Gee(e,t){return e&&t&&typeof e=="object"&&typeof t=="object"?Object.keys(e).length===Object.keys(t).length&&Object.keys(e).reduce(function(n,r){return n&&Gee(e[r],t[r])},!0):e===t}var CSe=P.createContext({hotkeys:[],enabledScopes:[],toggleScope:function(){},enableScope:function(){},disableScope:function(){}}),wSe=function(){return P.useContext(CSe)};function RSe(e){var t=P.useRef(void 0);return Gee(t.current,e)||(t.current=e),t.current}var xU=function(t){t.stopPropagation(),t.preventDefault(),t.stopImmediatePropagation()},ASe=typeof window<"u"?P.useLayoutEffect:P.useEffect;function eit(e,t,n,r){var i=P.useState(null),a=i[0],u=i[1],l=P.useRef(!1),o=n instanceof Array?r instanceof Array?void 0:r:n,s=sN(e)?e.join(o==null?void 0:o.splitKey):e,c=n instanceof Array?n:r instanceof Array?r:void 0,d=P.useCallback(t,c??[]),h=P.useRef(d);c?h.current=d:h.current=t;var m=RSe(o),y=wSe(),v=y.enabledScopes,S=$Se();return ASe(function(){if(!((m==null?void 0:m.enabled)===!1||!_Se(v,m==null?void 0:m.scopes))){var _=function(w,R){var F;if(R===void 0&&(R=!1),!(xSe(w)&&!Yee(w,m==null?void 0:m.enableOnFormTags))){if(a!==null){var I=a.getRootNode();if((I instanceof Document||I instanceof ShadowRoot)&&I.activeElement!==a&&!a.contains(I.activeElement)){xU(w);return}}(F=w.target)!=null&&F.isContentEditable&&!(m!=null&&m.enableOnContentEditable)||VT(s,m==null?void 0:m.splitKey).forEach(function(U){var q,V=UT(U,m==null?void 0:m.combinationKey);if(ESe(w,V,m==null?void 0:m.ignoreModifiers)||(q=V.keys)!=null&&q.includes("*")){if(m!=null&&m.ignoreEventWhen!=null&&m.ignoreEventWhen(w)||R&&l.current)return;if(bSe(w,V,m==null?void 0:m.preventDefault),!OSe(w,V,m==null?void 0:m.enabled)){xU(w);return}h.current(w,V),R||(l.current=!0)}})}},E=function(w){w.key!==void 0&&(Hee(Mv(w.code)),((m==null?void 0:m.keydown)===void 0&&(m==null?void 0:m.keyup)!==!0||m!=null&&m.keydown)&&_(w))},$=function(w){w.key!==void 0&&(Xee(Mv(w.code)),l.current=!1,m!=null&&m.keyup&&_(w,!0))},k=a||(o==null?void 0:o.document)||document;return k.addEventListener("keyup",$,o==null?void 0:o.eventListenerOptions),k.addEventListener("keydown",E,o==null?void 0:o.eventListenerOptions),S&&VT(s,m==null?void 0:m.splitKey).forEach(function(A){return S.addHotkey(UT(A,m==null?void 0:m.combinationKey,m==null?void 0:m.description))}),function(){k.removeEventListener("keyup",$,o==null?void 0:o.eventListenerOptions),k.removeEventListener("keydown",E,o==null?void 0:o.eventListenerOptions),S&&VT(s,m==null?void 0:m.splitKey).forEach(function(A){return S.removeHotkey(UT(A,m==null?void 0:m.combinationKey,m==null?void 0:m.description))})}}},[a,s,m,v]),u}const TSe=P.createContext(null);P.createContext(null);const kSe=P.createContext(null);P.createContext(null);P.createContext(null);const Kee=P.createContext({}),vx=P.createContext(null),LC=P.createContext(null),Hr=typeof document<"u"?be.useLayoutEffect:()=>{};var jT;const PSe=(jT=be.useInsertionEffect)!==null&&jT!==void 0?jT:Hr;function fi(e){const t=P.useRef(null);return PSe(()=>{t.current=e},[e]),P.useCallback((...n)=>{const r=t.current;return r==null?void 0:r(...n)},[])}function MSe(e){let[t,n]=P.useState(e),r=P.useRef(null),i=fi(()=>{if(!r.current)return;let u=r.current.next();if(u.done){r.current=null;return}t===u.value?i():n(u.value)});Hr(()=>{r.current&&i()});let a=fi(u=>{r.current=u(t),i()});return[t,a]}const Wee={prefix:String(Math.round(Math.random()*1e10)),current:0},Jee=be.createContext(Wee),FSe=be.createContext(!1);let ZT=new WeakMap;function NSe(e=!1){let t=P.useContext(Jee),n=P.useRef(null);if(n.current===null&&!e){var r,i;let a=(i=be.__SECRET_INTERNALS_DO_NOT_USE_OR_YOU_WILL_BE_FIRED)===null||i===void 0||(r=i.ReactCurrentOwner)===null||r===void 0?void 0:r.current;if(a){let u=ZT.get(a);u==null?ZT.set(a,{id:t.current,state:a.memoizedState}):a.memoizedState!==u.state&&(t.current=u.id,ZT.delete(a))}n.current=++t.current}return n.current}function ISe(e){let t=P.useContext(Jee),n=NSe(!!e),r=`react-aria${t.prefix}`;return e||`${r}-${n}`}function LSe(e){let t=be.useId(),[n]=P.useState(pp()),r=n?"react-aria":`react-aria${Wee.prefix}`;return e||`${r}-${t}`}const BSe=typeof be.useId=="function"?LSe:ISe;function QSe(){return!1}function zSe(){return!0}function qSe(e){return()=>{}}function pp(){return typeof be.useSyncExternalStore=="function"?be.useSyncExternalStore(qSe,QSe,zSe):P.useContext(FSe)}let VSe=!!(typeof window<"u"&&window.document&&window.document.createElement),r3=new Map,P2;typeof FinalizationRegistry<"u"&&(P2=new FinalizationRegistry(e=>{r3.delete(e)}));function vi(e){let[t,n]=P.useState(e),r=P.useRef(null),i=BSe(t),a=P.useRef(null);if(P2&&P2.register(a,i),VSe){const u=r3.get(i);u&&!u.includes(r)?u.push(r):r3.set(i,[r])}return Hr(()=>{let u=i;return()=>{P2&&P2.unregister(a),r3.delete(u)}},[i]),P.useEffect(()=>{let u=r.current;return u&&n(u),()=>{u&&(r.current=null)}}),i}function USe(e,t){if(e===t)return e;let n=r3.get(e);if(n)return n.forEach(i=>i.current=t),t;let r=r3.get(t);return r?(r.forEach(i=>i.current=e),e):t}function $0(e=[]){let t=vi(),[n,r]=MSe(t),i=P.useCallback(()=>{r(function*(){yield t,yield document.getElementById(t)?t:void 0})},[t,r]);return Hr(i,[t,i,...e]),n}function ih(...e){return(...t)=>{for(let n of e)typeof n=="function"&&n(...t)}}const hi=e=>{var t;return(t=e==null?void 0:e.ownerDocument)!==null&&t!==void 0?t:document},uf=e=>e&&"window"in e&&e.window===e?e:hi(e).defaultView||window;function jSe(e){return e!==null&&typeof e=="object"&&"nodeType"in e&&typeof e.nodeType=="number"}function ZSe(e){return jSe(e)&&e.nodeType===Node.DOCUMENT_FRAGMENT_NODE&&"host"in e}let HSe=!1;function BC(){return HSe}function Wc(e,t){if(!BC())return t&&e?e.contains(t):!1;if(!e||!t)return!1;let n=t;for(;n!==null;){if(n===e)return!0;n.tagName==="SLOT"&&n.assignedSlot?n=n.assignedSlot.parentNode:ZSe(n)?n=n.host:n=n.parentNode}return!1}const Zu=(e=document)=>{var t;if(!BC())return e.activeElement;let n=e.activeElement;for(;n&&"shadowRoot"in n&&(!((t=n.shadowRoot)===null||t===void 0)&&t.activeElement);)n=n.shadowRoot.activeElement;return n};function ju(e){return BC()&&e.target.shadowRoot&&e.composedPath?e.composedPath()[0]:e.target}class XSe{get currentNode(){return this._currentNode}set currentNode(t){if(!Wc(this.root,t))throw new Error("Cannot set currentNode to a node that is not contained by the root node.");const n=[];let r=t,i=t;for(this._currentNode=t;r&&r!==this.root;)if(r.nodeType===Node.DOCUMENT_FRAGMENT_NODE){const u=r,l=this._doc.createTreeWalker(u,this.whatToShow,{acceptNode:this._acceptNode});n.push(l),l.currentNode=i,this._currentSetFor.add(l),r=i=u.host}else r=r.parentNode;const a=this._doc.createTreeWalker(this.root,this.whatToShow,{acceptNode:this._acceptNode});n.push(a),a.currentNode=i,this._currentSetFor.add(a),this._walkerStack=n}get doc(){return this._doc}firstChild(){let t=this.currentNode,n=this.nextNode();return Wc(t,n)?(n&&(this.currentNode=n),n):(this.currentNode=t,null)}lastChild(){let n=this._walkerStack[0].lastChild();return n&&(this.currentNode=n),n}nextNode(){const t=this._walkerStack[0].nextNode();if(t){if(t.shadowRoot){var n;let i;if(typeof this.filter=="function"?i=this.filter(t):!((n=this.filter)===null||n===void 0)&&n.acceptNode&&(i=this.filter.acceptNode(t)),i===NodeFilter.FILTER_ACCEPT)return this.currentNode=t,t;let a=this.nextNode();return a&&(this.currentNode=a),a}return t&&(this.currentNode=t),t}else if(this._walkerStack.length>1){this._walkerStack.shift();let r=this.nextNode();return r&&(this.currentNode=r),r}else return null}previousNode(){const t=this._walkerStack[0];if(t.currentNode===t.root){if(this._currentSetFor.has(t))if(this._currentSetFor.delete(t),this._walkerStack.length>1){this._walkerStack.shift();let i=this.previousNode();return i&&(this.currentNode=i),i}else return null;return null}const n=t.previousNode();if(n){if(n.shadowRoot){var r;let a;if(typeof this.filter=="function"?a=this.filter(n):!((r=this.filter)===null||r===void 0)&&r.acceptNode&&(a=this.filter.acceptNode(n)),a===NodeFilter.FILTER_ACCEPT)return n&&(this.currentNode=n),n;let u=this.lastChild();return u&&(this.currentNode=u),u}return n&&(this.currentNode=n),n}else if(this._walkerStack.length>1){this._walkerStack.shift();let i=this.previousNode();return i&&(this.currentNode=i),i}else return null}nextSibling(){return null}previousSibling(){return null}parentNode(){return null}constructor(t,n,r,i){this._walkerStack=[],this._currentSetFor=new Set,this._acceptNode=u=>{if(u.nodeType===Node.ELEMENT_NODE){const o=u.shadowRoot;if(o){const s=this._doc.createTreeWalker(o,this.whatToShow,{acceptNode:this._acceptNode});return this._walkerStack.unshift(s),NodeFilter.FILTER_ACCEPT}else{var l;if(typeof this.filter=="function")return this.filter(u);if(!((l=this.filter)===null||l===void 0)&&l.acceptNode)return this.filter.acceptNode(u);if(this.filter===null)return NodeFilter.FILTER_ACCEPT}}return NodeFilter.FILTER_SKIP},this._doc=t,this.root=n,this.filter=i??null,this.whatToShow=r??NodeFilter.SHOW_ALL,this._currentNode=n,this._walkerStack.unshift(t.createTreeWalker(n,r,this._acceptNode));const a=n.shadowRoot;if(a){const u=this._doc.createTreeWalker(a,this.whatToShow,{acceptNode:this._acceptNode});this._walkerStack.unshift(u)}}}function YSe(e,t,n,r){return BC()?new XSe(e,t,n,r):e.createTreeWalker(t,n,r)}function ete(e){var t,n,r="";if(typeof e=="string"||typeof e=="number")r+=e;else if(typeof e=="object")if(Array.isArray(e)){var i=e.length;for(t=0;t<i;t++)e[t]&&(n=ete(e[t]))&&(r&&(r+=" "),r+=n)}else for(n in e)e[n]&&(r&&(r+=" "),r+=n);return r}function GSe(){for(var e,t,n=0,r="",i=arguments.length;n<i;n++)(e=arguments[n])&&(t=ete(e))&&(r&&(r+=" "),r+=t);return r}function Et(...e){let t={...e[0]};for(let n=1;n<e.length;n++){let r=e[n];for(let i in r){let a=t[i],u=r[i];typeof a=="function"&&typeof u=="function"&&i[0]==="o"&&i[1]==="n"&&i.charCodeAt(2)>=65&&i.charCodeAt(2)<=90?t[i]=ih(a,u):(i==="className"||i==="UNSAFE_className")&&typeof a=="string"&&typeof u=="string"?t[i]=GSe(a,u):i==="id"&&a&&u?t.id=USe(a,u):t[i]=u!==void 0?u:a}}return t}function yg(...e){return e.length===1&&e[0]?e[0]:t=>{let n=!1;const r=e.map(i=>{const a=SU(i,t);return n||(n=typeof a=="function"),a});if(n)return()=>{r.forEach((i,a)=>{typeof i=="function"?i():SU(e[a],null)})}}}function SU(e,t){if(typeof e=="function")return e(t);e!=null&&(e.current=t)}const KSe=new Set(["id"]),WSe=new Set(["aria-label","aria-labelledby","aria-describedby","aria-details"]),JSe=new Set(["href","hrefLang","target","rel","download","ping","referrerPolicy"]),e_e=new Set(["dir","lang","hidden","inert","translate"]),_U=new Set(["onClick","onAuxClick","onContextMenu","onDoubleClick","onMouseDown","onMouseEnter","onMouseLeave","onMouseMove","onMouseOut","onMouseOver","onMouseUp","onTouchCancel","onTouchEnd","onTouchMove","onTouchStart","onPointerDown","onPointerMove","onPointerUp","onPointerCancel","onPointerEnter","onPointerLeave","onPointerOver","onPointerOut","onGotPointerCapture","onLostPointerCapture","onScroll","onWheel","onAnimationStart","onAnimationEnd","onAnimationIteration","onTransitionCancel","onTransitionEnd","onTransitionRun","onTransitionStart"]),t_e=/^(data-.*)$/;function Gt(e,t={}){let{labelable:n,isLink:r,global:i,events:a=i,propNames:u}=t,l={};for(const o in e)Object.prototype.hasOwnProperty.call(e,o)&&(KSe.has(o)||n&&WSe.has(o)||r&&JSe.has(o)||i&&e_e.has(o)||a&&_U.has(o)||o.endsWith("Capture")&&_U.has(o.slice(0,-7))||u!=null&&u.has(o)||t_e.test(o))&&(l[o]=e[o]);return l}function $u(e){if(n_e())e.focus({preventScroll:!0});else{let t=r_e(e);e.focus(),i_e(t)}}let z_=null;function n_e(){if(z_==null){z_=!1;try{document.createElement("div").focus({get preventScroll(){return z_=!0,!0}})}catch{}}return z_}function r_e(e){let t=e.parentNode,n=[],r=document.scrollingElement||document.documentElement;for(;t instanceof HTMLElement&&t!==r;)(t.offsetHeight<t.scrollHeight||t.offsetWidth<t.scrollWidth)&&n.push({element:t,scrollTop:t.scrollTop,scrollLeft:t.scrollLeft}),t=t.parentNode;return r instanceof HTMLElement&&n.push({element:r,scrollTop:r.scrollTop,scrollLeft:r.scrollLeft}),n}function i_e(e){for(let{element:t,scrollTop:n,scrollLeft:r}of e)t.scrollTop=n,t.scrollLeft=r}function QC(e){var t;if(typeof window>"u"||window.navigator==null)return!1;let n=(t=window.navigator.userAgentData)===null||t===void 0?void 0:t.brands;return Array.isArray(n)&&n.some(r=>e.test(r.brand))||e.test(window.navigator.userAgent)}function cN(e){var t;return typeof window<"u"&&window.navigator!=null?e.test(((t=window.navigator.userAgentData)===null||t===void 0?void 0:t.platform)||window.navigator.platform):!1}function bg(e){let t=null;return()=>(t==null&&(t=e()),t)}const cg=bg(function(){return cN(/^Mac/i)}),tte=bg(function(){return cN(/^iPhone/i)}),nte=bg(function(){return cN(/^iPad/i)||cg()&&navigator.maxTouchPoints>1}),$1=bg(function(){return tte()||nte()}),JE=bg(function(){return cg()||$1()}),fN=bg(function(){return QC(/AppleWebKit/i)&&!dN()}),dN=bg(function(){return QC(/Chrome/i)}),fS=bg(function(){return QC(/Android/i)}),a_e=bg(function(){return QC(/Firefox/i)}),u_e=P.createContext({isNative:!0,open:s_e,useHref:e=>e});function X3(){return P.useContext(u_e)}function l_e(e,t){let n=e.getAttribute("target");return(!n||n==="_self")&&e.origin===location.origin&&!e.hasAttribute("download")&&!t.metaKey&&!t.ctrlKey&&!t.altKey&&!t.shiftKey}function a1(e,t,n=!0){var r,i;let{metaKey:a,ctrlKey:u,altKey:l,shiftKey:o}=t;a_e()&&(!((i=window.event)===null||i===void 0||(r=i.type)===null||r===void 0)&&r.startsWith("key"))&&e.target==="_blank"&&(cg()?a=!0:u=!0);let s=fN()&&cg()&&!nte()?new KeyboardEvent("keydown",{keyIdentifier:"Enter",metaKey:a,ctrlKey:u,altKey:l,shiftKey:o}):new MouseEvent("click",{metaKey:a,ctrlKey:u,altKey:l,shiftKey:o,bubbles:!0,cancelable:!0});a1.isOpening=n,$u(e),e.dispatchEvent(s),a1.isOpening=!1}a1.isOpening=!1;function o_e(e,t){if(e instanceof HTMLAnchorElement)t(e);else if(e.hasAttribute("data-href")){let n=document.createElement("a");n.href=e.getAttribute("data-href"),e.hasAttribute("data-target")&&(n.target=e.getAttribute("data-target")),e.hasAttribute("data-rel")&&(n.rel=e.getAttribute("data-rel")),e.hasAttribute("data-download")&&(n.download=e.getAttribute("data-download")),e.hasAttribute("data-ping")&&(n.ping=e.getAttribute("data-ping")),e.hasAttribute("data-referrer-policy")&&(n.referrerPolicy=e.getAttribute("data-referrer-policy")),e.appendChild(n),t(n),e.removeChild(n)}}function s_e(e,t){o_e(e,n=>a1(n,t))}function rte(e){let t=X3();var n;const r=t.useHref((n=e.href)!==null&&n!==void 0?n:"");return{"data-href":e.href?r:void 0,"data-target":e.target,"data-rel":e.rel,"data-download":e.download,"data-ping":e.ping,"data-referrer-policy":e.referrerPolicy}}function hN(e){let t=X3();var n;const r=t.useHref((n=e==null?void 0:e.href)!==null&&n!==void 0?n:"");return{href:e!=null&&e.href?r:void 0,target:e==null?void 0:e.target,rel:e==null?void 0:e.rel,download:e==null?void 0:e.download,ping:e==null?void 0:e.ping,referrerPolicy:e==null?void 0:e.referrerPolicy}}function c_e(e,t,n,r){!t.isNative&&e.currentTarget instanceof HTMLAnchorElement&&e.currentTarget.href&&!e.isDefaultPrevented()&&l_e(e.currentTarget,e)&&n&&(e.preventDefault(),t.open(e.currentTarget,e,n,r))}let wv=new Map,S9=new Set;function EU(){if(typeof window>"u")return;function e(r){return"propertyName"in r}let t=r=>{if(!e(r)||!r.target)return;let i=wv.get(r.target);i||(i=new Set,wv.set(r.target,i),r.target.addEventListener("transitioncancel",n,{once:!0})),i.add(r.propertyName)},n=r=>{if(!e(r)||!r.target)return;let i=wv.get(r.target);if(i&&(i.delete(r.propertyName),i.size===0&&(r.target.removeEventListener("transitioncancel",n),wv.delete(r.target)),wv.size===0)){for(let a of S9)a();S9.clear()}};document.body.addEventListener("transitionrun",t),document.body.addEventListener("transitionend",n)}typeof document<"u"&&(document.readyState!=="loading"?EU():document.addEventListener("DOMContentLoaded",EU));function f_e(){for(const[e]of wv)"isConnected"in e&&!e.isConnected&&wv.delete(e)}function ite(e){requestAnimationFrame(()=>{f_e(),wv.size===0?e():S9.add(e)})}function py(){let e=P.useRef(new Map),t=P.useCallback((i,a,u,l)=>{let o=l!=null&&l.once?(...s)=>{e.current.delete(u),u(...s)}:u;e.current.set(u,{type:a,eventTarget:i,fn:o,options:l}),i.addEventListener(a,o,l)},[]),n=P.useCallback((i,a,u,l)=>{var o;let s=((o=e.current.get(u))===null||o===void 0?void 0:o.fn)||u;i.removeEventListener(a,s,l),e.current.delete(u)},[]),r=P.useCallback(()=>{e.current.forEach((i,a)=>{n(i.eventTarget,i.type,a,i.options)})},[n]);return P.useEffect(()=>r,[r]),{addGlobalListener:t,removeGlobalListener:n,removeAllGlobalListeners:r}}function Gv(e,t){let{id:n,"aria-label":r,"aria-labelledby":i}=e;return n=vi(n),i&&r?i=[...new Set([n,...i.trim().split(/\s+/)])].join(" "):i&&(i=i.trim().split(/\s+/).join(" ")),!r&&!i&&t&&(r=t),{id:n,"aria-label":r,"aria-labelledby":i}}function wu(e){const t=P.useRef(null),n=P.useRef(void 0),r=P.useCallback(i=>{if(typeof e=="function"){const a=e,u=a(i);return()=>{typeof u=="function"?u():a(null)}}else if(e)return e.current=i,()=>{e.current=null}},[e]);return P.useMemo(()=>({get current(){return t.current},set current(i){t.current=i,n.current&&(n.current(),n.current=void 0),i!=null&&(n.current=r(i))}}),[r])}function ate(e,t){const n=P.useRef(!0),r=P.useRef(null);P.useEffect(()=>(n.current=!0,()=>{n.current=!1}),[]),P.useEffect(()=>{let i=r.current;n.current?n.current=!1:(!i||t.some((a,u)=>!Object.is(a,i[u])))&&e(),r.current=t},t)}function DU(e,t){const n=P.useRef(!0),r=P.useRef(null);Hr(()=>(n.current=!0,()=>{n.current=!1}),[]),Hr(()=>{n.current?n.current=!1:(!r.current||t.some((i,a)=>!Object.is(i,r[a])))&&e(),r.current=t},t)}function d_e(){return typeof window.ResizeObserver<"u"}function u1(e){const{ref:t,box:n,onResize:r}=e;P.useEffect(()=>{let i=t==null?void 0:t.current;if(i)if(d_e()){const a=new window.ResizeObserver(u=>{u.length&&r()});return a.observe(i,{box:n}),()=>{i&&a.unobserve(i)}}else return window.addEventListener("resize",r,!1),()=>{window.removeEventListener("resize",r,!1)}},[r,t,n])}function pN(e,t){Hr(()=>{if(e&&e.ref&&t)return e.ref.current=t.current,()=>{e.ref&&(e.ref.current=null)}})}function l1(e,t){if(!e)return!1;let n=window.getComputedStyle(e),r=/(auto|scroll)/.test(n.overflow+n.overflowX+n.overflowY);return r&&t&&(r=e.scrollHeight!==e.clientHeight||e.scrollWidth!==e.clientWidth),r}function m0(e,t){let n=e;for(l1(n,t)&&(n=n.parentElement);n&&!l1(n,t);)n=n.parentElement;return n||document.scrollingElement||document.documentElement}function h_e(e,t){const n=[];for(;e&&e!==document.documentElement;)l1(e,t)&&n.push(e),e=e.parentElement;return n}function Tb(e){return cg()?e.metaKey:e.ctrlKey}const p_e=new Set(["checkbox","radio","range","color","file","image","button","submit","reset"]);function j2(e){return e instanceof HTMLInputElement&&!p_e.has(e.type)||e instanceof HTMLTextAreaElement||e instanceof HTMLElement&&e.isContentEditable}let ef=typeof document<"u"&&window.visualViewport;function m_e(){let e=pp(),[t,n]=P.useState(()=>e?{width:0,height:0}:$U());return P.useEffect(()=>{let r=()=>{ef&&ef.scale>1||n(u=>{let l=$U();return l.width===u.width&&l.height===u.height?u:l})},i,a=u=>{ef&&ef.scale>1||j2(u.target)&&(i=requestAnimationFrame(()=>{(!document.activeElement||!j2(document.activeElement))&&n(l=>{let o={width:window.innerWidth,height:window.innerHeight};return o.width===l.width&&o.height===l.height?l:o})}))};return window.addEventListener("blur",a,!0),ef?ef.addEventListener("resize",r):window.addEventListener("resize",r),()=>{cancelAnimationFrame(i),window.removeEventListener("blur",a,!0),ef?ef.removeEventListener("resize",r):window.removeEventListener("resize",r)}},[]),t}function $U(){return{width:ef?ef.width*ef.scale:window.innerWidth,height:ef?ef.height*ef.scale:window.innerHeight}}let g_e=0;const HT=new Map;function zC(e){let[t,n]=P.useState();return Hr(()=>{if(!e)return;let r=HT.get(e);if(r)n(r.element.id);else{let i=`react-aria-description-${g_e++}`;n(i);let a=document.createElement("div");a.id=i,a.style.display="none",a.textContent=e,document.body.appendChild(a),r={refCount:0,element:a},HT.set(e,r)}return r.refCount++,()=>{r&&--r.refCount===0&&(r.element.remove(),HT.delete(e))}},[e]),{"aria-describedby":e?t:void 0}}function Yd(e,t,n,r){let i=fi(n),a=n==null;P.useEffect(()=>{if(a||!e.current)return;let u=e.current;return u.addEventListener(t,i,r),()=>{u.removeEventListener(t,i,r)}},[e,t,r,a,i])}function ute(e,t){let n=CU(e,t,"left"),r=CU(e,t,"top"),i=t.offsetWidth,a=t.offsetHeight,u=e.scrollLeft,l=e.scrollTop,{borderTopWidth:o,borderLeftWidth:s,scrollPaddingTop:c,scrollPaddingRight:d,scrollPaddingBottom:h,scrollPaddingLeft:m}=getComputedStyle(e),{scrollMarginTop:y,scrollMarginRight:v,scrollMarginBottom:S,scrollMarginLeft:_}=getComputedStyle(t),E=u+parseInt(s,10),$=l+parseInt(o,10),k=E+e.clientWidth,A=$+e.clientHeight,w=parseInt(c,10)||0,R=parseInt(h,10)||0,F=parseInt(d,10)||0,I=parseInt(m,10)||0,U=parseInt(y,10)||0,q=parseInt(S,10)||0,V=parseInt(v,10)||0,N=parseInt(_,10)||0,j=n-N,W=n+i+V,K=r-U,J=r+a+q,te=u+parseInt(s,10)+I,H=k-F,M=l+parseInt(o,10)+w,G=A-R;(j>te||W<H)&&(j<=u+I?u=j-parseInt(s,10)-I:W>k-F&&(u+=W-k+F)),(K>M||J<G)&&(K<=$+w?l=K-parseInt(o,10)-w:J>A-R&&(l+=J-A+R)),e.scrollTo({left:u,top:l})}function CU(e,t,n){const r=n==="left"?"offsetLeft":"offsetTop";let i=0;for(;t.offsetParent&&(i+=t[r],t.offsetParent!==e);){if(t.offsetParent.contains(e)){i-=e[r];break}t=t.offsetParent}return i}function Vm(e,t){if(e&&document.contains(e)){let u=document.scrollingElement||document.documentElement;if(!(window.getComputedStyle(u).overflow==="hidden")&&!dN()){var n;let{left:o,top:s}=e.getBoundingClientRect();e==null||(n=e.scrollIntoView)===null||n===void 0||n.call(e,{block:"nearest"});let{left:c,top:d}=e.getBoundingClientRect();if(Math.abs(o-c)>1||Math.abs(s-d)>1){var r,i,a;t==null||(i=t.containingElement)===null||i===void 0||(r=i.scrollIntoView)===null||r===void 0||r.call(i,{block:"center",inline:"center"}),(a=e.scrollIntoView)===null||a===void 0||a.call(e,{block:"nearest"})}}else{let o=h_e(e);for(let s of o)ute(s,e)}}}function lte(e){return e.pointerType===""&&e.isTrusted?!0:fS()&&e.pointerType?e.type==="click"&&e.buttons===1:e.detail===0&&!e.pointerType}function v_e(e){return!fS()&&e.width===0&&e.height===0||e.width===1&&e.height===1&&e.pressure===0&&e.detail===0&&e.pointerType==="mouse"}function y_e(e,t){let n=P.useRef(null);return e&&n.current&&t(e,n.current)&&(e=n.current),n.current=e,e}function my(e,t,n){let r=fi(()=>{n&&n(t)});P.useEffect(()=>{var i;let a=e==null||(i=e.current)===null||i===void 0?void 0:i.form;return a==null||a.addEventListener("reset",r),()=>{a==null||a.removeEventListener("reset",r)}},[e,r])}function b_e(e,t){let{collection:n,onLoadMore:r,scrollOffset:i=1}=e,a=P.useRef(null),u=fi(l=>{for(let o of l)o.isIntersecting&&r&&r()});Hr(()=>(t.current&&(a.current=new IntersectionObserver(u,{root:m0(t==null?void 0:t.current),rootMargin:`0px ${100*i}% ${100*i}% ${100*i}%`}),a.current.observe(t.current)),()=>{a.current&&a.current.disconnect()}),[n,u,t,i])}function _9(e){const t=P.version.split(".");return parseInt(t[0],10)>=19?e:e?"true":void 0}const ote="react-aria-clear-focus",E9="react-aria-focus";var Uf=AJ();const ste=da(Uf);function qC(e,t=!0){let[n,r]=P.useState(!0),i=n&&t;return Hr(()=>{if(i&&e.current&&"getAnimations"in e.current)for(let a of e.current.getAnimations())a instanceof CSSTransition&&a.cancel()},[e,i]),cte(e,i,P.useCallback(()=>r(!1),[])),i}function XD(e,t){let[n,r]=P.useState(t?"open":"closed");switch(n){case"open":t||r("exiting");break;case"closed":case"exiting":t&&r("open");break}let i=n==="exiting";return cte(e,i,P.useCallback(()=>{r(a=>a==="exiting"?"closed":a)},[])),i}function cte(e,t,n){Hr(()=>{if(t&&e.current){if(!("getAnimations"in e.current)){n();return}let r=e.current.getAnimations();if(r.length===0){n();return}let i=!1;return Promise.all(r.map(a=>a.finished)).then(()=>{i||Uf.flushSync(()=>{n()})}).catch(()=>{}),()=>{i=!0}}},[e,t,n])}const O_e=typeof Element<"u"&&"checkVisibility"in Element.prototype;function x_e(e){const t=uf(e);if(!(e instanceof t.HTMLElement)&&!(e instanceof t.SVGElement))return!1;let{display:n,visibility:r}=e.style,i=n!=="none"&&r!=="hidden"&&r!=="collapse";if(i){const{getComputedStyle:a}=e.ownerDocument.defaultView;let{display:u,visibility:l}=a(e);i=u!=="none"&&l!=="hidden"&&l!=="collapse"}return i}function S_e(e,t){return!e.hasAttribute("hidden")&&!e.hasAttribute("data-react-aria-prevent-focus")&&(e.nodeName==="DETAILS"&&t&&t.nodeName!=="SUMMARY"?e.hasAttribute("open"):!0)}function mN(e,t){return O_e?e.checkVisibility({visibilityProperty:!0})&&!e.closest("[data-react-aria-prevent-focus]"):e.nodeName!=="#comment"&&x_e(e)&&S_e(e,t)&&(!e.parentElement||mN(e.parentElement,e))}const gN=["input:not([disabled]):not([type=hidden])","select:not([disabled])","textarea:not([disabled])","button:not([disabled])","a[href]","area[href]","summary","iframe","object","embed","audio[controls]","video[controls]",'[contenteditable]:not([contenteditable^="false"])',"permission"],__e=gN.join(":not([hidden]),")+",[tabindex]:not([disabled]):not([hidden])";gN.push('[tabindex]:not([tabindex="-1"]):not([disabled])');const E_e=gN.join(':not([hidden]):not([tabindex="-1"]),');function fte(e){return e.matches(__e)&&mN(e)&&!dte(e)}function D_e(e){return e.matches(E_e)&&mN(e)&&!dte(e)}function dte(e){let t=e;for(;t!=null;){if(t instanceof t.ownerDocument.defaultView.HTMLElement&&t.inert)return!0;t=t.parentElement}return!1}function Yu(e,t,n){let[r,i]=P.useState(e||t),a=P.useRef(e!==void 0),u=e!==void 0;P.useEffect(()=>{a.current,a.current=u},[u]);let l=u?e:r,o=P.useCallback((s,...c)=>{let d=(h,...m)=>{n&&(Object.is(l,h)||n(h,...m)),u||(l=h)};typeof s=="function"?i((m,...y)=>{let v=s(u?l:m,...y);return d(v,...c),u?m:v}):(u||i(s),d(s,...c))},[u,l,n]);return[l,o]}function so(e,t=-1/0,n=1/0){return Math.min(Math.max(e,t),n)}function q_(e,t){let n=e,r=0,i=t.toString(),a=i.toLowerCase().indexOf("e-");if(a>0)r=Math.abs(Math.floor(Math.log10(Math.abs(t))))+a;else{let u=i.indexOf(".");u>=0&&(r=i.length-u)}if(r>0){let u=Math.pow(10,r);n=Math.round(n*u)/u}return n}function Jc(e,t,n,r){t=Number(t),n=Number(n);let i=(e-(isNaN(t)?0:t))%r,a=q_(Math.abs(i)*2>=r?e+Math.sign(i)*(r-Math.abs(i)):e-i,r);return isNaN(t)?!isNaN(n)&&a>n&&(a=Math.floor(q_(n/r,r))*r):a<t?a=t:!isNaN(n)&&a>n&&(a=t+Math.floor(q_((n-t)/r,r))*r),a=q_(a,r),a}function yl(e,t,n=10){const r=Math.pow(n,t);return Math.round(e*r)/r}const Kv=Symbol("default");function Ni({values:e,children:t}){for(let[n,r]of e)t=be.createElement(n.Provider,{value:r},t);return t}function Cn(e){let{className:t,style:n,children:r,defaultClassName:i,defaultChildren:a,defaultStyle:u,values:l}=e;return P.useMemo(()=>{let o,s,c;return typeof t=="function"?o=t({...l,defaultClassName:i}):o=t,typeof n=="function"?s=n({...l,defaultStyle:u||{}}):s=n,typeof r=="function"?c=r({...l,defaultChildren:a}):r==null?c=a:c=r,{className:o??i,style:s||u?{...u,...s}:void 0,children:c??a,"data-rac":""}},[t,n,r,i,a,u,l])}function $_e(e,t){return n=>t(typeof e=="function"?e(n):e,n)}function xs(e,t){let n=P.useContext(e);if(t===null)return null;if(n&&typeof n=="object"&&"slots"in n&&n.slots){let r=t||Kv;if(!n.slots[r]){let i=new Intl.ListFormat().format(Object.keys(n.slots).map(u=>`"${u}"`)),a=t?`Invalid slot "${t}".`:"A slot prop is required.";throw new Error(`${a} Valid slot names are ${i}.`)}return n.slots[r]}return n}function Nn(e,t,n){let r=xs(n,e.slot)||{},{ref:i,...a}=r,u=wu(P.useMemo(()=>yg(t,i),[t,i])),l=Et(a,e);return"style"in a&&a.style&&"style"in e&&e.style&&(typeof a.style=="function"||typeof e.style=="function"?l.style=o=>{let s=typeof a.style=="function"?a.style(o):a.style,c={...o.defaultStyle,...s},d=typeof e.style=="function"?e.style({...o,defaultStyle:c}):e.style;return{...c,...d}}:l.style={...a.style,...e.style}),[l,u]}function Ss(e=!0){let[t,n]=P.useState(e),r=P.useRef(!1),i=P.useCallback(a=>{r.current=!0,n(!!a)},[]);return Hr(()=>{r.current||n(!1)},[]),[i,t]}function pc(e){const t=/^(data-.*)$/;let n={};for(const r in e)t.test(r)||(n[r]=e[r]);return n}var hte={};hte={buttonLabel:"عرض المقترحات",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} خيار`,other:()=>`${t.number(e.optionCount)} خيارات`})} متاحة.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`المجموعة المدخلة ${e.groupTitle}, مع ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} خيار`,other:()=>`${t.number(e.groupCount)} خيارات`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", محدد",other:""},e.isSelected)}`,listboxLabel:"مقترحات",selectedAnnouncement:e=>`${e.optionText}، محدد`};var pte={};pte={buttonLabel:"Покажи предложения",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} опция`,other:()=>`${t.number(e.optionCount)} опции`})} на разположение.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Въведена група ${e.groupTitle}, с ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} опция`,other:()=>`${t.number(e.groupCount)} опции`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", избрани",other:""},e.isSelected)}`,listboxLabel:"Предложения",selectedAnnouncement:e=>`${e.optionText}, избрани`};var mte={};mte={buttonLabel:"Zobrazit doporučení",countAnnouncement:(e,t)=>`K dispozici ${t.plural(e.optionCount,{one:()=>`je ${t.number(e.optionCount)} možnost`,other:()=>`jsou/je ${t.number(e.optionCount)} možnosti/-í`})}.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Zadaná skupina „${e.groupTitle}“ ${t.plural(e.groupCount,{one:()=>`s ${t.number(e.groupCount)} možností`,other:()=>`se ${t.number(e.groupCount)} možnostmi`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:" (vybráno)",other:""},e.isSelected)}`,listboxLabel:"Návrhy",selectedAnnouncement:e=>`${e.optionText}, vybráno`};var gte={};gte={buttonLabel:"Vis forslag",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} mulighed tilgængelig`,other:()=>`${t.number(e.optionCount)} muligheder tilgængelige`})}.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Angivet gruppe ${e.groupTitle}, med ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} mulighed`,other:()=>`${t.number(e.groupCount)} muligheder`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", valgt",other:""},e.isSelected)}`,listboxLabel:"Forslag",selectedAnnouncement:e=>`${e.optionText}, valgt`};var vte={};vte={buttonLabel:"Empfehlungen anzeigen",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} Option`,other:()=>`${t.number(e.optionCount)} Optionen`})} verfügbar.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Eingetretene Gruppe ${e.groupTitle}, mit ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} Option`,other:()=>`${t.number(e.groupCount)} Optionen`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", ausgewählt",other:""},e.isSelected)}`,listboxLabel:"Empfehlungen",selectedAnnouncement:e=>`${e.optionText}, ausgewählt`};var yte={};yte={buttonLabel:"Προβολή προτάσεων",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} επιλογή`,other:()=>`${t.number(e.optionCount)} επιλογές `})} διαθέσιμες.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Εισαγμένη ομάδα ${e.groupTitle}, με ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} επιλογή`,other:()=>`${t.number(e.groupCount)} επιλογές`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", επιλεγμένο",other:""},e.isSelected)}`,listboxLabel:"Προτάσεις",selectedAnnouncement:e=>`${e.optionText}, επιλέχθηκε`};var bte={};bte={focusAnnouncement:(e,t)=>`${t.select({true:()=>`Entered group ${e.groupTitle}, with ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} option`,other:()=>`${t.number(e.groupCount)} options`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", selected",other:""},e.isSelected)}`,countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} option`,other:()=>`${t.number(e.optionCount)} options`})} available.`,selectedAnnouncement:e=>`${e.optionText}, selected`,buttonLabel:"Show suggestions",listboxLabel:"Suggestions"};var Ote={};Ote={buttonLabel:"Mostrar sugerencias",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opción`,other:()=>`${t.number(e.optionCount)} opciones`})} disponible(s).`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Se ha unido al grupo ${e.groupTitle}, con ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opción`,other:()=>`${t.number(e.groupCount)} opciones`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", seleccionado",other:""},e.isSelected)}`,listboxLabel:"Sugerencias",selectedAnnouncement:e=>`${e.optionText}, seleccionado`};var xte={};xte={buttonLabel:"Kuva soovitused",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} valik`,other:()=>`${t.number(e.optionCount)} valikud`})} saadaval.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Sisestatud rühm ${e.groupTitle}, valikuga ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} valik`,other:()=>`${t.number(e.groupCount)} valikud`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", valitud",other:""},e.isSelected)}`,listboxLabel:"Soovitused",selectedAnnouncement:e=>`${e.optionText}, valitud`};var Ste={};Ste={buttonLabel:"Näytä ehdotukset",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} vaihtoehto`,other:()=>`${t.number(e.optionCount)} vaihtoehdot`})} saatavilla.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Mentiin ryhmään ${e.groupTitle}, ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} vaihtoehdon`,other:()=>`${t.number(e.groupCount)} vaihtoehdon`})} kanssa.`,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", valittu",other:""},e.isSelected)}`,listboxLabel:"Ehdotukset",selectedAnnouncement:e=>`${e.optionText}, valittu`};var _te={};_te={buttonLabel:"Afficher les suggestions",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} option`,other:()=>`${t.number(e.optionCount)} options`})} disponible(s).`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Groupe ${e.groupTitle} rejoint, avec ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} option`,other:()=>`${t.number(e.groupCount)} options`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", sélectionné(s)",other:""},e.isSelected)}`,listboxLabel:"Suggestions",selectedAnnouncement:e=>`${e.optionText}, sélectionné`};var Ete={};Ete={buttonLabel:"הצג הצעות",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`אפשרות ${t.number(e.optionCount)}`,other:()=>`${t.number(e.optionCount)} אפשרויות`})} במצב זמין.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`נכנס לקבוצה ${e.groupTitle}, עם ${t.plural(e.groupCount,{one:()=>`אפשרות ${t.number(e.groupCount)}`,other:()=>`${t.number(e.groupCount)} אפשרויות`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", נבחר",other:""},e.isSelected)}`,listboxLabel:"הצעות",selectedAnnouncement:e=>`${e.optionText}, נבחר`};var Dte={};Dte={buttonLabel:"Prikaži prijedloge",countAnnouncement:(e,t)=>`Dostupno još: ${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opcija`,other:()=>`${t.number(e.optionCount)} opcije/a`})}.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Unesena skupina ${e.groupTitle}, s ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opcijom`,other:()=>`${t.number(e.groupCount)} opcije/a`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", odabranih",other:""},e.isSelected)}`,listboxLabel:"Prijedlozi",selectedAnnouncement:e=>`${e.optionText}, odabrano`};var $te={};$te={buttonLabel:"Javaslatok megjelenítése",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} lehetőség`,other:()=>`${t.number(e.optionCount)} lehetőség`})} áll rendelkezésre.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Belépett a(z) ${e.groupTitle} csoportba, amely ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} lehetőséget`,other:()=>`${t.number(e.groupCount)} lehetőséget`})} tartalmaz. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", kijelölve",other:""},e.isSelected)}`,listboxLabel:"Javaslatok",selectedAnnouncement:e=>`${e.optionText}, kijelölve`};var Cte={};Cte={buttonLabel:"Mostra suggerimenti",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opzione disponibile`,other:()=>`${t.number(e.optionCount)} opzioni disponibili`})}.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Ingresso nel gruppo ${e.groupTitle}, con ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opzione`,other:()=>`${t.number(e.groupCount)} opzioni`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", selezionato",other:""},e.isSelected)}`,listboxLabel:"Suggerimenti",selectedAnnouncement:e=>`${e.optionText}, selezionato`};var wte={};wte={buttonLabel:"候補を表示",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} 個のオプション`,other:()=>`${t.number(e.optionCount)} 個のオプション`})}を利用できます。`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`入力されたグループ ${e.groupTitle}、${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} 個のオプション`,other:()=>`${t.number(e.groupCount)} 個のオプション`})}を含む。`,other:""},e.isGroupChange)}${e.optionText}${t.select({true:"、選択済み",other:""},e.isSelected)}`,listboxLabel:"候補",selectedAnnouncement:e=>`${e.optionText}、選択済み`};var Rte={};Rte={buttonLabel:"제안 사항 표시",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)}개 옵션`,other:()=>`${t.number(e.optionCount)}개 옵션`})}을 사용할 수 있습니다.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`입력한 그룹 ${e.groupTitle}, ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)}개 옵션`,other:()=>`${t.number(e.groupCount)}개 옵션`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", 선택됨",other:""},e.isSelected)}`,listboxLabel:"제안",selectedAnnouncement:e=>`${e.optionText}, 선택됨`};var Ate={};Ate={buttonLabel:"Rodyti pasiūlymus",countAnnouncement:(e,t)=>`Yra ${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} parinktis`,other:()=>`${t.number(e.optionCount)} parinktys (-ių)`})}.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Įvesta grupė ${e.groupTitle}, su ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} parinktimi`,other:()=>`${t.number(e.groupCount)} parinktimis (-ių)`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", pasirinkta",other:""},e.isSelected)}`,listboxLabel:"Pasiūlymai",selectedAnnouncement:e=>`${e.optionText}, pasirinkta`};var Tte={};Tte={buttonLabel:"Rādīt ieteikumus",countAnnouncement:(e,t)=>`Pieejamo opciju skaits: ${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opcija`,other:()=>`${t.number(e.optionCount)} opcijas`})}.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Ievadīta grupa ${e.groupTitle}, ar ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opciju`,other:()=>`${t.number(e.groupCount)} opcijām`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", atlasīta",other:""},e.isSelected)}`,listboxLabel:"Ieteikumi",selectedAnnouncement:e=>`${e.optionText}, atlasīta`};var kte={};kte={buttonLabel:"Vis forslag",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} alternativ`,other:()=>`${t.number(e.optionCount)} alternativer`})} finnes.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Angitt gruppe ${e.groupTitle}, med ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} alternativ`,other:()=>`${t.number(e.groupCount)} alternativer`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", valgt",other:""},e.isSelected)}`,listboxLabel:"Forslag",selectedAnnouncement:e=>`${e.optionText}, valgt`};var Pte={};Pte={buttonLabel:"Suggesties weergeven",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} optie`,other:()=>`${t.number(e.optionCount)} opties`})} beschikbaar.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Groep ${e.groupTitle} ingevoerd met ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} optie`,other:()=>`${t.number(e.groupCount)} opties`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", geselecteerd",other:""},e.isSelected)}`,listboxLabel:"Suggesties",selectedAnnouncement:e=>`${e.optionText}, geselecteerd`};var Mte={};Mte={buttonLabel:"Wyświetlaj sugestie",countAnnouncement:(e,t)=>`dostępna/dostępne(-nych) ${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opcja`,other:()=>`${t.number(e.optionCount)} opcje(-i)`})}.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Dołączono do grupy ${e.groupTitle}, z ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opcją`,other:()=>`${t.number(e.groupCount)} opcjami`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", wybrano",other:""},e.isSelected)}`,listboxLabel:"Sugestie",selectedAnnouncement:e=>`${e.optionText}, wybrano`};var Fte={};Fte={buttonLabel:"Mostrar sugestões",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opção`,other:()=>`${t.number(e.optionCount)} opções`})} disponível.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Grupo inserido ${e.groupTitle}, com ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opção`,other:()=>`${t.number(e.groupCount)} opções`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", selecionado",other:""},e.isSelected)}`,listboxLabel:"Sugestões",selectedAnnouncement:e=>`${e.optionText}, selecionado`};var Nte={};Nte={buttonLabel:"Apresentar sugestões",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opção`,other:()=>`${t.number(e.optionCount)} opções`})} disponível.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Grupo introduzido ${e.groupTitle}, com ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opção`,other:()=>`${t.number(e.groupCount)} opções`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", selecionado",other:""},e.isSelected)}`,listboxLabel:"Sugestões",selectedAnnouncement:e=>`${e.optionText}, selecionado`};var Ite={};Ite={buttonLabel:"Afișare sugestii",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opțiune`,other:()=>`${t.number(e.optionCount)} opțiuni`})} disponibile.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Grup ${e.groupTitle} introdus, cu ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opțiune`,other:()=>`${t.number(e.groupCount)} opțiuni`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", selectat",other:""},e.isSelected)}`,listboxLabel:"Sugestii",selectedAnnouncement:e=>`${e.optionText}, selectat`};var Lte={};Lte={buttonLabel:"Показать предложения",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} параметр`,other:()=>`${t.number(e.optionCount)} параметров`})} доступно.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Введенная группа ${e.groupTitle}, с ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} параметром`,other:()=>`${t.number(e.groupCount)} параметрами`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", выбранными",other:""},e.isSelected)}`,listboxLabel:"Предложения",selectedAnnouncement:e=>`${e.optionText}, выбрано`};var Bte={};Bte={buttonLabel:"Zobraziť návrhy",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} možnosť`,other:()=>`${t.number(e.optionCount)} možnosti/-í`})} k dispozícii.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Zadaná skupina ${e.groupTitle}, s ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} možnosťou`,other:()=>`${t.number(e.groupCount)} možnosťami`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", vybraté",other:""},e.isSelected)}`,listboxLabel:"Návrhy",selectedAnnouncement:e=>`${e.optionText}, vybraté`};var Qte={};Qte={buttonLabel:"Prikaži predloge",countAnnouncement:(e,t)=>`Na voljo je ${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opcija`,other:()=>`${t.number(e.optionCount)} opcije`})}.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Vnesena skupina ${e.groupTitle}, z ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opcija`,other:()=>`${t.number(e.groupCount)} opcije`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", izbrano",other:""},e.isSelected)}`,listboxLabel:"Predlogi",selectedAnnouncement:e=>`${e.optionText}, izbrano`};var zte={};zte={buttonLabel:"Prikaži predloge",countAnnouncement:(e,t)=>`Dostupno još: ${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} opcija`,other:()=>`${t.number(e.optionCount)} opcije/a`})}.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Unesena grupa ${e.groupTitle}, s ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} opcijom`,other:()=>`${t.number(e.groupCount)} optione/a`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", izabranih",other:""},e.isSelected)}`,listboxLabel:"Predlozi",selectedAnnouncement:e=>`${e.optionText}, izabrano`};var qte={};qte={buttonLabel:"Visa förslag",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} alternativ`,other:()=>`${t.number(e.optionCount)} alternativ`})} tillgängliga.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Ingick i gruppen ${e.groupTitle} med ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} alternativ`,other:()=>`${t.number(e.groupCount)} alternativ`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", valda",other:""},e.isSelected)}`,listboxLabel:"Förslag",selectedAnnouncement:e=>`${e.optionText}, valda`};var Vte={};Vte={buttonLabel:"Önerileri göster",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} seçenek`,other:()=>`${t.number(e.optionCount)} seçenekler`})} kullanılabilir.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Girilen grup ${e.groupTitle}, ile ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} seçenek`,other:()=>`${t.number(e.groupCount)} seçenekler`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", seçildi",other:""},e.isSelected)}`,listboxLabel:"Öneriler",selectedAnnouncement:e=>`${e.optionText}, seçildi`};var Ute={};Ute={buttonLabel:"Показати пропозиції",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} параметр`,other:()=>`${t.number(e.optionCount)} параметри(-ів)`})} доступно.`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`Введена група ${e.groupTitle}, з ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} параметр`,other:()=>`${t.number(e.groupCount)} параметри(-ів)`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", вибрано",other:""},e.isSelected)}`,listboxLabel:"Пропозиції",selectedAnnouncement:e=>`${e.optionText}, вибрано`};var jte={};jte={buttonLabel:"显示建议",countAnnouncement:(e,t)=>`有 ${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} 个选项`,other:()=>`${t.number(e.optionCount)} 个选项`})}可用。`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`进入了 ${e.groupTitle} 组，其中有 ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} 个选项`,other:()=>`${t.number(e.groupCount)} 个选项`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", 已选择",other:""},e.isSelected)}`,listboxLabel:"建议",selectedAnnouncement:e=>`${e.optionText}, 已选择`};var Zte={};Zte={buttonLabel:"顯示建議",countAnnouncement:(e,t)=>`${t.plural(e.optionCount,{one:()=>`${t.number(e.optionCount)} 選項`,other:()=>`${t.number(e.optionCount)} 選項`})} 可用。`,focusAnnouncement:(e,t)=>`${t.select({true:()=>`輸入的群組 ${e.groupTitle}, 有 ${t.plural(e.groupCount,{one:()=>`${t.number(e.groupCount)} 選項`,other:()=>`${t.number(e.groupCount)} 選項`})}. `,other:""},e.isGroupChange)}${e.optionText}${t.select({true:", 已選取",other:""},e.isSelected)}`,listboxLabel:"建議",selectedAnnouncement:e=>`${e.optionText}, 已選取`};var Hte={};Hte={"ar-AE":hte,"bg-BG":pte,"cs-CZ":mte,"da-DK":gte,"de-DE":vte,"el-GR":yte,"en-US":bte,"es-ES":Ote,"et-EE":xte,"fi-FI":Ste,"fr-FR":_te,"he-IL":Ete,"hr-HR":Dte,"hu-HU":$te,"it-IT":Cte,"ja-JP":wte,"ko-KR":Rte,"lt-LT":Ate,"lv-LV":Tte,"nb-NO":kte,"nl-NL":Pte,"pl-PL":Mte,"pt-BR":Fte,"pt-PT":Nte,"ro-RO":Ite,"ru-RU":Lte,"sk-SK":Bte,"sl-SI":Qte,"sr-SP":zte,"sv-SE":qte,"tr-TR":Vte,"uk-UA":Ute,"zh-CN":jte,"zh-TW":Zte};const Xte=7e3;let Qd=null;function Zb(e,t="assertive",n=Xte){Qd?Qd.announce(e,t,n):(Qd=new w_e,(typeof IS_REACT_ACT_ENVIRONMENT=="boolean"?IS_REACT_ACT_ENVIRONMENT:typeof jest<"u")?Qd.announce(e,t,n):setTimeout(()=>{Qd!=null&&Qd.isAttached()&&(Qd==null||Qd.announce(e,t,n))},100))}function C_e(e){Qd&&Qd.clear(e)}class w_e{isAttached(){var t;return(t=this.node)===null||t===void 0?void 0:t.isConnected}createLog(t){let n=document.createElement("div");return n.setAttribute("role","log"),n.setAttribute("aria-live",t),n.setAttribute("aria-relevant","additions"),n}destroy(){this.node&&(document.body.removeChild(this.node),this.node=null)}announce(t,n="assertive",r=Xte){var i,a;if(!this.node)return;let u=document.createElement("div");typeof t=="object"?(u.setAttribute("role","img"),u.setAttribute("aria-labelledby",t["aria-labelledby"])):u.textContent=t,n==="assertive"?(i=this.assertiveLog)===null||i===void 0||i.appendChild(u):(a=this.politeLog)===null||a===void 0||a.appendChild(u),t!==""&&setTimeout(()=>{u.remove()},r)}clear(t){this.node&&((!t||t==="assertive")&&this.assertiveLog&&(this.assertiveLog.innerHTML=""),(!t||t==="polite")&&this.politeLog&&(this.politeLog.innerHTML=""))}constructor(){this.node=null,this.assertiveLog=null,this.politeLog=null,typeof document<"u"&&(this.node=document.createElement("div"),this.node.dataset.liveAnnouncer="true",Object.assign(this.node.style,{border:0,clip:"rect(0 0 0 0)",clipPath:"inset(50%)",height:"1px",margin:"-1px",overflow:"hidden",padding:0,position:"absolute",width:"1px",whiteSpace:"nowrap"}),this.assertiveLog=this.createLog("assertive"),this.node.appendChild(this.assertiveLog),this.politeLog=this.createLog("polite"),this.node.appendChild(this.politeLog),document.body.prepend(this.node))}}const y0={top:"top",bottom:"top",left:"left",right:"left"},YD={top:"bottom",bottom:"top",left:"right",right:"left"},R_e={top:"left",left:"top"},D9={top:"height",left:"width"},Yte={width:"totalWidth",height:"totalHeight"},V_={};let no=typeof document<"u"?window.visualViewport:null;function wU(e){let t=0,n=0,r=0,i=0,a=0,u=0,l={};var o;let s=((o=no==null?void 0:no.scale)!==null&&o!==void 0?o:1)>1;if(e.tagName==="BODY"){let y=document.documentElement;r=y.clientWidth,i=y.clientHeight;var c;t=(c=no==null?void 0:no.width)!==null&&c!==void 0?c:r;var d;n=(d=no==null?void 0:no.height)!==null&&d!==void 0?d:i,l.top=y.scrollTop||e.scrollTop,l.left=y.scrollLeft||e.scrollLeft,no&&(a=no.offsetTop,u=no.offsetLeft)}else({width:t,height:n,top:a,left:u}=i3(e,!1)),l.top=e.scrollTop,l.left=e.scrollLeft,r=t,i=n;if(fN()&&(e.tagName==="BODY"||e.tagName==="HTML")&&s){l.top=0,l.left=0;var h;a=(h=no==null?void 0:no.pageTop)!==null&&h!==void 0?h:0;var m;u=(m=no==null?void 0:no.pageLeft)!==null&&m!==void 0?m:0}return{width:t,height:n,totalWidth:r,totalHeight:i,scroll:l,top:a,left:u}}function A_e(e){return{top:e.scrollTop,left:e.scrollLeft,width:e.scrollWidth,height:e.scrollHeight}}function RU(e,t,n,r,i,a,u){var l;let o=(l=i.scroll[e])!==null&&l!==void 0?l:0,s=r[D9[e]],c=r.scroll[y0[e]]+a,d=s+r.scroll[y0[e]]-a,h=t-o+u[e]-r[y0[e]],m=t-o+n+u[e]-r[y0[e]];return h<c?c-h:m>d?Math.max(d-m,c-h):0}function T_e(e){let t=window.getComputedStyle(e);return{top:parseInt(t.marginTop,10)||0,bottom:parseInt(t.marginBottom,10)||0,left:parseInt(t.marginLeft,10)||0,right:parseInt(t.marginRight,10)||0}}function AU(e){if(V_[e])return V_[e];let[t,n]=e.split(" "),r=y0[t]||"right",i=R_e[r];y0[n]||(n="center");let a=D9[r],u=D9[i];return V_[e]={placement:t,crossPlacement:n,axis:r,crossAxis:i,size:a,crossSize:u},V_[e]}function XT(e,t,n,r,i,a,u,l,o,s){let{placement:c,crossPlacement:d,axis:h,crossAxis:m,size:y,crossSize:v}=r,S={};var _;S[m]=(_=e[m])!==null&&_!==void 0?_:0;var E,$,k,A;d==="center"?S[m]+=(((E=e[v])!==null&&E!==void 0?E:0)-(($=n[v])!==null&&$!==void 0?$:0))/2:d!==m&&(S[m]+=((k=e[v])!==null&&k!==void 0?k:0)-((A=n[v])!==null&&A!==void 0?A:0)),S[m]+=a;const w=e[m]-n[v]+o+s,R=e[m]+e[v]-o-s;if(S[m]=so(S[m],w,R),c===h){const F=l?u[y]:t[Yte[y]];S[YD[h]]=Math.floor(F-e[h]+i)}else S[h]=Math.floor(e[h]+e[y]+i);return S}function k_e(e,t,n,r,i,a,u,l){const o=r?n.height:t[Yte.height];var s;let c=e.top!=null?n.top+e.top:n.top+(o-((s=e.bottom)!==null&&s!==void 0?s:0)-u);var d,h,m,y,v,S;let _=l!=="top"?Math.max(0,t.height+t.top+((d=t.scroll.top)!==null&&d!==void 0?d:0)-c-(((h=i.top)!==null&&h!==void 0?h:0)+((m=i.bottom)!==null&&m!==void 0?m:0)+a)):Math.max(0,c+u-(t.top+((y=t.scroll.top)!==null&&y!==void 0?y:0))-(((v=i.top)!==null&&v!==void 0?v:0)+((S=i.bottom)!==null&&S!==void 0?S:0)+a));return Math.min(t.height-a*2,_)}function TU(e,t,n,r,i,a){let{placement:u,axis:l,size:o}=a;var s,c;if(u===l)return Math.max(0,n[l]-e[l]-((s=e.scroll[l])!==null&&s!==void 0?s:0)+t[l]-((c=r[l])!==null&&c!==void 0?c:0)-r[YD[l]]-i);var d;return Math.max(0,e[o]+e[l]+e.scroll[l]-t[l]-n[l]-n[o]-((d=r[l])!==null&&d!==void 0?d:0)-r[YD[l]]-i)}function P_e(e,t,n,r,i,a,u,l,o,s,c,d,h,m,y,v){let S=AU(e),{size:_,crossAxis:E,crossSize:$,placement:k,crossPlacement:A}=S,w=XT(t,l,n,S,c,d,s,h,y,v),R=c,F=TU(l,s,t,i,a+c,S);if(u&&r[_]>F){let de=AU(`${YD[k]} ${A}`),ge=XT(t,l,n,de,c,d,s,h,y,v);TU(l,s,t,i,a+c,de)>F&&(S=de,w=ge,R=c)}let I="bottom";S.axis==="top"?S.placement==="top"?I="top":S.placement==="bottom"&&(I="bottom"):S.crossAxis==="top"&&(S.crossPlacement==="top"?I="bottom":S.crossPlacement==="bottom"&&(I="top"));let U=RU(E,w[E],n[$],l,o,a,s);w[E]+=U;let q=k_e(w,l,s,h,i,a,n.height,I);m&&m<q&&(q=m),n.height=Math.min(n.height,q),w=XT(t,l,n,S,R,d,s,h,y,v),U=RU(E,w[E],n[$],l,o,a,s),w[E]+=U;let V={},N=t[E]-w[E]-i[y0[E]],j=N+.5*t[$];const W=y/2+v;var K,J,te,H;const M=y0[E]==="left"?((K=i.left)!==null&&K!==void 0?K:0)+((J=i.right)!==null&&J!==void 0?J:0):((te=i.top)!==null&&te!==void 0?te:0)+((H=i.bottom)!==null&&H!==void 0?H:0),G=n[$]-M-y/2-v,X=t[E]+y/2-(w[E]+i[y0[E]]),B=t[E]+t[$]-y/2-(w[E]+i[y0[E]]),ne=so(j,X,B);V[E]=so(ne,W,G),{placement:k,crossPlacement:A}=S,y?N=V[E]:A==="right"?N+=t[$]:A==="center"&&(N+=t[$]/2);let oe=k==="left"||k==="top"?n[_]:0,ie={x:k==="top"||k==="bottom"?N:oe,y:k==="left"||k==="right"?N:oe};return{position:w,maxHeight:q,arrowOffsetLeft:V.left,arrowOffsetTop:V.top,placement:k,triggerAnchorPoint:ie}}function M_e(e){let{placement:t,targetNode:n,overlayNode:r,scrollNode:i,padding:a,shouldFlip:u,boundaryElement:l,offset:o,crossOffset:s,maxHeight:c,arrowSize:d=0,arrowBoundaryOffset:h=0}=e,m=r instanceof HTMLElement?F_e(r):document.documentElement,y=m===document.documentElement;const v=window.getComputedStyle(m).position;let S=!!v&&v!=="static",_=y?i3(n,!1):kU(n,m,!1);if(!y){let{marginTop:V,marginLeft:N}=window.getComputedStyle(n);_.top+=parseInt(V,10)||0,_.left+=parseInt(N,10)||0}let E=i3(r,!0),$=T_e(r);var k,A;E.width+=((k=$.left)!==null&&k!==void 0?k:0)+((A=$.right)!==null&&A!==void 0?A:0);var w,R;E.height+=((w=$.top)!==null&&w!==void 0?w:0)+((R=$.bottom)!==null&&R!==void 0?R:0);let F=A_e(i),I=wU(l),U=wU(m),q=l.tagName==="BODY"?i3(m,!1):kU(m,l,!1);return m.tagName==="HTML"&&l.tagName==="BODY"&&(U.scroll.top=0,U.scroll.left=0),P_e(t,_,E,F,$,a,u,I,U,q,o,s,S,c,d,h)}function vN(e,t){let{top:n,left:r,width:i,height:a}=e.getBoundingClientRect();return t&&e instanceof e.ownerDocument.defaultView.HTMLElement&&(i=e.offsetWidth,a=e.offsetHeight),{top:n,left:r,width:i,height:a}}function i3(e,t){let{top:n,left:r,width:i,height:a}=vN(e,t),{scrollTop:u,scrollLeft:l,clientTop:o,clientLeft:s}=document.documentElement;return{top:n+u-o,left:r+l-s,width:i,height:a}}function kU(e,t,n){let r=window.getComputedStyle(e),i;if(r.position==="fixed")i=vN(e,n);else{i=i3(e,n);let a=i3(t,n),u=window.getComputedStyle(t);a.top+=(parseInt(u.borderTopWidth,10)||0)-t.scrollTop,a.left+=(parseInt(u.borderLeftWidth,10)||0)-t.scrollLeft,i.top-=a.top,i.left-=a.left}return i.top-=parseInt(r.marginTop,10)||0,i.left-=parseInt(r.marginLeft,10)||0,i}function F_e(e){let t=e.offsetParent;if(t&&t===document.body&&window.getComputedStyle(t).position==="static"&&!PU(t)&&(t=document.documentElement),t==null)for(t=e.parentElement;t&&!PU(t);)t=t.parentElement;return t||document.documentElement}function PU(e){let t=window.getComputedStyle(e);return t.transform!=="none"||/transform|perspective/.test(t.willChange)||t.filter!=="none"||t.contain==="paint"||"backdropFilter"in t&&t.backdropFilter!=="none"||"WebkitBackdropFilter"in t&&t.WebkitBackdropFilter!=="none"}const Gte=new WeakMap;function N_e(e){let{triggerRef:t,isOpen:n,onClose:r}=e;P.useEffect(()=>{if(!n||r===null)return;let i=a=>{let u=a.target;if(!t.current||u instanceof Node&&!u.contains(t.current)||a.target instanceof HTMLInputElement||a.target instanceof HTMLTextAreaElement)return;let l=r||Gte.get(t.current);l&&l()};return window.addEventListener("scroll",i,!0),()=>{window.removeEventListener("scroll",i,!0)}},[n,r,t])}const I_e=new Set(["Arab","Syrc","Samr","Mand","Thaa","Mend","Nkoo","Adlm","Rohg","Hebr"]),L_e=new Set(["ae","ar","arc","bcc","bqi","ckb","dv","fa","glk","he","ku","mzn","nqo","pnb","ps","sd","ug","ur","yi"]);function B_e(e){if(Intl.Locale){let n=new Intl.Locale(e).maximize(),r=typeof n.getTextInfo=="function"?n.getTextInfo():n.textInfo;if(r)return r.direction==="rtl";if(n.script)return I_e.has(n.script)}let t=e.split("-")[0];return L_e.has(t)}const Q_e=Symbol.for("react-aria.i18n.locale");function Kte(){let e=typeof window<"u"&&window[Q_e]||typeof navigator<"u"&&(navigator.language||navigator.userLanguage)||"en-US";try{Intl.DateTimeFormat.supportedLocalesOf([e])}catch{e="en-US"}return{locale:e,direction:B_e(e)?"rtl":"ltr"}}let $9=Kte(),M2=new Set;function MU(){$9=Kte();for(let e of M2)e($9)}function z_e(){let e=pp(),[t,n]=P.useState($9);return P.useEffect(()=>(M2.size===0&&window.addEventListener("languagechange",MU),M2.add(n),()=>{M2.delete(n),M2.size===0&&window.removeEventListener("languagechange",MU)}),[]),e?{locale:"en-US",direction:"ltr"}:t}const q_e=be.createContext(null);function Wi(){let e=z_e();return P.useContext(q_e)||e}var C9=function(e,t){return C9=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(n,r){n.__proto__=r}||function(n,r){for(var i in r)Object.prototype.hasOwnProperty.call(r,i)&&(n[i]=r[i])},C9(e,t)};function Wte(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Class extends value "+String(t)+" is not a constructor or null");C9(e,t);function n(){this.constructor=e}e.prototype=t===null?Object.create(t):(n.prototype=t.prototype,new n)}var kb=function(){return kb=Object.assign||function(t){for(var n,r=1,i=arguments.length;r<i;r++){n=arguments[r];for(var a in n)Object.prototype.hasOwnProperty.call(n,a)&&(t[a]=n[a])}return t},kb.apply(this,arguments)};function V_e(e,t,n,r){function i(a){return a instanceof n?a:new n(function(u){u(a)})}return new(n||(n=Promise))(function(a,u){function l(c){try{s(r.next(c))}catch(d){u(d)}}function o(c){try{s(r.throw(c))}catch(d){u(d)}}function s(c){c.done?a(c.value):i(c.value).then(l,o)}s((r=r.apply(e,t||[])).next())})}function U_e(e,t){var n={label:0,sent:function(){if(a[0]&1)throw a[1];return a[1]},trys:[],ops:[]},r,i,a,u=Object.create((typeof Iterator=="function"?Iterator:Object).prototype);return u.next=l(0),u.throw=l(1),u.return=l(2),typeof Symbol=="function"&&(u[Symbol.iterator]=function(){return this}),u;function l(s){return function(c){return o([s,c])}}function o(s){if(r)throw new TypeError("Generator is already executing.");for(;u&&(u=0,s[0]&&(n=0)),n;)try{if(r=1,i&&(a=s[0]&2?i.return:s[0]?i.throw||((a=i.return)&&a.call(i),0):i.next)&&!(a=a.call(i,s[1])).done)return a;switch(i=0,a&&(s=[s[0]&2,a.value]),s[0]){case 0:case 1:a=s;break;case 4:return n.label++,{value:s[1],done:!1};case 5:n.label++,i=s[1],s=[0];continue;case 7:s=n.ops.pop(),n.trys.pop();continue;default:if(a=n.trys,!(a=a.length>0&&a[a.length-1])&&(s[0]===6||s[0]===2)){n=0;continue}if(s[0]===3&&(!a||s[1]>a[0]&&s[1]<a[3])){n.label=s[1];break}if(s[0]===6&&n.label<a[1]){n.label=a[1],a=s;break}if(a&&n.label<a[2]){n.label=a[2],n.ops.push(s);break}a[2]&&n.ops.pop(),n.trys.pop();continue}s=t.call(e,n)}catch(c){s=[6,c],i=0}finally{r=a=0}if(s[0]&5)throw s[1];return{value:s[0]?s[1]:void 0,done:!0}}}function a3(e,t,n){if(n||arguments.length===2)for(var r=0,i=t.length,a;r<i;r++)(a||!(r in t))&&(a||(a=Array.prototype.slice.call(t,0,r)),a[r]=t[r]);return e.concat(a||Array.prototype.slice.call(t))}const j_e=Symbol.for("react-aria.i18n.locale"),Z_e=Symbol.for("react-aria.i18n.strings");let yO;class tp{getStringForLocale(t,n){let i=this.getStringsForLocale(n)[t];if(!i)throw new Error(`Could not find intl message ${t} in ${n} locale`);return i}getStringsForLocale(t){let n=this.strings[t];return n||(n=H_e(t,this.strings,this.defaultLocale),this.strings[t]=n),n}static getGlobalDictionaryForPackage(t){if(typeof window>"u")return null;let n=window[j_e];if(yO===void 0){let i=window[Z_e];if(!i)return null;yO={};for(let a in i)yO[a]=new tp({[n]:i[a]},n)}let r=yO==null?void 0:yO[t];if(!r)throw new Error(`Strings for package "${t}" were not included by LocalizedStringProvider. Please add it to the list passed to createLocalizedStringDictionary.`);return r}constructor(t,n="en-US"){this.strings=Object.fromEntries(Object.entries(t).filter(([,r])=>r)),this.defaultLocale=n}}function H_e(e,t,n="en-US"){if(t[e])return t[e];let r=X_e(e);if(t[r])return t[r];for(let i in t)if(i.startsWith(r+"-"))return t[i];return t[n]}function X_e(e){return Intl.Locale?new Intl.Locale(e).language:e.split("-")[0]}const FU=new Map,NU=new Map;class yN{format(t,n){let r=this.strings.getStringForLocale(t,this.locale);return typeof r=="function"?r(n,this):r}plural(t,n,r="cardinal"){let i=n["="+t];if(i)return typeof i=="function"?i():i;let a=this.locale+":"+r,u=FU.get(a);u||(u=new Intl.PluralRules(this.locale,{type:r}),FU.set(a,u));let l=u.select(t);return i=n[l]||n.other,typeof i=="function"?i():i}number(t){let n=NU.get(this.locale);return n||(n=new Intl.NumberFormat(this.locale),NU.set(this.locale,n)),n.format(t)}select(t,n){let r=t[n]||t.other;return typeof r=="function"?r():r}constructor(t,n){this.locale=t,this.strings=n}}const IU=new WeakMap;function Y_e(e){let t=IU.get(e);return t||(t=new tp(e),IU.set(e,t)),t}function Jte(e,t){return t&&tp.getGlobalDictionaryForPackage(t)||Y_e(e)}function Qo(e,t){let{locale:n}=Wi(),r=Jte(e,t);return P.useMemo(()=>new yN(n,r),[n,r])}function G_e(e={}){let{locale:t}=Wi();return P.useMemo(()=>new Intl.ListFormat(t,e),[t,e])}function u3(e,t){return e-t*Math.floor(e/t)}const ene=1721426;function Nb(e,t,n,r){t=dS(e,t);let i=t-1,a=-2;return n<=2?a=0:Nv(t)&&(a=-1),ene-1+365*i+Math.floor(i/4)-Math.floor(i/100)+Math.floor(i/400)+Math.floor((367*n-362)/12+a+r)}function Nv(e){return e%4===0&&(e%100!==0||e%400===0)}function dS(e,t){return e==="BC"?1-t:t}function VC(e){let t="AD";return e<=0&&(t="BC",e=1-e),[t,e]}const K_e={standard:[31,28,31,30,31,30,31,31,30,31,30,31],leapyear:[31,29,31,30,31,30,31,31,30,31,30,31]};class Xf{fromJulianDay(t){let n=t,r=n-ene,i=Math.floor(r/146097),a=u3(r,146097),u=Math.floor(a/36524),l=u3(a,36524),o=Math.floor(l/1461),s=u3(l,1461),c=Math.floor(s/365),d=i*400+u*100+o*4+c+(u!==4&&c!==4?1:0),[h,m]=VC(d),y=n-Nb(h,m,1,1),v=2;n<Nb(h,m,3,1)?v=0:Nv(m)&&(v=1);let S=Math.floor(((y+v)*12+373)/367),_=n-Nb(h,m,S,1)+1;return new Xu(h,m,S,_)}toJulianDay(t){return Nb(t.era,t.year,t.month,t.day)}getDaysInMonth(t){return K_e[Nv(t.year)?"leapyear":"standard"][t.month-1]}getMonthsInYear(t){return 12}getDaysInYear(t){return Nv(t.year)?366:365}getYearsInEra(t){return 9999}getEras(){return["BC","AD"]}isInverseEra(t){return t.era==="BC"}balanceDate(t){t.year<=0&&(t.era=t.era==="BC"?"AD":"BC",t.year=1-t.year)}constructor(){this.identifier="gregory"}}function tne(e,t){var n,r,i,a;return(a=(i=(n=e.isEqual)===null||n===void 0?void 0:n.call(e,t))!==null&&i!==void 0?i:(r=t.isEqual)===null||r===void 0?void 0:r.call(t,e))!==null&&a!==void 0?a:e.identifier===t.identifier}function nne(e){return Wm(Date.now(),e)}function W_e(e){return one(nne(e))}function rne(e,t){return e.calendar.toJulianDay(e)-t.calendar.toJulianDay(t)}function ine(e,t){return LU(e)-LU(t)}function LU(e){return e.hour*36e5+e.minute*6e4+e.second*1e3+e.millisecond}let YT=null;function hS(){return YT==null&&(YT=new Intl.DateTimeFormat().resolvedOptions().timeZone),YT}function J_e(e){return e.calendar.getMinimumMonthInYear?e.calendar.getMinimumMonthInYear(e):1}function eEe(e){return e.calendar.getMinimumDayInMonth?e.calendar.getMinimumDayInMonth(e):1}function R3(e){e=fc(e,new Xf);let t=dS(e.era,e.year);return ane(t,e.month,e.day,e.hour,e.minute,e.second,e.millisecond)}function ane(e,t,n,r,i,a,u){let l=new Date;return l.setUTCHours(r,i,a,u),l.setUTCFullYear(e,t-1,n),l.getTime()}function w9(e,t){if(t==="UTC")return 0;if(e>0&&t===hS())return new Date(e).getTimezoneOffset()*-6e4;let{year:n,month:r,day:i,hour:a,minute:u,second:l}=une(e,t);return ane(n,r,i,a,u,l,0)-Math.floor(e/1e3)*1e3}const BU=new Map;function une(e,t){let n=BU.get(t);n||(n=new Intl.DateTimeFormat("en-US",{timeZone:t,hour12:!1,era:"short",year:"numeric",month:"numeric",day:"numeric",hour:"numeric",minute:"numeric",second:"numeric"}),BU.set(t,n));let r=n.formatToParts(new Date(e)),i={};for(let a of r)a.type!=="literal"&&(i[a.type]=a.value);return{year:i.era==="BC"||i.era==="B"?-i.year+1:+i.year,month:+i.month,day:+i.day,hour:i.hour==="24"?0:+i.hour,minute:+i.minute,second:+i.second}}const QU=864e5;function tEe(e,t,n,r){return(n===r?[n]:[n,r]).filter(a=>nEe(e,t,a))}function nEe(e,t,n){let r=une(n,t);return e.year===r.year&&e.month===r.month&&e.day===r.day&&e.hour===r.hour&&e.minute===r.minute&&e.second===r.second}function jm(e,t,n="compatible"){let r=Wv(e);if(t==="UTC")return R3(r);if(t===hS()&&n==="compatible"){r=fc(r,new Xf);let o=new Date,s=dS(r.era,r.year);return o.setFullYear(s,r.month-1,r.day),o.setHours(r.hour,r.minute,r.second,r.millisecond),o.getTime()}let i=R3(r),a=w9(i-QU,t),u=w9(i+QU,t),l=tEe(r,t,i-a,i-u);if(l.length===1)return l[0];if(l.length>1)switch(n){case"compatible":case"earlier":return l[0];case"later":return l[l.length-1];case"reject":throw new RangeError("Multiple possible absolute times found")}switch(n){case"earlier":return Math.min(i-a,i-u);case"compatible":case"later":return Math.max(i-a,i-u);case"reject":throw new RangeError("No such absolute time found")}}function lne(e,t,n="compatible"){return new Date(jm(e,t,n))}function Wm(e,t){let n=w9(e,t),r=new Date(e+n),i=r.getUTCFullYear(),a=r.getUTCMonth()+1,u=r.getUTCDate(),l=r.getUTCHours(),o=r.getUTCMinutes(),s=r.getUTCSeconds(),c=r.getUTCMilliseconds();return new A3(i<1?"BC":"AD",i<1?-i+1:i,a,u,t,n,l,o,s,c)}function one(e){return new Xu(e.calendar,e.era,e.year,e.month,e.day)}function Wv(e,t){let n=0,r=0,i=0,a=0;if("timeZone"in e)({hour:n,minute:r,second:i,millisecond:a}=e);else if("hour"in e&&!t)return e;return t&&({hour:n,minute:r,second:i,millisecond:a}=t),new bx(e.calendar,e.era,e.year,e.month,e.day,n,r,i,a)}function zU(e){return new jC(e.hour,e.minute,e.second,e.millisecond)}function fc(e,t){if(tne(e.calendar,t))return e;let n=t.fromJulianDay(e.calendar.toJulianDay(e)),r=e.copy();return r.calendar=t,r.era=n.era,r.year=n.year,r.month=n.month,r.day=n.day,o1(r),r}function sne(e,t,n){if(e instanceof A3)return e.timeZone===t?e:cne(e,t);let r=jm(e,t,n);return Wm(r,t)}function rEe(e){let t=R3(e)-e.offset;return new Date(t)}function cne(e,t){let n=R3(e)-e.offset;return fc(Wm(n,t),e.calendar)}const K4=36e5;function UC(e,t){let n=e.copy(),r="hour"in n?pne(n,t):0;R9(n,t.years||0),n.calendar.balanceYearMonth&&n.calendar.balanceYearMonth(n,e),n.month+=t.months||0,A9(n),fne(n),n.day+=(t.weeks||0)*7,n.day+=t.days||0,n.day+=r,iEe(n),n.calendar.balanceDate&&n.calendar.balanceDate(n),n.year<1&&(n.year=1,n.month=1,n.day=1);let i=n.calendar.getYearsInEra(n);if(n.year>i){var a,u;let o=(a=(u=n.calendar).isInverseEra)===null||a===void 0?void 0:a.call(u,n);n.year=i,n.month=o?1:n.calendar.getMonthsInYear(n),n.day=o?1:n.calendar.getDaysInMonth(n)}n.month<1&&(n.month=1,n.day=1);let l=n.calendar.getMonthsInYear(n);return n.month>l&&(n.month=l,n.day=n.calendar.getDaysInMonth(n)),n.day=Math.max(1,Math.min(n.calendar.getDaysInMonth(n),n.day)),n}function R9(e,t){var n,r;!((n=(r=e.calendar).isInverseEra)===null||n===void 0)&&n.call(r,e)&&(t=-t),e.year+=t}function A9(e){for(;e.month<1;)R9(e,-1),e.month+=e.calendar.getMonthsInYear(e);let t=0;for(;e.month>(t=e.calendar.getMonthsInYear(e));)e.month-=t,R9(e,1)}function iEe(e){for(;e.day<1;)e.month--,A9(e),e.day+=e.calendar.getDaysInMonth(e);for(;e.day>e.calendar.getDaysInMonth(e);)e.day-=e.calendar.getDaysInMonth(e),e.month++,A9(e)}function fne(e){e.month=Math.max(1,Math.min(e.calendar.getMonthsInYear(e),e.month)),e.day=Math.max(1,Math.min(e.calendar.getDaysInMonth(e),e.day))}function o1(e){e.calendar.constrainDate&&e.calendar.constrainDate(e),e.year=Math.max(1,Math.min(e.calendar.getYearsInEra(e),e.year)),fne(e)}function bN(e){let t={};for(let n in e)typeof e[n]=="number"&&(t[n]=-e[n]);return t}function dne(e,t){return UC(e,bN(t))}function ON(e,t){let n=e.copy();return t.era!=null&&(n.era=t.era),t.year!=null&&(n.year=t.year),t.month!=null&&(n.month=t.month),t.day!=null&&(n.day=t.day),o1(n),n}function yx(e,t){let n=e.copy();return t.hour!=null&&(n.hour=t.hour),t.minute!=null&&(n.minute=t.minute),t.second!=null&&(n.second=t.second),t.millisecond!=null&&(n.millisecond=t.millisecond),hne(n),n}function aEe(e){e.second+=Math.floor(e.millisecond/1e3),e.millisecond=U_(e.millisecond,1e3),e.minute+=Math.floor(e.second/60),e.second=U_(e.second,60),e.hour+=Math.floor(e.minute/60),e.minute=U_(e.minute,60);let t=Math.floor(e.hour/24);return e.hour=U_(e.hour,24),t}function hne(e){e.millisecond=Math.max(0,Math.min(e.millisecond,1e3)),e.second=Math.max(0,Math.min(e.second,59)),e.minute=Math.max(0,Math.min(e.minute,59)),e.hour=Math.max(0,Math.min(e.hour,23))}function U_(e,t){let n=e%t;return n<0&&(n+=t),n}function pne(e,t){return e.hour+=t.hours||0,e.minute+=t.minutes||0,e.second+=t.seconds||0,e.millisecond+=t.milliseconds||0,aEe(e)}function mne(e,t){let n=e.copy();return pne(n,t),n}function uEe(e,t){return mne(e,bN(t))}function xN(e,t,n,r){let i=e.copy();switch(t){case"era":{let l=e.calendar.getEras(),o=l.indexOf(e.era);if(o<0)throw new Error("Invalid era: "+e.era);o=Jm(o,n,0,l.length-1,r==null?void 0:r.round),i.era=l[o],o1(i);break}case"year":var a,u;!((a=(u=i.calendar).isInverseEra)===null||a===void 0)&&a.call(u,i)&&(n=-n),i.year=Jm(e.year,n,-1/0,9999,r==null?void 0:r.round),i.year===-1/0&&(i.year=1),i.calendar.balanceYearMonth&&i.calendar.balanceYearMonth(i,e);break;case"month":i.month=Jm(e.month,n,1,e.calendar.getMonthsInYear(e),r==null?void 0:r.round);break;case"day":i.day=Jm(e.day,n,1,e.calendar.getDaysInMonth(e),r==null?void 0:r.round);break;default:throw new Error("Unsupported field "+t)}return e.calendar.balanceDate&&e.calendar.balanceDate(i),o1(i),i}function SN(e,t,n,r){let i=e.copy();switch(t){case"hour":{let a=e.hour,u=0,l=23;if((r==null?void 0:r.hourCycle)===12){let o=a>=12;u=o?12:0,l=o?23:11}i.hour=Jm(a,n,u,l,r==null?void 0:r.round);break}case"minute":i.minute=Jm(e.minute,n,0,59,r==null?void 0:r.round);break;case"second":i.second=Jm(e.second,n,0,59,r==null?void 0:r.round);break;case"millisecond":i.millisecond=Jm(e.millisecond,n,0,999,r==null?void 0:r.round);break;default:throw new Error("Unsupported field "+t)}return i}function Jm(e,t,n,r,i=!1){if(i){e+=Math.sign(t),e<n&&(e=r);let a=Math.abs(t);t>0?e=Math.ceil(e/a)*a:e=Math.floor(e/a)*a,e>r&&(e=n)}else e+=t,e<n?e=r-(n-e-1):e>r&&(e=n+(e-r-1));return e}function gne(e,t){let n;if(t.years!=null&&t.years!==0||t.months!=null&&t.months!==0||t.weeks!=null&&t.weeks!==0||t.days!=null&&t.days!==0){let i=UC(Wv(e),{years:t.years,months:t.months,weeks:t.weeks,days:t.days});n=jm(i,e.timeZone)}else n=R3(e)-e.offset;n+=t.milliseconds||0,n+=(t.seconds||0)*1e3,n+=(t.minutes||0)*6e4,n+=(t.hours||0)*36e5;let r=Wm(n,e.timeZone);return fc(r,e.calendar)}function lEe(e,t){return gne(e,bN(t))}function oEe(e,t,n,r){switch(t){case"hour":{let i=0,a=23;if((r==null?void 0:r.hourCycle)===12){let y=e.hour>=12;i=y?12:0,a=y?23:11}let u=Wv(e),l=fc(yx(u,{hour:i}),new Xf),o=[jm(l,e.timeZone,"earlier"),jm(l,e.timeZone,"later")].filter(y=>Wm(y,e.timeZone).day===l.day)[0],s=fc(yx(u,{hour:a}),new Xf),c=[jm(s,e.timeZone,"earlier"),jm(s,e.timeZone,"later")].filter(y=>Wm(y,e.timeZone).day===s.day).pop(),d=R3(e)-e.offset,h=Math.floor(d/K4),m=d%K4;return d=Jm(h,n,Math.floor(o/K4),Math.floor(c/K4),r==null?void 0:r.round)*K4+m,fc(Wm(d,e.timeZone),e.calendar)}case"minute":case"second":case"millisecond":return SN(e,t,n,r);case"era":case"year":case"month":case"day":{let i=xN(Wv(e),t,n,r),a=jm(i,e.timeZone);return fc(Wm(a,e.timeZone),e.calendar)}default:throw new Error("Unsupported field "+t)}}function sEe(e,t,n){let r=Wv(e),i=yx(ON(r,t),t);if(i.compare(r)===0)return e;let a=jm(i,e.timeZone,n);return fc(Wm(a,e.timeZone),e.calendar)}const cEe=/^([+-]\d{6}|\d{4})-(\d{2})-(\d{2})$/,fEe=/^([+-]\d{6}|\d{4})-(\d{2})-(\d{2})(?:T(\d{2}))?(?::(\d{2}))?(?::(\d{2}))?(\.\d+)?$/,dEe=/^([+-]\d{6}|\d{4})-(\d{2})-(\d{2})(?:T(\d{2}))?(?::(\d{2}))?(?::(\d{2}))?(\.\d+)?(?:(?:([+-]\d{2})(?::?(\d{2}))?)|Z)$/;function hEe(e){let t=e.match(cEe);if(!t)throw new Error("Invalid ISO 8601 date string: "+e);let n=new Xu(Al(t[1],0,9999),Al(t[2],1,12),1);return n.day=Al(t[3],1,n.calendar.getDaysInMonth(n)),n}function pEe(e){let t=e.match(fEe);if(!t)throw new Error("Invalid ISO 8601 date time string: "+e);let n=Al(t[1],-9999,9999),r=n<1?"BC":"AD",i=new bx(r,n<1?-n+1:n,Al(t[2],1,12),1,t[4]?Al(t[4],0,23):0,t[5]?Al(t[5],0,59):0,t[6]?Al(t[6],0,59):0,t[7]?Al(t[7],0,1/0)*1e3:0);return i.day=Al(t[3],0,i.calendar.getDaysInMonth(i)),i}function mEe(e,t){let n=e.match(dEe);if(!n)throw new Error("Invalid ISO 8601 date time string: "+e);let r=Al(n[1],-9999,9999),i=r<1?"BC":"AD",a=new A3(i,r<1?-r+1:r,Al(n[2],1,12),1,t,0,n[4]?Al(n[4],0,23):0,n[5]?Al(n[5],0,59):0,n[6]?Al(n[6],0,59):0,n[7]?Al(n[7],0,1/0)*1e3:0);a.day=Al(n[3],0,a.calendar.getDaysInMonth(a));var u;return n[8]&&(a.offset=Al(n[8],-23,23)*36e5+Al((u=n[9])!==null&&u!==void 0?u:"0",0,59)*6e4),cne(a,t)}function tit(e){return mEe(e,hS())}function Al(e,t,n){let r=Number(e);if(r<t||r>n)throw new RangeError(`Value out of range: ${t} <= ${r} <= ${n}`);return r}function vne(e){return`${String(e.hour).padStart(2,"0")}:${String(e.minute).padStart(2,"0")}:${String(e.second).padStart(2,"0")}${e.millisecond?String(e.millisecond/1e3).slice(1):""}`}function yne(e){let t=fc(e,new Xf),n;return t.era==="BC"?n=t.year===1?"0000":"-"+String(Math.abs(1-t.year)).padStart(6,"00"):n=String(t.year).padStart(4,"0"),`${n}-${String(t.month).padStart(2,"0")}-${String(t.day).padStart(2,"0")}`}function bne(e){return`${yne(e)}T${vne(e)}`}function gEe(e){let t=Math.sign(e)<0?"-":"+";e=Math.abs(e);let n=Math.floor(e/36e5),r=Math.floor(e%36e5/6e4),i=Math.floor(e%36e5%6e4/1e3),a=`${t}${String(n).padStart(2,"0")}:${String(r).padStart(2,"0")}`;return i!==0&&(a+=`:${String(i).padStart(2,"0")}`),a}function vEe(e){return`${bne(e)}${gEe(e.offset)}[${e.timeZone}]`}function yEe(e,t){if(t.has(e))throw new TypeError("Cannot initialize the same private elements twice on an object")}function pS(e,t,n){yEe(e,t),t.set(e,n)}function _N(e){let t=typeof e[0]=="object"?e.shift():new Xf,n;if(typeof e[0]=="string")n=e.shift();else{let u=t.getEras();n=u[u.length-1]}let r=e.shift(),i=e.shift(),a=e.shift();return[t,n,r,i,a]}var bEe=new WeakMap;class Xu{copy(){return this.era?new Xu(this.calendar,this.era,this.year,this.month,this.day):new Xu(this.calendar,this.year,this.month,this.day)}add(t){return UC(this,t)}subtract(t){return dne(this,t)}set(t){return ON(this,t)}cycle(t,n,r){return xN(this,t,n,r)}toDate(t){return lne(this,t)}toString(){return yne(this)}compare(t){return rne(this,t)}constructor(...t){pS(this,bEe,{writable:!0,value:void 0});let[n,r,i,a,u]=_N(t);this.calendar=n,this.era=r,this.year=i,this.month=a,this.day=u,o1(this)}}var OEe=new WeakMap;class jC{copy(){return new jC(this.hour,this.minute,this.second,this.millisecond)}add(t){return mne(this,t)}subtract(t){return uEe(this,t)}set(t){return yx(this,t)}cycle(t,n,r){return SN(this,t,n,r)}toString(){return vne(this)}compare(t){return ine(this,t)}constructor(t=0,n=0,r=0,i=0){pS(this,OEe,{writable:!0,value:void 0}),this.hour=t,this.minute=n,this.second=r,this.millisecond=i,hne(this)}}var xEe=new WeakMap;class bx{copy(){return this.era?new bx(this.calendar,this.era,this.year,this.month,this.day,this.hour,this.minute,this.second,this.millisecond):new bx(this.calendar,this.year,this.month,this.day,this.hour,this.minute,this.second,this.millisecond)}add(t){return UC(this,t)}subtract(t){return dne(this,t)}set(t){return ON(yx(this,t),t)}cycle(t,n,r){switch(t){case"era":case"year":case"month":case"day":return xN(this,t,n,r);default:return SN(this,t,n,r)}}toDate(t,n){return lne(this,t,n)}toString(){return bne(this)}compare(t){let n=rne(this,t);return n===0?ine(this,Wv(t)):n}constructor(...t){pS(this,xEe,{writable:!0,value:void 0});let[n,r,i,a,u]=_N(t);this.calendar=n,this.era=r,this.year=i,this.month=a,this.day=u,this.hour=t.shift()||0,this.minute=t.shift()||0,this.second=t.shift()||0,this.millisecond=t.shift()||0,o1(this)}}var SEe=new WeakMap;class A3{copy(){return this.era?new A3(this.calendar,this.era,this.year,this.month,this.day,this.timeZone,this.offset,this.hour,this.minute,this.second,this.millisecond):new A3(this.calendar,this.year,this.month,this.day,this.timeZone,this.offset,this.hour,this.minute,this.second,this.millisecond)}add(t){return gne(this,t)}subtract(t){return lEe(this,t)}set(t,n){return sEe(this,t,n)}cycle(t,n,r){return oEe(this,t,n,r)}toDate(){return rEe(this)}toString(){return vEe(this)}toAbsoluteString(){return this.toDate().toISOString()}compare(t){return this.toDate().getTime()-sne(t,this.timeZone).toDate().getTime()}constructor(...t){pS(this,SEe,{writable:!0,value:void 0});let[n,r,i,a,u]=_N(t),l=t.shift(),o=t.shift();this.calendar=n,this.era=r,this.year=i,this.month=a,this.day=u,this.timeZone=l,this.offset=o,this.hour=t.shift()||0,this.minute=t.shift()||0,this.second=t.shift()||0,this.millisecond=t.shift()||0,o1(this)}}const l3=[[1868,9,8],[1912,7,30],[1926,12,25],[1989,1,8],[2019,5,1]],_Ee=[[1912,7,29],[1926,12,24],[1989,1,7],[2019,4,30]],eD=[1867,1911,1925,1988,2018],xv=["meiji","taisho","showa","heisei","reiwa"];function qU(e){const t=l3.findIndex(([n,r,i])=>e.year<n||e.year===n&&e.month<r||e.year===n&&e.month===r&&e.day<i);return t===-1?l3.length-1:t===0?0:t-1}function GT(e){let t=eD[xv.indexOf(e.era)];if(!t)throw new Error("Unknown era: "+e.era);return new Xu(e.year+t,e.month,e.day)}class EEe extends Xf{fromJulianDay(t){let n=super.fromJulianDay(t),r=qU(n);return new Xu(this,xv[r],n.year-eD[r],n.month,n.day)}toJulianDay(t){return super.toJulianDay(GT(t))}balanceDate(t){let n=GT(t),r=qU(n);xv[r]!==t.era&&(t.era=xv[r],t.year=n.year-eD[r]),this.constrainDate(t)}constrainDate(t){let n=xv.indexOf(t.era),r=_Ee[n];if(r!=null){let[i,a,u]=r,l=i-eD[n];t.year=Math.max(1,Math.min(l,t.year)),t.year===l&&(t.month=Math.min(a,t.month),t.month===a&&(t.day=Math.min(u,t.day)))}if(t.year===1&&n>=0){let[,i,a]=l3[n];t.month=Math.max(i,t.month),t.month===i&&(t.day=Math.max(a,t.day))}}getEras(){return xv}getYearsInEra(t){let n=xv.indexOf(t.era),r=l3[n],i=l3[n+1];if(i==null)return 9999-r[0]+1;let a=i[0]-r[0];return(t.month<i[1]||t.month===i[1]&&t.day<i[2])&&a++,a}getDaysInMonth(t){return super.getDaysInMonth(GT(t))}getMinimumMonthInYear(t){let n=VU(t);return n?n[1]:1}getMinimumDayInMonth(t){let n=VU(t);return n&&t.month===n[1]?n[2]:1}constructor(...t){super(...t),this.identifier="japanese"}}function VU(e){if(e.year===1){let t=xv.indexOf(e.era);return l3[t]}}const One=-543;class DEe extends Xf{fromJulianDay(t){let n=super.fromJulianDay(t),r=dS(n.era,n.year);return new Xu(this,r-One,n.month,n.day)}toJulianDay(t){return super.toJulianDay(UU(t))}getEras(){return["BE"]}getDaysInMonth(t){return super.getDaysInMonth(UU(t))}balanceDate(){}constructor(...t){super(...t),this.identifier="buddhist"}}function UU(e){let[t,n]=VC(e.year+One);return new Xu(t,n,e.month,e.day)}const GD=1911;function xne(e){return e.era==="minguo"?e.year+GD:1-e.year+GD}function jU(e){let t=e-GD;return t>0?["minguo",t]:["before_minguo",1-t]}class $Ee extends Xf{fromJulianDay(t){let n=super.fromJulianDay(t),r=dS(n.era,n.year),[i,a]=jU(r);return new Xu(this,i,a,n.month,n.day)}toJulianDay(t){return super.toJulianDay(ZU(t))}getEras(){return["before_minguo","minguo"]}balanceDate(t){let[n,r]=jU(xne(t));t.era=n,t.year=r}isInverseEra(t){return t.era==="before_minguo"}getDaysInMonth(t){return super.getDaysInMonth(ZU(t))}getYearsInEra(t){return t.era==="before_minguo"?9999:9999-GD}constructor(...t){super(...t),this.identifier="roc"}}function ZU(e){let[t,n]=VC(xne(e));return new Xu(t,n,e.month,e.day)}const HU=1948320,XU=[0,31,62,93,124,155,186,216,246,276,306,336];class CEe{fromJulianDay(t){let n=t-HU,r=1+Math.floor((33*n+3)/12053),i=365*(r-1)+Math.floor((8*r+21)/33),a=n-i,u=a<216?Math.floor(a/31):Math.floor((a-6)/30),l=a-XU[u]+1;return new Xu(this,r,u+1,l)}toJulianDay(t){let n=HU-1+365*(t.year-1)+Math.floor((8*t.year+21)/33);return n+=XU[t.month-1],n+=t.day,n}getMonthsInYear(){return 12}getDaysInMonth(t){return t.month<=6?31:t.month<=11||u3(25*t.year+11,33)<8?30:29}getEras(){return["AP"]}getYearsInEra(){return 9377}constructor(){this.identifier="persian"}}const KT=78,YU=80;class wEe extends Xf{fromJulianDay(t){let n=super.fromJulianDay(t),r=n.year-KT,i=t-Nb(n.era,n.year,1,1),a;i<YU?(r--,a=Nv(n.year-1)?31:30,i+=a+155+90+10):(a=Nv(n.year)?31:30,i-=YU);let u,l;if(i<a)u=1,l=i+1;else{let o=i-a;o<155?(u=Math.floor(o/31)+2,l=o%31+1):(o-=155,u=Math.floor(o/30)+7,l=o%30+1)}return new Xu(this,r,u,l)}toJulianDay(t){let n=t.year+KT,[r,i]=VC(n),a,u;return Nv(i)?(a=31,u=Nb(r,i,3,21)):(a=30,u=Nb(r,i,3,22)),t.month===1?u+t.day-1:(u+=a+Math.min(t.month-2,5)*31,t.month>=8&&(u+=(t.month-7)*30),u+=t.day-1,u)}getDaysInMonth(t){return t.month===1&&Nv(t.year+KT)||t.month>=2&&t.month<=6?31:30}getYearsInEra(){return 9919}getEras(){return["saka"]}balanceDate(){}constructor(...t){super(...t),this.identifier="indian"}}const KD=1948440,GU=1948439,zd=1300,bO=1600,REe=460322;function WD(e,t,n,r){return r+Math.ceil(29.5*(n-1))+(t-1)*354+Math.floor((3+11*t)/30)+e-1}function Sne(e,t,n){let r=Math.floor((30*(n-t)+10646)/10631),i=Math.min(12,Math.ceil((n-(29+WD(t,r,1,1)))/29.5)+1),a=n-WD(t,r,i,1)+1;return new Xu(e,r,i,a)}function KU(e){return(14+11*e)%30<11}class EN{fromJulianDay(t){return Sne(this,KD,t)}toJulianDay(t){return WD(KD,t.year,t.month,t.day)}getDaysInMonth(t){let n=29+t.month%2;return t.month===12&&KU(t.year)&&n++,n}getMonthsInYear(){return 12}getDaysInYear(t){return KU(t.year)?355:354}getYearsInEra(){return 9665}getEras(){return["AH"]}constructor(){this.identifier="islamic-civil"}}class AEe extends EN{fromJulianDay(t){return Sne(this,GU,t)}toJulianDay(t){return WD(GU,t.year,t.month,t.day)}constructor(...t){super(...t),this.identifier="islamic-tbla"}}const TEe="qgpUDckO1AbqBmwDrQpVBakGkgepC9QF2gpcBS0NlQZKB1QLagutBa4ETwoXBYsGpQbVCtYCWwmdBE0KJg2VDawFtgm6AlsKKwWVCsoG6Qr0AnYJtgJWCcoKpAvSC9kF3AJtCU0FpQpSC6ULtAW2CVcFlwJLBaMGUgdlC2oFqworBZUMSg2lDcoF1gpXCasESwmlClILagt1BXYCtwhbBFUFqQW0BdoJ3QRuAjYJqgpUDbIN1QXaAlsJqwRVCkkLZAtxC7QFtQpVCiUNkg7JDtQG6QprCasEkwpJDaQNsg25CroEWworBZUKKgtVC1wFvQQ9Ah0JlQpKC1oLbQW2AjsJmwRVBqkGVAdqC2wFrQpVBSkLkgupC9QF2gpaBasKlQVJB2QHqgu1BbYCVgpNDiULUgtqC60FrgIvCZcESwalBqwG1gpdBZ0ETQoWDZUNqgW1BdoCWwmtBJUFygbkBuoK9QS2AlYJqgpUC9IL2QXqAm0JrQSVCkoLpQuyBbUJ1gSXCkcFkwZJB1ULagVrCisFiwpGDaMNygXWCtsEawJLCaUKUgtpC3UFdgG3CFsCKwVlBbQF2gntBG0BtgimClINqQ3UBdoKWwmrBFMGKQdiB6kLsgW1ClUFJQuSDckO0gbpCmsFqwRVCikNVA2qDbUJugQ7CpsETQqqCtUK2gJdCV4ELgqaDFUNsga5BroEXQotBZUKUguoC7QLuQXaAloJSgukDdEO6AZqC20FNQWVBkoNqA3UDdoGWwWdAisGFQtKC5ULqgWuCi4JjwwnBZUGqgbWCl0FnQI=";let T9,o3;function tD(e){return REe+o3[e-zd]}function F2(e,t){let n=e-zd,r=1<<11-(t-1);return(T9[n]&r)===0?29:30}function WU(e,t){let n=tD(e);for(let r=1;r<t;r++)n+=F2(e,r);return n}function JU(e){return o3[e+1-zd]-o3[e-zd]}class kEe extends EN{fromJulianDay(t){let n=t-KD,r=tD(zd),i=tD(bO);if(n<r||n>i)return super.fromJulianDay(t);{let a=zd-1,u=1,l=1;for(;l>0;){a++,l=n-tD(a)+1;let o=JU(a);if(l===o){u=12;break}else if(l<o){let s=F2(a,u);for(u=1;l>s;)l-=s,u++,s=F2(a,u);break}}return new Xu(this,a,u,n-WU(a,u)+1)}}toJulianDay(t){return t.year<zd||t.year>bO?super.toJulianDay(t):KD+WU(t.year,t.month)+(t.day-1)}getDaysInMonth(t){return t.year<zd||t.year>bO?super.getDaysInMonth(t):F2(t.year,t.month)}getDaysInYear(t){return t.year<zd||t.year>bO?super.getDaysInYear(t):JU(t.year)}constructor(){if(super(),this.identifier="islamic-umalqura",T9||(T9=new Uint16Array(Uint8Array.from(atob(TEe),t=>t.charCodeAt(0)).buffer)),!o3){o3=new Uint32Array(bO-zd+1);let t=0;for(let n=zd;n<=bO;n++){o3[n-zd]=t;for(let r=1;r<=12;r++)t+=F2(n,r)}}}}const ej=347997,_ne=1080,Ene=24*_ne,PEe=29,MEe=12*_ne+793,FEe=PEe*Ene+MEe;function Pb(e){return u3(e*7+1,19)<7}function nD(e){let t=Math.floor((235*e-234)/19),n=12084+13753*t,r=t*29+Math.floor(n/25920);return u3(3*(r+1),7)<3&&(r+=1),r}function NEe(e){let t=nD(e-1),n=nD(e);return nD(e+1)-n===356?2:n-t===382?1:0}function Z2(e){return nD(e)+NEe(e)}function Dne(e){return Z2(e+1)-Z2(e)}function IEe(e){let t=Dne(e);switch(t>380&&(t-=30),t){case 353:return 0;case 354:return 1;case 355:return 2}}function j_(e,t){if(t>=6&&!Pb(e)&&t++,t===4||t===7||t===9||t===11||t===13)return 29;let n=IEe(e);return t===2?n===2?30:29:t===3?n===0?29:30:t===6?Pb(e)?30:0:30}class LEe{fromJulianDay(t){let n=t-ej,r=n*Ene/FEe,i=Math.floor((19*r+234)/235)+1,a=Z2(i),u=Math.floor(n-a);for(;u<1;)i--,a=Z2(i),u=Math.floor(n-a);let l=1,o=0;for(;o<u;)o+=j_(i,l),l++;l--,o-=j_(i,l);let s=u-o;return new Xu(this,i,l,s)}toJulianDay(t){let n=Z2(t.year);for(let r=1;r<t.month;r++)n+=j_(t.year,r);return n+t.day+ej}getDaysInMonth(t){return j_(t.year,t.month)}getMonthsInYear(t){return Pb(t.year)?13:12}getDaysInYear(t){return Dne(t.year)}getYearsInEra(){return 9999}getEras(){return["AM"]}balanceYearMonth(t,n){n.year!==t.year&&(Pb(n.year)&&!Pb(t.year)&&n.month>6?t.month--:!Pb(n.year)&&Pb(t.year)&&n.month>6&&t.month++)}constructor(){this.identifier="hebrew"}}const k9=1723856,tj=1824665,P9=5500;function JD(e,t,n,r){return e+365*t+Math.floor(t/4)+30*(n-1)+r-1}function DN(e,t){let n=Math.floor(4*(t-e)/1461),r=1+Math.floor((t-JD(e,n,1,1))/30),i=t+1-JD(e,n,r,1);return[n,r,i]}function $ne(e){return Math.floor(e%4/3)}function Cne(e,t){return t%13!==0?30:$ne(e)+5}class $N{fromJulianDay(t){let[n,r,i]=DN(k9,t),a="AM";return n<=0&&(a="AA",n+=P9),new Xu(this,a,n,r,i)}toJulianDay(t){let n=t.year;return t.era==="AA"&&(n-=P9),JD(k9,n,t.month,t.day)}getDaysInMonth(t){return Cne(t.year,t.month)}getMonthsInYear(){return 13}getDaysInYear(t){return 365+$ne(t.year)}getYearsInEra(t){return t.era==="AA"?9999:9991}getEras(){return["AA","AM"]}constructor(){this.identifier="ethiopic"}}class BEe extends $N{fromJulianDay(t){let[n,r,i]=DN(k9,t);return n+=P9,new Xu(this,"AA",n,r,i)}getEras(){return["AA"]}getYearsInEra(){return 9999}constructor(...t){super(...t),this.identifier="ethioaa"}}class QEe extends $N{fromJulianDay(t){let[n,r,i]=DN(tj,t),a="CE";return n<=0&&(a="BCE",n=1-n),new Xu(this,a,n,r,i)}toJulianDay(t){let n=t.year;return t.era==="BCE"&&(n=1-n),JD(tj,n,t.month,t.day)}getDaysInMonth(t){let n=t.year;return t.era==="BCE"&&(n=1-n),Cne(n,t.month)}isInverseEra(t){return t.era==="BCE"}balanceDate(t){t.year<=0&&(t.era=t.era==="BCE"?"CE":"BCE",t.year=1-t.year)}getEras(){return["BCE","CE"]}getYearsInEra(t){return t.era==="BCE"?9999:9715}constructor(...t){super(...t),this.identifier="coptic"}}function wne(e){switch(e){case"buddhist":return new DEe;case"ethiopic":return new $N;case"ethioaa":return new BEe;case"coptic":return new QEe;case"hebrew":return new LEe;case"indian":return new wEe;case"islamic-civil":return new EN;case"islamic-tbla":return new AEe;case"islamic-umalqura":return new kEe;case"japanese":return new EEe;case"persian":return new CEe;case"roc":return new $Ee;case"gregory":default:return new Xf}}let WT=new Map;class UO{format(t){return this.formatter.format(t)}formatToParts(t){return this.formatter.formatToParts(t)}formatRange(t,n){if(typeof this.formatter.formatRange=="function")return this.formatter.formatRange(t,n);if(n<t)throw new RangeError("End date must be >= start date");return`${this.formatter.format(t)} – ${this.formatter.format(n)}`}formatRangeToParts(t,n){if(typeof this.formatter.formatRangeToParts=="function")return this.formatter.formatRangeToParts(t,n);if(n<t)throw new RangeError("End date must be >= start date");let r=this.formatter.formatToParts(t),i=this.formatter.formatToParts(n);return[...r.map(a=>({...a,source:"startRange"})),{type:"literal",value:" – ",source:"shared"},...i.map(a=>({...a,source:"endRange"}))]}resolvedOptions(){let t=this.formatter.resolvedOptions();return VEe()&&(this.resolvedHourCycle||(this.resolvedHourCycle=UEe(t.locale,this.options)),t.hourCycle=this.resolvedHourCycle,t.hour12=this.resolvedHourCycle==="h11"||this.resolvedHourCycle==="h12"),t.calendar==="ethiopic-amete-alem"&&(t.calendar="ethioaa"),t}constructor(t,n={}){this.formatter=Rne(t,n),this.options=n}}const zEe={true:{ja:"h11"},false:{}};function Rne(e,t={}){if(typeof t.hour12=="boolean"&&qEe()){t={...t};let i=zEe[String(t.hour12)][e.split("-")[0]],a=t.hour12?"h12":"h23";t.hourCycle=i??a,delete t.hour12}let n=e+(t?Object.entries(t).sort((i,a)=>i[0]<a[0]?-1:1).join():"");if(WT.has(n))return WT.get(n);let r=new Intl.DateTimeFormat(e,t);return WT.set(n,r),r}let JT=null;function qEe(){return JT==null&&(JT=new Intl.DateTimeFormat("en-US",{hour:"numeric",hour12:!1}).format(new Date(2020,2,3,0))==="24"),JT}let ek=null;function VEe(){return ek==null&&(ek=new Intl.DateTimeFormat("fr",{hour:"numeric",hour12:!1}).resolvedOptions().hourCycle==="h12"),ek}function UEe(e,t){if(!t.timeStyle&&!t.hour)return;e=e.replace(/(-u-)?-nu-[a-zA-Z0-9]+/,""),e+=(e.includes("-u-")?"":"-u")+"-nu-latn";let n=Rne(e,{...t,timeZone:void 0}),r=parseInt(n.formatToParts(new Date(2020,2,3,0)).find(a=>a.type==="hour").value,10),i=parseInt(n.formatToParts(new Date(2020,2,3,23)).find(a=>a.type==="hour").value,10);if(r===0&&i===23)return"h23";if(r===24&&i===23)return"h24";if(r===0&&i===11)return"h11";if(r===12&&i===11)return"h12";throw new Error("Unexpected hour cycle result")}function Z_(e){e=y_e(e??{},jEe);let{locale:t}=Wi();return P.useMemo(()=>new UO(t,e),[t,e])}function jEe(e,t){if(e===t)return!0;let n=Object.keys(e),r=Object.keys(t);if(n.length!==r.length)return!1;for(let i of n)if(t[i]!==e[i])return!1;return!0}let tk=new Map,M9=!1;try{M9=new Intl.NumberFormat("de-DE",{signDisplay:"exceptZero"}).resolvedOptions().signDisplay==="exceptZero"}catch{}let e$=!1;try{e$=new Intl.NumberFormat("de-DE",{style:"unit",unit:"degree"}).resolvedOptions().style==="unit"}catch{}const Ane={degree:{narrow:{default:"°","ja-JP":" 度","zh-TW":"度","sl-SI":" °"}}};class Jv{format(t){let n="";if(!M9&&this.options.signDisplay!=null?n=HEe(this.numberFormatter,this.options.signDisplay,t):n=this.numberFormatter.format(t),this.options.style==="unit"&&!e$){var r;let{unit:i,unitDisplay:a="short",locale:u}=this.resolvedOptions();if(!i)return n;let l=(r=Ane[i])===null||r===void 0?void 0:r[a];n+=l[u]||l.default}return n}formatToParts(t){return this.numberFormatter.formatToParts(t)}formatRange(t,n){if(typeof this.numberFormatter.formatRange=="function")return this.numberFormatter.formatRange(t,n);if(n<t)throw new RangeError("End date must be >= start date");return`${this.format(t)} – ${this.format(n)}`}formatRangeToParts(t,n){if(typeof this.numberFormatter.formatRangeToParts=="function")return this.numberFormatter.formatRangeToParts(t,n);if(n<t)throw new RangeError("End date must be >= start date");let r=this.numberFormatter.formatToParts(t),i=this.numberFormatter.formatToParts(n);return[...r.map(a=>({...a,source:"startRange"})),{type:"literal",value:" – ",source:"shared"},...i.map(a=>({...a,source:"endRange"}))]}resolvedOptions(){let t=this.numberFormatter.resolvedOptions();return!M9&&this.options.signDisplay!=null&&(t={...t,signDisplay:this.options.signDisplay}),!e$&&this.options.style==="unit"&&(t={...t,style:"unit",unit:this.options.unit,unitDisplay:this.options.unitDisplay}),t}constructor(t,n={}){this.numberFormatter=ZEe(t,n),this.options=n}}function ZEe(e,t={}){let{numberingSystem:n}=t;if(n&&e.includes("-nu-")&&(e.includes("-u-")||(e+="-u-"),e+=`-nu-${n}`),t.style==="unit"&&!e$){var r;let{unit:u,unitDisplay:l="short"}=t;if(!u)throw new Error('unit option must be provided with style: "unit"');if(!(!((r=Ane[u])===null||r===void 0)&&r[l]))throw new Error(`Unsupported unit ${u} with unitDisplay = ${l}`);t={...t,style:"decimal"}}let i=e+(t?Object.entries(t).sort((u,l)=>u[0]<l[0]?-1:1).join():"");if(tk.has(i))return tk.get(i);let a=new Intl.NumberFormat(e,t);return tk.set(i,a),a}function HEe(e,t,n){if(t==="auto")return e.format(n);if(t==="never")return e.format(Math.abs(n));{let r=!1;if(t==="always"?r=n>0||Object.is(n,0):t==="exceptZero"&&(Object.is(n,-0)||Object.is(n,0)?n=Math.abs(n):r=n>0),r){let i=e.format(-n),a=e.format(n),u=i.replace(a,"").replace(/\u200e|\u061C/,"");return[...u].length!==1&&console.warn("@react-aria/i18n polyfill for NumberFormat signDisplay: Unsupported case"),i.replace(a,"!!!").replace(u,"+").replace("!!!",a)}else return e.format(n)}}const XEe=new RegExp("^.*\\(.*\\).*$"),YEe=["latn","arab","hanidec","deva","beng","fullwide"];class CN{parse(t){return nk(this.locale,this.options,t).parse(t)}isValidPartialNumber(t,n,r){return nk(this.locale,this.options,t).isValidPartialNumber(t,n,r)}getNumberingSystem(t){return nk(this.locale,this.options,t).options.numberingSystem}constructor(t,n={}){this.locale=t,this.options=n}}const nj=new Map;function nk(e,t,n){let r=rj(e,t);if(!e.includes("-nu-")&&!r.isValidPartialNumber(n)){for(let i of YEe)if(i!==r.options.numberingSystem){let a=rj(e+(e.includes("-u-")?"-nu-":"-u-nu-")+i,t);if(a.isValidPartialNumber(n))return a}}return r}function rj(e,t){let n=e+(t?Object.entries(t).sort((i,a)=>i[0]<a[0]?-1:1).join():""),r=nj.get(n);return r||(r=new GEe(e,t),nj.set(n,r)),r}class GEe{parse(t){let n=this.sanitize(t);if(this.symbols.group&&(n=OO(n,this.symbols.group,"")),this.symbols.decimal&&(n=n.replace(this.symbols.decimal,".")),this.symbols.minusSign&&(n=n.replace(this.symbols.minusSign,"-")),n=n.replace(this.symbols.numeral,this.symbols.index),this.options.style==="percent"){let u=n.indexOf("-");n=n.replace("-",""),n=n.replace("+","");let l=n.indexOf(".");l===-1&&(l=n.length),n=n.replace(".",""),l-2===0?n=`0.${n}`:l-2===-1?n=`0.0${n}`:l-2===-2?n="0.00":n=`${n.slice(0,l-2)}.${n.slice(l-2)}`,u>-1&&(n=`-${n}`)}let r=n?+n:NaN;if(isNaN(r))return NaN;if(this.options.style==="percent"){var i,a;let u={...this.options,style:"decimal",minimumFractionDigits:Math.min(((i=this.options.minimumFractionDigits)!==null&&i!==void 0?i:0)+2,20),maximumFractionDigits:Math.min(((a=this.options.maximumFractionDigits)!==null&&a!==void 0?a:0)+2,20)};return new CN(this.locale,u).parse(new Jv(this.locale,u).format(r))}return this.options.currencySign==="accounting"&&XEe.test(t)&&(r=-1*r),r}sanitize(t){return t=t.replace(this.symbols.literals,""),this.symbols.minusSign&&(t=t.replace("-",this.symbols.minusSign)),this.options.numberingSystem==="arab"&&(this.symbols.decimal&&(t=t.replace(",",this.symbols.decimal),t=t.replace("،",this.symbols.decimal)),this.symbols.group&&(t=OO(t,".",this.symbols.group))),this.symbols.group==="’"&&t.includes("'")&&(t=OO(t,"'",this.symbols.group)),this.options.locale==="fr-FR"&&this.symbols.group&&(t=OO(t," ",this.symbols.group),t=OO(t,/\u00A0/g,this.symbols.group)),t}isValidPartialNumber(t,n=-1/0,r=1/0){return t=this.sanitize(t),this.symbols.minusSign&&t.startsWith(this.symbols.minusSign)&&n<0?t=t.slice(this.symbols.minusSign.length):this.symbols.plusSign&&t.startsWith(this.symbols.plusSign)&&r>0&&(t=t.slice(this.symbols.plusSign.length)),this.symbols.group&&t.startsWith(this.symbols.group)||this.symbols.decimal&&t.indexOf(this.symbols.decimal)>-1&&this.options.maximumFractionDigits===0?!1:(this.symbols.group&&(t=OO(t,this.symbols.group,"")),t=t.replace(this.symbols.numeral,""),this.symbols.decimal&&(t=t.replace(this.symbols.decimal,"")),t.length===0)}constructor(t,n={}){this.locale=t,n.roundingIncrement!==1&&n.roundingIncrement!=null&&(n.maximumFractionDigits==null&&n.minimumFractionDigits==null?(n.maximumFractionDigits=0,n.minimumFractionDigits=0):n.maximumFractionDigits==null?n.maximumFractionDigits=n.minimumFractionDigits:n.minimumFractionDigits==null&&(n.minimumFractionDigits=n.maximumFractionDigits)),this.formatter=new Intl.NumberFormat(t,n),this.options=this.formatter.resolvedOptions(),this.symbols=WEe(t,this.formatter,this.options,n);var r,i;this.options.style==="percent"&&(((r=this.options.minimumFractionDigits)!==null&&r!==void 0?r:0)>18||((i=this.options.maximumFractionDigits)!==null&&i!==void 0?i:0)>18)&&console.warn("NumberParser cannot handle percentages with greater than 18 decimal places, please reduce the number in your options.")}}const ij=new Set(["decimal","fraction","integer","minusSign","plusSign","group"]),KEe=[0,4,2,1,11,20,3,7,100,21,.1,1.1];function WEe(e,t,n,r){var i,a,u,l;let o=new Intl.NumberFormat(e,{...n,minimumSignificantDigits:1,maximumSignificantDigits:21,roundingIncrement:1,roundingPriority:"auto",roundingMode:"halfExpand"}),s=o.formatToParts(-10000.111),c=o.formatToParts(10000.111),d=KEe.map(U=>o.formatToParts(U));var h;let m=(h=(i=s.find(U=>U.type==="minusSign"))===null||i===void 0?void 0:i.value)!==null&&h!==void 0?h:"-",y=(a=c.find(U=>U.type==="plusSign"))===null||a===void 0?void 0:a.value;!y&&((r==null?void 0:r.signDisplay)==="exceptZero"||(r==null?void 0:r.signDisplay)==="always")&&(y="+");let S=(u=new Intl.NumberFormat(e,{...n,minimumFractionDigits:2,maximumFractionDigits:2}).formatToParts(.001).find(U=>U.type==="decimal"))===null||u===void 0?void 0:u.value,_=(l=s.find(U=>U.type==="group"))===null||l===void 0?void 0:l.value,E=s.filter(U=>!ij.has(U.type)).map(U=>aj(U.value)),$=d.flatMap(U=>U.filter(q=>!ij.has(q.type)).map(q=>aj(q.value))),k=[...new Set([...E,...$])].sort((U,q)=>q.length-U.length),A=k.length===0?new RegExp("[\\p{White_Space}]","gu"):new RegExp(`${k.join("|")}|[\\p{White_Space}]`,"gu"),w=[...new Intl.NumberFormat(n.locale,{useGrouping:!1}).format(9876543210)].reverse(),R=new Map(w.map((U,q)=>[U,q])),F=new RegExp(`[${w.join("")}]`,"g");return{minusSign:m,plusSign:y,decimal:S,group:_,literals:A,numeral:F,index:U=>String(R.get(U))}}function OO(e,t,n){return e.replaceAll?e.replaceAll(t,n):e.split(t).join(n)}function aj(e){return e.replace(/[.*+?^${}()|[\]\\]/g,"\\$&")}function t$(e={}){let{locale:t}=Wi();return P.useMemo(()=>new Jv(t,e),[t,e])}let rk=new Map;function ZC(e){let{locale:t}=Wi(),n=t+(e?Object.entries(e).sort((i,a)=>i[0]<a[0]?-1:1).join():"");if(rk.has(n))return rk.get(n);let r=new Intl.Collator(t,e);return rk.set(n,r),r}function Tne(e){let t=ZC({usage:"search",...e}),n=P.useCallback((a,u)=>u.length===0?!0:(a=a.normalize("NFC"),u=u.normalize("NFC"),t.compare(a.slice(0,u.length),u)===0),[t]),r=P.useCallback((a,u)=>u.length===0?!0:(a=a.normalize("NFC"),u=u.normalize("NFC"),t.compare(a.slice(-u.length),u)===0),[t]),i=P.useCallback((a,u)=>{if(u.length===0)return!0;a=a.normalize("NFC"),u=u.normalize("NFC");let l=0,o=u.length;for(;l+o<=a.length;l++){let s=a.slice(l,l+o);if(t.compare(u,s)===0)return!0}return!1},[t]);return P.useMemo(()=>({startsWith:n,endsWith:r,contains:i}),[n,r,i])}let qu=typeof document<"u"?window.visualViewport:null;function kne(e){let{direction:t}=Wi(),{arrowSize:n,targetRef:r,overlayRef:i,arrowRef:a,scrollRef:u=i,placement:l="bottom",containerPadding:o=12,shouldFlip:s=!0,boundaryElement:c=typeof document<"u"?document.body:null,offset:d=0,crossOffset:h=0,shouldUpdatePosition:m=!0,isOpen:y=!0,onClose:v,maxHeight:S,arrowBoundaryOffset:_=0}=e,[E,$]=P.useState(null),k=[m,l,i.current,r.current,a==null?void 0:a.current,u.current,o,s,c,d,h,y,t,S,_,n],A=P.useRef(qu==null?void 0:qu.scale);P.useEffect(()=>{y&&(A.current=qu==null?void 0:qu.scale)},[y]);let w=P.useCallback(()=>{if(m===!1||!y||!i.current||!r.current||!c||(qu==null?void 0:qu.scale)!==A.current)return;let V=null;if(u.current&&u.current.contains(document.activeElement)){var N;let M=(N=document.activeElement)===null||N===void 0?void 0:N.getBoundingClientRect(),G=u.current.getBoundingClientRect();var j;if(V={type:"top",offset:((j=M==null?void 0:M.top)!==null&&j!==void 0?j:0)-G.top},V.offset>G.height/2){V.type="bottom";var W;V.offset=((W=M==null?void 0:M.bottom)!==null&&W!==void 0?W:0)-G.bottom}}let K=i.current;if(!S&&i.current){var J;K.style.top="0px",K.style.bottom="";var te;K.style.maxHeight=((te=(J=window.visualViewport)===null||J===void 0?void 0:J.height)!==null&&te!==void 0?te:window.innerHeight)+"px"}let H=M_e({placement:eDe(l,t),overlayNode:i.current,targetNode:r.current,scrollNode:u.current||i.current,padding:o,shouldFlip:s,boundaryElement:c,offset:d,crossOffset:h,maxHeight:S,arrowSize:n??(a!=null&&a.current?vN(a.current,!0).width:0),arrowBoundaryOffset:_});if(H.position){if(K.style.top="",K.style.bottom="",K.style.left="",K.style.right="",Object.keys(H.position).forEach(M=>K.style[M]=H.position[M]+"px"),K.style.maxHeight=H.maxHeight!=null?H.maxHeight+"px":"",V&&document.activeElement&&u.current){let M=document.activeElement.getBoundingClientRect(),G=u.current.getBoundingClientRect(),X=M[V.type]-G[V.type];u.current.scrollTop+=X-V.offset}$(H)}},k);Hr(w,k),JEe(w),u1({ref:i,onResize:w}),u1({ref:r,onResize:w});let R=P.useRef(!1);Hr(()=>{let V,N=()=>{R.current=!0,clearTimeout(V),V=setTimeout(()=>{R.current=!1},500),w()},j=()=>{R.current&&N()};return qu==null||qu.addEventListener("resize",N),qu==null||qu.addEventListener("scroll",j),()=>{qu==null||qu.removeEventListener("resize",N),qu==null||qu.removeEventListener("scroll",j)}},[w]);let F=P.useCallback(()=>{R.current||v==null||v()},[v,R]);N_e({triggerRef:r,isOpen:y,onClose:v&&F});var I,U,q;return{overlayProps:{style:{position:E?"absolute":"fixed",top:E?void 0:0,left:E?void 0:0,zIndex:1e5,...E==null?void 0:E.position,maxHeight:(I=E==null?void 0:E.maxHeight)!==null&&I!==void 0?I:"100vh"}},placement:(U=E==null?void 0:E.placement)!==null&&U!==void 0?U:null,triggerAnchorPoint:(q=E==null?void 0:E.triggerAnchorPoint)!==null&&q!==void 0?q:null,arrowProps:{"aria-hidden":"true",role:"presentation",style:{left:E==null?void 0:E.arrowOffsetLeft,top:E==null?void 0:E.arrowOffsetTop}},updatePosition:w}}function JEe(e){Hr(()=>(window.addEventListener("resize",e,!1),()=>{window.removeEventListener("resize",e,!1)}),[e])}function eDe(e,t){return t==="rtl"?e.replace("start","right").replace("end","left"):e.replace("start","left").replace("end","right")}function wN(e){let t=e;return t.nativeEvent=e,t.isDefaultPrevented=()=>t.defaultPrevented,t.isPropagationStopped=()=>t.cancelBubble,t.persist=()=>{},t}function Pne(e,t){Object.defineProperty(e,"target",{value:t}),Object.defineProperty(e,"currentTarget",{value:t})}function Mne(e){let t=P.useRef({isFocused:!1,observer:null});Hr(()=>{const r=t.current;return()=>{r.observer&&(r.observer.disconnect(),r.observer=null)}},[]);let n=fi(r=>{e==null||e(r)});return P.useCallback(r=>{if(r.target instanceof HTMLButtonElement||r.target instanceof HTMLInputElement||r.target instanceof HTMLTextAreaElement||r.target instanceof HTMLSelectElement){t.current.isFocused=!0;let i=r.target,a=u=>{if(t.current.isFocused=!1,i.disabled){let l=wN(u);n(l)}t.current.observer&&(t.current.observer.disconnect(),t.current.observer=null)};i.addEventListener("focusout",a,{once:!0}),t.current.observer=new MutationObserver(()=>{if(t.current.isFocused&&i.disabled){var u;(u=t.current.observer)===null||u===void 0||u.disconnect();let l=i===document.activeElement?null:document.activeElement;i.dispatchEvent(new FocusEvent("blur",{relatedTarget:l})),i.dispatchEvent(new FocusEvent("focusout",{bubbles:!0,relatedTarget:l}))}}),t.current.observer.observe(i,{attributes:!0,attributeFilter:["disabled"]})}},[n])}let n$=!1;function tDe(e){for(;e&&!fte(e);)e=e.parentElement;let t=uf(e),n=t.document.activeElement;if(!n||n===e)return;n$=!0;let r=!1,i=c=>{(c.target===n||r)&&c.stopImmediatePropagation()},a=c=>{(c.target===n||r)&&(c.stopImmediatePropagation(),!e&&!r&&(r=!0,$u(n),o()))},u=c=>{(c.target===e||r)&&c.stopImmediatePropagation()},l=c=>{(c.target===e||r)&&(c.stopImmediatePropagation(),r||(r=!0,$u(n),o()))};t.addEventListener("blur",i,!0),t.addEventListener("focusout",a,!0),t.addEventListener("focusin",l,!0),t.addEventListener("focus",u,!0);let o=()=>{cancelAnimationFrame(s),t.removeEventListener("blur",i,!0),t.removeEventListener("focusout",a,!0),t.removeEventListener("focusin",l,!0),t.removeEventListener("focus",u,!0),n$=!1,r=!1},s=requestAnimationFrame(o);return o}let jO="default",F9="",rD=new WeakMap;function Fne(e){if($1()){if(jO==="default"){const t=hi(e);F9=t.documentElement.style.webkitUserSelect,t.documentElement.style.webkitUserSelect="none"}jO="disabled"}else if(e instanceof HTMLElement||e instanceof SVGElement){let t="userSelect"in e.style?"userSelect":"webkitUserSelect";rD.set(e,e.style[t]),e.style[t]="none"}}function N9(e){if($1()){if(jO!=="disabled")return;jO="restoring",setTimeout(()=>{ite(()=>{if(jO==="restoring"){const t=hi(e);t.documentElement.style.webkitUserSelect==="none"&&(t.documentElement.style.webkitUserSelect=F9||""),F9="",jO="default"}})},300)}else if((e instanceof HTMLElement||e instanceof SVGElement)&&e&&rD.has(e)){let t=rD.get(e),n="userSelect"in e.style?"userSelect":"webkitUserSelect";e.style[n]==="none"&&(e.style[n]=t),e.getAttribute("style")===""&&e.removeAttribute("style"),rD.delete(e)}}const Ox=be.createContext({register:()=>{}});Ox.displayName="PressResponderContext";function nDe(e,t){return t.get?t.get.call(e):t.value}function Nne(e,t,n){if(!t.has(e))throw new TypeError("attempted to "+n+" private field on non-instance");return t.get(e)}function rDe(e,t){var n=Nne(e,t,"get");return nDe(e,n)}function iDe(e,t,n){if(t.set)t.set.call(e,n);else{if(!t.writable)throw new TypeError("attempted to set read only private field");t.value=n}}function uj(e,t,n){var r=Nne(e,t,"set");return iDe(e,r,n),n}function aDe(e){let t=P.useContext(Ox);if(t){let{register:n,...r}=t;e=Et(r,e),n()}return pN(t,e.ref),e}var H_=new WeakMap;class X_{continuePropagation(){uj(this,H_,!1)}get shouldStopPropagation(){return rDe(this,H_)}constructor(t,n,r,i){pS(this,H_,{writable:!0,value:void 0}),uj(this,H_,!0);var a;let u=(a=i==null?void 0:i.target)!==null&&a!==void 0?a:r.currentTarget;const l=u==null?void 0:u.getBoundingClientRect();let o,s=0,c,d=null;r.clientX!=null&&r.clientY!=null&&(c=r.clientX,d=r.clientY),l&&(c!=null&&d!=null?(o=c-l.left,s=d-l.top):(o=l.width/2,s=l.height/2)),this.type=t,this.pointerType=n,this.target=r.currentTarget,this.shiftKey=r.shiftKey,this.metaKey=r.metaKey,this.ctrlKey=r.ctrlKey,this.altKey=r.altKey,this.x=o,this.y=s}}const lj=Symbol("linkClicked"),oj="react-aria-pressable-style",sj="data-react-aria-pressable";function lp(e){let{onPress:t,onPressChange:n,onPressStart:r,onPressEnd:i,onPressUp:a,onClick:u,isDisabled:l,isPressed:o,preventFocusOnPress:s,shouldCancelOnPointerExit:c,allowTextSelectionOnPress:d,ref:h,...m}=aDe(e),[y,v]=P.useState(!1),S=P.useRef({isPressed:!1,ignoreEmulatedMouseEvents:!1,didFirePressStart:!1,isTriggeringEvent:!1,activePointerId:null,target:null,isOverTarget:!1,pointerType:null,disposables:[]}),{addGlobalListener:_,removeAllGlobalListeners:E}=py(),$=fi((q,V)=>{let N=S.current;if(l||N.didFirePressStart)return!1;let j=!0;if(N.isTriggeringEvent=!0,r){let W=new X_("pressstart",V,q);r(W),j=W.shouldStopPropagation}return n&&n(!0),N.isTriggeringEvent=!1,N.didFirePressStart=!0,v(!0),j}),k=fi((q,V,N=!0)=>{let j=S.current;if(!j.didFirePressStart)return!1;j.didFirePressStart=!1,j.isTriggeringEvent=!0;let W=!0;if(i){let K=new X_("pressend",V,q);i(K),W=K.shouldStopPropagation}if(n&&n(!1),v(!1),t&&N&&!l){let K=new X_("press",V,q);t(K),W&&(W=K.shouldStopPropagation)}return j.isTriggeringEvent=!1,W}),A=fi((q,V)=>{let N=S.current;if(l)return!1;if(a){N.isTriggeringEvent=!0;let j=new X_("pressup",V,q);return a(j),N.isTriggeringEvent=!1,j.shouldStopPropagation}return!0}),w=fi(q=>{let V=S.current;if(V.isPressed&&V.target){V.didFirePressStart&&V.pointerType!=null&&k(mb(V.target,q),V.pointerType,!1),V.isPressed=!1,V.isOverTarget=!1,V.activePointerId=null,V.pointerType=null,E(),d||N9(V.target);for(let N of V.disposables)N();V.disposables=[]}}),R=fi(q=>{c&&w(q)}),F=fi(q=>{l||u==null||u(q)}),I=fi((q,V)=>{if(!l&&u){let N=new MouseEvent("click",q);Pne(N,V),u(wN(N))}}),U=P.useMemo(()=>{let q=S.current,V={onKeyDown(j){if(ik(j.nativeEvent,j.currentTarget)&&Wc(j.currentTarget,ju(j.nativeEvent))){var W;cj(ju(j.nativeEvent),j.key)&&j.preventDefault();let K=!0;if(!q.isPressed&&!j.repeat){q.target=j.currentTarget,q.isPressed=!0,q.pointerType="keyboard",K=$(j,"keyboard");let J=j.currentTarget,te=H=>{ik(H,J)&&!H.repeat&&Wc(J,ju(H))&&q.target&&A(mb(q.target,H),"keyboard")};_(hi(j.currentTarget),"keyup",ih(te,N),!0)}K&&j.stopPropagation(),j.metaKey&&cg()&&((W=q.metaKeyEvents)===null||W===void 0||W.set(j.key,j.nativeEvent))}else j.key==="Meta"&&(q.metaKeyEvents=new Map)},onClick(j){if(!(j&&!Wc(j.currentTarget,ju(j.nativeEvent)))&&j&&j.button===0&&!q.isTriggeringEvent&&!a1.isOpening){let W=!0;if(l&&j.preventDefault(),!q.ignoreEmulatedMouseEvents&&!q.isPressed&&(q.pointerType==="virtual"||lte(j.nativeEvent))){let K=$(j,"virtual"),J=A(j,"virtual"),te=k(j,"virtual");F(j),W=K&&J&&te}else if(q.isPressed&&q.pointerType!=="keyboard"){let K=q.pointerType||j.nativeEvent.pointerType||"virtual",J=A(mb(j.currentTarget,j),K),te=k(mb(j.currentTarget,j),K,!0);W=J&&te,q.isOverTarget=!1,F(j),w(j)}q.ignoreEmulatedMouseEvents=!1,W&&j.stopPropagation()}}},N=j=>{var W;if(q.isPressed&&q.target&&ik(j,q.target)){var K;cj(ju(j),j.key)&&j.preventDefault();let te=ju(j),H=Wc(q.target,ju(j));k(mb(q.target,j),"keyboard",H),H&&I(j,q.target),E(),j.key!=="Enter"&&RN(q.target)&&Wc(q.target,te)&&!j[lj]&&(j[lj]=!0,a1(q.target,j,!1)),q.isPressed=!1,(K=q.metaKeyEvents)===null||K===void 0||K.delete(j.key)}else if(j.key==="Meta"&&(!((W=q.metaKeyEvents)===null||W===void 0)&&W.size)){var J;let te=q.metaKeyEvents;q.metaKeyEvents=void 0;for(let H of te.values())(J=q.target)===null||J===void 0||J.dispatchEvent(new KeyboardEvent("keyup",H))}};if(typeof PointerEvent<"u"){V.onPointerDown=K=>{if(K.button!==0||!Wc(K.currentTarget,ju(K.nativeEvent)))return;if(v_e(K.nativeEvent)){q.pointerType="virtual";return}q.pointerType=K.pointerType;let J=!0;if(!q.isPressed){q.isPressed=!0,q.isOverTarget=!0,q.activePointerId=K.pointerId,q.target=K.currentTarget,d||Fne(q.target),J=$(K,q.pointerType);let te=ju(K.nativeEvent);"releasePointerCapture"in te&&te.releasePointerCapture(K.pointerId),_(hi(K.currentTarget),"pointerup",j,!1),_(hi(K.currentTarget),"pointercancel",W,!1)}J&&K.stopPropagation()},V.onMouseDown=K=>{if(Wc(K.currentTarget,ju(K.nativeEvent))&&K.button===0){if(s){let J=tDe(K.target);J&&q.disposables.push(J)}K.stopPropagation()}},V.onPointerUp=K=>{!Wc(K.currentTarget,ju(K.nativeEvent))||q.pointerType==="virtual"||K.button===0&&!q.isPressed&&A(K,q.pointerType||K.pointerType)},V.onPointerEnter=K=>{K.pointerId===q.activePointerId&&q.target&&!q.isOverTarget&&q.pointerType!=null&&(q.isOverTarget=!0,$(mb(q.target,K),q.pointerType))},V.onPointerLeave=K=>{K.pointerId===q.activePointerId&&q.target&&q.isOverTarget&&q.pointerType!=null&&(q.isOverTarget=!1,k(mb(q.target,K),q.pointerType,!1),R(K))};let j=K=>{if(K.pointerId===q.activePointerId&&q.isPressed&&K.button===0&&q.target){if(Wc(q.target,ju(K))&&q.pointerType!=null){let J=!1,te=setTimeout(()=>{q.isPressed&&q.target instanceof HTMLElement&&(J?w(K):($u(q.target),q.target.click()))},80);_(K.currentTarget,"click",()=>J=!0,!0),q.disposables.push(()=>clearTimeout(te))}else w(K);q.isOverTarget=!1}},W=K=>{w(K)};V.onDragStart=K=>{Wc(K.currentTarget,ju(K.nativeEvent))&&w(K)}}return V},[_,l,s,E,d,w,R,k,$,A,F,I]);return P.useEffect(()=>{if(!h)return;const q=hi(h.current);if(!q||!q.head||q.getElementById(oj))return;const V=q.createElement("style");V.id=oj,V.textContent=`
@layer {
  [${sj}] {
    touch-action: pan-x pan-y pinch-zoom;
  }
}
    `.trim(),q.head.prepend(V)},[h]),P.useEffect(()=>{let q=S.current;return()=>{var V;d||N9((V=q.target)!==null&&V!==void 0?V:void 0);for(let N of q.disposables)N();q.disposables=[]}},[d]),{isPressed:o||y,pressProps:Et(m,U,{[sj]:!0})}}function RN(e){return e.tagName==="A"&&e.hasAttribute("href")}function ik(e,t){const{key:n,code:r}=e,i=t,a=i.getAttribute("role");return(n==="Enter"||n===" "||n==="Spacebar"||r==="Space")&&!(i instanceof uf(i).HTMLInputElement&&!Ine(i,n)||i instanceof uf(i).HTMLTextAreaElement||i.isContentEditable)&&!((a==="link"||!a&&RN(i))&&n!=="Enter")}function mb(e,t){let n=t.clientX,r=t.clientY;return{currentTarget:e,shiftKey:t.shiftKey,ctrlKey:t.ctrlKey,metaKey:t.metaKey,altKey:t.altKey,clientX:n,clientY:r}}function uDe(e){return e instanceof HTMLInputElement?!1:e instanceof HTMLButtonElement?e.type!=="submit"&&e.type!=="reset":!RN(e)}function cj(e,t){return e instanceof HTMLInputElement?!Ine(e,t):uDe(e)}const lDe=new Set(["checkbox","radio","range","color","file","image","button","submit","reset"]);function Ine(e,t){return e.type==="checkbox"||e.type==="radio"?t===" ":lDe.has(e.type)}let fg=null,xx=new Set,H2=new Map,s1=!1,I9=!1;const oDe={Tab:!0,Escape:!0};function HC(e,t){for(let n of xx)n(e,t)}function sDe(e){return!(e.metaKey||!cg()&&e.altKey||e.ctrlKey||e.key==="Control"||e.key==="Shift"||e.key==="Meta")}function r$(e){s1=!0,sDe(e)&&(fg="keyboard",HC("keyboard",e))}function s3(e){fg="pointer",(e.type==="mousedown"||e.type==="pointerdown")&&(s1=!0,HC("pointer",e))}function Lne(e){lte(e)&&(s1=!0,fg="virtual")}function Bne(e){e.target===window||e.target===document||n$||!e.isTrusted||(!s1&&!I9&&(fg="virtual",HC("virtual",e)),s1=!1,I9=!1)}function Qne(){n$||(s1=!1,I9=!0)}function i$(e){if(typeof window>"u"||typeof document>"u"||H2.get(uf(e)))return;const t=uf(e),n=hi(e);let r=t.HTMLElement.prototype.focus;t.HTMLElement.prototype.focus=function(){s1=!0,r.apply(this,arguments)},n.addEventListener("keydown",r$,!0),n.addEventListener("keyup",r$,!0),n.addEventListener("click",Lne,!0),t.addEventListener("focus",Bne,!0),t.addEventListener("blur",Qne,!1),typeof PointerEvent<"u"&&(n.addEventListener("pointerdown",s3,!0),n.addEventListener("pointermove",s3,!0),n.addEventListener("pointerup",s3,!0)),t.addEventListener("beforeunload",()=>{zne(e)},{once:!0}),H2.set(t,{focus:r})}const zne=(e,t)=>{const n=uf(e),r=hi(e);t&&r.removeEventListener("DOMContentLoaded",t),H2.has(n)&&(n.HTMLElement.prototype.focus=H2.get(n).focus,r.removeEventListener("keydown",r$,!0),r.removeEventListener("keyup",r$,!0),r.removeEventListener("click",Lne,!0),n.removeEventListener("focus",Bne,!0),n.removeEventListener("blur",Qne,!1),typeof PointerEvent<"u"&&(r.removeEventListener("pointerdown",s3,!0),r.removeEventListener("pointermove",s3,!0),r.removeEventListener("pointerup",s3,!0)),H2.delete(n))};function cDe(e){const t=hi(e);let n;return t.readyState!=="loading"?i$(e):(n=()=>{i$(e)},t.addEventListener("DOMContentLoaded",n)),()=>zne(e,n)}typeof document<"u"&&cDe();function ey(){return fg!=="pointer"}function ig(){return fg}function AN(e){fg=e,HC(e,null)}function TN(){i$();let[e,t]=P.useState(fg);return P.useEffect(()=>{let n=()=>{t(fg)};return xx.add(n),()=>{xx.delete(n)}},[]),pp()?null:e}const fDe=new Set(["checkbox","radio","range","color","file","image","button","submit","reset"]);function dDe(e,t,n){let r=hi(n==null?void 0:n.target);const i=typeof window<"u"?uf(n==null?void 0:n.target).HTMLInputElement:HTMLInputElement,a=typeof window<"u"?uf(n==null?void 0:n.target).HTMLTextAreaElement:HTMLTextAreaElement,u=typeof window<"u"?uf(n==null?void 0:n.target).HTMLElement:HTMLElement,l=typeof window<"u"?uf(n==null?void 0:n.target).KeyboardEvent:KeyboardEvent;return e=e||r.activeElement instanceof i&&!fDe.has(r.activeElement.type)||r.activeElement instanceof a||r.activeElement instanceof u&&r.activeElement.isContentEditable,!(e&&t==="keyboard"&&n instanceof l&&!oDe[n.key])}function hDe(e,t,n){i$(),P.useEffect(()=>{let r=(i,a)=>{dDe(!!(n!=null&&n.isTextInput),i,a)&&e(ey())};return xx.add(r),()=>{xx.delete(r)}},t)}function lc(e){const t=hi(e),n=Zu(t);if(ig()==="virtual"){let r=n;ite(()=>{Zu(t)===r&&e.isConnected&&$u(e)})}else $u(e)}function XC(e){let{isDisabled:t,onFocus:n,onBlur:r,onFocusChange:i}=e;const a=P.useCallback(o=>{if(o.target===o.currentTarget)return r&&r(o),i&&i(!1),!0},[r,i]),u=Mne(a),l=P.useCallback(o=>{const s=hi(o.target),c=s?Zu(s):Zu();o.target===o.currentTarget&&c===ju(o.nativeEvent)&&(n&&n(o),i&&i(!0),u(o))},[i,n,u]);return{focusProps:{onFocus:!t&&(n||i||r)?l:void 0,onBlur:!t&&(r||i)?a:void 0}}}function fj(e){if(!e)return;let t=!0;return n=>{let r={...n,preventDefault(){n.preventDefault()},isDefaultPrevented(){return n.isDefaultPrevented()},stopPropagation(){t=!0},continuePropagation(){t=!1},isPropagationStopped(){return t}};e(r),t&&n.stopPropagation()}}function kN(e){return{keyboardProps:e.isDisabled?{}:{onKeyDown:fj(e.onKeyDown),onKeyUp:fj(e.onKeyUp)}}}let a$=be.createContext(null);function pDe(e){let t=P.useContext(a$)||{};pN(t,e);let{ref:n,...r}=t;return r}const mDe=be.forwardRef(function(t,n){let{children:r,...i}=t,a=wu(n),u={...i,ref:a};return be.createElement(a$.Provider,{value:u},r)});function I0(e,t){let{focusProps:n}=XC(e),{keyboardProps:r}=kN(e),i=Et(n,r),a=pDe(t),u=e.isDisabled?{}:a,l=P.useRef(e.autoFocus);P.useEffect(()=>{l.current&&t.current&&lc(t.current),l.current=!1},[t]);let o=e.excludeFromTabOrder?-1:0;return e.isDisabled&&(o=void 0),{focusableProps:Et({...i,tabIndex:o},u)}}const nit=P.forwardRef(({children:e,...t},n)=>{n=wu(n);let{focusableProps:r}=I0(t,n),i=be.Children.only(e);P.useEffect(()=>{},[n,t.isDisabled]);let a=parseInt(be.version,10)<19?i.ref:i.props.ref;return be.cloneElement(i,{...Et(r,i.props),ref:yg(a,n)})}),rit=be.forwardRef(({children:e,...t},n)=>{n=wu(n);let{pressProps:r}=lp({...t,ref:n}),{focusableProps:i}=I0(t,n),a=be.Children.only(e);P.useEffect(()=>{},[n,t.isDisabled]);let u=parseInt(be.version,10)<19?a.ref:a.props.ref;return be.cloneElement(a,{...Et(r,i,a.props),ref:yg(u,n)})}),qne=be.forwardRef(({children:e,...t},n)=>{let r=P.useRef(!1),i=P.useContext(Ox);n=wu(n||(i==null?void 0:i.ref));let a=Et(i||{},{...t,ref:n,register(){r.current=!0,i&&i.register()}});return pN(i,n),P.useEffect(()=>{r.current||(r.current=!0)},[]),be.createElement(Ox.Provider,{value:a},e)});function gDe({children:e}){let t=P.useMemo(()=>({register:()=>{}}),[]);return be.createElement(Ox.Provider,{value:t},e)}function L0(e){let{isDisabled:t,onBlurWithin:n,onFocusWithin:r,onFocusWithinChange:i}=e,a=P.useRef({isFocusWithin:!1}),{addGlobalListener:u,removeAllGlobalListeners:l}=py(),o=P.useCallback(d=>{d.currentTarget.contains(d.target)&&a.current.isFocusWithin&&!d.currentTarget.contains(d.relatedTarget)&&(a.current.isFocusWithin=!1,l(),n&&n(d),i&&i(!1))},[n,i,a,l]),s=Mne(o),c=P.useCallback(d=>{if(!d.currentTarget.contains(d.target))return;const h=hi(d.target),m=Zu(h);if(!a.current.isFocusWithin&&m===ju(d.nativeEvent)){r&&r(d),i&&i(!0),a.current.isFocusWithin=!0,s(d);let y=d.currentTarget;u(h,"focus",v=>{if(a.current.isFocusWithin&&!Wc(y,v.target)){let S=new h.defaultView.FocusEvent("blur",{relatedTarget:v.target});Pne(S,y);let _=wN(S);o(_)}},{capture:!0})}},[r,i,s,u,o]);return t?{focusWithinProps:{onFocus:void 0,onBlur:void 0}}:{focusWithinProps:{onFocus:c,onBlur:o}}}let L9=!1,Y_=0;function vDe(){L9=!0,setTimeout(()=>{L9=!1},50)}function dj(e){e.pointerType==="touch"&&vDe()}function yDe(){if(!(typeof document>"u"))return Y_===0&&typeof PointerEvent<"u"&&document.addEventListener("pointerup",dj),Y_++,()=>{Y_--,!(Y_>0)&&typeof PointerEvent<"u"&&document.removeEventListener("pointerup",dj)}}function Ru(e){let{onHoverStart:t,onHoverChange:n,onHoverEnd:r,isDisabled:i}=e,[a,u]=P.useState(!1),l=P.useRef({isHovered:!1,ignoreEmulatedMouseEvents:!1,pointerType:"",target:null}).current;P.useEffect(yDe,[]);let{addGlobalListener:o,removeAllGlobalListeners:s}=py(),{hoverProps:c,triggerHoverEnd:d}=P.useMemo(()=>{let h=(v,S)=>{if(l.pointerType=S,i||S==="touch"||l.isHovered||!v.currentTarget.contains(v.target))return;l.isHovered=!0;let _=v.currentTarget;l.target=_,o(hi(v.target),"pointerover",E=>{l.isHovered&&l.target&&!Wc(l.target,E.target)&&m(E,E.pointerType)},{capture:!0}),t&&t({type:"hoverstart",target:_,pointerType:S}),n&&n(!0),u(!0)},m=(v,S)=>{let _=l.target;l.pointerType="",l.target=null,!(S==="touch"||!l.isHovered||!_)&&(l.isHovered=!1,s(),r&&r({type:"hoverend",target:_,pointerType:S}),n&&n(!1),u(!1))},y={};return typeof PointerEvent<"u"&&(y.onPointerEnter=v=>{L9&&v.pointerType==="mouse"||h(v,v.pointerType)},y.onPointerLeave=v=>{!i&&v.currentTarget.contains(v.target)&&m(v,v.pointerType)}),{hoverProps:y,triggerHoverEnd:m}},[t,n,r,i,l,o,s]);return P.useEffect(()=>{i&&d({currentTarget:l.target},l.pointerType)},[i]),{hoverProps:c,isHovered:a}}function bDe(e){let{ref:t,onInteractOutside:n,isDisabled:r,onInteractOutsideStart:i}=e,a=P.useRef({isPointerDown:!1,ignoreEmulatedMouseEvents:!1}),u=fi(o=>{n&&hj(o,t)&&(i&&i(o),a.current.isPointerDown=!0)}),l=fi(o=>{n&&n(o)});P.useEffect(()=>{let o=a.current;if(r)return;const s=t.current,c=hi(s);if(typeof PointerEvent<"u"){let d=h=>{o.isPointerDown&&hj(h,t)&&l(h),o.isPointerDown=!1};return c.addEventListener("pointerdown",u,!0),c.addEventListener("click",d,!0),()=>{c.removeEventListener("pointerdown",u,!0),c.removeEventListener("click",d,!0)}}},[t,r,u,l])}function hj(e,t){if(e.button>0)return!1;if(e.target){const n=e.target.ownerDocument;if(!n||!n.documentElement.contains(e.target)||e.target.closest("[data-react-aria-top-layer]"))return!1}return t.current?!e.composedPath().includes(t.current):!1}function Vne(e){let{onMoveStart:t,onMove:n,onMoveEnd:r}=e,i=P.useRef({didMove:!1,lastPosition:null,id:null}),{addGlobalListener:a,removeGlobalListener:u}=py(),l=fi((c,d,h,m)=>{h===0&&m===0||(i.current.didMove||(i.current.didMove=!0,t==null||t({type:"movestart",pointerType:d,shiftKey:c.shiftKey,metaKey:c.metaKey,ctrlKey:c.ctrlKey,altKey:c.altKey})),n==null||n({type:"move",pointerType:d,deltaX:h,deltaY:m,shiftKey:c.shiftKey,metaKey:c.metaKey,ctrlKey:c.ctrlKey,altKey:c.altKey}))}),o=fi((c,d)=>{N9(),i.current.didMove&&(r==null||r({type:"moveend",pointerType:d,shiftKey:c.shiftKey,metaKey:c.metaKey,ctrlKey:c.ctrlKey,altKey:c.altKey}))});return{moveProps:P.useMemo(()=>{let c={},d=()=>{Fne(),i.current.didMove=!1};{let m=v=>{if(v.pointerId===i.current.id){var S,_;let k=v.pointerType||"mouse";var E,$;l(v,k,v.pageX-((E=(S=i.current.lastPosition)===null||S===void 0?void 0:S.pageX)!==null&&E!==void 0?E:0),v.pageY-(($=(_=i.current.lastPosition)===null||_===void 0?void 0:_.pageY)!==null&&$!==void 0?$:0)),i.current.lastPosition={pageX:v.pageX,pageY:v.pageY}}},y=v=>{if(v.pointerId===i.current.id){let S=v.pointerType||"mouse";o(v,S),i.current.id=null,u(window,"pointermove",m,!1),u(window,"pointerup",y,!1),u(window,"pointercancel",y,!1)}};c.onPointerDown=v=>{v.button===0&&i.current.id==null&&(d(),v.stopPropagation(),v.preventDefault(),i.current.lastPosition={pageX:v.pageX,pageY:v.pageY},i.current.id=v.pointerId,a(window,"pointermove",m,!1),a(window,"pointerup",y,!1),a(window,"pointercancel",y,!1))}}let h=(m,y,v)=>{d(),l(m,"keyboard",y,v),o(m,"keyboard")};return c.onKeyDown=m=>{switch(m.key){case"Left":case"ArrowLeft":m.preventDefault(),m.stopPropagation(),h(m,-1,0);break;case"Right":case"ArrowRight":m.preventDefault(),m.stopPropagation(),h(m,1,0);break;case"Up":case"ArrowUp":m.preventDefault(),m.stopPropagation(),h(m,0,-1);break;case"Down":case"ArrowDown":m.preventDefault(),m.stopPropagation(),h(m,0,1);break}},c},[i,a,u,l,o])}}function Une(e,t){let{onScroll:n,isDisabled:r}=e,i=P.useCallback(a=>{a.ctrlKey||(a.preventDefault(),a.stopPropagation(),n&&n({deltaX:a.deltaX,deltaY:a.deltaY}))},[n]);Yd(t,"wheel",r?void 0:i)}const ODe=500;function jne(e){let{isDisabled:t,onLongPressStart:n,onLongPressEnd:r,onLongPress:i,threshold:a=ODe,accessibilityDescription:u}=e;const l=P.useRef(void 0);let{addGlobalListener:o,removeGlobalListener:s}=py(),{pressProps:c}=lp({isDisabled:t,onPressStart(h){if(h.continuePropagation(),(h.pointerType==="mouse"||h.pointerType==="touch")&&(n&&n({...h,type:"longpressstart"}),l.current=setTimeout(()=>{h.target.dispatchEvent(new PointerEvent("pointercancel",{bubbles:!0})),hi(h.target).activeElement!==h.target&&$u(h.target),i&&i({...h,type:"longpress"}),l.current=void 0},a),h.pointerType==="touch")){let m=y=>{y.preventDefault()};o(h.target,"contextmenu",m,{once:!0}),o(window,"pointerup",()=>{setTimeout(()=>{s(h.target,"contextmenu",m)},30)},{once:!0})}},onPressEnd(h){l.current&&clearTimeout(l.current),r&&(h.pointerType==="mouse"||h.pointerType==="touch")&&r({...h,type:"longpressend"})}}),d=zC(i&&!t?u:void 0);return{longPressProps:Et(c,d)}}const B9=be.createContext(null),Q9="react-aria-focus-scope-restore";let Sa=null;function PN(e){let{children:t,contain:n,restoreFocus:r,autoFocus:i}=e,a=P.useRef(null),u=P.useRef(null),l=P.useRef([]),{parentNode:o}=P.useContext(B9)||{},s=P.useMemo(()=>new q9({scopeRef:l}),[l]);Hr(()=>{let h=o||vl.root;if(vl.getTreeNode(h.scopeRef)&&Sa&&!u$(Sa,h.scopeRef)){let m=vl.getTreeNode(Sa);m&&(h=m)}h.addChild(s),vl.addNode(s)},[s,o]),Hr(()=>{let h=vl.getTreeNode(l);h&&(h.contain=!!n)},[n]),Hr(()=>{var h;let m=(h=a.current)===null||h===void 0?void 0:h.nextSibling,y=[],v=S=>S.stopPropagation();for(;m&&m!==u.current;)y.push(m),m.addEventListener(Q9,v),m=m.nextSibling;return l.current=y,()=>{for(let S of y)S.removeEventListener(Q9,v)}},[t]),$De(l,r,n),_De(l,n),CDe(l,r,n),DDe(l,i),P.useEffect(()=>{const h=Zu(hi(l.current?l.current[0]:void 0));let m=null;if(Jd(h,l.current)){for(let y of vl.traverse())y.scopeRef&&Jd(h,y.scopeRef.current)&&(m=y);m===vl.getTreeNode(l)&&(Sa=m.scopeRef)}},[l]),Hr(()=>()=>{var h,m,y;let v=(y=(m=vl.getTreeNode(l))===null||m===void 0||(h=m.parent)===null||h===void 0?void 0:h.scopeRef)!==null&&y!==void 0?y:null;(l===Sa||u$(l,Sa))&&(!v||vl.getTreeNode(v))&&(Sa=v),vl.removeTreeNode(l)},[l]);let c=P.useMemo(()=>xDe(l),[]),d=P.useMemo(()=>({focusManager:c,parentNode:s}),[s,c]);return be.createElement(B9.Provider,{value:d},be.createElement("span",{"data-focus-scope-start":!0,hidden:!0,ref:a}),t,be.createElement("span",{"data-focus-scope-end":!0,hidden:!0,ref:u}))}function iit(){var e;return(e=P.useContext(B9))===null||e===void 0?void 0:e.focusManager}function xDe(e){return{focusNext(t={}){let n=e.current,{from:r,tabbable:i,wrap:a,accept:u}=t;var l;let o=r||Zu(hi((l=n[0])!==null&&l!==void 0?l:void 0)),s=n[0].previousElementSibling,c=Ib(n),d=Pl(c,{tabbable:i,accept:u},n);d.currentNode=Jd(o,n)?o:s;let h=d.nextNode();return!h&&a&&(d.currentNode=s,h=d.nextNode()),h&&af(h,!0),h},focusPrevious(t={}){let n=e.current,{from:r,tabbable:i,wrap:a,accept:u}=t;var l;let o=r||Zu(hi((l=n[0])!==null&&l!==void 0?l:void 0)),s=n[n.length-1].nextElementSibling,c=Ib(n),d=Pl(c,{tabbable:i,accept:u},n);d.currentNode=Jd(o,n)?o:s;let h=d.previousNode();return!h&&a&&(d.currentNode=s,h=d.previousNode()),h&&af(h,!0),h},focusFirst(t={}){let n=e.current,{tabbable:r,accept:i}=t,a=Ib(n),u=Pl(a,{tabbable:r,accept:i},n);u.currentNode=n[0].previousElementSibling;let l=u.nextNode();return l&&af(l,!0),l},focusLast(t={}){let n=e.current,{tabbable:r,accept:i}=t,a=Ib(n),u=Pl(a,{tabbable:r,accept:i},n);u.currentNode=n[n.length-1].nextElementSibling;let l=u.previousNode();return l&&af(l,!0),l}}}function Ib(e){return e[0].parentElement}function N2(e){let t=vl.getTreeNode(Sa);for(;t&&t.scopeRef!==e;){if(t.contain)return!1;t=t.parent}return!0}function SDe(e){if(e.checked)return!0;let t=[];if(!e.form)t=[...hi(e).querySelectorAll(`input[type="radio"][name="${CSS.escape(e.name)}"]`)].filter(a=>!a.form);else{var n,r;let a=(r=e.form)===null||r===void 0||(n=r.elements)===null||n===void 0?void 0:n.namedItem(e.name);t=[...a??[]]}return t?!t.some(a=>a.checked):!1}function _De(e,t){let n=P.useRef(void 0),r=P.useRef(void 0);Hr(()=>{let i=e.current;if(!t){r.current&&(cancelAnimationFrame(r.current),r.current=void 0);return}const a=hi(i?i[0]:void 0);let u=s=>{if(s.key!=="Tab"||s.altKey||s.ctrlKey||s.metaKey||!N2(e)||s.isComposing)return;let c=Zu(a),d=e.current;if(!d||!Jd(c,d))return;let h=Ib(d),m=Pl(h,{tabbable:!0},d);if(!c)return;m.currentNode=c;let y=s.shiftKey?m.previousNode():m.nextNode();y||(m.currentNode=s.shiftKey?d[d.length-1].nextElementSibling:d[0].previousElementSibling,y=s.shiftKey?m.previousNode():m.nextNode()),s.preventDefault(),y&&af(y,!0)},l=s=>{(!Sa||u$(Sa,e))&&Jd(ju(s),e.current)?(Sa=e,n.current=ju(s)):N2(e)&&!Iv(ju(s),e)?n.current?n.current.focus():Sa&&Sa.current&&z9(Sa.current):N2(e)&&(n.current=ju(s))},o=s=>{r.current&&cancelAnimationFrame(r.current),r.current=requestAnimationFrame(()=>{let c=ig(),d=(c==="virtual"||c===null)&&fS()&&dN(),h=Zu(a);if(!d&&h&&N2(e)&&!Iv(h,e)){Sa=e;let y=ju(s);if(y&&y.isConnected){var m;n.current=y,(m=n.current)===null||m===void 0||m.focus()}else Sa.current&&z9(Sa.current)}})};return a.addEventListener("keydown",u,!1),a.addEventListener("focusin",l,!1),i==null||i.forEach(s=>s.addEventListener("focusin",l,!1)),i==null||i.forEach(s=>s.addEventListener("focusout",o,!1)),()=>{a.removeEventListener("keydown",u,!1),a.removeEventListener("focusin",l,!1),i==null||i.forEach(s=>s.removeEventListener("focusin",l,!1)),i==null||i.forEach(s=>s.removeEventListener("focusout",o,!1))}},[e,t]),Hr(()=>()=>{r.current&&cancelAnimationFrame(r.current)},[r])}function Zne(e){return Iv(e)}function Jd(e,t){return!e||!t?!1:t.some(n=>n.contains(e))}function Iv(e,t=null){if(e instanceof Element&&e.closest("[data-react-aria-top-layer]"))return!0;for(let{scopeRef:n}of vl.traverse(vl.getTreeNode(t)))if(n&&Jd(e,n.current))return!0;return!1}function EDe(e){return Iv(e,Sa)}function u$(e,t){var n;let r=(n=vl.getTreeNode(t))===null||n===void 0?void 0:n.parent;for(;r;){if(r.scopeRef===e)return!0;r=r.parent}return!1}function af(e,t=!1){if(e!=null&&!t)try{lc(e)}catch{}else if(e!=null)try{e.focus()}catch{}}function Hne(e,t=!0){let n=e[0].previousElementSibling,r=Ib(e),i=Pl(r,{tabbable:t},e);i.currentNode=n;let a=i.nextNode();return t&&!a&&(r=Ib(e),i=Pl(r,{tabbable:!1},e),i.currentNode=n,a=i.nextNode()),a}function z9(e,t=!0){af(Hne(e,t))}function DDe(e,t){const n=be.useRef(t);P.useEffect(()=>{if(n.current){Sa=e;const r=hi(e.current?e.current[0]:void 0);!Jd(Zu(r),Sa.current)&&e.current&&z9(e.current)}n.current=!1},[e])}function $De(e,t,n){Hr(()=>{if(t||n)return;let r=e.current;const i=hi(r?r[0]:void 0);let a=u=>{let l=ju(u);Jd(l,e.current)?Sa=e:Zne(l)||(Sa=null)};return i.addEventListener("focusin",a,!1),r==null||r.forEach(u=>u.addEventListener("focusin",a,!1)),()=>{i.removeEventListener("focusin",a,!1),r==null||r.forEach(u=>u.removeEventListener("focusin",a,!1))}},[e,t,n])}function pj(e){let t=vl.getTreeNode(Sa);for(;t&&t.scopeRef!==e;){if(t.nodeToRestore)return!1;t=t.parent}return(t==null?void 0:t.scopeRef)===e}function CDe(e,t,n){const r=P.useRef(typeof document<"u"?Zu(hi(e.current?e.current[0]:void 0)):null);Hr(()=>{let i=e.current;const a=hi(i?i[0]:void 0);if(!t||n)return;let u=()=>{(!Sa||u$(Sa,e))&&Jd(Zu(a),e.current)&&(Sa=e)};return a.addEventListener("focusin",u,!1),i==null||i.forEach(l=>l.addEventListener("focusin",u,!1)),()=>{a.removeEventListener("focusin",u,!1),i==null||i.forEach(l=>l.removeEventListener("focusin",u,!1))}},[e,n]),Hr(()=>{const i=hi(e.current?e.current[0]:void 0);if(!t)return;let a=u=>{if(u.key!=="Tab"||u.altKey||u.ctrlKey||u.metaKey||!N2(e)||u.isComposing)return;let l=i.activeElement;if(!Iv(l,e)||!pj(e))return;let o=vl.getTreeNode(e);if(!o)return;let s=o.nodeToRestore,c=Pl(i.body,{tabbable:!0});c.currentNode=l;let d=u.shiftKey?c.previousNode():c.nextNode();if((!s||!s.isConnected||s===i.body)&&(s=void 0,o.nodeToRestore=void 0),(!d||!Iv(d,e))&&s){c.currentNode=s;do d=u.shiftKey?c.previousNode():c.nextNode();while(Iv(d,e));u.preventDefault(),u.stopPropagation(),d?af(d,!0):Zne(s)?af(s,!0):l.blur()}};return n||i.addEventListener("keydown",a,!0),()=>{n||i.removeEventListener("keydown",a,!0)}},[e,t,n]),Hr(()=>{const i=hi(e.current?e.current[0]:void 0);if(!t)return;let a=vl.getTreeNode(e);if(a){var u;return a.nodeToRestore=(u=r.current)!==null&&u!==void 0?u:void 0,()=>{let l=vl.getTreeNode(e);if(!l)return;let o=l.nodeToRestore,s=Zu(i);if(t&&o&&(s&&Iv(s,e)||s===i.body&&pj(e))){let c=vl.clone();requestAnimationFrame(()=>{if(i.activeElement===i.body){let d=c.getTreeNode(e);for(;d;){if(d.nodeToRestore&&d.nodeToRestore.isConnected){mj(d.nodeToRestore);return}d=d.parent}for(d=c.getTreeNode(e);d;){if(d.scopeRef&&d.scopeRef.current&&vl.getTreeNode(d.scopeRef)){let h=Hne(d.scopeRef.current,!0);mj(h);return}d=d.parent}}})}}}},[e,t])}function mj(e){e.dispatchEvent(new CustomEvent(Q9,{bubbles:!0,cancelable:!0}))&&af(e)}function Pl(e,t,n){let r=t!=null&&t.tabbable?D_e:fte,i=(e==null?void 0:e.nodeType)===Node.ELEMENT_NODE?e:null,a=hi(i),u=YSe(a,e||a,NodeFilter.SHOW_ELEMENT,{acceptNode(l){var o;return!(t==null||(o=t.from)===null||o===void 0)&&o.contains(l)||t!=null&&t.tabbable&&l.tagName==="INPUT"&&l.getAttribute("type")==="radio"&&(!SDe(l)||u.currentNode.tagName==="INPUT"&&u.currentNode.type==="radio"&&u.currentNode.name===l.name)?NodeFilter.FILTER_REJECT:r(l)&&(!n||Jd(l,n))&&(!(t!=null&&t.accept)||t.accept(l))?NodeFilter.FILTER_ACCEPT:NodeFilter.FILTER_SKIP}});return t!=null&&t.from&&(u.currentNode=t.from),u}function MN(e,t={}){return{focusNext(n={}){let r=e.current;if(!r)return null;let{from:i,tabbable:a=t.tabbable,wrap:u=t.wrap,accept:l=t.accept}=n,o=i||Zu(hi(r)),s=Pl(r,{tabbable:a,accept:l});r.contains(o)&&(s.currentNode=o);let c=s.nextNode();return!c&&u&&(s.currentNode=r,c=s.nextNode()),c&&af(c,!0),c},focusPrevious(n=t){let r=e.current;if(!r)return null;let{from:i,tabbable:a=t.tabbable,wrap:u=t.wrap,accept:l=t.accept}=n,o=i||Zu(hi(r)),s=Pl(r,{tabbable:a,accept:l});if(r.contains(o))s.currentNode=o;else{let d=ak(s);return d&&af(d,!0),d??null}let c=s.previousNode();if(!c&&u){s.currentNode=r;let d=ak(s);if(!d)return null;c=d}return c&&af(c,!0),c??null},focusFirst(n=t){let r=e.current;if(!r)return null;let{tabbable:i=t.tabbable,accept:a=t.accept}=n,l=Pl(r,{tabbable:i,accept:a}).nextNode();return l&&af(l,!0),l},focusLast(n=t){let r=e.current;if(!r)return null;let{tabbable:i=t.tabbable,accept:a=t.accept}=n,u=Pl(r,{tabbable:i,accept:a}),l=ak(u);return l&&af(l,!0),l??null}}}function ak(e){let t,n;do n=e.lastChild(),n&&(t=n);while(n);return t}class FN{get size(){return this.fastMap.size}getTreeNode(t){return this.fastMap.get(t)}addTreeNode(t,n,r){let i=this.fastMap.get(n??null);if(!i)return;let a=new q9({scopeRef:t});i.addChild(a),a.parent=i,this.fastMap.set(t,a),r&&(a.nodeToRestore=r)}addNode(t){this.fastMap.set(t.scopeRef,t)}removeTreeNode(t){if(t===null)return;let n=this.fastMap.get(t);if(!n)return;let r=n.parent;for(let a of this.traverse())a!==n&&n.nodeToRestore&&a.nodeToRestore&&n.scopeRef&&n.scopeRef.current&&Jd(a.nodeToRestore,n.scopeRef.current)&&(a.nodeToRestore=n.nodeToRestore);let i=n.children;r&&(r.removeChild(n),i.size>0&&i.forEach(a=>r&&r.addChild(a))),this.fastMap.delete(n.scopeRef)}*traverse(t=this.root){if(t.scopeRef!=null&&(yield t),t.children.size>0)for(let n of t.children)yield*this.traverse(n)}clone(){var t;let n=new FN;var r;for(let i of this.traverse())n.addTreeNode(i.scopeRef,(r=(t=i.parent)===null||t===void 0?void 0:t.scopeRef)!==null&&r!==void 0?r:null,i.nodeToRestore);return n}constructor(){this.fastMap=new Map,this.root=new q9({scopeRef:null}),this.fastMap.set(null,this.root)}}class q9{addChild(t){this.children.add(t),t.parent=this}removeChild(t){this.children.delete(t),t.parent=void 0}constructor(t){this.children=new Set,this.contain=!1,this.scopeRef=t.scopeRef}}let vl=new FN;function Gu(e={}){let{autoFocus:t=!1,isTextInput:n,within:r}=e,i=P.useRef({isFocused:!1,isFocusVisible:t||ey()}),[a,u]=P.useState(!1),[l,o]=P.useState(()=>i.current.isFocused&&i.current.isFocusVisible),s=P.useCallback(()=>o(i.current.isFocused&&i.current.isFocusVisible),[]),c=P.useCallback(m=>{i.current.isFocused=m,u(m),s()},[s]);hDe(m=>{i.current.isFocusVisible=m,s()},[],{isTextInput:n});let{focusProps:d}=XC({isDisabled:r,onFocusChange:c}),{focusWithinProps:h}=L0({isDisabled:!r,onFocusWithinChange:c});return{isFocused:a,isFocusVisible:l,focusProps:r?h:d}}function Xne(e,t){let n=t==null?void 0:t.isDisabled,[r,i]=P.useState(!1);return Hr(()=>{if(e!=null&&e.current&&!n){let a=()=>{if(e.current){let l=Pl(e.current,{tabbable:!0});i(!!l.nextNode())}};a();let u=new MutationObserver(a);return u.observe(e.current,{subtree:!0,childList:!0,attributes:!0,attributeFilter:["tabIndex","disabled"]}),()=>{u.disconnect()}}}),n?!1:r}function NN(e){let t=Yne(hi(e));t!==e&&(t&&V9(t,e),e&&Sx(e,t))}function V9(e,t){e.dispatchEvent(new FocusEvent("blur",{relatedTarget:t})),e.dispatchEvent(new FocusEvent("focusout",{bubbles:!0,relatedTarget:t}))}function Sx(e,t){e.dispatchEvent(new FocusEvent("focus",{relatedTarget:t})),e.dispatchEvent(new FocusEvent("focusin",{bubbles:!0,relatedTarget:t}))}function Yne(e){let t=Zu(e),n=t==null?void 0:t.getAttribute("aria-activedescendant");return n&&e.getElementById(n)||t}const a0=[];function Gne(e,t){let{onClose:n,shouldCloseOnBlur:r,isOpen:i,isDismissable:a=!1,isKeyboardDismissDisabled:u=!1,shouldCloseOnInteractOutside:l}=e;P.useEffect(()=>{if(i&&!a0.includes(t))return a0.push(t),()=>{let y=a0.indexOf(t);y>=0&&a0.splice(y,1)}},[i,t]);let o=()=>{a0[a0.length-1]===t&&n&&n()},s=y=>{(!l||l(y.target))&&a0[a0.length-1]===t&&(y.stopPropagation(),y.preventDefault())},c=y=>{(!l||l(y.target))&&(a0[a0.length-1]===t&&(y.stopPropagation(),y.preventDefault()),o())},d=y=>{y.key==="Escape"&&!u&&!y.nativeEvent.isComposing&&(y.stopPropagation(),y.preventDefault(),o())};bDe({ref:t,onInteractOutside:a&&i?c:void 0,onInteractOutsideStart:s});let{focusWithinProps:h}=L0({isDisabled:!r,onBlurWithin:y=>{!y.relatedTarget||EDe(y.relatedTarget)||(!l||l(y.relatedTarget))&&(n==null||n())}}),m=y=>{y.target===y.currentTarget&&y.preventDefault()};return{overlayProps:{onKeyDown:d,...h},underlayProps:{onPointerDown:m}}}function Kne(e,t,n){let{type:r}=e,{isOpen:i}=t;P.useEffect(()=>{n&&n.current&&Gte.set(n.current,t.close)});let a;r==="menu"?a=!0:r==="listbox"&&(a="listbox");let u=vi();return{triggerProps:{"aria-haspopup":a,"aria-expanded":i,"aria-controls":i?u:void 0,onPress:t.toggle},overlayProps:{id:u}}}const X2=typeof document<"u"&&window.visualViewport;let G_=0,uk;function Wne(e={}){let{isDisabled:t}=e;Hr(()=>{if(!t)return G_++,G_===1&&($1()?uk=RDe():uk=wDe()),()=>{G_--,G_===0&&uk()}},[t])}function wDe(){let e=window.innerWidth-document.documentElement.clientWidth;return ih(e>0&&("scrollbarGutter"in document.documentElement.style?lk(document.documentElement,"scrollbarGutter","stable"):lk(document.documentElement,"paddingRight",`${e}px`)),lk(document.documentElement,"overflow","hidden"))}function RDe(){let e,t=!1,n=o=>{let s=o.target;e=l1(s)?s:m0(s,!0),t=!1;let c=s.ownerDocument.defaultView.getSelection();c&&!c.isCollapsed&&c.containsNode(s,!0)&&(t=!0),"selectionStart"in s&&"selectionEnd"in s&&s.selectionStart<s.selectionEnd&&s.ownerDocument.activeElement===s&&(t=!0)},r=document.createElement("style");r.textContent=`
@layer {
  * {
    overscroll-behavior: contain;
  }
}`.trim(),document.head.prepend(r);let i=o=>{if(!(o.touches.length===2||t)){if(!e||e===document.documentElement||e===document.body){o.preventDefault();return}e.scrollHeight===e.clientHeight&&e.scrollWidth===e.clientWidth&&o.preventDefault()}},a=o=>{let s=o.target,c=o.relatedTarget;if(c&&j2(c))c.focus({preventScroll:!0}),gj(c,j2(s));else if(!c){var d;let h=(d=s.parentElement)===null||d===void 0?void 0:d.closest("[tabindex]");h==null||h.focus({preventScroll:!0})}},u=HTMLElement.prototype.focus;HTMLElement.prototype.focus=function(o){let s=document.activeElement!=null&&j2(document.activeElement);u.call(this,{...o,preventScroll:!0}),(!o||!o.preventScroll)&&gj(this,s)};let l=ih(ok(document,"touchstart",n,{passive:!1,capture:!0}),ok(document,"touchmove",i,{passive:!1,capture:!0}),ok(document,"blur",a,!0));return()=>{l(),r.remove(),HTMLElement.prototype.focus=u}}function lk(e,t,n){let r=e.style[t];return e.style[t]=n,()=>{e.style[t]=r}}function ok(e,t,n,r){return e.addEventListener(t,n,r),()=>{e.removeEventListener(t,n,r)}}function gj(e,t){t||!X2?vj(e):X2.addEventListener("resize",()=>vj(e),{once:!0})}function vj(e){let t=document.scrollingElement||document.documentElement,n=e;for(;n&&n!==t;){let r=m0(n);if(r!==document.documentElement&&r!==document.body&&r!==n){let i=r.getBoundingClientRect(),a=n.getBoundingClientRect();if(a.top<i.top||a.bottom>i.top+n.clientHeight){let u=i.bottom;X2&&(u=Math.min(u,X2.offsetTop+X2.height));let l=a.top-i.top-((u-i.top)/2-a.height/2);r.scrollTo({top:Math.max(0,Math.min(r.scrollHeight-r.clientHeight,r.scrollTop+l)),behavior:"smooth"})}}n=r.parentElement}}const ADe=P.createContext({});function IN(){var e;return(e=P.useContext(ADe))!==null&&e!==void 0?e:{}}const U9=be.createContext(null);function TDe(e){let{children:t}=e,n=P.useContext(U9),[r,i]=P.useState(0),a=P.useMemo(()=>({parent:n,modalCount:r,addModal(){i(u=>u+1),n&&n.addModal()},removeModal(){i(u=>u-1),n&&n.removeModal()}}),[n,r]);return be.createElement(U9.Provider,{value:a},t)}function kDe(){let e=P.useContext(U9);return{modalProviderProps:{"aria-hidden":e&&e.modalCount>0?!0:void 0}}}function PDe(e){let{modalProviderProps:t}=kDe();return be.createElement("div",{"data-overlay-container":!0,...e,...t})}function MDe(e){return be.createElement(TDe,null,be.createElement(PDe,e))}function FDe(e){let t=pp(),{portalContainer:n=t?null:document.body,...r}=e,{getContainer:i}=IN();if(!e.portalContainer&&i&&(n=i()),be.useEffect(()=>{if(n!=null&&n.closest("[data-overlay-container]"))throw new Error("An OverlayContainer must not be inside another container. Please change the portalContainer prop.")},[n]),!n)return null;let a=be.createElement(MDe,r);return ste.createPortal(a,n)}var Jne={};Jne={dismiss:"تجاهل"};var ere={};ere={dismiss:"Отхвърляне"};var tre={};tre={dismiss:"Odstranit"};var nre={};nre={dismiss:"Luk"};var rre={};rre={dismiss:"Schließen"};var ire={};ire={dismiss:"Απόρριψη"};var are={};are={dismiss:"Dismiss"};var ure={};ure={dismiss:"Descartar"};var lre={};lre={dismiss:"Lõpeta"};var ore={};ore={dismiss:"Hylkää"};var sre={};sre={dismiss:"Rejeter"};var cre={};cre={dismiss:"התעלם"};var fre={};fre={dismiss:"Odbaci"};var dre={};dre={dismiss:"Elutasítás"};var hre={};hre={dismiss:"Ignora"};var pre={};pre={dismiss:"閉じる"};var mre={};mre={dismiss:"무시"};var gre={};gre={dismiss:"Atmesti"};var vre={};vre={dismiss:"Nerādīt"};var yre={};yre={dismiss:"Lukk"};var bre={};bre={dismiss:"Negeren"};var Ore={};Ore={dismiss:"Zignoruj"};var xre={};xre={dismiss:"Descartar"};var Sre={};Sre={dismiss:"Dispensar"};var _re={};_re={dismiss:"Revocare"};var Ere={};Ere={dismiss:"Пропустить"};var Dre={};Dre={dismiss:"Zrušiť"};var $re={};$re={dismiss:"Opusti"};var Cre={};Cre={dismiss:"Odbaci"};var wre={};wre={dismiss:"Avvisa"};var Rre={};Rre={dismiss:"Kapat"};var Are={};Are={dismiss:"Скасувати"};var Tre={};Tre={dismiss:"取消"};var kre={};kre={dismiss:"關閉"};var Pre={};Pre={"ar-AE":Jne,"bg-BG":ere,"cs-CZ":tre,"da-DK":nre,"de-DE":rre,"el-GR":ire,"en-US":are,"es-ES":ure,"et-EE":lre,"fi-FI":ore,"fr-FR":sre,"he-IL":cre,"hr-HR":fre,"hu-HU":dre,"it-IT":hre,"ja-JP":pre,"ko-KR":mre,"lt-LT":gre,"lv-LV":vre,"nb-NO":yre,"nl-NL":bre,"pl-PL":Ore,"pt-BR":xre,"pt-PT":Sre,"ro-RO":_re,"ru-RU":Ere,"sk-SK":Dre,"sl-SI":$re,"sr-SP":Cre,"sv-SE":wre,"tr-TR":Rre,"uk-UA":Are,"zh-CN":Tre,"zh-TW":kre};const yj={border:0,clip:"rect(0 0 0 0)",clipPath:"inset(50%)",height:"1px",margin:"-1px",overflow:"hidden",padding:0,position:"absolute",width:"1px",whiteSpace:"nowrap"};function LN(e={}){let{style:t,isFocusable:n}=e,[r,i]=P.useState(!1),{focusWithinProps:a}=L0({isDisabled:!n,onFocusWithinChange:l=>i(l)}),u=P.useMemo(()=>r?t:t?{...yj,...t}:yj,[r]);return{visuallyHiddenProps:{...a,style:u}}}function mS(e){let{children:t,elementType:n="div",isFocusable:r,style:i,...a}=e,{visuallyHiddenProps:u}=LN(e);return be.createElement(n,Et(a,u),t)}function NDe(e){return e&&e.__esModule?e.default:e}function j9(e){let{onDismiss:t,...n}=e,r=Qo(NDe(Pre),"@react-aria/overlays"),i=Gv(n,r.format("dismiss")),a=()=>{t&&t()};return be.createElement(mS,null,be.createElement("button",{...i,tabIndex:-1,onClick:a,style:{width:1,height:1}}))}const IDe=typeof HTMLElement<"u"&&"inert"in HTMLElement.prototype;let W4=new WeakMap,Kc=[];function BN(e,t){let n=uf(e==null?void 0:e[0]),r=t instanceof n.Element?{root:t}:t;var i;let a=(i=r==null?void 0:r.root)!==null&&i!==void 0?i:document.body,u=(r==null?void 0:r.shouldUseInert)&&IDe,l=new Set(e),o=new Set,s=v=>u&&v instanceof n.HTMLElement?v.inert:v.getAttribute("aria-hidden")==="true",c=(v,S)=>{u&&v instanceof n.HTMLElement?v.inert=S:S?v.setAttribute("aria-hidden","true"):(v.removeAttribute("aria-hidden"),v instanceof n.HTMLElement&&(v.inert=!1))},d=v=>{for(let $ of v.querySelectorAll("[data-live-announcer], [data-react-aria-top-layer]"))l.add($);let S=$=>{if(o.has($)||l.has($)||$.parentElement&&o.has($.parentElement)&&$.parentElement.getAttribute("role")!=="row")return NodeFilter.FILTER_REJECT;for(let k of l)if($.contains(k))return NodeFilter.FILTER_SKIP;return NodeFilter.FILTER_ACCEPT},_=document.createTreeWalker(v,NodeFilter.SHOW_ELEMENT,{acceptNode:S}),E=S(v);if(E===NodeFilter.FILTER_ACCEPT&&h(v),E!==NodeFilter.FILTER_REJECT){let $=_.nextNode();for(;$!=null;)h($),$=_.nextNode()}},h=v=>{var S;let _=(S=W4.get(v))!==null&&S!==void 0?S:0;s(v)&&_===0||(_===0&&c(v,!0),o.add(v),W4.set(v,_+1))};Kc.length&&Kc[Kc.length-1].disconnect(),d(a);let m=new MutationObserver(v=>{for(let S of v)if(S.type==="childList"&&![...l,...o].some(_=>_.contains(S.target)))for(let _ of S.addedNodes)(_ instanceof HTMLElement||_ instanceof SVGElement)&&(_.dataset.liveAnnouncer==="true"||_.dataset.reactAriaTopLayer==="true")?l.add(_):_ instanceof Element&&d(_)});m.observe(a,{childList:!0,subtree:!0});let y={visibleNodes:l,hiddenNodes:o,observe(){m.observe(a,{childList:!0,subtree:!0})},disconnect(){m.disconnect()}};return Kc.push(y),()=>{m.disconnect();for(let v of o){let S=W4.get(v);S!=null&&(S===1?(c(v,!1),W4.delete(v)):W4.set(v,S-1))}y===Kc[Kc.length-1]?(Kc.pop(),Kc.length&&Kc[Kc.length-1].observe()):Kc.splice(Kc.indexOf(y),1)}}function LDe(e){let t=Kc[Kc.length-1];if(t&&!t.visibleNodes.has(e))return t.visibleNodes.add(e),()=>{t.visibleNodes.delete(e)}}function BDe(e,t){let{triggerRef:n,popoverRef:r,groupRef:i,isNonModal:a,isKeyboardDismissDisabled:u,shouldCloseOnInteractOutside:l,...o}=e,s=o.trigger==="SubmenuTrigger",{overlayProps:c,underlayProps:d}=Gne({isOpen:t.isOpen,onClose:t.close,shouldCloseOnBlur:!0,isDismissable:!a||s,isKeyboardDismissDisabled:u,shouldCloseOnInteractOutside:l},i??r),{overlayProps:h,arrowProps:m,placement:y,triggerAnchorPoint:v}=kne({...o,targetRef:n,overlayRef:r,isOpen:t.isOpen,onClose:a&&!s?t.close:null});return Wne({isDisabled:a||!t.isOpen}),P.useEffect(()=>{if(t.isOpen&&r.current){var S,_;return a?LDe((S=i==null?void 0:i.current)!==null&&S!==void 0?S:r.current):BN([(_=i==null?void 0:i.current)!==null&&_!==void 0?_:r.current],{shouldUseInert:!0})}},[a,t.isOpen,r,i]),{popoverProps:Et(c,h),arrowProps:m,underlayProps:d,placement:y,triggerAnchorPoint:v}}const Mre=be.createContext(null);function Z9(e){let t=pp(),{portalContainer:n=t?null:document.body,isExiting:r}=e,[i,a]=P.useState(!1),u=P.useMemo(()=>({contain:i,setContain:a}),[i,a]),{getContainer:l}=IN();if(!e.portalContainer&&l&&(n=l()),!n)return null;let o=e.children;return e.disableFocusManagement||(o=be.createElement(PN,{restoreFocus:!0,contain:(e.shouldContainFocus||i)&&!r},o)),o=be.createElement(Mre.Provider,{value:u},be.createElement(gDe,null,o)),ste.createPortal(o,n)}function Fre(){let e=P.useContext(Mre),t=e==null?void 0:e.setContain;Hr(()=>{t==null||t(!0)},[t])}function QDe(e,t,n){let{overlayProps:r,underlayProps:i}=Gne({...e,isOpen:t.isOpen,onClose:t.close},n);return Wne({isDisabled:!t.isOpen}),Fre(),P.useEffect(()=>{if(t.isOpen&&n.current)return BN([n.current],{shouldUseInert:!0})},[t.isOpen,n]),{modalProps:Et(r),underlayProps:i}}const YC=new WeakMap;function zDe(e){return typeof e=="string"?e.replace(/\s*/g,""):""+e}function Nre(e,t){let n=YC.get(e);if(!n)throw new Error("Unknown list");return`${n.id}-option-${zDe(t)}`}function gS(e){let{id:t,label:n,"aria-labelledby":r,"aria-label":i,labelElementType:a="label"}=e;t=vi(t);let u=vi(),l={};n&&(r=r?`${u} ${r}`:u,l={id:u,htmlFor:a==="label"?t:void 0});let o=Gv({id:t,"aria-label":i,"aria-labelledby":r});return{labelProps:l,fieldProps:o}}function vS(e){let{description:t,errorMessage:n,isInvalid:r,validationState:i}=e,{labelProps:a,fieldProps:u}=gS(e),l=$0([!!t,!!n,r,i]),o=$0([!!t,!!n,r,i]);return u=Et(u,{"aria-describedby":[l,o,e["aria-describedby"]].filter(Boolean).join(" ")||void 0}),{labelProps:a,fieldProps:u,descriptionProps:{id:l},errorMessageProps:{id:o}}}function H9(e){return JE()?e.altKey:e.ctrlKey}function iD(e,t){var n,r;let i=`[data-key="${CSS.escape(String(t))}"]`,a=(n=e.current)===null||n===void 0?void 0:n.dataset.collection;return a&&(i=`[data-collection="${CSS.escape(a)}"]${i}`),(r=e.current)===null||r===void 0?void 0:r.querySelector(i)}const Ire=new WeakMap;function qDe(e){let t=vi();return Ire.set(e,t),t}function VDe(e){return Ire.get(e)}const UDe=1e3;function Lre(e){let{keyboardDelegate:t,selectionManager:n,onTypeSelect:r}=e,i=P.useRef({search:"",timeout:void 0}).current,a=u=>{let l=jDe(u.key);if(!(!l||u.ctrlKey||u.metaKey||!u.currentTarget.contains(u.target)||i.search.length===0&&l===" ")){if(l===" "&&i.search.trim().length>0&&(u.preventDefault(),"continuePropagation"in u||u.stopPropagation()),i.search+=l,t.getKeyForSearch!=null){let o=t.getKeyForSearch(i.search,n.focusedKey);o==null&&(o=t.getKeyForSearch(i.search)),o!=null&&(n.setFocusedKey(o),r&&r(o))}clearTimeout(i.timeout),i.timeout=setTimeout(()=>{i.search=""},UDe)}};return{typeSelectProps:{onKeyDownCapture:t.getKeyForSearch?a:void 0}}}function jDe(e){return e.length===1||!/^[A-Z]/i.test(e)?e:""}function QN(e){let{selectionManager:t,keyboardDelegate:n,ref:r,autoFocus:i=!1,shouldFocusWrap:a=!1,disallowEmptySelection:u=!1,disallowSelectAll:l=!1,escapeKeyBehavior:o="clearSelection",selectOnFocus:s=t.selectionBehavior==="replace",disallowTypeAhead:c=!1,shouldUseVirtualFocus:d,allowsTabNavigation:h=!1,isVirtualized:m,scrollRef:y=r,linkBehavior:v="action"}=e,{direction:S}=Wi(),_=X3(),E=J=>{var te;if(J.altKey&&J.key==="Tab"&&J.preventDefault(),!(!((te=r.current)===null||te===void 0)&&te.contains(J.target)))return;const H=(se,pe)=>{if(se!=null){if(t.isLink(se)&&v==="selection"&&s&&!H9(J)){Uf.flushSync(()=>{t.setFocusedKey(se,pe)});let ve=iD(r,se),we=t.getItemProps(se);ve&&_.open(ve,J,we.href,we.routerOptions);return}if(t.setFocusedKey(se,pe),t.isLink(se)&&v==="override")return;J.shiftKey&&t.selectionMode==="multiple"?t.extendSelection(se):s&&!H9(J)&&t.replaceSelection(se)}};switch(J.key){case"ArrowDown":if(n.getKeyBelow){var M,G,X;let se=t.focusedKey!=null?(M=n.getKeyBelow)===null||M===void 0?void 0:M.call(n,t.focusedKey):(G=n.getFirstKey)===null||G===void 0?void 0:G.call(n);se==null&&a&&(se=(X=n.getFirstKey)===null||X===void 0?void 0:X.call(n,t.focusedKey)),se!=null&&(J.preventDefault(),H(se))}break;case"ArrowUp":if(n.getKeyAbove){var B,ne,oe;let se=t.focusedKey!=null?(B=n.getKeyAbove)===null||B===void 0?void 0:B.call(n,t.focusedKey):(ne=n.getLastKey)===null||ne===void 0?void 0:ne.call(n);se==null&&a&&(se=(oe=n.getLastKey)===null||oe===void 0?void 0:oe.call(n,t.focusedKey)),se!=null&&(J.preventDefault(),H(se))}break;case"ArrowLeft":if(n.getKeyLeftOf){var ie,de,ge;let se=t.focusedKey!=null?(ie=n.getKeyLeftOf)===null||ie===void 0?void 0:ie.call(n,t.focusedKey):null;se==null&&a&&(se=S==="rtl"?(de=n.getFirstKey)===null||de===void 0?void 0:de.call(n,t.focusedKey):(ge=n.getLastKey)===null||ge===void 0?void 0:ge.call(n,t.focusedKey)),se!=null&&(J.preventDefault(),H(se,S==="rtl"?"first":"last"))}break;case"ArrowRight":if(n.getKeyRightOf){var ye,me,Ee;let se=t.focusedKey!=null?(ye=n.getKeyRightOf)===null||ye===void 0?void 0:ye.call(n,t.focusedKey):null;se==null&&a&&(se=S==="rtl"?(me=n.getLastKey)===null||me===void 0?void 0:me.call(n,t.focusedKey):(Ee=n.getFirstKey)===null||Ee===void 0?void 0:Ee.call(n,t.focusedKey)),se!=null&&(J.preventDefault(),H(se,S==="rtl"?"last":"first"))}break;case"Home":if(n.getFirstKey){if(t.focusedKey===null&&J.shiftKey)return;J.preventDefault();let se=n.getFirstKey(t.focusedKey,Tb(J));t.setFocusedKey(se),se!=null&&(Tb(J)&&J.shiftKey&&t.selectionMode==="multiple"?t.extendSelection(se):s&&t.replaceSelection(se))}break;case"End":if(n.getLastKey){if(t.focusedKey===null&&J.shiftKey)return;J.preventDefault();let se=n.getLastKey(t.focusedKey,Tb(J));t.setFocusedKey(se),se!=null&&(Tb(J)&&J.shiftKey&&t.selectionMode==="multiple"?t.extendSelection(se):s&&t.replaceSelection(se))}break;case"PageDown":if(n.getKeyPageBelow&&t.focusedKey!=null){let se=n.getKeyPageBelow(t.focusedKey);se!=null&&(J.preventDefault(),H(se))}break;case"PageUp":if(n.getKeyPageAbove&&t.focusedKey!=null){let se=n.getKeyPageAbove(t.focusedKey);se!=null&&(J.preventDefault(),H(se))}break;case"a":Tb(J)&&t.selectionMode==="multiple"&&l!==!0&&(J.preventDefault(),t.selectAll());break;case"Escape":o==="clearSelection"&&!u&&t.selectedKeys.size!==0&&(J.stopPropagation(),J.preventDefault(),t.clearSelection());break;case"Tab":if(!h){if(J.shiftKey)r.current.focus();else{let se=Pl(r.current,{tabbable:!0}),pe,ve;do ve=se.lastChild(),ve&&(pe=ve);while(ve);pe&&!pe.contains(document.activeElement)&&$u(pe)}break}}},$=P.useRef({top:0,left:0});Yd(y,"scroll",m?void 0:()=>{var J,te,H,M;$.current={top:(H=(J=y.current)===null||J===void 0?void 0:J.scrollTop)!==null&&H!==void 0?H:0,left:(M=(te=y.current)===null||te===void 0?void 0:te.scrollLeft)!==null&&M!==void 0?M:0}});let k=J=>{if(t.isFocused){J.currentTarget.contains(J.target)||t.setFocused(!1);return}if(J.currentTarget.contains(J.target)){if(t.setFocused(!0),t.focusedKey==null){var te,H;let X=ne=>{ne!=null&&(t.setFocusedKey(ne),s&&!t.isSelected(ne)&&t.replaceSelection(ne))},B=J.relatedTarget;var M,G;B&&J.currentTarget.compareDocumentPosition(B)&Node.DOCUMENT_POSITION_FOLLOWING?X((M=t.lastSelectedKey)!==null&&M!==void 0?M:(te=n.getLastKey)===null||te===void 0?void 0:te.call(n)):X((G=t.firstSelectedKey)!==null&&G!==void 0?G:(H=n.getFirstKey)===null||H===void 0?void 0:H.call(n))}else!m&&y.current&&(y.current.scrollTop=$.current.top,y.current.scrollLeft=$.current.left);if(t.focusedKey!=null&&y.current){let X=iD(r,t.focusedKey);X instanceof HTMLElement&&(!X.contains(document.activeElement)&&!d&&$u(X),ig()==="keyboard"&&Vm(X,{containingElement:r.current}))}}},A=J=>{J.currentTarget.contains(J.relatedTarget)||t.setFocused(!1)},w=P.useRef(!1);Yd(r,E9,d?J=>{let{detail:te}=J;J.stopPropagation(),t.setFocused(!0),(te==null?void 0:te.focusStrategy)==="first"&&(w.current=!0)}:void 0);let R=fi(()=>{var J,te;let H=(te=(J=n.getFirstKey)===null||J===void 0?void 0:J.call(n))!==null&&te!==void 0?te:null;if(H==null){let M=Zu();NN(r.current),Sx(M,null),t.collection.size>0&&(w.current=!1)}else t.setFocusedKey(H),w.current=!1});DU(()=>{w.current&&R()},[t.collection,R]);let F=fi(()=>{t.collection.size>0&&(w.current=!1)});DU(()=>{F()},[t.focusedKey,F]),Yd(r,ote,d?J=>{var te;J.stopPropagation(),t.setFocused(!1),!((te=J.detail)===null||te===void 0)&&te.clearFocusKey&&t.setFocusedKey(null)}:void 0);const I=P.useRef(i),U=P.useRef(!1);P.useEffect(()=>{if(I.current){var J,te;let G=null;var H;i==="first"&&(G=(H=(J=n.getFirstKey)===null||J===void 0?void 0:J.call(n))!==null&&H!==void 0?H:null);var M;i==="last"&&(G=(M=(te=n.getLastKey)===null||te===void 0?void 0:te.call(n))!==null&&M!==void 0?M:null);let X=t.selectedKeys;if(X.size){for(let B of X)if(t.canSelectItem(B)){G=B;break}}t.setFocused(!0),t.setFocusedKey(G),G==null&&!d&&r.current&&lc(r.current),t.collection.size>0&&(I.current=!1,U.current=!0)}});let q=P.useRef(t.focusedKey),V=P.useRef(null);P.useEffect(()=>{if(t.isFocused&&t.focusedKey!=null&&(t.focusedKey!==q.current||U.current)&&y.current&&r.current){let J=ig(),te=iD(r,t.focusedKey);if(!(te instanceof HTMLElement))return;(J==="keyboard"||U.current)&&(V.current&&cancelAnimationFrame(V.current),V.current=requestAnimationFrame(()=>{y.current&&(ute(y.current,te),J!=="virtual"&&Vm(te,{containingElement:r.current}))}))}!d&&t.isFocused&&t.focusedKey==null&&q.current!=null&&r.current&&lc(r.current),q.current=t.focusedKey,U.current=!1}),P.useEffect(()=>()=>{V.current&&cancelAnimationFrame(V.current)},[]),Yd(r,"react-aria-focus-scope-restore",J=>{J.preventDefault(),t.setFocused(!0)});let N={onKeyDown:E,onFocus:k,onBlur:A,onMouseDown(J){y.current===J.target&&J.preventDefault()}},{typeSelectProps:j}=Lre({keyboardDelegate:n,selectionManager:t});c||(N=Et(j,N));let W;d||(W=t.focusedKey==null?0:-1);let K=qDe(t.collection);return{collectionProps:Et(N,{tabIndex:W,"data-collection":K})}}function GC(e){let{id:t,selectionManager:n,key:r,ref:i,shouldSelectOnPressUp:a,shouldUseVirtualFocus:u,focus:l,isDisabled:o,onAction:s,allowsDifferentPressOrigin:c,linkBehavior:d="action"}=e,h=X3();t=vi(t);let m=M=>{if(M.pointerType==="keyboard"&&H9(M))n.toggleSelection(r);else{if(n.selectionMode==="none")return;if(n.isLink(r)){if(d==="selection"&&i.current){let G=n.getItemProps(r);h.open(i.current,M,G.href,G.routerOptions),n.setSelectedKeys(n.selectedKeys);return}else if(d==="override"||d==="none")return}n.selectionMode==="single"?n.isSelected(r)&&!n.disallowEmptySelection?n.toggleSelection(r):n.replaceSelection(r):M&&M.shiftKey?n.extendSelection(r):n.selectionBehavior==="toggle"||M&&(Tb(M)||M.pointerType==="touch"||M.pointerType==="virtual")?n.toggleSelection(r):n.replaceSelection(r)}};P.useEffect(()=>{r===n.focusedKey&&n.isFocused&&(u?NN(i.current):l?l():document.activeElement!==i.current&&i.current&&lc(i.current))},[i,r,n.focusedKey,n.childFocusStrategy,n.isFocused,u]),o=o||n.isDisabled(r);let y={};!u&&!o?y={tabIndex:r===n.focusedKey?0:-1,onFocus(M){M.target===i.current&&n.setFocusedKey(r)}}:o&&(y.onMouseDown=M=>{M.preventDefault()});let v=n.isLink(r)&&d==="override",S=s&&e.UNSTABLE_itemBehavior==="action",_=n.isLink(r)&&d!=="selection"&&d!=="none",E=!o&&n.canSelectItem(r)&&!v&&!S,$=(s||_)&&!o,k=$&&(n.selectionBehavior==="replace"?!E:!E||n.isEmpty),A=$&&E&&n.selectionBehavior==="replace",w=k||A,R=P.useRef(null),F=w&&E,I=P.useRef(!1),U=P.useRef(!1),q=n.getItemProps(r),V=M=>{if(s){var G;s(),(G=i.current)===null||G===void 0||G.dispatchEvent(new CustomEvent("react-aria-item-action",{bubbles:!0}))}_&&i.current&&h.open(i.current,M,q.href,q.routerOptions)},N={ref:i};if(a?(N.onPressStart=M=>{R.current=M.pointerType,I.current=F,M.pointerType==="keyboard"&&(!w||Oj())&&m(M)},c?(N.onPressUp=k?void 0:M=>{M.pointerType==="mouse"&&E&&m(M)},N.onPress=k?V:M=>{M.pointerType!=="keyboard"&&M.pointerType!=="mouse"&&E&&m(M)}):N.onPress=M=>{if(k||A&&M.pointerType!=="mouse"){if(M.pointerType==="keyboard"&&!bj())return;V(M)}else M.pointerType!=="keyboard"&&E&&m(M)}):(N.onPressStart=M=>{R.current=M.pointerType,I.current=F,U.current=k,E&&(M.pointerType==="mouse"&&!k||M.pointerType==="keyboard"&&(!$||Oj()))&&m(M)},N.onPress=M=>{(M.pointerType==="touch"||M.pointerType==="pen"||M.pointerType==="virtual"||M.pointerType==="keyboard"&&w&&bj()||M.pointerType==="mouse"&&U.current)&&(w?V(M):E&&m(M))}),y["data-collection"]=VDe(n.collection),y["data-key"]=r,N.preventFocusOnPress=u,u&&(N=Et(N,{onPressStart(M){M.pointerType!=="touch"&&(n.setFocused(!0),n.setFocusedKey(r))},onPress(M){M.pointerType==="touch"&&(n.setFocused(!0),n.setFocusedKey(r))}})),q)for(let M of["onPressStart","onPressEnd","onPressChange","onPress","onPressUp","onClick"])q[M]&&(N[M]=ih(N[M],q[M]));let{pressProps:j,isPressed:W}=lp(N),K=A?M=>{R.current==="mouse"&&(M.stopPropagation(),M.preventDefault(),V(M))}:void 0,{longPressProps:J}=jne({isDisabled:!F,onLongPress(M){M.pointerType==="touch"&&(m(M),n.setSelectionBehavior("toggle"))}}),te=M=>{R.current==="touch"&&I.current&&M.preventDefault()},H=d!=="none"&&n.isLink(r)?M=>{a1.isOpening||M.preventDefault()}:void 0;return{itemProps:Et(y,E||k||u&&!o?j:{},F?J:{},{onDoubleClick:K,onDragStartCapture:te,onClick:H,id:t},u?{onMouseDown:M=>M.preventDefault()}:void 0),isPressed:W,isSelected:n.isSelected(r),isFocused:n.isFocused&&n.focusedKey===r,isDisabled:o,allowsSelection:E,hasAction:w}}function bj(){let e=window.event;return(e==null?void 0:e.key)==="Enter"}function Oj(){let e=window.event;return(e==null?void 0:e.key)===" "||(e==null?void 0:e.code)==="Space"}class xj{getItemRect(t){let n=this.ref.current;if(!n)return null;let r=t!=null?iD(this.ref,t):null;if(!r)return null;let i=n.getBoundingClientRect(),a=r.getBoundingClientRect();return{x:a.left-i.left-n.clientLeft+n.scrollLeft,y:a.top-i.top-n.clientTop+n.scrollTop,width:a.width,height:a.height}}getContentSize(){let t=this.ref.current;var n,r;return{width:(n=t==null?void 0:t.scrollWidth)!==null&&n!==void 0?n:0,height:(r=t==null?void 0:t.scrollHeight)!==null&&r!==void 0?r:0}}getVisibleRect(){let t=this.ref.current;var n,r,i,a;return{x:(n=t==null?void 0:t.scrollLeft)!==null&&n!==void 0?n:0,y:(r=t==null?void 0:t.scrollTop)!==null&&r!==void 0?r:0,width:(i=t==null?void 0:t.clientWidth)!==null&&i!==void 0?i:0,height:(a=t==null?void 0:t.clientHeight)!==null&&a!==void 0?a:0}}constructor(t){this.ref=t}}class yS{isDisabled(t){var n;return this.disabledBehavior==="all"&&(((n=t.props)===null||n===void 0?void 0:n.isDisabled)||this.disabledKeys.has(t.key))}findNextNonDisabled(t,n){let r=t;for(;r!=null;){let i=this.collection.getItem(r);if((i==null?void 0:i.type)==="item"&&!this.isDisabled(i))return r;r=n(r)}return null}getNextKey(t){let n=t;return n=this.collection.getKeyAfter(n),this.findNextNonDisabled(n,r=>this.collection.getKeyAfter(r))}getPreviousKey(t){let n=t;return n=this.collection.getKeyBefore(n),this.findNextNonDisabled(n,r=>this.collection.getKeyBefore(r))}findKey(t,n,r){let i=t,a=this.layoutDelegate.getItemRect(i);if(!a||i==null)return null;let u=a;do{if(i=n(i),i==null)break;a=this.layoutDelegate.getItemRect(i)}while(a&&r(u,a)&&i!=null);return i}isSameRow(t,n){return t.y===n.y||t.x!==n.x}isSameColumn(t,n){return t.x===n.x||t.y!==n.y}getKeyBelow(t){return this.layout==="grid"&&this.orientation==="vertical"?this.findKey(t,n=>this.getNextKey(n),this.isSameRow):this.getNextKey(t)}getKeyAbove(t){return this.layout==="grid"&&this.orientation==="vertical"?this.findKey(t,n=>this.getPreviousKey(n),this.isSameRow):this.getPreviousKey(t)}getNextColumn(t,n){return n?this.getPreviousKey(t):this.getNextKey(t)}getKeyRightOf(t){let n=this.direction==="ltr"?"getKeyRightOf":"getKeyLeftOf";return this.layoutDelegate[n]?(t=this.layoutDelegate[n](t),this.findNextNonDisabled(t,r=>this.layoutDelegate[n](r))):this.layout==="grid"?this.orientation==="vertical"?this.getNextColumn(t,this.direction==="rtl"):this.findKey(t,r=>this.getNextColumn(r,this.direction==="rtl"),this.isSameColumn):this.orientation==="horizontal"?this.getNextColumn(t,this.direction==="rtl"):null}getKeyLeftOf(t){let n=this.direction==="ltr"?"getKeyLeftOf":"getKeyRightOf";return this.layoutDelegate[n]?(t=this.layoutDelegate[n](t),this.findNextNonDisabled(t,r=>this.layoutDelegate[n](r))):this.layout==="grid"?this.orientation==="vertical"?this.getNextColumn(t,this.direction==="ltr"):this.findKey(t,r=>this.getNextColumn(r,this.direction==="ltr"),this.isSameColumn):this.orientation==="horizontal"?this.getNextColumn(t,this.direction==="ltr"):null}getFirstKey(){let t=this.collection.getFirstKey();return this.findNextNonDisabled(t,n=>this.collection.getKeyAfter(n))}getLastKey(){let t=this.collection.getLastKey();return this.findNextNonDisabled(t,n=>this.collection.getKeyBefore(n))}getKeyPageAbove(t){let n=this.ref.current,r=this.layoutDelegate.getItemRect(t);if(!r)return null;if(n&&!l1(n))return this.getFirstKey();let i=t;if(this.orientation==="horizontal"){let a=Math.max(0,r.x+r.width-this.layoutDelegate.getVisibleRect().width);for(;r&&r.x>a&&i!=null;)i=this.getKeyAbove(i),r=i==null?null:this.layoutDelegate.getItemRect(i)}else{let a=Math.max(0,r.y+r.height-this.layoutDelegate.getVisibleRect().height);for(;r&&r.y>a&&i!=null;)i=this.getKeyAbove(i),r=i==null?null:this.layoutDelegate.getItemRect(i)}return i??this.getFirstKey()}getKeyPageBelow(t){let n=this.ref.current,r=this.layoutDelegate.getItemRect(t);if(!r)return null;if(n&&!l1(n))return this.getLastKey();let i=t;if(this.orientation==="horizontal"){let a=Math.min(this.layoutDelegate.getContentSize().width,r.y-r.width+this.layoutDelegate.getVisibleRect().width);for(;r&&r.x<a&&i!=null;)i=this.getKeyBelow(i),r=i==null?null:this.layoutDelegate.getItemRect(i)}else{let a=Math.min(this.layoutDelegate.getContentSize().height,r.y-r.height+this.layoutDelegate.getVisibleRect().height);for(;r&&r.y<a&&i!=null;)i=this.getKeyBelow(i),r=i==null?null:this.layoutDelegate.getItemRect(i)}return i??this.getLastKey()}getKeyForSearch(t,n){if(!this.collator)return null;let r=this.collection,i=n||this.getFirstKey();for(;i!=null;){let a=r.getItem(i);if(!a)return null;let u=a.textValue.slice(0,t.length);if(a.textValue&&this.collator.compare(u,t)===0)return i;i=this.getNextKey(i)}return null}constructor(...t){if(t.length===1){let n=t[0];this.collection=n.collection,this.ref=n.ref,this.collator=n.collator,this.disabledKeys=n.disabledKeys||new Set,this.disabledBehavior=n.disabledBehavior||"all",this.orientation=n.orientation||"vertical",this.direction=n.direction,this.layout=n.layout||"stack",this.layoutDelegate=n.layoutDelegate||new xj(n.ref)}else this.collection=t[0],this.disabledKeys=t[1],this.ref=t[2],this.collator=t[3],this.layout="stack",this.orientation="vertical",this.disabledBehavior="all",this.layoutDelegate=new xj(this.ref);this.layout==="stack"&&this.orientation==="vertical"&&(this.getKeyLeftOf=void 0,this.getKeyRightOf=void 0)}}function zN(e){let{selectionManager:t,collection:n,disabledKeys:r,ref:i,keyboardDelegate:a,layoutDelegate:u}=e,l=ZC({usage:"search",sensitivity:"base"}),o=t.disabledBehavior,s=P.useMemo(()=>a||new yS({collection:n,disabledKeys:r,disabledBehavior:o,ref:i,collator:l,layoutDelegate:u}),[a,u,n,r,i,l,o]),{collectionProps:c}=QN({...e,ref:i,selectionManager:t,keyboardDelegate:s});return{listProps:c}}function ZDe(e,t,n){let r=Gt(e,{labelable:!0}),i=e.selectionBehavior||"toggle",a=e.linkBehavior||(i==="replace"?"action":"override");i==="toggle"&&a==="action"&&(a="override");let{listProps:u}=zN({...e,ref:n,selectionManager:t.selectionManager,collection:t.collection,disabledKeys:t.disabledKeys,linkBehavior:a}),{focusWithinProps:l}=L0({onFocusWithin:e.onFocus,onBlurWithin:e.onBlur,onFocusWithinChange:e.onFocusChange}),o=vi(e.id);YC.set(t,{id:o,shouldUseVirtualFocus:e.shouldUseVirtualFocus,shouldSelectOnPressUp:e.shouldSelectOnPressUp,shouldFocusOnHover:e.shouldFocusOnHover,isVirtualized:e.isVirtualized,onAction:e.onAction,linkBehavior:a,UNSTABLE_itemBehavior:e.UNSTABLE_itemBehavior});let{labelProps:s,fieldProps:c}=gS({...e,id:o,labelElementType:"span"});return{labelProps:s,listBoxProps:Et(r,l,t.selectionManager.selectionMode==="multiple"?{"aria-multiselectable":"true"}:{},{role:"listbox",...Et(c,u)})}}class HDe{build(t,n){return this.context=n,Sj(()=>this.iterateCollection(t))}*iterateCollection(t){let{children:n,items:r}=t;if(be.isValidElement(n)&&n.type===be.Fragment)yield*this.iterateCollection({children:n.props.children,items:r});else if(typeof n=="function"){if(!r)throw new Error("props.children was a function but props.items is missing");let i=0;for(let a of r)yield*this.getFullNode({value:a,index:i},{renderer:n}),i++}else{let i=[];be.Children.forEach(n,u=>{u&&i.push(u)});let a=0;for(let u of i){let l=this.getFullNode({element:u,index:a},{});for(let o of l)a++,yield o}}}getKey(t,n,r,i){if(t.key!=null)return t.key;if(n.type==="cell"&&n.key!=null)return`${i}${n.key}`;let a=n.value;if(a!=null){var u;let l=(u=a.key)!==null&&u!==void 0?u:a.id;if(l==null)throw new Error("No key found for item");return l}return i?`${i}.${n.index}`:`$.${n.index}`}getChildState(t,n){return{renderer:n.renderer||t.renderer}}*getFullNode(t,n,r,i){if(be.isValidElement(t.element)&&t.element.type===be.Fragment){let S=[];be.Children.forEach(t.element.props.children,E=>{S.push(E)});var a;let _=(a=t.index)!==null&&a!==void 0?a:0;for(const E of S)yield*this.getFullNode({element:E,index:_++},n,r,i);return}let u=t.element;if(!u&&t.value&&n&&n.renderer){let S=this.cache.get(t.value);if(S&&(!S.shouldInvalidate||!S.shouldInvalidate(this.context))){S.index=t.index,S.parentKey=i?i.key:null,yield S;return}u=n.renderer(t.value)}if(be.isValidElement(u)){let S=u.type;if(typeof S!="function"&&typeof S.getCollectionNode!="function"){let k=u.type;throw new Error(`Unknown element <${k}> in collection.`)}let _=S.getCollectionNode(u.props,this.context);var l;let E=(l=t.index)!==null&&l!==void 0?l:0,$=_.next();for(;!$.done&&$.value;){let k=$.value;t.index=E;var o;let A=(o=k.key)!==null&&o!==void 0?o:null;A==null&&(A=k.element?null:this.getKey(u,t,n,r));let R=[...this.getFullNode({...k,key:A,index:E,wrapper:XDe(t.wrapper,k.wrapper)},this.getChildState(n,k),r?`${r}${u.key}`:u.key,i)];for(let F of R){var s,c;F.value=(c=(s=k.value)!==null&&s!==void 0?s:t.value)!==null&&c!==void 0?c:null,F.value&&this.cache.set(F.value,F);var d;if(t.type&&F.type!==t.type)throw new Error(`Unsupported type <${sk(F.type)}> in <${sk((d=i==null?void 0:i.type)!==null&&d!==void 0?d:"unknown parent type")}>. Only <${sk(t.type)}> is supported.`);E++,yield F}$=_.next(R)}return}if(t.key==null||t.type==null)return;let h=this;var m,y;let v={type:t.type,props:t.props,key:t.key,parentKey:i?i.key:null,value:(m=t.value)!==null&&m!==void 0?m:null,level:i?i.level+1:0,index:t.index,rendered:t.rendered,textValue:(y=t.textValue)!==null&&y!==void 0?y:"","aria-label":t["aria-label"],wrapper:t.wrapper,shouldInvalidate:t.shouldInvalidate,hasChildNodes:t.hasChildNodes||!1,childNodes:Sj(function*(){if(!t.hasChildNodes||!t.childNodes)return;let S=0;for(let _ of t.childNodes()){_.key!=null&&(_.key=`${v.key}${_.key}`);let E=h.getFullNode({..._,index:S},h.getChildState(n,_),v.key,v);for(let $ of E)S++,yield $}})};yield v}constructor(){this.cache=new WeakMap}}function Sj(e){let t=[],n=null;return{*[Symbol.iterator](){for(let r of t)yield r;n||(n=e());for(let r of n)t.push(r),yield r}}}function XDe(e,t){if(e&&t)return n=>e(t(n));if(e)return e;if(t)return t}function sk(e){return e[0].toUpperCase()+e.slice(1)}function Bre(e,t,n){let r=P.useMemo(()=>new HDe,[]),{children:i,items:a,collection:u}=e;return P.useMemo(()=>{if(u)return u;let o=r.build({children:i,items:a},n);return t(o)},[r,i,a,u,n,t])}function KC(e,t){return typeof t.getChildren=="function"?t.getChildren(e.key):e.childNodes}function YDe(e){return GDe(e)}function GDe(e,t){for(let n of e)return n}function ck(e,t,n){if(t.parentKey===n.parentKey)return t.index-n.index;let r=[..._j(e,t),t],i=[..._j(e,n),n],a=r.slice(0,i.length).findIndex((u,l)=>u!==i[l]);return a!==-1?(t=r[a],n=i[a],t.index-n.index):r.findIndex(u=>u===n)>=0?1:(i.findIndex(u=>u===t)>=0,-1)}function _j(e,t){let n=[],r=t;for(;(r==null?void 0:r.parentKey)!=null;)r=e.getItem(r.parentKey),r&&n.unshift(r);return n}const Ej=new WeakMap;function qN(e){let t=Ej.get(e);if(t!=null)return t;let n=0,r=i=>{for(let a of i)a.type==="section"?r(KC(a,e)):a.type==="item"&&n++};return r(e),Ej.set(e,n),n}function KDe(e,t,n){var r,i;let{key:a}=e,u=YC.get(t);var l;let o=(l=e.isDisabled)!==null&&l!==void 0?l:t.selectionManager.isDisabled(a);var s;let c=(s=e.isSelected)!==null&&s!==void 0?s:t.selectionManager.isSelected(a);var d;let h=(d=e.shouldSelectOnPressUp)!==null&&d!==void 0?d:u==null?void 0:u.shouldSelectOnPressUp;var m;let y=(m=e.shouldFocusOnHover)!==null&&m!==void 0?m:u==null?void 0:u.shouldFocusOnHover;var v;let S=(v=e.shouldUseVirtualFocus)!==null&&v!==void 0?v:u==null?void 0:u.shouldUseVirtualFocus;var _;let E=(_=e.isVirtualized)!==null&&_!==void 0?_:u==null?void 0:u.isVirtualized,$=$0(),k=$0(),A={role:"option","aria-disabled":o||void 0,"aria-selected":t.selectionManager.selectionMode!=="none"?c:void 0};cg()&&fN()||(A["aria-label"]=e["aria-label"],A["aria-labelledby"]=$,A["aria-describedby"]=k);let w=t.collection.getItem(a);if(E){let J=Number(w==null?void 0:w.index);A["aria-posinset"]=Number.isNaN(J)?void 0:J+1,A["aria-setsize"]=qN(t.collection)}let R=u!=null&&u.onAction?()=>{var J;return u==null||(J=u.onAction)===null||J===void 0?void 0:J.call(u,a)}:void 0,F=Nre(t,a),{itemProps:I,isPressed:U,isFocused:q,hasAction:V,allowsSelection:N}=GC({selectionManager:t.selectionManager,key:a,ref:n,shouldSelectOnPressUp:h,allowsDifferentPressOrigin:h&&y,isVirtualized:E,shouldUseVirtualFocus:S,isDisabled:o,onAction:R||!(w==null||(r=w.props)===null||r===void 0)&&r.onAction?ih(w==null||(i=w.props)===null||i===void 0?void 0:i.onAction,R):void 0,linkBehavior:u==null?void 0:u.linkBehavior,UNSTABLE_itemBehavior:u==null?void 0:u.UNSTABLE_itemBehavior,id:F}),{hoverProps:j}=Ru({isDisabled:o||!y,onHoverStart(){ey()||(t.selectionManager.setFocused(!0),t.selectionManager.setFocusedKey(a))}}),W=Gt(w==null?void 0:w.props);delete W.id;let K=hN(w==null?void 0:w.props);return{optionProps:{...A,...Et(W,I,j,K),id:F},labelProps:{id:$},descriptionProps:{id:k},isFocused:q,isFocusVisible:q&&t.selectionManager.isFocused&&ey(),isSelected:c,isDisabled:o,isPressed:U,allowsSelection:N,hasAction:V}}function WDe(e){let{heading:t,"aria-label":n}=e,r=vi();return{itemProps:{role:"presentation"},headingProps:t?{id:r,role:"presentation"}:{},groupProps:{role:"group","aria-label":n,"aria-labelledby":t?r:void 0}}}const Qre={badInput:!1,customError:!1,patternMismatch:!1,rangeOverflow:!1,rangeUnderflow:!1,stepMismatch:!1,tooLong:!1,tooShort:!1,typeMismatch:!1,valueMissing:!1,valid:!0},zre={...Qre,customError:!0,valid:!1},IO={isInvalid:!1,validationDetails:Qre,validationErrors:[]},qre=P.createContext({}),c1="__formValidationState"+Date.now();function Og(e){if(e[c1]){let{realtimeValidation:t,displayValidation:n,updateValidation:r,resetValidation:i,commitValidation:a}=e[c1];return{realtimeValidation:t,displayValidation:n,updateValidation:r,resetValidation:i,commitValidation:a}}return JDe(e)}function JDe(e){let{isInvalid:t,validationState:n,name:r,value:i,builtinValidation:a,validate:u,validationBehavior:l="aria"}=e;n&&(t||(t=n==="invalid"));let o=t!==void 0?{isInvalid:t,validationErrors:[],validationDetails:zre}:null,s=P.useMemo(()=>{if(!u||i==null)return null;let U=e$e(u,i);return Dj(U)},[u,i]);a!=null&&a.validationDetails.valid&&(a=void 0);let c=P.useContext(qre),d=P.useMemo(()=>r?Array.isArray(r)?r.flatMap(U=>X9(c[U])):X9(c[r]):[],[c,r]),[h,m]=P.useState(c),[y,v]=P.useState(!1);c!==h&&(m(c),v(!1));let S=P.useMemo(()=>Dj(y?[]:d),[y,d]),_=P.useRef(IO),[E,$]=P.useState(IO),k=P.useRef(IO),A=()=>{if(!w)return;R(!1);let U=s||a||_.current;fk(U,k.current)||(k.current=U,$(U))},[w,R]=P.useState(!1);return P.useEffect(A),{realtimeValidation:o||S||s||a||IO,displayValidation:l==="native"?o||S||E:o||S||s||a||E,updateValidation(U){l==="aria"&&!fk(E,U)?$(U):_.current=U},resetValidation(){let U=IO;fk(U,k.current)||(k.current=U,$(U)),l==="native"&&R(!1),v(!0)},commitValidation(){l==="native"&&R(!0),v(!0)}}}function X9(e){return e?Array.isArray(e)?e:[e]:[]}function e$e(e,t){if(typeof e=="function"){let n=e(t);if(n&&typeof n!="boolean")return X9(n)}return[]}function Dj(e){return e.length?{isInvalid:!0,validationErrors:e,validationDetails:zre}:null}function fk(e,t){return e===t?!0:!!e&&!!t&&e.isInvalid===t.isInvalid&&e.validationErrors.length===t.validationErrors.length&&e.validationErrors.every((n,r)=>n===t.validationErrors[r])&&Object.entries(e.validationDetails).every(([n,r])=>t.validationDetails[n]===r)}var Vre={};Vre={longPressMessage:"اضغط مطولاً أو اضغط على Alt + السهم لأسفل لفتح القائمة"};var Ure={};Ure={longPressMessage:"Натиснете продължително или натиснете Alt+ стрелка надолу, за да отворите менюто"};var jre={};jre={longPressMessage:"Dlouhým stiskem nebo stisknutím kláves Alt + šipka dolů otevřete nabídku"};var Zre={};Zre={longPressMessage:"Langt tryk eller tryk på Alt + pil ned for at åbne menuen"};var Hre={};Hre={longPressMessage:"Drücken Sie lange oder drücken Sie Alt + Nach-unten, um das Menü zu öffnen"};var Xre={};Xre={longPressMessage:"Πιέστε παρατεταμένα ή πατήστε Alt + κάτω βέλος για να ανοίξετε το μενού"};var Yre={};Yre={longPressMessage:"Long press or press Alt + ArrowDown to open menu"};var Gre={};Gre={longPressMessage:"Mantenga pulsado o pulse Alt + flecha abajo para abrir el menú"};var Kre={};Kre={longPressMessage:"Menüü avamiseks vajutage pikalt või vajutage klahve Alt + allanool"};var Wre={};Wre={longPressMessage:"Avaa valikko painamalla pohjassa tai näppäinyhdistelmällä Alt + Alanuoli"};var Jre={};Jre={longPressMessage:"Appuyez de manière prolongée ou appuyez sur Alt + Flèche vers le bas pour ouvrir le menu."};var eie={};eie={longPressMessage:"לחץ לחיצה ארוכה או הקש Alt + ArrowDown כדי לפתוח את התפריט"};var tie={};tie={longPressMessage:"Dugo pritisnite ili pritisnite Alt + strelicu prema dolje za otvaranje izbornika"};var nie={};nie={longPressMessage:"Nyomja meg hosszan, vagy nyomja meg az Alt + lefele nyíl gombot a menü megnyitásához"};var rie={};rie={longPressMessage:"Premere a lungo o premere Alt + Freccia giù per aprire il menu"};var iie={};iie={longPressMessage:"長押しまたは Alt+下矢印キーでメニューを開く"};var aie={};aie={longPressMessage:"길게 누르거나 Alt + 아래쪽 화살표를 눌러 메뉴 열기"};var uie={};uie={longPressMessage:"Norėdami atidaryti meniu, nuspaudę palaikykite arba paspauskite „Alt + ArrowDown“."};var lie={};lie={longPressMessage:"Lai atvērtu izvēlni, turiet nospiestu vai nospiediet taustiņu kombināciju Alt + lejupvērstā bultiņa"};var oie={};oie={longPressMessage:"Langt trykk eller trykk Alt + PilNed for å åpne menyen"};var sie={};sie={longPressMessage:"Druk lang op Alt + pijl-omlaag of druk op Alt om het menu te openen"};var cie={};cie={longPressMessage:"Naciśnij i przytrzymaj lub naciśnij klawisze Alt + Strzałka w dół, aby otworzyć menu"};var fie={};fie={longPressMessage:"Pressione e segure ou pressione Alt + Seta para baixo para abrir o menu"};var die={};die={longPressMessage:"Prima continuamente ou prima Alt + Seta Para Baixo para abrir o menu"};var hie={};hie={longPressMessage:"Apăsați lung sau apăsați pe Alt + săgeată în jos pentru a deschide meniul"};var pie={};pie={longPressMessage:"Нажмите и удерживайте или нажмите Alt + Стрелка вниз, чтобы открыть меню"};var mie={};mie={longPressMessage:"Ponuku otvoríte dlhým stlačením alebo stlačením klávesu Alt + klávesu so šípkou nadol"};var gie={};gie={longPressMessage:"Za odprtje menija pritisnite in držite gumb ali pritisnite Alt+puščica navzdol"};var vie={};vie={longPressMessage:"Dugo pritisnite ili pritisnite Alt + strelicu prema dole da otvorite meni"};var yie={};yie={longPressMessage:"Håll nedtryckt eller tryck på Alt + pil nedåt för att öppna menyn"};var bie={};bie={longPressMessage:"Menüyü açmak için uzun basın veya Alt + Aşağı Ok tuşuna basın"};var Oie={};Oie={longPressMessage:"Довго або звичайно натисніть комбінацію клавіш Alt і стрілка вниз, щоб відкрити меню"};var xie={};xie={longPressMessage:"长按或按 Alt + 向下方向键以打开菜单"};var Sie={};Sie={longPressMessage:"長按或按 Alt+向下鍵以開啟功能表"};var _ie={};_ie={"ar-AE":Vre,"bg-BG":Ure,"cs-CZ":jre,"da-DK":Zre,"de-DE":Hre,"el-GR":Xre,"en-US":Yre,"es-ES":Gre,"et-EE":Kre,"fi-FI":Wre,"fr-FR":Jre,"he-IL":eie,"hr-HR":tie,"hu-HU":nie,"it-IT":rie,"ja-JP":iie,"ko-KR":aie,"lt-LT":uie,"lv-LV":lie,"nb-NO":oie,"nl-NL":sie,"pl-PL":cie,"pt-BR":fie,"pt-PT":die,"ro-RO":hie,"ru-RU":pie,"sk-SK":mie,"sl-SI":gie,"sr-SP":vie,"sv-SE":yie,"tr-TR":bie,"uk-UA":Oie,"zh-CN":xie,"zh-TW":Sie};function t$e(e){return e&&e.__esModule?e.default:e}function VN(e,t,n){let{type:r="menu",isDisabled:i,trigger:a="press"}=e,u=vi(),{triggerProps:l,overlayProps:o}=Kne({type:r},t,n),s=m=>{if(!i&&!(a==="longPress"&&!m.altKey)&&n&&n.current)switch(m.key){case"Enter":case" ":if(a==="longPress"||m.isDefaultPrevented())return;case"ArrowDown":"continuePropagation"in m||m.stopPropagation(),m.preventDefault(),t.toggle("first");break;case"ArrowUp":"continuePropagation"in m||m.stopPropagation(),m.preventDefault(),t.toggle("last");break;default:"continuePropagation"in m&&m.continuePropagation()}},c=Qo(t$e(_ie),"@react-aria/menu"),{longPressProps:d}=jne({isDisabled:i||a!=="longPress",accessibilityDescription:c.format("longPressMessage"),onLongPressStart(){t.close()},onLongPress(){t.open("first")}}),h={preventFocusOnPress:!0,onPressStart(m){m.pointerType!=="touch"&&m.pointerType!=="keyboard"&&!i&&($u(m.target),t.open(m.pointerType==="virtual"?"first":null))},onPress(m){m.pointerType==="touch"&&!i&&($u(m.target),t.toggle())}};return delete l.onPress,{menuTriggerProps:{...l,...a==="press"?h:d,id:u,onKeyDown:s},menuProps:{...o,"aria-labelledby":u,autoFocus:t.focusStrategy||!0,onClose:t.close}}}const Eie=new WeakMap;function n$e(e,t,n){let{shouldFocusWrap:r=!0,onKeyDown:i,onKeyUp:a,...u}=e;!e["aria-label"]&&e["aria-labelledby"];let l=Gt(e,{labelable:!0}),{listProps:o}=zN({...u,ref:n,selectionManager:t.selectionManager,collection:t.collection,disabledKeys:t.disabledKeys,shouldFocusWrap:r,linkBehavior:"override"});return Eie.set(t,{onClose:e.onClose,onAction:e.onAction,shouldUseVirtualFocus:e.shouldUseVirtualFocus}),{menuProps:Et(l,{onKeyDown:i,onKeyUp:a},{role:"menu",...o,onKeyDown:s=>{var c;(s.key!=="Escape"||e.shouldUseVirtualFocus)&&((c=o.onKeyDown)===null||c===void 0||c.call(o,s))}})}}function r$e(e,t,n){let{id:r,key:i,closeOnSelect:a,isVirtualized:u,"aria-haspopup":l,onPressStart:o,onPressUp:s,onPress:c,onPressChange:d,onPressEnd:h,onClick:m,onHoverStart:y,onHoverChange:v,onHoverEnd:S,onKeyDown:_,onKeyUp:E,onFocus:$,onFocusChange:k,onBlur:A,selectionManager:w=t.selectionManager}=e,R=!!l,F=R&&e["aria-expanded"]==="true";var I;let U=(I=e.isDisabled)!==null&&I!==void 0?I:w.isDisabled(i);var q;let V=(q=e.isSelected)!==null&&q!==void 0?q:w.isSelected(i),N=Eie.get(t),j=t.collection.getItem(i),W=e.onClose||N.onClose,K=X3(),J=()=>{var Re;if(!R&&(!(j==null||(Re=j.props)===null||Re===void 0)&&Re.onAction?j.props.onAction():e.onAction&&e.onAction(i),N.onAction)){let et=N.onAction;et(i)}},te="menuitem";R||(w.selectionMode==="single"?te="menuitemradio":w.selectionMode==="multiple"&&(te="menuitemcheckbox"));let H=$0(),M=$0(),G=$0(),X={id:r,"aria-disabled":U||void 0,role:te,"aria-label":e["aria-label"],"aria-labelledby":H,"aria-describedby":[M,G].filter(Boolean).join(" ")||void 0,"aria-controls":e["aria-controls"],"aria-haspopup":l,"aria-expanded":e["aria-expanded"]};w.selectionMode!=="none"&&!R&&(X["aria-checked"]=V),u&&(X["aria-posinset"]=j==null?void 0:j.index,X["aria-setsize"]=qN(t.collection));let B=P.useRef(!1),ne=Re=>{d==null||d(Re),B.current=Re},oe=P.useRef(null),ie=Re=>{Re.pointerType!=="keyboard"&&(oe.current={pointerType:Re.pointerType}),Re.pointerType==="mouse"&&(B.current||Re.target.click()),s==null||s(Re)},de=Re=>{var et,tt;m==null||m(Re),J(),c_e(Re,K,j.props.href,j==null?void 0:j.props.routerOptions);let Oe=((et=oe.current)===null||et===void 0?void 0:et.pointerType)==="keyboard"?((tt=oe.current)===null||tt===void 0?void 0:tt.key)==="Enter"||w.selectionMode==="none"||w.isLink(i):w.selectionMode!=="multiple"||w.isLink(i);Oe=a??Oe,W&&!R&&Oe&&W(),oe.current=null},{itemProps:ge,isFocused:ye}=GC({id:r,selectionManager:w,key:i,ref:n,shouldSelectOnPressUp:!0,allowsDifferentPressOrigin:!0,linkBehavior:"none",shouldUseVirtualFocus:N.shouldUseVirtualFocus}),{pressProps:me,isPressed:Ee}=lp({onPressStart:o,onPress:c,onPressUp:ie,onPressChange:ne,onPressEnd:h,isDisabled:U}),{hoverProps:se}=Ru({isDisabled:U,onHoverStart(Re){!ey()&&!(F&&l)&&(w.setFocused(!0),w.setFocusedKey(i)),y==null||y(Re)},onHoverChange:v,onHoverEnd:S}),{keyboardProps:pe}=kN({onKeyDown:Re=>{if(Re.repeat){Re.continuePropagation();return}switch(Re.key){case" ":oe.current={pointerType:"keyboard",key:" "},Re.target.click();break;case"Enter":oe.current={pointerType:"keyboard",key:"Enter"},Re.target.tagName!=="A"&&Re.target.click();break;default:R||Re.continuePropagation(),_==null||_(Re);break}},onKeyUp:E}),{focusProps:ve}=XC({onBlur:A,onFocus:$,onFocusChange:k}),we=Gt(j==null?void 0:j.props);delete we.id;let Me=hN(j==null?void 0:j.props);return{menuItemProps:{...X,...Et(we,Me,R?{onFocus:ge.onFocus,"data-collection":ge["data-collection"],"data-key":ge["data-key"]}:ge,me,se,pe,ve,N.shouldUseVirtualFocus||R?{onMouseDown:Re=>Re.preventDefault()}:void 0,U?void 0:{onClick:de}),tabIndex:ge.tabIndex!=null&&F&&!N.shouldUseVirtualFocus?-1:ge.tabIndex},labelProps:{id:H},descriptionProps:{id:M},keyboardShortcutProps:{id:G},isFocused:ye,isFocusVisible:ye&&w.isFocused&&ey()&&!F,isSelected:V,isPressed:Ee,isDisabled:U}}function i$e(e){let{heading:t,"aria-label":n}=e,r=vi();return{itemProps:{role:"presentation"},headingProps:t?{id:r,role:"presentation"}:{},groupProps:{role:"group","aria-label":n,"aria-labelledby":t?r:void 0}}}const K_=2,a$e=50,u$e=1e3,$j=Math.PI/12;function l$e(e){let{menuRef:t,submenuRef:n,isOpen:r,isDisabled:i}=e,a=P.useRef(void 0),u=P.useRef(void 0),l=P.useRef(0),o=P.useRef(void 0),s=P.useRef(void 0),c=P.useRef(void 0),d=P.useRef(2),[h,m]=P.useState(!1);u1({ref:n,onResize:()=>{n.current&&(u.current=n.current.getBoundingClientRect(),c.current=void 0)}});let v=()=>{m(!1),d.current=K_,a.current=void 0},S=TN(),_=fi(E=>{h&&E.preventDefault()});P.useEffect(()=>{h&&t.current?t.current.style.pointerEvents="none":t.current.style.pointerEvents=""},[t,h]),P.useEffect(()=>{let E=n.current,$=t.current;if(i||!E||!r||S!=="pointer"||!$){v();return}u.current=E.getBoundingClientRect();let k=A=>{if(A.pointerType==="touch"||A.pointerType==="pen")return;let w=Date.now();if(w-l.current<a$e)return;clearTimeout(o.current),clearTimeout(s.current);let{clientX:R,clientY:F}=A;if(!a.current){a.current={x:R,y:F};return}if(!u.current)return;if(c.current||(c.current=R>u.current.right?"left":"right"),R<$.getBoundingClientRect().left||R>$.getBoundingClientRect().right||F<$.getBoundingClientRect().top||F>$.getBoundingClientRect().bottom){v();return}let I=a.current.x,U=a.current.y,q=c.current==="right"?u.current.left-I:I-u.current.right,V=Math.atan2(U-u.current.top,q)+$j,N=Math.atan2(U-u.current.bottom,q)-$j,j=Math.atan2(U-F,c.current==="left"?-(R-I):R-I),W=j<V&&j>N;d.current=W?Math.min(d.current+1,K_):Math.max(d.current-1,0),d.current>=K_?m(!0):m(!1),l.current=w,a.current={x:R,y:F},W&&(o.current=setTimeout(()=>{v(),s.current=setTimeout(()=>{let K=document.elementFromPoint(R,F);K&&$.contains(K)&&K.dispatchEvent(new PointerEvent("pointerover",{bubbles:!0,cancelable:!0}))},100)},u$e))};return window.addEventListener("pointermove",k),window.addEventListener("pointerdown",_,!0),()=>{window.removeEventListener("pointermove",k),window.removeEventListener("pointerdown",_,!0),clearTimeout(o.current),clearTimeout(s.current),d.current=K_}},[i,r,t,S,m,_,n])}function o$e(e,t,n){let{parentMenuRef:r,submenuRef:i,type:a="menu",isDisabled:u,delay:l=200,shouldUseVirtualFocus:o}=e,s=vi(),c=vi(),{direction:d}=Wi(),h=P.useRef(void 0),m=P.useCallback(()=>{h.current&&(clearTimeout(h.current),h.current=void 0)},[h]),y=fi(F=>{m(),t.open(F)}),v=fi(()=>{m(),t.close()});Hr(()=>()=>{m()},[m]);let S=F=>{if(F.currentTarget.contains(document.activeElement))switch(F.key){case"ArrowLeft":d==="ltr"&&F.currentTarget.contains(F.target)&&(F.preventDefault(),F.stopPropagation(),v(),!o&&n.current&&$u(n.current));break;case"ArrowRight":d==="rtl"&&F.currentTarget.contains(F.target)&&(F.preventDefault(),F.stopPropagation(),v(),!o&&n.current&&$u(n.current));break;case"Escape":var I;!((I=i.current)===null||I===void 0)&&I.contains(F.target)&&(F.stopPropagation(),v(),!o&&n.current&&$u(n.current));break}};var _;let E={id:c,"aria-labelledby":s,submenuLevel:t.submenuLevel,...a==="menu"&&{onClose:t.closeAll,autoFocus:(_=t.focusStrategy)!==null&&_!==void 0?_:void 0,onKeyDown:S}},$=F=>{switch(F.key){case"ArrowRight":u||(d==="ltr"?(F.preventDefault(),t.isOpen||y("first"),a==="menu"&&(i!=null&&i.current)&&document.activeElement===(n==null?void 0:n.current)&&$u(i.current)):t.isOpen?v():F.continuePropagation());break;case"ArrowLeft":u||(d==="rtl"?(F.preventDefault(),t.isOpen||y("first"),a==="menu"&&(i!=null&&i.current)&&document.activeElement===(n==null?void 0:n.current)&&$u(i.current)):t.isOpen?v():F.continuePropagation());break;default:F.continuePropagation();break}},k=F=>{!u&&(F.pointerType==="virtual"||F.pointerType==="keyboard")&&y("first")},A=F=>{!u&&(F.pointerType==="touch"||F.pointerType==="mouse")&&y()},w=F=>{u||(F&&!t.isOpen?h.current||(h.current=setTimeout(()=>{y()},l)):F||m())};Yd(r,"focusin",F=>{var I;t.isOpen&&(!((I=r.current)===null||I===void 0)&&I.contains(F.target))&&F.target!==n.current&&v()});let R=F=>F!==n.current;return l$e({menuRef:r,submenuRef:i,isOpen:t.isOpen,isDisabled:u}),{submenuTriggerProps:{id:s,"aria-controls":t.isOpen?c:void 0,"aria-haspopup":u?void 0:a,"aria-expanded":t.isOpen?"true":"false",onPressStart:k,onPress:A,onHoverChange:w,onKeyDown:$,isOpen:t.isOpen},submenuProps:E,popoverProps:{isNonModal:!0,shouldCloseOnInteractOutside:R}}}function bS(e,t,n){let{validationBehavior:r,focus:i}=e;Hr(()=>{if(r==="native"&&(n!=null&&n.current)&&!n.current.disabled){let s=t.realtimeValidation.isInvalid?t.realtimeValidation.validationErrors.join(" ")||"Invalid value.":"";n.current.setCustomValidity(s),n.current.hasAttribute("title")||(n.current.title=""),t.realtimeValidation.isInvalid||t.updateValidation(c$e(n.current))}});let a=P.useRef(!1),u=fi(()=>{a.current||t.resetValidation()}),l=fi(s=>{var c;t.displayValidation.isInvalid||t.commitValidation();let d=n==null||(c=n.current)===null||c===void 0?void 0:c.form;if(!s.defaultPrevented&&n&&d&&f$e(d)===n.current){var h;i?i():(h=n.current)===null||h===void 0||h.focus(),AN("keyboard")}s.preventDefault()}),o=fi(()=>{t.commitValidation()});P.useEffect(()=>{let s=n==null?void 0:n.current;if(!s)return;let c=s.form,d=c==null?void 0:c.reset;return c&&(c.reset=()=>{a.current=!window.event||window.event.type==="message"&&window.event.target instanceof MessagePort,d==null||d.call(c),a.current=!1}),s.addEventListener("invalid",l),s.addEventListener("change",o),c==null||c.addEventListener("reset",u),()=>{s.removeEventListener("invalid",l),s.removeEventListener("change",o),c==null||c.removeEventListener("reset",u),c&&(c.reset=d)}},[n,l,o,u,r])}function s$e(e){let t=e.validity;return{badInput:t.badInput,customError:t.customError,patternMismatch:t.patternMismatch,rangeOverflow:t.rangeOverflow,rangeUnderflow:t.rangeUnderflow,stepMismatch:t.stepMismatch,tooLong:t.tooLong,tooShort:t.tooShort,typeMismatch:t.typeMismatch,valueMissing:t.valueMissing,valid:t.valid}}function c$e(e){return{isInvalid:!e.validity.valid,validationDetails:s$e(e),validationErrors:e.validationMessage?[e.validationMessage]:[]}}function f$e(e){for(let t=0;t<e.elements.length;t++){let n=e.elements[t];if(!n.validity.valid)return n}return null}function WC(e,t){let{inputElementType:n="input",isDisabled:r=!1,isRequired:i=!1,isReadOnly:a=!1,type:u="text",validationBehavior:l="aria"}=e,[o,s]=Yu(e.value,e.defaultValue||"",e.onChange),{focusableProps:c}=I0(e,t),d=Og({...e,value:o}),{isInvalid:h,validationErrors:m,validationDetails:y}=d.displayValidation,{labelProps:v,fieldProps:S,descriptionProps:_,errorMessageProps:E}=vS({...e,isInvalid:h,errorMessage:e.errorMessage||m}),$=Gt(e,{labelable:!0});const k={type:u,pattern:e.pattern};let[A]=P.useState(o);var w;return my(t,(w=e.defaultValue)!==null&&w!==void 0?w:A,s),bS(e,d,t),P.useEffect(()=>{if(t.current instanceof uf(t.current).HTMLTextAreaElement){let R=t.current;Object.defineProperty(R,"defaultValue",{get:()=>R.value,set:()=>{},configurable:!0})}},[t]),{labelProps:v,inputProps:Et($,n==="input"?k:void 0,{disabled:r,readOnly:a,required:i&&l==="native","aria-required":i&&l==="aria"||void 0,"aria-invalid":h||void 0,"aria-errormessage":e["aria-errormessage"],"aria-activedescendant":e["aria-activedescendant"],"aria-autocomplete":e["aria-autocomplete"],"aria-haspopup":e["aria-haspopup"],"aria-controls":e["aria-controls"],value:o,onChange:R=>s(R.target.value),autoComplete:e.autoComplete,autoCapitalize:e.autoCapitalize,maxLength:e.maxLength,minLength:e.minLength,name:e.name,form:e.form,placeholder:e.placeholder,inputMode:e.inputMode,autoCorrect:e.autoCorrect,spellCheck:e.spellCheck,[parseInt(be.version,10)>=17?"enterKeyHint":"enterkeyhint"]:e.enterKeyHint,onCopy:e.onCopy,onCut:e.onCut,onPaste:e.onPaste,onCompositionEnd:e.onCompositionEnd,onCompositionStart:e.onCompositionStart,onCompositionUpdate:e.onCompositionUpdate,onSelect:e.onSelect,onBeforeInput:e.onBeforeInput,onInput:e.onInput,...c,...S}),descriptionProps:_,errorMessageProps:E,isInvalid:h,validationErrors:m,validationDetails:y}}function Cj(){return typeof window<"u"&&window.InputEvent&&typeof InputEvent.prototype.getTargetRanges=="function"}function Die(e,t,n){let r=fi(d=>{let h=n.current;if(!h)return;let m=null;switch(d.inputType){case"historyUndo":case"historyRedo":return;case"insertLineBreak":return;case"deleteContent":case"deleteByCut":case"deleteByDrag":m=h.value.slice(0,h.selectionStart)+h.value.slice(h.selectionEnd);break;case"deleteContentForward":m=h.selectionEnd===h.selectionStart?h.value.slice(0,h.selectionStart)+h.value.slice(h.selectionEnd+1):h.value.slice(0,h.selectionStart)+h.value.slice(h.selectionEnd);break;case"deleteContentBackward":m=h.selectionEnd===h.selectionStart?h.value.slice(0,h.selectionStart-1)+h.value.slice(h.selectionStart):h.value.slice(0,h.selectionStart)+h.value.slice(h.selectionEnd);break;case"deleteSoftLineBackward":case"deleteHardLineBackward":m=h.value.slice(h.selectionStart);break;default:d.data!=null&&(m=h.value.slice(0,h.selectionStart)+d.data+h.value.slice(h.selectionEnd));break}(m==null||!t.validate(m))&&d.preventDefault()});P.useEffect(()=>{if(!Cj()||!n.current)return;let d=n.current;return d.addEventListener("beforeinput",r,!1),()=>{d.removeEventListener("beforeinput",r,!1)}},[n,r]);let i=Cj()?null:d=>{let h=d.target.value.slice(0,d.target.selectionStart)+d.data+d.target.value.slice(d.target.selectionEnd);t.validate(h)||d.preventDefault()},{labelProps:a,inputProps:u,descriptionProps:l,errorMessageProps:o,...s}=WC(e,n),c=P.useRef(null);return{inputProps:Et(u,{onBeforeInput:i,onCompositionStart(){let{value:d,selectionStart:h,selectionEnd:m}=n.current;c.current={value:d,selectionStart:h,selectionEnd:m}},onCompositionEnd(){if(n.current&&!t.validate(n.current.value)){let{value:d,selectionStart:h,selectionEnd:m}=c.current;n.current.value=d,n.current.setSelectionRange(h,m),t.setInputValue(d)}}}),labelProps:a,descriptionProps:l,errorMessageProps:o,...s}}function d$e(e){return e&&e.__esModule?e.default:e}function h$e(e,t){let{buttonRef:n,popoverRef:r,inputRef:i,listBoxRef:a,keyboardDelegate:u,layoutDelegate:l,shouldFocusWrap:o,isReadOnly:s,isDisabled:c}=e,d=P.useRef(null);n=n??d;let h=Qo(d$e(Hte),"@react-aria/combobox"),{menuTriggerProps:m,menuProps:y}=VN({type:"listbox",isDisabled:c||s},t,n);YC.set(t,{id:y.id});let{collection:v}=t,{disabledKeys:S}=t.selectionManager,_=P.useMemo(()=>u||new yS({collection:v,disabledKeys:S,ref:a,layoutDelegate:l}),[u,l,v,S,a]),{collectionProps:E}=QN({selectionManager:t.selectionManager,keyboardDelegate:_,disallowTypeAhead:!0,disallowEmptySelection:!0,shouldFocusWrap:o,ref:i,isVirtualized:!0}),$=X3(),k=Ee=>{if(!Ee.nativeEvent.isComposing)switch(Ee.key){case"Enter":case"Tab":if(t.isOpen&&Ee.key==="Enter"&&Ee.preventDefault(),t.isOpen&&a.current&&t.selectionManager.focusedKey!=null){let se=t.collection.getItem(t.selectionManager.focusedKey);if(se!=null&&se.props.href){let pe=a.current.querySelector(`[data-key="${CSS.escape(t.selectionManager.focusedKey.toString())}"]`);Ee.key==="Enter"&&pe instanceof HTMLAnchorElement&&$.open(pe,Ee,se.props.href,se.props.routerOptions),t.close();break}else if(se!=null&&se.props.onAction){se.props.onAction(),t.close();break}}t.commit();break;case"Escape":(t.selectedKey!==null||t.inputValue===""||e.allowsCustomValue)&&Ee.continuePropagation(),t.revert();break;case"ArrowDown":t.open("first","manual");break;case"ArrowUp":t.open("last","manual");break;case"ArrowLeft":case"ArrowRight":t.selectionManager.setFocusedKey(null);break}},A=Ee=>{var se;let pe=(n==null?void 0:n.current)&&n.current===Ee.relatedTarget,ve=(se=r.current)===null||se===void 0?void 0:se.contains(Ee.relatedTarget);pe||ve||(e.onBlur&&e.onBlur(Ee),t.setFocused(!1))},w=Ee=>{t.isFocused||(e.onFocus&&e.onFocus(Ee),t.setFocused(!0))},{isInvalid:R,validationErrors:F,validationDetails:I}=t.displayValidation,{labelProps:U,inputProps:q,descriptionProps:V,errorMessageProps:N}=WC({...e,onChange:t.setInputValue,onKeyDown:s?e.onKeyDown:ih(t.isOpen&&E.onKeyDown,k,e.onKeyDown),onBlur:A,value:t.inputValue,defaultValue:t.defaultInputValue,onFocus:w,autoComplete:"off",validate:void 0,[c1]:t},i),j=Ee=>{if(Ee.pointerType==="touch"){var se;(se=i.current)===null||se===void 0||se.focus(),t.toggle(null,"manual")}},W=Ee=>{if(Ee.pointerType!=="touch"){var se;(se=i.current)===null||se===void 0||se.focus(),t.toggle(Ee.pointerType==="keyboard"||Ee.pointerType==="virtual"?"first":null,"manual")}},K=Gv({id:m.id,"aria-label":h.format("buttonLabel"),"aria-labelledby":e["aria-labelledby"]||U.id}),J=Gv({id:y.id,"aria-label":h.format("listboxLabel"),"aria-labelledby":e["aria-labelledby"]||U.id}),te=P.useRef(0),H=Ee=>{if(c||s)return;if(Ee.timeStamp-te.current<500){var se;Ee.preventDefault(),(se=i.current)===null||se===void 0||se.focus();return}let pe=Ee.target.getBoundingClientRect(),ve=Ee.changedTouches[0],we=Math.ceil(pe.left+.5*pe.width),Me=Math.ceil(pe.top+.5*pe.height);if(ve.clientX===we&&ve.clientY===Me){var Re;Ee.preventDefault(),(Re=i.current)===null||Re===void 0||Re.focus(),t.toggle(null,"manual"),te.current=Ee.timeStamp}},M=t.selectionManager.focusedKey!=null&&t.isOpen?t.collection.getItem(t.selectionManager.focusedKey):void 0;var G;let X=(G=M==null?void 0:M.parentKey)!==null&&G!==void 0?G:null;var B;let ne=(B=t.selectionManager.focusedKey)!==null&&B!==void 0?B:null,oe=P.useRef(X),ie=P.useRef(ne);P.useEffect(()=>{if(JE()&&M!=null&&ne!=null&&ne!==ie.current){let se=t.selectionManager.isSelected(ne),pe=X!=null?t.collection.getItem(X):null,ve=(pe==null?void 0:pe["aria-label"])||(typeof(pe==null?void 0:pe.rendered)=="string"?pe.rendered:"")||"";var Ee;let we=h.format("focusAnnouncement",{isGroupChange:(Ee=pe&&X!==oe.current)!==null&&Ee!==void 0?Ee:!1,groupTitle:ve,groupCount:pe?[...KC(pe,t.collection)].length:0,optionText:M["aria-label"]||M.textValue||"",isSelected:se});Zb(we)}oe.current=X,ie.current=ne});let de=qN(t.collection),ge=P.useRef(de),ye=P.useRef(t.isOpen);P.useEffect(()=>{let Ee=t.isOpen!==ye.current&&(t.selectionManager.focusedKey==null||JE());if(t.isOpen&&(Ee||de!==ge.current)){let se=h.format("countAnnouncement",{optionCount:de});Zb(se)}ge.current=de,ye.current=t.isOpen});let me=P.useRef(t.selectedKey);return P.useEffect(()=>{if(JE()&&t.isFocused&&t.selectedItem&&t.selectedKey!==me.current){let Ee=t.selectedItem["aria-label"]||t.selectedItem.textValue||"",se=h.format("selectedAnnouncement",{optionText:Ee});Zb(se)}me.current=t.selectedKey}),P.useEffect(()=>{if(t.isOpen)return BN([i.current,r.current].filter(Ee=>Ee!=null))},[t.isOpen,i,r]),ate(()=>{!M&&i.current&&Zu(hi(i.current))===i.current&&Sx(i.current,null)},[M]),Yd(a,"react-aria-item-action",t.isOpen?()=>{t.close()}:void 0),{labelProps:U,buttonProps:{...m,...K,excludeFromTabOrder:!0,preventFocusOnPress:!0,onPress:j,onPressStart:W,isDisabled:c||s},inputProps:Et(q,{role:"combobox","aria-expanded":m["aria-expanded"],"aria-controls":t.isOpen?y.id:void 0,"aria-autocomplete":"list","aria-activedescendant":M?Nre(t,M.key):void 0,onTouchEnd:H,autoCorrect:"off",spellCheck:"false"}),listBoxProps:Et(y,J,{autoFocus:t.focusStrategy||!0,shouldUseVirtualFocus:!0,shouldSelectOnPressUp:!0,shouldFocusOnHover:!0,linkBehavior:"selection",UNSTABLE_itemBehavior:"action"}),descriptionProps:V,errorMessageProps:N,isInvalid:R,validationErrors:F,validationDetails:I}}var $ie={};$ie={"Clear search":"مسح البحث"};var Cie={};Cie={"Clear search":"Изчистване на търсене"};var wie={};wie={"Clear search":"Vymazat hledání"};var Rie={};Rie={"Clear search":"Ryd søgning"};var Aie={};Aie={"Clear search":"Suche zurücksetzen"};var Tie={};Tie={"Clear search":"Απαλοιφή αναζήτησης"};var kie={};kie={"Clear search":"Clear search"};var Pie={};Pie={"Clear search":"Borrar búsqueda"};var Mie={};Mie={"Clear search":"Tühjenda otsing"};var Fie={};Fie={"Clear search":"Tyhjennä haku"};var Nie={};Nie={"Clear search":"Effacer la recherche"};var Iie={};Iie={"Clear search":"נקה חיפוש"};var Lie={};Lie={"Clear search":"Obriši pretragu"};var Bie={};Bie={"Clear search":"Keresés törlése"};var Qie={};Qie={"Clear search":"Cancella ricerca"};var zie={};zie={"Clear search":"検索をクリア"};var qie={};qie={"Clear search":"검색 지우기"};var Vie={};Vie={"Clear search":"Išvalyti iešką"};var Uie={};Uie={"Clear search":"Notīrīt meklēšanu"};var jie={};jie={"Clear search":"Tøm søk"};var Zie={};Zie={"Clear search":"Zoekactie wissen"};var Hie={};Hie={"Clear search":"Wyczyść zawartość wyszukiwania"};var Xie={};Xie={"Clear search":"Limpar pesquisa"};var Yie={};Yie={"Clear search":"Limpar pesquisa"};var Gie={};Gie={"Clear search":"Ştergeţi căutarea"};var Kie={};Kie={"Clear search":"Очистить поиск"};var Wie={};Wie={"Clear search":"Vymazať vyhľadávanie"};var Jie={};Jie={"Clear search":"Počisti iskanje"};var eae={};eae={"Clear search":"Obriši pretragu"};var tae={};tae={"Clear search":"Rensa sökning"};var nae={};nae={"Clear search":"Aramayı temizle"};var rae={};rae={"Clear search":"Очистити пошук"};var iae={};iae={"Clear search":"清除搜索"};var aae={};aae={"Clear search":"清除搜尋條件"};var uae={};uae={"ar-AE":$ie,"bg-BG":Cie,"cs-CZ":wie,"da-DK":Rie,"de-DE":Aie,"el-GR":Tie,"en-US":kie,"es-ES":Pie,"et-EE":Mie,"fi-FI":Fie,"fr-FR":Nie,"he-IL":Iie,"hr-HR":Lie,"hu-HU":Bie,"it-IT":Qie,"ja-JP":zie,"ko-KR":qie,"lt-LT":Vie,"lv-LV":Uie,"nb-NO":jie,"nl-NL":Zie,"pl-PL":Hie,"pt-BR":Xie,"pt-PT":Yie,"ro-RO":Gie,"ru-RU":Kie,"sk-SK":Wie,"sl-SI":Jie,"sr-SP":eae,"sv-SE":tae,"tr-TR":nae,"uk-UA":rae,"zh-CN":iae,"zh-TW":aae};function p$e(e){return e&&e.__esModule?e.default:e}function m$e(e,t,n){let r=Qo(p$e(uae),"@react-aria/searchfield"),{isDisabled:i,isReadOnly:a,onSubmit:u,onClear:l,type:o="search"}=e,s=_=>{const E=_.key;E==="Enter"&&(i||a)&&_.preventDefault(),!(i||a)&&(E==="Enter"&&u&&(_.preventDefault(),u(t.value)),E==="Escape"&&(t.value===""&&(!n.current||n.current.value==="")?_.continuePropagation():(_.preventDefault(),t.setValue(""),l&&l())))},c=()=>{t.setValue(""),l&&l()},d=()=>{var _;(_=n.current)===null||_===void 0||_.focus()},{labelProps:h,inputProps:m,descriptionProps:y,errorMessageProps:v,...S}=WC({...e,value:t.value,onChange:t.setValue,onKeyDown:a?e.onKeyDown:ih(s,e.onKeyDown),type:o},n);return{labelProps:h,inputProps:{...m,defaultValue:void 0},clearButtonProps:{"aria-label":r.format("Clear search"),excludeFromTabOrder:!0,preventFocusOnPress:!0,isDisabled:i||a,onPress:c,onPressStart:d},descriptionProps:y,errorMessageProps:v,...S}}var lae={};lae={collectionLabel:"مقترحات"};var oae={};oae={collectionLabel:"Предложения"};var sae={};sae={collectionLabel:"Návrhy"};var cae={};cae={collectionLabel:"Forslag"};var fae={};fae={collectionLabel:"Empfehlungen"};var dae={};dae={collectionLabel:"Προτάσεις"};var hae={};hae={collectionLabel:"Suggestions"};var pae={};pae={collectionLabel:"Sugerencias"};var mae={};mae={collectionLabel:"Soovitused"};var gae={};gae={collectionLabel:"Ehdotukset"};var vae={};vae={collectionLabel:"Suggestions"};var yae={};yae={collectionLabel:"הצעות"};var bae={};bae={collectionLabel:"Prijedlozi"};var Oae={};Oae={collectionLabel:"Javaslatok"};var xae={};xae={collectionLabel:"Suggerimenti"};var Sae={};Sae={collectionLabel:"候補"};var _ae={};_ae={collectionLabel:"제안"};var Eae={};Eae={collectionLabel:"Pasiūlymai"};var Dae={};Dae={collectionLabel:"Ieteikumi"};var $ae={};$ae={collectionLabel:"Forslag"};var Cae={};Cae={collectionLabel:"Suggesties"};var wae={};wae={collectionLabel:"Sugestie"};var Rae={};Rae={collectionLabel:"Sugestões"};var Aae={};Aae={collectionLabel:"Sugestões"};var Tae={};Tae={collectionLabel:"Sugestii"};var kae={};kae={collectionLabel:"Предложения"};var Pae={};Pae={collectionLabel:"Návrhy"};var Mae={};Mae={collectionLabel:"Predlogi"};var Fae={};Fae={collectionLabel:"Predlozi"};var Nae={};Nae={collectionLabel:"Förslag"};var Iae={};Iae={collectionLabel:"Öneriler"};var Lae={};Lae={collectionLabel:"Пропозиції"};var Bae={};Bae={collectionLabel:"建议"};var Qae={};Qae={collectionLabel:"建議"};var zae={};zae={"ar-AE":lae,"bg-BG":oae,"cs-CZ":sae,"da-DK":cae,"de-DE":fae,"el-GR":dae,"en-US":hae,"es-ES":pae,"et-EE":mae,"fi-FI":gae,"fr-FR":vae,"he-IL":yae,"hr-HR":bae,"hu-HU":Oae,"it-IT":xae,"ja-JP":Sae,"ko-KR":_ae,"lt-LT":Eae,"lv-LV":Dae,"nb-NO":$ae,"nl-NL":Cae,"pl-PL":wae,"pt-BR":Rae,"pt-PT":Aae,"ro-RO":Tae,"ru-RU":kae,"sk-SK":Pae,"sl-SI":Mae,"sr-SP":Fae,"sv-SE":Nae,"tr-TR":Iae,"uk-UA":Lae,"zh-CN":Bae,"zh-TW":Qae};function g$e(e){return e&&e.__esModule?e.default:e}function v$e(e,t){let{inputRef:n,collectionRef:r,filter:i,disableAutoFocusFirst:a=!1,disableVirtualFocus:u=!1}=e,l=vi(),o=P.useRef(void 0),s=P.useRef(!1),c=P.useRef(null),d=P.useRef(null),h=ig()==="virtual"&&($1()||fS()),[m,y]=P.useState(!h&&!u),[v,S]=P.useState(!1);P.useEffect(()=>()=>clearTimeout(o.current),[]);let _=fi(H=>{!H.isTrusted&&m&&n.current&&Zu(hi(n.current))!==n.current&&n.current.focus();let M=H.target;H.isTrusted||!M||c.current===M.id||(clearTimeout(o.current),M!==r.current?s.current?(c.current=M.id,o.current=setTimeout(()=>{t.setFocusedNodeId(M.id)},500)):(c.current=M.id,t.setFocusedNodeId(M.id)):c.current&&!document.getElementById(c.current)&&(c.current=null,t.setFocusedNodeId(null)),s.current=!1)}),E=P.useCallback(H=>{if(H!=null){var M;(M=d.current)===null||M===void 0||M.removeEventListener("focusin",_),d.current=H,H.addEventListener("focusin",_),H.getAttribute("tabindex")!=null&&y(!1),S(!0)}else{var G;(G=d.current)===null||G===void 0||G.removeEventListener("focusin",_),S(!1)}},[_]),$=wu(P.useMemo(()=>yg(r,E),[r,E])),k=fi(()=>{var H;s.current=!0,(H=r.current)===null||H===void 0||H.dispatchEvent(new CustomEvent(E9,{cancelable:!0,bubbles:!0,detail:{focusStrategy:"first"}}))}),A=fi(H=>{var M;NN(Zu()),c.current=null,t.setFocusedNodeId(null);let G=new CustomEvent(ote,{cancelable:!0,bubbles:!0,detail:{clearFocusKey:H}});clearTimeout(o.current),s.current=!1,(M=r.current)===null||M===void 0||M.dispatchEvent(G)}),w=P.useRef("");Yd(n,"input",H=>{let{inputType:M}=H;w.current=M});let R=H=>{w.current==="insertText"&&!a?k():w.current&&(w.current.includes("insert")||w.current.includes("delete")||w.current.includes("history"))&&(A(!0),Yne(document)===n.current&&Sx(n.current,null)),t.setInputValue(H)},F=P.useRef(null),I=H=>{if(F.current=H.target,H.nativeEvent.isComposing)return;let M=c.current;switch(H.key){case"a":if(Tb(H))return;break;case"Escape":if(H.isDefaultPrevented())return;break;case" ":return;case"Tab":"continuePropagation"in H&&H.continuePropagation();return;case"Home":case"End":case"PageDown":case"PageUp":case"ArrowUp":case"ArrowDown":{var G;if((H.key==="Home"||H.key==="End")&&M==null&&H.shiftKey)return;H.preventDefault();let ne=new CustomEvent(E9,{cancelable:!0,bubbles:!0});(G=r.current)===null||G===void 0||G.dispatchEvent(ne);break}}H.isPropagationStopped()||H.stopPropagation();let X=!0;if(r.current!==null){var B;if(M==null)X=((B=r.current)===null||B===void 0?void 0:B.dispatchEvent(new KeyboardEvent(H.nativeEvent.type,H.nativeEvent)))||!1;else{let ne=document.getElementById(M);ne&&(X=(ne==null?void 0:ne.dispatchEvent(new KeyboardEvent(H.nativeEvent.type,H.nativeEvent)))||!1)}}if(X)switch(H.key){case"ArrowLeft":case"ArrowRight":A();break;case"Enter":if(M!=null){let ne=document.getElementById(M);ne==null||ne.click()}break}else H.preventDefault()},U=fi(H=>{if(H.target===F.current){var M;H.stopImmediatePropagation();let G=c.current;if(G==null)(M=r.current)===null||M===void 0||M.dispatchEvent(new KeyboardEvent(H.type,H));else{let X=document.getElementById(G);X==null||X.dispatchEvent(new KeyboardEvent(H.type,H))}}});P.useEffect(()=>(document.addEventListener("keyup",U,!0),()=>{document.removeEventListener("keyup",U,!0)}),[U]);let q=Qo(g$e(zae),"@react-aria/autocomplete"),V=Gv({id:l,"aria-label":q.format("collectionLabel")}),N=P.useCallback((H,M)=>i?i(H,t.inputValue,M):!0,[t.inputValue,i]),j=H=>{if(!H.isTrusted)return;let M=c.current?document.getElementById(c.current):null;M&&V9(M,H.relatedTarget)},W=H=>{if(!H.isTrusted)return;if(c.current?document.getElementById(c.current):null){let G=H.target;queueMicrotask(()=>{V9(G,r.current),Sx(r.current,G)})}},K={value:t.inputValue,onChange:R};var J;let te={onKeyDown:I,"aria-activedescendant":(J=t.focusedNodeId)!==null&&J!==void 0?J:void 0,onBlur:j,onFocus:W};return v&&(K={...K,...m&&te,enterKeyHint:"go","aria-controls":l,"aria-autocomplete":"list",autoCorrect:"off",spellCheck:"false",autoComplete:"off"}),{inputProps:K,collectionProps:Et(V,{shouldUseVirtualFocus:m,disallowTypeAhead:m}),collectionRef:$,filter:i!=null?N:void 0}}function y$e(e){let{onInputChange:t,inputValue:n,defaultInputValue:r=""}=e,i=s=>{t&&t(s)},[a,u]=P.useState(null),[l,o]=Yu(n,r,i);return{inputValue:l,setInputValue:o,focusedNodeId:a,setFocusedNodeId:u}}const b$e=P.createContext(null),O$e=P.createContext(null);function ait(e){let t=xs(b$e,e.slot);e=Et(t,e);let{filter:n,disableAutoFocusFirst:r}=e,i=y$e(e),a=P.useRef(null),u=P.useRef(null),{inputProps:l,collectionProps:o,collectionRef:s,filter:c}=v$e({...pc(e),filter:n,disableAutoFocusFirst:r,inputRef:a,collectionRef:u},i);return be.createElement(Ni,{values:[[O$e,i],[LC,{...l,ref:a}],[vx,{...o,filter:c,ref:s}]]},e.children)}class gy{get childNodes(){throw new Error("childNodes is not supported")}clone(){let t=new this.constructor(this.key);return t.value=this.value,t.level=this.level,t.hasChildNodes=this.hasChildNodes,t.rendered=this.rendered,t.textValue=this.textValue,t["aria-label"]=this["aria-label"],t.index=this.index,t.parentKey=this.parentKey,t.prevKey=this.prevKey,t.nextKey=this.nextKey,t.firstChildKey=this.firstChildKey,t.lastChildKey=this.lastChildKey,t.props=this.props,t.render=this.render,t.colSpan=this.colSpan,t.colIndex=this.colIndex,t}filter(t,n,r){let i=this.clone();return n.addDescendants(i,t),i}constructor(t){this.value=null,this.level=0,this.hasChildNodes=!1,this.rendered=null,this.textValue="",this["aria-label"]=void 0,this.index=0,this.parentKey=null,this.prevKey=null,this.nextKey=null,this.firstChildKey=null,this.lastChildKey=null,this.props={},this.colSpan=null,this.colIndex=null,this.type=this.constructor.type,this.key=t}}class qae extends gy{filter(t,n,r){let[i,a]=Uae(t,n,this.firstChildKey,r),u=this.clone();return u.firstChildKey=i,u.lastChildKey=a,u}}class x$e extends gy{}x$e.type="header";class Vae extends gy{}Vae.type="loader";class JC extends qae{filter(t,n,r){if(r(this.textValue,this)){let i=this.clone();return n.addDescendants(i,t),i}return null}}JC.type="item";class S$e extends qae{filter(t,n,r){let i=super.filter(t,n,r);if(i&&i.lastChildKey!==null){let a=t.getItem(i.lastChildKey);if(a&&a.type!=="header")return i}return null}}S$e.type="section";class _$e{get size(){return this.itemCount}getKeys(){return this.keyMap.keys()}*[Symbol.iterator](){let t=this.firstKey!=null?this.keyMap.get(this.firstKey):void 0;for(;t;)yield t,t=t.nextKey!=null?this.keyMap.get(t.nextKey):void 0}getChildren(t){let n=this.keyMap;return{*[Symbol.iterator](){let r=n.get(t),i=(r==null?void 0:r.firstChildKey)!=null?n.get(r.firstChildKey):null;for(;i;)yield i,i=i.nextKey!=null?n.get(i.nextKey):void 0}}}getKeyBefore(t){let n=this.keyMap.get(t);if(!n)return null;if(n.prevKey!=null){for(n=this.keyMap.get(n.prevKey);n&&n.type!=="item"&&n.lastChildKey!=null;)n=this.keyMap.get(n.lastChildKey);var r;return(r=n==null?void 0:n.key)!==null&&r!==void 0?r:null}return n.parentKey}getKeyAfter(t){let n=this.keyMap.get(t);if(!n)return null;if(n.type!=="item"&&n.firstChildKey!=null)return n.firstChildKey;for(;n;){if(n.nextKey!=null)return n.nextKey;if(n.parentKey!=null)n=this.keyMap.get(n.parentKey);else return null}return null}getFirstKey(){return this.firstKey}getLastKey(){let t=this.lastKey!=null?this.keyMap.get(this.lastKey):null;for(;(t==null?void 0:t.lastChildKey)!=null;)t=this.keyMap.get(t.lastChildKey);var n;return(n=t==null?void 0:t.key)!==null&&n!==void 0?n:null}getItem(t){var n;return(n=this.keyMap.get(t))!==null&&n!==void 0?n:null}at(){throw new Error("Not implemented")}clone(){let t=this.constructor,n=new t;return n.keyMap=new Map(this.keyMap),n.firstKey=this.firstKey,n.lastKey=this.lastKey,n.itemCount=this.itemCount,n}addNode(t){if(this.frozen)throw new Error("Cannot add a node to a frozen collection");t.type==="item"&&this.keyMap.get(t.key)==null&&this.itemCount++,this.keyMap.set(t.key,t)}addDescendants(t,n){this.addNode(t);let r=n.getChildren(t.key);for(let i of r)this.addDescendants(i,n)}removeNode(t){if(this.frozen)throw new Error("Cannot remove a node to a frozen collection");let n=this.keyMap.get(t);n!=null&&n.type==="item"&&this.itemCount--,this.keyMap.delete(t)}commit(t,n,r=!1){if(this.frozen)throw new Error("Cannot commit a frozen collection");this.firstKey=t,this.lastKey=n,this.frozen=!r}filter(t){let n=new this.constructor,[r,i]=Uae(this,n,this.firstKey,t);return n==null||n.commit(r,i),n}constructor(){this.keyMap=new Map,this.firstKey=null,this.lastKey=null,this.frozen=!1,this.itemCount=0}}function Uae(e,t,n,r){if(n==null)return[null,null];let i=null,a=null,u=e.getItem(n);for(;u!=null;){let s=u.filter(e,t,r);s!=null&&(s.nextKey=null,a&&(s.prevKey=a.key,a.nextKey=s.key),i==null&&(i=s),t.addNode(s),a=s),u=u.nextKey?e.getItem(u.nextKey):null}if(a&&a.type==="separator"){let s=a.prevKey;t.removeNode(a.key),s?(a=t.getItem(s),a.nextKey=null):a=null}var l,o;return[(l=i==null?void 0:i.key)!==null&&l!==void 0?l:null,(o=a==null?void 0:a.key)!==null&&o!==void 0?o:null]}class jae{*[Symbol.iterator](){let t=this.firstChild;for(;t;)yield t,t=t.nextSibling}get firstChild(){return this._firstChild}set firstChild(t){this._firstChild=t,this.ownerDocument.markDirty(this)}get lastChild(){return this._lastChild}set lastChild(t){this._lastChild=t,this.ownerDocument.markDirty(this)}get previousSibling(){return this._previousSibling}set previousSibling(t){this._previousSibling=t,this.ownerDocument.markDirty(this)}get nextSibling(){return this._nextSibling}set nextSibling(t){this._nextSibling=t,this.ownerDocument.markDirty(this)}get parentNode(){return this._parentNode}set parentNode(t){this._parentNode=t,this.ownerDocument.markDirty(this)}get isConnected(){var t;return((t=this.parentNode)===null||t===void 0?void 0:t.isConnected)||!1}invalidateChildIndices(t){(this._minInvalidChildIndex==null||!this._minInvalidChildIndex.isConnected||t.index<this._minInvalidChildIndex.index)&&(this._minInvalidChildIndex=t,this.ownerDocument.markDirty(this))}updateChildIndices(){let t=this._minInvalidChildIndex;for(;t;)t.index=t.previousSibling?t.previousSibling.index+1:0,t=t.nextSibling;this._minInvalidChildIndex=null}appendChild(t){t.parentNode&&t.parentNode.removeChild(t),this.firstChild==null&&(this.firstChild=t),this.lastChild?(this.lastChild.nextSibling=t,t.index=this.lastChild.index+1,t.previousSibling=this.lastChild):(t.previousSibling=null,t.index=0),t.parentNode=this,t.nextSibling=null,this.lastChild=t,this.ownerDocument.markDirty(this),this.isConnected&&this.ownerDocument.queueUpdate()}insertBefore(t,n){if(n==null)return this.appendChild(t);t.parentNode&&t.parentNode.removeChild(t),t.nextSibling=n,t.previousSibling=n.previousSibling,t.index=n.index-1,this.firstChild===n?this.firstChild=t:n.previousSibling&&(n.previousSibling.nextSibling=t),n.previousSibling=t,t.parentNode=n.parentNode,this.invalidateChildIndices(t),this.isConnected&&this.ownerDocument.queueUpdate()}removeChild(t){t.parentNode===this&&(this._minInvalidChildIndex===t&&(this._minInvalidChildIndex=null),t.nextSibling&&(this.invalidateChildIndices(t.nextSibling),t.nextSibling.previousSibling=t.previousSibling),t.previousSibling&&(t.previousSibling.nextSibling=t.nextSibling),this.firstChild===t&&(this.firstChild=t.nextSibling),this.lastChild===t&&(this.lastChild=t.previousSibling),t.parentNode=null,t.nextSibling=null,t.previousSibling=null,t.index=0,this.ownerDocument.markDirty(t),this.isConnected&&this.ownerDocument.queueUpdate())}addEventListener(){}removeEventListener(){}get previousVisibleSibling(){let t=this.previousSibling;for(;t&&t.isHidden;)t=t.previousSibling;return t}get nextVisibleSibling(){let t=this.nextSibling;for(;t&&t.isHidden;)t=t.nextSibling;return t}get firstVisibleChild(){let t=this.firstChild;for(;t&&t.isHidden;)t=t.nextSibling;return t}get lastVisibleChild(){let t=this.lastChild;for(;t&&t.isHidden;)t=t.previousSibling;return t}constructor(t){this._firstChild=null,this._lastChild=null,this._previousSibling=null,this._nextSibling=null,this._parentNode=null,this._minInvalidChildIndex=null,this.ownerDocument=t}}class c3 extends jae{get index(){return this._index}set index(t){this._index=t,this.ownerDocument.markDirty(this)}get level(){var t;return this.parentNode instanceof c3?this.parentNode.level+(((t=this.node)===null||t===void 0?void 0:t.type)==="item"?1:0):0}getMutableNode(){return this.node==null?null:(this.isMutated||(this.node=this.node.clone(),this.isMutated=!0),this.ownerDocument.markDirty(this),this.node)}updateNode(){var t,n,r,i,a,u,l,o;let s=this.nextVisibleSibling,c=this.getMutableNode();if(c!=null){c.index=this.index,c.level=this.level;var d;c.parentKey=this.parentNode instanceof c3&&(d=(t=this.parentNode.node)===null||t===void 0?void 0:t.key)!==null&&d!==void 0?d:null;var h;c.prevKey=(h=(r=this.previousVisibleSibling)===null||r===void 0||(n=r.node)===null||n===void 0?void 0:n.key)!==null&&h!==void 0?h:null;var m;c.nextKey=(m=s==null||(i=s.node)===null||i===void 0?void 0:i.key)!==null&&m!==void 0?m:null,c.hasChildNodes=!!this.firstChild;var y;c.firstChildKey=(y=(u=this.firstVisibleChild)===null||u===void 0||(a=u.node)===null||a===void 0?void 0:a.key)!==null&&y!==void 0?y:null;var v;if(c.lastChildKey=(v=(o=this.lastVisibleChild)===null||o===void 0||(l=o.node)===null||l===void 0?void 0:l.key)!==null&&v!==void 0?v:null,(c.colSpan!=null||c.colIndex!=null)&&s){var S,_;let E=((S=c.colIndex)!==null&&S!==void 0?S:c.index)+((_=c.colSpan)!==null&&_!==void 0?_:1);if(s.node!=null&&E!==s.node.colIndex){let $=s.getMutableNode();$.colIndex=E}}}}setProps(t,n,r,i,a){let u,{value:l,textValue:o,id:s,...c}=t;if(this.node==null?(u=new r(s??`react-aria-${++this.ownerDocument.nodeId}`),this.node=u):u=this.getMutableNode(),c.ref=n,u.props=c,u.rendered=i,u.render=a,u.value=l,u.textValue=o||(typeof c.children=="string"?c.children:"")||t["aria-label"]||"",s!=null&&s!==u.key)throw new Error("Cannot change the id of an item");c.colSpan!=null&&(u.colSpan=c.colSpan),this.isConnected&&this.ownerDocument.queueUpdate()}get style(){let t=this;return{get display(){return t.isHidden?"none":""},set display(n){let r=n==="none";if(t.isHidden!==r){var i,a;(((i=t.parentNode)===null||i===void 0?void 0:i.firstVisibleChild)===t||((a=t.parentNode)===null||a===void 0?void 0:a.lastVisibleChild)===t)&&t.ownerDocument.markDirty(t.parentNode);let u=t.previousVisibleSibling,l=t.nextVisibleSibling;u&&t.ownerDocument.markDirty(u),l&&t.ownerDocument.markDirty(l),t.isHidden=r,t.ownerDocument.markDirty(t)}}}}hasAttribute(){}setAttribute(){}setAttributeNS(){}removeAttribute(){}constructor(t,n){super(n),this.nodeType=8,this.isMutated=!0,this._index=0,this.isHidden=!1,this.node=null}}class E$e extends jae{get isConnected(){return!0}createElement(t){return new c3(t,this)}getMutableCollection(){return this.nextCollection||(this.nextCollection=this.collection.clone()),this.nextCollection}markDirty(t){this.dirtyNodes.add(t)}addNode(t){if(t.isHidden||t.node==null)return;let n=this.getMutableCollection();if(!n.getItem(t.node.key))for(let r of t)this.addNode(r);n.addNode(t.node)}removeNode(t){for(let n of t)this.removeNode(n);t.node&&this.getMutableCollection().removeNode(t.node.key)}getCollection(){return this.inSubscription?this.collection.clone():(this.queuedRender=!1,this.updateCollection(),this.collection)}updateCollection(){for(let l of this.dirtyNodes)l instanceof c3&&(!l.isConnected||l.isHidden)?this.removeNode(l):l.updateChildIndices();for(let l of this.dirtyNodes)l instanceof c3?(l.isConnected&&!l.isHidden&&(l.updateNode(),this.addNode(l)),l.node&&this.dirtyNodes.delete(l),l.isMutated=!1):this.dirtyNodes.delete(l);if(this.nextCollection){var t,n,r,i,a,u;this.nextCollection.commit((a=(n=this.firstVisibleChild)===null||n===void 0||(t=n.node)===null||t===void 0?void 0:t.key)!==null&&a!==void 0?a:null,(u=(i=this.lastVisibleChild)===null||i===void 0||(r=i.node)===null||r===void 0?void 0:r.key)!==null&&u!==void 0?u:null,this.isSSR),this.isSSR||(this.collection=this.nextCollection,this.nextCollection=null)}}queueUpdate(){if(!(this.dirtyNodes.size===0||this.queuedRender)){this.queuedRender=!0,this.inSubscription=!0;for(let t of this.subscriptions)t();this.inSubscription=!1}}subscribe(t){return this.subscriptions.add(t),()=>this.subscriptions.delete(t)}resetAfterSSR(){this.isSSR&&(this.isSSR=!1,this.firstChild=null,this.lastChild=null,this.nodeId=0)}constructor(t){super(null),this.nodeType=11,this.ownerDocument=this,this.dirtyNodes=new Set,this.isSSR=!1,this.nodeId=0,this.nodesByProps=new WeakMap,this.nextCollection=null,this.subscriptions=new Set,this.queuedRender=!1,this.inSubscription=!1,this.collection=t,this.nextCollection=t}}function Zae(e){let{children:t,items:n,idScope:r,addIdAndValue:i,dependencies:a=[]}=e,u=P.useMemo(()=>new WeakMap,a);return P.useMemo(()=>{if(n&&typeof t=="function"){let s=[];for(let c of n){let d=u.get(c);if(!d){d=t(c);var l,o;let h=(o=(l=d.props.id)!==null&&l!==void 0?l:c.key)!==null&&o!==void 0?o:c.id;if(h==null)throw new Error("Could not determine key for item");r&&(h=r+":"+h),d=P.cloneElement(d,i?{key:h,id:h,value:c}:{key:h}),u.set(c,d)}s.push(d)}return s}else if(typeof t!="function")return t},[t,n,u,r,i])}if(typeof HTMLTemplateElement<"u"){const e=Object.getOwnPropertyDescriptor(Node.prototype,"firstChild").get;Object.defineProperty(HTMLTemplateElement.prototype,"firstChild",{configurable:!0,enumerable:!0,get:function(){return this.dataset.reactAriaHidden?this.content.firstChild:e.call(this)}})}const l$=P.createContext(!1);function D$e(e){if(P.useContext(l$))return be.createElement(be.Fragment,null,e.children);let n=be.createElement(l$.Provider,{value:!0},e.children);return be.createElement("template",{"data-react-aria-hidden":!0},n)}function C1(e){let t=(n,r)=>P.useContext(l$)?null:e(n,r);return t.displayName=e.displayName||e.name,P.forwardRef(t)}function $$e(){return P.useContext(l$)}var UN=xee();const Hae=P.createContext(!1),_x=P.createContext(null);function w1(e){if(P.useContext(_x))return e.content;let{collection:n,document:r}=A$e(e.createCollection);return be.createElement(be.Fragment,null,be.createElement(D$e,null,be.createElement(_x.Provider,{value:r},e.content)),be.createElement(C$e,{render:e.children,collection:n}))}function C$e({collection:e,render:t}){return t(e)}function w$e(e,t,n){let r=pp(),i=P.useRef(r);i.current=r;let a=P.useCallback(()=>i.current?n():t(),[t,n]);return UN.useSyncExternalStore(e,a)}const R$e=typeof be.useSyncExternalStore=="function"?be.useSyncExternalStore:w$e;function A$e(e){let[t]=P.useState(()=>new E$e((e==null?void 0:e())||new _$e)),n=P.useCallback(u=>t.subscribe(u),[t]),r=P.useCallback(()=>{let u=t.getCollection();return t.isSSR&&t.resetAfterSSR(),u},[t]),i=P.useCallback(()=>(t.isSSR=!0,t.getCollection()),[t]);return{collection:R$e(n,r,i),document:t}}const Y9=P.createContext(null);function T$e(e){var t;return t=class extends gy{},t.type=e,t}function Xae(e,t,n,r,i,a){typeof e=="string"&&(e=T$e(e));let u=P.useCallback(o=>{o==null||o.setProps(t,n,e,r,a)},[t,n,r,a,e]),l=P.useContext(Y9);if(l){let o=l.ownerDocument.nodesByProps.get(t);return o||(o=l.ownerDocument.createElement(e.type),o.setProps(t,n,e,r,a),l.appendChild(o),l.ownerDocument.updateCollection(),l.ownerDocument.nodesByProps.set(t,o)),i?be.createElement(Y9.Provider,{value:o},i):null}return be.createElement(e.type,{ref:u},i)}function R1(e,t){let n=({node:i})=>t(i.props,i.props.ref,i),r=P.forwardRef((i,a)=>{let u=P.useContext(a$);if(!P.useContext(Hae)){if(t.length>=3)throw new Error(t.name+" cannot be rendered outside a collection.");return t(i,a)}return Xae(e,i,a,"children"in i?i.children:null,null,o=>be.createElement(a$.Provider,{value:u},be.createElement(n,{node:o})))});return r.displayName=t.name,r}function k$e(e,t,n=Yae){let r=({node:a})=>t(a.props,a.props.ref,a),i=P.forwardRef((a,u)=>{let l=n(a);var o;return(o=Xae(e,a,u,null,l,s=>be.createElement(r,{node:s})))!==null&&o!==void 0?o:be.createElement(be.Fragment,null)});return i.displayName=t.name,i}function Yae(e){return Zae({...e,addIdAndValue:!0})}const wj=P.createContext(null);function OS(e){let t=P.useContext(wj),n=((t==null?void 0:t.dependencies)||[]).concat(e.dependencies),r=e.idScope||(t==null?void 0:t.idScope),i=Yae({...e,idScope:r,dependencies:n});return P.useContext(_x)&&(i=be.createElement(P$e,null,i)),t=P.useMemo(()=>({dependencies:n,idScope:r}),[r,...n]),be.createElement(wj.Provider,{value:t},i)}function P$e({children:e}){let t=P.useContext(_x),n=P.useMemo(()=>be.createElement(_x.Provider,{value:null},be.createElement(Hae.Provider,{value:!0},e)),[e]);return pp()?be.createElement(Y9.Provider,{value:t},n):Uf.createPortal(n,t)}const Gae=P.createContext(null),jN={CollectionRoot({collection:e,renderDropIndicator:t}){return Rj(e,null,t)},CollectionBranch({collection:e,parent:t,renderDropIndicator:n}){return Rj(e,t,n)}};function Rj(e,t,n){return Zae({items:t?e.getChildren(t.key):e,dependencies:[n],children(r){let i=r.render(r);return!n||r.type!=="item"?i:be.createElement(be.Fragment,null,n({type:"item",key:r.key,dropPosition:"before"}),i,M$e(e,r,n))}})}function M$e(e,t,n){let r=t.key,i=e.getKeyAfter(r),a=i!=null?e.getItem(i):null;for(;a!=null&&a.type!=="item";)i=e.getKeyAfter(a.key),a=i!=null?e.getItem(i):null;let u=t.nextKey!=null?e.getItem(t.nextKey):null;for(;u!=null&&u.type!=="item";)u=u.nextKey!=null?e.getItem(u.nextKey):null;let l=[];if(u==null){let o=t;for(;o&&(!a||o.parentKey!==a.parentKey&&a.level<o.level);){let s=n({type:"item",key:o.key,dropPosition:"after"});P.isValidElement(s)&&l.push(P.cloneElement(s,{key:`${o.key}-after`})),o=o.parentKey!=null?e.getItem(o.parentKey):null}}return l}const mp=P.createContext(jN);function ZN(e){return P.useMemo(()=>e!=null?new Set([e]):null,[e])}var Kae={};Kae={breadcrumbs:"عناصر الواجهة"};var Wae={};Wae={breadcrumbs:"Трохи хляб"};var Jae={};Jae={breadcrumbs:"Popis cesty"};var eue={};eue={breadcrumbs:"Brødkrummer"};var tue={};tue={breadcrumbs:"Breadcrumbs"};var nue={};nue={breadcrumbs:"Πλοηγήσεις breadcrumb"};var rue={};rue={breadcrumbs:"Breadcrumbs"};var iue={};iue={breadcrumbs:"Migas de pan"};var aue={};aue={breadcrumbs:"Lingiread"};var uue={};uue={breadcrumbs:"Navigointilinkit"};var lue={};lue={breadcrumbs:"Chemin de navigation"};var oue={};oue={breadcrumbs:"שבילי ניווט"};var sue={};sue={breadcrumbs:"Navigacijski putovi"};var cue={};cue={breadcrumbs:"Morzsamenü"};var fue={};fue={breadcrumbs:"Breadcrumb"};var due={};due={breadcrumbs:"パンくずリスト"};var hue={};hue={breadcrumbs:"탐색 표시"};var pue={};pue={breadcrumbs:"Naršymo kelias"};var mue={};mue={breadcrumbs:"Atpakaļceļi"};var gue={};gue={breadcrumbs:"Navigasjonsstier"};var vue={};vue={breadcrumbs:"Broodkruimels"};var yue={};yue={breadcrumbs:"Struktura nawigacyjna"};var bue={};bue={breadcrumbs:"Caminho detalhado"};var Oue={};Oue={breadcrumbs:"Categorias"};var xue={};xue={breadcrumbs:"Miez de pâine"};var Sue={};Sue={breadcrumbs:"Навигация"};var _ue={};_ue={breadcrumbs:"Navigačné prvky Breadcrumbs"};var Eue={};Eue={breadcrumbs:"Drobtine"};var Due={};Due={breadcrumbs:"Putanje navigacije"};var $ue={};$ue={breadcrumbs:"Sökvägar"};var Cue={};Cue={breadcrumbs:"İçerik haritaları"};var wue={};wue={breadcrumbs:"Навігаційна стежка"};var Rue={};Rue={breadcrumbs:"导航栏"};var Aue={};Aue={breadcrumbs:"導覽列"};var Tue={};Tue={"ar-AE":Kae,"bg-BG":Wae,"cs-CZ":Jae,"da-DK":eue,"de-DE":tue,"el-GR":nue,"en-US":rue,"es-ES":iue,"et-EE":aue,"fi-FI":uue,"fr-FR":lue,"he-IL":oue,"hr-HR":sue,"hu-HU":cue,"it-IT":fue,"ja-JP":due,"ko-KR":hue,"lt-LT":pue,"lv-LV":mue,"nb-NO":gue,"nl-NL":vue,"pl-PL":yue,"pt-BR":bue,"pt-PT":Oue,"ro-RO":xue,"ru-RU":Sue,"sk-SK":_ue,"sl-SI":Eue,"sr-SP":Due,"sv-SE":$ue,"tr-TR":Cue,"uk-UA":wue,"zh-CN":Rue,"zh-TW":Aue};function F$e(e){return e&&e.__esModule?e.default:e}function N$e(e){let{"aria-label":t,...n}=e,r=Qo(F$e(Tue),"@react-aria/breadcrumbs");return{navProps:{...Gt(n,{labelable:!0}),"aria-label":t||r.format("breadcrumbs")}}}function kue(e,t){let{elementType:n="button",isDisabled:r,onPress:i,onPressStart:a,onPressEnd:u,onPressUp:l,onPressChange:o,preventFocusOnPress:s,allowFocusWhenDisabled:c,onClick:d,href:h,target:m,rel:y,type:v="button"}=e,S;n==="button"?S={type:v,disabled:r,form:e.form,formAction:e.formAction,formEncType:e.formEncType,formMethod:e.formMethod,formNoValidate:e.formNoValidate,formTarget:e.formTarget,name:e.name,value:e.value}:S={role:"button",href:n==="a"&&!r?h:void 0,target:n==="a"?m:void 0,type:n==="input"?v:void 0,disabled:n==="input"?r:void 0,"aria-disabled":!r||n==="input"?void 0:r,rel:n==="a"?y:void 0};let{pressProps:_,isPressed:E}=lp({onPressStart:a,onPressEnd:u,onPressChange:o,onPress:i,onPressUp:l,onClick:d,isDisabled:r,preventFocusOnPress:s,ref:t}),{focusableProps:$}=I0(e,t);c&&($.tabIndex=r?-1:$.tabIndex);let k=Et($,_,Gt(e,{labelable:!0}));return{isPressed:E,buttonProps:Et(S,k,{"aria-haspopup":e["aria-haspopup"],"aria-expanded":e["aria-expanded"],"aria-controls":e["aria-controls"],"aria-pressed":e["aria-pressed"],"aria-current":e["aria-current"],"aria-disabled":e["aria-disabled"]})}}function Pue(e,t,n){const{isSelected:r}=t,{isPressed:i,buttonProps:a}=kue({...e,onPress:ih(t.toggle,e.onPress)},n);return{isPressed:i,isSelected:r,isDisabled:e.isDisabled||!1,buttonProps:Et(a,{"aria-pressed":r})}}function Mue(e,t){const{"aria-label":n,"aria-labelledby":r,orientation:i="horizontal"}=e;let[a,u]=P.useState(!1);Hr(()=>{var y;u(!!(t.current&&(!((y=t.current.parentElement)===null||y===void 0)&&y.closest('[role="toolbar"]'))))});const{direction:l}=Wi(),o=l==="rtl"&&i==="horizontal";let s=MN(t);const c=y=>{if(y.currentTarget.contains(y.target)){if(i==="horizontal"&&y.key==="ArrowRight"||i==="vertical"&&y.key==="ArrowDown")o?s.focusPrevious():s.focusNext();else if(i==="horizontal"&&y.key==="ArrowLeft"||i==="vertical"&&y.key==="ArrowUp")o?s.focusNext():s.focusPrevious();else if(y.key==="Tab"){y.stopPropagation(),d.current=document.activeElement,y.shiftKey?s.focusFirst():s.focusLast();return}else return;y.stopPropagation(),y.preventDefault()}},d=P.useRef(null),h=y=>{!y.currentTarget.contains(y.relatedTarget)&&!d.current&&(d.current=y.target)},m=y=>{var v;if(d.current&&!y.currentTarget.contains(y.relatedTarget)&&(!((v=t.current)===null||v===void 0)&&v.contains(y.target))){var S;(S=d.current)===null||S===void 0||S.focus(),d.current=null}};return{toolbarProps:{...Gt(e,{labelable:!0}),role:a?"group":"toolbar","aria-orientation":i,"aria-label":n,"aria-labelledby":n==null?r:void 0,onKeyDownCapture:a?void 0:c,onFocusCapture:a?void 0:m,onBlurCapture:a?void 0:h}}}function I$e(e,t,n){let{isDisabled:r}=e,{toolbarProps:i}=Mue(e,n);return{groupProps:{...i,role:t.selectionMode==="single"?"radiogroup":i.role,"aria-disabled":r}}}function L$e(e,t,n){let r={isSelected:t.selectedKeys.has(e.id),defaultSelected:!1,setSelected(o){t.setSelected(e.id,o)},toggle(){t.toggleKey(e.id)}},{isPressed:i,isSelected:a,isDisabled:u,buttonProps:l}=Pue({...e,id:void 0,isDisabled:e.isDisabled||t.isDisabled},r,n);return t.selectionMode==="single"&&(l.role="radio",l["aria-checked"]=r.isSelected,delete l["aria-pressed"]),{isPressed:i,isSelected:a,isDisabled:u,buttonProps:l}}function Fue(e,t,n){let{isDisabled:r=!1,isReadOnly:i=!1,value:a,name:u,form:l,children:o,"aria-label":s,"aria-labelledby":c,validationState:d="valid",isInvalid:h,onPressStart:m,onPressEnd:y,onPressChange:v,onPress:S,onPressUp:_,onClick:E}=e,$=q=>{q.stopPropagation(),t.setSelected(q.target.checked)},{pressProps:k,isPressed:A}=lp({onPressStart:m,onPressEnd:y,onPressChange:v,onPress:S,onPressUp:_,onClick:E,isDisabled:r}),{pressProps:w,isPressed:R}=lp({onPressStart:m,onPressEnd:y,onPressChange:v,onPressUp:_,onClick:E,onPress(q){var V;S==null||S(q),t.toggle(),(V=n.current)===null||V===void 0||V.focus()},isDisabled:r||i}),{focusableProps:F}=I0(e,n),I=Et(k,F),U=Gt(e,{labelable:!0});return my(n,t.defaultSelected,t.setSelected),{labelProps:Et(w,{onClick:q=>q.preventDefault()}),inputProps:Et(U,{"aria-invalid":h||d==="invalid"||void 0,"aria-errormessage":e["aria-errormessage"],"aria-controls":e["aria-controls"],"aria-readonly":i||void 0,onChange:$,disabled:r,...a==null?{}:{value:a},name:u,form:l,type:"checkbox",...I}),isSelected:t.isSelected,isPressed:A||R,isDisabled:r,isReadOnly:i,isInvalid:h||d==="invalid"}}function Nue(e,t,n){let r=Og({...e,value:t.isSelected}),{isInvalid:i,validationErrors:a,validationDetails:u}=r.displayValidation,{labelProps:l,inputProps:o,isSelected:s,isPressed:c,isDisabled:d,isReadOnly:h}=Fue({...e,isInvalid:i},t,n);bS(e,r,n);let{isIndeterminate:m,isRequired:y,validationBehavior:v="aria"}=e;P.useEffect(()=>{n.current&&(n.current.indeterminate=!!m)});let{pressProps:S}=lp({isDisabled:d||h,onPress(){let{[c1]:_}=e,{commitValidation:E}=_||r;E()}});return{labelProps:Et(l,S,P.useMemo(()=>({onMouseDown:_=>_.preventDefault()}),[])),inputProps:{...o,checked:s,"aria-required":y&&v==="aria"||void 0,required:y&&v==="native"},isSelected:s,isPressed:c,isDisabled:d,isReadOnly:h,isInvalid:i,validationErrors:a,validationDetails:u}}const B$e=new WeakMap;function ew(e={}){let{isReadOnly:t}=e,[n,r]=Yu(e.isSelected,e.defaultSelected||!1,e.onChange),[i]=P.useState(n);function a(o){t||r(o)}function u(){t||r(!n)}var l;return{isSelected:n,defaultSelected:(l=e.defaultSelected)!==null&&l!==void 0?l:i,setSelected:a,toggle:u}}function Q$e(e){let{selectionMode:t="single",disallowEmptySelection:n,isDisabled:r=!1}=e,[i,a]=Yu(P.useMemo(()=>e.selectedKeys?new Set(e.selectedKeys):void 0,[e.selectedKeys]),P.useMemo(()=>e.defaultSelectedKeys?new Set(e.defaultSelectedKeys):new Set,[e.defaultSelectedKeys]),e.onSelectionChange);return{selectionMode:t,isDisabled:r,selectedKeys:i,setSelectedKeys:a,toggleKey(u){let l;t==="multiple"?(l=new Set(i),l.has(u)&&(!n||l.size>1)?l.delete(u):l.add(u)):l=new Set(i.has(u)&&!n?[]:[u]),a(l)},setSelected(u,l){l!==i.has(u)&&this.toggleKey(u)}}}function z$e(e,t,n){const r=ew({isReadOnly:e.isReadOnly||t.isReadOnly,isSelected:t.isSelected(e.value),defaultSelected:t.defaultValue.includes(e.value),onChange(_){_?t.addValue(e.value):t.removeValue(e.value),e.onChange&&e.onChange(_)}});let{name:i,form:a,descriptionId:u,errorMessageId:l,validationBehavior:o}=B$e.get(t);var s;o=(s=e.validationBehavior)!==null&&s!==void 0?s:o;let{realtimeValidation:c}=Og({...e,value:r.isSelected,name:void 0,validationBehavior:"aria"}),d=P.useRef(IO),h=()=>{t.setInvalid(e.value,c.isInvalid?c:d.current)};P.useEffect(h);let m=t.realtimeValidation.isInvalid?t.realtimeValidation:c,y=o==="native"?t.displayValidation:m;var v;let S=Nue({...e,isReadOnly:e.isReadOnly||t.isReadOnly,isDisabled:e.isDisabled||t.isDisabled,name:e.name||i,form:e.form||a,isRequired:(v=e.isRequired)!==null&&v!==void 0?v:t.isRequired,validationBehavior:o,[c1]:{realtimeValidation:m,displayValidation:y,resetValidation:t.resetValidation,commitValidation:t.commitValidation,updateValidation(_){d.current=_,h()}}},r,n);return{...S,inputProps:{...S.inputProps,"aria-describedby":[e["aria-describedby"],t.isInvalid?l:null,u].filter(Boolean).join(" ")||void 0}}}var Iue={};Iue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"أداة انتقاء اللون",colorSwatch:"تغيير الألوان",transparent:"شفاف",twoDimensionalSlider:"مُنزلق 2D"};var Lue={};Lue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Средство за избиране на цвят",colorSwatch:"цветна мостра",transparent:"прозрачен",twoDimensionalSlider:"2D плъзгач"};var Bue={};Bue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Výběr barvy",colorSwatch:"barevný vzorek",transparent:"průhledný",twoDimensionalSlider:"2D posuvník"};var Que={};Que={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Farvevælger",colorSwatch:"farveprøve",transparent:"gennemsigtig",twoDimensionalSlider:"2D-skyder"};var zue={};zue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Farbwähler",colorSwatch:"Farbfeld",transparent:"transparent",twoDimensionalSlider:"2D-Schieberegler"};var que={};que={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Επιλογέας χρωμάτων",colorSwatch:"χρωματικό δείγμα",transparent:"διαφανές",twoDimensionalSlider:"Ρυθμιστικό 2D"};var Vue={};Vue={colorPicker:"Color picker",twoDimensionalSlider:"2D slider",colorNameAndValue:e=>`${e.name}: ${e.value}`,colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorSwatch:"color swatch",transparent:"transparent"};var Uue={};Uue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Selector de color",colorSwatch:"muestra de color",transparent:"transparente",twoDimensionalSlider:"Regulador 2D"};var jue={};jue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Värvivalija",colorSwatch:"värvinäidis",transparent:"läbipaistev",twoDimensionalSlider:"2D-liugur"};var Zue={};Zue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Värimuokkain",colorSwatch:"värimalli",transparent:"läpinäkyvä",twoDimensionalSlider:"2D-liukusäädin"};var Hue={};Hue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name} : ${e.value}`,colorPicker:"Sélecteur de couleurs",colorSwatch:"Échantillon de couleurs",transparent:"Transparent",twoDimensionalSlider:"Curseur 2D"};var Xue={};Xue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"בוחר הצבעים",colorSwatch:"דוגמית צבע",transparent:"שקוף",twoDimensionalSlider:"מחוון דו מימדי"};var Yue={};Yue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Odabir boje",colorSwatch:"uzorak boje",transparent:"transparentno",twoDimensionalSlider:"2D klizač"};var Gue={};Gue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Színválasztó",colorSwatch:"színtár",transparent:"átlátszó",twoDimensionalSlider:"2D-csúszka"};var Kue={};Kue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Selettore colore",colorSwatch:"campione di colore",transparent:"trasparente",twoDimensionalSlider:"Cursore 2D"};var Wue={};Wue={colorInputLabel:e=>`${e.label}、${e.channelLabel}`,colorNameAndValue:e=>`${e.name} : ${e.value}`,colorPicker:"カラーピッカー",colorSwatch:"カラースウォッチ",transparent:"透明",twoDimensionalSlider:"2D スライダー"};var Jue={};Jue={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"색상 피커",colorSwatch:"색상 견본",transparent:"투명도",twoDimensionalSlider:"2D 슬라이더"};var ele={};ele={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Spalvų parinkiklis",colorSwatch:"spalvų pavyzdys",transparent:"skaidrus",twoDimensionalSlider:"2D slankiklis"};var tle={};tle={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Krāsu atlasītājs",colorSwatch:"krāsu paraugs",transparent:"caurspīdīgs",twoDimensionalSlider:"2D slīdnis"};var nle={};nle={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Fargevelger",colorSwatch:"fargekart",transparent:"gjennomsiktig",twoDimensionalSlider:"2D-glidebryter"};var rle={};rle={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Kleurkiezer",colorSwatch:"kleurstaal",transparent:"transparant",twoDimensionalSlider:"2D-schuifregelaar"};var ile={};ile={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Próbnik kolorów",colorSwatch:"próbka koloru",transparent:"przezroczysty",twoDimensionalSlider:"Suwak 2D"};var ale={};ale={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Seletor de cores",colorSwatch:"amostra de cores",transparent:"transparente",twoDimensionalSlider:"Controle deslizante 2D"};var ule={};ule={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Seletor de cores",colorSwatch:"amostra de cor",transparent:"transparente",twoDimensionalSlider:"Controle deslizante 2D"};var lle={};lle={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Selector de culori",colorSwatch:"specimen de culoare",transparent:"transparent",twoDimensionalSlider:"Glisor 2D"};var ole={};ole={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Палитра цветов",colorSwatch:"цветовой образец",transparent:"прозрачный",twoDimensionalSlider:"Ползунок 2D"};var sle={};sle={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Výber farieb",colorSwatch:"vzorkovník farieb",transparent:"transparentný",twoDimensionalSlider:"2D jazdec"};var cle={};cle={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Izbirnik barv",colorSwatch:"barvna paleta",transparent:"prozorno",twoDimensionalSlider:"2D drsnik"};var fle={};fle={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Birač boja",colorSwatch:"Uzorak boje",transparent:"providno",twoDimensionalSlider:"2D klizač"};var dle={};dle={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Färgväljaren",colorSwatch:"färgruta",transparent:"genomskinlig",twoDimensionalSlider:"2D-reglage"};var hle={};hle={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Renk Seçici",colorSwatch:"renk örneği",transparent:"saydam",twoDimensionalSlider:"2D sürgü"};var ple={};ple={colorInputLabel:e=>`${e.label}, ${e.channelLabel}`,colorNameAndValue:e=>`${e.name}: ${e.value}`,colorPicker:"Палітра кольорів",colorSwatch:"зразок кольору",transparent:"прозорий",twoDimensionalSlider:"Повзунок 2D"};var mle={};mle={colorInputLabel:e=>`${e.label}、${e.channelLabel}`,colorNameAndValue:e=>`${e.name}：${e.value}`,colorPicker:"拾色器",colorSwatch:"颜色色板",transparent:"透明",twoDimensionalSlider:"2D 滑块"};var gle={};gle={colorInputLabel:e=>`${e.label}，${e.channelLabel}`,colorNameAndValue:e=>`${e.name}：${e.value}`,colorPicker:"檢色器",colorSwatch:"色票",transparent:"透明",twoDimensionalSlider:"2D 滑桿"};var vle={};vle={"ar-AE":Iue,"bg-BG":Lue,"cs-CZ":Bue,"da-DK":Que,"de-DE":zue,"el-GR":que,"en-US":Vue,"es-ES":Uue,"et-EE":jue,"fi-FI":Zue,"fr-FR":Hue,"he-IL":Xue,"hr-HR":Yue,"hu-HU":Gue,"it-IT":Kue,"ja-JP":Wue,"ko-KR":Jue,"lt-LT":ele,"lv-LV":tle,"nb-NO":nle,"nl-NL":rle,"pl-PL":ile,"pt-BR":ale,"pt-PT":ule,"ro-RO":lle,"ru-RU":ole,"sk-SK":sle,"sl-SI":cle,"sr-SP":fle,"sv-SE":dle,"tr-TR":hle,"uk-UA":ple,"zh-CN":mle,"zh-TW":gle};var yle={};yle={alpha:"ألفا",black:"أسود",blue:"أزرق","blue purple":"أرجواني مزرق",brightness:"السطوع",brown:"بني","brown yellow":"أصفر بني",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"سماوي","cyan blue":"أزرق سماوي",dark:"داكن",gray:"رمادي",grayish:"مائل للرمادي",green:"أخضر","green cyan":"سماوي مخضر",hue:"درجة اللون",light:"فاتح",lightness:"الإضاءة",magenta:"أرجواني","magenta pink":"أرجواني وردي",orange:"برتقالي","orange yellow":"أصفر برتقالي",pale:"باهت",pink:"وردي","pink red":"أحمر وردي",purple:"أرجواني","purple magenta":"بنفسجي",red:"أحمر","red orange":"برتقالي محمر",saturation:"التشبع",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} شفاف`,"very dark":"داكن جدًا","very light":"فاتح جدًا",vibrant:"ساطع",white:"أبيض",yellow:"أصفر","yellow green":"أخضر مصفر"};var ble={};ble={alpha:"Алфа",black:"черно",blue:"Синьо","blue purple":"синьо лилаво",brightness:"Яркост",brown:"кафяво","brown yellow":"кафяво жълто",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"циан","cyan blue":"циан синьо",dark:"тъмно",gray:"сиво",grayish:"сивкав",green:"Зелено","green cyan":"зелен циан",hue:"Оттенък",light:"светло",lightness:"Лекота",magenta:"магента","magenta pink":"магента розово",orange:"оранжево","orange yellow":"оранжево жълто",pale:"бледо",pink:"розово","pink red":"розово червено",purple:"лилаво","purple magenta":"лилаво магента",red:"Червено","red orange":"червен портокал",saturation:"Наситеност",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} прозрачен`,"very dark":"много тъмно","very light":"много светло",vibrant:" ярко",white:"бяло",yellow:"жълто","yellow green":"жълто зелен"};var Ole={};Ole={alpha:"Alfa",black:"černá",blue:"Modrá","blue purple":"modrofialová",brightness:"Jas",brown:"hnědá","brown yellow":"hnědožlutá",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"tyrkysová","cyan blue":"tyrkysovomodrá",dark:"tmavá",gray:"šedá",grayish:"našedlá",green:"Zelená","green cyan":"zelenotyrkysová",hue:"Odstín",light:"světlá",lightness:"Světlost",magenta:"purpurová","magenta pink":"purpurově růžová",orange:"oranžová","orange yellow":"oranžovožlutá",pale:"bledá",pink:"růžová","pink red":"růžovočervená",purple:"fialová","purple magenta":"fialově purpurová",red:"Červená","red orange":"červenooranžová",saturation:"Sytost",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} průhledné`,"very dark":"velmi tmavá","very light":"velmi světlá",vibrant:"zářivá",white:"bílá",yellow:"žlutá","yellow green":"žlutozelená"};var xle={};xle={alpha:"Alfa",black:"sort",blue:"Blå","blue purple":"blålilla",brightness:"Lysstyrke",brown:"brun","brown yellow":"brungul",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"cyan","cyan blue":"cyan blå",dark:"mørk",gray:"grå",grayish:"grålig",green:"Grøn","green cyan":"grøn cyan",hue:"Tone",light:"lys",lightness:"Lyshed",magenta:"magenta","magenta pink":"magenta pink",orange:"orange","orange yellow":"orangegul",pale:"bleg",pink:"lyserød","pink red":"lyserødlig rød",purple:"lilla","purple magenta":"lilla magenta",red:"Rød","red orange":"rødorange",saturation:"Farvemætning",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} gennemsigtig`,"very dark":"meget mørk","very light":"meget lys",vibrant:"klar",white:"hvid",yellow:"gul","yellow green":"gulgrøn"};var Sle={};Sle={alpha:"Alpha",black:"Schwarz",blue:"Blau","blue purple":"Blaulila",brightness:"Helligkeit",brown:"Braun","brown yellow":"Braungelb",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"Cyan","cyan blue":"Cyanblau",dark:"dunkles",gray:"Grau",grayish:"gräuliches",green:"Grün","green cyan":"Grüncyan",hue:"Farbton",light:"helles",lightness:"Leuchtkraft",magenta:"Magenta","magenta pink":"Magentarosa",orange:"Orange","orange yellow":"Orangegelb",pale:"blasses",pink:"Rosa","pink red":"Rosarot",purple:"Lila","purple magenta":"Lilamagenta",red:"Rot","red orange":"Rotorange",saturation:"Sättigung",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, zu ${e.percentTransparent} transparent`,"very dark":"sehr dunkles","very light":"sehr helles",vibrant:"lebhaftes",white:"Weiß",yellow:"Gelb","yellow green":"Gelbgrün"};var _le={};_le={alpha:"Άλφα",black:"μαύρο",blue:"Μπλε","blue purple":"μπλε μωβ",brightness:"Φωτεινότητα",brown:"καφέ","brown yellow":"καφέ κίτρινο",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"κυανό","cyan blue":"κυανό μπλε",dark:"σκούρο",gray:"γκρι",grayish:"γκριζωπό",green:"Πράσινο","green cyan":"πράσινο κυανό",hue:"Τόνος",light:"ανοιχτό",lightness:"Φωτεινότητα",magenta:"ματζέντα","magenta pink":"ματζέντα ροζ",orange:"πορτοκαλί","orange yellow":"πορτοκαλί κίτρινο",pale:"ανοιχτό",pink:"ροζ","pink red":"ροζ κόκκινο",purple:"μωβ","purple magenta":"μωβ ματζέντα",red:"Κόκκινο","red orange":"κόκκινο πορτοκαλί",saturation:"Κορεσμός",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} διαφανές`,"very dark":"πολύ σκούρο","very light":"πολύ ανοιχτό",vibrant:"έντονο",white:"λευκό",yellow:"κίτρινο","yellow green":"κίτρινο πράσινο"};var Ele={};Ele={hue:"Hue",saturation:"Saturation",lightness:"Lightness",brightness:"Brightness",red:"Red",green:"Green",blue:"Blue",alpha:"Alpha",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} transparent`,"very dark":"very dark",dark:"dark",light:"light","very light":"very light",pale:"pale",grayish:"grayish",vibrant:"vibrant",black:"black",white:"white",gray:"gray",pink:"pink","pink red":"pink red","red orange":"red orange",brown:"brown",orange:"orange","orange yellow":"orange yellow","brown yellow":"brown yellow",yellow:"yellow","yellow green":"yellow green","green cyan":"green cyan",cyan:"cyan","cyan blue":"cyan blue","blue purple":"blue purple",purple:"purple","purple magenta":"purple magenta",magenta:"magenta","magenta pink":"magenta pink"};var Dle={};Dle={alpha:"Alpha",black:"negro",blue:"Azul","blue purple":"púrpura azulado",brightness:"Brillo",brown:"marrón","brown yellow":"amarillo amarronado",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"cian","cyan blue":"azul cian",dark:"oscuro",gray:"gris",grayish:"grisáceo",green:"Verde","green cyan":"cian verdoso",hue:"Tono",light:"claro",lightness:"Luminosidad",magenta:"magenta","magenta pink":"rosa magenta",orange:"naranja","orange yellow":"amarillo anaranjado",pale:"pálido",pink:"rosa","pink red":"rojo rosado",purple:"morado","purple magenta":"magenta violáceo",red:"Rojo","red orange":"naranja rojizo",saturation:"Saturación",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} transparente`,"very dark":"muy oscuro","very light":"muy claro",vibrant:"intenso",white:"blanco",yellow:"amarillo","yellow green":"verde amarillento"};var $le={};$le={alpha:"Alfa",black:"must",blue:"Sinine","blue purple":"sinakaslilla",brightness:"Heledus",brown:"pruun","brown yellow":"pruunikaskollane",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"tsüaan","cyan blue":"tsüaansinine",dark:"tume",gray:"hall",grayish:"hallikas",green:"Roheline","green cyan":"tsüaanroheline",hue:"Värv",light:"valgus",lightness:"Valgus",magenta:"magentapunane","magenta pink":"magentaroosa",orange:"oranž","orange yellow":"oranžikaskollane",pale:"kahvatu",pink:"roosa","pink red":"vaarikapunane",purple:"lilla","purple magenta":"purpurne magenta",red:"Punane","red orange":"punakasoranž",saturation:"Küllastus",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} läbipaistev`,"very dark":"väga tume","very light":"väga hele",vibrant:"ere",white:"valge",yellow:"kollane","yellow green":"kollakasroheline"};var Cle={};Cle={alpha:"Alpha",black:"musta",blue:"Sininen","blue purple":"sinivioletti",brightness:"Kirkkaus",brown:"ruskea","brown yellow":"ruskeankeltainen",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"syaani","cyan blue":"syaaninsininen",dark:"tumma",gray:"harmaa",grayish:"harmahtava",green:"Vihreä","green cyan":"vihreänsyaani",hue:"Sävy",light:"vaalea",lightness:"Valomäärä",magenta:"magenta","magenta pink":"magentapinkki",orange:"oranssi","orange yellow":"oranssinkeltainen",pale:"vaalea",pink:"pinkki","pink red":"vaaleanpunainen",purple:"violetti","purple magenta":"violettimagenta",red:"Punainen","red orange":"punaoranssi",saturation:"Värikylläisyys",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} läpinäkyvä`,"very dark":"hyvin tumma","very light":"erittäin vaalea",vibrant:"eloisa",white:"valkea",yellow:"keltainen","yellow green":"keltavihreä"};var wle={};wle={alpha:"Alpha",black:"Noir",blue:"Bleu","blue purple":"Violet bleu",brightness:"Luminosité",brown:"Brun","brown yellow":"Jaune brun",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"Cyan","cyan blue":"Bleu cyan",dark:"Sombre",gray:"Gris",grayish:"Grisâtre",green:"Vert","green cyan":"Cyan vert",hue:"Teinte",light:"Clair",lightness:"Luminosité",magenta:"Magenta","magenta pink":"Rose magenta",orange:"Orange","orange yellow":"Jaune orangé",pale:"Pâle",pink:"Rose","pink red":"Rouge rosé",purple:"Violet","purple magenta":"Magenta violet",red:"Rouge","red orange":"Orange rouge",saturation:"Saturation",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} transparent`,"very dark":"Très sombre","very light":"Très clair",vibrant:"Vif",white:"Blanc",yellow:"Jaune","yellow green":"Vert jaune"};var Rle={};Rle={alpha:"אלפא",black:"שחור",blue:"כחול","blue purple":"כחול סגול",brightness:"בהירות",brown:"חום","brown yellow":"חום צהוב",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"טורקיז","cyan blue":"כחול ציאן",dark:"כהה",gray:"אפור",grayish:"אפרפר",green:"ירוק","green cyan":"ציאן ירוק",hue:"גוון",light:"אור",lightness:"כמות אור",magenta:"מג'נטה","magenta pink":"ורוד מג'נטה",orange:"כתום","orange yellow":"כתום צהוב",pale:"חיוור",pink:"ורוד","pink red":"ורוד אדום",purple:"סגול","purple magenta":"מג'נטה סגול",red:"אדום","red orange":"כתום אדום",saturation:"רוויה",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} שקוף`,"very dark":"כהה מאוד","very light":"בהיר מאוד",vibrant:"תוסס",white:"לבן",yellow:"צהוב","yellow green":"צהוב ירוק"};var Ale={};Ale={alpha:"Alfa",black:"crno",blue:"Plava","blue purple":"plavo ljubičasta",brightness:"Svjetlina",brown:"smeđa","brown yellow":"smeđe žuta",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"cijan","cyan blue":"cijan plava",dark:"tamno",gray:"siva",grayish:"sivkasto",green:"Zelena","green cyan":"zelena cijan",hue:"Nijansa",light:"svjetlo",lightness:"Osvijetljenost",magenta:"magenta","magenta pink":"magentno ružičasta",orange:"narančasta","orange yellow":"narančasto žuta",pale:"blijeda",pink:"ružičasta","pink red":"ružičasto crvena",purple:"ljubičasta","purple magenta":"ljubičasta magenta",red:"Crvena","red orange":"crveno narančasta",saturation:"Zasićenost",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} prozirnosti`,"very dark":"jako tamna","very light":"vrlo svijetlo",vibrant:"vibrantna",white:"bijela",yellow:"žuto","yellow green":"žuto zelena"};var Tle={};Tle={alpha:"Alfa",black:"fekete",blue:"Kék","blue purple":"kék lila",brightness:"Fényesség",brown:"barna","brown yellow":"barna sárga",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"ciánkék","cyan blue":"ciánkék",dark:"sötét",gray:"szürke",grayish:"szürkés",green:"Zöld","green cyan":"zöld ciánkék",hue:"Színárnyalat",light:"világos",lightness:"Világosság",magenta:"bíbor","magenta pink":"bíbor rózsaszín",orange:"narancs","orange yellow":"narancssárga",pale:"halvány",pink:"rózsaszín","pink red":"rózsaszín piros",purple:"lila","purple magenta":"lila bíbor",red:"Piros","red orange":"piros narancs",saturation:"Telítettség",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} átlátszó`,"very dark":"nagyon sötét","very light":"nagyon világos",vibrant:"élénk",white:"fehér",yellow:"sárga","yellow green":"sárga zöld"};var kle={};kle={alpha:"Alfa",black:"nero",blue:"Blu","blue purple":"blu viola",brightness:"Luminosità",brown:"marrone","brown yellow":"giallo bruno",colorName:e=>`${e.hue} ${e.chroma} ${e.lightness}`,cyan:"ciano","cyan blue":"blu ciano",dark:"scuro",gray:"grigio",grayish:"grigiastro",green:"Verde","green cyan":"verde ciano",hue:"Tonalità",light:"chiaro",lightness:"Luminosità",magenta:"magenta","magenta pink":"rosa magenta",orange:"arancio","orange yellow":"giallo arancio",pale:"tenue",pink:"rosa","pink red":"rosa rosso",purple:"viola","purple magenta":"viola magenta",red:"Rosso","red orange":"rosso arancio",saturation:"Saturazione",transparentColorName:e=>`${e.hue} ${e.chroma} ${e.lightness}, trasparenza ${e.percentTransparent}`,"very dark":"molto scuro","very light":"molto chiaro",vibrant:"vivace",white:"bianco",yellow:"giallo","yellow green":"giallo verde"};var Ple={};Ple={alpha:"アルファ",black:"ブラック",blue:"青","blue purple":"ブルーパープル",brightness:"明るさ",brown:"ブラウン","brown yellow":"ブラウンイエロー",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"シアン","cyan blue":"シアンブルー",dark:"ダーク",gray:"グレー",grayish:"グレイッシュ",green:"緑","green cyan":"グリーンシアン",hue:"色相",light:"ライト",lightness:"明度",magenta:"マゼンタ","magenta pink":"マゼンタピンク",orange:"オレンジ","orange yellow":"オレンジイエロー",pale:"ペール",pink:"ピンク","pink red":"ピンクレッド",purple:"パープル","purple magenta":"パープルマゼンタ",red:"赤","red orange":"レッドオレンジ",saturation:"彩度",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} 透明`,"very dark":"最も暗い","very light":"ベリーライト",vibrant:"鮮やか",white:"ホワイト",yellow:"イエロー","yellow green":"イエローグリーン"};var Mle={};Mle={alpha:"알파",black:"검은색",blue:"파랑","blue purple":"청자색",brightness:"명도",brown:"갈색","brown yellow":"황갈색",colorName:e=>`${e.lightness}, ${e.chroma}, ${e.hue}`,cyan:"청록색","cyan blue":"청록색",dark:"다크",gray:"회색",grayish:"회갈색",green:"초록","green cyan":"청록색",hue:"색조",light:"라이트",lightness:"밝기",magenta:"자홍색","magenta pink":"마젠타 핑크",orange:"주황색","orange yellow":"불그스름한 노랑",pale:"흙색",pink:"분홍색","pink red":"핑크 레드",purple:"자주색","purple magenta":"보라빛 자홍색",red:"빨강","red orange":"붉은 주황색",saturation:"채도",transparentColorName:e=>`${e.lightness}, ${e.chroma}, ${e.hue}, ${e.percentTransparent} 투명도`,"very dark":"매우 어두운","very light":"매우 연함",vibrant:"강렬한",white:"흰색",yellow:"노란색","yellow green":"연두색"};var Fle={};Fle={alpha:"Alfa",black:"juoda",blue:"Mėlyna","blue purple":"melsvai violetinė",brightness:"Ryškumas",brown:"ruda","brown yellow":"rusvai geltona",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"žalsvai mėlyna","cyan blue":"žalsvai mėlyna",dark:"tamsi",gray:"pilka",grayish:"pilkšva",green:"Žalia","green cyan":"žalsvai mėlyna",hue:"Atspalvis",light:"šviesi",lightness:"Šviesumas",magenta:"rausvai raudona","magenta pink":"purpurinė",orange:"oranžinė","orange yellow":"oranžinio atspalvio geltona",pale:"blyški",pink:"rožinė","pink red":"rožinė raudona",purple:"violetinė","purple magenta":"purpurinė rausvai raudona",red:"Raudona","red orange":"rausvai oranžinė",saturation:"Įsotinimas",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} skaidri`,"very dark":"labai tamsi","very light":"labai šviesi",vibrant:"ryški",white:"balta",yellow:"geltona","yellow green":"gelsvai žalia"};var Nle={};Nle={alpha:"Alfa",black:"melns",blue:"Zila","blue purple":"zili violets",brightness:"Spilgtums",brown:"brūns","brown yellow":"brūni dzeltens",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"ciāns","cyan blue":"ciāna zils",dark:"tumšs",gray:"pelēks",grayish:"pelēcīgs",green:"Zaļa","green cyan":"zaļš ciāns",hue:"Nokrāsa",light:"gaišs",lightness:"Gaišums",magenta:"fuksiju","magenta pink":"fuksiju rozā",orange:"oranžs","orange yellow":"oranži dzeltens",pale:"bāls",pink:"rozā","pink red":"rozīgi sarkans",purple:"violets","purple magenta":"violets fuksiju",red:"Sarkana","red orange":"sarkanīgi oranžs",saturation:"Piesātinājums",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} caurspīdīgs`,"very dark":"ļoti tumšs","very light":"ļoti gaišs",vibrant:"košs",white:"balts",yellow:"dzeltens","yellow green":"dzelteni zaļš"};var Ile={};Ile={alpha:"Alfa",black:"svart",blue:"Blå","blue purple":"blålilla",brightness:"Lysstyrke",brown:"brun","brown yellow":"brungul",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"cyan","cyan blue":"cyanblå",dark:"mørk",gray:"grå",grayish:"gråaktig",green:"Grønn","green cyan":"grønncyan",hue:"Fargetone",light:"lys",lightness:"Lyshet",magenta:"magenta","magenta pink":"magentarosa",orange:"oransje","orange yellow":"oransjegul",pale:"blek",pink:"rosa","pink red":"rosarød",purple:"lilla","purple magenta":"lillamagenta",red:"Rød","red orange":"rødoransje",saturation:"Metning",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} gjennomsiktig`,"very dark":"svært mørk","very light":"svært lys",vibrant:"levende",white:"hvit",yellow:"gul","yellow green":"gulgrønn"};var Lle={};Lle={alpha:"Alfa",black:"zwart",blue:"Blauw","blue purple":"paarsblauw",brightness:"Helderheid",brown:"bruin","brown yellow":"bruingeel",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"cyaan","cyan blue":"cyaanblauw",dark:"donker",gray:"grijs",grayish:"grijsachtig",green:"Groen","green cyan":"cyaangroen",hue:"Kleurtoon",light:"licht",lightness:"Lichtsterkte",magenta:"magenta","magenta pink":"magentaroze",orange:"oranje","orange yellow":"oranjegeel",pale:"bleek",pink:"roze","pink red":"rozerood",purple:"paars","purple magenta":"magentapaars",red:"Rood","red orange":"roodoranje",saturation:"Verzadiging",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} transparant`,"very dark":"heel donker","very light":"heel licht",vibrant:"levendig",white:"wit",yellow:"geel","yellow green":"geelgroen"};var Ble={};Ble={alpha:"Alfa",black:"czarny",blue:"Niebieski","blue purple":"niebiesko-fioletowy",brightness:"Jasność",brown:"brązowy","brown yellow":"brązowo-żółty",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"cyjanowy","cyan blue":"cyjanowo-niebieski",dark:"ciemny",gray:"szary",grayish:"szarawy",green:"Zielony","green cyan":"zielono-cyjanowy",hue:"Odcień",light:"jasny",lightness:"Jaskrawość",magenta:"purpurowy","magenta pink":"purpurowo-różowy",orange:"pomarańczowy","orange yellow":"pomarańczowo-żółty",pale:"blady",pink:"różowy","pink red":"różowo-czerwony",purple:"fioletowy","purple magenta":"fioletowo-purpurowy",red:"Czerwony","red orange":"czerwono-pomarańczowy",saturation:"Nasycenie",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} przezroczystości`,"very dark":"bardzo ciemny","very light":"bardzo jasny",vibrant:"intensywny",white:"biały",yellow:"żółty","yellow green":"żółto-zielony"};var Qle={};Qle={alpha:"Alfa",black:"preto",blue:"Azul","blue purple":"roxo azulado",brightness:"Brilho",brown:"marrom","brown yellow":"marrom amarelado",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"ciano","cyan blue":"azul-ciano",dark:"escuro",gray:"cinza",grayish:"acinzentado",green:"Verde","green cyan":"verde-ciano",hue:"Matiz",light:"claro",lightness:"Luminosidade",magenta:"magenta","magenta pink":"rosa-magenta",orange:"laranja","orange yellow":"amarelo alaranjado",pale:"pálido",pink:"rosa","pink red":"rosa avermelhado",purple:"roxo","purple magenta":"roxo-magenta",red:"Vermelho","red orange":"laranja avermelhado",saturation:"Saturação",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} transparente`,"very dark":"muito escuro","very light":"muito claro",vibrant:"vibrante",white:"branco",yellow:"amarelo","yellow green":"verde amarelado"};var zle={};zle={alpha:"Alfa",black:"preto",blue:"Azul","blue purple":"azul-púrpura",brightness:"Luminosidade",brown:"castanho","brown yellow":"amarelo-castanho",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"ciano","cyan blue":"azul-ciano",dark:"escuro",gray:"cinzento",grayish:"acinzentado",green:"Verde","green cyan":"verde-ciano",hue:"Tonalidade",light:"claro",lightness:"Claridade",magenta:"magenta","magenta pink":"rosa-magenta",orange:"laranja","orange yellow":"amarelo-laranja",pale:"pálido",pink:"cor-de-rosa","pink red":"vermelho-rosa",purple:"púrpura","purple magenta":"púrpura-magenta",red:"Vermelho","red orange":"laranja-vermelho",saturation:"Saturação",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} transparente`,"very dark":"muito escuro","very light":"muito claro",vibrant:"vibrante",white:"branco",yellow:"amarelo","yellow green":"verde-amarelo"};var qle={};qle={alpha:"Alfa",black:"negru",blue:"Albastru","blue purple":"albastru-violet",brightness:"Luminozitate",brown:"maro","brown yellow":"galben maro",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"bleu","cyan blue":"albastru-bleu",dark:"închis",gray:"gri",grayish:"cenușiu",green:"Verde","green cyan":"verde bleu",hue:"Nuanță",light:"deschis",lightness:"Luminozitate",magenta:"fucsia","magenta pink":"roz-fucsia",orange:"portocaliu","orange yellow":"galben-portocaliu",pale:"pal",pink:"roz","pink red":"roz-roșu",purple:"violet","purple magenta":"violet-fucsia",red:"Roșu","red orange":"portocaliu-roșu",saturation:"Saturație",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} transparent`,"very dark":"foarte închis","very light":"foarte deschis",vibrant:"plin de viață",white:"alb",yellow:"galben","yellow green":"galben-verde"};var Vle={};Vle={alpha:"Альфа",black:"черный",blue:"Синий","blue purple":"сине-фиолетовый",brightness:"Яркость",brown:"коричневый","brown yellow":"коричнево-желтый",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"голубой","cyan blue":"цвет морской волны",dark:"темный",gray:"серый",grayish:"сероватый",green:"Зеленый","green cyan":"сине-зеленый",hue:"Оттенок",light:"светлый",lightness:"Освещенность",magenta:"пурпурный","magenta pink":"пурпурно-розовый",orange:"оранжевый","orange yellow":"оранжево-желтый",pale:"бледный",pink:"розовый","pink red":"розово-красный",purple:"фиолетовый","purple magenta":"фиолетово-пурпурный",red:"Красный","red orange":"красно-оранжевый",saturation:"Насыщенность",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, прозрачный на ${e.percentTransparent}`,"very dark":"очень темный","very light":"очень светлый",vibrant:"яркий",white:"белый",yellow:"желтый","yellow green":"желто-зеленый"};var Ule={};Ule={alpha:"Alfa",black:"čierna",blue:"Modrá","blue purple":"modrofialová",brightness:"Jas",brown:"hnedá","brown yellow":"hnedožltá",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"azúrová","cyan blue":"azúrová modrá",dark:"tmavá",gray:"sivá",grayish:"sivastá",green:"Zelená","green cyan":"zelená azúrová",hue:"Odtieň",light:"svetlá",lightness:"Svetlosť",magenta:"purpurová","magenta pink":"ružová purpurová",orange:"oranžová","orange yellow":"oranžovožltá",pale:"bledá",pink:"ružová","pink red":"ružovočervená",purple:"fialová","purple magenta":"fialová purpurová",red:"Červená","red orange":"červenooranžová",saturation:"Sýtosť",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} priehľadná`,"very dark":"veľmi tmavá","very light":"veľmi svetlá",vibrant:"energická",white:"biela",yellow:"žltá","yellow green":"žltozelená"};var jle={};jle={alpha:"Alfa",black:"črna",blue:"Modra","blue purple":"modro vijolična",brightness:"Svetlost",brown:"rjava","brown yellow":"rjavo rumena",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"cian","cyan blue":"cian modra",dark:"temna",gray:"siva",grayish:"sivkasta",green:"Zelena","green cyan":"zelena cian",hue:"Barva",light:"svetla",lightness:"Lahkost",magenta:"škrlatna","magenta pink":"škrlatno roza",orange:"oranžna","orange yellow":"oranžno rumena",pale:"bleda",pink:"roza","pink red":"roza rdeča",purple:"vijolična","purple magenta":"vijolično škrlatna",red:"Rdeča","red orange":"rdeče oranžna",saturation:"Nasičenost",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} prozorna`,"very dark":"zelo temna","very light":"zelo svetla",vibrant:"živahna",white:"bela",yellow:"rumena","yellow green":"rumeno zelena"};var Zle={};Zle={alpha:"Alfa",black:"crno",blue:"Plava","blue purple":"plavoljubičasta",brightness:"Osvetljenost",brown:"smeđa","brown yellow":"smeđežuta",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"cijan","cyan blue":"cijan plava",dark:"tamno",gray:"siva",grayish:"sivkasta",green:"Zelena","green cyan":"zeleno cijan",hue:"Nijansa",light:"svetla",lightness:"Osvetljenje",magenta:"purpurnocrvena","magenta pink":"magenta ružičasta",orange:"narandžasta","orange yellow":"narandžastožuta",pale:"bledo",pink:"ružičasta","pink red":"ružičastocrvena",purple:"ljubičasta","purple magenta":"ljubičasta magenta",red:"Crvena","red orange":"crvenonarandžasta",saturation:"Zasićenje",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} prozirna`,"very dark":"veoma tamno","very light":"vrlo svetlo",vibrant:"živopisna",white:"bela",yellow:"žuto","yellow green":"žutozelena"};var Hle={};Hle={alpha:"Alfa",black:"svart",blue:"Blått","blue purple":"blålila",brightness:"Ljusstyrka",brown:"brun","brown yellow":"brungul",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"cyan","cyan blue":"cyanblå",dark:"mörk",gray:"grå",grayish:"gråaktig",green:"Grönt","green cyan":"grön cyan",hue:"Nyans",light:"ljus",lightness:"Ljushet",magenta:"magenta","magenta pink":"magentarosa",orange:"orange","orange yellow":"orangegul",pale:"blek",pink:"rosa","pink red":"rosaröd",purple:"lila","purple magenta":"lila magenta",red:"Rött","red orange":"rödorange",saturation:"Mättnad",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} genomskinlig`,"very dark":"mycket mörk","very light":"mycket ljus",vibrant:"livfull",white:"vit",yellow:"gul","yellow green":"gulgrön"};var Xle={};Xle={alpha:"Alfa",black:"siyah",blue:"Mavi","blue purple":"mavi mor",brightness:"Parlaklık",brown:"kahverengi","brown yellow":"kahverengi sarı",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"camgöbeği","cyan blue":"camgöbeği mavi",dark:"koyu",gray:"gri",grayish:"grimsi",green:"Yeşil","green cyan":"yeşil camgöbeği",hue:"Ton",light:"açık",lightness:"Canlılık",magenta:"eflatun","magenta pink":"eflatun pembe",orange:"turuncu","orange yellow":"turuncu sarı",pale:"solgun",pink:"pembe","pink red":"pembe kırmızı",purple:"mor","purple magenta":"mor eflatun",red:"Kırmızı","red orange":"kırmızı portakal",saturation:"Doygunluk",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} saydam`,"very dark":"çok koyu","very light":"çok açık",vibrant:"canlı",white:"beyaz",yellow:"sarı","yellow green":"sarı yeşil"};var Yle={};Yle={alpha:"Альфа",black:"чорний",blue:"Синій","blue purple":"синьо-фіолетовий",brightness:"Яскравість",brown:"коричневий","brown yellow":"коричнево-жовтий",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"блакитний","cyan blue":"синьо-блакитний",dark:"темний",gray:"сірий",grayish:"сіруватий",green:"Зелений","green cyan":"зелено-блакитний",hue:"Тон",light:"світлий",lightness:"Освітленість",magenta:"пурпуровий","magenta pink":"пурпурово-рожевий",orange:"помаранчевий","orange yellow":"помаранчево-жовтий",pale:"блідий",pink:"рожевий","pink red":"рожево-червоний",purple:"фіолетовий","purple magenta":"фіолетово-пурпуровий",red:"Червоний","red orange":"червоно-помаранчевий",saturation:"Насиченість",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, прозорий на ${e.percentTransparent}`,"very dark":"дуже темний","very light":"дуже світлий",vibrant:"яскравий",white:"білий",yellow:"жовтий","yellow green":"жовто-зелений"};var Gle={};Gle={alpha:"Alpha",black:"黑色",blue:"蓝色","blue purple":"蓝紫色",brightness:"亮度",brown:"棕色的","brown yellow":"棕黄色",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"蓝绿色","cyan blue":"青蓝色",dark:"深色",gray:"灰色",grayish:"浅灰色的",green:"绿色","green cyan":"绿青色",hue:"色相",light:"浅色",lightness:"明亮度",magenta:"紫红色","magenta pink":"紫粉色",orange:"橙色","orange yellow":"橙黄色",pale:"苍白的",pink:"粉色","pink red":"粉红色",purple:"紫色","purple magenta":"紫洋红色",red:"红色","red orange":"红橙色",saturation:"饱和度",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} 透明`,"very dark":"很暗","very light":"很浅",vibrant:"生机勃勃",white:"白色",yellow:"黄色","yellow green":"黄色/绿色"};var Kle={};Kle={alpha:"Alpha",black:"黑",blue:"藍色","blue purple":"藍紫",brightness:"亮度",brown:"棕","brown yellow":"棕黃",colorName:e=>`${e.lightness} ${e.chroma} ${e.hue}`,cyan:"青","cyan blue":"青藍",dark:"暗",gray:"灰",grayish:"偏灰",green:"綠色","green cyan":"青綠",hue:"色相",light:"淺",lightness:"明亮",magenta:"洋紅","magenta pink":"淺洋紅",orange:"橙","orange yellow":"橙黃",pale:"淡",pink:"粉紅","pink red":"粉紅",purple:"紫","purple magenta":"紫洋紅",red:"紅色","red orange":"橙紅",saturation:"飽和度",transparentColorName:e=>`${e.lightness} ${e.chroma} ${e.hue}, ${e.percentTransparent} 透明`,"very dark":"很暗","very light":"很淺",vibrant:"鮮豔",white:"白",yellow:"黃","yellow green":"黃綠"};var Wle={};Wle={"ar-AE":yle,"bg-BG":ble,"cs-CZ":Ole,"da-DK":xle,"de-DE":Sle,"el-GR":_le,"en-US":Ele,"es-ES":Dle,"et-EE":$le,"fi-FI":Cle,"fr-FR":wle,"he-IL":Rle,"hr-HR":Ale,"hu-HU":Tle,"it-IT":kle,"ja-JP":Ple,"ko-KR":Mle,"lt-LT":Fle,"lv-LV":Nle,"nb-NO":Ile,"nl-NL":Lle,"pl-PL":Ble,"pt-BR":Qle,"pt-PT":zle,"ro-RO":qle,"ru-RU":Vle,"sk-SK":Ule,"sl-SI":jle,"sr-SP":Zle,"sv-SE":Hle,"tr-TR":Xle,"uk-UA":Yle,"zh-CN":Gle,"zh-TW":Kle};function q$e(e){return e&&e.__esModule?e.default:e}let dk=new tp(q$e(Wle));function vy(e){let t=qv.parse(e)||Vv.parse(e)||Uv.parse(e);if(t)return t;throw new Error("Invalid color value: "+e)}function Jle(e){return e===360?e:(e%360+360)%360}const Aj=.68,V$e=.85,Tj=.55,kj=.001,hk=[[0,"pink"],[15,"red"],[48,"orange"],[94,"yellow"],[135,"green"],[175,"cyan"],[264,"blue"],[284,"purple"],[320,"magenta"],[349,"pink"]];class HN{toHexInt(){return this.toFormat("rgb").toHexInt()}getChannelValue(t){if(t in this)return this[t];throw new Error("Unsupported color channel: "+t)}withChannelValue(t,n){if(t in this){let r=this.clone();return r[t]=n,r}throw new Error("Unsupported color channel: "+t)}getChannelName(t,n){return(tp.getGlobalDictionaryForPackage("@react-stately/color")||dk).getStringForLocale(t,n)}getColorSpaceAxes(t){let{xChannel:n,yChannel:r}=t,i=n||this.getColorChannels().find(l=>l!==r),a=r||this.getColorChannels().find(l=>l!==i),u=this.getColorChannels().find(l=>l!==i&&l!==a);return{xChannel:i,yChannel:a,zChannel:u}}getColorName(t){let[n,r,i]=Pj(this),a=tp.getGlobalDictionaryForPackage("@react-stately/color")||dk;if(n>.999)return a.getStringForLocale("white",t);if(n<.001)return a.getStringForLocale("black",t);let u;[u,n]=this.getOklchHue(n,r,i,t);let l="",o="";r<=.1&&r>=kj?n>=.7?o="pale":o="grayish":r>=.15&&(o="vibrant"),n<.3?l="very dark":n<Tj?l="dark":n<.7||(n<.85?l="light":l="very light"),o&&(o=a.getStringForLocale(o,t)),l&&(l=a.getStringForLocale(l,t));let s=this.getChannelValue("alpha"),c=new yN(t,a);if(s<1){let d=new Jv(t,{style:"percent"}).format(1-s);return c.format("transparentColorName",{lightness:l,chroma:o,hue:u,percentTransparent:d}).replace(/\s+/g," ").trim()}else return c.format("colorName",{lightness:l,chroma:o,hue:u}).replace(/\s+/g," ").trim()}getOklchHue(t,n,r,i){let a=tp.getGlobalDictionaryForPackage("@react-stately/color")||dk;if(n<kj)return[a.getStringForLocale("gray",i),t];for(let u=0;u<hk.length;u++){let[l,o]=hk[u],[s,c]=hk[u+1]||[360,"pink"];if(r>=l&&r<s)return o==="orange"&&(t<Aj?o="brown":t=t-Aj+Tj),r>l+(s-l)/2&&o!==c?o=`${o} ${c}`:o==="yellow"&&t<V$e&&(o="yellow green"),[a.getStringForLocale(o,i).toLocaleLowerCase(i),t]}throw new Error("Unexpected hue")}getHueName(t){let[n,r,i]=Pj(this),[a]=this.getOklchHue(n,r,i,t);return a}}class qv extends HN{static parse(t){let n=[];if(/^#[\da-f]+$/i.test(t)&&[4,5,7,9].includes(t.length)){const a=(t.length<6?t.replace(/[^#]/gi,"$&$&"):t).slice(1).split("");for(;a.length>0;)n.push(parseInt(a.splice(0,2).join(""),16));n[3]=n[3]!==void 0?n[3]/255:void 0}const r=t.match(/^rgba?\((.*)\)$/);if(r!=null&&r[1]&&(n=r[1].split(",").map(a=>Number(a.trim())),n=n.map((a,u)=>so(a??0,0,u<3?255:1))),!(n[0]===void 0||n[1]===void 0||n[2]===void 0)){var i;return n.length<3?void 0:new qv(n[0],n[1],n[2],(i=n[3])!==null&&i!==void 0?i:1)}}toString(t="css"){switch(t){case"hex":return"#"+(this.red.toString(16).padStart(2,"0")+this.green.toString(16).padStart(2,"0")+this.blue.toString(16).padStart(2,"0")).toUpperCase();case"hexa":return"#"+(this.red.toString(16).padStart(2,"0")+this.green.toString(16).padStart(2,"0")+this.blue.toString(16).padStart(2,"0")+Math.round(this.alpha*255).toString(16).padStart(2,"0")).toUpperCase();case"rgb":return`rgb(${this.red}, ${this.green}, ${this.blue})`;case"css":case"rgba":return`rgba(${this.red}, ${this.green}, ${this.blue}, ${this.alpha})`;default:return this.toFormat(t).toString(t)}}toFormat(t){switch(t){case"hex":case"hexa":case"rgb":case"rgba":return this;case"hsb":case"hsba":return this.toHSB();case"hsl":case"hsla":return this.toHSL();default:throw new Error("Unsupported color conversion: rgb -> "+t)}}toHexInt(){return this.red<<16|this.green<<8|this.blue}toHSB(){const t=this.red/255,n=this.green/255,r=this.blue/255,i=Math.min(t,n,r),a=Math.max(t,n,r),u=a-i,l=a===0?0:u/a;let o=0;if(u!==0){switch(a){case t:o=(n-r)/u+(n<r?6:0);break;case n:o=(r-t)/u+2;break;case r:o=(t-n)/u+4;break}o/=6}return new Vv(yl(o*360,2),yl(l*100,2),yl(a*100,2),this.alpha)}toHSL(){const t=this.red/255,n=this.green/255,r=this.blue/255,i=Math.min(t,n,r),a=Math.max(t,n,r),u=(a+i)/2,l=a-i;let o,s;if(l===0)o=s=0;else{switch(s=l/(u<.5?a+i:2-a-i),a){case t:o=(n-r)/l+(n<r?6:0);break;case n:o=(r-t)/l+2;break;case r:default:o=(t-n)/l+4;break}o/=6}return new Uv(yl(o*360,2),yl(s*100,2),yl(u*100,2),this.alpha)}clone(){return new qv(this.red,this.green,this.blue,this.alpha)}getChannelRange(t){switch(t){case"red":case"green":case"blue":return{minValue:0,maxValue:255,step:1,pageSize:17};case"alpha":return{minValue:0,maxValue:1,step:.01,pageSize:.1};default:throw new Error("Unknown color channel: "+t)}}getChannelFormatOptions(t){switch(t){case"red":case"green":case"blue":return{style:"decimal"};case"alpha":return{style:"percent"};default:throw new Error("Unknown color channel: "+t)}}formatChannelValue(t,n){let r=this.getChannelFormatOptions(t),i=this.getChannelValue(t);return new Jv(n,r).format(i)}getColorSpace(){return"rgb"}getColorChannels(){return qv.colorChannels}constructor(t,n,r,i){super(),this.red=t,this.green=n,this.blue=r,this.alpha=i}}qv.colorChannels=["red","green","blue"];const U$e=/hsb\(([-+]?\d+(?:.\d+)?\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d+(?:.\d+)?%)\)|hsba\(([-+]?\d+(?:.\d+)?\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d(.\d+)?)\)/;class Vv extends HN{static parse(t){let n;if(n=t.match(U$e)){var r;const[i,a,u,l]=((r=n[1])!==null&&r!==void 0?r:n[2]).split(",").map(o=>Number(o.trim().replace("%","")));return new Vv(Jle(i),so(a,0,100),so(u,0,100),so(l??1,0,1))}}toString(t="css"){switch(t){case"css":return this.toHSL().toString("css");case"hex":return this.toRGB().toString("hex");case"hexa":return this.toRGB().toString("hexa");case"hsb":return`hsb(${this.hue}, ${yl(this.saturation,2)}%, ${yl(this.brightness,2)}%)`;case"hsba":return`hsba(${this.hue}, ${yl(this.saturation,2)}%, ${yl(this.brightness,2)}%, ${this.alpha})`;default:return this.toFormat(t).toString(t)}}toFormat(t){switch(t){case"hsb":case"hsba":return this;case"hsl":case"hsla":return this.toHSL();case"rgb":case"rgba":return this.toRGB();default:throw new Error("Unsupported color conversion: hsb -> "+t)}}toHSL(){let t=this.saturation/100,n=this.brightness/100,r=n*(1-t/2);return t=r===0||r===1?0:(n-r)/Math.min(r,1-r),new Uv(yl(this.hue,2),yl(t*100,2),yl(r*100,2),this.alpha)}toRGB(){let t=this.hue,n=this.saturation/100,r=this.brightness/100,i=(a,u=(a+t/60)%6)=>r-n*r*Math.max(Math.min(u,4-u,1),0);return new qv(Math.round(i(5)*255),Math.round(i(3)*255),Math.round(i(1)*255),this.alpha)}clone(){return new Vv(this.hue,this.saturation,this.brightness,this.alpha)}getChannelRange(t){switch(t){case"hue":return{minValue:0,maxValue:360,step:1,pageSize:15};case"saturation":case"brightness":return{minValue:0,maxValue:100,step:1,pageSize:10};case"alpha":return{minValue:0,maxValue:1,step:.01,pageSize:.1};default:throw new Error("Unknown color channel: "+t)}}getChannelFormatOptions(t){switch(t){case"hue":return{style:"unit",unit:"degree",unitDisplay:"narrow"};case"saturation":case"brightness":case"alpha":return{style:"percent"};default:throw new Error("Unknown color channel: "+t)}}formatChannelValue(t,n){let r=this.getChannelFormatOptions(t),i=this.getChannelValue(t);return(t==="saturation"||t==="brightness")&&(i/=100),new Jv(n,r).format(i)}getColorSpace(){return"hsb"}getColorChannels(){return Vv.colorChannels}constructor(t,n,r,i){super(),this.hue=t,this.saturation=n,this.brightness=r,this.alpha=i}}Vv.colorChannels=["hue","saturation","brightness"];const j$e=/hsl\(([-+]?\d+(?:.\d+)?\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d+(?:.\d+)?%)\)|hsla\(([-+]?\d+(?:.\d+)?\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d(.\d+)?)\)/;class Uv extends HN{static parse(t){let n;if(n=t.match(j$e)){var r;const[i,a,u,l]=((r=n[1])!==null&&r!==void 0?r:n[2]).split(",").map(o=>Number(o.trim().replace("%","")));return new Uv(Jle(i),so(a,0,100),so(u,0,100),so(l??1,0,1))}}toString(t="css"){switch(t){case"hex":return this.toRGB().toString("hex");case"hexa":return this.toRGB().toString("hexa");case"hsl":return`hsl(${this.hue}, ${yl(this.saturation,2)}%, ${yl(this.lightness,2)}%)`;case"css":case"hsla":return`hsla(${this.hue}, ${yl(this.saturation,2)}%, ${yl(this.lightness,2)}%, ${this.alpha})`;default:return this.toFormat(t).toString(t)}}toFormat(t){switch(t){case"hsl":case"hsla":return this;case"hsb":case"hsba":return this.toHSB();case"rgb":case"rgba":return this.toRGB();default:throw new Error("Unsupported color conversion: hsl -> "+t)}}toHSB(){let t=this.saturation/100,n=this.lightness/100,r=n+t*Math.min(n,1-n);return t=r===0?0:2*(1-n/r),new Vv(yl(this.hue,2),yl(t*100,2),yl(r*100,2),this.alpha)}toRGB(){let t=this.hue,n=this.saturation/100,r=this.lightness/100,i=n*Math.min(r,1-r),a=(u,l=(u+t/30)%12)=>r-i*Math.max(Math.min(l-3,9-l,1),-1);return new qv(Math.round(a(0)*255),Math.round(a(8)*255),Math.round(a(4)*255),this.alpha)}clone(){return new Uv(this.hue,this.saturation,this.lightness,this.alpha)}getChannelRange(t){switch(t){case"hue":return{minValue:0,maxValue:360,step:1,pageSize:15};case"saturation":case"lightness":return{minValue:0,maxValue:100,step:1,pageSize:10};case"alpha":return{minValue:0,maxValue:1,step:.01,pageSize:.1};default:throw new Error("Unknown color channel: "+t)}}getChannelFormatOptions(t){switch(t){case"hue":return{style:"unit",unit:"degree",unitDisplay:"narrow"};case"saturation":case"lightness":case"alpha":return{style:"percent"};default:throw new Error("Unknown color channel: "+t)}}formatChannelValue(t,n){let r=this.getChannelFormatOptions(t),i=this.getChannelValue(t);return(t==="saturation"||t==="lightness")&&(i/=100),new Jv(n,r).format(i)}getColorSpace(){return"hsl"}getColorChannels(){return Uv.colorChannels}constructor(t,n,r,i){super(),this.hue=t,this.saturation=n,this.lightness=r,this.alpha=i}}Uv.colorChannels=["hue","saturation","lightness"];function Pj(e){let t=e.toFormat("rgb"),n=t.getChannelValue("red")/255,r=t.getChannelValue("green")/255,i=t.getChannelValue("blue")/255;[n,r,i]=H$e(n,r,i);let[a,u,l]=X$e(n,r,i),[o,s,c]=Y$e(a,u,l);return Z$e(o,s,c)}function Z$e(e,t,n){var r=Math.atan2(n,t)*180/Math.PI;return[e,Math.sqrt(t**2+n**2),r>=0?r:r+360]}function H$e(e,t,n){return[pk(e),pk(t),pk(n)]}function pk(e){let t=e<0?-1:1,n=Math.abs(e);return n<=.04045?e/12.92:t*Math.pow((n+.055)/1.055,2.4)}function X$e(e,t,n){return G9([.4123907992659595,.35758433938387796,.1804807884018343,.21263900587151036,.7151686787677559,.07219231536073371,.01933081871559185,.11919477979462599,.9505321522496606],e,t,n)}function Y$e(e,t,n){const r=[.819022437996703,.3619062600528904,-.1288737815209879,.0329836539323885,.9292868615863434,.0361446663506424,.0481771893596242,.2642395317527308,.6335478284694309],i=[.210454268309314,.7936177747023054,-.0040720430116193,1.9779985324311684,-2.42859224204858,.450593709617411,.0259040424655478,.7827717124575296,-.8086757549230774];let[a,u,l]=G9(r,e,t,n);return G9(i,Math.cbrt(a),Math.cbrt(u),Math.cbrt(l))}function G9(e,t,n,r){let i=e[0]*t+e[1]*n+e[2]*r,a=e[3]*t+e[4]*n+e[5]*r,u=e[6]*t+e[7]*n+e[8]*r;return[i,a,u]}const G$e=0,K$e=100,W$e=1;function J$e(e){const{isDisabled:t=!1,minValue:n=G$e,maxValue:r=K$e,numberFormatter:i,step:a=W$e,orientation:u="horizontal"}=e;let l=P.useMemo(()=>{let X=(r-n)/10;return X=Jc(X,0,X+a,a),Math.max(X,a)},[a,r,n]),o=P.useCallback(X=>X==null?void 0:X.map((B,ne)=>{let oe=ne===0?n:X[ne-1],ie=ne===X.length-1?r:X[ne+1];return Jc(B,oe,ie,a)}),[n,r,a]),s=P.useMemo(()=>o(Fj(e.value)),[e.value,o]),c=P.useMemo(()=>{var X;return o((X=Fj(e.defaultValue))!==null&&X!==void 0?X:[n])},[e.defaultValue,n,o]),d=Nj(e.value,e.defaultValue,e.onChange),h=Nj(e.value,e.defaultValue,e.onChangeEnd);const[m,y]=Yu(s,c,d);let[v]=P.useState(m);const[S,_]=P.useState(new Array(m.length).fill(!1)),E=P.useRef(new Array(m.length).fill(!0)),[$,k]=P.useState(void 0),A=P.useRef(m),w=P.useRef(S);let R=X=>{A.current=X,y(X)},F=X=>{w.current=X,_(X)};function I(X){return(X-n)/(r-n)}function U(X){return X===0?n:m[X-1]}function q(X){return X===m.length-1?r:m[X+1]}function V(X){return E.current[X]}function N(X,B){E.current[X]=B}function j(X,B){if(t||!V(X))return;const ne=U(X),oe=q(X);B=Jc(B,ne,oe,a);let ie=Mj(A.current,X,B);R(ie)}function W(X,B){if(t||!V(X))return;B&&(A.current=m);const ne=w.current[X];w.current=Mj(w.current,X,B),F(w.current),h&&ne&&!w.current.some(Boolean)&&h(A.current)}function K(X){return i.format(X)}function J(X,B){j(X,H(B))}function te(X){return Math.round((X-n)/a)*a+n}function H(X){const B=X*(r-n)+n;return so(te(B),n,r)}function M(X,B=1){let ne=Math.max(B,a);j(X,Jc(m[X]+ne,n,r,a))}function G(X,B=1){let ne=Math.max(B,a);j(X,Jc(m[X]-ne,n,r,a))}return{values:m,defaultValues:e.defaultValue!==void 0?c:v,getThumbValue:X=>m[X],setThumbValue:j,setThumbPercent:J,isThumbDragging:X=>S[X],setThumbDragging:W,focusedThumb:$,setFocusedThumb:k,getThumbPercent:X=>I(m[X]),getValuePercent:I,getThumbValueLabel:X=>K(m[X]),getFormattedValue:K,getThumbMinValue:U,getThumbMaxValue:q,getPercentValue:H,isThumbEditable:V,setThumbEditable:N,incrementThumb:M,decrementThumb:G,step:a,pageSize:l,orientation:u,isDisabled:t}}function Mj(e,t,n){return e[t]===n?e:[...e.slice(0,t),n,...e.slice(t+1)]}function Fj(e){if(e!=null)return Array.isArray(e)?e:[e]}function Nj(e,t,n){return r=>{typeof e=="number"||typeof t=="number"?n==null||n(r[0]):n==null||n(r)}}function f1(e){return P.useMemo(()=>{if(typeof e=="string")try{return vy(e)}catch{return}return e},[e])}const o$=vy("#000000"),eoe=vy("#FFFFFF"),eCe=o$.toHexInt(),tCe=eoe.toHexInt();function nCe(e){let{value:t,defaultValue:n,onChange:r}=e,{step:i}=o$.getChannelRange("red"),a=f1(n),[u,l]=Yu(f1(t),a,r),[o]=P.useState(u),[s,c]=P.useState(()=>(t||n)&&u?u.toString("hex"):""),d=Og({...e,value:u}),h=w=>{if(!u||!w){l(w);return}if(w.toHexInt()!==u.toHexInt()){l(w);return}},[m,y]=P.useState(u);m!==u&&(c(u?u.toString("hex"):""),y(u));let v=P.useMemo(()=>{let w;try{w=vy(s.startsWith("#")?s:`#${s}`)}catch{w=null}return w},[s]);return{...d,validate:w=>{var R;return w===""||!!(!((R=w.match(/^#?[0-9a-f]{0,6}$/i))===null||R===void 0)&&R[0])},colorValue:u,defaultColorValue:a??o,setColorValue:l,inputValue:s,setInputValue:c,commit:()=>{if(!s.length){h(null),c(t===void 0||u===null?"":u.toString("hex"));return}if(v==null){c(u?u.toString("hex"):"");return}h(v);let w="";u&&(w=u.toString("hex")),c(w),d.commitValidation()},increment:()=>{let w=Ij(v,i);w===u&&c(w.toString("hex")),h(w),d.commitValidation()},incrementToMax:()=>h(eoe),decrement:()=>{let w=Ij(v,-i);w===u&&c(w.toString("hex")),h(w),d.commitValidation()},decrementToMin:()=>h(o$)}}function Ij(e,t){let n=e||o$,r=n.toHexInt(),i=Math.min(Math.max(r+t,eCe),tCe);if(i!==r){let a=`#${i.toString(16).padStart(6,"0").toUpperCase()}`;n=vy(a)}return n}function toe(e){let{minValue:t,maxValue:n,step:r,formatOptions:i,value:a,defaultValue:u=NaN,onChange:l,locale:o,isDisabled:s,isReadOnly:c}=e;a===null&&(a=NaN),a!==void 0&&!isNaN(a)&&(r!==void 0&&!isNaN(r)?a=Jc(a,t,n,r):a=so(a,t,n)),isNaN(u)||(r!==void 0&&!isNaN(r)?u=Jc(u,t,n,r):u=so(u,t,n));let[d,h]=Yu(a,isNaN(u)?NaN:u,l),[m]=P.useState(d),[y,v]=P.useState(()=>isNaN(d)?"":new Jv(o,i).format(d)),S=P.useMemo(()=>new CN(o,i),[o,i]),_=P.useMemo(()=>S.getNumberingSystem(y),[S,y]),E=P.useMemo(()=>new Jv(o,{...i,numberingSystem:_}),[o,i,_]),$=P.useMemo(()=>E.resolvedOptions(),[E]),k=P.useCallback(B=>isNaN(B)||B===null?"":E.format(B),[E]),A=Og({...e,value:d}),w=r!==void 0&&!isNaN(r)?r:1;$.style==="percent"&&(r===void 0||isNaN(r))&&(w=.01);let[R,F]=P.useState(d),[I,U]=P.useState(o),[q,V]=P.useState(i);(!Object.is(d,R)||o!==I||i!==q)&&(v(k(d)),F(d),U(o),V(i));let N=P.useMemo(()=>S.parse(y),[S,y]),j=()=>{if(!y.length){h(NaN),v(a===void 0?"":k(d));return}if(isNaN(N)){v(k(d));return}let B;r===void 0||isNaN(r)?B=so(N,t,n):B=Jc(N,t,n,r),B=S.parse(k(B)),h(B),v(k(a===void 0?B:d)),A.commitValidation()},W=(B,ne=0)=>{let oe=N;if(isNaN(oe)){let ie=isNaN(ne)?0:ne;return Jc(ie,t,n,w)}else{let ie=Jc(oe,t,n,w);return B==="+"&&ie>oe||B==="-"&&ie<oe?ie:Jc(mk(B,oe,w),t,n,w)}},K=()=>{let B=W("+",t);B===d&&v(k(B)),h(B),A.commitValidation()},J=()=>{let B=W("-",n);B===d&&v(k(B)),h(B),A.commitValidation()},te=()=>{n!=null&&(h(Jc(n,t,n,w)),A.commitValidation())},H=()=>{t!=null&&(h(t),A.commitValidation())},M=P.useMemo(()=>!s&&!c&&(isNaN(N)||n===void 0||isNaN(n)||Jc(N,t,n,w)>N||mk("+",N,w)<=n),[s,c,t,n,w,N]),G=P.useMemo(()=>!s&&!c&&(isNaN(N)||t===void 0||isNaN(t)||Jc(N,t,n,w)<N||mk("-",N,w)>=t),[s,c,t,n,w,N]);return{...A,validate:B=>S.isValidPartialNumber(B,t,n),increment:K,incrementToMax:te,decrement:J,decrementToMin:H,canIncrement:M,canDecrement:G,minValue:t,maxValue:n,numberValue:N,defaultNumberValue:isNaN(u)?m:u,setNumberValue:h,setInputValue:v,inputValue:y,commit:j}}function mk(e,t,n){let r=e==="+"?t+n:t-n;if(t%1!==0||n%1!==0){const i=t.toString().split("."),a=n.toString().split("."),u=i[1]&&i[1].length||0,l=a[1]&&a[1].length||0,o=Math.pow(10,Math.max(u,l));t=Math.round(t*o),n=Math.round(n*o),r=e==="+"?t+n:t-n,r/=o}return r}function rCe(e){let{channel:t,colorSpace:n,locale:r}=e,i=f1(e.value),a=f1(e.defaultValue),[u,l]=Yu(i,a??null,e.onChange),o=Lj(u,n),[s]=P.useState(u),c=a??s,d=Lj(c,n),h=o.getChannelValue(t),m=o.getChannelRange(t),y=P.useMemo(()=>o.getChannelFormatOptions(t),[o,t]),v=y.style==="percent"&&m.maxValue===100?100:1;return{...toe({locale:r,value:u===null?NaN:h/v,defaultValue:c===null?NaN:d.getChannelValue(t)/v,onChange:_=>{Number.isNaN(_)?l(null):l(o.withChannelValue(t,_*v))},minValue:m.minValue/v,maxValue:m.maxValue/v,step:m.step/v,formatOptions:y}),colorValue:o,defaultColorValue:c,setColorValue:l}}function Lj(e,t){let n=f1("#000");return P.useMemo(()=>{let r=e||n;return t&&r?r.toFormat(t):r},[n,e,t])}function iCe(e){let t=f1(e.value),n=f1(e.defaultValue||"#000000"),[r,i]=Yu(t||void 0,n,e.onChange);return{color:r,setColor(a){a!=null&&i(a||vy("#000000"))}}}const XN=new WeakMap;function K9(e,t){let n=XN.get(e);if(!n)throw new Error("Unknown slider state");return`${n.id}-${t}`}function aCe(e,t,n){let{labelProps:r,fieldProps:i}=gS(e),a=e.orientation==="vertical";var u;XN.set(t,{id:(u=r.id)!==null&&u!==void 0?u:i.id,"aria-describedby":e["aria-describedby"],"aria-details":e["aria-details"]});let{direction:l}=Wi(),{addGlobalListener:o,removeGlobalListener:s}=py();const c=P.useRef(null),d=l==="rtl",h=P.useRef(null),{moveProps:m}=Vne({onMoveStart(){h.current=null},onMove({deltaX:_,deltaY:E}){if(!n.current)return;let{height:$,width:k}=n.current.getBoundingClientRect(),A=a?$:k;h.current==null&&c.current!=null&&(h.current=t.getThumbPercent(c.current)*A);let w=a?E:_;if((a||d)&&(w=-w),h.current+=w,c.current!=null&&n.current){const R=so(h.current/A,0,1);t.setThumbPercent(c.current,R)}},onMoveEnd(){c.current!=null&&(t.setThumbDragging(c.current,!1),c.current=null)}});let y=P.useRef(void 0),v=(_,E,$,k)=>{if(n.current&&!e.isDisabled&&t.values.every((A,w)=>!t.isThumbDragging(w))){let{height:A,width:w,top:R,left:F}=n.current.getBoundingClientRect(),I=a?A:w,N=((a?k:$)-(a?R:F))/I;(l==="rtl"||a)&&(N=1-N);let j=t.getPercentValue(N),W,K=t.values.findIndex(J=>j-J<0);if(K===0)W=K;else if(K===-1)W=t.values.length-1;else{let J=t.values[K-1],te=t.values[K];Math.abs(J-j)<Math.abs(te-j)?W=K-1:W=K}W>=0&&t.isThumbEditable(W)?(_.preventDefault(),c.current=W,t.setFocusedThumb(W),y.current=E,t.setThumbDragging(c.current,!0),t.setThumbValue(W,j),o(window,"mouseup",S,!1),o(window,"touchend",S,!1),o(window,"pointerup",S,!1)):c.current=null}},S=_=>{var E,$;(($=_.pointerId)!==null&&$!==void 0?$:(E=_.changedTouches)===null||E===void 0?void 0:E[0].identifier)===y.current&&(c.current!=null&&(t.setThumbDragging(c.current,!1),c.current=null),s(window,"mouseup",S,!1),s(window,"touchend",S,!1),s(window,"pointerup",S,!1))};return"htmlFor"in r&&r.htmlFor&&(delete r.htmlFor,r.onClick=()=>{var _;(_=document.getElementById(K9(t,0)))===null||_===void 0||_.focus(),AN("keyboard")}),{labelProps:r,groupProps:{role:"group",...i},trackProps:Et({onMouseDown(_){_.button!==0||_.altKey||_.ctrlKey||_.metaKey||v(_,void 0,_.clientX,_.clientY)},onPointerDown(_){_.pointerType==="mouse"&&(_.button!==0||_.altKey||_.ctrlKey||_.metaKey)||v(_,_.pointerId,_.clientX,_.clientY)},onTouchStart(_){v(_,_.changedTouches[0].identifier,_.changedTouches[0].clientX,_.changedTouches[0].clientY)},style:{position:"relative",touchAction:"none"}},m),outputProps:{htmlFor:t.values.map((_,E)=>K9(t,E)).join(" "),"aria-live":"off"}}}function uCe(e,t){let{index:n=0,isRequired:r,validationState:i,isInvalid:a,trackRef:u,inputRef:l,orientation:o=t.orientation,name:s,form:c}=e,d=e.isDisabled||t.isDisabled,h=o==="vertical",{direction:m}=Wi(),{addGlobalListener:y,removeGlobalListener:v}=py(),S=XN.get(t);var _;const{labelProps:E,fieldProps:$}=gS({...e,id:K9(t,n),"aria-labelledby":`${S.id} ${(_=e["aria-labelledby"])!==null&&_!==void 0?_:""}`.trim()}),k=t.values[n],A=P.useCallback(()=>{l.current&&$u(l.current)},[l]),w=t.focusedThumb===n;P.useEffect(()=>{w&&A()},[w,A]);let R=m==="rtl",F=P.useRef(null),{keyboardProps:I}=kN({onKeyDown(J){let{getThumbMaxValue:te,getThumbMinValue:H,decrementThumb:M,incrementThumb:G,setThumbValue:X,setThumbDragging:B,pageSize:ne}=t;if(!/^(PageUp|PageDown|Home|End)$/.test(J.key)){J.continuePropagation();return}switch(J.preventDefault(),B(n,!0),J.key){case"PageUp":G(n,ne);break;case"PageDown":M(n,ne);break;case"Home":X(n,H(n));break;case"End":X(n,te(n));break}B(n,!1)}}),{moveProps:U}=Vne({onMoveStart(){F.current=null,t.setThumbDragging(n,!0)},onMove({deltaX:J,deltaY:te,pointerType:H,shiftKey:M}){const{getThumbPercent:G,setThumbPercent:X,decrementThumb:B,incrementThumb:ne,step:oe,pageSize:ie}=t;if(!u.current)return;let{width:de,height:ge}=u.current.getBoundingClientRect(),ye=h?ge:de;if(F.current==null&&(F.current=G(n)*ye),H==="keyboard")J>0&&R||J<0&&!R||te>0?B(n,M?ie:oe):ne(n,M?ie:oe);else{let me=h?te:J;(h||R)&&(me=-me),F.current+=me,X(n,so(F.current/ye,0,1))}},onMoveEnd(){t.setThumbDragging(n,!1)}});t.setThumbEditable(n,!d);const{focusableProps:q}=I0(Et(e,{onFocus:()=>t.setFocusedThumb(n),onBlur:()=>t.setFocusedThumb(void 0)}),l);let V=P.useRef(void 0),N=J=>{A(),V.current=J,t.setThumbDragging(n,!0),y(window,"mouseup",j,!1),y(window,"touchend",j,!1),y(window,"pointerup",j,!1)},j=J=>{var te,H;((H=J.pointerId)!==null&&H!==void 0?H:(te=J.changedTouches)===null||te===void 0?void 0:te[0].identifier)===V.current&&(A(),t.setThumbDragging(n,!1),v(window,"mouseup",j,!1),v(window,"touchend",j,!1),v(window,"pointerup",j,!1))},W=t.getThumbPercent(n);(h||m==="rtl")&&(W=1-W);let K=d?{}:Et(I,U,{onMouseDown:J=>{J.button!==0||J.altKey||J.ctrlKey||J.metaKey||N()},onPointerDown:J=>{J.button!==0||J.altKey||J.ctrlKey||J.metaKey||N(J.pointerId)},onTouchStart:J=>{N(J.changedTouches[0].identifier)}});return my(l,t.defaultValues[n],J=>{t.setThumbValue(n,J)}),{inputProps:Et(q,$,{type:"range",tabIndex:d?void 0:0,min:t.getThumbMinValue(n),max:t.getThumbMaxValue(n),step:t.step,value:k,name:s,form:c,disabled:d,"aria-orientation":o,"aria-valuetext":t.getThumbValueLabel(n),"aria-required":r||void 0,"aria-invalid":a||i==="invalid"||void 0,"aria-errormessage":e["aria-errormessage"],"aria-describedby":[S["aria-describedby"],e["aria-describedby"]].filter(Boolean).join(" "),"aria-details":[S["aria-details"],e["aria-details"]].filter(Boolean).join(" "),onChange:J=>{t.setThumbValue(n,parseFloat(J.target.value))}}),thumbProps:{...K,style:{position:"absolute",[h?"top":"left"]:`${W*100}%`,transform:"translate(-50%, -50%)",touchAction:"none"}},labelProps:E,isDragging:t.isThumbDragging(n),isDisabled:d,isFocused:w}}var noe={};noe={Empty:"فارغ"};var roe={};roe={Empty:"Изпразни"};var ioe={};ioe={Empty:"Prázdné"};var aoe={};aoe={Empty:"Tom"};var uoe={};uoe={Empty:"Leer"};var loe={};loe={Empty:"Άδειο"};var ooe={};ooe={Empty:"Empty"};var soe={};soe={Empty:"Vacío"};var coe={};coe={Empty:"Tühjenda"};var foe={};foe={Empty:"Tyhjä"};var doe={};doe={Empty:"Vide"};var hoe={};hoe={Empty:"ריק"};var poe={};poe={Empty:"Prazno"};var moe={};moe={Empty:"Üres"};var goe={};goe={Empty:"Vuoto"};var voe={};voe={Empty:"空"};var yoe={};yoe={Empty:"비어 있음"};var boe={};boe={Empty:"Tuščias"};var Ooe={};Ooe={Empty:"Tukšs"};var xoe={};xoe={Empty:"Tom"};var Soe={};Soe={Empty:"Leeg"};var _oe={};_oe={Empty:"Pusty"};var Eoe={};Eoe={Empty:"Vazio"};var Doe={};Doe={Empty:"Vazio"};var $oe={};$oe={Empty:"Gol"};var Coe={};Coe={Empty:"Не заполнено"};var woe={};woe={Empty:"Prázdne"};var Roe={};Roe={Empty:"Prazen"};var Aoe={};Aoe={Empty:"Prazno"};var Toe={};Toe={Empty:"Tomt"};var koe={};koe={Empty:"Boş"};var Poe={};Poe={Empty:"Пусто"};var Moe={};Moe={Empty:"空"};var Foe={};Foe={Empty:"空白"};var Noe={};Noe={"ar-AE":noe,"bg-BG":roe,"cs-CZ":ioe,"da-DK":aoe,"de-DE":uoe,"el-GR":loe,"en-US":ooe,"es-ES":soe,"et-EE":coe,"fi-FI":foe,"fr-FR":doe,"he-IL":hoe,"hr-HR":poe,"hu-HU":moe,"it-IT":goe,"ja-JP":voe,"ko-KR":yoe,"lt-LT":boe,"lv-LV":Ooe,"nb-NO":xoe,"nl-NL":Soe,"pl-PL":_oe,"pt-BR":Eoe,"pt-PT":Doe,"ro-RO":$oe,"ru-RU":Coe,"sk-SK":woe,"sl-SI":Roe,"sr-SP":Aoe,"sv-SE":Toe,"tr-TR":koe,"uk-UA":Poe,"zh-CN":Moe,"zh-TW":Foe};function lCe(e){return e&&e.__esModule?e.default:e}function YN(e){const t=P.useRef(void 0);let{value:n,textValue:r,minValue:i,maxValue:a,isDisabled:u,isReadOnly:l,isRequired:o,onIncrement:s,onIncrementPage:c,onDecrement:d,onDecrementPage:h,onDecrementToMin:m,onIncrementToMax:y}=e;const v=Qo(lCe(Noe),"@react-aria/spinbutton"),S=()=>clearTimeout(t.current);P.useEffect(()=>()=>S(),[]);let _=q=>{if(!(q.ctrlKey||q.metaKey||q.shiftKey||q.altKey||l||q.nativeEvent.isComposing))switch(q.key){case"PageUp":if(c){q.preventDefault(),c==null||c();break}case"ArrowUp":case"Up":s&&(q.preventDefault(),s==null||s());break;case"PageDown":if(h){q.preventDefault(),h==null||h();break}case"ArrowDown":case"Down":d&&(q.preventDefault(),d==null||d());break;case"Home":m&&(q.preventDefault(),m==null||m());break;case"End":y&&(q.preventDefault(),y==null||y());break}},E=P.useRef(!1),$=()=>{E.current=!0},k=()=>{E.current=!1},A=r===""?v.format("Empty"):(r||`${n}`).replace("-","−");P.useEffect(()=>{E.current&&(C_e("assertive"),Zb(A,"assertive"))},[A]);const w=fi(q=>{S(),s==null||s(),t.current=window.setTimeout(()=>{(a===void 0||isNaN(a)||n===void 0||isNaN(n)||n<a)&&w(60)},q)}),R=fi(q=>{S(),d==null||d(),t.current=window.setTimeout(()=>{(i===void 0||isNaN(i)||n===void 0||isNaN(n)||n>i)&&R(60)},q)});let F=q=>{q.preventDefault()},{addGlobalListener:I,removeAllGlobalListeners:U}=py();return{spinButtonProps:{role:"spinbutton","aria-valuenow":n!==void 0&&!isNaN(n)?n:void 0,"aria-valuetext":A,"aria-valuemin":i,"aria-valuemax":a,"aria-disabled":u||void 0,"aria-readonly":l||void 0,"aria-required":o||void 0,onKeyDown:_,onFocus:$,onBlur:k},incrementButtonProps:{onPressStart:()=>{w(400),I(window,"contextmenu",F)},onPressEnd:()=>{S(),U()},onFocus:$,onBlur:k},decrementButtonProps:{onPressStart:()=>{R(400),I(window,"contextmenu",F)},onPressEnd:()=>{S(),U()},onFocus:$,onBlur:k}}}function oCe(e,t,n){let{isDisabled:r,isReadOnly:i,isRequired:a,isWheelDisabled:u,validationBehavior:l="aria"}=e,{colorValue:o,inputValue:s,increment:c,decrement:d,incrementToMax:h,decrementToMin:m,commit:y}=t,v=vi(),{spinButtonProps:S}=YN({isDisabled:r,isReadOnly:i,isRequired:a,maxValue:16777215,minValue:0,onIncrement:c,onIncrementToMax:h,onDecrement:d,onDecrementToMin:m,value:o?o.toHexInt():void 0,textValue:o?o.toString("hex"):void 0}),[_,E]=P.useState(!1),{focusWithinProps:$}=L0({isDisabled:r,onFocusWithinChange:E}),k=P.useCallback(I=>{Math.abs(I.deltaY)<=Math.abs(I.deltaX)||(I.deltaY>0?c():I.deltaY<0&&d())},[d,c]);Une({onScroll:k,isDisabled:u||r||i||!_},n);let w=I=>{t.validate(I)&&t.setInputValue(I)},{inputProps:R,...F}=Die({...e,id:v,value:s,defaultValue:"!",validate:void 0,[c1]:t,type:"text",autoComplete:"off",onChange:w},t,n);return my(n,t.defaultColorValue,t.setColorValue),R=Et(R,S,$,{role:"textbox","aria-valuemax":null,"aria-valuemin":null,"aria-valuenow":null,"aria-valuetext":null,autoCorrect:"off",spellCheck:"false",onBlur:y}),l==="native"&&(R["aria-required"]=void 0),{inputProps:R,...F}}function sCe(e){return e&&e.__esModule?e.default:e}function cCe(e){let{color:t,colorName:n}=e,r=t||"#fff0",i=P.useMemo(()=>typeof r=="string"?vy(r):r,[r]),{locale:a}=Wi(),u=Gt(e,{labelable:!0}),l=Qo(sCe(vle),"@react-aria/color"),o=vi(e.id);return n||(n=i.getChannelValue("alpha")===0?l.format("transparent"):i.getColorName(a)),{colorSwatchProps:{...u,role:"img","aria-roledescription":l.format("colorSwatch"),"aria-label":[n,e["aria-label"]||""].filter(Boolean).join(", "),"aria-labelledby":e["aria-labelledby"]?`${o} ${e["aria-labelledby"]}`:void 0,id:o,style:{backgroundColor:i.toString("css"),forcedColorAdjust:"none"}},color:i||null}}var Ioe={};Ioe={decrease:e=>`خفض ${e.fieldLabel}`,increase:e=>`زيادة ${e.fieldLabel}`,numberField:"حقل رقمي"};var Loe={};Loe={decrease:e=>`Намаляване ${e.fieldLabel}`,increase:e=>`Усилване ${e.fieldLabel}`,numberField:"Номер на полето"};var Boe={};Boe={decrease:e=>`Snížit ${e.fieldLabel}`,increase:e=>`Zvýšit ${e.fieldLabel}`,numberField:"Číselné pole"};var Qoe={};Qoe={decrease:e=>`Reducer ${e.fieldLabel}`,increase:e=>`Øg ${e.fieldLabel}`,numberField:"Talfelt"};var zoe={};zoe={decrease:e=>`${e.fieldLabel} verringern`,increase:e=>`${e.fieldLabel} erhöhen`,numberField:"Nummernfeld"};var qoe={};qoe={decrease:e=>`Μείωση ${e.fieldLabel}`,increase:e=>`Αύξηση ${e.fieldLabel}`,numberField:"Πεδίο αριθμού"};var Voe={};Voe={decrease:e=>`Decrease ${e.fieldLabel}`,increase:e=>`Increase ${e.fieldLabel}`,numberField:"Number field"};var Uoe={};Uoe={decrease:e=>`Reducir ${e.fieldLabel}`,increase:e=>`Aumentar ${e.fieldLabel}`,numberField:"Campo de número"};var joe={};joe={decrease:e=>`Vähenda ${e.fieldLabel}`,increase:e=>`Suurenda ${e.fieldLabel}`,numberField:"Numbri väli"};var Zoe={};Zoe={decrease:e=>`Vähennä ${e.fieldLabel}`,increase:e=>`Lisää ${e.fieldLabel}`,numberField:"Numerokenttä"};var Hoe={};Hoe={decrease:e=>`Diminuer ${e.fieldLabel}`,increase:e=>`Augmenter ${e.fieldLabel}`,numberField:"Champ de nombre"};var Xoe={};Xoe={decrease:e=>`הקטן ${e.fieldLabel}`,increase:e=>`הגדל ${e.fieldLabel}`,numberField:"שדה מספר"};var Yoe={};Yoe={decrease:e=>`Smanji ${e.fieldLabel}`,increase:e=>`Povećaj ${e.fieldLabel}`,numberField:"Polje broja"};var Goe={};Goe={decrease:e=>`${e.fieldLabel} csökkentése`,increase:e=>`${e.fieldLabel} növelése`,numberField:"Számmező"};var Koe={};Koe={decrease:e=>`Riduci ${e.fieldLabel}`,increase:e=>`Aumenta ${e.fieldLabel}`,numberField:"Campo numero"};var Woe={};Woe={decrease:e=>`${e.fieldLabel}を縮小`,increase:e=>`${e.fieldLabel}を拡大`,numberField:"数値フィールド"};var Joe={};Joe={decrease:e=>`${e.fieldLabel} 감소`,increase:e=>`${e.fieldLabel} 증가`,numberField:"번호 필드"};var ese={};ese={decrease:e=>`Sumažinti ${e.fieldLabel}`,increase:e=>`Padidinti ${e.fieldLabel}`,numberField:"Numerio laukas"};var tse={};tse={decrease:e=>`Samazināšana ${e.fieldLabel}`,increase:e=>`Palielināšana ${e.fieldLabel}`,numberField:"Skaitļu lauks"};var nse={};nse={decrease:e=>`Reduser ${e.fieldLabel}`,increase:e=>`Øk ${e.fieldLabel}`,numberField:"Tallfelt"};var rse={};rse={decrease:e=>`${e.fieldLabel} verlagen`,increase:e=>`${e.fieldLabel} verhogen`,numberField:"Getalveld"};var ise={};ise={decrease:e=>`Zmniejsz ${e.fieldLabel}`,increase:e=>`Zwiększ ${e.fieldLabel}`,numberField:"Pole numeru"};var ase={};ase={decrease:e=>`Diminuir ${e.fieldLabel}`,increase:e=>`Aumentar ${e.fieldLabel}`,numberField:"Campo de número"};var use={};use={decrease:e=>`Diminuir ${e.fieldLabel}`,increase:e=>`Aumentar ${e.fieldLabel}`,numberField:"Campo numérico"};var lse={};lse={decrease:e=>`Scădere ${e.fieldLabel}`,increase:e=>`Creștere ${e.fieldLabel}`,numberField:"Câmp numeric"};var ose={};ose={decrease:e=>`Уменьшение ${e.fieldLabel}`,increase:e=>`Увеличение ${e.fieldLabel}`,numberField:"Числовое поле"};var sse={};sse={decrease:e=>`Znížiť ${e.fieldLabel}`,increase:e=>`Zvýšiť ${e.fieldLabel}`,numberField:"Číselné pole"};var cse={};cse={decrease:e=>`Upadati ${e.fieldLabel}`,increase:e=>`Povečajte ${e.fieldLabel}`,numberField:"Številčno polje"};var fse={};fse={decrease:e=>`Smanji ${e.fieldLabel}`,increase:e=>`Povećaj ${e.fieldLabel}`,numberField:"Polje broja"};var dse={};dse={decrease:e=>`Minska ${e.fieldLabel}`,increase:e=>`Öka ${e.fieldLabel}`,numberField:"Nummerfält"};var hse={};hse={decrease:e=>`${e.fieldLabel} azalt`,increase:e=>`${e.fieldLabel} arttır`,numberField:"Sayı alanı"};var pse={};pse={decrease:e=>`Зменшити ${e.fieldLabel}`,increase:e=>`Збільшити ${e.fieldLabel}`,numberField:"Поле номера"};var mse={};mse={decrease:e=>`降低 ${e.fieldLabel}`,increase:e=>`提高 ${e.fieldLabel}`,numberField:"数字字段"};var gse={};gse={decrease:e=>`縮小 ${e.fieldLabel}`,increase:e=>`放大 ${e.fieldLabel}`,numberField:"數字欄位"};var vse={};vse={"ar-AE":Ioe,"bg-BG":Loe,"cs-CZ":Boe,"da-DK":Qoe,"de-DE":zoe,"el-GR":qoe,"en-US":Voe,"es-ES":Uoe,"et-EE":joe,"fi-FI":Zoe,"fr-FR":Hoe,"he-IL":Xoe,"hr-HR":Yoe,"hu-HU":Goe,"it-IT":Koe,"ja-JP":Woe,"ko-KR":Joe,"lt-LT":ese,"lv-LV":tse,"nb-NO":nse,"nl-NL":rse,"pl-PL":ise,"pt-BR":ase,"pt-PT":use,"ro-RO":lse,"ru-RU":ose,"sk-SK":sse,"sl-SI":cse,"sr-SP":fse,"sv-SE":dse,"tr-TR":hse,"uk-UA":pse,"zh-CN":mse,"zh-TW":gse};function fCe(e){return e&&e.__esModule?e.default:e}function yse(e,t,n){let{id:r,decrementAriaLabel:i,incrementAriaLabel:a,isDisabled:u,isReadOnly:l,isRequired:o,minValue:s,maxValue:c,autoFocus:d,label:h,formatOptions:m,onBlur:y=()=>{},onFocus:v,onFocusChange:S,onKeyDown:_,onKeyUp:E,description:$,errorMessage:k,isWheelDisabled:A,...w}=e,{increment:R,incrementToMax:F,decrement:I,decrementToMin:U,numberValue:q,inputValue:V,commit:N,commitValidation:j}=t;const W=Qo(fCe(vse),"@react-aria/numberfield");let K=vi(r),{focusProps:J}=XC({onBlur(){N()}}),te=t$(m),H=P.useMemo(()=>te.resolvedOptions(),[te]),M=t$({...m,currencySign:void 0}),G=P.useMemo(()=>isNaN(q)?"":M.format(q),[M,q]),{spinButtonProps:X,incrementButtonProps:B,decrementButtonProps:ne}=YN({isDisabled:u,isReadOnly:l,isRequired:o,maxValue:c,minValue:s,onIncrement:R,onIncrementToMax:F,onDecrement:I,onDecrementToMin:U,value:q,textValue:G}),[oe,ie]=P.useState(!1),{focusWithinProps:de}=L0({isDisabled:u,onFocusWithinChange:ie}),ge=P.useCallback(zt=>{Math.abs(zt.deltaY)<=Math.abs(zt.deltaX)||(zt.deltaY>0?R():zt.deltaY<0&&I())},[I,R]);Une({onScroll:ge,isDisabled:A||u||l||!oe},n);var me;let Ee=((me=H.maximumFractionDigits)!==null&&me!==void 0?me:0)>0,se=t.minValue===void 0||isNaN(t.minValue)||t.minValue<0,pe="numeric";tte()?se?pe="text":Ee&&(pe="decimal"):fS()&&(se?pe="numeric":Ee&&(pe="decimal"));let ve=zt=>{t.validate(zt)&&t.setInputValue(zt)},we=Gt(e),Me=P.useCallback(zt=>{zt.nativeEvent.isComposing||(zt.key==="Enter"?(N(),j()):zt.continuePropagation())},[N,j]),{isInvalid:Re,validationErrors:et,validationDetails:tt}=t.displayValidation,{labelProps:Oe,inputProps:$e,descriptionProps:Ae,errorMessageProps:He}=Die({...w,...we,name:void 0,form:void 0,label:h,autoFocus:d,isDisabled:u,isReadOnly:l,isRequired:o,validate:void 0,[c1]:t,value:V,defaultValue:"!",autoComplete:"off","aria-label":e["aria-label"]||void 0,"aria-labelledby":e["aria-labelledby"]||void 0,id:K,type:"text",inputMode:pe,onChange:ve,onBlur:y,onFocus:v,onFocusChange:S,onKeyDown:P.useMemo(()=>ih(Me,_),[Me,_]),onKeyUp:E,description:$,errorMessage:k},t,n);my(n,t.defaultNumberValue,t.setNumberValue);let je=Et(X,J,$e,{role:null,"aria-roledescription":$1()?null:W.format("numberField"),"aria-valuemax":null,"aria-valuemin":null,"aria-valuenow":null,"aria-valuetext":null,autoCorrect:"off",spellCheck:"false"});e.validationBehavior==="native"&&(je["aria-required"]=void 0);let Ye=zt=>{var qt;document.activeElement!==n.current&&(zt.pointerType==="mouse"?(qt=n.current)===null||qt===void 0||qt.focus():zt.target.focus())},Ne=e["aria-label"]||(typeof e.label=="string"?e.label:""),ze;Ne||(ze=e.label!=null?Oe.id:e["aria-labelledby"]);let at=vi(),mt=vi(),Mt=Et(B,{"aria-label":a||W.format("increase",{fieldLabel:Ne}).trim(),id:ze&&!a?at:null,"aria-labelledby":ze&&!a?`${at} ${ze}`:null,"aria-controls":K,excludeFromTabOrder:!0,preventFocusOnPress:!0,allowFocusWhenDisabled:!0,isDisabled:!t.canIncrement,onPressStart:Ye}),bt=Et(ne,{"aria-label":i||W.format("decrease",{fieldLabel:Ne}).trim(),id:ze&&!i?mt:null,"aria-labelledby":ze&&!i?`${mt} ${ze}`:null,"aria-controls":K,excludeFromTabOrder:!0,preventFocusOnPress:!0,allowFocusWhenDisabled:!0,isDisabled:!t.canDecrement,onPressStart:Ye});return{groupProps:{...de,role:"group","aria-disabled":u,"aria-invalid":Re?"true":void 0},labelProps:Oe,inputProps:je,incrementButtonProps:Mt,decrementButtonProps:bt,errorMessageProps:He,descriptionProps:Ae,isInvalid:Re,validationErrors:et,validationDetails:tt}}function dCe(e,t,n){let{locale:r}=Wi();return yse({...e,value:void 0,defaultValue:void 0,onChange:void 0,validate:void 0,"aria-label":e["aria-label"]||(e.label||e["aria-labelledby"]?void 0:t.colorValue.getChannelName(e.channel,r))},t,n)}var bse={};bse={calendar:"التقويم",day:"يوم",dayPeriod:"ص/م",endDate:"تاريخ الانتهاء",era:"العصر",hour:"الساعات",minute:"الدقائق",month:"الشهر",second:"الثواني",selectedDateDescription:e=>`تاريخ محدد: ${e.date}`,selectedRangeDescription:e=>`المدى الزمني المحدد: ${e.startDate} إلى ${e.endDate}`,selectedTimeDescription:e=>`الوقت المحدد: ${e.time}`,startDate:"تاريخ البدء",timeZoneName:"التوقيت",weekday:"اليوم",year:"السنة"};var Ose={};Ose={calendar:"Календар",day:"ден",dayPeriod:"пр.об./сл.об.",endDate:"Крайна дата",era:"ера",hour:"час",minute:"минута",month:"месец",second:"секунда",selectedDateDescription:e=>`Избрана дата: ${e.date}`,selectedRangeDescription:e=>`Избран диапазон: ${e.startDate} до ${e.endDate}`,selectedTimeDescription:e=>`Избрано време: ${e.time}`,startDate:"Начална дата",timeZoneName:"часова зона",weekday:"ден от седмицата",year:"година"};var xse={};xse={calendar:"Kalendář",day:"den",dayPeriod:"část dne",endDate:"Konečné datum",era:"letopočet",hour:"hodina",minute:"minuta",month:"měsíc",second:"sekunda",selectedDateDescription:e=>`Vybrané datum: ${e.date}`,selectedRangeDescription:e=>`Vybrané období: ${e.startDate} až ${e.endDate}`,selectedTimeDescription:e=>`Vybraný čas: ${e.time}`,startDate:"Počáteční datum",timeZoneName:"časové pásmo",weekday:"den v týdnu",year:"rok"};var Sse={};Sse={calendar:"Kalender",day:"dag",dayPeriod:"AM/PM",endDate:"Slutdato",era:"æra",hour:"time",minute:"minut",month:"måned",second:"sekund",selectedDateDescription:e=>`Valgt dato: ${e.date}`,selectedRangeDescription:e=>`Valgt interval: ${e.startDate} til ${e.endDate}`,selectedTimeDescription:e=>`Valgt tidspunkt: ${e.time}`,startDate:"Startdato",timeZoneName:"tidszone",weekday:"ugedag",year:"år"};var _se={};_se={calendar:"Kalender",day:"Tag",dayPeriod:"Tageshälfte",endDate:"Enddatum",era:"Epoche",hour:"Stunde",minute:"Minute",month:"Monat",second:"Sekunde",selectedDateDescription:e=>`Ausgewähltes Datum: ${e.date}`,selectedRangeDescription:e=>`Ausgewählter Bereich: ${e.startDate} bis ${e.endDate}`,selectedTimeDescription:e=>`Ausgewählte Zeit: ${e.time}`,startDate:"Startdatum",timeZoneName:"Zeitzone",weekday:"Wochentag",year:"Jahr"};var Ese={};Ese={calendar:"Ημερολόγιο",day:"ημέρα",dayPeriod:"π.μ./μ.μ.",endDate:"Ημερομηνία λήξης",era:"περίοδος",hour:"ώρα",minute:"λεπτό",month:"μήνας",second:"δευτερόλεπτο",selectedDateDescription:e=>`Επιλεγμένη ημερομηνία: ${e.date}`,selectedRangeDescription:e=>`Επιλεγμένο εύρος: ${e.startDate} έως ${e.endDate}`,selectedTimeDescription:e=>`Επιλεγμένη ώρα: ${e.time}`,startDate:"Ημερομηνία έναρξης",timeZoneName:"ζώνη ώρας",weekday:"καθημερινή",year:"έτος"};var Dse={};Dse={era:"era",year:"year",month:"month",day:"day",hour:"hour",minute:"minute",second:"second",dayPeriod:"AM/PM",calendar:"Calendar",startDate:"Start Date",endDate:"End Date",weekday:"day of the week",timeZoneName:"time zone",selectedDateDescription:e=>`Selected Date: ${e.date}`,selectedRangeDescription:e=>`Selected Range: ${e.startDate} to ${e.endDate}`,selectedTimeDescription:e=>`Selected Time: ${e.time}`};var $se={};$se={calendar:"Calendario",day:"día",dayPeriod:"a. m./p. m.",endDate:"Fecha final",era:"era",hour:"hora",minute:"minuto",month:"mes",second:"segundo",selectedDateDescription:e=>`Fecha seleccionada: ${e.date}`,selectedRangeDescription:e=>`Rango seleccionado: ${e.startDate} a ${e.endDate}`,selectedTimeDescription:e=>`Hora seleccionada: ${e.time}`,startDate:"Fecha de inicio",timeZoneName:"zona horaria",weekday:"día de la semana",year:"año"};var Cse={};Cse={calendar:"Kalender",day:"päev",dayPeriod:"enne/pärast lõunat",endDate:"Lõppkuupäev",era:"ajastu",hour:"tund",minute:"minut",month:"kuu",second:"sekund",selectedDateDescription:e=>`Valitud kuupäev: ${e.date}`,selectedRangeDescription:e=>`Valitud vahemik: ${e.startDate} kuni ${e.endDate}`,selectedTimeDescription:e=>`Valitud aeg: ${e.time}`,startDate:"Alguskuupäev",timeZoneName:"ajavöönd",weekday:"nädalapäev",year:"aasta"};var wse={};wse={calendar:"Kalenteri",day:"päivä",dayPeriod:"vuorokaudenaika",endDate:"Päättymispäivä",era:"aikakausi",hour:"tunti",minute:"minuutti",month:"kuukausi",second:"sekunti",selectedDateDescription:e=>`Valittu päivämäärä: ${e.date}`,selectedRangeDescription:e=>`Valittu aikaväli: ${e.startDate} – ${e.endDate}`,selectedTimeDescription:e=>`Valittu aika: ${e.time}`,startDate:"Alkamispäivä",timeZoneName:"aikavyöhyke",weekday:"viikonpäivä",year:"vuosi"};var Rse={};Rse={calendar:"Calendrier",day:"jour",dayPeriod:"cadran",endDate:"Date de fin",era:"ère",hour:"heure",minute:"minute",month:"mois",second:"seconde",selectedDateDescription:e=>`Date sélectionnée : ${e.date}`,selectedRangeDescription:e=>`Plage sélectionnée : ${e.startDate} au ${e.endDate}`,selectedTimeDescription:e=>`Heure choisie : ${e.time}`,startDate:"Date de début",timeZoneName:"fuseau horaire",weekday:"jour de la semaine",year:"année"};var Ase={};Ase={calendar:"לוח שנה",day:"יום",dayPeriod:"לפנה״צ/אחה״צ",endDate:"תאריך סיום",era:"תקופה",hour:"שעה",minute:"דקה",month:"חודש",second:"שנייה",selectedDateDescription:e=>`תאריך נבחר: ${e.date}`,selectedRangeDescription:e=>`טווח נבחר: ${e.startDate} עד ${e.endDate}`,selectedTimeDescription:e=>`זמן נבחר: ${e.time}`,startDate:"תאריך התחלה",timeZoneName:"אזור זמן",weekday:"יום בשבוע",year:"שנה"};var Tse={};Tse={calendar:"Kalendar",day:"dan",dayPeriod:"AM/PM",endDate:"Datum završetka",era:"era",hour:"sat",minute:"minuta",month:"mjesec",second:"sekunda",selectedDateDescription:e=>`Odabrani datum: ${e.date}`,selectedRangeDescription:e=>`Odabrani raspon: ${e.startDate} do ${e.endDate}`,selectedTimeDescription:e=>`Odabrano vrijeme: ${e.time}`,startDate:"Datum početka",timeZoneName:"vremenska zona",weekday:"dan u tjednu",year:"godina"};var kse={};kse={calendar:"Naptár",day:"nap",dayPeriod:"napszak",endDate:"Befejező dátum",era:"éra",hour:"óra",minute:"perc",month:"hónap",second:"másodperc",selectedDateDescription:e=>`Kijelölt dátum: ${e.date}`,selectedRangeDescription:e=>`Kijelölt tartomány: ${e.startDate}–${e.endDate}`,selectedTimeDescription:e=>`Kijelölt idő: ${e.time}`,startDate:"Kezdő dátum",timeZoneName:"időzóna",weekday:"hét napja",year:"év"};var Pse={};Pse={calendar:"Calendario",day:"giorno",dayPeriod:"AM/PM",endDate:"Data finale",era:"era",hour:"ora",minute:"minuto",month:"mese",second:"secondo",selectedDateDescription:e=>`Data selezionata: ${e.date}`,selectedRangeDescription:e=>`Intervallo selezionato: da ${e.startDate} a ${e.endDate}`,selectedTimeDescription:e=>`Ora selezionata: ${e.time}`,startDate:"Data iniziale",timeZoneName:"fuso orario",weekday:"giorno della settimana",year:"anno"};var Mse={};Mse={calendar:"カレンダー",day:"日",dayPeriod:"午前/午後",endDate:"終了日",era:"時代",hour:"時",minute:"分",month:"月",second:"秒",selectedDateDescription:e=>`選択した日付 : ${e.date}`,selectedRangeDescription:e=>`選択範囲 : ${e.startDate} から ${e.endDate}`,selectedTimeDescription:e=>`選択した時間 : ${e.time}`,startDate:"開始日",timeZoneName:"タイムゾーン",weekday:"曜日",year:"年"};var Fse={};Fse={calendar:"달력",day:"일",dayPeriod:"오전/오후",endDate:"종료일",era:"연호",hour:"시",minute:"분",month:"월",second:"초",selectedDateDescription:e=>`선택 일자: ${e.date}`,selectedRangeDescription:e=>`선택 범위: ${e.startDate} ~ ${e.endDate}`,selectedTimeDescription:e=>`선택 시간: ${e.time}`,startDate:"시작일",timeZoneName:"시간대",weekday:"요일",year:"년"};var Nse={};Nse={calendar:"Kalendorius",day:"diena",dayPeriod:"iki pietų / po pietų",endDate:"Pabaigos data",era:"era",hour:"valanda",minute:"minutė",month:"mėnuo",second:"sekundė",selectedDateDescription:e=>`Pasirinkta data: ${e.date}`,selectedRangeDescription:e=>`Pasirinktas intervalas: nuo ${e.startDate} iki ${e.endDate}`,selectedTimeDescription:e=>`Pasirinktas laikas: ${e.time}`,startDate:"Pradžios data",timeZoneName:"laiko juosta",weekday:"savaitės diena",year:"metai"};var Ise={};Ise={calendar:"Kalendārs",day:"diena",dayPeriod:"priekšpusdienā/pēcpusdienā",endDate:"Beigu datums",era:"ēra",hour:"stundas",minute:"minūtes",month:"mēnesis",second:"sekundes",selectedDateDescription:e=>`Atlasītais datums: ${e.date}`,selectedRangeDescription:e=>`Atlasītais diapazons: no ${e.startDate} līdz ${e.endDate}`,selectedTimeDescription:e=>`Atlasītais laiks: ${e.time}`,startDate:"Sākuma datums",timeZoneName:"laika josla",weekday:"nedēļas diena",year:"gads"};var Lse={};Lse={calendar:"Kalender",day:"dag",dayPeriod:"a.m./p.m.",endDate:"Sluttdato",era:"tidsalder",hour:"time",minute:"minutt",month:"måned",second:"sekund",selectedDateDescription:e=>`Valgt dato: ${e.date}`,selectedRangeDescription:e=>`Valgt område: ${e.startDate} til ${e.endDate}`,selectedTimeDescription:e=>`Valgt tid: ${e.time}`,startDate:"Startdato",timeZoneName:"tidssone",weekday:"ukedag",year:"år"};var Bse={};Bse={calendar:"Kalender",day:"dag",dayPeriod:"a.m./p.m.",endDate:"Einddatum",era:"tijdperk",hour:"uur",minute:"minuut",month:"maand",second:"seconde",selectedDateDescription:e=>`Geselecteerde datum: ${e.date}`,selectedRangeDescription:e=>`Geselecteerd bereik: ${e.startDate} tot ${e.endDate}`,selectedTimeDescription:e=>`Geselecteerde tijd: ${e.time}`,startDate:"Startdatum",timeZoneName:"tijdzone",weekday:"dag van de week",year:"jaar"};var Qse={};Qse={calendar:"Kalendarz",day:"dzień",dayPeriod:"rano / po południu / wieczorem",endDate:"Data końcowa",era:"era",hour:"godzina",minute:"minuta",month:"miesiąc",second:"sekunda",selectedDateDescription:e=>`Wybrana data: ${e.date}`,selectedRangeDescription:e=>`Wybrany zakres: ${e.startDate} do ${e.endDate}`,selectedTimeDescription:e=>`Wybrany czas: ${e.time}`,startDate:"Data początkowa",timeZoneName:"strefa czasowa",weekday:"dzień tygodnia",year:"rok"};var zse={};zse={calendar:"Calendário",day:"dia",dayPeriod:"AM/PM",endDate:"Data final",era:"era",hour:"hora",minute:"minuto",month:"mês",second:"segundo",selectedDateDescription:e=>`Data selecionada: ${e.date}`,selectedRangeDescription:e=>`Intervalo selecionado: ${e.startDate} a ${e.endDate}`,selectedTimeDescription:e=>`Hora selecionada: ${e.time}`,startDate:"Data inicial",timeZoneName:"fuso horário",weekday:"dia da semana",year:"ano"};var qse={};qse={calendar:"Calendário",day:"dia",dayPeriod:"am/pm",endDate:"Data de Término",era:"era",hour:"hora",minute:"minuto",month:"mês",second:"segundo",selectedDateDescription:e=>`Data selecionada: ${e.date}`,selectedRangeDescription:e=>`Intervalo selecionado: ${e.startDate} a ${e.endDate}`,selectedTimeDescription:e=>`Hora selecionada: ${e.time}`,startDate:"Data de Início",timeZoneName:"fuso horário",weekday:"dia da semana",year:"ano"};var Vse={};Vse={calendar:"Calendar",day:"zi",dayPeriod:"a.m/p.m.",endDate:"Dată final",era:"eră",hour:"oră",minute:"minut",month:"lună",second:"secundă",selectedDateDescription:e=>`Dată selectată: ${e.date}`,selectedRangeDescription:e=>`Interval selectat: de la ${e.startDate} până la ${e.endDate}`,selectedTimeDescription:e=>`Ora selectată: ${e.time}`,startDate:"Dată început",timeZoneName:"fus orar",weekday:"ziua din săptămână",year:"an"};var Use={};Use={calendar:"Календарь",day:"день",dayPeriod:"AM/PM",endDate:"Дата окончания",era:"эра",hour:"час",minute:"минута",month:"месяц",second:"секунда",selectedDateDescription:e=>`Выбранная дата: ${e.date}`,selectedRangeDescription:e=>`Выбранный диапазон: с ${e.startDate} по ${e.endDate}`,selectedTimeDescription:e=>`Выбранное время: ${e.time}`,startDate:"Дата начала",timeZoneName:"часовой пояс",weekday:"день недели",year:"год"};var jse={};jse={calendar:"Kalendár",day:"deň",dayPeriod:"AM/PM",endDate:"Dátum ukončenia",era:"letopočet",hour:"hodina",minute:"minúta",month:"mesiac",second:"sekunda",selectedDateDescription:e=>`Vybratý dátum: ${e.date}`,selectedRangeDescription:e=>`Vybratý rozsah: od ${e.startDate} do ${e.endDate}`,selectedTimeDescription:e=>`Vybratý čas: ${e.time}`,startDate:"Dátum začatia",timeZoneName:"časové pásmo",weekday:"deň týždňa",year:"rok"};var Zse={};Zse={calendar:"Koledar",day:"dan",dayPeriod:"dop/pop",endDate:"Datum konca",era:"doba",hour:"ura",minute:"minuta",month:"mesec",second:"sekunda",selectedDateDescription:e=>`Izbrani datum: ${e.date}`,selectedRangeDescription:e=>`Izbrano območje: ${e.startDate} do ${e.endDate}`,selectedTimeDescription:e=>`Izbrani čas: ${e.time}`,startDate:"Datum začetka",timeZoneName:"časovni pas",weekday:"dan v tednu",year:"leto"};var Hse={};Hse={calendar:"Kalendar",day:"дан",dayPeriod:"пре подне/по подне",endDate:"Datum završetka",era:"ера",hour:"сат",minute:"минут",month:"месец",second:"секунд",selectedDateDescription:e=>`Izabrani datum: ${e.date}`,selectedRangeDescription:e=>`Izabrani opseg: od ${e.startDate} do ${e.endDate}`,selectedTimeDescription:e=>`Izabrano vreme: ${e.time}`,startDate:"Datum početka",timeZoneName:"временска зона",weekday:"дан у недељи",year:"година"};var Xse={};Xse={calendar:"Kalender",day:"dag",dayPeriod:"fm/em",endDate:"Slutdatum",era:"era",hour:"timme",minute:"minut",month:"månad",second:"sekund",selectedDateDescription:e=>`Valt datum: ${e.date}`,selectedRangeDescription:e=>`Valt intervall: ${e.startDate} till ${e.endDate}`,selectedTimeDescription:e=>`Vald tid: ${e.time}`,startDate:"Startdatum",timeZoneName:"tidszon",weekday:"veckodag",year:"år"};var Yse={};Yse={calendar:"Takvim",day:"gün",dayPeriod:"ÖÖ/ÖS",endDate:"Bitiş Tarihi",era:"çağ",hour:"saat",minute:"dakika",month:"ay",second:"saniye",selectedDateDescription:e=>`Seçilen Tarih: ${e.date}`,selectedRangeDescription:e=>`Seçilen Aralık: ${e.startDate} - ${e.endDate}`,selectedTimeDescription:e=>`Seçilen Zaman: ${e.time}`,startDate:"Başlangıç Tarihi",timeZoneName:"saat dilimi",weekday:"haftanın günü",year:"yıl"};var Gse={};Gse={calendar:"Календар",day:"день",dayPeriod:"дп/пп",endDate:"Дата завершення",era:"ера",hour:"година",minute:"хвилина",month:"місяць",second:"секунда",selectedDateDescription:e=>`Вибрана дата: ${e.date}`,selectedRangeDescription:e=>`Вибраний діапазон: ${e.startDate} — ${e.endDate}`,selectedTimeDescription:e=>`Вибраний час: ${e.time}`,startDate:"Дата початку",timeZoneName:"часовий пояс",weekday:"день тижня",year:"рік"};var Kse={};Kse={calendar:"日历",day:"日",dayPeriod:"上午/下午",endDate:"结束日期",era:"纪元",hour:"小时",minute:"分钟",month:"月",second:"秒",selectedDateDescription:e=>`选定的日期：${e.date}`,selectedRangeDescription:e=>`选定的范围：${e.startDate} 至 ${e.endDate}`,selectedTimeDescription:e=>`选定的时间：${e.time}`,startDate:"开始日期",timeZoneName:"时区",weekday:"工作日",year:"年"};var Wse={};Wse={calendar:"日曆",day:"日",dayPeriod:"上午/下午",endDate:"結束日期",era:"纪元",hour:"小时",minute:"分钟",month:"月",second:"秒",selectedDateDescription:e=>`選定的日期：${e.date}`,selectedRangeDescription:e=>`選定的範圍：${e.startDate} 至 ${e.endDate}`,selectedTimeDescription:e=>`選定的時間：${e.time}`,startDate:"開始日期",timeZoneName:"时区",weekday:"工作日",year:"年"};var GN={};GN={"ar-AE":bse,"bg-BG":Ose,"cs-CZ":xse,"da-DK":Sse,"de-DE":_se,"el-GR":Ese,"en-US":Dse,"es-ES":$se,"et-EE":Cse,"fi-FI":wse,"fr-FR":Rse,"he-IL":Ase,"hr-HR":Tse,"hu-HU":kse,"it-IT":Pse,"ja-JP":Mse,"ko-KR":Fse,"lt-LT":Nse,"lv-LV":Ise,"nb-NO":Lse,"nl-NL":Bse,"pl-PL":Qse,"pt-BR":zse,"pt-PT":qse,"ro-RO":Vse,"ru-RU":Use,"sk-SK":jse,"sl-SI":Zse,"sr-SP":Hse,"sv-SE":Xse,"tr-TR":Yse,"uk-UA":Gse,"zh-CN":Kse,"zh-TW":Wse};function hCe(e,t,n){let{direction:r}=Wi(),i=P.useMemo(()=>MN(t),[t]),a=o=>{if(o.currentTarget.contains(o.target)&&(o.altKey&&(o.key==="ArrowDown"||o.key==="ArrowUp")&&"setOpen"in e&&(o.preventDefault(),o.stopPropagation(),e.setOpen(!0)),!n))switch(o.key){case"ArrowLeft":if(o.preventDefault(),o.stopPropagation(),r==="rtl"){if(t.current){let s=o.target,c=Bj(t.current,s.getBoundingClientRect().left,-1);c&&c.focus()}}else i.focusPrevious();break;case"ArrowRight":if(o.preventDefault(),o.stopPropagation(),r==="rtl"){if(t.current){let s=o.target,c=Bj(t.current,s.getBoundingClientRect().left,1);c&&c.focus()}}else i.focusNext();break}},u=()=>{var o;if(!t.current)return;let s=(o=window.event)===null||o===void 0?void 0:o.target,c=Pl(t.current,{tabbable:!0});if(s&&(c.currentNode=s,s=c.previousNode()),!s){let d;do d=c.lastChild(),d&&(s=d);while(d)}for(;s!=null&&s.hasAttribute("data-placeholder");){let d=c.previousNode();if(d&&d.hasAttribute("data-placeholder"))s=d;else break}s&&s.focus()},{pressProps:l}=lp({preventFocusOnPress:!0,allowTextSelectionOnPress:!0,onPressStart(o){o.pointerType==="mouse"&&u()},onPress(o){(o.pointerType==="touch"||o.pointerType==="pen")&&u()}});return Et(l,{onKeyDown:a})}function Bj(e,t,n){let r=Pl(e,{tabbable:!0}),i=r.nextNode(),a=null,u=1/0;for(;i;){let o=i.getBoundingClientRect().left-t,s=Math.abs(o);Math.sign(o)===n&&s<u&&(a=i,u=s),i=r.nextNode()}return a}function pCe(e){return e&&e.__esModule?e.default:e}const Jse=new WeakMap,gk="__role_"+Date.now(),mCe="__focusManager_"+Date.now();function KN(e,t,n){var r;let{isInvalid:i,validationErrors:a,validationDetails:u}=t.displayValidation,{labelProps:l,fieldProps:o,descriptionProps:s,errorMessageProps:c}=vS({...e,labelElementType:"span",isInvalid:i,errorMessage:e.errorMessage||a}),d=P.useRef(null),{focusWithinProps:h}=L0({...e,onFocusWithin(U){var q;d.current=t.value,(q=e.onFocus)===null||q===void 0||q.call(e,U)},onBlurWithin:U=>{var q;t.confirmPlaceholder(),t.value!==d.current&&t.commitValidation(),(q=e.onBlur)===null||q===void 0||q.call(e,U)},onFocusWithinChange:e.onFocusChange}),m=Qo(pCe(GN),"@react-aria/datepicker"),y=t.maxGranularity==="hour"?"selectedTimeDescription":"selectedDateDescription",v=t.maxGranularity==="hour"?"time":"date",S=t.value?m.format(y,{[v]:t.formatValue({month:"long"})}):"",_=zC(S),E=e[gk]==="presentation"?o["aria-describedby"]:[_["aria-describedby"],o["aria-describedby"]].filter(Boolean).join(" ")||void 0,$=e[mCe],k=P.useMemo(()=>$||MN(n),[$,n]),A=hCe(t,n,e[gk]==="presentation");Jse.set(t,{ariaLabel:e["aria-label"],ariaLabelledBy:[l.id,e["aria-labelledby"]].filter(Boolean).join(" ")||void 0,ariaDescribedBy:E,focusManager:k});let w=P.useRef(e.autoFocus),R;e[gk]==="presentation"?R={role:"presentation"}:R=Et(o,{role:"group","aria-disabled":e.isDisabled||void 0,"aria-describedby":E}),P.useEffect(()=>{w.current&&k.focusFirst(),w.current=!1},[k]),my(e.inputRef,t.defaultValue,t.setValue),bS({...e,focus(){k.focusFirst()}},t,e.inputRef);let F={type:"hidden",name:e.name,form:e.form,value:((r=t.value)===null||r===void 0?void 0:r.toString())||"",disabled:e.isDisabled};e.validationBehavior==="native"&&(F.type="text",F.hidden=!0,F.required=e.isRequired,F.onChange=()=>{});let I=Gt(e);return{labelProps:{...l,onClick:()=>{k.focusFirst()}},fieldProps:Et(I,R,A,h,{onKeyDown(U){e.onKeyDown&&e.onKeyDown(U)},onKeyUp(U){e.onKeyUp&&e.onKeyUp(U)},style:{unicodeBidi:"isolate"}}),inputProps:F,descriptionProps:s,errorMessageProps:c,isInvalid:i,validationErrors:a,validationDetails:u}}function gCe(e,t,n){var r;let i=KN(e,t,n);return i.inputProps.value=((r=t.timeValue)===null||r===void 0?void 0:r.toString())||"",i}function vCe(e){return e&&e.__esModule?e.default:e}function yCe(){let{locale:e}=Wi(),t=Jte(vCe(GN),"@react-aria/datepicker");return P.useMemo(()=>{try{return new Intl.DisplayNames(e,{type:"dateTimeField"})}catch{return new bCe(e,t)}},[e,t])}class bCe{of(t){return this.dictionary.getStringForLocale(t,this.locale)}constructor(t,n){this.locale=t,this.dictionary=n}}function OCe(e,t,n){let r=P.useRef(""),{locale:i,direction:a}=Wi(),u=yCe(),{ariaLabel:l,ariaLabelledBy:o,ariaDescribedBy:s,focusManager:c}=Jse.get(t),d=e.isPlaceholder?"":e.text,h=P.useMemo(()=>t.dateFormatter.resolvedOptions(),[t.dateFormatter]),m=Z_({month:"long",timeZone:h.timeZone}),y=Z_({hour:"numeric",hour12:h.hour12,timeZone:h.timeZone});if(e.type==="month"&&!e.isPlaceholder){let M=m.format(t.dateValue);d=M!==d?`${d} – ${M}`:M}else e.type==="hour"&&!e.isPlaceholder&&(d=y.format(t.dateValue));let{spinButtonProps:v}=YN({value:e.value,textValue:d,minValue:e.minValue,maxValue:e.maxValue,isDisabled:t.isDisabled,isReadOnly:t.isReadOnly||!e.isEditable,isRequired:t.isRequired,onIncrement:()=>{r.current="",t.increment(e.type)},onDecrement:()=>{r.current="",t.decrement(e.type)},onIncrementPage:()=>{r.current="",t.incrementPage(e.type)},onDecrementPage:()=>{r.current="",t.decrementPage(e.type)},onIncrementToMax:()=>{r.current="",e.maxValue!==void 0&&t.setSegment(e.type,e.maxValue)},onDecrementToMin:()=>{r.current="",e.minValue!==void 0&&t.setSegment(e.type,e.minValue)}}),S=P.useMemo(()=>new CN(i,{maximumFractionDigits:0}),[i]),_=()=>{if(e.text===e.placeholder&&c.focusPrevious(),S.isValidPartialNumber(e.text)&&!t.isReadOnly&&!e.isPlaceholder){let M=e.text.slice(0,-1),G=S.parse(M);M=G===0?"":M,M.length===0||G===0?t.clearSegment(e.type):t.setSegment(e.type,G),r.current=M}else e.type==="dayPeriod"&&t.clearSegment(e.type)},E=M=>{if(M.key==="a"&&(cg()?M.metaKey:M.ctrlKey)&&M.preventDefault(),!(M.ctrlKey||M.metaKey||M.shiftKey||M.altKey))switch(M.key){case"Backspace":case"Delete":M.preventDefault(),M.stopPropagation(),_();break}},{startsWith:$}=Tne({sensitivity:"base"}),k=Z_({hour:"numeric",hour12:!0}),A=P.useMemo(()=>{let M=new Date;return M.setHours(0),k.formatToParts(M).find(G=>G.type==="dayPeriod").value},[k]),w=P.useMemo(()=>{let M=new Date;return M.setHours(12),k.formatToParts(M).find(G=>G.type==="dayPeriod").value},[k]),R=Z_({year:"numeric",era:"narrow",timeZone:"UTC"}),F=P.useMemo(()=>{if(e.type!=="era")return[];let M=fc(new Xu(1,1,1),t.calendar),G=t.calendar.getEras().map(B=>{let ne=M.set({year:1,month:1,day:1,era:B}).toDate("UTC"),ie=R.formatToParts(ne).find(de=>de.type==="era").value;return{era:B,formatted:ie}}),X=xCe(G.map(B=>B.formatted));if(X)for(let B of G)B.formatted=B.formatted.slice(X);return G},[R,t.calendar,e.type]),I=M=>{if(t.isDisabled||t.isReadOnly)return;let G=r.current+M;switch(e.type){case"dayPeriod":if($(A,M))t.setSegment("dayPeriod",0);else if($(w,M))t.setSegment("dayPeriod",12);else break;c.focusNext();break;case"era":{let X=F.find(B=>$(B.formatted,M));X&&(t.setSegment("era",X.era),c.focusNext());break}case"day":case"hour":case"minute":case"second":case"month":case"year":{if(!S.isValidPartialNumber(G))return;let X=S.parse(G),B=X,ne=e.minValue===0;if(e.type==="hour"&&t.dateFormatter.resolvedOptions().hour12){switch(t.dateFormatter.resolvedOptions().hourCycle){case"h11":X>11&&(B=S.parse(M));break;case"h12":ne=!1,X>12&&(B=S.parse(M));break}e.value!==void 0&&e.value>=12&&X>1&&(X+=12)}else e.maxValue!==void 0&&X>e.maxValue&&(B=S.parse(M));if(isNaN(X))return;let oe=B!==0||ne;oe&&t.setSegment(e.type,B),e.maxValue!==void 0&&(+(X+"0")>e.maxValue||G.length>=String(e.maxValue).length)?(r.current="",oe&&c.focusNext()):r.current=G;break}}},U=()=>{r.current="",n.current&&Vm(n.current,{containingElement:m0(n.current)});let M=window.getSelection();M==null||M.collapse(n.current)},q=P.useRef(typeof document<"u"?document:null);Yd(q,"selectionchange",()=>{var M;let G=window.getSelection();G!=null&&G.anchorNode&&(!((M=n.current)===null||M===void 0)&&M.contains(G==null?void 0:G.anchorNode))&&G.collapse(n.current)});let V=P.useRef("");Yd(n,"beforeinput",M=>{if(n.current)switch(M.preventDefault(),M.inputType){case"deleteContentBackward":case"deleteContentForward":S.isValidPartialNumber(e.text)&&!t.isReadOnly&&_();break;case"insertCompositionText":V.current=n.current.textContent,n.current.textContent=n.current.textContent;break;default:M.data!=null&&I(M.data);break}}),Yd(n,"input",M=>{let{inputType:G,data:X}=M;switch(G){case"insertCompositionText":n.current&&(n.current.textContent=V.current),X!=null&&($(A,X)||$(w,X))&&I(X);break}}),Hr(()=>{let M=n.current;return()=>{document.activeElement===M&&(c.focusPrevious()||c.focusNext())}},[n,c]);let N=$1()||e.type==="timeZoneName"?{role:"textbox","aria-valuemax":null,"aria-valuemin":null,"aria-valuetext":null,"aria-valuenow":null}:{},j=P.useMemo(()=>t.segments.find(M=>M.isEditable),[t.segments]);e!==j&&!t.isInvalid&&(s=void 0);let W=vi(),K=!t.isDisabled&&!t.isReadOnly&&e.isEditable,J=e.type==="literal"?"":u.of(e.type),te=Gv({"aria-label":`${J}${l?`, ${l}`:""}${o?", ":""}`,"aria-labelledby":o});if(e.type==="literal")return{segmentProps:{"aria-hidden":!0}};let H={caretColor:"transparent"};if(a==="rtl"){H.unicodeBidi="embed";let M=h[e.type];(M==="numeric"||M==="2-digit")&&(H.direction="ltr")}return{segmentProps:Et(v,te,{id:W,...N,"aria-invalid":t.isInvalid?"true":void 0,"aria-describedby":s,"aria-readonly":t.isReadOnly||!e.isEditable?"true":void 0,"data-placeholder":e.isPlaceholder||void 0,contentEditable:K,suppressContentEditableWarning:K,spellCheck:K?"false":void 0,autoCorrect:K?"off":void 0,[parseInt(be.version,10)>=17?"enterKeyHint":"enterkeyhint"]:K?"next":void 0,inputMode:t.isDisabled||e.type==="dayPeriod"||e.type==="era"||!K?void 0:"numeric",tabIndex:t.isDisabled?void 0:0,onKeyDown:E,onFocus:U,style:H,onPointerDown(M){M.stopPropagation()},onMouseDown(M){M.stopPropagation()}})}}function xCe(e){e.sort();let t=e[0],n=e[e.length-1];for(let r=0;r<t.length;r++)if(t[r]!==n[r])return r;return 0}function SCe(e,t){let{role:n="dialog"}=e,r=$0();r=e["aria-label"]?void 0:r;let i=P.useRef(!1);return P.useEffect(()=>{if(t.current&&!t.current.contains(document.activeElement)){lc(t.current);let a=setTimeout(()=>{(document.activeElement===t.current||document.activeElement===document.body)&&(i.current=!0,t.current&&(t.current.blur(),lc(t.current)),i.current=!1)},500);return()=>{clearTimeout(a)}}},[t]),Fre(),{dialogProps:{...Gt(e,{labelable:!0}),role:n,tabIndex:-1,"aria-labelledby":e["aria-labelledby"]||r,onBlur:a=>{i.current&&a.stopPropagation()}},titleProps:{id:r}}}function _Ce(e,t,n){let{isDisabled:r}=e,i=vi(),a=vi(),u=pp(),l=P.useRef(null),o=P.useCallback(()=>{l.current=requestAnimationFrame(()=>{n.current&&n.current.setAttribute("hidden","until-found")}),Uf.flushSync(()=>{t.toggle()})},[n,t]);Yd(n,"beforematch",o);let s=P.useRef(null);return Hr(()=>{if(l.current&&cancelAnimationFrame(l.current),n.current&&!r&&!u){let c=n.current;s.current==null||typeof c.getAnimations!="function"?t.isExpanded?(c.removeAttribute("hidden"),c.style.setProperty("--disclosure-panel-width","auto"),c.style.setProperty("--disclosure-panel-height","auto")):(c.setAttribute("hidden","until-found"),c.style.setProperty("--disclosure-panel-width","0px"),c.style.setProperty("--disclosure-panel-height","0px")):t.isExpanded!==s.current&&(t.isExpanded?(c.removeAttribute("hidden"),c.style.setProperty("--disclosure-panel-width",c.scrollWidth+"px"),c.style.setProperty("--disclosure-panel-height",c.scrollHeight+"px"),Promise.all(c.getAnimations().map(d=>d.finished)).then(()=>{c.style.setProperty("--disclosure-panel-width","auto"),c.style.setProperty("--disclosure-panel-height","auto")}).catch(()=>{})):(c.style.setProperty("--disclosure-panel-width",c.scrollWidth+"px"),c.style.setProperty("--disclosure-panel-height",c.scrollHeight+"px"),window.getComputedStyle(c).height,c.style.setProperty("--disclosure-panel-width","0px"),c.style.setProperty("--disclosure-panel-height","0px"),Promise.all(c.getAnimations().map(d=>d.finished)).then(()=>c.setAttribute("hidden","until-found")).catch(()=>{}))),s.current=t.isExpanded}},[r,n,t.isExpanded,u]),P.useEffect(()=>()=>{l.current&&cancelAnimationFrame(l.current)},[]),{buttonProps:{id:i,"aria-expanded":t.isExpanded,"aria-controls":a,onPress:c=>{!r&&c.pointerType!=="keyboard"&&t.toggle()},isDisabled:r,onPressStart(c){c.pointerType==="keyboard"&&!r&&t.toggle()}},panelProps:{id:a,role:"group","aria-labelledby":i,"aria-hidden":!t.isExpanded,hidden:u||r?r||!t.isExpanded:void 0}}}const WN=new WeakMap;function Qj(e,t){var n;let{id:r}=(n=WN.get(e))!==null&&n!==void 0?n:{};if(!r)throw new Error("Unknown list");return`${r}-${ECe(t)}`}function ECe(e){return typeof e=="string"?e.replace(/\s*/g,""):""+e}var ece={};ece={deselectedItem:e=>`${e.item} غير المحدد`,longPressToSelect:"اضغط مطولًا للدخول إلى وضع التحديد.",select:"تحديد",selectedAll:"جميع العناصر المحددة.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"لم يتم تحديد عناصر",one:()=>`${t.number(e.count)} عنصر محدد`,other:()=>`${t.number(e.count)} عنصر محدد`})}.`,selectedItem:e=>`${e.item} المحدد`};var tce={};tce={deselectedItem:e=>`${e.item} не е избран.`,longPressToSelect:"Натиснете и задръжте за да влезете в избирателен режим.",select:"Изберете",selectedAll:"Всички елементи са избрани.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Няма избрани елементи",one:()=>`${t.number(e.count)} избран елемент`,other:()=>`${t.number(e.count)} избрани елементи`})}.`,selectedItem:e=>`${e.item} избран.`};var nce={};nce={deselectedItem:e=>`Položka ${e.item} není vybrána.`,longPressToSelect:"Dlouhým stisknutím přejdete do režimu výběru.",select:"Vybrat",selectedAll:"Vybrány všechny položky.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Nevybrány žádné položky",one:()=>`Vybrána ${t.number(e.count)} položka`,other:()=>`Vybráno ${t.number(e.count)} položek`})}.`,selectedItem:e=>`Vybrána položka ${e.item}.`};var rce={};rce={deselectedItem:e=>`${e.item} ikke valgt.`,longPressToSelect:"Lav et langt tryk for at aktivere valgtilstand.",select:"Vælg",selectedAll:"Alle elementer valgt.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Ingen elementer valgt",one:()=>`${t.number(e.count)} element valgt`,other:()=>`${t.number(e.count)} elementer valgt`})}.`,selectedItem:e=>`${e.item} valgt.`};var ice={};ice={deselectedItem:e=>`${e.item} nicht ausgewählt.`,longPressToSelect:"Gedrückt halten, um Auswahlmodus zu öffnen.",select:"Auswählen",selectedAll:"Alle Elemente ausgewählt.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Keine Elemente ausgewählt",one:()=>`${t.number(e.count)} Element ausgewählt`,other:()=>`${t.number(e.count)} Elemente ausgewählt`})}.`,selectedItem:e=>`${e.item} ausgewählt.`};var ace={};ace={deselectedItem:e=>`Δεν επιλέχθηκε το στοιχείο ${e.item}.`,longPressToSelect:"Πατήστε παρατεταμένα για να μπείτε σε λειτουργία επιλογής.",select:"Επιλογή",selectedAll:"Επιλέχθηκαν όλα τα στοιχεία.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Δεν επιλέχθηκαν στοιχεία",one:()=>`Επιλέχθηκε ${t.number(e.count)} στοιχείο`,other:()=>`Επιλέχθηκαν ${t.number(e.count)} στοιχεία`})}.`,selectedItem:e=>`Επιλέχθηκε το στοιχείο ${e.item}.`};var uce={};uce={deselectedItem:e=>`${e.item} not selected.`,select:"Select",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"No items selected",one:()=>`${t.number(e.count)} item selected`,other:()=>`${t.number(e.count)} items selected`})}.`,selectedAll:"All items selected.",selectedItem:e=>`${e.item} selected.`,longPressToSelect:"Long press to enter selection mode."};var lce={};lce={deselectedItem:e=>`${e.item} no seleccionado.`,longPressToSelect:"Mantenga pulsado para abrir el modo de selección.",select:"Seleccionar",selectedAll:"Todos los elementos seleccionados.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Ningún elemento seleccionado",one:()=>`${t.number(e.count)} elemento seleccionado`,other:()=>`${t.number(e.count)} elementos seleccionados`})}.`,selectedItem:e=>`${e.item} seleccionado.`};var oce={};oce={deselectedItem:e=>`${e.item} pole valitud.`,longPressToSelect:"Valikurežiimi sisenemiseks vajutage pikalt.",select:"Vali",selectedAll:"Kõik üksused valitud.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Üksusi pole valitud",one:()=>`${t.number(e.count)} üksus valitud`,other:()=>`${t.number(e.count)} üksust valitud`})}.`,selectedItem:e=>`${e.item} valitud.`};var sce={};sce={deselectedItem:e=>`Kohdetta ${e.item} ei valittu.`,longPressToSelect:"Siirry valintatilaan painamalla pitkään.",select:"Valitse",selectedAll:"Kaikki kohteet valittu.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Ei yhtään kohdetta valittu",one:()=>`${t.number(e.count)} kohde valittu`,other:()=>`${t.number(e.count)} kohdetta valittu`})}.`,selectedItem:e=>`${e.item} valittu.`};var cce={};cce={deselectedItem:e=>`${e.item} non sélectionné.`,longPressToSelect:"Appuyez de manière prolongée pour passer en mode de sélection.",select:"Sélectionner",selectedAll:"Tous les éléments sélectionnés.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Aucun élément sélectionné",one:()=>`${t.number(e.count)} élément sélectionné`,other:()=>`${t.number(e.count)} éléments sélectionnés`})}.`,selectedItem:e=>`${e.item} sélectionné.`};var fce={};fce={deselectedItem:e=>`${e.item} לא נבחר.`,longPressToSelect:"הקשה ארוכה לכניסה למצב בחירה.",select:"בחר",selectedAll:"כל הפריטים נבחרו.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"לא נבחרו פריטים",one:()=>`פריט ${t.number(e.count)} נבחר`,other:()=>`${t.number(e.count)} פריטים נבחרו`})}.`,selectedItem:e=>`${e.item} נבחר.`};var dce={};dce={deselectedItem:e=>`Stavka ${e.item} nije odabrana.`,longPressToSelect:"Dugo pritisnite za ulazak u način odabira.",select:"Odaberite",selectedAll:"Odabrane su sve stavke.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Nije odabrana nijedna stavka",one:()=>`Odabrana je ${t.number(e.count)} stavka`,other:()=>`Odabrano je ${t.number(e.count)} stavki`})}.`,selectedItem:e=>`Stavka ${e.item} je odabrana.`};var hce={};hce={deselectedItem:e=>`${e.item} nincs kijelölve.`,longPressToSelect:"Nyomja hosszan a kijelöléshez.",select:"Kijelölés",selectedAll:"Az összes elem kijelölve.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Egy elem sincs kijelölve",one:()=>`${t.number(e.count)} elem kijelölve`,other:()=>`${t.number(e.count)} elem kijelölve`})}.`,selectedItem:e=>`${e.item} kijelölve.`};var pce={};pce={deselectedItem:e=>`${e.item} non selezionato.`,longPressToSelect:"Premi a lungo per passare alla modalità di selezione.",select:"Seleziona",selectedAll:"Tutti gli elementi selezionati.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Nessun elemento selezionato",one:()=>`${t.number(e.count)} elemento selezionato`,other:()=>`${t.number(e.count)} elementi selezionati`})}.`,selectedItem:e=>`${e.item} selezionato.`};var mce={};mce={deselectedItem:e=>`${e.item} が選択されていません。`,longPressToSelect:"長押しして選択モードを開きます。",select:"選択",selectedAll:"すべての項目を選択しました。",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"項目が選択されていません",one:()=>`${t.number(e.count)} 項目を選択しました`,other:()=>`${t.number(e.count)} 項目を選択しました`})}。`,selectedItem:e=>`${e.item} を選択しました。`};var gce={};gce={deselectedItem:e=>`${e.item}이(가) 선택되지 않았습니다.`,longPressToSelect:"선택 모드로 들어가려면 길게 누르십시오.",select:"선택",selectedAll:"모든 항목이 선택되었습니다.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"선택된 항목이 없습니다",one:()=>`${t.number(e.count)}개 항목이 선택되었습니다`,other:()=>`${t.number(e.count)}개 항목이 선택되었습니다`})}.`,selectedItem:e=>`${e.item}이(가) 선택되었습니다.`};var vce={};vce={deselectedItem:e=>`${e.item} nepasirinkta.`,longPressToSelect:"Norėdami įjungti pasirinkimo režimą, paspauskite ir palaikykite.",select:"Pasirinkti",selectedAll:"Pasirinkti visi elementai.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Nepasirinktas nė vienas elementas",one:()=>`Pasirinktas ${t.number(e.count)} elementas`,other:()=>`Pasirinkta elementų: ${t.number(e.count)}`})}.`,selectedItem:e=>`Pasirinkta: ${e.item}.`};var yce={};yce={deselectedItem:e=>`Vienums ${e.item} nav atlasīts.`,longPressToSelect:"Ilgi turiet nospiestu. lai ieslēgtu atlases režīmu.",select:"Atlasīt",selectedAll:"Atlasīti visi vienumi.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Nav atlasīts neviens vienums",one:()=>`Atlasīto vienumu skaits: ${t.number(e.count)}`,other:()=>`Atlasīto vienumu skaits: ${t.number(e.count)}`})}.`,selectedItem:e=>`Atlasīts vienums ${e.item}.`};var bce={};bce={deselectedItem:e=>`${e.item} er ikke valgt.`,longPressToSelect:"Bruk et langt trykk for å gå inn i valgmodus.",select:"Velg",selectedAll:"Alle elementer er valgt.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Ingen elementer er valgt",one:()=>`${t.number(e.count)} element er valgt`,other:()=>`${t.number(e.count)} elementer er valgt`})}.`,selectedItem:e=>`${e.item} er valgt.`};var Oce={};Oce={deselectedItem:e=>`${e.item} niet geselecteerd.`,longPressToSelect:"Druk lang om de selectiemodus te openen.",select:"Selecteren",selectedAll:"Alle items geselecteerd.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Geen items geselecteerd",one:()=>`${t.number(e.count)} item geselecteerd`,other:()=>`${t.number(e.count)} items geselecteerd`})}.`,selectedItem:e=>`${e.item} geselecteerd.`};var xce={};xce={deselectedItem:e=>`Nie zaznaczono ${e.item}.`,longPressToSelect:"Naciśnij i przytrzymaj, aby wejść do trybu wyboru.",select:"Zaznacz",selectedAll:"Wszystkie zaznaczone elementy.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Nie zaznaczono żadnych elementów",one:()=>`${t.number(e.count)} zaznaczony element`,other:()=>`${t.number(e.count)} zaznaczonych elementów`})}.`,selectedItem:e=>`Zaznaczono ${e.item}.`};var Sce={};Sce={deselectedItem:e=>`${e.item} não selecionado.`,longPressToSelect:"Mantenha pressionado para entrar no modo de seleção.",select:"Selecionar",selectedAll:"Todos os itens selecionados.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Nenhum item selecionado",one:()=>`${t.number(e.count)} item selecionado`,other:()=>`${t.number(e.count)} itens selecionados`})}.`,selectedItem:e=>`${e.item} selecionado.`};var _ce={};_ce={deselectedItem:e=>`${e.item} não selecionado.`,longPressToSelect:"Prima continuamente para entrar no modo de seleção.",select:"Selecionar",selectedAll:"Todos os itens selecionados.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Nenhum item selecionado",one:()=>`${t.number(e.count)} item selecionado`,other:()=>`${t.number(e.count)} itens selecionados`})}.`,selectedItem:e=>`${e.item} selecionado.`};var Ece={};Ece={deselectedItem:e=>`${e.item} neselectat.`,longPressToSelect:"Apăsați lung pentru a intra în modul de selectare.",select:"Selectare",selectedAll:"Toate elementele selectate.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Niciun element selectat",one:()=>`${t.number(e.count)} element selectat`,other:()=>`${t.number(e.count)} elemente selectate`})}.`,selectedItem:e=>`${e.item} selectat.`};var Dce={};Dce={deselectedItem:e=>`${e.item} не выбрано.`,longPressToSelect:"Нажмите и удерживайте для входа в режим выбора.",select:"Выбрать",selectedAll:"Выбраны все элементы.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Нет выбранных элементов",one:()=>`${t.number(e.count)} элемент выбран`,other:()=>`${t.number(e.count)} элементов выбрано`})}.`,selectedItem:e=>`${e.item} выбрано.`};var $ce={};$ce={deselectedItem:e=>`Nevybraté položky: ${e.item}.`,longPressToSelect:"Dlhším stlačením prejdite do režimu výberu.",select:"Vybrať",selectedAll:"Všetky vybraté položky.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Žiadne vybraté položky",one:()=>`${t.number(e.count)} vybratá položka`,other:()=>`Počet vybratých položiek:${t.number(e.count)}`})}.`,selectedItem:e=>`Vybraté položky: ${e.item}.`};var Cce={};Cce={deselectedItem:e=>`Element ${e.item} ni izbran.`,longPressToSelect:"Za izbirni način pritisnite in dlje časa držite.",select:"Izberite",selectedAll:"Vsi elementi so izbrani.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Noben element ni izbran",one:()=>`${t.number(e.count)} element je izbran`,other:()=>`${t.number(e.count)} elementov je izbranih`})}.`,selectedItem:e=>`Element ${e.item} je izbran.`};var wce={};wce={deselectedItem:e=>`${e.item} nije izabrano.`,longPressToSelect:"Dugo pritisnite za ulazak u režim biranja.",select:"Izaberite",selectedAll:"Izabrane su sve stavke.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Nije izabrana nijedna stavka",one:()=>`Izabrana je ${t.number(e.count)} stavka`,other:()=>`Izabrano je ${t.number(e.count)} stavki`})}.`,selectedItem:e=>`${e.item} je izabrano.`};var Rce={};Rce={deselectedItem:e=>`${e.item} ej markerat.`,longPressToSelect:"Tryck länge när du vill öppna väljarläge.",select:"Markera",selectedAll:"Alla markerade objekt.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Inga markerade objekt",one:()=>`${t.number(e.count)} markerat objekt`,other:()=>`${t.number(e.count)} markerade objekt`})}.`,selectedItem:e=>`${e.item} markerat.`};var Ace={};Ace={deselectedItem:e=>`${e.item} seçilmedi.`,longPressToSelect:"Seçim moduna girmek için uzun basın.",select:"Seç",selectedAll:"Tüm ögeler seçildi.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Hiçbir öge seçilmedi",one:()=>`${t.number(e.count)} öge seçildi`,other:()=>`${t.number(e.count)} öge seçildi`})}.`,selectedItem:e=>`${e.item} seçildi.`};var Tce={};Tce={deselectedItem:e=>`${e.item} не вибрано.`,longPressToSelect:"Виконайте довге натиснення, щоб перейти в режим вибору.",select:"Вибрати",selectedAll:"Усі елементи вибрано.",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"Жодних елементів не вибрано",one:()=>`${t.number(e.count)} елемент вибрано`,other:()=>`Вибрано елементів: ${t.number(e.count)}`})}.`,selectedItem:e=>`${e.item} вибрано.`};var kce={};kce={deselectedItem:e=>`未选择 ${e.item}。`,longPressToSelect:"长按以进入选择模式。",select:"选择",selectedAll:"已选择所有项目。",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"未选择项目",one:()=>`已选择 ${t.number(e.count)} 个项目`,other:()=>`已选择 ${t.number(e.count)} 个项目`})}。`,selectedItem:e=>`已选择 ${e.item}。`};var Pce={};Pce={deselectedItem:e=>`未選取「${e.item}」。`,longPressToSelect:"長按以進入選擇模式。",select:"選取",selectedAll:"已選取所有項目。",selectedCount:(e,t)=>`${t.plural(e.count,{"=0":"未選取任何項目",one:()=>`已選取 ${t.number(e.count)} 個項目`,other:()=>`已選取 ${t.number(e.count)} 個項目`})}。`,selectedItem:e=>`已選取「${e.item}」。`};var JN={};JN={"ar-AE":ece,"bg-BG":tce,"cs-CZ":nce,"da-DK":rce,"de-DE":ice,"el-GR":ace,"en-US":uce,"es-ES":lce,"et-EE":oce,"fi-FI":sce,"fr-FR":cce,"he-IL":fce,"hr-HR":dce,"hu-HU":hce,"it-IT":pce,"ja-JP":mce,"ko-KR":gce,"lt-LT":vce,"lv-LV":yce,"nb-NO":bce,"nl-NL":Oce,"pl-PL":xce,"pt-BR":Sce,"pt-PT":_ce,"ro-RO":Ece,"ru-RU":Dce,"sk-SK":$ce,"sl-SI":Cce,"sr-SP":wce,"sv-SE":Rce,"tr-TR":Ace,"uk-UA":Tce,"zh-CN":kce,"zh-TW":Pce};function DCe(e){return e&&e.__esModule?e.default:e}function $Ce(e,t){let{getRowText:n=l=>{var o,s,c,d;return(d=(o=(s=t.collection).getTextValue)===null||o===void 0?void 0:o.call(s,l))!==null&&d!==void 0?d:(c=t.collection.getItem(l))===null||c===void 0?void 0:c.textValue}}=e,r=Qo(DCe(JN),"@react-aria/grid"),i=t.selectionManager.rawSelection,a=P.useRef(i),u=fi(()=>{var l;if(!t.selectionManager.isFocused||i===a.current){a.current=i;return}let o=zj(i,a.current),s=zj(a.current,i),c=t.selectionManager.selectionBehavior==="replace",d=[];if(t.selectionManager.selectedKeys.size===1&&c){let h=t.selectionManager.selectedKeys.keys().next().value;if(h!=null&&t.collection.getItem(h)){let m=n(h);m&&d.push(r.format("selectedItem",{item:m}))}}else if(o.size===1&&s.size===0){let h=o.keys().next().value;if(h!=null){let m=n(h);m&&d.push(r.format("selectedItem",{item:m}))}}else if(s.size===1&&o.size===0){let h=s.keys().next().value;if(h!=null&&t.collection.getItem(h)){let m=n(h);m&&d.push(r.format("deselectedItem",{item:m}))}}t.selectionManager.selectionMode==="multiple"&&(d.length===0||i==="all"||i.size>1||a.current==="all"||((l=a.current)===null||l===void 0?void 0:l.size)>1)&&d.push(i==="all"?r.format("selectedAll"):r.format("selectedCount",{count:i.size})),d.length>0&&Zb(d.join(" ")),a.current=i});ate(()=>{if(t.selectionManager.isFocused)u();else{let l=requestAnimationFrame(u);return()=>cancelAnimationFrame(l)}},[i,t.selectionManager.isFocused])}function zj(e,t){let n=new Set;if(e==="all"||t==="all")return n;for(let r of e.keys())t.has(r)||n.add(r);return n}function CCe(e){return e&&e.__esModule?e.default:e}function wCe(e){let t=Qo(CCe(JN),"@react-aria/grid"),n=TN(),r=(n==="pointer"||n==="virtual"||n==null)&&typeof window<"u"&&"ontouchstart"in window,i=P.useMemo(()=>{let u=e.selectionManager.selectionMode,l=e.selectionManager.selectionBehavior,o;return r&&(o=t.format("longPressToSelect")),l==="replace"&&u!=="none"&&e.hasItemActions?o:void 0},[e.selectionManager.selectionMode,e.selectionManager.selectionBehavior,e.hasItemActions,t,r]);return zC(i)}function RCe(e,t,n){let{isVirtualized:r,keyboardDelegate:i,layoutDelegate:a,onAction:u,disallowTypeAhead:l,linkBehavior:o="action",keyboardNavigationBehavior:s="arrow",escapeKeyBehavior:c="clearSelection",shouldSelectOnPressUp:d}=e;!e["aria-label"]&&!e["aria-labelledby"]&&console.warn("An aria-label or aria-labelledby prop is required for accessibility.");let{listProps:h}=zN({selectionManager:t.selectionManager,collection:t.collection,disabledKeys:t.disabledKeys,ref:n,keyboardDelegate:i,layoutDelegate:a,isVirtualized:r,selectOnFocus:t.selectionManager.selectionBehavior==="replace",shouldFocusWrap:e.shouldFocusWrap,linkBehavior:o,disallowTypeAhead:l,autoFocus:e.autoFocus,escapeKeyBehavior:c}),m=vi(e.id);WN.set(t,{id:m,onAction:u,linkBehavior:o,keyboardNavigationBehavior:s,shouldSelectOnPressUp:d});let y=wCe({selectionManager:t.selectionManager,hasItemActions:!!u}),v=Xne(n,{isDisabled:t.collection.size!==0}),S=Gt(e,{labelable:!0}),_=Et(S,{role:"grid",id:m,"aria-multiselectable":t.selectionManager.selectionMode==="multiple"?"true":void 0},t.collection.size===0?{tabIndex:v?-1:0}:h,y);return r&&(_["aria-rowcount"]=t.collection.size,_["aria-colcount"]=1),$Ce({},t),{gridProps:_}}const qj={expand:{ltr:"ArrowRight",rtl:"ArrowLeft"},collapse:{ltr:"ArrowLeft",rtl:"ArrowRight"}};function ACe(e,t,n){var r,i;let{node:a,isVirtualized:u}=e,{direction:l}=Wi(),{onAction:o,linkBehavior:s,keyboardNavigationBehavior:c,shouldSelectOnPressUp:d}=WN.get(t),h=$0(),m=P.useRef(null),y=()=>{var W;n.current!==null&&(m.current!=null&&a.key!==m.current||!(!((W=n.current)===null||W===void 0)&&W.contains(document.activeElement)))&&lc(n.current)},v={},S=e.hasChildItems,_=t.selectionManager.isLink(a.key);if(a!=null&&"expandedKeys"in t){var E,$;let W=(E=($=t.collection).getChildren)===null||E===void 0?void 0:E.call($,a.key);S=S||[...W??[]].length>1,o==null&&!_&&t.selectionManager.selectionMode==="none"&&S&&(o=()=>t.toggleKey(a.key));let K=S?t.expandedKeys.has(a.key):void 0,J=1;if(a.level>0&&(a==null?void 0:a.parentKey)!=null){let te=t.collection.getItem(a.parentKey);if(te){var k,A;J=[...(k=(A=t.collection).getChildren)===null||k===void 0?void 0:k.call(A,te.key)].filter(M=>M.type==="item").length}}else J=[...t.collection].filter(te=>te.level===0&&te.type==="item").length;v={"aria-expanded":K,"aria-level":a.level+1,"aria-posinset":(a==null?void 0:a.index)+1,"aria-setsize":J}}let{itemProps:w,...R}=GC({selectionManager:t.selectionManager,key:a.key,ref:n,isVirtualized:u,shouldSelectOnPressUp:e.shouldSelectOnPressUp||d,onAction:o||!((r=a.props)===null||r===void 0)&&r.onAction?ih((i=a.props)===null||i===void 0?void 0:i.onAction,o?()=>o(a.key):void 0):void 0,focus:y,linkBehavior:s}),F=W=>{if(!W.currentTarget.contains(W.target)||!n.current||!document.activeElement)return;let K=Pl(n.current);if(K.currentNode=document.activeElement,"expandedKeys"in t&&document.activeElement===n.current){if(W.key===qj.expand[l]&&t.selectionManager.focusedKey===a.key&&S&&!t.expandedKeys.has(a.key)){t.toggleKey(a.key),W.stopPropagation();return}else if(W.key===qj.collapse[l]&&t.selectionManager.focusedKey===a.key&&S&&t.expandedKeys.has(a.key)){t.toggleKey(a.key),W.stopPropagation();return}}switch(W.key){case"ArrowLeft":if(c==="arrow"){let te=l==="rtl"?K.nextNode():K.previousNode();if(te)W.preventDefault(),W.stopPropagation(),lc(te),Vm(te,{containingElement:m0(n.current)});else if(W.preventDefault(),W.stopPropagation(),l==="rtl")lc(n.current),Vm(n.current,{containingElement:m0(n.current)});else{K.currentNode=n.current;let H=Vj(K);H&&(lc(H),Vm(H,{containingElement:m0(n.current)}))}}break;case"ArrowRight":if(c==="arrow"){let te=l==="rtl"?K.previousNode():K.nextNode();if(te)W.preventDefault(),W.stopPropagation(),lc(te),Vm(te,{containingElement:m0(n.current)});else if(W.preventDefault(),W.stopPropagation(),l==="ltr")lc(n.current),Vm(n.current,{containingElement:m0(n.current)});else{K.currentNode=n.current;let H=Vj(K);H&&(lc(H),Vm(H,{containingElement:m0(n.current)}))}}break;case"ArrowUp":case"ArrowDown":if(!W.altKey&&n.current.contains(W.target)){var J;W.stopPropagation(),W.preventDefault(),(J=n.current.parentElement)===null||J===void 0||J.dispatchEvent(new KeyboardEvent(W.nativeEvent.type,W.nativeEvent))}break}},I=W=>{if(m.current=a.key,W.target!==n.current){ey()||t.selectionManager.setFocusedKey(a.key);return}},U=W=>{if(!(!W.currentTarget.contains(W.target)||!n.current||!document.activeElement))switch(W.key){case"Tab":if(c==="tab"){let K=Pl(n.current,{tabbable:!0});K.currentNode=document.activeElement,(W.shiftKey?K.previousNode():K.nextNode())&&W.stopPropagation()}}},q=rte(a.props),V=R.hasAction?q:{},N=Et(w,V,{role:"row",onKeyDownCapture:F,onKeyDown:U,onFocus:I,"aria-label":a.textValue||void 0,"aria-selected":t.selectionManager.canSelectItem(a.key)?t.selectionManager.isSelected(a.key):void 0,"aria-disabled":t.selectionManager.isDisabled(a.key)||void 0,"aria-labelledby":h&&a.textValue?`${Qj(t,a.key)} ${h}`:void 0,id:Qj(t,a.key)});if(u){let{collection:W}=t,K=[...W];N["aria-rowindex"]=K.find(J=>J.type==="section")?[...W.getKeys()].filter(J=>{var te;return((te=W.getItem(J))===null||te===void 0?void 0:te.type)!=="section"}).findIndex(J=>J===a.key)+1:a.index+1}let j={role:"gridcell","aria-colindex":1};return{rowProps:{...Et(N,v)},gridCellProps:j,descriptionProps:{id:h},...R}}function Vj(e){let t=null,n=null;do n=e.lastChild(),n&&(t=n);while(n);return t}const Mce=1,vk=Symbol.for("react-aria-landmark-manager");function TCe(e){return document.addEventListener("react-aria-landmark-manager-change",e),()=>document.removeEventListener("react-aria-landmark-manager-change",e)}function Uj(){if(typeof document>"u")return null;let e=document[vk];return e&&e.version>=Mce?e:(document[vk]=new PCe,document.dispatchEvent(new CustomEvent("react-aria-landmark-manager-change")),document[vk])}function kCe(){return UN.useSyncExternalStore(TCe,Uj,Uj)}class PCe{setupIfNeeded(){this.isListening||(document.addEventListener("keydown",this.f6Handler,{capture:!0}),document.addEventListener("focusin",this.focusinHandler,{capture:!0}),document.addEventListener("focusout",this.focusoutHandler,{capture:!0}),this.isListening=!0)}teardownIfNeeded(){!this.isListening||this.landmarks.length>0||this.refCount>0||(document.removeEventListener("keydown",this.f6Handler,{capture:!0}),document.removeEventListener("focusin",this.focusinHandler,{capture:!0}),document.removeEventListener("focusout",this.focusoutHandler,{capture:!0}),this.isListening=!1)}focusLandmark(t,n){var r,i;(i=this.landmarks.find(a=>a.ref.current===t))===null||i===void 0||(r=i.focus)===null||r===void 0||r.call(i,n)}getLandmarksByRole(t){return new Set(this.landmarks.filter(n=>n.role===t))}getLandmarkByRole(t){return this.landmarks.find(n=>n.role===t)}addLandmark(t){if(this.setupIfNeeded(),this.landmarks.find(i=>i.ref===t.ref)||!t.ref.current)return;if(this.landmarks.filter(i=>i.role==="main").length>1,this.landmarks.length===0){this.landmarks=[t],this.checkLabels(t.role);return}let n=0,r=this.landmarks.length-1;for(;n<=r;){let i=Math.floor((n+r)/2),a=t.ref.current.compareDocumentPosition(this.landmarks[i].ref.current);!!(a&Node.DOCUMENT_POSITION_PRECEDING||a&Node.DOCUMENT_POSITION_CONTAINS)?n=i+1:r=i-1}this.landmarks.splice(n,0,t),this.checkLabels(t.role)}updateLandmark(t){let n=this.landmarks.findIndex(r=>r.ref===t.ref);n>=0&&(this.landmarks[n]={...this.landmarks[n],...t},this.checkLabels(this.landmarks[n].role))}removeLandmark(t){this.landmarks=this.landmarks.filter(n=>n.ref!==t),this.teardownIfNeeded()}checkLabels(t){let n=this.getLandmarksByRole(t);n.size>1&&[...n].filter(r=>!r.label)}closestLandmark(t){let n=new Map(this.landmarks.map(i=>[i.ref.current,i])),r=t;for(;r&&!n.has(r)&&r!==document.body&&r.parentElement;)r=r.parentElement;return n.get(r)}getNextLandmark(t,{backward:n}){var r;let i=this.closestLandmark(t),a=n?this.landmarks.length-1:0;i&&(a=this.landmarks.indexOf(i)+(n?-1:1));let u=()=>{if(a<0){if(!t.dispatchEvent(new CustomEvent("react-aria-landmark-navigation",{detail:{direction:"backward"},bubbles:!0,cancelable:!0})))return!0;a=this.landmarks.length-1}else if(a>=this.landmarks.length){if(!t.dispatchEvent(new CustomEvent("react-aria-landmark-navigation",{detail:{direction:"forward"},bubbles:!0,cancelable:!0})))return!0;a=0}return a<0||a>=this.landmarks.length};if(u())return;let l=a;for(;!((r=this.landmarks[a].ref.current)===null||r===void 0)&&r.closest("[aria-hidden=true]");){if(a+=n?-1:1,u())return;if(a===l)break}return this.landmarks[a]}f6Handler(t){t.key==="F6"&&(t.altKey?this.focusMain():this.navigate(t.target,t.shiftKey))&&(t.preventDefault(),t.stopPropagation())}focusMain(){let t=this.getLandmarkByRole("main");return t&&t.ref.current&&document.contains(t.ref.current)?(this.focusLandmark(t.ref.current,"forward"),!0):!1}navigate(t,n){let r=this.getNextLandmark(t,{backward:n});if(!r)return!1;if(r.lastFocused){let i=r.lastFocused;if(document.body.contains(i))return i.focus(),!0}return r.ref.current&&document.contains(r.ref.current)?(this.focusLandmark(r.ref.current,n?"backward":"forward"),!0):!1}focusinHandler(t){let n=this.closestLandmark(t.target);n&&n.ref.current!==t.target&&this.updateLandmark({ref:n.ref,lastFocused:t.target});let r=t.relatedTarget;if(r){let i=this.closestLandmark(r);i&&i.ref.current===r&&i.blur()}}focusoutHandler(t){let n=t.target,r=t.relatedTarget;if(!r||r===document){let i=this.closestLandmark(n);i&&i.ref.current===n&&i.blur()}}createLandmarkController(){let t=this;return t.refCount++,t.setupIfNeeded(),{navigate(n,r){let i=(r==null?void 0:r.from)||document.activeElement;return t.navigate(i,n==="backward")},focusNext(n){let r=(n==null?void 0:n.from)||document.activeElement;return t.navigate(r,!1)},focusPrevious(n){let r=(n==null?void 0:n.from)||document.activeElement;return t.navigate(r,!0)},focusMain(){return t.focusMain()},dispose(){t&&(t.refCount--,t.teardownIfNeeded(),t=null)}}}registerLandmark(t){return this.landmarks.find(n=>n.ref===t.ref)?this.updateLandmark(t):this.addLandmark(t),()=>this.removeLandmark(t.ref)}constructor(){this.landmarks=[],this.isListening=!1,this.refCount=0,this.version=Mce,this.f6Handler=this.f6Handler.bind(this),this.focusinHandler=this.focusinHandler.bind(this),this.focusoutHandler=this.focusoutHandler.bind(this)}}function MCe(e,t){const{role:n,"aria-label":r,"aria-labelledby":i,focus:a}=e;let u=kCe(),l=r||i,[o,s]=P.useState(!1),c=P.useCallback(()=>{s(!0)},[s]),d=P.useCallback(()=>{s(!1)},[s]);return Hr(()=>{if(u)return u.registerLandmark({ref:t,label:l,role:n,focus:a||c,blur:d})},[u,l,t,n,a,c,d]),P.useEffect(()=>{var h;o&&((h=t.current)===null||h===void 0||h.focus())},[o,t]),{landmarkProps:{role:n,tabIndex:o?-1:void 0,"aria-label":r,"aria-labelledby":i}}}function FCe(e){let{value:t=0,minValue:n=0,maxValue:r=100,valueLabel:i,isIndeterminate:a,formatOptions:u={style:"percent"}}=e,l=Gt(e,{labelable:!0}),{labelProps:o,fieldProps:s}=gS({...e,labelElementType:"span"});t=so(t,n,r);let c=(t-n)/(r-n),d=t$(u);if(!a&&!i){let h=u.style==="percent"?c:t;i=d.format(h)}return{progressBarProps:Et(l,{...s,"aria-valuenow":a?void 0:t,"aria-valuemin":n,"aria-valuemax":r,"aria-valuetext":a?void 0:i,role:"progressbar"}),labelProps:o}}const Fce=new WeakMap;function NCe(e,t,n){let{value:r,children:i,"aria-label":a,"aria-labelledby":u,onPressStart:l,onPressEnd:o,onPressChange:s,onPress:c,onPressUp:d,onClick:h}=e;const m=e.isDisabled||t.isDisabled;let y=t.selectedValue===r,v=N=>{N.stopPropagation(),t.setSelectedValue(r)},{pressProps:S,isPressed:_}=lp({onPressStart:l,onPressEnd:o,onPressChange:s,onPress:c,onPressUp:d,onClick:h,isDisabled:m}),{pressProps:E,isPressed:$}=lp({onPressStart:l,onPressEnd:o,onPressChange:s,onPressUp:d,onClick:h,isDisabled:m,onPress(N){var j;c==null||c(N),t.setSelectedValue(r),(j=n.current)===null||j===void 0||j.focus()}}),{focusableProps:k}=I0(Et(e,{onFocus:()=>t.setLastFocusedValue(r)}),n),A=Et(S,k),w=Gt(e,{labelable:!0}),R=-1;t.selectedValue!=null?t.selectedValue===r&&(R=0):(t.lastFocusedValue===r||t.lastFocusedValue==null)&&(R=0),m&&(R=void 0);let{name:F,form:I,descriptionId:U,errorMessageId:q,validationBehavior:V}=Fce.get(t);return my(n,t.defaultSelectedValue,t.setSelectedValue),bS({validationBehavior:V},t,n),{labelProps:Et(E,P.useMemo(()=>({onClick:N=>N.preventDefault(),onMouseDown:N=>N.preventDefault()}),[])),inputProps:Et(w,{...A,type:"radio",name:F,form:I,tabIndex:R,disabled:m,required:t.isRequired&&V==="native",checked:y,value:r,onChange:v,"aria-describedby":[e["aria-describedby"],t.isInvalid?q:null,U].filter(Boolean).join(" ")||void 0}),isDisabled:m,isSelected:y,isPressed:_||$}}function ICe(e,t){let{name:n,form:r,isReadOnly:i,isRequired:a,isDisabled:u,orientation:l="vertical",validationBehavior:o="aria"}=e,{direction:s}=Wi(),{isInvalid:c,validationErrors:d,validationDetails:h}=t.displayValidation,{labelProps:m,fieldProps:y,descriptionProps:v,errorMessageProps:S}=vS({...e,labelElementType:"span",isInvalid:t.isInvalid,errorMessage:e.errorMessage||d}),_=Gt(e,{labelable:!0}),{focusWithinProps:E}=L0({onBlurWithin(A){var w;(w=e.onBlur)===null||w===void 0||w.call(e,A),t.selectedValue||t.setLastFocusedValue(null)},onFocusWithin:e.onFocus,onFocusWithinChange:e.onFocusChange}),$=A=>{let w;switch(A.key){case"ArrowRight":s==="rtl"&&l!=="vertical"?w="prev":w="next";break;case"ArrowLeft":s==="rtl"&&l!=="vertical"?w="next":w="prev";break;case"ArrowDown":w="next";break;case"ArrowUp":w="prev";break;default:return}A.preventDefault();let R=Pl(A.currentTarget,{from:A.target,accept:I=>I instanceof uf(I).HTMLInputElement&&I.type==="radio"}),F;w==="next"?(F=R.nextNode(),F||(R.currentNode=A.currentTarget,F=R.firstChild())):(F=R.previousNode(),F||(R.currentNode=A.currentTarget,F=R.lastChild())),F&&(F.focus(),t.setSelectedValue(F.value))},k=vi(n);return Fce.set(t,{name:k,form:r,descriptionId:v.id,errorMessageId:S.id,validationBehavior:o}),{radioGroupProps:Et(_,{role:"radiogroup",onKeyDown:$,"aria-invalid":t.isInvalid||void 0,"aria-errormessage":e["aria-errormessage"],"aria-readonly":i||void 0,"aria-required":a||void 0,"aria-disabled":u||void 0,"aria-orientation":l,...y,...E}),labelProps:m,descriptionProps:v,errorMessageProps:S,isInvalid:c,validationErrors:d,validationDetails:h}}const eI=new WeakMap;function LCe(e,t,n){let{keyboardDelegate:r,isDisabled:i,isRequired:a,name:u,form:l,validationBehavior:o="aria"}=e,s=ZC({usage:"search",sensitivity:"base"}),c=P.useMemo(()=>r||new yS(t.collection,t.disabledKeys,n,s),[r,t.collection,t.disabledKeys,s,n]),{menuTriggerProps:d,menuProps:h}=VN({isDisabled:i,type:"listbox"},t,n),m=I=>{if(t.selectionManager.selectionMode!=="multiple")switch(I.key){case"ArrowLeft":{var U,q;I.preventDefault();let j=t.selectedKey!=null?(U=c.getKeyAbove)===null||U===void 0?void 0:U.call(c,t.selectedKey):(q=c.getFirstKey)===null||q===void 0?void 0:q.call(c);j&&t.setSelectedKey(j);break}case"ArrowRight":{var V,N;I.preventDefault();let j=t.selectedKey!=null?(V=c.getKeyBelow)===null||V===void 0?void 0:V.call(c,t.selectedKey):(N=c.getFirstKey)===null||N===void 0?void 0:N.call(c);j&&t.setSelectedKey(j);break}}},{typeSelectProps:y}=Lre({keyboardDelegate:c,selectionManager:t.selectionManager,onTypeSelect(I){t.setSelectedKey(I)}}),{isInvalid:v,validationErrors:S,validationDetails:_}=t.displayValidation,{labelProps:E,fieldProps:$,descriptionProps:k,errorMessageProps:A}=vS({...e,labelElementType:"span",isInvalid:v,errorMessage:e.errorMessage||S});y.onKeyDown=y.onKeyDownCapture,delete y.onKeyDownCapture,t.selectionManager.selectionMode==="multiple"&&(y={});let w=Gt(e,{labelable:!0}),R=Et(y,d,$),F=vi();return eI.set(t,{isDisabled:i,isRequired:a,name:u,form:l,validationBehavior:o}),{labelProps:{...E,onClick:()=>{if(!e.isDisabled){var I;(I=n.current)===null||I===void 0||I.focus(),AN("keyboard")}}},triggerProps:Et(w,{...R,isDisabled:i,onKeyDown:ih(R.onKeyDown,m,e.onKeyDown),onKeyUp:e.onKeyUp,"aria-labelledby":[F,R["aria-labelledby"],R["aria-label"]&&!R["aria-labelledby"]?R.id:null].filter(Boolean).join(" "),onFocus(I){t.isFocused||(e.onFocus&&e.onFocus(I),e.onFocusChange&&e.onFocusChange(!0),t.setFocused(!0))},onBlur(I){t.isOpen||(e.onBlur&&e.onBlur(I),e.onFocusChange&&e.onFocusChange(!1),t.setFocused(!1))}}),valueProps:{id:F},menuProps:{...h,autoFocus:t.focusStrategy||!0,shouldSelectOnPressUp:!0,shouldFocusOnHover:!0,disallowEmptySelection:!0,linkBehavior:"selection",onBlur:I=>{I.currentTarget.contains(I.relatedTarget)||(e.onBlur&&e.onBlur(I),e.onFocusChange&&e.onFocusChange(!1),t.setFocused(!1))},"aria-labelledby":[$["aria-labelledby"],R["aria-label"]&&!$["aria-labelledby"]?R.id:null].filter(Boolean).join(" ")},descriptionProps:k,errorMessageProps:A,isInvalid:v,validationErrors:S,validationDetails:_,hiddenSelectProps:{isDisabled:i,name:u,label:e.label,state:t,triggerRef:n,form:l}}}function BCe(e,t,n){let r=eI.get(t)||{},{autoComplete:i,name:a=r.name,form:u=r.form,isDisabled:l=r.isDisabled}=e,{validationBehavior:o,isRequired:s}=r,{visuallyHiddenProps:c}=LN({style:{position:"fixed",top:0,left:0}});my(e.selectRef,t.defaultValue,t.setValue),bS({validationBehavior:o,focus:()=>{var y;return(y=n.current)===null||y===void 0?void 0:y.focus()}},t,e.selectRef);let d=t.setValue,h=P.useCallback(y=>{y.target.multiple?d(Array.from(y.target.selectedOptions,v=>v.value)):d(y.currentTarget.value)},[d]);var m;return{containerProps:{...c,"aria-hidden":!0,"data-react-aria-prevent-focus":!0,"data-a11y-ignore":"aria-hidden-focus"},inputProps:{style:{display:"none"}},selectProps:{tabIndex:-1,autoComplete:i,disabled:l,multiple:t.selectionManager.selectionMode==="multiple",required:o==="native"&&s,name:a,form:u,value:(m=t.value)!==null&&m!==void 0?m:"",onChange:h,onInput:h}}}function QCe(e){let{state:t,triggerRef:n,label:r,name:i,form:a,isDisabled:u}=e,l=P.useRef(null),o=P.useRef(null),{containerProps:s,selectProps:c}=BCe({...e,selectRef:t.collection.size<=300?l:o},t,n);if(t.collection.size<=300)return be.createElement("div",{...s,"data-testid":"hidden-select-container"},be.createElement("label",null,r,be.createElement("select",{...c,ref:l},be.createElement("option",null),[...t.collection.getKeys()].map(d=>{let h=t.collection.getItem(d);if(h&&h.type==="item")return be.createElement("option",{key:h.key,value:h.key},h.textValue)}))));if(i){let d=eI.get(t)||{},{validationBehavior:h}=d,m=Array.isArray(t.value)?t.value:[t.value];m.length===0&&(m=[null]);let y=m.map((v,S)=>{let _={type:"hidden",autoComplete:c.autoComplete,name:i,form:a,disabled:u,value:v??""};return h==="native"?be.createElement("input",{key:S,..._,ref:S===0?o:null,style:{display:"none"},type:"text",required:S===0?c.required:!1,onChange:()=>{}}):be.createElement("input",{key:S,..._,ref:S===0?o:null})});return be.createElement(be.Fragment,null,y)}return null}function zCe(e){let t=Gt(e,{labelable:!0}),n;return e.orientation==="vertical"&&(n="vertical"),e.elementType!=="hr"?{separatorProps:{...t,role:"separator","aria-orientation":n}}:{separatorProps:t}}function qCe(e,t,n){let{labelProps:r,inputProps:i,isSelected:a,isPressed:u,isDisabled:l,isReadOnly:o}=Fue(e,t,n);return{labelProps:r,inputProps:{...i,role:"switch",checked:a},isSelected:a,isPressed:u,isDisabled:l,isReadOnly:o}}const Nce=new WeakMap;function s$(e,t,n){return e?(typeof t=="string"&&(t=t.replace(/\s+/g,"")),`${Nce.get(e)}-${n}-${t}`):""}function VCe(e,t,n){let{key:r,isDisabled:i,shouldSelectOnPressUp:a}=e,{selectionManager:u,selectedKey:l}=t,o=r===l,s=i||t.isDisabled||t.selectionManager.isDisabled(r),{itemProps:c,isPressed:d}=GC({selectionManager:u,key:r,ref:n,isDisabled:s,shouldSelectOnPressUp:a,linkBehavior:"selection"}),h=s$(t,r,"tab"),m=s$(t,r,"tabpanel"),{tabIndex:y}=c,v=t.collection.getItem(r),S=Gt(v==null?void 0:v.props,{labelable:!0});delete S.id;let _=hN(v==null?void 0:v.props),{focusableProps:E}=I0({isDisabled:s},n);return{tabProps:Et(S,E,_,c,{id:h,"aria-selected":o,"aria-disabled":s||void 0,"aria-controls":o?m:void 0,tabIndex:s?void 0:y,role:"tab"}),isSelected:o,isDisabled:s,isPressed:d}}function UCe(e,t,n){let r=Xne(n)?void 0:0;var i;const a=s$(t,(i=e.id)!==null&&i!==void 0?i:t==null?void 0:t.selectedKey,"tabpanel"),u=Gv({...e,id:a,"aria-labelledby":s$(t,t==null?void 0:t.selectedKey,"tab")});return{tabPanelProps:Et(u,{tabIndex:r,role:"tabpanel","aria-describedby":e["aria-describedby"],"aria-details":e["aria-details"]})}}class jCe{getKeyLeftOf(t){return this.flipDirection?this.getNextKey(t):this.getPreviousKey(t)}getKeyRightOf(t){return this.flipDirection?this.getPreviousKey(t):this.getNextKey(t)}isDisabled(t){var n,r;return this.disabledKeys.has(t)||!!(!((r=this.collection.getItem(t))===null||r===void 0||(n=r.props)===null||n===void 0)&&n.isDisabled)}getFirstKey(){let t=this.collection.getFirstKey();return t!=null&&this.isDisabled(t)&&(t=this.getNextKey(t)),t}getLastKey(){let t=this.collection.getLastKey();return t!=null&&this.isDisabled(t)&&(t=this.getPreviousKey(t)),t}getKeyAbove(t){return this.tabDirection?null:this.getPreviousKey(t)}getKeyBelow(t){return this.tabDirection?null:this.getNextKey(t)}getNextKey(t){let n=t;do n=this.collection.getKeyAfter(n),n==null&&(n=this.collection.getFirstKey());while(n!=null&&this.isDisabled(n));return n}getPreviousKey(t){let n=t;do n=this.collection.getKeyBefore(n),n==null&&(n=this.collection.getLastKey());while(n!=null&&this.isDisabled(n));return n}constructor(t,n,r,i=new Set){this.collection=t,this.flipDirection=n==="rtl"&&r==="horizontal",this.disabledKeys=i,this.tabDirection=r==="horizontal"}}function ZCe(e,t,n){let{orientation:r="horizontal",keyboardActivation:i="automatic"}=e,{collection:a,selectionManager:u,disabledKeys:l}=t,{direction:o}=Wi(),s=P.useMemo(()=>new jCe(a,o,r,l),[a,l,r,o]),{collectionProps:c}=QN({ref:n,selectionManager:u,keyboardDelegate:s,selectOnFocus:i==="automatic",disallowEmptySelection:!0,scrollRef:n,linkBehavior:"selection"}),d=vi();Nce.set(t,d);let h=Gv({...e,id:d});return{tabListProps:{...Et(c,h),role:"tablist","aria-orientation":r,tabIndex:void 0}}}const Ice=new WeakMap;function HCe(e,t,n){let{direction:r}=Wi(),i=e.keyboardDelegate||new yS({collection:t.collection,ref:n,orientation:"horizontal",direction:r,disabledKeys:t.disabledKeys,disabledBehavior:t.selectionManager.disabledBehavior}),{labelProps:a,fieldProps:u,descriptionProps:l,errorMessageProps:o}=vS({...e,labelElementType:"span"}),{gridProps:s}=RCe({...e,...u,keyboardDelegate:i,shouldFocusWrap:!0,linkBehavior:"override",keyboardNavigationBehavior:"tab"},t,n),[c,d]=P.useState(!1),{focusWithinProps:h}=L0({onFocusWithinChange:d}),m=Gt(e),y=P.useRef(t.collection.size);return P.useEffect(()=>{n.current&&y.current>0&&t.collection.size===0&&c&&n.current.focus(),y.current=t.collection.size},[t.collection.size,c,n]),Ice.set(t,{onRemove:e.onRemove}),{gridProps:Et(s,m,{role:t.collection.size?"grid":"group","aria-atomic":!1,"aria-relevant":"additions","aria-live":c?"polite":"off",...h,...u}),labelProps:a,descriptionProps:l,errorMessageProps:o}}var Lce={};Lce={removeButtonLabel:"إزالة",removeDescription:"اضغط على مفتاح DELETE لإزالة علامة."};var Bce={};Bce={removeButtonLabel:"Премахване",removeDescription:"Натиснете Delete, за да премахнете маркера."};var Qce={};Qce={removeButtonLabel:"Odebrat",removeDescription:"Stisknutím klávesy Delete odeberete značku."};var zce={};zce={removeButtonLabel:"Fjern",removeDescription:"Tryk på Slet for at fjerne tag."};var qce={};qce={removeButtonLabel:"Entfernen",removeDescription:"Auf „Löschen“ drücken, um das Tag zu entfernen."};var Vce={};Vce={removeButtonLabel:"Κατάργηση",removeDescription:"Πατήστε Διαγραφή για να καταργήσετε την ετικέτα."};var Uce={};Uce={removeDescription:"Press Delete to remove tag.",removeButtonLabel:"Remove"};var jce={};jce={removeButtonLabel:"Quitar",removeDescription:"Pulse Eliminar para quitar la etiqueta."};var Zce={};Zce={removeButtonLabel:"Eemalda",removeDescription:"Sildi eemaldamiseks vajutage kustutusklahvi Delete."};var Hce={};Hce={removeButtonLabel:"Poista",removeDescription:"Poista tunniste painamalla Poista-painiketta."};var Xce={};Xce={removeButtonLabel:"Supprimer",removeDescription:"Appuyez sur Supprimer pour supprimer l’étiquette."};var Yce={};Yce={removeButtonLabel:"הסר",removeDescription:"לחץ על מחק כדי להסיר תג."};var Gce={};Gce={removeButtonLabel:"Ukloni",removeDescription:"Pritisnite Delete za uklanjanje oznake."};var Kce={};Kce={removeButtonLabel:"Eltávolítás",removeDescription:"Nyomja meg a Delete billentyűt a címke eltávolításához."};var Wce={};Wce={removeButtonLabel:"Rimuovi",removeDescription:"Premi Elimina per rimuovere il tag."};var Jce={};Jce={removeButtonLabel:"削除",removeDescription:"タグを削除するには、Delete キーを押します。"};var efe={};efe={removeButtonLabel:"제거",removeDescription:"태그를 제거하려면 Delete 키를 누르십시오."};var tfe={};tfe={removeButtonLabel:"Pašalinti",removeDescription:"Norėdami pašalinti žymą, paspauskite „Delete“ klavišą."};var nfe={};nfe={removeButtonLabel:"Noņemt",removeDescription:"Nospiediet Delete [Dzēst], lai noņemtu tagu."};var rfe={};rfe={removeButtonLabel:"Fjern",removeDescription:"Trykk på Slett for å fjerne taggen."};var ife={};ife={removeButtonLabel:"Verwijderen",removeDescription:"Druk op Verwijderen om de tag te verwijderen."};var afe={};afe={removeButtonLabel:"Usuń",removeDescription:"Naciśnij Usuń, aby usunąć znacznik."};var ufe={};ufe={removeButtonLabel:"Remover",removeDescription:"Pressione Delete para remover a tag."};var lfe={};lfe={removeButtonLabel:"Eliminar",removeDescription:"Prima Delete para eliminar a tag."};var ofe={};ofe={removeButtonLabel:"Îndepărtaţi",removeDescription:"Apăsați pe Delete (Ștergere) pentru a elimina eticheta."};var sfe={};sfe={removeButtonLabel:"Удалить",removeDescription:"Нажмите DELETE, чтобы удалить тег."};var cfe={};cfe={removeButtonLabel:"Odstrániť",removeDescription:"Ak chcete odstrániť značku, stlačte kláves Delete."};var ffe={};ffe={removeButtonLabel:"Odstrani",removeDescription:"Pritisnite Delete, da odstranite oznako."};var dfe={};dfe={removeButtonLabel:"Ukloni",removeDescription:"Pritisnite Obriši da biste uklonili oznaku."};var hfe={};hfe={removeButtonLabel:"Ta bort",removeDescription:"Tryck på Radera för att ta bort taggen."};var pfe={};pfe={removeButtonLabel:"Kaldır",removeDescription:"Etiketi kaldırmak için Sil tuşuna basın."};var mfe={};mfe={removeButtonLabel:"Вилучити",removeDescription:"Натисніть Delete, щоб вилучити тег."};var gfe={};gfe={removeButtonLabel:"删除",removeDescription:"按下“删除”以删除标记。"};var vfe={};vfe={removeButtonLabel:"移除",removeDescription:"按 Delete 鍵以移除標記。"};var yfe={};yfe={"ar-AE":Lce,"bg-BG":Bce,"cs-CZ":Qce,"da-DK":zce,"de-DE":qce,"el-GR":Vce,"en-US":Uce,"es-ES":jce,"et-EE":Zce,"fi-FI":Hce,"fr-FR":Xce,"he-IL":Yce,"hr-HR":Gce,"hu-HU":Kce,"it-IT":Wce,"ja-JP":Jce,"ko-KR":efe,"lt-LT":tfe,"lv-LV":nfe,"nb-NO":rfe,"nl-NL":ife,"pl-PL":afe,"pt-BR":ufe,"pt-PT":lfe,"ro-RO":ofe,"ru-RU":sfe,"sk-SK":cfe,"sl-SI":ffe,"sr-SP":dfe,"sv-SE":hfe,"tr-TR":pfe,"uk-UA":mfe,"zh-CN":gfe,"zh-TW":vfe};function XCe(e){return e&&e.__esModule?e.default:e}function YCe(e,t,n){let{item:r}=e,i=Qo(XCe(yfe),"@react-aria/tag"),a=vi(),{onRemove:u}=Ice.get(t)||{},{rowProps:l,gridCellProps:o,...s}=ACe({node:r},t,n),{descriptionProps:c,...d}=s,h=t.disabledKeys.has(r.key)||r.props.isDisabled,m=R=>{if(R.key==="Delete"||R.key==="Backspace"){if(h)return;R.preventDefault(),t.selectionManager.isSelected(r.key)?u==null||u(new Set(t.selectionManager.selectedKeys)):u==null||u(new Set([r.key]))}},y=TN();y==="virtual"&&typeof window<"u"&&"ontouchstart"in window&&(y="pointer");let v=u&&(y==="keyboard"||y==="virtual")?i.format("removeDescription"):"",S=zC(v),_=r.key===t.selectionManager.focusedKey,E=t.selectionManager.focusedKey!=null,$=-1;!h&&(_||!E)&&($=0);let k=Gt(r.props),A=rte(r.props),{focusableProps:w}=I0({isDisabled:h},n);return{removeButtonProps:{"aria-label":i.format("removeButtonLabel"),"aria-labelledby":`${a} ${l.id}`,isDisabled:h,id:a,onPress:()=>u?u(new Set([r.key])):null},rowProps:Et(w,l,k,A,{tabIndex:$,onKeyDown:u?m:void 0,"aria-describedby":S["aria-describedby"]}),gridCellProps:Et(o,{"aria-errormessage":e["aria-errormessage"],"aria-label":e["aria-label"]}),...d,allowsRemoving:!!u}}var bfe={};bfe={close:"إغلاق",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} إشعار`,other:()=>`${t.number(e.count)} إشعارات`})}.`};var Ofe={};Ofe={close:"Затвори",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} известие`,other:()=>`${t.number(e.count)} известия`})}.`};var xfe={};xfe={close:"Zavřít",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} oznámení`,other:()=>`${t.number(e.count)} oznámení`})}.`};var Sfe={};Sfe={close:"Luk",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} besked`,other:()=>`${t.number(e.count)} beskeder`})}.`};var _fe={};_fe={close:"Schließen",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} Benachrichtigung`,other:()=>`${t.number(e.count)} Benachrichtigungen`})}.`};var Efe={};Efe={close:"Κλείσιμο",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} ειδοποίηση`,other:()=>`${t.number(e.count)} ειδοποιήσεις`})}.`};var Dfe={};Dfe={close:"Close",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} notification`,other:()=>`${t.number(e.count)} notifications`})}.`};var $fe={};$fe={close:"Cerrar",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} notificación`,other:()=>`${t.number(e.count)} notificaciones`})}.`};var Cfe={};Cfe={close:"Sule",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} teatis`,other:()=>`${t.number(e.count)} teatist`})}.`};var wfe={};wfe={close:"Sulje",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} ilmoitus`,other:()=>`${t.number(e.count)} ilmoitusta`})}.`};var Rfe={};Rfe={close:"Fermer",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} notification`,other:()=>`${t.number(e.count)} notifications`})}.`};var Afe={};Afe={close:"סגור",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} התראה`,other:()=>`${t.number(e.count)} התראות`})}.`};var Tfe={};Tfe={close:"Zatvori",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} obavijest`,other:()=>`${t.number(e.count)} obavijesti`})}.`};var kfe={};kfe={close:"Bezárás",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} értesítés`,other:()=>`${t.number(e.count)} értesítés`})}.`};var Pfe={};Pfe={close:"Chiudi",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} notifica`,other:()=>`${t.number(e.count)} notifiche`})}.`};var Mfe={};Mfe={close:"閉じる",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} 個の通知`,other:()=>`${t.number(e.count)} 個の通知`})}。`};var Ffe={};Ffe={close:"닫기",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)}개 알림`,other:()=>`${t.number(e.count)}개 알림`})}.`};var Nfe={};Nfe={close:"Uždaryti",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} pranešimas`,other:()=>`${t.number(e.count)} pranešimai`})}.`};var Ife={};Ife={close:"Aizvērt",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} paziņojums`,other:()=>`${t.number(e.count)} paziņojumi`})}.`};var Lfe={};Lfe={close:"Lukk",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} varsling`,other:()=>`${t.number(e.count)} varsler`})}.`};var Bfe={};Bfe={close:"Sluiten",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} melding`,other:()=>`${t.number(e.count)} meldingen`})}.`};var Qfe={};Qfe={close:"Zamknij",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} powiadomienie`,few:()=>`${t.number(e.count)} powiadomienia`,many:()=>`${t.number(e.count)} powiadomień`,other:()=>`${t.number(e.count)} powiadomienia`})}.`};var zfe={};zfe={close:"Fechar",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} notificação`,other:()=>`${t.number(e.count)} notificações`})}.`};var qfe={};qfe={close:"Fechar",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} notificação`,other:()=>`${t.number(e.count)} notificações`})}.`};var Vfe={};Vfe={close:"Închideţi",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} notificare`,other:()=>`${t.number(e.count)} notificări`})}.`};var Ufe={};Ufe={close:"Закрыть",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} уведомление`,other:()=>`${t.number(e.count)} уведомления`})}.`};var jfe={};jfe={close:"Zatvoriť",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} oznámenie`,few:()=>`${t.number(e.count)} oznámenia`,other:()=>`${t.number(e.count)} oznámení`})}.`};var Zfe={};Zfe={close:"Zapri",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} obvestilo`,two:()=>`${t.number(e.count)} obvestili`,few:()=>`${t.number(e.count)} obvestila`,other:()=>`${t.number(e.count)} obvestil`})}.`};var Hfe={};Hfe={close:"Zatvori",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} obaveštenje`,other:()=>`${t.number(e.count)} obaveštenja`})}.`};var Xfe={};Xfe={close:"Stäng",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} meddelande`,other:()=>`${t.number(e.count)} meddelanden`})}.`};var Yfe={};Yfe={close:"Kapat",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} bildirim`,other:()=>`${t.number(e.count)} bildirim`})}.`};var Gfe={};Gfe={close:"Закрити",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} сповіщення`,other:()=>`${t.number(e.count)} сповіщення`})}.`};var Kfe={};Kfe={close:"关闭",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} 个通知`,other:()=>`${t.number(e.count)} 个通知`})}。`};var Wfe={};Wfe={close:"關閉",notifications:(e,t)=>`${t.plural(e.count,{one:()=>`${t.number(e.count)} 個通知`,other:()=>`${t.number(e.count)} 個通知`})}。`};var tI={};tI={"ar-AE":bfe,"bg-BG":Ofe,"cs-CZ":xfe,"da-DK":Sfe,"de-DE":_fe,"el-GR":Efe,"en-US":Dfe,"es-ES":$fe,"et-EE":Cfe,"fi-FI":wfe,"fr-FR":Rfe,"he-IL":Afe,"hr-HR":Tfe,"hu-HU":kfe,"it-IT":Pfe,"ja-JP":Mfe,"ko-KR":Ffe,"lt-LT":Nfe,"lv-LV":Ife,"nb-NO":Lfe,"nl-NL":Bfe,"pl-PL":Qfe,"pt-BR":zfe,"pt-PT":qfe,"ro-RO":Vfe,"ru-RU":Ufe,"sk-SK":jfe,"sl-SI":Zfe,"sr-SP":Hfe,"sv-SE":Xfe,"tr-TR":Yfe,"uk-UA":Gfe,"zh-CN":Kfe,"zh-TW":Wfe};function GCe(e){return e&&e.__esModule?e.default:e}function KCe(e,t,n){let{key:r,timer:i,timeout:a}=e.toast;P.useEffect(()=>{if(!(i==null||a==null))return i.reset(a),()=>{i.pause()}},[i,a]);let u=vi(),l=$0(),o=Qo(GCe(tI),"@react-aria/toast"),[s,c]=P.useState(!1);return P.useEffect(()=>{c(!0)},[]),{toastProps:{...Gt(e,{labelable:!0}),role:"alertdialog","aria-modal":"false","aria-labelledby":e["aria-labelledby"]||u,"aria-describedby":e["aria-describedby"]||l,tabIndex:0},contentProps:{role:"alert","aria-atomic":"true","aria-hidden":s?void 0:"true"},titleProps:{id:u},descriptionProps:{id:l},closeButtonProps:{"aria-label":o.format("close"),onPress:()=>t.close(r)}}}function WCe(e){return e&&e.__esModule?e.default:e}function JCe(e,t,n){let r=Qo(WCe(tI),"@react-aria/toast"),{landmarkProps:i}=MCe({role:"region","aria-label":e["aria-label"]||r.format("notifications",{count:t.visibleToasts.length})},n),a=P.useRef(!1),u=P.useRef(!1),l=fi(()=>{a.current||u.current?t.pauseAll():t.resumeAll()}),{hoverProps:o}=Ru({onHoverStart:()=>{a.current=!0,l()},onHoverEnd:()=>{a.current=!1,l()}}),s=P.useRef([]),c=P.useRef(t.visibleToasts),d=P.useRef(null);Hr(()=>{if(d.current===-1||t.visibleToasts.length===0||!n.current){s.current=[],c.current=t.visibleToasts;return}if(s.current=[...n.current.querySelectorAll('[role="alertdialog"]')],c.current.length===t.visibleToasts.length&&t.visibleToasts.every((_,E)=>_.key===c.current[E].key)){c.current=t.visibleToasts;return}let y=c.current.map((_,E)=>({..._,i:E,isRemoved:!t.visibleToasts.some($=>_.key===$.key)})),v=y.findIndex(_=>_.i===d.current&&_.isRemoved);if(v>-1){var S;if(ig()==="pointer"&&(!((S=h.current)===null||S===void 0)&&S.isConnected))$u(h.current);else{let _=0,E,$;for(;_<=v;)y[_].isRemoved||($=Math.max(0,_-1)),_++;for(;_<y.length;){if(!y[_].isRemoved){E=_-1;break}_++}$===void 0&&E===void 0&&($=0),$>=0&&$<s.current.length?$u(s.current[$]):E>=0&&E<s.current.length&&$u(s.current[E])}}c.current=t.visibleToasts},[t.visibleToasts,n,l]);let h=P.useRef(null),{focusWithinProps:m}=L0({onFocusWithin:y=>{u.current=!0,h.current=y.relatedTarget,l()},onBlurWithin:()=>{u.current=!1,h.current=null,l()}});return P.useEffect(()=>{var y;t.visibleToasts.length===0&&(!((y=h.current)===null||y===void 0)&&y.isConnected)&&(ig()==="pointer"?$u(h.current):h.current.focus(),h.current=null)},[n,t.visibleToasts.length]),P.useEffect(()=>()=>{var y;!((y=h.current)===null||y===void 0)&&y.isConnected&&(ig()==="pointer"?$u(h.current):h.current.focus(),h.current=null)},[n]),{regionProps:Et(i,o,m,{tabIndex:-1,"data-react-aria-top-layer":!0,onFocus:y=>{let v=y.target.closest('[role="alertdialog"]');d.current=s.current.findIndex(S=>S===v)},onBlur:()=>{d.current=-1}})}}function ewe(e,t){let n=Gt(e,{labelable:!0}),{hoverProps:r}=Ru({onHoverStart:()=>t==null?void 0:t.open(!0),onHoverEnd:()=>t==null?void 0:t.close()});return{tooltipProps:Et(n,r,{role:"tooltip"})}}function twe(e,t,n){let{isDisabled:r,trigger:i}=e,a=vi(),u=P.useRef(!1),l=P.useRef(!1),o=()=>{(u.current||l.current)&&t.open(l.current)},s=_=>{!u.current&&!l.current&&t.close(_)};P.useEffect(()=>{let _=E=>{n&&n.current&&E.key==="Escape"&&(E.stopPropagation(),t.close(!0))};if(t.isOpen)return document.addEventListener("keydown",_,!0),()=>{document.removeEventListener("keydown",_,!0)}},[n,t]);let c=()=>{i!=="focus"&&(ig()==="pointer"?u.current=!0:u.current=!1,o())},d=()=>{i!=="focus"&&(l.current=!1,u.current=!1,s())},h=()=>{l.current=!1,u.current=!1,s(!0)},m=()=>{ey()&&(l.current=!0,o())},y=()=>{l.current=!1,u.current=!1,s(!0)},{hoverProps:v}=Ru({isDisabled:r,onHoverStart:c,onHoverEnd:d}),{focusableProps:S}=I0({isDisabled:r,onFocus:m,onBlur:y},n);return{triggerProps:{"aria-describedby":t.isOpen?a:void 0,...Et(S,v,{onPointerDown:h,onKeyDown:h}),tabIndex:void 0},tooltipProps:{id:a}}}const nwe=P.createContext(null),W9=P.createContext(null),uit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,W9);let{CollectionRoot:r}=P.useContext(mp),{navProps:i}=N$e(t),a=Gt(t,{global:!0});return be.createElement(w1,{content:be.createElement(OS,t)},u=>{var l;return be.createElement("ol",{ref:n,...Et(a,i),slot:t.slot||void 0,style:t.style,className:(l=t.className)!==null&&l!==void 0?l:"react-aria-Breadcrumbs"},be.createElement(W9.Provider,{value:t},be.createElement(r,{collection:u})))})});class Jfe extends gy{}Jfe.type="item";const lit=R1(Jfe,function(t,n,r){let i=r.nextKey==null,{isDisabled:a,onAction:u}=xs(W9),l={"aria-current":i?"page":null,isDisabled:a||i,onPress:()=>u==null?void 0:u(r.key)},o=Cn({...r.props,children:r.rendered,values:{isDisabled:a||i,isCurrent:i},defaultClassName:"react-aria-Breadcrumb"}),s=Gt(t,{global:!0});return delete s.id,be.createElement("li",{...s,...o,ref:n,"data-disabled":a||i||void 0,"data-current":i||void 0},be.createElement(nwe.Provider,{value:l},o.children))}),_s=P.createContext({}),oit=C1(function(t,n){[t,n]=Nn(t,n,_s);let{elementType:r="label",...i}=t;return be.createElement(r,{className:"react-aria-Label",...i,ref:n})}),ede=P.createContext(null),sit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,ede);let{value:r=0,minValue:i=0,maxValue:a=100,isIndeterminate:u=!1}=t;r=so(r,i,a);let[l,o]=Ss(!t["aria-label"]&&!t["aria-labelledby"]),{progressBarProps:s,labelProps:c}=FCe({...t,label:o}),d=u?void 0:(r-i)/(a-i)*100,h=Cn({...t,defaultClassName:"react-aria-ProgressBar",values:{percentage:d,valueText:s["aria-valuetext"],isIndeterminate:u}}),m=Gt(t,{global:!0});return be.createElement("div",{...Et(m,h,s),ref:n,slot:t.slot||void 0},be.createElement(_s.Provider,{value:{...c,ref:l,elementType:"span"}},h.children))}),ah=P.createContext({}),cit=C1(function(t,n){[t,n]=Nn(t,n,ah);let r=t,{isPending:i}=r,{buttonProps:a,isPressed:u}=kue(t,n);a=rwe(a,i);let{focusProps:l,isFocused:o,isFocusVisible:s}=Gu(t),{hoverProps:c,isHovered:d}=Ru({...t,isDisabled:t.isDisabled||i}),h={isHovered:d,isPressed:(r.isPressed||u)&&!i,isFocused:o,isFocusVisible:s,isDisabled:t.isDisabled||!1,isPending:i??!1},m=Cn({...t,values:h,defaultClassName:"react-aria-Button"}),y=vi(a.id),v=vi(),S=a["aria-labelledby"];i&&(S?S=`${S} ${v}`:a["aria-label"]&&(S=`${y} ${v}`));let _=P.useRef(i);P.useEffect(()=>{let $={"aria-labelledby":S||y};(!_.current&&o&&i||_.current&&o&&!i)&&Zb($,"assertive"),_.current=i},[i,o,S,y]);let E=Gt(t,{global:!0});return delete E.onClick,be.createElement("button",{...Et(E,m,a,l,c),type:a.type==="submit"&&i?"button":a.type,id:y,ref:n,"aria-labelledby":S,slot:t.slot||void 0,"aria-disabled":i?"true":a["aria-disabled"],"data-disabled":t.isDisabled||void 0,"data-pressed":h.isPressed||void 0,"data-hovered":d||void 0,"data-focused":o||void 0,"data-pending":i||void 0,"data-focus-visible":s||void 0},be.createElement(ede.Provider,{value:{id:v}},m.children))});function rwe(e,t){if(t){for(const n in e)n.startsWith("on")&&!(n.includes("Focus")||n.includes("Blur"))&&(e[n]=void 0);e.href=void 0,e.target=void 0}return e}const zo=P.createContext({}),iwe=P.forwardRef(function(t,n){[t,n]=Nn(t,n,zo);let{elementType:r="span",...i}=t;return be.createElement(r,{className:"react-aria-Text",...i,ref:n})});class J9{*[Symbol.iterator](){yield*this.iterable}get size(){return this._size}getKeys(){return this.keyMap.keys()}getKeyBefore(t){let n=this.keyMap.get(t);var r;return n&&(r=n.prevKey)!==null&&r!==void 0?r:null}getKeyAfter(t){let n=this.keyMap.get(t);var r;return n&&(r=n.nextKey)!==null&&r!==void 0?r:null}getFirstKey(){return this.firstKey}getLastKey(){return this.lastKey}getItem(t){var n;return(n=this.keyMap.get(t))!==null&&n!==void 0?n:null}at(t){const n=[...this.getKeys()];return this.getItem(n[t])}getChildren(t){let n=this.keyMap.get(t);return(n==null?void 0:n.childNodes)||[]}constructor(t){this.keyMap=new Map,this.firstKey=null,this.lastKey=null,this.iterable=t;let n=l=>{if(this.keyMap.set(l.key,l),l.childNodes&&l.type==="section")for(let o of l.childNodes)n(o)};for(let l of t)n(l);let r=null,i=0,a=0;for(let[l,o]of this.keyMap)r?(r.nextKey=l,o.prevKey=r.key):(this.firstKey=l,o.prevKey=void 0),o.type==="item"&&(o.index=i++),(o.type==="section"||o.type==="item")&&a++,r=o,r.nextKey=void 0;this._size=a;var u;this.lastKey=(u=r==null?void 0:r.key)!==null&&u!==void 0?u:null}}class Gh extends Set{constructor(t,n,r){super(t),t instanceof Gh?(this.anchorKey=n??t.anchorKey,this.currentKey=r??t.currentKey):(this.anchorKey=n??null,this.currentKey=r??null)}}function awe(e,t){if(e.size!==t.size)return!1;for(let n of e)if(!t.has(n))return!1;return!0}function nI(e){let{selectionMode:t="none",disallowEmptySelection:n=!1,allowDuplicateSelectionEvents:r,selectionBehavior:i="toggle",disabledBehavior:a="all"}=e,u=P.useRef(!1),[,l]=P.useState(!1),o=P.useRef(null),s=P.useRef(null),[,c]=P.useState(null),d=P.useMemo(()=>jj(e.selectedKeys),[e.selectedKeys]),h=P.useMemo(()=>jj(e.defaultSelectedKeys,new Gh),[e.defaultSelectedKeys]),[m,y]=Yu(d,h,e.onSelectionChange),v=P.useMemo(()=>e.disabledKeys?new Set(e.disabledKeys):new Set,[e.disabledKeys]),[S,_]=P.useState(i);i==="replace"&&S==="toggle"&&typeof m=="object"&&m.size===0&&_("replace");let E=P.useRef(i);return P.useEffect(()=>{i!==E.current&&(_(i),E.current=i)},[i]),{selectionMode:t,disallowEmptySelection:n,selectionBehavior:S,setSelectionBehavior:_,get isFocused(){return u.current},setFocused($){u.current=$,l($)},get focusedKey(){return o.current},get childFocusStrategy(){return s.current},setFocusedKey($,k="first"){o.current=$,s.current=k,c($)},selectedKeys:m,setSelectedKeys($){(r||!awe($,m))&&y($)},disabledKeys:v,disabledBehavior:a}}function jj(e,t){return e?e==="all"?"all":new Gh(e):t}class xS{get selectionMode(){return this.state.selectionMode}get disallowEmptySelection(){return this.state.disallowEmptySelection}get selectionBehavior(){return this.state.selectionBehavior}setSelectionBehavior(t){this.state.setSelectionBehavior(t)}get isFocused(){return this.state.isFocused}setFocused(t){this.state.setFocused(t)}get focusedKey(){return this.state.focusedKey}get childFocusStrategy(){return this.state.childFocusStrategy}setFocusedKey(t,n){(t==null||this.collection.getItem(t))&&this.state.setFocusedKey(t,n)}get selectedKeys(){return this.state.selectedKeys==="all"?new Set(this.getSelectAllKeys()):this.state.selectedKeys}get rawSelection(){return this.state.selectedKeys}isSelected(t){if(this.state.selectionMode==="none")return!1;let n=this.getKey(t);return n==null?!1:this.state.selectedKeys==="all"?this.canSelectItem(n):this.state.selectedKeys.has(n)}get isEmpty(){return this.state.selectedKeys!=="all"&&this.state.selectedKeys.size===0}get isSelectAll(){if(this.isEmpty)return!1;if(this.state.selectedKeys==="all")return!0;if(this._isSelectAll!=null)return this._isSelectAll;let t=this.getSelectAllKeys(),n=this.state.selectedKeys;return this._isSelectAll=t.every(r=>n.has(r)),this._isSelectAll}get firstSelectedKey(){let t=null;for(let r of this.state.selectedKeys){let i=this.collection.getItem(r);(!t||i&&ck(this.collection,i,t)<0)&&(t=i)}var n;return(n=t==null?void 0:t.key)!==null&&n!==void 0?n:null}get lastSelectedKey(){let t=null;for(let r of this.state.selectedKeys){let i=this.collection.getItem(r);(!t||i&&ck(this.collection,i,t)>0)&&(t=i)}var n;return(n=t==null?void 0:t.key)!==null&&n!==void 0?n:null}get disabledKeys(){return this.state.disabledKeys}get disabledBehavior(){return this.state.disabledBehavior}extendSelection(t){if(this.selectionMode==="none")return;if(this.selectionMode==="single"){this.replaceSelection(t);return}let n=this.getKey(t);if(n==null)return;let r;if(this.state.selectedKeys==="all")r=new Gh([n],n,n);else{let u=this.state.selectedKeys;var i;let l=(i=u.anchorKey)!==null&&i!==void 0?i:n;r=new Gh(u,l,n);var a;for(let o of this.getKeyRange(l,(a=u.currentKey)!==null&&a!==void 0?a:n))r.delete(o);for(let o of this.getKeyRange(n,l))this.canSelectItem(o)&&r.add(o)}this.state.setSelectedKeys(r)}getKeyRange(t,n){let r=this.collection.getItem(t),i=this.collection.getItem(n);return r&&i?ck(this.collection,r,i)<=0?this.getKeyRangeInternal(t,n):this.getKeyRangeInternal(n,t):[]}getKeyRangeInternal(t,n){var r;if(!((r=this.layoutDelegate)===null||r===void 0)&&r.getKeyRange)return this.layoutDelegate.getKeyRange(t,n);let i=[],a=t;for(;a!=null;){let u=this.collection.getItem(a);if(u&&(u.type==="item"||u.type==="cell"&&this.allowsCellSelection)&&i.push(a),a===n)return i;a=this.collection.getKeyAfter(a)}return[]}getKey(t){let n=this.collection.getItem(t);if(!n||n.type==="cell"&&this.allowsCellSelection)return t;for(;n&&n.type!=="item"&&n.parentKey!=null;)n=this.collection.getItem(n.parentKey);return!n||n.type!=="item"?null:n.key}toggleSelection(t){if(this.selectionMode==="none")return;if(this.selectionMode==="single"&&!this.isSelected(t)){this.replaceSelection(t);return}let n=this.getKey(t);if(n==null)return;let r=new Gh(this.state.selectedKeys==="all"?this.getSelectAllKeys():this.state.selectedKeys);r.has(n)?r.delete(n):this.canSelectItem(n)&&(r.add(n),r.anchorKey=n,r.currentKey=n),!(this.disallowEmptySelection&&r.size===0)&&this.state.setSelectedKeys(r)}replaceSelection(t){if(this.selectionMode==="none")return;let n=this.getKey(t);if(n==null)return;let r=this.canSelectItem(n)?new Gh([n],n,n):new Gh;this.state.setSelectedKeys(r)}setSelectedKeys(t){if(this.selectionMode==="none")return;let n=new Gh;for(let r of t){let i=this.getKey(r);if(i!=null&&(n.add(i),this.selectionMode==="single"))break}this.state.setSelectedKeys(n)}getSelectAllKeys(){let t=[],n=r=>{for(;r!=null;){if(this.canSelectItem(r)){var i;let u=this.collection.getItem(r);(u==null?void 0:u.type)==="item"&&t.push(r);var a;u!=null&&u.hasChildNodes&&(this.allowsCellSelection||u.type!=="item")&&n((a=(i=YDe(KC(u,this.collection)))===null||i===void 0?void 0:i.key)!==null&&a!==void 0?a:null)}r=this.collection.getKeyAfter(r)}};return n(this.collection.getFirstKey()),t}selectAll(){!this.isSelectAll&&this.selectionMode==="multiple"&&this.state.setSelectedKeys("all")}clearSelection(){!this.disallowEmptySelection&&(this.state.selectedKeys==="all"||this.state.selectedKeys.size>0)&&this.state.setSelectedKeys(new Gh)}toggleSelectAll(){this.isSelectAll?this.clearSelection():this.selectAll()}select(t,n){this.selectionMode!=="none"&&(this.selectionMode==="single"?this.isSelected(t)&&!this.disallowEmptySelection?this.toggleSelection(t):this.replaceSelection(t):this.selectionBehavior==="toggle"||n&&(n.pointerType==="touch"||n.pointerType==="virtual")?this.toggleSelection(t):this.replaceSelection(t))}isSelectionEqual(t){if(t===this.state.selectedKeys)return!0;let n=this.selectedKeys;if(t.size!==n.size)return!1;for(let r of t)if(!n.has(r))return!1;for(let r of n)if(!t.has(r))return!1;return!0}canSelectItem(t){var n;if(this.state.selectionMode==="none"||this.state.disabledKeys.has(t))return!1;let r=this.collection.getItem(t);return!(!r||!(r==null||(n=r.props)===null||n===void 0)&&n.isDisabled||r.type==="cell"&&!this.allowsCellSelection)}isDisabled(t){var n,r;return this.state.disabledBehavior==="all"&&(this.state.disabledKeys.has(t)||!!(!((r=this.collection.getItem(t))===null||r===void 0||(n=r.props)===null||n===void 0)&&n.isDisabled))}isLink(t){var n,r;return!!(!((r=this.collection.getItem(t))===null||r===void 0||(n=r.props)===null||n===void 0)&&n.href)}getItemProps(t){var n;return(n=this.collection.getItem(t))===null||n===void 0?void 0:n.props}withCollection(t){return new xS(t,this.state,{allowsCellSelection:this.allowsCellSelection,layoutDelegate:this.layoutDelegate||void 0})}constructor(t,n,r){this.collection=t,this.state=n;var i;this.allowsCellSelection=(i=r==null?void 0:r.allowsCellSelection)!==null&&i!==void 0?i:!1,this._isSelectAll=null,this.layoutDelegate=(r==null?void 0:r.layoutDelegate)||null}}function tw(e){let{filter:t,layoutDelegate:n}=e,r=nI(e),i=P.useMemo(()=>e.disabledKeys?new Set(e.disabledKeys):new Set,[e.disabledKeys]),a=P.useCallback(s=>t?new J9(t(s)):new J9(s),[t]),u=P.useMemo(()=>({suppressTextValueWarning:e.suppressTextValueWarning}),[e.suppressTextValueWarning]),l=Bre(e,a,u),o=P.useMemo(()=>new xS(l,r,{layoutDelegate:n}),[l,r,n]);return nde(l,o),{collection:l,disabledKeys:i,selectionManager:o}}function tde(e,t){let n=P.useMemo(()=>t?e.collection.filter(t):e.collection,[e.collection,t]),r=e.selectionManager.withCollection(n);return nde(n,r),{collection:n,selectionManager:r,disabledKeys:e.disabledKeys}}function nde(e,t){const n=P.useRef(null);P.useEffect(()=>{if(t.focusedKey!=null&&!e.getItem(t.focusedKey)&&n.current){const c=n.current.getItem(t.focusedKey),d=[...n.current.getKeys()].map(_=>{const E=n.current.getItem(_);return(E==null?void 0:E.type)==="item"?E:null}).filter(_=>_!==null),h=[...e.getKeys()].map(_=>{const E=e.getItem(_);return(E==null?void 0:E.type)==="item"?E:null}).filter(_=>_!==null);var r,i;const m=((r=d==null?void 0:d.length)!==null&&r!==void 0?r:0)-((i=h==null?void 0:h.length)!==null&&i!==void 0?i:0);var a,u,l;let y=Math.min(m>1?Math.max(((a=c==null?void 0:c.index)!==null&&a!==void 0?a:0)-m+1,0):(u=c==null?void 0:c.index)!==null&&u!==void 0?u:0,((l=h==null?void 0:h.length)!==null&&l!==void 0?l:0)-1),v=null,S=!1;for(;y>=0;){if(!t.isDisabled(h[y].key)){v=h[y];break}if(y<h.length-1&&!S)y++;else{S=!0;var o,s;y>((o=c==null?void 0:c.index)!==null&&o!==void 0?o:0)&&(y=(s=c==null?void 0:c.index)!==null&&s!==void 0?s:0),y--}}t.setFocusedKey(v?v.key:null)}n.current=e},[e,t])}function rde(e){var t;let[n,r]=Yu(e.selectedKey,(t=e.defaultSelectedKey)!==null&&t!==void 0?t:null,e.onSelectionChange),i=P.useMemo(()=>n!=null?[n]:[],[n]),{collection:a,disabledKeys:u,selectionManager:l}=tw({...e,selectionMode:"single",disallowEmptySelection:!0,allowDuplicateSelectionEvents:!0,selectedKeys:i,onSelectionChange:s=>{if(s==="all")return;var c;let d=(c=s.values().next().value)!==null&&c!==void 0?c:null;d===n&&e.onSelectionChange&&e.onSelectionChange(d),r(d)}}),o=n!=null?a.getItem(n):null;return{collection:a,disabledKeys:u,selectionManager:l,selectedKey:n,setSelectedKey:r,selectedItem:o}}function Y3(e){let[t,n]=Yu(e.isOpen,e.defaultOpen||!1,e.onOpenChange);const r=P.useCallback(()=>{n(!0)},[n]),i=P.useCallback(()=>{n(!1)},[n]),a=P.useCallback(()=>{n(!t)},[n,t]);return{isOpen:t,setOpen:n,open:r,close:i,toggle:a}}function uwe(e){var t;let{defaultFilter:n,menuTrigger:r="input",allowsEmptyCollection:i=!1,allowsCustomValue:a,shouldCloseOnBlur:u=!0}=e,[l,o]=P.useState(!1),[s,c]=P.useState(!1),[d,h]=P.useState(null),m=$e=>{e.onSelectionChange&&e.onSelectionChange($e),$e===_&&(B(),M())};var y;let{collection:v,selectionManager:S,selectedKey:_,setSelectedKey:E,selectedItem:$,disabledKeys:k}=rde({...e,onSelectionChange:m,items:(y=e.items)!==null&&y!==void 0?y:e.defaultItems}),[A,w]=Yu(e.inputValue,Zj(e.defaultInputValue,_,v)||"",e.onInputChange),[R]=P.useState(_),[F]=P.useState(A),I=v,U=P.useMemo(()=>e.items!=null||!n?v:lwe(v,A,n),[v,A,n,e.items]),[q,V]=P.useState(U),N=P.useRef("focus"),W=Y3({...e,onOpenChange:$e=>{e.onOpenChange&&e.onOpenChange($e,$e?N.current:void 0),S.setFocused($e),$e||S.setFocusedKey(null)},isOpen:void 0,defaultOpen:void 0}),K=($e=null,Ae)=>{let He=Ae==="manual"||Ae==="focus"&&r==="focus";(i||U.size>0||He&&I.size>0||e.items)&&(He&&!W.isOpen&&e.items===void 0&&o(!0),N.current=Ae,h($e),W.open())},J=($e=null,Ae)=>{let He=Ae==="manual"||Ae==="focus"&&r==="focus";!(i||U.size>0||He&&I.size>0||e.items)&&!W.isOpen||(He&&!W.isOpen&&e.items===void 0&&o(!0),W.isOpen||(N.current=Ae),H($e))},te=P.useCallback(()=>{V(l?I:U)},[l,I,U]),H=P.useCallback(($e=null)=>{W.isOpen&&te(),h($e),W.toggle()},[W,te]),M=P.useCallback(()=>{W.isOpen&&(te(),W.close())},[W,te]),[G,X]=P.useState(A),B=()=>{var $e,Ae;let He=_!=null&&(Ae=($e=v.getItem(_))===null||$e===void 0?void 0:$e.textValue)!==null&&Ae!==void 0?Ae:"";X(He),w(He)};var ne,oe;let ie=P.useRef((oe=(ne=e.selectedKey)!==null&&ne!==void 0?ne:e.defaultSelectedKey)!==null&&oe!==void 0?oe:null);var de;let ge=P.useRef(_!=null&&(de=(t=v.getItem(_))===null||t===void 0?void 0:t.textValue)!==null&&de!==void 0?de:"");P.useEffect(()=>{var $e;s&&(U.size>0||i)&&!W.isOpen&&A!==G&&r!=="manual"&&K(null,"input"),!l&&!i&&W.isOpen&&U.size===0&&M(),_!=null&&_!==ie.current&&M(),A!==G&&(S.setFocusedKey(null),o(!1),A===""&&(e.inputValue===void 0||e.selectedKey===void 0)&&E(null)),_!==ie.current&&(e.inputValue===void 0||e.selectedKey===void 0)?B():G!==A&&X(A);var Ae;let He=_!=null&&(Ae=($e=v.getItem(_))===null||$e===void 0?void 0:$e.textValue)!==null&&Ae!==void 0?Ae:"";!s&&_!=null&&e.inputValue===void 0&&_===ie.current&&ge.current!==He&&(X(He),w(He)),ie.current=_,ge.current=He});let ye=Og({...e,value:P.useMemo(()=>({inputValue:A,selectedKey:_}),[A,_])}),me=()=>{a&&_==null?Ee():se()},Ee=()=>{ie.current=null,E(null),M()},se=()=>{if(e.selectedKey!==void 0&&e.inputValue!==void 0){var $e,Ae;($e=e.onSelectionChange)===null||$e===void 0||$e.call(e,_);var He;let je=_!=null&&(He=(Ae=v.getItem(_))===null||Ae===void 0?void 0:Ae.textValue)!==null&&He!==void 0?He:"";X(je),M()}else B(),M()};const pe=()=>{if(a){var $e,Ae;const He=_!=null&&(Ae=($e=v.getItem(_))===null||$e===void 0?void 0:$e.textValue)!==null&&Ae!==void 0?Ae:"";A===He?se():Ee()}else se()};let ve=()=>{W.isOpen&&S.focusedKey!=null?_===S.focusedKey?se():E(S.focusedKey):pe()},we=P.useRef(A),Me=$e=>{$e?(we.current=A,r==="focus"&&!e.isReadOnly&&K(null,"focus")):(u&&pe(),A!==we.current&&ye.commitValidation()),c($e)},Re=P.useMemo(()=>W.isOpen?l?I:U:q,[W.isOpen,I,U,l,q]);var et;let tt=(et=e.defaultSelectedKey)!==null&&et!==void 0?et:R;var Oe;return{...ye,...W,focusStrategy:d,toggle:J,open:K,close:pe,selectionManager:S,selectedKey:_,defaultSelectedKey:tt,setSelectedKey:E,disabledKeys:k,isFocused:s,setFocused:Me,selectedItem:$,collection:Re,inputValue:A,defaultInputValue:(Oe=Zj(e.defaultInputValue,tt,v))!==null&&Oe!==void 0?Oe:F,setInputValue:w,commit:ve,revert:me}}function lwe(e,t,n){return new J9(ide(e,e,t,n))}function ide(e,t,n,r){let i=[];for(let a of t)if(a.type==="section"&&a.hasChildNodes){let u=ide(e,KC(a,e),n,r);[...u].some(l=>l.type==="item")&&i.push({...a,childNodes:u})}else a.type==="item"&&r(a.textValue,n)?i.push({...a}):a.type!=="item"&&i.push({...a});return i}function Zj(e,t,n){if(e==null){var r,i;if(t!=null)return(i=(r=n.getItem(t))===null||r===void 0?void 0:r.textValue)!==null&&i!==void 0?i:""}return e}var ade={};ade={rangeOverflow:e=>`يجب أن تكون القيمة ${e.maxValue} أو قبل ذلك.`,rangeReversed:"تاريخ البدء يجب أن يكون قبل تاريخ الانتهاء.",rangeUnderflow:e=>`يجب أن تكون القيمة ${e.minValue} أو بعد ذلك.`,unavailableDate:"البيانات المحددة غير متاحة."};var ude={};ude={rangeOverflow:e=>`Стойността трябва да е ${e.maxValue} или по-ранна.`,rangeReversed:"Началната дата трябва да е преди крайната.",rangeUnderflow:e=>`Стойността трябва да е ${e.minValue} или по-късно.`,unavailableDate:"Избраната дата не е налична."};var lde={};lde={rangeOverflow:e=>`Hodnota musí být ${e.maxValue} nebo dřívější.`,rangeReversed:"Datum zahájení musí předcházet datu ukončení.",rangeUnderflow:e=>`Hodnota musí být ${e.minValue} nebo pozdější.`,unavailableDate:"Vybrané datum není k dispozici."};var ode={};ode={rangeOverflow:e=>`Værdien skal være ${e.maxValue} eller tidligere.`,rangeReversed:"Startdatoen skal være før slutdatoen.",rangeUnderflow:e=>`Værdien skal være ${e.minValue} eller nyere.`,unavailableDate:"Den valgte dato er ikke tilgængelig."};var sde={};sde={rangeOverflow:e=>`Der Wert muss ${e.maxValue} oder früher sein.`,rangeReversed:"Das Startdatum muss vor dem Enddatum liegen.",rangeUnderflow:e=>`Der Wert muss ${e.minValue} oder später sein.`,unavailableDate:"Das ausgewählte Datum ist nicht verfügbar."};var cde={};cde={rangeOverflow:e=>`Η τιμή πρέπει να είναι ${e.maxValue} ή παλαιότερη.`,rangeReversed:"Η ημερομηνία έναρξης πρέπει να είναι πριν από την ημερομηνία λήξης.",rangeUnderflow:e=>`Η τιμή πρέπει να είναι ${e.minValue} ή μεταγενέστερη.`,unavailableDate:"Η επιλεγμένη ημερομηνία δεν είναι διαθέσιμη."};var fde={};fde={rangeUnderflow:e=>`Value must be ${e.minValue} or later.`,rangeOverflow:e=>`Value must be ${e.maxValue} or earlier.`,rangeReversed:"Start date must be before end date.",unavailableDate:"Selected date unavailable."};var dde={};dde={rangeOverflow:e=>`El valor debe ser ${e.maxValue} o anterior.`,rangeReversed:"La fecha de inicio debe ser anterior a la fecha de finalización.",rangeUnderflow:e=>`El valor debe ser ${e.minValue} o posterior.`,unavailableDate:"Fecha seleccionada no disponible."};var hde={};hde={rangeOverflow:e=>`Väärtus peab olema ${e.maxValue} või varasem.`,rangeReversed:"Alguskuupäev peab olema enne lõppkuupäeva.",rangeUnderflow:e=>`Väärtus peab olema ${e.minValue} või hilisem.`,unavailableDate:"Valitud kuupäev pole saadaval."};var pde={};pde={rangeOverflow:e=>`Arvon on oltava ${e.maxValue} tai sitä aikaisempi.`,rangeReversed:"Aloituspäivän on oltava ennen lopetuspäivää.",rangeUnderflow:e=>`Arvon on oltava ${e.minValue} tai sitä myöhäisempi.`,unavailableDate:"Valittu päivämäärä ei ole käytettävissä."};var mde={};mde={rangeOverflow:e=>`La valeur doit être ${e.maxValue} ou antérieure.`,rangeReversed:"La date de début doit être antérieure à la date de fin.",rangeUnderflow:e=>`La valeur doit être ${e.minValue} ou ultérieure.`,unavailableDate:"La date sélectionnée n’est pas disponible."};var gde={};gde={rangeOverflow:e=>`הערך חייב להיות ${e.maxValue} או מוקדם יותר.`,rangeReversed:"תאריך ההתחלה חייב להיות לפני תאריך הסיום.",rangeUnderflow:e=>`הערך חייב להיות ${e.minValue} או מאוחר יותר.`,unavailableDate:"התאריך הנבחר אינו זמין."};var vde={};vde={rangeOverflow:e=>`Vrijednost mora biti ${e.maxValue} ili ranije.`,rangeReversed:"Datum početka mora biti prije datuma završetka.",rangeUnderflow:e=>`Vrijednost mora biti ${e.minValue} ili kasnije.`,unavailableDate:"Odabrani datum nije dostupan."};var yde={};yde={rangeOverflow:e=>`Az értéknek ${e.maxValue} vagy korábbinak kell lennie.`,rangeReversed:"A kezdő dátumnak a befejező dátumnál korábbinak kell lennie.",rangeUnderflow:e=>`Az értéknek ${e.minValue} vagy későbbinek kell lennie.`,unavailableDate:"A kiválasztott dátum nem érhető el."};var bde={};bde={rangeOverflow:e=>`Il valore deve essere ${e.maxValue} o precedente.`,rangeReversed:"La data di inizio deve essere antecedente alla data di fine.",rangeUnderflow:e=>`Il valore deve essere ${e.minValue} o successivo.`,unavailableDate:"Data selezionata non disponibile."};var Ode={};Ode={rangeOverflow:e=>`値は ${e.maxValue} 以下にする必要があります。`,rangeReversed:"開始日は終了日より前にする必要があります。",rangeUnderflow:e=>`値は ${e.minValue} 以上にする必要があります。`,unavailableDate:"選択した日付は使用できません。"};var xde={};xde={rangeOverflow:e=>`값은 ${e.maxValue} 이전이어야 합니다.`,rangeReversed:"시작일은 종료일 이전이어야 합니다.",rangeUnderflow:e=>`값은 ${e.minValue} 이상이어야 합니다.`,unavailableDate:"선택한 날짜를 사용할 수 없습니다."};var Sde={};Sde={rangeOverflow:e=>`Reikšmė turi būti ${e.maxValue} arba ankstesnė.`,rangeReversed:"Pradžios data turi būti ankstesnė nei pabaigos data.",rangeUnderflow:e=>`Reikšmė turi būti ${e.minValue} arba naujesnė.`,unavailableDate:"Pasirinkta data nepasiekiama."};var _de={};_de={rangeOverflow:e=>`Vērtībai ir jābūt ${e.maxValue} vai agrākai.`,rangeReversed:"Sākuma datumam ir jābūt pirms beigu datuma.",rangeUnderflow:e=>`Vērtībai ir jābūt ${e.minValue} vai vēlākai.`,unavailableDate:"Atlasītais datums nav pieejams."};var Ede={};Ede={rangeOverflow:e=>`Verdien må være ${e.maxValue} eller tidligere.`,rangeReversed:"Startdatoen må være før sluttdatoen.",rangeUnderflow:e=>`Verdien må være ${e.minValue} eller senere.`,unavailableDate:"Valgt dato utilgjengelig."};var Dde={};Dde={rangeOverflow:e=>`Waarde moet ${e.maxValue} of eerder zijn.`,rangeReversed:"De startdatum moet voor de einddatum liggen.",rangeUnderflow:e=>`Waarde moet ${e.minValue} of later zijn.`,unavailableDate:"Geselecteerde datum niet beschikbaar."};var $de={};$de={rangeOverflow:e=>`Wartość musi mieć wartość ${e.maxValue} lub wcześniejszą.`,rangeReversed:"Data rozpoczęcia musi być wcześniejsza niż data zakończenia.",rangeUnderflow:e=>`Wartość musi mieć wartość ${e.minValue} lub późniejszą.`,unavailableDate:"Wybrana data jest niedostępna."};var Cde={};Cde={rangeOverflow:e=>`O valor deve ser ${e.maxValue} ou anterior.`,rangeReversed:"A data inicial deve ser anterior à data final.",rangeUnderflow:e=>`O valor deve ser ${e.minValue} ou posterior.`,unavailableDate:"Data selecionada indisponível."};var wde={};wde={rangeOverflow:e=>`O valor tem de ser ${e.maxValue} ou anterior.`,rangeReversed:"A data de início deve ser anterior à data de fim.",rangeUnderflow:e=>`O valor tem de ser ${e.minValue} ou posterior.`,unavailableDate:"Data selecionada indisponível."};var Rde={};Rde={rangeOverflow:e=>`Valoarea trebuie să fie ${e.maxValue} sau anterioară.`,rangeReversed:"Data de început trebuie să fie anterioară datei de sfârșit.",rangeUnderflow:e=>`Valoarea trebuie să fie ${e.minValue} sau ulterioară.`,unavailableDate:"Data selectată nu este disponibilă."};var Ade={};Ade={rangeOverflow:e=>`Значение должно быть не позже ${e.maxValue}.`,rangeReversed:"Дата начала должна предшествовать дате окончания.",rangeUnderflow:e=>`Значение должно быть не раньше ${e.minValue}.`,unavailableDate:"Выбранная дата недоступна."};var Tde={};Tde={rangeOverflow:e=>`Hodnota musí byť ${e.maxValue} alebo skoršia.`,rangeReversed:"Dátum začiatku musí byť skorší ako dátum konca.",rangeUnderflow:e=>`Hodnota musí byť ${e.minValue} alebo neskoršia.`,unavailableDate:"Vybratý dátum je nedostupný."};var kde={};kde={rangeOverflow:e=>`Vrednost mora biti ${e.maxValue} ali starejša.`,rangeReversed:"Začetni datum mora biti pred končnim datumom.",rangeUnderflow:e=>`Vrednost mora biti ${e.minValue} ali novejša.`,unavailableDate:"Izbrani datum ni na voljo."};var Pde={};Pde={rangeOverflow:e=>`Vrednost mora da bude ${e.maxValue} ili starija.`,rangeReversed:"Datum početka mora biti pre datuma završetka.",rangeUnderflow:e=>`Vrednost mora da bude ${e.minValue} ili novija.`,unavailableDate:"Izabrani datum nije dostupan."};var Mde={};Mde={rangeOverflow:e=>`Värdet måste vara ${e.maxValue} eller tidigare.`,rangeReversed:"Startdatumet måste vara före slutdatumet.",rangeUnderflow:e=>`Värdet måste vara ${e.minValue} eller senare.`,unavailableDate:"Det valda datumet är inte tillgängligt."};var Fde={};Fde={rangeOverflow:e=>`Değer, ${e.maxValue} veya öncesi olmalıdır.`,rangeReversed:"Başlangıç tarihi bitiş tarihinden önce olmalıdır.",rangeUnderflow:e=>`Değer, ${e.minValue} veya sonrası olmalıdır.`,unavailableDate:"Seçilen tarih kullanılamıyor."};var Nde={};Nde={rangeOverflow:e=>`Значення має бути не пізніше ${e.maxValue}.`,rangeReversed:"Дата початку має передувати даті завершення.",rangeUnderflow:e=>`Значення має бути не раніше ${e.minValue}.`,unavailableDate:"Вибрана дата недоступна."};var Ide={};Ide={rangeOverflow:e=>`值必须是 ${e.maxValue} 或更早日期。`,rangeReversed:"开始日期必须早于结束日期。",rangeUnderflow:e=>`值必须是 ${e.minValue} 或更晚日期。`,unavailableDate:"所选日期不可用。"};var Lde={};Lde={rangeOverflow:e=>`值必須是 ${e.maxValue} 或更早。`,rangeReversed:"開始日期必須在結束日期之前。",rangeUnderflow:e=>`值必須是 ${e.minValue} 或更晚。`,unavailableDate:"所選日期無法使用。"};var Bde={};Bde={"ar-AE":ade,"bg-BG":ude,"cs-CZ":lde,"da-DK":ode,"de-DE":sde,"el-GR":cde,"en-US":fde,"es-ES":dde,"et-EE":hde,"fi-FI":pde,"fr-FR":mde,"he-IL":gde,"hr-HR":vde,"hu-HU":yde,"it-IT":bde,"ja-JP":Ode,"ko-KR":xde,"lt-LT":Sde,"lv-LV":_de,"nb-NO":Ede,"nl-NL":Dde,"pl-PL":$de,"pt-BR":Cde,"pt-PT":wde,"ro-RO":Rde,"ru-RU":Ade,"sk-SK":Tde,"sl-SI":kde,"sr-SP":Pde,"sv-SE":Mde,"tr-TR":Fde,"uk-UA":Nde,"zh-CN":Ide,"zh-TW":Lde};function owe(e){return e&&e.__esModule?e.default:e}const swe=new tp(owe(Bde));function cwe(){let e=typeof navigator<"u"&&(navigator.language||navigator.userLanguage)||"en-US";try{Intl.DateTimeFormat.supportedLocalesOf([e])}catch{e="en-US"}return e}function fwe(e,t,n,r,i){let a=e!=null&&n!=null&&e.compare(n)>0,u=e!=null&&t!=null&&e.compare(t)<0,l=e!=null&&(r==null?void 0:r(e))||!1,o=a||u||l,s=[];if(o){let c=cwe(),d=tp.getGlobalDictionaryForPackage("@react-stately/datepicker")||swe,h=new yN(c,d),m=new UO(c,aD({},i)),y=m.resolvedOptions().timeZone;u&&t!=null&&s.push(h.format("rangeUnderflow",{minValue:m.format(t.toDate(y))})),a&&n!=null&&s.push(h.format("rangeOverflow",{maxValue:m.format(n.toDate(y))})),l&&s.push(h.format("unavailableDate"))}return{isInvalid:o,validationErrors:s,validationDetails:{badInput:l,customError:!1,patternMismatch:!1,rangeOverflow:a,rangeUnderflow:u,stepMismatch:!1,tooLong:!1,tooShort:!1,typeMismatch:!1,valueMissing:!1,valid:!o}}}const dwe={year:"numeric",month:"numeric",day:"numeric",hour:"numeric",minute:"2-digit",second:"2-digit"},hwe={year:"numeric",month:"2-digit",day:"2-digit",hour:"2-digit",minute:"2-digit",second:"2-digit"};function aD(e,t){e={...t.shouldForceLeadingZeros?hwe:dwe,...e};let r=t.granularity||"minute",i=Object.keys(e);var a;let u=i.indexOf((a=t.maxGranularity)!==null&&a!==void 0?a:"year");u<0&&(u=0);let l=i.indexOf(r);if(l<0&&(l=2),u>l)throw new Error("maxGranularity must be greater than granularity");let o=i.slice(u,l+1).reduce((c,d)=>(c[d]=e[d],c),{});return t.hourCycle!=null&&(o.hour12=t.hourCycle===12),o.timeZone=t.timeZone||"UTC",(r==="hour"||r==="minute"||r==="second")&&t.timeZone&&!t.hideTimeZone&&(o.timeZoneName="short"),t.showEra&&u===0&&(o.era="short"),o}function Qde(e,t){if(e===null)return null;if(e)return fc(e,t)}function J4(e,t,n,r){if(e)return Qde(e,n);let i=fc(nne(r??hS()).set({hour:0,minute:0,second:0,millisecond:0}),n);return t==="year"||t==="month"||t==="day"?one(i):r?i:Wv(i)}function pwe(e,t){let n=e&&"timeZone"in e?e.timeZone:void 0,r=e&&"minute"in e?"minute":"day";if(e&&t&&!(t in e))throw new Error("Invalid granularity "+t+" for value "+e.toString());let[i,a]=P.useState([r,n]);e&&(i[0]!==r||i[1]!==n)&&a([r,n]),t||(t=e?r:i[0]);let u=e?n:i[1];return[t,u]}const mwe=new tp({ach:{year:"mwaka",month:"dwe",day:"nino"},af:{year:"jjjj",month:"mm",day:"dd"},am:{year:"ዓዓዓዓ",month:"ሚሜ",day:"ቀቀ"},an:{year:"aaaa",month:"mm",day:"dd"},ar:{year:"سنة",month:"شهر",day:"يوم"},ast:{year:"aaaa",month:"mm",day:"dd"},az:{year:"iiii",month:"aa",day:"gg"},be:{year:"гггг",month:"мм",day:"дд"},bg:{year:"гггг",month:"мм",day:"дд"},bn:{year:"yyyy",month:"মিমি",day:"dd"},br:{year:"bbbb",month:"mm",day:"dd"},bs:{year:"gggg",month:"mm",day:"dd"},ca:{year:"aaaa",month:"mm",day:"dd"},cak:{year:"jjjj",month:"ii",day:"q'q'"},ckb:{year:"ساڵ",month:"مانگ",day:"ڕۆژ"},cs:{year:"rrrr",month:"mm",day:"dd"},cy:{year:"bbbb",month:"mm",day:"dd"},da:{year:"åååå",month:"mm",day:"dd"},de:{year:"jjjj",month:"mm",day:"tt"},dsb:{year:"llll",month:"mm",day:"źź"},el:{year:"εεεε",month:"μμ",day:"ηη"},en:{year:"yyyy",month:"mm",day:"dd"},eo:{year:"jjjj",month:"mm",day:"tt"},es:{year:"aaaa",month:"mm",day:"dd"},et:{year:"aaaa",month:"kk",day:"pp"},eu:{year:"uuuu",month:"hh",day:"ee"},fa:{year:"سال",month:"ماه",day:"روز"},ff:{year:"hhhh",month:"ll",day:"ññ"},fi:{year:"vvvv",month:"kk",day:"pp"},fr:{year:"aaaa",month:"mm",day:"jj"},fy:{year:"jjjj",month:"mm",day:"dd"},ga:{year:"bbbb",month:"mm",day:"ll"},gd:{year:"bbbb",month:"mm",day:"ll"},gl:{year:"aaaa",month:"mm",day:"dd"},he:{year:"שנה",month:"חודש",day:"יום"},hr:{year:"gggg",month:"mm",day:"dd"},hsb:{year:"llll",month:"mm",day:"dd"},hu:{year:"éééé",month:"hh",day:"nn"},ia:{year:"aaaa",month:"mm",day:"dd"},id:{year:"tttt",month:"bb",day:"hh"},it:{year:"aaaa",month:"mm",day:"gg"},ja:{year:"年",month:"月",day:"日"},ka:{year:"წწწწ",month:"თთ",day:"რრ"},kk:{year:"жжжж",month:"аа",day:"кк"},kn:{year:"ವವವವ",month:"ಮಿಮೀ",day:"ದಿದಿ"},ko:{year:"연도",month:"월",day:"일"},lb:{year:"jjjj",month:"mm",day:"dd"},lo:{year:"ປປປປ",month:"ດດ",day:"ວວ"},lt:{year:"mmmm",month:"mm",day:"dd"},lv:{year:"gggg",month:"mm",day:"dd"},meh:{year:"aaaa",month:"mm",day:"dd"},ml:{year:"വർഷം",month:"മാസം",day:"തീയതി"},ms:{year:"tttt",month:"mm",day:"hh"},nb:{year:"åååå",month:"mm",day:"dd"},nl:{year:"jjjj",month:"mm",day:"dd"},nn:{year:"åååå",month:"mm",day:"dd"},no:{year:"åååå",month:"mm",day:"dd"},oc:{year:"aaaa",month:"mm",day:"jj"},pl:{year:"rrrr",month:"mm",day:"dd"},pt:{year:"aaaa",month:"mm",day:"dd"},rm:{year:"oooo",month:"mm",day:"dd"},ro:{year:"aaaa",month:"ll",day:"zz"},ru:{year:"гггг",month:"мм",day:"дд"},sc:{year:"aaaa",month:"mm",day:"dd"},scn:{year:"aaaa",month:"mm",day:"jj"},sk:{year:"rrrr",month:"mm",day:"dd"},sl:{year:"llll",month:"mm",day:"dd"},sr:{year:"гггг",month:"мм",day:"дд"},sv:{year:"åååå",month:"mm",day:"dd"},szl:{year:"rrrr",month:"mm",day:"dd"},tg:{year:"сссс",month:"мм",day:"рр"},th:{year:"ปปปป",month:"ดด",day:"วว"},tr:{year:"yyyy",month:"aa",day:"gg"},uk:{year:"рррр",month:"мм",day:"дд"},"zh-CN":{year:"年",month:"月",day:"日"},"zh-TW":{year:"年",month:"月",day:"日"}},"en");function gwe(e,t,n){return e==="era"||e==="dayPeriod"?t:e==="year"||e==="month"||e==="day"?mwe.getStringForLocale(e,n):"––"}const uD={year:!0,month:!0,day:!0,hour:!0,minute:!0,second:!0,dayPeriod:!0,era:!0},Hj={year:5,month:2,day:7,hour:2,minute:15,second:15},zde={dayperiod:"dayPeriod",relatedYear:"year",yearName:"literal",unknown:"literal"};function rI(e){let{locale:t,createCalendar:n,hideTimeZone:r,isDisabled:i=!1,isReadOnly:a=!1,isRequired:u=!1,minValue:l,maxValue:o,isDateUnavailable:s}=e,c=e.value||e.defaultValue||e.placeholderValue||null,[d,h]=pwe(c,e.granularity),m=h||"UTC";if(c&&!(d in c))throw new Error("Invalid granularity "+d+" for value "+c.toString());let y=P.useMemo(()=>new UO(t),[t]),v=P.useMemo(()=>n(y.resolvedOptions().calendar),[n,y]);var S;let[_,E]=Yu(e.value,(S=e.defaultValue)!==null&&S!==void 0?S:null,e.onChange),[$]=P.useState(_),k=P.useMemo(()=>{var me;return(me=Qde(_,v))!==null&&me!==void 0?me:null},[_,v]),[A,w]=P.useState(()=>J4(e.placeholderValue,d,v,h)),R=k||A,F=v.identifier==="gregory"&&R.era==="BC",I=P.useMemo(()=>{var me;return{granularity:d,maxGranularity:(me=e.maxGranularity)!==null&&me!==void 0?me:"year",timeZone:h,hideTimeZone:r,hourCycle:e.hourCycle,showEra:F,shouldForceLeadingZeros:e.shouldForceLeadingZeros}},[e.maxGranularity,d,e.hourCycle,e.shouldForceLeadingZeros,h,r,F]),U=P.useMemo(()=>aD({},I),[I]),q=P.useMemo(()=>new UO(t,U),[t,U]),V=P.useMemo(()=>q.resolvedOptions(),[q]),N=P.useMemo(()=>q.formatToParts(new Date).filter(me=>uD[me.type]).reduce((me,Ee)=>(me[zde[Ee.type]||Ee.type]=!0,me),{}),[q]),[j,W]=P.useState(()=>e.value||e.defaultValue?{...N}:{}),K=P.useRef(null),J=P.useRef(v);P.useEffect(()=>{tne(v,J.current)||(J.current=v,w(me=>Object.keys(j).length>0?fc(me,v):J4(e.placeholderValue,d,v,h)))},[v,d,j,h,e.placeholderValue]),_&&Object.keys(j).length<Object.keys(N).length&&(j={...N},W(j)),_==null&&Object.keys(j).length===Object.keys(N).length&&(j={},W(j),w(J4(e.placeholderValue,d,v,h)));let te=k&&Object.keys(j).length>=Object.keys(N).length?k:A,H=me=>{if(e.isDisabled||e.isReadOnly)return;let Ee=Object.keys(j),se=Object.keys(N);me==null?(E(null),w(J4(e.placeholderValue,d,v,h)),W({})):Ee.length===0&&K.current==null||Ee.length>=se.length||Ee.length===se.length-1&&N.dayPeriod&&!j.dayPeriod&&K.current!=="dayPeriod"?(Ee.length===0&&(j={...N},W(j)),me=fc(me,(c==null?void 0:c.calendar)||new Xf),E(me)):w(me),K.current=null},M=P.useMemo(()=>te.toDate(m),[te,m]),G=P.useMemo(()=>vwe(M,j,q,V,te,v,t,d),[M,j,q,V,te,v,t,d]);N.era&&j.year&&!j.era?(j.era=!0,W({...j})):!N.era&&j.era&&(delete j.era,W({...j}));let X=me=>{j[me]=!0,me==="year"&&N.era&&(j.era=!0),W({...j})},B=(me,Ee)=>{if(j[me])H(ywe(te,me,Ee,V));else{X(me);let se=Object.keys(j),pe=Object.keys(N);(se.length>=pe.length||se.length===pe.length-1&&N.dayPeriod&&!j.dayPeriod)&&H(te)}},ne=P.useMemo(()=>fwe(_,l,o,s,I),[_,l,o,s,I]),oe=Og({...e,value:_,builtinValidation:ne}),ie=oe.displayValidation.isInvalid,de=e.validationState||(ie?"invalid":null);var ge,ye;return{...oe,value:k,defaultValue:(ge=e.defaultValue)!==null&&ge!==void 0?ge:$,dateValue:M,calendar:v,setValue:H,segments:G,dateFormatter:q,validationState:de,isInvalid:ie,granularity:d,maxGranularity:(ye=e.maxGranularity)!==null&&ye!==void 0?ye:"year",isDisabled:i,isReadOnly:a,isRequired:u,increment(me){B(me,1)},decrement(me){B(me,-1)},incrementPage(me){B(me,Hj[me]||1)},decrementPage(me){B(me,-(Hj[me]||1))},setSegment(me,Ee){X(me),H(bwe(te,me,Ee,V))},confirmPlaceholder(){if(e.isDisabled||e.isReadOnly)return;let me=Object.keys(j),Ee=Object.keys(N);me.length===Ee.length-1&&N.dayPeriod&&!j.dayPeriod&&(j={...N},W(j),H(te.copy()))},clearSegment(me){delete j[me],K.current=me,W({...j});let Ee=J4(e.placeholderValue,d,v,h),se=te;if(me==="dayPeriod"&&"hour"in te&&"hour"in Ee){let pe=te.hour>=12,ve=Ee.hour>=12;pe&&!ve?se=te.set({hour:te.hour-12}):!pe&&ve&&(se=te.set({hour:te.hour+12}))}else me==="hour"&&"hour"in te&&te.hour>=12&&j.dayPeriod?se=te.set({hour:Ee.hour+12}):me in te&&(se=te.set({[me]:Ee[me]}));E(null),H(se)},formatValue(me){if(!k)return"";let Ee=aD(me,I);return new UO(t,Ee).format(M)},getDateFormatter(me,Ee){let se={...I,...Ee},pe=aD({},se);return new UO(me,pe)}}}function vwe(e,t,n,r,i,a,u,l){let o=["hour","minute","second"],s=n.formatToParts(e),c=[];for(let d of s){let h=zde[d.type]||d.type,m=uD[h];h==="era"&&a.getEras().length===1&&(m=!1);let y=uD[h]&&!t[h],v=uD[h]?gwe(h,d.value,u):null,S={type:h,text:y?v:d.value,...W_(i,h,r),isPlaceholder:y,placeholder:v,isEditable:m};h==="hour"?(c.push({type:"literal",text:"⁦",...W_(i,"literal",r),isPlaceholder:!1,placeholder:"",isEditable:!1}),c.push(S),h===l&&c.push({type:"literal",text:"⁩",...W_(i,"literal",r),isPlaceholder:!1,placeholder:"",isEditable:!1})):o.includes(h)&&h===l?(c.push(S),c.push({type:"literal",text:"⁩",...W_(i,"literal",r),isPlaceholder:!1,placeholder:"",isEditable:!1})):c.push(S)}return c}function W_(e,t,n){switch(t){case"era":{let r=e.calendar.getEras();return{value:r.indexOf(e.era),minValue:0,maxValue:r.length-1}}case"year":return{value:e.year,minValue:1,maxValue:e.calendar.getYearsInEra(e)};case"month":return{value:e.month,minValue:J_e(e),maxValue:e.calendar.getMonthsInYear(e)};case"day":return{value:e.day,minValue:eEe(e),maxValue:e.calendar.getDaysInMonth(e)}}if("hour"in e)switch(t){case"dayPeriod":return{value:e.hour>=12?12:0,minValue:0,maxValue:12};case"hour":if(n.hour12){let r=e.hour>=12;return{value:e.hour,minValue:r?12:0,maxValue:r?23:11}}return{value:e.hour,minValue:0,maxValue:23};case"minute":return{value:e.minute,minValue:0,maxValue:59};case"second":return{value:e.second,minValue:0,maxValue:59}}return{}}function ywe(e,t,n,r){switch(t){case"era":case"year":case"month":case"day":return e.cycle(t,n,{round:t==="year"})}if("hour"in e)switch(t){case"dayPeriod":{let i=e.hour,a=i>=12;return e.set({hour:a?i-12:i+12})}case"hour":case"minute":case"second":return e.cycle(t,n,{round:t!=="hour",hourCycle:r.hour12?12:24})}throw new Error("Unknown segment: "+t)}function bwe(e,t,n,r){switch(t){case"day":case"month":case"year":case"era":return e.set({[t]:n})}if("hour"in e&&typeof n=="number")switch(t){case"dayPeriod":{let i=e.hour,a=i>=12;return n>=12===a?e:e.set({hour:a?i-12:i+12})}case"hour":if(r.hour12){let a=e.hour>=12;!a&&n===12&&(n=0),a&&n<12&&(n+=12)}case"minute":case"second":return e.set({[t]:n})}throw new Error("Unknown segment: "+t)}function Owe(e){let{placeholderValue:t=new jC,minValue:n,maxValue:r,defaultValue:i,granularity:a,validate:u}=e,[l,o]=Yu(e.value,i??null,e.onChange),s=l||t,c=s&&"day"in s?s:void 0,d=i&&"timeZone"in i?i.timeZone:void 0,h=P.useMemo(()=>{let k=s&&"timeZone"in s?s.timeZone:void 0;return(k||d)&&t?sne(xO(t),k||d):xO(t)},[t,s,d]),m=P.useMemo(()=>xO(n,c),[n,c]),y=P.useMemo(()=>xO(r,c),[r,c]),v=P.useMemo(()=>l&&"day"in l?zU(l):l,[l]),S=P.useMemo(()=>l==null?null:xO(l),[l]),_=P.useMemo(()=>i==null?null:xO(i),[i]);return{...rI({...e,value:S,defaultValue:_,minValue:m,maxValue:y,onChange:k=>{o(c||d?k:k&&zU(k))},granularity:a||"minute",maxGranularity:"hour",placeholderValue:h??void 0,createCalendar:()=>new Xf,validate:P.useCallback(()=>u==null?void 0:u(l),[u,l])}),timeValue:v}}function xO(e,t=W_e(hS())){return e?"day"in e?e:Wv(t,e):null}function xwe(e){let[t,n]=Yu(e.isExpanded,e.defaultExpanded||!1,e.onExpandedChange);const r=P.useCallback(()=>{n(!0)},[n]),i=P.useCallback(()=>{n(!1)},[n]),a=P.useCallback(()=>{n(!t)},[n,t]);return{isExpanded:t,setExpanded:n,expand:r,collapse:i,toggle:a}}function Swe(e){let{allowsMultipleExpanded:t=!1,isDisabled:n=!1}=e,[r,i]=Yu(P.useMemo(()=>e.expandedKeys?new Set(e.expandedKeys):void 0,[e.expandedKeys]),P.useMemo(()=>e.defaultExpandedKeys?new Set(e.defaultExpandedKeys):new Set,[e.defaultExpandedKeys]),e.onExpandedChange);return P.useEffect(()=>{if(!t&&r.size>1){let a=r.values().next().value;a!=null&&i(new Set([a]))}}),{allowsMultipleExpanded:t,isDisabled:n,expandedKeys:r,setExpandedKeys:i,toggleKey(a){let u;t?(u=new Set(r),u.has(a)?u.delete(a):u.add(a)):u=new Set(r.has(a)?[]:[a]),i(u)}}}function iI(e){let t=Y3(e),[n,r]=P.useState(null),[i,a]=P.useState([]),u=()=>{a([]),t.close()};return{focusStrategy:n,...t,open(s=null){r(s),t.open()},toggle(s=null){r(s),t.toggle()},close(){u()},expandedKeysStack:i,openSubmenu:(s,c)=>{a(d=>c>d.length?d:[...d.slice(0,c),s])},closeSubmenu:(s,c)=>{a(d=>d[c]===s?d.slice(0,c):d)}}}function _we(e,t){let{triggerKey:n}=e,{expandedKeysStack:r,openSubmenu:i,closeSubmenu:a,close:u}=t,[l]=P.useState(r==null?void 0:r.length),o=P.useMemo(()=>r[l]===n,[r,n,l]),[s,c]=P.useState(null),d=P.useCallback(y=>{c(y??null),i(n,l)},[i,l,n]),h=P.useCallback(()=>{c(null),a(n,l)},[a,l,n]),m=P.useCallback(y=>{c(y??null),o?h():d(y)},[h,d,o]);return P.useMemo(()=>({focusStrategy:s,isOpen:o,open:d,close:h,closeAll:u,submenuLevel:l,setOpen:()=>{},toggle:m}),[o,d,h,u,s,m,l])}let Ewe=Math.round(Math.random()*1e10),Dwe=0;function $we(e){let t=P.useMemo(()=>e.name||`radio-group-${Ewe}-${++Dwe}`,[e.name]);var n;let[r,i]=Yu(e.value,(n=e.defaultValue)!==null&&n!==void 0?n:null,e.onChange),[a]=P.useState(r),[u,l]=P.useState(null),o=Og({...e,value:r}),s=h=>{!e.isReadOnly&&!e.isDisabled&&(i(h),o.commitValidation())},c=o.displayValidation.isInvalid;var d;return{...o,name:t,selectedValue:r,defaultSelectedValue:e.value!==void 0?a:(d=e.defaultValue)!==null&&d!==void 0?d:null,setSelectedValue:s,lastFocusedValue:u,setLastFocusedValue:l,isDisabled:e.isDisabled||!1,isReadOnly:e.isReadOnly||!1,isRequired:e.isRequired||!1,validationState:e.validationState||(c?"invalid":null),isInvalid:c}}function Cwe(e){let[t,n]=Yu(Xj(e.value),Xj(e.defaultValue)||"",e.onChange);return{value:t,setValue:n}}function Xj(e){if(e!=null)return e.toString()}function wwe(e){let{selectionMode:t="single"}=e,n=Y3(e),[r,i]=P.useState(null),a=P.useMemo(()=>{var k;return e.defaultValue!==void 0?e.defaultValue:t==="single"?(k=e.defaultSelectedKey)!==null&&k!==void 0?k:null:[]},[e.defaultValue,e.defaultSelectedKey,t]),u=P.useMemo(()=>e.value!==void 0?e.value:t==="single"?e.selectedKey:void 0,[e.value,e.selectedKey,t]),[l,o]=Yu(u,a,e.onChange),s=t==="single"&&Array.isArray(l)?l[0]:l,c=k=>{if(t==="single"){var A,w;let R=Array.isArray(k)?(w=k[0])!==null&&w!==void 0?w:null:k;o(R),R!==s&&((A=e.onSelectionChange)===null||A===void 0||A.call(e,R))}else{let R=[];Array.isArray(k)?R=k:k!=null&&(R=[k]),o(R)}},d=tw({...e,selectionMode:t,disallowEmptySelection:t==="single",allowDuplicateSelectionEvents:!0,selectedKeys:P.useMemo(()=>Rwe(s),[s]),onSelectionChange:k=>{if(k!=="all"){if(t==="single"){var A;let w=(A=k.values().next().value)!==null&&A!==void 0?A:null;c(w),n.close()}else c([...k]);y.commitValidation()}}}),h=d.selectionManager.firstSelectedKey,m=P.useMemo(()=>[...d.selectionManager.selectedKeys].map(k=>d.collection.getItem(k)).filter(k=>k!=null),[d.selectionManager.selectedKeys,d.collection]),y=Og({...e,value:Array.isArray(s)&&s.length===0?null:s}),[v,S]=P.useState(!1),[_]=P.useState(s);var E,$;return{...y,...d,...n,value:s,defaultValue:a??_,setValue:c,selectedKey:h,setSelectedKey:c,selectedItem:(E=m[0])!==null&&E!==void 0?E:null,selectedItems:m,defaultSelectedKey:($=e.defaultSelectedKey)!==null&&$!==void 0?$:e.selectionMode==="single"?_:null,focusStrategy:r,open(k=null){d.collection.size!==0&&(i(k),n.open())},toggle(k=null){d.collection.size!==0&&(i(k),n.toggle())},isFocused:v,setFocused:S}}function Rwe(e){if(e!==void 0)return e===null?[]:Array.isArray(e)?e:[e]}function Awe(e){var t,n;let r=rde({...e,onSelectionChange:e.onSelectionChange?o=>{var s;o!=null&&((s=e.onSelectionChange)===null||s===void 0||s.call(e,o))}:void 0,suppressTextValueWarning:!0,defaultSelectedKey:(n=(t=e.defaultSelectedKey)!==null&&t!==void 0?t:Yj(e.collection,e.disabledKeys?new Set(e.disabledKeys):new Set))!==null&&n!==void 0?n:void 0}),{selectionManager:i,collection:a,selectedKey:u}=r,l=P.useRef(u);return P.useEffect(()=>{let o=u;e.selectedKey==null&&(i.isEmpty||o==null||!a.getItem(o))&&(o=Yj(a,r.disabledKeys),o!=null&&i.setSelectedKeys([o])),(o!=null&&i.focusedKey==null||!i.isFocused&&o!==l.current)&&i.setFocusedKey(o),l.current=o}),{...r,isDisabled:e.isDisabled||!1}}function Yj(e,t){let n=null;if(e){var r,i,a,u;for(n=e.getFirstKey();n!=null&&(t.has(n)||!((i=e.getItem(n))===null||i===void 0||(r=i.props)===null||r===void 0)&&r.isDisabled)&&n!==e.getLastKey();)n=e.getKeyAfter(n);n!=null&&(t.has(n)||!((u=e.getItem(n))===null||u===void 0||(a=u.props)===null||a===void 0)&&a.isDisabled)&&n===e.getLastKey()&&(n=e.getFirstKey())}return n}function Twe(e){let t=P.useCallback(i=>e.subscribe(i),[e]),n=P.useCallback(()=>e.visibleToasts,[e]);return{visibleToasts:UN.useSyncExternalStore(t,n,n),add:(i,a)=>e.add(i,a),close:i=>e.close(i),pauseAll:()=>e.pauseAll(),resumeAll:()=>e.resumeAll()}}class fit{runWithWrapUpdate(t,n){this.wrapUpdate?this.wrapUpdate(t,n):t()}subscribe(t){return this.subscriptions.add(t),()=>this.subscriptions.delete(t)}add(t,n={}){let r="_"+Math.random().toString(36).slice(2),i={...n,content:t,key:r,timer:n.timeout?new kwe(()=>this.close(r),n.timeout):void 0};return this.queue.unshift(i),this.updateVisibleToasts("add"),r}close(t){let n=this.queue.findIndex(a=>a.key===t);if(n>=0){var r,i;(r=(i=this.queue[n]).onClose)===null||r===void 0||r.call(i),this.queue.splice(n,1)}this.updateVisibleToasts("remove")}updateVisibleToasts(t){this.visibleToasts=this.queue.slice(0,this.maxVisibleToasts),this.runWithWrapUpdate(()=>{for(let n of this.subscriptions)n()},t)}pauseAll(){for(let t of this.visibleToasts)t.timer&&t.timer.pause()}resumeAll(){for(let t of this.visibleToasts)t.timer&&t.timer.resume()}clear(){this.queue=[],this.updateVisibleToasts("clear")}constructor(t){this.queue=[],this.subscriptions=new Set,this.visibleToasts=[];var n;this.maxVisibleToasts=(n=t==null?void 0:t.maxVisibleToasts)!==null&&n!==void 0?n:1/0,this.wrapUpdate=t==null?void 0:t.wrapUpdate}}class kwe{reset(t){this.remaining=t,this.resume()}pause(){this.timerId!=null&&(clearTimeout(this.timerId),this.timerId=null,this.remaining-=Date.now()-this.startTime)}resume(){this.remaining<=0||(this.startTime=Date.now(),this.timerId=setTimeout(()=>{this.timerId=null,this.remaining=0,this.callback()},this.remaining))}constructor(t,n){this.startTime=null,this.remaining=n,this.callback=t}}const Pwe=1500,Gj=500;let gb={},Mwe=0,e2=!1,Rm=null,vb=null;function qde(e={}){let{delay:t=Pwe,closeDelay:n=Gj}=e,{isOpen:r,open:i,close:a}=Y3(e),u=P.useMemo(()=>`${++Mwe}`,[]),l=P.useRef(null),o=P.useRef(a),s=()=>{gb[u]=h},c=()=>{for(let y in gb)y!==u&&(gb[y](!0),delete gb[y])},d=()=>{l.current&&clearTimeout(l.current),l.current=null,c(),s(),e2=!0,i(),Rm&&(clearTimeout(Rm),Rm=null),vb&&(clearTimeout(vb),vb=null)},h=y=>{y||n<=0?(l.current&&clearTimeout(l.current),l.current=null,o.current()):l.current||(l.current=setTimeout(()=>{l.current=null,o.current()},n)),Rm&&(clearTimeout(Rm),Rm=null),e2&&(vb&&clearTimeout(vb),vb=setTimeout(()=>{delete gb[u],vb=null,e2=!1},Math.max(Gj,n)))},m=()=>{c(),s(),!r&&!Rm&&!e2?Rm=setTimeout(()=>{Rm=null,e2=!0,d()},t):r||d()};return P.useEffect(()=>{o.current=a},[a]),P.useEffect(()=>()=>{l.current&&clearTimeout(l.current),gb[u]&&delete gb[u]},[u]),{isOpen:r,open:y=>{!y&&t>0&&!l.current?m():d()},close:h}}class Fwe{*[Symbol.iterator](){yield*this.iterable}get size(){return this.keyMap.size}getKeys(){return this.keyMap.keys()}getKeyBefore(t){let n=this.keyMap.get(t);var r;return n&&(r=n.prevKey)!==null&&r!==void 0?r:null}getKeyAfter(t){let n=this.keyMap.get(t);var r;return n&&(r=n.nextKey)!==null&&r!==void 0?r:null}getFirstKey(){return this.firstKey}getLastKey(){return this.lastKey}getItem(t){var n;return(n=this.keyMap.get(t))!==null&&n!==void 0?n:null}at(t){const n=[...this.getKeys()];return this.getItem(n[t])}constructor(t,{expandedKeys:n}={}){this.keyMap=new Map,this.firstKey=null,this.lastKey=null,this.iterable=t,n=n||new Set;let r=l=>{if(this.keyMap.set(l.key,l),l.childNodes&&(l.type==="section"||n.has(l.key)))for(let o of l.childNodes)r(o)};for(let l of t)r(l);let i=null,a=0;for(let[l,o]of this.keyMap)i?(i.nextKey=l,o.prevKey=i.key):(this.firstKey=l,o.prevKey=void 0),o.type==="item"&&(o.index=a++),i=o,i.nextKey=void 0;var u;this.lastKey=(u=i==null?void 0:i.key)!==null&&u!==void 0?u:null}}function Nwe(e){let{onExpandedChange:t}=e,[n,r]=Yu(e.expandedKeys?new Set(e.expandedKeys):void 0,e.defaultExpandedKeys?new Set(e.defaultExpandedKeys):new Set,t),i=nI(e),a=P.useMemo(()=>e.disabledKeys?new Set(e.disabledKeys):new Set,[e.disabledKeys]),u=Bre(e,P.useCallback(o=>new Fwe(o,{expandedKeys:n}),[n]),null);return P.useEffect(()=>{i.focusedKey!=null&&!u.getItem(i.focusedKey)&&i.setFocusedKey(null)},[u,i.focusedKey]),{collection:u,expandedKeys:n,disabledKeys:a,toggleKey:o=>{r(Iwe(n,o))},setExpandedKeys:r,selectionManager:new xS(u,i)}}function Iwe(e,t){let n=new Set(e);return n.has(t)?n.delete(t):n.add(t),n}const gp=P.createContext(null),dit=P.forwardRef(function(t,n){let r=P.useContext(gp);return r!=null&&r.isInvalid?be.createElement(Lwe,{...t,ref:n}):null}),Lwe=P.forwardRef((e,t)=>{let n=P.useContext(gp),r=Gt(e,{global:!0}),i=Cn({...e,defaultClassName:"react-aria-FieldError",defaultChildren:n.validationErrors.length===0?void 0:n.validationErrors.join(" "),values:n});return i.children==null?null:be.createElement(iwe,{slot:"errorMessage",...r,...i,ref:t})}),op=P.createContext(null),hit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,op);let{validationErrors:r,validationBehavior:i="native",children:a,className:u,...l}=t;return be.createElement("form",{noValidate:i!=="native",...l,ref:n,className:u||"react-aria-Form"},be.createElement(op.Provider,{value:{...t,validationBehavior:i}},be.createElement(qre.Provider,{value:r??{}},a)))}),Bwe=P.createContext(null),pit=P.forwardRef(function(t,n){let{inputRef:r=null,...i}=t;[t,n]=Nn(i,n,TSe);let{validationBehavior:a}=xs(op)||{};var u,l;let o=(l=(u=t.validationBehavior)!==null&&u!==void 0?u:a)!==null&&l!==void 0?l:"native",s=P.useContext(Bwe),c=wu(P.useMemo(()=>yg(r,t.inputRef!==void 0?t.inputRef:null),[r,t.inputRef])),{labelProps:d,inputProps:h,isSelected:m,isDisabled:y,isReadOnly:v,isPressed:S,isInvalid:_}=s?z$e({...pc(t),value:t.value,children:typeof t.children=="function"?!0:t.children},s,c):Nue({...pc(t),children:typeof t.children=="function"?!0:t.children,validationBehavior:o},ew(t),c),{isFocused:E,isFocusVisible:$,focusProps:k}=Gu(),A=y||v,{hoverProps:w,isHovered:R}=Ru({...t,isDisabled:A}),F=Cn({...t,defaultClassName:"react-aria-Checkbox",values:{isSelected:m,isIndeterminate:t.isIndeterminate||!1,isPressed:S,isHovered:R,isFocused:E,isFocusVisible:$,isDisabled:y,isReadOnly:v,isInvalid:_,isRequired:t.isRequired||!1}}),I=Gt(t,{global:!0});return delete I.id,delete I.onClick,be.createElement("label",{...Et(I,d,w,F),ref:n,slot:t.slot||void 0,"data-selected":m||void 0,"data-indeterminate":t.isIndeterminate||void 0,"data-pressed":S||void 0,"data-hovered":R||void 0,"data-focused":E||void 0,"data-focus-visible":$||void 0,"data-disabled":y||void 0,"data-readonly":v||void 0,"data-invalid":_||void 0,"data-required":t.isRequired||void 0},be.createElement(mS,{elementType:"span"},be.createElement("input",{...Et(h,k),ref:c})),F.children)}),B0=P.createContext({}),Qwe=P.forwardRef(function(t,n){[t,n]=Nn(t,n,B0);let{isDisabled:r,isInvalid:i,isReadOnly:a,onHoverStart:u,onHoverChange:l,onHoverEnd:o,...s}=t,{hoverProps:c,isHovered:d}=Ru({onHoverStart:u,onHoverChange:l,onHoverEnd:o,isDisabled:r}),{isFocused:h,isFocusVisible:m,focusProps:y}=Gu({within:!0});r??(r=!!t["aria-disabled"]&&t["aria-disabled"]!=="false"),i??(i=!!t["aria-invalid"]&&t["aria-invalid"]!=="false");let v=Cn({...t,values:{isHovered:d,isFocusWithin:h,isFocusVisible:m,isDisabled:r,isInvalid:i},defaultClassName:"react-aria-Group"});var S,_;return be.createElement("div",{...Et(s,y,c),...v,ref:n,role:(S=t.role)!==null&&S!==void 0?S:"group",slot:(_=t.slot)!==null&&_!==void 0?_:void 0,"data-focus-within":h||void 0,"data-hovered":d||void 0,"data-focus-visible":m||void 0,"data-disabled":r||void 0,"data-invalid":i||void 0,"data-readonly":a||void 0},v.children)}),Q0=P.createContext({});let zwe=e=>{let{onHoverStart:t,onHoverChange:n,onHoverEnd:r,...i}=e;return i};const qwe=C1(function(t,n){[t,n]=Nn(t,n,Q0);let{hoverProps:r,isHovered:i}=Ru(t),{isFocused:a,isFocusVisible:u,focusProps:l}=Gu({isTextInput:!0,autoFocus:t.autoFocus}),o=!!t["aria-invalid"]&&t["aria-invalid"]!=="false",s=Cn({...t,values:{isHovered:i,isFocused:a,isFocusVisible:u,isDisabled:t.disabled||!1,isInvalid:o},defaultClassName:"react-aria-Input"});return be.createElement("input",{...Et(zwe(t),l,r),...s,ref:n,"data-focused":a||void 0,"data-disabled":t.disabled||void 0,"data-hovered":i||void 0,"data-focus-visible":u||void 0,"data-invalid":o||void 0})}),Vwe=P.createContext(null),mit=P.forwardRef(function(t,n){return[t,n]=Nn(t,n,kSe),t.channel?be.createElement(Uwe,{...t,channel:t.channel,forwardedRef:n}):be.createElement(jwe,{...t,forwardedRef:n})});function Uwe(e){let{locale:t}=Wi(),n=rCe({...e,locale:t}),r=P.useRef(null),[i,a]=Ss(!e["aria-label"]&&!e["aria-labelledby"]);var u;let{labelProps:l,inputProps:o,descriptionProps:s,errorMessageProps:c,...d}=dCe({...pc(e),label:a,validationBehavior:(u=e.validationBehavior)!==null&&u!==void 0?u:"native"},n,r);return be.createElement(be.Fragment,null,Vde(e,n,e.forwardedRef,o,r,l,i,s,c,d),e.name&&be.createElement("input",{type:"hidden",name:e.name,form:e.form,value:isNaN(n.numberValue)?"":n.numberValue}))}function jwe(e){var t;let n=nCe({...e,validationBehavior:(t=e.validationBehavior)!==null&&t!==void 0?t:"native"}),r=P.useRef(null),[i,a]=Ss(!e["aria-label"]&&!e["aria-labelledby"]);var u;let{labelProps:l,inputProps:o,descriptionProps:s,errorMessageProps:c,...d}=oCe({...pc(e),label:a,validationBehavior:(u=e.validationBehavior)!==null&&u!==void 0?u:"native"},n,r);return Vde(e,n,e.forwardedRef,o,r,l,i,s,c,d)}function Vde(e,t,n,r,i,a,u,l,o,s){let c=Cn({...e,values:{state:t,channel:e.channel||"hex",isDisabled:e.isDisabled||!1,isInvalid:s.isInvalid||!1},defaultClassName:"react-aria-ColorField"}),d=Gt(e,{global:!0});return delete d.id,be.createElement(Ni,{values:[[Vwe,t],[Q0,{...r,ref:i}],[_s,{...a,ref:u}],[B0,{role:"presentation",isInvalid:s.isInvalid,isDisabled:e.isDisabled||!1}],[zo,{slots:{description:l,errorMessage:o}}],[gp,s]]},be.createElement("div",{...d,...c,ref:n,slot:e.slot||void 0,"data-channel":e.channel||"hex","data-disabled":e.isDisabled||void 0,"data-invalid":s.isInvalid||void 0}))}const Ude=P.createContext(null),git=P.forwardRef(function(t,n){[t,n]=Nn(t,n,Ude);let{colorSwatchProps:r,color:i}=cCe(t),a=Cn({...t,defaultClassName:"react-aria-ColorSwatch",defaultStyle:r.style,values:{color:i}}),u=Gt(t,{global:!0});return be.createElement("div",{...Et(u,r,a),slot:t.slot||void 0,ref:n})});var jde={};jde={colorSwatchPicker:"تغييرات الألوان",dropzoneLabel:"DropZone",selectPlaceholder:"حدد عنصرًا",tableResizer:"أداة تغيير الحجم"};var Zde={};Zde={colorSwatchPicker:"Цветови мостри",dropzoneLabel:"DropZone",selectPlaceholder:"Изберете предмет",tableResizer:"Преоразмерител"};var Hde={};Hde={colorSwatchPicker:"Vzorky barev",dropzoneLabel:"Místo pro přetažení",selectPlaceholder:"Vyberte položku",tableResizer:"Změna velikosti"};var Xde={};Xde={colorSwatchPicker:"Farveprøver",dropzoneLabel:"DropZone",selectPlaceholder:"Vælg et element",tableResizer:"Størrelsesændring"};var Yde={};Yde={colorSwatchPicker:"Farbfelder",dropzoneLabel:"Ablegebereich",selectPlaceholder:"Element wählen",tableResizer:"Größenanpassung"};var Gde={};Gde={colorSwatchPicker:"Χρωματικά δείγματα",dropzoneLabel:"DropZone",selectPlaceholder:"Επιλέξτε ένα αντικείμενο",tableResizer:"Αλλαγή μεγέθους"};var Kde={};Kde={selectPlaceholder:"Select an item",tableResizer:"Resizer",dropzoneLabel:"DropZone",colorSwatchPicker:"Color swatches"};var Wde={};Wde={colorSwatchPicker:"Muestras de colores",dropzoneLabel:"DropZone",selectPlaceholder:"Seleccionar un artículo",tableResizer:"Cambiador de tamaño"};var Jde={};Jde={colorSwatchPicker:"Värvinäidised",dropzoneLabel:"DropZone",selectPlaceholder:"Valige üksus",tableResizer:"Suuruse muutja"};var ehe={};ehe={colorSwatchPicker:"Värimallit",dropzoneLabel:"DropZone",selectPlaceholder:"Valitse kohde",tableResizer:"Koon muuttaja"};var the={};the={colorSwatchPicker:"Échantillons de couleurs",dropzoneLabel:"DropZone",selectPlaceholder:"Sélectionner un élément",tableResizer:"Redimensionneur"};var nhe={};nhe={colorSwatchPicker:"דוגמיות צבע",dropzoneLabel:"DropZone",selectPlaceholder:"בחר פריט",tableResizer:"שינוי גודל"};var rhe={};rhe={colorSwatchPicker:"Uzorci boja",dropzoneLabel:"Zona spuštanja",selectPlaceholder:"Odaberite stavku",tableResizer:"Promjena veličine"};var ihe={};ihe={colorSwatchPicker:"Színtárak",dropzoneLabel:"DropZone",selectPlaceholder:"Válasszon ki egy elemet",tableResizer:"Átméretező"};var ahe={};ahe={colorSwatchPicker:"Campioni di colore",dropzoneLabel:"Zona di rilascio",selectPlaceholder:"Seleziona un elemento",tableResizer:"Ridimensionamento"};var uhe={};uhe={colorSwatchPicker:"カラースウォッチ",dropzoneLabel:"ドロップゾーン",selectPlaceholder:"項目を選択",tableResizer:"サイズ変更ツール"};var lhe={};lhe={colorSwatchPicker:"색상 견본",dropzoneLabel:"드롭 영역",selectPlaceholder:"항목 선택",tableResizer:"크기 조정기"};var ohe={};ohe={colorSwatchPicker:"Spalvų pavyzdžiai",dropzoneLabel:"„DropZone“",selectPlaceholder:"Pasirinkite elementą",tableResizer:"Dydžio keitiklis"};var she={};she={colorSwatchPicker:"Krāsu paraugi",dropzoneLabel:"DropZone",selectPlaceholder:"Izvēlēties vienumu",tableResizer:"Izmēra mainītājs"};var che={};che={colorSwatchPicker:"Fargekart",dropzoneLabel:"Droppsone",selectPlaceholder:"Velg et element",tableResizer:"Størrelsesendrer"};var fhe={};fhe={colorSwatchPicker:"kleurstalen",dropzoneLabel:"DropZone",selectPlaceholder:"Selecteer een item",tableResizer:"Resizer"};var dhe={};dhe={colorSwatchPicker:"Próbki kolorów",dropzoneLabel:"Strefa upuszczania",selectPlaceholder:"Wybierz element",tableResizer:"Zmiana rozmiaru"};var hhe={};hhe={colorSwatchPicker:"Amostras de cores",dropzoneLabel:"DropZone",selectPlaceholder:"Selecione um item",tableResizer:"Redimensionador"};var phe={};phe={colorSwatchPicker:"Amostras de cores",dropzoneLabel:"DropZone",selectPlaceholder:"Selecione um item",tableResizer:"Redimensionador"};var mhe={};mhe={colorSwatchPicker:"Specimene de culoare",dropzoneLabel:"Zonă de plasare",selectPlaceholder:"Selectați un element",tableResizer:"Instrument de redimensionare"};var ghe={};ghe={colorSwatchPicker:"Цветовые образцы",dropzoneLabel:"DropZone",selectPlaceholder:"Выберите элемент",tableResizer:"Средство изменения размера"};var vhe={};vhe={colorSwatchPicker:"Vzorkovníky farieb",dropzoneLabel:"DropZone",selectPlaceholder:"Vyberte položku",tableResizer:"Nástroj na zmenu veľkosti"};var yhe={};yhe={colorSwatchPicker:"Barvne palete",dropzoneLabel:"DropZone",selectPlaceholder:"Izberite element",tableResizer:"Spreminjanje velikosti"};var bhe={};bhe={colorSwatchPicker:"Uzorci boje",dropzoneLabel:"DropZone",selectPlaceholder:"Izaberite stavku",tableResizer:"Promena veličine"};var Ohe={};Ohe={colorSwatchPicker:"Färgrutor",dropzoneLabel:"DropZone",selectPlaceholder:"Välj en artikel",tableResizer:"Storleksändrare"};var xhe={};xhe={colorSwatchPicker:"Renk örnekleri",dropzoneLabel:"Bırakma Bölgesi",selectPlaceholder:"Bir öğe seçin",tableResizer:"Yeniden boyutlandırıcı"};var She={};She={colorSwatchPicker:"Зразки кольорів",dropzoneLabel:"DropZone",selectPlaceholder:"Виберіть елемент",tableResizer:"Засіб змінення розміру"};var _he={};_he={colorSwatchPicker:"颜色色板",dropzoneLabel:"放置区域",selectPlaceholder:"选择一个项目",tableResizer:"尺寸调整器"};var Ehe={};Ehe={colorSwatchPicker:"色票",dropzoneLabel:"放置區",selectPlaceholder:"選取項目",tableResizer:"大小調整器"};var aI={};aI={"ar-AE":jde,"bg-BG":Zde,"cs-CZ":Hde,"da-DK":Xde,"de-DE":Yde,"el-GR":Gde,"en-US":Kde,"es-ES":Wde,"et-EE":Jde,"fi-FI":ehe,"fr-FR":the,"he-IL":nhe,"hr-HR":rhe,"hu-HU":ihe,"it-IT":ahe,"ja-JP":uhe,"ko-KR":lhe,"lt-LT":ohe,"lv-LV":she,"nb-NO":che,"nl-NL":fhe,"pl-PL":dhe,"pt-BR":hhe,"pt-PT":phe,"ro-RO":mhe,"ru-RU":ghe,"sk-SK":vhe,"sl-SI":yhe,"sr-SP":bhe,"sv-SE":Ohe,"tr-TR":xhe,"uk-UA":She,"zh-CN":_he,"zh-TW":Ehe};const nw=P.createContext({}),Dhe=P.createContext(null),Zwe=P.forwardRef(function(t,n){let{render:r}=P.useContext(Dhe);return be.createElement(be.Fragment,null,r(t,n))});function $he(e,t){var n;let r=e==null?void 0:e.renderDropIndicator,i=e==null||(n=e.isVirtualDragging)===null||n===void 0?void 0:n.call(e),a=P.useCallback(u=>{if(i||t!=null&&t.isDropTarget(u))return r?r(u):be.createElement(Zwe,{target:u})},[t==null?void 0:t.target,i,r]);return e!=null&&e.useDropIndicator?a:void 0}function Hwe(e,t,n){var r,i;let a=e.focusedKey,u=null;if(!(t==null||(r=t.isVirtualDragging)===null||r===void 0)&&r.call(t)&&(n==null||(i=n.target)===null||i===void 0?void 0:i.type)==="item"&&(u=n.target.key,n.target.dropPosition==="after")){let d=n.collection.getKeyAfter(u),h=null;if(d!=null){var l,o;let m=(o=(l=n.collection.getItem(u))===null||l===void 0?void 0:l.level)!==null&&o!==void 0?o:0;for(;d;){let y=n.collection.getItem(d);if(!y)break;if(y.type!=="item"){d=n.collection.getKeyAfter(d);continue}var s;if(((s=y.level)!==null&&s!==void 0?s:0)<=m)break;h=d,d=n.collection.getKeyAfter(d)}}var c;u=(c=d??h)!==null&&c!==void 0?c:u}return P.useMemo(()=>new Set([a,u].filter(d=>d!=null)),[a,u])}const Che=P.createContext({}),Xwe=P.createContext(null);function G3(e){let t=P.useRef({});return be.createElement(Xwe.Provider,{value:t},e.children)}const K3=P.createContext({isSelected:!1}),uI=P.createContext({});class whe extends gy{filter(t,n){let r=n.getItem(this.prevKey);if(r&&r.type!=="separator"){let i=this.clone();return n.addDescendants(i,t),i}return null}}whe.type="separator";const vit=R1(whe,function(t,n){[t,n]=Nn(t,n,uI);let{elementType:r,orientation:i,style:a,className:u,slot:l,...o}=t,s=r||"hr";s==="hr"&&i==="vertical"&&(s="div");let{separatorProps:c}=zCe({...o,elementType:r,orientation:i}),d=Gt(t,{global:!0});return be.createElement(s,{...Et(d,c),style:a,className:u??"react-aria-Separator",ref:n,slot:l||void 0})}),SS=P.createContext(null),uh=P.createContext(null),Ywe=P.forwardRef(function(t,n){[t,n]=Nn(t,n,SS);let r=P.useContext(uh);return r?be.createElement(Rhe,{state:r,props:t,listBoxRef:n}):be.createElement(w1,{content:be.createElement(OS,t)},i=>be.createElement(Gwe,{props:t,listBoxRef:n,collection:i}))});function Gwe({props:e,listBoxRef:t,collection:n}){e={...e,collection:n,children:null,items:null};let{layoutDelegate:r}=P.useContext(mp),i=tw({...e,layoutDelegate:r});return be.createElement(Rhe,{state:i,props:e,listBoxRef:t})}function Rhe({state:e,props:t,listBoxRef:n}){[t,n]=Nn(t,n,vx);let{dragAndDropHooks:r,layout:i="stack",orientation:a="vertical",filter:u}=t,l=tde(e,u),{collection:o,selectionManager:s}=l,c=!!(r!=null&&r.useDraggableCollectionState),d=!!(r!=null&&r.useDroppableCollectionState),{direction:h}=Wi(),{disabledBehavior:m,disabledKeys:y}=s,v=ZC({usage:"search",sensitivity:"base"}),{isVirtualized:S,layoutDelegate:_,dropTargetDelegate:E,CollectionRoot:$}=P.useContext(mp),k=P.useMemo(()=>t.keyboardDelegate||new yS({collection:o,collator:v,ref:n,disabledKeys:y,disabledBehavior:m,layout:i,orientation:a,direction:h,layoutDelegate:_}),[o,v,n,m,y,a,h,t.keyboardDelegate,i,_]),{listBoxProps:A}=ZDe({...t,shouldSelectOnPressUp:c||t.shouldSelectOnPressUp,keyboardDelegate:k,isVirtualized:S},l,n);P.useRef(c),P.useRef(d),P.useEffect(()=>{},[c,d]);let w,R,F,I=!1,U=null,q=P.useRef(null);if(c&&r){w=r.useDraggableCollectionState({collection:o,selectionManager:s,preview:r.renderDragPreview?q:void 0}),r.useDraggableCollection({},w,n);let M=r.DragPreview;U=r.renderDragPreview?be.createElement(M,{ref:q},r.renderDragPreview):null}if(d&&r){R=r.useDroppableCollectionState({collection:o,selectionManager:s});let M=r.dropTargetDelegate||E||new r.ListDropTargetDelegate(o,n,{orientation:a,layout:i,direction:h});F=r.useDroppableCollection({keyboardDelegate:k,dropTargetDelegate:M},R,n),I=R.isDropTarget({type:"root"})}let{focusProps:V,isFocused:N,isFocusVisible:j}=Gu(),W=l.collection.size===0,K={isDropTarget:I,isEmpty:W,isFocused:N,isFocusVisible:j,layout:t.layout||"stack",state:l},J=Cn({className:t.className,style:t.style,defaultClassName:"react-aria-ListBox",values:K}),te=null;W&&t.renderEmptyState&&(te=be.createElement("div",{role:"option",style:{display:"contents"}},t.renderEmptyState(K)));let H=Gt(t,{global:!0});return be.createElement(PN,null,be.createElement("div",{...Et(H,J,A,V,F==null?void 0:F.collectionProps),ref:n,slot:t.slot||void 0,onScroll:t.onScroll,"data-drop-target":I||void 0,"data-empty":W||void 0,"data-focused":N||void 0,"data-focus-visible":j||void 0,"data-layout":t.layout||"stack","data-orientation":t.orientation||"vertical"},be.createElement(Ni,{values:[[SS,t],[uh,l],[nw,{dragAndDropHooks:r,dragState:w,dropState:R}],[uI,{elementType:"div"}],[Dhe,{render:Jwe}],[Gae,{name:"ListBoxSection",render:Kwe}]]},be.createElement(G3,null,be.createElement($,{collection:o,scrollRef:n,persistedKeys:Hwe(s,r,R),renderDropIndicator:$he(r,R)}))),te,U))}function Kwe(e,t,n,r="react-aria-ListBoxSection"){let i=P.useContext(uh),{dragAndDropHooks:a,dropState:u}=P.useContext(nw),{CollectionBranch:l}=P.useContext(mp),[o,s]=Ss();var c;let{headingProps:d,groupProps:h}=WDe({heading:s,"aria-label":(c=e["aria-label"])!==null&&c!==void 0?c:void 0}),m=Cn({defaultClassName:r,className:e.className,style:e.style,values:{}}),y=Gt(e,{global:!0});return delete y.id,be.createElement("section",{...Et(y,m,h),ref:t},be.createElement(Che.Provider,{value:{...d,ref:o}},be.createElement(l,{collection:i.collection,parent:n,renderDropIndicator:$he(a,u)})))}const Wwe=R1(JC,function(t,n,r){let i=wu(n),a=P.useContext(uh),{dragAndDropHooks:u,dragState:l,dropState:o}=P.useContext(nw),{optionProps:s,labelProps:c,descriptionProps:d,...h}=KDe({key:r.key,"aria-label":t==null?void 0:t["aria-label"]},a,i),{hoverProps:m,isHovered:y}=Ru({isDisabled:!h.allowsSelection&&!h.hasAction,onHoverStart:r.props.onHoverStart,onHoverChange:r.props.onHoverChange,onHoverEnd:r.props.onHoverEnd}),v=null;l&&u&&(v=u.useDraggableItem({key:r.key},l));let S=null;o&&u&&(S=u.useDroppableItem({target:{type:"item",key:r.key,dropPosition:"on"}},o,i));let _=l&&l.isDragging(r.key),E=Cn({...t,id:void 0,children:t.children,defaultClassName:"react-aria-ListBoxItem",values:{...h,isHovered:y,selectionMode:a.selectionManager.selectionMode,selectionBehavior:a.selectionManager.selectionBehavior,allowsDragging:!!l,isDragging:_,isDropTarget:S==null?void 0:S.isDropTarget}});P.useEffect(()=>{r.textValue},[r.textValue]);let $=t.href?"a":"div",k=Gt(t,{global:!0});return delete k.id,delete k.onClick,be.createElement($,{...Et(k,E,s,m,v==null?void 0:v.dragProps,S==null?void 0:S.dropProps),ref:i,"data-allows-dragging":!!l||void 0,"data-selected":h.isSelected||void 0,"data-disabled":h.isDisabled||void 0,"data-hovered":y||void 0,"data-focused":h.isFocused||void 0,"data-focus-visible":h.isFocusVisible||void 0,"data-pressed":h.isPressed||void 0,"data-dragging":_||void 0,"data-drop-target":(S==null?void 0:S.isDropTarget)||void 0,"data-selection-mode":a.selectionManager.selectionMode==="none"?void 0:a.selectionManager.selectionMode},be.createElement(Ni,{values:[[zo,{slots:{[Kv]:c,label:c,description:d}}],[K3,{isSelected:h.isSelected}]]},E.children))});function Jwe(e,t){t=wu(t);let{dragAndDropHooks:n,dropState:r}=P.useContext(nw),{dropIndicatorProps:i,isHidden:a,isDropTarget:u}=n.useDropIndicator(e,r,t);return a?null:be.createElement(t5e,{...e,dropIndicatorProps:i,isDropTarget:u,ref:t})}function e5e(e,t){let{dropIndicatorProps:n,isDropTarget:r,...i}=e,a=Cn({...i,defaultClassName:"react-aria-DropIndicator",values:{isDropTarget:r}});return be.createElement("div",{...n,...a,role:"option",ref:t,"data-drop-target":r||void 0})}const t5e=P.forwardRef(e5e);R1(Vae,function(t,n,r){let i=P.useContext(uh),{isLoading:a,onLoadMore:u,scrollOffset:l,...o}=t,s=P.useRef(null),c=P.useMemo(()=>({onLoadMore:u,collection:i==null?void 0:i.collection,sentinelRef:s,scrollOffset:l}),[u,l,i==null?void 0:i.collection]);b_e(c,s);let d=Cn({...o,id:void 0,children:r.rendered,defaultClassName:"react-aria-ListBoxLoadingIndicator",values:null}),h={tabIndex:-1};return be.createElement(be.Fragment,null,be.createElement("div",{style:{position:"relative",width:0,height:0},inert:_9(!0)},be.createElement("div",{"data-testid":"loadMoreSentinel",ref:s,style:{position:"absolute",height:1,width:1}})),a&&d.children&&be.createElement("div",{...Et(Gt(t,{global:!0}),h),...d,role:"option",ref:n},d.children))});function n5e(e){return e&&e.__esModule?e.default:e}const r5e=P.createContext(null),Ahe=P.createContext(null),yit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,r5e);let r=iCe(t),i=P.useMemo(()=>new Map,[]),a=Qo(n5e(aI),"react-aria-components");return be.createElement(Ywe,{...Gt(t,{labelable:!0}),ref:n,className:t.className||"react-aria-ColorSwatchPicker",style:t.style,"aria-label":t["aria-label"]||(t["aria-labelledby"]?void 0:a.format("colorSwatchPicker")),layout:t.layout||"grid",selectionMode:"single",selectedKeys:[r.color.toString("hexa")],onSelectionChange:u=>{u!=="all"&&r.setColor(i.get([...u][0]))},disallowEmptySelection:!0},be.createElement(Ahe.Provider,{value:i},t.children))}),bit=P.forwardRef(function(t,n){let r=t.color||"#0000",i=P.useMemo(()=>typeof r=="string"?vy(r):r,[r]),{locale:a}=Wi(),u=P.useContext(Ahe);P.useEffect(()=>{let o=i.toString("hexa");return u.set(o,i),()=>{u.delete(o)}},[i,u]);let l=o=>typeof o=="function"?s=>o({...s,color:i}):o;return be.createElement(Wwe,{...t,ref:n,id:i.toString("hexa"),textValue:i.getColorName(a),className:l(t.className||"react-aria-ColorSwatchPickerItem"),style:l(t.style)},$_e(l(t.children),o=>be.createElement(Ude.Provider,{value:{color:i}},o)))}),i5e=P.createContext(null),rw=P.createContext(null),lI=P.createContext(null),The=P.createContext(null),Oit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,i5e);let r=P.useRef(null),i=t$(t.formatOptions),a=J$e({...t,numberFormatter:i}),[u,l]=Ss(!t["aria-label"]&&!t["aria-labelledby"]),{groupProps:o,trackProps:s,labelProps:c,outputProps:d}=aCe({...t,label:l},a,r),h=Cn({...t,values:{orientation:a.orientation,isDisabled:a.isDisabled,state:a},defaultClassName:"react-aria-Slider"}),m=Gt(t,{global:!0});return delete m.id,be.createElement(Ni,{values:[[rw,a],[lI,{...s,ref:r}],[The,d],[_s,{...c,ref:u}]]},be.createElement("div",{...Et(m,h,o),ref:n,slot:t.slot||void 0,"data-orientation":a.orientation,"data-disabled":a.isDisabled||void 0}))}),xit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,The);let{children:r,style:i,className:a,...u}=t,l=P.useContext(rw),o=Cn({className:a,style:i,children:r,defaultChildren:l.getThumbValueLabel(0),defaultClassName:"react-aria-SliderOutput",values:{orientation:l.orientation,isDisabled:l.isDisabled,state:l}});return be.createElement("output",{...u,...o,ref:n,"data-orientation":l.orientation||void 0,"data-disabled":l.isDisabled||void 0})}),Sit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,lI);let r=P.useContext(rw),{onHoverStart:i,onHoverEnd:a,onHoverChange:u,...l}=t,{hoverProps:o,isHovered:s}=Ru({onHoverStart:i,onHoverEnd:a,onHoverChange:u}),c=Cn({...t,defaultClassName:"react-aria-SliderTrack",values:{orientation:r.orientation,isDisabled:r.isDisabled,isHovered:s,state:r}});return be.createElement("div",{...Et(l,o),...c,ref:n,"data-hovered":s||void 0,"data-orientation":r.orientation||void 0,"data-disabled":r.isDisabled||void 0})}),_it=P.forwardRef(function(t,n){let{inputRef:r=null}=t,i=P.useContext(rw),{ref:a}=xs(lI),{index:u=0}=t,l=P.useRef(null),o=r||l,[s,c]=Ss(!t["aria-label"]&&!t["aria-labelledby"]),{thumbProps:d,inputProps:h,labelProps:m,isDragging:y,isFocused:v,isDisabled:S}=uCe({...t,index:u,trackRef:a,inputRef:o,label:c},i),{focusProps:_,isFocusVisible:E}=Gu(),{hoverProps:$,isHovered:k}=Ru(t),A=Cn({...t,defaultClassName:"react-aria-SliderThumb",values:{state:i,isHovered:k,isDragging:y,isFocused:v,isFocusVisible:E,isDisabled:S}}),w=Gt(t,{global:!0});return delete w.id,be.createElement("div",{...Et(w,d,$),...A,ref:n,style:{...d.style,...A.style},"data-hovered":k||void 0,"data-dragging":y||void 0,"data-focused":v||void 0,"data-focus-visible":E||void 0,"data-disabled":S||void 0},be.createElement(mS,null,be.createElement("input",{ref:o,...Et(h,_)})),be.createElement(Ni,{values:[[_s,{...m,ref:s}]]},A.children))}),oI=P.createContext({placement:"bottom"}),Eit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,oI);let r=t.placement,i={position:"absolute",transform:r==="top"||r==="bottom"?"translateX(-50%)":"translateY(-50%)"};r!=null&&(i[r]="100%");let a=Cn({...t,defaultClassName:"react-aria-OverlayArrow",values:{placement:r}});a.style&&Object.keys(a.style).forEach(l=>a.style[l]===void 0&&delete a.style[l]);let u=Gt(t);return be.createElement("div",{...u,...a,style:{...i,...a.style},ref:n,"data-placement":r})}),W3=P.createContext(null),Kj=P.createContext(null),Dit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,W3);let r=P.useContext(z0),i=Y3(t),a=t.isOpen!=null||t.defaultOpen!=null||!r?i:r,u=XD(n,a.isOpen)||t.isExiting||!1,l=$$e(),{direction:o}=Wi();if(l){let s=t.children;return typeof s=="function"&&(s=s({trigger:t.trigger||null,placement:"bottom",isEntering:!1,isExiting:!1,defaultChildren:null})),be.createElement(be.Fragment,null,s)}return a&&!a.isOpen&&!u?null:be.createElement(a5e,{...t,triggerRef:t.triggerRef,state:a,popoverRef:n,isExiting:u,dir:o})});function a5e({state:e,isExiting:t,UNSTABLE_portalContainer:n,clearContexts:r,...i}){let a=P.useRef(null),u=P.useRef(null),l=P.useContext(Kj),o=l&&i.trigger==="SubmenuTrigger";var s;let{popoverProps:c,underlayProps:d,arrowProps:h,placement:m,triggerAnchorPoint:y}=BDe({...i,offset:(s=i.offset)!==null&&s!==void 0?s:8,arrowRef:a,groupRef:o?l:u},e),v=i.popoverRef,S=qC(v,!!m)||i.isEntering||!1,_=Cn({...i,defaultClassName:"react-aria-Popover",values:{trigger:i.trigger||null,placement:m,isEntering:S,isExiting:t}}),E=!i.isNonModal||i.trigger==="SubmenuTrigger",[$,k]=P.useState(!1);Hr(()=>{v.current&&k(E&&!v.current.querySelector("[role=dialog]"))},[v,E]),P.useEffect(()=>{$&&i.trigger!=="SubmenuTrigger"&&v.current&&!v.current.contains(document.activeElement)&&lc(v.current)},[$,v,i.trigger]);let A=P.useMemo(()=>{let I=_.children;if(r)for(let U of r)I=be.createElement(U.Provider,{value:null},I);return I},[_.children,r]),w={...c.style,"--trigger-anchor-point":y?`${y.x}px ${y.y}px`:void 0,..._.style},R=be.createElement("div",{...Et(Gt(i,{global:!0}),c),..._,role:$?"dialog":void 0,tabIndex:$?-1:void 0,"aria-label":i["aria-label"],"aria-labelledby":i["aria-labelledby"],ref:v,slot:i.slot||void 0,style:w,dir:i.dir,"data-trigger":i.trigger,"data-placement":m,"data-entering":S||void 0,"data-exiting":t||void 0},!i.isNonModal&&be.createElement(j9,{onDismiss:e.close}),be.createElement(oI.Provider,{value:{...h,placement:m,ref:a}},A),be.createElement(j9,{onDismiss:e.close}));if(!o)return be.createElement(Z9,{...i,shouldContainFocus:$,isExiting:t,portalContainer:n},!i.isNonModal&&e.isOpen&&be.createElement("div",{"data-testid":"underlay",...d,style:{position:"fixed",inset:0}}),be.createElement("div",{ref:u,style:{display:"contents"}},be.createElement(Kj.Provider,{value:u},R)));var F;return be.createElement(Z9,{...i,shouldContainFocus:$,isExiting:t,portalContainer:(F=n??(l==null?void 0:l.current))!==null&&F!==void 0?F:void 0},R)}const u5e=P.createContext({}),sI=P.createContext(null),iw=P.createContext(null),Ex=P.createContext(null),c$=P.createContext(null);function $it(e){let t=iI(e),n=P.useRef(null),{menuTriggerProps:r,menuProps:i}=VN({...e,type:"menu"},t,n),[a,u]=P.useState(null),l=P.useCallback(()=>{n.current&&u(n.current.offsetWidth+"px")},[n]);u1({ref:n,onResize:l});let o=P.useRef(null);return be.createElement(Ni,{values:[[sI,{...i,ref:o}],[z0,t],[Ex,t],[W3,{trigger:"MenuTrigger",triggerRef:n,scrollRef:o,placement:"bottom start",style:{"--trigger-width":a},"aria-labelledby":i["aria-labelledby"]}]]},be.createElement(qne,{...r,ref:n,isPressed:t.isOpen},e.children))}const khe=P.createContext(null);class Phe extends gy{filter(t,n,r){let i=t.getItem(this.firstChildKey);if(i&&r(i.textValue,this)){let a=this.clone();return n.addDescendants(a,t),a}return null}}Phe.type="submenutrigger";const Cit=k$e(Phe,(e,t,n)=>{let{CollectionBranch:r}=P.useContext(mp),i=P.useContext(iw),a=P.useContext(Ex),u=_we({triggerKey:n.key},a),l=P.useRef(null),o=wu(t),{parentMenuRef:s,shouldUseVirtualFocus:c}=P.useContext(khe),{submenuTriggerProps:d,submenuProps:h,popoverProps:m}=o$e({parentMenuRef:s,submenuRef:l,delay:e.delay,shouldUseVirtualFocus:c},u,o);return be.createElement(Ni,{values:[[f$,{...d,onAction:void 0,ref:o}],[sI,{ref:l,...h}],[z0,u],[W3,{trigger:"SubmenuTrigger",triggerRef:o,placement:"end top","aria-labelledby":h["aria-labelledby"],...m}]]},be.createElement(r,{collection:i.collection,parent:n}),e.children[1])},e=>e.children[0]),wit=P.forwardRef(function(t,n){return[t,n]=Nn(t,n,sI),be.createElement(w1,{content:be.createElement(OS,t)},r=>be.createElement(l5e,{props:t,collection:r,menuRef:n}))});function l5e({props:e,collection:t,menuRef:n}){[e,n]=Nn(e,n,vx);let{filter:r,...i}=e,a=P.useMemo(()=>r?t.filter(r):t,[t,r]),u=Nwe({...e,collection:a,children:void 0}),l=P.useContext(Ex),{isVirtualized:o,CollectionRoot:s}=P.useContext(mp),{menuProps:c}=n$e({...e,isVirtualized:o,onClose:e.onClose||(l==null?void 0:l.close)},u,n),d=Cn({defaultClassName:"react-aria-Menu",className:e.className,style:e.style,values:{isEmpty:u.collection.size===0}}),h=null;u.collection.size===0&&e.renderEmptyState&&(h=be.createElement("div",{role:"menuitem",style:{display:"contents"}},e.renderEmptyState()));let m=Gt(e,{global:!0});return be.createElement(PN,null,be.createElement("div",{...Et(m,d,c),ref:n,slot:e.slot||void 0,"data-empty":u.collection.size===0||void 0,onScroll:e.onScroll},be.createElement(Ni,{values:[[iw,u],[uI,{elementType:"div"}],[Gae,{name:"MenuSection",render:s5e}],[khe,{parentMenuRef:n,shouldUseVirtualFocus:i==null?void 0:i.shouldUseVirtualFocus}],[f$,null],[vx,null],[LC,null],[c$,u.selectionManager],[Ex,l??iI({})]]},be.createElement(G3,null,be.createElement(s,{collection:u.collection,persistedKeys:ZN(u.selectionManager.focusedKey),scrollRef:n}))),h))}class o5e extends xS{get focusedKey(){return this.parent.focusedKey}get isFocused(){return this.parent.isFocused}setFocusedKey(t,n){return this.parent.setFocusedKey(t,n)}setFocused(t){this.parent.setFocused(t)}get childFocusStrategy(){return this.parent.childFocusStrategy}constructor(t,n){super(t.collection,n),this.parent=t}}function s5e(e,t,n,r="react-aria-MenuSection"){var i,a;let u=P.useContext(iw),{CollectionBranch:l}=P.useContext(mp),[o,s]=Ss();var c;let{headingProps:d,groupProps:h}=i$e({heading:s,"aria-label":(c=n.props["aria-label"])!==null&&c!==void 0?c:void 0}),m=Cn({defaultClassName:r,className:(i=n.props)===null||i===void 0?void 0:i.className,style:(a=n.props)===null||a===void 0?void 0:a.style,values:{}}),y=P.useContext(c$),v=nI(e),S=e.selectionMode!=null?new o5e(y,v):y,_=Gt(e,{global:!0});return delete _.id,be.createElement("section",{...Et(_,m,h),ref:t},be.createElement(Ni,{values:[[Che,{...d,ref:o}],[c$,S]]},be.createElement(l,{collection:u.collection,parent:n})))}const f$=P.createContext(null),Rit=R1(JC,function(t,n,r){var i;[t,n]=Nn(t,n,f$);let a=(i=xs(f$))===null||i===void 0?void 0:i.id,u=P.useContext(iw),l=wu(n),o=P.useContext(c$),{menuItemProps:s,labelProps:c,descriptionProps:d,keyboardShortcutProps:h,...m}=r$e({...t,id:a,key:r.key,selectionManager:o},u,l),{hoverProps:y,isHovered:v}=Ru({isDisabled:m.isDisabled}),S=Cn({...t,id:void 0,children:r.rendered,defaultClassName:"react-aria-MenuItem",values:{...m,isHovered:v,isFocusVisible:m.isFocusVisible,selectionMode:o.selectionMode,selectionBehavior:o.selectionBehavior,hasSubmenu:!!t["aria-haspopup"],isOpen:t["aria-expanded"]==="true"}}),_=t.href?"a":"div",E=Gt(t,{global:!0});return delete E.id,delete E.onClick,be.createElement(_,{...Et(E,S,s,y),ref:l,"data-disabled":m.isDisabled||void 0,"data-hovered":v||void 0,"data-focused":m.isFocused||void 0,"data-focus-visible":m.isFocusVisible||void 0,"data-pressed":m.isPressed||void 0,"data-selected":m.isSelected||void 0,"data-selection-mode":o.selectionMode==="none"?void 0:o.selectionMode,"data-has-submenu":!!t["aria-haspopup"]||void 0,"data-open":t["aria-expanded"]==="true"||void 0},be.createElement(Ni,{values:[[zo,{slots:{[Kv]:c,label:c,description:d}}],[u5e,h],[K3,{isSelected:m.isSelected}]]},S.children))}),Mhe=P.createContext(null),z0=P.createContext(null);function Ait(e){let t=iI(e),n=P.useRef(null),{triggerProps:r,overlayProps:i}=Kne({type:"dialog"},t,n),[a,u]=P.useState(null),l=P.useCallback(()=>{n.current&&u(n.current.offsetWidth+"px")},[n]);return u1({ref:n,onResize:l}),r.id=vi(),i["aria-labelledby"]=r.id,be.createElement(Ni,{values:[[z0,t],[Ex,t],[Mhe,i],[W3,{trigger:"DialogTrigger",triggerRef:n,"aria-labelledby":i["aria-labelledby"],style:{"--trigger-width":a}}]]},be.createElement(qne,{...r,ref:n,isPressed:t.isOpen},e.children))}const Tit=P.forwardRef(function(t,n){let r=t["aria-labelledby"];[t,n]=Nn(t,n,Mhe);let{dialogProps:i,titleProps:a}=SCe({...t,"aria-labelledby":r},n),u=P.useContext(z0);!i["aria-label"]&&!i["aria-labelledby"]&&t["aria-labelledby"]&&(i["aria-labelledby"]=t["aria-labelledby"]);let l=Cn({defaultClassName:"react-aria-Dialog",className:t.className,style:t.style,children:t.children,values:{close:(u==null?void 0:u.close)||(()=>{})}}),o=Gt(t,{global:!0});return be.createElement("section",{...Et(o,l,i),ref:n,slot:t.slot||void 0},be.createElement(Ni,{values:[[Kee,{slots:{[Kv]:{},title:{...a,level:2}}}],[ah,{slots:{[Kv]:{},close:{onPress:()=>u==null?void 0:u.close()}}}]]},l.children))}),c5e=P.createContext(null),f5e=P.createContext(null),kit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,c5e);let{children:r,isDisabled:i=!1,isInvalid:a=!1,isRequired:u=!1}=t,l=P.useMemo(()=>{var o;return be.createElement(SS.Provider,{value:{items:(o=t.items)!==null&&o!==void 0?o:t.defaultItems}},typeof r=="function"?r({isOpen:!1,isDisabled:i,isInvalid:a,isRequired:u,defaultChildren:null}):r)},[r,i,a,u,t.items,t.defaultItems]);return be.createElement(w1,{content:l},o=>be.createElement(h5e,{props:t,collection:o,comboBoxRef:n}))}),d5e=[_s,ah,Q0,B0,zo];function h5e({props:e,collection:t,comboBoxRef:n}){let{name:r,formValue:i="key",allowsCustomValue:a}=e;a&&(i="text");let{validationBehavior:u}=xs(op)||{};var l,o;let s=(o=(l=e.validationBehavior)!==null&&l!==void 0?l:u)!==null&&o!==void 0?o:"native",{contains:c}=Tne({sensitivity:"base"}),d=uwe({defaultFilter:e.defaultFilter||c,...e,items:e.items,children:void 0,collection:t,validationBehavior:s}),h=P.useRef(null),m=P.useRef(null),y=P.useRef(null),v=P.useRef(null),[S,_]=Ss(!e["aria-label"]&&!e["aria-labelledby"]),{buttonProps:E,inputProps:$,listBoxProps:k,labelProps:A,descriptionProps:w,errorMessageProps:R,...F}=h$e({...pc(e),label:_,inputRef:m,buttonRef:h,listBoxRef:y,popoverRef:v,name:i==="text"?r:void 0,validationBehavior:s},d),[I,U]=P.useState(null),q=P.useCallback(()=>{if(m.current){var K;let J=(K=h.current)===null||K===void 0?void 0:K.getBoundingClientRect(),te=m.current.getBoundingClientRect(),H=J?Math.min(J.left,te.left):te.left,M=J?Math.max(J.right,te.right):te.right;U(M-H+"px")}},[h,m,U]);u1({ref:m,onResize:q});let V=P.useMemo(()=>({isOpen:d.isOpen,isDisabled:e.isDisabled||!1,isInvalid:F.isInvalid||!1,isRequired:e.isRequired||!1}),[d.isOpen,e.isDisabled,F.isInvalid,e.isRequired]),N=Cn({...e,values:V,defaultClassName:"react-aria-ComboBox"}),j=Gt(e,{global:!0});delete j.id;var W;return be.createElement(Ni,{values:[[f5e,d],[_s,{...A,ref:S}],[ah,{...E,ref:h,isPressed:d.isOpen}],[Q0,{...$,ref:m}],[z0,d],[W3,{ref:v,triggerRef:m,scrollRef:y,placement:"bottom start",isNonModal:!0,trigger:"ComboBox",style:{"--trigger-width":I},clearContexts:d5e}],[SS,{...k,ref:y}],[uh,d],[zo,{slots:{description:w,errorMessage:R}}],[B0,{isInvalid:F.isInvalid,isDisabled:e.isDisabled||!1}],[gp,F]]},be.createElement("div",{...j,...N,ref:n,slot:e.slot||void 0,"data-focused":d.isFocused||void 0,"data-open":d.isOpen||void 0,"data-disabled":e.isDisabled||void 0,"data-invalid":F.isInvalid||void 0,"data-required":e.isRequired||void 0}),r&&i==="key"&&be.createElement("input",{type:"hidden",name:r,form:e.form,value:(W=d.selectedKey)!==null&&W!==void 0?W:""}))}const p5e=["day","month","year"],m5e={hour:1,minute:2,second:3};function g5e(e,t){let{autoComplete:n,isDisabled:r,name:i}=e,{visuallyHiddenProps:a}=LN({style:{position:"fixed",top:0,left:0}}),u=60;t.granularity==="second"?u=1:t.granularity==="hour"&&(u=3600);let l=t.value==null?"":t.value.toString(),o=t.granularity==="day"?"date":"datetime-local",s=["hour","minute","second"],c=0;return s.includes(t.granularity)&&(c=m5e[t.granularity],s=s.slice(0,c)),{containerProps:{...a,"aria-hidden":!0,"data-react-aria-prevent-focus":!0,"data-a11y-ignore":"aria-hidden-focus"},inputProps:{tabIndex:-1,autoComplete:n,disabled:r,type:o,form:"",name:i,step:u,value:l,onChange:d=>{let h=d.target.value.toString();if(h)try{let m=pEe(h);if(t.granularity==="day"&&(m=hEe(h)),"setSegment"in t)for(let y in m)p5e.includes(y)&&t.setSegment(y,m[y]),s.includes(y)&&t.setSegment(y,m[y]);t.setValue(m)}catch{}}}}}function v5e(e){let{state:t}=e,{containerProps:n,inputProps:r}=g5e({...e},t);return be.createElement("div",{...n,"data-testid":"hidden-dateinput-container"},be.createElement("input",r))}const Fhe=P.createContext(null),y5e=P.createContext(null),_S=P.createContext(null),aw=P.createContext(null),Pit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,Fhe);let{validationBehavior:r}=xs(op)||{};var i,a;let u=(a=(i=t.validationBehavior)!==null&&i!==void 0?i:r)!==null&&a!==void 0?a:"native",{locale:l}=Wi(),o=rI({...t,locale:l,createCalendar:wne,validationBehavior:u}),s=P.useRef(null),[c,d]=Ss(!t["aria-label"]&&!t["aria-labelledby"]),h=P.useRef(null),{labelProps:m,fieldProps:y,inputProps:v,descriptionProps:S,errorMessageProps:_,...E}=KN({...pc(t),label:d,inputRef:h,validationBehavior:u},o,s),$=Cn({...pc(t),values:{state:o,isInvalid:o.isInvalid,isDisabled:o.isDisabled,isReadOnly:o.isReadOnly},defaultClassName:"react-aria-DateField"}),k=Gt(t,{global:!0});return delete k.id,be.createElement(Ni,{values:[[_S,o],[B0,{...y,ref:s,isInvalid:o.isInvalid,isDisabled:o.isDisabled}],[Q0,{...v,ref:h}],[_s,{...m,ref:c,elementType:"span"}],[zo,{slots:{description:S,errorMessage:_}}],[gp,E]]},be.createElement("div",{...k,...$,ref:n,slot:t.slot||void 0,"data-invalid":o.isInvalid||void 0,"data-disabled":o.isDisabled||void 0,"data-readonly":o.isReadOnly||void 0}),be.createElement(v5e,{autoComplete:t.autoComplete,name:t.name,isDisabled:t.isDisabled,state:o}))}),Mit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,y5e);let{validationBehavior:r}=xs(op)||{};var i,a;let u=(a=(i=t.validationBehavior)!==null&&i!==void 0?i:r)!==null&&a!==void 0?a:"native",{locale:l}=Wi(),o=Owe({...t,locale:l,validationBehavior:u}),s=P.useRef(null),[c,d]=Ss(!t["aria-label"]&&!t["aria-labelledby"]),h=P.useRef(null),{labelProps:m,fieldProps:y,inputProps:v,descriptionProps:S,errorMessageProps:_,...E}=gCe({...pc(t),label:d,inputRef:h,validationBehavior:u},o,s),$=Cn({...t,values:{state:o,isInvalid:o.isInvalid,isDisabled:o.isDisabled,isReadOnly:o.isReadOnly},defaultClassName:"react-aria-TimeField"}),k=Gt(t,{global:!0});return delete k.id,be.createElement(Ni,{values:[[aw,o],[B0,{...y,ref:s,isInvalid:o.isInvalid,isDisabled:o.isDisabled}],[Q0,{...v,ref:h}],[_s,{...m,ref:c,elementType:"span"}],[zo,{slots:{description:S,errorMessage:_}}],[gp,E]]},be.createElement("div",{...k,...$,ref:n,slot:t.slot||void 0,"data-invalid":o.isInvalid||void 0,"data-disabled":o.isDisabled||void 0,"data-readonly":o.isReadOnly||void 0}))}),Fit=P.forwardRef(function(t,n){let r=P.useContext(_S),i=P.useContext(aw);return r||i?be.createElement(Nhe,{...t,ref:n}):be.createElement(b5e,{...t,ref:n})}),b5e=P.forwardRef((e,t)=>{let[n,r]=Nn({slot:e.slot},t,Fhe),{locale:i}=Wi(),a=rI({...n,locale:i,createCalendar:wne}),u=P.useRef(null),{fieldProps:l,inputProps:o}=KN({...n,inputRef:u},a,r);return be.createElement(Ni,{values:[[_S,a],[Q0,{...o,ref:u}],[B0,{...l,ref:r,isInvalid:a.isInvalid,isDisabled:a.isDisabled}]]},be.createElement(Nhe,e))}),Nhe=P.forwardRef((e,t)=>{let{className:n,children:r}=e,i=P.useContext(_S),a=P.useContext(aw),u=i??a;return be.createElement(be.Fragment,null,be.createElement(Qwe,{...e,ref:t,slot:e.slot||void 0,className:n??"react-aria-DateInput",isReadOnly:u.isReadOnly,isInvalid:u.isInvalid,isDisabled:u.isDisabled},u.segments.map((l,o)=>P.cloneElement(r(l),{key:o}))),be.createElement(qwe,null))}),Nit=P.forwardRef(function({segment:t,...n},r){let i=P.useContext(_S),a=P.useContext(aw),u=i??a,l=wu(r),{segmentProps:o}=OCe(t,u,l),{focusProps:s,isFocused:c,isFocusVisible:d}=Gu(),{hoverProps:h,isHovered:m}=Ru({...n,isDisabled:u.isDisabled||t.type==="literal"}),y=Cn({...n,values:{...t,isReadOnly:u.isReadOnly,isInvalid:u.isInvalid,isDisabled:u.isDisabled,isHovered:m,isFocused:c,isFocusVisible:d},defaultChildren:t.text,defaultClassName:"react-aria-DateSegment"});return be.createElement("span",{...Et(Gt(n,{global:!0}),o,s,h),...y,style:o.style,ref:l,"data-placeholder":t.isPlaceholder||void 0,"data-invalid":u.isInvalid||void 0,"data-readonly":u.isReadOnly||void 0,"data-disabled":u.isDisabled||void 0,"data-type":t.type,"data-hovered":m||void 0,"data-focused":c||void 0,"data-focus-visible":d||void 0})}),Ihe=P.createContext(null),Iit=P.forwardRef(function(t,n){let r=Swe(t),i=Cn({...t,defaultClassName:"react-aria-DisclosureGroup",values:{isDisabled:r.isDisabled,state:r}}),a=Gt(t,{global:!0});return be.createElement("div",{...a,...i,ref:n,"data-disabled":t.isDisabled||void 0},be.createElement(Ihe.Provider,{value:r},i.children))}),O5e=P.createContext(null),x5e=P.createContext(null),Lhe=P.createContext(null),Lit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,O5e);let r=P.useContext(Ihe),{id:i,...a}=t,u=vi();i||(i=u);let l=r?r.expandedKeys.has(i):t.isExpanded,o=xwe({...t,isExpanded:l,onExpandedChange(_){var E;r&&r.toggleKey(i),(E=t.onExpandedChange)===null||E===void 0||E.call(t,_)}}),s=be.useRef(null),c=t.isDisabled||(r==null?void 0:r.isDisabled)||!1,{buttonProps:d,panelProps:h}=_Ce({...t,isExpanded:l,isDisabled:c},o,s),{isFocusVisible:m,focusProps:y}=Gu({within:!0}),v=Cn({...t,id:void 0,defaultClassName:"react-aria-Disclosure",values:{isExpanded:o.isExpanded,isDisabled:c,isFocusVisibleWithin:m,state:o}}),S=Gt(a,{global:!0});return be.createElement(Ni,{values:[[ah,{slots:{[Kv]:{},trigger:d}}],[Lhe,{panelProps:h,panelRef:s}],[x5e,o]]},be.createElement("div",{...Et(S,v,y),ref:n,"data-expanded":o.isExpanded||void 0,"data-disabled":c||void 0,"data-focus-visible-within":m||void 0},v.children))}),Bit=P.forwardRef(function(t,n){let{role:r="group"}=t,{panelProps:i,panelRef:a}=P.useContext(Lhe),{isFocusVisible:u,focusProps:l}=Gu({within:!0}),o=Cn({...t,defaultClassName:"react-aria-DisclosurePanel",values:{isFocusVisibleWithin:u}}),s=Gt(t,{global:!0});return be.createElement("div",{...Et(s,o,i,l),ref:yg(n,a),role:r,"data-focus-visible-within":u||void 0},be.createElement(Ni,{values:[[ah,null]]},t.children))}),Qit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,Kee);let{children:r,level:i=3,className:a,...u}=t,l=`h${i}`;return be.createElement(l,{...u,ref:n,className:a??"react-aria-Heading"},r)}),S5e=P.createContext(null),cI=P.createContext(null),zit=P.forwardRef(function(t,n){if(P.useContext(cI))return be.createElement(Wj,{...t,modalRef:n},t.children);let{isDismissable:i,isKeyboardDismissDisabled:a,isOpen:u,defaultOpen:l,onOpenChange:o,children:s,isEntering:c,isExiting:d,UNSTABLE_portalContainer:h,shouldCloseOnInteractOutside:m,...y}=t;return be.createElement(E5e,{isDismissable:i,isKeyboardDismissDisabled:a,isOpen:u,defaultOpen:l,onOpenChange:o,isEntering:c,isExiting:d,UNSTABLE_portalContainer:h,shouldCloseOnInteractOutside:m},be.createElement(Wj,{...y,modalRef:n},s))});function _5e(e,t){[e,t]=Nn(e,t,S5e);let n=P.useContext(z0),r=Y3(e),i=e.isOpen!=null||e.defaultOpen!=null||!n?r:n,a=wu(t),u=P.useRef(null),l=XD(a,i.isOpen),o=XD(u,i.isOpen),s=l||o||e.isExiting||!1,c=pp();return!i.isOpen&&!s||c?null:be.createElement(D5e,{...e,state:i,isExiting:s,overlayRef:a,modalRef:u})}const E5e=P.forwardRef(_5e);function D5e({UNSTABLE_portalContainer:e,...t}){let n=t.modalRef,{state:r}=t,{modalProps:i,underlayProps:a}=QDe(t,r,n),u=qC(t.overlayRef)||t.isEntering||!1,l=Cn({...t,defaultClassName:"react-aria-ModalOverlay",values:{isEntering:u,isExiting:t.isExiting,state:r}}),o=m_e(),s;if(typeof document<"u"){let d=l1(document.body)?document.body:document.scrollingElement||document.documentElement,h=d.getBoundingClientRect().height%1;s=d.scrollHeight-h}let c={...l.style,"--visual-viewport-height":o.height+"px","--page-height":s!==void 0?s+"px":void 0};return be.createElement(Z9,{isExiting:t.isExiting,portalContainer:e},be.createElement("div",{...Et(Gt(t,{global:!0}),a),...l,style:c,ref:t.overlayRef,"data-entering":u||void 0,"data-exiting":t.isExiting||void 0},be.createElement(Ni,{values:[[cI,{modalProps:i,modalRef:n,isExiting:t.isExiting,isDismissable:t.isDismissable}],[z0,r]]},l.children)))}function Wj(e){let{modalProps:t,modalRef:n,isExiting:r,isDismissable:i}=P.useContext(cI),a=P.useContext(z0),u=P.useMemo(()=>yg(e.modalRef,n),[e.modalRef,n]),l=wu(u),o=qC(l),s=Cn({...e,defaultClassName:"react-aria-Modal",values:{isEntering:o,isExiting:r,state:a}});return be.createElement("div",{...Et(Gt(e,{global:!0}),t),...s,ref:l,"data-entering":o||void 0,"data-exiting":r||void 0},i&&be.createElement(j9,{onDismiss:a.close}),s.children)}const $5e=P.createContext(null),C5e=P.createContext(null),qit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,$5e);let{validationBehavior:r}=xs(op)||{};var i,a;let u=(a=(i=t.validationBehavior)!==null&&i!==void 0?i:r)!==null&&a!==void 0?a:"native",{locale:l}=Wi(),o=toe({...t,locale:l,validationBehavior:u}),s=P.useRef(null),[c,d]=Ss(!t["aria-label"]&&!t["aria-labelledby"]),{labelProps:h,groupProps:m,inputProps:y,incrementButtonProps:v,decrementButtonProps:S,descriptionProps:_,errorMessageProps:E,...$}=yse({...pc(t),label:d,validationBehavior:u},o,s),k=Cn({...t,values:{state:o,isDisabled:t.isDisabled||!1,isInvalid:$.isInvalid||!1,isRequired:t.isRequired||!1},defaultClassName:"react-aria-NumberField"}),A=Gt(t,{global:!0});return delete A.id,be.createElement(Ni,{values:[[C5e,o],[B0,m],[Q0,{...y,ref:s}],[_s,{...h,ref:c}],[ah,{slots:{increment:v,decrement:S}}],[zo,{slots:{description:_,errorMessage:E}}],[gp,$]]},be.createElement("div",{...A,...k,ref:n,slot:t.slot||void 0,"data-disabled":t.isDisabled||void 0,"data-required":t.isRequired||void 0,"data-invalid":$.isInvalid||void 0}),t.name&&be.createElement("input",{type:"hidden",name:t.name,form:t.form,value:isNaN(o.numberValue)?"":o.numberValue,disabled:t.isDisabled||void 0}))}),w5e=P.createContext(null),R5e=P.createContext(null),Bhe=P.createContext(null),Vit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,w5e);let{validationBehavior:r}=xs(op)||{};var i,a;let u=(a=(i=t.validationBehavior)!==null&&i!==void 0?i:r)!==null&&a!==void 0?a:"native",l=$we({...t,validationBehavior:u}),[o,s]=Ss(!t["aria-label"]&&!t["aria-labelledby"]),{radioGroupProps:c,labelProps:d,descriptionProps:h,errorMessageProps:m,...y}=ICe({...t,label:s,validationBehavior:u},l),v=Cn({...t,values:{orientation:t.orientation||"vertical",isDisabled:l.isDisabled,isReadOnly:l.isReadOnly,isRequired:l.isRequired,isInvalid:l.isInvalid,state:l},defaultClassName:"react-aria-RadioGroup"}),S=Gt(t,{global:!0});return be.createElement("div",{...Et(S,v,c),ref:n,slot:t.slot||void 0,"data-orientation":t.orientation||"vertical","data-invalid":l.isInvalid||void 0,"data-disabled":l.isDisabled||void 0,"data-readonly":l.isReadOnly||void 0,"data-required":l.isRequired||void 0},be.createElement(Ni,{values:[[Bhe,l],[_s,{...d,ref:o,elementType:"span"}],[zo,{slots:{description:h,errorMessage:m}}],[gp,y]]},be.createElement(G3,null,v.children)))}),Uit=P.forwardRef(function(t,n){let{inputRef:r=null,...i}=t;[t,n]=Nn(i,n,R5e);let a=be.useContext(Bhe),u=wu(P.useMemo(()=>yg(r,t.inputRef!==void 0?t.inputRef:null),[r,t.inputRef])),{labelProps:l,inputProps:o,isSelected:s,isDisabled:c,isPressed:d}=NCe({...pc(t),children:typeof t.children=="function"?!0:t.children},a,u),{isFocused:h,isFocusVisible:m,focusProps:y}=Gu(),v=c||a.isReadOnly,{hoverProps:S,isHovered:_}=Ru({...t,isDisabled:v}),E=Cn({...t,defaultClassName:"react-aria-Radio",values:{isSelected:s,isPressed:d,isHovered:_,isFocused:h,isFocusVisible:m,isDisabled:c,isReadOnly:a.isReadOnly,isInvalid:a.isInvalid,isRequired:a.isRequired}}),$=Gt(t,{global:!0});return delete $.id,delete $.onClick,be.createElement("label",{...Et($,l,S,E),ref:n,"data-selected":s||void 0,"data-pressed":d||void 0,"data-hovered":_||void 0,"data-focused":h||void 0,"data-focus-visible":m||void 0,"data-disabled":c||void 0,"data-readonly":a.isReadOnly||void 0,"data-invalid":a.isInvalid||void 0,"data-required":a.isRequired||void 0},be.createElement(mS,{elementType:"span"},be.createElement("input",{...Et(o,y),ref:u})),be.createElement(K3.Provider,{value:{isSelected:s}},E.children))}),A5e=P.createContext(null),jit=C1(function(t,n){[t,n]=Nn(t,n,A5e);let{validationBehavior:r}=xs(op)||{};var i,a;let u=(a=(i=t.validationBehavior)!==null&&i!==void 0?i:r)!==null&&a!==void 0?a:"native",l=P.useRef(null);[t,l]=Nn(t,l,LC);let[o,s]=Ss(!t["aria-label"]&&!t["aria-labelledby"]),c=Cwe({...t,validationBehavior:u}),{labelProps:d,inputProps:h,clearButtonProps:m,descriptionProps:y,errorMessageProps:v,...S}=m$e({...pc(t),label:s,validationBehavior:u},c,l),_=Cn({...t,values:{isEmpty:c.value==="",isDisabled:t.isDisabled||!1,isInvalid:S.isInvalid||!1,state:c},defaultClassName:"react-aria-SearchField"}),E=Gt(t,{global:!0});return delete E.id,be.createElement("div",{...E,..._,ref:n,slot:t.slot||void 0,"data-empty":c.value===""||void 0,"data-disabled":t.isDisabled||void 0,"data-invalid":S.isInvalid||void 0},be.createElement(Ni,{values:[[_s,{...d,ref:o}],[Q0,{...h,ref:l}],[ah,m],[zo,{slots:{description:y,errorMessage:v}}],[B0,{isInvalid:S.isInvalid,isDisabled:t.isDisabled||!1}],[gp,S]]},_.children))});function T5e(e){return e&&e.__esModule?e.default:e}const fI=P.createContext(null),Qhe=P.createContext(null),Zit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,fI);let{children:r,isDisabled:i=!1,isInvalid:a=!1,isRequired:u=!1}=t,l=P.useMemo(()=>typeof r=="function"?r({isOpen:!1,isDisabled:i,isInvalid:a,isRequired:u,isFocused:!1,isFocusVisible:!1,defaultChildren:null}):r,[r,i,a,u]);return be.createElement(w1,{content:l},o=>be.createElement(P5e,{props:t,collection:o,selectRef:n}))}),k5e=[_s,ah,zo];function P5e({props:e,selectRef:t,collection:n}){let{validationBehavior:r}=xs(op)||{};var i,a;let u=(a=(i=e.validationBehavior)!==null&&i!==void 0?i:r)!==null&&a!==void 0?a:"native",l=wwe({...e,collection:n,children:void 0,validationBehavior:u}),{isFocusVisible:o,focusProps:s}=Gu({within:!0}),c=P.useRef(null),[d,h]=Ss(!e["aria-label"]&&!e["aria-labelledby"]),{labelProps:m,triggerProps:y,valueProps:v,menuProps:S,descriptionProps:_,errorMessageProps:E,hiddenSelectProps:$,...k}=LCe({...pc(e),label:h,validationBehavior:u},l,c),[A,w]=P.useState(null),R=P.useCallback(()=>{c.current&&w(c.current.offsetWidth+"px")},[c]);u1({ref:c,onResize:R});let F=P.useMemo(()=>({isOpen:l.isOpen,isFocused:l.isFocused,isFocusVisible:o,isDisabled:e.isDisabled||!1,isInvalid:k.isInvalid||!1,isRequired:e.isRequired||!1}),[l.isOpen,l.isFocused,o,e.isDisabled,k.isInvalid,e.isRequired]),I=Cn({...e,values:F,defaultClassName:"react-aria-Select"}),U=Gt(e,{global:!0});delete U.id;let q=P.useRef(null);return be.createElement(Ni,{values:[[fI,e],[Qhe,l],[zhe,v],[_s,{...m,ref:d,elementType:"span"}],[ah,{...y,ref:c,isPressed:l.isOpen,autoFocus:e.autoFocus}],[z0,l],[W3,{trigger:"Select",triggerRef:c,scrollRef:q,placement:"bottom start",style:{"--trigger-width":A},"aria-labelledby":S["aria-labelledby"],clearContexts:k5e}],[SS,{...S,ref:q}],[uh,l],[zo,{slots:{description:_,errorMessage:E}}],[gp,k]]},be.createElement("div",{...Et(U,I,s),ref:t,slot:e.slot||void 0,"data-focused":l.isFocused||void 0,"data-focus-visible":o||void 0,"data-open":l.isOpen||void 0,"data-disabled":e.isDisabled||void 0,"data-invalid":k.isInvalid||void 0,"data-required":e.isRequired||void 0},I.children,be.createElement(QCe,{...$,autoComplete:e.autoComplete})))}const zhe=P.createContext(null),Hit=C1(function(t,n){var r;[t,n]=Nn(t,n,zhe);let i=P.useContext(Qhe),{placeholder:a}=xs(fI),u=i.selectedItems.map(_=>{var E;let $=(E=_.props)===null||E===void 0?void 0:E.children;return typeof $=="function"&&($=$({isHovered:!1,isPressed:!1,isSelected:!1,isFocused:!1,isFocusVisible:!1,isDisabled:!1,selectionMode:"single",selectionBehavior:"toggle"})),$}),l=G_e(),o=P.useMemo(()=>i.selectedItems.map(_=>_==null?void 0:_.textValue),[i.selectedItems]),s=i.selectionManager.selectionMode,c=P.useMemo(()=>{var _;return s==="single"?(_=o[0])!==null&&_!==void 0?_:"":l.format(o)},[s,l,o]),d=P.useMemo(()=>{if(s==="single")return u[0];let _=l.formatToParts(o);if(_.length===0)return null;let E=0;return _.map($=>$.type==="element"?be.createElement(P.Fragment,{key:E},u[E++]):$.value)},[s,l,o,u]),h=Qo(T5e(aI),"react-aria-components");var m,y;let v=Cn({...t,defaultChildren:(m=d??a)!==null&&m!==void 0?m:h.format("selectPlaceholder"),defaultClassName:"react-aria-SelectValue",values:{selectedItem:(y=(r=i.selectedItems[0])===null||r===void 0?void 0:r.value)!==null&&y!==void 0?y:null,selectedItems:P.useMemo(()=>i.selectedItems.map(_=>{var E;return(E=_.value)!==null&&E!==void 0?E:null}),[i.selectedItems]),selectedText:c,isPlaceholder:i.selectedItems.length===0,state:i}}),S=Gt(t,{global:!0});return be.createElement("span",{ref:n,...S,...v,"data-placeholder":i.selectedItems.length===0||void 0},be.createElement(zo.Provider,{value:void 0},v.children))}),M5e=P.createContext(null),Xit=P.forwardRef(function(t,n){let{inputRef:r=null,...i}=t;[t,n]=Nn(i,n,M5e);let a=wu(yg(r,t.inputRef!==void 0?t.inputRef:null)),u=ew(t),{labelProps:l,inputProps:o,isSelected:s,isDisabled:c,isReadOnly:d,isPressed:h}=qCe({...pc(t),children:typeof t.children=="function"?!0:t.children},u,a),{isFocused:m,isFocusVisible:y,focusProps:v}=Gu(),S=t.isDisabled||t.isReadOnly,{hoverProps:_,isHovered:E}=Ru({...t,isDisabled:S}),$=Cn({...t,defaultClassName:"react-aria-Switch",values:{isSelected:s,isPressed:h,isHovered:E,isFocused:m,isFocusVisible:y,isDisabled:c,isReadOnly:d,state:u}}),k=Gt(t,{global:!0});return delete k.id,delete k.onClick,be.createElement("label",{...Et(k,l,_,$),ref:n,slot:t.slot||void 0,"data-selected":s||void 0,"data-pressed":h||void 0,"data-hovered":E||void 0,"data-focused":m||void 0,"data-focus-visible":y||void 0,"data-disabled":c||void 0,"data-readonly":d||void 0},be.createElement(mS,{elementType:"span"},be.createElement("input",{...Et(o,v),ref:a})),$.children)}),uw=P.createContext(null),T3=P.createContext(null),Yit=P.forwardRef(function(t,n){[t,n]=Nn(t,n,uw);let{children:r,orientation:i="horizontal"}=t;return r=P.useMemo(()=>typeof r=="function"?r({orientation:i,defaultChildren:null}):r,[r,i]),be.createElement(w1,{content:r},a=>be.createElement(F5e,{props:t,collection:a,tabsRef:n}))});function F5e({props:e,tabsRef:t,collection:n}){let{orientation:r="horizontal"}=e,i=Awe({...e,collection:n,children:void 0}),{focusProps:a,isFocused:u,isFocusVisible:l}=Gu({within:!0}),o=P.useMemo(()=>({orientation:r,isFocusWithin:u,isFocusVisible:l}),[r,u,l]),s=Cn({...e,defaultClassName:"react-aria-Tabs",values:o}),c=Gt(e,{global:!0});return be.createElement("div",{...Et(c,s,a),ref:t,slot:e.slot||void 0,"data-focused":u||void 0,"data-orientation":r,"data-focus-visible":l||void 0,"data-disabled":i.isDisabled||void 0},be.createElement(Ni,{values:[[uw,e],[T3,i]]},s.children))}const Git=P.forwardRef(function(t,n){return P.useContext(T3)?be.createElement(N5e,{props:t,forwardedRef:n}):be.createElement(OS,t)});function N5e({props:e,forwardedRef:t}){let n=P.useContext(T3),{CollectionRoot:r}=P.useContext(mp),{orientation:i="horizontal",keyboardActivation:a="automatic"}=xs(uw),u=wu(t),{tabListProps:l}=ZCe({...e,orientation:i,keyboardActivation:a},n,u),o=Cn({...e,children:null,defaultClassName:"react-aria-TabList",values:{orientation:i,state:n}}),s=Gt(e,{global:!0});return delete s.id,be.createElement("div",{...Et(s,o,l),ref:u,"data-orientation":i||void 0},be.createElement(G3,null,be.createElement(r,{collection:n.collection,persistedKeys:ZN(n.selectionManager.focusedKey)})))}class qhe extends gy{}qhe.type="item";const Kit=R1(qhe,(e,t,n)=>{let r=P.useContext(T3),i=wu(t),{tabProps:a,isSelected:u,isDisabled:l,isPressed:o}=VCe({key:n.key,...e},r,i),{focusProps:s,isFocused:c,isFocusVisible:d}=Gu(),{hoverProps:h,isHovered:m}=Ru({isDisabled:l,onHoverStart:e.onHoverStart,onHoverEnd:e.onHoverEnd,onHoverChange:e.onHoverChange}),y=Cn({...e,id:void 0,children:n.rendered,defaultClassName:"react-aria-Tab",values:{isSelected:u,isDisabled:l,isFocused:c,isFocusVisible:d,isPressed:o,isHovered:m}}),v=n.props.href?"a":"div",S=Gt(e,{global:!0});return delete S.id,delete S.onClick,be.createElement(v,{...Et(S,y,a,s,h),ref:i,"data-selected":u||void 0,"data-disabled":l||void 0,"data-focused":c||void 0,"data-focus-visible":d||void 0,"data-pressed":o||void 0,"data-hovered":m||void 0},be.createElement(K3.Provider,{value:{isSelected:u}},y.children))}),Wit=C1(function(t,n){const r=P.useContext(T3);let i=wu(n),{id:a,...u}=t,{tabPanelProps:l}=UCe(t,r,i),{focusProps:o,isFocused:s,isFocusVisible:c}=Gu(),d=r.selectedKey===t.id,h=Cn({...t,defaultClassName:"react-aria-TabPanel",values:{isFocused:s,isFocusVisible:c,isInert:_9(!d),state:r}});if(!d&&!t.shouldForceMount)return null;let m=Gt(u,{global:!0});delete m.id;let y=d?Et(m,l,o,h):h;return be.createElement("div",{...y,ref:i,"data-focused":s||void 0,"data-focus-visible":c||void 0,inert:_9(!d||t.inert),"data-inert":d?void 0:"true"},be.createElement(Ni,{values:[[uw,null],[T3,null]]},be.createElement(mp.Provider,{value:jN},h.children)))}),I5e=P.createContext(null),Vhe=P.createContext(null),Jit=P.forwardRef(function(t,n){return[t,n]=Nn(t,n,I5e),be.createElement(uh.Provider,{value:null},be.createElement(w1,{content:t.children},r=>be.createElement(L5e,{props:t,forwardedRef:n,collection:r})))});function L5e({props:e,forwardedRef:t,collection:n}){let r=P.useRef(null),{id:i,...a}=e;[a,r]=Nn(a,r,vx);let{filter:u,shouldUseVirtualFocus:l,...o}=a,[s,c]=Ss(!e["aria-label"]&&!e["aria-labelledby"]),d=tw({...o,children:void 0,collection:n}),h=tde(d,u),m=Gt(a,{global:!0}),y=Object.fromEntries(Object.entries(m).map(([k,A])=>[k,k==="id"?A:void 0])),{gridProps:v,labelProps:S,descriptionProps:_,errorMessageProps:E}=HCe({...o,...y,label:c},h,r);var $;return be.createElement("div",{...m,id:i,ref:t,slot:e.slot||void 0,className:($=e.className)!==null&&$!==void 0?$:"react-aria-TagGroup",style:e.style},be.createElement(Ni,{values:[[_s,{...S,elementType:"span",ref:s}],[Vhe,{...v,ref:r}],[uh,h],[zo,{slots:{description:_,errorMessage:E}}]]},e.children))}const eat=P.forwardRef(function(t,n){return P.useContext(uh)?be.createElement(B5e,{props:t,forwardedRef:n}):be.createElement(OS,t)});function B5e({props:e,forwardedRef:t}){let n=P.useContext(uh),{CollectionRoot:r}=P.useContext(mp),[i,a]=Nn({},t,Vhe),{focusProps:u,isFocused:l,isFocusVisible:o}=Gu(),s={isEmpty:n.collection.size===0,isFocused:l,isFocusVisible:o,state:n},c=Cn({className:e.className,style:e.style,defaultClassName:"react-aria-TagList",values:s}),d=ZN(n.selectionManager.focusedKey),h=Gt(e,{global:!0});return be.createElement("div",{...Et(h,c,i,u),ref:a,"data-empty":n.collection.size===0||void 0,"data-focused":l||void 0,"data-focus-visible":o||void 0},be.createElement(G3,null,n.collection.size===0&&e.renderEmptyState?e.renderEmptyState(s):be.createElement(r,{collection:n.collection,persistedKeys:d})))}const tat=R1(JC,(e,t,n)=>{let r=P.useContext(uh),i=wu(t),{focusProps:a,isFocusVisible:u}=Gu({within:!1}),{rowProps:l,gridCellProps:o,removeButtonProps:s,...c}=YCe({item:n},r,i),{hoverProps:d,isHovered:h}=Ru({isDisabled:!c.allowsSelection,onHoverStart:n.props.onHoverStart,onHoverChange:n.props.onHoverChange,onHoverEnd:n.props.onHoverEnd}),m=Cn({...e,id:void 0,children:n.rendered,defaultClassName:"react-aria-Tag",values:{...c,isFocusVisible:u,isHovered:h,selectionMode:r.selectionManager.selectionMode,selectionBehavior:r.selectionManager.selectionBehavior}});P.useEffect(()=>{n.textValue},[n.textValue]);let y=Gt(e,{global:!0});return delete y.id,delete y.onClick,be.createElement("div",{ref:i,...Et(y,m,l,a,d),"data-selected":c.isSelected||void 0,"data-disabled":c.isDisabled||void 0,"data-hovered":h||void 0,"data-focused":c.isFocused||void 0,"data-focus-visible":u||void 0,"data-pressed":c.isPressed||void 0,"data-allows-removing":c.allowsRemoving||void 0,"data-selection-mode":r.selectionManager.selectionMode==="none"?void 0:r.selectionManager.selectionMode},be.createElement("div",{...o,style:{display:"contents"}},be.createElement(Ni,{values:[[ah,{slots:{remove:s}}],[mp,jN],[K3,{isSelected:c.isSelected}]]},m.children)))}),Uhe=P.createContext({});let Q5e=e=>{let{onHoverStart:t,onHoverChange:n,onHoverEnd:r,...i}=e;return i};const nat=P.forwardRef(function(t,n){[t,n]=Nn(t,n,Uhe);let{hoverProps:r,isHovered:i}=Ru(t),{isFocused:a,isFocusVisible:u,focusProps:l}=Gu({isTextInput:!0,autoFocus:t.autoFocus}),o=!!t["aria-invalid"]&&t["aria-invalid"]!=="false",s=Cn({...t,values:{isHovered:i,isFocused:a,isFocusVisible:u,isDisabled:t.disabled||!1,isInvalid:o},defaultClassName:"react-aria-TextArea"});return be.createElement("textarea",{...Et(Q5e(t),l,r),...s,ref:n,"data-focused":a||void 0,"data-disabled":t.disabled||void 0,"data-hovered":i||void 0,"data-focus-visible":u||void 0,"data-invalid":o||void 0})}),z5e=P.createContext(null),rat=C1(function(t,n){[t,n]=Nn(t,n,z5e);let{validationBehavior:r}=xs(op)||{};var i,a;let u=(a=(i=t.validationBehavior)!==null&&i!==void 0?i:r)!==null&&a!==void 0?a:"native",l=P.useRef(null);[t,l]=Nn(t,l,LC);let[o,s]=Ss(!t["aria-label"]&&!t["aria-labelledby"]),[c,d]=P.useState("input"),{labelProps:h,inputProps:m,descriptionProps:y,errorMessageProps:v,...S}=WC({...pc(t),inputElementType:c,label:s,validationBehavior:u},l),_=P.useCallback(k=>{l.current=k,k&&d(k instanceof HTMLTextAreaElement?"textarea":"input")},[l]),E=Cn({...t,values:{isDisabled:t.isDisabled||!1,isInvalid:S.isInvalid,isReadOnly:t.isReadOnly||!1,isRequired:t.isRequired||!1},defaultClassName:"react-aria-TextField"}),$=Gt(t,{global:!0});return delete $.id,be.createElement("div",{...$,...E,ref:n,slot:t.slot||void 0,"data-disabled":t.isDisabled||void 0,"data-invalid":S.isInvalid||void 0,"data-readonly":t.isReadOnly||void 0,"data-required":t.isRequired||void 0},be.createElement(Ni,{values:[[_s,{...h,ref:o}],[Q0,{...m,ref:_}],[Uhe,{...m,ref:_}],[B0,{role:"presentation",isInvalid:S.isInvalid,isDisabled:t.isDisabled||!1}],[zo,{slots:{description:y,errorMessage:v}}],[gp,S]]},E.children))}),dI=P.createContext(null),iat=P.forwardRef(function(t,n){let r=pp(),i=Twe(t.queue),a=wu(n),{regionProps:u}=JCe(t,i,a),{focusProps:l,isFocused:o,isFocusVisible:s}=Gu(),{hoverProps:c,isHovered:d}=Ru({}),h=Cn({...t,children:void 0,defaultClassName:"react-aria-ToastRegion",values:{visibleToasts:i.visibleToasts,isHovered:d,isFocused:o,isFocusVisible:s}}),{direction:m}=Wi(),y,{getContainer:v}=IN();r||(y=document.body,v&&(y=v()));let S=Gt(t,{global:!0}),_=be.createElement(dI.Provider,{value:i},be.createElement("div",{...Et(S,h,u,l,c),dir:m,ref:a,"data-hovered":d||void 0,"data-focused":o||void 0,"data-focus-visible":s||void 0},typeof t.children=="function"?be.createElement(q5e,{...t,className:void 0,style:{display:"contents"}},t.children):t.children));return i.visibleToasts.length>0&&y?Uf.createPortal(_,y):null}),q5e=P.forwardRef(function(t,n){let r=P.useContext(dI),{hoverProps:i,isHovered:a}=Ru({}),u=Cn({...t,children:void 0,defaultClassName:"react-aria-ToastList",values:{visibleToasts:r.visibleToasts,isFocused:!1,isFocusVisible:!1,isHovered:a}});return be.createElement("ol",{...i,...u,ref:n},r.visibleToasts.map(l=>be.createElement("li",{key:l.key,style:{display:"contents"}},t.children({toast:l}))))}),aat=P.forwardRef(function(t,n){let r=P.useContext(dI),i=wu(n),{toastProps:a,contentProps:u,titleProps:l,descriptionProps:o,closeButtonProps:s}=KCe(t,r,i),{focusProps:c,isFocused:d,isFocusVisible:h}=Gu(),m=Cn({...t,defaultClassName:"react-aria-Toast",values:{toast:t.toast,isFocused:d,isFocusVisible:h}}),y=Gt(t,{global:!0});return be.createElement("div",{...Et(y,m,a,c),ref:i,"data-focused":d||void 0,"data-focus-visible":h||void 0},be.createElement(Ni,{values:[[jhe,u],[zo,{slots:{[Kv]:{},title:l,description:o}}],[ah,{slots:{[Kv]:{},close:s}}]]},m.children))}),jhe=P.createContext({}),uat=P.forwardRef(function(t,n){return[t,n]=Nn(t,n,jhe),be.createElement("div",{className:"react-aria-ToastContent",...t,ref:n},t.children)}),V5e=P.createContext({}),Zhe=P.createContext(null),lat=P.forwardRef(function(t,n){[t,n]=Nn(t,n,V5e);let r=Q$e(t),{groupProps:i}=I$e(t,r,n),a=Cn({...t,values:{isDisabled:r.isDisabled,state:r},defaultClassName:"react-aria-ToggleButtonGroup"}),u=Gt(t,{global:!0});return be.createElement("div",{...Et(u,a,i),ref:n,slot:t.slot||void 0,"data-orientation":t.orientation||"horizontal","data-disabled":t.isDisabled||void 0},be.createElement(Zhe.Provider,{value:r},be.createElement(G3,null,a.children)))}),U5e=P.createContext({}),oat=P.forwardRef(function(t,n){[t,n]=Nn(t,n,U5e);let r=P.useContext(Zhe),i=ew(r&&t.id!=null?{isSelected:r.selectedKeys.has(t.id),onChange(S){r.setSelected(t.id,S)}}:t),{buttonProps:a,isPressed:u,isSelected:l,isDisabled:o}=r&&t.id!=null?L$e({...t,id:t.id},r,n):Pue({...t,id:t.id!=null?String(t.id):void 0},i,n),{focusProps:s,isFocused:c,isFocusVisible:d}=Gu(t),{hoverProps:h,isHovered:m}=Ru(t),y=Cn({...t,id:void 0,values:{isHovered:m,isPressed:u,isFocused:c,isSelected:i.isSelected,isFocusVisible:d,isDisabled:o,state:i},defaultClassName:"react-aria-ToggleButton"}),v=Gt(t,{global:!0});return delete v.id,delete v.onClick,be.createElement("button",{...Et(v,y,a,s,h),ref:n,slot:t.slot||void 0,"data-focused":c||void 0,"data-disabled":o||void 0,"data-pressed":u||void 0,"data-selected":l||void 0,"data-hovered":m||void 0,"data-focus-visible":d||void 0},be.createElement(K3.Provider,{value:{isSelected:l}},y.children))}),j5e=P.createContext({}),sat=P.forwardRef(function(t,n){[t,n]=Nn(t,n,j5e);let{toolbarProps:r}=Mue(t,n),i=Cn({...t,values:{orientation:t.orientation||"horizontal"},defaultClassName:"react-aria-Toolbar"}),a=Gt(t,{global:!0});return delete a.id,be.createElement("div",{...Et(a,i,r),ref:n,slot:t.slot||void 0,"data-orientation":t.orientation||"horizontal"},i.children)}),hI=P.createContext(null),Hhe=P.createContext(null);function cat(e){let t=qde(e),n=P.useRef(null),{triggerProps:r,tooltipProps:i}=twe(e,t,n);return be.createElement(Ni,{values:[[hI,t],[Hhe,{...i,triggerRef:n}]]},be.createElement(mDe,{...r,ref:n},e.children))}const fat=P.forwardRef(function({UNSTABLE_portalContainer:t,...n},r){[n,r]=Nn(n,r,Hhe);let i=P.useContext(hI),a=qde(n),u=n.isOpen!=null||n.defaultOpen!=null||!i?a:i,l=XD(r,u.isOpen)||n.isExiting||!1;return!u.isOpen&&!l?null:be.createElement(FDe,{portalContainer:t},be.createElement(Z5e,{...n,tooltipRef:r,isExiting:l}))});function Z5e(e){let t=P.useContext(hI),n=P.useRef(null),{overlayProps:r,arrowProps:i,placement:a,triggerAnchorPoint:u}=kne({placement:e.placement||"top",targetRef:e.triggerRef,overlayRef:e.tooltipRef,arrowRef:n,offset:e.offset,crossOffset:e.crossOffset,isOpen:t.isOpen,arrowBoundaryOffset:e.arrowBoundaryOffset,shouldFlip:e.shouldFlip,containerPadding:e.containerPadding,onClose:()=>t.close(!0)}),l=qC(e.tooltipRef,!!a)||e.isEntering||!1,o=Cn({...e,defaultClassName:"react-aria-Tooltip",values:{placement:a,isEntering:l,isExiting:e.isExiting,state:t}});e=Et(e,r);let{tooltipProps:s}=ewe(e,t),c=Gt(e,{global:!0});return be.createElement("div",{...Et(c,o,s),ref:e.tooltipRef,style:{...r.style,"--trigger-anchor-point":u?`${u.x}px ${u.y}px`:void 0,...o.style},"data-placement":a??void 0,"data-entering":l||void 0,"data-exiting":e.isExiting||void 0},be.createElement(oI.Provider,{value:{...i,placement:a,ref:n}},o.children))}/**
 * react-router v7.12.0
 *
 * Copyright (c) Remix Software Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE.md file in the root directory of this source tree.
 *
 * @license MIT
 */var Xhe=e=>{throw TypeError(e)},H5e=(e,t,n)=>t.has(e)||Xhe("Cannot "+n),yk=(e,t,n)=>(H5e(e,t,"read from private field"),n?n.call(e):t.get(e)),X5e=(e,t,n)=>t.has(e)?Xhe("Cannot add the same private member more than once"):t instanceof WeakSet?t.add(e):t.set(e,n),Jj="popstate";function Y5e(e={}){function t(r,i){let{pathname:a,search:u,hash:l}=r.location;return Dx("",{pathname:a,search:u,hash:l},i.state&&i.state.usr||null,i.state&&i.state.key||"default")}function n(r,i){return typeof i=="string"?i:k0(i)}return K5e(t,n,null,e)}function gr(e,t){if(e===!1||e===null||typeof e>"u")throw new Error(t)}function fu(e,t){if(!e){typeof console<"u"&&console.warn(t);try{throw new Error(t)}catch{}}}function G5e(){return Math.random().toString(36).substring(2,10)}function eZ(e,t){return{usr:e.state,key:e.key,idx:t}}function Dx(e,t,n=null,r){return{pathname:typeof e=="string"?e:e.pathname,search:"",hash:"",...typeof t=="string"?yy(t):t,state:n,key:t&&t.key||r||G5e()}}function k0({pathname:e="/",search:t="",hash:n=""}){return t&&t!=="?"&&(e+=t.charAt(0)==="?"?t:"?"+t),n&&n!=="#"&&(e+=n.charAt(0)==="#"?n:"#"+n),e}function yy(e){let t={};if(e){let n=e.indexOf("#");n>=0&&(t.hash=e.substring(n),e=e.substring(0,n));let r=e.indexOf("?");r>=0&&(t.search=e.substring(r),e=e.substring(0,r)),e&&(t.pathname=e)}return t}function K5e(e,t,n,r={}){let{window:i=document.defaultView,v5Compat:a=!1}=r,u=i.history,l="POP",o=null,s=c();s==null&&(s=0,u.replaceState({...u.state,idx:s},""));function c(){return(u.state||{idx:null}).idx}function d(){l="POP";let S=c(),_=S==null?null:S-s;s=S,o&&o({action:l,location:v.location,delta:_})}function h(S,_){l="PUSH";let E=Dx(v.location,S,_);s=c()+1;let $=eZ(E,s),k=v.createHref(E);try{u.pushState($,"",k)}catch(A){if(A instanceof DOMException&&A.name==="DataCloneError")throw A;i.location.assign(k)}a&&o&&o({action:l,location:v.location,delta:1})}function m(S,_){l="REPLACE";let E=Dx(v.location,S,_);s=c();let $=eZ(E,s),k=v.createHref(E);u.replaceState($,"",k),a&&o&&o({action:l,location:v.location,delta:0})}function y(S){return Yhe(S)}let v={get action(){return l},get location(){return e(i,u)},listen(S){if(o)throw new Error("A history only accepts one active listener");return i.addEventListener(Jj,d),o=S,()=>{i.removeEventListener(Jj,d),o=null}},createHref(S){return t(i,S)},createURL:y,encodeLocation(S){let _=y(S);return{pathname:_.pathname,search:_.search,hash:_.hash}},push:h,replace:m,go(S){return u.go(S)}};return v}function Yhe(e,t=!1){let n="http://localhost";typeof window<"u"&&(n=window.location.origin!=="null"?window.location.origin:window.location.href),gr(n,"No window.location.(origin|href) available to create URL");let r=typeof e=="string"?e:k0(e);return r=r.replace(/ $/,"%20"),!t&&r.startsWith("//")&&(r=n+r),new URL(r,n)}var I2,tZ=class{constructor(e){if(X5e(this,I2,new Map),e)for(let[t,n]of e)this.set(t,n)}get(e){if(yk(this,I2).has(e))return yk(this,I2).get(e);if(e.defaultValue!==void 0)return e.defaultValue;throw new Error("No value found for context")}set(e,t){yk(this,I2).set(e,t)}};I2=new WeakMap;var W5e=new Set(["lazy","caseSensitive","path","id","index","children"]);function J5e(e){return W5e.has(e)}var e6e=new Set(["lazy","caseSensitive","path","id","index","middleware","children"]);function t6e(e){return e6e.has(e)}function n6e(e){return e.index===!0}function $x(e,t,n=[],r={},i=!1){return e.map((a,u)=>{let l=[...n,String(u)],o=typeof a.id=="string"?a.id:l.join("-");if(gr(a.index!==!0||!a.children,"Cannot specify children on an index route"),gr(i||!r[o],`Found a route id collision on id "${o}".  Route id's must be globally unique within Data Router usages`),n6e(a)){let s={...a,id:o};return r[o]=nZ(s,t(s)),s}else{let s={...a,id:o,children:void 0};return r[o]=nZ(s,t(s)),a.children&&(s.children=$x(a.children,t,l,r,i)),s}})}function nZ(e,t){return Object.assign(e,{...t,...typeof t.lazy=="object"&&t.lazy!=null?{lazy:{...e.lazy,...t.lazy}}:{}})}function Rv(e,t,n="/"){return L2(e,t,n,!1)}function L2(e,t,n,r){let i=typeof t=="string"?yy(t):t,a=ff(i.pathname||"/",n);if(a==null)return null;let u=Khe(e);r6e(u);let l=null;for(let o=0;l==null&&o<u.length;++o){let s=p6e(a);l=d6e(u[o],s,r)}return l}function Ghe(e,t){let{route:n,pathname:r,params:i}=e;return{id:n.id,pathname:r,params:i,data:t[n.id],loaderData:t[n.id],handle:n.handle}}function Khe(e,t=[],n=[],r="",i=!1){let a=(u,l,o=i,s)=>{let c={relativePath:s===void 0?u.path||"":s,caseSensitive:u.caseSensitive===!0,childrenIndex:l,route:u};if(c.relativePath.startsWith("/")){if(!c.relativePath.startsWith(r)&&o)return;gr(c.relativePath.startsWith(r),`Absolute route path "${c.relativePath}" nested under path "${r}" is not valid. An absolute child route path must start with the combined path of all its parent routes.`),c.relativePath=c.relativePath.slice(r.length)}let d=C0([r,c.relativePath]),h=n.concat(c);u.children&&u.children.length>0&&(gr(u.index!==!0,`Index routes must not have child routes. Please remove all child routes from route path "${d}".`),Khe(u.children,t,h,d,o)),!(u.path==null&&!u.index)&&t.push({path:d,score:c6e(d,u.index),routesMeta:h})};return e.forEach((u,l)=>{var o;if(u.path===""||!((o=u.path)!=null&&o.includes("?")))a(u,l);else for(let s of Whe(u.path))a(u,l,!0,s)}),t}function Whe(e){let t=e.split("/");if(t.length===0)return[];let[n,...r]=t,i=n.endsWith("?"),a=n.replace(/\?$/,"");if(r.length===0)return i?[a,""]:[a];let u=Whe(r.join("/")),l=[];return l.push(...u.map(o=>o===""?a:[a,o].join("/"))),i&&l.push(...u),l.map(o=>e.startsWith("/")&&o===""?"/":o)}function r6e(e){e.sort((t,n)=>t.score!==n.score?n.score-t.score:f6e(t.routesMeta.map(r=>r.childrenIndex),n.routesMeta.map(r=>r.childrenIndex)))}var i6e=/^:[\w-]+$/,a6e=3,u6e=2,l6e=1,o6e=10,s6e=-2,rZ=e=>e==="*";function c6e(e,t){let n=e.split("/"),r=n.length;return n.some(rZ)&&(r+=s6e),t&&(r+=u6e),n.filter(i=>!rZ(i)).reduce((i,a)=>i+(i6e.test(a)?a6e:a===""?l6e:o6e),r)}function f6e(e,t){return e.length===t.length&&e.slice(0,-1).every((r,i)=>r===t[i])?e[e.length-1]-t[t.length-1]:0}function d6e(e,t,n=!1){let{routesMeta:r}=e,i={},a="/",u=[];for(let l=0;l<r.length;++l){let o=r[l],s=l===r.length-1,c=a==="/"?t:t.slice(a.length)||"/",d=d$({path:o.relativePath,caseSensitive:o.caseSensitive,end:s},c),h=o.route;if(!d&&s&&n&&!r[r.length-1].route.index&&(d=d$({path:o.relativePath,caseSensitive:o.caseSensitive,end:!1},c)),!d)return null;Object.assign(i,d.params),u.push({params:i,pathname:C0([a,d.pathname]),pathnameBase:v6e(C0([a,d.pathnameBase])),route:h}),d.pathnameBase!=="/"&&(a=C0([a,d.pathnameBase]))}return u}function d$(e,t){typeof e=="string"&&(e={path:e,caseSensitive:!1,end:!0});let[n,r]=h6e(e.path,e.caseSensitive,e.end),i=t.match(n);if(!i)return null;let a=i[0],u=a.replace(/(.)\/+$/,"$1"),l=i.slice(1);return{params:r.reduce((s,{paramName:c,isOptional:d},h)=>{if(c==="*"){let y=l[h]||"";u=a.slice(0,a.length-y.length).replace(/(.)\/+$/,"$1")}const m=l[h];return d&&!m?s[c]=void 0:s[c]=(m||"").replace(/%2F/g,"/"),s},{}),pathname:a,pathnameBase:u,pattern:e}}function h6e(e,t=!1,n=!0){fu(e==="*"||!e.endsWith("*")||e.endsWith("/*"),`Route path "${e}" will be treated as if it were "${e.replace(/\*$/,"/*")}" because the \`*\` character must always follow a \`/\` in the pattern. To get rid of this warning, please change the route path to "${e.replace(/\*$/,"/*")}".`);let r=[],i="^"+e.replace(/\/*\*?$/,"").replace(/^\/*/,"/").replace(/[\\.*+^${}|()[\]]/g,"\\$&").replace(/\/:([\w-]+)(\?)?/g,(u,l,o)=>(r.push({paramName:l,isOptional:o!=null}),o?"/?([^\\/]+)?":"/([^\\/]+)")).replace(/\/([\w-]+)\?(\/|$)/g,"(/$1)?$2");return e.endsWith("*")?(r.push({paramName:"*"}),i+=e==="*"||e==="/*"?"(.*)$":"(?:\\/(.+)|\\/*)$"):n?i+="\\/*$":e!==""&&e!=="/"&&(i+="(?:(?=\\/|$))"),[new RegExp(i,t?void 0:"i"),r]}function p6e(e){try{return e.split("/").map(t=>decodeURIComponent(t).replace(/\//g,"%2F")).join("/")}catch(t){return fu(!1,`The URL path "${e}" could not be decoded because it is a malformed URL segment. This is probably due to a bad percent encoding (${t}).`),e}}function ff(e,t){if(t==="/")return e;if(!e.toLowerCase().startsWith(t.toLowerCase()))return null;let n=t.endsWith("/")?t.length-1:t.length,r=e.charAt(n);return r&&r!=="/"?null:e.slice(n)||"/"}function m6e({basename:e,pathname:t}){return t==="/"?e:C0([e,t])}var Jhe=/^(?:[a-z][a-z0-9+.-]*:|\/\/)/i,lw=e=>Jhe.test(e);function g6e(e,t="/"){let{pathname:n,search:r="",hash:i=""}=typeof e=="string"?yy(e):e,a;if(n)if(lw(n))a=n;else{if(n.includes("//")){let u=n;n=n.replace(/\/\/+/g,"/"),fu(!1,`Pathnames cannot have embedded double slashes - normalizing ${u} -> ${n}`)}n.startsWith("/")?a=iZ(n.substring(1),"/"):a=iZ(n,t)}else a=t;return{pathname:a,search:y6e(r),hash:b6e(i)}}function iZ(e,t){let n=t.replace(/\/+$/,"").split("/");return e.split("/").forEach(i=>{i===".."?n.length>1&&n.pop():i!=="."&&n.push(i)}),n.length>1?n.join("/"):"/"}function bk(e,t,n,r){return`Cannot include a '${e}' character in a manually specified \`to.${t}\` field [${JSON.stringify(r)}].  Please separate it out to the \`to.${n}\` field. Alternatively you may provide the full path as a string in <Link to="..."> and the router will parse it for you.`}function epe(e){return e.filter((t,n)=>n===0||t.route.path&&t.route.path.length>0)}function ow(e){let t=epe(e);return t.map((n,r)=>r===t.length-1?n.pathname:n.pathnameBase)}function sw(e,t,n,r=!1){let i;typeof e=="string"?i=yy(e):(i={...e},gr(!i.pathname||!i.pathname.includes("?"),bk("?","pathname","search",i)),gr(!i.pathname||!i.pathname.includes("#"),bk("#","pathname","hash",i)),gr(!i.search||!i.search.includes("#"),bk("#","search","hash",i)));let a=e===""||i.pathname==="",u=a?"/":i.pathname,l;if(u==null)l=n;else{let d=t.length-1;if(!r&&u.startsWith("..")){let h=u.split("/");for(;h[0]==="..";)h.shift(),d-=1;i.pathname=h.join("/")}l=d>=0?t[d]:"/"}let o=g6e(i,l),s=u&&u!=="/"&&u.endsWith("/"),c=(a||u===".")&&n.endsWith("/");return!o.pathname.endsWith("/")&&(s||c)&&(o.pathname+="/"),o}var C0=e=>e.join("/").replace(/\/\/+/g,"/"),v6e=e=>e.replace(/\/+$/,"").replace(/^\/*/,"/"),y6e=e=>!e||e==="?"?"":e.startsWith("?")?e:"?"+e,b6e=e=>!e||e==="#"?"":e.startsWith("#")?e:"#"+e,dat=(e,t=302)=>{let n=t;typeof n=="number"?n={status:n}:typeof n.status>"u"&&(n.status=302);let r=new Headers(n.headers);return r.set("Location",e),new Response(null,{...n,headers:r})},ES=class{constructor(e,t,n,r=!1){this.status=e,this.statusText=t||"",this.internal=r,n instanceof Error?(this.data=n.toString(),this.error=n):this.data=n}};function Cx(e){return e!=null&&typeof e.status=="number"&&typeof e.statusText=="string"&&typeof e.internal=="boolean"&&"data"in e}function DS(e){return e.map(t=>t.route.path).filter(Boolean).join("/").replace(/\/\/*/g,"/")||"/"}var tpe=typeof window<"u"&&typeof window.document<"u"&&typeof window.document.createElement<"u";function npe(e,t){let n=e;if(typeof n!="string"||!Jhe.test(n))return{absoluteURL:void 0,isExternal:!1,to:n};let r=n,i=!1;if(tpe)try{let a=new URL(window.location.href),u=n.startsWith("//")?new URL(a.protocol+n):new URL(n),l=ff(u.pathname,t);u.origin===a.origin&&l!=null?n=l+u.search+u.hash:i=!0}catch{fu(!1,`<Link to="${n}"> contains an invalid URL which will probably break when clicked - please update to a valid URL path.`)}return{absoluteURL:r,isExternal:i,to:n}}var Lv=Symbol("Uninstrumented");function O6e(e,t){let n={lazy:[],"lazy.loader":[],"lazy.action":[],"lazy.middleware":[],middleware:[],loader:[],action:[]};e.forEach(i=>i({id:t.id,index:t.index,path:t.path,instrument(a){let u=Object.keys(n);for(let l of u)a[l]&&n[l].push(a[l])}}));let r={};if(typeof t.lazy=="function"&&n.lazy.length>0){let i=ZO(n.lazy,t.lazy,()=>{});i&&(r.lazy=i)}if(typeof t.lazy=="object"){let i=t.lazy;["middleware","loader","action"].forEach(a=>{let u=i[a],l=n[`lazy.${a}`];if(typeof u=="function"&&l.length>0){let o=ZO(l,u,()=>{});o&&(r.lazy=Object.assign(r.lazy||{},{[a]:o}))}})}return["loader","action"].forEach(i=>{let a=t[i];if(typeof a=="function"&&n[i].length>0){let u=a[Lv]??a,l=ZO(n[i],u,(...o)=>aZ(o[0]));l&&(i==="loader"&&u.hydrate===!0&&(l.hydrate=!0),l[Lv]=u,r[i]=l)}}),t.middleware&&t.middleware.length>0&&n.middleware.length>0&&(r.middleware=t.middleware.map(i=>{let a=i[Lv]??i,u=ZO(n.middleware,a,(...l)=>aZ(l[0]));return u?(u[Lv]=a,u):i})),r}function x6e(e,t){let n={navigate:[],fetch:[]};if(t.forEach(r=>r({instrument(i){let a=Object.keys(i);for(let u of a)i[u]&&n[u].push(i[u])}})),n.navigate.length>0){let r=e.navigate[Lv]??e.navigate,i=ZO(n.navigate,r,(...a)=>{let[u,l]=a;return{to:typeof u=="number"||typeof u=="string"?u:u?k0(u):".",...uZ(e,l??{})}});i&&(i[Lv]=r,e.navigate=i)}if(n.fetch.length>0){let r=e.fetch[Lv]??e.fetch,i=ZO(n.fetch,r,(...a)=>{let[u,,l,o]=a;return{href:l??".",fetcherKey:u,...uZ(e,o??{})}});i&&(i[Lv]=r,e.fetch=i)}return e}function ZO(e,t,n){return e.length===0?null:async(...r)=>{let i=await rpe(e,n(...r),()=>t(...r),e.length-1);if(i.type==="error")throw i.value;return i.value}}async function rpe(e,t,n,r){let i=e[r],a;if(i){let u,l=async()=>(u?console.error("You cannot call instrumented handlers more than once"):u=rpe(e,t,n,r-1),a=await u,gr(a,"Expected a result"),a.type==="error"&&a.value instanceof Error?{status:"error",error:a.value}:{status:"success",error:void 0});try{await i(l,t)}catch(o){console.error("An instrumentation function threw an error:",o)}u||await l(),await u}else try{a={type:"success",value:await n()}}catch(u){a={type:"error",value:u}}return a||{type:"error",value:new Error("No result assigned in instrumentation chain.")}}function aZ(e){let{request:t,context:n,params:r,unstable_pattern:i}=e;return{request:S6e(t),params:{...r},unstable_pattern:i,context:_6e(n)}}function uZ(e,t){return{currentUrl:k0(e.state.location),..."formMethod"in t?{formMethod:t.formMethod}:{},..."formEncType"in t?{formEncType:t.formEncType}:{},..."formData"in t?{formData:t.formData}:{},..."body"in t?{body:t.body}:{}}}function S6e(e){return{method:e.method,url:e.url,headers:{get:(...t)=>e.headers.get(...t)}}}function _6e(e){if(D6e(e)){let t={...e};return Object.freeze(t),t}else return{get:t=>e.get(t)}}var E6e=Object.getOwnPropertyNames(Object.prototype).sort().join("\0");function D6e(e){if(e===null||typeof e!="object")return!1;const t=Object.getPrototypeOf(e);return t===Object.prototype||t===null||Object.getOwnPropertyNames(t).sort().join("\0")===E6e}var ipe=["POST","PUT","PATCH","DELETE"],$6e=new Set(ipe),C6e=["GET",...ipe],w6e=new Set(C6e),ape=new Set([301,302,303,307,308]),R6e=new Set([307,308]),Ok={state:"idle",location:void 0,formMethod:void 0,formAction:void 0,formEncType:void 0,formData:void 0,json:void 0,text:void 0},A6e={state:"idle",data:void 0,formMethod:void 0,formAction:void 0,formEncType:void 0,formData:void 0,json:void 0,text:void 0},LO={state:"unblocked",proceed:void 0,reset:void 0,location:void 0},T6e=e=>({hasErrorBoundary:!!e.hasErrorBoundary}),upe="remix-router-transitions",lpe=Symbol("ResetLoaderData");function k6e(e){const t=e.window?e.window:typeof window<"u"?window:void 0,n=typeof t<"u"&&typeof t.document<"u"&&typeof t.document.createElement<"u";gr(e.routes.length>0,"You must provide a non-empty routes array to createRouter");let r=e.hydrationRouteProperties||[],i=e.mapRouteProperties||T6e,a=i;if(e.unstable_instrumentations){let Te=e.unstable_instrumentations;a=Ve=>({...i(Ve),...O6e(Te.map(nt=>nt.route).filter(Boolean),Ve)})}let u={},l=$x(e.routes,a,void 0,u),o,s=e.basename||"/";s.startsWith("/")||(s=`/${s}`);let c=e.dataStrategy||I6e,d={...e.future},h=null,m=new Set,y=null,v=null,S=null,_=e.hydrationData!=null,E=Rv(l,e.history.location,s),$=!1,k=null,A;if(E==null&&!e.patchRoutesOnNavigation){let Te=qd(404,{pathname:e.history.location.pathname}),{matches:Ve,route:nt}=J_(l);A=!0,E=Ve,k={[nt.id]:Te}}else if(E&&!e.hydrationData&&wr(E,l,e.history.location.pathname).active&&(E=null),E)if(E.some(Te=>Te.route.lazy))A=!1;else if(!E.some(Te=>pI(Te.route)))A=!0;else{let Te=e.hydrationData?e.hydrationData.loaderData:null,Ve=e.hydrationData?e.hydrationData.errors:null;if(Ve){let nt=E.findIndex(gt=>Ve[gt.route.id]!==void 0);A=E.slice(0,nt+1).every(gt=>!tM(gt.route,Te,Ve))}else A=E.every(nt=>!tM(nt.route,Te,Ve))}else{A=!1,E=[];let Te=wr(null,l,e.history.location.pathname);Te.active&&Te.matches&&($=!0,E=Te.matches)}let w,R={historyAction:e.history.action,location:e.history.location,matches:E,initialized:A,navigation:Ok,restoreScrollPosition:e.hydrationData!=null?!1:null,preventScrollReset:!1,revalidation:"idle",loaderData:e.hydrationData&&e.hydrationData.loaderData||{},actionData:e.hydrationData&&e.hydrationData.actionData||null,errors:e.hydrationData&&e.hydrationData.errors||k,fetchers:new Map,blockers:new Map},F="POP",I=null,U=!1,q,V=!1,N=new Map,j=null,W=!1,K=!1,J=new Set,te=new Map,H=0,M=-1,G=new Map,X=new Set,B=new Map,ne=new Map,oe=new Set,ie=new Map,de,ge=null;function ye(){if(h=e.history.listen(({action:Te,location:Ve,delta:nt})=>{if(de){de(),de=void 0;return}fu(ie.size===0||nt!=null,"You are trying to use a blocker on a POP navigation to a location that was not created by @remix-run/router. This will fail silently in production. This can happen if you are navigating outside the router via `window.history.pushState`/`window.location.hash` instead of using router navigation APIs.  This can also happen if you are using createHashRouter and the user manually changes the URL.");let gt=In({currentLocation:R.location,nextLocation:Ve,historyAction:Te});if(gt&&nt!=null){let Ct=new Promise(Vt=>{de=Vt});e.history.go(nt*-1),en(gt,{state:"blocked",location:Ve,proceed(){en(gt,{state:"proceeding",proceed:void 0,reset:void 0,location:Ve}),Ct.then(()=>e.history.go(nt))},reset(){let Vt=new Map(R.blockers);Vt.set(gt,LO),se({blockers:Vt})}}),I==null||I.resolve(),I=null;return}return Me(Te,Ve)}),n){tRe(t,N);let Te=()=>nRe(t,N);t.addEventListener("pagehide",Te),j=()=>t.removeEventListener("pagehide",Te)}return R.initialized||Me("POP",R.location,{initialHydration:!0}),w}function me(){h&&h(),j&&j(),m.clear(),q&&q.abort(),R.fetchers.forEach((Te,Ve)=>zt(Ve)),R.blockers.forEach((Te,Ve)=>Qt(Ve))}function Ee(Te){return m.add(Te),()=>m.delete(Te)}function se(Te,Ve={}){Te.matches&&(Te.matches=Te.matches.map(Ct=>{let Vt=u[Ct.route.id],Yt=Ct.route;return Yt.element!==Vt.element||Yt.errorElement!==Vt.errorElement||Yt.hydrateFallbackElement!==Vt.hydrateFallbackElement?{...Ct,route:Vt}:Ct})),R={...R,...Te};let nt=[],gt=[];R.fetchers.forEach((Ct,Vt)=>{Ct.state==="idle"&&(oe.has(Vt)?nt.push(Vt):gt.push(Vt))}),oe.forEach(Ct=>{!R.fetchers.has(Ct)&&!te.has(Ct)&&nt.push(Ct)}),[...m].forEach(Ct=>Ct(R,{deletedFetchers:nt,newErrors:Te.errors??null,viewTransitionOpts:Ve.viewTransitionOpts,flushSync:Ve.flushSync===!0})),nt.forEach(Ct=>zt(Ct)),gt.forEach(Ct=>R.fetchers.delete(Ct))}function pe(Te,Ve,{flushSync:nt}={}){var _n,sn;let gt=R.actionData!=null&&R.navigation.formMethod!=null&&vs(R.navigation.formMethod)&&R.navigation.state==="loading"&&((_n=Te.state)==null?void 0:_n._isRedirect)!==!0,Ct;Ve.actionData?Object.keys(Ve.actionData).length>0?Ct=Ve.actionData:Ct=null:gt?Ct=R.actionData:Ct=null;let Vt=Ve.loaderData?gZ(R.loaderData,Ve.loaderData,Ve.matches||[],Ve.errors):R.loaderData,Yt=R.blockers;Yt.size>0&&(Yt=new Map(Yt),Yt.forEach((Ln,er)=>Yt.set(er,LO)));let Dt=W?!1:Cr(Te,Ve.matches||R.matches),Lt=U===!0||R.navigation.formMethod!=null&&vs(R.navigation.formMethod)&&((sn=Te.state)==null?void 0:sn._isRedirect)!==!0;o&&(l=o,o=void 0),W||F==="POP"||(F==="PUSH"?e.history.push(Te,Te.state):F==="REPLACE"&&e.history.replace(Te,Te.state));let an;if(F==="POP"){let Ln=N.get(R.location.pathname);Ln&&Ln.has(Te.pathname)?an={currentLocation:R.location,nextLocation:Te}:N.has(Te.pathname)&&(an={currentLocation:Te,nextLocation:R.location})}else if(V){let Ln=N.get(R.location.pathname);Ln?Ln.add(Te.pathname):(Ln=new Set([Te.pathname]),N.set(R.location.pathname,Ln)),an={currentLocation:R.location,nextLocation:Te}}se({...Ve,actionData:Ct,loaderData:Vt,historyAction:F,location:Te,initialized:!0,navigation:Ok,revalidation:"idle",restoreScrollPosition:Dt,preventScrollReset:Lt,blockers:Yt},{viewTransitionOpts:an,flushSync:nt===!0}),F="POP",U=!1,V=!1,W=!1,K=!1,I==null||I.resolve(),I=null,ge==null||ge.resolve(),ge=null}async function ve(Te,Ve){if(I==null||I.resolve(),I=null,typeof Te=="number"){I||(I=OZ());let er=I.promise;return e.history.go(Te),er}let nt=eM(R.location,R.matches,s,Te,Ve==null?void 0:Ve.fromRouteId,Ve==null?void 0:Ve.relative),{path:gt,submission:Ct,error:Vt}=lZ(!1,nt,Ve),Yt=R.location,Dt=Dx(R.location,gt,Ve&&Ve.state);Dt={...Dt,...e.history.encodeLocation(Dt)};let Lt=Ve&&Ve.replace!=null?Ve.replace:void 0,an="PUSH";Lt===!0?an="REPLACE":Lt===!1||Ct!=null&&vs(Ct.formMethod)&&Ct.formAction===R.location.pathname+R.location.search&&(an="REPLACE");let _n=Ve&&"preventScrollReset"in Ve?Ve.preventScrollReset===!0:void 0,sn=(Ve&&Ve.flushSync)===!0,Ln=In({currentLocation:Yt,nextLocation:Dt,historyAction:an});if(Ln){en(Ln,{state:"blocked",location:Dt,proceed(){en(Ln,{state:"proceeding",proceed:void 0,reset:void 0,location:Dt}),ve(Te,Ve)},reset(){let er=new Map(R.blockers);er.set(Ln,LO),se({blockers:er})}});return}await Me(an,Dt,{submission:Ct,pendingError:Vt,preventScrollReset:_n,replace:Ve&&Ve.replace,enableViewTransition:Ve&&Ve.viewTransition,flushSync:sn,callSiteDefaultShouldRevalidate:Ve&&Ve.unstable_defaultShouldRevalidate})}function we(){ge||(ge=OZ()),ze(),se({revalidation:"loading"});let Te=ge.promise;return R.navigation.state==="submitting"?Te:R.navigation.state==="idle"?(Me(R.historyAction,R.location,{startUninterruptedRevalidation:!0}),Te):(Me(F||R.historyAction,R.navigation.location,{overrideNavigation:R.navigation,enableViewTransition:V===!0}),Te)}async function Me(Te,Ve,nt){q&&q.abort(),q=null,F=Te,W=(nt&&nt.startUninterruptedRevalidation)===!0,Or(R.location,R.matches),U=(nt&&nt.preventScrollReset)===!0,V=(nt&&nt.enableViewTransition)===!0;let gt=o||l,Ct=nt&&nt.overrideNavigation,Vt=nt!=null&&nt.initialHydration&&R.matches&&R.matches.length>0&&!$?R.matches:Rv(gt,Ve,s),Yt=(nt&&nt.flushSync)===!0;if(Vt&&R.initialized&&!K&&j6e(R.location,Ve)&&!(nt&&nt.submission&&vs(nt.submission.formMethod))){pe(Ve,{matches:Vt},{flushSync:Yt});return}let Dt=wr(Vt,gt,Ve.pathname);if(Dt.active&&Dt.matches&&(Vt=Dt.matches),!Vt){let{error:$i,notFoundMatches:hr,route:wt}=rn(Ve.pathname);pe(Ve,{matches:hr,loaderData:{},errors:{[wt.id]:$i}},{flushSync:Yt});return}q=new AbortController;let Lt=BO(e.history,Ve,q.signal,nt&&nt.submission),an=e.getContext?await e.getContext():new tZ,_n;if(nt&&nt.pendingError)_n=[Av(Vt).route.id,{type:"error",error:nt.pendingError}];else if(nt&&nt.submission&&vs(nt.submission.formMethod)){let $i=await Re(Lt,Ve,nt.submission,Vt,an,Dt.active,nt&&nt.initialHydration===!0,{replace:nt.replace,flushSync:Yt});if($i.shortCircuited)return;if($i.pendingActionResult){let[hr,wt]=$i.pendingActionResult;if(Qf(wt)&&Cx(wt.error)&&wt.error.status===404){q=null,pe(Ve,{matches:$i.matches,loaderData:{},errors:{[hr]:wt.error}});return}}Vt=$i.matches||Vt,_n=$i.pendingActionResult,Ct=xk(Ve,nt.submission),Yt=!1,Dt.active=!1,Lt=BO(e.history,Lt.url,Lt.signal)}let{shortCircuited:sn,matches:Ln,loaderData:er,errors:Di}=await et(Lt,Ve,Vt,an,Dt.active,Ct,nt&&nt.submission,nt&&nt.fetcherSubmission,nt&&nt.replace,nt&&nt.initialHydration===!0,Yt,_n,nt&&nt.callSiteDefaultShouldRevalidate);sn||(q=null,pe(Ve,{matches:Ln||Vt,...vZ(_n),loaderData:er,errors:Di}))}async function Re(Te,Ve,nt,gt,Ct,Vt,Yt,Dt={}){ze();let Lt=J6e(Ve,nt);if(se({navigation:Lt},{flushSync:Dt.flushSync===!0}),Vt){let sn=await Gr(gt,Ve.pathname,Te.signal);if(sn.type==="aborted")return{shortCircuited:!0};if(sn.type==="error"){if(sn.partialMatches.length===0){let{matches:er,route:Di}=J_(l);return{matches:er,pendingActionResult:[Di.id,{type:"error",error:sn.error}]}}let Ln=Av(sn.partialMatches).route.id;return{matches:sn.partialMatches,pendingActionResult:[Ln,{type:"error",error:sn.error}]}}else if(sn.matches)gt=sn.matches;else{let{notFoundMatches:Ln,error:er,route:Di}=rn(Ve.pathname);return{matches:Ln,pendingActionResult:[Di.id,{type:"error",error:er}]}}}let an,_n=lD(gt,Ve);if(!_n.route.action&&!_n.route.lazy)an={type:"error",error:qd(405,{method:Te.method,pathname:Ve.pathname,routeId:_n.route.id})};else{let sn=f3(a,u,Te,gt,_n,Yt?[]:r,Ct),Ln=await Ye(Te,sn,Ct,null);if(an=Ln[_n.route.id],!an){for(let er of gt)if(Ln[er.route.id]){an=Ln[er.route.id];break}}if(Te.signal.aborted)return{shortCircuited:!0}}if(Lb(an)){let sn;return Dt&&Dt.replace!=null?sn=Dt.replace:sn=hZ(an.response.headers.get("Location"),new URL(Te.url),s,e.history)===R.location.pathname+R.location.search,await je(Te,an,!0,{submission:nt,replace:sn}),{shortCircuited:!0}}if(Qf(an)){let sn=Av(gt,_n.route.id);return(Dt&&Dt.replace)!==!0&&(F="PUSH"),{matches:gt,pendingActionResult:[sn.route.id,an,_n.route.id]}}return{matches:gt,pendingActionResult:[_n.route.id,an]}}async function et(Te,Ve,nt,gt,Ct,Vt,Yt,Dt,Lt,an,_n,sn,Ln){let er=Vt||xk(Ve,Yt),Di=Yt||Dt||bZ(er),$i=!W&&!an;if(Ct){if($i){let Rr=tt(sn);se({navigation:er,...Rr!==void 0?{actionData:Rr}:{}},{flushSync:_n})}let ur=await Gr(nt,Ve.pathname,Te.signal);if(ur.type==="aborted")return{shortCircuited:!0};if(ur.type==="error"){if(ur.partialMatches.length===0){let{matches:Sl,route:tr}=J_(l);return{matches:Sl,loaderData:{},errors:{[tr.id]:ur.error}}}let Rr=Av(ur.partialMatches).route.id;return{matches:ur.partialMatches,loaderData:{},errors:{[Rr]:ur.error}}}else if(ur.matches)nt=ur.matches;else{let{error:Rr,notFoundMatches:Sl,route:tr}=rn(Ve.pathname);return{matches:Sl,loaderData:{},errors:{[tr.id]:Rr}}}}let hr=o||l,{dsMatches:wt,revalidatingFetchers:Kr}=oZ(Te,gt,a,u,e.history,R,nt,Di,Ve,an?[]:r,an===!0,K,J,oe,B,X,hr,s,e.patchRoutesOnNavigation!=null,sn,Ln);if(M=++H,!e.dataStrategy&&!wt.some(ur=>ur.shouldLoad)&&!wt.some(ur=>ur.route.middleware&&ur.route.middleware.length>0)&&Kr.length===0){let ur=Fe();return pe(Ve,{matches:nt,loaderData:{},errors:sn&&Qf(sn[1])?{[sn[0]]:sn[1].error}:null,...vZ(sn),...ur?{fetchers:new Map(R.fetchers)}:{}},{flushSync:_n}),{shortCircuited:!0}}if($i){let ur={};if(!Ct){ur.navigation=er;let Rr=tt(sn);Rr!==void 0&&(ur.actionData=Rr)}Kr.length>0&&(ur.fetchers=Oe(Kr)),se(ur,{flushSync:_n})}Kr.forEach(ur=>{Xt(ur.key),ur.controller&&te.set(ur.key,ur.controller)});let Zi=()=>Kr.forEach(ur=>Xt(ur.key));q&&q.signal.addEventListener("abort",Zi);let{loaderResults:yi,fetcherResults:Ca}=await Ne(wt,Kr,Te,gt);if(Te.signal.aborted)return{shortCircuited:!0};q&&q.signal.removeEventListener("abort",Zi),Kr.forEach(ur=>te.delete(ur.key));let Xa=eE(yi);if(Xa)return await je(Te,Xa.result,!0,{replace:Lt}),{shortCircuited:!0};if(Xa=eE(Ca),Xa)return X.add(Xa.key),await je(Te,Xa.result,!0,{replace:Lt}),{shortCircuited:!0};let{loaderData:Au,errors:Tu}=mZ(R,nt,yi,sn,Kr,Ca);an&&R.errors&&(Tu={...R.errors,...Tu});let Nl=Fe(),gc=ht(M),xl=Nl||gc||Kr.length>0;return{matches:nt,loaderData:Au,errors:Tu,...xl?{fetchers:new Map(R.fetchers)}:{}}}function tt(Te){if(Te&&!Qf(Te[1]))return{[Te[0]]:Te[1].data};if(R.actionData)return Object.keys(R.actionData).length===0?null:R.actionData}function Oe(Te){return Te.forEach(Ve=>{let nt=R.fetchers.get(Ve.key),gt=t2(void 0,nt?nt.data:void 0);R.fetchers.set(Ve.key,gt)}),new Map(R.fetchers)}async function $e(Te,Ve,nt,gt){Xt(Te);let Ct=(gt&&gt.flushSync)===!0,Vt=o||l,Yt=eM(R.location,R.matches,s,nt,Ve,gt==null?void 0:gt.relative),Dt=Rv(Vt,Yt,s),Lt=wr(Dt,Vt,Yt);if(Lt.active&&Lt.matches&&(Dt=Lt.matches),!Dt){mt(Te,Ve,qd(404,{pathname:Yt}),{flushSync:Ct});return}let{path:an,submission:_n,error:sn}=lZ(!0,Yt,gt);if(sn){mt(Te,Ve,sn,{flushSync:Ct});return}let Ln=e.getContext?await e.getContext():new tZ,er=(gt&&gt.preventScrollReset)===!0;if(_n&&vs(_n.formMethod)){await Ae(Te,Ve,an,Dt,Ln,Lt.active,Ct,er,_n,gt&&gt.unstable_defaultShouldRevalidate);return}B.set(Te,{routeId:Ve,path:an}),await He(Te,Ve,an,Dt,Ln,Lt.active,Ct,er,_n)}async function Ae(Te,Ve,nt,gt,Ct,Vt,Yt,Dt,Lt,an){ze(),B.delete(Te);let _n=R.fetchers.get(Te);at(Te,eRe(Lt,_n),{flushSync:Yt});let sn=new AbortController,Ln=BO(e.history,nt,sn.signal,Lt);if(Vt){let Bn=await Gr(gt,new URL(Ln.url).pathname,Ln.signal,Te);if(Bn.type==="aborted")return;if(Bn.type==="error"){mt(Te,Ve,Bn.error,{flushSync:Yt});return}else if(Bn.matches)gt=Bn.matches;else{mt(Te,Ve,qd(404,{pathname:nt}),{flushSync:Yt});return}}let er=lD(gt,nt);if(!er.route.action&&!er.route.lazy){let Bn=qd(405,{method:Lt.formMethod,pathname:nt,routeId:Ve});mt(Te,Ve,Bn,{flushSync:Yt});return}te.set(Te,sn);let Di=H,$i=f3(a,u,Ln,gt,er,r,Ct),hr=await Ye(Ln,$i,Ct,Te),wt=hr[er.route.id];if(!wt){for(let Bn of $i)if(hr[Bn.route.id]){wt=hr[Bn.route.id];break}}if(Ln.signal.aborted){te.get(Te)===sn&&te.delete(Te);return}if(oe.has(Te)){if(Lb(wt)||Qf(wt)){at(Te,Pm(void 0));return}}else{if(Lb(wt))if(te.delete(Te),M>Di){at(Te,Pm(void 0));return}else return X.add(Te),at(Te,t2(Lt)),je(Ln,wt,!1,{fetcherSubmission:Lt,preventScrollReset:Dt});if(Qf(wt)){mt(Te,Ve,wt.error);return}}let Kr=R.navigation.location||R.location,Zi=BO(e.history,Kr,sn.signal),yi=o||l,Ca=R.navigation.state!=="idle"?Rv(yi,R.navigation.location,s):R.matches;gr(Ca,"Didn't find any matches after fetcher action");let Xa=++H;G.set(Te,Xa);let Au=t2(Lt,wt.data);R.fetchers.set(Te,Au);let{dsMatches:Tu,revalidatingFetchers:Nl}=oZ(Zi,Ct,a,u,e.history,R,Ca,Lt,Kr,r,!1,K,J,oe,B,X,yi,s,e.patchRoutesOnNavigation!=null,[er.route.id,wt],an);Nl.filter(Bn=>Bn.key!==Te).forEach(Bn=>{let ku=Bn.key,gf=R.fetchers.get(ku),Ds=t2(void 0,gf?gf.data:void 0);R.fetchers.set(ku,Ds),Xt(ku),Bn.controller&&te.set(ku,Bn.controller)}),se({fetchers:new Map(R.fetchers)});let gc=()=>Nl.forEach(Bn=>Xt(Bn.key));sn.signal.addEventListener("abort",gc);let{loaderResults:xl,fetcherResults:ur}=await Ne(Tu,Nl,Zi,Ct);if(sn.signal.aborted)return;if(sn.signal.removeEventListener("abort",gc),G.delete(Te),te.delete(Te),Nl.forEach(Bn=>te.delete(Bn.key)),R.fetchers.has(Te)){let Bn=Pm(wt.data);R.fetchers.set(Te,Bn)}let Rr=eE(xl);if(Rr)return je(Zi,Rr.result,!1,{preventScrollReset:Dt});if(Rr=eE(ur),Rr)return X.add(Rr.key),je(Zi,Rr.result,!1,{preventScrollReset:Dt});let{loaderData:Sl,errors:tr}=mZ(R,Ca,xl,void 0,Nl,ur);ht(Xa),R.navigation.state==="loading"&&Xa>M?(gr(F,"Expected pending action"),q&&q.abort(),pe(R.navigation.location,{matches:Ca,loaderData:Sl,errors:tr,fetchers:new Map(R.fetchers)})):(se({errors:tr,loaderData:gZ(R.loaderData,Sl,Ca,tr),fetchers:new Map(R.fetchers)}),K=!1)}async function He(Te,Ve,nt,gt,Ct,Vt,Yt,Dt,Lt){let an=R.fetchers.get(Te);at(Te,t2(Lt,an?an.data:void 0),{flushSync:Yt});let _n=new AbortController,sn=BO(e.history,nt,_n.signal);if(Vt){let wt=await Gr(gt,new URL(sn.url).pathname,sn.signal,Te);if(wt.type==="aborted")return;if(wt.type==="error"){mt(Te,Ve,wt.error,{flushSync:Yt});return}else if(wt.matches)gt=wt.matches;else{mt(Te,Ve,qd(404,{pathname:nt}),{flushSync:Yt});return}}let Ln=lD(gt,nt);te.set(Te,_n);let er=H,Di=f3(a,u,sn,gt,Ln,r,Ct),hr=(await Ye(sn,Di,Ct,Te))[Ln.route.id];if(te.get(Te)===_n&&te.delete(Te),!sn.signal.aborted){if(oe.has(Te)){at(Te,Pm(void 0));return}if(Lb(hr))if(M>er){at(Te,Pm(void 0));return}else{X.add(Te),await je(sn,hr,!1,{preventScrollReset:Dt});return}if(Qf(hr)){mt(Te,Ve,hr.error);return}at(Te,Pm(hr.data))}}async function je(Te,Ve,nt,{submission:gt,fetcherSubmission:Ct,preventScrollReset:Vt,replace:Yt}={}){nt||(I==null||I.resolve(),I=null),Ve.response.headers.has("X-Remix-Revalidate")&&(K=!0);let Dt=Ve.response.headers.get("Location");gr(Dt,"Expected a Location header on the redirect Response"),Dt=hZ(Dt,new URL(Te.url),s,e.history);let Lt=Dx(R.location,Dt,{_isRedirect:!0});if(n){let Di=!1;if(Ve.response.headers.has("X-Remix-Reload-Document"))Di=!0;else if(lw(Dt)){const $i=Yhe(Dt,!0);Di=$i.origin!==t.location.origin||ff($i.pathname,s)==null}if(Di){Yt?t.location.replace(Dt):t.location.assign(Dt);return}}q=null;let an=Yt===!0||Ve.response.headers.has("X-Remix-Replace")?"REPLACE":"PUSH",{formMethod:_n,formAction:sn,formEncType:Ln}=R.navigation;!gt&&!Ct&&_n&&sn&&Ln&&(gt=bZ(R.navigation));let er=gt||Ct;if(R6e.has(Ve.response.status)&&er&&vs(er.formMethod))await Me(an,Lt,{submission:{...er,formAction:Dt},preventScrollReset:Vt||U,enableViewTransition:nt?V:void 0});else{let Di=xk(Lt,gt);await Me(an,Lt,{overrideNavigation:Di,fetcherSubmission:Ct,preventScrollReset:Vt||U,enableViewTransition:nt?V:void 0})}}async function Ye(Te,Ve,nt,gt){var Yt;let Ct,Vt={};try{Ct=await B6e(c,Te,Ve,gt,nt,!1)}catch(Dt){return Ve.filter(Lt=>Lt.shouldLoad).forEach(Lt=>{Vt[Lt.route.id]={type:"error",error:Dt}}),Vt}if(Te.signal.aborted)return Vt;if(!vs(Te.method))for(let Dt of Ve){if(((Yt=Ct[Dt.route.id])==null?void 0:Yt.type)==="error")break;!Ct.hasOwnProperty(Dt.route.id)&&!R.loaderData.hasOwnProperty(Dt.route.id)&&(!R.errors||!R.errors.hasOwnProperty(Dt.route.id))&&Dt.shouldCallHandler()&&(Ct[Dt.route.id]={type:"error",result:new Error(`No result returned from dataStrategy for route ${Dt.route.id}`)})}for(let[Dt,Lt]of Object.entries(Ct))if(Y6e(Lt)){let an=Lt.result;Vt[Dt]={type:"redirect",response:V6e(an,Te,Dt,Ve,s)}}else Vt[Dt]=await q6e(Lt);return Vt}async function Ne(Te,Ve,nt,gt){let Ct=Ye(nt,Te,gt,null),Vt=Promise.all(Ve.map(async Lt=>{if(Lt.matches&&Lt.match&&Lt.request&&Lt.controller){let _n=(await Ye(Lt.request,Lt.matches,gt,Lt.key))[Lt.match.route.id];return{[Lt.key]:_n}}else return Promise.resolve({[Lt.key]:{type:"error",error:qd(404,{pathname:Lt.path})}})})),Yt=await Ct,Dt=(await Vt).reduce((Lt,an)=>Object.assign(Lt,an),{});return{loaderResults:Yt,fetcherResults:Dt}}function ze(){K=!0,B.forEach((Te,Ve)=>{te.has(Ve)&&J.add(Ve),Xt(Ve)})}function at(Te,Ve,nt={}){R.fetchers.set(Te,Ve),se({fetchers:new Map(R.fetchers)},{flushSync:(nt&&nt.flushSync)===!0})}function mt(Te,Ve,nt,gt={}){let Ct=Av(R.matches,Ve);zt(Te),se({errors:{[Ct.route.id]:nt},fetchers:new Map(R.fetchers)},{flushSync:(gt&&gt.flushSync)===!0})}function Mt(Te){return ne.set(Te,(ne.get(Te)||0)+1),oe.has(Te)&&oe.delete(Te),R.fetchers.get(Te)||A6e}function bt(Te,Ve){Xt(Te,Ve==null?void 0:Ve.reason),at(Te,Pm(null))}function zt(Te){let Ve=R.fetchers.get(Te);te.has(Te)&&!(Ve&&Ve.state==="loading"&&G.has(Te))&&Xt(Te),B.delete(Te),G.delete(Te),X.delete(Te),oe.delete(Te),J.delete(Te),R.fetchers.delete(Te)}function qt(Te){let Ve=(ne.get(Te)||0)-1;Ve<=0?(ne.delete(Te),oe.add(Te)):ne.set(Te,Ve),se({fetchers:new Map(R.fetchers)})}function Xt(Te,Ve){let nt=te.get(Te);nt&&(nt.abort(Ve),te.delete(Te))}function nn(Te){for(let Ve of Te){let nt=Mt(Ve),gt=Pm(nt.data);R.fetchers.set(Ve,gt)}}function Fe(){let Te=[],Ve=!1;for(let nt of X){let gt=R.fetchers.get(nt);gr(gt,`Expected fetcher: ${nt}`),gt.state==="loading"&&(X.delete(nt),Te.push(nt),Ve=!0)}return nn(Te),Ve}function ht(Te){let Ve=[];for(let[nt,gt]of G)if(gt<Te){let Ct=R.fetchers.get(nt);gr(Ct,`Expected fetcher: ${nt}`),Ct.state==="loading"&&(Xt(nt),G.delete(nt),Ve.push(nt))}return nn(Ve),Ve.length>0}function kt(Te,Ve){let nt=R.blockers.get(Te)||LO;return ie.get(Te)!==Ve&&ie.set(Te,Ve),nt}function Qt(Te){R.blockers.delete(Te),ie.delete(Te)}function en(Te,Ve){let nt=R.blockers.get(Te)||LO;gr(nt.state==="unblocked"&&Ve.state==="blocked"||nt.state==="blocked"&&Ve.state==="blocked"||nt.state==="blocked"&&Ve.state==="proceeding"||nt.state==="blocked"&&Ve.state==="unblocked"||nt.state==="proceeding"&&Ve.state==="unblocked",`Invalid blocker state transition: ${nt.state} -> ${Ve.state}`);let gt=new Map(R.blockers);gt.set(Te,Ve),se({blockers:gt})}function In({currentLocation:Te,nextLocation:Ve,historyAction:nt}){if(ie.size===0)return;ie.size>1&&fu(!1,"A router only supports one blocker at a time");let gt=Array.from(ie.entries()),[Ct,Vt]=gt[gt.length-1],Yt=R.blockers.get(Ct);if(!(Yt&&Yt.state==="proceeding")&&Vt({currentLocation:Te,nextLocation:Ve,historyAction:nt}))return Ct}function rn(Te){let Ve=qd(404,{pathname:Te}),nt=o||l,{matches:gt,route:Ct}=J_(nt);return{notFoundMatches:gt,route:Ct,error:Ve}}function Xn(Te,Ve,nt){if(y=Te,S=Ve,v=nt||null,!_&&R.navigation===Ok){_=!0;let gt=Cr(R.location,R.matches);gt!=null&&se({restoreScrollPosition:gt})}return()=>{y=null,S=null,v=null}}function br(Te,Ve){return v&&v(Te,Ve.map(gt=>Ghe(gt,R.loaderData)))||Te.key}function Or(Te,Ve){if(y&&S){let nt=br(Te,Ve);y[nt]=S()}}function Cr(Te,Ve){if(y){let nt=br(Te,Ve),gt=y[nt];if(typeof gt=="number")return gt}return null}function wr(Te,Ve,nt){if(e.patchRoutesOnNavigation)if(Te){if(Object.keys(Te[0].params).length>0)return{active:!0,matches:L2(Ve,nt,s,!0)}}else return{active:!0,matches:L2(Ve,nt,s,!0)||[]};return{active:!1,matches:null}}async function Gr(Te,Ve,nt,gt){if(!e.patchRoutesOnNavigation)return{type:"success",matches:Te};let Ct=Te;for(;;){let Vt=o==null,Yt=o||l,Dt=u;try{await e.patchRoutesOnNavigation({signal:nt,path:Ve,matches:Ct,fetcherKey:gt,patch:(_n,sn)=>{nt.aborted||sZ(_n,sn,Yt,Dt,a,!1)}})}catch(_n){return{type:"error",error:_n,partialMatches:Ct}}finally{Vt&&!nt.aborted&&(l=[...l])}if(nt.aborted)return{type:"aborted"};let Lt=Rv(Yt,Ve,s),an=null;if(Lt){if(Object.keys(Lt[0].params).length===0)return{type:"success",matches:Lt};if(an=L2(Yt,Ve,s,!0),!(an&&Ct.length<an.length&&ji(Ct,an.slice(0,Ct.length))))return{type:"success",matches:Lt}}if(an||(an=L2(Yt,Ve,s,!0)),!an||ji(Ct,an))return{type:"success",matches:null};Ct=an}}function ji(Te,Ve){return Te.length===Ve.length&&Te.every((nt,gt)=>nt.route.id===Ve[gt].route.id)}function ai(Te){u={},o=$x(Te,a,void 0,u)}function ar(Te,Ve,nt=!1){let gt=o==null;sZ(Te,Ve,o||l,u,a,nt),gt&&(l=[...l],se({}))}return w={get basename(){return s},get future(){return d},get state(){return R},get routes(){return l},get window(){return t},initialize:ye,subscribe:Ee,enableScrollRestoration:Xn,navigate:ve,fetch:$e,revalidate:we,createHref:Te=>e.history.createHref(Te),encodeLocation:Te=>e.history.encodeLocation(Te),getFetcher:Mt,resetFetcher:bt,deleteFetcher:qt,dispose:me,getBlocker:kt,deleteBlocker:Qt,patchRoutes:ar,_internalFetchControllers:te,_internalSetRoutes:ai,_internalSetStateDoNotUseOrYouWillBreakYourApp(Te){se(Te)}},e.unstable_instrumentations&&(w=x6e(w,e.unstable_instrumentations.map(Te=>Te.router).filter(Boolean))),w}function P6e(e){return e!=null&&("formData"in e&&e.formData!=null||"body"in e&&e.body!==void 0)}function eM(e,t,n,r,i,a){let u,l;if(i){u=[];for(let s of t)if(u.push(s),s.route.id===i){l=s;break}}else u=t,l=t[t.length-1];let o=sw(r||".",ow(u),ff(e.pathname,n)||e.pathname,a==="path");if(r==null&&(o.search=e.search,o.hash=e.hash),(r==null||r===""||r===".")&&l){let s=gI(o.search);if(l.route.index&&!s)o.search=o.search?o.search.replace(/^\?/,"?index&"):"?index";else if(!l.route.index&&s){let c=new URLSearchParams(o.search),d=c.getAll("index");c.delete("index"),d.filter(m=>m).forEach(m=>c.append("index",m));let h=c.toString();o.search=h?`?${h}`:""}}return n!=="/"&&(o.pathname=m6e({basename:n,pathname:o.pathname})),k0(o)}function lZ(e,t,n){if(!n||!P6e(n))return{path:t};if(n.formMethod&&!W6e(n.formMethod))return{path:t,error:qd(405,{method:n.formMethod})};let r=()=>({path:t,error:qd(400,{type:"invalid-body"})}),a=(n.formMethod||"get").toUpperCase(),u=hpe(t);if(n.body!==void 0){if(n.formEncType==="text/plain"){if(!vs(a))return r();let d=typeof n.body=="string"?n.body:n.body instanceof FormData||n.body instanceof URLSearchParams?Array.from(n.body.entries()).reduce((h,[m,y])=>`${h}${m}=${y}
`,""):String(n.body);return{path:t,submission:{formMethod:a,formAction:u,formEncType:n.formEncType,formData:void 0,json:void 0,text:d}}}else if(n.formEncType==="application/json"){if(!vs(a))return r();try{let d=typeof n.body=="string"?JSON.parse(n.body):n.body;return{path:t,submission:{formMethod:a,formAction:u,formEncType:n.formEncType,formData:void 0,json:d,text:void 0}}}catch{return r()}}}gr(typeof FormData=="function","FormData is not available in this environment");let l,o;if(n.formData)l=rM(n.formData),o=n.formData;else if(n.body instanceof FormData)l=rM(n.body),o=n.body;else if(n.body instanceof URLSearchParams)l=n.body,o=pZ(l);else if(n.body==null)l=new URLSearchParams,o=new FormData;else try{l=new URLSearchParams(n.body),o=pZ(l)}catch{return r()}let s={formMethod:a,formAction:u,formEncType:n&&n.formEncType||"application/x-www-form-urlencoded",formData:o,json:void 0,text:void 0};if(vs(s.formMethod))return{path:t,submission:s};let c=yy(t);return e&&c.search&&gI(c.search)&&l.append("index",""),c.search=`?${l}`,{path:k0(c),submission:s}}function oZ(e,t,n,r,i,a,u,l,o,s,c,d,h,m,y,v,S,_,E,$,k){var W;let A=$?Qf($[1])?$[1].error:$[1].data:void 0,w=i.createURL(a.location),R=i.createURL(o),F;if(c&&a.errors){let K=Object.keys(a.errors)[0];F=u.findIndex(J=>J.route.id===K)}else if($&&Qf($[1])){let K=$[0];F=u.findIndex(J=>J.route.id===K)-1}let I=$?$[1].statusCode:void 0,U=I&&I>=400,q={currentUrl:w,currentParams:((W=a.matches[0])==null?void 0:W.params)||{},nextUrl:R,nextParams:u[0].params,...l,actionResult:A,actionStatus:I},V=DS(u),N=u.map((K,J)=>{let{route:te}=K,H=null;if(F!=null&&J>F?H=!1:te.lazy?H=!0:pI(te)?c?H=tM(te,a.loaderData,a.errors):M6e(a.loaderData,a.matches[J],K)&&(H=!0):H=!1,H!==null)return nM(n,r,e,V,K,s,t,H);let M=!1;typeof k=="boolean"?M=k:U?M=!1:(d||w.pathname+w.search===R.pathname+R.search||w.search!==R.search||F6e(a.matches[J],K))&&(M=!0);let G={...q,defaultShouldRevalidate:M},X=Y2(K,G);return nM(n,r,e,V,K,s,t,X,G,k)}),j=[];return y.forEach((K,J)=>{if(c||!u.some(oe=>oe.route.id===K.routeId)||m.has(J))return;let te=a.fetchers.get(J),H=te&&te.state!=="idle"&&te.data===void 0,M=Rv(S,K.path,_);if(!M){if(E&&H)return;j.push({key:J,routeId:K.routeId,path:K.path,matches:null,match:null,request:null,controller:null});return}if(v.has(J))return;let G=lD(M,K.path),X=new AbortController,B=BO(i,K.path,X.signal),ne=null;if(h.has(J))h.delete(J),ne=f3(n,r,B,M,G,s,t);else if(H)d&&(ne=f3(n,r,B,M,G,s,t));else{let oe;typeof k=="boolean"?oe=k:U?oe=!1:oe=d;let ie={...q,defaultShouldRevalidate:oe};Y2(G,ie)&&(ne=f3(n,r,B,M,G,s,t,ie))}ne&&j.push({key:J,routeId:K.routeId,path:K.path,matches:ne,match:G,request:B,controller:X})}),{dsMatches:N,revalidatingFetchers:j}}function pI(e){return e.loader!=null||e.middleware!=null&&e.middleware.length>0}function tM(e,t,n){if(e.lazy)return!0;if(!pI(e))return!1;let r=t!=null&&e.id in t,i=n!=null&&n[e.id]!==void 0;return!r&&i?!1:typeof e.loader=="function"&&e.loader.hydrate===!0?!0:!r&&!i}function M6e(e,t,n){let r=!t||n.route.id!==t.route.id,i=!e.hasOwnProperty(n.route.id);return r||i}function F6e(e,t){let n=e.route.path;return e.pathname!==t.pathname||n!=null&&n.endsWith("*")&&e.params["*"]!==t.params["*"]}function Y2(e,t){if(e.route.shouldRevalidate){let n=e.route.shouldRevalidate(t);if(typeof n=="boolean")return n}return t.defaultShouldRevalidate}function sZ(e,t,n,r,i,a){let u;if(e){let s=r[e];gr(s,`No route found to patch children into: routeId = ${e}`),s.children||(s.children=[]),u=s.children}else u=n;let l=[],o=[];if(t.forEach(s=>{let c=u.find(d=>ope(s,d));c?o.push({existingRoute:c,newRoute:s}):l.push(s)}),l.length>0){let s=$x(l,i,[e||"_","patch",String((u==null?void 0:u.length)||"0")],r);u.push(...s)}if(a&&o.length>0)for(let s=0;s<o.length;s++){let{existingRoute:c,newRoute:d}=o[s],h=c,[m]=$x([d],i,[],{},!0);Object.assign(h,{element:m.element?m.element:h.element,errorElement:m.errorElement?m.errorElement:h.errorElement,hydrateFallbackElement:m.hydrateFallbackElement?m.hydrateFallbackElement:h.hydrateFallbackElement})}}function ope(e,t){return"id"in e&&"id"in t&&e.id===t.id?!0:e.index===t.index&&e.path===t.path&&e.caseSensitive===t.caseSensitive?(!e.children||e.children.length===0)&&(!t.children||t.children.length===0)?!0:e.children.every((n,r)=>{var i;return(i=t.children)==null?void 0:i.some(a=>ope(n,a))}):!1}var cZ=new WeakMap,spe=({key:e,route:t,manifest:n,mapRouteProperties:r})=>{let i=n[t.id];if(gr(i,"No route found in manifest"),!i.lazy||typeof i.lazy!="object")return;let a=i.lazy[e];if(!a)return;let u=cZ.get(i);u||(u={},cZ.set(i,u));let l=u[e];if(l)return l;let o=(async()=>{let s=J5e(e),d=i[e]!==void 0&&e!=="hasErrorBoundary";if(s)fu(!s,"Route property "+e+" is not a supported lazy route property. This property will be ignored."),u[e]=Promise.resolve();else if(d)fu(!1,`Route "${i.id}" has a static property "${e}" defined. The lazy property will be ignored.`);else{let h=await a();h!=null&&(Object.assign(i,{[e]:h}),Object.assign(i,r(i)))}typeof i.lazy=="object"&&(i.lazy[e]=void 0,Object.values(i.lazy).every(h=>h===void 0)&&(i.lazy=void 0))})();return u[e]=o,o},fZ=new WeakMap;function N6e(e,t,n,r,i){let a=n[e.id];if(gr(a,"No route found in manifest"),!e.lazy)return{lazyRoutePromise:void 0,lazyHandlerPromise:void 0};if(typeof e.lazy=="function"){let c=fZ.get(a);if(c)return{lazyRoutePromise:c,lazyHandlerPromise:c};let d=(async()=>{gr(typeof e.lazy=="function","No lazy route function found");let h=await e.lazy(),m={};for(let y in h){let v=h[y];if(v===void 0)continue;let S=t6e(y),E=a[y]!==void 0&&y!=="hasErrorBoundary";S?fu(!S,"Route property "+y+" is not a supported property to be returned from a lazy route function. This property will be ignored."):E?fu(!E,`Route "${a.id}" has a static property "${y}" defined but its lazy function is also returning a value for this property. The lazy route property "${y}" will be ignored.`):m[y]=v}Object.assign(a,m),Object.assign(a,{...r(a),lazy:void 0})})();return fZ.set(a,d),d.catch(()=>{}),{lazyRoutePromise:d,lazyHandlerPromise:d}}let u=Object.keys(e.lazy),l=[],o;for(let c of u){if(i&&i.includes(c))continue;let d=spe({key:c,route:e,manifest:n,mapRouteProperties:r});d&&(l.push(d),c===t&&(o=d))}let s=l.length>0?Promise.all(l).then(()=>{}):void 0;return s==null||s.catch(()=>{}),o==null||o.catch(()=>{}),{lazyRoutePromise:s,lazyHandlerPromise:o}}async function dZ(e){let t=e.matches.filter(i=>i.shouldLoad),n={};return(await Promise.all(t.map(i=>i.resolve()))).forEach((i,a)=>{n[t[a].route.id]=i}),n}async function I6e(e){return e.matches.some(t=>t.route.middleware)?cpe(e,()=>dZ(e)):dZ(e)}function cpe(e,t){return L6e(e,t,r=>{if(K6e(r))throw r;return r},H6e,n);function n(r,i,a){if(a)return Promise.resolve(Object.assign(a.value,{[i]:{type:"error",result:r}}));{let{matches:u}=e,l=Math.min(Math.max(u.findIndex(s=>s.route.id===i),0),Math.max(u.findIndex(s=>s.shouldCallHandler()),0)),o=Av(u,u[l].route.id).route.id;return Promise.resolve({[o]:{type:"error",result:r}})}}}async function L6e(e,t,n,r,i){let{matches:a,request:u,params:l,context:o,unstable_pattern:s}=e,c=a.flatMap(h=>h.route.middleware?h.route.middleware.map(m=>[h.route.id,m]):[]);return await fpe({request:u,params:l,context:o,unstable_pattern:s},c,t,n,r,i)}async function fpe(e,t,n,r,i,a,u=0){let{request:l}=e;if(l.signal.aborted)throw l.signal.reason??new Error(`Request aborted: ${l.method} ${l.url}`);let o=t[u];if(!o)return await n();let[s,c]=o,d,h=async()=>{if(d)throw new Error("You may only call `next()` once per middleware");try{return d={value:await fpe(e,t,n,r,i,a,u+1)},d.value}catch(m){return d={value:await a(m,s,d)},d.value}};try{let m=await c(e,h),y=m!=null?r(m):void 0;return i(y)?y:d?y??d.value:(d={value:await h()},d.value)}catch(m){return await a(m,s,d)}}function dpe(e,t,n,r,i){let a=spe({key:"middleware",route:r.route,manifest:t,mapRouteProperties:e}),u=N6e(r.route,vs(n.method)?"action":"loader",t,e,i);return{middleware:a,route:u.lazyRoutePromise,handler:u.lazyHandlerPromise}}function nM(e,t,n,r,i,a,u,l,o=null,s){let c=!1,d=dpe(e,t,n,i,a);return{...i,_lazyPromises:d,shouldLoad:l,shouldRevalidateArgs:o,shouldCallHandler(h){return c=!0,o?typeof s=="boolean"?Y2(i,{...o,defaultShouldRevalidate:s}):typeof h=="boolean"?Y2(i,{...o,defaultShouldRevalidate:h}):Y2(i,o):l},resolve(h){let{lazy:m,loader:y,middleware:v}=i.route,S=c||l||h&&!vs(n.method)&&(m||y),_=v&&v.length>0&&!y&&!m;return S&&(vs(n.method)||!_)?Q6e({request:n,unstable_pattern:r,match:i,lazyHandlerPromise:d==null?void 0:d.handler,lazyRoutePromise:d==null?void 0:d.route,handlerOverride:h,scopedContext:u}):Promise.resolve({type:"data",result:void 0})}}}function f3(e,t,n,r,i,a,u,l=null){return r.map(o=>o.route.id!==i.route.id?{...o,shouldLoad:!1,shouldRevalidateArgs:l,shouldCallHandler:()=>!1,_lazyPromises:dpe(e,t,n,o,a),resolve:()=>Promise.resolve({type:"data",result:void 0})}:nM(e,t,n,DS(r),o,a,u,!0,l))}async function B6e(e,t,n,r,i,a){n.some(s=>{var c;return(c=s._lazyPromises)==null?void 0:c.middleware})&&await Promise.all(n.map(s=>{var c;return(c=s._lazyPromises)==null?void 0:c.middleware}));let u={request:t,unstable_pattern:DS(n),params:n[0].params,context:i,matches:n},o=await e({...u,fetcherKey:r,runClientMiddleware:s=>{let c=u;return cpe(c,()=>s({...c,fetcherKey:r,runClientMiddleware:()=>{throw new Error("Cannot call `runClientMiddleware()` from within an `runClientMiddleware` handler")}}))}});try{await Promise.all(n.flatMap(s=>{var c,d;return[(c=s._lazyPromises)==null?void 0:c.handler,(d=s._lazyPromises)==null?void 0:d.route]}))}catch{}return o}async function Q6e({request:e,unstable_pattern:t,match:n,lazyHandlerPromise:r,lazyRoutePromise:i,handlerOverride:a,scopedContext:u}){let l,o,s=vs(e.method),c=s?"action":"loader",d=h=>{let m,y=new Promise((_,E)=>m=E);o=()=>m(),e.signal.addEventListener("abort",o);let v=_=>typeof h!="function"?Promise.reject(new Error(`You cannot call the handler for a route which defines a boolean "${c}" [routeId: ${n.route.id}]`)):h({request:e,unstable_pattern:t,params:n.params,context:u},..._!==void 0?[_]:[]),S=(async()=>{try{return{type:"data",result:await(a?a(E=>v(E)):v())}}catch(_){return{type:"error",result:_}}})();return Promise.race([S,y])};try{let h=s?n.route.action:n.route.loader;if(r||i)if(h){let m,[y]=await Promise.all([d(h).catch(v=>{m=v}),r,i]);if(m!==void 0)throw m;l=y}else{await r;let m=s?n.route.action:n.route.loader;if(m)[l]=await Promise.all([d(m),i]);else if(c==="action"){let y=new URL(e.url),v=y.pathname+y.search;throw qd(405,{method:e.method,pathname:v,routeId:n.route.id})}else return{type:"data",result:void 0}}else if(h)l=await d(h);else{let m=new URL(e.url),y=m.pathname+m.search;throw qd(404,{pathname:y})}}catch(h){return{type:"error",result:h}}finally{o&&e.signal.removeEventListener("abort",o)}return l}async function z6e(e){let t=e.headers.get("Content-Type");return t&&/\bapplication\/json\b/.test(t)?e.body==null?null:e.json():e.text()}async function q6e(e){var r,i,a,u,l;let{result:t,type:n}=e;if(mI(t)){let o;try{o=await z6e(t)}catch(s){return{type:"error",error:s}}return n==="error"?{type:"error",error:new ES(t.status,t.statusText,o),statusCode:t.status,headers:t.headers}:{type:"data",data:o,statusCode:t.status,headers:t.headers}}return n==="error"?yZ(t)?t.data instanceof Error?{type:"error",error:t.data,statusCode:(r=t.init)==null?void 0:r.status,headers:(i=t.init)!=null&&i.headers?new Headers(t.init.headers):void 0}:{type:"error",error:Z6e(t),statusCode:Cx(t)?t.status:void 0,headers:(a=t.init)!=null&&a.headers?new Headers(t.init.headers):void 0}:{type:"error",error:t,statusCode:Cx(t)?t.status:void 0}:yZ(t)?{type:"data",data:t.data,statusCode:(u=t.init)==null?void 0:u.status,headers:(l=t.init)!=null&&l.headers?new Headers(t.init.headers):void 0}:{type:"data",data:t}}function V6e(e,t,n,r,i){let a=e.headers.get("Location");if(gr(a,"Redirects returned/thrown from loaders/actions must have a Location header"),!lw(a)){let u=r.slice(0,r.findIndex(l=>l.route.id===n)+1);a=eM(new URL(t.url),u,i,a),e.headers.set("Location",a)}return e}function hZ(e,t,n,r){let i=["about:","blob:","chrome:","chrome-untrusted:","content:","data:","devtools:","file:","filesystem:","javascript:"];if(lw(e)){let a=e,u=a.startsWith("//")?new URL(t.protocol+a):new URL(a);if(i.includes(u.protocol))throw new Error("Invalid redirect location");let l=ff(u.pathname,n)!=null;if(u.origin===t.origin&&l)return u.pathname+u.search+u.hash}try{let a=r.createURL(e);if(i.includes(a.protocol))throw new Error("Invalid redirect location")}catch{}return e}function BO(e,t,n,r){let i=e.createURL(hpe(t)).toString(),a={signal:n};if(r&&vs(r.formMethod)){let{formMethod:u,formEncType:l}=r;a.method=u.toUpperCase(),l==="application/json"?(a.headers=new Headers({"Content-Type":l}),a.body=JSON.stringify(r.json)):l==="text/plain"?a.body=r.text:l==="application/x-www-form-urlencoded"&&r.formData?a.body=rM(r.formData):a.body=r.formData}return new Request(i,a)}function rM(e){let t=new URLSearchParams;for(let[n,r]of e.entries())t.append(n,typeof r=="string"?r:r.name);return t}function pZ(e){let t=new FormData;for(let[n,r]of e.entries())t.append(n,r);return t}function U6e(e,t,n,r=!1,i=!1){let a={},u=null,l,o=!1,s={},c=n&&Qf(n[1])?n[1].error:void 0;return e.forEach(d=>{if(!(d.route.id in t))return;let h=d.route.id,m=t[h];if(gr(!Lb(m),"Cannot handle redirect results in processLoaderData"),Qf(m)){let y=m.error;if(c!==void 0&&(y=c,c=void 0),u=u||{},i)u[h]=y;else{let v=Av(e,h);u[v.route.id]==null&&(u[v.route.id]=y)}r||(a[h]=lpe),o||(o=!0,l=Cx(m.error)?m.error.status:500),m.headers&&(s[h]=m.headers)}else a[h]=m.data,m.statusCode&&m.statusCode!==200&&!o&&(l=m.statusCode),m.headers&&(s[h]=m.headers)}),c!==void 0&&n&&(u={[n[0]]:c},n[2]&&(a[n[2]]=void 0)),{loaderData:a,errors:u,statusCode:l||200,loaderHeaders:s}}function mZ(e,t,n,r,i,a){let{loaderData:u,errors:l}=U6e(t,n,r);return i.filter(o=>!o.matches||o.matches.some(s=>s.shouldLoad)).forEach(o=>{let{key:s,match:c,controller:d}=o;if(d&&d.signal.aborted)return;let h=a[s];if(gr(h,"Did not find corresponding fetcher result"),Qf(h)){let m=Av(e.matches,c==null?void 0:c.route.id);l&&l[m.route.id]||(l={...l,[m.route.id]:h.error}),e.fetchers.delete(s)}else if(Lb(h))gr(!1,"Unhandled fetcher revalidation redirect");else{let m=Pm(h.data);e.fetchers.set(s,m)}}),{loaderData:u,errors:l}}function gZ(e,t,n,r){let i=Object.entries(t).filter(([,a])=>a!==lpe).reduce((a,[u,l])=>(a[u]=l,a),{});for(let a of n){let u=a.route.id;if(!t.hasOwnProperty(u)&&e.hasOwnProperty(u)&&a.route.loader&&(i[u]=e[u]),r&&r.hasOwnProperty(u))break}return i}function vZ(e){return e?Qf(e[1])?{actionData:{}}:{actionData:{[e[0]]:e[1].data}}:{}}function Av(e,t){return(t?e.slice(0,e.findIndex(r=>r.route.id===t)+1):[...e]).reverse().find(r=>r.route.hasErrorBoundary===!0)||e[0]}function J_(e){let t=e.length===1?e[0]:e.find(n=>n.index||!n.path||n.path==="/")||{id:"__shim-error-route__"};return{matches:[{params:{},pathname:"",pathnameBase:"",route:t}],route:t}}function qd(e,{pathname:t,routeId:n,method:r,type:i,message:a}={}){let u="Unknown Server Error",l="Unknown @remix-run/router error";return e===400?(u="Bad Request",r&&t&&n?l=`You made a ${r} request to "${t}" but did not provide a \`loader\` for route "${n}", so there is no way to handle the request.`:i==="invalid-body"&&(l="Unable to encode submission body")):e===403?(u="Forbidden",l=`Route "${n}" does not match URL "${t}"`):e===404?(u="Not Found",l=`No route matches URL "${t}"`):e===405&&(u="Method Not Allowed",r&&t&&n?l=`You made a ${r.toUpperCase()} request to "${t}" but did not provide an \`action\` for route "${n}", so there is no way to handle the request.`:r&&(l=`Invalid request method "${r.toUpperCase()}"`)),new ES(e||500,u,new Error(l),!0)}function eE(e){let t=Object.entries(e);for(let n=t.length-1;n>=0;n--){let[r,i]=t[n];if(Lb(i))return{key:r,result:i}}}function hpe(e){let t=typeof e=="string"?yy(e):e;return k0({...t,hash:""})}function j6e(e,t){return e.pathname!==t.pathname||e.search!==t.search?!1:e.hash===""?t.hash!=="":e.hash===t.hash?!0:t.hash!==""}function Z6e(e){var t,n;return new ES(((t=e.init)==null?void 0:t.status)??500,((n=e.init)==null?void 0:n.statusText)??"Internal Server Error",e.data)}function H6e(e){return e!=null&&typeof e=="object"&&Object.entries(e).every(([t,n])=>typeof t=="string"&&X6e(n))}function X6e(e){return e!=null&&typeof e=="object"&&"type"in e&&"result"in e&&(e.type==="data"||e.type==="error")}function Y6e(e){return mI(e.result)&&ape.has(e.result.status)}function Qf(e){return e.type==="error"}function Lb(e){return(e&&e.type)==="redirect"}function yZ(e){return typeof e=="object"&&e!=null&&"type"in e&&"data"in e&&"init"in e&&e.type==="DataWithResponseInit"}function mI(e){return e!=null&&typeof e.status=="number"&&typeof e.statusText=="string"&&typeof e.headers=="object"&&typeof e.body<"u"}function G6e(e){return ape.has(e)}function K6e(e){return mI(e)&&G6e(e.status)&&e.headers.has("Location")}function W6e(e){return w6e.has(e.toUpperCase())}function vs(e){return $6e.has(e.toUpperCase())}function gI(e){return new URLSearchParams(e).getAll("index").some(t=>t==="")}function lD(e,t){let n=typeof t=="string"?yy(t).search:t.search;if(e[e.length-1].route.index&&gI(n||""))return e[e.length-1];let r=epe(e);return r[r.length-1]}function bZ(e){let{formMethod:t,formAction:n,formEncType:r,text:i,formData:a,json:u}=e;if(!(!t||!n||!r)){if(i!=null)return{formMethod:t,formAction:n,formEncType:r,formData:void 0,json:void 0,text:i};if(a!=null)return{formMethod:t,formAction:n,formEncType:r,formData:a,json:void 0,text:void 0};if(u!==void 0)return{formMethod:t,formAction:n,formEncType:r,formData:void 0,json:u,text:void 0}}}function xk(e,t){return t?{state:"loading",location:e,formMethod:t.formMethod,formAction:t.formAction,formEncType:t.formEncType,formData:t.formData,json:t.json,text:t.text}:{state:"loading",location:e,formMethod:void 0,formAction:void 0,formEncType:void 0,formData:void 0,json:void 0,text:void 0}}function J6e(e,t){return{state:"submitting",location:e,formMethod:t.formMethod,formAction:t.formAction,formEncType:t.formEncType,formData:t.formData,json:t.json,text:t.text}}function t2(e,t){return e?{state:"loading",formMethod:e.formMethod,formAction:e.formAction,formEncType:e.formEncType,formData:e.formData,json:e.json,text:e.text,data:t}:{state:"loading",formMethod:void 0,formAction:void 0,formEncType:void 0,formData:void 0,json:void 0,text:void 0,data:t}}function eRe(e,t){return{state:"submitting",formMethod:e.formMethod,formAction:e.formAction,formEncType:e.formEncType,formData:e.formData,json:e.json,text:e.text,data:t?t.data:void 0}}function Pm(e){return{state:"idle",formMethod:void 0,formAction:void 0,formEncType:void 0,formData:void 0,json:void 0,text:void 0,data:e}}function tRe(e,t){try{let n=e.sessionStorage.getItem(upe);if(n){let r=JSON.parse(n);for(let[i,a]of Object.entries(r||{}))a&&Array.isArray(a)&&t.set(i,new Set(a||[]))}}catch{}}function nRe(e,t){if(t.size>0){let n={};for(let[r,i]of t)n[r]=[...i];try{e.sessionStorage.setItem(upe,JSON.stringify(n))}catch(r){fu(!1,`Failed to save applied view transitions in sessionStorage (${r}).`)}}}function OZ(){let e,t,n=new Promise((r,i)=>{e=async a=>{r(a);try{await n}catch{}},t=async a=>{i(a);try{await n}catch{}}});return{promise:n,resolve:e,reject:t}}var A1=P.createContext(null);A1.displayName="DataRouter";var $S=P.createContext(null);$S.displayName="DataRouterState";var ppe=P.createContext(!1);function rRe(){return P.useContext(ppe)}var vI=P.createContext({isTransitioning:!1});vI.displayName="ViewTransition";var mpe=P.createContext(new Map);mpe.displayName="Fetchers";var iRe=P.createContext(null);iRe.displayName="Await";var td=P.createContext(null);td.displayName="Navigation";var cw=P.createContext(null);cw.displayName="Location";var lh=P.createContext({outlet:null,matches:[],isDataRoute:!1});lh.displayName="Route";var yI=P.createContext(null);yI.displayName="RouteError";var gpe="REACT_ROUTER_ERROR",aRe="REDIRECT",uRe="ROUTE_ERROR_RESPONSE";function lRe(e){if(e.startsWith(`${gpe}:${aRe}:{`))try{let t=JSON.parse(e.slice(28));if(typeof t=="object"&&t&&typeof t.status=="number"&&typeof t.statusText=="string"&&typeof t.location=="string"&&typeof t.reloadDocument=="boolean"&&typeof t.replace=="boolean")return t}catch{}}function oRe(e){if(e.startsWith(`${gpe}:${uRe}:{`))try{let t=JSON.parse(e.slice(40));if(typeof t=="object"&&t&&typeof t.status=="number"&&typeof t.statusText=="string")return new ES(t.status,t.statusText,t.data)}catch{}}function sRe(e,{relative:t}={}){gr(J3(),"useHref() may be used only in the context of a <Router> component.");let{basename:n,navigator:r}=P.useContext(td),{hash:i,pathname:a,search:u}=CS(e,{relative:t}),l=a;return n!=="/"&&(l=a==="/"?n:C0([n,a])),r.createHref({pathname:l,search:u,hash:i})}function J3(){return P.useContext(cw)!=null}function xg(){return gr(J3(),"useLocation() may be used only in the context of a <Router> component."),P.useContext(cw).location}var vpe="You should call navigate() in a React.useEffect(), not when your component is first rendered.";function ype(e){P.useContext(td).static||P.useLayoutEffect(e)}function bI(){let{isDataRoute:e}=P.useContext(lh);return e?_Re():cRe()}function cRe(){gr(J3(),"useNavigate() may be used only in the context of a <Router> component.");let e=P.useContext(A1),{basename:t,navigator:n}=P.useContext(td),{matches:r}=P.useContext(lh),{pathname:i}=xg(),a=JSON.stringify(ow(r)),u=P.useRef(!1);return ype(()=>{u.current=!0}),P.useCallback((o,s={})=>{if(fu(u.current,vpe),!u.current)return;if(typeof o=="number"){n.go(o);return}let c=sw(o,JSON.parse(a),i,s.relative==="path");e==null&&t!=="/"&&(c.pathname=c.pathname==="/"?t:C0([t,c.pathname])),(s.replace?n.replace:n.push)(c,s.state,s)},[t,n,a,i,e])}var fRe=P.createContext(null);function dRe(e){let t=P.useContext(lh).outlet;return P.useMemo(()=>t&&P.createElement(fRe.Provider,{value:e},t),[t,e])}function hat(){let{matches:e}=P.useContext(lh),t=e[e.length-1];return t?t.params:{}}function CS(e,{relative:t}={}){let{matches:n}=P.useContext(lh),{pathname:r}=xg(),i=JSON.stringify(ow(n));return P.useMemo(()=>sw(e,JSON.parse(i),r,t==="path"),[e,i,r,t])}function hRe(e,t,n,r,i){gr(J3(),"useRoutes() may be used only in the context of a <Router> component.");let{navigator:a}=P.useContext(td),{matches:u}=P.useContext(lh),l=u[u.length-1],o=l?l.params:{},s=l?l.pathname:"/",c=l?l.pathnameBase:"/",d=l&&l.route;{let E=d&&d.path||"";Ope(s,!d||E.endsWith("*")||E.endsWith("*?"),`You rendered descendant <Routes> (or called \`useRoutes()\`) at "${s}" (under <Route path="${E}">) but the parent route path has no trailing "*". This means if you navigate deeper, the parent won't match anymore and therefore the child routes will never render.

Please change the parent <Route path="${E}"> to <Route path="${E==="/"?"*":`${E}/*`}">.`)}let h=xg(),m;m=h;let y=m.pathname||"/",v=y;if(c!=="/"){let E=c.replace(/^\//,"").split("/");v="/"+y.replace(/^\//,"").split("/").slice(E.length).join("/")}let S=Rv(e,{pathname:v});return fu(d||S!=null,`No routes matched location "${m.pathname}${m.search}${m.hash}" `),fu(S==null||S[S.length-1].route.element!==void 0||S[S.length-1].route.Component!==void 0||S[S.length-1].route.lazy!==void 0,`Matched leaf route at location "${m.pathname}${m.search}${m.hash}" does not have an element or Component. This means it will render an <Outlet /> with a null value by default resulting in an "empty" page.`),yRe(S&&S.map(E=>Object.assign({},E,{params:Object.assign({},o,E.params),pathname:C0([c,a.encodeLocation?a.encodeLocation(E.pathname.replace(/\?/g,"%3F").replace(/#/g,"%23")).pathname:E.pathname]),pathnameBase:E.pathnameBase==="/"?c:C0([c,a.encodeLocation?a.encodeLocation(E.pathnameBase.replace(/\?/g,"%3F").replace(/#/g,"%23")).pathname:E.pathnameBase])})),u,n,r,i)}function pRe(){let e=xRe(),t=Cx(e)?`${e.status} ${e.statusText}`:e instanceof Error?e.message:JSON.stringify(e),n=e instanceof Error?e.stack:null,r="rgba(200,200,200, 0.5)",i={padding:"0.5rem",backgroundColor:r},a={padding:"2px 4px",backgroundColor:r},u=null;return console.error("Error handled by React Router default ErrorBoundary:",e),u=P.createElement(P.Fragment,null,P.createElement("p",null,"💿 Hey developer 👋"),P.createElement("p",null,"You can provide a way better UX than this when your app throws errors by providing your own ",P.createElement("code",{style:a},"ErrorBoundary")," or"," ",P.createElement("code",{style:a},"errorElement")," prop on your route.")),P.createElement(P.Fragment,null,P.createElement("h2",null,"Unexpected Application Error!"),P.createElement("h3",{style:{fontStyle:"italic"}},t),n?P.createElement("pre",{style:i},n):null,u)}var mRe=P.createElement(pRe,null),bpe=class extends P.Component{constructor(e){super(e),this.state={location:e.location,revalidation:e.revalidation,error:e.error}}static getDerivedStateFromError(e){return{error:e}}static getDerivedStateFromProps(e,t){return t.location!==e.location||t.revalidation!=="idle"&&e.revalidation==="idle"?{error:e.error,location:e.location,revalidation:e.revalidation}:{error:e.error!==void 0?e.error:t.error,location:t.location,revalidation:e.revalidation||t.revalidation}}componentDidCatch(e,t){this.props.onError?this.props.onError(e,t):console.error("React Router caught the following error during render",e)}render(){let e=this.state.error;if(this.context&&typeof e=="object"&&e&&"digest"in e&&typeof e.digest=="string"){const n=oRe(e.digest);n&&(e=n)}let t=e!==void 0?P.createElement(lh.Provider,{value:this.props.routeContext},P.createElement(yI.Provider,{value:e,children:this.props.component})):this.props.children;return this.context?P.createElement(gRe,{error:e},t):t}};bpe.contextType=ppe;var Sk=new WeakMap;function gRe({children:e,error:t}){let{basename:n}=P.useContext(td);if(typeof t=="object"&&t&&"digest"in t&&typeof t.digest=="string"){let r=lRe(t.digest);if(r){let i=Sk.get(t);if(i)throw i;let a=npe(r.location,n);if(tpe&&!Sk.get(t))if(a.isExternal||r.reloadDocument)window.location.href=a.absoluteURL||a.to;else{const u=Promise.resolve().then(()=>window.__reactRouterDataRouter.navigate(a.to,{replace:r.replace}));throw Sk.set(t,u),u}return P.createElement("meta",{httpEquiv:"refresh",content:`0;url=${a.absoluteURL||a.to}`})}}return e}function vRe({routeContext:e,match:t,children:n}){let r=P.useContext(A1);return r&&r.static&&r.staticContext&&(t.route.errorElement||t.route.ErrorBoundary)&&(r.staticContext._deepestRenderedBoundaryId=t.route.id),P.createElement(lh.Provider,{value:e},n)}function yRe(e,t=[],n=null,r=null,i=null){if(e==null){if(!n)return null;if(n.errors)e=n.matches;else if(t.length===0&&!n.initialized&&n.matches.length>0)e=n.matches;else return null}let a=e,u=n==null?void 0:n.errors;if(u!=null){let c=a.findIndex(d=>d.route.id&&(u==null?void 0:u[d.route.id])!==void 0);gr(c>=0,`Could not find a matching route for errors on route IDs: ${Object.keys(u).join(",")}`),a=a.slice(0,Math.min(a.length,c+1))}let l=!1,o=-1;if(n)for(let c=0;c<a.length;c++){let d=a[c];if((d.route.HydrateFallback||d.route.hydrateFallbackElement)&&(o=c),d.route.id){let{loaderData:h,errors:m}=n,y=d.route.loader&&!h.hasOwnProperty(d.route.id)&&(!m||m[d.route.id]===void 0);if(d.route.lazy||y){l=!0,o>=0?a=a.slice(0,o+1):a=[a[0]];break}}}let s=n&&r?(c,d)=>{var h,m;r(c,{location:n.location,params:((m=(h=n.matches)==null?void 0:h[0])==null?void 0:m.params)??{},unstable_pattern:DS(n.matches),errorInfo:d})}:void 0;return a.reduceRight((c,d,h)=>{let m,y=!1,v=null,S=null;n&&(m=u&&d.route.id?u[d.route.id]:void 0,v=d.route.errorElement||mRe,l&&(o<0&&h===0?(Ope("route-fallback",!1,"No `HydrateFallback` element provided to render during initial hydration"),y=!0,S=null):o===h&&(y=!0,S=d.route.hydrateFallbackElement||null)));let _=t.concat(a.slice(0,h+1)),E=()=>{let $;return m?$=v:y?$=S:d.route.Component?$=P.createElement(d.route.Component,null):d.route.element?$=d.route.element:$=c,P.createElement(vRe,{match:d,routeContext:{outlet:c,matches:_,isDataRoute:n!=null},children:$})};return n&&(d.route.ErrorBoundary||d.route.errorElement||h===0)?P.createElement(bpe,{location:n.location,revalidation:n.revalidation,component:v,error:m,children:E(),routeContext:{outlet:null,matches:_,isDataRoute:!0},onError:s}):E()},null)}function OI(e){return`${e} must be used within a data router.  See https://reactrouter.com/en/main/routers/picking-a-router.`}function xI(e){let t=P.useContext(A1);return gr(t,OI(e)),t}function e4(e){let t=P.useContext($S);return gr(t,OI(e)),t}function bRe(e){let t=P.useContext(lh);return gr(t,OI(e)),t}function fw(e){let t=bRe(e),n=t.matches[t.matches.length-1];return gr(n.route.id,`${e} can only be used on routes that contain a unique "id"`),n.route.id}function ORe(){return fw("useRouteId")}function pat(){let e=xI("useRevalidator"),t=e4("useRevalidator"),n=P.useCallback(async()=>{await e.router.revalidate()},[e.router]);return P.useMemo(()=>({revalidate:n,state:t.revalidation}),[n,t.revalidation])}function mat(){let{matches:e,loaderData:t}=e4("useMatches");return P.useMemo(()=>e.map(n=>Ghe(n,t)),[e,t])}function gat(){let e=e4("useLoaderData"),t=fw("useLoaderData");return e.loaderData[t]}function vat(e){return e4("useRouteLoaderData").loaderData[e]}function xRe(){var r;let e=P.useContext(yI),t=e4("useRouteError"),n=fw("useRouteError");return e!==void 0?e:(r=t.errors)==null?void 0:r[n]}var SRe=0;function yat(e){let{router:t,basename:n}=xI("useBlocker"),r=e4("useBlocker"),[i,a]=P.useState(""),u=P.useCallback(l=>{if(typeof e!="function")return!!e;if(n==="/")return e(l);let{currentLocation:o,nextLocation:s,historyAction:c}=l;return e({currentLocation:{...o,pathname:ff(o.pathname,n)||o.pathname},nextLocation:{...s,pathname:ff(s.pathname,n)||s.pathname},historyAction:c})},[n,e]);return P.useEffect(()=>{let l=String(++SRe);return a(l),()=>t.deleteBlocker(l)},[t]),P.useEffect(()=>{i!==""&&t.getBlocker(i,u)},[t,i,u]),i&&r.blockers.has(i)?r.blockers.get(i):LO}function _Re(){let{router:e}=xI("useNavigate"),t=fw("useNavigate"),n=P.useRef(!1);return ype(()=>{n.current=!0}),P.useCallback(async(i,a={})=>{fu(n.current,vpe),n.current&&(typeof i=="number"?await e.navigate(i):await e.navigate(i,{fromRouteId:t,...a}))},[e,t])}var xZ={};function Ope(e,t,n){!t&&!xZ[e]&&(xZ[e]=!0,fu(!1,n))}var SZ={};function _Z(e,t){!e&&!SZ[t]&&(SZ[t]=!0,console.warn(t))}var ERe="useOptimistic",EZ=wD[ERe],DRe=()=>{};function $Re(e){return EZ?EZ(e):[e,DRe]}function CRe(e){let t={hasErrorBoundary:e.hasErrorBoundary||e.ErrorBoundary!=null||e.errorElement!=null};return e.Component&&(e.element&&fu(!1,"You should not include both `Component` and `element` on your route - `Component` will be used."),Object.assign(t,{element:P.createElement(e.Component),Component:void 0})),e.HydrateFallback&&(e.hydrateFallbackElement&&fu(!1,"You should not include both `HydrateFallback` and `hydrateFallbackElement` on your route - `HydrateFallback` will be used."),Object.assign(t,{hydrateFallbackElement:P.createElement(e.HydrateFallback),HydrateFallback:void 0})),e.ErrorBoundary&&(e.errorElement&&fu(!1,"You should not include both `ErrorBoundary` and `errorElement` on your route - `ErrorBoundary` will be used."),Object.assign(t,{errorElement:P.createElement(e.ErrorBoundary),ErrorBoundary:void 0})),t}var wRe=["HydrateFallback","hydrateFallbackElement"],RRe=class{constructor(){this.status="pending",this.promise=new Promise((e,t)=>{this.resolve=n=>{this.status==="pending"&&(this.status="resolved",e(n))},this.reject=n=>{this.status==="pending"&&(this.status="rejected",t(n))}})}};function ARe({router:e,flushSync:t,onError:n,unstable_useTransitions:r}){r=rRe()||r;let[a,u]=P.useState(e.state),[l,o]=$Re(a),[s,c]=P.useState(),[d,h]=P.useState({isTransitioning:!1}),[m,y]=P.useState(),[v,S]=P.useState(),[_,E]=P.useState(),$=P.useRef(new Map),k=P.useCallback((F,{deletedFetchers:I,newErrors:U,flushSync:q,viewTransitionOpts:V})=>{U&&n&&Object.values(U).forEach(j=>{var W;return n(j,{location:F.location,params:((W=F.matches[0])==null?void 0:W.params)??{},unstable_pattern:DS(F.matches)})}),F.fetchers.forEach((j,W)=>{j.data!==void 0&&$.current.set(W,j.data)}),I.forEach(j=>$.current.delete(j)),_Z(q===!1||t!=null,'You provided the `flushSync` option to a router update, but you are not using the `<RouterProvider>` from `react-router/dom` so `ReactDOM.flushSync()` is unavailable.  Please update your app to `import { RouterProvider } from "react-router/dom"` and ensure you have `react-dom` installed as a dependency to use the `flushSync` option.');let N=e.window!=null&&e.window.document!=null&&typeof e.window.document.startViewTransition=="function";if(_Z(V==null||N,"You provided the `viewTransition` option to a router update, but you do not appear to be running in a DOM environment as `window.startViewTransition` is not available."),!V||!N){t&&q?t(()=>u(F)):r===!1?u(F):P.startTransition(()=>{r===!0&&o(j=>DZ(j,F)),u(F)});return}if(t&&q){t(()=>{v&&(m==null||m.resolve(),v.skipTransition()),h({isTransitioning:!0,flushSync:!0,currentLocation:V.currentLocation,nextLocation:V.nextLocation})});let j=e.window.document.startViewTransition(()=>{t(()=>u(F))});j.finished.finally(()=>{t(()=>{y(void 0),S(void 0),c(void 0),h({isTransitioning:!1})})}),t(()=>S(j));return}v?(m==null||m.resolve(),v.skipTransition(),E({state:F,currentLocation:V.currentLocation,nextLocation:V.nextLocation})):(c(F),h({isTransitioning:!0,flushSync:!1,currentLocation:V.currentLocation,nextLocation:V.nextLocation}))},[e.window,t,v,m,r,o,n]);P.useLayoutEffect(()=>e.subscribe(k),[e,k]),P.useEffect(()=>{d.isTransitioning&&!d.flushSync&&y(new RRe)},[d]),P.useEffect(()=>{if(m&&s&&e.window){let F=s,I=m.promise,U=e.window.document.startViewTransition(async()=>{r===!1?u(F):P.startTransition(()=>{r===!0&&o(q=>DZ(q,F)),u(F)}),await I});U.finished.finally(()=>{y(void 0),S(void 0),c(void 0),h({isTransitioning:!1})}),S(U)}},[s,m,e.window,r,o]),P.useEffect(()=>{m&&s&&l.location.key===s.location.key&&m.resolve()},[m,v,l.location,s]),P.useEffect(()=>{!d.isTransitioning&&_&&(c(_.state),h({isTransitioning:!0,flushSync:!1,currentLocation:_.currentLocation,nextLocation:_.nextLocation}),E(void 0))},[d.isTransitioning,_]);let A=P.useMemo(()=>({createHref:e.createHref,encodeLocation:e.encodeLocation,go:F=>e.navigate(F),push:(F,I,U)=>e.navigate(F,{state:I,preventScrollReset:U==null?void 0:U.preventScrollReset}),replace:(F,I,U)=>e.navigate(F,{replace:!0,state:I,preventScrollReset:U==null?void 0:U.preventScrollReset})}),[e]),w=e.basename||"/",R=P.useMemo(()=>({router:e,navigator:A,static:!1,basename:w,onError:n}),[e,A,w,n]);return P.createElement(P.Fragment,null,P.createElement(A1.Provider,{value:R},P.createElement($S.Provider,{value:l},P.createElement(mpe.Provider,{value:$.current},P.createElement(vI.Provider,{value:d},P.createElement(MRe,{basename:w,location:l.location,navigationType:l.historyAction,navigator:A,unstable_useTransitions:r},P.createElement(TRe,{routes:e.routes,future:e.future,state:l,onError:n})))))),null)}function DZ(e,t){return{...e,navigation:t.navigation.state!=="idle"?t.navigation:e.navigation,revalidation:t.revalidation!=="idle"?t.revalidation:e.revalidation,actionData:t.navigation.state!=="submitting"?t.actionData:e.actionData,fetchers:t.fetchers}}var TRe=P.memo(kRe);function kRe({routes:e,future:t,state:n,onError:r}){return hRe(e,void 0,n,r,t)}function bat({to:e,replace:t,state:n,relative:r}){gr(J3(),"<Navigate> may be used only in the context of a <Router> component.");let{static:i}=P.useContext(td);fu(!i,"<Navigate> must not be used on the initial render in a <StaticRouter>. This is a no-op, but you should modify your code so the <Navigate> is only ever rendered in response to some user interaction or state change.");let{matches:a}=P.useContext(lh),{pathname:u}=xg(),l=bI(),o=sw(e,ow(a),u,r==="path"),s=JSON.stringify(o);return P.useEffect(()=>{l(JSON.parse(s),{replace:t,state:n,relative:r})},[l,s,r,t,n]),null}function Oat(e){return dRe(e.context)}function PRe(e){gr(!1,"A <Route> is only ever to be used as the child of <Routes> element, never rendered directly. Please wrap your <Route> in a <Routes>.")}function MRe({basename:e="/",children:t=null,location:n,navigationType:r="POP",navigator:i,static:a=!1,unstable_useTransitions:u}){gr(!J3(),"You cannot render a <Router> inside another <Router>. You should never have more than one in your app.");let l=e.replace(/^\/*/,"/"),o=P.useMemo(()=>({basename:l,navigator:i,static:a,unstable_useTransitions:u,future:{}}),[l,i,a,u]);typeof n=="string"&&(n=yy(n));let{pathname:s="/",search:c="",hash:d="",state:h=null,key:m="default"}=n,y=P.useMemo(()=>{let v=ff(s,l);return v==null?null:{location:{pathname:v,search:c,hash:d,state:h,key:m},navigationType:r}},[l,s,c,d,h,m,r]);return fu(y!=null,`<Router basename="${l}"> is not able to match the URL "${s}${c}${d}" because it does not start with the basename, so the <Router> won't render anything.`),y==null?null:P.createElement(td.Provider,{value:o},P.createElement(cw.Provider,{children:t,value:y}))}function iM(e,t=[]){let n=[];return P.Children.forEach(e,(r,i)=>{if(!P.isValidElement(r))return;let a=[...t,i];if(r.type===P.Fragment){n.push.apply(n,iM(r.props.children,a));return}gr(r.type===PRe,`[${typeof r.type=="string"?r.type:r.type.name}] is not a <Route> component. All component children of <Routes> must be a <Route> or <React.Fragment>`),gr(!r.props.index||!r.props.children,"An index route cannot have child routes.");let u={id:r.props.id||a.join("-"),caseSensitive:r.props.caseSensitive,element:r.props.element,Component:r.props.Component,index:r.props.index,path:r.props.path,middleware:r.props.middleware,loader:r.props.loader,action:r.props.action,hydrateFallbackElement:r.props.hydrateFallbackElement,HydrateFallback:r.props.HydrateFallback,errorElement:r.props.errorElement,ErrorBoundary:r.props.ErrorBoundary,hasErrorBoundary:r.props.hasErrorBoundary===!0||r.props.ErrorBoundary!=null||r.props.errorElement!=null,shouldRevalidate:r.props.shouldRevalidate,handle:r.props.handle,lazy:r.props.lazy};r.props.children&&(u.children=iM(r.props.children,a)),n.push(u)}),n}var xat=iM,oD="get",sD="application/x-www-form-urlencoded";function dw(e){return typeof HTMLElement<"u"&&e instanceof HTMLElement}function FRe(e){return dw(e)&&e.tagName.toLowerCase()==="button"}function NRe(e){return dw(e)&&e.tagName.toLowerCase()==="form"}function IRe(e){return dw(e)&&e.tagName.toLowerCase()==="input"}function LRe(e){return!!(e.metaKey||e.altKey||e.ctrlKey||e.shiftKey)}function BRe(e,t){return e.button===0&&(!t||t==="_self")&&!LRe(e)}function aM(e=""){return new URLSearchParams(typeof e=="string"||Array.isArray(e)||e instanceof URLSearchParams?e:Object.keys(e).reduce((t,n)=>{let r=e[n];return t.concat(Array.isArray(r)?r.map(i=>[n,i]):[[n,r]])},[]))}function QRe(e,t){let n=aM(e);return t&&t.forEach((r,i)=>{n.has(i)||t.getAll(i).forEach(a=>{n.append(i,a)})}),n}var tE=null;function zRe(){if(tE===null)try{new FormData(document.createElement("form"),0),tE=!1}catch{tE=!0}return tE}var qRe=new Set(["application/x-www-form-urlencoded","multipart/form-data","text/plain"]);function _k(e){return e!=null&&!qRe.has(e)?(fu(!1,`"${e}" is not a valid \`encType\` for \`<Form>\`/\`<fetcher.Form>\` and will default to "${sD}"`),null):e}function VRe(e,t){let n,r,i,a,u;if(NRe(e)){let l=e.getAttribute("action");r=l?ff(l,t):null,n=e.getAttribute("method")||oD,i=_k(e.getAttribute("enctype"))||sD,a=new FormData(e)}else if(FRe(e)||IRe(e)&&(e.type==="submit"||e.type==="image")){let l=e.form;if(l==null)throw new Error('Cannot submit a <button> or <input type="submit"> without a <form>');let o=e.getAttribute("formaction")||l.getAttribute("action");if(r=o?ff(o,t):null,n=e.getAttribute("formmethod")||l.getAttribute("method")||oD,i=_k(e.getAttribute("formenctype"))||_k(l.getAttribute("enctype"))||sD,a=new FormData(l,e),!zRe()){let{name:s,type:c,value:d}=e;if(c==="image"){let h=s?`${s}.`:"";a.append(`${h}x`,"0"),a.append(`${h}y`,"0")}else s&&a.append(s,d)}}else{if(dw(e))throw new Error('Cannot submit element that is not <form>, <button>, or <input type="submit|image">');n=oD,r=null,i=sD,u=e}return a&&i==="text/plain"&&(u=a,a=void 0),{action:r,method:n.toLowerCase(),encType:i,formData:a,body:u}}Object.getOwnPropertyNames(Object.prototype).sort().join("\0");function SI(e,t){if(e===!1||e===null||typeof e>"u")throw new Error(t)}function URe(e,t,n,r){let i=typeof e=="string"?new URL(e,typeof window>"u"?"server://singlefetch/":window.location.origin):e;return n?i.pathname.endsWith("/")?i.pathname=`${i.pathname}_.${r}`:i.pathname=`${i.pathname}.${r}`:i.pathname==="/"?i.pathname=`_root.${r}`:t&&ff(i.pathname,t)==="/"?i.pathname=`${t.replace(/\/$/,"")}/_root.${r}`:i.pathname=`${i.pathname.replace(/\/$/,"")}.${r}`,i}async function jRe(e,t){if(e.id in t)return t[e.id];try{let n=await import(e.module);return t[e.id]=n,n}catch(n){return console.error(`Error loading route module \`${e.module}\`, reloading page...`),console.error(n),window.__reactRouterContext&&window.__reactRouterContext.isSpaMode,window.location.reload(),new Promise(()=>{})}}function ZRe(e){return e==null?!1:e.href==null?e.rel==="preload"&&typeof e.imageSrcSet=="string"&&typeof e.imageSizes=="string":typeof e.rel=="string"&&typeof e.href=="string"}async function HRe(e,t,n){let r=await Promise.all(e.map(async i=>{let a=t.routes[i.route.id];if(a){let u=await jRe(a,n);return u.links?u.links():[]}return[]}));return KRe(r.flat(1).filter(ZRe).filter(i=>i.rel==="stylesheet"||i.rel==="preload").map(i=>i.rel==="stylesheet"?{...i,rel:"prefetch",as:"style"}:{...i,rel:"prefetch"}))}function $Z(e,t,n,r,i,a){let u=(o,s)=>n[s]?o.route.id!==n[s].route.id:!0,l=(o,s)=>{var c;return n[s].pathname!==o.pathname||((c=n[s].route.path)==null?void 0:c.endsWith("*"))&&n[s].params["*"]!==o.params["*"]};return a==="assets"?t.filter((o,s)=>u(o,s)||l(o,s)):a==="data"?t.filter((o,s)=>{var d;let c=r.routes[o.route.id];if(!c||!c.hasLoader)return!1;if(u(o,s)||l(o,s))return!0;if(o.route.shouldRevalidate){let h=o.route.shouldRevalidate({currentUrl:new URL(i.pathname+i.search+i.hash,window.origin),currentParams:((d=n[0])==null?void 0:d.params)||{},nextUrl:new URL(e,window.origin),nextParams:o.params,defaultShouldRevalidate:!0});if(typeof h=="boolean")return h}return!0}):[]}function XRe(e,t,{includeHydrateFallback:n}={}){return YRe(e.map(r=>{let i=t.routes[r.route.id];if(!i)return[];let a=[i.module];return i.clientActionModule&&(a=a.concat(i.clientActionModule)),i.clientLoaderModule&&(a=a.concat(i.clientLoaderModule)),n&&i.hydrateFallbackModule&&(a=a.concat(i.hydrateFallbackModule)),i.imports&&(a=a.concat(i.imports)),a}).flat(1))}function YRe(e){return[...new Set(e)]}function GRe(e){let t={},n=Object.keys(e).sort();for(let r of n)t[r]=e[r];return t}function KRe(e,t){let n=new Set;return new Set(t),e.reduce((r,i)=>{let a=JSON.stringify(GRe(i));return n.has(a)||(n.add(a),r.push({key:a,link:i})),r},[])}function xpe(){let e=P.useContext(A1);return SI(e,"You must render this element inside a <DataRouterContext.Provider> element"),e}function WRe(){let e=P.useContext($S);return SI(e,"You must render this element inside a <DataRouterStateContext.Provider> element"),e}var _I=P.createContext(void 0);_I.displayName="FrameworkContext";function Spe(){let e=P.useContext(_I);return SI(e,"You must render this element inside a <HydratedRouter> element"),e}function JRe(e,t){let n=P.useContext(_I),[r,i]=P.useState(!1),[a,u]=P.useState(!1),{onFocus:l,onBlur:o,onMouseEnter:s,onMouseLeave:c,onTouchStart:d}=t,h=P.useRef(null);P.useEffect(()=>{if(e==="render"&&u(!0),e==="viewport"){let v=_=>{_.forEach(E=>{u(E.isIntersecting)})},S=new IntersectionObserver(v,{threshold:.5});return h.current&&S.observe(h.current),()=>{S.disconnect()}}},[e]),P.useEffect(()=>{if(r){let v=setTimeout(()=>{u(!0)},100);return()=>{clearTimeout(v)}}},[r]);let m=()=>{i(!0)},y=()=>{i(!1),u(!1)};return n?e!=="intent"?[a,h,{}]:[a,h,{onFocus:n2(l,m),onBlur:n2(o,y),onMouseEnter:n2(s,m),onMouseLeave:n2(c,y),onTouchStart:n2(d,m)}]:[!1,h,{}]}function n2(e,t){return n=>{e&&e(n),n.defaultPrevented||t(n)}}function eAe({page:e,...t}){let{router:n}=xpe(),r=P.useMemo(()=>Rv(n.routes,e,n.basename),[n.routes,e,n.basename]);return r?P.createElement(nAe,{page:e,matches:r,...t}):null}function tAe(e){let{manifest:t,routeModules:n}=Spe(),[r,i]=P.useState([]);return P.useEffect(()=>{let a=!1;return HRe(e,t,n).then(u=>{a||i(u)}),()=>{a=!0}},[e,t,n]),r}function nAe({page:e,matches:t,...n}){let r=xg(),{future:i,manifest:a,routeModules:u}=Spe(),{basename:l}=xpe(),{loaderData:o,matches:s}=WRe(),c=P.useMemo(()=>$Z(e,t,s,a,r,"data"),[e,t,s,a,r]),d=P.useMemo(()=>$Z(e,t,s,a,r,"assets"),[e,t,s,a,r]),h=P.useMemo(()=>{if(e===r.pathname+r.search+r.hash)return[];let v=new Set,S=!1;if(t.forEach(E=>{var k;let $=a.routes[E.route.id];!$||!$.hasLoader||(!c.some(A=>A.route.id===E.route.id)&&E.route.id in o&&((k=u[E.route.id])!=null&&k.shouldRevalidate)||$.hasClientLoader?S=!0:v.add(E.route.id))}),v.size===0)return[];let _=URe(e,l,i.unstable_trailingSlashAwareDataRequests,"data");return S&&v.size>0&&_.searchParams.set("_routes",t.filter(E=>v.has(E.route.id)).map(E=>E.route.id).join(",")),[_.pathname+_.search]},[l,i.unstable_trailingSlashAwareDataRequests,o,r,a,c,t,e,u]),m=P.useMemo(()=>XRe(d,a),[d,a]),y=tAe(d);return P.createElement(P.Fragment,null,h.map(v=>P.createElement("link",{key:v,rel:"prefetch",as:"fetch",href:v,...n})),m.map(v=>P.createElement("link",{key:v,rel:"modulepreload",href:v,...n})),y.map(({key:v,link:S})=>P.createElement("link",{key:v,nonce:n.nonce,...S})))}function rAe(...e){return t=>{e.forEach(n=>{typeof n=="function"?n(t):n!=null&&(n.current=t)})}}var iAe=typeof window<"u"&&typeof window.document<"u"&&typeof window.document.createElement<"u";try{iAe&&(window.__reactRouterVersion="7.12.0")}catch{}function Sat(e,t){return k6e({basename:t==null?void 0:t.basename,getContext:t==null?void 0:t.getContext,future:t==null?void 0:t.future,history:Y5e({window:t==null?void 0:t.window}),hydrationData:(t==null?void 0:t.hydrationData)||aAe(),routes:e,mapRouteProperties:CRe,hydrationRouteProperties:wRe,dataStrategy:t==null?void 0:t.dataStrategy,patchRoutesOnNavigation:t==null?void 0:t.patchRoutesOnNavigation,window:t==null?void 0:t.window,unstable_instrumentations:t==null?void 0:t.unstable_instrumentations}).initialize()}function aAe(){let e=window==null?void 0:window.__staticRouterHydrationData;return e&&e.errors&&(e={...e,errors:uAe(e.errors)}),e}function uAe(e){if(!e)return null;let t=Object.entries(e),n={};for(let[r,i]of t)if(i&&i.__type==="RouteErrorResponse")n[r]=new ES(i.status,i.statusText,i.data,i.internal===!0);else if(i&&i.__type==="Error"){if(i.__subType){let a=window[i.__subType];if(typeof a=="function")try{let u=new a(i.message);u.stack="",n[r]=u}catch{}}if(n[r]==null){let a=new Error(i.message);a.stack="",n[r]=a}}else n[r]=i;return n}var _pe=/^(?:[a-z][a-z0-9+.-]*:|\/\/)/i,Epe=P.forwardRef(function({onClick:t,discover:n="render",prefetch:r="none",relative:i,reloadDocument:a,replace:u,state:l,target:o,to:s,preventScrollReset:c,viewTransition:d,unstable_defaultShouldRevalidate:h,...m},y){let{basename:v,unstable_useTransitions:S}=P.useContext(td),_=typeof s=="string"&&_pe.test(s),E=npe(s,v);s=E.to;let $=sRe(s,{relative:i}),[k,A,w]=JRe(r,m),R=cAe(s,{replace:u,state:l,target:o,preventScrollReset:c,relative:i,viewTransition:d,unstable_defaultShouldRevalidate:h,unstable_useTransitions:S});function F(U){t&&t(U),U.defaultPrevented||R(U)}let I=P.createElement("a",{...m,...w,href:E.absoluteURL||$,onClick:E.isExternal||a?t:F,ref:rAe(y,A),target:o,"data-discover":!_&&n==="render"?"true":void 0});return k&&!_?P.createElement(P.Fragment,null,I,P.createElement(eAe,{page:$})):I});Epe.displayName="Link";var lAe=P.forwardRef(function({"aria-current":t="page",caseSensitive:n=!1,className:r="",end:i=!1,style:a,to:u,viewTransition:l,children:o,...s},c){let d=CS(u,{relative:s.relative}),h=xg(),m=P.useContext($S),{navigator:y,basename:v}=P.useContext(td),S=m!=null&&mAe(d)&&l===!0,_=y.encodeLocation?y.encodeLocation(d).pathname:d.pathname,E=h.pathname,$=m&&m.navigation&&m.navigation.location?m.navigation.location.pathname:null;n||(E=E.toLowerCase(),$=$?$.toLowerCase():null,_=_.toLowerCase()),$&&v&&($=ff($,v)||$);const k=_!=="/"&&_.endsWith("/")?_.length-1:_.length;let A=E===_||!i&&E.startsWith(_)&&E.charAt(k)==="/",w=$!=null&&($===_||!i&&$.startsWith(_)&&$.charAt(_.length)==="/"),R={isActive:A,isPending:w,isTransitioning:S},F=A?t:void 0,I;typeof r=="function"?I=r(R):I=[r,A?"active":null,w?"pending":null,S?"transitioning":null].filter(Boolean).join(" ");let U=typeof a=="function"?a(R):a;return P.createElement(Epe,{...s,"aria-current":F,className:I,ref:c,style:U,to:u,viewTransition:l},typeof o=="function"?o(R):o)});lAe.displayName="NavLink";var oAe=P.forwardRef(({discover:e="render",fetcherKey:t,navigate:n,reloadDocument:r,replace:i,state:a,method:u=oD,action:l,onSubmit:o,relative:s,preventScrollReset:c,viewTransition:d,unstable_defaultShouldRevalidate:h,...m},y)=>{let{unstable_useTransitions:v}=P.useContext(td),S=hAe(),_=pAe(l,{relative:s}),E=u.toLowerCase()==="get"?"get":"post",$=typeof l=="string"&&_pe.test(l),k=A=>{if(o&&o(A),A.defaultPrevented)return;A.preventDefault();let w=A.nativeEvent.submitter,R=(w==null?void 0:w.getAttribute("formmethod"))||u,F=()=>S(w||A.currentTarget,{fetcherKey:t,method:R,navigate:n,replace:i,state:a,relative:s,preventScrollReset:c,viewTransition:d,unstable_defaultShouldRevalidate:h});v&&n!==!1?P.startTransition(()=>F()):F()};return P.createElement("form",{ref:y,method:E,action:_,onSubmit:r?o:k,...m,"data-discover":!$&&e==="render"?"true":void 0})});oAe.displayName="Form";function sAe(e){return`${e} must be used within a data router.  See https://reactrouter.com/en/main/routers/picking-a-router.`}function Dpe(e){let t=P.useContext(A1);return gr(t,sAe(e)),t}function cAe(e,{target:t,replace:n,state:r,preventScrollReset:i,relative:a,viewTransition:u,unstable_defaultShouldRevalidate:l,unstable_useTransitions:o}={}){let s=bI(),c=xg(),d=CS(e,{relative:a});return P.useCallback(h=>{if(BRe(h,t)){h.preventDefault();let m=n!==void 0?n:k0(c)===k0(d),y=()=>s(e,{replace:m,state:r,preventScrollReset:i,relative:a,viewTransition:u,unstable_defaultShouldRevalidate:l});o?P.startTransition(()=>y()):y()}},[c,s,d,n,r,t,e,i,a,u,l,o])}function _at(e){fu(typeof URLSearchParams<"u","You cannot use the `useSearchParams` hook in a browser that does not support the URLSearchParams API. If you need to support Internet Explorer 11, we recommend you load a polyfill such as https://github.com/ungap/url-search-params.");let t=P.useRef(aM(e)),n=P.useRef(!1),r=xg(),i=P.useMemo(()=>QRe(r.search,n.current?null:t.current),[r.search]),a=bI(),u=P.useCallback((l,o)=>{const s=aM(typeof l=="function"?l(new URLSearchParams(i)):l);n.current=!0,a("?"+s,o)},[a,i]);return[i,u]}var fAe=0,dAe=()=>`__${String(++fAe)}__`;function hAe(){let{router:e}=Dpe("useSubmit"),{basename:t}=P.useContext(td),n=ORe(),r=e.fetch,i=e.navigate;return P.useCallback(async(a,u={})=>{let{action:l,method:o,encType:s,formData:c,body:d}=VRe(a,t);if(u.navigate===!1){let h=u.fetcherKey||dAe();await r(h,n,u.action||l,{unstable_defaultShouldRevalidate:u.unstable_defaultShouldRevalidate,preventScrollReset:u.preventScrollReset,formData:c,body:d,formMethod:u.method||o,formEncType:u.encType||s,flushSync:u.flushSync})}else await i(u.action||l,{unstable_defaultShouldRevalidate:u.unstable_defaultShouldRevalidate,preventScrollReset:u.preventScrollReset,formData:c,body:d,formMethod:u.method||o,formEncType:u.encType||s,replace:u.replace,state:u.state,fromRouteId:n,flushSync:u.flushSync,viewTransition:u.viewTransition})},[r,i,t,n])}function pAe(e,{relative:t}={}){let{basename:n}=P.useContext(td),r=P.useContext(lh);gr(r,"useFormAction must be used inside a RouteContext");let[i]=r.matches.slice(-1),a={...CS(e||".",{relative:t})},u=xg();if(e==null){a.search=u.search;let l=new URLSearchParams(a.search),o=l.getAll("index");if(o.some(c=>c==="")){l.delete("index"),o.filter(d=>d).forEach(d=>l.append("index",d));let c=l.toString();a.search=c?`?${c}`:""}}return(!e||e===".")&&i.route.index&&(a.search=a.search?a.search.replace(/^\?/,"?index&"):"?index"),n!=="/"&&(a.pathname=a.pathname==="/"?n:C0([n,a.pathname])),k0(a)}function mAe(e,{relative:t}={}){let n=P.useContext(vI);gr(n!=null,"`useViewTransitionState` must be used within `react-router-dom`'s `RouterProvider`.  Did you accidentally import `RouterProvider` from `react-router`?");let{basename:r}=Dpe("useViewTransitionState"),i=CS(e,{relative:t});if(!n.isTransitioning)return!1;let a=ff(n.currentLocation.pathname,r)||n.currentLocation.pathname,u=ff(n.nextLocation.pathname,r)||n.nextLocation.pathname;return d$(i.pathname,u)!=null||d$(i.pathname,a)!=null}var CZ=function(t,n){var r=arguments;if(n==null||!lS.call(n,"css"))return P.createElement.apply(void 0,r);var i=r.length,a=new Array(i);a[0]=zF,a[1]=QF(t,n);for(var u=2;u<i;u++)a[u]=r[u];return P.createElement.apply(null,a)};(function(e){var t;t||(t=e.JSX||(e.JSX={}))})(CZ||(CZ={}));var Eat=wJ(function(e,t){var n=e.styles,r=BF([n],void 0,P.useContext(RJ)),i=P.useRef();return iQ(function(){var a=t.key+"-global",u=new t.sheet.constructor({key:a,nonce:t.sheet.nonce,container:t.sheet.container,speedy:t.sheet.isSpeedy}),l=!1,o=document.querySelector('style[data-emotion="'+a+" "+r.name+'"]');return t.sheet.tags.length&&(u.before=t.sheet.tags[0]),o!==null&&(l=!0,o.setAttribute("data-emotion",a),u.hydrate([o])),i.current=[u,l],function(){u.flush()}},[t]),iQ(function(){var a=i.current,u=a[0],l=a[1];if(l){a[1]=!1;return}if(r.next!==void 0&&EJ(t,r.next,!0),u.tags.length){var o=u.tags[u.tags.length-1].nextElementSibling;u.before=o,u.flush()}t.insert("",r,u,!1)},[t,r.name]),null});function gAe(){for(var e=arguments.length,t=new Array(e),n=0;n<e;n++)t[n]=arguments[n];return BF(t)}function Dat(){var e=gAe.apply(void 0,arguments),t="animation-"+e.name;return{name:t,styles:"@keyframes "+t+"{"+e.styles+"}",anim:1,toString:function(){return"_EMO_"+this.name+"_"+this.styles+"_EMO_"}}}var Ek,wZ;function vAe(){return wZ||(wZ=1,Ek=function(){var e=document.getSelection();if(!e.rangeCount)return function(){};for(var t=document.activeElement,n=[],r=0;r<e.rangeCount;r++)n.push(e.getRangeAt(r));switch(t.tagName.toUpperCase()){case"INPUT":case"TEXTAREA":t.blur();break;default:t=null;break}return e.removeAllRanges(),function(){e.type==="Caret"&&e.removeAllRanges(),e.rangeCount||n.forEach(function(i){e.addRange(i)}),t&&t.focus()}}),Ek}var Dk,RZ;function yAe(){if(RZ)return Dk;RZ=1;var e=vAe(),t={"text/plain":"Text","text/html":"Url",default:"Text"},n="Copy to clipboard: #{key}, Enter";function r(a){var u=(/mac os x/i.test(navigator.userAgent)?"⌘":"Ctrl")+"+C";return a.replace(/#{\s*key\s*}/g,u)}function i(a,u){var l,o,s,c,d,h,m=!1;u||(u={}),l=u.debug||!1;try{s=e(),c=document.createRange(),d=document.getSelection(),h=document.createElement("span"),h.textContent=a,h.ariaHidden="true",h.style.all="unset",h.style.position="fixed",h.style.top=0,h.style.clip="rect(0, 0, 0, 0)",h.style.whiteSpace="pre",h.style.webkitUserSelect="text",h.style.MozUserSelect="text",h.style.msUserSelect="text",h.style.userSelect="text",h.addEventListener("copy",function(v){if(v.stopPropagation(),u.format)if(v.preventDefault(),typeof v.clipboardData>"u"){l&&console.warn("unable to use e.clipboardData"),l&&console.warn("trying IE specific stuff"),window.clipboardData.clearData();var S=t[u.format]||t.default;window.clipboardData.setData(S,a)}else v.clipboardData.clearData(),v.clipboardData.setData(u.format,a);u.onCopy&&(v.preventDefault(),u.onCopy(v.clipboardData))}),document.body.appendChild(h),c.selectNodeContents(h),d.addRange(c);var y=document.execCommand("copy");if(!y)throw new Error("copy command was unsuccessful");m=!0}catch(v){l&&console.error("unable to copy using execCommand: ",v),l&&console.warn("trying IE specific stuff");try{window.clipboardData.setData(u.format||"text",a),u.onCopy&&u.onCopy(window.clipboardData),m=!0}catch(S){l&&console.error("unable to copy using clipboardData: ",S),l&&console.error("falling back to prompt"),o=r("message"in u?u.message:n),window.prompt(o,a)}}finally{d&&(typeof d.removeRange=="function"?d.removeRange(c):d.removeAllRanges()),h&&document.body.removeChild(h),s()}return m}return Dk=i,Dk}var bAe=yAe();const $at=da(bAe),$pe=6048e5,OAe=864e5,xAe=6e4,SAe=36e5,nE=43200,AZ=1440,TZ=Symbol.for("constructDateFrom");function sp(e,t){return typeof e=="function"?e(t):e&&typeof e=="object"&&TZ in e?e[TZ](t):e instanceof Date?new e.constructor(t):new Date(t)}function $a(e,t){return sp(t||e,e)}function _Ae(e,t,n){const r=$a(e,n==null?void 0:n.in);return isNaN(t)?sp(e,NaN):(t&&r.setDate(r.getDate()+t),r)}function EAe(e,t,n){return sp(e,+$a(e)+t)}function DAe(e,t,n){return EAe(e,t*SAe)}let $Ae={};function t4(){return $Ae}function wx(e,t){var l,o,s,c;const n=t4(),r=(t==null?void 0:t.weekStartsOn)??((o=(l=t==null?void 0:t.locale)==null?void 0:l.options)==null?void 0:o.weekStartsOn)??n.weekStartsOn??((c=(s=n.locale)==null?void 0:s.options)==null?void 0:c.weekStartsOn)??0,i=$a(e,t==null?void 0:t.in),a=i.getDay(),u=(a<r?7:0)+a-r;return i.setDate(i.getDate()-u),i.setHours(0,0,0,0),i}function h$(e,t){return wx(e,{...t,weekStartsOn:1})}function Cpe(e,t){const n=$a(e,t==null?void 0:t.in),r=n.getFullYear(),i=sp(n,0);i.setFullYear(r+1,0,4),i.setHours(0,0,0,0);const a=h$(i),u=sp(n,0);u.setFullYear(r,0,4),u.setHours(0,0,0,0);const l=h$(u);return n.getTime()>=a.getTime()?r+1:n.getTime()>=l.getTime()?r:r-1}function p$(e){const t=$a(e),n=new Date(Date.UTC(t.getFullYear(),t.getMonth(),t.getDate(),t.getHours(),t.getMinutes(),t.getSeconds(),t.getMilliseconds()));return n.setUTCFullYear(t.getFullYear()),+e-+n}function wS(e,...t){const n=sp.bind(null,e||t.find(r=>typeof r=="object"));return t.map(n)}function kZ(e,t){const n=$a(e,t==null?void 0:t.in);return n.setHours(0,0,0,0),n}function wpe(e,t,n){const[r,i]=wS(n==null?void 0:n.in,e,t),a=kZ(r),u=kZ(i),l=+a-p$(a),o=+u-p$(u);return Math.round((l-o)/OAe)}function CAe(e,t){const n=Cpe(e,t),r=sp(e,0);return r.setFullYear(n,0,4),r.setHours(0,0,0,0),h$(r)}function wAe(e,t,n){const r=$a(e,n==null?void 0:n.in);return r.setTime(r.getTime()+t*xAe),r}function cD(e,t){const n=+$a(e)-+$a(t);return n<0?-1:n>0?1:n}function RAe(e){return e instanceof Date||typeof e=="object"&&Object.prototype.toString.call(e)==="[object Date]"}function AAe(e){return!(!RAe(e)&&typeof e!="number"||isNaN(+$a(e)))}function TAe(e,t,n){const[r,i]=wS(n==null?void 0:n.in,e,t),a=r.getFullYear()-i.getFullYear(),u=r.getMonth()-i.getMonth();return a*12+u}function Rpe(e){return t=>{const r=(e?Math[e]:Math.trunc)(t);return r===0?0:r}}function kAe(e,t){return+$a(e)-+$a(t)}function PAe(e,t){const n=$a(e,t==null?void 0:t.in);return n.setHours(23,59,59,999),n}function MAe(e,t){const n=$a(e,t==null?void 0:t.in),r=n.getMonth();return n.setFullYear(n.getFullYear(),r+1,0),n.setHours(23,59,59,999),n}function FAe(e,t){const n=$a(e,t==null?void 0:t.in);return+PAe(n,t)==+MAe(n,t)}function NAe(e,t,n){const[r,i,a]=wS(n==null?void 0:n.in,e,e,t),u=cD(i,a),l=Math.abs(TAe(i,a));if(l<1)return 0;i.getMonth()===1&&i.getDate()>27&&i.setDate(30),i.setMonth(i.getMonth()-u*l);let o=cD(i,a)===-u;FAe(r)&&l===1&&cD(r,a)===1&&(o=!1);const s=u*(l-+o);return s===0?0:s}function IAe(e,t,n){const r=kAe(e,t)/1e3;return Rpe(n==null?void 0:n.roundingMethod)(r)}function LAe(e,t){const n=$a(e,t==null?void 0:t.in);return n.setFullYear(n.getFullYear(),0,1),n.setHours(0,0,0,0),n}function Cat(e,t){const n=$a(e,t==null?void 0:t.in);return n.setMinutes(59,59,999),n}function wat(e,t){const n=$a(e,t==null?void 0:t.in);return n.setSeconds(59,999),n}const BAe={lessThanXSeconds:{one:"less than a second",other:"less than {{count}} seconds"},xSeconds:{one:"1 second",other:"{{count}} seconds"},halfAMinute:"half a minute",lessThanXMinutes:{one:"less than a minute",other:"less than {{count}} minutes"},xMinutes:{one:"1 minute",other:"{{count}} minutes"},aboutXHours:{one:"about 1 hour",other:"about {{count}} hours"},xHours:{one:"1 hour",other:"{{count}} hours"},xDays:{one:"1 day",other:"{{count}} days"},aboutXWeeks:{one:"about 1 week",other:"about {{count}} weeks"},xWeeks:{one:"1 week",other:"{{count}} weeks"},aboutXMonths:{one:"about 1 month",other:"about {{count}} months"},xMonths:{one:"1 month",other:"{{count}} months"},aboutXYears:{one:"about 1 year",other:"about {{count}} years"},xYears:{one:"1 year",other:"{{count}} years"},overXYears:{one:"over 1 year",other:"over {{count}} years"},almostXYears:{one:"almost 1 year",other:"almost {{count}} years"}},QAe=(e,t,n)=>{let r;const i=BAe[e];return typeof i=="string"?r=i:t===1?r=i.one:r=i.other.replace("{{count}}",t.toString()),n!=null&&n.addSuffix?n.comparison&&n.comparison>0?"in "+r:r+" ago":r};function $k(e){return(t={})=>{const n=t.width?String(t.width):e.defaultWidth;return e.formats[n]||e.formats[e.defaultWidth]}}const zAe={full:"EEEE, MMMM do, y",long:"MMMM do, y",medium:"MMM d, y",short:"MM/dd/yyyy"},qAe={full:"h:mm:ss a zzzz",long:"h:mm:ss a z",medium:"h:mm:ss a",short:"h:mm a"},VAe={full:"{{date}} 'at' {{time}}",long:"{{date}} 'at' {{time}}",medium:"{{date}}, {{time}}",short:"{{date}}, {{time}}"},UAe={date:$k({formats:zAe,defaultWidth:"full"}),time:$k({formats:qAe,defaultWidth:"full"}),dateTime:$k({formats:VAe,defaultWidth:"full"})},jAe={lastWeek:"'last' eeee 'at' p",yesterday:"'yesterday at' p",today:"'today at' p",tomorrow:"'tomorrow at' p",nextWeek:"eeee 'at' p",other:"P"},ZAe=(e,t,n,r)=>jAe[e];function r2(e){return(t,n)=>{const r=n!=null&&n.context?String(n.context):"standalone";let i;if(r==="formatting"&&e.formattingValues){const u=e.defaultFormattingWidth||e.defaultWidth,l=n!=null&&n.width?String(n.width):u;i=e.formattingValues[l]||e.formattingValues[u]}else{const u=e.defaultWidth,l=n!=null&&n.width?String(n.width):e.defaultWidth;i=e.values[l]||e.values[u]}const a=e.argumentCallback?e.argumentCallback(t):t;return i[a]}}const HAe={narrow:["B","A"],abbreviated:["BC","AD"],wide:["Before Christ","Anno Domini"]},XAe={narrow:["1","2","3","4"],abbreviated:["Q1","Q2","Q3","Q4"],wide:["1st quarter","2nd quarter","3rd quarter","4th quarter"]},YAe={narrow:["J","F","M","A","M","J","J","A","S","O","N","D"],abbreviated:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],wide:["January","February","March","April","May","June","July","August","September","October","November","December"]},GAe={narrow:["S","M","T","W","T","F","S"],short:["Su","Mo","Tu","We","Th","Fr","Sa"],abbreviated:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],wide:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"]},KAe={narrow:{am:"a",pm:"p",midnight:"mi",noon:"n",morning:"morning",afternoon:"afternoon",evening:"evening",night:"night"},abbreviated:{am:"AM",pm:"PM",midnight:"midnight",noon:"noon",morning:"morning",afternoon:"afternoon",evening:"evening",night:"night"},wide:{am:"a.m.",pm:"p.m.",midnight:"midnight",noon:"noon",morning:"morning",afternoon:"afternoon",evening:"evening",night:"night"}},WAe={narrow:{am:"a",pm:"p",midnight:"mi",noon:"n",morning:"in the morning",afternoon:"in the afternoon",evening:"in the evening",night:"at night"},abbreviated:{am:"AM",pm:"PM",midnight:"midnight",noon:"noon",morning:"in the morning",afternoon:"in the afternoon",evening:"in the evening",night:"at night"},wide:{am:"a.m.",pm:"p.m.",midnight:"midnight",noon:"noon",morning:"in the morning",afternoon:"in the afternoon",evening:"in the evening",night:"at night"}},JAe=(e,t)=>{const n=Number(e),r=n%100;if(r>20||r<10)switch(r%10){case 1:return n+"st";case 2:return n+"nd";case 3:return n+"rd"}return n+"th"},eTe={ordinalNumber:JAe,era:r2({values:HAe,defaultWidth:"wide"}),quarter:r2({values:XAe,defaultWidth:"wide",argumentCallback:e=>e-1}),month:r2({values:YAe,defaultWidth:"wide"}),day:r2({values:GAe,defaultWidth:"wide"}),dayPeriod:r2({values:KAe,defaultWidth:"wide",formattingValues:WAe,defaultFormattingWidth:"wide"})};function i2(e){return(t,n={})=>{const r=n.width,i=r&&e.matchPatterns[r]||e.matchPatterns[e.defaultMatchWidth],a=t.match(i);if(!a)return null;const u=a[0],l=r&&e.parsePatterns[r]||e.parsePatterns[e.defaultParseWidth],o=Array.isArray(l)?nTe(l,d=>d.test(u)):tTe(l,d=>d.test(u));let s;s=e.valueCallback?e.valueCallback(o):o,s=n.valueCallback?n.valueCallback(s):s;const c=t.slice(u.length);return{value:s,rest:c}}}function tTe(e,t){for(const n in e)if(Object.prototype.hasOwnProperty.call(e,n)&&t(e[n]))return n}function nTe(e,t){for(let n=0;n<e.length;n++)if(t(e[n]))return n}function rTe(e){return(t,n={})=>{const r=t.match(e.matchPattern);if(!r)return null;const i=r[0],a=t.match(e.parsePattern);if(!a)return null;let u=e.valueCallback?e.valueCallback(a[0]):a[0];u=n.valueCallback?n.valueCallback(u):u;const l=t.slice(i.length);return{value:u,rest:l}}}const iTe=/^(\d+)(th|st|nd|rd)?/i,aTe=/\d+/i,uTe={narrow:/^(b|a)/i,abbreviated:/^(b\.?\s?c\.?|b\.?\s?c\.?\s?e\.?|a\.?\s?d\.?|c\.?\s?e\.?)/i,wide:/^(before christ|before common era|anno domini|common era)/i},lTe={any:[/^b/i,/^(a|c)/i]},oTe={narrow:/^[1234]/i,abbreviated:/^q[1234]/i,wide:/^[1234](th|st|nd|rd)? quarter/i},sTe={any:[/1/i,/2/i,/3/i,/4/i]},cTe={narrow:/^[jfmasond]/i,abbreviated:/^(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)/i,wide:/^(january|february|march|april|may|june|july|august|september|october|november|december)/i},fTe={narrow:[/^j/i,/^f/i,/^m/i,/^a/i,/^m/i,/^j/i,/^j/i,/^a/i,/^s/i,/^o/i,/^n/i,/^d/i],any:[/^ja/i,/^f/i,/^mar/i,/^ap/i,/^may/i,/^jun/i,/^jul/i,/^au/i,/^s/i,/^o/i,/^n/i,/^d/i]},dTe={narrow:/^[smtwf]/i,short:/^(su|mo|tu|we|th|fr|sa)/i,abbreviated:/^(sun|mon|tue|wed|thu|fri|sat)/i,wide:/^(sunday|monday|tuesday|wednesday|thursday|friday|saturday)/i},hTe={narrow:[/^s/i,/^m/i,/^t/i,/^w/i,/^t/i,/^f/i,/^s/i],any:[/^su/i,/^m/i,/^tu/i,/^w/i,/^th/i,/^f/i,/^sa/i]},pTe={narrow:/^(a|p|mi|n|(in the|at) (morning|afternoon|evening|night))/i,any:/^([ap]\.?\s?m\.?|midnight|noon|(in the|at) (morning|afternoon|evening|night))/i},mTe={any:{am:/^a/i,pm:/^p/i,midnight:/^mi/i,noon:/^no/i,morning:/morning/i,afternoon:/afternoon/i,evening:/evening/i,night:/night/i}},gTe={ordinalNumber:rTe({matchPattern:iTe,parsePattern:aTe,valueCallback:e=>parseInt(e,10)}),era:i2({matchPatterns:uTe,defaultMatchWidth:"wide",parsePatterns:lTe,defaultParseWidth:"any"}),quarter:i2({matchPatterns:oTe,defaultMatchWidth:"wide",parsePatterns:sTe,defaultParseWidth:"any",valueCallback:e=>e+1}),month:i2({matchPatterns:cTe,defaultMatchWidth:"wide",parsePatterns:fTe,defaultParseWidth:"any"}),day:i2({matchPatterns:dTe,defaultMatchWidth:"wide",parsePatterns:hTe,defaultParseWidth:"any"}),dayPeriod:i2({matchPatterns:pTe,defaultMatchWidth:"any",parsePatterns:mTe,defaultParseWidth:"any"})},EI={code:"en-US",formatDistance:QAe,formatLong:UAe,formatRelative:ZAe,localize:eTe,match:gTe,options:{weekStartsOn:0,firstWeekContainsDate:1}};function vTe(e,t){const n=$a(e,t==null?void 0:t.in);return wpe(n,LAe(n))+1}function yTe(e,t){const n=$a(e,t==null?void 0:t.in),r=+h$(n)-+CAe(n);return Math.round(r/$pe)+1}function Ape(e,t){var c,d,h,m;const n=$a(e,t==null?void 0:t.in),r=n.getFullYear(),i=t4(),a=(t==null?void 0:t.firstWeekContainsDate)??((d=(c=t==null?void 0:t.locale)==null?void 0:c.options)==null?void 0:d.firstWeekContainsDate)??i.firstWeekContainsDate??((m=(h=i.locale)==null?void 0:h.options)==null?void 0:m.firstWeekContainsDate)??1,u=sp((t==null?void 0:t.in)||e,0);u.setFullYear(r+1,0,a),u.setHours(0,0,0,0);const l=wx(u,t),o=sp((t==null?void 0:t.in)||e,0);o.setFullYear(r,0,a),o.setHours(0,0,0,0);const s=wx(o,t);return+n>=+l?r+1:+n>=+s?r:r-1}function bTe(e,t){var l,o,s,c;const n=t4(),r=(t==null?void 0:t.firstWeekContainsDate)??((o=(l=t==null?void 0:t.locale)==null?void 0:l.options)==null?void 0:o.firstWeekContainsDate)??n.firstWeekContainsDate??((c=(s=n.locale)==null?void 0:s.options)==null?void 0:c.firstWeekContainsDate)??1,i=Ape(e,t),a=sp((t==null?void 0:t.in)||e,0);return a.setFullYear(i,0,r),a.setHours(0,0,0,0),wx(a,t)}function OTe(e,t){const n=$a(e,t==null?void 0:t.in),r=+wx(n,t)-+bTe(n,t);return Math.round(r/$pe)+1}function Gi(e,t){const n=e<0?"-":"",r=Math.abs(e).toString().padStart(t,"0");return n+r}const hv={y(e,t){const n=e.getFullYear(),r=n>0?n:1-n;return Gi(t==="yy"?r%100:r,t.length)},M(e,t){const n=e.getMonth();return t==="M"?String(n+1):Gi(n+1,2)},d(e,t){return Gi(e.getDate(),t.length)},a(e,t){const n=e.getHours()/12>=1?"pm":"am";switch(t){case"a":case"aa":return n.toUpperCase();case"aaa":return n;case"aaaaa":return n[0];case"aaaa":default:return n==="am"?"a.m.":"p.m."}},h(e,t){return Gi(e.getHours()%12||12,t.length)},H(e,t){return Gi(e.getHours(),t.length)},m(e,t){return Gi(e.getMinutes(),t.length)},s(e,t){return Gi(e.getSeconds(),t.length)},S(e,t){const n=t.length,r=e.getMilliseconds(),i=Math.trunc(r*Math.pow(10,n-3));return Gi(i,t.length)}},SO={midnight:"midnight",noon:"noon",morning:"morning",afternoon:"afternoon",evening:"evening",night:"night"},PZ={G:function(e,t,n){const r=e.getFullYear()>0?1:0;switch(t){case"G":case"GG":case"GGG":return n.era(r,{width:"abbreviated"});case"GGGGG":return n.era(r,{width:"narrow"});case"GGGG":default:return n.era(r,{width:"wide"})}},y:function(e,t,n){if(t==="yo"){const r=e.getFullYear(),i=r>0?r:1-r;return n.ordinalNumber(i,{unit:"year"})}return hv.y(e,t)},Y:function(e,t,n,r){const i=Ape(e,r),a=i>0?i:1-i;if(t==="YY"){const u=a%100;return Gi(u,2)}return t==="Yo"?n.ordinalNumber(a,{unit:"year"}):Gi(a,t.length)},R:function(e,t){const n=Cpe(e);return Gi(n,t.length)},u:function(e,t){const n=e.getFullYear();return Gi(n,t.length)},Q:function(e,t,n){const r=Math.ceil((e.getMonth()+1)/3);switch(t){case"Q":return String(r);case"QQ":return Gi(r,2);case"Qo":return n.ordinalNumber(r,{unit:"quarter"});case"QQQ":return n.quarter(r,{width:"abbreviated",context:"formatting"});case"QQQQQ":return n.quarter(r,{width:"narrow",context:"formatting"});case"QQQQ":default:return n.quarter(r,{width:"wide",context:"formatting"})}},q:function(e,t,n){const r=Math.ceil((e.getMonth()+1)/3);switch(t){case"q":return String(r);case"qq":return Gi(r,2);case"qo":return n.ordinalNumber(r,{unit:"quarter"});case"qqq":return n.quarter(r,{width:"abbreviated",context:"standalone"});case"qqqqq":return n.quarter(r,{width:"narrow",context:"standalone"});case"qqqq":default:return n.quarter(r,{width:"wide",context:"standalone"})}},M:function(e,t,n){const r=e.getMonth();switch(t){case"M":case"MM":return hv.M(e,t);case"Mo":return n.ordinalNumber(r+1,{unit:"month"});case"MMM":return n.month(r,{width:"abbreviated",context:"formatting"});case"MMMMM":return n.month(r,{width:"narrow",context:"formatting"});case"MMMM":default:return n.month(r,{width:"wide",context:"formatting"})}},L:function(e,t,n){const r=e.getMonth();switch(t){case"L":return String(r+1);case"LL":return Gi(r+1,2);case"Lo":return n.ordinalNumber(r+1,{unit:"month"});case"LLL":return n.month(r,{width:"abbreviated",context:"standalone"});case"LLLLL":return n.month(r,{width:"narrow",context:"standalone"});case"LLLL":default:return n.month(r,{width:"wide",context:"standalone"})}},w:function(e,t,n,r){const i=OTe(e,r);return t==="wo"?n.ordinalNumber(i,{unit:"week"}):Gi(i,t.length)},I:function(e,t,n){const r=yTe(e);return t==="Io"?n.ordinalNumber(r,{unit:"week"}):Gi(r,t.length)},d:function(e,t,n){return t==="do"?n.ordinalNumber(e.getDate(),{unit:"date"}):hv.d(e,t)},D:function(e,t,n){const r=vTe(e);return t==="Do"?n.ordinalNumber(r,{unit:"dayOfYear"}):Gi(r,t.length)},E:function(e,t,n){const r=e.getDay();switch(t){case"E":case"EE":case"EEE":return n.day(r,{width:"abbreviated",context:"formatting"});case"EEEEE":return n.day(r,{width:"narrow",context:"formatting"});case"EEEEEE":return n.day(r,{width:"short",context:"formatting"});case"EEEE":default:return n.day(r,{width:"wide",context:"formatting"})}},e:function(e,t,n,r){const i=e.getDay(),a=(i-r.weekStartsOn+8)%7||7;switch(t){case"e":return String(a);case"ee":return Gi(a,2);case"eo":return n.ordinalNumber(a,{unit:"day"});case"eee":return n.day(i,{width:"abbreviated",context:"formatting"});case"eeeee":return n.day(i,{width:"narrow",context:"formatting"});case"eeeeee":return n.day(i,{width:"short",context:"formatting"});case"eeee":default:return n.day(i,{width:"wide",context:"formatting"})}},c:function(e,t,n,r){const i=e.getDay(),a=(i-r.weekStartsOn+8)%7||7;switch(t){case"c":return String(a);case"cc":return Gi(a,t.length);case"co":return n.ordinalNumber(a,{unit:"day"});case"ccc":return n.day(i,{width:"abbreviated",context:"standalone"});case"ccccc":return n.day(i,{width:"narrow",context:"standalone"});case"cccccc":return n.day(i,{width:"short",context:"standalone"});case"cccc":default:return n.day(i,{width:"wide",context:"standalone"})}},i:function(e,t,n){const r=e.getDay(),i=r===0?7:r;switch(t){case"i":return String(i);case"ii":return Gi(i,t.length);case"io":return n.ordinalNumber(i,{unit:"day"});case"iii":return n.day(r,{width:"abbreviated",context:"formatting"});case"iiiii":return n.day(r,{width:"narrow",context:"formatting"});case"iiiiii":return n.day(r,{width:"short",context:"formatting"});case"iiii":default:return n.day(r,{width:"wide",context:"formatting"})}},a:function(e,t,n){const i=e.getHours()/12>=1?"pm":"am";switch(t){case"a":case"aa":return n.dayPeriod(i,{width:"abbreviated",context:"formatting"});case"aaa":return n.dayPeriod(i,{width:"abbreviated",context:"formatting"}).toLowerCase();case"aaaaa":return n.dayPeriod(i,{width:"narrow",context:"formatting"});case"aaaa":default:return n.dayPeriod(i,{width:"wide",context:"formatting"})}},b:function(e,t,n){const r=e.getHours();let i;switch(r===12?i=SO.noon:r===0?i=SO.midnight:i=r/12>=1?"pm":"am",t){case"b":case"bb":return n.dayPeriod(i,{width:"abbreviated",context:"formatting"});case"bbb":return n.dayPeriod(i,{width:"abbreviated",context:"formatting"}).toLowerCase();case"bbbbb":return n.dayPeriod(i,{width:"narrow",context:"formatting"});case"bbbb":default:return n.dayPeriod(i,{width:"wide",context:"formatting"})}},B:function(e,t,n){const r=e.getHours();let i;switch(r>=17?i=SO.evening:r>=12?i=SO.afternoon:r>=4?i=SO.morning:i=SO.night,t){case"B":case"BB":case"BBB":return n.dayPeriod(i,{width:"abbreviated",context:"formatting"});case"BBBBB":return n.dayPeriod(i,{width:"narrow",context:"formatting"});case"BBBB":default:return n.dayPeriod(i,{width:"wide",context:"formatting"})}},h:function(e,t,n){if(t==="ho"){let r=e.getHours()%12;return r===0&&(r=12),n.ordinalNumber(r,{unit:"hour"})}return hv.h(e,t)},H:function(e,t,n){return t==="Ho"?n.ordinalNumber(e.getHours(),{unit:"hour"}):hv.H(e,t)},K:function(e,t,n){const r=e.getHours()%12;return t==="Ko"?n.ordinalNumber(r,{unit:"hour"}):Gi(r,t.length)},k:function(e,t,n){let r=e.getHours();return r===0&&(r=24),t==="ko"?n.ordinalNumber(r,{unit:"hour"}):Gi(r,t.length)},m:function(e,t,n){return t==="mo"?n.ordinalNumber(e.getMinutes(),{unit:"minute"}):hv.m(e,t)},s:function(e,t,n){return t==="so"?n.ordinalNumber(e.getSeconds(),{unit:"second"}):hv.s(e,t)},S:function(e,t){return hv.S(e,t)},X:function(e,t,n){const r=e.getTimezoneOffset();if(r===0)return"Z";switch(t){case"X":return FZ(r);case"XXXX":case"XX":return Db(r);case"XXXXX":case"XXX":default:return Db(r,":")}},x:function(e,t,n){const r=e.getTimezoneOffset();switch(t){case"x":return FZ(r);case"xxxx":case"xx":return Db(r);case"xxxxx":case"xxx":default:return Db(r,":")}},O:function(e,t,n){const r=e.getTimezoneOffset();switch(t){case"O":case"OO":case"OOO":return"GMT"+MZ(r,":");case"OOOO":default:return"GMT"+Db(r,":")}},z:function(e,t,n){const r=e.getTimezoneOffset();switch(t){case"z":case"zz":case"zzz":return"GMT"+MZ(r,":");case"zzzz":default:return"GMT"+Db(r,":")}},t:function(e,t,n){const r=Math.trunc(+e/1e3);return Gi(r,t.length)},T:function(e,t,n){return Gi(+e,t.length)}};function MZ(e,t=""){const n=e>0?"-":"+",r=Math.abs(e),i=Math.trunc(r/60),a=r%60;return a===0?n+String(i):n+String(i)+t+Gi(a,2)}function FZ(e,t){return e%60===0?(e>0?"-":"+")+Gi(Math.abs(e)/60,2):Db(e,t)}function Db(e,t=""){const n=e>0?"-":"+",r=Math.abs(e),i=Gi(Math.trunc(r/60),2),a=Gi(r%60,2);return n+i+t+a}const NZ=(e,t)=>{switch(e){case"P":return t.date({width:"short"});case"PP":return t.date({width:"medium"});case"PPP":return t.date({width:"long"});case"PPPP":default:return t.date({width:"full"})}},Tpe=(e,t)=>{switch(e){case"p":return t.time({width:"short"});case"pp":return t.time({width:"medium"});case"ppp":return t.time({width:"long"});case"pppp":default:return t.time({width:"full"})}},xTe=(e,t)=>{const n=e.match(/(P+)(p+)?/)||[],r=n[1],i=n[2];if(!i)return NZ(e,t);let a;switch(r){case"P":a=t.dateTime({width:"short"});break;case"PP":a=t.dateTime({width:"medium"});break;case"PPP":a=t.dateTime({width:"long"});break;case"PPPP":default:a=t.dateTime({width:"full"});break}return a.replace("{{date}}",NZ(r,t)).replace("{{time}}",Tpe(i,t))},STe={p:Tpe,P:xTe},_Te=/^D+$/,ETe=/^Y+$/,DTe=["D","DD","YY","YYYY"];function $Te(e){return _Te.test(e)}function CTe(e){return ETe.test(e)}function wTe(e,t,n){const r=RTe(e,t,n);if(console.warn(r),DTe.includes(e))throw new RangeError(r)}function RTe(e,t,n){const r=e[0]==="Y"?"years":"days of the month";return`Use \`${e.toLowerCase()}\` instead of \`${e}\` (in \`${t}\`) for formatting ${r} to the input \`${n}\`; see: https://github.com/date-fns/date-fns/blob/master/docs/unicodeTokens.md`}const ATe=/[yYQqMLwIdDecihHKkms]o|(\w)\1*|''|'(''|[^'])+('|$)|./g,TTe=/P+p+|P+|p+|''|'(''|[^'])+('|$)|./g,kTe=/^'([^]*?)'?$/,PTe=/''/g,MTe=/[a-zA-Z]/;function FTe(e,t,n){var c,d,h,m,y,v,S,_;const r=t4(),i=(n==null?void 0:n.locale)??r.locale??EI,a=(n==null?void 0:n.firstWeekContainsDate)??((d=(c=n==null?void 0:n.locale)==null?void 0:c.options)==null?void 0:d.firstWeekContainsDate)??r.firstWeekContainsDate??((m=(h=r.locale)==null?void 0:h.options)==null?void 0:m.firstWeekContainsDate)??1,u=(n==null?void 0:n.weekStartsOn)??((v=(y=n==null?void 0:n.locale)==null?void 0:y.options)==null?void 0:v.weekStartsOn)??r.weekStartsOn??((_=(S=r.locale)==null?void 0:S.options)==null?void 0:_.weekStartsOn)??0,l=$a(e,n==null?void 0:n.in);if(!AAe(l))throw new RangeError("Invalid time value");let o=t.match(TTe).map(E=>{const $=E[0];if($==="p"||$==="P"){const k=STe[$];return k(E,i.formatLong)}return E}).join("").match(ATe).map(E=>{if(E==="''")return{isToken:!1,value:"'"};const $=E[0];if($==="'")return{isToken:!1,value:NTe(E)};if(PZ[$])return{isToken:!0,value:E};if($.match(MTe))throw new RangeError("Format string contains an unescaped latin alphabet character `"+$+"`");return{isToken:!1,value:E}});i.localize.preprocessor&&(o=i.localize.preprocessor(l,o));const s={firstWeekContainsDate:a,weekStartsOn:u,locale:i};return o.map(E=>{if(!E.isToken)return E.value;const $=E.value;(!(n!=null&&n.useAdditionalWeekYearTokens)&&CTe($)||!(n!=null&&n.useAdditionalDayOfYearTokens)&&$Te($))&&wTe($,t,String(e));const k=PZ[$[0]];return k(l,$,i.localize,s)}).join("")}function NTe(e){const t=e.match(kTe);return t?t[1].replace(PTe,"'"):e}function Rat(e,t,n){const r=t4(),i=(n==null?void 0:n.locale)??r.locale??EI,a=2520,u=cD(e,t);if(isNaN(u))throw new RangeError("Invalid time value");const l=Object.assign({},n,{addSuffix:n==null?void 0:n.addSuffix,comparison:u}),[o,s]=wS(n==null?void 0:n.in,...u>0?[t,e]:[e,t]),c=IAe(s,o),d=(p$(s)-p$(o))/1e3,h=Math.round((c-d)/60);let m;if(h<2)return n!=null&&n.includeSeconds?c<5?i.formatDistance("lessThanXSeconds",5,l):c<10?i.formatDistance("lessThanXSeconds",10,l):c<20?i.formatDistance("lessThanXSeconds",20,l):c<40?i.formatDistance("halfAMinute",0,l):c<60?i.formatDistance("lessThanXMinutes",1,l):i.formatDistance("xMinutes",1,l):h===0?i.formatDistance("lessThanXMinutes",1,l):i.formatDistance("xMinutes",h,l);if(h<45)return i.formatDistance("xMinutes",h,l);if(h<90)return i.formatDistance("aboutXHours",1,l);if(h<AZ){const y=Math.round(h/60);return i.formatDistance("aboutXHours",y,l)}else{if(h<a)return i.formatDistance("xDays",1,l);if(h<nE){const y=Math.round(h/AZ);return i.formatDistance("xDays",y,l)}else if(h<nE*2)return m=Math.round(h/nE),i.formatDistance("aboutXMonths",m,l)}if(m=NAe(s,o),m<12){const y=Math.round(h/nE);return i.formatDistance("xMonths",y,l)}else{const y=m%12,v=Math.trunc(m/12);return y<3?i.formatDistance("aboutXYears",v,l):y<9?i.formatDistance("overXYears",v,l):i.formatDistance("almostXYears",v+1,l)}}function Aat(e,t,n){var d,h;const[r,i]=wS(n==null?void 0:n.in,e,t),a=t4(),u=a.locale??EI,l=a.weekStartsOn??((h=(d=a.locale)==null?void 0:d.options)==null?void 0:h.weekStartsOn)??0,o=wpe(r,i);if(isNaN(o))throw new RangeError("Invalid time value");let s;o<-6?s="other":o<-1?s="lastWeek":o<0?s="yesterday":o<1?s="today":o<2?s="tomorrow":o<7?s="nextWeek":s="other";const c=u.formatRelative(s,r,i,{locale:u,weekStartsOn:l});return FTe(r,c,{locale:u,weekStartsOn:l})}function Tat(e,t){const n=$a(e,t==null?void 0:t.in);return n.setMinutes(0,0,0),n}function kat(e,t){const n=$a(e,t==null?void 0:t.in);return n.setSeconds(0,0),n}function Pat(e,t,n){return _Ae(e,-t,n)}function Mat(e,t){const n=(t==null?void 0:t.nearestTo)??1;if(n<1||n>30)return sp(e,NaN);const r=$a(e,t==null?void 0:t.in),i=r.getSeconds()/60,a=r.getMilliseconds()/1e3/60,u=r.getMinutes()+i+a,l=(t==null?void 0:t.roundingMethod)??"round",s=Rpe(l)(u/n)*n;return r.setMinutes(s,0,0),r}function Fat(e,t,n){return DAe(e,-t)}function Nat(e,t,n){return wAe(e,-15,n)}var Ck,IZ;function DI(){if(IZ)return Ck;IZ=1;function e(t){var n=typeof t;return t!=null&&(n=="object"||n=="function")}return Ck=e,Ck}var wk,LZ;function ITe(){if(LZ)return wk;LZ=1;var e=typeof Xd=="object"&&Xd&&Xd.Object===Object&&Xd;return wk=e,wk}var Rk,BZ;function kpe(){if(BZ)return Rk;BZ=1;var e=ITe(),t=typeof self=="object"&&self&&self.Object===Object&&self,n=e||t||Function("return this")();return Rk=n,Rk}var Ak,QZ;function LTe(){if(QZ)return Ak;QZ=1;var e=kpe(),t=function(){return e.Date.now()};return Ak=t,Ak}var Tk,zZ;function BTe(){if(zZ)return Tk;zZ=1;var e=/\s/;function t(n){for(var r=n.length;r--&&e.test(n.charAt(r)););return r}return Tk=t,Tk}var kk,qZ;function QTe(){if(qZ)return kk;qZ=1;var e=BTe(),t=/^\s+/;function n(r){return r&&r.slice(0,e(r)+1).replace(t,"")}return kk=n,kk}var Pk,VZ;function Ppe(){if(VZ)return Pk;VZ=1;var e=kpe(),t=e.Symbol;return Pk=t,Pk}var Mk,UZ;function zTe(){if(UZ)return Mk;UZ=1;var e=Ppe(),t=Object.prototype,n=t.hasOwnProperty,r=t.toString,i=e?e.toStringTag:void 0;function a(u){var l=n.call(u,i),o=u[i];try{u[i]=void 0;var s=!0}catch{}var c=r.call(u);return s&&(l?u[i]=o:delete u[i]),c}return Mk=a,Mk}var Fk,jZ;function qTe(){if(jZ)return Fk;jZ=1;var e=Object.prototype,t=e.toString;function n(r){return t.call(r)}return Fk=n,Fk}var Nk,ZZ;function VTe(){if(ZZ)return Nk;ZZ=1;var e=Ppe(),t=zTe(),n=qTe(),r="[object Null]",i="[object Undefined]",a=e?e.toStringTag:void 0;function u(l){return l==null?l===void 0?i:r:a&&a in Object(l)?t(l):n(l)}return Nk=u,Nk}var Ik,HZ;function UTe(){if(HZ)return Ik;HZ=1;function e(t){return t!=null&&typeof t=="object"}return Ik=e,Ik}var Lk,XZ;function jTe(){if(XZ)return Lk;XZ=1;var e=VTe(),t=UTe(),n="[object Symbol]";function r(i){return typeof i=="symbol"||t(i)&&e(i)==n}return Lk=r,Lk}var Bk,YZ;function ZTe(){if(YZ)return Bk;YZ=1;var e=QTe(),t=DI(),n=jTe(),r=NaN,i=/^[-+]0x[0-9a-f]+$/i,a=/^0b[01]+$/i,u=/^0o[0-7]+$/i,l=parseInt;function o(s){if(typeof s=="number")return s;if(n(s))return r;if(t(s)){var c=typeof s.valueOf=="function"?s.valueOf():s;s=t(c)?c+"":c}if(typeof s!="string")return s===0?s:+s;s=e(s);var d=a.test(s);return d||u.test(s)?l(s.slice(2),d?2:8):i.test(s)?r:+s}return Bk=o,Bk}var Qk,GZ;function HTe(){if(GZ)return Qk;GZ=1;var e=DI(),t=LTe(),n=ZTe(),r="Expected a function",i=Math.max,a=Math.min;function u(l,o,s){var c,d,h,m,y,v,S=0,_=!1,E=!1,$=!0;if(typeof l!="function")throw new TypeError(r);o=n(o)||0,e(s)&&(_=!!s.leading,E="maxWait"in s,h=E?i(n(s.maxWait)||0,o):h,$="trailing"in s?!!s.trailing:$);function k(N){var j=c,W=d;return c=d=void 0,S=N,m=l.apply(W,j),m}function A(N){return S=N,y=setTimeout(F,o),_?k(N):m}function w(N){var j=N-v,W=N-S,K=o-j;return E?a(K,h-W):K}function R(N){var j=N-v,W=N-S;return v===void 0||j>=o||j<0||E&&W>=h}function F(){var N=t();if(R(N))return I(N);y=setTimeout(F,w(N))}function I(N){return y=void 0,$&&c?k(N):(c=d=void 0,m)}function U(){y!==void 0&&clearTimeout(y),S=0,c=v=d=y=void 0}function q(){return y===void 0?m:I(t())}function V(){var N=t(),j=R(N);if(c=arguments,d=this,v=N,j){if(y===void 0)return A(v);if(E)return clearTimeout(y),y=setTimeout(F,o),k(v)}return y===void 0&&(y=setTimeout(F,o)),m}return V.cancel=U,V.flush=q,V}return Qk=u,Qk}var XTe=HTe();const Iat=da(XTe);var RS=e=>e.type==="checkbox",Bb=e=>e instanceof Date,ac=e=>e==null;const Mpe=e=>typeof e=="object";var Hu=e=>!ac(e)&&!Array.isArray(e)&&Mpe(e)&&!Bb(e),Fpe=e=>Hu(e)&&e.target?RS(e.target)?e.target.checked:e.target.value:e,YTe=e=>e.substring(0,e.search(/\.\d+(\.|$)/))||e,Npe=(e,t)=>e.has(YTe(t)),GTe=e=>{const t=e.constructor&&e.constructor.prototype;return Hu(t)&&t.hasOwnProperty("isPrototypeOf")},$I=typeof window<"u"&&typeof window.HTMLElement<"u"&&typeof document<"u";function Na(e){let t;const n=Array.isArray(e),r=typeof FileList<"u"?e instanceof FileList:!1;if(e instanceof Date)t=new Date(e);else if(!($I&&(e instanceof Blob||r))&&(n||Hu(e)))if(t=n?[]:Object.create(Object.getPrototypeOf(e)),!n&&!GTe(e))t=e;else for(const i in e)e.hasOwnProperty(i)&&(t[i]=Na(e[i]));else return e;return t}var hw=e=>/^\w*$/.test(e),La=e=>e===void 0,pw=e=>Array.isArray(e)?e.filter(Boolean):[],CI=e=>pw(e.replace(/["|']|\]/g,"").split(/\.|\[/)),Ut=(e,t,n)=>{if(!t||!Hu(e))return n;const r=(hw(t)?[t]:CI(t)).reduce((i,a)=>ac(i)?i:i[a],e);return La(r)||r===e?La(e[t])?n:e[t]:r},zf=e=>typeof e=="boolean",na=(e,t,n)=>{let r=-1;const i=hw(t)?[t]:CI(t),a=i.length,u=a-1;for(;++r<a;){const l=i[r];let o=n;if(r!==u){const s=e[l];o=Hu(s)||Array.isArray(s)?s:isNaN(+i[r+1])?{}:[]}if(l==="__proto__"||l==="constructor"||l==="prototype")return;e[l]=o,e=e[l]}};const m$={BLUR:"blur",FOCUS_OUT:"focusout",CHANGE:"change"},Hd={onBlur:"onBlur",onChange:"onChange",onSubmit:"onSubmit",onTouched:"onTouched",all:"all"},Am={max:"max",min:"min",maxLength:"maxLength",minLength:"minLength",pattern:"pattern",required:"required",validate:"validate"},wI=be.createContext(null);wI.displayName="HookFormContext";const mw=()=>be.useContext(wI),Lat=e=>{const{children:t,...n}=e;return be.createElement(wI.Provider,{value:n},t)};var Ipe=(e,t,n,r=!0)=>{const i={defaultValues:t._defaultValues};for(const a in e)Object.defineProperty(i,a,{get:()=>{const u=a;return t._proxyFormState[u]!==Hd.all&&(t._proxyFormState[u]=!r||Hd.all),n&&(n[u]=!0),e[u]}});return i};const gw=typeof window<"u"?be.useLayoutEffect:be.useEffect;function KTe(e){const t=mw(),{control:n=t.control,disabled:r,name:i,exact:a}=e||{},[u,l]=be.useState(n._formState),o=be.useRef({isDirty:!1,isLoading:!1,dirtyFields:!1,touchedFields:!1,validatingFields:!1,isValidating:!1,isValid:!1,errors:!1});return gw(()=>n._subscribe({name:i,formState:o.current,exact:a,callback:s=>{!r&&l({...n._formState,...s})}}),[i,r,a]),be.useEffect(()=>{o.current.isValid&&n._setValid(!0)},[n]),be.useMemo(()=>Ipe(u,n,o.current,!1),[u,n])}var b0=e=>typeof e=="string",Lpe=(e,t,n,r,i)=>b0(e)?(r&&t.watch.add(e),Ut(n,e,i)):Array.isArray(e)?e.map(a=>(r&&t.watch.add(a),Ut(n,a))):(r&&(t.watchAll=!0),n),uM=e=>ac(e)||!Mpe(e);function Zm(e,t,n=new WeakSet){if(uM(e)||uM(t))return e===t;if(Bb(e)&&Bb(t))return e.getTime()===t.getTime();const r=Object.keys(e),i=Object.keys(t);if(r.length!==i.length)return!1;if(n.has(e)||n.has(t))return!0;n.add(e),n.add(t);for(const a of r){const u=e[a];if(!i.includes(a))return!1;if(a!=="ref"){const l=t[a];if(Bb(u)&&Bb(l)||Hu(u)&&Hu(l)||Array.isArray(u)&&Array.isArray(l)?!Zm(u,l,n):u!==l)return!1}}return!0}function WTe(e){const t=mw(),{control:n=t.control,name:r,defaultValue:i,disabled:a,exact:u,compute:l}=e||{},o=be.useRef(i),s=be.useRef(l),c=be.useRef(void 0);s.current=l;const d=be.useMemo(()=>n._getWatch(r,o.current),[n,r]),[h,m]=be.useState(s.current?s.current(d):d);return gw(()=>n._subscribe({name:r,formState:{values:!0},exact:u,callback:y=>{if(!a){const v=Lpe(r,n._names,y.values||n._formValues,!1,o.current);if(s.current){const S=s.current(v);Zm(S,c.current)||(m(S),c.current=S)}else m(v)}}}),[n,a,r,u]),be.useEffect(()=>n._removeUnmounted()),h}function JTe(e){const t=mw(),{name:n,disabled:r,control:i=t.control,shouldUnregister:a,defaultValue:u}=e,l=Npe(i._names.array,n),o=be.useMemo(()=>Ut(i._formValues,n,Ut(i._defaultValues,n,u)),[i,n,u]),s=WTe({control:i,name:n,defaultValue:o,exact:!0}),c=KTe({control:i,name:n,exact:!0}),d=be.useRef(e),h=be.useRef(i.register(n,{...e.rules,value:s,...zf(e.disabled)?{disabled:e.disabled}:{}}));d.current=e;const m=be.useMemo(()=>Object.defineProperties({},{invalid:{enumerable:!0,get:()=>!!Ut(c.errors,n)},isDirty:{enumerable:!0,get:()=>!!Ut(c.dirtyFields,n)},isTouched:{enumerable:!0,get:()=>!!Ut(c.touchedFields,n)},isValidating:{enumerable:!0,get:()=>!!Ut(c.validatingFields,n)},error:{enumerable:!0,get:()=>Ut(c.errors,n)}}),[c,n]),y=be.useCallback(E=>h.current.onChange({target:{value:Fpe(E),name:n},type:m$.CHANGE}),[n]),v=be.useCallback(()=>h.current.onBlur({target:{value:Ut(i._formValues,n),name:n},type:m$.BLUR}),[n,i._formValues]),S=be.useCallback(E=>{const $=Ut(i._fields,n);$&&E&&($._f.ref={focus:()=>E.focus&&E.focus(),select:()=>E.select&&E.select(),setCustomValidity:k=>E.setCustomValidity(k),reportValidity:()=>E.reportValidity()})},[i._fields,n]),_=be.useMemo(()=>({name:n,value:s,...zf(r)||c.disabled?{disabled:c.disabled||r}:{},onChange:y,onBlur:v,ref:S}),[n,r,c.disabled,y,v,S,s]);return be.useEffect(()=>{const E=i._options.shouldUnregister||a;i.register(n,{...d.current.rules,...zf(d.current.disabled)?{disabled:d.current.disabled}:{}});const $=(k,A)=>{const w=Ut(i._fields,k);w&&w._f&&(w._f.mount=A)};if($(n,!0),E){const k=Na(Ut(i._options.defaultValues,n));na(i._defaultValues,n,k),La(Ut(i._formValues,n))&&na(i._formValues,n,k)}return!l&&i.register(n),()=>{(l?E&&!i._state.action:E)?i.unregister(n):$(n,!1)}},[n,i,l,a]),be.useEffect(()=>{i._setDisabledField({disabled:r,name:n})},[r,n,i]),be.useMemo(()=>({field:_,formState:c,fieldState:m}),[_,c,m])}const Bat=e=>e.render(JTe(e));var eke=(e,t,n,r,i)=>t?{...n[e],types:{...n[e]&&n[e].types?n[e].types:{},[r]:i||!0}}:{},lf=e=>Array.isArray(e)?e:[e],KZ=()=>{let e=[];return{get observers(){return e},next:i=>{for(const a of e)a.next&&a.next(i)},subscribe:i=>(e.push(i),{unsubscribe:()=>{e=e.filter(a=>a!==i)}}),unsubscribe:()=>{e=[]}}},ic=e=>Hu(e)&&!Object.keys(e).length,RI=e=>e.type==="file",Kh=e=>typeof e=="function",g$=e=>{if(!$I)return!1;const t=e?e.ownerDocument:0;return e instanceof(t&&t.defaultView?t.defaultView.HTMLElement:HTMLElement)},Bpe=e=>e.type==="select-multiple",AI=e=>e.type==="radio",tke=e=>AI(e)||RS(e),zk=e=>g$(e)&&e.isConnected;function nke(e,t){const n=t.slice(0,-1).length;let r=0;for(;r<n;)e=La(e)?r++:e[t[r++]];return e}function rke(e){for(const t in e)if(e.hasOwnProperty(t)&&!La(e[t]))return!1;return!0}function Vu(e,t){const n=Array.isArray(t)?t:hw(t)?[t]:CI(t),r=n.length===1?e:nke(e,n),i=n.length-1,a=n[i];return r&&delete r[a],i!==0&&(Hu(r)&&ic(r)||Array.isArray(r)&&rke(r))&&Vu(e,n.slice(0,-1)),e}var Qpe=e=>{for(const t in e)if(Kh(e[t]))return!0;return!1};function v$(e,t={}){const n=Array.isArray(e);if(Hu(e)||n)for(const r in e)Array.isArray(e[r])||Hu(e[r])&&!Qpe(e[r])?(t[r]=Array.isArray(e[r])?[]:{},v$(e[r],t[r])):ac(e[r])||(t[r]=!0);return t}function zpe(e,t,n){const r=Array.isArray(e);if(Hu(e)||r)for(const i in e)Array.isArray(e[i])||Hu(e[i])&&!Qpe(e[i])?La(t)||uM(n[i])?n[i]=Array.isArray(e[i])?v$(e[i],[]):{...v$(e[i])}:zpe(e[i],ac(t)?{}:t[i],n[i]):n[i]=!Zm(e[i],t[i]);return n}var a2=(e,t)=>zpe(e,t,v$(t));const WZ={value:!1,isValid:!1},JZ={value:!0,isValid:!0};var qpe=e=>{if(Array.isArray(e)){if(e.length>1){const t=e.filter(n=>n&&n.checked&&!n.disabled).map(n=>n.value);return{value:t,isValid:!!t.length}}return e[0].checked&&!e[0].disabled?e[0].attributes&&!La(e[0].attributes.value)?La(e[0].value)||e[0].value===""?JZ:{value:e[0].value,isValid:!0}:JZ:WZ}return WZ},Vpe=(e,{valueAsNumber:t,valueAsDate:n,setValueAs:r})=>La(e)?e:t?e===""?NaN:e&&+e:n&&b0(e)?new Date(e):r?r(e):e;const eH={isValid:!1,value:null};var Upe=e=>Array.isArray(e)?e.reduce((t,n)=>n&&n.checked&&!n.disabled?{isValid:!0,value:n.value}:t,eH):eH;function tH(e){const t=e.ref;return RI(t)?t.files:AI(t)?Upe(e.refs).value:Bpe(t)?[...t.selectedOptions].map(({value:n})=>n):RS(t)?qpe(e.refs).value:Vpe(La(t.value)?e.ref.value:t.value,e)}var ike=(e,t,n,r)=>{const i={};for(const a of e){const u=Ut(t,a);u&&na(i,a,u._f)}return{criteriaMode:n,names:[...e],fields:i,shouldUseNativeValidation:r}},y$=e=>e instanceof RegExp,u2=e=>La(e)?e:y$(e)?e.source:Hu(e)?y$(e.value)?e.value.source:e.value:e,HO=e=>({isOnSubmit:!e||e===Hd.onSubmit,isOnBlur:e===Hd.onBlur,isOnChange:e===Hd.onChange,isOnAll:e===Hd.all,isOnTouch:e===Hd.onTouched});const nH="AsyncFunction";var ake=e=>!!e&&!!e.validate&&!!(Kh(e.validate)&&e.validate.constructor.name===nH||Hu(e.validate)&&Object.values(e.validate).find(t=>t.constructor.name===nH)),uke=e=>e.mount&&(e.required||e.min||e.max||e.maxLength||e.minLength||e.pattern||e.validate),lM=(e,t,n)=>!n&&(t.watchAll||t.watch.has(e)||[...t.watch].some(r=>e.startsWith(r)&&/^\.\w+/.test(e.slice(r.length))));const d3=(e,t,n,r)=>{for(const i of n||Object.keys(e)){const a=Ut(e,i);if(a){const{_f:u,...l}=a;if(u){if(u.refs&&u.refs[0]&&t(u.refs[0],i)&&!r)return!0;if(u.ref&&t(u.ref,u.name)&&!r)return!0;if(d3(l,t))break}else if(Hu(l)&&d3(l,t))break}}};function rH(e,t,n){const r=Ut(e,n);if(r||hw(n))return{error:r,name:n};const i=n.split(".");for(;i.length;){const a=i.join("."),u=Ut(t,a),l=Ut(e,a);if(u&&!Array.isArray(u)&&n!==a)return{name:n};if(l&&l.type)return{name:a,error:l};if(l&&l.root&&l.root.type)return{name:`${a}.root`,error:l.root};i.pop()}return{name:n}}var lke=(e,t,n,r)=>{n(e);const{name:i,...a}=e;return ic(a)||Object.keys(a).length>=Object.keys(t).length||Object.keys(a).find(u=>t[u]===(!r||Hd.all))},oke=(e,t,n)=>!e||!t||e===t||lf(e).some(r=>r&&(n?r===t:r.startsWith(t)||t.startsWith(r))),ske=(e,t,n,r,i)=>i.isOnAll?!1:!n&&i.isOnTouch?!(t||e):(n?r.isOnBlur:i.isOnBlur)?!e:(n?r.isOnChange:i.isOnChange)?e:!0,cke=(e,t)=>!pw(Ut(e,t)).length&&Vu(e,t),jpe=(e,t,n)=>{const r=lf(Ut(e,n));return na(r,"root",t[n]),na(e,n,r),e},fD=e=>b0(e);function iH(e,t,n="validate"){if(fD(e)||Array.isArray(e)&&e.every(fD)||zf(e)&&!e)return{type:n,message:fD(e)?e:"",ref:t}}var _O=e=>Hu(e)&&!y$(e)?e:{value:e,message:""},oM=async(e,t,n,r,i,a)=>{const{ref:u,refs:l,required:o,maxLength:s,minLength:c,min:d,max:h,pattern:m,validate:y,name:v,valueAsNumber:S,mount:_}=e._f,E=Ut(n,v);if(!_||t.has(v))return{};const $=l?l[0]:u,k=V=>{i&&$.reportValidity&&($.setCustomValidity(zf(V)?"":V||""),$.reportValidity())},A={},w=AI(u),R=RS(u),F=w||R,I=(S||RI(u))&&La(u.value)&&La(E)||g$(u)&&u.value===""||E===""||Array.isArray(E)&&!E.length,U=eke.bind(null,v,r,A),q=(V,N,j,W=Am.maxLength,K=Am.minLength)=>{const J=V?N:j;A[v]={type:V?W:K,message:J,ref:u,...U(V?W:K,J)}};if(a?!Array.isArray(E)||!E.length:o&&(!F&&(I||ac(E))||zf(E)&&!E||R&&!qpe(l).isValid||w&&!Upe(l).isValid)){const{value:V,message:N}=fD(o)?{value:!!o,message:o}:_O(o);if(V&&(A[v]={type:Am.required,message:N,ref:$,...U(Am.required,N)},!r))return k(N),A}if(!I&&(!ac(d)||!ac(h))){let V,N;const j=_O(h),W=_O(d);if(!ac(E)&&!isNaN(E)){const K=u.valueAsNumber||E&&+E;ac(j.value)||(V=K>j.value),ac(W.value)||(N=K<W.value)}else{const K=u.valueAsDate||new Date(E),J=M=>new Date(new Date().toDateString()+" "+M),te=u.type=="time",H=u.type=="week";b0(j.value)&&E&&(V=te?J(E)>J(j.value):H?E>j.value:K>new Date(j.value)),b0(W.value)&&E&&(N=te?J(E)<J(W.value):H?E<W.value:K<new Date(W.value))}if((V||N)&&(q(!!V,j.message,W.message,Am.max,Am.min),!r))return k(A[v].message),A}if((s||c)&&!I&&(b0(E)||a&&Array.isArray(E))){const V=_O(s),N=_O(c),j=!ac(V.value)&&E.length>+V.value,W=!ac(N.value)&&E.length<+N.value;if((j||W)&&(q(j,V.message,N.message),!r))return k(A[v].message),A}if(m&&!I&&b0(E)){const{value:V,message:N}=_O(m);if(y$(V)&&!E.match(V)&&(A[v]={type:Am.pattern,message:N,ref:u,...U(Am.pattern,N)},!r))return k(N),A}if(y){if(Kh(y)){const V=await y(E,n),N=iH(V,$);if(N&&(A[v]={...N,...U(Am.validate,N.message)},!r))return k(N.message),A}else if(Hu(y)){let V={};for(const N in y){if(!ic(V)&&!r)break;const j=iH(await y[N](E,n),$,N);j&&(V={...j,...U(N,j.message)},k(j.message),r&&(A[v]=V))}if(!ic(V)&&(A[v]={ref:$,...V},!r))return A}}return k(!0),A};const fke={mode:Hd.onSubmit,reValidateMode:Hd.onChange,shouldFocusError:!0};function dke(e={}){let t={...fke,...e},n={submitCount:0,isDirty:!1,isReady:!1,isLoading:Kh(t.defaultValues),isValidating:!1,isSubmitted:!1,isSubmitting:!1,isSubmitSuccessful:!1,isValid:!1,touchedFields:{},dirtyFields:{},validatingFields:{},errors:t.errors||{},disabled:t.disabled||!1},r={},i=Hu(t.defaultValues)||Hu(t.values)?Na(t.defaultValues||t.values)||{}:{},a=t.shouldUnregister?{}:Na(i),u={action:!1,mount:!1,watch:!1},l={mount:new Set,disabled:new Set,unMount:new Set,array:new Set,watch:new Set},o,s=0;const c={isDirty:!1,dirtyFields:!1,validatingFields:!1,touchedFields:!1,isValidating:!1,isValid:!1,errors:!1};let d={...c};const h={array:KZ(),state:KZ()},m=t.criteriaMode===Hd.all,y=Oe=>$e=>{clearTimeout(s),s=setTimeout(Oe,$e)},v=async Oe=>{if(!t.disabled&&(c.isValid||d.isValid||Oe)){const $e=t.resolver?ic((await R()).errors):await I(r,!0);$e!==n.isValid&&h.state.next({isValid:$e})}},S=(Oe,$e)=>{!t.disabled&&(c.isValidating||c.validatingFields||d.isValidating||d.validatingFields)&&((Oe||Array.from(l.mount)).forEach(Ae=>{Ae&&($e?na(n.validatingFields,Ae,$e):Vu(n.validatingFields,Ae))}),h.state.next({validatingFields:n.validatingFields,isValidating:!ic(n.validatingFields)}))},_=(Oe,$e=[],Ae,He,je=!0,Ye=!0)=>{if(He&&Ae&&!t.disabled){if(u.action=!0,Ye&&Array.isArray(Ut(r,Oe))){const Ne=Ae(Ut(r,Oe),He.argA,He.argB);je&&na(r,Oe,Ne)}if(Ye&&Array.isArray(Ut(n.errors,Oe))){const Ne=Ae(Ut(n.errors,Oe),He.argA,He.argB);je&&na(n.errors,Oe,Ne),cke(n.errors,Oe)}if((c.touchedFields||d.touchedFields)&&Ye&&Array.isArray(Ut(n.touchedFields,Oe))){const Ne=Ae(Ut(n.touchedFields,Oe),He.argA,He.argB);je&&na(n.touchedFields,Oe,Ne)}(c.dirtyFields||d.dirtyFields)&&(n.dirtyFields=a2(i,a)),h.state.next({name:Oe,isDirty:q(Oe,$e),dirtyFields:n.dirtyFields,errors:n.errors,isValid:n.isValid})}else na(a,Oe,$e)},E=(Oe,$e)=>{na(n.errors,Oe,$e),h.state.next({errors:n.errors})},$=Oe=>{n.errors=Oe,h.state.next({errors:n.errors,isValid:!1})},k=(Oe,$e,Ae,He)=>{const je=Ut(r,Oe);if(je){const Ye=Ut(a,Oe,La(Ae)?Ut(i,Oe):Ae);La(Ye)||He&&He.defaultChecked||$e?na(a,Oe,$e?Ye:tH(je._f)):j(Oe,Ye),u.mount&&v()}},A=(Oe,$e,Ae,He,je)=>{let Ye=!1,Ne=!1;const ze={name:Oe};if(!t.disabled){if(!Ae||He){(c.isDirty||d.isDirty)&&(Ne=n.isDirty,n.isDirty=ze.isDirty=q(),Ye=Ne!==ze.isDirty);const at=Zm(Ut(i,Oe),$e);Ne=!!Ut(n.dirtyFields,Oe),at?Vu(n.dirtyFields,Oe):na(n.dirtyFields,Oe,!0),ze.dirtyFields=n.dirtyFields,Ye=Ye||(c.dirtyFields||d.dirtyFields)&&Ne!==!at}if(Ae){const at=Ut(n.touchedFields,Oe);at||(na(n.touchedFields,Oe,Ae),ze.touchedFields=n.touchedFields,Ye=Ye||(c.touchedFields||d.touchedFields)&&at!==Ae)}Ye&&je&&h.state.next(ze)}return Ye?ze:{}},w=(Oe,$e,Ae,He)=>{const je=Ut(n.errors,Oe),Ye=(c.isValid||d.isValid)&&zf($e)&&n.isValid!==$e;if(t.delayError&&Ae?(o=y(()=>E(Oe,Ae)),o(t.delayError)):(clearTimeout(s),o=null,Ae?na(n.errors,Oe,Ae):Vu(n.errors,Oe)),(Ae?!Zm(je,Ae):je)||!ic(He)||Ye){const Ne={...He,...Ye&&zf($e)?{isValid:$e}:{},errors:n.errors,name:Oe};n={...n,...Ne},h.state.next(Ne)}},R=async Oe=>{S(Oe,!0);const $e=await t.resolver(a,t.context,ike(Oe||l.mount,r,t.criteriaMode,t.shouldUseNativeValidation));return S(Oe),$e},F=async Oe=>{const{errors:$e}=await R(Oe);if(Oe)for(const Ae of Oe){const He=Ut($e,Ae);He?na(n.errors,Ae,He):Vu(n.errors,Ae)}else n.errors=$e;return $e},I=async(Oe,$e,Ae={valid:!0})=>{for(const He in Oe){const je=Oe[He];if(je){const{_f:Ye,...Ne}=je;if(Ye){const ze=l.array.has(Ye.name),at=je._f&&ake(je._f);at&&c.validatingFields&&S([He],!0);const mt=await oM(je,l.disabled,a,m,t.shouldUseNativeValidation&&!$e,ze);if(at&&c.validatingFields&&S([He]),mt[Ye.name]&&(Ae.valid=!1,$e))break;!$e&&(Ut(mt,Ye.name)?ze?jpe(n.errors,mt,Ye.name):na(n.errors,Ye.name,mt[Ye.name]):Vu(n.errors,Ye.name))}!ic(Ne)&&await I(Ne,$e,Ae)}}return Ae.valid},U=()=>{for(const Oe of l.unMount){const $e=Ut(r,Oe);$e&&($e._f.refs?$e._f.refs.every(Ae=>!zk(Ae)):!zk($e._f.ref))&&de(Oe)}l.unMount=new Set},q=(Oe,$e)=>!t.disabled&&(Oe&&$e&&na(a,Oe,$e),!Zm(M(),i)),V=(Oe,$e,Ae)=>Lpe(Oe,l,{...u.mount?a:La($e)?i:b0(Oe)?{[Oe]:$e}:$e},Ae,$e),N=Oe=>pw(Ut(u.mount?a:i,Oe,t.shouldUnregister?Ut(i,Oe,[]):[])),j=(Oe,$e,Ae={})=>{const He=Ut(r,Oe);let je=$e;if(He){const Ye=He._f;Ye&&(!Ye.disabled&&na(a,Oe,Vpe($e,Ye)),je=g$(Ye.ref)&&ac($e)?"":$e,Bpe(Ye.ref)?[...Ye.ref.options].forEach(Ne=>Ne.selected=je.includes(Ne.value)):Ye.refs?RS(Ye.ref)?Ye.refs.forEach(Ne=>{(!Ne.defaultChecked||!Ne.disabled)&&(Array.isArray(je)?Ne.checked=!!je.find(ze=>ze===Ne.value):Ne.checked=je===Ne.value||!!je)}):Ye.refs.forEach(Ne=>Ne.checked=Ne.value===je):RI(Ye.ref)?Ye.ref.value="":(Ye.ref.value=je,Ye.ref.type||h.state.next({name:Oe,values:Na(a)})))}(Ae.shouldDirty||Ae.shouldTouch)&&A(Oe,je,Ae.shouldTouch,Ae.shouldDirty,!0),Ae.shouldValidate&&H(Oe)},W=(Oe,$e,Ae)=>{for(const He in $e){if(!$e.hasOwnProperty(He))return;const je=$e[He],Ye=Oe+"."+He,Ne=Ut(r,Ye);(l.array.has(Oe)||Hu(je)||Ne&&!Ne._f)&&!Bb(je)?W(Ye,je,Ae):j(Ye,je,Ae)}},K=(Oe,$e,Ae={})=>{const He=Ut(r,Oe),je=l.array.has(Oe),Ye=Na($e);na(a,Oe,Ye),je?(h.array.next({name:Oe,values:Na(a)}),(c.isDirty||c.dirtyFields||d.isDirty||d.dirtyFields)&&Ae.shouldDirty&&h.state.next({name:Oe,dirtyFields:a2(i,a),isDirty:q(Oe,Ye)})):He&&!He._f&&!ac(Ye)?W(Oe,Ye,Ae):j(Oe,Ye,Ae),lM(Oe,l)&&h.state.next({...n,name:Oe}),h.state.next({name:u.mount?Oe:void 0,values:Na(a)})},J=async Oe=>{u.mount=!0;const $e=Oe.target;let Ae=$e.name,He=!0;const je=Ut(r,Ae),Ye=at=>{He=Number.isNaN(at)||Bb(at)&&isNaN(at.getTime())||Zm(at,Ut(a,Ae,at))},Ne=HO(t.mode),ze=HO(t.reValidateMode);if(je){let at,mt;const Mt=$e.type?tH(je._f):Fpe(Oe),bt=Oe.type===m$.BLUR||Oe.type===m$.FOCUS_OUT,zt=!uke(je._f)&&!t.resolver&&!Ut(n.errors,Ae)&&!je._f.deps||ske(bt,Ut(n.touchedFields,Ae),n.isSubmitted,ze,Ne),qt=lM(Ae,l,bt);na(a,Ae,Mt),bt?(!$e||!$e.readOnly)&&(je._f.onBlur&&je._f.onBlur(Oe),o&&o(0)):je._f.onChange&&je._f.onChange(Oe);const Xt=A(Ae,Mt,bt),nn=!ic(Xt)||qt;if(!bt&&h.state.next({name:Ae,type:Oe.type,values:Na(a)}),zt)return(c.isValid||d.isValid)&&(t.mode==="onBlur"?bt&&v():bt||v()),nn&&h.state.next({name:Ae,...qt?{}:Xt});if(!bt&&qt&&h.state.next({...n}),t.resolver){const{errors:Fe}=await R([Ae]);if(Ye(Mt),He){const ht=rH(n.errors,r,Ae),kt=rH(Fe,r,ht.name||Ae);at=kt.error,Ae=kt.name,mt=ic(Fe)}}else S([Ae],!0),at=(await oM(je,l.disabled,a,m,t.shouldUseNativeValidation))[Ae],S([Ae]),Ye(Mt),He&&(at?mt=!1:(c.isValid||d.isValid)&&(mt=await I(r,!0)));He&&(je._f.deps&&H(je._f.deps),w(Ae,mt,at,Xt))}},te=(Oe,$e)=>{if(Ut(n.errors,$e)&&Oe.focus)return Oe.focus(),1},H=async(Oe,$e={})=>{let Ae,He;const je=lf(Oe);if(t.resolver){const Ye=await F(La(Oe)?Oe:je);Ae=ic(Ye),He=Oe?!je.some(Ne=>Ut(Ye,Ne)):Ae}else Oe?(He=(await Promise.all(je.map(async Ye=>{const Ne=Ut(r,Ye);return await I(Ne&&Ne._f?{[Ye]:Ne}:Ne)}))).every(Boolean),!(!He&&!n.isValid)&&v()):He=Ae=await I(r);return h.state.next({...!b0(Oe)||(c.isValid||d.isValid)&&Ae!==n.isValid?{}:{name:Oe},...t.resolver||!Oe?{isValid:Ae}:{},errors:n.errors}),$e.shouldFocus&&!He&&d3(r,te,Oe?je:l.mount),He},M=Oe=>{const $e={...u.mount?a:i};return La(Oe)?$e:b0(Oe)?Ut($e,Oe):Oe.map(Ae=>Ut($e,Ae))},G=(Oe,$e)=>({invalid:!!Ut(($e||n).errors,Oe),isDirty:!!Ut(($e||n).dirtyFields,Oe),error:Ut(($e||n).errors,Oe),isValidating:!!Ut(n.validatingFields,Oe),isTouched:!!Ut(($e||n).touchedFields,Oe)}),X=Oe=>{Oe&&lf(Oe).forEach($e=>Vu(n.errors,$e)),h.state.next({errors:Oe?n.errors:{}})},B=(Oe,$e,Ae)=>{const He=(Ut(r,Oe,{_f:{}})._f||{}).ref,je=Ut(n.errors,Oe)||{},{ref:Ye,message:Ne,type:ze,...at}=je;na(n.errors,Oe,{...at,...$e,ref:He}),h.state.next({name:Oe,errors:n.errors,isValid:!1}),Ae&&Ae.shouldFocus&&He&&He.focus&&He.focus()},ne=(Oe,$e)=>Kh(Oe)?h.state.subscribe({next:Ae=>"values"in Ae&&Oe(V(void 0,$e),Ae)}):V(Oe,$e,!0),oe=Oe=>h.state.subscribe({next:$e=>{oke(Oe.name,$e.name,Oe.exact)&&lke($e,Oe.formState||c,Re,Oe.reRenderRoot)&&Oe.callback({values:{...a},...n,...$e,defaultValues:i})}}).unsubscribe,ie=Oe=>(u.mount=!0,d={...d,...Oe.formState},oe({...Oe,formState:d})),de=(Oe,$e={})=>{for(const Ae of Oe?lf(Oe):l.mount)l.mount.delete(Ae),l.array.delete(Ae),$e.keepValue||(Vu(r,Ae),Vu(a,Ae)),!$e.keepError&&Vu(n.errors,Ae),!$e.keepDirty&&Vu(n.dirtyFields,Ae),!$e.keepTouched&&Vu(n.touchedFields,Ae),!$e.keepIsValidating&&Vu(n.validatingFields,Ae),!t.shouldUnregister&&!$e.keepDefaultValue&&Vu(i,Ae);h.state.next({values:Na(a)}),h.state.next({...n,...$e.keepDirty?{isDirty:q()}:{}}),!$e.keepIsValid&&v()},ge=({disabled:Oe,name:$e})=>{(zf(Oe)&&u.mount||Oe||l.disabled.has($e))&&(Oe?l.disabled.add($e):l.disabled.delete($e))},ye=(Oe,$e={})=>{let Ae=Ut(r,Oe);const He=zf($e.disabled)||zf(t.disabled);return na(r,Oe,{...Ae||{},_f:{...Ae&&Ae._f?Ae._f:{ref:{name:Oe}},name:Oe,mount:!0,...$e}}),l.mount.add(Oe),Ae?ge({disabled:zf($e.disabled)?$e.disabled:t.disabled,name:Oe}):k(Oe,!0,$e.value),{...He?{disabled:$e.disabled||t.disabled}:{},...t.progressive?{required:!!$e.required,min:u2($e.min),max:u2($e.max),minLength:u2($e.minLength),maxLength:u2($e.maxLength),pattern:u2($e.pattern)}:{},name:Oe,onChange:J,onBlur:J,ref:je=>{if(je){ye(Oe,$e),Ae=Ut(r,Oe);const Ye=La(je.value)&&je.querySelectorAll&&je.querySelectorAll("input,select,textarea")[0]||je,Ne=tke(Ye),ze=Ae._f.refs||[];if(Ne?ze.find(at=>at===Ye):Ye===Ae._f.ref)return;na(r,Oe,{_f:{...Ae._f,...Ne?{refs:[...ze.filter(zk),Ye,...Array.isArray(Ut(i,Oe))?[{}]:[]],ref:{type:Ye.type,name:Oe}}:{ref:Ye}}}),k(Oe,!1,void 0,Ye)}else Ae=Ut(r,Oe,{}),Ae._f&&(Ae._f.mount=!1),(t.shouldUnregister||$e.shouldUnregister)&&!(Npe(l.array,Oe)&&u.action)&&l.unMount.add(Oe)}}},me=()=>t.shouldFocusError&&d3(r,te,l.mount),Ee=Oe=>{zf(Oe)&&(h.state.next({disabled:Oe}),d3(r,($e,Ae)=>{const He=Ut(r,Ae);He&&($e.disabled=He._f.disabled||Oe,Array.isArray(He._f.refs)&&He._f.refs.forEach(je=>{je.disabled=He._f.disabled||Oe}))},0,!1))},se=(Oe,$e)=>async Ae=>{let He;Ae&&(Ae.preventDefault&&Ae.preventDefault(),Ae.persist&&Ae.persist());let je=Na(a);if(h.state.next({isSubmitting:!0}),t.resolver){const{errors:Ye,values:Ne}=await R();n.errors=Ye,je=Na(Ne)}else await I(r);if(l.disabled.size)for(const Ye of l.disabled)Vu(je,Ye);if(Vu(n.errors,"root"),ic(n.errors)){h.state.next({errors:{}});try{await Oe(je,Ae)}catch(Ye){He=Ye}}else $e&&await $e({...n.errors},Ae),me(),setTimeout(me);if(h.state.next({isSubmitted:!0,isSubmitting:!1,isSubmitSuccessful:ic(n.errors)&&!He,submitCount:n.submitCount+1,errors:n.errors}),He)throw He},pe=(Oe,$e={})=>{Ut(r,Oe)&&(La($e.defaultValue)?K(Oe,Na(Ut(i,Oe))):(K(Oe,$e.defaultValue),na(i,Oe,Na($e.defaultValue))),$e.keepTouched||Vu(n.touchedFields,Oe),$e.keepDirty||(Vu(n.dirtyFields,Oe),n.isDirty=$e.defaultValue?q(Oe,Na(Ut(i,Oe))):q()),$e.keepError||(Vu(n.errors,Oe),c.isValid&&v()),h.state.next({...n}))},ve=(Oe,$e={})=>{const Ae=Oe?Na(Oe):i,He=Na(Ae),je=ic(Oe),Ye=je?i:He;if($e.keepDefaultValues||(i=Ae),!$e.keepValues){if($e.keepDirtyValues){const Ne=new Set([...l.mount,...Object.keys(a2(i,a))]);for(const ze of Array.from(Ne))Ut(n.dirtyFields,ze)?na(Ye,ze,Ut(a,ze)):K(ze,Ut(Ye,ze))}else{if($I&&La(Oe))for(const Ne of l.mount){const ze=Ut(r,Ne);if(ze&&ze._f){const at=Array.isArray(ze._f.refs)?ze._f.refs[0]:ze._f.ref;if(g$(at)){const mt=at.closest("form");if(mt){mt.reset();break}}}}if($e.keepFieldsRef)for(const Ne of l.mount)K(Ne,Ut(Ye,Ne));else r={}}a=t.shouldUnregister?$e.keepDefaultValues?Na(i):{}:Na(Ye),h.array.next({values:{...Ye}}),h.state.next({values:{...Ye}})}l={mount:$e.keepDirtyValues?l.mount:new Set,unMount:new Set,array:new Set,disabled:new Set,watch:new Set,watchAll:!1,focus:""},u.mount=!c.isValid||!!$e.keepIsValid||!!$e.keepDirtyValues,u.watch=!!t.shouldUnregister,h.state.next({submitCount:$e.keepSubmitCount?n.submitCount:0,isDirty:je?!1:$e.keepDirty?n.isDirty:!!($e.keepDefaultValues&&!Zm(Oe,i)),isSubmitted:$e.keepIsSubmitted?n.isSubmitted:!1,dirtyFields:je?{}:$e.keepDirtyValues?$e.keepDefaultValues&&a?a2(i,a):n.dirtyFields:$e.keepDefaultValues&&Oe?a2(i,Oe):$e.keepDirty?n.dirtyFields:{},touchedFields:$e.keepTouched?n.touchedFields:{},errors:$e.keepErrors?n.errors:{},isSubmitSuccessful:$e.keepIsSubmitSuccessful?n.isSubmitSuccessful:!1,isSubmitting:!1,defaultValues:i})},we=(Oe,$e)=>ve(Kh(Oe)?Oe(a):Oe,$e),Me=(Oe,$e={})=>{const Ae=Ut(r,Oe),He=Ae&&Ae._f;if(He){const je=He.refs?He.refs[0]:He.ref;je.focus&&(je.focus(),$e.shouldSelect&&Kh(je.select)&&je.select())}},Re=Oe=>{n={...n,...Oe}},tt={control:{register:ye,unregister:de,getFieldState:G,handleSubmit:se,setError:B,_subscribe:oe,_runSchema:R,_focusError:me,_getWatch:V,_getDirty:q,_setValid:v,_setFieldArray:_,_setDisabledField:ge,_setErrors:$,_getFieldArray:N,_reset:ve,_resetDefaultValues:()=>Kh(t.defaultValues)&&t.defaultValues().then(Oe=>{we(Oe,t.resetOptions),h.state.next({isLoading:!1})}),_removeUnmounted:U,_disableForm:Ee,_subjects:h,_proxyFormState:c,get _fields(){return r},get _formValues(){return a},get _state(){return u},set _state(Oe){u=Oe},get _defaultValues(){return i},get _names(){return l},set _names(Oe){l=Oe},get _formState(){return n},get _options(){return t},set _options(Oe){t={...t,...Oe}}},subscribe:ie,trigger:H,register:ye,handleSubmit:se,watch:ne,setValue:K,getValues:M,reset:we,resetField:pe,clearErrors:X,unregister:de,setError:B,setFocus:Me,getFieldState:G};return{...tt,formControl:tt}}var pv=()=>{if(typeof crypto<"u"&&crypto.randomUUID)return crypto.randomUUID();const e=typeof performance>"u"?Date.now():performance.now()*1e3;return"xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace(/[xy]/g,t=>{const n=(Math.random()*16+e)%16|0;return(t=="x"?n:n&3|8).toString(16)})},qk=(e,t,n={})=>n.shouldFocus||La(n.shouldFocus)?n.focusName||`${e}.${La(n.focusIndex)?t:n.focusIndex}.`:"",Vk=(e,t)=>[...e,...lf(t)],Uk=e=>Array.isArray(e)?e.map(()=>{}):void 0;function jk(e,t,n){return[...e.slice(0,t),...lf(n),...e.slice(t)]}var Zk=(e,t,n)=>Array.isArray(e)?(La(e[n])&&(e[n]=void 0),e.splice(n,0,e.splice(t,1)[0]),e):[],Hk=(e,t)=>[...lf(t),...lf(e)];function hke(e,t){let n=0;const r=[...e];for(const i of t)r.splice(i-n,1),n++;return pw(r).length?r:[]}var Xk=(e,t)=>La(t)?[]:hke(e,lf(t).sort((n,r)=>n-r)),Yk=(e,t,n)=>{[e[t],e[n]]=[e[n],e[t]]},aH=(e,t,n)=>(e[t]=n,e);function Qat(e){const t=mw(),{control:n=t.control,name:r,keyName:i="id",shouldUnregister:a,rules:u}=e,[l,o]=be.useState(n._getFieldArray(r)),s=be.useRef(n._getFieldArray(r).map(pv)),c=be.useRef(l),d=be.useRef(!1);c.current=l,n._names.array.add(r),be.useMemo(()=>u&&n.register(r,u),[n,u,r]),gw(()=>n._subjects.array.subscribe({next:({values:A,name:w})=>{if(w===r||!w){const R=Ut(A,r);Array.isArray(R)&&(o(R),s.current=R.map(pv))}}}).unsubscribe,[n,r]);const h=be.useCallback(A=>{d.current=!0,n._setFieldArray(r,A)},[n,r]),m=(A,w)=>{const R=lf(Na(A)),F=Vk(n._getFieldArray(r),R);n._names.focus=qk(r,F.length-1,w),s.current=Vk(s.current,R.map(pv)),h(F),o(F),n._setFieldArray(r,F,Vk,{argA:Uk(A)})},y=(A,w)=>{const R=lf(Na(A)),F=Hk(n._getFieldArray(r),R);n._names.focus=qk(r,0,w),s.current=Hk(s.current,R.map(pv)),h(F),o(F),n._setFieldArray(r,F,Hk,{argA:Uk(A)})},v=A=>{const w=Xk(n._getFieldArray(r),A);s.current=Xk(s.current,A),h(w),o(w),!Array.isArray(Ut(n._fields,r))&&na(n._fields,r,void 0),n._setFieldArray(r,w,Xk,{argA:A})},S=(A,w,R)=>{const F=lf(Na(w)),I=jk(n._getFieldArray(r),A,F);n._names.focus=qk(r,A,R),s.current=jk(s.current,A,F.map(pv)),h(I),o(I),n._setFieldArray(r,I,jk,{argA:A,argB:Uk(w)})},_=(A,w)=>{const R=n._getFieldArray(r);Yk(R,A,w),Yk(s.current,A,w),h(R),o(R),n._setFieldArray(r,R,Yk,{argA:A,argB:w},!1)},E=(A,w)=>{const R=n._getFieldArray(r);Zk(R,A,w),Zk(s.current,A,w),h(R),o(R),n._setFieldArray(r,R,Zk,{argA:A,argB:w},!1)},$=(A,w)=>{const R=Na(w),F=aH(n._getFieldArray(r),A,R);s.current=[...F].map((I,U)=>!I||U===A?pv():s.current[U]),h(F),o([...F]),n._setFieldArray(r,F,aH,{argA:A,argB:R},!0,!1)},k=A=>{const w=lf(Na(A));s.current=w.map(pv),h([...w]),o([...w]),n._setFieldArray(r,[...w],R=>R,{},!0,!1)};return be.useEffect(()=>{if(n._state.action=!1,lM(r,n._names)&&n._subjects.state.next({...n._formState}),d.current&&(!HO(n._options.mode).isOnSubmit||n._formState.isSubmitted)&&!HO(n._options.reValidateMode).isOnSubmit)if(n._options.resolver)n._runSchema([r]).then(A=>{const w=Ut(A.errors,r),R=Ut(n._formState.errors,r);(R?!w&&R.type||w&&(R.type!==w.type||R.message!==w.message):w&&w.type)&&(w?na(n._formState.errors,r,w):Vu(n._formState.errors,r),n._subjects.state.next({errors:n._formState.errors}))});else{const A=Ut(n._fields,r);A&&A._f&&!(HO(n._options.reValidateMode).isOnSubmit&&HO(n._options.mode).isOnSubmit)&&oM(A,n._names.disabled,n._formValues,n._options.criteriaMode===Hd.all,n._options.shouldUseNativeValidation,!0).then(w=>!ic(w)&&n._subjects.state.next({errors:jpe(n._formState.errors,w,r)}))}n._subjects.state.next({name:r,values:Na(n._formValues)}),n._names.focus&&d3(n._fields,(A,w)=>{if(n._names.focus&&w.startsWith(n._names.focus)&&A.focus)return A.focus(),1}),n._names.focus="",n._setValid(),d.current=!1},[l,r,n]),be.useEffect(()=>(!Ut(n._formValues,r)&&n._setFieldArray(r),()=>{const A=(w,R)=>{const F=Ut(n._fields,w);F&&F._f&&(F._f.mount=R)};n._options.shouldUnregister||a?n.unregister(r):A(r,!1)}),[r,n,i,a]),{swap:be.useCallback(_,[h,r,n]),move:be.useCallback(E,[h,r,n]),prepend:be.useCallback(y,[h,r,n]),append:be.useCallback(m,[h,r,n]),remove:be.useCallback(v,[h,r,n]),insert:be.useCallback(S,[h,r,n]),update:be.useCallback($,[h,r,n]),replace:be.useCallback(k,[h,r,n]),fields:be.useMemo(()=>l.map((A,w)=>({...A,[i]:s.current[w]||pv()})),[l,i])}}function zat(e={}){const t=be.useRef(void 0),n=be.useRef(void 0),[r,i]=be.useState({isDirty:!1,isValidating:!1,isLoading:Kh(e.defaultValues),isSubmitted:!1,isSubmitting:!1,isSubmitSuccessful:!1,isValid:!1,submitCount:0,dirtyFields:{},touchedFields:{},validatingFields:{},errors:e.errors||{},disabled:e.disabled||!1,isReady:!1,defaultValues:Kh(e.defaultValues)?void 0:e.defaultValues});if(!t.current)if(e.formControl)t.current={...e.formControl,formState:r},e.defaultValues&&!Kh(e.defaultValues)&&e.formControl.reset(e.defaultValues,e.resetOptions);else{const{formControl:u,...l}=dke(e);t.current={...l,formState:r}}const a=t.current.control;return a._options=e,gw(()=>{const u=a._subscribe({formState:a._proxyFormState,callback:()=>i({...a._formState}),reRenderRoot:!0});return i(l=>({...l,isReady:!0})),a._formState.isReady=!0,u},[a]),be.useEffect(()=>a._disableForm(e.disabled),[a,e.disabled]),be.useEffect(()=>{e.mode&&(a._options.mode=e.mode),e.reValidateMode&&(a._options.reValidateMode=e.reValidateMode)},[a,e.mode,e.reValidateMode]),be.useEffect(()=>{e.errors&&(a._setErrors(e.errors),a._focusError())},[a,e.errors]),be.useEffect(()=>{e.shouldUnregister&&a._subjects.state.next({values:a._getWatch()})},[a,e.shouldUnregister]),be.useEffect(()=>{if(a._proxyFormState.isDirty){const u=a._getDirty();u!==r.isDirty&&a._subjects.state.next({isDirty:u})}},[a,r.isDirty]),be.useEffect(()=>{e.values&&!Zm(e.values,n.current)?(a._reset(e.values,{keepFieldsRef:!0,...a._options.resetOptions}),n.current=e.values,i(u=>({...u}))):a._resetDefaultValues()},[a,e.values]),be.useEffect(()=>{a._state.mount||(a._setValid(),a._state.mount=!0),a._state.watch&&(a._state.watch=!1,a._subjects.state.next({...a._formState})),a._removeUnmounted()}),t.current.formState=Ipe(r,a),t.current}var Gk="Invariant Violation",uH=Object.setPrototypeOf,pke=uH===void 0?function(e,t){return e.__proto__=t,e}:uH,mke=(function(e){Wte(t,e);function t(n){n===void 0&&(n=Gk);var r=e.call(this,typeof n=="number"?Gk+": "+n+" (see https://github.com/apollographql/invariant-packages)":n)||this;return r.framesToPop=1,r.name=Gk,pke(r,t.prototype),r}return t})(Error);function Mb(e,t){if(!e)throw new mke(t)}var Zpe=["debug","log","warn","error","silent"],gke=Zpe.indexOf("log");function rE(e){return function(){if(Zpe.indexOf(e)>=gke){var t=console[e]||console.log;return t.apply(console,arguments)}}}(function(e){e.debug=rE("debug"),e.log=rE("log"),e.warn=rE("warn"),e.error=rE("error")})(Mb||(Mb={}));var Hpe="3.13.9";function Um(e){try{return e()}catch{}}const lH=Um(function(){return globalThis})||Um(function(){return window})||Um(function(){return self})||Um(function(){return global})||Um(function(){return Um.constructor("return this")()});var oH=new Map;function vke(e){var t=oH.get(e)||1;return oH.set(e,t+1),"".concat(e,":").concat(t,":").concat(Math.random().toString(36).slice(2))}function yke(e,t){var n=vke("stringifyForDisplay");return JSON.stringify(e,function(r,i){return i===void 0?n:i},t).split(JSON.stringify(n)).join("<undefined>")}function iE(e){return function(t){for(var n=[],r=1;r<arguments.length;r++)n[r-1]=arguments[r];if(typeof t=="number"){var i=t;t=Ype(i),t||(t=Gpe(i,n),n=[])}e.apply(void 0,[t].concat(n))}}Object.assign(function(t,n){for(var r=[],i=2;i<arguments.length;i++)r[i-2]=arguments[i];t||Mb(t,Ype(n,r)||Gpe(n,r))},{debug:iE(Mb.debug),log:iE(Mb.log),warn:iE(Mb.warn),error:iE(Mb.error)});var sH=Symbol.for("ApolloErrorMessageHandler_"+Hpe);function Xpe(e){if(typeof e=="string")return e;try{return yke(e,2).slice(0,1e3)}catch{return"<non-serializable>"}}function Ype(e,t){if(t===void 0&&(t=[]),!!e)return lH[sH]&&lH[sH](e,t.map(Xpe))}function Gpe(e,t){if(t===void 0&&(t=[]),!!e)return"An error occurred! For more details, see the full error text at https://go.apollo.dev/c/err#".concat(encodeURIComponent(JSON.stringify({version:Hpe,message:e,args:t.map(Xpe)})))}Um(function(){return navigator.product})=="ReactNative";var bke=typeof Symbol=="function"&&typeof Symbol.for=="function",vw=bke&&Symbol.asyncIterator;Um(function(){return window.document.createElement});Um(function(){return navigator.userAgent.indexOf("jsdom")>=0});function Kpe(e){return e!==null&&typeof e=="object"}function Oke(e){return Kpe(e)&&"payload"in e}function xke(e){var t,n=e[Symbol.asyncIterator]();return t={next:function(){return n.next()}},t[Symbol.asyncIterator]=function(){return this},t}function Ske(e){var t=null,n=null,r=!1,i=[],a=[];function u(d){if(!n){if(a.length){var h=a.shift();if(Array.isArray(h)&&h[0])return h[0]({value:d,done:!1})}i.push(d)}}function l(d){n=d;var h=a.slice();h.forEach(function(m){m[1](d)}),!t||t()}function o(){r=!0;var d=a.slice();d.forEach(function(h){h[0]({value:void 0,done:!0})}),!t||t()}t=function(){t=null,e.removeListener("data",u),e.removeListener("error",l),e.removeListener("end",o),e.removeListener("finish",o),e.removeListener("close",o)},e.on("data",u),e.on("error",l),e.on("end",o),e.on("finish",o),e.on("close",o);function s(){return new Promise(function(d,h){if(n)return h(n);if(i.length)return d({value:i.shift(),done:!1});if(r)return d({value:void 0,done:!0});a.push([d,h])})}var c={next:function(){return s()}};return vw&&(c[Symbol.asyncIterator]=function(){return this}),c}function _ke(e){var t=!1,n={next:function(){return t?Promise.resolve({value:void 0,done:!0}):(t=!0,new Promise(function(r,i){e.then(function(a){r({value:a,done:!1})}).catch(i)}))}};return vw&&(n[Symbol.asyncIterator]=function(){return this}),n}function cH(e){var t={next:function(){return e.read()}};return vw&&(t[Symbol.asyncIterator]=function(){return this}),t}function Eke(e){return!!e.body}function Dke(e){return!!e.getReader}function $ke(e){return!!(vw&&e[Symbol.asyncIterator])}function Cke(e){return!!e.stream}function wke(e){return!!e.arrayBuffer}function Rke(e){return!!e.pipe}function Ake(e){var t=e;if(Eke(e)&&(t=e.body),$ke(t))return xke(t);if(Dke(t))return cH(t.getReader());if(Cke(t))return cH(t.stream().getReader());if(wke(t))return _ke(t.arrayBuffer());if(Rke(t))return Ske(t);throw new Error("Unknown body type for responseIterator. Please pass a streamable response.")}var Tke=function(e,t,n){var r=new Error(n);throw r.name="ServerError",r.response=e,r.statusCode=e.status,r.result=t,r},kke=Symbol(),Pke=function(e){var t=a3(a3(a3([],e.graphQLErrors,!0),e.clientErrors,!0),e.protocolErrors,!0);return e.networkError&&t.push(e.networkError),t.map(function(n){return Kpe(n)&&n.message||"Error message not found."}).join(`
`)};(function(e){Wte(t,e);function t(n){var r=n.graphQLErrors,i=n.protocolErrors,a=n.clientErrors,u=n.networkError,l=n.errorMessage,o=n.extraInfo,s=e.call(this,l)||this;return s.name="ApolloError",s.graphQLErrors=r||[],s.protocolErrors=i||[],s.clientErrors=a||[],s.networkError=u||null,s.message=l||Pke(s),s.extraInfo=o,s.cause=a3(a3(a3([u],r||[],!0),i||[],!0),a||[],!0).find(function(c){return!!c})||null,s.__proto__=t.prototype,s}return t})(Error);function qat(e,t){return V_e(this,void 0,void 0,function(){var n,r,i,a,u,l,o,s,c,d,h,m,y,v,S,_,E,$,k,A,w,R,F,I;return U_e(this,function(U){switch(U.label){case 0:if(TextDecoder===void 0)throw new Error("TextDecoder must be defined in the environment: please import a polyfill.");n=new TextDecoder("utf-8"),r=(I=e.headers)===null||I===void 0?void 0:I.get("content-type"),i="boundary=",a=r!=null&&r.includes(i)?r==null?void 0:r.substring((r==null?void 0:r.indexOf(i))+i.length).replace(/['"]/g,"").replace(/\;(.*)/gm,"").trim():"-",u=`\r
--`.concat(a),l="",o=Ake(e),s=!0,U.label=1;case 1:return s?[4,o.next()]:[3,3];case 2:for(c=U.sent(),d=c.value,h=c.done,m=typeof d=="string"?d:n.decode(d),y=l.length-u.length+1,s=!h,l+=m,v=l.indexOf(u,y);v>-1;){if(S=void 0,R=[l.slice(0,v),l.slice(v+u.length)],S=R[0],l=R[1],_=S.indexOf(`\r
\r
`),E=Mke(S.slice(0,_)),$=E["content-type"],$&&$.toLowerCase().indexOf("application/json")===-1)throw new Error("Unsupported patch content type: application/json is required.");if(k=S.slice(_),k){if(A=Fke(e,k),Object.keys(A).length>1||"data"in A||"incremental"in A||"errors"in A||"payload"in A)if(Oke(A)){if(w={},"payload"in A){if(Object.keys(A).length===1&&A.payload===null)return[2];w=kb({},A.payload)}"errors"in A&&(w=kb(kb({},w),{extensions:kb(kb({},"extensions"in w?w.extensions:null),(F={},F[kke]=A.errors,F))})),t(w)}else t(A);else if(Object.keys(A).length===1&&"hasNext"in A&&!A.hasNext)return[2]}v=l.indexOf(u)}return[3,1];case 3:return[2]}})})}function Mke(e){var t={};return e.split(`
`).forEach(function(n){var r=n.indexOf(":");if(r>-1){var i=n.slice(0,r).trim().toLowerCase(),a=n.slice(r+1).trim();t[i]=a}}),t}function Fke(e,t){if(e.status>=300){var n=function(){try{return JSON.parse(t)}catch{return t}};Tke(e,n(),"Response not successful: Received status code ".concat(e.status))}try{return JSON.parse(t)}catch(i){var r=i;throw r.name="ServerParseError",r.response=e,r.statusCode=e.status,r.bodyText=t,r}}var Vat=ii(),Nke="Invariant failed";function Uat(e,t){if(!e)throw new Error(Nke)}const Ike="modulepreload",Lke=function(e){return"/"+e},fH={},jat=function(t,n,r){let i=Promise.resolve();if(n&&n.length>0){let u=function(s){return Promise.all(s.map(c=>Promise.resolve(c).then(d=>({status:"fulfilled",value:d}),d=>({status:"rejected",reason:d}))))};document.getElementsByTagName("link");const l=document.querySelector("meta[property=csp-nonce]"),o=(l==null?void 0:l.nonce)||(l==null?void 0:l.getAttribute("nonce"));i=u(n.map(s=>{if(s=Lke(s),s in fH)return;fH[s]=!0;const c=s.endsWith(".css"),d=c?'[rel="stylesheet"]':"";if(document.querySelector(`link[href="${s}"]${d}`))return;const h=document.createElement("link");if(h.rel=c?"stylesheet":Ike,c||(h.as="script"),h.crossOrigin="",h.href=s,o&&h.setAttribute("nonce",o),document.head.appendChild(h),c)return new Promise((m,y)=>{h.addEventListener("load",m),h.addEventListener("error",()=>y(new Error(`Unable to preload CSS for ${s}`)))})}))}function a(u){const l=new Event("vite:preloadError",{cancelable:!0});if(l.payload=u,window.dispatchEvent(l),!l.defaultPrevented)throw u}return i.then(u=>{for(const l of u||[])l.status==="rejected"&&a(l.reason);return t().catch(a)})};function Zat(e){return P.createElement(ARe,{flushSync:Uf.flushSync,...e})}/**
   * table-core
   *
   * Copyright (c) TanStack
   *
   * This source code is licensed under the MIT license found in the
   * LICENSE.md file in the root directory of this source tree.
   *
   * @license MIT
   */function Hat(){return{accessor:(e,t)=>typeof e=="function"?{...t,accessorFn:e}:{...t,accessorKey:e},display:e=>e,group:e=>e}}function Bv(e,t){return typeof e=="function"?e(t):e}function Yf(e,t){return n=>{t.setState(r=>({...r,[e]:Bv(n,r[e])}))}}function yw(e){return e instanceof Function}function Bke(e){return Array.isArray(e)&&e.every(t=>typeof t=="number")}function Qke(e,t){const n=[],r=i=>{i.forEach(a=>{n.push(a);const u=t(a);u!=null&&u.length&&r(u)})};return r(e),n}function Mn(e,t,n){let r=[],i;return a=>{let u;n.key&&n.debug&&(u=Date.now());const l=e(a);if(!(l.length!==r.length||l.some((c,d)=>r[d]!==c)))return i;r=l;let s;if(n.key&&n.debug&&(s=Date.now()),i=t(...l),n==null||n.onChange==null||n.onChange(i),n.key&&n.debug&&n!=null&&n.debug()){const c=Math.round((Date.now()-u)*100)/100,d=Math.round((Date.now()-s)*100)/100,h=d/16,m=(y,v)=>{for(y=String(y);y.length<v;)y=" "+y;return y};console.info(`%c⏱ ${m(d,5)} /${m(c,5)} ms`,`
            font-size: .6rem;
            font-weight: bold;
            color: hsl(${Math.max(0,Math.min(120-120*h,120))}deg 100% 31%);`,n==null?void 0:n.key)}return i}}function Fn(e,t,n,r){return{debug:()=>{var i;return(i=e==null?void 0:e.debugAll)!=null?i:e[t]},key:!1,onChange:r}}function zke(e,t,n,r){const i=()=>{var u;return(u=a.getValue())!=null?u:e.options.renderFallbackValue},a={id:`${t.id}_${n.id}`,row:t,column:n,getValue:()=>t.getValue(r),renderValue:i,getContext:Mn(()=>[e,n,t,a],(u,l,o,s)=>({table:u,column:l,row:o,cell:s,getValue:s.getValue,renderValue:s.renderValue}),Fn(e.options,"debugCells"))};return e._features.forEach(u=>{u.createCell==null||u.createCell(a,n,t,e)},{}),a}function qke(e,t,n,r){var i,a;const l={...e._getDefaultColumnDef(),...t},o=l.accessorKey;let s=(i=(a=l.id)!=null?a:o?typeof String.prototype.replaceAll=="function"?o.replaceAll(".","_"):o.replace(/\./g,"_"):void 0)!=null?i:typeof l.header=="string"?l.header:void 0,c;if(l.accessorFn?c=l.accessorFn:o&&(o.includes(".")?c=h=>{let m=h;for(const v of o.split(".")){var y;m=(y=m)==null?void 0:y[v]}return m}:c=h=>h[l.accessorKey]),!s)throw new Error;let d={id:`${String(s)}`,accessorFn:c,parent:r,depth:n,columnDef:l,columns:[],getFlatColumns:Mn(()=>[!0],()=>{var h;return[d,...(h=d.columns)==null?void 0:h.flatMap(m=>m.getFlatColumns())]},Fn(e.options,"debugColumns")),getLeafColumns:Mn(()=>[e._getOrderColumnsFn()],h=>{var m;if((m=d.columns)!=null&&m.length){let y=d.columns.flatMap(v=>v.getLeafColumns());return h(y)}return[d]},Fn(e.options,"debugColumns"))};for(const h of e._features)h.createColumn==null||h.createColumn(d,e);return d}const ms="debugHeaders";function dH(e,t,n){var r;let a={id:(r=n.id)!=null?r:t.id,column:t,index:n.index,isPlaceholder:!!n.isPlaceholder,placeholderId:n.placeholderId,depth:n.depth,subHeaders:[],colSpan:0,rowSpan:0,headerGroup:null,getLeafHeaders:()=>{const u=[],l=o=>{o.subHeaders&&o.subHeaders.length&&o.subHeaders.map(l),u.push(o)};return l(a),u},getContext:()=>({table:e,header:a,column:t})};return e._features.forEach(u=>{u.createHeader==null||u.createHeader(a,e)}),a}const Vke={createTable:e=>{e.getHeaderGroups=Mn(()=>[e.getAllColumns(),e.getVisibleLeafColumns(),e.getState().columnPinning.left,e.getState().columnPinning.right],(t,n,r,i)=>{var a,u;const l=(a=r==null?void 0:r.map(d=>n.find(h=>h.id===d)).filter(Boolean))!=null?a:[],o=(u=i==null?void 0:i.map(d=>n.find(h=>h.id===d)).filter(Boolean))!=null?u:[],s=n.filter(d=>!(r!=null&&r.includes(d.id))&&!(i!=null&&i.includes(d.id)));return aE(t,[...l,...s,...o],e)},Fn(e.options,ms)),e.getCenterHeaderGroups=Mn(()=>[e.getAllColumns(),e.getVisibleLeafColumns(),e.getState().columnPinning.left,e.getState().columnPinning.right],(t,n,r,i)=>(n=n.filter(a=>!(r!=null&&r.includes(a.id))&&!(i!=null&&i.includes(a.id))),aE(t,n,e,"center")),Fn(e.options,ms)),e.getLeftHeaderGroups=Mn(()=>[e.getAllColumns(),e.getVisibleLeafColumns(),e.getState().columnPinning.left],(t,n,r)=>{var i;const a=(i=r==null?void 0:r.map(u=>n.find(l=>l.id===u)).filter(Boolean))!=null?i:[];return aE(t,a,e,"left")},Fn(e.options,ms)),e.getRightHeaderGroups=Mn(()=>[e.getAllColumns(),e.getVisibleLeafColumns(),e.getState().columnPinning.right],(t,n,r)=>{var i;const a=(i=r==null?void 0:r.map(u=>n.find(l=>l.id===u)).filter(Boolean))!=null?i:[];return aE(t,a,e,"right")},Fn(e.options,ms)),e.getFooterGroups=Mn(()=>[e.getHeaderGroups()],t=>[...t].reverse(),Fn(e.options,ms)),e.getLeftFooterGroups=Mn(()=>[e.getLeftHeaderGroups()],t=>[...t].reverse(),Fn(e.options,ms)),e.getCenterFooterGroups=Mn(()=>[e.getCenterHeaderGroups()],t=>[...t].reverse(),Fn(e.options,ms)),e.getRightFooterGroups=Mn(()=>[e.getRightHeaderGroups()],t=>[...t].reverse(),Fn(e.options,ms)),e.getFlatHeaders=Mn(()=>[e.getHeaderGroups()],t=>t.map(n=>n.headers).flat(),Fn(e.options,ms)),e.getLeftFlatHeaders=Mn(()=>[e.getLeftHeaderGroups()],t=>t.map(n=>n.headers).flat(),Fn(e.options,ms)),e.getCenterFlatHeaders=Mn(()=>[e.getCenterHeaderGroups()],t=>t.map(n=>n.headers).flat(),Fn(e.options,ms)),e.getRightFlatHeaders=Mn(()=>[e.getRightHeaderGroups()],t=>t.map(n=>n.headers).flat(),Fn(e.options,ms)),e.getCenterLeafHeaders=Mn(()=>[e.getCenterFlatHeaders()],t=>t.filter(n=>{var r;return!((r=n.subHeaders)!=null&&r.length)}),Fn(e.options,ms)),e.getLeftLeafHeaders=Mn(()=>[e.getLeftFlatHeaders()],t=>t.filter(n=>{var r;return!((r=n.subHeaders)!=null&&r.length)}),Fn(e.options,ms)),e.getRightLeafHeaders=Mn(()=>[e.getRightFlatHeaders()],t=>t.filter(n=>{var r;return!((r=n.subHeaders)!=null&&r.length)}),Fn(e.options,ms)),e.getLeafHeaders=Mn(()=>[e.getLeftHeaderGroups(),e.getCenterHeaderGroups(),e.getRightHeaderGroups()],(t,n,r)=>{var i,a,u,l,o,s;return[...(i=(a=t[0])==null?void 0:a.headers)!=null?i:[],...(u=(l=n[0])==null?void 0:l.headers)!=null?u:[],...(o=(s=r[0])==null?void 0:s.headers)!=null?o:[]].map(c=>c.getLeafHeaders()).flat()},Fn(e.options,ms))}};function aE(e,t,n,r){var i,a;let u=0;const l=function(h,m){m===void 0&&(m=1),u=Math.max(u,m),h.filter(y=>y.getIsVisible()).forEach(y=>{var v;(v=y.columns)!=null&&v.length&&l(y.columns,m+1)},0)};l(e);let o=[];const s=(h,m)=>{const y={depth:m,id:[r,`${m}`].filter(Boolean).join("_"),headers:[]},v=[];h.forEach(S=>{const _=[...v].reverse()[0],E=S.column.depth===y.depth;let $,k=!1;if(E&&S.column.parent?$=S.column.parent:($=S.column,k=!0),_&&(_==null?void 0:_.column)===$)_.subHeaders.push(S);else{const A=dH(n,$,{id:[r,m,$.id,S==null?void 0:S.id].filter(Boolean).join("_"),isPlaceholder:k,placeholderId:k?`${v.filter(w=>w.column===$).length}`:void 0,depth:m,index:v.length});A.subHeaders.push(S),v.push(A)}y.headers.push(S),S.headerGroup=y}),o.push(y),m>0&&s(v,m-1)},c=t.map((h,m)=>dH(n,h,{depth:u,index:m}));s(c,u-1),o.reverse();const d=h=>h.filter(y=>y.column.getIsVisible()).map(y=>{let v=0,S=0,_=[0];y.subHeaders&&y.subHeaders.length?(_=[],d(y.subHeaders).forEach($=>{let{colSpan:k,rowSpan:A}=$;v+=k,_.push(A)})):v=1;const E=Math.min(..._);return S=S+E,y.colSpan=v,y.rowSpan=S,{colSpan:v,rowSpan:S}});return d((i=(a=o[0])==null?void 0:a.headers)!=null?i:[]),o}const TI=(e,t,n,r,i,a,u)=>{let l={id:t,index:r,original:n,depth:i,parentId:u,_valuesCache:{},_uniqueValuesCache:{},getValue:o=>{if(l._valuesCache.hasOwnProperty(o))return l._valuesCache[o];const s=e.getColumn(o);if(s!=null&&s.accessorFn)return l._valuesCache[o]=s.accessorFn(l.original,r),l._valuesCache[o]},getUniqueValues:o=>{if(l._uniqueValuesCache.hasOwnProperty(o))return l._uniqueValuesCache[o];const s=e.getColumn(o);if(s!=null&&s.accessorFn)return s.columnDef.getUniqueValues?(l._uniqueValuesCache[o]=s.columnDef.getUniqueValues(l.original,r),l._uniqueValuesCache[o]):(l._uniqueValuesCache[o]=[l.getValue(o)],l._uniqueValuesCache[o])},renderValue:o=>{var s;return(s=l.getValue(o))!=null?s:e.options.renderFallbackValue},subRows:[],getLeafRows:()=>Qke(l.subRows,o=>o.subRows),getParentRow:()=>l.parentId?e.getRow(l.parentId,!0):void 0,getParentRows:()=>{let o=[],s=l;for(;;){const c=s.getParentRow();if(!c)break;o.push(c),s=c}return o.reverse()},getAllCells:Mn(()=>[e.getAllLeafColumns()],o=>o.map(s=>zke(e,l,s,s.id)),Fn(e.options,"debugRows")),_getAllCellsByColumnId:Mn(()=>[l.getAllCells()],o=>o.reduce((s,c)=>(s[c.column.id]=c,s),{}),Fn(e.options,"debugRows"))};for(let o=0;o<e._features.length;o++){const s=e._features[o];s==null||s.createRow==null||s.createRow(l,e)}return l},Uke={createColumn:(e,t)=>{e._getFacetedRowModel=t.options.getFacetedRowModel&&t.options.getFacetedRowModel(t,e.id),e.getFacetedRowModel=()=>e._getFacetedRowModel?e._getFacetedRowModel():t.getPreFilteredRowModel(),e._getFacetedUniqueValues=t.options.getFacetedUniqueValues&&t.options.getFacetedUniqueValues(t,e.id),e.getFacetedUniqueValues=()=>e._getFacetedUniqueValues?e._getFacetedUniqueValues():new Map,e._getFacetedMinMaxValues=t.options.getFacetedMinMaxValues&&t.options.getFacetedMinMaxValues(t,e.id),e.getFacetedMinMaxValues=()=>{if(e._getFacetedMinMaxValues)return e._getFacetedMinMaxValues()}}},Wpe=(e,t,n)=>{var r,i;const a=n==null||(r=n.toString())==null?void 0:r.toLowerCase();return!!(!((i=e.getValue(t))==null||(i=i.toString())==null||(i=i.toLowerCase())==null)&&i.includes(a))};Wpe.autoRemove=e=>np(e);const Jpe=(e,t,n)=>{var r;return!!(!((r=e.getValue(t))==null||(r=r.toString())==null)&&r.includes(n))};Jpe.autoRemove=e=>np(e);const e0e=(e,t,n)=>{var r;return((r=e.getValue(t))==null||(r=r.toString())==null?void 0:r.toLowerCase())===(n==null?void 0:n.toLowerCase())};e0e.autoRemove=e=>np(e);const t0e=(e,t,n)=>{var r;return(r=e.getValue(t))==null?void 0:r.includes(n)};t0e.autoRemove=e=>np(e);const n0e=(e,t,n)=>!n.some(r=>{var i;return!((i=e.getValue(t))!=null&&i.includes(r))});n0e.autoRemove=e=>np(e)||!(e!=null&&e.length);const r0e=(e,t,n)=>n.some(r=>{var i;return(i=e.getValue(t))==null?void 0:i.includes(r)});r0e.autoRemove=e=>np(e)||!(e!=null&&e.length);const i0e=(e,t,n)=>e.getValue(t)===n;i0e.autoRemove=e=>np(e);const a0e=(e,t,n)=>e.getValue(t)==n;a0e.autoRemove=e=>np(e);const kI=(e,t,n)=>{let[r,i]=n;const a=e.getValue(t);return a>=r&&a<=i};kI.resolveFilterValue=e=>{let[t,n]=e,r=typeof t!="number"?parseFloat(t):t,i=typeof n!="number"?parseFloat(n):n,a=t===null||Number.isNaN(r)?-1/0:r,u=n===null||Number.isNaN(i)?1/0:i;if(a>u){const l=a;a=u,u=l}return[a,u]};kI.autoRemove=e=>np(e)||np(e[0])&&np(e[1]);const Bm={includesString:Wpe,includesStringSensitive:Jpe,equalsString:e0e,arrIncludes:t0e,arrIncludesAll:n0e,arrIncludesSome:r0e,equals:i0e,weakEquals:a0e,inNumberRange:kI};function np(e){return e==null||e===""}const jke={getDefaultColumnDef:()=>({filterFn:"auto"}),getInitialState:e=>({columnFilters:[],...e}),getDefaultOptions:e=>({onColumnFiltersChange:Yf("columnFilters",e),filterFromLeafRows:!1,maxLeafRowFilterDepth:100}),createColumn:(e,t)=>{e.getAutoFilterFn=()=>{const n=t.getCoreRowModel().flatRows[0],r=n==null?void 0:n.getValue(e.id);return typeof r=="string"?Bm.includesString:typeof r=="number"?Bm.inNumberRange:typeof r=="boolean"||r!==null&&typeof r=="object"?Bm.equals:Array.isArray(r)?Bm.arrIncludes:Bm.weakEquals},e.getFilterFn=()=>{var n,r;return yw(e.columnDef.filterFn)?e.columnDef.filterFn:e.columnDef.filterFn==="auto"?e.getAutoFilterFn():(n=(r=t.options.filterFns)==null?void 0:r[e.columnDef.filterFn])!=null?n:Bm[e.columnDef.filterFn]},e.getCanFilter=()=>{var n,r,i;return((n=e.columnDef.enableColumnFilter)!=null?n:!0)&&((r=t.options.enableColumnFilters)!=null?r:!0)&&((i=t.options.enableFilters)!=null?i:!0)&&!!e.accessorFn},e.getIsFiltered=()=>e.getFilterIndex()>-1,e.getFilterValue=()=>{var n;return(n=t.getState().columnFilters)==null||(n=n.find(r=>r.id===e.id))==null?void 0:n.value},e.getFilterIndex=()=>{var n,r;return(n=(r=t.getState().columnFilters)==null?void 0:r.findIndex(i=>i.id===e.id))!=null?n:-1},e.setFilterValue=n=>{t.setColumnFilters(r=>{const i=e.getFilterFn(),a=r==null?void 0:r.find(c=>c.id===e.id),u=Bv(n,a?a.value:void 0);if(hH(i,u,e)){var l;return(l=r==null?void 0:r.filter(c=>c.id!==e.id))!=null?l:[]}const o={id:e.id,value:u};if(a){var s;return(s=r==null?void 0:r.map(c=>c.id===e.id?o:c))!=null?s:[]}return r!=null&&r.length?[...r,o]:[o]})}},createRow:(e,t)=>{e.columnFilters={},e.columnFiltersMeta={}},createTable:e=>{e.setColumnFilters=t=>{const n=e.getAllLeafColumns(),r=i=>{var a;return(a=Bv(t,i))==null?void 0:a.filter(u=>{const l=n.find(o=>o.id===u.id);if(l){const o=l.getFilterFn();if(hH(o,u.value,l))return!1}return!0})};e.options.onColumnFiltersChange==null||e.options.onColumnFiltersChange(r)},e.resetColumnFilters=t=>{var n,r;e.setColumnFilters(t?[]:(n=(r=e.initialState)==null?void 0:r.columnFilters)!=null?n:[])},e.getPreFilteredRowModel=()=>e.getCoreRowModel(),e.getFilteredRowModel=()=>(!e._getFilteredRowModel&&e.options.getFilteredRowModel&&(e._getFilteredRowModel=e.options.getFilteredRowModel(e)),e.options.manualFiltering||!e._getFilteredRowModel?e.getPreFilteredRowModel():e._getFilteredRowModel())}};function hH(e,t,n){return(e&&e.autoRemove?e.autoRemove(t,n):!1)||typeof t>"u"||typeof t=="string"&&!t}const Zke=(e,t,n)=>n.reduce((r,i)=>{const a=i.getValue(e);return r+(typeof a=="number"?a:0)},0),Hke=(e,t,n)=>{let r;return n.forEach(i=>{const a=i.getValue(e);a!=null&&(r>a||r===void 0&&a>=a)&&(r=a)}),r},Xke=(e,t,n)=>{let r;return n.forEach(i=>{const a=i.getValue(e);a!=null&&(r<a||r===void 0&&a>=a)&&(r=a)}),r},Yke=(e,t,n)=>{let r,i;return n.forEach(a=>{const u=a.getValue(e);u!=null&&(r===void 0?u>=u&&(r=i=u):(r>u&&(r=u),i<u&&(i=u)))}),[r,i]},Gke=(e,t)=>{let n=0,r=0;if(t.forEach(i=>{let a=i.getValue(e);a!=null&&(a=+a)>=a&&(++n,r+=a)}),n)return r/n},Kke=(e,t)=>{if(!t.length)return;const n=t.map(a=>a.getValue(e));if(!Bke(n))return;if(n.length===1)return n[0];const r=Math.floor(n.length/2),i=n.sort((a,u)=>a-u);return n.length%2!==0?i[r]:(i[r-1]+i[r])/2},Wke=(e,t)=>Array.from(new Set(t.map(n=>n.getValue(e))).values()),Jke=(e,t)=>new Set(t.map(n=>n.getValue(e))).size,ePe=(e,t)=>t.length,Kk={sum:Zke,min:Hke,max:Xke,extent:Yke,mean:Gke,median:Kke,unique:Wke,uniqueCount:Jke,count:ePe},tPe={getDefaultColumnDef:()=>({aggregatedCell:e=>{var t,n;return(t=(n=e.getValue())==null||n.toString==null?void 0:n.toString())!=null?t:null},aggregationFn:"auto"}),getInitialState:e=>({grouping:[],...e}),getDefaultOptions:e=>({onGroupingChange:Yf("grouping",e),groupedColumnMode:"reorder"}),createColumn:(e,t)=>{e.toggleGrouping=()=>{t.setGrouping(n=>n!=null&&n.includes(e.id)?n.filter(r=>r!==e.id):[...n??[],e.id])},e.getCanGroup=()=>{var n,r;return((n=e.columnDef.enableGrouping)!=null?n:!0)&&((r=t.options.enableGrouping)!=null?r:!0)&&(!!e.accessorFn||!!e.columnDef.getGroupingValue)},e.getIsGrouped=()=>{var n;return(n=t.getState().grouping)==null?void 0:n.includes(e.id)},e.getGroupedIndex=()=>{var n;return(n=t.getState().grouping)==null?void 0:n.indexOf(e.id)},e.getToggleGroupingHandler=()=>{const n=e.getCanGroup();return()=>{n&&e.toggleGrouping()}},e.getAutoAggregationFn=()=>{const n=t.getCoreRowModel().flatRows[0],r=n==null?void 0:n.getValue(e.id);if(typeof r=="number")return Kk.sum;if(Object.prototype.toString.call(r)==="[object Date]")return Kk.extent},e.getAggregationFn=()=>{var n,r;if(!e)throw new Error;return yw(e.columnDef.aggregationFn)?e.columnDef.aggregationFn:e.columnDef.aggregationFn==="auto"?e.getAutoAggregationFn():(n=(r=t.options.aggregationFns)==null?void 0:r[e.columnDef.aggregationFn])!=null?n:Kk[e.columnDef.aggregationFn]}},createTable:e=>{e.setGrouping=t=>e.options.onGroupingChange==null?void 0:e.options.onGroupingChange(t),e.resetGrouping=t=>{var n,r;e.setGrouping(t?[]:(n=(r=e.initialState)==null?void 0:r.grouping)!=null?n:[])},e.getPreGroupedRowModel=()=>e.getFilteredRowModel(),e.getGroupedRowModel=()=>(!e._getGroupedRowModel&&e.options.getGroupedRowModel&&(e._getGroupedRowModel=e.options.getGroupedRowModel(e)),e.options.manualGrouping||!e._getGroupedRowModel?e.getPreGroupedRowModel():e._getGroupedRowModel())},createRow:(e,t)=>{e.getIsGrouped=()=>!!e.groupingColumnId,e.getGroupingValue=n=>{if(e._groupingValuesCache.hasOwnProperty(n))return e._groupingValuesCache[n];const r=t.getColumn(n);return r!=null&&r.columnDef.getGroupingValue?(e._groupingValuesCache[n]=r.columnDef.getGroupingValue(e.original),e._groupingValuesCache[n]):e.getValue(n)},e._groupingValuesCache={}},createCell:(e,t,n,r)=>{e.getIsGrouped=()=>t.getIsGrouped()&&t.id===n.groupingColumnId,e.getIsPlaceholder=()=>!e.getIsGrouped()&&t.getIsGrouped(),e.getIsAggregated=()=>{var i;return!e.getIsGrouped()&&!e.getIsPlaceholder()&&!!((i=n.subRows)!=null&&i.length)}}};function nPe(e,t,n){if(!(t!=null&&t.length)||!n)return e;const r=e.filter(a=>!t.includes(a.id));return n==="remove"?r:[...t.map(a=>e.find(u=>u.id===a)).filter(Boolean),...r]}const rPe={getInitialState:e=>({columnOrder:[],...e}),getDefaultOptions:e=>({onColumnOrderChange:Yf("columnOrder",e)}),createColumn:(e,t)=>{e.getIndex=Mn(n=>[G2(t,n)],n=>n.findIndex(r=>r.id===e.id),Fn(t.options,"debugColumns")),e.getIsFirstColumn=n=>{var r;return((r=G2(t,n)[0])==null?void 0:r.id)===e.id},e.getIsLastColumn=n=>{var r;const i=G2(t,n);return((r=i[i.length-1])==null?void 0:r.id)===e.id}},createTable:e=>{e.setColumnOrder=t=>e.options.onColumnOrderChange==null?void 0:e.options.onColumnOrderChange(t),e.resetColumnOrder=t=>{var n;e.setColumnOrder(t?[]:(n=e.initialState.columnOrder)!=null?n:[])},e._getOrderColumnsFn=Mn(()=>[e.getState().columnOrder,e.getState().grouping,e.options.groupedColumnMode],(t,n,r)=>i=>{let a=[];if(!(t!=null&&t.length))a=i;else{const u=[...t],l=[...i];for(;l.length&&u.length;){const o=u.shift(),s=l.findIndex(c=>c.id===o);s>-1&&a.push(l.splice(s,1)[0])}a=[...a,...l]}return nPe(a,n,r)},Fn(e.options,"debugTable"))}},Wk=()=>({left:[],right:[]}),iPe={getInitialState:e=>({columnPinning:Wk(),...e}),getDefaultOptions:e=>({onColumnPinningChange:Yf("columnPinning",e)}),createColumn:(e,t)=>{e.pin=n=>{const r=e.getLeafColumns().map(i=>i.id).filter(Boolean);t.setColumnPinning(i=>{var a,u;if(n==="right"){var l,o;return{left:((l=i==null?void 0:i.left)!=null?l:[]).filter(d=>!(r!=null&&r.includes(d))),right:[...((o=i==null?void 0:i.right)!=null?o:[]).filter(d=>!(r!=null&&r.includes(d))),...r]}}if(n==="left"){var s,c;return{left:[...((s=i==null?void 0:i.left)!=null?s:[]).filter(d=>!(r!=null&&r.includes(d))),...r],right:((c=i==null?void 0:i.right)!=null?c:[]).filter(d=>!(r!=null&&r.includes(d)))}}return{left:((a=i==null?void 0:i.left)!=null?a:[]).filter(d=>!(r!=null&&r.includes(d))),right:((u=i==null?void 0:i.right)!=null?u:[]).filter(d=>!(r!=null&&r.includes(d)))}})},e.getCanPin=()=>e.getLeafColumns().some(r=>{var i,a,u;return((i=r.columnDef.enablePinning)!=null?i:!0)&&((a=(u=t.options.enableColumnPinning)!=null?u:t.options.enablePinning)!=null?a:!0)}),e.getIsPinned=()=>{const n=e.getLeafColumns().map(l=>l.id),{left:r,right:i}=t.getState().columnPinning,a=n.some(l=>r==null?void 0:r.includes(l)),u=n.some(l=>i==null?void 0:i.includes(l));return a?"left":u?"right":!1},e.getPinnedIndex=()=>{var n,r;const i=e.getIsPinned();return i?(n=(r=t.getState().columnPinning)==null||(r=r[i])==null?void 0:r.indexOf(e.id))!=null?n:-1:0}},createRow:(e,t)=>{e.getCenterVisibleCells=Mn(()=>[e._getAllVisibleCells(),t.getState().columnPinning.left,t.getState().columnPinning.right],(n,r,i)=>{const a=[...r??[],...i??[]];return n.filter(u=>!a.includes(u.column.id))},Fn(t.options,"debugRows")),e.getLeftVisibleCells=Mn(()=>[e._getAllVisibleCells(),t.getState().columnPinning.left],(n,r)=>(r??[]).map(a=>n.find(u=>u.column.id===a)).filter(Boolean).map(a=>({...a,position:"left"})),Fn(t.options,"debugRows")),e.getRightVisibleCells=Mn(()=>[e._getAllVisibleCells(),t.getState().columnPinning.right],(n,r)=>(r??[]).map(a=>n.find(u=>u.column.id===a)).filter(Boolean).map(a=>({...a,position:"right"})),Fn(t.options,"debugRows"))},createTable:e=>{e.setColumnPinning=t=>e.options.onColumnPinningChange==null?void 0:e.options.onColumnPinningChange(t),e.resetColumnPinning=t=>{var n,r;return e.setColumnPinning(t?Wk():(n=(r=e.initialState)==null?void 0:r.columnPinning)!=null?n:Wk())},e.getIsSomeColumnsPinned=t=>{var n;const r=e.getState().columnPinning;if(!t){var i,a;return!!((i=r.left)!=null&&i.length||(a=r.right)!=null&&a.length)}return!!((n=r[t])!=null&&n.length)},e.getLeftLeafColumns=Mn(()=>[e.getAllLeafColumns(),e.getState().columnPinning.left],(t,n)=>(n??[]).map(r=>t.find(i=>i.id===r)).filter(Boolean),Fn(e.options,"debugColumns")),e.getRightLeafColumns=Mn(()=>[e.getAllLeafColumns(),e.getState().columnPinning.right],(t,n)=>(n??[]).map(r=>t.find(i=>i.id===r)).filter(Boolean),Fn(e.options,"debugColumns")),e.getCenterLeafColumns=Mn(()=>[e.getAllLeafColumns(),e.getState().columnPinning.left,e.getState().columnPinning.right],(t,n,r)=>{const i=[...n??[],...r??[]];return t.filter(a=>!i.includes(a.id))},Fn(e.options,"debugColumns"))}};function aPe(e){return e||(typeof document<"u"?document:null)}const uE={size:150,minSize:20,maxSize:Number.MAX_SAFE_INTEGER},Jk=()=>({startOffset:null,startSize:null,deltaOffset:null,deltaPercentage:null,isResizingColumn:!1,columnSizingStart:[]}),uPe={getDefaultColumnDef:()=>uE,getInitialState:e=>({columnSizing:{},columnSizingInfo:Jk(),...e}),getDefaultOptions:e=>({columnResizeMode:"onEnd",columnResizeDirection:"ltr",onColumnSizingChange:Yf("columnSizing",e),onColumnSizingInfoChange:Yf("columnSizingInfo",e)}),createColumn:(e,t)=>{e.getSize=()=>{var n,r,i;const a=t.getState().columnSizing[e.id];return Math.min(Math.max((n=e.columnDef.minSize)!=null?n:uE.minSize,(r=a??e.columnDef.size)!=null?r:uE.size),(i=e.columnDef.maxSize)!=null?i:uE.maxSize)},e.getStart=Mn(n=>[n,G2(t,n),t.getState().columnSizing],(n,r)=>r.slice(0,e.getIndex(n)).reduce((i,a)=>i+a.getSize(),0),Fn(t.options,"debugColumns")),e.getAfter=Mn(n=>[n,G2(t,n),t.getState().columnSizing],(n,r)=>r.slice(e.getIndex(n)+1).reduce((i,a)=>i+a.getSize(),0),Fn(t.options,"debugColumns")),e.resetSize=()=>{t.setColumnSizing(n=>{let{[e.id]:r,...i}=n;return i})},e.getCanResize=()=>{var n,r;return((n=e.columnDef.enableResizing)!=null?n:!0)&&((r=t.options.enableColumnResizing)!=null?r:!0)},e.getIsResizing=()=>t.getState().columnSizingInfo.isResizingColumn===e.id},createHeader:(e,t)=>{e.getSize=()=>{let n=0;const r=i=>{if(i.subHeaders.length)i.subHeaders.forEach(r);else{var a;n+=(a=i.column.getSize())!=null?a:0}};return r(e),n},e.getStart=()=>{if(e.index>0){const n=e.headerGroup.headers[e.index-1];return n.getStart()+n.getSize()}return 0},e.getResizeHandler=n=>{const r=t.getColumn(e.column.id),i=r==null?void 0:r.getCanResize();return a=>{if(!r||!i||(a.persist==null||a.persist(),eP(a)&&a.touches&&a.touches.length>1))return;const u=e.getSize(),l=e?e.getLeafHeaders().map(_=>[_.column.id,_.column.getSize()]):[[r.id,r.getSize()]],o=eP(a)?Math.round(a.touches[0].clientX):a.clientX,s={},c=(_,E)=>{typeof E=="number"&&(t.setColumnSizingInfo($=>{var k,A;const w=t.options.columnResizeDirection==="rtl"?-1:1,R=(E-((k=$==null?void 0:$.startOffset)!=null?k:0))*w,F=Math.max(R/((A=$==null?void 0:$.startSize)!=null?A:0),-.999999);return $.columnSizingStart.forEach(I=>{let[U,q]=I;s[U]=Math.round(Math.max(q+q*F,0)*100)/100}),{...$,deltaOffset:R,deltaPercentage:F}}),(t.options.columnResizeMode==="onChange"||_==="end")&&t.setColumnSizing($=>({...$,...s})))},d=_=>c("move",_),h=_=>{c("end",_),t.setColumnSizingInfo(E=>({...E,isResizingColumn:!1,startOffset:null,startSize:null,deltaOffset:null,deltaPercentage:null,columnSizingStart:[]}))},m=aPe(n),y={moveHandler:_=>d(_.clientX),upHandler:_=>{m==null||m.removeEventListener("mousemove",y.moveHandler),m==null||m.removeEventListener("mouseup",y.upHandler),h(_.clientX)}},v={moveHandler:_=>(_.cancelable&&(_.preventDefault(),_.stopPropagation()),d(_.touches[0].clientX),!1),upHandler:_=>{var E;m==null||m.removeEventListener("touchmove",v.moveHandler),m==null||m.removeEventListener("touchend",v.upHandler),_.cancelable&&(_.preventDefault(),_.stopPropagation()),h((E=_.touches[0])==null?void 0:E.clientX)}},S=lPe()?{passive:!1}:!1;eP(a)?(m==null||m.addEventListener("touchmove",v.moveHandler,S),m==null||m.addEventListener("touchend",v.upHandler,S)):(m==null||m.addEventListener("mousemove",y.moveHandler,S),m==null||m.addEventListener("mouseup",y.upHandler,S)),t.setColumnSizingInfo(_=>({..._,startOffset:o,startSize:u,deltaOffset:0,deltaPercentage:0,columnSizingStart:l,isResizingColumn:r.id}))}}},createTable:e=>{e.setColumnSizing=t=>e.options.onColumnSizingChange==null?void 0:e.options.onColumnSizingChange(t),e.setColumnSizingInfo=t=>e.options.onColumnSizingInfoChange==null?void 0:e.options.onColumnSizingInfoChange(t),e.resetColumnSizing=t=>{var n;e.setColumnSizing(t?{}:(n=e.initialState.columnSizing)!=null?n:{})},e.resetHeaderSizeInfo=t=>{var n;e.setColumnSizingInfo(t?Jk():(n=e.initialState.columnSizingInfo)!=null?n:Jk())},e.getTotalSize=()=>{var t,n;return(t=(n=e.getHeaderGroups()[0])==null?void 0:n.headers.reduce((r,i)=>r+i.getSize(),0))!=null?t:0},e.getLeftTotalSize=()=>{var t,n;return(t=(n=e.getLeftHeaderGroups()[0])==null?void 0:n.headers.reduce((r,i)=>r+i.getSize(),0))!=null?t:0},e.getCenterTotalSize=()=>{var t,n;return(t=(n=e.getCenterHeaderGroups()[0])==null?void 0:n.headers.reduce((r,i)=>r+i.getSize(),0))!=null?t:0},e.getRightTotalSize=()=>{var t,n;return(t=(n=e.getRightHeaderGroups()[0])==null?void 0:n.headers.reduce((r,i)=>r+i.getSize(),0))!=null?t:0}}};let lE=null;function lPe(){if(typeof lE=="boolean")return lE;let e=!1;try{const t={get passive(){return e=!0,!1}},n=()=>{};window.addEventListener("test",n,t),window.removeEventListener("test",n)}catch{e=!1}return lE=e,lE}function eP(e){return e.type==="touchstart"}const oPe={getInitialState:e=>({columnVisibility:{},...e}),getDefaultOptions:e=>({onColumnVisibilityChange:Yf("columnVisibility",e)}),createColumn:(e,t)=>{e.toggleVisibility=n=>{e.getCanHide()&&t.setColumnVisibility(r=>({...r,[e.id]:n??!e.getIsVisible()}))},e.getIsVisible=()=>{var n,r;const i=e.columns;return(n=i.length?i.some(a=>a.getIsVisible()):(r=t.getState().columnVisibility)==null?void 0:r[e.id])!=null?n:!0},e.getCanHide=()=>{var n,r;return((n=e.columnDef.enableHiding)!=null?n:!0)&&((r=t.options.enableHiding)!=null?r:!0)},e.getToggleVisibilityHandler=()=>n=>{e.toggleVisibility==null||e.toggleVisibility(n.target.checked)}},createRow:(e,t)=>{e._getAllVisibleCells=Mn(()=>[e.getAllCells(),t.getState().columnVisibility],n=>n.filter(r=>r.column.getIsVisible()),Fn(t.options,"debugRows")),e.getVisibleCells=Mn(()=>[e.getLeftVisibleCells(),e.getCenterVisibleCells(),e.getRightVisibleCells()],(n,r,i)=>[...n,...r,...i],Fn(t.options,"debugRows"))},createTable:e=>{const t=(n,r)=>Mn(()=>[r(),r().filter(i=>i.getIsVisible()).map(i=>i.id).join("_")],i=>i.filter(a=>a.getIsVisible==null?void 0:a.getIsVisible()),Fn(e.options,"debugColumns"));e.getVisibleFlatColumns=t("getVisibleFlatColumns",()=>e.getAllFlatColumns()),e.getVisibleLeafColumns=t("getVisibleLeafColumns",()=>e.getAllLeafColumns()),e.getLeftVisibleLeafColumns=t("getLeftVisibleLeafColumns",()=>e.getLeftLeafColumns()),e.getRightVisibleLeafColumns=t("getRightVisibleLeafColumns",()=>e.getRightLeafColumns()),e.getCenterVisibleLeafColumns=t("getCenterVisibleLeafColumns",()=>e.getCenterLeafColumns()),e.setColumnVisibility=n=>e.options.onColumnVisibilityChange==null?void 0:e.options.onColumnVisibilityChange(n),e.resetColumnVisibility=n=>{var r;e.setColumnVisibility(n?{}:(r=e.initialState.columnVisibility)!=null?r:{})},e.toggleAllColumnsVisible=n=>{var r;n=(r=n)!=null?r:!e.getIsAllColumnsVisible(),e.setColumnVisibility(e.getAllLeafColumns().reduce((i,a)=>({...i,[a.id]:n||!(a.getCanHide!=null&&a.getCanHide())}),{}))},e.getIsAllColumnsVisible=()=>!e.getAllLeafColumns().some(n=>!(n.getIsVisible!=null&&n.getIsVisible())),e.getIsSomeColumnsVisible=()=>e.getAllLeafColumns().some(n=>n.getIsVisible==null?void 0:n.getIsVisible()),e.getToggleAllColumnsVisibilityHandler=()=>n=>{var r;e.toggleAllColumnsVisible((r=n.target)==null?void 0:r.checked)}}};function G2(e,t){return t?t==="center"?e.getCenterVisibleLeafColumns():t==="left"?e.getLeftVisibleLeafColumns():e.getRightVisibleLeafColumns():e.getVisibleLeafColumns()}const sPe={createTable:e=>{e._getGlobalFacetedRowModel=e.options.getFacetedRowModel&&e.options.getFacetedRowModel(e,"__global__"),e.getGlobalFacetedRowModel=()=>e.options.manualFiltering||!e._getGlobalFacetedRowModel?e.getPreFilteredRowModel():e._getGlobalFacetedRowModel(),e._getGlobalFacetedUniqueValues=e.options.getFacetedUniqueValues&&e.options.getFacetedUniqueValues(e,"__global__"),e.getGlobalFacetedUniqueValues=()=>e._getGlobalFacetedUniqueValues?e._getGlobalFacetedUniqueValues():new Map,e._getGlobalFacetedMinMaxValues=e.options.getFacetedMinMaxValues&&e.options.getFacetedMinMaxValues(e,"__global__"),e.getGlobalFacetedMinMaxValues=()=>{if(e._getGlobalFacetedMinMaxValues)return e._getGlobalFacetedMinMaxValues()}}},cPe={getInitialState:e=>({globalFilter:void 0,...e}),getDefaultOptions:e=>({onGlobalFilterChange:Yf("globalFilter",e),globalFilterFn:"auto",getColumnCanGlobalFilter:t=>{var n;const r=(n=e.getCoreRowModel().flatRows[0])==null||(n=n._getAllCellsByColumnId()[t.id])==null?void 0:n.getValue();return typeof r=="string"||typeof r=="number"}}),createColumn:(e,t)=>{e.getCanGlobalFilter=()=>{var n,r,i,a;return((n=e.columnDef.enableGlobalFilter)!=null?n:!0)&&((r=t.options.enableGlobalFilter)!=null?r:!0)&&((i=t.options.enableFilters)!=null?i:!0)&&((a=t.options.getColumnCanGlobalFilter==null?void 0:t.options.getColumnCanGlobalFilter(e))!=null?a:!0)&&!!e.accessorFn}},createTable:e=>{e.getGlobalAutoFilterFn=()=>Bm.includesString,e.getGlobalFilterFn=()=>{var t,n;const{globalFilterFn:r}=e.options;return yw(r)?r:r==="auto"?e.getGlobalAutoFilterFn():(t=(n=e.options.filterFns)==null?void 0:n[r])!=null?t:Bm[r]},e.setGlobalFilter=t=>{e.options.onGlobalFilterChange==null||e.options.onGlobalFilterChange(t)},e.resetGlobalFilter=t=>{e.setGlobalFilter(t?void 0:e.initialState.globalFilter)}}},fPe={getInitialState:e=>({expanded:{},...e}),getDefaultOptions:e=>({onExpandedChange:Yf("expanded",e),paginateExpandedRows:!0}),createTable:e=>{let t=!1,n=!1;e._autoResetExpanded=()=>{var r,i;if(!t){e._queue(()=>{t=!0});return}if((r=(i=e.options.autoResetAll)!=null?i:e.options.autoResetExpanded)!=null?r:!e.options.manualExpanding){if(n)return;n=!0,e._queue(()=>{e.resetExpanded(),n=!1})}},e.setExpanded=r=>e.options.onExpandedChange==null?void 0:e.options.onExpandedChange(r),e.toggleAllRowsExpanded=r=>{r??!e.getIsAllRowsExpanded()?e.setExpanded(!0):e.setExpanded({})},e.resetExpanded=r=>{var i,a;e.setExpanded(r?{}:(i=(a=e.initialState)==null?void 0:a.expanded)!=null?i:{})},e.getCanSomeRowsExpand=()=>e.getPrePaginationRowModel().flatRows.some(r=>r.getCanExpand()),e.getToggleAllRowsExpandedHandler=()=>r=>{r.persist==null||r.persist(),e.toggleAllRowsExpanded()},e.getIsSomeRowsExpanded=()=>{const r=e.getState().expanded;return r===!0||Object.values(r).some(Boolean)},e.getIsAllRowsExpanded=()=>{const r=e.getState().expanded;return typeof r=="boolean"?r===!0:!(!Object.keys(r).length||e.getRowModel().flatRows.some(i=>!i.getIsExpanded()))},e.getExpandedDepth=()=>{let r=0;return(e.getState().expanded===!0?Object.keys(e.getRowModel().rowsById):Object.keys(e.getState().expanded)).forEach(a=>{const u=a.split(".");r=Math.max(r,u.length)}),r},e.getPreExpandedRowModel=()=>e.getSortedRowModel(),e.getExpandedRowModel=()=>(!e._getExpandedRowModel&&e.options.getExpandedRowModel&&(e._getExpandedRowModel=e.options.getExpandedRowModel(e)),e.options.manualExpanding||!e._getExpandedRowModel?e.getPreExpandedRowModel():e._getExpandedRowModel())},createRow:(e,t)=>{e.toggleExpanded=n=>{t.setExpanded(r=>{var i;const a=r===!0?!0:!!(r!=null&&r[e.id]);let u={};if(r===!0?Object.keys(t.getRowModel().rowsById).forEach(l=>{u[l]=!0}):u=r,n=(i=n)!=null?i:!a,!a&&n)return{...u,[e.id]:!0};if(a&&!n){const{[e.id]:l,...o}=u;return o}return r})},e.getIsExpanded=()=>{var n;const r=t.getState().expanded;return!!((n=t.options.getIsRowExpanded==null?void 0:t.options.getIsRowExpanded(e))!=null?n:r===!0||r!=null&&r[e.id])},e.getCanExpand=()=>{var n,r,i;return(n=t.options.getRowCanExpand==null?void 0:t.options.getRowCanExpand(e))!=null?n:((r=t.options.enableExpanding)!=null?r:!0)&&!!((i=e.subRows)!=null&&i.length)},e.getIsAllParentsExpanded=()=>{let n=!0,r=e;for(;n&&r.parentId;)r=t.getRow(r.parentId,!0),n=r.getIsExpanded();return n},e.getToggleExpandedHandler=()=>{const n=e.getCanExpand();return()=>{n&&e.toggleExpanded()}}}},sM=0,cM=10,tP=()=>({pageIndex:sM,pageSize:cM}),dPe={getInitialState:e=>({...e,pagination:{...tP(),...e==null?void 0:e.pagination}}),getDefaultOptions:e=>({onPaginationChange:Yf("pagination",e)}),createTable:e=>{let t=!1,n=!1;e._autoResetPageIndex=()=>{var r,i;if(!t){e._queue(()=>{t=!0});return}if((r=(i=e.options.autoResetAll)!=null?i:e.options.autoResetPageIndex)!=null?r:!e.options.manualPagination){if(n)return;n=!0,e._queue(()=>{e.resetPageIndex(),n=!1})}},e.setPagination=r=>{const i=a=>Bv(r,a);return e.options.onPaginationChange==null?void 0:e.options.onPaginationChange(i)},e.resetPagination=r=>{var i;e.setPagination(r?tP():(i=e.initialState.pagination)!=null?i:tP())},e.setPageIndex=r=>{e.setPagination(i=>{let a=Bv(r,i.pageIndex);const u=typeof e.options.pageCount>"u"||e.options.pageCount===-1?Number.MAX_SAFE_INTEGER:e.options.pageCount-1;return a=Math.max(0,Math.min(a,u)),{...i,pageIndex:a}})},e.resetPageIndex=r=>{var i,a;e.setPageIndex(r?sM:(i=(a=e.initialState)==null||(a=a.pagination)==null?void 0:a.pageIndex)!=null?i:sM)},e.resetPageSize=r=>{var i,a;e.setPageSize(r?cM:(i=(a=e.initialState)==null||(a=a.pagination)==null?void 0:a.pageSize)!=null?i:cM)},e.setPageSize=r=>{e.setPagination(i=>{const a=Math.max(1,Bv(r,i.pageSize)),u=i.pageSize*i.pageIndex,l=Math.floor(u/a);return{...i,pageIndex:l,pageSize:a}})},e.setPageCount=r=>e.setPagination(i=>{var a;let u=Bv(r,(a=e.options.pageCount)!=null?a:-1);return typeof u=="number"&&(u=Math.max(-1,u)),{...i,pageCount:u}}),e.getPageOptions=Mn(()=>[e.getPageCount()],r=>{let i=[];return r&&r>0&&(i=[...new Array(r)].fill(null).map((a,u)=>u)),i},Fn(e.options,"debugTable")),e.getCanPreviousPage=()=>e.getState().pagination.pageIndex>0,e.getCanNextPage=()=>{const{pageIndex:r}=e.getState().pagination,i=e.getPageCount();return i===-1?!0:i===0?!1:r<i-1},e.previousPage=()=>e.setPageIndex(r=>r-1),e.nextPage=()=>e.setPageIndex(r=>r+1),e.firstPage=()=>e.setPageIndex(0),e.lastPage=()=>e.setPageIndex(e.getPageCount()-1),e.getPrePaginationRowModel=()=>e.getExpandedRowModel(),e.getPaginationRowModel=()=>(!e._getPaginationRowModel&&e.options.getPaginationRowModel&&(e._getPaginationRowModel=e.options.getPaginationRowModel(e)),e.options.manualPagination||!e._getPaginationRowModel?e.getPrePaginationRowModel():e._getPaginationRowModel()),e.getPageCount=()=>{var r;return(r=e.options.pageCount)!=null?r:Math.ceil(e.getRowCount()/e.getState().pagination.pageSize)},e.getRowCount=()=>{var r;return(r=e.options.rowCount)!=null?r:e.getPrePaginationRowModel().rows.length}}},nP=()=>({top:[],bottom:[]}),hPe={getInitialState:e=>({rowPinning:nP(),...e}),getDefaultOptions:e=>({onRowPinningChange:Yf("rowPinning",e)}),createRow:(e,t)=>{e.pin=(n,r,i)=>{const a=r?e.getLeafRows().map(o=>{let{id:s}=o;return s}):[],u=i?e.getParentRows().map(o=>{let{id:s}=o;return s}):[],l=new Set([...u,e.id,...a]);t.setRowPinning(o=>{var s,c;if(n==="bottom"){var d,h;return{top:((d=o==null?void 0:o.top)!=null?d:[]).filter(v=>!(l!=null&&l.has(v))),bottom:[...((h=o==null?void 0:o.bottom)!=null?h:[]).filter(v=>!(l!=null&&l.has(v))),...Array.from(l)]}}if(n==="top"){var m,y;return{top:[...((m=o==null?void 0:o.top)!=null?m:[]).filter(v=>!(l!=null&&l.has(v))),...Array.from(l)],bottom:((y=o==null?void 0:o.bottom)!=null?y:[]).filter(v=>!(l!=null&&l.has(v)))}}return{top:((s=o==null?void 0:o.top)!=null?s:[]).filter(v=>!(l!=null&&l.has(v))),bottom:((c=o==null?void 0:o.bottom)!=null?c:[]).filter(v=>!(l!=null&&l.has(v)))}})},e.getCanPin=()=>{var n;const{enableRowPinning:r,enablePinning:i}=t.options;return typeof r=="function"?r(e):(n=r??i)!=null?n:!0},e.getIsPinned=()=>{const n=[e.id],{top:r,bottom:i}=t.getState().rowPinning,a=n.some(l=>r==null?void 0:r.includes(l)),u=n.some(l=>i==null?void 0:i.includes(l));return a?"top":u?"bottom":!1},e.getPinnedIndex=()=>{var n,r;const i=e.getIsPinned();if(!i)return-1;const a=(n=i==="top"?t.getTopRows():t.getBottomRows())==null?void 0:n.map(u=>{let{id:l}=u;return l});return(r=a==null?void 0:a.indexOf(e.id))!=null?r:-1}},createTable:e=>{e.setRowPinning=t=>e.options.onRowPinningChange==null?void 0:e.options.onRowPinningChange(t),e.resetRowPinning=t=>{var n,r;return e.setRowPinning(t?nP():(n=(r=e.initialState)==null?void 0:r.rowPinning)!=null?n:nP())},e.getIsSomeRowsPinned=t=>{var n;const r=e.getState().rowPinning;if(!t){var i,a;return!!((i=r.top)!=null&&i.length||(a=r.bottom)!=null&&a.length)}return!!((n=r[t])!=null&&n.length)},e._getPinnedRows=(t,n,r)=>{var i;return((i=e.options.keepPinnedRows)==null||i?(n??[]).map(u=>{const l=e.getRow(u,!0);return l.getIsAllParentsExpanded()?l:null}):(n??[]).map(u=>t.find(l=>l.id===u))).filter(Boolean).map(u=>({...u,position:r}))},e.getTopRows=Mn(()=>[e.getRowModel().rows,e.getState().rowPinning.top],(t,n)=>e._getPinnedRows(t,n,"top"),Fn(e.options,"debugRows")),e.getBottomRows=Mn(()=>[e.getRowModel().rows,e.getState().rowPinning.bottom],(t,n)=>e._getPinnedRows(t,n,"bottom"),Fn(e.options,"debugRows")),e.getCenterRows=Mn(()=>[e.getRowModel().rows,e.getState().rowPinning.top,e.getState().rowPinning.bottom],(t,n,r)=>{const i=new Set([...n??[],...r??[]]);return t.filter(a=>!i.has(a.id))},Fn(e.options,"debugRows"))}},pPe={getInitialState:e=>({rowSelection:{},...e}),getDefaultOptions:e=>({onRowSelectionChange:Yf("rowSelection",e),enableRowSelection:!0,enableMultiRowSelection:!0,enableSubRowSelection:!0}),createTable:e=>{e.setRowSelection=t=>e.options.onRowSelectionChange==null?void 0:e.options.onRowSelectionChange(t),e.resetRowSelection=t=>{var n;return e.setRowSelection(t?{}:(n=e.initialState.rowSelection)!=null?n:{})},e.toggleAllRowsSelected=t=>{e.setRowSelection(n=>{t=typeof t<"u"?t:!e.getIsAllRowsSelected();const r={...n},i=e.getPreGroupedRowModel().flatRows;return t?i.forEach(a=>{a.getCanSelect()&&(r[a.id]=!0)}):i.forEach(a=>{delete r[a.id]}),r})},e.toggleAllPageRowsSelected=t=>e.setRowSelection(n=>{const r=typeof t<"u"?t:!e.getIsAllPageRowsSelected(),i={...n};return e.getRowModel().rows.forEach(a=>{fM(i,a.id,r,!0,e)}),i}),e.getPreSelectedRowModel=()=>e.getCoreRowModel(),e.getSelectedRowModel=Mn(()=>[e.getState().rowSelection,e.getCoreRowModel()],(t,n)=>Object.keys(t).length?rP(e,n):{rows:[],flatRows:[],rowsById:{}},Fn(e.options,"debugTable")),e.getFilteredSelectedRowModel=Mn(()=>[e.getState().rowSelection,e.getFilteredRowModel()],(t,n)=>Object.keys(t).length?rP(e,n):{rows:[],flatRows:[],rowsById:{}},Fn(e.options,"debugTable")),e.getGroupedSelectedRowModel=Mn(()=>[e.getState().rowSelection,e.getSortedRowModel()],(t,n)=>Object.keys(t).length?rP(e,n):{rows:[],flatRows:[],rowsById:{}},Fn(e.options,"debugTable")),e.getIsAllRowsSelected=()=>{const t=e.getFilteredRowModel().flatRows,{rowSelection:n}=e.getState();let r=!!(t.length&&Object.keys(n).length);return r&&t.some(i=>i.getCanSelect()&&!n[i.id])&&(r=!1),r},e.getIsAllPageRowsSelected=()=>{const t=e.getPaginationRowModel().flatRows.filter(i=>i.getCanSelect()),{rowSelection:n}=e.getState();let r=!!t.length;return r&&t.some(i=>!n[i.id])&&(r=!1),r},e.getIsSomeRowsSelected=()=>{var t;const n=Object.keys((t=e.getState().rowSelection)!=null?t:{}).length;return n>0&&n<e.getFilteredRowModel().flatRows.length},e.getIsSomePageRowsSelected=()=>{const t=e.getPaginationRowModel().flatRows;return e.getIsAllPageRowsSelected()?!1:t.filter(n=>n.getCanSelect()).some(n=>n.getIsSelected()||n.getIsSomeSelected())},e.getToggleAllRowsSelectedHandler=()=>t=>{e.toggleAllRowsSelected(t.target.checked)},e.getToggleAllPageRowsSelectedHandler=()=>t=>{e.toggleAllPageRowsSelected(t.target.checked)}},createRow:(e,t)=>{e.toggleSelected=(n,r)=>{const i=e.getIsSelected();t.setRowSelection(a=>{var u;if(n=typeof n<"u"?n:!i,e.getCanSelect()&&i===n)return a;const l={...a};return fM(l,e.id,n,(u=r==null?void 0:r.selectChildren)!=null?u:!0,t),l})},e.getIsSelected=()=>{const{rowSelection:n}=t.getState();return PI(e,n)},e.getIsSomeSelected=()=>{const{rowSelection:n}=t.getState();return dM(e,n)==="some"},e.getIsAllSubRowsSelected=()=>{const{rowSelection:n}=t.getState();return dM(e,n)==="all"},e.getCanSelect=()=>{var n;return typeof t.options.enableRowSelection=="function"?t.options.enableRowSelection(e):(n=t.options.enableRowSelection)!=null?n:!0},e.getCanSelectSubRows=()=>{var n;return typeof t.options.enableSubRowSelection=="function"?t.options.enableSubRowSelection(e):(n=t.options.enableSubRowSelection)!=null?n:!0},e.getCanMultiSelect=()=>{var n;return typeof t.options.enableMultiRowSelection=="function"?t.options.enableMultiRowSelection(e):(n=t.options.enableMultiRowSelection)!=null?n:!0},e.getToggleSelectedHandler=()=>{const n=e.getCanSelect();return r=>{var i;n&&e.toggleSelected((i=r.target)==null?void 0:i.checked)}}}},fM=(e,t,n,r,i)=>{var a;const u=i.getRow(t,!0);n?(u.getCanMultiSelect()||Object.keys(e).forEach(l=>delete e[l]),u.getCanSelect()&&(e[t]=!0)):delete e[t],r&&(a=u.subRows)!=null&&a.length&&u.getCanSelectSubRows()&&u.subRows.forEach(l=>fM(e,l.id,n,r,i))};function rP(e,t){const n=e.getState().rowSelection,r=[],i={},a=function(u,l){return u.map(o=>{var s;const c=PI(o,n);if(c&&(r.push(o),i[o.id]=o),(s=o.subRows)!=null&&s.length&&(o={...o,subRows:a(o.subRows)}),c)return o}).filter(Boolean)};return{rows:a(t.rows),flatRows:r,rowsById:i}}function PI(e,t){var n;return(n=t[e.id])!=null?n:!1}function dM(e,t,n){var r;if(!((r=e.subRows)!=null&&r.length))return!1;let i=!0,a=!1;return e.subRows.forEach(u=>{if(!(a&&!i)&&(u.getCanSelect()&&(PI(u,t)?a=!0:i=!1),u.subRows&&u.subRows.length)){const l=dM(u,t);l==="all"?a=!0:(l==="some"&&(a=!0),i=!1)}}),i?"all":a?"some":!1}const hM=/([0-9]+)/gm,mPe=(e,t,n)=>u0e(ty(e.getValue(n)).toLowerCase(),ty(t.getValue(n)).toLowerCase()),gPe=(e,t,n)=>u0e(ty(e.getValue(n)),ty(t.getValue(n))),vPe=(e,t,n)=>MI(ty(e.getValue(n)).toLowerCase(),ty(t.getValue(n)).toLowerCase()),yPe=(e,t,n)=>MI(ty(e.getValue(n)),ty(t.getValue(n))),bPe=(e,t,n)=>{const r=e.getValue(n),i=t.getValue(n);return r>i?1:r<i?-1:0},OPe=(e,t,n)=>MI(e.getValue(n),t.getValue(n));function MI(e,t){return e===t?0:e>t?1:-1}function ty(e){return typeof e=="number"?isNaN(e)||e===1/0||e===-1/0?"":String(e):typeof e=="string"?e:""}function u0e(e,t){const n=e.split(hM).filter(Boolean),r=t.split(hM).filter(Boolean);for(;n.length&&r.length;){const i=n.shift(),a=r.shift(),u=parseInt(i,10),l=parseInt(a,10),o=[u,l].sort();if(isNaN(o[0])){if(i>a)return 1;if(a>i)return-1;continue}if(isNaN(o[1]))return isNaN(u)?-1:1;if(u>l)return 1;if(l>u)return-1}return n.length-r.length}const l2={alphanumeric:mPe,alphanumericCaseSensitive:gPe,text:vPe,textCaseSensitive:yPe,datetime:bPe,basic:OPe},xPe={getInitialState:e=>({sorting:[],...e}),getDefaultColumnDef:()=>({sortingFn:"auto",sortUndefined:1}),getDefaultOptions:e=>({onSortingChange:Yf("sorting",e),isMultiSortEvent:t=>t.shiftKey}),createColumn:(e,t)=>{e.getAutoSortingFn=()=>{const n=t.getFilteredRowModel().flatRows.slice(10);let r=!1;for(const i of n){const a=i==null?void 0:i.getValue(e.id);if(Object.prototype.toString.call(a)==="[object Date]")return l2.datetime;if(typeof a=="string"&&(r=!0,a.split(hM).length>1))return l2.alphanumeric}return r?l2.text:l2.basic},e.getAutoSortDir=()=>{const n=t.getFilteredRowModel().flatRows[0];return typeof(n==null?void 0:n.getValue(e.id))=="string"?"asc":"desc"},e.getSortingFn=()=>{var n,r;if(!e)throw new Error;return yw(e.columnDef.sortingFn)?e.columnDef.sortingFn:e.columnDef.sortingFn==="auto"?e.getAutoSortingFn():(n=(r=t.options.sortingFns)==null?void 0:r[e.columnDef.sortingFn])!=null?n:l2[e.columnDef.sortingFn]},e.toggleSorting=(n,r)=>{const i=e.getNextSortingOrder(),a=typeof n<"u"&&n!==null;t.setSorting(u=>{const l=u==null?void 0:u.find(m=>m.id===e.id),o=u==null?void 0:u.findIndex(m=>m.id===e.id);let s=[],c,d=a?n:i==="desc";if(u!=null&&u.length&&e.getCanMultiSort()&&r?l?c="toggle":c="add":u!=null&&u.length&&o!==u.length-1?c="replace":l?c="toggle":c="replace",c==="toggle"&&(a||i||(c="remove")),c==="add"){var h;s=[...u,{id:e.id,desc:d}],s.splice(0,s.length-((h=t.options.maxMultiSortColCount)!=null?h:Number.MAX_SAFE_INTEGER))}else c==="toggle"?s=u.map(m=>m.id===e.id?{...m,desc:d}:m):c==="remove"?s=u.filter(m=>m.id!==e.id):s=[{id:e.id,desc:d}];return s})},e.getFirstSortDir=()=>{var n,r;return((n=(r=e.columnDef.sortDescFirst)!=null?r:t.options.sortDescFirst)!=null?n:e.getAutoSortDir()==="desc")?"desc":"asc"},e.getNextSortingOrder=n=>{var r,i;const a=e.getFirstSortDir(),u=e.getIsSorted();return u?u!==a&&((r=t.options.enableSortingRemoval)==null||r)&&(!(n&&(i=t.options.enableMultiRemove)!=null)||i)?!1:u==="desc"?"asc":"desc":a},e.getCanSort=()=>{var n,r;return((n=e.columnDef.enableSorting)!=null?n:!0)&&((r=t.options.enableSorting)!=null?r:!0)&&!!e.accessorFn},e.getCanMultiSort=()=>{var n,r;return(n=(r=e.columnDef.enableMultiSort)!=null?r:t.options.enableMultiSort)!=null?n:!!e.accessorFn},e.getIsSorted=()=>{var n;const r=(n=t.getState().sorting)==null?void 0:n.find(i=>i.id===e.id);return r?r.desc?"desc":"asc":!1},e.getSortIndex=()=>{var n,r;return(n=(r=t.getState().sorting)==null?void 0:r.findIndex(i=>i.id===e.id))!=null?n:-1},e.clearSorting=()=>{t.setSorting(n=>n!=null&&n.length?n.filter(r=>r.id!==e.id):[])},e.getToggleSortingHandler=()=>{const n=e.getCanSort();return r=>{n&&(r.persist==null||r.persist(),e.toggleSorting==null||e.toggleSorting(void 0,e.getCanMultiSort()?t.options.isMultiSortEvent==null?void 0:t.options.isMultiSortEvent(r):!1))}}},createTable:e=>{e.setSorting=t=>e.options.onSortingChange==null?void 0:e.options.onSortingChange(t),e.resetSorting=t=>{var n,r;e.setSorting(t?[]:(n=(r=e.initialState)==null?void 0:r.sorting)!=null?n:[])},e.getPreSortedRowModel=()=>e.getGroupedRowModel(),e.getSortedRowModel=()=>(!e._getSortedRowModel&&e.options.getSortedRowModel&&(e._getSortedRowModel=e.options.getSortedRowModel(e)),e.options.manualSorting||!e._getSortedRowModel?e.getPreSortedRowModel():e._getSortedRowModel())}},SPe=[Vke,oPe,rPe,iPe,Uke,jke,sPe,cPe,xPe,tPe,fPe,dPe,hPe,pPe,uPe];function _Pe(e){var t,n;const r=[...SPe,...(t=e._features)!=null?t:[]];let i={_features:r};const a=i._features.reduce((h,m)=>Object.assign(h,m.getDefaultOptions==null?void 0:m.getDefaultOptions(i)),{}),u=h=>i.options.mergeOptions?i.options.mergeOptions(a,h):{...a,...h};let o={...{},...(n=e.initialState)!=null?n:{}};i._features.forEach(h=>{var m;o=(m=h.getInitialState==null?void 0:h.getInitialState(o))!=null?m:o});const s=[];let c=!1;const d={_features:r,options:{...a,...e},initialState:o,_queue:h=>{s.push(h),c||(c=!0,Promise.resolve().then(()=>{for(;s.length;)s.shift()();c=!1}).catch(m=>setTimeout(()=>{throw m})))},reset:()=>{i.setState(i.initialState)},setOptions:h=>{const m=Bv(h,i.options);i.options=u(m)},getState:()=>i.options.state,setState:h=>{i.options.onStateChange==null||i.options.onStateChange(h)},_getRowId:(h,m,y)=>{var v;return(v=i.options.getRowId==null?void 0:i.options.getRowId(h,m,y))!=null?v:`${y?[y.id,m].join("."):m}`},getCoreRowModel:()=>(i._getCoreRowModel||(i._getCoreRowModel=i.options.getCoreRowModel(i)),i._getCoreRowModel()),getRowModel:()=>i.getPaginationRowModel(),getRow:(h,m)=>{let y=(m?i.getPrePaginationRowModel():i.getRowModel()).rowsById[h];if(!y&&(y=i.getCoreRowModel().rowsById[h],!y))throw new Error;return y},_getDefaultColumnDef:Mn(()=>[i.options.defaultColumn],h=>{var m;return h=(m=h)!=null?m:{},{header:y=>{const v=y.header.column.columnDef;return v.accessorKey?v.accessorKey:v.accessorFn?v.id:null},cell:y=>{var v,S;return(v=(S=y.renderValue())==null||S.toString==null?void 0:S.toString())!=null?v:null},...i._features.reduce((y,v)=>Object.assign(y,v.getDefaultColumnDef==null?void 0:v.getDefaultColumnDef()),{}),...h}},Fn(e,"debugColumns")),_getColumnDefs:()=>i.options.columns,getAllColumns:Mn(()=>[i._getColumnDefs()],h=>{const m=function(y,v,S){return S===void 0&&(S=0),y.map(_=>{const E=qke(i,_,S,v),$=_;return E.columns=$.columns?m($.columns,E,S+1):[],E})};return m(h)},Fn(e,"debugColumns")),getAllFlatColumns:Mn(()=>[i.getAllColumns()],h=>h.flatMap(m=>m.getFlatColumns()),Fn(e,"debugColumns")),_getAllFlatColumnsById:Mn(()=>[i.getAllFlatColumns()],h=>h.reduce((m,y)=>(m[y.id]=y,m),{}),Fn(e,"debugColumns")),getAllLeafColumns:Mn(()=>[i.getAllColumns(),i._getOrderColumnsFn()],(h,m)=>{let y=h.flatMap(v=>v.getLeafColumns());return m(y)},Fn(e,"debugColumns")),getColumn:h=>i._getAllFlatColumnsById()[h]};Object.assign(i,d);for(let h=0;h<i._features.length;h++){const m=i._features[h];m==null||m.createTable==null||m.createTable(i)}return i}function Xat(){return e=>Mn(()=>[e.options.data],t=>{const n={rows:[],flatRows:[],rowsById:{}},r=function(i,a,u){a===void 0&&(a=0);const l=[];for(let s=0;s<i.length;s++){const c=TI(e,e._getRowId(i[s],s,u),i[s],s,a,void 0,u==null?void 0:u.id);if(n.flatRows.push(c),n.rowsById[c.id]=c,l.push(c),e.options.getSubRows){var o;c.originalSubRows=e.options.getSubRows(i[s],s),(o=c.originalSubRows)!=null&&o.length&&(c.subRows=r(c.originalSubRows,a+1,c))}}return l};return n.rows=r(t),n},Fn(e.options,"debugTable","getRowModel",()=>e._autoResetPageIndex()))}function Yat(){return e=>Mn(()=>[e.getState().expanded,e.getPreExpandedRowModel(),e.options.paginateExpandedRows],(t,n,r)=>!n.rows.length||t!==!0&&!Object.keys(t??{}).length||!r?n:l0e(n),Fn(e.options,"debugTable"))}function l0e(e){const t=[],n=r=>{var i;t.push(r),(i=r.subRows)!=null&&i.length&&r.getIsExpanded()&&r.subRows.forEach(n)};return e.rows.forEach(n),{rows:t,flatRows:e.flatRows,rowsById:e.rowsById}}function EPe(e,t,n){return n.options.filterFromLeafRows?DPe(e,t,n):$Pe(e,t,n)}function DPe(e,t,n){var r;const i=[],a={},u=(r=n.options.maxLeafRowFilterDepth)!=null?r:100,l=function(o,s){s===void 0&&(s=0);const c=[];for(let h=0;h<o.length;h++){var d;let m=o[h];const y=TI(n,m.id,m.original,m.index,m.depth,void 0,m.parentId);if(y.columnFilters=m.columnFilters,(d=m.subRows)!=null&&d.length&&s<u){if(y.subRows=l(m.subRows,s+1),m=y,t(m)&&!y.subRows.length){c.push(m),a[m.id]=m,i.push(m);continue}if(t(m)||y.subRows.length){c.push(m),a[m.id]=m,i.push(m);continue}}else m=y,t(m)&&(c.push(m),a[m.id]=m,i.push(m))}return c};return{rows:l(e),flatRows:i,rowsById:a}}function $Pe(e,t,n){var r;const i=[],a={},u=(r=n.options.maxLeafRowFilterDepth)!=null?r:100,l=function(o,s){s===void 0&&(s=0);const c=[];for(let h=0;h<o.length;h++){let m=o[h];if(t(m)){var d;if((d=m.subRows)!=null&&d.length&&s<u){const v=TI(n,m.id,m.original,m.index,m.depth,void 0,m.parentId);v.subRows=l(m.subRows,s+1),m=v}c.push(m),i.push(m),a[m.id]=m}}return c};return{rows:l(e),flatRows:i,rowsById:a}}function Gat(){return e=>Mn(()=>[e.getPreFilteredRowModel(),e.getState().columnFilters,e.getState().globalFilter],(t,n,r)=>{if(!t.rows.length||!(n!=null&&n.length)&&!r){for(let h=0;h<t.flatRows.length;h++)t.flatRows[h].columnFilters={},t.flatRows[h].columnFiltersMeta={};return t}const i=[],a=[];(n??[]).forEach(h=>{var m;const y=e.getColumn(h.id);if(!y)return;const v=y.getFilterFn();v&&i.push({id:h.id,filterFn:v,resolvedValue:(m=v.resolveFilterValue==null?void 0:v.resolveFilterValue(h.value))!=null?m:h.value})});const u=(n??[]).map(h=>h.id),l=e.getGlobalFilterFn(),o=e.getAllLeafColumns().filter(h=>h.getCanGlobalFilter());r&&l&&o.length&&(u.push("__global__"),o.forEach(h=>{var m;a.push({id:h.id,filterFn:l,resolvedValue:(m=l.resolveFilterValue==null?void 0:l.resolveFilterValue(r))!=null?m:r})}));let s,c;for(let h=0;h<t.flatRows.length;h++){const m=t.flatRows[h];if(m.columnFilters={},i.length)for(let y=0;y<i.length;y++){s=i[y];const v=s.id;m.columnFilters[v]=s.filterFn(m,v,s.resolvedValue,S=>{m.columnFiltersMeta[v]=S})}if(a.length){for(let y=0;y<a.length;y++){c=a[y];const v=c.id;if(c.filterFn(m,v,c.resolvedValue,S=>{m.columnFiltersMeta[v]=S})){m.columnFilters.__global__=!0;break}}m.columnFilters.__global__!==!0&&(m.columnFilters.__global__=!1)}}const d=h=>{for(let m=0;m<u.length;m++)if(h.columnFilters[u[m]]===!1)return!1;return!0};return EPe(t.rows,d,e)},Fn(e.options,"debugTable","getFilteredRowModel",()=>e._autoResetPageIndex()))}function Kat(e){return t=>Mn(()=>[t.getState().pagination,t.getPrePaginationRowModel(),t.options.paginateExpandedRows?void 0:t.getState().expanded],(n,r)=>{if(!r.rows.length)return r;const{pageSize:i,pageIndex:a}=n;let{rows:u,flatRows:l,rowsById:o}=r;const s=i*a,c=s+i;u=u.slice(s,c);let d;t.options.paginateExpandedRows?d={rows:u,flatRows:l,rowsById:o}:d=l0e({rows:u,flatRows:l,rowsById:o}),d.flatRows=[];const h=m=>{d.flatRows.push(m),m.subRows.length&&m.subRows.forEach(h)};return d.rows.forEach(h),d},Fn(t.options,"debugTable"))}function Wat(){return e=>Mn(()=>[e.getState().sorting,e.getPreSortedRowModel()],(t,n)=>{if(!n.rows.length||!(t!=null&&t.length))return n;const r=e.getState().sorting,i=[],a=r.filter(o=>{var s;return(s=e.getColumn(o.id))==null?void 0:s.getCanSort()}),u={};a.forEach(o=>{const s=e.getColumn(o.id);s&&(u[o.id]={sortUndefined:s.columnDef.sortUndefined,invertSorting:s.columnDef.invertSorting,sortingFn:s.getSortingFn()})});const l=o=>{const s=o.map(c=>({...c}));return s.sort((c,d)=>{for(let m=0;m<a.length;m+=1){var h;const y=a[m],v=u[y.id],S=v.sortUndefined,_=(h=y==null?void 0:y.desc)!=null?h:!1;let E=0;if(S){const $=c.getValue(y.id),k=d.getValue(y.id),A=$===void 0,w=k===void 0;if(A||w){if(S==="first")return A?-1:1;if(S==="last")return A?1:-1;E=A&&w?0:A?S:-S}}if(E===0&&(E=v.sortingFn(c,d,y.id)),E!==0)return _&&(E*=-1),v.invertSorting&&(E*=-1),E}return c.index-d.index}),s.forEach(c=>{var d;i.push(c),(d=c.subRows)!=null&&d.length&&(c.subRows=l(c.subRows))}),s};return{rows:l(n.rows),flatRows:i,rowsById:n.rowsById}},Fn(e.options,"debugTable","getSortedRowModel",()=>e._autoResetPageIndex()))}/**
   * react-table
   *
   * Copyright (c) TanStack
   *
   * This source code is licensed under the MIT license found in the
   * LICENSE.md file in the root directory of this source tree.
   *
   * @license MIT
   */function Jat(e,t){return e?CPe(e)?P.createElement(e,t):e:null}function CPe(e){return wPe(e)||typeof e=="function"||RPe(e)}function wPe(e){return typeof e=="function"&&(()=>{const t=Object.getPrototypeOf(e);return t.prototype&&t.prototype.isReactComponent})()}function RPe(e){return typeof e=="object"&&typeof e.$$typeof=="symbol"&&["react.memo","react.forward_ref"].includes(e.$$typeof.description)}function eut(e){const t={state:{},onStateChange:()=>{},renderFallbackValue:null,...e},[n]=P.useState(()=>({current:_Pe(t)})),[r,i]=P.useState(()=>n.current.initialState);return n.current.setOptions(a=>({...a,...e,state:{...r,...e.state},onStateChange:u=>{i(u),e.onStateChange==null||e.onStateChange(u)}})),n.current}function FI(e){for(var t=e.length/6|0,n=new Array(t),r=0;r<t;)n[r]="#"+e.slice(r*6,++r*6);return n}const tut=FI("1f77b4ff7f0e2ca02cd627289467bd8c564be377c27f7f7fbcbd2217becf"),nut=FI("a6cee31f78b4b2df8a33a02cfb9a99e31a1cfdbf6fff7f00cab2d66a3d9affff99b15928"),rut=FI("66c2a5fc8d628da0cbe78ac3a6d854ffd92fe5c494b3b3b3");function bw(e,t,n){e.prototype=t.prototype=n,n.constructor=e}function NI(e,t){var n=Object.create(e.prototype);for(var r in t)n[r]=t[r];return n}function n4(){}var d1=.7,k3=1/d1,h3="\\s*([+-]?\\d+)\\s*",Rx="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)\\s*",w0="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)%\\s*",APe=/^#([0-9a-f]{3,8})$/,TPe=new RegExp(`^rgb\\(${h3},${h3},${h3}\\)$`),kPe=new RegExp(`^rgb\\(${w0},${w0},${w0}\\)$`),PPe=new RegExp(`^rgba\\(${h3},${h3},${h3},${Rx}\\)$`),MPe=new RegExp(`^rgba\\(${w0},${w0},${w0},${Rx}\\)$`),FPe=new RegExp(`^hsl\\(${Rx},${w0},${w0}\\)$`),NPe=new RegExp(`^hsla\\(${Rx},${w0},${w0},${Rx}\\)$`),pH={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};bw(n4,Ax,{copy(e){return Object.assign(new this.constructor,this,e)},displayable(){return this.rgb().displayable()},hex:mH,formatHex:mH,formatHex8:IPe,formatHsl:LPe,formatRgb:gH,toString:gH});function mH(){return this.rgb().formatHex()}function IPe(){return this.rgb().formatHex8()}function LPe(){return s0e(this).formatHsl()}function gH(){return this.rgb().formatRgb()}function Ax(e){var t,n;return e=(e+"").trim().toLowerCase(),(t=APe.exec(e))?(n=t[1].length,t=parseInt(t[1],16),n===6?vH(t):n===3?new ys(t>>8&15|t>>4&240,t>>4&15|t&240,(t&15)<<4|t&15,1):n===8?oE(t>>24&255,t>>16&255,t>>8&255,(t&255)/255):n===4?oE(t>>12&15|t>>8&240,t>>8&15|t>>4&240,t>>4&15|t&240,((t&15)<<4|t&15)/255):null):(t=TPe.exec(e))?new ys(t[1],t[2],t[3],1):(t=kPe.exec(e))?new ys(t[1]*255/100,t[2]*255/100,t[3]*255/100,1):(t=PPe.exec(e))?oE(t[1],t[2],t[3],t[4]):(t=MPe.exec(e))?oE(t[1]*255/100,t[2]*255/100,t[3]*255/100,t[4]):(t=FPe.exec(e))?OH(t[1],t[2]/100,t[3]/100,1):(t=NPe.exec(e))?OH(t[1],t[2]/100,t[3]/100,t[4]):pH.hasOwnProperty(e)?vH(pH[e]):e==="transparent"?new ys(NaN,NaN,NaN,0):null}function vH(e){return new ys(e>>16&255,e>>8&255,e&255,1)}function oE(e,t,n,r){return r<=0&&(e=t=n=NaN),new ys(e,t,n,r)}function o0e(e){return e instanceof n4||(e=Ax(e)),e?(e=e.rgb(),new ys(e.r,e.g,e.b,e.opacity)):new ys}function b$(e,t,n,r){return arguments.length===1?o0e(e):new ys(e,t,n,r??1)}function ys(e,t,n,r){this.r=+e,this.g=+t,this.b=+n,this.opacity=+r}bw(ys,b$,NI(n4,{brighter(e){return e=e==null?k3:Math.pow(k3,e),new ys(this.r*e,this.g*e,this.b*e,this.opacity)},darker(e){return e=e==null?d1:Math.pow(d1,e),new ys(this.r*e,this.g*e,this.b*e,this.opacity)},rgb(){return this},clamp(){return new ys(Hb(this.r),Hb(this.g),Hb(this.b),O$(this.opacity))},displayable(){return-.5<=this.r&&this.r<255.5&&-.5<=this.g&&this.g<255.5&&-.5<=this.b&&this.b<255.5&&0<=this.opacity&&this.opacity<=1},hex:yH,formatHex:yH,formatHex8:BPe,formatRgb:bH,toString:bH}));function yH(){return`#${Qb(this.r)}${Qb(this.g)}${Qb(this.b)}`}function BPe(){return`#${Qb(this.r)}${Qb(this.g)}${Qb(this.b)}${Qb((isNaN(this.opacity)?1:this.opacity)*255)}`}function bH(){const e=O$(this.opacity);return`${e===1?"rgb(":"rgba("}${Hb(this.r)}, ${Hb(this.g)}, ${Hb(this.b)}${e===1?")":`, ${e})`}`}function O$(e){return isNaN(e)?1:Math.max(0,Math.min(1,e))}function Hb(e){return Math.max(0,Math.min(255,Math.round(e)||0))}function Qb(e){return e=Hb(e),(e<16?"0":"")+e.toString(16)}function OH(e,t,n,r){return r<=0?e=t=n=NaN:n<=0||n>=1?e=t=NaN:t<=0&&(e=NaN),new Wh(e,t,n,r)}function s0e(e){if(e instanceof Wh)return new Wh(e.h,e.s,e.l,e.opacity);if(e instanceof n4||(e=Ax(e)),!e)return new Wh;if(e instanceof Wh)return e;e=e.rgb();var t=e.r/255,n=e.g/255,r=e.b/255,i=Math.min(t,n,r),a=Math.max(t,n,r),u=NaN,l=a-i,o=(a+i)/2;return l?(t===a?u=(n-r)/l+(n<r)*6:n===a?u=(r-t)/l+2:u=(t-n)/l+4,l/=o<.5?a+i:2-a-i,u*=60):l=o>0&&o<1?0:u,new Wh(u,l,o,e.opacity)}function QPe(e,t,n,r){return arguments.length===1?s0e(e):new Wh(e,t,n,r??1)}function Wh(e,t,n,r){this.h=+e,this.s=+t,this.l=+n,this.opacity=+r}bw(Wh,QPe,NI(n4,{brighter(e){return e=e==null?k3:Math.pow(k3,e),new Wh(this.h,this.s,this.l*e,this.opacity)},darker(e){return e=e==null?d1:Math.pow(d1,e),new Wh(this.h,this.s,this.l*e,this.opacity)},rgb(){var e=this.h%360+(this.h<0)*360,t=isNaN(e)||isNaN(this.s)?0:this.s,n=this.l,r=n+(n<.5?n:1-n)*t,i=2*n-r;return new ys(iP(e>=240?e-240:e+120,i,r),iP(e,i,r),iP(e<120?e+240:e-120,i,r),this.opacity)},clamp(){return new Wh(xH(this.h),sE(this.s),sE(this.l),O$(this.opacity))},displayable(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1},formatHsl(){const e=O$(this.opacity);return`${e===1?"hsl(":"hsla("}${xH(this.h)}, ${sE(this.s)*100}%, ${sE(this.l)*100}%${e===1?")":`, ${e})`}`}}));function xH(e){return e=(e||0)%360,e<0?e+360:e}function sE(e){return Math.max(0,Math.min(1,e||0))}function iP(e,t,n){return(e<60?t+(n-t)*e/60:e<180?n:e<240?t+(n-t)*(240-e)/60:t)*255}const zPe=Math.PI/180,qPe=180/Math.PI;var c0e=-.14861,II=1.78277,LI=-.29227,Ow=-.90649,Tx=1.97294,SH=Tx*Ow,_H=Tx*II,EH=II*LI-Ow*c0e;function VPe(e){if(e instanceof Xb)return new Xb(e.h,e.s,e.l,e.opacity);e instanceof ys||(e=o0e(e));var t=e.r/255,n=e.g/255,r=e.b/255,i=(EH*r+SH*t-_H*n)/(EH+SH-_H),a=r-i,u=(Tx*(n-i)-LI*a)/Ow,l=Math.sqrt(u*u+a*a)/(Tx*i*(1-i)),o=l?Math.atan2(u,a)*qPe-120:NaN;return new Xb(o<0?o+360:o,l,i,e.opacity)}function ny(e,t,n,r){return arguments.length===1?VPe(e):new Xb(e,t,n,r??1)}function Xb(e,t,n,r){this.h=+e,this.s=+t,this.l=+n,this.opacity=+r}bw(Xb,ny,NI(n4,{brighter(e){return e=e==null?k3:Math.pow(k3,e),new Xb(this.h,this.s,this.l*e,this.opacity)},darker(e){return e=e==null?d1:Math.pow(d1,e),new Xb(this.h,this.s,this.l*e,this.opacity)},rgb(){var e=isNaN(this.h)?0:(this.h+120)*zPe,t=+this.l,n=isNaN(this.s)?0:this.s*t*(1-t),r=Math.cos(e),i=Math.sin(e);return new ys(255*(t+n*(c0e*r+II*i)),255*(t+n*(LI*r+Ow*i)),255*(t+n*(Tx*r)),this.opacity)}}));const xw=e=>()=>e;function f0e(e,t){return function(n){return e+n*t}}function UPe(e,t,n){return e=Math.pow(e,n),t=Math.pow(t,n)-e,n=1/n,function(r){return Math.pow(e+r*t,n)}}function jPe(e,t){var n=t-e;return n?f0e(e,n>180||n<-180?n-360*Math.round(n/360):n):xw(isNaN(e)?t:e)}function ZPe(e){return(e=+e)==1?p3:function(t,n){return n-t?UPe(t,n,e):xw(isNaN(t)?n:t)}}function p3(e,t){var n=t-e;return n?f0e(e,n):xw(isNaN(e)?t:e)}const DH=(function e(t){var n=ZPe(t);function r(i,a){var u=n((i=b$(i)).r,(a=b$(a)).r),l=n(i.g,a.g),o=n(i.b,a.b),s=p3(i.opacity,a.opacity);return function(c){return i.r=u(c),i.g=l(c),i.b=o(c),i.opacity=s(c),i+""}}return r.gamma=e,r})(1);function HPe(e,t){t||(t=[]);var n=e?Math.min(t.length,e.length):0,r=t.slice(),i;return function(a){for(i=0;i<n;++i)r[i]=e[i]*(1-a)+t[i]*a;return r}}function XPe(e){return ArrayBuffer.isView(e)&&!(e instanceof DataView)}function YPe(e,t){var n=t?t.length:0,r=e?Math.min(n,e.length):0,i=new Array(r),a=new Array(n),u;for(u=0;u<r;++u)i[u]=r4(e[u],t[u]);for(;u<n;++u)a[u]=t[u];return function(l){for(u=0;u<r;++u)a[u]=i[u](l);return a}}function GPe(e,t){var n=new Date;return e=+e,t=+t,function(r){return n.setTime(e*(1-r)+t*r),n}}function x$(e,t){return e=+e,t=+t,function(n){return e*(1-n)+t*n}}function KPe(e,t){var n={},r={},i;(e===null||typeof e!="object")&&(e={}),(t===null||typeof t!="object")&&(t={});for(i in t)i in e?n[i]=r4(e[i],t[i]):r[i]=t[i];return function(a){for(i in n)r[i]=n[i](a);return r}}var pM=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,aP=new RegExp(pM.source,"g");function WPe(e){return function(){return e}}function JPe(e){return function(t){return e(t)+""}}function e8e(e,t){var n=pM.lastIndex=aP.lastIndex=0,r,i,a,u=-1,l=[],o=[];for(e=e+"",t=t+"";(r=pM.exec(e))&&(i=aP.exec(t));)(a=i.index)>n&&(a=t.slice(n,a),l[u]?l[u]+=a:l[++u]=a),(r=r[0])===(i=i[0])?l[u]?l[u]+=i:l[++u]=i:(l[++u]=null,o.push({i:u,x:x$(r,i)})),n=aP.lastIndex;return n<t.length&&(a=t.slice(n),l[u]?l[u]+=a:l[++u]=a),l.length<2?o[0]?JPe(o[0].x):WPe(t):(t=o.length,function(s){for(var c=0,d;c<t;++c)l[(d=o[c]).i]=d.x(s);return l.join("")})}function r4(e,t){var n=typeof t,r;return t==null||n==="boolean"?xw(t):(n==="number"?x$:n==="string"?(r=Ax(t))?(t=r,DH):e8e:t instanceof Ax?DH:t instanceof Date?GPe:XPe(t)?HPe:Array.isArray(t)?YPe:typeof t.valueOf!="function"&&typeof t.toString!="function"||isNaN(t)?KPe:x$)(e,t)}function BI(e,t){return e=+e,t=+t,function(n){return Math.round(e*(1-n)+t*n)}}function d0e(e){return(function t(n){n=+n;function r(i,a){var u=e((i=ny(i)).h,(a=ny(a)).h),l=p3(i.s,a.s),o=p3(i.l,a.l),s=p3(i.opacity,a.opacity);return function(c){return i.h=u(c),i.s=l(c),i.l=o(Math.pow(c,n)),i.opacity=s(c),i+""}}return r.gamma=t,r})(1)}d0e(jPe);var h0e=d0e(p3);function t8e(e,t){t===void 0&&(t=e,e=r4);for(var n=0,r=t.length-1,i=t[0],a=new Array(r<0?0:r);n<r;)a[n]=e(i,i=t[++n]);return function(u){var l=Math.max(0,Math.min(r-1,Math.floor(u*=r)));return a[l](u-l)}}h0e(ny(-100,.75,.35),ny(80,1.5,.8));var iut=h0e(ny(260,.75,.35),ny(80,1.5,.8));ny();var cE=b$(),n8e=Math.PI/3,r8e=Math.PI*2/3;function aut(e){var t;return e=(.5-e)*Math.PI,cE.r=255*(t=Math.sin(e))*t,cE.g=255*(t=Math.sin(e+n8e))*t,cE.b=255*(t=Math.sin(e+r8e))*t,cE+""}function i8e(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function kx(e,t){return kx=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(n,r){return n.__proto__=r,n},kx(e,t)}function a8e(e,t){e.prototype=Object.create(t.prototype),e.prototype.constructor=e,kx(e,t)}function mM(e){return mM=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(t){return t.__proto__||Object.getPrototypeOf(t)},mM(e)}function u8e(e){try{return Function.toString.call(e).indexOf("[native code]")!==-1}catch{return typeof e=="function"}}function p0e(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(p0e=function(){return!!e})()}function l8e(e,t,n){if(p0e())return Reflect.construct.apply(null,arguments);var r=[null];r.push.apply(r,t);var i=new(e.bind.apply(e,r));return n&&kx(i,n.prototype),i}function gM(e){var t=typeof Map=="function"?new Map:void 0;return gM=function(r){if(r===null||!u8e(r))return r;if(typeof r!="function")throw new TypeError("Super expression must either be null or a function");if(t!==void 0){if(t.has(r))return t.get(r);t.set(r,i)}function i(){return l8e(r,arguments,mM(this).constructor)}return i.prototype=Object.create(r.prototype,{constructor:{value:i,enumerable:!1,writable:!0,configurable:!0}}),kx(i,r)},gM(e)}var O0=(function(e){a8e(t,e);function t(n){var r;return r=e.call(this,"An error occurred. See https://github.com/styled-components/polished/blob/main/src/internalHelpers/errors.md#"+n+" for more information.")||this,i8e(r)}return t})(gM(Error));function uP(e){return Math.round(e*255)}function o8e(e,t,n){return uP(e)+","+uP(t)+","+uP(n)}function Px(e,t,n,r){if(r===void 0&&(r=o8e),t===0)return r(n,n,n);var i=(e%360+360)%360/60,a=(1-Math.abs(2*n-1))*t,u=a*(1-Math.abs(i%2-1)),l=0,o=0,s=0;i>=0&&i<1?(l=a,o=u):i>=1&&i<2?(l=u,o=a):i>=2&&i<3?(o=a,s=u):i>=3&&i<4?(o=u,s=a):i>=4&&i<5?(l=u,s=a):i>=5&&i<6&&(l=a,s=u);var c=n-a/2,d=l+c,h=o+c,m=s+c;return r(d,h,m)}var $H={aliceblue:"f0f8ff",antiquewhite:"faebd7",aqua:"00ffff",aquamarine:"7fffd4",azure:"f0ffff",beige:"f5f5dc",bisque:"ffe4c4",black:"000",blanchedalmond:"ffebcd",blue:"0000ff",blueviolet:"8a2be2",brown:"a52a2a",burlywood:"deb887",cadetblue:"5f9ea0",chartreuse:"7fff00",chocolate:"d2691e",coral:"ff7f50",cornflowerblue:"6495ed",cornsilk:"fff8dc",crimson:"dc143c",cyan:"00ffff",darkblue:"00008b",darkcyan:"008b8b",darkgoldenrod:"b8860b",darkgray:"a9a9a9",darkgreen:"006400",darkgrey:"a9a9a9",darkkhaki:"bdb76b",darkmagenta:"8b008b",darkolivegreen:"556b2f",darkorange:"ff8c00",darkorchid:"9932cc",darkred:"8b0000",darksalmon:"e9967a",darkseagreen:"8fbc8f",darkslateblue:"483d8b",darkslategray:"2f4f4f",darkslategrey:"2f4f4f",darkturquoise:"00ced1",darkviolet:"9400d3",deeppink:"ff1493",deepskyblue:"00bfff",dimgray:"696969",dimgrey:"696969",dodgerblue:"1e90ff",firebrick:"b22222",floralwhite:"fffaf0",forestgreen:"228b22",fuchsia:"ff00ff",gainsboro:"dcdcdc",ghostwhite:"f8f8ff",gold:"ffd700",goldenrod:"daa520",gray:"808080",green:"008000",greenyellow:"adff2f",grey:"808080",honeydew:"f0fff0",hotpink:"ff69b4",indianred:"cd5c5c",indigo:"4b0082",ivory:"fffff0",khaki:"f0e68c",lavender:"e6e6fa",lavenderblush:"fff0f5",lawngreen:"7cfc00",lemonchiffon:"fffacd",lightblue:"add8e6",lightcoral:"f08080",lightcyan:"e0ffff",lightgoldenrodyellow:"fafad2",lightgray:"d3d3d3",lightgreen:"90ee90",lightgrey:"d3d3d3",lightpink:"ffb6c1",lightsalmon:"ffa07a",lightseagreen:"20b2aa",lightskyblue:"87cefa",lightslategray:"789",lightslategrey:"789",lightsteelblue:"b0c4de",lightyellow:"ffffe0",lime:"0f0",limegreen:"32cd32",linen:"faf0e6",magenta:"f0f",maroon:"800000",mediumaquamarine:"66cdaa",mediumblue:"0000cd",mediumorchid:"ba55d3",mediumpurple:"9370db",mediumseagreen:"3cb371",mediumslateblue:"7b68ee",mediumspringgreen:"00fa9a",mediumturquoise:"48d1cc",mediumvioletred:"c71585",midnightblue:"191970",mintcream:"f5fffa",mistyrose:"ffe4e1",moccasin:"ffe4b5",navajowhite:"ffdead",navy:"000080",oldlace:"fdf5e6",olive:"808000",olivedrab:"6b8e23",orange:"ffa500",orangered:"ff4500",orchid:"da70d6",palegoldenrod:"eee8aa",palegreen:"98fb98",paleturquoise:"afeeee",palevioletred:"db7093",papayawhip:"ffefd5",peachpuff:"ffdab9",peru:"cd853f",pink:"ffc0cb",plum:"dda0dd",powderblue:"b0e0e6",purple:"800080",rebeccapurple:"639",red:"f00",rosybrown:"bc8f8f",royalblue:"4169e1",saddlebrown:"8b4513",salmon:"fa8072",sandybrown:"f4a460",seagreen:"2e8b57",seashell:"fff5ee",sienna:"a0522d",silver:"c0c0c0",skyblue:"87ceeb",slateblue:"6a5acd",slategray:"708090",slategrey:"708090",snow:"fffafa",springgreen:"00ff7f",steelblue:"4682b4",tan:"d2b48c",teal:"008080",thistle:"d8bfd8",tomato:"ff6347",turquoise:"40e0d0",violet:"ee82ee",wheat:"f5deb3",white:"fff",whitesmoke:"f5f5f5",yellow:"ff0",yellowgreen:"9acd32"};function s8e(e){if(typeof e!="string")return e;var t=e.toLowerCase();return $H[t]?"#"+$H[t]:e}var c8e=/^#[a-fA-F0-9]{6}$/,f8e=/^#[a-fA-F0-9]{8}$/,d8e=/^#[a-fA-F0-9]{3}$/,h8e=/^#[a-fA-F0-9]{4}$/,lP=/^rgb\(\s*(\d{1,3})\s*(?:,)?\s*(\d{1,3})\s*(?:,)?\s*(\d{1,3})\s*\)$/i,p8e=/^rgb(?:a)?\(\s*(\d{1,3})\s*(?:,)?\s*(\d{1,3})\s*(?:,)?\s*(\d{1,3})\s*(?:,|\/)\s*([-+]?\d*[.]?\d+[%]?)\s*\)$/i,m8e=/^hsl\(\s*(\d{0,3}[.]?[0-9]+(?:deg)?)\s*(?:,)?\s*(\d{1,3}[.]?[0-9]?)%\s*(?:,)?\s*(\d{1,3}[.]?[0-9]?)%\s*\)$/i,g8e=/^hsl(?:a)?\(\s*(\d{0,3}[.]?[0-9]+(?:deg)?)\s*(?:,)?\s*(\d{1,3}[.]?[0-9]?)%\s*(?:,)?\s*(\d{1,3}[.]?[0-9]?)%\s*(?:,|\/)\s*([-+]?\d*[.]?\d+[%]?)\s*\)$/i;function h1(e){if(typeof e!="string")throw new O0(3);var t=s8e(e);if(t.match(c8e))return{red:parseInt(""+t[1]+t[2],16),green:parseInt(""+t[3]+t[4],16),blue:parseInt(""+t[5]+t[6],16)};if(t.match(f8e)){var n=parseFloat((parseInt(""+t[7]+t[8],16)/255).toFixed(2));return{red:parseInt(""+t[1]+t[2],16),green:parseInt(""+t[3]+t[4],16),blue:parseInt(""+t[5]+t[6],16),alpha:n}}if(t.match(d8e))return{red:parseInt(""+t[1]+t[1],16),green:parseInt(""+t[2]+t[2],16),blue:parseInt(""+t[3]+t[3],16)};if(t.match(h8e)){var r=parseFloat((parseInt(""+t[4]+t[4],16)/255).toFixed(2));return{red:parseInt(""+t[1]+t[1],16),green:parseInt(""+t[2]+t[2],16),blue:parseInt(""+t[3]+t[3],16),alpha:r}}var i=lP.exec(t);if(i)return{red:parseInt(""+i[1],10),green:parseInt(""+i[2],10),blue:parseInt(""+i[3],10)};var a=p8e.exec(t.substring(0,50));if(a)return{red:parseInt(""+a[1],10),green:parseInt(""+a[2],10),blue:parseInt(""+a[3],10),alpha:parseFloat(""+a[4])>1?parseFloat(""+a[4])/100:parseFloat(""+a[4])};var u=m8e.exec(t);if(u){var l=parseInt(""+u[1],10),o=parseInt(""+u[2],10)/100,s=parseInt(""+u[3],10)/100,c="rgb("+Px(l,o,s)+")",d=lP.exec(c);if(!d)throw new O0(4,t,c);return{red:parseInt(""+d[1],10),green:parseInt(""+d[2],10),blue:parseInt(""+d[3],10)}}var h=g8e.exec(t.substring(0,50));if(h){var m=parseInt(""+h[1],10),y=parseInt(""+h[2],10)/100,v=parseInt(""+h[3],10)/100,S="rgb("+Px(m,y,v)+")",_=lP.exec(S);if(!_)throw new O0(4,t,S);return{red:parseInt(""+_[1],10),green:parseInt(""+_[2],10),blue:parseInt(""+_[3],10),alpha:parseFloat(""+h[4])>1?parseFloat(""+h[4])/100:parseFloat(""+h[4])}}throw new O0(5)}function v8e(e){var t=e.red/255,n=e.green/255,r=e.blue/255,i=Math.max(t,n,r),a=Math.min(t,n,r),u=(i+a)/2;if(i===a)return e.alpha!==void 0?{hue:0,saturation:0,lightness:u,alpha:e.alpha}:{hue:0,saturation:0,lightness:u};var l,o=i-a,s=u>.5?o/(2-i-a):o/(i+a);switch(i){case t:l=(n-r)/o+(n<r?6:0);break;case n:l=(r-t)/o+2;break;default:l=(t-n)/o+4;break}return l*=60,e.alpha!==void 0?{hue:l,saturation:s,lightness:u,alpha:e.alpha}:{hue:l,saturation:s,lightness:u}}function by(e){return v8e(h1(e))}var y8e=function(t){return t.length===7&&t[1]===t[2]&&t[3]===t[4]&&t[5]===t[6]?"#"+t[1]+t[3]+t[5]:t},vM=y8e;function $b(e){var t=e.toString(16);return t.length===1?"0"+t:t}function oP(e){return $b(Math.round(e*255))}function b8e(e,t,n){return vM("#"+oP(e)+oP(t)+oP(n))}function S$(e,t,n){return Px(e,t,n,b8e)}function O8e(e,t,n){if(typeof e=="number"&&typeof t=="number"&&typeof n=="number")return S$(e,t,n);if(typeof e=="object"&&t===void 0&&n===void 0)return S$(e.hue,e.saturation,e.lightness);throw new O0(1)}function x8e(e,t,n,r){if(typeof e=="number"&&typeof t=="number"&&typeof n=="number"&&typeof r=="number")return r>=1?S$(e,t,n):"rgba("+Px(e,t,n)+","+r+")";if(typeof e=="object"&&t===void 0&&n===void 0&&r===void 0)return e.alpha>=1?S$(e.hue,e.saturation,e.lightness):"rgba("+Px(e.hue,e.saturation,e.lightness)+","+e.alpha+")";throw new O0(2)}function yM(e,t,n){if(typeof e=="number"&&typeof t=="number"&&typeof n=="number")return vM("#"+$b(e)+$b(t)+$b(n));if(typeof e=="object"&&t===void 0&&n===void 0)return vM("#"+$b(e.red)+$b(e.green)+$b(e.blue));throw new O0(6)}function Sw(e,t,n,r){if(typeof e=="string"&&typeof t=="number"){var i=h1(e);return"rgba("+i.red+","+i.green+","+i.blue+","+t+")"}else{if(typeof e=="number"&&typeof t=="number"&&typeof n=="number"&&typeof r=="number")return r>=1?yM(e,t,n):"rgba("+e+","+t+","+n+","+r+")";if(typeof e=="object"&&t===void 0&&n===void 0&&r===void 0)return e.alpha>=1?yM(e.red,e.green,e.blue):"rgba("+e.red+","+e.green+","+e.blue+","+e.alpha+")"}throw new O0(7)}var S8e=function(t){return typeof t.red=="number"&&typeof t.green=="number"&&typeof t.blue=="number"&&(typeof t.alpha!="number"||typeof t.alpha>"u")},_8e=function(t){return typeof t.red=="number"&&typeof t.green=="number"&&typeof t.blue=="number"&&typeof t.alpha=="number"},E8e=function(t){return typeof t.hue=="number"&&typeof t.saturation=="number"&&typeof t.lightness=="number"&&(typeof t.alpha!="number"||typeof t.alpha>"u")},D8e=function(t){return typeof t.hue=="number"&&typeof t.saturation=="number"&&typeof t.lightness=="number"&&typeof t.alpha=="number"};function Oy(e){if(typeof e!="object")throw new O0(8);if(_8e(e))return Sw(e);if(S8e(e))return yM(e);if(D8e(e))return x8e(e);if(E8e(e))return O8e(e);throw new O0(8)}function m0e(e,t,n){return function(){var i=n.concat(Array.prototype.slice.call(arguments));return i.length>=t?e.apply(this,i):m0e(e,t,i)}}function nd(e){return m0e(e,e.length,[])}function $8e(e,t){if(t==="transparent")return t;var n=by(t);return Oy(sf({},n,{hue:n.hue+parseFloat(e)}))}nd($8e);function i4(e,t,n){return Math.max(e,Math.min(t,n))}function C8e(e,t){if(t==="transparent")return t;var n=by(t);return Oy(sf({},n,{lightness:i4(0,1,n.lightness-parseFloat(e))}))}var w8e=nd(C8e),uut=w8e;function R8e(e,t){if(t==="transparent")return t;var n=by(t);return Oy(sf({},n,{saturation:i4(0,1,n.saturation-parseFloat(e))}))}nd(R8e);function CH(e){if(e==="transparent")return 0;var t=h1(e),n=Object.keys(t).map(function(u){var l=t[u]/255;return l<=.03928?l/12.92:Math.pow((l+.055)/1.055,2.4)}),r=n[0],i=n[1],a=n[2];return parseFloat((.2126*r+.7152*i+.0722*a).toFixed(3))}function lut(e,t){var n=CH(e),r=CH(t);return parseFloat((n>r?(n+.05)/(r+.05):(r+.05)/(n+.05)).toFixed(2))}function A8e(e,t){if(t==="transparent")return t;var n=by(t);return Oy(sf({},n,{lightness:i4(0,1,n.lightness+parseFloat(e))}))}var T8e=nd(A8e),out=T8e;function k8e(e,t,n){if(t==="transparent")return n;if(n==="transparent")return t;if(e===0)return n;var r=h1(t),i=sf({},r,{alpha:typeof r.alpha=="number"?r.alpha:1}),a=h1(n),u=sf({},a,{alpha:typeof a.alpha=="number"?a.alpha:1}),l=i.alpha-u.alpha,o=parseFloat(e)*2-1,s=o*l===-1?o:o+l,c=1+o*l,d=(s/c+1)/2,h=1-d,m={red:Math.floor(i.red*d+u.red*h),green:Math.floor(i.green*d+u.green*h),blue:Math.floor(i.blue*d+u.blue*h),alpha:i.alpha*parseFloat(e)+u.alpha*(1-parseFloat(e))};return Sw(m)}var P8e=nd(k8e),g0e=P8e;function M8e(e,t){if(t==="transparent")return t;var n=h1(t),r=typeof n.alpha=="number"?n.alpha:1,i=sf({},n,{alpha:i4(0,1,(r*100+parseFloat(e)*100)/100)});return Sw(i)}nd(M8e);function F8e(e,t){if(t==="transparent")return t;var n=by(t);return Oy(sf({},n,{saturation:i4(0,1,n.saturation+parseFloat(e))}))}nd(F8e);function N8e(e,t){return t==="transparent"?t:Oy(sf({},by(t),{hue:parseFloat(e)}))}nd(N8e);function I8e(e,t){return t==="transparent"?t:Oy(sf({},by(t),{lightness:parseFloat(e)}))}nd(I8e);function L8e(e,t){return t==="transparent"?t:Oy(sf({},by(t),{saturation:parseFloat(e)}))}nd(L8e);function B8e(e,t){return t==="transparent"?t:g0e(parseFloat(e),"rgb(0, 0, 0)",t)}var Q8e=nd(B8e),sut=Q8e;function z8e(e,t){return t==="transparent"?t:g0e(parseFloat(e),"rgb(255, 255, 255)",t)}nd(z8e);function q8e(e,t){if(t==="transparent")return t;var n=h1(t),r=typeof n.alpha=="number"?n.alpha:1,i=sf({},n,{alpha:i4(0,1,+(r*100-parseFloat(e)*100).toFixed(2)/100)});return Sw(i)}var V8e=nd(q8e),cut=V8e;function U8e(e){return Math.abs(e=Math.round(e))>=1e21?e.toLocaleString("en").replace(/,/g,""):e.toString(10)}function _$(e,t){if((n=(e=t?e.toExponential(t-1):e.toExponential()).indexOf("e"))<0)return null;var n,r=e.slice(0,n);return[r.length>1?r[0]+r.slice(2):r,+e.slice(n+1)]}function P3(e){return e=_$(Math.abs(e)),e?e[1]:NaN}function j8e(e,t){return function(n,r){for(var i=n.length,a=[],u=0,l=e[0],o=0;i>0&&l>0&&(o+l+1>r&&(l=Math.max(1,r-o)),a.push(n.substring(i-=l,i+l)),!((o+=l+1)>r));)l=e[u=(u+1)%e.length];return a.reverse().join(t)}}function Z8e(e){return function(t){return t.replace(/[0-9]/g,function(n){return e[+n]})}}var H8e=/^(?:(.)?([<>=^]))?([+\-( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?(~)?([a-z%])?$/i;function Mx(e){if(!(t=H8e.exec(e)))throw new Error("invalid format: "+e);var t;return new QI({fill:t[1],align:t[2],sign:t[3],symbol:t[4],zero:t[5],width:t[6],comma:t[7],precision:t[8]&&t[8].slice(1),trim:t[9],type:t[10]})}Mx.prototype=QI.prototype;function QI(e){this.fill=e.fill===void 0?" ":e.fill+"",this.align=e.align===void 0?">":e.align+"",this.sign=e.sign===void 0?"-":e.sign+"",this.symbol=e.symbol===void 0?"":e.symbol+"",this.zero=!!e.zero,this.width=e.width===void 0?void 0:+e.width,this.comma=!!e.comma,this.precision=e.precision===void 0?void 0:+e.precision,this.trim=!!e.trim,this.type=e.type===void 0?"":e.type+""}QI.prototype.toString=function(){return this.fill+this.align+this.sign+this.symbol+(this.zero?"0":"")+(this.width===void 0?"":Math.max(1,this.width|0))+(this.comma?",":"")+(this.precision===void 0?"":"."+Math.max(0,this.precision|0))+(this.trim?"~":"")+this.type};function X8e(e){e:for(var t=e.length,n=1,r=-1,i;n<t;++n)switch(e[n]){case".":r=i=n;break;case"0":r===0&&(r=n),i=n;break;default:if(!+e[n])break e;r>0&&(r=0);break}return r>0?e.slice(0,r)+e.slice(i+1):e}var v0e;function Y8e(e,t){var n=_$(e,t);if(!n)return e+"";var r=n[0],i=n[1],a=i-(v0e=Math.max(-8,Math.min(8,Math.floor(i/3)))*3)+1,u=r.length;return a===u?r:a>u?r+new Array(a-u+1).join("0"):a>0?r.slice(0,a)+"."+r.slice(a):"0."+new Array(1-a).join("0")+_$(e,Math.max(0,t+a-1))[0]}function wH(e,t){var n=_$(e,t);if(!n)return e+"";var r=n[0],i=n[1];return i<0?"0."+new Array(-i).join("0")+r:r.length>i+1?r.slice(0,i+1)+"."+r.slice(i+1):r+new Array(i-r.length+2).join("0")}const RH={"%":(e,t)=>(e*100).toFixed(t),b:e=>Math.round(e).toString(2),c:e=>e+"",d:U8e,e:(e,t)=>e.toExponential(t),f:(e,t)=>e.toFixed(t),g:(e,t)=>e.toPrecision(t),o:e=>Math.round(e).toString(8),p:(e,t)=>wH(e*100,t),r:wH,s:Y8e,X:e=>Math.round(e).toString(16).toUpperCase(),x:e=>Math.round(e).toString(16)};function AH(e){return e}var TH=Array.prototype.map,kH=["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"];function G8e(e){var t=e.grouping===void 0||e.thousands===void 0?AH:j8e(TH.call(e.grouping,Number),e.thousands+""),n=e.currency===void 0?"":e.currency[0]+"",r=e.currency===void 0?"":e.currency[1]+"",i=e.decimal===void 0?".":e.decimal+"",a=e.numerals===void 0?AH:Z8e(TH.call(e.numerals,String)),u=e.percent===void 0?"%":e.percent+"",l=e.minus===void 0?"−":e.minus+"",o=e.nan===void 0?"NaN":e.nan+"";function s(d){d=Mx(d);var h=d.fill,m=d.align,y=d.sign,v=d.symbol,S=d.zero,_=d.width,E=d.comma,$=d.precision,k=d.trim,A=d.type;A==="n"?(E=!0,A="g"):RH[A]||($===void 0&&($=12),k=!0,A="g"),(S||h==="0"&&m==="=")&&(S=!0,h="0",m="=");var w=v==="$"?n:v==="#"&&/[boxX]/.test(A)?"0"+A.toLowerCase():"",R=v==="$"?r:/[%p]/.test(A)?u:"",F=RH[A],I=/[defgprs%]/.test(A);$=$===void 0?6:/[gprs]/.test(A)?Math.max(1,Math.min(21,$)):Math.max(0,Math.min(20,$));function U(q){var V=w,N=R,j,W,K;if(A==="c")N=F(q)+N,q="";else{q=+q;var J=q<0||1/q<0;if(q=isNaN(q)?o:F(Math.abs(q),$),k&&(q=X8e(q)),J&&+q==0&&y!=="+"&&(J=!1),V=(J?y==="("?y:l:y==="-"||y==="("?"":y)+V,N=(A==="s"?kH[8+v0e/3]:"")+N+(J&&y==="("?")":""),I){for(j=-1,W=q.length;++j<W;)if(K=q.charCodeAt(j),48>K||K>57){N=(K===46?i+q.slice(j+1):q.slice(j))+N,q=q.slice(0,j);break}}}E&&!S&&(q=t(q,1/0));var te=V.length+q.length+N.length,H=te<_?new Array(_-te+1).join(h):"";switch(E&&S&&(q=t(H+q,H.length?_-N.length:1/0),H=""),m){case"<":q=V+q+N+H;break;case"=":q=V+H+q+N;break;case"^":q=H.slice(0,te=H.length>>1)+V+q+N+H.slice(te);break;default:q=H+V+q+N;break}return a(q)}return U.toString=function(){return d+""},U}function c(d,h){var m=s((d=Mx(d),d.type="f",d)),y=Math.max(-8,Math.min(8,Math.floor(P3(h)/3)))*3,v=Math.pow(10,-y),S=kH[8+y/3];return function(_){return m(v*_)+S}}return{format:s,formatPrefix:c}}var fE,zI,y0e;K8e({thousands:",",grouping:[3],currency:["$",""]});function K8e(e){return fE=G8e(e),zI=fE.format,y0e=fE.formatPrefix,fE}function W8e(e){return Math.max(0,-P3(Math.abs(e)))}function J8e(e,t){return Math.max(0,Math.max(-8,Math.min(8,Math.floor(P3(t)/3)))*3-P3(Math.abs(e)))}function e7e(e,t){return e=Math.abs(e),t=Math.abs(t)-e,Math.max(0,P3(t)-P3(e))+1}const t7e=oN({String:st.string,Number:st.number,"True False":st.bool,PropertyName:st.propertyName,Null:st.null,", :":st.separator,"[ ]":st.squareBracket,"{ }":st.brace}),fut=w3.deserialize({version:14,states:"$bOVQPOOOOQO'#Cb'#CbOnQPO'#CeOvQPO'#ClOOQO'#Cr'#CrQOQPOOOOQO'#Cg'#CgO}QPO'#CfO!SQPO'#CtOOQO,59P,59PO![QPO,59PO!aQPO'#CuOOQO,59W,59WO!iQPO,59WOVQPO,59QOqQPO'#CmO!nQPO,59`OOQO1G.k1G.kOVQPO'#CnO!vQPO,59aOOQO1G.r1G.rOOQO1G.l1G.lOOQO,59X,59XOOQO-E6k-E6kOOQO,59Y,59YOOQO-E6l-E6l",stateData:"#O~OeOS~OQSORSOSSOTSOWQO_ROgPO~OVXOgUO~O^[O~PVO[^O~O]_OVhX~OVaO~O]bO^iX~O^dO~O]_OVha~O]bO^ia~O",goto:"!kjPPPPPPkPPkqwPPPPk{!RPPP!XP!e!hXSOR^bQWQRf_TVQ_Q`WRg`QcZRicQTOQZRQe^RhbRYQR]R",nodeNames:"⚠ JsonText True False Null Number String } { Object Property PropertyName : , ] [ Array",maxTerm:25,nodeProps:[["isolate",-2,6,11,""],["openedBy",7,"{",14,"["],["closedBy",8,"}",15,"]"]],propSources:[t7e],skippedNodes:[0],repeatNodeCount:2,tokenData:"(|~RaXY!WYZ!W]^!Wpq!Wrs!]|}$u}!O$z!Q!R%T!R![&c![!]&t!}#O&y#P#Q'O#Y#Z'T#b#c'r#h#i(Z#o#p(r#q#r(w~!]Oe~~!`Wpq!]qr!]rs!xs#O!]#O#P!}#P;'S!];'S;=`$o<%lO!]~!}Og~~#QXrs!]!P!Q!]#O#P!]#U#V!]#Y#Z!]#b#c!]#f#g!]#h#i!]#i#j#m~#pR!Q![#y!c!i#y#T#Z#y~#|R!Q![$V!c!i$V#T#Z$V~$YR!Q![$c!c!i$c#T#Z$c~$fR!Q![!]!c!i!]#T#Z!]~$rP;=`<%l!]~$zO]~~$}Q!Q!R%T!R![&c~%YRT~!O!P%c!g!h%w#X#Y%w~%fP!Q![%i~%nRT~!Q![%i!g!h%w#X#Y%w~%zR{|&T}!O&T!Q![&Z~&WP!Q![&Z~&`PT~!Q![&Z~&hST~!O!P%c!Q![&c!g!h%w#X#Y%w~&yO[~~'OO_~~'TO^~~'WP#T#U'Z~'^P#`#a'a~'dP#g#h'g~'jP#X#Y'm~'rOR~~'uP#i#j'x~'{P#`#a(O~(RP#`#a(U~(ZOS~~(^P#f#g(a~(dP#i#j(g~(jP#X#Y(m~(rOQ~~(wOW~~(|OV~",tokenizers:[0],topRules:{JsonText:[0,1]},tokenPrec:0});function ry(){return ry=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)({}).hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},ry.apply(null,arguments)}function dut(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.indexOf(r)!==-1)continue;n[r]=e[r]}return n}const n7e=1024;let r7e=0;class sP{constructor(t,n){this.from=t,this.to=n}}class du{constructor(t={}){this.id=r7e++,this.perNode=!!t.perNode,this.deserialize=t.deserialize||(()=>{throw new Error("This node type doesn't define a deserialize function")}),this.combine=t.combine||null}add(t){if(this.perNode)throw new RangeError("Can't add per-node props to node types");return typeof t!="function"&&(t=dg.match(t)),n=>{let r=t(n);return r===void 0?null:[this,r]}}}du.closedBy=new du({deserialize:e=>e.split(" ")});du.openedBy=new du({deserialize:e=>e.split(" ")});du.group=new du({deserialize:e=>e.split(" ")});du.isolate=new du({deserialize:e=>{if(e&&e!="rtl"&&e!="ltr"&&e!="auto")throw new RangeError("Invalid value for isolate: "+e);return e||"auto"}});du.contextHash=new du({perNode:!0});du.lookAhead=new du({perNode:!0});du.mounted=new du({perNode:!0});class K2{constructor(t,n,r,i=!1){this.tree=t,this.overlay=n,this.parser=r,this.bracketed=i}static get(t){return t&&t.props&&t.props[du.mounted.id]}}const i7e=Object.create(null);class dg{constructor(t,n,r,i=0){this.name=t,this.props=n,this.id=r,this.flags=i}static define(t){let n=t.props&&t.props.length?Object.create(null):i7e,r=(t.top?1:0)|(t.skipped?2:0)|(t.error?4:0)|(t.name==null?8:0),i=new dg(t.name||"",n,t.id,r);if(t.props){for(let a of t.props)if(Array.isArray(a)||(a=a(i)),a){if(a[0].perNode)throw new RangeError("Can't store a per-node prop on a node type");n[a[0].id]=a[1]}}return i}prop(t){return this.props[t.id]}get isTop(){return(this.flags&1)>0}get isSkipped(){return(this.flags&2)>0}get isError(){return(this.flags&4)>0}get isAnonymous(){return(this.flags&8)>0}is(t){if(typeof t=="string"){if(this.name==t)return!0;let n=this.prop(du.group);return n?n.indexOf(t)>-1:!1}return this.id==t}static match(t){let n=Object.create(null);for(let r in t)for(let i of r.split(" "))n[i]=t[r];return r=>{for(let i=r.prop(du.group),a=-1;a<(i?i.length:0);a++){let u=n[a<0?r.name:i[a]];if(u)return u}}}}dg.none=new dg("",Object.create(null),0,8);const dE=new WeakMap,PH=new WeakMap;var Tl;(function(e){e[e.ExcludeBuffers=1]="ExcludeBuffers",e[e.IncludeAnonymous=2]="IncludeAnonymous",e[e.IgnoreMounts=4]="IgnoreMounts",e[e.IgnoreOverlays=8]="IgnoreOverlays",e[e.EnterBracketed=16]="EnterBracketed"})(Tl||(Tl={}));class rp{constructor(t,n,r,i,a){if(this.type=t,this.children=n,this.positions=r,this.length=i,this.props=null,a&&a.length){this.props=Object.create(null);for(let[u,l]of a)this.props[typeof u=="number"?u:u.id]=l}}toString(){let t=K2.get(this);if(t&&!t.overlay)return t.tree.toString();let n="";for(let r of this.children){let i=r.toString();i&&(n&&(n+=","),n+=i)}return this.type.name?(/\W/.test(this.type.name)&&!this.type.isError?JSON.stringify(this.type.name):this.type.name)+(n.length?"("+n+")":""):n}cursor(t=0){return new OM(this.topNode,t)}cursorAt(t,n=0,r=0){let i=dE.get(this)||this.topNode,a=new OM(i);return a.moveTo(t,n),dE.set(this,a._tree),a}get topNode(){return new Kd(this,0,0,null)}resolve(t,n=0){let r=Fx(dE.get(this)||this.topNode,t,n,!1);return dE.set(this,r),r}resolveInner(t,n=0){let r=Fx(PH.get(this)||this.topNode,t,n,!0);return PH.set(this,r),r}resolveStack(t,n=0){return l7e(this,t,n)}iterate(t){let{enter:n,leave:r,from:i=0,to:a=this.length}=t,u=t.mode||0,l=(u&Tl.IncludeAnonymous)>0;for(let o=this.cursor(u|Tl.IncludeAnonymous);;){let s=!1;if(o.from<=a&&o.to>=i&&(!l&&o.type.isAnonymous||n(o)!==!1)){if(o.firstChild())continue;s=!0}for(;s&&r&&(l||!o.type.isAnonymous)&&r(o),!o.nextSibling();){if(!o.parent())return;s=!0}}}prop(t){return t.perNode?this.props?this.props[t.id]:void 0:this.type.prop(t)}get propValues(){let t=[];if(this.props)for(let n in this.props)t.push([+n,this.props[n]]);return t}balance(t={}){return this.children.length<=8?this:UI(dg.none,this.children,this.positions,0,this.children.length,0,this.length,(n,r,i)=>new rp(this.type,n,r,i,this.propValues),t.makeTree||((n,r,i)=>new rp(dg.none,n,r,i)))}static build(t){return o7e(t)}}rp.empty=new rp(dg.none,[],[],0);class qI{constructor(t,n){this.buffer=t,this.index=n}get id(){return this.buffer[this.index-4]}get start(){return this.buffer[this.index-3]}get end(){return this.buffer[this.index-2]}get size(){return this.buffer[this.index-1]}get pos(){return this.index}next(){this.index-=4}fork(){return new qI(this.buffer,this.index)}}class iy{constructor(t,n,r){this.buffer=t,this.length=n,this.set=r}get type(){return dg.none}toString(){let t=[];for(let n=0;n<this.buffer.length;)t.push(this.childString(n)),n=this.buffer[n+3];return t.join(",")}childString(t){let n=this.buffer[t],r=this.buffer[t+3],i=this.set.types[n],a=i.name;if(/\W/.test(a)&&!i.isError&&(a=JSON.stringify(a)),t+=4,r==t)return a;let u=[];for(;t<r;)u.push(this.childString(t)),t=this.buffer[t+3];return a+"("+u.join(",")+")"}findChild(t,n,r,i,a){let{buffer:u}=this,l=-1;for(let o=t;o!=n&&!(b0e(a,i,u[o+1],u[o+2])&&(l=o,r>0));o=u[o+3]);return l}slice(t,n,r){let i=this.buffer,a=new Uint16Array(n-t),u=0;for(let l=t,o=0;l<n;){a[o++]=i[l++],a[o++]=i[l++]-r;let s=a[o++]=i[l++]-r;a[o++]=i[l++]-t,u=Math.max(u,s)}return new iy(a,u,this.set)}}function b0e(e,t,n,r){switch(e){case-2:return n<t;case-1:return r>=t&&n<t;case 0:return n<t&&r>t;case 1:return n<=t&&r>t;case 2:return r>t;case 4:return!0}}function Fx(e,t,n,r){for(var i;e.from==e.to||(n<1?e.from>=t:e.from>t)||(n>-1?e.to<=t:e.to<t);){let u=!r&&e instanceof Kd&&e.index<0?null:e.parent;if(!u)return e;e=u}let a=r?0:Tl.IgnoreOverlays;if(r)for(let u=e,l=u.parent;l;u=l,l=u.parent)u instanceof Kd&&u.index<0&&((i=l.enter(t,n,a))===null||i===void 0?void 0:i.from)!=u.from&&(e=l);for(;;){let u=e.enter(t,n,a);if(!u)return e;e=u}}class O0e{cursor(t=0){return new OM(this,t)}getChild(t,n=null,r=null){let i=MH(this,t,n,r);return i.length?i[0]:null}getChildren(t,n=null,r=null){return MH(this,t,n,r)}resolve(t,n=0){return Fx(this,t,n,!1)}resolveInner(t,n=0){return Fx(this,t,n,!0)}matchContext(t){return bM(this.parent,t)}enterUnfinishedNodesBefore(t){let n=this.childBefore(t),r=this;for(;n;){let i=n.lastChild;if(!i||i.to!=n.to)break;i.type.isError&&i.from==i.to?(r=n,n=i.prevSibling):n=i}return r}get node(){return this}get next(){return this.parent}}class Kd extends O0e{constructor(t,n,r,i){super(),this._tree=t,this.from=n,this.index=r,this._parent=i}get type(){return this._tree.type}get name(){return this._tree.type.name}get to(){return this.from+this._tree.length}nextChild(t,n,r,i,a=0){var u;for(let l=this;;){for(let{children:o,positions:s}=l._tree,c=n>0?o.length:-1;t!=c;t+=n){let d=o[t],h=s[t]+l.from;if(!(!(a&Tl.EnterBracketed&&d instanceof rp&&((u=K2.get(d))===null||u===void 0?void 0:u.overlay)===null&&(h>=r||h+d.length<=r))&&!b0e(i,r,h,h+d.length))){if(d instanceof iy){if(a&Tl.ExcludeBuffers)continue;let m=d.findChild(0,d.buffer.length,n,r-h,i);if(m>-1)return new Qv(new a7e(l,d,t,h),null,m)}else if(a&Tl.IncludeAnonymous||!d.type.isAnonymous||VI(d)){let m;if(!(a&Tl.IgnoreMounts)&&(m=K2.get(d))&&!m.overlay)return new Kd(m.tree,h,t,l);let y=new Kd(d,h,t,l);return a&Tl.IncludeAnonymous||!y.type.isAnonymous?y:y.nextChild(n<0?d.children.length-1:0,n,r,i,a)}}}if(a&Tl.IncludeAnonymous||!l.type.isAnonymous||(l.index>=0?t=l.index+n:t=n<0?-1:l._parent._tree.children.length,l=l._parent,!l))return null}}get firstChild(){return this.nextChild(0,1,0,4)}get lastChild(){return this.nextChild(this._tree.children.length-1,-1,0,4)}childAfter(t){return this.nextChild(0,1,t,2)}childBefore(t){return this.nextChild(this._tree.children.length-1,-1,t,-2)}prop(t){return this._tree.prop(t)}enter(t,n,r=0){let i;if(!(r&Tl.IgnoreOverlays)&&(i=K2.get(this._tree))&&i.overlay){let a=t-this.from,u=r&Tl.EnterBracketed&&i.bracketed;for(let{from:l,to:o}of i.overlay)if((n>0||u?l<=a:l<a)&&(n<0||u?o>=a:o>a))return new Kd(i.tree,i.overlay[0].from+this.from,-1,this)}return this.nextChild(0,1,t,n,r)}nextSignificantParent(){let t=this;for(;t.type.isAnonymous&&t._parent;)t=t._parent;return t}get parent(){return this._parent?this._parent.nextSignificantParent():null}get nextSibling(){return this._parent&&this.index>=0?this._parent.nextChild(this.index+1,1,0,4):null}get prevSibling(){return this._parent&&this.index>=0?this._parent.nextChild(this.index-1,-1,0,4):null}get tree(){return this._tree}toTree(){return this._tree}toString(){return this._tree.toString()}}function MH(e,t,n,r){let i=e.cursor(),a=[];if(!i.firstChild())return a;if(n!=null){for(let u=!1;!u;)if(u=i.type.is(n),!i.nextSibling())return a}for(;;){if(r!=null&&i.type.is(r))return a;if(i.type.is(t)&&a.push(i.node),!i.nextSibling())return r==null?a:[]}}function bM(e,t,n=t.length-1){for(let r=e;n>=0;r=r.parent){if(!r)return!1;if(!r.type.isAnonymous){if(t[n]&&t[n]!=r.name)return!1;n--}}return!0}class a7e{constructor(t,n,r,i){this.parent=t,this.buffer=n,this.index=r,this.start=i}}class Qv extends O0e{get name(){return this.type.name}get from(){return this.context.start+this.context.buffer.buffer[this.index+1]}get to(){return this.context.start+this.context.buffer.buffer[this.index+2]}constructor(t,n,r){super(),this.context=t,this._parent=n,this.index=r,this.type=t.buffer.set.types[t.buffer.buffer[r]]}child(t,n,r){let{buffer:i}=this.context,a=i.findChild(this.index+4,i.buffer[this.index+3],t,n-this.context.start,r);return a<0?null:new Qv(this.context,this,a)}get firstChild(){return this.child(1,0,4)}get lastChild(){return this.child(-1,0,4)}childAfter(t){return this.child(1,t,2)}childBefore(t){return this.child(-1,t,-2)}prop(t){return this.type.prop(t)}enter(t,n,r=0){if(r&Tl.ExcludeBuffers)return null;let{buffer:i}=this.context,a=i.findChild(this.index+4,i.buffer[this.index+3],n>0?1:-1,t-this.context.start,n);return a<0?null:new Qv(this.context,this,a)}get parent(){return this._parent||this.context.parent.nextSignificantParent()}externalSibling(t){return this._parent?null:this.context.parent.nextChild(this.context.index+t,t,0,4)}get nextSibling(){let{buffer:t}=this.context,n=t.buffer[this.index+3];return n<(this._parent?t.buffer[this._parent.index+3]:t.buffer.length)?new Qv(this.context,this._parent,n):this.externalSibling(1)}get prevSibling(){let{buffer:t}=this.context,n=this._parent?this._parent.index+4:0;return this.index==n?this.externalSibling(-1):new Qv(this.context,this._parent,t.findChild(n,this.index,-1,0,4))}get tree(){return null}toTree(){let t=[],n=[],{buffer:r}=this.context,i=this.index+4,a=r.buffer[this.index+3];if(a>i){let u=r.buffer[this.index+1];t.push(r.slice(i,a,u)),n.push(0)}return new rp(this.type,t,n,this.to-this.from)}toString(){return this.context.buffer.childString(this.index)}}function x0e(e){if(!e.length)return null;let t=0,n=e[0];for(let a=1;a<e.length;a++){let u=e[a];(u.from>n.from||u.to<n.to)&&(n=u,t=a)}let r=n instanceof Kd&&n.index<0?null:n.parent,i=e.slice();return r?i[t]=r:i.splice(t,1),new u7e(i,n)}class u7e{constructor(t,n){this.heads=t,this.node=n}get next(){return x0e(this.heads)}}function l7e(e,t,n){let r=e.resolveInner(t,n),i=null;for(let a=r instanceof Kd?r:r.context.parent;a;a=a.parent)if(a.index<0){let u=a.parent;(i||(i=[r])).push(u.resolve(t,n)),a=u}else{let u=K2.get(a.tree);if(u&&u.overlay&&u.overlay[0].from<=t&&u.overlay[u.overlay.length-1].to>=t){let l=new Kd(u.tree,u.overlay[0].from+a.from,-1,a);(i||(i=[r])).push(Fx(l,t,n,!1))}}return i?x0e(i):r}class OM{get name(){return this.type.name}constructor(t,n=0){if(this.buffer=null,this.stack=[],this.index=0,this.bufferNode=null,this.mode=n&~Tl.EnterBracketed,t instanceof Kd)this.yieldNode(t);else{this._tree=t.context.parent,this.buffer=t.context;for(let r=t._parent;r;r=r._parent)this.stack.unshift(r.index);this.bufferNode=t,this.yieldBuf(t.index)}}yieldNode(t){return t?(this._tree=t,this.type=t.type,this.from=t.from,this.to=t.to,!0):!1}yieldBuf(t,n){this.index=t;let{start:r,buffer:i}=this.buffer;return this.type=n||i.set.types[i.buffer[t]],this.from=r+i.buffer[t+1],this.to=r+i.buffer[t+2],!0}yield(t){return t?t instanceof Kd?(this.buffer=null,this.yieldNode(t)):(this.buffer=t.context,this.yieldBuf(t.index,t.type)):!1}toString(){return this.buffer?this.buffer.buffer.childString(this.index):this._tree.toString()}enterChild(t,n,r){if(!this.buffer)return this.yield(this._tree.nextChild(t<0?this._tree._tree.children.length-1:0,t,n,r,this.mode));let{buffer:i}=this.buffer,a=i.findChild(this.index+4,i.buffer[this.index+3],t,n-this.buffer.start,r);return a<0?!1:(this.stack.push(this.index),this.yieldBuf(a))}firstChild(){return this.enterChild(1,0,4)}lastChild(){return this.enterChild(-1,0,4)}childAfter(t){return this.enterChild(1,t,2)}childBefore(t){return this.enterChild(-1,t,-2)}enter(t,n,r=this.mode){return this.buffer?r&Tl.ExcludeBuffers?!1:this.enterChild(1,t,n):this.yield(this._tree.enter(t,n,r))}parent(){if(!this.buffer)return this.yieldNode(this.mode&Tl.IncludeAnonymous?this._tree._parent:this._tree.parent);if(this.stack.length)return this.yieldBuf(this.stack.pop());let t=this.mode&Tl.IncludeAnonymous?this.buffer.parent:this.buffer.parent.nextSignificantParent();return this.buffer=null,this.yieldNode(t)}sibling(t){if(!this.buffer)return this._tree._parent?this.yield(this._tree.index<0?null:this._tree._parent.nextChild(this._tree.index+t,t,0,4,this.mode)):!1;let{buffer:n}=this.buffer,r=this.stack.length-1;if(t<0){let i=r<0?0:this.stack[r]+4;if(this.index!=i)return this.yieldBuf(n.findChild(i,this.index,-1,0,4))}else{let i=n.buffer[this.index+3];if(i<(r<0?n.buffer.length:n.buffer[this.stack[r]+3]))return this.yieldBuf(i)}return r<0?this.yield(this.buffer.parent.nextChild(this.buffer.index+t,t,0,4,this.mode)):!1}nextSibling(){return this.sibling(1)}prevSibling(){return this.sibling(-1)}atLastNode(t){let n,r,{buffer:i}=this;if(i){if(t>0){if(this.index<i.buffer.buffer.length)return!1}else for(let a=0;a<this.index;a++)if(i.buffer.buffer[a+3]<this.index)return!1;({index:n,parent:r}=i)}else({index:n,_parent:r}=this._tree);for(;r;{index:n,_parent:r}=r)if(n>-1)for(let a=n+t,u=t<0?-1:r._tree.children.length;a!=u;a+=t){let l=r._tree.children[a];if(this.mode&Tl.IncludeAnonymous||l instanceof iy||!l.type.isAnonymous||VI(l))return!1}return!0}move(t,n){if(n&&this.enterChild(t,0,4))return!0;for(;;){if(this.sibling(t))return!0;if(this.atLastNode(t)||!this.parent())return!1}}next(t=!0){return this.move(1,t)}prev(t=!0){return this.move(-1,t)}moveTo(t,n=0){for(;(this.from==this.to||(n<1?this.from>=t:this.from>t)||(n>-1?this.to<=t:this.to<t))&&this.parent(););for(;this.enterChild(1,t,n););return this}get node(){if(!this.buffer)return this._tree;let t=this.bufferNode,n=null,r=0;if(t&&t.context==this.buffer)e:for(let i=this.index,a=this.stack.length;a>=0;){for(let u=t;u;u=u._parent)if(u.index==i){if(i==this.index)return u;n=u,r=a+1;break e}i=this.stack[--a]}for(let i=r;i<this.stack.length;i++)n=new Qv(this.buffer,n,this.stack[i]);return this.bufferNode=new Qv(this.buffer,n,this.index)}get tree(){return this.buffer?null:this._tree._tree}iterate(t,n){for(let r=0;;){let i=!1;if(this.type.isAnonymous||t(this)!==!1){if(this.firstChild()){r++;continue}this.type.isAnonymous||(i=!0)}for(;;){if(i&&n&&n(this),i=this.type.isAnonymous,!r)return;if(this.nextSibling())break;this.parent(),r--,i=!0}}}matchContext(t){if(!this.buffer)return bM(this.node.parent,t);let{buffer:n}=this.buffer,{types:r}=n.set;for(let i=t.length-1,a=this.stack.length-1;i>=0;a--){if(a<0)return bM(this._tree,t,i);let u=r[n.buffer[this.stack[a]]];if(!u.isAnonymous){if(t[i]&&t[i]!=u.name)return!1;i--}}return!0}}function VI(e){return e.children.some(t=>t instanceof iy||!t.type.isAnonymous||VI(t))}function o7e(e){var t;let{buffer:n,nodeSet:r,maxBufferLength:i=n7e,reused:a=[],minRepeatType:u=r.types.length}=e,l=Array.isArray(n)?new qI(n,n.length):n,o=r.types,s=0,c=0;function d(A,w,R,F,I,U){let{id:q,start:V,end:N,size:j}=l,W=c,K=s;if(j<0)if(l.next(),j==-1){let G=a[q];R.push(G),F.push(V-A);return}else if(j==-3){s=q;return}else if(j==-4){c=q;return}else throw new RangeError(`Unrecognized record size: ${j}`);let J=o[q],te,H,M=V-A;if(N-V<=i&&(H=S(l.pos-w,I))){let G=new Uint16Array(H.size-H.skip),X=l.pos-H.size,B=G.length;for(;l.pos>X;)B=_(H.start,G,B);te=new iy(G,N-H.start,r),M=H.start-A}else{let G=l.pos-j;l.next();let X=[],B=[],ne=q>=u?q:-1,oe=0,ie=N;for(;l.pos>G;)ne>=0&&l.id==ne&&l.size>=0?(l.end<=ie-i&&(y(X,B,V,oe,l.end,ie,ne,W,K),oe=X.length,ie=l.end),l.next()):U>2500?h(V,G,X,B):d(V,G,X,B,ne,U+1);if(ne>=0&&oe>0&&oe<X.length&&y(X,B,V,oe,V,ie,ne,W,K),X.reverse(),B.reverse(),ne>-1&&oe>0){let de=m(J,K);te=UI(J,X,B,0,X.length,0,N-V,de,de)}else te=v(J,X,B,N-V,W-N,K)}R.push(te),F.push(M)}function h(A,w,R,F){let I=[],U=0,q=-1;for(;l.pos>w;){let{id:V,start:N,end:j,size:W}=l;if(W>4)l.next();else{if(q>-1&&N<q)break;q<0&&(q=j-i),I.push(V,N,j),U++,l.next()}}if(U){let V=new Uint16Array(U*4),N=I[I.length-2];for(let j=I.length-3,W=0;j>=0;j-=3)V[W++]=I[j],V[W++]=I[j+1]-N,V[W++]=I[j+2]-N,V[W++]=W;R.push(new iy(V,I[2]-N,r)),F.push(N-A)}}function m(A,w){return(R,F,I)=>{let U=0,q=R.length-1,V,N;if(q>=0&&(V=R[q])instanceof rp){if(!q&&V.type==A&&V.length==I)return V;(N=V.prop(du.lookAhead))&&(U=F[q]+V.length+N)}return v(A,R,F,I,U,w)}}function y(A,w,R,F,I,U,q,V,N){let j=[],W=[];for(;A.length>F;)j.push(A.pop()),W.push(w.pop()+R-I);A.push(v(r.types[q],j,W,U-I,V-U,N)),w.push(I-R)}function v(A,w,R,F,I,U,q){if(U){let V=[du.contextHash,U];q=q?[V].concat(q):[V]}if(I>25){let V=[du.lookAhead,I];q=q?[V].concat(q):[V]}return new rp(A,w,R,F,q)}function S(A,w){let R=l.fork(),F=0,I=0,U=0,q=R.end-i,V={size:0,start:0,skip:0};e:for(let N=R.pos-A;R.pos>N;){let j=R.size;if(R.id==w&&j>=0){V.size=F,V.start=I,V.skip=U,U+=4,F+=4,R.next();continue}let W=R.pos-j;if(j<0||W<N||R.start<q)break;let K=R.id>=u?4:0,J=R.start;for(R.next();R.pos>W;){if(R.size<0)if(R.size==-3||R.size==-4)K+=4;else break e;else R.id>=u&&(K+=4);R.next()}I=J,F+=j,U+=K}return(w<0||F==A)&&(V.size=F,V.start=I,V.skip=U),V.size>4?V:void 0}function _(A,w,R){let{id:F,start:I,end:U,size:q}=l;if(l.next(),q>=0&&F<u){let V=R;if(q>4){let N=l.pos-(q-4);for(;l.pos>N;)R=_(A,w,R)}w[--R]=V,w[--R]=U-A,w[--R]=I-A,w[--R]=F}else q==-3?s=F:q==-4&&(c=F);return R}let E=[],$=[];for(;l.pos>0;)d(e.start||0,e.bufferStart||0,E,$,-1,0);let k=(t=e.length)!==null&&t!==void 0?t:E.length?$[0]+E[0].length:0;return new rp(o[e.topID],E.reverse(),$.reverse(),k)}const FH=new WeakMap;function dD(e,t){if(!e.isAnonymous||t instanceof iy||t.type!=e)return 1;let n=FH.get(t);if(n==null){n=1;for(let r of t.children){if(r.type!=e||!(r instanceof rp)){n=1;break}n+=dD(e,r)}FH.set(t,n)}return n}function UI(e,t,n,r,i,a,u,l,o){let s=0;for(let y=r;y<i;y++)s+=dD(e,t[y]);let c=Math.ceil(s*1.5/8),d=[],h=[];function m(y,v,S,_,E){for(let $=S;$<_;){let k=$,A=v[$],w=dD(e,y[$]);for($++;$<_;$++){let R=dD(e,y[$]);if(w+R>=c)break;w+=R}if($==k+1){if(w>c){let R=y[k];m(R.children,R.positions,0,R.children.length,v[k]+E);continue}d.push(y[k])}else{let R=v[$-1]+y[$-1].length-A;d.push(UI(e,y,v,k,$,A,R,null,o))}h.push(A+E-a)}}return m(t,n,r,i,0),(l||o)(d,h,u)}class xM{constructor(t,n,r,i,a=!1,u=!1){this.from=t,this.to=n,this.tree=r,this.offset=i,this.open=(a?1:0)|(u?2:0)}get openStart(){return(this.open&1)>0}get openEnd(){return(this.open&2)>0}static addTree(t,n=[],r=!1){let i=[new xM(0,t.length,t,0,!1,r)];for(let a of n)a.to>t.length&&i.push(a);return i}static applyChanges(t,n,r=128){if(!n.length)return t;let i=[],a=1,u=t.length?t[0]:null;for(let l=0,o=0,s=0;;l++){let c=l<n.length?n[l]:null,d=c?c.fromA:1e9;if(d-o>=r)for(;u&&u.from<d;){let h=u;if(o>=h.from||d<=h.to||s){let m=Math.max(h.from,o)-s,y=Math.min(h.to,d)-s;h=m>=y?null:new xM(m,y,h.tree,h.offset+s,l>0,!!c)}if(h&&i.push(h),u.to>d)break;u=a<t.length?t[a++]:null}if(!c)break;o=c.toA,s=c.toA-c.toB}return i}}let hut=class{startParse(t,n,r){return typeof t=="string"&&(t=new s7e(t)),r=r?r.length?r.map(i=>new sP(i.from,i.to)):[new sP(0,0)]:[new sP(0,t.length)],this.createParse(t,n||[],r)}parse(t,n,r){let i=this.startParse(t,n,r);for(;;){let a=i.advance();if(a)return a}}};class s7e{constructor(t){this.string=t}get length(){return this.string.length}chunk(t){return this.string.slice(t)}get lineChunks(){return!1}read(t,n){return this.string.slice(t,n)}}new du({perNode:!0});let c7e=0;class jd{constructor(t,n,r,i){this.name=t,this.set=n,this.base=r,this.modified=i,this.id=c7e++}toString(){let{name:t}=this;for(let n of this.modified)n.name&&(t=`${n.name}(${t})`);return t}static define(t,n){let r=typeof t=="string"?t:"?";if(t instanceof jd&&(n=t),n!=null&&n.base)throw new Error("Can not derive from a modified tag");let i=new jd(r,[],null,[]);if(i.set.push(i),n)for(let a of n.set)i.set.push(a);return i}static defineModifier(t){let n=new E$(t);return r=>r.modified.indexOf(n)>-1?r:E$.get(r.base||r,r.modified.concat(n).sort((i,a)=>i.id-a.id))}}let f7e=0;class E${constructor(t){this.name=t,this.instances=[],this.id=f7e++}static get(t,n){if(!n.length)return t;let r=n[0].instances.find(l=>l.base==t&&d7e(n,l.modified));if(r)return r;let i=[],a=new jd(t.name,i,t,n);for(let l of n)l.instances.push(a);let u=h7e(n);for(let l of t.set)if(!l.modified.length)for(let o of u)i.push(E$.get(l,o));return a}}function d7e(e,t){return e.length==t.length&&e.every((n,r)=>n==t[r])}function h7e(e){let t=[[]];for(let n=0;n<e.length;n++)for(let r=0,i=t.length;r<i;r++)t.push(t[r].concat(e[n]));return t.sort((n,r)=>r.length-n.length)}function mut(e){let t=Object.create(null);for(let n in e){let r=e[n];Array.isArray(r)||(r=[r]);for(let i of n.split(" "))if(i){let a=[],u=2,l=i;for(let d=0;;){if(l=="..."&&d>0&&d+3==i.length){u=1;break}let h=/^"(?:[^"\\]|\\.)*?"|[^\/!]+/.exec(l);if(!h)throw new RangeError("Invalid path: "+i);if(a.push(h[0]=="*"?"":h[0][0]=='"'?JSON.parse(h[0]):h[0]),d+=h[0].length,d==i.length)break;let m=i[d++];if(d==i.length&&m=="!"){u=0;break}if(m!="/")throw new RangeError("Invalid path: "+i);l=i.slice(d)}let o=a.length-1,s=a[o];if(!s)throw new RangeError("Invalid path: "+i);let c=new Nx(r,u,o>0?a.slice(0,o):null);t[s]=c.sort(t[s])}}return S0e.add(t)}const S0e=new du({combine(e,t){let n,r,i;for(;e||t;){if(!e||t&&e.depth>=t.depth?(i=t,t=t.next):(i=e,e=e.next),n&&n.mode==i.mode&&!i.context&&!n.context)continue;let a=new Nx(i.tags,i.mode,i.context);n?n.next=a:r=a,n=a}return r}});class Nx{constructor(t,n,r,i){this.tags=t,this.mode=n,this.context=r,this.next=i}get opaque(){return this.mode==0}get inherit(){return this.mode==1}sort(t){return!t||t.depth<this.depth?(this.next=t,this):(t.next=this.sort(t.next),t)}get depth(){return this.context?this.context.length:0}}Nx.empty=new Nx([],2,null);function p7e(e,t){let n=Object.create(null);for(let a of e)if(!Array.isArray(a.tag))n[a.tag.id]=a.class;else for(let u of a.tag)n[u.id]=a.class;let{scope:r,all:i=null}=t||{};return{style:a=>{let u=i;for(let l of a)for(let o of l.set){let s=n[o.id];if(s){u=u?u+" "+s:s;break}}return u},scope:r}}function m7e(e,t){let n=null;for(let r of e){let i=r.style(t);i&&(n=n?n+" "+i:i)}return n}function gut(e,t,n,r=0,i=e.length){let a=new g7e(r,Array.isArray(t)?t:[t],n);a.highlightRange(e.cursor(),r,i,"",a.highlighters),a.flush(i)}class g7e{constructor(t,n,r){this.at=t,this.highlighters=n,this.span=r,this.class=""}startSpan(t,n){n!=this.class&&(this.flush(t),t>this.at&&(this.at=t),this.class=n)}flush(t){t>this.at&&this.class&&this.span(this.at,t,this.class)}highlightRange(t,n,r,i,a){let{type:u,from:l,to:o}=t;if(l>=r||o<=n)return;u.isTop&&(a=this.highlighters.filter(m=>!m.scope||m.scope(u)));let s=i,c=v7e(t)||Nx.empty,d=m7e(a,c.tags);if(d&&(s&&(s+=" "),s+=d,c.mode==1&&(i+=(i?" ":"")+d)),this.startSpan(Math.max(n,l),s),c.opaque)return;let h=t.tree&&t.tree.prop(du.mounted);if(h&&h.overlay){let m=t.node.enter(h.overlay[0].from+l,1),y=this.highlighters.filter(S=>!S.scope||S.scope(h.tree.type)),v=t.firstChild();for(let S=0,_=l;;S++){let E=S<h.overlay.length?h.overlay[S]:null,$=E?E.from+l:o,k=Math.max(n,_),A=Math.min(r,$);if(k<A&&v)for(;t.from<A&&(this.highlightRange(t,k,A,i,a),this.startSpan(Math.min(A,t.to),s),!(t.to>=$||!t.nextSibling())););if(!E||$>r)break;_=E.to+l,_>n&&(this.highlightRange(m.cursor(),Math.max(n,E.from+l),Math.min(r,_),"",y),this.startSpan(Math.min(r,_),s))}v&&t.parent()}else if(t.firstChild()){h&&(i="");do if(!(t.to<=n)){if(t.from>=r)break;this.highlightRange(t,n,r,i,a),this.startSpan(Math.min(r,t.to),s)}while(t.nextSibling());t.parent()}}}function v7e(e){let t=e.type.prop(S0e);for(;t&&t.context&&!e.matchContext(t.context);)t=t.next;return t||null}const Ht=jd.define,hE=Ht(),Sv=Ht(),NH=Ht(Sv),IH=Ht(Sv),_v=Ht(),pE=Ht(_v),cP=Ht(_v),f0=Ht(),yb=Ht(f0),u0=Ht(),l0=Ht(),SM=Ht(),o2=Ht(SM),mE=Ht(),Pr={comment:hE,lineComment:Ht(hE),blockComment:Ht(hE),docComment:Ht(hE),name:Sv,variableName:Ht(Sv),typeName:NH,tagName:Ht(NH),propertyName:IH,attributeName:Ht(IH),className:Ht(Sv),labelName:Ht(Sv),namespace:Ht(Sv),macroName:Ht(Sv),literal:_v,string:pE,docString:Ht(pE),character:Ht(pE),attributeValue:Ht(pE),number:cP,integer:Ht(cP),float:Ht(cP),bool:Ht(_v),regexp:Ht(_v),escape:Ht(_v),color:Ht(_v),url:Ht(_v),keyword:u0,self:Ht(u0),null:Ht(u0),atom:Ht(u0),unit:Ht(u0),modifier:Ht(u0),operatorKeyword:Ht(u0),controlKeyword:Ht(u0),definitionKeyword:Ht(u0),moduleKeyword:Ht(u0),operator:l0,derefOperator:Ht(l0),arithmeticOperator:Ht(l0),logicOperator:Ht(l0),bitwiseOperator:Ht(l0),compareOperator:Ht(l0),updateOperator:Ht(l0),definitionOperator:Ht(l0),typeOperator:Ht(l0),controlOperator:Ht(l0),punctuation:SM,separator:Ht(SM),bracket:o2,angleBracket:Ht(o2),squareBracket:Ht(o2),paren:Ht(o2),brace:Ht(o2),content:f0,heading:yb,heading1:Ht(yb),heading2:Ht(yb),heading3:Ht(yb),heading4:Ht(yb),heading5:Ht(yb),heading6:Ht(yb),contentSeparator:Ht(f0),list:Ht(f0),quote:Ht(f0),emphasis:Ht(f0),strong:Ht(f0),link:Ht(f0),monospace:Ht(f0),strikethrough:Ht(f0),inserted:Ht(),deleted:Ht(),changed:Ht(),invalid:Ht(),meta:mE,documentMeta:Ht(mE),annotation:Ht(mE),processingInstruction:Ht(mE),definition:jd.defineModifier("definition"),constant:jd.defineModifier("constant"),function:jd.defineModifier("function"),standard:jd.defineModifier("standard"),local:jd.defineModifier("local"),special:jd.defineModifier("special")};for(let e in Pr){let t=Pr[e];t instanceof jd&&(t.name=e)}p7e([{tag:Pr.link,class:"tok-link"},{tag:Pr.heading,class:"tok-heading"},{tag:Pr.emphasis,class:"tok-emphasis"},{tag:Pr.strong,class:"tok-strong"},{tag:Pr.keyword,class:"tok-keyword"},{tag:Pr.atom,class:"tok-atom"},{tag:Pr.bool,class:"tok-bool"},{tag:Pr.url,class:"tok-url"},{tag:Pr.labelName,class:"tok-labelName"},{tag:Pr.inserted,class:"tok-inserted"},{tag:Pr.deleted,class:"tok-deleted"},{tag:Pr.literal,class:"tok-literal"},{tag:Pr.string,class:"tok-string"},{tag:Pr.number,class:"tok-number"},{tag:[Pr.regexp,Pr.escape,Pr.special(Pr.string)],class:"tok-string2"},{tag:Pr.variableName,class:"tok-variableName"},{tag:Pr.local(Pr.variableName),class:"tok-variableName tok-local"},{tag:Pr.definition(Pr.variableName),class:"tok-variableName tok-definition"},{tag:Pr.special(Pr.variableName),class:"tok-variableName2"},{tag:Pr.definition(Pr.propertyName),class:"tok-propertyName tok-definition"},{tag:Pr.typeName,class:"tok-typeName"},{tag:Pr.namespace,class:"tok-namespace"},{tag:Pr.className,class:"tok-className"},{tag:Pr.macroName,class:"tok-macroName"},{tag:Pr.propertyName,class:"tok-propertyName"},{tag:Pr.operator,class:"tok-operator"},{tag:Pr.comment,class:"tok-comment"},{tag:Pr.meta,class:"tok-meta"},{tag:Pr.invalid,class:"tok-invalid"},{tag:Pr.punctuation,class:"tok-punctuation"}]);const _M="ͼ",LH=typeof Symbol>"u"?"__"+_M:Symbol.for(_M),EM=typeof Symbol>"u"?"__styleSet"+Math.floor(Math.random()*1e8):Symbol("styleSet"),BH=typeof globalThis<"u"?globalThis:typeof window<"u"?window:{};class vut{constructor(t,n){this.rules=[];let{finish:r}=n||{};function i(u){return/^@/.test(u)?[u]:u.split(/,\s*/)}function a(u,l,o,s){let c=[],d=/^@(\w+)\b/.exec(u[0]),h=d&&d[1]=="keyframes";if(d&&l==null)return o.push(u[0]+";");for(let m in l){let y=l[m];if(/&/.test(m))a(m.split(/,\s*/).map(v=>u.map(S=>v.replace(/&/,S))).reduce((v,S)=>v.concat(S)),y,o);else if(y&&typeof y=="object"){if(!d)throw new RangeError("The value of a property ("+m+") should be a primitive value.");a(i(m),y,c,h)}else y!=null&&c.push(m.replace(/_.*/,"").replace(/[A-Z]/g,v=>"-"+v.toLowerCase())+": "+y+";")}(c.length||h)&&o.push((r&&!d&&!s?u.map(r):u).join(", ")+" {"+c.join(" ")+"}")}for(let u in t)a(i(u),t[u],this.rules)}getRules(){return this.rules.join(`
`)}static newName(){let t=BH[LH]||1;return BH[LH]=t+1,_M+t.toString(36)}static mount(t,n,r){let i=t[EM],a=r&&r.nonce;i?a&&i.setNonce(a):i=new y7e(t,a),i.mount(Array.isArray(n)?n:[n],t)}}let QH=new Map;class y7e{constructor(t,n){let r=t.ownerDocument||t,i=r.defaultView;if(!t.head&&t.adoptedStyleSheets&&i.CSSStyleSheet){let a=QH.get(r);if(a)return t[EM]=a;this.sheet=new i.CSSStyleSheet,QH.set(r,this)}else this.styleTag=r.createElement("style"),n&&this.styleTag.setAttribute("nonce",n);this.modules=[],t[EM]=this}mount(t,n){let r=this.sheet,i=0,a=0;for(let u=0;u<t.length;u++){let l=t[u],o=this.modules.indexOf(l);if(o<a&&o>-1&&(this.modules.splice(o,1),a--,o=-1),o==-1){if(this.modules.splice(a++,0,l),r)for(let s=0;s<l.rules.length;s++)r.insertRule(l.rules[s],i++)}else{for(;a<o;)i+=this.modules[a++].rules.length;i+=l.rules.length,a++}}if(r)n.adoptedStyleSheets.indexOf(this.sheet)<0&&(n.adoptedStyleSheets=[this.sheet,...n.adoptedStyleSheets]);else{let u="";for(let o=0;o<this.modules.length;o++)u+=this.modules[o].getRules()+`
`;this.styleTag.textContent=u;let l=n.head||n;this.styleTag.parentNode!=l&&l.insertBefore(this.styleTag,l.firstChild)}}setNonce(t){this.styleTag&&this.styleTag.getAttribute("nonce")!=t&&this.styleTag.setAttribute("nonce",t)}}var hD={exports:{}},b7e=hD.exports,zH;function O7e(){return zH||(zH=1,(function(e){(function(t,n){e.exports?e.exports=n():t.log=n()})(b7e,function(){var t=function(){},n="undefined",r=typeof window!==n&&typeof window.navigator!==n&&/Trident\/|MSIE /.test(window.navigator.userAgent),i=["trace","debug","info","warn","error"],a={},u=null;function l(v,S){var _=v[S];if(typeof _.bind=="function")return _.bind(v);try{return Function.prototype.bind.call(_,v)}catch{return function(){return Function.prototype.apply.apply(_,[v,arguments])}}}function o(){console.log&&(console.log.apply?console.log.apply(console,arguments):Function.prototype.apply.apply(console.log,[console,arguments])),console.trace&&console.trace()}function s(v){return v==="debug"&&(v="log"),typeof console===n?!1:v==="trace"&&r?o:console[v]!==void 0?l(console,v):console.log!==void 0?l(console,"log"):t}function c(){for(var v=this.getLevel(),S=0;S<i.length;S++){var _=i[S];this[_]=S<v?t:this.methodFactory(_,v,this.name)}if(this.log=this.debug,typeof console===n&&v<this.levels.SILENT)return"No console available for logging"}function d(v){return function(){typeof console!==n&&(c.call(this),this[v].apply(this,arguments))}}function h(v,S,_){return s(v)||d.apply(this,arguments)}function m(v,S){var _=this,E,$,k,A="loglevel";typeof v=="string"?A+=":"+v:typeof v=="symbol"&&(A=void 0);function w(q){var V=(i[q]||"silent").toUpperCase();if(!(typeof window===n||!A)){try{window.localStorage[A]=V;return}catch{}try{window.document.cookie=encodeURIComponent(A)+"="+V+";"}catch{}}}function R(){var q;if(!(typeof window===n||!A)){try{q=window.localStorage[A]}catch{}if(typeof q===n)try{var V=window.document.cookie,N=encodeURIComponent(A),j=V.indexOf(N+"=");j!==-1&&(q=/^([^;]+)/.exec(V.slice(j+N.length+1))[1])}catch{}return _.levels[q]===void 0&&(q=void 0),q}}function F(){if(!(typeof window===n||!A)){try{window.localStorage.removeItem(A)}catch{}try{window.document.cookie=encodeURIComponent(A)+"=; expires=Thu, 01 Jan 1970 00:00:00 UTC"}catch{}}}function I(q){var V=q;if(typeof V=="string"&&_.levels[V.toUpperCase()]!==void 0&&(V=_.levels[V.toUpperCase()]),typeof V=="number"&&V>=0&&V<=_.levels.SILENT)return V;throw new TypeError("log.setLevel() called with invalid level: "+q)}_.name=v,_.levels={TRACE:0,DEBUG:1,INFO:2,WARN:3,ERROR:4,SILENT:5},_.methodFactory=S||h,_.getLevel=function(){return k??$??E},_.setLevel=function(q,V){return k=I(q),V!==!1&&w(k),c.call(_)},_.setDefaultLevel=function(q){$=I(q),R()||_.setLevel(q,!1)},_.resetLevel=function(){k=null,F(),c.call(_)},_.enableAll=function(q){_.setLevel(_.levels.TRACE,q)},_.disableAll=function(q){_.setLevel(_.levels.SILENT,q)},_.rebuild=function(){if(u!==_&&(E=I(u.getLevel())),c.call(_),u===_)for(var q in a)a[q].rebuild()},E=I(u?u.getLevel():"WARN");var U=R();U!=null&&(k=I(U)),c.call(_)}u=new m,u.getLogger=function(S){if(typeof S!="symbol"&&typeof S!="string"||S==="")throw new TypeError("You must supply a name when creating a logger.");var _=a[S];return _||(_=a[S]=new m(S,u.methodFactory)),_};var y=typeof window!==n?window.log:void 0;return u.noConflict=function(){return typeof window!==n&&window.log===u&&(window.log=y),u},u.getLoggers=function(){return a},u.default=u,u})})(hD)),hD.exports}var x7e=O7e();const yut=da(x7e),S7e={AdditionalItemsError:"Array at `{{pointer}}` may not have an additional item `{{key}}`",AdditionalPropertiesError:"Additional property `{{property}}` on `{{pointer}}` does not match schema `{{schema}}`",AllOfError:"Value `{{value}}` at `{{pointer}}` does not match schema of `{{allOf}}`",AnyOfError:"Value `{{value}}` at `{{pointer}}` does not match any schema of `{{anyOf}}`",ConstError:"Expected value at `{{pointer}}` to be `{{expected}}`, but value given is `{{value}}`",containsAnyError:"The array at `{{pointer}}` must contain at least one item",ContainsArrayError:"The property at `{{pointer}}` must not be an array",ContainsError:"The array at `{{pointer}}` must contain an element that matches `{{schema}}`",EnumError:"Expected given value `{{value}}` in `{{pointer}}` to be one of `{{values}}`",ForbiddenPropertyError:"Property name `{{property}}` at `{{pointer}}` is not allowed",FormatDateError:"Value `{{value}}` at `{{pointer}}` is not a valid date",FormatDateTimeError:"Value `{{value}}` at `{{pointer}}` is not a valid date-time",FormatEmailError:"Value `{{value}}` at `{{pointer}}` is not a valid email",FormatHostnameError:"Value `{{value}}` at `{{pointer}}` is not a valid hostname",FormatIPV4Error:"Value `{{value}}` at `{{pointer}}` is not a valid IPv4 address",FormatIPV4LeadingZeroError:"IPv4 addresses starting with zero are invalid, since they are interpreted as octals",FormatIPV6Error:"Value `{{value}}` at `{{pointer}}` is not a valid IPv6 address",FormatIPV6LeadingZeroError:"IPv6 addresses starting with zero are invalid, since they are interpreted as octals",FormatJsonPointerError:"Value `{{value}}` at `{{pointer}}` is not a valid json-pointer",FormatRegExError:"Value `{{value}}` at `{{pointer}}` is not a valid regular expression",FormatTimeError:"Value `{{value}}` at `{{pointer}}` is not a valid time",FormatURIError:"Value `{{value}}` at `{{pointer}}` is not a valid uri",FormatURIReferenceError:"Value `{{value}}` at `{{pointer}}` is not a valid uri-reference",FormatURITemplateError:"Value `{{value}}` at `{{pointer}}` is not a valid uri-template",FormatURLError:"Value `{{value}}` at `{{pointer}}` is not a valid url",InvalidDataError:"No value may be specified in `{{pointer}}`",InvalidPropertyNameError:"Invalid property name `{{property}}` at `{{pointer}}`",MaximumError:"Value in `{{pointer}}` is `{{length}}`, but should be `{{maximum}}` at maximum",MaxItemsError:"Too many items in `{{pointer}}`, should be `{{maximum}}` at most, but got `{{length}}`",MaxLengthError:"Value `{{pointer}}` should have a maximum length of `{{maxLength}}`, but got `{{length}}`.",MaxPropertiesError:"Too many properties in `{{pointer}}`, should be `{{maximum}}` at most, but got `{{length}}`",MinimumError:"Value in `{{pointer}}` is `{{length}}`, but should be `{{minimum}}` at minimum",MinItemsError:"Too few items in `{{pointer}}`, should be at least `{{minItems}}`, but got `{{length}}`",MinItemsOneError:"At least one item is required in `{{pointer}}`",MinLengthError:"Value `{{pointer}}` should have a minimum length of `{{minLength}}`, but got `{{length}}`.",MinLengthOneError:"A value is required in `{{pointer}}`",MinPropertiesError:"Too few properties in `{{pointer}}`, should be at least `{{minimum}}`, but got `{{length}}`",MissingDependencyError:"The required propery '{{missingProperty}}' in `{{pointer}}` is missing",MissingOneOfPropertyError:"Value at `{{pointer}}` property: `{{property}}`",MultipleOfError:"Expected `{{value}}` in `{{pointer}}` to be multiple of `{{multipleOf}}`",MultipleOneOfError:"Value `{{value}}` should not match multiple schemas in oneOf `{{matches}}`",NoAdditionalPropertiesError:"Additional property `{{property}}` in `{{pointer}}` is not allowed",NotError:"Value `{{value}}` at pointer should not match schema `{{not}}`",OneOfError:"Value `{{value}}` in `{{pointer}}` does not match any given oneof schema",OneOfPropertyError:"Failed finding a matching oneOfProperty schema in `{{pointer}}` where `{{property}}` matches `{{value}}`",PatternError:"Value in `{{pointer}}` should match `{{description}}`, but received `{{received}}`",PatternPropertiesError:"Property `{{key}}` does not match any patterns in `{{pointer}}`. Valid patterns are: {{patterns}}",RequiredPropertyError:"The required property `{{key}}` is missing at `{{pointer}}`",SchemaWarning:"Failed retrieving a schema from '{{pointer}}' to key '{{key}}'",TypeError:"Expected `{{value}}` ({{received}}) in `{{pointer}}` to be of type `{{expected}}`",UndefinedValueError:"Value must not be undefined in `{{pointer}}`",UniqueItemsError:"Items in array must be unique. Value `{{value}}` in `{{pointer}}` is a duplicate of {{duplicatePointer}}.",UnknownPropertyError:"Could not find a valid schema for property `{{pointer}}` within object",ValueNotEmptyError:"A value for `{{property}}` is required at `{{pointer}}`"},_7e=Object.prototype.toString;function gi(e){const t=_7e.call(e).match(/\s([^\]]+)\]/).pop().toLowerCase();return t==="file"?"object":t}const E7e="object",D7e="array";function $7e(e,t={}){return e.replace(/\{\{\w+\}\}/g,n=>{const r=n.replace(/[{}]/g,""),i=t[r],a=gi(i);return a===E7e||a===D7e?JSON.stringify(i):i})}function C7e(e,t,n=e){var r;const i=(r=S7e[e])!==null&&r!==void 0?r:n;return $7e(i,t)}function w7e(e){return e.replace(/([a-z])([A-Z])/g,"$1-$2").toLowerCase()}function R7e(e,t){return{type:"error",name:e,code:w7e(e),message:C7e(e,t),data:t}}function Dn(e){return R7e.bind(null,e)}function Ix(e,t=[]){for(let n=0;n<e.length;n+=1){const r=e[n];Array.isArray(r)?Ix(r,t):t.push(r)}return t}const a4={DECLARATOR_ONEOF:"oneOfProperty",GET_TEMPLATE_RECURSION_LIMIT:1,propertyBlacklist:["_id"],templateDefaultOptions:{addOptionalProps:!1,removeInvalidData:!1,extendDefaults:!0}};function W2(e,t,n){const r={...t};return Object.defineProperty(r,"getOneOfOrigin",{enumerable:!1,value:()=>({index:n,schema:e})}),r}function Gf(e){return(e==null?void 0:e.type)==="error"}function A7e(e){return e instanceof Promise}function D$(e){return Gf(e)||A7e(e)}const{DECLARATOR_ONEOF:m3}=a4;function _0e(e,t,n=e.rootSchema,r="#"){if(t!=null&&n[m3]){const u=[],l=n[m3],o=t[n[m3]];if(o===void 0)return e.errors.missingOneOfPropertyError({property:l,pointer:r,schema:n,value:t});for(let s=0;s<n.oneOf.length;s+=1){const c=e.resolveRef(n.oneOf[s]),d=e.step(l,c,t,r);if(Gf(d))return d;let h=Ix(e.validate(o,d,r));if(h=h.filter(D$),h.length>0)u.push(...h);else return W2(n,c,s)}return e.errors.oneOfPropertyError({property:l,value:o,pointer:r,schema:n,errors:u})}const i=[],a=[];for(let u=0;u<n.oneOf.length;u+=1){const l=e.resolveRef(n.oneOf[u]);let o=Ix(e.validate(t,l,r));o=o.filter(D$),o.length>0?a.push(...o):i.push({index:u,schema:l})}return i.length===1?W2(n,i[0].schema,i[0].index):i.length>1?e.errors.multipleOneOfError({value:t,pointer:r,schema:n,matches:i}):e.errors.oneOfError({value:JSON.stringify(t),pointer:r,schema:n,oneOf:n.oneOf,errors:a})}function T7e(e,t,n,r){if(n==null||t.properties==null)return-1;let i=0;const a=Object.keys(t.properties);for(let u=0;u<a.length;u+=1){const l=a[u];n[l]!=null&&e.isValid(n[l],t.properties[l],r)&&(i+=1)}return i}function jI(e,t,n=e.rootSchema,r="#"){if(t!=null&&n[m3]){const a=[],u=n[m3],l=t[n[m3]];if(l===void 0)return e.errors.missingOneOfPropertyError({property:u,pointer:r,schema:n,value:t});for(let o=0;o<n.oneOf.length;o+=1){const s=e.resolveRef(n.oneOf[o]),c=e.step(u,s,t,r);if(Gf(c))return c;let d=Ix(e.validate(l,c,r));if(d=d.filter(D$),d.length>0)a.push(...d);else return W2(n,s,o)}return e.errors.oneOfPropertyError({property:u,value:l,pointer:r,schema:n,errors:a})}const i=[];for(let a=0;a<n.oneOf.length;a+=1){const u=e.resolveRef(n.oneOf[a]);e.isValid(t,u,r)&&i.push({schema:u,index:a})}if(i.length===1)return W2(n,i[0].schema,i[0].index);if(gi(t)==="object"){let a,u=-1,l=0;for(let o=0;o<n.oneOf.length;o+=1){const s=e.resolveRef(n.oneOf[o]),c=T7e(e,s,t);l<c&&(l=c,a=n.oneOf[o],u=o)}return a===void 0?e.errors.oneOfError({value:JSON.stringify(t),pointer:r,schema:n,oneOf:n.oneOf}):W2(n,a,u)}return i.length>1?e.errors.multipleOneOfError({matches:i,pointer:r,schema:n,value:t}):e.errors.oneOfError({value:JSON.stringify(t),pointer:r,schema:n,oneOf:n.oneOf})}const k7e=(e,t,n,r)=>{if(Array.isArray(t.oneOf)){const i=e.resolveOneOf(n,t,r);if(Gf(i))return i}};var fP,qH;function P7e(){if(qH)return fP;qH=1;var e=function(E){return t(E)&&!n(E)};function t(_){return!!_&&typeof _=="object"}function n(_){var E=Object.prototype.toString.call(_);return E==="[object RegExp]"||E==="[object Date]"||a(_)}var r=typeof Symbol=="function"&&Symbol.for,i=r?Symbol.for("react.element"):60103;function a(_){return _.$$typeof===i}function u(_){return Array.isArray(_)?[]:{}}function l(_,E){return E.clone!==!1&&E.isMergeableObject(_)?v(u(_),_,E):_}function o(_,E,$){return _.concat(E).map(function(k){return l(k,$)})}function s(_,E){if(!E.customMerge)return v;var $=E.customMerge(_);return typeof $=="function"?$:v}function c(_){return Object.getOwnPropertySymbols?Object.getOwnPropertySymbols(_).filter(function(E){return Object.propertyIsEnumerable.call(_,E)}):[]}function d(_){return Object.keys(_).concat(c(_))}function h(_,E){try{return E in _}catch{return!1}}function m(_,E){return h(_,E)&&!(Object.hasOwnProperty.call(_,E)&&Object.propertyIsEnumerable.call(_,E))}function y(_,E,$){var k={};return $.isMergeableObject(_)&&d(_).forEach(function(A){k[A]=l(_[A],$)}),d(E).forEach(function(A){m(_,A)||(h(_,A)&&$.isMergeableObject(E[A])?k[A]=s(A,$)(_[A],E[A],$):k[A]=l(E[A],$))}),k}function v(_,E,$){$=$||{},$.arrayMerge=$.arrayMerge||o,$.isMergeableObject=$.isMergeableObject||e,$.cloneUnlessOtherwiseSpecified=l;var k=Array.isArray(E),A=Array.isArray(_),w=k===A;return w?k?$.arrayMerge(_,E,$):y(_,E,$):l(E,$)}v.all=function(E,$){if(!Array.isArray(E))throw new Error("first argument should be an array");return E.reduce(function(k,A){return v(k,A,$)},{})};var S=v;return fP=S,fP}var M7e=P7e();const E0e=da(M7e),F7e=(e,t)=>t,ZI=(e,t)=>E0e(e,t,{arrayMerge:F7e}),N7e=(e,t)=>{const n=e.concat(t);return n.filter((r,i)=>n.indexOf(r)===i)},I7e=(e,t)=>E0e(e,t,{arrayMerge:N7e});function oc(e,t){const n=gi(e),r=gi(t);if(n!==r)return e;const i=I7e(e,t);return e.getOneOfOrigin?Object.defineProperty(i,"getOneOfOrigin",{enumerable:!1,value:e.getOneOfOrigin}):t.getOneOfOrigin&&Object.defineProperty(i,"getOneOfOrigin",{enumerable:!1,value:t.getOneOfOrigin}),i}function Lx(e,...t){const n={};return Object.keys(e).forEach(r=>{t.includes(r)||(n[r]=e[r])}),e.getOneOfOrigin&&Object.defineProperty(n,"getOneOfOrigin",{enumerable:!1,value:e.getOneOfOrigin}),n}var L7e=Function.prototype.toString,dP=Object.create,B7e=Object.prototype.toString,Q7e=(function(){function e(){this._keys=[],this._values=[]}return e.prototype.has=function(t){return!!~this._keys.indexOf(t)},e.prototype.get=function(t){return this._values[this._keys.indexOf(t)]},e.prototype.set=function(t,n){this._keys.push(t),this._values.push(n)},e})();function z7e(){return new Q7e}function q7e(){return new WeakMap}var V7e=typeof WeakMap<"u"?q7e:z7e;function HI(e){if(!e)return dP(null);var t=e.constructor;if(t===Object)return e===Object.prototype?{}:dP(e);if(t&&~L7e.call(t).indexOf("[native code]"))try{return new t}catch{}return dP(e)}function U7e(e){var t="";return e.global&&(t+="g"),e.ignoreCase&&(t+="i"),e.multiline&&(t+="m"),e.unicode&&(t+="u"),e.sticky&&(t+="y"),t}function j7e(e){return e.flags}var Z7e=/test/g.flags==="g"?j7e:U7e;function D0e(e){var t=B7e.call(e);return t.substring(8,t.length-1)}function H7e(e){return e[Symbol.toStringTag]||D0e(e)}var X7e=typeof Symbol<"u"?H7e:D0e,Y7e=Object.defineProperty,G7e=Object.getOwnPropertyDescriptor,$0e=Object.getOwnPropertyNames,XI=Object.getOwnPropertySymbols,C0e=Object.prototype,w0e=C0e.hasOwnProperty,K7e=C0e.propertyIsEnumerable,R0e=typeof XI=="function";function W7e(e){return $0e(e).concat(XI(e))}var J7e=R0e?W7e:$0e;function _w(e,t,n){for(var r=J7e(e),i=0,a=r.length,u=void 0,l=void 0;i<a;++i)if(u=r[i],!(u==="callee"||u==="caller")){if(l=G7e(e,u),!l){t[u]=n.copier(e[u],n);continue}!l.get&&!l.set&&(l.value=n.copier(l.value,n));try{Y7e(t,u,l)}catch{t[u]=l.value}}return t}function e9e(e,t){var n=new t.Constructor;t.cache.set(e,n);for(var r=0,i=e.length;r<i;++r)n[r]=t.copier(e[r],t);return n}function t9e(e,t){var n=new t.Constructor;return t.cache.set(e,n),_w(e,n,t)}function A0e(e,t){return e.slice(0)}function n9e(e,t){return e.slice(0,e.size,e.type)}function r9e(e,t){return new t.Constructor(A0e(e.buffer))}function i9e(e,t){return new t.Constructor(e.getTime())}function T0e(e,t){var n=new t.Constructor;return t.cache.set(e,n),e.forEach(function(r,i){n.set(i,t.copier(r,t))}),n}function a9e(e,t){return _w(e,T0e(e,t),t)}function u9e(e,t){var n=HI(t.prototype);t.cache.set(e,n);for(var r in e)w0e.call(e,r)&&(n[r]=t.copier(e[r],t));return n}function l9e(e,t){var n=HI(t.prototype);t.cache.set(e,n);for(var r in e)w0e.call(e,r)&&(n[r]=t.copier(e[r],t));for(var i=XI(e),a=0,u=i.length,l=void 0;a<u;++a)l=i[a],K7e.call(e,l)&&(n[l]=t.copier(e[l],t));return n}var o9e=R0e?l9e:u9e;function s9e(e,t){var n=HI(t.prototype);return t.cache.set(e,n),_w(e,n,t)}function hP(e,t){return new t.Constructor(e.valueOf())}function c9e(e,t){var n=new t.Constructor(e.source,Z7e(e));return n.lastIndex=e.lastIndex,n}function pD(e,t){return e}function k0e(e,t){var n=new t.Constructor;return t.cache.set(e,n),e.forEach(function(r){n.add(t.copier(r,t))}),n}function f9e(e,t){return _w(e,k0e(e,t),t)}var d9e=Array.isArray,YI=Object.assign,h9e=Object.getPrototypeOf||(function(e){return e.__proto__}),P0e={array:e9e,arrayBuffer:A0e,blob:n9e,dataView:r9e,date:i9e,error:pD,map:T0e,object:o9e,regExp:c9e,set:k0e},p9e=YI({},P0e,{array:t9e,map:a9e,object:s9e,set:f9e});function m9e(e){return{Arguments:e.object,Array:e.array,ArrayBuffer:e.arrayBuffer,Blob:e.blob,Boolean:hP,DataView:e.dataView,Date:e.date,Error:e.error,Float32Array:e.arrayBuffer,Float64Array:e.arrayBuffer,Int8Array:e.arrayBuffer,Int16Array:e.arrayBuffer,Int32Array:e.arrayBuffer,Map:e.map,Number:hP,Object:e.object,Promise:pD,RegExp:e.regExp,Set:e.set,String:hP,WeakMap:pD,WeakSet:pD,Uint8Array:e.arrayBuffer,Uint8ClampedArray:e.arrayBuffer,Uint16Array:e.arrayBuffer,Uint32Array:e.arrayBuffer,Uint64Array:e.arrayBuffer}}function M0e(e){var t=YI({},P0e,e),n=m9e(t),r=n.Array,i=n.Object;function a(u,l){if(l.prototype=l.Constructor=void 0,!u||typeof u!="object")return u;if(l.cache.has(u))return l.cache.get(u);if(l.prototype=h9e(u),l.Constructor=l.prototype&&l.prototype.constructor,!l.Constructor||l.Constructor===Object)return i(u,l);if(d9e(u))return r(u,l);var o=n[X7e(u)];return o?o(u,l):typeof u.then=="function"?u:i(u,l)}return function(l){return a(l,{Constructor:void 0,cache:V7e(),copier:a,prototype:void 0})}}function g9e(e){return M0e(YI({},p9e,e))}g9e({});var $$=M0e({});function Ew(e,t,n){if(t.if!=null){if(t.if===!1)return t.else;if(t.if&&(t.then||t.else)){const r=e.validate(n,e.resolveRef(t.if));if(r.length===0&&t.then)return e.resolveRef(t.then);if(r.length!==0&&t.else)return e.resolveRef(t.else)}}}const v9e=(e,t,n,r)=>{const i=Ew(e,t,n);if(i)return e.validate(n,i,r)};function F0e(e,t,n){var r;const i={...(r=e.resolveRef(t))!==null&&r!==void 0?r:{}},a=Ew(e,i,n);return a||Lx(i,"if","then","else")}function N0e(e,t,n=e.rootSchema){let r=$$(n);for(let i=0;i<n.allOf.length;i+=1){const a=F0e(e,n.allOf[i],t);r=oc(r,a)}return delete r.allOf,r}function I0e(e,t){const{allOf:n}=t;if(!Array.isArray(n)||n.length===0)return;let r={};return n.forEach(i=>{r=oc(r,e.resolveRef(i))}),r}const y9e=(e,t,n,r)=>{const{allOf:i}=t;if(!Array.isArray(i)||i.length===0)return;const a=[];return t.allOf.forEach(u=>{a.push(...e.validate(n,u,r))}),a};function L0e(e,t){return e==null||e.$ref==null?e:e.getRoot?e.getRoot().getRef(e):t.getRef(e)}function b9e(e){return e.filter((t,n)=>e.indexOf(t)===n)}function jv(e){return gi(e)==="object"}function B0e(e,t,n){const{dependencies:r}=t;if(!jv(r)||!jv(n))return;let i=!1,a={required:[]};if(Object.keys(r).forEach(u=>{var l,o;if(n[u]==null&&!(!((l=t.required)===null||l===void 0)&&l.includes(u)||!((o=a.required)===null||o===void 0)&&o.includes(u)))return;const s=r[u];if(Array.isArray(s)){i=!0,a.required.push(...s);return}if(jv(s)){i=!0,a=oc(a,e.resolveRef(s));return}}),i)return a.required=b9e(a.required),a}const O9e=(e,t,n,r)=>{if(gi(t.dependencies)!=="object")return;const i=[];return Object.keys(n).forEach(a=>{if(t.dependencies[a]===void 0||t.dependencies[a]===!0)return;if(t.dependencies[a]===!1){i.push(e.errors.missingDependencyError({pointer:r,schema:t,value:n}));return}let u;const l=gi(t.dependencies[a]);if(l==="array")u=t.dependencies[a].filter(o=>n[o]===void 0).map(o=>e.errors.missingDependencyError({missingProperty:o,pointer:r,schema:t,value:n}));else if(l==="object")u=e.validate(n,t.dependencies[a],r);else throw new Error(`Invalid dependency definition for ${r}/${a}. Must be string[] or schema`);i.push(...u)}),i.length>0?i:void 0};function Q0e(e,t,n){if(!Array.isArray(t.anyOf)||t.anyOf.length===0)return;let r;return t.anyOf.forEach(i=>{i=e.resolveRef(i),e.isValid(n,i)&&(r=r?oc(r,i):i)}),r}function z0e(e,t,n=e.rootSchema,r="#"){const{anyOf:i}=n;if(!Array.isArray(i)||i.length===0)return n;const a=Q0e(e,n,t);if(a==null)return e.errors.anyOfError({pointer:r,schema:n,value:t,anyOf:JSON.stringify(i)});const u=oc(n,a);return Lx(u,"anyOf")}const x9e=(e,t,n,r)=>{if(!(!Array.isArray(t.anyOf)||t.anyOf.length===0)){for(let i=0;i<t.anyOf.length;i+=1)if(e.isValid(n,t.anyOf[i]))return;return e.errors.anyOfError({pointer:r,schema:t,value:n,anyOf:t.anyOf})}},VH=["allOf","anyOf","oneOf","dependencies","if","then","else"],S9e=["allOf","anyOf","oneOf","dependencies","if"];function _9e(e){const t=Object.keys(e);return S9e.findIndex(n=>t.includes(n))!==-1}function DM(e,t,n,r){let i,a;if(t=e.resolveRef(t),t.oneOf){const c=jI(e,n,t,r);Gf(c)?a=c:c&&(i=oc(i??{},c))}if(Array.isArray(t.allOf)){const c=t.allOf.map(d=>{if(_9e(d)){let h=DM(e,d,n,r);return h?(h=oc(d,h),Lx(h,...VH)):void 0}return d});if(c.length>0){const d=I0e(e,{allOf:c});i=oc(i??{},d)}}const u=Q0e(e,t,n);u&&(i=oc(i??{},u));const l=B0e(e,t,n);l&&(i=oc(i??{},l));const o=Ew(e,t,n);if(o&&(i=oc(i??{},o)),i==null)return a;const s=DM(e,i,n,r);return s&&(i=oc(i,s)),Lx(i,...VH)}const E9e=["allOf","anyOf","oneOf","dependencies","if","then","else"];function UH(e,t,n,r){let i=DM(e,t,n,r);return i?(i=oc(t,i),Lx(i,...E9e)):t}class q0e{constructor(t,n){this.remotes={},this.errors={},this.typeKeywords={},this.validateKeyword={},this.validateType={},this.validateFormat={},this.config=t,this.typeKeywords=$$(t.typeKeywords),this.validateKeyword=Object.assign({},t.validateKeyword),this.validateType=Object.assign({},t.validateType),this.validateFormat=Object.assign({},t.validateFormat),this.errors=Object.assign({},t.errors),this.setSchema(n)}get rootSchema(){return this.__rootSchema}set rootSchema(t){t!=null&&(this.__rootSchema=this.config.compileSchema(this,t))}addRemoteSchema(t,n){this.config.addRemoteSchema(this,t,n)}compileSchema(t){var n;return this.config.compileSchema(this,t,(n=this.rootSchema)!==null&&n!==void 0?n:t)}createSchemaOf(t){return this.config.createSchemaOf(t)}each(t,n,r,i){return this.config.each(this,t,n,r,i)}eachSchema(t,n=this.rootSchema){return this.config.eachSchema(n,t)}getChildSchemaSelection(t,n){return this.config.getChildSchemaSelection(this,t,n)}getSchema(t){return this.config.getSchema(this,t)}getTemplate(t,n,r=this.config.templateDefaultOptions){return this.config.getTemplate(this,t,n,r)}isValid(t,n,r){return this.config.isValid(this,t,n,r)}resolveAnyOf(t,n,r){return this.config.resolveAnyOf(this,t,n,r)}resolveAllOf(t,n){return this.config.resolveAllOf(this,t,n)}resolveRef(t){return this.config.resolveRef(t,this.rootSchema)}resolveOneOf(t,n,r){return this.config.resolveOneOf(this,t,n,r)}setSchema(t){this.rootSchema=t}step(t,n,r,i){return this.config.step(this,t,n,r,i)}validate(t,n,r){return this.config.validate(this,t,n,r)}}function V0e(e,t,n){n.id=n.id||t,e.remotes[t]=e.compileSchema(n)}var mD={exports:{}},D9e=mD.exports,jH;function $9e(){return jH||(jH=1,(function(e,t){(function(n,r){e.exports=r()})(typeof self<"u"?self:D9e,(()=>(()=>{var n={d:(N,j)=>{for(var W in j)n.o(j,W)&&!n.o(N,W)&&Object.defineProperty(N,W,{enumerable:!0,get:j[W]})},o:(N,j)=>Object.prototype.hasOwnProperty.call(N,j),r:N=>{typeof Symbol<"u"&&Symbol.toStringTag&&Object.defineProperty(N,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(N,"__esModule",{value:!0})}},r={};function i(N){return N==="#"||N===""||Array.isArray(N)&&N.length===0||!1}n.r(r),n.d(r,{default:()=>V,get:()=>h,isRoot:()=>i,join:()=>U,remove:()=>A,removeUndefinedItems:()=>k,set:()=>_,split:()=>d,splitLast:()=>q});const a=/~1/g,u=/~0/g,l=/\/+/g,o=/(^[#/]*|\/+$)/g;function s(N){return N.replace(a,"/").replace(u,"~")}function c(N){return s(decodeURIComponent(N))}function d(N){if(N==null||typeof N!="string"||i(N))return Array.isArray(N)?N:[];const j=N.indexOf("#")>=0?c:s,W=(N=(N=N.replace(l,"/")).replace(o,"")).split("/");for(let K=0,J=W.length;K<J;K+=1)W[K]=j(W[K]);return W}function h(N,j,W=void 0){if(j==null||N==null)return W;if(i(j))return N;const K=m(N,d(j));return K===void 0?W:K}function m(N,j){const W=j.shift();if(N!==void 0)return W!==void 0?m(N[W],j):N}const y=/^\[.*\]$/,v=/^\[(.+)\]$/;function S(N,j){return N==="__proto__"||N=="constructor"&&j.length>0&&j[0]=="prototype"}function _(N,j,W){if(j==null)return N;const K=d(j);if(K.length===0)return N;N==null&&(N=y.test(K[0])?[]:{});let J,te,H=N;for(;K.length>1;)J=K.shift(),te=y.test(K[0]),S(J,K)||(H=$(H,J,te));return J=K.pop(),E(H,J,W),N}function E(N,j,W){let K;const J=j.match(v);j==="[]"&&Array.isArray(N)?N.push(W):J?(K=J.pop(),N[K]=W):N[j]=W}function $(N,j,W){if(N[j]!=null)return N[j];const K=W?[]:{};return E(N,j,K),K}function k(N){let j=0,W=0;for(;j+W<N.length;)N[j+W]===void 0&&(W+=1),N[j]=N[j+W],j+=1;return N.length=N.length-W,N}function A(N,j,W){const K=d(j),J=K.pop(),te=h(N,K);return te&&delete te[J],Array.isArray(te)&&W!==!0&&k(te),N}const w=/\/+/g,R=/~/g,F=/\//g;function I(N,j){if(N.length===0)return j?"#":"";for(let W=0,K=N.length;W<K;W+=1)N[W]=N[W].replace(R,"~0").replace(F,"~1"),j&&(N[W]=encodeURIComponent(N[W]));return((j?"#/":"/")+N.join("/")).replace(w,"/")}function U(N,...j){const W=[];if(Array.isArray(N))return I(N,arguments[1]===!0);const K=arguments[arguments.length-1],J=typeof K=="boolean"?K:N&&N[0]==="#";for(let H=0,M=arguments.length;H<M;H+=1)W.push.apply(W,d(arguments[H]));const te=[];for(let H=0,M=W.length;H<M;H+=1)if(W[H]===".."){if(te.length===0)return J?"#":"";te.pop()}else te.push(W[H]);return I(te,J)}function q(N){const j=d(N);if(j.length===0)return typeof N=="string"&&N[0]==="#"?["#",j[0]]:["",void 0];if(j.length===1)return N[0]==="#"?["#",j[0]]:["",j[0]];const W=j.pop();return[U(j,N[0]==="#"),W]}const V={get:h,set:_,remove:A,join:U,split:d,splitLast:q,isRoot:i,removeUndefinedItems:k};return r})()))})(mD)),mD.exports}var zb=$9e();const Bx=da(zb);var gE={},Xc={},s2={},ZH;function C9e(){if(ZH)return s2;ZH=1,Object.defineProperty(s2,"__esModule",{value:!0}),s2.TokenError=void 0;let e=class extends Error{constructor(n,r){if(super(n),this.message=n,this.token=r,r&&r.errors)r.errors.push(this);else throw this}inspect(){return"SyntaxError: "+this.message}};return s2.TokenError=e,s2}var HH;function w9e(){if(HH)return Xc;HH=1,Object.defineProperty(Xc,"__esModule",{value:!0}),Xc.Parser=Xc.findRuleByName=Xc.parseRuleName=Xc.escapeRegExp=Xc.readToken=void 0;const e=/^[A-Z0-9_]+$/,t=/(\?|\+|\*)$/,n=/^(@|&|!)/,r="WS",i=C9e();function a(v,S){let _=S.exec(v);return _&&_.index==0?_[0].length==0&&S.source.length>0?null:{type:null,text:_[0],rest:v.substr(_[0].length),start:0,end:_[0].length-1,fullText:_[0],errors:[],children:[],parent:null}:null}Xc.readToken=a;function u(v){return v.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g,"\\$&")}Xc.escapeRegExp=u;function l(v){v.rest="",v.children&&v.children.forEach(S=>l(S))}function o(v,S){v.start+=S,v.end+=S,v.children&&v.children.forEach(_=>o(_,v.start))}function s(v,S){S.errors&&S.errors.length&&S.errors.forEach(_=>v.push(_)),S.children&&S.children.forEach(_=>s(v,_))}function c(v){let S=t.exec(v),_=n.exec(v),E=S&&S[0]||"",$=_&&_[0]||"",k={raw:v,name:v.replace(t,"").replace(n,""),isOptional:E=="?"||E=="*",allowRepetition:E=="+"||E=="*",atLeastOne:E=="+",lookupPositive:$=="&",lookupNegative:$=="!",pinned:$=="@",lookup:!1,isLiteral:!1};return k.isLiteral=k.name[0]=="'"||k.name[0]=='"',k.lookup=k.lookupNegative||k.lookupPositive,k}Xc.parseRuleName=c;function d(v,S){let _=c(v);return S.cachedRules[_.name]||null}Xc.findRuleByName=d;function h(v,S){if(v.children){let _=v.children.filter(E=>E.type&&S.test(E.type));for(let E=0;E<_.length;E++){let $=v.children.indexOf(_[E]);$!=-1&&v.children.splice($,1)}v.children.forEach(E=>h(E,S))}}const m=["EOF"];class y{constructor(S,_){this.grammarRules=S,this.options=_,this.cachedRules={},this.debug=_?_.debug===!0:!1;let E=[],$=[];if(S.forEach(k=>{let A=c(k.name);if(A.name in this.cachedRules){E.push("Duplicated rule "+A.name);return}else this.cachedRules[A.name]=k;if(!k.bnf||!k.bnf.length){let w="Missing rule content, rule: "+k.name;E.indexOf(w)==-1&&E.push(w)}else k.bnf.forEach(w=>{if(typeof w[0]=="string"&&c(w[0]).name==k.name){let F="Left recursion is not allowed, rule: "+k.name;E.indexOf(F)==-1&&E.push(F)}w.forEach(R=>{if(typeof R=="string"){let F=c(R);!F.isLiteral&&$.indexOf(F.name)==-1&&m.indexOf(F.name)==-1&&$.push(F.name)}})});r==k.name&&(k.implicitWs=!1),k.implicitWs&&$.indexOf(r)==-1&&$.push(r),k.recover&&$.indexOf(k.recover)==-1&&$.push(k.recover)}),$.forEach(k=>{k in this.cachedRules||E.push("Missing rule "+k)}),E.length)throw new Error(E.join(`
`))}getAST(S,_){_||(_=this.grammarRules.filter($=>!$.fragment&&$.name.indexOf("%")!=0)[0].name);let E=this.parse(S,_);if(E){s(E.errors,E),o(E,0),h(E,/^%/),(!this.options||!this.options.keepUpperRules)&&h(E,e);let $=E.rest;$&&new i.TokenError(`Unexpected end of input: 
`+$,E),l(E),E.rest=$}return E}emitSource(){return"CANNOT EMIT SOURCE FROM BASE Parser"}parse(S,_,E=0){let $=null,k=c(_),A,w=this.debug&&!e.test(k.name);w&&console.log(new Array(E).join("│  ")+"Trying to get "+_+" from "+JSON.stringify(S.split(`
`)[0]));let R=k.name,F=d(k.name,this);if(k.name=="EOF"){if(S.length)return null;if(S.length==0)return{type:"EOF",text:"",rest:"",start:0,end:0,fullText:"",errors:[],children:[],parent:null}}try{if(!F&&k.isLiteral){let I=k.name.trim();if(I.startsWith('"')?I=JSON.parse(I):I.startsWith("'")&&(I=I.replace(/^'(.+)'$/,"$1").replace(/\\'/g,"'")),I==="")return{type:"%%EMPTY%%",text:"",rest:S,start:0,end:0,fullText:"",errors:[],children:[],parent:null};A=new RegExp(u(I)),R=null}}catch(I){return I instanceof ReferenceError&&console.error(I),null}if(A){let I=a(S,A);if(I)return I.type=R,I}else{let I=F.bnf;I instanceof Array&&I.forEach(U=>{if($)return;let q=null,V={type:k.name,text:"",children:[],end:0,errors:[],fullText:"",parent:null,start:0,rest:S};F.fragment&&(V.fragment=!0);let N=S,j=0,W=U.length>0,K=!1;for(let J=0;J<U.length;J++)if(typeof U[J]=="string"){let te=c(U[J]);W=W&&te.isOptional;let H,M=!1;do{if(H=null,F.implicitWs&&(H=this.parse(N,te.name,E+1),!H)){let G;do if(G=this.parse(N,r,E+1),G)V.text=V.text+G.text,V.end=V.text.length,G.parent=V,V.children.push(G),N=N.substr(G.text.length),j+=G.text.length;else break;while(G&&G.text.length)}if(H=H||this.parse(N,te.name,E+1),te.lookupNegative){if(H)return;break}if(te.lookupPositive&&!H)return;if(!H&&(te.isOptional||te.atLeastOne&&M))break;if(H&&F.pinned==J+1&&(q=H,w&&console.log(new Array(E+1).join("│  ")+"└─ "+H.type+" PINNED")),H||(H=this.parseRecovery(F,N,E+1)),!H)if(q)$=V,H={type:"SyntaxError",text:N,children:[],end:N.length,errors:[],fullText:"",parent:null,start:0,rest:""},N.length?new i.TokenError(`Unexpected end of input. Expecting ${te.name} Got: ${N}`,H):new i.TokenError(`Unexpected end of input. Missing ${te.name}`,H),w&&console.log(new Array(E+1).join("│  ")+"└─ "+H.type+" "+JSON.stringify(H.text));else return;if(M=!0,K=!0,H.type=="%%EMPTY%%")break;H.start+=j,H.end+=j,!te.lookupPositive&&H.type&&(H.fragment?H.children&&H.children.forEach(G=>{G.start+=j,G.end+=j,G.parent=V,V.children.push(G)}):(H.parent=V,V.children.push(H))),te.lookup&&(H.lookup=!0),w&&console.log(new Array(E+1).join("│  ")+"└─ "+H.type+" "+JSON.stringify(H.text)),!te.lookup&&!H.lookup&&(V.text=V.text+H.text,V.end=V.text.length,N=N.substr(H.text.length),j+=H.text.length),V.rest=N}while(H&&te.allowRepetition&&N.length&&!H.lookup)}else{let te=a(N,U[J]);if(!te)return;w&&console.log(new Array(E+1).join("│  ")+"└> "+JSON.stringify(te.text)+U[J].source),K=!0,te.start+=j,te.end+=j,V.text=V.text+te.text,V.end=V.text.length,N=N.substr(te.text.length),j+=te.text.length,V.rest=N}K&&($=V,w&&console.log(new Array(E).join("│  ")+"├<─┴< PUSHING "+$.type+" "+JSON.stringify($.text)))}),$&&F.simplifyWhenOneChildren&&$.children.length==1&&($=$.children[0])}return $||w&&console.log(_+" NOT RESOLVED FROM "+S),$}parseRecovery(S,_,E){if(S.recover&&_.length){let $=this.debug;$&&console.log(new Array(E+1).join("│  ")+"Trying to recover until token "+S.recover+" from "+JSON.stringify(_.split(`
`)[0]+_.split(`
`)[1]));let k={type:"SyntaxError",text:"",children:[],end:0,errors:[],fullText:"",parent:null,start:0,rest:""},A;do if(A=this.parse(_,S.recover,E+1),A){new i.TokenError('Unexpected input: "'+k.text+`" Expecting: ${S.name}`,k);break}else k.text=k.text+_[0],k.end=k.text.length,_=_.substr(1);while(!A&&_.length>0);if(k.text.length>0&&A)return $&&console.log(new Array(E+1).join("│  ")+"Recovered text: "+JSON.stringify(k.text)),k}return null}}return Xc.Parser=y,Xc.default=y,Xc}var XH;function R9e(){if(XH)return gE;XH=1,Object.defineProperty(gE,"__esModule",{value:!0});const e=w9e();var t;return(function(n){n.RULES=[{name:"Grammar",bnf:[["RULE_S*","%Atomic*","EOF"]]},{name:"%Atomic",bnf:[["Production","RULE_S*"]],fragment:!0},{name:"Production",bnf:[["NCName","RULE_S*",'"::="',"RULE_WHITESPACE*","Choice","RULE_WHITESPACE*","RULE_EOL+","RULE_S*"]]},{name:"NCName",bnf:[[/[a-zA-Z][a-zA-Z_0-9]*/]]},{name:"Choice",bnf:[["SequenceOrDifference","%_Choice_1*"]],fragment:!0},{name:"%_Choice_1",bnf:[["RULE_WHITESPACE*",'"|"',"RULE_WHITESPACE*","SequenceOrDifference"]],fragment:!0},{name:"SequenceOrDifference",bnf:[["Item","RULE_WHITESPACE*","%_Item_1?"]]},{name:"%_Item_1",bnf:[["Minus","Item"],["Item*"]],fragment:!0},{name:"Minus",bnf:[['"-"']]},{name:"Item",bnf:[["RULE_WHITESPACE*","%Primary","PrimaryDecoration?"]],fragment:!0},{name:"PrimaryDecoration",bnf:[['"?"'],['"*"'],['"+"']]},{name:"DecorationName",bnf:[['"ebnf://"',/[^\x5D#]+/]]},{name:"%Primary",bnf:[["NCName"],["StringLiteral"],["CharCode"],["CharClass"],["SubItem"]],fragment:!0},{name:"SubItem",bnf:[['"("',"RULE_WHITESPACE*","Choice","RULE_WHITESPACE*",'")"']]},{name:"StringLiteral",bnf:[[`'"'`,/[^"]*/,`'"'`],[`"'"`,/[^']*/,`"'"`]],pinned:1},{name:"CharCode",bnf:[['"#x"',/[0-9a-zA-Z]+/]]},{name:"CharClass",bnf:[["'['","'^'?","%RULE_CharClass_1+",'"]"']]},{name:"%RULE_CharClass_1",bnf:[["CharCodeRange"],["CharRange"],["CharCode"],["RULE_Char"]],fragment:!0},{name:"RULE_Char",bnf:[[/\x09/],[/\x0A/],[/\x0D/],[/[\x20-\x5c]/],[/[\x5e-\uD7FF]/],[/[\uE000-\uFFFD]/]]},{name:"CharRange",bnf:[["RULE_Char",'"-"',"RULE_Char"]]},{name:"CharCodeRange",bnf:[["CharCode",'"-"',"CharCode"]]},{name:"RULE_WHITESPACE",bnf:[["%RULE_WHITESPACE_CHAR*"],["Comment","RULE_WHITESPACE*"]]},{name:"RULE_S",bnf:[["RULE_WHITESPACE","RULE_S*"],["RULE_EOL","RULE_S*"]]},{name:"%RULE_WHITESPACE_CHAR",bnf:[[/\x09/],[/\x20/]],fragment:!0},{name:"Comment",bnf:[['"/*"',"%RULE_Comment_Body*",'"*/"']]},{name:"%RULE_Comment_Body",bnf:[['!"*/"',/[^*]/]],fragment:!0},{name:"RULE_EOL",bnf:[[/\x0D/,/\x0A/],[/\x0A/],[/\x0D/]]},{name:"Link",bnf:[["'['","Url","']'"]]},{name:"Url",bnf:[[/[^\x5D:/?#]/,'"://"',/[^\x5D#]+/,"%Url1?"]]},{name:"%Url1",bnf:[['"#"',"NCName"]],fragment:!0}],n.defaultParser=new e.Parser(n.RULES,{debug:!1});const r=/^(!|&)/,i=/(\?|\+|\*)$/,a=/^%/;function u($,k){if(typeof $=="string"){if(r.test($))return"";if(a.test($)){let w=i.exec($),R=w?w[0]+" ":"";return l($,k)?s($,k)+R:"("+s($,k)+")"+R}return $}else return $.source.replace(/\\(?:x|u)([a-zA-Z0-9]+)/g,"#x$1").replace(/\[\\(?:x|u)([a-zA-Z0-9]+)-\\(?:x|u)([a-zA-Z0-9]+)\]/g,"[#x$1-#x$2]")}function l($,k){let A=e.findRuleByName($,k);return A&&A.bnf.length==1&&A.bnf[0].length==1&&(A.bnf[0][0]instanceof RegExp||A.bnf[0][0][0]=='"'||A.bnf[0][0][0]=="'")}function o($,k){return $.map(A=>u(A,k)).join(" ")}function s($,k){let A=e.findRuleByName($,k);return A?A.bnf.map(w=>o(w,k)).join(" | "):"RULE_NOT_FOUND {"+$+"}"}function c($){let k=[];return $.grammarRules.forEach(A=>{if(!/^%/.test(A.name)){let w=A.recover?" /* { recoverUntil="+A.recover+" } */":"";k.push(A.name+" ::= "+s(A.name,$)+w)}}),k.join(`
`)}n.emit=c;let d=0;function h($,k){throw console.log("reberia restar "+k+" a "+$),new Error("Difference not supported yet")}function m($){return new RegExp($.replace(/#x([a-zA-Z0-9]{4})/g,"\\u$1").replace(/#x([a-zA-Z0-9]{3})/g,"\\u0$1").replace(/#x([a-zA-Z0-9]{2})/g,"\\x$1").replace(/#x([a-zA-Z0-9]{1})/g,"\\x0$1"))}function y($,k,A){let w=null,R=[];return k.children.forEach((F,I)=>{F.type=="Minus"&&h(w,F);let U=k.children[I+1];U=U&&U.type=="PrimaryDecoration"&&U.text||"";let q="";switch(F.type){case"SubItem":let V="%"+(A+d++);v($,F,V),R.push(q+V+U);break;case"NCName":case"StringLiteral":R.push(q+F.text+U);break;case"CharCode":case"CharClass":if(U||q){let N={name:"%"+(A+d++),bnf:[[m(F.text)]]};$.push(N),R.push(q+N.name+U)}else R.push(m(F.text));break;case"PrimaryDecoration":break;default:throw new Error(" HOW SHOULD I PARSE THIS? "+F.type+" -> "+JSON.stringify(F.text))}w=F}),R}function v($,k,A){let w=k.children.filter(I=>I.type=="SequenceOrDifference").map(I=>y($,I,A)),R={name:A,bnf:w},F=null;w.forEach(I=>{F=F||I.recover,delete I.recover}),A.indexOf("%")==0&&(R.fragment=!0),F&&(R.recover=F),$.push(R)}function S($,k=n.defaultParser){let A=k.getAST($);if(!A)throw new Error("Could not parse "+$);if(A.errors&&A.errors.length)throw A.errors[0];let w=[];return A.children.filter(R=>R.type=="Production").map(R=>{let F=R.children.filter(I=>I.type=="NCName")[0].text;v(w,R,F)}),w}n.getRules=S;function _($,k=n.defaultParser){return S($.join(""),k)}n.Transform=_;class E extends e.Parser{constructor(k,A){const w=A&&A.debugRulesParser===!0?new e.Parser(n.RULES,{debug:!0}):n.defaultParser;super(S(k,w),A)}emitSource(){return c(this)}}n.Parser=E})(t||(t={})),gE.default=t,gE}var A9e=R9e();const T9e=da(A9e),k9e="[^?/{}*,()#]+",P9e=`
root ::= ("#" recursion | recursion | (query | pattern) recursion* | "#" SEP? | SEP)
recursion ::= (SEP query | pattern)*

query ::= (ESC escaped ESC | property | all | any | regex) typecheck? lookahead?
property ::= ${k9e}
regex ::= "{" [^}]+ "}"
SEP ::= "/"
all ::= "**"
any ::= "*"

typecheck ::= "?:" ("value" | "boolean" | "string" | "number" | "object" | "array")
lookahead ::= "?" expression ((andExpr | orExpr) expression)*
andExpr ::= S? "&&" S?
orExpr ::= S? "||" S?

expression ::= (exprProperty | ESC escaped ESC) ((isnot | is) (exprProperty | regex | ESC escaped ESC))*
exprProperty ::= [a-zA-Z0-9-_ $]+
escaped ::= [^"]+
is ::= ":"
isnot ::= ":!"
ESC ::= '"'

pattern ::= S? "(" (SEP query | pattern (orPattern? pattern)*)* ")" quantifier? S? lookahead?
quantifier ::= "+" | "*" | [0-9]+
orPattern ::= S? "," S?

S ::= [ ]*
`,M9e=new T9e.Parser(P9e),F9e=e=>M9e.getAST(e),nf=0,M3=3,C$=(e,t)=>`${e}/${t}`,U0e=Object.prototype.toString,N9e=/Object|Array/,$M=e=>N9e.test(U0e.call(e)),I9e=e=>U0e.call(e).match(/\s([^\]]+)\]/).pop().toLowerCase();function j0e(e){return new RegExp(e.text.replace(/(^{|}$)/g,""))}function L9e(e,t){Array.isArray(e)?e.forEach(t):Object.prototype.toString.call(e)==="[object Object]"&&Object.keys(e).forEach(function(n){t(e[n],n,e)})}function YH(e){return Array.isArray(e)?e.map(function(t,n){return`${n}`}):Object.prototype.toString.call(e)==="[object Object]"?Object.keys(e):[]}const Zv={mem:[],get(e,t){const n=e[nf][t];if(!Zv.mem.includes(n))return $M(n)&&Zv.mem.push(n),[n,t,e[nf],C$(e[M3],t)]},reset(){Zv.mem.length=0}},CM={any(e,t){const n=t[nf];return YH(n).map(r=>[n[r],r,n,C$(t[M3],r)])},all(e,t){const n=[t];return L9e(t[nf],(r,i)=>{const a=Zv.get(t,i);a&&n.push(...CM.all(e,a))}),n},regex(e,t){const n=j0e(e),r=t[nf];return YH(r).filter(i=>n.test(i)).map(i=>[r[i],i,r,C$(t[M3],i)])}},w$={escaped:(e,t)=>w$.property(e,t),property:(e,t)=>{const n=e.text;if(t[nf]&&t[nf][n]!==void 0)return[t[nf][n],n,t[nf],C$(t[M3],n)]},typecheck:(e,t)=>{const n=e.text.replace(/^\?:/,"");if(n==="value")return $M(t[nf])?void 0:t;if(I9e(t[nf])===n)return t},lookahead:(e,t)=>{let n=!0,r=!1;return e.children.forEach(i=>{if(i.type==="expression"){const a=w$.expression(i,t)!==void 0;n=r===!0?n||a:n&&a}else r=i.type==="orExpr"}),n?t:void 0},expression:(e,t)=>{const n=e.children[0].text,r=e.children[1],i=e.children[2],a=t[nf];if($M(a)!==!1)return B9e(a[n],r,i)?t:void 0}};function B9e(e,t,n){if(t===void 0)return e!==void 0;let r;const i=`${e}`;return n.type==="regex"?r=j0e(n).test(i):r=i===n.text,t.type==="isnot"&&(r=r===!1&&e!==void 0),r}function Q9e(e,t,n,r){const i=[];for(let a=0,u=t.length;a<u;a+=1)i.push(...e(n,t[a],n,r));return i}function z9e(e,t,n,r){const i=[];for(let a=0,u=t.length;a<u;a+=1){const l=e(n,t[a],r);l&&i.push(l)}return i}function q9e(e,t,n){let r=e;return t.children.forEach(i=>{if(CM[i.type])r=Q9e(CM[i.type],r,i,n);else if(w$[i.type])r=z9e(w$[i.type],r,i,n);else throw new Error(`Unknown filter ${i.type}`)}),r}function V9e(e,t,n){const r=[];let i=e;return t.children.forEach(a=>{if(a.type==="orPattern"){r.push(...i),i=e;return}i=GI(i,a,n)}),r.push(...i),r}function U9e(e){if(e==null)return 1;if(e==="*"||e==="+")return 1/0;const t=parseInt(e);return isNaN(t)?1:t}function j9e(e,t,n){const r=[],i=t.children.find(o=>o.type==="quantifier"),a=U9e(i&&i.text);let u=e;i&&i.text==="*"&&r.push(...u);let l=0;for(;u.length>0&&l<a;)u=V9e(u,t,n),r.push(...u),l+=1;return r}function Z9e(e,t,n){let r=e;return t.children.forEach(i=>r=GI(r,i,n)),r}function GI(e,t,n){let r;return t.type==="query"?r=q9e(e,t,n):t.type==="pattern"?r=j9e(e,t,n):r=Z9e(e,t,n),Zv.reset(),Zv.mem.push(e),r}function H9e(e,t){return Zv.reset(),Zv.mem.push(e),GI([[e,null,null,"#"]],t)}const GH={value:e=>e.map(t=>t[nf]),pointer:e=>e.map(t=>t[M3]),all:e=>e,map:e=>{const t={};return e.forEach(n=>t[n[M3]]=n[nf]),t}};var p1;(function(e){e.POINTER="pointer",e.VALUE="value",e.ALL="all",e.MAP="map"})(p1||(p1={}));AS.POINTER=p1.POINTER;AS.VALUE=p1.VALUE;AS.ALL=p1.ALL;AS.MAP=p1.MAP;function AS(e,t,n=p1.VALUE){if(t==null)return[];t=t.replace(/(\/$)/g,""),t===""&&(t="#");const r=F9e(t);if(r==null)throw new Error(`empty ast for '${t}'`);if(r.rest!=="")throw new Error(`Failed parsing queryString from: '${r.rest}'`);const i=H9e(e,r);return typeof n=="function"?i.map(a=>n(...a)):GH[n]?GH[n](i):i}var R$;(function(e){e.REPLACE_ITEMS="replace",e.INSERT_ITEMS="insert"})(R$||(R$={}));R$.REPLACE_ITEMS;R$.INSERT_ITEMS;const Hm={$ref:{type:!1},allOf:{type:!1,definitions:["allOf/*"]},anyOf:{type:!1,definitions:["anyOf/*"]},array:{type:!0,definitions:["allOf/*","anyOf/*","oneOf/*","not","items","items/*","additionalItems"],validationKeywords:["minItems","maxItems","uniqueItems"],keywords:["items","additionalItems","minItems","maxItems","uniqueItems"]},boolean:{type:!0},enum:{type:!1},integer:{type:!0,definitions:["allOf/*","anyOf/*","oneOf/*","not"],validationKeywords:["minimum","maximum","multipleOf"]},not:{type:!1,definitions:["not"]},number:{type:!0,definitions:["allOf/*","anyOf/*","oneOf/*","not"],validationKeywords:["minimum","maximum","multipleOf"]},null:{type:!0},object:{type:!0,definitions:["allOf/*","anyOf/*","oneOf/*","not","properties/*","additionalProperties","patternProperties/*","dependencies/*"],validationKeywords:["minProperties","maxProperties","required"],keywords:["properties","additionalProperties","patternProperties","dependencies","minProperties","maxProperties","required"]},oneOf:{type:!1,definitions:["oneOf/*"]},string:{type:!0,definitions:["allOf/*","anyOf/*","oneOf/*","not"],validationKeywords:["minLength","maxLength","pattern"]}},X9e=Object.keys(Hm).filter(e=>Hm[e].type===!1),KH=Object.prototype.hasOwnProperty;function WH(e){if(jv(e)===!1)return;if(e.enum)return"enum";const t=e.type;if(Array.isArray(t)||Hm[t])return t;const n=X9e.filter(r=>e[r]);if(n.length===1)return n[0];if(n.length===0){for(let r=0,i=Hm.object.keywords.length;r<i;r+=1){const a=Hm.object.keywords[r];if(KH.call(e,a))return"object"}for(let r=0,i=Hm.array.keywords.length;r<i;r+=1){const a=Hm.array.keywords[r];if(KH.call(e,a))return"array"}return}throw new Error(`Mutiple typeIds [${n.join(", ")}] matched in ${JSON.stringify(e)}`)}function Y9e(e){const t=[],n=WH(e);if(n==null)return t;let r;if(Array.isArray(n)){r={};for(let i=0,a=n.length;i<a;i+=1)Object.assign(r,Hm[n[i]])}else r=Hm[n];return r.definitions==null||r.definitions.forEach(i=>{AS(e,i,(a,u,l,o)=>{jv(a)&&WH(a)&&t.push({pointer:Bx.join(Bx.split(o),!1),def:a})})}),t}function G9e(e,t){if(this.callback(e,t)===!0)return;Y9e(e).forEach(r=>this.nextTypeDefs(r.def,Bx.join(t,r.pointer,!1)))}function vE(e,t,n,r="definitions"){const i=t[r];Object.keys(i).forEach(a=>{if(i[a]===!1||jv(i[a])){e.nextTypeDefs(i[a],Bx.join(n,r,a,!1));return}})}function Dw(e,t,n="#"){const r={callback:t,nextTypeDefs:G9e};r.nextTypeDefs(e,n),e.definitions!=null&&(r.callback=(i,a)=>{t(i,a),i.definitions!=null&&vE(r,i,a)},vE(r,e,n)),e.$defs!=null&&(r.callback=(i,a)=>{t(i,a),i.definitions!=null&&vE(r,i,a)},vE(r,e,n,"$defs"))}const K9e=/(#|\/)+$/,yE=/#$/,W9e=/^[^:]+:\/\/[^/]+\//,J9e=/\/[^/]*$/,eMe=/#.*$/;function A$(e,t){return e==null&&t==null?"#":t==null?e.replace(yE,""):e==null?t.replace(yE,""):t[0]==="#"?`${e.replace(eMe,"")}${t.replace(K9e,"")}`:W9e.test(t)?t.replace(yE,""):`${e.replace(J9e,"")}/${t.replace(yE,"")}`}const c2=/(#|\/)+$/g,tMe=["",null,"#"];function nMe(e){if(tMe.includes(e))return[];if(e=e.replace(c2,""),e.indexOf("#")===-1)return[e.replace(c2,"")];if(e.indexOf("#")===0)return[e.replace(c2,"")];const t=e.split("#");return t[0]=t[0].replace(c2,""),t[1]=`#${t[1].replace(c2,"")}`,t}const rMe=/(#|\/)+$/g,iMe=e=>gi(e)==="object";function Qm(e,t,n){if(iMe(n)&&(n=n.__ref||n.$ref),n==null)return t;let r;const i=n.replace(rMe,"");if(e.remotes[i])return r=e.remotes[i],r&&r.$ref?Qm(e,t,r.$ref):r;if(e.ids[n])return r=zb.get(t,e.ids[n]),r&&r.$ref?Qm(e,t,r.$ref):r;const a=nMe(n);if(a.length===0)return t;if(a.length===1){if(n=a[0],e.remotes[n])return r=e.remotes[n],Qm(e,t,r.$ref);if(e.ids[n])return r=zb.get(t,e.ids[n]),r&&r.$ref?Qm(e,t,r.$ref):r}if(a.length===2){const u=a[0];if(n=a[1],e.remotes[u])return e.remotes[u].getRef?e.remotes[u].getRef(n):Qm(e,e.remotes[u],n);if(e.ids[u])return Qm(e,zb.get(t,e.ids[u]),n)}return r=zb.get(t,e.ids[n]||n),r&&r.$ref?Qm(e,t,r.$ref):r}const JH="__compiled",eX="__ref",aMe="getRef",uMe="getRoot",lMe=/(#|\/)+$/g;function oMe(e,t,n=t,r=!1){if(!t||t[JH]!==void 0)return t;const i={ids:{},remotes:e.remotes},a=JSON.stringify(t),u=JSON.parse(a);if(Object.defineProperty(u,JH,{enumerable:!1,value:!0}),Object.defineProperty(u,aMe,{enumerable:!1,value:Qm.bind(null,i,u)}),r===!1&&a.includes("$ref")===!1)return u;t!==n&&Object.defineProperty(u,"definitions",{enumerable:!1,value:Object.assign({},n.definitions,n.$defs,t.definitions,t.$defs)});const l={},o=()=>u;return Dw(u,(s,c)=>{var d;if(s.id){if(s.id.startsWith("http")&&/(allOf|anyOf|oneOf)\/\d+$/.test(c)){const S=c.replace(/\/(allOf|anyOf|oneOf)\/\d+$/,""),_=zb.get(u,S);s.id=(d=_.id)!==null&&d!==void 0?d:s.id}i.ids[s.id.replace(lMe,"")]=c}c=`#${c}`.replace(/##+/,"#");const h=c.replace(/\/[^/]+$/,""),m=c.replace(/\/[^/]+\/[^/]+$/,""),y=l[h]||l[m],v=A$(y,s.id);l[c]=v,i.ids[v]==null&&(i.ids[v]=c),s.$ref&&!s[eX]&&(Object.defineProperty(s,eX,{enumerable:!1,value:A$(v,s.$ref)}),Object.defineProperty(s,uMe,{enumerable:!1,value:o}))}),u}function Z0e(e,t,n,r=e.rootSchema,i="#"){r=e.resolveRef(r),n(r,t,i);const a=gi(t);a==="object"?Object.keys(t).forEach(u=>{const l=e.step(u,r,t,i),o=t[u];e.each(o,n,l,`${i}/${u}`)}):a==="array"&&t.forEach((u,l)=>{const o=e.step(l,r,t,i);e.each(u,n,o,`${i}/${l}`)})}const H0e={additionalItemsError:Dn("AdditionalItemsError"),additionalPropertiesError:Dn("AdditionalPropertiesError"),anyOfError:Dn("AnyOfError"),allOfError:Dn("AllOfError"),constError:Dn("ConstError"),containsError:Dn("ContainsError"),containsArrayError:Dn("ContainsArrayError"),containsAnyError:Dn("ContainsAnyError"),enumError:Dn("EnumError"),forbiddenPropertyError:Dn("ForbiddenPropertyError"),formatURLError:Dn("FormatURLError"),formatURIError:Dn("FormatURIError"),formatURIReferenceError:Dn("FormatURIReferenceError"),formatURITemplateError:Dn("FormatURITemplateError"),formatDateError:Dn("FormatDateError"),formatDateTimeError:Dn("FormatDateTimeError"),formatEmailError:Dn("FormatEmailError"),formatHostnameError:Dn("FormatHostnameError"),formatIPV4Error:Dn("FormatIPV4Error"),formatIPV4LeadingZeroError:Dn("FormatIPV4LeadingZeroError"),formatIPV6Error:Dn("FormatIPV6Error"),formatIPV6LeadingZeroError:Dn("FormatIPV6LeadingZeroError"),formatJsonPointerError:Dn("FormatJsonPointerError"),formatRegExError:Dn("FormatRegExError"),formatTimeError:Dn("FormatTimeError"),invalidSchemaError:Dn("InvalidSchemaError"),invalidDataError:Dn("InvalidDataError"),invalidTypeError:Dn("InvalidTypeError"),invalidPropertyNameError:Dn("InvalidPropertyNameError"),maximumError:Dn("MaximumError"),maxItemsError:Dn("MaxItemsError"),maxLengthError:Dn("MaxLengthError"),maxPropertiesError:Dn("MaxPropertiesError"),minimumError:Dn("MinimumError"),minItemsError:Dn("MinItemsError"),minItemsOneError:Dn("MinItemsOneError"),minLengthError:Dn("MinLengthError"),minLengthOneError:Dn("MinLengthOneError"),minPropertiesError:Dn("MinPropertiesError"),missingDependencyError:Dn("MissingDependencyError"),missingOneOfPropertyError:Dn("MissingOneOfPropertyError"),multipleOfError:Dn("MultipleOfError"),multipleOneOfError:Dn("MultipleOneOfError"),noAdditionalPropertiesError:Dn("NoAdditionalPropertiesError"),notError:Dn("NotError"),oneOfError:Dn("OneOfError"),oneOfPropertyError:Dn("OneOfPropertyError"),patternError:Dn("PatternError"),patternPropertiesError:Dn("PatternPropertiesError"),requiredPropertyError:Dn("RequiredPropertyError"),schemaWarning:Dn("SchemaWarning"),typeError:Dn("TypeError"),undefinedValueError:Dn("UndefinedValueError"),uniqueItemsError:Dn("UniqueItemsError"),unknownPropertyError:Dn("UnknownPropertyError"),valueNotEmptyError:Dn("ValueNotEmptyError")};var pP={exports:{}},tX;function sMe(){return tX||(tX=1,(function(e){(function(t){t.exports.is_uri=r,t.exports.is_http_uri=i,t.exports.is_https_uri=a,t.exports.is_web_uri=u,t.exports.isUri=r,t.exports.isHttpUri=i,t.exports.isHttpsUri=a,t.exports.isWebUri=u;var n=function(l){var o=l.match(/(?:([^:\/?#]+):)?(?:\/\/([^\/?#]*))?([^?#]*)(?:\?([^#]*))?(?:#(.*))?/);return o};function r(l){if(l&&!/[^a-z0-9\:\/\?\#\[\]\@\!\$\&\'\(\)\*\+\,\;\=\.\-\_\~\%]/i.test(l)&&!/%[^0-9a-f]/i.test(l)&&!/%[0-9a-f](:?[^0-9a-f]|$)/i.test(l)){var o=[],s="",c="",d="",h="",m="",y="";if(o=n(l),s=o[1],c=o[2],d=o[3],h=o[4],m=o[5],!!(s&&s.length&&d.length>=0)){if(c&&c.length){if(!(d.length===0||/^\//.test(d)))return}else if(/^\/\//.test(d))return;if(/^[a-z][a-z0-9\+\-\.]*$/.test(s.toLowerCase()))return y+=s+":",c&&c.length&&(y+="//"+c),y+=d,h&&h.length&&(y+="?"+h),m&&m.length&&(y+="#"+m),y}}}function i(l,o){if(r(l)){var s=[],c="",d="",h="",m="",y="",v="",S="";if(s=n(l),c=s[1],d=s[2],h=s[3],y=s[4],v=s[5],!!c){if(o){if(c.toLowerCase()!="https")return}else if(c.toLowerCase()!="http")return;if(d)return/:(\d+)$/.test(d)&&(m=d.match(/:(\d+)$/)[0],d=d.replace(/:\d+$/,"")),S+=c+":",S+="//"+d,m&&(S+=m),S+=h,y&&y.length&&(S+="?"+y),v&&v.length&&(S+="#"+v),S}}}function a(l){return i(l,!0)}function u(l){return i(l)||a(l)}})(e)})(pP)),pP.exports}var cMe=sMe();const nX=da(cMe);var rc={},gD={exports:{}},fMe=gD.exports,rX;function dMe(){return rX||(rX=1,(function(e){(function(t,n){e.exports?e.exports=n():t.nearley=n()})(fMe,function(){function t(s,c,d){return this.id=++t.highestId,this.name=s,this.symbols=c,this.postprocess=d,this}t.highestId=0,t.prototype.toString=function(s){var c=typeof s>"u"?this.symbols.map(o).join(" "):this.symbols.slice(0,s).map(o).join(" ")+" ● "+this.symbols.slice(s).map(o).join(" ");return this.name+" → "+c};function n(s,c,d,h){this.rule=s,this.dot=c,this.reference=d,this.data=[],this.wantedBy=h,this.isComplete=this.dot===s.symbols.length}n.prototype.toString=function(){return"{"+this.rule.toString(this.dot)+"}, from: "+(this.reference||0)},n.prototype.nextState=function(s){var c=new n(this.rule,this.dot+1,this.reference,this.wantedBy);return c.left=this,c.right=s,c.isComplete&&(c.data=c.build(),c.right=void 0),c},n.prototype.build=function(){var s=[],c=this;do s.push(c.right.data),c=c.left;while(c.left);return s.reverse(),s},n.prototype.finish=function(){this.rule.postprocess&&(this.data=this.rule.postprocess(this.data,this.reference,u.fail))};function r(s,c){this.grammar=s,this.index=c,this.states=[],this.wants={},this.scannable=[],this.completed={}}r.prototype.process=function(s){for(var c=this.states,d=this.wants,h=this.completed,m=0;m<c.length;m++){var y=c[m];if(y.isComplete){if(y.finish(),y.data!==u.fail){for(var v=y.wantedBy,S=v.length;S--;){var _=v[S];this.complete(_,y)}if(y.reference===this.index){var E=y.rule.name;(this.completed[E]=this.completed[E]||[]).push(y)}}}else{var E=y.rule.symbols[y.dot];if(typeof E!="string"){this.scannable.push(y);continue}if(d[E]){if(d[E].push(y),h.hasOwnProperty(E))for(var $=h[E],S=0;S<$.length;S++){var k=$[S];this.complete(y,k)}}else d[E]=[y],this.predict(E)}}},r.prototype.predict=function(s){for(var c=this.grammar.byName[s]||[],d=0;d<c.length;d++){var h=c[d],m=this.wants[s],y=new n(h,0,this.index,m);this.states.push(y)}},r.prototype.complete=function(s,c){var d=s.nextState(c);this.states.push(d)};function i(s,c){this.rules=s,this.start=c||this.rules[0].name;var d=this.byName={};this.rules.forEach(function(h){d.hasOwnProperty(h.name)||(d[h.name]=[]),d[h.name].push(h)})}i.fromCompiled=function(h,c){var d=h.Lexer;h.ParserStart&&(c=h.ParserStart,h=h.ParserRules);var h=h.map(function(y){return new t(y.name,y.symbols,y.postprocess)}),m=new i(h,c);return m.lexer=d,m};function a(){this.reset("")}a.prototype.reset=function(s,c){this.buffer=s,this.index=0,this.line=c?c.line:1,this.lastLineBreak=c?-c.col:0},a.prototype.next=function(){if(this.index<this.buffer.length){var s=this.buffer[this.index++];return s===`
`&&(this.line+=1,this.lastLineBreak=this.index),{value:s}}},a.prototype.save=function(){return{line:this.line,col:this.index-this.lastLineBreak}},a.prototype.formatError=function(s,c){var d=this.buffer;if(typeof d=="string"){var h=d.split(`
`).slice(Math.max(0,this.line-5),this.line),m=d.indexOf(`
`,this.index);m===-1&&(m=d.length);var y=this.index-this.lastLineBreak,v=String(this.line).length;return c+=" at line "+this.line+" col "+y+`:

`,c+=h.map(function(_,E){return S(this.line-h.length+E+1,v)+" "+_},this).join(`
`),c+=`
`+S("",v+y)+`^
`,c}else return c+" at index "+(this.index-1);function S(_,E){var $=String(_);return Array(E-$.length+1).join(" ")+$}};function u(s,c,d){if(s instanceof i)var h=s,d=c;else var h=i.fromCompiled(s,c);this.grammar=h,this.options={keepHistory:!1,lexer:h.lexer||new a};for(var m in d||{})this.options[m]=d[m];this.lexer=this.options.lexer,this.lexerState=void 0;var y=new r(h,0);this.table=[y],y.wants[h.start]=[],y.predict(h.start),y.process(),this.current=0}u.fail={},u.prototype.feed=function(s){var c=this.lexer;c.reset(s,this.lexerState);for(var d;;){try{if(d=c.next(),!d)break}catch(R){var v=new r(this.grammar,this.current+1);this.table.push(v);var h=new Error(this.reportLexerError(R));throw h.offset=this.current,h.token=R.token,h}var m=this.table[this.current];this.options.keepHistory||delete this.table[this.current-1];var y=this.current+1,v=new r(this.grammar,y);this.table.push(v);for(var S=d.text!==void 0?d.text:d.value,_=c.constructor===a?d.value:d,E=m.scannable,$=E.length;$--;){var k=E[$],A=k.rule.symbols[k.dot];if(A.test?A.test(_):A.type?A.type===d.type:A.literal===S){var w=k.nextState({data:_,token:d,isToken:!0,reference:y-1});v.states.push(w)}}if(v.process(),v.states.length===0){var h=new Error(this.reportError(d));throw h.offset=this.current,h.token=d,h}this.options.keepHistory&&(m.lexerState=c.save()),this.current++}return m&&(this.lexerState=c.save()),this.results=this.finish(),this},u.prototype.reportLexerError=function(s){var c,d,h=s.token;return h?(c="input "+JSON.stringify(h.text[0])+" (lexer error)",d=this.lexer.formatError(h,"Syntax error")):(c="input (lexer error)",d=s.message),this.reportErrorCommon(d,c)},u.prototype.reportError=function(s){var c=(s.type?s.type+" token: ":"")+JSON.stringify(s.value!==void 0?s.value:s),d=this.lexer.formatError(s,"Syntax error");return this.reportErrorCommon(d,c)},u.prototype.reportErrorCommon=function(s,c){var d=[];d.push(s);var h=this.table.length-2,m=this.table[h],y=m.states.filter(function(S){var _=S.rule.symbols[S.dot];return _&&typeof _!="string"});if(y.length===0)d.push("Unexpected "+c+`. I did not expect any more input. Here is the state of my parse table:
`),this.displayStateStack(m.states,d);else{d.push("Unexpected "+c+`. Instead, I was expecting to see one of the following:
`);var v=y.map(function(S){return this.buildFirstStateStack(S,[])||[S]},this);v.forEach(function(S){var _=S[0],E=_.rule.symbols[_.dot],$=this.getSymbolDisplay(E);d.push("A "+$+" based on:"),this.displayStateStack(S,d)},this)}return d.push(""),d.join(`
`)},u.prototype.displayStateStack=function(s,c){for(var d,h=0,m=0;m<s.length;m++){var y=s[m],v=y.rule.toString(y.dot);v===d?h++:(h>0&&c.push("    ^ "+h+" more lines identical to this"),h=0,c.push("    "+v)),d=v}},u.prototype.getSymbolDisplay=function(s){return l(s)},u.prototype.buildFirstStateStack=function(s,c){if(c.indexOf(s)!==-1)return null;if(s.wantedBy.length===0)return[s];var d=s.wantedBy[0],h=[s].concat(c),m=this.buildFirstStateStack(d,h);return m===null?null:[s].concat(m)},u.prototype.save=function(){var s=this.table[this.current];return s.lexerState=this.lexerState,s},u.prototype.restore=function(s){var c=s.index;this.current=c,this.table[c]=s,this.table.splice(c+1),this.lexerState=s.lexerState,this.results=this.finish()},u.prototype.rewind=function(s){if(!this.options.keepHistory)throw new Error("set option `keepHistory` to enable rewinding");this.restore(this.table[s])},u.prototype.finish=function(){var s=[],c=this.grammar.start,d=this.table[this.table.length-1];return d.states.forEach(function(h){h.rule.name===c&&h.dot===h.rule.symbols.length&&h.reference===0&&h.data!==u.fail&&s.push(h)}),s.map(function(h){return h.data})};function l(s){var c=typeof s;if(c==="string")return s;if(c==="object"){if(s.literal)return JSON.stringify(s.literal);if(s instanceof RegExp)return"character matching "+s;if(s.type)return s.type+" token";if(s.test)return"token matching "+String(s.test);throw new Error("Unknown symbol type: "+s)}}function o(s){var c=typeof s;if(c==="string")return s;if(c==="object"){if(s.literal)return JSON.stringify(s.literal);if(s instanceof RegExp)return s.toString();if(s.type)return"%"+s.type;if(s.test)return"<"+String(s.test)+">";throw new Error("Unknown symbol type: "+s)}}return{Parser:u,Grammar:i,Rule:t}})})(gD)),gD.exports}var bE={},iX;function hMe(){if(iX)return bE;iX=1,Object.defineProperty(bE,"__esModule",{value:!0});function e(i){return i[0]}const t=i=>[].concat(...i.map(a=>Array.isArray(a)?t(a):a));function n(i){return i?Array.isArray(i)?t(i).join(""):i:""}const r={Lexer:void 0,ParserRules:[{name:"Reverse_path",symbols:["Path"]},{name:"Reverse_path$string$1",symbols:[{literal:"<"},{literal:">"}],postprocess:i=>i.join("")},{name:"Reverse_path",symbols:["Reverse_path$string$1"]},{name:"Forward_path$subexpression$1$subexpression$1",symbols:[{literal:"<"},/[pP]/,/[oO]/,/[sS]/,/[tT]/,/[mM]/,/[aA]/,/[sS]/,/[tT]/,/[eE]/,/[rR]/,{literal:"@"}],postprocess:function(i){return i.join("")}},{name:"Forward_path$subexpression$1",symbols:["Forward_path$subexpression$1$subexpression$1","Domain",{literal:">"}]},{name:"Forward_path",symbols:["Forward_path$subexpression$1"]},{name:"Forward_path$subexpression$2",symbols:[{literal:"<"},/[pP]/,/[oO]/,/[sS]/,/[tT]/,/[mM]/,/[aA]/,/[sS]/,/[tT]/,/[eE]/,/[rR]/,{literal:">"}],postprocess:function(i){return i.join("")}},{name:"Forward_path",symbols:["Forward_path$subexpression$2"]},{name:"Forward_path",symbols:["Path"]},{name:"Path$ebnf$1$subexpression$1",symbols:["A_d_l",{literal:":"}]},{name:"Path$ebnf$1",symbols:["Path$ebnf$1$subexpression$1"],postprocess:e},{name:"Path$ebnf$1",symbols:[],postprocess:()=>null},{name:"Path",symbols:[{literal:"<"},"Path$ebnf$1","Mailbox",{literal:">"}]},{name:"A_d_l$ebnf$1",symbols:[]},{name:"A_d_l$ebnf$1$subexpression$1",symbols:[{literal:","},"At_domain"]},{name:"A_d_l$ebnf$1",symbols:["A_d_l$ebnf$1","A_d_l$ebnf$1$subexpression$1"],postprocess:i=>i[0].concat([i[1]])},{name:"A_d_l",symbols:["At_domain","A_d_l$ebnf$1"]},{name:"At_domain",symbols:[{literal:"@"},"Domain"]},{name:"Domain$ebnf$1",symbols:[]},{name:"Domain$ebnf$1$subexpression$1",symbols:[{literal:"."},"sub_domain"]},{name:"Domain$ebnf$1",symbols:["Domain$ebnf$1","Domain$ebnf$1$subexpression$1"],postprocess:i=>i[0].concat([i[1]])},{name:"Domain",symbols:["sub_domain","Domain$ebnf$1"]},{name:"sub_domain",symbols:["U_label"]},{name:"Let_dig",symbols:["ALPHA_DIGIT"],postprocess:e},{name:"Ldh_str$ebnf$1",symbols:[]},{name:"Ldh_str$ebnf$1",symbols:["Ldh_str$ebnf$1","ALPHA_DIG_DASH"],postprocess:i=>i[0].concat([i[1]])},{name:"Ldh_str",symbols:["Ldh_str$ebnf$1","Let_dig"]},{name:"U_Let_dig",symbols:["ALPHA_DIGIT_U"],postprocess:e},{name:"U_Ldh_str$ebnf$1",symbols:[]},{name:"U_Ldh_str$ebnf$1",symbols:["U_Ldh_str$ebnf$1","ALPHA_DIG_DASH_U"],postprocess:i=>i[0].concat([i[1]])},{name:"U_Ldh_str",symbols:["U_Ldh_str$ebnf$1","U_Let_dig"]},{name:"U_label$ebnf$1$subexpression$1",symbols:["U_Ldh_str"]},{name:"U_label$ebnf$1",symbols:["U_label$ebnf$1$subexpression$1"],postprocess:e},{name:"U_label$ebnf$1",symbols:[],postprocess:()=>null},{name:"U_label",symbols:["U_Let_dig","U_label$ebnf$1"]},{name:"address_literal$subexpression$1",symbols:["IPv4_address_literal"]},{name:"address_literal$subexpression$1",symbols:["IPv6_address_literal"]},{name:"address_literal$subexpression$1",symbols:["General_address_literal"]},{name:"address_literal",symbols:[{literal:"["},"address_literal$subexpression$1",{literal:"]"}]},{name:"non_local_part",symbols:["Domain"],postprocess:function(i){return{DomainName:n(i[0])}}},{name:"non_local_part",symbols:["address_literal"],postprocess:function(i){return{AddressLiteral:n(i[0])}}},{name:"Mailbox",symbols:["Local_part",{literal:"@"},"non_local_part"],postprocess:function(i){return{localPart:n(i[0]),domainPart:n(i[2])}}},{name:"Local_part",symbols:["Dot_string"],postprocess:function(i){return{DotString:n(i[0])}}},{name:"Local_part",symbols:["Quoted_string"],postprocess:function(i){return{QuotedString:n(i[0])}}},{name:"Dot_string$ebnf$1",symbols:[]},{name:"Dot_string$ebnf$1$subexpression$1",symbols:[{literal:"."},"Atom"]},{name:"Dot_string$ebnf$1",symbols:["Dot_string$ebnf$1","Dot_string$ebnf$1$subexpression$1"],postprocess:i=>i[0].concat([i[1]])},{name:"Dot_string",symbols:["Atom","Dot_string$ebnf$1"]},{name:"Atom$ebnf$1",symbols:[/[0-9A-Za-z!#$%&'*+\-/=?^_`{|}~\u0080-\uFFFF/]/]},{name:"Atom$ebnf$1",symbols:["Atom$ebnf$1",/[0-9A-Za-z!#$%&'*+\-/=?^_`{|}~\u0080-\uFFFF/]/],postprocess:i=>i[0].concat([i[1]])},{name:"Atom",symbols:["Atom$ebnf$1"]},{name:"Quoted_string$ebnf$1",symbols:[]},{name:"Quoted_string$ebnf$1",symbols:["Quoted_string$ebnf$1","QcontentSMTP"],postprocess:i=>i[0].concat([i[1]])},{name:"Quoted_string",symbols:["DQUOTE","Quoted_string$ebnf$1","DQUOTE"]},{name:"QcontentSMTP",symbols:["qtextSMTP"]},{name:"QcontentSMTP",symbols:["quoted_pairSMTP"]},{name:"quoted_pairSMTP",symbols:[{literal:"\\"},/[\x20-\x7e]/]},{name:"qtextSMTP",symbols:[/[\x20-\x21\x23-\x5b\x5d-\x7e\u0080-\uFFFF]/],postprocess:e},{name:"IPv4_address_literal$macrocall$2",symbols:[{literal:"."},"Snum"]},{name:"IPv4_address_literal$macrocall$1",symbols:["IPv4_address_literal$macrocall$2","IPv4_address_literal$macrocall$2","IPv4_address_literal$macrocall$2"]},{name:"IPv4_address_literal",symbols:["Snum","IPv4_address_literal$macrocall$1"]},{name:"IPv6_address_literal$subexpression$1",symbols:[/[iI]/,/[pP]/,/[vV]/,{literal:"6"},{literal:":"}],postprocess:function(i){return i.join("")}},{name:"IPv6_address_literal",symbols:["IPv6_address_literal$subexpression$1","IPv6_addr"]},{name:"General_address_literal$ebnf$1",symbols:["dcontent"]},{name:"General_address_literal$ebnf$1",symbols:["General_address_literal$ebnf$1","dcontent"],postprocess:i=>i[0].concat([i[1]])},{name:"General_address_literal",symbols:["Standardized_tag",{literal:":"},"General_address_literal$ebnf$1"]},{name:"Standardized_tag",symbols:["Ldh_str"]},{name:"dcontent",symbols:[/[\x21-\x5a\x5e-\x7e]/],postprocess:e},{name:"Snum",symbols:["DIGIT"]},{name:"Snum$subexpression$1",symbols:[/[1-9]/,"DIGIT"]},{name:"Snum",symbols:["Snum$subexpression$1"]},{name:"Snum$subexpression$2",symbols:[{literal:"1"},"DIGIT","DIGIT"]},{name:"Snum",symbols:["Snum$subexpression$2"]},{name:"Snum$subexpression$3",symbols:[{literal:"2"},/[0-4]/,"DIGIT"]},{name:"Snum",symbols:["Snum$subexpression$3"]},{name:"Snum$subexpression$4",symbols:[{literal:"2"},{literal:"5"},/[0-5]/]},{name:"Snum",symbols:["Snum$subexpression$4"]},{name:"IPv6_addr",symbols:["IPv6_full"]},{name:"IPv6_addr",symbols:["IPv6_comp"]},{name:"IPv6_addr",symbols:["IPv6v4_full"]},{name:"IPv6_addr",symbols:["IPv6v4_comp"]},{name:"IPv6_hex",symbols:["HEXDIG"]},{name:"IPv6_hex$subexpression$1",symbols:["HEXDIG","HEXDIG"]},{name:"IPv6_hex",symbols:["IPv6_hex$subexpression$1"]},{name:"IPv6_hex$subexpression$2",symbols:["HEXDIG","HEXDIG","HEXDIG"]},{name:"IPv6_hex",symbols:["IPv6_hex$subexpression$2"]},{name:"IPv6_hex$subexpression$3",symbols:["HEXDIG","HEXDIG","HEXDIG","HEXDIG"]},{name:"IPv6_hex",symbols:["IPv6_hex$subexpression$3"]},{name:"IPv6_full$macrocall$2",symbols:[{literal:":"},"IPv6_hex"]},{name:"IPv6_full$macrocall$1",symbols:["IPv6_full$macrocall$2","IPv6_full$macrocall$2","IPv6_full$macrocall$2","IPv6_full$macrocall$2","IPv6_full$macrocall$2","IPv6_full$macrocall$2","IPv6_full$macrocall$2"]},{name:"IPv6_full",symbols:["IPv6_hex","IPv6_full$macrocall$1"]},{name:"IPv6_comp$ebnf$1$subexpression$1$macrocall$2",symbols:[{literal:":"},"IPv6_hex"]},{name:"IPv6_comp$ebnf$1$subexpression$1$macrocall$1",symbols:["IPv6_comp$ebnf$1$subexpression$1$macrocall$2","IPv6_comp$ebnf$1$subexpression$1$macrocall$2","IPv6_comp$ebnf$1$subexpression$1$macrocall$2","IPv6_comp$ebnf$1$subexpression$1$macrocall$2","IPv6_comp$ebnf$1$subexpression$1$macrocall$2"]},{name:"IPv6_comp$ebnf$1$subexpression$1",symbols:["IPv6_hex","IPv6_comp$ebnf$1$subexpression$1$macrocall$1"]},{name:"IPv6_comp$ebnf$1",symbols:["IPv6_comp$ebnf$1$subexpression$1"],postprocess:e},{name:"IPv6_comp$ebnf$1",symbols:[],postprocess:()=>null},{name:"IPv6_comp$string$1",symbols:[{literal:":"},{literal:":"}],postprocess:i=>i.join("")},{name:"IPv6_comp$ebnf$2$subexpression$1$macrocall$2",symbols:[{literal:":"},"IPv6_hex"]},{name:"IPv6_comp$ebnf$2$subexpression$1$macrocall$1",symbols:["IPv6_comp$ebnf$2$subexpression$1$macrocall$2","IPv6_comp$ebnf$2$subexpression$1$macrocall$2","IPv6_comp$ebnf$2$subexpression$1$macrocall$2","IPv6_comp$ebnf$2$subexpression$1$macrocall$2","IPv6_comp$ebnf$2$subexpression$1$macrocall$2"]},{name:"IPv6_comp$ebnf$2$subexpression$1",symbols:["IPv6_hex","IPv6_comp$ebnf$2$subexpression$1$macrocall$1"]},{name:"IPv6_comp$ebnf$2",symbols:["IPv6_comp$ebnf$2$subexpression$1"],postprocess:e},{name:"IPv6_comp$ebnf$2",symbols:[],postprocess:()=>null},{name:"IPv6_comp",symbols:["IPv6_comp$ebnf$1","IPv6_comp$string$1","IPv6_comp$ebnf$2"]},{name:"IPv6v4_full$macrocall$2",symbols:[{literal:":"},"IPv6_hex"]},{name:"IPv6v4_full$macrocall$1",symbols:["IPv6v4_full$macrocall$2","IPv6v4_full$macrocall$2","IPv6v4_full$macrocall$2","IPv6v4_full$macrocall$2","IPv6v4_full$macrocall$2"]},{name:"IPv6v4_full",symbols:["IPv6_hex","IPv6v4_full$macrocall$1",{literal:":"},"IPv4_address_literal"]},{name:"IPv6v4_comp$ebnf$1$subexpression$1$macrocall$2",symbols:[{literal:":"},"IPv6_hex"]},{name:"IPv6v4_comp$ebnf$1$subexpression$1$macrocall$1",symbols:["IPv6v4_comp$ebnf$1$subexpression$1$macrocall$2","IPv6v4_comp$ebnf$1$subexpression$1$macrocall$2","IPv6v4_comp$ebnf$1$subexpression$1$macrocall$2"]},{name:"IPv6v4_comp$ebnf$1$subexpression$1",symbols:["IPv6_hex","IPv6v4_comp$ebnf$1$subexpression$1$macrocall$1"]},{name:"IPv6v4_comp$ebnf$1",symbols:["IPv6v4_comp$ebnf$1$subexpression$1"],postprocess:e},{name:"IPv6v4_comp$ebnf$1",symbols:[],postprocess:()=>null},{name:"IPv6v4_comp$string$1",symbols:[{literal:":"},{literal:":"}],postprocess:i=>i.join("")},{name:"IPv6v4_comp$ebnf$2$subexpression$1$macrocall$2",symbols:[{literal:":"},"IPv6_hex"]},{name:"IPv6v4_comp$ebnf$2$subexpression$1$macrocall$1",symbols:["IPv6v4_comp$ebnf$2$subexpression$1$macrocall$2","IPv6v4_comp$ebnf$2$subexpression$1$macrocall$2","IPv6v4_comp$ebnf$2$subexpression$1$macrocall$2"]},{name:"IPv6v4_comp$ebnf$2$subexpression$1",symbols:["IPv6_hex","IPv6v4_comp$ebnf$2$subexpression$1$macrocall$1",{literal:":"}]},{name:"IPv6v4_comp$ebnf$2",symbols:["IPv6v4_comp$ebnf$2$subexpression$1"],postprocess:e},{name:"IPv6v4_comp$ebnf$2",symbols:[],postprocess:()=>null},{name:"IPv6v4_comp",symbols:["IPv6v4_comp$ebnf$1","IPv6v4_comp$string$1","IPv6v4_comp$ebnf$2","IPv4_address_literal"]},{name:"DIGIT",symbols:[/[0-9]/],postprocess:e},{name:"ALPHA_DIGIT_U",symbols:[/[0-9A-Za-z\u0080-\uFFFF]/],postprocess:e},{name:"ALPHA_DIGIT",symbols:[/[0-9A-Za-z]/],postprocess:e},{name:"ALPHA_DIG_DASH",symbols:[/[-0-9A-Za-z]/],postprocess:e},{name:"ALPHA_DIG_DASH_U",symbols:[/[-0-9A-Za-z\u0080-\uFFFF]/],postprocess:e},{name:"HEXDIG",symbols:[/[0-9A-Fa-f]/],postprocess:e},{name:"DQUOTE",symbols:[{literal:'"'}],postprocess:e}],ParserStart:"Reverse_path"};return bE.default=r,bE}var aX;function pMe(){if(aX)return rc;aX=1;var e=rc&&rc.__importDefault||function(s){return s&&s.__esModule?s:{default:s}};Object.defineProperty(rc,"__esModule",{value:!0}),rc.canonicalize=rc.canonicalize_quoted_string=rc.normalize=rc.normalize_dot_string=rc.parse=void 0;const t=dMe(),n=e(hMe());n.default.ParserStart="Mailbox";const r=t.Grammar.fromCompiled(n.default);function i(s){const c=new t.Parser(r);if(c.feed(s),c.results.length!==1)throw new Error("address parsing failed: ambiguous grammar");return c.results[0]}rc.parse=i;function a(s){return(function(){const h=s.indexOf("+");return h===-1?s:s.substr(0,h)})().replace(/\./g,"").toLowerCase()}rc.normalize_dot_string=a;function u(s){var c,d;const h=i(s),m=(c=h.domainPart.AddressLiteral)!==null&&c!==void 0?c:h.domainPart.DomainName.toLowerCase();return`${(d=h.localPart.QuotedString)!==null&&d!==void 0?d:a(h.localPart.DotString)}@${m}`}rc.normalize=u;function l(s){return`"${s.substr(1).substr(0,s.length-2).replace(/(?:\\(.))/g,"$1").replace(/(?:(["\\]))/g,"\\$1")}"`}rc.canonicalize_quoted_string=l;function o(s){var c;const d=i(s),h=(c=d.domainPart.AddressLiteral)!==null&&c!==void 0?c:d.domainPart.DomainName.toLowerCase();return`${d.localPart.QuotedString?l(d.localPart.QuotedString):d.localPart.DotString}@${h}`}return rc.canonicalize=o,rc}var mMe=pMe();const gMe=new RegExp("^([0-9]+)-(0[1-9]|1[012])-(0[1-9]|[12][0-9]|3[01])[Tt]([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9]|60)(\\.[0-9]+)?(([Zz])|([\\+|\\-]([01][0-9]|2[0-3]):[0-5][0-9]))$"),vMe=/^(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?)$/,yMe=/^((([0-9a-f]{1,4}:){7}([0-9a-f]{1,4}|:))|(([0-9a-f]{1,4}:){6}(:[0-9a-f]{1,4}|((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9a-f]{1,4}:){5}(((:[0-9a-f]{1,4}){1,2})|:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9a-f]{1,4}:){4}(((:[0-9a-f]{1,4}){1,3})|((:[0-9a-f]{1,4})?:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9a-f]{1,4}:){3}(((:[0-9a-f]{1,4}){1,4})|((:[0-9a-f]{1,4}){0,2}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9a-f]{1,4}:){2}(((:[0-9a-f]{1,4}){1,5})|((:[0-9a-f]{1,4}){0,3}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9a-f]{1,4}:){1}(((:[0-9a-f]{1,4}){1,6})|((:[0-9a-f]{1,4}){0,4}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(:(((:[0-9a-f]{1,4}){1,7})|((:[0-9a-f]{1,4}){0,5}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:)))$/i,bMe=/^(?=.{1,255}$)[0-9A-Za-z](?:(?:[0-9A-Za-z]|-){0,61}[0-9A-Za-z])?(?:\.[0-9A-Za-z](?:(?:[0-9A-Za-z]|-){0,61}[0-9A-Za-z])?)*\.?$/,OMe=/^(\d\d\d\d)-(\d\d)-(\d\d)$/,xMe=/^(?:[0-2]\d:[0-5]\d:[0-5]\d|23:59:60)(?:\.\d+)?(?:z|[+-]\d\d(?::?\d\d)?)?$/i,SMe=[0,31,28,31,30,31,30,31,31,30,31,30,31],_Me=/^(?:\/(?:[^~/]|~0|~1)*)*$/,EMe=/^(?:0|[1-9][0-9]*)(?:#|(?:\/(?:[^~/]|~0|~1)*)*)$/,DMe=/^(?:[a-z][a-z0-9+\-.]*:)?(?:\/?\/(?:(?:[a-z0-9\-._~!$&'()*+,;=:]|%[0-9a-f]{2})*@)?(?:\[(?:(?:(?:(?:[0-9a-f]{1,4}:){6}|::(?:[0-9a-f]{1,4}:){5}|(?:[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){4}|(?:(?:[0-9a-f]{1,4}:){0,1}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){3}|(?:(?:[0-9a-f]{1,4}:){0,2}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){2}|(?:(?:[0-9a-f]{1,4}:){0,3}[0-9a-f]{1,4})?::[0-9a-f]{1,4}:|(?:(?:[0-9a-f]{1,4}:){0,4}[0-9a-f]{1,4})?::)(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?))|(?:(?:[0-9a-f]{1,4}:){0,5}[0-9a-f]{1,4})?::[0-9a-f]{1,4}|(?:(?:[0-9a-f]{1,4}:){0,6}[0-9a-f]{1,4})?::)|[Vv][0-9a-f]+\.[a-z0-9\-._~!$&'()*+,;=:]+)\]|(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?)|(?:[a-z0-9\-._~!$&'"()*+,;=]|%[0-9a-f]{2})*)(?::\d*)?(?:\/(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})*)*|\/(?:(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})+(?:\/(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})*)*)?|(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})+(?:\/(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})*)*)?(?:\?(?:[a-z0-9\-._~!$&'"()*+,;=:@/?]|%[0-9a-f]{2})*)?(?:#(?:[a-z0-9\-._~!$&'"()*+,;=:@/?]|%[0-9a-f]{2})*)?$/i,$Me=/^(?:(?:[^\x00-\x20"'<>%\\^`{|}]|%[0-9a-f]{2})|\{[+#./;?&=,!@|]?(?:[a-z0-9_]|%[0-9a-f]{2})+(?::[1-9][0-9]{0,3}|\*)?(?:,(?:[a-z0-9_]|%[0-9a-f]{2})+(?::[1-9][0-9]{0,3}|\*)?)*\})*$/i,X0e={date:(e,t,n,r)=>{if(typeof n!="string"||n==="")return;const i=n.match(OMe);if(!i)return e.errors.formatDateTimeError({value:n,pointer:r,schema:t});const a=+i[1],u=+i[2],l=+i[3],o=a%4===0&&(a%100!==0||a%400===0);if(!(u>=1&&u<=12&&l>=1&&l<=(u==2&&o?29:SMe[u])))return e.errors.formatDateError({value:n,pointer:r,schema:t})},"date-time":(e,t,n,r)=>{if(!(typeof n!="string"||n===""))return n===""||gMe.test(n)?new Date(n).toString()==="Invalid Date"?e.errors.formatDateTimeError({value:n,pointer:r,schema:t}):void 0:e.errors.formatDateTimeError({value:n,pointer:r,schema:t})},email:(e,t,n,r)=>{if(typeof n!="string"||n==="")return;if(n[0]==='"')return e.errors.formatEmailError({value:n,pointer:r,schema:t});const[i,a,...u]=n.split("@");if(!i||!a||u.length!==0||i.length>64||a.length>253)return e.errors.formatEmailError({value:n,pointer:r,schema:t});if(i[0]==="."||i.endsWith(".")||i.includes(".."))return e.errors.formatEmailError({value:n,pointer:r,schema:t});if(!/^[a-z0-9.-]+$/i.test(a)||!/^[a-z0-9.!#$%&'*+/=?^_`{|}~-]+$/i.test(i))return e.errors.formatEmailError({value:n,pointer:r,schema:t});if(!a.split(".").every(l=>/^[a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?$/i.test(l)))return e.errors.formatEmailError({value:n,pointer:r,schema:t})},"idn-email":(e,t,n,r)=>{if(!(typeof n!="string"||n===""))try{mMe.parse(n);return}catch{return e.errors.formatEmailError({value:n,pointer:r,schema:t})}},hostname:(e,t,n,r)=>{if(typeof n=="string"&&!(n===""||bMe.test(n)))return e.errors.formatHostnameError({value:n,pointer:r,schema:t})},ipv4:(e,t,n,r)=>{if(!(typeof n!="string"||n==="")){if(n&&n[0]==="0")return e.errors.formatIPV4LeadingZeroError({value:n,pointer:r,schema:t});if(!(n.length<=15&&vMe.test(n)))return e.errors.formatIPV4Error({value:n,pointer:r,schema:t})}},ipv6:(e,t,n,r)=>{if(!(typeof n!="string"||n==="")){if(n&&n[0]==="0")return e.errors.formatIPV6LeadingZeroError({value:n,pointer:r,schema:t});if(!(n.length<=45&&yMe.test(n)))return e.errors.formatIPV6Error({value:n,pointer:r,schema:t})}},"json-pointer":(e,t,n,r)=>{if(!(typeof n!="string"||n==="")&&!_Me.test(n))return e.errors.formatJsonPointerError({value:n,pointer:r,schema:t})},"relative-json-pointer":(e,t,n,r)=>{if(!(typeof n!="string"||n==="")&&!EMe.test(n))return e.errors.formatJsonPointerError({value:n,pointer:r,schema:t})},regex:(e,t,n,r)=>{if(typeof n=="string"&&/\\Z$/.test(n)===!1){try{new RegExp(n);return}catch{}return e.errors.formatRegExError({value:n,pointer:r,schema:t})}if(!(typeof n=="object"||typeof n=="number"||Array.isArray(n)))return e.errors.formatRegExError({value:n,pointer:r,schema:t})},time:(e,t,n,r)=>typeof n!="string"||n===""||n.match(xMe)?void 0:e.errors.formatDateTimeError({value:n,pointer:r,schema:t}),uri:(e,t,n,r)=>{if(!(typeof n!="string"||n==="")&&!nX.isUri(n))return e.errors.formatURIError({value:n,pointer:r,schema:t})},"uri-reference":(e,t,n,r)=>{if(!(typeof n!="string"||n==="")&&!DMe.test(n))return e.errors.formatURIReferenceError({value:n,pointer:r,schema:t})},"uri-template":(e,t,n,r)=>{if(!(typeof n!="string"||n==="")&&!$Me.test(n))return e.errors.formatURITemplateError({value:n,pointer:r,schema:t})},url:(e,t,n,r)=>{if(!(n===""||nX.isWebUri(n)))return e.errors.formatURLError({value:n,pointer:r,schema:t})}},Y0e={};function G0e(e,t=Y0e){const{pointer:n="#",data:r,schema:i=e.rootSchema,withSchemaWarning:a=!1}=t,u=Bx.split(n),l=K0e(e,e.resolveRef(i),u,"#",r);if(!(!a&&(l==null?void 0:l.code)==="schema-warning"))return l}function K0e(e,t,n,r,i=Y0e){if(n.length===0)return e.resolveRef(t);const a=n.shift();return t=e.step(a,t,i,r),Gf(t)?t:(i=i[a],K0e(e,t,n,`${r}/${a}`,i))}function CMe(e){switch(gi(e)){case"string":case"array":return(e==null?void 0:e.length)===0;case"null":case"undefined":return!0;case"object":return Object.keys(e).length===0;default:return!1}}const uX=a4.templateDefaultOptions;let Xm;function mP(e,t){const{$ref:n}=e;return n==null?!0:(Xm[t]==null||Xm[t][n]==null?0:Xm[t][n])<a4.GET_TEMPLATE_RECURSION_LIMIT}function lX(e,t,n){const{$ref:r}=t;return r==null?t:(Xm[n]=Xm[n]||{},Xm[n][r]=Xm[n][r]||0,Xm[n][r]+=1,e.resolveRef(t))}function wMe(e,t){if(e==="string")return JSON.stringify(t);if(typeof t!="string")return null;try{if(t=JSON.parse(t),typeof t===e)return t}catch{}return null}function W0e(e,t,n,r,i){if(gi(t)!=="object")return Object.assign({pointer:r},t);if(mP(t,r)===!1&&n==null)return!1;let a=$$(lX(e,t,r));if(Array.isArray(t.anyOf)&&t.anyOf.length>0){if(mP(t.anyOf[0],`${r}/anyOf/0`)){const u=lX(e,t.anyOf[0],`${r}/anyOf/0`);a=ZI(a,u),a.pointer=t.anyOf[0].$ref||a.pointer}delete a.anyOf}if(Array.isArray(t.allOf)&&t.allOf.map((l,o)=>mP(l,`${r}/allOf/${o}`)).reduceRight((l,o)=>l&&o,!0)){const l=[];let o=$$(n);for(let c=0;c<t.allOf.length;c+=1)l.push(F0e(e,t.allOf[c],o)),o=Yh(e,o,{type:t.type,...l[c]},`${r}/allOf/${c}`,i);const s=I0e(e,{allOf:l});s&&(a=oc(a,s))}return a.pointer=a.pointer||t.$ref||r,a}const oX=e=>e&&typeof e=="object";function Yh(e,t,n,r,i){var a;if(n==null)throw new Error(`getTemplate: missing schema for data: ${JSON.stringify(t)}`);if(r==null)throw new Error("Missing pointer");let u=W0e(e,n,t,r,i);if(!oX(u))return;if(r=u.pointer,u!=null&&u.const)return u.const;if(Array.isArray(u.oneOf))if(CMe(t)){const c=u.oneOf[0].type||u.type||u.const&&typeof u.const||gi(t);u={...u.oneOf[0],type:c}}else{const c=jI(e,t,u);if(Gf(c)){if(t!=null&&i.removeInvalidData!==!0)return t;u=u.oneOf[0],t=void 0}else c.type=(a=c.type)!==null&&a!==void 0?a:u.type,u=c}if(!oX(u)||u.type==null)return;if(t instanceof File)return t;const l=Array.isArray(u.type)?RMe(u.type,t,u.default):u.type,o=gi(t);return t!=null&&o!==l&&!(o==="number"&&l==="integer")&&(t=wMe(l,t)),sX[l]==null?i.removeInvalidData?void 0:t:sX[l](e,u,t,r,i)}function RMe(e,t,n){if(t==null){if(n!=null){const i=gi(n);if(e.includes(i))return i}return e[0]}const r=gi(t);return e.includes(r)?r:e[0]}const sX={null:(e,t,n)=>f2(t,n,null),string:(e,t,n)=>f2(t,n,""),number:(e,t,n)=>f2(t,n,0),integer:(e,t,n)=>f2(t,n,0),boolean:(e,t,n)=>f2(t,n,!1),object:(e,t,n,r,i)=>{var a;const u=t.default===void 0?{}:t.default,l={},o=i.extendDefaults===!1&&t.default!==void 0?[]:(a=t.required)!==null&&a!==void 0?a:[];t.properties&&Object.keys(t.properties).forEach(d=>{const h=n==null||n[d]==null?u[d]:n[d],m=o.includes(d);(h!=null||m||i.addOptionalProps)&&(l[d]=Yh(e,h,t.properties[d],`${r}/properties/${d}`,i))});let s=B0e(e,t,l);if(s){s=oc(t,s),delete s.dependencies;const d=Yh(e,n,s,`${r}/dependencies`,i);Object.assign(l,d)}n&&(i.removeInvalidData===!0&&(t.additionalProperties===!1||gi(t.additionalProperties)==="object")?gi(t.additionalProperties)==="object"&&Object.keys(n).forEach(d=>{l[d]==null&&e.isValid(n[d],t.additionalProperties)&&(l[d]=n[d])}):Object.keys(n).forEach(d=>l[d]==null&&(l[d]=n[d])));const c=Ew(e,t,l);if(c){const d=Yh(e,l,{type:"object",...c},r,i);Object.assign(l,d)}return l},array:(e,t,n,r,i)=>{var a,u;if(t.items==null)return n||[];const l=t.default===void 0?[]:t.default,o=n||l,s=i.extendDefaults===!1&&t.default!==void 0?0:t.minItems||0;if(Array.isArray(t.items)){for(let d=0,h=Math.max(s??0,(u=(a=t.items)===null||a===void 0?void 0:a.length)!==null&&u!==void 0?u:0);d<h;d+=1)o[d]=Yh(e,o[d]==null?l[d]:o[d],t.items[d],`${r}/items/${d}`,i);return o}if(gi(t.items)!=="object")return o;const c=W0e(e,t.items,n,r,i);if(c===!1)return o;if(r=c.pointer||r,c.oneOf&&o.length===0){const d=c.oneOf[0];for(let h=0;h<s;h+=1)o[h]=Yh(e,o[h]==null?l[h]:o[h],d,`${r}/oneOf/0`,i);return o}if(c.oneOf&&o.length>0){const d=Math.max(s,o.length);for(let h=0;h<d;h+=1){let m=o[h]==null?l[h]:o[h],y=jI(e,m,c);y==null||Gf(y)?m!=null&&i.removeInvalidData!==!0?o[h]=m:(m=void 0,y=c.oneOf[0],o[h]=Yh(e,m,y,`${r}/oneOf/${h}`,i)):o[h]=Yh(e,m,y,`${r}/oneOf/${h}`,i)}return o}if(c.type){for(let d=0,h=Math.max(s,o.length);d<h;d+=1)o[d]=Yh(e,o[d]==null?l[d]:o[d],c,`${r}/items`,i);return o}return o}};function f2(e,t,n){return t??(e.const?e.const:e.default===void 0&&Array.isArray(e.enum)?e.enum[0]:e.default===void 0?n:e.default)}const J0e=(e,t,n=e.rootSchema,r)=>(Xm={},r?Yh(e,t,n,"#",{...uX,...r}):Yh(e,t,n,"#",uX));function eme(e,t,n=e.rootSchema,r="#"){return e.validate(t,n,r).length===0}function cX(e){const t=[];let n=0;const r=e.length;for(;n<r;){const i=e.charCodeAt(n++);if(i>=55296&&i<=56319&&n<r){const a=e.charCodeAt(n++);(a&64512)==56320?t.push(((i&1023)<<10)+(a&1023)+65536):(t.push(i),n--)}else t.push(i)}return t}function fX(e){const t=`${e}`,n=t.indexOf(".");return n===-1?0:t.length-(n+1)}var gP,dX;function AMe(){return dX||(dX=1,gP=function e(t,n){if(t===n)return!0;if(t&&n&&typeof t=="object"&&typeof n=="object"){if(t.constructor!==n.constructor)return!1;var r,i,a;if(Array.isArray(t)){if(r=t.length,r!=n.length)return!1;for(i=r;i--!==0;)if(!e(t[i],n[i]))return!1;return!0}if(t.constructor===RegExp)return t.source===n.source&&t.flags===n.flags;if(t.valueOf!==Object.prototype.valueOf)return t.valueOf()===n.valueOf();if(t.toString!==Object.prototype.toString)return t.toString()===n.toString();if(a=Object.keys(t),r=a.length,r!==Object.keys(n).length)return!1;for(i=r;i--!==0;)if(!Object.prototype.hasOwnProperty.call(n,a[i]))return!1;for(i=r;i--!==0;){var u=a[i];if(!e(t[u],n[u]))return!1}return!0}return t!==t&&n!==n}),gP}var TMe=AMe();const tme=da(TMe),kMe=Object.prototype.hasOwnProperty,hX=(e,t)=>!(e[t]===void 0||!kMe.call(e,t)),nme={additionalProperties:(e,t,n,r)=>{if(t.additionalProperties===!0||t.additionalProperties==null||gi(t.patternProperties)==="object"&&t.additionalProperties===!1)return;const i=[];let a=Object.keys(n).filter(l=>a4.propertyBlacklist.includes(l)===!1);const u=Object.keys(t.properties||{});if(gi(t.patternProperties)==="object"){const l=Object.keys(t.patternProperties).map(o=>new RegExp(o));a=a.filter(o=>{for(let s=0;s<l.length;s+=1)if(l[s].test(o))return!1;return!0})}for(let l=0,o=a.length;l<o;l+=1){const s=a[l];if(u.indexOf(s)===-1){const c=jv(t.additionalProperties);if(c&&Array.isArray(t.additionalProperties.oneOf)){const d=e.resolveOneOf(n[s],t.additionalProperties,`${r}/${s}`);Gf(d)?i.push(e.errors.additionalPropertiesError({pointer:r,schema:t.additionalProperties,value:n,property:a[l],properties:u,errors:d.data.errors})):i.push(...e.validate(n[s],d,r))}else c?i.push(...e.validate(n[s],t.additionalProperties,`${r}/${s}`)):i.push(e.errors.noAdditionalPropertiesError({pointer:r,schema:t,value:n,property:a[l],properties:u}))}}return i},allOf:y9e,anyOf:x9e,dependencies:O9e,enum:(e,t,n,r)=>{const i=gi(n);if(i==="object"||i==="array"){const a=JSON.stringify(n);for(let u=0;u<t.enum.length;u+=1)if(JSON.stringify(t.enum[u])===a)return}else if(t.enum.includes(n))return;return e.errors.enumError({pointer:r,schema:t,value:n,values:t.enum})},format:(e,t,n,r)=>{if(e.validateFormat[t.format])return e.validateFormat[t.format](e,t,n,r)},items:(e,t,n,r)=>{if(t.items===!1)return Array.isArray(n)&&n.length===0?void 0:e.errors.invalidDataError({pointer:r,value:n,schema:t});const i=[];for(let a=0;a<n.length;a+=1){const u=n[a],l=e.step(a,t,n,r);if(Gf(l))return[l];const o=e.validate(u,l,`${r}/${a}`);i.push(...o)}return i},maximum:(e,t,n,r)=>{if(!isNaN(t.maximum)){if(t.maximum&&t.maximum<n)return e.errors.maximumError({maximum:t.maximum,length:n,value:n,pointer:r,schema:t});if(t.maximum&&t.exclusiveMaximum===!0&&t.maximum===n)return e.errors.maximumError({maximum:t.maximum,length:n,pointer:r,schema:t,value:n})}},maxItems:(e,t,n,r)=>{if(!isNaN(t.maxItems)&&t.maxItems<n.length)return e.errors.maxItemsError({maximum:t.maxItems,length:n.length,schema:t,value:n,pointer:r})},maxLength:(e,t,n,r)=>{if(isNaN(t.maxLength))return;const i=cX(n).length;if(t.maxLength<i)return e.errors.maxLengthError({maxLength:t.maxLength,length:i,pointer:r,schema:t,value:n})},maxProperties:(e,t,n,r)=>{const i=Object.keys(n).length;if(isNaN(t.maxProperties)===!1&&t.maxProperties<i)return e.errors.maxPropertiesError({maxProperties:t.maxProperties,length:i,pointer:r,schema:t,value:n})},minLength:(e,t,n,r)=>{if(isNaN(t.minLength))return;const i=cX(n).length;if(t.minLength>i)return t.minLength===1?e.errors.minLengthOneError({minLength:t.minLength,length:i,pointer:r,schema:t,value:n}):e.errors.minLengthError({minLength:t.minLength,length:i,pointer:r,schema:t,value:n})},minimum:(e,t,n,r)=>{if(!isNaN(t.minimum)){if(t.minimum>n)return e.errors.minimumError({minimum:t.minimum,length:n,pointer:r,schema:t,value:n});if(t.exclusiveMinimum===!0&&t.minimum===n)return e.errors.minimumError({minimum:t.minimum,length:n,pointer:r,schema:t,value:n})}},minItems:(e,t,n,r)=>{if(!isNaN(t.minItems)&&t.minItems>n.length)return t.minItems===1?e.errors.minItemsOneError({minItems:t.minItems,length:n.length,pointer:r,schema:t,value:n}):e.errors.minItemsError({minItems:t.minItems,length:n.length,pointer:r,schema:t,value:n})},minProperties:(e,t,n,r)=>{if(isNaN(t.minProperties))return;const i=Object.keys(n).length;if(t.minProperties>i)return e.errors.minPropertiesError({minProperties:t.minProperties,length:i,pointer:r,schema:t,value:n})},multipleOf:(e,t,n,r)=>{if(isNaN(t.multipleOf)||typeof n!="number")return;const i=fX(n),a=fX(t.multipleOf);if(i>a)return e.errors.multipleOfError({multipleOf:t.multipleOf,value:n,pointer:r,schema:t});const u=Math.pow(10,a),l=Math.round(n*u),o=Math.round(t.multipleOf*u);if(l%o/u!==0)return e.errors.multipleOfError({multipleOf:t.multipleOf,value:n,pointer:r,schema:t})},not:(e,t,n,r)=>{const i=[];return e.validate(n,t.not,r).length===0&&i.push(e.errors.notError({value:n,not:t.not,pointer:r,schema:t})),i},oneOf:k7e,pattern:(e,t,n,r)=>{if(new RegExp(t.pattern,"u").test(n)===!1)return e.errors.patternError({pattern:t.pattern,description:t.patternExample||t.pattern,received:n,schema:t,value:n,pointer:r})},patternProperties:(e,t,n,r)=>{const i=t.properties||{},a=t.patternProperties;if(gi(a)!=="object")return;const u=[],l=Object.keys(n),o=Object.keys(a).map(s=>({regex:new RegExp(s),patternSchema:a[s]}));return l.forEach(s=>{let c=!1;for(let d=0,h=o.length;d<h;d+=1)if(o[d].regex.test(s)){c=!0;const m=e.validate(n[s],o[d].patternSchema,`${r}/${s}`);m&&m.length>0&&u.push(...m)}i[s]||c===!1&&t.additionalProperties===!1&&u.push(e.errors.patternPropertiesError({key:s,pointer:r,schema:t,value:n,patterns:Object.keys(a).join(",")}))}),u},properties:(e,t,n,r)=>{const i=[],a=Object.keys(t.properties||{});for(let u=0;u<a.length;u+=1){const l=a[u];if(hX(n,l)){const o=e.step(l,t,n,r),s=e.validate(n[l],o,`${r}/${l}`);i.push(...s)}}return i},propertiesRequired:(e,t,n,r)=>{const i=[],a=Object.keys(t.properties||{});for(let u=0;u<a.length;u+=1){const l=a[u];if(n[l]===void 0)i.push(e.errors.requiredPropertyError({key:l,pointer:r,schema:t,value:n}));else{const o=e.step(l,t,n,r),s=e.validate(n[l],o,`${r}/${l}`);i.push(...s)}}return i},required:(e,t,n,r)=>{if(Array.isArray(t.required)!==!1)return t.required.map(i=>{if(!hX(n,i))return e.errors.requiredPropertyError({key:i,pointer:r,schema:t,value:n})})},requiredNotEmpty:(e,t,n,r)=>{if(Array.isArray(t.required)!==!1)return t.required.map(i=>{if(n[i]==null||n[i]==="")return e.errors.valueNotEmptyError({property:i,pointer:`${r}/${i}`,schema:t,value:n})})},uniqueItems:(e,t,n,r)=>{if((Array.isArray(n)&&t.uniqueItems)===!1)return;const i=[],a=[];return n.forEach((u,l)=>{for(let o=l+1;o<n.length;o+=1)tme(u,n[o])&&!i.includes(o)&&(a.push(e.errors.uniqueItemsError({pointer:`${r}/${o}`,duplicatePointer:`${r}/${l}`,arrayPointer:r,value:JSON.stringify(u),schema:t})),i.push(o))}),a}};function g0(e){if(e===void 0)return;const t={type:gi(e)};return t.type==="object"&&jv(e)&&(t.properties={},Object.keys(e).forEach(n=>t.properties[n]=g0(e[n]))),t.type==="array"&&Array.isArray(e)&&(e.length===1?t.items=g0(e[0]):t.items=e.map(g0)),t}const PMe={array:(e,t,n,r,i)=>{const a=r==null?void 0:r[t],u=gi(n.items);if(u==="object")return UH(e,n.items,a,`${i}/${t}`)||e.resolveRef(n.items);if(u==="array"){if(n.items[t]===!0)return g0(a);if(n.items[t]===!1)return e.errors.invalidDataError({key:t,value:a,pointer:i,schema:n});if(n.items[t])return e.resolveRef(n.items[t]);if(n.additionalItems===!1)return e.errors.additionalItemsError({key:t,value:a,pointer:i,schema:n});if(n.additionalItems===!0||n.additionalItems===void 0)return g0(a);if(gi(n.additionalItems)==="object")return n.additionalItems;throw new Error(`Invalid schema ${JSON.stringify(n,null,2)} for ${JSON.stringify(r,null,2)}`)}return n.additionalItems!==!1&&a?g0(a):new Error(`Invalid array schema for ${t} at ${i}`)},object:(e,t,n,r,i)=>{var a;n=UH(e,n,r,i);const u=(a=n==null?void 0:n.properties)===null||a===void 0?void 0:a[t];if(u!==void 0){if(u===!1)return e.errors.forbiddenPropertyError({property:t,value:r,pointer:i,schema:n});if(u===!0)return g0(r==null?void 0:r[t]);const s=e.resolveRef(u);if(Gf(s))return s;if(s&&Array.isArray(s.oneOf)){const c=e.resolveOneOf(r[t],s,`${i}/${t}`);for(const d in s)d!=="oneOf"&&c[d]===void 0&&(c[d]=s[d]);return c}if(s)return s}const{patternProperties:l}=n;if(gi(l)==="object"){let s;const c=Object.keys(l);for(let d=0,h=c.length;d<h;d+=1)if(s=new RegExp(c[d]),s.test(t))return l[c[d]]}const{additionalProperties:o}=n;return gi(o)==="object"?n.additionalProperties:r&&(o===void 0||o===!0)?g0(r[t]):e.errors.unknownPropertyError({property:t,value:r,pointer:`${i}`,schema:n})}};function rme(e,t,n,r,i="#"){var a;const u=gi(r);let l=(a=n.type)!==null&&a!==void 0?a:u;if(Array.isArray(l)){if(!l.includes(u))return e.errors.typeError({value:r,pointer:i,expected:n.type,received:u,schema:n});l=u}const o=PMe[l];if(o){const s=o(e,`${t}`,n,r,i);return s===void 0?e.errors.schemaWarning({pointer:i,value:r,schema:n,key:t}):s}return new Error(`Unsupported schema type ${n.type} for key ${t}`)}function ime(e,t,n=e.rootSchema){var r;if(n.oneOf)return n.oneOf.map(a=>e.resolveRef(a));if(!((r=n.items)===null||r===void 0)&&r.oneOf)return n.items.oneOf.map(a=>e.resolveRef(a));const i=e.step(t,n,{},"#");return Gf(i)?i:[i]}const MMe={array:(e,t,n,r)=>e.typeKeywords.array.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),object:(e,t,n,r)=>e.typeKeywords.object.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),string:(e,t,n,r)=>e.typeKeywords.string.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),integer:(e,t,n,r)=>e.typeKeywords.number.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),number:(e,t,n,r)=>e.typeKeywords.number.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),boolean:(e,t,n,r)=>e.typeKeywords.boolean.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),null:(e,t,n,r)=>e.typeKeywords.null.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r))};function FMe(e,t){const n=gi(e);return n==="number"&&(t==="integer"||Array.isArray(t)&&t.includes("integer"))?Number.isInteger(e)||isNaN(e)?"integer":"number":n}function ame(e,t,n=e.rootSchema,r="#"){if(n=e.resolveRef(n),gi(n)==="boolean")return n?[]:[e.errors.invalidDataError({pointer:r,schema:n,value:t})];if(Gf(n))return[n];if(n.const!==void 0)return tme(n.const,t)?[]:[e.errors.constError({pointer:r,schema:n,value:t,expected:n.const})];const i=FMe(t,n.type),a=n.type||i;return i!==a&&(!Array.isArray(a)||!a.includes(i))?[e.errors.typeError({pointer:r,schema:n,value:t,received:i,expected:a})]:e.validateType[i]==null?[e.errors.invalidTypeError({pointer:r,schema:n,value:t,receivedType:i})]:Ix(e.validateType[i](e,n,t,r)).filter(D$)}const NMe={typeKeywords:{array:["allOf","anyOf","enum","items","maxItems","minItems","not","oneOf","uniqueItems"],boolean:["enum","not","allOf","anyOf","oneOf"],object:["additionalProperties","dependencies","enum","format","minProperties","maxProperties","patternProperties","properties","required","not","oneOf","allOf","anyOf"],string:["allOf","anyOf","enum","format","maxLength","minLength","not","oneOf","pattern"],number:["allOf","anyOf","enum","format","maximum","minimum","multipleOf","not","oneOf"],null:["allOf","anyOf","enum","format","not","oneOf"]},validateKeyword:nme,validateType:MMe,validateFormat:X0e,errors:H0e,addRemoteSchema:V0e,compileSchema:oMe,createSchemaOf:g0,each:Z0e,eachSchema:Dw,getChildSchemaSelection:ime,getSchema:G0e,getTemplate:J0e,isValid:eme,resolveAllOf:N0e,resolveAnyOf:z0e,resolveOneOf:_0e,resolveRef:L0e,step:rme,validate:ame,templateDefaultOptions:a4.templateDefaultOptions};class Out extends q0e{constructor(t,n={}){super(ZI(NMe,n),t)}}const pX="__compiled",mX="__ref",IMe="getRef",LMe="getRoot",BMe=/(#|\/)+$/g;function QMe(e,t,n=t,r=!1){if(t===!0||t===!1||t===void 0||t[pX]!==void 0)return t;const i={ids:{},remotes:e.remotes},a=JSON.stringify(t),u=JSON.parse(a);if(Object.defineProperty(u,pX,{enumerable:!1,value:!0}),Object.defineProperty(u,IMe,{enumerable:!1,value:Qm.bind(null,i,u)}),r===!1&&a.includes("$ref")===!1)return u;u!==n&&Object.defineProperty(u,"$defs",{enumerable:!0,value:Object.assign({},n.definitions,n.$defs,u.definitions,u.$defs)});const l={},o=()=>u;return Dw(u,(s,c)=>{var d;if(s.$id){if(s.$id.startsWith("http")&&/(allOf|anyOf|oneOf)\/\d+$/.test(c)){const S=c.replace(/\/(allOf|anyOf|oneOf)\/\d+$/,""),_=zb.get(u,S);s.$id=(d=_.$id)!==null&&d!==void 0?d:s.$id}i.ids[s.$id.replace(BMe,"")]=c}c=`#${c}`.replace(/##+/,"#");const h=c.replace(/\/[^/]+$/,""),m=c.replace(/\/[^/]+\/[^/]+$/,""),y=l[h]||l[m],v=A$(y,s.$id);l[c]=v,i.ids[v]==null&&(i.ids[v]=c),s.$ref&&!s[mX]&&(Object.defineProperty(s,mX,{enumerable:!1,value:A$(v,s.$ref)}),Object.defineProperty(s,LMe,{enumerable:!1,value:o}))}),u}const zMe={...nme,contains:(e,t,n,r)=>{if(t.contains===!1)return e.errors.containsArrayError({pointer:r,value:n,schema:t});if(t.contains===!0)return Array.isArray(n)&&n.length===0?e.errors.containsAnyError({pointer:r,value:n,schema:t}):void 0;if(gi(t.contains)==="object"){for(let i=0;i<n.length;i+=1)if(e.isValid(n[i],t.contains))return;return e.errors.containsError({pointer:r,schema:t,value:n})}},exclusiveMaximum:(e,t,n,r)=>{if(!isNaN(t.exclusiveMaximum)&&t.exclusiveMaximum<=n)return e.errors.maximumError({maximum:t.exclusiveMaximum,length:n,pointer:r,schema:t,value:n})},exclusiveMinimum:(e,t,n,r)=>{if(!isNaN(t.exclusiveMinimum)&&t.exclusiveMinimum>=n)return e.errors.minimumError({minimum:t.exclusiveMinimum,length:n,pointer:r,schema:t,value:n})},if:v9e,maximum:(e,t,n,r)=>{if(!isNaN(t.maximum)&&t.maximum&&t.maximum<n)return e.errors.maximumError({maximum:t.maximum,length:n,pointer:r,schema:t,value:n})},minimum:(e,t,n,r)=>{if(!isNaN(t.minimum)&&t.minimum>n)return e.errors.minimumError({minimum:t.minimum,length:n,pointer:r,schema:t,value:n})},patternProperties:(e,t,n,r)=>{const i=t.properties||{},a=t.patternProperties;if(gi(a)!=="object")return;const u=[],l=Object.keys(n),o=Object.keys(a).map(s=>({regex:new RegExp(s),patternSchema:a[s]}));return l.forEach(s=>{let c=!1;for(let d=0,h=o.length;d<h;d+=1)if(o[d].regex.test(s)){if(c=!0,o[d].patternSchema===!1){u.push(e.errors.patternPropertiesError({key:s,pointer:r,patterns:Object.keys(a).join(","),schema:t,value:n}));return}const m=e.validate(n[s],o[d].patternSchema,`${r}/${s}`);m&&m.length>0&&u.push(...m)}i[s]||c===!1&&t.additionalProperties===!1&&u.push(e.errors.patternPropertiesError({key:s,pointer:r,patterns:Object.keys(a).join(","),schema:t,value:n}))}),u},propertyNames:(e,t,n,r)=>{if(t.propertyNames===!1)return Object.keys(n).length===0?void 0:e.errors.invalidPropertyNameError({property:Object.keys(n),pointer:r,value:n,schema:t});if(t.propertyNames===!0||gi(t.propertyNames)!=="object")return;const i=[],a=Object.keys(n),u={...t.propertyNames,type:"string"};return a.forEach(l=>{const o=e.validate(l,u,`${r}/${l}`);o.length>0&&i.push(e.errors.invalidPropertyNameError({property:l,pointer:r,validationError:o[0],value:n[l],schema:t}))}),i}},qMe={array:(e,t,n,r)=>e.typeKeywords.array.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),object:(e,t,n,r)=>e.typeKeywords.object.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),string:(e,t,n,r)=>e.typeKeywords.string.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),integer:(e,t,n,r)=>e.typeKeywords.number.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),number:(e,t,n,r)=>e.typeKeywords.number.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),boolean:(e,t,n,r)=>e.typeKeywords.boolean.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r)),null:(e,t,n,r)=>e.typeKeywords.null.filter(i=>t&&t[i]!=null).map(i=>e.validateKeyword[i](e,t,n,r))},ume={typeKeywords:{array:["allOf","anyOf","contains","enum","if","items","maxItems","minItems","not","oneOf","uniqueItems"],boolean:["allOf","anyOf","enum","not","oneOf"],object:["additionalProperties","allOf","anyOf","dependencies","enum","format","if","maxProperties","minProperties","not","oneOf","patternProperties","properties","propertyNames","required"],string:["allOf","anyOf","enum","format","if","maxLength","minLength","not","oneOf","pattern"],number:["allOf","anyOf","enum","exclusiveMaximum","exclusiveMinimum","format","if","maximum","minimum","multipleOf","not","oneOf"],null:["allOf","anyOf","enum","format","not","oneOf"]},validateKeyword:zMe,validateType:qMe,validateFormat:X0e,errors:H0e,addRemoteSchema:V0e,compileSchema:QMe,createSchemaOf:g0,each:Z0e,eachSchema:Dw,getChildSchemaSelection:ime,getSchema:G0e,getTemplate:J0e,isValid:eme,resolveAllOf:N0e,resolveAnyOf:z0e,resolveOneOf:_0e,resolveRef:L0e,step:rme,validate:ame,templateDefaultOptions:a4.templateDefaultOptions};class xut extends q0e{constructor(t,n={}){super(ZI(ume,n),t)}}({...ume});const gX={};function VMe(e){let t=gX[e];if(t)return t;t=gX[e]=[];for(let n=0;n<128;n++){const r=String.fromCharCode(n);t.push(r)}for(let n=0;n<e.length;n++){const r=e.charCodeAt(n);t[r]="%"+("0"+r.toString(16).toUpperCase()).slice(-2)}return t}function F3(e,t){typeof t!="string"&&(t=F3.defaultChars);const n=VMe(t);return e.replace(/(%[a-f0-9]{2})+/gi,function(r){let i="";for(let a=0,u=r.length;a<u;a+=3){const l=parseInt(r.slice(a+1,a+3),16);if(l<128){i+=n[l];continue}if((l&224)===192&&a+3<u){const o=parseInt(r.slice(a+4,a+6),16);if((o&192)===128){const s=l<<6&1984|o&63;s<128?i+="��":i+=String.fromCharCode(s),a+=3;continue}}if((l&240)===224&&a+6<u){const o=parseInt(r.slice(a+4,a+6),16),s=parseInt(r.slice(a+7,a+9),16);if((o&192)===128&&(s&192)===128){const c=l<<12&61440|o<<6&4032|s&63;c<2048||c>=55296&&c<=57343?i+="���":i+=String.fromCharCode(c),a+=6;continue}}if((l&248)===240&&a+9<u){const o=parseInt(r.slice(a+4,a+6),16),s=parseInt(r.slice(a+7,a+9),16),c=parseInt(r.slice(a+10,a+12),16);if((o&192)===128&&(s&192)===128&&(c&192)===128){let d=l<<18&1835008|o<<12&258048|s<<6&4032|c&63;d<65536||d>1114111?i+="����":(d-=65536,i+=String.fromCharCode(55296+(d>>10),56320+(d&1023))),a+=9;continue}}i+="�"}return i})}F3.defaultChars=";/?:@&=+$,#";F3.componentChars="";const vX={};function UMe(e){let t=vX[e];if(t)return t;t=vX[e]=[];for(let n=0;n<128;n++){const r=String.fromCharCode(n);/^[0-9a-z]$/i.test(r)?t.push(r):t.push("%"+("0"+n.toString(16).toUpperCase()).slice(-2))}for(let n=0;n<e.length;n++)t[e.charCodeAt(n)]=e[n];return t}function TS(e,t,n){typeof t!="string"&&(n=t,t=TS.defaultChars),typeof n>"u"&&(n=!0);const r=UMe(t);let i="";for(let a=0,u=e.length;a<u;a++){const l=e.charCodeAt(a);if(n&&l===37&&a+2<u&&/^[0-9a-f]{2}$/i.test(e.slice(a+1,a+3))){i+=e.slice(a,a+3),a+=2;continue}if(l<128){i+=r[l];continue}if(l>=55296&&l<=57343){if(l>=55296&&l<=56319&&a+1<u){const o=e.charCodeAt(a+1);if(o>=56320&&o<=57343){i+=encodeURIComponent(e[a]+e[a+1]),a++;continue}}i+="%EF%BF%BD";continue}i+=encodeURIComponent(e[a])}return i}TS.defaultChars=";/?:@&=+$,-_.!~*'()#";TS.componentChars="-_.!~*'()";function KI(e){let t="";return t+=e.protocol||"",t+=e.slashes?"//":"",t+=e.auth?e.auth+"@":"",e.hostname&&e.hostname.indexOf(":")!==-1?t+="["+e.hostname+"]":t+=e.hostname||"",t+=e.port?":"+e.port:"",t+=e.pathname||"",t+=e.search||"",t+=e.hash||"",t}function T$(){this.protocol=null,this.slashes=null,this.auth=null,this.port=null,this.hostname=null,this.hash=null,this.search=null,this.pathname=null}const jMe=/^([a-z0-9.+-]+:)/i,ZMe=/:[0-9]*$/,HMe=/^(\/\/?(?!\/)[^\?\s]*)(\?[^\s]*)?$/,XMe=["<",">",'"',"`"," ","\r",`
`,"	"],YMe=["{","}","|","\\","^","`"].concat(XMe),GMe=["'"].concat(YMe),yX=["%","/","?",";","#"].concat(GMe),bX=["/","?","#"],KMe=255,OX=/^[+a-z0-9A-Z_-]{0,63}$/,WMe=/^([+a-z0-9A-Z_-]{0,63})(.*)$/,xX={javascript:!0,"javascript:":!0},SX={http:!0,https:!0,ftp:!0,gopher:!0,file:!0,"http:":!0,"https:":!0,"ftp:":!0,"gopher:":!0,"file:":!0};function WI(e,t){if(e&&e instanceof T$)return e;const n=new T$;return n.parse(e,t),n}T$.prototype.parse=function(e,t){let n,r,i,a=e;if(a=a.trim(),!t&&e.split("#").length===1){const s=HMe.exec(a);if(s)return this.pathname=s[1],s[2]&&(this.search=s[2]),this}let u=jMe.exec(a);if(u&&(u=u[0],n=u.toLowerCase(),this.protocol=u,a=a.substr(u.length)),(t||u||a.match(/^\/\/[^@\/]+@[^@\/]+/))&&(i=a.substr(0,2)==="//",i&&!(u&&xX[u])&&(a=a.substr(2),this.slashes=!0)),!xX[u]&&(i||u&&!SX[u])){let s=-1;for(let y=0;y<bX.length;y++)r=a.indexOf(bX[y]),r!==-1&&(s===-1||r<s)&&(s=r);let c,d;s===-1?d=a.lastIndexOf("@"):d=a.lastIndexOf("@",s),d!==-1&&(c=a.slice(0,d),a=a.slice(d+1),this.auth=c),s=-1;for(let y=0;y<yX.length;y++)r=a.indexOf(yX[y]),r!==-1&&(s===-1||r<s)&&(s=r);s===-1&&(s=a.length),a[s-1]===":"&&s--;const h=a.slice(0,s);a=a.slice(s),this.parseHost(h),this.hostname=this.hostname||"";const m=this.hostname[0]==="["&&this.hostname[this.hostname.length-1]==="]";if(!m){const y=this.hostname.split(/\./);for(let v=0,S=y.length;v<S;v++){const _=y[v];if(_&&!_.match(OX)){let E="";for(let $=0,k=_.length;$<k;$++)_.charCodeAt($)>127?E+="x":E+=_[$];if(!E.match(OX)){const $=y.slice(0,v),k=y.slice(v+1),A=_.match(WMe);A&&($.push(A[1]),k.unshift(A[2])),k.length&&(a=k.join(".")+a),this.hostname=$.join(".");break}}}}this.hostname.length>KMe&&(this.hostname=""),m&&(this.hostname=this.hostname.substr(1,this.hostname.length-2))}const l=a.indexOf("#");l!==-1&&(this.hash=a.substr(l),a=a.slice(0,l));const o=a.indexOf("?");return o!==-1&&(this.search=a.substr(o),a=a.slice(0,o)),a&&(this.pathname=a),SX[n]&&this.hostname&&!this.pathname&&(this.pathname=""),this};T$.prototype.parseHost=function(e){let t=ZMe.exec(e);t&&(t=t[0],t!==":"&&(this.port=t.substr(1)),e=e.substr(0,e.length-t.length)),e&&(this.hostname=e)};const JMe=Object.freeze(Object.defineProperty({__proto__:null,decode:F3,encode:TS,format:KI,parse:WI},Symbol.toStringTag,{value:"Module"})),lme=/[\0-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF]/,ome=/[\0-\x1F\x7F-\x9F]/,eFe=/[\xAD\u0600-\u0605\u061C\u06DD\u070F\u0890\u0891\u08E2\u180E\u200B-\u200F\u202A-\u202E\u2060-\u2064\u2066-\u206F\uFEFF\uFFF9-\uFFFB]|\uD804[\uDCBD\uDCCD]|\uD80D[\uDC30-\uDC3F]|\uD82F[\uDCA0-\uDCA3]|\uD834[\uDD73-\uDD7A]|\uDB40[\uDC01\uDC20-\uDC7F]/,JI=/[!-#%-\*,-\/:;\?@\[-\]_\{\}\xA1\xA7\xAB\xB6\xB7\xBB\xBF\u037E\u0387\u055A-\u055F\u0589\u058A\u05BE\u05C0\u05C3\u05C6\u05F3\u05F4\u0609\u060A\u060C\u060D\u061B\u061D-\u061F\u066A-\u066D\u06D4\u0700-\u070D\u07F7-\u07F9\u0830-\u083E\u085E\u0964\u0965\u0970\u09FD\u0A76\u0AF0\u0C77\u0C84\u0DF4\u0E4F\u0E5A\u0E5B\u0F04-\u0F12\u0F14\u0F3A-\u0F3D\u0F85\u0FD0-\u0FD4\u0FD9\u0FDA\u104A-\u104F\u10FB\u1360-\u1368\u1400\u166E\u169B\u169C\u16EB-\u16ED\u1735\u1736\u17D4-\u17D6\u17D8-\u17DA\u1800-\u180A\u1944\u1945\u1A1E\u1A1F\u1AA0-\u1AA6\u1AA8-\u1AAD\u1B5A-\u1B60\u1B7D\u1B7E\u1BFC-\u1BFF\u1C3B-\u1C3F\u1C7E\u1C7F\u1CC0-\u1CC7\u1CD3\u2010-\u2027\u2030-\u2043\u2045-\u2051\u2053-\u205E\u207D\u207E\u208D\u208E\u2308-\u230B\u2329\u232A\u2768-\u2775\u27C5\u27C6\u27E6-\u27EF\u2983-\u2998\u29D8-\u29DB\u29FC\u29FD\u2CF9-\u2CFC\u2CFE\u2CFF\u2D70\u2E00-\u2E2E\u2E30-\u2E4F\u2E52-\u2E5D\u3001-\u3003\u3008-\u3011\u3014-\u301F\u3030\u303D\u30A0\u30FB\uA4FE\uA4FF\uA60D-\uA60F\uA673\uA67E\uA6F2-\uA6F7\uA874-\uA877\uA8CE\uA8CF\uA8F8-\uA8FA\uA8FC\uA92E\uA92F\uA95F\uA9C1-\uA9CD\uA9DE\uA9DF\uAA5C-\uAA5F\uAADE\uAADF\uAAF0\uAAF1\uABEB\uFD3E\uFD3F\uFE10-\uFE19\uFE30-\uFE52\uFE54-\uFE61\uFE63\uFE68\uFE6A\uFE6B\uFF01-\uFF03\uFF05-\uFF0A\uFF0C-\uFF0F\uFF1A\uFF1B\uFF1F\uFF20\uFF3B-\uFF3D\uFF3F\uFF5B\uFF5D\uFF5F-\uFF65]|\uD800[\uDD00-\uDD02\uDF9F\uDFD0]|\uD801\uDD6F|\uD802[\uDC57\uDD1F\uDD3F\uDE50-\uDE58\uDE7F\uDEF0-\uDEF6\uDF39-\uDF3F\uDF99-\uDF9C]|\uD803[\uDEAD\uDF55-\uDF59\uDF86-\uDF89]|\uD804[\uDC47-\uDC4D\uDCBB\uDCBC\uDCBE-\uDCC1\uDD40-\uDD43\uDD74\uDD75\uDDC5-\uDDC8\uDDCD\uDDDB\uDDDD-\uDDDF\uDE38-\uDE3D\uDEA9]|\uD805[\uDC4B-\uDC4F\uDC5A\uDC5B\uDC5D\uDCC6\uDDC1-\uDDD7\uDE41-\uDE43\uDE60-\uDE6C\uDEB9\uDF3C-\uDF3E]|\uD806[\uDC3B\uDD44-\uDD46\uDDE2\uDE3F-\uDE46\uDE9A-\uDE9C\uDE9E-\uDEA2\uDF00-\uDF09]|\uD807[\uDC41-\uDC45\uDC70\uDC71\uDEF7\uDEF8\uDF43-\uDF4F\uDFFF]|\uD809[\uDC70-\uDC74]|\uD80B[\uDFF1\uDFF2]|\uD81A[\uDE6E\uDE6F\uDEF5\uDF37-\uDF3B\uDF44]|\uD81B[\uDE97-\uDE9A\uDFE2]|\uD82F\uDC9F|\uD836[\uDE87-\uDE8B]|\uD83A[\uDD5E\uDD5F]/,sme=/[\$\+<->\^`\|~\xA2-\xA6\xA8\xA9\xAC\xAE-\xB1\xB4\xB8\xD7\xF7\u02C2-\u02C5\u02D2-\u02DF\u02E5-\u02EB\u02ED\u02EF-\u02FF\u0375\u0384\u0385\u03F6\u0482\u058D-\u058F\u0606-\u0608\u060B\u060E\u060F\u06DE\u06E9\u06FD\u06FE\u07F6\u07FE\u07FF\u0888\u09F2\u09F3\u09FA\u09FB\u0AF1\u0B70\u0BF3-\u0BFA\u0C7F\u0D4F\u0D79\u0E3F\u0F01-\u0F03\u0F13\u0F15-\u0F17\u0F1A-\u0F1F\u0F34\u0F36\u0F38\u0FBE-\u0FC5\u0FC7-\u0FCC\u0FCE\u0FCF\u0FD5-\u0FD8\u109E\u109F\u1390-\u1399\u166D\u17DB\u1940\u19DE-\u19FF\u1B61-\u1B6A\u1B74-\u1B7C\u1FBD\u1FBF-\u1FC1\u1FCD-\u1FCF\u1FDD-\u1FDF\u1FED-\u1FEF\u1FFD\u1FFE\u2044\u2052\u207A-\u207C\u208A-\u208C\u20A0-\u20C0\u2100\u2101\u2103-\u2106\u2108\u2109\u2114\u2116-\u2118\u211E-\u2123\u2125\u2127\u2129\u212E\u213A\u213B\u2140-\u2144\u214A-\u214D\u214F\u218A\u218B\u2190-\u2307\u230C-\u2328\u232B-\u2426\u2440-\u244A\u249C-\u24E9\u2500-\u2767\u2794-\u27C4\u27C7-\u27E5\u27F0-\u2982\u2999-\u29D7\u29DC-\u29FB\u29FE-\u2B73\u2B76-\u2B95\u2B97-\u2BFF\u2CE5-\u2CEA\u2E50\u2E51\u2E80-\u2E99\u2E9B-\u2EF3\u2F00-\u2FD5\u2FF0-\u2FFF\u3004\u3012\u3013\u3020\u3036\u3037\u303E\u303F\u309B\u309C\u3190\u3191\u3196-\u319F\u31C0-\u31E3\u31EF\u3200-\u321E\u322A-\u3247\u3250\u3260-\u327F\u328A-\u32B0\u32C0-\u33FF\u4DC0-\u4DFF\uA490-\uA4C6\uA700-\uA716\uA720\uA721\uA789\uA78A\uA828-\uA82B\uA836-\uA839\uAA77-\uAA79\uAB5B\uAB6A\uAB6B\uFB29\uFBB2-\uFBC2\uFD40-\uFD4F\uFDCF\uFDFC-\uFDFF\uFE62\uFE64-\uFE66\uFE69\uFF04\uFF0B\uFF1C-\uFF1E\uFF3E\uFF40\uFF5C\uFF5E\uFFE0-\uFFE6\uFFE8-\uFFEE\uFFFC\uFFFD]|\uD800[\uDD37-\uDD3F\uDD79-\uDD89\uDD8C-\uDD8E\uDD90-\uDD9C\uDDA0\uDDD0-\uDDFC]|\uD802[\uDC77\uDC78\uDEC8]|\uD805\uDF3F|\uD807[\uDFD5-\uDFF1]|\uD81A[\uDF3C-\uDF3F\uDF45]|\uD82F\uDC9C|\uD833[\uDF50-\uDFC3]|\uD834[\uDC00-\uDCF5\uDD00-\uDD26\uDD29-\uDD64\uDD6A-\uDD6C\uDD83\uDD84\uDD8C-\uDDA9\uDDAE-\uDDEA\uDE00-\uDE41\uDE45\uDF00-\uDF56]|\uD835[\uDEC1\uDEDB\uDEFB\uDF15\uDF35\uDF4F\uDF6F\uDF89\uDFA9\uDFC3]|\uD836[\uDC00-\uDDFF\uDE37-\uDE3A\uDE6D-\uDE74\uDE76-\uDE83\uDE85\uDE86]|\uD838[\uDD4F\uDEFF]|\uD83B[\uDCAC\uDCB0\uDD2E\uDEF0\uDEF1]|\uD83C[\uDC00-\uDC2B\uDC30-\uDC93\uDCA0-\uDCAE\uDCB1-\uDCBF\uDCC1-\uDCCF\uDCD1-\uDCF5\uDD0D-\uDDAD\uDDE6-\uDE02\uDE10-\uDE3B\uDE40-\uDE48\uDE50\uDE51\uDE60-\uDE65\uDF00-\uDFFF]|\uD83D[\uDC00-\uDED7\uDEDC-\uDEEC\uDEF0-\uDEFC\uDF00-\uDF76\uDF7B-\uDFD9\uDFE0-\uDFEB\uDFF0]|\uD83E[\uDC00-\uDC0B\uDC10-\uDC47\uDC50-\uDC59\uDC60-\uDC87\uDC90-\uDCAD\uDCB0\uDCB1\uDD00-\uDE53\uDE60-\uDE6D\uDE70-\uDE7C\uDE80-\uDE88\uDE90-\uDEBD\uDEBF-\uDEC5\uDECE-\uDEDB\uDEE0-\uDEE8\uDEF0-\uDEF8\uDF00-\uDF92\uDF94-\uDFCA]/,cme=/[ \xA0\u1680\u2000-\u200A\u2028\u2029\u202F\u205F\u3000]/,tFe=Object.freeze(Object.defineProperty({__proto__:null,Any:lme,Cc:ome,Cf:eFe,P:JI,S:sme,Z:cme},Symbol.toStringTag,{value:"Module"})),nFe=new Uint16Array('ᵁ<Õıʊҝջאٵ۞ޢߖࠏ੊ઑඡ๭༉༦჊ረዡᐕᒝᓃᓟᔥ\0\0\0\0\0\0ᕫᛍᦍᰒᷝ὾⁠↰⊍⏀⏻⑂⠤⤒ⴈ⹈⿎〖㊺㘹㞬㣾㨨㩱㫠㬮ࠀEMabcfglmnoprstu\\bfms¦³¹ÈÏlig耻Æ䃆P耻&䀦cute耻Á䃁reve;䄂Āiyx}rc耻Â䃂;䐐r;쀀𝔄rave耻À䃀pha;䎑acr;䄀d;橓Āgp¡on;䄄f;쀀𝔸plyFunction;恡ing耻Å䃅Ācs¾Ãr;쀀𝒜ign;扔ilde耻Ã䃃ml耻Ä䃄ЀaceforsuåûþėĜĢħĪĀcrêòkslash;或Ŷöø;櫧ed;挆y;䐑ƀcrtąċĔause;戵noullis;愬a;䎒r;쀀𝔅pf;쀀𝔹eve;䋘còēmpeq;扎܀HOacdefhilorsuōőŖƀƞƢƵƷƺǜȕɳɸɾcy;䐧PY耻©䂩ƀcpyŝŢźute;䄆Ā;iŧŨ拒talDifferentialD;慅leys;愭ȀaeioƉƎƔƘron;䄌dil耻Ç䃇rc;䄈nint;戰ot;䄊ĀdnƧƭilla;䂸terDot;䂷òſi;䎧rcleȀDMPTǇǋǑǖot;抙inus;抖lus;投imes;抗oĀcsǢǸkwiseContourIntegral;戲eCurlyĀDQȃȏoubleQuote;思uote;怙ȀlnpuȞȨɇɕonĀ;eȥȦ户;橴ƀgitȯȶȺruent;扡nt;戯ourIntegral;戮ĀfrɌɎ;愂oduct;成nterClockwiseContourIntegral;戳oss;樯cr;쀀𝒞pĀ;Cʄʅ拓ap;才րDJSZacefiosʠʬʰʴʸˋ˗ˡ˦̳ҍĀ;oŹʥtrahd;椑cy;䐂cy;䐅cy;䐏ƀgrsʿ˄ˇger;怡r;憡hv;櫤Āayː˕ron;䄎;䐔lĀ;t˝˞戇a;䎔r;쀀𝔇Āaf˫̧Ācm˰̢riticalȀADGT̖̜̀̆cute;䂴oŴ̋̍;䋙bleAcute;䋝rave;䁠ilde;䋜ond;拄ferentialD;慆Ѱ̽\0\0\0͔͂\0Ѕf;쀀𝔻ƀ;DE͈͉͍䂨ot;惜qual;扐blèCDLRUVͣͲ΂ϏϢϸontourIntegraìȹoɴ͹\0\0ͻ»͉nArrow;懓Āeo·ΤftƀARTΐΖΡrrow;懐ightArrow;懔eåˊngĀLRΫτeftĀARγιrrow;柸ightArrow;柺ightArrow;柹ightĀATϘϞrrow;懒ee;抨pɁϩ\0\0ϯrrow;懑ownArrow;懕erticalBar;戥ǹABLRTaВЪаўѿͼrrowƀ;BUНОТ憓ar;椓pArrow;懵reve;䌑eft˒к\0ц\0ѐightVector;楐eeVector;楞ectorĀ;Bљњ憽ar;楖ightǔѧ\0ѱeeVector;楟ectorĀ;BѺѻ懁ar;楗eeĀ;A҆҇护rrow;憧ĀctҒҗr;쀀𝒟rok;䄐ࠀNTacdfglmopqstuxҽӀӄӋӞӢӧӮӵԡԯԶՒ՝ՠեG;䅊H耻Ð䃐cute耻É䃉ƀaiyӒӗӜron;䄚rc耻Ê䃊;䐭ot;䄖r;쀀𝔈rave耻È䃈ement;戈ĀapӺӾcr;䄒tyɓԆ\0\0ԒmallSquare;旻erySmallSquare;斫ĀgpԦԪon;䄘f;쀀𝔼silon;䎕uĀaiԼՉlĀ;TՂՃ橵ilde;扂librium;懌Āci՗՚r;愰m;橳a;䎗ml耻Ë䃋Āipժկsts;戃onentialE;慇ʀcfiosօֈ֍ֲ׌y;䐤r;쀀𝔉lledɓ֗\0\0֣mallSquare;旼erySmallSquare;斪Ͱֺ\0ֿ\0\0ׄf;쀀𝔽All;戀riertrf;愱cò׋؀JTabcdfgorstר׬ׯ׺؀ؒؖ؛؝أ٬ٲcy;䐃耻>䀾mmaĀ;d׷׸䎓;䏜reve;䄞ƀeiy؇،ؐdil;䄢rc;䄜;䐓ot;䄠r;쀀𝔊;拙pf;쀀𝔾eater̀EFGLSTصلَٖٛ٦qualĀ;Lؾؿ扥ess;招ullEqual;执reater;檢ess;扷lantEqual;橾ilde;扳cr;쀀𝒢;扫ЀAacfiosuڅڋږڛڞڪھۊRDcy;䐪Āctڐڔek;䋇;䁞irc;䄤r;愌lbertSpace;愋ǰگ\0ڲf;愍izontalLine;攀Āctۃۅòکrok;䄦mpńېۘownHumðįqual;扏܀EJOacdfgmnostuۺ۾܃܇܎ܚܞܡܨ݄ݸދޏޕcy;䐕lig;䄲cy;䐁cute耻Í䃍Āiyܓܘrc耻Î䃎;䐘ot;䄰r;愑rave耻Ì䃌ƀ;apܠܯܿĀcgܴܷr;䄪inaryI;慈lieóϝǴ݉\0ݢĀ;eݍݎ戬Āgrݓݘral;戫section;拂isibleĀCTݬݲomma;恣imes;恢ƀgptݿރވon;䄮f;쀀𝕀a;䎙cr;愐ilde;䄨ǫޚ\0ޞcy;䐆l耻Ï䃏ʀcfosuެ޷޼߂ߐĀiyޱ޵rc;䄴;䐙r;쀀𝔍pf;쀀𝕁ǣ߇\0ߌr;쀀𝒥rcy;䐈kcy;䐄΀HJacfosߤߨ߽߬߱ࠂࠈcy;䐥cy;䐌ppa;䎚Āey߶߻dil;䄶;䐚r;쀀𝔎pf;쀀𝕂cr;쀀𝒦րJTaceflmostࠥࠩࠬࡐࡣ঳সে্਷ੇcy;䐉耻<䀼ʀcmnpr࠷࠼ࡁࡄࡍute;䄹bda;䎛g;柪lacetrf;愒r;憞ƀaeyࡗ࡜ࡡron;䄽dil;䄻;䐛Āfsࡨ॰tԀACDFRTUVarࡾࢩࢱࣦ࣠ࣼयज़ΐ४Ānrࢃ࢏gleBracket;柨rowƀ;BR࢙࢚࢞憐ar;懤ightArrow;懆eiling;挈oǵࢷ\0ࣃbleBracket;柦nǔࣈ\0࣒eeVector;楡ectorĀ;Bࣛࣜ懃ar;楙loor;挊ightĀAV࣯ࣵrrow;憔ector;楎Āerँगeƀ;AVउऊऐ抣rrow;憤ector;楚iangleƀ;BEतथऩ抲ar;槏qual;抴pƀDTVषूौownVector;楑eeVector;楠ectorĀ;Bॖॗ憿ar;楘ectorĀ;B॥०憼ar;楒ightáΜs̀EFGLSTॾঋকঝঢভqualGreater;拚ullEqual;扦reater;扶ess;檡lantEqual;橽ilde;扲r;쀀𝔏Ā;eঽা拘ftarrow;懚idot;䄿ƀnpw৔ਖਛgȀLRlr৞৷ਂਐeftĀAR০৬rrow;柵ightArrow;柷ightArrow;柶eftĀarγਊightáοightáϊf;쀀𝕃erĀLRਢਬeftArrow;憙ightArrow;憘ƀchtਾੀੂòࡌ;憰rok;䅁;扪Ѐacefiosuਗ਼੝੠੷੼અઋ઎p;椅y;䐜Ādl੥੯iumSpace;恟lintrf;愳r;쀀𝔐nusPlus;戓pf;쀀𝕄cò੶;䎜ҀJacefostuણધભીଔଙඑ඗ඞcy;䐊cute;䅃ƀaey઴હાron;䅇dil;䅅;䐝ƀgswે૰଎ativeƀMTV૓૟૨ediumSpace;怋hiĀcn૦૘ë૙eryThiî૙tedĀGL૸ଆreaterGreateòٳessLesóੈLine;䀊r;쀀𝔑ȀBnptଢନଷ଺reak;恠BreakingSpace;䂠f;愕ڀ;CDEGHLNPRSTV୕ୖ୪୼஡௫ఄ౞಄ದ೘ൡඅ櫬Āou୛୤ngruent;扢pCap;扭oubleVerticalBar;戦ƀlqxஃஊ஛ement;戉ualĀ;Tஒஓ扠ilde;쀀≂̸ists;戄reater΀;EFGLSTஶஷ஽௉௓௘௥扯qual;扱ullEqual;쀀≧̸reater;쀀≫̸ess;批lantEqual;쀀⩾̸ilde;扵umpń௲௽ownHump;쀀≎̸qual;쀀≏̸eĀfsఊధtTriangleƀ;BEచఛడ拪ar;쀀⧏̸qual;括s̀;EGLSTవశ఼ౄోౘ扮qual;扰reater;扸ess;쀀≪̸lantEqual;쀀⩽̸ilde;扴estedĀGL౨౹reaterGreater;쀀⪢̸essLess;쀀⪡̸recedesƀ;ESಒಓಛ技qual;쀀⪯̸lantEqual;拠ĀeiಫಹverseElement;戌ghtTriangleƀ;BEೋೌ೒拫ar;쀀⧐̸qual;拭ĀquೝഌuareSuĀbp೨೹setĀ;E೰ೳ쀀⊏̸qual;拢ersetĀ;Eഃആ쀀⊐̸qual;拣ƀbcpഓതൎsetĀ;Eഛഞ쀀⊂⃒qual;抈ceedsȀ;ESTലള഻െ抁qual;쀀⪰̸lantEqual;拡ilde;쀀≿̸ersetĀ;E൘൛쀀⊃⃒qual;抉ildeȀ;EFT൮൯൵ൿ扁qual;扄ullEqual;扇ilde;扉erticalBar;戤cr;쀀𝒩ilde耻Ñ䃑;䎝܀Eacdfgmoprstuvලෂ෉෕ෛ෠෧෼ขภยา฿ไlig;䅒cute耻Ó䃓Āiy෎ීrc耻Ô䃔;䐞blac;䅐r;쀀𝔒rave耻Ò䃒ƀaei෮ෲ෶cr;䅌ga;䎩cron;䎟pf;쀀𝕆enCurlyĀDQฎบoubleQuote;怜uote;怘;橔Āclวฬr;쀀𝒪ash耻Ø䃘iŬื฼de耻Õ䃕es;樷ml耻Ö䃖erĀBP๋๠Āar๐๓r;怾acĀek๚๜;揞et;掴arenthesis;揜Ҁacfhilors๿ງຊຏຒດຝະ໼rtialD;戂y;䐟r;쀀𝔓i;䎦;䎠usMinus;䂱Āipຢອncareplanåڝf;愙Ȁ;eio຺ູ໠໤檻cedesȀ;EST່້໏໚扺qual;檯lantEqual;扼ilde;找me;怳Ādp໩໮uct;戏ortionĀ;aȥ໹l;戝Āci༁༆r;쀀𝒫;䎨ȀUfos༑༖༛༟OT耻"䀢r;쀀𝔔pf;愚cr;쀀𝒬؀BEacefhiorsu༾གྷཇའཱིྦྷྪྭ႖ႩႴႾarr;椐G耻®䂮ƀcnrཎནབute;䅔g;柫rĀ;tཛྷཝ憠l;椖ƀaeyཧཬཱron;䅘dil;䅖;䐠Ā;vླྀཹ愜erseĀEUྂྙĀlq྇ྎement;戋uilibrium;懋pEquilibrium;楯r»ཹo;䎡ghtЀACDFTUVa࿁࿫࿳ဢဨၛႇϘĀnr࿆࿒gleBracket;柩rowƀ;BL࿜࿝࿡憒ar;懥eftArrow;懄eiling;按oǵ࿹\0စbleBracket;柧nǔည\0နeeVector;楝ectorĀ;Bဝသ懂ar;楕loor;挋Āerိ၃eƀ;AVဵံြ抢rrow;憦ector;楛iangleƀ;BEၐၑၕ抳ar;槐qual;抵pƀDTVၣၮၸownVector;楏eeVector;楜ectorĀ;Bႂႃ憾ar;楔ectorĀ;B႑႒懀ar;楓Āpuႛ႞f;愝ndImplies;楰ightarrow;懛ĀchႹႼr;愛;憱leDelayed;槴ڀHOacfhimoqstuფჱჷჽᄙᄞᅑᅖᅡᅧᆵᆻᆿĀCcჩხHcy;䐩y;䐨FTcy;䐬cute;䅚ʀ;aeiyᄈᄉᄎᄓᄗ檼ron;䅠dil;䅞rc;䅜;䐡r;쀀𝔖ortȀDLRUᄪᄴᄾᅉownArrow»ОeftArrow»࢚ightArrow»࿝pArrow;憑gma;䎣allCircle;战pf;쀀𝕊ɲᅭ\0\0ᅰt;戚areȀ;ISUᅻᅼᆉᆯ斡ntersection;抓uĀbpᆏᆞsetĀ;Eᆗᆘ抏qual;抑ersetĀ;Eᆨᆩ抐qual;抒nion;抔cr;쀀𝒮ar;拆ȀbcmpᇈᇛሉላĀ;sᇍᇎ拐etĀ;Eᇍᇕqual;抆ĀchᇠህeedsȀ;ESTᇭᇮᇴᇿ扻qual;檰lantEqual;扽ilde;承Tháྌ;我ƀ;esሒሓሣ拑rsetĀ;Eሜም抃qual;抇et»ሓրHRSacfhiorsሾቄ቉ቕ቞ቱቶኟዂወዑORN耻Þ䃞ADE;愢ĀHc቎ቒcy;䐋y;䐦Ābuቚቜ;䀉;䎤ƀaeyብቪቯron;䅤dil;䅢;䐢r;쀀𝔗Āeiቻ኉ǲኀ\0ኇefore;戴a;䎘Ācn኎ኘkSpace;쀀  Space;怉ldeȀ;EFTካኬኲኼ戼qual;扃ullEqual;扅ilde;扈pf;쀀𝕋ipleDot;惛Āctዖዛr;쀀𝒯rok;䅦ૡዷጎጚጦ\0ጬጱ\0\0\0\0\0ጸጽ፷ᎅ\0᏿ᐄᐊᐐĀcrዻጁute耻Ú䃚rĀ;oጇገ憟cir;楉rǣጓ\0጖y;䐎ve;䅬Āiyጞጣrc耻Û䃛;䐣blac;䅰r;쀀𝔘rave耻Ù䃙acr;䅪Ādiፁ፩erĀBPፈ፝Āarፍፐr;䁟acĀekፗፙ;揟et;掵arenthesis;揝onĀ;P፰፱拃lus;抎Āgp፻፿on;䅲f;쀀𝕌ЀADETadps᎕ᎮᎸᏄϨᏒᏗᏳrrowƀ;BDᅐᎠᎤar;椒ownArrow;懅ownArrow;憕quilibrium;楮eeĀ;AᏋᏌ报rrow;憥ownáϳerĀLRᏞᏨeftArrow;憖ightArrow;憗iĀ;lᏹᏺ䏒on;䎥ing;䅮cr;쀀𝒰ilde;䅨ml耻Ü䃜ҀDbcdefosvᐧᐬᐰᐳᐾᒅᒊᒐᒖash;披ar;櫫y;䐒ashĀ;lᐻᐼ抩;櫦Āerᑃᑅ;拁ƀbtyᑌᑐᑺar;怖Ā;iᑏᑕcalȀBLSTᑡᑥᑪᑴar;戣ine;䁼eparator;杘ilde;所ThinSpace;怊r;쀀𝔙pf;쀀𝕍cr;쀀𝒱dash;抪ʀcefosᒧᒬᒱᒶᒼirc;䅴dge;拀r;쀀𝔚pf;쀀𝕎cr;쀀𝒲Ȁfiosᓋᓐᓒᓘr;쀀𝔛;䎞pf;쀀𝕏cr;쀀𝒳ҀAIUacfosuᓱᓵᓹᓽᔄᔏᔔᔚᔠcy;䐯cy;䐇cy;䐮cute耻Ý䃝Āiyᔉᔍrc;䅶;䐫r;쀀𝔜pf;쀀𝕐cr;쀀𝒴ml;䅸ЀHacdefosᔵᔹᔿᕋᕏᕝᕠᕤcy;䐖cute;䅹Āayᕄᕉron;䅽;䐗ot;䅻ǲᕔ\0ᕛoWidtè૙a;䎖r;愨pf;愤cr;쀀𝒵௡ᖃᖊᖐ\0ᖰᖶᖿ\0\0\0\0ᗆᗛᗫᙟ᙭\0ᚕ᚛ᚲᚹ\0ᚾcute耻á䃡reve;䄃̀;Ediuyᖜᖝᖡᖣᖨᖭ戾;쀀∾̳;房rc耻â䃢te肻´̆;䐰lig耻æ䃦Ā;r²ᖺ;쀀𝔞rave耻à䃠ĀepᗊᗖĀfpᗏᗔsym;愵èᗓha;䎱ĀapᗟcĀclᗤᗧr;䄁g;樿ɤᗰ\0\0ᘊʀ;adsvᗺᗻᗿᘁᘇ戧nd;橕;橜lope;橘;橚΀;elmrszᘘᘙᘛᘞᘿᙏᙙ戠;榤e»ᘙsdĀ;aᘥᘦ戡ѡᘰᘲᘴᘶᘸᘺᘼᘾ;榨;榩;榪;榫;榬;榭;榮;榯tĀ;vᙅᙆ戟bĀ;dᙌᙍ抾;榝Āptᙔᙗh;戢»¹arr;捼Āgpᙣᙧon;䄅f;쀀𝕒΀;Eaeiop዁ᙻᙽᚂᚄᚇᚊ;橰cir;橯;扊d;手s;䀧roxĀ;e዁ᚒñᚃing耻å䃥ƀctyᚡᚦᚨr;쀀𝒶;䀪mpĀ;e዁ᚯñʈilde耻ã䃣ml耻ä䃤Āciᛂᛈoninôɲnt;樑ࠀNabcdefiklnoprsu᛭ᛱᜰ᜼ᝃᝈ᝸᝽០៦ᠹᡐᜍ᤽᥈ᥰot;櫭Ācrᛶ᜞kȀcepsᜀᜅᜍᜓong;扌psilon;䏶rime;怵imĀ;e᜚᜛戽q;拍Ŷᜢᜦee;抽edĀ;gᜬᜭ挅e»ᜭrkĀ;t፜᜷brk;掶Āoyᜁᝁ;䐱quo;怞ʀcmprtᝓ᝛ᝡᝤᝨausĀ;eĊĉptyv;榰séᜌnoõēƀahwᝯ᝱ᝳ;䎲;愶een;扬r;쀀𝔟g΀costuvwឍឝឳេ៕៛៞ƀaiuបពរðݠrc;旯p»፱ƀdptឤឨឭot;樀lus;樁imes;樂ɱឹ\0\0ើcup;樆ar;昅riangleĀdu៍្own;施p;斳plus;樄eåᑄåᒭarow;植ƀako៭ᠦᠵĀcn៲ᠣkƀlst៺֫᠂ozenge;槫riangleȀ;dlr᠒᠓᠘᠝斴own;斾eft;旂ight;斸k;搣Ʊᠫ\0ᠳƲᠯ\0ᠱ;斒;斑4;斓ck;斈ĀeoᠾᡍĀ;qᡃᡆ쀀=⃥uiv;쀀≡⃥t;挐Ȁptwxᡙᡞᡧᡬf;쀀𝕓Ā;tᏋᡣom»Ꮜtie;拈؀DHUVbdhmptuvᢅᢖᢪᢻᣗᣛᣬ᣿ᤅᤊᤐᤡȀLRlrᢎᢐᢒᢔ;敗;敔;敖;敓ʀ;DUduᢡᢢᢤᢦᢨ敐;敦;敩;敤;敧ȀLRlrᢳᢵᢷᢹ;敝;敚;敜;教΀;HLRhlrᣊᣋᣍᣏᣑᣓᣕ救;敬;散;敠;敫;敢;敟ox;槉ȀLRlrᣤᣦᣨᣪ;敕;敒;攐;攌ʀ;DUduڽ᣷᣹᣻᣽;敥;敨;攬;攴inus;抟lus;択imes;抠ȀLRlrᤙᤛᤝ᤟;敛;敘;攘;攔΀;HLRhlrᤰᤱᤳᤵᤷ᤻᤹攂;敪;敡;敞;攼;攤;攜Āevģ᥂bar耻¦䂦Ȁceioᥑᥖᥚᥠr;쀀𝒷mi;恏mĀ;e᜚᜜lƀ;bhᥨᥩᥫ䁜;槅sub;柈Ŭᥴ᥾lĀ;e᥹᥺怢t»᥺pƀ;Eeįᦅᦇ;檮Ā;qۜۛೡᦧ\0᧨ᨑᨕᨲ\0ᨷᩐ\0\0᪴\0\0᫁\0\0ᬡᬮ᭍᭒\0᯽\0ᰌƀcpr᦭ᦲ᧝ute;䄇̀;abcdsᦿᧀᧄ᧊᧕᧙戩nd;橄rcup;橉Āau᧏᧒p;橋p;橇ot;橀;쀀∩︀Āeo᧢᧥t;恁îړȀaeiu᧰᧻ᨁᨅǰ᧵\0᧸s;橍on;䄍dil耻ç䃧rc;䄉psĀ;sᨌᨍ橌m;橐ot;䄋ƀdmnᨛᨠᨦil肻¸ƭptyv;榲t脀¢;eᨭᨮ䂢räƲr;쀀𝔠ƀceiᨽᩀᩍy;䑇ckĀ;mᩇᩈ朓ark»ᩈ;䏇r΀;Ecefms᩟᩠ᩢᩫ᪤᪪᪮旋;槃ƀ;elᩩᩪᩭ䋆q;扗eɡᩴ\0\0᪈rrowĀlr᩼᪁eft;憺ight;憻ʀRSacd᪒᪔᪖᪚᪟»ཇ;擈st;抛irc;抚ash;抝nint;樐id;櫯cir;槂ubsĀ;u᪻᪼晣it»᪼ˬ᫇᫔᫺\0ᬊonĀ;eᫍᫎ䀺Ā;qÇÆɭ᫙\0\0᫢aĀ;t᫞᫟䀬;䁀ƀ;fl᫨᫩᫫戁îᅠeĀmx᫱᫶ent»᫩eóɍǧ᫾\0ᬇĀ;dኻᬂot;橭nôɆƀfryᬐᬔᬗ;쀀𝕔oäɔ脀©;sŕᬝr;愗Āaoᬥᬩrr;憵ss;朗Ācuᬲᬷr;쀀𝒸Ābpᬼ᭄Ā;eᭁᭂ櫏;櫑Ā;eᭉᭊ櫐;櫒dot;拯΀delprvw᭠᭬᭷ᮂᮬᯔ᯹arrĀlr᭨᭪;椸;椵ɰ᭲\0\0᭵r;拞c;拟arrĀ;p᭿ᮀ憶;椽̀;bcdosᮏᮐᮖᮡᮥᮨ截rcap;橈Āauᮛᮞp;橆p;橊ot;抍r;橅;쀀∪︀Ȁalrv᮵ᮿᯞᯣrrĀ;mᮼᮽ憷;椼yƀevwᯇᯔᯘqɰᯎ\0\0ᯒreã᭳uã᭵ee;拎edge;拏en耻¤䂤earrowĀlrᯮ᯳eft»ᮀight»ᮽeäᯝĀciᰁᰇoninôǷnt;戱lcty;挭ঀAHabcdefhijlorstuwz᰸᰻᰿ᱝᱩᱵᲊᲞᲬᲷ᳻᳿ᴍᵻᶑᶫᶻ᷆᷍rò΁ar;楥Ȁglrs᱈ᱍ᱒᱔ger;怠eth;愸òᄳhĀ;vᱚᱛ怐»ऊūᱡᱧarow;椏aã̕Āayᱮᱳron;䄏;䐴ƀ;ao̲ᱼᲄĀgrʿᲁr;懊tseq;橷ƀglmᲑᲔᲘ耻°䂰ta;䎴ptyv;榱ĀirᲣᲨsht;楿;쀀𝔡arĀlrᲳᲵ»ࣜ»သʀaegsv᳂͸᳖᳜᳠mƀ;oș᳊᳔ndĀ;ș᳑uit;晦amma;䏝in;拲ƀ;io᳧᳨᳸䃷de脀÷;o᳧ᳰntimes;拇nø᳷cy;䑒cɯᴆ\0\0ᴊrn;挞op;挍ʀlptuwᴘᴝᴢᵉᵕlar;䀤f;쀀𝕕ʀ;emps̋ᴭᴷᴽᵂqĀ;d͒ᴳot;扑inus;戸lus;戔quare;抡blebarwedgåúnƀadhᄮᵝᵧownarrowóᲃarpoonĀlrᵲᵶefôᲴighôᲶŢᵿᶅkaro÷གɯᶊ\0\0ᶎrn;挟op;挌ƀcotᶘᶣᶦĀryᶝᶡ;쀀𝒹;䑕l;槶rok;䄑Ādrᶰᶴot;拱iĀ;fᶺ᠖斿Āah᷀᷃ròЩaòྦangle;榦Āci᷒ᷕy;䑟grarr;柿ऀDacdefglmnopqrstuxḁḉḙḸոḼṉṡṾấắẽỡἪἷὄ὎὚ĀDoḆᴴoôᲉĀcsḎḔute耻é䃩ter;橮ȀaioyḢḧḱḶron;䄛rĀ;cḭḮ扖耻ê䃪lon;払;䑍ot;䄗ĀDrṁṅot;扒;쀀𝔢ƀ;rsṐṑṗ檚ave耻è䃨Ā;dṜṝ檖ot;檘Ȁ;ilsṪṫṲṴ檙nters;揧;愓Ā;dṹṺ檕ot;檗ƀapsẅẉẗcr;䄓tyƀ;svẒẓẕ戅et»ẓpĀ1;ẝẤĳạả;怄;怅怃ĀgsẪẬ;䅋p;怂ĀgpẴẸon;䄙f;쀀𝕖ƀalsỄỎỒrĀ;sỊị拕l;槣us;橱iƀ;lvỚớở䎵on»ớ;䏵ȀcsuvỪỳἋἣĀioữḱrc»Ḯɩỹ\0\0ỻíՈantĀglἂἆtr»ṝess»Ṻƀaeiἒ἖Ἒls;䀽st;扟vĀ;DȵἠD;橸parsl;槥ĀDaἯἳot;打rr;楱ƀcdiἾὁỸr;愯oô͒ĀahὉὋ;䎷耻ð䃰Āmrὓὗl耻ë䃫o;悬ƀcipὡὤὧl;䀡sôծĀeoὬὴctatioîՙnentialåչৡᾒ\0ᾞ\0ᾡᾧ\0\0ῆῌ\0ΐ\0ῦῪ \0 ⁚llingdotseñṄy;䑄male;晀ƀilrᾭᾳ῁lig;耀ﬃɩᾹ\0\0᾽g;耀ﬀig;耀ﬄ;쀀𝔣lig;耀ﬁlig;쀀fjƀaltῙ῜ῡt;晭ig;耀ﬂns;斱of;䆒ǰ΅\0ῳf;쀀𝕗ĀakֿῷĀ;vῼ´拔;櫙artint;樍Āao‌⁕Ācs‑⁒α‚‰‸⁅⁈\0⁐β•‥‧‪‬\0‮耻½䂽;慓耻¼䂼;慕;慙;慛Ƴ‴\0‶;慔;慖ʴ‾⁁\0\0⁃耻¾䂾;慗;慜5;慘ƶ⁌\0⁎;慚;慝8;慞l;恄wn;挢cr;쀀𝒻ࢀEabcdefgijlnorstv₂₉₟₥₰₴⃰⃵⃺⃿℃ℒℸ̗ℾ⅒↞Ā;lٍ₇;檌ƀcmpₐₕ₝ute;䇵maĀ;dₜ᳚䎳;檆reve;䄟Āiy₪₮rc;䄝;䐳ot;䄡Ȁ;lqsؾق₽⃉ƀ;qsؾٌ⃄lanô٥Ȁ;cdl٥⃒⃥⃕c;檩otĀ;o⃜⃝檀Ā;l⃢⃣檂;檄Ā;e⃪⃭쀀⋛︀s;檔r;쀀𝔤Ā;gٳ؛mel;愷cy;䑓Ȁ;Eajٚℌℎℐ;檒;檥;檤ȀEaesℛℝ℩ℴ;扩pĀ;p℣ℤ檊rox»ℤĀ;q℮ℯ檈Ā;q℮ℛim;拧pf;쀀𝕘Āci⅃ⅆr;愊mƀ;el٫ⅎ⅐;檎;檐茀>;cdlqr׮ⅠⅪⅮⅳⅹĀciⅥⅧ;檧r;橺ot;拗Par;榕uest;橼ʀadelsↄⅪ←ٖ↛ǰ↉\0↎proø₞r;楸qĀlqؿ↖lesó₈ií٫Āen↣↭rtneqq;쀀≩︀Å↪ԀAabcefkosy⇄⇇⇱⇵⇺∘∝∯≨≽ròΠȀilmr⇐⇔⇗⇛rsðᒄf»․ilôکĀdr⇠⇤cy;䑊ƀ;cwࣴ⇫⇯ir;楈;憭ar;意irc;䄥ƀalr∁∎∓rtsĀ;u∉∊晥it»∊lip;怦con;抹r;쀀𝔥sĀew∣∩arow;椥arow;椦ʀamopr∺∾≃≞≣rr;懿tht;戻kĀlr≉≓eftarrow;憩ightarrow;憪f;쀀𝕙bar;怕ƀclt≯≴≸r;쀀𝒽asè⇴rok;䄧Ābp⊂⊇ull;恃hen»ᱛૡ⊣\0⊪\0⊸⋅⋎\0⋕⋳\0\0⋸⌢⍧⍢⍿\0⎆⎪⎴cute耻í䃭ƀ;iyݱ⊰⊵rc耻î䃮;䐸Ācx⊼⊿y;䐵cl耻¡䂡ĀfrΟ⋉;쀀𝔦rave耻ì䃬Ȁ;inoܾ⋝⋩⋮Āin⋢⋦nt;樌t;戭fin;槜ta;愩lig;䄳ƀaop⋾⌚⌝ƀcgt⌅⌈⌗r;䄫ƀelpܟ⌏⌓inåގarôܠh;䄱f;抷ed;䆵ʀ;cfotӴ⌬⌱⌽⍁are;愅inĀ;t⌸⌹戞ie;槝doô⌙ʀ;celpݗ⍌⍐⍛⍡al;抺Āgr⍕⍙eróᕣã⍍arhk;樗rod;樼Ȁcgpt⍯⍲⍶⍻y;䑑on;䄯f;쀀𝕚a;䎹uest耻¿䂿Āci⎊⎏r;쀀𝒾nʀ;EdsvӴ⎛⎝⎡ӳ;拹ot;拵Ā;v⎦⎧拴;拳Ā;iݷ⎮lde;䄩ǫ⎸\0⎼cy;䑖l耻ï䃯̀cfmosu⏌⏗⏜⏡⏧⏵Āiy⏑⏕rc;䄵;䐹r;쀀𝔧ath;䈷pf;쀀𝕛ǣ⏬\0⏱r;쀀𝒿rcy;䑘kcy;䑔Ѐacfghjos␋␖␢␧␭␱␵␻ppaĀ;v␓␔䎺;䏰Āey␛␠dil;䄷;䐺r;쀀𝔨reen;䄸cy;䑅cy;䑜pf;쀀𝕜cr;쀀𝓀஀ABEHabcdefghjlmnoprstuv⑰⒁⒆⒍⒑┎┽╚▀♎♞♥♹♽⚚⚲⛘❝❨➋⟀⠁⠒ƀart⑷⑺⑼rò৆òΕail;椛arr;椎Ā;gঔ⒋;檋ar;楢ॣ⒥\0⒪\0⒱\0\0\0\0\0⒵Ⓔ\0ⓆⓈⓍ\0⓹ute;䄺mptyv;榴raîࡌbda;䎻gƀ;dlࢎⓁⓃ;榑åࢎ;檅uo耻«䂫rЀ;bfhlpst࢙ⓞⓦⓩ⓫⓮⓱⓵Ā;f࢝ⓣs;椟s;椝ë≒p;憫l;椹im;楳l;憢ƀ;ae⓿─┄檫il;椙Ā;s┉┊檭;쀀⪭︀ƀabr┕┙┝rr;椌rk;杲Āak┢┬cĀek┨┪;䁻;䁛Āes┱┳;榋lĀdu┹┻;榏;榍Ȁaeuy╆╋╖╘ron;䄾Ādi═╔il;䄼ìࢰâ┩;䐻Ȁcqrs╣╦╭╽a;椶uoĀ;rนᝆĀdu╲╷har;楧shar;楋h;憲ʀ;fgqs▋▌উ◳◿扤tʀahlrt▘▤▷◂◨rrowĀ;t࢙□aé⓶arpoonĀdu▯▴own»њp»०eftarrows;懇ightƀahs◍◖◞rrowĀ;sࣴࢧarpoonó྘quigarro÷⇰hreetimes;拋ƀ;qs▋ও◺lanôবʀ;cdgsব☊☍☝☨c;檨otĀ;o☔☕橿Ā;r☚☛檁;檃Ā;e☢☥쀀⋚︀s;檓ʀadegs☳☹☽♉♋pproøⓆot;拖qĀgq♃♅ôউgtò⒌ôছiíলƀilr♕࣡♚sht;楼;쀀𝔩Ā;Eজ♣;檑š♩♶rĀdu▲♮Ā;l॥♳;楪lk;斄cy;䑙ʀ;achtੈ⚈⚋⚑⚖rò◁orneòᴈard;楫ri;旺Āio⚟⚤dot;䅀ustĀ;a⚬⚭掰che»⚭ȀEaes⚻⚽⛉⛔;扨pĀ;p⛃⛄檉rox»⛄Ā;q⛎⛏檇Ā;q⛎⚻im;拦Ѐabnoptwz⛩⛴⛷✚✯❁❇❐Ānr⛮⛱g;柬r;懽rëࣁgƀlmr⛿✍✔eftĀar০✇ightá৲apsto;柼ightá৽parrowĀlr✥✩efô⓭ight;憬ƀafl✶✹✽r;榅;쀀𝕝us;樭imes;樴š❋❏st;戗áፎƀ;ef❗❘᠀旊nge»❘arĀ;l❤❥䀨t;榓ʀachmt❳❶❼➅➇ròࢨorneòᶌarĀ;d྘➃;業;怎ri;抿̀achiqt➘➝ੀ➢➮➻quo;怹r;쀀𝓁mƀ;egল➪➬;檍;檏Ābu┪➳oĀ;rฟ➹;怚rok;䅂萀<;cdhilqrࠫ⟒☹⟜⟠⟥⟪⟰Āci⟗⟙;檦r;橹reå◲mes;拉arr;楶uest;橻ĀPi⟵⟹ar;榖ƀ;ef⠀भ᠛旃rĀdu⠇⠍shar;楊har;楦Āen⠗⠡rtneqq;쀀≨︀Å⠞܀Dacdefhilnopsu⡀⡅⢂⢎⢓⢠⢥⢨⣚⣢⣤ઃ⣳⤂Dot;戺Ȁclpr⡎⡒⡣⡽r耻¯䂯Āet⡗⡙;時Ā;e⡞⡟朠se»⡟Ā;sျ⡨toȀ;dluျ⡳⡷⡻owîҌefôएðᏑker;斮Āoy⢇⢌mma;権;䐼ash;怔asuredangle»ᘦr;쀀𝔪o;愧ƀcdn⢯⢴⣉ro耻µ䂵Ȁ;acdᑤ⢽⣀⣄sôᚧir;櫰ot肻·Ƶusƀ;bd⣒ᤃ⣓戒Ā;uᴼ⣘;横ţ⣞⣡p;櫛ò−ðઁĀdp⣩⣮els;抧f;쀀𝕞Āct⣸⣽r;쀀𝓂pos»ᖝƀ;lm⤉⤊⤍䎼timap;抸ఀGLRVabcdefghijlmoprstuvw⥂⥓⥾⦉⦘⧚⧩⨕⨚⩘⩝⪃⪕⪤⪨⬄⬇⭄⭿⮮ⰴⱧⱼ⳩Āgt⥇⥋;쀀⋙̸Ā;v⥐௏쀀≫⃒ƀelt⥚⥲⥶ftĀar⥡⥧rrow;懍ightarrow;懎;쀀⋘̸Ā;v⥻ే쀀≪⃒ightarrow;懏ĀDd⦎⦓ash;抯ash;抮ʀbcnpt⦣⦧⦬⦱⧌la»˞ute;䅄g;쀀∠⃒ʀ;Eiop඄⦼⧀⧅⧈;쀀⩰̸d;쀀≋̸s;䅉roø඄urĀ;a⧓⧔普lĀ;s⧓ସǳ⧟\0⧣p肻 ଷmpĀ;e௹ఀʀaeouy⧴⧾⨃⨐⨓ǰ⧹\0⧻;橃on;䅈dil;䅆ngĀ;dൾ⨊ot;쀀⩭̸p;橂;䐽ash;怓΀;Aadqsxஒ⨩⨭⨻⩁⩅⩐rr;懗rĀhr⨳⨶k;椤Ā;oᏲᏰot;쀀≐̸uiöୣĀei⩊⩎ar;椨í஘istĀ;s஠டr;쀀𝔫ȀEest௅⩦⩹⩼ƀ;qs஼⩭௡ƀ;qs஼௅⩴lanô௢ií௪Ā;rஶ⪁»ஷƀAap⪊⪍⪑rò⥱rr;憮ar;櫲ƀ;svྍ⪜ྌĀ;d⪡⪢拼;拺cy;䑚΀AEadest⪷⪺⪾⫂⫅⫶⫹rò⥦;쀀≦̸rr;憚r;急Ȁ;fqs఻⫎⫣⫯tĀar⫔⫙rro÷⫁ightarro÷⪐ƀ;qs఻⪺⫪lanôౕĀ;sౕ⫴»శiíౝĀ;rవ⫾iĀ;eచథiäඐĀpt⬌⬑f;쀀𝕟膀¬;in⬙⬚⬶䂬nȀ;Edvஉ⬤⬨⬮;쀀⋹̸ot;쀀⋵̸ǡஉ⬳⬵;拷;拶iĀ;vಸ⬼ǡಸ⭁⭃;拾;拽ƀaor⭋⭣⭩rȀ;ast୻⭕⭚⭟lleì୻l;쀀⫽⃥;쀀∂̸lint;樔ƀ;ceಒ⭰⭳uåಥĀ;cಘ⭸Ā;eಒ⭽ñಘȀAait⮈⮋⮝⮧rò⦈rrƀ;cw⮔⮕⮙憛;쀀⤳̸;쀀↝̸ghtarrow»⮕riĀ;eೋೖ΀chimpqu⮽⯍⯙⬄୸⯤⯯Ȁ;cerല⯆ഷ⯉uå൅;쀀𝓃ortɭ⬅\0\0⯖ará⭖mĀ;e൮⯟Ā;q൴൳suĀbp⯫⯭å೸åഋƀbcp⯶ⰑⰙȀ;Ees⯿ⰀഢⰄ抄;쀀⫅̸etĀ;eഛⰋqĀ;qണⰀcĀ;eലⰗñസȀ;EesⰢⰣൟⰧ抅;쀀⫆̸etĀ;e൘ⰮqĀ;qൠⰣȀgilrⰽⰿⱅⱇìௗlde耻ñ䃱çృiangleĀlrⱒⱜeftĀ;eచⱚñదightĀ;eೋⱥñ೗Ā;mⱬⱭ䎽ƀ;esⱴⱵⱹ䀣ro;愖p;怇ҀDHadgilrsⲏⲔⲙⲞⲣⲰⲶⳓⳣash;抭arr;椄p;쀀≍⃒ash;抬ĀetⲨⲬ;쀀≥⃒;쀀>⃒nfin;槞ƀAetⲽⳁⳅrr;椂;쀀≤⃒Ā;rⳊⳍ쀀<⃒ie;쀀⊴⃒ĀAtⳘⳜrr;椃rie;쀀⊵⃒im;쀀∼⃒ƀAan⳰⳴ⴂrr;懖rĀhr⳺⳽k;椣Ā;oᏧᏥear;椧ቓ᪕\0\0\0\0\0\0\0\0\0\0\0\0\0ⴭ\0ⴸⵈⵠⵥ⵲ⶄᬇ\0\0ⶍⶫ\0ⷈⷎ\0ⷜ⸙⸫⸾⹃Ācsⴱ᪗ute耻ó䃳ĀiyⴼⵅrĀ;c᪞ⵂ耻ô䃴;䐾ʀabios᪠ⵒⵗǈⵚlac;䅑v;樸old;榼lig;䅓Ācr⵩⵭ir;榿;쀀𝔬ͯ⵹\0\0⵼\0ⶂn;䋛ave耻ò䃲;槁Ābmⶈ෴ar;榵Ȁacitⶕ⶘ⶥⶨrò᪀Āir⶝ⶠr;榾oss;榻nå๒;槀ƀaeiⶱⶵⶹcr;䅍ga;䏉ƀcdnⷀⷅǍron;䎿;榶pf;쀀𝕠ƀaelⷔ⷗ǒr;榷rp;榹΀;adiosvⷪⷫⷮ⸈⸍⸐⸖戨rò᪆Ȁ;efmⷷⷸ⸂⸅橝rĀ;oⷾⷿ愴f»ⷿ耻ª䂪耻º䂺gof;抶r;橖lope;橗;橛ƀclo⸟⸡⸧ò⸁ash耻ø䃸l;折iŬⸯ⸴de耻õ䃵esĀ;aǛ⸺s;樶ml耻ö䃶bar;挽ૡ⹞\0⹽\0⺀⺝\0⺢⺹\0\0⻋ຜ\0⼓\0\0⼫⾼\0⿈rȀ;astЃ⹧⹲຅脀¶;l⹭⹮䂶leìЃɩ⹸\0\0⹻m;櫳;櫽y;䐿rʀcimpt⺋⺏⺓ᡥ⺗nt;䀥od;䀮il;怰enk;怱r;쀀𝔭ƀimo⺨⺰⺴Ā;v⺭⺮䏆;䏕maô੶ne;明ƀ;tv⺿⻀⻈䏀chfork»´;䏖Āau⻏⻟nĀck⻕⻝kĀ;h⇴⻛;愎ö⇴sҀ;abcdemst⻳⻴ᤈ⻹⻽⼄⼆⼊⼎䀫cir;樣ir;樢Āouᵀ⼂;樥;橲n肻±ຝim;樦wo;樧ƀipu⼙⼠⼥ntint;樕f;쀀𝕡nd耻£䂣Ԁ;Eaceinosu່⼿⽁⽄⽇⾁⾉⾒⽾⾶;檳p;檷uå໙Ā;c໎⽌̀;acens່⽙⽟⽦⽨⽾pproø⽃urlyeñ໙ñ໎ƀaes⽯⽶⽺pprox;檹qq;檵im;拨iíໟmeĀ;s⾈ຮ怲ƀEas⽸⾐⽺ð⽵ƀdfp໬⾙⾯ƀals⾠⾥⾪lar;挮ine;挒urf;挓Ā;t໻⾴ï໻rel;抰Āci⿀⿅r;쀀𝓅;䏈ncsp;怈̀fiopsu⿚⋢⿟⿥⿫⿱r;쀀𝔮pf;쀀𝕢rime;恗cr;쀀𝓆ƀaeo⿸〉〓tĀei⿾々rnionóڰnt;樖stĀ;e【】䀿ñἙô༔઀ABHabcdefhilmnoprstux぀けさすムㄎㄫㅇㅢㅲㆎ㈆㈕㈤㈩㉘㉮㉲㊐㊰㊷ƀartぇおがròႳòϝail;検aròᱥar;楤΀cdenqrtとふへみわゔヌĀeuねぱ;쀀∽̱te;䅕iãᅮmptyv;榳gȀ;del࿑らるろ;榒;榥å࿑uo耻»䂻rր;abcfhlpstw࿜ガクシスゼゾダッデナp;極Ā;f࿠ゴs;椠;椳s;椞ë≝ð✮l;楅im;楴l;憣;憝Āaiパフil;椚oĀ;nホボ戶aló༞ƀabrョリヮrò៥rk;杳ĀakンヽcĀekヹ・;䁽;䁝Āes㄂㄄;榌lĀduㄊㄌ;榎;榐Ȁaeuyㄗㄜㄧㄩron;䅙Ādiㄡㄥil;䅗ì࿲âヺ;䑀Ȁclqsㄴㄷㄽㅄa;椷dhar;楩uoĀ;rȎȍh;憳ƀacgㅎㅟངlȀ;ipsླྀㅘㅛႜnåႻarôྩt;断ƀilrㅩဣㅮsht;楽;쀀𝔯ĀaoㅷㆆrĀduㅽㅿ»ѻĀ;l႑ㆄ;楬Ā;vㆋㆌ䏁;䏱ƀgns㆕ㇹㇼht̀ahlrstㆤㆰ㇂㇘㇤㇮rrowĀ;t࿜ㆭaéトarpoonĀduㆻㆿowîㅾp»႒eftĀah㇊㇐rrowó࿪arpoonóՑightarrows;應quigarro÷ニhreetimes;拌g;䋚ingdotseñἲƀahm㈍㈐㈓rò࿪aòՑ;怏oustĀ;a㈞㈟掱che»㈟mid;櫮Ȁabpt㈲㈽㉀㉒Ānr㈷㈺g;柭r;懾rëဃƀafl㉇㉊㉎r;榆;쀀𝕣us;樮imes;樵Āap㉝㉧rĀ;g㉣㉤䀩t;榔olint;樒arò㇣Ȁachq㉻㊀Ⴜ㊅quo;怺r;쀀𝓇Ābu・㊊oĀ;rȔȓƀhir㊗㊛㊠reåㇸmes;拊iȀ;efl㊪ၙᠡ㊫方tri;槎luhar;楨;愞ൡ㋕㋛㋟㌬㌸㍱\0㍺㎤\0\0㏬㏰\0㐨㑈㑚㒭㒱㓊㓱\0㘖\0\0㘳cute;䅛quï➺Ԁ;Eaceinpsyᇭ㋳㋵㋿㌂㌋㌏㌟㌦㌩;檴ǰ㋺\0㋼;檸on;䅡uåᇾĀ;dᇳ㌇il;䅟rc;䅝ƀEas㌖㌘㌛;檶p;檺im;择olint;樓iíሄ;䑁otƀ;be㌴ᵇ㌵担;橦΀Aacmstx㍆㍊㍗㍛㍞㍣㍭rr;懘rĀhr㍐㍒ë∨Ā;oਸ਼਴t耻§䂧i;䀻war;椩mĀin㍩ðnuóñt;朶rĀ;o㍶⁕쀀𝔰Ȁacoy㎂㎆㎑㎠rp;景Āhy㎋㎏cy;䑉;䑈rtɭ㎙\0\0㎜iäᑤaraì⹯耻­䂭Āgm㎨㎴maƀ;fv㎱㎲㎲䏃;䏂Ѐ;deglnprካ㏅㏉㏎㏖㏞㏡㏦ot;橪Ā;q኱ኰĀ;E㏓㏔檞;檠Ā;E㏛㏜檝;檟e;扆lus;樤arr;楲aròᄽȀaeit㏸㐈㐏㐗Āls㏽㐄lsetmé㍪hp;樳parsl;槤Ādlᑣ㐔e;挣Ā;e㐜㐝檪Ā;s㐢㐣檬;쀀⪬︀ƀflp㐮㐳㑂tcy;䑌Ā;b㐸㐹䀯Ā;a㐾㐿槄r;挿f;쀀𝕤aĀdr㑍ЂesĀ;u㑔㑕晠it»㑕ƀcsu㑠㑹㒟Āau㑥㑯pĀ;sᆈ㑫;쀀⊓︀pĀ;sᆴ㑵;쀀⊔︀uĀbp㑿㒏ƀ;esᆗᆜ㒆etĀ;eᆗ㒍ñᆝƀ;esᆨᆭ㒖etĀ;eᆨ㒝ñᆮƀ;afᅻ㒦ְrť㒫ֱ»ᅼaròᅈȀcemt㒹㒾㓂㓅r;쀀𝓈tmîñiì㐕aræᆾĀar㓎㓕rĀ;f㓔ឿ昆Āan㓚㓭ightĀep㓣㓪psiloîỠhé⺯s»⡒ʀbcmnp㓻㕞ሉ㖋㖎Ҁ;Edemnprs㔎㔏㔑㔕㔞㔣㔬㔱㔶抂;櫅ot;檽Ā;dᇚ㔚ot;櫃ult;櫁ĀEe㔨㔪;櫋;把lus;檿arr;楹ƀeiu㔽㕒㕕tƀ;en㔎㕅㕋qĀ;qᇚ㔏eqĀ;q㔫㔨m;櫇Ābp㕚㕜;櫕;櫓c̀;acensᇭ㕬㕲㕹㕻㌦pproø㋺urlyeñᇾñᇳƀaes㖂㖈㌛pproø㌚qñ㌗g;晪ڀ123;Edehlmnps㖩㖬㖯ሜ㖲㖴㗀㗉㗕㗚㗟㗨㗭耻¹䂹耻²䂲耻³䂳;櫆Āos㖹㖼t;檾ub;櫘Ā;dሢ㗅ot;櫄sĀou㗏㗒l;柉b;櫗arr;楻ult;櫂ĀEe㗤㗦;櫌;抋lus;櫀ƀeiu㗴㘉㘌tƀ;enሜ㗼㘂qĀ;qሢ㖲eqĀ;q㗧㗤m;櫈Ābp㘑㘓;櫔;櫖ƀAan㘜㘠㘭rr;懙rĀhr㘦㘨ë∮Ā;oਫ਩war;椪lig耻ß䃟௡㙑㙝㙠ዎ㙳㙹\0㙾㛂\0\0\0\0\0㛛㜃\0㜉㝬\0\0\0㞇ɲ㙖\0\0㙛get;挖;䏄rë๟ƀaey㙦㙫㙰ron;䅥dil;䅣;䑂lrec;挕r;쀀𝔱Ȁeiko㚆㚝㚵㚼ǲ㚋\0㚑eĀ4fኄኁaƀ;sv㚘㚙㚛䎸ym;䏑Ācn㚢㚲kĀas㚨㚮pproø዁im»ኬsðኞĀas㚺㚮ð዁rn耻þ䃾Ǭ̟㛆⋧es膀×;bd㛏㛐㛘䃗Ā;aᤏ㛕r;樱;樰ƀeps㛡㛣㜀á⩍Ȁ;bcf҆㛬㛰㛴ot;挶ir;櫱Ā;o㛹㛼쀀𝕥rk;櫚á㍢rime;怴ƀaip㜏㜒㝤dåቈ΀adempst㜡㝍㝀㝑㝗㝜㝟ngleʀ;dlqr㜰㜱㜶㝀㝂斵own»ᶻeftĀ;e⠀㜾ñम;扜ightĀ;e㊪㝋ñၚot;旬inus;樺lus;樹b;槍ime;樻ezium;揢ƀcht㝲㝽㞁Āry㝷㝻;쀀𝓉;䑆cy;䑛rok;䅧Āio㞋㞎xô᝷headĀlr㞗㞠eftarro÷ࡏightarrow»ཝऀAHabcdfghlmoprstuw㟐㟓㟗㟤㟰㟼㠎㠜㠣㠴㡑㡝㡫㢩㣌㣒㣪㣶ròϭar;楣Ācr㟜㟢ute耻ú䃺òᅐrǣ㟪\0㟭y;䑞ve;䅭Āiy㟵㟺rc耻û䃻;䑃ƀabh㠃㠆㠋ròᎭlac;䅱aòᏃĀir㠓㠘sht;楾;쀀𝔲rave耻ù䃹š㠧㠱rĀlr㠬㠮»ॗ»ႃlk;斀Āct㠹㡍ɯ㠿\0\0㡊rnĀ;e㡅㡆挜r»㡆op;挏ri;旸Āal㡖㡚cr;䅫肻¨͉Āgp㡢㡦on;䅳f;쀀𝕦̀adhlsuᅋ㡸㡽፲㢑㢠ownáᎳarpoonĀlr㢈㢌efô㠭ighô㠯iƀ;hl㢙㢚㢜䏅»ᏺon»㢚parrows;懈ƀcit㢰㣄㣈ɯ㢶\0\0㣁rnĀ;e㢼㢽挝r»㢽op;挎ng;䅯ri;旹cr;쀀𝓊ƀdir㣙㣝㣢ot;拰lde;䅩iĀ;f㜰㣨»᠓Āam㣯㣲rò㢨l耻ü䃼angle;榧ހABDacdeflnoprsz㤜㤟㤩㤭㦵㦸㦽㧟㧤㧨㧳㧹㧽㨁㨠ròϷarĀ;v㤦㤧櫨;櫩asèϡĀnr㤲㤷grt;榜΀eknprst㓣㥆㥋㥒㥝㥤㦖appá␕othinçẖƀhir㓫⻈㥙opô⾵Ā;hᎷ㥢ïㆍĀiu㥩㥭gmá㎳Ābp㥲㦄setneqĀ;q㥽㦀쀀⊊︀;쀀⫋︀setneqĀ;q㦏㦒쀀⊋︀;쀀⫌︀Āhr㦛㦟etá㚜iangleĀlr㦪㦯eft»थight»ၑy;䐲ash»ံƀelr㧄㧒㧗ƀ;beⷪ㧋㧏ar;抻q;扚lip;拮Ābt㧜ᑨaòᑩr;쀀𝔳tré㦮suĀbp㧯㧱»ജ»൙pf;쀀𝕧roð໻tré㦴Ācu㨆㨋r;쀀𝓋Ābp㨐㨘nĀEe㦀㨖»㥾nĀEe㦒㨞»㦐igzag;榚΀cefoprs㨶㨻㩖㩛㩔㩡㩪irc;䅵Ādi㩀㩑Ābg㩅㩉ar;機eĀ;qᗺ㩏;扙erp;愘r;쀀𝔴pf;쀀𝕨Ā;eᑹ㩦atèᑹcr;쀀𝓌ૣណ㪇\0㪋\0㪐㪛\0\0㪝㪨㪫㪯\0\0㫃㫎\0㫘ៜ៟tré៑r;쀀𝔵ĀAa㪔㪗ròσrò৶;䎾ĀAa㪡㪤ròθrò৫að✓is;拻ƀdptឤ㪵㪾Āfl㪺ឩ;쀀𝕩imåឲĀAa㫇㫊ròώròਁĀcq㫒ីr;쀀𝓍Āpt៖㫜ré។Ѐacefiosu㫰㫽㬈㬌㬑㬕㬛㬡cĀuy㫶㫻te耻ý䃽;䑏Āiy㬂㬆rc;䅷;䑋n耻¥䂥r;쀀𝔶cy;䑗pf;쀀𝕪cr;쀀𝓎Ācm㬦㬩y;䑎l耻ÿ䃿Ԁacdefhiosw㭂㭈㭔㭘㭤㭩㭭㭴㭺㮀cute;䅺Āay㭍㭒ron;䅾;䐷ot;䅼Āet㭝㭡træᕟa;䎶r;쀀𝔷cy;䐶grarr;懝pf;쀀𝕫cr;쀀𝓏Ājn㮅㮇;怍j;怌'.split("").map(e=>e.charCodeAt(0))),rFe=new Uint16Array("Ȁaglq	\x1Bɭ\0\0p;䀦os;䀧t;䀾t;䀼uot;䀢".split("").map(e=>e.charCodeAt(0)));var vP;const iFe=new Map([[0,65533],[128,8364],[130,8218],[131,402],[132,8222],[133,8230],[134,8224],[135,8225],[136,710],[137,8240],[138,352],[139,8249],[140,338],[142,381],[145,8216],[146,8217],[147,8220],[148,8221],[149,8226],[150,8211],[151,8212],[152,732],[153,8482],[154,353],[155,8250],[156,339],[158,382],[159,376]]),aFe=(vP=String.fromCodePoint)!==null&&vP!==void 0?vP:function(e){let t="";return e>65535&&(e-=65536,t+=String.fromCharCode(e>>>10&1023|55296),e=56320|e&1023),t+=String.fromCharCode(e),t};function uFe(e){var t;return e>=55296&&e<=57343||e>1114111?65533:(t=iFe.get(e))!==null&&t!==void 0?t:e}var oo;(function(e){e[e.NUM=35]="NUM",e[e.SEMI=59]="SEMI",e[e.EQUALS=61]="EQUALS",e[e.ZERO=48]="ZERO",e[e.NINE=57]="NINE",e[e.LOWER_A=97]="LOWER_A",e[e.LOWER_F=102]="LOWER_F",e[e.LOWER_X=120]="LOWER_X",e[e.LOWER_Z=122]="LOWER_Z",e[e.UPPER_A=65]="UPPER_A",e[e.UPPER_F=70]="UPPER_F",e[e.UPPER_Z=90]="UPPER_Z"})(oo||(oo={}));const lFe=32;var zv;(function(e){e[e.VALUE_LENGTH=49152]="VALUE_LENGTH",e[e.BRANCH_LENGTH=16256]="BRANCH_LENGTH",e[e.JUMP_TABLE=127]="JUMP_TABLE"})(zv||(zv={}));function wM(e){return e>=oo.ZERO&&e<=oo.NINE}function oFe(e){return e>=oo.UPPER_A&&e<=oo.UPPER_F||e>=oo.LOWER_A&&e<=oo.LOWER_F}function sFe(e){return e>=oo.UPPER_A&&e<=oo.UPPER_Z||e>=oo.LOWER_A&&e<=oo.LOWER_Z||wM(e)}function cFe(e){return e===oo.EQUALS||sFe(e)}var ro;(function(e){e[e.EntityStart=0]="EntityStart",e[e.NumericStart=1]="NumericStart",e[e.NumericDecimal=2]="NumericDecimal",e[e.NumericHex=3]="NumericHex",e[e.NamedEntity=4]="NamedEntity"})(ro||(ro={}));var Tv;(function(e){e[e.Legacy=0]="Legacy",e[e.Strict=1]="Strict",e[e.Attribute=2]="Attribute"})(Tv||(Tv={}));class fFe{constructor(t,n,r){this.decodeTree=t,this.emitCodePoint=n,this.errors=r,this.state=ro.EntityStart,this.consumed=1,this.result=0,this.treeIndex=0,this.excess=1,this.decodeMode=Tv.Strict}startEntity(t){this.decodeMode=t,this.state=ro.EntityStart,this.result=0,this.treeIndex=0,this.excess=1,this.consumed=1}write(t,n){switch(this.state){case ro.EntityStart:return t.charCodeAt(n)===oo.NUM?(this.state=ro.NumericStart,this.consumed+=1,this.stateNumericStart(t,n+1)):(this.state=ro.NamedEntity,this.stateNamedEntity(t,n));case ro.NumericStart:return this.stateNumericStart(t,n);case ro.NumericDecimal:return this.stateNumericDecimal(t,n);case ro.NumericHex:return this.stateNumericHex(t,n);case ro.NamedEntity:return this.stateNamedEntity(t,n)}}stateNumericStart(t,n){return n>=t.length?-1:(t.charCodeAt(n)|lFe)===oo.LOWER_X?(this.state=ro.NumericHex,this.consumed+=1,this.stateNumericHex(t,n+1)):(this.state=ro.NumericDecimal,this.stateNumericDecimal(t,n))}addToNumericResult(t,n,r,i){if(n!==r){const a=r-n;this.result=this.result*Math.pow(i,a)+parseInt(t.substr(n,a),i),this.consumed+=a}}stateNumericHex(t,n){const r=n;for(;n<t.length;){const i=t.charCodeAt(n);if(wM(i)||oFe(i))n+=1;else return this.addToNumericResult(t,r,n,16),this.emitNumericEntity(i,3)}return this.addToNumericResult(t,r,n,16),-1}stateNumericDecimal(t,n){const r=n;for(;n<t.length;){const i=t.charCodeAt(n);if(wM(i))n+=1;else return this.addToNumericResult(t,r,n,10),this.emitNumericEntity(i,2)}return this.addToNumericResult(t,r,n,10),-1}emitNumericEntity(t,n){var r;if(this.consumed<=n)return(r=this.errors)===null||r===void 0||r.absenceOfDigitsInNumericCharacterReference(this.consumed),0;if(t===oo.SEMI)this.consumed+=1;else if(this.decodeMode===Tv.Strict)return 0;return this.emitCodePoint(uFe(this.result),this.consumed),this.errors&&(t!==oo.SEMI&&this.errors.missingSemicolonAfterCharacterReference(),this.errors.validateNumericCharacterReference(this.result)),this.consumed}stateNamedEntity(t,n){const{decodeTree:r}=this;let i=r[this.treeIndex],a=(i&zv.VALUE_LENGTH)>>14;for(;n<t.length;n++,this.excess++){const u=t.charCodeAt(n);if(this.treeIndex=dFe(r,i,this.treeIndex+Math.max(1,a),u),this.treeIndex<0)return this.result===0||this.decodeMode===Tv.Attribute&&(a===0||cFe(u))?0:this.emitNotTerminatedNamedEntity();if(i=r[this.treeIndex],a=(i&zv.VALUE_LENGTH)>>14,a!==0){if(u===oo.SEMI)return this.emitNamedEntityData(this.treeIndex,a,this.consumed+this.excess);this.decodeMode!==Tv.Strict&&(this.result=this.treeIndex,this.consumed+=this.excess,this.excess=0)}}return-1}emitNotTerminatedNamedEntity(){var t;const{result:n,decodeTree:r}=this,i=(r[n]&zv.VALUE_LENGTH)>>14;return this.emitNamedEntityData(n,i,this.consumed),(t=this.errors)===null||t===void 0||t.missingSemicolonAfterCharacterReference(),this.consumed}emitNamedEntityData(t,n,r){const{decodeTree:i}=this;return this.emitCodePoint(n===1?i[t]&~zv.VALUE_LENGTH:i[t+1],r),n===3&&this.emitCodePoint(i[t+2],r),r}end(){var t;switch(this.state){case ro.NamedEntity:return this.result!==0&&(this.decodeMode!==Tv.Attribute||this.result===this.treeIndex)?this.emitNotTerminatedNamedEntity():0;case ro.NumericDecimal:return this.emitNumericEntity(0,2);case ro.NumericHex:return this.emitNumericEntity(0,3);case ro.NumericStart:return(t=this.errors)===null||t===void 0||t.absenceOfDigitsInNumericCharacterReference(this.consumed),0;case ro.EntityStart:return 0}}}function fme(e){let t="";const n=new fFe(e,r=>t+=aFe(r));return function(i,a){let u=0,l=0;for(;(l=i.indexOf("&",l))>=0;){t+=i.slice(u,l),n.startEntity(a);const s=n.write(i,l+1);if(s<0){u=l+n.end();break}u=l+s,l=s===0?u+1:u}const o=t+i.slice(u);return t="",o}}function dFe(e,t,n,r){const i=(t&zv.BRANCH_LENGTH)>>7,a=t&zv.JUMP_TABLE;if(i===0)return a!==0&&r===a?n:-1;if(a){const o=r-a;return o<0||o>=i?-1:e[n+o]-1}let u=n,l=u+i-1;for(;u<=l;){const o=u+l>>>1,s=e[o];if(s<r)u=o+1;else if(s>r)l=o-1;else return e[o+i]}return-1}const hFe=fme(nFe);fme(rFe);function dme(e,t=Tv.Legacy){return hFe(e,t)}function pFe(e){return Object.prototype.toString.call(e)}function eL(e){return pFe(e)==="[object String]"}const mFe=Object.prototype.hasOwnProperty;function gFe(e,t){return mFe.call(e,t)}function $w(e){return Array.prototype.slice.call(arguments,1).forEach(function(n){if(n){if(typeof n!="object")throw new TypeError(n+"must be object");Object.keys(n).forEach(function(r){e[r]=n[r]})}}),e}function hme(e,t,n){return[].concat(e.slice(0,t),n,e.slice(t+1))}function tL(e){return!(e>=55296&&e<=57343||e>=64976&&e<=65007||(e&65535)===65535||(e&65535)===65534||e>=0&&e<=8||e===11||e>=14&&e<=31||e>=127&&e<=159||e>1114111)}function k$(e){if(e>65535){e-=65536;const t=55296+(e>>10),n=56320+(e&1023);return String.fromCharCode(t,n)}return String.fromCharCode(e)}const pme=/\\([!"#$%&'()*+,\-./:;<=>?@[\\\]^_`{|}~])/g,vFe=/&([a-z#][a-z0-9]{1,31});/gi,yFe=new RegExp(pme.source+"|"+vFe.source,"gi"),bFe=/^#((?:x[a-f0-9]{1,8}|[0-9]{1,8}))$/i;function OFe(e,t){if(t.charCodeAt(0)===35&&bFe.test(t)){const r=t[1].toLowerCase()==="x"?parseInt(t.slice(2),16):parseInt(t.slice(1),10);return tL(r)?k$(r):e}const n=dme(e);return n!==e?n:e}function xFe(e){return e.indexOf("\\")<0?e:e.replace(pme,"$1")}function N3(e){return e.indexOf("\\")<0&&e.indexOf("&")<0?e:e.replace(yFe,function(t,n,r){return n||OFe(t,r)})}const SFe=/[&<>"]/,_Fe=/[&<>"]/g,EFe={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;"};function DFe(e){return EFe[e]}function ay(e){return SFe.test(e)?e.replace(_Fe,DFe):e}const $Fe=/[.?*+^$[\]\\(){}|-]/g;function CFe(e){return e.replace($Fe,"\\$&")}function _a(e){switch(e){case 9:case 32:return!0}return!1}function Qx(e){if(e>=8192&&e<=8202)return!0;switch(e){case 9:case 10:case 11:case 12:case 13:case 32:case 160:case 5760:case 8239:case 8287:case 12288:return!0}return!1}function zx(e){return JI.test(e)||sme.test(e)}function qx(e){switch(e){case 33:case 34:case 35:case 36:case 37:case 38:case 39:case 40:case 41:case 42:case 43:case 44:case 45:case 46:case 47:case 58:case 59:case 60:case 61:case 62:case 63:case 64:case 91:case 92:case 93:case 94:case 95:case 96:case 123:case 124:case 125:case 126:return!0;default:return!1}}function Cw(e){return e=e.trim().replace(/\s+/g," "),"ẞ".toLowerCase()==="Ṿ"&&(e=e.replace(/ẞ/g,"ß")),e.toLowerCase().toUpperCase()}const wFe={mdurl:JMe,ucmicro:tFe},RFe=Object.freeze(Object.defineProperty({__proto__:null,arrayReplaceAt:hme,assign:$w,escapeHtml:ay,escapeRE:CFe,fromCodePoint:k$,has:gFe,isMdAsciiPunct:qx,isPunctChar:zx,isSpace:_a,isString:eL,isValidEntityCode:tL,isWhiteSpace:Qx,lib:wFe,normalizeReference:Cw,unescapeAll:N3,unescapeMd:xFe},Symbol.toStringTag,{value:"Module"}));function AFe(e,t,n){let r,i,a,u;const l=e.posMax,o=e.pos;for(e.pos=t+1,r=1;e.pos<l;){if(a=e.src.charCodeAt(e.pos),a===93&&(r--,r===0)){i=!0;break}if(u=e.pos,e.md.inline.skipToken(e),a===91){if(u===e.pos-1)r++;else if(n)return e.pos=o,-1}}let s=-1;return i&&(s=e.pos),e.pos=o,s}function TFe(e,t,n){let r,i=t;const a={ok:!1,pos:0,str:""};if(e.charCodeAt(i)===60){for(i++;i<n;){if(r=e.charCodeAt(i),r===10||r===60)return a;if(r===62)return a.pos=i+1,a.str=N3(e.slice(t+1,i)),a.ok=!0,a;if(r===92&&i+1<n){i+=2;continue}i++}return a}let u=0;for(;i<n&&(r=e.charCodeAt(i),!(r===32||r<32||r===127));){if(r===92&&i+1<n){if(e.charCodeAt(i+1)===32)break;i+=2;continue}if(r===40&&(u++,u>32))return a;if(r===41){if(u===0)break;u--}i++}return t===i||u!==0||(a.str=N3(e.slice(t,i)),a.pos=i,a.ok=!0),a}function kFe(e,t,n,r){let i,a=t;const u={ok:!1,can_continue:!1,pos:0,str:"",marker:0};if(r)u.str=r.str,u.marker=r.marker;else{if(a>=n)return u;let l=e.charCodeAt(a);if(l!==34&&l!==39&&l!==40)return u;t++,a++,l===40&&(l=41),u.marker=l}for(;a<n;){if(i=e.charCodeAt(a),i===u.marker)return u.pos=a+1,u.str+=N3(e.slice(t,a)),u.ok=!0,u;if(i===40&&u.marker===41)return u;i===92&&a+1<n&&a++,a++}return u.can_continue=!0,u.str+=N3(e.slice(t,a)),u}const PFe=Object.freeze(Object.defineProperty({__proto__:null,parseLinkDestination:TFe,parseLinkLabel:AFe,parseLinkTitle:kFe},Symbol.toStringTag,{value:"Module"})),q0={};q0.code_inline=function(e,t,n,r,i){const a=e[t];return"<code"+i.renderAttrs(a)+">"+ay(a.content)+"</code>"};q0.code_block=function(e,t,n,r,i){const a=e[t];return"<pre"+i.renderAttrs(a)+"><code>"+ay(e[t].content)+`</code></pre>
`};q0.fence=function(e,t,n,r,i){const a=e[t],u=a.info?N3(a.info).trim():"";let l="",o="";if(u){const c=u.split(/(\s+)/g);l=c[0],o=c.slice(2).join("")}let s;if(n.highlight?s=n.highlight(a.content,l,o)||ay(a.content):s=ay(a.content),s.indexOf("<pre")===0)return s+`
`;if(u){const c=a.attrIndex("class"),d=a.attrs?a.attrs.slice():[];c<0?d.push(["class",n.langPrefix+l]):(d[c]=d[c].slice(),d[c][1]+=" "+n.langPrefix+l);const h={attrs:d};return`<pre><code${i.renderAttrs(h)}>${s}</code></pre>
`}return`<pre><code${i.renderAttrs(a)}>${s}</code></pre>
`};q0.image=function(e,t,n,r,i){const a=e[t];return a.attrs[a.attrIndex("alt")][1]=i.renderInlineAsText(a.children,n,r),i.renderToken(e,t,n)};q0.hardbreak=function(e,t,n){return n.xhtmlOut?`<br />
`:`<br>
`};q0.softbreak=function(e,t,n){return n.breaks?n.xhtmlOut?`<br />
`:`<br>
`:`
`};q0.text=function(e,t){return ay(e[t].content)};q0.html_block=function(e,t){return e[t].content};q0.html_inline=function(e,t){return e[t].content};function u4(){this.rules=$w({},q0)}u4.prototype.renderAttrs=function(t){let n,r,i;if(!t.attrs)return"";for(i="",n=0,r=t.attrs.length;n<r;n++)i+=" "+ay(t.attrs[n][0])+'="'+ay(t.attrs[n][1])+'"';return i};u4.prototype.renderToken=function(t,n,r){const i=t[n];let a="";if(i.hidden)return"";i.block&&i.nesting!==-1&&n&&t[n-1].hidden&&(a+=`
`),a+=(i.nesting===-1?"</":"<")+i.tag,a+=this.renderAttrs(i),i.nesting===0&&r.xhtmlOut&&(a+=" /");let u=!1;if(i.block&&(u=!0,i.nesting===1&&n+1<t.length)){const l=t[n+1];(l.type==="inline"||l.hidden||l.nesting===-1&&l.tag===i.tag)&&(u=!1)}return a+=u?`>
`:">",a};u4.prototype.renderInline=function(e,t,n){let r="";const i=this.rules;for(let a=0,u=e.length;a<u;a++){const l=e[a].type;typeof i[l]<"u"?r+=i[l](e,a,t,n,this):r+=this.renderToken(e,a,t)}return r};u4.prototype.renderInlineAsText=function(e,t,n){let r="";for(let i=0,a=e.length;i<a;i++)switch(e[i].type){case"text":r+=e[i].content;break;case"image":r+=this.renderInlineAsText(e[i].children,t,n);break;case"html_inline":case"html_block":r+=e[i].content;break;case"softbreak":case"hardbreak":r+=`
`;break}return r};u4.prototype.render=function(e,t,n){let r="";const i=this.rules;for(let a=0,u=e.length;a<u;a++){const l=e[a].type;l==="inline"?r+=this.renderInline(e[a].children,t,n):typeof i[l]<"u"?r+=i[l](e,a,t,n,this):r+=this.renderToken(e,a,t,n)}return r};function df(){this.__rules__=[],this.__cache__=null}df.prototype.__find__=function(e){for(let t=0;t<this.__rules__.length;t++)if(this.__rules__[t].name===e)return t;return-1};df.prototype.__compile__=function(){const e=this,t=[""];e.__rules__.forEach(function(n){n.enabled&&n.alt.forEach(function(r){t.indexOf(r)<0&&t.push(r)})}),e.__cache__={},t.forEach(function(n){e.__cache__[n]=[],e.__rules__.forEach(function(r){r.enabled&&(n&&r.alt.indexOf(n)<0||e.__cache__[n].push(r.fn))})})};df.prototype.at=function(e,t,n){const r=this.__find__(e),i=n||{};if(r===-1)throw new Error("Parser rule not found: "+e);this.__rules__[r].fn=t,this.__rules__[r].alt=i.alt||[],this.__cache__=null};df.prototype.before=function(e,t,n,r){const i=this.__find__(e),a=r||{};if(i===-1)throw new Error("Parser rule not found: "+e);this.__rules__.splice(i,0,{name:t,enabled:!0,fn:n,alt:a.alt||[]}),this.__cache__=null};df.prototype.after=function(e,t,n,r){const i=this.__find__(e),a=r||{};if(i===-1)throw new Error("Parser rule not found: "+e);this.__rules__.splice(i+1,0,{name:t,enabled:!0,fn:n,alt:a.alt||[]}),this.__cache__=null};df.prototype.push=function(e,t,n){const r=n||{};this.__rules__.push({name:e,enabled:!0,fn:t,alt:r.alt||[]}),this.__cache__=null};df.prototype.enable=function(e,t){Array.isArray(e)||(e=[e]);const n=[];return e.forEach(function(r){const i=this.__find__(r);if(i<0){if(t)return;throw new Error("Rules manager: invalid rule name "+r)}this.__rules__[i].enabled=!0,n.push(r)},this),this.__cache__=null,n};df.prototype.enableOnly=function(e,t){Array.isArray(e)||(e=[e]),this.__rules__.forEach(function(n){n.enabled=!1}),this.enable(e,t)};df.prototype.disable=function(e,t){Array.isArray(e)||(e=[e]);const n=[];return e.forEach(function(r){const i=this.__find__(r);if(i<0){if(t)return;throw new Error("Rules manager: invalid rule name "+r)}this.__rules__[i].enabled=!1,n.push(r)},this),this.__cache__=null,n};df.prototype.getRules=function(e){return this.__cache__===null&&this.__compile__(),this.__cache__[e]||[]};function vp(e,t,n){this.type=e,this.tag=t,this.attrs=null,this.map=null,this.nesting=n,this.level=0,this.children=null,this.content="",this.markup="",this.info="",this.meta=null,this.block=!1,this.hidden=!1}vp.prototype.attrIndex=function(t){if(!this.attrs)return-1;const n=this.attrs;for(let r=0,i=n.length;r<i;r++)if(n[r][0]===t)return r;return-1};vp.prototype.attrPush=function(t){this.attrs?this.attrs.push(t):this.attrs=[t]};vp.prototype.attrSet=function(t,n){const r=this.attrIndex(t),i=[t,n];r<0?this.attrPush(i):this.attrs[r]=i};vp.prototype.attrGet=function(t){const n=this.attrIndex(t);let r=null;return n>=0&&(r=this.attrs[n][1]),r};vp.prototype.attrJoin=function(t,n){const r=this.attrIndex(t);r<0?this.attrPush([t,n]):this.attrs[r][1]=this.attrs[r][1]+" "+n};function mme(e,t,n){this.src=e,this.env=n,this.tokens=[],this.inlineMode=!1,this.md=t}mme.prototype.Token=vp;const MFe=/\r\n?|\n/g,FFe=/\0/g;function NFe(e){let t;t=e.src.replace(MFe,`
`),t=t.replace(FFe,"�"),e.src=t}function IFe(e){let t;e.inlineMode?(t=new e.Token("inline","",0),t.content=e.src,t.map=[0,1],t.children=[],e.tokens.push(t)):e.md.block.parse(e.src,e.md,e.env,e.tokens)}function LFe(e){const t=e.tokens;for(let n=0,r=t.length;n<r;n++){const i=t[n];i.type==="inline"&&e.md.inline.parse(i.content,e.md,e.env,i.children)}}function BFe(e){return/^<a[>\s]/i.test(e)}function QFe(e){return/^<\/a\s*>/i.test(e)}function zFe(e){const t=e.tokens;if(e.md.options.linkify)for(let n=0,r=t.length;n<r;n++){if(t[n].type!=="inline"||!e.md.linkify.pretest(t[n].content))continue;let i=t[n].children,a=0;for(let u=i.length-1;u>=0;u--){const l=i[u];if(l.type==="link_close"){for(u--;i[u].level!==l.level&&i[u].type!=="link_open";)u--;continue}if(l.type==="html_inline"&&(BFe(l.content)&&a>0&&a--,QFe(l.content)&&a++),!(a>0)&&l.type==="text"&&e.md.linkify.test(l.content)){const o=l.content;let s=e.md.linkify.match(o);const c=[];let d=l.level,h=0;s.length>0&&s[0].index===0&&u>0&&i[u-1].type==="text_special"&&(s=s.slice(1));for(let m=0;m<s.length;m++){const y=s[m].url,v=e.md.normalizeLink(y);if(!e.md.validateLink(v))continue;let S=s[m].text;s[m].schema?s[m].schema==="mailto:"&&!/^mailto:/i.test(S)?S=e.md.normalizeLinkText("mailto:"+S).replace(/^mailto:/,""):S=e.md.normalizeLinkText(S):S=e.md.normalizeLinkText("http://"+S).replace(/^http:\/\//,"");const _=s[m].index;if(_>h){const A=new e.Token("text","",0);A.content=o.slice(h,_),A.level=d,c.push(A)}const E=new e.Token("link_open","a",1);E.attrs=[["href",v]],E.level=d++,E.markup="linkify",E.info="auto",c.push(E);const $=new e.Token("text","",0);$.content=S,$.level=d,c.push($);const k=new e.Token("link_close","a",-1);k.level=--d,k.markup="linkify",k.info="auto",c.push(k),h=s[m].lastIndex}if(h<o.length){const m=new e.Token("text","",0);m.content=o.slice(h),m.level=d,c.push(m)}t[n].children=i=hme(i,u,c)}}}}const gme=/\+-|\.\.|\?\?\?\?|!!!!|,,|--/,qFe=/\((c|tm|r)\)/i,VFe=/\((c|tm|r)\)/ig,UFe={c:"©",r:"®",tm:"™"};function jFe(e,t){return UFe[t.toLowerCase()]}function ZFe(e){let t=0;for(let n=e.length-1;n>=0;n--){const r=e[n];r.type==="text"&&!t&&(r.content=r.content.replace(VFe,jFe)),r.type==="link_open"&&r.info==="auto"&&t--,r.type==="link_close"&&r.info==="auto"&&t++}}function HFe(e){let t=0;for(let n=e.length-1;n>=0;n--){const r=e[n];r.type==="text"&&!t&&gme.test(r.content)&&(r.content=r.content.replace(/\+-/g,"±").replace(/\.{2,}/g,"…").replace(/([?!])…/g,"$1..").replace(/([?!]){4,}/g,"$1$1$1").replace(/,{2,}/g,",").replace(/(^|[^-])---(?=[^-]|$)/mg,"$1—").replace(/(^|\s)--(?=\s|$)/mg,"$1–").replace(/(^|[^-\s])--(?=[^-\s]|$)/mg,"$1–")),r.type==="link_open"&&r.info==="auto"&&t--,r.type==="link_close"&&r.info==="auto"&&t++}}function XFe(e){let t;if(e.md.options.typographer)for(t=e.tokens.length-1;t>=0;t--)e.tokens[t].type==="inline"&&(qFe.test(e.tokens[t].content)&&ZFe(e.tokens[t].children),gme.test(e.tokens[t].content)&&HFe(e.tokens[t].children))}const YFe=/['"]/,_X=/['"]/g,EX="’";function OE(e,t,n){return e.slice(0,t)+n+e.slice(t+1)}function GFe(e,t){let n;const r=[];for(let i=0;i<e.length;i++){const a=e[i],u=e[i].level;for(n=r.length-1;n>=0&&!(r[n].level<=u);n--);if(r.length=n+1,a.type!=="text")continue;let l=a.content,o=0,s=l.length;e:for(;o<s;){_X.lastIndex=o;const c=_X.exec(l);if(!c)break;let d=!0,h=!0;o=c.index+1;const m=c[0]==="'";let y=32;if(c.index-1>=0)y=l.charCodeAt(c.index-1);else for(n=i-1;n>=0&&!(e[n].type==="softbreak"||e[n].type==="hardbreak");n--)if(e[n].content){y=e[n].content.charCodeAt(e[n].content.length-1);break}let v=32;if(o<s)v=l.charCodeAt(o);else for(n=i+1;n<e.length&&!(e[n].type==="softbreak"||e[n].type==="hardbreak");n++)if(e[n].content){v=e[n].content.charCodeAt(0);break}const S=qx(y)||zx(String.fromCharCode(y)),_=qx(v)||zx(String.fromCharCode(v)),E=Qx(y),$=Qx(v);if($?d=!1:_&&(E||S||(d=!1)),E?h=!1:S&&($||_||(h=!1)),v===34&&c[0]==='"'&&y>=48&&y<=57&&(h=d=!1),d&&h&&(d=S,h=_),!d&&!h){m&&(a.content=OE(a.content,c.index,EX));continue}if(h)for(n=r.length-1;n>=0;n--){let k=r[n];if(r[n].level<u)break;if(k.single===m&&r[n].level===u){k=r[n];let A,w;m?(A=t.md.options.quotes[2],w=t.md.options.quotes[3]):(A=t.md.options.quotes[0],w=t.md.options.quotes[1]),a.content=OE(a.content,c.index,w),e[k.token].content=OE(e[k.token].content,k.pos,A),o+=w.length-1,k.token===i&&(o+=A.length-1),l=a.content,s=l.length,r.length=n;continue e}}d?r.push({token:i,pos:c.index,single:m,level:u}):h&&m&&(a.content=OE(a.content,c.index,EX))}}}function KFe(e){if(e.md.options.typographer)for(let t=e.tokens.length-1;t>=0;t--)e.tokens[t].type!=="inline"||!YFe.test(e.tokens[t].content)||GFe(e.tokens[t].children,e)}function WFe(e){let t,n;const r=e.tokens,i=r.length;for(let a=0;a<i;a++){if(r[a].type!=="inline")continue;const u=r[a].children,l=u.length;for(t=0;t<l;t++)u[t].type==="text_special"&&(u[t].type="text");for(t=n=0;t<l;t++)u[t].type==="text"&&t+1<l&&u[t+1].type==="text"?u[t+1].content=u[t].content+u[t+1].content:(t!==n&&(u[n]=u[t]),n++);t!==n&&(u.length=n)}}const yP=[["normalize",NFe],["block",IFe],["inline",LFe],["linkify",zFe],["replacements",XFe],["smartquotes",KFe],["text_join",WFe]];function nL(){this.ruler=new df;for(let e=0;e<yP.length;e++)this.ruler.push(yP[e][0],yP[e][1])}nL.prototype.process=function(e){const t=this.ruler.getRules("");for(let n=0,r=t.length;n<r;n++)t[n](e)};nL.prototype.State=mme;function V0(e,t,n,r){this.src=e,this.md=t,this.env=n,this.tokens=r,this.bMarks=[],this.eMarks=[],this.tShift=[],this.sCount=[],this.bsCount=[],this.blkIndent=0,this.line=0,this.lineMax=0,this.tight=!1,this.ddIndent=-1,this.listIndent=-1,this.parentType="root",this.level=0;const i=this.src;for(let a=0,u=0,l=0,o=0,s=i.length,c=!1;u<s;u++){const d=i.charCodeAt(u);if(!c)if(_a(d)){l++,d===9?o+=4-o%4:o++;continue}else c=!0;(d===10||u===s-1)&&(d!==10&&u++,this.bMarks.push(a),this.eMarks.push(u),this.tShift.push(l),this.sCount.push(o),this.bsCount.push(0),c=!1,l=0,o=0,a=u+1)}this.bMarks.push(i.length),this.eMarks.push(i.length),this.tShift.push(0),this.sCount.push(0),this.bsCount.push(0),this.lineMax=this.bMarks.length-1}V0.prototype.push=function(e,t,n){const r=new vp(e,t,n);return r.block=!0,n<0&&this.level--,r.level=this.level,n>0&&this.level++,this.tokens.push(r),r};V0.prototype.isEmpty=function(t){return this.bMarks[t]+this.tShift[t]>=this.eMarks[t]};V0.prototype.skipEmptyLines=function(t){for(let n=this.lineMax;t<n&&!(this.bMarks[t]+this.tShift[t]<this.eMarks[t]);t++);return t};V0.prototype.skipSpaces=function(t){for(let n=this.src.length;t<n;t++){const r=this.src.charCodeAt(t);if(!_a(r))break}return t};V0.prototype.skipSpacesBack=function(t,n){if(t<=n)return t;for(;t>n;)if(!_a(this.src.charCodeAt(--t)))return t+1;return t};V0.prototype.skipChars=function(t,n){for(let r=this.src.length;t<r&&this.src.charCodeAt(t)===n;t++);return t};V0.prototype.skipCharsBack=function(t,n,r){if(t<=r)return t;for(;t>r;)if(n!==this.src.charCodeAt(--t))return t+1;return t};V0.prototype.getLines=function(t,n,r,i){if(t>=n)return"";const a=new Array(n-t);for(let u=0,l=t;l<n;l++,u++){let o=0;const s=this.bMarks[l];let c=s,d;for(l+1<n||i?d=this.eMarks[l]+1:d=this.eMarks[l];c<d&&o<r;){const h=this.src.charCodeAt(c);if(_a(h))h===9?o+=4-(o+this.bsCount[l])%4:o++;else if(c-s<this.tShift[l])o++;else break;c++}o>r?a[u]=new Array(o-r+1).join(" ")+this.src.slice(c,d):a[u]=this.src.slice(c,d)}return a.join("")};V0.prototype.Token=vp;const JFe=65536;function bP(e,t){const n=e.bMarks[t]+e.tShift[t],r=e.eMarks[t];return e.src.slice(n,r)}function DX(e){const t=[],n=e.length;let r=0,i=e.charCodeAt(r),a=!1,u=0,l="";for(;r<n;)i===124&&(a?(l+=e.substring(u,r-1),u=r):(t.push(l+e.substring(u,r)),l="",u=r+1)),a=i===92,r++,i=e.charCodeAt(r);return t.push(l+e.substring(u)),t}function eNe(e,t,n,r){if(t+2>n)return!1;let i=t+1;if(e.sCount[i]<e.blkIndent||e.sCount[i]-e.blkIndent>=4)return!1;let a=e.bMarks[i]+e.tShift[i];if(a>=e.eMarks[i])return!1;const u=e.src.charCodeAt(a++);if(u!==124&&u!==45&&u!==58||a>=e.eMarks[i])return!1;const l=e.src.charCodeAt(a++);if(l!==124&&l!==45&&l!==58&&!_a(l)||u===45&&_a(l))return!1;for(;a<e.eMarks[i];){const k=e.src.charCodeAt(a);if(k!==124&&k!==45&&k!==58&&!_a(k))return!1;a++}let o=bP(e,t+1),s=o.split("|");const c=[];for(let k=0;k<s.length;k++){const A=s[k].trim();if(!A){if(k===0||k===s.length-1)continue;return!1}if(!/^:?-+:?$/.test(A))return!1;A.charCodeAt(A.length-1)===58?c.push(A.charCodeAt(0)===58?"center":"right"):A.charCodeAt(0)===58?c.push("left"):c.push("")}if(o=bP(e,t).trim(),o.indexOf("|")===-1||e.sCount[t]-e.blkIndent>=4)return!1;s=DX(o),s.length&&s[0]===""&&s.shift(),s.length&&s[s.length-1]===""&&s.pop();const d=s.length;if(d===0||d!==c.length)return!1;if(r)return!0;const h=e.parentType;e.parentType="table";const m=e.md.block.ruler.getRules("blockquote"),y=e.push("table_open","table",1),v=[t,0];y.map=v;const S=e.push("thead_open","thead",1);S.map=[t,t+1];const _=e.push("tr_open","tr",1);_.map=[t,t+1];for(let k=0;k<s.length;k++){const A=e.push("th_open","th",1);c[k]&&(A.attrs=[["style","text-align:"+c[k]]]);const w=e.push("inline","",0);w.content=s[k].trim(),w.children=[],e.push("th_close","th",-1)}e.push("tr_close","tr",-1),e.push("thead_close","thead",-1);let E,$=0;for(i=t+2;i<n&&!(e.sCount[i]<e.blkIndent);i++){let k=!1;for(let w=0,R=m.length;w<R;w++)if(m[w](e,i,n,!0)){k=!0;break}if(k||(o=bP(e,i).trim(),!o)||e.sCount[i]-e.blkIndent>=4||(s=DX(o),s.length&&s[0]===""&&s.shift(),s.length&&s[s.length-1]===""&&s.pop(),$+=d-s.length,$>JFe))break;if(i===t+2){const w=e.push("tbody_open","tbody",1);w.map=E=[t+2,0]}const A=e.push("tr_open","tr",1);A.map=[i,i+1];for(let w=0;w<d;w++){const R=e.push("td_open","td",1);c[w]&&(R.attrs=[["style","text-align:"+c[w]]]);const F=e.push("inline","",0);F.content=s[w]?s[w].trim():"",F.children=[],e.push("td_close","td",-1)}e.push("tr_close","tr",-1)}return E&&(e.push("tbody_close","tbody",-1),E[1]=i),e.push("table_close","table",-1),v[1]=i,e.parentType=h,e.line=i,!0}function tNe(e,t,n){if(e.sCount[t]-e.blkIndent<4)return!1;let r=t+1,i=r;for(;r<n;){if(e.isEmpty(r)){r++;continue}if(e.sCount[r]-e.blkIndent>=4){r++,i=r;continue}break}e.line=i;const a=e.push("code_block","code",0);return a.content=e.getLines(t,i,4+e.blkIndent,!1)+`
`,a.map=[t,e.line],!0}function nNe(e,t,n,r){let i=e.bMarks[t]+e.tShift[t],a=e.eMarks[t];if(e.sCount[t]-e.blkIndent>=4||i+3>a)return!1;const u=e.src.charCodeAt(i);if(u!==126&&u!==96)return!1;let l=i;i=e.skipChars(i,u);let o=i-l;if(o<3)return!1;const s=e.src.slice(l,i),c=e.src.slice(i,a);if(u===96&&c.indexOf(String.fromCharCode(u))>=0)return!1;if(r)return!0;let d=t,h=!1;for(;d++,!(d>=n||(i=l=e.bMarks[d]+e.tShift[d],a=e.eMarks[d],i<a&&e.sCount[d]<e.blkIndent));)if(e.src.charCodeAt(i)===u&&!(e.sCount[d]-e.blkIndent>=4)&&(i=e.skipChars(i,u),!(i-l<o)&&(i=e.skipSpaces(i),!(i<a)))){h=!0;break}o=e.sCount[t],e.line=d+(h?1:0);const m=e.push("fence","code",0);return m.info=c,m.content=e.getLines(t+1,d,o,!0),m.markup=s,m.map=[t,e.line],!0}function rNe(e,t,n,r){let i=e.bMarks[t]+e.tShift[t],a=e.eMarks[t];const u=e.lineMax;if(e.sCount[t]-e.blkIndent>=4||e.src.charCodeAt(i)!==62)return!1;if(r)return!0;const l=[],o=[],s=[],c=[],d=e.md.block.ruler.getRules("blockquote"),h=e.parentType;e.parentType="blockquote";let m=!1,y;for(y=t;y<n;y++){const $=e.sCount[y]<e.blkIndent;if(i=e.bMarks[y]+e.tShift[y],a=e.eMarks[y],i>=a)break;if(e.src.charCodeAt(i++)===62&&!$){let A=e.sCount[y]+1,w,R;e.src.charCodeAt(i)===32?(i++,A++,R=!1,w=!0):e.src.charCodeAt(i)===9?(w=!0,(e.bsCount[y]+A)%4===3?(i++,A++,R=!1):R=!0):w=!1;let F=A;for(l.push(e.bMarks[y]),e.bMarks[y]=i;i<a;){const I=e.src.charCodeAt(i);if(_a(I))I===9?F+=4-(F+e.bsCount[y]+(R?1:0))%4:F++;else break;i++}m=i>=a,o.push(e.bsCount[y]),e.bsCount[y]=e.sCount[y]+1+(w?1:0),s.push(e.sCount[y]),e.sCount[y]=F-A,c.push(e.tShift[y]),e.tShift[y]=i-e.bMarks[y];continue}if(m)break;let k=!1;for(let A=0,w=d.length;A<w;A++)if(d[A](e,y,n,!0)){k=!0;break}if(k){e.lineMax=y,e.blkIndent!==0&&(l.push(e.bMarks[y]),o.push(e.bsCount[y]),c.push(e.tShift[y]),s.push(e.sCount[y]),e.sCount[y]-=e.blkIndent);break}l.push(e.bMarks[y]),o.push(e.bsCount[y]),c.push(e.tShift[y]),s.push(e.sCount[y]),e.sCount[y]=-1}const v=e.blkIndent;e.blkIndent=0;const S=e.push("blockquote_open","blockquote",1);S.markup=">";const _=[t,0];S.map=_,e.md.block.tokenize(e,t,y);const E=e.push("blockquote_close","blockquote",-1);E.markup=">",e.lineMax=u,e.parentType=h,_[1]=e.line;for(let $=0;$<c.length;$++)e.bMarks[$+t]=l[$],e.tShift[$+t]=c[$],e.sCount[$+t]=s[$],e.bsCount[$+t]=o[$];return e.blkIndent=v,!0}function iNe(e,t,n,r){const i=e.eMarks[t];if(e.sCount[t]-e.blkIndent>=4)return!1;let a=e.bMarks[t]+e.tShift[t];const u=e.src.charCodeAt(a++);if(u!==42&&u!==45&&u!==95)return!1;let l=1;for(;a<i;){const s=e.src.charCodeAt(a++);if(s!==u&&!_a(s))return!1;s===u&&l++}if(l<3)return!1;if(r)return!0;e.line=t+1;const o=e.push("hr","hr",0);return o.map=[t,e.line],o.markup=Array(l+1).join(String.fromCharCode(u)),!0}function $X(e,t){const n=e.eMarks[t];let r=e.bMarks[t]+e.tShift[t];const i=e.src.charCodeAt(r++);if(i!==42&&i!==45&&i!==43)return-1;if(r<n){const a=e.src.charCodeAt(r);if(!_a(a))return-1}return r}function CX(e,t){const n=e.bMarks[t]+e.tShift[t],r=e.eMarks[t];let i=n;if(i+1>=r)return-1;let a=e.src.charCodeAt(i++);if(a<48||a>57)return-1;for(;;){if(i>=r)return-1;if(a=e.src.charCodeAt(i++),a>=48&&a<=57){if(i-n>=10)return-1;continue}if(a===41||a===46)break;return-1}return i<r&&(a=e.src.charCodeAt(i),!_a(a))?-1:i}function aNe(e,t){const n=e.level+2;for(let r=t+2,i=e.tokens.length-2;r<i;r++)e.tokens[r].level===n&&e.tokens[r].type==="paragraph_open"&&(e.tokens[r+2].hidden=!0,e.tokens[r].hidden=!0,r+=2)}function uNe(e,t,n,r){let i,a,u,l,o=t,s=!0;if(e.sCount[o]-e.blkIndent>=4||e.listIndent>=0&&e.sCount[o]-e.listIndent>=4&&e.sCount[o]<e.blkIndent)return!1;let c=!1;r&&e.parentType==="paragraph"&&e.sCount[o]>=e.blkIndent&&(c=!0);let d,h,m;if((m=CX(e,o))>=0){if(d=!0,u=e.bMarks[o]+e.tShift[o],h=Number(e.src.slice(u,m-1)),c&&h!==1)return!1}else if((m=$X(e,o))>=0)d=!1;else return!1;if(c&&e.skipSpaces(m)>=e.eMarks[o])return!1;if(r)return!0;const y=e.src.charCodeAt(m-1),v=e.tokens.length;d?(l=e.push("ordered_list_open","ol",1),h!==1&&(l.attrs=[["start",h]])):l=e.push("bullet_list_open","ul",1);const S=[o,0];l.map=S,l.markup=String.fromCharCode(y);let _=!1;const E=e.md.block.ruler.getRules("list"),$=e.parentType;for(e.parentType="list";o<n;){a=m,i=e.eMarks[o];const k=e.sCount[o]+m-(e.bMarks[o]+e.tShift[o]);let A=k;for(;a<i;){const W=e.src.charCodeAt(a);if(W===9)A+=4-(A+e.bsCount[o])%4;else if(W===32)A++;else break;a++}const w=a;let R;w>=i?R=1:R=A-k,R>4&&(R=1);const F=k+R;l=e.push("list_item_open","li",1),l.markup=String.fromCharCode(y);const I=[o,0];l.map=I,d&&(l.info=e.src.slice(u,m-1));const U=e.tight,q=e.tShift[o],V=e.sCount[o],N=e.listIndent;if(e.listIndent=e.blkIndent,e.blkIndent=F,e.tight=!0,e.tShift[o]=w-e.bMarks[o],e.sCount[o]=A,w>=i&&e.isEmpty(o+1)?e.line=Math.min(e.line+2,n):e.md.block.tokenize(e,o,n,!0),(!e.tight||_)&&(s=!1),_=e.line-o>1&&e.isEmpty(e.line-1),e.blkIndent=e.listIndent,e.listIndent=N,e.tShift[o]=q,e.sCount[o]=V,e.tight=U,l=e.push("list_item_close","li",-1),l.markup=String.fromCharCode(y),o=e.line,I[1]=o,o>=n||e.sCount[o]<e.blkIndent||e.sCount[o]-e.blkIndent>=4)break;let j=!1;for(let W=0,K=E.length;W<K;W++)if(E[W](e,o,n,!0)){j=!0;break}if(j)break;if(d){if(m=CX(e,o),m<0)break;u=e.bMarks[o]+e.tShift[o]}else if(m=$X(e,o),m<0)break;if(y!==e.src.charCodeAt(m-1))break}return d?l=e.push("ordered_list_close","ol",-1):l=e.push("bullet_list_close","ul",-1),l.markup=String.fromCharCode(y),S[1]=o,e.line=o,e.parentType=$,s&&aNe(e,v),!0}function lNe(e,t,n,r){let i=e.bMarks[t]+e.tShift[t],a=e.eMarks[t],u=t+1;if(e.sCount[t]-e.blkIndent>=4||e.src.charCodeAt(i)!==91)return!1;function l(E){const $=e.lineMax;if(E>=$||e.isEmpty(E))return null;let k=!1;if(e.sCount[E]-e.blkIndent>3&&(k=!0),e.sCount[E]<0&&(k=!0),!k){const R=e.md.block.ruler.getRules("reference"),F=e.parentType;e.parentType="reference";let I=!1;for(let U=0,q=R.length;U<q;U++)if(R[U](e,E,$,!0)){I=!0;break}if(e.parentType=F,I)return null}const A=e.bMarks[E]+e.tShift[E],w=e.eMarks[E];return e.src.slice(A,w+1)}let o=e.src.slice(i,a+1);a=o.length;let s=-1;for(i=1;i<a;i++){const E=o.charCodeAt(i);if(E===91)return!1;if(E===93){s=i;break}else if(E===10){const $=l(u);$!==null&&(o+=$,a=o.length,u++)}else if(E===92&&(i++,i<a&&o.charCodeAt(i)===10)){const $=l(u);$!==null&&(o+=$,a=o.length,u++)}}if(s<0||o.charCodeAt(s+1)!==58)return!1;for(i=s+2;i<a;i++){const E=o.charCodeAt(i);if(E===10){const $=l(u);$!==null&&(o+=$,a=o.length,u++)}else if(!_a(E))break}const c=e.md.helpers.parseLinkDestination(o,i,a);if(!c.ok)return!1;const d=e.md.normalizeLink(c.str);if(!e.md.validateLink(d))return!1;i=c.pos;const h=i,m=u,y=i;for(;i<a;i++){const E=o.charCodeAt(i);if(E===10){const $=l(u);$!==null&&(o+=$,a=o.length,u++)}else if(!_a(E))break}let v=e.md.helpers.parseLinkTitle(o,i,a);for(;v.can_continue;){const E=l(u);if(E===null)break;o+=E,i=a,a=o.length,u++,v=e.md.helpers.parseLinkTitle(o,i,a,v)}let S;for(i<a&&y!==i&&v.ok?(S=v.str,i=v.pos):(S="",i=h,u=m);i<a;){const E=o.charCodeAt(i);if(!_a(E))break;i++}if(i<a&&o.charCodeAt(i)!==10&&S)for(S="",i=h,u=m;i<a;){const E=o.charCodeAt(i);if(!_a(E))break;i++}if(i<a&&o.charCodeAt(i)!==10)return!1;const _=Cw(o.slice(1,s));return _?(r||(typeof e.env.references>"u"&&(e.env.references={}),typeof e.env.references[_]>"u"&&(e.env.references[_]={title:S,href:d}),e.line=u),!0):!1}const oNe=["address","article","aside","base","basefont","blockquote","body","caption","center","col","colgroup","dd","details","dialog","dir","div","dl","dt","fieldset","figcaption","figure","footer","form","frame","frameset","h1","h2","h3","h4","h5","h6","head","header","hr","html","iframe","legend","li","link","main","menu","menuitem","nav","noframes","ol","optgroup","option","p","param","search","section","summary","table","tbody","td","tfoot","th","thead","title","tr","track","ul"],sNe="[a-zA-Z_:][a-zA-Z0-9:._-]*",cNe="[^\"'=<>`\\x00-\\x20]+",fNe="'[^']*'",dNe='"[^"]*"',hNe="(?:"+cNe+"|"+fNe+"|"+dNe+")",pNe="(?:\\s+"+sNe+"(?:\\s*=\\s*"+hNe+")?)",vme="<[A-Za-z][A-Za-z0-9\\-]*"+pNe+"*\\s*\\/?>",yme="<\\/[A-Za-z][A-Za-z0-9\\-]*\\s*>",mNe="<!---?>|<!--(?:[^-]|-[^-]|--[^>])*-->",gNe="<[?][\\s\\S]*?[?]>",vNe="<![A-Za-z][^>]*>",yNe="<!\\[CDATA\\[[\\s\\S]*?\\]\\]>",bNe=new RegExp("^(?:"+vme+"|"+yme+"|"+mNe+"|"+gNe+"|"+vNe+"|"+yNe+")"),ONe=new RegExp("^(?:"+vme+"|"+yme+")"),EO=[[/^<(script|pre|style|textarea)(?=(\s|>|$))/i,/<\/(script|pre|style|textarea)>/i,!0],[/^<!--/,/-->/,!0],[/^<\?/,/\?>/,!0],[/^<![A-Z]/,/>/,!0],[/^<!\[CDATA\[/,/\]\]>/,!0],[new RegExp("^</?("+oNe.join("|")+")(?=(\\s|/?>|$))","i"),/^$/,!0],[new RegExp(ONe.source+"\\s*$"),/^$/,!1]];function xNe(e,t,n,r){let i=e.bMarks[t]+e.tShift[t],a=e.eMarks[t];if(e.sCount[t]-e.blkIndent>=4||!e.md.options.html||e.src.charCodeAt(i)!==60)return!1;let u=e.src.slice(i,a),l=0;for(;l<EO.length&&!EO[l][0].test(u);l++);if(l===EO.length)return!1;if(r)return EO[l][2];let o=t+1;if(!EO[l][1].test(u)){for(;o<n&&!(e.sCount[o]<e.blkIndent);o++)if(i=e.bMarks[o]+e.tShift[o],a=e.eMarks[o],u=e.src.slice(i,a),EO[l][1].test(u)){u.length!==0&&o++;break}}e.line=o;const s=e.push("html_block","",0);return s.map=[t,o],s.content=e.getLines(t,o,e.blkIndent,!0),!0}function SNe(e,t,n,r){let i=e.bMarks[t]+e.tShift[t],a=e.eMarks[t];if(e.sCount[t]-e.blkIndent>=4)return!1;let u=e.src.charCodeAt(i);if(u!==35||i>=a)return!1;let l=1;for(u=e.src.charCodeAt(++i);u===35&&i<a&&l<=6;)l++,u=e.src.charCodeAt(++i);if(l>6||i<a&&!_a(u))return!1;if(r)return!0;a=e.skipSpacesBack(a,i);const o=e.skipCharsBack(a,35,i);o>i&&_a(e.src.charCodeAt(o-1))&&(a=o),e.line=t+1;const s=e.push("heading_open","h"+String(l),1);s.markup="########".slice(0,l),s.map=[t,e.line];const c=e.push("inline","",0);c.content=e.src.slice(i,a).trim(),c.map=[t,e.line],c.children=[];const d=e.push("heading_close","h"+String(l),-1);return d.markup="########".slice(0,l),!0}function _Ne(e,t,n){const r=e.md.block.ruler.getRules("paragraph");if(e.sCount[t]-e.blkIndent>=4)return!1;const i=e.parentType;e.parentType="paragraph";let a=0,u,l=t+1;for(;l<n&&!e.isEmpty(l);l++){if(e.sCount[l]-e.blkIndent>3)continue;if(e.sCount[l]>=e.blkIndent){let m=e.bMarks[l]+e.tShift[l];const y=e.eMarks[l];if(m<y&&(u=e.src.charCodeAt(m),(u===45||u===61)&&(m=e.skipChars(m,u),m=e.skipSpaces(m),m>=y))){a=u===61?1:2;break}}if(e.sCount[l]<0)continue;let h=!1;for(let m=0,y=r.length;m<y;m++)if(r[m](e,l,n,!0)){h=!0;break}if(h)break}if(!a)return!1;const o=e.getLines(t,l,e.blkIndent,!1).trim();e.line=l+1;const s=e.push("heading_open","h"+String(a),1);s.markup=String.fromCharCode(u),s.map=[t,e.line];const c=e.push("inline","",0);c.content=o,c.map=[t,e.line-1],c.children=[];const d=e.push("heading_close","h"+String(a),-1);return d.markup=String.fromCharCode(u),e.parentType=i,!0}function ENe(e,t,n){const r=e.md.block.ruler.getRules("paragraph"),i=e.parentType;let a=t+1;for(e.parentType="paragraph";a<n&&!e.isEmpty(a);a++){if(e.sCount[a]-e.blkIndent>3||e.sCount[a]<0)continue;let s=!1;for(let c=0,d=r.length;c<d;c++)if(r[c](e,a,n,!0)){s=!0;break}if(s)break}const u=e.getLines(t,a,e.blkIndent,!1).trim();e.line=a;const l=e.push("paragraph_open","p",1);l.map=[t,e.line];const o=e.push("inline","",0);return o.content=u,o.map=[t,e.line],o.children=[],e.push("paragraph_close","p",-1),e.parentType=i,!0}const xE=[["table",eNe,["paragraph","reference"]],["code",tNe],["fence",nNe,["paragraph","reference","blockquote","list"]],["blockquote",rNe,["paragraph","reference","blockquote","list"]],["hr",iNe,["paragraph","reference","blockquote","list"]],["list",uNe,["paragraph","reference","blockquote"]],["reference",lNe],["html_block",xNe,["paragraph","reference","blockquote"]],["heading",SNe,["paragraph","reference","blockquote"]],["lheading",_Ne],["paragraph",ENe]];function ww(){this.ruler=new df;for(let e=0;e<xE.length;e++)this.ruler.push(xE[e][0],xE[e][1],{alt:(xE[e][2]||[]).slice()})}ww.prototype.tokenize=function(e,t,n){const r=this.ruler.getRules(""),i=r.length,a=e.md.options.maxNesting;let u=t,l=!1;for(;u<n&&(e.line=u=e.skipEmptyLines(u),!(u>=n||e.sCount[u]<e.blkIndent));){if(e.level>=a){e.line=n;break}const o=e.line;let s=!1;for(let c=0;c<i;c++)if(s=r[c](e,u,n,!1),s){if(o>=e.line)throw new Error("block rule didn't increment state.line");break}if(!s)throw new Error("none of the block rules matched");e.tight=!l,e.isEmpty(e.line-1)&&(l=!0),u=e.line,u<n&&e.isEmpty(u)&&(l=!0,u++,e.line=u)}};ww.prototype.parse=function(e,t,n,r){if(!e)return;const i=new this.State(e,t,n,r);this.tokenize(i,i.line,i.lineMax)};ww.prototype.State=V0;function kS(e,t,n,r){this.src=e,this.env=n,this.md=t,this.tokens=r,this.tokens_meta=Array(r.length),this.pos=0,this.posMax=this.src.length,this.level=0,this.pending="",this.pendingLevel=0,this.cache={},this.delimiters=[],this._prev_delimiters=[],this.backticks={},this.backticksScanned=!1,this.linkLevel=0}kS.prototype.pushPending=function(){const e=new vp("text","",0);return e.content=this.pending,e.level=this.pendingLevel,this.tokens.push(e),this.pending="",e};kS.prototype.push=function(e,t,n){this.pending&&this.pushPending();const r=new vp(e,t,n);let i=null;return n<0&&(this.level--,this.delimiters=this._prev_delimiters.pop()),r.level=this.level,n>0&&(this.level++,this._prev_delimiters.push(this.delimiters),this.delimiters=[],i={delimiters:this.delimiters}),this.pendingLevel=this.level,this.tokens.push(r),this.tokens_meta.push(i),r};kS.prototype.scanDelims=function(e,t){const n=this.posMax,r=this.src.charCodeAt(e),i=e>0?this.src.charCodeAt(e-1):32;let a=e;for(;a<n&&this.src.charCodeAt(a)===r;)a++;const u=a-e,l=a<n?this.src.charCodeAt(a):32,o=qx(i)||zx(String.fromCharCode(i)),s=qx(l)||zx(String.fromCharCode(l)),c=Qx(i),d=Qx(l),h=!d&&(!s||c||o),m=!c&&(!o||d||s);return{can_open:h&&(t||!m||o),can_close:m&&(t||!h||s),length:u}};kS.prototype.Token=vp;function DNe(e){switch(e){case 10:case 33:case 35:case 36:case 37:case 38:case 42:case 43:case 45:case 58:case 60:case 61:case 62:case 64:case 91:case 92:case 93:case 94:case 95:case 96:case 123:case 125:case 126:return!0;default:return!1}}function $Ne(e,t){let n=e.pos;for(;n<e.posMax&&!DNe(e.src.charCodeAt(n));)n++;return n===e.pos?!1:(t||(e.pending+=e.src.slice(e.pos,n)),e.pos=n,!0)}const CNe=/(?:^|[^a-z0-9.+-])([a-z][a-z0-9.+-]*)$/i;function wNe(e,t){if(!e.md.options.linkify||e.linkLevel>0)return!1;const n=e.pos,r=e.posMax;if(n+3>r||e.src.charCodeAt(n)!==58||e.src.charCodeAt(n+1)!==47||e.src.charCodeAt(n+2)!==47)return!1;const i=e.pending.match(CNe);if(!i)return!1;const a=i[1],u=e.md.linkify.matchAtStart(e.src.slice(n-a.length));if(!u)return!1;let l=u.url;if(l.length<=a.length)return!1;l=l.replace(/\*+$/,"");const o=e.md.normalizeLink(l);if(!e.md.validateLink(o))return!1;if(!t){e.pending=e.pending.slice(0,-a.length);const s=e.push("link_open","a",1);s.attrs=[["href",o]],s.markup="linkify",s.info="auto";const c=e.push("text","",0);c.content=e.md.normalizeLinkText(l);const d=e.push("link_close","a",-1);d.markup="linkify",d.info="auto"}return e.pos+=l.length-a.length,!0}function RNe(e,t){let n=e.pos;if(e.src.charCodeAt(n)!==10)return!1;const r=e.pending.length-1,i=e.posMax;if(!t)if(r>=0&&e.pending.charCodeAt(r)===32)if(r>=1&&e.pending.charCodeAt(r-1)===32){let a=r-1;for(;a>=1&&e.pending.charCodeAt(a-1)===32;)a--;e.pending=e.pending.slice(0,a),e.push("hardbreak","br",0)}else e.pending=e.pending.slice(0,-1),e.push("softbreak","br",0);else e.push("softbreak","br",0);for(n++;n<i&&_a(e.src.charCodeAt(n));)n++;return e.pos=n,!0}const rL=[];for(let e=0;e<256;e++)rL.push(0);"\\!\"#$%&'()*+,./:;<=>?@[]^_`{|}~-".split("").forEach(function(e){rL[e.charCodeAt(0)]=1});function ANe(e,t){let n=e.pos;const r=e.posMax;if(e.src.charCodeAt(n)!==92||(n++,n>=r))return!1;let i=e.src.charCodeAt(n);if(i===10){for(t||e.push("hardbreak","br",0),n++;n<r&&(i=e.src.charCodeAt(n),!!_a(i));)n++;return e.pos=n,!0}let a=e.src[n];if(i>=55296&&i<=56319&&n+1<r){const l=e.src.charCodeAt(n+1);l>=56320&&l<=57343&&(a+=e.src[n+1],n++)}const u="\\"+a;if(!t){const l=e.push("text_special","",0);i<256&&rL[i]!==0?l.content=a:l.content=u,l.markup=u,l.info="escape"}return e.pos=n+1,!0}function TNe(e,t){let n=e.pos;if(e.src.charCodeAt(n)!==96)return!1;const i=n;n++;const a=e.posMax;for(;n<a&&e.src.charCodeAt(n)===96;)n++;const u=e.src.slice(i,n),l=u.length;if(e.backticksScanned&&(e.backticks[l]||0)<=i)return t||(e.pending+=u),e.pos+=l,!0;let o=n,s;for(;(s=e.src.indexOf("`",o))!==-1;){for(o=s+1;o<a&&e.src.charCodeAt(o)===96;)o++;const c=o-s;if(c===l){if(!t){const d=e.push("code_inline","code",0);d.markup=u,d.content=e.src.slice(n,s).replace(/\n/g," ").replace(/^ (.+) $/,"$1")}return e.pos=o,!0}e.backticks[c]=s}return e.backticksScanned=!0,t||(e.pending+=u),e.pos+=l,!0}function kNe(e,t){const n=e.pos,r=e.src.charCodeAt(n);if(t||r!==126)return!1;const i=e.scanDelims(e.pos,!0);let a=i.length;const u=String.fromCharCode(r);if(a<2)return!1;let l;a%2&&(l=e.push("text","",0),l.content=u,a--);for(let o=0;o<a;o+=2)l=e.push("text","",0),l.content=u+u,e.delimiters.push({marker:r,length:0,token:e.tokens.length-1,end:-1,open:i.can_open,close:i.can_close});return e.pos+=i.length,!0}function wX(e,t){let n;const r=[],i=t.length;for(let a=0;a<i;a++){const u=t[a];if(u.marker!==126||u.end===-1)continue;const l=t[u.end];n=e.tokens[u.token],n.type="s_open",n.tag="s",n.nesting=1,n.markup="~~",n.content="",n=e.tokens[l.token],n.type="s_close",n.tag="s",n.nesting=-1,n.markup="~~",n.content="",e.tokens[l.token-1].type==="text"&&e.tokens[l.token-1].content==="~"&&r.push(l.token-1)}for(;r.length;){const a=r.pop();let u=a+1;for(;u<e.tokens.length&&e.tokens[u].type==="s_close";)u++;u--,a!==u&&(n=e.tokens[u],e.tokens[u]=e.tokens[a],e.tokens[a]=n)}}function PNe(e){const t=e.tokens_meta,n=e.tokens_meta.length;wX(e,e.delimiters);for(let r=0;r<n;r++)t[r]&&t[r].delimiters&&wX(e,t[r].delimiters)}const bme={tokenize:kNe,postProcess:PNe};function MNe(e,t){const n=e.pos,r=e.src.charCodeAt(n);if(t||r!==95&&r!==42)return!1;const i=e.scanDelims(e.pos,r===42);for(let a=0;a<i.length;a++){const u=e.push("text","",0);u.content=String.fromCharCode(r),e.delimiters.push({marker:r,length:i.length,token:e.tokens.length-1,end:-1,open:i.can_open,close:i.can_close})}return e.pos+=i.length,!0}function RX(e,t){const n=t.length;for(let r=n-1;r>=0;r--){const i=t[r];if(i.marker!==95&&i.marker!==42||i.end===-1)continue;const a=t[i.end],u=r>0&&t[r-1].end===i.end+1&&t[r-1].marker===i.marker&&t[r-1].token===i.token-1&&t[i.end+1].token===a.token+1,l=String.fromCharCode(i.marker),o=e.tokens[i.token];o.type=u?"strong_open":"em_open",o.tag=u?"strong":"em",o.nesting=1,o.markup=u?l+l:l,o.content="";const s=e.tokens[a.token];s.type=u?"strong_close":"em_close",s.tag=u?"strong":"em",s.nesting=-1,s.markup=u?l+l:l,s.content="",u&&(e.tokens[t[r-1].token].content="",e.tokens[t[i.end+1].token].content="",r--)}}function FNe(e){const t=e.tokens_meta,n=e.tokens_meta.length;RX(e,e.delimiters);for(let r=0;r<n;r++)t[r]&&t[r].delimiters&&RX(e,t[r].delimiters)}const Ome={tokenize:MNe,postProcess:FNe};function NNe(e,t){let n,r,i,a,u="",l="",o=e.pos,s=!0;if(e.src.charCodeAt(e.pos)!==91)return!1;const c=e.pos,d=e.posMax,h=e.pos+1,m=e.md.helpers.parseLinkLabel(e,e.pos,!0);if(m<0)return!1;let y=m+1;if(y<d&&e.src.charCodeAt(y)===40){for(s=!1,y++;y<d&&(n=e.src.charCodeAt(y),!(!_a(n)&&n!==10));y++);if(y>=d)return!1;if(o=y,i=e.md.helpers.parseLinkDestination(e.src,y,e.posMax),i.ok){for(u=e.md.normalizeLink(i.str),e.md.validateLink(u)?y=i.pos:u="",o=y;y<d&&(n=e.src.charCodeAt(y),!(!_a(n)&&n!==10));y++);if(i=e.md.helpers.parseLinkTitle(e.src,y,e.posMax),y<d&&o!==y&&i.ok)for(l=i.str,y=i.pos;y<d&&(n=e.src.charCodeAt(y),!(!_a(n)&&n!==10));y++);}(y>=d||e.src.charCodeAt(y)!==41)&&(s=!0),y++}if(s){if(typeof e.env.references>"u")return!1;if(y<d&&e.src.charCodeAt(y)===91?(o=y+1,y=e.md.helpers.parseLinkLabel(e,y),y>=0?r=e.src.slice(o,y++):y=m+1):y=m+1,r||(r=e.src.slice(h,m)),a=e.env.references[Cw(r)],!a)return e.pos=c,!1;u=a.href,l=a.title}if(!t){e.pos=h,e.posMax=m;const v=e.push("link_open","a",1),S=[["href",u]];v.attrs=S,l&&S.push(["title",l]),e.linkLevel++,e.md.inline.tokenize(e),e.linkLevel--,e.push("link_close","a",-1)}return e.pos=y,e.posMax=d,!0}function INe(e,t){let n,r,i,a,u,l,o,s,c="";const d=e.pos,h=e.posMax;if(e.src.charCodeAt(e.pos)!==33||e.src.charCodeAt(e.pos+1)!==91)return!1;const m=e.pos+2,y=e.md.helpers.parseLinkLabel(e,e.pos+1,!1);if(y<0)return!1;if(a=y+1,a<h&&e.src.charCodeAt(a)===40){for(a++;a<h&&(n=e.src.charCodeAt(a),!(!_a(n)&&n!==10));a++);if(a>=h)return!1;for(s=a,l=e.md.helpers.parseLinkDestination(e.src,a,e.posMax),l.ok&&(c=e.md.normalizeLink(l.str),e.md.validateLink(c)?a=l.pos:c=""),s=a;a<h&&(n=e.src.charCodeAt(a),!(!_a(n)&&n!==10));a++);if(l=e.md.helpers.parseLinkTitle(e.src,a,e.posMax),a<h&&s!==a&&l.ok)for(o=l.str,a=l.pos;a<h&&(n=e.src.charCodeAt(a),!(!_a(n)&&n!==10));a++);else o="";if(a>=h||e.src.charCodeAt(a)!==41)return e.pos=d,!1;a++}else{if(typeof e.env.references>"u")return!1;if(a<h&&e.src.charCodeAt(a)===91?(s=a+1,a=e.md.helpers.parseLinkLabel(e,a),a>=0?i=e.src.slice(s,a++):a=y+1):a=y+1,i||(i=e.src.slice(m,y)),u=e.env.references[Cw(i)],!u)return e.pos=d,!1;c=u.href,o=u.title}if(!t){r=e.src.slice(m,y);const v=[];e.md.inline.parse(r,e.md,e.env,v);const S=e.push("image","img",0),_=[["src",c],["alt",""]];S.attrs=_,S.children=v,S.content=r,o&&_.push(["title",o])}return e.pos=a,e.posMax=h,!0}const LNe=/^([a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*)$/,BNe=/^([a-zA-Z][a-zA-Z0-9+.-]{1,31}):([^<>\x00-\x20]*)$/;function QNe(e,t){let n=e.pos;if(e.src.charCodeAt(n)!==60)return!1;const r=e.pos,i=e.posMax;for(;;){if(++n>=i)return!1;const u=e.src.charCodeAt(n);if(u===60)return!1;if(u===62)break}const a=e.src.slice(r+1,n);if(BNe.test(a)){const u=e.md.normalizeLink(a);if(!e.md.validateLink(u))return!1;if(!t){const l=e.push("link_open","a",1);l.attrs=[["href",u]],l.markup="autolink",l.info="auto";const o=e.push("text","",0);o.content=e.md.normalizeLinkText(a);const s=e.push("link_close","a",-1);s.markup="autolink",s.info="auto"}return e.pos+=a.length+2,!0}if(LNe.test(a)){const u=e.md.normalizeLink("mailto:"+a);if(!e.md.validateLink(u))return!1;if(!t){const l=e.push("link_open","a",1);l.attrs=[["href",u]],l.markup="autolink",l.info="auto";const o=e.push("text","",0);o.content=e.md.normalizeLinkText(a);const s=e.push("link_close","a",-1);s.markup="autolink",s.info="auto"}return e.pos+=a.length+2,!0}return!1}function zNe(e){return/^<a[>\s]/i.test(e)}function qNe(e){return/^<\/a\s*>/i.test(e)}function VNe(e){const t=e|32;return t>=97&&t<=122}function UNe(e,t){if(!e.md.options.html)return!1;const n=e.posMax,r=e.pos;if(e.src.charCodeAt(r)!==60||r+2>=n)return!1;const i=e.src.charCodeAt(r+1);if(i!==33&&i!==63&&i!==47&&!VNe(i))return!1;const a=e.src.slice(r).match(bNe);if(!a)return!1;if(!t){const u=e.push("html_inline","",0);u.content=a[0],zNe(u.content)&&e.linkLevel++,qNe(u.content)&&e.linkLevel--}return e.pos+=a[0].length,!0}const jNe=/^&#((?:x[a-f0-9]{1,6}|[0-9]{1,7}));/i,ZNe=/^&([a-z][a-z0-9]{1,31});/i;function HNe(e,t){const n=e.pos,r=e.posMax;if(e.src.charCodeAt(n)!==38||n+1>=r)return!1;if(e.src.charCodeAt(n+1)===35){const a=e.src.slice(n).match(jNe);if(a){if(!t){const u=a[1][0].toLowerCase()==="x"?parseInt(a[1].slice(1),16):parseInt(a[1],10),l=e.push("text_special","",0);l.content=tL(u)?k$(u):k$(65533),l.markup=a[0],l.info="entity"}return e.pos+=a[0].length,!0}}else{const a=e.src.slice(n).match(ZNe);if(a){const u=dme(a[0]);if(u!==a[0]){if(!t){const l=e.push("text_special","",0);l.content=u,l.markup=a[0],l.info="entity"}return e.pos+=a[0].length,!0}}}return!1}function AX(e){const t={},n=e.length;if(!n)return;let r=0,i=-2;const a=[];for(let u=0;u<n;u++){const l=e[u];if(a.push(0),(e[r].marker!==l.marker||i!==l.token-1)&&(r=u),i=l.token,l.length=l.length||0,!l.close)continue;t.hasOwnProperty(l.marker)||(t[l.marker]=[-1,-1,-1,-1,-1,-1]);const o=t[l.marker][(l.open?3:0)+l.length%3];let s=r-a[r]-1,c=s;for(;s>o;s-=a[s]+1){const d=e[s];if(d.marker===l.marker&&d.open&&d.end<0){let h=!1;if((d.close||l.open)&&(d.length+l.length)%3===0&&(d.length%3!==0||l.length%3!==0)&&(h=!0),!h){const m=s>0&&!e[s-1].open?a[s-1]+1:0;a[u]=u-s+m,a[s]=m,l.open=!1,d.end=u,d.close=!1,c=-1,i=-2;break}}}c!==-1&&(t[l.marker][(l.open?3:0)+(l.length||0)%3]=c)}}function XNe(e){const t=e.tokens_meta,n=e.tokens_meta.length;AX(e.delimiters);for(let r=0;r<n;r++)t[r]&&t[r].delimiters&&AX(t[r].delimiters)}function YNe(e){let t,n,r=0;const i=e.tokens,a=e.tokens.length;for(t=n=0;t<a;t++)i[t].nesting<0&&r--,i[t].level=r,i[t].nesting>0&&r++,i[t].type==="text"&&t+1<a&&i[t+1].type==="text"?i[t+1].content=i[t].content+i[t+1].content:(t!==n&&(i[n]=i[t]),n++);t!==n&&(i.length=n)}const OP=[["text",$Ne],["linkify",wNe],["newline",RNe],["escape",ANe],["backticks",TNe],["strikethrough",bme.tokenize],["emphasis",Ome.tokenize],["link",NNe],["image",INe],["autolink",QNe],["html_inline",UNe],["entity",HNe]],xP=[["balance_pairs",XNe],["strikethrough",bme.postProcess],["emphasis",Ome.postProcess],["fragments_join",YNe]];function PS(){this.ruler=new df;for(let e=0;e<OP.length;e++)this.ruler.push(OP[e][0],OP[e][1]);this.ruler2=new df;for(let e=0;e<xP.length;e++)this.ruler2.push(xP[e][0],xP[e][1])}PS.prototype.skipToken=function(e){const t=e.pos,n=this.ruler.getRules(""),r=n.length,i=e.md.options.maxNesting,a=e.cache;if(typeof a[t]<"u"){e.pos=a[t];return}let u=!1;if(e.level<i){for(let l=0;l<r;l++)if(e.level++,u=n[l](e,!0),e.level--,u){if(t>=e.pos)throw new Error("inline rule didn't increment state.pos");break}}else e.pos=e.posMax;u||e.pos++,a[t]=e.pos};PS.prototype.tokenize=function(e){const t=this.ruler.getRules(""),n=t.length,r=e.posMax,i=e.md.options.maxNesting;for(;e.pos<r;){const a=e.pos;let u=!1;if(e.level<i){for(let l=0;l<n;l++)if(u=t[l](e,!1),u){if(a>=e.pos)throw new Error("inline rule didn't increment state.pos");break}}if(u){if(e.pos>=r)break;continue}e.pending+=e.src[e.pos++]}e.pending&&e.pushPending()};PS.prototype.parse=function(e,t,n,r){const i=new this.State(e,t,n,r);this.tokenize(i);const a=this.ruler2.getRules(""),u=a.length;for(let l=0;l<u;l++)a[l](i)};PS.prototype.State=kS;function GNe(e){const t={};e=e||{},t.src_Any=lme.source,t.src_Cc=ome.source,t.src_Z=cme.source,t.src_P=JI.source,t.src_ZPCc=[t.src_Z,t.src_P,t.src_Cc].join("|"),t.src_ZCc=[t.src_Z,t.src_Cc].join("|");const n="[><｜]";return t.src_pseudo_letter="(?:(?!"+n+"|"+t.src_ZPCc+")"+t.src_Any+")",t.src_ip4="(?:(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)",t.src_auth="(?:(?:(?!"+t.src_ZCc+"|[@/\\[\\]()]).)+@)?",t.src_port="(?::(?:6(?:[0-4]\\d{3}|5(?:[0-4]\\d{2}|5(?:[0-2]\\d|3[0-5])))|[1-5]?\\d{1,4}))?",t.src_host_terminator="(?=$|"+n+"|"+t.src_ZPCc+")(?!"+(e["---"]?"-(?!--)|":"-|")+"_|:\\d|\\.-|\\.(?!$|"+t.src_ZPCc+"))",t.src_path="(?:[/?#](?:(?!"+t.src_ZCc+"|"+n+`|[()[\\]{}.,"'?!\\-;]).|\\[(?:(?!`+t.src_ZCc+"|\\]).)*\\]|\\((?:(?!"+t.src_ZCc+"|[)]).)*\\)|\\{(?:(?!"+t.src_ZCc+'|[}]).)*\\}|\\"(?:(?!'+t.src_ZCc+`|["]).)+\\"|\\'(?:(?!`+t.src_ZCc+"|[']).)+\\'|\\'(?="+t.src_pseudo_letter+"|[-])|\\.{2,}[a-zA-Z0-9%/&]|\\.(?!"+t.src_ZCc+"|[.]|$)|"+(e["---"]?"\\-(?!--(?:[^-]|$))(?:-*)|":"\\-+|")+",(?!"+t.src_ZCc+"|$)|;(?!"+t.src_ZCc+"|$)|\\!+(?!"+t.src_ZCc+"|[!]|$)|\\?(?!"+t.src_ZCc+"|[?]|$))+|\\/)?",t.src_email_name='[\\-;:&=\\+\\$,\\.a-zA-Z0-9_][\\-;:&=\\+\\$,\\"\\.a-zA-Z0-9_]*',t.src_xn="xn--[a-z0-9\\-]{1,59}",t.src_domain_root="(?:"+t.src_xn+"|"+t.src_pseudo_letter+"{1,63})",t.src_domain="(?:"+t.src_xn+"|(?:"+t.src_pseudo_letter+")|(?:"+t.src_pseudo_letter+"(?:-|"+t.src_pseudo_letter+"){0,61}"+t.src_pseudo_letter+"))",t.src_host="(?:(?:(?:(?:"+t.src_domain+")\\.)*"+t.src_domain+"))",t.tpl_host_fuzzy="(?:"+t.src_ip4+"|(?:(?:(?:"+t.src_domain+")\\.)+(?:%TLDS%)))",t.tpl_host_no_ip_fuzzy="(?:(?:(?:"+t.src_domain+")\\.)+(?:%TLDS%))",t.src_host_strict=t.src_host+t.src_host_terminator,t.tpl_host_fuzzy_strict=t.tpl_host_fuzzy+t.src_host_terminator,t.src_host_port_strict=t.src_host+t.src_port+t.src_host_terminator,t.tpl_host_port_fuzzy_strict=t.tpl_host_fuzzy+t.src_port+t.src_host_terminator,t.tpl_host_port_no_ip_fuzzy_strict=t.tpl_host_no_ip_fuzzy+t.src_port+t.src_host_terminator,t.tpl_host_fuzzy_test="localhost|www\\.|\\.\\d{1,3}\\.|(?:\\.(?:%TLDS%)(?:"+t.src_ZPCc+"|>|$))",t.tpl_email_fuzzy="(^|"+n+'|"|\\(|'+t.src_ZCc+")("+t.src_email_name+"@"+t.tpl_host_fuzzy_strict+")",t.tpl_link_fuzzy="(^|(?![.:/\\-_@])(?:[$+<=>^`|｜]|"+t.src_ZPCc+"))((?![$+<=>^`|｜])"+t.tpl_host_port_fuzzy_strict+t.src_path+")",t.tpl_link_no_ip_fuzzy="(^|(?![.:/\\-_@])(?:[$+<=>^`|｜]|"+t.src_ZPCc+"))((?![$+<=>^`|｜])"+t.tpl_host_port_no_ip_fuzzy_strict+t.src_path+")",t}function RM(e){return Array.prototype.slice.call(arguments,1).forEach(function(n){n&&Object.keys(n).forEach(function(r){e[r]=n[r]})}),e}function Rw(e){return Object.prototype.toString.call(e)}function KNe(e){return Rw(e)==="[object String]"}function WNe(e){return Rw(e)==="[object Object]"}function JNe(e){return Rw(e)==="[object RegExp]"}function TX(e){return Rw(e)==="[object Function]"}function eIe(e){return e.replace(/[.?*+^$[\]\\(){}|-]/g,"\\$&")}const xme={fuzzyLink:!0,fuzzyEmail:!0,fuzzyIP:!1};function tIe(e){return Object.keys(e||{}).reduce(function(t,n){return t||xme.hasOwnProperty(n)},!1)}const nIe={"http:":{validate:function(e,t,n){const r=e.slice(t);return n.re.http||(n.re.http=new RegExp("^\\/\\/"+n.re.src_auth+n.re.src_host_port_strict+n.re.src_path,"i")),n.re.http.test(r)?r.match(n.re.http)[0].length:0}},"https:":"http:","ftp:":"http:","//":{validate:function(e,t,n){const r=e.slice(t);return n.re.no_http||(n.re.no_http=new RegExp("^"+n.re.src_auth+"(?:localhost|(?:(?:"+n.re.src_domain+")\\.)+"+n.re.src_domain_root+")"+n.re.src_port+n.re.src_host_terminator+n.re.src_path,"i")),n.re.no_http.test(r)?t>=3&&e[t-3]===":"||t>=3&&e[t-3]==="/"?0:r.match(n.re.no_http)[0].length:0}},"mailto:":{validate:function(e,t,n){const r=e.slice(t);return n.re.mailto||(n.re.mailto=new RegExp("^"+n.re.src_email_name+"@"+n.re.src_host_strict,"i")),n.re.mailto.test(r)?r.match(n.re.mailto)[0].length:0}}},rIe="a[cdefgilmnoqrstuwxz]|b[abdefghijmnorstvwyz]|c[acdfghiklmnoruvwxyz]|d[ejkmoz]|e[cegrstu]|f[ijkmor]|g[abdefghilmnpqrstuwy]|h[kmnrtu]|i[delmnoqrst]|j[emop]|k[eghimnprwyz]|l[abcikrstuvy]|m[acdeghklmnopqrstuvwxyz]|n[acefgilopruz]|om|p[aefghklmnrstwy]|qa|r[eosuw]|s[abcdeghijklmnortuvxyz]|t[cdfghjklmnortvwz]|u[agksyz]|v[aceginu]|w[fs]|y[et]|z[amw]",iIe="biz|com|edu|gov|net|org|pro|web|xxx|aero|asia|coop|info|museum|name|shop|рф".split("|");function aIe(e){e.__index__=-1,e.__text_cache__=""}function uIe(e){return function(t,n){const r=t.slice(n);return e.test(r)?r.match(e)[0].length:0}}function kX(){return function(e,t){t.normalize(e)}}function P$(e){const t=e.re=GNe(e.__opts__),n=e.__tlds__.slice();e.onCompile(),e.__tlds_replaced__||n.push(rIe),n.push(t.src_xn),t.src_tlds=n.join("|");function r(l){return l.replace("%TLDS%",t.src_tlds)}t.email_fuzzy=RegExp(r(t.tpl_email_fuzzy),"i"),t.link_fuzzy=RegExp(r(t.tpl_link_fuzzy),"i"),t.link_no_ip_fuzzy=RegExp(r(t.tpl_link_no_ip_fuzzy),"i"),t.host_fuzzy_test=RegExp(r(t.tpl_host_fuzzy_test),"i");const i=[];e.__compiled__={};function a(l,o){throw new Error('(LinkifyIt) Invalid schema "'+l+'": '+o)}Object.keys(e.__schemas__).forEach(function(l){const o=e.__schemas__[l];if(o===null)return;const s={validate:null,link:null};if(e.__compiled__[l]=s,WNe(o)){JNe(o.validate)?s.validate=uIe(o.validate):TX(o.validate)?s.validate=o.validate:a(l,o),TX(o.normalize)?s.normalize=o.normalize:o.normalize?a(l,o):s.normalize=kX();return}if(KNe(o)){i.push(l);return}a(l,o)}),i.forEach(function(l){e.__compiled__[e.__schemas__[l]]&&(e.__compiled__[l].validate=e.__compiled__[e.__schemas__[l]].validate,e.__compiled__[l].normalize=e.__compiled__[e.__schemas__[l]].normalize)}),e.__compiled__[""]={validate:null,normalize:kX()};const u=Object.keys(e.__compiled__).filter(function(l){return l.length>0&&e.__compiled__[l]}).map(eIe).join("|");e.re.schema_test=RegExp("(^|(?!_)(?:[><｜]|"+t.src_ZPCc+"))("+u+")","i"),e.re.schema_search=RegExp("(^|(?!_)(?:[><｜]|"+t.src_ZPCc+"))("+u+")","ig"),e.re.schema_at_start=RegExp("^"+e.re.schema_search.source,"i"),e.re.pretest=RegExp("("+e.re.schema_test.source+")|("+e.re.host_fuzzy_test.source+")|@","i"),aIe(e)}function lIe(e,t){const n=e.__index__,r=e.__last_index__,i=e.__text_cache__.slice(n,r);this.schema=e.__schema__.toLowerCase(),this.index=n+t,this.lastIndex=r+t,this.raw=i,this.text=i,this.url=i}function AM(e,t){const n=new lIe(e,t);return e.__compiled__[n.schema].normalize(n,e),n}function Kf(e,t){if(!(this instanceof Kf))return new Kf(e,t);t||tIe(e)&&(t=e,e={}),this.__opts__=RM({},xme,t),this.__index__=-1,this.__last_index__=-1,this.__schema__="",this.__text_cache__="",this.__schemas__=RM({},nIe,e),this.__compiled__={},this.__tlds__=iIe,this.__tlds_replaced__=!1,this.re={},P$(this)}Kf.prototype.add=function(t,n){return this.__schemas__[t]=n,P$(this),this};Kf.prototype.set=function(t){return this.__opts__=RM(this.__opts__,t),this};Kf.prototype.test=function(t){if(this.__text_cache__=t,this.__index__=-1,!t.length)return!1;let n,r,i,a,u,l,o,s,c;if(this.re.schema_test.test(t)){for(o=this.re.schema_search,o.lastIndex=0;(n=o.exec(t))!==null;)if(a=this.testSchemaAt(t,n[2],o.lastIndex),a){this.__schema__=n[2],this.__index__=n.index+n[1].length,this.__last_index__=n.index+n[0].length+a;break}}return this.__opts__.fuzzyLink&&this.__compiled__["http:"]&&(s=t.search(this.re.host_fuzzy_test),s>=0&&(this.__index__<0||s<this.__index__)&&(r=t.match(this.__opts__.fuzzyIP?this.re.link_fuzzy:this.re.link_no_ip_fuzzy))!==null&&(u=r.index+r[1].length,(this.__index__<0||u<this.__index__)&&(this.__schema__="",this.__index__=u,this.__last_index__=r.index+r[0].length))),this.__opts__.fuzzyEmail&&this.__compiled__["mailto:"]&&(c=t.indexOf("@"),c>=0&&(i=t.match(this.re.email_fuzzy))!==null&&(u=i.index+i[1].length,l=i.index+i[0].length,(this.__index__<0||u<this.__index__||u===this.__index__&&l>this.__last_index__)&&(this.__schema__="mailto:",this.__index__=u,this.__last_index__=l))),this.__index__>=0};Kf.prototype.pretest=function(t){return this.re.pretest.test(t)};Kf.prototype.testSchemaAt=function(t,n,r){return this.__compiled__[n.toLowerCase()]?this.__compiled__[n.toLowerCase()].validate(t,r,this):0};Kf.prototype.match=function(t){const n=[];let r=0;this.__index__>=0&&this.__text_cache__===t&&(n.push(AM(this,r)),r=this.__last_index__);let i=r?t.slice(r):t;for(;this.test(i);)n.push(AM(this,r)),i=i.slice(this.__last_index__),r+=this.__last_index__;return n.length?n:null};Kf.prototype.matchAtStart=function(t){if(this.__text_cache__=t,this.__index__=-1,!t.length)return null;const n=this.re.schema_at_start.exec(t);if(!n)return null;const r=this.testSchemaAt(t,n[2],n[0].length);return r?(this.__schema__=n[2],this.__index__=n.index+n[1].length,this.__last_index__=n.index+n[0].length+r,AM(this,0)):null};Kf.prototype.tlds=function(t,n){return t=Array.isArray(t)?t:[t],n?(this.__tlds__=this.__tlds__.concat(t).sort().filter(function(r,i,a){return r!==a[i-1]}).reverse(),P$(this),this):(this.__tlds__=t.slice(),this.__tlds_replaced__=!0,P$(this),this)};Kf.prototype.normalize=function(t){t.schema||(t.url="http://"+t.url),t.schema==="mailto:"&&!/^mailto:/i.test(t.url)&&(t.url="mailto:"+t.url)};Kf.prototype.onCompile=function(){};const g3=2147483647,x0=36,iL=1,Vx=26,oIe=38,sIe=700,Sme=72,_me=128,Eme="-",cIe=/^xn--/,fIe=/[^\0-\x7F]/,dIe=/[\x2E\u3002\uFF0E\uFF61]/g,hIe={overflow:"Overflow: input needs wider integers to process","not-basic":"Illegal input >= 0x80 (not a basic code point)","invalid-input":"Invalid input"},SP=x0-iL,S0=Math.floor,_P=String.fromCharCode;function Ev(e){throw new RangeError(hIe[e])}function pIe(e,t){const n=[];let r=e.length;for(;r--;)n[r]=t(e[r]);return n}function Dme(e,t){const n=e.split("@");let r="";n.length>1&&(r=n[0]+"@",e=n[1]),e=e.replace(dIe,".");const i=e.split("."),a=pIe(i,t).join(".");return r+a}function $me(e){const t=[];let n=0;const r=e.length;for(;n<r;){const i=e.charCodeAt(n++);if(i>=55296&&i<=56319&&n<r){const a=e.charCodeAt(n++);(a&64512)==56320?t.push(((i&1023)<<10)+(a&1023)+65536):(t.push(i),n--)}else t.push(i)}return t}const mIe=e=>String.fromCodePoint(...e),gIe=function(e){return e>=48&&e<58?26+(e-48):e>=65&&e<91?e-65:e>=97&&e<123?e-97:x0},PX=function(e,t){return e+22+75*(e<26)-((t!=0)<<5)},Cme=function(e,t,n){let r=0;for(e=n?S0(e/sIe):e>>1,e+=S0(e/t);e>SP*Vx>>1;r+=x0)e=S0(e/SP);return S0(r+(SP+1)*e/(e+oIe))},wme=function(e){const t=[],n=e.length;let r=0,i=_me,a=Sme,u=e.lastIndexOf(Eme);u<0&&(u=0);for(let l=0;l<u;++l)e.charCodeAt(l)>=128&&Ev("not-basic"),t.push(e.charCodeAt(l));for(let l=u>0?u+1:0;l<n;){const o=r;for(let c=1,d=x0;;d+=x0){l>=n&&Ev("invalid-input");const h=gIe(e.charCodeAt(l++));h>=x0&&Ev("invalid-input"),h>S0((g3-r)/c)&&Ev("overflow"),r+=h*c;const m=d<=a?iL:d>=a+Vx?Vx:d-a;if(h<m)break;const y=x0-m;c>S0(g3/y)&&Ev("overflow"),c*=y}const s=t.length+1;a=Cme(r-o,s,o==0),S0(r/s)>g3-i&&Ev("overflow"),i+=S0(r/s),r%=s,t.splice(r++,0,i)}return String.fromCodePoint(...t)},Rme=function(e){const t=[];e=$me(e);const n=e.length;let r=_me,i=0,a=Sme;for(const o of e)o<128&&t.push(_P(o));const u=t.length;let l=u;for(u&&t.push(Eme);l<n;){let o=g3;for(const c of e)c>=r&&c<o&&(o=c);const s=l+1;o-r>S0((g3-i)/s)&&Ev("overflow"),i+=(o-r)*s,r=o;for(const c of e)if(c<r&&++i>g3&&Ev("overflow"),c===r){let d=i;for(let h=x0;;h+=x0){const m=h<=a?iL:h>=a+Vx?Vx:h-a;if(d<m)break;const y=d-m,v=x0-m;t.push(_P(PX(m+y%v,0))),d=S0(y/v)}t.push(_P(PX(d,0))),a=Cme(i,s,l===u),i=0,++l}++i,++r}return t.join("")},vIe=function(e){return Dme(e,function(t){return cIe.test(t)?wme(t.slice(4).toLowerCase()):t})},yIe=function(e){return Dme(e,function(t){return fIe.test(t)?"xn--"+Rme(t):t})},Ame={version:"2.3.1",ucs2:{decode:$me,encode:mIe},decode:wme,encode:Rme,toASCII:yIe,toUnicode:vIe},bIe={options:{html:!1,xhtmlOut:!1,breaks:!1,langPrefix:"language-",linkify:!1,typographer:!1,quotes:"“”‘’",highlight:null,maxNesting:100},components:{core:{},block:{},inline:{}}},OIe={options:{html:!1,xhtmlOut:!1,breaks:!1,langPrefix:"language-",linkify:!1,typographer:!1,quotes:"“”‘’",highlight:null,maxNesting:20},components:{core:{rules:["normalize","block","inline","text_join"]},block:{rules:["paragraph"]},inline:{rules:["text"],rules2:["balance_pairs","fragments_join"]}}},xIe={options:{html:!0,xhtmlOut:!0,breaks:!1,langPrefix:"language-",linkify:!1,typographer:!1,quotes:"“”‘’",highlight:null,maxNesting:20},components:{core:{rules:["normalize","block","inline","text_join"]},block:{rules:["blockquote","code","fence","heading","hr","html_block","lheading","list","reference","paragraph"]},inline:{rules:["autolink","backticks","emphasis","entity","escape","html_inline","image","link","newline","text"],rules2:["balance_pairs","emphasis","fragments_join"]}}},SIe={default:bIe,zero:OIe,commonmark:xIe},_Ie=/^(vbscript|javascript|file|data):/,EIe=/^data:image\/(gif|png|jpeg|webp);/;function DIe(e){const t=e.trim().toLowerCase();return _Ie.test(t)?EIe.test(t):!0}const Tme=["http:","https:","mailto:"];function $Ie(e){const t=WI(e,!0);if(t.hostname&&(!t.protocol||Tme.indexOf(t.protocol)>=0))try{t.hostname=Ame.toASCII(t.hostname)}catch{}return TS(KI(t))}function CIe(e){const t=WI(e,!0);if(t.hostname&&(!t.protocol||Tme.indexOf(t.protocol)>=0))try{t.hostname=Ame.toUnicode(t.hostname)}catch{}return F3(KI(t),F3.defaultChars+"%")}function cp(e,t){if(!(this instanceof cp))return new cp(e,t);t||eL(e)||(t=e||{},e="default"),this.inline=new PS,this.block=new ww,this.core=new nL,this.renderer=new u4,this.linkify=new Kf,this.validateLink=DIe,this.normalizeLink=$Ie,this.normalizeLinkText=CIe,this.utils=RFe,this.helpers=$w({},PFe),this.options={},this.configure(e),t&&this.set(t)}cp.prototype.set=function(e){return $w(this.options,e),this};cp.prototype.configure=function(e){const t=this;if(eL(e)){const n=e;if(e=SIe[n],!e)throw new Error('Wrong `markdown-it` preset "'+n+'", check name')}if(!e)throw new Error("Wrong `markdown-it` preset, can't be empty");return e.options&&t.set(e.options),e.components&&Object.keys(e.components).forEach(function(n){e.components[n].rules&&t[n].ruler.enableOnly(e.components[n].rules),e.components[n].rules2&&t[n].ruler2.enableOnly(e.components[n].rules2)}),this};cp.prototype.enable=function(e,t){let n=[];Array.isArray(e)||(e=[e]),["core","block","inline"].forEach(function(i){n=n.concat(this[i].ruler.enable(e,!0))},this),n=n.concat(this.inline.ruler2.enable(e,!0));const r=e.filter(function(i){return n.indexOf(i)<0});if(r.length&&!t)throw new Error("MarkdownIt. Failed to enable unknown rule(s): "+r);return this};cp.prototype.disable=function(e,t){let n=[];Array.isArray(e)||(e=[e]),["core","block","inline"].forEach(function(i){n=n.concat(this[i].ruler.disable(e,!0))},this),n=n.concat(this.inline.ruler2.disable(e,!0));const r=e.filter(function(i){return n.indexOf(i)<0});if(r.length&&!t)throw new Error("MarkdownIt. Failed to disable unknown rule(s): "+r);return this};cp.prototype.use=function(e){const t=[this].concat(Array.prototype.slice.call(arguments,1));return e.apply(e,t),this};cp.prototype.parse=function(e,t){if(typeof e!="string")throw new Error("Input data should be a String");const n=new this.core.State(e,this,t);return this.core.process(n),n.tokens};cp.prototype.render=function(e,t){return t=t||{},this.renderer.render(this.parse(e,t),this.options,t)};cp.prototype.parseInline=function(e,t){const n=new this.core.State(e,this,t);return n.inlineMode=!0,this.core.process(n),n.tokens};cp.prototype.renderInline=function(e,t){return t=t||{},this.renderer.render(this.parseInline(e,t),this.options,t)};const wIe=["area","base","basefont","bgsound","br","col","command","embed","frame","hr","image","img","input","keygen","link","meta","param","source","track","wbr"];let MS=class{constructor(t,n,r){this.normal=n,this.property=t,r&&(this.space=r)}};MS.prototype.normal={};MS.prototype.property={};MS.prototype.space=void 0;function kme(e,t){const n={},r={};for(const i of e)Object.assign(n,i.property),Object.assign(r,i.normal);return new MS(n,r,t)}function TM(e){return e.toLowerCase()}class pf{constructor(t,n){this.attribute=n,this.property=t}}pf.prototype.attribute="";pf.prototype.booleanish=!1;pf.prototype.boolean=!1;pf.prototype.commaOrSpaceSeparated=!1;pf.prototype.commaSeparated=!1;pf.prototype.defined=!1;pf.prototype.mustUseProperty=!1;pf.prototype.number=!1;pf.prototype.overloadedBoolean=!1;pf.prototype.property="";pf.prototype.spaceSeparated=!1;pf.prototype.space=void 0;let RIe=0;const vr=T1(),ml=T1(),kM=T1(),Ft=T1(),Oa=T1(),v3=T1(),Lf=T1();function T1(){return 2**++RIe}const PM=Object.freeze(Object.defineProperty({__proto__:null,boolean:vr,booleanish:ml,commaOrSpaceSeparated:Lf,commaSeparated:v3,number:Ft,overloadedBoolean:kM,spaceSeparated:Oa},Symbol.toStringTag,{value:"Module"})),EP=Object.keys(PM);class aL extends pf{constructor(t,n,r,i){let a=-1;if(super(t,n),MX(this,"space",i),typeof r=="number")for(;++a<EP.length;){const u=EP[a];MX(this,EP[a],(r&PM[u])===PM[u])}}}aL.prototype.defined=!0;function MX(e,t,n){n&&(e[t]=n)}function l4(e){const t={},n={};for(const[r,i]of Object.entries(e.properties)){const a=new aL(r,e.transform(e.attributes||{},r),i,e.space);e.mustUseProperty&&e.mustUseProperty.includes(r)&&(a.mustUseProperty=!0),t[r]=a,n[TM(r)]=r,n[TM(a.attribute)]=r}return new MS(t,n,e.space)}const Pme=l4({properties:{ariaActiveDescendant:null,ariaAtomic:ml,ariaAutoComplete:null,ariaBusy:ml,ariaChecked:ml,ariaColCount:Ft,ariaColIndex:Ft,ariaColSpan:Ft,ariaControls:Oa,ariaCurrent:null,ariaDescribedBy:Oa,ariaDetails:null,ariaDisabled:ml,ariaDropEffect:Oa,ariaErrorMessage:null,ariaExpanded:ml,ariaFlowTo:Oa,ariaGrabbed:ml,ariaHasPopup:null,ariaHidden:ml,ariaInvalid:null,ariaKeyShortcuts:null,ariaLabel:null,ariaLabelledBy:Oa,ariaLevel:Ft,ariaLive:null,ariaModal:ml,ariaMultiLine:ml,ariaMultiSelectable:ml,ariaOrientation:null,ariaOwns:Oa,ariaPlaceholder:null,ariaPosInSet:Ft,ariaPressed:ml,ariaReadOnly:ml,ariaRelevant:null,ariaRequired:ml,ariaRoleDescription:Oa,ariaRowCount:Ft,ariaRowIndex:Ft,ariaRowSpan:Ft,ariaSelected:ml,ariaSetSize:Ft,ariaSort:null,ariaValueMax:Ft,ariaValueMin:Ft,ariaValueNow:Ft,ariaValueText:null,role:null},transform(e,t){return t==="role"?t:"aria-"+t.slice(4).toLowerCase()}});function Mme(e,t){return t in e?e[t]:t}function Fme(e,t){return Mme(e,t.toLowerCase())}const AIe=l4({attributes:{acceptcharset:"accept-charset",classname:"class",htmlfor:"for",httpequiv:"http-equiv"},mustUseProperty:["checked","multiple","muted","selected"],properties:{abbr:null,accept:v3,acceptCharset:Oa,accessKey:Oa,action:null,allow:null,allowFullScreen:vr,allowPaymentRequest:vr,allowUserMedia:vr,alt:null,as:null,async:vr,autoCapitalize:null,autoComplete:Oa,autoFocus:vr,autoPlay:vr,blocking:Oa,capture:null,charSet:null,checked:vr,cite:null,className:Oa,cols:Ft,colSpan:null,content:null,contentEditable:ml,controls:vr,controlsList:Oa,coords:Ft|v3,crossOrigin:null,data:null,dateTime:null,decoding:null,default:vr,defer:vr,dir:null,dirName:null,disabled:vr,download:kM,draggable:ml,encType:null,enterKeyHint:null,fetchPriority:null,form:null,formAction:null,formEncType:null,formMethod:null,formNoValidate:vr,formTarget:null,headers:Oa,height:Ft,hidden:kM,high:Ft,href:null,hrefLang:null,htmlFor:Oa,httpEquiv:Oa,id:null,imageSizes:null,imageSrcSet:null,inert:vr,inputMode:null,integrity:null,is:null,isMap:vr,itemId:null,itemProp:Oa,itemRef:Oa,itemScope:vr,itemType:Oa,kind:null,label:null,lang:null,language:null,list:null,loading:null,loop:vr,low:Ft,manifest:null,max:null,maxLength:Ft,media:null,method:null,min:null,minLength:Ft,multiple:vr,muted:vr,name:null,nonce:null,noModule:vr,noValidate:vr,onAbort:null,onAfterPrint:null,onAuxClick:null,onBeforeMatch:null,onBeforePrint:null,onBeforeToggle:null,onBeforeUnload:null,onBlur:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onContextLost:null,onContextMenu:null,onContextRestored:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnded:null,onError:null,onFocus:null,onFormData:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLanguageChange:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadEnd:null,onLoadStart:null,onMessage:null,onMessageError:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRejectionHandled:null,onReset:null,onResize:null,onScroll:null,onScrollEnd:null,onSecurityPolicyViolation:null,onSeeked:null,onSeeking:null,onSelect:null,onSlotChange:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnhandledRejection:null,onUnload:null,onVolumeChange:null,onWaiting:null,onWheel:null,open:vr,optimum:Ft,pattern:null,ping:Oa,placeholder:null,playsInline:vr,popover:null,popoverTarget:null,popoverTargetAction:null,poster:null,preload:null,readOnly:vr,referrerPolicy:null,rel:Oa,required:vr,reversed:vr,rows:Ft,rowSpan:Ft,sandbox:Oa,scope:null,scoped:vr,seamless:vr,selected:vr,shadowRootClonable:vr,shadowRootDelegatesFocus:vr,shadowRootMode:null,shape:null,size:Ft,sizes:null,slot:null,span:Ft,spellCheck:ml,src:null,srcDoc:null,srcLang:null,srcSet:null,start:Ft,step:null,style:null,tabIndex:Ft,target:null,title:null,translate:null,type:null,typeMustMatch:vr,useMap:null,value:ml,width:Ft,wrap:null,writingSuggestions:null,align:null,aLink:null,archive:Oa,axis:null,background:null,bgColor:null,border:Ft,borderColor:null,bottomMargin:Ft,cellPadding:null,cellSpacing:null,char:null,charOff:null,classId:null,clear:null,code:null,codeBase:null,codeType:null,color:null,compact:vr,declare:vr,event:null,face:null,frame:null,frameBorder:null,hSpace:Ft,leftMargin:Ft,link:null,longDesc:null,lowSrc:null,marginHeight:Ft,marginWidth:Ft,noResize:vr,noHref:vr,noShade:vr,noWrap:vr,object:null,profile:null,prompt:null,rev:null,rightMargin:Ft,rules:null,scheme:null,scrolling:ml,standby:null,summary:null,text:null,topMargin:Ft,valueType:null,version:null,vAlign:null,vLink:null,vSpace:Ft,allowTransparency:null,autoCorrect:null,autoSave:null,disablePictureInPicture:vr,disableRemotePlayback:vr,prefix:null,property:null,results:Ft,security:null,unselectable:null},space:"html",transform:Fme}),TIe=l4({attributes:{accentHeight:"accent-height",alignmentBaseline:"alignment-baseline",arabicForm:"arabic-form",baselineShift:"baseline-shift",capHeight:"cap-height",className:"class",clipPath:"clip-path",clipRule:"clip-rule",colorInterpolation:"color-interpolation",colorInterpolationFilters:"color-interpolation-filters",colorProfile:"color-profile",colorRendering:"color-rendering",crossOrigin:"crossorigin",dataType:"datatype",dominantBaseline:"dominant-baseline",enableBackground:"enable-background",fillOpacity:"fill-opacity",fillRule:"fill-rule",floodColor:"flood-color",floodOpacity:"flood-opacity",fontFamily:"font-family",fontSize:"font-size",fontSizeAdjust:"font-size-adjust",fontStretch:"font-stretch",fontStyle:"font-style",fontVariant:"font-variant",fontWeight:"font-weight",glyphName:"glyph-name",glyphOrientationHorizontal:"glyph-orientation-horizontal",glyphOrientationVertical:"glyph-orientation-vertical",hrefLang:"hreflang",horizAdvX:"horiz-adv-x",horizOriginX:"horiz-origin-x",horizOriginY:"horiz-origin-y",imageRendering:"image-rendering",letterSpacing:"letter-spacing",lightingColor:"lighting-color",markerEnd:"marker-end",markerMid:"marker-mid",markerStart:"marker-start",navDown:"nav-down",navDownLeft:"nav-down-left",navDownRight:"nav-down-right",navLeft:"nav-left",navNext:"nav-next",navPrev:"nav-prev",navRight:"nav-right",navUp:"nav-up",navUpLeft:"nav-up-left",navUpRight:"nav-up-right",onAbort:"onabort",onActivate:"onactivate",onAfterPrint:"onafterprint",onBeforePrint:"onbeforeprint",onBegin:"onbegin",onCancel:"oncancel",onCanPlay:"oncanplay",onCanPlayThrough:"oncanplaythrough",onChange:"onchange",onClick:"onclick",onClose:"onclose",onCopy:"oncopy",onCueChange:"oncuechange",onCut:"oncut",onDblClick:"ondblclick",onDrag:"ondrag",onDragEnd:"ondragend",onDragEnter:"ondragenter",onDragExit:"ondragexit",onDragLeave:"ondragleave",onDragOver:"ondragover",onDragStart:"ondragstart",onDrop:"ondrop",onDurationChange:"ondurationchange",onEmptied:"onemptied",onEnd:"onend",onEnded:"onended",onError:"onerror",onFocus:"onfocus",onFocusIn:"onfocusin",onFocusOut:"onfocusout",onHashChange:"onhashchange",onInput:"oninput",onInvalid:"oninvalid",onKeyDown:"onkeydown",onKeyPress:"onkeypress",onKeyUp:"onkeyup",onLoad:"onload",onLoadedData:"onloadeddata",onLoadedMetadata:"onloadedmetadata",onLoadStart:"onloadstart",onMessage:"onmessage",onMouseDown:"onmousedown",onMouseEnter:"onmouseenter",onMouseLeave:"onmouseleave",onMouseMove:"onmousemove",onMouseOut:"onmouseout",onMouseOver:"onmouseover",onMouseUp:"onmouseup",onMouseWheel:"onmousewheel",onOffline:"onoffline",onOnline:"ononline",onPageHide:"onpagehide",onPageShow:"onpageshow",onPaste:"onpaste",onPause:"onpause",onPlay:"onplay",onPlaying:"onplaying",onPopState:"onpopstate",onProgress:"onprogress",onRateChange:"onratechange",onRepeat:"onrepeat",onReset:"onreset",onResize:"onresize",onScroll:"onscroll",onSeeked:"onseeked",onSeeking:"onseeking",onSelect:"onselect",onShow:"onshow",onStalled:"onstalled",onStorage:"onstorage",onSubmit:"onsubmit",onSuspend:"onsuspend",onTimeUpdate:"ontimeupdate",onToggle:"ontoggle",onUnload:"onunload",onVolumeChange:"onvolumechange",onWaiting:"onwaiting",onZoom:"onzoom",overlinePosition:"overline-position",overlineThickness:"overline-thickness",paintOrder:"paint-order",panose1:"panose-1",pointerEvents:"pointer-events",referrerPolicy:"referrerpolicy",renderingIntent:"rendering-intent",shapeRendering:"shape-rendering",stopColor:"stop-color",stopOpacity:"stop-opacity",strikethroughPosition:"strikethrough-position",strikethroughThickness:"strikethrough-thickness",strokeDashArray:"stroke-dasharray",strokeDashOffset:"stroke-dashoffset",strokeLineCap:"stroke-linecap",strokeLineJoin:"stroke-linejoin",strokeMiterLimit:"stroke-miterlimit",strokeOpacity:"stroke-opacity",strokeWidth:"stroke-width",tabIndex:"tabindex",textAnchor:"text-anchor",textDecoration:"text-decoration",textRendering:"text-rendering",transformOrigin:"transform-origin",typeOf:"typeof",underlinePosition:"underline-position",underlineThickness:"underline-thickness",unicodeBidi:"unicode-bidi",unicodeRange:"unicode-range",unitsPerEm:"units-per-em",vAlphabetic:"v-alphabetic",vHanging:"v-hanging",vIdeographic:"v-ideographic",vMathematical:"v-mathematical",vectorEffect:"vector-effect",vertAdvY:"vert-adv-y",vertOriginX:"vert-origin-x",vertOriginY:"vert-origin-y",wordSpacing:"word-spacing",writingMode:"writing-mode",xHeight:"x-height",playbackOrder:"playbackorder",timelineBegin:"timelinebegin"},properties:{about:Lf,accentHeight:Ft,accumulate:null,additive:null,alignmentBaseline:null,alphabetic:Ft,amplitude:Ft,arabicForm:null,ascent:Ft,attributeName:null,attributeType:null,azimuth:Ft,bandwidth:null,baselineShift:null,baseFrequency:null,baseProfile:null,bbox:null,begin:null,bias:Ft,by:null,calcMode:null,capHeight:Ft,className:Oa,clip:null,clipPath:null,clipPathUnits:null,clipRule:null,color:null,colorInterpolation:null,colorInterpolationFilters:null,colorProfile:null,colorRendering:null,content:null,contentScriptType:null,contentStyleType:null,crossOrigin:null,cursor:null,cx:null,cy:null,d:null,dataType:null,defaultAction:null,descent:Ft,diffuseConstant:Ft,direction:null,display:null,dur:null,divisor:Ft,dominantBaseline:null,download:vr,dx:null,dy:null,edgeMode:null,editable:null,elevation:Ft,enableBackground:null,end:null,event:null,exponent:Ft,externalResourcesRequired:null,fill:null,fillOpacity:Ft,fillRule:null,filter:null,filterRes:null,filterUnits:null,floodColor:null,floodOpacity:null,focusable:null,focusHighlight:null,fontFamily:null,fontSize:null,fontSizeAdjust:null,fontStretch:null,fontStyle:null,fontVariant:null,fontWeight:null,format:null,fr:null,from:null,fx:null,fy:null,g1:v3,g2:v3,glyphName:v3,glyphOrientationHorizontal:null,glyphOrientationVertical:null,glyphRef:null,gradientTransform:null,gradientUnits:null,handler:null,hanging:Ft,hatchContentUnits:null,hatchUnits:null,height:null,href:null,hrefLang:null,horizAdvX:Ft,horizOriginX:Ft,horizOriginY:Ft,id:null,ideographic:Ft,imageRendering:null,initialVisibility:null,in:null,in2:null,intercept:Ft,k:Ft,k1:Ft,k2:Ft,k3:Ft,k4:Ft,kernelMatrix:Lf,kernelUnitLength:null,keyPoints:null,keySplines:null,keyTimes:null,kerning:null,lang:null,lengthAdjust:null,letterSpacing:null,lightingColor:null,limitingConeAngle:Ft,local:null,markerEnd:null,markerMid:null,markerStart:null,markerHeight:null,markerUnits:null,markerWidth:null,mask:null,maskContentUnits:null,maskUnits:null,mathematical:null,max:null,media:null,mediaCharacterEncoding:null,mediaContentEncodings:null,mediaSize:Ft,mediaTime:null,method:null,min:null,mode:null,name:null,navDown:null,navDownLeft:null,navDownRight:null,navLeft:null,navNext:null,navPrev:null,navRight:null,navUp:null,navUpLeft:null,navUpRight:null,numOctaves:null,observer:null,offset:null,onAbort:null,onActivate:null,onAfterPrint:null,onBeforePrint:null,onBegin:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnd:null,onEnded:null,onError:null,onFocus:null,onFocusIn:null,onFocusOut:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadStart:null,onMessage:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onMouseWheel:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRepeat:null,onReset:null,onResize:null,onScroll:null,onSeeked:null,onSeeking:null,onSelect:null,onShow:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnload:null,onVolumeChange:null,onWaiting:null,onZoom:null,opacity:null,operator:null,order:null,orient:null,orientation:null,origin:null,overflow:null,overlay:null,overlinePosition:Ft,overlineThickness:Ft,paintOrder:null,panose1:null,path:null,pathLength:Ft,patternContentUnits:null,patternTransform:null,patternUnits:null,phase:null,ping:Oa,pitch:null,playbackOrder:null,pointerEvents:null,points:null,pointsAtX:Ft,pointsAtY:Ft,pointsAtZ:Ft,preserveAlpha:null,preserveAspectRatio:null,primitiveUnits:null,propagate:null,property:Lf,r:null,radius:null,referrerPolicy:null,refX:null,refY:null,rel:Lf,rev:Lf,renderingIntent:null,repeatCount:null,repeatDur:null,requiredExtensions:Lf,requiredFeatures:Lf,requiredFonts:Lf,requiredFormats:Lf,resource:null,restart:null,result:null,rotate:null,rx:null,ry:null,scale:null,seed:null,shapeRendering:null,side:null,slope:null,snapshotTime:null,specularConstant:Ft,specularExponent:Ft,spreadMethod:null,spacing:null,startOffset:null,stdDeviation:null,stemh:null,stemv:null,stitchTiles:null,stopColor:null,stopOpacity:null,strikethroughPosition:Ft,strikethroughThickness:Ft,string:null,stroke:null,strokeDashArray:Lf,strokeDashOffset:null,strokeLineCap:null,strokeLineJoin:null,strokeMiterLimit:Ft,strokeOpacity:Ft,strokeWidth:null,style:null,surfaceScale:Ft,syncBehavior:null,syncBehaviorDefault:null,syncMaster:null,syncTolerance:null,syncToleranceDefault:null,systemLanguage:Lf,tabIndex:Ft,tableValues:null,target:null,targetX:Ft,targetY:Ft,textAnchor:null,textDecoration:null,textRendering:null,textLength:null,timelineBegin:null,title:null,transformBehavior:null,type:null,typeOf:Lf,to:null,transform:null,transformOrigin:null,u1:null,u2:null,underlinePosition:Ft,underlineThickness:Ft,unicode:null,unicodeBidi:null,unicodeRange:null,unitsPerEm:Ft,values:null,vAlphabetic:Ft,vMathematical:Ft,vectorEffect:null,vHanging:Ft,vIdeographic:Ft,version:null,vertAdvY:Ft,vertOriginX:Ft,vertOriginY:Ft,viewBox:null,viewTarget:null,visibility:null,width:null,widths:null,wordSpacing:null,writingMode:null,x:null,x1:null,x2:null,xChannelSelector:null,xHeight:Ft,y:null,y1:null,y2:null,yChannelSelector:null,z:null,zoomAndPan:null},space:"svg",transform:Mme}),Nme=l4({properties:{xLinkActuate:null,xLinkArcRole:null,xLinkHref:null,xLinkRole:null,xLinkShow:null,xLinkTitle:null,xLinkType:null},space:"xlink",transform(e,t){return"xlink:"+t.slice(5).toLowerCase()}}),Ime=l4({attributes:{xmlnsxlink:"xmlns:xlink"},properties:{xmlnsXLink:null,xmlns:null},space:"xmlns",transform:Fme}),Lme=l4({properties:{xmlBase:null,xmlLang:null,xmlSpace:null},space:"xml",transform(e,t){return"xml:"+t.slice(3).toLowerCase()}}),kIe={classId:"classID",dataType:"datatype",itemId:"itemID",strokeDashArray:"strokeDasharray",strokeDashOffset:"strokeDashoffset",strokeLineCap:"strokeLinecap",strokeLineJoin:"strokeLinejoin",strokeMiterLimit:"strokeMiterlimit",typeOf:"typeof",xLinkActuate:"xlinkActuate",xLinkArcRole:"xlinkArcrole",xLinkHref:"xlinkHref",xLinkRole:"xlinkRole",xLinkShow:"xlinkShow",xLinkTitle:"xlinkTitle",xLinkType:"xlinkType",xmlnsXLink:"xmlnsXlink"},PIe=/[A-Z]/g,FX=/-[a-z]/g,MIe=/^data[-\w.:]+$/i;function Bme(e,t){const n=TM(t);let r=t,i=pf;if(n in e.normal)return e.property[e.normal[n]];if(n.length>4&&n.slice(0,4)==="data"&&MIe.test(t)){if(t.charAt(4)==="-"){const a=t.slice(5).replace(FX,NIe);r="data"+a.charAt(0).toUpperCase()+a.slice(1)}else{const a=t.slice(4);if(!FX.test(a)){let u=a.replace(PIe,FIe);u.charAt(0)!=="-"&&(u="-"+u),t="data"+u}}i=aL}return new i(r,t)}function FIe(e){return"-"+e.toLowerCase()}function NIe(e){return e.charAt(1).toUpperCase()}const Qme=kme([Pme,AIe,Nme,Ime,Lme],"html"),FS=kme([Pme,TIe,Nme,Ime,Lme],"svg"),NX={}.hasOwnProperty;function IIe(e,t){const n=t||{};function r(i,...a){let u=r.invalid;const l=r.handlers;if(i&&NX.call(i,e)){const o=String(i[e]);u=NX.call(l,o)?l[o]:r.unknown}if(u)return u.call(this,i,...a)}return r.handlers=n.handlers||{},r.invalid=n.invalid,r.unknown=n.unknown,r}const LIe=/["&'<>`]/g,BIe=/[\uD800-\uDBFF][\uDC00-\uDFFF]/g,QIe=/[\x01-\t\v\f\x0E-\x1F\x7F\x81\x8D\x8F\x90\x9D\xA0-\uFFFF]/g,zIe=/[|\\{}()[\]^$+*?.]/g,IX=new WeakMap;function qIe(e,t){if(e=e.replace(t.subset?VIe(t.subset):LIe,r),t.subset||t.escapeOnly)return e;return e.replace(BIe,n).replace(QIe,r);function n(i,a,u){return t.format((i.charCodeAt(0)-55296)*1024+i.charCodeAt(1)-56320+65536,u.charCodeAt(a+2),t)}function r(i,a,u){return t.format(i.charCodeAt(0),u.charCodeAt(a+1),t)}}function VIe(e){let t=IX.get(e);return t||(t=UIe(e),IX.set(e,t)),t}function UIe(e){const t=[];let n=-1;for(;++n<e.length;)t.push(e[n].replace(zIe,"\\$&"));return new RegExp("(?:"+t.join("|")+")","g")}const jIe=/[\dA-Fa-f]/;function ZIe(e,t,n){const r="&#x"+e.toString(16).toUpperCase();return n&&t&&!jIe.test(String.fromCharCode(t))?r:r+";"}const HIe=/\d/;function XIe(e,t,n){const r="&#"+String(e);return n&&t&&!HIe.test(String.fromCharCode(t))?r:r+";"}const YIe=["AElig","AMP","Aacute","Acirc","Agrave","Aring","Atilde","Auml","COPY","Ccedil","ETH","Eacute","Ecirc","Egrave","Euml","GT","Iacute","Icirc","Igrave","Iuml","LT","Ntilde","Oacute","Ocirc","Ograve","Oslash","Otilde","Ouml","QUOT","REG","THORN","Uacute","Ucirc","Ugrave","Uuml","Yacute","aacute","acirc","acute","aelig","agrave","amp","aring","atilde","auml","brvbar","ccedil","cedil","cent","copy","curren","deg","divide","eacute","ecirc","egrave","eth","euml","frac12","frac14","frac34","gt","iacute","icirc","iexcl","igrave","iquest","iuml","laquo","lt","macr","micro","middot","nbsp","not","ntilde","oacute","ocirc","ograve","ordf","ordm","oslash","otilde","ouml","para","plusmn","pound","quot","raquo","reg","sect","shy","sup1","sup2","sup3","szlig","thorn","times","uacute","ucirc","ugrave","uml","uuml","yacute","yen","yuml"],DP={nbsp:" ",iexcl:"¡",cent:"¢",pound:"£",curren:"¤",yen:"¥",brvbar:"¦",sect:"§",uml:"¨",copy:"©",ordf:"ª",laquo:"«",not:"¬",shy:"­",reg:"®",macr:"¯",deg:"°",plusmn:"±",sup2:"²",sup3:"³",acute:"´",micro:"µ",para:"¶",middot:"·",cedil:"¸",sup1:"¹",ordm:"º",raquo:"»",frac14:"¼",frac12:"½",frac34:"¾",iquest:"¿",Agrave:"À",Aacute:"Á",Acirc:"Â",Atilde:"Ã",Auml:"Ä",Aring:"Å",AElig:"Æ",Ccedil:"Ç",Egrave:"È",Eacute:"É",Ecirc:"Ê",Euml:"Ë",Igrave:"Ì",Iacute:"Í",Icirc:"Î",Iuml:"Ï",ETH:"Ð",Ntilde:"Ñ",Ograve:"Ò",Oacute:"Ó",Ocirc:"Ô",Otilde:"Õ",Ouml:"Ö",times:"×",Oslash:"Ø",Ugrave:"Ù",Uacute:"Ú",Ucirc:"Û",Uuml:"Ü",Yacute:"Ý",THORN:"Þ",szlig:"ß",agrave:"à",aacute:"á",acirc:"â",atilde:"ã",auml:"ä",aring:"å",aelig:"æ",ccedil:"ç",egrave:"è",eacute:"é",ecirc:"ê",euml:"ë",igrave:"ì",iacute:"í",icirc:"î",iuml:"ï",eth:"ð",ntilde:"ñ",ograve:"ò",oacute:"ó",ocirc:"ô",otilde:"õ",ouml:"ö",divide:"÷",oslash:"ø",ugrave:"ù",uacute:"ú",ucirc:"û",uuml:"ü",yacute:"ý",thorn:"þ",yuml:"ÿ",fnof:"ƒ",Alpha:"Α",Beta:"Β",Gamma:"Γ",Delta:"Δ",Epsilon:"Ε",Zeta:"Ζ",Eta:"Η",Theta:"Θ",Iota:"Ι",Kappa:"Κ",Lambda:"Λ",Mu:"Μ",Nu:"Ν",Xi:"Ξ",Omicron:"Ο",Pi:"Π",Rho:"Ρ",Sigma:"Σ",Tau:"Τ",Upsilon:"Υ",Phi:"Φ",Chi:"Χ",Psi:"Ψ",Omega:"Ω",alpha:"α",beta:"β",gamma:"γ",delta:"δ",epsilon:"ε",zeta:"ζ",eta:"η",theta:"θ",iota:"ι",kappa:"κ",lambda:"λ",mu:"μ",nu:"ν",xi:"ξ",omicron:"ο",pi:"π",rho:"ρ",sigmaf:"ς",sigma:"σ",tau:"τ",upsilon:"υ",phi:"φ",chi:"χ",psi:"ψ",omega:"ω",thetasym:"ϑ",upsih:"ϒ",piv:"ϖ",bull:"•",hellip:"…",prime:"′",Prime:"″",oline:"‾",frasl:"⁄",weierp:"℘",image:"ℑ",real:"ℜ",trade:"™",alefsym:"ℵ",larr:"←",uarr:"↑",rarr:"→",darr:"↓",harr:"↔",crarr:"↵",lArr:"⇐",uArr:"⇑",rArr:"⇒",dArr:"⇓",hArr:"⇔",forall:"∀",part:"∂",exist:"∃",empty:"∅",nabla:"∇",isin:"∈",notin:"∉",ni:"∋",prod:"∏",sum:"∑",minus:"−",lowast:"∗",radic:"√",prop:"∝",infin:"∞",ang:"∠",and:"∧",or:"∨",cap:"∩",cup:"∪",int:"∫",there4:"∴",sim:"∼",cong:"≅",asymp:"≈",ne:"≠",equiv:"≡",le:"≤",ge:"≥",sub:"⊂",sup:"⊃",nsub:"⊄",sube:"⊆",supe:"⊇",oplus:"⊕",otimes:"⊗",perp:"⊥",sdot:"⋅",lceil:"⌈",rceil:"⌉",lfloor:"⌊",rfloor:"⌋",lang:"〈",rang:"〉",loz:"◊",spades:"♠",clubs:"♣",hearts:"♥",diams:"♦",quot:'"',amp:"&",lt:"<",gt:">",OElig:"Œ",oelig:"œ",Scaron:"Š",scaron:"š",Yuml:"Ÿ",circ:"ˆ",tilde:"˜",ensp:" ",emsp:" ",thinsp:" ",zwnj:"‌",zwj:"‍",lrm:"‎",rlm:"‏",ndash:"–",mdash:"—",lsquo:"‘",rsquo:"’",sbquo:"‚",ldquo:"“",rdquo:"”",bdquo:"„",dagger:"†",Dagger:"‡",permil:"‰",lsaquo:"‹",rsaquo:"›",euro:"€"},GIe=["cent","copy","divide","gt","lt","not","para","times"],zme={}.hasOwnProperty,MM={};let SE;for(SE in DP)zme.call(DP,SE)&&(MM[DP[SE]]=SE);const KIe=/[^\dA-Za-z]/;function WIe(e,t,n,r){const i=String.fromCharCode(e);if(zme.call(MM,i)){const a=MM[i],u="&"+a;return n&&YIe.includes(a)&&!GIe.includes(a)&&(!r||t&&t!==61&&KIe.test(String.fromCharCode(t)))?u:u+";"}return""}function JIe(e,t,n){let r=ZIe(e,t,n.omitOptionalSemicolons),i;if((n.useNamedReferences||n.useShortestReferences)&&(i=WIe(e,t,n.omitOptionalSemicolons,n.attribute)),(n.useShortestReferences||!i)&&n.useShortestReferences){const a=XIe(e,t,n.omitOptionalSemicolons);a.length<r.length&&(r=a)}return i&&(!n.useShortestReferences||i.length<r.length)?i:r}function y3(e,t){return qIe(e,Object.assign({format:JIe},t))}const eLe=/^>|^->|<!--|-->|--!>|<!-$/g,tLe=[">"],nLe=["<",">"];function rLe(e,t,n,r){return r.settings.bogusComments?"<?"+y3(e.value,Object.assign({},r.settings.characterReferences,{subset:tLe}))+">":"<!--"+e.value.replace(eLe,i)+"-->";function i(a){return y3(a,Object.assign({},r.settings.characterReferences,{subset:nLe}))}}function iLe(e,t,n,r){return"<!"+(r.settings.upperDoctype?"DOCTYPE":"doctype")+(r.settings.tightDoctype?"":" ")+"html>"}function M$(e,t){const n=String(e);if(typeof t!="string")throw new TypeError("Expected character");let r=0,i=n.indexOf(t);for(;i!==-1;)r++,i=n.indexOf(t,i+t.length);return r}function qme(e,t){const n=t||{};return(e[e.length-1]===""?[...e,""]:e).join((n.padRight?" ":"")+","+(n.padLeft===!1?"":" ")).trim()}function Vme(e){return e.join(" ").trim()}const aLe=/[ \t\n\f\r]/g;function Aw(e){return typeof e=="object"?e.type==="text"?LX(e.value):!1:LX(e)}function LX(e){return e.replace(aLe,"")===""}const Ml=jme(1),Ume=jme(-1),uLe=[];function jme(e){return t;function t(n,r,i){const a=n?n.children:uLe;let u=(r||0)+e,l=a[u];if(!i)for(;l&&Aw(l);)u+=e,l=a[u];return l}}const lLe={}.hasOwnProperty;function Zme(e){return t;function t(n,r,i){return lLe.call(e,n.tagName)&&e[n.tagName](n,r,i)}}const uL=Zme({body:sLe,caption:$P,colgroup:$P,dd:hLe,dt:dLe,head:$P,html:oLe,li:fLe,optgroup:pLe,option:mLe,p:cLe,rp:BX,rt:BX,tbody:vLe,td:QX,tfoot:yLe,th:QX,thead:gLe,tr:bLe});function $P(e,t,n){const r=Ml(n,t,!0);return!r||r.type!=="comment"&&!(r.type==="text"&&Aw(r.value.charAt(0)))}function oLe(e,t,n){const r=Ml(n,t);return!r||r.type!=="comment"}function sLe(e,t,n){const r=Ml(n,t);return!r||r.type!=="comment"}function cLe(e,t,n){const r=Ml(n,t);return r?r.type==="element"&&(r.tagName==="address"||r.tagName==="article"||r.tagName==="aside"||r.tagName==="blockquote"||r.tagName==="details"||r.tagName==="div"||r.tagName==="dl"||r.tagName==="fieldset"||r.tagName==="figcaption"||r.tagName==="figure"||r.tagName==="footer"||r.tagName==="form"||r.tagName==="h1"||r.tagName==="h2"||r.tagName==="h3"||r.tagName==="h4"||r.tagName==="h5"||r.tagName==="h6"||r.tagName==="header"||r.tagName==="hgroup"||r.tagName==="hr"||r.tagName==="main"||r.tagName==="menu"||r.tagName==="nav"||r.tagName==="ol"||r.tagName==="p"||r.tagName==="pre"||r.tagName==="section"||r.tagName==="table"||r.tagName==="ul"):!n||!(n.type==="element"&&(n.tagName==="a"||n.tagName==="audio"||n.tagName==="del"||n.tagName==="ins"||n.tagName==="map"||n.tagName==="noscript"||n.tagName==="video"))}function fLe(e,t,n){const r=Ml(n,t);return!r||r.type==="element"&&r.tagName==="li"}function dLe(e,t,n){const r=Ml(n,t);return!!(r&&r.type==="element"&&(r.tagName==="dt"||r.tagName==="dd"))}function hLe(e,t,n){const r=Ml(n,t);return!r||r.type==="element"&&(r.tagName==="dt"||r.tagName==="dd")}function BX(e,t,n){const r=Ml(n,t);return!r||r.type==="element"&&(r.tagName==="rp"||r.tagName==="rt")}function pLe(e,t,n){const r=Ml(n,t);return!r||r.type==="element"&&r.tagName==="optgroup"}function mLe(e,t,n){const r=Ml(n,t);return!r||r.type==="element"&&(r.tagName==="option"||r.tagName==="optgroup")}function gLe(e,t,n){const r=Ml(n,t);return!!(r&&r.type==="element"&&(r.tagName==="tbody"||r.tagName==="tfoot"))}function vLe(e,t,n){const r=Ml(n,t);return!r||r.type==="element"&&(r.tagName==="tbody"||r.tagName==="tfoot")}function yLe(e,t,n){return!Ml(n,t)}function bLe(e,t,n){const r=Ml(n,t);return!r||r.type==="element"&&r.tagName==="tr"}function QX(e,t,n){const r=Ml(n,t);return!r||r.type==="element"&&(r.tagName==="td"||r.tagName==="th")}const OLe=Zme({body:_Le,colgroup:ELe,head:SLe,html:xLe,tbody:DLe});function xLe(e){const t=Ml(e,-1);return!t||t.type!=="comment"}function SLe(e){const t=new Set;for(const r of e.children)if(r.type==="element"&&(r.tagName==="base"||r.tagName==="title")){if(t.has(r.tagName))return!1;t.add(r.tagName)}const n=e.children[0];return!n||n.type==="element"}function _Le(e){const t=Ml(e,-1,!0);return!t||t.type!=="comment"&&!(t.type==="text"&&Aw(t.value.charAt(0)))&&!(t.type==="element"&&(t.tagName==="meta"||t.tagName==="link"||t.tagName==="script"||t.tagName==="style"||t.tagName==="template"))}function ELe(e,t,n){const r=Ume(n,t),i=Ml(e,-1,!0);return n&&r&&r.type==="element"&&r.tagName==="colgroup"&&uL(r,n.children.indexOf(r),n)?!1:!!(i&&i.type==="element"&&i.tagName==="col")}function DLe(e,t,n){const r=Ume(n,t),i=Ml(e,-1);return n&&r&&r.type==="element"&&(r.tagName==="thead"||r.tagName==="tbody")&&uL(r,n.children.indexOf(r),n)?!1:!!(i&&i.type==="element"&&i.tagName==="tr")}const _E={name:[[`	
\f\r &/=>`.split(""),`	
\f\r "&'/=>\``.split("")],[`\0	
\f\r "&'/<=>`.split(""),`\0	
\f\r "&'/<=>\``.split("")]],unquoted:[[`	
\f\r &>`.split(""),`\0	
\f\r "&'<=>\``.split("")],[`\0	
\f\r "&'<=>\``.split(""),`\0	
\f\r "&'<=>\``.split("")]],single:[["&'".split(""),"\"&'`".split("")],["\0&'".split(""),"\0\"&'`".split("")]],double:[['"&'.split(""),"\"&'`".split("")],['\0"&'.split(""),"\0\"&'`".split("")]]};function $Le(e,t,n,r){const i=r.schema,a=i.space==="svg"?!1:r.settings.omitOptionalTags;let u=i.space==="svg"?r.settings.closeEmptyElements:r.settings.voids.includes(e.tagName.toLowerCase());const l=[];let o;i.space==="html"&&e.tagName==="svg"&&(r.schema=FS);const s=CLe(r,e.properties),c=r.all(i.space==="html"&&e.tagName==="template"?e.content:e);return r.schema=i,c&&(u=!1),(s||!a||!OLe(e,t,n))&&(l.push("<",e.tagName,s?" "+s:""),u&&(i.space==="svg"||r.settings.closeSelfClosing)&&(o=s.charAt(s.length-1),(!r.settings.tightSelfClosing||o==="/"||o&&o!=='"'&&o!=="'")&&l.push(" "),l.push("/")),l.push(">")),l.push(c),!u&&(!a||!uL(e,t,n))&&l.push("</"+e.tagName+">"),l.join("")}function CLe(e,t){const n=[];let r=-1,i;if(t){for(i in t)if(t[i]!==null&&t[i]!==void 0){const a=wLe(e,i,t[i]);a&&n.push(a)}}for(;++r<n.length;){const a=e.settings.tightAttributes?n[r].charAt(n[r].length-1):void 0;r!==n.length-1&&a!=='"'&&a!=="'"&&(n[r]+=" ")}return n.join("")}function wLe(e,t,n){const r=Bme(e.schema,t),i=e.settings.allowParseErrors&&e.schema.space==="html"?0:1,a=e.settings.allowDangerousCharacters?0:1;let u=e.quote,l;if(r.overloadedBoolean&&(n===r.attribute||n==="")?n=!0:(r.boolean||r.overloadedBoolean)&&(typeof n!="string"||n===r.attribute||n==="")&&(n=!!n),n==null||n===!1||typeof n=="number"&&Number.isNaN(n))return"";const o=y3(r.attribute,Object.assign({},e.settings.characterReferences,{subset:_E.name[i][a]}));return n===!0||(n=Array.isArray(n)?(r.commaSeparated?qme:Vme)(n,{padLeft:!e.settings.tightCommaSeparatedLists}):String(n),e.settings.collapseEmptyAttributes&&!n)?o:(e.settings.preferUnquoted&&(l=y3(n,Object.assign({},e.settings.characterReferences,{attribute:!0,subset:_E.unquoted[i][a]}))),l!==n&&(e.settings.quoteSmart&&M$(n,u)>M$(n,e.alternative)&&(u=e.alternative),l=u+y3(n,Object.assign({},e.settings.characterReferences,{subset:(u==="'"?_E.single:_E.double)[i][a],attribute:!0}))+u),o+(l&&"="+l))}const RLe=["<","&"];function Hme(e,t,n,r){return n&&n.type==="element"&&(n.tagName==="script"||n.tagName==="style")?e.value:y3(e.value,Object.assign({},r.settings.characterReferences,{subset:RLe}))}function ALe(e,t,n,r){return r.settings.allowDangerousHtml?e.value:Hme(e,t,n,r)}function TLe(e,t,n,r){return r.all(e)}const kLe=IIe("type",{invalid:PLe,unknown:MLe,handlers:{comment:rLe,doctype:iLe,element:$Le,raw:ALe,root:TLe,text:Hme}});function PLe(e){throw new Error("Expected node, not `"+e+"`")}function MLe(e){const t=e;throw new Error("Cannot compile unknown node `"+t.type+"`")}const FLe={},NLe={},ILe=[];function _ut(e,t){const n=FLe,r=n.quote||'"',i=r==='"'?"'":'"';if(r!=='"'&&r!=="'")throw new Error("Invalid quote `"+r+"`, expected `'` or `\"`");return{one:LLe,all:BLe,settings:{omitOptionalTags:n.omitOptionalTags||!1,allowParseErrors:n.allowParseErrors||!1,allowDangerousCharacters:n.allowDangerousCharacters||!1,quoteSmart:n.quoteSmart||!1,preferUnquoted:n.preferUnquoted||!1,tightAttributes:n.tightAttributes||!1,upperDoctype:n.upperDoctype||!1,tightDoctype:n.tightDoctype||!1,bogusComments:n.bogusComments||!1,tightCommaSeparatedLists:n.tightCommaSeparatedLists||!1,tightSelfClosing:n.tightSelfClosing||!1,collapseEmptyAttributes:n.collapseEmptyAttributes||!1,allowDangerousHtml:n.allowDangerousHtml||!1,voids:n.voids||wIe,characterReferences:n.characterReferences||NLe,closeSelfClosing:n.closeSelfClosing||!1,closeEmptyElements:n.closeEmptyElements||!1},schema:n.space==="svg"?FS:Qme,quote:r,alternative:i}.one(Array.isArray(e)?{type:"root",children:e}:e,void 0,void 0)}function LLe(e,t,n){return kLe(e,t,n,this)}function BLe(e){const t=[],n=e&&e.children||ILe;let r=-1;for(;++r<n.length;)t[r]=this.one(n[r],r,e);return t.join("")}var CP={},zX;function QLe(){return zX||(zX=1,(function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.parse=e.enableErrorLogging=e.disableErrorLogging=e.setErrorLogger=void 0;let t=console.error;function n(R){t=R}e.setErrorLogger=n;function r(){t=()=>{}}e.disableErrorLogging=r;function i(){t=console.error}e.enableErrorLogging=i;function a(R){if(R!==void 0){if(R===null)return null;if(R==="")return"";R=R.replace(/\\+$/,F=>F.length%2===0?F:F.slice(0,-1));try{return JSON.parse(R)}catch(F){const[I,U]=R.trimLeft()[0]===":"?u(R,F):u(R,F,_);if(a.lastParseReminding=U,a.onExtraToken&&U.length>0){const q=U.trimRight();a.lastParseReminding=q,q.length>0&&a.onExtraToken(R,I,q)}return I}}}e.parse=a,(function(R){R.onExtraToken=(F,I,U)=>{t("parsed json with extra tokens:",{text:F,data:I,reminding:U})}})(a=e.parse||(e.parse={}));function u(R,F,I){const U=o[R[0]]||I;if(!U)throw t(`no parser registered for ${JSON.stringify(R[0])}:`,{s:R}),F;return U(R,F)}function l(R,F,I){return R[0]==='"'?y(R):R[0]==="'"?S(R):_(R,F,I)}const o={};function s(R){return R.trimLeft()}o[" "]=c,o["\r"]=c,o[`
`]=c,o["	"]=c;function c(R,F){return R=s(R),u(R,F)}o["["]=d;function d(R,F){R=R.substr(1);const I=[];for(R=s(R);R.length>0;){if(R[0]==="]"){R=R.substr(1);break}const U=u(R,F,(q,V)=>_(q,V,[",","]"]));I.push(U[0]),R=U[1],R=s(R),R[0]===","&&(R=R.substring(1),R=s(R))}return[I,R]}for(const R of"0123456789.-".slice())o[R]=h;function h(R){for(let F=0;F<R.length;F++){const I=R[F];if(o[I]===h)continue;const U=R.substring(0,F);return R=R.substring(F),[m(U),R]}return[m(R),""]}function m(R){if(R==="-")return-0;const F=+R;return Number.isNaN(F)?R:F}o['"']=y;function y(R){for(let F=1;F<R.length;F++){const I=R[F];if(I==="\\"){F++;continue}if(I==='"'){const U=v(R.substring(0,F+1));return R=R.substring(F+1),[JSON.parse(U),R]}}return[JSON.parse(v(R)+'"'),""]}function v(R){return R.replace(/\n/g,"\\n").replace(/\t/g,"\\t").replace(/\r/g,"\\r")}o["'"]=S;function S(R){for(let F=1;F<R.length;F++){const I=R[F];if(I==="\\"){F++;continue}if(I==="'"){const U=v(R.substring(0,F+1));return R=R.substring(F+1),[JSON.parse('"'+U.slice(1,-1)+'"'),R]}}return[JSON.parse('"'+v(R.slice(1))+'"'),""]}function _(R,F,I=[" "]){const U=Math.min(...I.map(N=>{const j=R.indexOf(N);return j===-1?R.length:j})),q=R.substring(0,U).trim(),V=R.substring(U);return[q,V]}o["{"]=E;function E(R,F){R=R.substr(1);const I={};for(R=s(R);R.length>0;){if(R[0]==="}"){R=R.substr(1);break}const U=l(R,F,[":","}"]),q=U[0];if(R=U[1],R=s(R),R[0]!==":"){I[q]=void 0;break}if(R=R.substr(1),R=s(R),R.length===0){I[q]=void 0;break}const V=u(R,F);I[q]=V[0],R=V[1],R=s(R),R[0]===","&&(R=R.substr(1),R=s(R))}return[I,R]}o.t=$;function $(R,F){return w(R,"true",!0,F)}o.f=k;function k(R,F){return w(R,"false",!1,F)}o.n=A;function A(R,F){return w(R,"null",null,F)}function w(R,F,I,U){for(let q=F.length;q>=1;q--)if(R.startsWith(F.slice(0,q)))return[I,R.slice(q)];{const q=JSON.stringify(R.slice(0,F.length));throw t(`unknown token starting with ${q}:`,{s:R}),U}}})(CP)),CP}var Eut=QLe(),zLe=/[\u1680\u2000-\u200A\u202F\u205F\u3000]/,qLe=/[\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u0860-\u086A\u08A0-\u08B4\u08B6-\u08BD\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u09FC\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0AF9\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D\u0C58-\u0C5A\u0C60\u0C61\u0C80\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D54-\u0D56\u0D5F-\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u1884\u1887-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191E\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1C80-\u1C88\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2119-\u211D\u2124\u2126\u2128\u212A-\u212D\u212F-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2E2F\u3005-\u3007\u3021-\u3029\u3031-\u3035\u3038-\u303C\u3041-\u3096\u309D-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312E\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FEA\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA69D\uA6A0-\uA6EF\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AE\uA7B0-\uA7B7\uA7F7-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA8FD\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uA9E0-\uA9E4\uA9E6-\uA9EF\uA9FA-\uA9FE\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA7E-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDD40-\uDD74\uDE80-\uDE9C\uDEA0-\uDED0\uDF00-\uDF1F\uDF2D-\uDF4A\uDF50-\uDF75\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF\uDFD1-\uDFD5]|\uD801[\uDC00-\uDC9D\uDCB0-\uDCD3\uDCD8-\uDCFB\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00\uDE10-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE4\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC03-\uDC37\uDC83-\uDCAF\uDCD0-\uDCE8\uDD03-\uDD26\uDD50-\uDD72\uDD76\uDD83-\uDDB2\uDDC1-\uDDC4\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE2B\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEDE\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3D\uDF50\uDF5D-\uDF61]|\uD805[\uDC00-\uDC34\uDC47-\uDC4A\uDC80-\uDCAF\uDCC4\uDCC5\uDCC7\uDD80-\uDDAE\uDDD8-\uDDDB\uDE00-\uDE2F\uDE44\uDE80-\uDEAA\uDF00-\uDF19]|\uD806[\uDCA0-\uDCDF\uDCFF\uDE00\uDE0B-\uDE32\uDE3A\uDE50\uDE5C-\uDE83\uDE86-\uDE89\uDEC0-\uDEF8]|\uD807[\uDC00-\uDC08\uDC0A-\uDC2E\uDC40\uDC72-\uDC8F\uDD00-\uDD06\uDD08\uDD09\uDD0B-\uDD30\uDD46]|\uD808[\uDC00-\uDF99]|\uD809[\uDC00-\uDC6E\uDC80-\uDD43]|[\uD80C\uD81C-\uD820\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872\uD874-\uD879][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDED0-\uDEED\uDF00-\uDF2F\uDF40-\uDF43\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50\uDF93-\uDF9F\uDFE0\uDFE1]|\uD821[\uDC00-\uDFEC]|\uD822[\uDC00-\uDEF2]|\uD82C[\uDC00-\uDD1E\uDD70-\uDEFB]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB]|\uD83A[\uDC00-\uDCC4\uDD00-\uDD43]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1\uDEB0-\uDFFF]|\uD87A[\uDC00-\uDFE0]|\uD87E[\uDC00-\uDE1D]/,VLe=/[\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0300-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u0483-\u0487\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u0591-\u05BD\u05BF\u05C1\u05C2\u05C4\u05C5\u05C7\u05D0-\u05EA\u05F0-\u05F2\u0610-\u061A\u0620-\u0669\u066E-\u06D3\u06D5-\u06DC\u06DF-\u06E8\u06EA-\u06FC\u06FF\u0710-\u074A\u074D-\u07B1\u07C0-\u07F5\u07FA\u0800-\u082D\u0840-\u085B\u0860-\u086A\u08A0-\u08B4\u08B6-\u08BD\u08D4-\u08E1\u08E3-\u0963\u0966-\u096F\u0971-\u0983\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BC-\u09C4\u09C7\u09C8\u09CB-\u09CE\u09D7\u09DC\u09DD\u09DF-\u09E3\u09E6-\u09F1\u09FC\u0A01-\u0A03\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A3C\u0A3E-\u0A42\u0A47\u0A48\u0A4B-\u0A4D\u0A51\u0A59-\u0A5C\u0A5E\u0A66-\u0A75\u0A81-\u0A83\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABC-\u0AC5\u0AC7-\u0AC9\u0ACB-\u0ACD\u0AD0\u0AE0-\u0AE3\u0AE6-\u0AEF\u0AF9-\u0AFF\u0B01-\u0B03\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3C-\u0B44\u0B47\u0B48\u0B4B-\u0B4D\u0B56\u0B57\u0B5C\u0B5D\u0B5F-\u0B63\u0B66-\u0B6F\u0B71\u0B82\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BBE-\u0BC2\u0BC6-\u0BC8\u0BCA-\u0BCD\u0BD0\u0BD7\u0BE6-\u0BEF\u0C00-\u0C03\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D-\u0C44\u0C46-\u0C48\u0C4A-\u0C4D\u0C55\u0C56\u0C58-\u0C5A\u0C60-\u0C63\u0C66-\u0C6F\u0C80-\u0C83\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBC-\u0CC4\u0CC6-\u0CC8\u0CCA-\u0CCD\u0CD5\u0CD6\u0CDE\u0CE0-\u0CE3\u0CE6-\u0CEF\u0CF1\u0CF2\u0D00-\u0D03\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D44\u0D46-\u0D48\u0D4A-\u0D4E\u0D54-\u0D57\u0D5F-\u0D63\u0D66-\u0D6F\u0D7A-\u0D7F\u0D82\u0D83\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0DCA\u0DCF-\u0DD4\u0DD6\u0DD8-\u0DDF\u0DE6-\u0DEF\u0DF2\u0DF3\u0E01-\u0E3A\u0E40-\u0E4E\u0E50-\u0E59\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB9\u0EBB-\u0EBD\u0EC0-\u0EC4\u0EC6\u0EC8-\u0ECD\u0ED0-\u0ED9\u0EDC-\u0EDF\u0F00\u0F18\u0F19\u0F20-\u0F29\u0F35\u0F37\u0F39\u0F3E-\u0F47\u0F49-\u0F6C\u0F71-\u0F84\u0F86-\u0F97\u0F99-\u0FBC\u0FC6\u1000-\u1049\u1050-\u109D\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u135D-\u135F\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1714\u1720-\u1734\u1740-\u1753\u1760-\u176C\u176E-\u1770\u1772\u1773\u1780-\u17D3\u17D7\u17DC\u17DD\u17E0-\u17E9\u180B-\u180D\u1810-\u1819\u1820-\u1877\u1880-\u18AA\u18B0-\u18F5\u1900-\u191E\u1920-\u192B\u1930-\u193B\u1946-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u19D0-\u19D9\u1A00-\u1A1B\u1A20-\u1A5E\u1A60-\u1A7C\u1A7F-\u1A89\u1A90-\u1A99\u1AA7\u1AB0-\u1ABD\u1B00-\u1B4B\u1B50-\u1B59\u1B6B-\u1B73\u1B80-\u1BF3\u1C00-\u1C37\u1C40-\u1C49\u1C4D-\u1C7D\u1C80-\u1C88\u1CD0-\u1CD2\u1CD4-\u1CF9\u1D00-\u1DF9\u1DFB-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u203F\u2040\u2054\u2071\u207F\u2090-\u209C\u20D0-\u20DC\u20E1\u20E5-\u20F0\u2102\u2107\u210A-\u2113\u2115\u2119-\u211D\u2124\u2126\u2128\u212A-\u212D\u212F-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D7F-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2DE0-\u2DFF\u2E2F\u3005-\u3007\u3021-\u302F\u3031-\u3035\u3038-\u303C\u3041-\u3096\u3099\u309A\u309D-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312E\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FEA\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA62B\uA640-\uA66F\uA674-\uA67D\uA67F-\uA6F1\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AE\uA7B0-\uA7B7\uA7F7-\uA827\uA840-\uA873\uA880-\uA8C5\uA8D0-\uA8D9\uA8E0-\uA8F7\uA8FB\uA8FD\uA900-\uA92D\uA930-\uA953\uA960-\uA97C\uA980-\uA9C0\uA9CF-\uA9D9\uA9E0-\uA9FE\uAA00-\uAA36\uAA40-\uAA4D\uAA50-\uAA59\uAA60-\uAA76\uAA7A-\uAAC2\uAADB-\uAADD\uAAE0-\uAAEF\uAAF2-\uAAF6\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABEA\uABEC\uABED\uABF0-\uABF9\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE00-\uFE0F\uFE20-\uFE2F\uFE33\uFE34\uFE4D-\uFE4F\uFE70-\uFE74\uFE76-\uFEFC\uFF10-\uFF19\uFF21-\uFF3A\uFF3F\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDD40-\uDD74\uDDFD\uDE80-\uDE9C\uDEA0-\uDED0\uDEE0\uDF00-\uDF1F\uDF2D-\uDF4A\uDF50-\uDF7A\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF\uDFD1-\uDFD5]|\uD801[\uDC00-\uDC9D\uDCA0-\uDCA9\uDCB0-\uDCD3\uDCD8-\uDCFB\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00-\uDE03\uDE05\uDE06\uDE0C-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE38-\uDE3A\uDE3F\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE6\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC00-\uDC46\uDC66-\uDC6F\uDC7F-\uDCBA\uDCD0-\uDCE8\uDCF0-\uDCF9\uDD00-\uDD34\uDD36-\uDD3F\uDD50-\uDD73\uDD76\uDD80-\uDDC4\uDDCA-\uDDCC\uDDD0-\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE37\uDE3E\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEEA\uDEF0-\uDEF9\uDF00-\uDF03\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3C-\uDF44\uDF47\uDF48\uDF4B-\uDF4D\uDF50\uDF57\uDF5D-\uDF63\uDF66-\uDF6C\uDF70-\uDF74]|\uD805[\uDC00-\uDC4A\uDC50-\uDC59\uDC80-\uDCC5\uDCC7\uDCD0-\uDCD9\uDD80-\uDDB5\uDDB8-\uDDC0\uDDD8-\uDDDD\uDE00-\uDE40\uDE44\uDE50-\uDE59\uDE80-\uDEB7\uDEC0-\uDEC9\uDF00-\uDF19\uDF1D-\uDF2B\uDF30-\uDF39]|\uD806[\uDCA0-\uDCE9\uDCFF\uDE00-\uDE3E\uDE47\uDE50-\uDE83\uDE86-\uDE99\uDEC0-\uDEF8]|\uD807[\uDC00-\uDC08\uDC0A-\uDC36\uDC38-\uDC40\uDC50-\uDC59\uDC72-\uDC8F\uDC92-\uDCA7\uDCA9-\uDCB6\uDD00-\uDD06\uDD08\uDD09\uDD0B-\uDD36\uDD3A\uDD3C\uDD3D\uDD3F-\uDD47\uDD50-\uDD59]|\uD808[\uDC00-\uDF99]|\uD809[\uDC00-\uDC6E\uDC80-\uDD43]|[\uD80C\uD81C-\uD820\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872\uD874-\uD879][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDE60-\uDE69\uDED0-\uDEED\uDEF0-\uDEF4\uDF00-\uDF36\uDF40-\uDF43\uDF50-\uDF59\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50-\uDF7E\uDF8F-\uDF9F\uDFE0\uDFE1]|\uD821[\uDC00-\uDFEC]|\uD822[\uDC00-\uDEF2]|\uD82C[\uDC00-\uDD1E\uDD70-\uDEFB]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99\uDC9D\uDC9E]|\uD834[\uDD65-\uDD69\uDD6D-\uDD72\uDD7B-\uDD82\uDD85-\uDD8B\uDDAA-\uDDAD\uDE42-\uDE44]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB\uDFCE-\uDFFF]|\uD836[\uDE00-\uDE36\uDE3B-\uDE6C\uDE75\uDE84\uDE9B-\uDE9F\uDEA1-\uDEAF]|\uD838[\uDC00-\uDC06\uDC08-\uDC18\uDC1B-\uDC21\uDC23\uDC24\uDC26-\uDC2A]|\uD83A[\uDC00-\uDCC4\uDCD0-\uDCD6\uDD00-\uDD4A\uDD50-\uDD59]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1\uDEB0-\uDFFF]|\uD87A[\uDC00-\uDFE0]|\uD87E[\uDC00-\uDE1D]|\uDB40[\uDD00-\uDDEF]/,wP={Space_Separator:zLe,ID_Start:qLe,ID_Continue:VLe},Uu={isSpaceSeparator(e){return typeof e=="string"&&wP.Space_Separator.test(e)},isIdStartChar(e){return typeof e=="string"&&(e>="a"&&e<="z"||e>="A"&&e<="Z"||e==="$"||e==="_"||wP.ID_Start.test(e))},isIdContinueChar(e){return typeof e=="string"&&(e>="a"&&e<="z"||e>="A"&&e<="Z"||e>="0"&&e<="9"||e==="$"||e==="_"||e==="‌"||e==="‍"||wP.ID_Continue.test(e))},isDigit(e){return typeof e=="string"&&/[0-9]/.test(e)},isHexDigit(e){return typeof e=="string"&&/[0-9A-Fa-f]/.test(e)}};let FM,sc,Ym,F$,uy,ip,io,lL,J2;var ULe=function(t,n){FM=String(t),sc="start",Ym=[],F$=0,uy=1,ip=0,io=void 0,lL=void 0,J2=void 0;do io=jLe(),XLe[sc]();while(io.type!=="eof");return typeof n=="function"?NM({"":J2},"",n):J2};function NM(e,t,n){const r=e[t];if(r!=null&&typeof r=="object")if(Array.isArray(r))for(let i=0;i<r.length;i++){const a=String(i),u=NM(r,a,n);u===void 0?delete r[a]:Object.defineProperty(r,a,{value:u,writable:!0,enumerable:!0,configurable:!0})}else for(const i in r){const a=NM(r,i,n);a===void 0?delete r[i]:Object.defineProperty(r,i,{value:a,writable:!0,enumerable:!0,configurable:!0})}return n.call(e,t,r)}let Dr,dr,B2,Mm,Xr;function jLe(){for(Dr="default",dr="",B2=!1,Mm=1;;){Xr=ag();const e=Xme[Dr]();if(e)return e}}function ag(){if(FM[F$])return String.fromCodePoint(FM.codePointAt(F$))}function $t(){const e=ag();return e===`
`?(uy++,ip=0):e?ip+=e.length:ip++,e&&(F$+=e.length),e}const Xme={default(){switch(Xr){case"	":case"\v":case"\f":case" ":case" ":case"\uFEFF":case`
`:case"\r":case"\u2028":case"\u2029":$t();return;case"/":$t(),Dr="comment";return;case void 0:return $t(),Ma("eof")}if(Uu.isSpaceSeparator(Xr)){$t();return}return Xme[sc]()},comment(){switch(Xr){case"*":$t(),Dr="multiLineComment";return;case"/":$t(),Dr="singleLineComment";return}throw Fa($t())},multiLineComment(){switch(Xr){case"*":$t(),Dr="multiLineCommentAsterisk";return;case void 0:throw Fa($t())}$t()},multiLineCommentAsterisk(){switch(Xr){case"*":$t();return;case"/":$t(),Dr="default";return;case void 0:throw Fa($t())}$t(),Dr="multiLineComment"},singleLineComment(){switch(Xr){case`
`:case"\r":case"\u2028":case"\u2029":$t(),Dr="default";return;case void 0:return $t(),Ma("eof")}$t()},value(){switch(Xr){case"{":case"[":return Ma("punctuator",$t());case"n":return $t(),bb("ull"),Ma("null",null);case"t":return $t(),bb("rue"),Ma("boolean",!0);case"f":return $t(),bb("alse"),Ma("boolean",!1);case"-":case"+":$t()==="-"&&(Mm=-1),Dr="sign";return;case".":dr=$t(),Dr="decimalPointLeading";return;case"0":dr=$t(),Dr="zero";return;case"1":case"2":case"3":case"4":case"5":case"6":case"7":case"8":case"9":dr=$t(),Dr="decimalInteger";return;case"I":return $t(),bb("nfinity"),Ma("numeric",1/0);case"N":return $t(),bb("aN"),Ma("numeric",NaN);case'"':case"'":B2=$t()==='"',dr="",Dr="string";return}throw Fa($t())},identifierNameStartEscape(){if(Xr!=="u")throw Fa($t());$t();const e=IM();switch(e){case"$":case"_":break;default:if(!Uu.isIdStartChar(e))throw qX();break}dr+=e,Dr="identifierName"},identifierName(){switch(Xr){case"$":case"_":case"‌":case"‍":dr+=$t();return;case"\\":$t(),Dr="identifierNameEscape";return}if(Uu.isIdContinueChar(Xr)){dr+=$t();return}return Ma("identifier",dr)},identifierNameEscape(){if(Xr!=="u")throw Fa($t());$t();const e=IM();switch(e){case"$":case"_":case"‌":case"‍":break;default:if(!Uu.isIdContinueChar(e))throw qX();break}dr+=e,Dr="identifierName"},sign(){switch(Xr){case".":dr=$t(),Dr="decimalPointLeading";return;case"0":dr=$t(),Dr="zero";return;case"1":case"2":case"3":case"4":case"5":case"6":case"7":case"8":case"9":dr=$t(),Dr="decimalInteger";return;case"I":return $t(),bb("nfinity"),Ma("numeric",Mm*(1/0));case"N":return $t(),bb("aN"),Ma("numeric",NaN)}throw Fa($t())},zero(){switch(Xr){case".":dr+=$t(),Dr="decimalPoint";return;case"e":case"E":dr+=$t(),Dr="decimalExponent";return;case"x":case"X":dr+=$t(),Dr="hexadecimal";return}return Ma("numeric",Mm*0)},decimalInteger(){switch(Xr){case".":dr+=$t(),Dr="decimalPoint";return;case"e":case"E":dr+=$t(),Dr="decimalExponent";return}if(Uu.isDigit(Xr)){dr+=$t();return}return Ma("numeric",Mm*Number(dr))},decimalPointLeading(){if(Uu.isDigit(Xr)){dr+=$t(),Dr="decimalFraction";return}throw Fa($t())},decimalPoint(){switch(Xr){case"e":case"E":dr+=$t(),Dr="decimalExponent";return}if(Uu.isDigit(Xr)){dr+=$t(),Dr="decimalFraction";return}return Ma("numeric",Mm*Number(dr))},decimalFraction(){switch(Xr){case"e":case"E":dr+=$t(),Dr="decimalExponent";return}if(Uu.isDigit(Xr)){dr+=$t();return}return Ma("numeric",Mm*Number(dr))},decimalExponent(){switch(Xr){case"+":case"-":dr+=$t(),Dr="decimalExponentSign";return}if(Uu.isDigit(Xr)){dr+=$t(),Dr="decimalExponentInteger";return}throw Fa($t())},decimalExponentSign(){if(Uu.isDigit(Xr)){dr+=$t(),Dr="decimalExponentInteger";return}throw Fa($t())},decimalExponentInteger(){if(Uu.isDigit(Xr)){dr+=$t();return}return Ma("numeric",Mm*Number(dr))},hexadecimal(){if(Uu.isHexDigit(Xr)){dr+=$t(),Dr="hexadecimalInteger";return}throw Fa($t())},hexadecimalInteger(){if(Uu.isHexDigit(Xr)){dr+=$t();return}return Ma("numeric",Mm*Number(dr))},string(){switch(Xr){case"\\":$t(),dr+=ZLe();return;case'"':if(B2)return $t(),Ma("string",dr);dr+=$t();return;case"'":if(!B2)return $t(),Ma("string",dr);dr+=$t();return;case`
`:case"\r":throw Fa($t());case"\u2028":case"\u2029":YLe(Xr);break;case void 0:throw Fa($t())}dr+=$t()},start(){switch(Xr){case"{":case"[":return Ma("punctuator",$t())}Dr="value"},beforePropertyName(){switch(Xr){case"$":case"_":dr=$t(),Dr="identifierName";return;case"\\":$t(),Dr="identifierNameStartEscape";return;case"}":return Ma("punctuator",$t());case'"':case"'":B2=$t()==='"',Dr="string";return}if(Uu.isIdStartChar(Xr)){dr+=$t(),Dr="identifierName";return}throw Fa($t())},afterPropertyName(){if(Xr===":")return Ma("punctuator",$t());throw Fa($t())},beforePropertyValue(){Dr="value"},afterPropertyValue(){switch(Xr){case",":case"}":return Ma("punctuator",$t())}throw Fa($t())},beforeArrayValue(){if(Xr==="]")return Ma("punctuator",$t());Dr="value"},afterArrayValue(){switch(Xr){case",":case"]":return Ma("punctuator",$t())}throw Fa($t())},end(){throw Fa($t())}};function Ma(e,t){return{type:e,value:t,line:uy,column:ip}}function bb(e){for(const t of e){if(ag()!==t)throw Fa($t());$t()}}function ZLe(){switch(ag()){case"b":return $t(),"\b";case"f":return $t(),"\f";case"n":return $t(),`
`;case"r":return $t(),"\r";case"t":return $t(),"	";case"v":return $t(),"\v";case"0":if($t(),Uu.isDigit(ag()))throw Fa($t());return"\0";case"x":return $t(),HLe();case"u":return $t(),IM();case`
`:case"\u2028":case"\u2029":return $t(),"";case"\r":return $t(),ag()===`
`&&$t(),"";case"1":case"2":case"3":case"4":case"5":case"6":case"7":case"8":case"9":throw Fa($t());case void 0:throw Fa($t())}return $t()}function HLe(){let e="",t=ag();if(!Uu.isHexDigit(t)||(e+=$t(),t=ag(),!Uu.isHexDigit(t)))throw Fa($t());return e+=$t(),String.fromCodePoint(parseInt(e,16))}function IM(){let e="",t=4;for(;t-- >0;){const n=ag();if(!Uu.isHexDigit(n))throw Fa($t());e+=$t()}return String.fromCodePoint(parseInt(e,16))}const XLe={start(){if(io.type==="eof")throw Ob();RP()},beforePropertyName(){switch(io.type){case"identifier":case"string":lL=io.value,sc="afterPropertyName";return;case"punctuator":EE();return;case"eof":throw Ob()}},afterPropertyName(){if(io.type==="eof")throw Ob();sc="beforePropertyValue"},beforePropertyValue(){if(io.type==="eof")throw Ob();RP()},beforeArrayValue(){if(io.type==="eof")throw Ob();if(io.type==="punctuator"&&io.value==="]"){EE();return}RP()},afterPropertyValue(){if(io.type==="eof")throw Ob();switch(io.value){case",":sc="beforePropertyName";return;case"}":EE()}},afterArrayValue(){if(io.type==="eof")throw Ob();switch(io.value){case",":sc="beforeArrayValue";return;case"]":EE()}},end(){}};function RP(){let e;switch(io.type){case"punctuator":switch(io.value){case"{":e={};break;case"[":e=[];break}break;case"null":case"boolean":case"numeric":case"string":e=io.value;break}if(J2===void 0)J2=e;else{const t=Ym[Ym.length-1];Array.isArray(t)?t.push(e):Object.defineProperty(t,lL,{value:e,writable:!0,enumerable:!0,configurable:!0})}if(e!==null&&typeof e=="object")Ym.push(e),Array.isArray(e)?sc="beforeArrayValue":sc="beforePropertyName";else{const t=Ym[Ym.length-1];t==null?sc="end":Array.isArray(t)?sc="afterArrayValue":sc="afterPropertyValue"}}function EE(){Ym.pop();const e=Ym[Ym.length-1];e==null?sc="end":Array.isArray(e)?sc="afterArrayValue":sc="afterPropertyValue"}function Fa(e){return N$(e===void 0?`JSON5: invalid end of input at ${uy}:${ip}`:`JSON5: invalid character '${Yme(e)}' at ${uy}:${ip}`)}function Ob(){return N$(`JSON5: invalid end of input at ${uy}:${ip}`)}function qX(){return ip-=5,N$(`JSON5: invalid identifier character at ${uy}:${ip}`)}function YLe(e){console.warn(`JSON5: '${Yme(e)}' in strings is not valid ECMAScript; consider escaping`)}function Yme(e){const t={"'":"\\'",'"':'\\"',"\\":"\\\\","\b":"\\b","\f":"\\f","\n":"\\n","\r":"\\r","	":"\\t","\v":"\\v","\0":"\\0","\u2028":"\\u2028","\u2029":"\\u2029"};if(t[e])return t[e];if(e<" "){const n=e.charCodeAt(0).toString(16);return"\\x"+("00"+n).substring(n.length)}return e}function N$(e){const t=new SyntaxError(e);return t.lineNumber=uy,t.columnNumber=ip,t}var GLe=function(t,n,r){const i=[];let a="",u,l,o="",s;if(n!=null&&typeof n=="object"&&!Array.isArray(n)&&(r=n.space,s=n.quote,n=n.replacer),typeof n=="function")l=n;else if(Array.isArray(n)){u=[];for(const v of n){let S;typeof v=="string"?S=v:(typeof v=="number"||v instanceof String||v instanceof Number)&&(S=String(v)),S!==void 0&&u.indexOf(S)<0&&u.push(S)}}return r instanceof Number?r=Number(r):r instanceof String&&(r=String(r)),typeof r=="number"?r>0&&(r=Math.min(10,Math.floor(r)),o="          ".substr(0,r)):typeof r=="string"&&(o=r.substr(0,10)),c("",{"":t});function c(v,S){let _=S[v];switch(_!=null&&(typeof _.toJSON5=="function"?_=_.toJSON5(v):typeof _.toJSON=="function"&&(_=_.toJSON(v))),l&&(_=l.call(S,v,_)),_ instanceof Number?_=Number(_):_ instanceof String?_=String(_):_ instanceof Boolean&&(_=_.valueOf()),_){case null:return"null";case!0:return"true";case!1:return"false"}if(typeof _=="string")return d(_);if(typeof _=="number")return String(_);if(typeof _=="object")return Array.isArray(_)?y(_):h(_)}function d(v){const S={"'":.1,'"':.2},_={"'":"\\'",'"':'\\"',"\\":"\\\\","\b":"\\b","\f":"\\f","\n":"\\n","\r":"\\r","	":"\\t","\v":"\\v","\0":"\\0","\u2028":"\\u2028","\u2029":"\\u2029"};let E="";for(let k=0;k<v.length;k++){const A=v[k];switch(A){case"'":case'"':S[A]++,E+=A;continue;case"\0":if(Uu.isDigit(v[k+1])){E+="\\x00";continue}}if(_[A]){E+=_[A];continue}if(A<" "){let w=A.charCodeAt(0).toString(16);E+="\\x"+("00"+w).substring(w.length);continue}E+=A}const $=s||Object.keys(S).reduce((k,A)=>S[k]<S[A]?k:A);return E=E.replace(new RegExp($,"g"),_[$]),$+E+$}function h(v){if(i.indexOf(v)>=0)throw TypeError("Converting circular structure to JSON5");i.push(v);let S=a;a=a+o;let _=u||Object.keys(v),E=[];for(const k of _){const A=c(k,v);if(A!==void 0){let w=m(k)+":";o!==""&&(w+=" "),w+=A,E.push(w)}}let $;if(E.length===0)$="{}";else{let k;if(o==="")k=E.join(","),$="{"+k+"}";else{let A=`,
`+a;k=E.join(A),$=`{
`+a+k+`,
`+S+"}"}}return i.pop(),a=S,$}function m(v){if(v.length===0)return d(v);const S=String.fromCodePoint(v.codePointAt(0));if(!Uu.isIdStartChar(S))return d(v);for(let _=S.length;_<v.length;_++)if(!Uu.isIdContinueChar(String.fromCodePoint(v.codePointAt(_))))return d(v);return v}function y(v){if(i.indexOf(v)>=0)throw TypeError("Converting circular structure to JSON5");i.push(v);let S=a;a=a+o;let _=[];for(let $=0;$<v.length;$++){const k=c(String($),v);_.push(k!==void 0?k:"null")}let E;if(_.length===0)E="[]";else if(o==="")E="["+_.join(",")+"]";else{let $=`,
`+a,k=_.join($);E=`[
`+a+k+`,
`+S+"]"}return i.pop(),a=S,E}};const KLe={parse:ULe,stringify:GLe};var Dut=KLe;const oL=Symbol.for("yaml.alias"),LM=Symbol.for("yaml.document"),Hv=Symbol.for("yaml.map"),Gme=Symbol.for("yaml.pair"),P0=Symbol.for("yaml.scalar"),o4=Symbol.for("yaml.seq"),eh=Symbol.for("yaml.node.type"),xy=e=>!!e&&typeof e=="object"&&e[eh]===oL,k1=e=>!!e&&typeof e=="object"&&e[eh]===LM,s4=e=>!!e&&typeof e=="object"&&e[eh]===Hv,Ha=e=>!!e&&typeof e=="object"&&e[eh]===Gme,fa=e=>!!e&&typeof e=="object"&&e[eh]===P0,c4=e=>!!e&&typeof e=="object"&&e[eh]===o4;function cu(e){if(e&&typeof e=="object")switch(e[eh]){case Hv:case o4:return!0}return!1}function hu(e){if(e&&typeof e=="object")switch(e[eh]){case oL:case Hv:case P0:case o4:return!0}return!1}const Kme=e=>(fa(e)||cu(e))&&!!e.anchor,rf=Symbol("break visit"),Wme=Symbol("skip children"),R0=Symbol("remove node");function P1(e,t){const n=Jme(t);k1(e)?XO(null,e.contents,n,Object.freeze([e]))===R0&&(e.contents=null):XO(null,e,n,Object.freeze([]))}P1.BREAK=rf;P1.SKIP=Wme;P1.REMOVE=R0;function XO(e,t,n,r){const i=ege(e,t,n,r);if(hu(i)||Ha(i))return tge(e,r,i),XO(e,i,n,r);if(typeof i!="symbol"){if(cu(t)){r=Object.freeze(r.concat(t));for(let a=0;a<t.items.length;++a){const u=XO(a,t.items[a],n,r);if(typeof u=="number")a=u-1;else{if(u===rf)return rf;u===R0&&(t.items.splice(a,1),a-=1)}}}else if(Ha(t)){r=Object.freeze(r.concat(t));const a=XO("key",t.key,n,r);if(a===rf)return rf;a===R0&&(t.key=null);const u=XO("value",t.value,n,r);if(u===rf)return rf;u===R0&&(t.value=null)}}return i}async function Tw(e,t){const n=Jme(t);k1(e)?await YO(null,e.contents,n,Object.freeze([e]))===R0&&(e.contents=null):await YO(null,e,n,Object.freeze([]))}Tw.BREAK=rf;Tw.SKIP=Wme;Tw.REMOVE=R0;async function YO(e,t,n,r){const i=await ege(e,t,n,r);if(hu(i)||Ha(i))return tge(e,r,i),YO(e,i,n,r);if(typeof i!="symbol"){if(cu(t)){r=Object.freeze(r.concat(t));for(let a=0;a<t.items.length;++a){const u=await YO(a,t.items[a],n,r);if(typeof u=="number")a=u-1;else{if(u===rf)return rf;u===R0&&(t.items.splice(a,1),a-=1)}}}else if(Ha(t)){r=Object.freeze(r.concat(t));const a=await YO("key",t.key,n,r);if(a===rf)return rf;a===R0&&(t.key=null);const u=await YO("value",t.value,n,r);if(u===rf)return rf;u===R0&&(t.value=null)}}return i}function Jme(e){return typeof e=="object"&&(e.Collection||e.Node||e.Value)?Object.assign({Alias:e.Node,Map:e.Node,Scalar:e.Node,Seq:e.Node},e.Value&&{Map:e.Value,Scalar:e.Value,Seq:e.Value},e.Collection&&{Map:e.Collection,Seq:e.Collection},e):e}function ege(e,t,n,r){var i,a,u,l,o;if(typeof n=="function")return n(e,t,r);if(s4(t))return(i=n.Map)==null?void 0:i.call(n,e,t,r);if(c4(t))return(a=n.Seq)==null?void 0:a.call(n,e,t,r);if(Ha(t))return(u=n.Pair)==null?void 0:u.call(n,e,t,r);if(fa(t))return(l=n.Scalar)==null?void 0:l.call(n,e,t,r);if(xy(t))return(o=n.Alias)==null?void 0:o.call(n,e,t,r)}function tge(e,t,n){const r=t[t.length-1];if(cu(r))r.items[e]=n;else if(Ha(r))e==="key"?r.key=n:r.value=n;else if(k1(r))r.contents=n;else{const i=xy(r)?"alias":"scalar";throw new Error(`Cannot replace node with ${i} parent`)}}const WLe={"!":"%21",",":"%2C","[":"%5B","]":"%5D","{":"%7B","}":"%7D"},JLe=e=>e.replace(/[!,[\]{}]/g,t=>WLe[t]);class uc{constructor(t,n){this.docStart=null,this.docEnd=!1,this.yaml=Object.assign({},uc.defaultYaml,t),this.tags=Object.assign({},uc.defaultTags,n)}clone(){const t=new uc(this.yaml,this.tags);return t.docStart=this.docStart,t}atDocument(){const t=new uc(this.yaml,this.tags);switch(this.yaml.version){case"1.1":this.atNextDocument=!0;break;case"1.2":this.atNextDocument=!1,this.yaml={explicit:uc.defaultYaml.explicit,version:"1.2"},this.tags=Object.assign({},uc.defaultTags);break}return t}add(t,n){this.atNextDocument&&(this.yaml={explicit:uc.defaultYaml.explicit,version:"1.1"},this.tags=Object.assign({},uc.defaultTags),this.atNextDocument=!1);const r=t.trim().split(/[ \t]+/),i=r.shift();switch(i){case"%TAG":{if(r.length!==2&&(n(0,"%TAG directive should contain exactly two parts"),r.length<2))return!1;const[a,u]=r;return this.tags[a]=u,!0}case"%YAML":{if(this.yaml.explicit=!0,r.length!==1)return n(0,"%YAML directive should contain exactly one part"),!1;const[a]=r;if(a==="1.1"||a==="1.2")return this.yaml.version=a,!0;{const u=/^\d+\.\d+$/.test(a);return n(6,`Unsupported YAML version ${a}`,u),!1}}default:return n(0,`Unknown directive ${i}`,!0),!1}}tagName(t,n){if(t==="!")return"!";if(t[0]!=="!")return n(`Not a valid tag: ${t}`),null;if(t[1]==="<"){const u=t.slice(2,-1);return u==="!"||u==="!!"?(n(`Verbatim tags aren't resolved, so ${t} is invalid.`),null):(t[t.length-1]!==">"&&n("Verbatim tags must end with a >"),u)}const[,r,i]=t.match(/^(.*!)([^!]*)$/s);i||n(`The ${t} tag has no suffix`);const a=this.tags[r];if(a)try{return a+decodeURIComponent(i)}catch(u){return n(String(u)),null}return r==="!"?t:(n(`Could not resolve tag: ${t}`),null)}tagString(t){for(const[n,r]of Object.entries(this.tags))if(t.startsWith(r))return n+JLe(t.substring(r.length));return t[0]==="!"?t:`!<${t}>`}toString(t){const n=this.yaml.explicit?[`%YAML ${this.yaml.version||"1.2"}`]:[],r=Object.entries(this.tags);let i;if(t&&r.length>0&&hu(t.contents)){const a={};P1(t.contents,(u,l)=>{hu(l)&&l.tag&&(a[l.tag]=!0)}),i=Object.keys(a)}else i=[];for(const[a,u]of r)a==="!!"&&u==="tag:yaml.org,2002:"||(!t||i.some(l=>l.startsWith(u)))&&n.push(`%TAG ${a} ${u}`);return n.join(`
`)}}uc.defaultYaml={explicit:!1,version:"1.2"};uc.defaultTags={"!!":"tag:yaml.org,2002:"};function nge(e){if(/[\x00-\x19\s,[\]{}]/.test(e)){const n=`Anchor must not contain whitespace or control characters: ${JSON.stringify(e)}`;throw new Error(n)}return!0}function rge(e){const t=new Set;return P1(e,{Value(n,r){r.anchor&&t.add(r.anchor)}}),t}function ige(e,t){for(let n=1;;++n){const r=`${e}${n}`;if(!t.has(r))return r}}function eBe(e,t){const n=[],r=new Map;let i=null;return{onAnchor:a=>{n.push(a),i??(i=rge(e));const u=ige(t,i);return i.add(u),u},setAnchors:()=>{for(const a of n){const u=r.get(a);if(typeof u=="object"&&u.anchor&&(fa(u.node)||cu(u.node)))u.node.anchor=u.anchor;else{const l=new Error("Failed to resolve repeated object (this should not happen)");throw l.source=a,l}}},sourceObjects:r}}function GO(e,t,n,r){if(r&&typeof r=="object")if(Array.isArray(r))for(let i=0,a=r.length;i<a;++i){const u=r[i],l=GO(e,r,String(i),u);l===void 0?delete r[i]:l!==u&&(r[i]=l)}else if(r instanceof Map)for(const i of Array.from(r.keys())){const a=r.get(i),u=GO(e,r,i,a);u===void 0?r.delete(i):u!==a&&r.set(i,u)}else if(r instanceof Set)for(const i of Array.from(r)){const a=GO(e,r,i,i);a===void 0?r.delete(i):a!==i&&(r.delete(i),r.add(a))}else for(const[i,a]of Object.entries(r)){const u=GO(e,r,i,a);u===void 0?delete r[i]:u!==a&&(r[i]=u)}return e.call(t,n,r)}function Wd(e,t,n){if(Array.isArray(e))return e.map((r,i)=>Wd(r,String(i),n));if(e&&typeof e.toJSON=="function"){if(!n||!Kme(e))return e.toJSON(t,n);const r={aliasCount:0,count:1,res:void 0};n.anchors.set(e,r),n.onCreate=a=>{r.res=a,delete n.onCreate};const i=e.toJSON(t,n);return n.onCreate&&n.onCreate(i),i}return typeof e=="bigint"&&!(n!=null&&n.keep)?Number(e):e}class sL{constructor(t){Object.defineProperty(this,eh,{value:t})}clone(){const t=Object.create(Object.getPrototypeOf(this),Object.getOwnPropertyDescriptors(this));return this.range&&(t.range=this.range.slice()),t}toJS(t,{mapAsMap:n,maxAliasCount:r,onAnchor:i,reviver:a}={}){if(!k1(t))throw new TypeError("A document argument is required");const u={anchors:new Map,doc:t,keep:!0,mapAsMap:n===!0,mapKeyWarned:!1,maxAliasCount:typeof r=="number"?r:100},l=Wd(this,"",u);if(typeof i=="function")for(const{count:o,res:s}of u.anchors.values())i(s,o);return typeof a=="function"?GO(a,{"":l},"",l):l}}class kw extends sL{constructor(t){super(oL),this.source=t,Object.defineProperty(this,"tag",{set(){throw new Error("Alias nodes cannot have tags")}})}resolve(t,n){let r;n!=null&&n.aliasResolveCache?r=n.aliasResolveCache:(r=[],P1(t,{Node:(a,u)=>{(xy(u)||Kme(u))&&r.push(u)}}),n&&(n.aliasResolveCache=r));let i;for(const a of r){if(a===this)break;a.anchor===this.source&&(i=a)}return i}toJSON(t,n){if(!n)return{source:this.source};const{anchors:r,doc:i,maxAliasCount:a}=n,u=this.resolve(i,n);if(!u){const o=`Unresolved alias (the anchor must be set before the alias): ${this.source}`;throw new ReferenceError(o)}let l=r.get(u);if(l||(Wd(u,null,n),l=r.get(u)),!l||l.res===void 0){const o="This should not happen: Alias anchor was not resolved?";throw new ReferenceError(o)}if(a>=0&&(l.count+=1,l.aliasCount===0&&(l.aliasCount=vD(i,u,r)),l.count*l.aliasCount>a)){const o="Excessive alias count indicates a resource exhaustion attack";throw new ReferenceError(o)}return l.res}toString(t,n,r){const i=`*${this.source}`;if(t){if(nge(this.source),t.options.verifyAliasOrder&&!t.anchors.has(this.source)){const a=`Unresolved alias (the anchor must be set before the alias): ${this.source}`;throw new Error(a)}if(t.implicitKey)return`${i} `}return i}}function vD(e,t,n){if(xy(t)){const r=t.resolve(e),i=n&&r&&n.get(r);return i?i.count*i.aliasCount:0}else if(cu(t)){let r=0;for(const i of t.items){const a=vD(e,i,n);a>r&&(r=a)}return r}else if(Ha(t)){const r=vD(e,t.key,n),i=vD(e,t.value,n);return Math.max(r,i)}return 1}const age=e=>!e||typeof e!="function"&&typeof e!="object";class sr extends sL{constructor(t){super(P0),this.value=t}toJSON(t,n){return n!=null&&n.keep?this.value:Wd(this.value,t,n)}toString(){return String(this.value)}}sr.BLOCK_FOLDED="BLOCK_FOLDED";sr.BLOCK_LITERAL="BLOCK_LITERAL";sr.PLAIN="PLAIN";sr.QUOTE_DOUBLE="QUOTE_DOUBLE";sr.QUOTE_SINGLE="QUOTE_SINGLE";const tBe="tag:yaml.org,2002:";function nBe(e,t,n){if(t){const r=n.filter(a=>a.tag===t),i=r.find(a=>!a.format)??r[0];if(!i)throw new Error(`Tag ${t} not found`);return i}return n.find(r=>{var i;return((i=r.identify)==null?void 0:i.call(r,e))&&!r.format})}function Ux(e,t,n){var d,h,m;if(k1(e)&&(e=e.contents),hu(e))return e;if(Ha(e)){const y=(h=(d=n.schema[Hv]).createNode)==null?void 0:h.call(d,n.schema,null,n);return y.items.push(e),y}(e instanceof String||e instanceof Number||e instanceof Boolean||typeof BigInt<"u"&&e instanceof BigInt)&&(e=e.valueOf());const{aliasDuplicateObjects:r,onAnchor:i,onTagObj:a,schema:u,sourceObjects:l}=n;let o;if(r&&e&&typeof e=="object"){if(o=l.get(e),o)return o.anchor??(o.anchor=i(e)),new kw(o.anchor);o={anchor:null,node:null},l.set(e,o)}t!=null&&t.startsWith("!!")&&(t=tBe+t.slice(2));let s=nBe(e,t,u.tags);if(!s){if(e&&typeof e.toJSON=="function"&&(e=e.toJSON()),!e||typeof e!="object"){const y=new sr(e);return o&&(o.node=y),y}s=e instanceof Map?u[Hv]:Symbol.iterator in Object(e)?u[o4]:u[Hv]}a&&(a(s),delete n.onTagObj);const c=s!=null&&s.createNode?s.createNode(n.schema,e,n):typeof((m=s==null?void 0:s.nodeClass)==null?void 0:m.from)=="function"?s.nodeClass.from(n.schema,e,n):new sr(e);return t?c.tag=t:s.default||(c.tag=s.tag),o&&(o.node=c),c}function I$(e,t,n){let r=n;for(let i=t.length-1;i>=0;--i){const a=t[i];if(typeof a=="number"&&Number.isInteger(a)&&a>=0){const u=[];u[a]=r,r=u}else r=new Map([[a,r]])}return Ux(r,void 0,{aliasDuplicateObjects:!1,keepUndefined:!1,onAnchor:()=>{throw new Error("This should not happen, please report a bug.")},schema:e,sourceObjects:new Map})}const Q2=e=>e==null||typeof e=="object"&&!!e[Symbol.iterator]().next().done;class uge extends sL{constructor(t,n){super(t),Object.defineProperty(this,"schema",{value:n,configurable:!0,enumerable:!1,writable:!0})}clone(t){const n=Object.create(Object.getPrototypeOf(this),Object.getOwnPropertyDescriptors(this));return t&&(n.schema=t),n.items=n.items.map(r=>hu(r)||Ha(r)?r.clone(t):r),this.range&&(n.range=this.range.slice()),n}addIn(t,n){if(Q2(t))this.add(n);else{const[r,...i]=t,a=this.get(r,!0);if(cu(a))a.addIn(i,n);else if(a===void 0&&this.schema)this.set(r,I$(this.schema,i,n));else throw new Error(`Expected YAML collection at ${r}. Remaining path: ${i}`)}}deleteIn(t){const[n,...r]=t;if(r.length===0)return this.delete(n);const i=this.get(n,!0);if(cu(i))return i.deleteIn(r);throw new Error(`Expected YAML collection at ${n}. Remaining path: ${r}`)}getIn(t,n){const[r,...i]=t,a=this.get(r,!0);return i.length===0?!n&&fa(a)?a.value:a:cu(a)?a.getIn(i,n):void 0}hasAllNullValues(t){return this.items.every(n=>{if(!Ha(n))return!1;const r=n.value;return r==null||t&&fa(r)&&r.value==null&&!r.commentBefore&&!r.comment&&!r.tag})}hasIn(t){const[n,...r]=t;if(r.length===0)return this.has(n);const i=this.get(n,!0);return cu(i)?i.hasIn(r):!1}setIn(t,n){const[r,...i]=t;if(i.length===0)this.set(r,n);else{const a=this.get(r,!0);if(cu(a))a.setIn(i,n);else if(a===void 0&&this.schema)this.set(r,I$(this.schema,i,n));else throw new Error(`Expected YAML collection at ${r}. Remaining path: ${i}`)}}}const rBe=e=>e.replace(/^(?!$)(?: $)?/gm,"#");function eg(e,t){return/^\n+$/.test(e)?e.substring(1):t?e.replace(/^(?! *$)/gm,t):e}const qb=(e,t,n)=>e.endsWith(`
`)?eg(n,t):n.includes(`
`)?`
`+eg(n,t):(e.endsWith(" ")?"":" ")+n,lge="flow",BM="block",yD="quoted";function Pw(e,t,n="flow",{indentAtStart:r,lineWidth:i=80,minContentWidth:a=20,onFold:u,onOverflow:l}={}){if(!i||i<0)return e;i<a&&(a=0);const o=Math.max(1+a,1+i-t.length);if(e.length<=o)return e;const s=[],c={};let d=i-t.length;typeof r=="number"&&(r>i-Math.max(2,a)?s.push(0):d=i-r);let h,m,y=!1,v=-1,S=-1,_=-1;n===BM&&(v=VX(e,v,t.length),v!==-1&&(d=v+o));for(let $;$=e[v+=1];){if(n===yD&&$==="\\"){switch(S=v,e[v+1]){case"x":v+=3;break;case"u":v+=5;break;case"U":v+=9;break;default:v+=1}_=v}if($===`
`)n===BM&&(v=VX(e,v,t.length)),d=v+t.length+o,h=void 0;else{if($===" "&&m&&m!==" "&&m!==`
`&&m!=="	"){const k=e[v+1];k&&k!==" "&&k!==`
`&&k!=="	"&&(h=v)}if(v>=d)if(h)s.push(h),d=h+o,h=void 0;else if(n===yD){for(;m===" "||m==="	";)m=$,$=e[v+=1],y=!0;const k=v>_+1?v-2:S-1;if(c[k])return e;s.push(k),c[k]=!0,d=k+o,h=void 0}else y=!0}m=$}if(y&&l&&l(),s.length===0)return e;u&&u();let E=e.slice(0,s[0]);for(let $=0;$<s.length;++$){const k=s[$],A=s[$+1]||e.length;k===0?E=`
${t}${e.slice(0,A)}`:(n===yD&&c[k]&&(E+=`${e[k]}\\`),E+=`
${t}${e.slice(k+1,A)}`)}return E}function VX(e,t,n){let r=t,i=t+1,a=e[i];for(;a===" "||a==="	";)if(t<i+n)a=e[++t];else{do a=e[++t];while(a&&a!==`
`);r=t,i=t+1,a=e[i]}return r}const Mw=(e,t)=>({indentAtStart:t?e.indent.length:e.indentAtStart,lineWidth:e.options.lineWidth,minContentWidth:e.options.minContentWidth}),Fw=e=>/^(%|---|\.\.\.)/m.test(e);function iBe(e,t,n){if(!t||t<0)return!1;const r=t-n,i=e.length;if(i<=r)return!1;for(let a=0,u=0;a<i;++a)if(e[a]===`
`){if(a-u>r)return!0;if(u=a+1,i-u<=r)return!1}return!0}function ex(e,t){const n=JSON.stringify(e);if(t.options.doubleQuotedAsJSON)return n;const{implicitKey:r}=t,i=t.options.doubleQuotedMinMultiLineLength,a=t.indent||(Fw(e)?"  ":"");let u="",l=0;for(let o=0,s=n[o];s;s=n[++o])if(s===" "&&n[o+1]==="\\"&&n[o+2]==="n"&&(u+=n.slice(l,o)+"\\ ",o+=1,l=o,s="\\"),s==="\\")switch(n[o+1]){case"u":{u+=n.slice(l,o);const c=n.substr(o+2,4);switch(c){case"0000":u+="\\0";break;case"0007":u+="\\a";break;case"000b":u+="\\v";break;case"001b":u+="\\e";break;case"0085":u+="\\N";break;case"00a0":u+="\\_";break;case"2028":u+="\\L";break;case"2029":u+="\\P";break;default:c.substr(0,2)==="00"?u+="\\x"+c.substr(2):u+=n.substr(o,6)}o+=5,l=o+1}break;case"n":if(r||n[o+2]==='"'||n.length<i)o+=1;else{for(u+=n.slice(l,o)+`

`;n[o+2]==="\\"&&n[o+3]==="n"&&n[o+4]!=='"';)u+=`
`,o+=2;u+=a,n[o+2]===" "&&(u+="\\"),o+=1,l=o+1}break;default:o+=1}return u=l?u+n.slice(l):n,r?u:Pw(u,a,yD,Mw(t,!1))}function QM(e,t){if(t.options.singleQuote===!1||t.implicitKey&&e.includes(`
`)||/[ \t]\n|\n[ \t]/.test(e))return ex(e,t);const n=t.indent||(Fw(e)?"  ":""),r="'"+e.replace(/'/g,"''").replace(/\n+/g,`$&
${n}`)+"'";return t.implicitKey?r:Pw(r,n,lge,Mw(t,!1))}function KO(e,t){const{singleQuote:n}=t.options;let r;if(n===!1)r=ex;else{const i=e.includes('"'),a=e.includes("'");i&&!a?r=QM:a&&!i?r=ex:r=n?QM:ex}return r(e,t)}let zM;try{zM=new RegExp(`(^|(?<!
))
+(?!
|$)`,"g")}catch{zM=/\n+(?!\n|$)/g}function bD({comment:e,type:t,value:n},r,i,a){const{blockQuote:u,commentString:l,lineWidth:o}=r.options;if(!u||/\n[\t ]+$/.test(n)||/^\s*$/.test(n))return KO(n,r);const s=r.indent||(r.forceBlockIndent||Fw(n)?"  ":""),c=u==="literal"?!0:u==="folded"||t===sr.BLOCK_FOLDED?!1:t===sr.BLOCK_LITERAL?!0:!iBe(n,o,s.length);if(!n)return c?`|
`:`>
`;let d,h;for(h=n.length;h>0;--h){const A=n[h-1];if(A!==`
`&&A!=="	"&&A!==" ")break}let m=n.substring(h);const y=m.indexOf(`
`);y===-1?d="-":n===m||y!==m.length-1?(d="+",a&&a()):d="",m&&(n=n.slice(0,-m.length),m[m.length-1]===`
`&&(m=m.slice(0,-1)),m=m.replace(zM,`$&${s}`));let v=!1,S,_=-1;for(S=0;S<n.length;++S){const A=n[S];if(A===" ")v=!0;else if(A===`
`)_=S;else break}let E=n.substring(0,_<S?_+1:S);E&&(n=n.substring(E.length),E=E.replace(/\n+/g,`$&${s}`));let k=(v?s?"2":"1":"")+d;if(e&&(k+=" "+l(e.replace(/ ?[\r\n]+/g," ")),i&&i()),!c){const A=n.replace(/\n+/g,`
$&`).replace(/(?:^|\n)([\t ].*)(?:([\n\t ]*)\n(?![\n\t ]))?/g,"$1$2").replace(/\n+/g,`$&${s}`);let w=!1;const R=Mw(r,!0);u!=="folded"&&t!==sr.BLOCK_FOLDED&&(R.onOverflow=()=>{w=!0});const F=Pw(`${E}${A}${m}`,s,BM,R);if(!w)return`>${k}
${s}${F}`}return n=n.replace(/\n+/g,`$&${s}`),`|${k}
${s}${E}${n}${m}`}function aBe(e,t,n,r){const{type:i,value:a}=e,{actualString:u,implicitKey:l,indent:o,indentStep:s,inFlow:c}=t;if(l&&a.includes(`
`)||c&&/[[\]{},]/.test(a))return KO(a,t);if(/^[\n\t ,[\]{}#&*!|>'"%@`]|^[?-]$|^[?-][ \t]|[\n:][ \t]|[ \t]\n|[\n\t ]#|[\n\t :]$/.test(a))return l||c||!a.includes(`
`)?KO(a,t):bD(e,t,n,r);if(!l&&!c&&i!==sr.PLAIN&&a.includes(`
`))return bD(e,t,n,r);if(Fw(a)){if(o==="")return t.forceBlockIndent=!0,bD(e,t,n,r);if(l&&o===s)return KO(a,t)}const d=a.replace(/\n+/g,`$&
${o}`);if(u){const h=v=>{var S;return v.default&&v.tag!=="tag:yaml.org,2002:str"&&((S=v.test)==null?void 0:S.test(d))},{compat:m,tags:y}=t.doc.schema;if(y.some(h)||m!=null&&m.some(h))return KO(a,t)}return l?d:Pw(d,o,lge,Mw(t,!1))}function NS(e,t,n,r){const{implicitKey:i,inFlow:a}=t,u=typeof e.value=="string"?e:Object.assign({},e,{value:String(e.value)});let{type:l}=e;l!==sr.QUOTE_DOUBLE&&/[\x00-\x08\x0b-\x1f\x7f-\x9f\u{D800}-\u{DFFF}]/u.test(u.value)&&(l=sr.QUOTE_DOUBLE);const o=c=>{switch(c){case sr.BLOCK_FOLDED:case sr.BLOCK_LITERAL:return i||a?KO(u.value,t):bD(u,t,n,r);case sr.QUOTE_DOUBLE:return ex(u.value,t);case sr.QUOTE_SINGLE:return QM(u.value,t);case sr.PLAIN:return aBe(u,t,n,r);default:return null}};let s=o(l);if(s===null){const{defaultKeyType:c,defaultStringType:d}=t.options,h=i&&c||d;if(s=o(h),s===null)throw new Error(`Unsupported default string type ${h}`)}return s}function oge(e,t){const n=Object.assign({blockQuote:!0,commentString:rBe,defaultKeyType:null,defaultStringType:"PLAIN",directives:null,doubleQuotedAsJSON:!1,doubleQuotedMinMultiLineLength:40,falseStr:"false",flowCollectionPadding:!0,indentSeq:!0,lineWidth:80,minContentWidth:20,nullStr:"null",simpleKeys:!1,singleQuote:null,trueStr:"true",verifyAliasOrder:!0},e.schema.toStringOptions,t);let r;switch(n.collectionStyle){case"block":r=!1;break;case"flow":r=!0;break;default:r=null}return{anchors:new Set,doc:e,flowCollectionPadding:n.flowCollectionPadding?" ":"",indent:"",indentStep:typeof n.indent=="number"?" ".repeat(n.indent):"  ",inFlow:r,options:n}}function uBe(e,t){var i;if(t.tag){const a=e.filter(u=>u.tag===t.tag);if(a.length>0)return a.find(u=>u.format===t.format)??a[0]}let n,r;if(fa(t)){r=t.value;let a=e.filter(u=>{var l;return(l=u.identify)==null?void 0:l.call(u,r)});if(a.length>1){const u=a.filter(l=>l.test);u.length>0&&(a=u)}n=a.find(u=>u.format===t.format)??a.find(u=>!u.format)}else r=t,n=e.find(a=>a.nodeClass&&r instanceof a.nodeClass);if(!n){const a=((i=r==null?void 0:r.constructor)==null?void 0:i.name)??(r===null?"null":typeof r);throw new Error(`Tag not resolved for ${a} value`)}return n}function lBe(e,t,{anchors:n,doc:r}){if(!r.directives)return"";const i=[],a=(fa(e)||cu(e))&&e.anchor;a&&nge(a)&&(n.add(a),i.push(`&${a}`));const u=e.tag??(t.default?null:t.tag);return u&&i.push(r.directives.tagString(u)),i.join(" ")}function I3(e,t,n,r){var o;if(Ha(e))return e.toString(t,n,r);if(xy(e)){if(t.doc.directives)return e.toString(t);if((o=t.resolvedAliases)!=null&&o.has(e))throw new TypeError("Cannot stringify circular structure without alias nodes");t.resolvedAliases?t.resolvedAliases.add(e):t.resolvedAliases=new Set([e]),e=e.resolve(t.doc)}let i;const a=hu(e)?e:t.doc.createNode(e,{onTagObj:s=>i=s});i??(i=uBe(t.doc.schema.tags,a));const u=lBe(a,i,t);u.length>0&&(t.indentAtStart=(t.indentAtStart??0)+u.length+1);const l=typeof i.stringify=="function"?i.stringify(a,t,n,r):fa(a)?NS(a,t,n,r):a.toString(t,n,r);return u?fa(a)||l[0]==="{"||l[0]==="["?`${u} ${l}`:`${u}
${t.indent}${l}`:l}function oBe({key:e,value:t},n,r,i){const{allNullValues:a,doc:u,indent:l,indentStep:o,options:{commentString:s,indentSeq:c,simpleKeys:d}}=n;let h=hu(e)&&e.comment||null;if(d){if(h)throw new Error("With simple keys, key nodes cannot have comments");if(cu(e)||!hu(e)&&typeof e=="object"){const R="With simple keys, collection cannot be used as a key value";throw new Error(R)}}let m=!d&&(!e||h&&t==null&&!n.inFlow||cu(e)||(fa(e)?e.type===sr.BLOCK_FOLDED||e.type===sr.BLOCK_LITERAL:typeof e=="object"));n=Object.assign({},n,{allNullValues:!1,implicitKey:!m&&(d||!a),indent:l+o});let y=!1,v=!1,S=I3(e,n,()=>y=!0,()=>v=!0);if(!m&&!n.inFlow&&S.length>1024){if(d)throw new Error("With simple keys, single line scalar must not span more than 1024 characters");m=!0}if(n.inFlow){if(a||t==null)return y&&r&&r(),S===""?"?":m?`? ${S}`:S}else if(a&&!d||t==null&&m)return S=`? ${S}`,h&&!y?S+=qb(S,n.indent,s(h)):v&&i&&i(),S;y&&(h=null),m?(h&&(S+=qb(S,n.indent,s(h))),S=`? ${S}
${l}:`):(S=`${S}:`,h&&(S+=qb(S,n.indent,s(h))));let _,E,$;hu(t)?(_=!!t.spaceBefore,E=t.commentBefore,$=t.comment):(_=!1,E=null,$=null,t&&typeof t=="object"&&(t=u.createNode(t))),n.implicitKey=!1,!m&&!h&&fa(t)&&(n.indentAtStart=S.length+1),v=!1,!c&&o.length>=2&&!n.inFlow&&!m&&c4(t)&&!t.flow&&!t.tag&&!t.anchor&&(n.indent=n.indent.substring(2));let k=!1;const A=I3(t,n,()=>k=!0,()=>v=!0);let w=" ";if(h||_||E){if(w=_?`
`:"",E){const R=s(E);w+=`
${eg(R,n.indent)}`}A===""&&!n.inFlow?w===`
`&&(w=`

`):w+=`
${n.indent}`}else if(!m&&cu(t)){const R=A[0],F=A.indexOf(`
`),I=F!==-1,U=n.inFlow??t.flow??t.items.length===0;if(I||!U){let q=!1;if(I&&(R==="&"||R==="!")){let V=A.indexOf(" ");R==="&"&&V!==-1&&V<F&&A[V+1]==="!"&&(V=A.indexOf(" ",V+1)),(V===-1||F<V)&&(q=!0)}q||(w=`
${n.indent}`)}}else(A===""||A[0]===`
`)&&(w="");return S+=w+A,n.inFlow?k&&r&&r():$&&!k?S+=qb(S,n.indent,s($)):v&&i&&i(),S}function sge(e,t){(e==="debug"||e==="warn")&&console.warn(t)}const DE="<<",ug={identify:e=>e===DE||typeof e=="symbol"&&e.description===DE,default:"key",tag:"tag:yaml.org,2002:merge",test:/^<<$/,resolve:()=>Object.assign(new sr(Symbol(DE)),{addToJSMap:cge}),stringify:()=>DE},sBe=(e,t)=>(ug.identify(t)||fa(t)&&(!t.type||t.type===sr.PLAIN)&&ug.identify(t.value))&&(e==null?void 0:e.doc.schema.tags.some(n=>n.tag===ug.tag&&n.default));function cge(e,t,n){if(n=e&&xy(n)?n.resolve(e.doc):n,c4(n))for(const r of n.items)AP(e,t,r);else if(Array.isArray(n))for(const r of n)AP(e,t,r);else AP(e,t,n)}function AP(e,t,n){const r=e&&xy(n)?n.resolve(e.doc):n;if(!s4(r))throw new Error("Merge sources must be maps or map aliases");const i=r.toJSON(null,e,Map);for(const[a,u]of i)t instanceof Map?t.has(a)||t.set(a,u):t instanceof Set?t.add(a):Object.prototype.hasOwnProperty.call(t,a)||Object.defineProperty(t,a,{value:u,writable:!0,enumerable:!0,configurable:!0});return t}function fge(e,t,{key:n,value:r}){if(hu(n)&&n.addToJSMap)n.addToJSMap(e,t,r);else if(sBe(e,n))cge(e,t,r);else{const i=Wd(n,"",e);if(t instanceof Map)t.set(i,Wd(r,i,e));else if(t instanceof Set)t.add(i);else{const a=cBe(n,i,e),u=Wd(r,a,e);a in t?Object.defineProperty(t,a,{value:u,writable:!0,enumerable:!0,configurable:!0}):t[a]=u}}return t}function cBe(e,t,n){if(t===null)return"";if(typeof t!="object")return String(t);if(hu(e)&&(n!=null&&n.doc)){const r=oge(n.doc,{});r.anchors=new Set;for(const a of n.anchors.keys())r.anchors.add(a.anchor);r.inFlow=!0,r.inStringifyKey=!0;const i=e.toString(r);if(!n.mapKeyWarned){let a=JSON.stringify(i);a.length>40&&(a=a.substring(0,36)+'..."'),sge(n.doc.options.logLevel,`Keys with collection values will be stringified due to JS Object restrictions: ${a}. Set mapAsMap: true to use object keys.`),n.mapKeyWarned=!0}return i}return JSON.stringify(t)}function cL(e,t,n){const r=Ux(e,void 0,n),i=Ux(t,void 0,n);return new Os(r,i)}class Os{constructor(t,n=null){Object.defineProperty(this,eh,{value:Gme}),this.key=t,this.value=n}clone(t){let{key:n,value:r}=this;return hu(n)&&(n=n.clone(t)),hu(r)&&(r=r.clone(t)),new Os(n,r)}toJSON(t,n){const r=n!=null&&n.mapAsMap?new Map:{};return fge(n,r,this)}toString(t,n,r){return t!=null&&t.doc?oBe(this,t,n,r):JSON.stringify(this)}}function dge(e,t,n){return(t.inFlow??e.flow?dBe:fBe)(e,t,n)}function fBe({comment:e,items:t},n,{blockItemPrefix:r,flowChars:i,itemIndent:a,onChompKeep:u,onComment:l}){const{indent:o,options:{commentString:s}}=n,c=Object.assign({},n,{indent:a,type:null});let d=!1;const h=[];for(let y=0;y<t.length;++y){const v=t[y];let S=null;if(hu(v))!d&&v.spaceBefore&&h.push(""),L$(n,h,v.commentBefore,d),v.comment&&(S=v.comment);else if(Ha(v)){const E=hu(v.key)?v.key:null;E&&(!d&&E.spaceBefore&&h.push(""),L$(n,h,E.commentBefore,d))}d=!1;let _=I3(v,c,()=>S=null,()=>d=!0);S&&(_+=qb(_,a,s(S))),d&&S&&(d=!1),h.push(r+_)}let m;if(h.length===0)m=i.start+i.end;else{m=h[0];for(let y=1;y<h.length;++y){const v=h[y];m+=v?`
${o}${v}`:`
`}}return e?(m+=`
`+eg(s(e),o),l&&l()):d&&u&&u(),m}function dBe({items:e},t,{flowChars:n,itemIndent:r}){const{indent:i,indentStep:a,flowCollectionPadding:u,options:{commentString:l}}=t;r+=a;const o=Object.assign({},t,{indent:r,inFlow:!0,type:null});let s=!1,c=0;const d=[];for(let y=0;y<e.length;++y){const v=e[y];let S=null;if(hu(v))v.spaceBefore&&d.push(""),L$(t,d,v.commentBefore,!1),v.comment&&(S=v.comment);else if(Ha(v)){const E=hu(v.key)?v.key:null;E&&(E.spaceBefore&&d.push(""),L$(t,d,E.commentBefore,!1),E.comment&&(s=!0));const $=hu(v.value)?v.value:null;$?($.comment&&(S=$.comment),$.commentBefore&&(s=!0)):v.value==null&&(E!=null&&E.comment)&&(S=E.comment)}S&&(s=!0);let _=I3(v,o,()=>S=null);y<e.length-1&&(_+=","),S&&(_+=qb(_,r,l(S))),!s&&(d.length>c||_.includes(`
`))&&(s=!0),d.push(_),c=d.length}const{start:h,end:m}=n;if(d.length===0)return h+m;if(!s){const y=d.reduce((v,S)=>v+S.length+2,2);s=t.options.lineWidth>0&&y>t.options.lineWidth}if(s){let y=h;for(const v of d)y+=v?`
${a}${i}${v}`:`
`;return`${y}
${i}${m}`}else return`${h}${u}${d.join(" ")}${u}${m}`}function L$({indent:e,options:{commentString:t}},n,r,i){if(r&&i&&(r=r.replace(/^\n+/,"")),r){const a=eg(t(r),e);n.push(a.trimStart())}}function Vb(e,t){const n=fa(t)?t.value:t;for(const r of e)if(Ha(r)&&(r.key===t||r.key===n||fa(r.key)&&r.key.value===n))return r}class qf extends uge{static get tagName(){return"tag:yaml.org,2002:map"}constructor(t){super(Hv,t),this.items=[]}static from(t,n,r){const{keepUndefined:i,replacer:a}=r,u=new this(t),l=(o,s)=>{if(typeof a=="function")s=a.call(n,o,s);else if(Array.isArray(a)&&!a.includes(o))return;(s!==void 0||i)&&u.items.push(cL(o,s,r))};if(n instanceof Map)for(const[o,s]of n)l(o,s);else if(n&&typeof n=="object")for(const o of Object.keys(n))l(o,n[o]);return typeof t.sortMapEntries=="function"&&u.items.sort(t.sortMapEntries),u}add(t,n){var u;let r;Ha(t)?r=t:!t||typeof t!="object"||!("key"in t)?r=new Os(t,t==null?void 0:t.value):r=new Os(t.key,t.value);const i=Vb(this.items,r.key),a=(u=this.schema)==null?void 0:u.sortMapEntries;if(i){if(!n)throw new Error(`Key ${r.key} already set`);fa(i.value)&&age(r.value)?i.value.value=r.value:i.value=r.value}else if(a){const l=this.items.findIndex(o=>a(r,o)<0);l===-1?this.items.push(r):this.items.splice(l,0,r)}else this.items.push(r)}delete(t){const n=Vb(this.items,t);return n?this.items.splice(this.items.indexOf(n),1).length>0:!1}get(t,n){const r=Vb(this.items,t),i=r==null?void 0:r.value;return(!n&&fa(i)?i.value:i)??void 0}has(t){return!!Vb(this.items,t)}set(t,n){this.add(new Os(t,n),!0)}toJSON(t,n,r){const i=r?new r:n!=null&&n.mapAsMap?new Map:{};n!=null&&n.onCreate&&n.onCreate(i);for(const a of this.items)fge(n,i,a);return i}toString(t,n,r){if(!t)return JSON.stringify(this);for(const i of this.items)if(!Ha(i))throw new Error(`Map items must all be pairs; found ${JSON.stringify(i)} instead`);return!t.allNullValues&&this.hasAllNullValues(!1)&&(t=Object.assign({},t,{allNullValues:!0})),dge(this,t,{blockItemPrefix:"",flowChars:{start:"{",end:"}"},itemIndent:t.indent||"",onChompKeep:r,onComment:n})}}const f4={collection:"map",default:!0,nodeClass:qf,tag:"tag:yaml.org,2002:map",resolve(e,t){return s4(e)||t("Expected a mapping for this tag"),e},createNode:(e,t,n)=>qf.from(e,t,n)};class ly extends uge{static get tagName(){return"tag:yaml.org,2002:seq"}constructor(t){super(o4,t),this.items=[]}add(t){this.items.push(t)}delete(t){const n=$E(t);return typeof n!="number"?!1:this.items.splice(n,1).length>0}get(t,n){const r=$E(t);if(typeof r!="number")return;const i=this.items[r];return!n&&fa(i)?i.value:i}has(t){const n=$E(t);return typeof n=="number"&&n<this.items.length}set(t,n){const r=$E(t);if(typeof r!="number")throw new Error(`Expected a valid index, not ${t}.`);const i=this.items[r];fa(i)&&age(n)?i.value=n:this.items[r]=n}toJSON(t,n){const r=[];n!=null&&n.onCreate&&n.onCreate(r);let i=0;for(const a of this.items)r.push(Wd(a,String(i++),n));return r}toString(t,n,r){return t?dge(this,t,{blockItemPrefix:"- ",flowChars:{start:"[",end:"]"},itemIndent:(t.indent||"")+"  ",onChompKeep:r,onComment:n}):JSON.stringify(this)}static from(t,n,r){const{replacer:i}=r,a=new this(t);if(n&&Symbol.iterator in Object(n)){let u=0;for(let l of n){if(typeof i=="function"){const o=n instanceof Set?l:String(u++);l=i.call(n,o,l)}a.items.push(Ux(l,void 0,r))}}return a}}function $E(e){let t=fa(e)?e.value:e;return t&&typeof t=="string"&&(t=Number(t)),typeof t=="number"&&Number.isInteger(t)&&t>=0?t:null}const d4={collection:"seq",default:!0,nodeClass:ly,tag:"tag:yaml.org,2002:seq",resolve(e,t){return c4(e)||t("Expected a sequence for this tag"),e},createNode:(e,t,n)=>ly.from(e,t,n)},Nw={identify:e=>typeof e=="string",default:!0,tag:"tag:yaml.org,2002:str",resolve:e=>e,stringify(e,t,n,r){return t=Object.assign({actualString:!0},t),NS(e,t,n,r)}},Iw={identify:e=>e==null,createNode:()=>new sr(null),default:!0,tag:"tag:yaml.org,2002:null",test:/^(?:~|[Nn]ull|NULL)?$/,resolve:()=>new sr(null),stringify:({source:e},t)=>typeof e=="string"&&Iw.test.test(e)?e:t.options.nullStr},fL={identify:e=>typeof e=="boolean",default:!0,tag:"tag:yaml.org,2002:bool",test:/^(?:[Tt]rue|TRUE|[Ff]alse|FALSE)$/,resolve:e=>new sr(e[0]==="t"||e[0]==="T"),stringify({source:e,value:t},n){if(e&&fL.test.test(e)){const r=e[0]==="t"||e[0]==="T";if(t===r)return e}return t?n.options.trueStr:n.options.falseStr}};function yp({format:e,minFractionDigits:t,tag:n,value:r}){if(typeof r=="bigint")return String(r);const i=typeof r=="number"?r:Number(r);if(!isFinite(i))return isNaN(i)?".nan":i<0?"-.inf":".inf";let a=JSON.stringify(r);if(!e&&t&&(!n||n==="tag:yaml.org,2002:float")&&/^\d/.test(a)){let u=a.indexOf(".");u<0&&(u=a.length,a+=".");let l=t-(a.length-u-1);for(;l-- >0;)a+="0"}return a}const hge={identify:e=>typeof e=="number",default:!0,tag:"tag:yaml.org,2002:float",test:/^(?:[-+]?\.(?:inf|Inf|INF)|\.nan|\.NaN|\.NAN)$/,resolve:e=>e.slice(-3).toLowerCase()==="nan"?NaN:e[0]==="-"?Number.NEGATIVE_INFINITY:Number.POSITIVE_INFINITY,stringify:yp},pge={identify:e=>typeof e=="number",default:!0,tag:"tag:yaml.org,2002:float",format:"EXP",test:/^[-+]?(?:\.[0-9]+|[0-9]+(?:\.[0-9]*)?)[eE][-+]?[0-9]+$/,resolve:e=>parseFloat(e),stringify(e){const t=Number(e.value);return isFinite(t)?t.toExponential():yp(e)}},mge={identify:e=>typeof e=="number",default:!0,tag:"tag:yaml.org,2002:float",test:/^[-+]?(?:\.[0-9]+|[0-9]+\.[0-9]*)$/,resolve(e){const t=new sr(parseFloat(e)),n=e.indexOf(".");return n!==-1&&e[e.length-1]==="0"&&(t.minFractionDigits=e.length-n-1),t},stringify:yp},Lw=e=>typeof e=="bigint"||Number.isInteger(e),dL=(e,t,n,{intAsBigInt:r})=>r?BigInt(e):parseInt(e.substring(t),n);function gge(e,t,n){const{value:r}=e;return Lw(r)&&r>=0?n+r.toString(t):yp(e)}const vge={identify:e=>Lw(e)&&e>=0,default:!0,tag:"tag:yaml.org,2002:int",format:"OCT",test:/^0o[0-7]+$/,resolve:(e,t,n)=>dL(e,2,8,n),stringify:e=>gge(e,8,"0o")},yge={identify:Lw,default:!0,tag:"tag:yaml.org,2002:int",test:/^[-+]?[0-9]+$/,resolve:(e,t,n)=>dL(e,0,10,n),stringify:yp},bge={identify:e=>Lw(e)&&e>=0,default:!0,tag:"tag:yaml.org,2002:int",format:"HEX",test:/^0x[0-9a-fA-F]+$/,resolve:(e,t,n)=>dL(e,2,16,n),stringify:e=>gge(e,16,"0x")},hBe=[f4,d4,Nw,Iw,fL,vge,yge,bge,hge,pge,mge];function UX(e){return typeof e=="bigint"||Number.isInteger(e)}const CE=({value:e})=>JSON.stringify(e),pBe=[{identify:e=>typeof e=="string",default:!0,tag:"tag:yaml.org,2002:str",resolve:e=>e,stringify:CE},{identify:e=>e==null,createNode:()=>new sr(null),default:!0,tag:"tag:yaml.org,2002:null",test:/^null$/,resolve:()=>null,stringify:CE},{identify:e=>typeof e=="boolean",default:!0,tag:"tag:yaml.org,2002:bool",test:/^true$|^false$/,resolve:e=>e==="true",stringify:CE},{identify:UX,default:!0,tag:"tag:yaml.org,2002:int",test:/^-?(?:0|[1-9][0-9]*)$/,resolve:(e,t,{intAsBigInt:n})=>n?BigInt(e):parseInt(e,10),stringify:({value:e})=>UX(e)?e.toString():JSON.stringify(e)},{identify:e=>typeof e=="number",default:!0,tag:"tag:yaml.org,2002:float",test:/^-?(?:0|[1-9][0-9]*)(?:\.[0-9]*)?(?:[eE][-+]?[0-9]+)?$/,resolve:e=>parseFloat(e),stringify:CE}],mBe={default:!0,tag:"",test:/^/,resolve(e,t){return t(`Unresolved plain scalar ${JSON.stringify(e)}`),e}},gBe=[f4,d4].concat(pBe,mBe),hL={identify:e=>e instanceof Uint8Array,default:!1,tag:"tag:yaml.org,2002:binary",resolve(e,t){if(typeof atob=="function"){const n=atob(e.replace(/[\n\r]/g,"")),r=new Uint8Array(n.length);for(let i=0;i<n.length;++i)r[i]=n.charCodeAt(i);return r}else return t("This environment does not support reading binary tags; either Buffer or atob is required"),e},stringify({comment:e,type:t,value:n},r,i,a){if(!n)return"";const u=n;let l;if(typeof btoa=="function"){let o="";for(let s=0;s<u.length;++s)o+=String.fromCharCode(u[s]);l=btoa(o)}else throw new Error("This environment does not support writing binary tags; either Buffer or btoa is required");if(t??(t=sr.BLOCK_LITERAL),t!==sr.QUOTE_DOUBLE){const o=Math.max(r.options.lineWidth-r.indent.length,r.options.minContentWidth),s=Math.ceil(l.length/o),c=new Array(s);for(let d=0,h=0;d<s;++d,h+=o)c[d]=l.substr(h,o);l=c.join(t===sr.BLOCK_LITERAL?`
`:" ")}return NS({comment:e,type:t,value:l},r,i,a)}};function Oge(e,t){if(c4(e))for(let n=0;n<e.items.length;++n){let r=e.items[n];if(!Ha(r)){if(s4(r)){r.items.length>1&&t("Each pair must have its own sequence indicator");const i=r.items[0]||new Os(new sr(null));if(r.commentBefore&&(i.key.commentBefore=i.key.commentBefore?`${r.commentBefore}
${i.key.commentBefore}`:r.commentBefore),r.comment){const a=i.value??i.key;a.comment=a.comment?`${r.comment}
${a.comment}`:r.comment}r=i}e.items[n]=Ha(r)?r:new Os(r)}}else t("Expected a sequence for this tag");return e}function xge(e,t,n){const{replacer:r}=n,i=new ly(e);i.tag="tag:yaml.org,2002:pairs";let a=0;if(t&&Symbol.iterator in Object(t))for(let u of t){typeof r=="function"&&(u=r.call(t,String(a++),u));let l,o;if(Array.isArray(u))if(u.length===2)l=u[0],o=u[1];else throw new TypeError(`Expected [key, value] tuple: ${u}`);else if(u&&u instanceof Object){const s=Object.keys(u);if(s.length===1)l=s[0],o=u[l];else throw new TypeError(`Expected tuple with one key, not ${s.length} keys`)}else l=u;i.items.push(cL(l,o,n))}return i}const pL={collection:"seq",default:!1,tag:"tag:yaml.org,2002:pairs",resolve:Oge,createNode:xge};class b3 extends ly{constructor(){super(),this.add=qf.prototype.add.bind(this),this.delete=qf.prototype.delete.bind(this),this.get=qf.prototype.get.bind(this),this.has=qf.prototype.has.bind(this),this.set=qf.prototype.set.bind(this),this.tag=b3.tag}toJSON(t,n){if(!n)return super.toJSON(t);const r=new Map;n!=null&&n.onCreate&&n.onCreate(r);for(const i of this.items){let a,u;if(Ha(i)?(a=Wd(i.key,"",n),u=Wd(i.value,a,n)):a=Wd(i,"",n),r.has(a))throw new Error("Ordered maps must not include duplicate keys");r.set(a,u)}return r}static from(t,n,r){const i=xge(t,n,r),a=new this;return a.items=i.items,a}}b3.tag="tag:yaml.org,2002:omap";const mL={collection:"seq",identify:e=>e instanceof Map,nodeClass:b3,default:!1,tag:"tag:yaml.org,2002:omap",resolve(e,t){const n=Oge(e,t),r=[];for(const{key:i}of n.items)fa(i)&&(r.includes(i.value)?t(`Ordered maps must not include duplicate keys: ${i.value}`):r.push(i.value));return Object.assign(new b3,n)},createNode:(e,t,n)=>b3.from(e,t,n)};function Sge({value:e,source:t},n){return t&&(e?_ge:Ege).test.test(t)?t:e?n.options.trueStr:n.options.falseStr}const _ge={identify:e=>e===!0,default:!0,tag:"tag:yaml.org,2002:bool",test:/^(?:Y|y|[Yy]es|YES|[Tt]rue|TRUE|[Oo]n|ON)$/,resolve:()=>new sr(!0),stringify:Sge},Ege={identify:e=>e===!1,default:!0,tag:"tag:yaml.org,2002:bool",test:/^(?:N|n|[Nn]o|NO|[Ff]alse|FALSE|[Oo]ff|OFF)$/,resolve:()=>new sr(!1),stringify:Sge},vBe={identify:e=>typeof e=="number",default:!0,tag:"tag:yaml.org,2002:float",test:/^(?:[-+]?\.(?:inf|Inf|INF)|\.nan|\.NaN|\.NAN)$/,resolve:e=>e.slice(-3).toLowerCase()==="nan"?NaN:e[0]==="-"?Number.NEGATIVE_INFINITY:Number.POSITIVE_INFINITY,stringify:yp},yBe={identify:e=>typeof e=="number",default:!0,tag:"tag:yaml.org,2002:float",format:"EXP",test:/^[-+]?(?:[0-9][0-9_]*)?(?:\.[0-9_]*)?[eE][-+]?[0-9]+$/,resolve:e=>parseFloat(e.replace(/_/g,"")),stringify(e){const t=Number(e.value);return isFinite(t)?t.toExponential():yp(e)}},bBe={identify:e=>typeof e=="number",default:!0,tag:"tag:yaml.org,2002:float",test:/^[-+]?(?:[0-9][0-9_]*)?\.[0-9_]*$/,resolve(e){const t=new sr(parseFloat(e.replace(/_/g,""))),n=e.indexOf(".");if(n!==-1){const r=e.substring(n+1).replace(/_/g,"");r[r.length-1]==="0"&&(t.minFractionDigits=r.length)}return t},stringify:yp},IS=e=>typeof e=="bigint"||Number.isInteger(e);function Bw(e,t,n,{intAsBigInt:r}){const i=e[0];if((i==="-"||i==="+")&&(t+=1),e=e.substring(t).replace(/_/g,""),r){switch(n){case 2:e=`0b${e}`;break;case 8:e=`0o${e}`;break;case 16:e=`0x${e}`;break}const u=BigInt(e);return i==="-"?BigInt(-1)*u:u}const a=parseInt(e,n);return i==="-"?-1*a:a}function gL(e,t,n){const{value:r}=e;if(IS(r)){const i=r.toString(t);return r<0?"-"+n+i.substr(1):n+i}return yp(e)}const OBe={identify:IS,default:!0,tag:"tag:yaml.org,2002:int",format:"BIN",test:/^[-+]?0b[0-1_]+$/,resolve:(e,t,n)=>Bw(e,2,2,n),stringify:e=>gL(e,2,"0b")},xBe={identify:IS,default:!0,tag:"tag:yaml.org,2002:int",format:"OCT",test:/^[-+]?0[0-7_]+$/,resolve:(e,t,n)=>Bw(e,1,8,n),stringify:e=>gL(e,8,"0")},SBe={identify:IS,default:!0,tag:"tag:yaml.org,2002:int",test:/^[-+]?[0-9][0-9_]*$/,resolve:(e,t,n)=>Bw(e,0,10,n),stringify:yp},_Be={identify:IS,default:!0,tag:"tag:yaml.org,2002:int",format:"HEX",test:/^[-+]?0x[0-9a-fA-F_]+$/,resolve:(e,t,n)=>Bw(e,2,16,n),stringify:e=>gL(e,16,"0x")};class O3 extends qf{constructor(t){super(t),this.tag=O3.tag}add(t){let n;Ha(t)?n=t:t&&typeof t=="object"&&"key"in t&&"value"in t&&t.value===null?n=new Os(t.key,null):n=new Os(t,null),Vb(this.items,n.key)||this.items.push(n)}get(t,n){const r=Vb(this.items,t);return!n&&Ha(r)?fa(r.key)?r.key.value:r.key:r}set(t,n){if(typeof n!="boolean")throw new Error(`Expected boolean value for set(key, value) in a YAML set, not ${typeof n}`);const r=Vb(this.items,t);r&&!n?this.items.splice(this.items.indexOf(r),1):!r&&n&&this.items.push(new Os(t))}toJSON(t,n){return super.toJSON(t,n,Set)}toString(t,n,r){if(!t)return JSON.stringify(this);if(this.hasAllNullValues(!0))return super.toString(Object.assign({},t,{allNullValues:!0}),n,r);throw new Error("Set items must all have null values")}static from(t,n,r){const{replacer:i}=r,a=new this(t);if(n&&Symbol.iterator in Object(n))for(let u of n)typeof i=="function"&&(u=i.call(n,u,u)),a.items.push(cL(u,null,r));return a}}O3.tag="tag:yaml.org,2002:set";const vL={collection:"map",identify:e=>e instanceof Set,nodeClass:O3,default:!1,tag:"tag:yaml.org,2002:set",createNode:(e,t,n)=>O3.from(e,t,n),resolve(e,t){if(s4(e)){if(e.hasAllNullValues(!0))return Object.assign(new O3,e);t("Set items must all have null values")}else t("Expected a mapping for this tag");return e}};function yL(e,t){const n=e[0],r=n==="-"||n==="+"?e.substring(1):e,i=u=>t?BigInt(u):Number(u),a=r.replace(/_/g,"").split(":").reduce((u,l)=>u*i(60)+i(l),i(0));return n==="-"?i(-1)*a:a}function Dge(e){let{value:t}=e,n=u=>u;if(typeof t=="bigint")n=u=>BigInt(u);else if(isNaN(t)||!isFinite(t))return yp(e);let r="";t<0&&(r="-",t*=n(-1));const i=n(60),a=[t%i];return t<60?a.unshift(0):(t=(t-a[0])/i,a.unshift(t%i),t>=60&&(t=(t-a[0])/i,a.unshift(t))),r+a.map(u=>String(u).padStart(2,"0")).join(":").replace(/000000\d*$/,"")}const $ge={identify:e=>typeof e=="bigint"||Number.isInteger(e),default:!0,tag:"tag:yaml.org,2002:int",format:"TIME",test:/^[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+$/,resolve:(e,t,{intAsBigInt:n})=>yL(e,n),stringify:Dge},Cge={identify:e=>typeof e=="number",default:!0,tag:"tag:yaml.org,2002:float",format:"TIME",test:/^[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\.[0-9_]*$/,resolve:e=>yL(e,!1),stringify:Dge},Qw={identify:e=>e instanceof Date,default:!0,tag:"tag:yaml.org,2002:timestamp",test:RegExp("^([0-9]{4})-([0-9]{1,2})-([0-9]{1,2})(?:(?:t|T|[ \\t]+)([0-9]{1,2}):([0-9]{1,2}):([0-9]{1,2}(\\.[0-9]+)?)(?:[ \\t]*(Z|[-+][012]?[0-9](?::[0-9]{2})?))?)?$"),resolve(e){const t=e.match(Qw.test);if(!t)throw new Error("!!timestamp expects a date, starting with yyyy-mm-dd");const[,n,r,i,a,u,l]=t.map(Number),o=t[7]?Number((t[7]+"00").substr(1,3)):0;let s=Date.UTC(n,r-1,i,a||0,u||0,l||0,o);const c=t[8];if(c&&c!=="Z"){let d=yL(c,!1);Math.abs(d)<30&&(d*=60),s-=6e4*d}return new Date(s)},stringify:({value:e})=>(e==null?void 0:e.toISOString().replace(/(T00:00:00)?\.000Z$/,""))??""},jX=[f4,d4,Nw,Iw,_ge,Ege,OBe,xBe,SBe,_Be,vBe,yBe,bBe,hL,ug,mL,pL,vL,$ge,Cge,Qw],ZX=new Map([["core",hBe],["failsafe",[f4,d4,Nw]],["json",gBe],["yaml11",jX],["yaml-1.1",jX]]),HX={binary:hL,bool:fL,float:mge,floatExp:pge,floatNaN:hge,floatTime:Cge,int:yge,intHex:bge,intOct:vge,intTime:$ge,map:f4,merge:ug,null:Iw,omap:mL,pairs:pL,seq:d4,set:vL,timestamp:Qw},EBe={"tag:yaml.org,2002:binary":hL,"tag:yaml.org,2002:merge":ug,"tag:yaml.org,2002:omap":mL,"tag:yaml.org,2002:pairs":pL,"tag:yaml.org,2002:set":vL,"tag:yaml.org,2002:timestamp":Qw};function TP(e,t,n){const r=ZX.get(t);if(r&&!e)return n&&!r.includes(ug)?r.concat(ug):r.slice();let i=r;if(!i)if(Array.isArray(e))i=[];else{const a=Array.from(ZX.keys()).filter(u=>u!=="yaml11").map(u=>JSON.stringify(u)).join(", ");throw new Error(`Unknown schema "${t}"; use one of ${a} or define customTags array`)}if(Array.isArray(e))for(const a of e)i=i.concat(a);else typeof e=="function"&&(i=e(i.slice()));return n&&(i=i.concat(ug)),i.reduce((a,u)=>{const l=typeof u=="string"?HX[u]:u;if(!l){const o=JSON.stringify(u),s=Object.keys(HX).map(c=>JSON.stringify(c)).join(", ");throw new Error(`Unknown custom tag ${o}; use one of ${s}`)}return a.includes(l)||a.push(l),a},[])}const DBe=(e,t)=>e.key<t.key?-1:e.key>t.key?1:0;class zw{constructor({compat:t,customTags:n,merge:r,resolveKnownTags:i,schema:a,sortMapEntries:u,toStringDefaults:l}){this.compat=Array.isArray(t)?TP(t,"compat"):t?TP(null,t):null,this.name=typeof a=="string"&&a||"core",this.knownTags=i?EBe:{},this.tags=TP(n,this.name,r),this.toStringOptions=l??null,Object.defineProperty(this,Hv,{value:f4}),Object.defineProperty(this,P0,{value:Nw}),Object.defineProperty(this,o4,{value:d4}),this.sortMapEntries=typeof u=="function"?u:u===!0?DBe:null}clone(){const t=Object.create(zw.prototype,Object.getOwnPropertyDescriptors(this));return t.tags=this.tags.slice(),t}}function $Be(e,t){var o;const n=[];let r=t.directives===!0;if(t.directives!==!1&&e.directives){const s=e.directives.toString(e);s?(n.push(s),r=!0):e.directives.docStart&&(r=!0)}r&&n.push("---");const i=oge(e,t),{commentString:a}=i.options;if(e.commentBefore){n.length!==1&&n.unshift("");const s=a(e.commentBefore);n.unshift(eg(s,""))}let u=!1,l=null;if(e.contents){if(hu(e.contents)){if(e.contents.spaceBefore&&r&&n.push(""),e.contents.commentBefore){const d=a(e.contents.commentBefore);n.push(eg(d,""))}i.forceBlockIndent=!!e.comment,l=e.contents.comment}const s=l?void 0:()=>u=!0;let c=I3(e.contents,i,()=>l=null,s);l&&(c+=qb(c,"",a(l))),(c[0]==="|"||c[0]===">")&&n[n.length-1]==="---"?n[n.length-1]=`--- ${c}`:n.push(c)}else n.push(I3(e.contents,i));if((o=e.directives)!=null&&o.docEnd)if(e.comment){const s=a(e.comment);s.includes(`
`)?(n.push("..."),n.push(eg(s,""))):n.push(`... ${s}`)}else n.push("...");else{let s=e.comment;s&&u&&(s=s.replace(/^\n+/,"")),s&&((!u||l)&&n[n.length-1]!==""&&n.push(""),n.push(eg(a(s),"")))}return n.join(`
`)+`
`}let qw=class wge{constructor(t,n,r){this.commentBefore=null,this.comment=null,this.errors=[],this.warnings=[],Object.defineProperty(this,eh,{value:LM});let i=null;typeof n=="function"||Array.isArray(n)?i=n:r===void 0&&n&&(r=n,n=void 0);const a=Object.assign({intAsBigInt:!1,keepSourceTokens:!1,logLevel:"warn",prettyErrors:!0,strict:!0,stringKeys:!1,uniqueKeys:!0,version:"1.2"},r);this.options=a;let{version:u}=a;r!=null&&r._directives?(this.directives=r._directives.atDocument(),this.directives.yaml.explicit&&(u=this.directives.yaml.version)):this.directives=new uc({version:u}),this.setSchema(u,r),this.contents=t===void 0?null:this.createNode(t,i,r)}clone(){const t=Object.create(wge.prototype,{[eh]:{value:LM}});return t.commentBefore=this.commentBefore,t.comment=this.comment,t.errors=this.errors.slice(),t.warnings=this.warnings.slice(),t.options=Object.assign({},this.options),this.directives&&(t.directives=this.directives.clone()),t.schema=this.schema.clone(),t.contents=hu(this.contents)?this.contents.clone(t.schema):this.contents,this.range&&(t.range=this.range.slice()),t}add(t){DO(this.contents)&&this.contents.add(t)}addIn(t,n){DO(this.contents)&&this.contents.addIn(t,n)}createAlias(t,n){if(!t.anchor){const r=rge(this);t.anchor=!n||r.has(n)?ige(n||"a",r):n}return new kw(t.anchor)}createNode(t,n,r){let i;if(typeof n=="function")t=n.call({"":t},"",t),i=n;else if(Array.isArray(n)){const S=E=>typeof E=="number"||E instanceof String||E instanceof Number,_=n.filter(S).map(String);_.length>0&&(n=n.concat(_)),i=n}else r===void 0&&n&&(r=n,n=void 0);const{aliasDuplicateObjects:a,anchorPrefix:u,flow:l,keepUndefined:o,onTagObj:s,tag:c}=r??{},{onAnchor:d,setAnchors:h,sourceObjects:m}=eBe(this,u||"a"),y={aliasDuplicateObjects:a??!0,keepUndefined:o??!1,onAnchor:d,onTagObj:s,replacer:i,schema:this.schema,sourceObjects:m},v=Ux(t,c,y);return l&&cu(v)&&(v.flow=!0),h(),v}createPair(t,n,r={}){const i=this.createNode(t,null,r),a=this.createNode(n,null,r);return new Os(i,a)}delete(t){return DO(this.contents)?this.contents.delete(t):!1}deleteIn(t){return Q2(t)?this.contents==null?!1:(this.contents=null,!0):DO(this.contents)?this.contents.deleteIn(t):!1}get(t,n){return cu(this.contents)?this.contents.get(t,n):void 0}getIn(t,n){return Q2(t)?!n&&fa(this.contents)?this.contents.value:this.contents:cu(this.contents)?this.contents.getIn(t,n):void 0}has(t){return cu(this.contents)?this.contents.has(t):!1}hasIn(t){return Q2(t)?this.contents!==void 0:cu(this.contents)?this.contents.hasIn(t):!1}set(t,n){this.contents==null?this.contents=I$(this.schema,[t],n):DO(this.contents)&&this.contents.set(t,n)}setIn(t,n){Q2(t)?this.contents=n:this.contents==null?this.contents=I$(this.schema,Array.from(t),n):DO(this.contents)&&this.contents.setIn(t,n)}setSchema(t,n={}){typeof t=="number"&&(t=String(t));let r;switch(t){case"1.1":this.directives?this.directives.yaml.version="1.1":this.directives=new uc({version:"1.1"}),r={resolveKnownTags:!1,schema:"yaml-1.1"};break;case"1.2":case"next":this.directives?this.directives.yaml.version=t:this.directives=new uc({version:t}),r={resolveKnownTags:!0,schema:"core"};break;case null:this.directives&&delete this.directives,r=null;break;default:{const i=JSON.stringify(t);throw new Error(`Expected '1.1', '1.2' or null as first argument, but found: ${i}`)}}if(n.schema instanceof Object)this.schema=n.schema;else if(r)this.schema=new zw(Object.assign(r,n));else throw new Error("With a null YAML version, the { schema: Schema } option is required")}toJS({json:t,jsonArg:n,mapAsMap:r,maxAliasCount:i,onAnchor:a,reviver:u}={}){const l={anchors:new Map,doc:this,keep:!t,mapAsMap:r===!0,mapKeyWarned:!1,maxAliasCount:typeof i=="number"?i:100},o=Wd(this.contents,n??"",l);if(typeof a=="function")for(const{count:s,res:c}of l.anchors.values())a(c,s);return typeof u=="function"?GO(u,{"":o},"",o):o}toJSON(t,n){return this.toJS({json:!0,jsonArg:t,mapAsMap:!1,onAnchor:n})}toString(t={}){if(this.errors.length>0)throw new Error("Document with errors cannot be stringified");if("indent"in t&&(!Number.isInteger(t.indent)||Number(t.indent)<=0)){const n=JSON.stringify(t.indent);throw new Error(`"indent" option must be a positive integer, not ${n}`)}return $Be(this,t)}};function DO(e){if(cu(e))return!0;throw new Error("Expected a YAML collection as document contents")}class bL extends Error{constructor(t,n,r,i){super(),this.name=t,this.code=r,this.message=i,this.pos=n}}class Ub extends bL{constructor(t,n,r){super("YAMLParseError",t,n,r)}}class Rge extends bL{constructor(t,n,r){super("YAMLWarning",t,n,r)}}const B$=(e,t)=>n=>{if(n.pos[0]===-1)return;n.linePos=n.pos.map(l=>t.linePos(l));const{line:r,col:i}=n.linePos[0];n.message+=` at line ${r}, column ${i}`;let a=i-1,u=e.substring(t.lineStarts[r-1],t.lineStarts[r]).replace(/[\n\r]+$/,"");if(a>=60&&u.length>80){const l=Math.min(a-39,u.length-79);u="…"+u.substring(l),a-=l-1}if(u.length>80&&(u=u.substring(0,79)+"…"),r>1&&/^ *$/.test(u.substring(0,a))){let l=e.substring(t.lineStarts[r-2],t.lineStarts[r-1]);l.length>80&&(l=l.substring(0,79)+`…
`),u=l+u}if(/[^ ]/.test(u)){let l=1;const o=n.linePos[1];o&&o.line===r&&o.col>i&&(l=Math.max(1,Math.min(o.col-i,80-a)));const s=" ".repeat(a)+"^".repeat(l);n.message+=`:

${u}
${s}
`}};function L3(e,{flow:t,indicator:n,next:r,offset:i,onError:a,parentIndent:u,startOnNewline:l}){let o=!1,s=l,c=l,d="",h="",m=!1,y=!1,v=null,S=null,_=null,E=null,$=null,k=null,A=null;for(const F of e)switch(y&&(F.type!=="space"&&F.type!=="newline"&&F.type!=="comma"&&a(F.offset,"MISSING_CHAR","Tags and anchors must be separated from the next token by white space"),y=!1),v&&(s&&F.type!=="comment"&&F.type!=="newline"&&a(v,"TAB_AS_INDENT","Tabs are not allowed as indentation"),v=null),F.type){case"space":!t&&(n!=="doc-start"||(r==null?void 0:r.type)!=="flow-collection")&&F.source.includes("	")&&(v=F),c=!0;break;case"comment":{c||a(F,"MISSING_CHAR","Comments must be separated from other tokens by white space characters");const I=F.source.substring(1)||" ";d?d+=h+I:d=I,h="",s=!1;break}case"newline":s?d?d+=F.source:(!k||n!=="seq-item-ind")&&(o=!0):h+=F.source,s=!0,m=!0,(S||_)&&(E=F),c=!0;break;case"anchor":S&&a(F,"MULTIPLE_ANCHORS","A node can have at most one anchor"),F.source.endsWith(":")&&a(F.offset+F.source.length-1,"BAD_ALIAS","Anchor ending in : is ambiguous",!0),S=F,A??(A=F.offset),s=!1,c=!1,y=!0;break;case"tag":{_&&a(F,"MULTIPLE_TAGS","A node can have at most one tag"),_=F,A??(A=F.offset),s=!1,c=!1,y=!0;break}case n:(S||_)&&a(F,"BAD_PROP_ORDER",`Anchors and tags must be after the ${F.source} indicator`),k&&a(F,"UNEXPECTED_TOKEN",`Unexpected ${F.source} in ${t??"collection"}`),k=F,s=n==="seq-item-ind"||n==="explicit-key-ind",c=!1;break;case"comma":if(t){$&&a(F,"UNEXPECTED_TOKEN",`Unexpected , in ${t}`),$=F,s=!1,c=!1;break}default:a(F,"UNEXPECTED_TOKEN",`Unexpected ${F.type} token`),s=!1,c=!1}const w=e[e.length-1],R=w?w.offset+w.source.length:i;return y&&r&&r.type!=="space"&&r.type!=="newline"&&r.type!=="comma"&&(r.type!=="scalar"||r.source!=="")&&a(r.offset,"MISSING_CHAR","Tags and anchors must be separated from the next token by white space"),v&&(s&&v.indent<=u||(r==null?void 0:r.type)==="block-map"||(r==null?void 0:r.type)==="block-seq")&&a(v,"TAB_AS_INDENT","Tabs are not allowed as indentation"),{comma:$,found:k,spaceBefore:o,comment:d,hasNewline:m,anchor:S,tag:_,newlineAfterProp:E,end:R,start:A??R}}function jx(e){if(!e)return null;switch(e.type){case"alias":case"scalar":case"double-quoted-scalar":case"single-quoted-scalar":if(e.source.includes(`
`))return!0;if(e.end){for(const t of e.end)if(t.type==="newline")return!0}return!1;case"flow-collection":for(const t of e.items){for(const n of t.start)if(n.type==="newline")return!0;if(t.sep){for(const n of t.sep)if(n.type==="newline")return!0}if(jx(t.key)||jx(t.value))return!0}return!1;default:return!0}}function qM(e,t,n){if((t==null?void 0:t.type)==="flow-collection"){const r=t.end[0];r.indent===e&&(r.source==="]"||r.source==="}")&&jx(t)&&n(r,"BAD_INDENT","Flow end indicator should be more indented than parent",!0)}}function Age(e,t,n){const{uniqueKeys:r}=e.options;if(r===!1)return!1;const i=typeof r=="function"?r:(a,u)=>a===u||fa(a)&&fa(u)&&a.value===u.value;return t.some(a=>i(a.key,n))}const XX="All mapping items must start at the same column";function CBe({composeNode:e,composeEmptyNode:t},n,r,i,a){var c;const u=(a==null?void 0:a.nodeClass)??qf,l=new u(n.schema);n.atRoot&&(n.atRoot=!1);let o=r.offset,s=null;for(const d of r.items){const{start:h,key:m,sep:y,value:v}=d,S=L3(h,{indicator:"explicit-key-ind",next:m??(y==null?void 0:y[0]),offset:o,onError:i,parentIndent:r.indent,startOnNewline:!0}),_=!S.found;if(_){if(m&&(m.type==="block-seq"?i(o,"BLOCK_AS_IMPLICIT_KEY","A block sequence may not be used as an implicit map key"):"indent"in m&&m.indent!==r.indent&&i(o,"BAD_INDENT",XX)),!S.anchor&&!S.tag&&!y){s=S.end,S.comment&&(l.comment?l.comment+=`
`+S.comment:l.comment=S.comment);continue}(S.newlineAfterProp||jx(m))&&i(m??h[h.length-1],"MULTILINE_IMPLICIT_KEY","Implicit keys need to be on a single line")}else((c=S.found)==null?void 0:c.indent)!==r.indent&&i(o,"BAD_INDENT",XX);n.atKey=!0;const E=S.end,$=m?e(n,m,S,i):t(n,E,h,null,S,i);n.schema.compat&&qM(r.indent,m,i),n.atKey=!1,Age(n,l.items,$)&&i(E,"DUPLICATE_KEY","Map keys must be unique");const k=L3(y??[],{indicator:"map-value-ind",next:v,offset:$.range[2],onError:i,parentIndent:r.indent,startOnNewline:!m||m.type==="block-scalar"});if(o=k.end,k.found){_&&((v==null?void 0:v.type)==="block-map"&&!k.hasNewline&&i(o,"BLOCK_AS_IMPLICIT_KEY","Nested mappings are not allowed in compact mappings"),n.options.strict&&S.start<k.found.offset-1024&&i($.range,"KEY_OVER_1024_CHARS","The : indicator must be at most 1024 chars after the start of an implicit block mapping key"));const A=v?e(n,v,k,i):t(n,o,y,null,k,i);n.schema.compat&&qM(r.indent,v,i),o=A.range[2];const w=new Os($,A);n.options.keepSourceTokens&&(w.srcToken=d),l.items.push(w)}else{_&&i($.range,"MISSING_CHAR","Implicit map keys need to be followed by map values"),k.comment&&($.comment?$.comment+=`
`+k.comment:$.comment=k.comment);const A=new Os($);n.options.keepSourceTokens&&(A.srcToken=d),l.items.push(A)}}return s&&s<o&&i(s,"IMPOSSIBLE","Map comment with trailing content"),l.range=[r.offset,o,s??o],l}function wBe({composeNode:e,composeEmptyNode:t},n,r,i,a){const u=(a==null?void 0:a.nodeClass)??ly,l=new u(n.schema);n.atRoot&&(n.atRoot=!1),n.atKey&&(n.atKey=!1);let o=r.offset,s=null;for(const{start:c,value:d}of r.items){const h=L3(c,{indicator:"seq-item-ind",next:d,offset:o,onError:i,parentIndent:r.indent,startOnNewline:!0});if(!h.found)if(h.anchor||h.tag||d)d&&d.type==="block-seq"?i(h.end,"BAD_INDENT","All sequence items must start at the same column"):i(o,"MISSING_CHAR","Sequence item without - indicator");else{s=h.end,h.comment&&(l.comment=h.comment);continue}const m=d?e(n,d,h,i):t(n,h.end,c,null,h,i);n.schema.compat&&qM(r.indent,d,i),o=m.range[2],l.items.push(m)}return l.range=[r.offset,o,s??o],l}function LS(e,t,n,r){let i="";if(e){let a=!1,u="";for(const l of e){const{source:o,type:s}=l;switch(s){case"space":a=!0;break;case"comment":{n&&!a&&r(l,"MISSING_CHAR","Comments must be separated from other tokens by white space characters");const c=o.substring(1)||" ";i?i+=u+c:i=c,u="";break}case"newline":i&&(u+=o),a=!0;break;default:r(l,"UNEXPECTED_TOKEN",`Unexpected ${s} at node end`)}t+=o.length}}return{comment:i,offset:t}}const kP="Block collections are not allowed within flow collections",PP=e=>e&&(e.type==="block-map"||e.type==="block-seq");function RBe({composeNode:e,composeEmptyNode:t},n,r,i,a){const u=r.start.source==="{",l=u?"flow map":"flow sequence",o=(a==null?void 0:a.nodeClass)??(u?qf:ly),s=new o(n.schema);s.flow=!0;const c=n.atRoot;c&&(n.atRoot=!1),n.atKey&&(n.atKey=!1);let d=r.offset+r.start.source.length;for(let S=0;S<r.items.length;++S){const _=r.items[S],{start:E,key:$,sep:k,value:A}=_,w=L3(E,{flow:l,indicator:"explicit-key-ind",next:$??(k==null?void 0:k[0]),offset:d,onError:i,parentIndent:r.indent,startOnNewline:!1});if(!w.found){if(!w.anchor&&!w.tag&&!k&&!A){S===0&&w.comma?i(w.comma,"UNEXPECTED_TOKEN",`Unexpected , in ${l}`):S<r.items.length-1&&i(w.start,"UNEXPECTED_TOKEN",`Unexpected empty item in ${l}`),w.comment&&(s.comment?s.comment+=`
`+w.comment:s.comment=w.comment),d=w.end;continue}!u&&n.options.strict&&jx($)&&i($,"MULTILINE_IMPLICIT_KEY","Implicit keys of flow sequence pairs need to be on a single line")}if(S===0)w.comma&&i(w.comma,"UNEXPECTED_TOKEN",`Unexpected , in ${l}`);else if(w.comma||i(w.start,"MISSING_CHAR",`Missing , between ${l} items`),w.comment){let R="";e:for(const F of E)switch(F.type){case"comma":case"space":break;case"comment":R=F.source.substring(1);break e;default:break e}if(R){let F=s.items[s.items.length-1];Ha(F)&&(F=F.value??F.key),F.comment?F.comment+=`
`+R:F.comment=R,w.comment=w.comment.substring(R.length+1)}}if(!u&&!k&&!w.found){const R=A?e(n,A,w,i):t(n,w.end,k,null,w,i);s.items.push(R),d=R.range[2],PP(A)&&i(R.range,"BLOCK_IN_FLOW",kP)}else{n.atKey=!0;const R=w.end,F=$?e(n,$,w,i):t(n,R,E,null,w,i);PP($)&&i(F.range,"BLOCK_IN_FLOW",kP),n.atKey=!1;const I=L3(k??[],{flow:l,indicator:"map-value-ind",next:A,offset:F.range[2],onError:i,parentIndent:r.indent,startOnNewline:!1});if(I.found){if(!u&&!w.found&&n.options.strict){if(k)for(const V of k){if(V===I.found)break;if(V.type==="newline"){i(V,"MULTILINE_IMPLICIT_KEY","Implicit keys of flow sequence pairs need to be on a single line");break}}w.start<I.found.offset-1024&&i(I.found,"KEY_OVER_1024_CHARS","The : indicator must be at most 1024 chars after the start of an implicit flow sequence key")}}else A&&("source"in A&&A.source&&A.source[0]===":"?i(A,"MISSING_CHAR",`Missing space after : in ${l}`):i(I.start,"MISSING_CHAR",`Missing , or : between ${l} items`));const U=A?e(n,A,I,i):I.found?t(n,I.end,k,null,I,i):null;U?PP(A)&&i(U.range,"BLOCK_IN_FLOW",kP):I.comment&&(F.comment?F.comment+=`
`+I.comment:F.comment=I.comment);const q=new Os(F,U);if(n.options.keepSourceTokens&&(q.srcToken=_),u){const V=s;Age(n,V.items,F)&&i(R,"DUPLICATE_KEY","Map keys must be unique"),V.items.push(q)}else{const V=new qf(n.schema);V.flow=!0,V.items.push(q);const N=(U??F).range;V.range=[F.range[0],N[1],N[2]],s.items.push(V)}d=U?U.range[2]:I.end}}const h=u?"}":"]",[m,...y]=r.end;let v=d;if(m&&m.source===h)v=m.offset+m.source.length;else{const S=l[0].toUpperCase()+l.substring(1),_=c?`${S} must end with a ${h}`:`${S} in block collection must be sufficiently indented and end with a ${h}`;i(d,c?"MISSING_CHAR":"BAD_INDENT",_),m&&m.source.length!==1&&y.unshift(m)}if(y.length>0){const S=LS(y,v,n.options.strict,i);S.comment&&(s.comment?s.comment+=`
`+S.comment:s.comment=S.comment),s.range=[r.offset,v,S.offset]}else s.range=[r.offset,v,v];return s}function MP(e,t,n,r,i,a){const u=n.type==="block-map"?CBe(e,t,n,r,a):n.type==="block-seq"?wBe(e,t,n,r,a):RBe(e,t,n,r,a),l=u.constructor;return i==="!"||i===l.tagName?(u.tag=l.tagName,u):(i&&(u.tag=i),u)}function ABe(e,t,n,r,i){var h;const a=r.tag,u=a?t.directives.tagName(a.source,m=>i(a,"TAG_RESOLVE_FAILED",m)):null;if(n.type==="block-seq"){const{anchor:m,newlineAfterProp:y}=r,v=m&&a?m.offset>a.offset?m:a:m??a;v&&(!y||y.offset<v.offset)&&i(v,"MISSING_CHAR","Missing newline after block sequence props")}const l=n.type==="block-map"?"map":n.type==="block-seq"?"seq":n.start.source==="{"?"map":"seq";if(!a||!u||u==="!"||u===qf.tagName&&l==="map"||u===ly.tagName&&l==="seq")return MP(e,t,n,i,u);let o=t.schema.tags.find(m=>m.tag===u&&m.collection===l);if(!o){const m=t.schema.knownTags[u];if(m&&m.collection===l)t.schema.tags.push(Object.assign({},m,{default:!1})),o=m;else return m?i(a,"BAD_COLLECTION_TYPE",`${m.tag} used for ${l} collection, but expects ${m.collection??"scalar"}`,!0):i(a,"TAG_RESOLVE_FAILED",`Unresolved tag: ${u}`,!0),MP(e,t,n,i,u)}const s=MP(e,t,n,i,u,o),c=((h=o.resolve)==null?void 0:h.call(o,s,m=>i(a,"TAG_RESOLVE_FAILED",m),t.options))??s,d=hu(c)?c:new sr(c);return d.range=s.range,d.tag=u,o!=null&&o.format&&(d.format=o.format),d}function Tge(e,t,n){const r=t.offset,i=TBe(t,e.options.strict,n);if(!i)return{value:"",type:null,comment:"",range:[r,r,r]};const a=i.mode===">"?sr.BLOCK_FOLDED:sr.BLOCK_LITERAL,u=t.source?kBe(t.source):[];let l=u.length;for(let v=u.length-1;v>=0;--v){const S=u[v][1];if(S===""||S==="\r")l=v;else break}if(l===0){const v=i.chomp==="+"&&u.length>0?`
`.repeat(Math.max(1,u.length-1)):"";let S=r+i.length;return t.source&&(S+=t.source.length),{value:v,type:a,comment:i.comment,range:[r,S,S]}}let o=t.indent+i.indent,s=t.offset+i.length,c=0;for(let v=0;v<l;++v){const[S,_]=u[v];if(_===""||_==="\r")i.indent===0&&S.length>o&&(o=S.length);else{S.length<o&&n(s+S.length,"MISSING_CHAR","Block scalars with more-indented leading empty lines must use an explicit indentation indicator"),i.indent===0&&(o=S.length),c=v,o===0&&!e.atRoot&&n(s,"BAD_INDENT","Block scalar values in collections must be indented");break}s+=S.length+_.length+1}for(let v=u.length-1;v>=l;--v)u[v][0].length>o&&(l=v+1);let d="",h="",m=!1;for(let v=0;v<c;++v)d+=u[v][0].slice(o)+`
`;for(let v=c;v<l;++v){let[S,_]=u[v];s+=S.length+_.length+1;const E=_[_.length-1]==="\r";if(E&&(_=_.slice(0,-1)),_&&S.length<o){const k=`Block scalar lines must not be less indented than their ${i.indent?"explicit indentation indicator":"first line"}`;n(s-_.length-(E?2:1),"BAD_INDENT",k),S=""}a===sr.BLOCK_LITERAL?(d+=h+S.slice(o)+_,h=`
`):S.length>o||_[0]==="	"?(h===" "?h=`
`:!m&&h===`
`&&(h=`

`),d+=h+S.slice(o)+_,h=`
`,m=!0):_===""?h===`
`?d+=`
`:h=`
`:(d+=h+_,h=" ",m=!1)}switch(i.chomp){case"-":break;case"+":for(let v=l;v<u.length;++v)d+=`
`+u[v][0].slice(o);d[d.length-1]!==`
`&&(d+=`
`);break;default:d+=`
`}const y=r+i.length+t.source.length;return{value:d,type:a,comment:i.comment,range:[r,y,y]}}function TBe({offset:e,props:t},n,r){if(t[0].type!=="block-scalar-header")return r(t[0],"IMPOSSIBLE","Block scalar header not found"),null;const{source:i}=t[0],a=i[0];let u=0,l="",o=-1;for(let h=1;h<i.length;++h){const m=i[h];if(!l&&(m==="-"||m==="+"))l=m;else{const y=Number(m);!u&&y?u=y:o===-1&&(o=e+h)}}o!==-1&&r(o,"UNEXPECTED_TOKEN",`Block scalar header includes extra characters: ${i}`);let s=!1,c="",d=i.length;for(let h=1;h<t.length;++h){const m=t[h];switch(m.type){case"space":s=!0;case"newline":d+=m.source.length;break;case"comment":n&&!s&&r(m,"MISSING_CHAR","Comments must be separated from other tokens by white space characters"),d+=m.source.length,c=m.source.substring(1);break;case"error":r(m,"UNEXPECTED_TOKEN",m.message),d+=m.source.length;break;default:{const y=`Unexpected token in block scalar header: ${m.type}`;r(m,"UNEXPECTED_TOKEN",y);const v=m.source;v&&typeof v=="string"&&(d+=v.length)}}}return{mode:a,indent:u,chomp:l,comment:c,length:d}}function kBe(e){const t=e.split(/\n( *)/),n=t[0],r=n.match(/^( *)/),a=[r!=null&&r[1]?[r[1],n.slice(r[1].length)]:["",n]];for(let u=1;u<t.length;u+=2)a.push([t[u],t[u+1]]);return a}function kge(e,t,n){const{offset:r,type:i,source:a,end:u}=e;let l,o;const s=(h,m,y)=>n(r+h,m,y);switch(i){case"scalar":l=sr.PLAIN,o=PBe(a,s);break;case"single-quoted-scalar":l=sr.QUOTE_SINGLE,o=MBe(a,s);break;case"double-quoted-scalar":l=sr.QUOTE_DOUBLE,o=FBe(a,s);break;default:return n(e,"UNEXPECTED_TOKEN",`Expected a flow scalar value, but found: ${i}`),{value:"",type:null,comment:"",range:[r,r+a.length,r+a.length]}}const c=r+a.length,d=LS(u,c,t,n);return{value:o,type:l,comment:d.comment,range:[r,c,d.offset]}}function PBe(e,t){let n="";switch(e[0]){case"	":n="a tab character";break;case",":n="flow indicator character ,";break;case"%":n="directive indicator character %";break;case"|":case">":{n=`block scalar indicator ${e[0]}`;break}case"@":case"`":{n=`reserved character ${e[0]}`;break}}return n&&t(0,"BAD_SCALAR_START",`Plain value cannot start with ${n}`),Pge(e)}function MBe(e,t){return(e[e.length-1]!=="'"||e.length===1)&&t(e.length,"MISSING_CHAR","Missing closing 'quote"),Pge(e.slice(1,-1)).replace(/''/g,"'")}function Pge(e){let t,n;try{t=new RegExp(`(.*?)(?<![ 	])[ 	]*\r?
`,"sy"),n=new RegExp(`[ 	]*(.*?)(?:(?<![ 	])[ 	]*)?\r?
`,"sy")}catch{t=/(.*?)[ \t]*\r?\n/sy,n=/[ \t]*(.*?)[ \t]*\r?\n/sy}let r=t.exec(e);if(!r)return e;let i=r[1],a=" ",u=t.lastIndex;for(n.lastIndex=u;r=n.exec(e);)r[1]===""?a===`
`?i+=a:a=`
`:(i+=a+r[1],a=" "),u=n.lastIndex;const l=/[ \t]*(.*)/sy;return l.lastIndex=u,r=l.exec(e),i+a+((r==null?void 0:r[1])??"")}function FBe(e,t){let n="";for(let r=1;r<e.length-1;++r){const i=e[r];if(!(i==="\r"&&e[r+1]===`
`))if(i===`
`){const{fold:a,offset:u}=NBe(e,r);n+=a,r=u}else if(i==="\\"){let a=e[++r];const u=IBe[a];if(u)n+=u;else if(a===`
`)for(a=e[r+1];a===" "||a==="	";)a=e[++r+1];else if(a==="\r"&&e[r+1]===`
`)for(a=e[++r+1];a===" "||a==="	";)a=e[++r+1];else if(a==="x"||a==="u"||a==="U"){const l={x:2,u:4,U:8}[a];n+=LBe(e,r+1,l,t),r+=l}else{const l=e.substr(r-1,2);t(r-1,"BAD_DQ_ESCAPE",`Invalid escape sequence ${l}`),n+=l}}else if(i===" "||i==="	"){const a=r;let u=e[r+1];for(;u===" "||u==="	";)u=e[++r+1];u!==`
`&&!(u==="\r"&&e[r+2]===`
`)&&(n+=r>a?e.slice(a,r+1):i)}else n+=i}return(e[e.length-1]!=='"'||e.length===1)&&t(e.length,"MISSING_CHAR",'Missing closing "quote'),n}function NBe(e,t){let n="",r=e[t+1];for(;(r===" "||r==="	"||r===`
`||r==="\r")&&!(r==="\r"&&e[t+2]!==`
`);)r===`
`&&(n+=`
`),t+=1,r=e[t+1];return n||(n=" "),{fold:n,offset:t}}const IBe={0:"\0",a:"\x07",b:"\b",e:"\x1B",f:"\f",n:`
`,r:"\r",t:"	",v:"\v",N:"",_:" ",L:"\u2028",P:"\u2029"," ":" ",'"':'"',"/":"/","\\":"\\","	":"	"};function LBe(e,t,n,r){const i=e.substr(t,n),u=i.length===n&&/^[0-9a-fA-F]+$/.test(i)?parseInt(i,16):NaN;if(isNaN(u)){const l=e.substr(t-2,n+2);return r(t-2,"BAD_DQ_ESCAPE",`Invalid escape sequence ${l}`),l}return String.fromCodePoint(u)}function Mge(e,t,n,r){const{value:i,type:a,comment:u,range:l}=t.type==="block-scalar"?Tge(e,t,r):kge(t,e.options.strict,r),o=n?e.directives.tagName(n.source,d=>r(n,"TAG_RESOLVE_FAILED",d)):null;let s;e.options.stringKeys&&e.atKey?s=e.schema[P0]:o?s=BBe(e.schema,i,o,n,r):t.type==="scalar"?s=QBe(e,i,t,r):s=e.schema[P0];let c;try{const d=s.resolve(i,h=>r(n??t,"TAG_RESOLVE_FAILED",h),e.options);c=fa(d)?d:new sr(d)}catch(d){const h=d instanceof Error?d.message:String(d);r(n??t,"TAG_RESOLVE_FAILED",h),c=new sr(i)}return c.range=l,c.source=i,a&&(c.type=a),o&&(c.tag=o),s.format&&(c.format=s.format),u&&(c.comment=u),c}function BBe(e,t,n,r,i){var l;if(n==="!")return e[P0];const a=[];for(const o of e.tags)if(!o.collection&&o.tag===n)if(o.default&&o.test)a.push(o);else return o;for(const o of a)if((l=o.test)!=null&&l.test(t))return o;const u=e.knownTags[n];return u&&!u.collection?(e.tags.push(Object.assign({},u,{default:!1,test:void 0})),u):(i(r,"TAG_RESOLVE_FAILED",`Unresolved tag: ${n}`,n!=="tag:yaml.org,2002:str"),e[P0])}function QBe({atKey:e,directives:t,schema:n},r,i,a){const u=n.tags.find(l=>{var o;return(l.default===!0||e&&l.default==="key")&&((o=l.test)==null?void 0:o.test(r))})||n[P0];if(n.compat){const l=n.compat.find(o=>{var s;return o.default&&((s=o.test)==null?void 0:s.test(r))})??n[P0];if(u.tag!==l.tag){const o=t.tagString(u.tag),s=t.tagString(l.tag),c=`Value may be parsed as either ${o} or ${s}`;a(i,"TAG_RESOLVE_FAILED",c,!0)}}return u}function zBe(e,t,n){if(t){n??(n=t.length);for(let r=n-1;r>=0;--r){let i=t[r];switch(i.type){case"space":case"comment":case"newline":e-=i.source.length;continue}for(i=t[++r];(i==null?void 0:i.type)==="space";)e+=i.source.length,i=t[++r];break}}return e}const qBe={composeNode:Fge,composeEmptyNode:OL};function Fge(e,t,n,r){const i=e.atKey,{spaceBefore:a,comment:u,anchor:l,tag:o}=n;let s,c=!0;switch(t.type){case"alias":s=VBe(e,t,r),(l||o)&&r(t,"ALIAS_PROPS","An alias node must not specify any properties");break;case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":case"block-scalar":s=Mge(e,t,o,r),l&&(s.anchor=l.source.substring(1));break;case"block-map":case"block-seq":case"flow-collection":s=ABe(qBe,e,t,n,r),l&&(s.anchor=l.source.substring(1));break;default:{const d=t.type==="error"?t.message:`Unsupported token (type: ${t.type})`;r(t,"UNEXPECTED_TOKEN",d),s=OL(e,t.offset,void 0,null,n,r),c=!1}}return l&&s.anchor===""&&r(l,"BAD_ALIAS","Anchor cannot be an empty string"),i&&e.options.stringKeys&&(!fa(s)||typeof s.value!="string"||s.tag&&s.tag!=="tag:yaml.org,2002:str")&&r(o??t,"NON_STRING_KEY","With stringKeys, all keys must be strings"),a&&(s.spaceBefore=!0),u&&(t.type==="scalar"&&t.source===""?s.comment=u:s.commentBefore=u),e.options.keepSourceTokens&&c&&(s.srcToken=t),s}function OL(e,t,n,r,{spaceBefore:i,comment:a,anchor:u,tag:l,end:o},s){const c={type:"scalar",offset:zBe(t,n,r),indent:-1,source:""},d=Mge(e,c,l,s);return u&&(d.anchor=u.source.substring(1),d.anchor===""&&s(u,"BAD_ALIAS","Anchor cannot be an empty string")),i&&(d.spaceBefore=!0),a&&(d.comment=a,d.range[2]=o),d}function VBe({options:e},{offset:t,source:n,end:r},i){const a=new kw(n.substring(1));a.source===""&&i(t,"BAD_ALIAS","Alias cannot be an empty string"),a.source.endsWith(":")&&i(t+n.length-1,"BAD_ALIAS","Alias ending in : is ambiguous",!0);const u=t+n.length,l=LS(r,u,e.strict,i);return a.range=[t,u,l.offset],l.comment&&(a.comment=l.comment),a}function UBe(e,t,{offset:n,start:r,value:i,end:a},u){const l=Object.assign({_directives:t},e),o=new qw(void 0,l),s={atKey:!1,atRoot:!0,directives:o.directives,options:o.options,schema:o.schema},c=L3(r,{indicator:"doc-start",next:i??(a==null?void 0:a[0]),offset:n,onError:u,parentIndent:0,startOnNewline:!0});c.found&&(o.directives.docStart=!0,i&&(i.type==="block-map"||i.type==="block-seq")&&!c.hasNewline&&u(c.end,"MISSING_CHAR","Block collection cannot start on same line with directives-end marker")),o.contents=i?Fge(s,i,c,u):OL(s,c.end,r,null,c,u);const d=o.contents.range[2],h=LS(a,d,!1,u);return h.comment&&(o.comment=h.comment),o.range=[n,d,h.offset],o}function d2(e){if(typeof e=="number")return[e,e+1];if(Array.isArray(e))return e.length===2?e:[e[0],e[1]];const{offset:t,source:n}=e;return[t,t+(typeof n=="string"?n.length:1)]}function YX(e){var i;let t="",n=!1,r=!1;for(let a=0;a<e.length;++a){const u=e[a];switch(u[0]){case"#":t+=(t===""?"":r?`

`:`
`)+(u.substring(1)||" "),n=!0,r=!1;break;case"%":((i=e[a+1])==null?void 0:i[0])!=="#"&&(a+=1),n=!1;break;default:n||(r=!0),n=!1}}return{comment:t,afterEmptyLine:r}}class xL{constructor(t={}){this.doc=null,this.atDirectives=!1,this.prelude=[],this.errors=[],this.warnings=[],this.onError=(n,r,i,a)=>{const u=d2(n);a?this.warnings.push(new Rge(u,r,i)):this.errors.push(new Ub(u,r,i))},this.directives=new uc({version:t.version||"1.2"}),this.options=t}decorate(t,n){const{comment:r,afterEmptyLine:i}=YX(this.prelude);if(r){const a=t.contents;if(n)t.comment=t.comment?`${t.comment}
${r}`:r;else if(i||t.directives.docStart||!a)t.commentBefore=r;else if(cu(a)&&!a.flow&&a.items.length>0){let u=a.items[0];Ha(u)&&(u=u.key);const l=u.commentBefore;u.commentBefore=l?`${r}
${l}`:r}else{const u=a.commentBefore;a.commentBefore=u?`${r}
${u}`:r}}n?(Array.prototype.push.apply(t.errors,this.errors),Array.prototype.push.apply(t.warnings,this.warnings)):(t.errors=this.errors,t.warnings=this.warnings),this.prelude=[],this.errors=[],this.warnings=[]}streamInfo(){return{comment:YX(this.prelude).comment,directives:this.directives,errors:this.errors,warnings:this.warnings}}*compose(t,n=!1,r=-1){for(const i of t)yield*this.next(i);yield*this.end(n,r)}*next(t){switch(t.type){case"directive":this.directives.add(t.source,(n,r,i)=>{const a=d2(t);a[0]+=n,this.onError(a,"BAD_DIRECTIVE",r,i)}),this.prelude.push(t.source),this.atDirectives=!0;break;case"document":{const n=UBe(this.options,this.directives,t,this.onError);this.atDirectives&&!n.directives.docStart&&this.onError(t,"MISSING_CHAR","Missing directives-end/doc-start indicator line"),this.decorate(n,!1),this.doc&&(yield this.doc),this.doc=n,this.atDirectives=!1;break}case"byte-order-mark":case"space":break;case"comment":case"newline":this.prelude.push(t.source);break;case"error":{const n=t.source?`${t.message}: ${JSON.stringify(t.source)}`:t.message,r=new Ub(d2(t),"UNEXPECTED_TOKEN",n);this.atDirectives||!this.doc?this.errors.push(r):this.doc.errors.push(r);break}case"doc-end":{if(!this.doc){const r="Unexpected doc-end without preceding document";this.errors.push(new Ub(d2(t),"UNEXPECTED_TOKEN",r));break}this.doc.directives.docEnd=!0;const n=LS(t.end,t.offset+t.source.length,this.doc.options.strict,this.onError);if(this.decorate(this.doc,!0),n.comment){const r=this.doc.comment;this.doc.comment=r?`${r}
${n.comment}`:n.comment}this.doc.range[2]=n.offset;break}default:this.errors.push(new Ub(d2(t),"UNEXPECTED_TOKEN",`Unsupported token ${t.type}`))}}*end(t=!1,n=-1){if(this.doc)this.decorate(this.doc,!0),yield this.doc,this.doc=null;else if(t){const r=Object.assign({_directives:this.directives},this.options),i=new qw(void 0,r);this.atDirectives&&this.onError(n,"MISSING_CHAR","Missing directives-end indicator line"),i.range=[0,n,n],this.decorate(i,!1),yield i}}}function jBe(e,t=!0,n){if(e){const r=(i,a,u)=>{const l=typeof i=="number"?i:Array.isArray(i)?i[0]:i.offset;if(n)n(l,a,u);else throw new Ub([l,l+1],a,u)};switch(e.type){case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":return kge(e,t,r);case"block-scalar":return Tge({options:{strict:t}},e,r)}}return null}function ZBe(e,t){const{implicitKey:n=!1,indent:r,inFlow:i=!1,offset:a=-1,type:u="PLAIN"}=t,l=NS({type:u,value:e},{implicitKey:n,indent:r>0?" ".repeat(r):"",inFlow:i,options:{blockQuote:!0,lineWidth:-1}}),o=t.end??[{type:"newline",offset:-1,indent:r,source:`
`}];switch(l[0]){case"|":case">":{const s=l.indexOf(`
`),c=l.substring(0,s),d=l.substring(s+1)+`
`,h=[{type:"block-scalar-header",offset:a,indent:r,source:c}];return Nge(h,o)||h.push({type:"newline",offset:-1,indent:r,source:`
`}),{type:"block-scalar",offset:a,indent:r,props:h,source:d}}case'"':return{type:"double-quoted-scalar",offset:a,indent:r,source:l,end:o};case"'":return{type:"single-quoted-scalar",offset:a,indent:r,source:l,end:o};default:return{type:"scalar",offset:a,indent:r,source:l,end:o}}}function HBe(e,t,n={}){let{afterKey:r=!1,implicitKey:i=!1,inFlow:a=!1,type:u}=n,l="indent"in e?e.indent:null;if(r&&typeof l=="number"&&(l+=2),!u)switch(e.type){case"single-quoted-scalar":u="QUOTE_SINGLE";break;case"double-quoted-scalar":u="QUOTE_DOUBLE";break;case"block-scalar":{const s=e.props[0];if(s.type!=="block-scalar-header")throw new Error("Invalid block scalar header");u=s.source[0]===">"?"BLOCK_FOLDED":"BLOCK_LITERAL";break}default:u="PLAIN"}const o=NS({type:u,value:t},{implicitKey:i||l===null,indent:l!==null&&l>0?" ".repeat(l):"",inFlow:a,options:{blockQuote:!0,lineWidth:-1}});switch(o[0]){case"|":case">":XBe(e,o);break;case'"':FP(e,o,"double-quoted-scalar");break;case"'":FP(e,o,"single-quoted-scalar");break;default:FP(e,o,"scalar")}}function XBe(e,t){const n=t.indexOf(`
`),r=t.substring(0,n),i=t.substring(n+1)+`
`;if(e.type==="block-scalar"){const a=e.props[0];if(a.type!=="block-scalar-header")throw new Error("Invalid block scalar header");a.source=r,e.source=i}else{const{offset:a}=e,u="indent"in e?e.indent:-1,l=[{type:"block-scalar-header",offset:a,indent:u,source:r}];Nge(l,"end"in e?e.end:void 0)||l.push({type:"newline",offset:-1,indent:u,source:`
`});for(const o of Object.keys(e))o!=="type"&&o!=="offset"&&delete e[o];Object.assign(e,{type:"block-scalar",indent:u,props:l,source:i})}}function Nge(e,t){if(t)for(const n of t)switch(n.type){case"space":case"comment":e.push(n);break;case"newline":return e.push(n),!0}return!1}function FP(e,t,n){switch(e.type){case"scalar":case"double-quoted-scalar":case"single-quoted-scalar":e.type=n,e.source=t;break;case"block-scalar":{const r=e.props.slice(1);let i=t.length;e.props[0].type==="block-scalar-header"&&(i-=e.props[0].source.length);for(const a of r)a.offset+=i;delete e.props,Object.assign(e,{type:n,source:t,end:r});break}case"block-map":case"block-seq":{const i={type:"newline",offset:e.offset+t.length,indent:e.indent,source:`
`};delete e.items,Object.assign(e,{type:n,source:t,end:[i]});break}default:{const r="indent"in e?e.indent:-1,i="end"in e&&Array.isArray(e.end)?e.end.filter(a=>a.type==="space"||a.type==="comment"||a.type==="newline"):[];for(const a of Object.keys(e))a!=="type"&&a!=="offset"&&delete e[a];Object.assign(e,{type:n,indent:r,source:t,end:i})}}}const YBe=e=>"type"in e?Q$(e):OD(e);function Q$(e){switch(e.type){case"block-scalar":{let t="";for(const n of e.props)t+=Q$(n);return t+e.source}case"block-map":case"block-seq":{let t="";for(const n of e.items)t+=OD(n);return t}case"flow-collection":{let t=e.start.source;for(const n of e.items)t+=OD(n);for(const n of e.end)t+=n.source;return t}case"document":{let t=OD(e);if(e.end)for(const n of e.end)t+=n.source;return t}default:{let t=e.source;if("end"in e&&e.end)for(const n of e.end)t+=n.source;return t}}}function OD({start:e,key:t,sep:n,value:r}){let i="";for(const a of e)i+=a.source;if(t&&(i+=Q$(t)),n)for(const a of n)i+=a.source;return r&&(i+=Q$(r)),i}const VM=Symbol("break visit"),GBe=Symbol("skip children"),Ige=Symbol("remove item");function m1(e,t){"type"in e&&e.type==="document"&&(e={start:e.start,value:e.value}),Lge(Object.freeze([]),e,t)}m1.BREAK=VM;m1.SKIP=GBe;m1.REMOVE=Ige;m1.itemAtPath=(e,t)=>{let n=e;for(const[r,i]of t){const a=n==null?void 0:n[r];if(a&&"items"in a)n=a.items[i];else return}return n};m1.parentCollection=(e,t)=>{const n=m1.itemAtPath(e,t.slice(0,-1)),r=t[t.length-1][0],i=n==null?void 0:n[r];if(i&&"items"in i)return i;throw new Error("Parent collection not found")};function Lge(e,t,n){let r=n(t,e);if(typeof r=="symbol")return r;for(const i of["key","value"]){const a=t[i];if(a&&"items"in a){for(let u=0;u<a.items.length;++u){const l=Lge(Object.freeze(e.concat([[i,u]])),a.items[u],n);if(typeof l=="number")u=l-1;else{if(l===VM)return VM;l===Ige&&(a.items.splice(u,1),u-=1)}}typeof r=="function"&&i==="key"&&(r=r(t,e))}}return typeof r=="function"?r(t,e):r}const Vw="\uFEFF",Uw="",jw="",Zx="",KBe=e=>!!e&&"items"in e,WBe=e=>!!e&&(e.type==="scalar"||e.type==="single-quoted-scalar"||e.type==="double-quoted-scalar"||e.type==="block-scalar");function JBe(e){switch(e){case Vw:return"<BOM>";case Uw:return"<DOC>";case jw:return"<FLOW_END>";case Zx:return"<SCALAR>";default:return JSON.stringify(e)}}function Bge(e){switch(e){case Vw:return"byte-order-mark";case Uw:return"doc-mode";case jw:return"flow-error-end";case Zx:return"scalar";case"---":return"doc-start";case"...":return"doc-end";case"":case`
`:case`\r
`:return"newline";case"-":return"seq-item-ind";case"?":return"explicit-key-ind";case":":return"map-value-ind";case"{":return"flow-map-start";case"}":return"flow-map-end";case"[":return"flow-seq-start";case"]":return"flow-seq-end";case",":return"comma"}switch(e[0]){case" ":case"	":return"space";case"#":return"comment";case"%":return"directive-line";case"*":return"alias";case"&":return"anchor";case"!":return"tag";case"'":return"single-quoted-scalar";case'"':return"double-quoted-scalar";case"|":case">":return"block-scalar-header"}return null}const eQe=Object.freeze(Object.defineProperty({__proto__:null,BOM:Vw,DOCUMENT:Uw,FLOW_END:jw,SCALAR:Zx,createScalarToken:ZBe,isCollection:KBe,isScalar:WBe,prettyToken:JBe,resolveAsScalar:jBe,setScalarValue:HBe,stringify:YBe,tokenType:Bge,visit:m1},Symbol.toStringTag,{value:"Module"}));function Hh(e){switch(e){case void 0:case" ":case`
`:case"\r":case"	":return!0;default:return!1}}const GX=new Set("0123456789ABCDEFabcdef"),tQe=new Set("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-#;/?:@&=+$_.!~*'()"),wE=new Set(",[]{}"),nQe=new Set(` ,[]{}
\r	`),NP=e=>!e||nQe.has(e);class Qge{constructor(){this.atEnd=!1,this.blockScalarIndent=-1,this.blockScalarKeep=!1,this.buffer="",this.flowKey=!1,this.flowLevel=0,this.indentNext=0,this.indentValue=0,this.lineEndPos=null,this.next=null,this.pos=0}*lex(t,n=!1){if(t){if(typeof t!="string")throw TypeError("source is not a string");this.buffer=this.buffer?this.buffer+t:t,this.lineEndPos=null}this.atEnd=!n;let r=this.next??"stream";for(;r&&(n||this.hasChars(1));)r=yield*this.parseNext(r)}atLineEnd(){let t=this.pos,n=this.buffer[t];for(;n===" "||n==="	";)n=this.buffer[++t];return!n||n==="#"||n===`
`?!0:n==="\r"?this.buffer[t+1]===`
`:!1}charAt(t){return this.buffer[this.pos+t]}continueScalar(t){let n=this.buffer[t];if(this.indentNext>0){let r=0;for(;n===" ";)n=this.buffer[++r+t];if(n==="\r"){const i=this.buffer[r+t+1];if(i===`
`||!i&&!this.atEnd)return t+r+1}return n===`
`||r>=this.indentNext||!n&&!this.atEnd?t+r:-1}if(n==="-"||n==="."){const r=this.buffer.substr(t,3);if((r==="---"||r==="...")&&Hh(this.buffer[t+3]))return-1}return t}getLine(){let t=this.lineEndPos;return(typeof t!="number"||t!==-1&&t<this.pos)&&(t=this.buffer.indexOf(`
`,this.pos),this.lineEndPos=t),t===-1?this.atEnd?this.buffer.substring(this.pos):null:(this.buffer[t-1]==="\r"&&(t-=1),this.buffer.substring(this.pos,t))}hasChars(t){return this.pos+t<=this.buffer.length}setNext(t){return this.buffer=this.buffer.substring(this.pos),this.pos=0,this.lineEndPos=null,this.next=t,null}peek(t){return this.buffer.substr(this.pos,t)}*parseNext(t){switch(t){case"stream":return yield*this.parseStream();case"line-start":return yield*this.parseLineStart();case"block-start":return yield*this.parseBlockStart();case"doc":return yield*this.parseDocument();case"flow":return yield*this.parseFlowCollection();case"quoted-scalar":return yield*this.parseQuotedScalar();case"block-scalar":return yield*this.parseBlockScalar();case"plain-scalar":return yield*this.parsePlainScalar()}}*parseStream(){let t=this.getLine();if(t===null)return this.setNext("stream");if(t[0]===Vw&&(yield*this.pushCount(1),t=t.substring(1)),t[0]==="%"){let n=t.length,r=t.indexOf("#");for(;r!==-1;){const a=t[r-1];if(a===" "||a==="	"){n=r-1;break}else r=t.indexOf("#",r+1)}for(;;){const a=t[n-1];if(a===" "||a==="	")n-=1;else break}const i=(yield*this.pushCount(n))+(yield*this.pushSpaces(!0));return yield*this.pushCount(t.length-i),this.pushNewline(),"stream"}if(this.atLineEnd()){const n=yield*this.pushSpaces(!0);return yield*this.pushCount(t.length-n),yield*this.pushNewline(),"stream"}return yield Uw,yield*this.parseLineStart()}*parseLineStart(){const t=this.charAt(0);if(!t&&!this.atEnd)return this.setNext("line-start");if(t==="-"||t==="."){if(!this.atEnd&&!this.hasChars(4))return this.setNext("line-start");const n=this.peek(3);if((n==="---"||n==="...")&&Hh(this.charAt(3)))return yield*this.pushCount(3),this.indentValue=0,this.indentNext=0,n==="---"?"doc":"stream"}return this.indentValue=yield*this.pushSpaces(!1),this.indentNext>this.indentValue&&!Hh(this.charAt(1))&&(this.indentNext=this.indentValue),yield*this.parseBlockStart()}*parseBlockStart(){const[t,n]=this.peek(2);if(!n&&!this.atEnd)return this.setNext("block-start");if((t==="-"||t==="?"||t===":")&&Hh(n)){const r=(yield*this.pushCount(1))+(yield*this.pushSpaces(!0));return this.indentNext=this.indentValue+1,this.indentValue+=r,yield*this.parseBlockStart()}return"doc"}*parseDocument(){yield*this.pushSpaces(!0);const t=this.getLine();if(t===null)return this.setNext("doc");let n=yield*this.pushIndicators();switch(t[n]){case"#":yield*this.pushCount(t.length-n);case void 0:return yield*this.pushNewline(),yield*this.parseLineStart();case"{":case"[":return yield*this.pushCount(1),this.flowKey=!1,this.flowLevel=1,"flow";case"}":case"]":return yield*this.pushCount(1),"doc";case"*":return yield*this.pushUntil(NP),"doc";case'"':case"'":return yield*this.parseQuotedScalar();case"|":case">":return n+=yield*this.parseBlockScalarHeader(),n+=yield*this.pushSpaces(!0),yield*this.pushCount(t.length-n),yield*this.pushNewline(),yield*this.parseBlockScalar();default:return yield*this.parsePlainScalar()}}*parseFlowCollection(){let t,n,r=-1;do t=yield*this.pushNewline(),t>0?(n=yield*this.pushSpaces(!1),this.indentValue=r=n):n=0,n+=yield*this.pushSpaces(!0);while(t+n>0);const i=this.getLine();if(i===null)return this.setNext("flow");if((r!==-1&&r<this.indentNext&&i[0]!=="#"||r===0&&(i.startsWith("---")||i.startsWith("..."))&&Hh(i[3]))&&!(r===this.indentNext-1&&this.flowLevel===1&&(i[0]==="]"||i[0]==="}")))return this.flowLevel=0,yield jw,yield*this.parseLineStart();let a=0;for(;i[a]===",";)a+=yield*this.pushCount(1),a+=yield*this.pushSpaces(!0),this.flowKey=!1;switch(a+=yield*this.pushIndicators(),i[a]){case void 0:return"flow";case"#":return yield*this.pushCount(i.length-a),"flow";case"{":case"[":return yield*this.pushCount(1),this.flowKey=!1,this.flowLevel+=1,"flow";case"}":case"]":return yield*this.pushCount(1),this.flowKey=!0,this.flowLevel-=1,this.flowLevel?"flow":"doc";case"*":return yield*this.pushUntil(NP),"flow";case'"':case"'":return this.flowKey=!0,yield*this.parseQuotedScalar();case":":{const u=this.charAt(1);if(this.flowKey||Hh(u)||u===",")return this.flowKey=!1,yield*this.pushCount(1),yield*this.pushSpaces(!0),"flow"}default:return this.flowKey=!1,yield*this.parsePlainScalar()}}*parseQuotedScalar(){const t=this.charAt(0);let n=this.buffer.indexOf(t,this.pos+1);if(t==="'")for(;n!==-1&&this.buffer[n+1]==="'";)n=this.buffer.indexOf("'",n+2);else for(;n!==-1;){let a=0;for(;this.buffer[n-1-a]==="\\";)a+=1;if(a%2===0)break;n=this.buffer.indexOf('"',n+1)}const r=this.buffer.substring(0,n);let i=r.indexOf(`
`,this.pos);if(i!==-1){for(;i!==-1;){const a=this.continueScalar(i+1);if(a===-1)break;i=r.indexOf(`
`,a)}i!==-1&&(n=i-(r[i-1]==="\r"?2:1))}if(n===-1){if(!this.atEnd)return this.setNext("quoted-scalar");n=this.buffer.length}return yield*this.pushToIndex(n+1,!1),this.flowLevel?"flow":"doc"}*parseBlockScalarHeader(){this.blockScalarIndent=-1,this.blockScalarKeep=!1;let t=this.pos;for(;;){const n=this.buffer[++t];if(n==="+")this.blockScalarKeep=!0;else if(n>"0"&&n<="9")this.blockScalarIndent=Number(n)-1;else if(n!=="-")break}return yield*this.pushUntil(n=>Hh(n)||n==="#")}*parseBlockScalar(){let t=this.pos-1,n=0,r;e:for(let a=this.pos;r=this.buffer[a];++a)switch(r){case" ":n+=1;break;case`
`:t=a,n=0;break;case"\r":{const u=this.buffer[a+1];if(!u&&!this.atEnd)return this.setNext("block-scalar");if(u===`
`)break}default:break e}if(!r&&!this.atEnd)return this.setNext("block-scalar");if(n>=this.indentNext){this.blockScalarIndent===-1?this.indentNext=n:this.indentNext=this.blockScalarIndent+(this.indentNext===0?1:this.indentNext);do{const a=this.continueScalar(t+1);if(a===-1)break;t=this.buffer.indexOf(`
`,a)}while(t!==-1);if(t===-1){if(!this.atEnd)return this.setNext("block-scalar");t=this.buffer.length}}let i=t+1;for(r=this.buffer[i];r===" ";)r=this.buffer[++i];if(r==="	"){for(;r==="	"||r===" "||r==="\r"||r===`
`;)r=this.buffer[++i];t=i-1}else if(!this.blockScalarKeep)do{let a=t-1,u=this.buffer[a];u==="\r"&&(u=this.buffer[--a]);const l=a;for(;u===" ";)u=this.buffer[--a];if(u===`
`&&a>=this.pos&&a+1+n>l)t=a;else break}while(!0);return yield Zx,yield*this.pushToIndex(t+1,!0),yield*this.parseLineStart()}*parsePlainScalar(){const t=this.flowLevel>0;let n=this.pos-1,r=this.pos-1,i;for(;i=this.buffer[++r];)if(i===":"){const a=this.buffer[r+1];if(Hh(a)||t&&wE.has(a))break;n=r}else if(Hh(i)){let a=this.buffer[r+1];if(i==="\r"&&(a===`
`?(r+=1,i=`
`,a=this.buffer[r+1]):n=r),a==="#"||t&&wE.has(a))break;if(i===`
`){const u=this.continueScalar(r+1);if(u===-1)break;r=Math.max(r,u-2)}}else{if(t&&wE.has(i))break;n=r}return!i&&!this.atEnd?this.setNext("plain-scalar"):(yield Zx,yield*this.pushToIndex(n+1,!0),t?"flow":"doc")}*pushCount(t){return t>0?(yield this.buffer.substr(this.pos,t),this.pos+=t,t):0}*pushToIndex(t,n){const r=this.buffer.slice(this.pos,t);return r?(yield r,this.pos+=r.length,r.length):(n&&(yield""),0)}*pushIndicators(){switch(this.charAt(0)){case"!":return(yield*this.pushTag())+(yield*this.pushSpaces(!0))+(yield*this.pushIndicators());case"&":return(yield*this.pushUntil(NP))+(yield*this.pushSpaces(!0))+(yield*this.pushIndicators());case"-":case"?":case":":{const t=this.flowLevel>0,n=this.charAt(1);if(Hh(n)||t&&wE.has(n))return t?this.flowKey&&(this.flowKey=!1):this.indentNext=this.indentValue+1,(yield*this.pushCount(1))+(yield*this.pushSpaces(!0))+(yield*this.pushIndicators())}}return 0}*pushTag(){if(this.charAt(1)==="<"){let t=this.pos+2,n=this.buffer[t];for(;!Hh(n)&&n!==">";)n=this.buffer[++t];return yield*this.pushToIndex(n===">"?t+1:t,!1)}else{let t=this.pos+1,n=this.buffer[t];for(;n;)if(tQe.has(n))n=this.buffer[++t];else if(n==="%"&&GX.has(this.buffer[t+1])&&GX.has(this.buffer[t+2]))n=this.buffer[t+=3];else break;return yield*this.pushToIndex(t,!1)}}*pushNewline(){const t=this.buffer[this.pos];return t===`
`?yield*this.pushCount(1):t==="\r"&&this.charAt(1)===`
`?yield*this.pushCount(2):0}*pushSpaces(t){let n=this.pos-1,r;do r=this.buffer[++n];while(r===" "||t&&r==="	");const i=n-this.pos;return i>0&&(yield this.buffer.substr(this.pos,i),this.pos=n),i}*pushUntil(t){let n=this.pos,r=this.buffer[n];for(;!t(r);)r=this.buffer[++n];return yield*this.pushToIndex(n,!1)}}class zge{constructor(){this.lineStarts=[],this.addNewLine=t=>this.lineStarts.push(t),this.linePos=t=>{let n=0,r=this.lineStarts.length;for(;n<r;){const a=n+r>>1;this.lineStarts[a]<t?n=a+1:r=a}if(this.lineStarts[n]===t)return{line:n+1,col:1};if(n===0)return{line:0,col:t};const i=this.lineStarts[n-1];return{line:n,col:t-i+1}}}}function Dv(e,t){for(let n=0;n<e.length;++n)if(e[n].type===t)return!0;return!1}function KX(e){for(let t=0;t<e.length;++t)switch(e[t].type){case"space":case"comment":case"newline":break;default:return t}return-1}function qge(e){switch(e==null?void 0:e.type){case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":case"flow-collection":return!0;default:return!1}}function RE(e){switch(e.type){case"document":return e.start;case"block-map":{const t=e.items[e.items.length-1];return t.sep??t.start}case"block-seq":return e.items[e.items.length-1].start;default:return[]}}function $O(e){var n;if(e.length===0)return[];let t=e.length;e:for(;--t>=0;)switch(e[t].type){case"doc-start":case"explicit-key-ind":case"map-value-ind":case"seq-item-ind":case"newline":break e}for(;((n=e[++t])==null?void 0:n.type)==="space";);return e.splice(t,e.length)}function WX(e){if(e.start.type==="flow-seq-start")for(const t of e.items)t.sep&&!t.value&&!Dv(t.start,"explicit-key-ind")&&!Dv(t.sep,"map-value-ind")&&(t.key&&(t.value=t.key),delete t.key,qge(t.value)?t.value.end?Array.prototype.push.apply(t.value.end,t.sep):t.value.end=t.sep:Array.prototype.push.apply(t.start,t.sep),delete t.sep)}class SL{constructor(t){this.atNewLine=!0,this.atScalar=!1,this.indent=0,this.offset=0,this.onKeyLine=!1,this.stack=[],this.source="",this.type="",this.lexer=new Qge,this.onNewLine=t}*parse(t,n=!1){this.onNewLine&&this.offset===0&&this.onNewLine(0);for(const r of this.lexer.lex(t,n))yield*this.next(r);n||(yield*this.end())}*next(t){if(this.source=t,this.atScalar){this.atScalar=!1,yield*this.step(),this.offset+=t.length;return}const n=Bge(t);if(n)if(n==="scalar")this.atNewLine=!1,this.atScalar=!0,this.type="scalar";else{switch(this.type=n,yield*this.step(),n){case"newline":this.atNewLine=!0,this.indent=0,this.onNewLine&&this.onNewLine(this.offset+t.length);break;case"space":this.atNewLine&&t[0]===" "&&(this.indent+=t.length);break;case"explicit-key-ind":case"map-value-ind":case"seq-item-ind":this.atNewLine&&(this.indent+=t.length);break;case"doc-mode":case"flow-error-end":return;default:this.atNewLine=!1}this.offset+=t.length}else{const r=`Not a YAML token: ${t}`;yield*this.pop({type:"error",offset:this.offset,message:r,source:t}),this.offset+=t.length}}*end(){for(;this.stack.length>0;)yield*this.pop()}get sourceToken(){return{type:this.type,offset:this.offset,indent:this.indent,source:this.source}}*step(){const t=this.peek(1);if(this.type==="doc-end"&&(!t||t.type!=="doc-end")){for(;this.stack.length>0;)yield*this.pop();this.stack.push({type:"doc-end",offset:this.offset,source:this.source});return}if(!t)return yield*this.stream();switch(t.type){case"document":return yield*this.document(t);case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":return yield*this.scalar(t);case"block-scalar":return yield*this.blockScalar(t);case"block-map":return yield*this.blockMap(t);case"block-seq":return yield*this.blockSequence(t);case"flow-collection":return yield*this.flowCollection(t);case"doc-end":return yield*this.documentEnd(t)}yield*this.pop()}peek(t){return this.stack[this.stack.length-t]}*pop(t){const n=t??this.stack.pop();if(!n)yield{type:"error",offset:this.offset,source:"",message:"Tried to pop an empty stack"};else if(this.stack.length===0)yield n;else{const r=this.peek(1);switch(n.type==="block-scalar"?n.indent="indent"in r?r.indent:0:n.type==="flow-collection"&&r.type==="document"&&(n.indent=0),n.type==="flow-collection"&&WX(n),r.type){case"document":r.value=n;break;case"block-scalar":r.props.push(n);break;case"block-map":{const i=r.items[r.items.length-1];if(i.value){r.items.push({start:[],key:n,sep:[]}),this.onKeyLine=!0;return}else if(i.sep)i.value=n;else{Object.assign(i,{key:n,sep:[]}),this.onKeyLine=!i.explicitKey;return}break}case"block-seq":{const i=r.items[r.items.length-1];i.value?r.items.push({start:[],value:n}):i.value=n;break}case"flow-collection":{const i=r.items[r.items.length-1];!i||i.value?r.items.push({start:[],key:n,sep:[]}):i.sep?i.value=n:Object.assign(i,{key:n,sep:[]});return}default:yield*this.pop(),yield*this.pop(n)}if((r.type==="document"||r.type==="block-map"||r.type==="block-seq")&&(n.type==="block-map"||n.type==="block-seq")){const i=n.items[n.items.length-1];i&&!i.sep&&!i.value&&i.start.length>0&&KX(i.start)===-1&&(n.indent===0||i.start.every(a=>a.type!=="comment"||a.indent<n.indent))&&(r.type==="document"?r.end=i.start:r.items.push({start:i.start}),n.items.splice(-1,1))}}}*stream(){switch(this.type){case"directive-line":yield{type:"directive",offset:this.offset,source:this.source};return;case"byte-order-mark":case"space":case"comment":case"newline":yield this.sourceToken;return;case"doc-mode":case"doc-start":{const t={type:"document",offset:this.offset,start:[]};this.type==="doc-start"&&t.start.push(this.sourceToken),this.stack.push(t);return}}yield{type:"error",offset:this.offset,message:`Unexpected ${this.type} token in YAML stream`,source:this.source}}*document(t){if(t.value)return yield*this.lineEnd(t);switch(this.type){case"doc-start":{KX(t.start)!==-1?(yield*this.pop(),yield*this.step()):t.start.push(this.sourceToken);return}case"anchor":case"tag":case"space":case"comment":case"newline":t.start.push(this.sourceToken);return}const n=this.startBlockValue(t);n?this.stack.push(n):yield{type:"error",offset:this.offset,message:`Unexpected ${this.type} token in YAML document`,source:this.source}}*scalar(t){if(this.type==="map-value-ind"){const n=RE(this.peek(2)),r=$O(n);let i;t.end?(i=t.end,i.push(this.sourceToken),delete t.end):i=[this.sourceToken];const a={type:"block-map",offset:t.offset,indent:t.indent,items:[{start:r,key:t,sep:i}]};this.onKeyLine=!0,this.stack[this.stack.length-1]=a}else yield*this.lineEnd(t)}*blockScalar(t){switch(this.type){case"space":case"comment":case"newline":t.props.push(this.sourceToken);return;case"scalar":if(t.source=this.source,this.atNewLine=!0,this.indent=0,this.onNewLine){let n=this.source.indexOf(`
`)+1;for(;n!==0;)this.onNewLine(this.offset+n),n=this.source.indexOf(`
`,n)+1}yield*this.pop();break;default:yield*this.pop(),yield*this.step()}}*blockMap(t){var r;const n=t.items[t.items.length-1];switch(this.type){case"newline":if(this.onKeyLine=!1,n.value){const i="end"in n.value?n.value.end:void 0,a=Array.isArray(i)?i[i.length-1]:void 0;(a==null?void 0:a.type)==="comment"?i==null||i.push(this.sourceToken):t.items.push({start:[this.sourceToken]})}else n.sep?n.sep.push(this.sourceToken):n.start.push(this.sourceToken);return;case"space":case"comment":if(n.value)t.items.push({start:[this.sourceToken]});else if(n.sep)n.sep.push(this.sourceToken);else{if(this.atIndentedComment(n.start,t.indent)){const i=t.items[t.items.length-2],a=(r=i==null?void 0:i.value)==null?void 0:r.end;if(Array.isArray(a)){Array.prototype.push.apply(a,n.start),a.push(this.sourceToken),t.items.pop();return}}n.start.push(this.sourceToken)}return}if(this.indent>=t.indent){const i=!this.onKeyLine&&this.indent===t.indent,a=i&&(n.sep||n.explicitKey)&&this.type!=="seq-item-ind";let u=[];if(a&&n.sep&&!n.value){const l=[];for(let o=0;o<n.sep.length;++o){const s=n.sep[o];switch(s.type){case"newline":l.push(o);break;case"space":break;case"comment":s.indent>t.indent&&(l.length=0);break;default:l.length=0}}l.length>=2&&(u=n.sep.splice(l[1]))}switch(this.type){case"anchor":case"tag":a||n.value?(u.push(this.sourceToken),t.items.push({start:u}),this.onKeyLine=!0):n.sep?n.sep.push(this.sourceToken):n.start.push(this.sourceToken);return;case"explicit-key-ind":!n.sep&&!n.explicitKey?(n.start.push(this.sourceToken),n.explicitKey=!0):a||n.value?(u.push(this.sourceToken),t.items.push({start:u,explicitKey:!0})):this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:[this.sourceToken],explicitKey:!0}]}),this.onKeyLine=!0;return;case"map-value-ind":if(n.explicitKey)if(n.sep)if(n.value)t.items.push({start:[],key:null,sep:[this.sourceToken]});else if(Dv(n.sep,"map-value-ind"))this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:u,key:null,sep:[this.sourceToken]}]});else if(qge(n.key)&&!Dv(n.sep,"newline")){const l=$O(n.start),o=n.key,s=n.sep;s.push(this.sourceToken),delete n.key,delete n.sep,this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:l,key:o,sep:s}]})}else u.length>0?n.sep=n.sep.concat(u,this.sourceToken):n.sep.push(this.sourceToken);else if(Dv(n.start,"newline"))Object.assign(n,{key:null,sep:[this.sourceToken]});else{const l=$O(n.start);this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:l,key:null,sep:[this.sourceToken]}]})}else n.sep?n.value||a?t.items.push({start:u,key:null,sep:[this.sourceToken]}):Dv(n.sep,"map-value-ind")?this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:[],key:null,sep:[this.sourceToken]}]}):n.sep.push(this.sourceToken):Object.assign(n,{key:null,sep:[this.sourceToken]});this.onKeyLine=!0;return;case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":{const l=this.flowScalar(this.type);a||n.value?(t.items.push({start:u,key:l,sep:[]}),this.onKeyLine=!0):n.sep?this.stack.push(l):(Object.assign(n,{key:l,sep:[]}),this.onKeyLine=!0);return}default:{const l=this.startBlockValue(t);if(l){if(l.type==="block-seq"){if(!n.explicitKey&&n.sep&&!Dv(n.sep,"newline")){yield*this.pop({type:"error",offset:this.offset,message:"Unexpected block-seq-ind on same line with key",source:this.source});return}}else i&&t.items.push({start:u});this.stack.push(l);return}}}}yield*this.pop(),yield*this.step()}*blockSequence(t){var r;const n=t.items[t.items.length-1];switch(this.type){case"newline":if(n.value){const i="end"in n.value?n.value.end:void 0,a=Array.isArray(i)?i[i.length-1]:void 0;(a==null?void 0:a.type)==="comment"?i==null||i.push(this.sourceToken):t.items.push({start:[this.sourceToken]})}else n.start.push(this.sourceToken);return;case"space":case"comment":if(n.value)t.items.push({start:[this.sourceToken]});else{if(this.atIndentedComment(n.start,t.indent)){const i=t.items[t.items.length-2],a=(r=i==null?void 0:i.value)==null?void 0:r.end;if(Array.isArray(a)){Array.prototype.push.apply(a,n.start),a.push(this.sourceToken),t.items.pop();return}}n.start.push(this.sourceToken)}return;case"anchor":case"tag":if(n.value||this.indent<=t.indent)break;n.start.push(this.sourceToken);return;case"seq-item-ind":if(this.indent!==t.indent)break;n.value||Dv(n.start,"seq-item-ind")?t.items.push({start:[this.sourceToken]}):n.start.push(this.sourceToken);return}if(this.indent>t.indent){const i=this.startBlockValue(t);if(i){this.stack.push(i);return}}yield*this.pop(),yield*this.step()}*flowCollection(t){const n=t.items[t.items.length-1];if(this.type==="flow-error-end"){let r;do yield*this.pop(),r=this.peek(1);while(r&&r.type==="flow-collection")}else if(t.end.length===0){switch(this.type){case"comma":case"explicit-key-ind":!n||n.sep?t.items.push({start:[this.sourceToken]}):n.start.push(this.sourceToken);return;case"map-value-ind":!n||n.value?t.items.push({start:[],key:null,sep:[this.sourceToken]}):n.sep?n.sep.push(this.sourceToken):Object.assign(n,{key:null,sep:[this.sourceToken]});return;case"space":case"comment":case"newline":case"anchor":case"tag":!n||n.value?t.items.push({start:[this.sourceToken]}):n.sep?n.sep.push(this.sourceToken):n.start.push(this.sourceToken);return;case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":{const i=this.flowScalar(this.type);!n||n.value?t.items.push({start:[],key:i,sep:[]}):n.sep?this.stack.push(i):Object.assign(n,{key:i,sep:[]});return}case"flow-map-end":case"flow-seq-end":t.end.push(this.sourceToken);return}const r=this.startBlockValue(t);r?this.stack.push(r):(yield*this.pop(),yield*this.step())}else{const r=this.peek(2);if(r.type==="block-map"&&(this.type==="map-value-ind"&&r.indent===t.indent||this.type==="newline"&&!r.items[r.items.length-1].sep))yield*this.pop(),yield*this.step();else if(this.type==="map-value-ind"&&r.type!=="flow-collection"){const i=RE(r),a=$O(i);WX(t);const u=t.end.splice(1,t.end.length);u.push(this.sourceToken);const l={type:"block-map",offset:t.offset,indent:t.indent,items:[{start:a,key:t,sep:u}]};this.onKeyLine=!0,this.stack[this.stack.length-1]=l}else yield*this.lineEnd(t)}}flowScalar(t){if(this.onNewLine){let n=this.source.indexOf(`
`)+1;for(;n!==0;)this.onNewLine(this.offset+n),n=this.source.indexOf(`
`,n)+1}return{type:t,offset:this.offset,indent:this.indent,source:this.source}}startBlockValue(t){switch(this.type){case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":return this.flowScalar(this.type);case"block-scalar-header":return{type:"block-scalar",offset:this.offset,indent:this.indent,props:[this.sourceToken],source:""};case"flow-map-start":case"flow-seq-start":return{type:"flow-collection",offset:this.offset,indent:this.indent,start:this.sourceToken,items:[],end:[]};case"seq-item-ind":return{type:"block-seq",offset:this.offset,indent:this.indent,items:[{start:[this.sourceToken]}]};case"explicit-key-ind":{this.onKeyLine=!0;const n=RE(t),r=$O(n);return r.push(this.sourceToken),{type:"block-map",offset:this.offset,indent:this.indent,items:[{start:r,explicitKey:!0}]}}case"map-value-ind":{this.onKeyLine=!0;const n=RE(t),r=$O(n);return{type:"block-map",offset:this.offset,indent:this.indent,items:[{start:r,key:null,sep:[this.sourceToken]}]}}}return null}atIndentedComment(t,n){return this.type!=="comment"||this.indent<=n?!1:t.every(r=>r.type==="newline"||r.type==="space")}*documentEnd(t){this.type!=="doc-mode"&&(t.end?t.end.push(this.sourceToken):t.end=[this.sourceToken],this.type==="newline"&&(yield*this.pop()))}*lineEnd(t){switch(this.type){case"comma":case"doc-start":case"doc-end":case"flow-seq-end":case"flow-map-end":case"map-value-ind":yield*this.pop(),yield*this.step();break;case"newline":this.onKeyLine=!1;case"space":case"comment":default:t.end?t.end.push(this.sourceToken):t.end=[this.sourceToken],this.type==="newline"&&(yield*this.pop())}}}function Vge(e){const t=e.prettyErrors!==!1;return{lineCounter:e.lineCounter||t&&new zge||null,prettyErrors:t}}function rQe(e,t={}){const{lineCounter:n,prettyErrors:r}=Vge(t),i=new SL(n==null?void 0:n.addNewLine),a=new xL(t),u=Array.from(a.compose(i.parse(e)));if(r&&n)for(const l of u)l.errors.forEach(B$(e,n)),l.warnings.forEach(B$(e,n));return u.length>0?u:Object.assign([],{empty:!0},a.streamInfo())}function Uge(e,t={}){const{lineCounter:n,prettyErrors:r}=Vge(t),i=new SL(n==null?void 0:n.addNewLine),a=new xL(t);let u=null;for(const l of a.compose(i.parse(e),!0,e.length))if(!u)u=l;else if(u.options.logLevel!=="silent"){u.errors.push(new Ub(l.range.slice(0,2),"MULTIPLE_DOCS","Source contains multiple documents; please use YAML.parseAllDocuments()"));break}return r&&n&&(u.errors.forEach(B$(e,n)),u.warnings.forEach(B$(e,n))),u}function iQe(e,t,n){let r;typeof t=="function"?r=t:n===void 0&&t&&typeof t=="object"&&(n=t);const i=Uge(e,n);if(!i)return null;if(i.warnings.forEach(a=>sge(i.options.logLevel,a)),i.errors.length>0){if(i.options.logLevel!=="silent")throw i.errors[0];i.errors=[]}return i.toJS(Object.assign({reviver:r},n))}function aQe(e,t,n){let r=null;if(typeof t=="function"||Array.isArray(t)?r=t:n===void 0&&t&&(n=t),typeof n=="string"&&(n=n.length),typeof n=="number"){const i=Math.round(n);n=i<1?void 0:i>8?{indent:8}:{indent:i}}if(e===void 0){const{keepUndefined:i}=n??t??{};if(!i)return}return k1(e)&&!r?e.toString(n):new qw(e,r,n).toString(n)}const $ut=Object.freeze(Object.defineProperty({__proto__:null,Alias:kw,CST:eQe,Composer:xL,Document:qw,Lexer:Qge,LineCounter:zge,Pair:Os,Parser:SL,Scalar:sr,Schema:zw,YAMLError:bL,YAMLMap:qf,YAMLParseError:Ub,YAMLSeq:ly,YAMLWarning:Rge,isAlias:xy,isCollection:cu,isDocument:k1,isMap:s4,isNode:hu,isPair:Ha,isScalar:fa,isSeq:c4,parse:iQe,parseAllDocuments:rQe,parseDocument:Uge,stringify:aQe,visit:P1,visitAsync:Tw},Symbol.toStringTag,{value:"Module"})),uQe=1,jge=194,Zge=195,lQe=196,JX=197,oQe=198,sQe=199,cQe=200,fQe=2,Hge=3,eY=201,dQe=24,hQe=25,pQe=49,mQe=50,gQe=55,vQe=56,yQe=57,bQe=59,OQe=60,xQe=61,SQe=62,_Qe=63,EQe=65,DQe=238,$Qe=71,CQe=241,wQe=242,RQe=243,AQe=244,TQe=245,kQe=246,PQe=247,MQe=248,Xge=72,FQe=249,NQe=250,IQe=251,LQe=252,BQe=253,QQe=254,zQe=255,qQe=256,VQe=73,UQe=77,jQe=263,ZQe=112,HQe=130,XQe=151,YQe=152,GQe=155,g1=10,Hx=13,_L=32,Zw=9,EL=35,KQe=40,WQe=46,UM=123,tY=125,Yge=39,Gge=34,nY=92,JQe=111,eze=120,tze=78,nze=117,rze=85,ize=new Set([hQe,pQe,mQe,jQe,EQe,HQe,vQe,yQe,DQe,SQe,_Qe,Xge,VQe,UQe,OQe,xQe,XQe,YQe,GQe,ZQe]);function IP(e){return e==g1||e==Hx}function LP(e){return e>=48&&e<=57||e>=65&&e<=70||e>=97&&e<=102}const aze=new vg((e,t)=>{let n;if(e.next<0)e.acceptToken(sQe);else if(t.context.flags&xD)IP(e.next)&&e.acceptToken(oQe,1);else if(((n=e.peek(-1))<0||IP(n))&&t.canShift(JX)){let r=0;for(;e.next==_L||e.next==Zw;)e.advance(),r++;(e.next==g1||e.next==Hx||e.next==EL)&&e.acceptToken(JX,-r)}else IP(e.next)&&e.acceptToken(lQe,1)},{contextual:!0}),uze=new vg((e,t)=>{let n=t.context;if(n.flags)return;let r=e.peek(-1);if(r==g1||r==Hx){let i=0,a=0;for(;;){if(e.next==_L)i++;else if(e.next==Zw)i+=8-i%8;else break;e.advance(),a++}i!=n.indent&&e.next!=g1&&e.next!=Hx&&e.next!=EL&&(i<n.indent?e.acceptToken(Zge,-a):e.acceptToken(jge))}}),xD=1,Kge=2,Fm=4,Nm=8,Im=16,Lm=32;function SD(e,t,n){this.parent=e,this.indent=t,this.flags=n,this.hash=(e?e.hash+e.hash<<8:0)+t+(t<<4)+n+(n<<6)}const lze=new SD(null,0,0);function oze(e){let t=0;for(let n=0;n<e.length;n++)t+=e.charCodeAt(n)==Zw?8-t%8:1;return t}const rY=new Map([[CQe,0],[wQe,Fm],[RQe,Nm],[AQe,Nm|Fm],[TQe,Im],[kQe,Im|Fm],[PQe,Im|Nm],[MQe,Im|Nm|Fm],[FQe,Lm],[NQe,Lm|Fm],[IQe,Lm|Nm],[LQe,Lm|Nm|Fm],[BQe,Lm|Im],[QQe,Lm|Im|Fm],[zQe,Lm|Im|Nm],[qQe,Lm|Im|Nm|Fm]].map(([e,t])=>[e,t|Kge])),sze=new Bee({start:lze,reduce(e,t,n,r){return e.flags&xD&&ize.has(t)||(t==$Qe||t==Xge)&&e.flags&Kge?e.parent:e},shift(e,t,n,r){return t==jge?new SD(e,oze(r.read(r.pos,n.pos)),0):t==Zge?e.parent:t==dQe||t==gQe||t==bQe||t==Hge?new SD(e,0,xD):rY.has(t)?new SD(e,0,rY.get(t)|e.flags&xD):e},hash(e){return e.hash}}),cze=new vg(e=>{for(let t=0;t<5;t++){if(e.next!="print".charCodeAt(t))return;e.advance()}if(!/\w/.test(String.fromCharCode(e.next)))for(let t=0;;t++){let n=e.peek(t);if(!(n==_L||n==Zw)){n!=KQe&&n!=WQe&&n!=g1&&n!=Hx&&n!=EL&&e.acceptToken(uQe);return}}}),fze=new vg((e,t)=>{let{flags:n}=t.context,r=n&Fm?Gge:Yge,i=(n&Nm)>0,a=!(n&Im),u=(n&Lm)>0,l=e.pos;for(;!(e.next<0);)if(u&&e.next==UM)if(e.peek(1)==UM)e.advance(2);else{if(e.pos==l){e.acceptToken(Hge,1);return}break}else if(a&&e.next==nY){if(e.pos==l){e.advance();let o=e.next;o>=0&&(e.advance(),dze(e,o)),e.acceptToken(fQe);return}break}else if(e.next==nY&&!a&&e.peek(1)>-1)e.advance(2);else if(e.next==r&&(!i||e.peek(1)==r&&e.peek(2)==r)){if(e.pos==l){e.acceptToken(eY,i?3:1);return}break}else if(e.next==g1){if(i)e.advance();else if(e.pos==l){e.acceptToken(eY);return}break}else e.advance();e.pos>l&&e.acceptToken(cQe)});function dze(e,t){if(t==JQe)for(let n=0;n<2&&e.next>=48&&e.next<=55;n++)e.advance();else if(t==eze)for(let n=0;n<2&&LP(e.next);n++)e.advance();else if(t==nze)for(let n=0;n<4&&LP(e.next);n++)e.advance();else if(t==rze)for(let n=0;n<8&&LP(e.next);n++)e.advance();else if(t==tze&&e.next==UM){for(e.advance();e.next>=0&&e.next!=tY&&e.next!=Yge&&e.next!=Gge&&e.next!=g1;)e.advance();e.next==tY&&e.advance()}}const hze=oN({'async "*" "**" FormatConversion FormatSpec':st.modifier,"for while if elif else try except finally return raise break continue with pass assert await yield match case":st.controlKeyword,"in not and or is del":st.operatorKeyword,"from def class global nonlocal lambda":st.definitionKeyword,import:st.moduleKeyword,"with as print":st.keyword,Boolean:st.bool,None:st.null,VariableName:st.variableName,"CallExpression/VariableName":st.function(st.variableName),"FunctionDefinition/VariableName":st.function(st.definition(st.variableName)),"ClassDefinition/VariableName":st.definition(st.className),PropertyName:st.propertyName,"CallExpression/MemberExpression/PropertyName":st.function(st.propertyName),Comment:st.lineComment,Number:st.number,String:st.string,FormatString:st.special(st.string),Escape:st.escape,UpdateOp:st.updateOperator,"ArithOp!":st.arithmeticOperator,BitOp:st.bitwiseOperator,CompareOp:st.compareOperator,AssignOp:st.definitionOperator,Ellipsis:st.punctuation,At:st.meta,"( )":st.paren,"[ ]":st.squareBracket,"{ }":st.brace,".":st.derefOperator,", ;":st.separator}),pze={__proto__:null,await:44,or:54,and:56,in:60,not:62,is:64,if:70,else:72,lambda:76,yield:94,from:96,async:102,for:104,None:162,True:164,False:164,del:178,pass:182,break:186,continue:190,return:194,raise:202,import:206,as:208,global:212,nonlocal:214,assert:218,type:223,elif:236,while:240,try:246,except:248,finally:250,with:254,def:258,class:268,match:279,case:285},Cut=w3.deserialize({version:14,states:"##jO`QeOOP$}OSOOO&WQtO'#HUOOQS'#Co'#CoOOQS'#Cp'#CpO'vQdO'#CnO*UQtO'#HTOOQS'#HU'#HUOOQS'#DU'#DUOOQS'#HT'#HTO*rQdO'#D_O+VQdO'#DfO+gQdO'#DjO+zOWO'#DuO,VOWO'#DvO.[QtO'#GuOOQS'#Gu'#GuO'vQdO'#GtO0ZQtO'#GtOOQS'#Eb'#EbO0rQdO'#EcOOQS'#Gs'#GsO0|QdO'#GrOOQV'#Gr'#GrO1XQdO'#FYOOQS'#G^'#G^O1^QdO'#FXOOQV'#IS'#ISOOQV'#Gq'#GqOOQV'#Fq'#FqQ`QeOOO'vQdO'#CqO1lQdO'#C}O1sQdO'#DRO2RQdO'#HYO2cQtO'#EVO'vQdO'#EWOOQS'#EY'#EYOOQS'#E['#E[OOQS'#E^'#E^O2wQdO'#E`O3_QdO'#EdO3rQdO'#EfO3zQtO'#EfO1XQdO'#EiO0rQdO'#ElO1XQdO'#EnO0rQdO'#EtO0rQdO'#EwO4VQdO'#EyO4^QdO'#FOO4iQdO'#EzO0rQdO'#FOO1XQdO'#FQO1XQdO'#FVO4nQdO'#F[P4uOdO'#GpPOOO)CBd)CBdOOQS'#Ce'#CeOOQS'#Cf'#CfOOQS'#Cg'#CgOOQS'#Ch'#ChOOQS'#Ci'#CiOOQS'#Cj'#CjOOQS'#Cl'#ClO'vQdO,59OO'vQdO,59OO'vQdO,59OO'vQdO,59OO'vQdO,59OO'vQdO,59OO5TQdO'#DoOOQS,5:Y,5:YO5hQdO'#HdOOQS,5:],5:]O5uQ!fO,5:]O5zQtO,59YO1lQdO,59bO1lQdO,59bO1lQdO,59bO8jQdO,59bO8oQdO,59bO8vQdO,59jO8}QdO'#HTO:TQdO'#HSOOQS'#HS'#HSOOQS'#D['#D[O:lQdO,59aO'vQdO,59aO:zQdO,59aOOQS,59y,59yO;PQdO,5:RO'vQdO,5:ROOQS,5:Q,5:QO;_QdO,5:QO;dQdO,5:XO'vQdO,5:XO'vQdO,5:VOOQS,5:U,5:UO;uQdO,5:UO;zQdO,5:WOOOW'#Fy'#FyO<POWO,5:aOOQS,5:a,5:aO<[QdO'#HwOOOW'#Dw'#DwOOOW'#Fz'#FzO<lOWO,5:bOOQS,5:b,5:bOOQS'#F}'#F}O<zQtO,5:iO?lQtO,5=`O@VQ#xO,5=`O@vQtO,5=`OOQS,5:},5:}OA_QeO'#GWOBqQdO,5;^OOQV,5=^,5=^OB|QtO'#IPOCkQdO,5;tOOQS-E:[-E:[OOQV,5;s,5;sO4dQdO'#FQOOQV-E9o-E9oOCsQtO,59]OEzQtO,59iOFeQdO'#HVOFpQdO'#HVO1XQdO'#HVOF{QdO'#DTOGTQdO,59mOGYQdO'#HZO'vQdO'#HZO0rQdO,5=tOOQS,5=t,5=tO0rQdO'#EROOQS'#ES'#ESOGwQdO'#GPOHXQdO,58|OHXQdO,58|O*xQdO,5:oOHgQtO'#H]OOQS,5:r,5:rOOQS,5:z,5:zOHzQdO,5;OOI]QdO'#IOO1XQdO'#H}OOQS,5;Q,5;QOOQS'#GT'#GTOIqQtO,5;QOJPQdO,5;QOJUQdO'#IQOOQS,5;T,5;TOJdQdO'#H|OOQS,5;W,5;WOJuQdO,5;YO4iQdO,5;`O4iQdO,5;cOJ}QtO'#ITO'vQdO'#ITOKXQdO,5;eO4VQdO,5;eO0rQdO,5;jO1XQdO,5;lOK^QeO'#EuOLjQgO,5;fO!!kQdO'#IUO4iQdO,5;jO!!vQdO,5;lO!#OQdO,5;qO!#ZQtO,5;vO'vQdO,5;vPOOO,5=[,5=[P!#bOSO,5=[P!#jOdO,5=[O!&bQtO1G.jO!&iQtO1G.jO!)YQtO1G.jO!)dQtO1G.jO!+}QtO1G.jO!,bQtO1G.jO!,uQdO'#HcO!-TQtO'#GuO0rQdO'#HcO!-_QdO'#HbOOQS,5:Z,5:ZO!-gQdO,5:ZO!-lQdO'#HeO!-wQdO'#HeO!.[QdO,5>OOOQS'#Ds'#DsOOQS1G/w1G/wOOQS1G.|1G.|O!/[QtO1G.|O!/cQtO1G.|O1lQdO1G.|O!0OQdO1G/UOOQS'#DZ'#DZO0rQdO,59tOOQS1G.{1G.{O!0VQdO1G/eO!0gQdO1G/eO!0oQdO1G/fO'vQdO'#H[O!0tQdO'#H[O!0yQtO1G.{O!1ZQdO,59iO!2aQdO,5=zO!2qQdO,5=zO!2yQdO1G/mO!3OQtO1G/mOOQS1G/l1G/lO!3`QdO,5=uO!4VQdO,5=uO0rQdO1G/qO!4tQdO1G/sO!4yQtO1G/sO!5ZQtO1G/qOOQS1G/p1G/pOOQS1G/r1G/rOOOW-E9w-E9wOOQS1G/{1G/{O!5kQdO'#HxO0rQdO'#HxO!5|QdO,5>cOOOW-E9x-E9xOOQS1G/|1G/|OOQS-E9{-E9{O!6[Q#xO1G2zO!6{QtO1G2zO'vQdO,5<jOOQS,5<j,5<jOOQS-E9|-E9|OOQS,5<r,5<rOOQS-E:U-E:UOOQV1G0x1G0xO1XQdO'#GRO!7dQtO,5>kOOQS1G1`1G1`O!8RQdO1G1`OOQS'#DV'#DVO0rQdO,5=qOOQS,5=q,5=qO!8WQdO'#FrO!8cQdO,59oO!8kQdO1G/XO!8uQtO,5=uOOQS1G3`1G3`OOQS,5:m,5:mO!9fQdO'#GtOOQS,5<k,5<kOOQS-E9}-E9}O!9wQdO1G.hOOQS1G0Z1G0ZO!:VQdO,5=wO!:gQdO,5=wO0rQdO1G0jO0rQdO1G0jO!:xQdO,5>jO!;ZQdO,5>jO1XQdO,5>jO!;lQdO,5>iOOQS-E:R-E:RO!;qQdO1G0lO!;|QdO1G0lO!<RQdO,5>lO!<aQdO,5>lO!<oQdO,5>hO!=VQdO,5>hO!=hQdO'#EpO0rQdO1G0tO!=sQdO1G0tO!=xQgO1G0zO!AvQgO1G0}O!EqQdO,5>oO!E{QdO,5>oO!FTQtO,5>oO0rQdO1G1PO!F_QdO1G1PO4iQdO1G1UO!!vQdO1G1WOOQV,5;a,5;aO!FdQfO,5;aO!FiQgO1G1QO!JjQdO'#GZO4iQdO1G1QO4iQdO1G1QO!JzQdO,5>pO!KXQdO,5>pO1XQdO,5>pOOQV1G1U1G1UO!KaQdO'#FSO!KrQ!fO1G1WO!KzQdO1G1WOOQV1G1]1G1]O4iQdO1G1]O!LPQdO1G1]O!LXQdO'#F^OOQV1G1b1G1bO!#ZQtO1G1bPOOO1G2v1G2vP!L^OSO1G2vOOQS,5=},5=}OOQS'#Dp'#DpO0rQdO,5=}O!LfQdO,5=|O!LyQdO,5=|OOQS1G/u1G/uO!MRQdO,5>PO!McQdO,5>PO!MkQdO,5>PO!NOQdO,5>PO!N`QdO,5>POOQS1G3j1G3jOOQS7+$h7+$hO!8kQdO7+$pO#!RQdO1G.|O#!YQdO1G.|OOQS1G/`1G/`OOQS,5<`,5<`O'vQdO,5<`OOQS7+%P7+%PO#!aQdO7+%POOQS-E9r-E9rOOQS7+%Q7+%QO#!qQdO,5=vO'vQdO,5=vOOQS7+$g7+$gO#!vQdO7+%PO##OQdO7+%QO##TQdO1G3fOOQS7+%X7+%XO##eQdO1G3fO##mQdO7+%XOOQS,5<_,5<_O'vQdO,5<_O##rQdO1G3aOOQS-E9q-E9qO#$iQdO7+%]OOQS7+%_7+%_O#$wQdO1G3aO#%fQdO7+%_O#%kQdO1G3gO#%{QdO1G3gO#&TQdO7+%]O#&YQdO,5>dO#&sQdO,5>dO#&sQdO,5>dOOQS'#Dx'#DxO#'UO&jO'#DzO#'aO`O'#HyOOOW1G3}1G3}O#'fQdO1G3}O#'nQdO1G3}O#'yQ#xO7+(fO#(jQtO1G2UP#)TQdO'#GOOOQS,5<m,5<mOOQS-E:P-E:POOQS7+&z7+&zOOQS1G3]1G3]OOQS,5<^,5<^OOQS-E9p-E9pOOQS7+$s7+$sO#)bQdO,5=`O#){QdO,5=`O#*^QtO,5<aO#*qQdO1G3cOOQS-E9s-E9sOOQS7+&U7+&UO#+RQdO7+&UO#+aQdO,5<nO#+uQdO1G4UOOQS-E:Q-E:QO#,WQdO1G4UOOQS1G4T1G4TOOQS7+&W7+&WO#,iQdO7+&WOOQS,5<p,5<pO#,tQdO1G4WOOQS-E:S-E:SOOQS,5<l,5<lO#-SQdO1G4SOOQS-E:O-E:OO1XQdO'#EqO#-jQdO'#EqO#-uQdO'#IRO#-}QdO,5;[OOQS7+&`7+&`O0rQdO7+&`O#.SQgO7+&fO!JmQdO'#GXO4iQdO7+&fO4iQdO7+&iO#2QQtO,5<tO'vQdO,5<tO#2[QdO1G4ZOOQS-E:W-E:WO#2fQdO1G4ZO4iQdO7+&kO0rQdO7+&kOOQV7+&p7+&pO!KrQ!fO7+&rO!KzQdO7+&rO`QeO1G0{OOQV-E:X-E:XO4iQdO7+&lO4iQdO7+&lOOQV,5<u,5<uO#2nQdO,5<uO!JmQdO,5<uOOQV7+&l7+&lO#2yQgO7+&lO#6tQdO,5<vO#7PQdO1G4[OOQS-E:Y-E:YO#7^QdO1G4[O#7fQdO'#IWO#7tQdO'#IWO1XQdO'#IWOOQS'#IW'#IWO#8PQdO'#IVOOQS,5;n,5;nO#8XQdO,5;nO0rQdO'#FUOOQV7+&r7+&rO4iQdO7+&rOOQV7+&w7+&wO4iQdO7+&wO#8^QfO,5;xOOQV7+&|7+&|POOO7+(b7+(bO#8cQdO1G3iOOQS,5<c,5<cO#8qQdO1G3hOOQS-E9u-E9uO#9UQdO,5<dO#9aQdO,5<dO#9tQdO1G3kOOQS-E9v-E9vO#:UQdO1G3kO#:^QdO1G3kO#:nQdO1G3kO#:UQdO1G3kOOQS<<H[<<H[O#:yQtO1G1zOOQS<<Hk<<HkP#;WQdO'#FtO8vQdO1G3bO#;eQdO1G3bO#;jQdO<<HkOOQS<<Hl<<HlO#;zQdO7+)QOOQS<<Hs<<HsO#<[QtO1G1yP#<{QdO'#FsO#=YQdO7+)RO#=jQdO7+)RO#=rQdO<<HwO#=wQdO7+({OOQS<<Hy<<HyO#>nQdO,5<bO'vQdO,5<bOOQS-E9t-E9tOOQS<<Hw<<HwOOQS,5<g,5<gO0rQdO,5<gO#>sQdO1G4OOOQS-E9y-E9yO#?^QdO1G4OO<[QdO'#H{OOOO'#D{'#D{OOOO'#F|'#F|O#?oO&jO,5:fOOOW,5>e,5>eOOOW7+)i7+)iO#?zQdO7+)iO#@SQdO1G2zO#@mQdO1G2zP'vQdO'#FuO0rQdO<<IpO1XQdO1G2YP1XQdO'#GSO#AOQdO7+)pO#AaQdO7+)pOOQS<<Ir<<IrP1XQdO'#GUP0rQdO'#GQOOQS,5;],5;]O#ArQdO,5>mO#BQQdO,5>mOOQS1G0v1G0vOOQS<<Iz<<IzOOQV-E:V-E:VO4iQdO<<JQOOQV,5<s,5<sO4iQdO,5<sOOQV<<JQ<<JQOOQV<<JT<<JTO#BYQtO1G2`P#BdQdO'#GYO#BkQdO7+)uO#BuQgO<<JVO4iQdO<<JVOOQV<<J^<<J^O4iQdO<<J^O!KrQ!fO<<J^O#FpQgO7+&gOOQV<<JW<<JWO#FzQgO<<JWOOQV1G2a1G2aO1XQdO1G2aO#JuQdO1G2aO4iQdO<<JWO1XQdO1G2bP0rQdO'#G[O#KQQdO7+)vO#K_QdO7+)vOOQS'#FT'#FTO0rQdO,5>rO#KgQdO,5>rO#KrQdO,5>rO#K}QdO,5>qO#L`QdO,5>qOOQS1G1Y1G1YOOQS,5;p,5;pOOQV<<Jc<<JcO#LhQdO1G1dOOQS7+)T7+)TP#LmQdO'#FwO#L}QdO1G2OO#MbQdO1G2OO#MrQdO1G2OP#M}QdO'#FxO#N[QdO7+)VO#NlQdO7+)VO#NlQdO7+)VO#NtQdO7+)VO$ UQdO7+(|O8vQdO7+(|OOQSAN>VAN>VO$ oQdO<<LmOOQSAN>cAN>cO0rQdO1G1|O$!PQtO1G1|P$!ZQdO'#FvOOQS1G2R1G2RP$!hQdO'#F{O$!uQdO7+)jO$#`QdO,5>gOOOO-E9z-E9zOOOW<<MT<<MTO$#nQdO7+(fOOQSAN?[AN?[OOQS7+'t7+'tO$$XQdO<<M[OOQS,5<q,5<qO$$jQdO1G4XOOQS-E:T-E:TOOQVAN?lAN?lOOQV1G2_1G2_O4iQdOAN?qO$$xQgOAN?qOOQVAN?xAN?xO4iQdOAN?xOOQV<<JR<<JRO4iQdOAN?rO4iQdO7+'{OOQV7+'{7+'{O1XQdO7+'{OOQVAN?rAN?rOOQS7+'|7+'|O$(sQdO<<MbOOQS1G4^1G4^O0rQdO1G4^OOQS,5<w,5<wO$)QQdO1G4]OOQS-E:Z-E:ZOOQU'#G_'#G_O$)cQfO7+'OO$)nQdO'#F_O$*uQdO7+'jO$+VQdO7+'jOOQS7+'j7+'jO$+bQdO<<LqO$+rQdO<<LqO$+rQdO<<LqO$+zQdO'#H^OOQS<<Lh<<LhO$,UQdO<<LhOOQS7+'h7+'hOOQS'#D|'#D|OOOO1G4R1G4RO$,oQdO1G4RO$,wQdO1G4RP!=hQdO'#GVOOQVG25]G25]O4iQdOG25]OOQVG25dG25dOOQVG25^G25^OOQV<<Kg<<KgO4iQdO<<KgOOQS7+)x7+)xP$-SQdO'#G]OOQU-E:]-E:]OOQV<<Jj<<JjO$-vQtO'#FaOOQS'#Fc'#FcO$.WQdO'#FbO$.xQdO'#FbOOQS'#Fb'#FbO$.}QdO'#IYO$)nQdO'#FiO$)nQdO'#FiO$/fQdO'#FjO$)nQdO'#FkO$/mQdO'#IZOOQS'#IZ'#IZO$0[QdO,5;yOOQS<<KU<<KUO$0dQdO<<KUO$0tQdOANB]O$1UQdOANB]O$1^QdO'#H_OOQS'#H_'#H_O1sQdO'#DcO$1wQdO,5=xOOQSANBSANBSOOOO7+)m7+)mO$2`QdO7+)mOOQVLD*wLD*wOOQVANARANARO5uQ!fO'#GaO$2hQtO,5<SO$)nQdO'#FmOOQS,5<W,5<WOOQS'#Fd'#FdO$3YQdO,5;|O$3_QdO,5;|OOQS'#Fg'#FgO$)nQdO'#G`O$4PQdO,5<QO$4kQdO,5>tO$4{QdO,5>tO1XQdO,5<PO$5^QdO,5<TO$5cQdO,5<TO$)nQdO'#I[O$5hQdO'#I[O$5mQdO,5<UOOQS,5<V,5<VO0rQdO'#FpOOQU1G1e1G1eO4iQdO1G1eOOQSAN@pAN@pO$5rQdOG27wO$6SQdO,59}OOQS1G3d1G3dOOOO<<MX<<MXOOQS,5<{,5<{OOQS-E:_-E:_O$6XQtO'#FaO$6`QdO'#I]O$6nQdO'#I]O$6vQdO,5<XOOQS1G1h1G1hO$6{QdO1G1hO$7QQdO,5<zOOQS-E:^-E:^O$7lQdO,5=OO$8TQdO1G4`OOQS-E:b-E:bOOQS1G1k1G1kOOQS1G1o1G1oO$8eQdO,5>vO$)nQdO,5>vOOQS1G1p1G1pOOQS,5<[,5<[OOQU7+'P7+'PO$+zQdO1G/iO$)nQdO,5<YO$8sQdO,5>wO$8zQdO,5>wOOQS1G1s1G1sOOQS7+'S7+'SP$)nQdO'#GdO$9SQdO1G4bO$9^QdO1G4bO$9fQdO1G4bOOQS7+%T7+%TO$9tQdO1G1tO$:SQtO'#FaO$:ZQdO,5<}OOQS,5<},5<}O$:iQdO1G4cOOQS-E:a-E:aO$)nQdO,5<|O$:pQdO,5<|O$:uQdO7+)|OOQS-E:`-E:`O$;PQdO7+)|O$)nQdO,5<ZP$)nQdO'#GcO$;XQdO1G2hO$)nQdO1G2hP$;gQdO'#GbO$;nQdO<<MhO$;xQdO1G1uO$<WQdO7+(SO8vQdO'#C}O8vQdO,59bO8vQdO,59bO8vQdO,59bO$<fQtO,5=`O8vQdO1G.|O0rQdO1G/XO0rQdO7+$pP$<yQdO'#GOO'vQdO'#GtO$=WQdO,59bO$=]QdO,59bO$=dQdO,59mO$=iQdO1G/UO1sQdO'#DRO8vQdO,59j",stateData:"$>S~O%cOS%^OSSOS%]PQ~OPdOVaOfoOhYOopOs!POvqO!PrO!Q{O!T!SO!U!RO!XZO!][O!h`O!r`O!s`O!t`O!{tO!}uO#PvO#RwO#TxO#XyO#ZzO#^|O#_|O#a}O#c!OO#l!QO#o!TO#s!UO#u!VO#z!WO#}hO$P!XO%oRO%pRO%tSO%uWO&Z]O&[]O&]]O&^]O&_]O&`]O&a]O&b]O&c^O&d^O&e^O&f^O&g^O&h^O&i^O&j^O~O%]!YO~OV!aO_!aOa!bOh!iO!X!kO!f!mO%j![O%k!]O%l!^O%m!_O%n!_O%o!`O%p!`O%q!aO%r!aO%s!aO~Ok%xXl%xXm%xXn%xXo%xXp%xXs%xXz%xX{%xX!x%xX#g%xX%[%xX%_%xX%z%xXg%xX!T%xX!U%xX%{%xX!W%xX![%xX!Q%xX#[%xXt%xX!m%xX~P%SOfoOhYO!XZO!][O!h`O!r`O!s`O!t`O%oRO%pRO%tSO%uWO&Z]O&[]O&]]O&^]O&_]O&`]O&a]O&b]O&c^O&d^O&e^O&f^O&g^O&h^O&i^O&j^O~Oz%wX{%wX#g%wX%[%wX%_%wX%z%wX~Ok!pOl!qOm!oOn!oOo!rOp!sOs!tO!x%wX~P)pOV!zOg!|Oo0cOv0qO!PrO~P'vOV#OOo0cOv0qO!W#PO~P'vOV#SOa#TOo0cOv0qO![#UO~P'vOQ#XO%`#XO%a#ZO~OQ#^OR#[O%`#^O%a#`O~OV%iX_%iXa%iXh%iXk%iXl%iXm%iXn%iXo%iXp%iXs%iXz%iX!X%iX!f%iX%j%iX%k%iX%l%iX%m%iX%n%iX%o%iX%p%iX%q%iX%r%iX%s%iXg%iX!T%iX!U%iX~O&Z]O&[]O&]]O&^]O&_]O&`]O&a]O&b]O&c^O&d^O&e^O&f^O&g^O&h^O&i^O&j^O{%iX!x%iX#g%iX%[%iX%_%iX%z%iX%{%iX!W%iX![%iX!Q%iX#[%iXt%iX!m%iX~P,eOz#dO{%hX!x%hX#g%hX%[%hX%_%hX%z%hX~Oo0cOv0qO~P'vO#g#gO%[#iO%_#iO~O%uWO~O!T#nO#u!VO#z!WO#}hO~OopO~P'vOV#sOa#tO%uWO{wP~OV#xOo0cOv0qO!Q#yO~P'vO{#{O!x$QO%z#|O#g!yX%[!yX%_!yX~OV#xOo0cOv0qO#g#SX%[#SX%_#SX~P'vOo0cOv0qO#g#WX%[#WX%_#WX~P'vOh$WO%uWO~O!f$YO!r$YO%uWO~OV$eO~P'vO!U$gO#s$hO#u$iO~O{$jO~OV$qO~P'vOS$sO%[$rO%_$rO%c$tO~OV$}Oa$}Og%POo0cOv0qO~P'vOo0cOv0qO{%SO~P'vO&Y%UO~Oa!bOh!iO!X!kO!f!mOVba_bakbalbambanbaobapbasbazba{ba!xba#gba%[ba%_ba%jba%kba%lba%mba%nba%oba%pba%qba%rba%sba%zbagba!Tba!Uba%{ba!Wba![ba!Qba#[batba!mba~On%ZO~Oo%ZO~P'vOo0cO~P'vOk0eOl0fOm0dOn0dOo0mOp0nOs0rOg%wX!T%wX!U%wX%{%wX!W%wX![%wX!Q%wX#[%wX!m%wX~P)pO%{%]Og%vXz%vX!T%vX!U%vX!W%vX{%vX~Og%_Oz%`O!T%dO!U%cO~Og%_O~Oz%gO!T%dO!U%cO!W&SX~O!W%kO~Oz%lO{%nO!T%dO!U%cO![%}X~O![%rO~O![%sO~OQ#XO%`#XO%a%uO~OV%wOo0cOv0qO!PrO~P'vOQ#^OR#[O%`#^O%a%zO~OV!qa_!qaa!qah!qak!qal!qam!qan!qao!qap!qas!qaz!qa{!qa!X!qa!f!qa!x!qa#g!qa%[!qa%_!qa%j!qa%k!qa%l!qa%m!qa%n!qa%o!qa%p!qa%q!qa%r!qa%s!qa%z!qag!qa!T!qa!U!qa%{!qa!W!qa![!qa!Q!qa#[!qat!qa!m!qa~P#yOz%|O{%ha!x%ha#g%ha%[%ha%_%ha%z%ha~P%SOV&OOopOvqO{%ha!x%ha#g%ha%[%ha%_%ha%z%ha~P'vOz%|O{%ha!x%ha#g%ha%[%ha%_%ha%z%ha~OPdOVaOopOvqO!PrO!Q{O!{tO!}uO#PvO#RwO#TxO#XyO#ZzO#^|O#_|O#a}O#c!OO#g$zX%[$zX%_$zX~P'vO#g#gO%[&TO%_&TO~O!f&UOh&sX%[&sXz&sX#[&sX#g&sX%_&sX#Z&sXg&sX~Oh!iO%[&WO~Okealeameaneaoeapeaseazea{ea!xea#gea%[ea%_ea%zeagea!Tea!Uea%{ea!Wea![ea!Qea#[eatea!mea~P%SOsqazqa{qa#gqa%[qa%_qa%zqa~Ok!pOl!qOm!oOn!oOo!rOp!sO!xqa~PEcO%z&YOz%yX{%yX~O%uWOz%yX{%yX~Oz&]O{wX~O{&_O~Oz%lO#g%}X%[%}X%_%}Xg%}X{%}X![%}X!m%}X%z%}X~OV0lOo0cOv0qO!PrO~P'vO%z#|O#gUa%[Ua%_Ua~Oz&hO#g&PX%[&PX%_&PXn&PX~P%SOz&kO!Q&jO#g#Wa%[#Wa%_#Wa~Oz&lO#[&nO#g&rX%[&rX%_&rXg&rX~O!f$YO!r$YO#Z&qO%uWO~O#Z&qO~Oz&sO#g&tX%[&tX%_&tX~Oz&uO#g&pX%[&pX%_&pX{&pX~O!X&wO%z&xO~Oz&|On&wX~P%SOn'PO~OPdOVaOopOvqO!PrO!Q{O!{tO!}uO#PvO#RwO#TxO#XyO#ZzO#^|O#_|O#a}O#c!OO%['UO~P'vOt'YO#p'WO#q'XOP#naV#naf#nah#nao#nas#nav#na!P#na!Q#na!T#na!U#na!X#na!]#na!h#na!r#na!s#na!t#na!{#na!}#na#P#na#R#na#T#na#X#na#Z#na#^#na#_#na#a#na#c#na#l#na#o#na#s#na#u#na#z#na#}#na$P#na%X#na%o#na%p#na%t#na%u#na&Z#na&[#na&]#na&^#na&_#na&`#na&a#na&b#na&c#na&d#na&e#na&f#na&g#na&h#na&i#na&j#na%Z#na%_#na~Oz'ZO#[']O{&xX~Oh'_O!X&wO~Oh!iO{$jO!X&wO~O{'eO~P%SO%['hO%_'hO~OS'iO%['hO%_'hO~OV!aO_!aOa!bOh!iO!X!kO!f!mO%l!^O%m!_O%n!_O%o!`O%p!`O%q!aO%r!aO%s!aOkWilWimWinWioWipWisWizWi{Wi!xWi#gWi%[Wi%_Wi%jWi%zWigWi!TWi!UWi%{Wi!WWi![Wi!QWi#[WitWi!mWi~O%k!]O~P!#uO%kWi~P!#uOV!aO_!aOa!bOh!iO!X!kO!f!mO%o!`O%p!`O%q!aO%r!aO%s!aOkWilWimWinWioWipWisWizWi{Wi!xWi#gWi%[Wi%_Wi%jWi%kWi%lWi%zWigWi!TWi!UWi%{Wi!WWi![Wi!QWi#[WitWi!mWi~O%m!_O%n!_O~P!&pO%mWi%nWi~P!&pOa!bOh!iO!X!kO!f!mOkWilWimWinWioWipWisWizWi{Wi!xWi#gWi%[Wi%_Wi%jWi%kWi%lWi%mWi%nWi%oWi%pWi%zWigWi!TWi!UWi%{Wi!WWi![Wi!QWi#[WitWi!mWi~OV!aO_!aO%q!aO%r!aO%s!aO~P!)nOVWi_Wi%qWi%rWi%sWi~P!)nO!T%dO!U%cOg&VXz&VX~O%z'kO%{'kO~P,eOz'mOg&UX~Og'oO~Oz'pO{'rO!W&XX~Oo0cOv0qOz'pO{'sO!W&XX~P'vO!W'uO~Om!oOn!oOo!rOp!sOkjisjizji{ji!xji#gji%[ji%_ji%zji~Ol!qO~P!.aOlji~P!.aOk0eOl0fOm0dOn0dOo0mOp0nO~Ot'wO~P!/jOV'|Og'}Oo0cOv0qO~P'vOg'}Oz(OO~Og(QO~O!U(SO~Og(TOz(OO!T%dO!U%cO~P%SOk0eOl0fOm0dOn0dOo0mOp0nOgqa!Tqa!Uqa%{qa!Wqa![qa!Qqa#[qatqa!mqa~PEcOV'|Oo0cOv0qO!W&Sa~P'vOz(WO!W&Sa~O!W(XO~Oz(WO!T%dO!U%cO!W&Sa~P%SOV(]Oo0cOv0qO![%}a#g%}a%[%}a%_%}ag%}a{%}a!m%}a%z%}a~P'vOz(^O![%}a#g%}a%[%}a%_%}ag%}a{%}a!m%}a%z%}a~O![(aO~Oz(^O!T%dO!U%cO![%}a~P%SOz(dO!T%dO!U%cO![&Ta~P%SOz(gO{&lX![&lX!m&lX%z&lX~O{(kO![(mO!m(nO%z(jO~OV&OOopOvqO{%hi!x%hi#g%hi%[%hi%_%hi%z%hi~P'vOz(pO{%hi!x%hi#g%hi%[%hi%_%hi%z%hi~O!f&UOh&sa%[&saz&sa#[&sa#g&sa%_&sa#Z&sag&sa~O%[(uO~OV#sOa#tO%uWO~Oz&]O{wa~OopOvqO~P'vOz(^O#g%}a%[%}a%_%}ag%}a{%}a![%}a!m%}a%z%}a~P%SOz(zO#g%hX%[%hX%_%hX%z%hX~O%z#|O#gUi%[Ui%_Ui~O#g&Pa%[&Pa%_&Pan&Pa~P'vOz(}O#g&Pa%[&Pa%_&Pan&Pa~O%uWO#g&ra%[&ra%_&rag&ra~Oz)SO#g&ra%[&ra%_&rag&ra~Og)VO~OV)WOh$WO%uWO~O#Z)XO~O%uWO#g&ta%[&ta%_&ta~Oz)ZO#g&ta%[&ta%_&ta~Oo0cOv0qO#g&pa%[&pa%_&pa{&pa~P'vOz)^O#g&pa%[&pa%_&pa{&pa~OV)`Oa)`O%uWO~O%z)eO~Ot)hO#j)gOP#hiV#hif#hih#hio#his#hiv#hi!P#hi!Q#hi!T#hi!U#hi!X#hi!]#hi!h#hi!r#hi!s#hi!t#hi!{#hi!}#hi#P#hi#R#hi#T#hi#X#hi#Z#hi#^#hi#_#hi#a#hi#c#hi#l#hi#o#hi#s#hi#u#hi#z#hi#}#hi$P#hi%X#hi%o#hi%p#hi%t#hi%u#hi&Z#hi&[#hi&]#hi&^#hi&_#hi&`#hi&a#hi&b#hi&c#hi&d#hi&e#hi&f#hi&g#hi&h#hi&i#hi&j#hi%Z#hi%_#hi~Ot)iOP#kiV#kif#kih#kio#kis#kiv#ki!P#ki!Q#ki!T#ki!U#ki!X#ki!]#ki!h#ki!r#ki!s#ki!t#ki!{#ki!}#ki#P#ki#R#ki#T#ki#X#ki#Z#ki#^#ki#_#ki#a#ki#c#ki#l#ki#o#ki#s#ki#u#ki#z#ki#}#ki$P#ki%X#ki%o#ki%p#ki%t#ki%u#ki&Z#ki&[#ki&]#ki&^#ki&_#ki&`#ki&a#ki&b#ki&c#ki&d#ki&e#ki&f#ki&g#ki&h#ki&i#ki&j#ki%Z#ki%_#ki~OV)kOn&wa~P'vOz)lOn&wa~Oz)lOn&wa~P%SOn)pO~O%Y)tO~Ot)wO#p'WO#q)vOP#niV#nif#nih#nio#nis#niv#ni!P#ni!Q#ni!T#ni!U#ni!X#ni!]#ni!h#ni!r#ni!s#ni!t#ni!{#ni!}#ni#P#ni#R#ni#T#ni#X#ni#Z#ni#^#ni#_#ni#a#ni#c#ni#l#ni#o#ni#s#ni#u#ni#z#ni#}#ni$P#ni%X#ni%o#ni%p#ni%t#ni%u#ni&Z#ni&[#ni&]#ni&^#ni&_#ni&`#ni&a#ni&b#ni&c#ni&d#ni&e#ni&f#ni&g#ni&h#ni&i#ni&j#ni%Z#ni%_#ni~OV)zOo0cOv0qO{$jO~P'vOo0cOv0qO{&xa~P'vOz*OO{&xa~OV*SOa*TOg*WO%q*UO%uWO~O{$jO&{*YO~Oh'_O~Oh!iO{$jO~O%[*_O~O%[*aO%_*aO~OV$}Oa$}Oo0cOv0qOg&Ua~P'vOz*dOg&Ua~Oo0cOv0qO{*gO!W&Xa~P'vOz*hO!W&Xa~Oo0cOv0qOz*hO{*kO!W&Xa~P'vOo0cOv0qOz*hO!W&Xa~P'vOz*hO{*kO!W&Xa~Om0dOn0dOo0mOp0nOgjikjisjizji!Tji!Uji%{ji!Wji{ji![ji#gji%[ji%_ji!Qji#[jitji!mji%zji~Ol0fO~P!NkOlji~P!NkOV'|Og*pOo0cOv0qO~P'vOn*rO~Og*pOz*tO~Og*uO~OV'|Oo0cOv0qO!W&Si~P'vOz*vO!W&Si~O!W*wO~OV(]Oo0cOv0qO![%}i#g%}i%[%}i%_%}ig%}i{%}i!m%}i%z%}i~P'vOz*zO!T%dO!U%cO![&Ti~Oz*}O![%}i#g%}i%[%}i%_%}ig%}i{%}i!m%}i%z%}i~O![+OO~Oa+QOo0cOv0qO![&Ti~P'vOz*zO![&Ti~O![+SO~OV+UOo0cOv0qO{&la![&la!m&la%z&la~P'vOz+VO{&la![&la!m&la%z&la~O!]+YO&n+[O![!nX~O![+^O~O{(kO![+_O~O{(kO![+_O!m+`O~OV&OOopOvqO{%hq!x%hq#g%hq%[%hq%_%hq%z%hq~P'vOz$ri{$ri!x$ri#g$ri%[$ri%_$ri%z$ri~P%SOV&OOopOvqO~P'vOV&OOo0cOv0qO#g%ha%[%ha%_%ha%z%ha~P'vOz+aO#g%ha%[%ha%_%ha%z%ha~Oz$ia#g$ia%[$ia%_$ian$ia~P%SO#g&Pi%[&Pi%_&Pin&Pi~P'vOz+dO#g#Wq%[#Wq%_#Wq~O#[+eOz$va#g$va%[$va%_$vag$va~O%uWO#g&ri%[&ri%_&rig&ri~Oz+gO#g&ri%[&ri%_&rig&ri~OV+iOh$WO%uWO~O%uWO#g&ti%[&ti%_&ti~Oo0cOv0qO#g&pi%[&pi%_&pi{&pi~P'vO{#{Oz#eX!W#eX~Oz+mO!W&uX~O!W+oO~Ot+rO#j)gOP#hqV#hqf#hqh#hqo#hqs#hqv#hq!P#hq!Q#hq!T#hq!U#hq!X#hq!]#hq!h#hq!r#hq!s#hq!t#hq!{#hq!}#hq#P#hq#R#hq#T#hq#X#hq#Z#hq#^#hq#_#hq#a#hq#c#hq#l#hq#o#hq#s#hq#u#hq#z#hq#}#hq$P#hq%X#hq%o#hq%p#hq%t#hq%u#hq&Z#hq&[#hq&]#hq&^#hq&_#hq&`#hq&a#hq&b#hq&c#hq&d#hq&e#hq&f#hq&g#hq&h#hq&i#hq&j#hq%Z#hq%_#hq~On$|az$|a~P%SOV)kOn&wi~P'vOz+yOn&wi~Oz,TO{$jO#[,TO~O#q,VOP#nqV#nqf#nqh#nqo#nqs#nqv#nq!P#nq!Q#nq!T#nq!U#nq!X#nq!]#nq!h#nq!r#nq!s#nq!t#nq!{#nq!}#nq#P#nq#R#nq#T#nq#X#nq#Z#nq#^#nq#_#nq#a#nq#c#nq#l#nq#o#nq#s#nq#u#nq#z#nq#}#nq$P#nq%X#nq%o#nq%p#nq%t#nq%u#nq&Z#nq&[#nq&]#nq&^#nq&_#nq&`#nq&a#nq&b#nq&c#nq&d#nq&e#nq&f#nq&g#nq&h#nq&i#nq&j#nq%Z#nq%_#nq~O#[,WOz%Oa{%Oa~Oo0cOv0qO{&xi~P'vOz,YO{&xi~O{#{O%z,[Og&zXz&zX~O%uWOg&zXz&zX~Oz,`Og&yX~Og,bO~O%Y,eO~O!T%dO!U%cOg&Viz&Vi~OV$}Oa$}Oo0cOv0qOg&Ui~P'vO{,hOz$la!W$la~Oo0cOv0qO{,iOz$la!W$la~P'vOo0cOv0qO{*gO!W&Xi~P'vOz,lO!W&Xi~Oo0cOv0qOz,lO!W&Xi~P'vOz,lO{,oO!W&Xi~Og$hiz$hi!W$hi~P%SOV'|Oo0cOv0qO~P'vOn,qO~OV'|Og,rOo0cOv0qO~P'vOV'|Oo0cOv0qO!W&Sq~P'vOz$gi![$gi#g$gi%[$gi%_$gig$gi{$gi!m$gi%z$gi~P%SOV(]Oo0cOv0qO~P'vOa+QOo0cOv0qO![&Tq~P'vOz,sO![&Tq~O![,tO~OV(]Oo0cOv0qO![%}q#g%}q%[%}q%_%}qg%}q{%}q!m%}q%z%}q~P'vO{,uO~OV+UOo0cOv0qO{&li![&li!m&li%z&li~P'vOz,zO{&li![&li!m&li%z&li~O!]+YO&n+[O![!na~O{(kO![,}O~OV&OOo0cOv0qO#g%hi%[%hi%_%hi%z%hi~P'vOz-OO#g%hi%[%hi%_%hi%z%hi~O%uWO#g&rq%[&rq%_&rqg&rq~Oz-RO#g&rq%[&rq%_&rqg&rq~OV)`Oa)`O%uWO!W&ua~Oz-TO!W&ua~On$|iz$|i~P%SOV)kO~P'vOV)kOn&wq~P'vOt-XOP#myV#myf#myh#myo#mys#myv#my!P#my!Q#my!T#my!U#my!X#my!]#my!h#my!r#my!s#my!t#my!{#my!}#my#P#my#R#my#T#my#X#my#Z#my#^#my#_#my#a#my#c#my#l#my#o#my#s#my#u#my#z#my#}#my$P#my%X#my%o#my%p#my%t#my%u#my&Z#my&[#my&]#my&^#my&_#my&`#my&a#my&b#my&c#my&d#my&e#my&f#my&g#my&h#my&i#my&j#my%Z#my%_#my~O%Z-]O%_-]O~P`O#q-^OP#nyV#nyf#nyh#nyo#nys#nyv#ny!P#ny!Q#ny!T#ny!U#ny!X#ny!]#ny!h#ny!r#ny!s#ny!t#ny!{#ny!}#ny#P#ny#R#ny#T#ny#X#ny#Z#ny#^#ny#_#ny#a#ny#c#ny#l#ny#o#ny#s#ny#u#ny#z#ny#}#ny$P#ny%X#ny%o#ny%p#ny%t#ny%u#ny&Z#ny&[#ny&]#ny&^#ny&_#ny&`#ny&a#ny&b#ny&c#ny&d#ny&e#ny&f#ny&g#ny&h#ny&i#ny&j#ny%Z#ny%_#ny~Oz-aO{$jO#[-aO~Oo0cOv0qO{&xq~P'vOz-dO{&xq~O%z,[Og&zaz&za~O{#{Og&zaz&za~OV*SOa*TO%q*UO%uWOg&ya~Oz-hOg&ya~O$S-lO~OV$}Oa$}Oo0cOv0qO~P'vOo0cOv0qO{-mOz$li!W$li~P'vOo0cOv0qOz$li!W$li~P'vO{-mOz$li!W$li~Oo0cOv0qO{*gO~P'vOo0cOv0qO{*gO!W&Xq~P'vOz-pO!W&Xq~Oo0cOv0qOz-pO!W&Xq~P'vOs-sO!T%dO!U%cOg&Oq!W&Oq![&Oqz&Oq~P!/jOa+QOo0cOv0qO![&Ty~P'vOz$ji![$ji~P%SOa+QOo0cOv0qO~P'vOV+UOo0cOv0qO~P'vOV+UOo0cOv0qO{&lq![&lq!m&lq%z&lq~P'vO{(kO![-xO!m-yO%z-wO~OV&OOo0cOv0qO#g%hq%[%hq%_%hq%z%hq~P'vO%uWO#g&ry%[&ry%_&ryg&ry~OV)`Oa)`O%uWO!W&ui~Ot-}OP#m!RV#m!Rf#m!Rh#m!Ro#m!Rs#m!Rv#m!R!P#m!R!Q#m!R!T#m!R!U#m!R!X#m!R!]#m!R!h#m!R!r#m!R!s#m!R!t#m!R!{#m!R!}#m!R#P#m!R#R#m!R#T#m!R#X#m!R#Z#m!R#^#m!R#_#m!R#a#m!R#c#m!R#l#m!R#o#m!R#s#m!R#u#m!R#z#m!R#}#m!R$P#m!R%X#m!R%o#m!R%p#m!R%t#m!R%u#m!R&Z#m!R&[#m!R&]#m!R&^#m!R&_#m!R&`#m!R&a#m!R&b#m!R&c#m!R&d#m!R&e#m!R&f#m!R&g#m!R&h#m!R&i#m!R&j#m!R%Z#m!R%_#m!R~Oo0cOv0qO{&xy~P'vOV*SOa*TO%q*UO%uWOg&yi~O$S-lO%Z.VO%_.VO~OV.aOh._O!X.^O!].`O!h.YO!s.[O!t.[O%p.XO%uWO&Z]O&[]O&]]O&^]O&_]O&`]O&a]O&b]O~Oo0cOv0qOz$lq!W$lq~P'vO{.fOz$lq!W$lq~Oo0cOv0qO{*gO!W&Xy~P'vOz.gO!W&Xy~Oo0cOv.kO~P'vOs-sO!T%dO!U%cOg&Oy!W&Oy![&Oyz&Oy~P!/jO{(kO![.nO~O{(kO![.nO!m.oO~OV*SOa*TO%q*UO%uWO~Oh.tO!f.rOz$TX#[$TX%j$TXg$TX~Os$TX{$TX!W$TX![$TX~P$-bO%o.vO%p.vOs$UXz$UX{$UX#[$UX%j$UX!W$UXg$UX![$UX~O!h.xO~Oz.|O#[/OO%j.yOs&|X{&|X!W&|Xg&|X~Oa/RO~P$)zOh.tOs&}Xz&}X{&}X#[&}X%j&}X!W&}Xg&}X![&}X~Os/VO{$jO~Oo0cOv0qOz$ly!W$ly~P'vOo0cOv0qO{*gO!W&X!R~P'vOz/ZO!W&X!R~Og&RXs&RX!T&RX!U&RX!W&RX![&RXz&RX~P!/jOs-sO!T%dO!U%cOg&Qa!W&Qa![&Qaz&Qa~O{(kO![/^O~O!f.rOh$[as$[az$[a{$[a#[$[a%j$[a!W$[ag$[a![$[a~O!h/eO~O%o.vO%p.vOs$Uaz$Ua{$Ua#[$Ua%j$Ua!W$Uag$Ua![$Ua~O%j.yOs$Yaz$Ya{$Ya#[$Ya!W$Yag$Ya![$Ya~Os&|a{&|a!W&|ag&|a~P$)nOz/jOs&|a{&|a!W&|ag&|a~O!W/mO~Og/mO~O{/oO~O![/pO~Oo0cOv0qO{*gO!W&X!Z~P'vO{/sO~O%z/tO~P$-bOz/uO#[/OO%j.yOg'PX~Oz/uOg'PX~Og/wO~O!h/xO~O#[/OOs%Saz%Sa{%Sa%j%Sa!W%Sag%Sa![%Sa~O#[/OO%j.yOs%Waz%Wa{%Wa!W%Wag%Wa~Os&|i{&|i!W&|ig&|i~P$)nOz/zO#[/OO%j.yO!['Oa~Og'Pa~P$)nOz0SOg'Pa~Oa0UO!['Oi~P$)zOz0WO!['Oi~Oz0WO#[/OO%j.yO!['Oi~O#[/OO%j.yOg$biz$bi~O%z0ZO~P$-bO#[/OO%j.yOg%Vaz%Va~Og'Pi~P$)nO{0^O~Oa0UO!['Oq~P$)zOz0`O!['Oq~O#[/OO%j.yOz%Ui![%Ui~Oa0UO~P$)zOa0UO!['Oy~P$)zO#[/OO%j.yOg$ciz$ci~O#[/OO%j.yOz%Uq![%Uq~Oz+aO#g%ha%[%ha%_%ha%z%ha~P%SOV&OOo0cOv0qO~P'vOn0hO~Oo0hO~P'vO{0iO~Ot0jO~P!/jO&]&Z&j&h&i&g&f&d&e&c&b&`&a&_&^&[%u~",goto:"!=j'QPPPPPP'RP'Z*s+[+t,_,y-fP.SP'Z.r.r'ZPPP'Z2[PPPPPP2[5PPP5PP7b7k=sPP=v>h>kPP'Z'ZPP>zPP'Z'ZPP'Z'Z'Z'Z'Z?O?w'ZP?zP@QDXGuGyPG|HWH['ZPPPH_Hk'RP'R'RP'RP'RP'RP'RP'R'R'RP'RPP'RPP'RP'RPHqH}IVPI^IdPI^PI^I^PPPI^PKrPK{LVL]KrPI^LfPI^PLmLsPLwM]MzNeLwLwNkNxLwLwLwLw! ^! d! g! l! o! y!!P!!]!!o!!u!#P!#V!#s!#y!$P!$Z!$a!$g!$y!%T!%Z!%a!%k!%q!%w!%}!&T!&Z!&e!&k!&u!&{!'U!'[!'k!'s!'}!(UPPPPPPPPPPP!([!(_!(e!(n!(x!)TPPPPPPPPPPPP!-u!/Z!3^!6oPP!6w!7W!7a!8Y!8P!8c!8i!8l!8o!8r!8z!9jPPPPPPPPPPPPPPPPP!9m!9q!9wP!:]!:a!:m!:v!;S!;j!;m!;p!;v!;|!<S!<VP!<_!<h!=d!=g]eOn#g$j)t,P'}`OTYZ[adnoprtxy}!P!Q!R!U!X!c!d!e!f!g!h!i!k!o!p!q!s!t!z#O#S#T#[#d#g#x#y#{#}$Q$e$g$h$j$q$}%S%Z%^%`%c%g%l%n%w%|&O&Z&_&h&j&k&u&x&|'P'W'Z'l'm'p'r's'w'|(O(S(W(](^(d(g(p(r(z(})^)e)g)k)l)p)t)z*O*Y*d*g*h*k*q*r*t*v*y*z*}+Q+U+V+Y+a+c+d+k+x+y,P,X,Y,],g,h,i,k,l,o,q,s,u,w,y,z-O-d-f-m-p-s.f.g/V/Z/s0c0d0e0f0h0i0j0k0l0n0r{!cQ#c#p$R$d$p%e%j%p%q&`'O'g(q(|)j*o*x+w,v0g}!dQ#c#p$R$d$p$u%e%j%p%q&`'O'g(q(|)j*o*x+w,v0g!P!eQ#c#p$R$d$p$u$v%e%j%p%q&`'O'g(q(|)j*o*x+w,v0g!R!fQ#c#p$R$d$p$u$v$w%e%j%p%q&`'O'g(q(|)j*o*x+w,v0g!T!gQ#c#p$R$d$p$u$v$w$x%e%j%p%q&`'O'g(q(|)j*o*x+w,v0g!V!hQ#c#p$R$d$p$u$v$w$x$y%e%j%p%q&`'O'g(q(|)j*o*x+w,v0g!Z!hQ!n#c#p$R$d$p$u$v$w$x$y$z%e%j%p%q&`'O'g(q(|)j*o*x+w,v0g'}TOTYZ[adnoprtxy}!P!Q!R!U!X!c!d!e!f!g!h!i!k!o!p!q!s!t!z#O#S#T#[#d#g#x#y#{#}$Q$e$g$h$j$q$}%S%Z%^%`%c%g%l%n%w%|&O&Z&_&h&j&k&u&x&|'P'W'Z'l'm'p'r's'w'|(O(S(W(](^(d(g(p(r(z(})^)e)g)k)l)p)t)z*O*Y*d*g*h*k*q*r*t*v*y*z*}+Q+U+V+Y+a+c+d+k+x+y,P,X,Y,],g,h,i,k,l,o,q,s,u,w,y,z-O-d-f-m-p-s.f.g/V/Z/s0c0d0e0f0h0i0j0k0l0n0r&eVOYZ[dnprxy}!P!Q!U!i!k!o!p!q!s!t#[#d#g#y#{#}$Q$h$j$}%S%Z%^%`%g%l%n%w%|&Z&_&j&k&u&x'P'W'Z'l'm'p'r's'w(O(W(^(d(g(p(r(z)^)e)g)p)t)z*O*Y*d*g*h*k*q*r*t*v*y*z*}+U+V+Y+a+d+k,P,X,Y,],g,h,i,k,l,o,q,s,u,w,y,z-O-d-f-m-p-s.f.g/V/Z/s0c0d0e0f0h0i0j0k0n0r%oXOYZ[dnrxy}!P!Q!U!i!k#[#d#g#y#{#}$Q$h$j$}%S%^%`%g%l%n%w%|&Z&_&j&k&u&x'P'W'Z'l'm'p'r's'w(O(W(^(d(g(p(r(z)^)e)g)p)t)z*O*Y*d*g*h*k*q*t*v*y*z*}+U+V+Y+a+d+k,P,X,Y,],g,h,i,k,l,o,s,u,w,y,z-O-d-f-m-p.f.g/V/Z0i0j0kQ#vqQ/[.kR0o0q't`OTYZ[adnoprtxy}!P!Q!R!U!X!c!d!e!f!g!h!k!o!p!q!s!t!z#O#S#T#[#d#g#x#y#{#}$Q$e$g$h$j$q$}%S%Z%^%`%c%g%l%n%w%|&O&Z&_&h&j&k&u&x&|'P'W'Z'l'p'r's'w'|(O(S(W(](^(d(g(p(r(z(})^)e)g)k)l)p)t)z*O*Y*g*h*k*q*r*t*v*y*z*}+Q+U+V+Y+a+c+d+k+x+y,P,X,Y,],h,i,k,l,o,q,s,u,w,y,z-O-d-f-m-p-s.f.g/V/Z/s0c0d0e0f0h0i0j0k0l0n0rh#jhz{$W$Z&l&q)S)X+f+g-RW#rq&].k0qQ$]|Q$a!OQ$n!VQ$o!WW$|!i'm*d,gS&[#s#tQ'S$iQ(s&UQ)U&nU)Y&s)Z+jW)a&w+m-T-{Q*Q']W*R'_,`-h.TQ+l)`S,_*S*TQ-Q+eQ-_,TQ-c,WQ.R-al.W-l.^._.a.z.|/R/j/o/t/y0U0Z0^Q/S.`Q/a.tQ/l/OU0P/u0S0[X0V/z0W0_0`R&Z#r!_!wYZ!P!Q!k%S%`%g'p'r's(O(W)g*g*h*k*q*t*v,h,i,k,l,o-m-p.f.g/ZR%^!vQ!{YQ%x#[Q&d#}Q&g$QR,{+YT.j-s/s!Y!jQ!n#c#p$R$d$p$u$v$w$x$y$z%e%j%p%q&`'O'g(q(|)j*o*x+w,v0gQ&X#kQ'c$oR*^'dR'l$|Q%V!mR/_.r'|_OTYZ[adnoprtxy}!P!Q!R!U!X!c!d!e!f!g!h!i!k!o!p!q!s!t!z#O#S#T#[#d#g#x#y#{#}$Q$e$g$h$j$q$}%S%Z%^%`%c%g%l%n%w%|&O&Z&_&h&j&k&u&x&|'P'W'Z'l'm'p'r's'w'|(O(S(W(](^(d(g(p(r(z(})^)e)g)k)l)p)t)z*O*Y*d*g*h*k*q*r*t*v*y*z*}+Q+U+V+Y+a+c+d+k+x+y,P,X,Y,],g,h,i,k,l,o,q,s,u,w,y,z-O-d-f-m-p-s.f.g/V/Z/s0c0d0e0f0h0i0j0k0l0n0rS#a_#b!P.[-l.^._.`.a.t.z.|/R/j/o/t/u/y/z0S0U0W0Z0[0^0_0`'|_OTYZ[adnoprtxy}!P!Q!R!U!X!c!d!e!f!g!h!i!k!o!p!q!s!t!z#O#S#T#[#d#g#x#y#{#}$Q$e$g$h$j$q$}%S%Z%^%`%c%g%l%n%w%|&O&Z&_&h&j&k&u&x&|'P'W'Z'l'm'p'r's'w'|(O(S(W(](^(d(g(p(r(z(})^)e)g)k)l)p)t)z*O*Y*d*g*h*k*q*r*t*v*y*z*}+Q+U+V+Y+a+c+d+k+x+y,P,X,Y,],g,h,i,k,l,o,q,s,u,w,y,z-O-d-f-m-p-s.f.g/V/Z/s0c0d0e0f0h0i0j0k0l0n0rT#a_#bT#^^#_R(o%xa(l%x(n(o+`,{-y-z.oT+[(k+]R-z,{Q$PsQ+l)aQ,^*RR-e,_X#}s$O$P&fQ&y$aQ'a$nQ'd$oR)s'SQ)b&wV-S+m-T-{ZgOn$j)t,PXkOn)t,PQ$k!TQ&z$bQ&{$cQ'^$mQ'b$oQ)q'RQ)x'WQ){'XQ)|'YQ*Z'`S*]'c'dQ+s)gQ+u)hQ+v)iQ+z)oS+|)r*[Q,Q)vQ,R)wS,S)y)zQ,d*^Q-V+rQ-W+tQ-Y+{S-Z+},OQ-`,UQ-b,VQ-|-XQ.O-[Q.P-^Q.Q-_Q.p-}Q.q.RQ/W.dR/r/XWkOn)t,PR#mjQ'`$nS)r'S'aR,O)sQ,]*RR-f,^Q*['`Q+})rR-[,OZiOjn)t,PQ'f$pR*`'gT-j,e-ku.c-l.^._.a.t.z.|/R/j/o/t/u/y0S0U0Z0[0^t.c-l.^._.a.t.z.|/R/j/o/t/u/y0S0U0Z0[0^Q/S.`X0V/z0W0_0`!P.Z-l.^._.`.a.t.z.|/R/j/o/t/u/y/z0S0U0W0Z0[0^0_0`Q.w.YR/f.xg.z.].{/b/i/n/|0O0Q0]0a0bu.b-l.^._.a.t.z.|/R/j/o/t/u/y0S0U0Z0[0^X.u.W.b/a0PR/c.tV0R/u0S0[R/X.dQnOS#on,PR,P)tQ&^#uR(x&^S%m#R#wS(_%m(bT(b%p&`Q%a!yQ%h!}W(P%a%h(U(YQ(U%eR(Y%jQ&i$RR)O&iQ(e%qQ*{(`T+R(e*{Q'n%OR*e'nS'q%R%SY*i'q*j,m-q.hU*j'r's'tU,m*k*l*mS-q,n,oR.h-rQ#Y]R%t#YQ#_^R%y#_Q(h%vS+W(h+XR+X(iQ+](kR,|+]Q#b_R%{#bQ#ebQ%}#cW&Q#e%}({+bQ({&cR+b0gQ$OsS&e$O&fR&f$PQ&v$_R)_&vQ&V#jR(t&VQ&m$VS)T&m+hR+h)UQ$Z{R&p$ZQ&t$]R)[&tQ+n)bR-U+nQ#hfR&S#hQ)f&zR+q)fQ&}$dS)m&})nR)n'OQ'V$kR)u'VQ'[$lS*P'[,ZR,Z*QQ,a*VR-i,aWjOn)t,PR#ljQ-k,eR.U-kd.{.]/b/i/n/|0O0Q0]0a0bR/h.{U.s.W/a0PR/`.sQ/{/nS0X/{0YR0Y/|S/v/b/cR0T/vQ.}.]R/k.}R!ZPXmOn)t,PWlOn)t,PR'T$jYfOn$j)t,PR&R#g[sOn#g$j)t,PR&d#}&dQOYZ[dnprxy}!P!Q!U!i!k!o!p!q!s!t#[#d#g#y#{#}$Q$h$j$}%S%Z%^%`%g%l%n%w%|&Z&_&j&k&u&x'P'W'Z'l'm'p'r's'w(O(W(^(d(g(p(r(z)^)e)g)p)t)z*O*Y*d*g*h*k*q*r*t*v*y*z*}+U+V+Y+a+d+k,P,X,Y,],g,h,i,k,l,o,q,s,u,w,y,z-O-d-f-m-p-s.f.g/V/Z/s0c0d0e0f0h0i0j0k0n0rQ!nTQ#caQ#poU$Rt%c(SS$d!R$gQ$p!XQ$u!cQ$v!dQ$w!eQ$x!fQ$y!gQ$z!hQ%e!zQ%j#OQ%p#SQ%q#TQ&`#xQ'O$eQ'g$qQ(q&OU(|&h(}+cW)j&|)l+x+yQ*o'|Q*x(]Q+w)kQ,v+QR0g0lQ!yYQ!}ZQ$b!PQ$c!QQ%R!kQ't%S^'{%`%g(O(W*q*t*v^*f'p*h,k,l-p.g/ZQ*l'rQ*m'sQ+t)gQ,j*gQ,n*kQ-n,hQ-o,iQ-r,oQ.e-mR/Y.f[bOn#g$j)t,P!^!vYZ!P!Q!k%S%`%g'p'r's(O(W)g*g*h*k*q*t*v,h,i,k,l,o-m-p.f.g/ZQ#R[Q#fdS#wrxQ$UyW$_}$Q'P)pS$l!U$hW${!i'm*d,gS%v#[+Y`&P#d%|(p(r(z+a-O0kQ&a#yQ&b#{Q&c#}Q'j$}Q'z%^W([%l(^*y*}Q(`%nQ(i%wQ(v&ZS(y&_0iQ)P&jQ)Q&kU)]&u)^+kQ)d&xQ)y'WY)}'Z*O,X,Y-dQ*b'lS*n'w0jW+P(d*z,s,wW+T(g+V,y,zQ+p)eQ,U)zQ,c*YQ,x+UQ-P+dQ-e,]Q-v,uQ.S-fR/q/VhUOn#d#g$j%|&_'w(p(r)t,P%U!uYZ[drxy}!P!Q!U!i!k#[#y#{#}$Q$h$}%S%^%`%g%l%n%w&Z&j&k&u&x'P'W'Z'l'm'p'r's(O(W(^(d(g(z)^)e)g)p)z*O*Y*d*g*h*k*q*t*v*y*z*}+U+V+Y+a+d+k,X,Y,],g,h,i,k,l,o,s,u,w,y,z-O-d-f-m-p.f.g/V/Z0i0j0kQ#qpW%W!o!s0d0nQ%X!pQ%Y!qQ%[!tQ%f0cS'v%Z0hQ'x0eQ'y0fQ,p*rQ-u,qS.i-s/sR0p0rU#uq.k0qR(w&][cOn#g$j)t,PZ!xY#[#}$Q+YQ#W[Q#zrR$TxQ%b!yQ%i!}Q%o#RQ'j${Q(V%eQ(Z%jQ(c%pQ(f%qQ*|(`Q,f*bQ-t,pQ.m-uR/].lQ$StQ(R%cR*s(SQ.l-sR/}/sR#QZR#V[R%Q!iQ%O!iV*c'm*d,g!Z!lQ!n#c#p$R$d$p$u$v$w$x$y$z%e%j%p%q&`'O'g(q(|)j*o*x+w,v0gR%T!kT#]^#_Q%x#[R,{+YQ(m%xS+_(n(oQ,}+`Q-x,{S.n-y-zR/^.oT+Z(k+]Q$`}Q&g$QQ)o'PR+{)pQ$XzQ)W&qR+i)XQ$XzQ&o$WQ)W&qR+i)XQ#khW$Vz$W&q)XQ$[{Q&r$ZZ)R&l)S+f+g-RR$^|R)c&wXlOn)t,PQ$f!RR'Q$gQ$m!UR'R$hR*X'_Q*V'_V-g,`-h.TQ.d-lQ/P.^R/Q._U.]-l.^._Q/U.aQ/b.tQ/g.zU/i.|/j/yQ/n/RQ/|/oQ0O/tU0Q/u0S0[Q0]0UQ0a0ZR0b0^R/T.`R/d.t",nodeNames:"⚠ print Escape { Comment Script AssignStatement * BinaryExpression BitOp BitOp BitOp BitOp ArithOp ArithOp @ ArithOp ** UnaryExpression ArithOp BitOp AwaitExpression await ) ( ParenthesizedExpression BinaryExpression or and CompareOp in not is UnaryExpression ConditionalExpression if else LambdaExpression lambda ParamList VariableName AssignOp , : NamedExpression AssignOp YieldExpression yield from TupleExpression ComprehensionExpression async for LambdaExpression ] [ ArrayExpression ArrayComprehensionExpression } { DictionaryExpression DictionaryComprehensionExpression SetExpression SetComprehensionExpression CallExpression ArgList AssignOp MemberExpression . PropertyName Number String FormatString FormatReplacement FormatSelfDoc FormatConversion FormatSpec FormatReplacement FormatSelfDoc ContinuedString Ellipsis None Boolean TypeDef AssignOp UpdateStatement UpdateOp ExpressionStatement DeleteStatement del PassStatement pass BreakStatement break ContinueStatement continue ReturnStatement return YieldStatement PrintStatement RaiseStatement raise ImportStatement import as ScopeStatement global nonlocal AssertStatement assert TypeDefinition type TypeParamList TypeParam StatementGroup ; IfStatement Body elif WhileStatement while ForStatement TryStatement try except finally WithStatement with FunctionDefinition def ParamList AssignOp TypeDef ClassDefinition class DecoratedStatement Decorator At MatchStatement match MatchBody MatchClause case CapturePattern LiteralPattern ArithOp ArithOp AsPattern OrPattern LogicOp AttributePattern SequencePattern MappingPattern StarPattern ClassPattern PatternArgList KeywordPattern KeywordPattern Guard",maxTerm:277,context:sze,nodeProps:[["isolate",-5,4,71,72,73,77,""],["group",-15,6,85,87,88,90,92,94,96,98,99,100,102,105,108,110,"Statement Statement",-22,8,18,21,25,40,49,50,56,57,60,61,62,63,64,67,70,71,72,79,80,81,82,"Expression",-10,114,116,119,121,122,126,128,133,135,138,"Statement",-9,143,144,147,148,150,151,152,153,154,"Pattern"],["openedBy",23,"(",54,"[",58,"{"],["closedBy",24,")",55,"]",59,"}"]],propSources:[hze],skippedNodes:[0,4],repeatNodeCount:34,tokenData:"!2|~R!`OX%TXY%oY[%T[]%o]p%Tpq%oqr'ars)Yst*xtu%Tuv,dvw-hwx.Uxy/tyz0[z{0r{|2S|}2p}!O3W!O!P4_!P!Q:Z!Q!R;k!R![>_![!]Do!]!^Es!^!_FZ!_!`Gk!`!aHX!a!b%T!b!cIf!c!dJU!d!eK^!e!hJU!h!i!#f!i!tJU!t!u!,|!u!wJU!w!x!.t!x!}JU!}#O!0S#O#P&o#P#Q!0j#Q#R!1Q#R#SJU#S#T%T#T#UJU#U#VK^#V#YJU#Y#Z!#f#Z#fJU#f#g!,|#g#iJU#i#j!.t#j#oJU#o#p!1n#p#q!1s#q#r!2a#r#s!2f#s$g%T$g;'SJU;'S;=`KW<%lOJU`%YT&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%T`%lP;=`<%l%To%v]&n`%c_OX%TXY%oY[%T[]%o]p%Tpq%oq#O%T#O#P&o#P#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%To&tX&n`OY%TYZ%oZ]%T]^%o^#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tc'f[&n`O!_%T!_!`([!`#T%T#T#U(r#U#f%T#f#g(r#g#h(r#h#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tc(cTmR&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tc(yT!mR&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk)aV&n`&[ZOr%Trs)vs#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk){V&n`Or%Trs*bs#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk*iT&n`&^ZO#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%To+PZS_&n`OY*xYZ%TZ]*x]^%T^#o*x#o#p+r#p#q*x#q#r+r#r;'S*x;'S;=`,^<%lO*x_+wTS_OY+rZ]+r^;'S+r;'S;=`,W<%lO+r_,ZP;=`<%l+ro,aP;=`<%l*xj,kV%rQ&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tj-XT!xY&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tj-oV%lQ&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk.]V&n`&ZZOw%Twx.rx#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk.wV&n`Ow%Twx/^x#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk/eT&n`&]ZO#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk/{ThZ&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tc0cTgR&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk0yXVZ&n`Oz%Tz{1f{!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk1mVaR&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk2ZV%oZ&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tc2wTzR&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%To3_W%pZ&n`O!_%T!_!`-Q!`!a3w!a#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Td4OT&{S&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk4fX!fQ&n`O!O%T!O!P5R!P!Q%T!Q![6T![#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk5WV&n`O!O%T!O!P5m!P#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk5tT!rZ&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti6[a!hX&n`O!Q%T!Q![6T![!g%T!g!h7a!h!l%T!l!m9s!m#R%T#R#S6T#S#X%T#X#Y7a#Y#^%T#^#_9s#_#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti7fZ&n`O{%T{|8X|}%T}!O8X!O!Q%T!Q![8s![#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti8^V&n`O!Q%T!Q![8s![#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti8z]!hX&n`O!Q%T!Q![8s![!l%T!l!m9s!m#R%T#R#S8s#S#^%T#^#_9s#_#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti9zT!hX&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk:bX%qR&n`O!P%T!P!Q:}!Q!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tj;UV%sQ&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti;ro!hX&n`O!O%T!O!P=s!P!Q%T!Q![>_![!d%T!d!e?q!e!g%T!g!h7a!h!l%T!l!m9s!m!q%T!q!rA]!r!z%T!z!{Bq!{#R%T#R#S>_#S#U%T#U#V?q#V#X%T#X#Y7a#Y#^%T#^#_9s#_#c%T#c#dA]#d#l%T#l#mBq#m#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti=xV&n`O!Q%T!Q![6T![#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti>fc!hX&n`O!O%T!O!P=s!P!Q%T!Q![>_![!g%T!g!h7a!h!l%T!l!m9s!m#R%T#R#S>_#S#X%T#X#Y7a#Y#^%T#^#_9s#_#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti?vY&n`O!Q%T!Q!R@f!R!S@f!S#R%T#R#S@f#S#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Ti@mY!hX&n`O!Q%T!Q!R@f!R!S@f!S#R%T#R#S@f#S#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TiAbX&n`O!Q%T!Q!YA}!Y#R%T#R#SA}#S#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TiBUX!hX&n`O!Q%T!Q!YA}!Y#R%T#R#SA}#S#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TiBv]&n`O!Q%T!Q![Co![!c%T!c!iCo!i#R%T#R#SCo#S#T%T#T#ZCo#Z#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TiCv]!hX&n`O!Q%T!Q![Co![!c%T!c!iCo!i#R%T#R#SCo#S#T%T#T#ZCo#Z#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%ToDvV{_&n`O!_%T!_!`E]!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TcEdT%{R&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TkEzT#gZ&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TkFbXmR&n`O!^%T!^!_F}!_!`([!`!a([!a#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TjGUV%mQ&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TkGrV%zZ&n`O!_%T!_!`([!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TkH`WmR&n`O!_%T!_!`([!`!aHx!a#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TjIPV%nQ&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TkIoV_Q#}P&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%ToJ_]&n`&YS%uZO!Q%T!Q![JU![!c%T!c!}JU!}#R%T#R#SJU#S#T%T#T#oJU#p#q%T#r$g%T$g;'SJU;'S;=`KW<%lOJUoKZP;=`<%lJUoKge&n`&YS%uZOr%Trs)Ysw%Twx.Ux!Q%T!Q![JU![!c%T!c!tJU!t!uLx!u!}JU!}#R%T#R#SJU#S#T%T#T#fJU#f#gLx#g#oJU#p#q%T#r$g%T$g;'SJU;'S;=`KW<%lOJUoMRa&n`&YS%uZOr%TrsNWsw%Twx! vx!Q%T!Q![JU![!c%T!c!}JU!}#R%T#R#SJU#S#T%T#T#oJU#p#q%T#r$g%T$g;'SJU;'S;=`KW<%lOJUkN_V&n`&`ZOr%TrsNts#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%TkNyV&n`Or%Trs! `s#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk! gT&n`&bZO#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk! }V&n`&_ZOw%Twx!!dx#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!!iV&n`Ow%Twx!#Ox#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!#VT&n`&aZO#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%To!#oe&n`&YS%uZOr%Trs!%Qsw%Twx!&px!Q%T!Q![JU![!c%T!c!tJU!t!u!(`!u!}JU!}#R%T#R#SJU#S#T%T#T#fJU#f#g!(`#g#oJU#p#q%T#r$g%T$g;'SJU;'S;=`KW<%lOJUk!%XV&n`&dZOr%Trs!%ns#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!%sV&n`Or%Trs!&Ys#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!&aT&n`&fZO#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!&wV&n`&cZOw%Twx!'^x#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!'cV&n`Ow%Twx!'xx#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!(PT&n`&eZO#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%To!(ia&n`&YS%uZOr%Trs!)nsw%Twx!+^x!Q%T!Q![JU![!c%T!c!}JU!}#R%T#R#SJU#S#T%T#T#oJU#p#q%T#r$g%T$g;'SJU;'S;=`KW<%lOJUk!)uV&n`&hZOr%Trs!*[s#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!*aV&n`Or%Trs!*vs#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!*}T&n`&jZO#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!+eV&n`&gZOw%Twx!+zx#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!,PV&n`Ow%Twx!,fx#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tk!,mT&n`&iZO#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%To!-Vi&n`&YS%uZOr%TrsNWsw%Twx! vx!Q%T!Q![JU![!c%T!c!dJU!d!eLx!e!hJU!h!i!(`!i!}JU!}#R%T#R#SJU#S#T%T#T#UJU#U#VLx#V#YJU#Y#Z!(`#Z#oJU#p#q%T#r$g%T$g;'SJU;'S;=`KW<%lOJUo!.}a&n`&YS%uZOr%Trs)Ysw%Twx.Ux!Q%T!Q![JU![!c%T!c!}JU!}#R%T#R#SJU#S#T%T#T#oJU#p#q%T#r$g%T$g;'SJU;'S;=`KW<%lOJUk!0ZT!XZ&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tc!0qT!WR&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%Tj!1XV%kQ&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%T~!1sO!]~k!1zV%jR&n`O!_%T!_!`-Q!`#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%T~!2fO![~i!2mT%tX&n`O#o%T#p#q%T#r;'S%T;'S;=`%i<%lO%T",tokenizers:[cze,uze,aze,fze,0,1,2,3,4],topRules:{Script:[0,5]},specialized:[{term:221,get:e=>pze[e]||-1}],tokenPrec:7668}),mze=315,gze=316,iY=1,vze=2,yze=3,bze=4,Oze=317,xze=319,Sze=320,_ze=5,Eze=6,Dze=0,jM=[9,10,11,12,13,32,133,160,5760,8192,8193,8194,8195,8196,8197,8198,8199,8200,8201,8202,8232,8233,8239,8287,12288],Wge=125,$ze=59,ZM=47,Cze=42,wze=43,Rze=45,Aze=60,Tze=44,kze=63,Pze=46,Mze=91,Fze=new Bee({start:!1,shift(e,t){return t==_ze||t==Eze||t==xze?e:t==Sze},strict:!1}),Nze=new vg((e,t)=>{let{next:n}=e;(n==Wge||n==-1||t.context)&&e.acceptToken(Oze)},{contextual:!0,fallback:!0}),Ize=new vg((e,t)=>{let{next:n}=e,r;jM.indexOf(n)>-1||n==ZM&&((r=e.peek(1))==ZM||r==Cze)||n!=Wge&&n!=$ze&&n!=-1&&!t.context&&e.acceptToken(mze)},{contextual:!0}),Lze=new vg((e,t)=>{e.next==Mze&&!t.context&&e.acceptToken(gze)},{contextual:!0}),Bze=new vg((e,t)=>{let{next:n}=e;if(n==wze||n==Rze){if(e.advance(),n==e.next){e.advance();let r=!t.context&&t.canShift(iY);e.acceptToken(r?iY:vze)}}else n==kze&&e.peek(1)==Pze&&(e.advance(),e.advance(),(e.next<48||e.next>57)&&e.acceptToken(yze))},{contextual:!0});function BP(e,t){return e>=65&&e<=90||e>=97&&e<=122||e==95||e>=192||!t&&e>=48&&e<=57}const Qze=new vg((e,t)=>{if(e.next!=Aze||!t.dialectEnabled(Dze)||(e.advance(),e.next==ZM))return;let n=0;for(;jM.indexOf(e.next)>-1;)e.advance(),n++;if(BP(e.next,!0)){for(e.advance(),n++;BP(e.next,!1);)e.advance(),n++;for(;jM.indexOf(e.next)>-1;)e.advance(),n++;if(e.next==Tze)return;for(let r=0;;r++){if(r==7){if(!BP(e.next,!0))return;break}if(e.next!="extends".charCodeAt(r))break;e.advance(),n++}}e.acceptToken(bze,-n)}),zze=oN({"get set async static":st.modifier,"for while do if else switch try catch finally return throw break continue default case":st.controlKeyword,"in of await yield void typeof delete instanceof as satisfies":st.operatorKeyword,"let var const using function class extends":st.definitionKeyword,"import export from":st.moduleKeyword,"with debugger new":st.keyword,TemplateString:st.special(st.string),super:st.atom,BooleanLiteral:st.bool,this:st.self,null:st.null,Star:st.modifier,VariableName:st.variableName,"CallExpression/VariableName TaggedTemplateExpression/VariableName":st.function(st.variableName),VariableDefinition:st.definition(st.variableName),Label:st.labelName,PropertyName:st.propertyName,PrivatePropertyName:st.special(st.propertyName),"CallExpression/MemberExpression/PropertyName":st.function(st.propertyName),"FunctionDeclaration/VariableDefinition":st.function(st.definition(st.variableName)),"ClassDeclaration/VariableDefinition":st.definition(st.className),"NewExpression/VariableName":st.className,PropertyDefinition:st.definition(st.propertyName),PrivatePropertyDefinition:st.definition(st.special(st.propertyName)),UpdateOp:st.updateOperator,"LineComment Hashbang":st.lineComment,BlockComment:st.blockComment,Number:st.number,String:st.string,Escape:st.escape,ArithOp:st.arithmeticOperator,LogicOp:st.logicOperator,BitOp:st.bitwiseOperator,CompareOp:st.compareOperator,RegExp:st.regexp,Equals:st.definitionOperator,Arrow:st.function(st.punctuation),": Spread":st.punctuation,"( )":st.paren,"[ ]":st.squareBracket,"{ }":st.brace,"InterpolationStart InterpolationEnd":st.special(st.brace),".":st.derefOperator,", ;":st.separator,"@":st.meta,TypeName:st.typeName,TypeDefinition:st.definition(st.typeName),"type enum interface implements namespace module declare":st.definitionKeyword,"abstract global Privacy readonly override":st.modifier,"is keyof unique infer asserts":st.operatorKeyword,JSXAttributeValue:st.attributeValue,JSXText:st.content,"JSXStartTag JSXStartCloseTag JSXSelfCloseEndTag JSXEndTag":st.angleBracket,"JSXIdentifier JSXNameSpacedName":st.tagName,"JSXAttribute/JSXIdentifier JSXAttribute/JSXNameSpacedName":st.attributeName,"JSXBuiltin/JSXIdentifier":st.standard(st.tagName)}),qze={__proto__:null,export:20,as:25,from:33,default:36,async:41,function:42,in:52,out:55,const:56,extends:60,this:64,true:72,false:72,null:84,void:88,typeof:92,super:108,new:142,delete:154,yield:163,await:167,class:172,public:235,private:235,protected:235,readonly:237,instanceof:256,satisfies:259,import:292,keyof:349,unique:353,infer:359,asserts:395,is:397,abstract:417,implements:419,type:421,let:424,var:426,using:429,interface:435,enum:439,namespace:445,module:447,declare:451,global:455,for:474,of:483,while:486,with:490,do:494,if:498,else:500,switch:504,case:510,try:516,catch:520,finally:524,return:528,throw:532,break:536,continue:540,debugger:544},Vze={__proto__:null,async:129,get:131,set:133,declare:195,public:197,private:197,protected:197,static:199,abstract:201,override:203,readonly:209,accessor:211,new:401},Uze={__proto__:null,"<":193},wut=w3.deserialize({version:14,states:"$EOQ%TQlOOO%[QlOOO'_QpOOP(lO`OOO*zQ!0MxO'#CiO+RO#tO'#CjO+aO&jO'#CjO+oO#@ItO'#DaO.QQlO'#DgO.bQlO'#DrO%[QlO'#DzO0fQlO'#ESOOQ!0Lf'#E['#E[O1PQ`O'#EXOOQO'#Ep'#EpOOQO'#Ik'#IkO1XQ`O'#GsO1dQ`O'#EoO1iQ`O'#EoO3hQ!0MxO'#JqO6[Q!0MxO'#JrO6uQ`O'#F]O6zQ,UO'#FtOOQ!0Lf'#Ff'#FfO7VO7dO'#FfO7eQMhO'#F|O9[Q`O'#F{OOQ!0Lf'#Jr'#JrOOQ!0Lb'#Jq'#JqO9aQ`O'#GwOOQ['#K^'#K^O9lQ`O'#IXO9qQ!0LrO'#IYOOQ['#J_'#J_OOQ['#I^'#I^Q`QlOOQ`QlOOO9yQ!L^O'#DvO:QQlO'#EOO:XQlO'#EQO9gQ`O'#GsO:`QMhO'#CoO:nQ`O'#EnO:yQ`O'#EyO;OQMhO'#FeO;mQ`O'#GsOOQO'#K_'#K_O;rQ`O'#K_O<QQ`O'#G{O<QQ`O'#G|O<QQ`O'#HOO9gQ`O'#HRO<wQ`O'#HUO>`Q`O'#CeO>pQ`O'#HbO>xQ`O'#HhO>xQ`O'#HjO`QlO'#HlO>xQ`O'#HnO>xQ`O'#HqO>}Q`O'#HwO?SQ!0LsO'#H}O%[QlO'#IPO?_Q!0LsO'#IRO?jQ!0LsO'#ITO9qQ!0LrO'#IVO?uQ!0MxO'#CiO@wQpO'#DlQOQ`OOO%[QlO'#EQOA_Q`O'#ETO:`QMhO'#EnOAjQ`O'#EnOAuQ!bO'#FeOOQ['#Cg'#CgOOQ!0Lb'#Dq'#DqOOQ!0Lb'#Ju'#JuO%[QlO'#JuOOQO'#Jx'#JxOOQO'#Ig'#IgOBuQpO'#EgOOQ!0Lb'#Ef'#EfOOQ!0Lb'#J|'#J|OCqQ!0MSO'#EgOC{QpO'#EWOOQO'#Jw'#JwODaQpO'#JxOEnQpO'#EWOC{QpO'#EgPE{O&2DjO'#CbPOOO)CD|)CD|OOOO'#I_'#I_OFWO#tO,59UOOQ!0Lh,59U,59UOOOO'#I`'#I`OFfO&jO,59UOFtQ!L^O'#DcOOOO'#Ib'#IbOF{O#@ItO,59{OOQ!0Lf,59{,59{OGZQlO'#IcOGnQ`O'#JsOImQ!fO'#JsO+}QlO'#JsOItQ`O,5:ROJ[Q`O'#EpOJiQ`O'#KSOJtQ`O'#KROJtQ`O'#KROJ|Q`O,5;^OKRQ`O'#KQOOQ!0Ln,5:^,5:^OKYQlO,5:^OMWQ!0MxO,5:fOMwQ`O,5:nONbQ!0LrO'#KPONiQ`O'#KOO9aQ`O'#KOON}Q`O'#KOO! VQ`O,5;]O! [Q`O'#KOO!#aQ!fO'#JrOOQ!0Lh'#Ci'#CiO%[QlO'#ESO!$PQ!fO,5:sOOQS'#Jy'#JyOOQO-E<i-E<iO9gQ`O,5=_O!$gQ`O,5=_O!$lQlO,5;ZO!&oQMhO'#EkO!(YQ`O,5;ZO!(_QlO'#DyO!(iQpO,5;dO!(qQpO,5;dO%[QlO,5;dOOQ['#FT'#FTOOQ['#FV'#FVO%[QlO,5;eO%[QlO,5;eO%[QlO,5;eO%[QlO,5;eO%[QlO,5;eO%[QlO,5;eO%[QlO,5;eO%[QlO,5;eO%[QlO,5;eO%[QlO,5;eOOQ['#FZ'#FZO!)PQlO,5;tOOQ!0Lf,5;y,5;yOOQ!0Lf,5;z,5;zOOQ!0Lf,5;|,5;|O%[QlO'#IoO!+SQ!0LrO,5<iO%[QlO,5;eO!&oQMhO,5;eO!+qQMhO,5;eO!-cQMhO'#E^O%[QlO,5;wOOQ!0Lf,5;{,5;{O!-jQ,UO'#FjO!.gQ,UO'#KWO!.RQ,UO'#KWO!.nQ,UO'#KWOOQO'#KW'#KWO!/SQ,UO,5<SOOOW,5<`,5<`O!/eQlO'#FvOOOW'#In'#InO7VO7dO,5<QO!/lQ,UO'#FxOOQ!0Lf,5<Q,5<QO!0]Q$IUO'#CyOOQ!0Lh'#C}'#C}O!0pO#@ItO'#DRO!1^QMjO,5<eO!1eQ`O,5<hO!3QQ(CWO'#GXO!3_Q`O'#GYO!3dQ`O'#GYO!5SQ(CWO'#G^O!6XQpO'#GbOOQO'#Gn'#GnO!+xQMhO'#GmOOQO'#Gp'#GpO!+xQMhO'#GoO!6zQ$IUO'#JkOOQ!0Lh'#Jk'#JkO!7UQ`O'#JjO!7dQ`O'#JiO!7lQ`O'#CuOOQ!0Lh'#C{'#C{O!7}Q`O'#C}OOQ!0Lh'#DV'#DVOOQ!0Lh'#DX'#DXO1SQ`O'#DZO!+xQMhO'#GPO!+xQMhO'#GRO!8SQ`O'#GTO!8XQ`O'#GUO!3dQ`O'#G[O!+xQMhO'#GaO<QQ`O'#JjO!8^Q`O'#EqO!8{Q`O,5<gOOQ!0Lb'#Cr'#CrO!9TQ`O'#ErO!9}QpO'#EsOOQ!0Lb'#KQ'#KQO!:UQ!0LrO'#K`O9qQ!0LrO,5=cO`QlO,5>sOOQ['#Jg'#JgOOQ[,5>t,5>tOOQ[-E<[-E<[O!<TQ!0MxO,5:bO!9xQpO,5:`O!>nQ!0MxO,5:jO%[QlO,5:jO!AUQ!0MxO,5:lOOQO,5@y,5@yO!AuQMhO,5=_O!BTQ!0LrO'#JhO9[Q`O'#JhO!BfQ!0LrO,59ZO!BqQpO,59ZO!ByQMhO,59ZO:`QMhO,59ZO!CUQ`O,5;ZO!C^Q`O'#HaO!CrQ`O'#KcO%[QlO,5;}O!9xQpO,5<PO!CzQ`O,5=zO!DPQ`O,5=zO!DUQ`O,5=zO9qQ!0LrO,5=zO<QQ`O,5=jOOQO'#Cy'#CyO!DdQpO,5=gO!DlQMhO,5=hO!DwQ`O,5=jO!D|Q!bO,5=mO!EUQ`O'#K_O>}Q`O'#HWO9gQ`O'#HYO!EZQ`O'#HYO:`QMhO'#H[O!E`Q`O'#H[OOQ[,5=p,5=pO!EeQ`O'#H]O!EvQ`O'#CoO!E{Q`O,59PO!FVQ`O,59PO!H[QlO,59POOQ[,59P,59PO!HlQ!0LrO,59PO%[QlO,59PO!JwQlO'#HdOOQ['#He'#HeOOQ['#Hf'#HfO`QlO,5=|O!K_Q`O,5=|O`QlO,5>SO`QlO,5>UO!KdQ`O,5>WO`QlO,5>YO!KiQ`O,5>]O!KnQlO,5>cOOQ[,5>i,5>iO%[QlO,5>iO9qQ!0LrO,5>kOOQ[,5>m,5>mO# xQ`O,5>mOOQ[,5>o,5>oO# xQ`O,5>oOOQ[,5>q,5>qO#!fQpO'#D_O%[QlO'#JuO##XQpO'#JuO##cQpO'#DmO##tQpO'#DmO#&VQlO'#DmO#&^Q`O'#JtO#&fQ`O,5:WO#&kQ`O'#EtO#&yQ`O'#KTO#'RQ`O,5;_O#'WQpO'#DmO#'eQpO'#EVOOQ!0Lf,5:o,5:oO%[QlO,5:oO#'lQ`O,5:oO>}Q`O,5;YO!BqQpO,5;YO!ByQMhO,5;YO:`QMhO,5;YO#'tQ`O,5@aO#'yQ07dO,5:sOOQO-E<e-E<eO#)PQ!0MSO,5;ROC{QpO,5:rO#)ZQpO,5:rOC{QpO,5;RO!BfQ!0LrO,5:rOOQ!0Lb'#Ej'#EjOOQO,5;R,5;RO%[QlO,5;RO#)hQ!0LrO,5;RO#)sQ!0LrO,5;RO!BqQpO,5:rOOQO,5;X,5;XO#*RQ!0LrO,5;RPOOO'#I]'#I]P#*gO&2DjO,58|POOO,58|,58|OOOO-E<]-E<]OOQ!0Lh1G.p1G.pOOOO-E<^-E<^OOOO,59},59}O#*rQ!bO,59}OOOO-E<`-E<`OOQ!0Lf1G/g1G/gO#*wQ!fO,5>}O+}QlO,5>}OOQO,5?T,5?TO#+RQlO'#IcOOQO-E<a-E<aO#+`Q`O,5@_O#+hQ!fO,5@_O#+oQ`O,5@mOOQ!0Lf1G/m1G/mO%[QlO,5@nO#+wQ`O'#IiOOQO-E<g-E<gO#+oQ`O,5@mOOQ!0Lb1G0x1G0xOOQ!0Ln1G/x1G/xOOQ!0Ln1G0Y1G0YO%[QlO,5@kO#,]Q!0LrO,5@kO#,nQ!0LrO,5@kO#,uQ`O,5@jO9aQ`O,5@jO#,}Q`O,5@jO#-]Q`O'#IlO#,uQ`O,5@jOOQ!0Lb1G0w1G0wO!(iQpO,5:uO!(tQpO,5:uOOQS,5:w,5:wO#-}QdO,5:wO#.VQMhO1G2yO9gQ`O1G2yOOQ!0Lf1G0u1G0uO#.eQ!0MxO1G0uO#/jQ!0MvO,5;VOOQ!0Lh'#GW'#GWO#0WQ!0MzO'#JkO!$lQlO1G0uO#2cQ!fO'#JvO%[QlO'#JvO#2mQ`O,5:eOOQ!0Lh'#D_'#D_OOQ!0Lf1G1O1G1OO%[QlO1G1OOOQ!0Lf1G1f1G1fO#2rQ`O1G1OO#5WQ!0MxO1G1PO#5_Q!0MxO1G1PO#7uQ!0MxO1G1PO#7|Q!0MxO1G1PO#:dQ!0MxO1G1PO#<zQ!0MxO1G1PO#=RQ!0MxO1G1PO#=YQ!0MxO1G1PO#?pQ!0MxO1G1PO#?wQ!0MxO1G1PO#BUQ?MtO'#CiO#DPQ?MtO1G1`O#DWQ?MtO'#JrO#DkQ!0MxO,5?ZOOQ!0Lb-E<m-E<mO#FxQ!0MxO1G1PO#GuQ!0MzO1G1POOQ!0Lf1G1P1G1PO#HxQMjO'#J{O#ISQ`O,5:xO#IXQ!0MxO1G1cO#I{Q,UO,5<WO#JTQ,UO,5<XO#J]Q,UO'#FoO#JtQ`O'#FnOOQO'#KX'#KXOOQO'#Im'#ImO#JyQ,UO1G1nOOQ!0Lf1G1n1G1nOOOW1G1y1G1yO#K[Q?MtO'#JqO#KfQ`O,5<bO!)PQlO,5<bOOOW-E<l-E<lOOQ!0Lf1G1l1G1lO#KkQpO'#KWOOQ!0Lf,5<d,5<dO#KsQpO,5<dO#KxQMhO'#DTOOOO'#Ia'#IaO#LPO#@ItO,59mOOQ!0Lh,59m,59mO%[QlO1G2PO!8XQ`O'#IqO#L[Q`O,5<zOOQ!0Lh,5<w,5<wO!+xQMhO'#ItO#LxQMjO,5=XO!+xQMhO'#IvO#MkQMjO,5=ZO!&oQMhO,5=]OOQO1G2S1G2SO#MuQ!dO'#CrO#NYQ(CWO'#ErO$ _QpO'#GbO$ uQ!dO,5<sO$ |Q`O'#KZO9aQ`O'#KZO$![Q`O,5<uO!+xQMhO,5<tO$!aQ`O'#GZO$!rQ`O,5<tO$!wQ!dO'#GWO$#UQ!dO'#K[O$#`Q`O'#K[O!&oQMhO'#K[O$#eQ`O,5<xO$#jQlO'#JuO$#tQpO'#GcO##tQpO'#GcO$$VQ`O'#GgO!3dQ`O'#GkO$$[Q!0LrO'#IsO$$gQpO,5<|OOQ!0Lp,5<|,5<|O$$nQpO'#GcO$${QpO'#GdO$%^QpO'#GdO$%cQMjO,5=XO$%sQMjO,5=ZOOQ!0Lh,5=^,5=^O!+xQMhO,5@UO!+xQMhO,5@UO$&TQ`O'#IxO$&iQ`O,5@TO$&qQ`O,59aOOQ!0Lh,59i,59iO$'hQ$IYO,59uOOQ!0Lh'#Jo'#JoO$(ZQMjO,5<kO$(|QMjO,5<mO@oQ`O,5<oOOQ!0Lh,5<p,5<pO$)WQ`O,5<vO$)]QMjO,5<{O$)mQ`O,5@UO$){Q`O'#KOO!$lQlO1G2RO$*QQ`O1G2RO9aQ`O'#KRO9aQ`O'#EtO%[QlO'#EtO9aQ`O'#IzO$*VQ!0LrO,5@zOOQ[1G2}1G2}OOQ[1G4_1G4_OOQ!0Lf1G/|1G/|OOQ!0Lf1G/z1G/zO$,XQ!0MxO1G0UOOQ[1G2y1G2yO!&oQMhO1G2yO%[QlO1G2yO#.YQ`O1G2yO$.]QMhO'#EkOOQ!0Lb,5@S,5@SO$.jQ!0LrO,5@SOOQ[1G.u1G.uO!BfQ!0LrO1G.uO!BqQpO1G.uO!ByQMhO1G.uO$.{Q`O1G0uO$/QQ`O'#CiO$/]Q`O'#KdO$/eQ`O,5={O$/jQ`O'#KdO$/oQ`O'#KdO$/}Q`O'#JQO$0]Q`O,5@}O$0eQ!fO1G1iOOQ!0Lf1G1k1G1kO9gQ`O1G3fO@oQ`O1G3fO$0lQ`O1G3fO$0qQ`O1G3fOOQ[1G3f1G3fO!DwQ`O1G3UO!&oQMhO1G3RO$0vQ`O1G3ROOQ[1G3S1G3SO!&oQMhO1G3SO$0{Q`O1G3SO$1TQpO'#HQOOQ[1G3U1G3UO!6SQpO'#I|O!D|Q!bO1G3XOOQ[1G3X1G3XOOQ[,5=r,5=rO$1]QMhO,5=tO9gQ`O,5=tO$$VQ`O,5=vO9[Q`O,5=vO!BqQpO,5=vO!ByQMhO,5=vO:`QMhO,5=vO$1kQ`O'#KbO$1vQ`O,5=wOOQ[1G.k1G.kO$1{Q!0LrO1G.kO@oQ`O1G.kO$2WQ`O1G.kO9qQ!0LrO1G.kO$4`Q!fO,5APO$4mQ`O,5APO9aQ`O,5APO$4xQlO,5>OO$5PQ`O,5>OOOQ[1G3h1G3hO`QlO1G3hOOQ[1G3n1G3nOOQ[1G3p1G3pO>xQ`O1G3rO$5UQlO1G3tO$9YQlO'#HsOOQ[1G3w1G3wO$9gQ`O'#HyO>}Q`O'#H{OOQ[1G3}1G3}O$9oQlO1G3}O9qQ!0LrO1G4TOOQ[1G4V1G4VOOQ!0Lb'#G_'#G_O9qQ!0LrO1G4XO9qQ!0LrO1G4ZO$=vQ`O,5@aO!)PQlO,5;`O9aQ`O,5;`O>}Q`O,5:XO!)PQlO,5:XO!BqQpO,5:XO$={Q?MtO,5:XOOQO,5;`,5;`O$>VQpO'#IdO$>mQ`O,5@`OOQ!0Lf1G/r1G/rO$>uQpO'#IjO$?PQ`O,5@oOOQ!0Lb1G0y1G0yO##tQpO,5:XOOQO'#If'#IfO$?XQpO,5:qOOQ!0Ln,5:q,5:qO#'oQ`O1G0ZOOQ!0Lf1G0Z1G0ZO%[QlO1G0ZOOQ!0Lf1G0t1G0tO>}Q`O1G0tO!BqQpO1G0tO!ByQMhO1G0tOOQ!0Lb1G5{1G5{O!BfQ!0LrO1G0^OOQO1G0m1G0mO%[QlO1G0mO$?`Q!0LrO1G0mO$?kQ!0LrO1G0mO!BqQpO1G0^OC{QpO1G0^O$?yQ!0LrO1G0mOOQO1G0^1G0^O$@_Q!0MxO1G0mPOOO-E<Z-E<ZPOOO1G.h1G.hOOOO1G/i1G/iO$@iQ!bO,5<iO$@qQ!fO1G4iOOQO1G4o1G4oO%[QlO,5>}O$@{Q`O1G5yO$ATQ`O1G6XO$A]Q!fO1G6YO9aQ`O,5?TO$AgQ!0MxO1G6VO%[QlO1G6VO$AwQ!0LrO1G6VO$BYQ`O1G6UO$BYQ`O1G6UO9aQ`O1G6UO$BbQ`O,5?WO9aQ`O,5?WOOQO,5?W,5?WO$BvQ`O,5?WO$){Q`O,5?WOOQO-E<j-E<jOOQS1G0a1G0aOOQS1G0c1G0cO#.QQ`O1G0cOOQ[7+(e7+(eO!&oQMhO7+(eO%[QlO7+(eO$CUQ`O7+(eO$CaQMhO7+(eO$CoQ!0MzO,5=XO$EzQ!0MzO,5=ZO$HVQ!0MzO,5=XO$JhQ!0MzO,5=ZO$LyQ!0MzO,59uO% OQ!0MzO,5<kO%#ZQ!0MzO,5<mO%%fQ!0MzO,5<{OOQ!0Lf7+&a7+&aO%'wQ!0MxO7+&aO%(kQlO'#IeO%(xQ`O,5@bO%)QQ!fO,5@bOOQ!0Lf1G0P1G0PO%)[Q`O7+&jOOQ!0Lf7+&j7+&jO%)aQ?MtO,5:fO%[QlO7+&zO%)kQ?MtO,5:bO%)xQ?MtO,5:jO%*SQ?MtO,5:lO%*^QMhO'#IhO%*hQ`O,5@gOOQ!0Lh1G0d1G0dOOQO1G1r1G1rOOQO1G1s1G1sO%*pQ!jO,5<ZO!)PQlO,5<YOOQO-E<k-E<kOOQ!0Lf7+'Y7+'YOOOW7+'e7+'eOOOW1G1|1G1|O%*{Q`O1G1|OOQ!0Lf1G2O1G2OOOOO,59o,59oO%+QQ!dO,59oOOOO-E<_-E<_OOQ!0Lh1G/X1G/XO%+XQ!0MxO7+'kOOQ!0Lh,5?],5?]O%+{QMhO1G2fP%,SQ`O'#IqPOQ!0Lh-E<o-E<oO%,pQMjO,5?`OOQ!0Lh-E<r-E<rO%-cQMjO,5?bOOQ!0Lh-E<t-E<tO%-mQ!dO1G2wO%-tQ!dO'#CrO%.[QMhO'#KRO$#jQlO'#JuOOQ!0Lh1G2_1G2_O%.cQ`O'#IpO%.wQ`O,5@uO%.wQ`O,5@uO%/PQ`O,5@uO%/[Q`O,5@uOOQO1G2a1G2aO%/jQMjO1G2`O!+xQMhO1G2`O%/zQ(CWO'#IrO%0XQ`O,5@vO!&oQMhO,5@vO%0aQ!dO,5@vOOQ!0Lh1G2d1G2dO%2qQ!fO'#CiO%2{Q`O,5=POOQ!0Lb,5<},5<}O%3TQpO,5<}OOQ!0Lb,5=O,5=OOClQ`O,5<}O%3`QpO,5<}OOQ!0Lb,5=R,5=RO$){Q`O,5=VOOQO,5?_,5?_OOQO-E<q-E<qOOQ!0Lp1G2h1G2hO##tQpO,5<}O$#jQlO,5=PO%3nQ`O,5=OO%3yQpO,5=OO!+xQMhO'#ItO%4sQMjO1G2sO!+xQMhO'#IvO%5fQMjO1G2uO%5pQMjO1G5pO%5zQMjO1G5pOOQO,5?d,5?dOOQO-E<v-E<vOOQO1G.{1G.{O!9xQpO,59wO%[QlO,59wOOQ!0Lh,5<j,5<jO%6XQ`O1G2ZO!+xQMhO1G2bO!+xQMhO1G5pO!+xQMhO1G5pO%6^Q!0MxO7+'mOOQ!0Lf7+'m7+'mO!$lQlO7+'mO%7QQ`O,5;`OOQ!0Lb,5?f,5?fOOQ!0Lb-E<x-E<xO%7VQ!dO'#K]O#'oQ`O7+(eO4UQ!fO7+(eO$CXQ`O7+(eO%7aQ!0MvO'#CiO%7tQ!0MvO,5=SO%8fQ`O,5=SO%8nQ`O,5=SOOQ!0Lb1G5n1G5nOOQ[7+$a7+$aO!BfQ!0LrO7+$aO!BqQpO7+$aO!$lQlO7+&aO%8sQ`O'#JPO%9[Q`O,5AOOOQO1G3g1G3gO9gQ`O,5AOO%9[Q`O,5AOO%9dQ`O,5AOOOQO,5?l,5?lOOQO-E=O-E=OOOQ!0Lf7+'T7+'TO%9iQ`O7+)QO9qQ!0LrO7+)QO9gQ`O7+)QO@oQ`O7+)QOOQ[7+(p7+(pO%9nQ!0MvO7+(mO!&oQMhO7+(mO!DrQ`O7+(nOOQ[7+(n7+(nO!&oQMhO7+(nO%9xQ`O'#KaO%:TQ`O,5=lOOQO,5?h,5?hOOQO-E<z-E<zOOQ[7+(s7+(sO%;gQpO'#HZOOQ[1G3`1G3`O!&oQMhO1G3`O%[QlO1G3`O%;nQ`O1G3`O%;yQMhO1G3`O9qQ!0LrO1G3bO$$VQ`O1G3bO9[Q`O1G3bO!BqQpO1G3bO!ByQMhO1G3bO%<XQ`O'#JOO%<mQ`O,5@|O%<uQpO,5@|OOQ!0Lb1G3c1G3cOOQ[7+$V7+$VO@oQ`O7+$VO9qQ!0LrO7+$VO%=QQ`O7+$VO%[QlO1G6kO%[QlO1G6lO%=VQ!0LrO1G6kO%=aQlO1G3jO%=hQ`O1G3jO%=mQlO1G3jOOQ[7+)S7+)SO9qQ!0LrO7+)^O`QlO7+)`OOQ['#Kg'#KgOOQ['#JR'#JRO%=tQlO,5>_OOQ[,5>_,5>_O%[QlO'#HtO%>RQ`O'#HvOOQ[,5>e,5>eO9aQ`O,5>eOOQ[,5>g,5>gOOQ[7+)i7+)iOOQ[7+)o7+)oOOQ[7+)s7+)sOOQ[7+)u7+)uO%>WQpO1G5{O%>rQ?MtO1G0zO%>|Q`O1G0zOOQO1G/s1G/sO%?XQ?MtO1G/sO>}Q`O1G/sO!)PQlO'#DmOOQO,5?O,5?OOOQO-E<b-E<bOOQO,5?U,5?UOOQO-E<h-E<hO!BqQpO1G/sOOQO-E<d-E<dOOQ!0Ln1G0]1G0]OOQ!0Lf7+%u7+%uO#'oQ`O7+%uOOQ!0Lf7+&`7+&`O>}Q`O7+&`O!BqQpO7+&`OOQO7+%x7+%xO$@_Q!0MxO7+&XOOQO7+&X7+&XO%[QlO7+&XO%?cQ!0LrO7+&XO!BfQ!0LrO7+%xO!BqQpO7+%xO%?nQ!0LrO7+&XO%?|Q!0MxO7++qO%[QlO7++qO%@^Q`O7++pO%@^Q`O7++pOOQO1G4r1G4rO9aQ`O1G4rO%@fQ`O1G4rOOQS7+%}7+%}O#'oQ`O<<LPO4UQ!fO<<LPO%@tQ`O<<LPOOQ[<<LP<<LPO!&oQMhO<<LPO%[QlO<<LPO%@|Q`O<<LPO%AXQ!0MzO,5?`O%CdQ!0MzO,5?bO%EoQ!0MzO1G2`O%HQQ!0MzO1G2sO%J]Q!0MzO1G2uO%LhQ!fO,5?PO%[QlO,5?POOQO-E<c-E<cO%LrQ`O1G5|OOQ!0Lf<<JU<<JUO%LzQ?MtO1G0uO& RQ?MtO1G1PO& YQ?MtO1G1PO&#ZQ?MtO1G1PO&#bQ?MtO1G1PO&%cQ?MtO1G1PO&'dQ?MtO1G1PO&'kQ?MtO1G1PO&'rQ?MtO1G1PO&)sQ?MtO1G1PO&)zQ?MtO1G1PO&*RQ!0MxO<<JfO&+yQ?MtO1G1PO&,vQ?MvO1G1PO&-yQ?MvO'#JkO&0PQ?MtO1G1cO&0^Q?MtO1G0UO&0hQMjO,5?SOOQO-E<f-E<fO!)PQlO'#FqOOQO'#KY'#KYOOQO1G1u1G1uO&0rQ`O1G1tO&0wQ?MtO,5?ZOOOW7+'h7+'hOOOO1G/Z1G/ZO&1RQ!dO1G4wOOQ!0Lh7+(Q7+(QP!&oQMhO,5?]O!+xQMhO7+(cO&1YQ`O,5?[O9aQ`O,5?[OOQO-E<n-E<nO&1hQ`O1G6aO&1hQ`O1G6aO&1pQ`O1G6aO&1{QMjO7+'zO&2]Q!dO,5?^O&2gQ`O,5?^O!&oQMhO,5?^OOQO-E<p-E<pO&2lQ!dO1G6bO&2vQ`O1G6bO&3OQ`O1G2kO!&oQMhO1G2kOOQ!0Lb1G2i1G2iOOQ!0Lb1G2j1G2jO%3TQpO1G2iO!BqQpO1G2iOClQ`O1G2iOOQ!0Lb1G2q1G2qO&3TQpO1G2iO&3cQ`O1G2kO$){Q`O1G2jOClQ`O1G2jO$#jQlO1G2kO&3kQ`O1G2jO&4_QMjO,5?`OOQ!0Lh-E<s-E<sO&5QQMjO,5?bOOQ!0Lh-E<u-E<uO!+xQMhO7++[OOQ!0Lh1G/c1G/cO&5[Q`O1G/cOOQ!0Lh7+'u7+'uO&5aQMjO7+'|O&5qQMjO7++[O&5{QMjO7++[O&6YQ!0MxO<<KXOOQ!0Lf<<KX<<KXO&6|Q`O1G0zO!&oQMhO'#IyO&7RQ`O,5@wO&9TQ!fO<<LPO!&oQMhO1G2nO&9[Q!0LrO1G2nOOQ[<<G{<<G{O!BfQ!0LrO<<G{O&9mQ!0MxO<<I{OOQ!0Lf<<I{<<I{OOQO,5?k,5?kO&:aQ`O,5?kO&:fQ`O,5?kOOQO-E<}-E<}O&:tQ`O1G6jO&:tQ`O1G6jO9gQ`O1G6jO@oQ`O<<LlOOQ[<<Ll<<LlO&:|Q`O<<LlO9qQ!0LrO<<LlOOQ[<<LX<<LXO%9nQ!0MvO<<LXOOQ[<<LY<<LYO!DrQ`O<<LYO&;RQpO'#I{O&;^Q`O,5@{O!)PQlO,5@{OOQ[1G3W1G3WOOQO'#I}'#I}O9qQ!0LrO'#I}O&;fQpO,5=uOOQ[,5=u,5=uO&;mQpO'#EgO&;tQpO'#GeO&;yQ`O7+(zO&<OQ`O7+(zOOQ[7+(z7+(zO!&oQMhO7+(zO%[QlO7+(zO&<WQ`O7+(zOOQ[7+(|7+(|O9qQ!0LrO7+(|O$$VQ`O7+(|O9[Q`O7+(|O!BqQpO7+(|O&<cQ`O,5?jOOQO-E<|-E<|OOQO'#H^'#H^O&<nQ`O1G6hO9qQ!0LrO<<GqOOQ[<<Gq<<GqO@oQ`O<<GqO&<vQ`O7+,VO&<{Q`O7+,WO%[QlO7+,VO%[QlO7+,WOOQ[7+)U7+)UO&=QQ`O7+)UO&=VQlO7+)UO&=^Q`O7+)UOOQ[<<Lx<<LxOOQ[<<Lz<<LzOOQ[-E=P-E=POOQ[1G3y1G3yO&=cQ`O,5>`OOQ[,5>b,5>bO&=hQ`O1G4PO9aQ`O7+&fO!)PQlO7+&fOOQO7+%_7+%_O&=mQ?MtO1G6YO>}Q`O7+%_OOQ!0Lf<<Ia<<IaOOQ!0Lf<<Iz<<IzO>}Q`O<<IzOOQO<<Is<<IsO$@_Q!0MxO<<IsO%[QlO<<IsOOQO<<Id<<IdO!BfQ!0LrO<<IdO&=wQ!0LrO<<IsO&>SQ!0MxO<= ]O&>dQ`O<= [OOQO7+*^7+*^O9aQ`O7+*^OOQ[ANAkANAkO&>lQ!fOANAkO!&oQMhOANAkO#'oQ`OANAkO4UQ!fOANAkO&>sQ`OANAkO%[QlOANAkO&>{Q!0MzO7+'zO&A^Q!0MzO,5?`O&CiQ!0MzO,5?bO&EtQ!0MzO7+'|O&HVQ!fO1G4kO&HaQ?MtO7+&aO&JeQ?MvO,5=XO&LlQ?MvO,5=ZO&L|Q?MvO,5=XO&M^Q?MvO,5=ZO&MnQ?MvO,59uO' tQ?MvO,5<kO'#wQ?MvO,5<mO'&]Q?MvO,5<{O'(RQ?MtO7+'kO'(`Q?MtO7+'mO'(mQ`O,5<]OOQO7+'`7+'`OOQ!0Lh7+*c7+*cO'(rQMjO<<K}OOQO1G4v1G4vO'(yQ`O1G4vO')UQ`O1G4vO')dQ`O7++{O')dQ`O7++{O!&oQMhO1G4xO')lQ!dO1G4xO')vQ`O7++|O'*OQ`O7+(VO'*ZQ!dO7+(VOOQ!0Lb7+(T7+(TOOQ!0Lb7+(U7+(UO!BqQpO7+(TOClQ`O7+(TO'*eQ`O7+(VO!&oQMhO7+(VO$){Q`O7+(UO'*jQ`O7+(VOClQ`O7+(UO'*rQMjO<<NvOOQ!0Lh7+$}7+$}O!+xQMhO<<NvO'*|Q!dO,5?eOOQO-E<w-E<wO'+WQ!0MvO7+(YO!&oQMhO7+(YOOQ[AN=gAN=gO9gQ`O1G5VOOQO1G5V1G5VO'+hQ`O1G5VO'+mQ`O7+,UO'+mQ`O7+,UO9qQ!0LrOANBWO@oQ`OANBWOOQ[ANBWANBWOOQ[ANAsANAsOOQ[ANAtANAtO'+uQ`O,5?gOOQO-E<y-E<yO',QQ?MtO1G6gOOQO,5?i,5?iOOQO-E<{-E<{OOQ[1G3a1G3aO',[Q`O,5=POOQ[<<Lf<<LfO!&oQMhO<<LfO&;yQ`O<<LfO',aQ`O<<LfO%[QlO<<LfOOQ[<<Lh<<LhO9qQ!0LrO<<LhO$$VQ`O<<LhO9[Q`O<<LhO',iQpO1G5UO',tQ`O7+,SOOQ[AN=]AN=]O9qQ!0LrOAN=]OOQ[<= q<= qOOQ[<= r<= rO',|Q`O<= qO'-RQ`O<= rOOQ[<<Lp<<LpO'-WQ`O<<LpO'-]QlO<<LpOOQ[1G3z1G3zO>}Q`O7+)kO'-dQ`O<<JQO'-oQ?MtO<<JQOOQO<<Hy<<HyOOQ!0LfAN?fAN?fOOQOAN?_AN?_O$@_Q!0MxOAN?_OOQOAN?OAN?OO%[QlOAN?_OOQO<<Mx<<MxOOQ[G27VG27VO!&oQMhOG27VO#'oQ`OG27VO'-yQ!fOG27VO4UQ!fOG27VO'.QQ`OG27VO'.YQ?MtO<<JfO'.gQ?MvO1G2`O'0]Q?MvO,5?`O'2`Q?MvO,5?bO'4cQ?MvO1G2sO'6fQ?MvO1G2uO'8iQ?MtO<<KXO'8vQ?MtO<<I{OOQO1G1w1G1wO!+xQMhOANAiOOQO7+*b7+*bO'9TQ`O7+*bO'9`Q`O<= gO'9hQ!dO7+*dOOQ!0Lb<<Kq<<KqO$){Q`O<<KqOClQ`O<<KqO'9rQ`O<<KqO!&oQMhO<<KqOOQ!0Lb<<Ko<<KoO!BqQpO<<KoO'9}Q!dO<<KqOOQ!0Lb<<Kp<<KpO':XQ`O<<KqO!&oQMhO<<KqO$){Q`O<<KpO':^QMjOANDbO':hQ!0MvO<<KtOOQO7+*q7+*qO9gQ`O7+*qO':xQ`O<= pOOQ[G27rG27rO9qQ!0LrOG27rO!)PQlO1G5RO';QQ`O7+,RO';YQ`O1G2kO&;yQ`OANBQOOQ[ANBQANBQO!&oQMhOANBQO';_Q`OANBQOOQ[ANBSANBSO9qQ!0LrOANBSO$$VQ`OANBSOOQO'#H_'#H_OOQO7+*p7+*pOOQ[G22wG22wOOQ[ANE]ANE]OOQ[ANE^ANE^OOQ[ANB[ANB[O';gQ`OANB[OOQ[<<MV<<MVO!)PQlOAN?lOOQOG24yG24yO$@_Q!0MxOG24yO#'oQ`OLD,qOOQ[LD,qLD,qO!&oQMhOLD,qO';lQ!fOLD,qO';sQ?MvO7+'zO'=iQ?MvO,5?`O'?lQ?MvO,5?bO'AoQ?MvO7+'|O'CeQMjOG27TOOQO<<M|<<M|OOQ!0LbANA]ANA]O$){Q`OANA]OClQ`OANA]O'CuQ!dOANA]OOQ!0LbANAZANAZO'C|Q`OANA]O!&oQMhOANA]O'DXQ!dOANA]OOQ!0LbANA[ANA[OOQO<<N]<<N]OOQ[LD-^LD-^O'DcQ?MtO7+*mOOQO'#Gf'#GfOOQ[G27lG27lO&;yQ`OG27lO!&oQMhOG27lOOQ[G27nG27nO9qQ!0LrOG27nOOQ[G27vG27vO'DmQ?MtOG25WOOQOLD*eLD*eOOQ[!$(!]!$(!]O#'oQ`O!$(!]O!&oQMhO!$(!]O'DwQ!0MzOG27TOOQ!0LbG26wG26wO$){Q`OG26wO'GYQ`OG26wOClQ`OG26wO'GeQ!dOG26wO!&oQMhOG26wOOQ[LD-WLD-WO&;yQ`OLD-WOOQ[LD-YLD-YOOQ[!)9Ew!)9EwO#'oQ`O!)9EwOOQ!0LbLD,cLD,cO$){Q`OLD,cOClQ`OLD,cO'GlQ`OLD,cO'GwQ!dOLD,cOOQ[!$(!r!$(!rOOQ[!.K;c!.K;cO'HOQ?MvOG27TOOQ!0Lb!$( }!$( }O$){Q`O!$( }OClQ`O!$( }O'ItQ`O!$( }OOQ!0Lb!)9Ei!)9EiO$){Q`O!)9EiOClQ`O!)9EiOOQ!0Lb!.K;T!.K;TO$){Q`O!.K;TOOQ!0Lb!4/0o!4/0oO!)PQlO'#DzO1PQ`O'#EXO'JPQ!fO'#JqO'JWQ!L^O'#DvO'J_QlO'#EOO'JfQ!fO'#CiO'L|Q!fO'#CiO!)PQlO'#EQO'M^QlO,5;ZO!)PQlO,5;eO!)PQlO,5;eO!)PQlO,5;eO!)PQlO,5;eO!)PQlO,5;eO!)PQlO,5;eO!)PQlO,5;eO!)PQlO,5;eO!)PQlO,5;eO!)PQlO,5;eO!)PQlO'#IoO( aQ`O,5<iO!)PQlO,5;eO( iQMhO,5;eO(#SQMhO,5;eO!)PQlO,5;wO!&oQMhO'#GmO( iQMhO'#GmO!&oQMhO'#GoO( iQMhO'#GoO1SQ`O'#DZO1SQ`O'#DZO!&oQMhO'#GPO( iQMhO'#GPO!&oQMhO'#GRO( iQMhO'#GRO!&oQMhO'#GaO( iQMhO'#GaO!)PQlO,5:jO(#ZQpO'#D_O(#eQpO'#JuO!)PQlO,5@nO'M^QlO1G0uO(#oQ?MtO'#CiO!)PQlO1G2PO!&oQMhO'#ItO( iQMhO'#ItO!&oQMhO'#IvO( iQMhO'#IvO(#yQ!dO'#CrO!&oQMhO,5<tO( iQMhO,5<tO'M^QlO1G2RO!)PQlO7+&zO!&oQMhO1G2`O( iQMhO1G2`O!&oQMhO'#ItO( iQMhO'#ItO!&oQMhO'#IvO( iQMhO'#IvO!&oQMhO1G2bO( iQMhO1G2bO'M^QlO7+'mO'M^QlO7+&aO!&oQMhOANAiO( iQMhOANAiO($^Q`O'#EoO($cQ`O'#EoO($kQ`O'#F]O($pQ`O'#EyO($uQ`O'#KSO(%QQ`O'#KQO(%]Q`O,5;ZO(%bQMjO,5<eO(%iQ`O'#GYO(%nQ`O'#GYO(%sQ`O,5<gO(%{Q`O,5;ZO(&TQ?MtO1G1`O(&[Q`O,5<tO(&aQ`O,5<tO(&fQ`O,5<vO(&kQ`O,5<vO(&pQ`O1G2RO(&uQ`O1G0uO(&zQMjO<<K}O('RQMjO<<K}O7eQMhO'#F|O9[Q`O'#F{OAjQ`O'#EnO!)PQlO,5;tO!3dQ`O'#GYO!3dQ`O'#GYO!3dQ`O'#G[O!3dQ`O'#G[O!+xQMhO7+(cO!+xQMhO7+(cO%-mQ!dO1G2wO%-mQ!dO1G2wO!&oQMhO,5=]O!&oQMhO,5=]",stateData:"((X~O'{OS'|OSTOS'}RQ~OPYOQYOSfOY!VOaqOdzOeyOl!POpkOrYOskOtkOzkO|YO!OYO!SWO!WkO!XkO!_XO!iuO!lZO!oYO!pYO!qYO!svO!uwO!xxO!|]O$W|O$niO%h}O%j!QO%l!OO%m!OO%n!OO%q!RO%s!SO%v!TO%w!TO%y!UO&V!WO&]!XO&_!YO&a!ZO&c![O&f!]O&l!^O&r!_O&t!`O&v!aO&x!bO&z!cO(SSO(UTO(XUO(`VO(n[O~OWtO~P`OPYOQYOSfOd!jOe!iOpkOrYOskOtkOzkO|YO!OYO!SWO!WkO!XkO!_!eO!iuO!lZO!oYO!pYO!qYO!svO!u!gO!x!hO$W!kO$niO(S!dO(UTO(XUO(`VO(n[O~Oa!wOs!nO!S!oO!b!yO!c!vO!d!vO!|;wO#T!pO#U!pO#V!xO#W!pO#X!pO#[!zO#]!zO(T!lO(UTO(XUO(d!mO(n!sO~O'}!{O~OP]XR]X[]Xa]Xj]Xr]X!Q]X!S]X!]]X!l]X!p]X#R]X#S]X#`]X#kfX#n]X#o]X#p]X#q]X#r]X#s]X#t]X#u]X#v]X#x]X#z]X#{]X$Q]X'y]X(`]X(q]X(x]X(y]X~O!g%RX~P(qO_!}O(U#PO(V!}O(W#PO~O_#QO(W#PO(X#PO(Y#QO~Ox#SO!U#TO(a#TO(b#VO~OPYOQYOSfOd!jOe!iOpkOrYOskOtkOzkO|YO!OYO!SWO!WkO!XkO!_!eO!iuO!lZO!oYO!pYO!qYO!svO!u!gO!x!hO$W!kO$niO(S;{O(UTO(XUO(`VO(n[O~O![#ZO!]#WO!Y(gP!Y(uP~P+}O!^#cO~P`OPYOQYOSfOd!jOe!iOrYOskOtkOzkO|YO!OYO!SWO!WkO!XkO!_!eO!iuO!lZO!oYO!pYO!qYO!svO!u!gO!x!hO$W!kO$niO(UTO(XUO(`VO(n[O~Op#mO![#iO!|]O#i#lO#j#iO(S;|O!k(rP~P.iO!l#oO(S#nO~O!x#sO!|]O%h#tO~O#k#uO~O!g#vO#k#uO~OP$[OR#zO[$cOj$ROr$aO!Q#yO!S#{O!]$_O!l#xO!p$[O#R$RO#n$OO#o$PO#p$PO#q$PO#r$QO#s$RO#t$RO#u$bO#v$SO#x$UO#z$WO#{$XO(`VO(q$YO(x#|O(y#}O~Oa(eX'y(eX'v(eX!k(eX!Y(eX!_(eX%i(eX!g(eX~P1qO#S$dO#`$eO$Q$eOP(fXR(fX[(fXj(fXr(fX!Q(fX!S(fX!](fX!l(fX!p(fX#R(fX#n(fX#o(fX#p(fX#q(fX#r(fX#s(fX#t(fX#u(fX#v(fX#x(fX#z(fX#{(fX(`(fX(q(fX(x(fX(y(fX!_(fX%i(fX~Oa(fX'y(fX'v(fX!Y(fX!k(fXv(fX!g(fX~P4UO#`$eO~O$]$hO$_$gO$f$mO~OSfO!_$nO$i$oO$k$qO~Oh%VOj%cOk%cOl%cOp%WOr%XOs$tOt$tOz%YO|%ZO!O%[O!S${O!_$|O!i%aO!l$xO#j%bO$W%_O$t%]O$v%^O$y%`O(S$sO(UTO(XUO(`$uO(x$}O(y%POg(]P~O!l%dO~O!S%gO!_%hO(S%fO~O!g%lO~Oa%mO'y%mO~O!Q%qO~P%[O(T!lO~P%[O%n%uO~P%[Oh%VO!l%dO(S%fO(T!lO~Oe%|O!l%dO(S%fO~Oj$RO~O!Q&RO!_&OO!l&QO%j&UO(S%fO(T!lO(UTO(XUO`)VP~O!x#sO~O%s&WO!S)RX!_)RX(S)RX~O(S&XO~Ol!PO!u&^O%j!QO%l!OO%m!OO%n!OO%q!RO%s!SO%v!TO%w!TO~Od&cOe&bO!x&`O%h&aO%{&_O~P<VOd&fOeyOl!PO!_&eO!u&^O!xxO!|]O%h}O%l!OO%m!OO%n!OO%q!RO%s!SO%v!TO%w!TO%y!UO~Ob&iO#`&lO%j&gO(T!lO~P=[O!l&mO!u&qO~O!l#oO~O!_XO~Oa%mO'w&yO'y%mO~Oa%mO'w&|O'y%mO~Oa%mO'w'OO'y%mO~O'v]X!Y]Xv]X!k]X&Z]X!_]X%i]X!g]X~P(qO!b']O!c'UO!d'UO(T!lO(UTO(XUO~Os'SO!S'RO!['VO(d'QO!^(hP!^(wP~P@cOn'`O!_'^O(S%fO~Oe'eO!l%dO(S%fO~O!Q&RO!l&QO~Os!nO!S!oO!|;wO#T!pO#U!pO#W!pO#X!pO(T!lO(UTO(XUO(d!mO(n!sO~O!b'kO!c'jO!d'jO#V!pO#['lO#]'lO~PA}Oa%mOh%VO!g#vO!l%dO'y%mO(q'nO~O!p'rO#`'pO~PC]Os!nO!S!oO(UTO(XUO(d!mO(n!sO~O!_XOs(lX!S(lX!b(lX!c(lX!d(lX!|(lX#T(lX#U(lX#V(lX#W(lX#X(lX#[(lX#](lX(T(lX(U(lX(X(lX(d(lX(n(lX~O!c'jO!d'jO(T!lO~PC{O(O'vO(P'vO(Q'xO~O_!}O(U'zO(V!}O(W'zO~O_#QO(W'zO(X'zO(Y#QO~Ov'|O~P%[Ox#SO!U#TO(a#TO(b(PO~O![(RO!Y'VX!Y']X!]'VX!]']X~P+}O!](TO!Y(gX~OP$[OR#zO[$cOj$ROr$aO!Q#yO!S#{O!](TO!l#xO!p$[O#R$RO#n$OO#o$PO#p$PO#q$PO#r$QO#s$RO#t$RO#u$bO#v$SO#x$UO#z$WO#{$XO(`VO(q$YO(x#|O(y#}O~O!Y(gX~PGvO!Y(YO~O!Y(tX!](tX!g(tX!k(tX(q(tX~O#`(tX#k#dX!^(tX~PIyO#`(ZO!Y(vX!](vX~O!]([O!Y(uX~O!Y(_O~O#`$eO~PIyO!^(`O~P`OR#zO!Q#yO!S#{O!l#xO(`VOP!na[!naj!nar!na!]!na!p!na#R!na#n!na#o!na#p!na#q!na#r!na#s!na#t!na#u!na#v!na#x!na#z!na#{!na(q!na(x!na(y!na~Oa!na'y!na'v!na!Y!na!k!nav!na!_!na%i!na!g!na~PKaO!k(aO~O!g#vO#`(bO(q'nO!](sXa(sX'y(sX~O!k(sX~PM|O!S%gO!_%hO!|]O#i(gO#j(fO(S%fO~O!](hO!k(rX~O!k(jO~O!S%gO!_%hO#j(fO(S%fO~OP(fXR(fX[(fXj(fXr(fX!Q(fX!S(fX!](fX!l(fX!p(fX#R(fX#n(fX#o(fX#p(fX#q(fX#r(fX#s(fX#t(fX#u(fX#v(fX#x(fX#z(fX#{(fX(`(fX(q(fX(x(fX(y(fX~O!g#vO!k(fX~P! jOR(lO!Q(kO!l#xO#S$dO!|!{a!S!{a~O!x!{a%h!{a!_!{a#i!{a#j!{a(S!{a~P!#kO!x(pO~OPYOQYOSfOd!jOe!iOpkOrYOskOtkOzkO|YO!OYO!SWO!WkO!XkO!_XO!iuO!lZO!oYO!pYO!qYO!svO!u!gO!x!hO$W!kO$niO(S!dO(UTO(XUO(`VO(n[O~Oh%VOp%WOr%XOs$tOt$tOz%YO|%ZO!O<eO!S${O!_$|O!i=vO!l$xO#j<kO$W%_O$t<gO$v<iO$y%`O(S(tO(UTO(XUO(`$uO(x$}O(y%PO~O#k(vO~O![(xO!k(jP~P%[O(d(zO(n[O~O!S(|O!l#xO(d(zO(n[O~OP;vOQ;vOSfOd=rOe!iOpkOr;vOskOtkOzkO|;vO!O;vO!SWO!WkO!XkO!_!eO!i;yO!lZO!o;vO!p;vO!q;vO!s;zO!u;}O!x!hO$W!kO$n=pO(S)ZO(UTO(XUO(`VO(n[O~O!]$_Oa$qa'y$qa'v$qa!k$qa!Y$qa!_$qa%i$qa!g$qa~Ol)bO~P!&oOh%VOp%WOr%XOs$tOt$tOz%YO|%ZO!O%[O!S${O!_$|O!i%aO!l$xO#j%bO$W%_O$t%]O$v%^O$y%`O(S(tO(UTO(XUO(`$uO(x$}O(y%PO~Og(oP~P!+xO!Q)gO!g)fO!_$^X$Z$^X$]$^X$_$^X$f$^X~O!g)fO!_(zX$Z(zX$](zX$_(zX$f(zX~O!Q)gO~P!.RO!Q)gO!_(zX$Z(zX$](zX$_(zX$f(zX~O!_)iO$Z)mO$])hO$_)hO$f)nO~O![)qO~P!)PO$]$hO$_$gO$f)uO~On$zX!Q$zX#S$zX'x$zX(x$zX(y$zX~OgmXg$zXnmX!]mX#`mX~P!/wOx)wO(a)xO(b)zO~On*TO!Q)|O'x)}O(x$}O(y%PO~Og){O~P!0{Og*UO~Oh%VOp%WOr%XOs$tOt$tOz%YO|%ZO!O<eO!S*WO!_*XO!i=vO!l$xO#j<kO$W%_O$t<gO$v<iO$y%`O(UTO(XUO(`$uO(x$}O(y%PO~O![*[O(S*VO!k(}P~P!1jO#k*^O~O!l*_O~Oh%VOp%WOr%XOs$tOt$tOz%YO|%ZO!O<eO!S${O!_$|O!i=vO!l$xO#j<kO$W%_O$t<gO$v<iO$y%`O(S*aO(UTO(XUO(`$uO(x$}O(y%PO~O![*dO!Y)OP~P!3iOr*pOs!nO!S*fO!b*nO!c*hO!d*hO!l*_O#[*oO%`*jO(T!lO(UTO(XUO(d!mO~O!^*mO~P!5^O#S$dOn(_X!Q(_X'x(_X(x(_X(y(_X!](_X#`(_X~Og(_X$O(_X~P!6`On*uO#`*tOg(^X!](^X~O!]*vOg(]X~Oj%cOk%cOl%cO(S&XOg(]P~Os*yO~O!l+OO~O(S(tO~Op+TO!S%gO![#iO!_%hO!|]O#i#lO#j#iO(S%fO!k(rP~O!g#vO#k+UO~O!S%gO![+WO!]([O!_%hO(S%fO!Y(uP~Os'YO!S+YO![+XO(UTO(XUO(d(zO~O!^(wP~P!9iO!]+ZOa)SX'y)SX~OP$[OR#zO[$cOj$ROr$aO!Q#yO!S#{O!l#xO!p$[O#R$RO#n$OO#o$PO#p$PO#q$PO#r$QO#s$RO#t$RO#u$bO#v$SO#x$UO#z$WO#{$XO(`VO(q$YO(x#|O(y#}O~Oa!ja!]!ja'y!ja'v!ja!Y!ja!k!jav!ja!_!ja%i!ja!g!ja~P!:aOR#zO!Q#yO!S#{O!l#xO(`VOP!ra[!raj!rar!ra!]!ra!p!ra#R!ra#n!ra#o!ra#p!ra#q!ra#r!ra#s!ra#t!ra#u!ra#v!ra#x!ra#z!ra#{!ra(q!ra(x!ra(y!ra~Oa!ra'y!ra'v!ra!Y!ra!k!rav!ra!_!ra%i!ra!g!ra~P!<wOR#zO!Q#yO!S#{O!l#xO(`VOP!ta[!taj!tar!ta!]!ta!p!ta#R!ta#n!ta#o!ta#p!ta#q!ta#r!ta#s!ta#t!ta#u!ta#v!ta#x!ta#z!ta#{!ta(q!ta(x!ta(y!ta~Oa!ta'y!ta'v!ta!Y!ta!k!tav!ta!_!ta%i!ta!g!ta~P!?_Oh%VOn+dO!_'^O%i+cO~O!g+fOa([X!_([X'y([X!]([X~Oa%mO!_XO'y%mO~Oh%VO!l%dO~Oh%VO!l%dO(S%fO~O!g#vO#k(vO~Ob+qO%j+rO(S+nO(UTO(XUO!^)WP~O!]+sO`)VX~O[+wO~O`+xO~O!_&OO(S%fO(T!lO`)VP~Oh%VO#`+}O~Oh%VOn,QO!_$|O~O!_,SO~O!Q,UO!_XO~O%n%uO~O!x,ZO~Oe,`O~Ob,aO(S#nO(UTO(XUO!^)UP~Oe%|O~O%j!QO(S&XO~P=[O[,fO`,eO~OPYOQYOSfOdzOeyOpkOrYOskOtkOzkO|YO!OYO!SWO!WkO!XkO!iuO!lZO!oYO!pYO!qYO!svO!xxO!|]O$niO%h}O(UTO(XUO(`VO(n[O~O!_!eO!u!gO$W!kO(S!dO~P!F_O`,eOa%mO'y%mO~OPYOQYOSfOd!jOe!iOpkOrYOskOtkOzkO|YO!OYO!SWO!WkO!XkO!_!eO!iuO!lZO!oYO!pYO!qYO!svO!x!hO$W!kO$niO(S!dO(UTO(XUO(`VO(n[O~Oa,kOl!OO!uwO%l!OO%m!OO%n!OO~P!HwO!l&mO~O&],qO~O!_,sO~O&n,uO&p,vOP&kaQ&kaS&kaY&kaa&kad&kae&kal&kap&kar&kas&kat&kaz&ka|&ka!O&ka!S&ka!W&ka!X&ka!_&ka!i&ka!l&ka!o&ka!p&ka!q&ka!s&ka!u&ka!x&ka!|&ka$W&ka$n&ka%h&ka%j&ka%l&ka%m&ka%n&ka%q&ka%s&ka%v&ka%w&ka%y&ka&V&ka&]&ka&_&ka&a&ka&c&ka&f&ka&l&ka&r&ka&t&ka&v&ka&x&ka&z&ka'v&ka(S&ka(U&ka(X&ka(`&ka(n&ka!^&ka&d&kab&ka&i&ka~O(S,{O~Oh!eX!]!RX!^!RX!g!RX!g!eX!l!eX#`!RX~O!]!eX!^!eX~P# }O!g-QO#`-POh(iX!]#hX!^#hX!g(iX!l(iX~O!](iX!^(iX~P#!pOh%VO!g-SO!l%dO!]!aX!^!aX~Os!nO!S!oO(UTO(XUO(d!mO~OP;vOQ;vOSfOd=rOe!iOpkOr;vOskOtkOzkO|;vO!O;vO!SWO!WkO!XkO!_!eO!i;yO!lZO!o;vO!p;vO!q;vO!s;zO!u;}O!x!hO$W!kO$n=pO(UTO(XUO(`VO(n[O~O(S<rO~P#$VO!]-WO!^(hX~O!^-YO~O!g-QO#`-PO!]#hX!^#hX~O!]-ZO!^(wX~O!^-]O~O!c-^O!d-^O(T!lO~P##tO!^-aO~P'_On-dO!_'^O~O!Y-iO~Os!{a!b!{a!c!{a!d!{a#T!{a#U!{a#V!{a#W!{a#X!{a#[!{a#]!{a(T!{a(U!{a(X!{a(d!{a(n!{a~P!#kO!p-nO#`-lO~PC]O!c-pO!d-pO(T!lO~PC{Oa%mO#`-lO'y%mO~Oa%mO!g#vO#`-lO'y%mO~Oa%mO!g#vO!p-nO#`-lO'y%mO(q'nO~O(O'vO(P'vO(Q-uO~Ov-vO~O!Y'Va!]'Va~P!:aO![-zO!Y'VX!]'VX~P%[O!](TO!Y(ga~O!Y(ga~PGvO!]([O!Y(ua~O!S%gO![.OO!_%hO(S%fO!Y']X!]']X~O#`.QO!](sa!k(saa(sa'y(sa~O!g#vO~P#,]O!](hO!k(ra~O!S%gO!_%hO#j.UO(S%fO~Op.ZO!S%gO![.WO!_%hO!|]O#i.YO#j.WO(S%fO!]'`X!k'`X~OR._O!l#xO~Oh%VOn.bO!_'^O%i.aO~Oa#ci!]#ci'y#ci'v#ci!Y#ci!k#civ#ci!_#ci%i#ci!g#ci~P!:aOn=|O!Q)|O'x)}O(x$}O(y%PO~O#k#_aa#_a#`#_a'y#_a!]#_a!k#_a!_#_a!Y#_a~P#/XO#k(_XP(_XR(_X[(_Xa(_Xj(_Xr(_X!S(_X!l(_X!p(_X#R(_X#n(_X#o(_X#p(_X#q(_X#r(_X#s(_X#t(_X#u(_X#v(_X#x(_X#z(_X#{(_X'y(_X(`(_X(q(_X!k(_X!Y(_X'v(_Xv(_X!_(_X%i(_X!g(_X~P!6`O!].oO!k(jX~P!:aO!k.rO~O!Y.tO~OP$[OR#zO!Q#yO!S#{O!l#xO!p$[O(`VO[#mia#mij#mir#mi!]#mi#R#mi#o#mi#p#mi#q#mi#r#mi#s#mi#t#mi#u#mi#v#mi#x#mi#z#mi#{#mi'y#mi(q#mi(x#mi(y#mi'v#mi!Y#mi!k#miv#mi!_#mi%i#mi!g#mi~O#n#mi~P#2wO#n$OO~P#2wOP$[OR#zOr$aO!Q#yO!S#{O!l#xO!p$[O#n$OO#o$PO#p$PO#q$PO(`VO[#mia#mij#mi!]#mi#R#mi#s#mi#t#mi#u#mi#v#mi#x#mi#z#mi#{#mi'y#mi(q#mi(x#mi(y#mi'v#mi!Y#mi!k#miv#mi!_#mi%i#mi!g#mi~O#r#mi~P#5fO#r$QO~P#5fOP$[OR#zO[$cOj$ROr$aO!Q#yO!S#{O!l#xO!p$[O#R$RO#n$OO#o$PO#p$PO#q$PO#r$QO#s$RO#t$RO#u$bO(`VOa#mi!]#mi#x#mi#z#mi#{#mi'y#mi(q#mi(x#mi(y#mi'v#mi!Y#mi!k#miv#mi!_#mi%i#mi!g#mi~O#v#mi~P#8TOP$[OR#zO[$cOj$ROr$aO!Q#yO!S#{O!l#xO!p$[O#R$RO#n$OO#o$PO#p$PO#q$PO#r$QO#s$RO#t$RO#u$bO#v$SO(`VO(y#}Oa#mi!]#mi#z#mi#{#mi'y#mi(q#mi(x#mi'v#mi!Y#mi!k#miv#mi!_#mi%i#mi!g#mi~O#x$UO~P#:kO#x#mi~P#:kO#v$SO~P#8TOP$[OR#zO[$cOj$ROr$aO!Q#yO!S#{O!l#xO!p$[O#R$RO#n$OO#o$PO#p$PO#q$PO#r$QO#s$RO#t$RO#u$bO#v$SO#x$UO(`VO(x#|O(y#}Oa#mi!]#mi#{#mi'y#mi(q#mi'v#mi!Y#mi!k#miv#mi!_#mi%i#mi!g#mi~O#z#mi~P#=aO#z$WO~P#=aOP]XR]X[]Xj]Xr]X!Q]X!S]X!l]X!p]X#R]X#S]X#`]X#kfX#n]X#o]X#p]X#q]X#r]X#s]X#t]X#u]X#v]X#x]X#z]X#{]X$Q]X(`]X(q]X(x]X(y]X!]]X!^]X~O$O]X~P#@OOP$[OR#zO[<_Oj<SOr<]O!Q#yO!S#{O!l#xO!p$[O#R<SO#n<PO#o<QO#p<QO#q<QO#r<RO#s<SO#t<SO#u<^O#v<TO#x<VO#z<XO#{<YO(`VO(q$YO(x#|O(y#}O~O$O.vO~P#B]O#S$dO#`<`O$Q<`O$O(fX!^(fX~P! jOa'ca!]'ca'y'ca'v'ca!k'ca!Y'cav'ca!_'ca%i'ca!g'ca~P!:aO[#mia#mij#mir#mi!]#mi#R#mi#r#mi#s#mi#t#mi#u#mi#v#mi#x#mi#z#mi#{#mi'y#mi(q#mi'v#mi!Y#mi!k#miv#mi!_#mi%i#mi!g#mi~OP$[OR#zO!Q#yO!S#{O!l#xO!p$[O#n$OO#o$PO#p$PO#q$PO(`VO(x#mi(y#mi~P#E_On=|O!Q)|O'x)}O(x$}O(y%POP#miR#mi!S#mi!l#mi!p#mi#n#mi#o#mi#p#mi#q#mi(`#mi~P#E_O!].zOg(oX~P!0{Og.|O~Oa$Pi!]$Pi'y$Pi'v$Pi!Y$Pi!k$Piv$Pi!_$Pi%i$Pi!g$Pi~P!:aO$].}O$_.}O~O$]/OO$_/OO~O!g)fO#`/PO!_$cX$Z$cX$]$cX$_$cX$f$cX~O![/QO~O!_)iO$Z/SO$])hO$_)hO$f/TO~O!]<ZO!^(eX~P#B]O!^/UO~O!g)fO$f(zX~O$f/WO~Ov/XO~P!&oOx)wO(a)xO(b/[O~O!S/_O~O(x$}On%aa!Q%aa'x%aa(y%aa!]%aa#`%aa~Og%aa$O%aa~P#LaO(y%POn%ca!Q%ca'x%ca(x%ca!]%ca#`%ca~Og%ca$O%ca~P#MSO!]fX!gfX!kfX!k$zX(qfX~P!/wO![/hO!]([O(S/gO!Y(uP!Y)OP~P!1jOr*pO!b*nO!c*hO!d*hO!l*_O#[*oO%`*jO(T!lO(UTO(XUO~Os<oO!S/iO![+XO!^*mO(d<nO!^(wP~P#NmO!k/jO~P#/XO!]/kO!g#vO(q'nO!k(}X~O!k/pO~O!S%gO![*[O!_%hO(S%fO!k(}P~O#k/rO~O!Y$zX!]$zX!g%RX~P!/wO!]/sO!Y)OX~P#/XO!g/uO~O!Y/wO~OpkO(S/xO~P.iOh%VOr/}O!g#vO!l%dO(q'nO~O!g+fO~Oa%mO!]0RO'y%mO~O!^0TO~P!5^O!c0UO!d0UO(T!lO~P##tOs!nO!S0VO(UTO(XUO(d!mO~O#[0XO~Og%aa!]%aa#`%aa$O%aa~P!0{Og%ca!]%ca#`%ca$O%ca~P!0{Oj%cOk%cOl%cO(S&XOg'lX!]'lX~O!]*vOg(]a~Og0bO~OR0cO!Q0cO!S0dO#S$dOn}a'x}a(x}a(y}a!]}a#`}a~Og}a$O}a~P$&vO!Q)|O'x)}On$sa(x$sa(y$sa!]$sa#`$sa~Og$sa$O$sa~P$'rO!Q)|O'x)}On$ua(x$ua(y$ua!]$ua#`$ua~Og$ua$O$ua~P$(eO#k0gO~Og%Ta!]%Ta#`%Ta$O%Ta~P!0{On0iO#`0hOg(^a!](^a~O!g#vO~O#k0lO~O!]+ZOa)Sa'y)Sa~OR#zO!Q#yO!S#{O!l#xO(`VOP!ri[!rij!rir!ri!]!ri!p!ri#R!ri#n!ri#o!ri#p!ri#q!ri#r!ri#s!ri#t!ri#u!ri#v!ri#x!ri#z!ri#{!ri(q!ri(x!ri(y!ri~Oa!ri'y!ri'v!ri!Y!ri!k!riv!ri!_!ri%i!ri!g!ri~P$*bOh%VOr%XOs$tOt$tOz%YO|%ZO!O<eO!S${O!_$|O!i=vO!l$xO#j<kO$W%_O$t<gO$v<iO$y%`O(UTO(XUO(`$uO(x$}O(y%PO~Op0uO%]0vO(S0tO~P$,xO!g+fOa([a!_([a'y([a!]([a~O#k0|O~O[]X!]fX!^fX~O!]0}O!^)WX~O!^1PO~O[1QO~Ob1SO(S+nO(UTO(XUO~O!_&OO(S%fO`'tX!]'tX~O!]+sO`)Va~O!k1VO~P!:aO[1YO~O`1ZO~O#`1^O~On1aO!_$|O~O(d(zO!^)TP~Oh%VOn1jO!_1gO%i1iO~O[1tO!]1rO!^)UX~O!^1uO~O`1wOa%mO'y%mO~O(S#nO(UTO(XUO~O#S$dO#`$eO$Q$eOP(fXR(fX[(fXr(fX!Q(fX!S(fX!](fX!l(fX!p(fX#R(fX#n(fX#o(fX#p(fX#q(fX#r(fX#s(fX#t(fX#u(fX#v(fX#x(fX#z(fX#{(fX(`(fX(q(fX(x(fX(y(fX~Oj1zO&Z1{Oa(fX~P$2cOj1zO#`$eO&Z1{O~Oa1}O~P%[Oa2PO~O&d2SOP&biQ&biS&biY&bia&bid&bie&bil&bip&bir&bis&bit&biz&bi|&bi!O&bi!S&bi!W&bi!X&bi!_&bi!i&bi!l&bi!o&bi!p&bi!q&bi!s&bi!u&bi!x&bi!|&bi$W&bi$n&bi%h&bi%j&bi%l&bi%m&bi%n&bi%q&bi%s&bi%v&bi%w&bi%y&bi&V&bi&]&bi&_&bi&a&bi&c&bi&f&bi&l&bi&r&bi&t&bi&v&bi&x&bi&z&bi'v&bi(S&bi(U&bi(X&bi(`&bi(n&bi!^&bib&bi&i&bi~Ob2YO!^2WO&i2XO~P`O!_XO!l2[O~O&p,vOP&kiQ&kiS&kiY&kia&kid&kie&kil&kip&kir&kis&kit&kiz&ki|&ki!O&ki!S&ki!W&ki!X&ki!_&ki!i&ki!l&ki!o&ki!p&ki!q&ki!s&ki!u&ki!x&ki!|&ki$W&ki$n&ki%h&ki%j&ki%l&ki%m&ki%n&ki%q&ki%s&ki%v&ki%w&ki%y&ki&V&ki&]&ki&_&ki&a&ki&c&ki&f&ki&l&ki&r&ki&t&ki&v&ki&x&ki&z&ki'v&ki(S&ki(U&ki(X&ki(`&ki(n&ki!^&ki&d&kib&ki&i&ki~O!Y2bO~O!]!aa!^!aa~P#B]Os!nO!S!oO![2hO(d!mO!]'WX!^'WX~P@cO!]-WO!^(ha~O!]'^X!^'^X~P!9iO!]-ZO!^(wa~O!^2oO~P'_Oa%mO#`2xO'y%mO~Oa%mO!g#vO#`2xO'y%mO~Oa%mO!g#vO!p2|O#`2xO'y%mO(q'nO~Oa%mO'y%mO~P!:aO!]$_Ov$qa~O!Y'Vi!]'Vi~P!:aO!](TO!Y(gi~O!]([O!Y(ui~O!Y(vi!](vi~P!:aO!](si!k(sia(si'y(si~P!:aO#`3OO!](si!k(sia(si'y(si~O!](hO!k(ri~O!S%gO!_%hO!|]O#i3TO#j3SO(S%fO~O!S%gO!_%hO#j3SO(S%fO~On3[O!_'^O%i3ZO~Oh%VOn3[O!_'^O%i3ZO~O#k%aaP%aaR%aa[%aaa%aaj%aar%aa!S%aa!l%aa!p%aa#R%aa#n%aa#o%aa#p%aa#q%aa#r%aa#s%aa#t%aa#u%aa#v%aa#x%aa#z%aa#{%aa'y%aa(`%aa(q%aa!k%aa!Y%aa'v%aav%aa!_%aa%i%aa!g%aa~P#LaO#k%caP%caR%ca[%caa%caj%car%ca!S%ca!l%ca!p%ca#R%ca#n%ca#o%ca#p%ca#q%ca#r%ca#s%ca#t%ca#u%ca#v%ca#x%ca#z%ca#{%ca'y%ca(`%ca(q%ca!k%ca!Y%ca'v%cav%ca!_%ca%i%ca!g%ca~P#MSO#k%aaP%aaR%aa[%aaa%aaj%aar%aa!S%aa!]%aa!l%aa!p%aa#R%aa#n%aa#o%aa#p%aa#q%aa#r%aa#s%aa#t%aa#u%aa#v%aa#x%aa#z%aa#{%aa'y%aa(`%aa(q%aa!k%aa!Y%aa'v%aa#`%aav%aa!_%aa%i%aa!g%aa~P#/XO#k%caP%caR%ca[%caa%caj%car%ca!S%ca!]%ca!l%ca!p%ca#R%ca#n%ca#o%ca#p%ca#q%ca#r%ca#s%ca#t%ca#u%ca#v%ca#x%ca#z%ca#{%ca'y%ca(`%ca(q%ca!k%ca!Y%ca'v%ca#`%cav%ca!_%ca%i%ca!g%ca~P#/XO#k}aP}a[}aa}aj}ar}a!l}a!p}a#R}a#n}a#o}a#p}a#q}a#r}a#s}a#t}a#u}a#v}a#x}a#z}a#{}a'y}a(`}a(q}a!k}a!Y}a'v}av}a!_}a%i}a!g}a~P$&vO#k$saP$saR$sa[$saa$saj$sar$sa!S$sa!l$sa!p$sa#R$sa#n$sa#o$sa#p$sa#q$sa#r$sa#s$sa#t$sa#u$sa#v$sa#x$sa#z$sa#{$sa'y$sa(`$sa(q$sa!k$sa!Y$sa'v$sav$sa!_$sa%i$sa!g$sa~P$'rO#k$uaP$uaR$ua[$uaa$uaj$uar$ua!S$ua!l$ua!p$ua#R$ua#n$ua#o$ua#p$ua#q$ua#r$ua#s$ua#t$ua#u$ua#v$ua#x$ua#z$ua#{$ua'y$ua(`$ua(q$ua!k$ua!Y$ua'v$uav$ua!_$ua%i$ua!g$ua~P$(eO#k%TaP%TaR%Ta[%Taa%Taj%Tar%Ta!S%Ta!]%Ta!l%Ta!p%Ta#R%Ta#n%Ta#o%Ta#p%Ta#q%Ta#r%Ta#s%Ta#t%Ta#u%Ta#v%Ta#x%Ta#z%Ta#{%Ta'y%Ta(`%Ta(q%Ta!k%Ta!Y%Ta'v%Ta#`%Tav%Ta!_%Ta%i%Ta!g%Ta~P#/XOa#cq!]#cq'y#cq'v#cq!Y#cq!k#cqv#cq!_#cq%i#cq!g#cq~P!:aO![3dO!]'XX!k'XX~P%[O!].oO!k(ja~O!].oO!k(ja~P!:aO!Y3gO~O$O!na!^!na~PKaO$O!ja!]!ja!^!ja~P#B]O$O!ra!^!ra~P!<wO$O!ta!^!ta~P!?_Og'[X!]'[X~P!+xO!].zOg(oa~OSfO!_3{O$d3|O~O!^4QO~Ov4RO~P#/XOa$mq!]$mq'y$mq'v$mq!Y$mq!k$mqv$mq!_$mq%i$mq!g$mq~P!:aO!Y4TO~P!&oO!S4UO~O!Q)|O'x)}O(y%POn'ha(x'ha!]'ha#`'ha~Og'ha$O'ha~P%,XO!Q)|O'x)}On'ja(x'ja(y'ja!]'ja#`'ja~Og'ja$O'ja~P%,zO(q$YO~P#/XO!YfX!Y$zX!]fX!]$zX!g%RX#`fX~P!/wO(S<xO~P!1jO!S%gO![4XO!_%hO(S%fO!]'dX!k'dX~O!]/kO!k(}a~O!]/kO!g#vO!k(}a~O!]/kO!g#vO(q'nO!k(}a~Og$|i!]$|i#`$|i$O$|i~P!0{O![4aO!Y'fX!]'fX~P!3iO!]/sO!Y)Oa~O!]/sO!Y)Oa~P#/XOP]XR]X[]Xj]Xr]X!Q]X!S]X!Y]X!]]X!l]X!p]X#R]X#S]X#`]X#kfX#n]X#o]X#p]X#q]X#r]X#s]X#t]X#u]X#v]X#x]X#z]X#{]X$Q]X(`]X(q]X(x]X(y]X~Oj%YX!g%YX~P%0kOj4fO!g#vO~Oh%VO!g#vO!l%dO~Oh%VOr4kO!l%dO(q'nO~Or4pO!g#vO(q'nO~Os!nO!S4qO(UTO(XUO(d!mO~O(x$}On%ai!Q%ai'x%ai(y%ai!]%ai#`%ai~Og%ai$O%ai~P%4[O(y%POn%ci!Q%ci'x%ci(x%ci!]%ci#`%ci~Og%ci$O%ci~P%4}Og(^i!](^i~P!0{O#`4wOg(^i!](^i~P!0{O!k4zO~Oa$oq!]$oq'y$oq'v$oq!Y$oq!k$oqv$oq!_$oq%i$oq!g$oq~P!:aO!Y5QO~O!]5RO!_)PX~P#/XOa$zX!_$zX%^]X'y$zX!]$zX~P!/wO%^5UOaoXnoX!QoX!_oX'xoX'yoX(xoX(yoX!]oX~Op5VO(S#nO~O%^5UO~Ob5]O%j5^O(S+nO(UTO(XUO!]'sX!^'sX~O!]0}O!^)Wa~O[5bO~O`5cO~Oa%mO'y%mO~P#/XO!]5kO#`5mO!^)TX~O!^5nO~Or5tOs!nO!S*fO!b!yO!c!vO!d!vO!|;wO#T!pO#U!pO#V!pO#W!pO#X!pO#[5sO#]!zO(T!lO(UTO(XUO(d!mO(n!sO~O!^5rO~P%:YOn5yO!_1gO%i5xO~Oh%VOn5yO!_1gO%i5xO~Ob6QO(S#nO(UTO(XUO!]'rX!^'rX~O!]1rO!^)Ua~O(UTO(XUO(d6SO~O`6WO~Oj6ZO&Z6[O~PM|O!k6]O~P%[Oa6_O~Oa6_O~P%[Ob2YO!^6dO&i2XO~P`O!g6fO~O!g6hOh(ii!](ii!^(ii!g(ii!l(iir(ii(q(ii~O!]#hi!^#hi~P#B]O#`6iO!]#hi!^#hi~O!]!ai!^!ai~P#B]Oa%mO#`6rO'y%mO~Oa%mO!g#vO#`6rO'y%mO~O!](sq!k(sqa(sq'y(sq~P!:aO!](hO!k(rq~O!S%gO!_%hO#j6yO(S%fO~O!_'^O%i6|O~On7QO!_'^O%i6|O~O#k'haP'haR'ha['haa'haj'har'ha!S'ha!l'ha!p'ha#R'ha#n'ha#o'ha#p'ha#q'ha#r'ha#s'ha#t'ha#u'ha#v'ha#x'ha#z'ha#{'ha'y'ha(`'ha(q'ha!k'ha!Y'ha'v'hav'ha!_'ha%i'ha!g'ha~P%,XO#k'jaP'jaR'ja['jaa'jaj'jar'ja!S'ja!l'ja!p'ja#R'ja#n'ja#o'ja#p'ja#q'ja#r'ja#s'ja#t'ja#u'ja#v'ja#x'ja#z'ja#{'ja'y'ja(`'ja(q'ja!k'ja!Y'ja'v'jav'ja!_'ja%i'ja!g'ja~P%,zO#k$|iP$|iR$|i[$|ia$|ij$|ir$|i!S$|i!]$|i!l$|i!p$|i#R$|i#n$|i#o$|i#p$|i#q$|i#r$|i#s$|i#t$|i#u$|i#v$|i#x$|i#z$|i#{$|i'y$|i(`$|i(q$|i!k$|i!Y$|i'v$|i#`$|iv$|i!_$|i%i$|i!g$|i~P#/XO#k%aiP%aiR%ai[%aia%aij%air%ai!S%ai!l%ai!p%ai#R%ai#n%ai#o%ai#p%ai#q%ai#r%ai#s%ai#t%ai#u%ai#v%ai#x%ai#z%ai#{%ai'y%ai(`%ai(q%ai!k%ai!Y%ai'v%aiv%ai!_%ai%i%ai!g%ai~P%4[O#k%ciP%ciR%ci[%cia%cij%cir%ci!S%ci!l%ci!p%ci#R%ci#n%ci#o%ci#p%ci#q%ci#r%ci#s%ci#t%ci#u%ci#v%ci#x%ci#z%ci#{%ci'y%ci(`%ci(q%ci!k%ci!Y%ci'v%civ%ci!_%ci%i%ci!g%ci~P%4}O!]'Xa!k'Xa~P!:aO!].oO!k(ji~O$O#ci!]#ci!^#ci~P#B]OP$[OR#zO!Q#yO!S#{O!l#xO!p$[O(`VO[#mij#mir#mi#R#mi#o#mi#p#mi#q#mi#r#mi#s#mi#t#mi#u#mi#v#mi#x#mi#z#mi#{#mi$O#mi(q#mi(x#mi(y#mi!]#mi!^#mi~O#n#mi~P%MXO#n<PO~P%MXOP$[OR#zOr<]O!Q#yO!S#{O!l#xO!p$[O#n<PO#o<QO#p<QO#q<QO(`VO[#mij#mi#R#mi#s#mi#t#mi#u#mi#v#mi#x#mi#z#mi#{#mi$O#mi(q#mi(x#mi(y#mi!]#mi!^#mi~O#r#mi~P& aO#r<RO~P& aOP$[OR#zO[<_Oj<SOr<]O!Q#yO!S#{O!l#xO!p$[O#R<SO#n<PO#o<QO#p<QO#q<QO#r<RO#s<SO#t<SO#u<^O(`VO#x#mi#z#mi#{#mi$O#mi(q#mi(x#mi(y#mi!]#mi!^#mi~O#v#mi~P&#iOP$[OR#zO[<_Oj<SOr<]O!Q#yO!S#{O!l#xO!p$[O#R<SO#n<PO#o<QO#p<QO#q<QO#r<RO#s<SO#t<SO#u<^O#v<TO(`VO(y#}O#z#mi#{#mi$O#mi(q#mi(x#mi!]#mi!^#mi~O#x<VO~P&%jO#x#mi~P&%jO#v<TO~P&#iOP$[OR#zO[<_Oj<SOr<]O!Q#yO!S#{O!l#xO!p$[O#R<SO#n<PO#o<QO#p<QO#q<QO#r<RO#s<SO#t<SO#u<^O#v<TO#x<VO(`VO(x#|O(y#}O#{#mi$O#mi(q#mi!]#mi!^#mi~O#z#mi~P&'yO#z<XO~P&'yOa#|y!]#|y'y#|y'v#|y!Y#|y!k#|yv#|y!_#|y%i#|y!g#|y~P!:aO[#mij#mir#mi#R#mi#r#mi#s#mi#t#mi#u#mi#v#mi#x#mi#z#mi#{#mi$O#mi(q#mi!]#mi!^#mi~OP$[OR#zO!Q#yO!S#{O!l#xO!p$[O#n<PO#o<QO#p<QO#q<QO(`VO(x#mi(y#mi~P&*uOn=}O!Q)|O'x)}O(x$}O(y%POP#miR#mi!S#mi!l#mi!p#mi#n#mi#o#mi#p#mi#q#mi(`#mi~P&*uO#S$dOP(_XR(_X[(_Xj(_Xn(_Xr(_X!Q(_X!S(_X!l(_X!p(_X#R(_X#n(_X#o(_X#p(_X#q(_X#r(_X#s(_X#t(_X#u(_X#v(_X#x(_X#z(_X#{(_X$O(_X'x(_X(`(_X(q(_X(x(_X(y(_X!](_X!^(_X~O$O$Pi!]$Pi!^$Pi~P#B]O$O!ri!^!ri~P$*bOg'[a!]'[a~P!0{O!^7dO~O!]'ca!^'ca~P#B]O!Y7eO~P#/XO!g#vO(q'nO!]'da!k'da~O!]/kO!k(}i~O!]/kO!g#vO!k(}i~Og$|q!]$|q#`$|q$O$|q~P!0{O!Y'fa!]'fa~P#/XO!g7lO~O!]/sO!Y)Oi~P#/XO!]/sO!Y)Oi~O!Y7oO~Oh%VOr7tO!l%dO(q'nO~Oj7vO!g#vO~Or7yO!g#vO(q'nO~O!Q)|O'x)}O(y%POn'ia(x'ia!]'ia#`'ia~Og'ia$O'ia~P&3vO!Q)|O'x)}On'ka(x'ka(y'ka!]'ka#`'ka~Og'ka$O'ka~P&4iO!Y7{O~Og%Oq!]%Oq#`%Oq$O%Oq~P!0{Og(^q!](^q~P!0{O#`7|Og(^q!](^q~P!0{Oa$oy!]$oy'y$oy'v$oy!Y$oy!k$oyv$oy!_$oy%i$oy!g$oy~P!:aO!g6hO~O!]5RO!_)Pa~O!_'^OP$TaR$Ta[$Taj$Tar$Ta!Q$Ta!S$Ta!]$Ta!l$Ta!p$Ta#R$Ta#n$Ta#o$Ta#p$Ta#q$Ta#r$Ta#s$Ta#t$Ta#u$Ta#v$Ta#x$Ta#z$Ta#{$Ta(`$Ta(q$Ta(x$Ta(y$Ta~O%i6|O~P&7ZO%^8QOa%[i!_%[i'y%[i!]%[i~Oa#cy!]#cy'y#cy'v#cy!Y#cy!k#cyv#cy!_#cy%i#cy!g#cy~P!:aO[8SO~Ob8UO(S+nO(UTO(XUO~O!]0}O!^)Wi~O`8YO~O(d(zO!]'oX!^'oX~O!]5kO!^)Ta~O!^8cO~P%:YO(n!sO~P$${O#[8dO~O!_1gO~O!_1gO%i8fO~On8iO!_1gO%i8fO~O[8nO!]'ra!^'ra~O!]1rO!^)Ui~O!k8rO~O!k8sO~O!k8vO~O!k8vO~P%[Oa8xO~O!g8yO~O!k8zO~O!](vi!^(vi~P#B]Oa%mO#`9SO'y%mO~O!](sy!k(sya(sy'y(sy~P!:aO!](hO!k(ry~O%i9VO~P&7ZO!_'^O%i9VO~O#k$|qP$|qR$|q[$|qa$|qj$|qr$|q!S$|q!]$|q!l$|q!p$|q#R$|q#n$|q#o$|q#p$|q#q$|q#r$|q#s$|q#t$|q#u$|q#v$|q#x$|q#z$|q#{$|q'y$|q(`$|q(q$|q!k$|q!Y$|q'v$|q#`$|qv$|q!_$|q%i$|q!g$|q~P#/XO#k'iaP'iaR'ia['iaa'iaj'iar'ia!S'ia!l'ia!p'ia#R'ia#n'ia#o'ia#p'ia#q'ia#r'ia#s'ia#t'ia#u'ia#v'ia#x'ia#z'ia#{'ia'y'ia(`'ia(q'ia!k'ia!Y'ia'v'iav'ia!_'ia%i'ia!g'ia~P&3vO#k'kaP'kaR'ka['kaa'kaj'kar'ka!S'ka!l'ka!p'ka#R'ka#n'ka#o'ka#p'ka#q'ka#r'ka#s'ka#t'ka#u'ka#v'ka#x'ka#z'ka#{'ka'y'ka(`'ka(q'ka!k'ka!Y'ka'v'kav'ka!_'ka%i'ka!g'ka~P&4iO#k%OqP%OqR%Oq[%Oqa%Oqj%Oqr%Oq!S%Oq!]%Oq!l%Oq!p%Oq#R%Oq#n%Oq#o%Oq#p%Oq#q%Oq#r%Oq#s%Oq#t%Oq#u%Oq#v%Oq#x%Oq#z%Oq#{%Oq'y%Oq(`%Oq(q%Oq!k%Oq!Y%Oq'v%Oq#`%Oqv%Oq!_%Oq%i%Oq!g%Oq~P#/XO!]'Xi!k'Xi~P!:aO$O#cq!]#cq!^#cq~P#B]O(x$}OP%aaR%aa[%aaj%aar%aa!S%aa!l%aa!p%aa#R%aa#n%aa#o%aa#p%aa#q%aa#r%aa#s%aa#t%aa#u%aa#v%aa#x%aa#z%aa#{%aa$O%aa(`%aa(q%aa!]%aa!^%aa~On%aa!Q%aa'x%aa(y%aa~P&HnO(y%POP%caR%ca[%caj%car%ca!S%ca!l%ca!p%ca#R%ca#n%ca#o%ca#p%ca#q%ca#r%ca#s%ca#t%ca#u%ca#v%ca#x%ca#z%ca#{%ca$O%ca(`%ca(q%ca!]%ca!^%ca~On%ca!Q%ca'x%ca(x%ca~P&JuOn=}O!Q)|O'x)}O(y%PO~P&HnOn=}O!Q)|O'x)}O(x$}O~P&JuOR0cO!Q0cO!S0dO#S$dOP}a[}aj}an}ar}a!l}a!p}a#R}a#n}a#o}a#p}a#q}a#r}a#s}a#t}a#u}a#v}a#x}a#z}a#{}a$O}a'x}a(`}a(q}a(x}a(y}a!]}a!^}a~O!Q)|O'x)}OP$saR$sa[$saj$san$sar$sa!S$sa!l$sa!p$sa#R$sa#n$sa#o$sa#p$sa#q$sa#r$sa#s$sa#t$sa#u$sa#v$sa#x$sa#z$sa#{$sa$O$sa(`$sa(q$sa(x$sa(y$sa!]$sa!^$sa~O!Q)|O'x)}OP$uaR$ua[$uaj$uan$uar$ua!S$ua!l$ua!p$ua#R$ua#n$ua#o$ua#p$ua#q$ua#r$ua#s$ua#t$ua#u$ua#v$ua#x$ua#z$ua#{$ua$O$ua(`$ua(q$ua(x$ua(y$ua!]$ua!^$ua~On=}O!Q)|O'x)}O(x$}O(y%PO~OP%TaR%Ta[%Taj%Tar%Ta!S%Ta!l%Ta!p%Ta#R%Ta#n%Ta#o%Ta#p%Ta#q%Ta#r%Ta#s%Ta#t%Ta#u%Ta#v%Ta#x%Ta#z%Ta#{%Ta$O%Ta(`%Ta(q%Ta!]%Ta!^%Ta~P'%zO$O$mq!]$mq!^$mq~P#B]O$O$oq!]$oq!^$oq~P#B]O!^9dO~O$O9eO~P!0{O!g#vO!]'di!k'di~O!g#vO(q'nO!]'di!k'di~O!]/kO!k(}q~O!Y'fi!]'fi~P#/XO!]/sO!Y)Oq~Or9lO!g#vO(q'nO~O[9nO!Y9mO~P#/XO!Y9mO~Oj9tO!g#vO~Og(^y!](^y~P!0{O!]'ma!_'ma~P#/XOa%[q!_%[q'y%[q!]%[q~P#/XO[9yO~O!]0}O!^)Wq~O#`9}O!]'oa!^'oa~O!]5kO!^)Ti~P#B]O!S:PO~O!_1gO%i:SO~O(UTO(XUO(d:XO~O!]1rO!^)Uq~O!k:[O~O!k:]O~O!k:^O~O!k:^O~P%[O#`:aO!]#hy!^#hy~O!]#hy!^#hy~P#B]O%i:fO~P&7ZO!_'^O%i:fO~O$O#|y!]#|y!^#|y~P#B]OP$|iR$|i[$|ij$|ir$|i!S$|i!l$|i!p$|i#R$|i#n$|i#o$|i#p$|i#q$|i#r$|i#s$|i#t$|i#u$|i#v$|i#x$|i#z$|i#{$|i$O$|i(`$|i(q$|i!]$|i!^$|i~P'%zO!Q)|O'x)}O(y%POP'haR'ha['haj'han'har'ha!S'ha!l'ha!p'ha#R'ha#n'ha#o'ha#p'ha#q'ha#r'ha#s'ha#t'ha#u'ha#v'ha#x'ha#z'ha#{'ha$O'ha(`'ha(q'ha(x'ha!]'ha!^'ha~O!Q)|O'x)}OP'jaR'ja['jaj'jan'jar'ja!S'ja!l'ja!p'ja#R'ja#n'ja#o'ja#p'ja#q'ja#r'ja#s'ja#t'ja#u'ja#v'ja#x'ja#z'ja#{'ja$O'ja(`'ja(q'ja(x'ja(y'ja!]'ja!^'ja~O(x$}OP%aiR%ai[%aij%ain%air%ai!Q%ai!S%ai!l%ai!p%ai#R%ai#n%ai#o%ai#p%ai#q%ai#r%ai#s%ai#t%ai#u%ai#v%ai#x%ai#z%ai#{%ai$O%ai'x%ai(`%ai(q%ai(y%ai!]%ai!^%ai~O(y%POP%ciR%ci[%cij%cin%cir%ci!Q%ci!S%ci!l%ci!p%ci#R%ci#n%ci#o%ci#p%ci#q%ci#r%ci#s%ci#t%ci#u%ci#v%ci#x%ci#z%ci#{%ci$O%ci'x%ci(`%ci(q%ci(x%ci!]%ci!^%ci~O$O$oy!]$oy!^$oy~P#B]O$O#cy!]#cy!^#cy~P#B]O!g#vO!]'dq!k'dq~O!]/kO!k(}y~O!Y'fq!]'fq~P#/XOr:pO!g#vO(q'nO~O[:tO!Y:sO~P#/XO!Y:sO~Og(^!R!](^!R~P!0{Oa%[y!_%[y'y%[y!]%[y~P#/XO!]0}O!^)Wy~O!]5kO!^)Tq~O(S:zO~O!_1gO%i:}O~O!k;QO~O%i;VO~P&7ZOP$|qR$|q[$|qj$|qr$|q!S$|q!l$|q!p$|q#R$|q#n$|q#o$|q#p$|q#q$|q#r$|q#s$|q#t$|q#u$|q#v$|q#x$|q#z$|q#{$|q$O$|q(`$|q(q$|q!]$|q!^$|q~P'%zO!Q)|O'x)}O(y%POP'iaR'ia['iaj'ian'iar'ia!S'ia!l'ia!p'ia#R'ia#n'ia#o'ia#p'ia#q'ia#r'ia#s'ia#t'ia#u'ia#v'ia#x'ia#z'ia#{'ia$O'ia(`'ia(q'ia(x'ia!]'ia!^'ia~O!Q)|O'x)}OP'kaR'ka['kaj'kan'kar'ka!S'ka!l'ka!p'ka#R'ka#n'ka#o'ka#p'ka#q'ka#r'ka#s'ka#t'ka#u'ka#v'ka#x'ka#z'ka#{'ka$O'ka(`'ka(q'ka(x'ka(y'ka!]'ka!^'ka~OP%OqR%Oq[%Oqj%Oqr%Oq!S%Oq!l%Oq!p%Oq#R%Oq#n%Oq#o%Oq#p%Oq#q%Oq#r%Oq#s%Oq#t%Oq#u%Oq#v%Oq#x%Oq#z%Oq#{%Oq$O%Oq(`%Oq(q%Oq!]%Oq!^%Oq~P'%zOg%e!Z!]%e!Z#`%e!Z$O%e!Z~P!0{O!Y;ZO~P#/XOr;[O!g#vO(q'nO~O[;^O!Y;ZO~P#/XO!]'oq!^'oq~P#B]O!]#h!Z!^#h!Z~P#B]O#k%e!ZP%e!ZR%e!Z[%e!Za%e!Zj%e!Zr%e!Z!S%e!Z!]%e!Z!l%e!Z!p%e!Z#R%e!Z#n%e!Z#o%e!Z#p%e!Z#q%e!Z#r%e!Z#s%e!Z#t%e!Z#u%e!Z#v%e!Z#x%e!Z#z%e!Z#{%e!Z'y%e!Z(`%e!Z(q%e!Z!k%e!Z!Y%e!Z'v%e!Z#`%e!Zv%e!Z!_%e!Z%i%e!Z!g%e!Z~P#/XOr;fO!g#vO(q'nO~O!Y;gO~P#/XOr;nO!g#vO(q'nO~O!Y;oO~P#/XOP%e!ZR%e!Z[%e!Zj%e!Zr%e!Z!S%e!Z!l%e!Z!p%e!Z#R%e!Z#n%e!Z#o%e!Z#p%e!Z#q%e!Z#r%e!Z#s%e!Z#t%e!Z#u%e!Z#v%e!Z#x%e!Z#z%e!Z#{%e!Z$O%e!Z(`%e!Z(q%e!Z!]%e!Z!^%e!Z~P'%zOr;rO!g#vO(q'nO~Ov(eX~P1qO!Q%qO~P!)PO(T!lO~P!)PO!YfX!]fX#`fX~P%0kOP]XR]X[]Xj]Xr]X!Q]X!S]X!]]X!]fX!l]X!p]X#R]X#S]X#`]X#`fX#kfX#n]X#o]X#p]X#q]X#r]X#s]X#t]X#u]X#v]X#x]X#z]X#{]X$Q]X(`]X(q]X(x]X(y]X~O!gfX!k]X!kfX(qfX~P'JsOP;vOQ;vOSfOd=rOe!iOpkOr;vOskOtkOzkO|;vO!O;vO!SWO!WkO!XkO!_XO!i;yO!lZO!o;vO!p;vO!q;vO!s;zO!u;}O!x!hO$W!kO$n=pO(S)ZO(UTO(XUO(`VO(n[O~O!]<ZO!^$qa~Oh%VOp%WOr%XOs$tOt$tOz%YO|%ZO!O<fO!S${O!_$|O!i=wO!l$xO#j<lO$W%_O$t<hO$v<jO$y%`O(S(tO(UTO(XUO(`$uO(x$}O(y%PO~Ol)bO~P( iOr!eX(q!eX~P# }Or(iX(q(iX~P#!pO!^]X!^fX~P'JsO!YfX!Y$zX!]fX!]$zX#`fX~P!/wO#k<OO~O!g#vO#k<OO~O#`<`O~Oj<SO~O#`<pO!](vX!^(vX~O#`<`O!](tX!^(tX~O#k<qO~Og<sO~P!0{O#k<yO~O#k<zO~O!g#vO#k<{O~O!g#vO#k<qO~O$O<|O~P#B]O#k<}O~O#k=OO~O#k=TO~O#k=UO~O#k=VO~O#k=WO~O$O=XO~P!0{O$O=YO~P!0{Ok#S#T#U#W#X#[#i#j#u$n$t$v$y%]%^%h%i%j%q%s%v%w%y%{~'}T#o!X'{(T#ps#n#qr!Q'|$]'|(S$_(d~",goto:"$8g)[PPPPPP)]PP)`P)qP+R/WPPPP6bPP6xPP<pPPP@dP@zP@zPPP@zPCSP@zP@zP@zPCWPC]PCzPHtPPPHxPPPPHxK{PPPLRLsPHxPHxPP! RHxPPPHxPHxP!#YHxP!&p!'u!(OP!(r!(v!(r!,TPPPPPPP!,t!'uPP!-U!.vP!2SHxHx!2X!5e!:R!:R!>QPPP!>YHxPPPPPPPPP!AiP!BvPPHx!DXPHxPHxHxHxHxHxPHx!EkP!HuP!K{P!LP!LZ!L_!L_P!HrP!Lc!LcP# iP# mHxPHx# s#$xCW@zP@zP@z@zP#&V@z@z#(i@z#+a@z#-m@z@z#.]#0q#0q#0v#1P#0q#1[PP#0qP@z#1t@z#5s@z@z6bPPP#9xPPP#:c#:cP#:cP#:y#:cPP#;PP#:vP#:v#;d#:v#<O#<U#<X)`#<[)`P#<c#<c#<cP)`P)`P)`P)`PP)`P#<i#<lP#<l)`P#<pP#<sP)`P)`P)`P)`P)`P)`)`PP#<y#=P#=[#=b#=h#=n#=t#>S#>Y#>d#>j#>t#>z#?[#?b#@S#@f#@l#@r#AQ#Ag#C[#Cj#Cq#E]#Ek#G]#Gk#Gq#Gw#G}#HX#H_#He#Ho#IR#IXPPPPPPPPPPP#I_PPPPPPP#JS#MZ#Ns#Nz$ SPPP$&nP$&w$)p$0Z$0^$0a$1`$1c$1j$1rP$1x$1{P$2i$2m$3e$4s$4x$5`PP$5e$5k$5o$5r$5v$5z$6v$7_$7v$7z$7}$8Q$8W$8Z$8_$8cR!|RoqOXst!Z#d%l&p&r&s&u,n,s2S2VY!vQ'^-`1g5qQ%svQ%{yQ&S|Q&h!VS'U!e-WQ'd!iS'j!r!yU*h$|*X*lQ+l%|Q+y&UQ,_&bQ-^']Q-h'eQ-p'kQ0U*nQ1q,`R<m;z%SdOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y,k,n,s-d-l-z.Q.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3d4q5y6Z6[6_6r8i8x9SS#q];w!r)]$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sU*{%[<e<fQ+q&OQ,a&eQ,h&mQ0r+dQ0w+fQ1S+rQ1y,fQ3W.bQ5V0vQ5]0}Q6Q1rQ7O3[Q8U5^R9Y7Q'QkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=s!S!nQ!r!v!y!z$|'U']'^'j'k'l*h*l*n*o-W-^-`-p0U0X1g5q5s%[$ti#v$b$c$d$x${%O%Q%]%^%b)w*P*R*T*W*^*d*t*u+c+f+},Q.a.z/_/h/r/s/u0Y0[0g0h0i1^1a1i3Z4U4V4a4f4w5R5U5x6|7l7v7|8Q8f9V9e9n9t:S:f:t:};V;^<^<_<a<b<c<d<g<h<i<j<k<l<t<u<v<w<y<z<}=O=P=Q=R=S=T=U=X=Y=p=x=y=|=}Q&V|Q'S!eS'Y%h-ZQ+q&OQ,a&eQ0f+OQ1S+rQ1X+xQ1x,eQ1y,fQ5]0}Q5f1ZQ6Q1rQ6T1tQ6U1wQ8U5^Q8X5cQ8q6WQ9|8YQ:Y8nR<o*XrnOXst!V!Z#d%l&g&p&r&s&u,n,s2S2VR,c&i&z^OPXYstuvwz!Z!`!g!j!o#S#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=r=s[#]WZ#W#Z'V(R!b%im#h#i#l$x%d%g([(f(g(h*W*[*_+W+X+Z,j-Q.O.U.V.W.Y/h/k2[3S3T4X6h6yQ%vxQ%zyS&P|&UQ&]!TQ'a!hQ'c!iQ(o#sS+k%{%|Q+o&OQ,Y&`Q,^&bS-g'd'eQ.d(pQ0{+lQ1R+rQ1T+sQ1W+wQ1l,ZS1p,_,`Q2t-hQ5[0}Q5`1QQ5e1YQ6P1qQ8T5^Q8W5bQ9x8SR:w9y!U$zi$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=y!^%xy!i!u%z%{%|'T'c'd'e'i's*g+k+l-T-g-h-o/{0O0{2m2t2{4i4j4m7s9pQ+e%vQ,O&YQ,R&ZQ,]&bQ.c(oQ1k,YU1o,^,_,`Q3].dQ5z1lS6O1p1qQ8m6P#f=t#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}o=u<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=YW%Ti%V*v=pS&Y!Q&gQ&Z!RQ&[!SQ+S%cR+|&W%]%Si#v$b$c$d$x${%O%Q%]%^%b)w*P*R*T*W*^*d*t*u+c+f+},Q.a.z/_/h/r/s/u0Y0[0g0h0i1^1a1i3Z4U4V4a4f4w5R5U5x6|7l7v7|8Q8f9V9e9n9t:S:f:t:};V;^<^<_<a<b<c<d<g<h<i<j<k<l<t<u<v<w<y<z<}=O=P=Q=R=S=T=U=X=Y=p=x=y=|=}T)x$u)yV*{%[<e<fW'Y!e%h*X-ZS({#y#zQ+`%qQ+v&RS.](k(lQ1b,SQ4x0cR8^5k'QkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=s$i$^c#Y#e%p%r%t(Q(W(r(w)P)Q)R)S)T)U)V)W)X)Y)[)^)`)e)o+a+u-U-s-x-}.P.n.q.u.w.x.y/]0j2c2f2v2}3c3h3i3j3k3l3m3n3o3p3q3r3s3t3w3x4P5O5Y6k6q6v7V7W7a7b8`8|9Q9[9b9c:c:y;R;x=gT#TV#U'RkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sQ'W!eR2i-W!W!nQ!e!r!v!y!z$|'U']'^'j'k'l*X*h*l*n*o-W-^-`-p0U0X1g5q5sR1d,UnqOXst!Z#d%l&p&r&s&u,n,s2S2VQ&w!^Q't!xS(q#u<OQ+i%yQ,W&]Q,X&_Q-e'bQ-r'mS.m(v<qS0k+U<{Q0y+jQ1f,VQ2Z,uQ2],vQ2e-RQ2r-fQ2u-jS5P0l=VQ5W0zS5Z0|=WQ6j2gQ6n2sQ6s2zQ8R5XQ8}6lQ9O6oQ9R6tR:`8z$d$]c#Y#e%r%t(Q(W(r(w)P)Q)R)S)T)U)V)W)X)Y)[)^)`)e)o+a+u-U-s-x-}.P.n.q.u.x.y/]0j2c2f2v2}3c3h3i3j3k3l3m3n3o3p3q3r3s3t3w3x4P5O5Y6k6q6v7V7W7a7b8`8|9Q9[9b9c:c:y;R;x=gS(m#p'gQ(}#zS+_%p.wS.^(l(nR3U._'QkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sS#q];wQ&r!XQ&s!YQ&u![Q&v!]R2R,qQ'_!hQ+b%vQ-c'aS.`(o+eQ2p-bW3Y.c.d0q0sQ6m2qW6z3V3X3]5TU9U6{6}7PU:e9W9X9ZS;T:d:gQ;b;UR;j;cU!wQ'^-`T5o1g5q!Q_OXZ`st!V!Z#d#h%d%l&g&i&p&r&s&u(h,n,s.V2S2V]!pQ!r'^-`1g5qT#q];w%^{OPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9SS({#y#zS.](k(l!s=^$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sU$fd)],hS(n#p'gU*s%R(u3vU0e*z.i7]Q5T0rQ6{3WQ9X7OR:g9Ym!tQ!r!v!y!z'^'j'k'l-`-p1g5q5sQ'r!uS(d#g1|S-n'i'uQ/n*ZQ/{*gQ2|-qQ4]/oQ4i/}Q4j0OQ4o0WQ7h4WS7s4k4mS7w4p4rQ9g7iQ9k7oQ9p7tQ9u7yS:o9l9mS;Y:p:sS;e;Z;[S;m;f;gS;q;n;oR;t;rQ#wbQ'q!uS(c#g1|S(e#m+TQ+V%eQ+g%wQ+m%}U-m'i'r'uQ.R(dQ/m*ZQ/|*gQ0P*iQ0x+hQ1m,[S2y-n-qQ3R.ZS4[/n/oQ4e/yS4h/{0WQ4l0QQ5|1nQ6u2|Q7g4WQ7k4]U7r4i4o4rQ7u4nQ8k5}S9f7h7iQ9j7oQ9r7wQ9s7xQ:V8lQ:m9gS:n9k9mQ:v9uQ;P:WS;X:o:sS;d;Y;ZS;l;e;gS;p;m;oQ;s;qQ;u;tQ=a=[Q=l=eR=m=fV!wQ'^-`%^aOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9SS#wz!j!r=Z$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sR=a=r%^bOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9SQ%ej!^%wy!i!u%z%{%|'T'c'd'e'i's*g+k+l-T-g-h-o/{0O0{2m2t2{4i4j4m7s9pS%}z!jQ+h%xQ,[&bW1n,],^,_,`U5}1o1p1qS8l6O6PQ:W8m!r=[$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sQ=e=qR=f=r%QeOPXYstuvw!Z!`!g!o#S#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&p&r&s&u&y'R'`'p(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9SY#bWZ#W#Z(R!b%im#h#i#l$x%d%g([(f(g(h*W*[*_+W+X+Z,j-Q.O.U.V.W.Y/h/k2[3S3T4X6h6yQ,i&m!p=]$Z$n)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sR=`'VU'Z!e%h*XR2k-Z%SdOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y,k,n,s-d-l-z.Q.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3d4q5y6Z6[6_6r8i8x9S!r)]$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sQ,h&mQ0r+dQ3W.bQ7O3[R9Y7Q!b$Tc#Y%p(Q(W(r(w)X)Y)^)e+u-s-x-}.P.n.q/]0j2v2}3c3s5O5Y6q6v7V9Q:c;x!P<U)[)o-U.w2c2f3h3q3r3w4P6k7W7a7b8`8|9[9b9c:y;R=g!f$Vc#Y%p(Q(W(r(w)U)V)X)Y)^)e+u-s-x-}.P.n.q/]0j2v2}3c3s5O5Y6q6v7V9Q:c;x!T<W)[)o-U.w2c2f3h3n3o3q3r3w4P6k7W7a7b8`8|9[9b9c:y;R=g!^$Zc#Y%p(Q(W(r(w)^)e+u-s-x-}.P.n.q/]0j2v2}3c3s5O5Y6q6v7V9Q:c;xQ4V/fz=s)[)o-U.w2c2f3h3w4P6k7W7a7b8`8|9[9b9c:y;R=gQ=x=zR=y={'QkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sS$oh$pR3|/P'XgOPWXYZhstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n$p%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/P/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sT$kf$qQ$ifS)h$l)lR)t$qT$jf$qT)j$l)l'XhOPWXYZhstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n$p%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/P/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sT$oh$pQ$rhR)s$p%^jOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9S!s=q$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=s#glOPXZst!Z!`!o#S#d#o#{$n%l&i&l&m&p&r&s&u&y'R'`(|)q*f+Y+d,k,n,s-d.b/Q/i0V0d1j1z1{1}2P2S2V2X3[3{4q5y6Z6[6_7Q8i8x!U%Ri$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=y#f(u#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}Q+P%`Q/^)|o3v<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=Y!U$yi$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=yQ*`$zU*i$|*X*lQ+Q%aQ0Q*j#f=c#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}n=d<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=YQ=h=tQ=i=uQ=j=vR=k=w!U%Ri$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=y#f(u#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}o3v<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=YnoOXst!Z#d%l&p&r&s&u,n,s2S2VS*c${*WQ,|&|Q,}'OR4`/s%[%Si#v$b$c$d$x${%O%Q%]%^%b)w*P*R*T*W*^*d*t*u+c+f+},Q.a.z/_/h/r/s/u0Y0[0g0h0i1^1a1i3Z4U4V4a4f4w5R5U5x6|7l7v7|8Q8f9V9e9n9t:S:f:t:};V;^<^<_<a<b<c<d<g<h<i<j<k<l<t<u<v<w<y<z<}=O=P=Q=R=S=T=U=X=Y=p=x=y=|=}Q,P&ZQ1`,RQ5i1_R8]5jV*k$|*X*lU*k$|*X*lT5p1g5qS/y*f/iQ4n0VT7x4q:PQ+g%wQ0P*iQ0x+hQ1m,[Q5|1nQ8k5}Q:V8lR;P:W!U%Oi$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=yx*P$v)c*Q*r+R/q0^0_3y4^4{4|4}7f7z9v:l=b=n=oS0Y*q0Z#f<a#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}n<b<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=Y!d<t(s)a*Y*b.e.h.l/Y/f/v0p1]3`4S4_4c5h7R7U7m7p7}8P9i9q9w:q:u;W;];h=z={`<u3u7X7[7`9]:h:k;kS=P.g3aT=Q7Z9`!U%Qi$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=y|*R$v)c*S*q+R/b/q0^0_3y4^4s4{4|4}7f7z9v:l=b=n=oS0[*r0]#f<c#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}n<d<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=Y!h<v(s)a*Y*b.f.g.l/Y/f/v0p1]3^3`4S4_4c5h7R7S7U7m7p7}8P9i9q9w:q:u;W;];h=z={d<w3u7Y7Z7`9]9^:h:i:k;kS=R.h3bT=S7[9arnOXst!V!Z#d%l&g&p&r&s&u,n,s2S2VQ&d!UR,k&mrnOXst!V!Z#d%l&g&p&r&s&u,n,s2S2VR&d!UQ,T&[R1[+|snOXst!V!Z#d%l&g&p&r&s&u,n,s2S2VQ1h,YS5w1k1lU8e5u5v5zS:R8g8hS:{:Q:TQ;_:|R;i;`Q&k!VR,d&gR6T1tR:Y8nS&P|&UR1T+sQ&p!WR,n&qR,t&vT2T,s2VR,x&wQ,w&wR2^,xQ'w!{R-t'wSsOtQ#dXT%os#dQ#OTR'y#OQ#RUR'{#RQ)y$uR/Z)yQ#UVR(O#UQ#XWU(U#X(V-{Q(V#YR-{(WQ-X'WR2j-XQ.p(wS3e.p3fR3f.qQ-`'^R2n-`Y!rQ'^-`1g5qR'h!rQ.{)cR3z.{U#_W%g*WU(]#_(^-|Q(^#`R-|(XQ-['ZR2l-[t`OXst!V!Z#d%l&g&i&p&r&s&u,n,s2S2VS#hZ%dU#r`#h.VR.V(hQ(i#jQ.S(eW.[(i.S3P6wQ3P.TR6w3QQ)l$lR/R)lQ$phR)r$pQ$`cU)_$`-w<[Q-w;xR<[)oQ/l*ZW4Y/l4Z7j9hU4Z/m/n/oS7j4[4]R9h7k$e*O$v(s)a)c*Y*b*q*r*|*}+R.g.h.j.k.l/Y/b/d/f/q/v0^0_0p1]3^3_3`3u3y4S4^4_4c4s4u4{4|4}5h7R7S7T7U7Z7[7^7_7`7f7m7p7z7}8P9]9^9_9i9q9v9w:h:i:j:k:l:q:u;W;];h;k=b=n=o=z={Q/t*bU4b/t4d7nQ4d/vR7n4cS*l$|*XR0S*lx*Q$v)c*q*r+R/q0^0_3y4^4{4|4}7f7z9v:l=b=n=o!d.e(s)a*Y*b.g.h.l/Y/f/v0p1]3`4S4_4c5h7R7U7m7p7}8P9i9q9w:q:u;W;];h=z={U/c*Q.e7Xa7X3u7Z7[7`9]:h:k;kQ0Z*qQ3a.gU4t0Z3a9`R9`7Z|*S$v)c*q*r+R/b/q0^0_3y4^4s4{4|4}7f7z9v:l=b=n=o!h.f(s)a*Y*b.g.h.l/Y/f/v0p1]3^3`4S4_4c5h7R7S7U7m7p7}8P9i9q9w:q:u;W;];h=z={U/e*S.f7Ye7Y3u7Z7[7`9]9^:h:i:k;kQ0]*rQ3b.hU4v0]3b9aR9a7[Q*w%UR0a*wQ5S0pR8O5SQ+[%jR0o+[Q5l1bS8_5l:OR:O8`Q,V&]R1e,VQ5q1gR8b5qQ1s,aS6R1s8oR8o6TQ1O+oW5_1O5a8V9zQ5a1RQ8V5`R9z8WQ+t&PR1U+tQ2V,sR6c2VYrOXst#dQ&t!ZQ+^%lQ,m&pQ,o&rQ,p&sQ,r&uQ2Q,nS2T,s2VR6b2SQ%npQ&x!_Q&{!aQ&}!bQ'P!cQ'o!uQ+]%kQ+i%yQ+{&VQ,c&kQ,z&zW-k'i'q'r'uQ-r'mQ0R*kQ0y+jS1v,d,gQ2_,yQ2`,|Q2a,}Q2u-jW2w-m-n-q-sQ5W0zQ5d1XQ5g1]Q5{1mQ6V1xQ6a2RU6p2v2y2|Q6s2zQ8R5XQ8Z5fQ8[5hQ8a5pQ8j5|Q8p6US9P6q6uQ9R6tQ9{8XQ:U8kQ:Z8qQ:b9QQ:x9|Q;O:VQ;S:cR;a;PQ%yyQ'b!iQ'm!uU+j%z%{%|Q-R'TU-f'c'd'eS-j'i'sQ/z*gS0z+k+lQ2g-TS2s-g-hQ2z-oS4g/{0OQ5X0{Q6l2mQ6o2tQ6t2{U7q4i4j4mQ9o7sR:r9pS$wi=pR*x%VU%Ui%V=pR0`*vQ$viS(s#v+fS)a$b$cQ)c$dQ*Y$xS*b${*WQ*q%OQ*r%QQ*|%]Q*}%^Q+R%bQ.g<aQ.h<cQ.j<gQ.k<iQ.l<kQ/Y)wQ/b*PQ/d*RQ/f*TQ/q*^S/v*d/hQ0^*tQ0_*ul0p+c,Q.a1a1i3Z5x6|8f9V:S:f:};VQ1]+}Q3^<tQ3_<vQ3`<yS3u<^<_Q3y.zS4S/_4UQ4^/rQ4_/sQ4c/uQ4s0YQ4u0[Q4{0gQ4|0hQ4}0iQ5h1^Q7R<}Q7S=PQ7T=RQ7U=TQ7Z<bQ7[<dQ7^<hQ7_<jQ7`<lQ7f4VQ7m4aQ7p4fQ7z4wQ7}5RQ8P5UQ9]<zQ9^<uQ9_<wQ9i7lQ9q7vQ9v7|Q9w8QQ:h=OQ:i=QQ:j=SQ:k=UQ:l9eQ:q9nQ:u9tQ;W=XQ;]:tQ;h;^Q;k=YQ=b=pQ=n=xQ=o=yQ=z=|R={=}Q*z%[Q.i<eR7]<fnpOXst!Z#d%l&p&r&s&u,n,s2S2VQ!fPS#fZ#oQ&z!`W'f!o*f0V4qQ'}#SQ)O#{Q)p$nS,g&i&lQ,l&mQ,y&yS-O'R/iQ-b'`Q.s(|Q/V)qQ0m+YQ0s+dQ2O,kQ2q-dQ3X.bQ4O/QQ4y0dQ5v1jQ6X1zQ6Y1{Q6^1}Q6`2PQ6e2XQ7P3[Q7c3{Q8h5yQ8t6ZQ8u6[Q8w6_Q9Z7QQ:T8iR:_8x#[cOPXZst!Z!`!o#d#o#{%l&i&l&m&p&r&s&u&y'R'`(|*f+Y+d,k,n,s-d.b/i0V0d1j1z1{1}2P2S2V2X3[4q5y6Z6[6_7Q8i8xQ#YWQ#eYQ%puQ%rvS%tw!gS(Q#W(TQ(W#ZQ(r#uQ(w#xQ)P$OQ)Q$PQ)R$QQ)S$RQ)T$SQ)U$TQ)V$UQ)W$VQ)X$WQ)Y$XQ)[$ZQ)^$_Q)`$aQ)e$eW)o$n)q/Q3{Q+a%sQ+u&QS-U'V2hQ-s'pS-x(R-zQ-}(ZQ.P(bQ.n(vQ.q(xQ.u;vQ.w;yQ.x;zQ.y;}Q/]){Q0j+UQ2c-PQ2f-SQ2v-lQ2}.QQ3c.oQ3h<OQ3i<PQ3j<QQ3k<RQ3l<SQ3m<TQ3n<UQ3o<VQ3p<WQ3q<XQ3r<YQ3s.vQ3t<]Q3w<`Q3x<mQ4P<ZQ5O0lQ5Y0|Q6k<pQ6q2xQ6v3OQ7V3dQ7W<qQ7a<sQ7b<{Q8`5mQ8|6iQ9Q6rQ9[<|Q9b=VQ9c=WQ:c9SQ:y9}Q;R:aQ;x#SR=g=sR#[WR'X!el!tQ!r!v!y!z'^'j'k'l-`-p1g5q5sS'T!e-WU*g$|*X*lS-T'U']S0O*h*nQ0W*oQ2m-^Q4m0UR4r0XR(y#xQ!fQT-_'^-`]!qQ!r'^-`1g5qQ#p]R'g;wR)d$dY!uQ'^-`1g5qQ'i!rS's!v!yS'u!z5sS-o'j'kQ-q'lR2{-pT#kZ%dS#jZ%dS%jm,jU(e#h#i#lS.T(f(gQ.X(hQ0n+ZQ3Q.UU3R.V.W.YS6x3S3TR9T6yd#^W#W#Z%g(R([*W+W.O/hr#gZm#h#i#l%d(f(g(h+Z.U.V.W.Y3S3T6yS*Z$x*_Q/o*[Q1|,jQ2d-QQ4W/kQ6g2[Q7i4XQ8{6hT=_'V+XV#aW%g*WU#`W%g*WS(S#W([U(X#Z+W/hS-V'V+XT-y(R.OV'[!e%h*XQ$lfR)v$qT)k$l)lR3}/PT*]$x*_T*e${*WQ0q+cQ1_,QQ3V.aQ5j1aQ5u1iQ6}3ZQ8g5xQ9W6|Q:Q8fQ:d9VQ:|:SQ;U:fQ;`:}R;c;VnqOXst!Z#d%l&p&r&s&u,n,s2S2VQ&j!VR,c&gtmOXst!U!V!Z#d%l&g&p&r&s&u,n,s2S2VR,j&mT%km,jR1c,SR,b&eQ&T|R+z&UR+p&OT&n!W&qT&o!W&qT2U,s2V",nodeNames:"⚠ ArithOp ArithOp ?. JSXStartTag LineComment BlockComment Script Hashbang ExportDeclaration export Star as VariableName String Escape from ; default FunctionDeclaration async function VariableDefinition > < TypeParamList in out const TypeDefinition extends ThisType this LiteralType ArithOp Number BooleanLiteral TemplateType InterpolationEnd Interpolation InterpolationStart NullType null VoidType void TypeofType typeof MemberExpression . PropertyName [ TemplateString Escape Interpolation super RegExp ] ArrayExpression Spread , } { ObjectExpression Property async get set PropertyDefinition Block : NewTarget new NewExpression ) ( ArgList UnaryExpression delete LogicOp BitOp YieldExpression yield AwaitExpression await ParenthesizedExpression ClassExpression class ClassBody MethodDeclaration Decorator @ MemberExpression PrivatePropertyName CallExpression TypeArgList CompareOp < declare Privacy static abstract override PrivatePropertyDefinition PropertyDeclaration readonly accessor Optional TypeAnnotation Equals StaticBlock FunctionExpression ArrowFunction ParamList ParamList ArrayPattern ObjectPattern PatternProperty Privacy readonly Arrow MemberExpression BinaryExpression ArithOp ArithOp ArithOp ArithOp BitOp CompareOp instanceof satisfies CompareOp BitOp BitOp BitOp LogicOp LogicOp ConditionalExpression LogicOp LogicOp AssignmentExpression UpdateOp PostfixExpression CallExpression InstantiationExpression TaggedTemplateExpression DynamicImport import ImportMeta JSXElement JSXSelfCloseEndTag JSXSelfClosingTag JSXIdentifier JSXBuiltin JSXIdentifier JSXNamespacedName JSXMemberExpression JSXSpreadAttribute JSXAttribute JSXAttributeValue JSXEscape JSXEndTag JSXOpenTag JSXFragmentTag JSXText JSXEscape JSXStartCloseTag JSXCloseTag PrefixCast < ArrowFunction TypeParamList SequenceExpression InstantiationExpression KeyofType keyof UniqueType unique ImportType InferredType infer TypeName ParenthesizedType FunctionSignature ParamList NewSignature IndexedType TupleType Label ArrayType ReadonlyType ObjectType MethodType PropertyType IndexSignature PropertyDefinition CallSignature TypePredicate asserts is NewSignature new UnionType LogicOp IntersectionType LogicOp ConditionalType ParameterizedType ClassDeclaration abstract implements type VariableDeclaration let var using TypeAliasDeclaration InterfaceDeclaration interface EnumDeclaration enum EnumBody NamespaceDeclaration namespace module AmbientDeclaration declare GlobalDeclaration global ClassDeclaration ClassBody AmbientFunctionDeclaration ExportGroup VariableName VariableName ImportDeclaration ImportGroup ForStatement for ForSpec ForInSpec ForOfSpec of WhileStatement while WithStatement with DoStatement do IfStatement if else SwitchStatement switch SwitchBody CaseLabel case DefaultLabel TryStatement try CatchClause catch FinallyClause finally ReturnStatement return ThrowStatement throw BreakStatement break ContinueStatement continue DebuggerStatement debugger LabeledStatement ExpressionStatement SingleExpression SingleClassItem",maxTerm:379,context:Fze,nodeProps:[["isolate",-8,5,6,14,37,39,51,53,55,""],["group",-26,9,17,19,68,207,211,215,216,218,221,224,234,236,242,244,246,248,251,257,263,265,267,269,271,273,274,"Statement",-34,13,14,32,35,36,42,51,54,55,57,62,70,72,76,80,82,84,85,110,111,120,121,136,139,141,142,143,144,145,147,148,167,169,171,"Expression",-23,31,33,37,41,43,45,173,175,177,178,180,181,182,184,185,186,188,189,190,201,203,205,206,"Type",-3,88,103,109,"ClassItem"],["openedBy",23,"<",38,"InterpolationStart",56,"[",60,"{",73,"(",160,"JSXStartCloseTag"],["closedBy",-2,24,168,">",40,"InterpolationEnd",50,"]",61,"}",74,")",165,"JSXEndTag"]],propSources:[zze],skippedNodes:[0,5,6,277],repeatNodeCount:37,tokenData:"$Fq07[R!bOX%ZXY+gYZ-yZ[+g[]%Z]^.c^p%Zpq+gqr/mrs3cst:_tuEruvJSvwLkwx! Yxy!'iyz!(sz{!)}{|!,q|}!.O}!O!,q!O!P!/Y!P!Q!9j!Q!R#:O!R![#<_![!]#I_!]!^#Jk!^!_#Ku!_!`$![!`!a$$v!a!b$*T!b!c$,r!c!}Er!}#O$-|#O#P$/W#P#Q$4o#Q#R$5y#R#SEr#S#T$7W#T#o$8b#o#p$<r#p#q$=h#q#r$>x#r#s$@U#s$f%Z$f$g+g$g#BYEr#BY#BZ$A`#BZ$ISEr$IS$I_$A`$I_$I|Er$I|$I}$Dk$I}$JO$Dk$JO$JTEr$JT$JU$A`$JU$KVEr$KV$KW$A`$KW&FUEr&FU&FV$A`&FV;'SEr;'S;=`I|<%l?HTEr?HT?HU$A`?HUOEr(n%d_$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z&j&hT$i&jO!^&c!_#o&c#p;'S&c;'S;=`&w<%lO&c&j&zP;=`<%l&c'|'U]$i&j(Y!bOY&}YZ&cZw&}wx&cx!^&}!^!_'}!_#O&}#O#P&c#P#o&}#o#p'}#p;'S&};'S;=`(l<%lO&}!b(SU(Y!bOY'}Zw'}x#O'}#P;'S'};'S;=`(f<%lO'}!b(iP;=`<%l'}'|(oP;=`<%l&}'[(y]$i&j(VpOY(rYZ&cZr(rrs&cs!^(r!^!_)r!_#O(r#O#P&c#P#o(r#o#p)r#p;'S(r;'S;=`*a<%lO(rp)wU(VpOY)rZr)rs#O)r#P;'S)r;'S;=`*Z<%lO)rp*^P;=`<%l)r'[*dP;=`<%l(r#S*nX(Vp(Y!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g#S+^P;=`<%l*g(n+dP;=`<%l%Z07[+rq$i&j(Vp(Y!b'{0/lOX%ZXY+gYZ&cZ[+g[p%Zpq+gqr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p$f%Z$f$g+g$g#BY%Z#BY#BZ+g#BZ$IS%Z$IS$I_+g$I_$JT%Z$JT$JU+g$JU$KV%Z$KV$KW+g$KW&FU%Z&FU&FV+g&FV;'S%Z;'S;=`+a<%l?HT%Z?HT?HU+g?HUO%Z07[.ST(W#S$i&j'|0/lO!^&c!_#o&c#p;'S&c;'S;=`&w<%lO&c07[.n_$i&j(Vp(Y!b'|0/lOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z)3p/x`$i&j!p),Q(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`0z!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW1V`#v(Ch$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`2X!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW2d_#v(Ch$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'At3l_(U':f$i&j(Y!bOY4kYZ5qZr4krs7nsw4kwx5qx!^4k!^!_8p!_#O4k#O#P5q#P#o4k#o#p8p#p;'S4k;'S;=`:X<%lO4k(^4r_$i&j(Y!bOY4kYZ5qZr4krs7nsw4kwx5qx!^4k!^!_8p!_#O4k#O#P5q#P#o4k#o#p8p#p;'S4k;'S;=`:X<%lO4k&z5vX$i&jOr5qrs6cs!^5q!^!_6y!_#o5q#o#p6y#p;'S5q;'S;=`7h<%lO5q&z6jT$d`$i&jO!^&c!_#o&c#p;'S&c;'S;=`&w<%lO&c`6|TOr6yrs7]s;'S6y;'S;=`7b<%lO6y`7bO$d``7eP;=`<%l6y&z7kP;=`<%l5q(^7w]$d`$i&j(Y!bOY&}YZ&cZw&}wx&cx!^&}!^!_'}!_#O&}#O#P&c#P#o&}#o#p'}#p;'S&};'S;=`(l<%lO&}!r8uZ(Y!bOY8pYZ6yZr8prs9hsw8pwx6yx#O8p#O#P6y#P;'S8p;'S;=`:R<%lO8p!r9oU$d`(Y!bOY'}Zw'}x#O'}#P;'S'};'S;=`(f<%lO'}!r:UP;=`<%l8p(^:[P;=`<%l4k%9[:hh$i&j(Vp(Y!bOY%ZYZ&cZq%Zqr<Srs&}st%ZtuCruw%Zwx(rx!^%Z!^!_*g!_!c%Z!c!}Cr!}#O%Z#O#P&c#P#R%Z#R#SCr#S#T%Z#T#oCr#o#p*g#p$g%Z$g;'SCr;'S;=`El<%lOCr(r<__WS$i&j(Vp(Y!bOY<SYZ&cZr<Srs=^sw<Swx@nx!^<S!^!_Bm!_#O<S#O#P>`#P#o<S#o#pBm#p;'S<S;'S;=`Cl<%lO<S(Q=g]WS$i&j(Y!bOY=^YZ&cZw=^wx>`x!^=^!^!_?q!_#O=^#O#P>`#P#o=^#o#p?q#p;'S=^;'S;=`@h<%lO=^&n>gXWS$i&jOY>`YZ&cZ!^>`!^!_?S!_#o>`#o#p?S#p;'S>`;'S;=`?k<%lO>`S?XSWSOY?SZ;'S?S;'S;=`?e<%lO?SS?hP;=`<%l?S&n?nP;=`<%l>`!f?xWWS(Y!bOY?qZw?qwx?Sx#O?q#O#P?S#P;'S?q;'S;=`@b<%lO?q!f@eP;=`<%l?q(Q@kP;=`<%l=^'`@w]WS$i&j(VpOY@nYZ&cZr@nrs>`s!^@n!^!_Ap!_#O@n#O#P>`#P#o@n#o#pAp#p;'S@n;'S;=`Bg<%lO@ntAwWWS(VpOYApZrAprs?Ss#OAp#O#P?S#P;'SAp;'S;=`Ba<%lOAptBdP;=`<%lAp'`BjP;=`<%l@n#WBvYWS(Vp(Y!bOYBmZrBmrs?qswBmwxApx#OBm#O#P?S#P;'SBm;'S;=`Cf<%lOBm#WCiP;=`<%lBm(rCoP;=`<%l<S%9[C}i$i&j(n%1l(Vp(Y!bOY%ZYZ&cZr%Zrs&}st%ZtuCruw%Zwx(rx!Q%Z!Q![Cr![!^%Z!^!_*g!_!c%Z!c!}Cr!}#O%Z#O#P&c#P#R%Z#R#SCr#S#T%Z#T#oCr#o#p*g#p$g%Z$g;'SCr;'S;=`El<%lOCr%9[EoP;=`<%lCr07[FRk$i&j(Vp(Y!b$]#t(S,2j(d$I[OY%ZYZ&cZr%Zrs&}st%ZtuEruw%Zwx(rx}%Z}!OGv!O!Q%Z!Q![Er![!^%Z!^!_*g!_!c%Z!c!}Er!}#O%Z#O#P&c#P#R%Z#R#SEr#S#T%Z#T#oEr#o#p*g#p$g%Z$g;'SEr;'S;=`I|<%lOEr+dHRk$i&j(Vp(Y!b$]#tOY%ZYZ&cZr%Zrs&}st%ZtuGvuw%Zwx(rx}%Z}!OGv!O!Q%Z!Q![Gv![!^%Z!^!_*g!_!c%Z!c!}Gv!}#O%Z#O#P&c#P#R%Z#R#SGv#S#T%Z#T#oGv#o#p*g#p$g%Z$g;'SGv;'S;=`Iv<%lOGv+dIyP;=`<%lGv07[JPP;=`<%lEr(KWJ_`$i&j(Vp(Y!b#p(ChOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KWKl_$i&j$Q(Ch(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z,#xLva(y+JY$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sv%ZvwM{wx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KWNW`$i&j#z(Ch(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'At! c_(X';W$i&j(VpOY!!bYZ!#hZr!!brs!#hsw!!bwx!$xx!^!!b!^!_!%z!_#O!!b#O#P!#h#P#o!!b#o#p!%z#p;'S!!b;'S;=`!'c<%lO!!b'l!!i_$i&j(VpOY!!bYZ!#hZr!!brs!#hsw!!bwx!$xx!^!!b!^!_!%z!_#O!!b#O#P!#h#P#o!!b#o#p!%z#p;'S!!b;'S;=`!'c<%lO!!b&z!#mX$i&jOw!#hwx6cx!^!#h!^!_!$Y!_#o!#h#o#p!$Y#p;'S!#h;'S;=`!$r<%lO!#h`!$]TOw!$Ywx7]x;'S!$Y;'S;=`!$l<%lO!$Y`!$oP;=`<%l!$Y&z!$uP;=`<%l!#h'l!%R]$d`$i&j(VpOY(rYZ&cZr(rrs&cs!^(r!^!_)r!_#O(r#O#P&c#P#o(r#o#p)r#p;'S(r;'S;=`*a<%lO(r!Q!&PZ(VpOY!%zYZ!$YZr!%zrs!$Ysw!%zwx!&rx#O!%z#O#P!$Y#P;'S!%z;'S;=`!']<%lO!%z!Q!&yU$d`(VpOY)rZr)rs#O)r#P;'S)r;'S;=`*Z<%lO)r!Q!'`P;=`<%l!%z'l!'fP;=`<%l!!b/5|!'t_!l/.^$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z#&U!)O_!k!Lf$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z-!n!*[b$i&j(Vp(Y!b(T%&f#q(ChOY%ZYZ&cZr%Zrs&}sw%Zwx(rxz%Zz{!+d{!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW!+o`$i&j(Vp(Y!b#n(ChOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z+;x!,|`$i&j(Vp(Y!br+4YOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z,$U!.Z_!]+Jf$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z07[!/ec$i&j(Vp(Y!b!Q.2^OY%ZYZ&cZr%Zrs&}sw%Zwx(rx!O%Z!O!P!0p!P!Q%Z!Q![!3Y![!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z#%|!0ya$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!O%Z!O!P!2O!P!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z#%|!2Z_![!L^$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad!3eg$i&j(Vp(Y!bs'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![!3Y![!^%Z!^!_*g!_!g%Z!g!h!4|!h#O%Z#O#P&c#P#R%Z#R#S!3Y#S#X%Z#X#Y!4|#Y#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad!5Vg$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx{%Z{|!6n|}%Z}!O!6n!O!Q%Z!Q![!8S![!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S!8S#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad!6wc$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![!8S![!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S!8S#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad!8_c$i&j(Vp(Y!bs'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![!8S![!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S!8S#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z07[!9uf$i&j(Vp(Y!b#o(ChOY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcxz!;Zz{#-}{!P!;Z!P!Q#/d!Q!^!;Z!^!_#(i!_!`#7S!`!a#8i!a!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;Z?O!;fb$i&j(Vp(Y!b!X7`OY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcx!P!;Z!P!Q#&`!Q!^!;Z!^!_#(i!_!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;Z>^!<w`$i&j(Y!b!X7`OY!<nYZ&cZw!<nwx!=yx!P!<n!P!Q!Eq!Q!^!<n!^!_!Gr!_!}!<n!}#O!KS#O#P!Dy#P#o!<n#o#p!Gr#p;'S!<n;'S;=`!L]<%lO!<n<z!>Q^$i&j!X7`OY!=yYZ&cZ!P!=y!P!Q!>|!Q!^!=y!^!_!@c!_!}!=y!}#O!CW#O#P!Dy#P#o!=y#o#p!@c#p;'S!=y;'S;=`!Ek<%lO!=y<z!?Td$i&j!X7`O!^&c!_#W&c#W#X!>|#X#Z&c#Z#[!>|#[#]&c#]#^!>|#^#a&c#a#b!>|#b#g&c#g#h!>|#h#i&c#i#j!>|#j#k!>|#k#m&c#m#n!>|#n#o&c#p;'S&c;'S;=`&w<%lO&c7`!@hX!X7`OY!@cZ!P!@c!P!Q!AT!Q!}!@c!}#O!Ar#O#P!Bq#P;'S!@c;'S;=`!CQ<%lO!@c7`!AYW!X7`#W#X!AT#Z#[!AT#]#^!AT#a#b!AT#g#h!AT#i#j!AT#j#k!AT#m#n!AT7`!AuVOY!ArZ#O!Ar#O#P!B[#P#Q!@c#Q;'S!Ar;'S;=`!Bk<%lO!Ar7`!B_SOY!ArZ;'S!Ar;'S;=`!Bk<%lO!Ar7`!BnP;=`<%l!Ar7`!BtSOY!@cZ;'S!@c;'S;=`!CQ<%lO!@c7`!CTP;=`<%l!@c<z!C][$i&jOY!CWYZ&cZ!^!CW!^!_!Ar!_#O!CW#O#P!DR#P#Q!=y#Q#o!CW#o#p!Ar#p;'S!CW;'S;=`!Ds<%lO!CW<z!DWX$i&jOY!CWYZ&cZ!^!CW!^!_!Ar!_#o!CW#o#p!Ar#p;'S!CW;'S;=`!Ds<%lO!CW<z!DvP;=`<%l!CW<z!EOX$i&jOY!=yYZ&cZ!^!=y!^!_!@c!_#o!=y#o#p!@c#p;'S!=y;'S;=`!Ek<%lO!=y<z!EnP;=`<%l!=y>^!Ezl$i&j(Y!b!X7`OY&}YZ&cZw&}wx&cx!^&}!^!_'}!_#O&}#O#P&c#P#W&}#W#X!Eq#X#Z&}#Z#[!Eq#[#]&}#]#^!Eq#^#a&}#a#b!Eq#b#g&}#g#h!Eq#h#i&}#i#j!Eq#j#k!Eq#k#m&}#m#n!Eq#n#o&}#o#p'}#p;'S&};'S;=`(l<%lO&}8r!GyZ(Y!b!X7`OY!GrZw!Grwx!@cx!P!Gr!P!Q!Hl!Q!}!Gr!}#O!JU#O#P!Bq#P;'S!Gr;'S;=`!J|<%lO!Gr8r!Hse(Y!b!X7`OY'}Zw'}x#O'}#P#W'}#W#X!Hl#X#Z'}#Z#[!Hl#[#]'}#]#^!Hl#^#a'}#a#b!Hl#b#g'}#g#h!Hl#h#i'}#i#j!Hl#j#k!Hl#k#m'}#m#n!Hl#n;'S'};'S;=`(f<%lO'}8r!JZX(Y!bOY!JUZw!JUwx!Arx#O!JU#O#P!B[#P#Q!Gr#Q;'S!JU;'S;=`!Jv<%lO!JU8r!JyP;=`<%l!JU8r!KPP;=`<%l!Gr>^!KZ^$i&j(Y!bOY!KSYZ&cZw!KSwx!CWx!^!KS!^!_!JU!_#O!KS#O#P!DR#P#Q!<n#Q#o!KS#o#p!JU#p;'S!KS;'S;=`!LV<%lO!KS>^!LYP;=`<%l!KS>^!L`P;=`<%l!<n=l!Ll`$i&j(Vp!X7`OY!LcYZ&cZr!Lcrs!=ys!P!Lc!P!Q!Mn!Q!^!Lc!^!_# o!_!}!Lc!}#O#%P#O#P!Dy#P#o!Lc#o#p# o#p;'S!Lc;'S;=`#&Y<%lO!Lc=l!Mwl$i&j(Vp!X7`OY(rYZ&cZr(rrs&cs!^(r!^!_)r!_#O(r#O#P&c#P#W(r#W#X!Mn#X#Z(r#Z#[!Mn#[#](r#]#^!Mn#^#a(r#a#b!Mn#b#g(r#g#h!Mn#h#i(r#i#j!Mn#j#k!Mn#k#m(r#m#n!Mn#n#o(r#o#p)r#p;'S(r;'S;=`*a<%lO(r8Q# vZ(Vp!X7`OY# oZr# ors!@cs!P# o!P!Q#!i!Q!}# o!}#O#$R#O#P!Bq#P;'S# o;'S;=`#$y<%lO# o8Q#!pe(Vp!X7`OY)rZr)rs#O)r#P#W)r#W#X#!i#X#Z)r#Z#[#!i#[#])r#]#^#!i#^#a)r#a#b#!i#b#g)r#g#h#!i#h#i)r#i#j#!i#j#k#!i#k#m)r#m#n#!i#n;'S)r;'S;=`*Z<%lO)r8Q#$WX(VpOY#$RZr#$Rrs!Ars#O#$R#O#P!B[#P#Q# o#Q;'S#$R;'S;=`#$s<%lO#$R8Q#$vP;=`<%l#$R8Q#$|P;=`<%l# o=l#%W^$i&j(VpOY#%PYZ&cZr#%Prs!CWs!^#%P!^!_#$R!_#O#%P#O#P!DR#P#Q!Lc#Q#o#%P#o#p#$R#p;'S#%P;'S;=`#&S<%lO#%P=l#&VP;=`<%l#%P=l#&]P;=`<%l!Lc?O#&kn$i&j(Vp(Y!b!X7`OY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#W%Z#W#X#&`#X#Z%Z#Z#[#&`#[#]%Z#]#^#&`#^#a%Z#a#b#&`#b#g%Z#g#h#&`#h#i%Z#i#j#&`#j#k#&`#k#m%Z#m#n#&`#n#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z9d#(r](Vp(Y!b!X7`OY#(iZr#(irs!Grsw#(iwx# ox!P#(i!P!Q#)k!Q!}#(i!}#O#+`#O#P!Bq#P;'S#(i;'S;=`#,`<%lO#(i9d#)th(Vp(Y!b!X7`OY*gZr*grs'}sw*gwx)rx#O*g#P#W*g#W#X#)k#X#Z*g#Z#[#)k#[#]*g#]#^#)k#^#a*g#a#b#)k#b#g*g#g#h#)k#h#i*g#i#j#)k#j#k#)k#k#m*g#m#n#)k#n;'S*g;'S;=`+Z<%lO*g9d#+gZ(Vp(Y!bOY#+`Zr#+`rs!JUsw#+`wx#$Rx#O#+`#O#P!B[#P#Q#(i#Q;'S#+`;'S;=`#,Y<%lO#+`9d#,]P;=`<%l#+`9d#,cP;=`<%l#(i?O#,o`$i&j(Vp(Y!bOY#,fYZ&cZr#,frs!KSsw#,fwx#%Px!^#,f!^!_#+`!_#O#,f#O#P!DR#P#Q!;Z#Q#o#,f#o#p#+`#p;'S#,f;'S;=`#-q<%lO#,f?O#-tP;=`<%l#,f?O#-zP;=`<%l!;Z07[#.[b$i&j(Vp(Y!b'}0/l!X7`OY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcx!P!;Z!P!Q#&`!Q!^!;Z!^!_#(i!_!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;Z07[#/o_$i&j(Vp(Y!bT0/lOY#/dYZ&cZr#/drs#0nsw#/dwx#4Ox!^#/d!^!_#5}!_#O#/d#O#P#1p#P#o#/d#o#p#5}#p;'S#/d;'S;=`#6|<%lO#/d06j#0w]$i&j(Y!bT0/lOY#0nYZ&cZw#0nwx#1px!^#0n!^!_#3R!_#O#0n#O#P#1p#P#o#0n#o#p#3R#p;'S#0n;'S;=`#3x<%lO#0n05W#1wX$i&jT0/lOY#1pYZ&cZ!^#1p!^!_#2d!_#o#1p#o#p#2d#p;'S#1p;'S;=`#2{<%lO#1p0/l#2iST0/lOY#2dZ;'S#2d;'S;=`#2u<%lO#2d0/l#2xP;=`<%l#2d05W#3OP;=`<%l#1p01O#3YW(Y!bT0/lOY#3RZw#3Rwx#2dx#O#3R#O#P#2d#P;'S#3R;'S;=`#3r<%lO#3R01O#3uP;=`<%l#3R06j#3{P;=`<%l#0n05x#4X]$i&j(VpT0/lOY#4OYZ&cZr#4Ors#1ps!^#4O!^!_#5Q!_#O#4O#O#P#1p#P#o#4O#o#p#5Q#p;'S#4O;'S;=`#5w<%lO#4O00^#5XW(VpT0/lOY#5QZr#5Qrs#2ds#O#5Q#O#P#2d#P;'S#5Q;'S;=`#5q<%lO#5Q00^#5tP;=`<%l#5Q05x#5zP;=`<%l#4O01p#6WY(Vp(Y!bT0/lOY#5}Zr#5}rs#3Rsw#5}wx#5Qx#O#5}#O#P#2d#P;'S#5};'S;=`#6v<%lO#5}01p#6yP;=`<%l#5}07[#7PP;=`<%l#/d)3h#7ab$i&j$Q(Ch(Vp(Y!b!X7`OY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcx!P!;Z!P!Q#&`!Q!^!;Z!^!_#(i!_!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;ZAt#8vb$Z#t$i&j(Vp(Y!b!X7`OY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcx!P!;Z!P!Q#&`!Q!^!;Z!^!_#(i!_!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;Z'Ad#:Zp$i&j(Vp(Y!bs'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!O%Z!O!P!3Y!P!Q%Z!Q![#<_![!^%Z!^!_*g!_!g%Z!g!h!4|!h#O%Z#O#P&c#P#R%Z#R#S#<_#S#U%Z#U#V#?i#V#X%Z#X#Y!4|#Y#b%Z#b#c#>_#c#d#Bq#d#l%Z#l#m#Es#m#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#<jk$i&j(Vp(Y!bs'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!O%Z!O!P!3Y!P!Q%Z!Q![#<_![!^%Z!^!_*g!_!g%Z!g!h!4|!h#O%Z#O#P&c#P#R%Z#R#S#<_#S#X%Z#X#Y!4|#Y#b%Z#b#c#>_#c#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#>j_$i&j(Vp(Y!bs'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#?rd$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q!R#AQ!R!S#AQ!S!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S#AQ#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#A]f$i&j(Vp(Y!bs'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q!R#AQ!R!S#AQ!S!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S#AQ#S#b%Z#b#c#>_#c#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#Bzc$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q!Y#DV!Y!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S#DV#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#Dbe$i&j(Vp(Y!bs'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q!Y#DV!Y!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S#DV#S#b%Z#b#c#>_#c#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#E|g$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![#Ge![!^%Z!^!_*g!_!c%Z!c!i#Ge!i#O%Z#O#P&c#P#R%Z#R#S#Ge#S#T%Z#T#Z#Ge#Z#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#Gpi$i&j(Vp(Y!bs'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![#Ge![!^%Z!^!_*g!_!c%Z!c!i#Ge!i#O%Z#O#P&c#P#R%Z#R#S#Ge#S#T%Z#T#Z#Ge#Z#b%Z#b#c#>_#c#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z*)x#Il_!g$b$i&j$O)Lv(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z)[#Jv_al$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z04f#LS^h#)`#R-<U(Vp(Y!b$n7`OY*gZr*grs'}sw*gwx)rx!P*g!P!Q#MO!Q!^*g!^!_#Mt!_!`$ f!`#O*g#P;'S*g;'S;=`+Z<%lO*g(n#MXX$k&j(Vp(Y!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g(El#M}Z#r(Ch(Vp(Y!bOY*gZr*grs'}sw*gwx)rx!_*g!_!`#Np!`#O*g#P;'S*g;'S;=`+Z<%lO*g(El#NyX$Q(Ch(Vp(Y!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g(El$ oX#s(Ch(Vp(Y!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g*)x$!ga#`*!Y$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`0z!`!a$#l!a#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(K[$#w_#k(Cl$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z*)x$%Vag!*r#s(Ch$f#|$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`$&[!`!a$'f!a#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$&g_#s(Ch$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$'qa#r(Ch$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`!a$(v!a#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$)R`#r(Ch$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(Kd$*`a(q(Ct$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!a%Z!a!b$+e!b#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$+p`$i&j#{(Ch(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z%#`$,}_!|$Ip$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z04f$.X_!S0,v$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(n$/]Z$i&jO!^$0O!^!_$0f!_#i$0O#i#j$0k#j#l$0O#l#m$2^#m#o$0O#o#p$0f#p;'S$0O;'S;=`$4i<%lO$0O(n$0VT_#S$i&jO!^&c!_#o&c#p;'S&c;'S;=`&w<%lO&c#S$0kO_#S(n$0p[$i&jO!Q&c!Q![$1f![!^&c!_!c&c!c!i$1f!i#T&c#T#Z$1f#Z#o&c#o#p$3|#p;'S&c;'S;=`&w<%lO&c(n$1kZ$i&jO!Q&c!Q![$2^![!^&c!_!c&c!c!i$2^!i#T&c#T#Z$2^#Z#o&c#p;'S&c;'S;=`&w<%lO&c(n$2cZ$i&jO!Q&c!Q![$3U![!^&c!_!c&c!c!i$3U!i#T&c#T#Z$3U#Z#o&c#p;'S&c;'S;=`&w<%lO&c(n$3ZZ$i&jO!Q&c!Q![$0O![!^&c!_!c&c!c!i$0O!i#T&c#T#Z$0O#Z#o&c#p;'S&c;'S;=`&w<%lO&c#S$4PR!Q![$4Y!c!i$4Y#T#Z$4Y#S$4]S!Q![$4Y!c!i$4Y#T#Z$4Y#q#r$0f(n$4lP;=`<%l$0O#1[$4z_!Y#)l$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$6U`#x(Ch$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z+;p$7c_$i&j(Vp(Y!b(`+4QOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z07[$8qk$i&j(Vp(Y!b(S,2j$_#t(d$I[OY%ZYZ&cZr%Zrs&}st%Ztu$8buw%Zwx(rx}%Z}!O$:f!O!Q%Z!Q![$8b![!^%Z!^!_*g!_!c%Z!c!}$8b!}#O%Z#O#P&c#P#R%Z#R#S$8b#S#T%Z#T#o$8b#o#p*g#p$g%Z$g;'S$8b;'S;=`$<l<%lO$8b+d$:qk$i&j(Vp(Y!b$_#tOY%ZYZ&cZr%Zrs&}st%Ztu$:fuw%Zwx(rx}%Z}!O$:f!O!Q%Z!Q![$:f![!^%Z!^!_*g!_!c%Z!c!}$:f!}#O%Z#O#P&c#P#R%Z#R#S$:f#S#T%Z#T#o$:f#o#p*g#p$g%Z$g;'S$:f;'S;=`$<f<%lO$:f+d$<iP;=`<%l$:f07[$<oP;=`<%l$8b#Jf$<{X!_#Hb(Vp(Y!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g,#x$=sa(x+JY$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p#q$+e#q;'S%Z;'S;=`+a<%lO%Z)>v$?V_!^(CdvBr$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z?O$@a_!q7`$i&j(Vp(Y!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z07[$Aq|$i&j(Vp(Y!b'{0/l$]#t(S,2j(d$I[OX%ZXY+gYZ&cZ[+g[p%Zpq+gqr%Zrs&}st%ZtuEruw%Zwx(rx}%Z}!OGv!O!Q%Z!Q![Er![!^%Z!^!_*g!_!c%Z!c!}Er!}#O%Z#O#P&c#P#R%Z#R#SEr#S#T%Z#T#oEr#o#p*g#p$f%Z$f$g+g$g#BYEr#BY#BZ$A`#BZ$ISEr$IS$I_$A`$I_$JTEr$JT$JU$A`$JU$KVEr$KV$KW$A`$KW&FUEr&FU&FV$A`&FV;'SEr;'S;=`I|<%l?HTEr?HT?HU$A`?HUOEr07[$D|k$i&j(Vp(Y!b'|0/l$]#t(S,2j(d$I[OY%ZYZ&cZr%Zrs&}st%ZtuEruw%Zwx(rx}%Z}!OGv!O!Q%Z!Q![Er![!^%Z!^!_*g!_!c%Z!c!}Er!}#O%Z#O#P&c#P#R%Z#R#SEr#S#T%Z#T#oEr#o#p*g#p$g%Z$g;'SEr;'S;=`I|<%lOEr",tokenizers:[Ize,Lze,Bze,Qze,2,3,4,5,6,7,8,9,10,11,12,13,14,Nze,new O9("$S~RRtu[#O#Pg#S#T#|~_P#o#pb~gOx~~jVO#i!P#i#j!U#j#l!P#l#m!q#m;'S!P;'S;=`#v<%lO!P~!UO!U~~!XS!Q![!e!c!i!e#T#Z!e#o#p#Z~!hR!Q![!q!c!i!q#T#Z!q~!tR!Q![!}!c!i!}#T#Z!}~#QR!Q![!P!c!i!P#T#Z!P~#^R!Q![#g!c!i#g#T#Z#g~#jS!Q![#g!c!i#g#T#Z#g#q#r!P~#yP;=`<%l!P~$RO(b~~",141,339),new O9("j~RQYZXz{^~^O(P~~aP!P!Qd~iO(Q~~",25,322)],topRules:{Script:[0,7],SingleExpression:[1,275],SingleClassItem:[2,276]},dialects:{jsx:0,ts:15098},dynamicPrecedences:{80:1,82:1,94:1,169:1,199:1},specialized:[{term:326,get:e=>qze[e]||-1},{term:342,get:e=>Vze[e]||-1},{term:95,get:e=>Uze[e]||-1}],tokenPrec:15124});var xb={},If={},h2={},Sb={},QP={},aY;function jze(){return aY||(aY=1,Object.defineProperty(QP,"__esModule",{value:!0})),QP}var p2={},m2={},uY;function M1(){var t,n,r,i,a,Jge,l;if(uY)return m2;uY=1,Object.defineProperty(m2,"__esModule",{value:!0}),m2.CronField=void 0;let e=(l=class{constructor(s,c={rawValue:""}){Yi(this,a);Yi(this,t,!1);Yi(this,n,!1);Yi(this,r,!1);Yi(this,i,[]);P_(this,"options",{rawValue:""});if(!Array.isArray(s))throw new Error(`${this.constructor.name} Validation error, values is not an array`);if(!(s.length>0))throw new Error(`${this.constructor.name} Validation error, values contains no values`);this.options={...c,rawValue:c.rawValue??""},wn(this,i,s.sort(l.sorter)),wn(this,r,this.options.wildcard!==void 0?this.options.wildcard:Zn(this,a,Jge).call(this)),wn(this,t,this.options.rawValue.includes("L")),wn(this,n,this.options.rawValue.includes("?"))}static get min(){throw new Error("min must be overridden")}static get max(){throw new Error("max must be overridden")}static get chars(){return Object.freeze([])}static get validChars(){return/^[?,*\dH/-]+$|^.*H\(\d+-\d+\)\/\d+.*$|^.*H\(\d+-\d+\).*$|^.*H\/\d+.*$/}static get constraints(){return{min:this.min,max:this.max,chars:this.chars,validChars:this.validChars}}get min(){return this.constructor.min}get max(){return this.constructor.max}get chars(){return this.constructor.chars}get hasLastChar(){return Ot(this,t)}get hasQuestionMarkChar(){return Ot(this,n)}get isWildcard(){return Ot(this,r)}get values(){return Ot(this,i)}static sorter(s,c){const d=typeof s=="number",h=typeof c=="number";return d&&h?s-c:!d&&!h?s.localeCompare(c):d?-1:1}serialize(){return{wildcard:Ot(this,r),values:Ot(this,i)}}validate(){let s;const c=this.chars.length>0?` or chars ${this.chars.join("")}`:"",d=v=>S=>new RegExp(`^\\d{0,2}${S}$`).test(v),h=v=>(s=v,typeof v=="number"?v>=this.min&&v<=this.max:this.chars.some(d(v)));if(!Ot(this,i).every(h))throw new Error(`${this.constructor.name} Validation error, got value ${s} expected range ${this.min}-${this.max}${c}`);const y=Ot(this,i).find((v,S)=>Ot(this,i).indexOf(v)!==S);if(y)throw new Error(`${this.constructor.name} Validation error, duplicate values found: ${y}`)}},t=new WeakMap,n=new WeakMap,r=new WeakMap,i=new WeakMap,a=new WeakSet,Jge=function(){return this.options.rawValue.length>0?["*","?"].includes(this.options.rawValue):Array.from({length:this.max-this.min+1},(s,c)=>c+this.min).every(s=>Ot(this,i).includes(s))},l);return m2.CronField=e,m2}var lY;function Zze(){if(lY)return p2;lY=1,Object.defineProperty(p2,"__esModule",{value:!0}),p2.CronDayOfMonth=void 0;const e=M1(),t=1,n=31,r=Object.freeze(["L"]);let i=class extends e.CronField{static get min(){return t}static get max(){return n}static get chars(){return r}static get validChars(){return/^[?,*\dLH/-]+$|^.*H\(\d+-\d+\)\/\d+.*$|^.*H\(\d+-\d+\).*$|^.*H\/\d+.*$/}constructor(u,l){super(u,l),this.validate()}get values(){return super.values}};return p2.CronDayOfMonth=i,p2}var g2={},oY;function Hze(){if(oY)return g2;oY=1,Object.defineProperty(g2,"__esModule",{value:!0}),g2.CronDayOfWeek=void 0;const e=M1(),t=0,n=7,r=Object.freeze(["L"]);let i=class extends e.CronField{static get min(){return t}static get max(){return n}static get chars(){return r}static get validChars(){return/^[?,*\dLH#/-]+$|^.*H\(\d+-\d+\)\/\d+.*$|^.*H\(\d+-\d+\).*$|^.*H\/\d+.*$/}constructor(u,l){super(u,l),this.validate()}get values(){return super.values}get nthDay(){return this.options.nthDayOfWeek??0}};return g2.CronDayOfWeek=i,g2}var v2={},sY;function Xze(){if(sY)return v2;sY=1,Object.defineProperty(v2,"__esModule",{value:!0}),v2.CronHour=void 0;const e=M1(),t=0,n=23,r=Object.freeze([]);let i=class extends e.CronField{static get min(){return t}static get max(){return n}static get chars(){return r}constructor(u,l){super(u,l),this.validate()}get values(){return super.values}};return v2.CronHour=i,v2}var y2={},cY;function Yze(){if(cY)return y2;cY=1,Object.defineProperty(y2,"__esModule",{value:!0}),y2.CronMinute=void 0;const e=M1(),t=0,n=59,r=Object.freeze([]);let i=class extends e.CronField{static get min(){return t}static get max(){return n}static get chars(){return r}constructor(u,l){super(u,l),this.validate()}get values(){return super.values}};return y2.CronMinute=i,y2}var b2={},zP={},Yc={},fY;function Gze(){if(fY)return Yc;fY=1,Object.defineProperty(Yc,"__esModule",{value:!0});class e extends Error{}class t extends e{constructor(L){super(`Invalid DateTime: ${L.toMessage()}`)}}class n extends e{constructor(L){super(`Invalid Interval: ${L.toMessage()}`)}}class r extends e{constructor(L){super(`Invalid Duration: ${L.toMessage()}`)}}class i extends e{}class a extends e{constructor(L){super(`Invalid unit ${L}`)}}class u extends e{}class l extends e{constructor(){super("Zone is an abstract class")}}const o="numeric",s="short",c="long",d={year:o,month:o,day:o},h={year:o,month:s,day:o},m={year:o,month:s,day:o,weekday:s},y={year:o,month:c,day:o},v={year:o,month:c,day:o,weekday:c},S={hour:o,minute:o},_={hour:o,minute:o,second:o},E={hour:o,minute:o,second:o,timeZoneName:s},$={hour:o,minute:o,second:o,timeZoneName:c},k={hour:o,minute:o,hourCycle:"h23"},A={hour:o,minute:o,second:o,hourCycle:"h23"},w={hour:o,minute:o,second:o,hourCycle:"h23",timeZoneName:s},R={hour:o,minute:o,second:o,hourCycle:"h23",timeZoneName:c},F={year:o,month:o,day:o,hour:o,minute:o},I={year:o,month:o,day:o,hour:o,minute:o,second:o},U={year:o,month:s,day:o,hour:o,minute:o},q={year:o,month:s,day:o,hour:o,minute:o,second:o},V={year:o,month:s,day:o,weekday:s,hour:o,minute:o},N={year:o,month:c,day:o,hour:o,minute:o,timeZoneName:s},j={year:o,month:c,day:o,hour:o,minute:o,second:o,timeZoneName:s},W={year:o,month:c,day:o,weekday:c,hour:o,minute:o,timeZoneName:c},K={year:o,month:c,day:o,weekday:c,hour:o,minute:o,second:o,timeZoneName:c};class J{get type(){throw new l}get name(){throw new l}get ianaName(){return this.name}get isUniversal(){throw new l}offsetName(L,re){throw new l}formatOffset(L,re){throw new l}offset(L){throw new l}equals(L){throw new l}get isValid(){throw new l}}let te=null;class H extends J{static get instance(){return te===null&&(te=new H),te}get type(){return"system"}get name(){return new Intl.DateTimeFormat().resolvedOptions().timeZone}get isUniversal(){return!1}offsetName(L,{format:re,locale:fe}){return vc(L,re,fe)}formatOffset(L,re){return Cs(this.offset(L),re)}offset(L){return-new Date(L).getTimezoneOffset()}equals(L){return L.type==="system"}get isValid(){return!0}}const M=new Map;function G(ue){let L=M.get(ue);return L===void 0&&(L=new Intl.DateTimeFormat("en-US",{hour12:!1,timeZone:ue,year:"numeric",month:"2-digit",day:"2-digit",hour:"2-digit",minute:"2-digit",second:"2-digit",era:"short"}),M.set(ue,L)),L}const X={year:0,month:1,day:2,era:3,hour:4,minute:5,second:6};function B(ue,L){const re=ue.format(L).replace(/\u200E/g,""),fe=/(\d+)\/(\d+)\/(\d+) (AD|BC),? (\d+):(\d+):(\d+)/.exec(re),[,xe,ke,Qe,Be,Ze,Rt,Kt]=fe;return[Qe,xe,ke,Be,Ze,Rt,Kt]}function ne(ue,L){const re=ue.formatToParts(L),fe=[];for(let xe=0;xe<re.length;xe++){const{type:ke,value:Qe}=re[xe],Be=X[ke];ke==="era"?fe[Be]=Qe:wt(Be)||(fe[Be]=parseInt(Qe,10))}return fe}const oe=new Map;class ie extends J{static create(L){let re=oe.get(L);return re===void 0&&oe.set(L,re=new ie(L)),re}static resetCache(){oe.clear(),M.clear()}static isValidSpecifier(L){return this.isValidZone(L)}static isValidZone(L){if(!L)return!1;try{return new Intl.DateTimeFormat("en-US",{timeZone:L}).format(),!0}catch{return!1}}constructor(L){super(),this.zoneName=L,this.valid=ie.isValidZone(L)}get type(){return"iana"}get name(){return this.zoneName}get isUniversal(){return!1}offsetName(L,{format:re,locale:fe}){return vc(L,re,fe,this.name)}formatOffset(L,re){return Cs(this.offset(L),re)}offset(L){if(!this.valid)return NaN;const re=new Date(L);if(isNaN(re))return NaN;const fe=G(this.name);let[xe,ke,Qe,Be,Ze,Rt,Kt]=fe.formatToParts?ne(fe,re):B(fe,re);Be==="BC"&&(xe=-Math.abs(xe)+1);const ln=Ya({year:xe,month:ke,day:Qe,hour:Ze===24?0:Ze,minute:Rt,second:Kt,millisecond:0});let mn=+re;const Ri=mn%1e3;return mn-=Ri>=0?Ri:1e3+Ri,(ln-mn)/(60*1e3)}equals(L){return L.type==="iana"&&L.name===this.name}get isValid(){return this.valid}}let de={};function ge(ue,L={}){const re=JSON.stringify([ue,L]);let fe=de[re];return fe||(fe=new Intl.ListFormat(ue,L),de[re]=fe),fe}const ye=new Map;function me(ue,L={}){const re=JSON.stringify([ue,L]);let fe=ye.get(re);return fe===void 0&&(fe=new Intl.DateTimeFormat(ue,L),ye.set(re,fe)),fe}const Ee=new Map;function se(ue,L={}){const re=JSON.stringify([ue,L]);let fe=Ee.get(re);return fe===void 0&&(fe=new Intl.NumberFormat(ue,L),Ee.set(re,fe)),fe}const pe=new Map;function ve(ue,L={}){const{base:re,...fe}=L,xe=JSON.stringify([ue,fe]);let ke=pe.get(xe);return ke===void 0&&(ke=new Intl.RelativeTimeFormat(ue,L),pe.set(xe,ke)),ke}let we=null;function Me(){return we||(we=new Intl.DateTimeFormat().resolvedOptions().locale,we)}const Re=new Map;function et(ue){let L=Re.get(ue);return L===void 0&&(L=new Intl.DateTimeFormat(ue).resolvedOptions(),Re.set(ue,L)),L}const tt=new Map;function Oe(ue){let L=tt.get(ue);if(!L){const re=new Intl.Locale(ue);L="getWeekInfo"in re?re.getWeekInfo():re.weekInfo,"minimalDays"in L||(L={...Mt,...L}),tt.set(ue,L)}return L}function $e(ue){const L=ue.indexOf("-x-");L!==-1&&(ue=ue.substring(0,L));const re=ue.indexOf("-u-");if(re===-1)return[ue];{let fe,xe;try{fe=me(ue).resolvedOptions(),xe=ue}catch{const Ze=ue.substring(0,re);fe=me(Ze).resolvedOptions(),xe=Ze}const{numberingSystem:ke,calendar:Qe}=fe;return[xe,ke,Qe]}}function Ae(ue,L,re){return(re||L)&&(ue.includes("-u-")||(ue+="-u"),re&&(ue+=`-ca-${re}`),L&&(ue+=`-nu-${L}`)),ue}function He(ue){const L=[];for(let re=1;re<=12;re++){const fe=dn.utc(2009,re,1);L.push(ue(fe))}return L}function je(ue){const L=[];for(let re=1;re<=7;re++){const fe=dn.utc(2016,11,13+re);L.push(ue(fe))}return L}function Ye(ue,L,re,fe){const xe=ue.listingMode();return xe==="error"?null:xe==="en"?re(L):fe(L)}function Ne(ue){return ue.numberingSystem&&ue.numberingSystem!=="latn"?!1:ue.numberingSystem==="latn"||!ue.locale||ue.locale.startsWith("en")||et(ue.locale).numberingSystem==="latn"}class ze{constructor(L,re,fe){this.padTo=fe.padTo||0,this.floor=fe.floor||!1;const{padTo:xe,floor:ke,...Qe}=fe;if(!re||Object.keys(Qe).length>0){const Be={useGrouping:!1,...fe};fe.padTo>0&&(Be.minimumIntegerDigits=fe.padTo),this.inf=se(L,Be)}}format(L){if(this.inf){const re=this.floor?Math.floor(L):L;return this.inf.format(re)}else{const re=this.floor?Math.floor(L):Ds(L,3);return tr(re,this.padTo)}}}class at{constructor(L,re,fe){this.opts=fe,this.originalZone=void 0;let xe;if(this.opts.timeZone)this.dt=L;else if(L.zone.type==="fixed"){const Qe=-1*(L.offset/60),Be=Qe>=0?`Etc/GMT+${Qe}`:`Etc/GMT${Qe}`;L.offset!==0&&ie.create(Be).valid?(xe=Be,this.dt=L):(xe="UTC",this.dt=L.offset===0?L:L.setZone("UTC").plus({minutes:L.offset}),this.originalZone=L.zone)}else L.zone.type==="system"?this.dt=L:L.zone.type==="iana"?(this.dt=L,xe=L.zone.name):(xe="UTC",this.dt=L.setZone("UTC").plus({minutes:L.offset}),this.originalZone=L.zone);const ke={...this.opts};ke.timeZone=ke.timeZone||xe,this.dtf=me(re,ke)}format(){return this.originalZone?this.formatToParts().map(({value:L})=>L).join(""):this.dtf.format(this.dt.toJSDate())}formatToParts(){const L=this.dtf.formatToParts(this.dt.toJSDate());return this.originalZone?L.map(re=>{if(re.type==="timeZoneName"){const fe=this.originalZone.offsetName(this.dt.ts,{locale:this.dt.locale,format:this.opts.timeZoneName});return{...re,value:fe}}else return re}):L}resolvedOptions(){return this.dtf.resolvedOptions()}}class mt{constructor(L,re,fe){this.opts={style:"long",...fe},!re&&Xa()&&(this.rtf=ve(L,fe))}format(L,re){return this.rtf?this.rtf.format(L,re):Oc(re,L,this.opts.numeric,this.opts.style!=="long")}formatToParts(L,re){return this.rtf?this.rtf.formatToParts(L,re):[]}}const Mt={firstDay:1,minimalDays:4,weekend:[6,7]};class bt{static fromOpts(L){return bt.create(L.locale,L.numberingSystem,L.outputCalendar,L.weekSettings,L.defaultToEN)}static create(L,re,fe,xe,ke=!1){const Qe=L||ar.defaultLocale,Be=Qe||(ke?"en-US":Me()),Ze=re||ar.defaultNumberingSystem,Rt=fe||ar.defaultOutputCalendar,Kt=ur(xe)||ar.defaultWeekSettings;return new bt(Be,Ze,Rt,Kt,Qe)}static resetCache(){we=null,ye.clear(),Ee.clear(),pe.clear(),Re.clear(),tt.clear()}static fromObject({locale:L,numberingSystem:re,outputCalendar:fe,weekSettings:xe}={}){return bt.create(L,re,fe,xe)}constructor(L,re,fe,xe,ke){const[Qe,Be,Ze]=$e(L);this.locale=Qe,this.numberingSystem=re||Be||null,this.outputCalendar=fe||Ze||null,this.weekSettings=xe,this.intl=Ae(this.locale,this.numberingSystem,this.outputCalendar),this.weekdaysCache={format:{},standalone:{}},this.monthsCache={format:{},standalone:{}},this.meridiemCache=null,this.eraCache={},this.specifiedLocale=ke,this.fastNumbersCached=null}get fastNumbers(){return this.fastNumbersCached==null&&(this.fastNumbersCached=Ne(this)),this.fastNumbersCached}listingMode(){const L=this.isEnglish(),re=(this.numberingSystem===null||this.numberingSystem==="latn")&&(this.outputCalendar===null||this.outputCalendar==="gregory");return L&&re?"en":"intl"}clone(L){return!L||Object.getOwnPropertyNames(L).length===0?this:bt.create(L.locale||this.specifiedLocale,L.numberingSystem||this.numberingSystem,L.outputCalendar||this.outputCalendar,ur(L.weekSettings)||this.weekSettings,L.defaultToEN||!1)}redefaultToEN(L={}){return this.clone({...L,defaultToEN:!0})}redefaultToSystem(L={}){return this.clone({...L,defaultToEN:!1})}months(L,re=!1){return Ye(this,L,Op,()=>{const fe=this.intl==="ja"||this.intl.startsWith("ja-");re&=!fe;const xe=re?{month:L,day:"numeric"}:{month:L},ke=re?"format":"standalone";if(!this.monthsCache[ke][L]){const Qe=fe?Be=>this.dtFormatter(Be,xe).format():Be=>this.extract(Be,xe,"month");this.monthsCache[ke][L]=He(Qe)}return this.monthsCache[ke][L]})}weekdays(L,re=!1){return Ye(this,L,fh,()=>{const fe=re?{weekday:L,year:"numeric",month:"long",day:"numeric"}:{weekday:L},xe=re?"format":"standalone";return this.weekdaysCache[xe][L]||(this.weekdaysCache[xe][L]=je(ke=>this.extract(ke,fe,"weekday"))),this.weekdaysCache[xe][L]})}meridiems(){return Ye(this,void 0,()=>Rs,()=>{if(!this.meridiemCache){const L={hour:"numeric",hourCycle:"h12"};this.meridiemCache=[dn.utc(2016,11,13,9),dn.utc(2016,11,13,19)].map(re=>this.extract(re,L,"dayperiod"))}return this.meridiemCache})}eras(L){return Ye(this,L,Sp,()=>{const re={era:L};return this.eraCache[L]||(this.eraCache[L]=[dn.utc(-40,1,1),dn.utc(2017,1,1)].map(fe=>this.extract(fe,re,"era"))),this.eraCache[L]})}extract(L,re,fe){const xe=this.dtFormatter(L,re),ke=xe.formatToParts(),Qe=ke.find(Be=>Be.type.toLowerCase()===fe);return Qe?Qe.value:null}numberFormatter(L={}){return new ze(this.intl,L.forceSimple||this.fastNumbers,L)}dtFormatter(L,re={}){return new at(L,this.intl,re)}relFormatter(L={}){return new mt(this.intl,this.isEnglish(),L)}listFormatter(L={}){return ge(this.intl,L)}isEnglish(){return this.locale==="en"||this.locale.toLowerCase()==="en-us"||et(this.intl).locale.startsWith("en-us")}getWeekSettings(){return this.weekSettings?this.weekSettings:Au()?Oe(this.locale):Mt}getStartOfWeek(){return this.getWeekSettings().firstDay}getMinDaysInFirstWeek(){return this.getWeekSettings().minimalDays}getWeekendDays(){return this.getWeekSettings().weekend}equals(L){return this.locale===L.locale&&this.numberingSystem===L.numberingSystem&&this.outputCalendar===L.outputCalendar}toString(){return`Locale(${this.locale}, ${this.numberingSystem}, ${this.outputCalendar})`}}let zt=null;class qt extends J{static get utcInstance(){return zt===null&&(zt=new qt(0)),zt}static instance(L){return L===0?qt.utcInstance:new qt(L)}static parseSpecifier(L){if(L){const re=L.match(/^utc(?:([+-]\d{1,2})(?::(\d{2}))?)?$/i);if(re)return new qt(yc(re[1],re[2]))}return null}constructor(L){super(),this.fixed=L}get type(){return"fixed"}get name(){return this.fixed===0?"UTC":`UTC${Cs(this.fixed,"narrow")}`}get ianaName(){return this.fixed===0?"Etc/UTC":`Etc/GMT${Cs(-this.fixed,"narrow")}`}offsetName(){return this.name}formatOffset(L,re){return Cs(this.fixed,re)}get isUniversal(){return!0}offset(){return this.fixed}equals(L){return L.type==="fixed"&&L.fixed===this.fixed}get isValid(){return!0}}class Xt extends J{constructor(L){super(),this.zoneName=L}get type(){return"invalid"}get name(){return this.zoneName}get isUniversal(){return!1}offsetName(){return null}formatOffset(){return""}offset(){return NaN}equals(){return!1}get isValid(){return!1}}function nn(ue,L){if(wt(ue)||ue===null)return L;if(ue instanceof J)return ue;if(yi(ue)){const re=ue.toLowerCase();return re==="default"?L:re==="local"||re==="system"?H.instance:re==="utc"||re==="gmt"?qt.utcInstance:qt.parseSpecifier(re)||ie.create(ue)}else return Kr(ue)?qt.instance(ue):typeof ue=="object"&&"offset"in ue&&typeof ue.offset=="function"?ue:new Xt(ue)}const Fe={arab:"[٠-٩]",arabext:"[۰-۹]",bali:"[᭐-᭙]",beng:"[০-৯]",deva:"[०-९]",fullwide:"[０-９]",gujr:"[૦-૯]",hanidec:"[〇|一|二|三|四|五|六|七|八|九]",khmr:"[០-៩]",knda:"[೦-೯]",laoo:"[໐-໙]",limb:"[᥆-᥏]",mlym:"[൦-൯]",mong:"[᠐-᠙]",mymr:"[၀-၉]",orya:"[୦-୯]",tamldec:"[௦-௯]",telu:"[౦-౯]",thai:"[๐-๙]",tibt:"[༠-༩]",latn:"\\d"},ht={arab:[1632,1641],arabext:[1776,1785],bali:[6992,7001],beng:[2534,2543],deva:[2406,2415],fullwide:[65296,65303],gujr:[2790,2799],khmr:[6112,6121],knda:[3302,3311],laoo:[3792,3801],limb:[6470,6479],mlym:[3430,3439],mong:[6160,6169],mymr:[4160,4169],orya:[2918,2927],tamldec:[3046,3055],telu:[3174,3183],thai:[3664,3673],tibt:[3872,3881]},kt=Fe.hanidec.replace(/[\[|\]]/g,"").split("");function Qt(ue){let L=parseInt(ue,10);if(isNaN(L)){L="";for(let re=0;re<ue.length;re++){const fe=ue.charCodeAt(re);if(ue[re].search(Fe.hanidec)!==-1)L+=kt.indexOf(ue[re]);else for(const xe in ht){const[ke,Qe]=ht[xe];fe>=ke&&fe<=Qe&&(L+=fe-ke)}}return parseInt(L,10)}else return L}const en=new Map;function In(){en.clear()}function rn({numberingSystem:ue},L=""){const re=ue||"latn";let fe=en.get(re);fe===void 0&&(fe=new Map,en.set(re,fe));let xe=fe.get(L);return xe===void 0&&(xe=new RegExp(`${Fe[re]}${L}`),fe.set(L,xe)),xe}let Xn=()=>Date.now(),br="system",Or=null,Cr=null,wr=null,Gr=60,ji,ai=null;class ar{static get now(){return Xn}static set now(L){Xn=L}static set defaultZone(L){br=L}static get defaultZone(){return nn(br,H.instance)}static get defaultLocale(){return Or}static set defaultLocale(L){Or=L}static get defaultNumberingSystem(){return Cr}static set defaultNumberingSystem(L){Cr=L}static get defaultOutputCalendar(){return wr}static set defaultOutputCalendar(L){wr=L}static get defaultWeekSettings(){return ai}static set defaultWeekSettings(L){ai=ur(L)}static get twoDigitCutoffYear(){return Gr}static set twoDigitCutoffYear(L){Gr=L%100}static get throwOnInvalid(){return ji}static set throwOnInvalid(L){ji=L}static resetCaches(){bt.resetCache(),ie.resetCache(),dn.resetCache(),In()}}class Te{constructor(L,re){this.reason=L,this.explanation=re}toMessage(){return this.explanation?`${this.reason}: ${this.explanation}`:this.reason}}const Ve=[0,31,59,90,120,151,181,212,243,273,304,334],nt=[0,31,60,91,121,152,182,213,244,274,305,335];function gt(ue,L){return new Te("unit out of range",`you specified ${L} (of type ${typeof L}) as a ${ue}, which is invalid`)}function Ct(ue,L,re){const fe=new Date(Date.UTC(ue,L-1,re));ue<100&&ue>=0&&fe.setUTCFullYear(fe.getUTCFullYear()-1900);const xe=fe.getUTCDay();return xe===0?7:xe}function Vt(ue,L,re){return re+(qo(ue)?nt:Ve)[L-1]}function Yt(ue,L){const re=qo(ue)?nt:Ve,fe=re.findIndex(ke=>ke<L),xe=L-re[fe];return{month:fe+1,day:xe}}function Dt(ue,L){return(ue-L+7)%7+1}function Lt(ue,L=4,re=1){const{year:fe,month:xe,day:ke}=ue,Qe=Vt(fe,xe,ke),Be=Dt(Ct(fe,xe,ke),re);let Ze=Math.floor((Qe-Be+14-L)/7),Rt;return Ze<1?(Rt=fe-1,Ze=$s(Rt,L,re)):Ze>$s(fe,L,re)?(Rt=fe+1,Ze=1):Rt=fe,{weekYear:Rt,weekNumber:Ze,weekday:Be,...ws(ue)}}function an(ue,L=4,re=1){const{weekYear:fe,weekNumber:xe,weekday:ke}=ue,Qe=Dt(Ct(fe,1,L),re),Be=Il(fe);let Ze=xe*7+ke-Qe-7+L,Rt;Ze<1?(Rt=fe-1,Ze+=Il(Rt)):Ze>Be?(Rt=fe+1,Ze-=Il(fe)):Rt=fe;const{month:Kt,day:zn}=Yt(Rt,Ze);return{year:Rt,month:Kt,day:zn,...ws(ue)}}function _n(ue){const{year:L,month:re,day:fe}=ue,xe=Vt(L,re,fe);return{year:L,ordinal:xe,...ws(ue)}}function sn(ue){const{year:L,ordinal:re}=ue,{month:fe,day:xe}=Yt(L,re);return{year:L,month:fe,day:xe,...ws(ue)}}function Ln(ue,L){if(!wt(ue.localWeekday)||!wt(ue.localWeekNumber)||!wt(ue.localWeekYear)){if(!wt(ue.weekday)||!wt(ue.weekNumber)||!wt(ue.weekYear))throw new i("Cannot mix locale-based week fields with ISO-based week fields");return wt(ue.localWeekday)||(ue.weekday=ue.localWeekday),wt(ue.localWeekNumber)||(ue.weekNumber=ue.localWeekNumber),wt(ue.localWeekYear)||(ue.weekYear=ue.localWeekYear),delete ue.localWeekday,delete ue.localWeekNumber,delete ue.localWeekYear,{minDaysInFirstWeek:L.getMinDaysInFirstWeek(),startOfWeek:L.getStartOfWeek()}}else return{minDaysInFirstWeek:4,startOfWeek:1}}function er(ue,L=4,re=1){const fe=Zi(ue.weekYear),xe=Rr(ue.weekNumber,1,$s(ue.weekYear,L,re)),ke=Rr(ue.weekday,1,7);return fe?xe?ke?!1:gt("weekday",ue.weekday):gt("week",ue.weekNumber):gt("weekYear",ue.weekYear)}function Di(ue){const L=Zi(ue.year),re=Rr(ue.ordinal,1,Il(ue.year));return L?re?!1:gt("ordinal",ue.ordinal):gt("year",ue.year)}function $i(ue){const L=Zi(ue.year),re=Rr(ue.month,1,12),fe=Rr(ue.day,1,sh(ue.year,ue.month));return L?re?fe?!1:gt("day",ue.day):gt("month",ue.month):gt("year",ue.year)}function hr(ue){const{hour:L,minute:re,second:fe,millisecond:xe}=ue,ke=Rr(L,0,23)||L===24&&re===0&&fe===0&&xe===0,Qe=Rr(re,0,59),Be=Rr(fe,0,59),Ze=Rr(xe,0,999);return ke?Qe?Be?Ze?!1:gt("millisecond",xe):gt("second",fe):gt("minute",re):gt("hour",L)}function wt(ue){return typeof ue>"u"}function Kr(ue){return typeof ue=="number"}function Zi(ue){return typeof ue=="number"&&ue%1===0}function yi(ue){return typeof ue=="string"}function Ca(ue){return Object.prototype.toString.call(ue)==="[object Date]"}function Xa(){try{return typeof Intl<"u"&&!!Intl.RelativeTimeFormat}catch{return!1}}function Au(){try{return typeof Intl<"u"&&!!Intl.Locale&&("weekInfo"in Intl.Locale.prototype||"getWeekInfo"in Intl.Locale.prototype)}catch{return!1}}function Tu(ue){return Array.isArray(ue)?ue:[ue]}function Nl(ue,L,re){if(ue.length!==0)return ue.reduce((fe,xe)=>{const ke=[L(xe),xe];return fe&&re(fe[0],ke[0])===fe[0]?fe:ke},null)[1]}function gc(ue,L){return L.reduce((re,fe)=>(re[fe]=ue[fe],re),{})}function xl(ue,L){return Object.prototype.hasOwnProperty.call(ue,L)}function ur(ue){if(ue==null)return null;if(typeof ue!="object")throw new u("Week settings must be an object");if(!Rr(ue.firstDay,1,7)||!Rr(ue.minimalDays,1,7)||!Array.isArray(ue.weekend)||ue.weekend.some(L=>!Rr(L,1,7)))throw new u("Invalid week settings");return{firstDay:ue.firstDay,minimalDays:ue.minimalDays,weekend:Array.from(ue.weekend)}}function Rr(ue,L,re){return Zi(ue)&&ue>=L&&ue<=re}function Sl(ue,L){return ue-L*Math.floor(ue/L)}function tr(ue,L=2){const re=ue<0;let fe;return re?fe="-"+(""+-ue).padStart(L,"0"):fe=(""+ue).padStart(L,"0"),fe}function Bn(ue){if(!(wt(ue)||ue===null||ue===""))return parseInt(ue,10)}function ku(ue){if(!(wt(ue)||ue===null||ue===""))return parseFloat(ue)}function gf(ue){if(!(wt(ue)||ue===null||ue==="")){const L=parseFloat("0."+ue)*1e3;return Math.floor(L)}}function Ds(ue,L,re="round"){const fe=10**L;switch(re){case"expand":return ue>0?Math.ceil(ue*fe)/fe:Math.floor(ue*fe)/fe;case"trunc":return Math.trunc(ue*fe)/fe;case"round":return Math.round(ue*fe)/fe;case"floor":return Math.floor(ue*fe)/fe;case"ceil":return Math.ceil(ue*fe)/fe;default:throw new RangeError(`Value rounding ${re} is out of range`)}}function qo(ue){return ue%4===0&&(ue%100!==0||ue%400===0)}function Il(ue){return qo(ue)?366:365}function sh(ue,L){const re=Sl(L-1,12)+1,fe=ue+(L-re)/12;return re===2?qo(fe)?29:28:[31,null,31,30,31,30,31,31,30,31,30,31][re-1]}function Ya(ue){let L=Date.UTC(ue.year,ue.month-1,ue.day,ue.hour,ue.minute,ue.second,ue.millisecond);return ue.year<100&&ue.year>=0&&(L=new Date(L),L.setUTCFullYear(ue.year,ue.month-1,ue.day)),+L}function rd(ue,L,re){return-Dt(Ct(ue,1,L),re)+L-1}function $s(ue,L=4,re=1){const fe=rd(ue,L,re),xe=rd(ue+1,L,re);return(Il(ue)-fe+xe)/7}function id(ue){return ue>99?ue:ue>ar.twoDigitCutoffYear?1900+ue:2e3+ue}function vc(ue,L,re,fe=null){const xe=new Date(ue),ke={hourCycle:"h23",year:"numeric",month:"2-digit",day:"2-digit",hour:"2-digit",minute:"2-digit"};fe&&(ke.timeZone=fe);const Qe={timeZoneName:L,...ke},Be=new Intl.DateTimeFormat(re,Qe).formatToParts(xe).find(Ze=>Ze.type.toLowerCase()==="timezonename");return Be?Be.value:null}function yc(ue,L){let re=parseInt(ue,10);Number.isNaN(re)&&(re=0);const fe=parseInt(L,10)||0,xe=re<0||Object.is(re,-0)?-fe:fe;return re*60+xe}function Vo(ue){const L=Number(ue);if(typeof ue=="boolean"||ue===""||!Number.isFinite(L))throw new u(`Invalid unit value ${ue}`);return L}function Pu(ue,L){const re={};for(const fe in ue)if(xl(ue,fe)){const xe=ue[fe];if(xe==null)continue;re[L(fe)]=Vo(xe)}return re}function Cs(ue,L){const re=Math.trunc(Math.abs(ue/60)),fe=Math.trunc(Math.abs(ue%60)),xe=ue>=0?"+":"-";switch(L){case"short":return`${xe}${tr(re,2)}:${tr(fe,2)}`;case"narrow":return`${xe}${re}${fe>0?`:${fe}`:""}`;case"techie":return`${xe}${tr(re,2)}${tr(fe,2)}`;default:throw new RangeError(`Value format ${L} is out of range for property format`)}}function ws(ue){return gc(ue,["hour","minute","second","millisecond"])}const j0=["January","February","March","April","May","June","July","August","September","October","November","December"],vf=["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],Z0=["J","F","M","A","M","J","J","A","S","O","N","D"];function Op(ue){switch(ue){case"narrow":return[...Z0];case"short":return[...vf];case"long":return[...j0];case"numeric":return["1","2","3","4","5","6","7","8","9","10","11","12"];case"2-digit":return["01","02","03","04","05","06","07","08","09","10","11","12"];default:return null}}const yf=["Monday","Tuesday","Wednesday","Thursday","Friday","Saturday","Sunday"],ho=["Mon","Tue","Wed","Thu","Fri","Sat","Sun"],ch=["M","T","W","T","F","S","S"];function fh(ue){switch(ue){case"narrow":return[...ch];case"short":return[...ho];case"long":return[...yf];case"numeric":return["1","2","3","4","5","6","7"];default:return null}}const Rs=["AM","PM"],bc=["Before Christ","Anno Domini"],H0=["BC","AD"],xp=["B","A"];function Sp(ue){switch(ue){case"narrow":return[...xp];case"short":return[...H0];case"long":return[...bc];default:return null}}function As(ue){return Rs[ue.hour<12?0:1]}function Uo(ue,L){return fh(L)[ue.weekday-1]}function _p(ue,L){return Op(L)[ue.month-1]}function dh(ue,L){return Sp(L)[ue.year<0?0:1]}function Oc(ue,L,re="always",fe=!1){const xe={years:["year","yr."],quarters:["quarter","qtr."],months:["month","mo."],weeks:["week","wk."],days:["day","day","days"],hours:["hour","hr."],minutes:["minute","min."],seconds:["second","sec."]},ke=["hours","minutes","seconds"].indexOf(ue)===-1;if(re==="auto"&&ke){const zn=ue==="days";switch(L){case 1:return zn?"tomorrow":`next ${xe[ue][0]}`;case-1:return zn?"yesterday":`last ${xe[ue][0]}`;case 0:return zn?"today":`this ${xe[ue][0]}`}}const Qe=Object.is(L,-0)||L<0,Be=Math.abs(L),Ze=Be===1,Rt=xe[ue],Kt=fe?Ze?Rt[1]:Rt[2]||Rt[1]:Ze?xe[ue][0]:ue;return Qe?`${Be} ${Kt} ago`:`in ${Be} ${Kt}`}function hh(ue,L){let re="";for(const fe of ue)fe.literal?re+=fe.val:re+=L(fe.val);return re}const Mr={D:d,DD:h,DDD:y,DDDD:v,t:S,tt:_,ttt:E,tttt:$,T:k,TT:A,TTT:w,TTTT:R,f:F,ff:U,fff:N,ffff:W,F:I,FF:q,FFF:j,FFFF:K};class Qn{static create(L,re={}){return new Qn(L,re)}static parseFormat(L){let re=null,fe="",xe=!1;const ke=[];for(let Qe=0;Qe<L.length;Qe++){const Be=L.charAt(Qe);Be==="'"?((fe.length>0||xe)&&ke.push({literal:xe||/^\s+$/.test(fe),val:fe===""?"'":fe}),re=null,fe="",xe=!xe):xe||Be===re?fe+=Be:(fe.length>0&&ke.push({literal:/^\s+$/.test(fe),val:fe}),fe=Be,re=Be)}return fe.length>0&&ke.push({literal:xe||/^\s+$/.test(fe),val:fe}),ke}static macroTokenToFormatOpts(L){return Mr[L]}constructor(L,re){this.opts=re,this.loc=L,this.systemLoc=null}formatWithSystemDefault(L,re){return this.systemLoc===null&&(this.systemLoc=this.loc.redefaultToSystem()),this.systemLoc.dtFormatter(L,{...this.opts,...re}).format()}dtFormatter(L,re={}){return this.loc.dtFormatter(L,{...this.opts,...re})}formatDateTime(L,re){return this.dtFormatter(L,re).format()}formatDateTimeParts(L,re){return this.dtFormatter(L,re).formatToParts()}formatInterval(L,re){return this.dtFormatter(L.start,re).dtf.formatRange(L.start.toJSDate(),L.end.toJSDate())}resolvedOptions(L,re){return this.dtFormatter(L,re).resolvedOptions()}num(L,re=0,fe=void 0){if(this.opts.forceSimple)return tr(L,re);const xe={...this.opts};return re>0&&(xe.padTo=re),fe&&(xe.signDisplay=fe),this.loc.numberFormatter(xe).format(L)}formatDateTimeFromString(L,re){const fe=this.loc.listingMode()==="en",xe=this.loc.outputCalendar&&this.loc.outputCalendar!=="gregory",ke=(mn,Ri)=>this.loc.extract(L,mn,Ri),Qe=mn=>L.isOffsetFixed&&L.offset===0&&mn.allowZ?"Z":L.isValid?L.zone.formatOffset(L.ts,mn.format):"",Be=()=>fe?As(L):ke({hour:"numeric",hourCycle:"h12"},"dayperiod"),Ze=(mn,Ri)=>fe?_p(L,mn):ke(Ri?{month:mn}:{month:mn,day:"numeric"},"month"),Rt=(mn,Ri)=>fe?Uo(L,mn):ke(Ri?{weekday:mn}:{weekday:mn,month:"long",day:"numeric"},"weekday"),Kt=mn=>{const Ri=Qn.macroTokenToFormatOpts(mn);return Ri?this.formatWithSystemDefault(L,Ri):mn},zn=mn=>fe?dh(L,mn):ke({era:mn},"era"),ln=mn=>{switch(mn){case"S":return this.num(L.millisecond);case"u":case"SSS":return this.num(L.millisecond,3);case"s":return this.num(L.second);case"ss":return this.num(L.second,2);case"uu":return this.num(Math.floor(L.millisecond/10),2);case"uuu":return this.num(Math.floor(L.millisecond/100));case"m":return this.num(L.minute);case"mm":return this.num(L.minute,2);case"h":return this.num(L.hour%12===0?12:L.hour%12);case"hh":return this.num(L.hour%12===0?12:L.hour%12,2);case"H":return this.num(L.hour);case"HH":return this.num(L.hour,2);case"Z":return Qe({format:"narrow",allowZ:this.opts.allowZ});case"ZZ":return Qe({format:"short",allowZ:this.opts.allowZ});case"ZZZ":return Qe({format:"techie",allowZ:this.opts.allowZ});case"ZZZZ":return L.zone.offsetName(L.ts,{format:"short",locale:this.loc.locale});case"ZZZZZ":return L.zone.offsetName(L.ts,{format:"long",locale:this.loc.locale});case"z":return L.zoneName;case"a":return Be();case"d":return xe?ke({day:"numeric"},"day"):this.num(L.day);case"dd":return xe?ke({day:"2-digit"},"day"):this.num(L.day,2);case"c":return this.num(L.weekday);case"ccc":return Rt("short",!0);case"cccc":return Rt("long",!0);case"ccccc":return Rt("narrow",!0);case"E":return this.num(L.weekday);case"EEE":return Rt("short",!1);case"EEEE":return Rt("long",!1);case"EEEEE":return Rt("narrow",!1);case"L":return xe?ke({month:"numeric",day:"numeric"},"month"):this.num(L.month);case"LL":return xe?ke({month:"2-digit",day:"numeric"},"month"):this.num(L.month,2);case"LLL":return Ze("short",!0);case"LLLL":return Ze("long",!0);case"LLLLL":return Ze("narrow",!0);case"M":return xe?ke({month:"numeric"},"month"):this.num(L.month);case"MM":return xe?ke({month:"2-digit"},"month"):this.num(L.month,2);case"MMM":return Ze("short",!1);case"MMMM":return Ze("long",!1);case"MMMMM":return Ze("narrow",!1);case"y":return xe?ke({year:"numeric"},"year"):this.num(L.year);case"yy":return xe?ke({year:"2-digit"},"year"):this.num(L.year.toString().slice(-2),2);case"yyyy":return xe?ke({year:"numeric"},"year"):this.num(L.year,4);case"yyyyyy":return xe?ke({year:"numeric"},"year"):this.num(L.year,6);case"G":return zn("short");case"GG":return zn("long");case"GGGGG":return zn("narrow");case"kk":return this.num(L.weekYear.toString().slice(-2),2);case"kkkk":return this.num(L.weekYear,4);case"W":return this.num(L.weekNumber);case"WW":return this.num(L.weekNumber,2);case"n":return this.num(L.localWeekNumber);case"nn":return this.num(L.localWeekNumber,2);case"ii":return this.num(L.localWeekYear.toString().slice(-2),2);case"iiii":return this.num(L.localWeekYear,4);case"o":return this.num(L.ordinal);case"ooo":return this.num(L.ordinal,3);case"q":return this.num(L.quarter);case"qq":return this.num(L.quarter,2);case"X":return this.num(Math.floor(L.ts/1e3));case"x":return this.num(L.ts);default:return Kt(mn)}};return hh(Qn.parseFormat(re),ln)}formatDurationFromString(L,re){const fe=this.opts.signMode==="negativeLargestOnly"?-1:1,xe=Kt=>{switch(Kt[0]){case"S":return"milliseconds";case"s":return"seconds";case"m":return"minutes";case"h":return"hours";case"d":return"days";case"w":return"weeks";case"M":return"months";case"y":return"years";default:return null}},ke=(Kt,zn)=>ln=>{const mn=xe(ln);if(mn){const Ri=zn.isNegativeDuration&&mn!==zn.largestUnit?fe:1;let za;return this.opts.signMode==="negativeLargestOnly"&&mn!==zn.largestUnit?za="never":this.opts.signMode==="all"?za="always":za="auto",this.num(Kt.get(mn)*Ri,ln.length,za)}else return ln},Qe=Qn.parseFormat(re),Be=Qe.reduce((Kt,{literal:zn,val:ln})=>zn?Kt:Kt.concat(ln),[]),Ze=L.shiftTo(...Be.map(xe).filter(Kt=>Kt)),Rt={isNegativeDuration:Ze<0,largestUnit:Object.keys(Ze.values)[0]};return hh(Qe,ke(Ze,Rt))}}const xc=/[A-Za-z_+-]{1,256}(?::?\/[A-Za-z0-9_+-]{1,256}(?:\/[A-Za-z0-9_+-]{1,256})?)?/;function jo(...ue){const L=ue.reduce((re,fe)=>re+fe.source,"");return RegExp(`^${L}$`)}function Sc(...ue){return L=>ue.reduce(([re,fe,xe],ke)=>{const[Qe,Be,Ze]=ke(L,xe);return[{...re,...Qe},Be||fe,Ze]},[{},null,1]).slice(0,2)}function ra(ue,...L){if(ue==null)return[null,null];for(const[re,fe]of L){const xe=re.exec(ue);if(xe)return fe(xe)}return[null,null]}function Ts(...ue){return(L,re)=>{const fe={};let xe;for(xe=0;xe<ue.length;xe++)fe[ue[xe]]=Bn(L[re+xe]);return[fe,null,re+xe]}}const ad=/(?:([Zz])|([+-]\d\d)(?::?(\d\d))?)/,Ll=`(?:${ad.source}?(?:\\[(${xc.source})\\])?)?`,ud=/(\d\d)(?::?(\d\d)(?::?(\d\d)(?:[.,](\d{1,30}))?)?)?/,ia=RegExp(`${ud.source}${Ll}`),ks=RegExp(`(?:[Tt]${ia.source})?`),Bl=/([+-]\d{6}|\d{4})(?:-?(\d\d)(?:-?(\d\d))?)?/,ld=/(\d{4})-?W(\d\d)(?:-?(\d))?/,Ep=/(\d{4})-?(\d{3})/,ui=Ts("weekYear","weekNumber","weekDay"),X0=Ts("year","ordinal"),Y0=/(\d{4})-(\d\d)-(\d\d)/,od=RegExp(`${ud.source} ?(?:${ad.source}|(${xc.source}))?`),G0=RegExp(`(?: ${od.source})?`);function _c(ue,L,re){const fe=ue[L];return wt(fe)?re:Bn(fe)}function sd(ue,L){return[{year:_c(ue,L),month:_c(ue,L+1,1),day:_c(ue,L+2,1)},null,L+3]}function ha(ue,L){return[{hours:_c(ue,L,0),minutes:_c(ue,L+1,0),seconds:_c(ue,L+2,0),milliseconds:gf(ue[L+3])},null,L+4]}function cd(ue,L){const re=!ue[L]&&!ue[L+1],fe=yc(ue[L+1],ue[L+2]),xe=re?null:qt.instance(fe);return[{},xe,L+3]}function Ba(ue,L){const re=ue[L]?ie.create(ue[L]):null;return[{},re,L+1]}const Eg=RegExp(`^T?${ud.source}$`),Dp=/^-?P(?:(?:(-?\d{1,20}(?:\.\d{1,20})?)Y)?(?:(-?\d{1,20}(?:\.\d{1,20})?)M)?(?:(-?\d{1,20}(?:\.\d{1,20})?)W)?(?:(-?\d{1,20}(?:\.\d{1,20})?)D)?(?:T(?:(-?\d{1,20}(?:\.\d{1,20})?)H)?(?:(-?\d{1,20}(?:\.\d{1,20})?)M)?(?:(-?\d{1,20})(?:[.,](-?\d{1,20}))?S)?)?)$/;function Zo(ue){const[L,re,fe,xe,ke,Qe,Be,Ze,Rt]=ue,Kt=L[0]==="-",zn=Ze&&Ze[0]==="-",ln=(mn,Ri=!1)=>mn!==void 0&&(Ri||mn&&Kt)?-mn:mn;return[{years:ln(ku(re)),months:ln(ku(fe)),weeks:ln(ku(xe)),days:ln(ku(ke)),hours:ln(ku(Qe)),minutes:ln(ku(Be)),seconds:ln(ku(Ze),Ze==="-0"),milliseconds:ln(gf(Rt),zn)}]}const ph={GMT:0,EDT:-240,EST:-300,CDT:-300,CST:-360,MDT:-360,MST:-420,PDT:-420,PST:-480};function Ps(ue,L,re,fe,xe,ke,Qe){const Be={year:L.length===2?id(Bn(L)):Bn(L),month:vf.indexOf(re)+1,day:Bn(fe),hour:Bn(xe),minute:Bn(ke)};return Qe&&(Be.second=Bn(Qe)),ue&&(Be.weekday=ue.length>3?yf.indexOf(ue)+1:ho.indexOf(ue)+1),Be}const bi=/^(?:(Mon|Tue|Wed|Thu|Fri|Sat|Sun),\s)?(\d{1,2})\s(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s(\d{2,4})\s(\d\d):(\d\d)(?::(\d\d))?\s(?:(UT|GMT|[ECMP][SD]T)|([Zz])|(?:([+-]\d\d)(\d\d)))$/;function Ho(ue){const[,L,re,fe,xe,ke,Qe,Be,Ze,Rt,Kt,zn]=ue,ln=Ps(L,xe,fe,re,ke,Qe,Be);let mn;return Ze?mn=ph[Ze]:Rt?mn=0:mn=yc(Kt,zn),[ln,new qt(mn)]}function $p(ue){return ue.replace(/\([^()]*\)|[\n\t]/g," ").replace(/(\s\s+)/g," ").trim()}const Cp=/^(Mon|Tue|Wed|Thu|Fri|Sat|Sun), (\d\d) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) (\d{4}) (\d\d):(\d\d):(\d\d) GMT$/,wp=/^(Monday|Tuesday|Wednesday|Thursday|Friday|Saturday|Sunday), (\d\d)-(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)-(\d\d) (\d\d):(\d\d):(\d\d) GMT$/,Ms=/^(Mon|Tue|Wed|Thu|Fri|Sat|Sun) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) ( \d|\d\d) (\d\d):(\d\d):(\d\d) (\d{4})$/;function Rp(ue){const[,L,re,fe,xe,ke,Qe,Be]=ue;return[Ps(L,xe,fe,re,ke,Qe,Be),qt.utcInstance]}function Dy(ue){const[,L,re,fe,xe,ke,Qe,Be]=ue;return[Ps(L,Be,re,fe,xe,ke,Qe),qt.utcInstance]}const K0=jo(Bl,ks),po=jo(ld,ks),Ku=jo(Ep,ks),W0=jo(ia),mh=Sc(sd,ha,cd,Ba),Ec=Sc(ui,ha,cd,Ba),Xo=Sc(X0,ha,cd,Ba),Fs=Sc(ha,cd,Ba);function Qa(ue){return ra(ue,[K0,mh],[po,Ec],[Ku,Xo],[W0,Fs])}function gh(ue){return ra($p(ue),[bi,Ho])}function vh(ue){return ra(ue,[Cp,Rp],[wp,Rp],[Ms,Dy])}function fd(ue){return ra(ue,[Dp,Zo])}const dd=Sc(ha);function Ns(ue){return ra(ue,[Eg,dd])}const pu=jo(Y0,G0),Ap=jo(od),Ga=Sc(ha,cd,Ba);function Tp(ue){return ra(ue,[pu,mh],[Ap,Ga])}const hd="Invalid Duration",bf={weeks:{days:7,hours:168,minutes:10080,seconds:10080*60,milliseconds:10080*60*1e3},days:{hours:24,minutes:1440,seconds:1440*60,milliseconds:1440*60*1e3},hours:{minutes:60,seconds:3600,milliseconds:3600*1e3},minutes:{seconds:60,milliseconds:60*1e3},seconds:{milliseconds:1e3}},pd={years:{quarters:4,months:12,weeks:52,days:365,hours:365*24,minutes:365*24*60,seconds:365*24*60*60,milliseconds:365*24*60*60*1e3},quarters:{months:3,weeks:13,days:91,hours:2184,minutes:2184*60,seconds:2184*60*60,milliseconds:2184*60*60*1e3},months:{weeks:4,days:30,hours:720,minutes:720*60,seconds:720*60*60,milliseconds:720*60*60*1e3},...bf},Wu=146097/400,mo=146097/4800,J0={years:{quarters:4,months:12,weeks:Wu/7,days:Wu,hours:Wu*24,minutes:Wu*24*60,seconds:Wu*24*60*60,milliseconds:Wu*24*60*60*1e3},quarters:{months:3,weeks:Wu/28,days:Wu/4,hours:Wu*24/4,minutes:Wu*24*60/4,seconds:Wu*24*60*60/4,milliseconds:Wu*24*60*60*1e3/4},months:{weeks:mo/7,days:mo,hours:mo*24,minutes:mo*24*60,seconds:mo*24*60*60,milliseconds:mo*24*60*60*1e3},...bf},mu=["years","quarters","months","weeks","days","hours","minutes","seconds","milliseconds"],Of=mu.slice(0).reverse();function Ar(ue,L,re=!1){const fe={values:re?L.values:{...ue.values,...L.values||{}},loc:ue.loc.clone(L.loc),conversionAccuracy:L.conversionAccuracy||ue.conversionAccuracy,matrix:L.matrix||ue.matrix};return new Kn(fe)}function yh(ue,L){var re;let fe=(re=L.milliseconds)!=null?re:0;for(const xe of Of.slice(1))L[xe]&&(fe+=L[xe]*ue[xe].milliseconds);return fe}function kp(ue,L){const re=yh(ue,L)<0?-1:1;mu.reduceRight((fe,xe)=>{if(wt(L[xe]))return fe;if(fe){const ke=L[fe]*re,Qe=ue[xe][fe],Be=Math.floor(ke/Qe);L[xe]+=Be*re,L[fe]-=Be*Qe*re}return xe},null),mu.reduce((fe,xe)=>{if(wt(L[xe]))return fe;if(fe){const ke=L[fe]%1;L[fe]-=ke,L[xe]+=ke*ue[fe][xe]}return xe},null)}function go(ue){const L={};for(const[re,fe]of Object.entries(ue))fe!==0&&(L[re]=fe);return L}class Kn{constructor(L){const re=L.conversionAccuracy==="longterm"||!1;let fe=re?J0:pd;L.matrix&&(fe=L.matrix),this.values=L.values,this.loc=L.loc||bt.create(),this.conversionAccuracy=re?"longterm":"casual",this.invalid=L.invalid||null,this.matrix=fe,this.isLuxonDuration=!0}static fromMillis(L,re){return Kn.fromObject({milliseconds:L},re)}static fromObject(L,re={}){if(L==null||typeof L!="object")throw new u(`Duration.fromObject: argument expected to be an object, got ${L===null?"null":typeof L}`);return new Kn({values:Pu(L,Kn.normalizeUnit),loc:bt.fromObject(re),conversionAccuracy:re.conversionAccuracy,matrix:re.matrix})}static fromDurationLike(L){if(Kr(L))return Kn.fromMillis(L);if(Kn.isDuration(L))return L;if(typeof L=="object")return Kn.fromObject(L);throw new u(`Unknown duration argument ${L} of type ${typeof L}`)}static fromISO(L,re){const[fe]=fd(L);return fe?Kn.fromObject(fe,re):Kn.invalid("unparsable",`the input "${L}" can't be parsed as ISO 8601`)}static fromISOTime(L,re){const[fe]=Ns(L);return fe?Kn.fromObject(fe,re):Kn.invalid("unparsable",`the input "${L}" can't be parsed as ISO 8601`)}static invalid(L,re=null){if(!L)throw new u("need to specify a reason the Duration is invalid");const fe=L instanceof Te?L:new Te(L,re);if(ar.throwOnInvalid)throw new r(fe);return new Kn({invalid:fe})}static normalizeUnit(L){const re={year:"years",years:"years",quarter:"quarters",quarters:"quarters",month:"months",months:"months",week:"weeks",weeks:"weeks",day:"days",days:"days",hour:"hours",hours:"hours",minute:"minutes",minutes:"minutes",second:"seconds",seconds:"seconds",millisecond:"milliseconds",milliseconds:"milliseconds"}[L&&L.toLowerCase()];if(!re)throw new a(L);return re}static isDuration(L){return L&&L.isLuxonDuration||!1}get locale(){return this.isValid?this.loc.locale:null}get numberingSystem(){return this.isValid?this.loc.numberingSystem:null}toFormat(L,re={}){const fe={...re,floor:re.round!==!1&&re.floor!==!1};return this.isValid?Qn.create(this.loc,fe).formatDurationFromString(this,L):hd}toHuman(L={}){if(!this.isValid)return hd;const re=L.showZeros!==!1,fe=mu.map(xe=>{const ke=this.values[xe];return wt(ke)||ke===0&&!re?null:this.loc.numberFormatter({style:"unit",unitDisplay:"long",...L,unit:xe.slice(0,-1)}).format(ke)}).filter(xe=>xe);return this.loc.listFormatter({type:"conjunction",style:L.listStyle||"narrow",...L}).format(fe)}toObject(){return this.isValid?{...this.values}:{}}toISO(){if(!this.isValid)return null;let L="P";return this.years!==0&&(L+=this.years+"Y"),(this.months!==0||this.quarters!==0)&&(L+=this.months+this.quarters*3+"M"),this.weeks!==0&&(L+=this.weeks+"W"),this.days!==0&&(L+=this.days+"D"),(this.hours!==0||this.minutes!==0||this.seconds!==0||this.milliseconds!==0)&&(L+="T"),this.hours!==0&&(L+=this.hours+"H"),this.minutes!==0&&(L+=this.minutes+"M"),(this.seconds!==0||this.milliseconds!==0)&&(L+=Ds(this.seconds+this.milliseconds/1e3,3)+"S"),L==="P"&&(L+="T0S"),L}toISOTime(L={}){if(!this.isValid)return null;const re=this.toMillis();return re<0||re>=864e5?null:(L={suppressMilliseconds:!1,suppressSeconds:!1,includePrefix:!1,format:"extended",...L,includeOffset:!1},dn.fromMillis(re,{zone:"UTC"}).toISOTime(L))}toJSON(){return this.toISO()}toString(){return this.toISO()}[Symbol.for("nodejs.util.inspect.custom")](){return this.isValid?`Duration { values: ${JSON.stringify(this.values)} }`:`Duration { Invalid, reason: ${this.invalidReason} }`}toMillis(){return this.isValid?yh(this.matrix,this.values):NaN}valueOf(){return this.toMillis()}plus(L){if(!this.isValid)return this;const re=Kn.fromDurationLike(L),fe={};for(const xe of mu)(xl(re.values,xe)||xl(this.values,xe))&&(fe[xe]=re.get(xe)+this.get(xe));return Ar(this,{values:fe},!0)}minus(L){if(!this.isValid)return this;const re=Kn.fromDurationLike(L);return this.plus(re.negate())}mapUnits(L){if(!this.isValid)return this;const re={};for(const fe of Object.keys(this.values))re[fe]=Vo(L(this.values[fe],fe));return Ar(this,{values:re},!0)}get(L){return this[Kn.normalizeUnit(L)]}set(L){if(!this.isValid)return this;const re={...this.values,...Pu(L,Kn.normalizeUnit)};return Ar(this,{values:re})}reconfigure({locale:L,numberingSystem:re,conversionAccuracy:fe,matrix:xe}={}){const Qe={loc:this.loc.clone({locale:L,numberingSystem:re}),matrix:xe,conversionAccuracy:fe};return Ar(this,Qe)}as(L){return this.isValid?this.shiftTo(L).get(L):NaN}normalize(){if(!this.isValid)return this;const L=this.toObject();return kp(this.matrix,L),Ar(this,{values:L},!0)}rescale(){if(!this.isValid)return this;const L=go(this.normalize().shiftToAll().toObject());return Ar(this,{values:L},!0)}shiftTo(...L){if(!this.isValid)return this;if(L.length===0)return this;L=L.map(Qe=>Kn.normalizeUnit(Qe));const re={},fe={},xe=this.toObject();let ke;for(const Qe of mu)if(L.indexOf(Qe)>=0){ke=Qe;let Be=0;for(const Rt in fe)Be+=this.matrix[Rt][Qe]*fe[Rt],fe[Rt]=0;Kr(xe[Qe])&&(Be+=xe[Qe]);const Ze=Math.trunc(Be);re[Qe]=Ze,fe[Qe]=(Be*1e3-Ze*1e3)/1e3}else Kr(xe[Qe])&&(fe[Qe]=xe[Qe]);for(const Qe in fe)fe[Qe]!==0&&(re[ke]+=Qe===ke?fe[Qe]:fe[Qe]/this.matrix[ke][Qe]);return kp(this.matrix,re),Ar(this,{values:re},!0)}shiftToAll(){return this.isValid?this.shiftTo("years","months","weeks","days","hours","minutes","seconds","milliseconds"):this}negate(){if(!this.isValid)return this;const L={};for(const re of Object.keys(this.values))L[re]=this.values[re]===0?0:-this.values[re];return Ar(this,{values:L},!0)}removeZeros(){if(!this.isValid)return this;const L=go(this.values);return Ar(this,{values:L},!0)}get years(){return this.isValid?this.values.years||0:NaN}get quarters(){return this.isValid?this.values.quarters||0:NaN}get months(){return this.isValid?this.values.months||0:NaN}get weeks(){return this.isValid?this.values.weeks||0:NaN}get days(){return this.isValid?this.values.days||0:NaN}get hours(){return this.isValid?this.values.hours||0:NaN}get minutes(){return this.isValid?this.values.minutes||0:NaN}get seconds(){return this.isValid?this.values.seconds||0:NaN}get milliseconds(){return this.isValid?this.values.milliseconds||0:NaN}get isValid(){return this.invalid===null}get invalidReason(){return this.invalid?this.invalid.reason:null}get invalidExplanation(){return this.invalid?this.invalid.explanation:null}equals(L){if(!this.isValid||!L.isValid||!this.loc.equals(L.loc))return!1;function re(fe,xe){return fe===void 0||fe===0?xe===void 0||xe===0:fe===xe}for(const fe of mu)if(!re(this.values[fe],L.values[fe]))return!1;return!0}}const _l="Invalid Interval";function Dg(ue,L){return!ue||!ue.isValid?Ii.invalid("missing or invalid start"):!L||!L.isValid?Ii.invalid("missing or invalid end"):L<ue?Ii.invalid("end before start",`The end of an interval must be after its start, but you had start=${ue.toISO()} and end=${L.toISO()}`):null}class Ii{constructor(L){this.s=L.start,this.e=L.end,this.invalid=L.invalid||null,this.isLuxonInterval=!0}static invalid(L,re=null){if(!L)throw new u("need to specify a reason the Interval is invalid");const fe=L instanceof Te?L:new Te(L,re);if(ar.throwOnInvalid)throw new n(fe);return new Ii({invalid:fe})}static fromDateTimes(L,re){const fe=Nu(L),xe=Nu(re),ke=Dg(fe,xe);return ke??new Ii({start:fe,end:xe})}static after(L,re){const fe=Kn.fromDurationLike(re),xe=Nu(L);return Ii.fromDateTimes(xe,xe.plus(fe))}static before(L,re){const fe=Kn.fromDurationLike(re),xe=Nu(L);return Ii.fromDateTimes(xe.minus(fe),xe)}static fromISO(L,re){const[fe,xe]=(L||"").split("/",2);if(fe&&xe){let ke,Qe;try{ke=dn.fromISO(fe,re),Qe=ke.isValid}catch{Qe=!1}let Be,Ze;try{Be=dn.fromISO(xe,re),Ze=Be.isValid}catch{Ze=!1}if(Qe&&Ze)return Ii.fromDateTimes(ke,Be);if(Qe){const Rt=Kn.fromISO(xe,re);if(Rt.isValid)return Ii.after(ke,Rt)}else if(Ze){const Rt=Kn.fromISO(fe,re);if(Rt.isValid)return Ii.before(Be,Rt)}}return Ii.invalid("unparsable",`the input "${L}" can't be parsed as ISO 8601`)}static isInterval(L){return L&&L.isLuxonInterval||!1}get start(){return this.isValid?this.s:null}get end(){return this.isValid?this.e:null}get lastDateTime(){return this.isValid&&this.e?this.e.minus(1):null}get isValid(){return this.invalidReason===null}get invalidReason(){return this.invalid?this.invalid.reason:null}get invalidExplanation(){return this.invalid?this.invalid.explanation:null}length(L="milliseconds"){return this.isValid?this.toDuration(L).get(L):NaN}count(L="milliseconds",re){if(!this.isValid)return NaN;const fe=this.start.startOf(L,re);let xe;return re!=null&&re.useLocaleWeeks?xe=this.end.reconfigure({locale:fe.locale}):xe=this.end,xe=xe.startOf(L,re),Math.floor(xe.diff(fe,L).get(L))+(xe.valueOf()!==this.end.valueOf())}hasSame(L){return this.isValid?this.isEmpty()||this.e.minus(1).hasSame(this.s,L):!1}isEmpty(){return this.s.valueOf()===this.e.valueOf()}isAfter(L){return this.isValid?this.s>L:!1}isBefore(L){return this.isValid?this.e<=L:!1}contains(L){return this.isValid?this.s<=L&&this.e>L:!1}set({start:L,end:re}={}){return this.isValid?Ii.fromDateTimes(L||this.s,re||this.e):this}splitAt(...L){if(!this.isValid)return[];const re=L.map(Nu).filter(Qe=>this.contains(Qe)).sort((Qe,Be)=>Qe.toMillis()-Be.toMillis()),fe=[];let{s:xe}=this,ke=0;for(;xe<this.e;){const Qe=re[ke]||this.e,Be=+Qe>+this.e?this.e:Qe;fe.push(Ii.fromDateTimes(xe,Be)),xe=Be,ke+=1}return fe}splitBy(L){const re=Kn.fromDurationLike(L);if(!this.isValid||!re.isValid||re.as("milliseconds")===0)return[];let{s:fe}=this,xe=1,ke;const Qe=[];for(;fe<this.e;){const Be=this.start.plus(re.mapUnits(Ze=>Ze*xe));ke=+Be>+this.e?this.e:Be,Qe.push(Ii.fromDateTimes(fe,ke)),fe=ke,xe+=1}return Qe}divideEqually(L){return this.isValid?this.splitBy(this.length()/L).slice(0,L):[]}overlaps(L){return this.e>L.s&&this.s<L.e}abutsStart(L){return this.isValid?+this.e==+L.s:!1}abutsEnd(L){return this.isValid?+L.e==+this.s:!1}engulfs(L){return this.isValid?this.s<=L.s&&this.e>=L.e:!1}equals(L){return!this.isValid||!L.isValid?!1:this.s.equals(L.s)&&this.e.equals(L.e)}intersection(L){if(!this.isValid)return this;const re=this.s>L.s?this.s:L.s,fe=this.e<L.e?this.e:L.e;return re>=fe?null:Ii.fromDateTimes(re,fe)}union(L){if(!this.isValid)return this;const re=this.s<L.s?this.s:L.s,fe=this.e>L.e?this.e:L.e;return Ii.fromDateTimes(re,fe)}static merge(L){const[re,fe]=L.sort((xe,ke)=>xe.s-ke.s).reduce(([xe,ke],Qe)=>ke?ke.overlaps(Qe)||ke.abutsStart(Qe)?[xe,ke.union(Qe)]:[xe.concat([ke]),Qe]:[xe,Qe],[[],null]);return fe&&re.push(fe),re}static xor(L){let re=null,fe=0;const xe=[],ke=L.map(Ze=>[{time:Ze.s,type:"s"},{time:Ze.e,type:"e"}]),Qe=Array.prototype.concat(...ke),Be=Qe.sort((Ze,Rt)=>Ze.time-Rt.time);for(const Ze of Be)fe+=Ze.type==="s"?1:-1,fe===1?re=Ze.time:(re&&+re!=+Ze.time&&xe.push(Ii.fromDateTimes(re,Ze.time)),re=null);return Ii.merge(xe)}difference(...L){return Ii.xor([this].concat(L)).map(re=>this.intersection(re)).filter(re=>re&&!re.isEmpty())}toString(){return this.isValid?`[${this.s.toISO()} – ${this.e.toISO()})`:_l}[Symbol.for("nodejs.util.inspect.custom")](){return this.isValid?`Interval { start: ${this.s.toISO()}, end: ${this.e.toISO()} }`:`Interval { Invalid, reason: ${this.invalidReason} }`}toLocaleString(L=d,re={}){return this.isValid?Qn.create(this.s.loc.clone(re),L).formatInterval(this):_l}toISO(L){return this.isValid?`${this.s.toISO(L)}/${this.e.toISO(L)}`:_l}toISODate(){return this.isValid?`${this.s.toISODate()}/${this.e.toISODate()}`:_l}toISOTime(L){return this.isValid?`${this.s.toISOTime(L)}/${this.e.toISOTime(L)}`:_l}toFormat(L,{separator:re=" – "}={}){return this.isValid?`${this.s.toFormat(L)}${re}${this.e.toFormat(L)}`:_l}toDuration(L,re){return this.isValid?this.e.diff(this.s,L,re):Kn.invalid(this.invalidReason)}mapEndpoints(L){return Ii.fromDateTimes(L(this.s),L(this.e))}}class gu{static hasDST(L=ar.defaultZone){const re=dn.now().setZone(L).set({month:12});return!L.isUniversal&&re.offset!==re.set({month:6}).offset}static isValidIANAZone(L){return ie.isValidZone(L)}static normalizeZone(L){return nn(L,ar.defaultZone)}static getStartOfWeek({locale:L=null,locObj:re=null}={}){return(re||bt.create(L)).getStartOfWeek()}static getMinimumDaysInFirstWeek({locale:L=null,locObj:re=null}={}){return(re||bt.create(L)).getMinDaysInFirstWeek()}static getWeekendWeekdays({locale:L=null,locObj:re=null}={}){return(re||bt.create(L)).getWeekendDays().slice()}static months(L="long",{locale:re=null,numberingSystem:fe=null,locObj:xe=null,outputCalendar:ke="gregory"}={}){return(xe||bt.create(re,fe,ke)).months(L)}static monthsFormat(L="long",{locale:re=null,numberingSystem:fe=null,locObj:xe=null,outputCalendar:ke="gregory"}={}){return(xe||bt.create(re,fe,ke)).months(L,!0)}static weekdays(L="long",{locale:re=null,numberingSystem:fe=null,locObj:xe=null}={}){return(xe||bt.create(re,fe,null)).weekdays(L)}static weekdaysFormat(L="long",{locale:re=null,numberingSystem:fe=null,locObj:xe=null}={}){return(xe||bt.create(re,fe,null)).weekdays(L,!0)}static meridiems({locale:L=null}={}){return bt.create(L).meridiems()}static eras(L="short",{locale:re=null}={}){return bt.create(re,null,"gregory").eras(L)}static features(){return{relative:Xa(),localeWeek:Au()}}}function pr(ue,L){const re=xe=>xe.toUTC(0,{keepLocalTime:!0}).startOf("day").valueOf(),fe=re(L)-re(ue);return Math.floor(Kn.fromMillis(fe).as("days"))}function xf(ue,L,re){const fe=[["years",(Ze,Rt)=>Rt.year-Ze.year],["quarters",(Ze,Rt)=>Rt.quarter-Ze.quarter+(Rt.year-Ze.year)*4],["months",(Ze,Rt)=>Rt.month-Ze.month+(Rt.year-Ze.year)*12],["weeks",(Ze,Rt)=>{const Kt=pr(Ze,Rt);return(Kt-Kt%7)/7}],["days",pr]],xe={},ke=ue;let Qe,Be;for(const[Ze,Rt]of fe)re.indexOf(Ze)>=0&&(Qe=Ze,xe[Ze]=Rt(ue,L),Be=ke.plus(xe),Be>L?(xe[Ze]--,ue=ke.plus(xe),ue>L&&(Be=ue,xe[Ze]--,ue=ke.plus(xe))):ue=Be);return[ue,xe,Be,Qe]}function em(ue,L,re,fe){let[xe,ke,Qe,Be]=xf(ue,L,re);const Ze=L-xe,Rt=re.filter(zn=>["hours","minutes","seconds","milliseconds"].indexOf(zn)>=0);Rt.length===0&&(Qe<L&&(Qe=xe.plus({[Be]:1})),Qe!==xe&&(ke[Be]=(ke[Be]||0)+Ze/(Qe-xe)));const Kt=Kn.fromObject(ke,fe);return Rt.length>0?Kn.fromMillis(Ze,fe).shiftTo(...Rt).plus(Kt):Kt}const tm="missing Intl.DateTimeFormat.formatToParts support";function xr(ue,L=re=>re){return{regex:ue,deser:([re])=>L(Qt(re))}}const Sf="[  ]",Dc=new RegExp(Sf,"g");function Xe(ue){return ue.replace(/\./g,"\\.?").replace(Dc,Sf)}function dt(ue){return ue.replace(/\./g,"").replace(Dc," ").toLowerCase()}function rt(ue,L){return ue===null?null:{regex:RegExp(ue.map(Xe).join("|")),deser:([re])=>ue.findIndex(fe=>dt(re)===dt(fe))+L}}function Bt(ue,L){return{regex:ue,deser:([,re,fe])=>yc(re,fe),groups:L}}function En(ue){return{regex:ue,deser:([L])=>L}}function lr(ue){return ue.replace(/[\-\[\]{}()*+?.,\\\^$|#\s]/g,"\\$&")}function Li(ue,L){const re=rn(L),fe=rn(L,"{2}"),xe=rn(L,"{3}"),ke=rn(L,"{4}"),Qe=rn(L,"{6}"),Be=rn(L,"{1,2}"),Ze=rn(L,"{1,3}"),Rt=rn(L,"{1,6}"),Kt=rn(L,"{1,9}"),zn=rn(L,"{2,4}"),ln=rn(L,"{4,6}"),mn=Se=>({regex:RegExp(lr(Se.val)),deser:([El])=>El,literal:!0}),za=(Se=>{if(ue.literal)return mn(Se);switch(Se.val){case"G":return rt(L.eras("short"),0);case"GG":return rt(L.eras("long"),0);case"y":return xr(Rt);case"yy":return xr(zn,id);case"yyyy":return xr(ke);case"yyyyy":return xr(ln);case"yyyyyy":return xr(Qe);case"M":return xr(Be);case"MM":return xr(fe);case"MMM":return rt(L.months("short",!0),1);case"MMMM":return rt(L.months("long",!0),1);case"L":return xr(Be);case"LL":return xr(fe);case"LLL":return rt(L.months("short",!1),1);case"LLLL":return rt(L.months("long",!1),1);case"d":return xr(Be);case"dd":return xr(fe);case"o":return xr(Ze);case"ooo":return xr(xe);case"HH":return xr(fe);case"H":return xr(Be);case"hh":return xr(fe);case"h":return xr(Be);case"mm":return xr(fe);case"m":return xr(Be);case"q":return xr(Be);case"qq":return xr(fe);case"s":return xr(Be);case"ss":return xr(fe);case"S":return xr(Ze);case"SSS":return xr(xe);case"u":return En(Kt);case"uu":return En(Be);case"uuu":return xr(re);case"a":return rt(L.meridiems(),0);case"kkkk":return xr(ke);case"kk":return xr(zn,id);case"W":return xr(Be);case"WW":return xr(fe);case"E":case"c":return xr(re);case"EEE":return rt(L.weekdays("short",!1),1);case"EEEE":return rt(L.weekdays("long",!1),1);case"ccc":return rt(L.weekdays("short",!0),1);case"cccc":return rt(L.weekdays("long",!0),1);case"Z":case"ZZ":return Bt(new RegExp(`([+-]${Be.source})(?::(${fe.source}))?`),2);case"ZZZ":return Bt(new RegExp(`([+-]${Be.source})(${fe.source})?`),2);case"z":return En(/[a-z_+-/]{1,256}?/i);case" ":return En(/[^\S\n\r]/);default:return mn(Se)}})(ue)||{invalidReason:tm};return za.token=ue,za}const zr={year:{"2-digit":"yy",numeric:"yyyyy"},month:{numeric:"M","2-digit":"MM",short:"MMM",long:"MMMM"},day:{numeric:"d","2-digit":"dd"},weekday:{short:"EEE",long:"EEEE"},dayperiod:"a",dayPeriod:"a",hour12:{numeric:"h","2-digit":"hh"},hour24:{numeric:"H","2-digit":"HH"},minute:{numeric:"m","2-digit":"mm"},second:{numeric:"s","2-digit":"ss"},timeZoneName:{long:"ZZZZZ",short:"ZZZ"}};function Ka(ue,L,re){const{type:fe,value:xe}=ue;if(fe==="literal"){const Ze=/^\s+$/.test(xe);return{literal:!Ze,val:Ze?" ":xe}}const ke=L[fe];let Qe=fe;fe==="hour"&&(L.hour12!=null?Qe=L.hour12?"hour12":"hour24":L.hourCycle!=null?L.hourCycle==="h11"||L.hourCycle==="h12"?Qe="hour12":Qe="hour24":Qe=re.hour12?"hour12":"hour24");let Be=zr[Qe];if(typeof Be=="object"&&(Be=Be[ke]),Be)return{literal:!1,val:Be}}function $c(ue){return[`^${ue.map(re=>re.regex).reduce((re,fe)=>`${re}(${fe.source})`,"")}$`,ue]}function Ci(ue,L,re){const fe=ue.match(L);if(fe){const xe={};let ke=1;for(const Qe in re)if(xl(re,Qe)){const Be=re[Qe],Ze=Be.groups?Be.groups+1:1;!Be.literal&&Be.token&&(xe[Be.token.val[0]]=Be.deser(fe.slice(ke,ke+Ze))),ke+=Ze}return[fe,xe]}else return[fe,{}]}function md(ue){const L=ke=>{switch(ke){case"S":return"millisecond";case"s":return"second";case"m":return"minute";case"h":case"H":return"hour";case"d":return"day";case"o":return"ordinal";case"L":case"M":return"month";case"y":return"year";case"E":case"c":return"weekday";case"W":return"weekNumber";case"k":return"weekYear";case"q":return"quarter";default:return null}};let re=null,fe;return wt(ue.z)||(re=ie.create(ue.z)),wt(ue.Z)||(re||(re=new qt(ue.Z)),fe=ue.Z),wt(ue.q)||(ue.M=(ue.q-1)*3+1),wt(ue.h)||(ue.h<12&&ue.a===1?ue.h+=12:ue.h===12&&ue.a===0&&(ue.h=0)),ue.G===0&&ue.y&&(ue.y=-ue.y),wt(ue.u)||(ue.S=gf(ue.u)),[Object.keys(ue).reduce((ke,Qe)=>{const Be=L(Qe);return Be&&(ke[Be]=ue[Qe]),ke},{}),re,fe]}let Ql=null;function vo(){return Ql||(Ql=dn.fromMillis(1555555555555)),Ql}function yo(ue,L){if(ue.literal)return ue;const re=Qn.macroTokenToFormatOpts(ue.val),fe=pa(re,L);return fe==null||fe.includes(void 0)?ue:fe}function _f(ue,L){return Array.prototype.concat(...ue.map(re=>yo(re,L)))}class Sr{constructor(L,re){if(this.locale=L,this.format=re,this.tokens=_f(Qn.parseFormat(re),L),this.units=this.tokens.map(fe=>Li(fe,L)),this.disqualifyingUnit=this.units.find(fe=>fe.invalidReason),!this.disqualifyingUnit){const[fe,xe]=$c(this.units);this.regex=RegExp(fe,"i"),this.handlers=xe}}explainFromTokens(L){if(this.isValid){const[re,fe]=Ci(L,this.regex,this.handlers),[xe,ke,Qe]=fe?md(fe):[null,null,void 0];if(xl(fe,"a")&&xl(fe,"H"))throw new i("Can't include meridiem when specifying 24-hour format");return{input:L,tokens:this.tokens,regex:this.regex,rawMatches:re,matches:fe,result:xe,zone:ke,specificOffset:Qe}}else return{input:L,tokens:this.tokens,invalidReason:this.invalidReason}}get isValid(){return!this.disqualifyingUnit}get invalidReason(){return this.disqualifyingUnit?this.disqualifyingUnit.invalidReason:null}}function Yo(ue,L,re){return new Sr(ue,re).explainFromTokens(L)}function bh(ue,L,re){const{result:fe,zone:xe,specificOffset:ke,invalidReason:Qe}=Yo(ue,L,re);return[fe,xe,ke,Qe]}function pa(ue,L){if(!ue)return null;const fe=Qn.create(L,ue).dtFormatter(vo()),xe=fe.formatToParts(),ke=fe.resolvedOptions();return xe.map(Qe=>Ka(Qe,ue,ke))}const Go="Invalid DateTime",wa=864e13;function bo(ue){return new Te("unsupported zone",`the zone "${ue.name}" is not supported`)}function Is(ue){return ue.weekData===null&&(ue.weekData=Lt(ue.c)),ue.weekData}function Wa(ue){return ue.localWeekData===null&&(ue.localWeekData=Lt(ue.c,ue.loc.getMinDaysInFirstWeek(),ue.loc.getStartOfWeek())),ue.localWeekData}function Ju(ue,L){const re={ts:ue.ts,zone:ue.zone,c:ue.c,o:ue.o,loc:ue.loc,invalid:ue.invalid};return new dn({...re,...L,old:re})}function Cc(ue,L,re){let fe=ue-L*60*1e3;const xe=re.offset(fe);if(L===xe)return[fe,L];fe-=(xe-L)*60*1e3;const ke=re.offset(fe);return xe===ke?[fe,xe]:[ue-Math.min(xe,ke)*60*1e3,Math.max(xe,ke)]}function el(ue,L){ue+=L*60*1e3;const re=new Date(ue);return{year:re.getUTCFullYear(),month:re.getUTCMonth()+1,day:re.getUTCDate(),hour:re.getUTCHours(),minute:re.getUTCMinutes(),second:re.getUTCSeconds(),millisecond:re.getUTCMilliseconds()}}function wc(ue,L,re){return Cc(Ya(ue),L,re)}function Rc(ue,L){const re=ue.o,fe=ue.c.year+Math.trunc(L.years),xe=ue.c.month+Math.trunc(L.months)+Math.trunc(L.quarters)*3,ke={...ue.c,year:fe,month:xe,day:Math.min(ue.c.day,sh(fe,xe))+Math.trunc(L.days)+Math.trunc(L.weeks)*7},Qe=Kn.fromObject({years:L.years-Math.trunc(L.years),quarters:L.quarters-Math.trunc(L.quarters),months:L.months-Math.trunc(L.months),weeks:L.weeks-Math.trunc(L.weeks),days:L.days-Math.trunc(L.days),hours:L.hours,minutes:L.minutes,seconds:L.seconds,milliseconds:L.milliseconds}).as("milliseconds"),Be=Ya(ke);let[Ze,Rt]=Cc(Be,re,ue.zone);return Qe!==0&&(Ze+=Qe,Rt=ue.zone.offset(Ze)),{ts:Ze,o:Rt}}function vu(ue,L,re,fe,xe,ke){const{setZone:Qe,zone:Be}=re;if(ue&&Object.keys(ue).length!==0||L){const Ze=L||Be,Rt=dn.fromObject(ue,{...re,zone:Ze,specificOffset:ke});return Qe?Rt:Rt.setZone(Be)}else return dn.invalid(new Te("unparsable",`the input "${xe}" can't be parsed as ${fe}`))}function gd(ue,L,re=!0){return ue.isValid?Qn.create(bt.create("en-US"),{allowZ:re,forceSimple:!0}).formatDateTimeFromString(ue,L):null}function aa(ue,L,re){const fe=ue.c.year>9999||ue.c.year<0;let xe="";if(fe&&ue.c.year>=0&&(xe+="+"),xe+=tr(ue.c.year,fe?6:4),re==="year")return xe;if(L){if(xe+="-",xe+=tr(ue.c.month),re==="month")return xe;xe+="-"}else if(xe+=tr(ue.c.month),re==="month")return xe;return xe+=tr(ue.c.day),xe}function Ji(ue,L,re,fe,xe,ke,Qe){let Be=!re||ue.c.millisecond!==0||ue.c.second!==0,Ze="";switch(Qe){case"day":case"month":case"year":break;default:if(Ze+=tr(ue.c.hour),Qe==="hour")break;if(L){if(Ze+=":",Ze+=tr(ue.c.minute),Qe==="minute")break;Be&&(Ze+=":",Ze+=tr(ue.c.second))}else{if(Ze+=tr(ue.c.minute),Qe==="minute")break;Be&&(Ze+=tr(ue.c.second))}if(Qe==="second")break;Be&&(!fe||ue.c.millisecond!==0)&&(Ze+=".",Ze+=tr(ue.c.millisecond,3))}return xe&&(ue.isOffsetFixed&&ue.offset===0&&!ke?Ze+="Z":ue.o<0?(Ze+="-",Ze+=tr(Math.trunc(-ue.o/60)),Ze+=":",Ze+=tr(Math.trunc(-ue.o%60))):(Ze+="+",Ze+=tr(Math.trunc(ue.o/60)),Ze+=":",Ze+=tr(Math.trunc(ue.o%60)))),ke&&(Ze+="["+ue.zone.ianaName+"]"),Ze}const tl={month:1,day:1,hour:0,minute:0,second:0,millisecond:0},yu={weekNumber:1,weekday:1,hour:0,minute:0,second:0,millisecond:0},Oh={ordinal:1,hour:0,minute:0,second:0,millisecond:0},zl=["year","month","day","hour","minute","second","millisecond"],Ef=["weekYear","weekNumber","weekday","hour","minute","second","millisecond"],xh=["year","ordinal","hour","minute","second","millisecond"];function bu(ue){const L={year:"year",years:"year",month:"month",months:"month",day:"day",days:"day",hour:"hour",hours:"hour",minute:"minute",minutes:"minute",quarter:"quarter",quarters:"quarter",second:"second",seconds:"second",millisecond:"millisecond",milliseconds:"millisecond",weekday:"weekday",weekdays:"weekday",weeknumber:"weekNumber",weeksnumber:"weekNumber",weeknumbers:"weekNumber",weekyear:"weekYear",weekyears:"weekYear",ordinal:"ordinal"}[ue.toLowerCase()];if(!L)throw new a(ue);return L}function Pp(ue){switch(ue.toLowerCase()){case"localweekday":case"localweekdays":return"localWeekday";case"localweeknumber":case"localweeknumbers":return"localWeekNumber";case"localweekyear":case"localweekyears":return"localWeekYear";default:return bu(ue)}}function Mu(ue){if(wi===void 0&&(wi=ar.now()),ue.type!=="iana")return ue.offset(wi);const L=ue.name;let re=Fu.get(L);return re===void 0&&(re=ue.offset(wi),Fu.set(L,re)),re}function ql(ue,L){const re=nn(L.zone,ar.defaultZone);if(!re.isValid)return dn.invalid(bo(re));const fe=bt.fromObject(L);let xe,ke;if(wt(ue.year))xe=ar.now();else{for(const Ze of zl)wt(ue[Ze])&&(ue[Ze]=tl[Ze]);const Qe=$i(ue)||hr(ue);if(Qe)return dn.invalid(Qe);const Be=Mu(re);[xe,ke]=wc(ue,Be,re)}return new dn({ts:xe,zone:re,loc:fe,o:ke})}function Ac(ue,L,re){const fe=wt(re.round)?!0:re.round,xe=wt(re.rounding)?"trunc":re.rounding,ke=(Be,Ze)=>(Be=Ds(Be,fe||re.calendary?0:2,re.calendary?"round":xe),L.loc.clone(re).relFormatter(re).format(Be,Ze)),Qe=Be=>re.calendary?L.hasSame(ue,Be)?0:L.startOf(Be).diff(ue.startOf(Be),Be).get(Be):L.diff(ue,Be).get(Be);if(re.unit)return ke(Qe(re.unit),re.unit);for(const Be of re.units){const Ze=Qe(Be);if(Math.abs(Ze)>=1)return ke(Ze,Be)}return ke(ue>L?-0:0,re.units[re.units.length-1])}function Bi(ue){let L={},re;return ue.length>0&&typeof ue[ue.length-1]=="object"?(L=ue[ue.length-1],re=Array.from(ue).slice(0,ue.length-1)):re=Array.from(ue),[L,re]}let wi;const Fu=new Map;class dn{constructor(L){const re=L.zone||ar.defaultZone;let fe=L.invalid||(Number.isNaN(L.ts)?new Te("invalid input"):null)||(re.isValid?null:bo(re));this.ts=wt(L.ts)?ar.now():L.ts;let xe=null,ke=null;if(!fe)if(L.old&&L.old.ts===this.ts&&L.old.zone.equals(re))[xe,ke]=[L.old.c,L.old.o];else{const Be=Kr(L.o)&&!L.old?L.o:re.offset(this.ts);xe=el(this.ts,Be),fe=Number.isNaN(xe.year)?new Te("invalid input"):null,xe=fe?null:xe,ke=fe?null:Be}this._zone=re,this.loc=L.loc||bt.create(),this.invalid=fe,this.weekData=null,this.localWeekData=null,this.c=xe,this.o=ke,this.isLuxonDateTime=!0}static now(){return new dn({})}static local(){const[L,re]=Bi(arguments),[fe,xe,ke,Qe,Be,Ze,Rt]=re;return ql({year:fe,month:xe,day:ke,hour:Qe,minute:Be,second:Ze,millisecond:Rt},L)}static utc(){const[L,re]=Bi(arguments),[fe,xe,ke,Qe,Be,Ze,Rt]=re;return L.zone=qt.utcInstance,ql({year:fe,month:xe,day:ke,hour:Qe,minute:Be,second:Ze,millisecond:Rt},L)}static fromJSDate(L,re={}){const fe=Ca(L)?L.valueOf():NaN;if(Number.isNaN(fe))return dn.invalid("invalid input");const xe=nn(re.zone,ar.defaultZone);return xe.isValid?new dn({ts:fe,zone:xe,loc:bt.fromObject(re)}):dn.invalid(bo(xe))}static fromMillis(L,re={}){if(Kr(L))return L<-wa||L>wa?dn.invalid("Timestamp out of range"):new dn({ts:L,zone:nn(re.zone,ar.defaultZone),loc:bt.fromObject(re)});throw new u(`fromMillis requires a numerical input, but received a ${typeof L} with value ${L}`)}static fromSeconds(L,re={}){if(Kr(L))return new dn({ts:L*1e3,zone:nn(re.zone,ar.defaultZone),loc:bt.fromObject(re)});throw new u("fromSeconds requires a numerical input")}static fromObject(L,re={}){L=L||{};const fe=nn(re.zone,ar.defaultZone);if(!fe.isValid)return dn.invalid(bo(fe));const xe=bt.fromObject(re),ke=Pu(L,Pp),{minDaysInFirstWeek:Qe,startOfWeek:Be}=Ln(ke,xe),Ze=ar.now(),Rt=wt(re.specificOffset)?fe.offset(Ze):re.specificOffset,Kt=!wt(ke.ordinal),zn=!wt(ke.year),ln=!wt(ke.month)||!wt(ke.day),mn=zn||ln,Ri=ke.weekYear||ke.weekNumber;if((mn||Kt)&&Ri)throw new i("Can't mix weekYear/weekNumber units with year/month/day or ordinals");if(ln&&Kt)throw new i("Can't mix ordinal dates with month/day");const za=Ri||ke.weekday&&!mn;let Se,El,Vl=el(Ze,Rt);za?(Se=Ef,El=yu,Vl=Lt(Vl,Qe,Be)):Kt?(Se=xh,El=Oh,Vl=_n(Vl)):(Se=zl,El=tl);let Iu=!1;for(const jl of Se){const vd=ke[jl];wt(vd)?Iu?ke[jl]=El[jl]:ke[jl]=Vl[jl]:Iu=!0}const qn=za?er(ke,Qe,Be):Kt?Di(ke):$i(ke),Ja=qn||hr(ke);if(Ja)return dn.invalid(Ja);const Ls=za?an(ke,Qe,Be):Kt?sn(ke):ke,[Ul,Lu]=wc(Ls,Rt,fe),Tc=new dn({ts:Ul,zone:fe,o:Lu,loc:xe});return ke.weekday&&mn&&L.weekday!==Tc.weekday?dn.invalid("mismatched weekday",`you can't specify both a weekday of ${ke.weekday} and a date of ${Tc.toISO()}`):Tc.isValid?Tc:dn.invalid(Tc.invalid)}static fromISO(L,re={}){const[fe,xe]=Qa(L);return vu(fe,xe,re,"ISO 8601",L)}static fromRFC2822(L,re={}){const[fe,xe]=gh(L);return vu(fe,xe,re,"RFC 2822",L)}static fromHTTP(L,re={}){const[fe,xe]=vh(L);return vu(fe,xe,re,"HTTP",re)}static fromFormat(L,re,fe={}){if(wt(L)||wt(re))throw new u("fromFormat requires an input string and a format");const{locale:xe=null,numberingSystem:ke=null}=fe,Qe=bt.fromOpts({locale:xe,numberingSystem:ke,defaultToEN:!0}),[Be,Ze,Rt,Kt]=bh(Qe,L,re);return Kt?dn.invalid(Kt):vu(Be,Ze,fe,`format ${re}`,L,Rt)}static fromString(L,re,fe={}){return dn.fromFormat(L,re,fe)}static fromSQL(L,re={}){const[fe,xe]=Tp(L);return vu(fe,xe,re,"SQL",L)}static invalid(L,re=null){if(!L)throw new u("need to specify a reason the DateTime is invalid");const fe=L instanceof Te?L:new Te(L,re);if(ar.throwOnInvalid)throw new t(fe);return new dn({invalid:fe})}static isDateTime(L){return L&&L.isLuxonDateTime||!1}static parseFormatForOpts(L,re={}){const fe=pa(L,bt.fromObject(re));return fe?fe.map(xe=>xe?xe.val:null).join(""):null}static expandFormat(L,re={}){return _f(Qn.parseFormat(L),bt.fromObject(re)).map(xe=>xe.val).join("")}static resetCache(){wi=void 0,Fu.clear()}get(L){return this[L]}get isValid(){return this.invalid===null}get invalidReason(){return this.invalid?this.invalid.reason:null}get invalidExplanation(){return this.invalid?this.invalid.explanation:null}get locale(){return this.isValid?this.loc.locale:null}get numberingSystem(){return this.isValid?this.loc.numberingSystem:null}get outputCalendar(){return this.isValid?this.loc.outputCalendar:null}get zone(){return this._zone}get zoneName(){return this.isValid?this.zone.name:null}get year(){return this.isValid?this.c.year:NaN}get quarter(){return this.isValid?Math.ceil(this.c.month/3):NaN}get month(){return this.isValid?this.c.month:NaN}get day(){return this.isValid?this.c.day:NaN}get hour(){return this.isValid?this.c.hour:NaN}get minute(){return this.isValid?this.c.minute:NaN}get second(){return this.isValid?this.c.second:NaN}get millisecond(){return this.isValid?this.c.millisecond:NaN}get weekYear(){return this.isValid?Is(this).weekYear:NaN}get weekNumber(){return this.isValid?Is(this).weekNumber:NaN}get weekday(){return this.isValid?Is(this).weekday:NaN}get isWeekend(){return this.isValid&&this.loc.getWeekendDays().includes(this.weekday)}get localWeekday(){return this.isValid?Wa(this).weekday:NaN}get localWeekNumber(){return this.isValid?Wa(this).weekNumber:NaN}get localWeekYear(){return this.isValid?Wa(this).weekYear:NaN}get ordinal(){return this.isValid?_n(this.c).ordinal:NaN}get monthShort(){return this.isValid?gu.months("short",{locObj:this.loc})[this.month-1]:null}get monthLong(){return this.isValid?gu.months("long",{locObj:this.loc})[this.month-1]:null}get weekdayShort(){return this.isValid?gu.weekdays("short",{locObj:this.loc})[this.weekday-1]:null}get weekdayLong(){return this.isValid?gu.weekdays("long",{locObj:this.loc})[this.weekday-1]:null}get offset(){return this.isValid?+this.o:NaN}get offsetNameShort(){return this.isValid?this.zone.offsetName(this.ts,{format:"short",locale:this.locale}):null}get offsetNameLong(){return this.isValid?this.zone.offsetName(this.ts,{format:"long",locale:this.locale}):null}get isOffsetFixed(){return this.isValid?this.zone.isUniversal:null}get isInDST(){return this.isOffsetFixed?!1:this.offset>this.set({month:1,day:1}).offset||this.offset>this.set({month:5}).offset}getPossibleOffsets(){if(!this.isValid||this.isOffsetFixed)return[this];const L=864e5,re=6e4,fe=Ya(this.c),xe=this.zone.offset(fe-L),ke=this.zone.offset(fe+L),Qe=this.zone.offset(fe-xe*re),Be=this.zone.offset(fe-ke*re);if(Qe===Be)return[this];const Ze=fe-Qe*re,Rt=fe-Be*re,Kt=el(Ze,Qe),zn=el(Rt,Be);return Kt.hour===zn.hour&&Kt.minute===zn.minute&&Kt.second===zn.second&&Kt.millisecond===zn.millisecond?[Ju(this,{ts:Ze}),Ju(this,{ts:Rt})]:[this]}get isInLeapYear(){return qo(this.year)}get daysInMonth(){return sh(this.year,this.month)}get daysInYear(){return this.isValid?Il(this.year):NaN}get weeksInWeekYear(){return this.isValid?$s(this.weekYear):NaN}get weeksInLocalWeekYear(){return this.isValid?$s(this.localWeekYear,this.loc.getMinDaysInFirstWeek(),this.loc.getStartOfWeek()):NaN}resolvedLocaleOptions(L={}){const{locale:re,numberingSystem:fe,calendar:xe}=Qn.create(this.loc.clone(L),L).resolvedOptions(this);return{locale:re,numberingSystem:fe,outputCalendar:xe}}toUTC(L=0,re={}){return this.setZone(qt.instance(L),re)}toLocal(){return this.setZone(ar.defaultZone)}setZone(L,{keepLocalTime:re=!1,keepCalendarTime:fe=!1}={}){if(L=nn(L,ar.defaultZone),L.equals(this.zone))return this;if(L.isValid){let xe=this.ts;if(re||fe){const ke=L.offset(this.ts),Qe=this.toObject();[xe]=wc(Qe,ke,L)}return Ju(this,{ts:xe,zone:L})}else return dn.invalid(bo(L))}reconfigure({locale:L,numberingSystem:re,outputCalendar:fe}={}){const xe=this.loc.clone({locale:L,numberingSystem:re,outputCalendar:fe});return Ju(this,{loc:xe})}setLocale(L){return this.reconfigure({locale:L})}set(L){if(!this.isValid)return this;const re=Pu(L,Pp),{minDaysInFirstWeek:fe,startOfWeek:xe}=Ln(re,this.loc),ke=!wt(re.weekYear)||!wt(re.weekNumber)||!wt(re.weekday),Qe=!wt(re.ordinal),Be=!wt(re.year),Ze=!wt(re.month)||!wt(re.day),Rt=Be||Ze,Kt=re.weekYear||re.weekNumber;if((Rt||Qe)&&Kt)throw new i("Can't mix weekYear/weekNumber units with year/month/day or ordinals");if(Ze&&Qe)throw new i("Can't mix ordinal dates with month/day");let zn;ke?zn=an({...Lt(this.c,fe,xe),...re},fe,xe):wt(re.ordinal)?(zn={...this.toObject(),...re},wt(re.day)&&(zn.day=Math.min(sh(zn.year,zn.month),zn.day))):zn=sn({..._n(this.c),...re});const[ln,mn]=wc(zn,this.o,this.zone);return Ju(this,{ts:ln,o:mn})}plus(L){if(!this.isValid)return this;const re=Kn.fromDurationLike(L);return Ju(this,Rc(this,re))}minus(L){if(!this.isValid)return this;const re=Kn.fromDurationLike(L).negate();return Ju(this,Rc(this,re))}startOf(L,{useLocaleWeeks:re=!1}={}){if(!this.isValid)return this;const fe={},xe=Kn.normalizeUnit(L);switch(xe){case"years":fe.month=1;case"quarters":case"months":fe.day=1;case"weeks":case"days":fe.hour=0;case"hours":fe.minute=0;case"minutes":fe.second=0;case"seconds":fe.millisecond=0;break}if(xe==="weeks")if(re){const ke=this.loc.getStartOfWeek(),{weekday:Qe}=this;Qe<ke&&(fe.weekNumber=this.weekNumber-1),fe.weekday=ke}else fe.weekday=1;if(xe==="quarters"){const ke=Math.ceil(this.month/3);fe.month=(ke-1)*3+1}return this.set(fe)}endOf(L,re){return this.isValid?this.plus({[L]:1}).startOf(L,re).minus(1):this}toFormat(L,re={}){return this.isValid?Qn.create(this.loc.redefaultToEN(re)).formatDateTimeFromString(this,L):Go}toLocaleString(L=d,re={}){return this.isValid?Qn.create(this.loc.clone(re),L).formatDateTime(this):Go}toLocaleParts(L={}){return this.isValid?Qn.create(this.loc.clone(L),L).formatDateTimeParts(this):[]}toISO({format:L="extended",suppressSeconds:re=!1,suppressMilliseconds:fe=!1,includeOffset:xe=!0,extendedZone:ke=!1,precision:Qe="milliseconds"}={}){if(!this.isValid)return null;Qe=bu(Qe);const Be=L==="extended";let Ze=aa(this,Be,Qe);return zl.indexOf(Qe)>=3&&(Ze+="T"),Ze+=Ji(this,Be,re,fe,xe,ke,Qe),Ze}toISODate({format:L="extended",precision:re="day"}={}){return this.isValid?aa(this,L==="extended",bu(re)):null}toISOWeekDate(){return gd(this,"kkkk-'W'WW-c")}toISOTime({suppressMilliseconds:L=!1,suppressSeconds:re=!1,includeOffset:fe=!0,includePrefix:xe=!1,extendedZone:ke=!1,format:Qe="extended",precision:Be="milliseconds"}={}){return this.isValid?(Be=bu(Be),(xe&&zl.indexOf(Be)>=3?"T":"")+Ji(this,Qe==="extended",re,L,fe,ke,Be)):null}toRFC2822(){return gd(this,"EEE, dd LLL yyyy HH:mm:ss ZZZ",!1)}toHTTP(){return gd(this.toUTC(),"EEE, dd LLL yyyy HH:mm:ss 'GMT'")}toSQLDate(){return this.isValid?aa(this,!0):null}toSQLTime({includeOffset:L=!0,includeZone:re=!1,includeOffsetSpace:fe=!0}={}){let xe="HH:mm:ss.SSS";return(re||L)&&(fe&&(xe+=" "),re?xe+="z":L&&(xe+="ZZ")),gd(this,xe,!0)}toSQL(L={}){return this.isValid?`${this.toSQLDate()} ${this.toSQLTime(L)}`:null}toString(){return this.isValid?this.toISO():Go}[Symbol.for("nodejs.util.inspect.custom")](){return this.isValid?`DateTime { ts: ${this.toISO()}, zone: ${this.zone.name}, locale: ${this.locale} }`:`DateTime { Invalid, reason: ${this.invalidReason} }`}valueOf(){return this.toMillis()}toMillis(){return this.isValid?this.ts:NaN}toSeconds(){return this.isValid?this.ts/1e3:NaN}toUnixInteger(){return this.isValid?Math.floor(this.ts/1e3):NaN}toJSON(){return this.toISO()}toBSON(){return this.toJSDate()}toObject(L={}){if(!this.isValid)return{};const re={...this.c};return L.includeConfig&&(re.outputCalendar=this.outputCalendar,re.numberingSystem=this.loc.numberingSystem,re.locale=this.loc.locale),re}toJSDate(){return new Date(this.isValid?this.ts:NaN)}diff(L,re="milliseconds",fe={}){if(!this.isValid||!L.isValid)return Kn.invalid("created by diffing an invalid DateTime");const xe={locale:this.locale,numberingSystem:this.numberingSystem,...fe},ke=Tu(re).map(Kn.normalizeUnit),Qe=L.valueOf()>this.valueOf(),Be=Qe?this:L,Ze=Qe?L:this,Rt=em(Be,Ze,ke,xe);return Qe?Rt.negate():Rt}diffNow(L="milliseconds",re={}){return this.diff(dn.now(),L,re)}until(L){return this.isValid?Ii.fromDateTimes(this,L):this}hasSame(L,re,fe){if(!this.isValid)return!1;const xe=L.valueOf(),ke=this.setZone(L.zone,{keepLocalTime:!0});return ke.startOf(re,fe)<=xe&&xe<=ke.endOf(re,fe)}equals(L){return this.isValid&&L.isValid&&this.valueOf()===L.valueOf()&&this.zone.equals(L.zone)&&this.loc.equals(L.loc)}toRelative(L={}){if(!this.isValid)return null;const re=L.base||dn.fromObject({},{zone:this.zone}),fe=L.padding?this<re?-L.padding:L.padding:0;let xe=["years","months","days","hours","minutes","seconds"],ke=L.unit;return Array.isArray(L.unit)&&(xe=L.unit,ke=void 0),Ac(re,this.plus(fe),{...L,numeric:"always",units:xe,unit:ke})}toRelativeCalendar(L={}){return this.isValid?Ac(L.base||dn.fromObject({},{zone:this.zone}),this,{...L,numeric:"auto",units:["years","months","days"],calendary:!0}):null}static min(...L){if(!L.every(dn.isDateTime))throw new u("min requires all arguments be DateTimes");return Nl(L,re=>re.valueOf(),Math.min)}static max(...L){if(!L.every(dn.isDateTime))throw new u("max requires all arguments be DateTimes");return Nl(L,re=>re.valueOf(),Math.max)}static fromFormatExplain(L,re,fe={}){const{locale:xe=null,numberingSystem:ke=null}=fe,Qe=bt.fromOpts({locale:xe,numberingSystem:ke,defaultToEN:!0});return Yo(Qe,L,re)}static fromStringExplain(L,re,fe={}){return dn.fromFormatExplain(L,re,fe)}static buildFormatParser(L,re={}){const{locale:fe=null,numberingSystem:xe=null}=re,ke=bt.fromOpts({locale:fe,numberingSystem:xe,defaultToEN:!0});return new Sr(ke,L)}static fromFormatParser(L,re,fe={}){if(wt(L)||wt(re))throw new u("fromFormatParser requires an input string and a format parser");const{locale:xe=null,numberingSystem:ke=null}=fe,Qe=bt.fromOpts({locale:xe,numberingSystem:ke,defaultToEN:!0});if(!Qe.equals(re.locale))throw new u(`fromFormatParser called with a locale of ${Qe}, but the format parser was created for ${re.locale}`);const{result:Be,zone:Ze,specificOffset:Rt,invalidReason:Kt}=re.explainFromTokens(L);return Kt?dn.invalid(Kt):vu(Be,Ze,fe,`format ${re.format}`,L,Rt)}static get DATE_SHORT(){return d}static get DATE_MED(){return h}static get DATE_MED_WITH_WEEKDAY(){return m}static get DATE_FULL(){return y}static get DATE_HUGE(){return v}static get TIME_SIMPLE(){return S}static get TIME_WITH_SECONDS(){return _}static get TIME_WITH_SHORT_OFFSET(){return E}static get TIME_WITH_LONG_OFFSET(){return $}static get TIME_24_SIMPLE(){return k}static get TIME_24_WITH_SECONDS(){return A}static get TIME_24_WITH_SHORT_OFFSET(){return w}static get TIME_24_WITH_LONG_OFFSET(){return R}static get DATETIME_SHORT(){return F}static get DATETIME_SHORT_WITH_SECONDS(){return I}static get DATETIME_MED(){return U}static get DATETIME_MED_WITH_SECONDS(){return q}static get DATETIME_MED_WITH_WEEKDAY(){return V}static get DATETIME_FULL(){return N}static get DATETIME_FULL_WITH_SECONDS(){return j}static get DATETIME_HUGE(){return W}static get DATETIME_HUGE_WITH_SECONDS(){return K}}function Nu(ue){if(dn.isDateTime(ue))return ue;if(ue&&ue.valueOf&&Kr(ue.valueOf()))return dn.fromJSDate(ue);if(ue&&typeof ue=="object")return dn.fromObject(ue);throw new u(`Unknown datetime argument: ${ue}, of type ${typeof ue}`)}const Ko="3.7.1";return Yc.DateTime=dn,Yc.Duration=Kn,Yc.FixedOffsetZone=qt,Yc.IANAZone=ie,Yc.Info=gu,Yc.Interval=Ii,Yc.InvalidZone=Xt,Yc.Settings=ar,Yc.SystemZone=H,Yc.VERSION=Ko,Yc.Zone=J,Yc}var dY;function Hw(){return dY||(dY=1,(function(e){var a,u,l,o,s,HM,d,gv;Object.defineProperty(e,"__esModule",{value:!0}),e.CronDate=e.DAYS_IN_MONTH=e.DateMathOp=e.TimeUnit=void 0;const t=Gze();var n;(function(y){y.Second="Second",y.Minute="Minute",y.Hour="Hour",y.Day="Day",y.Month="Month",y.Year="Year"})(n||(e.TimeUnit=n={}));var r;(function(y){y.Add="Add",y.Subtract="Subtract"})(r||(e.DateMathOp=r={})),e.DAYS_IN_MONTH=Object.freeze([31,29,31,30,31,30,31,31,30,31,30,31]);const m=class m{constructor(v,S){Yi(this,d);Yi(this,a);Yi(this,u,null);Yi(this,l,null);Yi(this,o,{add:{[n.Year]:this.addYear.bind(this),[n.Month]:this.addMonth.bind(this),[n.Day]:this.addDay.bind(this),[n.Hour]:this.addHour.bind(this),[n.Minute]:this.addMinute.bind(this),[n.Second]:this.addSecond.bind(this)},subtract:{[n.Year]:this.subtractYear.bind(this),[n.Month]:this.subtractMonth.bind(this),[n.Day]:this.subtractDay.bind(this),[n.Hour]:this.subtractHour.bind(this),[n.Minute]:this.subtractMinute.bind(this),[n.Second]:this.subtractSecond.bind(this)}});const _={zone:S};if(v?v instanceof m?(wn(this,a,Ot(v,a)),wn(this,u,Ot(v,u)),wn(this,l,Ot(v,l))):v instanceof Date?wn(this,a,t.DateTime.fromJSDate(v,_)):typeof v=="number"?wn(this,a,t.DateTime.fromMillis(v,_)):(wn(this,a,t.DateTime.fromISO(v,_)),Ot(this,a).isValid||wn(this,a,t.DateTime.fromRFC2822(v,_)),Ot(this,a).isValid||wn(this,a,t.DateTime.fromSQL(v,_)),Ot(this,a).isValid||wn(this,a,t.DateTime.fromFormat(v,"EEE, d MMM yyyy HH:mm:ss",_))):wn(this,a,t.DateTime.local()),!Ot(this,a).isValid)throw new Error(`CronDate: unhandled timestamp: ${v}`);S&&S!==Ot(this,a).zoneName&&wn(this,a,Ot(this,a).setZone(S))}get dstStart(){return Ot(this,u)}set dstStart(v){wn(this,u,v)}get dstEnd(){return Ot(this,l)}set dstEnd(v){wn(this,l,v)}addYear(){wn(this,a,Ot(this,a).plus({years:1}))}addMonth(){wn(this,a,Ot(this,a).plus({months:1}).startOf("month"))}addDay(){wn(this,a,Ot(this,a).plus({days:1}).startOf("day"))}addHour(){wn(this,a,Ot(this,a).plus({hours:1}).startOf("hour"))}addMinute(){wn(this,a,Ot(this,a).plus({minutes:1}).startOf("minute"))}addSecond(){wn(this,a,Ot(this,a).plus({seconds:1}))}subtractYear(){wn(this,a,Ot(this,a).minus({years:1}))}subtractMonth(){wn(this,a,Ot(this,a).minus({months:1}).endOf("month").startOf("second"))}subtractDay(){wn(this,a,Ot(this,a).minus({days:1}).endOf("day").startOf("second"))}subtractHour(){wn(this,a,Ot(this,a).minus({hours:1}).endOf("hour").startOf("second"))}subtractMinute(){wn(this,a,Ot(this,a).minus({minutes:1}).endOf("minute").startOf("second"))}subtractSecond(){wn(this,a,Ot(this,a).minus({seconds:1}))}addUnit(v){Ot(this,o).add[v]()}subtractUnit(v){Ot(this,o).subtract[v]()}invokeDateOperation(v,S){if(v===r.Add){this.addUnit(S);return}if(v===r.Subtract){this.subtractUnit(S);return}throw new Error(`Invalid verb: ${v}`)}getDate(){return Ot(this,a).day}getFullYear(){return Ot(this,a).year}getDay(){const v=Ot(this,a).weekday;return v===7?0:v}getMonth(){return Ot(this,a).month-1}getHours(){return Ot(this,a).hour}getMinutes(){return Ot(this,a).minute}getSeconds(){return Ot(this,a).second}getMilliseconds(){return Ot(this,a).millisecond}getTime(){return Ot(this,a).valueOf()}getUTCDate(){return Zn(this,d,gv).call(this).day}getUTCFullYear(){return Zn(this,d,gv).call(this).year}getUTCDay(){const v=Zn(this,d,gv).call(this).weekday;return v===7?0:v}getUTCMonth(){return Zn(this,d,gv).call(this).month-1}getUTCHours(){return Zn(this,d,gv).call(this).hour}getUTCMinutes(){return Zn(this,d,gv).call(this).minute}getUTCSeconds(){return Zn(this,d,gv).call(this).second}toISOString(){return Ot(this,a).toUTC().toISO()}toJSON(){return Ot(this,a).toJSON()}setDate(v){wn(this,a,Ot(this,a).set({day:v}))}setFullYear(v){wn(this,a,Ot(this,a).set({year:v}))}setDay(v){wn(this,a,Ot(this,a).set({weekday:v}))}setMonth(v){wn(this,a,Ot(this,a).set({month:v+1}))}setHours(v){wn(this,a,Ot(this,a).set({hour:v}))}setMinutes(v){wn(this,a,Ot(this,a).set({minute:v}))}setSeconds(v){wn(this,a,Ot(this,a).set({second:v}))}setMilliseconds(v){wn(this,a,Ot(this,a).set({millisecond:v}))}toString(){return this.toDate().toString()}toDate(){return Ot(this,a).toJSDate()}isLastDayOfMonth(){var _;const{day:v,month:S}=Ot(this,a);if(S===2){const E=Zn(_=m,s,HM).call(_,Ot(this,a).year);return v===e.DAYS_IN_MONTH[S-1]-(E?0:1)}return v===e.DAYS_IN_MONTH[S-1]}isLastWeekdayOfMonth(){var E;const{day:v,month:S}=Ot(this,a);let _;return S===2?_=e.DAYS_IN_MONTH[S-1]-(Zn(E=m,s,HM).call(E,Ot(this,a).year)?0:1):_=e.DAYS_IN_MONTH[S-1],v>_-7}applyDateOperation(v,S,_){if(S===n.Month||S===n.Day){this.invokeDateOperation(v,S);return}const E=this.getHours();this.invokeDateOperation(v,S);const $=this.getHours(),k=$-E;k===2?_!==24&&(this.dstStart=$):k===0&&this.getMinutes()===0&&this.getSeconds()===0&&_!==24&&(this.dstEnd=$)}};a=new WeakMap,u=new WeakMap,l=new WeakMap,o=new WeakMap,s=new WeakSet,HM=function(v){return v%4===0&&v%100!==0||v%400===0},d=new WeakSet,gv=function(){return Ot(this,a).toUTC()},Yi(m,s);let i=m;e.CronDate=i,e.default=i})(zP)),zP}var hY;function Kze(){if(hY)return b2;hY=1,Object.defineProperty(b2,"__esModule",{value:!0}),b2.CronMonth=void 0;const e=Hw(),t=M1(),n=1,r=12,i=Object.freeze([]);let a=class extends t.CronField{static get min(){return n}static get max(){return r}static get chars(){return i}static get daysInMonth(){return e.DAYS_IN_MONTH}constructor(l,o){super(l,o),this.validate()}get values(){return super.values}};return b2.CronMonth=a,b2}var O2={},pY;function Wze(){if(pY)return O2;pY=1,Object.defineProperty(O2,"__esModule",{value:!0}),O2.CronSecond=void 0;const e=M1(),t=0,n=59,r=Object.freeze([]);let i=class extends e.CronField{static get min(){return t}static get max(){return n}static get chars(){return r}constructor(u,l){super(u,l),this.validate()}get values(){return super.values}};return O2.CronSecond=i,O2}var mY;function DL(){return mY||(mY=1,(function(e){var t=Sb&&Sb.__createBinding||(Object.create?(function(r,i,a,u){u===void 0&&(u=a);var l=Object.getOwnPropertyDescriptor(i,a);(!l||("get"in l?!i.__esModule:l.writable||l.configurable))&&(l={enumerable:!0,get:function(){return i[a]}}),Object.defineProperty(r,u,l)}):(function(r,i,a,u){u===void 0&&(u=a),r[u]=i[a]})),n=Sb&&Sb.__exportStar||function(r,i){for(var a in r)a!=="default"&&!Object.prototype.hasOwnProperty.call(i,a)&&t(i,r,a)};Object.defineProperty(e,"__esModule",{value:!0}),n(jze(),e),n(Zze(),e),n(Hze(),e),n(M1(),e),n(Xze(),e),n(Yze(),e),n(Kze(),e),n(Wze(),e)})(Sb)),Sb}var gY;function nve(){var n,r,i,a,u,l,o,s,eve,tve;if(gY)return h2;gY=1,Object.defineProperty(h2,"__esModule",{value:!0}),h2.CronFieldCollection=void 0;const e=DL();let t=(s=class{constructor({second:m,minute:y,hour:v,dayOfMonth:S,month:_,dayOfWeek:E}){Yi(this,n);Yi(this,r);Yi(this,i);Yi(this,a);Yi(this,u);Yi(this,l);if(!m)throw new Error("Validation error, Field second is missing");if(!y)throw new Error("Validation error, Field minute is missing");if(!v)throw new Error("Validation error, Field hour is missing");if(!S)throw new Error("Validation error, Field dayOfMonth is missing");if(!_)throw new Error("Validation error, Field month is missing");if(!E)throw new Error("Validation error, Field dayOfWeek is missing");if(_.values.length===1&&!S.hasLastChar&&!(parseInt(S.values[0],10)<=e.CronMonth.daysInMonth[_.values[0]-1]))throw new Error("Invalid explicit day of month definition");wn(this,n,m),wn(this,r,y),wn(this,i,v),wn(this,u,_),wn(this,l,E),wn(this,a,S)}static from(m,y){return new s({second:this.resolveField(e.CronSecond,m.second,y.second),minute:this.resolveField(e.CronMinute,m.minute,y.minute),hour:this.resolveField(e.CronHour,m.hour,y.hour),dayOfMonth:this.resolveField(e.CronDayOfMonth,m.dayOfMonth,y.dayOfMonth),month:this.resolveField(e.CronMonth,m.month,y.month),dayOfWeek:this.resolveField(e.CronDayOfWeek,m.dayOfWeek,y.dayOfWeek)})}static resolveField(m,y,v){return v?v instanceof e.CronField?v:new m(v):y}get second(){return Ot(this,n)}get minute(){return Ot(this,r)}get hour(){return Ot(this,i)}get dayOfMonth(){return Ot(this,a)}get month(){return Ot(this,u)}get dayOfWeek(){return Ot(this,l)}static compactField(m){if(m.length===0)return[];const y=[];let v;return m.forEach((S,_,E)=>{if(v===void 0){v={start:S,count:1};return}const $=E[_-1]||v.start,k=E[_+1];if(S==="L"||S==="W"){y.push(v),y.push({start:S,count:1}),v=void 0;return}if(v.step===void 0&&k!==void 0){const A=S-$,w=k-S;if(A<=w){v={...v,count:2,end:S,step:A};return}v.step=1}S-(v.end??0)===v.step?(v.count++,v.end=S):(v.count===1?y.push({start:v.start,count:1}):v.count===2?(y.push({start:v.start,count:1}),y.push({start:v.end??$,count:1})):y.push(v),v={start:S,count:1})}),v&&y.push(v),y}stringifyField(m){var _;let y=m.max,v=m.values;if(m instanceof e.CronDayOfWeek){y=6;const E=Ot(this,l).values;v=E[E.length-1]===7?E.slice(0,-1):E}m instanceof e.CronDayOfMonth&&(y=Ot(this,u).values.length===1?e.CronMonth.daysInMonth[Ot(this,u).values[0]-1]:m.max);const S=s.compactField(v);if(S.length===1){const E=Zn(_=s,o,eve).call(_,m,S[0],y);if(E)return E}return S.map(E=>{var k;const $=E.count===1?E.start.toString():Zn(k=s,o,tve).call(k,E,y);return m instanceof e.CronDayOfWeek&&m.nthDay>0?`${$}#${m.nthDay}`:$}).join(",")}stringify(m=!1){const y=[];return m&&y.push(this.stringifyField(Ot(this,n))),y.push(this.stringifyField(Ot(this,r)),this.stringifyField(Ot(this,i)),this.stringifyField(Ot(this,a)),this.stringifyField(Ot(this,u)),this.stringifyField(Ot(this,l))),y.join(" ")}serialize(){return{second:Ot(this,n).serialize(),minute:Ot(this,r).serialize(),hour:Ot(this,i).serialize(),dayOfMonth:Ot(this,a).serialize(),month:Ot(this,u).serialize(),dayOfWeek:Ot(this,l).serialize()}}},n=new WeakMap,r=new WeakMap,i=new WeakMap,a=new WeakMap,u=new WeakMap,l=new WeakMap,o=new WeakSet,eve=function(m,y,v){const S=y.step;return S?S===1&&y.start===m.min&&y.end&&y.end>=v?m.hasQuestionMarkChar?"?":"*":S!==1&&y.start===m.min&&y.end&&y.end>=v-S+1?`*/${S}`:null:null},tve=function(m,y){const v=m.step;if(v===1)return`${m.start}-${m.end}`;const S=m.start===0?m.count-1:m.count;if(!v)throw new Error("Unexpected range step");if(!m.end)throw new Error("Unexpected range end");if(v*S>m.end){const _=($,k)=>{if(typeof m.start!="number")throw new Error("Unexpected range start");return k%v===0?m.start+k:null};if(typeof m.start!="number")throw new Error("Unexpected range start");const E={length:m.end-m.start+1};return Array.from(E,_).filter($=>$!==null).join(",")}return m.end===y-v+1?`${m.start}/${v}`:`${m.start}-${m.end}/${v}`},Yi(s,o),s);return h2.CronFieldCollection=t,h2}var CO={},vY;function ave(){var r,i,a,u,l,o,s,c,vv,rve,m,XM,ive,QO;if(vY)return CO;vY=1,Object.defineProperty(CO,"__esModule",{value:!0}),CO.CronExpression=void 0;const e=Hw(),t=1e4;let n=(c=class{constructor(E,$){Yi(this,m);Yi(this,r);Yi(this,i);Yi(this,a);Yi(this,u);Yi(this,l);Yi(this,o);wn(this,r,$),wn(this,i,$.tz),wn(this,a,new e.CronDate($.currentDate,Ot(this,i))),wn(this,u,$.startDate?new e.CronDate($.startDate,Ot(this,i)):null),wn(this,l,$.endDate?new e.CronDate($.endDate,Ot(this,i)):null),wn(this,o,E)}get fields(){return Ot(this,o)}static fieldsToExpression(E,$){return new c(E,$||{})}next(){return Zn(this,m,QO).call(this)}prev(){return Zn(this,m,QO).call(this,!0)}hasNext(){const E=Ot(this,a);try{return Zn(this,m,QO).call(this),!0}catch{return!1}finally{wn(this,a,E)}}hasPrev(){const E=Ot(this,a);try{return Zn(this,m,QO).call(this,!0),!0}catch{return!1}finally{wn(this,a,E)}}take(E){const $=[];if(E>=0)for(let k=0;k<E;k++)try{$.push(this.next())}catch{return $}else for(let k=0;k>E;k--)try{$.push(this.prev())}catch{return $}return $}reset(E){wn(this,a,new e.CronDate(E||Ot(this,r).currentDate))}stringify(E=!1){return Ot(this,o).stringify(E)}includesDate(E){const{second:$,minute:k,hour:A,month:w}=Ot(this,o),R=new e.CronDate(E,Ot(this,i));return!(!$.values.includes(R.getSeconds())||!k.values.includes(R.getMinutes())||!A.values.includes(R.getHours())||!w.values.includes(R.getMonth()+1)||!Zn(this,m,XM).call(this,R)||Ot(this,o).dayOfWeek.nthDay>0&&Math.ceil(R.getDate()/7)!==Ot(this,o).dayOfWeek.nthDay)}toString(){return Ot(this,r).expression||this.stringify(!0)}[Symbol.iterator](){return{next:()=>({value:Zn(this,m,QO).call(this),done:!this.hasNext()})}}},r=new WeakMap,i=new WeakMap,a=new WeakMap,u=new WeakMap,l=new WeakMap,o=new WeakMap,s=new WeakSet,vv=function(E,$){return $.some(k=>k===E)},rve=function(E,$){const k=$.isLastWeekdayOfMonth();return E.some(A=>{const w=parseInt(A.toString().charAt(0),10)%7;if(Number.isNaN(w))throw new Error(`Invalid last weekday of the month expression: ${A}`);return $.getDay()===w&&k})},m=new WeakSet,XM=function(E){var I,U,q;const $=Ot(this,o).dayOfMonth.isWildcard,k=!$,A=Ot(this,o).dayOfWeek.isWildcard,w=!A,R=Zn(I=c,s,vv).call(I,E.getDate(),Ot(this,o).dayOfMonth.values)||Ot(this,o).dayOfMonth.hasLastChar&&E.isLastDayOfMonth(),F=Zn(U=c,s,vv).call(U,E.getDay(),Ot(this,o).dayOfWeek.values)||Ot(this,o).dayOfWeek.hasLastChar&&Zn(q=c,s,rve).call(q,Ot(this,o).dayOfWeek.values,E);return!!(k&&w&&(R||F)||R&&!w||$&&!A&&F)},ive=function(E,$,k){var I,U;const A=E.getHours(),w=Zn(I=c,s,vv).call(I,A,Ot(this,o).hour.values),R=E.dstStart===A,F=E.dstEnd===A;return!w&&!R?(E.dstStart=null,E.applyDateOperation($,e.TimeUnit.Hour,Ot(this,o).hour.values.length),!1):R&&!Zn(U=c,s,vv).call(U,A-1,Ot(this,o).hour.values)?(E.invokeDateOperation($,e.TimeUnit.Hour),!1):F&&!k?(E.dstEnd=null,E.applyDateOperation(e.DateMathOp.Add,e.TimeUnit.Hour,Ot(this,o).hour.values.length),!1):!0},QO=function(E=!1){var I,U,q;const $=E?e.DateMathOp.Subtract:e.DateMathOp.Add,k=new e.CronDate(Ot(this,a)),A=Ot(this,u),w=Ot(this,l),R=k.getTime();let F=0;for(;F++<t;){if(F>t)throw new Error("Invalid expression, loop limit exceeded");if(E&&A&&A.getTime()>k.getTime())throw new Error("Out of the timespan range");if(!E&&w&&k.getTime()>w.getTime())throw new Error("Out of the timespan range");if(!Zn(this,m,XM).call(this,k)){k.applyDateOperation($,e.TimeUnit.Day,Ot(this,o).hour.values.length);continue}if(!(Ot(this,o).dayOfWeek.nthDay<=0||Math.ceil(k.getDate()/7)===Ot(this,o).dayOfWeek.nthDay)){k.applyDateOperation($,e.TimeUnit.Day,Ot(this,o).hour.values.length);continue}if(!Zn(I=c,s,vv).call(I,k.getMonth()+1,Ot(this,o).month.values)){k.applyDateOperation($,e.TimeUnit.Month,Ot(this,o).hour.values.length);continue}if(Zn(this,m,ive).call(this,k,$,E)){if(!Zn(U=c,s,vv).call(U,k.getMinutes(),Ot(this,o).minute.values)){k.applyDateOperation($,e.TimeUnit.Minute,Ot(this,o).hour.values.length);continue}if(!Zn(q=c,s,vv).call(q,k.getSeconds(),Ot(this,o).second.values)){k.applyDateOperation($,e.TimeUnit.Second,Ot(this,o).hour.values.length);continue}if(R===k.getTime()){($==="Add"||k.getMilliseconds()===0)&&k.applyDateOperation($,e.TimeUnit.Second,Ot(this,o).hour.values.length);continue}break}}return k.getMilliseconds()!==0&&k.setMilliseconds(0),wn(this,a,k),k},Yi(c,s),c);return CO.CronExpression=n,CO.default=n,CO}var AE={},yY;function Jze(){if(yY)return AE;yY=1,Object.defineProperty(AE,"__esModule",{value:!0}),AE.seededRandom=n;function e(r){let i=2166136261;for(let a=0;a<r.length;a++)i^=r.charCodeAt(a),i=Math.imul(i,16777619);return()=>i>>>0}function t(r){return()=>{let i=r+=1831565813;return i=Math.imul(i^i>>>15,i|1),i^=i+Math.imul(i^i>>>7,i|61),((i^i>>>14)>>>0)/4294967296}}function n(r){const i=r?e(r)():Math.floor(Math.random()*1e10);return t(i)}return AE}var bY;function GM(){var c,uve,Cb,lve,ove,sve,cve,fve,dve,hve,YM,pve,mve;if(bY)return If;bY=1,Object.defineProperty(If,"__esModule",{value:!0}),If.CronExpressionParser=If.DayOfWeek=If.Months=If.CronUnit=If.PredefinedExpressions=void 0;const e=nve(),t=Hw(),n=ave(),r=Jze(),i=DL();var a;(function(F){F["@yearly"]="0 0 0 1 1 *",F["@annually"]="0 0 0 1 1 *",F["@monthly"]="0 0 0 1 * *",F["@weekly"]="0 0 0 * * 0",F["@daily"]="0 0 0 * * *",F["@hourly"]="0 0 * * * *",F["@minutely"]="0 * * * * *",F["@secondly"]="* * * * * *",F["@weekdays"]="0 0 0 * * 1-5",F["@weekends"]="0 0 0 * * 0,6"})(a||(If.PredefinedExpressions=a={}));var u;(function(F){F.Second="Second",F.Minute="Minute",F.Hour="Hour",F.DayOfMonth="DayOfMonth",F.Month="Month",F.DayOfWeek="DayOfWeek"})(u||(If.CronUnit=u={}));var l;(function(F){F[F.jan=1]="jan",F[F.feb=2]="feb",F[F.mar=3]="mar",F[F.apr=4]="apr",F[F.may=5]="may",F[F.jun=6]="jun",F[F.jul=7]="jul",F[F.aug=8]="aug",F[F.sep=9]="sep",F[F.oct=10]="oct",F[F.nov=11]="nov",F[F.dec=12]="dec"})(l||(If.Months=l={}));var o;(function(F){F[F.sun=0]="sun",F[F.mon=1]="mon",F[F.tue=2]="tue",F[F.wed=3]="wed",F[F.thu=4]="thu",F[F.fri=5]="fri",F[F.sat=6]="sat"})(o||(If.DayOfWeek=o={}));const R=class R{static parse(I,U={}){var ne,oe,ie,de,ge,ye,me,Ee;const{strict:q=!1}=U,V=U.currentDate||new t.CronDate,N=(0,r.seededRandom)(U.hashSeed);I=a[I]||I;const j=Zn(ne=R,c,uve).call(ne,I,q);if(!(j.dayOfMonth==="*"||j.dayOfWeek==="*"||!q))throw new Error("Cannot use both dayOfMonth and dayOfWeek together in strict mode!");const W=Zn(oe=R,c,Cb).call(oe,u.Second,j.second,i.CronSecond.constraints,N),K=Zn(ie=R,c,Cb).call(ie,u.Minute,j.minute,i.CronMinute.constraints,N),J=Zn(de=R,c,Cb).call(de,u.Hour,j.hour,i.CronHour.constraints,N),te=Zn(ge=R,c,Cb).call(ge,u.Month,j.month,i.CronMonth.constraints,N),H=Zn(ye=R,c,Cb).call(ye,u.DayOfMonth,j.dayOfMonth,i.CronDayOfMonth.constraints,N),{dayOfWeek:M,nthDayOfWeek:G}=Zn(me=R,c,pve).call(me,j.dayOfWeek),X=Zn(Ee=R,c,Cb).call(Ee,u.DayOfWeek,M,i.CronDayOfWeek.constraints,N),B=new e.CronFieldCollection({second:new i.CronSecond(W,{rawValue:j.second}),minute:new i.CronMinute(K,{rawValue:j.minute}),hour:new i.CronHour(J,{rawValue:j.hour}),dayOfMonth:new i.CronDayOfMonth(H,{rawValue:j.dayOfMonth}),month:new i.CronMonth(te,{rawValue:j.month}),dayOfWeek:new i.CronDayOfWeek(X,{rawValue:j.dayOfWeek,nthDayOfWeek:G})});return new n.CronExpression(B,{...U,expression:I,currentDate:V})}};c=new WeakSet,uve=function(I,U){if(U&&!I.length)throw new Error("Invalid cron expression");I=I||"0 * * * * *";const q=I.trim().split(/\s+/);if(U&&q.length<6)throw new Error("Invalid cron expression, expected 6 fields");if(q.length>6)throw new Error("Invalid cron expression, too many fields");const V=["*","*","*","*","*","0"];q.length<V.length&&q.unshift(...V.slice(q.length));const[N,j,W,K,J,te]=q;return{second:N,minute:j,hour:W,dayOfMonth:K,month:J,dayOfWeek:te}},Cb=function(I,U,q,V){if((I===u.Month||I===u.DayOfWeek)&&(U=U.replace(/[a-z]{3}/gi,N=>{N=N.toLowerCase();const j=l[N]||o[N];if(j===void 0)throw new Error(`Validation error, cannot resolve alias "${N}"`);return j.toString()})),!q.validChars.test(U))throw new Error(`Invalid characters, got value: ${U}`);return U=Zn(this,c,lve).call(this,U,q),U=Zn(this,c,ove).call(this,U,q,V),Zn(this,c,sve).call(this,I,U,q)},lve=function(I,U){return I.replace(/[*?]/g,U.min+"-"+U.max)},ove=function(I,U,q){const V=q();return I.replace(/H(?:\((\d+)-(\d+)\))?(?:\/(\d+))?/g,(N,j,W,K)=>{if(j&&W&&K){const J=parseInt(j,10),te=parseInt(W,10),H=parseInt(K,10);if(J>te)throw new Error(`Invalid range: ${J}-${te}, min > max`);if(H<=0)throw new Error(`Invalid step: ${H}, must be positive`);const M=Math.max(J,U.min),G=Math.floor(V*H),X=[];for(let B=Math.floor(M/H)*H+G;B<=te;B+=H)B>=M&&X.push(B);return X.join(",")}else if(j&&W){const J=parseInt(j,10),te=parseInt(W,10);if(J>te)throw new Error(`Invalid range: ${J}-${te}, min > max`);return String(Math.floor(V*(te-J+1))+J)}else if(K){const J=parseInt(K,10);if(J<=0)throw new Error(`Invalid step: ${J}, must be positive`);const te=Math.floor(V*J),H=[];for(let M=Math.floor(U.min/J)*J+te;M<=U.max;M+=J)M>=U.min&&H.push(M);return H.join(",")}else return String(Math.floor(V*(U.max-U.min+1)+U.min))})},sve=function(I,U,q){const V=[];function N(W,K){var J;if(Array.isArray(W))V.push(...W);else if(Zn(J=R,c,mve).call(J,K,W))V.push(W);else{const te=parseInt(W.toString(),10);if(!(te>=K.min&&te<=K.max))throw new Error(`Constraint error, got value ${W} expected range ${K.min}-${K.max}`);V.push(I===u.DayOfWeek?te%7:W)}}return U.split(",").forEach(W=>{var K;if(!(W.length>0))throw new Error("Invalid list value format");N(Zn(K=R,c,cve).call(K,I,W,q),q)}),V},cve=function(I,U,q){var N,j;const V=U.split("/");if(V.length>2)throw new Error(`Invalid repeat: ${U}`);return V.length===2?(isNaN(parseInt(V[0],10))||(V[0]=`${V[0]}-${q.max}`),Zn(N=R,c,YM).call(N,I,V[0],parseInt(V[1],10),q)):Zn(j=R,c,YM).call(j,I,U,1,q)},fve=function(I,U,q){if(!(!isNaN(I)&&!isNaN(U)&&I>=q.min&&U<=q.max))throw new Error(`Constraint error, got range ${I}-${U} expected range ${q.min}-${q.max}`);if(I>U)throw new Error(`Invalid range: ${I}-${U}, min(${I}) > max(${U})`)},dve=function(I){if(!(!isNaN(I)&&I>0))throw new Error(`Constraint error, cannot repeat at every ${I} time.`)},hve=function(I,U,q,V){const N=[];I===u.DayOfWeek&&q%7===0&&N.push(0);for(let j=U;j<=q;j+=V)N.indexOf(j)===-1&&N.push(j);return N},YM=function(I,U,q,V){const N=U.split("-");if(N.length<=1)return isNaN(+U)?U:+U;const[j,W]=N.map(K=>parseInt(K,10));return Zn(this,c,fve).call(this,j,W,V),Zn(this,c,dve).call(this,q),Zn(this,c,hve).call(this,I,j,W,q)},pve=function(I){const U=I.split("#");if(U.length<=1)return{dayOfWeek:U[0]};const q=+U[U.length-1],V=I.match(/([,-/])/);if(V!==null)throw new Error(`Constraint error, invalid dayOfWeek \`#\` and \`${V==null?void 0:V[0]}\` special characters are incompatible`);if(!(U.length<=2&&!isNaN(q)&&q>=1&&q<=5))throw new Error("Constraint error, invalid dayOfWeek occurrence number (#)");return{dayOfWeek:U[0],nthDayOfWeek:q}},mve=function(I,U){return I.chars.some(q=>U.toString().includes(q))},Yi(R,c);let s=R;return If.CronExpressionParser=s,If}var Xh={};const eqe={},tqe=Object.freeze(Object.defineProperty({__proto__:null,default:eqe},Symbol.toStringTag,{value:"Module"})),OY=zOe(tqe);var xY;function nqe(){var a,u,KM,gve;if(xY)return Xh;xY=1;var e=Xh&&Xh.__createBinding||(Object.create?(function(s,c,d,h){h===void 0&&(h=d);var m=Object.getOwnPropertyDescriptor(c,d);(!m||("get"in m?!c.__esModule:m.writable||m.configurable))&&(m={enumerable:!0,get:function(){return c[d]}}),Object.defineProperty(s,h,m)}):(function(s,c,d,h){h===void 0&&(h=d),s[h]=c[d]})),t=Xh&&Xh.__setModuleDefault||(Object.create?(function(s,c){Object.defineProperty(s,"default",{enumerable:!0,value:c})}):function(s,c){s.default=c}),n=Xh&&Xh.__importStar||(function(){var s=function(c){return s=Object.getOwnPropertyNames||function(d){var h=[];for(var m in d)Object.prototype.hasOwnProperty.call(d,m)&&(h[h.length]=m);return h},s(c)};return function(c){if(c&&c.__esModule)return c;var d={};if(c!=null)for(var h=s(c),m=0;m<h.length;m++)h[m]!=="default"&&e(d,c,h[m]);return t(d,c),d}})();Object.defineProperty(Xh,"__esModule",{value:!0}),Xh.CronFileParser=void 0;const r=GM();let i=(u=class{static async parseFile(c){var m;const{readFile:d}=await Promise.resolve().then(()=>n(OY)),h=await d(c,"utf8");return Zn(m=u,a,KM).call(m,h)}static parseFileSync(c){var m;const{readFileSync:d}=OY,h=d(c,"utf8");return Zn(m=u,a,KM).call(m,h)}},a=new WeakSet,KM=function(c){var m;const d=c.split(`
`),h={variables:{},expressions:[],errors:{}};for(const y of d){const v=y.trim();if(v.length===0||v.startsWith("#"))continue;const S=v.match(/^(.*)=(.*)$/);if(S){const[,_,E]=S;h.variables[_]=E.replace(/["']/g,"");continue}try{const _=Zn(m=u,a,gve).call(m,v);h.expressions.push(_.interval)}catch(_){h.errors[v]=_}}return h},gve=function(c){const d=c.split(" ");return{interval:r.CronExpressionParser.parse(d.slice(0,5).join(" ")),command:d.slice(5,d.length)}},Yi(u,a),u);return Xh.CronFileParser=i,Xh}var SY;function rqe(){return SY||(SY=1,(function(e){var t=xb&&xb.__createBinding||(Object.create?(function(s,c,d,h){h===void 0&&(h=d);var m=Object.getOwnPropertyDescriptor(c,d);(!m||("get"in m?!c.__esModule:m.writable||m.configurable))&&(m={enumerable:!0,get:function(){return c[d]}}),Object.defineProperty(s,h,m)}):(function(s,c,d,h){h===void 0&&(h=d),s[h]=c[d]})),n=xb&&xb.__exportStar||function(s,c){for(var d in s)d!=="default"&&!Object.prototype.hasOwnProperty.call(c,d)&&t(c,s,d)};Object.defineProperty(e,"__esModule",{value:!0}),e.CronFileParser=e.CronExpressionParser=e.CronExpression=e.CronFieldCollection=e.CronDate=void 0;const r=GM();var i=Hw();Object.defineProperty(e,"CronDate",{enumerable:!0,get:function(){return i.CronDate}});var a=nve();Object.defineProperty(e,"CronFieldCollection",{enumerable:!0,get:function(){return a.CronFieldCollection}});var u=ave();Object.defineProperty(e,"CronExpression",{enumerable:!0,get:function(){return u.CronExpression}});var l=GM();Object.defineProperty(e,"CronExpressionParser",{enumerable:!0,get:function(){return l.CronExpressionParser}});var o=nqe();Object.defineProperty(e,"CronFileParser",{enumerable:!0,get:function(){return o.CronFileParser}}),n(DL(),e),e.default=r.CronExpressionParser})(xb)),xb}var iqe=rqe();const Fut=da(iqe);var qP={exports:{}},_Y;function aqe(){return _Y||(_Y=1,(function(e,t){(function(r,i){e.exports=i()})(globalThis,()=>(()=>{var n={949:((u,l,o)=>{Object.defineProperty(l,"__esModule",{value:!0}),l.CronParser=void 0;var s=o(515),c=(function(){function d(h,m,y){m===void 0&&(m=!0),y===void 0&&(y=!1),this.expression=h,this.dayOfWeekStartIndexZero=m,this.monthStartIndexZero=y}return d.prototype.parse=function(){var h,m,y=(h=this.expression)!==null&&h!==void 0?h:"";if(y.startsWith("@")){var v=this.parseSpecial(this.expression);m=this.extractParts(v)}else m=this.extractParts(this.expression);return this.normalize(m),this.validate(m),m},d.prototype.parseSpecial=function(h){var m={"@yearly":"0 0 1 1 *","@annually":"0 0 1 1 *","@monthly":"0 0 1 * *","@weekly":"0 0 * * 0","@daily":"0 0 * * *","@midnight":"0 0 * * *","@hourly":"0 * * * *"},y=m[h];if(!y)throw new Error("Unknown special expression.");return y},d.prototype.extractParts=function(h){if(!this.expression)throw new Error("cron expression is empty");for(var m=h.trim().split(/[ ]+/),y=0;y<m.length;y++)if(m[y].includes(",")){var v=m[y].split(",").map(function(_){return _.trim()}).filter(function(_){return _!==""}).map(function(_){return isNaN(Number(_))?_:Number(_)}).filter(function(_){return _!==null&&_!==""});v.length===0&&v.push("*"),v.sort(function(_,E){return _!==null&&E!==null?_-E:0}),m[y]=v.map(function(_){return _!==null?_.toString():""}).join(",")}if(m.length<5)throw new Error("Expression has only ".concat(m.length," part").concat(m.length==1?"":"s",". At least 5 parts are required."));if(m.length==5)m.unshift(""),m.push("");else if(m.length==6){var S=/\d{4}$/.test(m[5])||m[4]=="?"||m[2]=="?";S?m.unshift(""):m.push("")}else if(m.length>7)throw new Error("Expression has ".concat(m.length," parts; too many!"));return m},d.prototype.normalize=function(h){var m=this;if(h[3]=h[3].replace("?","*"),h[5]=h[5].replace("?","*"),h[2]=h[2].replace("?","*"),h[0].indexOf("0/")==0&&(h[0]=h[0].replace("0/","*/")),h[1].indexOf("0/")==0&&(h[1]=h[1].replace("0/","*/")),h[2].indexOf("0/")==0&&(h[2]=h[2].replace("0/","*/")),h[3].indexOf("1/")==0&&(h[3]=h[3].replace("1/","*/")),h[4].indexOf("1/")==0&&(h[4]=h[4].replace("1/","*/")),h[6].indexOf("1/")==0&&(h[6]=h[6].replace("1/","*/")),h[5]=h[5].replace(/(^\d)|([^#/\s]\d)/g,function(A){var w=A.replace(/\D/,""),R=w;return m.dayOfWeekStartIndexZero?w=="7"&&(R="0"):R=(parseInt(w)-1).toString(),A.replace(w,R)}),h[5]=="L"&&(h[5]="6"),h[3]=="?"&&(h[3]="*"),h[3].indexOf("W")>-1&&(h[3].indexOf(",")>-1||h[3].indexOf("-")>-1))throw new Error("The 'W' character can be specified only when the day-of-month is a single day, not a range or list of days.");var y={SUN:0,MON:1,TUE:2,WED:3,THU:4,FRI:5,SAT:6};for(var v in y)h[5]=h[5].replace(new RegExp(v,"gi"),y[v].toString());h[4]=h[4].replace(/(^\d{1,2})|([^#/\s]\d{1,2})/g,function(A){var w=A.replace(/\D/,""),R=w;return m.monthStartIndexZero&&(R=(parseInt(w)+1).toString()),A.replace(w,R)});var S={JAN:1,FEB:2,MAR:3,APR:4,MAY:5,JUN:6,JUL:7,AUG:8,SEP:9,OCT:10,NOV:11,DEC:12};for(var _ in S)h[4]=h[4].replace(new RegExp(_,"gi"),S[_].toString());h[0]=="0"&&(h[0]=""),!/\*|\-|\,|\//.test(h[2])&&(/\*|\//.test(h[1])||/\*|\//.test(h[0]))&&(h[2]+="-".concat(h[2]));for(var E=0;E<h.length;E++)if(h[E].indexOf(",")!=-1&&(h[E]=h[E].split(",").filter(function(A){return A!==""}).join(",")||"*"),h[E]=="*/1"&&(h[E]="*"),h[E].indexOf("/")>-1&&!/^\*|\-|\,/.test(h[E])){var $=null;switch(E){case 4:$="12";break;case 5:$="6";break;case 6:$="9999";break;default:$=null;break}if($!==null){var k=h[E].split("/");h[E]="".concat(k[0],"-").concat($,"/").concat(k[1])}}},d.prototype.validate=function(h){var m="0-9,\\-*/";this.validateOnlyExpectedCharactersFound(h[0],m),this.validateOnlyExpectedCharactersFound(h[1],m),this.validateOnlyExpectedCharactersFound(h[2],m),this.validateOnlyExpectedCharactersFound(h[3],"0-9,\\-*/LW"),this.validateOnlyExpectedCharactersFound(h[4],m),this.validateOnlyExpectedCharactersFound(h[5],"0-9,\\-*/L#"),this.validateOnlyExpectedCharactersFound(h[6],m),this.validateAnyRanges(h)},d.prototype.validateAnyRanges=function(h){s.default.secondRange(h[0]),s.default.minuteRange(h[1]),s.default.hourRange(h[2]),s.default.dayOfMonthRange(h[3]),s.default.monthRange(h[4],this.monthStartIndexZero),s.default.dayOfWeekRange(h[5],this.dayOfWeekStartIndexZero)},d.prototype.validateOnlyExpectedCharactersFound=function(h,m){var y=h.match(new RegExp("[^".concat(m,"]+"),"gi"));if(y&&y.length)throw new Error("Expression contains invalid values: '".concat(y.toString(),"'"))},d})();l.CronParser=c}),333:((u,l,o)=>{Object.defineProperty(l,"__esModule",{value:!0}),l.ExpressionDescriptor=void 0;var s=o(823),c=o(949),d=(function(){function h(m,y){if(this.expression=m,this.options=y,this.expressionParts=new Array(5),!this.options.locale&&h.defaultLocale&&(this.options.locale=h.defaultLocale),!h.locales[this.options.locale]){var v=Object.keys(h.locales)[0];console.warn("Locale '".concat(this.options.locale,"' could not be found; falling back to '").concat(v,"'.")),this.options.locale=v}this.i18n=h.locales[this.options.locale],y.use24HourTimeFormat===void 0&&(y.use24HourTimeFormat=this.i18n.use24HourTimeFormatByDefault())}return h.toString=function(m,y){var v=y===void 0?{}:y,S=v.throwExceptionOnParseError,_=S===void 0?!0:S,E=v.verbose,$=E===void 0?!1:E,k=v.dayOfWeekStartIndexZero,A=k===void 0?!0:k,w=v.monthStartIndexZero,R=w===void 0?!1:w,F=v.use24HourTimeFormat,I=v.locale,U=I===void 0?null:I,q=v.tzOffset,V=q===void 0?0:q,N={throwExceptionOnParseError:_,verbose:$,dayOfWeekStartIndexZero:A,monthStartIndexZero:R,use24HourTimeFormat:F,locale:U,tzOffset:V};N.tzOffset&&console.warn("'tzOffset' option has been deprecated and will be removed in a future release.");var j=new h(m,N);return j.getFullDescription()},h.initialize=function(m,y){y===void 0&&(y="en"),h.specialCharacters=["/","-",",","*"],h.defaultLocale=y,m.load(h.locales)},h.prototype.getFullDescription=function(){var m="";try{var y=new c.CronParser(this.expression,this.options.dayOfWeekStartIndexZero,this.options.monthStartIndexZero);this.expressionParts=y.parse();var v=this.getTimeOfDayDescription(),S=this.getDayOfMonthDescription(),_=this.getMonthDescription(),E=this.getDayOfWeekDescription(),$=this.getYearDescription();m+=v+S+E+_+$,m=this.transformVerbosity(m,!!this.options.verbose),m=m.charAt(0).toLocaleUpperCase()+m.substr(1)}catch(k){if(!this.options.throwExceptionOnParseError)m=this.i18n.anErrorOccuredWhenGeneratingTheExpressionD();else throw"".concat(k)}return m},h.prototype.getTimeOfDayDescription=function(){var m=this.expressionParts[0],y=this.expressionParts[1],v=this.expressionParts[2],S="";if(!s.StringUtilities.containsAny(y,h.specialCharacters)&&!s.StringUtilities.containsAny(v,h.specialCharacters)&&!s.StringUtilities.containsAny(m,h.specialCharacters))S+=this.i18n.atSpace()+this.formatTime(v,y,m);else if(!m&&y.indexOf("-")>-1&&!(y.indexOf(",")>-1)&&!(y.indexOf("/")>-1)&&!s.StringUtilities.containsAny(v,h.specialCharacters)){var _=y.split("-");S+=s.StringUtilities.format(this.i18n.everyMinuteBetweenX0AndX1(),this.formatTime(v,_[0],""),this.formatTime(v,_[1],""))}else if(!m&&v.indexOf(",")>-1&&v.indexOf("-")==-1&&v.indexOf("/")==-1&&!s.StringUtilities.containsAny(y,h.specialCharacters)){var E=v.split(",");S+=this.i18n.at();for(var $=0;$<E.length;$++)S+=" ",S+=this.formatTime(E[$],y,""),$<E.length-2&&(S+=","),$==E.length-2&&(S+=this.i18n.spaceAnd())}else{var k=this.getSecondsDescription(),A=this.getMinutesDescription(),w=this.getHoursDescription();if(S+=k,S&&A&&(S+=", "),S+=A,A===w)return S;S&&w&&(S+=", "),S+=w}return S},h.prototype.getSecondsDescription=function(){var m=this,y=this.getSegmentDescription(this.expressionParts[0],this.i18n.everySecond(),function(v){return v},function(v){return s.StringUtilities.format(m.i18n.everyX0Seconds(v),v)},function(v){return m.i18n.secondsX0ThroughX1PastTheMinute()},function(v){return v=="0"?"":parseInt(v)<20?m.i18n.atX0SecondsPastTheMinute(v):m.i18n.atX0SecondsPastTheMinuteGt20()||m.i18n.atX0SecondsPastTheMinute(v)});return y},h.prototype.getMinutesDescription=function(){var m=this,y=this.expressionParts[0],v=this.expressionParts[2],S=this.getSegmentDescription(this.expressionParts[1],this.i18n.everyMinute(),function(_){return _},function(_){return s.StringUtilities.format(m.i18n.everyX0Minutes(_),_)},function(_){return m.i18n.minutesX0ThroughX1PastTheHour()},function(_){try{return _=="0"&&v.indexOf("/")==-1&&y==""?m.i18n.everyHour():parseInt(_)<20?m.i18n.atX0MinutesPastTheHour(_):m.i18n.atX0MinutesPastTheHourGt20()||m.i18n.atX0MinutesPastTheHour(_)}catch{return m.i18n.atX0MinutesPastTheHour(_)}});return S},h.prototype.getHoursDescription=function(){var m=this,y=this.expressionParts[2],v=0,S=[];y.split("/")[0].split(",").forEach(function($){var k=$.split("-");k.length===2&&S.push({value:k[1],index:v+1}),v+=k.length});var _=0,E=this.getSegmentDescription(y,this.i18n.everyHour(),function($){var k=S.find(function(w){return w.value===$&&w.index===_}),A=k&&m.expressionParts[1]!=="0";return _++,A?m.formatTime($,"59",""):m.formatTime($,"0","")},function($){return s.StringUtilities.format(m.i18n.everyX0Hours($),$)},function($){return m.i18n.betweenX0AndX1()},function($){return m.i18n.atX0()});return E},h.prototype.getDayOfWeekDescription=function(){var m=this,y=this.i18n.daysOfTheWeek(),v=null;return this.expressionParts[5]=="*"?v="":v=this.getSegmentDescription(this.expressionParts[5],this.i18n.commaEveryDay(),function(S,_){var E=S;S.indexOf("#")>-1?E=S.substring(0,S.indexOf("#")):S.indexOf("L")>-1&&(E=E.replace("L",""));var $=parseInt(E);if(m.options.tzOffset){var k=m.expressionParts[2],A=parseInt(k)+(m.options.tzOffset?m.options.tzOffset:0);A>=24?$++:A<0&&$--,$>6?$=0:$<0&&($=6)}var w=m.i18n.daysOfTheWeekInCase?m.i18n.daysOfTheWeekInCase(_)[$]:y[$];if(S.indexOf("#")>-1){var R=null,F=S.substring(S.indexOf("#")+1),I=S.substring(0,S.indexOf("#"));switch(F){case"1":R=m.i18n.first(I);break;case"2":R=m.i18n.second(I);break;case"3":R=m.i18n.third(I);break;case"4":R=m.i18n.fourth(I);break;case"5":R=m.i18n.fifth(I);break}w=R+" "+w}return w},function(S){return parseInt(S)==1?"":s.StringUtilities.format(m.i18n.commaEveryX0DaysOfTheWeek(S),S)},function(S){var _=S.substring(0,S.indexOf("-")),E=m.expressionParts[3]!="*";return E?m.i18n.commaAndX0ThroughX1(_):m.i18n.commaX0ThroughX1(_)},function(S){var _=null;if(S.indexOf("#")>-1){var E=S.substring(S.indexOf("#")+1),$=S.substring(0,S.indexOf("#"));_=m.i18n.commaOnThe(E,$).trim()+m.i18n.spaceX0OfTheMonth()}else if(S.indexOf("L")>-1)_=m.i18n.commaOnTheLastX0OfTheMonth(S.replace("L",""));else{var k=m.expressionParts[3]!="*";_=k?m.i18n.commaAndOnX0():m.i18n.commaOnlyOnX0(S)}return _}),v},h.prototype.getMonthDescription=function(){var m=this,y=this.i18n.monthsOfTheYear(),v=this.getSegmentDescription(this.expressionParts[4],"",function(S,_){return _&&m.i18n.monthsOfTheYearInCase?m.i18n.monthsOfTheYearInCase(_)[parseInt(S)-1]:y[parseInt(S)-1]},function(S){return parseInt(S)==1?"":s.StringUtilities.format(m.i18n.commaEveryX0Months(S),S)},function(S){return m.i18n.commaMonthX0ThroughMonthX1()||m.i18n.commaX0ThroughX1()},function(S){return m.i18n.commaOnlyInMonthX0?m.i18n.commaOnlyInMonthX0():m.i18n.commaOnlyInX0()});return v},h.prototype.getDayOfMonthDescription=function(){var m=this,y=null,v=this.expressionParts[3];switch(v){case"L":y=this.i18n.commaOnTheLastDayOfTheMonth();break;case"WL":case"LW":y=this.i18n.commaOnTheLastWeekdayOfTheMonth();break;default:var S=v.match(/(\d{1,2}W)|(W\d{1,2})/);if(S){var _=parseInt(S[0].replace("W","")),E=_==1?this.i18n.firstWeekday():s.StringUtilities.format(this.i18n.weekdayNearestDayX0(),_.toString());y=s.StringUtilities.format(this.i18n.commaOnTheX0OfTheMonth(),E);break}else{var $=v.match(/L-(\d{1,2})/);if($){var k=$[1];y=s.StringUtilities.format(this.i18n.commaDaysBeforeTheLastDayOfTheMonth(k),k);break}else{if(v=="*"&&this.expressionParts[5]!="*")return"";y=this.getSegmentDescription(v,this.i18n.commaEveryDay(),function(A){return A=="L"?m.i18n.lastDay():m.i18n.dayX0?s.StringUtilities.format(m.i18n.dayX0(),A):A},function(A){return A=="1"?m.i18n.commaEveryDay():m.i18n.commaEveryX0Days(A)},function(A){return m.i18n.commaBetweenDayX0AndX1OfTheMonth(A)},function(A){return m.i18n.commaOnDayX0OfTheMonth(A)})}break}}return y},h.prototype.getYearDescription=function(){var m=this,y=this.getSegmentDescription(this.expressionParts[6],"",function(v){return/^\d+$/.test(v)?new Date(parseInt(v),1).getFullYear().toString():v},function(v){return s.StringUtilities.format(m.i18n.commaEveryX0Years(v),v)},function(v){return m.i18n.commaYearX0ThroughYearX1()||m.i18n.commaX0ThroughX1()},function(v){return m.i18n.commaOnlyInYearX0?m.i18n.commaOnlyInYearX0():m.i18n.commaOnlyInX0()});return y},h.prototype.getSegmentDescription=function(m,y,v,S,_,E){var $=null,k=m.indexOf("/")>-1,A=m.indexOf("-")>-1,w=m.indexOf(",")>-1;if(!m)$="";else if(m==="*")$=y;else if(!k&&!A&&!w)$=s.StringUtilities.format(E(m),v(m));else if(w){for(var R=m.split(","),F="",I=0;I<R.length;I++)if(I>0&&R.length>2&&(F+=",",I<R.length-1&&(F+=" ")),I>0&&R.length>1&&(I==R.length-1||R.length==2)&&(F+="".concat(this.i18n.spaceAnd()," ")),R[I].indexOf("/")>-1||R[I].indexOf("-")>-1){var U=R[I].indexOf("-")>-1&&R[I].indexOf("/")==-1,q=this.getSegmentDescription(R[I],y,v,S,U?this.i18n.commaX0ThroughX1:_,E);U&&(q=q.replace(", ","")),F+=q}else k?F+=this.getSegmentDescription(R[I],y,v,S,_,E):F+=v(R[I]);k?$=F:$=s.StringUtilities.format(E(m),F)}else if(k){var R=m.split("/");if($=s.StringUtilities.format(S(R[1]),R[1]),R[0].indexOf("-")>-1){var V=this.generateRangeSegmentDescription(R[0],_,v);V.indexOf(", ")!=0&&($+=", "),$+=V}else if(R[0].indexOf("*")==-1){var N=s.StringUtilities.format(E(R[0]),v(R[0]));N=N.replace(", ",""),$+=s.StringUtilities.format(this.i18n.commaStartingX0(),N)}}else A&&($=this.generateRangeSegmentDescription(m,_,v));return $},h.prototype.generateRangeSegmentDescription=function(m,y,v){var S="",_=m.split("-"),E=v(_[0],1),$=v(_[1],2),k=y(m);return S+=s.StringUtilities.format(k,E,$),S},h.prototype.formatTime=function(m,y,v){var S=0,_=0;this.options.tzOffset&&(S=this.options.tzOffset>0?Math.floor(this.options.tzOffset):Math.ceil(this.options.tzOffset),_=parseFloat((this.options.tzOffset%1).toFixed(2)),_!=0&&(_*=60));var E=parseInt(m)+S,$=parseInt(y)+_;$>=60?($-=60,E+=1):$<0&&($+=60,E-=1),E>=24?E=E-24:E<0&&(E=24+E);var k="",A=!1;this.options.use24HourTimeFormat||(A=!!(this.i18n.setPeriodBeforeTime&&this.i18n.setPeriodBeforeTime()),k=A?"".concat(this.getPeriod(E)," "):" ".concat(this.getPeriod(E)),E>12&&(E-=12),E===0&&(E=12));var w="";return v&&(w=":".concat(("00"+v).substring(v.length))),"".concat(A?k:"").concat(("00"+E.toString()).substring(E.toString().length),":").concat(("00"+$.toString()).substring($.toString().length)).concat(w).concat(A?"":k)},h.prototype.transformVerbosity=function(m,y){if(!y&&(m=m.replace(new RegExp(", ".concat(this.i18n.everyMinute()),"g"),""),m=m.replace(new RegExp(", ".concat(this.i18n.everyHour()),"g"),""),m=m.replace(new RegExp(this.i18n.commaEveryDay(),"g"),""),m=m.replace(/\, ?$/,""),this.i18n.conciseVerbosityReplacements))for(var v=0,S=Object.entries(this.i18n.conciseVerbosityReplacements());v<S.length;v++){var _=S[v],E=_[0],$=_[1];m=m.replace(new RegExp(E,"g"),$)}return m},h.prototype.getPeriod=function(m){return m>=12?this.i18n.pm&&this.i18n.pm()||"PM":this.i18n.am&&this.i18n.am()||"AM"},h.locales={},h})();l.ExpressionDescriptor=d}),747:((u,l,o)=>{Object.defineProperty(l,"__esModule",{value:!0}),l.enLocaleLoader=void 0;var s=o(486),c=(function(){function d(){}return d.prototype.load=function(h){h.en=new s.en},d})();l.enLocaleLoader=c}),486:((u,l)=>{Object.defineProperty(l,"__esModule",{value:!0}),l.en=void 0;var o=(function(){function s(){}return s.prototype.atX0SecondsPastTheMinuteGt20=function(){return null},s.prototype.atX0MinutesPastTheHourGt20=function(){return null},s.prototype.commaMonthX0ThroughMonthX1=function(){return null},s.prototype.commaYearX0ThroughYearX1=function(){return null},s.prototype.use24HourTimeFormatByDefault=function(){return!1},s.prototype.anErrorOccuredWhenGeneratingTheExpressionD=function(){return"An error occurred when generating the expression description. Check the cron expression syntax."},s.prototype.everyMinute=function(){return"every minute"},s.prototype.everyHour=function(){return"every hour"},s.prototype.atSpace=function(){return"At "},s.prototype.everyMinuteBetweenX0AndX1=function(){return"Every minute between %s and %s"},s.prototype.at=function(){return"At"},s.prototype.spaceAnd=function(){return" and"},s.prototype.everySecond=function(){return"every second"},s.prototype.everyX0Seconds=function(){return"every %s seconds"},s.prototype.secondsX0ThroughX1PastTheMinute=function(){return"seconds %s through %s past the minute"},s.prototype.atX0SecondsPastTheMinute=function(){return"at %s seconds past the minute"},s.prototype.everyX0Minutes=function(){return"every %s minutes"},s.prototype.minutesX0ThroughX1PastTheHour=function(){return"minutes %s through %s past the hour"},s.prototype.atX0MinutesPastTheHour=function(){return"at %s minutes past the hour"},s.prototype.everyX0Hours=function(){return"every %s hours"},s.prototype.betweenX0AndX1=function(){return"between %s and %s"},s.prototype.atX0=function(){return"at %s"},s.prototype.commaEveryDay=function(){return", every day"},s.prototype.commaEveryX0DaysOfTheWeek=function(){return", every %s days of the week"},s.prototype.commaX0ThroughX1=function(){return", %s through %s"},s.prototype.commaAndX0ThroughX1=function(){return", %s through %s"},s.prototype.first=function(){return"first"},s.prototype.second=function(){return"second"},s.prototype.third=function(){return"third"},s.prototype.fourth=function(){return"fourth"},s.prototype.fifth=function(){return"fifth"},s.prototype.commaOnThe=function(){return", on the "},s.prototype.spaceX0OfTheMonth=function(){return" %s of the month"},s.prototype.lastDay=function(){return"the last day"},s.prototype.commaOnTheLastX0OfTheMonth=function(){return", on the last %s of the month"},s.prototype.commaOnlyOnX0=function(){return", only on %s"},s.prototype.commaAndOnX0=function(){return", and on %s"},s.prototype.commaEveryX0Months=function(){return", every %s months"},s.prototype.commaOnlyInX0=function(){return", only in %s"},s.prototype.commaOnTheLastDayOfTheMonth=function(){return", on the last day of the month"},s.prototype.commaOnTheLastWeekdayOfTheMonth=function(){return", on the last weekday of the month"},s.prototype.commaDaysBeforeTheLastDayOfTheMonth=function(){return", %s days before the last day of the month"},s.prototype.firstWeekday=function(){return"first weekday"},s.prototype.weekdayNearestDayX0=function(){return"weekday nearest day %s"},s.prototype.commaOnTheX0OfTheMonth=function(){return", on the %s of the month"},s.prototype.commaEveryX0Days=function(){return", every %s days"},s.prototype.commaBetweenDayX0AndX1OfTheMonth=function(){return", between day %s and %s of the month"},s.prototype.commaOnDayX0OfTheMonth=function(){return", on day %s of the month"},s.prototype.commaEveryHour=function(){return", every hour"},s.prototype.commaEveryX0Years=function(){return", every %s years"},s.prototype.commaStartingX0=function(){return", starting %s"},s.prototype.daysOfTheWeek=function(){return["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"]},s.prototype.monthsOfTheYear=function(){return["January","February","March","April","May","June","July","August","September","October","November","December"]},s})();l.en=o}),515:((u,l)=>{Object.defineProperty(l,"__esModule",{value:!0});function o(c,d){if(!c)throw new Error(d)}var s=(function(){function c(){}return c.secondRange=function(d){for(var h=d.split(","),m=0;m<h.length;m++)if(!isNaN(parseInt(h[m],10))){var y=parseInt(h[m],10);o(y>=0&&y<=59,"seconds part must be >= 0 and <= 59")}},c.minuteRange=function(d){for(var h=d.split(","),m=0;m<h.length;m++)if(!isNaN(parseInt(h[m],10))){var y=parseInt(h[m],10);o(y>=0&&y<=59,"minutes part must be >= 0 and <= 59")}},c.hourRange=function(d){for(var h=d.split(","),m=0;m<h.length;m++)if(!isNaN(parseInt(h[m],10))){var y=parseInt(h[m],10);o(y>=0&&y<=23,"hours part must be >= 0 and <= 23")}},c.dayOfMonthRange=function(d){for(var h=d.split(","),m=0;m<h.length;m++)if(!isNaN(parseInt(h[m],10))){var y=parseInt(h[m],10);o(y>=1&&y<=31,"DOM part must be >= 1 and <= 31")}},c.monthRange=function(d,h){for(var m=d.split(","),y=0;y<m.length;y++)if(!isNaN(parseInt(m[y],10))){var v=parseInt(m[y],10);o(v>=1&&v<=12,h?"month part must be >= 0 and <= 11":"month part must be >= 1 and <= 12")}},c.dayOfWeekRange=function(d,h){for(var m=d.split(","),y=0;y<m.length;y++)if(!isNaN(parseInt(m[y],10))){var v=parseInt(m[y],10);o(v>=0&&v<=6,h?"DOW part must be >= 0 and <= 6":"DOW part must be >= 1 and <= 7")}},c})();l.default=s}),823:((u,l)=>{Object.defineProperty(l,"__esModule",{value:!0}),l.StringUtilities=void 0;var o=(function(){function s(){}return s.format=function(c){for(var d=[],h=1;h<arguments.length;h++)d[h-1]=arguments[h];return c.replace(/%s/g,function(m){return d.shift()})},s.containsAny=function(c,d){return d.some(function(h){return c.indexOf(h)>-1})},s})();l.StringUtilities=o})},r={};function i(u){var l=r[u];if(l!==void 0)return l.exports;var o=r[u]={exports:{}};return n[u](o,o.exports,i),o.exports}var a={};return(()=>{var u=a;Object.defineProperty(u,"__esModule",{value:!0}),u.toString=void 0;var l=i(333),o=i(747);l.ExpressionDescriptor.initialize(new o.enLocaleLoader),u.default=l.ExpressionDescriptor;var s=l.ExpressionDescriptor.toString;u.toString=s})(),a})())})(qP)),qP.exports}var uqe=aqe();const Nut=da(uqe);var VP={},UP={},EY;function lqe(){return EY||(EY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return n==="__proto__"}e.isUnsafeProperty=t})(UP)),UP}var jP={},DY;function vve(){return DY||(DY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){switch(typeof n){case"number":case"symbol":return!1;case"string":return n.includes(".")||n.includes("[")||n.includes("]")}}e.isDeepKey=t})(jP)),jP}var ZP={},$Y;function yve(){return $Y||($Y=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){var r;return typeof n=="string"||typeof n=="symbol"?n:Object.is((r=n==null?void 0:n.valueOf)==null?void 0:r.call(n),-0)?"-0":String(n)}e.toKey=t})(ZP)),ZP}var HP={},CY;function $L(){return CY||(CY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){const r=[],i=n.length;if(i===0)return r;let a=0,u="",l="",o=!1;for(n.charCodeAt(0)===46&&(r.push(""),a++);a<i;){const s=n[a];l?s==="\\"&&a+1<i?(a++,u+=n[a]):s===l?l="":u+=s:o?s==='"'||s==="'"?l=s:s==="]"?(o=!1,r.push(u),u=""):u+=s:s==="["?(o=!0,u&&(r.push(u),u="")):s==="."?u&&(r.push(u),u=""):u+=s,a++}return u&&r.push(u),r}e.toPath=t})(HP)),HP}var wY;function CL(){return wY||(wY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=lqe(),n=vve(),r=yve(),i=$L();function a(l,o,s){if(l==null)return s;switch(typeof o){case"string":{if(t.isUnsafeProperty(o))return s;const c=l[o];return c===void 0?n.isDeepKey(o)?a(l,i.toPath(o),s):s:c}case"number":case"symbol":{typeof o=="number"&&(o=r.toKey(o));const c=l[o];return c===void 0?s:c}default:{if(Array.isArray(o))return u(l,o,s);if(Object.is(o==null?void 0:o.valueOf(),-0)?o="-0":o=String(o),t.isUnsafeProperty(o))return s;const c=l[o];return c===void 0?s:c}}}function u(l,o,s){if(o.length===0)return s;let c=l;for(let d=0;d<o.length;d++){if(c==null||t.isUnsafeProperty(o[d]))return s;c=c[o[d]]}return c===void 0?s:c}e.get=a})(VP)),VP}var XP,RY;function oqe(){return RY||(RY=1,XP=CL().get),XP}var sqe=oqe();const Iut=da(sqe);var YP={exports:{}},ta={};/** @license React v17.0.2
 * react-is.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var AY;function cqe(){if(AY)return ta;AY=1;var e=60103,t=60106,n=60107,r=60108,i=60114,a=60109,u=60110,l=60112,o=60113,s=60120,c=60115,d=60116,h=60121,m=60122,y=60117,v=60129,S=60131;if(typeof Symbol=="function"&&Symbol.for){var _=Symbol.for;e=_("react.element"),t=_("react.portal"),n=_("react.fragment"),r=_("react.strict_mode"),i=_("react.profiler"),a=_("react.provider"),u=_("react.context"),l=_("react.forward_ref"),o=_("react.suspense"),s=_("react.suspense_list"),c=_("react.memo"),d=_("react.lazy"),h=_("react.block"),m=_("react.server.block"),y=_("react.fundamental"),v=_("react.debug_trace_mode"),S=_("react.legacy_hidden")}function E(N){if(typeof N=="object"&&N!==null){var j=N.$$typeof;switch(j){case e:switch(N=N.type,N){case n:case i:case r:case o:case s:return N;default:switch(N=N&&N.$$typeof,N){case u:case l:case d:case c:case a:return N;default:return j}}case t:return j}}}var $=a,k=e,A=l,w=n,R=d,F=c,I=t,U=i,q=r,V=o;return ta.ContextConsumer=u,ta.ContextProvider=$,ta.Element=k,ta.ForwardRef=A,ta.Fragment=w,ta.Lazy=R,ta.Memo=F,ta.Portal=I,ta.Profiler=U,ta.StrictMode=q,ta.Suspense=V,ta.isAsyncMode=function(){return!1},ta.isConcurrentMode=function(){return!1},ta.isContextConsumer=function(N){return E(N)===u},ta.isContextProvider=function(N){return E(N)===a},ta.isElement=function(N){return typeof N=="object"&&N!==null&&N.$$typeof===e},ta.isForwardRef=function(N){return E(N)===l},ta.isFragment=function(N){return E(N)===n},ta.isLazy=function(N){return E(N)===d},ta.isMemo=function(N){return E(N)===c},ta.isPortal=function(N){return E(N)===t},ta.isProfiler=function(N){return E(N)===i},ta.isStrictMode=function(N){return E(N)===r},ta.isSuspense=function(N){return E(N)===o},ta.isValidElementType=function(N){return typeof N=="string"||typeof N=="function"||N===n||N===i||N===v||N===r||N===o||N===s||N===S||typeof N=="object"&&N!==null&&(N.$$typeof===d||N.$$typeof===c||N.$$typeof===a||N.$$typeof===u||N.$$typeof===l||N.$$typeof===y||N.$$typeof===h||N[0]===m)},ta.typeOf=E,ta}var TY;function fqe(){return TY||(TY=1,YP.exports=cqe()),YP.exports}var Lut=fqe();function xa(e){return function(){return e}}const bve=Math.cos,z$=Math.sin,bp=Math.sqrt,q$=Math.PI,Xw=2*q$,WM=Math.PI,JM=2*WM,wb=1e-6,dqe=JM-wb;function Ove(e){this._+=e[0];for(let t=1,n=e.length;t<n;++t)this._+=arguments[t]+e[t]}function hqe(e){let t=Math.floor(e);if(!(t>=0))throw new Error(`invalid digits: ${e}`);if(t>15)return Ove;const n=10**t;return function(r){this._+=r[0];for(let i=1,a=r.length;i<a;++i)this._+=Math.round(arguments[i]*n)/n+r[i]}}class pqe{constructor(t){this._x0=this._y0=this._x1=this._y1=null,this._="",this._append=t==null?Ove:hqe(t)}moveTo(t,n){this._append`M${this._x0=this._x1=+t},${this._y0=this._y1=+n}`}closePath(){this._x1!==null&&(this._x1=this._x0,this._y1=this._y0,this._append`Z`)}lineTo(t,n){this._append`L${this._x1=+t},${this._y1=+n}`}quadraticCurveTo(t,n,r,i){this._append`Q${+t},${+n},${this._x1=+r},${this._y1=+i}`}bezierCurveTo(t,n,r,i,a,u){this._append`C${+t},${+n},${+r},${+i},${this._x1=+a},${this._y1=+u}`}arcTo(t,n,r,i,a){if(t=+t,n=+n,r=+r,i=+i,a=+a,a<0)throw new Error(`negative radius: ${a}`);let u=this._x1,l=this._y1,o=r-t,s=i-n,c=u-t,d=l-n,h=c*c+d*d;if(this._x1===null)this._append`M${this._x1=t},${this._y1=n}`;else if(h>wb)if(!(Math.abs(d*o-s*c)>wb)||!a)this._append`L${this._x1=t},${this._y1=n}`;else{let m=r-u,y=i-l,v=o*o+s*s,S=m*m+y*y,_=Math.sqrt(v),E=Math.sqrt(h),$=a*Math.tan((WM-Math.acos((v+h-S)/(2*_*E)))/2),k=$/E,A=$/_;Math.abs(k-1)>wb&&this._append`L${t+k*c},${n+k*d}`,this._append`A${a},${a},0,0,${+(d*m>c*y)},${this._x1=t+A*o},${this._y1=n+A*s}`}}arc(t,n,r,i,a,u){if(t=+t,n=+n,r=+r,u=!!u,r<0)throw new Error(`negative radius: ${r}`);let l=r*Math.cos(i),o=r*Math.sin(i),s=t+l,c=n+o,d=1^u,h=u?i-a:a-i;this._x1===null?this._append`M${s},${c}`:(Math.abs(this._x1-s)>wb||Math.abs(this._y1-c)>wb)&&this._append`L${s},${c}`,r&&(h<0&&(h=h%JM+JM),h>dqe?this._append`A${r},${r},0,1,${d},${t-l},${n-o}A${r},${r},0,1,${d},${this._x1=s},${this._y1=c}`:h>wb&&this._append`A${r},${r},0,${+(h>=WM)},${d},${this._x1=t+r*Math.cos(a)},${this._y1=n+r*Math.sin(a)}`)}rect(t,n,r,i){this._append`M${this._x0=this._x1=+t},${this._y0=this._y1=+n}h${r=+r}v${+i}h${-r}Z`}toString(){return this._}}function wL(e){let t=3;return e.digits=function(n){if(!arguments.length)return t;if(n==null)t=null;else{const r=Math.floor(n);if(!(r>=0))throw new RangeError(`invalid digits: ${n}`);t=r}return e},()=>new pqe(t)}function RL(e){return typeof e=="object"&&"length"in e?e:Array.from(e)}function xve(e){this._context=e}xve.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;default:this._context.lineTo(e,t);break}}};function Sve(e){return new xve(e)}function _ve(e){return e[0]}function Eve(e){return e[1]}function mqe(e,t){var n=xa(!0),r=null,i=Sve,a=null,u=wL(l);e=typeof e=="function"?e:e===void 0?_ve:xa(e),t=typeof t=="function"?t:t===void 0?Eve:xa(t);function l(o){var s,c=(o=RL(o)).length,d,h=!1,m;for(r==null&&(a=i(m=u())),s=0;s<=c;++s)!(s<c&&n(d=o[s],s,o))===h&&((h=!h)?a.lineStart():a.lineEnd()),h&&a.point(+e(d,s,o),+t(d,s,o));if(m)return a=null,m+""||null}return l.x=function(o){return arguments.length?(e=typeof o=="function"?o:xa(+o),l):e},l.y=function(o){return arguments.length?(t=typeof o=="function"?o:xa(+o),l):t},l.defined=function(o){return arguments.length?(n=typeof o=="function"?o:xa(!!o),l):n},l.curve=function(o){return arguments.length?(i=o,r!=null&&(a=i(r)),l):i},l.context=function(o){return arguments.length?(o==null?r=a=null:a=i(r=o),l):r},l}function But(e,t,n){var r=null,i=xa(!0),a=null,u=Sve,l=null,o=wL(s);e=typeof e=="function"?e:e===void 0?_ve:xa(+e),t=typeof t=="function"?t:xa(t===void 0?0:+t),n=typeof n=="function"?n:n===void 0?Eve:xa(+n);function s(d){var h,m,y,v=(d=RL(d)).length,S,_=!1,E,$=new Array(v),k=new Array(v);for(a==null&&(l=u(E=o())),h=0;h<=v;++h){if(!(h<v&&i(S=d[h],h,d))===_)if(_=!_)m=h,l.areaStart(),l.lineStart();else{for(l.lineEnd(),l.lineStart(),y=h-1;y>=m;--y)l.point($[y],k[y]);l.lineEnd(),l.areaEnd()}_&&($[h]=+e(S,h,d),k[h]=+t(S,h,d),l.point(r?+r(S,h,d):$[h],n?+n(S,h,d):k[h]))}if(E)return l=null,E+""||null}function c(){return mqe().defined(i).curve(u).context(a)}return s.x=function(d){return arguments.length?(e=typeof d=="function"?d:xa(+d),r=null,s):e},s.x0=function(d){return arguments.length?(e=typeof d=="function"?d:xa(+d),s):e},s.x1=function(d){return arguments.length?(r=d==null?null:typeof d=="function"?d:xa(+d),s):r},s.y=function(d){return arguments.length?(t=typeof d=="function"?d:xa(+d),n=null,s):t},s.y0=function(d){return arguments.length?(t=typeof d=="function"?d:xa(+d),s):t},s.y1=function(d){return arguments.length?(n=d==null?null:typeof d=="function"?d:xa(+d),s):n},s.lineX0=s.lineY0=function(){return c().x(e).y(t)},s.lineY1=function(){return c().x(e).y(n)},s.lineX1=function(){return c().x(r).y(t)},s.defined=function(d){return arguments.length?(i=typeof d=="function"?d:xa(!!d),s):i},s.curve=function(d){return arguments.length?(u=d,a!=null&&(l=u(a)),s):u},s.context=function(d){return arguments.length?(d==null?a=l=null:l=u(a=d),s):a},s}class Dve{constructor(t,n){this._context=t,this._x=n}areaStart(){this._line=0}areaEnd(){this._line=NaN}lineStart(){this._point=0}lineEnd(){(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line}point(t,n){switch(t=+t,n=+n,this._point){case 0:{this._point=1,this._line?this._context.lineTo(t,n):this._context.moveTo(t,n);break}case 1:this._point=2;default:{this._x?this._context.bezierCurveTo(this._x0=(this._x0+t)/2,this._y0,this._x0,n,t,n):this._context.bezierCurveTo(this._x0,this._y0=(this._y0+n)/2,t,this._y0,t,n);break}}this._x0=t,this._y0=n}}function Qut(e){return new Dve(e,!0)}function zut(e){return new Dve(e,!1)}const gqe={draw(e,t){const n=bp(t/q$);e.moveTo(n,0),e.arc(0,0,n,0,Xw)}},qut={draw(e,t){const n=bp(t/5)/2;e.moveTo(-3*n,-n),e.lineTo(-n,-n),e.lineTo(-n,-3*n),e.lineTo(n,-3*n),e.lineTo(n,-n),e.lineTo(3*n,-n),e.lineTo(3*n,n),e.lineTo(n,n),e.lineTo(n,3*n),e.lineTo(-n,3*n),e.lineTo(-n,n),e.lineTo(-3*n,n),e.closePath()}},$ve=bp(1/3),vqe=$ve*2,Vut={draw(e,t){const n=bp(t/vqe),r=n*$ve;e.moveTo(0,-n),e.lineTo(r,0),e.lineTo(0,n),e.lineTo(-r,0),e.closePath()}},Uut={draw(e,t){const n=bp(t),r=-n/2;e.rect(r,r,n,n)}},yqe=.8908130915292852,Cve=z$(q$/10)/z$(7*q$/10),bqe=z$(Xw/10)*Cve,Oqe=-bve(Xw/10)*Cve,jut={draw(e,t){const n=bp(t*yqe),r=bqe*n,i=Oqe*n;e.moveTo(0,-n),e.lineTo(r,i);for(let a=1;a<5;++a){const u=Xw*a/5,l=bve(u),o=z$(u);e.lineTo(o*n,-l*n),e.lineTo(l*r-o*i,o*r+l*i)}e.closePath()}},GP=bp(3),Zut={draw(e,t){const n=-bp(t/(GP*3));e.moveTo(0,n*2),e.lineTo(-GP*n,-n),e.lineTo(GP*n,-n),e.closePath()}},Ld=-.5,Bd=bp(3)/2,eF=1/bp(12),xqe=(eF/2+1)*3,Hut={draw(e,t){const n=bp(t/xqe),r=n/2,i=n*eF,a=r,u=n*eF+n,l=-a,o=u;e.moveTo(r,i),e.lineTo(a,u),e.lineTo(l,o),e.lineTo(Ld*r-Bd*i,Bd*r+Ld*i),e.lineTo(Ld*a-Bd*u,Bd*a+Ld*u),e.lineTo(Ld*l-Bd*o,Bd*l+Ld*o),e.lineTo(Ld*r+Bd*i,Ld*i-Bd*r),e.lineTo(Ld*a+Bd*u,Ld*u-Bd*a),e.lineTo(Ld*l+Bd*o,Ld*o-Bd*l),e.closePath()}};function Xut(e,t){let n=null,r=wL(i);e=typeof e=="function"?e:xa(e||gqe),t=typeof t=="function"?t:xa(t===void 0?64:+t);function i(){let a;if(n||(n=a=r()),e.apply(this,arguments).draw(n,+t.apply(this,arguments)),a)return n=null,a+""||null}return i.type=function(a){return arguments.length?(e=typeof a=="function"?a:xa(a),i):e},i.size=function(a){return arguments.length?(t=typeof a=="function"?a:xa(+a),i):t},i.context=function(a){return arguments.length?(n=a??null,i):n},i}function V$(){}function U$(e,t,n){e._context.bezierCurveTo((2*e._x0+e._x1)/3,(2*e._y0+e._y1)/3,(e._x0+2*e._x1)/3,(e._y0+2*e._y1)/3,(e._x0+4*e._x1+t)/6,(e._y0+4*e._y1+n)/6)}function wve(e){this._context=e}wve.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){switch(this._point){case 3:U$(this,this._x1,this._y1);case 2:this._context.lineTo(this._x1,this._y1);break}(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;break;case 2:this._point=3,this._context.lineTo((5*this._x0+this._x1)/6,(5*this._y0+this._y1)/6);default:U$(this,e,t);break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}};function Yut(e){return new wve(e)}function Rve(e){this._context=e}Rve.prototype={areaStart:V$,areaEnd:V$,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._y0=this._y1=this._y2=this._y3=this._y4=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:{this._context.moveTo(this._x2,this._y2),this._context.closePath();break}case 2:{this._context.moveTo((this._x2+2*this._x3)/3,(this._y2+2*this._y3)/3),this._context.lineTo((this._x3+2*this._x2)/3,(this._y3+2*this._y2)/3),this._context.closePath();break}case 3:{this.point(this._x2,this._y2),this.point(this._x3,this._y3),this.point(this._x4,this._y4);break}}},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._x2=e,this._y2=t;break;case 1:this._point=2,this._x3=e,this._y3=t;break;case 2:this._point=3,this._x4=e,this._y4=t,this._context.moveTo((this._x0+4*this._x1+e)/6,(this._y0+4*this._y1+t)/6);break;default:U$(this,e,t);break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}};function Gut(e){return new Rve(e)}function Ave(e){this._context=e}Ave.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){(this._line||this._line!==0&&this._point===3)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3;var n=(this._x0+4*this._x1+e)/6,r=(this._y0+4*this._y1+t)/6;this._line?this._context.lineTo(n,r):this._context.moveTo(n,r);break;case 3:this._point=4;default:U$(this,e,t);break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}};function Kut(e){return new Ave(e)}function Tve(e){this._context=e}Tve.prototype={areaStart:V$,areaEnd:V$,lineStart:function(){this._point=0},lineEnd:function(){this._point&&this._context.closePath()},point:function(e,t){e=+e,t=+t,this._point?this._context.lineTo(e,t):(this._point=1,this._context.moveTo(e,t))}};function Wut(e){return new Tve(e)}function kY(e){return e<0?-1:1}function PY(e,t,n){var r=e._x1-e._x0,i=t-e._x1,a=(e._y1-e._y0)/(r||i<0&&-0),u=(n-e._y1)/(i||r<0&&-0),l=(a*i+u*r)/(r+i);return(kY(a)+kY(u))*Math.min(Math.abs(a),Math.abs(u),.5*Math.abs(l))||0}function MY(e,t){var n=e._x1-e._x0;return n?(3*(e._y1-e._y0)/n-t)/2:t}function KP(e,t,n){var r=e._x0,i=e._y0,a=e._x1,u=e._y1,l=(a-r)/3;e._context.bezierCurveTo(r+l,i+l*t,a-l,u-l*n,a,u)}function j$(e){this._context=e}j$.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=this._t0=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x1,this._y1);break;case 3:KP(this,this._t0,MY(this,this._t0));break}(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){var n=NaN;if(e=+e,t=+t,!(e===this._x1&&t===this._y1)){switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;break;case 2:this._point=3,KP(this,MY(this,n=PY(this,e,t)),n);break;default:KP(this,this._t0,n=PY(this,e,t));break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t,this._t0=n}}};function kve(e){this._context=new Pve(e)}(kve.prototype=Object.create(j$.prototype)).point=function(e,t){j$.prototype.point.call(this,t,e)};function Pve(e){this._context=e}Pve.prototype={moveTo:function(e,t){this._context.moveTo(t,e)},closePath:function(){this._context.closePath()},lineTo:function(e,t){this._context.lineTo(t,e)},bezierCurveTo:function(e,t,n,r,i,a){this._context.bezierCurveTo(t,e,r,n,a,i)}};function Jut(e){return new j$(e)}function elt(e){return new kve(e)}function Mve(e){this._context=e}Mve.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=[],this._y=[]},lineEnd:function(){var e=this._x,t=this._y,n=e.length;if(n)if(this._line?this._context.lineTo(e[0],t[0]):this._context.moveTo(e[0],t[0]),n===2)this._context.lineTo(e[1],t[1]);else for(var r=FY(e),i=FY(t),a=0,u=1;u<n;++a,++u)this._context.bezierCurveTo(r[0][a],i[0][a],r[1][a],i[1][a],e[u],t[u]);(this._line||this._line!==0&&n===1)&&this._context.closePath(),this._line=1-this._line,this._x=this._y=null},point:function(e,t){this._x.push(+e),this._y.push(+t)}};function FY(e){var t,n=e.length-1,r,i=new Array(n),a=new Array(n),u=new Array(n);for(i[0]=0,a[0]=2,u[0]=e[0]+2*e[1],t=1;t<n-1;++t)i[t]=1,a[t]=4,u[t]=4*e[t]+2*e[t+1];for(i[n-1]=2,a[n-1]=7,u[n-1]=8*e[n-1]+e[n],t=1;t<n;++t)r=i[t]/a[t-1],a[t]-=r,u[t]-=r*u[t-1];for(i[n-1]=u[n-1]/a[n-1],t=n-2;t>=0;--t)i[t]=(u[t]-i[t+1])/a[t];for(a[n-1]=(e[n]+i[n-1])/2,t=0;t<n-1;++t)a[t]=2*e[t+1]-i[t+1];return[i,a]}function tlt(e){return new Mve(e)}function Yw(e,t){this._context=e,this._t=t}Yw.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=this._y=NaN,this._point=0},lineEnd:function(){0<this._t&&this._t<1&&this._point===2&&this._context.lineTo(this._x,this._y),(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line>=0&&(this._t=1-this._t,this._line=1-this._line)},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;default:{if(this._t<=0)this._context.lineTo(this._x,t),this._context.lineTo(e,t);else{var n=this._x*(1-this._t)+e*this._t;this._context.lineTo(n,this._y),this._context.lineTo(n,t)}break}}this._x=e,this._y=t}};function nlt(e){return new Yw(e,.5)}function rlt(e){return new Yw(e,0)}function ilt(e){return new Yw(e,1)}function Xx(e,t){if((u=e.length)>1)for(var n=1,r,i,a=e[t[0]],u,l=a.length;n<u;++n)for(i=a,a=e[t[n]],r=0;r<l;++r)a[r][1]+=a[r][0]=isNaN(i[r][1])?i[r][0]:i[r][1]}function NY(e){for(var t=e.length,n=new Array(t);--t>=0;)n[t]=t;return n}function Sqe(e,t){return e[t]}function _qe(e){const t=[];return t.key=e,t}function alt(){var e=xa([]),t=NY,n=Xx,r=Sqe;function i(a){var u=Array.from(e.apply(this,arguments),_qe),l,o=u.length,s=-1,c;for(const d of a)for(l=0,++s;l<o;++l)(u[l][s]=[0,+r(d,u[l].key,s,a)]).data=d;for(l=0,c=RL(t(u));l<o;++l)u[c[l]].index=l;return n(u,c),u}return i.keys=function(a){return arguments.length?(e=typeof a=="function"?a:xa(Array.from(a)),i):e},i.value=function(a){return arguments.length?(r=typeof a=="function"?a:xa(+a),i):r},i.order=function(a){return arguments.length?(t=a==null?NY:typeof a=="function"?a:xa(Array.from(a)),i):t},i.offset=function(a){return arguments.length?(n=a??Xx,i):n},i}function ult(e,t){if((r=e.length)>0){for(var n,r,i=0,a=e[0].length,u;i<a;++i){for(u=n=0;n<r;++n)u+=e[n][i][1]||0;if(u)for(n=0;n<r;++n)e[n][i][1]/=u}Xx(e,t)}}function llt(e,t){if((i=e.length)>0){for(var n=0,r=e[t[0]],i,a=r.length;n<a;++n){for(var u=0,l=0;u<i;++u)l+=e[u][n][1]||0;r[n][1]+=r[n][0]=-l/2}Xx(e,t)}}function olt(e,t){if(!(!((u=e.length)>0)||!((a=(i=e[t[0]]).length)>0))){for(var n=0,r=1,i,a,u;r<a;++r){for(var l=0,o=0,s=0;l<u;++l){for(var c=e[t[l]],d=c[r][1]||0,h=c[r-1][1]||0,m=(d-h)/2,y=0;y<l;++y){var v=e[t[y]],S=v[r][1]||0,_=v[r-1][1]||0;m+=S-_}o+=d,s+=m*d}i[r-1][1]+=i[r-1][0]=n,o&&(n-=s/o)}i[r-1][1]+=i[r-1][0]=n,Xx(e,t)}}var WP={},JP={},IY;function Eqe(){return IY||(IY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n,r){const i=new Map;for(let a=0;a<n.length;a++){const u=n[a],l=r(u);i.has(l)||i.set(l,u)}return Array.from(i.values())}e.uniqBy=t})(JP)),JP}var e8={},LY;function Fve(){return LY||(LY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return n}e.identity=t})(e8)),e8}var t8={},n8={},r8={},BY;function Dqe(){return BY||(BY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return Number.isSafeInteger(n)&&n>=0}e.isLength=t})(r8)),r8}var QY;function AL(){return QY||(QY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Dqe();function n(r){return r!=null&&typeof r!="function"&&t.isLength(r.length)}e.isArrayLike=n})(n8)),n8}var i8={},zY;function $qe(){return zY||(zY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return typeof n=="object"&&n!==null}e.isObjectLike=t})(i8)),i8}var qY;function Cqe(){return qY||(qY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=AL(),n=$qe();function r(i){return n.isObjectLike(i)&&t.isArrayLike(i)}e.isArrayLikeObject=r})(t8)),t8}var a8={},u8={},VY;function wqe(){return VY||(VY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=CL();function n(r){return function(i){return t.get(i,r)}}e.property=n})(u8)),u8}var l8={},o8={},s8={},c8={},UY;function Nve(){return UY||(UY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return n!==null&&(typeof n=="object"||typeof n=="function")}e.isObject=t})(c8)),c8}var f8={},jY;function Ive(){return jY||(jY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return n==null||typeof n!="object"&&typeof n!="function"}e.isPrimitive=t})(f8)),f8}var d8={},ZY;function TL(){return ZY||(ZY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n,r){return n===r||Number.isNaN(n)&&Number.isNaN(r)}e.eq=t})(d8)),d8}var HY;function Rqe(){return HY||(HY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=kL(),n=Nve(),r=Ive(),i=TL();function a(d,h,m){return typeof m!="function"?t.isMatch(d,h):u(d,h,function y(v,S,_,E,$,k){const A=m(v,S,_,E,$,k);return A!==void 0?!!A:u(v,S,y,k)},new Map)}function u(d,h,m,y){if(h===d)return!0;switch(typeof h){case"object":return l(d,h,m,y);case"function":return Object.keys(h).length>0?u(d,{...h},m,y):i.eq(d,h);default:return n.isObject(d)?typeof h=="string"?h==="":!0:i.eq(d,h)}}function l(d,h,m,y){if(h==null)return!0;if(Array.isArray(h))return s(d,h,m,y);if(h instanceof Map)return o(d,h,m,y);if(h instanceof Set)return c(d,h,m,y);const v=Object.keys(h);if(d==null)return v.length===0;if(v.length===0)return!0;if(y&&y.has(h))return y.get(h)===d;y&&y.set(h,d);try{for(let S=0;S<v.length;S++){const _=v[S];if(!r.isPrimitive(d)&&!(_ in d)||h[_]===void 0&&d[_]!==void 0||h[_]===null&&d[_]!==null||!m(d[_],h[_],_,d,h,y))return!1}return!0}finally{y&&y.delete(h)}}function o(d,h,m,y){if(h.size===0)return!0;if(!(d instanceof Map))return!1;for(const[v,S]of h.entries()){const _=d.get(v);if(m(_,S,v,d,h,y)===!1)return!1}return!0}function s(d,h,m,y){if(h.length===0)return!0;if(!Array.isArray(d))return!1;const v=new Set;for(let S=0;S<h.length;S++){const _=h[S];let E=!1;for(let $=0;$<d.length;$++){if(v.has($))continue;const k=d[$];let A=!1;if(m(k,_,S,d,h,y)&&(A=!0),A){v.add($),E=!0;break}}if(!E)return!1}return!0}function c(d,h,m,y){return h.size===0?!0:d instanceof Set?s([...d],[...h],m,y):!1}e.isMatchWith=a,e.isSetMatch=c})(s8)),s8}var XY;function kL(){return XY||(XY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Rqe();function n(r,i){return t.isMatchWith(r,i,()=>{})}e.isMatch=n})(o8)),o8}var h8={},p8={},m8={},YY;function Lve(){return YY||(YY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return Object.getOwnPropertySymbols(n).filter(r=>Object.prototype.propertyIsEnumerable.call(n,r))}e.getSymbols=t})(m8)),m8}var g8={},GY;function PL(){return GY||(GY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return n==null?n===void 0?"[object Undefined]":"[object Null]":Object.prototype.toString.call(n)}e.getTag=t})(g8)),g8}var v8={},KY;function ML(){return KY||(KY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t="[object RegExp]",n="[object String]",r="[object Number]",i="[object Boolean]",a="[object Arguments]",u="[object Symbol]",l="[object Date]",o="[object Map]",s="[object Set]",c="[object Array]",d="[object Function]",h="[object ArrayBuffer]",m="[object Object]",y="[object Error]",v="[object DataView]",S="[object Uint8Array]",_="[object Uint8ClampedArray]",E="[object Uint16Array]",$="[object Uint32Array]",k="[object BigUint64Array]",A="[object Int8Array]",w="[object Int16Array]",R="[object Int32Array]",F="[object BigInt64Array]",I="[object Float32Array]",U="[object Float64Array]";e.argumentsTag=a,e.arrayBufferTag=h,e.arrayTag=c,e.bigInt64ArrayTag=F,e.bigUint64ArrayTag=k,e.booleanTag=i,e.dataViewTag=v,e.dateTag=l,e.errorTag=y,e.float32ArrayTag=I,e.float64ArrayTag=U,e.functionTag=d,e.int16ArrayTag=w,e.int32ArrayTag=R,e.int8ArrayTag=A,e.mapTag=o,e.numberTag=r,e.objectTag=m,e.regexpTag=t,e.setTag=s,e.stringTag=n,e.symbolTag=u,e.uint16ArrayTag=E,e.uint32ArrayTag=$,e.uint8ArrayTag=S,e.uint8ClampedArrayTag=_})(v8)),v8}var y8={},WY;function Aqe(){return WY||(WY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return ArrayBuffer.isView(n)&&!(n instanceof DataView)}e.isTypedArray=t})(y8)),y8}var JY;function Bve(){return JY||(JY=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Lve(),n=PL(),r=ML(),i=Ive(),a=Aqe();function u(c,d){return l(c,void 0,c,new Map,d)}function l(c,d,h,m=new Map,y=void 0){const v=y==null?void 0:y(c,d,h,m);if(v!=null)return v;if(i.isPrimitive(c))return c;if(m.has(c))return m.get(c);if(Array.isArray(c)){const S=new Array(c.length);m.set(c,S);for(let _=0;_<c.length;_++)S[_]=l(c[_],_,h,m,y);return Object.hasOwn(c,"index")&&(S.index=c.index),Object.hasOwn(c,"input")&&(S.input=c.input),S}if(c instanceof Date)return new Date(c.getTime());if(c instanceof RegExp){const S=new RegExp(c.source,c.flags);return S.lastIndex=c.lastIndex,S}if(c instanceof Map){const S=new Map;m.set(c,S);for(const[_,E]of c)S.set(_,l(E,_,h,m,y));return S}if(c instanceof Set){const S=new Set;m.set(c,S);for(const _ of c)S.add(l(_,void 0,h,m,y));return S}if(typeof Buffer<"u"&&Buffer.isBuffer(c))return c.subarray();if(a.isTypedArray(c)){const S=new(Object.getPrototypeOf(c)).constructor(c.length);m.set(c,S);for(let _=0;_<c.length;_++)S[_]=l(c[_],_,h,m,y);return S}if(c instanceof ArrayBuffer||typeof SharedArrayBuffer<"u"&&c instanceof SharedArrayBuffer)return c.slice(0);if(c instanceof DataView){const S=new DataView(c.buffer.slice(0),c.byteOffset,c.byteLength);return m.set(c,S),o(S,c,h,m,y),S}if(typeof File<"u"&&c instanceof File){const S=new File([c],c.name,{type:c.type});return m.set(c,S),o(S,c,h,m,y),S}if(c instanceof Blob){const S=new Blob([c],{type:c.type});return m.set(c,S),o(S,c,h,m,y),S}if(c instanceof Error){const S=new c.constructor;return m.set(c,S),S.message=c.message,S.name=c.name,S.stack=c.stack,S.cause=c.cause,o(S,c,h,m,y),S}if(typeof c=="object"&&s(c)){const S=Object.create(Object.getPrototypeOf(c));return m.set(c,S),o(S,c,h,m,y),S}return c}function o(c,d,h=c,m,y){const v=[...Object.keys(d),...t.getSymbols(d)];for(let S=0;S<v.length;S++){const _=v[S],E=Object.getOwnPropertyDescriptor(c,_);(E==null||E.writable)&&(c[_]=l(d[_],_,h,m,y))}}function s(c){switch(n.getTag(c)){case r.argumentsTag:case r.arrayTag:case r.arrayBufferTag:case r.dataViewTag:case r.booleanTag:case r.dateTag:case r.float32ArrayTag:case r.float64ArrayTag:case r.int8ArrayTag:case r.int16ArrayTag:case r.int32ArrayTag:case r.mapTag:case r.numberTag:case r.objectTag:case r.regexpTag:case r.setTag:case r.stringTag:case r.symbolTag:case r.uint8ArrayTag:case r.uint8ClampedArrayTag:case r.uint16ArrayTag:case r.uint32ArrayTag:return!0;default:return!1}}e.cloneDeepWith=u,e.cloneDeepWithImpl=l,e.copyProperties=o})(p8)),p8}var eG;function Tqe(){return eG||(eG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Bve();function n(r){return t.cloneDeepWithImpl(r,void 0,r,new Map,void 0)}e.cloneDeep=n})(h8)),h8}var tG;function kqe(){return tG||(tG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=kL(),n=Tqe();function r(i){return i=n.cloneDeep(i),a=>t.isMatch(a,i)}e.matches=r})(l8)),l8}var b8={},O8={},x8={},nG;function Pqe(){return nG||(nG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Bve(),n=ML();function r(i,a){return t.cloneDeepWith(i,(u,l,o,s)=>{const c=a==null?void 0:a(u,l,o,s);if(c!=null)return c;if(typeof i=="object")switch(Object.prototype.toString.call(i)){case n.numberTag:case n.stringTag:case n.booleanTag:{const d=new i.constructor(i==null?void 0:i.valueOf());return t.copyProperties(d,i),d}case n.argumentsTag:{const d={};return t.copyProperties(d,i),d.length=i.length,d[Symbol.iterator]=i[Symbol.iterator],d}default:return}})}e.cloneDeepWith=r})(x8)),x8}var rG;function Mqe(){return rG||(rG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Pqe();function n(r){return t.cloneDeepWith(r)}e.cloneDeep=n})(O8)),O8}var S8={},_8={},iG;function Qve(){return iG||(iG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=/^(?:0|[1-9]\d*)$/;function n(r,i=Number.MAX_SAFE_INTEGER){switch(typeof r){case"number":return Number.isInteger(r)&&r>=0&&r<i;case"symbol":return!1;case"string":return t.test(r)}}e.isIndex=n})(_8)),_8}var E8={},aG;function Fqe(){return aG||(aG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=PL();function n(r){return r!==null&&typeof r=="object"&&t.getTag(r)==="[object Arguments]"}e.isArguments=n})(E8)),E8}var uG;function Nqe(){return uG||(uG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=vve(),n=Qve(),r=Fqe(),i=$L();function a(u,l){let o;if(Array.isArray(l)?o=l:typeof l=="string"&&t.isDeepKey(l)&&(u==null?void 0:u[l])==null?o=i.toPath(l):o=[l],o.length===0)return!1;let s=u;for(let c=0;c<o.length;c++){const d=o[c];if((s==null||!Object.hasOwn(s,d))&&!((Array.isArray(s)||r.isArguments(s))&&n.isIndex(d)&&d<s.length))return!1;s=s[d]}return!0}e.has=a})(S8)),S8}var lG;function Iqe(){return lG||(lG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=kL(),n=yve(),r=Mqe(),i=CL(),a=Nqe();function u(l,o){switch(typeof l){case"object":{Object.is(l==null?void 0:l.valueOf(),-0)&&(l="-0");break}case"number":{l=n.toKey(l);break}}return o=r.cloneDeep(o),function(s){const c=i.get(s,l);return c===void 0?a.has(s,l):o===void 0?c===void 0:t.isMatch(c,o)}}e.matchesProperty=u})(b8)),b8}var oG;function Lqe(){return oG||(oG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Fve(),n=wqe(),r=kqe(),i=Iqe();function a(u){if(u==null)return t.identity;switch(typeof u){case"function":return u;case"object":return Array.isArray(u)&&u.length===2?i.matchesProperty(u[0],u[1]):r.matches(u);case"string":case"symbol":case"number":return n.property(u)}}e.iteratee=a})(a8)),a8}var sG;function Bqe(){return sG||(sG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Eqe(),n=Fve(),r=Cqe(),i=Lqe();function a(u,l=n.identity){return r.isArrayLikeObject(u)?t.uniqBy(Array.from(u),i.iteratee(l)):[]}e.uniqBy=a})(WP)),WP}var D8,cG;function Qqe(){return cG||(cG=1,D8=Bqe().uniqBy),D8}var zqe=Qqe();const slt=da(zqe);function qqe(e,t=`expected a function, instead received ${typeof e}`){if(typeof e!="function")throw new TypeError(t)}function Vqe(e,t=`expected an object, instead received ${typeof e}`){if(typeof e!="object")throw new TypeError(t)}function Uqe(e,t="expected all items to be functions, instead received the following types: "){if(!e.every(n=>typeof n=="function")){const n=e.map(r=>typeof r=="function"?`function ${r.name||"unnamed"}()`:typeof r).join(", ");throw new TypeError(`${t}[${n}]`)}}var fG=e=>Array.isArray(e)?e:[e];function jqe(e){const t=Array.isArray(e[0])?e[0]:e;return Uqe(t,"createSelector expects all input-selectors to be functions, but received the following types: "),t}function Zqe(e,t){const n=[],{length:r}=e;for(let i=0;i<r;i++)n.push(e[i].apply(null,t));return n}var Hqe=class{constructor(e){this.value=e}deref(){return this.value}},Xqe=typeof WeakRef<"u"?WeakRef:Hqe,Yqe=0,dG=1;function TE(){return{s:Yqe,v:void 0,o:null,p:null}}function zve(e,t={}){let n=TE();const{resultEqualityCheck:r}=t;let i,a=0;function u(){var d;let l=n;const{length:o}=arguments;for(let h=0,m=o;h<m;h++){const y=arguments[h];if(typeof y=="function"||typeof y=="object"&&y!==null){let v=l.o;v===null&&(l.o=v=new WeakMap);const S=v.get(y);S===void 0?(l=TE(),v.set(y,l)):l=S}else{let v=l.p;v===null&&(l.p=v=new Map);const S=v.get(y);S===void 0?(l=TE(),v.set(y,l)):l=S}}const s=l;let c;if(l.s===dG)c=l.v;else if(c=e.apply(null,arguments),a++,r){const h=((d=i==null?void 0:i.deref)==null?void 0:d.call(i))??i;h!=null&&r(h,c)&&(c=h,a!==0&&a--),i=typeof c=="object"&&c!==null||typeof c=="function"?new Xqe(c):c}return s.s=dG,s.v=c,c}return u.clearCache=()=>{n=TE(),u.resetResultsCount()},u.resultsCount=()=>a,u.resetResultsCount=()=>{a=0},u}function Gqe(e,...t){const n=typeof e=="function"?{memoize:e,memoizeOptions:t}:e,r=(...i)=>{let a=0,u=0,l,o={},s=i.pop();typeof s=="object"&&(o=s,s=i.pop()),qqe(s,`createSelector expects an output function after the inputs, but received: [${typeof s}]`);const c={...n,...o},{memoize:d,memoizeOptions:h=[],argsMemoize:m=zve,argsMemoizeOptions:y=[]}=c,v=fG(h),S=fG(y),_=jqe(i),E=d(function(){return a++,s.apply(null,arguments)},...v),$=m(function(){u++;const A=Zqe(_,arguments);return l=E.apply(null,A),l},...S);return Object.assign($,{resultFunc:s,memoizedResultFunc:E,dependencies:_,dependencyRecomputations:()=>u,resetDependencyRecomputations:()=>{u=0},lastResult:()=>l,recomputations:()=>a,resetRecomputations:()=>{a=0},memoize:d,argsMemoize:m})};return Object.assign(r,{withTypes:()=>r}),r}var Kqe=Gqe(zve),Wqe=Object.assign((e,t=Kqe)=>{Vqe(e,`createStructuredSelector expects first argument to be an object where each property is a selector, instead received a ${typeof e}`);const n=Object.keys(e),r=n.map(a=>e[a]);return t(r,(...a)=>a.reduce((u,l,o)=>(u[n[o]]=l,u),{}))},{withTypes:()=>Wqe}),$8={},C8={},w8={},hG;function Jqe(){return hG||(hG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return typeof r=="symbol"?1:r===null?2:r===void 0?3:r!==r?4:0}const n=(r,i,a)=>{if(r!==i){const u=t(r),l=t(i);if(u===l&&u===0){if(r<i)return a==="desc"?1:-1;if(r>i)return a==="desc"?-1:1}return a==="desc"?l-u:u-l}return 0};e.compareValues=n})(w8)),w8}var R8={},A8={},pG;function qve(){return pG||(pG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return typeof n=="symbol"||n instanceof Symbol}e.isSymbol=t})(A8)),A8}var mG;function eVe(){return mG||(mG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=qve(),n=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,r=/^\w*$/;function i(a,u){return Array.isArray(a)?!1:typeof a=="number"||typeof a=="boolean"||a==null||t.isSymbol(a)?!0:typeof a=="string"&&(r.test(a)||!n.test(a))||u!=null&&Object.hasOwn(u,a)}e.isKey=i})(R8)),R8}var gG;function tVe(){return gG||(gG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Jqe(),n=eVe(),r=$L();function i(a,u,l,o){if(a==null)return[];l=o?void 0:l,Array.isArray(a)||(a=Object.values(a)),Array.isArray(u)||(u=u==null?[null]:[u]),u.length===0&&(u=[null]),Array.isArray(l)||(l=l==null?[]:[l]),l=l.map(m=>String(m));const s=(m,y)=>{let v=m;for(let S=0;S<y.length&&v!=null;++S)v=v[y[S]];return v},c=(m,y)=>y==null||m==null?y:typeof m=="object"&&"key"in m?Object.hasOwn(y,m.key)?y[m.key]:s(y,m.path):typeof m=="function"?m(y):Array.isArray(m)?s(y,m):typeof y=="object"?y[m]:y,d=u.map(m=>(Array.isArray(m)&&m.length===1&&(m=m[0]),m==null||typeof m=="function"||Array.isArray(m)||n.isKey(m)?m:{key:m,path:r.toPath(m)}));return a.map(m=>({original:m,criteria:d.map(y=>c(y,m))})).slice().sort((m,y)=>{for(let v=0;v<d.length;v++){const S=t.compareValues(m.criteria[v],y.criteria[v],l[v]);if(S!==0)return S}return 0}).map(m=>m.original)}e.orderBy=i})(C8)),C8}var T8={},vG;function nVe(){return vG||(vG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n,r=1){const i=[],a=Math.floor(r),u=(l,o)=>{for(let s=0;s<l.length;s++){const c=l[s];Array.isArray(c)&&o<a?u(c,o+1):i.push(c)}};return u(n,0),i}e.flatten=t})(T8)),T8}var k8={},yG;function Vve(){return yG||(yG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Qve(),n=AL(),r=Nve(),i=TL();function a(u,l,o){return r.isObject(o)&&(typeof l=="number"&&n.isArrayLike(o)&&t.isIndex(l)&&l<o.length||typeof l=="string"&&l in o)?i.eq(o[l],u):!1}e.isIterateeCall=a})(k8)),k8}var bG;function rVe(){return bG||(bG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=tVe(),n=nVe(),r=Vve();function i(a,...u){const l=u.length;return l>1&&r.isIterateeCall(a,u[0],u[1])?u=[]:l>2&&r.isIterateeCall(u[0],u[1],u[2])&&(u=[u[0]]),t.orderBy(a,n.flatten(u),["asc"])}e.sortBy=i})($8)),$8}var P8,OG;function iVe(){return OG||(OG=1,P8=rVe().sortBy),P8}var aVe=iVe();const clt=da(aVe);function Fo(e){return`Minified Redux error #${e}; visit https://redux.js.org/Errors?code=${e} for the full message or use the non-minified dev environment for full errors. `}var uVe=typeof Symbol=="function"&&Symbol.observable||"@@observable",xG=uVe,M8=()=>Math.random().toString(36).substring(7).split("").join("."),lVe={INIT:`@@redux/INIT${M8()}`,REPLACE:`@@redux/REPLACE${M8()}`,PROBE_UNKNOWN_ACTION:()=>`@@redux/PROBE_UNKNOWN_ACTION${M8()}`},Z$=lVe;function FL(e){if(typeof e!="object"||e===null)return!1;let t=e;for(;Object.getPrototypeOf(t)!==null;)t=Object.getPrototypeOf(t);return Object.getPrototypeOf(e)===t||Object.getPrototypeOf(e)===null}function Uve(e,t,n){if(typeof e!="function")throw new Error(Fo(2));if(typeof t=="function"&&typeof n=="function"||typeof n=="function"&&typeof arguments[3]=="function")throw new Error(Fo(0));if(typeof t=="function"&&typeof n>"u"&&(n=t,t=void 0),typeof n<"u"){if(typeof n!="function")throw new Error(Fo(1));return n(Uve)(e,t)}let r=e,i=t,a=new Map,u=a,l=0,o=!1;function s(){u===a&&(u=new Map,a.forEach((S,_)=>{u.set(_,S)}))}function c(){if(o)throw new Error(Fo(3));return i}function d(S){if(typeof S!="function")throw new Error(Fo(4));if(o)throw new Error(Fo(5));let _=!0;s();const E=l++;return u.set(E,S),function(){if(_){if(o)throw new Error(Fo(6));_=!1,s(),u.delete(E),a=null}}}function h(S){if(!FL(S))throw new Error(Fo(7));if(typeof S.type>"u")throw new Error(Fo(8));if(typeof S.type!="string")throw new Error(Fo(17));if(o)throw new Error(Fo(9));try{o=!0,i=r(i,S)}finally{o=!1}return(a=u).forEach(E=>{E()}),S}function m(S){if(typeof S!="function")throw new Error(Fo(10));r=S,h({type:Z$.REPLACE})}function y(){const S=d;return{subscribe(_){if(typeof _!="object"||_===null)throw new Error(Fo(11));function E(){const k=_;k.next&&k.next(c())}return E(),{unsubscribe:S(E)}},[xG](){return this}}}return h({type:Z$.INIT}),{dispatch:h,subscribe:d,getState:c,replaceReducer:m,[xG]:y}}function oVe(e){Object.keys(e).forEach(t=>{const n=e[t];if(typeof n(void 0,{type:Z$.INIT})>"u")throw new Error(Fo(12));if(typeof n(void 0,{type:Z$.PROBE_UNKNOWN_ACTION()})>"u")throw new Error(Fo(13))})}function sVe(e){const t=Object.keys(e),n={};for(let a=0;a<t.length;a++){const u=t[a];typeof e[u]=="function"&&(n[u]=e[u])}const r=Object.keys(n);let i;try{oVe(n)}catch(a){i=a}return function(u={},l){if(i)throw i;let o=!1;const s={};for(let c=0;c<r.length;c++){const d=r[c],h=n[d],m=u[d],y=h(m,l);if(typeof y>"u")throw l&&l.type,new Error(Fo(14));s[d]=y,o=o||y!==m}return o=o||r.length!==Object.keys(u).length,o?s:u}}function H$(...e){return e.length===0?t=>t:e.length===1?e[0]:e.reduce((t,n)=>(...r)=>t(n(...r)))}function cVe(...e){return t=>(n,r)=>{const i=t(n,r);let a=()=>{throw new Error(Fo(15))};const u={getState:i.getState,dispatch:(o,...s)=>a(o,...s)},l=e.map(o=>o(u));return a=H$(...l)(i.dispatch),{...i,dispatch:a}}}function jve(e){return FL(e)&&"type"in e&&typeof e.type=="string"}var Zve=Symbol.for("immer-nothing"),SG=Symbol.for("immer-draftable"),Wf=Symbol.for("immer-state");function Jh(e,...t){throw new Error(`[Immer] minified error nr: ${e}. Full error at: https://bit.ly/3cXEKWf`)}var B3=Object.getPrototypeOf;function v1(e){return!!e&&!!e[Wf]}function hg(e){var t;return e?Hve(e)||Array.isArray(e)||!!e[SG]||!!((t=e.constructor)!=null&&t[SG])||Kw(e)||Ww(e):!1}var fVe=Object.prototype.constructor.toString();function Hve(e){if(!e||typeof e!="object")return!1;const t=B3(e);if(t===null)return!0;const n=Object.hasOwnProperty.call(t,"constructor")&&t.constructor;return n===Object?!0:typeof n=="function"&&Function.toString.call(n)===fVe}function X$(e,t){Gw(e)===0?Reflect.ownKeys(e).forEach(n=>{t(n,e[n],e)}):e.forEach((n,r)=>t(r,n,e))}function Gw(e){const t=e[Wf];return t?t.type_:Array.isArray(e)?1:Kw(e)?2:Ww(e)?3:0}function tF(e,t){return Gw(e)===2?e.has(t):Object.prototype.hasOwnProperty.call(e,t)}function Xve(e,t,n){const r=Gw(e);r===2?e.set(t,n):r===3?e.add(n):e[t]=n}function dVe(e,t){return e===t?e!==0||1/e===1/t:e!==e&&t!==t}function Kw(e){return e instanceof Map}function Ww(e){return e instanceof Set}function Rb(e){return e.copy_||e.base_}function nF(e,t){if(Kw(e))return new Map(e);if(Ww(e))return new Set(e);if(Array.isArray(e))return Array.prototype.slice.call(e);const n=Hve(e);if(t===!0||t==="class_only"&&!n){const r=Object.getOwnPropertyDescriptors(e);delete r[Wf];let i=Reflect.ownKeys(r);for(let a=0;a<i.length;a++){const u=i[a],l=r[u];l.writable===!1&&(l.writable=!0,l.configurable=!0),(l.get||l.set)&&(r[u]={configurable:!0,writable:!0,enumerable:l.enumerable,value:e[u]})}return Object.create(B3(e),r)}else{const r=B3(e);if(r!==null&&n)return{...e};const i=Object.create(r);return Object.assign(i,e)}}function NL(e,t=!1){return Jw(e)||v1(e)||!hg(e)||(Gw(e)>1&&(e.set=e.add=e.clear=e.delete=hVe),Object.freeze(e),t&&Object.entries(e).forEach(([n,r])=>NL(r,!0))),e}function hVe(){Jh(2)}function Jw(e){return Object.isFrozen(e)}var pVe={};function y1(e){const t=pVe[e];return t||Jh(0,e),t}var Yx;function Yve(){return Yx}function mVe(e,t){return{drafts_:[],parent_:e,immer_:t,canAutoFreeze_:!0,unfinalizedDrafts_:0}}function _G(e,t){t&&(y1("Patches"),e.patches_=[],e.inversePatches_=[],e.patchListener_=t)}function rF(e){iF(e),e.drafts_.forEach(gVe),e.drafts_=null}function iF(e){e===Yx&&(Yx=e.parent_)}function EG(e){return Yx=mVe(Yx,e)}function gVe(e){const t=e[Wf];t.type_===0||t.type_===1?t.revoke_():t.revoked_=!0}function DG(e,t){t.unfinalizedDrafts_=t.drafts_.length;const n=t.drafts_[0];return e!==void 0&&e!==n?(n[Wf].modified_&&(rF(t),Jh(4)),hg(e)&&(e=Y$(t,e),t.parent_||G$(t,e)),t.patches_&&y1("Patches").generateReplacementPatches_(n[Wf].base_,e,t.patches_,t.inversePatches_)):e=Y$(t,n,[]),rF(t),t.patches_&&t.patchListener_(t.patches_,t.inversePatches_),e!==Zve?e:void 0}function Y$(e,t,n){if(Jw(t))return t;const r=t[Wf];if(!r)return X$(t,(i,a)=>$G(e,r,t,i,a,n)),t;if(r.scope_!==e)return t;if(!r.modified_)return G$(e,r.base_,!0),r.base_;if(!r.finalized_){r.finalized_=!0,r.scope_.unfinalizedDrafts_--;const i=r.copy_;let a=i,u=!1;r.type_===3&&(a=new Set(i),i.clear(),u=!0),X$(a,(l,o)=>$G(e,r,i,l,o,n,u)),G$(e,i,!1),n&&e.patches_&&y1("Patches").generatePatches_(r,n,e.patches_,e.inversePatches_)}return r.copy_}function $G(e,t,n,r,i,a,u){if(v1(i)){const l=a&&t&&t.type_!==3&&!tF(t.assigned_,r)?a.concat(r):void 0,o=Y$(e,i,l);if(Xve(n,r,o),v1(o))e.canAutoFreeze_=!1;else return}else u&&n.add(i);if(hg(i)&&!Jw(i)){if(!e.immer_.autoFreeze_&&e.unfinalizedDrafts_<1)return;Y$(e,i),(!t||!t.scope_.parent_)&&typeof r!="symbol"&&Object.prototype.propertyIsEnumerable.call(n,r)&&G$(e,i)}}function G$(e,t,n=!1){!e.parent_&&e.immer_.autoFreeze_&&e.canAutoFreeze_&&NL(t,n)}function vVe(e,t){const n=Array.isArray(e),r={type_:n?1:0,scope_:t?t.scope_:Yve(),modified_:!1,finalized_:!1,assigned_:{},parent_:t,base_:e,draft_:null,copy_:null,revoke_:null,isManual_:!1};let i=r,a=IL;n&&(i=[r],a=Gx);const{revoke:u,proxy:l}=Proxy.revocable(i,a);return r.draft_=l,r.revoke_=u,l}var IL={get(e,t){if(t===Wf)return e;const n=Rb(e);if(!tF(n,t))return yVe(e,n,t);const r=n[t];return e.finalized_||!hg(r)?r:r===F8(e.base_,t)?(N8(e),e.copy_[t]=uF(r,e)):r},has(e,t){return t in Rb(e)},ownKeys(e){return Reflect.ownKeys(Rb(e))},set(e,t,n){const r=Gve(Rb(e),t);if(r!=null&&r.set)return r.set.call(e.draft_,n),!0;if(!e.modified_){const i=F8(Rb(e),t),a=i==null?void 0:i[Wf];if(a&&a.base_===n)return e.copy_[t]=n,e.assigned_[t]=!1,!0;if(dVe(n,i)&&(n!==void 0||tF(e.base_,t)))return!0;N8(e),aF(e)}return e.copy_[t]===n&&(n!==void 0||t in e.copy_)||Number.isNaN(n)&&Number.isNaN(e.copy_[t])||(e.copy_[t]=n,e.assigned_[t]=!0),!0},deleteProperty(e,t){return F8(e.base_,t)!==void 0||t in e.base_?(e.assigned_[t]=!1,N8(e),aF(e)):delete e.assigned_[t],e.copy_&&delete e.copy_[t],!0},getOwnPropertyDescriptor(e,t){const n=Rb(e),r=Reflect.getOwnPropertyDescriptor(n,t);return r&&{writable:!0,configurable:e.type_!==1||t!=="length",enumerable:r.enumerable,value:n[t]}},defineProperty(){Jh(11)},getPrototypeOf(e){return B3(e.base_)},setPrototypeOf(){Jh(12)}},Gx={};X$(IL,(e,t)=>{Gx[e]=function(){return arguments[0]=arguments[0][0],t.apply(this,arguments)}});Gx.deleteProperty=function(e,t){return Gx.set.call(this,e,t,void 0)};Gx.set=function(e,t,n){return IL.set.call(this,e[0],t,n,e[0])};function F8(e,t){const n=e[Wf];return(n?Rb(n):e)[t]}function yVe(e,t,n){var i;const r=Gve(t,n);return r?"value"in r?r.value:(i=r.get)==null?void 0:i.call(e.draft_):void 0}function Gve(e,t){if(!(t in e))return;let n=B3(e);for(;n;){const r=Object.getOwnPropertyDescriptor(n,t);if(r)return r;n=B3(n)}}function aF(e){e.modified_||(e.modified_=!0,e.parent_&&aF(e.parent_))}function N8(e){e.copy_||(e.copy_=nF(e.base_,e.scope_.immer_.useStrictShallowCopy_))}var bVe=class{constructor(e){this.autoFreeze_=!0,this.useStrictShallowCopy_=!1,this.produce=(t,n,r)=>{if(typeof t=="function"&&typeof n!="function"){const a=n;n=t;const u=this;return function(o=a,...s){return u.produce(o,c=>n.call(this,c,...s))}}typeof n!="function"&&Jh(6),r!==void 0&&typeof r!="function"&&Jh(7);let i;if(hg(t)){const a=EG(this),u=uF(t,void 0);let l=!0;try{i=n(u),l=!1}finally{l?rF(a):iF(a)}return _G(a,r),DG(i,a)}else if(!t||typeof t!="object"){if(i=n(t),i===void 0&&(i=t),i===Zve&&(i=void 0),this.autoFreeze_&&NL(i,!0),r){const a=[],u=[];y1("Patches").generateReplacementPatches_(t,i,a,u),r(a,u)}return i}else Jh(1,t)},this.produceWithPatches=(t,n)=>{if(typeof t=="function")return(u,...l)=>this.produceWithPatches(u,o=>t(o,...l));let r,i;return[this.produce(t,n,(u,l)=>{r=u,i=l}),r,i]},typeof(e==null?void 0:e.autoFreeze)=="boolean"&&this.setAutoFreeze(e.autoFreeze),typeof(e==null?void 0:e.useStrictShallowCopy)=="boolean"&&this.setUseStrictShallowCopy(e.useStrictShallowCopy)}createDraft(e){hg(e)||Jh(8),v1(e)&&(e=OVe(e));const t=EG(this),n=uF(e,void 0);return n[Wf].isManual_=!0,iF(t),n}finishDraft(e,t){const n=e&&e[Wf];(!n||!n.isManual_)&&Jh(9);const{scope_:r}=n;return _G(r,t),DG(void 0,r)}setAutoFreeze(e){this.autoFreeze_=e}setUseStrictShallowCopy(e){this.useStrictShallowCopy_=e}applyPatches(e,t){let n;for(n=t.length-1;n>=0;n--){const i=t[n];if(i.path.length===0&&i.op==="replace"){e=i.value;break}}n>-1&&(t=t.slice(n+1));const r=y1("Patches").applyPatches_;return v1(e)?r(e,t):this.produce(e,i=>r(i,t))}};function uF(e,t){const n=Kw(e)?y1("MapSet").proxyMap_(e,t):Ww(e)?y1("MapSet").proxySet_(e,t):vVe(e,t);return(t?t.scope_:Yve()).drafts_.push(n),n}function OVe(e){return v1(e)||Jh(10,e),Kve(e)}function Kve(e){if(!hg(e)||Jw(e))return e;const t=e[Wf];let n;if(t){if(!t.modified_)return t.base_;t.finalized_=!0,n=nF(e,t.scope_.immer_.useStrictShallowCopy_)}else n=nF(e,!0);return X$(n,(r,i)=>{Xve(n,r,Kve(i))}),t&&(t.finalized_=!1),n}var Jf=new bVe,Wve=Jf.produce;Jf.produceWithPatches.bind(Jf);Jf.setAutoFreeze.bind(Jf);Jf.setUseStrictShallowCopy.bind(Jf);Jf.applyPatches.bind(Jf);Jf.createDraft.bind(Jf);Jf.finishDraft.bind(Jf);function flt(e){return e}function Jve(e){return({dispatch:n,getState:r})=>i=>a=>typeof a=="function"?a(n,r,e):i(a)}var xVe=Jve(),SVe=Jve,_Ve=typeof window<"u"&&window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__?window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__:function(){if(arguments.length!==0)return typeof arguments[0]=="object"?H$:H$.apply(null,arguments)};function Q3(e,t){function n(...r){if(t){let i=t(...r);if(!i)throw new Error(Zf(0));return{type:e,payload:i.payload,..."meta"in i&&{meta:i.meta},..."error"in i&&{error:i.error}}}return{type:e,payload:r[0]}}return n.toString=()=>`${e}`,n.type=e,n.match=r=>jve(r)&&r.type===e,n}var eye=class z2 extends Array{constructor(...t){super(...t),Object.setPrototypeOf(this,z2.prototype)}static get[Symbol.species](){return z2}concat(...t){return super.concat.apply(this,t)}prepend(...t){return t.length===1&&Array.isArray(t[0])?new z2(...t[0].concat(this)):new z2(...t.concat(this))}};function CG(e){return hg(e)?Wve(e,()=>{}):e}function kE(e,t,n){return e.has(t)?e.get(t):e.set(t,n(t)).get(t)}function EVe(e){return typeof e=="boolean"}var DVe=()=>function(t){const{thunk:n=!0,immutableCheck:r=!0,serializableCheck:i=!0,actionCreatorCheck:a=!0}=t??{};let u=new eye;return n&&(EVe(n)?u.push(xVe):u.push(SVe(n.extraArgument))),u},$Ve="RTK_autoBatch",wG=e=>t=>{setTimeout(t,e)},CVe=(e={type:"raf"})=>t=>(...n)=>{const r=t(...n);let i=!0,a=!1,u=!1;const l=new Set,o=e.type==="tick"?queueMicrotask:e.type==="raf"?typeof window<"u"&&window.requestAnimationFrame?window.requestAnimationFrame:wG(10):e.type==="callback"?e.queueNotification:wG(e.timeout),s=()=>{u=!1,a&&(a=!1,l.forEach(c=>c()))};return Object.assign({},r,{subscribe(c){const d=()=>i&&c(),h=r.subscribe(d);return l.add(c),()=>{h(),l.delete(c)}},dispatch(c){var d;try{return i=!((d=c==null?void 0:c.meta)!=null&&d[$Ve]),a=!i,a&&(u||(u=!0,o(s))),r.dispatch(c)}finally{i=!0}}})},wVe=e=>function(n){const{autoBatch:r=!0}=n??{};let i=new eye(e);return r&&i.push(CVe(typeof r=="object"?r:void 0)),i};function dlt(e){const t=DVe(),{reducer:n=void 0,middleware:r,devTools:i=!0,preloadedState:a=void 0,enhancers:u=void 0}=e||{};let l;if(typeof n=="function")l=n;else if(FL(n))l=sVe(n);else throw new Error(Zf(1));let o;typeof r=="function"?o=r(t):o=t();let s=H$;i&&(s=_Ve({trace:!1,...typeof i=="object"&&i}));const c=cVe(...o),d=wVe(c);let h=typeof u=="function"?u(d):d();const m=s(...h);return Uve(l,a,m)}function tye(e){const t={},n=[];let r;const i={addCase(a,u){const l=typeof a=="string"?a:a.type;if(!l)throw new Error(Zf(28));if(l in t)throw new Error(Zf(29));return t[l]=u,i},addMatcher(a,u){return n.push({matcher:a,reducer:u}),i},addDefaultCase(a){return r=a,i}};return e(i),[t,n,r]}function RVe(e){return typeof e=="function"}function AVe(e,t){let[n,r,i]=tye(t),a;if(RVe(e))a=()=>CG(e());else{const l=CG(e);a=()=>l}function u(l=a(),o){let s=[n[o.type],...r.filter(({matcher:c})=>c(o)).map(({reducer:c})=>c)];return s.filter(c=>!!c).length===0&&(s=[i]),s.reduce((c,d)=>{if(d)if(v1(c)){const m=d(c,o);return m===void 0?c:m}else{if(hg(c))return Wve(c,h=>d(h,o));{const h=d(c,o);if(h===void 0){if(c===null)return c;throw Error("A case reducer on a non-draftable value must not return undefined")}return h}}return c},l)}return u.getInitialState=a,u}var TVe="ModuleSymbhasOwnPr-0123456789ABCDEFGHNRVfgctiUvz_KqYTJkLxpZXIjQW",kVe=(e=21)=>{let t="",n=e;for(;n--;)t+=TVe[Math.random()*64|0];return t},PVe=Symbol.for("rtk-slice-createasyncthunk");function MVe(e,t){return`${e}/${t}`}function FVe({creators:e}={}){var n;const t=(n=e==null?void 0:e.asyncThunk)==null?void 0:n[PVe];return function(i){const{name:a,reducerPath:u=a}=i;if(!a)throw new Error(Zf(11));const l=(typeof i.reducers=="function"?i.reducers(IVe()):i.reducers)||{},o=Object.keys(l),s={sliceCaseReducersByName:{},sliceCaseReducersByType:{},actionCreators:{},sliceMatchers:[]},c={addCase(k,A){const w=typeof k=="string"?k:k.type;if(!w)throw new Error(Zf(12));if(w in s.sliceCaseReducersByType)throw new Error(Zf(13));return s.sliceCaseReducersByType[w]=A,c},addMatcher(k,A){return s.sliceMatchers.push({matcher:k,reducer:A}),c},exposeAction(k,A){return s.actionCreators[k]=A,c},exposeCaseReducer(k,A){return s.sliceCaseReducersByName[k]=A,c}};o.forEach(k=>{const A=l[k],w={reducerName:k,type:MVe(a,k),createNotation:typeof i.reducers=="function"};BVe(A)?zVe(w,A,c,t):LVe(w,A,c)});function d(){const[k={},A=[],w=void 0]=typeof i.extraReducers=="function"?tye(i.extraReducers):[i.extraReducers],R={...k,...s.sliceCaseReducersByType};return AVe(i.initialState,F=>{for(let I in R)F.addCase(I,R[I]);for(let I of s.sliceMatchers)F.addMatcher(I.matcher,I.reducer);for(let I of A)F.addMatcher(I.matcher,I.reducer);w&&F.addDefaultCase(w)})}const h=k=>k,m=new Map,y=new WeakMap;let v;function S(k,A){return v||(v=d()),v(k,A)}function _(){return v||(v=d()),v.getInitialState()}function E(k,A=!1){function w(F){let I=F[k];return typeof I>"u"&&A&&(I=kE(y,w,_)),I}function R(F=h){const I=kE(m,A,()=>new WeakMap);return kE(I,F,()=>{const U={};for(const[q,V]of Object.entries(i.selectors??{}))U[q]=NVe(V,F,()=>kE(y,F,_),A);return U})}return{reducerPath:k,getSelectors:R,get selectors(){return R(w)},selectSlice:w}}const $={name:a,reducer:S,actions:s.actionCreators,caseReducers:s.sliceCaseReducersByName,getInitialState:_,...E(u),injectInto(k,{reducerPath:A,...w}={}){const R=A??u;return k.inject({reducerPath:R,reducer:S},w),{...$,...E(R,!0)}}};return $}}function NVe(e,t,n,r){function i(a,...u){let l=t(a);return typeof l>"u"&&r&&(l=n()),e(l,...u)}return i.unwrapped=e,i}var hlt=FVe();function IVe(){function e(t,n){return{_reducerDefinitionType:"asyncThunk",payloadCreator:t,...n}}return e.withTypes=()=>e,{reducer(t){return Object.assign({[t.name](...n){return t(...n)}}[t.name],{_reducerDefinitionType:"reducer"})},preparedReducer(t,n){return{_reducerDefinitionType:"reducerWithPrepare",prepare:t,reducer:n}},asyncThunk:e}}function LVe({type:e,reducerName:t,createNotation:n},r,i){let a,u;if("reducer"in r){if(n&&!QVe(r))throw new Error(Zf(17));a=r.reducer,u=r.prepare}else a=r;i.addCase(e,a).exposeCaseReducer(t,a).exposeAction(t,u?Q3(e,u):Q3(e))}function BVe(e){return e._reducerDefinitionType==="asyncThunk"}function QVe(e){return e._reducerDefinitionType==="reducerWithPrepare"}function zVe({type:e,reducerName:t},n,r,i){if(!i)throw new Error(Zf(18));const{payloadCreator:a,fulfilled:u,pending:l,rejected:o,settled:s,options:c}=n,d=i(e,a,c);r.exposeAction(t,d),u&&r.addCase(d.fulfilled,u),l&&r.addCase(d.pending,l),o&&r.addCase(d.rejected,o),s&&r.addMatcher(d.settled,s),r.exposeCaseReducer(t,{fulfilled:u||PE,pending:l||PE,rejected:o||PE,settled:s||PE})}function PE(){}var qVe="task",nye="listener",rye="completed",LL="cancelled",VVe=`task-${LL}`,UVe=`task-${rye}`,lF=`${nye}-${LL}`,jVe=`${nye}-${rye}`,e5=class{constructor(e){P_(this,"name","TaskAbortError");P_(this,"message");this.code=e,this.message=`${qVe} ${LL} (reason: ${e})`}},BL=(e,t)=>{if(typeof e!="function")throw new TypeError(Zf(32))},K$=()=>{},iye=(e,t=K$)=>(e.catch(t),e),aye=(e,t)=>(e.addEventListener("abort",t,{once:!0}),()=>e.removeEventListener("abort",t)),Yb=(e,t)=>{const n=e.signal;n.aborted||("reason"in n||Object.defineProperty(n,"reason",{enumerable:!0,value:t,configurable:!0,writable:!0}),e.abort(t))},Gb=e=>{if(e.aborted){const{reason:t}=e;throw new e5(t)}};function uye(e,t){let n=K$;return new Promise((r,i)=>{const a=()=>i(new e5(e.reason));if(e.aborted){a();return}n=aye(e,a),t.finally(()=>n()).then(r,i)}).finally(()=>{n=K$})}var ZVe=async(e,t)=>{try{return await Promise.resolve(),{status:"ok",value:await e()}}catch(n){return{status:n instanceof e5?"cancelled":"rejected",error:n}}finally{t==null||t()}},W$=e=>t=>iye(uye(e,t).then(n=>(Gb(e),n))),lye=e=>{const t=W$(e);return n=>t(new Promise(r=>setTimeout(r,n)))},{assign:x3}=Object,RG={},t5="listenerMiddleware",HVe=(e,t)=>{const n=r=>aye(e,()=>Yb(r,e.reason));return(r,i)=>{BL(r);const a=new AbortController;n(a);const u=ZVe(async()=>{Gb(e),Gb(a.signal);const l=await r({pause:W$(a.signal),delay:lye(a.signal),signal:a.signal});return Gb(a.signal),l},()=>Yb(a,UVe));return i!=null&&i.autoJoin&&t.push(u.catch(K$)),{result:W$(e)(u),cancel(){Yb(a,VVe)}}}},XVe=(e,t)=>{const n=async(r,i)=>{Gb(t);let a=()=>{};const l=[new Promise((o,s)=>{let c=e({predicate:r,effect:(d,h)=>{h.unsubscribe(),o([d,h.getState(),h.getOriginalState()])}});a=()=>{c(),s()}})];i!=null&&l.push(new Promise(o=>setTimeout(o,i,null)));try{const o=await uye(t,Promise.race(l));return Gb(t),o}finally{a()}};return(r,i)=>iye(n(r,i))},oye=e=>{let{type:t,actionCreator:n,matcher:r,predicate:i,effect:a}=e;if(t)i=Q3(t).match;else if(n)t=n.type,i=n.match;else if(r)i=r;else if(!i)throw new Error(Zf(21));return BL(a),{predicate:i,type:t,effect:a}},sye=x3(e=>{const{type:t,predicate:n,effect:r}=oye(e);return{id:kVe(),effect:r,type:t,predicate:n,pending:new Set,unsubscribe:()=>{throw new Error(Zf(22))}}},{withTypes:()=>sye}),AG=(e,t)=>{const{type:n,effect:r,predicate:i}=oye(t);return Array.from(e.values()).find(a=>(typeof n=="string"?a.type===n:a.predicate===i)&&a.effect===r)},oF=e=>{e.pending.forEach(t=>{Yb(t,lF)})},YVe=e=>()=>{e.forEach(oF),e.clear()},TG=(e,t,n)=>{try{e(t,n)}catch(r){setTimeout(()=>{throw r},0)}},cye=x3(Q3(`${t5}/add`),{withTypes:()=>cye}),GVe=Q3(`${t5}/removeAll`),fye=x3(Q3(`${t5}/remove`),{withTypes:()=>fye}),KVe=(...e)=>{console.error(`${t5}/error`,...e)},plt=(e={})=>{const t=new Map,{extra:n,onError:r=KVe}=e;BL(r);const i=c=>(c.unsubscribe=()=>t.delete(c.id),t.set(c.id,c),d=>{c.unsubscribe(),d!=null&&d.cancelActive&&oF(c)}),a=c=>{const d=AG(t,c)??sye(c);return i(d)};x3(a,{withTypes:()=>a});const u=c=>{const d=AG(t,c);return d&&(d.unsubscribe(),c.cancelActive&&oF(d)),!!d};x3(u,{withTypes:()=>u});const l=async(c,d,h,m)=>{const y=new AbortController,v=XVe(a,y.signal),S=[];try{c.pending.add(y),await Promise.resolve(c.effect(d,x3({},h,{getOriginalState:m,condition:(_,E)=>v(_,E).then(Boolean),take:v,delay:lye(y.signal),pause:W$(y.signal),extra:n,signal:y.signal,fork:HVe(y.signal,S),unsubscribe:c.unsubscribe,subscribe:()=>{t.set(c.id,c)},cancelActiveListeners:()=>{c.pending.forEach((_,E,$)=>{_!==y&&(Yb(_,lF),$.delete(_))})},cancel:()=>{Yb(y,lF),c.pending.delete(y)},throwIfCancelled:()=>{Gb(y.signal)}})))}catch(_){_ instanceof e5||TG(r,_,{raisedBy:"effect"})}finally{await Promise.all(S),Yb(y,jVe),c.pending.delete(y)}},o=YVe(t);return{middleware:c=>d=>h=>{if(!jve(h))return d(h);if(cye.match(h))return a(h.payload);if(GVe.match(h)){o();return}if(fye.match(h))return u(h.payload);let m=c.getState();const y=()=>{if(m===RG)throw new Error(Zf(23));return m};let v;try{if(v=d(h),t.size>0){const S=c.getState(),_=Array.from(t.values());for(const E of _){let $=!1;try{$=E.predicate(h,S,m)}catch(k){$=!1,TG(r,k,{raisedBy:"predicate"})}$&&l(E,h,c,y)}}}finally{m=RG}return v},startListening:a,stopListening:u,clearListeners:o}};function Zf(e){return`Minified Redux Toolkit error #${e}; visit https://redux-toolkit.js.org/Errors?code=${e} for the full message or use the non-minified dev environment for full errors. `}var I8={},L8={},B8={},kG;function WVe(){return kG||(kG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){if(!n||typeof n!="object")return!1;const r=Object.getPrototypeOf(n);return r===null||r===Object.prototype||Object.getPrototypeOf(r)===null?Object.prototype.toString.call(n)==="[object Object]":!1}e.isPlainObject=t})(B8)),B8}var PG;function JVe(){return PG||(PG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=WVe(),n=Lve(),r=PL(),i=ML(),a=TL();function u(s,c,d){return l(s,c,void 0,void 0,void 0,void 0,d)}function l(s,c,d,h,m,y,v){const S=v(s,c,d,h,m,y);if(S!==void 0)return S;if(typeof s==typeof c)switch(typeof s){case"bigint":case"string":case"boolean":case"symbol":case"undefined":return s===c;case"number":return s===c||Object.is(s,c);case"function":return s===c;case"object":return o(s,c,y,v)}return o(s,c,y,v)}function o(s,c,d,h){if(Object.is(s,c))return!0;let m=r.getTag(s),y=r.getTag(c);if(m===i.argumentsTag&&(m=i.objectTag),y===i.argumentsTag&&(y=i.objectTag),m!==y)return!1;switch(m){case i.stringTag:return s.toString()===c.toString();case i.numberTag:{const _=s.valueOf(),E=c.valueOf();return a.eq(_,E)}case i.booleanTag:case i.dateTag:case i.symbolTag:return Object.is(s.valueOf(),c.valueOf());case i.regexpTag:return s.source===c.source&&s.flags===c.flags;case i.functionTag:return s===c}d=d??new Map;const v=d.get(s),S=d.get(c);if(v!=null&&S!=null)return v===c;d.set(s,c),d.set(c,s);try{switch(m){case i.mapTag:{if(s.size!==c.size)return!1;for(const[_,E]of s.entries())if(!c.has(_)||!l(E,c.get(_),_,s,c,d,h))return!1;return!0}case i.setTag:{if(s.size!==c.size)return!1;const _=Array.from(s.values()),E=Array.from(c.values());for(let $=0;$<_.length;$++){const k=_[$],A=E.findIndex(w=>l(k,w,void 0,s,c,d,h));if(A===-1)return!1;E.splice(A,1)}return!0}case i.arrayTag:case i.uint8ArrayTag:case i.uint8ClampedArrayTag:case i.uint16ArrayTag:case i.uint32ArrayTag:case i.bigUint64ArrayTag:case i.int8ArrayTag:case i.int16ArrayTag:case i.int32ArrayTag:case i.bigInt64ArrayTag:case i.float32ArrayTag:case i.float64ArrayTag:{if(typeof Buffer<"u"&&Buffer.isBuffer(s)!==Buffer.isBuffer(c)||s.length!==c.length)return!1;for(let _=0;_<s.length;_++)if(!l(s[_],c[_],_,s,c,d,h))return!1;return!0}case i.arrayBufferTag:return s.byteLength!==c.byteLength?!1:o(new Uint8Array(s),new Uint8Array(c),d,h);case i.dataViewTag:return s.byteLength!==c.byteLength||s.byteOffset!==c.byteOffset?!1:o(new Uint8Array(s),new Uint8Array(c),d,h);case i.errorTag:return s.name===c.name&&s.message===c.message;case i.objectTag:{if(!(o(s.constructor,c.constructor,d,h)||t.isPlainObject(s)&&t.isPlainObject(c)))return!1;const E=[...Object.keys(s),...n.getSymbols(s)],$=[...Object.keys(c),...n.getSymbols(c)];if(E.length!==$.length)return!1;for(let k=0;k<E.length;k++){const A=E[k],w=s[A];if(!Object.hasOwn(c,A))return!1;const R=c[A];if(!l(w,R,A,s,c,d,h))return!1}return!0}default:return!1}}finally{d.delete(s),d.delete(c)}}e.isEqualWith=u})(L8)),L8}var Q8={},MG;function eUe(){return MG||(MG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(){}e.noop=t})(Q8)),Q8}var FG;function tUe(){return FG||(FG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=JVe(),n=eUe();function r(i,a){return t.isEqualWith(i,a,n.noop)}e.isEqual=r})(I8)),I8}var z8,NG;function nUe(){return NG||(NG=1,z8=tUe().isEqual),z8}var rUe=nUe();const mlt=da(rUe);var q8={},V8={},U8={},IG;function iUe(){return IG||(IG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=qve();function n(r){return t.isSymbol(r)?NaN:Number(r)}e.toNumber=n})(U8)),U8}var LG;function aUe(){return LG||(LG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=iUe();function n(r){return r?(r=t.toNumber(r),r===1/0||r===-1/0?(r<0?-1:1)*Number.MAX_VALUE:r===r?r:0):r===0?r:0}e.toFinite=n})(V8)),V8}var BG;function uUe(){return BG||(BG=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Vve(),n=aUe();function r(i,a,u){u&&typeof u!="number"&&t.isIterateeCall(i,a,u)&&(a=u=void 0),i=n.toFinite(i),a===void 0?(a=i,i=0):a=n.toFinite(a),u=u===void 0?i<a?1:-1:n.toFinite(u);const l=Math.max(Math.ceil((a-i)/(u||1)),0),o=new Array(l);for(let s=0;s<l;s++)o[s]=i,i+=u;return o}e.range=r})(q8)),q8}var j8,QG;function lUe(){return QG||(QG=1,j8=uUe().range),j8}var oUe=lUe();const glt=da(oUe);function Xv(e,t){return e==null||t==null?NaN:e<t?-1:e>t?1:e>=t?0:NaN}function sUe(e,t){return e==null||t==null?NaN:t<e?-1:t>e?1:t>=e?0:NaN}function QL(e){let t,n,r;e.length!==2?(t=Xv,n=(l,o)=>Xv(e(l),o),r=(l,o)=>e(l)-o):(t=e===Xv||e===sUe?e:cUe,n=e,r=e);function i(l,o,s=0,c=l.length){if(s<c){if(t(o,o)!==0)return c;do{const d=s+c>>>1;n(l[d],o)<0?s=d+1:c=d}while(s<c)}return s}function a(l,o,s=0,c=l.length){if(s<c){if(t(o,o)!==0)return c;do{const d=s+c>>>1;n(l[d],o)<=0?s=d+1:c=d}while(s<c)}return s}function u(l,o,s=0,c=l.length){const d=i(l,o,s,c-1);return d>s&&r(l[d-1],o)>-r(l[d],o)?d-1:d}return{left:i,center:u,right:a}}function cUe(){return 0}function dye(e){return e===null?NaN:+e}function*fUe(e,t){for(let n of e)n!=null&&(n=+n)>=n&&(yield n)}const dUe=QL(Xv),BS=dUe.right;QL(dye).center;class zG extends Map{constructor(t,n=mUe){if(super(),Object.defineProperties(this,{_intern:{value:new Map},_key:{value:n}}),t!=null)for(const[r,i]of t)this.set(r,i)}get(t){return super.get(qG(this,t))}has(t){return super.has(qG(this,t))}set(t,n){return super.set(hUe(this,t),n)}delete(t){return super.delete(pUe(this,t))}}function qG({_intern:e,_key:t},n){const r=t(n);return e.has(r)?e.get(r):n}function hUe({_intern:e,_key:t},n){const r=t(n);return e.has(r)?e.get(r):(e.set(r,n),n)}function pUe({_intern:e,_key:t},n){const r=t(n);return e.has(r)&&(n=e.get(r),e.delete(r)),n}function mUe(e){return e!==null&&typeof e=="object"?e.valueOf():e}function gUe(e=Xv){if(e===Xv)return hye;if(typeof e!="function")throw new TypeError("compare is not a function");return(t,n)=>{const r=e(t,n);return r||r===0?r:(e(n,n)===0)-(e(t,t)===0)}}function hye(e,t){return(e==null||!(e>=e))-(t==null||!(t>=t))||(e<t?-1:e>t?1:0)}const vUe=Math.sqrt(50),yUe=Math.sqrt(10),bUe=Math.sqrt(2);function J$(e,t,n){const r=(t-e)/Math.max(0,n),i=Math.floor(Math.log10(r)),a=r/Math.pow(10,i),u=a>=vUe?10:a>=yUe?5:a>=bUe?2:1;let l,o,s;return i<0?(s=Math.pow(10,-i)/u,l=Math.round(e*s),o=Math.round(t*s),l/s<e&&++l,o/s>t&&--o,s=-s):(s=Math.pow(10,i)*u,l=Math.round(e/s),o=Math.round(t/s),l*s<e&&++l,o*s>t&&--o),o<l&&.5<=n&&n<2?J$(e,t,n*2):[l,o,s]}function sF(e,t,n){if(t=+t,e=+e,n=+n,!(n>0))return[];if(e===t)return[e];const r=t<e,[i,a,u]=r?J$(t,e,n):J$(e,t,n);if(!(a>=i))return[];const l=a-i+1,o=new Array(l);if(r)if(u<0)for(let s=0;s<l;++s)o[s]=(a-s)/-u;else for(let s=0;s<l;++s)o[s]=(a-s)*u;else if(u<0)for(let s=0;s<l;++s)o[s]=(i+s)/-u;else for(let s=0;s<l;++s)o[s]=(i+s)*u;return o}function cF(e,t,n){return t=+t,e=+e,n=+n,J$(e,t,n)[2]}function fF(e,t,n){t=+t,e=+e,n=+n;const r=t<e,i=r?cF(t,e,n):cF(e,t,n);return(r?-1:1)*(i<0?1/-i:i)}function VG(e,t){let n;for(const r of e)r!=null&&(n<r||n===void 0&&r>=r)&&(n=r);return n}function UG(e,t){let n;for(const r of e)r!=null&&(n>r||n===void 0&&r>=r)&&(n=r);return n}function pye(e,t,n=0,r=1/0,i){if(t=Math.floor(t),n=Math.floor(Math.max(0,n)),r=Math.floor(Math.min(e.length-1,r)),!(n<=t&&t<=r))return e;for(i=i===void 0?hye:gUe(i);r>n;){if(r-n>600){const o=r-n+1,s=t-n+1,c=Math.log(o),d=.5*Math.exp(2*c/3),h=.5*Math.sqrt(c*d*(o-d)/o)*(s-o/2<0?-1:1),m=Math.max(n,Math.floor(t-s*d/o+h)),y=Math.min(r,Math.floor(t+(o-s)*d/o+h));pye(e,t,m,y,i)}const a=e[t];let u=n,l=r;for(x2(e,n,t),i(e[r],a)>0&&x2(e,n,r);u<l;){for(x2(e,u,l),++u,--l;i(e[u],a)<0;)++u;for(;i(e[l],a)>0;)--l}i(e[n],a)===0?x2(e,n,l):(++l,x2(e,l,r)),l<=t&&(n=l+1),t<=l&&(r=l-1)}return e}function x2(e,t,n){const r=e[t];e[t]=e[n],e[n]=r}function OUe(e,t,n){if(e=Float64Array.from(fUe(e)),!(!(r=e.length)||isNaN(t=+t))){if(t<=0||r<2)return UG(e);if(t>=1)return VG(e);var r,i=(r-1)*t,a=Math.floor(i),u=VG(pye(e,a).subarray(0,a+1)),l=UG(e.subarray(a+1));return u+(l-u)*(i-a)}}function xUe(e,t,n=dye){if(!(!(r=e.length)||isNaN(t=+t))){if(t<=0||r<2)return+n(e[0],0,e);if(t>=1)return+n(e[r-1],r-1,e);var r,i=(r-1)*t,a=Math.floor(i),u=+n(e[a],a,e),l=+n(e[a+1],a+1,e);return u+(l-u)*(i-a)}}function SUe(e,t,n){e=+e,t=+t,n=(i=arguments.length)<2?(t=e,e=0,1):i<3?1:+n;for(var r=-1,i=Math.max(0,Math.ceil((t-e)/n))|0,a=new Array(i);++r<i;)a[r]=e+r*n;return a}function oh(e,t){switch(arguments.length){case 0:break;case 1:this.range(e);break;default:this.range(t).domain(e);break}return this}function Sg(e,t){switch(arguments.length){case 0:break;case 1:{typeof e=="function"?this.interpolator(e):this.range(e);break}default:{this.domain(e),typeof t=="function"?this.interpolator(t):this.range(t);break}}return this}const dF=Symbol("implicit");function zL(){var e=new zG,t=[],n=[],r=dF;function i(a){let u=e.get(a);if(u===void 0){if(r!==dF)return r;e.set(a,u=t.push(a)-1)}return n[u%n.length]}return i.domain=function(a){if(!arguments.length)return t.slice();t=[],e=new zG;for(const u of a)e.has(u)||e.set(u,t.push(u)-1);return i},i.range=function(a){return arguments.length?(n=Array.from(a),i):n.slice()},i.unknown=function(a){return arguments.length?(r=a,i):r},i.copy=function(){return zL(t,n).unknown(r)},oh.apply(i,arguments),i}function qL(){var e=zL().unknown(void 0),t=e.domain,n=e.range,r=0,i=1,a,u,l=!1,o=0,s=0,c=.5;delete e.unknown;function d(){var h=t().length,m=i<r,y=m?i:r,v=m?r:i;a=(v-y)/Math.max(1,h-o+s*2),l&&(a=Math.floor(a)),y+=(v-y-a*(h-o))*c,u=a*(1-o),l&&(y=Math.round(y),u=Math.round(u));var S=SUe(h).map(function(_){return y+a*_});return n(m?S.reverse():S)}return e.domain=function(h){return arguments.length?(t(h),d()):t()},e.range=function(h){return arguments.length?([r,i]=h,r=+r,i=+i,d()):[r,i]},e.rangeRound=function(h){return[r,i]=h,r=+r,i=+i,l=!0,d()},e.bandwidth=function(){return u},e.step=function(){return a},e.round=function(h){return arguments.length?(l=!!h,d()):l},e.padding=function(h){return arguments.length?(o=Math.min(1,s=+h),d()):o},e.paddingInner=function(h){return arguments.length?(o=Math.min(1,h),d()):o},e.paddingOuter=function(h){return arguments.length?(s=+h,d()):s},e.align=function(h){return arguments.length?(c=Math.max(0,Math.min(1,h)),d()):c},e.copy=function(){return qL(t(),[r,i]).round(l).paddingInner(o).paddingOuter(s).align(c)},oh.apply(d(),arguments)}function mye(e){var t=e.copy;return e.padding=e.paddingOuter,delete e.paddingInner,delete e.paddingOuter,e.copy=function(){return mye(t())},e}function _Ue(){return mye(qL.apply(null,arguments).paddingInner(1))}function EUe(e){return function(){return e}}function eC(e){return+e}var jG=[0,1];function dc(e){return e}function hF(e,t){return(t-=e=+e)?function(n){return(n-e)/t}:EUe(isNaN(t)?NaN:.5)}function DUe(e,t){var n;return e>t&&(n=e,e=t,t=n),function(r){return Math.max(e,Math.min(t,r))}}function $Ue(e,t,n){var r=e[0],i=e[1],a=t[0],u=t[1];return i<r?(r=hF(i,r),a=n(u,a)):(r=hF(r,i),a=n(a,u)),function(l){return a(r(l))}}function CUe(e,t,n){var r=Math.min(e.length,t.length)-1,i=new Array(r),a=new Array(r),u=-1;for(e[r]<e[0]&&(e=e.slice().reverse(),t=t.slice().reverse());++u<r;)i[u]=hF(e[u],e[u+1]),a[u]=n(t[u],t[u+1]);return function(l){var o=BS(e,l,1,r)-1;return a[o](i[o](l))}}function QS(e,t){return t.domain(e.domain()).range(e.range()).interpolate(e.interpolate()).clamp(e.clamp()).unknown(e.unknown())}function n5(){var e=jG,t=jG,n=r4,r,i,a,u=dc,l,o,s;function c(){var h=Math.min(e.length,t.length);return u!==dc&&(u=DUe(e[0],e[h-1])),l=h>2?CUe:$Ue,o=s=null,d}function d(h){return h==null||isNaN(h=+h)?a:(o||(o=l(e.map(r),t,n)))(r(u(h)))}return d.invert=function(h){return u(i((s||(s=l(t,e.map(r),x$)))(h)))},d.domain=function(h){return arguments.length?(e=Array.from(h,eC),c()):e.slice()},d.range=function(h){return arguments.length?(t=Array.from(h),c()):t.slice()},d.rangeRound=function(h){return t=Array.from(h),n=BI,c()},d.clamp=function(h){return arguments.length?(u=h?!0:dc,c()):u!==dc},d.interpolate=function(h){return arguments.length?(n=h,c()):n},d.unknown=function(h){return arguments.length?(a=h,d):a},function(h,m){return r=h,i=m,c()}}function VL(){return n5()(dc,dc)}function gye(e,t,n,r){var i=fF(e,t,n),a;switch(r=Mx(r??",f"),r.type){case"s":{var u=Math.max(Math.abs(e),Math.abs(t));return r.precision==null&&!isNaN(a=J8e(i,u))&&(r.precision=a),y0e(r,u)}case"":case"e":case"g":case"p":case"r":{r.precision==null&&!isNaN(a=e7e(i,Math.max(Math.abs(e),Math.abs(t))))&&(r.precision=a-(r.type==="e"));break}case"f":case"%":{r.precision==null&&!isNaN(a=W8e(i))&&(r.precision=a-(r.type==="%")*2);break}}return zI(r)}function Sy(e){var t=e.domain;return e.ticks=function(n){var r=t();return sF(r[0],r[r.length-1],n??10)},e.tickFormat=function(n,r){var i=t();return gye(i[0],i[i.length-1],n??10,r)},e.nice=function(n){n==null&&(n=10);var r=t(),i=0,a=r.length-1,u=r[i],l=r[a],o,s,c=10;for(l<u&&(s=u,u=l,l=s,s=i,i=a,a=s);c-- >0;){if(s=cF(u,l,n),s===o)return r[i]=u,r[a]=l,t(r);if(s>0)u=Math.floor(u/s)*s,l=Math.ceil(l/s)*s;else if(s<0)u=Math.ceil(u*s)/s,l=Math.floor(l*s)/s;else break;o=s}return e},e}function vye(){var e=VL();return e.copy=function(){return QS(e,vye())},oh.apply(e,arguments),Sy(e)}function yye(e){var t;function n(r){return r==null||isNaN(r=+r)?t:r}return n.invert=n,n.domain=n.range=function(r){return arguments.length?(e=Array.from(r,eC),n):e.slice()},n.unknown=function(r){return arguments.length?(t=r,n):t},n.copy=function(){return yye(e).unknown(t)},e=arguments.length?Array.from(e,eC):[0,1],Sy(n)}function bye(e,t){e=e.slice();var n=0,r=e.length-1,i=e[n],a=e[r],u;return a<i&&(u=n,n=r,r=u,u=i,i=a,a=u),e[n]=t.floor(i),e[r]=t.ceil(a),e}function ZG(e){return Math.log(e)}function HG(e){return Math.exp(e)}function wUe(e){return-Math.log(-e)}function RUe(e){return-Math.exp(-e)}function AUe(e){return isFinite(e)?+("1e"+e):e<0?0:e}function TUe(e){return e===10?AUe:e===Math.E?Math.exp:t=>Math.pow(e,t)}function kUe(e){return e===Math.E?Math.log:e===10&&Math.log10||e===2&&Math.log2||(e=Math.log(e),t=>Math.log(t)/e)}function XG(e){return(t,n)=>-e(-t,n)}function UL(e){const t=e(ZG,HG),n=t.domain;let r=10,i,a;function u(){return i=kUe(r),a=TUe(r),n()[0]<0?(i=XG(i),a=XG(a),e(wUe,RUe)):e(ZG,HG),t}return t.base=function(l){return arguments.length?(r=+l,u()):r},t.domain=function(l){return arguments.length?(n(l),u()):n()},t.ticks=l=>{const o=n();let s=o[0],c=o[o.length-1];const d=c<s;d&&([s,c]=[c,s]);let h=i(s),m=i(c),y,v;const S=l==null?10:+l;let _=[];if(!(r%1)&&m-h<S){if(h=Math.floor(h),m=Math.ceil(m),s>0){for(;h<=m;++h)for(y=1;y<r;++y)if(v=h<0?y/a(-h):y*a(h),!(v<s)){if(v>c)break;_.push(v)}}else for(;h<=m;++h)for(y=r-1;y>=1;--y)if(v=h>0?y/a(-h):y*a(h),!(v<s)){if(v>c)break;_.push(v)}_.length*2<S&&(_=sF(s,c,S))}else _=sF(h,m,Math.min(m-h,S)).map(a);return d?_.reverse():_},t.tickFormat=(l,o)=>{if(l==null&&(l=10),o==null&&(o=r===10?"s":","),typeof o!="function"&&(!(r%1)&&(o=Mx(o)).precision==null&&(o.trim=!0),o=zI(o)),l===1/0)return o;const s=Math.max(1,r*l/t.ticks().length);return c=>{let d=c/a(Math.round(i(c)));return d*r<r-.5&&(d*=r),d<=s?o(c):""}},t.nice=()=>n(bye(n(),{floor:l=>a(Math.floor(i(l))),ceil:l=>a(Math.ceil(i(l)))})),t}function Oye(){const e=UL(n5()).domain([1,10]);return e.copy=()=>QS(e,Oye()).base(e.base()),oh.apply(e,arguments),e}function YG(e){return function(t){return Math.sign(t)*Math.log1p(Math.abs(t/e))}}function GG(e){return function(t){return Math.sign(t)*Math.expm1(Math.abs(t))*e}}function jL(e){var t=1,n=e(YG(t),GG(t));return n.constant=function(r){return arguments.length?e(YG(t=+r),GG(t)):t},Sy(n)}function xye(){var e=jL(n5());return e.copy=function(){return QS(e,xye()).constant(e.constant())},oh.apply(e,arguments)}function KG(e){return function(t){return t<0?-Math.pow(-t,e):Math.pow(t,e)}}function PUe(e){return e<0?-Math.sqrt(-e):Math.sqrt(e)}function MUe(e){return e<0?-e*e:e*e}function ZL(e){var t=e(dc,dc),n=1;function r(){return n===1?e(dc,dc):n===.5?e(PUe,MUe):e(KG(n),KG(1/n))}return t.exponent=function(i){return arguments.length?(n=+i,r()):n},Sy(t)}function HL(){var e=ZL(n5());return e.copy=function(){return QS(e,HL()).exponent(e.exponent())},oh.apply(e,arguments),e}function FUe(){return HL.apply(null,arguments).exponent(.5)}function WG(e){return Math.sign(e)*e*e}function NUe(e){return Math.sign(e)*Math.sqrt(Math.abs(e))}function Sye(){var e=VL(),t=[0,1],n=!1,r;function i(a){var u=NUe(e(a));return isNaN(u)?r:n?Math.round(u):u}return i.invert=function(a){return e.invert(WG(a))},i.domain=function(a){return arguments.length?(e.domain(a),i):e.domain()},i.range=function(a){return arguments.length?(e.range((t=Array.from(a,eC)).map(WG)),i):t.slice()},i.rangeRound=function(a){return i.range(a).round(!0)},i.round=function(a){return arguments.length?(n=!!a,i):n},i.clamp=function(a){return arguments.length?(e.clamp(a),i):e.clamp()},i.unknown=function(a){return arguments.length?(r=a,i):r},i.copy=function(){return Sye(e.domain(),t).round(n).clamp(e.clamp()).unknown(r)},oh.apply(i,arguments),Sy(i)}function _ye(){var e=[],t=[],n=[],r;function i(){var u=0,l=Math.max(1,t.length);for(n=new Array(l-1);++u<l;)n[u-1]=xUe(e,u/l);return a}function a(u){return u==null||isNaN(u=+u)?r:t[BS(n,u)]}return a.invertExtent=function(u){var l=t.indexOf(u);return l<0?[NaN,NaN]:[l>0?n[l-1]:e[0],l<n.length?n[l]:e[e.length-1]]},a.domain=function(u){if(!arguments.length)return e.slice();e=[];for(let l of u)l!=null&&!isNaN(l=+l)&&e.push(l);return e.sort(Xv),i()},a.range=function(u){return arguments.length?(t=Array.from(u),i()):t.slice()},a.unknown=function(u){return arguments.length?(r=u,a):r},a.quantiles=function(){return n.slice()},a.copy=function(){return _ye().domain(e).range(t).unknown(r)},oh.apply(a,arguments)}function Eye(){var e=0,t=1,n=1,r=[.5],i=[0,1],a;function u(o){return o!=null&&o<=o?i[BS(r,o,0,n)]:a}function l(){var o=-1;for(r=new Array(n);++o<n;)r[o]=((o+1)*t-(o-n)*e)/(n+1);return u}return u.domain=function(o){return arguments.length?([e,t]=o,e=+e,t=+t,l()):[e,t]},u.range=function(o){return arguments.length?(n=(i=Array.from(o)).length-1,l()):i.slice()},u.invertExtent=function(o){var s=i.indexOf(o);return s<0?[NaN,NaN]:s<1?[e,r[0]]:s>=n?[r[n-1],t]:[r[s-1],r[s]]},u.unknown=function(o){return arguments.length&&(a=o),u},u.thresholds=function(){return r.slice()},u.copy=function(){return Eye().domain([e,t]).range(i).unknown(a)},oh.apply(Sy(u),arguments)}function Dye(){var e=[.5],t=[0,1],n,r=1;function i(a){return a!=null&&a<=a?t[BS(e,a,0,r)]:n}return i.domain=function(a){return arguments.length?(e=Array.from(a),r=Math.min(e.length,t.length-1),i):e.slice()},i.range=function(a){return arguments.length?(t=Array.from(a),r=Math.min(e.length,t.length-1),i):t.slice()},i.invertExtent=function(a){var u=t.indexOf(a);return[e[u-1],e[u]]},i.unknown=function(a){return arguments.length?(n=a,i):n},i.copy=function(){return Dye().domain(e).range(t).unknown(n)},oh.apply(i,arguments)}const Z8=new Date,H8=new Date;function Fl(e,t,n,r){function i(a){return e(a=arguments.length===0?new Date:new Date(+a)),a}return i.floor=a=>(e(a=new Date(+a)),a),i.ceil=a=>(e(a=new Date(a-1)),t(a,1),e(a),a),i.round=a=>{const u=i(a),l=i.ceil(a);return a-u<l-a?u:l},i.offset=(a,u)=>(t(a=new Date(+a),u==null?1:Math.floor(u)),a),i.range=(a,u,l)=>{const o=[];if(a=i.ceil(a),l=l==null?1:Math.floor(l),!(a<u)||!(l>0))return o;let s;do o.push(s=new Date(+a)),t(a,l),e(a);while(s<a&&a<u);return o},i.filter=a=>Fl(u=>{if(u>=u)for(;e(u),!a(u);)u.setTime(u-1)},(u,l)=>{if(u>=u)if(l<0)for(;++l<=0;)for(;t(u,-1),!a(u););else for(;--l>=0;)for(;t(u,1),!a(u););}),n&&(i.count=(a,u)=>(Z8.setTime(+a),H8.setTime(+u),e(Z8),e(H8),Math.floor(n(Z8,H8))),i.every=a=>(a=Math.floor(a),!isFinite(a)||!(a>0)?null:a>1?i.filter(r?u=>r(u)%a===0:u=>i.count(0,u)%a===0):i)),i}const tC=Fl(()=>{},(e,t)=>{e.setTime(+e+t)},(e,t)=>t-e);tC.every=e=>(e=Math.floor(e),!isFinite(e)||!(e>0)?null:e>1?Fl(t=>{t.setTime(Math.floor(t/e)*e)},(t,n)=>{t.setTime(+t+n*e)},(t,n)=>(n-t)/e):tC);tC.range;const tg=1e3,Gd=tg*60,ng=Gd*60,pg=ng*24,XL=pg*7,JG=pg*30,X8=pg*365,jb=Fl(e=>{e.setTime(e-e.getMilliseconds())},(e,t)=>{e.setTime(+e+t*tg)},(e,t)=>(t-e)/tg,e=>e.getUTCSeconds());jb.range;const YL=Fl(e=>{e.setTime(e-e.getMilliseconds()-e.getSeconds()*tg)},(e,t)=>{e.setTime(+e+t*Gd)},(e,t)=>(t-e)/Gd,e=>e.getMinutes());YL.range;const GL=Fl(e=>{e.setUTCSeconds(0,0)},(e,t)=>{e.setTime(+e+t*Gd)},(e,t)=>(t-e)/Gd,e=>e.getUTCMinutes());GL.range;const KL=Fl(e=>{e.setTime(e-e.getMilliseconds()-e.getSeconds()*tg-e.getMinutes()*Gd)},(e,t)=>{e.setTime(+e+t*ng)},(e,t)=>(t-e)/ng,e=>e.getHours());KL.range;const WL=Fl(e=>{e.setUTCMinutes(0,0,0)},(e,t)=>{e.setTime(+e+t*ng)},(e,t)=>(t-e)/ng,e=>e.getUTCHours());WL.range;const zS=Fl(e=>e.setHours(0,0,0,0),(e,t)=>e.setDate(e.getDate()+t),(e,t)=>(t-e-(t.getTimezoneOffset()-e.getTimezoneOffset())*Gd)/pg,e=>e.getDate()-1);zS.range;const r5=Fl(e=>{e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCDate(e.getUTCDate()+t)},(e,t)=>(t-e)/pg,e=>e.getUTCDate()-1);r5.range;const $ye=Fl(e=>{e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCDate(e.getUTCDate()+t)},(e,t)=>(t-e)/pg,e=>Math.floor(e/pg));$ye.range;function F1(e){return Fl(t=>{t.setDate(t.getDate()-(t.getDay()+7-e)%7),t.setHours(0,0,0,0)},(t,n)=>{t.setDate(t.getDate()+n*7)},(t,n)=>(n-t-(n.getTimezoneOffset()-t.getTimezoneOffset())*Gd)/XL)}const i5=F1(0),nC=F1(1),IUe=F1(2),LUe=F1(3),z3=F1(4),BUe=F1(5),QUe=F1(6);i5.range;nC.range;IUe.range;LUe.range;z3.range;BUe.range;QUe.range;function N1(e){return Fl(t=>{t.setUTCDate(t.getUTCDate()-(t.getUTCDay()+7-e)%7),t.setUTCHours(0,0,0,0)},(t,n)=>{t.setUTCDate(t.getUTCDate()+n*7)},(t,n)=>(n-t)/XL)}const a5=N1(0),rC=N1(1),zUe=N1(2),qUe=N1(3),q3=N1(4),VUe=N1(5),UUe=N1(6);a5.range;rC.range;zUe.range;qUe.range;q3.range;VUe.range;UUe.range;const JL=Fl(e=>{e.setDate(1),e.setHours(0,0,0,0)},(e,t)=>{e.setMonth(e.getMonth()+t)},(e,t)=>t.getMonth()-e.getMonth()+(t.getFullYear()-e.getFullYear())*12,e=>e.getMonth());JL.range;const eB=Fl(e=>{e.setUTCDate(1),e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCMonth(e.getUTCMonth()+t)},(e,t)=>t.getUTCMonth()-e.getUTCMonth()+(t.getUTCFullYear()-e.getUTCFullYear())*12,e=>e.getUTCMonth());eB.range;const mg=Fl(e=>{e.setMonth(0,1),e.setHours(0,0,0,0)},(e,t)=>{e.setFullYear(e.getFullYear()+t)},(e,t)=>t.getFullYear()-e.getFullYear(),e=>e.getFullYear());mg.every=e=>!isFinite(e=Math.floor(e))||!(e>0)?null:Fl(t=>{t.setFullYear(Math.floor(t.getFullYear()/e)*e),t.setMonth(0,1),t.setHours(0,0,0,0)},(t,n)=>{t.setFullYear(t.getFullYear()+n*e)});mg.range;const gg=Fl(e=>{e.setUTCMonth(0,1),e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCFullYear(e.getUTCFullYear()+t)},(e,t)=>t.getUTCFullYear()-e.getUTCFullYear(),e=>e.getUTCFullYear());gg.every=e=>!isFinite(e=Math.floor(e))||!(e>0)?null:Fl(t=>{t.setUTCFullYear(Math.floor(t.getUTCFullYear()/e)*e),t.setUTCMonth(0,1),t.setUTCHours(0,0,0,0)},(t,n)=>{t.setUTCFullYear(t.getUTCFullYear()+n*e)});gg.range;function Cye(e,t,n,r,i,a){const u=[[jb,1,tg],[jb,5,5*tg],[jb,15,15*tg],[jb,30,30*tg],[a,1,Gd],[a,5,5*Gd],[a,15,15*Gd],[a,30,30*Gd],[i,1,ng],[i,3,3*ng],[i,6,6*ng],[i,12,12*ng],[r,1,pg],[r,2,2*pg],[n,1,XL],[t,1,JG],[t,3,3*JG],[e,1,X8]];function l(s,c,d){const h=c<s;h&&([s,c]=[c,s]);const m=d&&typeof d.range=="function"?d:o(s,c,d),y=m?m.range(s,+c+1):[];return h?y.reverse():y}function o(s,c,d){const h=Math.abs(c-s)/d,m=QL(([,,S])=>S).right(u,h);if(m===u.length)return e.every(fF(s/X8,c/X8,d));if(m===0)return tC.every(Math.max(fF(s,c,d),1));const[y,v]=u[h/u[m-1][2]<u[m][2]/h?m-1:m];return y.every(v)}return[l,o]}const[jUe,ZUe]=Cye(gg,eB,a5,$ye,WL,GL),[HUe,XUe]=Cye(mg,JL,i5,zS,KL,YL);function Y8(e){if(0<=e.y&&e.y<100){var t=new Date(-1,e.m,e.d,e.H,e.M,e.S,e.L);return t.setFullYear(e.y),t}return new Date(e.y,e.m,e.d,e.H,e.M,e.S,e.L)}function G8(e){if(0<=e.y&&e.y<100){var t=new Date(Date.UTC(-1,e.m,e.d,e.H,e.M,e.S,e.L));return t.setUTCFullYear(e.y),t}return new Date(Date.UTC(e.y,e.m,e.d,e.H,e.M,e.S,e.L))}function S2(e,t,n){return{y:e,m:t,d:n,H:0,M:0,S:0,L:0}}function YUe(e){var t=e.dateTime,n=e.date,r=e.time,i=e.periods,a=e.days,u=e.shortDays,l=e.months,o=e.shortMonths,s=_2(i),c=E2(i),d=_2(a),h=E2(a),m=_2(u),y=E2(u),v=_2(l),S=E2(l),_=_2(o),E=E2(o),$={a:J,A:te,b:H,B:M,c:null,d:aK,e:aK,f:yje,g:wje,G:Aje,H:mje,I:gje,j:vje,L:wye,m:bje,M:Oje,p:G,q:X,Q:oK,s:sK,S:xje,u:Sje,U:_je,V:Eje,w:Dje,W:$je,x:null,X:null,y:Cje,Y:Rje,Z:Tje,"%":lK},k={a:B,A:ne,b:oe,B:ie,c:null,d:uK,e:uK,f:Fje,g:jje,G:Hje,H:kje,I:Pje,j:Mje,L:Aye,m:Nje,M:Ije,p:de,q:ge,Q:oK,s:sK,S:Lje,u:Bje,U:Qje,V:zje,w:qje,W:Vje,x:null,X:null,y:Uje,Y:Zje,Z:Xje,"%":lK},A={a:U,A:q,b:V,B:N,c:j,d:rK,e:rK,f:fje,g:nK,G:tK,H:iK,I:iK,j:lje,L:cje,m:uje,M:oje,p:I,q:aje,Q:hje,s:pje,S:sje,u:eje,U:tje,V:nje,w:JUe,W:rje,x:W,X:K,y:nK,Y:tK,Z:ije,"%":dje};$.x=w(n,$),$.X=w(r,$),$.c=w(t,$),k.x=w(n,k),k.X=w(r,k),k.c=w(t,k);function w(ye,me){return function(Ee){var se=[],pe=-1,ve=0,we=ye.length,Me,Re,et;for(Ee instanceof Date||(Ee=new Date(+Ee));++pe<we;)ye.charCodeAt(pe)===37&&(se.push(ye.slice(ve,pe)),(Re=eK[Me=ye.charAt(++pe)])!=null?Me=ye.charAt(++pe):Re=Me==="e"?" ":"0",(et=me[Me])&&(Me=et(Ee,Re)),se.push(Me),ve=pe+1);return se.push(ye.slice(ve,pe)),se.join("")}}function R(ye,me){return function(Ee){var se=S2(1900,void 0,1),pe=F(se,ye,Ee+="",0),ve,we;if(pe!=Ee.length)return null;if("Q"in se)return new Date(se.Q);if("s"in se)return new Date(se.s*1e3+("L"in se?se.L:0));if(me&&!("Z"in se)&&(se.Z=0),"p"in se&&(se.H=se.H%12+se.p*12),se.m===void 0&&(se.m="q"in se?se.q:0),"V"in se){if(se.V<1||se.V>53)return null;"w"in se||(se.w=1),"Z"in se?(ve=G8(S2(se.y,0,1)),we=ve.getUTCDay(),ve=we>4||we===0?rC.ceil(ve):rC(ve),ve=r5.offset(ve,(se.V-1)*7),se.y=ve.getUTCFullYear(),se.m=ve.getUTCMonth(),se.d=ve.getUTCDate()+(se.w+6)%7):(ve=Y8(S2(se.y,0,1)),we=ve.getDay(),ve=we>4||we===0?nC.ceil(ve):nC(ve),ve=zS.offset(ve,(se.V-1)*7),se.y=ve.getFullYear(),se.m=ve.getMonth(),se.d=ve.getDate()+(se.w+6)%7)}else("W"in se||"U"in se)&&("w"in se||(se.w="u"in se?se.u%7:"W"in se?1:0),we="Z"in se?G8(S2(se.y,0,1)).getUTCDay():Y8(S2(se.y,0,1)).getDay(),se.m=0,se.d="W"in se?(se.w+6)%7+se.W*7-(we+5)%7:se.w+se.U*7-(we+6)%7);return"Z"in se?(se.H+=se.Z/100|0,se.M+=se.Z%100,G8(se)):Y8(se)}}function F(ye,me,Ee,se){for(var pe=0,ve=me.length,we=Ee.length,Me,Re;pe<ve;){if(se>=we)return-1;if(Me=me.charCodeAt(pe++),Me===37){if(Me=me.charAt(pe++),Re=A[Me in eK?me.charAt(pe++):Me],!Re||(se=Re(ye,Ee,se))<0)return-1}else if(Me!=Ee.charCodeAt(se++))return-1}return se}function I(ye,me,Ee){var se=s.exec(me.slice(Ee));return se?(ye.p=c.get(se[0].toLowerCase()),Ee+se[0].length):-1}function U(ye,me,Ee){var se=m.exec(me.slice(Ee));return se?(ye.w=y.get(se[0].toLowerCase()),Ee+se[0].length):-1}function q(ye,me,Ee){var se=d.exec(me.slice(Ee));return se?(ye.w=h.get(se[0].toLowerCase()),Ee+se[0].length):-1}function V(ye,me,Ee){var se=_.exec(me.slice(Ee));return se?(ye.m=E.get(se[0].toLowerCase()),Ee+se[0].length):-1}function N(ye,me,Ee){var se=v.exec(me.slice(Ee));return se?(ye.m=S.get(se[0].toLowerCase()),Ee+se[0].length):-1}function j(ye,me,Ee){return F(ye,t,me,Ee)}function W(ye,me,Ee){return F(ye,n,me,Ee)}function K(ye,me,Ee){return F(ye,r,me,Ee)}function J(ye){return u[ye.getDay()]}function te(ye){return a[ye.getDay()]}function H(ye){return o[ye.getMonth()]}function M(ye){return l[ye.getMonth()]}function G(ye){return i[+(ye.getHours()>=12)]}function X(ye){return 1+~~(ye.getMonth()/3)}function B(ye){return u[ye.getUTCDay()]}function ne(ye){return a[ye.getUTCDay()]}function oe(ye){return o[ye.getUTCMonth()]}function ie(ye){return l[ye.getUTCMonth()]}function de(ye){return i[+(ye.getUTCHours()>=12)]}function ge(ye){return 1+~~(ye.getUTCMonth()/3)}return{format:function(ye){var me=w(ye+="",$);return me.toString=function(){return ye},me},parse:function(ye){var me=R(ye+="",!1);return me.toString=function(){return ye},me},utcFormat:function(ye){var me=w(ye+="",k);return me.toString=function(){return ye},me},utcParse:function(ye){var me=R(ye+="",!0);return me.toString=function(){return ye},me}}}var eK={"-":"",_:" ",0:"0"},fo=/^\s*\d+/,GUe=/^%/,KUe=/[\\^$*+?|[\]().{}]/g;function Ei(e,t,n){var r=e<0?"-":"",i=(r?-e:e)+"",a=i.length;return r+(a<n?new Array(n-a+1).join(t)+i:i)}function WUe(e){return e.replace(KUe,"\\$&")}function _2(e){return new RegExp("^(?:"+e.map(WUe).join("|")+")","i")}function E2(e){return new Map(e.map((t,n)=>[t.toLowerCase(),n]))}function JUe(e,t,n){var r=fo.exec(t.slice(n,n+1));return r?(e.w=+r[0],n+r[0].length):-1}function eje(e,t,n){var r=fo.exec(t.slice(n,n+1));return r?(e.u=+r[0],n+r[0].length):-1}function tje(e,t,n){var r=fo.exec(t.slice(n,n+2));return r?(e.U=+r[0],n+r[0].length):-1}function nje(e,t,n){var r=fo.exec(t.slice(n,n+2));return r?(e.V=+r[0],n+r[0].length):-1}function rje(e,t,n){var r=fo.exec(t.slice(n,n+2));return r?(e.W=+r[0],n+r[0].length):-1}function tK(e,t,n){var r=fo.exec(t.slice(n,n+4));return r?(e.y=+r[0],n+r[0].length):-1}function nK(e,t,n){var r=fo.exec(t.slice(n,n+2));return r?(e.y=+r[0]+(+r[0]>68?1900:2e3),n+r[0].length):-1}function ije(e,t,n){var r=/^(Z)|([+-]\d\d)(?::?(\d\d))?/.exec(t.slice(n,n+6));return r?(e.Z=r[1]?0:-(r[2]+(r[3]||"00")),n+r[0].length):-1}function aje(e,t,n){var r=fo.exec(t.slice(n,n+1));return r?(e.q=r[0]*3-3,n+r[0].length):-1}function uje(e,t,n){var r=fo.exec(t.slice(n,n+2));return r?(e.m=r[0]-1,n+r[0].length):-1}function rK(e,t,n){var r=fo.exec(t.slice(n,n+2));return r?(e.d=+r[0],n+r[0].length):-1}function lje(e,t,n){var r=fo.exec(t.slice(n,n+3));return r?(e.m=0,e.d=+r[0],n+r[0].length):-1}function iK(e,t,n){var r=fo.exec(t.slice(n,n+2));return r?(e.H=+r[0],n+r[0].length):-1}function oje(e,t,n){var r=fo.exec(t.slice(n,n+2));return r?(e.M=+r[0],n+r[0].length):-1}function sje(e,t,n){var r=fo.exec(t.slice(n,n+2));return r?(e.S=+r[0],n+r[0].length):-1}function cje(e,t,n){var r=fo.exec(t.slice(n,n+3));return r?(e.L=+r[0],n+r[0].length):-1}function fje(e,t,n){var r=fo.exec(t.slice(n,n+6));return r?(e.L=Math.floor(r[0]/1e3),n+r[0].length):-1}function dje(e,t,n){var r=GUe.exec(t.slice(n,n+1));return r?n+r[0].length:-1}function hje(e,t,n){var r=fo.exec(t.slice(n));return r?(e.Q=+r[0],n+r[0].length):-1}function pje(e,t,n){var r=fo.exec(t.slice(n));return r?(e.s=+r[0],n+r[0].length):-1}function aK(e,t){return Ei(e.getDate(),t,2)}function mje(e,t){return Ei(e.getHours(),t,2)}function gje(e,t){return Ei(e.getHours()%12||12,t,2)}function vje(e,t){return Ei(1+zS.count(mg(e),e),t,3)}function wye(e,t){return Ei(e.getMilliseconds(),t,3)}function yje(e,t){return wye(e,t)+"000"}function bje(e,t){return Ei(e.getMonth()+1,t,2)}function Oje(e,t){return Ei(e.getMinutes(),t,2)}function xje(e,t){return Ei(e.getSeconds(),t,2)}function Sje(e){var t=e.getDay();return t===0?7:t}function _je(e,t){return Ei(i5.count(mg(e)-1,e),t,2)}function Rye(e){var t=e.getDay();return t>=4||t===0?z3(e):z3.ceil(e)}function Eje(e,t){return e=Rye(e),Ei(z3.count(mg(e),e)+(mg(e).getDay()===4),t,2)}function Dje(e){return e.getDay()}function $je(e,t){return Ei(nC.count(mg(e)-1,e),t,2)}function Cje(e,t){return Ei(e.getFullYear()%100,t,2)}function wje(e,t){return e=Rye(e),Ei(e.getFullYear()%100,t,2)}function Rje(e,t){return Ei(e.getFullYear()%1e4,t,4)}function Aje(e,t){var n=e.getDay();return e=n>=4||n===0?z3(e):z3.ceil(e),Ei(e.getFullYear()%1e4,t,4)}function Tje(e){var t=e.getTimezoneOffset();return(t>0?"-":(t*=-1,"+"))+Ei(t/60|0,"0",2)+Ei(t%60,"0",2)}function uK(e,t){return Ei(e.getUTCDate(),t,2)}function kje(e,t){return Ei(e.getUTCHours(),t,2)}function Pje(e,t){return Ei(e.getUTCHours()%12||12,t,2)}function Mje(e,t){return Ei(1+r5.count(gg(e),e),t,3)}function Aye(e,t){return Ei(e.getUTCMilliseconds(),t,3)}function Fje(e,t){return Aye(e,t)+"000"}function Nje(e,t){return Ei(e.getUTCMonth()+1,t,2)}function Ije(e,t){return Ei(e.getUTCMinutes(),t,2)}function Lje(e,t){return Ei(e.getUTCSeconds(),t,2)}function Bje(e){var t=e.getUTCDay();return t===0?7:t}function Qje(e,t){return Ei(a5.count(gg(e)-1,e),t,2)}function Tye(e){var t=e.getUTCDay();return t>=4||t===0?q3(e):q3.ceil(e)}function zje(e,t){return e=Tye(e),Ei(q3.count(gg(e),e)+(gg(e).getUTCDay()===4),t,2)}function qje(e){return e.getUTCDay()}function Vje(e,t){return Ei(rC.count(gg(e)-1,e),t,2)}function Uje(e,t){return Ei(e.getUTCFullYear()%100,t,2)}function jje(e,t){return e=Tye(e),Ei(e.getUTCFullYear()%100,t,2)}function Zje(e,t){return Ei(e.getUTCFullYear()%1e4,t,4)}function Hje(e,t){var n=e.getUTCDay();return e=n>=4||n===0?q3(e):q3.ceil(e),Ei(e.getUTCFullYear()%1e4,t,4)}function Xje(){return"+0000"}function lK(){return"%"}function oK(e){return+e}function sK(e){return Math.floor(+e/1e3)}var wO,kye,Pye;Yje({dateTime:"%x, %X",date:"%-m/%-d/%Y",time:"%-I:%M:%S %p",periods:["AM","PM"],days:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],shortDays:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],months:["January","February","March","April","May","June","July","August","September","October","November","December"],shortMonths:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"]});function Yje(e){return wO=YUe(e),kye=wO.format,wO.parse,Pye=wO.utcFormat,wO.utcParse,wO}function Gje(e){return new Date(e)}function Kje(e){return e instanceof Date?+e:+new Date(+e)}function tB(e,t,n,r,i,a,u,l,o,s){var c=VL(),d=c.invert,h=c.domain,m=s(".%L"),y=s(":%S"),v=s("%I:%M"),S=s("%I %p"),_=s("%a %d"),E=s("%b %d"),$=s("%B"),k=s("%Y");function A(w){return(o(w)<w?m:l(w)<w?y:u(w)<w?v:a(w)<w?S:r(w)<w?i(w)<w?_:E:n(w)<w?$:k)(w)}return c.invert=function(w){return new Date(d(w))},c.domain=function(w){return arguments.length?h(Array.from(w,Kje)):h().map(Gje)},c.ticks=function(w){var R=h();return e(R[0],R[R.length-1],w??10)},c.tickFormat=function(w,R){return R==null?A:s(R)},c.nice=function(w){var R=h();return(!w||typeof w.range!="function")&&(w=t(R[0],R[R.length-1],w??10)),w?h(bye(R,w)):c},c.copy=function(){return QS(c,tB(e,t,n,r,i,a,u,l,o,s))},c}function Wje(){return oh.apply(tB(HUe,XUe,mg,JL,i5,zS,KL,YL,jb,kye).domain([new Date(2e3,0,1),new Date(2e3,0,2)]),arguments)}function Jje(){return oh.apply(tB(jUe,ZUe,gg,eB,a5,r5,WL,GL,jb,Pye).domain([Date.UTC(2e3,0,1),Date.UTC(2e3,0,2)]),arguments)}function u5(){var e=0,t=1,n,r,i,a,u=dc,l=!1,o;function s(d){return d==null||isNaN(d=+d)?o:u(i===0?.5:(d=(a(d)-n)*i,l?Math.max(0,Math.min(1,d)):d))}s.domain=function(d){return arguments.length?([e,t]=d,n=a(e=+e),r=a(t=+t),i=n===r?0:1/(r-n),s):[e,t]},s.clamp=function(d){return arguments.length?(l=!!d,s):l},s.interpolator=function(d){return arguments.length?(u=d,s):u};function c(d){return function(h){var m,y;return arguments.length?([m,y]=h,u=d(m,y),s):[u(0),u(1)]}}return s.range=c(r4),s.rangeRound=c(BI),s.unknown=function(d){return arguments.length?(o=d,s):o},function(d){return a=d,n=d(e),r=d(t),i=n===r?0:1/(r-n),s}}function _y(e,t){return t.domain(e.domain()).interpolator(e.interpolator()).clamp(e.clamp()).unknown(e.unknown())}function Mye(){var e=Sy(u5()(dc));return e.copy=function(){return _y(e,Mye())},Sg.apply(e,arguments)}function Fye(){var e=UL(u5()).domain([1,10]);return e.copy=function(){return _y(e,Fye()).base(e.base())},Sg.apply(e,arguments)}function Nye(){var e=jL(u5());return e.copy=function(){return _y(e,Nye()).constant(e.constant())},Sg.apply(e,arguments)}function nB(){var e=ZL(u5());return e.copy=function(){return _y(e,nB()).exponent(e.exponent())},Sg.apply(e,arguments)}function eZe(){return nB.apply(null,arguments).exponent(.5)}function Iye(){var e=[],t=dc;function n(r){if(r!=null&&!isNaN(r=+r))return t((BS(e,r,1)-1)/(e.length-1))}return n.domain=function(r){if(!arguments.length)return e.slice();e=[];for(let i of r)i!=null&&!isNaN(i=+i)&&e.push(i);return e.sort(Xv),n},n.interpolator=function(r){return arguments.length?(t=r,n):t},n.range=function(){return e.map((r,i)=>t(i/(e.length-1)))},n.quantiles=function(r){return Array.from({length:r+1},(i,a)=>OUe(e,a/r))},n.copy=function(){return Iye(t).domain(e)},Sg.apply(n,arguments)}function l5(){var e=0,t=.5,n=1,r=1,i,a,u,l,o,s=dc,c,d=!1,h;function m(v){return isNaN(v=+v)?h:(v=.5+((v=+c(v))-a)*(r*v<r*a?l:o),s(d?Math.max(0,Math.min(1,v)):v))}m.domain=function(v){return arguments.length?([e,t,n]=v,i=c(e=+e),a=c(t=+t),u=c(n=+n),l=i===a?0:.5/(a-i),o=a===u?0:.5/(u-a),r=a<i?-1:1,m):[e,t,n]},m.clamp=function(v){return arguments.length?(d=!!v,m):d},m.interpolator=function(v){return arguments.length?(s=v,m):s};function y(v){return function(S){var _,E,$;return arguments.length?([_,E,$]=S,s=t8e(v,[_,E,$]),m):[s(0),s(.5),s(1)]}}return m.range=y(r4),m.rangeRound=y(BI),m.unknown=function(v){return arguments.length?(h=v,m):h},function(v){return c=v,i=v(e),a=v(t),u=v(n),l=i===a?0:.5/(a-i),o=a===u?0:.5/(u-a),r=a<i?-1:1,m}}function Lye(){var e=Sy(l5()(dc));return e.copy=function(){return _y(e,Lye())},Sg.apply(e,arguments)}function Bye(){var e=UL(l5()).domain([.1,1,10]);return e.copy=function(){return _y(e,Bye()).base(e.base())},Sg.apply(e,arguments)}function Qye(){var e=jL(l5());return e.copy=function(){return _y(e,Qye()).constant(e.constant())},Sg.apply(e,arguments)}function rB(){var e=ZL(l5());return e.copy=function(){return _y(e,rB()).exponent(e.exponent())},Sg.apply(e,arguments)}function tZe(){return rB.apply(null,arguments).exponent(.5)}const vlt=Object.freeze(Object.defineProperty({__proto__:null,scaleBand:qL,scaleDiverging:Lye,scaleDivergingLog:Bye,scaleDivergingPow:rB,scaleDivergingSqrt:tZe,scaleDivergingSymlog:Qye,scaleIdentity:yye,scaleImplicit:dF,scaleLinear:vye,scaleLog:Oye,scaleOrdinal:zL,scalePoint:_Ue,scalePow:HL,scaleQuantile:_ye,scaleQuantize:Eye,scaleRadial:Sye,scaleSequential:Mye,scaleSequentialLog:Fye,scaleSequentialPow:nB,scaleSequentialQuantile:Iye,scaleSequentialSqrt:eZe,scaleSequentialSymlog:Nye,scaleSqrt:FUe,scaleSymlog:xye,scaleThreshold:Dye,scaleTime:Wje,scaleUtc:Jje,tickFormat:gye},Symbol.toStringTag,{value:"Module"}));var h4=1e9,nZe={precision:20,rounding:4,toExpNeg:-7,toExpPos:21,LN10:"2.302585092994045684017991454684364207601101488628772976033327900967572609677352480235997205089598298341967784042286"},aB,Za=!0,th="[DecimalError] ",Kb=th+"Invalid argument: ",iB=th+"Exponent out of range: ",p4=Math.floor,Ab=Math.pow,rZe=/^(\d+(\.\d*)?|\.\d+)(e[+-]?\d+)?$/i,Vf,uo=1e7,Ia=7,zye=9007199254740991,iC=p4(zye/Ia),fn={};fn.absoluteValue=fn.abs=function(){var e=new this.constructor(this);return e.s&&(e.s=1),e};fn.comparedTo=fn.cmp=function(e){var t,n,r,i,a=this;if(e=new a.constructor(e),a.s!==e.s)return a.s||-e.s;if(a.e!==e.e)return a.e>e.e^a.s<0?1:-1;for(r=a.d.length,i=e.d.length,t=0,n=r<i?r:i;t<n;++t)if(a.d[t]!==e.d[t])return a.d[t]>e.d[t]^a.s<0?1:-1;return r===i?0:r>i^a.s<0?1:-1};fn.decimalPlaces=fn.dp=function(){var e=this,t=e.d.length-1,n=(t-e.e)*Ia;if(t=e.d[t],t)for(;t%10==0;t/=10)n--;return n<0?0:n};fn.dividedBy=fn.div=function(e){return lg(this,new this.constructor(e))};fn.dividedToIntegerBy=fn.idiv=function(e){var t=this,n=t.constructor;return sa(lg(t,new n(e),0,1),n.precision)};fn.equals=fn.eq=function(e){return!this.cmp(e)};fn.exponent=function(){return Ol(this)};fn.greaterThan=fn.gt=function(e){return this.cmp(e)>0};fn.greaterThanOrEqualTo=fn.gte=function(e){return this.cmp(e)>=0};fn.isInteger=fn.isint=function(){return this.e>this.d.length-2};fn.isNegative=fn.isneg=function(){return this.s<0};fn.isPositive=fn.ispos=function(){return this.s>0};fn.isZero=function(){return this.s===0};fn.lessThan=fn.lt=function(e){return this.cmp(e)<0};fn.lessThanOrEqualTo=fn.lte=function(e){return this.cmp(e)<1};fn.logarithm=fn.log=function(e){var t,n=this,r=n.constructor,i=r.precision,a=i+5;if(e===void 0)e=new r(10);else if(e=new r(e),e.s<1||e.eq(Vf))throw Error(th+"NaN");if(n.s<1)throw Error(th+(n.s?"NaN":"-Infinity"));return n.eq(Vf)?new r(0):(Za=!1,t=lg(Kx(n,a),Kx(e,a),a),Za=!0,sa(t,i))};fn.minus=fn.sub=function(e){var t=this;return e=new t.constructor(e),t.s==e.s?Uye(t,e):qye(t,(e.s=-e.s,e))};fn.modulo=fn.mod=function(e){var t,n=this,r=n.constructor,i=r.precision;if(e=new r(e),!e.s)throw Error(th+"NaN");return n.s?(Za=!1,t=lg(n,e,0,1).times(e),Za=!0,n.minus(t)):sa(new r(n),i)};fn.naturalExponential=fn.exp=function(){return Vye(this)};fn.naturalLogarithm=fn.ln=function(){return Kx(this)};fn.negated=fn.neg=function(){var e=new this.constructor(this);return e.s=-e.s||0,e};fn.plus=fn.add=function(e){var t=this;return e=new t.constructor(e),t.s==e.s?qye(t,e):Uye(t,(e.s=-e.s,e))};fn.precision=fn.sd=function(e){var t,n,r,i=this;if(e!==void 0&&e!==!!e&&e!==1&&e!==0)throw Error(Kb+e);if(t=Ol(i)+1,r=i.d.length-1,n=r*Ia+1,r=i.d[r],r){for(;r%10==0;r/=10)n--;for(r=i.d[0];r>=10;r/=10)n++}return e&&t>n?t:n};fn.squareRoot=fn.sqrt=function(){var e,t,n,r,i,a,u,l=this,o=l.constructor;if(l.s<1){if(!l.s)return new o(0);throw Error(th+"NaN")}for(e=Ol(l),Za=!1,i=Math.sqrt(+l),i==0||i==1/0?(t=_0(l.d),(t.length+e)%2==0&&(t+="0"),i=Math.sqrt(t),e=p4((e+1)/2)-(e<0||e%2),i==1/0?t="5e"+e:(t=i.toExponential(),t=t.slice(0,t.indexOf("e")+1)+e),r=new o(t)):r=new o(i.toString()),n=o.precision,i=u=n+3;;)if(a=r,r=a.plus(lg(l,a,u+2)).times(.5),_0(a.d).slice(0,u)===(t=_0(r.d)).slice(0,u)){if(t=t.slice(u-3,u+1),i==u&&t=="4999"){if(sa(a,n+1,0),a.times(a).eq(l)){r=a;break}}else if(t!="9999")break;u+=4}return Za=!0,sa(r,n)};fn.times=fn.mul=function(e){var t,n,r,i,a,u,l,o,s,c=this,d=c.constructor,h=c.d,m=(e=new d(e)).d;if(!c.s||!e.s)return new d(0);for(e.s*=c.s,n=c.e+e.e,o=h.length,s=m.length,o<s&&(a=h,h=m,m=a,u=o,o=s,s=u),a=[],u=o+s,r=u;r--;)a.push(0);for(r=s;--r>=0;){for(t=0,i=o+r;i>r;)l=a[i]+m[r]*h[i-r-1]+t,a[i--]=l%uo|0,t=l/uo|0;a[i]=(a[i]+t)%uo|0}for(;!a[--u];)a.pop();return t?++n:a.shift(),e.d=a,e.e=n,Za?sa(e,d.precision):e};fn.toDecimalPlaces=fn.todp=function(e,t){var n=this,r=n.constructor;return n=new r(n),e===void 0?n:(M0(e,0,h4),t===void 0?t=r.rounding:M0(t,0,8),sa(n,e+Ol(n)+1,t))};fn.toExponential=function(e,t){var n,r=this,i=r.constructor;return e===void 0?n=b1(r,!0):(M0(e,0,h4),t===void 0?t=i.rounding:M0(t,0,8),r=sa(new i(r),e+1,t),n=b1(r,!0,e+1)),n};fn.toFixed=function(e,t){var n,r,i=this,a=i.constructor;return e===void 0?b1(i):(M0(e,0,h4),t===void 0?t=a.rounding:M0(t,0,8),r=sa(new a(i),e+Ol(i)+1,t),n=b1(r.abs(),!1,e+Ol(r)+1),i.isneg()&&!i.isZero()?"-"+n:n)};fn.toInteger=fn.toint=function(){var e=this,t=e.constructor;return sa(new t(e),Ol(e)+1,t.rounding)};fn.toNumber=function(){return+this};fn.toPower=fn.pow=function(e){var t,n,r,i,a,u,l=this,o=l.constructor,s=12,c=+(e=new o(e));if(!e.s)return new o(Vf);if(l=new o(l),!l.s){if(e.s<1)throw Error(th+"Infinity");return l}if(l.eq(Vf))return l;if(r=o.precision,e.eq(Vf))return sa(l,r);if(t=e.e,n=e.d.length-1,u=t>=n,a=l.s,u){if((n=c<0?-c:c)<=zye){for(i=new o(Vf),t=Math.ceil(r/Ia+4),Za=!1;n%2&&(i=i.times(l),fK(i.d,t)),n=p4(n/2),n!==0;)l=l.times(l),fK(l.d,t);return Za=!0,e.s<0?new o(Vf).div(i):sa(i,r)}}else if(a<0)throw Error(th+"NaN");return a=a<0&&e.d[Math.max(t,n)]&1?-1:1,l.s=1,Za=!1,i=e.times(Kx(l,r+s)),Za=!0,i=Vye(i),i.s=a,i};fn.toPrecision=function(e,t){var n,r,i=this,a=i.constructor;return e===void 0?(n=Ol(i),r=b1(i,n<=a.toExpNeg||n>=a.toExpPos)):(M0(e,1,h4),t===void 0?t=a.rounding:M0(t,0,8),i=sa(new a(i),e,t),n=Ol(i),r=b1(i,e<=n||n<=a.toExpNeg,e)),r};fn.toSignificantDigits=fn.tosd=function(e,t){var n=this,r=n.constructor;return e===void 0?(e=r.precision,t=r.rounding):(M0(e,1,h4),t===void 0?t=r.rounding:M0(t,0,8)),sa(new r(n),e,t)};fn.toString=fn.valueOf=fn.val=fn.toJSON=fn[Symbol.for("nodejs.util.inspect.custom")]=function(){var e=this,t=Ol(e),n=e.constructor;return b1(e,t<=n.toExpNeg||t>=n.toExpPos)};function qye(e,t){var n,r,i,a,u,l,o,s,c=e.constructor,d=c.precision;if(!e.s||!t.s)return t.s||(t=new c(e)),Za?sa(t,d):t;if(o=e.d,s=t.d,u=e.e,i=t.e,o=o.slice(),a=u-i,a){for(a<0?(r=o,a=-a,l=s.length):(r=s,i=u,l=o.length),u=Math.ceil(d/Ia),l=u>l?u+1:l+1,a>l&&(a=l,r.length=1),r.reverse();a--;)r.push(0);r.reverse()}for(l=o.length,a=s.length,l-a<0&&(a=l,r=s,s=o,o=r),n=0;a;)n=(o[--a]=o[a]+s[a]+n)/uo|0,o[a]%=uo;for(n&&(o.unshift(n),++i),l=o.length;o[--l]==0;)o.pop();return t.d=o,t.e=i,Za?sa(t,d):t}function M0(e,t,n){if(e!==~~e||e<t||e>n)throw Error(Kb+e)}function _0(e){var t,n,r,i=e.length-1,a="",u=e[0];if(i>0){for(a+=u,t=1;t<i;t++)r=e[t]+"",n=Ia-r.length,n&&(a+=$v(n)),a+=r;u=e[t],r=u+"",n=Ia-r.length,n&&(a+=$v(n))}else if(u===0)return"0";for(;u%10===0;)u/=10;return a+u}var lg=(function(){function e(r,i){var a,u=0,l=r.length;for(r=r.slice();l--;)a=r[l]*i+u,r[l]=a%uo|0,u=a/uo|0;return u&&r.unshift(u),r}function t(r,i,a,u){var l,o;if(a!=u)o=a>u?1:-1;else for(l=o=0;l<a;l++)if(r[l]!=i[l]){o=r[l]>i[l]?1:-1;break}return o}function n(r,i,a){for(var u=0;a--;)r[a]-=u,u=r[a]<i[a]?1:0,r[a]=u*uo+r[a]-i[a];for(;!r[0]&&r.length>1;)r.shift()}return function(r,i,a,u){var l,o,s,c,d,h,m,y,v,S,_,E,$,k,A,w,R,F,I=r.constructor,U=r.s==i.s?1:-1,q=r.d,V=i.d;if(!r.s)return new I(r);if(!i.s)throw Error(th+"Division by zero");for(o=r.e-i.e,R=V.length,A=q.length,m=new I(U),y=m.d=[],s=0;V[s]==(q[s]||0);)++s;if(V[s]>(q[s]||0)&&--o,a==null?E=a=I.precision:u?E=a+(Ol(r)-Ol(i))+1:E=a,E<0)return new I(0);if(E=E/Ia+2|0,s=0,R==1)for(c=0,V=V[0],E++;(s<A||c)&&E--;s++)$=c*uo+(q[s]||0),y[s]=$/V|0,c=$%V|0;else{for(c=uo/(V[0]+1)|0,c>1&&(V=e(V,c),q=e(q,c),R=V.length,A=q.length),k=R,v=q.slice(0,R),S=v.length;S<R;)v[S++]=0;F=V.slice(),F.unshift(0),w=V[0],V[1]>=uo/2&&++w;do c=0,l=t(V,v,R,S),l<0?(_=v[0],R!=S&&(_=_*uo+(v[1]||0)),c=_/w|0,c>1?(c>=uo&&(c=uo-1),d=e(V,c),h=d.length,S=v.length,l=t(d,v,h,S),l==1&&(c--,n(d,R<h?F:V,h))):(c==0&&(l=c=1),d=V.slice()),h=d.length,h<S&&d.unshift(0),n(v,d,S),l==-1&&(S=v.length,l=t(V,v,R,S),l<1&&(c++,n(v,R<S?F:V,S))),S=v.length):l===0&&(c++,v=[0]),y[s++]=c,l&&v[0]?v[S++]=q[k]||0:(v=[q[k]],S=1);while((k++<A||v[0]!==void 0)&&E--)}return y[0]||y.shift(),m.e=o,sa(m,u?a+Ol(m)+1:a)}})();function Vye(e,t){var n,r,i,a,u,l,o=0,s=0,c=e.constructor,d=c.precision;if(Ol(e)>16)throw Error(iB+Ol(e));if(!e.s)return new c(Vf);for(Za=!1,l=d,u=new c(.03125);e.abs().gte(.1);)e=e.times(u),s+=5;for(r=Math.log(Ab(2,s))/Math.LN10*2+5|0,l+=r,n=i=a=new c(Vf),c.precision=l;;){if(i=sa(i.times(e),l),n=n.times(++o),u=a.plus(lg(i,n,l)),_0(u.d).slice(0,l)===_0(a.d).slice(0,l)){for(;s--;)a=sa(a.times(a),l);return c.precision=d,t==null?(Za=!0,sa(a,d)):a}a=u}}function Ol(e){for(var t=e.e*Ia,n=e.d[0];n>=10;n/=10)t++;return t}function K8(e,t,n){if(t>e.LN10.sd())throw Za=!0,n&&(e.precision=n),Error(th+"LN10 precision limit exceeded");return sa(new e(e.LN10),t)}function $v(e){for(var t="";e--;)t+="0";return t}function Kx(e,t){var n,r,i,a,u,l,o,s,c,d=1,h=10,m=e,y=m.d,v=m.constructor,S=v.precision;if(m.s<1)throw Error(th+(m.s?"NaN":"-Infinity"));if(m.eq(Vf))return new v(0);if(t==null?(Za=!1,s=S):s=t,m.eq(10))return t==null&&(Za=!0),K8(v,s);if(s+=h,v.precision=s,n=_0(y),r=n.charAt(0),a=Ol(m),Math.abs(a)<15e14){for(;r<7&&r!=1||r==1&&n.charAt(1)>3;)m=m.times(e),n=_0(m.d),r=n.charAt(0),d++;a=Ol(m),r>1?(m=new v("0."+n),a++):m=new v(r+"."+n.slice(1))}else return o=K8(v,s+2,S).times(a+""),m=Kx(new v(r+"."+n.slice(1)),s-h).plus(o),v.precision=S,t==null?(Za=!0,sa(m,S)):m;for(l=u=m=lg(m.minus(Vf),m.plus(Vf),s),c=sa(m.times(m),s),i=3;;){if(u=sa(u.times(c),s),o=l.plus(lg(u,new v(i),s)),_0(o.d).slice(0,s)===_0(l.d).slice(0,s))return l=l.times(2),a!==0&&(l=l.plus(K8(v,s+2,S).times(a+""))),l=lg(l,new v(d),s),v.precision=S,t==null?(Za=!0,sa(l,S)):l;l=o,i+=2}}function cK(e,t){var n,r,i;for((n=t.indexOf("."))>-1&&(t=t.replace(".","")),(r=t.search(/e/i))>0?(n<0&&(n=r),n+=+t.slice(r+1),t=t.substring(0,r)):n<0&&(n=t.length),r=0;t.charCodeAt(r)===48;)++r;for(i=t.length;t.charCodeAt(i-1)===48;)--i;if(t=t.slice(r,i),t){if(i-=r,n=n-r-1,e.e=p4(n/Ia),e.d=[],r=(n+1)%Ia,n<0&&(r+=Ia),r<i){for(r&&e.d.push(+t.slice(0,r)),i-=Ia;r<i;)e.d.push(+t.slice(r,r+=Ia));t=t.slice(r),r=Ia-t.length}else r-=i;for(;r--;)t+="0";if(e.d.push(+t),Za&&(e.e>iC||e.e<-iC))throw Error(iB+n)}else e.s=0,e.e=0,e.d=[0];return e}function sa(e,t,n){var r,i,a,u,l,o,s,c,d=e.d;for(u=1,a=d[0];a>=10;a/=10)u++;if(r=t-u,r<0)r+=Ia,i=t,s=d[c=0];else{if(c=Math.ceil((r+1)/Ia),a=d.length,c>=a)return e;for(s=a=d[c],u=1;a>=10;a/=10)u++;r%=Ia,i=r-Ia+u}if(n!==void 0&&(a=Ab(10,u-i-1),l=s/a%10|0,o=t<0||d[c+1]!==void 0||s%a,o=n<4?(l||o)&&(n==0||n==(e.s<0?3:2)):l>5||l==5&&(n==4||o||n==6&&(r>0?i>0?s/Ab(10,u-i):0:d[c-1])%10&1||n==(e.s<0?8:7))),t<1||!d[0])return o?(a=Ol(e),d.length=1,t=t-a-1,d[0]=Ab(10,(Ia-t%Ia)%Ia),e.e=p4(-t/Ia)||0):(d.length=1,d[0]=e.e=e.s=0),e;if(r==0?(d.length=c,a=1,c--):(d.length=c+1,a=Ab(10,Ia-r),d[c]=i>0?(s/Ab(10,u-i)%Ab(10,i)|0)*a:0),o)for(;;)if(c==0){(d[0]+=a)==uo&&(d[0]=1,++e.e);break}else{if(d[c]+=a,d[c]!=uo)break;d[c--]=0,a=1}for(r=d.length;d[--r]===0;)d.pop();if(Za&&(e.e>iC||e.e<-iC))throw Error(iB+Ol(e));return e}function Uye(e,t){var n,r,i,a,u,l,o,s,c,d,h=e.constructor,m=h.precision;if(!e.s||!t.s)return t.s?t.s=-t.s:t=new h(e),Za?sa(t,m):t;if(o=e.d,d=t.d,r=t.e,s=e.e,o=o.slice(),u=s-r,u){for(c=u<0,c?(n=o,u=-u,l=d.length):(n=d,r=s,l=o.length),i=Math.max(Math.ceil(m/Ia),l)+2,u>i&&(u=i,n.length=1),n.reverse(),i=u;i--;)n.push(0);n.reverse()}else{for(i=o.length,l=d.length,c=i<l,c&&(l=i),i=0;i<l;i++)if(o[i]!=d[i]){c=o[i]<d[i];break}u=0}for(c&&(n=o,o=d,d=n,t.s=-t.s),l=o.length,i=d.length-l;i>0;--i)o[l++]=0;for(i=d.length;i>u;){if(o[--i]<d[i]){for(a=i;a&&o[--a]===0;)o[a]=uo-1;--o[a],o[i]+=uo}o[i]-=d[i]}for(;o[--l]===0;)o.pop();for(;o[0]===0;o.shift())--r;return o[0]?(t.d=o,t.e=r,Za?sa(t,m):t):new h(0)}function b1(e,t,n){var r,i=Ol(e),a=_0(e.d),u=a.length;return t?(n&&(r=n-u)>0?a=a.charAt(0)+"."+a.slice(1)+$v(r):u>1&&(a=a.charAt(0)+"."+a.slice(1)),a=a+(i<0?"e":"e+")+i):i<0?(a="0."+$v(-i-1)+a,n&&(r=n-u)>0&&(a+=$v(r))):i>=u?(a+=$v(i+1-u),n&&(r=n-i-1)>0&&(a=a+"."+$v(r))):((r=i+1)<u&&(a=a.slice(0,r)+"."+a.slice(r)),n&&(r=n-u)>0&&(i+1===u&&(a+="."),a+=$v(r))),e.s<0?"-"+a:a}function fK(e,t){if(e.length>t)return e.length=t,!0}function jye(e){var t,n,r;function i(a){var u=this;if(!(u instanceof i))return new i(a);if(u.constructor=i,a instanceof i){u.s=a.s,u.e=a.e,u.d=(a=a.d)?a.slice():a;return}if(typeof a=="number"){if(a*0!==0)throw Error(Kb+a);if(a>0)u.s=1;else if(a<0)a=-a,u.s=-1;else{u.s=0,u.e=0,u.d=[0];return}if(a===~~a&&a<1e7){u.e=0,u.d=[a];return}return cK(u,a.toString())}else if(typeof a!="string")throw Error(Kb+a);if(a.charCodeAt(0)===45?(a=a.slice(1),u.s=-1):u.s=1,rZe.test(a))cK(u,a);else throw Error(Kb+a)}if(i.prototype=fn,i.ROUND_UP=0,i.ROUND_DOWN=1,i.ROUND_CEIL=2,i.ROUND_FLOOR=3,i.ROUND_HALF_UP=4,i.ROUND_HALF_DOWN=5,i.ROUND_HALF_EVEN=6,i.ROUND_HALF_CEIL=7,i.ROUND_HALF_FLOOR=8,i.clone=jye,i.config=i.set=iZe,e===void 0&&(e={}),e)for(r=["precision","rounding","toExpNeg","toExpPos","LN10"],t=0;t<r.length;)e.hasOwnProperty(n=r[t++])||(e[n]=this[n]);return i.config(e),i}function iZe(e){if(!e||typeof e!="object")throw Error(th+"Object expected");var t,n,r,i=["precision",1,h4,"rounding",0,8,"toExpNeg",-1/0,0,"toExpPos",0,1/0];for(t=0;t<i.length;t+=3)if((r=e[n=i[t]])!==void 0)if(p4(r)===r&&r>=i[t+1]&&r<=i[t+2])this[n]=r;else throw Error(Kb+n+": "+r);if((r=e[n="LN10"])!==void 0)if(r==Math.LN10)this[n]=new this(r);else throw Error(Kb+n+": "+r);return this}var aB=jye(nZe);Vf=new aB(1);const ylt=aB;var W8={exports:{}},dK;function aZe(){return dK||(dK=1,(function(e){var t=Object.prototype.hasOwnProperty,n="~";function r(){}Object.create&&(r.prototype=Object.create(null),new r().__proto__||(n=!1));function i(o,s,c){this.fn=o,this.context=s,this.once=c||!1}function a(o,s,c,d,h){if(typeof c!="function")throw new TypeError("The listener must be a function");var m=new i(c,d||o,h),y=n?n+s:s;return o._events[y]?o._events[y].fn?o._events[y]=[o._events[y],m]:o._events[y].push(m):(o._events[y]=m,o._eventsCount++),o}function u(o,s){--o._eventsCount===0?o._events=new r:delete o._events[s]}function l(){this._events=new r,this._eventsCount=0}l.prototype.eventNames=function(){var s=[],c,d;if(this._eventsCount===0)return s;for(d in c=this._events)t.call(c,d)&&s.push(n?d.slice(1):d);return Object.getOwnPropertySymbols?s.concat(Object.getOwnPropertySymbols(c)):s},l.prototype.listeners=function(s){var c=n?n+s:s,d=this._events[c];if(!d)return[];if(d.fn)return[d.fn];for(var h=0,m=d.length,y=new Array(m);h<m;h++)y[h]=d[h].fn;return y},l.prototype.listenerCount=function(s){var c=n?n+s:s,d=this._events[c];return d?d.fn?1:d.length:0},l.prototype.emit=function(s,c,d,h,m,y){var v=n?n+s:s;if(!this._events[v])return!1;var S=this._events[v],_=arguments.length,E,$;if(S.fn){switch(S.once&&this.removeListener(s,S.fn,void 0,!0),_){case 1:return S.fn.call(S.context),!0;case 2:return S.fn.call(S.context,c),!0;case 3:return S.fn.call(S.context,c,d),!0;case 4:return S.fn.call(S.context,c,d,h),!0;case 5:return S.fn.call(S.context,c,d,h,m),!0;case 6:return S.fn.call(S.context,c,d,h,m,y),!0}for($=1,E=new Array(_-1);$<_;$++)E[$-1]=arguments[$];S.fn.apply(S.context,E)}else{var k=S.length,A;for($=0;$<k;$++)switch(S[$].once&&this.removeListener(s,S[$].fn,void 0,!0),_){case 1:S[$].fn.call(S[$].context);break;case 2:S[$].fn.call(S[$].context,c);break;case 3:S[$].fn.call(S[$].context,c,d);break;case 4:S[$].fn.call(S[$].context,c,d,h);break;default:if(!E)for(A=1,E=new Array(_-1);A<_;A++)E[A-1]=arguments[A];S[$].fn.apply(S[$].context,E)}}return!0},l.prototype.on=function(s,c,d){return a(this,s,c,d,!1)},l.prototype.once=function(s,c,d){return a(this,s,c,d,!0)},l.prototype.removeListener=function(s,c,d,h){var m=n?n+s:s;if(!this._events[m])return this;if(!c)return u(this,m),this;var y=this._events[m];if(y.fn)y.fn===c&&(!h||y.once)&&(!d||y.context===d)&&u(this,m);else{for(var v=0,S=[],_=y.length;v<_;v++)(y[v].fn!==c||h&&!y[v].once||d&&y[v].context!==d)&&S.push(y[v]);S.length?this._events[m]=S.length===1?S[0]:S:u(this,m)}return this},l.prototype.removeAllListeners=function(s){var c;return s?(c=n?n+s:s,this._events[c]&&u(this,c)):(this._events=new r,this._eventsCount=0),this},l.prototype.off=l.prototype.removeListener,l.prototype.addListener=l.prototype.on,l.prefixed=n,l.EventEmitter=l,e.exports=l})(W8)),W8.exports}var uZe=aZe();const blt=da(uZe);var J8={},e7={},t7={},hK;function lZe(){return hK||(hK=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n,r,{signal:i,edges:a}={}){let u,l=null;const o=a!=null&&a.includes("leading"),s=a==null||a.includes("trailing"),c=()=>{l!==null&&(n.apply(u,l),u=void 0,l=null)},d=()=>{s&&c(),v()};let h=null;const m=()=>{h!=null&&clearTimeout(h),h=setTimeout(()=>{h=null,d()},r)},y=()=>{h!==null&&(clearTimeout(h),h=null)},v=()=>{y(),u=void 0,l=null},S=()=>{c()},_=function(...E){if(i!=null&&i.aborted)return;u=this,l=E;const $=h==null;m(),o&&$&&c()};return _.schedule=m,_.cancel=v,_.flush=S,i==null||i.addEventListener("abort",v,{once:!0}),_}e.debounce=t})(t7)),t7}var pK;function oZe(){return pK||(pK=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=lZe();function n(r,i=0,a={}){typeof a!="object"&&(a={});const{leading:u=!1,trailing:l=!0,maxWait:o}=a,s=Array(2);u&&(s[0]="leading"),l&&(s[1]="trailing");let c,d=null;const h=t.debounce(function(...v){c=r.apply(this,v),d=null},i,{edges:s}),m=function(...v){return o!=null&&(d===null&&(d=Date.now()),Date.now()-d>=o)?(c=r.apply(this,v),d=Date.now(),h.cancel(),h.schedule(),c):(h.apply(this,v),c)},y=()=>(h.flush(),c);return m.cancel=h.cancel,m.flush=y,m}e.debounce=n})(e7)),e7}var mK;function sZe(){return mK||(mK=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=oZe();function n(r,i=0,a={}){const{leading:u=!0,trailing:l=!0}=a;return t.debounce(r,i,{leading:u,maxWait:i,trailing:l})}e.throttle=n})(J8)),J8}var n7,gK;function cZe(){return gK||(gK=1,n7=sZe().throttle),n7}var fZe=cZe();const Olt=da(fZe);var r7={},i7={},vK;function dZe(){return vK||(vK=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return n[n.length-1]}e.last=t})(i7)),i7}var a7={},yK;function hZe(){return yK||(yK=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){return Array.isArray(n)?n:Array.from(n)}e.toArray=t})(a7)),a7}var bK;function pZe(){return bK||(bK=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=dZe(),n=hZe(),r=AL();function i(a){if(r.isArrayLike(a))return t.last(n.toArray(a))}e.last=i})(r7)),r7}var u7,OK;function mZe(){return OK||(OK=1,u7=pZe().last),u7}var gZe=mZe();const xlt=da(gZe);var l7={},xK;function vZe(){return xK||(xK=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(n){var i;if(typeof n!="object"||n==null)return!1;if(Object.getPrototypeOf(n)===null)return!0;if(Object.prototype.toString.call(n)!=="[object Object]"){const a=n[Symbol.toStringTag];return a==null||!((i=Object.getOwnPropertyDescriptor(n,Symbol.toStringTag))!=null&&i.writable)?!1:n.toString()===`[object ${a}]`}let r=n;for(;Object.getPrototypeOf(r)!==null;)r=Object.getPrototypeOf(r);return Object.getPrototypeOf(n)===r}e.isPlainObject=t})(l7)),l7}var o7,SK;function yZe(){return SK||(SK=1,o7=vZe().isPlainObject),o7}var bZe=yZe();const Slt=da(bZe);var s7={exports:{}},c7={};/**
 * @license React
 * use-sync-external-store-with-selector.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var _K;function OZe(){if(_K)return c7;_K=1;var e=ir();function t(o,s){return o===s&&(o!==0||1/o===1/s)||o!==o&&s!==s}var n=typeof Object.is=="function"?Object.is:t,r=e.useSyncExternalStore,i=e.useRef,a=e.useEffect,u=e.useMemo,l=e.useDebugValue;return c7.useSyncExternalStoreWithSelector=function(o,s,c,d,h){var m=i(null);if(m.current===null){var y={hasValue:!1,value:null};m.current=y}else y=m.current;m=u(function(){function S(A){if(!_){if(_=!0,E=A,A=d(A),h!==void 0&&y.hasValue){var w=y.value;if(h(w,A))return $=w}return $=A}if(w=$,n(E,A))return w;var R=d(A);return h!==void 0&&h(w,R)?(E=A,w):(E=A,$=R)}var _=!1,E,$,k=c===void 0?null:c;return[function(){return S(s())},k===null?void 0:function(){return S(k())}]},[s,c,d,h]);var v=r(o,m[0],m[1]);return a(function(){y.hasValue=!0,y.value=v},[v]),l(v),v},c7}var EK;function xZe(){return EK||(EK=1,s7.exports=OZe()),s7.exports}xZe();function SZe(e){e()}function _Ze(){let e=null,t=null;return{clear(){e=null,t=null},notify(){SZe(()=>{let n=e;for(;n;)n.callback(),n=n.next})},get(){const n=[];let r=e;for(;r;)n.push(r),r=r.next;return n},subscribe(n){let r=!0;const i=t={callback:n,next:null,prev:t};return i.prev?i.prev.next=i:e=i,function(){!r||e===null||(r=!1,i.next?i.next.prev=i.prev:t=i.prev,i.prev?i.prev.next=i.next:e=i.next)}}}}var DK={notify(){},get:()=>[]};function EZe(e,t){let n,r=DK,i=0,a=!1;function u(v){c();const S=r.subscribe(v);let _=!1;return()=>{_||(_=!0,S(),d())}}function l(){r.notify()}function o(){y.onStateChange&&y.onStateChange()}function s(){return a}function c(){i++,n||(n=e.subscribe(o),r=_Ze())}function d(){i--,n&&i===0&&(n(),n=void 0,r.clear(),r=DK)}function h(){a||(a=!0,c())}function m(){a&&(a=!1,d())}const y={addNestedSub:u,notifyNestedSubs:l,handleChangeWrapper:o,isSubscribed:s,trySubscribe:h,tryUnsubscribe:m,getListeners:()=>r};return y}var DZe=()=>typeof window<"u"&&typeof window.document<"u"&&typeof window.document.createElement<"u",$Ze=DZe(),CZe=()=>typeof navigator<"u"&&navigator.product==="ReactNative",wZe=CZe(),RZe=()=>$Ze||wZe?P.useLayoutEffect:P.useEffect,AZe=RZe(),f7=Symbol.for("react-redux-context"),d7=typeof globalThis<"u"?globalThis:{};function TZe(){if(!P.createContext)return{};const e=d7[f7]??(d7[f7]=new Map);let t=e.get(P.createContext);return t||(t=P.createContext(null),e.set(P.createContext,t)),t}var kZe=TZe();function PZe(e){const{children:t,context:n,serverState:r,store:i}=e,a=P.useMemo(()=>{const o=EZe(i);return{store:i,subscription:o,getServerState:r?()=>r:void 0}},[i,r]),u=P.useMemo(()=>i.getState(),[i]);AZe(()=>{const{subscription:o}=a;return o.onStateChange=o.notifyNestedSubs,o.trySubscribe(),u!==i.getState()&&o.notifyNestedSubs(),()=>{o.tryUnsubscribe(),o.onStateChange=void 0}},[a,u]);const l=n||kZe;return P.createElement(l.Provider,{value:a},t)}var _lt=PZe,q2={exports:{}};/**
 * @license
 * Lodash <https://lodash.com/>
 * Copyright OpenJS Foundation and other contributors <https://openjsf.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */var MZe=q2.exports,$K;function FZe(){return $K||($K=1,(function(e,t){(function(){var n,r="4.17.23",i=200,a="Unsupported core-js use. Try https://npms.io/search?q=ponyfill.",u="Expected a function",l="Invalid `variable` option passed into `_.template`",o="__lodash_hash_undefined__",s=500,c="__lodash_placeholder__",d=1,h=2,m=4,y=1,v=2,S=1,_=2,E=4,$=8,k=16,A=32,w=64,R=128,F=256,I=512,U=30,q="...",V=800,N=16,j=1,W=2,K=3,J=1/0,te=9007199254740991,H=17976931348623157e292,M=NaN,G=4294967295,X=G-1,B=G>>>1,ne=[["ary",R],["bind",S],["bindKey",_],["curry",$],["curryRight",k],["flip",I],["partial",A],["partialRight",w],["rearg",F]],oe="[object Arguments]",ie="[object Array]",de="[object AsyncFunction]",ge="[object Boolean]",ye="[object Date]",me="[object DOMException]",Ee="[object Error]",se="[object Function]",pe="[object GeneratorFunction]",ve="[object Map]",we="[object Number]",Me="[object Null]",Re="[object Object]",et="[object Promise]",tt="[object Proxy]",Oe="[object RegExp]",$e="[object Set]",Ae="[object String]",He="[object Symbol]",je="[object Undefined]",Ye="[object WeakMap]",Ne="[object WeakSet]",ze="[object ArrayBuffer]",at="[object DataView]",mt="[object Float32Array]",Mt="[object Float64Array]",bt="[object Int8Array]",zt="[object Int16Array]",qt="[object Int32Array]",Xt="[object Uint8Array]",nn="[object Uint8ClampedArray]",Fe="[object Uint16Array]",ht="[object Uint32Array]",kt=/\b__p \+= '';/g,Qt=/\b(__p \+=) '' \+/g,en=/(__e\(.*?\)|\b__t\)) \+\n'';/g,In=/&(?:amp|lt|gt|quot|#39);/g,rn=/[&<>"']/g,Xn=RegExp(In.source),br=RegExp(rn.source),Or=/<%-([\s\S]+?)%>/g,Cr=/<%([\s\S]+?)%>/g,wr=/<%=([\s\S]+?)%>/g,Gr=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,ji=/^\w*$/,ai=/[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g,ar=/[\\^$.*+?()[\]{}|]/g,Te=RegExp(ar.source),Ve=/^\s+/,nt=/\s/,gt=/\{(?:\n\/\* \[wrapped with .+\] \*\/)?\n?/,Ct=/\{\n\/\* \[wrapped with (.+)\] \*/,Vt=/,? & /,Yt=/[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g,Dt=/[()=,{}\[\]\/\s]/,Lt=/\\(\\)?/g,an=/\$\{([^\\}]*(?:\\.[^\\}]*)*)\}/g,_n=/\w*$/,sn=/^[-+]0x[0-9a-f]+$/i,Ln=/^0b[01]+$/i,er=/^\[object .+?Constructor\]$/,Di=/^0o[0-7]+$/i,$i=/^(?:0|[1-9]\d*)$/,hr=/[\xc0-\xd6\xd8-\xf6\xf8-\xff\u0100-\u017f]/g,wt=/($^)/,Kr=/['\n\r\u2028\u2029\\]/g,Zi="\\ud800-\\udfff",yi="\\u0300-\\u036f",Ca="\\ufe20-\\ufe2f",Xa="\\u20d0-\\u20ff",Au=yi+Ca+Xa,Tu="\\u2700-\\u27bf",Nl="a-z\\xdf-\\xf6\\xf8-\\xff",gc="\\xac\\xb1\\xd7\\xf7",xl="\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf",ur="\\u2000-\\u206f",Rr=" \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000",Sl="A-Z\\xc0-\\xd6\\xd8-\\xde",tr="\\ufe0e\\ufe0f",Bn=gc+xl+ur+Rr,ku="['’]",gf="["+Zi+"]",Ds="["+Bn+"]",qo="["+Au+"]",Il="\\d+",sh="["+Tu+"]",Ya="["+Nl+"]",rd="[^"+Zi+Bn+Il+Tu+Nl+Sl+"]",$s="\\ud83c[\\udffb-\\udfff]",id="(?:"+qo+"|"+$s+")",vc="[^"+Zi+"]",yc="(?:\\ud83c[\\udde6-\\uddff]){2}",Vo="[\\ud800-\\udbff][\\udc00-\\udfff]",Pu="["+Sl+"]",Cs="\\u200d",ws="(?:"+Ya+"|"+rd+")",j0="(?:"+Pu+"|"+rd+")",vf="(?:"+ku+"(?:d|ll|m|re|s|t|ve))?",Z0="(?:"+ku+"(?:D|LL|M|RE|S|T|VE))?",Op=id+"?",yf="["+tr+"]?",ho="(?:"+Cs+"(?:"+[vc,yc,Vo].join("|")+")"+yf+Op+")*",ch="\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])",fh="\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])",Rs=yf+Op+ho,bc="(?:"+[sh,yc,Vo].join("|")+")"+Rs,H0="(?:"+[vc+qo+"?",qo,yc,Vo,gf].join("|")+")",xp=RegExp(ku,"g"),Sp=RegExp(qo,"g"),As=RegExp($s+"(?="+$s+")|"+H0+Rs,"g"),Uo=RegExp([Pu+"?"+Ya+"+"+vf+"(?="+[Ds,Pu,"$"].join("|")+")",j0+"+"+Z0+"(?="+[Ds,Pu+ws,"$"].join("|")+")",Pu+"?"+ws+"+"+vf,Pu+"+"+Z0,fh,ch,Il,bc].join("|"),"g"),_p=RegExp("["+Cs+Zi+Au+tr+"]"),dh=/[a-z][A-Z]|[A-Z]{2}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/,Oc=["Array","Buffer","DataView","Date","Error","Float32Array","Float64Array","Function","Int8Array","Int16Array","Int32Array","Map","Math","Object","Promise","RegExp","Set","String","Symbol","TypeError","Uint8Array","Uint8ClampedArray","Uint16Array","Uint32Array","WeakMap","_","clearTimeout","isFinite","parseInt","setTimeout"],hh=-1,Mr={};Mr[mt]=Mr[Mt]=Mr[bt]=Mr[zt]=Mr[qt]=Mr[Xt]=Mr[nn]=Mr[Fe]=Mr[ht]=!0,Mr[oe]=Mr[ie]=Mr[ze]=Mr[ge]=Mr[at]=Mr[ye]=Mr[Ee]=Mr[se]=Mr[ve]=Mr[we]=Mr[Re]=Mr[Oe]=Mr[$e]=Mr[Ae]=Mr[Ye]=!1;var Qn={};Qn[oe]=Qn[ie]=Qn[ze]=Qn[at]=Qn[ge]=Qn[ye]=Qn[mt]=Qn[Mt]=Qn[bt]=Qn[zt]=Qn[qt]=Qn[ve]=Qn[we]=Qn[Re]=Qn[Oe]=Qn[$e]=Qn[Ae]=Qn[He]=Qn[Xt]=Qn[nn]=Qn[Fe]=Qn[ht]=!0,Qn[Ee]=Qn[se]=Qn[Ye]=!1;var xc={À:"A",Á:"A",Â:"A",Ã:"A",Ä:"A",Å:"A",à:"a",á:"a",â:"a",ã:"a",ä:"a",å:"a",Ç:"C",ç:"c",Ð:"D",ð:"d",È:"E",É:"E",Ê:"E",Ë:"E",è:"e",é:"e",ê:"e",ë:"e",Ì:"I",Í:"I",Î:"I",Ï:"I",ì:"i",í:"i",î:"i",ï:"i",Ñ:"N",ñ:"n",Ò:"O",Ó:"O",Ô:"O",Õ:"O",Ö:"O",Ø:"O",ò:"o",ó:"o",ô:"o",õ:"o",ö:"o",ø:"o",Ù:"U",Ú:"U",Û:"U",Ü:"U",ù:"u",ú:"u",û:"u",ü:"u",Ý:"Y",ý:"y",ÿ:"y",Æ:"Ae",æ:"ae",Þ:"Th",þ:"th",ß:"ss",Ā:"A",Ă:"A",Ą:"A",ā:"a",ă:"a",ą:"a",Ć:"C",Ĉ:"C",Ċ:"C",Č:"C",ć:"c",ĉ:"c",ċ:"c",č:"c",Ď:"D",Đ:"D",ď:"d",đ:"d",Ē:"E",Ĕ:"E",Ė:"E",Ę:"E",Ě:"E",ē:"e",ĕ:"e",ė:"e",ę:"e",ě:"e",Ĝ:"G",Ğ:"G",Ġ:"G",Ģ:"G",ĝ:"g",ğ:"g",ġ:"g",ģ:"g",Ĥ:"H",Ħ:"H",ĥ:"h",ħ:"h",Ĩ:"I",Ī:"I",Ĭ:"I",Į:"I",İ:"I",ĩ:"i",ī:"i",ĭ:"i",į:"i",ı:"i",Ĵ:"J",ĵ:"j",Ķ:"K",ķ:"k",ĸ:"k",Ĺ:"L",Ļ:"L",Ľ:"L",Ŀ:"L",Ł:"L",ĺ:"l",ļ:"l",ľ:"l",ŀ:"l",ł:"l",Ń:"N",Ņ:"N",Ň:"N",Ŋ:"N",ń:"n",ņ:"n",ň:"n",ŋ:"n",Ō:"O",Ŏ:"O",Ő:"O",ō:"o",ŏ:"o",ő:"o",Ŕ:"R",Ŗ:"R",Ř:"R",ŕ:"r",ŗ:"r",ř:"r",Ś:"S",Ŝ:"S",Ş:"S",Š:"S",ś:"s",ŝ:"s",ş:"s",š:"s",Ţ:"T",Ť:"T",Ŧ:"T",ţ:"t",ť:"t",ŧ:"t",Ũ:"U",Ū:"U",Ŭ:"U",Ů:"U",Ű:"U",Ų:"U",ũ:"u",ū:"u",ŭ:"u",ů:"u",ű:"u",ų:"u",Ŵ:"W",ŵ:"w",Ŷ:"Y",ŷ:"y",Ÿ:"Y",Ź:"Z",Ż:"Z",Ž:"Z",ź:"z",ż:"z",ž:"z",Ĳ:"IJ",ĳ:"ij",Œ:"Oe",œ:"oe",ŉ:"'n",ſ:"s"},jo={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;"},Sc={"&amp;":"&","&lt;":"<","&gt;":">","&quot;":'"',"&#39;":"'"},ra={"\\":"\\","'":"'","\n":"n","\r":"r","\u2028":"u2028","\u2029":"u2029"},Ts=parseFloat,ad=parseInt,Ll=typeof Xd=="object"&&Xd&&Xd.Object===Object&&Xd,ud=typeof self=="object"&&self&&self.Object===Object&&self,ia=Ll||ud||Function("return this")(),ks=t&&!t.nodeType&&t,Bl=ks&&!0&&e&&!e.nodeType&&e,ld=Bl&&Bl.exports===ks,Ep=ld&&Ll.process,ui=(function(){try{var Xe=Bl&&Bl.require&&Bl.require("util").types;return Xe||Ep&&Ep.binding&&Ep.binding("util")}catch{}})(),X0=ui&&ui.isArrayBuffer,Y0=ui&&ui.isDate,od=ui&&ui.isMap,G0=ui&&ui.isRegExp,_c=ui&&ui.isSet,sd=ui&&ui.isTypedArray;function ha(Xe,dt,rt){switch(rt.length){case 0:return Xe.call(dt);case 1:return Xe.call(dt,rt[0]);case 2:return Xe.call(dt,rt[0],rt[1]);case 3:return Xe.call(dt,rt[0],rt[1],rt[2])}return Xe.apply(dt,rt)}function cd(Xe,dt,rt,Bt){for(var En=-1,lr=Xe==null?0:Xe.length;++En<lr;){var Li=Xe[En];dt(Bt,Li,rt(Li),Xe)}return Bt}function Ba(Xe,dt){for(var rt=-1,Bt=Xe==null?0:Xe.length;++rt<Bt&&dt(Xe[rt],rt,Xe)!==!1;);return Xe}function Eg(Xe,dt){for(var rt=Xe==null?0:Xe.length;rt--&&dt(Xe[rt],rt,Xe)!==!1;);return Xe}function Dp(Xe,dt){for(var rt=-1,Bt=Xe==null?0:Xe.length;++rt<Bt;)if(!dt(Xe[rt],rt,Xe))return!1;return!0}function Zo(Xe,dt){for(var rt=-1,Bt=Xe==null?0:Xe.length,En=0,lr=[];++rt<Bt;){var Li=Xe[rt];dt(Li,rt,Xe)&&(lr[En++]=Li)}return lr}function ph(Xe,dt){var rt=Xe==null?0:Xe.length;return!!rt&&Ku(Xe,dt,0)>-1}function Ps(Xe,dt,rt){for(var Bt=-1,En=Xe==null?0:Xe.length;++Bt<En;)if(rt(dt,Xe[Bt]))return!0;return!1}function bi(Xe,dt){for(var rt=-1,Bt=Xe==null?0:Xe.length,En=Array(Bt);++rt<Bt;)En[rt]=dt(Xe[rt],rt,Xe);return En}function Ho(Xe,dt){for(var rt=-1,Bt=dt.length,En=Xe.length;++rt<Bt;)Xe[En+rt]=dt[rt];return Xe}function $p(Xe,dt,rt,Bt){var En=-1,lr=Xe==null?0:Xe.length;for(Bt&&lr&&(rt=Xe[++En]);++En<lr;)rt=dt(rt,Xe[En],En,Xe);return rt}function Cp(Xe,dt,rt,Bt){var En=Xe==null?0:Xe.length;for(Bt&&En&&(rt=Xe[--En]);En--;)rt=dt(rt,Xe[En],En,Xe);return rt}function wp(Xe,dt){for(var rt=-1,Bt=Xe==null?0:Xe.length;++rt<Bt;)if(dt(Xe[rt],rt,Xe))return!0;return!1}var Ms=Xo("length");function Rp(Xe){return Xe.split("")}function Dy(Xe){return Xe.match(Yt)||[]}function K0(Xe,dt,rt){var Bt;return rt(Xe,function(En,lr,Li){if(dt(En,lr,Li))return Bt=lr,!1}),Bt}function po(Xe,dt,rt,Bt){for(var En=Xe.length,lr=rt+(Bt?1:-1);Bt?lr--:++lr<En;)if(dt(Xe[lr],lr,Xe))return lr;return-1}function Ku(Xe,dt,rt){return dt===dt?Dg(Xe,dt,rt):po(Xe,mh,rt)}function W0(Xe,dt,rt,Bt){for(var En=rt-1,lr=Xe.length;++En<lr;)if(Bt(Xe[En],dt))return En;return-1}function mh(Xe){return Xe!==Xe}function Ec(Xe,dt){var rt=Xe==null?0:Xe.length;return rt?vh(Xe,dt)/rt:M}function Xo(Xe){return function(dt){return dt==null?n:dt[Xe]}}function Fs(Xe){return function(dt){return Xe==null?n:Xe[dt]}}function Qa(Xe,dt,rt,Bt,En){return En(Xe,function(lr,Li,zr){rt=Bt?(Bt=!1,lr):dt(rt,lr,Li,zr)}),rt}function gh(Xe,dt){var rt=Xe.length;for(Xe.sort(dt);rt--;)Xe[rt]=Xe[rt].value;return Xe}function vh(Xe,dt){for(var rt,Bt=-1,En=Xe.length;++Bt<En;){var lr=dt(Xe[Bt]);lr!==n&&(rt=rt===n?lr:rt+lr)}return rt}function fd(Xe,dt){for(var rt=-1,Bt=Array(Xe);++rt<Xe;)Bt[rt]=dt(rt);return Bt}function dd(Xe,dt){return bi(dt,function(rt){return[rt,Xe[rt]]})}function Ns(Xe){return Xe&&Xe.slice(0,xf(Xe)+1).replace(Ve,"")}function pu(Xe){return function(dt){return Xe(dt)}}function Ap(Xe,dt){return bi(dt,function(rt){return Xe[rt]})}function Ga(Xe,dt){return Xe.has(dt)}function Tp(Xe,dt){for(var rt=-1,Bt=Xe.length;++rt<Bt&&Ku(dt,Xe[rt],0)>-1;);return rt}function hd(Xe,dt){for(var rt=Xe.length;rt--&&Ku(dt,Xe[rt],0)>-1;);return rt}function bf(Xe,dt){for(var rt=Xe.length,Bt=0;rt--;)Xe[rt]===dt&&++Bt;return Bt}var pd=Fs(xc),Wu=Fs(jo);function mo(Xe){return"\\"+ra[Xe]}function J0(Xe,dt){return Xe==null?n:Xe[dt]}function mu(Xe){return _p.test(Xe)}function Of(Xe){return dh.test(Xe)}function Ar(Xe){for(var dt,rt=[];!(dt=Xe.next()).done;)rt.push(dt.value);return rt}function yh(Xe){var dt=-1,rt=Array(Xe.size);return Xe.forEach(function(Bt,En){rt[++dt]=[En,Bt]}),rt}function kp(Xe,dt){return function(rt){return Xe(dt(rt))}}function go(Xe,dt){for(var rt=-1,Bt=Xe.length,En=0,lr=[];++rt<Bt;){var Li=Xe[rt];(Li===dt||Li===c)&&(Xe[rt]=c,lr[En++]=rt)}return lr}function Kn(Xe){var dt=-1,rt=Array(Xe.size);return Xe.forEach(function(Bt){rt[++dt]=Bt}),rt}function _l(Xe){var dt=-1,rt=Array(Xe.size);return Xe.forEach(function(Bt){rt[++dt]=[Bt,Bt]}),rt}function Dg(Xe,dt,rt){for(var Bt=rt-1,En=Xe.length;++Bt<En;)if(Xe[Bt]===dt)return Bt;return-1}function Ii(Xe,dt,rt){for(var Bt=rt+1;Bt--;)if(Xe[Bt]===dt)return Bt;return Bt}function gu(Xe){return mu(Xe)?tm(Xe):Ms(Xe)}function pr(Xe){return mu(Xe)?xr(Xe):Rp(Xe)}function xf(Xe){for(var dt=Xe.length;dt--&&nt.test(Xe.charAt(dt)););return dt}var em=Fs(Sc);function tm(Xe){for(var dt=As.lastIndex=0;As.test(Xe);)++dt;return dt}function xr(Xe){return Xe.match(As)||[]}function nm(Xe){return Xe.match(Uo)||[]}var Sf=(function Xe(dt){dt=dt==null?ia:Dc.defaults(ia.Object(),dt,Dc.pick(ia,Oc));var rt=dt.Array,Bt=dt.Date,En=dt.Error,lr=dt.Function,Li=dt.Math,zr=dt.Object,Ka=dt.RegExp,$c=dt.String,Ci=dt.TypeError,md=rt.prototype,Ql=lr.prototype,vo=zr.prototype,yo=dt["__core-js_shared__"],_f=Ql.toString,Sr=vo.hasOwnProperty,Yo=0,bh=(function(){var x=/[^.]+$/.exec(yo&&yo.keys&&yo.keys.IE_PROTO||"");return x?"Symbol(src)_1."+x:""})(),pa=vo.toString,Go=_f.call(zr),wa=ia._,bo=Ka("^"+_f.call(Sr).replace(ar,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$"),Is=ld?dt.Buffer:n,Wa=dt.Symbol,Ju=dt.Uint8Array,Cc=Is?Is.allocUnsafe:n,el=kp(zr.getPrototypeOf,zr),wc=zr.create,Rc=vo.propertyIsEnumerable,vu=md.splice,gd=Wa?Wa.isConcatSpreadable:n,aa=Wa?Wa.iterator:n,Ji=Wa?Wa.toStringTag:n,tl=(function(){try{var x=iu(zr,"defineProperty");return x({},"",{}),x}catch{}})(),yu=dt.clearTimeout!==ia.clearTimeout&&dt.clearTimeout,Oh=Bt&&Bt.now!==ia.Date.now&&Bt.now,zl=dt.setTimeout!==ia.setTimeout&&dt.setTimeout,Ef=Li.ceil,xh=Li.floor,bu=zr.getOwnPropertySymbols,Pp=Is?Is.isBuffer:n,Mu=dt.isFinite,ql=md.join,Ac=kp(zr.keys,zr),Bi=Li.max,wi=Li.min,Fu=Bt.now,dn=dt.parseInt,Nu=Li.random,Ko=md.reverse,ue=iu(dt,"DataView"),L=iu(dt,"Map"),re=iu(dt,"Promise"),fe=iu(dt,"Set"),xe=iu(dt,"WeakMap"),ke=iu(zr,"create"),Qe=xe&&new xe,Be={},Ze=uu(ue),Rt=uu(L),Kt=uu(re),zn=uu(fe),ln=uu(xe),mn=Wa?Wa.prototype:n,Ri=mn?mn.valueOf:n,za=mn?mn.toString:n;function Se(x){if(ja(x)&&!Wn(x)&&!(x instanceof qn)){if(x instanceof Iu)return x;if(Sr.call(x,"__wrapped__"))return wo(x)}return new Iu(x)}var El=(function(){function x(){}return function(C){if(!ba(C))return{};if(wc)return wc(C);x.prototype=C;var Z=new x;return x.prototype=n,Z}})();function Vl(){}function Iu(x,C){this.__wrapped__=x,this.__actions__=[],this.__chain__=!!C,this.__index__=0,this.__values__=n}Se.templateSettings={escape:Or,evaluate:Cr,interpolate:wr,variable:"",imports:{_:Se}},Se.prototype=Vl.prototype,Se.prototype.constructor=Se,Iu.prototype=El(Vl.prototype),Iu.prototype.constructor=Iu;function qn(x){this.__wrapped__=x,this.__actions__=[],this.__dir__=1,this.__filtered__=!1,this.__iteratees__=[],this.__takeCount__=G,this.__views__=[]}function Ja(){var x=new qn(this.__wrapped__);return x.__actions__=ul(this.__actions__),x.__dir__=this.__dir__,x.__filtered__=this.__filtered__,x.__iteratees__=ul(this.__iteratees__),x.__takeCount__=this.__takeCount__,x.__views__=ul(this.__views__),x}function Ls(){if(this.__filtered__){var x=new qn(this);x.__dir__=-1,x.__filtered__=!0}else x=this.clone(),x.__dir__*=-1;return x}function Ul(){var x=this.__wrapped__.value(),C=this.__dir__,Z=Wn(x),ae=C<0,he=Z?x.length:0,Ce=mm(0,he,this.__views__),Pe=Ce.start,qe=Ce.end,Je=qe-Pe,St=ae?qe:Pe-1,_t=this.__iteratees__,At=_t.length,Jt=0,yn=wi(Je,this.__takeCount__);if(!Z||!ae&&he==Je&&yn==Je)return zy(x,this.__actions__);var Un=[];e:for(;Je--&&Jt<yn;){St+=C;for(var Er=-1,jn=x[St];++Er<At;){var jr=_t[Er],ni=jr.iteratee,Mf=jr.type,tc=ni(jn);if(Mf==W)jn=tc;else if(!tc){if(Mf==j)continue e;break e}}Un[Jt++]=jn}return Un}qn.prototype=El(Vl.prototype),qn.prototype.constructor=qn;function Lu(x){var C=-1,Z=x==null?0:x.length;for(this.clear();++C<Z;){var ae=x[C];this.set(ae[0],ae[1])}}function Tc(){this.__data__=ke?ke(null):{},this.size=0}function jl(x){var C=this.has(x)&&delete this.__data__[x];return this.size-=C?1:0,C}function vd(x){var C=this.__data__;if(ke){var Z=C[x];return Z===o?n:Z}return Sr.call(C,x)?C[x]:n}function Sh(x){var C=this.__data__;return ke?C[x]!==n:Sr.call(C,x)}function Ou(x,C){var Z=this.__data__;return this.size+=this.has(x)?0:1,Z[x]=ke&&C===n?o:C,this}Lu.prototype.clear=Tc,Lu.prototype.delete=jl,Lu.prototype.get=vd,Lu.prototype.has=Sh,Lu.prototype.set=Ou;function Oo(x){var C=-1,Z=x==null?0:x.length;for(this.clear();++C<Z;){var ae=x[C];this.set(ae[0],ae[1])}}function $y(){this.__data__=[],this.size=0}function I1(x){var C=this.__data__,Z=xn(C,x);if(Z<0)return!1;var ae=C.length-1;return Z==ae?C.pop():vu.call(C,Z,1),--this.size,!0}function L1(x){var C=this.__data__,Z=xn(C,x);return Z<0?n:C[Z][1]}function Cy(x){return xn(this.__data__,x)>-1}function gn(x,C){var Z=this.__data__,ae=xn(Z,x);return ae<0?(++this.size,Z.push([x,C])):Z[ae][1]=C,this}Oo.prototype.clear=$y,Oo.prototype.delete=I1,Oo.prototype.get=L1,Oo.prototype.has=Cy,Oo.prototype.set=gn;function Bs(x){var C=-1,Z=x==null?0:x.length;for(this.clear();++C<Z;){var ae=x[C];this.set(ae[0],ae[1])}}function Bu(){this.size=0,this.__data__={hash:new Lu,map:new(L||Oo),string:new Lu}}function _h(x){var C=Td(this,x).delete(x);return this.size-=C?1:0,C}function wy(x){return Td(this,x).get(x)}function Df(x){return Td(this,x).has(x)}function Qs(x,C){var Z=Td(this,x),ae=Z.size;return Z.set(x,C),this.size+=Z.size==ae?0:1,this}Bs.prototype.clear=Bu,Bs.prototype.delete=_h,Bs.prototype.get=wy,Bs.prototype.has=Df,Bs.prototype.set=Qs;function Wo(x){var C=-1,Z=x==null?0:x.length;for(this.__data__=new Bs;++C<Z;)this.add(x[C])}function Ry(x){return this.__data__.set(x,o),this}function Ay(x){return this.__data__.has(x)}Wo.prototype.add=Wo.prototype.push=Ry,Wo.prototype.has=Ay;function Jo(x){var C=this.__data__=new Oo(x);this.size=C.size}function yd(){this.__data__=new Oo,this.size=0}function kc(x){var C=this.__data__,Z=C.delete(x);return this.size=C.size,Z}function Mp(x){return this.__data__.get(x)}function $g(x){return this.__data__.has(x)}function bd(x,C){var Z=this.__data__;if(Z instanceof Oo){var ae=Z.__data__;if(!L||ae.length<i-1)return ae.push([x,C]),this.size=++Z.size,this;Z=this.__data__=new Bs(ae)}return Z.set(x,C),this.size=Z.size,this}Jo.prototype.clear=yd,Jo.prototype.delete=kc,Jo.prototype.get=Mp,Jo.prototype.has=$g,Jo.prototype.set=bd;function Fp(x,C){var Z=Wn(x),ae=!Z&&$m(x),he=!Z&&!ae&&Yp(x),Ce=!Z&&!ae&&!he&&Gp(x),Pe=Z||ae||he||Ce,qe=Pe?fd(x.length,$c):[],Je=qe.length;for(var St in x)(C||Sr.call(x,St))&&!(Pe&&(St=="length"||he&&(St=="offset"||St=="parent")||Ce&&(St=="buffer"||St=="byteLength"||St=="byteOffset")||Do(St,Je)))&&qe.push(St);return qe}function es(x){var C=x.length;return C?x[Qp(0,C-1)]:n}function rm(x,C){return Xs(ul(x),qs(C,0,x.length))}function B1(x){return Xs(ul(x))}function im(x,C,Z){(Z!==n&&!Ws(x[C],Z)||Z===n&&!(C in x))&&nl(x,C,Z)}function zs(x,C,Z){var ae=x[C];(!(Sr.call(x,C)&&Ws(ae,Z))||Z===n&&!(C in x))&&nl(x,C,Z)}function xn(x,C){for(var Z=x.length;Z--;)if(Ws(x[Z][0],C))return Z;return-1}function Cg(x,C,Z,ae){return Pc(x,function(he,Ce,Pe){C(ae,he,Z(he),Pe)}),ae}function wg(x,C){return x&&as(C,pl(C),x)}function Od(x,C){return x&&as(C,hs(C),x)}function nl(x,C,Z){C=="__proto__"&&tl?tl(x,C,{configurable:!0,enumerable:!0,value:Z,writable:!0}):x[C]=Z}function Np(x,C){for(var Z=-1,ae=C.length,he=rt(ae),Ce=x==null;++Z<ae;)he[Z]=Ce?n:Kp(x,C[Z]);return he}function qs(x,C,Z){return x===x&&(Z!==n&&(x=x<=Z?x:Z),C!==n&&(x=x>=C?x:C)),x}function rl(x,C,Z,ae,he,Ce){var Pe,qe=C&d,Je=C&h,St=C&m;if(Z&&(Pe=he?Z(x,ae,he,Ce):Z(x)),Pe!==n)return Pe;if(!ba(x))return x;var _t=Wn(x);if(_t){if(Pe=sl(x),!qe)return ul(x,Pe)}else{var At=qr(x),Jt=At==se||At==pe;if(Yp(x))return Uy(x,qe);if(At==Re||At==oe||Jt&&!he){if(Pe=Je||Jt?{}:Eu(x),!qe)return Je?Zy(x,Od(Pe,x)):jy(x,wg(Pe,x))}else{if(!Qn[At])return he?x:{};Pe=Vr(x,At,qe)}}Ce||(Ce=new Jo);var yn=Ce.get(x);if(yn)return yn;Ce.set(x,Pe),lb(x)?x.forEach(function(jn){Pe.add(rl(jn,C,Z,jn,x,Ce))}):n_(x)&&x.forEach(function(jn,jr){Pe.set(jr,rl(jn,C,Z,jr,x,Ce))});var Un=St?Je?Dl:ol:Je?hs:pl,Er=_t?n:Un(x);return Ba(Er||x,function(jn,jr){Er&&(jr=jn,jn=x[jr]),zs(Pe,jr,rl(jn,C,Z,jr,x,Ce))}),Pe}function Ip(x){var C=pl(x);return function(Z){return xd(Z,x,C)}}function xd(x,C,Z){var ae=Z.length;if(x==null)return!ae;for(x=zr(x);ae--;){var he=Z[ae],Ce=C[he],Pe=x[he];if(Pe===n&&!(he in x)||!Ce(Pe))return!1}return!0}function Ty(x,C,Z){if(typeof x!="function")throw new Ci(u);return Md(function(){x.apply(n,Z)},C)}function ts(x,C,Z,ae){var he=-1,Ce=ph,Pe=!0,qe=x.length,Je=[],St=C.length;if(!qe)return Je;Z&&(C=bi(C,pu(Z))),ae?(Ce=Ps,Pe=!1):C.length>=i&&(Ce=Ga,Pe=!1,C=new Wo(C));e:for(;++he<qe;){var _t=x[he],At=Z==null?_t:Z(_t);if(_t=ae||_t!==0?_t:0,Pe&&At===At){for(var Jt=St;Jt--;)if(C[Jt]===At)continue e;Je.push(_t)}else Ce(C,At,ae)||Je.push(_t)}return Je}var Pc=Hy(tu),Rg=Hy(nu,!0);function Mc(x,C){var Z=!0;return Pc(x,function(ae,he,Ce){return Z=!!C(ae,he,Ce),Z}),Z}function ns(x,C,Z){for(var ae=-1,he=x.length;++ae<he;){var Ce=x[ae],Pe=C(Ce);if(Pe!=null&&(qe===n?Pe===Pe&&!ds(Pe):Z(Pe,qe)))var qe=Pe,Je=Ce}return Je}function Sd(x,C,Z,ae){var he=x.length;for(Z=Vn(Z),Z<0&&(Z=-Z>he?0:he+Z),ae=ae===n||ae>he?he:Vn(ae),ae<0&&(ae+=he),ae=Z>ae?0:N4(ae);Z<ae;)x[Z++]=C;return x}function am(x,C){var Z=[];return Pc(x,function(ae,he,Ce){C(ae,he,Ce)&&Z.push(ae)}),Z}function eu(x,C,Z,ae,he){var Ce=-1,Pe=x.length;for(Z||(Z=ls),he||(he=[]);++Ce<Pe;){var qe=x[Ce];C>0&&Z(qe)?C>1?eu(qe,C-1,Z,ae,he):Ho(he,qe):ae||(he[he.length]=qe)}return he}var Eh=Zg(),_d=Zg(!0);function tu(x,C){return x&&Eh(x,C,pl)}function nu(x,C){return x&&_d(x,C,pl)}function il(x,C){return Zo(C,function(Z){return Js(x[Z])})}function Ra(x,C){C=So(C,x);for(var Z=0,ae=C.length;x!=null&&Z<ae;)x=x[au(C[Z++])];return Z&&Z==ae?x:n}function Ag(x,C,Z){var ae=C(x);return Wn(x)?ae:Ho(ae,Z(x))}function Qu(x){return x==null?x===n?je:Me:Ji&&Ji in zr(x)?js(x):Pd(x)}function xu(x,C){return x>C}function ru(x,C){return x!=null&&Sr.call(x,C)}function rs(x,C){return x!=null&&C in zr(x)}function Ed(x,C,Z){return x>=wi(C,Z)&&x<Bi(C,Z)}function um(x,C,Z){for(var ae=Z?Ps:ph,he=x[0].length,Ce=x.length,Pe=Ce,qe=rt(Ce),Je=1/0,St=[];Pe--;){var _t=x[Pe];Pe&&C&&(_t=bi(_t,pu(C))),Je=wi(_t.length,Je),qe[Pe]=!Z&&(C||he>=120&&_t.length>=120)?new Wo(Pe&&_t):n}_t=x[0];var At=-1,Jt=qe[0];e:for(;++At<he&&St.length<Je;){var yn=_t[At],Un=C?C(yn):yn;if(yn=Z||yn!==0?yn:0,!(Jt?Ga(Jt,Un):ae(St,Un,Z))){for(Pe=Ce;--Pe;){var Er=qe[Pe];if(!(Er?Ga(Er,Un):ae(x[Pe],Un,Z)))continue e}Jt&&Jt.push(Un),St.push(yn)}}return St}function Q1(x,C,Z,ae){return tu(x,function(he,Ce,Pe){C(ae,Z(he),Ce,Pe)}),ae}function xo(x,C,Z){C=So(C,x),x=Mh(x,C);var ae=x==null?x:x[au(Ur(C))];return ae==null?n:ha(ae,x,Z)}function Zl(x){return ja(x)&&Qu(x)==oe}function ma(x){return ja(x)&&Qu(x)==ze}function Tg(x){return ja(x)&&Qu(x)==ye}function Su(x,C,Z,ae,he){return x===C?!0:x==null||C==null||!ja(x)&&!ja(C)?x!==x&&C!==C:An(x,C,Z,ae,Su,he)}function An(x,C,Z,ae,he,Ce){var Pe=Wn(x),qe=Wn(C),Je=Pe?ie:qr(x),St=qe?ie:qr(C);Je=Je==oe?Re:Je,St=St==oe?Re:St;var _t=Je==Re,At=St==Re,Jt=Je==St;if(Jt&&Yp(x)){if(!Yp(C))return!1;Pe=!0,_t=!1}if(Jt&&!_t)return Ce||(Ce=new Jo),Pe||Gp(x)?pm(x,C,Z,ae,he,Ce):X1(x,C,Je,Z,ae,he,Ce);if(!(Z&y)){var yn=_t&&Sr.call(x,"__wrapped__"),Un=At&&Sr.call(C,"__wrapped__");if(yn||Un){var Er=yn?x.value():x,jn=Un?C.value():C;return Ce||(Ce=new Jo),he(Er,jn,Z,ae,Ce)}}return Jt?(Ce||(Ce=new Jo),Wy(x,C,Z,ae,he,Ce)):!1}function Oi(x){return ja(x)&&qr(x)==ve}function ga(x,C,Z,ae){var he=Z.length,Ce=he,Pe=!ae;if(x==null)return!Ce;for(x=zr(x);he--;){var qe=Z[he];if(Pe&&qe[2]?qe[1]!==x[qe[0]]:!(qe[0]in x))return!1}for(;++he<Ce;){qe=Z[he];var Je=qe[0],St=x[Je],_t=qe[1];if(Pe&&qe[2]){if(St===n&&!(Je in x))return!1}else{var At=new Jo;if(ae)var Jt=ae(St,_t,Je,x,C,At);if(!(Jt===n?Su(_t,St,y|v,ae,At):Jt))return!1}}return!0}function lm(x){if(!ba(x)||Ph(x))return!1;var C=Js(x)?bo:er;return C.test(uu(x))}function Fc(x){return ja(x)&&Qu(x)==Oe}function Dh(x){return ja(x)&&qr(x)==$e}function kg(x){return ja(x)&&tO(x.length)&&!!Mr[Qu(x)]}function Lp(x){return typeof x=="function"?x:x==null?ko:typeof x=="object"?Wn(x)?ky(x[0],x[1]):Mg(x):_e(x)}function is(x){if(!$f(x))return Ac(x);var C=[];for(var Z in zr(x))Sr.call(x,Z)&&Z!="constructor"&&C.push(Z);return C}function O4(x){if(!ba(x))return Kl(x);var C=$f(x),Z=[];for(var ae in x)ae=="constructor"&&(C||!Sr.call(x,ae))||Z.push(ae);return Z}function ea(x,C){return x<C}function Pg(x,C){var Z=-1,ae=fs(x)?rt(x.length):[];return Pc(x,function(he,Ce,Pe){ae[++Z]=C(he,Ce,Pe)}),ae}function Mg(x){var C=$l(x);return C.length==1&&C[0][2]?Cf(C[0][0],C[0][1]):function(Z){return Z===x||ga(Z,x,C)}}function ky(x,C){return Bc(x)&&os(C)?Cf(au(x),C):function(Z){var ae=Kp(Z,x);return ae===n&&ae===C?L4(Z,x):Su(C,ae,y|v)}}function om(x,C,Z,ae,he){x!==C&&Eh(C,function(Ce,Pe){if(he||(he=new Jo),ba(Ce))x4(x,C,Pe,Z,om,ae,he);else{var qe=ae?ae(Wr(x,Pe),Ce,Pe+"",x,C,he):n;qe===n&&(qe=Ce),im(x,Pe,qe)}},hs)}function x4(x,C,Z,ae,he,Ce,Pe){var qe=Wr(x,Z),Je=Wr(C,Z),St=Pe.get(Je);if(St){im(x,Z,St);return}var _t=Ce?Ce(qe,Je,Z+"",x,C,Pe):n,At=_t===n;if(At){var Jt=Wn(Je),yn=!Jt&&Yp(Je),Un=!Jt&&!yn&&Gp(Je);_t=Je,Jt||yn||Un?Wn(qe)?_t=qe:ou(qe)?_t=ul(qe):yn?(At=!1,_t=Uy(Je,!0)):Un?(At=!1,_t=Yl(Je,!0)):_t=[]:ab(Je)||$m(Je)?(_t=qe,$m(qe)?_t=nO(qe):(!ba(qe)||Js(qe))&&(_t=Eu(Je))):At=!1}At&&(Pe.set(Je,_t),he(_t,Je,ae,Ce,Pe),Pe.delete(Je)),im(x,Z,_t)}function Fg(x,C){var Z=x.length;if(Z)return C+=C<0?Z:0,Do(C,Z)?x[C]:n}function Ng(x,C,Z){C.length?C=bi(C,function(Ce){return Wn(Ce)?function(Pe){return Ra(Pe,Ce.length===1?Ce[0]:Ce)}:Ce}):C=[ko];var ae=-1;C=bi(C,pu(On()));var he=Pg(x,function(Ce,Pe,qe){var Je=bi(C,function(St){return St(Ce)});return{criteria:Je,index:++ae,value:Ce}});return gh(he,function(Ce,Pe){return Vg(Ce,Pe,Z)})}function Py(x,C){return al(x,C,function(Z,ae){return L4(x,ae)})}function al(x,C,Z){for(var ae=-1,he=C.length,Ce={};++ae<he;){var Pe=C[ae],qe=Ra(x,Pe);Z(qe,Pe)&&Ch(Ce,So(Pe,x),qe)}return Ce}function Aa(x){return function(C){return Ra(C,x)}}function $h(x,C,Z,ae){var he=ae?W0:Ku,Ce=-1,Pe=C.length,qe=x;for(x===C&&(C=ul(C)),Z&&(qe=bi(x,pu(Z)));++Ce<Pe;)for(var Je=0,St=C[Ce],_t=Z?Z(St):St;(Je=he(qe,_t,Je,ae))>-1;)qe!==x&&vu.call(qe,Je,1),vu.call(x,Je,1);return x}function Bp(x,C){for(var Z=x?C.length:0,ae=Z-1;Z--;){var he=C[Z];if(Z==ae||he!==Ce){var Ce=he;Do(he)?vu.call(x,he,1):zp(x,he)}}return x}function Qp(x,C){return x+xh(Nu()*(C-x+1))}function My(x,C,Z,ae){for(var he=-1,Ce=Bi(Ef((C-x)/(Z||1)),0),Pe=rt(Ce);Ce--;)Pe[ae?Ce:++he]=x,x+=Z;return Pe}function Vs(x,C){var Z="";if(!x||C<1||C>te)return Z;do C%2&&(Z+=x),C=xh(C/2),C&&(x+=x);while(C);return Z}function Jn(x,C){return wf(Qc(x,C,ko),x+"")}function Fy(x){return es(sv(x))}function Ig(x,C){var Z=sv(x);return Xs(Z,qs(C,0,Z.length))}function Ch(x,C,Z,ae){if(!ba(x))return x;C=So(C,x);for(var he=-1,Ce=C.length,Pe=Ce-1,qe=x;qe!=null&&++he<Ce;){var Je=au(C[he]),St=Z;if(Je==="__proto__"||Je==="constructor"||Je==="prototype")return x;if(he!=Pe){var _t=qe[Je];St=ae?ae(_t,Je,qe):n,St===n&&(St=ba(_t)?_t:Do(C[he+1])?[]:{})}zs(qe,Je,St),qe=qe[Je]}return x}var Ny=Qe?function(x,C){return Qe.set(x,C),x}:ko,Iy=tl?function(x,C){return tl(x,"toString",{configurable:!0,enumerable:!1,value:fv(C),writable:!0})}:ko;function Ly(x){return Xs(sv(x))}function Hl(x,C,Z){var ae=-1,he=x.length;C<0&&(C=-C>he?0:he+C),Z=Z>he?he:Z,Z<0&&(Z+=he),he=C>Z?0:Z-C>>>0,C>>>=0;for(var Ce=rt(he);++ae<he;)Ce[ae]=x[ae+C];return Ce}function Lg(x,C){var Z;return Pc(x,function(ae,he,Ce){return Z=C(ae,he,Ce),!Z}),!!Z}function Dd(x,C,Z){var ae=0,he=x==null?ae:x.length;if(typeof C=="number"&&C===C&&he<=B){for(;ae<he;){var Ce=ae+he>>>1,Pe=x[Ce];Pe!==null&&!ds(Pe)&&(Z?Pe<=C:Pe<C)?ae=Ce+1:he=Ce}return he}return Bg(x,C,ko,Z)}function Bg(x,C,Z,ae){var he=0,Ce=x==null?0:x.length;if(Ce===0)return 0;C=Z(C);for(var Pe=C!==C,qe=C===null,Je=ds(C),St=C===n;he<Ce;){var _t=xh((he+Ce)/2),At=Z(x[_t]),Jt=At!==n,yn=At===null,Un=At===At,Er=ds(At);if(Pe)var jn=ae||Un;else St?jn=Un&&(ae||Jt):qe?jn=Un&&Jt&&(ae||!yn):Je?jn=Un&&Jt&&!yn&&(ae||!Er):yn||Er?jn=!1:jn=ae?At<=C:At<C;jn?he=_t+1:Ce=_t}return wi(Ce,X)}function z1(x,C){for(var Z=-1,ae=x.length,he=0,Ce=[];++Z<ae;){var Pe=x[Z],qe=C?C(Pe):Pe;if(!Z||!Ws(qe,Je)){var Je=qe;Ce[he++]=Pe===0?0:Pe}}return Ce}function By(x){return typeof x=="number"?x:ds(x)?M:+x}function Xl(x){if(typeof x=="string")return x;if(Wn(x))return bi(x,Xl)+"";if(ds(x))return za?za.call(x):"";var C=x+"";return C=="0"&&1/x==-J?"-0":C}function Nc(x,C,Z){var ae=-1,he=ph,Ce=x.length,Pe=!0,qe=[],Je=qe;if(Z)Pe=!1,he=Ps;else if(Ce>=i){var St=C?null:H1(x);if(St)return Kn(St);Pe=!1,he=Ga,Je=new Wo}else Je=C?[]:qe;e:for(;++ae<Ce;){var _t=x[ae],At=C?C(_t):_t;if(_t=Z||_t!==0?_t:0,Pe&&At===At){for(var Jt=Je.length;Jt--;)if(Je[Jt]===At)continue e;C&&Je.push(At),qe.push(_t)}else he(Je,At,Z)||(Je!==qe&&Je.push(At),qe.push(_t))}return qe}function zp(x,C){C=So(C,x);var Z=-1,ae=C.length;if(!ae)return!0;for(var he=x==null||typeof x!="object"&&typeof x!="function";++Z<ae;){var Ce=C[Z];if(typeof Ce=="string"){if(Ce==="__proto__"&&!Sr.call(x,"__proto__"))return!1;if(Ce==="constructor"&&Z+1<ae&&typeof C[Z+1]=="string"&&C[Z+1]==="prototype"){if(he&&Z===0)continue;return!1}}}var Pe=Mh(x,C);return Pe==null||delete Pe[au(Ur(C))]}function Qy(x,C,Z,ae){return Ch(x,C,Z(Ra(x,C)),ae)}function sm(x,C,Z,ae){for(var he=x.length,Ce=ae?he:-1;(ae?Ce--:++Ce<he)&&C(x[Ce],Ce,x););return Z?Hl(x,ae?0:Ce,ae?Ce+1:he):Hl(x,ae?Ce+1:0,ae?he:Ce)}function zy(x,C){var Z=x;return Z instanceof qn&&(Z=Z.value()),$p(C,function(ae,he){return he.func.apply(he.thisArg,Ho([ae],he.args))},Z)}function Qg(x,C,Z){var ae=x.length;if(ae<2)return ae?Nc(x[0]):[];for(var he=-1,Ce=rt(ae);++he<ae;)for(var Pe=x[he],qe=-1;++qe<ae;)qe!=he&&(Ce[he]=ts(Ce[he]||Pe,x[qe],C,Z));return Nc(eu(Ce,1),C,Z)}function qy(x,C,Z){for(var ae=-1,he=x.length,Ce=C.length,Pe={};++ae<he;){var qe=ae<Ce?C[ae]:n;Z(Pe,x[ae],qe)}return Pe}function Vy(x){return ou(x)?x:[]}function zg(x){return typeof x=="function"?x:ko}function So(x,C){return Wn(x)?x:Bc(x,C)?[x]:Kg(_i(x))}var q1=Jn;function Us(x,C,Z){var ae=x.length;return Z=Z===n?ae:Z,!C&&Z>=ae?x:Hl(x,C,Z)}var cm=yu||function(x){return ia.clearTimeout(x)};function Uy(x,C){if(C)return x.slice();var Z=x.length,ae=Cc?Cc(Z):new x.constructor(Z);return x.copy(ae),ae}function fm(x){var C=new x.constructor(x.byteLength);return new Ju(C).set(new Ju(x)),C}function V1(x,C){var Z=C?fm(x.buffer):x.buffer;return new x.constructor(Z,x.byteOffset,x.byteLength)}function U1(x){var C=new x.constructor(x.source,_n.exec(x));return C.lastIndex=x.lastIndex,C}function j1(x){return Ri?zr(Ri.call(x)):{}}function Yl(x,C){var Z=C?fm(x.buffer):x.buffer;return new x.constructor(Z,x.byteOffset,x.length)}function qg(x,C){if(x!==C){var Z=x!==n,ae=x===null,he=x===x,Ce=ds(x),Pe=C!==n,qe=C===null,Je=C===C,St=ds(C);if(!qe&&!St&&!Ce&&x>C||Ce&&Pe&&Je&&!qe&&!St||ae&&Pe&&Je||!Z&&Je||!he)return 1;if(!ae&&!Ce&&!St&&x<C||St&&Z&&he&&!ae&&!Ce||qe&&Z&&he||!Pe&&he||!Je)return-1}return 0}function Vg(x,C,Z){for(var ae=-1,he=x.criteria,Ce=C.criteria,Pe=he.length,qe=Z.length;++ae<Pe;){var Je=qg(he[ae],Ce[ae]);if(Je){if(ae>=qe)return Je;var St=Z[ae];return Je*(St=="desc"?-1:1)}}return x.index-C.index}function dm(x,C,Z,ae){for(var he=-1,Ce=x.length,Pe=Z.length,qe=-1,Je=C.length,St=Bi(Ce-Pe,0),_t=rt(Je+St),At=!ae;++qe<Je;)_t[qe]=C[qe];for(;++he<Pe;)(At||he<Ce)&&(_t[Z[he]]=x[he]);for(;St--;)_t[qe++]=x[he++];return _t}function Ug(x,C,Z,ae){for(var he=-1,Ce=x.length,Pe=-1,qe=Z.length,Je=-1,St=C.length,_t=Bi(Ce-qe,0),At=rt(_t+St),Jt=!ae;++he<_t;)At[he]=x[he];for(var yn=he;++Je<St;)At[yn+Je]=C[Je];for(;++Pe<qe;)(Jt||he<Ce)&&(At[yn+Z[Pe]]=x[he++]);return At}function ul(x,C){var Z=-1,ae=x.length;for(C||(C=rt(ae));++Z<ae;)C[Z]=x[Z];return C}function as(x,C,Z,ae){var he=!Z;Z||(Z={});for(var Ce=-1,Pe=C.length;++Ce<Pe;){var qe=C[Ce],Je=ae?ae(Z[qe],x[qe],qe,Z,x):n;Je===n&&(Je=x[qe]),he?nl(Z,qe,Je):zs(Z,qe,Je)}return Z}function jy(x,C){return as(x,us(x),C)}function Zy(x,C){return as(x,_o(x),C)}function jg(x,C){return function(Z,ae){var he=Wn(Z)?cd:Cg,Ce=C?C():{};return he(Z,x,On(ae,2),Ce)}}function $d(x){return Jn(function(C,Z){var ae=-1,he=Z.length,Ce=he>1?Z[he-1]:n,Pe=he>2?Z[2]:n;for(Ce=x.length>3&&typeof Ce=="function"?(he--,Ce):n,Pe&&Ti(Z[0],Z[1],Pe)&&(Ce=he<3?n:Ce,he=1),C=zr(C);++ae<he;){var qe=Z[ae];qe&&x(C,qe,ae,Ce)}return C})}function Hy(x,C){return function(Z,ae){if(Z==null)return Z;if(!fs(Z))return x(Z,ae);for(var he=Z.length,Ce=C?he:-1,Pe=zr(Z);(C?Ce--:++Ce<he)&&ae(Pe[Ce],Ce,Pe)!==!1;);return Z}}function Zg(x){return function(C,Z,ae){for(var he=-1,Ce=zr(C),Pe=ae(C),qe=Pe.length;qe--;){var Je=Pe[x?qe:++he];if(Z(Ce[Je],Je,Ce)===!1)break}return C}}function Xy(x,C,Z){var ae=C&S,he=hm(x);function Ce(){var Pe=this&&this!==ia&&this instanceof Ce?he:x;return Pe.apply(ae?Z:this,arguments)}return Ce}function _u(x){return function(C){C=_i(C);var Z=mu(C)?pr(C):n,ae=Z?Z[0]:C.charAt(0),he=Z?Us(Z,1).join(""):C.slice(1);return ae[x]()+he}}function Cd(x){return function(C){return $p(R_(cv(C).replace(xp,"")),x,"")}}function hm(x){return function(){var C=arguments;switch(C.length){case 0:return new x;case 1:return new x(C[0]);case 2:return new x(C[0],C[1]);case 3:return new x(C[0],C[1],C[2]);case 4:return new x(C[0],C[1],C[2],C[3]);case 5:return new x(C[0],C[1],C[2],C[3],C[4]);case 6:return new x(C[0],C[1],C[2],C[3],C[4],C[5]);case 7:return new x(C[0],C[1],C[2],C[3],C[4],C[5],C[6])}var Z=El(x.prototype),ae=x.apply(Z,C);return ba(ae)?ae:Z}}function S4(x,C,Z){var ae=hm(x);function he(){for(var Ce=arguments.length,Pe=rt(Ce),qe=Ce,Je=Ai(he);qe--;)Pe[qe]=arguments[qe];var St=Ce<3&&Pe[0]!==Je&&Pe[Ce-1]!==Je?[]:go(Pe,Je);if(Ce-=St.length,Ce<Z)return Ah(x,C,wd,he.placeholder,n,Pe,St,n,n,Z-Ce);var _t=this&&this!==ia&&this instanceof he?ae:x;return ha(_t,this,Pe)}return he}function Yy(x){return function(C,Z,ae){var he=zr(C);if(!fs(C)){var Ce=On(Z,3);C=pl(C),Z=function(qe){return Ce(he[qe],qe,he)}}var Pe=x(C,Z,ae);return Pe>-1?he[Ce?C[Pe]:Pe]:n}}function wh(x){return Gl(function(C){var Z=C.length,ae=Z,he=Iu.prototype.thru;for(x&&C.reverse();ae--;){var Ce=C[ae];if(typeof Ce!="function")throw new Ci(u);if(he&&!Pe&&xi(Ce)=="wrapper")var Pe=new Iu([],!0)}for(ae=Pe?ae:Z;++ae<Z;){Ce=C[ae];var qe=xi(Ce),Je=qe=="wrapper"?qp(Ce):n;Je&&kd(Je[0])&&Je[1]==(R|$|A|F)&&!Je[4].length&&Je[9]==1?Pe=Pe[xi(Je[0])].apply(Pe,Je[3]):Pe=Ce.length==1&&kd(Ce)?Pe[qe]():Pe.thru(Ce)}return function(){var St=arguments,_t=St[0];if(Pe&&St.length==1&&Wn(_t))return Pe.plant(_t).value();for(var At=0,Jt=Z?C[At].apply(this,St):_t;++At<Z;)Jt=C[At].call(this,Jt);return Jt}})}function wd(x,C,Z,ae,he,Ce,Pe,qe,Je,St){var _t=C&R,At=C&S,Jt=C&_,yn=C&($|k),Un=C&I,Er=Jt?n:hm(x);function jn(){for(var jr=arguments.length,ni=rt(jr),Mf=jr;Mf--;)ni[Mf]=arguments[Mf];if(yn)var tc=Ai(jn),Ff=bf(ni,tc);if(ae&&(ni=dm(ni,ae,he,yn)),Ce&&(ni=Ug(ni,Ce,Pe,yn)),jr-=Ff,yn&&jr<St){var zu=go(ni,tc);return Ah(x,C,wd,jn.placeholder,Z,ni,zu,qe,Je,St-jr)}var Uh=At?Z:this,wm=Jt?Uh[x]:x;return jr=ni.length,qe?ni=ym(ni,qe):Un&&jr>1&&ni.reverse(),_t&&Je<jr&&(ni.length=Je),this&&this!==ia&&this instanceof jn&&(wm=Er||hm(wm)),wm.apply(Uh,ni)}return jn}function Ic(x,C){return function(Z,ae){return Q1(Z,x,C(ae),{})}}function Lc(x,C){return function(Z,ae){var he;if(Z===n&&ae===n)return C;if(Z!==n&&(he=Z),ae!==n){if(he===n)return ae;typeof Z=="string"||typeof ae=="string"?(Z=Xl(Z),ae=Xl(ae)):(Z=By(Z),ae=By(ae)),he=x(Z,ae)}return he}}function ll(x){return Gl(function(C){return C=bi(C,pu(On())),Jn(function(Z){var ae=this;return x(C,function(he){return ha(he,ae,Z)})})})}function Rd(x,C){C=C===n?" ":Xl(C);var Z=C.length;if(Z<2)return Z?Vs(C,x):C;var ae=Vs(C,Ef(x/gu(C)));return mu(C)?Us(pr(ae),0,x).join(""):ae.slice(0,x)}function Z1(x,C,Z,ae){var he=C&S,Ce=hm(x);function Pe(){for(var qe=-1,Je=arguments.length,St=-1,_t=ae.length,At=rt(_t+Je),Jt=this&&this!==ia&&this instanceof Pe?Ce:x;++St<_t;)At[St]=ae[St];for(;Je--;)At[St++]=arguments[++qe];return ha(Jt,he?Z:this,At)}return Pe}function Rh(x){return function(C,Z,ae){return ae&&typeof ae!="number"&&Ti(C,Z,ae)&&(Z=ae=n),C=qh(C),Z===n?(Z=C,C=0):Z=qh(Z),ae=ae===n?C<Z?1:-1:qh(ae),My(C,Z,ae,x)}}function Ad(x){return function(C,Z){return typeof C=="string"&&typeof Z=="string"||(C=ec(C),Z=ec(Z)),x(C,Z)}}function Ah(x,C,Z,ae,he,Ce,Pe,qe,Je,St){var _t=C&$,At=_t?Pe:n,Jt=_t?n:Pe,yn=_t?Ce:n,Un=_t?n:Ce;C|=_t?A:w,C&=~(_t?w:A),C&E||(C&=-4);var Er=[x,C,he,yn,At,Un,Jt,qe,Je,St],jn=Z.apply(n,Er);return kd(x)&&Gg(jn,Er),jn.placeholder=ae,Hs(jn,x,C)}function Th(x){var C=Li[x];return function(Z,ae){if(Z=ec(Z),ae=ae==null?0:wi(Vn(ae),292),ae&&Mu(Z)){var he=(_i(Z)+"e").split("e"),Ce=C(he[0]+"e"+(+he[1]+ae));return he=(_i(Ce)+"e").split("e"),+(he[0]+"e"+(+he[1]-ae))}return C(Z)}}var H1=fe&&1/Kn(new fe([,-0]))[1]==J?function(x){return new fe(x)}:O;function Gy(x){return function(C){var Z=qr(C);return Z==ve?yh(C):Z==$e?_l(C):dd(C,x(C))}}function Hi(x,C,Z,ae,he,Ce,Pe,qe){var Je=C&_;if(!Je&&typeof x!="function")throw new Ci(u);var St=ae?ae.length:0;if(St||(C&=-97,ae=he=n),Pe=Pe===n?Pe:Bi(Vn(Pe),0),qe=qe===n?qe:Vn(qe),St-=he?he.length:0,C&w){var _t=ae,At=he;ae=he=n}var Jt=Je?n:qp(x),yn=[x,C,Z,ae,he,_t,At,Ce,Pe,qe];if(Jt&&Yg(yn,Jt),x=yn[0],C=yn[1],Z=yn[2],ae=yn[3],he=yn[4],qe=yn[9]=yn[9]===n?Je?0:x.length:Bi(yn[9]-St,0),!qe&&C&($|k)&&(C&=-25),!C||C==S)var Un=Xy(x,C,Z);else C==$||C==k?Un=S4(x,C,qe):(C==A||C==(S|A))&&!he.length?Un=Z1(x,C,Z,ae):Un=wd.apply(n,yn);var Er=Jt?Ny:Gg;return Hs(Er(Un,yn),x,C)}function Hg(x,C,Z,ae){return x===n||Ws(x,vo[Z])&&!Sr.call(ae,Z)?C:x}function Ky(x,C,Z,ae,he,Ce){return ba(x)&&ba(C)&&(Ce.set(C,x),om(x,C,n,Ky,Ce),Ce.delete(C)),x}function Xg(x){return ab(x)?n:x}function pm(x,C,Z,ae,he,Ce){var Pe=Z&y,qe=x.length,Je=C.length;if(qe!=Je&&!(Pe&&Je>qe))return!1;var St=Ce.get(x),_t=Ce.get(C);if(St&&_t)return St==C&&_t==x;var At=-1,Jt=!0,yn=Z&v?new Wo:n;for(Ce.set(x,C),Ce.set(C,x);++At<qe;){var Un=x[At],Er=C[At];if(ae)var jn=Pe?ae(Er,Un,At,C,x,Ce):ae(Un,Er,At,x,C,Ce);if(jn!==n){if(jn)continue;Jt=!1;break}if(yn){if(!wp(C,function(jr,ni){if(!Ga(yn,ni)&&(Un===jr||he(Un,jr,Z,ae,Ce)))return yn.push(ni)})){Jt=!1;break}}else if(!(Un===Er||he(Un,Er,Z,ae,Ce))){Jt=!1;break}}return Ce.delete(x),Ce.delete(C),Jt}function X1(x,C,Z,ae,he,Ce,Pe){switch(Z){case at:if(x.byteLength!=C.byteLength||x.byteOffset!=C.byteOffset)return!1;x=x.buffer,C=C.buffer;case ze:return!(x.byteLength!=C.byteLength||!Ce(new Ju(x),new Ju(C)));case ge:case ye:case we:return Ws(+x,+C);case Ee:return x.name==C.name&&x.message==C.message;case Oe:case Ae:return x==C+"";case ve:var qe=yh;case $e:var Je=ae&y;if(qe||(qe=Kn),x.size!=C.size&&!Je)return!1;var St=Pe.get(x);if(St)return St==C;ae|=v,Pe.set(x,C);var _t=pm(qe(x),qe(C),ae,he,Ce,Pe);return Pe.delete(x),_t;case He:if(Ri)return Ri.call(x)==Ri.call(C)}return!1}function Wy(x,C,Z,ae,he,Ce){var Pe=Z&y,qe=ol(x),Je=qe.length,St=ol(C),_t=St.length;if(Je!=_t&&!Pe)return!1;for(var At=Je;At--;){var Jt=qe[At];if(!(Pe?Jt in C:Sr.call(C,Jt)))return!1}var yn=Ce.get(x),Un=Ce.get(C);if(yn&&Un)return yn==C&&Un==x;var Er=!0;Ce.set(x,C),Ce.set(C,x);for(var jn=Pe;++At<Je;){Jt=qe[At];var jr=x[Jt],ni=C[Jt];if(ae)var Mf=Pe?ae(ni,jr,Jt,C,x,Ce):ae(jr,ni,Jt,x,C,Ce);if(!(Mf===n?jr===ni||he(jr,ni,Z,ae,Ce):Mf)){Er=!1;break}jn||(jn=Jt=="constructor")}if(Er&&!jn){var tc=x.constructor,Ff=C.constructor;tc!=Ff&&"constructor"in x&&"constructor"in C&&!(typeof tc=="function"&&tc instanceof tc&&typeof Ff=="function"&&Ff instanceof Ff)&&(Er=!1)}return Ce.delete(x),Ce.delete(C),Er}function Gl(x){return wf(Qc(x,n,Fr),x+"")}function ol(x){return Ag(x,pl,us)}function Dl(x){return Ag(x,hs,_o)}var qp=Qe?function(x){return Qe.get(x)}:O;function xi(x){for(var C=x.name+"",Z=Be[C],ae=Sr.call(Be,C)?Z.length:0;ae--;){var he=Z[ae],Ce=he.func;if(Ce==null||Ce==x)return he.name}return C}function Ai(x){var C=Sr.call(Se,"placeholder")?Se:x;return C.placeholder}function On(){var x=Se.iteratee||hb;return x=x===hb?Lp:x,arguments.length?x(arguments[0],arguments[1]):x}function Td(x,C){var Z=x.__data__;return Vp(C)?Z[typeof C=="string"?"string":"hash"]:Z.map}function $l(x){for(var C=pl(x),Z=C.length;Z--;){var ae=C[Z],he=x[ae];C[Z]=[ae,he,os(he)]}return C}function iu(x,C){var Z=J0(x,C);return lm(Z)?Z:n}function js(x){var C=Sr.call(x,Ji),Z=x[Ji];try{x[Ji]=n;var ae=!0}catch{}var he=pa.call(x);return ae&&(C?x[Ji]=Z:delete x[Ji]),he}var us=bu?function(x){return x==null?[]:(x=zr(x),Zo(bu(x),function(C){return Rc.call(x,C)}))}:xt,_o=bu?function(x){for(var C=[];x;)Ho(C,us(x)),x=el(x);return C}:xt,qr=Qu;(ue&&qr(new ue(new ArrayBuffer(1)))!=at||L&&qr(new L)!=ve||re&&qr(re.resolve())!=et||fe&&qr(new fe)!=$e||xe&&qr(new xe)!=Ye)&&(qr=function(x){var C=Qu(x),Z=C==Re?x.constructor:n,ae=Z?uu(Z):"";if(ae)switch(ae){case Ze:return at;case Rt:return ve;case Kt:return et;case zn:return $e;case ln:return Ye}return C});function mm(x,C,Z){for(var ae=-1,he=Z.length;++ae<he;){var Ce=Z[ae],Pe=Ce.size;switch(Ce.type){case"drop":x+=Pe;break;case"dropRight":C-=Pe;break;case"take":C=wi(C,x+Pe);break;case"takeRight":x=Bi(x,C-Pe);break}}return{start:x,end:C}}function Zs(x){var C=x.match(Ct);return C?C[1].split(Vt):[]}function kh(x,C,Z){C=So(C,x);for(var ae=-1,he=C.length,Ce=!1;++ae<he;){var Pe=au(C[ae]);if(!(Ce=x!=null&&Z(x,Pe)))break;x=x[Pe]}return Ce||++ae!=he?Ce:(he=x==null?0:x.length,!!he&&tO(he)&&Do(Pe,he)&&(Wn(x)||$m(x)))}function sl(x){var C=x.length,Z=new x.constructor(C);return C&&typeof x[0]=="string"&&Sr.call(x,"index")&&(Z.index=x.index,Z.input=x.input),Z}function Eu(x){return typeof x.constructor=="function"&&!$f(x)?El(el(x)):{}}function Vr(x,C,Z){var ae=x.constructor;switch(C){case ze:return fm(x);case ge:case ye:return new ae(+x);case at:return V1(x,Z);case mt:case Mt:case bt:case zt:case qt:case Xt:case nn:case Fe:case ht:return Yl(x,Z);case ve:return new ae;case we:case Ae:return new ae(x);case Oe:return U1(x);case $e:return new ae;case He:return j1(x)}}function Eo(x,C){var Z=C.length;if(!Z)return x;var ae=Z-1;return C[ae]=(Z>1?"& ":"")+C[ae],C=C.join(Z>2?", ":" "),x.replace(gt,`{
/* [wrapped with `+C+`] */
`)}function ls(x){return Wn(x)||$m(x)||!!(gd&&x&&x[gd])}function Do(x,C){var Z=typeof x;return C=C??te,!!C&&(Z=="number"||Z!="symbol"&&$i.test(x))&&x>-1&&x%1==0&&x<C}function Ti(x,C,Z){if(!ba(Z))return!1;var ae=typeof C;return(ae=="number"?fs(Z)&&Do(C,Z.length):ae=="string"&&C in Z)?Ws(Z[C],x):!1}function Bc(x,C){if(Wn(x))return!1;var Z=typeof x;return Z=="number"||Z=="symbol"||Z=="boolean"||x==null||ds(x)?!0:ji.test(x)||!Gr.test(x)||C!=null&&x in zr(C)}function Vp(x){var C=typeof x;return C=="string"||C=="number"||C=="symbol"||C=="boolean"?x!=="__proto__":x===null}function kd(x){var C=xi(x),Z=Se[C];if(typeof Z!="function"||!(C in qn.prototype))return!1;if(x===Z)return!0;var ae=qp(Z);return!!ae&&x===ae[0]}function Ph(x){return!!bh&&bh in x}var gm=yo?Js:ot;function $f(x){var C=x&&x.constructor,Z=typeof C=="function"&&C.prototype||vo;return x===Z}function os(x){return x===x&&!ba(x)}function Cf(x,C){return function(Z){return Z==null?!1:Z[x]===C&&(C!==n||x in zr(Z))}}function vm(x){var C=tb(x,function(ae){return Z.size===s&&Z.clear(),ae}),Z=C.cache;return C}function Yg(x,C){var Z=x[1],ae=C[1],he=Z|ae,Ce=he<(S|_|R),Pe=ae==R&&Z==$||ae==R&&Z==F&&x[7].length<=C[8]||ae==(R|F)&&C[7].length<=C[8]&&Z==$;if(!(Ce||Pe))return x;ae&S&&(x[2]=C[2],he|=Z&S?0:E);var qe=C[3];if(qe){var Je=x[3];x[3]=Je?dm(Je,qe,C[4]):qe,x[4]=Je?go(x[3],c):C[4]}return qe=C[5],qe&&(Je=x[5],x[5]=Je?Ug(Je,qe,C[6]):qe,x[6]=Je?go(x[5],c):C[6]),qe=C[7],qe&&(x[7]=qe),ae&R&&(x[8]=x[8]==null?C[8]:wi(x[8],C[8])),x[9]==null&&(x[9]=C[9]),x[0]=C[0],x[1]=he,x}function Kl(x){var C=[];if(x!=null)for(var Z in zr(x))C.push(Z);return C}function Pd(x){return pa.call(x)}function Qc(x,C,Z){return C=Bi(C===n?x.length-1:C,0),function(){for(var ae=arguments,he=-1,Ce=Bi(ae.length-C,0),Pe=rt(Ce);++he<Ce;)Pe[he]=ae[C+he];he=-1;for(var qe=rt(C+1);++he<C;)qe[he]=ae[he];return qe[C]=Z(Pe),ha(x,this,qe)}}function Mh(x,C){return C.length<2?x:Ra(x,Hl(C,0,-1))}function ym(x,C){for(var Z=x.length,ae=wi(C.length,Z),he=ul(x);ae--;){var Ce=C[ae];x[ae]=Do(Ce,Z)?he[Ce]:n}return x}function Wr(x,C){if(!(C==="constructor"&&typeof x[C]=="function")&&C!="__proto__")return x[C]}var Gg=$o(Ny),Md=zl||function(x,C){return ia.setTimeout(x,C)},wf=$o(Iy);function Hs(x,C,Z){var ae=C+"";return wf(x,Eo(ae,Co(Zs(ae),Z)))}function $o(x){var C=0,Z=0;return function(){var ae=Fu(),he=N-(ae-Z);if(Z=ae,he>0){if(++C>=V)return arguments[0]}else C=0;return x.apply(n,arguments)}}function Xs(x,C){var Z=-1,ae=x.length,he=ae-1;for(C=C===n?ae:C;++Z<C;){var Ce=Qp(Z,he),Pe=x[Ce];x[Ce]=x[Z],x[Z]=Pe}return x.length=C,x}var Kg=vm(function(x){var C=[];return x.charCodeAt(0)===46&&C.push(""),x.replace(ai,function(Z,ae,he,Ce){C.push(he?Ce.replace(Lt,"$1"):ae||Z)}),C});function au(x){if(typeof x=="string"||ds(x))return x;var C=x+"";return C=="0"&&1/x==-J?"-0":C}function uu(x){if(x!=null){try{return _f.call(x)}catch{}try{return x+""}catch{}}return""}function Co(x,C){return Ba(ne,function(Z){var ae="_."+Z[0];C&Z[1]&&!ph(x,ae)&&x.push(ae)}),x.sort()}function wo(x){if(x instanceof qn)return x.clone();var C=new Iu(x.__wrapped__,x.__chain__);return C.__actions__=ul(x.__actions__),C.__index__=x.__index__,C.__values__=x.__values__,C}function qa(x,C,Z){(Z?Ti(x,C,Z):C===n)?C=1:C=Bi(Vn(C),0);var ae=x==null?0:x.length;if(!ae||C<1)return[];for(var he=0,Ce=0,Pe=rt(Ef(ae/C));he<ae;)Pe[Ce++]=Hl(x,he,he+=C);return Pe}function zc(x){for(var C=-1,Z=x==null?0:x.length,ae=0,he=[];++C<Z;){var Ce=x[C];Ce&&(he[ae++]=Ce)}return he}function nr(){var x=arguments.length;if(!x)return[];for(var C=rt(x-1),Z=arguments[0],ae=x;ae--;)C[ae-1]=arguments[ae];return Ho(Wn(Z)?ul(Z):[Z],eu(C,1))}var Jr=Jn(function(x,C){return ou(x)?ts(x,eu(C,1,ou,!0)):[]}),Ta=Jn(function(x,C){var Z=Ur(C);return ou(Z)&&(Z=n),ou(x)?ts(x,eu(C,1,ou,!0),On(Z,2)):[]}),ss=Jn(function(x,C){var Z=Ur(C);return ou(Z)&&(Z=n),ou(x)?ts(x,eu(C,1,ou,!0),n,Z):[]});function Xi(x,C,Z){var ae=x==null?0:x.length;return ae?(C=Z||C===n?1:Vn(C),Hl(x,C<0?0:C,ae)):[]}function Rf(x,C,Z){var ae=x==null?0:x.length;return ae?(C=Z||C===n?1:Vn(C),C=ae-C,Hl(x,0,C<0?0:C)):[]}function bm(x,C){return x&&x.length?sm(x,On(C,3),!0,!0):[]}function va(x,C){return x&&x.length?sm(x,On(C,3),!0):[]}function Up(x,C,Z,ae){var he=x==null?0:x.length;return he?(Z&&typeof Z!="number"&&Ti(x,C,Z)&&(Z=0,ae=he),Sd(x,C,Z,ae)):[]}function Jy(x,C,Z){var ae=x==null?0:x.length;if(!ae)return-1;var he=Z==null?0:Vn(Z);return he<0&&(he=Bi(ae+he,0)),po(x,On(C,3),he)}function Om(x,C,Z){var ae=x==null?0:x.length;if(!ae)return-1;var he=ae-1;return Z!==n&&(he=Vn(Z),he=Z<0?Bi(ae+he,0):wi(he,ae-1)),po(x,On(C,3),he,!0)}function Fr(x){var C=x==null?0:x.length;return C?eu(x,1):[]}function Va(x){var C=x==null?0:x.length;return C?eu(x,J):[]}function cl(x,C){var Z=x==null?0:x.length;return Z?(C=C===n?1:Vn(C),eu(x,C)):[]}function Fd(x){for(var C=-1,Z=x==null?0:x.length,ae={};++C<Z;){var he=x[C];ae[he[0]]=he[1]}return ae}function xm(x){return x&&x.length?x[0]:n}function eb(x,C,Z){var ae=x==null?0:x.length;if(!ae)return-1;var he=Z==null?0:Vn(Z);return he<0&&(he=Bi(ae+he,0)),Ku(x,C,he)}function Y1(x){var C=x==null?0:x.length;return C?Hl(x,0,-1):[]}var ka=Jn(function(x){var C=bi(x,Vy);return C.length&&C[0]===x[0]?um(C):[]}),fl=Jn(function(x){var C=Ur(x),Z=bi(x,Vy);return C===Ur(Z)?C=n:Z.pop(),Z.length&&Z[0]===x[0]?um(Z,On(C,2)):[]}),Wl=Jn(function(x){var C=Ur(x),Z=bi(x,Vy);return C=typeof C=="function"?C:n,C&&Z.pop(),Z.length&&Z[0]===x[0]?um(Z,n,C):[]});function Nd(x,C){return x==null?"":ql.call(x,C)}function Ur(x){var C=x==null?0:x.length;return C?x[C-1]:n}function G1(x,C,Z){var ae=x==null?0:x.length;if(!ae)return-1;var he=ae;return Z!==n&&(he=Vn(Z),he=he<0?Bi(ae+he,0):wi(he,ae-1)),C===C?Ii(x,C,he):po(x,mh,he,!0)}function Af(x,C){return x&&x.length?Fg(x,Vn(C)):n}var Fh=Jn(ei);function ei(x,C){return x&&x.length&&C&&C.length?$h(x,C):x}function Wg(x,C,Z){return x&&x.length&&C&&C.length?$h(x,C,On(Z,2)):x}function Tf(x,C,Z){return x&&x.length&&C&&C.length?$h(x,C,n,Z):x}var lu=Gl(function(x,C){var Z=x==null?0:x.length,ae=Np(x,C);return Bp(x,bi(C,function(he){return Do(he,Z)?+he:he}).sort(qg)),ae});function ua(x,C){var Z=[];if(!(x&&x.length))return Z;var ae=-1,he=[],Ce=x.length;for(C=On(C,3);++ae<Ce;){var Pe=x[ae];C(Pe,ae,x)&&(Z.push(Pe),he.push(ae))}return Bp(x,he),Z}function qc(x){return x==null?x:Ko.call(x)}function jp(x,C,Z){var ae=x==null?0:x.length;return ae?(Z&&typeof Z!="number"&&Ti(x,C,Z)?(C=0,Z=ae):(C=C==null?0:Vn(C),Z=Z===n?ae:Vn(Z)),Hl(x,C,Z)):[]}function ya(x,C){return Dd(x,C)}function Nh(x,C,Z){return Bg(x,C,On(Z,2))}function Pa(x,C){var Z=x==null?0:x.length;if(Z){var ae=Dd(x,C);if(ae<Z&&Ws(x[ae],C))return ae}return-1}function Jl(x,C){return Dd(x,C,!0)}function Cl(x,C,Z){return Bg(x,C,On(Z,2),!0)}function Ih(x,C){var Z=x==null?0:x.length;if(Z){var ae=Dd(x,C,!0)-1;if(Ws(x[ae],C))return ae}return-1}function Jg(x){return x&&x.length?z1(x):[]}function Sm(x,C){return x&&x.length?z1(x,On(C,2)):[]}function ev(x){var C=x==null?0:x.length;return C?Hl(x,1,C):[]}function tv(x,C,Z){return x&&x.length?(C=Z||C===n?1:Vn(C),Hl(x,0,C<0?0:C)):[]}function Nr(x,C,Z){var ae=x==null?0:x.length;return ae?(C=Z||C===n?1:Vn(C),C=ae-C,Hl(x,C<0?0:C,ae)):[]}function li(x,C){return x&&x.length?sm(x,On(C,3),!1,!0):[]}function rr(x,C){return x&&x.length?sm(x,On(C,3)):[]}var _r=Jn(function(x){return Nc(eu(x,1,ou,!0))}),vn=Jn(function(x){var C=Ur(x);return ou(C)&&(C=n),Nc(eu(x,1,ou,!0),On(C,2))}),mr=Jn(function(x){var C=Ur(x);return C=typeof C=="function"?C:n,Nc(eu(x,1,ou,!0),n,C)});function Yn(x){return x&&x.length?Nc(x):[]}function cr(x,C){return x&&x.length?Nc(x,On(C,2)):[]}function Si(x,C){return C=typeof C=="function"?C:n,x&&x.length?Nc(x,n,C):[]}function la(x){if(!(x&&x.length))return[];var C=0;return x=Zo(x,function(Z){if(ou(Z))return C=Bi(Z.length,C),!0}),fd(C,function(Z){return bi(x,Xo(Z))})}function ti(x,C){if(!(x&&x.length))return[];var Z=la(x);return C==null?Z:bi(Z,function(ae){return ha(C,n,ae)})}var Ro=Jn(function(x,C){return ou(x)?ts(x,C):[]}),kf=Jn(function(x){return Qg(Zo(x,ou))}),Ys=Jn(function(x){var C=Ur(x);return ou(C)&&(C=n),Qg(Zo(x,ou),On(C,2))}),Ir=Jn(function(x){var C=Ur(x);return C=typeof C=="function"?C:n,Qg(Zo(x,ou),n,C)}),Ao=Jn(la);function Gs(x,C){return qy(x||[],C||[],zs)}function dl(x,C){return qy(x||[],C||[],Ch)}var eo=Jn(function(x){var C=x.length,Z=C>1?x[C-1]:n;return Z=typeof Z=="function"?(x.pop(),Z):n,ti(x,Z)});function Vc(x){var C=Se(x);return C.__chain__=!0,C}function Zp(x,C){return C(x),x}function hl(x,C){return C(x)}var Hp=Gl(function(x){var C=x.length,Z=C?x[0]:0,ae=this.__wrapped__,he=function(Ce){return Np(Ce,x)};return C>1||this.__actions__.length||!(ae instanceof qn)||!Do(Z)?this.thru(he):(ae=ae.slice(Z,+Z+(C?1:0)),ae.__actions__.push({func:hl,args:[he],thisArg:n}),new Iu(ae,this.__chain__).thru(function(Ce){return C&&!Ce.length&&Ce.push(n),Ce}))});function Uc(){return Vc(this)}function oa(){return new Iu(this.value(),this.__chain__)}function Ks(){this.__values__===n&&(this.__values__=l_(this.value()));var x=this.__index__>=this.__values__.length,C=x?n:this.__values__[this.__index__++];return{done:x,value:C}}function To(){return this}function cs(x){for(var C,Z=this;Z instanceof Vl;){var ae=wo(Z);ae.__index__=0,ae.__values__=n,C?he.__wrapped__=ae:C=ae;var he=ae;Z=Z.__wrapped__}return he.__wrapped__=x,C}function Pf(){var x=this.__wrapped__;if(x instanceof qn){var C=x;return this.__actions__.length&&(C=new qn(this)),C=C.reverse(),C.__actions__.push({func:hl,args:[qc],thisArg:n}),new Iu(C,this.__chain__)}return this.thru(qc)}function Lh(){return zy(this.__wrapped__,this.__actions__)}var nv=jg(function(x,C,Z){Sr.call(x,Z)?++x[Z]:nl(x,Z,1)});function Bh(x,C,Z){var ae=Wn(x)?Dp:Mc;return Z&&Ti(x,C,Z)&&(C=n),ae(x,On(C,3))}function Qh(x,C){var Z=Wn(x)?Zo:am;return Z(x,On(C,3))}var wl=Yy(Jy),g=Yy(Om);function b(x,C){return eu(Pt(x,C),1)}function T(x,C){return eu(Pt(x,C),J)}function z(x,C,Z){return Z=Z===n?1:Vn(Z),eu(Pt(x,C),Z)}function ee(x,C){var Z=Wn(x)?Ba:Pc;return Z(x,On(C,3))}function le(x,C){var Z=Wn(x)?Eg:Rg;return Z(x,On(C,3))}var De=jg(function(x,C,Z){Sr.call(x,Z)?x[Z].push(C):nl(x,Z,[C])});function Ie(x,C,Z,ae){x=fs(x)?x:sv(x),Z=Z&&!ae?Vn(Z):0;var he=x.length;return Z<0&&(Z=Bi(he+Z,0)),ob(x)?Z<=he&&x.indexOf(C,Z)>-1:!!he&&Ku(x,C,Z)>-1}var We=Jn(function(x,C,Z){var ae=-1,he=typeof C=="function",Ce=fs(x)?rt(x.length):[];return Pc(x,function(Pe){Ce[++ae]=he?ha(C,Pe,Z):xo(Pe,C,Z)}),Ce}),vt=jg(function(x,C,Z){nl(x,Z,C)});function Pt(x,C){var Z=Wn(x)?bi:Pg;return Z(x,On(C,3))}function jt(x,C,Z,ae){return x==null?[]:(Wn(C)||(C=C==null?[]:[C]),Z=ae?n:Z,Wn(Z)||(Z=Z==null?[]:[Z]),Ng(x,C,Z))}var Wt=jg(function(x,C,Z){x[Z?0:1].push(C)},function(){return[[],[]]});function Hn(x,C,Z){var ae=Wn(x)?$p:Qa,he=arguments.length<3;return ae(x,On(C,4),Z,he,Pc)}function to(x,C,Z){var ae=Wn(x)?Cp:Qa,he=arguments.length<3;return ae(x,On(C,4),Z,he,Rg)}function _m(x,C){var Z=Wn(x)?Zo:am;return Z(x,K1(On(C,3)))}function zh(x){var C=Wn(x)?es:Fy;return C(x)}function lt(x,C,Z){(Z?Ti(x,C,Z):C===n)?C=1:C=Vn(C);var ae=Wn(x)?rm:Ig;return ae(x,C)}function Ge(x){var C=Wn(x)?B1:Ly;return C(x)}function ct(x){if(x==null)return 0;if(fs(x))return ob(x)?gu(x):x.length;var C=qr(x);return C==ve||C==$e?x.size:is(x).length}function Tt(x,C,Z){var ae=Wn(x)?wp:Lg;return Z&&Ti(x,C,Z)&&(C=n),ae(x,On(C,3))}var pn=Jn(function(x,C){if(x==null)return[];var Z=C.length;return Z>1&&Ti(x,C[0],C[1])?C=[]:Z>2&&Ti(C[0],C[1],C[2])&&(C=[C[0]]),Ng(x,eu(C,1),[])}),pi=Oh||function(){return ia.Date.now()};function Tn(x,C){if(typeof C!="function")throw new Ci(u);return x=Vn(x),function(){if(--x<1)return C.apply(this,arguments)}}function Lr(x,C,Z){return C=Z?n:C,C=x&&C==null?x.length:C,Hi(x,R,n,n,n,n,C)}function Ua(x,C){var Z;if(typeof C!="function")throw new Ci(u);return x=Vn(x),function(){return--x>0&&(Z=C.apply(this,arguments)),x<=1&&(C=n),Z}}var Tr=Jn(function(x,C,Z){var ae=S;if(Z.length){var he=go(Z,Ai(Tr));ae|=A}return Hi(x,ae,C,Z,he)}),jc=Jn(function(x,C,Z){var ae=S|_;if(Z.length){var he=go(Z,Ai(jc));ae|=A}return Hi(C,ae,x,Z,he)});function Em(x,C,Z){C=Z?n:C;var ae=Hi(x,$,n,n,n,n,n,C);return ae.placeholder=Em.placeholder,ae}function _4(x,C,Z){C=Z?n:C;var ae=Hi(x,k,n,n,n,n,n,C);return ae.placeholder=_4.placeholder,ae}function E4(x,C,Z){var ae,he,Ce,Pe,qe,Je,St=0,_t=!1,At=!1,Jt=!0;if(typeof x!="function")throw new Ci(u);C=ec(C)||0,ba(Z)&&(_t=!!Z.leading,At="maxWait"in Z,Ce=At?Bi(ec(Z.maxWait)||0,C):Ce,Jt="trailing"in Z?!!Z.trailing:Jt);function yn(zu){var Uh=ae,wm=he;return ae=he=n,St=zu,Pe=x.apply(wm,Uh),Pe}function Un(zu){return St=zu,qe=Md(jr,C),_t?yn(zu):Pe}function Er(zu){var Uh=zu-Je,wm=zu-St,LB=C-Uh;return At?wi(LB,Ce-wm):LB}function jn(zu){var Uh=zu-Je,wm=zu-St;return Je===n||Uh>=C||Uh<0||At&&wm>=Ce}function jr(){var zu=pi();if(jn(zu))return ni(zu);qe=Md(jr,Er(zu))}function ni(zu){return qe=n,Jt&&ae?yn(zu):(ae=he=n,Pe)}function Mf(){qe!==n&&cm(qe),St=0,ae=Je=he=qe=n}function tc(){return qe===n?Pe:ni(pi())}function Ff(){var zu=pi(),Uh=jn(zu);if(ae=arguments,he=this,Je=zu,Uh){if(qe===n)return Un(Je);if(At)return cm(qe),qe=Md(jr,C),yn(Je)}return qe===n&&(qe=Md(jr,C)),Pe}return Ff.cancel=Mf,Ff.flush=tc,Ff}var ki=Jn(function(x,C){return Ty(x,1,C)}),D4=Jn(function(x,C,Z){return Ty(x,ec(C)||0,Z)});function _5(x){return Hi(x,I)}function tb(x,C){if(typeof x!="function"||C!=null&&typeof C!="function")throw new Ci(u);var Z=function(){var ae=arguments,he=C?C.apply(this,ae):ae[0],Ce=Z.cache;if(Ce.has(he))return Ce.get(he);var Pe=x.apply(this,ae);return Z.cache=Ce.set(he,Pe)||Ce,Pe};return Z.cache=new(tb.Cache||Bs),Z}tb.Cache=Bs;function K1(x){if(typeof x!="function")throw new Ci(u);return function(){var C=arguments;switch(C.length){case 0:return!x.call(this);case 1:return!x.call(this,C[0]);case 2:return!x.call(this,C[0],C[1]);case 3:return!x.call(this,C[0],C[1],C[2])}return!x.apply(this,C)}}function E5(x){return Ua(2,x)}var D5=q1(function(x,C){C=C.length==1&&Wn(C[0])?bi(C[0],pu(On())):bi(eu(C,1),pu(On()));var Z=C.length;return Jn(function(ae){for(var he=-1,Ce=wi(ae.length,Z);++he<Ce;)ae[he]=C[he].call(this,ae[he]);return ha(x,this,ae)})}),rv=Jn(function(x,C){var Z=go(C,Ai(rv));return Hi(x,A,n,C,Z)}),Dm=Jn(function(x,C){var Z=go(C,Ai(Dm));return Hi(x,w,n,C,Z)}),$4=Gl(function(x,C){return Hi(x,F,n,n,n,C)});function W1(x,C){if(typeof x!="function")throw new Ci(u);return C=C===n?C:Vn(C),Jn(x,C)}function C4(x,C){if(typeof x!="function")throw new Ci(u);return C=C==null?0:Bi(Vn(C),0),Jn(function(Z){var ae=Z[C],he=Us(Z,0,C);return ae&&Ho(he,ae),ha(x,this,he)})}function Xp(x,C,Z){var ae=!0,he=!0;if(typeof x!="function")throw new Ci(u);return ba(Z)&&(ae="leading"in Z?!!Z.leading:ae,he="trailing"in Z?!!Z.trailing:he),E4(x,C,{leading:ae,maxWait:C,trailing:he})}function Id(x){return Lr(x,1)}function nb(x,C){return rv(zg(C),x)}function $5(){if(!arguments.length)return[];var x=arguments[0];return Wn(x)?x:[x]}function YS(x){return rl(x,m)}function GS(x,C){return C=typeof C=="function"?C:n,rl(x,m,C)}function KS(x){return rl(x,d|m)}function WS(x,C){return C=typeof C=="function"?C:n,rl(x,d|m,C)}function C5(x,C){return C==null||xd(x,C,pl(C))}function Ws(x,C){return x===C||x!==x&&C!==C}var JS=Ad(xu),e_=Ad(function(x,C){return x>=C}),$m=Zl((function(){return arguments})())?Zl:function(x){return ja(x)&&Sr.call(x,"callee")&&!Rc.call(x,"callee")},Wn=rt.isArray,w4=X0?pu(X0):ma;function fs(x){return x!=null&&tO(x.length)&&!Js(x)}function ou(x){return ja(x)&&fs(x)}function rb(x){return x===!0||x===!1||ja(x)&&Qu(x)==ge}var Yp=Pp||ot,t_=Y0?pu(Y0):Tg;function kr(x){return ja(x)&&x.nodeType===1&&!ab(x)}function R4(x){if(x==null)return!0;if(fs(x)&&(Wn(x)||typeof x=="string"||typeof x.splice=="function"||Yp(x)||Gp(x)||$m(x)))return!x.length;var C=qr(x);if(C==ve||C==$e)return!x.size;if($f(x))return!is(x).length;for(var Z in x)if(Sr.call(x,Z))return!1;return!0}function J1(x,C){return Su(x,C)}function A4(x,C,Z){Z=typeof Z=="function"?Z:n;var ae=Z?Z(x,C):n;return ae===n?Su(x,C,n,Z):!!ae}function eO(x){if(!ja(x))return!1;var C=Qu(x);return C==Ee||C==me||typeof x.message=="string"&&typeof x.name=="string"&&!ab(x)}function T4(x){return typeof x=="number"&&Mu(x)}function Js(x){if(!ba(x))return!1;var C=Qu(x);return C==se||C==pe||C==de||C==tt}function ib(x){return typeof x=="number"&&x==Vn(x)}function tO(x){return typeof x=="number"&&x>-1&&x%1==0&&x<=te}function ba(x){var C=typeof x;return x!=null&&(C=="object"||C=="function")}function ja(x){return x!=null&&typeof x=="object"}var n_=od?pu(od):Oi;function r_(x,C){return x===C||ga(x,C,$l(C))}function i_(x,C,Z){return Z=typeof Z=="function"?Z:n,ga(x,C,$l(C),Z)}function Qi(x){return P4(x)&&x!=+x}function k4(x){if(gm(x))throw new En(a);return lm(x)}function Rl(x){return x===null}function w5(x){return x==null}function P4(x){return typeof x=="number"||ja(x)&&Qu(x)==we}function ab(x){if(!ja(x)||Qu(x)!=Re)return!1;var C=el(x);if(C===null)return!0;var Z=Sr.call(C,"constructor")&&C.constructor;return typeof Z=="function"&&Z instanceof Z&&_f.call(Z)==Go}var ub=G0?pu(G0):Fc;function M4(x){return ib(x)&&x>=-te&&x<=te}var lb=_c?pu(_c):Dh;function ob(x){return typeof x=="string"||!Wn(x)&&ja(x)&&Qu(x)==Ae}function ds(x){return typeof x=="symbol"||ja(x)&&Qu(x)==He}var Gp=sd?pu(sd):kg;function F4(x){return x===n}function R5(x){return ja(x)&&qr(x)==Ye}function a_(x){return ja(x)&&Qu(x)==Ne}var A5=Ad(ea),u_=Ad(function(x,C){return x<=C});function l_(x){if(!x)return[];if(fs(x))return ob(x)?pr(x):ul(x);if(aa&&x[aa])return Ar(x[aa]());var C=qr(x),Z=C==ve?yh:C==$e?Kn:sv;return Z(x)}function qh(x){if(!x)return x===0?x:0;if(x=ec(x),x===J||x===-J){var C=x<0?-1:1;return C*H}return x===x?x:0}function Vn(x){var C=qh(x),Z=C%1;return C===C?Z?C-Z:C:0}function N4(x){return x?qs(Vn(x),0,G):0}function ec(x){if(typeof x=="number")return x;if(ds(x))return M;if(ba(x)){var C=typeof x.valueOf=="function"?x.valueOf():x;x=ba(C)?C+"":C}if(typeof x!="string")return x===0?x:+x;x=Ns(x);var Z=Ln.test(x);return Z||Di.test(x)?ad(x.slice(2),Z?2:8):sn.test(x)?M:+x}function nO(x){return as(x,hs(x))}function T5(x){return x?qs(Vn(x),-te,te):x===0?x:0}function _i(x){return x==null?"":Xl(x)}var o_=$d(function(x,C){if($f(C)||fs(C)){as(C,pl(C),x);return}for(var Z in C)Sr.call(C,Z)&&zs(x,Z,C[Z])}),rO=$d(function(x,C){as(C,hs(C),x)}),iv=$d(function(x,C,Z,ae){as(C,hs(C),x,ae)}),k5=$d(function(x,C,Z,ae){as(C,pl(C),x,ae)}),Zc=Gl(Np);function I4(x,C){var Z=El(x);return C==null?Z:wg(Z,C)}var s_=Jn(function(x,C){x=zr(x);var Z=-1,ae=C.length,he=ae>2?C[2]:n;for(he&&Ti(C[0],C[1],he)&&(ae=1);++Z<ae;)for(var Ce=C[Z],Pe=hs(Ce),qe=-1,Je=Pe.length;++qe<Je;){var St=Pe[qe],_t=x[St];(_t===n||Ws(_t,vo[St])&&!Sr.call(x,St))&&(x[St]=Ce[St])}return x}),c_=Jn(function(x){return x.push(n,Ky),ha(uv,n,x)});function f_(x,C){return K0(x,On(C,3),tu)}function sb(x,C){return K0(x,On(C,3),nu)}function Hc(x,C){return x==null?x:Eh(x,On(C,3),hs)}function d_(x,C){return x==null?x:_d(x,On(C,3),hs)}function iO(x,C){return x&&tu(x,On(C,3))}function Vh(x,C){return x&&nu(x,On(C,3))}function P5(x){return x==null?[]:il(x,pl(x))}function M5(x){return x==null?[]:il(x,hs(x))}function Kp(x,C,Z){var ae=x==null?n:Ra(x,C);return ae===n?Z:ae}function h_(x,C){return x!=null&&kh(x,C,ru)}function L4(x,C){return x!=null&&kh(x,C,rs)}var F5=Ic(function(x,C,Z){C!=null&&typeof C.toString!="function"&&(C=pa.call(C)),x[C]=Z},fv(ko)),N5=Ic(function(x,C,Z){C!=null&&typeof C.toString!="function"&&(C=pa.call(C)),Sr.call(x,C)?x[C].push(Z):x[C]=[Z]},On),I5=Jn(xo);function pl(x){return fs(x)?Fp(x):is(x)}function hs(x){return fs(x)?Fp(x,!0):O4(x)}function L5(x,C){var Z={};return C=On(C,3),tu(x,function(ae,he,Ce){nl(Z,C(ae,he,Ce),ae)}),Z}function p_(x,C){var Z={};return C=On(C,3),tu(x,function(ae,he,Ce){nl(Z,he,C(ae,he,Ce))}),Z}var av=$d(function(x,C,Z){om(x,C,Z)}),uv=$d(function(x,C,Z,ae){om(x,C,Z,ae)}),m_=Gl(function(x,C){var Z={};if(x==null)return Z;var ae=!1;C=bi(C,function(Ce){return Ce=So(Ce,x),ae||(ae=Ce.length>1),Ce}),as(x,Dl(x),Z),ae&&(Z=rl(Z,d|h|m,Xg));for(var he=C.length;he--;)zp(Z,C[he]);return Z});function B5(x,C){return ov(x,K1(On(C)))}var lv=Gl(function(x,C){return x==null?{}:Py(x,C)});function ov(x,C){if(x==null)return{};var Z=bi(Dl(x),function(ae){return[ae]});return C=On(C),al(x,Z,function(ae,he){return C(ae,he[0])})}function g_(x,C,Z){C=So(C,x);var ae=-1,he=C.length;for(he||(he=1,x=n);++ae<he;){var Ce=x==null?n:x[au(C[ae])];Ce===n&&(ae=he,Ce=Z),x=Js(Ce)?Ce.call(x):Ce}return x}function aO(x,C,Z){return x==null?x:Ch(x,C,Z)}function B4(x,C,Z,ae){return ae=typeof ae=="function"?ae:n,x==null?x:Ch(x,C,Z,ae)}var uO=Gy(pl),cb=Gy(hs);function v_(x,C,Z){var ae=Wn(x),he=ae||Yp(x)||Gp(x);if(C=On(C,4),Z==null){var Ce=x&&x.constructor;he?Z=ae?new Ce:[]:ba(x)?Z=Js(Ce)?El(el(x)):{}:Z={}}return(he?Ba:tu)(x,function(Pe,qe,Je){return C(Z,Pe,qe,Je)}),Z}function y_(x,C){return x==null?!0:zp(x,C)}function Q5(x,C,Z){return x==null?x:Qy(x,C,zg(Z))}function b_(x,C,Z,ae){return ae=typeof ae=="function"?ae:n,x==null?x:Qy(x,C,zg(Z),ae)}function sv(x){return x==null?[]:Ap(x,pl(x))}function Q4(x){return x==null?[]:Ap(x,hs(x))}function z5(x,C,Z){return Z===n&&(Z=C,C=n),Z!==n&&(Z=ec(Z),Z=Z===Z?Z:0),C!==n&&(C=ec(C),C=C===C?C:0),qs(ec(x),C,Z)}function lO(x,C,Z){return C=qh(C),Z===n?(Z=C,C=0):Z=qh(Z),x=ec(x),Ed(x,C,Z)}function oO(x,C,Z){if(Z&&typeof Z!="boolean"&&Ti(x,C,Z)&&(C=Z=n),Z===n&&(typeof C=="boolean"?(Z=C,C=n):typeof x=="boolean"&&(Z=x,x=n)),x===n&&C===n?(x=0,C=1):(x=qh(x),C===n?(C=x,x=0):C=qh(C)),x>C){var ae=x;x=C,C=ae}if(Z||x%1||C%1){var he=Nu();return wi(x+he*(C-x+Ts("1e-"+((he+"").length-1))),C)}return Qp(x,C)}var sO=Cd(function(x,C,Z){return C=C.toLowerCase(),x+(Z?O_(C):C)});function O_(x){return Cm(_i(x).toLowerCase())}function cv(x){return x=_i(x),x&&x.replace(hr,pd).replace(Sp,"")}function q5(x,C,Z){x=_i(x),C=Xl(C);var ae=x.length;Z=Z===n?ae:qs(Vn(Z),0,ae);var he=Z;return Z-=C.length,Z>=0&&x.slice(Z,he)==C}function x_(x){return x=_i(x),x&&br.test(x)?x.replace(rn,Wu):x}function S_(x){return x=_i(x),x&&Te.test(x)?x.replace(ar,"\\$&"):x}var __=Cd(function(x,C,Z){return x+(Z?"-":"")+C.toLowerCase()}),E_=Cd(function(x,C,Z){return x+(Z?" ":"")+C.toLowerCase()}),z4=_u("toLowerCase");function D_(x,C,Z){x=_i(x),C=Vn(C);var ae=C?gu(x):0;if(!C||ae>=C)return x;var he=(C-ae)/2;return Rd(xh(he),Z)+x+Rd(Ef(he),Z)}function $_(x,C,Z){x=_i(x),C=Vn(C);var ae=C?gu(x):0;return C&&ae<C?x+Rd(C-ae,Z):x}function cO(x,C,Z){x=_i(x),C=Vn(C);var ae=C?gu(x):0;return C&&ae<C?Rd(C-ae,Z)+x:x}function V5(x,C,Z){return Z||C==null?C=0:C&&(C=+C),dn(_i(x).replace(Ve,""),C||0)}function U5(x,C,Z){return(Z?Ti(x,C,Z):C===n)?C=1:C=Vn(C),Vs(_i(x),C)}function q4(){var x=arguments,C=_i(x[0]);return x.length<3?C:C.replace(x[1],x[2])}var V4=Cd(function(x,C,Z){return x+(Z?"_":"")+C.toLowerCase()});function fO(x,C,Z){return Z&&typeof Z!="number"&&Ti(x,C,Z)&&(C=Z=n),Z=Z===n?G:Z>>>0,Z?(x=_i(x),x&&(typeof C=="string"||C!=null&&!ub(C))&&(C=Xl(C),!C&&mu(x))?Us(pr(x),0,Z):x.split(C,Z)):[]}var U4=Cd(function(x,C,Z){return x+(Z?" ":"")+Cm(C)});function C_(x,C,Z){return x=_i(x),Z=Z==null?0:qs(Vn(Z),0,x.length),C=Xl(C),x.slice(Z,Z+C.length)==C}function j4(x,C,Z){var ae=Se.templateSettings;Z&&Ti(x,C,Z)&&(C=n),x=_i(x),C=iv({},C,ae,Hg);var he=iv({},C.imports,ae.imports,Hg),Ce=pl(he),Pe=Ap(he,Ce),qe,Je,St=0,_t=C.interpolate||wt,At="__p += '",Jt=Ka((C.escape||wt).source+"|"+_t.source+"|"+(_t===wr?an:wt).source+"|"+(C.evaluate||wt).source+"|$","g"),yn="//# sourceURL="+(Sr.call(C,"sourceURL")?(C.sourceURL+"").replace(/\s/g," "):"lodash.templateSources["+ ++hh+"]")+`
`;x.replace(Jt,function(jn,jr,ni,Mf,tc,Ff){return ni||(ni=Mf),At+=x.slice(St,Ff).replace(Kr,mo),jr&&(qe=!0,At+=`' +
__e(`+jr+`) +
'`),tc&&(Je=!0,At+=`';
`+tc+`;
__p += '`),ni&&(At+=`' +
((__t = (`+ni+`)) == null ? '' : __t) +
'`),St=Ff+jn.length,jn}),At+=`';
`;var Un=Sr.call(C,"variable")&&C.variable;if(!Un)At=`with (obj) {
`+At+`
}
`;else if(Dt.test(Un))throw new En(l);At=(Je?At.replace(kt,""):At).replace(Qt,"$1").replace(en,"$1;"),At="function("+(Un||"obj")+`) {
`+(Un?"":`obj || (obj = {});
`)+"var __t, __p = ''"+(qe?", __e = _.escape":"")+(Je?`, __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
`:`;
`)+At+`return __p
}`;var Er=Z4(function(){return lr(Ce,yn+"return "+At).apply(n,Pe)});if(Er.source=At,eO(Er))throw Er;return Er}function Wp(x){return _i(x).toLowerCase()}function Jp(x){return _i(x).toUpperCase()}function e0(x,C,Z){if(x=_i(x),x&&(Z||C===n))return Ns(x);if(!x||!(C=Xl(C)))return x;var ae=pr(x),he=pr(C),Ce=Tp(ae,he),Pe=hd(ae,he)+1;return Us(ae,Ce,Pe).join("")}function fb(x,C,Z){if(x=_i(x),x&&(Z||C===n))return x.slice(0,xf(x)+1);if(!x||!(C=Xl(C)))return x;var ae=pr(x),he=hd(ae,pr(C))+1;return Us(ae,0,he).join("")}function db(x,C,Z){if(x=_i(x),x&&(Z||C===n))return x.replace(Ve,"");if(!x||!(C=Xl(C)))return x;var ae=pr(x),he=Tp(ae,pr(C));return Us(ae,he).join("")}function t0(x,C){var Z=U,ae=q;if(ba(C)){var he="separator"in C?C.separator:he;Z="length"in C?Vn(C.length):Z,ae="omission"in C?Xl(C.omission):ae}x=_i(x);var Ce=x.length;if(mu(x)){var Pe=pr(x);Ce=Pe.length}if(Z>=Ce)return x;var qe=Z-gu(ae);if(qe<1)return ae;var Je=Pe?Us(Pe,0,qe).join(""):x.slice(0,qe);if(he===n)return Je+ae;if(Pe&&(qe+=Je.length-qe),ub(he)){if(x.slice(qe).search(he)){var St,_t=Je;for(he.global||(he=Ka(he.source,_i(_n.exec(he))+"g")),he.lastIndex=0;St=he.exec(_t);)var At=St.index;Je=Je.slice(0,At===n?qe:At)}}else if(x.indexOf(Xl(he),qe)!=qe){var Jt=Je.lastIndexOf(he);Jt>-1&&(Je=Je.slice(0,Jt))}return Je+ae}function j5(x){return x=_i(x),x&&Xn.test(x)?x.replace(In,em):x}var w_=Cd(function(x,C,Z){return x+(Z?" ":"")+C.toUpperCase()}),Cm=_u("toUpperCase");function R_(x,C,Z){return x=_i(x),C=Z?n:C,C===n?Of(x)?nm(x):Dy(x):x.match(C)||[]}var Z4=Jn(function(x,C){try{return ha(x,n,C)}catch(Z){return eO(Z)?Z:new En(Z)}}),dO=Gl(function(x,C){return Ba(C,function(Z){Z=au(Z),nl(x,Z,Tr(x[Z],x))}),x});function A_(x){var C=x==null?0:x.length,Z=On();return x=C?bi(x,function(ae){if(typeof ae[1]!="function")throw new Ci(u);return[Z(ae[0]),ae[1]]}):[],Jn(function(ae){for(var he=-1;++he<C;){var Ce=x[he];if(ha(Ce[0],this,ae))return ha(Ce[1],this,ae)}})}function Z5(x){return Ip(rl(x,d))}function fv(x){return function(){return x}}function hO(x,C){return x==null||x!==x?C:x}var T_=wh(),dv=wh(!0);function ko(x){return x}function hb(x){return Lp(typeof x=="function"?x:rl(x,d))}function pO(x){return Mg(rl(x,d))}function k_(x,C){return ky(x,rl(C,d))}var H5=Jn(function(x,C){return function(Z){return xo(Z,x,C)}}),mO=Jn(function(x,C){return function(Z){return xo(x,Z,C)}});function f(x,C,Z){var ae=pl(C),he=il(C,ae);Z==null&&!(ba(C)&&(he.length||!ae.length))&&(Z=C,C=x,x=this,he=il(C,pl(C)));var Ce=!(ba(Z)&&"chain"in Z)||!!Z.chain,Pe=Js(x);return Ba(he,function(qe){var Je=C[qe];x[qe]=Je,Pe&&(x.prototype[qe]=function(){var St=this.__chain__;if(Ce||St){var _t=x(this.__wrapped__),At=_t.__actions__=ul(this.__actions__);return At.push({func:Je,args:arguments,thisArg:x}),_t.__chain__=St,_t}return Je.apply(x,Ho([this.value()],arguments))})}),x}function p(){return ia._===this&&(ia._=wa),this}function O(){}function D(x){return x=Vn(x),Jn(function(C){return Fg(C,x)})}var Q=ll(bi),Y=ll(Dp),ce=ll(wp);function _e(x){return Bc(x)?Xo(au(x)):Aa(x)}function Le(x){return function(C){return x==null?n:Ra(x,C)}}var ut=Rh(),pt=Rh(!0);function xt(){return[]}function ot(){return!1}function ft(){return{}}function on(){return""}function Rn(){return!0}function zi(x,C){if(x=Vn(x),x<1||x>te)return[];var Z=G,ae=wi(x,G);C=On(C),x-=G;for(var he=fd(ae,C);++Z<x;)C(Z);return he}function Ke(x){return Wn(x)?bi(x,au):ds(x)?[x]:ul(Kg(_i(x)))}function Ue(x){var C=++Yo;return _i(x)+C}var it=Lc(function(x,C){return x+C},0),yt=Th("ceil"),Sn=Lc(function(x,C){return x/C},1),oi=Th("floor");function hn(x){return x&&x.length?ns(x,ko,xu):n}function fr(x,C){return x&&x.length?ns(x,On(C,2),xu):n}function Br(x){return Ec(x,ko)}function si(x,C){return Ec(x,On(C,2))}function n0(x){return x&&x.length?ns(x,ko,ea):n}function X5(x,C){return x&&x.length?ns(x,On(C,2),ea):n}var rOe=Lc(function(x,C){return x*C},1),iOe=Th("round"),aOe=Lc(function(x,C){return x-C},0);function uOe(x){return x&&x.length?vh(x,ko):0}function lOe(x,C){return x&&x.length?vh(x,On(C,2)):0}return Se.after=Tn,Se.ary=Lr,Se.assign=o_,Se.assignIn=rO,Se.assignInWith=iv,Se.assignWith=k5,Se.at=Zc,Se.before=Ua,Se.bind=Tr,Se.bindAll=dO,Se.bindKey=jc,Se.castArray=$5,Se.chain=Vc,Se.chunk=qa,Se.compact=zc,Se.concat=nr,Se.cond=A_,Se.conforms=Z5,Se.constant=fv,Se.countBy=nv,Se.create=I4,Se.curry=Em,Se.curryRight=_4,Se.debounce=E4,Se.defaults=s_,Se.defaultsDeep=c_,Se.defer=ki,Se.delay=D4,Se.difference=Jr,Se.differenceBy=Ta,Se.differenceWith=ss,Se.drop=Xi,Se.dropRight=Rf,Se.dropRightWhile=bm,Se.dropWhile=va,Se.fill=Up,Se.filter=Qh,Se.flatMap=b,Se.flatMapDeep=T,Se.flatMapDepth=z,Se.flatten=Fr,Se.flattenDeep=Va,Se.flattenDepth=cl,Se.flip=_5,Se.flow=T_,Se.flowRight=dv,Se.fromPairs=Fd,Se.functions=P5,Se.functionsIn=M5,Se.groupBy=De,Se.initial=Y1,Se.intersection=ka,Se.intersectionBy=fl,Se.intersectionWith=Wl,Se.invert=F5,Se.invertBy=N5,Se.invokeMap=We,Se.iteratee=hb,Se.keyBy=vt,Se.keys=pl,Se.keysIn=hs,Se.map=Pt,Se.mapKeys=L5,Se.mapValues=p_,Se.matches=pO,Se.matchesProperty=k_,Se.memoize=tb,Se.merge=av,Se.mergeWith=uv,Se.method=H5,Se.methodOf=mO,Se.mixin=f,Se.negate=K1,Se.nthArg=D,Se.omit=m_,Se.omitBy=B5,Se.once=E5,Se.orderBy=jt,Se.over=Q,Se.overArgs=D5,Se.overEvery=Y,Se.overSome=ce,Se.partial=rv,Se.partialRight=Dm,Se.partition=Wt,Se.pick=lv,Se.pickBy=ov,Se.property=_e,Se.propertyOf=Le,Se.pull=Fh,Se.pullAll=ei,Se.pullAllBy=Wg,Se.pullAllWith=Tf,Se.pullAt=lu,Se.range=ut,Se.rangeRight=pt,Se.rearg=$4,Se.reject=_m,Se.remove=ua,Se.rest=W1,Se.reverse=qc,Se.sampleSize=lt,Se.set=aO,Se.setWith=B4,Se.shuffle=Ge,Se.slice=jp,Se.sortBy=pn,Se.sortedUniq=Jg,Se.sortedUniqBy=Sm,Se.split=fO,Se.spread=C4,Se.tail=ev,Se.take=tv,Se.takeRight=Nr,Se.takeRightWhile=li,Se.takeWhile=rr,Se.tap=Zp,Se.throttle=Xp,Se.thru=hl,Se.toArray=l_,Se.toPairs=uO,Se.toPairsIn=cb,Se.toPath=Ke,Se.toPlainObject=nO,Se.transform=v_,Se.unary=Id,Se.union=_r,Se.unionBy=vn,Se.unionWith=mr,Se.uniq=Yn,Se.uniqBy=cr,Se.uniqWith=Si,Se.unset=y_,Se.unzip=la,Se.unzipWith=ti,Se.update=Q5,Se.updateWith=b_,Se.values=sv,Se.valuesIn=Q4,Se.without=Ro,Se.words=R_,Se.wrap=nb,Se.xor=kf,Se.xorBy=Ys,Se.xorWith=Ir,Se.zip=Ao,Se.zipObject=Gs,Se.zipObjectDeep=dl,Se.zipWith=eo,Se.entries=uO,Se.entriesIn=cb,Se.extend=rO,Se.extendWith=iv,f(Se,Se),Se.add=it,Se.attempt=Z4,Se.camelCase=sO,Se.capitalize=O_,Se.ceil=yt,Se.clamp=z5,Se.clone=YS,Se.cloneDeep=KS,Se.cloneDeepWith=WS,Se.cloneWith=GS,Se.conformsTo=C5,Se.deburr=cv,Se.defaultTo=hO,Se.divide=Sn,Se.endsWith=q5,Se.eq=Ws,Se.escape=x_,Se.escapeRegExp=S_,Se.every=Bh,Se.find=wl,Se.findIndex=Jy,Se.findKey=f_,Se.findLast=g,Se.findLastIndex=Om,Se.findLastKey=sb,Se.floor=oi,Se.forEach=ee,Se.forEachRight=le,Se.forIn=Hc,Se.forInRight=d_,Se.forOwn=iO,Se.forOwnRight=Vh,Se.get=Kp,Se.gt=JS,Se.gte=e_,Se.has=h_,Se.hasIn=L4,Se.head=xm,Se.identity=ko,Se.includes=Ie,Se.indexOf=eb,Se.inRange=lO,Se.invoke=I5,Se.isArguments=$m,Se.isArray=Wn,Se.isArrayBuffer=w4,Se.isArrayLike=fs,Se.isArrayLikeObject=ou,Se.isBoolean=rb,Se.isBuffer=Yp,Se.isDate=t_,Se.isElement=kr,Se.isEmpty=R4,Se.isEqual=J1,Se.isEqualWith=A4,Se.isError=eO,Se.isFinite=T4,Se.isFunction=Js,Se.isInteger=ib,Se.isLength=tO,Se.isMap=n_,Se.isMatch=r_,Se.isMatchWith=i_,Se.isNaN=Qi,Se.isNative=k4,Se.isNil=w5,Se.isNull=Rl,Se.isNumber=P4,Se.isObject=ba,Se.isObjectLike=ja,Se.isPlainObject=ab,Se.isRegExp=ub,Se.isSafeInteger=M4,Se.isSet=lb,Se.isString=ob,Se.isSymbol=ds,Se.isTypedArray=Gp,Se.isUndefined=F4,Se.isWeakMap=R5,Se.isWeakSet=a_,Se.join=Nd,Se.kebabCase=__,Se.last=Ur,Se.lastIndexOf=G1,Se.lowerCase=E_,Se.lowerFirst=z4,Se.lt=A5,Se.lte=u_,Se.max=hn,Se.maxBy=fr,Se.mean=Br,Se.meanBy=si,Se.min=n0,Se.minBy=X5,Se.stubArray=xt,Se.stubFalse=ot,Se.stubObject=ft,Se.stubString=on,Se.stubTrue=Rn,Se.multiply=rOe,Se.nth=Af,Se.noConflict=p,Se.noop=O,Se.now=pi,Se.pad=D_,Se.padEnd=$_,Se.padStart=cO,Se.parseInt=V5,Se.random=oO,Se.reduce=Hn,Se.reduceRight=to,Se.repeat=U5,Se.replace=q4,Se.result=g_,Se.round=iOe,Se.runInContext=Xe,Se.sample=zh,Se.size=ct,Se.snakeCase=V4,Se.some=Tt,Se.sortedIndex=ya,Se.sortedIndexBy=Nh,Se.sortedIndexOf=Pa,Se.sortedLastIndex=Jl,Se.sortedLastIndexBy=Cl,Se.sortedLastIndexOf=Ih,Se.startCase=U4,Se.startsWith=C_,Se.subtract=aOe,Se.sum=uOe,Se.sumBy=lOe,Se.template=j4,Se.times=zi,Se.toFinite=qh,Se.toInteger=Vn,Se.toLength=N4,Se.toLower=Wp,Se.toNumber=ec,Se.toSafeInteger=T5,Se.toString=_i,Se.toUpper=Jp,Se.trim=e0,Se.trimEnd=fb,Se.trimStart=db,Se.truncate=t0,Se.unescape=j5,Se.uniqueId=Ue,Se.upperCase=w_,Se.upperFirst=Cm,Se.each=ee,Se.eachRight=le,Se.first=xm,f(Se,(function(){var x={};return tu(Se,function(C,Z){Sr.call(Se.prototype,Z)||(x[Z]=C)}),x})(),{chain:!1}),Se.VERSION=r,Ba(["bind","bindKey","curry","curryRight","partial","partialRight"],function(x){Se[x].placeholder=Se}),Ba(["drop","take"],function(x,C){qn.prototype[x]=function(Z){Z=Z===n?1:Bi(Vn(Z),0);var ae=this.__filtered__&&!C?new qn(this):this.clone();return ae.__filtered__?ae.__takeCount__=wi(Z,ae.__takeCount__):ae.__views__.push({size:wi(Z,G),type:x+(ae.__dir__<0?"Right":"")}),ae},qn.prototype[x+"Right"]=function(Z){return this.reverse()[x](Z).reverse()}}),Ba(["filter","map","takeWhile"],function(x,C){var Z=C+1,ae=Z==j||Z==K;qn.prototype[x]=function(he){var Ce=this.clone();return Ce.__iteratees__.push({iteratee:On(he,3),type:Z}),Ce.__filtered__=Ce.__filtered__||ae,Ce}}),Ba(["head","last"],function(x,C){var Z="take"+(C?"Right":"");qn.prototype[x]=function(){return this[Z](1).value()[0]}}),Ba(["initial","tail"],function(x,C){var Z="drop"+(C?"":"Right");qn.prototype[x]=function(){return this.__filtered__?new qn(this):this[Z](1)}}),qn.prototype.compact=function(){return this.filter(ko)},qn.prototype.find=function(x){return this.filter(x).head()},qn.prototype.findLast=function(x){return this.reverse().find(x)},qn.prototype.invokeMap=Jn(function(x,C){return typeof x=="function"?new qn(this):this.map(function(Z){return xo(Z,x,C)})}),qn.prototype.reject=function(x){return this.filter(K1(On(x)))},qn.prototype.slice=function(x,C){x=Vn(x);var Z=this;return Z.__filtered__&&(x>0||C<0)?new qn(Z):(x<0?Z=Z.takeRight(-x):x&&(Z=Z.drop(x)),C!==n&&(C=Vn(C),Z=C<0?Z.dropRight(-C):Z.take(C-x)),Z)},qn.prototype.takeRightWhile=function(x){return this.reverse().takeWhile(x).reverse()},qn.prototype.toArray=function(){return this.take(G)},tu(qn.prototype,function(x,C){var Z=/^(?:filter|find|map|reject)|While$/.test(C),ae=/^(?:head|last)$/.test(C),he=Se[ae?"take"+(C=="last"?"Right":""):C],Ce=ae||/^find/.test(C);he&&(Se.prototype[C]=function(){var Pe=this.__wrapped__,qe=ae?[1]:arguments,Je=Pe instanceof qn,St=qe[0],_t=Je||Wn(Pe),At=function(jr){var ni=he.apply(Se,Ho([jr],qe));return ae&&Jt?ni[0]:ni};_t&&Z&&typeof St=="function"&&St.length!=1&&(Je=_t=!1);var Jt=this.__chain__,yn=!!this.__actions__.length,Un=Ce&&!Jt,Er=Je&&!yn;if(!Ce&&_t){Pe=Er?Pe:new qn(this);var jn=x.apply(Pe,qe);return jn.__actions__.push({func:hl,args:[At],thisArg:n}),new Iu(jn,Jt)}return Un&&Er?x.apply(this,qe):(jn=this.thru(At),Un?ae?jn.value()[0]:jn.value():jn)})}),Ba(["pop","push","shift","sort","splice","unshift"],function(x){var C=md[x],Z=/^(?:push|sort|unshift)$/.test(x)?"tap":"thru",ae=/^(?:pop|shift)$/.test(x);Se.prototype[x]=function(){var he=arguments;if(ae&&!this.__chain__){var Ce=this.value();return C.apply(Wn(Ce)?Ce:[],he)}return this[Z](function(Pe){return C.apply(Wn(Pe)?Pe:[],he)})}}),tu(qn.prototype,function(x,C){var Z=Se[C];if(Z){var ae=Z.name+"";Sr.call(Be,ae)||(Be[ae]=[]),Be[ae].push({name:C,func:Z})}}),Be[wd(n,_).name]=[{name:"wrapper",func:n}],qn.prototype.clone=Ja,qn.prototype.reverse=Ls,qn.prototype.value=Ul,Se.prototype.at=Hp,Se.prototype.chain=Uc,Se.prototype.commit=oa,Se.prototype.next=Ks,Se.prototype.plant=cs,Se.prototype.reverse=Pf,Se.prototype.toJSON=Se.prototype.valueOf=Se.prototype.value=Lh,Se.prototype.first=Se.prototype.head,aa&&(Se.prototype[aa]=To),Se}),Dc=Sf();Bl?((Bl.exports=Dc)._=Dc,ks._=Dc):ia._=Dc}).call(MZe)})(q2,q2.exports)),q2.exports}var Elt=FZe(),NZe=DI();const Dlt=da(NZe),o5=P.createContext(null);o5.displayName="PanelGroupContext";const Du={group:"data-panel-group",groupDirection:"data-panel-group-direction",groupId:"data-panel-group-id",panel:"data-panel",panelCollapsible:"data-panel-collapsible",panelId:"data-panel-id",panelSize:"data-panel-size",resizeHandle:"data-resize-handle",resizeHandleActive:"data-resize-handle-active",resizeHandleEnabled:"data-panel-resize-handle-enabled",resizeHandleId:"data-panel-resize-handle-id",resizeHandleState:"data-resize-handle-state"},uB=10,Wb=P.useLayoutEffect,CK=wD.useId,IZe=typeof CK=="function"?CK:()=>null;let LZe=0;function lB(e=null){const t=IZe(),n=P.useRef(e||t||null);return n.current===null&&(n.current=""+LZe++),e??n.current}function Zye({children:e,className:t="",collapsedSize:n,collapsible:r,defaultSize:i,forwardedRef:a,id:u,maxSize:l,minSize:o,onCollapse:s,onExpand:c,onResize:d,order:h,style:m,tagName:y="div",...v}){const S=P.useContext(o5);if(S===null)throw Error("Panel components must be rendered within a PanelGroup container");const{collapsePanel:_,expandPanel:E,getPanelSize:$,getPanelStyle:k,groupId:A,isPanelCollapsed:w,reevaluatePanelConstraints:R,registerPanel:F,resizePanel:I,unregisterPanel:U}=S,q=lB(u),V=P.useRef({callbacks:{onCollapse:s,onExpand:c,onResize:d},constraints:{collapsedSize:n,collapsible:r,defaultSize:i,maxSize:l,minSize:o},id:q,idIsFromProps:u!==void 0,order:h});P.useRef({didLogMissingDefaultSizeWarning:!1}),Wb(()=>{const{callbacks:j,constraints:W}=V.current,K={...W};V.current.id=q,V.current.idIsFromProps=u!==void 0,V.current.order=h,j.onCollapse=s,j.onExpand=c,j.onResize=d,W.collapsedSize=n,W.collapsible=r,W.defaultSize=i,W.maxSize=l,W.minSize=o,(K.collapsedSize!==W.collapsedSize||K.collapsible!==W.collapsible||K.maxSize!==W.maxSize||K.minSize!==W.minSize)&&R(V.current,K)}),Wb(()=>{const j=V.current;return F(j),()=>{U(j)}},[h,q,F,U]),P.useImperativeHandle(a,()=>({collapse:()=>{_(V.current)},expand:j=>{E(V.current,j)},getId(){return q},getSize(){return $(V.current)},isCollapsed(){return w(V.current)},isExpanded(){return!w(V.current)},resize:j=>{I(V.current,j)}}),[_,E,$,w,q,I]);const N=k(V.current,i);return P.createElement(y,{...v,children:e,className:t,id:q,style:{...N,...m},[Du.groupId]:A,[Du.panel]:"",[Du.panelCollapsible]:r||void 0,[Du.panelId]:q,[Du.panelSize]:parseFloat(""+N.flexGrow).toFixed(1)})}const BZe=P.forwardRef((e,t)=>P.createElement(Zye,{...e,forwardedRef:t}));Zye.displayName="Panel";BZe.displayName="forwardRef(Panel)";let pF=null,_D=-1,kv=null;function QZe(e,t){if(t){const n=(t&Kye)!==0,r=(t&Wye)!==0,i=(t&Jye)!==0,a=(t&ebe)!==0;if(n)return i?"se-resize":a?"ne-resize":"e-resize";if(r)return i?"sw-resize":a?"nw-resize":"w-resize";if(i)return"s-resize";if(a)return"n-resize"}switch(e){case"horizontal":return"ew-resize";case"intersection":return"move";case"vertical":return"ns-resize"}}function zZe(){kv!==null&&(document.head.removeChild(kv),pF=null,kv=null,_D=-1)}function h7(e,t){var n,r;const i=QZe(e,t);if(pF!==i){if(pF=i,kv===null&&(kv=document.createElement("style"),document.head.appendChild(kv)),_D>=0){var a;(a=kv.sheet)===null||a===void 0||a.removeRule(_D)}_D=(n=(r=kv.sheet)===null||r===void 0?void 0:r.insertRule(`*{cursor: ${i} !important;}`))!==null&&n!==void 0?n:-1}}function Hye(e){return e.type==="keydown"}function Xye(e){return e.type.startsWith("pointer")}function Yye(e){return e.type.startsWith("mouse")}function s5(e){if(Xye(e)){if(e.isPrimary)return{x:e.clientX,y:e.clientY}}else if(Yye(e))return{x:e.clientX,y:e.clientY};return{x:1/0,y:1/0}}function qZe(){if(typeof matchMedia=="function")return matchMedia("(pointer:coarse)").matches?"coarse":"fine"}function VZe(e,t,n){return e.x<t.x+t.width&&e.x+e.width>t.x&&e.y<t.y+t.height&&e.y+e.height>t.y}function UZe(e,t){if(e===t)throw new Error("Cannot compare node with itself");const n={a:AK(e),b:AK(t)};let r;for(;n.a.at(-1)===n.b.at(-1);)e=n.a.pop(),t=n.b.pop(),r=e;$r(r,"Stacking order can only be calculated for elements with a common ancestor");const i={a:RK(wK(n.a)),b:RK(wK(n.b))};if(i.a===i.b){const a=r.childNodes,u={a:n.a.at(-1),b:n.b.at(-1)};let l=a.length;for(;l--;){const o=a[l];if(o===u.a)return 1;if(o===u.b)return-1}}return Math.sign(i.a-i.b)}const jZe=/\b(?:position|zIndex|opacity|transform|webkitTransform|mixBlendMode|filter|webkitFilter|isolation)\b/;function ZZe(e){var t;const n=getComputedStyle((t=Gye(e))!==null&&t!==void 0?t:e).display;return n==="flex"||n==="inline-flex"}function HZe(e){const t=getComputedStyle(e);return!!(t.position==="fixed"||t.zIndex!=="auto"&&(t.position!=="static"||ZZe(e))||+t.opacity<1||"transform"in t&&t.transform!=="none"||"webkitTransform"in t&&t.webkitTransform!=="none"||"mixBlendMode"in t&&t.mixBlendMode!=="normal"||"filter"in t&&t.filter!=="none"||"webkitFilter"in t&&t.webkitFilter!=="none"||"isolation"in t&&t.isolation==="isolate"||jZe.test(t.willChange)||t.webkitOverflowScrolling==="touch")}function wK(e){let t=e.length;for(;t--;){const n=e[t];if($r(n,"Missing node"),HZe(n))return n}return null}function RK(e){return e&&Number(getComputedStyle(e).zIndex)||0}function AK(e){const t=[];for(;e;)t.push(e),e=Gye(e);return t}function Gye(e){const{parentNode:t}=e;return t&&t instanceof ShadowRoot?t.host:t}const Kye=1,Wye=2,Jye=4,ebe=8,XZe=qZe()==="coarse";let ap=[],S3=!1,Fb=new Map,c5=new Map;const Wx=new Set;function YZe(e,t,n,r,i){var a;const{ownerDocument:u}=t,l={direction:n,element:t,hitAreaMargins:r,setResizeHandlerState:i},o=(a=Fb.get(u))!==null&&a!==void 0?a:0;return Fb.set(u,o+1),Wx.add(l),aC(),function(){var c;c5.delete(e),Wx.delete(l);const d=(c=Fb.get(u))!==null&&c!==void 0?c:1;if(Fb.set(u,d-1),aC(),d===1&&Fb.delete(u),ap.includes(l)){const h=ap.indexOf(l);h>=0&&ap.splice(h,1),sB(),i("up",!0,null)}}}function GZe(e){const{target:t}=e,{x:n,y:r}=s5(e);S3=!0,oB({target:t,x:n,y:r}),aC(),ap.length>0&&(uC("down",e),e.preventDefault(),tbe(t)||e.stopImmediatePropagation())}function p7(e){const{x:t,y:n}=s5(e);if(S3&&e.buttons===0&&(S3=!1,uC("up",e)),!S3){const{target:r}=e;oB({target:r,x:t,y:n})}uC("move",e),sB(),ap.length>0&&e.preventDefault()}function m7(e){const{target:t}=e,{x:n,y:r}=s5(e);c5.clear(),S3=!1,ap.length>0&&(e.preventDefault(),tbe(t)||e.stopImmediatePropagation()),uC("up",e),oB({target:t,x:n,y:r}),sB(),aC()}function tbe(e){let t=e;for(;t;){if(t.hasAttribute(Du.resizeHandle))return!0;t=t.parentElement}return!1}function oB({target:e,x:t,y:n}){ap.splice(0);let r=null;(e instanceof HTMLElement||e instanceof SVGElement)&&(r=e),Wx.forEach(i=>{const{element:a,hitAreaMargins:u}=i,l=a.getBoundingClientRect(),{bottom:o,left:s,right:c,top:d}=l,h=XZe?u.coarse:u.fine;if(t>=s-h&&t<=c+h&&n>=d-h&&n<=o+h){if(r!==null&&document.contains(r)&&a!==r&&!a.contains(r)&&!r.contains(a)&&UZe(r,a)>0){let y=r,v=!1;for(;y&&!y.contains(a);){if(VZe(y.getBoundingClientRect(),l)){v=!0;break}y=y.parentElement}if(v)return}ap.push(i)}})}function g7(e,t){c5.set(e,t)}function sB(){let e=!1,t=!1;ap.forEach(r=>{const{direction:i}=r;i==="horizontal"?e=!0:t=!0});let n=0;c5.forEach(r=>{n|=r}),e&&t?h7("intersection",n):e?h7("horizontal",n):t?h7("vertical",n):zZe()}let v7=new AbortController;function aC(){v7.abort(),v7=new AbortController;const e={capture:!0,signal:v7.signal};Wx.size&&(S3?(ap.length>0&&Fb.forEach((t,n)=>{const{body:r}=n;t>0&&(r.addEventListener("contextmenu",m7,e),r.addEventListener("pointerleave",p7,e),r.addEventListener("pointermove",p7,e))}),window.addEventListener("pointerup",m7,e),window.addEventListener("pointercancel",m7,e)):Fb.forEach((t,n)=>{const{body:r}=n;t>0&&(r.addEventListener("pointerdown",GZe,e),r.addEventListener("pointermove",p7,e))}))}function uC(e,t){Wx.forEach(n=>{const{setResizeHandlerState:r}=n,i=ap.includes(n);r(e,i,t)})}function KZe(){const[e,t]=P.useState(0);return P.useCallback(()=>t(n=>n+1),[])}function $r(e,t){if(!e)throw console.error(t),Error(t)}function O1(e,t,n=uB){return e.toFixed(n)===t.toFixed(n)?0:e>t?1:-1}function Gm(e,t,n=uB){return O1(e,t,n)===0}function Bf(e,t,n){return O1(e,t,n)===0}function WZe(e,t,n){if(e.length!==t.length)return!1;for(let r=0;r<e.length;r++){const i=e[r],a=t[r];if(!Bf(i,a,n))return!1}return!0}function WO({panelConstraints:e,panelIndex:t,size:n}){const r=e[t];$r(r!=null,`Panel constraints not found for index ${t}`);let{collapsedSize:i=0,collapsible:a,maxSize:u=100,minSize:l=0}=r;if(O1(n,l)<0)if(a){const o=(i+l)/2;O1(n,o)<0?n=i:n=l}else n=l;return n=Math.min(u,n),n=parseFloat(n.toFixed(uB)),n}function V2({delta:e,initialLayout:t,panelConstraints:n,pivotIndices:r,prevLayout:i,trigger:a}){if(Bf(e,0))return t;const u=[...t],[l,o]=r;$r(l!=null,"Invalid first pivot index"),$r(o!=null,"Invalid second pivot index");let s=0;if(a==="keyboard"){{const d=e<0?o:l,h=n[d];$r(h,`Panel constraints not found for index ${d}`);const{collapsedSize:m=0,collapsible:y,minSize:v=0}=h;if(y){const S=t[d];if($r(S!=null,`Previous layout not found for panel index ${d}`),Bf(S,m)){const _=v-S;O1(_,Math.abs(e))>0&&(e=e<0?0-_:_)}}}{const d=e<0?l:o,h=n[d];$r(h,`No panel constraints found for index ${d}`);const{collapsedSize:m=0,collapsible:y,minSize:v=0}=h;if(y){const S=t[d];if($r(S!=null,`Previous layout not found for panel index ${d}`),Bf(S,v)){const _=S-m;O1(_,Math.abs(e))>0&&(e=e<0?0-_:_)}}}}{const d=e<0?1:-1;let h=e<0?o:l,m=0;for(;;){const v=t[h];$r(v!=null,`Previous layout not found for panel index ${h}`);const _=WO({panelConstraints:n,panelIndex:h,size:100})-v;if(m+=_,h+=d,h<0||h>=n.length)break}const y=Math.min(Math.abs(e),Math.abs(m));e=e<0?0-y:y}{let h=e<0?l:o;for(;h>=0&&h<n.length;){const m=Math.abs(e)-Math.abs(s),y=t[h];$r(y!=null,`Previous layout not found for panel index ${h}`);const v=y-m,S=WO({panelConstraints:n,panelIndex:h,size:v});if(!Bf(y,S)&&(s+=y-S,u[h]=S,s.toPrecision(3).localeCompare(Math.abs(e).toPrecision(3),void 0,{numeric:!0})>=0))break;e<0?h--:h++}}if(WZe(i,u))return i;{const d=e<0?o:l,h=t[d];$r(h!=null,`Previous layout not found for panel index ${d}`);const m=h+s,y=WO({panelConstraints:n,panelIndex:d,size:m});if(u[d]=y,!Bf(y,m)){let v=m-y,_=e<0?o:l;for(;_>=0&&_<n.length;){const E=u[_];$r(E!=null,`Previous layout not found for panel index ${_}`);const $=E+v,k=WO({panelConstraints:n,panelIndex:_,size:$});if(Bf(E,k)||(v-=k-E,u[_]=k),Bf(v,0))break;e>0?_--:_++}}}const c=u.reduce((d,h)=>h+d,0);return Bf(c,100)?u:i}function JZe({layout:e,panelsArray:t,pivotIndices:n}){let r=0,i=100,a=0,u=0;const l=n[0];$r(l!=null,"No pivot index found"),t.forEach((d,h)=>{const{constraints:m}=d,{maxSize:y=100,minSize:v=0}=m;h===l?(r=v,i=y):(a+=v,u+=y)});const o=Math.min(i,100-a),s=Math.max(r,100-u),c=e[l];return{valueMax:o,valueMin:s,valueNow:c}}function Jx(e,t=document){return Array.from(t.querySelectorAll(`[${Du.resizeHandleId}][data-panel-group-id="${e}"]`))}function nbe(e,t,n=document){const i=Jx(e,n).findIndex(a=>a.getAttribute(Du.resizeHandleId)===t);return i??null}function rbe(e,t,n){const r=nbe(e,t,n);return r!=null?[r,r+1]:[-1,-1]}function ibe(e,t=document){var n;if(t instanceof HTMLElement&&(t==null||(n=t.dataset)===null||n===void 0?void 0:n.panelGroupId)==e)return t;const r=t.querySelector(`[data-panel-group][data-panel-group-id="${e}"]`);return r||null}function f5(e,t=document){const n=t.querySelector(`[${Du.resizeHandleId}="${e}"]`);return n||null}function eHe(e,t,n,r=document){var i,a,u,l;const o=f5(t,r),s=Jx(e,r),c=o?s.indexOf(o):-1,d=(i=(a=n[c])===null||a===void 0?void 0:a.id)!==null&&i!==void 0?i:null,h=(u=(l=n[c+1])===null||l===void 0?void 0:l.id)!==null&&u!==void 0?u:null;return[d,h]}function tHe({committedValuesRef:e,eagerValuesRef:t,groupId:n,layout:r,panelDataArray:i,panelGroupElement:a,setLayout:u}){P.useRef({didWarnAboutMissingResizeHandle:!1}),Wb(()=>{if(!a)return;const l=Jx(n,a);for(let o=0;o<i.length-1;o++){const{valueMax:s,valueMin:c,valueNow:d}=JZe({layout:r,panelsArray:i,pivotIndices:[o,o+1]}),h=l[o];if(h!=null){const m=i[o];$r(m,`No panel data found for index "${o}"`),h.setAttribute("aria-controls",m.id),h.setAttribute("aria-valuemax",""+Math.round(s)),h.setAttribute("aria-valuemin",""+Math.round(c)),h.setAttribute("aria-valuenow",d!=null?""+Math.round(d):"")}}return()=>{l.forEach((o,s)=>{o.removeAttribute("aria-controls"),o.removeAttribute("aria-valuemax"),o.removeAttribute("aria-valuemin"),o.removeAttribute("aria-valuenow")})}},[n,r,i,a]),P.useEffect(()=>{if(!a)return;const l=t.current;$r(l,"Eager values not found");const{panelDataArray:o}=l,s=ibe(n,a);$r(s!=null,`No group found for id "${n}"`);const c=Jx(n,a);$r(c,`No resize handles found for group id "${n}"`);const d=c.map(h=>{const m=h.getAttribute(Du.resizeHandleId);$r(m,"Resize handle element has no handle id attribute");const[y,v]=eHe(n,m,o,a);if(y==null||v==null)return()=>{};const S=_=>{if(!_.defaultPrevented)switch(_.key){case"Enter":{_.preventDefault();const E=o.findIndex($=>$.id===y);if(E>=0){const $=o[E];$r($,`No panel data found for index ${E}`);const k=r[E],{collapsedSize:A=0,collapsible:w,minSize:R=0}=$.constraints;if(k!=null&&w){const F=V2({delta:Bf(k,A)?R-A:A-k,initialLayout:r,panelConstraints:o.map(I=>I.constraints),pivotIndices:rbe(n,m,a),prevLayout:r,trigger:"keyboard"});r!==F&&u(F)}}break}}};return h.addEventListener("keydown",S),()=>{h.removeEventListener("keydown",S)}});return()=>{d.forEach(h=>h())}},[a,e,t,n,r,i,u])}function TK(e,t){if(e.length!==t.length)return!1;for(let n=0;n<e.length;n++)if(e[n]!==t[n])return!1;return!0}function abe(e,t){const n=e==="horizontal",{x:r,y:i}=s5(t);return n?r:i}function nHe(e,t,n,r,i){const a=n==="horizontal",u=f5(t,i);$r(u,`No resize handle element found for id "${t}"`);const l=u.getAttribute(Du.groupId);$r(l,"Resize handle element has no group id attribute");let{initialCursorPosition:o}=r;const s=abe(n,e),c=ibe(l,i);$r(c,`No group element found for id "${l}"`);const d=c.getBoundingClientRect(),h=a?d.width:d.height;return(s-o)/h*100}function rHe(e,t,n,r,i,a){if(Hye(e)){const u=n==="horizontal";let l=0;e.shiftKey?l=100:i!=null?l=i:l=10;let o=0;switch(e.key){case"ArrowDown":o=u?0:l;break;case"ArrowLeft":o=u?-l:0;break;case"ArrowRight":o=u?l:0;break;case"ArrowUp":o=u?0:-l;break;case"End":o=100;break;case"Home":o=-100;break}return o}else return r==null?0:nHe(e,t,n,r,a)}function iHe({panelDataArray:e}){const t=Array(e.length),n=e.map(a=>a.constraints);let r=0,i=100;for(let a=0;a<e.length;a++){const u=n[a];$r(u,`Panel constraints not found for index ${a}`);const{defaultSize:l}=u;l!=null&&(r++,t[a]=l,i-=l)}for(let a=0;a<e.length;a++){const u=n[a];$r(u,`Panel constraints not found for index ${a}`);const{defaultSize:l}=u;if(l!=null)continue;const o=e.length-r,s=i/o;r++,t[a]=s,i-=s}return t}function RO(e,t,n){t.forEach((r,i)=>{const a=e[i];$r(a,`Panel data not found for index ${i}`);const{callbacks:u,constraints:l,id:o}=a,{collapsedSize:s=0,collapsible:c}=l,d=n[o];if(d==null||r!==d){n[o]=r;const{onCollapse:h,onExpand:m,onResize:y}=u;y&&y(r,d),c&&(h||m)&&(m&&(d==null||Gm(d,s))&&!Gm(r,s)&&m(),h&&(d==null||!Gm(d,s))&&Gm(r,s)&&h())}})}function ME(e,t){if(e.length!==t.length)return!1;for(let n=0;n<e.length;n++)if(e[n]!=t[n])return!1;return!0}function aHe({defaultSize:e,dragState:t,layout:n,panelData:r,panelIndex:i,precision:a=3}){const u=n[i];let l;return u==null?l=e!=null?e.toPrecision(a):"1":r.length===1?l="1":l=u.toPrecision(a),{flexBasis:0,flexGrow:l,flexShrink:1,overflow:"hidden",pointerEvents:t!==null?"none":void 0}}function uHe(e,t=10){let n=null;return(...i)=>{n!==null&&clearTimeout(n),n=setTimeout(()=>{e(...i)},t)}}function kK(e){try{if(typeof localStorage<"u")e.getItem=t=>localStorage.getItem(t),e.setItem=(t,n)=>{localStorage.setItem(t,n)};else throw new Error("localStorage not supported in this environment")}catch(t){console.error(t),e.getItem=()=>null,e.setItem=()=>{}}}function ube(e){return`react-resizable-panels:${e}`}function lbe(e){return e.map(t=>{const{constraints:n,id:r,idIsFromProps:i,order:a}=t;return i?r:a?`${a}:${JSON.stringify(n)}`:JSON.stringify(n)}).sort((t,n)=>t.localeCompare(n)).join(",")}function obe(e,t){try{const n=ube(e),r=t.getItem(n);if(r){const i=JSON.parse(r);if(typeof i=="object"&&i!=null)return i}}catch{}return null}function lHe(e,t,n){var r,i;const a=(r=obe(e,n))!==null&&r!==void 0?r:{},u=lbe(t);return(i=a[u])!==null&&i!==void 0?i:null}function oHe(e,t,n,r,i){var a;const u=ube(e),l=lbe(t),o=(a=obe(e,i))!==null&&a!==void 0?a:{};o[l]={expandToSizes:Object.fromEntries(n.entries()),layout:r};try{i.setItem(u,JSON.stringify(o))}catch(s){console.error(s)}}function PK({layout:e,panelConstraints:t}){const n=[...e],r=n.reduce((a,u)=>a+u,0);if(n.length!==t.length)throw Error(`Invalid ${t.length} panel layout: ${n.map(a=>`${a}%`).join(", ")}`);if(!Bf(r,100)&&n.length>0)for(let a=0;a<t.length;a++){const u=n[a];$r(u!=null,`No layout data found for index ${a}`);const l=100/r*u;n[a]=l}let i=0;for(let a=0;a<t.length;a++){const u=n[a];$r(u!=null,`No layout data found for index ${a}`);const l=WO({panelConstraints:t,panelIndex:a,size:u});u!=l&&(i+=u-l,n[a]=l)}if(!Bf(i,0))for(let a=0;a<t.length;a++){const u=n[a];$r(u!=null,`No layout data found for index ${a}`);const l=u+i,o=WO({panelConstraints:t,panelIndex:a,size:l});if(u!==o&&(i-=o-u,n[a]=o,Bf(i,0)))break}return n}const sHe=100,U2={getItem:e=>(kK(U2),U2.getItem(e)),setItem:(e,t)=>{kK(U2),U2.setItem(e,t)}},MK={};function sbe({autoSaveId:e=null,children:t,className:n="",direction:r,forwardedRef:i,id:a=null,onLayout:u=null,keyboardResizeBy:l=null,storage:o=U2,style:s,tagName:c="div",...d}){const h=lB(a),m=P.useRef(null),[y,v]=P.useState(null),[S,_]=P.useState([]),E=KZe(),$=P.useRef({}),k=P.useRef(new Map),A=P.useRef(0),w=P.useRef({autoSaveId:e,direction:r,dragState:y,id:h,keyboardResizeBy:l,onLayout:u,storage:o}),R=P.useRef({layout:S,panelDataArray:[],panelDataArrayChanged:!1});P.useRef({didLogIdAndOrderWarning:!1,didLogPanelConstraintsWarning:!1,prevPanelIds:[]}),P.useImperativeHandle(i,()=>({getId:()=>w.current.id,getLayout:()=>{const{layout:B}=R.current;return B},setLayout:B=>{const{onLayout:ne}=w.current,{layout:oe,panelDataArray:ie}=R.current,de=PK({layout:B,panelConstraints:ie.map(ge=>ge.constraints)});TK(oe,de)||(_(de),R.current.layout=de,ne&&ne(de),RO(ie,de,$.current))}}),[]),Wb(()=>{w.current.autoSaveId=e,w.current.direction=r,w.current.dragState=y,w.current.id=h,w.current.onLayout=u,w.current.storage=o}),tHe({committedValuesRef:w,eagerValuesRef:R,groupId:h,layout:S,panelDataArray:R.current.panelDataArray,setLayout:_,panelGroupElement:m.current}),P.useEffect(()=>{const{panelDataArray:B}=R.current;if(e){if(S.length===0||S.length!==B.length)return;let ne=MK[e];ne==null&&(ne=uHe(oHe,sHe),MK[e]=ne);const oe=[...B],ie=new Map(k.current);ne(e,oe,ie,S,o)}},[e,S,o]),P.useEffect(()=>{});const F=P.useCallback(B=>{const{onLayout:ne}=w.current,{layout:oe,panelDataArray:ie}=R.current;if(B.constraints.collapsible){const de=ie.map(Ee=>Ee.constraints),{collapsedSize:ge=0,panelSize:ye,pivotIndices:me}=_b(ie,B,oe);if($r(ye!=null,`Panel size not found for panel "${B.id}"`),!Gm(ye,ge)){k.current.set(B.id,ye);const se=zO(ie,B)===ie.length-1?ye-ge:ge-ye,pe=V2({delta:se,initialLayout:oe,panelConstraints:de,pivotIndices:me,prevLayout:oe,trigger:"imperative-api"});ME(oe,pe)||(_(pe),R.current.layout=pe,ne&&ne(pe),RO(ie,pe,$.current))}}},[]),I=P.useCallback((B,ne)=>{const{onLayout:oe}=w.current,{layout:ie,panelDataArray:de}=R.current;if(B.constraints.collapsible){const ge=de.map(ve=>ve.constraints),{collapsedSize:ye=0,panelSize:me=0,minSize:Ee=0,pivotIndices:se}=_b(de,B,ie),pe=ne??Ee;if(Gm(me,ye)){const ve=k.current.get(B.id),we=ve!=null&&ve>=pe?ve:pe,Re=zO(de,B)===de.length-1?me-we:we-me,et=V2({delta:Re,initialLayout:ie,panelConstraints:ge,pivotIndices:se,prevLayout:ie,trigger:"imperative-api"});ME(ie,et)||(_(et),R.current.layout=et,oe&&oe(et),RO(de,et,$.current))}}},[]),U=P.useCallback(B=>{const{layout:ne,panelDataArray:oe}=R.current,{panelSize:ie}=_b(oe,B,ne);return $r(ie!=null,`Panel size not found for panel "${B.id}"`),ie},[]),q=P.useCallback((B,ne)=>{const{panelDataArray:oe}=R.current,ie=zO(oe,B);return aHe({defaultSize:ne,dragState:y,layout:S,panelData:oe,panelIndex:ie})},[y,S]),V=P.useCallback(B=>{const{layout:ne,panelDataArray:oe}=R.current,{collapsedSize:ie=0,collapsible:de,panelSize:ge}=_b(oe,B,ne);return $r(ge!=null,`Panel size not found for panel "${B.id}"`),de===!0&&Gm(ge,ie)},[]),N=P.useCallback(B=>{const{layout:ne,panelDataArray:oe}=R.current,{collapsedSize:ie=0,collapsible:de,panelSize:ge}=_b(oe,B,ne);return $r(ge!=null,`Panel size not found for panel "${B.id}"`),!de||O1(ge,ie)>0},[]),j=P.useCallback(B=>{const{panelDataArray:ne}=R.current;ne.push(B),ne.sort((oe,ie)=>{const de=oe.order,ge=ie.order;return de==null&&ge==null?0:de==null?-1:ge==null?1:de-ge}),R.current.panelDataArrayChanged=!0,E()},[E]);Wb(()=>{if(R.current.panelDataArrayChanged){R.current.panelDataArrayChanged=!1;const{autoSaveId:B,onLayout:ne,storage:oe}=w.current,{layout:ie,panelDataArray:de}=R.current;let ge=null;if(B){const me=lHe(B,de,oe);me&&(k.current=new Map(Object.entries(me.expandToSizes)),ge=me.layout)}ge==null&&(ge=iHe({panelDataArray:de}));const ye=PK({layout:ge,panelConstraints:de.map(me=>me.constraints)});TK(ie,ye)||(_(ye),R.current.layout=ye,ne&&ne(ye),RO(de,ye,$.current))}}),Wb(()=>{const B=R.current;return()=>{B.layout=[]}},[]);const W=P.useCallback(B=>{let ne=!1;const oe=m.current;return oe&&window.getComputedStyle(oe,null).getPropertyValue("direction")==="rtl"&&(ne=!0),function(de){de.preventDefault();const ge=m.current;if(!ge)return()=>null;const{direction:ye,dragState:me,id:Ee,keyboardResizeBy:se,onLayout:pe}=w.current,{layout:ve,panelDataArray:we}=R.current,{initialLayout:Me}=me??{},Re=rbe(Ee,B,ge);let et=rHe(de,B,ye,me,se,ge);const tt=ye==="horizontal";tt&&ne&&(et=-et);const Oe=we.map(He=>He.constraints),$e=V2({delta:et,initialLayout:Me??ve,panelConstraints:Oe,pivotIndices:Re,prevLayout:ve,trigger:Hye(de)?"keyboard":"mouse-or-touch"}),Ae=!ME(ve,$e);(Xye(de)||Yye(de))&&A.current!=et&&(A.current=et,!Ae&&et!==0?tt?g7(B,et<0?Kye:Wye):g7(B,et<0?Jye:ebe):g7(B,0)),Ae&&(_($e),R.current.layout=$e,pe&&pe($e),RO(we,$e,$.current))}},[]),K=P.useCallback((B,ne)=>{const{onLayout:oe}=w.current,{layout:ie,panelDataArray:de}=R.current,ge=de.map(ve=>ve.constraints),{panelSize:ye,pivotIndices:me}=_b(de,B,ie);$r(ye!=null,`Panel size not found for panel "${B.id}"`);const se=zO(de,B)===de.length-1?ye-ne:ne-ye,pe=V2({delta:se,initialLayout:ie,panelConstraints:ge,pivotIndices:me,prevLayout:ie,trigger:"imperative-api"});ME(ie,pe)||(_(pe),R.current.layout=pe,oe&&oe(pe),RO(de,pe,$.current))},[]),J=P.useCallback((B,ne)=>{const{layout:oe,panelDataArray:ie}=R.current,{collapsedSize:de=0,collapsible:ge}=ne,{collapsedSize:ye=0,collapsible:me,maxSize:Ee=100,minSize:se=0}=B.constraints,{panelSize:pe}=_b(ie,B,oe);pe!=null&&(ge&&me&&Gm(pe,de)?Gm(de,ye)||K(B,ye):pe<se?K(B,se):pe>Ee&&K(B,Ee))},[K]),te=P.useCallback((B,ne)=>{const{direction:oe}=w.current,{layout:ie}=R.current;if(!m.current)return;const de=f5(B,m.current);$r(de,`Drag handle element not found for id "${B}"`);const ge=abe(oe,ne);v({dragHandleId:B,dragHandleRect:de.getBoundingClientRect(),initialCursorPosition:ge,initialLayout:ie})},[]),H=P.useCallback(()=>{v(null)},[]),M=P.useCallback(B=>{const{panelDataArray:ne}=R.current,oe=zO(ne,B);oe>=0&&(ne.splice(oe,1),delete $.current[B.id],R.current.panelDataArrayChanged=!0,E())},[E]),G=P.useMemo(()=>({collapsePanel:F,direction:r,dragState:y,expandPanel:I,getPanelSize:U,getPanelStyle:q,groupId:h,isPanelCollapsed:V,isPanelExpanded:N,reevaluatePanelConstraints:J,registerPanel:j,registerResizeHandle:W,resizePanel:K,startDragging:te,stopDragging:H,unregisterPanel:M,panelGroupElement:m.current}),[F,y,r,I,U,q,h,V,N,J,j,W,K,te,H,M]),X={display:"flex",flexDirection:r==="horizontal"?"row":"column",height:"100%",overflow:"hidden",width:"100%"};return P.createElement(o5.Provider,{value:G},P.createElement(c,{...d,children:t,className:n,id:a,ref:m,style:{...X,...s},[Du.group]:"",[Du.groupDirection]:r,[Du.groupId]:h}))}const cHe=P.forwardRef((e,t)=>P.createElement(sbe,{...e,forwardedRef:t}));sbe.displayName="PanelGroup";cHe.displayName="forwardRef(PanelGroup)";function zO(e,t){return e.findIndex(n=>n===t||n.id===t.id)}function _b(e,t,n){const r=zO(e,t),a=r===e.length-1?[r-1,r]:[r,r+1],u=n[r];return{...t.constraints,panelSize:u,pivotIndices:a}}function fHe({disabled:e,handleId:t,resizeHandler:n,panelGroupElement:r}){P.useEffect(()=>{if(e||n==null||r==null)return;const i=f5(t,r);if(i==null)return;const a=u=>{if(!u.defaultPrevented)switch(u.key){case"ArrowDown":case"ArrowLeft":case"ArrowRight":case"ArrowUp":case"End":case"Home":{u.preventDefault(),n(u);break}case"F6":{u.preventDefault();const l=i.getAttribute(Du.groupId);$r(l,`No group element found for id "${l}"`);const o=Jx(l,r),s=nbe(l,t,r);$r(s!==null,`No resize element found for id "${t}"`);const c=u.shiftKey?s>0?s-1:o.length-1:s+1<o.length?s+1:0;o[c].focus();break}}};return i.addEventListener("keydown",a),()=>{i.removeEventListener("keydown",a)}},[r,e,t,n])}function dHe({children:e=null,className:t="",disabled:n=!1,hitAreaMargins:r,id:i,onBlur:a,onClick:u,onDragging:l,onFocus:o,onPointerDown:s,onPointerUp:c,style:d={},tabIndex:h=0,tagName:m="div",...y}){var v,S;const _=P.useRef(null),E=P.useRef({onClick:u,onDragging:l,onPointerDown:s,onPointerUp:c});P.useEffect(()=>{E.current.onClick=u,E.current.onDragging=l,E.current.onPointerDown=s,E.current.onPointerUp=c});const $=P.useContext(o5);if($===null)throw Error("PanelResizeHandle components must be rendered within a PanelGroup container");const{direction:k,groupId:A,registerResizeHandle:w,startDragging:R,stopDragging:F,panelGroupElement:I}=$,U=lB(i),[q,V]=P.useState("inactive"),[N,j]=P.useState(!1),[W,K]=P.useState(null),J=P.useRef({state:q});Wb(()=>{J.current.state=q}),P.useEffect(()=>{if(n)K(null);else{const G=w(U);K(()=>G)}},[n,U,w]);const te=(v=r==null?void 0:r.coarse)!==null&&v!==void 0?v:15,H=(S=r==null?void 0:r.fine)!==null&&S!==void 0?S:5;P.useEffect(()=>{if(n||W==null)return;const G=_.current;$r(G,"Element ref not attached");let X=!1;return YZe(U,G,k,{coarse:te,fine:H},(ne,oe,ie)=>{if(!oe){V("inactive");return}switch(ne){case"down":{V("drag"),X=!1,$r(ie,'Expected event to be defined for "down" action'),R(U,ie);const{onDragging:de,onPointerDown:ge}=E.current;de==null||de(!0),ge==null||ge();break}case"move":{const{state:de}=J.current;X=!0,de!=="drag"&&V("hover"),$r(ie,'Expected event to be defined for "move" action'),W(ie);break}case"up":{V("hover"),F();const{onClick:de,onDragging:ge,onPointerUp:ye}=E.current;ge==null||ge(!1),ye==null||ye(),X||de==null||de();break}}})},[te,k,n,H,w,U,W,R,F]),fHe({disabled:n,handleId:U,resizeHandler:W,panelGroupElement:I});const M={touchAction:"none",userSelect:"none"};return P.createElement(m,{...y,children:e,className:t,id:i,onBlur:()=>{j(!1),a==null||a()},onFocus:()=>{j(!0),o==null||o()},ref:_,role:"separator",style:{...M,...d},tabIndex:h,[Du.groupDirection]:k,[Du.groupId]:A,[Du.resizeHandle]:"",[Du.resizeHandleActive]:q==="drag"?"pointer":N?"keyboard":void 0,[Du.resizeHandleEnabled]:!n,[Du.resizeHandleId]:U,[Du.resizeHandleState]:q})}dHe.displayName="PanelResizeHandle";function AO(e,t,n){let r=n.initialDeps??[],i;function a(){var u,l,o,s;let c;n.key&&((u=n.debug)!=null&&u.call(n))&&(c=Date.now());const d=e();if(!(d.length!==r.length||d.some((y,v)=>r[v]!==y)))return i;r=d;let m;if(n.key&&((l=n.debug)!=null&&l.call(n))&&(m=Date.now()),i=t(...d),n.key&&((o=n.debug)!=null&&o.call(n))){const y=Math.round((Date.now()-c)*100)/100,v=Math.round((Date.now()-m)*100)/100,S=v/16,_=(E,$)=>{for(E=String(E);E.length<$;)E=" "+E;return E};console.info(`%c⏱ ${_(v,5)} /${_(y,5)} ms`,`
            font-size: .6rem;
            font-weight: bold;
            color: hsl(${Math.max(0,Math.min(120-120*S,120))}deg 100% 31%);`,n==null?void 0:n.key)}return(s=n==null?void 0:n.onChange)==null||s.call(n,i),i}return a.updateDeps=u=>{r=u},a}function FK(e,t){if(e===void 0)throw new Error("Unexpected undefined");return e}const hHe=(e,t)=>Math.abs(e-t)<1.01,pHe=(e,t,n)=>{let r;return function(...i){e.clearTimeout(r),r=e.setTimeout(()=>t.apply(this,i),n)}},NK=e=>{const{offsetWidth:t,offsetHeight:n}=e;return{width:t,height:n}},mHe=e=>e,gHe=e=>{const t=Math.max(e.startIndex-e.overscan,0),n=Math.min(e.endIndex+e.overscan,e.count-1),r=[];for(let i=t;i<=n;i++)r.push(i);return r},vHe=(e,t)=>{const n=e.scrollElement;if(!n)return;const r=e.targetWindow;if(!r)return;const i=u=>{const{width:l,height:o}=u;t({width:Math.round(l),height:Math.round(o)})};if(i(NK(n)),!r.ResizeObserver)return()=>{};const a=new r.ResizeObserver(u=>{const l=()=>{const o=u[0];if(o!=null&&o.borderBoxSize){const s=o.borderBoxSize[0];if(s){i({width:s.inlineSize,height:s.blockSize});return}}i(NK(n))};e.options.useAnimationFrameWithResizeObserver?requestAnimationFrame(l):l()});return a.observe(n,{box:"border-box"}),()=>{a.unobserve(n)}},IK={passive:!0},LK=typeof window>"u"?!0:"onscrollend"in window,yHe=(e,t)=>{const n=e.scrollElement;if(!n)return;const r=e.targetWindow;if(!r)return;let i=0;const a=e.options.useScrollendEvent&&LK?()=>{}:pHe(r,()=>{t(i,!1)},e.options.isScrollingResetDelay),u=c=>()=>{const{horizontal:d,isRtl:h}=e.options;i=d?n.scrollLeft*(h&&-1||1):n.scrollTop,a(),t(i,c)},l=u(!0),o=u(!1);o(),n.addEventListener("scroll",l,IK);const s=e.options.useScrollendEvent&&LK;return s&&n.addEventListener("scrollend",o,IK),()=>{n.removeEventListener("scroll",l),s&&n.removeEventListener("scrollend",o)}},bHe=(e,t,n)=>{if(t!=null&&t.borderBoxSize){const r=t.borderBoxSize[0];if(r)return Math.round(r[n.options.horizontal?"inlineSize":"blockSize"])}return e[n.options.horizontal?"offsetWidth":"offsetHeight"]},OHe=(e,{adjustments:t=0,behavior:n},r)=>{var i,a;const u=e+t;(a=(i=r.scrollElement)==null?void 0:i.scrollTo)==null||a.call(i,{[r.options.horizontal?"left":"top"]:u,behavior:n})};class xHe{constructor(t){this.unsubs=[],this.scrollElement=null,this.targetWindow=null,this.isScrolling=!1,this.measurementsCache=[],this.itemSizeCache=new Map,this.pendingMeasuredCacheIndexes=[],this.scrollRect=null,this.scrollOffset=null,this.scrollDirection=null,this.scrollAdjustments=0,this.elementsCache=new Map,this.observer=(()=>{let n=null;const r=()=>n||(!this.targetWindow||!this.targetWindow.ResizeObserver?null:n=new this.targetWindow.ResizeObserver(i=>{i.forEach(a=>{const u=()=>{this._measureElement(a.target,a)};this.options.useAnimationFrameWithResizeObserver?requestAnimationFrame(u):u()})}));return{disconnect:()=>{var i;(i=r())==null||i.disconnect(),n=null},observe:i=>{var a;return(a=r())==null?void 0:a.observe(i,{box:"border-box"})},unobserve:i=>{var a;return(a=r())==null?void 0:a.unobserve(i)}}})(),this.range=null,this.setOptions=n=>{Object.entries(n).forEach(([r,i])=>{typeof i>"u"&&delete n[r]}),this.options={debug:!1,initialOffset:0,overscan:1,paddingStart:0,paddingEnd:0,scrollPaddingStart:0,scrollPaddingEnd:0,horizontal:!1,getItemKey:mHe,rangeExtractor:gHe,onChange:()=>{},measureElement:bHe,initialRect:{width:0,height:0},scrollMargin:0,gap:0,indexAttribute:"data-index",initialMeasurementsCache:[],lanes:1,isScrollingResetDelay:150,enabled:!0,isRtl:!1,useScrollendEvent:!1,useAnimationFrameWithResizeObserver:!1,...n}},this.notify=n=>{var r,i;(i=(r=this.options).onChange)==null||i.call(r,this,n)},this.maybeNotify=AO(()=>(this.calculateRange(),[this.isScrolling,this.range?this.range.startIndex:null,this.range?this.range.endIndex:null]),n=>{this.notify(n)},{key:!1,debug:()=>this.options.debug,initialDeps:[this.isScrolling,this.range?this.range.startIndex:null,this.range?this.range.endIndex:null]}),this.cleanup=()=>{this.unsubs.filter(Boolean).forEach(n=>n()),this.unsubs=[],this.observer.disconnect(),this.scrollElement=null,this.targetWindow=null},this._didMount=()=>()=>{this.cleanup()},this._willUpdate=()=>{var n;const r=this.options.enabled?this.options.getScrollElement():null;if(this.scrollElement!==r){if(this.cleanup(),!r){this.maybeNotify();return}this.scrollElement=r,this.scrollElement&&"ownerDocument"in this.scrollElement?this.targetWindow=this.scrollElement.ownerDocument.defaultView:this.targetWindow=((n=this.scrollElement)==null?void 0:n.window)??null,this.elementsCache.forEach(i=>{this.observer.observe(i)}),this._scrollToOffset(this.getScrollOffset(),{adjustments:void 0,behavior:void 0}),this.unsubs.push(this.options.observeElementRect(this,i=>{this.scrollRect=i,this.maybeNotify()})),this.unsubs.push(this.options.observeElementOffset(this,(i,a)=>{this.scrollAdjustments=0,this.scrollDirection=a?this.getScrollOffset()<i?"forward":"backward":null,this.scrollOffset=i,this.isScrolling=a,this.maybeNotify()}))}},this.getSize=()=>this.options.enabled?(this.scrollRect=this.scrollRect??this.options.initialRect,this.scrollRect[this.options.horizontal?"width":"height"]):(this.scrollRect=null,0),this.getScrollOffset=()=>this.options.enabled?(this.scrollOffset=this.scrollOffset??(typeof this.options.initialOffset=="function"?this.options.initialOffset():this.options.initialOffset),this.scrollOffset):(this.scrollOffset=null,0),this.getFurthestMeasurement=(n,r)=>{const i=new Map,a=new Map;for(let u=r-1;u>=0;u--){const l=n[u];if(i.has(l.lane))continue;const o=a.get(l.lane);if(o==null||l.end>o.end?a.set(l.lane,l):l.end<o.end&&i.set(l.lane,!0),i.size===this.options.lanes)break}return a.size===this.options.lanes?Array.from(a.values()).sort((u,l)=>u.end===l.end?u.index-l.index:u.end-l.end)[0]:void 0},this.getMeasurementOptions=AO(()=>[this.options.count,this.options.paddingStart,this.options.scrollMargin,this.options.getItemKey,this.options.enabled],(n,r,i,a,u)=>(this.pendingMeasuredCacheIndexes=[],{count:n,paddingStart:r,scrollMargin:i,getItemKey:a,enabled:u}),{key:!1}),this.getMeasurements=AO(()=>[this.getMeasurementOptions(),this.itemSizeCache],({count:n,paddingStart:r,scrollMargin:i,getItemKey:a,enabled:u},l)=>{if(!u)return this.measurementsCache=[],this.itemSizeCache.clear(),[];this.measurementsCache.length===0&&(this.measurementsCache=this.options.initialMeasurementsCache,this.measurementsCache.forEach(c=>{this.itemSizeCache.set(c.key,c.size)}));const o=this.pendingMeasuredCacheIndexes.length>0?Math.min(...this.pendingMeasuredCacheIndexes):0;this.pendingMeasuredCacheIndexes=[];const s=this.measurementsCache.slice(0,o);for(let c=o;c<n;c++){const d=a(c),h=this.options.lanes===1?s[c-1]:this.getFurthestMeasurement(s,c),m=h?h.end+this.options.gap:r+i,y=l.get(d),v=typeof y=="number"?y:this.options.estimateSize(c),S=m+v,_=h?h.lane:c%this.options.lanes;s[c]={index:c,start:m,size:v,end:S,key:d,lane:_}}return this.measurementsCache=s,s},{key:!1,debug:()=>this.options.debug}),this.calculateRange=AO(()=>[this.getMeasurements(),this.getSize(),this.getScrollOffset(),this.options.lanes],(n,r,i,a)=>this.range=n.length>0&&r>0?SHe({measurements:n,outerSize:r,scrollOffset:i,lanes:a}):null,{key:!1,debug:()=>this.options.debug}),this.getVirtualIndexes=AO(()=>{let n=null,r=null;const i=this.calculateRange();return i&&(n=i.startIndex,r=i.endIndex),this.maybeNotify.updateDeps([this.isScrolling,n,r]),[this.options.rangeExtractor,this.options.overscan,this.options.count,n,r]},(n,r,i,a,u)=>a===null||u===null?[]:n({startIndex:a,endIndex:u,overscan:r,count:i}),{key:!1,debug:()=>this.options.debug}),this.indexFromElement=n=>{const r=this.options.indexAttribute,i=n.getAttribute(r);return i?parseInt(i,10):(console.warn(`Missing attribute name '${r}={index}' on measured element.`),-1)},this._measureElement=(n,r)=>{const i=this.indexFromElement(n),a=this.measurementsCache[i];if(!a)return;const u=a.key,l=this.elementsCache.get(u);l!==n&&(l&&this.observer.unobserve(l),this.observer.observe(n),this.elementsCache.set(u,n)),n.isConnected&&this.resizeItem(i,this.options.measureElement(n,r,this))},this.resizeItem=(n,r)=>{const i=this.measurementsCache[n];if(!i)return;const a=this.itemSizeCache.get(i.key)??i.size,u=r-a;u!==0&&((this.shouldAdjustScrollPositionOnItemSizeChange!==void 0?this.shouldAdjustScrollPositionOnItemSizeChange(i,u,this):i.start<this.getScrollOffset()+this.scrollAdjustments)&&this._scrollToOffset(this.getScrollOffset(),{adjustments:this.scrollAdjustments+=u,behavior:void 0}),this.pendingMeasuredCacheIndexes.push(i.index),this.itemSizeCache=new Map(this.itemSizeCache.set(i.key,r)),this.notify(!1))},this.measureElement=n=>{if(!n){this.elementsCache.forEach((r,i)=>{r.isConnected||(this.observer.unobserve(r),this.elementsCache.delete(i))});return}this._measureElement(n,void 0)},this.getVirtualItems=AO(()=>[this.getVirtualIndexes(),this.getMeasurements()],(n,r)=>{const i=[];for(let a=0,u=n.length;a<u;a++){const l=n[a],o=r[l];i.push(o)}return i},{key:!1,debug:()=>this.options.debug}),this.getVirtualItemForOffset=n=>{const r=this.getMeasurements();if(r.length!==0)return FK(r[cbe(0,r.length-1,i=>FK(r[i]).start,n)])},this.getOffsetForAlignment=(n,r,i=0)=>{const a=this.getSize(),u=this.getScrollOffset();r==="auto"&&(r=n>=u+a?"end":"start"),r==="center"?n+=(i-a)/2:r==="end"&&(n-=a);const l=this.getTotalSize()+this.options.scrollMargin-a;return Math.max(Math.min(l,n),0)},this.getOffsetForIndex=(n,r="auto")=>{n=Math.max(0,Math.min(n,this.options.count-1));const i=this.measurementsCache[n];if(!i)return;const a=this.getSize(),u=this.getScrollOffset();if(r==="auto")if(i.end>=u+a-this.options.scrollPaddingEnd)r="end";else if(i.start<=u+this.options.scrollPaddingStart)r="start";else return[u,r];const l=r==="end"?i.end+this.options.scrollPaddingEnd:i.start-this.options.scrollPaddingStart;return[this.getOffsetForAlignment(l,r,i.size),r]},this.isDynamicMode=()=>this.elementsCache.size>0,this.scrollToOffset=(n,{align:r="start",behavior:i}={})=>{i==="smooth"&&this.isDynamicMode()&&console.warn("The `smooth` scroll behavior is not fully supported with dynamic size."),this._scrollToOffset(this.getOffsetForAlignment(n,r),{adjustments:void 0,behavior:i})},this.scrollToIndex=(n,{align:r="auto",behavior:i}={})=>{i==="smooth"&&this.isDynamicMode()&&console.warn("The `smooth` scroll behavior is not fully supported with dynamic size."),n=Math.max(0,Math.min(n,this.options.count-1));let a=0;const u=10,l=s=>{if(!this.targetWindow)return;const c=this.getOffsetForIndex(n,s);if(!c){console.warn("Failed to get offset for index:",n);return}const[d,h]=c;this._scrollToOffset(d,{adjustments:void 0,behavior:i}),this.targetWindow.requestAnimationFrame(()=>{const m=this.getScrollOffset(),y=this.getOffsetForIndex(n,h);if(!y){console.warn("Failed to get offset for index:",n);return}hHe(y[0],m)||o(h)})},o=s=>{this.targetWindow&&(a++,a<u?this.targetWindow.requestAnimationFrame(()=>l(s)):console.warn(`Failed to scroll to index ${n} after ${u} attempts.`))};l(r)},this.scrollBy=(n,{behavior:r}={})=>{r==="smooth"&&this.isDynamicMode()&&console.warn("The `smooth` scroll behavior is not fully supported with dynamic size."),this._scrollToOffset(this.getScrollOffset()+n,{adjustments:void 0,behavior:r})},this.getTotalSize=()=>{var n;const r=this.getMeasurements();let i;if(r.length===0)i=this.options.paddingStart;else if(this.options.lanes===1)i=((n=r[r.length-1])==null?void 0:n.end)??0;else{const a=Array(this.options.lanes).fill(null);let u=r.length-1;for(;u>=0&&a.some(l=>l===null);){const l=r[u];a[l.lane]===null&&(a[l.lane]=l.end),u--}i=Math.max(...a.filter(l=>l!==null))}return Math.max(i-this.options.scrollMargin+this.options.paddingEnd,0)},this._scrollToOffset=(n,{adjustments:r,behavior:i})=>{this.options.scrollToFn(n,{behavior:i,adjustments:r},this)},this.measure=()=>{this.itemSizeCache=new Map,this.notify(!1)},this.setOptions(t)}}const cbe=(e,t,n,r)=>{for(;e<=t;){const i=(e+t)/2|0,a=n(i);if(a<r)e=i+1;else if(a>r)t=i-1;else return i}return e>0?e-1:0};function SHe({measurements:e,outerSize:t,scrollOffset:n,lanes:r}){const i=e.length-1,a=o=>e[o].start;if(e.length<=r)return{startIndex:0,endIndex:i};let u=cbe(0,i,a,n),l=u;if(r===1)for(;l<i&&e[l].end<n+t;)l++;else if(r>1){const o=Array(r).fill(0);for(;l<i&&o.some(c=>c<n+t);){const c=e[l];o[c.lane]=c.end,l++}const s=Array(r).fill(n+t);for(;u>=0&&s.some(c=>c>=n);){const c=e[u];s[c.lane]=c.start,u--}u=Math.max(0,u-u%r),l=Math.min(i,l+(r-1-l%r))}return{startIndex:u,endIndex:l}}const BK=typeof document<"u"?P.useLayoutEffect:P.useEffect;function _He(e){const t=P.useReducer(()=>({}),{})[1],n={...e,onChange:(i,a)=>{var u;a?Uf.flushSync(t):t(),(u=e.onChange)==null||u.call(e,i,a)}},[r]=P.useState(()=>new xHe(n));return r.setOptions(n),BK(()=>r._didMount(),[]),BK(()=>r._willUpdate()),r}function $lt(e){return _He({observeElementRect:vHe,observeElementOffset:yHe,scrollToFn:OHe,...e})}const EHe=/^[$_\p{ID_Start}][$_\u{200C}\u{200D}\p{ID_Continue}]*$/u,DHe=/^[$_\p{ID_Start}][-$_\u{200C}\u{200D}\p{ID_Continue}]*$/u,$He={};function QK(e,t){return($He.jsx?DHe:EHe).test(e)}var TO={},y7,zK;function CHe(){if(zK)return y7;zK=1;var e=/\/\*[^*]*\*+([^/*][^*]*\*+)*\//g,t=/\n/g,n=/^\s*/,r=/^(\*?[-#/*\\\w]+(\[[0-9a-z_-]+\])?)\s*/,i=/^:\s*/,a=/^((?:'(?:\\'|.)*?'|"(?:\\"|.)*?"|\([^)]*?\)|[^};])+)/,u=/^[;\s]*/,l=/^\s+|\s+$/g,o=`
`,s="/",c="*",d="",h="comment",m="declaration";y7=function(v,S){if(typeof v!="string")throw new TypeError("First argument must be a string");if(!v)return[];S=S||{};var _=1,E=1;function $(N){var j=N.match(t);j&&(_+=j.length);var W=N.lastIndexOf(o);E=~W?N.length-W:E+N.length}function k(){var N={line:_,column:E};return function(j){return j.position=new A(N),F(),j}}function A(N){this.start=N,this.end={line:_,column:E},this.source=S.source}A.prototype.content=v;function w(N){var j=new Error(S.source+":"+_+":"+E+": "+N);if(j.reason=N,j.filename=S.source,j.line=_,j.column=E,j.source=v,!S.silent)throw j}function R(N){var j=N.exec(v);if(j){var W=j[0];return $(W),v=v.slice(W.length),j}}function F(){R(n)}function I(N){var j;for(N=N||[];j=U();)j!==!1&&N.push(j);return N}function U(){var N=k();if(!(s!=v.charAt(0)||c!=v.charAt(1))){for(var j=2;d!=v.charAt(j)&&(c!=v.charAt(j)||s!=v.charAt(j+1));)++j;if(j+=2,d===v.charAt(j-1))return w("End of comment missing");var W=v.slice(2,j-2);return E+=2,$(W),v=v.slice(j),E+=2,N({type:h,comment:W})}}function q(){var N=k(),j=R(r);if(j){if(U(),!R(i))return w("property missing ':'");var W=R(a),K=N({type:m,property:y(j[0].replace(e,d)),value:W?y(W[0].replace(e,d)):d});return R(u),K}}function V(){var N=[];I(N);for(var j;j=q();)j!==!1&&(N.push(j),I(N));return N}return F(),V()};function y(v){return v?v.replace(l,d):d}return y7}var qK;function wHe(){if(qK)return TO;qK=1;var e=TO&&TO.__importDefault||function(r){return r&&r.__esModule?r:{default:r}};Object.defineProperty(TO,"__esModule",{value:!0}),TO.default=n;var t=e(CHe());function n(r,i){var a=null;if(!r||typeof r!="string")return a;var u=(0,t.default)(r),l=typeof i=="function";return u.forEach(function(o){if(o.type==="declaration"){var s=o.property,c=o.value;l?i(s,c,o):c&&(a=a||{},a[s]=c)}}),a}return TO}var D2={},VK;function RHe(){if(VK)return D2;VK=1,Object.defineProperty(D2,"__esModule",{value:!0}),D2.camelCase=void 0;var e=/^--[a-zA-Z0-9_-]+$/,t=/-([a-z])/g,n=/^[^-]+$/,r=/^-(webkit|moz|ms|o|khtml)-/,i=/^-(ms)-/,a=function(s){return!s||n.test(s)||e.test(s)},u=function(s,c){return c.toUpperCase()},l=function(s,c){return"".concat(c,"-")},o=function(s,c){return c===void 0&&(c={}),a(s)?s:(s=s.toLowerCase(),c.reactCompat?s=s.replace(i,l):s=s.replace(r,l),s.replace(t,u))};return D2.camelCase=o,D2}var $2,UK;function AHe(){if(UK)return $2;UK=1;var e=$2&&$2.__importDefault||function(i){return i&&i.__esModule?i:{default:i}},t=e(wHe()),n=RHe();function r(i,a){var u={};return!i||typeof i!="string"||(0,t.default)(i,function(l,o){l&&o&&(u[(0,n.camelCase)(l,a)]=o)}),u}return r.default=r,$2=r,$2}var THe=AHe();const kHe=da(THe),fbe=dbe("end"),cB=dbe("start");function dbe(e){return t;function t(n){const r=n&&n.position&&n.position[e]||{};if(typeof r.line=="number"&&r.line>0&&typeof r.column=="number"&&r.column>0)return{line:r.line,column:r.column,offset:typeof r.offset=="number"&&r.offset>-1?r.offset:void 0}}}function PHe(e){const t=cB(e),n=fbe(e);if(t&&n)return{start:t,end:n}}function tx(e){return!e||typeof e!="object"?"":"position"in e||"type"in e?jK(e.position):"start"in e||"end"in e?jK(e):"line"in e||"column"in e?mF(e):""}function mF(e){return ZK(e&&e.line)+":"+ZK(e&&e.column)}function jK(e){return mF(e&&e.start)+"-"+mF(e&&e.end)}function ZK(e){return e&&typeof e=="number"?e:1}class Es extends Error{constructor(t,n,r){super(),typeof n=="string"&&(r=n,n=void 0);let i="",a={},u=!1;if(n&&("line"in n&&"column"in n?a={place:n}:"start"in n&&"end"in n?a={place:n}:"type"in n?a={ancestors:[n],place:n.position}:a={...n}),typeof t=="string"?i=t:!a.cause&&t&&(u=!0,i=t.message,a.cause=t),!a.ruleId&&!a.source&&typeof r=="string"){const o=r.indexOf(":");o===-1?a.ruleId=r:(a.source=r.slice(0,o),a.ruleId=r.slice(o+1))}if(!a.place&&a.ancestors&&a.ancestors){const o=a.ancestors[a.ancestors.length-1];o&&(a.place=o.position)}const l=a.place&&"start"in a.place?a.place.start:a.place;this.ancestors=a.ancestors||void 0,this.cause=a.cause||void 0,this.column=l?l.column:void 0,this.fatal=void 0,this.file="",this.message=i,this.line=l?l.line:void 0,this.name=tx(a.place)||"1:1",this.place=a.place||void 0,this.reason=this.message,this.ruleId=a.ruleId||void 0,this.source=a.source||void 0,this.stack=u&&a.cause&&typeof a.cause.stack=="string"?a.cause.stack:"",this.actual=void 0,this.expected=void 0,this.note=void 0,this.url=void 0}}Es.prototype.file="";Es.prototype.name="";Es.prototype.reason="";Es.prototype.message="";Es.prototype.stack="";Es.prototype.column=void 0;Es.prototype.line=void 0;Es.prototype.ancestors=void 0;Es.prototype.cause=void 0;Es.prototype.fatal=void 0;Es.prototype.place=void 0;Es.prototype.ruleId=void 0;Es.prototype.source=void 0;const fB={}.hasOwnProperty,MHe=new Map,FHe=/[A-Z]/g,NHe=new Set(["table","tbody","thead","tfoot","tr"]),IHe=new Set(["td","th"]),hbe="https://github.com/syntax-tree/hast-util-to-jsx-runtime";function LHe(e,t){if(!t||t.Fragment===void 0)throw new TypeError("Expected `Fragment` in options");const n=t.filePath||void 0;let r;if(t.development){if(typeof t.jsxDEV!="function")throw new TypeError("Expected `jsxDEV` in options when `development: true`");r=ZHe(n,t.jsxDEV)}else{if(typeof t.jsx!="function")throw new TypeError("Expected `jsx` in production options");if(typeof t.jsxs!="function")throw new TypeError("Expected `jsxs` in production options");r=jHe(n,t.jsx,t.jsxs)}const i={Fragment:t.Fragment,ancestors:[],components:t.components||{},create:r,elementAttributeNameCase:t.elementAttributeNameCase||"react",evaluater:t.createEvaluater?t.createEvaluater():void 0,filePath:n,ignoreInvalidStyle:t.ignoreInvalidStyle||!1,passKeys:t.passKeys!==!1,passNode:t.passNode||!1,schema:t.space==="svg"?FS:Qme,stylePropertyNameCase:t.stylePropertyNameCase||"dom",tableCellAlignToStyle:t.tableCellAlignToStyle!==!1},a=pbe(i,e,void 0);return a&&typeof a!="string"?a:i.create(e,i.Fragment,{children:a||void 0},void 0)}function pbe(e,t,n){if(t.type==="element")return BHe(e,t,n);if(t.type==="mdxFlowExpression"||t.type==="mdxTextExpression")return QHe(e,t);if(t.type==="mdxJsxFlowElement"||t.type==="mdxJsxTextElement")return qHe(e,t,n);if(t.type==="mdxjsEsm")return zHe(e,t);if(t.type==="root")return VHe(e,t,n);if(t.type==="text")return UHe(e,t)}function BHe(e,t,n){const r=e.schema;let i=r;t.tagName.toLowerCase()==="svg"&&r.space==="html"&&(i=FS,e.schema=i),e.ancestors.push(t);const a=gbe(e,t.tagName,!1),u=HHe(e,t);let l=hB(e,t);return NHe.has(t.tagName)&&(l=l.filter(function(o){return typeof o=="string"?!Aw(o):!0})),mbe(e,u,a,t),dB(u,l),e.ancestors.pop(),e.schema=r,e.create(t,a,u,n)}function QHe(e,t){if(t.data&&t.data.estree&&e.evaluater){const r=t.data.estree.body[0];return r.type,e.evaluater.evaluateExpression(r.expression)}eS(e,t.position)}function zHe(e,t){if(t.data&&t.data.estree&&e.evaluater)return e.evaluater.evaluateProgram(t.data.estree);eS(e,t.position)}function qHe(e,t,n){const r=e.schema;let i=r;t.name==="svg"&&r.space==="html"&&(i=FS,e.schema=i),e.ancestors.push(t);const a=t.name===null?e.Fragment:gbe(e,t.name,!0),u=XHe(e,t),l=hB(e,t);return mbe(e,u,a,t),dB(u,l),e.ancestors.pop(),e.schema=r,e.create(t,a,u,n)}function VHe(e,t,n){const r={};return dB(r,hB(e,t)),e.create(t,e.Fragment,r,n)}function UHe(e,t){return t.value}function mbe(e,t,n,r){typeof n!="string"&&n!==e.Fragment&&e.passNode&&(t.node=r)}function dB(e,t){if(t.length>0){const n=t.length>1?t:t[0];n&&(e.children=n)}}function jHe(e,t,n){return r;function r(i,a,u,l){const s=Array.isArray(u.children)?n:t;return l?s(a,u,l):s(a,u)}}function ZHe(e,t){return n;function n(r,i,a,u){const l=Array.isArray(a.children),o=cB(r);return t(i,a,u,l,{columnNumber:o?o.column-1:void 0,fileName:e,lineNumber:o?o.line:void 0},void 0)}}function HHe(e,t){const n={};let r,i;for(i in t.properties)if(i!=="children"&&fB.call(t.properties,i)){const a=YHe(e,i,t.properties[i]);if(a){const[u,l]=a;e.tableCellAlignToStyle&&u==="align"&&typeof l=="string"&&IHe.has(t.tagName)?r=l:n[u]=l}}if(r){const a=n.style||(n.style={});a[e.stylePropertyNameCase==="css"?"text-align":"textAlign"]=r}return n}function XHe(e,t){const n={};for(const r of t.attributes)if(r.type==="mdxJsxExpressionAttribute")if(r.data&&r.data.estree&&e.evaluater){const a=r.data.estree.body[0];a.type;const u=a.expression;u.type;const l=u.properties[0];l.type,Object.assign(n,e.evaluater.evaluateExpression(l.argument))}else eS(e,t.position);else{const i=r.name;let a;if(r.value&&typeof r.value=="object")if(r.value.data&&r.value.data.estree&&e.evaluater){const l=r.value.data.estree.body[0];l.type,a=e.evaluater.evaluateExpression(l.expression)}else eS(e,t.position);else a=r.value===null?!0:r.value;n[i]=a}return n}function hB(e,t){const n=[];let r=-1;const i=e.passKeys?new Map:MHe;for(;++r<t.children.length;){const a=t.children[r];let u;if(e.passKeys){const o=a.type==="element"?a.tagName:a.type==="mdxJsxFlowElement"||a.type==="mdxJsxTextElement"?a.name:void 0;if(o){const s=i.get(o)||0;u=o+"-"+s,i.set(o,s+1)}}const l=pbe(e,a,u);l!==void 0&&n.push(l)}return n}function YHe(e,t,n){const r=Bme(e.schema,t);if(!(n==null||typeof n=="number"&&Number.isNaN(n))){if(Array.isArray(n)&&(n=r.commaSeparated?qme(n):Vme(n)),r.property==="style"){let i=typeof n=="object"?n:GHe(e,String(n));return e.stylePropertyNameCase==="css"&&(i=KHe(i)),["style",i]}return[e.elementAttributeNameCase==="react"&&r.space?kIe[r.property]||r.property:r.attribute,n]}}function GHe(e,t){try{return kHe(t,{reactCompat:!0})}catch(n){if(e.ignoreInvalidStyle)return{};const r=n,i=new Es("Cannot parse `style` attribute",{ancestors:e.ancestors,cause:r,ruleId:"style",source:"hast-util-to-jsx-runtime"});throw i.file=e.filePath||void 0,i.url=hbe+"#cannot-parse-style-attribute",i}}function gbe(e,t,n){let r;if(!n)r={type:"Literal",value:t};else if(t.includes(".")){const i=t.split(".");let a=-1,u;for(;++a<i.length;){const l=QK(i[a])?{type:"Identifier",name:i[a]}:{type:"Literal",value:i[a]};u=u?{type:"MemberExpression",object:u,property:l,computed:!!(a&&l.type==="Literal"),optional:!1}:l}r=u}else r=QK(t)&&!/^[a-z]/.test(t)?{type:"Identifier",name:t}:{type:"Literal",value:t};if(r.type==="Literal"){const i=r.value;return fB.call(e.components,i)?e.components[i]:i}if(e.evaluater)return e.evaluater.evaluateExpression(r);eS(e)}function eS(e,t){const n=new Es("Cannot handle MDX estrees without `createEvaluater`",{ancestors:e.ancestors,place:t,ruleId:"mdx-estree",source:"hast-util-to-jsx-runtime"});throw n.file=e.filePath||void 0,n.url=hbe+"#cannot-handle-mdx-estrees-without-createevaluater",n}function KHe(e){const t={};let n;for(n in e)fB.call(e,n)&&(t[WHe(n)]=e[n]);return t}function WHe(e){let t=e.replace(FHe,JHe);return t.slice(0,3)==="ms-"&&(t="-"+t),t}function JHe(e){return"-"+e.toLowerCase()}const b7={action:["form"],cite:["blockquote","del","ins","q"],data:["object"],formAction:["button","input"],href:["a","area","base","link"],icon:["menuitem"],itemId:null,manifest:["html"],ping:["a","area"],poster:["video"],src:["audio","embed","iframe","img","input","script","source","track","video"]},eXe={};function pB(e,t){const n=eXe,r=typeof n.includeImageAlt=="boolean"?n.includeImageAlt:!0,i=typeof n.includeHtml=="boolean"?n.includeHtml:!0;return vbe(e,r,i)}function vbe(e,t,n){if(tXe(e)){if("value"in e)return e.type==="html"&&!n?"":e.value;if(t&&"alt"in e&&e.alt)return e.alt;if("children"in e)return HK(e.children,t,n)}return Array.isArray(e)?HK(e,t,n):""}function HK(e,t,n){const r=[];let i=-1;for(;++i<e.length;)r[i]=vbe(e[i],t,n);return r.join("")}function tXe(e){return!!(e&&typeof e=="object")}const XK=document.createElement("i");function mB(e){const t="&"+e+";";XK.innerHTML=t;const n=XK.textContent;return n.charCodeAt(n.length-1)===59&&e!=="semi"||n===t?!1:n}function Hf(e,t,n,r){const i=e.length;let a=0,u;if(t<0?t=-t>i?0:i+t:t=t>i?i:t,n=n>0?n:0,r.length<1e4)u=Array.from(r),u.unshift(t,n),e.splice(...u);else for(n&&e.splice(t,n);a<r.length;)u=r.slice(a,a+1e4),u.unshift(t,0),e.splice(...u),a+=1e4,t+=1e4}function Zd(e,t){return e.length>0?(Hf(e,e.length,0,t),e):t}const YK={}.hasOwnProperty;function ybe(e){const t={};let n=-1;for(;++n<e.length;)nXe(t,e[n]);return t}function nXe(e,t){let n;for(n in t){const i=(YK.call(e,n)?e[n]:void 0)||(e[n]={}),a=t[n];let u;if(a)for(u in a){YK.call(i,u)||(i[u]=[]);const l=a[u];rXe(i[u],Array.isArray(l)?l:l?[l]:[])}}}function rXe(e,t){let n=-1;const r=[];for(;++n<t.length;)(t[n].add==="after"?e:r).push(t[n]);Hf(e,0,0,r)}function bbe(e,t){const n=Number.parseInt(e,t);return n<9||n===11||n>13&&n<32||n>126&&n<160||n>55295&&n<57344||n>64975&&n<65008||(n&65535)===65535||(n&65535)===65534||n>1114111?"�":String.fromCodePoint(n)}function up(e){return e.replace(/[\t\n\r ]+/g," ").replace(/^ | $/g,"").toLowerCase().toUpperCase()}const cc=Ey(/[A-Za-z]/),bs=Ey(/[\dA-Za-z]/),iXe=Ey(/[#-'*+\--9=?A-Z^-~]/);function lC(e){return e!==null&&(e<32||e===127)}const gF=Ey(/\d/),aXe=Ey(/[\dA-Fa-f]/),uXe=Ey(/[!-/:-@[-`{-~]/);function Pn(e){return e!==null&&e<-2}function ca(e){return e!==null&&(e<0||e===32)}function Zr(e){return e===-2||e===-1||e===32}const d5=Ey(new RegExp("\\p{P}|\\p{S}","u")),x1=Ey(/\s/);function Ey(e){return t;function t(n){return n!==null&&n>-1&&e.test(String.fromCharCode(n))}}function m4(e){const t=[];let n=-1,r=0,i=0;for(;++n<e.length;){const a=e.charCodeAt(n);let u="";if(a===37&&bs(e.charCodeAt(n+1))&&bs(e.charCodeAt(n+2)))i=2;else if(a<128)/[!#$&-;=?-Z_a-z~]/.test(String.fromCharCode(a))||(u=String.fromCharCode(a));else if(a>55295&&a<57344){const l=e.charCodeAt(n+1);a<56320&&l>56319&&l<57344?(u=String.fromCharCode(a,l),i=1):u="�"}else u=String.fromCharCode(a);u&&(t.push(e.slice(r,n),encodeURIComponent(u)),r=n+i+1,u=""),i&&(n+=i,i=0)}return t.join("")+e.slice(r)}function di(e,t,n,r){const i=r?r-1:Number.POSITIVE_INFINITY;let a=0;return u;function u(o){return Zr(o)?(e.enter(n),l(o)):t(o)}function l(o){return Zr(o)&&a++<i?(e.consume(o),l):(e.exit(n),t(o))}}const lXe={tokenize:oXe};function oXe(e){const t=e.attempt(this.parser.constructs.contentInitial,r,i);let n;return t;function r(l){if(l===null){e.consume(l);return}return e.enter("lineEnding"),e.consume(l),e.exit("lineEnding"),di(e,t,"linePrefix")}function i(l){return e.enter("paragraph"),a(l)}function a(l){const o=e.enter("chunkText",{contentType:"text",previous:n});return n&&(n.next=o),n=o,u(l)}function u(l){if(l===null){e.exit("chunkText"),e.exit("paragraph"),e.consume(l);return}return Pn(l)?(e.consume(l),e.exit("chunkText"),a):(e.consume(l),u)}}const sXe={tokenize:cXe},GK={tokenize:fXe};function cXe(e){const t=this,n=[];let r=0,i,a,u;return l;function l($){if(r<n.length){const k=n[r];return t.containerState=k[1],e.attempt(k[0].continuation,o,s)($)}return s($)}function o($){if(r++,t.containerState._closeFlow){t.containerState._closeFlow=void 0,i&&E();const k=t.events.length;let A=k,w;for(;A--;)if(t.events[A][0]==="exit"&&t.events[A][1].type==="chunkFlow"){w=t.events[A][1].end;break}_(r);let R=k;for(;R<t.events.length;)t.events[R][1].end={...w},R++;return Hf(t.events,A+1,0,t.events.slice(k)),t.events.length=R,s($)}return l($)}function s($){if(r===n.length){if(!i)return h($);if(i.currentConstruct&&i.currentConstruct.concrete)return y($);t.interrupt=!!(i.currentConstruct&&!i._gfmTableDynamicInterruptHack)}return t.containerState={},e.check(GK,c,d)($)}function c($){return i&&E(),_(r),h($)}function d($){return t.parser.lazy[t.now().line]=r!==n.length,u=t.now().offset,y($)}function h($){return t.containerState={},e.attempt(GK,m,y)($)}function m($){return r++,n.push([t.currentConstruct,t.containerState]),h($)}function y($){if($===null){i&&E(),_(0),e.consume($);return}return i=i||t.parser.flow(t.now()),e.enter("chunkFlow",{_tokenizer:i,contentType:"flow",previous:a}),v($)}function v($){if($===null){S(e.exit("chunkFlow"),!0),_(0),e.consume($);return}return Pn($)?(e.consume($),S(e.exit("chunkFlow")),r=0,t.interrupt=void 0,l):(e.consume($),v)}function S($,k){const A=t.sliceStream($);if(k&&A.push(null),$.previous=a,a&&(a.next=$),a=$,i.defineSkip($.start),i.write(A),t.parser.lazy[$.start.line]){let w=i.events.length;for(;w--;)if(i.events[w][1].start.offset<u&&(!i.events[w][1].end||i.events[w][1].end.offset>u))return;const R=t.events.length;let F=R,I,U;for(;F--;)if(t.events[F][0]==="exit"&&t.events[F][1].type==="chunkFlow"){if(I){U=t.events[F][1].end;break}I=!0}for(_(r),w=R;w<t.events.length;)t.events[w][1].end={...U},w++;Hf(t.events,F+1,0,t.events.slice(R)),t.events.length=w}}function _($){let k=n.length;for(;k-- >$;){const A=n[k];t.containerState=A[1],A[0].exit.call(t,e)}n.length=$}function E(){i.write([null]),a=void 0,i=void 0,t.containerState._closeFlow=void 0}}function fXe(e,t,n){return di(e,e.attempt(this.parser.constructs.document,t,n),"linePrefix",this.parser.constructs.disable.null.includes("codeIndented")?void 0:4)}function V3(e){if(e===null||ca(e)||x1(e))return 1;if(d5(e))return 2}function h5(e,t,n){const r=[];let i=-1;for(;++i<e.length;){const a=e[i].resolveAll;a&&!r.includes(a)&&(t=a(t,n),r.push(a))}return t}const vF={name:"attention",resolveAll:dXe,tokenize:hXe};function dXe(e,t){let n=-1,r,i,a,u,l,o,s,c;for(;++n<e.length;)if(e[n][0]==="enter"&&e[n][1].type==="attentionSequence"&&e[n][1]._close){for(r=n;r--;)if(e[r][0]==="exit"&&e[r][1].type==="attentionSequence"&&e[r][1]._open&&t.sliceSerialize(e[r][1]).charCodeAt(0)===t.sliceSerialize(e[n][1]).charCodeAt(0)){if((e[r][1]._close||e[n][1]._open)&&(e[n][1].end.offset-e[n][1].start.offset)%3&&!((e[r][1].end.offset-e[r][1].start.offset+e[n][1].end.offset-e[n][1].start.offset)%3))continue;o=e[r][1].end.offset-e[r][1].start.offset>1&&e[n][1].end.offset-e[n][1].start.offset>1?2:1;const d={...e[r][1].end},h={...e[n][1].start};KK(d,-o),KK(h,o),u={type:o>1?"strongSequence":"emphasisSequence",start:d,end:{...e[r][1].end}},l={type:o>1?"strongSequence":"emphasisSequence",start:{...e[n][1].start},end:h},a={type:o>1?"strongText":"emphasisText",start:{...e[r][1].end},end:{...e[n][1].start}},i={type:o>1?"strong":"emphasis",start:{...u.start},end:{...l.end}},e[r][1].end={...u.start},e[n][1].start={...l.end},s=[],e[r][1].end.offset-e[r][1].start.offset&&(s=Zd(s,[["enter",e[r][1],t],["exit",e[r][1],t]])),s=Zd(s,[["enter",i,t],["enter",u,t],["exit",u,t],["enter",a,t]]),s=Zd(s,h5(t.parser.constructs.insideSpan.null,e.slice(r+1,n),t)),s=Zd(s,[["exit",a,t],["enter",l,t],["exit",l,t],["exit",i,t]]),e[n][1].end.offset-e[n][1].start.offset?(c=2,s=Zd(s,[["enter",e[n][1],t],["exit",e[n][1],t]])):c=0,Hf(e,r-1,n-r+3,s),n=r+s.length-c-2;break}}for(n=-1;++n<e.length;)e[n][1].type==="attentionSequence"&&(e[n][1].type="data");return e}function hXe(e,t){const n=this.parser.constructs.attentionMarkers.null,r=this.previous,i=V3(r);let a;return u;function u(o){return a=o,e.enter("attentionSequence"),l(o)}function l(o){if(o===a)return e.consume(o),l;const s=e.exit("attentionSequence"),c=V3(o),d=!c||c===2&&i||n.includes(o),h=!i||i===2&&c||n.includes(r);return s._open=!!(a===42?d:d&&(i||!h)),s._close=!!(a===42?h:h&&(c||!d)),t(o)}}function KK(e,t){e.column+=t,e.offset+=t,e._bufferIndex+=t}const pXe={name:"autolink",tokenize:mXe};function mXe(e,t,n){let r=0;return i;function i(m){return e.enter("autolink"),e.enter("autolinkMarker"),e.consume(m),e.exit("autolinkMarker"),e.enter("autolinkProtocol"),a}function a(m){return cc(m)?(e.consume(m),u):m===64?n(m):s(m)}function u(m){return m===43||m===45||m===46||bs(m)?(r=1,l(m)):s(m)}function l(m){return m===58?(e.consume(m),r=0,o):(m===43||m===45||m===46||bs(m))&&r++<32?(e.consume(m),l):(r=0,s(m))}function o(m){return m===62?(e.exit("autolinkProtocol"),e.enter("autolinkMarker"),e.consume(m),e.exit("autolinkMarker"),e.exit("autolink"),t):m===null||m===32||m===60||lC(m)?n(m):(e.consume(m),o)}function s(m){return m===64?(e.consume(m),c):iXe(m)?(e.consume(m),s):n(m)}function c(m){return bs(m)?d(m):n(m)}function d(m){return m===46?(e.consume(m),r=0,c):m===62?(e.exit("autolinkProtocol").type="autolinkEmail",e.enter("autolinkMarker"),e.consume(m),e.exit("autolinkMarker"),e.exit("autolink"),t):h(m)}function h(m){if((m===45||bs(m))&&r++<63){const y=m===45?h:d;return e.consume(m),y}return n(m)}}const qS={partial:!0,tokenize:gXe};function gXe(e,t,n){return r;function r(a){return Zr(a)?di(e,i,"linePrefix")(a):i(a)}function i(a){return a===null||Pn(a)?t(a):n(a)}}const Obe={continuation:{tokenize:yXe},exit:bXe,name:"blockQuote",tokenize:vXe};function vXe(e,t,n){const r=this;return i;function i(u){if(u===62){const l=r.containerState;return l.open||(e.enter("blockQuote",{_container:!0}),l.open=!0),e.enter("blockQuotePrefix"),e.enter("blockQuoteMarker"),e.consume(u),e.exit("blockQuoteMarker"),a}return n(u)}function a(u){return Zr(u)?(e.enter("blockQuotePrefixWhitespace"),e.consume(u),e.exit("blockQuotePrefixWhitespace"),e.exit("blockQuotePrefix"),t):(e.exit("blockQuotePrefix"),t(u))}}function yXe(e,t,n){const r=this;return i;function i(u){return Zr(u)?di(e,a,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(u):a(u)}function a(u){return e.attempt(Obe,t,n)(u)}}function bXe(e){e.exit("blockQuote")}const xbe={name:"characterEscape",tokenize:OXe};function OXe(e,t,n){return r;function r(a){return e.enter("characterEscape"),e.enter("escapeMarker"),e.consume(a),e.exit("escapeMarker"),i}function i(a){return uXe(a)?(e.enter("characterEscapeValue"),e.consume(a),e.exit("characterEscapeValue"),e.exit("characterEscape"),t):n(a)}}const Sbe={name:"characterReference",tokenize:xXe};function xXe(e,t,n){const r=this;let i=0,a,u;return l;function l(d){return e.enter("characterReference"),e.enter("characterReferenceMarker"),e.consume(d),e.exit("characterReferenceMarker"),o}function o(d){return d===35?(e.enter("characterReferenceMarkerNumeric"),e.consume(d),e.exit("characterReferenceMarkerNumeric"),s):(e.enter("characterReferenceValue"),a=31,u=bs,c(d))}function s(d){return d===88||d===120?(e.enter("characterReferenceMarkerHexadecimal"),e.consume(d),e.exit("characterReferenceMarkerHexadecimal"),e.enter("characterReferenceValue"),a=6,u=aXe,c):(e.enter("characterReferenceValue"),a=7,u=gF,c(d))}function c(d){if(d===59&&i){const h=e.exit("characterReferenceValue");return u===bs&&!mB(r.sliceSerialize(h))?n(d):(e.enter("characterReferenceMarker"),e.consume(d),e.exit("characterReferenceMarker"),e.exit("characterReference"),t)}return u(d)&&i++<a?(e.consume(d),c):n(d)}}const WK={partial:!0,tokenize:_Xe},JK={concrete:!0,name:"codeFenced",tokenize:SXe};function SXe(e,t,n){const r=this,i={partial:!0,tokenize:A};let a=0,u=0,l;return o;function o(w){return s(w)}function s(w){const R=r.events[r.events.length-1];return a=R&&R[1].type==="linePrefix"?R[2].sliceSerialize(R[1],!0).length:0,l=w,e.enter("codeFenced"),e.enter("codeFencedFence"),e.enter("codeFencedFenceSequence"),c(w)}function c(w){return w===l?(u++,e.consume(w),c):u<3?n(w):(e.exit("codeFencedFenceSequence"),Zr(w)?di(e,d,"whitespace")(w):d(w))}function d(w){return w===null||Pn(w)?(e.exit("codeFencedFence"),r.interrupt?t(w):e.check(WK,v,k)(w)):(e.enter("codeFencedFenceInfo"),e.enter("chunkString",{contentType:"string"}),h(w))}function h(w){return w===null||Pn(w)?(e.exit("chunkString"),e.exit("codeFencedFenceInfo"),d(w)):Zr(w)?(e.exit("chunkString"),e.exit("codeFencedFenceInfo"),di(e,m,"whitespace")(w)):w===96&&w===l?n(w):(e.consume(w),h)}function m(w){return w===null||Pn(w)?d(w):(e.enter("codeFencedFenceMeta"),e.enter("chunkString",{contentType:"string"}),y(w))}function y(w){return w===null||Pn(w)?(e.exit("chunkString"),e.exit("codeFencedFenceMeta"),d(w)):w===96&&w===l?n(w):(e.consume(w),y)}function v(w){return e.attempt(i,k,S)(w)}function S(w){return e.enter("lineEnding"),e.consume(w),e.exit("lineEnding"),_}function _(w){return a>0&&Zr(w)?di(e,E,"linePrefix",a+1)(w):E(w)}function E(w){return w===null||Pn(w)?e.check(WK,v,k)(w):(e.enter("codeFlowValue"),$(w))}function $(w){return w===null||Pn(w)?(e.exit("codeFlowValue"),E(w)):(e.consume(w),$)}function k(w){return e.exit("codeFenced"),t(w)}function A(w,R,F){let I=0;return U;function U(W){return w.enter("lineEnding"),w.consume(W),w.exit("lineEnding"),q}function q(W){return w.enter("codeFencedFence"),Zr(W)?di(w,V,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(W):V(W)}function V(W){return W===l?(w.enter("codeFencedFenceSequence"),N(W)):F(W)}function N(W){return W===l?(I++,w.consume(W),N):I>=u?(w.exit("codeFencedFenceSequence"),Zr(W)?di(w,j,"whitespace")(W):j(W)):F(W)}function j(W){return W===null||Pn(W)?(w.exit("codeFencedFence"),R(W)):F(W)}}}function _Xe(e,t,n){const r=this;return i;function i(u){return u===null?n(u):(e.enter("lineEnding"),e.consume(u),e.exit("lineEnding"),a)}function a(u){return r.parser.lazy[r.now().line]?n(u):t(u)}}const O7={name:"codeIndented",tokenize:DXe},EXe={partial:!0,tokenize:$Xe};function DXe(e,t,n){const r=this;return i;function i(s){return e.enter("codeIndented"),di(e,a,"linePrefix",5)(s)}function a(s){const c=r.events[r.events.length-1];return c&&c[1].type==="linePrefix"&&c[2].sliceSerialize(c[1],!0).length>=4?u(s):n(s)}function u(s){return s===null?o(s):Pn(s)?e.attempt(EXe,u,o)(s):(e.enter("codeFlowValue"),l(s))}function l(s){return s===null||Pn(s)?(e.exit("codeFlowValue"),u(s)):(e.consume(s),l)}function o(s){return e.exit("codeIndented"),t(s)}}function $Xe(e,t,n){const r=this;return i;function i(u){return r.parser.lazy[r.now().line]?n(u):Pn(u)?(e.enter("lineEnding"),e.consume(u),e.exit("lineEnding"),i):di(e,a,"linePrefix",5)(u)}function a(u){const l=r.events[r.events.length-1];return l&&l[1].type==="linePrefix"&&l[2].sliceSerialize(l[1],!0).length>=4?t(u):Pn(u)?i(u):n(u)}}const CXe={name:"codeText",previous:RXe,resolve:wXe,tokenize:AXe};function wXe(e){let t=e.length-4,n=3,r,i;if((e[n][1].type==="lineEnding"||e[n][1].type==="space")&&(e[t][1].type==="lineEnding"||e[t][1].type==="space")){for(r=n;++r<t;)if(e[r][1].type==="codeTextData"){e[n][1].type="codeTextPadding",e[t][1].type="codeTextPadding",n+=2,t-=2;break}}for(r=n-1,t++;++r<=t;)i===void 0?r!==t&&e[r][1].type!=="lineEnding"&&(i=r):(r===t||e[r][1].type==="lineEnding")&&(e[i][1].type="codeTextData",r!==i+2&&(e[i][1].end=e[r-1][1].end,e.splice(i+2,r-i-2),t-=r-i-2,r=i+2),i=void 0);return e}function RXe(e){return e!==96||this.events[this.events.length-1][1].type==="characterEscape"}function AXe(e,t,n){let r=0,i,a;return u;function u(d){return e.enter("codeText"),e.enter("codeTextSequence"),l(d)}function l(d){return d===96?(e.consume(d),r++,l):(e.exit("codeTextSequence"),o(d))}function o(d){return d===null?n(d):d===32?(e.enter("space"),e.consume(d),e.exit("space"),o):d===96?(a=e.enter("codeTextSequence"),i=0,c(d)):Pn(d)?(e.enter("lineEnding"),e.consume(d),e.exit("lineEnding"),o):(e.enter("codeTextData"),s(d))}function s(d){return d===null||d===32||d===96||Pn(d)?(e.exit("codeTextData"),o(d)):(e.consume(d),s)}function c(d){return d===96?(e.consume(d),i++,c):i===r?(e.exit("codeTextSequence"),e.exit("codeText"),t(d)):(a.type="codeTextData",s(d))}}class TXe{constructor(t){this.left=t?[...t]:[],this.right=[]}get(t){if(t<0||t>=this.left.length+this.right.length)throw new RangeError("Cannot access index `"+t+"` in a splice buffer of size `"+(this.left.length+this.right.length)+"`");return t<this.left.length?this.left[t]:this.right[this.right.length-t+this.left.length-1]}get length(){return this.left.length+this.right.length}shift(){return this.setCursor(0),this.right.pop()}slice(t,n){const r=n??Number.POSITIVE_INFINITY;return r<this.left.length?this.left.slice(t,r):t>this.left.length?this.right.slice(this.right.length-r+this.left.length,this.right.length-t+this.left.length).reverse():this.left.slice(t).concat(this.right.slice(this.right.length-r+this.left.length).reverse())}splice(t,n,r){const i=n||0;this.setCursor(Math.trunc(t));const a=this.right.splice(this.right.length-i,Number.POSITIVE_INFINITY);return r&&C2(this.left,r),a.reverse()}pop(){return this.setCursor(Number.POSITIVE_INFINITY),this.left.pop()}push(t){this.setCursor(Number.POSITIVE_INFINITY),this.left.push(t)}pushMany(t){this.setCursor(Number.POSITIVE_INFINITY),C2(this.left,t)}unshift(t){this.setCursor(0),this.right.push(t)}unshiftMany(t){this.setCursor(0),C2(this.right,t.reverse())}setCursor(t){if(!(t===this.left.length||t>this.left.length&&this.right.length===0||t<0&&this.left.length===0))if(t<this.left.length){const n=this.left.splice(t,Number.POSITIVE_INFINITY);C2(this.right,n.reverse())}else{const n=this.right.splice(this.left.length+this.right.length-t,Number.POSITIVE_INFINITY);C2(this.left,n.reverse())}}}function C2(e,t){let n=0;if(t.length<1e4)e.push(...t);else for(;n<t.length;)e.push(...t.slice(n,n+1e4)),n+=1e4}function _be(e){const t={};let n=-1,r,i,a,u,l,o,s;const c=new TXe(e);for(;++n<c.length;){for(;n in t;)n=t[n];if(r=c.get(n),n&&r[1].type==="chunkFlow"&&c.get(n-1)[1].type==="listItemPrefix"&&(o=r[1]._tokenizer.events,a=0,a<o.length&&o[a][1].type==="lineEndingBlank"&&(a+=2),a<o.length&&o[a][1].type==="content"))for(;++a<o.length&&o[a][1].type!=="content";)o[a][1].type==="chunkText"&&(o[a][1]._isInFirstContentOfListItem=!0,a++);if(r[0]==="enter")r[1].contentType&&(Object.assign(t,kXe(c,n)),n=t[n],s=!0);else if(r[1]._container){for(a=n,i=void 0;a--;)if(u=c.get(a),u[1].type==="lineEnding"||u[1].type==="lineEndingBlank")u[0]==="enter"&&(i&&(c.get(i)[1].type="lineEndingBlank"),u[1].type="lineEnding",i=a);else if(!(u[1].type==="linePrefix"||u[1].type==="listItemIndent"))break;i&&(r[1].end={...c.get(i)[1].start},l=c.slice(i,n),l.unshift(r),c.splice(i,n-i+1,l))}}return Hf(e,0,Number.POSITIVE_INFINITY,c.slice(0)),!s}function kXe(e,t){const n=e.get(t)[1],r=e.get(t)[2];let i=t-1;const a=[];let u=n._tokenizer;u||(u=r.parser[n.contentType](n.start),n._contentTypeTextTrailing&&(u._contentTypeTextTrailing=!0));const l=u.events,o=[],s={};let c,d,h=-1,m=n,y=0,v=0;const S=[v];for(;m;){for(;e.get(++i)[1]!==m;);a.push(i),m._tokenizer||(c=r.sliceStream(m),m.next||c.push(null),d&&u.defineSkip(m.start),m._isInFirstContentOfListItem&&(u._gfmTasklistFirstContentOfListItem=!0),u.write(c),m._isInFirstContentOfListItem&&(u._gfmTasklistFirstContentOfListItem=void 0)),d=m,m=m.next}for(m=n;++h<l.length;)l[h][0]==="exit"&&l[h-1][0]==="enter"&&l[h][1].type===l[h-1][1].type&&l[h][1].start.line!==l[h][1].end.line&&(v=h+1,S.push(v),m._tokenizer=void 0,m.previous=void 0,m=m.next);for(u.events=[],m?(m._tokenizer=void 0,m.previous=void 0):S.pop(),h=S.length;h--;){const _=l.slice(S[h],S[h+1]),E=a.pop();o.push([E,E+_.length-1]),e.splice(E,2,_)}for(o.reverse(),h=-1;++h<o.length;)s[y+o[h][0]]=y+o[h][1],y+=o[h][1]-o[h][0]-1;return s}const PXe={resolve:FXe,tokenize:NXe},MXe={partial:!0,tokenize:IXe};function FXe(e){return _be(e),e}function NXe(e,t){let n;return r;function r(l){return e.enter("content"),n=e.enter("chunkContent",{contentType:"content"}),i(l)}function i(l){return l===null?a(l):Pn(l)?e.check(MXe,u,a)(l):(e.consume(l),i)}function a(l){return e.exit("chunkContent"),e.exit("content"),t(l)}function u(l){return e.consume(l),e.exit("chunkContent"),n.next=e.enter("chunkContent",{contentType:"content",previous:n}),n=n.next,i}}function IXe(e,t,n){const r=this;return i;function i(u){return e.exit("chunkContent"),e.enter("lineEnding"),e.consume(u),e.exit("lineEnding"),di(e,a,"linePrefix")}function a(u){if(u===null||Pn(u))return n(u);const l=r.events[r.events.length-1];return!r.parser.constructs.disable.null.includes("codeIndented")&&l&&l[1].type==="linePrefix"&&l[2].sliceSerialize(l[1],!0).length>=4?t(u):e.interrupt(r.parser.constructs.flow,n,t)(u)}}function Ebe(e,t,n,r,i,a,u,l,o){const s=o||Number.POSITIVE_INFINITY;let c=0;return d;function d(_){return _===60?(e.enter(r),e.enter(i),e.enter(a),e.consume(_),e.exit(a),h):_===null||_===32||_===41||lC(_)?n(_):(e.enter(r),e.enter(u),e.enter(l),e.enter("chunkString",{contentType:"string"}),v(_))}function h(_){return _===62?(e.enter(a),e.consume(_),e.exit(a),e.exit(i),e.exit(r),t):(e.enter(l),e.enter("chunkString",{contentType:"string"}),m(_))}function m(_){return _===62?(e.exit("chunkString"),e.exit(l),h(_)):_===null||_===60||Pn(_)?n(_):(e.consume(_),_===92?y:m)}function y(_){return _===60||_===62||_===92?(e.consume(_),m):m(_)}function v(_){return!c&&(_===null||_===41||ca(_))?(e.exit("chunkString"),e.exit(l),e.exit(u),e.exit(r),t(_)):c<s&&_===40?(e.consume(_),c++,v):_===41?(e.consume(_),c--,v):_===null||_===32||_===40||lC(_)?n(_):(e.consume(_),_===92?S:v)}function S(_){return _===40||_===41||_===92?(e.consume(_),v):v(_)}}function Dbe(e,t,n,r,i,a){const u=this;let l=0,o;return s;function s(m){return e.enter(r),e.enter(i),e.consume(m),e.exit(i),e.enter(a),c}function c(m){return l>999||m===null||m===91||m===93&&!o||m===94&&!l&&"_hiddenFootnoteSupport"in u.parser.constructs?n(m):m===93?(e.exit(a),e.enter(i),e.consume(m),e.exit(i),e.exit(r),t):Pn(m)?(e.enter("lineEnding"),e.consume(m),e.exit("lineEnding"),c):(e.enter("chunkString",{contentType:"string"}),d(m))}function d(m){return m===null||m===91||m===93||Pn(m)||l++>999?(e.exit("chunkString"),c(m)):(e.consume(m),o||(o=!Zr(m)),m===92?h:d)}function h(m){return m===91||m===92||m===93?(e.consume(m),l++,d):d(m)}}function $be(e,t,n,r,i,a){let u;return l;function l(h){return h===34||h===39||h===40?(e.enter(r),e.enter(i),e.consume(h),e.exit(i),u=h===40?41:h,o):n(h)}function o(h){return h===u?(e.enter(i),e.consume(h),e.exit(i),e.exit(r),t):(e.enter(a),s(h))}function s(h){return h===u?(e.exit(a),o(u)):h===null?n(h):Pn(h)?(e.enter("lineEnding"),e.consume(h),e.exit("lineEnding"),di(e,s,"linePrefix")):(e.enter("chunkString",{contentType:"string"}),c(h))}function c(h){return h===u||h===null||Pn(h)?(e.exit("chunkString"),s(h)):(e.consume(h),h===92?d:c)}function d(h){return h===u||h===92?(e.consume(h),c):c(h)}}function nx(e,t){let n;return r;function r(i){return Pn(i)?(e.enter("lineEnding"),e.consume(i),e.exit("lineEnding"),n=!0,r):Zr(i)?di(e,r,n?"linePrefix":"lineSuffix")(i):t(i)}}const LXe={name:"definition",tokenize:QXe},BXe={partial:!0,tokenize:zXe};function QXe(e,t,n){const r=this;let i;return a;function a(m){return e.enter("definition"),u(m)}function u(m){return Dbe.call(r,e,l,n,"definitionLabel","definitionLabelMarker","definitionLabelString")(m)}function l(m){return i=up(r.sliceSerialize(r.events[r.events.length-1][1]).slice(1,-1)),m===58?(e.enter("definitionMarker"),e.consume(m),e.exit("definitionMarker"),o):n(m)}function o(m){return ca(m)?nx(e,s)(m):s(m)}function s(m){return Ebe(e,c,n,"definitionDestination","definitionDestinationLiteral","definitionDestinationLiteralMarker","definitionDestinationRaw","definitionDestinationString")(m)}function c(m){return e.attempt(BXe,d,d)(m)}function d(m){return Zr(m)?di(e,h,"whitespace")(m):h(m)}function h(m){return m===null||Pn(m)?(e.exit("definition"),r.parser.defined.push(i),t(m)):n(m)}}function zXe(e,t,n){return r;function r(l){return ca(l)?nx(e,i)(l):n(l)}function i(l){return $be(e,a,n,"definitionTitle","definitionTitleMarker","definitionTitleString")(l)}function a(l){return Zr(l)?di(e,u,"whitespace")(l):u(l)}function u(l){return l===null||Pn(l)?t(l):n(l)}}const qXe={name:"hardBreakEscape",tokenize:VXe};function VXe(e,t,n){return r;function r(a){return e.enter("hardBreakEscape"),e.consume(a),i}function i(a){return Pn(a)?(e.exit("hardBreakEscape"),t(a)):n(a)}}const UXe={name:"headingAtx",resolve:jXe,tokenize:ZXe};function jXe(e,t){let n=e.length-2,r=3,i,a;return e[r][1].type==="whitespace"&&(r+=2),n-2>r&&e[n][1].type==="whitespace"&&(n-=2),e[n][1].type==="atxHeadingSequence"&&(r===n-1||n-4>r&&e[n-2][1].type==="whitespace")&&(n-=r+1===n?2:4),n>r&&(i={type:"atxHeadingText",start:e[r][1].start,end:e[n][1].end},a={type:"chunkText",start:e[r][1].start,end:e[n][1].end,contentType:"text"},Hf(e,r,n-r+1,[["enter",i,t],["enter",a,t],["exit",a,t],["exit",i,t]])),e}function ZXe(e,t,n){let r=0;return i;function i(c){return e.enter("atxHeading"),a(c)}function a(c){return e.enter("atxHeadingSequence"),u(c)}function u(c){return c===35&&r++<6?(e.consume(c),u):c===null||ca(c)?(e.exit("atxHeadingSequence"),l(c)):n(c)}function l(c){return c===35?(e.enter("atxHeadingSequence"),o(c)):c===null||Pn(c)?(e.exit("atxHeading"),t(c)):Zr(c)?di(e,l,"whitespace")(c):(e.enter("atxHeadingText"),s(c))}function o(c){return c===35?(e.consume(c),o):(e.exit("atxHeadingSequence"),l(c))}function s(c){return c===null||c===35||ca(c)?(e.exit("atxHeadingText"),l(c)):(e.consume(c),s)}}const HXe=["address","article","aside","base","basefont","blockquote","body","caption","center","col","colgroup","dd","details","dialog","dir","div","dl","dt","fieldset","figcaption","figure","footer","form","frame","frameset","h1","h2","h3","h4","h5","h6","head","header","hr","html","iframe","legend","li","link","main","menu","menuitem","nav","noframes","ol","optgroup","option","p","param","search","section","summary","table","tbody","td","tfoot","th","thead","title","tr","track","ul"],eW=["pre","script","style","textarea"],XXe={concrete:!0,name:"htmlFlow",resolveTo:KXe,tokenize:WXe},YXe={partial:!0,tokenize:eYe},GXe={partial:!0,tokenize:JXe};function KXe(e){let t=e.length;for(;t--&&!(e[t][0]==="enter"&&e[t][1].type==="htmlFlow"););return t>1&&e[t-2][1].type==="linePrefix"&&(e[t][1].start=e[t-2][1].start,e[t+1][1].start=e[t-2][1].start,e.splice(t-2,2)),e}function WXe(e,t,n){const r=this;let i,a,u,l,o;return s;function s(B){return c(B)}function c(B){return e.enter("htmlFlow"),e.enter("htmlFlowData"),e.consume(B),d}function d(B){return B===33?(e.consume(B),h):B===47?(e.consume(B),a=!0,v):B===63?(e.consume(B),i=3,r.interrupt?t:M):cc(B)?(e.consume(B),u=String.fromCharCode(B),S):n(B)}function h(B){return B===45?(e.consume(B),i=2,m):B===91?(e.consume(B),i=5,l=0,y):cc(B)?(e.consume(B),i=4,r.interrupt?t:M):n(B)}function m(B){return B===45?(e.consume(B),r.interrupt?t:M):n(B)}function y(B){const ne="CDATA[";return B===ne.charCodeAt(l++)?(e.consume(B),l===ne.length?r.interrupt?t:V:y):n(B)}function v(B){return cc(B)?(e.consume(B),u=String.fromCharCode(B),S):n(B)}function S(B){if(B===null||B===47||B===62||ca(B)){const ne=B===47,oe=u.toLowerCase();return!ne&&!a&&eW.includes(oe)?(i=1,r.interrupt?t(B):V(B)):HXe.includes(u.toLowerCase())?(i=6,ne?(e.consume(B),_):r.interrupt?t(B):V(B)):(i=7,r.interrupt&&!r.parser.lazy[r.now().line]?n(B):a?E(B):$(B))}return B===45||bs(B)?(e.consume(B),u+=String.fromCharCode(B),S):n(B)}function _(B){return B===62?(e.consume(B),r.interrupt?t:V):n(B)}function E(B){return Zr(B)?(e.consume(B),E):U(B)}function $(B){return B===47?(e.consume(B),U):B===58||B===95||cc(B)?(e.consume(B),k):Zr(B)?(e.consume(B),$):U(B)}function k(B){return B===45||B===46||B===58||B===95||bs(B)?(e.consume(B),k):A(B)}function A(B){return B===61?(e.consume(B),w):Zr(B)?(e.consume(B),A):$(B)}function w(B){return B===null||B===60||B===61||B===62||B===96?n(B):B===34||B===39?(e.consume(B),o=B,R):Zr(B)?(e.consume(B),w):F(B)}function R(B){return B===o?(e.consume(B),o=null,I):B===null||Pn(B)?n(B):(e.consume(B),R)}function F(B){return B===null||B===34||B===39||B===47||B===60||B===61||B===62||B===96||ca(B)?A(B):(e.consume(B),F)}function I(B){return B===47||B===62||Zr(B)?$(B):n(B)}function U(B){return B===62?(e.consume(B),q):n(B)}function q(B){return B===null||Pn(B)?V(B):Zr(B)?(e.consume(B),q):n(B)}function V(B){return B===45&&i===2?(e.consume(B),K):B===60&&i===1?(e.consume(B),J):B===62&&i===4?(e.consume(B),G):B===63&&i===3?(e.consume(B),M):B===93&&i===5?(e.consume(B),H):Pn(B)&&(i===6||i===7)?(e.exit("htmlFlowData"),e.check(YXe,X,N)(B)):B===null||Pn(B)?(e.exit("htmlFlowData"),N(B)):(e.consume(B),V)}function N(B){return e.check(GXe,j,X)(B)}function j(B){return e.enter("lineEnding"),e.consume(B),e.exit("lineEnding"),W}function W(B){return B===null||Pn(B)?N(B):(e.enter("htmlFlowData"),V(B))}function K(B){return B===45?(e.consume(B),M):V(B)}function J(B){return B===47?(e.consume(B),u="",te):V(B)}function te(B){if(B===62){const ne=u.toLowerCase();return eW.includes(ne)?(e.consume(B),G):V(B)}return cc(B)&&u.length<8?(e.consume(B),u+=String.fromCharCode(B),te):V(B)}function H(B){return B===93?(e.consume(B),M):V(B)}function M(B){return B===62?(e.consume(B),G):B===45&&i===2?(e.consume(B),M):V(B)}function G(B){return B===null||Pn(B)?(e.exit("htmlFlowData"),X(B)):(e.consume(B),G)}function X(B){return e.exit("htmlFlow"),t(B)}}function JXe(e,t,n){const r=this;return i;function i(u){return Pn(u)?(e.enter("lineEnding"),e.consume(u),e.exit("lineEnding"),a):n(u)}function a(u){return r.parser.lazy[r.now().line]?n(u):t(u)}}function eYe(e,t,n){return r;function r(i){return e.enter("lineEnding"),e.consume(i),e.exit("lineEnding"),e.attempt(qS,t,n)}}const tYe={name:"htmlText",tokenize:nYe};function nYe(e,t,n){const r=this;let i,a,u;return l;function l(M){return e.enter("htmlText"),e.enter("htmlTextData"),e.consume(M),o}function o(M){return M===33?(e.consume(M),s):M===47?(e.consume(M),A):M===63?(e.consume(M),$):cc(M)?(e.consume(M),F):n(M)}function s(M){return M===45?(e.consume(M),c):M===91?(e.consume(M),a=0,y):cc(M)?(e.consume(M),E):n(M)}function c(M){return M===45?(e.consume(M),m):n(M)}function d(M){return M===null?n(M):M===45?(e.consume(M),h):Pn(M)?(u=d,J(M)):(e.consume(M),d)}function h(M){return M===45?(e.consume(M),m):d(M)}function m(M){return M===62?K(M):M===45?h(M):d(M)}function y(M){const G="CDATA[";return M===G.charCodeAt(a++)?(e.consume(M),a===G.length?v:y):n(M)}function v(M){return M===null?n(M):M===93?(e.consume(M),S):Pn(M)?(u=v,J(M)):(e.consume(M),v)}function S(M){return M===93?(e.consume(M),_):v(M)}function _(M){return M===62?K(M):M===93?(e.consume(M),_):v(M)}function E(M){return M===null||M===62?K(M):Pn(M)?(u=E,J(M)):(e.consume(M),E)}function $(M){return M===null?n(M):M===63?(e.consume(M),k):Pn(M)?(u=$,J(M)):(e.consume(M),$)}function k(M){return M===62?K(M):$(M)}function A(M){return cc(M)?(e.consume(M),w):n(M)}function w(M){return M===45||bs(M)?(e.consume(M),w):R(M)}function R(M){return Pn(M)?(u=R,J(M)):Zr(M)?(e.consume(M),R):K(M)}function F(M){return M===45||bs(M)?(e.consume(M),F):M===47||M===62||ca(M)?I(M):n(M)}function I(M){return M===47?(e.consume(M),K):M===58||M===95||cc(M)?(e.consume(M),U):Pn(M)?(u=I,J(M)):Zr(M)?(e.consume(M),I):K(M)}function U(M){return M===45||M===46||M===58||M===95||bs(M)?(e.consume(M),U):q(M)}function q(M){return M===61?(e.consume(M),V):Pn(M)?(u=q,J(M)):Zr(M)?(e.consume(M),q):I(M)}function V(M){return M===null||M===60||M===61||M===62||M===96?n(M):M===34||M===39?(e.consume(M),i=M,N):Pn(M)?(u=V,J(M)):Zr(M)?(e.consume(M),V):(e.consume(M),j)}function N(M){return M===i?(e.consume(M),i=void 0,W):M===null?n(M):Pn(M)?(u=N,J(M)):(e.consume(M),N)}function j(M){return M===null||M===34||M===39||M===60||M===61||M===96?n(M):M===47||M===62||ca(M)?I(M):(e.consume(M),j)}function W(M){return M===47||M===62||ca(M)?I(M):n(M)}function K(M){return M===62?(e.consume(M),e.exit("htmlTextData"),e.exit("htmlText"),t):n(M)}function J(M){return e.exit("htmlTextData"),e.enter("lineEnding"),e.consume(M),e.exit("lineEnding"),te}function te(M){return Zr(M)?di(e,H,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(M):H(M)}function H(M){return e.enter("htmlTextData"),u(M)}}const gB={name:"labelEnd",resolveAll:uYe,resolveTo:lYe,tokenize:oYe},rYe={tokenize:sYe},iYe={tokenize:cYe},aYe={tokenize:fYe};function uYe(e){let t=-1;const n=[];for(;++t<e.length;){const r=e[t][1];if(n.push(e[t]),r.type==="labelImage"||r.type==="labelLink"||r.type==="labelEnd"){const i=r.type==="labelImage"?4:2;r.type="data",t+=i}}return e.length!==n.length&&Hf(e,0,e.length,n),e}function lYe(e,t){let n=e.length,r=0,i,a,u,l;for(;n--;)if(i=e[n][1],a){if(i.type==="link"||i.type==="labelLink"&&i._inactive)break;e[n][0]==="enter"&&i.type==="labelLink"&&(i._inactive=!0)}else if(u){if(e[n][0]==="enter"&&(i.type==="labelImage"||i.type==="labelLink")&&!i._balanced&&(a=n,i.type!=="labelLink")){r=2;break}}else i.type==="labelEnd"&&(u=n);const o={type:e[a][1].type==="labelLink"?"link":"image",start:{...e[a][1].start},end:{...e[e.length-1][1].end}},s={type:"label",start:{...e[a][1].start},end:{...e[u][1].end}},c={type:"labelText",start:{...e[a+r+2][1].end},end:{...e[u-2][1].start}};return l=[["enter",o,t],["enter",s,t]],l=Zd(l,e.slice(a+1,a+r+3)),l=Zd(l,[["enter",c,t]]),l=Zd(l,h5(t.parser.constructs.insideSpan.null,e.slice(a+r+4,u-3),t)),l=Zd(l,[["exit",c,t],e[u-2],e[u-1],["exit",s,t]]),l=Zd(l,e.slice(u+1)),l=Zd(l,[["exit",o,t]]),Hf(e,a,e.length,l),e}function oYe(e,t,n){const r=this;let i=r.events.length,a,u;for(;i--;)if((r.events[i][1].type==="labelImage"||r.events[i][1].type==="labelLink")&&!r.events[i][1]._balanced){a=r.events[i][1];break}return l;function l(h){return a?a._inactive?d(h):(u=r.parser.defined.includes(up(r.sliceSerialize({start:a.end,end:r.now()}))),e.enter("labelEnd"),e.enter("labelMarker"),e.consume(h),e.exit("labelMarker"),e.exit("labelEnd"),o):n(h)}function o(h){return h===40?e.attempt(rYe,c,u?c:d)(h):h===91?e.attempt(iYe,c,u?s:d)(h):u?c(h):d(h)}function s(h){return e.attempt(aYe,c,d)(h)}function c(h){return t(h)}function d(h){return a._balanced=!0,n(h)}}function sYe(e,t,n){return r;function r(d){return e.enter("resource"),e.enter("resourceMarker"),e.consume(d),e.exit("resourceMarker"),i}function i(d){return ca(d)?nx(e,a)(d):a(d)}function a(d){return d===41?c(d):Ebe(e,u,l,"resourceDestination","resourceDestinationLiteral","resourceDestinationLiteralMarker","resourceDestinationRaw","resourceDestinationString",32)(d)}function u(d){return ca(d)?nx(e,o)(d):c(d)}function l(d){return n(d)}function o(d){return d===34||d===39||d===40?$be(e,s,n,"resourceTitle","resourceTitleMarker","resourceTitleString")(d):c(d)}function s(d){return ca(d)?nx(e,c)(d):c(d)}function c(d){return d===41?(e.enter("resourceMarker"),e.consume(d),e.exit("resourceMarker"),e.exit("resource"),t):n(d)}}function cYe(e,t,n){const r=this;return i;function i(l){return Dbe.call(r,e,a,u,"reference","referenceMarker","referenceString")(l)}function a(l){return r.parser.defined.includes(up(r.sliceSerialize(r.events[r.events.length-1][1]).slice(1,-1)))?t(l):n(l)}function u(l){return n(l)}}function fYe(e,t,n){return r;function r(a){return e.enter("reference"),e.enter("referenceMarker"),e.consume(a),e.exit("referenceMarker"),i}function i(a){return a===93?(e.enter("referenceMarker"),e.consume(a),e.exit("referenceMarker"),e.exit("reference"),t):n(a)}}const dYe={name:"labelStartImage",resolveAll:gB.resolveAll,tokenize:hYe};function hYe(e,t,n){const r=this;return i;function i(l){return e.enter("labelImage"),e.enter("labelImageMarker"),e.consume(l),e.exit("labelImageMarker"),a}function a(l){return l===91?(e.enter("labelMarker"),e.consume(l),e.exit("labelMarker"),e.exit("labelImage"),u):n(l)}function u(l){return l===94&&"_hiddenFootnoteSupport"in r.parser.constructs?n(l):t(l)}}const pYe={name:"labelStartLink",resolveAll:gB.resolveAll,tokenize:mYe};function mYe(e,t,n){const r=this;return i;function i(u){return e.enter("labelLink"),e.enter("labelMarker"),e.consume(u),e.exit("labelMarker"),e.exit("labelLink"),a}function a(u){return u===94&&"_hiddenFootnoteSupport"in r.parser.constructs?n(u):t(u)}}const x7={name:"lineEnding",tokenize:gYe};function gYe(e,t){return n;function n(r){return e.enter("lineEnding"),e.consume(r),e.exit("lineEnding"),di(e,t,"linePrefix")}}const ED={name:"thematicBreak",tokenize:vYe};function vYe(e,t,n){let r=0,i;return a;function a(s){return e.enter("thematicBreak"),u(s)}function u(s){return i=s,l(s)}function l(s){return s===i?(e.enter("thematicBreakSequence"),o(s)):r>=3&&(s===null||Pn(s))?(e.exit("thematicBreak"),t(s)):n(s)}function o(s){return s===i?(e.consume(s),r++,o):(e.exit("thematicBreakSequence"),Zr(s)?di(e,l,"whitespace")(s):l(s))}}const Gc={continuation:{tokenize:xYe},exit:_Ye,name:"list",tokenize:OYe},yYe={partial:!0,tokenize:EYe},bYe={partial:!0,tokenize:SYe};function OYe(e,t,n){const r=this,i=r.events[r.events.length-1];let a=i&&i[1].type==="linePrefix"?i[2].sliceSerialize(i[1],!0).length:0,u=0;return l;function l(m){const y=r.containerState.type||(m===42||m===43||m===45?"listUnordered":"listOrdered");if(y==="listUnordered"?!r.containerState.marker||m===r.containerState.marker:gF(m)){if(r.containerState.type||(r.containerState.type=y,e.enter(y,{_container:!0})),y==="listUnordered")return e.enter("listItemPrefix"),m===42||m===45?e.check(ED,n,s)(m):s(m);if(!r.interrupt||m===49)return e.enter("listItemPrefix"),e.enter("listItemValue"),o(m)}return n(m)}function o(m){return gF(m)&&++u<10?(e.consume(m),o):(!r.interrupt||u<2)&&(r.containerState.marker?m===r.containerState.marker:m===41||m===46)?(e.exit("listItemValue"),s(m)):n(m)}function s(m){return e.enter("listItemMarker"),e.consume(m),e.exit("listItemMarker"),r.containerState.marker=r.containerState.marker||m,e.check(qS,r.interrupt?n:c,e.attempt(yYe,h,d))}function c(m){return r.containerState.initialBlankLine=!0,a++,h(m)}function d(m){return Zr(m)?(e.enter("listItemPrefixWhitespace"),e.consume(m),e.exit("listItemPrefixWhitespace"),h):n(m)}function h(m){return r.containerState.size=a+r.sliceSerialize(e.exit("listItemPrefix"),!0).length,t(m)}}function xYe(e,t,n){const r=this;return r.containerState._closeFlow=void 0,e.check(qS,i,a);function i(l){return r.containerState.furtherBlankLines=r.containerState.furtherBlankLines||r.containerState.initialBlankLine,di(e,t,"listItemIndent",r.containerState.size+1)(l)}function a(l){return r.containerState.furtherBlankLines||!Zr(l)?(r.containerState.furtherBlankLines=void 0,r.containerState.initialBlankLine=void 0,u(l)):(r.containerState.furtherBlankLines=void 0,r.containerState.initialBlankLine=void 0,e.attempt(bYe,t,u)(l))}function u(l){return r.containerState._closeFlow=!0,r.interrupt=void 0,di(e,e.attempt(Gc,t,n),"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(l)}}function SYe(e,t,n){const r=this;return di(e,i,"listItemIndent",r.containerState.size+1);function i(a){const u=r.events[r.events.length-1];return u&&u[1].type==="listItemIndent"&&u[2].sliceSerialize(u[1],!0).length===r.containerState.size?t(a):n(a)}}function _Ye(e){e.exit(this.containerState.type)}function EYe(e,t,n){const r=this;return di(e,i,"listItemPrefixWhitespace",r.parser.constructs.disable.null.includes("codeIndented")?void 0:5);function i(a){const u=r.events[r.events.length-1];return!Zr(a)&&u&&u[1].type==="listItemPrefixWhitespace"?t(a):n(a)}}const tW={name:"setextUnderline",resolveTo:DYe,tokenize:$Ye};function DYe(e,t){let n=e.length,r,i,a;for(;n--;)if(e[n][0]==="enter"){if(e[n][1].type==="content"){r=n;break}e[n][1].type==="paragraph"&&(i=n)}else e[n][1].type==="content"&&e.splice(n,1),!a&&e[n][1].type==="definition"&&(a=n);const u={type:"setextHeading",start:{...e[r][1].start},end:{...e[e.length-1][1].end}};return e[i][1].type="setextHeadingText",a?(e.splice(i,0,["enter",u,t]),e.splice(a+1,0,["exit",e[r][1],t]),e[r][1].end={...e[a][1].end}):e[r][1]=u,e.push(["exit",u,t]),e}function $Ye(e,t,n){const r=this;let i;return a;function a(s){let c=r.events.length,d;for(;c--;)if(r.events[c][1].type!=="lineEnding"&&r.events[c][1].type!=="linePrefix"&&r.events[c][1].type!=="content"){d=r.events[c][1].type==="paragraph";break}return!r.parser.lazy[r.now().line]&&(r.interrupt||d)?(e.enter("setextHeadingLine"),i=s,u(s)):n(s)}function u(s){return e.enter("setextHeadingLineSequence"),l(s)}function l(s){return s===i?(e.consume(s),l):(e.exit("setextHeadingLineSequence"),Zr(s)?di(e,o,"lineSuffix")(s):o(s))}function o(s){return s===null||Pn(s)?(e.exit("setextHeadingLine"),t(s)):n(s)}}const CYe={tokenize:wYe};function wYe(e){const t=this,n=e.attempt(qS,r,e.attempt(this.parser.constructs.flowInitial,i,di(e,e.attempt(this.parser.constructs.flow,i,e.attempt(PXe,i)),"linePrefix")));return n;function r(a){if(a===null){e.consume(a);return}return e.enter("lineEndingBlank"),e.consume(a),e.exit("lineEndingBlank"),t.currentConstruct=void 0,n}function i(a){if(a===null){e.consume(a);return}return e.enter("lineEnding"),e.consume(a),e.exit("lineEnding"),t.currentConstruct=void 0,n}}const RYe={resolveAll:wbe()},AYe=Cbe("string"),TYe=Cbe("text");function Cbe(e){return{resolveAll:wbe(e==="text"?kYe:void 0),tokenize:t};function t(n){const r=this,i=this.parser.constructs[e],a=n.attempt(i,u,l);return u;function u(c){return s(c)?a(c):l(c)}function l(c){if(c===null){n.consume(c);return}return n.enter("data"),n.consume(c),o}function o(c){return s(c)?(n.exit("data"),a(c)):(n.consume(c),o)}function s(c){if(c===null)return!0;const d=i[c];let h=-1;if(d)for(;++h<d.length;){const m=d[h];if(!m.previous||m.previous.call(r,r.previous))return!0}return!1}}}function wbe(e){return t;function t(n,r){let i=-1,a;for(;++i<=n.length;)a===void 0?n[i]&&n[i][1].type==="data"&&(a=i,i++):(!n[i]||n[i][1].type!=="data")&&(i!==a+2&&(n[a][1].end=n[i-1][1].end,n.splice(a+2,i-a-2),i=a+2),a=void 0);return e?e(n,r):n}}function kYe(e,t){let n=0;for(;++n<=e.length;)if((n===e.length||e[n][1].type==="lineEnding")&&e[n-1][1].type==="data"){const r=e[n-1][1],i=t.sliceStream(r);let a=i.length,u=-1,l=0,o;for(;a--;){const s=i[a];if(typeof s=="string"){for(u=s.length;s.charCodeAt(u-1)===32;)l++,u--;if(u)break;u=-1}else if(s===-2)o=!0,l++;else if(s!==-1){a++;break}}if(t._contentTypeTextTrailing&&n===e.length&&(l=0),l){const s={type:n===e.length||o||l<2?"lineSuffix":"hardBreakTrailing",start:{_bufferIndex:a?u:r.start._bufferIndex+u,_index:r.start._index+a,line:r.end.line,column:r.end.column-l,offset:r.end.offset-l},end:{...r.end}};r.end={...s.start},r.start.offset===r.end.offset?Object.assign(r,s):(e.splice(n,0,["enter",s,t],["exit",s,t]),n+=2)}n++}return e}const PYe={42:Gc,43:Gc,45:Gc,48:Gc,49:Gc,50:Gc,51:Gc,52:Gc,53:Gc,54:Gc,55:Gc,56:Gc,57:Gc,62:Obe},MYe={91:LXe},FYe={[-2]:O7,[-1]:O7,32:O7},NYe={35:UXe,42:ED,45:[tW,ED],60:XXe,61:tW,95:ED,96:JK,126:JK},IYe={38:Sbe,92:xbe},LYe={[-5]:x7,[-4]:x7,[-3]:x7,33:dYe,38:Sbe,42:vF,60:[pXe,tYe],91:pYe,92:[qXe,xbe],93:gB,95:vF,96:CXe},BYe={null:[vF,RYe]},QYe={null:[42,95]},zYe={null:[]},qYe=Object.freeze(Object.defineProperty({__proto__:null,attentionMarkers:QYe,contentInitial:MYe,disable:zYe,document:PYe,flow:NYe,flowInitial:FYe,insideSpan:BYe,string:IYe,text:LYe},Symbol.toStringTag,{value:"Module"}));function VYe(e,t,n){let r={_bufferIndex:-1,_index:0,line:n&&n.line||1,column:n&&n.column||1,offset:n&&n.offset||0};const i={},a=[];let u=[],l=[];const o={attempt:R(A),check:R(w),consume:E,enter:$,exit:k,interrupt:R(w,{interrupt:!0})},s={code:null,containerState:{},defineSkip:v,events:[],now:y,parser:e,previous:null,sliceSerialize:h,sliceStream:m,write:d};let c=t.tokenize.call(s,o);return t.resolveAll&&a.push(t),s;function d(q){return u=Zd(u,q),S(),u[u.length-1]!==null?[]:(F(t,0),s.events=h5(a,s.events,s),s.events)}function h(q,V){return jYe(m(q),V)}function m(q){return UYe(u,q)}function y(){const{_bufferIndex:q,_index:V,line:N,column:j,offset:W}=r;return{_bufferIndex:q,_index:V,line:N,column:j,offset:W}}function v(q){i[q.line]=q.column,U()}function S(){let q;for(;r._index<u.length;){const V=u[r._index];if(typeof V=="string")for(q=r._index,r._bufferIndex<0&&(r._bufferIndex=0);r._index===q&&r._bufferIndex<V.length;)_(V.charCodeAt(r._bufferIndex));else _(V)}}function _(q){c=c(q)}function E(q){Pn(q)?(r.line++,r.column=1,r.offset+=q===-3?2:1,U()):q!==-1&&(r.column++,r.offset++),r._bufferIndex<0?r._index++:(r._bufferIndex++,r._bufferIndex===u[r._index].length&&(r._bufferIndex=-1,r._index++)),s.previous=q}function $(q,V){const N=V||{};return N.type=q,N.start=y(),s.events.push(["enter",N,s]),l.push(N),N}function k(q){const V=l.pop();return V.end=y(),s.events.push(["exit",V,s]),V}function A(q,V){F(q,V.from)}function w(q,V){V.restore()}function R(q,V){return N;function N(j,W,K){let J,te,H,M;return Array.isArray(j)?X(j):"tokenize"in j?X([j]):G(j);function G(ie){return de;function de(ge){const ye=ge!==null&&ie[ge],me=ge!==null&&ie.null,Ee=[...Array.isArray(ye)?ye:ye?[ye]:[],...Array.isArray(me)?me:me?[me]:[]];return X(Ee)(ge)}}function X(ie){return J=ie,te=0,ie.length===0?K:B(ie[te])}function B(ie){return de;function de(ge){return M=I(),H=ie,ie.partial||(s.currentConstruct=ie),ie.name&&s.parser.constructs.disable.null.includes(ie.name)?oe():ie.tokenize.call(V?Object.assign(Object.create(s),V):s,o,ne,oe)(ge)}}function ne(ie){return q(H,M),W}function oe(ie){return M.restore(),++te<J.length?B(J[te]):K}}}function F(q,V){q.resolveAll&&!a.includes(q)&&a.push(q),q.resolve&&Hf(s.events,V,s.events.length-V,q.resolve(s.events.slice(V),s)),q.resolveTo&&(s.events=q.resolveTo(s.events,s))}function I(){const q=y(),V=s.previous,N=s.currentConstruct,j=s.events.length,W=Array.from(l);return{from:j,restore:K};function K(){r=q,s.previous=V,s.currentConstruct=N,s.events.length=j,l=W,U()}}function U(){r.line in i&&r.column<2&&(r.column=i[r.line],r.offset+=i[r.line]-1)}}function UYe(e,t){const n=t.start._index,r=t.start._bufferIndex,i=t.end._index,a=t.end._bufferIndex;let u;if(n===i)u=[e[n].slice(r,a)];else{if(u=e.slice(n,i),r>-1){const l=u[0];typeof l=="string"?u[0]=l.slice(r):u.shift()}a>0&&u.push(e[i].slice(0,a))}return u}function jYe(e,t){let n=-1;const r=[];let i;for(;++n<e.length;){const a=e[n];let u;if(typeof a=="string")u=a;else switch(a){case-5:{u="\r";break}case-4:{u=`
`;break}case-3:{u=`\r
`;break}case-2:{u=t?" ":"	";break}case-1:{if(!t&&i)continue;u=" ";break}default:u=String.fromCharCode(a)}i=a===-2,r.push(u)}return r.join("")}function ZYe(e){const r={constructs:ybe([qYe,...(e||{}).extensions||[]]),content:i(lXe),defined:[],document:i(sXe),flow:i(CYe),lazy:{},string:i(AYe),text:i(TYe)};return r;function i(a){return u;function u(l){return VYe(r,a,l)}}}function HYe(e){for(;!_be(e););return e}const nW=/[\0\t\n\r]/g;function XYe(){let e=1,t="",n=!0,r;return i;function i(a,u,l){const o=[];let s,c,d,h,m;for(a=t+(typeof a=="string"?a.toString():new TextDecoder(u||void 0).decode(a)),d=0,t="",n&&(a.charCodeAt(0)===65279&&d++,n=void 0);d<a.length;){if(nW.lastIndex=d,s=nW.exec(a),h=s&&s.index!==void 0?s.index:a.length,m=a.charCodeAt(h),!s){t=a.slice(d);break}if(m===10&&d===h&&r)o.push(-3),r=void 0;else switch(r&&(o.push(-5),r=void 0),d<h&&(o.push(a.slice(d,h)),e+=h-d),m){case 0:{o.push(65533),e++;break}case 9:{for(c=Math.ceil(e/4)*4,o.push(-2);e++<c;)o.push(-1);break}case 10:{o.push(-4),e=1;break}default:r=!0,e=1}d=h+1}return l&&(r&&o.push(-5),t&&o.push(t),o.push(null)),o}}const YYe=/\\([!-/:-@[-`{-~])|&(#(?:\d{1,7}|x[\da-f]{1,6})|[\da-z]{1,31});/gi;function GYe(e){return e.replace(YYe,KYe)}function KYe(e,t,n){if(t)return t;if(n.charCodeAt(0)===35){const i=n.charCodeAt(1),a=i===120||i===88;return bbe(n.slice(a?2:1),a?16:10)}return mB(n)||e}const Rbe={}.hasOwnProperty;function WYe(e,t,n){return typeof t!="string"&&(n=t,t=void 0),JYe(n)(HYe(ZYe(n).document().write(XYe()(e,t,!0))))}function JYe(e){const t={transforms:[],canContainEols:["emphasis","fragment","heading","paragraph","strong"],enter:{autolink:a(tt),autolinkProtocol:I,autolinkEmail:I,atxHeading:a(we),blockQuote:a(me),characterEscape:I,characterReference:I,codeFenced:a(Ee),codeFencedFenceInfo:u,codeFencedFenceMeta:u,codeIndented:a(Ee,u),codeText:a(se,u),codeTextData:I,data:I,codeFlowValue:I,definition:a(pe),definitionDestinationString:u,definitionLabelString:u,definitionTitleString:u,emphasis:a(ve),hardBreakEscape:a(Me),hardBreakTrailing:a(Me),htmlFlow:a(Re,u),htmlFlowData:I,htmlText:a(Re,u),htmlTextData:I,image:a(et),label:u,link:a(tt),listItem:a($e),listItemValue:h,listOrdered:a(Oe,d),listUnordered:a(Oe),paragraph:a(Ae),reference:B,referenceString:u,resourceDestinationString:u,resourceTitleString:u,setextHeading:a(we),strong:a(He),thematicBreak:a(Ye)},exit:{atxHeading:o(),atxHeadingSequence:A,autolink:o(),autolinkEmail:ye,autolinkProtocol:ge,blockQuote:o(),characterEscapeValue:U,characterReferenceMarkerHexadecimal:oe,characterReferenceMarkerNumeric:oe,characterReferenceValue:ie,characterReference:de,codeFenced:o(S),codeFencedFence:v,codeFencedFenceInfo:m,codeFencedFenceMeta:y,codeFlowValue:U,codeIndented:o(_),codeText:o(W),codeTextData:U,data:U,definition:o(),definitionDestinationString:k,definitionLabelString:E,definitionTitleString:$,emphasis:o(),hardBreakEscape:o(V),hardBreakTrailing:o(V),htmlFlow:o(N),htmlFlowData:U,htmlText:o(j),htmlTextData:U,image:o(J),label:H,labelText:te,lineEnding:q,link:o(K),listItem:o(),listOrdered:o(),listUnordered:o(),paragraph:o(),referenceString:ne,resourceDestinationString:M,resourceTitleString:G,resource:X,setextHeading:o(F),setextHeadingLineSequence:R,setextHeadingText:w,strong:o(),thematicBreak:o()}};Abe(t,(e||{}).mdastExtensions||[]);const n={};return r;function r(Ne){let ze={type:"root",children:[]};const at={stack:[ze],tokenStack:[],config:t,enter:l,exit:s,buffer:u,resume:c,data:n},mt=[];let Mt=-1;for(;++Mt<Ne.length;)if(Ne[Mt][1].type==="listOrdered"||Ne[Mt][1].type==="listUnordered")if(Ne[Mt][0]==="enter")mt.push(Mt);else{const bt=mt.pop();Mt=i(Ne,bt,Mt)}for(Mt=-1;++Mt<Ne.length;){const bt=t[Ne[Mt][0]];Rbe.call(bt,Ne[Mt][1].type)&&bt[Ne[Mt][1].type].call(Object.assign({sliceSerialize:Ne[Mt][2].sliceSerialize},at),Ne[Mt][1])}if(at.tokenStack.length>0){const bt=at.tokenStack[at.tokenStack.length-1];(bt[1]||rW).call(at,void 0,bt[0])}for(ze.position={start:mv(Ne.length>0?Ne[0][1].start:{line:1,column:1,offset:0}),end:mv(Ne.length>0?Ne[Ne.length-2][1].end:{line:1,column:1,offset:0})},Mt=-1;++Mt<t.transforms.length;)ze=t.transforms[Mt](ze)||ze;return ze}function i(Ne,ze,at){let mt=ze-1,Mt=-1,bt=!1,zt,qt,Xt,nn;for(;++mt<=at;){const Fe=Ne[mt];switch(Fe[1].type){case"listUnordered":case"listOrdered":case"blockQuote":{Fe[0]==="enter"?Mt++:Mt--,nn=void 0;break}case"lineEndingBlank":{Fe[0]==="enter"&&(zt&&!nn&&!Mt&&!Xt&&(Xt=mt),nn=void 0);break}case"linePrefix":case"listItemValue":case"listItemMarker":case"listItemPrefix":case"listItemPrefixWhitespace":break;default:nn=void 0}if(!Mt&&Fe[0]==="enter"&&Fe[1].type==="listItemPrefix"||Mt===-1&&Fe[0]==="exit"&&(Fe[1].type==="listUnordered"||Fe[1].type==="listOrdered")){if(zt){let ht=mt;for(qt=void 0;ht--;){const kt=Ne[ht];if(kt[1].type==="lineEnding"||kt[1].type==="lineEndingBlank"){if(kt[0]==="exit")continue;qt&&(Ne[qt][1].type="lineEndingBlank",bt=!0),kt[1].type="lineEnding",qt=ht}else if(!(kt[1].type==="linePrefix"||kt[1].type==="blockQuotePrefix"||kt[1].type==="blockQuotePrefixWhitespace"||kt[1].type==="blockQuoteMarker"||kt[1].type==="listItemIndent"))break}Xt&&(!qt||Xt<qt)&&(zt._spread=!0),zt.end=Object.assign({},qt?Ne[qt][1].start:Fe[1].end),Ne.splice(qt||mt,0,["exit",zt,Fe[2]]),mt++,at++}if(Fe[1].type==="listItemPrefix"){const ht={type:"listItem",_spread:!1,start:Object.assign({},Fe[1].start),end:void 0};zt=ht,Ne.splice(mt,0,["enter",ht,Fe[2]]),mt++,at++,Xt=void 0,nn=!0}}}return Ne[ze][1]._spread=bt,at}function a(Ne,ze){return at;function at(mt){l.call(this,Ne(mt),mt),ze&&ze.call(this,mt)}}function u(){this.stack.push({type:"fragment",children:[]})}function l(Ne,ze,at){this.stack[this.stack.length-1].children.push(Ne),this.stack.push(Ne),this.tokenStack.push([ze,at||void 0]),Ne.position={start:mv(ze.start),end:void 0}}function o(Ne){return ze;function ze(at){Ne&&Ne.call(this,at),s.call(this,at)}}function s(Ne,ze){const at=this.stack.pop(),mt=this.tokenStack.pop();if(mt)mt[0].type!==Ne.type&&(ze?ze.call(this,Ne,mt[0]):(mt[1]||rW).call(this,Ne,mt[0]));else throw new Error("Cannot close `"+Ne.type+"` ("+tx({start:Ne.start,end:Ne.end})+"): it’s not open");at.position.end=mv(Ne.end)}function c(){return pB(this.stack.pop())}function d(){this.data.expectingFirstListItemValue=!0}function h(Ne){if(this.data.expectingFirstListItemValue){const ze=this.stack[this.stack.length-2];ze.start=Number.parseInt(this.sliceSerialize(Ne),10),this.data.expectingFirstListItemValue=void 0}}function m(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.lang=Ne}function y(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.meta=Ne}function v(){this.data.flowCodeInside||(this.buffer(),this.data.flowCodeInside=!0)}function S(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.value=Ne.replace(/^(\r?\n|\r)|(\r?\n|\r)$/g,""),this.data.flowCodeInside=void 0}function _(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.value=Ne.replace(/(\r?\n|\r)$/g,"")}function E(Ne){const ze=this.resume(),at=this.stack[this.stack.length-1];at.label=ze,at.identifier=up(this.sliceSerialize(Ne)).toLowerCase()}function $(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.title=Ne}function k(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.url=Ne}function A(Ne){const ze=this.stack[this.stack.length-1];if(!ze.depth){const at=this.sliceSerialize(Ne).length;ze.depth=at}}function w(){this.data.setextHeadingSlurpLineEnding=!0}function R(Ne){const ze=this.stack[this.stack.length-1];ze.depth=this.sliceSerialize(Ne).codePointAt(0)===61?1:2}function F(){this.data.setextHeadingSlurpLineEnding=void 0}function I(Ne){const at=this.stack[this.stack.length-1].children;let mt=at[at.length-1];(!mt||mt.type!=="text")&&(mt=je(),mt.position={start:mv(Ne.start),end:void 0},at.push(mt)),this.stack.push(mt)}function U(Ne){const ze=this.stack.pop();ze.value+=this.sliceSerialize(Ne),ze.position.end=mv(Ne.end)}function q(Ne){const ze=this.stack[this.stack.length-1];if(this.data.atHardBreak){const at=ze.children[ze.children.length-1];at.position.end=mv(Ne.end),this.data.atHardBreak=void 0;return}!this.data.setextHeadingSlurpLineEnding&&t.canContainEols.includes(ze.type)&&(I.call(this,Ne),U.call(this,Ne))}function V(){this.data.atHardBreak=!0}function N(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.value=Ne}function j(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.value=Ne}function W(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.value=Ne}function K(){const Ne=this.stack[this.stack.length-1];if(this.data.inReference){const ze=this.data.referenceType||"shortcut";Ne.type+="Reference",Ne.referenceType=ze,delete Ne.url,delete Ne.title}else delete Ne.identifier,delete Ne.label;this.data.referenceType=void 0}function J(){const Ne=this.stack[this.stack.length-1];if(this.data.inReference){const ze=this.data.referenceType||"shortcut";Ne.type+="Reference",Ne.referenceType=ze,delete Ne.url,delete Ne.title}else delete Ne.identifier,delete Ne.label;this.data.referenceType=void 0}function te(Ne){const ze=this.sliceSerialize(Ne),at=this.stack[this.stack.length-2];at.label=GYe(ze),at.identifier=up(ze).toLowerCase()}function H(){const Ne=this.stack[this.stack.length-1],ze=this.resume(),at=this.stack[this.stack.length-1];if(this.data.inReference=!0,at.type==="link"){const mt=Ne.children;at.children=mt}else at.alt=ze}function M(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.url=Ne}function G(){const Ne=this.resume(),ze=this.stack[this.stack.length-1];ze.title=Ne}function X(){this.data.inReference=void 0}function B(){this.data.referenceType="collapsed"}function ne(Ne){const ze=this.resume(),at=this.stack[this.stack.length-1];at.label=ze,at.identifier=up(this.sliceSerialize(Ne)).toLowerCase(),this.data.referenceType="full"}function oe(Ne){this.data.characterReferenceType=Ne.type}function ie(Ne){const ze=this.sliceSerialize(Ne),at=this.data.characterReferenceType;let mt;at?(mt=bbe(ze,at==="characterReferenceMarkerNumeric"?10:16),this.data.characterReferenceType=void 0):mt=mB(ze);const Mt=this.stack[this.stack.length-1];Mt.value+=mt}function de(Ne){const ze=this.stack.pop();ze.position.end=mv(Ne.end)}function ge(Ne){U.call(this,Ne);const ze=this.stack[this.stack.length-1];ze.url=this.sliceSerialize(Ne)}function ye(Ne){U.call(this,Ne);const ze=this.stack[this.stack.length-1];ze.url="mailto:"+this.sliceSerialize(Ne)}function me(){return{type:"blockquote",children:[]}}function Ee(){return{type:"code",lang:null,meta:null,value:""}}function se(){return{type:"inlineCode",value:""}}function pe(){return{type:"definition",identifier:"",label:null,title:null,url:""}}function ve(){return{type:"emphasis",children:[]}}function we(){return{type:"heading",depth:0,children:[]}}function Me(){return{type:"break"}}function Re(){return{type:"html",value:""}}function et(){return{type:"image",title:null,url:"",alt:null}}function tt(){return{type:"link",title:null,url:"",children:[]}}function Oe(Ne){return{type:"list",ordered:Ne.type==="listOrdered",start:null,spread:Ne._spread,children:[]}}function $e(Ne){return{type:"listItem",spread:Ne._spread,checked:null,children:[]}}function Ae(){return{type:"paragraph",children:[]}}function He(){return{type:"strong",children:[]}}function je(){return{type:"text",value:""}}function Ye(){return{type:"thematicBreak"}}}function mv(e){return{line:e.line,column:e.column,offset:e.offset}}function Abe(e,t){let n=-1;for(;++n<t.length;){const r=t[n];Array.isArray(r)?Abe(e,r):eGe(e,r)}}function eGe(e,t){let n;for(n in t)if(Rbe.call(t,n))switch(n){case"canContainEols":{const r=t[n];r&&e[n].push(...r);break}case"transforms":{const r=t[n];r&&e[n].push(...r);break}case"enter":case"exit":{const r=t[n];r&&Object.assign(e[n],r);break}}}function rW(e,t){throw e?new Error("Cannot close `"+e.type+"` ("+tx({start:e.start,end:e.end})+"): a different token (`"+t.type+"`, "+tx({start:t.start,end:t.end})+") is open"):new Error("Cannot close document, a token (`"+t.type+"`, "+tx({start:t.start,end:t.end})+") is still open")}function tGe(e){const t=this;t.parser=n;function n(r){return WYe(r,{...t.data("settings"),...e,extensions:t.data("micromarkExtensions")||[],mdastExtensions:t.data("fromMarkdownExtensions")||[]})}}function nGe(e,t){const n={type:"element",tagName:"blockquote",properties:{},children:e.wrap(e.all(t),!0)};return e.patch(t,n),e.applyData(t,n)}function rGe(e,t){const n={type:"element",tagName:"br",properties:{},children:[]};return e.patch(t,n),[e.applyData(t,n),{type:"text",value:`
`}]}function iGe(e,t){const n=t.value?t.value+`
`:"",r={};t.lang&&(r.className=["language-"+t.lang]);let i={type:"element",tagName:"code",properties:r,children:[{type:"text",value:n}]};return t.meta&&(i.data={meta:t.meta}),e.patch(t,i),i=e.applyData(t,i),i={type:"element",tagName:"pre",properties:{},children:[i]},e.patch(t,i),i}function aGe(e,t){const n={type:"element",tagName:"del",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function uGe(e,t){const n={type:"element",tagName:"em",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function lGe(e,t){const n=typeof e.options.clobberPrefix=="string"?e.options.clobberPrefix:"user-content-",r=String(t.identifier).toUpperCase(),i=m4(r.toLowerCase()),a=e.footnoteOrder.indexOf(r);let u,l=e.footnoteCounts.get(r);l===void 0?(l=0,e.footnoteOrder.push(r),u=e.footnoteOrder.length):u=a+1,l+=1,e.footnoteCounts.set(r,l);const o={type:"element",tagName:"a",properties:{href:"#"+n+"fn-"+i,id:n+"fnref-"+i+(l>1?"-"+l:""),dataFootnoteRef:!0,ariaDescribedBy:["footnote-label"]},children:[{type:"text",value:String(u)}]};e.patch(t,o);const s={type:"element",tagName:"sup",properties:{},children:[o]};return e.patch(t,s),e.applyData(t,s)}function oGe(e,t){const n={type:"element",tagName:"h"+t.depth,properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function sGe(e,t){if(e.options.allowDangerousHtml){const n={type:"raw",value:t.value};return e.patch(t,n),e.applyData(t,n)}}function Tbe(e,t){const n=t.referenceType;let r="]";if(n==="collapsed"?r+="[]":n==="full"&&(r+="["+(t.label||t.identifier)+"]"),t.type==="imageReference")return[{type:"text",value:"!["+t.alt+r}];const i=e.all(t),a=i[0];a&&a.type==="text"?a.value="["+a.value:i.unshift({type:"text",value:"["});const u=i[i.length-1];return u&&u.type==="text"?u.value+=r:i.push({type:"text",value:r}),i}function cGe(e,t){const n=String(t.identifier).toUpperCase(),r=e.definitionById.get(n);if(!r)return Tbe(e,t);const i={src:m4(r.url||""),alt:t.alt};r.title!==null&&r.title!==void 0&&(i.title=r.title);const a={type:"element",tagName:"img",properties:i,children:[]};return e.patch(t,a),e.applyData(t,a)}function fGe(e,t){const n={src:m4(t.url)};t.alt!==null&&t.alt!==void 0&&(n.alt=t.alt),t.title!==null&&t.title!==void 0&&(n.title=t.title);const r={type:"element",tagName:"img",properties:n,children:[]};return e.patch(t,r),e.applyData(t,r)}function dGe(e,t){const n={type:"text",value:t.value.replace(/\r?\n|\r/g," ")};e.patch(t,n);const r={type:"element",tagName:"code",properties:{},children:[n]};return e.patch(t,r),e.applyData(t,r)}function hGe(e,t){const n=String(t.identifier).toUpperCase(),r=e.definitionById.get(n);if(!r)return Tbe(e,t);const i={href:m4(r.url||"")};r.title!==null&&r.title!==void 0&&(i.title=r.title);const a={type:"element",tagName:"a",properties:i,children:e.all(t)};return e.patch(t,a),e.applyData(t,a)}function pGe(e,t){const n={href:m4(t.url)};t.title!==null&&t.title!==void 0&&(n.title=t.title);const r={type:"element",tagName:"a",properties:n,children:e.all(t)};return e.patch(t,r),e.applyData(t,r)}function mGe(e,t,n){const r=e.all(t),i=n?gGe(n):kbe(t),a={},u=[];if(typeof t.checked=="boolean"){const c=r[0];let d;c&&c.type==="element"&&c.tagName==="p"?d=c:(d={type:"element",tagName:"p",properties:{},children:[]},r.unshift(d)),d.children.length>0&&d.children.unshift({type:"text",value:" "}),d.children.unshift({type:"element",tagName:"input",properties:{type:"checkbox",checked:t.checked,disabled:!0},children:[]}),a.className=["task-list-item"]}let l=-1;for(;++l<r.length;){const c=r[l];(i||l!==0||c.type!=="element"||c.tagName!=="p")&&u.push({type:"text",value:`
`}),c.type==="element"&&c.tagName==="p"&&!i?u.push(...c.children):u.push(c)}const o=r[r.length-1];o&&(i||o.type!=="element"||o.tagName!=="p")&&u.push({type:"text",value:`
`});const s={type:"element",tagName:"li",properties:a,children:u};return e.patch(t,s),e.applyData(t,s)}function gGe(e){let t=!1;if(e.type==="list"){t=e.spread||!1;const n=e.children;let r=-1;for(;!t&&++r<n.length;)t=kbe(n[r])}return t}function kbe(e){const t=e.spread;return t??e.children.length>1}function vGe(e,t){const n={},r=e.all(t);let i=-1;for(typeof t.start=="number"&&t.start!==1&&(n.start=t.start);++i<r.length;){const u=r[i];if(u.type==="element"&&u.tagName==="li"&&u.properties&&Array.isArray(u.properties.className)&&u.properties.className.includes("task-list-item")){n.className=["contains-task-list"];break}}const a={type:"element",tagName:t.ordered?"ol":"ul",properties:n,children:e.wrap(r,!0)};return e.patch(t,a),e.applyData(t,a)}function yGe(e,t){const n={type:"element",tagName:"p",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function bGe(e,t){const n={type:"root",children:e.wrap(e.all(t))};return e.patch(t,n),e.applyData(t,n)}function OGe(e,t){const n={type:"element",tagName:"strong",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function xGe(e,t){const n=e.all(t),r=n.shift(),i=[];if(r){const u={type:"element",tagName:"thead",properties:{},children:e.wrap([r],!0)};e.patch(t.children[0],u),i.push(u)}if(n.length>0){const u={type:"element",tagName:"tbody",properties:{},children:e.wrap(n,!0)},l=cB(t.children[1]),o=fbe(t.children[t.children.length-1]);l&&o&&(u.position={start:l,end:o}),i.push(u)}const a={type:"element",tagName:"table",properties:{},children:e.wrap(i,!0)};return e.patch(t,a),e.applyData(t,a)}function SGe(e,t,n){const r=n?n.children:void 0,a=(r?r.indexOf(t):1)===0?"th":"td",u=n&&n.type==="table"?n.align:void 0,l=u?u.length:t.children.length;let o=-1;const s=[];for(;++o<l;){const d=t.children[o],h={},m=u?u[o]:void 0;m&&(h.align=m);let y={type:"element",tagName:a,properties:h,children:[]};d&&(y.children=e.all(d),e.patch(d,y),y=e.applyData(d,y)),s.push(y)}const c={type:"element",tagName:"tr",properties:{},children:e.wrap(s,!0)};return e.patch(t,c),e.applyData(t,c)}function _Ge(e,t){const n={type:"element",tagName:"td",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}const iW=9,aW=32;function EGe(e){const t=String(e),n=/\r?\n|\r/g;let r=n.exec(t),i=0;const a=[];for(;r;)a.push(uW(t.slice(i,r.index),i>0,!0),r[0]),i=r.index+r[0].length,r=n.exec(t);return a.push(uW(t.slice(i),i>0,!1)),a.join("")}function uW(e,t,n){let r=0,i=e.length;if(t){let a=e.codePointAt(r);for(;a===iW||a===aW;)r++,a=e.codePointAt(r)}if(n){let a=e.codePointAt(i-1);for(;a===iW||a===aW;)i--,a=e.codePointAt(i-1)}return i>r?e.slice(r,i):""}function DGe(e,t){const n={type:"text",value:EGe(String(t.value))};return e.patch(t,n),e.applyData(t,n)}function $Ge(e,t){const n={type:"element",tagName:"hr",properties:{},children:[]};return e.patch(t,n),e.applyData(t,n)}const CGe={blockquote:nGe,break:rGe,code:iGe,delete:aGe,emphasis:uGe,footnoteReference:lGe,heading:oGe,html:sGe,imageReference:cGe,image:fGe,inlineCode:dGe,linkReference:hGe,link:pGe,listItem:mGe,list:vGe,paragraph:yGe,root:bGe,strong:OGe,table:xGe,tableCell:_Ge,tableRow:SGe,text:DGe,thematicBreak:$Ge,toml:FE,yaml:FE,definition:FE,footnoteDefinition:FE};function FE(){}const Pbe=-1,p5=0,rx=1,oC=2,vB=3,yB=4,bB=5,OB=6,Mbe=7,Fbe=8,lW=typeof self=="object"?self:globalThis,wGe=(e,t)=>{const n=(i,a)=>(e.set(a,i),i),r=i=>{if(e.has(i))return e.get(i);const[a,u]=t[i];switch(a){case p5:case Pbe:return n(u,i);case rx:{const l=n([],i);for(const o of u)l.push(r(o));return l}case oC:{const l=n({},i);for(const[o,s]of u)l[r(o)]=r(s);return l}case vB:return n(new Date(u),i);case yB:{const{source:l,flags:o}=u;return n(new RegExp(l,o),i)}case bB:{const l=n(new Map,i);for(const[o,s]of u)l.set(r(o),r(s));return l}case OB:{const l=n(new Set,i);for(const o of u)l.add(r(o));return l}case Mbe:{const{name:l,message:o}=u;return n(new lW[l](o),i)}case Fbe:return n(BigInt(u),i);case"BigInt":return n(Object(BigInt(u)),i);case"ArrayBuffer":return n(new Uint8Array(u).buffer,u);case"DataView":{const{buffer:l}=new Uint8Array(u);return n(new DataView(l),u)}}return n(new lW[a](u),i)};return r},oW=e=>wGe(new Map,e)(0),kO="",{toString:RGe}={},{keys:AGe}=Object,w2=e=>{const t=typeof e;if(t!=="object"||!e)return[p5,t];const n=RGe.call(e).slice(8,-1);switch(n){case"Array":return[rx,kO];case"Object":return[oC,kO];case"Date":return[vB,kO];case"RegExp":return[yB,kO];case"Map":return[bB,kO];case"Set":return[OB,kO];case"DataView":return[rx,n]}return n.includes("Array")?[rx,n]:n.includes("Error")?[Mbe,n]:[oC,n]},NE=([e,t])=>e===p5&&(t==="function"||t==="symbol"),TGe=(e,t,n,r)=>{const i=(u,l)=>{const o=r.push(u)-1;return n.set(l,o),o},a=u=>{if(n.has(u))return n.get(u);let[l,o]=w2(u);switch(l){case p5:{let c=u;switch(o){case"bigint":l=Fbe,c=u.toString();break;case"function":case"symbol":if(e)throw new TypeError("unable to serialize "+o);c=null;break;case"undefined":return i([Pbe],u)}return i([l,c],u)}case rx:{if(o){let h=u;return o==="DataView"?h=new Uint8Array(u.buffer):o==="ArrayBuffer"&&(h=new Uint8Array(u)),i([o,[...h]],u)}const c=[],d=i([l,c],u);for(const h of u)c.push(a(h));return d}case oC:{if(o)switch(o){case"BigInt":return i([o,u.toString()],u);case"Boolean":case"Number":case"String":return i([o,u.valueOf()],u)}if(t&&"toJSON"in u)return a(u.toJSON());const c=[],d=i([l,c],u);for(const h of AGe(u))(e||!NE(w2(u[h])))&&c.push([a(h),a(u[h])]);return d}case vB:return i([l,u.toISOString()],u);case yB:{const{source:c,flags:d}=u;return i([l,{source:c,flags:d}],u)}case bB:{const c=[],d=i([l,c],u);for(const[h,m]of u)(e||!(NE(w2(h))||NE(w2(m))))&&c.push([a(h),a(m)]);return d}case OB:{const c=[],d=i([l,c],u);for(const h of u)(e||!NE(w2(h)))&&c.push(a(h));return d}}const{message:s}=u;return i([l,{name:o,message:s}],u)};return a},sW=(e,{json:t,lossy:n}={})=>{const r=[];return TGe(!(t||n),!!t,new Map,r)(e),r},sC=typeof structuredClone=="function"?(e,t)=>t&&("json"in t||"lossy"in t)?oW(sW(e,t)):structuredClone(e):(e,t)=>oW(sW(e,t));function kGe(e,t){const n=[{type:"text",value:"↩"}];return t>1&&n.push({type:"element",tagName:"sup",properties:{},children:[{type:"text",value:String(t)}]}),n}function PGe(e,t){return"Back to reference "+(e+1)+(t>1?"-"+t:"")}function MGe(e){const t=typeof e.options.clobberPrefix=="string"?e.options.clobberPrefix:"user-content-",n=e.options.footnoteBackContent||kGe,r=e.options.footnoteBackLabel||PGe,i=e.options.footnoteLabel||"Footnotes",a=e.options.footnoteLabelTagName||"h2",u=e.options.footnoteLabelProperties||{className:["sr-only"]},l=[];let o=-1;for(;++o<e.footnoteOrder.length;){const s=e.footnoteById.get(e.footnoteOrder[o]);if(!s)continue;const c=e.all(s),d=String(s.identifier).toUpperCase(),h=m4(d.toLowerCase());let m=0;const y=[],v=e.footnoteCounts.get(d);for(;v!==void 0&&++m<=v;){y.length>0&&y.push({type:"text",value:" "});let E=typeof n=="string"?n:n(o,m);typeof E=="string"&&(E={type:"text",value:E}),y.push({type:"element",tagName:"a",properties:{href:"#"+t+"fnref-"+h+(m>1?"-"+m:""),dataFootnoteBackref:"",ariaLabel:typeof r=="string"?r:r(o,m),className:["data-footnote-backref"]},children:Array.isArray(E)?E:[E]})}const S=c[c.length-1];if(S&&S.type==="element"&&S.tagName==="p"){const E=S.children[S.children.length-1];E&&E.type==="text"?E.value+=" ":S.children.push({type:"text",value:" "}),S.children.push(...y)}else c.push(...y);const _={type:"element",tagName:"li",properties:{id:t+"fn-"+h},children:e.wrap(c,!0)};e.patch(s,_),l.push(_)}if(l.length!==0)return{type:"element",tagName:"section",properties:{dataFootnotes:!0,className:["footnotes"]},children:[{type:"element",tagName:a,properties:{...sC(u),id:"footnote-label"},children:[{type:"text",value:i}]},{type:"text",value:`
`},{type:"element",tagName:"ol",properties:{},children:e.wrap(l,!0)},{type:"text",value:`
`}]}}const m5=(function(e){if(e==null)return LGe;if(typeof e=="function")return g5(e);if(typeof e=="object")return Array.isArray(e)?FGe(e):NGe(e);if(typeof e=="string")return IGe(e);throw new Error("Expected function, string, or object as test")});function FGe(e){const t=[];let n=-1;for(;++n<e.length;)t[n]=m5(e[n]);return g5(r);function r(...i){let a=-1;for(;++a<t.length;)if(t[a].apply(this,i))return!0;return!1}}function NGe(e){const t=e;return g5(n);function n(r){const i=r;let a;for(a in e)if(i[a]!==t[a])return!1;return!0}}function IGe(e){return g5(t);function t(n){return n&&n.type===e}}function g5(e){return t;function t(n,r,i){return!!(BGe(n)&&e.call(this,n,typeof r=="number"?r:void 0,i||void 0))}}function LGe(){return!0}function BGe(e){return e!==null&&typeof e=="object"&&"type"in e}const Nbe=[],QGe=!0,yF=!1,zGe="skip";function Ibe(e,t,n,r){let i;typeof t=="function"&&typeof n!="function"?(r=n,n=t):i=t;const a=m5(i),u=r?-1:1;l(e,void 0,[])();function l(o,s,c){const d=o&&typeof o=="object"?o:{};if(typeof d.type=="string"){const m=typeof d.tagName=="string"?d.tagName:typeof d.name=="string"?d.name:void 0;Object.defineProperty(h,"name",{value:"node ("+(o.type+(m?"<"+m+">":""))+")"})}return h;function h(){let m=Nbe,y,v,S;if((!t||a(o,s,c[c.length-1]||void 0))&&(m=qGe(n(o,c)),m[0]===yF))return m;if("children"in o&&o.children){const _=o;if(_.children&&m[0]!==zGe)for(v=(r?_.children.length:-1)+u,S=c.concat(_);v>-1&&v<_.children.length;){const E=_.children[v];if(y=l(E,v,S)(),y[0]===yF)return y;v=typeof y[1]=="number"?y[1]:v+u}}return m}}}function qGe(e){return Array.isArray(e)?e:typeof e=="number"?[QGe,e]:e==null?Nbe:[e]}function xB(e,t,n,r){let i,a,u;typeof t=="function"&&typeof n!="function"?(a=void 0,u=t,i=n):(a=t,u=n,i=r),Ibe(e,a,l,i);function l(o,s){const c=s[s.length-1],d=c?c.children.indexOf(o):void 0;return u(o,d,c)}}const bF={}.hasOwnProperty,VGe={};function UGe(e,t){const n=t||VGe,r=new Map,i=new Map,a=new Map,u={...CGe,...n.handlers},l={all:s,applyData:ZGe,definitionById:r,footnoteById:i,footnoteCounts:a,footnoteOrder:[],handlers:u,one:o,options:n,patch:jGe,wrap:XGe};return xB(e,function(c){if(c.type==="definition"||c.type==="footnoteDefinition"){const d=c.type==="definition"?r:i,h=String(c.identifier).toUpperCase();d.has(h)||d.set(h,c)}}),l;function o(c,d){const h=c.type,m=l.handlers[h];if(bF.call(l.handlers,h)&&m)return m(l,c,d);if(l.options.passThrough&&l.options.passThrough.includes(h)){if("children"in c){const{children:v,...S}=c,_=sC(S);return _.children=l.all(c),_}return sC(c)}return(l.options.unknownHandler||HGe)(l,c,d)}function s(c){const d=[];if("children"in c){const h=c.children;let m=-1;for(;++m<h.length;){const y=l.one(h[m],c);if(y){if(m&&h[m-1].type==="break"&&(!Array.isArray(y)&&y.type==="text"&&(y.value=cW(y.value)),!Array.isArray(y)&&y.type==="element")){const v=y.children[0];v&&v.type==="text"&&(v.value=cW(v.value))}Array.isArray(y)?d.push(...y):d.push(y)}}}return d}}function jGe(e,t){e.position&&(t.position=PHe(e))}function ZGe(e,t){let n=t;if(e&&e.data){const r=e.data.hName,i=e.data.hChildren,a=e.data.hProperties;if(typeof r=="string")if(n.type==="element")n.tagName=r;else{const u="children"in n?n.children:[n];n={type:"element",tagName:r,properties:{},children:u}}n.type==="element"&&a&&Object.assign(n.properties,sC(a)),"children"in n&&n.children&&i!==null&&i!==void 0&&(n.children=i)}return n}function HGe(e,t){const n=t.data||{},r="value"in t&&!(bF.call(n,"hProperties")||bF.call(n,"hChildren"))?{type:"text",value:t.value}:{type:"element",tagName:"div",properties:{},children:e.all(t)};return e.patch(t,r),e.applyData(t,r)}function XGe(e,t){const n=[];let r=-1;for(t&&n.push({type:"text",value:`
`});++r<e.length;)r&&n.push({type:"text",value:`
`}),n.push(e[r]);return t&&e.length>0&&n.push({type:"text",value:`
`}),n}function cW(e){let t=0,n=e.charCodeAt(t);for(;n===9||n===32;)t++,n=e.charCodeAt(t);return e.slice(t)}function fW(e,t){const n=UGe(e,t),r=n.one(e,void 0),i=MGe(n),a=Array.isArray(r)?{type:"root",children:r}:r||{type:"root",children:[]};return i&&a.children.push({type:"text",value:`
`},i),a}function YGe(e,t){return e&&"run"in e?async function(n,r){const i=fW(n,{file:r,...t});await e.run(i,r)}:function(n,r){return fW(n,{file:r,...e||t})}}function dW(e){if(e)throw e}var S7,hW;function GGe(){if(hW)return S7;hW=1;var e=Object.prototype.hasOwnProperty,t=Object.prototype.toString,n=Object.defineProperty,r=Object.getOwnPropertyDescriptor,i=function(s){return typeof Array.isArray=="function"?Array.isArray(s):t.call(s)==="[object Array]"},a=function(s){if(!s||t.call(s)!=="[object Object]")return!1;var c=e.call(s,"constructor"),d=s.constructor&&s.constructor.prototype&&e.call(s.constructor.prototype,"isPrototypeOf");if(s.constructor&&!c&&!d)return!1;var h;for(h in s);return typeof h>"u"||e.call(s,h)},u=function(s,c){n&&c.name==="__proto__"?n(s,c.name,{enumerable:!0,configurable:!0,value:c.newValue,writable:!0}):s[c.name]=c.newValue},l=function(s,c){if(c==="__proto__")if(e.call(s,c)){if(r)return r(s,c).value}else return;return s[c]};return S7=function o(){var s,c,d,h,m,y,v=arguments[0],S=1,_=arguments.length,E=!1;for(typeof v=="boolean"&&(E=v,v=arguments[1]||{},S=2),(v==null||typeof v!="object"&&typeof v!="function")&&(v={});S<_;++S)if(s=arguments[S],s!=null)for(c in s)d=l(v,c),h=l(s,c),v!==h&&(E&&h&&(a(h)||(m=i(h)))?(m?(m=!1,y=d&&i(d)?d:[]):y=d&&a(d)?d:{},u(v,{name:c,newValue:o(E,y,h)})):typeof h<"u"&&u(v,{name:c,newValue:h}));return v},S7}var KGe=GGe();const _7=da(KGe);function OF(e){if(typeof e!="object"||e===null)return!1;const t=Object.getPrototypeOf(e);return(t===null||t===Object.prototype||Object.getPrototypeOf(t)===null)&&!(Symbol.toStringTag in e)&&!(Symbol.iterator in e)}function WGe(){const e=[],t={run:n,use:r};return t;function n(...i){let a=-1;const u=i.pop();if(typeof u!="function")throw new TypeError("Expected function as last argument, not "+u);l(null,...i);function l(o,...s){const c=e[++a];let d=-1;if(o){u(o);return}for(;++d<i.length;)(s[d]===null||s[d]===void 0)&&(s[d]=i[d]);i=s,c?JGe(c,l)(...s):u(null,...s)}}function r(i){if(typeof i!="function")throw new TypeError("Expected `middelware` to be a function, not "+i);return e.push(i),t}}function JGe(e,t){let n;return r;function r(...u){const l=e.length>u.length;let o;l&&u.push(i);try{o=e.apply(this,u)}catch(s){const c=s;if(l&&n)throw c;return i(c)}l||(o&&o.then&&typeof o.then=="function"?o.then(a,i):o instanceof Error?i(o):a(o))}function i(u,...l){n||(n=!0,t(u,...l))}function a(u){i(null,u)}}const d0={basename:eKe,dirname:tKe,extname:nKe,join:rKe,sep:"/"};function eKe(e,t){if(t!==void 0&&typeof t!="string")throw new TypeError('"ext" argument must be a string');VS(e);let n=0,r=-1,i=e.length,a;if(t===void 0||t.length===0||t.length>e.length){for(;i--;)if(e.codePointAt(i)===47){if(a){n=i+1;break}}else r<0&&(a=!0,r=i+1);return r<0?"":e.slice(n,r)}if(t===e)return"";let u=-1,l=t.length-1;for(;i--;)if(e.codePointAt(i)===47){if(a){n=i+1;break}}else u<0&&(a=!0,u=i+1),l>-1&&(e.codePointAt(i)===t.codePointAt(l--)?l<0&&(r=i):(l=-1,r=u));return n===r?r=u:r<0&&(r=e.length),e.slice(n,r)}function tKe(e){if(VS(e),e.length===0)return".";let t=-1,n=e.length,r;for(;--n;)if(e.codePointAt(n)===47){if(r){t=n;break}}else r||(r=!0);return t<0?e.codePointAt(0)===47?"/":".":t===1&&e.codePointAt(0)===47?"//":e.slice(0,t)}function nKe(e){VS(e);let t=e.length,n=-1,r=0,i=-1,a=0,u;for(;t--;){const l=e.codePointAt(t);if(l===47){if(u){r=t+1;break}continue}n<0&&(u=!0,n=t+1),l===46?i<0?i=t:a!==1&&(a=1):i>-1&&(a=-1)}return i<0||n<0||a===0||a===1&&i===n-1&&i===r+1?"":e.slice(i,n)}function rKe(...e){let t=-1,n;for(;++t<e.length;)VS(e[t]),e[t]&&(n=n===void 0?e[t]:n+"/"+e[t]);return n===void 0?".":iKe(n)}function iKe(e){VS(e);const t=e.codePointAt(0)===47;let n=aKe(e,!t);return n.length===0&&!t&&(n="."),n.length>0&&e.codePointAt(e.length-1)===47&&(n+="/"),t?"/"+n:n}function aKe(e,t){let n="",r=0,i=-1,a=0,u=-1,l,o;for(;++u<=e.length;){if(u<e.length)l=e.codePointAt(u);else{if(l===47)break;l=47}if(l===47){if(!(i===u-1||a===1))if(i!==u-1&&a===2){if(n.length<2||r!==2||n.codePointAt(n.length-1)!==46||n.codePointAt(n.length-2)!==46){if(n.length>2){if(o=n.lastIndexOf("/"),o!==n.length-1){o<0?(n="",r=0):(n=n.slice(0,o),r=n.length-1-n.lastIndexOf("/")),i=u,a=0;continue}}else if(n.length>0){n="",r=0,i=u,a=0;continue}}t&&(n=n.length>0?n+"/..":"..",r=2)}else n.length>0?n+="/"+e.slice(i+1,u):n=e.slice(i+1,u),r=u-i-1;i=u,a=0}else l===46&&a>-1?a++:a=-1}return n}function VS(e){if(typeof e!="string")throw new TypeError("Path must be a string. Received "+JSON.stringify(e))}const uKe={cwd:lKe};function lKe(){return"/"}function xF(e){return!!(e!==null&&typeof e=="object"&&"href"in e&&e.href&&"protocol"in e&&e.protocol&&e.auth===void 0)}function oKe(e){if(typeof e=="string")e=new URL(e);else if(!xF(e)){const t=new TypeError('The "path" argument must be of type string or an instance of URL. Received `'+e+"`");throw t.code="ERR_INVALID_ARG_TYPE",t}if(e.protocol!=="file:"){const t=new TypeError("The URL must be of scheme file");throw t.code="ERR_INVALID_URL_SCHEME",t}return sKe(e)}function sKe(e){if(e.hostname!==""){const r=new TypeError('File URL host must be "localhost" or empty on darwin');throw r.code="ERR_INVALID_FILE_URL_HOST",r}const t=e.pathname;let n=-1;for(;++n<t.length;)if(t.codePointAt(n)===37&&t.codePointAt(n+1)===50){const r=t.codePointAt(n+2);if(r===70||r===102){const i=new TypeError("File URL path must not include encoded / characters");throw i.code="ERR_INVALID_FILE_URL_PATH",i}}return decodeURIComponent(t)}const E7=["history","path","basename","stem","extname","dirname"];class Lbe{constructor(t){let n;t?xF(t)?n={path:t}:typeof t=="string"||cKe(t)?n={value:t}:n=t:n={},this.cwd="cwd"in n?"":uKe.cwd(),this.data={},this.history=[],this.messages=[],this.value,this.map,this.result,this.stored;let r=-1;for(;++r<E7.length;){const a=E7[r];a in n&&n[a]!==void 0&&n[a]!==null&&(this[a]=a==="history"?[...n[a]]:n[a])}let i;for(i in n)E7.includes(i)||(this[i]=n[i])}get basename(){return typeof this.path=="string"?d0.basename(this.path):void 0}set basename(t){$7(t,"basename"),D7(t,"basename"),this.path=d0.join(this.dirname||"",t)}get dirname(){return typeof this.path=="string"?d0.dirname(this.path):void 0}set dirname(t){pW(this.basename,"dirname"),this.path=d0.join(t||"",this.basename)}get extname(){return typeof this.path=="string"?d0.extname(this.path):void 0}set extname(t){if(D7(t,"extname"),pW(this.dirname,"extname"),t){if(t.codePointAt(0)!==46)throw new Error("`extname` must start with `.`");if(t.includes(".",1))throw new Error("`extname` cannot contain multiple dots")}this.path=d0.join(this.dirname,this.stem+(t||""))}get path(){return this.history[this.history.length-1]}set path(t){xF(t)&&(t=oKe(t)),$7(t,"path"),this.path!==t&&this.history.push(t)}get stem(){return typeof this.path=="string"?d0.basename(this.path,this.extname):void 0}set stem(t){$7(t,"stem"),D7(t,"stem"),this.path=d0.join(this.dirname||"",t+(this.extname||""))}fail(t,n,r){const i=this.message(t,n,r);throw i.fatal=!0,i}info(t,n,r){const i=this.message(t,n,r);return i.fatal=void 0,i}message(t,n,r){const i=new Es(t,n,r);return this.path&&(i.name=this.path+":"+i.name,i.file=this.path),i.fatal=!1,this.messages.push(i),i}toString(t){return this.value===void 0?"":typeof this.value=="string"?this.value:new TextDecoder(t||void 0).decode(this.value)}}function D7(e,t){if(e&&e.includes(d0.sep))throw new Error("`"+t+"` cannot be a path: did not expect `"+d0.sep+"`")}function $7(e,t){if(!e)throw new Error("`"+t+"` cannot be empty")}function pW(e,t){if(!e)throw new Error("Setting `"+t+"` requires `path` to be set too")}function cKe(e){return!!(e&&typeof e=="object"&&"byteLength"in e&&"byteOffset"in e)}const fKe=(function(e){const r=this.constructor.prototype,i=r[e],a=function(){return i.apply(a,arguments)};return Object.setPrototypeOf(a,r),a}),dKe={}.hasOwnProperty;class SB extends fKe{constructor(){super("copy"),this.Compiler=void 0,this.Parser=void 0,this.attachers=[],this.compiler=void 0,this.freezeIndex=-1,this.frozen=void 0,this.namespace={},this.parser=void 0,this.transformers=WGe()}copy(){const t=new SB;let n=-1;for(;++n<this.attachers.length;){const r=this.attachers[n];t.use(...r)}return t.data(_7(!0,{},this.namespace)),t}data(t,n){return typeof t=="string"?arguments.length===2?(R7("data",this.frozen),this.namespace[t]=n,this):dKe.call(this.namespace,t)&&this.namespace[t]||void 0:t?(R7("data",this.frozen),this.namespace=t,this):this.namespace}freeze(){if(this.frozen)return this;const t=this;for(;++this.freezeIndex<this.attachers.length;){const[n,...r]=this.attachers[this.freezeIndex];if(r[0]===!1)continue;r[0]===!0&&(r[0]=void 0);const i=n.call(t,...r);typeof i=="function"&&this.transformers.use(i)}return this.frozen=!0,this.freezeIndex=Number.POSITIVE_INFINITY,this}parse(t){this.freeze();const n=IE(t),r=this.parser||this.Parser;return C7("parse",r),r(String(n),n)}process(t,n){const r=this;return this.freeze(),C7("process",this.parser||this.Parser),w7("process",this.compiler||this.Compiler),n?i(void 0,n):new Promise(i);function i(a,u){const l=IE(t),o=r.parse(l);r.run(o,l,function(c,d,h){if(c||!d||!h)return s(c);const m=d,y=r.stringify(m,h);mKe(y)?h.value=y:h.result=y,s(c,h)});function s(c,d){c||!d?u(c):a?a(d):n(void 0,d)}}}processSync(t){let n=!1,r;return this.freeze(),C7("processSync",this.parser||this.Parser),w7("processSync",this.compiler||this.Compiler),this.process(t,i),gW("processSync","process",n),r;function i(a,u){n=!0,dW(a),r=u}}run(t,n,r){mW(t),this.freeze();const i=this.transformers;return!r&&typeof n=="function"&&(r=n,n=void 0),r?a(void 0,r):new Promise(a);function a(u,l){const o=IE(n);i.run(t,o,s);function s(c,d,h){const m=d||t;c?l(c):u?u(m):r(void 0,m,h)}}}runSync(t,n){let r=!1,i;return this.run(t,n,a),gW("runSync","run",r),i;function a(u,l){dW(u),i=l,r=!0}}stringify(t,n){this.freeze();const r=IE(n),i=this.compiler||this.Compiler;return w7("stringify",i),mW(t),i(t,r)}use(t,...n){const r=this.attachers,i=this.namespace;if(R7("use",this.frozen),t!=null)if(typeof t=="function")o(t,n);else if(typeof t=="object")Array.isArray(t)?l(t):u(t);else throw new TypeError("Expected usable value, not `"+t+"`");return this;function a(s){if(typeof s=="function")o(s,[]);else if(typeof s=="object")if(Array.isArray(s)){const[c,...d]=s;o(c,d)}else u(s);else throw new TypeError("Expected usable value, not `"+s+"`")}function u(s){if(!("plugins"in s)&&!("settings"in s))throw new Error("Expected usable value but received an empty preset, which is probably a mistake: presets typically come with `plugins` and sometimes with `settings`, but this has neither");l(s.plugins),s.settings&&(i.settings=_7(!0,i.settings,s.settings))}function l(s){let c=-1;if(s!=null)if(Array.isArray(s))for(;++c<s.length;){const d=s[c];a(d)}else throw new TypeError("Expected a list of plugins, not `"+s+"`")}function o(s,c){let d=-1,h=-1;for(;++d<r.length;)if(r[d][0]===s){h=d;break}if(h===-1)r.push([s,...c]);else if(c.length>0){let[m,...y]=c;const v=r[h][1];OF(v)&&OF(m)&&(m=_7(!0,v,m)),r[h]=[s,m,...y]}}}}const hKe=new SB().freeze();function C7(e,t){if(typeof t!="function")throw new TypeError("Cannot `"+e+"` without `parser`")}function w7(e,t){if(typeof t!="function")throw new TypeError("Cannot `"+e+"` without `compiler`")}function R7(e,t){if(t)throw new Error("Cannot call `"+e+"` on a frozen processor.\nCreate a new processor first, by calling it: use `processor()` instead of `processor`.")}function mW(e){if(!OF(e)||typeof e.type!="string")throw new TypeError("Expected node, got `"+e+"`")}function gW(e,t,n){if(!n)throw new Error("`"+e+"` finished async. Use `"+t+"` instead")}function IE(e){return pKe(e)?e:new Lbe(e)}function pKe(e){return!!(e&&typeof e=="object"&&"message"in e&&"messages"in e)}function mKe(e){return typeof e=="string"||gKe(e)}function gKe(e){return!!(e&&typeof e=="object"&&"byteLength"in e&&"byteOffset"in e)}const vKe="https://github.com/remarkjs/react-markdown/blob/main/changelog.md",vW=[],yW={allowDangerousHtml:!0},yKe=/^(https?|ircs?|mailto|xmpp)$/i,bKe=[{from:"astPlugins",id:"remove-buggy-html-in-markdown-parser"},{from:"allowDangerousHtml",id:"remove-buggy-html-in-markdown-parser"},{from:"allowNode",id:"replace-allownode-allowedtypes-and-disallowedtypes",to:"allowElement"},{from:"allowedTypes",id:"replace-allownode-allowedtypes-and-disallowedtypes",to:"allowedElements"},{from:"disallowedTypes",id:"replace-allownode-allowedtypes-and-disallowedtypes",to:"disallowedElements"},{from:"escapeHtml",id:"remove-buggy-html-in-markdown-parser"},{from:"includeElementIndex",id:"#remove-includeelementindex"},{from:"includeNodeIndex",id:"change-includenodeindex-to-includeelementindex"},{from:"linkTarget",id:"remove-linktarget"},{from:"plugins",id:"change-plugins-to-remarkplugins",to:"remarkPlugins"},{from:"rawSourcePos",id:"#remove-rawsourcepos"},{from:"renderers",id:"change-renderers-to-components",to:"components"},{from:"source",id:"change-source-to-children",to:"children"},{from:"sourcePos",id:"#remove-sourcepos"},{from:"transformImageUri",id:"#add-urltransform",to:"urlTransform"},{from:"transformLinkUri",id:"#add-urltransform",to:"urlTransform"}];function Clt(e){const t=OKe(e),n=xKe(e);return SKe(t.runSync(t.parse(n),n),e)}function OKe(e){const t=e.rehypePlugins||vW,n=e.remarkPlugins||vW,r=e.remarkRehypeOptions?{...e.remarkRehypeOptions,...yW}:yW;return hKe().use(tGe).use(n).use(YGe,r).use(t)}function xKe(e){const t=e.children||"",n=new Lbe;return typeof t=="string"&&(n.value=t),n}function SKe(e,t){const n=t.allowedElements,r=t.allowElement,i=t.components,a=t.disallowedElements,u=t.skipHtml,l=t.unwrapDisallowed,o=t.urlTransform||_Ke;for(const c of bKe)Object.hasOwn(t,c.from)&&(""+c.from+(c.to?"use `"+c.to+"` instead":"remove it")+vKe+c.id,void 0);return t.className&&(e={type:"element",tagName:"div",properties:{className:t.className},children:e.type==="root"?e.children:[e]}),xB(e,s),LHe(e,{Fragment:Qr.Fragment,components:i,ignoreInvalidStyle:!0,jsx:Qr.jsx,jsxs:Qr.jsxs,passKeys:!0,passNode:!0});function s(c,d,h){if(c.type==="raw"&&h&&typeof d=="number")return u?h.children.splice(d,1):h.children[d]={type:"text",value:c.value},d;if(c.type==="element"){let m;for(m in b7)if(Object.hasOwn(b7,m)&&Object.hasOwn(c.properties,m)){const y=c.properties[m],v=b7[m];(v===null||v.includes(c.tagName))&&(c.properties[m]=o(String(y||""),m,c))}}if(c.type==="element"){let m=n?!n.includes(c.tagName):a?a.includes(c.tagName):!1;if(!m&&r&&typeof d=="number"&&(m=!r(c,d,h)),m&&h&&typeof d=="number")return l&&c.children?h.children.splice(d,1,...c.children):h.children.splice(d,1),d}}}function _Ke(e){const t=e.indexOf(":"),n=e.indexOf("?"),r=e.indexOf("#"),i=e.indexOf("/");return t===-1||i!==-1&&t>i||n!==-1&&t>n||r!==-1&&t>r||yKe.test(e.slice(0,t))?e:""}function EKe(e){if(typeof e!="string")throw new TypeError("Expected a string");return e.replace(/[|\\{}()[\]^$+*?.]/g,"\\$&").replace(/-/g,"\\x2d")}function DKe(e,t,n){const i=m5((n||{}).ignore||[]),a=$Ke(t);let u=-1;for(;++u<a.length;)Ibe(e,"text",l);function l(s,c){let d=-1,h;for(;++d<c.length;){const m=c[d],y=h?h.children:void 0;if(i(m,y?y.indexOf(m):void 0,h))return;h=m}if(h)return o(s,c)}function o(s,c){const d=c[c.length-1],h=a[u][0],m=a[u][1];let y=0;const S=d.children.indexOf(s);let _=!1,E=[];h.lastIndex=0;let $=h.exec(s.value);for(;$;){const k=$.index,A={index:$.index,input:$.input,stack:[...c,s]};let w=m(...$,A);if(typeof w=="string"&&(w=w.length>0?{type:"text",value:w}:void 0),w===!1?h.lastIndex=k+1:(y!==k&&E.push({type:"text",value:s.value.slice(y,k)}),Array.isArray(w)?E.push(...w):w&&E.push(w),y=k+$[0].length,_=!0),!h.global)break;$=h.exec(s.value)}return _?(y<s.value.length&&E.push({type:"text",value:s.value.slice(y)}),d.children.splice(S,1,...E)):E=[s],S+E.length}}function $Ke(e){const t=[];if(!Array.isArray(e))throw new TypeError("Expected find and replace tuple or list of tuples");const n=!e[0]||Array.isArray(e[0])?e:[e];let r=-1;for(;++r<n.length;){const i=n[r];t.push([CKe(i[0]),wKe(i[1])])}return t}function CKe(e){return typeof e=="string"?new RegExp(EKe(e),"g"):e}function wKe(e){return typeof e=="function"?e:function(){return e}}const A7="phrasing",T7=["autolink","link","image","label"];function RKe(){return{transforms:[NKe],enter:{literalAutolink:TKe,literalAutolinkEmail:k7,literalAutolinkHttp:k7,literalAutolinkWww:k7},exit:{literalAutolink:FKe,literalAutolinkEmail:MKe,literalAutolinkHttp:kKe,literalAutolinkWww:PKe}}}function AKe(){return{unsafe:[{character:"@",before:"[+\\-.\\w]",after:"[\\-.\\w]",inConstruct:A7,notInConstruct:T7},{character:".",before:"[Ww]",after:"[\\-.\\w]",inConstruct:A7,notInConstruct:T7},{character:":",before:"[ps]",after:"\\/",inConstruct:A7,notInConstruct:T7}]}}function TKe(e){this.enter({type:"link",title:null,url:"",children:[]},e)}function k7(e){this.config.enter.autolinkProtocol.call(this,e)}function kKe(e){this.config.exit.autolinkProtocol.call(this,e)}function PKe(e){this.config.exit.data.call(this,e);const t=this.stack[this.stack.length-1];t.type,t.url="http://"+this.sliceSerialize(e)}function MKe(e){this.config.exit.autolinkEmail.call(this,e)}function FKe(e){this.exit(e)}function NKe(e){DKe(e,[[/(https?:\/\/|www(?=\.))([-.\w]+)([^ \t\r\n]*)/gi,IKe],[new RegExp("(?<=^|\\s|\\p{P}|\\p{S})([-.\\w+]+)@([-\\w]+(?:\\.[-\\w]+)+)","gu"),LKe]],{ignore:["link","linkReference"]})}function IKe(e,t,n,r,i){let a="";if(!Bbe(i)||(/^w/i.test(t)&&(n=t+n,t="",a="http://"),!BKe(n)))return!1;const u=QKe(n+r);if(!u[0])return!1;const l={type:"link",title:null,url:a+t+u[0],children:[{type:"text",value:t+u[0]}]};return u[1]?[l,{type:"text",value:u[1]}]:l}function LKe(e,t,n,r){return!Bbe(r,!0)||/[-\d_]$/.test(n)?!1:{type:"link",title:null,url:"mailto:"+t+"@"+n,children:[{type:"text",value:t+"@"+n}]}}function BKe(e){const t=e.split(".");return!(t.length<2||t[t.length-1]&&(/_/.test(t[t.length-1])||!/[a-zA-Z\d]/.test(t[t.length-1]))||t[t.length-2]&&(/_/.test(t[t.length-2])||!/[a-zA-Z\d]/.test(t[t.length-2])))}function QKe(e){const t=/[!"&'),.:;<>?\]}]+$/.exec(e);if(!t)return[e,void 0];e=e.slice(0,t.index);let n=t[0],r=n.indexOf(")");const i=M$(e,"(");let a=M$(e,")");for(;r!==-1&&i>a;)e+=n.slice(0,r+1),n=n.slice(r+1),r=n.indexOf(")"),a++;return[e,n]}function Bbe(e,t){const n=e.input.charCodeAt(e.index-1);return(e.index===0||x1(n)||d5(n))&&(!t||n!==47)}Qbe.peek=YKe;function zKe(){this.buffer()}function qKe(e){this.enter({type:"footnoteReference",identifier:"",label:""},e)}function VKe(){this.buffer()}function UKe(e){this.enter({type:"footnoteDefinition",identifier:"",label:"",children:[]},e)}function jKe(e){const t=this.resume(),n=this.stack[this.stack.length-1];n.type,n.identifier=up(this.sliceSerialize(e)).toLowerCase(),n.label=t}function ZKe(e){this.exit(e)}function HKe(e){const t=this.resume(),n=this.stack[this.stack.length-1];n.type,n.identifier=up(this.sliceSerialize(e)).toLowerCase(),n.label=t}function XKe(e){this.exit(e)}function YKe(){return"["}function Qbe(e,t,n,r){const i=n.createTracker(r);let a=i.move("[^");const u=n.enter("footnoteReference"),l=n.enter("reference");return a+=i.move(n.safe(n.associationId(e),{after:"]",before:a})),l(),u(),a+=i.move("]"),a}function GKe(){return{enter:{gfmFootnoteCallString:zKe,gfmFootnoteCall:qKe,gfmFootnoteDefinitionLabelString:VKe,gfmFootnoteDefinition:UKe},exit:{gfmFootnoteCallString:jKe,gfmFootnoteCall:ZKe,gfmFootnoteDefinitionLabelString:HKe,gfmFootnoteDefinition:XKe}}}function KKe(e){let t=!1;return e&&e.firstLineBlank&&(t=!0),{handlers:{footnoteDefinition:n,footnoteReference:Qbe},unsafe:[{character:"[",inConstruct:["label","phrasing","reference"]}]};function n(r,i,a,u){const l=a.createTracker(u);let o=l.move("[^");const s=a.enter("footnoteDefinition"),c=a.enter("label");return o+=l.move(a.safe(a.associationId(r),{before:o,after:"]"})),c(),o+=l.move("]:"),r.children&&r.children.length>0&&(l.shift(4),o+=l.move((t?`
`:" ")+a.indentLines(a.containerFlow(r,l.current()),t?zbe:WKe))),s(),o}}function WKe(e,t,n){return t===0?e:zbe(e,t,n)}function zbe(e,t,n){return(n?"":"    ")+e}const JKe=["autolink","destinationLiteral","destinationRaw","reference","titleQuote","titleApostrophe"];qbe.peek=iWe;function eWe(){return{canContainEols:["delete"],enter:{strikethrough:nWe},exit:{strikethrough:rWe}}}function tWe(){return{unsafe:[{character:"~",inConstruct:"phrasing",notInConstruct:JKe}],handlers:{delete:qbe}}}function nWe(e){this.enter({type:"delete",children:[]},e)}function rWe(e){this.exit(e)}function qbe(e,t,n,r){const i=n.createTracker(r),a=n.enter("strikethrough");let u=i.move("~~");return u+=n.containerPhrasing(e,{...i.current(),before:u,after:"~"}),u+=i.move("~~"),a(),u}function iWe(){return"~"}function aWe(e){return e.length}function uWe(e,t){const n=t||{},r=(n.align||[]).concat(),i=n.stringLength||aWe,a=[],u=[],l=[],o=[];let s=0,c=-1;for(;++c<e.length;){const v=[],S=[];let _=-1;for(e[c].length>s&&(s=e[c].length);++_<e[c].length;){const E=lWe(e[c][_]);if(n.alignDelimiters!==!1){const $=i(E);S[_]=$,(o[_]===void 0||$>o[_])&&(o[_]=$)}v.push(E)}u[c]=v,l[c]=S}let d=-1;if(typeof r=="object"&&"length"in r)for(;++d<s;)a[d]=bW(r[d]);else{const v=bW(r);for(;++d<s;)a[d]=v}d=-1;const h=[],m=[];for(;++d<s;){const v=a[d];let S="",_="";v===99?(S=":",_=":"):v===108?S=":":v===114&&(_=":");let E=n.alignDelimiters===!1?1:Math.max(1,o[d]-S.length-_.length);const $=S+"-".repeat(E)+_;n.alignDelimiters!==!1&&(E=S.length+E+_.length,E>o[d]&&(o[d]=E),m[d]=E),h[d]=$}u.splice(1,0,h),l.splice(1,0,m),c=-1;const y=[];for(;++c<u.length;){const v=u[c],S=l[c];d=-1;const _=[];for(;++d<s;){const E=v[d]||"";let $="",k="";if(n.alignDelimiters!==!1){const A=o[d]-(S[d]||0),w=a[d];w===114?$=" ".repeat(A):w===99?A%2?($=" ".repeat(A/2+.5),k=" ".repeat(A/2-.5)):($=" ".repeat(A/2),k=$):k=" ".repeat(A)}n.delimiterStart!==!1&&!d&&_.push("|"),n.padding!==!1&&!(n.alignDelimiters===!1&&E==="")&&(n.delimiterStart!==!1||d)&&_.push(" "),n.alignDelimiters!==!1&&_.push($),_.push(E),n.alignDelimiters!==!1&&_.push(k),n.padding!==!1&&_.push(" "),(n.delimiterEnd!==!1||d!==s-1)&&_.push("|")}y.push(n.delimiterEnd===!1?_.join("").replace(/ +$/,""):_.join(""))}return y.join(`
`)}function lWe(e){return e==null?"":String(e)}function bW(e){const t=typeof e=="string"?e.codePointAt(0):0;return t===67||t===99?99:t===76||t===108?108:t===82||t===114?114:0}function oWe(e,t,n,r){const i=n.enter("blockquote"),a=n.createTracker(r);a.move("> "),a.shift(2);const u=n.indentLines(n.containerFlow(e,a.current()),sWe);return i(),u}function sWe(e,t,n){return">"+(n?"":" ")+e}function cWe(e,t){return OW(e,t.inConstruct,!0)&&!OW(e,t.notInConstruct,!1)}function OW(e,t,n){if(typeof t=="string"&&(t=[t]),!t||t.length===0)return n;let r=-1;for(;++r<t.length;)if(e.includes(t[r]))return!0;return!1}function xW(e,t,n,r){let i=-1;for(;++i<n.unsafe.length;)if(n.unsafe[i].character===`
`&&cWe(n.stack,n.unsafe[i]))return/[ \t]/.test(r.before)?"":" ";return`\\
`}function fWe(e,t){const n=String(e);let r=n.indexOf(t),i=r,a=0,u=0;if(typeof t!="string")throw new TypeError("Expected substring");for(;r!==-1;)r===i?++a>u&&(u=a):a=1,i=r+t.length,r=n.indexOf(t,i);return u}function dWe(e,t){return!!(t.options.fences===!1&&e.value&&!e.lang&&/[^ \r\n]/.test(e.value)&&!/^[\t ]*(?:[\r\n]|$)|(?:^|[\r\n])[\t ]*$/.test(e.value))}function hWe(e){const t=e.options.fence||"`";if(t!=="`"&&t!=="~")throw new Error("Cannot serialize code with `"+t+"` for `options.fence`, expected `` ` `` or `~`");return t}function pWe(e,t,n,r){const i=hWe(n),a=e.value||"",u=i==="`"?"GraveAccent":"Tilde";if(dWe(e,n)){const d=n.enter("codeIndented"),h=n.indentLines(a,mWe);return d(),h}const l=n.createTracker(r),o=i.repeat(Math.max(fWe(a,i)+1,3)),s=n.enter("codeFenced");let c=l.move(o);if(e.lang){const d=n.enter(`codeFencedLang${u}`);c+=l.move(n.safe(e.lang,{before:c,after:" ",encode:["`"],...l.current()})),d()}if(e.lang&&e.meta){const d=n.enter(`codeFencedMeta${u}`);c+=l.move(" "),c+=l.move(n.safe(e.meta,{before:c,after:`
`,encode:["`"],...l.current()})),d()}return c+=l.move(`
`),a&&(c+=l.move(a+`
`)),c+=l.move(o),s(),c}function mWe(e,t,n){return(n?"":"    ")+e}function _B(e){const t=e.options.quote||'"';if(t!=='"'&&t!=="'")throw new Error("Cannot serialize title with `"+t+"` for `options.quote`, expected `\"`, or `'`");return t}function gWe(e,t,n,r){const i=_B(n),a=i==='"'?"Quote":"Apostrophe",u=n.enter("definition");let l=n.enter("label");const o=n.createTracker(r);let s=o.move("[");return s+=o.move(n.safe(n.associationId(e),{before:s,after:"]",...o.current()})),s+=o.move("]: "),l(),!e.url||/[\0- \u007F]/.test(e.url)?(l=n.enter("destinationLiteral"),s+=o.move("<"),s+=o.move(n.safe(e.url,{before:s,after:">",...o.current()})),s+=o.move(">")):(l=n.enter("destinationRaw"),s+=o.move(n.safe(e.url,{before:s,after:e.title?" ":`
`,...o.current()}))),l(),e.title&&(l=n.enter(`title${a}`),s+=o.move(" "+i),s+=o.move(n.safe(e.title,{before:s,after:i,...o.current()})),s+=o.move(i),l()),u(),s}function vWe(e){const t=e.options.emphasis||"*";if(t!=="*"&&t!=="_")throw new Error("Cannot serialize emphasis with `"+t+"` for `options.emphasis`, expected `*`, or `_`");return t}function tS(e){return"&#x"+e.toString(16).toUpperCase()+";"}function cC(e,t,n){const r=V3(e),i=V3(t);return r===void 0?i===void 0?n==="_"?{inside:!0,outside:!0}:{inside:!1,outside:!1}:i===1?{inside:!0,outside:!0}:{inside:!1,outside:!0}:r===1?i===void 0?{inside:!1,outside:!1}:i===1?{inside:!0,outside:!0}:{inside:!1,outside:!1}:i===void 0?{inside:!1,outside:!1}:i===1?{inside:!0,outside:!1}:{inside:!1,outside:!1}}Vbe.peek=yWe;function Vbe(e,t,n,r){const i=vWe(n),a=n.enter("emphasis"),u=n.createTracker(r),l=u.move(i);let o=u.move(n.containerPhrasing(e,{after:i,before:l,...u.current()}));const s=o.charCodeAt(0),c=cC(r.before.charCodeAt(r.before.length-1),s,i);c.inside&&(o=tS(s)+o.slice(1));const d=o.charCodeAt(o.length-1),h=cC(r.after.charCodeAt(0),d,i);h.inside&&(o=o.slice(0,-1)+tS(d));const m=u.move(i);return a(),n.attentionEncodeSurroundingInfo={after:h.outside,before:c.outside},l+o+m}function yWe(e,t,n){return n.options.emphasis||"*"}function bWe(e,t){let n=!1;return xB(e,function(r){if("value"in r&&/\r?\n|\r/.test(r.value)||r.type==="break")return n=!0,yF}),!!((!e.depth||e.depth<3)&&pB(e)&&(t.options.setext||n))}function OWe(e,t,n,r){const i=Math.max(Math.min(6,e.depth||1),1),a=n.createTracker(r);if(bWe(e,n)){const c=n.enter("headingSetext"),d=n.enter("phrasing"),h=n.containerPhrasing(e,{...a.current(),before:`
`,after:`
`});return d(),c(),h+`
`+(i===1?"=":"-").repeat(h.length-(Math.max(h.lastIndexOf("\r"),h.lastIndexOf(`
`))+1))}const u="#".repeat(i),l=n.enter("headingAtx"),o=n.enter("phrasing");a.move(u+" ");let s=n.containerPhrasing(e,{before:"# ",after:`
`,...a.current()});return/^[\t ]/.test(s)&&(s=tS(s.charCodeAt(0))+s.slice(1)),s=s?u+" "+s:u,n.options.closeAtx&&(s+=" "+u),o(),l(),s}Ube.peek=xWe;function Ube(e){return e.value||""}function xWe(){return"<"}jbe.peek=SWe;function jbe(e,t,n,r){const i=_B(n),a=i==='"'?"Quote":"Apostrophe",u=n.enter("image");let l=n.enter("label");const o=n.createTracker(r);let s=o.move("![");return s+=o.move(n.safe(e.alt,{before:s,after:"]",...o.current()})),s+=o.move("]("),l(),!e.url&&e.title||/[\0- \u007F]/.test(e.url)?(l=n.enter("destinationLiteral"),s+=o.move("<"),s+=o.move(n.safe(e.url,{before:s,after:">",...o.current()})),s+=o.move(">")):(l=n.enter("destinationRaw"),s+=o.move(n.safe(e.url,{before:s,after:e.title?" ":")",...o.current()}))),l(),e.title&&(l=n.enter(`title${a}`),s+=o.move(" "+i),s+=o.move(n.safe(e.title,{before:s,after:i,...o.current()})),s+=o.move(i),l()),s+=o.move(")"),u(),s}function SWe(){return"!"}Zbe.peek=_We;function Zbe(e,t,n,r){const i=e.referenceType,a=n.enter("imageReference");let u=n.enter("label");const l=n.createTracker(r);let o=l.move("![");const s=n.safe(e.alt,{before:o,after:"]",...l.current()});o+=l.move(s+"]["),u();const c=n.stack;n.stack=[],u=n.enter("reference");const d=n.safe(n.associationId(e),{before:o,after:"]",...l.current()});return u(),n.stack=c,a(),i==="full"||!s||s!==d?o+=l.move(d+"]"):i==="shortcut"?o=o.slice(0,-1):o+=l.move("]"),o}function _We(){return"!"}Hbe.peek=EWe;function Hbe(e,t,n){let r=e.value||"",i="`",a=-1;for(;new RegExp("(^|[^`])"+i+"([^`]|$)").test(r);)i+="`";for(/[^ \r\n]/.test(r)&&(/^[ \r\n]/.test(r)&&/[ \r\n]$/.test(r)||/^`|`$/.test(r))&&(r=" "+r+" ");++a<n.unsafe.length;){const u=n.unsafe[a],l=n.compilePattern(u);let o;if(u.atBreak)for(;o=l.exec(r);){let s=o.index;r.charCodeAt(s)===10&&r.charCodeAt(s-1)===13&&s--,r=r.slice(0,s)+" "+r.slice(o.index+1)}}return i+r+i}function EWe(){return"`"}function Xbe(e,t){const n=pB(e);return!!(!t.options.resourceLink&&e.url&&!e.title&&e.children&&e.children.length===1&&e.children[0].type==="text"&&(n===e.url||"mailto:"+n===e.url)&&/^[a-z][a-z+.-]+:/i.test(e.url)&&!/[\0- <>\u007F]/.test(e.url))}Ybe.peek=DWe;function Ybe(e,t,n,r){const i=_B(n),a=i==='"'?"Quote":"Apostrophe",u=n.createTracker(r);let l,o;if(Xbe(e,n)){const c=n.stack;n.stack=[],l=n.enter("autolink");let d=u.move("<");return d+=u.move(n.containerPhrasing(e,{before:d,after:">",...u.current()})),d+=u.move(">"),l(),n.stack=c,d}l=n.enter("link"),o=n.enter("label");let s=u.move("[");return s+=u.move(n.containerPhrasing(e,{before:s,after:"](",...u.current()})),s+=u.move("]("),o(),!e.url&&e.title||/[\0- \u007F]/.test(e.url)?(o=n.enter("destinationLiteral"),s+=u.move("<"),s+=u.move(n.safe(e.url,{before:s,after:">",...u.current()})),s+=u.move(">")):(o=n.enter("destinationRaw"),s+=u.move(n.safe(e.url,{before:s,after:e.title?" ":")",...u.current()}))),o(),e.title&&(o=n.enter(`title${a}`),s+=u.move(" "+i),s+=u.move(n.safe(e.title,{before:s,after:i,...u.current()})),s+=u.move(i),o()),s+=u.move(")"),l(),s}function DWe(e,t,n){return Xbe(e,n)?"<":"["}Gbe.peek=$We;function Gbe(e,t,n,r){const i=e.referenceType,a=n.enter("linkReference");let u=n.enter("label");const l=n.createTracker(r);let o=l.move("[");const s=n.containerPhrasing(e,{before:o,after:"]",...l.current()});o+=l.move(s+"]["),u();const c=n.stack;n.stack=[],u=n.enter("reference");const d=n.safe(n.associationId(e),{before:o,after:"]",...l.current()});return u(),n.stack=c,a(),i==="full"||!s||s!==d?o+=l.move(d+"]"):i==="shortcut"?o=o.slice(0,-1):o+=l.move("]"),o}function $We(){return"["}function EB(e){const t=e.options.bullet||"*";if(t!=="*"&&t!=="+"&&t!=="-")throw new Error("Cannot serialize items with `"+t+"` for `options.bullet`, expected `*`, `+`, or `-`");return t}function CWe(e){const t=EB(e),n=e.options.bulletOther;if(!n)return t==="*"?"-":"*";if(n!=="*"&&n!=="+"&&n!=="-")throw new Error("Cannot serialize items with `"+n+"` for `options.bulletOther`, expected `*`, `+`, or `-`");if(n===t)throw new Error("Expected `bullet` (`"+t+"`) and `bulletOther` (`"+n+"`) to be different");return n}function wWe(e){const t=e.options.bulletOrdered||".";if(t!=="."&&t!==")")throw new Error("Cannot serialize items with `"+t+"` for `options.bulletOrdered`, expected `.` or `)`");return t}function Kbe(e){const t=e.options.rule||"*";if(t!=="*"&&t!=="-"&&t!=="_")throw new Error("Cannot serialize rules with `"+t+"` for `options.rule`, expected `*`, `-`, or `_`");return t}function RWe(e,t,n,r){const i=n.enter("list"),a=n.bulletCurrent;let u=e.ordered?wWe(n):EB(n);const l=e.ordered?u==="."?")":".":CWe(n);let o=t&&n.bulletLastUsed?u===n.bulletLastUsed:!1;if(!e.ordered){const c=e.children?e.children[0]:void 0;if((u==="*"||u==="-")&&c&&(!c.children||!c.children[0])&&n.stack[n.stack.length-1]==="list"&&n.stack[n.stack.length-2]==="listItem"&&n.stack[n.stack.length-3]==="list"&&n.stack[n.stack.length-4]==="listItem"&&n.indexStack[n.indexStack.length-1]===0&&n.indexStack[n.indexStack.length-2]===0&&n.indexStack[n.indexStack.length-3]===0&&(o=!0),Kbe(n)===u&&c){let d=-1;for(;++d<e.children.length;){const h=e.children[d];if(h&&h.type==="listItem"&&h.children&&h.children[0]&&h.children[0].type==="thematicBreak"){o=!0;break}}}}o&&(u=l),n.bulletCurrent=u;const s=n.containerFlow(e,r);return n.bulletLastUsed=u,n.bulletCurrent=a,i(),s}function AWe(e){const t=e.options.listItemIndent||"one";if(t!=="tab"&&t!=="one"&&t!=="mixed")throw new Error("Cannot serialize items with `"+t+"` for `options.listItemIndent`, expected `tab`, `one`, or `mixed`");return t}function TWe(e,t,n,r){const i=AWe(n);let a=n.bulletCurrent||EB(n);t&&t.type==="list"&&t.ordered&&(a=(typeof t.start=="number"&&t.start>-1?t.start:1)+(n.options.incrementListMarker===!1?0:t.children.indexOf(e))+a);let u=a.length+1;(i==="tab"||i==="mixed"&&(t&&t.type==="list"&&t.spread||e.spread))&&(u=Math.ceil(u/4)*4);const l=n.createTracker(r);l.move(a+" ".repeat(u-a.length)),l.shift(u);const o=n.enter("listItem"),s=n.indentLines(n.containerFlow(e,l.current()),c);return o(),s;function c(d,h,m){return h?(m?"":" ".repeat(u))+d:(m?a:a+" ".repeat(u-a.length))+d}}function kWe(e,t,n,r){const i=n.enter("paragraph"),a=n.enter("phrasing"),u=n.containerPhrasing(e,r);return a(),i(),u}const PWe=m5(["break","delete","emphasis","footnote","footnoteReference","image","imageReference","inlineCode","inlineMath","link","linkReference","mdxJsxTextElement","mdxTextExpression","strong","text","textDirective"]);function MWe(e,t,n,r){return(e.children.some(function(u){return PWe(u)})?n.containerPhrasing:n.containerFlow).call(n,e,r)}function FWe(e){const t=e.options.strong||"*";if(t!=="*"&&t!=="_")throw new Error("Cannot serialize strong with `"+t+"` for `options.strong`, expected `*`, or `_`");return t}Wbe.peek=NWe;function Wbe(e,t,n,r){const i=FWe(n),a=n.enter("strong"),u=n.createTracker(r),l=u.move(i+i);let o=u.move(n.containerPhrasing(e,{after:i,before:l,...u.current()}));const s=o.charCodeAt(0),c=cC(r.before.charCodeAt(r.before.length-1),s,i);c.inside&&(o=tS(s)+o.slice(1));const d=o.charCodeAt(o.length-1),h=cC(r.after.charCodeAt(0),d,i);h.inside&&(o=o.slice(0,-1)+tS(d));const m=u.move(i+i);return a(),n.attentionEncodeSurroundingInfo={after:h.outside,before:c.outside},l+o+m}function NWe(e,t,n){return n.options.strong||"*"}function IWe(e,t,n,r){return n.safe(e.value,r)}function LWe(e){const t=e.options.ruleRepetition||3;if(t<3)throw new Error("Cannot serialize rules with repetition `"+t+"` for `options.ruleRepetition`, expected `3` or more");return t}function BWe(e,t,n){const r=(Kbe(n)+(n.options.ruleSpaces?" ":"")).repeat(LWe(n));return n.options.ruleSpaces?r.slice(0,-1):r}const Jbe={blockquote:oWe,break:xW,code:pWe,definition:gWe,emphasis:Vbe,hardBreak:xW,heading:OWe,html:Ube,image:jbe,imageReference:Zbe,inlineCode:Hbe,link:Ybe,linkReference:Gbe,list:RWe,listItem:TWe,paragraph:kWe,root:MWe,strong:Wbe,text:IWe,thematicBreak:BWe};function QWe(){return{enter:{table:zWe,tableData:SW,tableHeader:SW,tableRow:VWe},exit:{codeText:UWe,table:qWe,tableData:P7,tableHeader:P7,tableRow:P7}}}function zWe(e){const t=e._align;this.enter({type:"table",align:t.map(function(n){return n==="none"?null:n}),children:[]},e),this.data.inTable=!0}function qWe(e){this.exit(e),this.data.inTable=void 0}function VWe(e){this.enter({type:"tableRow",children:[]},e)}function P7(e){this.exit(e)}function SW(e){this.enter({type:"tableCell",children:[]},e)}function UWe(e){let t=this.resume();this.data.inTable&&(t=t.replace(/\\([\\|])/g,jWe));const n=this.stack[this.stack.length-1];n.type,n.value=t,this.exit(e)}function jWe(e,t){return t==="|"?t:e}function ZWe(e){const t=e||{},n=t.tableCellPadding,r=t.tablePipeAlign,i=t.stringLength,a=n?" ":"|";return{unsafe:[{character:"\r",inConstruct:"tableCell"},{character:`
`,inConstruct:"tableCell"},{atBreak:!0,character:"|",after:"[	 :-]"},{character:"|",inConstruct:"tableCell"},{atBreak:!0,character:":",after:"-"},{atBreak:!0,character:"-",after:"[:|-]"}],handlers:{inlineCode:h,table:u,tableCell:o,tableRow:l}};function u(m,y,v,S){return s(c(m,v,S),m.align)}function l(m,y,v,S){const _=d(m,v,S),E=s([_]);return E.slice(0,E.indexOf(`
`))}function o(m,y,v,S){const _=v.enter("tableCell"),E=v.enter("phrasing"),$=v.containerPhrasing(m,{...S,before:a,after:a});return E(),_(),$}function s(m,y){return uWe(m,{align:y,alignDelimiters:r,padding:n,stringLength:i})}function c(m,y,v){const S=m.children;let _=-1;const E=[],$=y.enter("table");for(;++_<S.length;)E[_]=d(S[_],y,v);return $(),E}function d(m,y,v){const S=m.children;let _=-1;const E=[],$=y.enter("tableRow");for(;++_<S.length;)E[_]=o(S[_],m,y,v);return $(),E}function h(m,y,v){let S=Jbe.inlineCode(m,y,v);return v.stack.includes("tableCell")&&(S=S.replace(/\|/g,"\\$&")),S}}function HWe(){return{exit:{taskListCheckValueChecked:_W,taskListCheckValueUnchecked:_W,paragraph:YWe}}}function XWe(){return{unsafe:[{atBreak:!0,character:"-",after:"[:|-]"}],handlers:{listItem:GWe}}}function _W(e){const t=this.stack[this.stack.length-2];t.type,t.checked=e.type==="taskListCheckValueChecked"}function YWe(e){const t=this.stack[this.stack.length-2];if(t&&t.type==="listItem"&&typeof t.checked=="boolean"){const n=this.stack[this.stack.length-1];n.type;const r=n.children[0];if(r&&r.type==="text"){const i=t.children;let a=-1,u;for(;++a<i.length;){const l=i[a];if(l.type==="paragraph"){u=l;break}}u===n&&(r.value=r.value.slice(1),r.value.length===0?n.children.shift():n.position&&r.position&&typeof r.position.start.offset=="number"&&(r.position.start.column++,r.position.start.offset++,n.position.start=Object.assign({},r.position.start)))}}this.exit(e)}function GWe(e,t,n,r){const i=e.children[0],a=typeof e.checked=="boolean"&&i&&i.type==="paragraph",u="["+(e.checked?"x":" ")+"] ",l=n.createTracker(r);a&&l.move(u);let o=Jbe.listItem(e,t,n,{...r,...l.current()});return a&&(o=o.replace(/^(?:[*+-]|\d+\.)([\r\n]| {1,3})/,s)),o;function s(c){return c+u}}function KWe(){return[RKe(),GKe(),eWe(),QWe(),HWe()]}function WWe(e){return{extensions:[AKe(),KKe(e),tWe(),ZWe(e),XWe()]}}const JWe={tokenize:aJe,partial:!0},e1e={tokenize:uJe,partial:!0},t1e={tokenize:lJe,partial:!0},n1e={tokenize:oJe,partial:!0},eJe={tokenize:sJe,partial:!0},r1e={name:"wwwAutolink",tokenize:rJe,previous:a1e},i1e={name:"protocolAutolink",tokenize:iJe,previous:u1e},_g={name:"emailAutolink",tokenize:nJe,previous:l1e},U0={};function tJe(){return{text:U0}}let Eb=48;for(;Eb<123;)U0[Eb]=_g,Eb++,Eb===58?Eb=65:Eb===91&&(Eb=97);U0[43]=_g;U0[45]=_g;U0[46]=_g;U0[95]=_g;U0[72]=[_g,i1e];U0[104]=[_g,i1e];U0[87]=[_g,r1e];U0[119]=[_g,r1e];function nJe(e,t,n){const r=this;let i,a;return u;function u(d){return!SF(d)||!l1e.call(r,r.previous)||DB(r.events)?n(d):(e.enter("literalAutolink"),e.enter("literalAutolinkEmail"),l(d))}function l(d){return SF(d)?(e.consume(d),l):d===64?(e.consume(d),o):n(d)}function o(d){return d===46?e.check(eJe,c,s)(d):d===45||d===95||bs(d)?(a=!0,e.consume(d),o):c(d)}function s(d){return e.consume(d),i=!0,o}function c(d){return a&&i&&cc(r.previous)?(e.exit("literalAutolinkEmail"),e.exit("literalAutolink"),t(d)):n(d)}}function rJe(e,t,n){const r=this;return i;function i(u){return u!==87&&u!==119||!a1e.call(r,r.previous)||DB(r.events)?n(u):(e.enter("literalAutolink"),e.enter("literalAutolinkWww"),e.check(JWe,e.attempt(e1e,e.attempt(t1e,a),n),n)(u))}function a(u){return e.exit("literalAutolinkWww"),e.exit("literalAutolink"),t(u)}}function iJe(e,t,n){const r=this;let i="",a=!1;return u;function u(d){return(d===72||d===104)&&u1e.call(r,r.previous)&&!DB(r.events)?(e.enter("literalAutolink"),e.enter("literalAutolinkHttp"),i+=String.fromCodePoint(d),e.consume(d),l):n(d)}function l(d){if(cc(d)&&i.length<5)return i+=String.fromCodePoint(d),e.consume(d),l;if(d===58){const h=i.toLowerCase();if(h==="http"||h==="https")return e.consume(d),o}return n(d)}function o(d){return d===47?(e.consume(d),a?s:(a=!0,o)):n(d)}function s(d){return d===null||lC(d)||ca(d)||x1(d)||d5(d)?n(d):e.attempt(e1e,e.attempt(t1e,c),n)(d)}function c(d){return e.exit("literalAutolinkHttp"),e.exit("literalAutolink"),t(d)}}function aJe(e,t,n){let r=0;return i;function i(u){return(u===87||u===119)&&r<3?(r++,e.consume(u),i):u===46&&r===3?(e.consume(u),a):n(u)}function a(u){return u===null?n(u):t(u)}}function uJe(e,t,n){let r,i,a;return u;function u(s){return s===46||s===95?e.check(n1e,o,l)(s):s===null||ca(s)||x1(s)||s!==45&&d5(s)?o(s):(a=!0,e.consume(s),u)}function l(s){return s===95?r=!0:(i=r,r=void 0),e.consume(s),u}function o(s){return i||r||!a?n(s):t(s)}}function lJe(e,t){let n=0,r=0;return i;function i(u){return u===40?(n++,e.consume(u),i):u===41&&r<n?a(u):u===33||u===34||u===38||u===39||u===41||u===42||u===44||u===46||u===58||u===59||u===60||u===63||u===93||u===95||u===126?e.check(n1e,t,a)(u):u===null||ca(u)||x1(u)?t(u):(e.consume(u),i)}function a(u){return u===41&&r++,e.consume(u),i}}function oJe(e,t,n){return r;function r(l){return l===33||l===34||l===39||l===41||l===42||l===44||l===46||l===58||l===59||l===63||l===95||l===126?(e.consume(l),r):l===38?(e.consume(l),a):l===93?(e.consume(l),i):l===60||l===null||ca(l)||x1(l)?t(l):n(l)}function i(l){return l===null||l===40||l===91||ca(l)||x1(l)?t(l):r(l)}function a(l){return cc(l)?u(l):n(l)}function u(l){return l===59?(e.consume(l),r):cc(l)?(e.consume(l),u):n(l)}}function sJe(e,t,n){return r;function r(a){return e.consume(a),i}function i(a){return bs(a)?n(a):t(a)}}function a1e(e){return e===null||e===40||e===42||e===95||e===91||e===93||e===126||ca(e)}function u1e(e){return!cc(e)}function l1e(e){return!(e===47||SF(e))}function SF(e){return e===43||e===45||e===46||e===95||bs(e)}function DB(e){let t=e.length,n=!1;for(;t--;){const r=e[t][1];if((r.type==="labelLink"||r.type==="labelImage")&&!r._balanced){n=!0;break}if(r._gfmAutolinkLiteralWalkedInto){n=!1;break}}return e.length>0&&!n&&(e[e.length-1][1]._gfmAutolinkLiteralWalkedInto=!0),n}const cJe={tokenize:yJe,partial:!0};function fJe(){return{document:{91:{name:"gfmFootnoteDefinition",tokenize:mJe,continuation:{tokenize:gJe},exit:vJe}},text:{91:{name:"gfmFootnoteCall",tokenize:pJe},93:{name:"gfmPotentialFootnoteCall",add:"after",tokenize:dJe,resolveTo:hJe}}}}function dJe(e,t,n){const r=this;let i=r.events.length;const a=r.parser.gfmFootnotes||(r.parser.gfmFootnotes=[]);let u;for(;i--;){const o=r.events[i][1];if(o.type==="labelImage"){u=o;break}if(o.type==="gfmFootnoteCall"||o.type==="labelLink"||o.type==="label"||o.type==="image"||o.type==="link")break}return l;function l(o){if(!u||!u._balanced)return n(o);const s=up(r.sliceSerialize({start:u.end,end:r.now()}));return s.codePointAt(0)!==94||!a.includes(s.slice(1))?n(o):(e.enter("gfmFootnoteCallLabelMarker"),e.consume(o),e.exit("gfmFootnoteCallLabelMarker"),t(o))}}function hJe(e,t){let n=e.length;for(;n--;)if(e[n][1].type==="labelImage"&&e[n][0]==="enter"){e[n][1];break}e[n+1][1].type="data",e[n+3][1].type="gfmFootnoteCallLabelMarker";const r={type:"gfmFootnoteCall",start:Object.assign({},e[n+3][1].start),end:Object.assign({},e[e.length-1][1].end)},i={type:"gfmFootnoteCallMarker",start:Object.assign({},e[n+3][1].end),end:Object.assign({},e[n+3][1].end)};i.end.column++,i.end.offset++,i.end._bufferIndex++;const a={type:"gfmFootnoteCallString",start:Object.assign({},i.end),end:Object.assign({},e[e.length-1][1].start)},u={type:"chunkString",contentType:"string",start:Object.assign({},a.start),end:Object.assign({},a.end)},l=[e[n+1],e[n+2],["enter",r,t],e[n+3],e[n+4],["enter",i,t],["exit",i,t],["enter",a,t],["enter",u,t],["exit",u,t],["exit",a,t],e[e.length-2],e[e.length-1],["exit",r,t]];return e.splice(n,e.length-n+1,...l),e}function pJe(e,t,n){const r=this,i=r.parser.gfmFootnotes||(r.parser.gfmFootnotes=[]);let a=0,u;return l;function l(d){return e.enter("gfmFootnoteCall"),e.enter("gfmFootnoteCallLabelMarker"),e.consume(d),e.exit("gfmFootnoteCallLabelMarker"),o}function o(d){return d!==94?n(d):(e.enter("gfmFootnoteCallMarker"),e.consume(d),e.exit("gfmFootnoteCallMarker"),e.enter("gfmFootnoteCallString"),e.enter("chunkString").contentType="string",s)}function s(d){if(a>999||d===93&&!u||d===null||d===91||ca(d))return n(d);if(d===93){e.exit("chunkString");const h=e.exit("gfmFootnoteCallString");return i.includes(up(r.sliceSerialize(h)))?(e.enter("gfmFootnoteCallLabelMarker"),e.consume(d),e.exit("gfmFootnoteCallLabelMarker"),e.exit("gfmFootnoteCall"),t):n(d)}return ca(d)||(u=!0),a++,e.consume(d),d===92?c:s}function c(d){return d===91||d===92||d===93?(e.consume(d),a++,s):s(d)}}function mJe(e,t,n){const r=this,i=r.parser.gfmFootnotes||(r.parser.gfmFootnotes=[]);let a,u=0,l;return o;function o(y){return e.enter("gfmFootnoteDefinition")._container=!0,e.enter("gfmFootnoteDefinitionLabel"),e.enter("gfmFootnoteDefinitionLabelMarker"),e.consume(y),e.exit("gfmFootnoteDefinitionLabelMarker"),s}function s(y){return y===94?(e.enter("gfmFootnoteDefinitionMarker"),e.consume(y),e.exit("gfmFootnoteDefinitionMarker"),e.enter("gfmFootnoteDefinitionLabelString"),e.enter("chunkString").contentType="string",c):n(y)}function c(y){if(u>999||y===93&&!l||y===null||y===91||ca(y))return n(y);if(y===93){e.exit("chunkString");const v=e.exit("gfmFootnoteDefinitionLabelString");return a=up(r.sliceSerialize(v)),e.enter("gfmFootnoteDefinitionLabelMarker"),e.consume(y),e.exit("gfmFootnoteDefinitionLabelMarker"),e.exit("gfmFootnoteDefinitionLabel"),h}return ca(y)||(l=!0),u++,e.consume(y),y===92?d:c}function d(y){return y===91||y===92||y===93?(e.consume(y),u++,c):c(y)}function h(y){return y===58?(e.enter("definitionMarker"),e.consume(y),e.exit("definitionMarker"),i.includes(a)||i.push(a),di(e,m,"gfmFootnoteDefinitionWhitespace")):n(y)}function m(y){return t(y)}}function gJe(e,t,n){return e.check(qS,t,e.attempt(cJe,t,n))}function vJe(e){e.exit("gfmFootnoteDefinition")}function yJe(e,t,n){const r=this;return di(e,i,"gfmFootnoteDefinitionIndent",5);function i(a){const u=r.events[r.events.length-1];return u&&u[1].type==="gfmFootnoteDefinitionIndent"&&u[2].sliceSerialize(u[1],!0).length===4?t(a):n(a)}}function bJe(e){let n=(e||{}).singleTilde;const r={name:"strikethrough",tokenize:a,resolveAll:i};return n==null&&(n=!0),{text:{126:r},insideSpan:{null:[r]},attentionMarkers:{null:[126]}};function i(u,l){let o=-1;for(;++o<u.length;)if(u[o][0]==="enter"&&u[o][1].type==="strikethroughSequenceTemporary"&&u[o][1]._close){let s=o;for(;s--;)if(u[s][0]==="exit"&&u[s][1].type==="strikethroughSequenceTemporary"&&u[s][1]._open&&u[o][1].end.offset-u[o][1].start.offset===u[s][1].end.offset-u[s][1].start.offset){u[o][1].type="strikethroughSequence",u[s][1].type="strikethroughSequence";const c={type:"strikethrough",start:Object.assign({},u[s][1].start),end:Object.assign({},u[o][1].end)},d={type:"strikethroughText",start:Object.assign({},u[s][1].end),end:Object.assign({},u[o][1].start)},h=[["enter",c,l],["enter",u[s][1],l],["exit",u[s][1],l],["enter",d,l]],m=l.parser.constructs.insideSpan.null;m&&Hf(h,h.length,0,h5(m,u.slice(s+1,o),l)),Hf(h,h.length,0,[["exit",d,l],["enter",u[o][1],l],["exit",u[o][1],l],["exit",c,l]]),Hf(u,s-1,o-s+3,h),o=s+h.length-2;break}}for(o=-1;++o<u.length;)u[o][1].type==="strikethroughSequenceTemporary"&&(u[o][1].type="data");return u}function a(u,l,o){const s=this.previous,c=this.events;let d=0;return h;function h(y){return s===126&&c[c.length-1][1].type!=="characterEscape"?o(y):(u.enter("strikethroughSequenceTemporary"),m(y))}function m(y){const v=V3(s);if(y===126)return d>1?o(y):(u.consume(y),d++,m);if(d<2&&!n)return o(y);const S=u.exit("strikethroughSequenceTemporary"),_=V3(y);return S._open=!_||_===2&&!!v,S._close=!v||v===2&&!!_,l(y)}}}class OJe{constructor(){this.map=[]}add(t,n,r){xJe(this,t,n,r)}consume(t){if(this.map.sort(function(a,u){return a[0]-u[0]}),this.map.length===0)return;let n=this.map.length;const r=[];for(;n>0;)n-=1,r.push(t.slice(this.map[n][0]+this.map[n][1]),this.map[n][2]),t.length=this.map[n][0];r.push(t.slice()),t.length=0;let i=r.pop();for(;i;){for(const a of i)t.push(a);i=r.pop()}this.map.length=0}}function xJe(e,t,n,r){let i=0;if(!(n===0&&r.length===0)){for(;i<e.map.length;){if(e.map[i][0]===t){e.map[i][1]+=n,e.map[i][2].push(...r);return}i+=1}e.map.push([t,n,r])}}function SJe(e,t){let n=!1;const r=[];for(;t<e.length;){const i=e[t];if(n){if(i[0]==="enter")i[1].type==="tableContent"&&r.push(e[t+1][1].type==="tableDelimiterMarker"?"left":"none");else if(i[1].type==="tableContent"){if(e[t-1][1].type==="tableDelimiterMarker"){const a=r.length-1;r[a]=r[a]==="left"?"center":"right"}}else if(i[1].type==="tableDelimiterRow")break}else i[0]==="enter"&&i[1].type==="tableDelimiterRow"&&(n=!0);t+=1}return r}function _Je(){return{flow:{null:{name:"table",tokenize:EJe,resolveAll:DJe}}}}function EJe(e,t,n){const r=this;let i=0,a=0,u;return l;function l(U){let q=r.events.length-1;for(;q>-1;){const j=r.events[q][1].type;if(j==="lineEnding"||j==="linePrefix")q--;else break}const V=q>-1?r.events[q][1].type:null,N=V==="tableHead"||V==="tableRow"?w:o;return N===w&&r.parser.lazy[r.now().line]?n(U):N(U)}function o(U){return e.enter("tableHead"),e.enter("tableRow"),s(U)}function s(U){return U===124||(u=!0,a+=1),c(U)}function c(U){return U===null?n(U):Pn(U)?a>1?(a=0,r.interrupt=!0,e.exit("tableRow"),e.enter("lineEnding"),e.consume(U),e.exit("lineEnding"),m):n(U):Zr(U)?di(e,c,"whitespace")(U):(a+=1,u&&(u=!1,i+=1),U===124?(e.enter("tableCellDivider"),e.consume(U),e.exit("tableCellDivider"),u=!0,c):(e.enter("data"),d(U)))}function d(U){return U===null||U===124||ca(U)?(e.exit("data"),c(U)):(e.consume(U),U===92?h:d)}function h(U){return U===92||U===124?(e.consume(U),d):d(U)}function m(U){return r.interrupt=!1,r.parser.lazy[r.now().line]?n(U):(e.enter("tableDelimiterRow"),u=!1,Zr(U)?di(e,y,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(U):y(U))}function y(U){return U===45||U===58?S(U):U===124?(u=!0,e.enter("tableCellDivider"),e.consume(U),e.exit("tableCellDivider"),v):A(U)}function v(U){return Zr(U)?di(e,S,"whitespace")(U):S(U)}function S(U){return U===58?(a+=1,u=!0,e.enter("tableDelimiterMarker"),e.consume(U),e.exit("tableDelimiterMarker"),_):U===45?(a+=1,_(U)):U===null||Pn(U)?k(U):A(U)}function _(U){return U===45?(e.enter("tableDelimiterFiller"),E(U)):A(U)}function E(U){return U===45?(e.consume(U),E):U===58?(u=!0,e.exit("tableDelimiterFiller"),e.enter("tableDelimiterMarker"),e.consume(U),e.exit("tableDelimiterMarker"),$):(e.exit("tableDelimiterFiller"),$(U))}function $(U){return Zr(U)?di(e,k,"whitespace")(U):k(U)}function k(U){return U===124?y(U):U===null||Pn(U)?!u||i!==a?A(U):(e.exit("tableDelimiterRow"),e.exit("tableHead"),t(U)):A(U)}function A(U){return n(U)}function w(U){return e.enter("tableRow"),R(U)}function R(U){return U===124?(e.enter("tableCellDivider"),e.consume(U),e.exit("tableCellDivider"),R):U===null||Pn(U)?(e.exit("tableRow"),t(U)):Zr(U)?di(e,R,"whitespace")(U):(e.enter("data"),F(U))}function F(U){return U===null||U===124||ca(U)?(e.exit("data"),R(U)):(e.consume(U),U===92?I:F)}function I(U){return U===92||U===124?(e.consume(U),F):F(U)}}function DJe(e,t){let n=-1,r=!0,i=0,a=[0,0,0,0],u=[0,0,0,0],l=!1,o=0,s,c,d;const h=new OJe;for(;++n<e.length;){const m=e[n],y=m[1];m[0]==="enter"?y.type==="tableHead"?(l=!1,o!==0&&(EW(h,t,o,s,c),c=void 0,o=0),s={type:"table",start:Object.assign({},y.start),end:Object.assign({},y.end)},h.add(n,0,[["enter",s,t]])):y.type==="tableRow"||y.type==="tableDelimiterRow"?(r=!0,d=void 0,a=[0,0,0,0],u=[0,n+1,0,0],l&&(l=!1,c={type:"tableBody",start:Object.assign({},y.start),end:Object.assign({},y.end)},h.add(n,0,[["enter",c,t]])),i=y.type==="tableDelimiterRow"?2:c?3:1):i&&(y.type==="data"||y.type==="tableDelimiterMarker"||y.type==="tableDelimiterFiller")?(r=!1,u[2]===0&&(a[1]!==0&&(u[0]=u[1],d=LE(h,t,a,i,void 0,d),a=[0,0,0,0]),u[2]=n)):y.type==="tableCellDivider"&&(r?r=!1:(a[1]!==0&&(u[0]=u[1],d=LE(h,t,a,i,void 0,d)),a=u,u=[a[1],n,0,0])):y.type==="tableHead"?(l=!0,o=n):y.type==="tableRow"||y.type==="tableDelimiterRow"?(o=n,a[1]!==0?(u[0]=u[1],d=LE(h,t,a,i,n,d)):u[1]!==0&&(d=LE(h,t,u,i,n,d)),i=0):i&&(y.type==="data"||y.type==="tableDelimiterMarker"||y.type==="tableDelimiterFiller")&&(u[3]=n)}for(o!==0&&EW(h,t,o,s,c),h.consume(t.events),n=-1;++n<t.events.length;){const m=t.events[n];m[0]==="enter"&&m[1].type==="table"&&(m[1]._align=SJe(t.events,n))}return e}function LE(e,t,n,r,i,a){const u=r===1?"tableHeader":r===2?"tableDelimiter":"tableData",l="tableContent";n[0]!==0&&(a.end=Object.assign({},qO(t.events,n[0])),e.add(n[0],0,[["exit",a,t]]));const o=qO(t.events,n[1]);if(a={type:u,start:Object.assign({},o),end:Object.assign({},o)},e.add(n[1],0,[["enter",a,t]]),n[2]!==0){const s=qO(t.events,n[2]),c=qO(t.events,n[3]),d={type:l,start:Object.assign({},s),end:Object.assign({},c)};if(e.add(n[2],0,[["enter",d,t]]),r!==2){const h=t.events[n[2]],m=t.events[n[3]];if(h[1].end=Object.assign({},m[1].end),h[1].type="chunkText",h[1].contentType="text",n[3]>n[2]+1){const y=n[2]+1,v=n[3]-n[2]-1;e.add(y,v,[])}}e.add(n[3]+1,0,[["exit",d,t]])}return i!==void 0&&(a.end=Object.assign({},qO(t.events,i)),e.add(i,0,[["exit",a,t]]),a=void 0),a}function EW(e,t,n,r,i){const a=[],u=qO(t.events,n);i&&(i.end=Object.assign({},u),a.push(["exit",i,t])),r.end=Object.assign({},u),a.push(["exit",r,t]),e.add(n+1,0,a)}function qO(e,t){const n=e[t],r=n[0]==="enter"?"start":"end";return n[1][r]}const $Je={name:"tasklistCheck",tokenize:wJe};function CJe(){return{text:{91:$Je}}}function wJe(e,t,n){const r=this;return i;function i(o){return r.previous!==null||!r._gfmTasklistFirstContentOfListItem?n(o):(e.enter("taskListCheck"),e.enter("taskListCheckMarker"),e.consume(o),e.exit("taskListCheckMarker"),a)}function a(o){return ca(o)?(e.enter("taskListCheckValueUnchecked"),e.consume(o),e.exit("taskListCheckValueUnchecked"),u):o===88||o===120?(e.enter("taskListCheckValueChecked"),e.consume(o),e.exit("taskListCheckValueChecked"),u):n(o)}function u(o){return o===93?(e.enter("taskListCheckMarker"),e.consume(o),e.exit("taskListCheckMarker"),e.exit("taskListCheck"),l):n(o)}function l(o){return Pn(o)?t(o):Zr(o)?e.check({tokenize:RJe},t,n)(o):n(o)}}function RJe(e,t,n){return di(e,r,"whitespace");function r(i){return i===null?n(i):t(i)}}function AJe(e){return ybe([tJe(),fJe(),bJe(e),_Je(),CJe()])}const TJe={};function wlt(e){const t=this,n=e||TJe,r=t.data(),i=r.micromarkExtensions||(r.micromarkExtensions=[]),a=r.fromMarkdownExtensions||(r.fromMarkdownExtensions=[]),u=r.toMarkdownExtensions||(r.toMarkdownExtensions=[]);i.push(AJe(n)),a.push(KWe()),u.push(WWe(n))}function kJe(){for(var e=arguments.length,t=new Array(e),n=0;n<e;n++)t[n]=arguments[n];return P.useMemo(()=>r=>{t.forEach(i=>i(r))},t)}const v5=typeof window<"u"&&typeof window.document<"u"&&typeof window.document.createElement<"u";function g4(e){const t=Object.prototype.toString.call(e);return t==="[object Window]"||t==="[object global]"}function $B(e){return"nodeType"in e}function mf(e){var t,n;return e?g4(e)?e:$B(e)&&(t=(n=e.ownerDocument)==null?void 0:n.defaultView)!=null?t:window:window}function CB(e){const{Document:t}=mf(e);return e instanceof t}function US(e){return g4(e)?!1:e instanceof mf(e).HTMLElement}function o1e(e){return e instanceof mf(e).SVGElement}function v4(e){return e?g4(e)?e.document:$B(e)?CB(e)?e:US(e)||o1e(e)?e.ownerDocument:document:document:document}const F0=v5?P.useLayoutEffect:P.useEffect;function wB(e){const t=P.useRef(e);return F0(()=>{t.current=e}),P.useCallback(function(){for(var n=arguments.length,r=new Array(n),i=0;i<n;i++)r[i]=arguments[i];return t.current==null?void 0:t.current(...r)},[])}function PJe(){const e=P.useRef(null),t=P.useCallback((r,i)=>{e.current=setInterval(r,i)},[]),n=P.useCallback(()=>{e.current!==null&&(clearInterval(e.current),e.current=null)},[]);return[t,n]}function nS(e,t){t===void 0&&(t=[e]);const n=P.useRef(e);return F0(()=>{n.current!==e&&(n.current=e)},t),n}function jS(e,t){const n=P.useRef();return P.useMemo(()=>{const r=e(n.current);return n.current=r,r},[...t])}function fC(e){const t=wB(e),n=P.useRef(null),r=P.useCallback(i=>{i!==n.current&&(t==null||t(i,n.current)),n.current=i},[]);return[n,r]}function _F(e){const t=P.useRef();return P.useEffect(()=>{t.current=e},[e]),t.current}let M7={};function ZS(e,t){return P.useMemo(()=>{if(t)return t;const n=M7[e]==null?0:M7[e]+1;return M7[e]=n,e+"-"+n},[e,t])}function s1e(e){return function(t){for(var n=arguments.length,r=new Array(n>1?n-1:0),i=1;i<n;i++)r[i-1]=arguments[i];return r.reduce((a,u)=>{const l=Object.entries(u);for(const[o,s]of l){const c=a[o];c!=null&&(a[o]=c+e*s)}return a},{...t})}}const _3=s1e(1),rS=s1e(-1);function MJe(e){return"clientX"in e&&"clientY"in e}function RB(e){if(!e)return!1;const{KeyboardEvent:t}=mf(e.target);return t&&e instanceof t}function FJe(e){if(!e)return!1;const{TouchEvent:t}=mf(e.target);return t&&e instanceof t}function EF(e){if(FJe(e)){if(e.touches&&e.touches.length){const{clientX:t,clientY:n}=e.touches[0];return{x:t,y:n}}else if(e.changedTouches&&e.changedTouches.length){const{clientX:t,clientY:n}=e.changedTouches[0];return{x:t,y:n}}}return MJe(e)?{x:e.clientX,y:e.clientY}:null}const dC=Object.freeze({Translate:{toString(e){if(!e)return;const{x:t,y:n}=e;return"translate3d("+(t?Math.round(t):0)+"px, "+(n?Math.round(n):0)+"px, 0)"}},Scale:{toString(e){if(!e)return;const{scaleX:t,scaleY:n}=e;return"scaleX("+t+") scaleY("+n+")"}},Transform:{toString(e){if(e)return[dC.Translate.toString(e),dC.Scale.toString(e)].join(" ")}},Transition:{toString(e){let{property:t,duration:n,easing:r}=e;return t+" "+n+"ms "+r}}}),DW="a,frame,iframe,input:not([type=hidden]):not(:disabled),select:not(:disabled),textarea:not(:disabled),button:not(:disabled),*[tabindex]";function NJe(e){return e.matches(DW)?e:e.querySelector(DW)}const IJe={display:"none"};function LJe(e){let{id:t,value:n}=e;return be.createElement("div",{id:t,style:IJe},n)}function BJe(e){let{id:t,announcement:n,ariaLiveType:r="assertive"}=e;const i={position:"fixed",top:0,left:0,width:1,height:1,margin:-1,border:0,padding:0,overflow:"hidden",clip:"rect(0 0 0 0)",clipPath:"inset(100%)",whiteSpace:"nowrap"};return be.createElement("div",{id:t,style:i,role:"status","aria-live":r,"aria-atomic":!0},n)}function QJe(){const[e,t]=P.useState("");return{announce:P.useCallback(r=>{r!=null&&t(r)},[]),announcement:e}}const c1e=P.createContext(null);function zJe(e){const t=P.useContext(c1e);P.useEffect(()=>{if(!t)throw new Error("useDndMonitor must be used within a children of <DndContext>");return t(e)},[e,t])}function qJe(){const[e]=P.useState(()=>new Set),t=P.useCallback(r=>(e.add(r),()=>e.delete(r)),[e]);return[P.useCallback(r=>{let{type:i,event:a}=r;e.forEach(u=>{var l;return(l=u[i])==null?void 0:l.call(u,a)})},[e]),t]}const VJe={draggable:`
    To pick up a draggable item, press the space bar.
    While dragging, use the arrow keys to move the item.
    Press space again to drop the item in its new position, or press escape to cancel.
  `},UJe={onDragStart(e){let{active:t}=e;return"Picked up draggable item "+t.id+"."},onDragOver(e){let{active:t,over:n}=e;return n?"Draggable item "+t.id+" was moved over droppable area "+n.id+".":"Draggable item "+t.id+" is no longer over a droppable area."},onDragEnd(e){let{active:t,over:n}=e;return n?"Draggable item "+t.id+" was dropped over droppable area "+n.id:"Draggable item "+t.id+" was dropped."},onDragCancel(e){let{active:t}=e;return"Dragging was cancelled. Draggable item "+t.id+" was dropped."}};function jJe(e){let{announcements:t=UJe,container:n,hiddenTextDescribedById:r,screenReaderInstructions:i=VJe}=e;const{announce:a,announcement:u}=QJe(),l=ZS("DndLiveRegion"),[o,s]=P.useState(!1);if(P.useEffect(()=>{s(!0)},[]),zJe(P.useMemo(()=>({onDragStart(d){let{active:h}=d;a(t.onDragStart({active:h}))},onDragMove(d){let{active:h,over:m}=d;t.onDragMove&&a(t.onDragMove({active:h,over:m}))},onDragOver(d){let{active:h,over:m}=d;a(t.onDragOver({active:h,over:m}))},onDragEnd(d){let{active:h,over:m}=d;a(t.onDragEnd({active:h,over:m}))},onDragCancel(d){let{active:h,over:m}=d;a(t.onDragCancel({active:h,over:m}))}}),[a,t])),!o)return null;const c=be.createElement(be.Fragment,null,be.createElement(LJe,{id:r,value:i.draggable}),be.createElement(BJe,{id:l,announcement:u}));return n?Uf.createPortal(c,n):c}var kl;(function(e){e.DragStart="dragStart",e.DragMove="dragMove",e.DragEnd="dragEnd",e.DragCancel="dragCancel",e.DragOver="dragOver",e.RegisterDroppable="registerDroppable",e.SetDroppableDisabled="setDroppableDisabled",e.UnregisterDroppable="unregisterDroppable"})(kl||(kl={}));function hC(){}function Rlt(e,t){return P.useMemo(()=>({sensor:e,options:t??{}}),[e,t])}function Alt(){for(var e=arguments.length,t=new Array(e),n=0;n<e;n++)t[n]=arguments[n];return P.useMemo(()=>[...t].filter(r=>r!=null),[...t])}const fp=Object.freeze({x:0,y:0});function ZJe(e,t){return Math.sqrt(Math.pow(e.x-t.x,2)+Math.pow(e.y-t.y,2))}function HJe(e,t){let{data:{value:n}}=e,{data:{value:r}}=t;return n-r}function XJe(e,t){let{data:{value:n}}=e,{data:{value:r}}=t;return r-n}function $W(e){let{left:t,top:n,height:r,width:i}=e;return[{x:t,y:n},{x:t+i,y:n},{x:t,y:n+r},{x:t+i,y:n+r}]}function f1e(e,t){if(!e||e.length===0)return null;const[n]=e;return n[t]}const YJe=e=>{let{collisionRect:t,droppableRects:n,droppableContainers:r}=e;const i=$W(t),a=[];for(const u of r){const{id:l}=u,o=n.get(l);if(o){const s=$W(o),c=i.reduce((h,m,y)=>h+ZJe(s[y],m),0),d=Number((c/4).toFixed(4));a.push({id:l,data:{droppableContainer:u,value:d}})}}return a.sort(HJe)};function GJe(e,t){const n=Math.max(t.top,e.top),r=Math.max(t.left,e.left),i=Math.min(t.left+t.width,e.left+e.width),a=Math.min(t.top+t.height,e.top+e.height),u=i-r,l=a-n;if(r<i&&n<a){const o=t.width*t.height,s=e.width*e.height,c=u*l,d=c/(o+s-c);return Number(d.toFixed(4))}return 0}const KJe=e=>{let{collisionRect:t,droppableRects:n,droppableContainers:r}=e;const i=[];for(const a of r){const{id:u}=a,l=n.get(u);if(l){const o=GJe(l,t);o>0&&i.push({id:u,data:{droppableContainer:a,value:o}})}}return i.sort(XJe)};function WJe(e,t,n){return{...e,scaleX:t&&n?t.width/n.width:1,scaleY:t&&n?t.height/n.height:1}}function d1e(e,t){return e&&t?{x:e.left-t.left,y:e.top-t.top}:fp}function JJe(e){return function(n){for(var r=arguments.length,i=new Array(r>1?r-1:0),a=1;a<r;a++)i[a-1]=arguments[a];return i.reduce((u,l)=>({...u,top:u.top+e*l.y,bottom:u.bottom+e*l.y,left:u.left+e*l.x,right:u.right+e*l.x}),{...n})}}const eet=JJe(1);function tet(e){if(e.startsWith("matrix3d(")){const t=e.slice(9,-1).split(/, /);return{x:+t[12],y:+t[13],scaleX:+t[0],scaleY:+t[5]}}else if(e.startsWith("matrix(")){const t=e.slice(7,-1).split(/, /);return{x:+t[4],y:+t[5],scaleX:+t[0],scaleY:+t[3]}}return null}function net(e,t,n){const r=tet(t);if(!r)return e;const{scaleX:i,scaleY:a,x:u,y:l}=r,o=e.left-u-(1-i)*parseFloat(n),s=e.top-l-(1-a)*parseFloat(n.slice(n.indexOf(" ")+1)),c=i?e.width/i:e.width,d=a?e.height/a:e.height;return{width:c,height:d,top:s,right:o+c,bottom:s+d,left:o}}const ret={ignoreTransform:!1};function y4(e,t){t===void 0&&(t=ret);let n=e.getBoundingClientRect();if(t.ignoreTransform){const{transform:s,transformOrigin:c}=mf(e).getComputedStyle(e);s&&(n=net(n,s,c))}const{top:r,left:i,width:a,height:u,bottom:l,right:o}=n;return{top:r,left:i,width:a,height:u,bottom:l,right:o}}function CW(e){return y4(e,{ignoreTransform:!0})}function iet(e){const t=e.innerWidth,n=e.innerHeight;return{top:0,left:0,right:t,bottom:n,width:t,height:n}}function aet(e,t){return t===void 0&&(t=mf(e).getComputedStyle(e)),t.position==="fixed"}function uet(e,t){t===void 0&&(t=mf(e).getComputedStyle(e));const n=/(auto|scroll|overlay)/;return["overflow","overflowX","overflowY"].some(i=>{const a=t[i];return typeof a=="string"?n.test(a):!1})}function y5(e,t){const n=[];function r(i){if(t!=null&&n.length>=t||!i)return n;if(CB(i)&&i.scrollingElement!=null&&!n.includes(i.scrollingElement))return n.push(i.scrollingElement),n;if(!US(i)||o1e(i)||n.includes(i))return n;const a=mf(e).getComputedStyle(i);return i!==e&&uet(i,a)&&n.push(i),aet(i,a)?n:r(i.parentNode)}return e?r(e):n}function h1e(e){const[t]=y5(e,1);return t??null}function F7(e){return!v5||!e?null:g4(e)?e:$B(e)?CB(e)||e===v4(e).scrollingElement?window:US(e)?e:null:null}function p1e(e){return g4(e)?e.scrollX:e.scrollLeft}function m1e(e){return g4(e)?e.scrollY:e.scrollTop}function DF(e){return{x:p1e(e),y:m1e(e)}}var lo;(function(e){e[e.Forward=1]="Forward",e[e.Backward=-1]="Backward"})(lo||(lo={}));function g1e(e){return!v5||!e?!1:e===document.scrollingElement}function v1e(e){const t={x:0,y:0},n=g1e(e)?{height:window.innerHeight,width:window.innerWidth}:{height:e.clientHeight,width:e.clientWidth},r={x:e.scrollWidth-n.width,y:e.scrollHeight-n.height},i=e.scrollTop<=t.y,a=e.scrollLeft<=t.x,u=e.scrollTop>=r.y,l=e.scrollLeft>=r.x;return{isTop:i,isLeft:a,isBottom:u,isRight:l,maxScroll:r,minScroll:t}}const oet={x:.2,y:.2};function set(e,t,n,r,i){let{top:a,left:u,right:l,bottom:o}=n;r===void 0&&(r=10),i===void 0&&(i=oet);const{isTop:s,isBottom:c,isLeft:d,isRight:h}=v1e(e),m={x:0,y:0},y={x:0,y:0},v={height:t.height*i.y,width:t.width*i.x};return!s&&a<=t.top+v.height?(m.y=lo.Backward,y.y=r*Math.abs((t.top+v.height-a)/v.height)):!c&&o>=t.bottom-v.height&&(m.y=lo.Forward,y.y=r*Math.abs((t.bottom-v.height-o)/v.height)),!h&&l>=t.right-v.width?(m.x=lo.Forward,y.x=r*Math.abs((t.right-v.width-l)/v.width)):!d&&u<=t.left+v.width&&(m.x=lo.Backward,y.x=r*Math.abs((t.left+v.width-u)/v.width)),{direction:m,speed:y}}function cet(e){if(e===document.scrollingElement){const{innerWidth:a,innerHeight:u}=window;return{top:0,left:0,right:a,bottom:u,width:a,height:u}}const{top:t,left:n,right:r,bottom:i}=e.getBoundingClientRect();return{top:t,left:n,right:r,bottom:i,width:e.clientWidth,height:e.clientHeight}}function y1e(e){return e.reduce((t,n)=>_3(t,DF(n)),fp)}function fet(e){return e.reduce((t,n)=>t+p1e(n),0)}function det(e){return e.reduce((t,n)=>t+m1e(n),0)}function het(e,t){if(t===void 0&&(t=y4),!e)return;const{top:n,left:r,bottom:i,right:a}=t(e);h1e(e)&&(i<=0||a<=0||n>=window.innerHeight||r>=window.innerWidth)&&e.scrollIntoView({block:"center",inline:"center"})}const pet=[["x",["left","right"],fet],["y",["top","bottom"],det]];class AB{constructor(t,n){this.rect=void 0,this.width=void 0,this.height=void 0,this.top=void 0,this.bottom=void 0,this.right=void 0,this.left=void 0;const r=y5(n),i=y1e(r);this.rect={...t},this.width=t.width,this.height=t.height;for(const[a,u,l]of pet)for(const o of u)Object.defineProperty(this,o,{get:()=>{const s=l(r),c=i[a]-s;return this.rect[o]+c},enumerable:!0});Object.defineProperty(this,"rect",{enumerable:!1})}}class ix{constructor(t){this.target=void 0,this.listeners=[],this.removeAll=()=>{this.listeners.forEach(n=>{var r;return(r=this.target)==null?void 0:r.removeEventListener(...n)})},this.target=t}add(t,n,r){var i;(i=this.target)==null||i.addEventListener(t,n,r),this.listeners.push([t,n,r])}}function met(e){const{EventTarget:t}=mf(e);return e instanceof t?e:v4(e)}function N7(e,t){const n=Math.abs(e.x),r=Math.abs(e.y);return typeof t=="number"?Math.sqrt(n**2+r**2)>t:"x"in t&&"y"in t?n>t.x&&r>t.y:"x"in t?n>t.x:"y"in t?r>t.y:!1}var Ud;(function(e){e.Click="click",e.DragStart="dragstart",e.Keydown="keydown",e.ContextMenu="contextmenu",e.Resize="resize",e.SelectionChange="selectionchange",e.VisibilityChange="visibilitychange"})(Ud||(Ud={}));function wW(e){e.preventDefault()}function get(e){e.stopPropagation()}var ri;(function(e){e.Space="Space",e.Down="ArrowDown",e.Right="ArrowRight",e.Left="ArrowLeft",e.Up="ArrowUp",e.Esc="Escape",e.Enter="Enter",e.Tab="Tab"})(ri||(ri={}));const b1e={start:[ri.Space,ri.Enter],cancel:[ri.Esc],end:[ri.Space,ri.Enter,ri.Tab]},vet=(e,t)=>{let{currentCoordinates:n}=t;switch(e.code){case ri.Right:return{...n,x:n.x+25};case ri.Left:return{...n,x:n.x-25};case ri.Down:return{...n,y:n.y+25};case ri.Up:return{...n,y:n.y-25}}};class O1e{constructor(t){this.props=void 0,this.autoScrollEnabled=!1,this.referenceCoordinates=void 0,this.listeners=void 0,this.windowListeners=void 0,this.props=t;const{event:{target:n}}=t;this.props=t,this.listeners=new ix(v4(n)),this.windowListeners=new ix(mf(n)),this.handleKeyDown=this.handleKeyDown.bind(this),this.handleCancel=this.handleCancel.bind(this),this.attach()}attach(){this.handleStart(),this.windowListeners.add(Ud.Resize,this.handleCancel),this.windowListeners.add(Ud.VisibilityChange,this.handleCancel),setTimeout(()=>this.listeners.add(Ud.Keydown,this.handleKeyDown))}handleStart(){const{activeNode:t,onStart:n}=this.props,r=t.node.current;r&&het(r),n(fp)}handleKeyDown(t){if(RB(t)){const{active:n,context:r,options:i}=this.props,{keyboardCodes:a=b1e,coordinateGetter:u=vet,scrollBehavior:l="smooth"}=i,{code:o}=t;if(a.end.includes(o)){this.handleEnd(t);return}if(a.cancel.includes(o)){this.handleCancel(t);return}const{collisionRect:s}=r.current,c=s?{x:s.left,y:s.top}:fp;this.referenceCoordinates||(this.referenceCoordinates=c);const d=u(t,{active:n,context:r.current,currentCoordinates:c});if(d){const h=rS(d,c),m={x:0,y:0},{scrollableAncestors:y}=r.current;for(const v of y){const S=t.code,{isTop:_,isRight:E,isLeft:$,isBottom:k,maxScroll:A,minScroll:w}=v1e(v),R=cet(v),F={x:Math.min(S===ri.Right?R.right-R.width/2:R.right,Math.max(S===ri.Right?R.left:R.left+R.width/2,d.x)),y:Math.min(S===ri.Down?R.bottom-R.height/2:R.bottom,Math.max(S===ri.Down?R.top:R.top+R.height/2,d.y))},I=S===ri.Right&&!E||S===ri.Left&&!$,U=S===ri.Down&&!k||S===ri.Up&&!_;if(I&&F.x!==d.x){const q=v.scrollLeft+h.x,V=S===ri.Right&&q<=A.x||S===ri.Left&&q>=w.x;if(V&&!h.y){v.scrollTo({left:q,behavior:l});return}V?m.x=v.scrollLeft-q:m.x=S===ri.Right?v.scrollLeft-A.x:v.scrollLeft-w.x,m.x&&v.scrollBy({left:-m.x,behavior:l});break}else if(U&&F.y!==d.y){const q=v.scrollTop+h.y,V=S===ri.Down&&q<=A.y||S===ri.Up&&q>=w.y;if(V&&!h.x){v.scrollTo({top:q,behavior:l});return}V?m.y=v.scrollTop-q:m.y=S===ri.Down?v.scrollTop-A.y:v.scrollTop-w.y,m.y&&v.scrollBy({top:-m.y,behavior:l});break}}this.handleMove(t,_3(rS(d,this.referenceCoordinates),m))}}}handleMove(t,n){const{onMove:r}=this.props;t.preventDefault(),r(n)}handleEnd(t){const{onEnd:n}=this.props;t.preventDefault(),this.detach(),n()}handleCancel(t){const{onCancel:n}=this.props;t.preventDefault(),this.detach(),n()}detach(){this.listeners.removeAll(),this.windowListeners.removeAll()}}O1e.activators=[{eventName:"onKeyDown",handler:(e,t,n)=>{let{keyboardCodes:r=b1e,onActivation:i}=t,{active:a}=n;const{code:u}=e.nativeEvent;if(r.start.includes(u)){const l=a.activatorNode.current;return l&&e.target!==l?!1:(e.preventDefault(),i==null||i({event:e.nativeEvent}),!0)}return!1}}];function RW(e){return!!(e&&"distance"in e)}function AW(e){return!!(e&&"delay"in e)}class TB{constructor(t,n,r){var i;r===void 0&&(r=met(t.event.target)),this.props=void 0,this.events=void 0,this.autoScrollEnabled=!0,this.document=void 0,this.activated=!1,this.initialCoordinates=void 0,this.timeoutId=null,this.listeners=void 0,this.documentListeners=void 0,this.windowListeners=void 0,this.props=t,this.events=n;const{event:a}=t,{target:u}=a;this.props=t,this.events=n,this.document=v4(u),this.documentListeners=new ix(this.document),this.listeners=new ix(r),this.windowListeners=new ix(mf(u)),this.initialCoordinates=(i=EF(a))!=null?i:fp,this.handleStart=this.handleStart.bind(this),this.handleMove=this.handleMove.bind(this),this.handleEnd=this.handleEnd.bind(this),this.handleCancel=this.handleCancel.bind(this),this.handleKeydown=this.handleKeydown.bind(this),this.removeTextSelection=this.removeTextSelection.bind(this),this.attach()}attach(){const{events:t,props:{options:{activationConstraint:n,bypassActivationConstraint:r}}}=this;if(this.listeners.add(t.move.name,this.handleMove,{passive:!1}),this.listeners.add(t.end.name,this.handleEnd),t.cancel&&this.listeners.add(t.cancel.name,this.handleCancel),this.windowListeners.add(Ud.Resize,this.handleCancel),this.windowListeners.add(Ud.DragStart,wW),this.windowListeners.add(Ud.VisibilityChange,this.handleCancel),this.windowListeners.add(Ud.ContextMenu,wW),this.documentListeners.add(Ud.Keydown,this.handleKeydown),n){if(r!=null&&r({event:this.props.event,activeNode:this.props.activeNode,options:this.props.options}))return this.handleStart();if(AW(n)){this.timeoutId=setTimeout(this.handleStart,n.delay),this.handlePending(n);return}if(RW(n)){this.handlePending(n);return}}this.handleStart()}detach(){this.listeners.removeAll(),this.windowListeners.removeAll(),setTimeout(this.documentListeners.removeAll,50),this.timeoutId!==null&&(clearTimeout(this.timeoutId),this.timeoutId=null)}handlePending(t,n){const{active:r,onPending:i}=this.props;i(r,t,this.initialCoordinates,n)}handleStart(){const{initialCoordinates:t}=this,{onStart:n}=this.props;t&&(this.activated=!0,this.documentListeners.add(Ud.Click,get,{capture:!0}),this.removeTextSelection(),this.documentListeners.add(Ud.SelectionChange,this.removeTextSelection),n(t))}handleMove(t){var n;const{activated:r,initialCoordinates:i,props:a}=this,{onMove:u,options:{activationConstraint:l}}=a;if(!i)return;const o=(n=EF(t))!=null?n:fp,s=rS(i,o);if(!r&&l){if(RW(l)){if(l.tolerance!=null&&N7(s,l.tolerance))return this.handleCancel();if(N7(s,l.distance))return this.handleStart()}if(AW(l)&&N7(s,l.tolerance))return this.handleCancel();this.handlePending(l,s);return}t.cancelable&&t.preventDefault(),u(o)}handleEnd(){const{onAbort:t,onEnd:n}=this.props;this.detach(),this.activated||t(this.props.active),n()}handleCancel(){const{onAbort:t,onCancel:n}=this.props;this.detach(),this.activated||t(this.props.active),n()}handleKeydown(t){t.code===ri.Esc&&this.handleCancel()}removeTextSelection(){var t;(t=this.document.getSelection())==null||t.removeAllRanges()}}const yet={cancel:{name:"pointercancel"},move:{name:"pointermove"},end:{name:"pointerup"}};class x1e extends TB{constructor(t){const{event:n}=t,r=v4(n.target);super(t,yet,r)}}x1e.activators=[{eventName:"onPointerDown",handler:(e,t)=>{let{nativeEvent:n}=e,{onActivation:r}=t;return!n.isPrimary||n.button!==0?!1:(r==null||r({event:n}),!0)}}];const bet={move:{name:"mousemove"},end:{name:"mouseup"}};var $F;(function(e){e[e.RightClick=2]="RightClick"})($F||($F={}));class Oet extends TB{constructor(t){super(t,bet,v4(t.event.target))}}Oet.activators=[{eventName:"onMouseDown",handler:(e,t)=>{let{nativeEvent:n}=e,{onActivation:r}=t;return n.button===$F.RightClick?!1:(r==null||r({event:n}),!0)}}];const I7={cancel:{name:"touchcancel"},move:{name:"touchmove"},end:{name:"touchend"}};class xet extends TB{constructor(t){super(t,I7)}static setup(){return window.addEventListener(I7.move.name,t,{capture:!1,passive:!1}),function(){window.removeEventListener(I7.move.name,t)};function t(){}}}xet.activators=[{eventName:"onTouchStart",handler:(e,t)=>{let{nativeEvent:n}=e,{onActivation:r}=t;const{touches:i}=n;return i.length>1?!1:(r==null||r({event:n}),!0)}}];var ax;(function(e){e[e.Pointer=0]="Pointer",e[e.DraggableRect=1]="DraggableRect"})(ax||(ax={}));var pC;(function(e){e[e.TreeOrder=0]="TreeOrder",e[e.ReversedTreeOrder=1]="ReversedTreeOrder"})(pC||(pC={}));function _et(e){let{acceleration:t,activator:n=ax.Pointer,canScroll:r,draggingRect:i,enabled:a,interval:u=5,order:l=pC.TreeOrder,pointerCoordinates:o,scrollableAncestors:s,scrollableAncestorRects:c,delta:d,threshold:h}=e;const m=Det({delta:d,disabled:!a}),[y,v]=PJe(),S=P.useRef({x:0,y:0}),_=P.useRef({x:0,y:0}),E=P.useMemo(()=>{switch(n){case ax.Pointer:return o?{top:o.y,bottom:o.y,left:o.x,right:o.x}:null;case ax.DraggableRect:return i}},[n,i,o]),$=P.useRef(null),k=P.useCallback(()=>{const w=$.current;if(!w)return;const R=S.current.x*_.current.x,F=S.current.y*_.current.y;w.scrollBy(R,F)},[]),A=P.useMemo(()=>l===pC.TreeOrder?[...s].reverse():s,[l,s]);P.useEffect(()=>{if(!a||!s.length||!E){v();return}for(const w of A){if((r==null?void 0:r(w))===!1)continue;const R=s.indexOf(w),F=c[R];if(!F)continue;const{direction:I,speed:U}=set(w,F,E,t,h);for(const q of["x","y"])m[q][I[q]]||(U[q]=0,I[q]=0);if(U.x>0||U.y>0){v(),$.current=w,y(k,u),S.current=U,_.current=I;return}}S.current={x:0,y:0},_.current={x:0,y:0},v()},[t,k,r,v,a,u,JSON.stringify(E),JSON.stringify(m),y,s,A,c,JSON.stringify(h)])}const Eet={x:{[lo.Backward]:!1,[lo.Forward]:!1},y:{[lo.Backward]:!1,[lo.Forward]:!1}};function Det(e){let{delta:t,disabled:n}=e;const r=_F(t);return jS(i=>{if(n||!r||!i)return Eet;const a={x:Math.sign(t.x-r.x),y:Math.sign(t.y-r.y)};return{x:{[lo.Backward]:i.x[lo.Backward]||a.x===-1,[lo.Forward]:i.x[lo.Forward]||a.x===1},y:{[lo.Backward]:i.y[lo.Backward]||a.y===-1,[lo.Forward]:i.y[lo.Forward]||a.y===1}}},[n,t,r])}function $et(e,t){const n=t!=null?e.get(t):void 0,r=n?n.node.current:null;return jS(i=>{var a;return t==null?null:(a=r??i)!=null?a:null},[r,t])}function Cet(e,t){return P.useMemo(()=>e.reduce((n,r)=>{const{sensor:i}=r,a=i.activators.map(u=>({eventName:u.eventName,handler:t(u.handler,r)}));return[...n,...a]},[]),[e,t])}var iS;(function(e){e[e.Always=0]="Always",e[e.BeforeDragging=1]="BeforeDragging",e[e.WhileDragging=2]="WhileDragging"})(iS||(iS={}));var CF;(function(e){e.Optimized="optimized"})(CF||(CF={}));const TW=new Map;function wet(e,t){let{dragging:n,dependencies:r,config:i}=t;const[a,u]=P.useState(null),{frequency:l,measure:o,strategy:s}=i,c=P.useRef(e),d=S(),h=nS(d),m=P.useCallback(function(_){_===void 0&&(_=[]),!h.current&&u(E=>E===null?_:E.concat(_.filter($=>!E.includes($))))},[h]),y=P.useRef(null),v=jS(_=>{if(d&&!n)return TW;if(!_||_===TW||c.current!==e||a!=null){const E=new Map;for(let $ of e){if(!$)continue;if(a&&a.length>0&&!a.includes($.id)&&$.rect.current){E.set($.id,$.rect.current);continue}const k=$.node.current,A=k?new AB(o(k),k):null;$.rect.current=A,A&&E.set($.id,A)}return E}return _},[e,a,n,d,o]);return P.useEffect(()=>{c.current=e},[e]),P.useEffect(()=>{d||m()},[n,d]),P.useEffect(()=>{a&&a.length>0&&u(null)},[JSON.stringify(a)]),P.useEffect(()=>{d||typeof l!="number"||y.current!==null||(y.current=setTimeout(()=>{m(),y.current=null},l))},[l,d,m,...r]),{droppableRects:v,measureDroppableContainers:m,measuringScheduled:a!=null};function S(){switch(s){case iS.Always:return!1;case iS.BeforeDragging:return n;default:return!n}}}function S1e(e,t){return jS(n=>e?n||(typeof t=="function"?t(e):e):null,[t,e])}function Ret(e,t){return S1e(e,t)}function Aet(e){let{callback:t,disabled:n}=e;const r=wB(t),i=P.useMemo(()=>{if(n||typeof window>"u"||typeof window.MutationObserver>"u")return;const{MutationObserver:a}=window;return new a(r)},[r,n]);return P.useEffect(()=>()=>i==null?void 0:i.disconnect(),[i]),i}function b5(e){let{callback:t,disabled:n}=e;const r=wB(t),i=P.useMemo(()=>{if(n||typeof window>"u"||typeof window.ResizeObserver>"u")return;const{ResizeObserver:a}=window;return new a(r)},[n]);return P.useEffect(()=>()=>i==null?void 0:i.disconnect(),[i]),i}function Tet(e){return new AB(y4(e),e)}function kW(e,t,n){t===void 0&&(t=Tet);const[r,i]=P.useState(null);function a(){i(o=>{if(!e)return null;if(e.isConnected===!1){var s;return(s=o??n)!=null?s:null}const c=t(e);return JSON.stringify(o)===JSON.stringify(c)?o:c})}const u=Aet({callback(o){if(e)for(const s of o){const{type:c,target:d}=s;if(c==="childList"&&d instanceof HTMLElement&&d.contains(e)){a();break}}}}),l=b5({callback:a});return F0(()=>{a(),e?(l==null||l.observe(e),u==null||u.observe(document.body,{childList:!0,subtree:!0})):(l==null||l.disconnect(),u==null||u.disconnect())},[e]),r}function ket(e){const t=S1e(e);return d1e(e,t)}const PW=[];function Pet(e){const t=P.useRef(e),n=jS(r=>e?r&&r!==PW&&e&&t.current&&e.parentNode===t.current.parentNode?r:y5(e):PW,[e]);return P.useEffect(()=>{t.current=e},[e]),n}function Met(e){const[t,n]=P.useState(null),r=P.useRef(e),i=P.useCallback(a=>{const u=F7(a.target);u&&n(l=>l?(l.set(u,DF(u)),new Map(l)):null)},[]);return P.useEffect(()=>{const a=r.current;if(e!==a){u(a);const l=e.map(o=>{const s=F7(o);return s?(s.addEventListener("scroll",i,{passive:!0}),[s,DF(s)]):null}).filter(o=>o!=null);n(l.length?new Map(l):null),r.current=e}return()=>{u(e),u(a)};function u(l){l.forEach(o=>{const s=F7(o);s==null||s.removeEventListener("scroll",i)})}},[i,e]),P.useMemo(()=>e.length?t?Array.from(t.values()).reduce((a,u)=>_3(a,u),fp):y1e(e):fp,[e,t])}function MW(e,t){t===void 0&&(t=[]);const n=P.useRef(null);return P.useEffect(()=>{n.current=null},t),P.useEffect(()=>{const r=e!==fp;r&&!n.current&&(n.current=e),!r&&n.current&&(n.current=null)},[e]),n.current?rS(e,n.current):fp}function Fet(e){P.useEffect(()=>{if(!v5)return;const t=e.map(n=>{let{sensor:r}=n;return r.setup==null?void 0:r.setup()});return()=>{for(const n of t)n==null||n()}},e.map(t=>{let{sensor:n}=t;return n}))}function Net(e,t){return P.useMemo(()=>e.reduce((n,r)=>{let{eventName:i,handler:a}=r;return n[i]=u=>{a(u,t)},n},{}),[e,t])}function _1e(e){return P.useMemo(()=>e?iet(e):null,[e])}const FW=[];function Iet(e,t){t===void 0&&(t=y4);const[n]=e,r=_1e(n?mf(n):null),[i,a]=P.useState(FW);function u(){a(()=>e.length?e.map(o=>g1e(o)?r:new AB(t(o),o)):FW)}const l=b5({callback:u});return F0(()=>{l==null||l.disconnect(),u(),e.forEach(o=>l==null?void 0:l.observe(o))},[e]),i}function Let(e){if(!e)return null;if(e.children.length>1)return e;const t=e.children[0];return US(t)?t:e}function Bet(e){let{measure:t}=e;const[n,r]=P.useState(null),i=P.useCallback(s=>{for(const{target:c}of s)if(US(c)){r(d=>{const h=t(c);return d?{...d,width:h.width,height:h.height}:h});break}},[t]),a=b5({callback:i}),u=P.useCallback(s=>{const c=Let(s);a==null||a.disconnect(),c&&(a==null||a.observe(c)),r(c?t(c):null)},[t,a]),[l,o]=fC(u);return P.useMemo(()=>({nodeRef:l,rect:n,setRef:o}),[n,l,o])}const Qet=[{sensor:x1e,options:{}},{sensor:O1e,options:{}}],zet={current:{}},DD={draggable:{measure:CW},droppable:{measure:CW,strategy:iS.WhileDragging,frequency:CF.Optimized},dragOverlay:{measure:y4}};class ux extends Map{get(t){var n;return t!=null&&(n=super.get(t))!=null?n:void 0}toArray(){return Array.from(this.values())}getEnabled(){return this.toArray().filter(t=>{let{disabled:n}=t;return!n})}getNodeFor(t){var n,r;return(n=(r=this.get(t))==null?void 0:r.node.current)!=null?n:void 0}}const qet={activatorEvent:null,active:null,activeNode:null,activeNodeRect:null,collisions:null,containerNodeRect:null,draggableNodes:new Map,droppableRects:new Map,droppableContainers:new ux,over:null,dragOverlay:{nodeRef:{current:null},rect:null,setRef:hC},scrollableAncestors:[],scrollableAncestorRects:[],measuringConfiguration:DD,measureDroppableContainers:hC,windowRect:null,measuringScheduled:!1},Vet={activatorEvent:null,activators:[],active:null,activeNodeRect:null,ariaDescribedById:{draggable:""},dispatch:hC,draggableNodes:new Map,over:null,measureDroppableContainers:hC},O5=P.createContext(Vet),E1e=P.createContext(qet);function Uet(){return{draggable:{active:null,initialCoordinates:{x:0,y:0},nodes:new Map,translate:{x:0,y:0}},droppable:{containers:new ux}}}function jet(e,t){switch(t.type){case kl.DragStart:return{...e,draggable:{...e.draggable,initialCoordinates:t.initialCoordinates,active:t.active}};case kl.DragMove:return e.draggable.active==null?e:{...e,draggable:{...e.draggable,translate:{x:t.coordinates.x-e.draggable.initialCoordinates.x,y:t.coordinates.y-e.draggable.initialCoordinates.y}}};case kl.DragEnd:case kl.DragCancel:return{...e,draggable:{...e.draggable,active:null,initialCoordinates:{x:0,y:0},translate:{x:0,y:0}}};case kl.RegisterDroppable:{const{element:n}=t,{id:r}=n,i=new ux(e.droppable.containers);return i.set(r,n),{...e,droppable:{...e.droppable,containers:i}}}case kl.SetDroppableDisabled:{const{id:n,key:r,disabled:i}=t,a=e.droppable.containers.get(n);if(!a||r!==a.key)return e;const u=new ux(e.droppable.containers);return u.set(n,{...a,disabled:i}),{...e,droppable:{...e.droppable,containers:u}}}case kl.UnregisterDroppable:{const{id:n,key:r}=t,i=e.droppable.containers.get(n);if(!i||r!==i.key)return e;const a=new ux(e.droppable.containers);return a.delete(n),{...e,droppable:{...e.droppable,containers:a}}}default:return e}}function Zet(e){let{disabled:t}=e;const{active:n,activatorEvent:r,draggableNodes:i}=P.useContext(O5),a=_F(r),u=_F(n==null?void 0:n.id);return P.useEffect(()=>{if(!t&&!r&&a&&u!=null){if(!RB(a)||document.activeElement===a.target)return;const l=i.get(u);if(!l)return;const{activatorNode:o,node:s}=l;if(!o.current&&!s.current)return;requestAnimationFrame(()=>{for(const c of[o.current,s.current]){if(!c)continue;const d=NJe(c);if(d){d.focus();break}}})}},[r,t,i,u,a]),null}function Het(e,t){let{transform:n,...r}=t;return e!=null&&e.length?e.reduce((i,a)=>a({transform:i,...r}),n):n}function Xet(e){return P.useMemo(()=>({draggable:{...DD.draggable,...e==null?void 0:e.draggable},droppable:{...DD.droppable,...e==null?void 0:e.droppable},dragOverlay:{...DD.dragOverlay,...e==null?void 0:e.dragOverlay}}),[e==null?void 0:e.draggable,e==null?void 0:e.droppable,e==null?void 0:e.dragOverlay])}function Yet(e){let{activeNode:t,measure:n,initialRect:r,config:i=!0}=e;const a=P.useRef(!1),{x:u,y:l}=typeof i=="boolean"?{x:i,y:i}:i;F0(()=>{if(!u&&!l||!t){a.current=!1;return}if(a.current||!r)return;const s=t==null?void 0:t.node.current;if(!s||s.isConnected===!1)return;const c=n(s),d=d1e(c,r);if(u||(d.x=0),l||(d.y=0),a.current=!0,Math.abs(d.x)>0||Math.abs(d.y)>0){const h=h1e(s);h&&h.scrollBy({top:d.y,left:d.x})}},[t,u,l,r,n])}const D1e=P.createContext({...fp,scaleX:1,scaleY:1});var Cv;(function(e){e[e.Uninitialized=0]="Uninitialized",e[e.Initializing=1]="Initializing",e[e.Initialized=2]="Initialized"})(Cv||(Cv={}));const Tlt=P.memo(function(t){var n,r,i,a;let{id:u,accessibility:l,autoScroll:o=!0,children:s,sensors:c=Qet,collisionDetection:d=KJe,measuring:h,modifiers:m,...y}=t;const v=P.useReducer(jet,void 0,Uet),[S,_]=v,[E,$]=qJe(),[k,A]=P.useState(Cv.Uninitialized),w=k===Cv.Initialized,{draggable:{active:R,nodes:F,translate:I},droppable:{containers:U}}=S,q=R!=null?F.get(R):null,V=P.useRef({initial:null,translated:null}),N=P.useMemo(()=>{var rn;return R!=null?{id:R,data:(rn=q==null?void 0:q.data)!=null?rn:zet,rect:V}:null},[R,q]),j=P.useRef(null),[W,K]=P.useState(null),[J,te]=P.useState(null),H=nS(y,Object.values(y)),M=ZS("DndDescribedBy",u),G=P.useMemo(()=>U.getEnabled(),[U]),X=Xet(h),{droppableRects:B,measureDroppableContainers:ne,measuringScheduled:oe}=wet(G,{dragging:w,dependencies:[I.x,I.y],config:X.droppable}),ie=$et(F,R),de=P.useMemo(()=>J?EF(J):null,[J]),ge=In(),ye=Ret(ie,X.draggable.measure);Yet({activeNode:R!=null?F.get(R):null,config:ge.layoutShiftCompensation,initialRect:ye,measure:X.draggable.measure});const me=kW(ie,X.draggable.measure,ye),Ee=kW(ie?ie.parentElement:null),se=P.useRef({activatorEvent:null,active:null,activeNode:ie,collisionRect:null,collisions:null,droppableRects:B,draggableNodes:F,draggingNode:null,draggingNodeRect:null,droppableContainers:U,over:null,scrollableAncestors:[],scrollAdjustedTranslate:null}),pe=U.getNodeFor((n=se.current.over)==null?void 0:n.id),ve=Bet({measure:X.dragOverlay.measure}),we=(r=ve.nodeRef.current)!=null?r:ie,Me=w?(i=ve.rect)!=null?i:me:null,Re=!!(ve.nodeRef.current&&ve.rect),et=ket(Re?null:me),tt=_1e(we?mf(we):null),Oe=Pet(w?pe??ie:null),$e=Iet(Oe),Ae=Het(m,{transform:{x:I.x-et.x,y:I.y-et.y,scaleX:1,scaleY:1},activatorEvent:J,active:N,activeNodeRect:me,containerNodeRect:Ee,draggingNodeRect:Me,over:se.current.over,overlayNodeRect:ve.rect,scrollableAncestors:Oe,scrollableAncestorRects:$e,windowRect:tt}),He=de?_3(de,I):null,je=Met(Oe),Ye=MW(je),Ne=MW(je,[me]),ze=_3(Ae,Ye),at=Me?eet(Me,Ae):null,mt=N&&at?d({active:N,collisionRect:at,droppableRects:B,droppableContainers:G,pointerCoordinates:He}):null,Mt=f1e(mt,"id"),[bt,zt]=P.useState(null),qt=Re?Ae:_3(Ae,Ne),Xt=WJe(qt,(a=bt==null?void 0:bt.rect)!=null?a:null,me),nn=P.useRef(null),Fe=P.useCallback((rn,Xn)=>{let{sensor:br,options:Or}=Xn;if(j.current==null)return;const Cr=F.get(j.current);if(!Cr)return;const wr=rn.nativeEvent,Gr=new br({active:j.current,activeNode:Cr,event:wr,options:Or,context:se,onAbort(ai){if(!F.get(ai))return;const{onDragAbort:Te}=H.current,Ve={id:ai};Te==null||Te(Ve),E({type:"onDragAbort",event:Ve})},onPending(ai,ar,Te,Ve){if(!F.get(ai))return;const{onDragPending:gt}=H.current,Ct={id:ai,constraint:ar,initialCoordinates:Te,offset:Ve};gt==null||gt(Ct),E({type:"onDragPending",event:Ct})},onStart(ai){const ar=j.current;if(ar==null)return;const Te=F.get(ar);if(!Te)return;const{onDragStart:Ve}=H.current,nt={activatorEvent:wr,active:{id:ar,data:Te.data,rect:V}};Uf.unstable_batchedUpdates(()=>{Ve==null||Ve(nt),A(Cv.Initializing),_({type:kl.DragStart,initialCoordinates:ai,active:ar}),E({type:"onDragStart",event:nt}),K(nn.current),te(wr)})},onMove(ai){_({type:kl.DragMove,coordinates:ai})},onEnd:ji(kl.DragEnd),onCancel:ji(kl.DragCancel)});nn.current=Gr;function ji(ai){return async function(){const{active:Te,collisions:Ve,over:nt,scrollAdjustedTranslate:gt}=se.current;let Ct=null;if(Te&&gt){const{cancelDrop:Vt}=H.current;Ct={activatorEvent:wr,active:Te,collisions:Ve,delta:gt,over:nt},ai===kl.DragEnd&&typeof Vt=="function"&&await Promise.resolve(Vt(Ct))&&(ai=kl.DragCancel)}j.current=null,Uf.unstable_batchedUpdates(()=>{_({type:ai}),A(Cv.Uninitialized),zt(null),K(null),te(null),nn.current=null;const Vt=ai===kl.DragEnd?"onDragEnd":"onDragCancel";if(Ct){const Yt=H.current[Vt];Yt==null||Yt(Ct),E({type:Vt,event:Ct})}})}}},[F]),ht=P.useCallback((rn,Xn)=>(br,Or)=>{const Cr=br.nativeEvent,wr=F.get(Or);if(j.current!==null||!wr||Cr.dndKit||Cr.defaultPrevented)return;const Gr={active:wr};rn(br,Xn.options,Gr)===!0&&(Cr.dndKit={capturedBy:Xn.sensor},j.current=Or,Fe(br,Xn))},[F,Fe]),kt=Cet(c,ht);Fet(c),F0(()=>{me&&k===Cv.Initializing&&A(Cv.Initialized)},[me,k]),P.useEffect(()=>{const{onDragMove:rn}=H.current,{active:Xn,activatorEvent:br,collisions:Or,over:Cr}=se.current;if(!Xn||!br)return;const wr={active:Xn,activatorEvent:br,collisions:Or,delta:{x:ze.x,y:ze.y},over:Cr};Uf.unstable_batchedUpdates(()=>{rn==null||rn(wr),E({type:"onDragMove",event:wr})})},[ze.x,ze.y]),P.useEffect(()=>{const{active:rn,activatorEvent:Xn,collisions:br,droppableContainers:Or,scrollAdjustedTranslate:Cr}=se.current;if(!rn||j.current==null||!Xn||!Cr)return;const{onDragOver:wr}=H.current,Gr=Or.get(Mt),ji=Gr&&Gr.rect.current?{id:Gr.id,rect:Gr.rect.current,data:Gr.data,disabled:Gr.disabled}:null,ai={active:rn,activatorEvent:Xn,collisions:br,delta:{x:Cr.x,y:Cr.y},over:ji};Uf.unstable_batchedUpdates(()=>{zt(ji),wr==null||wr(ai),E({type:"onDragOver",event:ai})})},[Mt]),F0(()=>{se.current={activatorEvent:J,active:N,activeNode:ie,collisionRect:at,collisions:mt,droppableRects:B,draggableNodes:F,draggingNode:we,draggingNodeRect:Me,droppableContainers:U,over:bt,scrollableAncestors:Oe,scrollAdjustedTranslate:ze},V.current={initial:Me,translated:at}},[N,ie,mt,at,F,we,Me,B,U,bt,Oe,ze]),_et({...ge,delta:I,draggingRect:at,pointerCoordinates:He,scrollableAncestors:Oe,scrollableAncestorRects:$e});const Qt=P.useMemo(()=>({active:N,activeNode:ie,activeNodeRect:me,activatorEvent:J,collisions:mt,containerNodeRect:Ee,dragOverlay:ve,draggableNodes:F,droppableContainers:U,droppableRects:B,over:bt,measureDroppableContainers:ne,scrollableAncestors:Oe,scrollableAncestorRects:$e,measuringConfiguration:X,measuringScheduled:oe,windowRect:tt}),[N,ie,me,J,mt,Ee,ve,F,U,B,bt,ne,Oe,$e,X,oe,tt]),en=P.useMemo(()=>({activatorEvent:J,activators:kt,active:N,activeNodeRect:me,ariaDescribedById:{draggable:M},dispatch:_,draggableNodes:F,over:bt,measureDroppableContainers:ne}),[J,kt,N,me,_,M,F,bt,ne]);return be.createElement(c1e.Provider,{value:$},be.createElement(O5.Provider,{value:en},be.createElement(E1e.Provider,{value:Qt},be.createElement(D1e.Provider,{value:Xt},s)),be.createElement(Zet,{disabled:(l==null?void 0:l.restoreFocus)===!1})),be.createElement(jJe,{...l,hiddenTextDescribedById:M}));function In(){const rn=(W==null?void 0:W.autoScrollEnabled)===!1,Xn=typeof o=="object"?o.enabled===!1:o===!1,br=w&&!rn&&!Xn;return typeof o=="object"?{...o,enabled:br}:{enabled:br}}}),Get=P.createContext(null),NW="button",Ket="Draggable";function Wet(e){let{id:t,data:n,disabled:r=!1,attributes:i}=e;const a=ZS(Ket),{activators:u,activatorEvent:l,active:o,activeNodeRect:s,ariaDescribedById:c,draggableNodes:d,over:h}=P.useContext(O5),{role:m=NW,roleDescription:y="draggable",tabIndex:v=0}=i??{},S=(o==null?void 0:o.id)===t,_=P.useContext(S?D1e:Get),[E,$]=fC(),[k,A]=fC(),w=Net(u,t),R=nS(n);F0(()=>(d.set(t,{id:t,key:a,node:E,activatorNode:k,data:R}),()=>{const I=d.get(t);I&&I.key===a&&d.delete(t)}),[d,t]);const F=P.useMemo(()=>({role:m,tabIndex:v,"aria-disabled":r,"aria-pressed":S&&m===NW?!0:void 0,"aria-roledescription":y,"aria-describedby":c.draggable}),[r,m,v,S,y,c.draggable]);return{active:o,activatorEvent:l,activeNodeRect:s,attributes:F,isDragging:S,listeners:r?void 0:w,node:E,over:h,setNodeRef:$,setActivatorNodeRef:A,transform:_}}function Jet(){return P.useContext(E1e)}const ett="Droppable",ttt={timeout:25};function ntt(e){let{data:t,disabled:n=!1,id:r,resizeObserverConfig:i}=e;const a=ZS(ett),{active:u,dispatch:l,over:o,measureDroppableContainers:s}=P.useContext(O5),c=P.useRef({disabled:n}),d=P.useRef(!1),h=P.useRef(null),m=P.useRef(null),{disabled:y,updateMeasurementsFor:v,timeout:S}={...ttt,...i},_=nS(v??r),E=P.useCallback(()=>{if(!d.current){d.current=!0;return}m.current!=null&&clearTimeout(m.current),m.current=setTimeout(()=>{s(Array.isArray(_.current)?_.current:[_.current]),m.current=null},S)},[S]),$=b5({callback:E,disabled:y||!u}),k=P.useCallback((F,I)=>{$&&(I&&($.unobserve(I),d.current=!1),F&&$.observe(F))},[$]),[A,w]=fC(k),R=nS(t);return P.useEffect(()=>{!$||!A.current||($.disconnect(),d.current=!1,$.observe(A.current))},[A,$]),P.useEffect(()=>(l({type:kl.RegisterDroppable,element:{id:r,key:a,disabled:n,node:A,rect:h,data:R}}),()=>l({type:kl.UnregisterDroppable,key:a,id:r})),[r]),P.useEffect(()=>{n!==c.current.disabled&&(l({type:kl.SetDroppableDisabled,id:r,key:a,disabled:n}),c.current.disabled=n)},[r,a,n,l]),{active:u,rect:h,isOver:(o==null?void 0:o.id)===r,node:A,over:o,setNodeRef:w}}function $1e(e,t,n){const r=e.slice();return r.splice(n<0?r.length+n:n,0,r.splice(t,1)[0]),r}function rtt(e,t){return e.reduce((n,r,i)=>{const a=t.get(r);return a&&(n[i]=a),n},Array(e.length))}function BE(e){return e!==null&&e>=0}function itt(e,t){if(e===t)return!0;if(e.length!==t.length)return!1;for(let n=0;n<e.length;n++)if(e[n]!==t[n])return!1;return!0}function att(e){return typeof e=="boolean"?{draggable:e,droppable:e}:e}const C1e=e=>{let{rects:t,activeIndex:n,overIndex:r,index:i}=e;const a=$1e(t,r,n),u=t[i],l=a[i];return!l||!u?null:{x:l.left-u.left,y:l.top-u.top,scaleX:l.width/u.width,scaleY:l.height/u.height}},w1e="Sortable",R1e=be.createContext({activeIndex:-1,containerId:w1e,disableTransforms:!1,items:[],overIndex:-1,useDragOverlay:!1,sortedRects:[],strategy:C1e,disabled:{draggable:!1,droppable:!1}});function klt(e){let{children:t,id:n,items:r,strategy:i=C1e,disabled:a=!1}=e;const{active:u,dragOverlay:l,droppableRects:o,over:s,measureDroppableContainers:c}=Jet(),d=ZS(w1e,n),h=l.rect!==null,m=P.useMemo(()=>r.map(w=>typeof w=="object"&&"id"in w?w.id:w),[r]),y=u!=null,v=u?m.indexOf(u.id):-1,S=s?m.indexOf(s.id):-1,_=P.useRef(m),E=!itt(m,_.current),$=S!==-1&&v===-1||E,k=att(a);F0(()=>{E&&y&&c(m)},[E,m,y,c]),P.useEffect(()=>{_.current=m},[m]);const A=P.useMemo(()=>({activeIndex:v,containerId:d,disabled:k,disableTransforms:$,items:m,overIndex:S,useDragOverlay:h,sortedRects:rtt(m,o),strategy:i}),[v,d,k.draggable,k.droppable,$,m,S,o,h,i]);return be.createElement(R1e.Provider,{value:A},t)}const utt=e=>{let{id:t,items:n,activeIndex:r,overIndex:i}=e;return $1e(n,r,i).indexOf(t)},ltt=e=>{let{containerId:t,isSorting:n,wasDragging:r,index:i,items:a,newIndex:u,previousItems:l,previousContainerId:o,transition:s}=e;return!s||!r||l!==a&&i===u?!1:n?!0:u!==i&&t===o},ott={duration:200,easing:"ease"},A1e="transform",stt=dC.Transition.toString({property:A1e,duration:0,easing:"linear"}),ctt={roleDescription:"sortable"};function ftt(e){let{disabled:t,index:n,node:r,rect:i}=e;const[a,u]=P.useState(null),l=P.useRef(n);return F0(()=>{if(!t&&n!==l.current&&r.current){const o=i.current;if(o){const s=y4(r.current,{ignoreTransform:!0}),c={x:o.left-s.left,y:o.top-s.top,scaleX:o.width/s.width,scaleY:o.height/s.height};(c.x||c.y)&&u(c)}}n!==l.current&&(l.current=n)},[t,n,r,i]),P.useEffect(()=>{a&&u(null)},[a]),a}function Plt(e){let{animateLayoutChanges:t=ltt,attributes:n,disabled:r,data:i,getNewIndex:a=utt,id:u,strategy:l,resizeObserverConfig:o,transition:s=ott}=e;const{items:c,containerId:d,activeIndex:h,disabled:m,disableTransforms:y,sortedRects:v,overIndex:S,useDragOverlay:_,strategy:E}=P.useContext(R1e),$=dtt(r,m),k=c.indexOf(u),A=P.useMemo(()=>({sortable:{containerId:d,index:k,items:c},...i}),[d,i,k,c]),w=P.useMemo(()=>c.slice(c.indexOf(u)),[c,u]),{rect:R,node:F,isOver:I,setNodeRef:U}=ntt({id:u,data:A,disabled:$.droppable,resizeObserverConfig:{updateMeasurementsFor:w,...o}}),{active:q,activatorEvent:V,activeNodeRect:N,attributes:j,setNodeRef:W,listeners:K,isDragging:J,over:te,setActivatorNodeRef:H,transform:M}=Wet({id:u,data:A,attributes:{...ctt,...n},disabled:$.draggable}),G=kJe(U,W),X=!!q,B=X&&!y&&BE(h)&&BE(S),ne=!_&&J,oe=ne&&B?M:null,de=B?oe??(l??E)({rects:v,activeNodeRect:N,activeIndex:h,overIndex:S,index:k}):null,ge=BE(h)&&BE(S)?a({id:u,items:c,activeIndex:h,overIndex:S}):k,ye=q==null?void 0:q.id,me=P.useRef({activeId:ye,items:c,newIndex:ge,containerId:d}),Ee=c!==me.current.items,se=t({active:q,containerId:d,isDragging:J,isSorting:X,id:u,index:k,items:c,newIndex:me.current.newIndex,previousItems:me.current.items,previousContainerId:me.current.containerId,transition:s,wasDragging:me.current.activeId!=null}),pe=ftt({disabled:!se,index:k,node:F,rect:R});return P.useEffect(()=>{X&&me.current.newIndex!==ge&&(me.current.newIndex=ge),d!==me.current.containerId&&(me.current.containerId=d),c!==me.current.items&&(me.current.items=c)},[X,ge,d,c]),P.useEffect(()=>{if(ye===me.current.activeId)return;if(ye&&!me.current.activeId){me.current.activeId=ye;return}const we=setTimeout(()=>{me.current.activeId=ye},50);return()=>clearTimeout(we)},[ye]),{active:q,activeIndex:h,attributes:j,data:A,rect:R,index:k,newIndex:ge,items:c,isOver:I,isSorting:X,isDragging:J,listeners:K,node:F,overIndex:S,over:te,setNodeRef:G,setActivatorNodeRef:H,setDroppableNodeRef:U,setDraggableNodeRef:W,transform:pe??de,transition:ve()};function ve(){if(pe||Ee&&me.current.newIndex===k)return stt;if(!(ne&&!RB(V)||!s)&&(X||se))return dC.Transition.toString({...s,property:A1e})}}function dtt(e,t){var n,r;return typeof e=="boolean"?{draggable:e,droppable:!1}:{draggable:(n=e==null?void 0:e.draggable)!=null?n:t.draggable,droppable:(r=e==null?void 0:e.droppable)!=null?r:t.droppable}}function mC(e){if(!e)return!1;const t=e.data.current;return!!(t&&"sortable"in t&&typeof t.sortable=="object"&&"containerId"in t.sortable&&"items"in t.sortable&&"index"in t.sortable)}const htt=[ri.Down,ri.Right,ri.Up,ri.Left],Mlt=(e,t)=>{let{context:{active:n,collisionRect:r,droppableRects:i,droppableContainers:a,over:u,scrollableAncestors:l}}=t;if(htt.includes(e.code)){if(e.preventDefault(),!n||!r)return;const o=[];a.getEnabled().forEach(d=>{if(!d||d!=null&&d.disabled)return;const h=i.get(d.id);if(h)switch(e.code){case ri.Down:r.top<h.top&&o.push(d);break;case ri.Up:r.top>h.top&&o.push(d);break;case ri.Left:r.left>h.left&&o.push(d);break;case ri.Right:r.left<h.left&&o.push(d);break}});const s=YJe({collisionRect:r,droppableRects:i,droppableContainers:o});let c=f1e(s,"id");if(c===(u==null?void 0:u.id)&&s.length>1&&(c=s[1].id),c!=null){const d=a.get(n.id),h=a.get(c),m=h?i.get(h.id):null,y=h==null?void 0:h.node.current;if(y&&m&&d&&h){const S=y5(y).some((w,R)=>l[R]!==w),_=T1e(d,h),E=ptt(d,h),$=S||!_?{x:0,y:0}:{x:E?r.width-m.width:0,y:E?r.height-m.height:0},k={x:m.left,y:m.top};return $.x&&$.y?k:rS(k,$)}}}};function T1e(e,t){return!mC(e)||!mC(t)?!1:e.data.current.sortable.containerId===t.data.current.sortable.containerId}function ptt(e,t){return!mC(e)||!mC(t)||!T1e(e,t)?!1:e.data.current.sortable.index<t.data.current.sortable.index}var IW=Object.prototype.hasOwnProperty;function LW(e,t,n){for(n of e.keys())if(lx(n,t))return n}function lx(e,t){var n,r,i;if(e===t)return!0;if(e&&t&&(n=e.constructor)===t.constructor){if(n===Date)return e.getTime()===t.getTime();if(n===RegExp)return e.toString()===t.toString();if(n===Array){if((r=e.length)===t.length)for(;r--&&lx(e[r],t[r]););return r===-1}if(n===Set){if(e.size!==t.size)return!1;for(r of e)if(i=r,i&&typeof i=="object"&&(i=LW(t,i),!i)||!t.has(i))return!1;return!0}if(n===Map){if(e.size!==t.size)return!1;for(r of e)if(i=r[0],i&&typeof i=="object"&&(i=LW(t,i),!i)||!lx(r[1],t.get(i)))return!1;return!0}if(n===ArrayBuffer)e=new Uint8Array(e),t=new Uint8Array(t);else if(n===DataView){if((r=e.byteLength)===t.byteLength)for(;r--&&e.getInt8(r)===t.getInt8(r););return r===-1}if(ArrayBuffer.isView(e)){if((r=e.byteLength)===t.byteLength)for(;r--&&e[r]===t[r];);return r===-1}if(!n||typeof e=="object"){r=0;for(n in e)if(IW.call(e,n)&&++r&&!IW.call(t,n)||!(n in t)||!lx(e[n],t[n]))return!1;return Object.keys(t).length===r}}return e!==e&&t!==t}function mtt(e){var t=P.useRef(e),n=P.useRef(0);return lx(e,t.current)||(t.current=e,n.current+=1),P.useMemo(function(){return t.current},[n.current])}function Flt(e,t){return P.useEffect(e,mtt(t))}var L7={exports:{}},Tm={};/**
 * @license React
 * react-reconciler-constants.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var BW;function gtt(){return BW||(BW=1,Tm.ConcurrentRoot=1,Tm.ContinuousEventPriority=8,Tm.DefaultEventPriority=32,Tm.DiscreteEventPriority=2,Tm.IdleEventPriority=268435456,Tm.LegacyRoot=0,Tm.NoEventPriority=0),Tm}var QW;function vtt(){return QW||(QW=1,L7.exports=gtt()),L7.exports}var $D=vtt();const zW=e=>{let t;const n=new Set,r=(s,c)=>{const d=typeof s=="function"?s(t):s;if(!Object.is(d,t)){const h=t;t=c??(typeof d!="object"||d===null)?d:Object.assign({},t,d),n.forEach(m=>m(t,h))}},i=()=>t,l={setState:r,getState:i,getInitialState:()=>o,subscribe:s=>(n.add(s),()=>n.delete(s))},o=t=e(r,i,l);return l},ytt=e=>e?zW(e):zW,{useSyncExternalStoreWithSelector:btt}=See,Ott=e=>e;function xtt(e,t=Ott,n){const r=btt(e.subscribe,e.getState,e.getInitialState,t,n);return be.useDebugValue(r),r}const qW=(e,t)=>{const n=ytt(e),r=(i,a=t)=>xtt(n,i,a);return Object.assign(r,n),r},Stt=(e,t)=>e?qW(e,t):qW;var B7={exports:{}},Q7={exports:{}},z7={exports:{}},q7={};/**
 * @license React
 * scheduler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var VW;function _tt(){return VW||(VW=1,(function(e){function t(K,J){var te=K.length;K.push(J);e:for(;0<te;){var H=te-1>>>1,M=K[H];if(0<i(M,J))K[H]=J,K[te]=M,te=H;else break e}}function n(K){return K.length===0?null:K[0]}function r(K){if(K.length===0)return null;var J=K[0],te=K.pop();if(te!==J){K[0]=te;e:for(var H=0,M=K.length,G=M>>>1;H<G;){var X=2*(H+1)-1,B=K[X],ne=X+1,oe=K[ne];if(0>i(B,te))ne<M&&0>i(oe,B)?(K[H]=oe,K[ne]=te,H=ne):(K[H]=B,K[X]=te,H=X);else if(ne<M&&0>i(oe,te))K[H]=oe,K[ne]=te,H=ne;else break e}}return J}function i(K,J){var te=K.sortIndex-J.sortIndex;return te!==0?te:K.id-J.id}if(e.unstable_now=void 0,typeof performance=="object"&&typeof performance.now=="function"){var a=performance;e.unstable_now=function(){return a.now()}}else{var u=Date,l=u.now();e.unstable_now=function(){return u.now()-l}}var o=[],s=[],c=1,d=null,h=3,m=!1,y=!1,v=!1,S=typeof setTimeout=="function"?setTimeout:null,_=typeof clearTimeout=="function"?clearTimeout:null,E=typeof setImmediate<"u"?setImmediate:null;function $(K){for(var J=n(s);J!==null;){if(J.callback===null)r(s);else if(J.startTime<=K)r(s),J.sortIndex=J.expirationTime,t(o,J);else break;J=n(s)}}function k(K){if(v=!1,$(K),!y)if(n(o)!==null)y=!0,j();else{var J=n(s);J!==null&&W(k,J.startTime-K)}}var A=!1,w=-1,R=5,F=-1;function I(){return!(e.unstable_now()-F<R)}function U(){if(A){var K=e.unstable_now();F=K;var J=!0;try{e:{y=!1,v&&(v=!1,_(w),w=-1),m=!0;var te=h;try{t:{for($(K),d=n(o);d!==null&&!(d.expirationTime>K&&I());){var H=d.callback;if(typeof H=="function"){d.callback=null,h=d.priorityLevel;var M=H(d.expirationTime<=K);if(K=e.unstable_now(),typeof M=="function"){d.callback=M,$(K),J=!0;break t}d===n(o)&&r(o),$(K)}else r(o);d=n(o)}if(d!==null)J=!0;else{var G=n(s);G!==null&&W(k,G.startTime-K),J=!1}}break e}finally{d=null,h=te,m=!1}J=void 0}}finally{J?q():A=!1}}}var q;if(typeof E=="function")q=function(){E(U)};else if(typeof MessageChannel<"u"){var V=new MessageChannel,N=V.port2;V.port1.onmessage=U,q=function(){N.postMessage(null)}}else q=function(){S(U,0)};function j(){A||(A=!0,q())}function W(K,J){w=S(function(){K(e.unstable_now())},J)}e.unstable_IdlePriority=5,e.unstable_ImmediatePriority=1,e.unstable_LowPriority=4,e.unstable_NormalPriority=3,e.unstable_Profiling=null,e.unstable_UserBlockingPriority=2,e.unstable_cancelCallback=function(K){K.callback=null},e.unstable_continueExecution=function(){y||m||(y=!0,j())},e.unstable_forceFrameRate=function(K){0>K||125<K?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):R=0<K?Math.floor(1e3/K):5},e.unstable_getCurrentPriorityLevel=function(){return h},e.unstable_getFirstCallbackNode=function(){return n(o)},e.unstable_next=function(K){switch(h){case 1:case 2:case 3:var J=3;break;default:J=h}var te=h;h=J;try{return K()}finally{h=te}},e.unstable_pauseExecution=function(){},e.unstable_requestPaint=function(){},e.unstable_runWithPriority=function(K,J){switch(K){case 1:case 2:case 3:case 4:case 5:break;default:K=3}var te=h;h=K;try{return J()}finally{h=te}},e.unstable_scheduleCallback=function(K,J,te){var H=e.unstable_now();switch(typeof te=="object"&&te!==null?(te=te.delay,te=typeof te=="number"&&0<te?H+te:H):te=H,K){case 1:var M=-1;break;case 2:M=250;break;case 5:M=1073741823;break;case 4:M=1e4;break;default:M=5e3}return M=te+M,K={id:c++,callback:J,priorityLevel:K,startTime:te,expirationTime:M,sortIndex:-1},te>H?(K.sortIndex=te,t(s,K),n(o)===null&&K===n(s)&&(v?(_(w),w=-1):v=!0,W(k,te-H))):(K.sortIndex=M,t(o,K),y||m||(y=!0,j())),K},e.unstable_shouldYield=I,e.unstable_wrapCallback=function(K){var J=h;return function(){var te=h;h=J;try{return K.apply(this,arguments)}finally{h=te}}}})(q7)),q7}var UW;function k1e(){return UW||(UW=1,z7.exports=_tt()),z7.exports}/**
 * @license React
 * react-reconciler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var jW;function Ett(){return jW||(jW=1,(function(e){e.exports=function(t){function n(g,b,T,z){return new Iu(g,b,T,z)}function r(){}function i(g){var b="https://react.dev/errors/"+g;if(1<arguments.length){b+="?args[]="+encodeURIComponent(arguments[1]);for(var T=2;T<arguments.length;T++)b+="&args[]="+encodeURIComponent(arguments[T])}return"Minified React error #"+g+"; visit "+b+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function a(g){return g===null||typeof g!="object"?null:(g=im&&g[im]||g["@@iterator"],typeof g=="function"?g:null)}function u(g){if(g==null)return null;if(typeof g=="function")return g.$$typeof===zs?null:g.displayName||g.name||null;if(typeof g=="string")return g;switch(g){case Wo:return"Fragment";case Qs:return"Portal";case Ay:return"Profiler";case Ry:return"StrictMode";case $g:return"Suspense";case bd:return"SuspenseList"}if(typeof g=="object")switch(g.$$typeof){case kc:return(g.displayName||"Context")+".Provider";case yd:return(g._context.displayName||"Context")+".Consumer";case Mp:var b=g.render;return g=g.displayName,g||(g=b.displayName||b.name||"",g=g!==""?"ForwardRef("+g+")":"ForwardRef"),g;case Fp:return b=g.displayName||null,b!==null?b:u(g.type)||"Memo";case es:b=g._payload,g=g._init;try{return u(g(b))}catch{}}return null}function l(g){if(Cg===void 0)try{throw Error()}catch(T){var b=T.stack.trim().match(/\n( *(at )?)/);Cg=b&&b[1]||"",wg=-1<T.stack.indexOf(`
    at`)?" (<anonymous>)":-1<T.stack.indexOf("@")?"@unknown:0:0":""}return`
`+Cg+g+wg}function o(g,b){if(!g||Od)return"";Od=!0;var T=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var z={DetermineComponentFrameRoot:function(){try{if(b){var jt=function(){throw Error()};if(Object.defineProperty(jt.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(jt,[])}catch(Hn){var Wt=Hn}Reflect.construct(g,[],jt)}else{try{jt.call()}catch(Hn){Wt=Hn}g.call(jt.prototype)}}else{try{throw Error()}catch(Hn){Wt=Hn}(jt=g())&&typeof jt.catch=="function"&&jt.catch(function(){})}}catch(Hn){if(Hn&&Wt&&typeof Hn.stack=="string")return[Hn.stack,Wt.stack]}return[null,null]}};z.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var ee=Object.getOwnPropertyDescriptor(z.DetermineComponentFrameRoot,"name");ee&&ee.configurable&&Object.defineProperty(z.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var le=z.DetermineComponentFrameRoot(),De=le[0],Ie=le[1];if(De&&Ie){var We=De.split(`
`),vt=Ie.split(`
`);for(ee=z=0;z<We.length&&!We[z].includes("DetermineComponentFrameRoot");)z++;for(;ee<vt.length&&!vt[ee].includes("DetermineComponentFrameRoot");)ee++;if(z===We.length||ee===vt.length)for(z=We.length-1,ee=vt.length-1;1<=z&&0<=ee&&We[z]!==vt[ee];)ee--;for(;1<=z&&0<=ee;z--,ee--)if(We[z]!==vt[ee]){if(z!==1||ee!==1)do if(z--,ee--,0>ee||We[z]!==vt[ee]){var Pt=`
`+We[z].replace(" at new "," at ");return g.displayName&&Pt.includes("<anonymous>")&&(Pt=Pt.replace("<anonymous>",g.displayName)),Pt}while(1<=z&&0<=ee);break}}}finally{Od=!1,Error.prepareStackTrace=T}return(T=g?g.displayName||g.name:"")?l(T):""}function s(g){switch(g.tag){case 26:case 27:case 5:return l(g.type);case 16:return l("Lazy");case 13:return l("Suspense");case 19:return l("SuspenseList");case 0:case 15:return g=o(g.type,!1),g;case 11:return g=o(g.type.render,!1),g;case 1:return g=o(g.type,!0),g;default:return""}}function c(g){try{var b="";do b+=s(g),g=g.return;while(g);return b}catch(T){return`
Error generating stack: `+T.message+`
`+T.stack}}function d(g){var b=g,T=g;if(g.alternate)for(;b.return;)b=b.return;else{g=b;do b=g,(b.flags&4098)!==0&&(T=b.return),g=b.return;while(g)}return b.tag===3?T:null}function h(g){if(d(g)!==g)throw Error(i(188))}function m(g){var b=g.alternate;if(!b){if(b=d(g),b===null)throw Error(i(188));return b!==g?null:g}for(var T=g,z=b;;){var ee=T.return;if(ee===null)break;var le=ee.alternate;if(le===null){if(z=ee.return,z!==null){T=z;continue}break}if(ee.child===le.child){for(le=ee.child;le;){if(le===T)return h(ee),g;if(le===z)return h(ee),b;le=le.sibling}throw Error(i(188))}if(T.return!==z.return)T=ee,z=le;else{for(var De=!1,Ie=ee.child;Ie;){if(Ie===T){De=!0,T=ee,z=le;break}if(Ie===z){De=!0,z=ee,T=le;break}Ie=Ie.sibling}if(!De){for(Ie=le.child;Ie;){if(Ie===T){De=!0,T=le,z=ee;break}if(Ie===z){De=!0,z=le,T=ee;break}Ie=Ie.sibling}if(!De)throw Error(i(189))}}if(T.alternate!==z)throw Error(i(190))}if(T.tag!==3)throw Error(i(188));return T.stateNode.current===T?g:b}function y(g){var b=g.tag;if(b===5||b===26||b===27||b===6)return g;for(g=g.child;g!==null;){if(b=y(g),b!==null)return b;g=g.sibling}return null}function v(g){var b=g.tag;if(b===5||b===26||b===27||b===6)return g;for(g=g.child;g!==null;){if(g.tag!==4&&(b=v(g),b!==null))return b;g=g.sibling}return null}function S(g){return{current:g}}function _(g){0>Ic||(g.current=wd[Ic],wd[Ic]=null,Ic--)}function E(g,b){Ic++,wd[Ic]=g.current,g.current=b}function $(g){return g>>>=0,g===0?32:31-(Rd(g)/Z1|0)|0}function k(g){var b=g&42;if(b!==0)return b;switch(g&-g){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return g&4194176;case 4194304:case 8388608:case 16777216:case 33554432:return g&62914560;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return g}}function A(g,b){var T=g.pendingLanes;if(T===0)return 0;var z=0,ee=g.suspendedLanes,le=g.pingedLanes,De=g.warmLanes;g=g.finishedLanes!==0;var Ie=T&134217727;return Ie!==0?(T=Ie&~ee,T!==0?z=k(T):(le&=Ie,le!==0?z=k(le):g||(De=Ie&~De,De!==0&&(z=k(De))))):(Ie=T&~ee,Ie!==0?z=k(Ie):le!==0?z=k(le):g||(De=T&~De,De!==0&&(z=k(De)))),z===0?0:b!==0&&b!==z&&(b&ee)===0&&(ee=z&-z,De=b&-b,ee>=De||ee===32&&(De&4194176)!==0)?b:z}function w(g,b){return(g.pendingLanes&~(g.suspendedLanes&~g.pingedLanes)&b)===0}function R(g,b){switch(g){case 1:case 2:case 4:case 8:return b+250;case 16:case 32:case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return b+5e3;case 4194304:case 8388608:case 16777216:case 33554432:return-1;case 67108864:case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function F(){var g=Rh;return Rh<<=1,(Rh&4194176)===0&&(Rh=128),g}function I(){var g=Ad;return Ad<<=1,(Ad&62914560)===0&&(Ad=4194304),g}function U(g){for(var b=[],T=0;31>T;T++)b.push(g);return b}function q(g,b){g.pendingLanes|=b,b!==268435456&&(g.suspendedLanes=0,g.pingedLanes=0,g.warmLanes=0)}function V(g,b,T,z,ee,le){var De=g.pendingLanes;g.pendingLanes=T,g.suspendedLanes=0,g.pingedLanes=0,g.warmLanes=0,g.expiredLanes&=T,g.entangledLanes&=T,g.errorRecoveryDisabledLanes&=T,g.shellSuspendCounter=0;var Ie=g.entanglements,We=g.expirationTimes,vt=g.hiddenUpdates;for(T=De&~T;0<T;){var Pt=31-ll(T),jt=1<<Pt;Ie[Pt]=0,We[Pt]=-1;var Wt=vt[Pt];if(Wt!==null)for(vt[Pt]=null,Pt=0;Pt<Wt.length;Pt++){var Hn=Wt[Pt];Hn!==null&&(Hn.lane&=-536870913)}T&=~jt}z!==0&&N(g,z,0),le!==0&&ee===0&&g.tag!==0&&(g.suspendedLanes|=le&~(De&~b))}function N(g,b,T){g.pendingLanes|=b,g.suspendedLanes&=~b;var z=31-ll(b);g.entangledLanes|=b,g.entanglements[z]=g.entanglements[z]|1073741824|T&4194218}function j(g,b){var T=g.entangledLanes|=b;for(g=g.entanglements;T;){var z=31-ll(T),ee=1<<z;ee&b|g[z]&b&&(g[z]|=b),T&=~ee}}function W(g){return g&=-g,2<g?8<g?(g&134217727)!==0?32:268435456:8:2}function K(g){if(ol&&typeof ol.onCommitFiberRoot=="function")try{ol.onCommitFiberRoot(Gl,g,void 0,(g.current.flags&128)===128)}catch{}}function J(g){if(typeof X1=="function"&&Wy(g),ol&&typeof ol.setStrictMode=="function")try{ol.setStrictMode(Gl,g)}catch{}}function te(g,b){return g===b&&(g!==0||1/g===1/b)||g!==g&&b!==b}function H(g,b){if(typeof g=="object"&&g!==null){var T=qp.get(g);return T!==void 0?T:(b={value:g,source:b,stack:c(b)},qp.set(g,b),b)}return{value:g,source:b,stack:c(b)}}function M(g,b){xi[Ai++]=Td,xi[Ai++]=On,On=g,Td=b}function G(g,b,T){$l[iu++]=us,$l[iu++]=_o,$l[iu++]=js,js=g;var z=us;g=_o;var ee=32-ll(z)-1;z&=~(1<<ee),T+=1;var le=32-ll(b)+ee;if(30<le){var De=ee-ee%5;le=(z&(1<<De)-1).toString(32),z>>=De,ee-=De,us=1<<32-ll(b)+ee|T<<ee|z,_o=le+g}else us=1<<le|T<<ee|z,_o=g}function X(g){g.return!==null&&(M(g,1),G(g,1,0))}function B(g){for(;g===On;)On=xi[--Ai],xi[Ai]=null,Td=xi[--Ai],xi[Ai]=null;for(;g===js;)js=$l[--iu],$l[iu]=null,_o=$l[--iu],$l[iu]=null,us=$l[--iu],$l[iu]=null}function ne(g,b){E(Zs,b),E(mm,g),E(qr,null),g=xd(b),_(qr),E(qr,g)}function oe(){_(qr),_(mm),_(Zs)}function ie(g){g.memoizedState!==null&&E(kh,g);var b=qr.current,T=Ty(b,g.type);b!==T&&(E(mm,g),E(qr,T))}function de(g){mm.current===g&&(_(qr),_(mm)),kh.current===g&&(_(kh),tu?An._currentValue=Su:An._currentValue2=Su)}function ge(g){var b=Error(i(418,""));throw pe(H(b,g)),Do}function ye(g,b){if(!Ra)throw Error(i(175));Vy(g.stateNode,g.type,g.memoizedProps,b,g)||ge(g)}function me(g){for(sl=g.return;sl;)switch(sl.tag){case 3:case 27:ls=!0;return;case 5:case 13:ls=!1;return;default:sl=sl.return}}function Ee(g){if(!Ra||g!==sl)return!1;if(!Vr)return me(g),Vr=!0,!1;var b=!1;if(_u?g.tag!==3&&g.tag!==27&&(g.tag!==5||V1(g.type)&&!Sd(g.type,g.memoizedProps))&&(b=!0):g.tag!==3&&(g.tag!==5||V1(g.type)&&!Sd(g.type,g.memoizedProps))&&(b=!0),b&&Eu&&ge(g),me(g),g.tag===13){if(!Ra)throw Error(i(316));if(g=g.memoizedState,g=g!==null?g.dehydrated:null,!g)throw Error(i(317));Eu=q1(g)}else Eu=sl?Nc(g.stateNode):null;return!0}function se(){Ra&&(Eu=sl=null,Vr=!1)}function pe(g){Eo===null?Eo=[g]:Eo.push(g)}function ve(){for(var g=Bc,b=Vp=Bc=0;b<g;){var T=Ti[b];Ti[b++]=null;var z=Ti[b];Ti[b++]=null;var ee=Ti[b];Ti[b++]=null;var le=Ti[b];if(Ti[b++]=null,z!==null&&ee!==null){var De=z.pending;De===null?ee.next=ee:(ee.next=De.next,De.next=ee),z.pending=ee}le!==0&&et(T,ee,le)}}function we(g,b,T,z){Ti[Bc++]=g,Ti[Bc++]=b,Ti[Bc++]=T,Ti[Bc++]=z,Vp|=z,g.lanes|=z,g=g.alternate,g!==null&&(g.lanes|=z)}function Me(g,b,T,z){return we(g,b,T,z),tt(g)}function Re(g,b){return we(g,null,null,b),tt(g)}function et(g,b,T){g.lanes|=T;var z=g.alternate;z!==null&&(z.lanes|=T);for(var ee=!1,le=g.return;le!==null;)le.childLanes|=T,z=le.alternate,z!==null&&(z.childLanes|=T),le.tag===22&&(g=le.stateNode,g===null||g._visibility&1||(ee=!0)),g=le,le=le.return;ee&&b!==null&&g.tag===3&&(le=g.stateNode,ee=31-ll(T),le=le.hiddenUpdates,g=le[ee],g===null?le[ee]=[b]:g.push(b),b.lane=T|536870912)}function tt(g){if(50<Qh)throw Qh=0,wl=null,Error(i(185));for(var b=g.return;b!==null;)g=b,b=g.return;return g.tag===3?g.stateNode:null}function Oe(g){g!==Ph&&g.next===null&&(Ph===null?kd=Ph=g:Ph=Ph.next=g),$f=!0,gm||(gm=!0,Ne(Ae))}function $e(g,b){if(!os&&$f){os=!0;do for(var T=!1,z=kd;z!==null;){if(g!==0){var ee=z.pendingLanes;if(ee===0)var le=0;else{var De=z.suspendedLanes,Ie=z.pingedLanes;le=(1<<31-ll(42|g)+1)-1,le&=ee&~(De&~Ie),le=le&201326677?le&201326677|1:le?le|2:0}le!==0&&(T=!0,Ye(z,le))}else le=cr,le=A(z,z===mr?le:0),(le&3)===0||w(z,le)||(T=!0,Ye(z,le));z=z.next}while(T);os=!1}}function Ae(){$f=gm=!1;var g=0;Cf!==0&&(Ed()&&(g=Cf),Cf=0);for(var b=Hi(),T=null,z=kd;z!==null;){var ee=z.next,le=He(z,b);le===0?(z.next=null,T===null?kd=ee:T.next=ee,ee===null&&(Ph=T)):(T=z,(g!==0||(le&3)!==0)&&($f=!0)),z=ee}$e(g)}function He(g,b){for(var T=g.suspendedLanes,z=g.pingedLanes,ee=g.expirationTimes,le=g.pendingLanes&-62914561;0<le;){var De=31-ll(le),Ie=1<<De,We=ee[De];We===-1?((Ie&T)===0||(Ie&z)!==0)&&(ee[De]=R(Ie,b)):We<=b&&(g.expiredLanes|=Ie),le&=~Ie}if(b=mr,T=cr,T=A(g,g===b?T:0),z=g.callbackNode,T===0||g===b&&Si===2||g.cancelPendingCommit!==null)return z!==null&&z!==null&&Th(z),g.callbackNode=null,g.callbackPriority=0;if((T&3)===0||w(g,T)){if(b=T&-T,b===g.callbackPriority)return b;switch(z!==null&&Th(z),W(T)){case 2:case 8:T=Ky;break;case 32:T=Xg;break;case 268435456:T=pm;break;default:T=Xg}return z=je.bind(null,g),T=Ah(T,z),g.callbackPriority=b,g.callbackNode=T,b}return z!==null&&z!==null&&Th(z),g.callbackPriority=2,g.callbackNode=null,2}function je(g,b){var T=g.callbackNode;if(Kt()&&g.callbackNode!==T)return null;var z=cr;return z=A(g,g===mr?z:0),z===0?null:(Oh(g,z,b),He(g,Hi()),g.callbackNode!=null&&g.callbackNode===T?je.bind(null,g):null)}function Ye(g,b){if(Kt())return null;Oh(g,b,!0)}function Ne(g){ga?lm(function(){(vn&6)!==0?Ah(Hg,g):g()}):Ah(Hg,g)}function ze(){return Cf===0&&(Cf=F()),Cf}function at(g,b){if(vm===null){var T=vm=[];Yg=0,Kl=ze(),Pd={status:"pending",value:void 0,then:function(z){T.push(z)}}}return Yg++,b.then(mt,mt),b}function mt(){if(--Yg===0&&vm!==null){Pd!==null&&(Pd.status="fulfilled");var g=vm;vm=null,Kl=0,Pd=null;for(var b=0;b<g.length;b++)(0,g[b])()}}function Mt(g,b){var T=[],z={status:"pending",value:null,reason:null,then:function(ee){T.push(ee)}};return g.then(function(){z.status="fulfilled",z.value=b;for(var ee=0;ee<T.length;ee++)(0,T[ee])(b)},function(ee){for(z.status="rejected",z.reason=ee,ee=0;ee<T.length;ee++)(0,T[ee])(void 0)}),z}function bt(g){g.updateQueue={baseState:g.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function zt(g,b){g=g.updateQueue,b.updateQueue===g&&(b.updateQueue={baseState:g.baseState,firstBaseUpdate:g.firstBaseUpdate,lastBaseUpdate:g.lastBaseUpdate,shared:g.shared,callbacks:null})}function qt(g){return{lane:g,tag:0,payload:null,callback:null,next:null}}function Xt(g,b,T){var z=g.updateQueue;if(z===null)return null;if(z=z.shared,(vn&2)!==0){var ee=z.pending;return ee===null?b.next=b:(b.next=ee.next,ee.next=b),z.pending=b,b=tt(g),et(g,null,T),b}return we(g,z,b,T),tt(g)}function nn(g,b,T){if(b=b.updateQueue,b!==null&&(b=b.shared,(T&4194176)!==0)){var z=b.lanes;z&=g.pendingLanes,T|=z,b.lanes=T,j(g,T)}}function Fe(g,b){var T=g.updateQueue,z=g.alternate;if(z!==null&&(z=z.updateQueue,T===z)){var ee=null,le=null;if(T=T.firstBaseUpdate,T!==null){do{var De={lane:T.lane,tag:T.tag,payload:T.payload,callback:null,next:null};le===null?ee=le=De:le=le.next=De,T=T.next}while(T!==null);le===null?ee=le=b:le=le.next=b}else ee=le=b;T={baseState:z.baseState,firstBaseUpdate:ee,lastBaseUpdate:le,shared:z.shared,callbacks:z.callbacks},g.updateQueue=T;return}g=T.lastBaseUpdate,g===null?T.firstBaseUpdate=b:g.next=b,T.lastBaseUpdate=b}function ht(){if(Mh){var g=Pd;if(g!==null)throw g}}function kt(g,b,T,z){Mh=!1;var ee=g.updateQueue;Qc=!1;var le=ee.firstBaseUpdate,De=ee.lastBaseUpdate,Ie=ee.shared.pending;if(Ie!==null){ee.shared.pending=null;var We=Ie,vt=We.next;We.next=null,De===null?le=vt:De.next=vt,De=We;var Pt=g.alternate;Pt!==null&&(Pt=Pt.updateQueue,Ie=Pt.lastBaseUpdate,Ie!==De&&(Ie===null?Pt.firstBaseUpdate=vt:Ie.next=vt,Pt.lastBaseUpdate=We))}if(le!==null){var jt=ee.baseState;De=0,Pt=vt=We=null,Ie=le;do{var Wt=Ie.lane&-536870913,Hn=Wt!==Ie.lane;if(Hn?(cr&Wt)===Wt:(z&Wt)===Wt){Wt!==0&&Wt===Kl&&(Mh=!0),Pt!==null&&(Pt=Pt.next={lane:0,tag:Ie.tag,payload:Ie.payload,callback:null,next:null});e:{var to=g,_m=Ie;Wt=b;var zh=T;switch(_m.tag){case 1:if(to=_m.payload,typeof to=="function"){jt=to.call(zh,jt,Wt);break e}jt=to;break e;case 3:to.flags=to.flags&-65537|128;case 0:if(to=_m.payload,Wt=typeof to=="function"?to.call(zh,jt,Wt):to,Wt==null)break e;jt=_h({},jt,Wt);break e;case 2:Qc=!0}}Wt=Ie.callback,Wt!==null&&(g.flags|=64,Hn&&(g.flags|=8192),Hn=ee.callbacks,Hn===null?ee.callbacks=[Wt]:Hn.push(Wt))}else Hn={lane:Wt,tag:Ie.tag,payload:Ie.payload,callback:Ie.callback,next:null},Pt===null?(vt=Pt=Hn,We=jt):Pt=Pt.next=Hn,De|=Wt;if(Ie=Ie.next,Ie===null){if(Ie=ee.shared.pending,Ie===null)break;Hn=Ie,Ie=Hn.next,Hn.next=null,ee.lastBaseUpdate=Hn,ee.shared.pending=null}}while(!0);Pt===null&&(We=jt),ee.baseState=We,ee.firstBaseUpdate=vt,ee.lastBaseUpdate=Pt,le===null&&(ee.shared.lanes=0),Ao|=De,g.lanes=De,g.memoizedState=jt}}function Qt(g,b){if(typeof g!="function")throw Error(i(191,g));g.call(b)}function en(g,b){var T=g.callbacks;if(T!==null)for(g.callbacks=null,g=0;g<T.length;g++)Qt(T[g],b)}function In(g,b){if(Dl(g,b))return!0;if(typeof g!="object"||g===null||typeof b!="object"||b===null)return!1;var T=Object.keys(g),z=Object.keys(b);if(T.length!==z.length)return!1;for(z=0;z<T.length;z++){var ee=T[z];if(!ym.call(b,ee)||!Dl(g[ee],b[ee]))return!1}return!0}function rn(g){return g=g.status,g==="fulfilled"||g==="rejected"}function Xn(){}function br(g,b,T){switch(T=g[T],T===void 0?g.push(b):T!==b&&(b.then(Xn,Xn),b=T),b.status){case"fulfilled":return b.value;case"rejected":throw g=b.reason,g===Wr?Error(i(483)):g;default:if(typeof b.status=="string")b.then(Xn,Xn);else{if(g=mr,g!==null&&100<g.shellSuspendCounter)throw Error(i(482));g=b,g.status="pending",g.then(function(z){if(b.status==="pending"){var ee=b;ee.status="fulfilled",ee.value=z}},function(z){if(b.status==="pending"){var ee=b;ee.status="rejected",ee.reason=z}})}switch(b.status){case"fulfilled":return b.value;case"rejected":throw g=b.reason,g===Wr?Error(i(483)):g}throw wf=b,Wr}}function Or(){if(wf===null)throw Error(i(459));var g=wf;return wf=null,g}function Cr(g){var b=$o;return $o+=1,Hs===null&&(Hs=[]),br(Hs,g,b)}function wr(g,b){b=b.props.ref,g.ref=b!==void 0?b:null}function Gr(g,b){throw b.$$typeof===wy?Error(i(525)):(g=Object.prototype.toString.call(b),Error(i(31,g==="[object Object]"?"object with keys {"+Object.keys(b).join(", ")+"}":g)))}function ji(g){var b=g._init;return b(g._payload)}function ai(g){function b(lt,Ge){if(g){var ct=lt.deletions;ct===null?(lt.deletions=[Ge],lt.flags|=16):ct.push(Ge)}}function T(lt,Ge){if(!g)return null;for(;Ge!==null;)b(lt,Ge),Ge=Ge.sibling;return null}function z(lt){for(var Ge=new Map;lt!==null;)lt.key!==null?Ge.set(lt.key,lt):Ge.set(lt.index,lt),lt=lt.sibling;return Ge}function ee(lt,Ge){return lt=Ja(lt,Ge),lt.index=0,lt.sibling=null,lt}function le(lt,Ge,ct){return lt.index=ct,g?(ct=lt.alternate,ct!==null?(ct=ct.index,ct<Ge?(lt.flags|=33554434,Ge):ct):(lt.flags|=33554434,Ge)):(lt.flags|=1048576,Ge)}function De(lt){return g&&lt.alternate===null&&(lt.flags|=33554434),lt}function Ie(lt,Ge,ct,Tt){return Ge===null||Ge.tag!==6?(Ge=jl(ct,lt.mode,Tt),Ge.return=lt,Ge):(Ge=ee(Ge,ct),Ge.return=lt,Ge)}function We(lt,Ge,ct,Tt){var pn=ct.type;return pn===Wo?Pt(lt,Ge,ct.props.children,Tt,ct.key):Ge!==null&&(Ge.elementType===pn||typeof pn=="object"&&pn!==null&&pn.$$typeof===es&&ji(pn)===Ge.type)?(Ge=ee(Ge,ct.props),wr(Ge,ct),Ge.return=lt,Ge):(Ge=Ul(ct.type,ct.key,ct.props,null,lt.mode,Tt),wr(Ge,ct),Ge.return=lt,Ge)}function vt(lt,Ge,ct,Tt){return Ge===null||Ge.tag!==4||Ge.stateNode.containerInfo!==ct.containerInfo||Ge.stateNode.implementation!==ct.implementation?(Ge=vd(ct,lt.mode,Tt),Ge.return=lt,Ge):(Ge=ee(Ge,ct.children||[]),Ge.return=lt,Ge)}function Pt(lt,Ge,ct,Tt,pn){return Ge===null||Ge.tag!==7?(Ge=Lu(ct,lt.mode,Tt,pn),Ge.return=lt,Ge):(Ge=ee(Ge,ct),Ge.return=lt,Ge)}function jt(lt,Ge,ct){if(typeof Ge=="string"&&Ge!==""||typeof Ge=="number"||typeof Ge=="bigint")return Ge=jl(""+Ge,lt.mode,ct),Ge.return=lt,Ge;if(typeof Ge=="object"&&Ge!==null){switch(Ge.$$typeof){case Df:return ct=Ul(Ge.type,Ge.key,Ge.props,null,lt.mode,ct),wr(ct,Ge),ct.return=lt,ct;case Qs:return Ge=vd(Ge,lt.mode,ct),Ge.return=lt,Ge;case es:var Tt=Ge._init;return Ge=Tt(Ge._payload),jt(lt,Ge,ct)}if(nl(Ge)||a(Ge))return Ge=Lu(Ge,lt.mode,ct,null),Ge.return=lt,Ge;if(typeof Ge.then=="function")return jt(lt,Cr(Ge),ct);if(Ge.$$typeof===kc)return jt(lt,gh(lt,Ge),ct);Gr(lt,Ge)}return null}function Wt(lt,Ge,ct,Tt){var pn=Ge!==null?Ge.key:null;if(typeof ct=="string"&&ct!==""||typeof ct=="number"||typeof ct=="bigint")return pn!==null?null:Ie(lt,Ge,""+ct,Tt);if(typeof ct=="object"&&ct!==null){switch(ct.$$typeof){case Df:return ct.key===pn?We(lt,Ge,ct,Tt):null;case Qs:return ct.key===pn?vt(lt,Ge,ct,Tt):null;case es:return pn=ct._init,ct=pn(ct._payload),Wt(lt,Ge,ct,Tt)}if(nl(ct)||a(ct))return pn!==null?null:Pt(lt,Ge,ct,Tt,null);if(typeof ct.then=="function")return Wt(lt,Ge,Cr(ct),Tt);if(ct.$$typeof===kc)return Wt(lt,Ge,gh(lt,ct),Tt);Gr(lt,ct)}return null}function Hn(lt,Ge,ct,Tt,pn){if(typeof Tt=="string"&&Tt!==""||typeof Tt=="number"||typeof Tt=="bigint")return lt=lt.get(ct)||null,Ie(Ge,lt,""+Tt,pn);if(typeof Tt=="object"&&Tt!==null){switch(Tt.$$typeof){case Df:return lt=lt.get(Tt.key===null?ct:Tt.key)||null,We(Ge,lt,Tt,pn);case Qs:return lt=lt.get(Tt.key===null?ct:Tt.key)||null,vt(Ge,lt,Tt,pn);case es:var pi=Tt._init;return Tt=pi(Tt._payload),Hn(lt,Ge,ct,Tt,pn)}if(nl(Tt)||a(Tt))return lt=lt.get(ct)||null,Pt(Ge,lt,Tt,pn,null);if(typeof Tt.then=="function")return Hn(lt,Ge,ct,Cr(Tt),pn);if(Tt.$$typeof===kc)return Hn(lt,Ge,ct,gh(Ge,Tt),pn);Gr(Ge,Tt)}return null}function to(lt,Ge,ct,Tt){for(var pn=null,pi=null,Tn=Ge,Lr=Ge=0,Ua=null;Tn!==null&&Lr<ct.length;Lr++){Tn.index>Lr?(Ua=Tn,Tn=null):Ua=Tn.sibling;var Tr=Wt(lt,Tn,ct[Lr],Tt);if(Tr===null){Tn===null&&(Tn=Ua);break}g&&Tn&&Tr.alternate===null&&b(lt,Tn),Ge=le(Tr,Ge,Lr),pi===null?pn=Tr:pi.sibling=Tr,pi=Tr,Tn=Ua}if(Lr===ct.length)return T(lt,Tn),Vr&&M(lt,Lr),pn;if(Tn===null){for(;Lr<ct.length;Lr++)Tn=jt(lt,ct[Lr],Tt),Tn!==null&&(Ge=le(Tn,Ge,Lr),pi===null?pn=Tn:pi.sibling=Tn,pi=Tn);return Vr&&M(lt,Lr),pn}for(Tn=z(Tn);Lr<ct.length;Lr++)Ua=Hn(Tn,lt,Lr,ct[Lr],Tt),Ua!==null&&(g&&Ua.alternate!==null&&Tn.delete(Ua.key===null?Lr:Ua.key),Ge=le(Ua,Ge,Lr),pi===null?pn=Ua:pi.sibling=Ua,pi=Ua);return g&&Tn.forEach(function(jc){return b(lt,jc)}),Vr&&M(lt,Lr),pn}function _m(lt,Ge,ct,Tt){if(ct==null)throw Error(i(151));for(var pn=null,pi=null,Tn=Ge,Lr=Ge=0,Ua=null,Tr=ct.next();Tn!==null&&!Tr.done;Lr++,Tr=ct.next()){Tn.index>Lr?(Ua=Tn,Tn=null):Ua=Tn.sibling;var jc=Wt(lt,Tn,Tr.value,Tt);if(jc===null){Tn===null&&(Tn=Ua);break}g&&Tn&&jc.alternate===null&&b(lt,Tn),Ge=le(jc,Ge,Lr),pi===null?pn=jc:pi.sibling=jc,pi=jc,Tn=Ua}if(Tr.done)return T(lt,Tn),Vr&&M(lt,Lr),pn;if(Tn===null){for(;!Tr.done;Lr++,Tr=ct.next())Tr=jt(lt,Tr.value,Tt),Tr!==null&&(Ge=le(Tr,Ge,Lr),pi===null?pn=Tr:pi.sibling=Tr,pi=Tr);return Vr&&M(lt,Lr),pn}for(Tn=z(Tn);!Tr.done;Lr++,Tr=ct.next())Tr=Hn(Tn,lt,Lr,Tr.value,Tt),Tr!==null&&(g&&Tr.alternate!==null&&Tn.delete(Tr.key===null?Lr:Tr.key),Ge=le(Tr,Ge,Lr),pi===null?pn=Tr:pi.sibling=Tr,pi=Tr);return g&&Tn.forEach(function(Em){return b(lt,Em)}),Vr&&M(lt,Lr),pn}function zh(lt,Ge,ct,Tt){if(typeof ct=="object"&&ct!==null&&ct.type===Wo&&ct.key===null&&(ct=ct.props.children),typeof ct=="object"&&ct!==null){switch(ct.$$typeof){case Df:e:{for(var pn=ct.key;Ge!==null;){if(Ge.key===pn){if(pn=ct.type,pn===Wo){if(Ge.tag===7){T(lt,Ge.sibling),Tt=ee(Ge,ct.props.children),Tt.return=lt,lt=Tt;break e}}else if(Ge.elementType===pn||typeof pn=="object"&&pn!==null&&pn.$$typeof===es&&ji(pn)===Ge.type){T(lt,Ge.sibling),Tt=ee(Ge,ct.props),wr(Tt,ct),Tt.return=lt,lt=Tt;break e}T(lt,Ge);break}else b(lt,Ge);Ge=Ge.sibling}ct.type===Wo?(Tt=Lu(ct.props.children,lt.mode,Tt,ct.key),Tt.return=lt,lt=Tt):(Tt=Ul(ct.type,ct.key,ct.props,null,lt.mode,Tt),wr(Tt,ct),Tt.return=lt,lt=Tt)}return De(lt);case Qs:e:{for(pn=ct.key;Ge!==null;){if(Ge.key===pn)if(Ge.tag===4&&Ge.stateNode.containerInfo===ct.containerInfo&&Ge.stateNode.implementation===ct.implementation){T(lt,Ge.sibling),Tt=ee(Ge,ct.children||[]),Tt.return=lt,lt=Tt;break e}else{T(lt,Ge);break}else b(lt,Ge);Ge=Ge.sibling}Tt=vd(ct,lt.mode,Tt),Tt.return=lt,lt=Tt}return De(lt);case es:return pn=ct._init,ct=pn(ct._payload),zh(lt,Ge,ct,Tt)}if(nl(ct))return to(lt,Ge,ct,Tt);if(a(ct)){if(pn=a(ct),typeof pn!="function")throw Error(i(150));return ct=pn.call(ct),_m(lt,Ge,ct,Tt)}if(typeof ct.then=="function")return zh(lt,Ge,Cr(ct),Tt);if(ct.$$typeof===kc)return zh(lt,Ge,gh(lt,ct),Tt);Gr(lt,ct)}return typeof ct=="string"&&ct!==""||typeof ct=="number"||typeof ct=="bigint"?(ct=""+ct,Ge!==null&&Ge.tag===6?(T(lt,Ge.sibling),Tt=ee(Ge,ct),Tt.return=lt,lt=Tt):(T(lt,Ge),Tt=jl(ct,lt.mode,Tt),Tt.return=lt,lt=Tt),De(lt)):T(lt,Ge)}return function(lt,Ge,ct,Tt){try{$o=0;var pn=zh(lt,Ge,ct,Tt);return Hs=null,pn}catch(Tn){if(Tn===Wr)throw Tn;var pi=n(29,Tn,null,lt.mode);return pi.lanes=Tt,pi.return=lt,pi}finally{}}}function ar(g,b){g=Ys,E(uu,g),E(au,b),Ys=g|b.baseLanes}function Te(){E(uu,Ys),E(au,au.current)}function Ve(){Ys=uu.current,_(au),_(uu)}function nt(g){var b=g.alternate;E(qa,qa.current&1),E(Co,g),wo===null&&(b===null||au.current!==null||b.memoizedState!==null)&&(wo=g)}function gt(g){if(g.tag===22){if(E(qa,qa.current),E(Co,g),wo===null){var b=g.alternate;b!==null&&b.memoizedState!==null&&(wo=g)}}else Ct()}function Ct(){E(qa,qa.current),E(Co,Co.current)}function Vt(g){_(Co),wo===g&&(wo=null),_(qa)}function Yt(g){for(var b=g;b!==null;){if(b.tag===13){var T=b.memoizedState;if(T!==null&&(T=T.dehydrated,T===null||Lg(T)||Dd(T)))return b}else if(b.tag===19&&b.memoizedProps.revealOrder!==void 0){if((b.flags&128)!==0)return b}else if(b.child!==null){b.child.return=b,b=b.child;continue}if(b===g)break;for(;b.sibling===null;){if(b.return===null||b.return===g)return null;b=b.return}b.sibling.return=b.return,b=b.sibling}return null}function Dt(){throw Error(i(321))}function Lt(g,b){if(b===null)return!1;for(var T=0;T<b.length&&T<g.length;T++)if(!Dl(g[T],b[T]))return!1;return!0}function an(g,b,T,z,ee,le){return zc=le,nr=b,b.memoizedState=null,b.updateQueue=null,b.lanes=0,xn.H=g===null||g.memoizedState===null?Va:cl,Rf=!1,le=T(z,ee),Rf=!1,Xi&&(le=sn(b,T,z,ee)),_n(g),le}function _n(g){xn.H=Fr;var b=Jr!==null&&Jr.next!==null;if(zc=0,Ta=Jr=nr=null,ss=!1,va=0,Up=null,b)throw Error(i(300));g===null||ka||(g=g.dependencies,g!==null&&Xo(g)&&(ka=!0))}function sn(g,b,T,z){nr=g;var ee=0;do{if(Xi&&(Up=null),va=0,Xi=!1,25<=ee)throw Error(i(301));if(ee+=1,Ta=Jr=null,g.updateQueue!=null){var le=g.updateQueue;le.lastEffect=null,le.events=null,le.stores=null,le.memoCache!=null&&(le.memoCache.index=0)}xn.H=Fd,le=b(T,z)}while(Xi);return le}function Ln(){var g=xn.H,b=g.useState()[0];return b=typeof b.then=="function"?Kr(b):b,g=g.useState()[0],(Jr!==null?Jr.memoizedState:null)!==g&&(nr.flags|=1024),b}function er(){var g=bm!==0;return bm=0,g}function Di(g,b,T){b.updateQueue=g.updateQueue,b.flags&=-2053,g.lanes&=~T}function $i(g){if(ss){for(g=g.memoizedState;g!==null;){var b=g.queue;b!==null&&(b.pending=null),g=g.next}ss=!1}zc=0,Ta=Jr=nr=null,Xi=!1,va=bm=0,Up=null}function hr(){var g={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return Ta===null?nr.memoizedState=Ta=g:Ta=Ta.next=g,Ta}function wt(){if(Jr===null){var g=nr.alternate;g=g!==null?g.memoizedState:null}else g=Jr.next;var b=Ta===null?nr.memoizedState:Ta.next;if(b!==null)Ta=b,Jr=g;else{if(g===null)throw nr.alternate===null?Error(i(467)):Error(i(310));Jr=g,g={memoizedState:Jr.memoizedState,baseState:Jr.baseState,baseQueue:Jr.baseQueue,queue:Jr.queue,next:null},Ta===null?nr.memoizedState=Ta=g:Ta=Ta.next=g}return Ta}function Kr(g){var b=va;return va+=1,Up===null&&(Up=[]),g=br(Up,g,b),b=nr,(Ta===null?b.memoizedState:Ta.next)===null&&(b=b.alternate,xn.H=b===null||b.memoizedState===null?Va:cl),g}function Zi(g){if(g!==null&&typeof g=="object"){if(typeof g.then=="function")return Kr(g);if(g.$$typeof===kc)return Qa(g)}throw Error(i(438,String(g)))}function yi(g){var b=null,T=nr.updateQueue;if(T!==null&&(b=T.memoCache),b==null){var z=nr.alternate;z!==null&&(z=z.updateQueue,z!==null&&(z=z.memoCache,z!=null&&(b={data:z.data.map(function(ee){return ee.slice()}),index:0})))}if(b==null&&(b={data:[],index:0}),T===null&&(T=Om(),nr.updateQueue=T),T.memoCache=b,T=b.data[b.index],T===void 0)for(T=b.data[b.index]=Array(g),z=0;z<g;z++)T[z]=B1;return b.index++,T}function Ca(g,b){return typeof b=="function"?b(g):b}function Xa(g){var b=wt();return Au(b,Jr,g)}function Au(g,b,T){var z=g.queue;if(z===null)throw Error(i(311));z.lastRenderedReducer=T;var ee=g.baseQueue,le=z.pending;if(le!==null){if(ee!==null){var De=ee.next;ee.next=le.next,le.next=De}b.baseQueue=ee=le,z.pending=null}if(le=g.baseState,ee===null)g.memoizedState=le;else{b=ee.next;var Ie=De=null,We=null,vt=b,Pt=!1;do{var jt=vt.lane&-536870913;if(jt!==vt.lane?(cr&jt)===jt:(zc&jt)===jt){var Wt=vt.revertLane;if(Wt===0)We!==null&&(We=We.next={lane:0,revertLane:0,action:vt.action,hasEagerState:vt.hasEagerState,eagerState:vt.eagerState,next:null}),jt===Kl&&(Pt=!0);else if((zc&Wt)===Wt){vt=vt.next,Wt===Kl&&(Pt=!0);continue}else jt={lane:0,revertLane:vt.revertLane,action:vt.action,hasEagerState:vt.hasEagerState,eagerState:vt.eagerState,next:null},We===null?(Ie=We=jt,De=le):We=We.next=jt,nr.lanes|=Wt,Ao|=Wt;jt=vt.action,Rf&&T(le,jt),le=vt.hasEagerState?vt.eagerState:T(le,jt)}else Wt={lane:jt,revertLane:vt.revertLane,action:vt.action,hasEagerState:vt.hasEagerState,eagerState:vt.eagerState,next:null},We===null?(Ie=We=Wt,De=le):We=We.next=Wt,nr.lanes|=jt,Ao|=jt;vt=vt.next}while(vt!==null&&vt!==b);if(We===null?De=le:We.next=Ie,!Dl(le,g.memoizedState)&&(ka=!0,Pt&&(T=Pd,T!==null)))throw T;g.memoizedState=le,g.baseState=De,g.baseQueue=We,z.lastRenderedState=le}return ee===null&&(z.lanes=0),[g.memoizedState,z.dispatch]}function Tu(g){var b=wt(),T=b.queue;if(T===null)throw Error(i(311));T.lastRenderedReducer=g;var z=T.dispatch,ee=T.pending,le=b.memoizedState;if(ee!==null){T.pending=null;var De=ee=ee.next;do le=g(le,De.action),De=De.next;while(De!==ee);Dl(le,b.memoizedState)||(ka=!0),b.memoizedState=le,b.baseQueue===null&&(b.baseState=le),T.lastRenderedState=le}return[le,z]}function Nl(g,b,T){var z=nr,ee=wt(),le=Vr;if(le){if(T===void 0)throw Error(i(407));T=T()}else T=b();var De=!Dl((Jr||ee).memoizedState,T);if(De&&(ee.memoizedState=T,ka=!0),ee=ee.queue,vf(ur.bind(null,z,ee,g),[g]),ee.getSnapshot!==b||De||Ta!==null&&Ta.memoizedState.tag&1){if(z.flags|=2048,Vo(9,xl.bind(null,z,ee,T,b),{destroy:void 0},null),mr===null)throw Error(i(349));le||(zc&60)!==0||gc(z,b,T)}return T}function gc(g,b,T){g.flags|=16384,g={getSnapshot:b,value:T},b=nr.updateQueue,b===null?(b=Om(),nr.updateQueue=b,b.stores=[g]):(T=b.stores,T===null?b.stores=[g]:T.push(g))}function xl(g,b,T,z){b.value=T,b.getSnapshot=z,Rr(b)&&Sl(g)}function ur(g,b,T){return T(function(){Rr(b)&&Sl(g)})}function Rr(g){var b=g.getSnapshot;g=g.value;try{var T=b();return!Dl(g,T)}catch{return!0}}function Sl(g){var b=Re(g,2);b!==null&&yu(b,g,2)}function tr(g){var b=hr();if(typeof g=="function"){var T=g;if(g=T(),Rf){J(!0);try{T()}finally{J(!1)}}}return b.memoizedState=b.baseState=g,b.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Ca,lastRenderedState:g},b}function Bn(g,b,T,z){return g.baseState=T,Au(g,Jr,typeof z=="function"?z:Ca)}function ku(g,b,T,z,ee){if(xc(g))throw Error(i(485));if(g=b.action,g!==null){var le={payload:ee,action:g,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(De){le.listeners.push(De)}};xn.T!==null?T(!0):le.isTransition=!1,z(le),T=b.pending,T===null?(le.next=b.pending=le,gf(b,le)):(le.next=T.next,b.pending=T.next=le)}}function gf(g,b){var T=b.action,z=b.payload,ee=g.state;if(b.isTransition){var le=xn.T,De={};xn.T=De;try{var Ie=T(ee,z),We=xn.S;We!==null&&We(De,Ie),Ds(g,b,Ie)}catch(vt){Il(g,b,vt)}finally{xn.T=le}}else try{le=T(ee,z),Ds(g,b,le)}catch(vt){Il(g,b,vt)}}function Ds(g,b,T){T!==null&&typeof T=="object"&&typeof T.then=="function"?T.then(function(z){qo(g,b,z)},function(z){return Il(g,b,z)}):qo(g,b,T)}function qo(g,b,T){b.status="fulfilled",b.value=T,sh(b),g.state=T,b=g.pending,b!==null&&(T=b.next,T===b?g.pending=null:(T=T.next,b.next=T,gf(g,T)))}function Il(g,b,T){var z=g.pending;if(g.pending=null,z!==null){z=z.next;do b.status="rejected",b.reason=T,sh(b),b=b.next;while(b!==z)}g.action=null}function sh(g){g=g.listeners;for(var b=0;b<g.length;b++)(0,g[b])()}function Ya(g,b){return b}function rd(g,b){if(Vr){var T=mr.formState;if(T!==null){e:{var z=nr;if(Vr){if(Eu){var ee=By(Eu,ls);if(ee){Eu=Nc(ee),z=Xl(ee);break e}}ge(z)}z=!1}z&&(b=T[0])}}T=hr(),T.memoizedState=T.baseState=b,z={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Ya,lastRenderedState:b},T.queue=z,T=hh.bind(null,nr,z),z.dispatch=T,z=tr(!1);var le=Qn.bind(null,nr,!1,z.queue);return z=hr(),ee={state:b,dispatch:null,action:g,pending:null},z.queue=ee,T=ku.bind(null,nr,ee,le,T),ee.dispatch=T,z.memoizedState=g,[b,T,!1]}function $s(g){var b=wt();return id(b,Jr,g)}function id(g,b,T){b=Au(g,b,Ya)[0],g=Xa(Ca)[0],b=typeof b=="object"&&b!==null&&typeof b.then=="function"?Kr(b):b;var z=wt(),ee=z.queue,le=ee.dispatch;return T!==z.memoizedState&&(nr.flags|=2048,Vo(9,vc.bind(null,ee,T),{destroy:void 0},null)),[b,le,g]}function vc(g,b){g.action=b}function yc(g){var b=wt(),T=Jr;if(T!==null)return id(b,T,g);wt(),b=b.memoizedState,T=wt();var z=T.queue.dispatch;return T.memoizedState=g,[b,z,!1]}function Vo(g,b,T,z){return g={tag:g,create:b,inst:T,deps:z,next:null},b=nr.updateQueue,b===null&&(b=Om(),nr.updateQueue=b),T=b.lastEffect,T===null?b.lastEffect=g.next=g:(z=T.next,T.next=g,g.next=z,b.lastEffect=g),g}function Pu(){return wt().memoizedState}function Cs(g,b,T,z){var ee=hr();nr.flags|=g,ee.memoizedState=Vo(1|b,T,{destroy:void 0},z===void 0?null:z)}function ws(g,b,T,z){var ee=wt();z=z===void 0?null:z;var le=ee.memoizedState.inst;Jr!==null&&z!==null&&Lt(z,Jr.memoizedState.deps)?ee.memoizedState=Vo(b,T,le,z):(nr.flags|=g,ee.memoizedState=Vo(1|b,T,le,z))}function j0(g,b){Cs(8390656,8,g,b)}function vf(g,b){ws(2048,8,g,b)}function Z0(g,b){return ws(4,2,g,b)}function Op(g,b){return ws(4,4,g,b)}function yf(g,b){if(typeof b=="function"){g=g();var T=b(g);return function(){typeof T=="function"?T():b(null)}}if(b!=null)return g=g(),b.current=g,function(){b.current=null}}function ho(g,b,T){T=T!=null?T.concat([g]):null,ws(4,4,yf.bind(null,b,g),T)}function ch(){}function fh(g,b){var T=wt();b=b===void 0?null:b;var z=T.memoizedState;return b!==null&&Lt(b,z[1])?z[0]:(T.memoizedState=[g,b],g)}function Rs(g,b){var T=wt();b=b===void 0?null:b;var z=T.memoizedState;if(b!==null&&Lt(b,z[1]))return z[0];if(z=g(),Rf){J(!0);try{g()}finally{J(!1)}}return T.memoizedState=[z,b],z}function bc(g,b,T){return T===void 0||(zc&1073741824)!==0?g.memoizedState=b:(g.memoizedState=T,g=tl(),nr.lanes|=g,Ao|=g,T)}function H0(g,b,T,z){return Dl(T,b)?T:au.current!==null?(g=bc(g,T,z),Dl(g,b)||(ka=!0),g):(zc&42)===0?(ka=!0,g.memoizedState=T):(g=tl(),nr.lanes|=g,Ao|=g,b)}function xp(g,b,T,z,ee){var le=ru();xu(le!==0&&8>le?le:8);var De=xn.T,Ie={};xn.T=Ie,Qn(g,!1,b,T);try{var We=ee(),vt=xn.S;if(vt!==null&&vt(Ie,We),We!==null&&typeof We=="object"&&typeof We.then=="function"){var Pt=Mt(We,z);Mr(g,b,Pt,Ji(g))}else Mr(g,b,z,Ji(g))}catch(jt){Mr(g,b,{then:function(){},status:"rejected",reason:jt},Ji())}finally{xu(le),xn.T=De}}function Sp(g){var b=g.memoizedState;if(b!==null)return b;b={memoizedState:Su,baseState:Su,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Ca,lastRenderedState:Su},next:null};var T={};return b.next={memoizedState:T,baseState:T,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Ca,lastRenderedState:T},next:null},g.memoizedState=b,g=g.alternate,g!==null&&(g.memoizedState=b),b}function As(){return Qa(An)}function Uo(){return wt().memoizedState}function _p(){return wt().memoizedState}function dh(g){for(var b=g.return;b!==null;){switch(b.tag){case 24:case 3:var T=Ji();g=qt(T);var z=Xt(b,g,T);z!==null&&(yu(z,b,T),nn(z,b,T)),b={cache:fd()},g.payload=b;return}b=b.return}}function Oc(g,b,T){var z=Ji();T={lane:z,revertLane:0,action:T,hasEagerState:!1,eagerState:null,next:null},xc(g)?jo(b,T):(T=Me(g,b,T,z),T!==null&&(yu(T,g,z),Sc(T,b,z)))}function hh(g,b,T){var z=Ji();Mr(g,b,T,z)}function Mr(g,b,T,z){var ee={lane:z,revertLane:0,action:T,hasEagerState:!1,eagerState:null,next:null};if(xc(g))jo(b,ee);else{var le=g.alternate;if(g.lanes===0&&(le===null||le.lanes===0)&&(le=b.lastRenderedReducer,le!==null))try{var De=b.lastRenderedState,Ie=le(De,T);if(ee.hasEagerState=!0,ee.eagerState=Ie,Dl(Ie,De))return we(g,b,ee,0),mr===null&&ve(),!1}catch{}finally{}if(T=Me(g,b,ee,z),T!==null)return yu(T,g,z),Sc(T,b,z),!0}return!1}function Qn(g,b,T,z){if(z={lane:2,revertLane:ze(),action:z,hasEagerState:!1,eagerState:null,next:null},xc(g)){if(b)throw Error(i(479))}else b=Me(g,T,z,2),b!==null&&yu(b,g,2)}function xc(g){var b=g.alternate;return g===nr||b!==null&&b===nr}function jo(g,b){Xi=ss=!0;var T=g.pending;T===null?b.next=b:(b.next=T.next,T.next=b),g.pending=b}function Sc(g,b,T){if((T&4194176)!==0){var z=b.lanes;z&=g.pendingLanes,T|=z,b.lanes=T,j(g,T)}}function ra(g,b,T,z){b=g.memoizedState,T=T(z,b),T=T==null?b:_h({},b,T),g.memoizedState=T,g.lanes===0&&(g.updateQueue.baseState=T)}function Ts(g,b,T,z,ee,le,De){return g=g.stateNode,typeof g.shouldComponentUpdate=="function"?g.shouldComponentUpdate(z,le,De):b.prototype&&b.prototype.isPureReactComponent?!In(T,z)||!In(ee,le):!0}function ad(g,b,T,z){g=b.state,typeof b.componentWillReceiveProps=="function"&&b.componentWillReceiveProps(T,z),typeof b.UNSAFE_componentWillReceiveProps=="function"&&b.UNSAFE_componentWillReceiveProps(T,z),b.state!==g&&xm.enqueueReplaceState(b,b.state,null)}function Ll(g,b){var T=b;if("ref"in b){T={};for(var z in b)z!=="ref"&&(T[z]=b[z])}if(g=g.defaultProps){T===b&&(T=_h({},T));for(var ee in g)T[ee]===void 0&&(T[ee]=g[ee])}return T}function ud(g,b){try{var T=g.onUncaughtError;T(b.value,{componentStack:b.stack})}catch(z){setTimeout(function(){throw z})}}function ia(g,b,T){try{var z=g.onCaughtError;z(T.value,{componentStack:T.stack,errorBoundary:b.tag===1?b.stateNode:null})}catch(ee){setTimeout(function(){throw ee})}}function ks(g,b,T){return T=qt(T),T.tag=3,T.payload={element:null},T.callback=function(){ud(g,b)},T}function Bl(g){return g=qt(g),g.tag=3,g}function ld(g,b,T,z){var ee=T.type.getDerivedStateFromError;if(typeof ee=="function"){var le=z.value;g.payload=function(){return ee(le)},g.callback=function(){ia(b,T,z)}}var De=T.stateNode;De!==null&&typeof De.componentDidCatch=="function"&&(g.callback=function(){ia(b,T,z),typeof ee!="function"&&(To===null?To=new Set([this]):To.add(this));var Ie=z.stack;this.componentDidCatch(z.value,{componentStack:Ie!==null?Ie:""})})}function Ep(g,b,T,z,ee){if(T.flags|=32768,z!==null&&typeof z=="object"&&typeof z.then=="function"){if(b=T.alternate,b!==null&&Ec(b,T,ee,!0),T=Co.current,T!==null){switch(T.tag){case 13:return wo===null?dn():T.alternate===null&&Ir===0&&(Ir=3),T.flags&=-257,T.flags|=65536,T.lanes=ee,z===Md?T.flags|=16384:(b=T.updateQueue,b===null?T.updateQueue=new Set([z]):b.add(z),mn(g,z,ee)),!1;case 22:return T.flags|=65536,z===Md?T.flags|=16384:(b=T.updateQueue,b===null?(b={transitions:null,markerInstances:null,retryQueue:new Set([z])},T.updateQueue=b):(T=b.retryQueue,T===null?b.retryQueue=new Set([z]):T.add(z)),mn(g,z,ee)),!1}throw Error(i(435,T.tag))}return mn(g,z,ee),dn(),!1}if(Vr)return b=Co.current,b!==null?((b.flags&65536)===0&&(b.flags|=256),b.flags|=65536,b.lanes=ee,z!==Do&&(g=Error(i(422),{cause:z}),pe(H(g,T)))):(z!==Do&&(b=Error(i(423),{cause:z}),pe(H(b,T))),g=g.current.alternate,g.flags|=65536,ee&=-ee,g.lanes|=ee,z=H(z,T),ee=ks(g.stateNode,z,ee),Fe(g,ee),Ir!==4&&(Ir=2)),!1;var le=Error(i(520),{cause:z});if(le=H(le,T),Zp===null?Zp=[le]:Zp.push(le),Ir!==4&&(Ir=2),b===null)return!0;z=H(z,T),T=b;do{switch(T.tag){case 3:return T.flags|=65536,g=ee&-ee,T.lanes|=g,g=ks(T.stateNode,z,g),Fe(T,g),!1;case 1:if(b=T.type,le=T.stateNode,(T.flags&128)===0&&(typeof b.getDerivedStateFromError=="function"||le!==null&&typeof le.componentDidCatch=="function"&&(To===null||!To.has(le))))return T.flags|=65536,ee&=-ee,T.lanes|=ee,ee=Bl(ee),ld(ee,g,T,z),Fe(T,ee),!1}T=T.return}while(T!==null);return!1}function ui(g,b,T,z){b.child=g===null?Kg(b,null,T,z):Xs(b,g.child,T,z)}function X0(g,b,T,z,ee){T=T.render;var le=b.ref;if("ref"in z){var De={};for(var Ie in z)Ie!=="ref"&&(De[Ie]=z[Ie])}else De=z;return Fs(b),z=an(g,b,T,De,le,ee),Ie=er(),g!==null&&!ka?(Di(g,b,ee),Ms(g,b,ee)):(Vr&&Ie&&X(b),b.flags|=1,ui(g,b,z,ee),b.child)}function Y0(g,b,T,z,ee){if(g===null){var le=T.type;return typeof le=="function"&&!qn(le)&&le.defaultProps===void 0&&T.compare===null?(b.tag=15,b.type=le,od(g,b,le,z,ee)):(g=Ul(T.type,null,z,b,b.mode,ee),g.ref=b.ref,g.return=b,b.child=g)}if(le=g.child,!Rp(g,ee)){var De=le.memoizedProps;if(T=T.compare,T=T!==null?T:In,T(De,z)&&g.ref===b.ref)return Ms(g,b,ee)}return b.flags|=1,g=Ja(le,z),g.ref=b.ref,g.return=b,b.child=g}function od(g,b,T,z,ee){if(g!==null){var le=g.memoizedProps;if(In(le,z)&&g.ref===b.ref)if(ka=!1,b.pendingProps=z=le,Rp(g,ee))(g.flags&131072)!==0&&(ka=!0);else return b.lanes=g.lanes,Ms(g,b,ee)}return ha(g,b,T,z,ee)}function G0(g,b,T){var z=b.pendingProps,ee=z.children,le=(b.stateNode._pendingVisibility&2)!==0,De=g!==null?g.memoizedState:null;if(sd(g,b),z.mode==="hidden"||le){if((b.flags&128)!==0){if(z=De!==null?De.baseLanes|T:T,g!==null){for(ee=b.child=g.child,le=0;ee!==null;)le=le|ee.lanes|ee.childLanes,ee=ee.sibling;b.childLanes=le&~z}else b.childLanes=0,b.child=null;return _c(g,b,z,T)}if((T&536870912)!==0)b.memoizedState={baseLanes:0,cachePool:null},g!==null&&pu(b,De!==null?De.cachePool:null),De!==null?ar(b,De):Te(),gt(b);else return b.lanes=b.childLanes=536870912,_c(g,b,De!==null?De.baseLanes|T:T,T)}else De!==null?(pu(b,De.cachePool),ar(b,De),Ct(),b.memoizedState=null):(g!==null&&pu(b,null),Te(),Ct());return ui(g,b,ee,T),b.child}function _c(g,b,T,z){var ee=Ns();return ee=ee===null?null:{parent:tu?ei._currentValue:ei._currentValue2,pool:ee},b.memoizedState={baseLanes:T,cachePool:ee},g!==null&&pu(b,null),Te(),gt(b),g!==null&&Ec(g,b,z,!0),null}function sd(g,b){var T=b.ref;if(T===null)g!==null&&g.ref!==null&&(b.flags|=2097664);else{if(typeof T!="function"&&typeof T!="object")throw Error(i(284));(g===null||g.ref!==T)&&(b.flags|=2097664)}}function ha(g,b,T,z,ee){return Fs(b),T=an(g,b,T,z,void 0,ee),z=er(),g!==null&&!ka?(Di(g,b,ee),Ms(g,b,ee)):(Vr&&z&&X(b),b.flags|=1,ui(g,b,T,ee),b.child)}function cd(g,b,T,z,ee,le){return Fs(b),b.updateQueue=null,T=sn(b,z,T,ee),_n(g),z=er(),g!==null&&!ka?(Di(g,b,le),Ms(g,b,le)):(Vr&&z&&X(b),b.flags|=1,ui(g,b,T,le),b.child)}function Ba(g,b,T,z,ee){if(Fs(b),b.stateNode===null){var le=Lc,De=T.contextType;typeof De=="object"&&De!==null&&(le=Qa(De)),le=new T(z,le),b.memoizedState=le.state!==null&&le.state!==void 0?le.state:null,le.updater=xm,b.stateNode=le,le._reactInternals=b,le=b.stateNode,le.props=z,le.state=b.memoizedState,le.refs={},bt(b),De=T.contextType,le.context=typeof De=="object"&&De!==null?Qa(De):Lc,le.state=b.memoizedState,De=T.getDerivedStateFromProps,typeof De=="function"&&(ra(b,T,De,z),le.state=b.memoizedState),typeof T.getDerivedStateFromProps=="function"||typeof le.getSnapshotBeforeUpdate=="function"||typeof le.UNSAFE_componentWillMount!="function"&&typeof le.componentWillMount!="function"||(De=le.state,typeof le.componentWillMount=="function"&&le.componentWillMount(),typeof le.UNSAFE_componentWillMount=="function"&&le.UNSAFE_componentWillMount(),De!==le.state&&xm.enqueueReplaceState(le,le.state,null),kt(b,z,le,ee),ht(),le.state=b.memoizedState),typeof le.componentDidMount=="function"&&(b.flags|=4194308),z=!0}else if(g===null){le=b.stateNode;var Ie=b.memoizedProps,We=Ll(T,Ie);le.props=We;var vt=le.context,Pt=T.contextType;De=Lc,typeof Pt=="object"&&Pt!==null&&(De=Qa(Pt));var jt=T.getDerivedStateFromProps;Pt=typeof jt=="function"||typeof le.getSnapshotBeforeUpdate=="function",Ie=b.pendingProps!==Ie,Pt||typeof le.UNSAFE_componentWillReceiveProps!="function"&&typeof le.componentWillReceiveProps!="function"||(Ie||vt!==De)&&ad(b,le,z,De),Qc=!1;var Wt=b.memoizedState;le.state=Wt,kt(b,z,le,ee),ht(),vt=b.memoizedState,Ie||Wt!==vt||Qc?(typeof jt=="function"&&(ra(b,T,jt,z),vt=b.memoizedState),(We=Qc||Ts(b,T,We,z,Wt,vt,De))?(Pt||typeof le.UNSAFE_componentWillMount!="function"&&typeof le.componentWillMount!="function"||(typeof le.componentWillMount=="function"&&le.componentWillMount(),typeof le.UNSAFE_componentWillMount=="function"&&le.UNSAFE_componentWillMount()),typeof le.componentDidMount=="function"&&(b.flags|=4194308)):(typeof le.componentDidMount=="function"&&(b.flags|=4194308),b.memoizedProps=z,b.memoizedState=vt),le.props=z,le.state=vt,le.context=De,z=We):(typeof le.componentDidMount=="function"&&(b.flags|=4194308),z=!1)}else{le=b.stateNode,zt(g,b),De=b.memoizedProps,Pt=Ll(T,De),le.props=Pt,jt=b.pendingProps,Wt=le.context,vt=T.contextType,We=Lc,typeof vt=="object"&&vt!==null&&(We=Qa(vt)),Ie=T.getDerivedStateFromProps,(vt=typeof Ie=="function"||typeof le.getSnapshotBeforeUpdate=="function")||typeof le.UNSAFE_componentWillReceiveProps!="function"&&typeof le.componentWillReceiveProps!="function"||(De!==jt||Wt!==We)&&ad(b,le,z,We),Qc=!1,Wt=b.memoizedState,le.state=Wt,kt(b,z,le,ee),ht();var Hn=b.memoizedState;De!==jt||Wt!==Hn||Qc||g!==null&&g.dependencies!==null&&Xo(g.dependencies)?(typeof Ie=="function"&&(ra(b,T,Ie,z),Hn=b.memoizedState),(Pt=Qc||Ts(b,T,Pt,z,Wt,Hn,We)||g!==null&&g.dependencies!==null&&Xo(g.dependencies))?(vt||typeof le.UNSAFE_componentWillUpdate!="function"&&typeof le.componentWillUpdate!="function"||(typeof le.componentWillUpdate=="function"&&le.componentWillUpdate(z,Hn,We),typeof le.UNSAFE_componentWillUpdate=="function"&&le.UNSAFE_componentWillUpdate(z,Hn,We)),typeof le.componentDidUpdate=="function"&&(b.flags|=4),typeof le.getSnapshotBeforeUpdate=="function"&&(b.flags|=1024)):(typeof le.componentDidUpdate!="function"||De===g.memoizedProps&&Wt===g.memoizedState||(b.flags|=4),typeof le.getSnapshotBeforeUpdate!="function"||De===g.memoizedProps&&Wt===g.memoizedState||(b.flags|=1024),b.memoizedProps=z,b.memoizedState=Hn),le.props=z,le.state=Hn,le.context=We,z=Pt):(typeof le.componentDidUpdate!="function"||De===g.memoizedProps&&Wt===g.memoizedState||(b.flags|=4),typeof le.getSnapshotBeforeUpdate!="function"||De===g.memoizedProps&&Wt===g.memoizedState||(b.flags|=1024),z=!1)}return le=z,sd(g,b),z=(b.flags&128)!==0,le||z?(le=b.stateNode,T=z&&typeof T.getDerivedStateFromError!="function"?null:le.render(),b.flags|=1,g!==null&&z?(b.child=Xs(b,g.child,null,ee),b.child=Xs(b,null,T,ee)):ui(g,b,T,ee),b.memoizedState=le.state,g=b.child):g=Ms(g,b,ee),g}function Eg(g,b,T,z){return se(),b.flags|=256,ui(g,b,T,z),b.child}function Dp(g){return{baseLanes:g,cachePool:Ap()}}function Zo(g,b,T){return g=g!==null?g.childLanes&~T:0,b&&(g|=eo),g}function ph(g,b,T){var z=b.pendingProps,ee=!1,le=(b.flags&128)!==0,De;if((De=le)||(De=g!==null&&g.memoizedState===null?!1:(qa.current&2)!==0),De&&(ee=!0,b.flags&=-129),De=(b.flags&32)!==0,b.flags&=-33,g===null){if(Vr){if(ee?nt(b):Ct(),Vr){var Ie=Eu,We;(We=Ie)&&(Ie=qy(Ie,ls),Ie!==null?(b.memoizedState={dehydrated:Ie,treeContext:js!==null?{id:us,overflow:_o}:null,retryLane:536870912},We=n(18,null,null,0),We.stateNode=Ie,We.return=b,b.child=We,sl=b,Eu=null,We=!0):We=!1),We||ge(b)}if(Ie=b.memoizedState,Ie!==null&&(Ie=Ie.dehydrated,Ie!==null))return Dd(Ie)?b.lanes=16:b.lanes=536870912,null;Vt(b)}return Ie=z.children,z=z.fallback,ee?(Ct(),ee=b.mode,Ie=bi({mode:"hidden",children:Ie},ee),z=Lu(z,ee,T,null),Ie.return=b,z.return=b,Ie.sibling=z,b.child=Ie,ee=b.child,ee.memoizedState=Dp(T),ee.childLanes=Zo(g,De,T),b.memoizedState=fl,z):(nt(b),Ps(b,Ie))}if(We=g.memoizedState,We!==null&&(Ie=We.dehydrated,Ie!==null)){if(le)b.flags&256?(nt(b),b.flags&=-257,b=Ho(g,b,T)):b.memoizedState!==null?(Ct(),b.child=g.child,b.flags|=128,b=null):(Ct(),ee=z.fallback,Ie=b.mode,z=bi({mode:"visible",children:z.children},Ie),ee=Lu(ee,Ie,T,null),ee.flags|=2,z.return=b,ee.return=b,z.sibling=ee,b.child=z,Xs(b,g.child,null,T),z=b.child,z.memoizedState=Dp(T),z.childLanes=Zo(g,De,T),b.memoizedState=fl,b=ee);else if(nt(b),Dd(Ie))De=Bg(Ie).digest,z=Error(i(419)),z.stack="",z.digest=De,pe({value:z,source:null,stack:null}),b=Ho(g,b,T);else if(ka||Ec(g,b,T,!1),De=(T&g.childLanes)!==0,ka||De){if(De=mr,De!==null){if(z=T&-T,(z&42)!==0)z=1;else switch(z){case 2:z=1;break;case 8:z=4;break;case 32:z=16;break;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:z=64;break;case 268435456:z=134217728;break;default:z=0}if(z=(z&(De.suspendedLanes|T))!==0?0:z,z!==0&&z!==We.retryLane)throw We.retryLane=z,Re(g,z),yu(De,g,z),Y1}Lg(Ie)||dn(),b=Ho(g,b,T)}else Lg(Ie)?(b.flags|=128,b.child=g.child,b=Se.bind(null,g),z1(Ie,b),b=null):(g=We.treeContext,Ra&&(Eu=sm(Ie),sl=b,Vr=!0,Eo=null,ls=!1,g!==null&&($l[iu++]=us,$l[iu++]=_o,$l[iu++]=js,us=g.id,_o=g.overflow,js=b)),b=Ps(b,z.children),b.flags|=4096);return b}return ee?(Ct(),ee=z.fallback,Ie=b.mode,We=g.child,le=We.sibling,z=Ja(We,{mode:"hidden",children:z.children}),z.subtreeFlags=We.subtreeFlags&31457280,le!==null?ee=Ja(le,ee):(ee=Lu(ee,Ie,T,null),ee.flags|=2),ee.return=b,z.return=b,z.sibling=ee,b.child=z,z=ee,ee=b.child,Ie=g.child.memoizedState,Ie===null?Ie=Dp(T):(We=Ie.cachePool,We!==null?(le=tu?ei._currentValue:ei._currentValue2,We=We.parent!==le?{parent:le,pool:le}:We):We=Ap(),Ie={baseLanes:Ie.baseLanes|T,cachePool:We}),ee.memoizedState=Ie,ee.childLanes=Zo(g,De,T),b.memoizedState=fl,z):(nt(b),T=g.child,g=T.sibling,T=Ja(T,{mode:"visible",children:z.children}),T.return=b,T.sibling=null,g!==null&&(De=b.deletions,De===null?(b.deletions=[g],b.flags|=16):De.push(g)),b.child=T,b.memoizedState=null,T)}function Ps(g,b){return b=bi({mode:"visible",children:b},g.mode),b.return=g,g.child=b}function bi(g,b){return Tc(g,b,0,null)}function Ho(g,b,T){return Xs(b,g.child,null,T),g=Ps(b,b.pendingProps.children),g.flags|=2,b.memoizedState=null,g}function $p(g,b,T){g.lanes|=b;var z=g.alternate;z!==null&&(z.lanes|=b),W0(g.return,b,T)}function Cp(g,b,T,z,ee){var le=g.memoizedState;le===null?g.memoizedState={isBackwards:b,rendering:null,renderingStartTime:0,last:z,tail:T,tailMode:ee}:(le.isBackwards=b,le.rendering=null,le.renderingStartTime=0,le.last=z,le.tail=T,le.tailMode=ee)}function wp(g,b,T){var z=b.pendingProps,ee=z.revealOrder,le=z.tail;if(ui(g,b,z.children,T),z=qa.current,(z&2)!==0)z=z&1|2,b.flags|=128;else{if(g!==null&&(g.flags&128)!==0)e:for(g=b.child;g!==null;){if(g.tag===13)g.memoizedState!==null&&$p(g,T,b);else if(g.tag===19)$p(g,T,b);else if(g.child!==null){g.child.return=g,g=g.child;continue}if(g===b)break e;for(;g.sibling===null;){if(g.return===null||g.return===b)break e;g=g.return}g.sibling.return=g.return,g=g.sibling}z&=1}switch(E(qa,z),ee){case"forwards":for(T=b.child,ee=null;T!==null;)g=T.alternate,g!==null&&Yt(g)===null&&(ee=T),T=T.sibling;T=ee,T===null?(ee=b.child,b.child=null):(ee=T.sibling,T.sibling=null),Cp(b,!1,ee,T,le);break;case"backwards":for(T=null,ee=b.child,b.child=null;ee!==null;){if(g=ee.alternate,g!==null&&Yt(g)===null){b.child=ee;break}g=ee.sibling,ee.sibling=T,T=ee,ee=g}Cp(b,!0,T,null,le);break;case"together":Cp(b,!1,null,null,void 0);break;default:b.memoizedState=null}return b.child}function Ms(g,b,T){if(g!==null&&(b.dependencies=g.dependencies),Ao|=b.lanes,(T&b.childLanes)===0)if(g!==null){if(Ec(g,b,T,!1),(T&b.childLanes)===0)return null}else return null;if(g!==null&&b.child!==g.child)throw Error(i(153));if(b.child!==null){for(g=b.child,T=Ja(g,g.pendingProps),b.child=T,T.return=b;g.sibling!==null;)g=g.sibling,T=T.sibling=Ja(g,g.pendingProps),T.return=b;T.sibling=null}return b.child}function Rp(g,b){return(g.lanes&b)!==0?!0:(g=g.dependencies,!!(g!==null&&Xo(g)))}function Dy(g,b,T){switch(b.tag){case 3:ne(b,b.stateNode.containerInfo),po(b,ei,g.memoizedState.cache),se();break;case 27:case 5:ie(b);break;case 4:ne(b,b.stateNode.containerInfo);break;case 10:po(b,b.type,b.memoizedProps.value);break;case 13:var z=b.memoizedState;if(z!==null)return z.dehydrated!==null?(nt(b),b.flags|=128,null):(T&b.child.childLanes)!==0?ph(g,b,T):(nt(b),g=Ms(g,b,T),g!==null?g.sibling:null);nt(b);break;case 19:var ee=(g.flags&128)!==0;if(z=(T&b.childLanes)!==0,z||(Ec(g,b,T,!1),z=(T&b.childLanes)!==0),ee){if(z)return wp(g,b,T);b.flags|=128}if(ee=b.memoizedState,ee!==null&&(ee.rendering=null,ee.tail=null,ee.lastEffect=null),E(qa,qa.current),z)break;return null;case 22:case 23:return b.lanes=0,G0(g,b,T);case 24:po(b,ei,g.memoizedState.cache)}return Ms(g,b,T)}function K0(g,b,T){if(g!==null)if(g.memoizedProps!==b.pendingProps)ka=!0;else{if(!Rp(g,T)&&(b.flags&128)===0)return ka=!1,Dy(g,b,T);ka=(g.flags&131072)!==0}else ka=!1,Vr&&(b.flags&1048576)!==0&&G(b,Td,b.index);switch(b.lanes=0,b.tag){case 16:e:{g=b.pendingProps;var z=b.elementType,ee=z._init;if(z=ee(z._payload),b.type=z,typeof z=="function")qn(z)?(g=Ll(z,g),b.tag=1,b=Ba(null,b,z,g,T)):(b.tag=0,b=ha(null,b,z,g,T));else{if(z!=null){if(ee=z.$$typeof,ee===Mp){b.tag=11,b=X0(null,b,z,g,T);break e}else if(ee===Fp){b.tag=14,b=Y0(null,b,z,g,T);break e}}throw b=u(z)||z,Error(i(306,b,""))}}return b;case 0:return ha(g,b,b.type,b.pendingProps,T);case 1:return z=b.type,ee=Ll(z,b.pendingProps),Ba(g,b,z,ee,T);case 3:e:{if(ne(b,b.stateNode.containerInfo),g===null)throw Error(i(387));var le=b.pendingProps;ee=b.memoizedState,z=ee.element,zt(g,b),kt(b,le,null,T);var De=b.memoizedState;if(le=De.cache,po(b,ei,le),le!==ee.cache&&mh(b,[ei],T,!0),ht(),le=De.element,Ra&&ee.isDehydrated)if(ee={element:le,isDehydrated:!1,cache:De.cache},b.updateQueue.baseState=ee,b.memoizedState=ee,b.flags&256){b=Eg(g,b,le,T);break e}else if(le!==z){z=H(Error(i(424)),b),pe(z),b=Eg(g,b,le,T);break e}else for(Ra&&(Eu=Qy(b.stateNode.containerInfo),sl=b,Vr=!0,Eo=null,ls=!0),T=Kg(b,null,le,T),b.child=T;T;)T.flags=T.flags&-3|4096,T=T.sibling;else{if(se(),le===z){b=Ms(g,b,T);break e}ui(g,b,le,T)}b=b.child}return b;case 26:if(Yl)return sd(g,b),g===null?(T=dm(b.type,null,b.pendingProps,null))?b.memoizedState=T:Vr||(b.stateNode=jg(b.type,b.pendingProps,Zs.current,b)):b.memoizedState=dm(b.type,g.memoizedProps,b.pendingProps,g.memoizedState),null;case 27:if(_u)return ie(b),g===null&&_u&&Vr&&(z=b.stateNode=Cd(b.type,b.pendingProps,Zs.current,qr.current,!1),sl=b,ls=!0,Eu=zp(z)),z=b.pendingProps.children,g!==null||Vr?ui(g,b,z,T):b.child=Xs(b,null,z,T),sd(g,b),b.child;case 5:return g===null&&Vr&&(U1(b.type,b.pendingProps,qr.current),(ee=z=Eu)&&(z=zy(z,b.type,b.pendingProps,ls),z!==null?(b.stateNode=z,sl=b,Eu=zp(z),ls=!1,ee=!0):ee=!1),ee||ge(b)),ie(b),ee=b.type,le=b.pendingProps,De=g!==null?g.memoizedProps:null,z=le.children,Sd(ee,le)?z=null:De!==null&&Sd(ee,De)&&(b.flags|=32),b.memoizedState!==null&&(ee=an(g,b,Ln,null,null,T),tu?An._currentValue=ee:An._currentValue2=ee),sd(g,b),ui(g,b,z,T),b.child;case 6:return g===null&&Vr&&(j1(b.pendingProps,qr.current),(g=T=Eu)&&(T=Qg(T,b.pendingProps,ls),T!==null?(b.stateNode=T,sl=b,Eu=null,g=!0):g=!1),g||ge(b)),null;case 13:return ph(g,b,T);case 4:return ne(b,b.stateNode.containerInfo),z=b.pendingProps,g===null?b.child=Xs(b,null,z,T):ui(g,b,z,T),b.child;case 11:return X0(g,b,b.type,b.pendingProps,T);case 7:return ui(g,b,b.pendingProps,T),b.child;case 8:return ui(g,b,b.pendingProps.children,T),b.child;case 12:return ui(g,b,b.pendingProps.children,T),b.child;case 10:return z=b.pendingProps,po(b,b.type,z.value),ui(g,b,z.children,T),b.child;case 9:return ee=b.type._context,z=b.pendingProps.children,Fs(b),ee=Qa(ee),z=z(ee),b.flags|=1,ui(g,b,z,T),b.child;case 14:return Y0(g,b,b.type,b.pendingProps,T);case 15:return od(g,b,b.type,b.pendingProps,T);case 19:return wp(g,b,T);case 22:return G0(g,b,T);case 24:return Fs(b),z=Qa(ei),g===null?(ee=Ns(),ee===null&&(ee=mr,le=fd(),ee.pooledCache=le,le.refCount++,le!==null&&(ee.pooledCacheLanes|=T),ee=le),b.memoizedState={parent:z,cache:ee},bt(b),po(b,ei,ee)):((g.lanes&T)!==0&&(zt(g,b),kt(b,null,null,T),ht()),ee=g.memoizedState,le=b.memoizedState,ee.parent!==z?(ee={parent:z,cache:z},b.memoizedState=ee,b.lanes===0&&(b.memoizedState=b.updateQueue.baseState=ee),po(b,ei,z)):(z=le.cache,po(b,ei,z),z!==ee.cache&&mh(b,[ei],T,!0))),ui(g,b,b.pendingProps.children,T),b.child;case 29:throw b.pendingProps}throw Error(i(156,b.tag))}function po(g,b,T){tu?(E(Wl,b._currentValue),b._currentValue=T):(E(Wl,b._currentValue2),b._currentValue2=T)}function Ku(g){var b=Wl.current;tu?g._currentValue=b:g._currentValue2=b,_(Wl)}function W0(g,b,T){for(;g!==null;){var z=g.alternate;if((g.childLanes&b)!==b?(g.childLanes|=b,z!==null&&(z.childLanes|=b)):z!==null&&(z.childLanes&b)!==b&&(z.childLanes|=b),g===T)break;g=g.return}}function mh(g,b,T,z){var ee=g.child;for(ee!==null&&(ee.return=g);ee!==null;){var le=ee.dependencies;if(le!==null){var De=ee.child;le=le.firstContext;e:for(;le!==null;){var Ie=le;le=ee;for(var We=0;We<b.length;We++)if(Ie.context===b[We]){le.lanes|=T,Ie=le.alternate,Ie!==null&&(Ie.lanes|=T),W0(le.return,T,g),z||(De=null);break e}le=Ie.next}}else if(ee.tag===18){if(De=ee.return,De===null)throw Error(i(341));De.lanes|=T,le=De.alternate,le!==null&&(le.lanes|=T),W0(De,T,g),De=null}else De=ee.child;if(De!==null)De.return=ee;else for(De=ee;De!==null;){if(De===g){De=null;break}if(ee=De.sibling,ee!==null){ee.return=De.return,De=ee;break}De=De.return}ee=De}}function Ec(g,b,T,z){g=null;for(var ee=b,le=!1;ee!==null;){if(!le){if((ee.flags&524288)!==0)le=!0;else if((ee.flags&262144)!==0)break}if(ee.tag===10){var De=ee.alternate;if(De===null)throw Error(i(387));if(De=De.memoizedProps,De!==null){var Ie=ee.type;Dl(ee.pendingProps.value,De.value)||(g!==null?g.push(Ie):g=[Ie])}}else if(ee===kh.current){if(De=ee.alternate,De===null)throw Error(i(387));De.memoizedState.memoizedState!==ee.memoizedState.memoizedState&&(g!==null?g.push(An):g=[An])}ee=ee.return}g!==null&&mh(b,g,T,z),b.flags|=262144}function Xo(g){for(g=g.firstContext;g!==null;){var b=g.context;if(!Dl(tu?b._currentValue:b._currentValue2,g.memoizedValue))return!0;g=g.next}return!1}function Fs(g){Nd=g,Ur=null,g=g.dependencies,g!==null&&(g.firstContext=null)}function Qa(g){return vh(Nd,g)}function gh(g,b){return Nd===null&&Fs(g),vh(g,b)}function vh(g,b){var T=tu?b._currentValue:b._currentValue2;if(b={context:b,memoizedValue:T,next:null},Ur===null){if(g===null)throw Error(i(308));Ur=b,g.dependencies={lanes:0,firstContext:b},g.flags|=524288}else Ur=Ur.next=b;return T}function fd(){return{controller:new G1,data:new Map,refCount:0}}function dd(g){g.refCount--,g.refCount===0&&Af(Fh,function(){g.controller.abort()})}function Ns(){var g=Tf.current;return g!==null?g:mr.pooledCache}function pu(g,b){b===null?E(Tf,Tf.current):E(Tf,b.pool)}function Ap(){var g=Ns();return g===null?null:{parent:tu?ei._currentValue:ei._currentValue2,pool:g}}function Ga(g){g.flags|=4}function Tp(g,b){if(g!==null&&g.child===b.child)return!1;if((b.flags&16)!==0)return!0;for(g=b.child;g!==null;){if((g.flags&13878)!==0||(g.subtreeFlags&13878)!==0)return!0;g=g.sibling}return!1}function hd(g,b,T,z){if(nu)for(T=b.child;T!==null;){if(T.tag===5||T.tag===6)Mc(g,T.stateNode);else if(!(T.tag===4||_u&&T.tag===27)&&T.child!==null){T.child.return=T,T=T.child;continue}if(T===b)break;for(;T.sibling===null;){if(T.return===null||T.return===b)return;T=T.return}T.sibling.return=T.return,T=T.sibling}else if(il)for(var ee=b.child;ee!==null;){if(ee.tag===5){var le=ee.stateNode;T&&z&&(le=Ly(le,ee.type,ee.memoizedProps)),Mc(g,le)}else if(ee.tag===6)le=ee.stateNode,T&&z&&(le=Hl(le,ee.memoizedProps)),Mc(g,le);else if(ee.tag!==4){if(ee.tag===22&&ee.memoizedState!==null)le=ee.child,le!==null&&(le.return=ee),hd(g,ee,!0,!0);else if(ee.child!==null){ee.child.return=ee,ee=ee.child;continue}}if(ee===b)break;for(;ee.sibling===null;){if(ee.return===null||ee.return===b)return;ee=ee.return}ee.sibling.return=ee.return,ee=ee.sibling}}function bf(g,b,T,z){if(il)for(var ee=b.child;ee!==null;){if(ee.tag===5){var le=ee.stateNode;T&&z&&(le=Ly(le,ee.type,ee.memoizedProps)),Ch(g,le)}else if(ee.tag===6)le=ee.stateNode,T&&z&&(le=Hl(le,ee.memoizedProps)),Ch(g,le);else if(ee.tag!==4){if(ee.tag===22&&ee.memoizedState!==null)le=ee.child,le!==null&&(le.return=ee),bf(g,ee,!(ee.memoizedProps!==null&&ee.memoizedProps.mode==="manual"),!0);else if(ee.child!==null){ee.child.return=ee,ee=ee.child;continue}}if(ee===b)break;for(;ee.sibling===null;){if(ee.return===null||ee.return===b)return;ee=ee.return}ee.sibling.return=ee.return,ee=ee.sibling}}function pd(g,b){if(il&&Tp(g,b)){g=b.stateNode;var T=g.containerInfo,z=Ig();bf(z,b,!1,!1),g.pendingChildren=z,Ga(b),Ny(T,z)}}function Wu(g,b,T,z){if(nu)g.memoizedProps!==z&&Ga(b);else if(il){var ee=g.stateNode,le=g.memoizedProps;if((g=Tp(g,b))||le!==z){var De=qr.current;le=Fy(ee,T,le,z,!g,null),le===ee?b.stateNode=ee:(ns(le,T,z,De)&&Ga(b),b.stateNode=le,g?hd(le,b,!1,!1):Ga(b))}else b.stateNode=ee}}function mo(g,b,T){if(Q1(b,T)){if(g.flags|=16777216,!xo(b,T))if(Bi())g.flags|=8192;else throw wf=Md,Gg}else g.flags&=-16777217}function J0(g,b){if(Hy(b)){if(g.flags|=16777216,!Zg(b))if(Bi())g.flags|=8192;else throw wf=Md,Gg}else g.flags&=-16777217}function mu(g,b){b!==null&&(g.flags|=4),g.flags&16384&&(b=g.tag!==22?I():536870912,g.lanes|=b,Vc|=b)}function Of(g,b){if(!Vr)switch(g.tailMode){case"hidden":b=g.tail;for(var T=null;b!==null;)b.alternate!==null&&(T=b),b=b.sibling;T===null?g.tail=null:T.sibling=null;break;case"collapsed":T=g.tail;for(var z=null;T!==null;)T.alternate!==null&&(z=T),T=T.sibling;z===null?b||g.tail===null?g.tail=null:g.tail.sibling=null:z.sibling=null}}function Ar(g){var b=g.alternate!==null&&g.alternate.child===g.child,T=0,z=0;if(b)for(var ee=g.child;ee!==null;)T|=ee.lanes|ee.childLanes,z|=ee.subtreeFlags&31457280,z|=ee.flags&31457280,ee.return=g,ee=ee.sibling;else for(ee=g.child;ee!==null;)T|=ee.lanes|ee.childLanes,z|=ee.subtreeFlags,z|=ee.flags,ee.return=g,ee=ee.sibling;return g.subtreeFlags|=z,g.childLanes=T,b}function yh(g,b,T){var z=b.pendingProps;switch(B(b),b.tag){case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return Ar(b),null;case 1:return Ar(b),null;case 3:return T=b.stateNode,z=null,g!==null&&(z=g.memoizedState.cache),b.memoizedState.cache!==z&&(b.flags|=2048),Ku(ei),oe(),T.pendingContext&&(T.context=T.pendingContext,T.pendingContext=null),(g===null||g.child===null)&&(Ee(b)?Ga(b):g===null||g.memoizedState.isDehydrated&&(b.flags&256)===0||(b.flags|=1024,Eo!==null&&(zl(Eo),Eo=null))),pd(g,b),Ar(b),null;case 26:if(Yl){T=b.type;var ee=b.memoizedState;return g===null?(Ga(b),ee!==null?(Ar(b),J0(b,ee)):(Ar(b),mo(b,T,z))):ee?ee!==g.memoizedState?(Ga(b),Ar(b),J0(b,ee)):(Ar(b),b.flags&=-16777217):(nu?g.memoizedProps!==z&&Ga(b):Wu(g,b,T,z),Ar(b),mo(b,T,z)),null}case 27:if(_u){if(de(b),T=Zs.current,ee=b.type,g!==null&&b.stateNode!=null)nu?g.memoizedProps!==z&&Ga(b):Wu(g,b,ee,z);else{if(!z){if(b.stateNode===null)throw Error(i(166));return Ar(b),null}g=qr.current,Ee(b)?ye(b,g):(g=Cd(ee,z,T,g,!0),b.stateNode=g,Ga(b))}return Ar(b),null}case 5:if(de(b),T=b.type,g!==null&&b.stateNode!=null)Wu(g,b,T,z);else{if(!z){if(b.stateNode===null)throw Error(i(166));return Ar(b),null}g=qr.current,Ee(b)?ye(b,g):(ee=Rg(T,z,Zs.current,g,b),hd(ee,b,!1,!1),b.stateNode=ee,ns(ee,T,z,g)&&Ga(b))}return Ar(b),mo(b,b.type,b.pendingProps),null;case 6:if(g&&b.stateNode!=null)T=g.memoizedProps,nu?T!==z&&Ga(b):il&&(T!==z?(b.stateNode=am(z,Zs.current,qr.current,b),Ga(b)):b.stateNode=g.stateNode);else{if(typeof z!="string"&&b.stateNode===null)throw Error(i(166));if(g=Zs.current,T=qr.current,Ee(b)){if(!Ra)throw Error(i(176));if(g=b.stateNode,T=b.memoizedProps,z=null,ee=sl,ee!==null)switch(ee.tag){case 27:case 5:z=ee.memoizedProps}zg(g,T,b,z)||ge(b)}else b.stateNode=am(z,g,T,b)}return Ar(b),null;case 13:if(z=b.memoizedState,g===null||g.memoizedState!==null&&g.memoizedState.dehydrated!==null){if(ee=Ee(b),z!==null&&z.dehydrated!==null){if(g===null){if(!ee)throw Error(i(318));if(!Ra)throw Error(i(344));if(ee=b.memoizedState,ee=ee!==null?ee.dehydrated:null,!ee)throw Error(i(317));So(ee,b)}else se(),(b.flags&128)===0&&(b.memoizedState=null),b.flags|=4;Ar(b),ee=!1}else Eo!==null&&(zl(Eo),Eo=null),ee=!0;if(!ee)return b.flags&256?(Vt(b),b):(Vt(b),null)}if(Vt(b),(b.flags&128)!==0)return b.lanes=T,b;if(T=z!==null,g=g!==null&&g.memoizedState!==null,T){z=b.child,ee=null,z.alternate!==null&&z.alternate.memoizedState!==null&&z.alternate.memoizedState.cachePool!==null&&(ee=z.alternate.memoizedState.cachePool.pool);var le=null;z.memoizedState!==null&&z.memoizedState.cachePool!==null&&(le=z.memoizedState.cachePool.pool),le!==ee&&(z.flags|=2048)}return T!==g&&T&&(b.child.flags|=8192),mu(b,b.updateQueue),Ar(b),null;case 4:return oe(),pd(g,b),g===null&&Qu(b.stateNode.containerInfo),Ar(b),null;case 10:return Ku(b.type),Ar(b),null;case 19:if(_(qa),ee=b.memoizedState,ee===null)return Ar(b),null;if(z=(b.flags&128)!==0,le=ee.rendering,le===null)if(z)Of(ee,!1);else{if(Ir!==0||g!==null&&(g.flags&128)!==0)for(g=b.child;g!==null;){if(le=Yt(g),le!==null){for(b.flags|=128,Of(ee,!1),g=le.updateQueue,b.updateQueue=g,mu(b,g),b.subtreeFlags=0,g=T,T=b.child;T!==null;)Ls(T,g),T=T.sibling;return E(qa,qa.current&1|2),b.child}g=g.sibling}ee.tail!==null&&Hi()>oa&&(b.flags|=128,z=!0,Of(ee,!1),b.lanes=4194304)}else{if(!z)if(g=Yt(le),g!==null){if(b.flags|=128,z=!0,g=g.updateQueue,b.updateQueue=g,mu(b,g),Of(ee,!0),ee.tail===null&&ee.tailMode==="hidden"&&!le.alternate&&!Vr)return Ar(b),null}else 2*Hi()-ee.renderingStartTime>oa&&T!==536870912&&(b.flags|=128,z=!0,Of(ee,!1),b.lanes=4194304);ee.isBackwards?(le.sibling=b.child,b.child=le):(g=ee.last,g!==null?g.sibling=le:b.child=le,ee.last=le)}return ee.tail!==null?(b=ee.tail,ee.rendering=b,ee.tail=b.sibling,ee.renderingStartTime=Hi(),b.sibling=null,g=qa.current,E(qa,z?g&1|2:g&1),b):(Ar(b),null);case 22:case 23:return Vt(b),Ve(),z=b.memoizedState!==null,g!==null?g.memoizedState!==null!==z&&(b.flags|=8192):z&&(b.flags|=8192),z?(T&536870912)!==0&&(b.flags&128)===0&&(Ar(b),b.subtreeFlags&6&&(b.flags|=8192)):Ar(b),T=b.updateQueue,T!==null&&mu(b,T.retryQueue),T=null,g!==null&&g.memoizedState!==null&&g.memoizedState.cachePool!==null&&(T=g.memoizedState.cachePool.pool),z=null,b.memoizedState!==null&&b.memoizedState.cachePool!==null&&(z=b.memoizedState.cachePool.pool),z!==T&&(b.flags|=2048),g!==null&&_(Tf),null;case 24:return T=null,g!==null&&(T=g.memoizedState.cache),b.memoizedState.cache!==T&&(b.flags|=2048),Ku(ei),Ar(b),null;case 25:return null}throw Error(i(156,b.tag))}function kp(g,b){switch(B(b),b.tag){case 1:return g=b.flags,g&65536?(b.flags=g&-65537|128,b):null;case 3:return Ku(ei),oe(),g=b.flags,(g&65536)!==0&&(g&128)===0?(b.flags=g&-65537|128,b):null;case 26:case 27:case 5:return de(b),null;case 13:if(Vt(b),g=b.memoizedState,g!==null&&g.dehydrated!==null){if(b.alternate===null)throw Error(i(340));se()}return g=b.flags,g&65536?(b.flags=g&-65537|128,b):null;case 19:return _(qa),null;case 4:return oe(),null;case 10:return Ku(b.type),null;case 22:case 23:return Vt(b),Ve(),g!==null&&_(Tf),g=b.flags,g&65536?(b.flags=g&-65537|128,b):null;case 24:return Ku(ei),null;case 25:return null;default:return null}}function go(g,b){switch(B(b),b.tag){case 3:Ku(ei),oe();break;case 26:case 27:case 5:de(b);break;case 4:oe();break;case 13:Vt(b);break;case 19:_(qa);break;case 10:Ku(b.type);break;case 22:case 23:Vt(b),Ve(),g!==null&&_(Tf);break;case 24:Ku(ei)}}function Kn(g,b){try{var T=b.updateQueue,z=T!==null?T.lastEffect:null;if(z!==null){var ee=z.next;T=ee;do{if((T.tag&g)===g){z=void 0;var le=T.create,De=T.inst;z=le(),De.destroy=z}T=T.next}while(T!==ee)}}catch(Ie){ln(b,b.return,Ie)}}function _l(g,b,T){try{var z=b.updateQueue,ee=z!==null?z.lastEffect:null;if(ee!==null){var le=ee.next;z=le;do{if((z.tag&g)===g){var De=z.inst,Ie=De.destroy;if(Ie!==void 0){De.destroy=void 0,ee=b;var We=T;try{Ie()}catch(vt){ln(ee,We,vt)}}}z=z.next}while(z!==le)}}catch(vt){ln(b,b.return,vt)}}function Dg(g){var b=g.updateQueue;if(b!==null){var T=g.stateNode;try{en(b,T)}catch(z){ln(g,g.return,z)}}}function Ii(g,b,T){T.props=Ll(g.type,g.memoizedProps),T.state=g.memoizedState;try{T.componentWillUnmount()}catch(z){ln(g,b,z)}}function gu(g,b){try{var T=g.ref;if(T!==null){var z=g.stateNode;switch(g.tag){case 26:case 27:case 5:var ee=Ip(z);break;default:ee=z}typeof T=="function"?g.refCleanup=T(ee):T.current=ee}}catch(le){ln(g,b,le)}}function pr(g,b){var T=g.ref,z=g.refCleanup;if(T!==null)if(typeof z=="function")try{z()}catch(ee){ln(g,b,ee)}finally{g.refCleanup=null,g=g.alternate,g!=null&&(g.refCleanup=null)}else if(typeof T=="function")try{T(null)}catch(ee){ln(g,b,ee)}else T.current=null}function xf(g){var b=g.type,T=g.memoizedProps,z=g.stateNode;try{x4(z,b,T,g)}catch(ee){ln(g,g.return,ee)}}function em(g,b,T){try{Fg(g.stateNode,g.type,T,b,g)}catch(z){ln(g,g.return,z)}}function tm(g){return g.tag===5||g.tag===3||(Yl?g.tag===26:!1)||(_u?g.tag===27:!1)||g.tag===4}function xr(g){e:for(;;){for(;g.sibling===null;){if(g.return===null||tm(g.return))return null;g=g.return}for(g.sibling.return=g.return,g=g.sibling;g.tag!==5&&g.tag!==6&&(!_u||g.tag!==27)&&g.tag!==18;){if(g.flags&2||g.child===null||g.tag===4)continue e;g.child.return=g,g=g.child}if(!(g.flags&2))return g.stateNode}}function nm(g,b,T){var z=g.tag;if(z===5||z===6)g=g.stateNode,b?Py(T,g,b):ky(T,g);else if(!(z===4||_u&&z===27)&&(g=g.child,g!==null))for(nm(g,b,T),g=g.sibling;g!==null;)nm(g,b,T),g=g.sibling}function Sf(g,b,T){var z=g.tag;if(z===5||z===6)g=g.stateNode,b?Ng(T,g,b):Mg(T,g);else if(!(z===4||_u&&z===27)&&(g=g.child,g!==null))for(Sf(g,b,T),g=g.sibling;g!==null;)Sf(g,b,T),g=g.sibling}function Dc(g,b,T){g=g.containerInfo;try{Iy(g,T)}catch(z){ln(b,b.return,z)}}function Xe(g,b){for(ts(g.containerInfo),ya=b;ya!==null;)if(g=ya,b=g.child,(g.subtreeFlags&1028)!==0&&b!==null)b.return=g,ya=b;else for(;ya!==null;){g=ya;var T=g.alternate;switch(b=g.flags,g.tag){case 0:break;case 11:case 15:break;case 1:if((b&1024)!==0&&T!==null){b=void 0;var z=g,ee=T.memoizedProps;T=T.memoizedState;var le=z.stateNode;try{var De=Ll(z.type,ee,z.elementType===z.type);b=le.getSnapshotBeforeUpdate(De,T),le.__reactInternalSnapshotBeforeUpdate=b}catch(Ie){ln(z,z.return,Ie)}}break;case 3:(b&1024)!==0&&nu&&Jn(g.stateNode.containerInfo);break;case 5:case 26:case 27:case 6:case 4:case 17:break;default:if((b&1024)!==0)throw Error(i(163))}if(b=g.sibling,b!==null){b.return=g.return,ya=b;break}ya=g.return}return De=Nh,Nh=!1,De}function dt(g,b,T){var z=T.flags;switch(T.tag){case 0:case 11:case 15:Ql(g,T),z&4&&Kn(5,T);break;case 1:if(Ql(g,T),z&4)if(g=T.stateNode,b===null)try{g.componentDidMount()}catch(Ie){ln(T,T.return,Ie)}else{var ee=Ll(T.type,b.memoizedProps);b=b.memoizedState;try{g.componentDidUpdate(ee,b,g.__reactInternalSnapshotBeforeUpdate)}catch(Ie){ln(T,T.return,Ie)}}z&64&&Dg(T),z&512&&gu(T,T.return);break;case 3:if(Ql(g,T),z&64&&(z=T.updateQueue,z!==null)){if(g=null,T.child!==null)switch(T.child.tag){case 27:case 5:g=Ip(T.child.stateNode);break;case 1:g=T.child.stateNode}try{en(z,g)}catch(Ie){ln(T,T.return,Ie)}}break;case 26:if(Yl){Ql(g,T),z&512&&gu(T,T.return);break}case 27:case 5:Ql(g,T),b===null&&z&4&&xf(T),z&512&&gu(T,T.return);break;case 12:Ql(g,T);break;case 13:Ql(g,T),z&4&&lr(g,T);break;case 22:if(ee=T.memoizedState!==null||lu,!ee){b=b!==null&&b.memoizedState!==null||ua;var le=lu,De=ua;lu=ee,(ua=b)&&!De?yo(g,T,(T.subtreeFlags&8772)!==0):Ql(g,T),lu=le,ua=De}z&512&&(T.memoizedProps.mode==="manual"?gu(T,T.return):pr(T,T.return));break;default:Ql(g,T)}}function rt(g){var b=g.alternate;b!==null&&(g.alternate=null,rt(b)),g.child=null,g.deletions=null,g.sibling=null,g.tag===5&&(b=g.stateNode,b!==null&&um(b)),g.stateNode=null,g.return=null,g.dependencies=null,g.memoizedProps=null,g.memoizedState=null,g.pendingProps=null,g.stateNode=null,g.updateQueue=null}function Bt(g,b,T){for(T=T.child;T!==null;)En(g,b,T),T=T.sibling}function En(g,b,T){if(ol&&typeof ol.onCommitFiberUnmount=="function")try{ol.onCommitFiberUnmount(Gl,T)}catch{}switch(T.tag){case 26:if(Yl){ua||pr(T,b),Bt(g,b,T),T.memoizedState?ul(T.memoizedState):T.stateNode&&Zy(T.stateNode);break}case 27:if(_u){ua||pr(T,b);var z=Pa,ee=Jl;Pa=T.stateNode,Bt(g,b,T),Yy(T.stateNode),Pa=z,Jl=ee;break}case 5:ua||pr(T,b);case 6:if(nu){if(z=Pa,ee=Jl,Pa=null,Bt(g,b,T),Pa=z,Jl=ee,Pa!==null)if(Jl)try{Aa(Pa,T.stateNode)}catch(le){ln(T,b,le)}else try{al(Pa,T.stateNode)}catch(le){ln(T,b,le)}}else Bt(g,b,T);break;case 18:nu&&Pa!==null&&(Jl?fm(Pa,T.stateNode):Uy(Pa,T.stateNode));break;case 4:nu?(z=Pa,ee=Jl,Pa=T.stateNode.containerInfo,Jl=!0,Bt(g,b,T),Pa=z,Jl=ee):(il&&Dc(T.stateNode,T,Ig()),Bt(g,b,T));break;case 0:case 11:case 14:case 15:ua||_l(2,T,b),ua||_l(4,T,b),Bt(g,b,T);break;case 1:ua||(pr(T,b),z=T.stateNode,typeof z.componentWillUnmount=="function"&&Ii(T,b,z)),Bt(g,b,T);break;case 21:Bt(g,b,T);break;case 22:ua||pr(T,b),ua=(z=ua)||T.memoizedState!==null,Bt(g,b,T),ua=z;break;default:Bt(g,b,T)}}function lr(g,b){if(Ra&&b.memoizedState===null&&(g=b.alternate,g!==null&&(g=g.memoizedState,g!==null&&(g=g.dehydrated,g!==null))))try{cm(g)}catch(T){ln(b,b.return,T)}}function Li(g){switch(g.tag){case 13:case 19:var b=g.stateNode;return b===null&&(b=g.stateNode=new jp),b;case 22:return g=g.stateNode,b=g._retryCache,b===null&&(b=g._retryCache=new jp),b;default:throw Error(i(435,g.tag))}}function zr(g,b){var T=Li(g);b.forEach(function(z){var ee=El.bind(null,g,z);T.has(z)||(T.add(z),z.then(ee,ee))})}function Ka(g,b){var T=b.deletions;if(T!==null)for(var z=0;z<T.length;z++){var ee=T[z],le=g,De=b;if(nu){var Ie=De;e:for(;Ie!==null;){switch(Ie.tag){case 27:case 5:Pa=Ie.stateNode,Jl=!1;break e;case 3:Pa=Ie.stateNode.containerInfo,Jl=!0;break e;case 4:Pa=Ie.stateNode.containerInfo,Jl=!0;break e}Ie=Ie.return}if(Pa===null)throw Error(i(160));En(le,De,ee),Pa=null,Jl=!1}else En(le,De,ee);le=ee.alternate,le!==null&&(le.return=null),ee.return=null}if(b.subtreeFlags&13878)for(b=b.child;b!==null;)$c(b,g),b=b.sibling}function $c(g,b){var T=g.alternate,z=g.flags;switch(g.tag){case 0:case 11:case 14:case 15:Ka(b,g),Ci(g),z&4&&(_l(3,g,g.return),Kn(3,g),_l(5,g,g.return));break;case 1:Ka(b,g),Ci(g),z&512&&(ua||T===null||pr(T,T.return)),z&64&&lu&&(g=g.updateQueue,g!==null&&(z=g.callbacks,z!==null&&(T=g.shared.hiddenCallbacks,g.shared.hiddenCallbacks=T===null?z:T.concat(z))));break;case 26:if(Yl){var ee=Cl;Ka(b,g),Ci(g),z&512&&(ua||T===null||pr(T,T.return)),z&4&&(z=T!==null?T.memoizedState:null,b=g.memoizedState,T===null?b===null?g.stateNode===null?g.stateNode=as(ee,g.type,g.memoizedProps,g):jy(ee,g.type,g.stateNode):g.stateNode=Ug(ee,b,g.memoizedProps):z!==b?(z===null?T.stateNode!==null&&Zy(T.stateNode):ul(z),b===null?jy(ee,g.type,g.stateNode):Ug(ee,b,g.memoizedProps)):b===null&&g.stateNode!==null&&em(g,g.memoizedProps,T.memoizedProps));break}case 27:if(_u&&z&4&&g.alternate===null){ee=g.stateNode;var le=g.memoizedProps;try{hm(ee),S4(g.type,le,ee,g)}catch(Pt){ln(g,g.return,Pt)}}case 5:if(Ka(b,g),Ci(g),z&512&&(ua||T===null||pr(T,T.return)),nu){if(g.flags&32){b=g.stateNode;try{$h(b)}catch(Pt){ln(g,g.return,Pt)}}z&4&&g.stateNode!=null&&(b=g.memoizedProps,em(g,b,T!==null?T.memoizedProps:b)),z&1024&&(qc=!0)}break;case 6:if(Ka(b,g),Ci(g),z&4&&nu){if(g.stateNode===null)throw Error(i(162));z=g.memoizedProps,T=T!==null?T.memoizedProps:z,b=g.stateNode;try{om(b,T,z)}catch(Pt){ln(g,g.return,Pt)}}break;case 3:if(Yl?($d(),ee=Cl,Cl=Vg(b.containerInfo),Ka(b,g),Cl=ee):Ka(b,g),Ci(g),z&4){if(nu&&Ra&&T!==null&&T.memoizedState.isDehydrated)try{Us(b.containerInfo)}catch(Pt){ln(g,g.return,Pt)}if(il){z=b.containerInfo,T=b.pendingChildren;try{Iy(z,T)}catch(Pt){ln(g,g.return,Pt)}}}qc&&(qc=!1,md(g));break;case 4:Yl?(T=Cl,Cl=Vg(g.stateNode.containerInfo),Ka(b,g),Ci(g),Cl=T):(Ka(b,g),Ci(g)),z&4&&il&&Dc(g.stateNode,g,g.stateNode.pendingChildren);break;case 12:Ka(b,g),Ci(g);break;case 13:Ka(b,g),Ci(g),g.child.flags&8192&&g.memoizedState!==null!=(T!==null&&T.memoizedState!==null)&&(Uc=Hi()),z&4&&(z=g.updateQueue,z!==null&&(g.updateQueue=null,zr(g,z)));break;case 22:z&512&&(ua||T===null||pr(T,T.return)),ee=g.memoizedState!==null;var De=T!==null&&T.memoizedState!==null,Ie=lu,We=ua;if(lu=Ie||ee,ua=We||De,Ka(b,g),ua=We,lu=Ie,Ci(g),b=g.stateNode,b._current=g,b._visibility&=-3,b._visibility|=b._pendingVisibility&2,z&8192&&(b._visibility=ee?b._visibility&-2:b._visibility|1,ee&&(b=lu||ua,T===null||De||b||vo(g)),nu&&(g.memoizedProps===null||g.memoizedProps.mode!=="manual"))){e:if(T=null,nu)for(b=g;;){if(b.tag===5||Yl&&b.tag===26||_u&&b.tag===27){if(T===null){De=T=b;try{le=De.stateNode,ee?Bp(le):My(De.stateNode,De.memoizedProps)}catch(Pt){ln(De,De.return,Pt)}}}else if(b.tag===6){if(T===null){De=b;try{var vt=De.stateNode;ee?Qp(vt):Vs(vt,De.memoizedProps)}catch(Pt){ln(De,De.return,Pt)}}}else if((b.tag!==22&&b.tag!==23||b.memoizedState===null||b===g)&&b.child!==null){b.child.return=b,b=b.child;continue}if(b===g)break e;for(;b.sibling===null;){if(b.return===null||b.return===g)break e;T===b&&(T=null),b=b.return}T===b&&(T=null),b.sibling.return=b.return,b=b.sibling}}z&4&&(z=g.updateQueue,z!==null&&(T=z.retryQueue,T!==null&&(z.retryQueue=null,zr(g,T))));break;case 19:Ka(b,g),Ci(g),z&4&&(z=g.updateQueue,z!==null&&(g.updateQueue=null,zr(g,z)));break;case 21:break;default:Ka(b,g),Ci(g)}}function Ci(g){var b=g.flags;if(b&2){try{if(nu&&(!_u||g.tag!==27)){e:{for(var T=g.return;T!==null;){if(tm(T)){var z=T;break e}T=T.return}throw Error(i(160))}switch(z.tag){case 27:if(_u){var ee=z.stateNode,le=xr(g);Sf(g,le,ee);break}case 5:var De=z.stateNode;z.flags&32&&($h(De),z.flags&=-33);var Ie=xr(g);Sf(g,Ie,De);break;case 3:case 4:var We=z.stateNode.containerInfo,vt=xr(g);nm(g,vt,We);break;default:throw Error(i(161))}}}catch(Pt){ln(g,g.return,Pt)}g.flags&=-3}b&4096&&(g.flags&=-4097)}function md(g){if(g.subtreeFlags&1024)for(g=g.child;g!==null;){var b=g;md(b),b.tag===5&&b.flags&1024&&Oi(b.stateNode),g=g.sibling}}function Ql(g,b){if(b.subtreeFlags&8772)for(b=b.child;b!==null;)dt(g,b.alternate,b),b=b.sibling}function vo(g){for(g=g.child;g!==null;){var b=g;switch(b.tag){case 0:case 11:case 14:case 15:_l(4,b,b.return),vo(b);break;case 1:pr(b,b.return);var T=b.stateNode;typeof T.componentWillUnmount=="function"&&Ii(b,b.return,T),vo(b);break;case 26:case 27:case 5:pr(b,b.return),vo(b);break;case 22:pr(b,b.return),b.memoizedState===null&&vo(b);break;default:vo(b)}g=g.sibling}}function yo(g,b,T){for(T=T&&(b.subtreeFlags&8772)!==0,b=b.child;b!==null;){var z=b.alternate,ee=g,le=b,De=le.flags;switch(le.tag){case 0:case 11:case 15:yo(ee,le,T),Kn(4,le);break;case 1:if(yo(ee,le,T),z=le,ee=z.stateNode,typeof ee.componentDidMount=="function")try{ee.componentDidMount()}catch(vt){ln(z,z.return,vt)}if(z=le,ee=z.updateQueue,ee!==null){var Ie=z.stateNode;try{var We=ee.shared.hiddenCallbacks;if(We!==null)for(ee.shared.hiddenCallbacks=null,ee=0;ee<We.length;ee++)Qt(We[ee],Ie)}catch(vt){ln(z,z.return,vt)}}T&&De&64&&Dg(le),gu(le,le.return);break;case 26:case 27:case 5:yo(ee,le,T),T&&z===null&&De&4&&xf(le),gu(le,le.return);break;case 12:yo(ee,le,T);break;case 13:yo(ee,le,T),T&&De&4&&lr(ee,le);break;case 22:le.memoizedState===null&&yo(ee,le,T),gu(le,le.return);break;default:yo(ee,le,T)}b=b.sibling}}function _f(g,b){var T=null;g!==null&&g.memoizedState!==null&&g.memoizedState.cachePool!==null&&(T=g.memoizedState.cachePool.pool),g=null,b.memoizedState!==null&&b.memoizedState.cachePool!==null&&(g=b.memoizedState.cachePool.pool),g!==T&&(g!=null&&g.refCount++,T!=null&&dd(T))}function Sr(g,b){g=null,b.alternate!==null&&(g=b.alternate.memoizedState.cache),b=b.memoizedState.cache,b!==g&&(b.refCount++,g!=null&&dd(g))}function Yo(g,b,T,z){if(b.subtreeFlags&10256)for(b=b.child;b!==null;)bh(g,b,T,z),b=b.sibling}function bh(g,b,T,z){var ee=b.flags;switch(b.tag){case 0:case 11:case 15:Yo(g,b,T,z),ee&2048&&Kn(9,b);break;case 3:Yo(g,b,T,z),ee&2048&&(g=null,b.alternate!==null&&(g=b.alternate.memoizedState.cache),b=b.memoizedState.cache,b!==g&&(b.refCount++,g!=null&&dd(g)));break;case 12:if(ee&2048){Yo(g,b,T,z),g=b.stateNode;try{var le=b.memoizedProps,De=le.id,Ie=le.onPostCommit;typeof Ie=="function"&&Ie(De,b.alternate===null?"mount":"update",g.passiveEffectDuration,-0)}catch(We){ln(b,b.return,We)}}else Yo(g,b,T,z);break;case 23:break;case 22:le=b.stateNode,b.memoizedState!==null?le._visibility&4?Yo(g,b,T,z):Go(g,b):le._visibility&4?Yo(g,b,T,z):(le._visibility|=4,pa(g,b,T,z,(b.subtreeFlags&10256)!==0)),ee&2048&&_f(b.alternate,b);break;case 24:Yo(g,b,T,z),ee&2048&&Sr(b.alternate,b);break;default:Yo(g,b,T,z)}}function pa(g,b,T,z,ee){for(ee=ee&&(b.subtreeFlags&10256)!==0,b=b.child;b!==null;){var le=g,De=b,Ie=T,We=z,vt=De.flags;switch(De.tag){case 0:case 11:case 15:pa(le,De,Ie,We,ee),Kn(8,De);break;case 23:break;case 22:var Pt=De.stateNode;De.memoizedState!==null?Pt._visibility&4?pa(le,De,Ie,We,ee):Go(le,De):(Pt._visibility|=4,pa(le,De,Ie,We,ee)),ee&&vt&2048&&_f(De.alternate,De);break;case 24:pa(le,De,Ie,We,ee),ee&&vt&2048&&Sr(De.alternate,De);break;default:pa(le,De,Ie,We,ee)}b=b.sibling}}function Go(g,b){if(b.subtreeFlags&10256)for(b=b.child;b!==null;){var T=g,z=b,ee=z.flags;switch(z.tag){case 22:Go(T,z),ee&2048&&_f(z.alternate,z);break;case 24:Go(T,z),ee&2048&&Sr(z.alternate,z);break;default:Go(T,z)}b=b.sibling}}function wa(g){if(g.subtreeFlags&Ih)for(g=g.child;g!==null;)bo(g),g=g.sibling}function bo(g){switch(g.tag){case 26:wa(g),g.flags&Ih&&(g.memoizedState!==null?Xy(Cl,g.memoizedState,g.memoizedProps):ma(g.type,g.memoizedProps));break;case 5:wa(g),g.flags&Ih&&ma(g.type,g.memoizedProps);break;case 3:case 4:if(Yl){var b=Cl;Cl=Vg(g.stateNode.containerInfo),wa(g),Cl=b}else wa(g);break;case 22:g.memoizedState===null&&(b=g.alternate,b!==null&&b.memoizedState!==null?(b=Ih,Ih=16777216,wa(g),Ih=b):wa(g));break;default:wa(g)}}function Is(g){var b=g.alternate;if(b!==null&&(g=b.child,g!==null)){b.child=null;do b=g.sibling,g.sibling=null,g=b;while(g!==null)}}function Wa(g){var b=g.deletions;if((g.flags&16)!==0){if(b!==null)for(var T=0;T<b.length;T++){var z=b[T];ya=z,el(z,g)}Is(g)}if(g.subtreeFlags&10256)for(g=g.child;g!==null;)Ju(g),g=g.sibling}function Ju(g){switch(g.tag){case 0:case 11:case 15:Wa(g),g.flags&2048&&_l(9,g,g.return);break;case 3:Wa(g);break;case 12:Wa(g);break;case 22:var b=g.stateNode;g.memoizedState!==null&&b._visibility&4&&(g.return===null||g.return.tag!==13)?(b._visibility&=-5,Cc(g)):Wa(g);break;default:Wa(g)}}function Cc(g){var b=g.deletions;if((g.flags&16)!==0){if(b!==null)for(var T=0;T<b.length;T++){var z=b[T];ya=z,el(z,g)}Is(g)}for(g=g.child;g!==null;){switch(b=g,b.tag){case 0:case 11:case 15:_l(8,b,b.return),Cc(b);break;case 22:T=b.stateNode,T._visibility&4&&(T._visibility&=-5,Cc(b));break;default:Cc(b)}g=g.sibling}}function el(g,b){for(;ya!==null;){var T=ya;switch(T.tag){case 0:case 11:case 15:_l(8,T,b);break;case 23:case 22:if(T.memoizedState!==null&&T.memoizedState.cachePool!==null){var z=T.memoizedState.cachePool.pool;z!=null&&z.refCount++}break;case 24:dd(T.memoizedState.cache)}if(z=T.child,z!==null)z.return=T,ya=z;else e:for(T=g;ya!==null;){z=ya;var ee=z.sibling,le=z.return;if(rt(z),z===T){ya=null;break e}if(ee!==null){ee.return=le,ya=ee;break e}ya=le}}}function wc(g){var b=Ag(g);if(b!=null){if(typeof b.memoizedProps["data-testname"]!="string")throw Error(i(364));return b}if(g=Dh(g),g===null)throw Error(i(362));return g.stateNode.current}function Rc(g,b){var T=g.tag;switch(b.$$typeof){case Sm:if(g.type===b.value)return!0;break;case ev:e:{for(b=b.value,g=[g,0],T=0;T<g.length;){var z=g[T++],ee=z.tag,le=g[T++],De=b[le];if(ee!==5&&ee!==26&&ee!==27||!is(z)){for(;De!=null&&Rc(z,De);)le++,De=b[le];if(le===b.length){b=!0;break e}else for(z=z.child;z!==null;)g.push(z,le),z=z.sibling}}b=!1}return b;case tv:if((T===5||T===26||T===27)&&O4(g.stateNode,b.value))return!0;break;case li:if((T===5||T===6||T===26||T===27)&&(g=Lp(g),g!==null&&0<=g.indexOf(b.value)))return!0;break;case Nr:if((T===5||T===26||T===27)&&(g=g.memoizedProps["data-testname"],typeof g=="string"&&g.toLowerCase()===b.value.toLowerCase()))return!0;break;default:throw Error(i(365))}return!1}function vu(g){switch(g.$$typeof){case Sm:return"<"+(u(g.value)||"Unknown")+">";case ev:return":has("+(vu(g)||"")+")";case tv:return'[role="'+g.value+'"]';case li:return'"'+g.value+'"';case Nr:return'[data-testname="'+g.value+'"]';default:throw Error(i(365))}}function gd(g,b){var T=[];g=[g,0];for(var z=0;z<g.length;){var ee=g[z++],le=ee.tag,De=g[z++],Ie=b[De];if(le!==5&&le!==26&&le!==27||!is(ee)){for(;Ie!=null&&Rc(ee,Ie);)De++,Ie=b[De];if(De===b.length)T.push(ee);else for(ee=ee.child;ee!==null;)g.push(ee,De),ee=ee.sibling}}return T}function aa(g,b){if(!Fc)throw Error(i(363));g=wc(g),g=gd(g,b),b=[],g=Array.from(g);for(var T=0;T<g.length;){var z=g[T++],ee=z.tag;if(ee===5||ee===26||ee===27)is(z)||b.push(z.stateNode);else for(z=z.child;z!==null;)g.push(z),z=z.sibling}return b}function Ji(){if((vn&2)!==0&&cr!==0)return cr&-cr;if(xn.T!==null){var g=Kl;return g!==0?g:ze()}return rs()}function tl(){eo===0&&(eo=(cr&536870912)===0||Vr?F():536870912);var g=Co.current;return g!==null&&(g.flags|=32),eo}function yu(g,b,T){(g===mr&&Si===2||g.cancelPendingCommit!==null)&&(ql(g,0),bu(g,cr,eo,!1)),q(g,T),((vn&2)===0||g!==mr)&&(g===mr&&((vn&2)===0&&(Gs|=T),Ir===4&&bu(g,cr,eo,!1)),Oe(g))}function Oh(g,b,T){if((vn&6)!==0)throw Error(i(327));var z=!T&&(b&60)===0&&(b&g.expiredLanes)===0||w(g,b),ee=z?ue(g,b):Nu(g,b,!0),le=z;do{if(ee===0){Ro&&!z&&bu(g,b,0,!1);break}else if(ee===6)bu(g,b,0,!ti);else{if(T=g.current.alternate,le&&!xh(T)){ee=Nu(g,b,!1),le=!1;continue}if(ee===2){if(le=b,g.errorRecoveryDisabledLanes&le)var De=0;else De=g.pendingLanes&-536870913,De=De!==0?De:De&536870912?536870912:0;if(De!==0){b=De;e:{var Ie=g;ee=Zp;var We=Ra&&Ie.current.memoizedState.isDehydrated;if(We&&(ql(Ie,De).flags|=256),De=Nu(Ie,De,!1),De!==2){if(kf&&!We){Ie.errorRecoveryDisabledLanes|=le,Gs|=le,ee=4;break e}le=hl,hl=ee,le!==null&&zl(le)}ee=De}if(le=!1,ee!==2)continue}}if(ee===1){ql(g,0),bu(g,b,0,!0);break}e:{switch(z=g,ee){case 0:case 1:throw Error(i(345));case 4:if((b&4194176)===b){bu(z,b,eo,!ti);break e}break;case 2:hl=null;break;case 3:case 5:break;default:throw Error(i(329))}if(z.finishedWork=T,z.finishedLanes=b,(b&62914560)===b&&(le=Uc+300-Hi(),10<le)){if(bu(z,b,eo,!ti),A(z,0)!==0)break e;z.timeoutHandle=eu(Ef.bind(null,z,T,hl,Ks,Hp,b,eo,Gs,Vc,ti,2,-0,0),le);break e}Ef(z,T,hl,Ks,Hp,b,eo,Gs,Vc,ti,0,-0,0)}}break}while(!0);Oe(g)}function zl(g){hl===null?hl=g:hl.push.apply(hl,g)}function Ef(g,b,T,z,ee,le,De,Ie,We,vt,Pt,jt,Wt){var Hn=b.subtreeFlags;if((Hn&8192||(Hn&16785408)===16785408)&&(Zl(),bo(b),b=Tg(),b!==null)){g.cancelPendingCommit=b(Be.bind(null,g,T,z,ee,De,Ie,We,1,jt,Wt)),bu(g,le,De,!vt);return}Be(g,T,z,ee,De,Ie,We,Pt,jt,Wt)}function xh(g){for(var b=g;;){var T=b.tag;if((T===0||T===11||T===15)&&b.flags&16384&&(T=b.updateQueue,T!==null&&(T=T.stores,T!==null)))for(var z=0;z<T.length;z++){var ee=T[z],le=ee.getSnapshot;ee=ee.value;try{if(!Dl(le(),ee))return!1}catch{return!1}}if(T=b.child,b.subtreeFlags&16384&&T!==null)T.return=b,b=T;else{if(b===g)break;for(;b.sibling===null;){if(b.return===null||b.return===g)return!0;b=b.return}b.sibling.return=b.return,b=b.sibling}}return!0}function bu(g,b,T,z){b&=~dl,b&=~Gs,g.suspendedLanes|=b,g.pingedLanes&=~b,z&&(g.warmLanes|=b),z=g.expirationTimes;for(var ee=b;0<ee;){var le=31-ll(ee),De=1<<le;z[le]=-1,ee&=~De}T!==0&&N(g,T,b)}function Pp(){return(vn&6)===0?($e(0),!1):!0}function Mu(){if(Yn!==null){if(Si===0)var g=Yn.return;else g=Yn,Ur=Nd=null,$i(g),Hs=null,$o=0,g=Yn;for(;g!==null;)go(g.alternate,g),g=g.return;Yn=null}}function ql(g,b){g.finishedWork=null,g.finishedLanes=0;var T=g.timeoutHandle;T!==_d&&(g.timeoutHandle=_d,Eh(T)),T=g.cancelPendingCommit,T!==null&&(g.cancelPendingCommit=null,T()),Mu(),mr=g,Yn=T=Ja(g.current,null),cr=b,Si=0,la=null,ti=!1,Ro=w(g,b),kf=!1,Vc=eo=dl=Gs=Ao=Ir=0,hl=Zp=null,Hp=!1,(b&8)!==0&&(b|=b&32);var z=g.entangledLanes;if(z!==0)for(g=g.entanglements,z&=b;0<z;){var ee=31-ll(z),le=1<<ee;b|=g[ee],z&=~le}return Ys=b,ve(),T}function Ac(g,b){nr=null,xn.H=Fr,b===Wr?(b=Or(),Si=3):b===Gg?(b=Or(),Si=4):Si=b===Y1?8:b!==null&&typeof b=="object"&&typeof b.then=="function"?6:1,la=b,Yn===null&&(Ir=1,ud(g,H(b,g.current)))}function Bi(){var g=Co.current;return g===null?!0:(cr&4194176)===cr?wo===null:(cr&62914560)===cr||(cr&536870912)!==0?g===wo:!1}function wi(){var g=xn.H;return xn.H=Fr,g===null?Fr:g}function Fu(){var g=xn.A;return xn.A=Jg,g}function dn(){Ir=4,ti||(cr&4194176)!==cr&&Co.current!==null||(Ro=!0),(Ao&134217727)===0&&(Gs&134217727)===0||mr===null||bu(mr,cr,eo,!1)}function Nu(g,b,T){var z=vn;vn|=2;var ee=wi(),le=Fu();(mr!==g||cr!==b)&&(Ks=null,ql(g,b)),b=!1;var De=Ir;e:do try{if(Si!==0&&Yn!==null){var Ie=Yn,We=la;switch(Si){case 8:Mu(),De=6;break e;case 3:case 2:case 6:Co.current===null&&(b=!0);var vt=Si;if(Si=0,la=null,xe(g,Ie,We,vt),T&&Ro){De=0;break e}break;default:vt=Si,Si=0,la=null,xe(g,Ie,We,vt)}}Ko(),De=Ir;break}catch(Pt){Ac(g,Pt)}while(!0);return b&&g.shellSuspendCounter++,Ur=Nd=null,vn=z,xn.H=ee,xn.A=le,Yn===null&&(mr=null,cr=0,ve()),De}function Ko(){for(;Yn!==null;)re(Yn)}function ue(g,b){var T=vn;vn|=2;var z=wi(),ee=Fu();mr!==g||cr!==b?(Ks=null,oa=Hi()+500,ql(g,b)):Ro=w(g,b);e:do try{if(Si!==0&&Yn!==null){b=Yn;var le=la;t:switch(Si){case 1:Si=0,la=null,xe(g,b,le,1);break;case 2:if(rn(le)){Si=0,la=null,fe(b);break}b=function(){Si===2&&mr===g&&(Si=7),Oe(g)},le.then(b,b);break e;case 3:Si=7;break e;case 4:Si=5;break e;case 7:rn(le)?(Si=0,la=null,fe(b)):(Si=0,la=null,xe(g,b,le,7));break;case 5:var De=null;switch(Yn.tag){case 26:De=Yn.memoizedState;case 5:case 27:var Ie=Yn,We=Ie.type,vt=Ie.pendingProps;if(De?Zg(De):xo(We,vt)){Si=0,la=null;var Pt=Ie.sibling;if(Pt!==null)Yn=Pt;else{var jt=Ie.return;jt!==null?(Yn=jt,ke(jt)):Yn=null}break t}}Si=0,la=null,xe(g,b,le,5);break;case 6:Si=0,la=null,xe(g,b,le,6);break;case 8:Mu(),Ir=6;break e;default:throw Error(i(462))}}L();break}catch(Wt){Ac(g,Wt)}while(!0);return Ur=Nd=null,xn.H=z,xn.A=ee,vn=T,Yn!==null?0:(mr=null,cr=0,ve(),Ir)}function L(){for(;Yn!==null&&!H1();)re(Yn)}function re(g){var b=K0(g.alternate,g,Ys);g.memoizedProps=g.pendingProps,b===null?ke(g):Yn=b}function fe(g){var b=g,T=b.alternate;switch(b.tag){case 15:case 0:b=cd(T,b,b.pendingProps,b.type,void 0,cr);break;case 11:b=cd(T,b,b.pendingProps,b.type.render,b.ref,cr);break;case 5:$i(b);default:go(T,b),b=Yn=Ls(b,Ys),b=K0(T,b,Ys)}g.memoizedProps=g.pendingProps,b===null?ke(g):Yn=b}function xe(g,b,T,z){Ur=Nd=null,$i(b),Hs=null,$o=0;var ee=b.return;try{if(Ep(g,ee,b,T,cr)){Ir=1,ud(g,H(T,g.current)),Yn=null;return}}catch(le){if(ee!==null)throw Yn=ee,le;Ir=1,ud(g,H(T,g.current)),Yn=null;return}b.flags&32768?(Vr||z===1?g=!0:Ro||(cr&536870912)!==0?g=!1:(ti=g=!0,(z===2||z===3||z===6)&&(z=Co.current,z!==null&&z.tag===13&&(z.flags|=16384))),Qe(b,g)):ke(b)}function ke(g){var b=g;do{if((b.flags&32768)!==0){Qe(b,ti);return}g=b.return;var T=yh(b.alternate,b,Ys);if(T!==null){Yn=T;return}if(b=b.sibling,b!==null){Yn=b;return}Yn=b=g}while(b!==null);Ir===0&&(Ir=5)}function Qe(g,b){do{var T=kp(g.alternate,g);if(T!==null){T.flags&=32767,Yn=T;return}if(T=g.return,T!==null&&(T.flags|=32768,T.subtreeFlags=0,T.deletions=null),!b&&(g=g.sibling,g!==null)){Yn=g;return}Yn=g=T}while(g!==null);Ir=6,Yn=null}function Be(g,b,T,z,ee,le,De,Ie,We,vt){var Pt=xn.T,jt=ru();try{xu(2),xn.T=null,Ze(g,b,T,z,jt,ee,le,De,Ie,We,vt)}finally{xn.T=Pt,xu(jt)}}function Ze(g,b,T,z,ee,le,De,Ie){do Kt();while(Pf!==null);if((vn&6)!==0)throw Error(i(327));var We=g.finishedWork;if(z=g.finishedLanes,We===null)return null;if(g.finishedWork=null,g.finishedLanes=0,We===g.current)throw Error(i(177));g.callbackNode=null,g.callbackPriority=0,g.cancelPendingCommit=null;var vt=We.lanes|We.childLanes;if(vt|=Vp,V(g,z,vt,le,De,Ie),g===mr&&(Yn=mr=null,cr=0),(We.subtreeFlags&10256)===0&&(We.flags&10256)===0||cs||(cs=!0,nv=vt,Bh=T,Vl(Xg,function(){return Kt(),null})),T=(We.flags&15990)!==0,(We.subtreeFlags&15990)!==0||T?(T=xn.T,xn.T=null,le=ru(),xu(2),De=vn,vn|=4,Xe(g,We),$c(We,g),Pc(g.containerInfo),g.current=We,dt(g,We.alternate,We),Gy(),vn=De,xu(le),xn.T=T):g.current=We,cs?(cs=!1,Pf=g,Lh=z):Rt(g,vt),vt=g.pendingLanes,vt===0&&(To=null),K(We.stateNode),Oe(g),b!==null)for(ee=g.onRecoverableError,We=0;We<b.length;We++)vt=b[We],ee(vt.value,{componentStack:vt.stack});return(Lh&3)!==0&&Kt(),vt=g.pendingLanes,(z&4194218)!==0&&(vt&42)!==0?g===wl?Qh++:(Qh=0,wl=g):Qh=0,$e(0),null}function Rt(g,b){(g.pooledCacheLanes&=b)===0&&(b=g.pooledCache,b!=null&&(g.pooledCache=null,dd(b)))}function Kt(){if(Pf!==null){var g=Pf,b=nv;nv=0;var T=W(Lh),z=32>T?32:T;T=xn.T;var ee=ru();try{if(xu(z),xn.T=null,Pf===null)var le=!1;else{z=Bh,Bh=null;var De=Pf,Ie=Lh;if(Pf=null,Lh=0,(vn&6)!==0)throw Error(i(331));var We=vn;if(vn|=4,Ju(De.current),bh(De,De.current,Ie,z),vn=We,$e(0,!1),ol&&typeof ol.onPostCommitFiberRoot=="function")try{ol.onPostCommitFiberRoot(Gl,De)}catch{}le=!0}return le}finally{xu(ee),xn.T=T,Rt(g,b)}}return!1}function zn(g,b,T){b=H(T,b),b=ks(g.stateNode,b,2),g=Xt(g,b,2),g!==null&&(q(g,2),Oe(g))}function ln(g,b,T){if(g.tag===3)zn(g,g,T);else for(;b!==null;){if(b.tag===3){zn(b,g,T);break}else if(b.tag===1){var z=b.stateNode;if(typeof b.type.getDerivedStateFromError=="function"||typeof z.componentDidCatch=="function"&&(To===null||!To.has(z))){g=H(T,g),T=Bl(2),z=Xt(b,T,2),z!==null&&(ld(T,z,b,g),q(z,2),Oe(z));break}}b=b.return}}function mn(g,b,T){var z=g.pingCache;if(z===null){z=g.pingCache=new _r;var ee=new Set;z.set(b,ee)}else ee=z.get(b),ee===void 0&&(ee=new Set,z.set(b,ee));ee.has(T)||(kf=!0,ee.add(T),g=Ri.bind(null,g,b,T),b.then(g,g))}function Ri(g,b,T){var z=g.pingCache;z!==null&&z.delete(b),g.pingedLanes|=g.suspendedLanes&T,g.warmLanes&=~T,mr===g&&(cr&T)===T&&(Ir===4||Ir===3&&(cr&62914560)===cr&&300>Hi()-Uc?(vn&2)===0&&ql(g,0):dl|=T,Vc===cr&&(Vc=0)),Oe(g)}function za(g,b){b===0&&(b=I()),g=Re(g,b),g!==null&&(q(g,b),Oe(g))}function Se(g){var b=g.memoizedState,T=0;b!==null&&(T=b.retryLane),za(g,T)}function El(g,b){var T=0;switch(g.tag){case 13:var z=g.stateNode,ee=g.memoizedState;ee!==null&&(T=ee.retryLane);break;case 19:z=g.stateNode;break;case 22:z=g.stateNode._retryCache;break;default:throw Error(i(314))}z!==null&&z.delete(b),za(g,T)}function Vl(g,b){return Ah(g,b)}function Iu(g,b,T,z){this.tag=g,this.key=T,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=b,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=z,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function qn(g){return g=g.prototype,!(!g||!g.isReactComponent)}function Ja(g,b){var T=g.alternate;return T===null?(T=n(g.tag,b,g.key,g.mode),T.elementType=g.elementType,T.type=g.type,T.stateNode=g.stateNode,T.alternate=g,g.alternate=T):(T.pendingProps=b,T.type=g.type,T.flags=0,T.subtreeFlags=0,T.deletions=null),T.flags=g.flags&31457280,T.childLanes=g.childLanes,T.lanes=g.lanes,T.child=g.child,T.memoizedProps=g.memoizedProps,T.memoizedState=g.memoizedState,T.updateQueue=g.updateQueue,b=g.dependencies,T.dependencies=b===null?null:{lanes:b.lanes,firstContext:b.firstContext},T.sibling=g.sibling,T.index=g.index,T.ref=g.ref,T.refCleanup=g.refCleanup,T}function Ls(g,b){g.flags&=31457282;var T=g.alternate;return T===null?(g.childLanes=0,g.lanes=b,g.child=null,g.subtreeFlags=0,g.memoizedProps=null,g.memoizedState=null,g.updateQueue=null,g.dependencies=null,g.stateNode=null):(g.childLanes=T.childLanes,g.lanes=T.lanes,g.child=T.child,g.subtreeFlags=0,g.deletions=null,g.memoizedProps=T.memoizedProps,g.memoizedState=T.memoizedState,g.updateQueue=T.updateQueue,g.type=T.type,b=T.dependencies,g.dependencies=b===null?null:{lanes:b.lanes,firstContext:b.firstContext}),g}function Ul(g,b,T,z,ee,le){var De=0;if(z=g,typeof g=="function")qn(g)&&(De=1);else if(typeof g=="string")De=Yl&&_u?qg(g,T,qr.current)?26:wh(g)?27:5:Yl?qg(g,T,qr.current)?26:5:_u&&wh(g)?27:5;else e:switch(g){case Wo:return Lu(T.children,ee,le,b);case Ry:De=8,ee|=24;break;case Ay:return g=n(12,T,b,ee|2),g.elementType=Ay,g.lanes=le,g;case $g:return g=n(13,T,b,ee),g.elementType=$g,g.lanes=le,g;case bd:return g=n(19,T,b,ee),g.elementType=bd,g.lanes=le,g;case rm:return Tc(T,ee,le,b);default:if(typeof g=="object"&&g!==null)switch(g.$$typeof){case Jo:case kc:De=10;break e;case yd:De=9;break e;case Mp:De=11;break e;case Fp:De=14;break e;case es:De=16,z=null;break e}De=29,T=Error(i(130,g===null?"null":typeof g,"")),z=null}return b=n(De,T,b,ee),b.elementType=g,b.type=z,b.lanes=le,b}function Lu(g,b,T,z){return g=n(7,g,z,b),g.lanes=T,g}function Tc(g,b,T,z){g=n(22,g,z,b),g.elementType=rm,g.lanes=T;var ee={_visibility:1,_pendingVisibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null,_current:null,detach:function(){var le=ee._current;if(le===null)throw Error(i(456));if((ee._pendingVisibility&2)===0){var De=Re(le,2);De!==null&&(ee._pendingVisibility|=2,yu(De,le,2))}},attach:function(){var le=ee._current;if(le===null)throw Error(i(456));if((ee._pendingVisibility&2)!==0){var De=Re(le,2);De!==null&&(ee._pendingVisibility&=-3,yu(De,le,2))}}};return g.stateNode=ee,g}function jl(g,b,T){return g=n(6,g,null,b),g.lanes=T,g}function vd(g,b,T){return b=n(4,g.children!==null?g.children:[],g.key,b),b.lanes=T,b.stateNode={containerInfo:g.containerInfo,pendingChildren:null,implementation:g.implementation},b}function Sh(g,b,T,z,ee,le,De,Ie){this.tag=1,this.containerInfo=g,this.finishedWork=this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=_d,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=U(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.finishedLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=U(0),this.hiddenUpdates=U(null),this.identifierPrefix=z,this.onUncaughtError=ee,this.onCaughtError=le,this.onRecoverableError=De,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=Ie,this.incompleteTransitions=new Map}function Ou(g,b,T,z,ee,le,De,Ie,We,vt,Pt,jt){return g=new Sh(g,b,T,De,Ie,We,vt,jt),b=1,le===!0&&(b|=24),le=n(3,null,null,b),g.current=le,le.stateNode=g,b=fd(),b.refCount++,g.pooledCache=b,b.refCount++,le.memoizedState={element:z,isDehydrated:T,cache:b},bt(le),g}function Oo(g){return g?(g=Lc,g):Lc}function $y(g){var b=g._reactInternals;if(b===void 0)throw typeof g.render=="function"?Error(i(188)):(g=Object.keys(g).join(","),Error(i(268,g)));return g=m(b),g=g!==null?y(g):null,g===null?null:Ip(g.stateNode)}function I1(g,b,T,z,ee,le){ee=Oo(ee),z.context===null?z.context=ee:z.pendingContext=ee,z=qt(b),z.payload={element:T},le=le===void 0?null:le,le!==null&&(z.callback=le),T=Xt(g,z,b),T!==null&&(yu(T,g,b),nn(T,g,b))}function L1(g,b){if(g=g.memoizedState,g!==null&&g.dehydrated!==null){var T=g.retryLane;g.retryLane=T!==0&&T<b?T:b}}function Cy(g,b){L1(g,b),(g=g.alternate)&&L1(g,b)}var gn={},Bs=ir(),Bu=k1e(),_h=Object.assign,wy=Symbol.for("react.element"),Df=Symbol.for("react.transitional.element"),Qs=Symbol.for("react.portal"),Wo=Symbol.for("react.fragment"),Ry=Symbol.for("react.strict_mode"),Ay=Symbol.for("react.profiler"),Jo=Symbol.for("react.provider"),yd=Symbol.for("react.consumer"),kc=Symbol.for("react.context"),Mp=Symbol.for("react.forward_ref"),$g=Symbol.for("react.suspense"),bd=Symbol.for("react.suspense_list"),Fp=Symbol.for("react.memo"),es=Symbol.for("react.lazy"),rm=Symbol.for("react.offscreen"),B1=Symbol.for("react.memo_cache_sentinel"),im=Symbol.iterator,zs=Symbol.for("react.client.reference"),xn=Bs.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,Cg,wg,Od=!1,nl=Array.isArray,Np=t.rendererVersion,qs=t.rendererPackageName,rl=t.extraDevToolsConfig,Ip=t.getPublicInstance,xd=t.getRootHostContext,Ty=t.getChildHostContext,ts=t.prepareForCommit,Pc=t.resetAfterCommit,Rg=t.createInstance,Mc=t.appendInitialChild,ns=t.finalizeInitialChildren,Sd=t.shouldSetTextContent,am=t.createTextInstance,eu=t.scheduleTimeout,Eh=t.cancelTimeout,_d=t.noTimeout,tu=t.isPrimaryRenderer;t.warnsIfNotActing;var nu=t.supportsMutation,il=t.supportsPersistence,Ra=t.supportsHydration,Ag=t.getInstanceFromNode;t.beforeActiveInstanceBlur,t.afterActiveInstanceBlur;var Qu=t.preparePortalMount;t.prepareScopeUpdate,t.getInstanceFromScope;var xu=t.setCurrentUpdatePriority,ru=t.getCurrentUpdatePriority,rs=t.resolveUpdatePriority;t.resolveEventType,t.resolveEventTimeStamp;var Ed=t.shouldAttemptEagerTransition,um=t.detachDeletedInstance;t.requestPostPaintCallback;var Q1=t.maySuspendCommit,xo=t.preloadInstance,Zl=t.startSuspendingCommit,ma=t.suspendInstance,Tg=t.waitForCommitToBeReady,Su=t.NotPendingTransition,An=t.HostTransitionContext,Oi=t.resetFormInstance;t.bindToConsole;var ga=t.supportsMicrotasks,lm=t.scheduleMicrotask,Fc=t.supportsTestSelectors,Dh=t.findFiberRoot,kg=t.getBoundingRect,Lp=t.getTextContent,is=t.isHiddenSubtree,O4=t.matchAccessibilityRole,ea=t.setFocusIfFocusable,Pg=t.setupIntersectionObserver,Mg=t.appendChild,ky=t.appendChildToContainer,om=t.commitTextUpdate,x4=t.commitMount,Fg=t.commitUpdate,Ng=t.insertBefore,Py=t.insertInContainerBefore,al=t.removeChild,Aa=t.removeChildFromContainer,$h=t.resetTextContent,Bp=t.hideInstance,Qp=t.hideTextInstance,My=t.unhideInstance,Vs=t.unhideTextInstance,Jn=t.clearContainer,Fy=t.cloneInstance,Ig=t.createContainerChildSet,Ch=t.appendChildToContainerChildSet,Ny=t.finalizeContainerChildren,Iy=t.replaceContainerChildren,Ly=t.cloneHiddenInstance,Hl=t.cloneHiddenTextInstance,Lg=t.isSuspenseInstancePending,Dd=t.isSuspenseInstanceFallback,Bg=t.getSuspenseInstanceFallbackErrorDetails,z1=t.registerSuspenseInstanceRetry,By=t.canHydrateFormStateMarker,Xl=t.isFormStateMarkerMatching,Nc=t.getNextHydratableSibling,zp=t.getFirstHydratableChild,Qy=t.getFirstHydratableChildWithinContainer,sm=t.getFirstHydratableChildWithinSuspenseInstance,zy=t.canHydrateInstance,Qg=t.canHydrateTextInstance,qy=t.canHydrateSuspenseInstance,Vy=t.hydrateInstance,zg=t.hydrateTextInstance,So=t.hydrateSuspenseInstance,q1=t.getNextHydratableInstanceAfterSuspenseInstance,Us=t.commitHydratedContainer,cm=t.commitHydratedSuspenseInstance,Uy=t.clearSuspenseBoundary,fm=t.clearSuspenseBoundaryFromContainer,V1=t.shouldDeleteUnhydratedTailInstances;t.diffHydratedPropsForDevWarnings,t.diffHydratedTextForDevWarnings,t.describeHydratableInstanceForDevWarnings;var U1=t.validateHydratableInstance,j1=t.validateHydratableTextInstance,Yl=t.supportsResources,qg=t.isHostHoistableType,Vg=t.getHoistableRoot,dm=t.getResource,Ug=t.acquireResource,ul=t.releaseResource,as=t.hydrateHoistable,jy=t.mountHoistable,Zy=t.unmountHoistable,jg=t.createHoistableInstance,$d=t.prepareToCommitHoistables,Hy=t.mayResourceSuspendCommit,Zg=t.preloadResource,Xy=t.suspendResource,_u=t.supportsSingletons,Cd=t.resolveSingletonInstance,hm=t.clearSingleton,S4=t.acquireSingletonInstance,Yy=t.releaseSingletonInstance,wh=t.isHostSingletonType,wd=[],Ic=-1,Lc={},ll=Math.clz32?Math.clz32:$,Rd=Math.log,Z1=Math.LN2,Rh=128,Ad=4194304,Ah=Bu.unstable_scheduleCallback,Th=Bu.unstable_cancelCallback,H1=Bu.unstable_shouldYield,Gy=Bu.unstable_requestPaint,Hi=Bu.unstable_now,Hg=Bu.unstable_ImmediatePriority,Ky=Bu.unstable_UserBlockingPriority,Xg=Bu.unstable_NormalPriority,pm=Bu.unstable_IdlePriority,X1=Bu.log,Wy=Bu.unstable_setDisableYieldValue,Gl=null,ol=null,Dl=typeof Object.is=="function"?Object.is:te,qp=new WeakMap,xi=[],Ai=0,On=null,Td=0,$l=[],iu=0,js=null,us=1,_o="",qr=S(null),mm=S(null),Zs=S(null),kh=S(null),sl=null,Eu=null,Vr=!1,Eo=null,ls=!1,Do=Error(i(519)),Ti=[],Bc=0,Vp=0,kd=null,Ph=null,gm=!1,$f=!1,os=!1,Cf=0,vm=null,Yg=0,Kl=0,Pd=null,Qc=!1,Mh=!1,ym=Object.prototype.hasOwnProperty,Wr=Error(i(460)),Gg=Error(i(474)),Md={then:function(){}},wf=null,Hs=null,$o=0,Xs=ai(!0),Kg=ai(!1),au=S(null),uu=S(0),Co=S(null),wo=null,qa=S(0),zc=0,nr=null,Jr=null,Ta=null,ss=!1,Xi=!1,Rf=!1,bm=0,va=0,Up=null,Jy=0,Om=function(){return{lastEffect:null,events:null,stores:null,memoCache:null}},Fr={readContext:Qa,use:Zi,useCallback:Dt,useContext:Dt,useEffect:Dt,useImperativeHandle:Dt,useLayoutEffect:Dt,useInsertionEffect:Dt,useMemo:Dt,useReducer:Dt,useRef:Dt,useState:Dt,useDebugValue:Dt,useDeferredValue:Dt,useTransition:Dt,useSyncExternalStore:Dt,useId:Dt};Fr.useCacheRefresh=Dt,Fr.useMemoCache=Dt,Fr.useHostTransitionStatus=Dt,Fr.useFormState=Dt,Fr.useActionState=Dt,Fr.useOptimistic=Dt;var Va={readContext:Qa,use:Zi,useCallback:function(g,b){return hr().memoizedState=[g,b===void 0?null:b],g},useContext:Qa,useEffect:j0,useImperativeHandle:function(g,b,T){T=T!=null?T.concat([g]):null,Cs(4194308,4,yf.bind(null,b,g),T)},useLayoutEffect:function(g,b){return Cs(4194308,4,g,b)},useInsertionEffect:function(g,b){Cs(4,2,g,b)},useMemo:function(g,b){var T=hr();b=b===void 0?null:b;var z=g();if(Rf){J(!0);try{g()}finally{J(!1)}}return T.memoizedState=[z,b],z},useReducer:function(g,b,T){var z=hr();if(T!==void 0){var ee=T(b);if(Rf){J(!0);try{T(b)}finally{J(!1)}}}else ee=b;return z.memoizedState=z.baseState=ee,g={pending:null,lanes:0,dispatch:null,lastRenderedReducer:g,lastRenderedState:ee},z.queue=g,g=g.dispatch=Oc.bind(null,nr,g),[z.memoizedState,g]},useRef:function(g){var b=hr();return g={current:g},b.memoizedState=g},useState:function(g){g=tr(g);var b=g.queue,T=hh.bind(null,nr,b);return b.dispatch=T,[g.memoizedState,T]},useDebugValue:ch,useDeferredValue:function(g,b){var T=hr();return bc(T,g,b)},useTransition:function(){var g=tr(!1);return g=xp.bind(null,nr,g.queue,!0,!1),hr().memoizedState=g,[!1,g]},useSyncExternalStore:function(g,b,T){var z=nr,ee=hr();if(Vr){if(T===void 0)throw Error(i(407));T=T()}else{if(T=b(),mr===null)throw Error(i(349));(cr&60)!==0||gc(z,b,T)}ee.memoizedState=T;var le={value:T,getSnapshot:b};return ee.queue=le,j0(ur.bind(null,z,le,g),[g]),z.flags|=2048,Vo(9,xl.bind(null,z,le,T,b),{destroy:void 0},null),T},useId:function(){var g=hr(),b=mr.identifierPrefix;if(Vr){var T=_o,z=us;T=(z&~(1<<32-ll(z)-1)).toString(32)+T,b=":"+b+"R"+T,T=bm++,0<T&&(b+="H"+T.toString(32)),b+=":"}else T=Jy++,b=":"+b+"r"+T.toString(32)+":";return g.memoizedState=b},useCacheRefresh:function(){return hr().memoizedState=dh.bind(null,nr)}};Va.useMemoCache=yi,Va.useHostTransitionStatus=As,Va.useFormState=rd,Va.useActionState=rd,Va.useOptimistic=function(g){var b=hr();b.memoizedState=b.baseState=g;var T={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return b.queue=T,b=Qn.bind(null,nr,!0,T),T.dispatch=b,[g,b]};var cl={readContext:Qa,use:Zi,useCallback:fh,useContext:Qa,useEffect:vf,useImperativeHandle:ho,useInsertionEffect:Z0,useLayoutEffect:Op,useMemo:Rs,useReducer:Xa,useRef:Pu,useState:function(){return Xa(Ca)},useDebugValue:ch,useDeferredValue:function(g,b){var T=wt();return H0(T,Jr.memoizedState,g,b)},useTransition:function(){var g=Xa(Ca)[0],b=wt().memoizedState;return[typeof g=="boolean"?g:Kr(g),b]},useSyncExternalStore:Nl,useId:Uo};cl.useCacheRefresh=_p,cl.useMemoCache=yi,cl.useHostTransitionStatus=As,cl.useFormState=$s,cl.useActionState=$s,cl.useOptimistic=function(g,b){var T=wt();return Bn(T,Jr,g,b)};var Fd={readContext:Qa,use:Zi,useCallback:fh,useContext:Qa,useEffect:vf,useImperativeHandle:ho,useInsertionEffect:Z0,useLayoutEffect:Op,useMemo:Rs,useReducer:Tu,useRef:Pu,useState:function(){return Tu(Ca)},useDebugValue:ch,useDeferredValue:function(g,b){var T=wt();return Jr===null?bc(T,g,b):H0(T,Jr.memoizedState,g,b)},useTransition:function(){var g=Tu(Ca)[0],b=wt().memoizedState;return[typeof g=="boolean"?g:Kr(g),b]},useSyncExternalStore:Nl,useId:Uo};Fd.useCacheRefresh=_p,Fd.useMemoCache=yi,Fd.useHostTransitionStatus=As,Fd.useFormState=yc,Fd.useActionState=yc,Fd.useOptimistic=function(g,b){var T=wt();return Jr!==null?Bn(T,Jr,g,b):(T.baseState=g,[g,T.queue.dispatch])};var xm={isMounted:function(g){return(g=g._reactInternals)?d(g)===g:!1},enqueueSetState:function(g,b,T){g=g._reactInternals;var z=Ji(),ee=qt(z);ee.payload=b,T!=null&&(ee.callback=T),b=Xt(g,ee,z),b!==null&&(yu(b,g,z),nn(b,g,z))},enqueueReplaceState:function(g,b,T){g=g._reactInternals;var z=Ji(),ee=qt(z);ee.tag=1,ee.payload=b,T!=null&&(ee.callback=T),b=Xt(g,ee,z),b!==null&&(yu(b,g,z),nn(b,g,z))},enqueueForceUpdate:function(g,b){g=g._reactInternals;var T=Ji(),z=qt(T);z.tag=2,b!=null&&(z.callback=b),b=Xt(g,z,T),b!==null&&(yu(b,g,T),nn(b,g,T))}},eb=typeof reportError=="function"?reportError:function(g){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var b=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof g=="object"&&g!==null&&typeof g.message=="string"?String(g.message):String(g),error:g});if(!window.dispatchEvent(b))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",g);return}console.error(g)},Y1=Error(i(461)),ka=!1,fl={dehydrated:null,treeContext:null,retryLane:0},Wl=S(null),Nd=null,Ur=null,G1=typeof AbortController<"u"?AbortController:function(){var g=[],b=this.signal={aborted:!1,addEventListener:function(T,z){g.push(z)}};this.abort=function(){b.aborted=!0,g.forEach(function(T){return T()})}},Af=Bu.unstable_scheduleCallback,Fh=Bu.unstable_NormalPriority,ei={$$typeof:kc,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0},Wg=xn.S;xn.S=function(g,b){typeof b=="object"&&b!==null&&typeof b.then=="function"&&at(g,b),Wg!==null&&Wg(g,b)};var Tf=S(null),lu=!1,ua=!1,qc=!1,jp=typeof WeakSet=="function"?WeakSet:Set,ya=null,Nh=!1,Pa=null,Jl=!1,Cl=null,Ih=8192,Jg={getCacheForType:function(g){var b=Qa(ei),T=b.data.get(g);return T===void 0&&(T=g(),b.data.set(g,T)),T}},Sm=0,ev=1,tv=2,Nr=3,li=4;if(typeof Symbol=="function"&&Symbol.for){var rr=Symbol.for;Sm=rr("selector.component"),ev=rr("selector.has_pseudo_class"),tv=rr("selector.role"),Nr=rr("selector.test_id"),li=rr("selector.text")}var _r=typeof WeakMap=="function"?WeakMap:Map,vn=0,mr=null,Yn=null,cr=0,Si=0,la=null,ti=!1,Ro=!1,kf=!1,Ys=0,Ir=0,Ao=0,Gs=0,dl=0,eo=0,Vc=0,Zp=null,hl=null,Hp=!1,Uc=0,oa=1/0,Ks=null,To=null,cs=!1,Pf=null,Lh=0,nv=0,Bh=null,Qh=0,wl=null;return gn.attemptContinuousHydration=function(g){if(g.tag===13){var b=Re(g,67108864);b!==null&&yu(b,g,67108864),Cy(g,67108864)}},gn.attemptHydrationAtCurrentPriority=function(g){if(g.tag===13){var b=Ji(),T=Re(g,b);T!==null&&yu(T,g,b),Cy(g,b)}},gn.attemptSynchronousHydration=function(g){switch(g.tag){case 3:if(g=g.stateNode,g.current.memoizedState.isDehydrated){var b=k(g.pendingLanes);if(b!==0){for(g.pendingLanes|=2,g.entangledLanes|=2;b;){var T=1<<31-ll(b);g.entanglements[1]|=T,b&=~T}Oe(g),(vn&6)===0&&(oa=Hi()+500,$e(0))}}break;case 13:b=Re(g,2),b!==null&&yu(b,g,2),Pp(),Cy(g,2)}},gn.batchedUpdates=function(g,b){return g(b)},gn.createComponentSelector=function(g){return{$$typeof:Sm,value:g}},gn.createContainer=function(g,b,T,z,ee,le,De,Ie,We,vt){return Ou(g,b,!1,null,T,z,le,De,Ie,We,vt,null)},gn.createHasPseudoClassSelector=function(g){return{$$typeof:ev,value:g}},gn.createHydrationContainer=function(g,b,T,z,ee,le,De,Ie,We,vt,Pt,jt,Wt){return g=Ou(T,z,!0,g,ee,le,Ie,We,vt,Pt,jt,Wt),g.context=Oo(null),T=g.current,z=Ji(),ee=qt(z),ee.callback=b??null,Xt(T,ee,z),g.current.lanes=z,q(g,z),Oe(g),g},gn.createPortal=function(g,b,T){var z=3<arguments.length&&arguments[3]!==void 0?arguments[3]:null;return{$$typeof:Qs,key:z==null?null:""+z,children:g,containerInfo:b,implementation:T}},gn.createRoleSelector=function(g){return{$$typeof:tv,value:g}},gn.createTestNameSelector=function(g){return{$$typeof:Nr,value:g}},gn.createTextSelector=function(g){return{$$typeof:li,value:g}},gn.defaultOnCaughtError=function(g){console.error(g)},gn.defaultOnRecoverableError=function(g){eb(g)},gn.defaultOnUncaughtError=function(g){eb(g)},gn.deferredUpdates=function(g){var b=xn.T,T=ru();try{return xu(32),xn.T=null,g()}finally{xu(T),xn.T=b}},gn.discreteUpdates=function(g,b,T,z,ee){var le=xn.T,De=ru();try{return xu(2),xn.T=null,g(b,T,z,ee)}finally{xu(De),xn.T=le,vn===0&&(oa=Hi()+500)}},gn.findAllNodes=aa,gn.findBoundingRects=function(g,b){if(!Fc)throw Error(i(363));b=aa(g,b),g=[];for(var T=0;T<b.length;T++)g.push(kg(b[T]));for(b=g.length-1;0<b;b--){T=g[b];for(var z=T.x,ee=z+T.width,le=T.y,De=le+T.height,Ie=b-1;0<=Ie;Ie--)if(b!==Ie){var We=g[Ie],vt=We.x,Pt=vt+We.width,jt=We.y,Wt=jt+We.height;if(z>=vt&&le>=jt&&ee<=Pt&&De<=Wt){g.splice(b,1);break}else if(z!==vt||T.width!==We.width||Wt<le||jt>De){if(!(le!==jt||T.height!==We.height||Pt<z||vt>ee)){vt>z&&(We.width+=vt-z,We.x=z),Pt<ee&&(We.width=ee-vt),g.splice(b,1);break}}else{jt>le&&(We.height+=jt-le,We.y=le),Wt<De&&(We.height=De-jt),g.splice(b,1);break}}}return g},gn.findHostInstance=$y,gn.findHostInstanceWithNoPortals=function(g){return g=m(g),g=g!==null?v(g):null,g===null?null:Ip(g.stateNode)},gn.findHostInstanceWithWarning=function(g){return $y(g)},gn.flushPassiveEffects=Kt,gn.flushSyncFromReconciler=function(g){var b=vn;vn|=1;var T=xn.T,z=ru();try{if(xu(2),xn.T=null,g)return g()}finally{xu(z),xn.T=T,vn=b,(vn&6)===0&&$e(0)}},gn.flushSyncWork=Pp,gn.focusWithin=function(g,b){if(!Fc)throw Error(i(363));for(g=wc(g),b=gd(g,b),b=Array.from(b),g=0;g<b.length;){var T=b[g++],z=T.tag;if(!is(T)){if((z===5||z===26||z===27)&&ea(T.stateNode))return!0;for(T=T.child;T!==null;)b.push(T),T=T.sibling}}return!1},gn.getFindAllNodesFailureDescription=function(g,b){if(!Fc)throw Error(i(363));var T=0,z=[];g=[wc(g),0];for(var ee=0;ee<g.length;){var le=g[ee++],De=le.tag,Ie=g[ee++],We=b[Ie];if((De!==5&&De!==26&&De!==27||!is(le))&&(Rc(le,We)&&(z.push(vu(We)),Ie++,Ie>T&&(T=Ie)),Ie<b.length))for(le=le.child;le!==null;)g.push(le,Ie),le=le.sibling}if(T<b.length){for(g=[];T<b.length;T++)g.push(vu(b[T]));return`findAllNodes was able to match part of the selector:
  `+(z.join(" > ")+`

No matching component was found for:
  `)+g.join(" > ")}return null},gn.getPublicRootInstance=function(g){if(g=g.current,!g.child)return null;switch(g.child.tag){case 27:case 5:return Ip(g.child.stateNode);default:return g.child.stateNode}},gn.injectIntoDevTools=function(){var g={bundleType:0,version:Np,rendererPackageName:qs,currentDispatcherRef:xn,findFiberByHostInstance:Ag,reconcilerVersion:"19.0.0"};if(rl!==null&&(g.rendererConfig=rl),typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u")g=!1;else{var b=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(b.isDisabled||!b.supportsFiber)g=!0;else{try{Gl=b.inject(g),ol=b}catch{}g=!!b.checkDCE}}return g},gn.isAlreadyRendering=function(){return!1},gn.observeVisibleRects=function(g,b,T,z){if(!Fc)throw Error(i(363));g=aa(g,b);var ee=Pg(g,T,z).disconnect;return{disconnect:function(){ee()}}},gn.shouldError=function(){return null},gn.shouldSuspend=function(){return!1},gn.startHostTransition=function(g,b,T,z){if(g.tag!==5)throw Error(i(476));var ee=Sp(g).queue;xp(g,ee,b,Su,T===null?r:function(){var le=Sp(g).next.queue;return Mr(g,le,{},Ji()),T(z)})},gn.updateContainer=function(g,b,T,z){var ee=b.current,le=Ji();return I1(ee,le,g,b,T,z),le},gn.updateContainerSync=function(g,b,T,z){return b.tag===0&&Kt(),I1(b.current,2,g,b,T,z),2},gn},e.exports.default=e.exports,Object.defineProperty(e.exports,"__esModule",{value:!0})})(Q7)),Q7.exports}var ZW;function Dtt(){return ZW||(ZW=1,B7.exports=Ett()),B7.exports}var $tt=Dtt();const Ctt=da($tt);var HW=k1e();function kB(e,t,n){if(!e)return;if(n(e)===!0)return e;let r=t?e.return:e.child;for(;r;){const i=kB(r,t,n);if(i)return i;r=t?null:r.sibling}}function P1e(e){try{return Object.defineProperties(e,{_currentRenderer:{get(){return null},set(){}},_currentRenderer2:{get(){return null},set(){}}})}catch{return e}}const PB=P1e(P.createContext(null));class M1e extends P.Component{render(){return P.createElement(PB.Provider,{value:this._reactInternals},this.props.children)}}function F1e(){const e=P.useContext(PB);if(e===null)throw new Error("its-fine: useFiber must be called within a <FiberProvider />!");const t=P.useId();return P.useMemo(()=>{for(const n of[e,e==null?void 0:e.alternate]){if(!n)continue;const r=kB(n,!1,i=>{let a=i.memoizedState;for(;a;){if(a.memoizedState===t)return!0;a=a.next}});if(r)return r}},[e,t])}const wtt=Symbol.for("react.context"),Rtt=e=>e!==null&&typeof e=="object"&&"$$typeof"in e&&e.$$typeof===wtt;function Att(){const e=F1e(),[t]=P.useState(()=>new Map);t.clear();let n=e;for(;n;){const r=n.type;Rtt(r)&&r!==PB&&!t.has(r)&&t.set(r,P.use(P1e(r))),n=n.return}return t}function Ttt(){const e=Att();return P.useMemo(()=>Array.from(e.keys()).reduce((t,n)=>r=>P.createElement(t,null,P.createElement(n.Provider,{...r,value:e.get(n)})),t=>P.createElement(M1e,{...t})),[e])}function N1e(e){let t=e.root;for(;t.getState().previousRoot;)t=t.getState().previousRoot;return t}const I1e=e=>e&&e.isOrthographicCamera,ktt=e=>e&&e.hasOwnProperty("current"),Ptt=e=>e!=null&&(typeof e=="string"||typeof e=="number"||e.isColor),HS=((e,t)=>typeof window<"u"&&(((e=window.document)==null?void 0:e.createElement)||((t=window.navigator)==null?void 0:t.product)==="ReactNative"))()?P.useLayoutEffect:P.useEffect;function L1e(e){const t=P.useRef(e);return HS(()=>void(t.current=e),[e]),t}function Mtt(){const e=F1e(),t=Ttt();return P.useMemo(()=>({children:n})=>{const i=!!kB(e,!0,a=>a.type===P.StrictMode)?P.StrictMode:P.Fragment;return Qr.jsx(i,{children:Qr.jsx(t,{children:n})})},[e,t])}function Ftt({set:e}){return HS(()=>(e(new Promise(()=>null)),()=>e(!1)),[e]),null}const Ntt=(e=>(e=class extends P.Component{constructor(...n){super(...n),this.state={error:!1}}componentDidCatch(n){this.props.set(n)}render(){return this.state.error?null:this.props.children}},e.getDerivedStateFromError=()=>({error:!0}),e))();function B1e(e){var t;const n=typeof window<"u"?(t=window.devicePixelRatio)!=null?t:2:1;return Array.isArray(e)?Math.min(Math.max(e[0],n),e[1]):e}function PO(e){var t;return(t=e.__r3f)==null?void 0:t.root.getState()}const su={obj:e=>e===Object(e)&&!su.arr(e)&&typeof e!="function",fun:e=>typeof e=="function",str:e=>typeof e=="string",num:e=>typeof e=="number",boo:e=>typeof e=="boolean",und:e=>e===void 0,nul:e=>e===null,arr:e=>Array.isArray(e),equ(e,t,{arrays:n="shallow",objects:r="reference",strict:i=!0}={}){if(typeof e!=typeof t||!!e!=!!t)return!1;if(su.str(e)||su.num(e)||su.boo(e))return e===t;const a=su.obj(e);if(a&&r==="reference")return e===t;const u=su.arr(e);if(u&&n==="reference")return e===t;if((u||a)&&e===t)return!0;let l;for(l in e)if(!(l in t))return!1;if(a&&n==="shallow"&&r==="shallow"){for(l in i?t:e)if(!su.equ(e[l],t[l],{strict:i,objects:"reference"}))return!1}else for(l in i?t:e)if(e[l]!==t[l])return!1;if(su.und(l)){if(u&&e.length===0&&t.length===0||a&&Object.keys(e).length===0&&Object.keys(t).length===0)return!0;if(e!==t)return!1}return!0}};function Itt(e){e.type!=="Scene"&&(e.dispose==null||e.dispose());for(const t in e){const n=e[t];(n==null?void 0:n.type)!=="Scene"&&(n==null||n.dispose==null||n.dispose())}}const Q1e=["children","key","ref"];function Ltt(e){const t={};for(const n in e)Q1e.includes(n)||(t[n]=e[n]);return t}function gC(e,t,n,r){const i=e;let a=i==null?void 0:i.__r3f;return a||(a={root:t,type:n,parent:null,children:[],props:Ltt(r),object:i,eventCount:0,handlers:{},isHidden:!1},i&&(i.__r3f=a)),a}function aS(e,t){let n=e[t];if(!t.includes("-"))return{root:e,key:t,target:n};n=e;for(const i of t.split("-")){var r;t=i,e=n,n=(r=n)==null?void 0:r[t]}return{root:e,key:t,target:n}}const XW=/-\d+$/;function vC(e,t){if(su.str(t.props.attach)){if(XW.test(t.props.attach)){const i=t.props.attach.replace(XW,""),{root:a,key:u}=aS(e.object,i);Array.isArray(a[u])||(a[u]=[])}const{root:n,key:r}=aS(e.object,t.props.attach);t.previousAttach=n[r],n[r]=t.object}else su.fun(t.props.attach)&&(t.previousAttach=t.props.attach(e.object,t.object))}function yC(e,t){if(su.str(t.props.attach)){const{root:n,key:r}=aS(e.object,t.props.attach),i=t.previousAttach;i===void 0?delete n[r]:n[r]=i}else t.previousAttach==null||t.previousAttach(e.object,t.object);delete t.previousAttach}const wF=[...Q1e,"args","dispose","attach","object","onUpdate","dispose"],YW=new Map;function Btt(e){let t=YW.get(e.constructor);try{t||(t=new e.constructor,YW.set(e.constructor,t))}catch{}return t}function Qtt(e,t){const n={};for(const r in t)if(!wF.includes(r)&&!su.equ(t[r],e.props[r])){n[r]=t[r];for(const i in t)i.startsWith(`${r}-`)&&(n[i]=t[i])}for(const r in e.props){if(wF.includes(r)||t.hasOwnProperty(r))continue;const{root:i,key:a}=aS(e.object,r);if(i.constructor&&i.constructor.length===0){const u=Btt(i);su.und(u)||(n[a]=u[a])}else n[a]=0}return n}const ztt=["map","emissiveMap","sheenColorMap","specularColorMap","envMap"],qtt=/^on(Pointer|Click|DoubleClick|ContextMenu|Wheel)/;function Pv(e,t){var n;const r=e.__r3f,i=r&&N1e(r).getState(),a=r==null?void 0:r.eventCount;for(const l in t){let o=t[l];if(wF.includes(l))continue;if(r&&qtt.test(l)){typeof o=="function"?r.handlers[l]=o:delete r.handlers[l],r.eventCount=Object.keys(r.handlers).length;continue}if(o===void 0)continue;let{root:s,key:c,target:d}=aS(e,l);if(d instanceof QB&&o instanceof QB)d.mask=o.mask;else if(d instanceof OC&&Ptt(o))d.set(o);else if(d!==null&&typeof d=="object"&&typeof d.set=="function"&&typeof d.copy=="function"&&o!=null&&o.constructor&&d.constructor===o.constructor)d.copy(o);else if(d!==null&&typeof d=="object"&&typeof d.set=="function"&&Array.isArray(o))typeof d.fromArray=="function"?d.fromArray(o):d.set(...o);else if(d!==null&&typeof d=="object"&&typeof d.set=="function"&&typeof o=="number")typeof d.setScalar=="function"?d.setScalar(o):d.set(o);else{var u;s[c]=o,i&&!i.linear&&ztt.includes(c)&&(u=s[c])!=null&&u.isTexture&&s[c].format===OOe&&s[c].type===xOe&&(s[c].colorSpace=hJ)}}if(r!=null&&r.parent&&i!=null&&i.internal&&(n=r.object)!=null&&n.isObject3D&&a!==r.eventCount){const l=r.object,o=i.internal.interaction.indexOf(l);o>-1&&i.internal.interaction.splice(o,1),r.eventCount&&l.raycast!==null&&i.internal.interaction.push(l)}return r&&r.props.attach===void 0&&(r.object.isBufferGeometry?r.props.attach="geometry":r.object.isMaterial&&(r.props.attach="material")),r&&b4(r),e}function b4(e){var t;if(!e.parent)return;e.props.onUpdate==null||e.props.onUpdate(e.object);const n=(t=e.root)==null||t.getState==null?void 0:t.getState();n&&n.internal.frames===0&&n.invalidate()}function Vtt(e,t){e.manual||(I1e(e)?(e.left=t.width/-2,e.right=t.width/2,e.top=t.height/2,e.bottom=t.height/-2):e.aspect=t.width/t.height,e.updateProjectionMatrix())}const tf=e=>e==null?void 0:e.isObject3D;function QE(e){return(e.eventObject||e.object).uuid+"/"+e.index+e.instanceId}function z1e(e,t,n,r){const i=n.get(t);i&&(n.delete(t),n.size===0&&(e.delete(r),i.target.releasePointerCapture(r)))}function Utt(e,t){const{internal:n}=e.getState();n.interaction=n.interaction.filter(r=>r!==t),n.initialHits=n.initialHits.filter(r=>r!==t),n.hovered.forEach((r,i)=>{(r.eventObject===t||r.object===t)&&n.hovered.delete(i)}),n.capturedMap.forEach((r,i)=>{z1e(n.capturedMap,t,r,i)})}function jtt(e){function t(o){const{internal:s}=e.getState(),c=o.offsetX-s.initialClick[0],d=o.offsetY-s.initialClick[1];return Math.round(Math.sqrt(c*c+d*d))}function n(o){return o.filter(s=>["Move","Over","Enter","Out","Leave"].some(c=>{var d;return(d=s.__r3f)==null?void 0:d.handlers["onPointer"+c]}))}function r(o,s){const c=e.getState(),d=new Set,h=[],m=s?s(c.internal.interaction):c.internal.interaction;for(let _=0;_<m.length;_++){const E=PO(m[_]);E&&(E.raycaster.camera=void 0)}c.previousRoot||c.events.compute==null||c.events.compute(o,c);function y(_){const E=PO(_);if(!E||!E.events.enabled||E.raycaster.camera===null)return[];if(E.raycaster.camera===void 0){var $;E.events.compute==null||E.events.compute(o,E,($=E.previousRoot)==null?void 0:$.getState()),E.raycaster.camera===void 0&&(E.raycaster.camera=null)}return E.raycaster.camera?E.raycaster.intersectObject(_,!0):[]}let v=m.flatMap(y).sort((_,E)=>{const $=PO(_.object),k=PO(E.object);return!$||!k?_.distance-E.distance:k.events.priority-$.events.priority||_.distance-E.distance}).filter(_=>{const E=QE(_);return d.has(E)?!1:(d.add(E),!0)});c.events.filter&&(v=c.events.filter(v,c));for(const _ of v){let E=_.object;for(;E;){var S;(S=E.__r3f)!=null&&S.eventCount&&h.push({..._,eventObject:E}),E=E.parent}}if("pointerId"in o&&c.internal.capturedMap.has(o.pointerId))for(let _ of c.internal.capturedMap.get(o.pointerId).values())d.has(QE(_.intersection))||h.push(_.intersection);return h}function i(o,s,c,d){if(o.length){const h={stopped:!1};for(const m of o){let y=PO(m.object);if(y||m.object.traverseAncestors(v=>{const S=PO(v);if(S)return y=S,!1}),y){const{raycaster:v,pointer:S,camera:_,internal:E}=y,$=new ci(S.x,S.y,0).unproject(_),k=I=>{var U,q;return(U=(q=E.capturedMap.get(I))==null?void 0:q.has(m.eventObject))!=null?U:!1},A=I=>{const U={intersection:m,target:s.target};E.capturedMap.has(I)?E.capturedMap.get(I).set(m.eventObject,U):E.capturedMap.set(I,new Map([[m.eventObject,U]])),s.target.setPointerCapture(I)},w=I=>{const U=E.capturedMap.get(I);U&&z1e(E.capturedMap,m.eventObject,U,I)};let R={};for(let I in s){let U=s[I];typeof U!="function"&&(R[I]=U)}let F={...m,...R,pointer:S,intersections:o,stopped:h.stopped,delta:c,unprojectedPoint:$,ray:v.ray,camera:_,stopPropagation(){const I="pointerId"in s&&E.capturedMap.get(s.pointerId);if((!I||I.has(m.eventObject))&&(F.stopped=h.stopped=!0,E.hovered.size&&Array.from(E.hovered.values()).find(U=>U.eventObject===m.eventObject))){const U=o.slice(0,o.indexOf(m));a([...U,m])}},target:{hasPointerCapture:k,setPointerCapture:A,releasePointerCapture:w},currentTarget:{hasPointerCapture:k,setPointerCapture:A,releasePointerCapture:w},nativeEvent:s};if(d(F),h.stopped===!0)break}}}return o}function a(o){const{internal:s}=e.getState();for(const c of s.hovered.values())if(!o.length||!o.find(d=>d.object===c.object&&d.index===c.index&&d.instanceId===c.instanceId)){const h=c.eventObject.__r3f;if(s.hovered.delete(QE(c)),h!=null&&h.eventCount){const m=h.handlers,y={...c,intersections:o};m.onPointerOut==null||m.onPointerOut(y),m.onPointerLeave==null||m.onPointerLeave(y)}}}function u(o,s){for(let c=0;c<s.length;c++){const d=s[c].__r3f;d==null||d.handlers.onPointerMissed==null||d.handlers.onPointerMissed(o)}}function l(o){switch(o){case"onPointerLeave":case"onPointerCancel":return()=>a([]);case"onLostPointerCapture":return s=>{const{internal:c}=e.getState();"pointerId"in s&&c.capturedMap.has(s.pointerId)&&requestAnimationFrame(()=>{c.capturedMap.has(s.pointerId)&&(c.capturedMap.delete(s.pointerId),a([]))})}}return function(c){const{onPointerMissed:d,internal:h}=e.getState();h.lastEvent.current=c;const m=o==="onPointerMove",y=o==="onClick"||o==="onContextMenu"||o==="onDoubleClick",S=r(c,m?n:void 0),_=y?t(c):0;o==="onPointerDown"&&(h.initialClick=[c.offsetX,c.offsetY],h.initialHits=S.map($=>$.eventObject)),y&&!S.length&&_<=2&&(u(c,h.interaction),d&&d(c)),m&&a(S);function E($){const k=$.eventObject,A=k.__r3f;if(!(A!=null&&A.eventCount))return;const w=A.handlers;if(m){if(w.onPointerOver||w.onPointerEnter||w.onPointerOut||w.onPointerLeave){const R=QE($),F=h.hovered.get(R);F?F.stopped&&$.stopPropagation():(h.hovered.set(R,$),w.onPointerOver==null||w.onPointerOver($),w.onPointerEnter==null||w.onPointerEnter($))}w.onPointerMove==null||w.onPointerMove($)}else{const R=w[o];R?(!y||h.initialHits.includes(k))&&(u(c,h.interaction.filter(F=>!h.initialHits.includes(F))),R($)):y&&h.initialHits.includes(k)&&u(c,h.interaction.filter(F=>!h.initialHits.includes(F)))}}i(S,c,_,E)}}return{handlePointer:l}}const GW=e=>!!(e!=null&&e.render),q1e=P.createContext(null),Ztt=(e,t)=>{const n=Stt((l,o)=>{const s=new ci,c=new ci,d=new ci;function h(_=o().camera,E=c,$=o().size){const{width:k,height:A,top:w,left:R}=$,F=k/A;E.isVector3?d.copy(E):d.set(...E);const I=_.getWorldPosition(s).distanceTo(d);if(I1e(_))return{width:k/_.zoom,height:A/_.zoom,top:w,left:R,factor:1,distance:I,aspect:F};{const U=_.fov*Math.PI/180,q=2*Math.tan(U/2)*I,V=q*(k/A);return{width:V,height:q,top:w,left:R,factor:k/V,distance:I,aspect:F}}}let m;const y=_=>l(E=>({performance:{...E.performance,current:_}})),v=new ao;return{set:l,get:o,gl:null,camera:null,raycaster:null,events:{priority:1,enabled:!0,connected:!1},scene:null,xr:null,invalidate:(_=1)=>e(o(),_),advance:(_,E)=>t(_,E,o()),legacy:!1,linear:!1,flat:!1,controls:null,clock:new SOe,pointer:v,mouse:v,frameloop:"always",onPointerMissed:void 0,performance:{current:1,min:.5,max:1,debounce:200,regress:()=>{const _=o();m&&clearTimeout(m),_.performance.current!==_.performance.min&&y(_.performance.min),m=setTimeout(()=>y(o().performance.max),_.performance.debounce)}},size:{width:0,height:0,top:0,left:0},viewport:{initialDpr:0,dpr:0,width:0,height:0,top:0,left:0,aspect:0,distance:0,factor:0,getCurrentViewport:h},setEvents:_=>l(E=>({...E,events:{...E.events,..._}})),setSize:(_,E,$=0,k=0)=>{const A=o().camera,w={width:_,height:E,top:$,left:k};l(R=>({size:w,viewport:{...R.viewport,...h(A,c,w)}}))},setDpr:_=>l(E=>{const $=B1e(_);return{viewport:{...E.viewport,dpr:$,initialDpr:E.viewport.initialDpr||$}}}),setFrameloop:(_="always")=>{const E=o().clock;E.stop(),E.elapsedTime=0,_!=="never"&&(E.start(),E.elapsedTime=0),l(()=>({frameloop:_}))},previousRoot:void 0,internal:{interaction:[],hovered:new Map,subscribers:[],initialClick:[0,0],initialHits:[],capturedMap:new Map,lastEvent:P.createRef(),active:!1,frames:0,priority:0,subscribe:(_,E,$)=>{const k=o().internal;return k.priority=k.priority+(E>0?1:0),k.subscribers.push({ref:_,priority:E,store:$}),k.subscribers=k.subscribers.sort((A,w)=>A.priority-w.priority),()=>{const A=o().internal;A!=null&&A.subscribers&&(A.priority=A.priority-(E>0?1:0),A.subscribers=A.subscribers.filter(w=>w.ref!==_))}}}}}),r=n.getState();let i=r.size,a=r.viewport.dpr,u=r.camera;return n.subscribe(()=>{const{camera:l,size:o,viewport:s,gl:c,set:d}=n.getState();if(o.width!==i.width||o.height!==i.height||s.dpr!==a){i=o,a=s.dpr,Vtt(l,o),s.dpr>0&&c.setPixelRatio(s.dpr);const h=typeof HTMLCanvasElement<"u"&&c.domElement instanceof HTMLCanvasElement;c.setSize(o.width,o.height,h)}l!==u&&(u=l,d(h=>({viewport:{...h.viewport,...h.viewport.getCurrentViewport(l)}})))}),n.subscribe(l=>e(l)),n};function V1e(){const e=P.useContext(q1e);if(!e)throw new Error("R3F: Hooks can only be used within the Canvas component!");return e}function Vd(e=n=>n,t){return V1e()(e,t)}function x5(e,t=0){const n=V1e(),r=n.getState().internal.subscribe,i=L1e(e);return HS(()=>r(i,t,n),[t,r,n]),null}function Htt(e){const t=Ctt(e);return t.injectIntoDevTools({bundleType:0,rendererPackageName:"@react-three/fiber",version:P.version}),t}const U1e=0,U3={},Xtt=/^three(?=[A-Z])/,S5=e=>`${e[0].toUpperCase()}${e.slice(1)}`;let Ytt=0;const Gtt=e=>typeof e=="function";function Ktt(e){if(Gtt(e)){const t=`${Ytt++}`;return U3[t]=e,t}else Object.assign(U3,e)}function j1e(e,t){const n=S5(e),r=U3[n];if(e!=="primitive"&&!r)throw new Error(`R3F: ${n} is not part of the THREE namespace! Did you forget to extend? See: https://docs.pmnd.rs/react-three-fiber/api/objects#using-3rd-party-objects-declaratively`);if(e==="primitive"&&!t.object)throw new Error("R3F: Primitives without 'object' are invalid!");if(t.args!==void 0&&!Array.isArray(t.args))throw new Error("R3F: The args prop must be an array!")}function Wtt(e,t,n){var r;return e=S5(e)in U3?e:e.replace(Xtt,""),j1e(e,t),e==="primitive"&&(r=t.object)!=null&&r.__r3f&&delete t.object.__r3f,gC(t.object,n,e,t)}function Jtt(e){if(!e.isHidden){var t;e.props.attach&&(t=e.parent)!=null&&t.object?yC(e.parent,e):tf(e.object)&&(e.object.visible=!1),e.isHidden=!0,b4(e)}}function Z1e(e){if(e.isHidden){var t;e.props.attach&&(t=e.parent)!=null&&t.object?vC(e.parent,e):tf(e.object)&&e.props.visible!==!1&&(e.object.visible=!0),e.isHidden=!1,b4(e)}}function MB(e,t,n){const r=t.root.getState();if(!(!e.parent&&e.object!==r.scene)){if(!t.object){var i,a;const u=U3[S5(t.type)];t.object=(i=t.props.object)!=null?i:new u(...(a=t.props.args)!=null?a:[]),t.object.__r3f=t}if(Pv(t.object,t.props),t.props.attach)vC(e,t);else if(tf(t.object)&&tf(e.object)){const u=e.object.children.indexOf(n==null?void 0:n.object);if(n&&u!==-1){const l=e.object.children.indexOf(t.object);if(l!==-1){e.object.children.splice(l,1);const o=l<u?u-1:u;e.object.children.splice(o,0,t.object)}else t.object.parent=e.object,e.object.children.splice(u,0,t.object),t.object.dispatchEvent({type:"added"}),e.object.dispatchEvent({type:"childadded",child:t.object})}else e.object.add(t.object)}for(const u of t.children)MB(t,u);b4(t)}}function V7(e,t){t&&(t.parent=e,e.children.push(t),MB(e,t))}function KW(e,t,n){if(!t||!n)return;t.parent=e;const r=e.children.indexOf(n);r!==-1?e.children.splice(r,0,t):e.children.push(t),MB(e,t,n)}function H1e(e){if(typeof e.dispose=="function"){const t=()=>{try{e.dispose()}catch{}};typeof IS_REACT_ACT_ENVIRONMENT<"u"?t():HW.unstable_scheduleCallback(HW.unstable_IdlePriority,t)}}function RF(e,t,n){if(!t)return;t.parent=null;const r=e.children.indexOf(t);r!==-1&&e.children.splice(r,1),t.props.attach?yC(e,t):tf(t.object)&&tf(e.object)&&(e.object.remove(t.object),Utt(N1e(t),t.object));const i=t.props.dispose!==null&&n!==!1;for(let a=t.children.length-1;a>=0;a--){const u=t.children[a];RF(t,u,i)}t.children.length=0,delete t.object.__r3f,i&&t.type!=="primitive"&&t.object.type!=="Scene"&&H1e(t.object),n===void 0&&b4(t)}function ent(e,t){for(const n of[e,e.alternate])if(n!==null)if(typeof n.ref=="function"){n.refCleanup==null||n.refCleanup();const r=n.ref(t);typeof r=="function"&&(n.refCleanup=r)}else n.ref&&(n.ref.current=t)}const CD=[];function tnt(){for(const[n]of CD){const r=n.parent;if(r){n.props.attach?yC(r,n):tf(n.object)&&tf(r.object)&&r.object.remove(n.object);for(const i of n.children)i.props.attach?yC(n,i):tf(i.object)&&tf(n.object)&&n.object.remove(i.object)}n.isHidden&&Z1e(n),n.object.__r3f&&delete n.object.__r3f,n.type!=="primitive"&&H1e(n.object)}for(const[n,r,i]of CD){n.props=r;const a=n.parent;if(a){var e,t;const u=U3[S5(n.type)];n.object=(e=n.props.object)!=null?e:new u(...(t=n.props.args)!=null?t:[]),n.object.__r3f=n,ent(i,n.object),Pv(n.object,n.props),n.props.attach?vC(a,n):tf(n.object)&&tf(a.object)&&a.object.add(n.object);for(const l of n.children)l.props.attach?vC(n,l):tf(l.object)&&tf(n.object)&&n.object.add(l.object);b4(n)}}CD.length=0}const U7=()=>{},WW={};let zE=U1e;const nnt=0,rnt=4,AF=Htt({isPrimaryRenderer:!1,warnsIfNotActing:!1,supportsMutation:!0,supportsPersistence:!1,supportsHydration:!1,createInstance:Wtt,removeChild:RF,appendChild:V7,appendInitialChild:V7,insertBefore:KW,appendChildToContainer(e,t){const n=e.getState().scene.__r3f;!t||!n||V7(n,t)},removeChildFromContainer(e,t){const n=e.getState().scene.__r3f;!t||!n||RF(n,t)},insertInContainerBefore(e,t,n){const r=e.getState().scene.__r3f;!t||!n||!r||KW(r,t,n)},getRootHostContext:()=>WW,getChildHostContext:()=>WW,commitUpdate(e,t,n,r,i){var a,u,l;j1e(t,r);let o=!1;if((e.type==="primitive"&&n.object!==r.object||((a=r.args)==null?void 0:a.length)!==((u=n.args)==null?void 0:u.length)||(l=r.args)!=null&&l.some((c,d)=>{var h;return c!==((h=n.args)==null?void 0:h[d])}))&&(o=!0),o)CD.push([e,{...r},i]);else{const c=Qtt(e,r);Object.keys(c).length&&(Object.assign(e.props,c),Pv(e.object,c))}(i.sibling===null||(i.flags&rnt)===nnt)&&tnt()},finalizeInitialChildren:()=>!1,commitMount(){},getPublicInstance:e=>e==null?void 0:e.object,prepareForCommit:()=>null,preparePortalMount:e=>gC(e.getState().scene,e,"",{}),resetAfterCommit:()=>{},shouldSetTextContent:()=>!1,clearContainer:()=>!1,hideInstance:Jtt,unhideInstance:Z1e,createTextInstance:U7,hideTextInstance:U7,unhideTextInstance:U7,scheduleTimeout:typeof setTimeout=="function"?setTimeout:void 0,cancelTimeout:typeof clearTimeout=="function"?clearTimeout:void 0,noTimeout:-1,getInstanceFromNode:()=>null,beforeActiveInstanceBlur(){},afterActiveInstanceBlur(){},detachDeletedInstance(){},prepareScopeUpdate(){},getInstanceFromScope:()=>null,shouldAttemptEagerTransition:()=>!1,trackSchedulerEvent:()=>{},resolveEventType:()=>null,resolveEventTimeStamp:()=>-1.1,requestPostPaintCallback(){},maySuspendCommit:()=>!1,preloadInstance:()=>!0,startSuspendingCommit(){},suspendInstance(){},waitForCommitToBeReady:()=>null,NotPendingTransition:null,HostTransitionContext:P.createContext(null),setCurrentUpdatePriority(e){zE=e},getCurrentUpdatePriority(){return zE},resolveUpdatePriority(){var e;if(zE!==U1e)return zE;switch(typeof window<"u"&&((e=window.event)==null?void 0:e.type)){case"click":case"contextmenu":case"dblclick":case"pointercancel":case"pointerdown":case"pointerup":return $D.DiscreteEventPriority;case"pointermove":case"pointerout":case"pointerover":case"pointerenter":case"pointerleave":case"wheel":return $D.ContinuousEventPriority;default:return $D.DefaultEventPriority}},resetFormInstance(){}}),S1=new Map,MO={objects:"shallow",strict:!1};function int(e,t){if(!t&&typeof HTMLCanvasElement<"u"&&e instanceof HTMLCanvasElement&&e.parentElement){const{width:n,height:r,top:i,left:a}=e.parentElement.getBoundingClientRect();return{width:n,height:r,top:i,left:a}}else if(!t&&typeof OffscreenCanvas<"u"&&e instanceof OffscreenCanvas)return{width:e.width,height:e.height,top:0,left:0};return{width:0,height:0,top:0,left:0,...t}}function ant(e){const t=S1.get(e),n=t==null?void 0:t.fiber,r=t==null?void 0:t.store;t&&console.warn("R3F.createRoot should only be called once!");const i=typeof reportError=="function"?reportError:console.error,a=r||Ztt(PF,eJ),u=n||AF.createContainer(a,$D.ConcurrentRoot,null,!1,null,"",i,i,i,null);t||S1.set(e,{fiber:u,store:a});let l,o,s=!1,c=null;return{async configure(d={}){let h;c=new Promise(G=>h=G);let{gl:m,size:y,scene:v,events:S,onCreated:_,shadows:E=!1,linear:$=!1,flat:k=!1,legacy:A=!1,orthographic:w=!1,frameloop:R="always",dpr:F=[1,2],performance:I,raycaster:U,camera:q,onPointerMissed:V}=d,N=a.getState(),j=N.gl;if(!N.gl){const G={canvas:e,powerPreference:"high-performance",antialias:!0,alpha:!0},X=typeof m=="function"?await m(G):m;GW(X)?j=X:j=new cOe({...G,...m}),N.set({gl:j})}let W=N.raycaster;W||N.set({raycaster:W=new fOe});const{params:K,...J}=U||{};if(su.equ(J,W,MO)||Pv(W,{...J}),su.equ(K,W.params,MO)||Pv(W,{params:{...W.params,...K}}),!N.camera||N.camera===o&&!su.equ(o,q,MO)){o=q;const G=q==null?void 0:q.isCamera,X=G?q:w?new JO(0,0,0,0,.1,1e3):new e3(75,0,.1,1e3);G||(X.position.z=5,q&&(Pv(X,q),X.manual||("aspect"in q||"left"in q||"right"in q||"bottom"in q||"top"in q)&&(X.manual=!0,X.updateProjectionMatrix())),!N.camera&&!(q!=null&&q.rotation)&&X.lookAt(0,0,0)),N.set({camera:X}),W.camera=X}if(!N.scene){let G;v!=null&&v.isScene?(G=v,gC(G,a,"",{})):(G=new dOe,gC(G,a,"",{}),v&&Pv(G,v)),N.set({scene:G})}S&&!N.events.handlers&&N.set({events:S(a)});const te=int(e,y);if(su.equ(te,N.size,MO)||N.setSize(te.width,te.height,te.top,te.left),F&&N.viewport.dpr!==B1e(F)&&N.setDpr(F),N.frameloop!==R&&N.setFrameloop(R),N.onPointerMissed||N.set({onPointerMissed:V}),I&&!su.equ(I,N.performance,MO)&&N.set(G=>({performance:{...G.performance,...I}})),!N.xr){var H;const G=(ne,oe)=>{const ie=a.getState();ie.frameloop!=="never"&&eJ(ne,!0,ie,oe)},X=()=>{const ne=a.getState();ne.gl.xr.enabled=ne.gl.xr.isPresenting,ne.gl.xr.setAnimationLoop(ne.gl.xr.isPresenting?G:null),ne.gl.xr.isPresenting||PF(ne)},B={connect(){const ne=a.getState().gl;ne.xr.addEventListener("sessionstart",X),ne.xr.addEventListener("sessionend",X)},disconnect(){const ne=a.getState().gl;ne.xr.removeEventListener("sessionstart",X),ne.xr.removeEventListener("sessionend",X)}};typeof((H=j.xr)==null?void 0:H.addEventListener)=="function"&&B.connect(),N.set({xr:B})}if(j.shadowMap){const G=j.shadowMap.enabled,X=j.shadowMap.type;if(j.shadowMap.enabled=!!E,su.boo(E))j.shadowMap.type=G5;else if(su.str(E)){var M;const B={basic:mOe,percentage:pOe,soft:G5,variance:hOe};j.shadowMap.type=(M=B[E])!=null?M:G5}else su.obj(E)&&Object.assign(j.shadowMap,E);(G!==j.shadowMap.enabled||X!==j.shadowMap.type)&&(j.shadowMap.needsUpdate=!0)}return gOe.enabled=!A,s||(j.outputColorSpace=$?vOe:hJ,j.toneMapping=k?yOe:bOe),N.legacy!==A&&N.set(()=>({legacy:A})),N.linear!==$&&N.set(()=>({linear:$})),N.flat!==k&&N.set(()=>({flat:k})),m&&!su.fun(m)&&!GW(m)&&!su.equ(m,j,MO)&&Pv(j,m),l=_,s=!0,h(),this},render(d){return!s&&!c&&this.configure(),c.then(()=>{AF.updateContainer(Qr.jsx(unt,{store:a,children:d,onCreated:l,rootElement:e}),u,null,()=>{})}),a},unmount(){X1e(e)}}}function unt({store:e,children:t,onCreated:n,rootElement:r}){return HS(()=>{const i=e.getState();i.set(a=>({internal:{...a.internal,active:!0}})),n&&n(i),e.getState().events.connected||i.events.connect==null||i.events.connect(r)},[]),Qr.jsx(q1e.Provider,{value:e,children:t})}function X1e(e,t){const n=S1.get(e),r=n==null?void 0:n.fiber;if(r){const i=n==null?void 0:n.store.getState();i&&(i.internal.active=!1),AF.updateContainer(null,r,null,()=>{i&&setTimeout(()=>{try{var a,u,l,o;i.events.disconnect==null||i.events.disconnect(),(a=i.gl)==null||(u=a.renderLists)==null||u.dispose==null||u.dispose(),(l=i.gl)==null||l.forceContextLoss==null||l.forceContextLoss(),(o=i.gl)!=null&&o.xr&&i.xr.disconnect(),Itt(i.scene),S1.delete(e)}catch{}},500)})}}const lnt=new Set,ont=new Set,snt=new Set;function j7(e,t){if(e.size)for(const{callback:n}of e.values())n(t)}function ox(e,t){switch(e){case"before":return j7(lnt,t);case"after":return j7(ont,t);case"tail":return j7(snt,t)}}let Z7,H7;function TF(e,t,n){let r=t.clock.getDelta();t.frameloop==="never"&&typeof e=="number"&&(r=e-t.clock.elapsedTime,t.clock.oldTime=t.clock.elapsedTime,t.clock.elapsedTime=e),Z7=t.internal.subscribers;for(let i=0;i<Z7.length;i++)H7=Z7[i],H7.ref.current(H7.store.getState(),r,n);return!t.internal.priority&&t.gl.render&&t.gl.render(t.scene,t.camera),t.internal.frames=Math.max(0,t.internal.frames-1),t.frameloop==="always"?1:t.internal.frames}let bC=!1,kF=!1,X7,JW,FO;function Y1e(e){JW=requestAnimationFrame(Y1e),bC=!0,X7=0,ox("before",e),kF=!0;for(const n of S1.values()){var t;FO=n.store.getState(),FO.internal.active&&(FO.frameloop==="always"||FO.internal.frames>0)&&!((t=FO.gl.xr)!=null&&t.isPresenting)&&(X7+=TF(e,FO))}if(kF=!1,ox("after",e),X7===0)return ox("tail",e),bC=!1,cancelAnimationFrame(JW)}function PF(e,t=1){var n;if(!e)return S1.forEach(r=>PF(r.store.getState(),t));(n=e.gl.xr)!=null&&n.isPresenting||!e.internal.active||e.frameloop==="never"||(t>1?e.internal.frames=Math.min(60,e.internal.frames+t):kF?e.internal.frames=2:e.internal.frames=1,bC||(bC=!0,requestAnimationFrame(Y1e)))}function eJ(e,t=!0,n,r){if(t&&ox("before",e),n)TF(e,n,r);else for(const i of S1.values())TF(e,i.store.getState());t&&ox("after",e)}const Y7={onClick:["click",!1],onContextMenu:["contextmenu",!1],onDoubleClick:["dblclick",!1],onWheel:["wheel",!0],onPointerDown:["pointerdown",!0],onPointerUp:["pointerup",!0],onPointerLeave:["pointerleave",!0],onPointerMove:["pointermove",!0],onPointerCancel:["pointercancel",!0],onLostPointerCapture:["lostpointercapture",!0]};function cnt(e){const{handlePointer:t}=jtt(e);return{priority:1,enabled:!0,compute(n,r,i){r.pointer.set(n.offsetX/r.size.width*2-1,-(n.offsetY/r.size.height)*2+1),r.raycaster.setFromCamera(r.pointer,r.camera)},connected:void 0,handlers:Object.keys(Y7).reduce((n,r)=>({...n,[r]:t(r)}),{}),update:()=>{var n;const{events:r,internal:i}=e.getState();(n=i.lastEvent)!=null&&n.current&&r.handlers&&r.handlers.onPointerMove(i.lastEvent.current)},connect:n=>{const{set:r,events:i}=e.getState();if(i.disconnect==null||i.disconnect(),r(a=>({events:{...a.events,connected:n}})),i.handlers)for(const a in i.handlers){const u=i.handlers[a],[l,o]=Y7[a];n.addEventListener(l,u,{passive:o})}},disconnect:()=>{const{set:n,events:r}=e.getState();if(r.connected){if(r.handlers)for(const i in r.handlers){const a=r.handlers[i],[u]=Y7[i];r.connected.removeEventListener(u,a)}n(i=>({events:{...i.events,connected:void 0}}))}}}}function tJ(e,t){let n;return(...r)=>{window.clearTimeout(n),n=window.setTimeout(()=>e(...r),t)}}function fnt({debounce:e,scroll:t,polyfill:n,offsetSize:r}={debounce:0,scroll:!1,offsetSize:!1}){const i=n||(typeof window>"u"?class{}:window.ResizeObserver);if(!i)throw new Error("This browser does not support ResizeObserver out of the box. See: https://github.com/react-spring/react-use-measure/#resize-observer-polyfills");const[a,u]=P.useState({left:0,top:0,width:0,height:0,bottom:0,right:0,x:0,y:0}),l=P.useRef({element:null,scrollContainers:null,resizeObserver:null,lastBounds:a,orientationHandler:null}),o=e?typeof e=="number"?e:e.scroll:null,s=e?typeof e=="number"?e:e.resize:null,c=P.useRef(!1);P.useEffect(()=>(c.current=!0,()=>void(c.current=!1)));const[d,h,m]=P.useMemo(()=>{const _=()=>{if(!l.current.element)return;const{left:E,top:$,width:k,height:A,bottom:w,right:R,x:F,y:I}=l.current.element.getBoundingClientRect(),U={left:E,top:$,width:k,height:A,bottom:w,right:R,x:F,y:I};l.current.element instanceof HTMLElement&&r&&(U.height=l.current.element.offsetHeight,U.width=l.current.element.offsetWidth),Object.freeze(U),c.current&&!mnt(l.current.lastBounds,U)&&u(l.current.lastBounds=U)};return[_,s?tJ(_,s):_,o?tJ(_,o):_]},[u,r,o,s]);function y(){l.current.scrollContainers&&(l.current.scrollContainers.forEach(_=>_.removeEventListener("scroll",m,!0)),l.current.scrollContainers=null),l.current.resizeObserver&&(l.current.resizeObserver.disconnect(),l.current.resizeObserver=null),l.current.orientationHandler&&("orientation"in screen&&"removeEventListener"in screen.orientation?screen.orientation.removeEventListener("change",l.current.orientationHandler):"onorientationchange"in window&&window.removeEventListener("orientationchange",l.current.orientationHandler))}function v(){l.current.element&&(l.current.resizeObserver=new i(m),l.current.resizeObserver.observe(l.current.element),t&&l.current.scrollContainers&&l.current.scrollContainers.forEach(_=>_.addEventListener("scroll",m,{capture:!0,passive:!0})),l.current.orientationHandler=()=>{m()},"orientation"in screen&&"addEventListener"in screen.orientation?screen.orientation.addEventListener("change",l.current.orientationHandler):"onorientationchange"in window&&window.addEventListener("orientationchange",l.current.orientationHandler))}const S=_=>{!_||_===l.current.element||(y(),l.current.element=_,l.current.scrollContainers=G1e(_),v())};return hnt(m,!!t),dnt(h),P.useEffect(()=>{y(),v()},[t,m,h]),P.useEffect(()=>y,[]),[S,a,d]}function dnt(e){P.useEffect(()=>{const t=e;return window.addEventListener("resize",t),()=>void window.removeEventListener("resize",t)},[e])}function hnt(e,t){P.useEffect(()=>{if(t){const n=e;return window.addEventListener("scroll",n,{capture:!0,passive:!0}),()=>void window.removeEventListener("scroll",n,!0)}},[e,t])}function G1e(e){const t=[];if(!e||e===document.body)return t;const{overflow:n,overflowX:r,overflowY:i}=window.getComputedStyle(e);return[n,r,i].some(a=>a==="auto"||a==="scroll")&&t.push(e),[...t,...G1e(e.parentElement)]}const pnt=["x","y","top","bottom","left","right","width","height"],mnt=(e,t)=>pnt.every(n=>e[n]===t[n]);function gnt({ref:e,children:t,fallback:n,resize:r,style:i,gl:a,events:u=cnt,eventSource:l,eventPrefix:o,shadows:s,linear:c,flat:d,legacy:h,orthographic:m,frameloop:y,dpr:v,performance:S,raycaster:_,camera:E,scene:$,onPointerMissed:k,onCreated:A,...w}){P.useMemo(()=>Ktt(_Oe),[]);const R=Mtt(),[F,I]=fnt({scroll:!0,debounce:{scroll:50,resize:0},...r}),U=P.useRef(null),q=P.useRef(null);P.useImperativeHandle(e,()=>U.current);const V=L1e(k),[N,j]=P.useState(!1),[W,K]=P.useState(!1);if(N)throw N;if(W)throw W;const J=P.useRef(null);HS(()=>{const H=U.current;if(I.width>0&&I.height>0&&H){J.current||(J.current=ant(H));async function M(){await J.current.configure({gl:a,scene:$,events:u,shadows:s,linear:c,flat:d,legacy:h,orthographic:m,frameloop:y,dpr:v,performance:S,raycaster:_,camera:E,size:I,onPointerMissed:(...G)=>V.current==null?void 0:V.current(...G),onCreated:G=>{G.events.connect==null||G.events.connect(l?ktt(l)?l.current:l:q.current),o&&G.setEvents({compute:(X,B)=>{const ne=X[o+"X"],oe=X[o+"Y"];B.pointer.set(ne/B.size.width*2-1,-(oe/B.size.height)*2+1),B.raycaster.setFromCamera(B.pointer,B.camera)}}),A==null||A(G)}}),J.current.render(Qr.jsx(R,{children:Qr.jsx(Ntt,{set:K,children:Qr.jsx(P.Suspense,{fallback:Qr.jsx(Ftt,{set:j}),children:t??null})})}))}M()}}),P.useEffect(()=>{const H=U.current;if(H)return()=>X1e(H)},[]);const te=l?"none":"auto";return Qr.jsx("div",{ref:q,style:{position:"relative",width:"100%",height:"100%",overflow:"hidden",pointerEvents:te,...i},...w,children:Qr.jsx("div",{ref:F,style:{width:"100%",height:"100%"},children:Qr.jsx("canvas",{ref:U,style:{display:"block"},children:n})})})}function vnt(e){return Qr.jsx(M1e,{children:Qr.jsx(gnt,{...e})})}const XS=new ci,FB=new ci,ynt=new ci,nJ=new ao;function bnt(e,t,n){const r=XS.setFromMatrixPosition(e.matrixWorld);r.project(t);const i=n.width/2,a=n.height/2;return[r.x*i+i,-(r.y*a)+a]}function Ont(e,t){const n=XS.setFromMatrixPosition(e.matrixWorld),r=FB.setFromMatrixPosition(t.matrixWorld),i=n.sub(r),a=t.getWorldDirection(ynt);return i.angleTo(a)>Math.PI/2}function xnt(e,t,n,r){const i=XS.setFromMatrixPosition(e.matrixWorld),a=i.clone();a.project(t),nJ.set(a.x,a.y),n.setFromCamera(nJ,t);const u=n.intersectObjects(r,!0);if(u.length){const l=u[0].distance;return i.distanceTo(n.ray.origin)<l}return!0}function Snt(e,t){if(t instanceof JO)return t.zoom;if(t instanceof e3){const n=XS.setFromMatrixPosition(e.matrixWorld),r=FB.setFromMatrixPosition(t.matrixWorld),i=t.fov*Math.PI/180,a=n.distanceTo(r);return 1/(2*Math.tan(i/2)*a)}else return 1}function _nt(e,t,n){if(t instanceof e3||t instanceof JO){const r=XS.setFromMatrixPosition(e.matrixWorld),i=FB.setFromMatrixPosition(t.matrixWorld),a=r.distanceTo(i),u=(n[1]-n[0])/(t.far-t.near),l=n[1]-u*t.far;return Math.round(u*a+l)}}const MF=e=>Math.abs(e)<1e-10?0:e;function K1e(e,t,n=""){let r="matrix3d(";for(let i=0;i!==16;i++)r+=MF(t[i]*e.elements[i])+(i!==15?",":")");return n+r}const Ent=(e=>t=>K1e(t,e))([1,-1,1,1,1,-1,1,1,1,-1,1,1,1,-1,1,1]),Dnt=(e=>(t,n)=>K1e(t,e(n),"translate(-50%,-50%)"))(e=>[1/e,1/e,1/e,1,-1/e,-1/e,-1/e,-1,1/e,1/e,1/e,1,1,1,1,1]);function $nt(e){return e&&typeof e=="object"&&"current"in e}const Ilt=P.forwardRef(({children:e,eps:t=.001,style:n,className:r,prepend:i,center:a,fullscreen:u,portal:l,distanceFactor:o,sprite:s=!1,transform:c=!1,occlude:d,onOcclude:h,castShadow:m,receiveShadow:y,material:v,geometry:S,zIndexRange:_=[16777271,0],calculatePosition:E=bnt,as:$="div",wrapperClass:k,pointerEvents:A="auto",...w},R)=>{const{gl:F,camera:I,scene:U,size:q,raycaster:V,events:N,viewport:j}=Vd(),[W]=P.useState(()=>document.createElement($)),K=P.useRef(null),J=P.useRef(null),te=P.useRef(0),H=P.useRef([0,0]),M=P.useRef(null),G=P.useRef(null),X=(l==null?void 0:l.current)||N.connected||F.domElement.parentNode,B=P.useRef(null),ne=P.useRef(!1),oe=P.useMemo(()=>d&&d!=="blending"||Array.isArray(d)&&d.length&&$nt(d[0]),[d]);P.useLayoutEffect(()=>{const me=F.domElement;d&&d==="blending"?(me.style.zIndex=`${Math.floor(_[0]/2)}`,me.style.position="absolute",me.style.pointerEvents="none"):(me.style.zIndex=null,me.style.position=null,me.style.pointerEvents=null)},[d]),P.useLayoutEffect(()=>{if(J.current){const me=K.current=TJ.createRoot(W);if(U.updateMatrixWorld(),c)W.style.cssText="position:absolute;top:0;left:0;pointer-events:none;overflow:hidden;";else{const Ee=E(J.current,I,q);W.style.cssText=`position:absolute;top:0;left:0;transform:translate3d(${Ee[0]}px,${Ee[1]}px,0);transform-origin:0 0;`}return X&&(i?X.prepend(W):X.appendChild(W)),()=>{X&&X.removeChild(W),me.unmount()}}},[X,c]),P.useLayoutEffect(()=>{k&&(W.className=k)},[k]);const ie=P.useMemo(()=>c?{position:"absolute",top:0,left:0,width:q.width,height:q.height,transformStyle:"preserve-3d",pointerEvents:"none"}:{position:"absolute",transform:a?"translate3d(-50%,-50%,0)":"none",...u&&{top:-q.height/2,left:-q.width/2,width:q.width,height:q.height},...n},[n,a,u,q,c]),de=P.useMemo(()=>({position:"absolute",pointerEvents:A}),[A]);P.useLayoutEffect(()=>{if(ne.current=!1,c){var me;(me=K.current)==null||me.render(P.createElement("div",{ref:M,style:ie},P.createElement("div",{ref:G,style:de},P.createElement("div",{ref:R,className:r,style:n,children:e}))))}else{var Ee;(Ee=K.current)==null||Ee.render(P.createElement("div",{ref:R,style:ie,className:r,children:e}))}});const ge=P.useRef(!0);x5(me=>{if(J.current){I.updateMatrixWorld(),J.current.updateWorldMatrix(!0,!1);const Ee=c?H.current:E(J.current,I,q);if(c||Math.abs(te.current-I.zoom)>t||Math.abs(H.current[0]-Ee[0])>t||Math.abs(H.current[1]-Ee[1])>t){const se=Ont(J.current,I);let pe=!1;oe&&(Array.isArray(d)?pe=d.map(Re=>Re.current):d!=="blending"&&(pe=[U]));const ve=ge.current;if(pe){const Re=xnt(J.current,I,V,pe);ge.current=Re&&!se}else ge.current=!se;ve!==ge.current&&(h?h(!ge.current):W.style.display=ge.current?"block":"none");const we=Math.floor(_[0]/2),Me=d?oe?[_[0],we]:[we-1,0]:_;if(W.style.zIndex=`${_nt(J.current,I,Me)}`,c){const[Re,et]=[q.width/2,q.height/2],tt=I.projectionMatrix.elements[5]*et,{isOrthographicCamera:Oe,top:$e,left:Ae,bottom:He,right:je}=I,Ye=Ent(I.matrixWorldInverse),Ne=Oe?`scale(${tt})translate(${MF(-(je+Ae)/2)}px,${MF(($e+He)/2)}px)`:`translateZ(${tt}px)`;let ze=J.current.matrixWorld;s&&(ze=I.matrixWorldInverse.clone().transpose().copyPosition(ze).scale(J.current.scale),ze.elements[3]=ze.elements[7]=ze.elements[11]=0,ze.elements[15]=1),W.style.width=q.width+"px",W.style.height=q.height+"px",W.style.perspective=Oe?"":`${tt}px`,M.current&&G.current&&(M.current.style.transform=`${Ne}${Ye}translate(${Re}px,${et}px)`,G.current.style.transform=Dnt(ze,1/((o||10)/400)))}else{const Re=o===void 0?1:Snt(J.current,I)*o;W.style.transform=`translate3d(${Ee[0]}px,${Ee[1]}px,0) scale(${Re})`}H.current=Ee,te.current=I.zoom}}if(!oe&&B.current&&!ne.current)if(c){if(M.current){const Ee=M.current.children[0];if(Ee!=null&&Ee.clientWidth&&Ee!=null&&Ee.clientHeight){const{isOrthographicCamera:se}=I;if(se||S)w.scale&&(Array.isArray(w.scale)?w.scale instanceof ci?B.current.scale.copy(w.scale.clone().divideScalar(1)):B.current.scale.set(1/w.scale[0],1/w.scale[1],1/w.scale[2]):B.current.scale.setScalar(1/w.scale));else{const pe=(o||10)/400,ve=Ee.clientWidth*pe,we=Ee.clientHeight*pe;B.current.scale.set(ve,we,1)}ne.current=!0}}}else{const Ee=W.children[0];if(Ee!=null&&Ee.clientWidth&&Ee!=null&&Ee.clientHeight){const se=1/j.factor,pe=Ee.clientWidth*se,ve=Ee.clientHeight*se;B.current.scale.set(pe,ve,1),ne.current=!0}B.current.lookAt(me.camera.position)}});const ye=P.useMemo(()=>({vertexShader:c?void 0:`
          /*
            This shader is from the THREE's SpriteMaterial.
            We need to turn the backing plane into a Sprite
            (make it always face the camera) if "transfrom"
            is false.
          */
          #include <common>

          void main() {
            vec2 center = vec2(0., 1.);
            float rotation = 0.0;

            // This is somewhat arbitrary, but it seems to work well
            // Need to figure out how to derive this dynamically if it even matters
            float size = 0.03;

            vec4 mvPosition = modelViewMatrix * vec4( 0.0, 0.0, 0.0, 1.0 );
            vec2 scale;
            scale.x = length( vec3( modelMatrix[ 0 ].x, modelMatrix[ 0 ].y, modelMatrix[ 0 ].z ) );
            scale.y = length( vec3( modelMatrix[ 1 ].x, modelMatrix[ 1 ].y, modelMatrix[ 1 ].z ) );

            bool isPerspective = isPerspectiveMatrix( projectionMatrix );
            if ( isPerspective ) scale *= - mvPosition.z;

            vec2 alignedPosition = ( position.xy - ( center - vec2( 0.5 ) ) ) * scale * size;
            vec2 rotatedPosition;
            rotatedPosition.x = cos( rotation ) * alignedPosition.x - sin( rotation ) * alignedPosition.y;
            rotatedPosition.y = sin( rotation ) * alignedPosition.x + cos( rotation ) * alignedPosition.y;
            mvPosition.xy += rotatedPosition;

            gl_Position = projectionMatrix * mvPosition;
          }
      `,fragmentShader:`
        void main() {
          gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0);
        }
      `}),[c]);return P.createElement("group",ry({},w,{ref:J}),d&&!oe&&P.createElement("mesh",{castShadow:m,receiveShadow:y,ref:B},S||P.createElement("planeGeometry",null),v||P.createElement("shaderMaterial",{side:EOe,vertexShader:ye.vertexShader,fragmentShader:ye.fragmentShader})))}),W1e=parseInt(DOe.replace(/\D+/g,"")),J1e=W1e>=125?"uv1":"uv2",Cnt=(e,t)=>{const n=e[0].index!==null,r=new Set(Object.keys(e[0].attributes)),i=new Set(Object.keys(e[0].morphAttributes)),a={},u={},l=e[0].morphTargetsRelative,o=new $Oe;if(e.forEach((s,c)=>{let d=0;if(n!==(s.index!==null))return console.error("THREE.BufferGeometryUtils: .mergeBufferGeometries() failed with geometry at index "+c+". All geometries must have compatible attributes; make sure index attribute exists among all geometries, or in none of them."),null;for(let h in s.attributes){if(!r.has(h))return console.error("THREE.BufferGeometryUtils: .mergeBufferGeometries() failed with geometry at index "+c+'. All geometries must have compatible attributes; make sure "'+h+'" attribute exists among all geometries, or in none of them.'),null;a[h]===void 0&&(a[h]=[]),a[h].push(s.attributes[h]),d++}if(d!==r.size)return console.error("THREE.BufferGeometryUtils: .mergeBufferGeometries() failed with geometry at index "+c+". Make sure all geometries have the same number of attributes."),null;if(l!==s.morphTargetsRelative)return console.error("THREE.BufferGeometryUtils: .mergeBufferGeometries() failed with geometry at index "+c+". .morphTargetsRelative must be consistent throughout all geometries."),null;for(let h in s.morphAttributes){if(!i.has(h))return console.error("THREE.BufferGeometryUtils: .mergeBufferGeometries() failed with geometry at index "+c+".  .morphAttributes must be consistent throughout all geometries."),null;u[h]===void 0&&(u[h]=[]),u[h].push(s.morphAttributes[h])}o.userData.mergedUserData=o.userData.mergedUserData||[],o.userData.mergedUserData.push(s.userData)}),n){let s=0;const c=[];e.forEach(d=>{const h=d.index;for(let m=0;m<h.count;++m)c.push(h.getX(m)+s);s+=d.attributes.position.count}),o.setIndex(c)}for(let s in a){const c=rJ(a[s]);if(!c)return console.error("THREE.BufferGeometryUtils: .mergeBufferGeometries() failed while trying to merge the "+s+" attribute."),null;o.setAttribute(s,c)}for(let s in u){const c=u[s][0].length;if(c===0)break;o.morphAttributes=o.morphAttributes||{},o.morphAttributes[s]=[];for(let d=0;d<c;++d){const h=[];for(let y=0;y<u[s].length;++y)h.push(u[s][y][d]);const m=rJ(h);if(!m)return console.error("THREE.BufferGeometryUtils: .mergeBufferGeometries() failed while trying to merge the "+s+" morphAttribute."),null;o.morphAttributes[s].push(m)}}return o},rJ=e=>{let t,n,r,i=0;if(e.forEach(a=>{if(t===void 0&&(t=a.array.constructor),t!==a.array.constructor)return console.error("THREE.BufferGeometryUtils: .mergeBufferAttributes() failed. BufferAttribute.array must be of consistent array types across matching attributes."),null;if(n===void 0&&(n=a.itemSize),n!==a.itemSize)return console.error("THREE.BufferGeometryUtils: .mergeBufferAttributes() failed. BufferAttribute.itemSize must be consistent across matching attributes."),null;if(r===void 0&&(r=a.normalized),r!==a.normalized)return console.error("THREE.BufferGeometryUtils: .mergeBufferAttributes() failed. BufferAttribute.normalized must be consistent across matching attributes."),null;i+=a.array.length}),t&&n){const a=new t(i);let u=0;return e.forEach(l=>{a.set(l.array,u),u+=l.array.length}),new COe(a,n,r)}};var wnt=Object.defineProperty,Rnt=(e,t,n)=>t in e?wnt(e,t,{enumerable:!0,configurable:!0,writable:!0,value:n}):e[t]=n,Ant=(e,t,n)=>(Rnt(e,t+"",n),n);class Tnt{constructor(){Ant(this,"_listeners")}addEventListener(t,n){this._listeners===void 0&&(this._listeners={});const r=this._listeners;r[t]===void 0&&(r[t]=[]),r[t].indexOf(n)===-1&&r[t].push(n)}hasEventListener(t,n){if(this._listeners===void 0)return!1;const r=this._listeners;return r[t]!==void 0&&r[t].indexOf(n)!==-1}removeEventListener(t,n){if(this._listeners===void 0)return;const i=this._listeners[t];if(i!==void 0){const a=i.indexOf(n);a!==-1&&i.splice(a,1)}}dispatchEvent(t){if(this._listeners===void 0)return;const r=this._listeners[t.type];if(r!==void 0){t.target=this;const i=r.slice(0);for(let a=0,u=i.length;a<u;a++)i[a].call(this,t);t.target=null}}}var knt=Object.defineProperty,Pnt=(e,t,n)=>t in e?knt(e,t,{enumerable:!0,configurable:!0,writable:!0,value:n}):e[t]=n,kn=(e,t,n)=>(Pnt(e,typeof t!="symbol"?t+"":t,n),n);const qE=new wOe,iJ=new ROe,Mnt=Math.cos(70*(Math.PI/180)),aJ=(e,t)=>(e%t+t)%t;let Fnt=class extends Tnt{constructor(t,n){super(),kn(this,"object"),kn(this,"domElement"),kn(this,"enabled",!0),kn(this,"target",new ci),kn(this,"minDistance",0),kn(this,"maxDistance",1/0),kn(this,"minZoom",0),kn(this,"maxZoom",1/0),kn(this,"minPolarAngle",0),kn(this,"maxPolarAngle",Math.PI),kn(this,"minAzimuthAngle",-1/0),kn(this,"maxAzimuthAngle",1/0),kn(this,"enableDamping",!1),kn(this,"dampingFactor",.05),kn(this,"enableZoom",!0),kn(this,"zoomSpeed",1),kn(this,"enableRotate",!0),kn(this,"rotateSpeed",1),kn(this,"enablePan",!0),kn(this,"panSpeed",1),kn(this,"screenSpacePanning",!0),kn(this,"keyPanSpeed",7),kn(this,"zoomToCursor",!1),kn(this,"autoRotate",!1),kn(this,"autoRotateSpeed",2),kn(this,"reverseOrbit",!1),kn(this,"reverseHorizontalOrbit",!1),kn(this,"reverseVerticalOrbit",!1),kn(this,"keys",{LEFT:"ArrowLeft",UP:"ArrowUp",RIGHT:"ArrowRight",BOTTOM:"ArrowDown"}),kn(this,"mouseButtons",{LEFT:gO.ROTATE,MIDDLE:gO.DOLLY,RIGHT:gO.PAN}),kn(this,"touches",{ONE:vO.ROTATE,TWO:vO.DOLLY_PAN}),kn(this,"target0"),kn(this,"position0"),kn(this,"zoom0"),kn(this,"_domElementKeyEvents",null),kn(this,"getPolarAngle"),kn(this,"getAzimuthalAngle"),kn(this,"setPolarAngle"),kn(this,"setAzimuthalAngle"),kn(this,"getDistance"),kn(this,"getZoomScale"),kn(this,"listenToKeyEvents"),kn(this,"stopListenToKeyEvents"),kn(this,"saveState"),kn(this,"reset"),kn(this,"update"),kn(this,"connect"),kn(this,"dispose"),kn(this,"dollyIn"),kn(this,"dollyOut"),kn(this,"getScale"),kn(this,"setScale"),this.object=t,this.domElement=n,this.target0=this.target.clone(),this.position0=this.object.position.clone(),this.zoom0=this.object.zoom,this.getPolarAngle=()=>c.phi,this.getAzimuthalAngle=()=>c.theta,this.setPolarAngle=Fe=>{let ht=aJ(Fe,2*Math.PI),kt=c.phi;kt<0&&(kt+=2*Math.PI),ht<0&&(ht+=2*Math.PI);let Qt=Math.abs(ht-kt);2*Math.PI-Qt<Qt&&(ht<kt?ht+=2*Math.PI:kt+=2*Math.PI),d.phi=ht-kt,r.update()},this.setAzimuthalAngle=Fe=>{let ht=aJ(Fe,2*Math.PI),kt=c.theta;kt<0&&(kt+=2*Math.PI),ht<0&&(ht+=2*Math.PI);let Qt=Math.abs(ht-kt);2*Math.PI-Qt<Qt&&(ht<kt?ht+=2*Math.PI:kt+=2*Math.PI),d.theta=ht-kt,r.update()},this.getDistance=()=>r.object.position.distanceTo(r.target),this.listenToKeyEvents=Fe=>{Fe.addEventListener("keydown",at),this._domElementKeyEvents=Fe},this.stopListenToKeyEvents=()=>{this._domElementKeyEvents.removeEventListener("keydown",at),this._domElementKeyEvents=null},this.saveState=()=>{r.target0.copy(r.target),r.position0.copy(r.object.position),r.zoom0=r.object.zoom},this.reset=()=>{r.target.copy(r.target0),r.object.position.copy(r.position0),r.object.zoom=r.zoom0,r.object.updateProjectionMatrix(),r.dispatchEvent(i),r.update(),o=l.NONE},this.update=(()=>{const Fe=new ci,ht=new ci(0,1,0),kt=new qB().setFromUnitVectors(t.up,ht),Qt=kt.clone().invert(),en=new ci,In=new qB,rn=2*Math.PI;return function(){const br=r.object.position;kt.setFromUnitVectors(t.up,ht),Qt.copy(kt).invert(),Fe.copy(br).sub(r.target),Fe.applyQuaternion(kt),c.setFromVector3(Fe),r.autoRotate&&o===l.NONE&&j(V()),r.enableDamping?(c.theta+=d.theta*r.dampingFactor,c.phi+=d.phi*r.dampingFactor):(c.theta+=d.theta,c.phi+=d.phi);let Or=r.minAzimuthAngle,Cr=r.maxAzimuthAngle;isFinite(Or)&&isFinite(Cr)&&(Or<-Math.PI?Or+=rn:Or>Math.PI&&(Or-=rn),Cr<-Math.PI?Cr+=rn:Cr>Math.PI&&(Cr-=rn),Or<=Cr?c.theta=Math.max(Or,Math.min(Cr,c.theta)):c.theta=c.theta>(Or+Cr)/2?Math.max(Or,c.theta):Math.min(Cr,c.theta)),c.phi=Math.max(r.minPolarAngle,Math.min(r.maxPolarAngle,c.phi)),c.makeSafe(),r.enableDamping===!0?r.target.addScaledVector(m,r.dampingFactor):r.target.add(m),r.zoomToCursor&&I||r.object.isOrthographicCamera?c.radius=B(c.radius):c.radius=B(c.radius*h),Fe.setFromSpherical(c),Fe.applyQuaternion(Qt),br.copy(r.target).add(Fe),r.object.matrixAutoUpdate||r.object.updateMatrix(),r.object.lookAt(r.target),r.enableDamping===!0?(d.theta*=1-r.dampingFactor,d.phi*=1-r.dampingFactor,m.multiplyScalar(1-r.dampingFactor)):(d.set(0,0,0),m.set(0,0,0));let wr=!1;if(r.zoomToCursor&&I){let Gr=null;if(r.object instanceof e3&&r.object.isPerspectiveCamera){const ji=Fe.length();Gr=B(ji*h);const ai=ji-Gr;r.object.position.addScaledVector(R,ai),r.object.updateMatrixWorld()}else if(r.object.isOrthographicCamera){const ji=new ci(F.x,F.y,0);ji.unproject(r.object),r.object.zoom=Math.max(r.minZoom,Math.min(r.maxZoom,r.object.zoom/h)),r.object.updateProjectionMatrix(),wr=!0;const ai=new ci(F.x,F.y,0);ai.unproject(r.object),r.object.position.sub(ai).add(ji),r.object.updateMatrixWorld(),Gr=Fe.length()}else console.warn("WARNING: OrbitControls.js encountered an unknown camera type - zoom to cursor disabled."),r.zoomToCursor=!1;Gr!==null&&(r.screenSpacePanning?r.target.set(0,0,-1).transformDirection(r.object.matrix).multiplyScalar(Gr).add(r.object.position):(qE.origin.copy(r.object.position),qE.direction.set(0,0,-1).transformDirection(r.object.matrix),Math.abs(r.object.up.dot(qE.direction))<Mnt?t.lookAt(r.target):(iJ.setFromNormalAndCoplanarPoint(r.object.up,r.target),qE.intersectPlane(iJ,r.target))))}else r.object instanceof JO&&r.object.isOrthographicCamera&&(wr=h!==1,wr&&(r.object.zoom=Math.max(r.minZoom,Math.min(r.maxZoom,r.object.zoom/h)),r.object.updateProjectionMatrix()));return h=1,I=!1,wr||en.distanceToSquared(r.object.position)>s||8*(1-In.dot(r.object.quaternion))>s?(r.dispatchEvent(i),en.copy(r.object.position),In.copy(r.object.quaternion),wr=!1,!0):!1}})(),this.connect=Fe=>{r.domElement=Fe,r.domElement.style.touchAction="none",r.domElement.addEventListener("contextmenu",bt),r.domElement.addEventListener("pointerdown",Ae),r.domElement.addEventListener("pointercancel",je),r.domElement.addEventListener("wheel",ze)},this.dispose=()=>{var Fe,ht,kt,Qt,en,In;r.domElement&&(r.domElement.style.touchAction="auto"),(Fe=r.domElement)==null||Fe.removeEventListener("contextmenu",bt),(ht=r.domElement)==null||ht.removeEventListener("pointerdown",Ae),(kt=r.domElement)==null||kt.removeEventListener("pointercancel",je),(Qt=r.domElement)==null||Qt.removeEventListener("wheel",ze),(en=r.domElement)==null||en.ownerDocument.removeEventListener("pointermove",He),(In=r.domElement)==null||In.ownerDocument.removeEventListener("pointerup",je),r._domElementKeyEvents!==null&&r._domElementKeyEvents.removeEventListener("keydown",at)};const r=this,i={type:"change"},a={type:"start"},u={type:"end"},l={NONE:-1,ROTATE:0,DOLLY:1,PAN:2,TOUCH_ROTATE:3,TOUCH_PAN:4,TOUCH_DOLLY_PAN:5,TOUCH_DOLLY_ROTATE:6};let o=l.NONE;const s=1e-6,c=new zB,d=new zB;let h=1;const m=new ci,y=new ao,v=new ao,S=new ao,_=new ao,E=new ao,$=new ao,k=new ao,A=new ao,w=new ao,R=new ci,F=new ao;let I=!1;const U=[],q={};function V(){return 2*Math.PI/60/60*r.autoRotateSpeed}function N(){return Math.pow(.95,r.zoomSpeed)}function j(Fe){r.reverseOrbit||r.reverseHorizontalOrbit?d.theta+=Fe:d.theta-=Fe}function W(Fe){r.reverseOrbit||r.reverseVerticalOrbit?d.phi+=Fe:d.phi-=Fe}const K=(()=>{const Fe=new ci;return function(kt,Qt){Fe.setFromMatrixColumn(Qt,0),Fe.multiplyScalar(-kt),m.add(Fe)}})(),J=(()=>{const Fe=new ci;return function(kt,Qt){r.screenSpacePanning===!0?Fe.setFromMatrixColumn(Qt,1):(Fe.setFromMatrixColumn(Qt,0),Fe.crossVectors(r.object.up,Fe)),Fe.multiplyScalar(kt),m.add(Fe)}})(),te=(()=>{const Fe=new ci;return function(kt,Qt){const en=r.domElement;if(en&&r.object instanceof e3&&r.object.isPerspectiveCamera){const In=r.object.position;Fe.copy(In).sub(r.target);let rn=Fe.length();rn*=Math.tan(r.object.fov/2*Math.PI/180),K(2*kt*rn/en.clientHeight,r.object.matrix),J(2*Qt*rn/en.clientHeight,r.object.matrix)}else en&&r.object instanceof JO&&r.object.isOrthographicCamera?(K(kt*(r.object.right-r.object.left)/r.object.zoom/en.clientWidth,r.object.matrix),J(Qt*(r.object.top-r.object.bottom)/r.object.zoom/en.clientHeight,r.object.matrix)):(console.warn("WARNING: OrbitControls.js encountered an unknown camera type - pan disabled."),r.enablePan=!1)}})();function H(Fe){r.object instanceof e3&&r.object.isPerspectiveCamera||r.object instanceof JO&&r.object.isOrthographicCamera?h=Fe:(console.warn("WARNING: OrbitControls.js encountered an unknown camera type - dolly/zoom disabled."),r.enableZoom=!1)}function M(Fe){H(h/Fe)}function G(Fe){H(h*Fe)}function X(Fe){if(!r.zoomToCursor||!r.domElement)return;I=!0;const ht=r.domElement.getBoundingClientRect(),kt=Fe.clientX-ht.left,Qt=Fe.clientY-ht.top,en=ht.width,In=ht.height;F.x=kt/en*2-1,F.y=-(Qt/In)*2+1,R.set(F.x,F.y,1).unproject(r.object).sub(r.object.position).normalize()}function B(Fe){return Math.max(r.minDistance,Math.min(r.maxDistance,Fe))}function ne(Fe){y.set(Fe.clientX,Fe.clientY)}function oe(Fe){X(Fe),k.set(Fe.clientX,Fe.clientY)}function ie(Fe){_.set(Fe.clientX,Fe.clientY)}function de(Fe){v.set(Fe.clientX,Fe.clientY),S.subVectors(v,y).multiplyScalar(r.rotateSpeed);const ht=r.domElement;ht&&(j(2*Math.PI*S.x/ht.clientHeight),W(2*Math.PI*S.y/ht.clientHeight)),y.copy(v),r.update()}function ge(Fe){A.set(Fe.clientX,Fe.clientY),w.subVectors(A,k),w.y>0?M(N()):w.y<0&&G(N()),k.copy(A),r.update()}function ye(Fe){E.set(Fe.clientX,Fe.clientY),$.subVectors(E,_).multiplyScalar(r.panSpeed),te($.x,$.y),_.copy(E),r.update()}function me(Fe){X(Fe),Fe.deltaY<0?G(N()):Fe.deltaY>0&&M(N()),r.update()}function Ee(Fe){let ht=!1;switch(Fe.code){case r.keys.UP:te(0,r.keyPanSpeed),ht=!0;break;case r.keys.BOTTOM:te(0,-r.keyPanSpeed),ht=!0;break;case r.keys.LEFT:te(r.keyPanSpeed,0),ht=!0;break;case r.keys.RIGHT:te(-r.keyPanSpeed,0),ht=!0;break}ht&&(Fe.preventDefault(),r.update())}function se(){if(U.length==1)y.set(U[0].pageX,U[0].pageY);else{const Fe=.5*(U[0].pageX+U[1].pageX),ht=.5*(U[0].pageY+U[1].pageY);y.set(Fe,ht)}}function pe(){if(U.length==1)_.set(U[0].pageX,U[0].pageY);else{const Fe=.5*(U[0].pageX+U[1].pageX),ht=.5*(U[0].pageY+U[1].pageY);_.set(Fe,ht)}}function ve(){const Fe=U[0].pageX-U[1].pageX,ht=U[0].pageY-U[1].pageY,kt=Math.sqrt(Fe*Fe+ht*ht);k.set(0,kt)}function we(){r.enableZoom&&ve(),r.enablePan&&pe()}function Me(){r.enableZoom&&ve(),r.enableRotate&&se()}function Re(Fe){if(U.length==1)v.set(Fe.pageX,Fe.pageY);else{const kt=nn(Fe),Qt=.5*(Fe.pageX+kt.x),en=.5*(Fe.pageY+kt.y);v.set(Qt,en)}S.subVectors(v,y).multiplyScalar(r.rotateSpeed);const ht=r.domElement;ht&&(j(2*Math.PI*S.x/ht.clientHeight),W(2*Math.PI*S.y/ht.clientHeight)),y.copy(v)}function et(Fe){if(U.length==1)E.set(Fe.pageX,Fe.pageY);else{const ht=nn(Fe),kt=.5*(Fe.pageX+ht.x),Qt=.5*(Fe.pageY+ht.y);E.set(kt,Qt)}$.subVectors(E,_).multiplyScalar(r.panSpeed),te($.x,$.y),_.copy(E)}function tt(Fe){const ht=nn(Fe),kt=Fe.pageX-ht.x,Qt=Fe.pageY-ht.y,en=Math.sqrt(kt*kt+Qt*Qt);A.set(0,en),w.set(0,Math.pow(A.y/k.y,r.zoomSpeed)),M(w.y),k.copy(A)}function Oe(Fe){r.enableZoom&&tt(Fe),r.enablePan&&et(Fe)}function $e(Fe){r.enableZoom&&tt(Fe),r.enableRotate&&Re(Fe)}function Ae(Fe){var ht,kt;r.enabled!==!1&&(U.length===0&&((ht=r.domElement)==null||ht.ownerDocument.addEventListener("pointermove",He),(kt=r.domElement)==null||kt.ownerDocument.addEventListener("pointerup",je)),zt(Fe),Fe.pointerType==="touch"?mt(Fe):Ye(Fe))}function He(Fe){r.enabled!==!1&&(Fe.pointerType==="touch"?Mt(Fe):Ne(Fe))}function je(Fe){var ht,kt,Qt;qt(Fe),U.length===0&&((ht=r.domElement)==null||ht.releasePointerCapture(Fe.pointerId),(kt=r.domElement)==null||kt.ownerDocument.removeEventListener("pointermove",He),(Qt=r.domElement)==null||Qt.ownerDocument.removeEventListener("pointerup",je)),r.dispatchEvent(u),o=l.NONE}function Ye(Fe){let ht;switch(Fe.button){case 0:ht=r.mouseButtons.LEFT;break;case 1:ht=r.mouseButtons.MIDDLE;break;case 2:ht=r.mouseButtons.RIGHT;break;default:ht=-1}switch(ht){case gO.DOLLY:if(r.enableZoom===!1)return;oe(Fe),o=l.DOLLY;break;case gO.ROTATE:if(Fe.ctrlKey||Fe.metaKey||Fe.shiftKey){if(r.enablePan===!1)return;ie(Fe),o=l.PAN}else{if(r.enableRotate===!1)return;ne(Fe),o=l.ROTATE}break;case gO.PAN:if(Fe.ctrlKey||Fe.metaKey||Fe.shiftKey){if(r.enableRotate===!1)return;ne(Fe),o=l.ROTATE}else{if(r.enablePan===!1)return;ie(Fe),o=l.PAN}break;default:o=l.NONE}o!==l.NONE&&r.dispatchEvent(a)}function Ne(Fe){if(r.enabled!==!1)switch(o){case l.ROTATE:if(r.enableRotate===!1)return;de(Fe);break;case l.DOLLY:if(r.enableZoom===!1)return;ge(Fe);break;case l.PAN:if(r.enablePan===!1)return;ye(Fe);break}}function ze(Fe){r.enabled===!1||r.enableZoom===!1||o!==l.NONE&&o!==l.ROTATE||(Fe.preventDefault(),r.dispatchEvent(a),me(Fe),r.dispatchEvent(u))}function at(Fe){r.enabled===!1||r.enablePan===!1||Ee(Fe)}function mt(Fe){switch(Xt(Fe),U.length){case 1:switch(r.touches.ONE){case vO.ROTATE:if(r.enableRotate===!1)return;se(),o=l.TOUCH_ROTATE;break;case vO.PAN:if(r.enablePan===!1)return;pe(),o=l.TOUCH_PAN;break;default:o=l.NONE}break;case 2:switch(r.touches.TWO){case vO.DOLLY_PAN:if(r.enableZoom===!1&&r.enablePan===!1)return;we(),o=l.TOUCH_DOLLY_PAN;break;case vO.DOLLY_ROTATE:if(r.enableZoom===!1&&r.enableRotate===!1)return;Me(),o=l.TOUCH_DOLLY_ROTATE;break;default:o=l.NONE}break;default:o=l.NONE}o!==l.NONE&&r.dispatchEvent(a)}function Mt(Fe){switch(Xt(Fe),o){case l.TOUCH_ROTATE:if(r.enableRotate===!1)return;Re(Fe),r.update();break;case l.TOUCH_PAN:if(r.enablePan===!1)return;et(Fe),r.update();break;case l.TOUCH_DOLLY_PAN:if(r.enableZoom===!1&&r.enablePan===!1)return;Oe(Fe),r.update();break;case l.TOUCH_DOLLY_ROTATE:if(r.enableZoom===!1&&r.enableRotate===!1)return;$e(Fe),r.update();break;default:o=l.NONE}}function bt(Fe){r.enabled!==!1&&Fe.preventDefault()}function zt(Fe){U.push(Fe)}function qt(Fe){delete q[Fe.pointerId];for(let ht=0;ht<U.length;ht++)if(U[ht].pointerId==Fe.pointerId){U.splice(ht,1);return}}function Xt(Fe){let ht=q[Fe.pointerId];ht===void 0&&(ht=new ao,q[Fe.pointerId]=ht),ht.set(Fe.pageX,Fe.pageY)}function nn(Fe){const ht=Fe.pointerId===U[0].pointerId?U[1]:U[0];return q[ht.pointerId]}this.dollyIn=(Fe=N())=>{G(Fe),r.update()},this.dollyOut=(Fe=N())=>{M(Fe),r.update()},this.getScale=()=>h,this.setScale=Fe=>{H(Fe),r.update()},this.getZoomScale=()=>N(),n!==void 0&&this.connect(n),this.update()}};const uJ=new FF,VE=new ci;class NB extends AOe{constructor(){super(),this.isLineSegmentsGeometry=!0,this.type="LineSegmentsGeometry";const t=[-1,2,0,1,2,0,-1,1,0,1,1,0,-1,0,0,1,0,0,-1,-1,0,1,-1,0],n=[-1,2,1,2,-1,1,1,1,-1,-1,1,-1,-1,-2,1,-2],r=[0,2,1,2,3,1,2,4,3,4,5,3,4,6,5,6,7,5];this.setIndex(r),this.setAttribute("position",new e9(t,3)),this.setAttribute("uv",new e9(n,2))}applyMatrix4(t){const n=this.attributes.instanceStart,r=this.attributes.instanceEnd;return n!==void 0&&(n.applyMatrix4(t),r.applyMatrix4(t),n.needsUpdate=!0),this.boundingBox!==null&&this.computeBoundingBox(),this.boundingSphere!==null&&this.computeBoundingSphere(),this}setPositions(t){let n;t instanceof Float32Array?n=t:Array.isArray(t)&&(n=new Float32Array(t));const r=new t9(n,6,1);return this.setAttribute("instanceStart",new VO(r,3,0)),this.setAttribute("instanceEnd",new VO(r,3,3)),this.computeBoundingBox(),this.computeBoundingSphere(),this}setColors(t,n=3){let r;t instanceof Float32Array?r=t:Array.isArray(t)&&(r=new Float32Array(t));const i=new t9(r,n*2,1);return this.setAttribute("instanceColorStart",new VO(i,n,0)),this.setAttribute("instanceColorEnd",new VO(i,n,n)),this}fromWireframeGeometry(t){return this.setPositions(t.attributes.position.array),this}fromEdgesGeometry(t){return this.setPositions(t.attributes.position.array),this}fromMesh(t){return this.fromWireframeGeometry(new TOe(t.geometry)),this}fromLineSegments(t){const n=t.geometry;return this.setPositions(n.attributes.position.array),this}computeBoundingBox(){this.boundingBox===null&&(this.boundingBox=new FF);const t=this.attributes.instanceStart,n=this.attributes.instanceEnd;t!==void 0&&n!==void 0&&(this.boundingBox.setFromBufferAttribute(t),uJ.setFromBufferAttribute(n),this.boundingBox.union(uJ))}computeBoundingSphere(){this.boundingSphere===null&&(this.boundingSphere=new pJ),this.boundingBox===null&&this.computeBoundingBox();const t=this.attributes.instanceStart,n=this.attributes.instanceEnd;if(t!==void 0&&n!==void 0){const r=this.boundingSphere.center;this.boundingBox.getCenter(r);let i=0;for(let a=0,u=t.count;a<u;a++)VE.fromBufferAttribute(t,a),i=Math.max(i,r.distanceToSquared(VE)),VE.fromBufferAttribute(n,a),i=Math.max(i,r.distanceToSquared(VE));this.boundingSphere.radius=Math.sqrt(i),isNaN(this.boundingSphere.radius)&&console.error("THREE.LineSegmentsGeometry.computeBoundingSphere(): Computed radius is NaN. The instanced position data is likely to have NaN values.",this)}}toJSON(){}applyMatrix(t){return console.warn("THREE.LineSegmentsGeometry: applyMatrix() has been renamed to applyMatrix4()."),this.applyMatrix4(t)}}class eOe extends NB{constructor(){super(),this.isLineGeometry=!0,this.type="LineGeometry"}setPositions(t){const n=t.length-3,r=new Float32Array(2*n);for(let i=0;i<n;i+=3)r[2*i]=t[i],r[2*i+1]=t[i+1],r[2*i+2]=t[i+2],r[2*i+3]=t[i+3],r[2*i+4]=t[i+4],r[2*i+5]=t[i+5];return super.setPositions(r),this}setColors(t,n=3){const r=t.length-n,i=new Float32Array(2*r);if(n===3)for(let a=0;a<r;a+=n)i[2*a]=t[a],i[2*a+1]=t[a+1],i[2*a+2]=t[a+2],i[2*a+3]=t[a+3],i[2*a+4]=t[a+4],i[2*a+5]=t[a+5];else for(let a=0;a<r;a+=n)i[2*a]=t[a],i[2*a+1]=t[a+1],i[2*a+2]=t[a+2],i[2*a+3]=t[a+3],i[2*a+4]=t[a+4],i[2*a+5]=t[a+5],i[2*a+6]=t[a+6],i[2*a+7]=t[a+7];return super.setColors(i,n),this}fromLine(t){const n=t.geometry;return this.setPositions(n.attributes.position.array),this}}class IB extends kOe{constructor(t){super({type:"LineMaterial",uniforms:VB.clone(VB.merge([UB.common,UB.fog,{worldUnits:{value:1},linewidth:{value:1},resolution:{value:new ao(1,1)},dashOffset:{value:0},dashScale:{value:1},dashSize:{value:1},gapSize:{value:1}}])),vertexShader:`
				#include <common>
				#include <fog_pars_vertex>
				#include <logdepthbuf_pars_vertex>
				#include <clipping_planes_pars_vertex>

				uniform float linewidth;
				uniform vec2 resolution;

				attribute vec3 instanceStart;
				attribute vec3 instanceEnd;

				#ifdef USE_COLOR
					#ifdef USE_LINE_COLOR_ALPHA
						varying vec4 vLineColor;
						attribute vec4 instanceColorStart;
						attribute vec4 instanceColorEnd;
					#else
						varying vec3 vLineColor;
						attribute vec3 instanceColorStart;
						attribute vec3 instanceColorEnd;
					#endif
				#endif

				#ifdef WORLD_UNITS

					varying vec4 worldPos;
					varying vec3 worldStart;
					varying vec3 worldEnd;

					#ifdef USE_DASH

						varying vec2 vUv;

					#endif

				#else

					varying vec2 vUv;

				#endif

				#ifdef USE_DASH

					uniform float dashScale;
					attribute float instanceDistanceStart;
					attribute float instanceDistanceEnd;
					varying float vLineDistance;

				#endif

				void trimSegment( const in vec4 start, inout vec4 end ) {

					// trim end segment so it terminates between the camera plane and the near plane

					// conservative estimate of the near plane
					float a = projectionMatrix[ 2 ][ 2 ]; // 3nd entry in 3th column
					float b = projectionMatrix[ 3 ][ 2 ]; // 3nd entry in 4th column
					float nearEstimate = - 0.5 * b / a;

					float alpha = ( nearEstimate - start.z ) / ( end.z - start.z );

					end.xyz = mix( start.xyz, end.xyz, alpha );

				}

				void main() {

					#ifdef USE_COLOR

						vLineColor = ( position.y < 0.5 ) ? instanceColorStart : instanceColorEnd;

					#endif

					#ifdef USE_DASH

						vLineDistance = ( position.y < 0.5 ) ? dashScale * instanceDistanceStart : dashScale * instanceDistanceEnd;
						vUv = uv;

					#endif

					float aspect = resolution.x / resolution.y;

					// camera space
					vec4 start = modelViewMatrix * vec4( instanceStart, 1.0 );
					vec4 end = modelViewMatrix * vec4( instanceEnd, 1.0 );

					#ifdef WORLD_UNITS

						worldStart = start.xyz;
						worldEnd = end.xyz;

					#else

						vUv = uv;

					#endif

					// special case for perspective projection, and segments that terminate either in, or behind, the camera plane
					// clearly the gpu firmware has a way of addressing this issue when projecting into ndc space
					// but we need to perform ndc-space calculations in the shader, so we must address this issue directly
					// perhaps there is a more elegant solution -- WestLangley

					bool perspective = ( projectionMatrix[ 2 ][ 3 ] == - 1.0 ); // 4th entry in the 3rd column

					if ( perspective ) {

						if ( start.z < 0.0 && end.z >= 0.0 ) {

							trimSegment( start, end );

						} else if ( end.z < 0.0 && start.z >= 0.0 ) {

							trimSegment( end, start );

						}

					}

					// clip space
					vec4 clipStart = projectionMatrix * start;
					vec4 clipEnd = projectionMatrix * end;

					// ndc space
					vec3 ndcStart = clipStart.xyz / clipStart.w;
					vec3 ndcEnd = clipEnd.xyz / clipEnd.w;

					// direction
					vec2 dir = ndcEnd.xy - ndcStart.xy;

					// account for clip-space aspect ratio
					dir.x *= aspect;
					dir = normalize( dir );

					#ifdef WORLD_UNITS

						// get the offset direction as perpendicular to the view vector
						vec3 worldDir = normalize( end.xyz - start.xyz );
						vec3 offset;
						if ( position.y < 0.5 ) {

							offset = normalize( cross( start.xyz, worldDir ) );

						} else {

							offset = normalize( cross( end.xyz, worldDir ) );

						}

						// sign flip
						if ( position.x < 0.0 ) offset *= - 1.0;

						float forwardOffset = dot( worldDir, vec3( 0.0, 0.0, 1.0 ) );

						// don't extend the line if we're rendering dashes because we
						// won't be rendering the endcaps
						#ifndef USE_DASH

							// extend the line bounds to encompass  endcaps
							start.xyz += - worldDir * linewidth * 0.5;
							end.xyz += worldDir * linewidth * 0.5;

							// shift the position of the quad so it hugs the forward edge of the line
							offset.xy -= dir * forwardOffset;
							offset.z += 0.5;

						#endif

						// endcaps
						if ( position.y > 1.0 || position.y < 0.0 ) {

							offset.xy += dir * 2.0 * forwardOffset;

						}

						// adjust for linewidth
						offset *= linewidth * 0.5;

						// set the world position
						worldPos = ( position.y < 0.5 ) ? start : end;
						worldPos.xyz += offset;

						// project the worldpos
						vec4 clip = projectionMatrix * worldPos;

						// shift the depth of the projected points so the line
						// segments overlap neatly
						vec3 clipPose = ( position.y < 0.5 ) ? ndcStart : ndcEnd;
						clip.z = clipPose.z * clip.w;

					#else

						vec2 offset = vec2( dir.y, - dir.x );
						// undo aspect ratio adjustment
						dir.x /= aspect;
						offset.x /= aspect;

						// sign flip
						if ( position.x < 0.0 ) offset *= - 1.0;

						// endcaps
						if ( position.y < 0.0 ) {

							offset += - dir;

						} else if ( position.y > 1.0 ) {

							offset += dir;

						}

						// adjust for linewidth
						offset *= linewidth;

						// adjust for clip-space to screen-space conversion // maybe resolution should be based on viewport ...
						offset /= resolution.y;

						// select end
						vec4 clip = ( position.y < 0.5 ) ? clipStart : clipEnd;

						// back to clip space
						offset *= clip.w;

						clip.xy += offset;

					#endif

					gl_Position = clip;

					vec4 mvPosition = ( position.y < 0.5 ) ? start : end; // this is an approximation

					#include <logdepthbuf_vertex>
					#include <clipping_planes_vertex>
					#include <fog_vertex>

				}
			`,fragmentShader:`
				uniform vec3 diffuse;
				uniform float opacity;
				uniform float linewidth;

				#ifdef USE_DASH

					uniform float dashOffset;
					uniform float dashSize;
					uniform float gapSize;

				#endif

				varying float vLineDistance;

				#ifdef WORLD_UNITS

					varying vec4 worldPos;
					varying vec3 worldStart;
					varying vec3 worldEnd;

					#ifdef USE_DASH

						varying vec2 vUv;

					#endif

				#else

					varying vec2 vUv;

				#endif

				#include <common>
				#include <fog_pars_fragment>
				#include <logdepthbuf_pars_fragment>
				#include <clipping_planes_pars_fragment>

				#ifdef USE_COLOR
					#ifdef USE_LINE_COLOR_ALPHA
						varying vec4 vLineColor;
					#else
						varying vec3 vLineColor;
					#endif
				#endif

				vec2 closestLineToLine(vec3 p1, vec3 p2, vec3 p3, vec3 p4) {

					float mua;
					float mub;

					vec3 p13 = p1 - p3;
					vec3 p43 = p4 - p3;

					vec3 p21 = p2 - p1;

					float d1343 = dot( p13, p43 );
					float d4321 = dot( p43, p21 );
					float d1321 = dot( p13, p21 );
					float d4343 = dot( p43, p43 );
					float d2121 = dot( p21, p21 );

					float denom = d2121 * d4343 - d4321 * d4321;

					float numer = d1343 * d4321 - d1321 * d4343;

					mua = numer / denom;
					mua = clamp( mua, 0.0, 1.0 );
					mub = ( d1343 + d4321 * ( mua ) ) / d4343;
					mub = clamp( mub, 0.0, 1.0 );

					return vec2( mua, mub );

				}

				void main() {

					#include <clipping_planes_fragment>

					#ifdef USE_DASH

						if ( vUv.y < - 1.0 || vUv.y > 1.0 ) discard; // discard endcaps

						if ( mod( vLineDistance + dashOffset, dashSize + gapSize ) > dashSize ) discard; // todo - FIX

					#endif

					float alpha = opacity;

					#ifdef WORLD_UNITS

						// Find the closest points on the view ray and the line segment
						vec3 rayEnd = normalize( worldPos.xyz ) * 1e5;
						vec3 lineDir = worldEnd - worldStart;
						vec2 params = closestLineToLine( worldStart, worldEnd, vec3( 0.0, 0.0, 0.0 ), rayEnd );

						vec3 p1 = worldStart + lineDir * params.x;
						vec3 p2 = rayEnd * params.y;
						vec3 delta = p1 - p2;
						float len = length( delta );
						float norm = len / linewidth;

						#ifndef USE_DASH

							#ifdef USE_ALPHA_TO_COVERAGE

								float dnorm = fwidth( norm );
								alpha = 1.0 - smoothstep( 0.5 - dnorm, 0.5 + dnorm, norm );

							#else

								if ( norm > 0.5 ) {

									discard;

								}

							#endif

						#endif

					#else

						#ifdef USE_ALPHA_TO_COVERAGE

							// artifacts appear on some hardware if a derivative is taken within a conditional
							float a = vUv.x;
							float b = ( vUv.y > 0.0 ) ? vUv.y - 1.0 : vUv.y + 1.0;
							float len2 = a * a + b * b;
							float dlen = fwidth( len2 );

							if ( abs( vUv.y ) > 1.0 ) {

								alpha = 1.0 - smoothstep( 1.0 - dlen, 1.0 + dlen, len2 );

							}

						#else

							if ( abs( vUv.y ) > 1.0 ) {

								float a = vUv.x;
								float b = ( vUv.y > 0.0 ) ? vUv.y - 1.0 : vUv.y + 1.0;
								float len2 = a * a + b * b;

								if ( len2 > 1.0 ) discard;

							}

						#endif

					#endif

					vec4 diffuseColor = vec4( diffuse, alpha );
					#ifdef USE_COLOR
						#ifdef USE_LINE_COLOR_ALPHA
							diffuseColor *= vLineColor;
						#else
							diffuseColor.rgb *= vLineColor;
						#endif
					#endif

					#include <logdepthbuf_fragment>

					gl_FragColor = diffuseColor;

					#include <tonemapping_fragment>
					#include <${W1e>=154?"colorspace_fragment":"encodings_fragment"}>
					#include <fog_fragment>
					#include <premultiplied_alpha_fragment>

				}
			`,clipping:!0}),this.isLineMaterial=!0,this.onBeforeCompile=function(){this.transparent?this.defines.USE_LINE_COLOR_ALPHA="1":delete this.defines.USE_LINE_COLOR_ALPHA},Object.defineProperties(this,{color:{enumerable:!0,get:function(){return this.uniforms.diffuse.value},set:function(n){this.uniforms.diffuse.value=n}},worldUnits:{enumerable:!0,get:function(){return"WORLD_UNITS"in this.defines},set:function(n){n===!0?this.defines.WORLD_UNITS="":delete this.defines.WORLD_UNITS}},linewidth:{enumerable:!0,get:function(){return this.uniforms.linewidth.value},set:function(n){this.uniforms.linewidth.value=n}},dashed:{enumerable:!0,get:function(){return"USE_DASH"in this.defines},set(n){!!n!="USE_DASH"in this.defines&&(this.needsUpdate=!0),n===!0?this.defines.USE_DASH="":delete this.defines.USE_DASH}},dashScale:{enumerable:!0,get:function(){return this.uniforms.dashScale.value},set:function(n){this.uniforms.dashScale.value=n}},dashSize:{enumerable:!0,get:function(){return this.uniforms.dashSize.value},set:function(n){this.uniforms.dashSize.value=n}},dashOffset:{enumerable:!0,get:function(){return this.uniforms.dashOffset.value},set:function(n){this.uniforms.dashOffset.value=n}},gapSize:{enumerable:!0,get:function(){return this.uniforms.gapSize.value},set:function(n){this.uniforms.gapSize.value=n}},opacity:{enumerable:!0,get:function(){return this.uniforms.opacity.value},set:function(n){this.uniforms.opacity.value=n}},resolution:{enumerable:!0,get:function(){return this.uniforms.resolution.value},set:function(n){this.uniforms.resolution.value.copy(n)}},alphaToCoverage:{enumerable:!0,get:function(){return"USE_ALPHA_TO_COVERAGE"in this.defines},set:function(n){!!n!="USE_ALPHA_TO_COVERAGE"in this.defines&&(this.needsUpdate=!0),n===!0?(this.defines.USE_ALPHA_TO_COVERAGE="",this.extensions.derivatives=!0):(delete this.defines.USE_ALPHA_TO_COVERAGE,this.extensions.derivatives=!1)}}}),this.setValues(t)}}const G7=new j3,lJ=new ci,oJ=new ci,Po=new j3,Mo=new j3,o0=new j3,K7=new ci,W7=new FOe,No=new MOe,sJ=new ci,UE=new FF,jE=new pJ,s0=new j3;let v0,Jb;function cJ(e,t,n){return s0.set(0,0,-t,1).applyMatrix4(e.projectionMatrix),s0.multiplyScalar(1/s0.w),s0.x=Jb/n.width,s0.y=Jb/n.height,s0.applyMatrix4(e.projectionMatrixInverse),s0.multiplyScalar(1/s0.w),Math.abs(Math.max(s0.x,s0.y))}function Nnt(e,t){const n=e.matrixWorld,r=e.geometry,i=r.attributes.instanceStart,a=r.attributes.instanceEnd,u=Math.min(r.instanceCount,i.count);for(let l=0,o=u;l<o;l++){No.start.fromBufferAttribute(i,l),No.end.fromBufferAttribute(a,l),No.applyMatrix4(n);const s=new ci,c=new ci;v0.distanceSqToSegment(No.start,No.end,c,s),c.distanceTo(s)<Jb*.5&&t.push({point:c,pointOnLine:s,distance:v0.origin.distanceTo(c),object:e,face:null,faceIndex:l,uv:null,[J1e]:null})}}function Int(e,t,n){const r=t.projectionMatrix,a=e.material.resolution,u=e.matrixWorld,l=e.geometry,o=l.attributes.instanceStart,s=l.attributes.instanceEnd,c=Math.min(l.instanceCount,o.count),d=-t.near;v0.at(1,o0),o0.w=1,o0.applyMatrix4(t.matrixWorldInverse),o0.applyMatrix4(r),o0.multiplyScalar(1/o0.w),o0.x*=a.x/2,o0.y*=a.y/2,o0.z=0,K7.copy(o0),W7.multiplyMatrices(t.matrixWorldInverse,u);for(let h=0,m=c;h<m;h++){if(Po.fromBufferAttribute(o,h),Mo.fromBufferAttribute(s,h),Po.w=1,Mo.w=1,Po.applyMatrix4(W7),Mo.applyMatrix4(W7),Po.z>d&&Mo.z>d)continue;if(Po.z>d){const $=Po.z-Mo.z,k=(Po.z-d)/$;Po.lerp(Mo,k)}else if(Mo.z>d){const $=Mo.z-Po.z,k=(Mo.z-d)/$;Mo.lerp(Po,k)}Po.applyMatrix4(r),Mo.applyMatrix4(r),Po.multiplyScalar(1/Po.w),Mo.multiplyScalar(1/Mo.w),Po.x*=a.x/2,Po.y*=a.y/2,Mo.x*=a.x/2,Mo.y*=a.y/2,No.start.copy(Po),No.start.z=0,No.end.copy(Mo),No.end.z=0;const v=No.closestPointToPointParameter(K7,!0);No.at(v,sJ);const S=mJ.lerp(Po.z,Mo.z,v),_=S>=-1&&S<=1,E=K7.distanceTo(sJ)<Jb*.5;if(_&&E){No.start.fromBufferAttribute(o,h),No.end.fromBufferAttribute(s,h),No.start.applyMatrix4(u),No.end.applyMatrix4(u);const $=new ci,k=new ci;v0.distanceSqToSegment(No.start,No.end,k,$),n.push({point:k,pointOnLine:$,distance:v0.origin.distanceTo(k),object:e,face:null,faceIndex:h,uv:null,[J1e]:null})}}}class tOe extends POe{constructor(t=new NB,n=new IB({color:Math.random()*16777215})){super(t,n),this.isLineSegments2=!0,this.type="LineSegments2"}computeLineDistances(){const t=this.geometry,n=t.attributes.instanceStart,r=t.attributes.instanceEnd,i=new Float32Array(2*n.count);for(let u=0,l=0,o=n.count;u<o;u++,l+=2)lJ.fromBufferAttribute(n,u),oJ.fromBufferAttribute(r,u),i[l]=l===0?0:i[l-1],i[l+1]=i[l]+lJ.distanceTo(oJ);const a=new t9(i,2,1);return t.setAttribute("instanceDistanceStart",new VO(a,1,0)),t.setAttribute("instanceDistanceEnd",new VO(a,1,1)),this}raycast(t,n){const r=this.material.worldUnits,i=t.camera;i===null&&!r&&console.error('LineSegments2: "Raycaster.camera" needs to be set in order to raycast against LineSegments2 while worldUnits is set to false.');const a=t.params.Line2!==void 0&&t.params.Line2.threshold||0;v0=t.ray;const u=this.matrixWorld,l=this.geometry,o=this.material;Jb=o.linewidth+a,l.boundingSphere===null&&l.computeBoundingSphere(),jE.copy(l.boundingSphere).applyMatrix4(u);let s;if(r)s=Jb*.5;else{const d=Math.max(i.near,jE.distanceToPoint(v0.origin));s=cJ(i,d,o.resolution)}if(jE.radius+=s,v0.intersectsSphere(jE)===!1)return;l.boundingBox===null&&l.computeBoundingBox(),UE.copy(l.boundingBox).applyMatrix4(u);let c;if(r)c=Jb*.5;else{const d=Math.max(i.near,UE.distanceToPoint(v0.origin));c=cJ(i,d,o.resolution)}UE.expandByScalar(c),v0.intersectsBox(UE)!==!1&&(r?Nnt(this,n):Int(this,i,n))}onBeforeRender(t){const n=this.material.uniforms;n&&n.resolution&&(t.getViewport(G7),this.material.uniforms.resolution.value.set(G7.z,G7.w))}}class Lnt extends tOe{constructor(t=new eOe,n=new IB({color:Math.random()*16777215})){super(t,n),this.isLine2=!0,this.type="Line2"}}const Bnt=P.forwardRef(function({points:t,color:n=16777215,vertexColors:r,linewidth:i,lineWidth:a,segments:u,dashed:l,...o},s){var c,d;const h=Vd(_=>_.size),m=P.useMemo(()=>u?new tOe:new Lnt,[u]),[y]=P.useState(()=>new IB),v=(r==null||(c=r[0])==null?void 0:c.length)===4?4:3,S=P.useMemo(()=>{const _=u?new NB:new eOe,E=t.map($=>{const k=Array.isArray($);return $ instanceof ci||$ instanceof j3?[$.x,$.y,$.z]:$ instanceof ao?[$.x,$.y,0]:k&&$.length===3?[$[0],$[1],$[2]]:k&&$.length===2?[$[0],$[1],0]:$});if(_.setPositions(E.flat()),r){n=16777215;const $=r.map(k=>k instanceof OC?k.toArray():k);_.setColors($.flat(),v)}return _},[t,u,r,v]);return P.useLayoutEffect(()=>{m.computeLineDistances()},[t,m]),P.useLayoutEffect(()=>{l?y.defines.USE_DASH="":delete y.defines.USE_DASH,y.needsUpdate=!0},[l,y]),P.useEffect(()=>()=>{S.dispose(),y.dispose()},[S]),P.createElement("primitive",ry({object:m,ref:s},o),P.createElement("primitive",{object:S,attach:"geometry"}),P.createElement("primitive",ry({object:y,attach:"material",color:n,vertexColors:!!r,resolution:[h.width,h.height],linewidth:(d=i??a)!==null&&d!==void 0?d:1,dashed:l,transparent:v===4},o)))}),Qnt=new ci,Blt=P.forwardRef(function({start:t=[0,0,0],end:n=[0,0,0],mid:r,segments:i=20,...a},u){const l=P.useRef(null);P.useImperativeHandle(u,()=>l.current);const[o]=P.useState(()=>new NOe(void 0,void 0,void 0)),s=P.useCallback((d,h,m,y=20)=>(d instanceof ci?o.v0.copy(d):o.v0.set(...d),h instanceof ci?o.v2.copy(h):o.v2.set(...h),m instanceof ci?o.v1.copy(m):Array.isArray(m)?o.v1.set(...m):o.v1.copy(o.v0.clone().add(o.v2.clone().sub(o.v0)).add(Qnt.set(0,o.v0.y-o.v2.y,0))),o.getPoints(y)),[]);P.useLayoutEffect(()=>{l.current.setPoints=(d,h,m)=>{const y=s(d,h,m);l.current.geometry&&l.current.geometry.setPositions(y.map(v=>v.toArray()).flat())}},[]);const c=P.useMemo(()=>s(t,n,r,i),[t,n,r,i]);return P.createElement(Bnt,ry({ref:l,points:c},a))}),znt=P.forwardRef(({makeDefault:e,camera:t,regress:n,domElement:r,enableDamping:i=!0,keyEvents:a=!1,onChange:u,onStart:l,onEnd:o,...s},c)=>{const d=Vd(w=>w.invalidate),h=Vd(w=>w.camera),m=Vd(w=>w.gl),y=Vd(w=>w.events),v=Vd(w=>w.setEvents),S=Vd(w=>w.set),_=Vd(w=>w.get),E=Vd(w=>w.performance),$=t||h,k=r||y.connected||m.domElement,A=P.useMemo(()=>new Fnt($),[$]);return x5(()=>{A.enabled&&A.update()},-1),P.useEffect(()=>(a&&A.connect(a===!0?k:a),A.connect(k),()=>void A.dispose()),[a,k,n,A,d]),P.useEffect(()=>{const w=I=>{d(),n&&E.regress(),u&&u(I)},R=I=>{l&&l(I)},F=I=>{o&&o(I)};return A.addEventListener("change",w),A.addEventListener("start",R),A.addEventListener("end",F),()=>{A.removeEventListener("start",R),A.removeEventListener("end",F),A.removeEventListener("change",w)}},[u,l,o,A,d,v]),P.useEffect(()=>{if(e){const w=_().controls;return S({controls:A}),()=>S({controls:w})}},[e,A]),P.createElement("primitive",ry({ref:c,object:A,enableDamping:i},s))});function Qlt(...e){const t=P.useRef([]);return t.current=e.map(n=>P.useContext(n)),P.useMemo(()=>({children:n})=>e.reduceRight((r,i,a)=>P.createElement(i.Provider,{value:t.current[a],children:r}),n),[])}function qnt(e,t){const n=e+"Geometry";return P.forwardRef(({args:r,children:i,...a},u)=>{const l=P.useRef(null);return P.useImperativeHandle(u,()=>l.current),P.useLayoutEffect(()=>void(t==null?void 0:t(l.current))),P.createElement("mesh",ry({ref:l},a),P.createElement(n,{attach:"geometry",args:r}),i)})}const zlt=qnt("sphere");function Vnt(e,t){const n=e[0],r=e[1];let i=!1;for(let a=0,u=t.length-1;a<t.length;u=a++){const l=t[a][0],o=t[a][1],s=t[u][0],c=t[u][1];o>r!=c>r&&n<(s-l)*(r-o)/(c-o)+l&&(i=!i)}return i}const Unt={minX:0,maxX:0,minY:0,maxY:0,minZ:0,maxZ:0};function qlt(e,t){if(e.length===0)return(t==null?void 0:t.fallbackBounds)||Unt;let n=1/0,r=1/0,i=-1/0,a=-1/0,u=1/0,l=-1/0;return e.forEach(o=>{n=Math.min(n,o[0]),r=Math.min(r,o[1]),u=Math.min(u,o[2]),i=Math.max(i,o[0]),a=Math.max(a,o[1]),l=Math.max(l,o[2])}),{minX:n,maxX:i,minY:r,maxY:a,minZ:u,maxZ:l}}function jnt({minX:e,maxX:t,minY:n,maxY:r,minZ:i,maxZ:a}){return[e+(t-e)/2,n+(r-n)/2,i+(a-i)/2]}function Znt({minX:e,maxX:t,minY:n,maxY:r,minZ:i,maxZ:a}){return Math.max(Math.abs(e),Math.abs(t),Math.abs(n),Math.abs(r),Math.abs(i),Math.abs(a))}function nOe(e){return e.isOrthographicCamera}P.createContext({bounds:{minX:-10,maxX:10,minY:-10,maxY:10},center:[0,0]});function Vlt({children:e,camera:t,orthographic:n=!1,...r}){return Qr.jsx(vnt,{orthographic:n,camera:{position:[0,0,1],zoom:1,up:[0,1,0],...t},...r,children:e})}function Ult({enabled:e=!0,zoomSpeed:t=1,enableZoom:n=!0,enablePan:r=!0,enableRotate:i=!0,autoRotate:a=!1,autoRotateSpeed:u=5,...l}){return Qr.jsx(znt,{enabled:e,zoomSpeed:t,enableZoom:n,enablePan:r,enableRotate:i,autoRotate:a,autoRotateSpeed:u,...l})}const Hnt=P.createContext({bounds:{minX:-10,maxX:10,minY:-10,maxY:10,minZ:-10,maxZ:10},center:[0,0,0]}),fJ=10;function jlt({bounds:e,children:t,offset:n=1.5,boundsZoomPaddingFactor:r=.9}){const i=jnt(e),{camera:a,size:{width:u,height:l}}=Vd();return P.useEffect(()=>{const{minX:o,maxX:s,minY:c,maxY:d,minZ:h,maxZ:m}=e,y=s-o,v=d-c,S=m-h,_=Math.max(y,v,S);if(a.position.x=y/2+o,a.position.y=v/2+c,nOe(a)){a.zoom=Math.min(u/y,l/v)*r;const E=Znt(e),$=Math.abs(E)+fJ;a.position.z=$;const k=Math.max(a.far,$+E+fJ);a.far=k}else{const E=a.fov*(Math.PI/180),$=_/(2*Math.tan(E/2));a.position.z=($+i[2])*n;const k=h<0?-h+a.position.z:a.position.z-h;a.far=k*3}a.updateProjectionMatrix(),a.lookAt(...i)},[e]),Qr.jsx(Hnt.Provider,{value:{bounds:e,center:i},children:t})}const Zlt=({size:e,color:t="#ffffff"})=>{const n=P.useRef(null);return P.useEffect(()=>{if(n.current){const r=new OC(t);n.current.setColors(r,r,r)}},[t]),Qr.jsx("axesHelper",{ref:n,args:[e]})},Xnt={color:"blue",scale:1},R2=new BOe,J7=new OC;function Hlt({data:e,pointProps:t=Xnt,onPointsClicked:n,onPointClicked:r,onPointHovered:i,onPointerLeave:a,pointShape:u="sphere",opacity:l=1,material:o="meshMatcap"}){const s=P.useCallback(y=>typeof t.color=="function"?t.color(y):t.color,[t.color]),c=P.useMemo(()=>Float32Array.from(new Array(e.length).fill(null).flatMap((y,v)=>J7.set(s(e[v])).toArray())),[e]),d=P.useRef(null);x5(()=>{e.forEach(({position:y},v)=>{if(R2.position.set(y[0],y[1],y[2]||0),t.scale&&R2.scale.set(t.scale,t.scale,t.scale),d.current){d.current.setMatrixAt(v,R2.matrix),d.current.instanceMatrix.needsUpdate=!0;const S=s(e[v]);J7.set(S),J7.toArray(c,v*3),d.current.geometry.attributes.color.needsUpdate=!0,R2.updateMatrix(),d.current.setMatrixAt(v,R2.matrix)}})});const h=P.useMemo(()=>{switch(u){case"sphere":return Qr.jsx("sphereGeometry",{args:[t.radius||.02,20,20],children:Qr.jsx("instancedBufferAttribute",{attach:"attributes-color",args:[c,3]})});case"cube":{const y=typeof(t==null?void 0:t.size)=="number"?[t.size,t.size,t.size]:[.03,.03,.03];return Qr.jsx("boxGeometry",{args:y,children:Qr.jsx("instancedBufferAttribute",{attach:"attributes-color",args:[c,3]})})}case"octahedron":return Qr.jsx("octahedronGeometry",{args:[t.radius||.02,0],children:Qr.jsx("instancedBufferAttribute",{attach:"attributes-color",args:[c,3]})});default:throw new Error(`Unsupported point shape: ${u}`)}},[u,t]),m=P.useMemo(()=>{switch(o){case"meshMatcap":return Qr.jsx("meshMatcapMaterial",{vertexColors:!0,opacity:l,transparent:l<1});default:return Qr.jsx("meshStandardMaterial",{vertexColors:!0,opacity:l,transparent:l<1})}},[o]);return Qr.jsxs("instancedMesh",{args:[void 0,void 0,e.length],ref:d,onPointerUp:y=>{if(y.intersections){const v=y.intersections.map(S=>S==null?void 0:S.instanceId).filter(S=>S!=null);n&&n(v.map(S=>e[S])),v.length>0&&r&&r(e[v[0]])}},onPointerOver:y=>{if(y.intersections){const v=y.intersections.map(S=>S==null?void 0:S.instanceId).filter(S=>S!=null);v.length>0&&i&&i(e[v[0]])}},onPointerLeave:()=>{a&&a()},children:[h,m]})}const Ynt=.1;function Xlt({data:e,pointRadius:t=Ynt,color:n="#999999",opacity:r=.2,wireframe:i=!1}){const a=P.useMemo(()=>{const u=[],l=new Set;return e.forEach(s=>{const{position:c}=s;if(!l.has(c.join(","))){const d=new LOe(t,12,12);d.translate(c[0],c[1],c[2]||0),u.push(d),l.add(c.join(","))}}),Cnt(u)},[e]);return Qr.jsx("mesh",{geometry:a??void 0,visible:r>0,children:Qr.jsx("meshBasicMaterial",{opacity:r,transparent:r<1,color:n,wireframe:i})})}function Gnt(){const e=new IOe;return e.renderOrder=1,e.position.z=-.2,e.depthTest=!1,e.scale.setScalar(1),e}function Ylt({children:e,lineColor:t="#53d7fb",onChange:n,points:r,enabled:i}){const{camera:a,raycaster:u,gl:l,controls:o,size:s,scene:c}=Vd(),d=P.useRef(Gnt()),h=P.useRef({selectionPoints:[],isDragging:!1,selectionNeedsUpdate:!1,selectionShapeNeedsUpdate:!1});return P.useEffect(()=>{d.current.material.color.set(t).convertSRGBToLinear(),a.add(d.current),c.add(a)},[]),P.useEffect(()=>{let m=-1/0,y=-1/0;const v=new ao,S=new ao,_=new ao;function E(A){if(i){const w=l.domElement.getClientRects()[0];m=A.clientX-w.left,y=A.clientY-w.top,h.current.selectionPoints.length=0,h.current.isDragging=!0}}function $(A){if(h.current.isDragging){const{selectionPoints:w}=h.current,R=l.domElement.getClientRects()[0],F=A.clientX-R.left,I=A.clientY-R.top,U=F/R.width*2-1,q=-(I/R.height*2-1);if(Math.abs(F-m)>=3||Math.abs(I-y)>=3){const N=(w.length/3-1)*3;let j=!1;w.length>3&&(v.set(w[N-3],w[N-3+1]),S.set(w[N],w[N+1]),S.sub(v).normalize(),v.set(w[N],w[N+1]),_.set(U,q),_.sub(v).normalize(),j=S.dot(_)>.99),j?(w[N]=U,w[N+1]=q):w.push(U,q,0),h.current.selectionShapeNeedsUpdate=!0,d.current.visible=!0,m=F,y=I}}}function k(){i&&h.current.isDragging&&h.current.selectionPoints.length&&(h.current.selectionShapeNeedsUpdate=!0,h.current.selectionNeedsUpdate=!0),d.current.visible=!1,h.current.isDragging=!1}return l.domElement.addEventListener("pointerdown",E,{passive:!0}),document.addEventListener("pointermove",$,{passive:!0,capture:!0}),document.addEventListener("pointerup",k,{passive:!0}),()=>{l.domElement.removeEventListener("pointerdown",E),document.removeEventListener("pointermove",$),document.removeEventListener("pointerup",k)}},[s,u,a,o,l,i]),x5(({camera:m})=>{const{selectionShapeNeedsUpdate:y,selectionPoints:v}=h.current;if(y){const S=v.length;v.push(v[0],v[1],v[2]),d.current.geometry.setAttribute("position",new e9(v,3,!1)),v.length=S,d.current.frustumCulled=!1,h.current.selectionShapeNeedsUpdate=!1}if(h.current.selectionNeedsUpdate&&v.length>0&&(h.current.selectionNeedsUpdate=!1,n(Knt({points:r,camera:m,selectionPoints:h.current.selectionPoints}))),nOe(m))d.current.scale.set((m.right-m.left)/m.zoom/2,(m.top-m.bottom)/m.zoom/2,1);else{const S=Math.tan(mJ.DEG2RAD*m.fov/2)*d.current.position.z;d.current.scale.set(-S*m.aspect,-S,1)}}),Qr.jsx(Qr.Fragment,{children:e})}function Knt({points:e,camera:t,selectionPoints:n}){const r=[],i=[];for(let u=0;u<n.length;u+=3)i.push([n[u],n[u+1]]);const a=new ci;return e.forEach(u=>{const l=u.position.length===3,o=l?u.position:[u.position[0],u.position[1],0];a.fromArray(o),l||a.sub(t.position).applyMatrix4(t.matrixWorld),a.project(t),Vnt([a.x,a.y],i)&&r.push(u)}),r}const dJ={Orange:"#ff7847",LightBlue:"#23abd8"},Wnt={LightBlueOrange:[dJ.LightBlue,dJ.Orange]},Glt={Discrete2:Wnt};export{cat as $,Tat as A,kat as B,DAe as C,Pat as D,Cat as E,wat as F,Fat as G,fit as H,Uf as I,lrt as J,srt as K,w3 as L,ort as M,urt as N,art as O,Dat as P,Iit as Q,be as R,Lit as S,cit as T,Bit as U,trt as V,Wwe as W,oit as X,qwe as Y,dit as Z,Ywe as _,O9 as a,uut as a$,Dit as a0,kit as a1,Gt as a2,Qwe as a3,rat as a4,jit as a5,qit as a6,Iat as a7,tat as a8,Jit as a9,_s as aA,Yit as aB,Git as aC,Kit as aD,Wit as aE,Zit as aF,uat as aG,aat as aH,fat as aI,rit as aJ,lit as aK,uit as aL,sit as aM,sat as aN,vit as aO,pit as aP,Xit as aQ,$it as aR,wit as aS,Rit as aT,Uat as aU,Vat as aV,qat as aW,iat as aX,Hit as aY,aut as aZ,lut as a_,eat as aa,Eit as ab,zit as ac,E5e as ad,Nat as ae,Tit as af,Ait as ag,Pit as ah,Mit as ai,hS as aj,zat as ak,Fit as al,Nit as am,Bat as an,hit as ao,mEe as ap,Vit as aq,Uit as ar,oat as as,lat as at,xit as au,_it as av,Sit as aw,Oit as ax,rw as ay,xs as az,crt as b,OU as b$,out as b0,zI as b1,oAe as b2,nat as b3,eut as b4,Jat as b5,Wat as b6,Kat as b7,Xat as b8,Elt as b9,Xlt as bA,zlt as bB,x5 as bC,Blt as bD,sut as bE,Hlt as bF,Ilt as bG,qlt as bH,Qlt as bI,Ult as bJ,jlt as bK,Ylt as bL,Zlt as bM,Vlt as bN,RJ as bO,frt as bP,hrt as bQ,Ort as bR,brt as bS,Urt as bT,prt as bU,jrt as bV,Krt as bW,Zrt as bX,Nt as bY,_rt as bZ,Grt as b_,lAe as ba,mat as bb,mit as bc,yit as bd,bit as be,Tne as bf,ait as bg,Cit as bh,rut as bi,cut as bj,BZe as bk,dHe as bl,cHe as bm,Clt as bn,wlt as bo,PN as bp,iit as bq,eit as br,WTe as bs,Lat as bt,mw as bu,xg as bv,bI as bw,Glt as bx,tut as by,iut as bz,drt as c,rp as c$,Oat as c0,gat as c1,Qat as c2,Fut as c3,Nut as c4,iqe as c5,nut as c6,tit as c7,pat as c8,nit as c9,Yat as cA,xRe as cB,AAe as cC,Rat as cD,Hat as cE,OS as cF,Flt as cG,Frt as cH,Nrt as cI,Brt as cJ,i1 as cK,VD as cL,Ea as cM,ep as cN,Lo as cO,krt as cP,dxe as cQ,sg as cR,xxe as cS,qrt as cT,fut as cU,Qrt as cV,sf as cW,du as cX,p7e as cY,vut as cZ,gut as c_,Gat as ca,hat as cb,Dlt as cc,Aat as cd,vat as ce,_at as cf,Hrt as cg,Xrt as ch,Yrt as ci,Wrt as cj,Jrt as ck,$lt as cl,CZ as cm,Alt as cn,Rlt as co,klt as cp,$1e as cq,Tlt as cr,Mlt as cs,O1e as ct,x1e as cu,Plt as cv,dC as cw,yat as cx,dat as cy,bat as cz,yrt as d,glt as d$,Tl as d0,xM as d1,hut as d2,dg as d3,Pr as d4,mut as d5,dut as d6,Qr as d7,ry as d8,yut as d9,hlt as dA,alt as dB,NY as dC,olt as dD,llt as dE,Xx as dF,ult as dG,OVe as dH,flt as dI,But as dJ,mqe as dK,rlt as dL,ilt as dM,nlt as dN,tlt as dO,elt as dP,Jut as dQ,Sve as dR,Wut as dS,zut as dT,Qut as dU,Yut as dV,Kut as dW,Gut as dX,mlt as dY,ylt as dZ,vlt as d_,cp as da,jat as db,Eut as dc,Dut as dd,$ut as de,xut as df,Gf as dg,Out as dh,Cut as di,Trt as dj,wut as dk,Iut as dl,Lut as dm,Xut as dn,Hut as dp,Zut as dq,jut as dr,Uut as ds,Vut as dt,qut as du,gqe as dv,slt as dw,_2e as dx,Kqe as dy,clt as dz,mrt as e,blt as e0,Olt as e1,xlt as e2,Slt as e3,Q3 as e4,plt as e5,dlt as e6,sVe as e7,_lt as e8,_ut as e9,Eat as ea,Sat as eb,xat as ec,Zat as ed,PRe as ee,irt as ef,Srt as f,Epe as g,gAe as h,rrt as i,nrt as j,$at as k,xrt as l,Wi as m,GSe as n,grt as o,iwe as p,Qit as q,P as r,oN as s,st as t,vrt as u,git as v,Mat as w,PAe as x,kZ as y,_Ae as z};
